# buildJar.tcl: build a Jar file of Diva code
#
# $Id: buildJar.tcl,v 1.2 2000/05/02 00:01:27 johnr Exp $
#
# Copyright (c) 1998-2000 The Regents of the University of California.
# All rights reserved. See the file COPYRIGHT for details.
#


### Assumed global variables:
# debug: print debug information if on
# silent: don't print anything if on
# DIVA: the root of the Diva tree


#######################################################################
#### buildJar
#
# Build the diva Jar file, given a list of package names to be included.
# This works properly only if run on a clean build.
#
proc buildJar {DIVA args} {
    global debug silent

    # Allow default package set
    if { $args == "" } {
	set args  [eval divaPackages $DIVA \
		PACKAGES \
		UNRELEASED_PACKAGES \
		AUXILIARY_PACKAGES]
    }
    if !$silent {
	puts "Building Jar file for $args"
    }

    # Collect file names
    set classes {}
    cd [file dirname $DIVA]
    foreach p $args {
	set dir [split $p .]
	set cs [glob -nocomplain [eval file join $dir *.class]]
	if !$silent {
	    if { $cs == "" } {
		puts "$p -- no class files"
	    } else {
		puts "$p -- ok"
	    }
	}
	set classes [concat $classes $cs]
    }

    # The command to execute
    set javahome [getMakevar JAVAHOME]
    set cmd [concat [file join $javahome bin jar] \
	cf \
	[file join $DIVA diva.jar] \
	$classes]

    if !$silent {
	# Don't put the whole command, there are too many classes...
	puts "[lrange $cmd 0 6]..."
    }

    catch {eval exec $cmd} result
    if !$silent {
	puts $result
	puts Done!
    }
}







