# buildSnapshot.tcl: build a snapshot release of Diva
#
# $Id: buildSnapshot.tcl,v 1.25 2000/05/02 00:01:27 johnr Exp $
#
# Copyright (c) 1998-2000 The Regents of the University of California.
# All rights reserved. See the file COPYRIGHT for details.
#


### Assumed global variables:
# debug: print debug information if on
# silent: don't print anything if on

# The startup message
set snapshotMessage {
    ### Building Diva snapshot.                 ###
    ###                                         ###
    ### Have you run the build and smoke test?  ###
}


#######################################################################
#### buildSnapshot
#
# Build a snapshot release of Diva. This is a release that contains
# packages in the following variables (in vars.mk):
#
#   PACKAGES
#   AUXILIARY_PACKAGES
#   UNRELEASED_PACKAGES
#
# The release is built in the directory Builds/Snapshot relative
# to the Diva root of the current directory.
#
proc buildSnapshot {DIVA} {
    global silent debug snapshotMessage

    if !$silent {
	puts $snapshotMessage
    }

    #
    # Export the repository into Builds/Snapshot
    #
    set dir [file join $DIVA Builds Snapshot]
    set diva [file join $dir diva]
    exportDiva $DIVA $dir
    divaConfigure $DIVA $diva

    #
    # The packages we want to build/release/document/etc.
    #
    set packages [eval divaPackages $diva \
	    PACKAGES \
	    UNRELEASED_PACKAGES \
	    AUXILIARY_PACKAGES]

    #
    # Build it
    #
    set time [clock seconds]
    set date [clock format $time -format "%d%b%y"]
    _buildCommon $diva $date $packages
}








