# documentMerge.tcl: Merge a document into a single printable document
# $Id: documentMerge.tcl,v 1.3 2000/05/02 00:01:28 johnr Exp $
#
# Author: John Reekie
#
# Copyright (c) 1998-2000 The Regents of the University of California.
# All rights reserved. See the file COPYRIGHT for details.
#


# Make sure we're initialized
document::init


##########################################################################
#### documentMerge
#
# Procedure to merge a document generated from a TIM skeleton.
#
proc documentMerge {DIVA {file {}} args} {
    # Read the document file
    if [string match -* $file] {
        set args [lindex $args $file]
        set file {}
    }
    if { $file == "" } {
	set file [file join [pwd] _document.tim]
    } else {
	# set file [::document::expandPath $file]
	# FIXME
	set file $file
    }
    
    # Create the document model
    set model document[incr document::counter]
    eval document::create $model -modelpath $file $args
    set model document::$model
    set fd [open $file]
    $model parse [read $fd]
    close $fd
    $model merge
    $model destroy
}

########################################################################
#### merge
#
# Generate the merged HTML file. The files will be generated in the
# same directory as this model resides in, and named "printable.html".
#
proc document::merge {ns} {
    # Make sure options are set, using pwd if necessary
    if { [$ns cget -modelpath] == "" } {
        $ns configure -modelpath [file join [pwd] _document.tim]
    }
    if { [$ns cget -directory] == "" } {
        $ns configure -directory [file dirname [$ns cget -modelpath]]
    }
    puts [$ns cget -directory]:

    # Make a backup of the output if it already exists
    set directory [$ns cget -directory]
    set filename [file join $directory printable.html]
    if [file exists $filename] {
        if ![file exists [file join $directory backup]] {
            file mkdir [file join $directory backup]
        }
        file copy -force $filename [file join $directory backup printable.html]
    }

    # Call recursive helper method
    set text [_merge $ns "index"]

    # Write the output file
    puts "Writing $filename"
    set fd [open $filename "w"]
    puts $fd [string trim $text "\n"]
    close $fd
}

########################################################################
#### _merge
#
# Helper method for <b>merge</b>. This method generates a single file.
#
proc document::_merge {ns section} {
    set directory [$ns cget -directory]
    set suffix [$ns cget -suffix]

    # Filename
    set filename [file join $directory \
            [lindex $section end]$suffix]
    puts "  [lindex $section end]$suffix"

    # Get the data of the file
    set fd [open $filename]
    set ${ns}::_data [read $fd]
    close $fd

    # If it's a regular file, extract the body and strip out stuff
    if { $section != "index" } {
	set text [extractTag ${ns}::_data "body"]
	substComment text "contents" ""
	substComment text "customfooter" ""
	substComment text "footer" ""
	substComment text "navigation" ""
	substComment text "navbar" ""
    } else {
	# If it's an index file, strip out contents
	set text [set ${ns}::_data]
	substComment text "contents" ""
   }

    # Now do all the child sections
    if { $section == "index" } {
        set context ""
    } else {
        set context $section
    }
    set moretext ""
    if { [$ns match section $context] != "" } {
	foreach sec [$ns match section $context] {
	    # Build this level
	    set subsection [concat $context $sec]
	    append moretext \n\n[_merge $ns $subsection]
	}
    }
    set ${ns}::_data ""

    # If it's an index file, insert the text before /body
    # FIXME: Needs to be before the customfooter
    if { $section == "index" } {
	insertBefore text "/body" \n$moretext
    } else {
	# Otherwise append to text
	set text $text\n\n$moretext
    }
    return $text
}







