# buildSnapshot.tcl: build a snapshot release of Diva
#
# $Id: publish.tcl,v 1.3 2000/05/02 00:01:29 johnr Exp $
#
# Copyright (c) 1998-2000 The Regents of the University of California.
# All rights reserved. See the file COPYRIGHT for details.
#


### Assumed global variables:
# debug: print debug information if on
# silent: don't print anything if on

# The startup message
set snapshotMessage {
    ### Publishing Diva snapshot                ###
    ###                                         ###
    ### Are you sure?                           ###
}

# The release message
set releaseMessage {
    ### Publishing Diva relase                  ###
    ###                                         ###
    ### Are you sure?                           ###
}


#######################################################################
#### publishSnapshot
#
# Copy snapshot files and documentation (generated by
# buildSnapshot) to the web page, from the directory 
# Builds/Snapshot relative to the Diva root of the current 
# directory.  The destination argument is the location
# of the release.
#
proc publishSnapshot {DIVA {destination ""}} {
    global silent debug snapshotMessage

    if !$silent {
	puts $snapshotMessage
    }

    set source [file join $DIVA Builds Snapshot]
    set date [clock format [clock seconds] -format "%d%b%y"]
    if {$destination == ""} {
	set destination "/home/www/releases/diva/snapshots/";
    }
    _publishCommon $DIVA $source $date $destination
}

#######################################################################
#### publishRelease
#
# Copy release files and documentation (generated by
# buildRelease) to the web page, from the directory 
# Builds/Release relative to the Diva root of the current 
# directory.  The destination argument is the location
# of the release.
#
proc publishRelease {DIVA version {destination ""} } {
    global silent debug releaseMessage

    if !$silent {
	puts $releaseMessage
    }

    set source [file join $DIVA Builds Release]
    if {$destination == ""} {
	set destination "/home/www/releases/diva";
    }
    _publishCommon $DIVA $source $version $destination
}


#######################################################################
#### publishCommon
#
# This is the common function for publishing snapshot/release
# documentation on the web.  
#
# It first moves the $destination
# directory to $destination-OLD and then recursively copies the
# directory $source/diva-$suffix-doc to $destination. 
#
# For example:
#
# $DIVA/Builds/Snapshot/diva-DATE-doc ==> /www/diva/snapshot
#
# It then copies the distribution files:
#   diva-$suffix.tar.gz
#   diva-$suffix.zip
#   diva-$suffix.jar
#   diva-$suffix.jar.zip
# 
# to the destination directory.
#
# For example:
#
# $DIVA/Builds/Snapshot/diva-DATE.tar.gz ==> /www/diva/snapshot
# etc.
#
proc _publishCommon {DIVA source suffix destination} {
    puts "#### Publishing: $source $suffix $destination"
    if [file exists "$destination/diva-$suffix"] {
	file delete -force "$destination/diva-$suffix"
    }
    file copy -force "$source/diva-$suffix" $destination
    file delete -force "$destination/latest"
    exec ln -s $destination/diva-$suffix $destination/latest
}







