package diva.canvas.interactor;

import java.awt.event.MouseEvent;

import diva.canvas.event.LayerEvent;
import diva.canvas.event.LayerEventMulticaster;
import diva.canvas.event.LayerListener;

public class RightClickInteractor extends CompositeInteractor {

	/** Layer listeners
	 */
	private transient LayerListener _layerListener;

	/** Add the given layer listener to this interactor.  Any event that is
	 * received by this interactor will be passed on to the listener after
	 * it is handled by this interactor.
	 */
	public void addLayerListener(LayerListener l) {
		System.out.println("RightClickInteractor.addLayerListener(l)");
		_layerListener = LayerEventMulticaster.add(_layerListener,l);
	}

	/** Fire a layer event.
	 */
	public void fireLayerEvent (LayerEvent event) {
		System.out.println("RightClickInteractor.fireLayerEvent(e)");
		if (_layerListener != null) {
			int id = event.getID();
			switch(id) {
			case MouseEvent.MOUSE_PRESSED:
				_layerListener.mousePressed(event);
				break;
			case MouseEvent.MOUSE_DRAGGED:
				_layerListener.mouseDragged(event);
				break;
			case MouseEvent.MOUSE_RELEASED:
				_layerListener.mouseReleased(event);
				break;
			}
		}
	}
	
	/** Handle a mouse press on a figure or layer. Set the target
	 * to be the figure contained in the event. Call the setup()
	 * method in case there is additional setup to do, then
	 * constrain the point and remember it. 
	 * Nothing happens if the interactor is not enabled, or if it
	 * is "selective enabled" but not in the selection.
	 */
	public void mousePressed (LayerEvent e) {
		//if (!isEnabled()
		//		|| (_selectiveEnabled && !SelectionInteractor.isSelected(e))) {
		//	return;
		//}
		System.out.println("RightClickInteractor.mousePressed(e)");
		if(getMouseFilter() == null || getMouseFilter().accept(e)) {      
            
            
			// Inform listeners
			fireLayerEvent(e);
            
			// Consume the event
			if (isConsuming()) {
				e.consume();
			}
		}
	}

	/** Handle a mouse released event.
	 * Nothing happens if the interactor is not enabled, if if it
	 * is "selective enabled" but not in the selection.
	 */
	public void mouseReleased (LayerEvent e) {
		//if (!isEnabled()
		//		|| (_selectiveEnabled && !SelectionInteractor.isSelected(e))) {
		//	return;
		//}
		System.out.println("RightClickInteractor.mouseReleased(e)");
		if(getMouseFilter() == null || getMouseFilter().accept(e)) {
			openStats();
			
			fireLayerEvent(e);
   
			// Consume the event
			if (isConsuming()) {
				e.consume();
			}
		}
	}	
	
	public void openStats(){
		System.out.println("RightClickInteractor.openStats()");
	}
}