/*
 * $Id: BasicEdgeRenderer.java,v 1.4 2004/04/27 03:54:08 mnwaltz Exp $
 *
 * Copyright (c) 1998-2000 The Regents of the University of California.
 * All rights reserved. See the file COPYRIGHT for details.
 */
package diva.graph.basic;
import java.awt.Color;
import java.awt.Paint;

import diva.canvas.Figure;
import diva.canvas.Site;
import diva.canvas.connector.AbstractConnector;
import diva.canvas.connector.ArcConnector;
import diva.canvas.connector.Arrowhead;
import diva.canvas.connector.Connector;
import diva.canvas.connector.StraightConnector;
import diva.graph.EdgeRenderer;
import diva.graph.modular.Edge;

/**
 * A basic implementation of the EdgeRenderer interface.
 * This renderer creates straight-line connectors with
 * an arrow at the head.
 *
 * @author Michael Shilman  (michaels@eecs.berkeley.edu)
 * @author Revised by Maylene Waltz to accomodate LabelFigures
 * @version $Revision: 1.4 $
 * @rating Red
 */
public class BasicEdgeRenderer implements EdgeRenderer {    
	
	/** The color of the edge's label. Default is BLACK.
	 */
	private Paint _labelTextColor = Color.black;
	
	/** The text size of the label. Default is 12.
	 */
	private int _labelTextSize = 12;
	
	//ADDED by Maylene Waltz
	public Paint getLabelTextColor(){
		return _labelTextColor;
	}
	
	//ADDED by Maylene Waltz - WPI
	public int getLabelTextSize(){
		return _labelTextSize;
	}
	
	/**
	 * Render a visual representation of the given edge.
	 */
	public Connector render(Object edge, Site tailSite, Site headSite) {
		AbstractConnector c;
		Figure tf = tailSite.getFigure();
		Figure hf = headSite.getFigure();
		Edge edgePeer = (Edge)edge;
		
		//if the edge is a self loop, create an ArcConnector instead!
		if((tf != null)&&(hf != null)&&(tf == hf)){
			c = new ArcConnector(tailSite, headSite);
		}
		else {
			c = new StraightConnector(tailSite, headSite);
			//ADDED by Maylene Waltz - WPI
			if(edgePeer.getLabelFigure() != null){
				c.setLabelFigure(edgePeer.getLabelFigure());
				if(c.getLabelFigure() != null){
					c.getLabelFigure().setFillPaint(_labelTextColor);
					c.getLabelFigure().setSize(_labelTextSize);
				}
			}
		}
		Arrowhead arrow = new Arrowhead(
				headSite.getX(), headSite.getY(), headSite.getNormal());
		c.setHeadEnd(arrow);
		
		//ADDED by Maylene Waltz - WPI
		c.setToolTipText(edgePeer.getToolTipText());
		
		return c;
	}
	
	//ADDED by Maylene Waltz
	public void setLabelTextColor(Paint p){
		_labelTextColor = p;
	}
	
	//ADDED by Maylene Waltz - WPI
	public void setLabelTextSize(int size){
		_labelTextSize = size;
	}
}


