/*
 * $Id: BasicGraphController.java,v 1.14 2004/04/27 03:54:08 mnwaltz Exp $
 *
 * Copyright (c) 1998-2000 The Regents of the University of California.
 * All rights reserved. See the file COPYRIGHT for details.
 */
package diva.graph.basic;

import java.awt.event.InputEvent;

import diva.canvas.event.LayerEvent;
import diva.canvas.event.MouseFilter;
import diva.canvas.interactor.AbstractInteractor;
import diva.canvas.interactor.SelectionDragger;
import diva.canvas.interactor.SelectionInteractor;
import diva.graph.EdgeController;
import diva.graph.EdgeCreator;
import diva.graph.GraphPane;
import diva.graph.NodeController;
import diva.graph.NodeInteractor;
import diva.graph.SimpleGraphController;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.listeners.StatsInteractor;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils.JTabbedPaneCAPE;

/**
 * A basic implementation of GraphController, which works with
 * simple graphs that have edges connecting simple nodes. It
 * sets up some simple interaction on its view's pane.
 *
 * @author 	Michael Shilman (michaels@eecs.berkeley.edu)
 * @version	$Revision: 1.14 $
 * @rating      Red
 */
public class BasicGraphController extends SimpleGraphController {
	
	/** The CAPE_GUI */
	private CAPE_GUI _gui;
	
    /**
     * The global count for the default node/edge creation.
     */
    private int _globalCount = 0;

    /** The selection interactor for drag-selecting nodes
     */
    //private SelectionDragger _selectionDragger;
    
	/** The interactor that opens a stats window
	 * when right-clicking a node.
	 * @author Maylene Waltz - WPI
	 */
	private StatsInteractor _statsInteractor;

   /** The interactor for creating new nodes
     */
    //private NodeCreator _nodeCreator;

    /** The interactor that interactively creates edges
     */
    //private EdgeCreator _edgeCreator;
    
    /** The tabbed pane for the statistic table(s) */
    private JTabbedPaneCAPE _statsTabbedPane;

    /** The filter for control operations
     */
    //private MouseFilter _controlFilter = new MouseFilter (
    //        InputEvent.BUTTON1_MASK,
    //        InputEvent.CTRL_MASK);
    
    /** The filter for right clicks
     * @author Maylene Waltz - WPI
     */     
	private MouseFilter _statsFilter = new MouseFilter(
			InputEvent.BUTTON2_MASK);

    /**
     * Create a new basic controller with default node and edge controllers.
     */
    public BasicGraphController() {
	NodeController nc = new NodeController(this);
	nc.setNodeRenderer(new BasicNodeRenderer(this));
	setNodeController(nc);
	
	EdgeController ec = new EdgeController(this);
	ec.setEdgeRenderer(new BasicEdgeRenderer());
	setEdgeController(ec);

        //addGraphViewListener(new IncrementalLayoutListener(new IncrLayoutAdapter(new LevelLayout(new BasicLayoutTarget(this))), null));
    }
    
	/**
	 * Create a new basic controller with default node and edge controllers.
	 */
	public BasicGraphController(CAPE_GUI gui) {
	NodeController nc = new NodeController(this);
	nc.setNodeRenderer(new BasicNodeRenderer(this));
	setNodeController(nc);
	
	EdgeController ec = new EdgeController(this);
	ec.setEdgeRenderer(new BasicEdgeRenderer());
	setEdgeController(ec);

	_gui = gui;
	//addGraphViewListener(new IncrementalLayoutListener(new IncrLayoutAdapter(new LevelLayout(new BasicLayoutTarget(this))), null));
	}

    /**
     * Initialize all interaction on the graph pane. This method
     * is called by the setGraphPane() method of the superclass.
     * This initialization cannot be done in the constructor because
     * the controller does not yet have a reference to its pane
     * at that time.
     */
    protected void initializeInteraction () {
        GraphPane pane = getGraphPane();
        
        // Create and set up the selection dragger
        /* COMMENTED OUT to disable dragging and node creation by CAPE, DSRG, at WPI
        _selectionDragger = new SelectionDragger(pane);
        _selectionDragger.addSelectionInteractor(
                (SelectionInteractor)getEdgeController().getEdgeInteractor());
		_selectionDragger.addSelectionInteractor(
                (SelectionInteractor)getNodeController().getNodeInteractor());
		
        // Create a listener that creates new nodes
        _nodeCreator = new NodeCreator();
        _nodeCreator.setMouseFilter(_controlFilter);
        pane.getBackgroundEventLayer().addInteractor(_nodeCreator);

        // Create the interactor that drags new edges.
		_edgeCreator = new EdgeCreator(this) {
	    	public Object createEdge() {	
				Object semanticObject = new Integer(_globalCount++);
				BasicGraphModel bgm = (BasicGraphModel)getGraphModel();
				return bgm.createEdge(semanticObject);
	   		}
		};
		_edgeCreator.setMouseFilter(_controlFilter);
		((NodeInteractor)getNodeController().getNodeInteractor()).addInteractor(_edgeCreator);
		*/
		//Create the interactor that opens statistics windows
		//when a node is right-clicked.
		//ADDED by Maylene Waltz - WPI
		_statsInteractor = new StatsInteractor(this);
		_statsInteractor.setMouseFilter(_statsFilter);
		pane.getForegroundEventLayer().addInteractor(_statsInteractor);
    }


	/** Returns the CAPE_GUI associated with this graph controller. */
	public CAPE_GUI getGUI(){
		return _gui;
	}
	
	/** Returns the statistics tabbed pane. 
	 * @return java.awt.Component */
	public java.awt.Component getStatsTabbedPane(){
		return _statsTabbedPane;
	}
	
	/** Sets the given <code>statsTabbedPane</code> for this controller. */
	public void setStatsTabbedPane(JTabbedPaneCAPE statsTabbedPane){
		_statsTabbedPane = statsTabbedPane;
	}
	
    ///////////////////////////////////////////////////////////////
    //// NodeCreator

    /** An inner class that places a node at the clicked-on point
     * on the screen, if control-clicked with mouse button 1. This
     * needs to be made more customizable.
     */
    protected class NodeCreator extends AbstractInteractor {
        public void mousePressed(LayerEvent e) {
            Object semanticObject = new Integer(_globalCount++);
			BasicGraphModel bgm = (BasicGraphModel)getGraphModel();
            Object node = bgm.createNode(semanticObject);
	    	addNode(node,  e.getLayerX(), e.getLayerY());
		}
    }
}


