/*
 * $Id: Node.java,v 1.4 2004/03/12 18:27:54 tims Exp $
 *
 * Copyright (c) 2000 The Regents of the University of California.
 * All rights reserved.  See the file COPYRIGHT for details.
 */
package diva.graph.modular;
import java.util.Iterator;

import diva.canvas.toolbox.LabelFigure;
import diva.util.PropertyContainer;
import diva.util.SemanticObjectContainer;

/**
 * A node is an object that is contained by a graph
 * and is connected to other nodes by edges.  A node
 * has a semantic object that is its semantic equivalent
 * in the application and may have a visual object which
 * is its syntactic representation in the user interface.
 * 
 * @author Michael Shilman  (michaels@eecs.berkeley.edu)
 * @version $Revision: 1.4 $
 * @rating Red
 */
public interface Node extends SemanticObjectContainer, PropertyContainer {
    /**
     * Return an iterator over the edges coming into this node.
     */
    public Iterator inEdges();

    /**
     * Return an iterator over the edges coming out of this node.
     */
    public Iterator outEdges();

    /**
     * Return the graph parent of this node.
     */
    public Graph getParent();

	/**
	 * Return the node's label.
	 * @author Maylene Waltz
	 */
	public LabelFigure getLabelFigure();
	
	/**
	 * Return the node's MachineID label.
	 * @author Maylene Waltz
	 */
	public LabelFigure getLabelFigureMachineID();
	
	/** 
	 * Return the node's <code>toolTipText</code>.
	 * @author Maylene Waltz
	 */
	public String getToolTipText();
	
   	/**
   	 * Set the node's label.
   	 * @author Maylene Waltz
   	 */
	public void setLabelFigure(LabelFigure label);
	
	public void setLabelFigureMachineID(LabelFigure label);
	
	/**
	 * Set the node's <code>toolTipText</code>.
	 * @author Maylene Waltz
	 */
	public void setToolTipText(String tip);
 
    /**
     * Set the graph parent of this node.  Implementors of this method
     * are also responsible for insuring that it is set properly as
     * the child of the graph in the graph.
     */
    public void setParent(Graph parent);
}

