package edu.wpi.cs.dsrg.utils;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import java.lang.reflect.Array;

import edu.wpi.cs.dsrg.xmldb.xat.common.dompattern.DOMPatternNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Expose;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.FOR;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.IfExpr;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Merge;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumn;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.SQLStmt;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Source;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.XQueryStmt;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.Function;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.CartesianProduct;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Distinct;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupBy;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Intersection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Join;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.LeftOuterJoin;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.OrderBy;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Project;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.RightOuterJoin;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Select;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Union;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.Delete;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.Insert;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.Rename;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.Replace;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Aggregate;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavCollection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavUnnest;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Tagger;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.XMLUnion;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.NavigationStep;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.InheritableVisitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * This class is used to print out the XAT.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */

public class Debugger extends InheritableVisitor {

	/**
	 * The XATTree debugging.
	 * @since 1.0
	 */
    protected XATTree tree;
	/**
	 * The XATTree depth.
	 * @since 1.0
	 */    
    private int level;

/**
 * The default Debugger constructor.
 * @since 1.0
 */
public Debugger(XATTree xattree) {
    tree = xattree;
    level = 0;
}
public Object dispatch() {
    //System.out.println(indent() + "NULL");
    return null;
}
public Object dispatch(Object obj) {
    System.out.println("Dispatch!! " + obj.getClass().toString());
    return null;
}
/**
 * Get the XML algebra tree.
 * @return XATTree
 * @since 1.0
 */
public XATTree getTree() {
    return tree;
}
/**
 * Print out the indent space.
 * @return String The indent string.
 * @since 1.0
 */
String indent() {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < level; i++)
        sb.append("  ");
    return sb.toString();

}
/**
 * Prints a node in the XAT
 * If the root node of an XAT is the input of this method,
 * the whole tree will be printed out.
 * @param node The node printing.
 * @since 1.0
 */
private void printNode(XATNode node) {
    XATNode[] children = node.getChildren();
    System.out.println(indent() + "Node id:" + node.hashCode());
    visit(node.getOperator());

    for (int i = 0; i < Array.getLength(children); i++) {
        if (children.length > 1) {
            System.out.println(indent() + "_____________");
            System.out.println(indent() + "Child " + i + ": ");
            printNode(children[i]);
            System.out.println(indent() + "_____________");
        } else
            printNode(children[i]);
    }
}
/**
 * Prints out the tagger tree.
 * @param node The Tagger node.
 * @since 1.0
 */
private void printTagger(DOMPatternNode node) {
    DOMPatternNode[] children = node.getChildren();
    for (int i = 0; i < Array.getLength(children); i++) {
        System.out.println(
            indent() + children[i].toString() + " parent: " + children[i].getParent());
        printTagger(children[i]);
    }

}
/**
 * Print the whole XATTree.
 * @since 1.0
 */
public void printTree() {
    System.out.println("\n\nXAT: with Nodes(" + tree.size() + ")\n");
    printNode(tree.getRoot());
    System.out.println(tree.getBindingTable().toString());
}
/**
 * Set the XML algebra tree.
 * @param xattree The XATTree set.
 * @since 1.0
 */
public void setTree(XATTree xattree) {
    tree = xattree;
}
/**
 * Output of the Expose node in an XAT.
 * @reutrn null
 * @param expose The Expose for output.
 * @since 1.0
 */
 
public Object visit(Expose expose) {
    System.out.println(indent() + "Expose: " + expose.getColumnName());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the FOR node in an XAT.
 * @reutrn null
 * @param forNode The FOR for output.
 * @since 1.0
 */

public Object visit(FOR forNode) {
    String[] binding = forNode.getColumnNames();
    System.out.println(
        indent() + "For Node: \n" + indent() + "binding: " + binding[0]);
    System.out.println(indent() + "\toutput col: " + forNode.getOutputColumnName());
    level += 3;
    System.out.println(indent() + "FOR tree: ");
    printNode(forNode.getXtreFunctions().getRoot());
    System.out.println(indent() + "End for tree");
    level -= 3;
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the Function node in an XAT.
 * @reutrn null
 * @param function The Function for output.
 * @since 1.0
 */

public Object visit(Function function) {
    System.out.println(indent() + "Function node:\n" + indent() + "data: ");
    System.out.println(
        indent() + "  output col: " + function.getOutputColumnName());
    System.out.println(indent() + "  function type: " + function.getName() + "()");
    System.out.println(indent() + "  params:");
    Object[] params = function.getParams();
    if (params != null) {
        for (int i = 0; i < params.length; i++) {
            if (params[i] != null) {
                if (params[i] instanceof NavigationStep) {
                    NavigationStep navStep = (NavigationStep) params[i];
                    System.out.println(indent() + "\t" + navStep.getStepValue());
                } else
                    System.out.println(indent() + "\t" + params[i].toString());
            }
        }
    }
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the IfExpr node in an XAT.
 * @reutrn null
 * @param IfExpr The IfExpr for output.
 * @since 1.0
 */

public Object visit(IfExpr ifExpr) {
    System.out.println(indent() + "IfExpr node");
    System.out.println(indent() + "\toutput col: " + ifExpr.getOutputColumnName());
    System.out.println(indent() + "\tCondition: " + ifExpr.getIfExpr());
    System.out.println(indent() + "--------------------");
    System.out.println(indent() + "Then Subtree: ");
    System.out.println(indent() + "_____________ ");
    level++;
    printNode(ifExpr.getThenXAT().getRoot());
    level--;
    System.out.println(indent() + "--------------------");
    System.out.println(indent() + "--------------------");
    System.out.println(indent() + "Else Subtree: ");
    System.out.println(indent() + "_____________ ");
    level++;
    printNode(ifExpr.getElseXAT().getRoot());
    level--;
    System.out.println(indent() + "--------------------");
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the Merge node in an XAT.
 * @reutrn null
 * @param merge The Merge for output.
 * @since 1.0
 */
 
public Object visit(Merge merge) {
    /**
    * Print out the operator.
    */
    System.out.println(indent() + "Merge");
    System.out.println(indent() + "Output Column:" + merge.getOutputColumnName());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the NameColumn node in an XAT.
 * @reutrn null
 * @param nameColumn The NameColumn for output.
 * @since 1.0
 */
 
public Object visit(NameColumn nameColumn) {
    StringBuffer sb = new StringBuffer();
    try {
        sb.append(indent() + "NameColumnNode: \n");
        sb.append(
            indent() + " originalName: " + nameColumn.getOriginalColumnName() + "\n");
        sb.append(indent() + " newName: " + nameColumn.getNewColumnName());
        System.out.println(sb.toString());
        System.out.println(indent() + "\t|");
    } catch (NullPointerException e) {
        System.out.println(sb.toString());
        System.out.println(indent() + "\t|");
    }

    return null;

}
/**
 * Output of the Source node in an XAT.
 * @reutrn null
 * @param source The Source for output.
 * @since 1.0
 */
 
public Object visit(Source source) {
    System.out.println(indent() + "___________________");
    System.out.println(indent() + "   Source node     ");
    System.out.println(indent() + " Data: " + source.getName());
    System.out.println(indent() + "\toutput col: " + source.getOutputColumnName());
    System.out.println(indent() + "===================\n");
    return null;
}
/**
 * Output of the SQLStmt node in an XAT.
 * @reutrn null
 * @param sqlS The SQLStmt for output.
 * @since 1.0
 */
 
public Object visit(SQLStmt sqlS) {
    System.out.println(indent() + "SQL Node: \n" + indent() + "  SQLStatement: ");
    System.out.println(indent() + sqlS.getStmt());

    return null;
}
/**
 * Output of the XQueryStmt node in an XAT.
 * @reutrn null
 * @param xqueryS The XQueryStmt for output.
 * @since 1.0
 */
 
public Object visit(XQueryStmt xqueryS) {
    System.out.println(
        indent() + "XQuery Node: \n" + indent() + "  XQueryStatement: ");
    System.out.println(indent() + xqueryS.getStmt());

    return null;
}
/**
 * Output of the CartesianProduct node in an XAT.
 * @reutrn null
 * @param cartProd The CartesianProduct for output.
 * @since 1.0
 */

public Object visit(CartesianProduct cartProd) {
	System.out.println(indent() + "Cartesian Product node\n" + indent() + "   children: ");
    System.out.println(indent() + "\t|");
	return null;
}
/**
 * Output of the Distinct node in an XAT.
 * @reutrn null
 * @param distinct The Distinct for output.
 * @since 1.0
 */
 
public Object visit(Distinct distinct) {
	System.out.println(indent() + "Distinct node\n" + indent() + "   data: ");
	System.out.println(indent() + "\tinputColumn: " + distinct.getInputColumnName());
	System.out.println(indent() + "\t|");
	return null;
}
/**
 * Output of the GroupBy node in an XAT.
 * @reutrn null
 * @param groupBy The GroupBy for output.
 * @since 1.0
 */

public Object visit(GroupBy groupBy) {
    String[] cols = groupBy.getColumnNames();

    System.out.println(indent() + "Group By node:");
    System.out.print(indent() + "\tcolumns for grouping:");
    for (int i = 0; i < +cols.length; i++)
        System.out.print(" " + cols[i]);
    System.out.println(
        "\n" + indent() + "\toutput col: " + groupBy.getOutputColumnName());
    try {
        System.out.println(indent() + "XATTree associated with group by: ");
        level++;
        printNode(groupBy.getXATTree().getRoot());

        //System.out.println(indent() + "\tfunction node: " + function.getName());
        System.out.println(indent() + "===========");
        level--;
    } catch (Exception e) {
        System.out.println(indent() + "no function node associated");
    }
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the Intersection node in an XAT.
 * @reutrn null
 * @param intersection The Intersection for output.
 * @since 1.0
 */
 
public Object visit(Intersection intersect) {
    System.out.println(indent() + "Intersect node: \n" + indent() + "  children: ");
    System.out.println(
        indent() + "\toutput col: " + intersect.getOutputColumnName());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the Join node in an XAT.
 * @reutrn null
 * @param join The Join for output.
 * @since 1.0
 */
 
public Object visit(Join join) {
    System.out.println(indent() + "Join node: \n" + indent() + "  children: ");
    System.out.println(indent() + "\toutput col: " + join.getOutputColumnName());
    System.out.println(indent() + "\tcondition: " + join.getExpression());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the LeftOuterJoin node in an XAT.
 * @reutrn null
 * @param join The LeftOuterJoin for output.
 * @since 1.0
 */
public Object visit(LeftOuterJoin join) {
    System.out.println(
        indent() + "Left Outer Join node: \n" + indent() + "  children: ");
    System.out.println(indent() + "\toutput col: " + join.getOutputColumnName());
    System.out.println(indent() + "\tcondition: " + join.getExpression());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the OrderBy  node in an XAT.
 * @reutrn null
 * @param order The OrderBy  for output.
 * @since 1.0
 */
 
public Object visit(OrderBy order) {
    String[] cols = order.getColumnNames();
    System.out.println(indent() + "OrderBy Node: ");
    System.out.println(indent() + "\tdata: ");
    for (int i = 0; i < cols.length; i++) {
        System.out.println(indent() + "\t\t" + cols[i]);
    }
    System.out.println(indent() + "\toutput col: " + order.getOutputColumnName());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the Project node in an XAT.
 * @reutrn null
 * @param project The Project for output.
 * @since 1.0
 */
 
public Object visit(Project project) {
    StringBuffer sb = new StringBuffer();
    sb.append("Project: \n  Column names: \n");
    String[] colList = project.getColumnNames();
    for (int i = 0; i < colList.length; i++) {
        sb.append("Column" + i + ":" + colList[i] + "\n");
    }

    System.out.print(indent() + sb.toString());
    System.out.println(indent() + "\t|");
    return null;

}
/**
 * Output of the RightOuterJoin node in an XAT.
 * @reutrn null
 * @param join The RightOuterJoin for output.
 * @since 1.0
 */
 
public Object visit(RightOuterJoin join) {
    System.out.println(
        indent() + "Right Outer Join node: \n" + indent() + "  children: ");
    System.out.println(indent() + "\toutput col: " + join.getOutputColumnName());
    System.out.println(indent() + "\tcondition: " + join.getExpression());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the Select node in an XAT.
 * @reutrn null
 * @param select The Select for output.
 * @since 1.0
 */
 
public Object visit(Select select) {
    System.out.println(indent() + "Select node:\n" + indent() + "  data: ");
    /*
    System.out.println(indent() + "Left side: " + ((BinCOMPExpression)select.getExpression()).getLeft());
    System.out.println(indent() + "Right side: " + ((BinCOMPExpression)select.getExpression()).getRight());
    */
    System.out.println(indent() + "Expression: " + select.getExpression());
    //System.out.println(indent() + "\toutput col: " + select.getOutputColumnName());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the Union node in an XAT.
 * @reutrn null
 * @param union The Union for output.
 * @since 1.0
 */
 
public Object visit(Union union) {
    System.out.println(indent() + "Union node: \n" + indent() + "  children: ");
    System.out.println(indent() + "\toutput col: " + union.getOutputColumnName());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the Delete node in an XAT.
 * @reutrn null
 * @param delete The Delete for output.
 * @since 1.0
 */

public Object visit(Delete delete) {
    System.out.println(indent() + "Delete node:");
    System.out.println(indent() + "\tdata: ");
    System.out.println(indent() + "\tContext: " + delete.getContext());
    System.out.println(indent() + "\tVariable: " + delete.getVariable());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the Insert node in an XAT.
 * @reutrn null
 * @param insert The Insert for output.
 * @since 1.0
 */

public Object visit(Insert insert) {
    System.out.println(indent() + "Insert node:");
    System.out.println(indent() + "\tdata: ");
    System.out.println(indent() + "\tContext: " + insert.getContext());
    // if there is a variable in the insert node	
    if (insert.getVariable() != null)
        System.out.println(indent() + "\tVariable: " + insert.getVariable());

    // if there appears an update function in the insert node     
    if (insert.getUpdateFunction() != 2) {
        String[] params = insert.getFuncArgs();
        if (insert.getUpdateFunction() == 0)
            System.out.println(
                indent() + "\tnew_attribute(" + params[0] + "," + params[1] + ")");
        else
            System.out.println(indent() + "\tnew_ref(" + params[0] + "," + params[1] + ")");
    }

    // if there is a relative variable in the insert node
    if (insert.getRelativeVariable() != null) {
        if (insert.getOrderPredicate() == 0)
            System.out.println(indent() + "\tBEFORE");
        else
            if (insert.getOrderPredicate() == 1)
                System.out.println(indent() + "\tAFTER");
        System.out.println(
            indent() + "\tRelativeVariable: " + insert.getRelativeVariable());
    }
    System.out.println(indent() + "\t|");

    return null;
}
/**
 * Output of the Rename node in an XAT.
 * @reutrn null
 * @param rename The Rename for output.
 * @since 1.0
 */
 
public Object visit(Rename rename) {
    System.out.println(indent() + "Rename node:");
    System.out.println(indent() + "\tdata: ");
    System.out.println(indent() + "\tContext: " + rename.getContext());
    System.out.println(indent() + "\tVariable: " + rename.getName());
    System.out.println(indent() + "\tContent: " + rename.getVariable());
    System.out.println(indent() + "\t|");

    return null;

}
/**
 * Output of the Replace node in an XAT.
 * @reutrn null
 * @param replace The Replace for output.
 * @since 1.0
 */
 
public Object visit(Replace replace) {
    System.out.println(indent() + "Replace node:");
    System.out.println(indent() + "\tdata: ");
    System.out.println(indent() + "\tContext: " + replace.getContext());
    System.out.println(indent() + "\tVariable: " + replace.getVariable());

    // if there is an update function in the replace node 
    if (replace.getUpdateFunction() != 2) {
        String[] params = replace.getFuncArgs();
        if (replace.getUpdateFunction() == 0)
            System.out.println(
                indent() + "\tnew_attribute(" + params[0] + "," + params[1] + ")");
        else
            System.out.println(indent() + "\tnew_ref(" + params[0] + "," + params[1] + ")");
    }

    if (replace.getPCDATAContent() != null)
        System.out.println(indent() + "\tContent: " + replace.getPCDATAContent());
    System.out.println(indent() + "\t|");

    return null;
}
/**
 * Output of the Aggregate node in an XAT.
 * @reutrn null
 * @param agg The aggregate for output.
 * @since 1.0
 */
public Object visit(Aggregate agg) {
    System.out.println(indent() + "Aggregate node\n" + indent() + "   data: ");
    System.out.println(indent() + "\toutput col: " + agg.getOutputColumnName());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the NavCollection node in an XAT.
 * @reutrn null
 * @param navigate The NavCollection for output.
 * @since 1.0
 */
 
public Object visit(NavCollection navigate) {
    System.out.println(
        indent() + "NavCollection Node: \n" + indent() + "  Entry Point: ");
    try {

        for (int i = 0; i < navigate.getEntryPoint().length; i++) {
            NavigationStep[] entry = navigate.getEntryPoint();
            if(entry[i]!=null)
	            System.out.println(indent() + "\t " + entry[i].toString());
        }
        System.out.println(indent() + "  Destination: ");
        if (navigate.getDestination() != null) {
            for (int j = 0; j < navigate.getDestination().length; j++) {
                NavigationStep[] dest = navigate.getDestination();
                if(dest[j] != null)
	                System.out.println(indent() + "\t " + dest[j].toString());
            }
        }
    } catch (NullPointerException e) {
        //there is no entry point or destination so just print out nothing
    }
    System.out.println(indent() + "  Binding: " + navigate.getBinding());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the NavUnnest node in an XAT.
 * @reutrn null
 * @param navigate The NavUnnest for output.
 * @since 1.0
 */
 
public Object visit(NavUnnest navigate) {
    System.out.println(indent() + "NavUnnest Node: \n" + indent() + "  entry: ");
    try {

        for (int i = 0; i < navigate.getEntryPoint().length; i++) {
            NavigationStep[] entry = navigate.getEntryPoint();
            System.out.println(indent() + "\t " + entry[i].toString());
        }
        System.out.println(indent() + "  destination: ");
        if (navigate.getDestination() != null) {
            for (int j = 0; j < navigate.getDestination().length; j++) {
                NavigationStep[] dest = navigate.getDestination();
                System.out.println(indent() + "\t " + dest[j].toString());
            }
        }
    } catch (NullPointerException e) {
        //there is no entry point or destination so just print out nothing
    }
    System.out.println(indent() + "  binding: " + navigate.getBinding());
    System.out.println(indent() + "\t|");
    return null;
}
/**
 * Output of the Tagger node in an XAT.
 * @reutrn null
 * @param tagVisit The Tagger for output.
 * @since 1.0
 */
 
public Object visit(Tagger tagVisit) {
	
    System.out.println(indent() + tagVisit.toString());
	System.out.println(indent() + "\t|");

    return null;
}
/**
 * Output of the XMLUnion node in an XAT.
 * @reutrn null
 * @param xu The XMLUnion for output.
 * @since 1.0
 */
 
public Object visit(XMLUnion xu) {
    System.out.print(indent() + "XML Union node: \n" + indent() + "  children: ");
    String[] names = xu.getColumnNames();
    for (int i = 0; i < names.length; i++)
        System.out.print(names[i] + ' ');
    System.out.println();
    System.out.println(indent() + "\toutput col: " + xu.getOutputColumnName());
    System.out.println(indent() + "\t|");
    return null;
}
}
