package edu.wpi.cs.dsrg.utils;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

/**
 * This class is make a deep copy for given object.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */

public class DeepCopyMaker {

/**
 * Default DeepCopyMaker constructor.
 * @since 1.0
 */
private DeepCopyMaker() {
    //I made constructor private so that DeepCopyMaker could not be created
}
/**
 * Get the value of the propertity 'DEBUG_DeepCopyMaker'.
 * @return boolean True: With debug information. False: No debug information.
 * @since 1.0
 */
 
public static boolean isDebug() {
	return System.getProperty("DEBUG_DeepCopyMaker")!=null;
}
/**
 * Doing deep copy using given object.
 * @param obj2DeepCopy The source object.
 * @return Object The copied result.
 * @since 1.0
 */
static public Object makeDeepCopy(Object obj2DeepCopy) {
    // test how slow is this one.
    long start = System.currentTimeMillis();
    //obj2DeepCopy must be serializable
    ObjectOutputStream outStream = null;
    ObjectInputStream inStream = null;

    try {
        ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
        outStream = new ObjectOutputStream(byteOut);
        // serialize and write obj2DeepCopy to byteOut
        outStream.writeObject(obj2DeepCopy);
        //always flush your stream
        outStream.flush();

        ByteArrayInputStream byteIn = new ByteArrayInputStream(byteOut.toByteArray());
        inStream = new ObjectInputStream(byteIn);
        // read the serialized, and deep copied, object and return it
        return inStream.readObject();
    } catch (IOException ioe) {
        //handle the exception
        ioe.printStackTrace(System.err);
        return null;
    } catch (ClassNotFoundException cnfe) {
        //handle the exception
        cnfe.printStackTrace(System.err);
        return null;
    } finally {
        try {
            //always close your streams in finally clauses
            if (outStream != null)
                outStream.close();
            if (inStream != null)
                inStream.close();
            long end = System.currentTimeMillis();
            //if (isDebug())
            //    System.err.println("Cloning:" + (end - start) / 1000.0);
        } catch (IOException ioe) {
            ioe.printStackTrace(System.err);
        }
    }
}
}
