package edu.wpi.cs.dsrg.utils;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;

/**
 * This class is used to store a number to a temporary file.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */

public class SequenceNumber {
	/**
	 * The file name.
	 * @since 1.0
	 */
    String filename;
/**
 * Specify how many steps should taken to finish one get() function call.
 * @since 1.0
 */
long step;
/**
 * The SequenceNumber constructor.
 * @param filename The file name.
 * @since 1.0
 */
public SequenceNumber(String filename) {
	super();
	this.filename = filename;
	this.step = 0;
}
/**
 * The SequenceNumber constructor.
 * @param filename The file name inluding full path.
 * @param step Specify once the file opened, how many steps there for each get() function call. 
 * @since 1.0
 */

public SequenceNumber(String filename, long step) {
	super();
	this.filename = filename;
	this.step = step;
}
/**
 * Get next unique number.
 * @return long The unique number got from file.
 * @since 1.0
 */
public long get() {
    try {
        File f = new File(this.filename);
        long number = 0;
        // if file exist read from it.
        if (f.exists()) {
            DataInputStream in = new DataInputStream(new FileInputStream(this.filename));
            number = in.readLong();
            in.close();
        }

        // then we can write it to the file again.
        DataOutputStream out =
            new DataOutputStream(new FileOutputStream(this.filename));
        out.writeLong(number + step);
        out.close();
        if (System.getProperty("DEBUG_SequenceNumber") != null)
            System.out.println("get " + number);
        return number;
    } catch (Exception e) {
        System.out.println(e.getMessage());
    }
    return 0;
}
/**
 * Insert the method's description here.
 * Creation date: (5/22/2002 3:30:09 PM)
 * @param args java.lang.String[]
 */
public static void main(String[] args) {
	SequenceNumber seq = new SequenceNumber("counter1", 1000);
	for(int i = 0; i < 10; i++)
		System.out.println(seq.get());
}
/**
 * Set the file name as new one.
 * @param newFilename The new file name.
 * @since 1.0
 */
public void setFilename(String newFilename) {
	this.filename = newFilename;
}
}
