package edu.wpi.cs.dsrg.xmldb.xat.common.expression;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.sqlgeneration.OperatorEvaluator;

/**
 * This class defines a binary OR expression.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 */

public class BinORExpressionImp
    extends BinExpressionImp
    implements BinORExpression {

    /**
     * The left and right expression.
     * @since 1.0
     */
    Expression eLeft, eRight;
/**
 * The default binary OR expression constructor.
 * @since 1.0
 */

public BinORExpressionImp() {
    eLeft = null;
    eRight = null;
}
/**
 * The binary OR expression constructor.
 * @param left  The left part of binary expression.
 * @param right The right part of binary expression.
 * @since 1.0
 */

public BinORExpressionImp(Expression left, Expression right) {
    eLeft = left;
    eRight = right;
}
/**
 * Evaluate the expression of the XATOperator.
 * @return The evaluate result of expression.
 * @param xopOper The operator evaluated.
 * @param iPositions The position of tuple in the XATTable.
 * @since 1.0
 */
 
public Object eval(XATOperator xopOper, int[] iPositions) {
    try {

        boolean c1 =
            ((XATDataValue) eLeft.eval(xopOper, iPositions))
                .convertToBoolean()
                .booleanValue();

        boolean c2 =
            ((XATDataValue) eRight.eval(xopOper, iPositions))
                .convertToBoolean()
                .booleanValue();

        return new XATDataValueImp(new Boolean(c1 || c2));
    } catch (ClassCastException e) {
        /* um, fail? */
        return new XATDataValueImp(Boolean.FALSE);
    }
}
/**
 * Get the left part of expression.
 * @return Expression The left part of binary expression
 * @since 1.0
 */
 
public Expression getLeft() {
    return eLeft;
}
/**
 * Get the right part of expression.
 * @return Expression The right part of binary expression
 * @since 1.0
 */
public Expression getRight() {
    return eRight;
}
/**
 * Set the left part of expression.
 * @param left The left part of a binary expression.
 * @since 1.0
 */
 
public void setLeft(Expression left) {
    eLeft = left;
}
/**
 * Set the right part of expression.
 * @param left The right part of a binary expression.
 * @since 1.0
 */

public void setRight(Expression right) {
    eRight = right;
}
/**
 * Convert the expression as a SQL statement.
 * @return java.lang.String The converted string.
 * @param soe The operator evaluator for SQL generation.
 * @since 1.0
 */

public String toSQL(OperatorEvaluator soe) {
	return eLeft.toSQL(soe) + " OR " + eRight.toSQL(soe);
}
/**
 * Output a string of this expression.
 * @return String the output result.
 * @since 1.0
 */

public String toString() {
    return "(" + eLeft.toString() + " OR " + eRight.toString() + ")";
}
}
