package edu.wpi.cs.dsrg.xmldb.xat.common.expression;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.ColumnName;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.Visitable;
import edu.wpi.cs.dsrg.xmldb.xat.component.sqlgeneration.OperatorEvaluator;

/**
 * This class defines an expression.
 * Expression can be
 * used in the Join and Select operators as a condition.
 * It can also be used in the LET clause to do a variable change.
 * We take the design from the xacute.quilt.Operator and their
 * subclasses. 
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 */

public interface Expression extends Visitable {
    /**
    * <p>
    * You don't really need a visitor pattern to implement the execution
    * of it. Also, it's not using the XATTree.
    * So, just implement the eval() function.
    */
/**
 * Evaluate the expression of the XATOperator.
 * @return The evaluate result of expression, it could be Integer, Float, Double, String, Boolean.
 * @param xopOper The operator evaluated.
 * @param iPositions The position of tuple in the XATTable.
 * @since 1.0
 */
 
public abstract Object eval(XATOperator xopOper, int[] iPositions);
/**
 * get column names related with expression.
 * @return ColumnName[] The array containing all the related column names.
 * @since 1.0
 */

ColumnName[] getRelatedColumnNames();
/**
 * Convert the expression as a SQL statement.
 * @return java.lang.String The converted string.
 * @param soe The operator evaluator for SQL generation.
 * @since 1.0
 */
 
String toSQL(OperatorEvaluator soe); 
}
