package edu.wpi.cs.dsrg.xmldb.xat.common.operator;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Set;

import edu.wpi.cs.dsrg.utils.DeepCopyMaker;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATNodeDoesNotExistException;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableEmptyException;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableOutOfRangeException;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataSource;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup.SchemaCleanupStatistics;

/**
 * Implementation for interface XATOperator. <P>
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0    
 */

public abstract class XATOperatorImp
    extends XATQueryObjectImp
    implements XATOperator, Serializable {
    protected XATNode xnodOwner;
    protected XATDataSource xtabOutput;
    protected String outputColumnName;

    protected int columnsFiltered = 0;

    /**
     * Assign a node to this operator .
     * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
     * @since   1.0        
     */
    public XATOperatorImp(XATNode node) {
        super();
        xnodOwner = node;
        node.setOperator(this);
        xtabOutput = null;
    }
    /**
     * Clones the XATOperator.
     * @since   1.0        
     */

    public Object clone() {
        try {
            return super.clone();
        } catch (CloneNotSupportedException e) {
            System.err.println(e.getMessage());
            return DeepCopyMaker.makeDeepCopy(this);
        }
    }
    /**
     * Gets column filtersd.
     * @return columnsFiltered.
     * @since   1.0        
     */
    public int getColumnsFiltered() {
        return columnsFiltered;
    }
    /**
     * Gets the node.
     * @return xnodOwner.
     * @since   1.0        
     */
    public XATNode getNode() {
        return this.xnodOwner;
    }
    /**
     * Gets the output of this Operator.
     * @return the output XATTable of this operator.
     * @since   1.0        
     * @deprecated use XATOperator.getOutputDataSource, then cast to a table or a queue
     */
    public XATTable getOutput() {
        return (XATTable) this.xtabOutput;
    }
    /**
     * Gets output column name.     
     * @return output column name
     * @since   1.0        
     */
    public String getOutputColumnName() {
        return this.outputColumnName;
    }
    /**
     * Get the output of this Operator.
     * @return the output DataSource of this operator.
     * @since   1.0        
     */
    public XATDataSource getOutputDataSource() {
        return this.xtabOutput;
    }
    /**
     * Get all output data sources for this XATOperator
     * @return Iterator of data sources.
     * @since   1.0        
     */
    public java.util.Iterator getOutputDataSources() {
        ArrayList list = new ArrayList();
        if (xtabOutput != null) {
            list.add(xtabOutput);
        }
        return list.iterator();
    }
    /**
     * Get primary data source. The primary data source will be the first data source in all the operators. For example:
     * For all the Join the primary data source is source l. For all the Union operators is the first source.
     * For CO is the source S1.
     * @return XATDataSource of the primary data source.
     * @throws XATNodeDoesNotExistException
     * @since   1.0        
     */
    public XATDataSource getPrimaryDataSource()
        throws XATNodeDoesNotExistException {
        return xnodOwner.getChild(0).getOperator().getOutputDataSource();

    }
    /**
    * Get primary data source. The primary data source will be the first data source in all the operators. For example:
    * For all the Join the primary data source is source l. For all the Union operators is the first source.
    * For CO is the source S1.
    * @return XATTable of the primary data source.
    * @throws XATNodeDoesNotExistException
     * @since   1.0       
    * @deprecated use XATOperator.getPrimaryDataSource, then cast to a table or a queue
    */
    public XATTable getPrimarySource() throws XATNodeDoesNotExistException {
        return (XATTable) xnodOwner.getChild(0).getOperator().getOutputDataSource();

    }
    /**
     * This will set the default properties that all XATOperators will have.
     * @since   1.0        
     */
    protected void setDefaultProperties() {
        super.setDefaultProperties();
        addProperty("TIME_TOOK_TO_RUN", "0");
        addProperty("TIME_TOOK_TO_RUN_TOTAL", "0");
        /** 
          * this is a unique ID assigned to each operator
          * this will help to keep track of operators from
          * one experiment to the next.  The idea is that
          * if a query plan is generated the same way from 
          * experiment to experiment, then its ID will be the same always.   
        	*/
        addProperty("OBJECT_ID", "-1");

    }
    /**
     * Sets the opertor's node.
     * @param newNode edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
     * @since   1.0        
     */
    public void setNode(XATNode node) {
        xnodOwner = node;
    }
    /**
     * Sets the ouput data source.
     * @param newoutput output data source
     * @since   1.0        
     */
    public void setOutput(XATDataSource output) {
        this.xtabOutput = output;
        Set schema = xnodOwner.getSchema();

        if (schema != null) {
            try {
                // to protect the wrong computed schema.
                // if the schema cannot be found in the table.
                // then, the columns will not be filtered by this schema.
                String[] schema_name = (String[]) schema.toArray(new String[] {
                });
                for (int i = 0; i < schema_name.length; i++) {
                    // test if the schema exists in the current table.
                    if (!((XATTable) output).getColNames().contains(schema_name[i])) {
                        // that means it doens't exist.
                        // then, do nothing for the output's schema.
                        return;
                    }
                }

                SchemaCleanupStatistics.filteredNodes.add(this);
                columnsFiltered = 0;
                for (int i = 1; i <= ((XATTable) xtabOutput).getNumberOfColumns();) {
                    String name;
                    try {
                        name = ((XATTable) xtabOutput).getColumnName(i);
                    } catch (XATTableOutOfRangeException e) {
                        name = null;
                    }
                    if (schema.contains(name)) {
                        ++i;
                    } else {
                        columnsFiltered++;
                        try {
                            ((XATTable) xtabOutput).deleteColumn(i);
                        } catch (XATTableOutOfRangeException e) {

                        } catch (XATTableEmptyException e) {

                        }
                    }
                }
                /** if the XATDataSource is not a table*/
            } catch (ClassCastException e) {

            }
        }
    }
    /**
     * Sets output columns.
     * @param OutputColumnName java.lang.String
     * @since   1.0        
     */
    public void setOutputColumnName(String OutputColumnName) {
        this.outputColumnName = OutputColumnName;
    }
    /**
    * sets the given property to the new value.  This method is responsible
    * for validating the given property.  An exception is thrown if
    * the value is not correct.
    * Validation should be done in the following manner: 
    * First the method should check if the object supports the given property by calling
    * <tt>isValidPropertyName(String)<tt>.  If it is not a valid property name, then an exception should
    * be thrown. <p>
    *
    * Once it is known that the object supports the property, the value must be validated.  Each implementor of
    * XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
    * For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
    * has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
    * to set a common type, it should be validated by the super class by calling <tt>super.setProperty(name, value)<tt>.  This
    * will reduce duplicate code. <p>
    *
    * When it is determined the name / value pair is correct, the pair is updated using <tt>setNewPropertyValue()<tt> <p>
    * <tt>XATStatisticsImp<tt> provides default validation by checking to see if the property's value is
    * >=0.  propertyValues that are Strings will just be set.  If the property requires more strict constraints than this, 
    * it is up to the subclass that created the property to validate it (as described above).
    * @param name the name of the property to set
    * @param value the new value of the property
    * @throws Exception if the given property cannot be found or
    * if the value is not valid for the given property.
    * @param name the name of the property to set
    * @param value the new value of the property
    * @throws Exception if the given property cannot be found or
    * if the value is not valid for the given property.
    * @since   1.0       
    */
    public void setProperty(String name, String value) throws java.lang.Exception {
        super.setProperty(name, value);
    }
    /**
     * returns a string representation 
     * @return java.lang.String
     * @since   1.0        
     */
    public String toString() {
        String sName = this.getClass().getName();
        //find the last dot.
        int index = 0;
        for (index = sName.length() - 1; index >= 0; index--)
            if (sName.charAt(index) == '.')
                break;
        sName = sName.substring(index + 1);
        return sName + "(" + this.getOutputColumnName() + ")";
    }
}
