package edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.Visitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * Implementation for interface Source.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0      
 */

public class SourceImp extends XATOperatorImp implements Source {
    protected String sName;
    protected int iType;

    /**
     * Assigns a node to this operator, sets the source type to XML, and source name to "" .
     * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
     * @since   1.0          
     */
    public SourceImp(XATNode xnodInit) {
        super(xnodInit);
        iType = XML;
        sName = "";
    }
    /**
     * Assigns a node to this operator, sets the source type and source name .
     * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
     * @param sNameInit is source type
     * @param iTypeInit	is source name 
     * @since   1.0          
     */
    public SourceImp(XATNode xnodInit, String sNameInit, int iTypeInit) {
        super(xnodInit);
        iType = iTypeInit;
        sName = sNameInit;
    }
    /** 
     * It used to implement a visitor pattern. 
     * @param visitor for the visitor pattern
     * @since   1.0         
    */

    public void accept(Visitor visitor) {
    }
    /**
     * get the current name of this source.
     * @return the current name.
     * @since   1.0          
     */
    public String getName() {
        return sName;
    }
    /** 
     * Return the type of this source. It could be either XML or TABLE. 
     * @return the type of this source.
     * @since   1.0          
    */
    public int getType() {
        return iType;
    }
    /**
     * Set the new name of this source.
     * @param The new name.
     * @return original name.
     * @since   1.0          
     */
    public String setName(String newName) {
        String sTemp = sName;
        sName = newName;
        return sTemp;
    }
    /**
     * Set the type of this source.
     * @param newType the new type of this source, could be either Source.XML or Source.TABLE.
     * @return original type.
     * @since   1.0          
     */
    public int setType(int newType) {
        int iTemp = iType;
        iType = newType;
        return iTemp;
    }
    /**
     * returns a string representation 
     * @return java.lang.String
     * @since   1.0         
    */
    public String toString() {
        return "Source ("
            + this.getOutputColumnName()
            + ")[type: "
            + iType
            + ", name: "
            + sName
            + "]";
    }
}
