package edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function;
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.expression.Expression;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATTable;

/**
 * It represents an arithmatic evaluation in the XQuery.
 * It used to represent all the possible evaluations by expressions in the common.expression package.
 * For example, it can be used to represent a string, a double, and a + b, etc.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
*/
public class FunArith extends FunctionImp {
   /**
    * Assigns a node to this operator and the name of the opertor to "Arith".
    * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
    * @since   1.0        
    */
    public FunArith(XATNode node) {
        super(node);
        this.setName("Arith");
    }

   /**
     * Execution method for FUNArith.
     * @param deRefParamV are the parameters 
     * @since   1.0      
    */
    public void execute() {
        try {
            XATTable xatInput = (XATTable) this.getPrimaryDataSource();
            XATTable xatOutput = (XATTable) xatInput.deepCloneDataSource();
            int resultTupleNum = xatOutput.getNumberOfTuples();
            // original parameters.
            Object[] paramV = this.getParams();
            if (paramV.length != 1)
                throw new Exception("Only one params are allowed!");
            else {
                Vector resultV = new Vector();
                for (int i = 1; i <= resultTupleNum; i++) {
                    int[] iPosition = new int[1];
                    iPosition[0] = i;
                    Object o =
                        ((Expression) paramV[0]).eval(this.getNode().getOperator(), iPosition);
                    if (o instanceof XATDataValue)
                        resultV.add(o);
                    else
                        resultV.add(new XATDataValueImp(o));
                }
                xatOutput.appendColumn(this.getOutputColumnName(), resultV);
                this.setOutput(xatOutput);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }

    }
}