package edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function;
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATTable;

/**
 * Implementation for the Function calss.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0   
 */
public class FunctionImp extends XATOperatorImp implements Function {

    /**
    * Function's name.
    */
    protected String sName;

    /**
    * Function's parameters.
    */
    protected Object[] oaParams;

    /**
     * Constuctor.
     */
    protected FunctionImp() {
        super(null);
    }

    /**
     * Assign a node to this operator .
     * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
     * @since   1.0        
     */
    public FunctionImp(XATNode node) {
        super(node);
    }

    /**
     * It used to dynamically load an Function implementation,
     * and then, execute the specific function.
     * We assume the function class will start with "fun".
     * @since   1.0         
     */
    public void execute() {
        try {
            FunCall f =
                (FunCall) Class
                    .forName(this.getPackage() + "Fun" + this.sName.toUpperCase())
                    .newInstance();
            f.eval(this);
        } catch (ClassNotFoundException cnfe) {
            cnfe.printStackTrace();
        } catch (InstantiationException ie) {
            ie.printStackTrace();
        } catch (IllegalAccessException iae) {
            iae.printStackTrace();
        }
    }

    /**
     * Get the name of this function.
     * @return the Name of this Function.
     * @since   1.0              
     */
    public String getName() {
        return sName;
    }

    /**
     * Used to extract the package name.
     * @since   1.0                  
    */
    String getPackage() {
        String className = this.getClass().getName();
        String packageName = className.substring(0, className.indexOf("Function"));
        return packageName;
    }

    /**
     * It returns all the parameters in an array.
     * @return The parameters.
     * @since   1.0                   
     */
    public Object[] getParams() {
        return oaParams;
    }

    /**
     * Get all the sources of this function.
     * @return XATTable of all the sources of this function.
     * @since   1.0                   
     */
    public XATTable[] getSources() {
        //int iNumChildren = Array.getLength(xnodOwner.getChildren());
        /** Get all the children nodes of this node. */
        //XATNode[] xnodChildren = new XATNode[iNumChildren];
        //xnodChildren = xnodOwner.getChildren();

        Vector vXtables = new Vector();
        String[] sReturn = {
        };
        //int iNumNodes = iNumChildren;
        /* for( int i=0; i<iNumNodes ;i++)
         {
           vXtables.addElement(xnodChildren[i].getOperator().getOutput());
         }
         return (XATTable[]) vXtables.toArray(sReturn);
        */
        return null;
    }

    /**
     * Insert the method's description here.
     * Creation date: (1/27/02 10:09:52 PM)
     * @param newSName java.lang.String
     * @since   1.0                   
     */
    public void setName(java.lang.String newSName) {
        sName = newSName;
    }

    /**
     * Set the new parameters of this function.
     * @return original parameters.
     * @since   1.0                   
     */
    public Object[] setParams(Object[] newParams) {
        Object[] oaOrig = oaParams;
        oaParams = newParams;
        return oaOrig;
    }

    /**
     * Insert the method's description here.
     * Creation date: (8/28/02 3:23:50 PM)
     * @return java.lang.String
     * @since   1.0              	 
     */
    public String toString() {
        String result = this.sName + "(";
        for (int i = 0; i < this.oaParams.length; i++) {
            result += oaParams[i].toString() + " ";
        }
        return result + ")";
    }
}