package edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.ColumnNameListImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.Visitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * Implementation for the GroupBy interface.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0   
 */
public class GroupByImp extends ColumnNameListImp implements GroupBy {

    protected XATTree xtreFunctions;

    /**
     * Assigns a node to this operator .
     * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
     * @since   1.0   
     */
    public GroupByImp(XATNode xnodInit) {
        super(xnodInit);
    }
    /**
     * Assigns a node to this operator and an XAT tree (subquery) .
     * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
     * @param xtreFuncInit is the groupby XAT tree     
     * @since   1.0   
     */
    public GroupByImp(XATNode xnodInit, XATTree xtreFuncInit) {
        super(xnodInit);
        xtreFunctions = xtreFuncInit;
    }
    /** 
     * Used to implement a visitor pattern. 
     * @param visitor for the visitor pattern
     * @since   1.0       
     */
    public void accept(Visitor visitor) {

    }
    /** 
     * Gets a the XAT tree associtaed with this GroupBy (the groupby subquery). 
     * @return the GroupBy XAT tree
     * @since   1.0         
    */
    public XATTree getXATTree() {
        return xtreFunctions;
    }
    /** 
     * Gets a the XAT tree associtaed with this GroupBy (the groupby subquery). 
     * @return the GroupBy XAT tree
     * @since   1.0         
    */
    public XATTree setXATTree(XATTree newXAT) {
        XATTree xtreFuncTemp = xtreFunctions;
        xtreFunctions = newXAT;
        return xtreFuncTemp;
    }
    /**
    * Returns a String that represents the value of this object.
    * @return a string representation of the receiver
    * @since   1.0          
    */
    public String toString() {
        StringBuffer output = new StringBuffer("GroupBy");
        // add those gb columns.
        String[] cols = this.getColumnNames();
        output.append("[");
        if (cols != null) {
            for (int i = 0; i < cols.length - 1; i++)
                output.append(cols[i] + ", ");
            output.append(cols[cols.length - 1]);
        }
        output.append("]");
        // add the focus column.
        output.append(":");
        output.append(this.getOutputColumnName());

        return output.toString();
    }
}
