package edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator;
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATNodeDoesNotExistException;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATTable;

/**
 * The Insert interface.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0   
 */
public interface Insert extends UpdateOperator {

    // constants indicating the order predicate in the insert query
    public static final int AFTER = 1;
    public static final int BEFORE = 0;

    // constants indicating the type of update function in the insert query
    public static final int NEW_ATTRIBUTE = 0;
    public static final int NEW_REF = 1;

    // unknown type.
    public static final int UNKNOWN = 2;
    public String[] getFuncArgs();

    /**
     * Gets the order predicate of the insert statement.
     * @return int
     * @since   1.0       
     */
    public int getOrderPredicate();

    /**
     * Gets the PCDATA content to be inserted.
     * @return String
     * @since   1.0       
     */
    public String getPCDATAContent();

    /**
     * Gets the Relative Variable value.
     * @return String
     * @since   1.0       
     */
    public String getRelativeVariable();

    /**
      * Gets the element contents that should replace the original element
      * @return XATTAble containing columns representing the new element
    	 * @throws XATNodeDoesNotExistException     
      * @since   1.0            
      */
    public XATTable getRight() throws XATNodeDoesNotExistException;

    /**
       * Gets the Update Function.
       * @return int
       * @since   1.0            
       */
    public int getUpdateFunction();

    /**
     * Sets the function argument, eg. INSERT $a BEFORE|AFTER $b.
     * @param funcArgs is an array of string for the arguments.
     * @since   1.0       
     */
    public void setFuncArgs(String[] funcArgs);

    /**
     * Sets the order predicate of the insert statement, eg. INSERT $a BEFORE|AFTER $b
     * @param orderPredicate int
     * @since   1.0        
     */
    public void setOrderPredicate(int orderPredicate);

    /**
      * set the PCDATA content to be inserted
      * @param PCDATAContent String
      * @since   1.0             
      */
    public void setPCDATAContent(String PCDATAContent);

    /**
      * set the Relative Variable value.
      * @param relVar String
      * @since   1.0             
      */
    public void setRelativeVariable(String relVar);

    /**
      * set the Update Function.
      * @param updateFunc int
      * @since   1.0             
      */
    public void setUpdateFunction(int updateFunc);
}