package edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator;
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATNodeDoesNotExistException;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATTable;

/**
 * The Replace interface.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0   
 */
public interface Replace extends UpdateOperator {

    // constants indicating the type of update function in the replace update
    public static final int NEW_ATTRIBUTE = 0;
    public static final int NEW_REF = 1;

    /**
     * Gets the right (secondary) data source.
     * @return XATTAble containing columns representing the new element
     * throws XATNodeDoesNotExistException     
     * @since   1.0       
    */
    public XATTable getRight() throws XATNodeDoesNotExistException;

    /** 
     * Gets the PCDATA content of the operator 
     * @return int
     * @since   1.0        
     */
    public String getPCDATAContent();

    /** 
     * Gets function argument 
     * @return String
     * @since   1.0        
     */
    public String[] getFuncArgs();

    /** 
     * Gets the update function
     * @return int
     * @since   1.0        
     */
    public int getUpdateFunction();

    /** 
     * Gets the replacing varibale. 
     * @return string
     * @since   1.0        
     */
    public String getReplacingVariable();

    /**
     * Sets the PCDATA content that should replace the original content
     * @param String
     * @since   1.0       
     */
    public void setPCDATAContent(String PCDATAContent);

    /**
     * Sets the PCDATA content that should replace the original content
     * @param funcArgs an array of String
     * @since   1.0       
     */
    public void setFuncArgs(String[] funcArgs);

    /**
     * Sets the update function
     * @param funcType int
     * @since   1.0       
     */
    public void setUpdateFunction(int funcType);

    /**
     * Sets the replacing varible.
     * @param newVar String
     * @since   1.0       
     */
    public void setReplacingVariable(String newVar);

}