package edu.wpi.cs.dsrg.xmldb.xat.common.treestructure;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.ColumnName;

/**
 * This Class implemented the ListPattern Interface.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */

public class ListPatternImp implements ListPattern
{
	/**
	 * The list pattern.
	 * @since 1.0
	 */
    protected Vector listp;

/**
 * The default ListPatternImp Constructor.
 * @since 1.0
 */
public ListPatternImp() {
    listp = new Vector();
}
/**
 * The ListPatternImp Constructor.
 * @param listp The list pattern used to construct the ListPatternNode.
 * @deprecated It should use Vector as it's parameter. It can not enforce the type of the items in the Vector.
 * @link add(String) and add(ColumnName)
 * @since 1.0
 */
public ListPatternImp(Vector listp) {
    this.listp = listp;
}
/**
 * Add the given column name into a ListPattern.
 * @return void
 * @param param The ColumnName adding to.
 * @since 1.0
 */
public void add(ColumnName col) {
    listp.add(col);
}
/**
 * Connect with another list pattern.
 * @return void     
 * @param text java.lang.String
 * @since 1.0     
 */
public void add(ListPattern pattern) {
    listp.addAll(pattern.getList());
}
/**
 * Add text in list pattern.
 * Creation date: (2/9/2003 10:26:56 PM)
 * @param text java.lang.String
 * @since 1.0     
 */
public void add(String text) {
    listp.add(text);
}
/**
 * Test to see if this DOMPattern contains a specific column name in its
 * pattern tree.
 * @return boolean True: The listPattern contains the given columnName. False: otherwise.
 * @param columnName The columnName searching for.
 * @since 1.0
 */
public boolean containsColumn(java.lang.String columnName) {
	// then iterature through the listPattern to see if the column is in it.
	Vector list = this.getList();
	for(int i = 0; i < list.size(); i++){
		Object item = list.elementAt(i);
		if(item instanceof ColumnName){
			if(((ColumnName)item).getColumnName().equals(columnName))
				return true;
		}
	}
	return false;
}
/**
 * Get all the column names within the ListPattern.
 * @return String[] The vector of columns within the list pattern.
 * @since 1.0
 */
public String[] getColumnNames() {
    return (String[]) listp.toArray(new String[] {
    });
}
/**
 * Get the list.
 * @return Vector The vector of the list pattern.
 * @since 1.0     
 */

public Vector getList() {
    return listp;
}
/**
 * Set the list in the ListPattern.
 * @return void
 * @param newPattern  
 * @since 1.0     
 */

public void setList(Vector lpattern) {
    this.listp = lpattern;
}
/**
 * Convert the ListPattern into String.
 * @return String The result of converting.
 * @since 1.0     
 */

public String toString() {
    StringBuffer result = new StringBuffer();
    for (int i = 0; i < listp.size(); i++) {
        if (listp.elementAt(i) instanceof ColumnName) {
            result.append("{" + listp.elementAt(i).toString() + "}");
        } else
            result.append(listp.elementAt(i).toString());
    }
    return result.toString();
}
}
