package edu.wpi.cs.dsrg.xmldb.xat.common.xatnode;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.apache.xerces.parsers.DOMParser;
import org.apache.xml.serialize.DOMSerializer;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.Serializer;
import org.apache.xml.serialize.XMLSerializer;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;

/**
 * <pre>
 * The class that represents the value of one cell in an XATTable
 * The value is an atomic value or a Collection and 
 * can be a Dom or a String datatype.  Methods are provided that will allow
 * one type to be converted to another.  
 * </pre>
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public class XATDataValueImp
    extends edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObjectImp
    implements XATDataValue {
    /**
     * The type be a vector of XATDataValue.
     * @since 1.0
     */
    protected Vector collection;

    /** 
     * The type be an atomic value instead of a collection.
     * @since 1.0     
     */
    protected Object value;
/**
 * The default XATDataValueImp constructor.
 * @since 1.0
 */
public XATDataValueImp() {
	super();
	collection = new Vector();
	value = null;
}
/**
 * The XATDataValueImp constructor.
 * This is the only place you can creat an atomic valued XATDataValue.
 * @param obj The Object used to construct the XATDataValueImp.
 * @since 1.0 
 */
public XATDataValueImp(Object obj) {
	this();
	// if the input obj is a "collection",
	// then, call the add() function to put each of the items 
	// of this collection into this XATDataValue.
	if(obj instanceof Vector){
		this.add(obj);
	}else if(obj instanceof XATDataValue){
		this.collection = ((XATDataValueImp)obj).collection;
		this.value = ((XATDataValueImp)obj).value;
	}else {
		value=obj;
	}
}
/**
 * If you use add, means this XATDataValue is a collection.
 * This is the only method that can convert an atomic value into
 * a collection.
 * @return void
 * @param obj Add a given Object into XATDataValue.
 * @since 1.0
 */
public void add(Object obj) {
	if(obj == null)
		return;

	// if it is not a collection originally, but an atomic data.
	if(value != null){
		// make it a collection.
		this.collection.add(value);
		// clean the value.
		this.value = null;
	}

	// it's a "collection".
	// now, add the new item in this XATDataValue.
	if(obj instanceof Vector){
		// if to be added is a collection.
		Vector vec = (Vector)obj;
		// don't add the nested collection, but
		// contact two collections together.
		for(int i = 0; i < vec.size(); i++){
			this.add(vec.elementAt(i));
		}
	}else if(obj instanceof XATDataValueImp){
		// merge the collections.
		collection.addAll(((XATDataValue)obj).convertToCollection());
	}else {// it's an atomic value.
		collection.add(new XATDataValueImp(obj));
	}
}
/**
 * Converts the XATValue into a Boolean object.
 * @return Boolean A boolean object after convert.
 * @since 1.0
 */
public Boolean convertToBoolean() {
	// NOTE: We use the simple logic here.
    if ("TRUE".equals(toString().toUpperCase())) {
        return Boolean.TRUE;
    } else
        return Boolean.FALSE;

}
/**
 * This method will take whatever the value is and converts it into
 * a Collection (Vector)
 * @return Vector The convert result.
 * @since 1.0
 */
public Vector convertToCollection(){
	// NOTE: The XATDataValue itself still same as before. It will not
    // really convert this XATDataValue into a collection.

	// if it's not a collection,
	if(value != null){
		Vector tempCollection = new Vector();
		tempCollection.addElement(this);
		return tempCollection;
	}else
		return this.collection;	
}
/**
 * Convert the XATDataValue as a DOM object, and construct as an XML Node.
 * @return Node The convert result.
 * @since 1.0
 */
public Node convertToDOM() {
    Vector result;
    if (value != null)
        return (Node)convertToDOM(this.value);
    else
        result = (Vector) convertToDOM(this.collection);

    // we assume there is only one element in the Vector.
    // otherwise, we need to create an XML fragment to contain all those nodes.
    if (result != null) {
        if (result.size() == 1)
            return (Node) result.elementAt(0);
        else {
            // create a fragment.
            DocumentFragment fragment = new org.apache.xerces.dom.DocumentFragmentImpl();
            for (int i = 0; i < result.size(); i++)
                fragment.appendChild((Node) result.elementAt(i));
            return fragment;
        }
    } else // return an empty fragment.
        return new org.apache.xerces.dom.DocumentFragmentImpl();
      
}
/**
 * Convert the given object as a DOM object, and construct as an XML Node.
 * @return Node The convert result.
 * @since 1.0
 */

protected Object convertToDOM(Object o) {
    if (o instanceof Vector) {
        Vector v = (Vector) o;
        for (int i = 0; i < v.size(); i++) {
            v.setElementAt(convertToDOM(v.elementAt(i)), i);
        }
        return v;
    }else if (o instanceof XATDataValue){
	    
	    return convertToDOM(((XATDataValue)o).convertToDOM());
	    
    }else if (o instanceof String) {
        /* convert it... */
        InputSource is = new InputSource(new StringReader((String)o));
        DOMParser parser = new DOMParser();
        try {
            parser.parse(is);
        } catch (java.io.IOException ioe) {
            ioe.printStackTrace(System.err);
        } catch (org.xml.sax.SAXException sae) {
            sae.printStackTrace(System.err);
        }
        Document dDocument = parser.getDocument();
        return dDocument.getDocumentElement();
    }else if (o instanceof Node) {
        return o;
    }
    return o;
}
/**
 * Convert the XATDataValue into a Double object.
 * @return Double The convert result.
 * @since 1.0
 */
public Double convertToDouble() {
	if(value != null)
		return new Double(value.toString());
	else if(collection.size() == 1)
		return ((XATDataValue)collection.elementAt(0)).convertToDouble();
	else
		throw new NumberFormatException();
}
/**
 * Converts the XATValue into a String.  
 * If the Value is a collection, then one large string is created.
 * @return String The convert result.
 * @since 1.0
 */

public String convertToString() {

	// a tmp variable to hold the result.
	Vector result = new Vector();
	if(this.isCollection())
    	result = (Vector) convertToString(this.collection);
    else
    	result.addElement(convertToString(this.value));

    // return string.
    String toReturn = new String();

    if (result.size() > 0) {

        for (int i = 0; i < result.size() - 1; i++) {
            toReturn = toReturn + result.elementAt(i) + " ";
        }

        toReturn = toReturn + result.elementAt(result.size() - 1).toString();
    }

    return toReturn;

}
/**
 * Converts the given object into a String.  
 * If the Value is a collection, then one large string is created.
 * @return String The convert result.
 * @since 1.0
 */
protected Object convertToString(Object o) {

    if (o instanceof String) {
        return o;
    } else
        if (o instanceof Vector) {
            Vector v = (Vector) o;
            for (int i = 0; i < v.size(); i++) {
                v.setElementAt(convertToString(v.elementAt(i)), i);
            }
            return v;

        } else
            /**
            * Elements and DocumentFragments are both accounted for here
            */
            if (o instanceof Node) {
                Document dDOMTree = ((Node) o).getOwnerDocument();
                OutputFormat format = new OutputFormat(dDOMTree); //Serialize DOM
                StringWriter stringOut = new StringWriter(); //Writer will be a String
                Serializer serial = new XMLSerializer();
                serial.setOutputFormat(format);
                serial.setOutputCharStream(stringOut);
                try {
                    DOMSerializer ds = serial.asDOMSerializer();
                    ds.serialize((Element) o);
                } catch (java.io.IOException ioe) {
                    ioe.printStackTrace(System.err);
                }
                String s = stringOut.toString();
                if (s.substring(0, 5).equals("<?xml")) {
                    int i = s.indexOf('\n');
                    s = s.substring(i + 1);
                }

                return s;
            } else
                if (o instanceof Attr) {
                    return ((Attr) o).getValue();
                }

    return o;
}
/**
 * Get the Hash code of this XATDataValue.
 * @return String The Hash code of this XATDataValue.
 * @since 1.0
 */

public java.lang.String getId() {
	return this.hashCode()+"";
}
/**
 * Get the value of this XATDataValue.
 * @return String The value of this XATDataValue.
 * @since 1.0
 */

public Object getValue() {
	return this.value;
}
/**
 * Check whether this XATDataValueImp is a collection or an atomic value.
 * @return boolean True: This XATDataValue is a collection. False: otherwise.
 * @since 1.0
 */
 
public boolean isCollection() {
	/** NOTE: It might be an empty collection, where value is null and
	 * also the collection is an empty vector.
	 */

	if(value != null)
		return false;
	else
		return true;
}
/**
 * Check whether this XATDataValueImp is an atomic value.
 * @return boolean True: This XATDataValue is an atomic node. False: otherwise.
 * @since 1.0
 */

public boolean isNode() {
    if (value != null && value instanceof Node)
        return true;
    else
        return false;
}
/**
 * Remove the duplicated values in a vector.
 * The comparison is done by values.
 * @return void
 * @since 1.0
 */
public void removeDuplicates() {
    if (!isCollection())
        return;

    if (this.collection.size() == 1)
        return;
    Hashtable index = new Hashtable();
    // to keep the order, we use hashtable.
    for (int i = 0; i < this.collection.size(); i++) {
        XATDataValue obj = (XATDataValue) this.collection.elementAt(i);
        if (index.get(obj.toString()) != null) {
            this.collection.removeElementAt(i);
            i--;

        } else {
            index.put(obj.toString(), obj);
        }
    }
}
/**
 * Convert the value of this XATDataValueImp into a String with XML format.
 * @return String The convert result.
 * @since 1.0
 */

public String toString() {
	if(value != null)
		return toXML(value);
	else
		return toXML(collection);
}
/**
 * Convert the value of this XATDataValueImp into a String with XML format.
 * @return String The convert result.
 * @since 1.0
 */

public String toXML() {
    return toXML(value);
}
/**
 * Convert the given object into a String with XML format.
 * @return String The convert result.
 * @param o The given object will be converted.
 * @since 1.0
 */
public String toXML(Object o) {

    if (o == null)
        return "";

    if (o instanceof Vector) {
        StringBuffer sb = new StringBuffer();
        for (Enumeration e = ((Vector) o).elements(); e.hasMoreElements();) {
            sb.append(toXML(e.nextElement()));
        }
        return sb.toString();
    }
    if (o instanceof Element) {
        try {
            Document dDomTree = ((Element) o).getOwnerDocument();
            OutputFormat format = new OutputFormat(dDomTree);
            StringWriter stringOut = new StringWriter();
            Serializer serial = new XMLSerializer();
            serial.setOutputFormat(format);
            serial.setOutputCharStream(stringOut);
            DOMSerializer ds = serial.asDOMSerializer();
            ds.serialize((Element) o);
            String s = stringOut.toString();
            if (s.substring(0, 5).equals("<?xml")) {
                int i = s.indexOf('\n');
                s = s.substring(i + 1);
            }
            return s;
        } catch (IOException e) {
            e.printStackTrace();
            return "";
        }
    }

    return o.toString();
}
}
