package edu.wpi.cs.dsrg.xmldb.xat.common.xatnode;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.io.ByteArrayInputStream;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.Vector;

import org.apache.xerces.parsers.DOMParser;
import org.apache.xml.serialize.DOMSerializer;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.Serializer;
import org.apache.xml.serialize.XMLSerializer;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableColumnNoExistException;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableEmptyException;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableException;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableOutOfRangeException;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableSchemaNoMatchException;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableTupleNoMatchException;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.Visitor;

/**
 * The class defines the data model of the XAT, which is a table of XML fragments.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public class XATTableImp extends XATDataSourceImp implements XATTable
{
	/**
	 * The name of XATTable.
	 * @since 1.0
	 */
	protected String sTableName;
	/**
	 * The name of columns within XATTable.
	 * @since 1.0
	 */
	protected Vector vColNames;
	/**
	 * The content of columns within XATTable.
	 * @since 1.0
     */
	protected Vector vColList;
/**
 * The default XATTableImp Constructor.
 * @since 1.0
 */
public XATTableImp() {
    sTableName = null;
    vColNames = new Vector();
    vColList = new Vector();
}
/**
 * The XATTableImp Constructor.
 * @param xtabInit The XATTable to initialize this XATTable.
 * @since 1.0
 */

public XATTableImp(XATTable xtabInit) {
    sTableName = null;
    vColNames = new Vector();
    vColList = new Vector();

    XATTable xtabTemp = (XATTable) xtabInit.deepCloneDataSource();
    int iNumCols = xtabTemp.getNumberOfColumns();

    for (int i = 1; i <= iNumCols; i++) {
        try {
            vColNames.addElement(xtabTemp.getColumnName(i));
            vColList.addElement(xtabTemp.getColumn(i));
        } catch (XATTableOutOfRangeException e) {

        }
    }
    sTableName = xtabInit.getName();
}
/**
 * The XATTableImp constructor.
 * @param sName The Table name.
 * @param oValue The table content.
 * @since 1.0
 */
public XATTableImp(String sName, Object oValue) {
    vColList = new Vector();
    sTableName = null;
    Vector vColumn = new Vector();
    vColumn.setElementAt(oValue, 0);
    // FIXME - this.appendColumn(sName,vColumn);
}
/** 
 * This method is used to implement a visitor pattern.
 * @return void
 * @param visitor
 * @since 1.0
 */
public void accept(Visitor visitor) {
}
/**
 * Append an empty column to the end of the table.
 * When we append the column, we have to make sure the column never exists in the XAT table before,
 * Otherwise, delete the original column, and print out an error message.
 * @return int The column position of the appended new column.
 * @param new_column The new column going to be appended.
 * @throws XATTableTupleNoMatchException
 * @since 1.0
 */
public int appendColumn(String sColumnName)
    throws XATTableTupleNoMatchException {
    int colPos = 0;
    Vector vNewColumn = new Vector();
    for (int i = 0; i < getNumberOfTuples(); i++)
        vNewColumn.addElement(new XATDataValueImp());

    if (this.getColNames().contains(sColumnName)) {
        // if exists, warning, and then make it an empty column.
        if (isDebug()) {
            System.out.println(
                "WARNING: XAT Table has duplicated column names, the original column ("
                    + sColumnName
                    + ")has been overlapped!");
        }
        try {
            this.updateColumn(this.getColNames().indexOf(sColumnName) + 1, vNewColumn);
        } catch (XATTableOutOfRangeException oe) {
            System.out.println("!??! It shouldn't be!");
            oe.printStackTrace(System.err);
        } catch (XATTableEmptyException ee) {
            System.out.println("!??! It shouldn't be!");
            ee.printStackTrace(System.err);
        }
    } else {
        vColNames.addElement(sColumnName);
        vColList.addElement(vNewColumn);
    }
    return vColList.size();
}
/**
 * Append a new column to the end of the table.
 * @return int The column position of the new appended column.
 * @param sName The name of the new appended column.
 * @param new_column The new column going to be appended.
 * @since 1.0
 */

public int appendColumn(String sName, Vector new_column)
    throws XATTableTupleNoMatchException {

    validate(new_column);
    if (vColList.size() > 0) {
        if (new_column.size() != ((Vector) vColList.elementAt(0)).size()) {
            throw new XATTableTupleNoMatchException();
        }
    }
    vColNames.addElement(sName);
    // have to clone the vector and put it into the table.
    vColList.addElement(new_column.clone());
    return vColList.size();
}
/**
 * Append this table to given table.
 * @return void  
 * @param sour The source table will be appended.
 * @since 1.0
 */
public void appendTable(XATTable sour) {
    for (int i = 1; i <= sour.getNumberOfTuples(); i++) {
        this.appendTuple(sour.getTuple(i));

    }
}
/**
 * Append a new tuple to the end of the table.
 * @return int The position of the new appended tuple.
 * @param new_tuple The new tuple going to be appended.
 * @since 1.0
 */
public int appendTuple(Vector new_tuple)
    throws XATTableSchemaNoMatchException {

    validate(new_tuple);
    if (vColList.size() > 0) {
        if (new_tuple.size() != vColList.size()) {
            throw new XATTableSchemaNoMatchException();
        }
    }
    int iNumCols = vColList.size();

    for (int i = 0; i < iNumCols; i++) {

        ((Vector) vColList.elementAt(i)).addElement(new_tuple.elementAt(i));
    }

    return ((Vector) vColList.elementAt(0)).size();
}
/**
 * Construct a new XATTable which has the same column names as this XATTable.
 * @return XATTable The new constructed XATTable.
 * @since 1.0
 */
public XATTable cloneSchema() {
	XATTableImp clone = new XATTableImp();
	clone.vColNames = (Vector)this.vColNames.clone();
	for(int i = 0; i < vColList.size(); i++){
		clone.vColList.addElement(new Vector());
	}
	return clone;
}
/**
 * Remove duplicates from a collection.
 * @return Vector The result Vector.
 * @param vecDis The Vector clearing for.
 * @since 1.0
 */
public static Vector collectionDistinct(Vector vecDis) {
    Vector retVector = new Vector();

    for (int i = 0; i < vecDis.size(); i++) {
        //add only items not in the return vector to the return vector
        if (!retVector.contains(vecDis.elementAt(i))) {
            retVector.add(vecDis.elementAt(i));
        }
    }

    return retVector;
}
/**
 * Convert the given object as a DOM object.
 * @return Object The result DOM object.
 * @param o The given object need to be converted.
 * @since 1.0
 */
private Object convertToDOM(Object o) {
    if (o instanceof Vector) {
        Vector v = (Vector) o;
        for (int i = 0; i < v.size(); i++) {
            v.setElementAt(convertToDOM(v.elementAt(i)), i);
        }
        return v;
    } else
        if (o instanceof String) {
            /* convert it... */
            try {
                byte[] ba = ((String) o).getBytes();
                ByteArrayInputStream bais = new ByteArrayInputStream(ba);
                InputSource is = new InputSource(bais);
                DOMParser parser = new DOMParser();
                parser.parse(is);
                Document dDocument = parser.getDocument();
                Element e = dDocument.getDocumentElement();
                return e;
            } catch (Exception e) {
                return o;
            }
        }
    return o;
}
/**
 * Convert the given object as a String object.
 * @return Object The result String object.
 * @param o The given object need to be converted.
 * @since 1.0
 */
 
private Object convertToString(Object o) {
    if (o instanceof Vector) {
        Vector v = (Vector) o;
        for (int i = 0; i < v.size(); i++) {
            v.setElementAt(convertToString(v.elementAt(i)), i);
        }
        return v;
    } else
        if (o instanceof Element) {
            try {
                Document dDOMTree = ((Element) o).getOwnerDocument();
                OutputFormat format = new OutputFormat(dDOMTree); //Serialize DOM
                StringWriter stringOut = new StringWriter(); //Writer will be a String
                Serializer serial = new XMLSerializer();
                serial.setOutputFormat(format);
                serial.setOutputCharStream(stringOut);
                DOMSerializer ds = serial.asDOMSerializer();
                ds.serialize((Element) o);
                String s = stringOut.toString();
                if (s.substring(0, 5).equals("<?xml")) {
                    int i = s.indexOf('\n');
                    s = s.substring(i + 1);
                }

                return s;
            } catch (Exception e) {

            }
        } else
            if (o instanceof Attr) {
                return ((Attr) o).getValue();
            }
    return o;
}
/**
 * Delete all the tuples in this XATTable.
 * @return void
 * @since 1.0
 */
public void deleteAllTuples() {
    for (Enumeration e = vColList.elements(); e.hasMoreElements();) {
        try {
            Vector vColumn = (Vector) e.nextElement();
            vColumn.setSize(0);
        } catch (ClassCastException ex) {
        }
    }
}
/**
 * Delete a column from the specific position.
 * Note: Col Index starts from 1.
 * @return Vector The content of the deleted column.
 * @param position The position of column going to be deleted.
 * @since 1.0
 */

public Vector deleteColumn(int position)
    throws XATTableOutOfRangeException, XATTableEmptyException {

    if (vColList.size() == 0) {
        throw new XATTableEmptyException();
    }

    Vector vTemp = new Vector();
    if (position > 0 && position <= vColList.size()) {
        // save the column at the specified position
        vTemp = (Vector) vColList.elementAt(position - 1);
        // remove the column at the position
        vColList.removeElementAt(position - 1);
        // removes the column name from the vector
        vColNames.removeElementAt(position - 1);
    } else {
        throw new XATTableOutOfRangeException();
    }

    return vTemp;

}
    /**
     * Delete a tuple of the specified position.
     * @return Vector The content of the deleted tuple.
     * @param position The position of the tuple going to be deleted.
     * @since 1.0
     */

       public Vector deleteTuple(int position) throws XATTableOutOfRangeException, XATTableEmptyException
	{
		if(vColList.size()==0)
		{
			throw new XATTableEmptyException();
		}

                int iNumCols = vColList.size();
		Vector vTemp = new Vector();
		if(position>0 && position<=((Vector)vColList.elementAt(0)).size())
		{
			for( int i=0; i<iNumCols ;i++)
			{
				// save the cell at the specified position
				vTemp.addElement(((Vector)vColList.elementAt(i)).elementAt(position-1));
				// remove the cell at the position
				((Vector)vColList.elementAt(i)).removeElementAt(position-1);
			}
		}
		else
		{
			throw new XATTableOutOfRangeException();
		}

		return vTemp;



	}
/**
 * Convert the XML element with the cell of XAT into a String.
 * @return String The result string.
 * @param o The element will be converted.
 * @since 1.0
 */
public static String elementToString(Element o) {
    try {
        Document dDOMTree = ((Element) o).getOwnerDocument();
        OutputFormat format = new OutputFormat(dDOMTree); //Serialize DOM
        StringWriter stringOut = new StringWriter(); //Writer will be a String
        Serializer serial = new XMLSerializer();
        serial.setOutputFormat(format);
        serial.setOutputCharStream(stringOut);
        DOMSerializer ds = serial.asDOMSerializer();
        ds.serialize((Element) o);
        String s = stringOut.toString();
        if (s.substring(0, 5).equals("<?xml")) {
            int i = s.indexOf('\n');
            s = s.substring(i + 1);
        }

        return s;
    } catch (Exception e) {
        e.printStackTrace(System.err);
        return o.toString();
    }
}
	/**
     * Get the vector containing all the columns content within the XATTable.
     * @return Vector The vector containing the columns content.
     * @since 1.0
     */
	public Vector getColList() {
		return vColList;
	}
    /**
     * Get the vector containing all the column names within the XATTable.
     * @return Vector The Vector containing the column names.
     * @since 1.0
     */
	public Vector getColNames() {
		return (Vector)vColNames.clone();
	}
    /**
     * Get one column content from the specified position.
     * @return a Vector of cells in one column.
     * @param The position of the column start from 1.
     */
    // updated 01/13/02
    public Vector getColumn(int position) throws XATTableOutOfRangeException
	{
		if(position>0 && position<=vColList.size())
		{
			Vector vTemp = (Vector)vColList.elementAt(position-1);
			return vTemp;
		}
		else
		{
			throw new XATTableOutOfRangeException();
		}

	}
    /**
     * Get the column name for a specific column position.
     * @return column Name for a give position.
     * @param The column position.
     */
    // updated 01/13/02
    public String getColumnName(int position) throws XATTableOutOfRangeException
	{
		if(position> vColNames.size())
		{
			throw new XATTableOutOfRangeException();
		}
		else
		{
			return (String)vColNames.elementAt(position-1);
		}
	}
/**
 * Get the Position of a column with the given Column Name.
 * Note: The column is not case sensitive.
 * @return int The position of the column.
 * @param column_name The column name searching for.
 * @since 1.0
 */

public int getColumnPosition(String column_name)
    throws XATTableColumnNoExistException {
    for (int i = 0; i < vColNames.size(); i++) {
        if (column_name
            .toUpperCase()
            .equals(((String) vColNames.elementAt(i)).toUpperCase())) {
            return i + 1;
        }
    }
    throw new XATTableColumnNoExistException();
}
/**
 * This method takes two XATTables, finds all matching columns. 
 * @return Vector The vector containning all common columns of two XATTable.
 * @param tab2 The another source XATTable.
 * @since 1.0
 */
public Vector getCommonColumnNames(XATTable tab2) {

    int iColumnCount = this.getNumberOfColumns();
    Vector outColNames = this.getColNames();
    
    int inColumns = tab2.getNumberOfColumns();
    Vector inColNames = tab2.getColNames();
    
    Vector comColNames = new Vector();
    
    //First find all common columns
    for (int l = 0; l < inColumns; l++) {
        for (int m = 0; m < iColumnCount; m++) {
            if (inColNames.elementAt(l).equals(outColNames.elementAt(m))) {
                //found a matching column name
                if (comColNames.indexOf(outColNames.elementAt(m)) < 0) {
                    comColNames.add(outColNames.elementAt(m));
                }
            } 
        }
    }

    return comColNames;
}
/**
 * Get the Name of the XATTable.
 * @return String The name of the XATTable.
 * @since 1.0
 */

public String getName() {
    return sTableName;
}
/** 
 * Get total number of columns in this table.
 * @return int The total number of columns in this XATTable.
 * @since 1.0
 */

public int getNumberOfColumns() {
    return vColList.size();
}
/** 
 * Get total number of tuples in this table.
 * @return int The total number of tuples with this XATTable.
 * @since 1.0
 */

public int getNumberOfTuples() {
    if (vColList.size() > 0) {
        return ((Vector) vColList.elementAt(0)).size();
    }
    return 0;
}
/**
 * Get the vector containing the columns contents.
 * @return Vector The Vector containing the columns contents.
 * @since 1.0
 */
public Vector getRowList() {
    Vector result = new Vector();
    try {
        for (int i = 1; i <= this.getNumberOfTuples(); i++) {
            result.addElement(this.getTuple(i));
        }
    } catch (Exception e) {
        e.printStackTrace();
    }
    return result;
}
/**
 * Get a tuple in a specified position.
 * @return Vector The Vector of cells in one tuple.
 * @param position The position of the tuple start from 1.
 * @since 1.0
 */

public Vector getTuple(int position)
    throws XATTableEmptyException, XATTableOutOfRangeException {
    if (vColList.size() == 0) {
        throw new XATTableEmptyException();
    }

    Vector vTuple = new Vector();
    int iNumCols = vColList.size();

    if (position > 0 && position <= ((Vector) vColList.elementAt(0)).size()) {
        for (int i = 0; i < iNumCols; i++) {
            vTuple.addElement(((Vector) vColList.elementAt(i)).elementAt(position - 1));
        }
    } else {
        throw new XATTableOutOfRangeException();
    }

    return vTuple;

}
/**
 * This method takes two XATTables, and finds all un-matching columns. 
 * @return Vector Containing all the un-matching columns.
 * @param tab2 The another source XATTable.
 * @since 1.0
 */
public Vector getUnCommonColumnNames(XATTable tab2) {

    int iColumnCount = this.getNumberOfColumns();
    Vector outColNames = this.getColNames();

    int inColumns = tab2.getNumberOfColumns();
    Vector inColNames = tab2.getColNames();

    Vector comColNames = new Vector();

    //First find all uncommon columns
    for (int l = 0; l < iColumnCount; l++) {
        if (inColNames.indexOf(outColNames.elementAt(l)) < 0) {
            comColNames.add(outColNames.elementAt(l));
        }
    }
    for (int l = 0; l < inColumns; l++) {
        if (outColNames.indexOf(inColNames.elementAt(l)) < 0) {
            comColNames.add(inColNames.elementAt(l));
        }
    }

    return comColNames;
}
/**
 * Get a cell by column and tuple positions.
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATValue - The value in the cell 
 *   									specified by column_positon and tuple_positon.
 * @param column_position The column position of the cell we are looking for.
 * @param tuple_position  The tuple position of the cell we are looking for.
 * @throws XATTableOutOfRangeException
 * @throws XATTableEmptyException
 * @since 1.0
 */
public XATValue getValue(int column_position, int tuple_position)
    throws
        edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableEmptyException,
        edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableOutOfRangeException {

    if (vColList.size() == 0) {
        throw new XATTableEmptyException();
    }
    if (column_position > 0 && column_position <= vColList.size()) {
        if (tuple_position > 0
            && tuple_position <= ((Vector) vColList.elementAt(0)).size()) {
            return (XATValue) ((Vector) vColList.elementAt(column_position - 1)).elementAt(
                tuple_position - 1);
        } else {
            throw new XATTableOutOfRangeException();
        }
    } else {
        throw new XATTableOutOfRangeException();
    }

}
/**
 * Insert a new column to the table before specific position.
 * @return int The number of Columns after the insertion.
 * @param position The position for new inserted column.
 * @param sName The new column name.
 * @param new_column The new column content.
 * @since 1.0
 * @throws XATTableOutOfRangeException
 * @throws XATTableTupleNoMatchException
 */

public int insertColumn(int position, String sName, Vector new_column)
    throws XATTableOutOfRangeException, XATTableTupleNoMatchException {
    validate(new_column);
    if (position > 0 && position <= vColList.size()) {
        if (new_column.size() != ((Vector) vColList.elementAt(0)).size()) {
            throw new XATTableTupleNoMatchException();
        }

        vColList.insertElementAt(new_column, position - 1);
        vColNames.insertElementAt(sName, position - 1);
    } else {
        throw new XATTableOutOfRangeException();
    }
    return vColList.size() - position;
}
/**
 * Insert a new tuple to the table before specific position.
 * @return int The number of tuples after the insertion.
 * @param positoin The inserted tuple position.
 * @param new_tuple The new tuple content.
 * @since 1.0
 */

public int insertTuple(int position, Vector ntuple)
    throws XATTableOutOfRangeException, XATTableSchemaNoMatchException {
    validate(ntuple);
    // clone the tuple.
    Vector new_tuple = (Vector) ntuple.clone();

    if (new_tuple.size() > vColList.size()) {
        throw new XATTableSchemaNoMatchException();
    }

    int iNumCols = vColList.size();
    if (position == 1 && getNumberOfTuples() == 0) {
        // insert a new tuple in an empty table.
        appendTuple(new_tuple);
    } else
        if (position > 0 && position <= ((Vector) vColList.elementAt(0)).size()) {
            for (int i = 0; i < iNumCols; i++) {
                ((Vector) vColList.elementAt(i)).insertElementAt(
                    new_tuple.elementAt(i),
                    position - 1);
            }
        } else {
            throw new XATTableOutOfRangeException();
        }

    return ((Vector) vColList.elementAt(0)).size() - (position);

}
/**
 * Get the value of the propertity 'DEBUG_NavTaggerCancelout'.
 * @return boolean True: With debug information. False: No debug information.
 * @since 1.0
 */
public static boolean isDebug() {
	return System.getProperty("DEBUG_XATTableImp")!=null;
}
/**
 * Merge the column names of two XATTables.
 * This method takes two XATTables, and try to find all matching columns, adds those to the head of a vector,
 *  and appends all other column names at the end. 
 * @return java.util.Vector The result column name vector.
 * @param tab2 The another source XATTable.
 * @since 1.0
 */
public Vector mergeColumnNames(XATTable tab2) {

    int iColumnCount = this.getNumberOfColumns();
    int iTupleCount = this.getNumberOfTuples();
    Vector outColNames = this.getColNames();
    int inColumns = tab2.getNumberOfColumns();

    Vector inColNames = tab2.getColNames();
    int inputTupleCount = tab2.getNumberOfTuples();

    Vector comColNames = new Vector();
    Vector fColNames = new Vector();

    //First find all common columns
    for (int l = 0; l < inColumns; l++) {
        for (int m = 0; m < iColumnCount; m++) {
            if (inColNames.elementAt(l).equals(outColNames.elementAt(m))) {
                //found a matching column name
                if (comColNames.indexOf(outColNames.elementAt(m)) < 0) {
                    comColNames.add(outColNames.elementAt(m));
                }
            } else {
                if (fColNames.indexOf(outColNames.elementAt(m)) < 0) {
                    fColNames.add(outColNames.elementAt(m));
                }
            }
        }
        if (fColNames.indexOf(inColNames.elementAt(l)) < 0) {
            fColNames.add(inColNames.elementAt(l));
        }
    }

    //System.out.println(comColNames + "" + fColNames);

    int newSize = fColNames.size();
    //remove all common column names, these will be added to the beginning of fColNames
    for (int k = 0; k < comColNames.size(); k++) {
        for (int l = 0; l < newSize; l++) {
            if (comColNames.elementAt(k).equals(fColNames.elementAt(l))) {
                fColNames.removeElementAt(l);
                newSize--;
                l--;
            }
        }
    }

    //System.out.println(comColNames + "" + fColNames);

    //add common items to head of final column names
    if (comColNames.size() != 0) {
        for (int k = 0; k < comColNames.size(); k++) {
            if (k < newSize) {
                fColNames.insertElementAt(comColNames.elementAt(k), k);
            } else {
                //if there are no uncommon columns, just append to fColNames
                fColNames.add(comColNames.elementAt(k));
            }
        }
    }

    return fColNames;
}
/**
 * Move a column to a specific position.
 * @return void
 * @param from_position Specify the position of column to be moved.
 * @param to_position Identfiy the dest position moved to.
 * @throws XATTableOutOfRangeException
 * @throws XATTableEmptyException
 * @throws XATTableTupleNoMatchException
 * @since 1.0
 */

public void moveColumn(int from_position, int to_position)
    throws XATTableOutOfRangeException, XATTableEmptyException {

    try {
        String sColName = getColumnName(from_position);
        // deletes the column from the position
        Vector vTemp = new Vector(deleteColumn(from_position));
        // inserts the tuple into the new position
        insertColumn(to_position, sColName, vTemp);
        //setColumnName(sColName, to_position);
    } catch (XATTableOutOfRangeException e) {
        throw e;
    } catch (XATTableEmptyException e) {
        throw e;
    } catch (XATTableTupleNoMatchException e) {

    }

}
/**
 * Move a tuple to a specific position.
 * @return void
 * @param from_position The position of tuple to be moved.
 * @param to_position Identify the dest position moved to.
 * @since 1.0
 */

public void moveTuple(int from_position, int to_position)
    throws XATTableOutOfRangeException, XATTableEmptyException {

    try {
        // deletes the tuple from the position
        Vector vTemp = deleteTuple(from_position);
        // inserts the tuple into the new position
        insertTuple(to_position, vTemp);
    } catch (XATTableOutOfRangeException e) {
        throw e;
    } catch (XATTableEmptyException e) {
        throw e;
    } catch (XATTableSchemaNoMatchException e) {

    }

}
/**
 * Prints the XATTable
 * @return void 
 * @since 1.0
 */
public void print() {
    if (vColList.size() > 0) {
        int iNumCols = vColList.size();
        int iNumTuples = ((Vector) vColList.elementAt(0)).size();
        for (int j = 0; j < iNumTuples; j++) {
            for (int k = 0; k < iNumCols; k++) {
                System.out.print(
                    ((String) ((Vector) vColList.elementAt(k)).elementAt(j)) + "             ");
            }
            System.out.println("");
        }
    }

}
/** 
 * Set Column Name in the specified position.
 * @return void
 * @param newName The new name for column.
 * @param position The position for this new column name.
 * @since 1.0
 */

public void setColumnName(String newName, int position)
    throws XATTableOutOfRangeException {
    if (vColNames.size() < position || position <= 0)
        throw new XATTableOutOfRangeException();
    else
        vColNames.setElementAt(newName, position - 1);
}
/**
 * Set the Name of the Table.
 * @return The old name of the XATTable.
 * @param newName The new name of the XATTable.
 * @since 1.0
 */

public String setName(String newName) {
    String sTemp = sTableName;
    sTableName = newName;
    return sTemp;
}
/**
 * Convert the XATTable into a String for output.
 * @return String The result string.
 * @since 1.0
 */
public String toString() {
    int iNumCols = vColList.size();
    int iNumTuples = this.getNumberOfTuples();
    StringBuffer sOutput = new StringBuffer();
    Vector vRow = new Vector();
    String sAddToOutput = "";

    //place table name at top
    sOutput.append("Table Name: " + sTableName + "\n");

    //place column names
    for (int n = 1; n <= iNumCols; n++) {
        try {
            sOutput.append("Column(" + getColumnName(iNumCols - n + 1) + ") | ");
        } catch (XATTableOutOfRangeException e) {
        }
    }
    sOutput.append("\n");

    try {
        //place column info
        for (int i = 1; i <= iNumTuples; i++) {

            vRow = this.getTuple(i);

            for (int j = 0; j < iNumCols; j++) {
                sOutput.append(vRow.elementAt(iNumCols - j - 1) + " | ");

            }
            sOutput.append("\n");
        }
    } catch (XATTableOutOfRangeException e) {
    } catch (XATTableEmptyException e1) {
    }

    //return string
    return sOutput.toString();

}
/**
 * Update a column in the specified position.
 * @return Vector The old column.
 * @param position The position of updated column.
 * @param new_column The new column.
 * @since 1.0
 */

public Vector updateColumn(int position, Vector new_column)
    throws
        XATTableOutOfRangeException,
        XATTableEmptyException,
        XATTableTupleNoMatchException {
    validate(new_column);
    if (vColList.size() == 0) {
        throw new XATTableEmptyException();
    }
    if (new_column.size() > ((Vector) vColList.elementAt(0)).size()) {
        throw new XATTableTupleNoMatchException();
    }

    Vector vTempCol = new Vector();

    if (position > 0 && position <= vColList.size()) {
        // saves the original column
        vTempCol = ((Vector) vColList.elementAt(position - 1));
        // set the column to the new column
        vColList.setElementAt(new_column, position - 1);
    } else {
        throw new XATTableOutOfRangeException();
    }

    return vTempCol;
}
/**
 * Update a tuple.
 * @return Vector The old tuple updated. 
 * @param position The position of tuple to be updated.
 * @param new_tuple The new tuple.
 * @since 1.0
 */

public Vector updateTuple(int position, Vector new_tuple)
    throws
        XATTableOutOfRangeException,
        XATTableEmptyException,
        XATTableSchemaNoMatchException {
    validate(new_tuple);
    if (vColList.size() == 0) {
        throw new XATTableEmptyException();
    }
    if (new_tuple.size() > vColList.size()) {
        throw new XATTableSchemaNoMatchException();
    }

    Vector vTempTuple = new Vector();
    int iNumCols = vColList.size();

    if (position > 0 && position <= ((Vector) vColList.elementAt(0)).size()) {
        for (int i = 0; i < iNumCols; i++) {
            // saves the cell in the new vector
            vTempTuple.addElement(((Vector) vColList.elementAt(i)).elementAt(position - 1));
            // sets the new value for the cell
            ((Vector) vColList.elementAt(i)).setElementAt(
                new_tuple.elementAt(i),
                position - 1);
        }
    } else {
        throw new XATTableOutOfRangeException();
    }

    return vTempTuple;

}
/**
 * Update a cell with a specific object.
 * @return XATValue The old value in the cell.
 * @param column_position The Column position of the cell.
 * @param tuple_position The Tuple position of the cell.
 * @param new_value The new value of the cell.
 * @since 1.0
 */
public XATValue updateValue(
    int column_position,
    int tuple_position,
    XATDataValue new_value)
    throws XATTableOutOfRangeException, XATTableEmptyException {
    XATValue oTempRet;
    if (vColList.size() == 0) {
        throw new XATTableEmptyException();
    }
    if (column_position > 0 && column_position <= vColList.size()) {
        if (tuple_position > 0
            && tuple_position <= ((Vector) vColList.elementAt(0)).size()) {
            // saves the cell
            oTempRet =
                (XATValue) ((Vector) vColList.elementAt(column_position - 1)).elementAt(
                    tuple_position - 1);
            // set the new value for the cell
            ((Vector) vColList.elementAt(column_position - 1)).setElementAt(
                new_value,
                tuple_position - 1);
            return oTempRet;
        } else {
            throw new XATTableOutOfRangeException();
        }
    } else {
        throw new XATTableOutOfRangeException();
    }

}
/**
 * Update a cell with a specific object.
 * @return XATValue The old value in the cell.
 * @param column_position The Column position.
 * @param tuple_position The Tuple position.
 * @param new_value The new value of the cell.
 * @since 1.0
 */
public XATValue updateValue(
    int column_position,
    int tuple_position,
    XATValue new_value)
    throws XATTableOutOfRangeException, XATTableEmptyException {
    XATValue oTempRet;
    if (vColList.size() == 0) {
        throw new XATTableEmptyException();
    }
    if (column_position > 0 && column_position <= vColList.size()) {
        if (tuple_position > 0
            && tuple_position <= ((Vector) vColList.elementAt(0)).size()) {
            // saves the cell
            oTempRet =
                (XATValue) ((Vector) vColList.elementAt(column_position - 1)).elementAt(
                    tuple_position - 1);
            // set the new value for the cell
            ((Vector) vColList.elementAt(column_position - 1)).setElementAt(
                new_value,
                tuple_position - 1);
            return oTempRet;
        } else {
            throw new XATTableOutOfRangeException();
        }
    } else {
        throw new XATTableOutOfRangeException();
    }

}
/**
 * This method is used to validate all the items in this Vector is of type XATDataValue.
 * Otherwise, throws runtime exception.
 * @return boolean True: Valid XATTable. False: otherwise.
 * @param vec The Vector tested for.
 * @since 1.0
 */
void validate(Vector vec) {
    if (vec == null)
        return;

    if (vec.size() == 0)
        return;

    try {
        // check if they can be converted in to an array of XATDataValue.
        XATDataValue[] values = (XATDataValue[]) vec.toArray(new XATDataValue[] {
        });
        // if not, an run time exception will be throwed.
    } catch (ArrayStoreException ase) {
        throw new XATTableException("One of the values you want to put into the XATTable is not of type XATDataValue!");
    }
}
}
