package edu.wpi.cs.dsrg.xmldb.xat.component.operatorfactory;

import edu.wpi.cs.dsrg.xmldb.xat.common.expression.Expression;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumn;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumnImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Source;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.SourceImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.Function;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.FunctionImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.CartesianProduct;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.CartesianProductImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Distinct;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.DistinctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupBy;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupByImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.LeftOuterJoin;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.LeftOuterJoinImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.OrderBy;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.OrderByImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Project;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.ProjectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Select;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.SelectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Aggregate;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.AggregateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavCollection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavCollectionImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavUnnest;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavUnnestImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Tagger;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.TaggerImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.Pattern;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
 * Insert the type's description here.
 * @author: Brad Pielech
 */
public class PersistentOperatorBuilder extends OperatorBuilder {
/**
 * StaticOperatorBuilder constructor comment.
 */
protected PersistentOperatorBuilder() {
	super();
}
/**
 * Returns a handle to the operatorBuilder
 * This is a friend method so it can only be invoked by the factory
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.operatorfactory.OperatorBuilder
 */
static OperatorBuilder getOperatorBuilder() {
	if(builder == null){
		builder = new PersistentOperatorBuilder();
	}
	return builder;
}
/**
 * Create a new Aggregate
 * @return Aggregate
 * @param node XATNode
 */
public Aggregate newAggregate(XATNode node) {
	return new AggregateImp(node);
}
/**
 * Create a new CartesianProduct
 * @return CartesianProduct
 * @param node XATNode
 */
public CartesianProduct newCartesianProduct(XATNode node) {
	return new CartesianProductImp(node);
}
/**
 * Create a new Distinct
 * @return Distinct
 * @param node XATNode
 */
public Distinct newDistinct(XATNode node) {
	return new DistinctImp(node);
}
/**
 * Create a new Function
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.Function
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public Function newFunction(XATNode node) {
	return new FunctionImp(node);
}
/**
 * Create a new groupBy
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupBy
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public GroupBy newGroupBy(XATNode node) {
	return new GroupByImp(node);
}
/**
 * Create a new LeftOuterJoin
 * @return LeftOuterJoin
 * @param node XATNode
 * @param exp Expression
 */
public LeftOuterJoin newLeftOuterJoin(XATNode node, Expression exp) {
	return new LeftOuterJoinImp(node, exp);
}
/**
 * Create a new NameColumn
 * @return NameColumn
 * @param node XATNode
 */
public NameColumn newNameColumn(XATNode node) {
	return new NameColumnImp(node);
}
/**
 * Create a new NavCollection operator 
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavCollection
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public NavCollection newNavCollection(XATNode node) {
	return (NavCollection)new NavCollectionImp(node);
}
/**
 * Create a new NavCollection operator 
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavUnnest
 * @param XATNode node the XATNode to associate with this operator
 */
public NavUnnest newNavUnnest(XATNode node) {
	return (NavUnnest) new NavUnnestImp(node);
}
/**
 * Create a new OrderBy
 * @return OrderBy
 * @param node XATNode
 */
public OrderBy newOrderBy(XATNode node) {
	return new OrderByImp(node);
}
/**
 * Create a new Project
 * @return Project
 * @param node XATNode
 */
public Project newProject(XATNode node) {
	return new ProjectImp(node);
}
/**
 * Insert the method's description here.
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Select
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public Select newSelect(XATNode node, Expression exp) {
	return (Select) new SelectImp(node, exp);
}
/**
 * Create a new Source
 * @return Source
 * @param node XATNode
 */
public Source newSource(XATNode node) {
	return new SourceImp(node);
}
/**
 * Insert the method's description here.
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Tagger
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 * @param pattern Pattern 
 */
public Tagger newTagger(XATNode node, Pattern pattern) {
	return (Tagger) new TaggerImp(node, pattern);
}
}
