package edu.wpi.cs.dsrg.xmldb.xat.component.querymerger;

import java.lang.reflect.Array;

import edu.wpi.cs.dsrg.utils.BindingTableComputer;
import edu.wpi.cs.dsrg.utils.RewriteUtilities;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATNodeDoesNotExistException;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.ExposeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumn;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumnImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Source;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.TwoParamVisitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNodeImp;


public class MergeVisitor extends TwoParamVisitor
{
	protected XATTree userXAT;
	protected XATTree viewXAT;
/**
 * Insert the method's description here.
 * 
 */
public MergeVisitor(XATTree userTree, XATTree viewTree) {
		userXAT = userTree;
		viewXAT = viewTree;
	
	
	}
    public Object dispatch()
    {
        //System.out.println(indent() + "NULL");
        return null;
    }
    public Object dispatch(Object obj)
    {
        return null;
    }
 public Object dispatch(Object obj1, Object obj2)
    {
        return null;
    }
/**
 * Insert the method's description here.
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree getUserXAT() {
	return userXAT;
}
/**
 * Insert the method's description here.
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree getViewXAT() {
	return viewXAT;
}
    /**
     *  xtrObj1 will be the user query or upper tree
     *  xtrObj2 will be the mapping query or bottom tree
	 *  Also the binding tables should be merged.
     */

    public void mergeTrees()
    {
		mergeTrees(true);
     }
    /**
     *  xtrObj1 will be the user query or upper tree
     *  xtrObj2 will be the mapping query or bottom tree
	 *  Also the binding tables should be merged.
	 * @param mergeTagger: whether merge will merge all the taggers or not.
     */

    public void mergeTrees(boolean mergeTagger)
    {

        XATNode[] xnaLeaves = userXAT.getLeaves();
        for(int i=0; i<Array.getLength(xnaLeaves); i++)
        {
               visit(xnaLeaves[i].getOperator(), viewXAT);
        }
		// cleanup the merged tree, which is the userXAT.
		RewriteUtilities.cleanXATAfterGeneration(userXAT, mergeTagger);
        // update the userXAT's binding table.
		userXAT.setBindingTable(new BindingTableComputer().computeBindingTable(userXAT));
       return ;
     }
/**
 * Insert the method's description here.
 * 
 * @param newUserXAT edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public void setUserXAT(edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree newUserXAT) {
	userXAT = newUserXAT;
}
/**
 * Insert the method's description here.
 * 
 * @param newViewXAT edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public void setViewXAT(edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree newViewXAT) {
	viewXAT = newViewXAT;
}
/**
* Connect the bottom of the user query to the top of the view query
* Add a Rename that takes the output of the root of the View query and renames the
* Output of the Source s into that.
* Also, get rid of the expose operator for the view tree.
*/
public Object visit(Source s, XATTree viewTree)
    throws XATNodeDoesNotExistException {

    XATNode xnoRename = new XATNodeImp();

    try {
        s.getNode().getParent().setChild(0, xnoRename);
    } catch (XATNodeDoesNotExistException e) {
        throw new XATNodeDoesNotExistException();
    }

    // the rename operator should rename the exposed column into the source name.
    String viewOutput = viewTree.getRoot().getOperator().getOutputColumnName();
    xnoRename.setParent(s.getNode().getParent());
    // here rename from the viewOutput into the source's binding.
    NameColumn ncRename = new NameColumnImp(xnoRename, viewOutput, s.getOutputColumnName());
    xnoRename.setOperator(ncRename);
    XATNode viewRoot = viewTree.getRoot();
    // test if the viewRoot is an expose operator.
    if(viewRoot.getOperator() instanceof ExposeImp){
	    // then, we get rid of this node by setting its first child as the root.
	    viewTree.setRoot(viewRoot.getChild(0));
	    // get the new viewRoot.
	    viewRoot = viewTree.getRoot();
    }
    viewRoot.setParent(xnoRename);
    xnoRename.addChild(viewRoot);

    // validate the merged Tree.
    // userXAT.validate();
    
    return null;
}
}
