package edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.HashSet;
import java.util.Set;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Expose;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.FOR;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Merge;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumn;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.SQLStmt;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Source;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.Function;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.CartesianProduct;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Distinct;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupBy;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Intersection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Join;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.OrderBy;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.OuterUnion;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Project;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Select;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Union;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Aggregate;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Composer;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavCollection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavUnnest;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Tagger;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.XMLUnion;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.NavigationStep;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.InheritableVisitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * This class is used to compute the consumed columns for each type of operator. 
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public class ModifiedVisitor extends InheritableVisitor
{
/**
 * Put all computed modified columns together.
 * @return Object The set or boolean object contains all consumed 
 * column by given operator.
 * @param node The XATNode computing for.
 * @since 1.0
 */
 
private Object aggregateModified(XATNode node) {
    Object o = visit(node);
    if (o instanceof Set) {
        Set s = new HashSet((Set) o);

        XATNode[] children = node.getChildren();

        for (int i = 0; i < children.length; i++) {
            Object childmod = aggregateModified(children[i]);
            if (childmod instanceof Set) {
                s.addAll((Set) aggregateModified(children[i]));
            } else
                if (childmod instanceof Boolean) {
                    return childmod;
                } else
                    return new Boolean(true);
        }
        return s;
    }
    if (o instanceof Boolean) {
        return o;
    }
    return new Boolean(true);
}
public Object dispatch() {
    if (isDebug()) {
        System.out.println("ModifiedVisitor.dispatch(null)");
    }
    return new Boolean(true);
}
public Object dispatch(Object o) {
    if (isDebug()) {
        System.out.println("ModifiedVisitor.dispatch(" + o.getClass() + ")");
    }
    return new Boolean(true);
}
/**
 * Get the value of the propertity "DEBUG_SchemaCleanUp".
 * @return boolean True: With debug information. False: No debug information.
 * @since 1.0
 */
public static boolean isDebug() {
	return System.getProperty("DEBUG_SchemaCleanUp")!=null;
}
/**
 * Construct an set object using given object o.
 * @return Set The constructed object.
 * @param  o   The object used to construct the Set object.
 * @since 1.0
 */

private Set oneElementSet(Object o) {
    Set hs = new HashSet();
    hs.add(o);
    return hs;
}
/**
 * Compute the modified columns for Expose operator.
 * @return Object The object constining modified columns of given operator.
 * @param  e	The Expose operator computing for.
 * @since 1.0
 * @see Expose
 */
 
public Object visit(Expose node) {
    /* unsure of this */
    return oneElementSet(node.getColumnName());
}
/**
 * Compute the modifeed columns for FOR operator.
 * @return Object The object constining modified columns of given operator.
 * @param  f	The FOR operator computing for.
 * @since 1.0
 * @see FOR
 */

public Object visit(FOR node) {
    /*
    	XATTree tree = node.getXtreFunctions();
    	return aggregateModified(tree.getRoot());
    */
    return new HashSet();
}
/**
 * Compute the modified columns for Function operator.
 * @return Object The object constining modified columns of given operator.
 * @param  f	The Function operator computing for.
 * @since 1.0
 * @see Function
 */

public Object visit(Function node) {
    return new HashSet();
}
/**
 * Compute the modified columns for Merge operator.
 * @return Object The object constining modified columns of given operator.
 * @param  m	The Merge operator computing for.
 * @since 1.0
 * @see Merge
 */

public Object visit(Merge node) {
    return new Boolean(true);
}
/**
 * Compute the modified columns for NameColumn operator.
 * @return Object The object constining modified columns of given operator.
 * @param  nc	The NameColumn operator computing for.
 * @since 1.0
 * @see NameColumn
 */

public Object visit(NameColumn node) {
    return new HashSet();
}
/**
 * Compute the modified columns for Source operator.
 * @return Object The object constining modified columns of given operator.
 * @param  ou	The Source operator computing for.
 * @since 1.0
 * @see Source
 */

public Object visit(Source node) {
    return new HashSet();
}
/**
 * Compute the modified columns for SQLStmt operator.
 * @return Object The object constining modified columns of given operator.
 * @param  ou	The SQLStmt operator computing for.
 * @since 1.0
 * @see SQLStmt
 */

public Object visit(SQLStmt node) {
    return new HashSet();
}
/**
 * Compute the modified columns for CartesianProduct operator.
 * @return Object The HashSet constining modified columns of given operator.
 * @param  cp	The CartesianProduct operator computing for.
 * @since 1.0
 * @see CartesianProduct
 */

public Object visit(CartesianProduct node) {
    return new HashSet();
}
/**
 * Compute the modified columns for Distinct operator.
 * @return Object The object constining modified columns of given operator.
 * @param  d	The Distinct operator computing for.
 * @since 1.0
 * @see Distinct
 */
public Object visit(Distinct node) {
    return oneElementSet(node.getInputColumnName());
}
/**
 * Compute the modified columns for GroupBy operator.
 * @return Object The object constining modified columns of given operator.
 * @param  gb	The GroupBy operator computing for.
 * @since 1.0
 * @see GroupBy
 */

public Object visit(GroupBy node) {
    XATTree tree = node.getXATTree();
    return aggregateModified(tree.getRoot());
}
/**
 * Compute the modified columns for Intersection operator.
 * @return Object The object constining modified columns of given operator.
 * @param  i	The Intersection operator computing for.
 * @since 1.0
 * @see Intersection
 */
public Object visit(Intersection node) {
    return new Boolean(true);
}
/**
 * Compute the modified columns for Join operator.
 * @return Object The object constining modified columns of given operator.
 * @param  j	The Join operator computing for.
 * @since 1.0
 * @see Join
 */

public Object visit(Join node) {
    return new Boolean(true);
}
/**
 * Compute the modified columns for OrderBy operator.
 * @return Object The object constining modified columns of given operator.
 * @param  ob	The OrderBy operator computing for.
 * @since 1.0
 * @see OrderBy
 */

public Object visit(OrderBy node) {
    return new Boolean(true);
}
/**
 * Compute the modified columns for OuterUnion operator.
 * @return Object The object constining modified columns of given operator.
 * @param  ou	The OuterUnion operator computing for.
 * @since 1.0
 * @see OuterUnion
 */

public Object visit(OuterUnion node) {
    /* unsure of this */
    return new HashSet();
}
/**
 * Compute the modified columns for Project operator.
 * @return Object The object constining modified columns of given operator.
 * @param  ou	The Project operator computing for.
 * @since 1.0
 * @see Project
 */
 
public Object visit(Project node) {
    return new HashSet();
}
/**
 * Compute the modified columns for Select operator.
 * @return Object The object constining modified columns of given operator.
 * @param  ou	The Select operator computing for.
 * @since 1.0
 * @see Select
 */
 
public Object visit(Select node) {
    return new Boolean(true);
}
/**
 * Compute the modified columns for Union operator.
 * @return Object The object constining modified columns of given operator.
 * @param  ou	The Union operator computing for.
 * @since 1.0
 * @see Union
 */
 
public Object visit(Union node) {
    return new Boolean(true);
}
/**
 * Compute the modified columns for Aggregate operator.
 * @return Object The HashSet constining modified columns of given operator.
 * @param  a	The Aggregate operator computing for.
 * @since 1.0
 * @see Aggregate
 */
 
public Object visit(Aggregate node) {
    return new Boolean(true);
}
/**
 * Compute the modified columns for Composer operator.
 * @return Object 
 * @param  c The Composer operator computing for.
 * @since 1.0
 * @see Composer
 */
 
public Object visit(Composer node) {
    return new HashSet();
}
/**
 * Compute the modified columns for NavCollection operator.
 * @return Object The object constining modified columns of given operator.
 * @param  n	The NavCollection operator computing for.
 * @since 1.0
 * @see NavCollection
 */

public Object visit(NavCollection node) {
    return new HashSet();
}
/**
 * Compute the modified columns for NavigationStep operator.
 * @return Object The object constining modified columns of given operator.
 * @param  n	The NavigationStep operator computing for.
 * @since 1.0
 * @see NavUnnest
 */

public Object visit(NavigationStep node) {
    return new HashSet();
}
/**
 * Compute the modified columns for NavUnnest operator.
 * @return Object The object constining modified columns of given operator.
 * @param  n	The NavUnnest operator computing for.
 * @since 1.0
 * @see NavUnnest
 */
 
public Object visit(NavUnnest node) {
    return new Boolean(true);
}
/**
 * Compute the modified columns for Tagger operator.
 * @return Object The object constining modified columns of given operator.
 * @param  ou	The Tagger operator computing for.
 * @since 1.0
 * @see Tagger
 */

public Object visit(Tagger node) {
    return new HashSet();
}
/**
 * Compute the modified columns for XMLUnion operator.
 * @return Object The object constining modified columns of given operator.
 * @param  ou	The XMLUnion operator computing for.
 * @since 1.0
 * @see XMLUnion
 */
 
public Object visit(XMLUnion node) {
    return new HashSet();
}
/**
 * Compute the modified columns for operator in given XATNode.
 * @return Object The object constining modified columns of given operator.
 * @param  ou	The XATNode computing for.
 * @since 1.0
 * @see XATNode
 */

public Object visit(XATNode node) {
    return visit(node.getOperator());
}
}
