package edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.HashSet;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableOutOfRangeException;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Expose;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.FOR;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumn;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.SQLStmt;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Source;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.Function;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.CartesianProduct;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Distinct;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupBy;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Intersection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Join;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.OrderBy;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.OuterUnion;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Project;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Select;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Union;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Aggregate;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Composer;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavCollection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavUnnest;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Tagger;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.XMLUnion;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.VarStep;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.InheritableVisitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATTable;

/**
 * This class is used to compute the Produce columns for each XATNode. 
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public class ProducedVisitor extends InheritableVisitor {

/**
 * Put all computed produced columns together.
 * @return Set The set contains all produced columns by given operator.
 * @param node The XATNode computing for.
 * @since 1.0
 */
 
private Set aggregateProduced(XATNode node) {
    Set s = new HashSet((Set) visit(node));

    XATNode[] children = node.getChildren();

    for (int i = 0; i < children.length; i++) {
        s.addAll(aggregateProduced(children[i]));
    }
    return s;
}
public Object dispatch() {
    return new HashSet();
}
public Object dispatch(Object o) {
    return new HashSet();
}
/**
* Get the value of the propertity "DEBUG_SchemaCleanUp".
* @return boolean True: With debug information. False: No debug information.
* @since 1.0
*/
public static boolean isDebug() {
    return System.getProperty("DEBUG_SchemaCleanUp") != null;
}
/**
 * Construct an set object using given object o.
 * @return Set The constructed object.
 * @param  o   The object used to construct the Set object.
 * @since 1.0
 */
 
private Set oneElementSet(Object o) {
    Set hs = new HashSet();
    hs.add(o);
    return hs;
}
/**
 * Compute the produced columns for Expose operator.
 * @return Object The object constining produced columns of given operator.
 * @param  e	The Expose operator computing for.
 * @since 1.0
 * @see Expose
 */
 
public Object visit(Expose node) {
    String sName = node.getColumnName();
    return oneElementSet(sName);
}
/**
 * Compute the produced columns for FOR operator.
 * @return Object The object constining produced columns of given operator.
 * @param  f	The FOR operator computing for.
 * @since 1.0
 * @see FOR
 */
 
public Object visit(FOR node) {
    /**
    	XATTree tree = node.getXtreFunctions();
    	return aggregateProduced(tree.getRoot());
    */
    return new HashSet();
}
/**
 * Compute the produced columns for Function operator.
 * @return Object The object constining produced columns of given operator.
 * @param  f	The Function operator computing for.
 * @since 1.0
 * @see Function
 */

public Object visit(Function node) {
    String sName = node.getOutputColumnName();
    return oneElementSet(sName);
}
/**
 * Compute the produced columns for NameColumn operator.
 * @return Object The object constining produced columns of given operator.
 * @param  nc	The NameColumn operator computing for.
 * @since 1.0
 * @see NameColumn
 */

public Object visit(NameColumn node) {
    String sName = node.getNewColumnName();
    return oneElementSet(sName);
}
/**
 * Compute the produced columns for Source operator.
 * @return Object The object constining produced columns of given operator.
 * @param  s	The Source operator computing for.
 * @since 1.0
 * @see Source
 */

public Object visit(Source node) {
    switch (node.getType()) {
        case Source.TABLE :
        case Source.TEMP :
            Set s = new HashSet();
            XATTable table = (XATTable) node.getOutputDataSource();
            int cols = table.getNumberOfColumns();
            for (int i = 1; i <= cols; i++) {
                try {
                    s.add(table.getColumnName(i));
                } catch (XATTableOutOfRangeException e) {
                }
            }
            return s;
        default :
            String sName = node.getOutputColumnName();
            return oneElementSet(sName);

    }
}
/**
 * Compute the produced columns for SQLStmt operator.
 * @return Object The object constining produced columns of given operator.
 * @param  sqlstmt The SQLStmt operator computing for.
 * @since 1.0
 * @see SQLStmt
 */

public Object visit(SQLStmt node) {
    Set s = new HashSet();

    String sql = node.getStmt().toLowerCase();

    String selClause = sql.substring(0, sql.indexOf("from") - 1);

    //now selClause contains the Select Clause of the query with keyword SELECT
    // in the beginning of the string

    selClause = selClause.substring(new String("select").length() + 1);

    //System.out.println("Before distinct " + selClause);

    //now selClause contains just the column names
    // and possible "distinct" keyword
    int dindex = -1;
    if ((dindex = selClause.indexOf("distinct")) > 0)
        selClause = selClause.substring(new String("distinct").length() + 1);
    ;
    // if the distinct keyword was found, it was extracted from the clause

    //System.out.println("After distinct " + selClause);

    StringTokenizer strtok = new StringTokenizer(selClause, ",");

    Vector columns = new Vector();

    while (strtok.hasMoreTokens()) {
        columns.addElement((String) strtok.nextToken());
    }

    //	System.out.println("DEGUG" + columns.toString());
    String colName;
    int pos;

    for (int i = 0; i < columns.size(); i++) {
        String element = columns.elementAt(i).toString();
        colName = element;
        System.out.println(colName);
        pos = element.indexOf("as ");

        if (pos != -1) {
            colName = element.substring(pos);
            colName = colName.substring(new String("as").length() + 1);
        } else {
            pos = colName.indexOf('.');
            if (pos != -1)
                colName = colName.substring(pos + 1);
        }

        //System.out.println("out "+ colName);
        // remove the whitespace around columnName
        colName.trim();
        // remove the unnecessary quotes]

        //System.out.println("trim "+ colName);

        if (colName.charAt(0) == '"' && colName.charAt(colName.length() - 1) == '"')
            colName = colName.substring(1, colName.length() - 1);

        //colName = colName.substring(colName.indexOf('"')+1, colName.lastIndexOf('"')-1);
        if (isDebug()) {
            System.out.println("minus quotes " + colName);
        }
        s.add(colName);
    }

    return s;
}
/**
 * Compute the produced columns for CartesianProduct operator.
 * @return Object The HashSet constining produced columns of given operator.
 * @param  cp	The CartesianProduct operator computing for.
 * @since 1.0
 * @see CartesianProduct
 */
 
public Object visit(CartesianProduct node) {
    return new HashSet();
}
/**
 * Compute the produced columns for Distinct operator.
 * @return Object The object constining produced columns of given operator.
 * @param  d	The Distinct operator computing for.
 * @since 1.0
 * @see Distinct
 */
 
public Object visit(Distinct node) {
    return new HashSet();
}
/**
 * Compute the produced columns for GroupBy operator.
 * @return Object The object constining produced columns of given operator.
 * @param  gb	The GroupBy operator computing for.
 * @since 1.0
 * @see GroupBy
 */
 
public Object visit(GroupBy node) {
    XATTree tree = node.getXATTree();
    return aggregateProduced(tree.getRoot());
}
/**
 * Compute the produced columns for Intersection operator.
 * @return Object The object constining produced columns of given operator.
 * @param  i	The Intersection operator computing for.
 * @since 1.0
 * @see Intersection
 */
 
public Object visit(Intersection node) {
    return new HashSet();
}
/**
 * Compute the produced columns for Join operator.
 * @return Object The object constining produced columns of given operator.
 * @param  j	The Join operator computing for.
 * @since 1.0
 * @see Join
 */

public Object visit(Join node) {
    return new HashSet();
}
/**
 * Compute the produced columns for OrderBy operator.
 * @return Object The object constining produced columns of given operator.
 * @param  ob	The OrderBy operator computing for.
 * @since 1.0
 * @see OrderBy
 */

public Object visit(OrderBy node) {
    return new HashSet();
}
/**
 * Compute the produced columns for OuterUnion operator.
 * @return Object The object constining produced columns of given operator.
 * @param  ou	The OuterUnion operator computing for.
 * @since 1.0
 * @see OuterUnion
 */
 
public Object visit(OuterUnion node) {
    return new HashSet();
}
/**
 * Compute the produced columns for Project operator.
 * @return Object The object constining produced columns of given operator.
 * @param  p	The Project operator computing for.
 * @since 1.0
 * @see Project
 */

public Object visit(Project node) {
    return new HashSet();
}
/**
 * Compute the produced columns for Select operator.
 * @return Object The object constining produced columns of given operator.
 * @param  s	The Select operator computing for.
 * @since 1.0
 * @see Select
 */

public Object visit(Select node) {
    return new HashSet();
}
/**
 * Compute the produced columns for Union operator.
 * @return Object The object constining produced columns of given operator.
 * @param  u The Union operator computing for.
 * @since 1.0
 * @see Union
 */

public Object visit(Union node) {
    return new HashSet();
}
/**
 * Compute the produced columns for Aggregate operator.
 * @return Object The HashSet constining produced columns of given operator.
 * @param  a	The Aggregate operator computing for.
 * @since 1.0
 * @see Aggregate
 */
 
public Object visit(Aggregate node) {
    return new HashSet();
}
/**
 * Compute the produced columns for Composer operator.
 * @return Object 
 * @param  c The Composer operator computing for.
 * @since 1.0
 * @see Composer
 */
 
public Object visit(Composer node) {
    return oneElementSet(node.getOutputColumnName());
}
/**
 * Compute the produced columns for NavCollection operator.
 * @return Object The object constining produced columns of given operator.
 * @param  n	The NavCollection operator computing for.
 * @since 1.0
 * @see NavCollection
 */

public Object visit(NavCollection node) {
    String sName = node.getOutputColumnName();
    return oneElementSet(sName);
}
/**
 * Compute the produced columns for VarStep operator.
 * @return Object The object constining produced columns of given operator.
 * @param  node The VarStep operator computing for.
 * @since 1.0
 * @see VarStep
 */

public Object visit(VarStep node) {
    return new HashSet();
}
/**
 * Compute the produced columns for NavUnnest operator.
 * @return Object The object constining produced columns of given operator.
 * @param  n	The NavUnnest operator computing for.
 * @since 1.0
 * @see NavUnnest
 */
 
public Object visit(NavUnnest node) {
    String sName = node.getOutputColumnName();
    return oneElementSet(sName);
}
/**
 * Compute the produced columns for Tagger operator.
 * @return Object The object constining produced columns of given operator.
 * @param  t The Tagger operator computing for.
 * @since 1.0
 * @see Tagger
 */

public Object visit(Tagger node) {
    return oneElementSet(node.getOutputColumnName());
}
/**
 * Compute the produced columns for XMLUnion operator.
 * @return Object The object constining produced columns of given operator.
 * @param  xu The XMLUnion operator computing for.
 * @since 1.0
 * @see XMLUnion
 */

public Object visit(XMLUnion node) {
    return oneElementSet(node.getOutputColumnName());
}
/**
 * Compute the produced columns for XATNode operator.
 * @return Object The object constining produced columns of given operator.
 * @param  node The XATNode operator computing for.
 * @since 1.0
 * @see XATNode
 */

public Object visit(XATNode node) {
    return visit(node.getOperator());
}
}
