package edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.LeftOuterJoin;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.InheritableVisitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * This class is used to compute the Required Schema.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public class RequiredVisitor extends InheritableVisitor
{
/**
  * The hash table to keep the map between XATNode and its corresponding SchemaCleanUpNode.
  * @since 1.0     
  */

Hashtable hash;

/**
 * Default RequiredVisitor Contructor.
 * @param  hash_init The hash table to keep the map between 
 * XATNode and its corresponding SchemaCleanUpNode.
 * @since 1.0
 */

public RequiredVisitor(Hashtable hash_init) {
    hash = hash_init;
}
public Object dispatch() {
    return null;
}
public Object dispatch(Object o) {
    return o;
}
/**
 * Get all produced columns of given root node.
 * @return Set The set contains all produced columns by given operator.
 * @param node The XATNode computing for.
 * @since 1.0
 */

private Set getAllProduced(XATNode root) {
    SchemaCleanupNode scn = (SchemaCleanupNode) hash.get(root);
    Set s = new HashSet();

    if (scn.produced != null) {
        s.addAll(scn.produced);
    }

    XATNode[] children = root.getChildren();

    for (int i = 0; i < children.length; i++) {
        s.addAll(getAllProduced(children[i]));
    }

    return s;

}
/**
 * Get the value of the propertity "DEBUG_SchemaCleanUp".
 * @return boolean True: With debug information. False: No debug information.
 * @since 1.0
 */
public static boolean isDebug() {
	return System.getProperty("DEBUG_SchemaCleanUp")!=null;
}
/**
 * Compute the SchemaCleanupNode of given XATNode.
 * This method compute the SchemaCleanupNode for a given XATNode, and put the result
 * into hash table.
 * @return Object 
 * @param  node   The XATNode being cleaned.
 * @since 1.0
 * @see SchemaCleanupNode
 */

public Object visit(XATNode node) {
    SchemaCleanupNode scn = (SchemaCleanupNode) hash.get(node);

    Set s = new HashSet();

    if (scn.schema != null) {
        s = new HashSet(scn.schema);
    }
    s.removeAll(scn.produced);
    if (scn.consumed != null)
        s.addAll(scn.consumed);

    //Here we assume all LEFT TREE of LeftOuterJoin will be needed.
    //This is true for LOJ generated in decorrelation.
    //children[0] is the LEFT TREE.
    XATOperator xatop = node.getOperator();
    if (xatop instanceof LeftOuterJoin) {
        XATNode left = node.getChild(0);
        s.addAll(getAllProduced(left));
    }

    scn.required = s;
    return s;
}
}
