package edu.wpi.cs.dsrg.xmldb.xat.component.sqlgeneration;

import edu.wpi.cs.dsrg.utils.BindingTable;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.NavigationStep;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.VarStep;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.InheritableVisitor;
/**
 * Insert the type's description here.
 * @author: Brad Pielech
 */
public abstract class OperatorEvaluator extends InheritableVisitor{ 

protected int heuristicType;
	protected BindingTable variableTable; 
	/**
	* When evaluating an Operator, it is important to make sure that a Source
	* has been visited.  If it hasnt, then the operator cannot go to SQL
	* if it has, then proceed as normal
	*/
	protected boolean sourceVisited;
/**
 * SQLOperatorEvaluator constructor comment.
 */
public OperatorEvaluator(int heuristic, BindingTable varTable) {
	heuristicType = heuristic;
	variableTable = varTable;
	sourceVisited = false;
}
/**
 * SQLOperatorEvaluator constructor comment.
 */
public OperatorEvaluator(BindingTable varTable) {
	variableTable = varTable;
	heuristicType = -1;
	sourceVisited = false;
}
    /**
     * Put the entryPoint and destination step arrays of a Navigate Operator into 1 bigger 
     * array to make it easier to manipulate
     * @return NavigationStep[]
     * @param entryPoint 
     * @param destination 
     */
    protected NavigationStep[] combineSteps(
        NavigationStep[] entryPoint,
        NavigationStep[] destination) {

        int combinedLength = entryPoint.length + destination.length;
        NavigationStep[] combinedPath = new NavigationStep[combinedLength];

        //put in all of the entryPoint steps	
        for (int x = 0; x < entryPoint.length; x++) {
            if (entryPoint[x] != null) {
                combinedPath[x] = entryPoint[x];
            }
        }
        //now put in all the destination steps
        for (int y = entryPoint.length; y < combinedLength; y++) {
            if (destination[y - entryPoint.length] != null) {
                combinedPath[y] = destination[y - entryPoint.length];
            }
        }

        return combinedPath;
    }
    /**
     * createString accepts an array of NavigationSteps and will convert these 
     * steps into one long string.
     * If a VarStep is found, it will be replaced by the actual column name(s) as found
     * in this.variableTable
     * The array is formatted with "/" between steps.  A "/" is added at the end of the string as well
     * @return String corresponding to step
     * @param step NavigationStep from which to create a string 
     */
    public String createString(NavigationStep[] step) {
        String returnString = new String();
        String stepValue = "";
        NavigationStep[] varValue;

        for (int i = 0; i < step.length; i++) {
/*
            if (step[i] instanceof VarStep) {
	            
                varValue = getVariableValue(step[i].getStepValue(), false);
                for (int j = 0; j < varValue.length; j++) {                
                    stepValue = stepValue + varValue[j].getStepValue() + "/";
                }
                

            } else {
            */
                stepValue = step[i].getStepValue() + "/";
//            }

            returnString = returnString + stepValue;
            stepValue = "";

        }

        return returnString;
    }
    /**
     * An alternate definition for createString
     * This definition accepts a string, it then looks up the value in the varTable
     * and returns the representation with "." instead of "/" between steps
     * steps into one long string.
     * If a variable is found, it only looks 1 level deep
     * @return String corresponding to step
     * @param varName String from which to create a string 
     */
    public String createString(String varName) {
        String returnString = new String("");

        NavigationStep[] varValue;
	     
        varValue = getVariableTable().getVariableValue(varName, false);
        for (int j = 0; j < varValue.length; j++) {
	        if(varValue[j] instanceof VarStep){
      	        returnString = returnString + "\"" + varValue[j].getStepValue() + "\".";
	        } else{
		        if(!varValue[j].getStepValue().equals("text()") && !varValue[j].getStepValue().equals("#text")){
			        returnString = returnString + varValue[j].getStepValue() + ".";
		        }
	        }
        }
        //trim the last "."
        returnString = returnString.substring(0, (returnString.length() - 1));
        return returnString;
    }
/**
 * Insert the method's description here.
 * 
 * @param o1 java.lang.Object
 */
public Object dispatch() {
	
	
	return null;
	
	
	}
/**
 * Do nothing
 * 
 * @param o1 java.lang.Object
 */
public Object dispatch(Object o1) {
	
	QueryResult s1 = new QueryResult();
	s1.setcommitQuery(true);
	return s1;
	
	
	}
/*
*	evaluateOperator calls the appropriate visit method for the parameter nextOperator
*   the QueryResult returned from visit method is returned
*/
public QueryResult evaluateOperator(XATOperator nextOperator){ 
	
	return (QueryResult)visit(nextOperator);
}
/**
 * Insert the method's description here.
 * 
 * @param 
 */
public BindingTable getVariableTable() {
	return variableTable;
}
/**
 * Insert the method's description here.
 * 
 * @return boolean
 */
public boolean isSourceVisited() {
	return sourceVisited;
}
/**
 * 
 * @param newVariableTable java.util.Hashtable
 */
public void setHeuristicType(int newHeuristic) {
	heuristicType = newHeuristic;
}
/**
 * Insert the method's description here.
 * 
 * @param newSourceVisited boolean
 */
public void setSourceVisited(boolean newSourceVisited) {
	sourceVisited = newSourceVisited;
}
/**
 * Insert the method's description here.
 * 
 * @param newVariableTable BindingTable
 */
public void setVariableTable(BindingTable newVariableTable) {
	variableTable = newVariableTable;
}
}
