package edu.wpi.cs.dsrg.xmldb.xat.component.sqlgeneration;
import java.util.Vector;

/**
 * <pre>
 * A QueryResult is returned after an operator is evaluated
 * Each result has 3 parameters,
 * 1. Was the operator evaluated
 * 2. Should the incremental statement be committed to the tree
 * 3. The Query Fragment of the evaluated operator
 * </pre>
 * @author: Brad Pielech
 */
public class QueryResult {
	private boolean evaluated;
	private Vector fragmentList;
	private boolean commitQuery;
/**
 * Default constuctor
 */
public QueryResult() {
	evaluated = false;
	commitQuery = false; 
	fragmentList = new Vector();
	
}
/**
 * Constuctor that sets evaluated and the fragment
 */
public QueryResult(boolean eval, QueryFragment newFragment) {
	evaluated = eval;
	commitQuery = false;	  
	fragmentList = new Vector();
	fragmentList.addElement(newFragment);
	
	
}
/**
 * add a fragment to the result
 * 
 * @param newFragment edu.wpi.cs.dsrg.xmldb.xat.component.xatrewrite.QueryFragment
 */
public void addFragment(QueryFragment newFragment) {
	fragmentList.addElement(newFragment);
}
/**
 * returns if the operator was evaluated
 * 
 * @return boolean
 */
public boolean isEvaluated() {
	return evaluated;
}
/**
 * Set whether the operator was evaluated
 * 
 * @param newEvaluated boolean
 */
public void setEvaluated(boolean newEvaluated) {
	evaluated = newEvaluated;
}

/**
 * returns the fragments
 * 
 * @return SQLFragment
 */
public QueryFragment[] getFragment() {
	return (QueryFragment[]) fragmentList.toArray(new QueryFragment[] {});
}

/**
 * Returns if the incremental statement should be committed
 * 
 * @return boolean
 */
public boolean iscommitQuery() {
	return commitQuery;
}

/**
 * Set whether the statement should be committed
 * 
 * @param newcommitQuery boolean
 */
public void setcommitQuery(boolean newcommitQuery) {
	commitQuery = newcommitQuery;
}
}
