package edu.wpi.cs.dsrg.xmldb.xat.component.sqlgeneration;

import java.util.Vector;

import edu.wpi.cs.dsrg.utils.BindingTable;
import edu.wpi.cs.dsrg.xmldb.xat.common.dompattern.ColumnNameNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dompattern.DOMPatternNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dompattern.TagNodeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumnImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.SourceImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.FunctionImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.DistinctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupByImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.JoinImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.OrderByImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.ProjectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.SelectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavUnnestImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Navigate;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.TaggerImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.NavigationStep;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.DOMPattern;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
 * Insert the type's description here.
 * @author: Brad Pielech
 */
public class XQueryOperatorEvaluator extends OperatorEvaluator {
/**
 * XQueryOperatorEvaluator constructor comment.
 * @param heuristic int
 * @param varTable java.util.Hashtable
 */
public XQueryOperatorEvaluator(int heuristic, BindingTable varTable) {
	super(heuristic, varTable);
}
/**
 * XQueryOperatorEvaluator constructor comment.
 * @param varTable BindingTable
 */
public XQueryOperatorEvaluator(BindingTable varTable) {
	super(varTable);
}
/**
 * deleteTagger will iterate over a Tagger operator and determine if the
 * operator can be deleted or not.  A Tagger may be deleted when all of its nodes
 * have been cancelout
 * 
 * @return boolean true if the tagger can be deleted, false otherwise
 * @param rootNode root of the DOMPattern to start iterating over
 */
protected boolean traverseTagger(
    DOMPatternNode rootNode,
    Vector columnName,
    Vector tagName) {

    //if this is a rootNode, then we dont check for cancelout
    if (rootNode instanceof TagNodeImp) {
        tagName.addElement(rootNode.getTagName());
    } else
        if (rootNode instanceof ColumnNameNode) {
            columnName.addElement(rootNode.getTagValue());
        }
    DOMPatternNode[] children = rootNode.getChildren();
    for (int i = 0; i < children.length; i++) {
        return traverseTagger(children[i], columnName, tagName);
    }


return true;
}
    /**
     * Evaluates a Function Operator and determines the SQL equivalent 
     * 
     * @return java.lang.Object
     * @param param Function
     */
    public Object visit(FunctionImp param) {
        QueryResult s1 = new QueryResult();
        if (isSourceVisited()) {
            s1.setEvaluated(true);
            String funName = param.getName();
            NavigationStep[] funParams = (NavigationStep[])param.getParams();
            String paramList = "";
            for (int i = 0; i < funParams.length; i++) {
	            paramList = paramList + funParams[i].getStepValue() + ".";
            }
            //trim the last "."
            paramList = paramList.substring(0, (paramList.length() - 1));
            s1.addFragment(new QueryFragment(QueryTypes.FUNCTION, funName, paramList));
        } else {
            s1.setEvaluated(false);
        }
        return s1;
    }
    /**
     * Evaluates a NameColumn Operator and determines the SQL equivalent of the destination
     * the old column name is return in the name parameter of the fragment, 
     * the new name is the binding
     * 
     * @return java.lang.Object
     * @param param NameColumnImp
     * @deprecated NameColumn operators can now be pushed down so they will not occur in the tree to generate
     * a query from
     */
    public Object visit(NameColumnImp param) {
        QueryResult s1 = new QueryResult();
        s1.setEvaluated(true);
        String newC = param.getNewColumnName();
        String oldC = param.getOriginalColumnName();
        s1.addFragment(new QueryFragment(1, oldC, newC));
        return s1;
    }
/**
 * Evaluates a Source Operator and determines the XQuery equivalent of the destination
 * If a source has already been visited, then add this to the LET, otherwise, add it to FOR
 * 
 * @return java.lang.Object
 * @param param Source
 */
public Object visit(SourceImp param) {
    QueryResult s1 = new QueryResult();
    //set the sourceVisited to TRUE
    if (isSourceVisited()) {

        s1.setEvaluated(true);
        //create a QueryFragment for the LET clause corresponding to the output of the Source
        String name = "document(\"" + param.getName() + "\")";
        String binding = param.getOutputColumnName();
        QueryFragment frag = new QueryFragment(QueryTypes.LET, name, binding);
        s1.addFragment(frag);

    } else {
	    setSourceVisited(true);
		s1.setEvaluated(true);
        //create a QueryFragment for the FOR clause
        String name = param.getName();
        String binding = param.getOutputColumnName();
        QueryFragment frag = new QueryFragment(QueryTypes.FOR, name, binding);
        s1.addFragment(frag);
    }
    return s1;
}

/*****
***** Not sure what to do here
*/

/**
     * Evaluates a Distinct Operator and determines the SQL equivalent of the destination
     * 
     * @return java.lang.Object
     * @param param Distinct
     */
    public Object visit(DistinctImp param) {
        QueryResult s1 = new QueryResult();
        //call the visit method for Distinct's child (Navigate)

        String name = param.getInputColumnName();
        name = "DISTINCT(" + name + ")";
        //create an empty QueryFragment and then return the result object
        QueryFragment frag = new QueryFragment(QueryTypes.SELECT, name);
        s1.addFragment(frag);
        s1.setEvaluated(true);
        return s1;

    }
    /**
     * Evaluates an GroupBy Operator and determines the SQL equivalent of the destination
     * take all of the columns names in the orderBy and add them to the oderBy clause
     * the result will look like GroupBy col1, col2, etc
     *
     * GroupBy returns multiple SQL fragments, 1 for the groupBy and 1 for each of the operators in its subTree
     * Therefore, we first visit the subTree first
     * @return java.lang.Object
     * @param param GroupBy
     */
    public Object visit(GroupByImp param) {
        QueryResult s1 = new QueryResult();
		//first visit the subTree
		/***
		**  traverse the tree better
		*/
		XATTree t1 = param.getXATTree(); 
		XATNode x1 = t1.getRoot();
		QueryResult subTreeR = (QueryResult) visit(x1.getOperator());  		
		 
        if (isSourceVisited()) {
            s1.setEvaluated(true);
            /*
            String[] columns = param.getColumnNames();
            String text = "";
            for (int i = 0; i < columns.length; i++) {
                text = text + columns[i] + ",";
            }
            //chop off the last ,
            if (text.length() > 0) {
                text = text.substring(0, (text.length() - 1));
            }
            */
            QueryFragment frag = new QueryFragment(-1, "");
            s1.addFragment(frag);
        } else {
            s1.setEvaluated(false);
        }
        //before returning s1, add all fragments from the subTreeR into s1
        QueryFragment[] f1 = subTreeR.getFragment();
        for (int i = 0; i < f1.length; i++){
        	s1.addFragment(f1[i]);
        }
        return s1;
    }
    /**
     * Evaluates a Join Operator and determines the SQL equivalent of the destination
     * This adds to the where clause, it will look like Where $a = '10' 
     *
     * Same thing as select
     * @return java.lang.Object
     * @param param Join
     */
    public Object visit(JoinImp param) {
        QueryResult s1 = new QueryResult();
        if (isSourceVisited()) {
            try {
                String exp = (String) param.getExpression().toSQL(this);
                s1.addFragment(new QueryFragment(QueryTypes.WHERE, exp));
                s1.setEvaluated(true);
                return s1;

            } catch (NullPointerException e1) {
                //create an empty QueryFragment and then return the result object
                QueryFragment frag = new QueryFragment(-1, "");
                s1.addFragment(frag);
                return s1;
            }
        } else {
            s1.setEvaluated(false);
        }
        return s1;
    }
    /**
     * Evaluates an OrderBy Operator and determines the SQL equivalent of the destination
     * take all of the columns names in the orderBy and add them to the oderBy clause
     * the result will look like OrderBy col1, col2, etc
     * 
     * @return java.lang.Object
     * @param param OrderBy
     * @deprecated until I figure out what to do with it
     */
    public Object visit(OrderByImp param) {
        QueryResult s1 = new QueryResult();
        s1.setEvaluated(true);
        //create an empty QueryFragment and then return the result object
        String[] columns = param.getColumnNames();
        String text = "";
        for (int i = 0; i < columns.length; i++) {
            text = text + columns[i] + ",";
        }
        //chop off the last ,
        text = text.substring(0, (text.length() - 1));
        QueryFragment frag = new QueryFragment(QueryTypes.ORDERBY, text);
        s1.addFragment(frag);
        return s1;
    }
    /**
     * Evaluates a Project Operator and determines the SQL equivalent of the destination
     * 
     * @return java.lang.Object
     * @param param Project
     */
    public Object visit(ProjectImp param) {
        QueryResult s1 = new QueryResult();
        s1.setEvaluated(true);
        //create an empty QueryFragment and then return the result object
        QueryFragment frag = new QueryFragment(-1, "");
        s1.addFragment(frag);
        return s1;
    }
    /**
     * Evaluates a Select Operator and determines the SQL equivalent of the destination
     * This adds to the where clause, it will look like Where $a = '10' 
     * 
     * @return java.lang.Object
     * @param param Select
     */
    public Object visit(SelectImp param) {
        QueryResult s1 = new QueryResult();
        if (isSourceVisited()) {
            String exp = (String) param.getExpression().toSQL(this);
            //replace all " with blanks
            exp = exp.replace('\"',' ');
            //replace all . with /
            exp = exp.replace('.','/');
            s1.addFragment(new QueryFragment(QueryTypes.WHERE, exp));
            s1.setEvaluated(true);
        } else {
            s1.setEvaluated(false);
        }
        return s1;

    }
/**
 * Evaluates a Navigate Operator and determines the XQuery equivalent of the destination
 * The Navigate's info is just put in a LET binding
 * 
 * @return java.lang.Object
 * @param param Navigate
 */
public Object visit(NavUnnestImp param) {

    Navigate x1 = param;
    QueryResult s1 = new QueryResult();
    if (isSourceVisited()) {
        NavigationStep[] entryPoint = x1.getEntryPoint();
        NavigationStep[] destination = x1.getDestination();
        NavigationStep[] combinedSteps = combineSteps(entryPoint, destination);
        String name = createString(combinedSteps);
        //trim the last /
        if (name.endsWith("/")) {
            name = name.substring(0, (name.length() - 1));
        }

        String binding = x1.getBinding();
        s1.addFragment(new QueryFragment(QueryTypes.LET, name, binding));
        s1.setEvaluated(true);
    } else {
        s1.setEvaluated(false);
    }
    return s1;
}
/**
 * Evaluates a Tagger Operator and determines the XQuery equivalent of the destination
 * 
 * @return java.lang.Object
 * @param param Tagger
 */
public Object visit(TaggerImp param) {
    QueryResult s1 = new QueryResult();
    //call the traverseTagger to determine what this tag contains
    Vector tagNameV = new Vector();
    Vector columnNames = new Vector();
    DOMPattern d1 = (DOMPattern) param.getPattern();
    DOMPatternNode root = d1.getRoot();
    traverseTagger(root, columnNames, tagNameV);
    String tagName = (String)tagNameV.elementAt(0);
    String columns = "";
    //now determine if this tagger has columns as its value or other tags
    for (int i = 0; i < columnNames.size(); i++) {
	    NavigationStep[] n1 = (NavigationStep[])columnNames.elementAt(i);
        String tagVarValue = createString(n1);
        if ((tagVarValue.equals("TAGGER")) || (tagVarValue.equals("AGGREGATE"))) {
            s1.setEvaluated(false); 
            s1.setcommitQuery(true); 
        } else {
            s1.setEvaluated(true);
            columns = columns + tagVarValue + ",";            
        }
    }
        //trim the last comma
    columns = columns.substring(0, (columns.length() - 2));
    String tag = "<" + tagName + "> " + columns + " </" + tagName +">";
    s1.addFragment(new QueryFragment(QueryTypes.INNER_RETURN, tag, columns));

    return s1;
}
}
