package edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
 * Calculates the (true, not weighed) average workload for each operator in the system.
 * This will allow for the system to determine how fully each operator
 * is being used by the scheduler
 * @author: Brad Pielech
 */
public class OperatorUtilization extends StatisticsObject {
/**
 * OperatorUtilization constructor comment.
 */
public OperatorUtilization() {
	super();
}
/**
 * OperatorUtilization constructor comment.
 * @param name java.lang.String
 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
 */
public OperatorUtilization(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) {
	super(name, queryObject);
}
/**
 * Calculates the new Value of the given property of this operator.  This
 * should be overwritten by subclasses who can tailor this method to calculating 1
 * specific property. <p>
 * This method should update at least the local propertyValue and possibly the
 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
 *
 * The procedure for calculating new values should be as follows:
 *
 * <ol>
 * <li> Consult the statsTable for any propertyValues that are needed.
 * <li> Perform the required calculations using the propertyValues from step 1
 * <li> Update the local propertyValue.
 * 
 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
 * </ol> 
 */
public void calculateNewValue() {
	double util = 0;
	/**
	 take the average of the average utilization for all operators in the query plan
	*/
	XATNode[] nodes = ((XATTree)this.getAssociatedQueryObject()).convertToArray();
	for (int i = 0; i < nodes.length; i++){
		XATOperator operator = nodes[i].getOperator();
		util += Double.parseDouble(stats.getValue("NUMBER_OF_TUPLES_DEQUEUED_TOTAL", operator, StatisticsReference.AVERAGE));
	}
	double totalRun = Double.parseDouble(stats.getValue("NUMBER_OF_TIMES_OPERATORS_WERE_RUN", this.getAssociatedQueryObject()));
	util = util / totalRun;
	setPropertyValue("" + util);	
}
}
