package edu.wpi.cs.dsrg.xmldb.xat.testfiles;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableOutOfRangeException;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Expose;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.LeftSemiJoin;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Project;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.RightSemiJoin;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Union;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.InheritableVisitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup.ProducedVisitor;

/**
 * This Class is used to test the Statistics Gatiering component.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */

public class StatisticsGatheringVisitor extends InheritableVisitor {
    Hashtable hash = new Hashtable();

    ProducedVisitor p = new ProducedVisitor();

    static Vector tempTrees = new Vector();

    public static void addTempTree(XATTree tree) {
        tempTrees.add(tree);
    }
    public Object dispatch() {
        return new HashSet();
    }
    public Object dispatch(Object o) {
        return new HashSet();
    }
    public int getNumberOfNodes() {
        return hash.size();
    }
    public int getNumberOfSavedColumns() {
        Set keys = hash.keySet();
        int count = 0;
        for (Iterator i = keys.iterator(); i.hasNext();) {
            try {
                XATNode node = (XATNode) i.next();
                Set columns = (Set) hash.get(node);
                XATTable table = (XATTable) node.getOperator().getOutputDataSource();
                for (int j = 1; j <= table.getNumberOfColumns(); j++) {
                    try {
                        columns.remove(table.getColumnName(j));
                    } catch (XATTableOutOfRangeException e) {
                    }
                }
                //System.err.println(columns.toString());
                count += columns.size();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        return count;
    }
    public static void resetTempTrees() {
        tempTrees = new Vector();
    }
    public Object visit(XATTree tree) {
        return visit(tree.getRoot());
    }
    public Object visit(XATNode node) {
        XATNode[] children = node.getChildren();
        XATOperator op = node.getOperator();
        Set s = new HashSet();
        for (int i = 0; i < children.length; i++) {
            try {
                s.addAll((Set) visit(children[i]));
            } catch (ClassCastException e) {
                e.printStackTrace();
            }
        }
        if (op instanceof Project) {
            s = new HashSet();
            String[] columns = ((Project) op).getColumnNames();
            for (int i = 0; i < columns.length; i++) {
                s.add(columns[i]);
            }
        } else
            if (op instanceof Expose) {
                s = new HashSet();
                s.add(((Expose) op).getColumnName());
            } else
                if ((op instanceof LeftSemiJoin) || (op instanceof Union)) {
                    s = new HashSet();
                    try {
                        s = (Set) visit((Set) children[0]);
                    } catch (ClassCastException e) {
                        e.printStackTrace();
                    }
                } else
                    if (op instanceof RightSemiJoin) {
                        s = new HashSet();
                        try {
                            s = (Set) visit(children[children.length - 1]);
                        } catch (ClassCastException e) {
                            e.printStackTrace();
                        }
                    } else {
                        try {
                            s.addAll((Set) p.visit(op));
                        } catch (ClassCastException e) {
                            e.printStackTrace();
                        }
                    }
        hash.put(node, s);
        return s;
    }
    public void visitTempTrees() {
        for (Iterator i = tempTrees.iterator(); i.hasNext();) {
            visit(i.next());
        }
    }
}