package edu.wpi.cs.dsrg.xmldb.xat.common.dompattern;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.*;

/**
 * This Class defines the structure of nodes may appear in DOMPattern.
 * Different combinations of the tagName and tagValue has different meanings:
 * 			tagName					tagValue
 * Attr: 	name of attribute 		value of attribute.
 * Element:	tag name				null
 * Text:		#text					content of the text node.
 * Column:	#col					ColumnName.
 * 
 * !NOTE!
 * In order to support the variable name as the attribute name, value and tag name, the Attr.tagName, Attr.tagValue and Element.tagName
 * Can be of type ColumnName.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */

public interface DOMPatternNode extends Visitable
{
/**
 * Add one child to this node. 
 * @return void
 * @param new_child The new child node adding to.
 * @since 1.0
 */
public void addChild(DOMPatternNode new_child);
/**
 * Add a child to the node at a certain position
 * @return void
 * @param new_child The new child node adding to.
 * @param position The position for this new node.
 * @since 1.0
 */
public void addChild(DOMPatternNode new_child, int position);
/**
 * Delete a child from this node.
 * @return int The number of children.
 * @param child The to-be-deleted child.
 * @since 1.0
 */
public int deleteChild(DOMPatternNode child);
/**
 * Get a specific child node of this node by their position.
 * @return DOMPatternNode The child node specified.
 * @param position The position of the child node.
 * @since 1.0
 */
public DOMPatternNode getChild(int position);
/** 
 * Get all the children nodes of this node.
 * @return DOMPatternNode[] The array containing all children nodes.
 * @since 1.0
 */
public DOMPatternNode[] getChildren();
/** 
 * Get the parent node of this node.
 * @param DOMPatternNode The parent node.
 * @since 1.0
 */
public DOMPatternNode getParent();
/**
 * Get the tag name of this DOMPatternNode.
 * @return String The tag name of this DOMPatternNode.
 * @since 1.0
 */
public String getTagName();
/**
 * Get the tag value of this DOMPatternNode.
 * @return Object The tag value of this DOMPatternNode.
 * @since 1.0
 */
 
public Object getTagValue();
/**
 * Check if the DOMPattern node has been canceled out.
 * @return boolean True: if the DOMPattern node has been canceled out. False: otherwise.
 * @since 1.0
 */    
public boolean isCanceledOut();
/**
 * Set the canceledOut property.
 * @return void
 * @param newCanceledOut The boolean indicate if the DOMPattern Node is canceled or not.
 * @since 1.0
 */
 public void setCanceledOut(boolean newCanceledOut);
/**
 * Set a specific child to be a new child node.
 * @return void
 * @param position The position of the child node.
 * @param new_child The new child node.
 * @since 1.0
 */
public void setChild(int position, DOMPatternNode new_child);
/**
 * Set the parent node of this node.
 * @return void
 * @param new_parent The new parent of this node.
 * @since 1.0
 */
public void setParent(DOMPatternNode new_parent);
/**
 * Set the tag name of this DOMPatternNode.
 * @return void
 * @param name The name for tag.
 * @since 1.0
 */ 
public void setTagName(String name);
/**
 * Set the tag value of DOMPatternNode.
 * @return void
 * @param param The tag value of this DOMPatternNode.
 * @since 1.0
 */
void setTagValue(NavigationStep[] param);
/**
 * Construct a NavigationStep using given value string. Then set the tag value
 * of DOMPatternNode as new value.
 * @return void
 * @param value The string of new value.
 * @since 1.0
 */ 
public void setTagValue(String value);
}
