package edu.wpi.cs.dsrg.xmldb.xat.common.expression;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.sqlgeneration.*;

/**
 * This class defines a arithmatic expression.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 */

public class BinArithExpressionImp
    extends BinExpressionImp
    implements BinArithExpression {

	/**
	 * The left and right expression.
	 * @since 1.0
	 */
    Expression eLeft, eRight;
    /**
     * The type of arithmatic expression.
     * @since 1.0
     */
    int iType;

/**
 * The default binary arithmatic expression constructor.
 * @since 1.0
 */
public BinArithExpressionImp() {
    eLeft = null;
    eRight = null;
}
/**
 * The binary arithmatic expression constructor.
 * @param left  The left part of binary expression.
 * @param right The right part of binary expression.
 * @since 1.0
 */

public BinArithExpressionImp(Expression left, Expression right) {
    eLeft = left;
    eRight = right;
}
/**
 * Evaluate the expression of the XATOperator.
 * @return The evaluate result of expression.
 * @param xopOper The operator evaluated.
 * @param iPositions The position of tuple in the XATTable.
 * @since 1.0
 */
 
public Object eval(XATOperator xopOper, int[] iPositions) {
    try {
        //evaluate the expression and convert result 
        Object lResult = eLeft.eval(xopOper, iPositions);
        Object rResult = eRight.eval(xopOper, iPositions);

        if ((((XATDataValue) lResult).convertToCollection().size() == 0)
            || (((XATDataValue) rResult).convertToCollection().size() == 0))
            return null;

        double dLeft = 0.0;
        double dRight = 0.0;

        dLeft = ((XATDataValue) lResult).convertToDouble().doubleValue();

        dRight = ((XATDataValue) rResult).convertToDouble().doubleValue();

        double dReturn;
        switch (iType) {
            case PLUS :
                dReturn = dLeft + dRight;
                break;
            case MINUS :
                dReturn = dLeft - dRight;
                break;
            case MULTIPLE :
                dReturn = dLeft * dRight;
                break;
            case DIVIDE :
                dReturn = dLeft / dRight;
                break;
            case MOD :
                dReturn = dLeft % dRight;
                break;
            default :
                /* fail */
                return null;
        }
        return new XATDataValueImp(new Double(dReturn));
    } catch (ClassCastException e2) {
        /* um, fail?  */
        return new XATDataValueImp(new Double(0.0));
    }
}
/**
 * Get the left part of expression.
 * @return Expression The left part of binary expression
 * @since 1.0
 */
 
public Expression getLeft() {
    return eLeft;
}
/**
 * Get the right part of expression.
 * @return Expression The right part of binary expression
 * @since 1.0
 */

public Expression getRight() {
    return eRight;
}
/**
 * Get the type of expression.
 * @return int The type of expression.
 * @since 1.0
 */
public int getType() {
    return iType;
}
/**
 * Set the left part of expression.
 * @param left The left part of a binary expression.
 * @since 1.0
 */
 
public void setLeft(Expression left) {
    eLeft = left;
}
/**
 * Set the right part of expression.
 * @param left The right part of a binary expression.
 * @since 1.0
 */
public void setRight(Expression right) {
    eRight = right;
}
/**
 * Set the type of the expression.
 * @param type The number used to represent the type of expression.
 * @since 1.0
 */
public void setType(int type) {
    iType = type;
}
/**
 * Convert the expression as a SQL statement.
 * @return java.lang.String The converted string.
 * @param soe The operator evaluator for SQL generation.
 * @since 1.0
 */
 
public String toSQL(OperatorEvaluator soe) {
	return null;
}
/**
 * Output a string of this expression.
 * @return String the output result.
 * @since 1.0
 */
 
public String toString() {
    switch (iType) {
        case PLUS :
            return "(" + eLeft.toString() + " + " + eRight.toString() + ")";
        case MINUS :
            return "(" + eLeft.toString() + " - " + eRight.toString() + ")";
        case MULTIPLE :
            return "(" + eLeft.toString() + " * " + eRight.toString() + ")";
        case DIVIDE :
            return "(" + eLeft.toString() + " / " + eRight.toString() + ")";
        case MOD :
            return "(" + eLeft.toString() + " % " + eRight.toString() + ")";
        default :
            return "("
                + eLeft.toString()
                + " Arith("
                + Integer.toString(iType)
                + ") "
                + eRight.toString()
                + ")";
    }
}
}
