package edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.Vector;
import java.lang.reflect.*;

/**
 * This class looks over the interfaces of the objects
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 */
public abstract class TwoParamVisitor extends InheritableVisitor
{
    /** 
     * To visit two objects.
     * @since 1.0
     */
    public abstract Object dispatch(Object obj1, Object obj2);
/**
 * This method try to find the method for both classes.
 * @return Method The common method for both classes.
 * @param c1 The first class.
 * @param c2 The second class.
 * @since 1.0
 */
protected Method getMethod(Class c1, Class c2) {

    if (c1 == null || c2 == null)
        return null;

    Method m = null;

    // A vector for all the ancestors include itself of c1.
    Vector ancestors1 = getAncestors(c1);

    // A vector for all the ancestors include itself of c2.
    Vector ancestors2 = getAncestors(c2);

    // then add all the super class into the Vector.
    for (int i1 = 0; i1 < ancestors1.size(); i1++) {

        for (int i2 = 0; i2 < ancestors2.size(); i2++) {

            // get the next to be procesed Class.
            Class curClass1 = (Class) ancestors1.elementAt(i1);
            Class curClass2 = (Class) ancestors2.elementAt(i2);
            // check if we can find a method for them.
            try {
                m = getClass().getMethod("visit", new Class[] { curClass1, curClass2 });
                // return the method if we found it.
                if (m != null) {
                    return m;
                }
            } catch (NoSuchMethodException e) {
            }
        }
    }

    // then, we can only call the dispatch method.
    try {
        m =
            getClass().getMethod("dispatch", new Class[] { Object.class, Object.class });
    } catch (NoSuchMethodException e) {
        System.out.println("Cannot Happend!");
    }
    return m;
}
    /**
     * In some case, the visit() will call the visitable, which is
     * the highest interface in the visitor pattern. In that case, we just
     * lead it to the visit(object) method to dynamically find out where
     * you should go.
     * @return Object
     * @param obj1 The first visitable object.
     * @param obj2 The second visitable object.
     * @since 1.0
     */
    public Object visit(Visitable obj1, Visitable obj2){
      return visit((Object)obj1, (Object)obj2);
    }
    /** 
     * If there is an inheretance between the visited objects, it will only match the 
     * highest class.
     * @return Object
     * @param obj1 The first object.
     * @param obj2 The second object.
     * @since 1.0     
     */
    public Object visit(Object obj1, Object obj2)
    {
        if(obj1 == null && obj2 == null)
          return dispatch();

        if(obj1 == null)
          return visit(obj2);

        if(obj2 == null)
          return visit(obj1);

        Method m = getMethod(obj1.getClass(), obj2.getClass());
        if(m == null){
            // then get the dispatch object.
            try
            {
                m = getClass().getMethod("dispatch",
                    new Class[] { Object.class, Object.class });
            }
            catch (NoSuchMethodException nsme)
            {
                System.out.println("Cannot Happend!");
            }
        }
        try{
            return m.invoke(this, new Object[] { obj1, obj2 });
        }catch(InvocationTargetException e){
          return dispatch(obj1, obj2);
        }catch(IllegalAccessException e){
          return dispatch(obj1, obj2);
        }catch(IllegalArgumentException e){
          return dispatch(obj1, obj2);
        }
     }
}
