package edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.InheritableVisitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.*;

/**
 * This class is used to construct a Project node to extract the required columns out. 
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public class ProjectFillVisitor extends InheritableVisitor
{
/**
  * The hash table to keep the map between XATNode and its corresponding SchemaCleanUpNode.
  * @since 1.0     
  */
  Hashtable hash;
    
/**
 * Default ProjectFillVisitor Contructor.
 * @param  hash_init The hash table to keep the map between 
 * XATNode and its corresponding SchemaCleanUpNode.
 * @since 1.0
 */
public ProjectFillVisitor(Hashtable hash_init) {
    hash = hash_init;
}
public Object dispatch() {
    return null;
}
public Object dispatch(Object o) {
    return o;
}
/**
* Get the value of the propertity "DEBUG_SchemaCleanUp".
* @return boolean True: With debug information. False: No debug information.
* @since 1.0
*/
public static boolean isDebug() {
    return System.getProperty("DEBUG_SchemaCleanUp") != null;
}
/**
 * Construct a Project node to extract the required columns out, 
 * and put it as parent of current node. 
 * @return Object 
 * @param  node   The XATNode being cleaned.
 * @since 1.0
 * @see SchemaCleanupNode
 */

public Object visit(XATNode node) {
    SchemaCleanupNode scn = (SchemaCleanupNode) hash.get(node);
    XATNode parent = node.getParent();
    if (parent == null) {
        return null;
    }
    XATNode projectNode = new XATNodeImp();
    Project projectOperator = new ProjectImp(projectNode);
    projectNode.setOperator(projectOperator);
    /* Here we have an essential problem.  We have the columns in set
     * form, which leaves us with no order.  In order to put them in
     * order, i think we'd have to traverse the tree bottom-up, but that
     * would break the algorithm.  for now i'm going to hope it's not a
     * problem.
     */

    projectOperator.setColumnNames((String[]) scn.schema.toArray(new String[] {
    }));
    projectNode.addChild(node);
    int position = parent.deleteChild(node);
    parent.addChild(projectNode, position);
    return projectNode;
}
}
