package edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.Hashtable;
import java.util.Set;
import java.util.HashSet;

import edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup.SchemaCleanupNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.InheritableVisitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;

/**
 * This class is uded to compute the Minimum Schema.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public class SchemaVisitor extends InheritableVisitor
{
    /**
      * The hash table to keep the map between XATNode and its corresponding SchemaCleanUpNode.
      * @since 1.0     
      */	
    Hashtable hash;

/**
 * The default SchemaVisitor constructor.
 * @param  hash_init The hash table to keep the map between 
 * XATNode and its corresponding SchemaCleanUpNode.
 * @since 1.0
 */
public SchemaVisitor(Hashtable hash_init) {
    hash = hash_init;
}
public Object dispatch() {
    return null;
}
public Object dispatch(Object o) {
    return o;
}
/**
 * Get the value of the propertity "DEBUG_SchemaCleanUp".
 * @return boolean True: With debug information. False: No debug information.
 * @since 1.0
 */
public static boolean isDebug() {
	return System.getProperty("DEBUG_SchemaCleanUp")!=null;
}
/**
 * Construct an set object using given object o.
 * @return Set The constructed object.
 * @param  o   The object used to construct the Set object.
 * @since 1.0
 */
 
protected Set oneElementSet(Object o) {
    Set s = new HashSet();
    s.add(o);
    return s;
}
/**
 * Compute the SchemaCleanupNode of given XATNode.
 * This method compute the SchemaCleanupNode for a given XATNode, and put the result
 * into hash table.
 * @return Object 
 * @param  node   The XATNode being cleaned.
 * @since 1.0
 * @see SchemaCleanupNode
 */

public Object visit(XATNode node) {
    XATNode parent = node.getParent();
    SchemaCleanupNode scn = (SchemaCleanupNode) hash.get(node);
    if (parent == null) {
        /* base case.  we need to find what's _really_ required. */

        /* we output the last column.  so what's the last column? */

        XATNode currentNode = node;
        SchemaCleanupNode currentSCN = scn;

        while (true) {
            if (!currentSCN.produced.isEmpty()) {
                scn.schema = currentSCN.produced;
                node.setSchema(currentSCN.produced);
                return scn.schema;
            }
            XATOperator op = currentNode.getOperator();
            if (op instanceof Project) {
                String[] strings = ((Project) op).getColumnNames();
                Set s = oneElementSet(strings[strings.length - 1]);
                scn.schema = s;
                node.setSchema(s);
                return s;
            }
            if (op instanceof Expose) {
                String column = ((Expose) op).getColumnName();
                Set s = oneElementSet(column);
                scn.schema = s;
                node.setSchema(s);
                return s;
            }
            XATNode children[] = currentNode.getChildren();
            for (int i = 0; i < children.length; i++) {
                visit(children[i]);
            }
            return null;
        }

    } else {
        /* inductive case. */
        SchemaCleanupNode parentSCN = (SchemaCleanupNode) hash.get(parent);

        scn.schema = parentSCN.required;
        node.setSchema(scn.schema);
        return parentSCN.required;
    }
}
}
