package edu.wpi.cs.dsrg.xmldb.xat.component.sqlgeneration;

/**
 * <pre>
 * QueryFragment contains a fragment type,
 * a name, a binding, and in some cases, a list of parameters
 * The name is the name of the table or column name to use
 * The binding is what to rename the name to, 
 * so SELCECT xxx AS yyy, has a name of xxx and a binding of yyy and the
 * clause type is select
 * </pre> 
 * @author: Brad Pielech
 */
public class QueryFragment implements Cloneable{
	private int clauseType;
	private String binding;
	private String name;
	private java.lang.String params;
/**
 * Create a QueryFragment with a fragment type and a name
 */
public QueryFragment(int fragmentType, String fragmentText) {
	clauseType = fragmentType;
	name = fragmentText;
	binding = "";
}
/**
 * Create a QueryFragment with a fragment type, a binding, and a name
 */
public QueryFragment(int fragmentType, String fragmentText, String fragmentBinding) {
	clauseType = fragmentType;
	name = fragmentText;
	binding = fragmentBinding;
}
/**
 * Create a QueryFragment with a fragment type, a binding, a name, and parameters
 */
public QueryFragment(int fragmentType, String fragmentText, String fragmentBinding, String fragmentParams) {
	clauseType = fragmentType;
	name = fragmentText;
	binding = fragmentBinding;
	params = fragmentParams;
}
/**
 * Returns a clone of this object
 * 
 * @return Object
 */
public Object clone() {
	try {
		return super.clone();	
	} catch (Exception e) {
		//
	} 
	return null;
	
}
/**
 * Returns the binding
 * 
 * @return java.lang.String
 */
public String getBinding() {
	return binding;
}
/**
 * Returns the fragment type
 * 
 * @return int
 */
public int getClauseType() {
	return clauseType;
}
/**
 * Returns the name
 * 
 * @return java.lang.String
 */
public String getName() {
	return name;
}
/**
 * Returns the parameters
 * 
 * @return java.lang.String
 */
public java.lang.String getParams() {
	return params;
}
/**
 * Sets the binding
 * 
 * @param newBinding java.lang.String
 */
public void setBinding(java.lang.String newBinding) {
	binding = newBinding;
}
/**
 * Sets the name
 * 
 * @param newName java.lang.String
 */
public void setName(java.lang.String newName) {
	name = newName;
}
/**
 * Sets the parameters
 * 
 * @param newParams java.lang.String
 */
public void setParams(java.lang.String newParams) {
	params = newParams;
}
}
