package edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer;

/**
 * Statistics that wish to keep an aggregate value (total), mean, min and max, but report
 * their values in increments (such as the statistic from the last running) as opposed to
 * the overall total, should derive from this class instead of AggregateStatisticsObject.
 * @author: Brad Pielech
 */
public abstract class AggregateDifference extends AggregateStatisticsObject {
/**
 * AggregateDifference constructor comment.
 */
public AggregateDifference() {
	super();
}
/**
 * AggregateDifference constructor comment.
 * @param name java.lang.String
 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
 */
public AggregateDifference(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) {
	super(name, queryObject);
}
/**
 * Sets the property value.  For AggregateStatisticsObjects, this method
 * will calculate the new weighed mean and set the value to what was
 * passed in.  This assumes the newly passed in value represents the
 * total property value and not the difference from when it was calculated
 * previously. <p>
 *
 * For example, the THROUGHPUT calculator will report
 * how many tuples the tree has outputted in total, not just how many
 * were outputted since the last time it ran.  OTOH, the NUMBER_OF_TUPLES_OUTPUTTED
 * calculator will report how many tuples have been outputted since it last ran.  Therefore,
 * that calculator should derive from AggregateDifference
 * @param newPropertyValue java.lang.String
 */ 
public void setPropertyValue(String newPropertyValue) {
	double thisValue = Double.parseDouble(newPropertyValue);
	double oldValue = 0;
	String oldValueS = getPropertyValue();
	if(oldValueS != null){ 
		oldValue = Double.parseDouble(oldValueS);
	}
	double total = thisValue + oldValue;	
    super.setPropertyValue("" + total);
    
}
}
