package edu.wpi.cs.dsrg.xmldb.xat.component.xatrewrite;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.*;

/**
 * This Class is the rewrite rules used to cancel out two navigate operators 
 * within XAT.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */
public class CancelOutNavigateRules extends ReWriteRules {
/**
 * Default CancelOutNavigateRules constructor.
 * Constructs a CancelOutNavigateRules object.
 * @since 1.0   
 */
public CancelOutNavigateRules() {
    super();
}
/**
 * Compare and cancel out two navigation operators (if possible).
 * This method first compares two navigate nodes. If they have the same
 * entry point and destination point, the parent is deleted, and replaced
 * with a rename node. The new rename node renames the bindings.
 * @return java.lang.Object
 * @param ancestor   The ancester navigate operator.
 * @param descendent The decendent navigate operator.
 * @since 1.0
 * @see   edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigate     
 */
public Object visit(Navigate ancestor, Navigate descendent) {
    ReWriteResult p1 = new ReWriteResult();
    XATNode x1 = descendent.getNode();
    NameColumnImp r1;

    //Get all navigation steps for ancestor, and descendent

    NavigationStep[] ancestorEntryPoint = ancestor.getEntryPoint();
    NavigationStep[] ancestorDestination = ancestor.getDestination();

    NavigationStep[] descendentEntryPoint = descendent.getEntryPoint();
    NavigationStep[] descendentDestination = descendent.getDestination();

    //obtain a string representation of the destination and entry points

    String ancestorEntryString = createString(ancestorEntryPoint);
    String ancestorDestinationString = createString(ancestorDestination);

    String descendentEntryString = createString(descendentEntryPoint);
    String descendentDestinationString = createString(descendentDestination);

    //If the entry points and destination points are the same, delete the ancestor
    if (ancestorEntryString.equals(descendentEntryString)
        && ancestorDestinationString.equals(descendentDestinationString)) {

        //String newEntryPoint = ancestor.getBinding();
        String newColumnNameString = ancestor.getBinding();
        String origColumnNameString = descendent.getBinding();
        if (isDebug()) {
            System.out.println("Found a match");
            System.out.println(
                "Renaming " + origColumnNameString + " into " + newColumnNameString);
        }
        //replace the navigate with a rename operator

        r1 = new NameColumnImp(x1, newColumnNameString, origColumnNameString);
        x1.setOperator(r1);
        NavigationStep[] ns1 =
            getVariableTable().getVariableValue(origColumnNameString);
        r1.setColumnPath(ns1);
        /*
                    NavigationStep[] ns1 = getVariableValue(newEntryPoint);
        			descendent.setEntryPoint(ns1);
        			descendent.setDestination(null);
        			//binding is all set
        			XATNode newNode = x1.getParent();
        			XATNode n1;
        			//Traverse up the tree attempting to get the root node.
        			while(true){
        				n1 = newNode;
        				System.out.println("this is the attempt to find the root " + n1.getOperator().toString());
        				newNode = newNode.getParent();
        				//no parent?  This is the root node.
        				if(newNode==null){
        					break;
        				}
        			}
        			//Move to the root, so that it can then be pushed down with rename push down
        //			moveNodeToRoot(n1,x1);
        */
        p1.setFinalPosition(true);
    } else {
        p1.setSwapped(false);
        //false
        p1.setFinalPosition(false);
    }

    return p1;
}
}
