/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.Mjoin;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;

import edu.wpi.cs.dsrg.eddy.datastructure.EddyQueue;
import edu.wpi.cs.dsrg.eddy.datastructure.EddyTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
//Karen Works

//This structure stores the data the MJoin searches on 

public class MJoinData 
{
	int MAX_WINDOW_SIZE = 1000; //minimize window size
	// maximum number of tuples to store

	//links to tuples that are referenced
	private ArrayList<HashMap<String, TupleQueue>> StoredTuples;

	// number of indexes this table is joined against
	int intNumIndexes; 
	
	// Array number of the data source
	int intSTREAMId; 
	
	EddyQueue TupleList;
	//Array of positions in tuple that need to be indexed on
	ArrayList<Integer> HashKeyIndex;  
	// Current number of stored tuples
	int intCurrNumberofStoredTuples = 0; 

	public void init(Integer STREAMId,ArrayList<Integer> HashKeys )
	{
		// set up required variables
		intNumIndexes = HashKeys.size();
		StoredTuples = new ArrayList<HashMap<String, TupleQueue>>(intNumIndexes);
		for (int i = 0; i < intNumIndexes; i++)
		{
			StoredTuples.add(i,new HashMap<String, TupleQueue>());
		}
		
		intSTREAMId = STREAMId;
		HashKeyIndex = HashKeys;
		TupleList = new EddyQueue();
	}
	
	public void Insert(XATTuple tuple)
	{
		// if adding this tuple will make the total#
		// > than the window purge the oldest tuple
		if ((intCurrNumberofStoredTuples +1) > MAX_WINDOW_SIZE)
		{	
			PurgeTuple();
		}
		
		// add the tuple to the list
		TupleList.enqueue(tuple);
		
		// for each index
		for (int i = 0; i < intNumIndexes; i++)
		{
			// compute the key for this tuple
			String key = "";
			key += tuple.getCell(HashKeyIndex.get(i)).convertToString()+"---";
			
			//for testing only			
			//System.out.println("adding  "+key+" to "+intSTREAMId);
			
			// store the tuple in the appropriate queue
			TupleQueue Tq = StoredTuples.get(i).get(key);
			if(Tq != null)
			{
				Tq.enqueue(tuple);	
			}
			else
			{
				TupleQueue<Object> tq = new TupleQueue<Object>();
				
				tq.enqueue(tuple);
				StoredTuples.get(i).put(key.toString(), tq);
			}

		}
		intCurrNumberofStoredTuples++;		
	}
	
	
	private void PurgeTuple()
	{
		//while the number of tuples+1 is > the window
		while ((intCurrNumberofStoredTuples +1) > MAX_WINDOW_SIZE)
		{	
			// find the oldest tuple & remove it
			XATTuple DeleteTup = (XATTuple)TupleList.dequeue();
			
			// remove this tuple from all the index queues ad well
			for (int i = 0;i < intNumIndexes;i++)
			{
				String key = "";
				key += DeleteTup.getCell(HashKeyIndex.get(i)).convertToString()+"---";
					 	
				StoredTuples.get(i).remove(key);				
			}
		
			intCurrNumberofStoredTuples--;
		}	
		
	}

	// this function returns the STREAM is for this STeM Data
	public int getSTREAMId()
	{
		return intSTREAMId;
	}

	// this function probes the STEM data and returns results 
	public ArrayList <XATTuple> Probe(XATTuple tup, int indexProber, int indexProbee)
	{
		// validate indexes
		if ((indexProber > intNumIndexes) || (indexProber < 0 ))
		{	
			System.out.println("Error STeM Probe - NumSTeMIndex out of range");
			return null;
		}
		
		ArrayList <XATTuple> results = new ArrayList();
		
		// compute key of probe tuple
		String key = "";
		key += tup.getCell(indexProber).convertToString()+"---";
		
		
		//for testing only			
		//System.out.println("looking for  "+key+" From "+tup.getStreamID()+"in "+intSTREAMId);
				
		// find matches to probe tuple key
		TupleQueue Tq = StoredTuples.get(indexProbee).get(key.toString());
		
		System.out.println("probe tuple ts: "+tup.getCell(tup.getLength() - 1));
		/*
		//rimma debug: print results:
		System.out.println("\nprobe key: "+ key);				
		System.out.println("distinct keys in the state: "+StoredTuples.get(indexProbee).keySet().size());		
		Iterator<String> keys = StoredTuples.get(indexProbee).keySet().iterator();						
		int p = 1;
		while (keys.hasNext())
		{							
		String currKey = keys.next().toString();
		System.out.println("\t" + p++ +". key: "+currKey);	
		if (currKey.equalsIgnoreCase(key))
		{
		System.out.println("\t\tMATCHING KEY FOUND");
		}
		}
		*/
		
		if(Tq != null) // if = null no tuples have been added yet
		{
			// loop through tuple(s) found
			//boolean bCont = true;
			for (Iterator <XATTuple>it = Tq.iterator(); it.hasNext();)// && bCont;) 
			{
		        XATTuple TupFound = it.next();
		        
		        // if the tuples meet the window req
		        // return them
		        	
		        //rimma begin -- 100 tuple window join
	    		Object strTupleCounter1 = tup.getCell(tup.getLength() - 1);
	    		long tupleCounter1 = Long.parseLong(strTupleCounter1.toString());
	    		
	    		Object strTupleCounter2 = TupFound.getCell(TupFound.getLength() - 1);
	    		long tupleCounter2 = Long.parseLong(strTupleCounter2.toString());
	    		//rimma end -- 100 tuple window join
	    		System.out.println("\t\tmatching key ts: " + tupleCounter2);
	    	
	    		//System.out.println("matching tuple ts: "+tupleCounter2);	    		
	    		
	    		long diff = Math.abs(tupleCounter1 - tupleCounter2);
		        //if (TupFound.getMinTimeStamp() < tup.getMinTimeStamp())
	    		if (diff <= 5)// && (TupFound.getMinTimeStamp() <= tup.getMinTimeStamp()))
		        {
		        	results.add(TupFound);
		        }
		       /*else
		       {
		    	   bCont = false;
		       }
		       */
		    }

		}
		else
		{
			results = null;
		}
		
		if (results != null)
		{	
			if (results.size() > 0)
			{				
				/*for (int k=0; k<results.size();k++)
				{
					System.out.print(results.get(k));
				}				
				System.out.print("# of results: "+ results.size());
				*/
				return results;
			}
			else
			{
				return null;
			}	
		}
		else
		{
			return null;
		}
	}
}
