/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.Mjoin;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Iterator;
import java.util.BitSet;

import edu.wpi.cs.dsrg.STeM.STeMData;
import edu.wpi.cs.dsrg.eddy.datastructure.*;
import edu.wpi.cs.dsrg.eddy.operator.EddyOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.*;
//Karen Works

//This is the MJoinProbe operator 
public class MJoinProbeOperatorImp extends XATMultiQueueWindowStreamOperatorImp implements MJoinProbeOperator
{
	//window size
	//public int windowSize;
	
    //number of streams
    private int NumSTREAMS = 0;	
	
	// data set associated with MJoin
	private HashMap<Integer,MJoinData> MJoinStreamData;
	private ArrayList<Integer> MJoinStreamDataSTREAMIds;
	

	//path that the stream will take
	private HashMap<Integer, ArrayList<Integer>> STREAMPaths;
	private HashMap<Integer, ArrayList<JoinEval>> STREAMPathsEval; 
	private ArrayList<JoinEval> Evals; //array of evals
				
	//initialization
	public MJoinProbeOperatorImp(XATNode init) 
	{
		super(init);
		MJoinStreamData = new HashMap<Integer,MJoinData>();
		MJoinStreamDataSTREAMIds = new ArrayList<Integer>();
		
		STREAMPaths= new HashMap<Integer, ArrayList<Integer>>(); 
		STREAMPathsEval= new HashMap<Integer, ArrayList<JoinEval>>();
		Evals = new ArrayList<JoinEval>();
	}
	
	String outFile = "C:\\Rimma\\Research\\QueryMesh\\execution_stats\\MJOIN_PROBE.txt";
	public static void WriteToFile(String fileLocation, String stringToWrite, boolean doNewLine)
	{
		try
		{
			//Create file 
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			if (doNewLine)
			{
				out.newLine();
			}
			//Close the output stream
			out.close();			    
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}

	public boolean run() throws Exception 
	{
		//input queue from STREAM
		XATQueueSchema inSchema;
		int inLength;
		if ((inSchema = this.getInputQueue().getSchema()) == null)
		{
			return false;
		}
		inLength = inSchema.getLength();

		if (this.getOutputQueue().getSchema() == null) 
		{
			init();
		} 
					
		//int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		int howMany  = 5; //rimma temp "hack"
		
		//compute the schema				
		for (int j = 0; j < (NumSTREAMS); j++) 
		{			
			int numResultsByStream = 0;
			System.out.println("\nProcessing stream [" + j + "]");
			/** get the tuples */
			
			System.out.println("\t# of tuples in the queue: " + this.getInputQueue(j).getQueueSize());			
			XATTuple[] tuples = this.getInputQueue(j).dequeue(howMany, this);											
			MJoinData mjHash = MJoinStreamData.get(j);
			System.out.println("\t# of tuples in the Mjoin hash: " + mjHash.TupleList.size());	
			System.out.println("\tDequeued: " + tuples.length + " tuples.");			
			
			//set stream id
			int CurrentSTREAMId = j;
			
			//get the eval list for the current tuple stream
			ArrayList<JoinEval> CurrentStreamEvalList = STREAMPathsEval.get(CurrentSTREAMId);
						
			for(int i = 0; i < tuples.length; i++)
			{
				//System.out.println("Processing..."+j+"-"+i+" of "+tuples.length);
				XATTuple ProbeTuple = tuples[i];
				ProbeTuple.setStreamID(j);
				
				//insert tuple into local stream hash table
				MJoinData mjsd = MJoinStreamData.get(ProbeTuple.getStreamID());
				mjsd.Insert(ProbeTuple);
				
				BitSet inAnswer = new BitSet(NumSTREAMS);
				inAnswer.clear();
				inAnswer.set(CurrentSTREAMId, true);
				
				// search the data for matches
				boolean bAnswerFound = true;
				boolean bFirstStream = true;
				boolean bProducedCompleteResults = false;
				boolean bReproduceResults = true;
				
				ArrayList<MJoinTuple> result = null;
										
				ArrayList <XATTuple> r = null;					
					
				//hashtable of intermediate results for the current base stream probe tuple
				Hashtable resultsByStreamHash = new Hashtable(); //hash of <MJoinTupleImp>										
				Hashtable intermResultsHash = new Hashtable(); //hash of <MJoinTupleImp>
					
				intermResultsHash.put(0, ProbeTuple);
				resultsByStreamHash.put(CurrentSTREAMId, intermResultsHash);
				
				//System.out.println("size of resultsByStreamHash: " + resultsByStreamHash.size());
				//System.out.println("size of intermResultsHash: " + intermResultsHash.size());
				String tabs = "";	
				
				//go through the list of all evals
				for (int b = 0; b < CurrentStreamEvalList.size(); b++)
				{		
					tabs += "  ";
					//get the current evaluation from the list of evals
					JoinEval procStreamIndex = CurrentStreamEvalList.get(b);
						
					int proberStream = 0;
					int probeeStream = 0;
					int proberIndex = 0;
					int probeeIndex = 0;
						
					//figure out if we need to probe on STREAM1 or STREAM2
					if (inAnswer.get(procStreamIndex.StreamId1))
					{		
						proberStream = procStreamIndex.StreamId1;
						probeeStream = procStreamIndex.StreamId2;
						proberIndex = procStreamIndex.Index1;
						probeeIndex = procStreamIndex.Index2;							
					}
					else
					{
						proberStream = procStreamIndex.StreamId2;
						probeeStream = procStreamIndex.StreamId1;
						proberIndex = procStreamIndex.Index2;
						probeeIndex = procStreamIndex.Index1;						
					}										
					
					System.out.println(tabs + "(" +proberStream + " join "+ probeeStream +"):");
					
					
					MJoinData mjHashProber = MJoinStreamData.get(proberStream);
					MJoinData mjHashProbee = MJoinStreamData.get(probeeStream);
					System.out.println("\t# of tuples in the Mjoin PROBER: " + mjHashProber.TupleList.size());	
					System.out.println("\t# of tuples in the Mjoin PROBEE: " + mjHashProbee.TupleList.size());	
					
					if (bFirstStream)
					{
						//run probe & get results  
						r = MJoinStreamData.get(probeeStream).Probe(ProbeTuple, proberIndex, probeeIndex);
						inAnswer.set(probeeStream, true);		
						
						if (r == null)
						{
							bFirstStream = true;
							bReproduceResults = false;
							
							System.out.println(tabs +"\t" + "0 results");							
							break;
						}
								
						bFirstStream = false;
						intermResultsHash = new Hashtable();
						//process the (intermediate) results further
						for (int d = 0; d < r.size(); d++)
						{
							intermResultsHash.put(d, r.get(d));								
						}	
						resultsByStreamHash.put(probeeStream, intermResultsHash); //add results according to the stream id
						System.out.println(tabs +"  " + r.size()+" results");
					}
					else
					{
						int numIntermRedultsProduced = 0;
						//probe is based on intermediate result values		
						Hashtable intermResults = (Hashtable)resultsByStreamHash.get(proberStream);
						
						if (intermResults == null)
						{
							bReproduceResults = false;
							System.out.println("want to catch this");
						}
						//iteration through intermediate results
						Enumeration keys = intermResults.keys();
						ArrayList <XATTuple> rInterm = null;
						int w = 1;
						while (keys.hasMoreElements())
						{		
							System.out.print(w++ + ". ");
							int key = Integer.parseInt(keys.nextElement().toString());	
							XATTuple intermProbeTuple = (XATTuple)intermResults.get(key);
							
							//run probe & get results  
							rInterm = MJoinStreamData.get(probeeStream).Probe(intermProbeTuple, proberIndex, probeeIndex);
							
							if (rInterm == null)
							{
								System.out.println(tabs +"  " + "0 results");
							}
							else
							{
								System.out.println(tabs + "  "+ rInterm.size()+" results");
							}
							
							if (rInterm != null)
							{
								inAnswer.set(probeeStream, true);
								
								intermResultsHash = new Hashtable();
								//process the (intermediate) results further
								for (int d = 0; d < rInterm.size(); d++)
								{
									intermResultsHash.put(d, rInterm.get(d));									
								}	
								resultsByStreamHash.put(probeeStream, intermResultsHash); //add results according to the stream id
								
								if (b == (CurrentStreamEvalList.size() - 1))//if there are results in the last evaluation, then there are physical results
								{
									bProducedCompleteResults = true; //there are results for this tuple
								}
								numIntermRedultsProduced += rInterm.size();
							}							
						}//end while
						
						if (numIntermRedultsProduced == 0)
						{							
							bReproduceResults = false;
							break;						
						}
					}//end else (processing intermediate results)																		
				}// end of iterating throught the list of all evals (in the stream path) for the base probe tuple
					
				
				if (bReproduceResults)
				{
					System.out.println("Iterating through results...");
					//test check				
					Enumeration keys = resultsByStreamHash.keys();				
					while (keys.hasMoreElements())
					{							
						int key = Integer.parseInt(keys.nextElement().toString());	
						Hashtable irHash = (Hashtable)resultsByStreamHash.get(key);
			
						String numImResults =  key + "--> number of results is " + irHash.size();
						System.out.println(numImResults);
						WriteToFile(outFile, numImResults, true);
					}
				}
				
				//now convert into the regular "CAPE tuple" and enqueue					
				if (bProducedCompleteResults)
				{
					System.out.println("Complete results were produced.");
					int proberStream = 0;
					int probeeStream = 0;
					int proberIndex = 0;
					int probeeIndex = 0;
					
					//determine how many physical CAPE result tuples have to be created
					int totalPhysicalCAPEOutputTuplesCount = 0;
					
					ArrayList arrayOfResults = new ArrayList();
					inAnswer.clear();
					inAnswer.set(CurrentSTREAMId, true);						
					totalPhysicalCAPEOutputTuplesCount = 1; // 1 for the original base probe tuple
										
					ArrayList resultContent = new ArrayList();
					resultContent.add(ProbeTuple);
					arrayOfResults.add(resultContent);
										
					for (int c = 0; c < CurrentStreamEvalList.size(); c++)
					{						
						//get the current evaluation from the list of evals
						JoinEval procStreamIndex = CurrentStreamEvalList.get(c);
							
						if (!inAnswer.get(procStreamIndex.StreamId2))
						{
							//get the hash of the intermediate results
							Hashtable imResultsHash = (Hashtable)resultsByStreamHash.get(procStreamIndex.StreamId2);							
							totalPhysicalCAPEOutputTuplesCount *= imResultsHash.size();
							inAnswer.set(procStreamIndex.StreamId2, true);
														
							//iterate through results
							if (imResultsHash.size() == 1)
							{								
								XATTuple intermProbeTuple = (XATTuple)imResultsHash.get(0);
								arrayOfResults = AddSubTupleResult(arrayOfResults, intermProbeTuple);
							}
							else
							{
								arrayOfResults = ExtendResultsByMultiplicativeFactor(arrayOfResults, imResultsHash);								
							}														
						}
						else //!StreamId1
						{
							assert (inAnswer.get(procStreamIndex.StreamId1) == false);
							Hashtable imResultsHash = (Hashtable)resultsByStreamHash.get(procStreamIndex.StreamId1);							
							totalPhysicalCAPEOutputTuplesCount *= imResultsHash.size();
							inAnswer.set(procStreamIndex.StreamId1, true);
							
							//iterate through results
							if (imResultsHash.size() == 1)
							{								
								XATTuple intermProbeTuple = (XATTuple)imResultsHash.get(0);
								arrayOfResults = AddSubTupleResult(arrayOfResults, intermProbeTuple);
							}
							else
							{
								arrayOfResults = ExtendResultsByMultiplicativeFactor(arrayOfResults, imResultsHash);								
							}	
						}
					}
						
					//String numResults = "number of physical results for tuple " + j + " is " + totalPhysicalCAPEOutputTuplesCount;
					//WriteToFile(outFile, numResults, true);
					
					XATTuple[] outputTuples = new XATTuple[totalPhysicalCAPEOutputTuplesCount];
					numResultsByStream += totalPhysicalCAPEOutputTuplesCount;		
					
					/*
					
					for(int y = 0; y < arrayOfResults.size(); y++)
					{						
						int numOutputAttr = 0;
						ArrayList mJoinTuple = (ArrayList)arrayOfResults.get(y);
						//iterate through subtuples						
						for (int p = 0; p < mJoinTuple.size(); p++)
						{
							XATTuple subTuple = (XATTuple)mJoinTuple.get(p);							
							numOutputAttr += subTuple.getArrayData().length;							
						}
												
						XATDataValue[] resultTupleContent = new XATDataValueImp[numOutputAttr];
						
						int x = 0;
						for (int p = 0; p < mJoinTuple.size(); p++)
						{
							XATTuple subTuple = (XATTuple)mJoinTuple.get(p);							
							for(int w=0; w < subTuple.getArrayData().length; w++)
							{
								resultTupleContent[x++] = subTuple.getArrayData()[w];
							}
						}						
						XATTuple resultTuple = new XATTupleImp(resultTupleContent);
						
						//enqueue result
						this.getOutputQueue().enqueue(resultTuple, this);
					}
					*/
										
					for(int f = 0; f < totalPhysicalCAPEOutputTuplesCount; f++)
					{
						XATTuple resultTuple = new XATTupleImp();
						this.getOutputQueue().enqueue(resultTuple, this);
					}
					
					String numResults = "value of totalPhysicalCAPEOutputTuplesCount is " + totalPhysicalCAPEOutputTuplesCount;
					System.out.println(numResults);
					totalPhysicalCAPEOutputTuplesCount = 0;
				}						
			}	
						
			//write here			
			String numResults = "number of results for stream " + j + " is " + numResultsByStream;
			System.out.println(numResults);
			WriteToFile(outFile, numResults, true);			
		}
		
		WriteToFile(outFile, "", true);
					
		System.gc();
		return true;
	}	
	
	private ArrayList AddSubTupleResult(ArrayList arrayOfResults, XATTuple intermProbeTuple)
	{
		for (int i=0; i<arrayOfResults.size();i++)
		{
			ArrayList arrayListContent = (ArrayList)arrayOfResults.get(i);
			arrayListContent.add(intermProbeTuple);
		}
		return arrayOfResults;
	}
	
	private ArrayList ExtendResultsByMultiplicativeFactor(ArrayList arrayOfResults, Hashtable hashOfSubResults)
	{
		int newArrayOfResultsSize = arrayOfResults.size() * hashOfSubResults.size();
		ArrayList newArrayOfResults = new ArrayList();
		
		Enumeration keys = hashOfSubResults.keys();
		ArrayList <XATTuple> rInterm = null;
		while (keys.hasMoreElements())
		{							
			int key = Integer.parseInt(keys.nextElement().toString());	
			XATTuple intermProbeTuple = (XATTuple)hashOfSubResults.get(key);
			
			for (int i = 0; i<arrayOfResults.size();i++)
			{
				ArrayList arrayListContent = (ArrayList)arrayOfResults.get(i);
				ArrayList newArrayListContent = new ArrayList();
				newArrayListContent.addAll(arrayListContent);
				newArrayListContent.add(intermProbeTuple);
				
				newArrayOfResults.add(newArrayListContent);
			}
		}
		return newArrayOfResults;
	}
		
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return false;
	}

	public boolean cleanAllAccumulatedTuples() 
	{
		return false;
	}

	public boolean matchStates(XATStreamOperator op) 
	{
		return false;
	}

	public boolean hasUnmatchedState() 
	{
		return false;
	}

	public boolean recomputeUnmatchedState() 
	{
		return false;
	}

	public int setStateID(int startID) 
	{
		return 0;
	}

	public String toToolTipString() 
	{
		return null;
	}

	public boolean finalizeOperation() 
	{
		return false;
	}

	public void setMJoinStreamData(HashMap<Integer,MJoinData>  mjd)
	{
		MJoinStreamData = mjd;
	}

///////////////////////////////////////////
//Info about stored stem data
			
	
	public void setNumStreams(int Ns)
	{
		NumSTREAMS = Ns;
	}

	public ArrayList<Integer> getSTREAMIds()
	{
		return MJoinStreamDataSTREAMIds;
	}	
	
	public void AddEval(int Id, int StreamId1, String Index1, int StreamId2, String Index2)
	{
		while (Evals.size() <= Id)
		{	
			JoinEval  si = new JoinEval();
			Evals.add(si);
		}
		
		JoinEval  si = new JoinEval();
		
		si.StreamId1 = StreamId1;		
		si.Index1 = Integer.parseInt(Index1);		

		si.StreamId2 = StreamId2;
		si.Index2 = Integer.parseInt(Index2);			    
		Evals.set(Id, si);		
	}

	
	public void AddSTREAMPath(int StreamId, String Path)
	{
		ArrayList<Integer> PathList  = new ArrayList<Integer>(); 
		StringTokenizer st = new StringTokenizer(Path);
	    while (st.hasMoreElements()) 
	    {
	    	PathList.add(Integer.parseInt(st.nextElement().toString()));
	    }
		STREAMPaths.put(StreamId, PathList);
		
		//now set up stream evals;
		ArrayList<JoinEval> EvalList  = new ArrayList<JoinEval>();
		for (int i= 0;i < PathList.size();i++)
		{
			EvalList.add(Evals.get(PathList.get(i)));
		}		
		STREAMPathsEval.put(StreamId, EvalList);
	}
	
	 public ArrayList<JoinEval> getEvals()
	 {
		 return Evals;
	 }
}
