/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.Mjoin;
import java.util.Iterator;
import java.util.NoSuchElementException;

//Karen Works

//This structure stores the information associated with 
//a tuple queue and mjoin operation

public class TupleQueue<Item> implements Iterable<Item> {
	    private int N;         // number of elements on queue
	    private Node first;    // beginning of queue
	    private Node last;     // end of queue

	    // helper linked list class
	    private class Node {
	        private Item item;
	        private Node next;
	    }

	    // create an empty queue
	    public TupleQueue() {
	        first = null;
	        last  = null;
	    }

	    // is the queue empty?
	    public boolean isEmpty() { return first == null; }
	    public int length()      { return N;             }
	    public int size()        { return N;             }

	    // add an item to the queue
	    public void enqueue(Item item) {
	        Node x = new Node();
	        x.item = item;
	        if (isEmpty()) { first = x;     last = x; }
	        else           { last.next = x; last = x; }
	        N++;
	    }

	    // remove and return the least recently added item
	    public Item dequeue() {
	        if (isEmpty()) throw new RuntimeException("Queue underflow");
	        Item item = first.item;
	        first = first.next;
	        N--;
	        return item;
	    }

	    // remove and return the least recently added item
	    public String toString() {
	        String s = "";
	        for (Node x = first; x != null; x = x.next)
	            s += x.item + " ";
	        return s;
	    }

	    public Iterator<Item> iterator()  { return new QueueIterator();  }

	    // an iterator, doesn't implement remove() since it's optional
	    private class QueueIterator implements Iterator<Item> {
	        private Node current = first;

	        public boolean hasNext()  { return current != null;                     }
	        public void remove()      { throw new UnsupportedOperationException();  }

	        public Item next() {
	            if (!hasNext()) throw new NoSuchElementException();
	            Item item = current.item;
	            current = current.next; 
	            return item;
	        }
	    }

	}





