/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

/*
 * Created on Jan 23, 2006
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.STeM.Partition.IMPIndex;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;

/**
 * @author lisading
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class MultiHashKeyIndex extends WindowGroupbyIndex {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	
	int[] hashFactors;
	int[] indexAttrPos;
	DataBucket[] dataBuckets;
	int totalNumTuples;
	int[] intervals;
	
	//for testing
	int numWildcards = 0;
	
	public MultiHashKeyIndex() {
		super();
	}
	
	private int hash(XATDataValue[] tupleValue) {
		int numHashAttrs = indexAttrPos.length;
		int hashValue = 0;
		for (int i=0; i<numHashAttrs; i++) {
			int hashKey = Math.abs(((Number)tupleValue[indexAttrPos[i]].getValue()).intValue());
			hashValue += (hashKey % hashFactors[i]) * intervals[i]; 
		}
		return hashValue;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.pullbasedgroupby.WindowGroupbyIndex#insertTuple(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple)
	 */
	public void insertTuple(XATTuple tuple) {
		XATDataValue[] tupleValue = tuple.getArrayData();
		int bucketNo = hash(tupleValue);
		dataBuckets[bucketNo].insertTuple(tuple);
		totalNumTuples ++;
//		if (totalNumTuples>maxTotalNumTuples) {
//			long windowStart = Math.max(0, tuple.getMinTimeStamp()-window);
//			removeExpiredTuples(windowStart);
//		}
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.pullbasedgroupby.WindowGroupbyIndex#insertTupleByTime(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple)
	 */
	public void insertTupleByTime(XATTuple tuple) {
		// TODO Auto-generated method stub

	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.pullbasedgroupby.WindowGroupbyIndex#probe(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct)
	 */
	public Vector probe(XATPunct query) {
		int[] punctTypeArray = query.getPunctTypeArray();

		int numConstantVals = 0;
		int[] constantAttrPos = new int[numGroupAttrs];
		int[] constantAttrVal = new int[numGroupAttrs];
		for (int i=0; i<numGroupAttrs; i++) {
			if (punctTypeArray[groupAttrPosArray[i]] == XATPunct.SINGLE_VALUE) {
				constantAttrPos[numConstantVals] = groupAttrPosArray[i];
				constantAttrVal[numConstantVals] = ((Integer)query.getCell(groupAttrPosArray[i]).getValue()).intValue();;
				numConstantVals ++;
			}
		}
		boolean equalityQuery = true;
		int numAttrs = indexAttrPos.length;
		for (int i=0; i<numAttrs; i++) {
			if (punctTypeArray[indexAttrPos[i]] != XATPunct.SINGLE_VALUE) {
				equalityQuery = false;
				break;
			}
		}

		DataBucket[] dataBucketArray = null;
		if (equalityQuery) {
			int bucketNo = hash(query.getArrayData());
			
			dataBucketArray = new DataBucket[] {dataBuckets[bucketNo]};
		} else {
			//DEBUG - BEGIN
			//numWildcards ++;
			//DEBUG - END
			dataBucketArray = dataBuckets;
		}
		
		Vector result = new Vector();
		int numResultBuckets = dataBucketArray.length;
		long windowStart = Math.max(0, query.getMinTimeStamp() - window);
		HashMap tempResult = new HashMap(100);
		for (int i=0; i<numResultBuckets; i++) {
			DataBucket dBucket = dataBucketArray[i];
			// remove expired tuples
			dBucket.removeExpiredTuples(windowStart);
			
			int numTuples = dBucket.size();
			tempResult.clear();
			for (int j=0; j<numTuples; j++) {
				XATTuple t = dBucket.getTuple(j);
				XATDataValue[] tupleValue = t.getArrayData();
				// check whether the tuple match the query pattern
				if (!this.matchVals(tupleValue, numConstantVals, constantAttrVal, constantAttrPos)) {
					continue;
				}
				int[] vals = new int[numGroupAttrs];
				for (int k=0; k<numGroupAttrs; k++) {
					vals[k] = ((Integer)tupleValue[groupAttrPosArray[k]].getValue()).intValue();
				}
				Integer tempHashKey = new Integer(vals[0]);
				Object obj = tempResult.get(tempHashKey);
				if (obj == null) {
					Vector tempBucket = new Vector();
					TempHashEntry tempEntry = new TempHashEntry(vals);
					tempEntry.insertTuple(t);
					tempBucket.add(tempEntry);
					tempResult.put(tempHashKey, tempBucket);
				} else {
					boolean foundMatchEntry = false;
					Vector tempBucket = (Vector)obj;
					int numTempEntries = tempBucket.size();
					for (int m=0; m<numTempEntries; m++) {
						TempHashEntry tempEntry = (TempHashEntry) tempBucket.get(m);
						if (tempEntry.ifMatch(vals)) {
							foundMatchEntry = true;
							tempEntry.insertTuple(t);
							break;
						}
					}
					if (!foundMatchEntry) {
						TempHashEntry tempEntry = new TempHashEntry(vals);
						tempEntry.insertTuple(t);
						tempBucket.add(tempEntry);
					}
				}
			} // end for j
			Iterator it = tempResult.values().iterator();
			while (it.hasNext()) {
				Vector tempBucket = (Vector)it.next();
				int numTempEntries = tempBucket.size();
				for (int m=0; m<numTempEntries; m++) {
					TempHashEntry tempEntry = (TempHashEntry)tempBucket.get(m);
					result.add(tempEntry);
				}
			}
		} // end for i
		return result;
	}

	public boolean probe2(XATPunct query, long tupleMaxCounter, long tupleMinCounter,ArrayList<XATTuple> results) 
	{
		int[] punctTypeArray = query.getPunctTypeArray();
		int numConstantVals = 0;
		int[] constantAttrPos = new int[numGroupAttrs];
		int[] constantAttrVal = new int[numGroupAttrs];
		
		for (int i=0; i<numGroupAttrs; i++) 
		{
			if (punctTypeArray[groupAttrPosArray[i]] == XATPunct.SINGLE_VALUE) {
				constantAttrPos[numConstantVals] = groupAttrPosArray[i];
				constantAttrVal[numConstantVals] = Math.abs(((Integer)query.getCell(groupAttrPosArray[i]).getValue()).intValue());
				numConstantVals ++;
			}
		}
		boolean equalityQuery = true;
		int numAttrs = indexAttrPos.length;
		for (int i=0; i<numAttrs; i++) 
		{
			if (punctTypeArray[indexAttrPos[i]] != XATPunct.SINGLE_VALUE) 
			{
				equalityQuery = false;
				break;
			}
		}

		DataBucket[] dataBucketArray = null;
		if (equalityQuery) 
		{
			int bucketNo = hash(query.getArrayData());
			dataBucketArray = new DataBucket[] {dataBuckets[bucketNo]};
		} 
		else 
		{
			dataBucketArray = dataBuckets;
		}
		
		boolean bCont = true;
		int numResultBuckets = dataBucketArray.length;
		for (int i=0; i<numResultBuckets; i++) 
		{
			DataBucket dBucket = dataBucketArray[i];
			
			for (int j=0; j<dBucket.size(); j++) {
				
				XATTuple t = dBucket.getTuple(j);				
	    	    long tupleCounter2 = Long.parseLong(t.getCell(t.getLength() - 2).toString());
	    	    
    		    if ((tupleCounter2 >= tupleMinCounter) && (tupleCounter2 <= tupleMaxCounter))
				{	
    		    	if (this.matchVals(t.getArrayData(), numConstantVals, constantAttrVal, constantAttrPos))
					{	
//	    		    	  System.out.println(t.toString());
						  results.add(t);
	    		    }
        		    else
        		    	continue;
				}
    		    else if (tupleCounter2 > tupleMaxCounter)
    		    {
    		    	j=dBucket.size();
    		    	bCont = false;
    		    	continue;
    		    }
				
			} // end for j
		} // end for i
		return bCont;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.pullbasedgroupby.WindowGroupbyIndex#removeExpiredTuples(long)
	 */
	protected void removeExpiredTuples(long windowStart) {
		int numDataBuckets = dataBuckets.length;
		for (int i=0; i<numDataBuckets; i++) {
			int numDeletedTuples = dataBuckets[i].removeExpiredTuples(windowStart);
			totalNumTuples -= numDeletedTuples;
		}
	}
	private boolean matchVals(XATDataValue[] tupleVal, int numSearchKeys, 
			int[] searchKeys, int[] searchAttrPos) {
		for (int i=0; i<numSearchKeys; i++) {
			int intTupleVal = Math.abs(((Integer)tupleVal[searchAttrPos[i]].getValue()).intValue());
			if(intTupleVal != Math.abs(searchKeys[i])) {
				return false;
			}
		}
		return true;
	}

	public int[] getHashFactors() {
		return hashFactors; 
	}
	
	public void setHashFactors(int[] hFactors) {
		totalNumTuples = 0;
		int numHashEntries = 1;
		int numFactors = hFactors.length;
		hashFactors = new int[numFactors];
		for (int j=0; j<numFactors; j++) {
			hashFactors[j] = hFactors[j];
			numHashEntries *= hFactors[j];

		}
		dataBuckets = new DataBucket[numHashEntries];
		for (int j=0; j<numHashEntries; j++) {
			dataBuckets[j] = new DataBucket();
		}
		
		intervals = new int[numFactors];
		intervals[0] = numHashEntries / hashFactors[0];
		for (int j=1; j<numFactors; j++) {
			intervals[j] = intervals[j-1] / hashFactors[j];
		}
	}
	
	public void setIndexAttrPosArray(int[] iAttrPos) {
		int numAttrs = iAttrPos.length;
		indexAttrPos = new int[numAttrs];
		for (int j=0; j<numAttrs; j++) {
			indexAttrPos[j] = iAttrPos[j];
		}
	}
	
	public int[] getIndexAttrPosArray() {
		return indexAttrPos;
	}
	
	///////////////////////////////
	public int getNumWildcards() {
		return numWildcards;
	}

	////////////////////////////
	public void clearState() {
		// TODO Auto-generated method stub
		
	}

	//clear state
	public int clearState2() {
		int nNumTupsRemoved = 0;
		
		hashFactors = null;
		indexAttrPos = null;
		intervals = null;
		
		intervals= null;
		for (int i=0; i<dataBuckets.length; i++) {
			nNumTupsRemoved += dataBuckets[i].getTupleList().size();
			dataBuckets[i].getTupleList().clear();
		}
		return nNumTupsRemoved;
	}
	
}
