/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.STeM.Partition.IMPIndex;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import java.util.HashMap;
import java.util.Vector;
import java.util.Iterator;
import java.util.Random;
import java.util.ArrayList;

/**
 * @author lisading
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */

public class PriorityHashTreeIndex extends WindowGroupbyIndex {
	int numLevels;
	int[] indexAttrPosArray;
	HashFunction[] hashFuncs;
	int[] numIndexEntries;
	int[] intervals;
	int numDataBuckets;
	DataBucket[] dataBucketArray;
	
/////////////////////////////////////////
	//Deletion type
		//only set one type to true
	boolean bWindowDeletion = false;
	boolean bActProcTupsDeletion = true;

	public PriorityHashTreeIndex() {
		super();
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState#copyStateContent(edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState)
	 */
	public void copyStateContent(OperatorState newstate) {
		// TODO Auto-generated method stub

	}
	
	public void insertTuple(XATTuple tuple) {
		int offset = 0;
	
		//key
		for (int i=0; i<numLevels; i++) {
			int val = Math.abs(((Integer)tuple.getCell(indexAttrPosArray[i]).getValue()).intValue());
			
			int hashVal = hashFuncs[i].hash(val);
			offset += hashVal * intervals[i];
		}
		
		//insert tuple
		DataBucket dBucket = dataBucketArray[offset];
		dBucket.insertTuple(tuple);
		totalNumTuples ++;

	}

	public void insertTupleByTime(XATTuple tuple) {
		int offset = 0;
		for (int i=0; i<numLevels; i++) {
			int val = ((Integer)tuple.getCell(indexAttrPosArray[i]).getValue()).intValue();
			int hashVal = hashFuncs[i].hash(val);
			offset += hashVal * intervals[i];
		}
		DataBucket dBucket = dataBucketArray[offset];
		dBucket.insertTupleByTime(tuple);
		
		//long windowStart = Math.max(0, tuple.getMinTimeStamp()-window);
		//int numDeletedTuples = dBucket.removeExpiredTuples(windowStart);
		//totalNumTuples = totalNumTuples + 1 - numDeletedTuples;
	}
	public Vector probe(XATPunct query) {
		int[] offsets = new int[]{0};
		int numConstantVals = 0;
		int[] constantAttrPos = new int[numGroupAttrs];
		int[] constantAttrVal = new int[numGroupAttrs];
		
		
		for (int i=0; i<numGroupAttrs; i++) {
			if (query.getPunctType(groupAttrPosArray[i]) == XATPunct.SINGLE_VALUE) {
				constantAttrPos[numConstantVals] = groupAttrPosArray[i];
				constantAttrVal[numConstantVals] = ((Integer)query.getCell(groupAttrPosArray[i]).getValue()).intValue();
				numConstantVals ++;
			}			
		}
		
		
		for (int i=0; i<numLevels; i++) {
			int val = 0;
			if (query.getPunctType(indexAttrPosArray[i]) == XATPunct.SINGLE_VALUE) {
				val = Math.abs(((Integer)query.getCell(indexAttrPosArray[i]).getValue()).intValue());
				
				int hashVal = hashFuncs[i].hash(val);
				
				int increment = hashVal * intervals[i];
				int numOffsets = offsets.length;
				for (int k=0; k<numOffsets; k++) {
					offsets[k] += increment;
				}
			} else {
				int numOffsets = offsets.length;
				int[] newOffsets = new int[numOffsets*numIndexEntries[i]];
				int cursor = 0;
				for (int k=0; k<numOffsets; k++) {
					for (int m=0; m<numIndexEntries[i]; m++) {
						newOffsets[cursor] = offsets[k] + m * intervals[i];
						cursor ++;
					}
				}
				offsets = newOffsets;
			}
		}
		Vector result = new Vector();
		int numResultBuckets = offsets.length;
		
//removed kew		
//		long windowStart = Math.max(0, query.getMinTimeStamp() - window);
		HashMap tempResult = new HashMap();
		for (int i=0; i<numResultBuckets; i++) {
			DataBucket dBucket = dataBucketArray[offsets[i]];
			
			int numTuples = dBucket.size();
			tempResult.clear();
			for (int j=0; j<numTuples; j++) {
				XATTuple t = dBucket.getTuple(j);
				XATDataValue[] tupleValue = t.getArrayData();
				// check whether the tuple match the query pattern
				if (!this.matchVals(tupleValue, numConstantVals, constantAttrVal, constantAttrPos)) 
				{
					continue;
				}
				int[] vals = new int[numGroupAttrs];
				for (int k=0; k<numGroupAttrs; k++) 
				{
					vals[k] = ((Integer)tupleValue[groupAttrPosArray[k]].getValue()).intValue();
				}
				Integer tempHashKey = new Integer(vals[0]);
				Object obj = tempResult.get(tempHashKey);
				if (obj == null) {
					Vector tempBucket = new Vector();
					TempHashEntry tempEntry = new TempHashEntry(vals);
					tempEntry.insertTuple(t);
					tempBucket.add(tempEntry);
					tempResult.put(tempHashKey, tempBucket);
				} 
				else {
					boolean foundMatchEntry = false;
					Vector tempBucket = (Vector)obj;
					int numTempEntries = tempBucket.size();
					for (int m=0; m<numTempEntries; m++) 
					{
						TempHashEntry tempEntry = (TempHashEntry) tempBucket.get(m);
						if (tempEntry.ifMatch(vals)) 
						{
							foundMatchEntry = true;
							tempEntry.insertTuple(t);
							break;
						}
					}
					if (!foundMatchEntry) 
					{
						TempHashEntry tempEntry = new TempHashEntry(vals);
						tempEntry.insertTuple(t);
						tempBucket.add(tempEntry);
					}
				}
			} // end for j
			Iterator it = tempResult.values().iterator();
			while (it.hasNext()) 
			{
				Vector tempBucket = (Vector)it.next();
				int numTempEntries = tempBucket.size();
				for (int m=0; m<numTempEntries; m++) 
				{
					TempHashEntry tempEntry = (TempHashEntry)tempBucket.get(m);
					result.add(tempEntry);
				}
			}
		} // end for i
		return result;
	}
	
	public boolean probe2(XATPunct query, long tupleMaxCounter, long tupleMinCounter,ArrayList<XATTuple> results) 
	{
		int[] offsets = new int[]{0};
		int numConstantVals = 0;
		int[] constantAttrPos = new int[numGroupAttrs];
		int[] constantAttrVal = new int[numGroupAttrs];
		boolean bCont = true;
		
		for (int i=0; i<numGroupAttrs; i++) 
		{
			if (query.getPunctType(groupAttrPosArray[i]) == XATPunct.SINGLE_VALUE) 
			{
				constantAttrPos[numConstantVals] = groupAttrPosArray[i];
				constantAttrVal[numConstantVals] = Math.abs(((Integer)query.getCell(groupAttrPosArray[i]).getValue()).intValue());
				numConstantVals ++;
			}			
		}
		
		
		
		for (int i=0; i<numLevels; i++) 
		{
			int val = 0;
			if (query.getPunctType(indexAttrPosArray[i]) == XATPunct.SINGLE_VALUE) 
			{
				
				val = Math.abs(((Integer)query.getCell(indexAttrPosArray[i]).getValue()).intValue());
				
				int hashVal = hashFuncs[i].hash(val);				
				int increment = hashVal * intervals[i];
				int numOffsets = offsets.length;
				for (int k=0; k<numOffsets; k++) 
				{
					offsets[k] += increment;
				}				
			} 
			else 
			{
				int numOffsets = offsets.length;
				int[] newOffsets = new int[numOffsets*numIndexEntries[i]];
				int cursor = 0;
				for (int k=0; k<numOffsets; k++) 
				{
					for (int m=0; m<numIndexEntries[i]; m++) 
					{
						newOffsets[cursor] = offsets[k] + m * intervals[i];
						cursor ++;
					}
				}
				offsets = newOffsets;
			}
		}
		
//		System.out.println("offsets.length="+offsets.length);
//		System.out.println("dataBucketArray="+dataBucketArray);
		
		for (int i=0; i<offsets.length; i++) 
		{
			DataBucket dBucket = null;
			dBucket = dataBucketArray[offsets[i]];
			
			for (int j=0; j<dBucket.size(); j++) 
			{
				XATTuple t = dBucket.getTuple(j);
	    	    long tupleCounter2 = Long.parseLong(t.getCell(t.getLength() - 2).toString());
	    	    
    		    if ((tupleCounter2 >= tupleMinCounter) && (tupleCounter2 <= tupleMaxCounter))
				{	
    		    	if (this.matchVals(t.getArrayData(), numConstantVals, constantAttrVal, constantAttrPos))
					{	
//	    		    	  System.out.println(t.toString());
						  results.add(t);
	    		    }
    		    	else
        		    	continue;
				}
    		    else if (tupleCounter2 > tupleMaxCounter)
    		    {
    		    	j=dBucket.size();
    		    	bCont = false;    		    	
    		    	continue;
    		    }
			}	
		} // end for i
		
		return bCont;
	}

	public int deleteTuple(XATPunct query) {
		int[] offsets = new int[]{0};
		int numConstantVals = 0;
		int[] constantAttrPos = new int[numGroupAttrs];
		int[] constantAttrVal = new int[numGroupAttrs];
		
		int nNumTupsRemoved = 0;
		
		for (int i=0; i<numGroupAttrs; i++) {
			if (query.getPunctType(groupAttrPosArray[i]) == XATPunct.SINGLE_VALUE) {
				constantAttrPos[numConstantVals] = groupAttrPosArray[i];
				constantAttrVal[numConstantVals] = ((Integer)query.getCell(groupAttrPosArray[i]).getValue()).intValue();;
				numConstantVals ++;
			}			
		}
		
		
		for (int i=0; i<numLevels; i++) {
			int val = 0;
			if (query.getPunctType(indexAttrPosArray[i]) == XATPunct.SINGLE_VALUE) {
				val = Math.abs(((Integer)query.getCell(indexAttrPosArray[i]).getValue()).intValue());
				
				int hashVal = hashFuncs[i].hash(val);
				
				int increment = hashVal * intervals[i];
				int numOffsets = offsets.length;
				for (int k=0; k<numOffsets; k++) {
					offsets[k] += increment;
				}
			} else {
				int numOffsets = offsets.length;
				int[] newOffsets = new int[numOffsets*numIndexEntries[i]];
				int cursor = 0;
				for (int k=0; k<numOffsets; k++) {
					for (int m=0; m<numIndexEntries[i]; m++) {
						newOffsets[cursor] = offsets[k] + m * intervals[i];
						cursor ++;
					}
				}
				offsets = newOffsets;
			}
		}
		Vector result = new Vector();
		int numResultBuckets = offsets.length;
		
		
		long windowStart = Math.max(0, query.getMinTimeStamp() - window);
		HashMap tempResult = new HashMap(100);
		for (int i=0; i<numResultBuckets; i++) {
			DataBucket dBucket = dataBucketArray[offsets[i]];
			
			int numTuples = dBucket.size();
			tempResult.clear();
			int nPosOffset = 0;
			for (int j=0; j<numTuples; j++) {
				XATTuple t = dBucket.getTuple(j - nPosOffset);
				XATDataValue[] tupleValue = t.getArrayData();
				// check whether the tuple match the query pattern
				if (!this.matchVals(tupleValue, numConstantVals, constantAttrVal, constantAttrPos)) {
					continue;
				}
				else
				{	
					dBucket.removeTuple(j - nPosOffset);
					nPosOffset++;
					nNumTupsRemoved++;
				}	
			} // end for j
		} // end for i
		return nNumTupsRemoved;
	}
	
	public boolean delete(int partitionNumber) {
		//System.out.println("Deleting ..."+partitionNumber);
		int[] offsets = new int[]{0};
		int numConstantVals = 0;
		int[] constantAttrVal = new int[numGroupAttrs];
		
		//partition number
		constantAttrVal[numConstantVals] = partitionNumber;
		numConstantVals++;

		for (int i=0; i<numLevels; i++) {
			int val = 0;
			if (i == numLevels) {
				val = partitionNumber;
				int hashVal = hashFuncs[i].hash(val);
				int increment = hashVal * intervals[i];
				int numOffsets = offsets.length;
				for (int k=0; k<numOffsets; k++) {
					offsets[k] += increment;
				}
			} else {
				int numOffsets = offsets.length;
				int[] newOffsets = new int[numOffsets*numIndexEntries[i]];
				int cursor = 0;
				for (int k=0; k<numOffsets; k++) {
					for (int m=0; m<numIndexEntries[i]; m++) {
						newOffsets[cursor] = offsets[k] + m * intervals[i];
						cursor ++;
					}
				}
				offsets = newOffsets;
			}
		}
		//is databucket empty?
		int numResultBuckets = offsets.length;
		HashMap tempResult = new HashMap(100);
		boolean bCont = true;
		boolean bRemove = true;
		for (int i=0; i<numResultBuckets&bCont; i++) {
			DataBucket dBucket = dataBucketArray[offsets[i]];
			
			if (bRemove & (dBucket.numTuples==0))
			{	
				dBucket.removeOldestTuple();
				bRemove = false;
			}
			
			if (dBucket.numTuples==0)
				bCont = false;
			else
				dBucket = null;
			
		} // end for i
		return bCont;
	}
	
	public void delete() {
//		System.out.println("Deleting Bucket...");
		//randomly choose a data bucket and remove the earliest tuple from it.
		Random generator = new Random( 19580427 );
		int d = generator.nextInt(numDataBuckets)+1;
		while(dataBucketArray[d].numTuples>0)
		{
			d = generator.nextInt(numDataBuckets)+1;
		}
		dataBucketArray[d].removeOldestTuple();
	}
	
	private boolean matchVals(XATDataValue[] tupleVal, int numSearchKeys, 
			int[] searchKeys, int[] searchAttrPos) {
		for (int i=0; i<numSearchKeys; i++) {
			if(Math.abs(((Integer)tupleVal[searchAttrPos[i]].getValue()).intValue()) != Math.abs(searchKeys[i])) {
				return false;
			}
		}
		return true;
	}

	
	protected void removeExpiredTuples(long windowStart) {
		if(bWindowDeletion)
			for (int i=0; i<numDataBuckets; i++) {
				int numDeletedTuples = dataBucketArray[i].removeExpiredTuples(windowStart);
				totalNumTuples -= numDeletedTuples;
			}
	}

	public DataBucket[] getDataBuckets() {
		return dataBucketArray;
	}
	public int[] getIndexAttrPosArray() {
		return indexAttrPosArray;
	}
	public int getNumDataBuckets() {
		return numDataBuckets;
	}
	public int[] getNumIndexEntries() {
		return numIndexEntries;
	}
	public int getNumLevels() {
		return numLevels;
	}
	public void setDataBucketArray(DataBucket[] dBucketArray) {
		dataBucketArray = dBucketArray;
		numDataBuckets = dBucketArray.length;
	}
	public void setNumIndexEntries(int[] nIndexEntries) {
		numLevels = nIndexEntries.length;
		numIndexEntries = new int[numLevels];
		numDataBuckets = 1;
		for (int i=0; i<numLevels; i++) {
			numIndexEntries[i] = nIndexEntries[i];
			numDataBuckets *= nIndexEntries[i];
		}
		// initialize data buckets
		dataBucketArray = new DataBucket[numDataBuckets];
		for (int i=0; i<numDataBuckets; i++) {
			DataBucket dBucket = new DataBucket();
			dataBucketArray[i] = dBucket;
		}
		// initialize hash intervals
		intervals = new int[numLevels];
		intervals[0] = numDataBuckets / numIndexEntries[0];
		for (int i=1; i<numLevels; i++) {
			intervals[i] = intervals[i-1] / numIndexEntries[i];
		}
	}
	public void setIndexAttrPosArray(int[] iAttrPosArray) {
		int nLevels = iAttrPosArray.length;
		indexAttrPosArray = new int[nLevels];
		for (int i=0; i<nLevels; i++) {
			indexAttrPosArray[i] = iAttrPosArray[i];
		}
	}
	public void setHashFuncs(HashFunction[] hFuncs) {
		int nLevels = hFuncs.length;
		hashFuncs = new HashFunction[nLevels];
		for (int i=0; i<nLevels; i++) {
			hashFuncs[i] = hFuncs[i];
		}
	}
	
	public void clearState() {
	}
	
	public int clearState2() {
		// TODO Auto-generated method stub
		int nNumTupsRemoved = 0;
		indexAttrPosArray= null;
		hashFuncs = null;
		numIndexEntries= null;
		intervals= null;
		for (int i=0; i<numDataBuckets; i++) {
			nNumTupsRemoved += dataBucketArray[i].getTupleList().size();
			dataBucketArray[i].getTupleList().clear();
		}
		return nNumTupsRemoved;
	}
}
