/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

/*
 * Created on Feb 17, 2006
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.STeM.Partition.IMPIndexHeuristic;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * @author lisading
 *
 * [AU79] Optimal Partial-Match Retrieval When Fields are Independently Specified
 * Alfred Aho and Jeffrey Ullman
 * ACM Transactions on Database Systems, Vol. 4, No. 2, June 1979, Pages 168-179 
 */
public class IndependentPartitionHeuristic extends PartitionHeuristic {
	float[] factor1; //log(p/1-p)
	float[] factor2; //sum_i(log(p/1-p))
	float coefficient;
	
	public IndependentPartitionHeuristic() {
	}
	
	public int getNumConsideredAttrs() {
		return numConsideredAttrs;
	}
//	protected void computeQueryPatterns() {
//		numQueryPatterns = (int)Math.pow(2, numAttrs);
//		queryPatterns = new QueryPattern[numQueryPatterns];
//		for (int i=0; i<numQueryPatterns; i++) {
//			float qWeight = 1;
//			for (int j=0; j<numAttrs; j++) {
//				int mask = (int)Math.pow(2, j);
//				if ((i & mask) == mask) {
//					qWeight *= occurrenceWeight[j]; 
//				} else {
//					qWeight *= (1 - occurrenceWeight[j]);
//				}
//			}
//			queryPatterns[i] = new QueryPattern(i, qWeight);
//		}
//	}
	
	// coefficient P = (1-p_1)...(1-p_k)
	protected void computeCoefficient() {
		coefficient = 1;
		for (int i=0; i<numAttrs; i++) {
			coefficient *= (1-occurrenceWeight[i]);
		}
	}
	public void computeOptimalConfig() {
		computeCoefficient();
		//initialize optimalConfig 
		int[] bitsNoIndex = new int[numAttrs];
		for (int i=0; i<numAttrs; i++) {
			bitsNoIndex[i] = 0;
		}
		optimalConfig = new IndexConfig(bitsNoIndex, computeNoIndexCost());
		
		for (int i=0; i<numAttrs; i++) {
			int attrID = nextAttrWithMaxOW();
			if (attrID == -1) {
				return;
			}
			consideredAttrs[i] = attrID;
			
			float weight = occurrenceWeight[attrID];
			factor1[i] = (float)(Math.log((float)weight/(float)(1-weight))/Math.log(2));
			if (i==0) {
				factor2[i] = factor1[i];
			} else {
				factor2[i] = factor2[i-1] + factor1[i];
			}
			//compute bits
			int[] intBits = new int[numAttrs];
			float[] fractions = new float[numAttrs];
			Vector notRoundedAttrs = new Vector();
			for (int j=0; j<numAttrs; j++) {
				intBits[j] = 0;
				fractions[j] = 0;
				notRoundedAttrs.add(new Integer(j));
			}
			float totalFractionalVal = 0;
			for (int j=0; j<=i; j++) {
				float b = (float)numBits/(float)(i+1) 
						+ factor1[j] - (factor2[i]/(float)(i+1));
				intBits[consideredAttrs[j]] = (int)b;
				fractions[consideredAttrs[j]] = (b - (int)b);
				totalFractionalVal += (b - (int)b);
			}
			
			//adjust bits 
			int numRoundUps = Math.round(totalFractionalVal);
			for (int j=0; j<numRoundUps; j++) {
				//get next attribute with largest bit fraction
				int attrPos = 0;
				int attrWithMaxFraction = 0;
				float maxFraction = 0;
				int numNotRoundedAttrs = notRoundedAttrs.size();
				for (int k=0; k<numNotRoundedAttrs; k++) {
					int nextAttr = ((Integer)notRoundedAttrs.get(k)).intValue();
					if (fractions[nextAttr]>maxFraction) {
						attrPos = k;
						attrWithMaxFraction = nextAttr;
						maxFraction = fractions[nextAttr];
					}
				}
				intBits[attrWithMaxFraction] ++;
				notRoundedAttrs.remove(attrPos);
			}
			
			//stop when newly considered attr is assigned 0 bits
			if (intBits[consideredAttrs[i]]==0) { 
				numConsideredAttrs --;
				return;
			}
			// update cost - not including query hash cost
			// compute product
			float probeFactor = 1;
			for (int k=0; k<numAttrs; k++) {
				float p = occurrenceWeight[k];
				probeFactor *= (1 + (p/(1-p))/Math.pow(2, intBits[k]));
			}
			float tupleCost = numTuples * hashCost * (i+1);
			float queryCost = numQueries * compareCost * coefficient * probeFactor * stateSize;
			float cost = tupleCost + queryCost;
			if (cost < optimalConfig.getCost()) {
				optimalConfig = new IndexConfig(intBits);
				optimalConfig.setCost(cost);
				updateNonConsideredAttrs(attrID);
			} else {
				return;
			}
		}
	}
	public void optimizedComputeOptimalConfig() {
		
	}
	public void initFromFile(String configFile) {
		int nAttrs = 0, nBits = 0;

		try {
			String thisLine = null;  
			BufferedReader br = new BufferedReader(new FileReader(configFile));
			
			thisLine = br.readLine();
			StringTokenizer st = new StringTokenizer(thisLine, " ");
			st.nextToken();
			nAttrs = Integer.parseInt(st.nextToken());
			
			thisLine = br.readLine();
			st = new StringTokenizer(thisLine, " ");
			st.nextToken();
			nBits = Integer.parseInt(st.nextToken());
			
			occurrenceWeight = new float[nAttrs];
			nonConsideredAttrs = new Vector(nAttrs);
			for (int i=0; i<nAttrs; i++) {
				thisLine = br.readLine();
				st = new StringTokenizer(thisLine, " ");
				float weight = Float.parseFloat(st.nextToken());
				occurrenceWeight[i] = weight;
				nonConsideredAttrs.add(new Integer(i));
			}
		} catch(IOException e) {
			e.printStackTrace();
		}
		numAttrs = nAttrs;
		numBits = nBits;
		consideredAttrs = new int[numAttrs];
		numConsideredAttrs = 0;
		numNonConsideredAttrs = numAttrs;
		factor1 = new float[numAttrs];
		factor2 = new float[numAttrs];
	}
}
