/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

/*
 * Created on Feb 17, 2006
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.STeM.Partition.IMPIndexHeuristic;

import java.util.HashMap;
import java.util.Vector;
/**
 * @author lisading
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class OWQCLeadingHeuristic extends QCLeadingHeuristic {
	HashMap savedResult = new HashMap(1000);
	
	public OWQCLeadingHeuristic() {
		super();
	}
	// for reusing cost computations
	protected int mapAttrSet(int newAttrID) {
		int setID = 0;
		int i = 0;
		for ( ; i<numConsideredAttrs; i++) 
		{
			setID += consideredAttrs[i] * Math.pow(2, i);
		}
		setID += newAttrID * Math.pow(2, i);
		return setID;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.pullbasedgroupbyheuristics.PartitionHeuristics#computeOptimalConfigs()
	 */
	public void computeOptimalConfig() {
		//initialize optimalConfig 
		int[] bitsNoIndex = new int[numAttrs];
		for (int i=0; i<numAttrs; i++) 
		{
			bitsNoIndex[i] = 0;
		}
		
		float noIndexCost =  computeNoIndexCost();
		optimalConfig = new IndexConfig(bitsNoIndex, noIndexCost);
		
		computeOccurrenceWeight();
		// run OW-Leading
		while (numNonConsideredAttrs > 0) 
		{
			int attrID = nextAttrWithMaxOW();
			if (attrID == -1) 
			{
				break;
			}
			
			buildIndexConfigs(attrID);
			IndexConfig[] indexConfigs = configGroups[numConsideredAttrs][numBits].getConfigs();
			int numIndexConfigs = indexConfigs.length;
			
			boolean betterConfigFound = false;			
			for (int j=0; j<numIndexConfigs; j++) 
			{
				IndexConfig ic = indexConfigs[j];
				if (arrangeCost(ic)) {
					betterConfigFound = true;
				}
			}
			
			if (!betterConfigFound) 
			{
				break;
			} 
			else {
				savedResult.put(new Integer(mapAttrSet(attrID)), optimalConfig);
			}
			consideredAttrs[numConsideredAttrs] = attrID;
			numConsideredAttrs ++;
			updateNonConsideredAttrs(attrID);
		}
		
		IndexConfig OWOptimalConfig = new IndexConfig(optimalConfig.getNumBits(), 
				optimalConfig.getCost());
		
		//RESET 
		optimalConfig = new IndexConfig(bitsNoIndex, noIndexCost);
		configGroups = new IndexConfigGroup[numAttrs][numBits+1];
		consideredAttrs = new int[numAttrs];
		nonConsideredAttrs = new Vector(numAttrs);
		numConsideredAttrs = 0;
		numNonConsideredAttrs = numAttrs;
		
//		for (int i=0; i<numAttrs; i++) {
//			nonConsideredAttrs.add(new Integer(i));
//		}

		// run QC-Leading
		queryCovered = new boolean[numQueryPatterns];
		for (int i=0; i<numQueryPatterns; i++) 
		{
			queryCovered[i] = false;
		}
		
		queryCoverage = new float[numAttrs];
		for (int i=0; i<numAttrs; i++) 
		{
			queryCoverage[i] = occurrenceWeight[i];
			nonConsideredAttrs.add(new Integer(i));
		}
		while (numNonConsideredAttrs > 0) 
		{
			int attrID = nextAttrWithMaxQC();
			if (attrID == -1) 
			{
				break;
			}
			
			int attrSetID = mapAttrSet(attrID);
			Object obj = savedResult.get(new Integer(attrSetID));
			if (obj == null) 
			{
				buildIndexConfigs(attrID);
				IndexConfig[] indexConfigs = configGroups[numConsideredAttrs][numBits].getConfigs();
				int numIndexConfigs = indexConfigs.length;
				boolean betterConfigFound = false;
				for (int j=0; j<numIndexConfigs; j++) 
				{
					IndexConfig ic = indexConfigs[j];
					if (arrangeCost(ic)) 
					{
						betterConfigFound = true;
					}
				}
				if (!betterConfigFound) {
					break;
				}
			} else {
				if (numConsideredAttrs>0) {
					buildAssistantConfigGroups(); //for later use, TO BE CHECKED 
				}
				IndexConfig newConfig = (IndexConfig)obj;
				float newCost = newConfig.getCost();
				if (newCost<optimalConfig.getCost()) {
					newConfig.setCost(newCost);
					optimalConfig = newConfig;
				} else {
					break;
				}
			}

			consideredAttrs[numConsideredAttrs] = attrID;
			numConsideredAttrs ++;
			updateNonConsideredAttrs(attrID);
			updateQueryCoverage(attrID);
		}
		
		if (OWOptimalConfig.getCost()<optimalConfig.getCost()) 
		{
			optimalConfig = OWOptimalConfig;
		}
		
		//System.out.println("# Configs Checked: "+numConfigsChecked);
	}
	
	public void optimizedComputeOptimalConfig() {
		//initialize optimalConfig 
		int[] bitsNoIndex = new int[numAttrs];
		for (int i=0; i<numAttrs; i++) 
		{
			bitsNoIndex[i] = 0;
		}
		float noIndexCost =  computeNoIndexCost();
		optimalConfig = new IndexConfig(bitsNoIndex, noIndexCost);
		
		computeOccurrenceWeight();
		// run OW-Leading
		while (numNonConsideredAttrs > 0) 
		{
			int attrID = nextAttrWithMaxOW();
			if (attrID == -1) 
				break;
			
			buildIndexConfigs(attrID);
			IndexConfig[] indexConfigs = configGroups[numConsideredAttrs][numBits].getConfigs();
			int numIndexConfigs = indexConfigs.length;
			boolean betterConfigFound = false;
			for (int j=0; j<numIndexConfigs; j++) {
				IndexConfig ic = indexConfigs[j];
				int[] icBits = ic.getNumBits();
				boolean skip = false;
				for (int k=0; k<numConsideredAttrs; k++) {
					if (icBits[consideredAttrs[k]]>numBitsLimits[consideredAttrs[k]]) {
						skip = true;
						break;
					}
				}
				if (!skip) {
					if (arrangeCost(ic)) {
						betterConfigFound = true;
					}
				}
			}
			if (!betterConfigFound) {
				break;
			} else {
				savedResult.put(new Integer(mapAttrSet(attrID)), optimalConfig);
			}
			consideredAttrs[numConsideredAttrs] = attrID;
			for (int j=0; j<=numConsideredAttrs; j++) {
				numBitsLimits[consideredAttrs[j]] = 
					optimalConfig.getNumBits()[consideredAttrs[j]];
			}
			numConsideredAttrs ++;
			updateNonConsideredAttrs(attrID);
		}
		IndexConfig OWOptimalConfig = new IndexConfig(optimalConfig.getNumBits(), 
				optimalConfig.getCost());
		
		//RESET 
		optimalConfig = new IndexConfig(bitsNoIndex, noIndexCost);
		configGroups = new IndexConfigGroup[numAttrs][numBits+1];
		consideredAttrs = new int[numAttrs];
		nonConsideredAttrs = new Vector(numAttrs);
		numConsideredAttrs = 0;
		numNonConsideredAttrs = numAttrs;
//		for (int i=0; i<numAttrs; i++) {
//			nonConsideredAttrs.add(new Integer(i));
//		}

		// run QC-Leading
		queryCovered = new boolean[numQueryPatterns];
		for (int i=0; i<numQueryPatterns; i++) {
			queryCovered[i] = false;
		}
		queryCoverage = new float[numAttrs];
		for (int i=0; i<numAttrs; i++) {
			queryCoverage[i] = occurrenceWeight[i];
			nonConsideredAttrs.add(new Integer(i));
		}
		while (numNonConsideredAttrs > 0) {
			int attrID = nextAttrWithMaxQC();
			if (attrID == -1) 
				break;
			
			
			int attrSetID = mapAttrSet(attrID);
			Object obj = savedResult.get(new Integer(attrSetID));
			if (obj == null) {
				buildIndexConfigs(attrID);
				IndexConfig[] indexConfigs = configGroups[numConsideredAttrs][numBits].getConfigs();
				int numIndexConfigs = indexConfigs.length;
				boolean betterConfigFound = false;
				for (int j=0; j<numIndexConfigs; j++) 
				{
					IndexConfig ic = indexConfigs[j];
					int[] icBits = ic.getNumBits();
					boolean skip = false;
					for (int k=0; k<numConsideredAttrs; k++) 
					{
						if (icBits[consideredAttrs[k]]>numBitsLimits[consideredAttrs[k]]) 
						{
							skip = true;
							break;
						}
					}
					if (!skip) 
					{
						if (arrangeCost(ic)) 
						{
							betterConfigFound = true;
						}
					}
				}
				if (!betterConfigFound) 				
					break;
				
			} else {
				if (numConsideredAttrs>0) {
					buildAssistantConfigGroups(); //for later use, TO BE CHECKED 
				}
				IndexConfig newConfig = (IndexConfig)obj;
				float newCost = newConfig.getCost();
				if (newCost<optimalConfig.getCost()) {
					newConfig.setCost(newCost);
					optimalConfig = newConfig;
				} else 
					break;
				
			}

			consideredAttrs[numConsideredAttrs] = attrID;
			for (int j=0; j<=numConsideredAttrs; j++) {
				numBitsLimits[consideredAttrs[j]] = 
					optimalConfig.getNumBits()[consideredAttrs[j]];
			}
			numConsideredAttrs ++;
			updateNonConsideredAttrs(attrID);
			updateQueryCoverage(attrID);
		}
		
		if (OWOptimalConfig.getCost()<optimalConfig.getCost()) 
			optimalConfig = OWOptimalConfig;
		
		
		//System.out.println("# Configs Checked: "+numConfigsChecked);
	}
////////////////////////////////////////////
	// KEW
		public float computeQueryCost(IndexConfig ic,int nPR, int nBR) {
			int[] nBitsAssigned = ic.getNumBits();
			//System.out.println("numBits="+nBitsAssigned.length);
//			System.out.println(" Length nBitsAssigned[]="+nBitsAssigned.length);
			
			float cost = 0;
			computeOccurrenceWeightOnly();
			
			float coefficient = 1;
			for (int i=0; i<numAttrs; i++) {
				if(occurrenceWeight[i] != 1)
					coefficient *= (1-occurrenceWeight[i]);
//				System.out.println("i="+i+" occurrenceWeight[i]="+occurrenceWeight[i]);
			}
			
			float probeFactor = 1;
			int nBitPos = 0;
			for (int k=0; k<numAttrs; k++) {
				float p = occurrenceWeight[k];
				if ((occurrenceWeight[k] > 0) &&(nBitsAssigned[nBitPos]>0))
				{	
//					System.out.println("k="+k+" occurrenceWeight[k]="+occurrenceWeight[k]);
//					System.out.println("nBitPos="+nBitPos);
//					System.out.println(" nBitsAssigned[k]="+nBitsAssigned[k]);
					probeFactor *= (1 + (p/(1-p))/Math.pow(2, nBitsAssigned[k]));
//					System.out.println("k="+k+" nBitsAssigned[nBitPos]="+nBitsAssigned[nBitPos]);
					nBitPos++;
				}
			}
			
			
			//numQueries = # probe tups per window
			numQueries = nPR;
			//stateSize = # build tups per window
			stateSize = nBR;		
			//cost to compare
			compareCost=1;
			
			if(coefficient>0.0)
				cost = numQueries * compareCost * coefficient * probeFactor * stateSize;
			else
				cost = numQueries * compareCost * probeFactor * stateSize;
				
			return cost;
		}
	
		public float computeEstUseQueryCost(IndexConfig ic,int nPR, int nBR, float estUse) {
			int[] nBitsAssigned = ic.getNumBits();
			//System.out.println("numBits="+nBitsAssigned.length);
//			System.out.println(" Length nBitsAssigned[]="+nBitsAssigned.length);
			
			float cost = 0;
			computeOccurrenceWeightOnly();
			
			float coefficient = 1;
			for (int i=0; i<numAttrs; i++) {
				coefficient *= (1-occurrenceWeight[i]);
//				System.out.println("i="+i+" occurrenceWeight[i]="+occurrenceWeight[i]);
			}
			
			float probeFactor = 1;
			int nBitPos = 0;
			for (int k=0; k<numAttrs; k++) {
				float p = occurrenceWeight[k];
				if ((occurrenceWeight[k] > 0) &&(nBitsAssigned[k]>0))
				{	
//					System.out.println("k="+k+" occurrenceWeight[k]="+occurrenceWeight[k]);
//					System.out.println("nBitPos="+nBitPos);
//					System.out.println(" nBitsAssigned[k]="+nBitsAssigned[k]);
					probeFactor *= (1 + (p/(1-p))/Math.pow(2, nBitsAssigned[k]));
					//System.out.println("k="+k+" nBitsAssigned[nBitPos]="+nBitsAssigned[nBitPos]);
					nBitPos++;
				}
			}
			
			
			//numQueries = # probe tups per window
			numQueries = nPR;
			//stateSize = # build tups per window
			stateSize = nBR;		
			//cost to compare
			compareCost = 1;
			
			cost = (estUse* numQueries) * compareCost * coefficient * probeFactor * stateSize;
			return cost;
		}
		
		/////////////////////////////	
		public float computetupleCost(int nBR) {
			//cost to hash
			hashCost = 1;
			
			float tupleCost = numTuples * hashCost * numAttrs;
			return tupleCost;
		}	

}
