/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

/*
 * Created on Feb 17, 2006
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.STeM.Partition.IMPIndexHeuristic;

import java.util.ArrayList;
import java.util.Vector;

/**
 * @author lisading
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public abstract class PartitionHeuristic {
	float compareCost = (float)148;
	float hashCost=(float)39;
	float numTuples; // # tuples arrived within a time unit
	float numQueries; // # queries arrived within a time unit
	float window = (float)1.0; // size in # time units
	float stateSize;
	float bucketSize;

	int numAttrs;
	int numBits;
	float[] occurrenceWeight;
	IndexConfig optimalConfig;

	int numQueryPatterns;
	edu.wpi.cs.dsrg.STeM.Partition.IMPIndexHeuristic.QueryPattern[] queryPatterns;

	ArrayList <Integer> EvalQP;
	
	public int numConsideredAttrs;
	int[] consideredAttrs;
	int numNonConsideredAttrs;
	Vector nonConsideredAttrs;
	
	int numConfigsChecked = 0;

	public int getNumBits() {
		return numBits;
	}
	public IndexConfig getOptimalConfig() {
		return optimalConfig;
	}
	public int nextAttrWithMaxOW() {
		int attrWithMaxOW = ((Integer)nonConsideredAttrs.get(0)).intValue();
		float maxOW = occurrenceWeight[attrWithMaxOW];
		for (int i=1; i<numNonConsideredAttrs; i++) {
			int nextAttrID = ((Integer)nonConsideredAttrs.get(i)).intValue();
			float nextOW = occurrenceWeight[nextAttrID];
			if (nextOW > maxOW) {
				attrWithMaxOW = nextAttrID;
				maxOW = nextOW;
			}
		}
		if (maxOW == 0) {
			return -1;
		} else {
			return attrWithMaxOW;
		}
	}
	public void updateNonConsideredAttrs(int attrID) {
		for (int i=0; i<numNonConsideredAttrs; i++) {
			if (((Integer)nonConsideredAttrs.get(i)).intValue() == attrID) {
				nonConsideredAttrs.remove(i);
				numNonConsideredAttrs --;
				return;
			}
		}
	}
	public abstract int getNumConsideredAttrs();
	
	public void init(String configFile, float cCost, float hCost, 
			float nTuples, float nQueries, float win) {
		initFromFile(configFile);
		compareCost = cCost;
		hashCost = hCost;
		numTuples = nTuples;
		numQueries = nQueries;
		window = win;
		stateSize = numTuples * window;
		bucketSize = (float)(stateSize/2.0);
	}
	
	public void init(float cCost, float hCost, 
			float nTuples, float nQueries, float d,int numBits, int PartitionSize) {
		compareCost = cCost;
		hashCost = hCost;
		numTuples = nTuples;
		numQueries = nQueries;
		window = d;
//		stateSize = numTuples * window;
		stateSize = PartitionSize;
		bucketSize = (float)(stateSize/2.0);
		
	}
	
	public abstract void initFromFile(String configFile);
	public abstract void computeOptimalConfig();
	public abstract void optimizedComputeOptimalConfig();
	
	public long computeNumConfigs() {
		double numIndexConfigs = 0;
		int numIterations = Math.min(numAttrs, numBits);
		for (int i=1; i<=numIterations; i++) {
			double p = choose(numAttrs, i);
			double q = choose(numBits-1, i-1);
			numIndexConfigs += (p * q);
		}
		//System.out.println((long)numIndexConfigs);
		return (long)numIndexConfigs;
	}
	
	//choose n out of m - an approximate result
	private double choose (int n, int m) {
		double result=1;
		for (int i=0; i<m; i++)
			result *= ((double)(n-i)/(double)(m-i));
		return result;
	}

	public void setVals(int nQ, int nT, int w) {
		numQueries = nQ;
		numTuples = nT;
		window  = w;
	}

	protected float computeNoIndexCost() {
		//System.out.println("computeNoIndexCost="+numQueries * numTuples * window * compareCost);
		return numQueries * numTuples * window * compareCost;
	}
	public int getNumConfigsChecked() {
		return numConfigsChecked;
	}
}
