/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

/*
 * Created on Feb 13, 2006
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.STeM.Partition.IMPIndexHeuristic;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.Vector;
import edu.wpi.cs.dsrg.STeM.Partition.IMPIndexHeuristic.QueryPattern;
import edu.wpi.cs.dsrg.STeM.Partition.IMPIndexHeuristic.IndexConfig;

/**
 * @author lisading
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public abstract class QueryBasedPartitionHeuristic extends PartitionHeuristic {
	int[] numBitsLimits;
	
	int[] numPatternOccurred;
	int[][] occurredPatternID; // by attribute
	int[] numAttrOccurred;
	int[][] occurredAttrID; // by pattern
	
	IndexConfigGroup[][] configGroups;
	ArrayList <Integer> EvalQP;
	
	public QueryBasedPartitionHeuristic() {
		super();
		
		EvalQP = new ArrayList <Integer>();

	}
	public void initFromFile(String configFile) {
		int nAttrs = 0, nBits = 0;
		QueryPattern[] qptns = null;

		try {
			String thisLine = null;  
			BufferedReader br = new BufferedReader(new FileReader(configFile));
			
			thisLine = br.readLine();
			StringTokenizer st = new StringTokenizer(thisLine, " ");
			st.nextToken();
			nAttrs = Integer.parseInt(st.nextToken());
			
			thisLine = br.readLine();
			st = new StringTokenizer(thisLine, " ");
			st.nextToken();
			nBits = Integer.parseInt(st.nextToken());
			
			thisLine = br.readLine();
			st = new StringTokenizer(thisLine, " ");
			st.nextToken();
			int nPatterns = Integer.parseInt(st.nextToken());
			
			qptns = new QueryPattern[nPatterns];
			int[] pidArray = new int[nPatterns];
			float[] weightArray = new float[nPatterns];
			float totalWeight = 0;
			for (int i=0; i<nPatterns; i++) {
				thisLine = br.readLine();
				st = new StringTokenizer(thisLine, " ");
				st.nextToken();
				pidArray[i] = Integer.parseInt(st.nextToken());
				weightArray[i] = Float.parseFloat(st.nextToken());
				totalWeight += weightArray[i];
			}
			for (int i=0; i<nPatterns; i++) {
				qptns[i] = new QueryPattern(pidArray[i], weightArray[i]/totalWeight);
			}
		} catch(IOException e) {
			e.printStackTrace();
		}
		numAttrs = nAttrs;
		numBits = nBits;
		numBitsLimits = new int[numAttrs];
		occurrenceWeight = new float[numAttrs];
		consideredAttrs = new int[numAttrs];
		nonConsideredAttrs = new Vector(numAttrs);
		numConsideredAttrs = 0;
		numNonConsideredAttrs = numAttrs;
		for (int i=0; i<numAttrs; i++) {
			occurrenceWeight[i] = 0;
			nonConsideredAttrs.add(new Integer(i));
		}
		configGroups = new IndexConfigGroup[nAttrs][nBits+1];
		setQueryPatterns(qptns);
	}
	
	public void init(int nAttrs, int nBits) {
		numAttrs = nAttrs;
		numBits = nBits;
		numBitsLimits = new int[numAttrs];
		occurrenceWeight = new float[numAttrs];
		consideredAttrs = new int[numAttrs];
		nonConsideredAttrs = new Vector(numAttrs);
		numConsideredAttrs = 0;
		numNonConsideredAttrs = numAttrs;
		for (int i=0; i<numAttrs; i++) {
			occurrenceWeight[i] = 0;
			nonConsideredAttrs.add(new Integer(i));
		}
		configGroups = new IndexConfigGroup[nAttrs][nBits+1];
	}
	
	public void setQueryPatterns(QueryPattern[] qptns) {
		numQueryPatterns = qptns.length;

		numPatternOccurred = new int[numAttrs];
		occurredPatternID = new int[numAttrs][];
		EvalQP = new ArrayList<Integer>();
		
		for (int i=0; i<numAttrs; i++) {
			numPatternOccurred[i] = 0;
			occurredPatternID[i] = new int[numQueryPatterns];
			for (int j=0; j<numQueryPatterns;j++) 
			{
				occurredPatternID[i][j] = Integer.MAX_VALUE;
			}
		}
		
		numAttrOccurred = new int[numQueryPatterns];
		occurredAttrID = new int[numQueryPatterns][];
		queryPatterns = new QueryPattern[numQueryPatterns];
		for (int i=0; i<numQueryPatterns;i++) 
		{
			queryPatterns[i] = new QueryPattern(qptns[i]);
			if(queryPatterns[i].weight>0)
			{				
				EvalQP.add(i);
			}
			              
			numAttrOccurred[i] = 0;
			occurredAttrID[i] = new int[numAttrs];
			for (int j=0; j<numAttrs; j++) 
			{
				occurredAttrID[i][j] = Integer.MAX_VALUE;
			}
		}
	}
	
	// compute occurrenceWeight, initialize occurredInPattern array
	protected void computeOccurrenceWeight() {
		//EvalQP.clear();
		
		for (int i=0; i<numAttrs; i++) {
			int mask = (int)Math.pow(2, i);
			for (int j=0; j<numQueryPatterns; j++) 
			{
				QueryPattern qp = queryPatterns[j];
				if ((qp.getPatternID() & mask) == mask) 
				{
					occurrenceWeight[i] += qp.getWeight();
					//if(qp.getWeight()>0)
					//	EvalQP.add(j);				
					occurredPatternID[i][numPatternOccurred[i]] = j;
					numPatternOccurred[i] ++;
					numAttrOccurred[j] ++;
				}
			}
		}
	}
	
	// compute occurrenceWeight, initialize occurredInPattern array
	protected void computeOccurrenceWeightOnly() 
	{
		for (int i=0; i<numAttrs; i++) 
		{
			int mask = (int)Math.pow(2, i);
			for (int j=0; j<numQueryPatterns; j++) 
			{
				QueryPattern qp = queryPatterns[j];
				if ((qp.getPatternID() & mask) == mask) 
				{
					occurrenceWeight[i] += qp.getWeight();
				}
			}
		}
	}
	
	public int getNumConsideredAttrs() {
		return numConsideredAttrs;
	}
	
	//for heuristic-based enumeration
	protected void buildIndexConfigs(int attrID) {
		if (numConsideredAttrs == 0) 
		{
			int[] bitsAssigned = new int[numAttrs];
			
			for (int k=0; k<numAttrs; k++) 
			{
				bitsAssigned[k] = 0;
			}
			bitsAssigned[attrID] = numBits;
			IndexConfig ic = new IndexConfig(bitsAssigned);
			IndexConfigGroup icg = new IndexConfigGroup(new IndexConfig[] {ic});
			configGroups[0][numBits] = icg;
		} 
		else 
		{
			buildAssistantConfigGroups();
			Vector newConfigs = new Vector();
			for (int k=numConsideredAttrs; k<numBits; k++) 
			{
				IndexConfig[] cfg = configGroups[numConsideredAttrs-1][k].getConfigs();
				int nConfigs = cfg.length;
				for (int l=0; l<nConfigs; l++) {
//					int[] numBitsArray = cfg[l].getNumBits();
//					int[] newNumBitsArray = new int[numAttrs];					
//					for (int m=0; m<numAttrs; m++) {
//						newNumBitsArray[m] = numBitsArray[m];
//					}
					int[] newNumBitsArray =cfg[l].getNumBits().clone();
					
					newNumBitsArray[attrID] = numBits-k;
					IndexConfig newcfg = new IndexConfig(newNumBitsArray);
					newConfigs.add(newcfg);
				}
			}
			configGroups[numConsideredAttrs][numBits] = new IndexConfigGroup((IndexConfig[])newConfigs.toArray(new IndexConfig[]{}));
		}
	}
	protected void buildAssistantConfigGroups() {
		int prevAttrID = consideredAttrs[numConsideredAttrs-1];
		if (numConsideredAttrs == 1) {
			for (int i=1; i<numBits; i++) {
				int[] bitsAssigned = new int[numAttrs];
				for (int k=0; k<numAttrs; k++) {
					bitsAssigned[k] = 0;
				}
				bitsAssigned[prevAttrID] = i;
				IndexConfig ic = new IndexConfig(bitsAssigned);
				IndexConfigGroup icg = new IndexConfigGroup(new IndexConfig[] {ic});
				configGroups[0][i] = icg;
			}
		} 
		else 
		{
			int[] bitsAssigned = new int[numAttrs];
			for (int k=0; k<numAttrs; k++) {
				bitsAssigned[k] = 0;
			}
			
			for (int j=numConsideredAttrs-1; j<numBits; j++) 
			{
				Vector newConfigs = new Vector();
				for (int k=numConsideredAttrs-1; k<j; k++) 
				{
					IndexConfig[] cfg = configGroups[numConsideredAttrs-2][k].getConfigs();
					int nConfigs = cfg.length;
					for (int l=0; l<nConfigs; l++) 
					{
//						int[] numBitsArray = cfg[l].getNumBits();
//						int[] newNumBitsArray = new int[numAttrs];
//						for (int m=0; m<numAttrs; m++) {
//							newNumBitsArray[m] = numBitsArray[m];
//						}
						
						int[] newNumBitsArray = cfg[l].getNumBits().clone();
						
						newNumBitsArray[prevAttrID] = j-k;
						IndexConfig newcfg = new IndexConfig(newNumBitsArray);
						newConfigs.add(newcfg);
					}
				}
				IndexConfigGroup newicg = new IndexConfigGroup((IndexConfig[])newConfigs.toArray(new IndexConfig[]{}));
				configGroups[numConsideredAttrs-1][j] = newicg;
				newConfigs.clear();
			}
		}
	}
	
	//for heuristic-based enumeration
	protected void buildIndexConfigsLessBit(int attrID) {
		if (numConsideredAttrs == 0) {
			int[] bitsAssigned = new int[numAttrs];
			for (int k=0; k<numAttrs; k++) {
				bitsAssigned[k] = 0;
			}
			bitsAssigned[attrID] = numBits;
			IndexConfig ic = new IndexConfig(bitsAssigned);
			IndexConfigGroup icg = new IndexConfigGroup(new IndexConfig[] {ic});
			configGroups[0][numBits] = icg;
		} else {
			buildAssistantConfigGroupsLessBit();
			Vector newConfigs = new Vector();
			for (int k=numConsideredAttrs; k<numBits; k++) 
			{
				IndexConfig[] cfg = configGroups[numConsideredAttrs-1][k].getConfigs();
				int nConfigs = cfg.length;
				for (int l=0; l<nConfigs; l++) 
				{
//					int[] numBitsArray = cfg[l].getNumBits();
//					int[] newNumBitsArray = new int[numAttrs];
//					for (int m=0; m<numAttrs; m++) {
//						newNumBitsArray[m] = numBitsArray[m];
//					}
					
					int[] newNumBitsArray = cfg[l].getNumBits().clone();
					newNumBitsArray[attrID] = numBits-k;
					IndexConfig newcfg = new IndexConfig(newNumBitsArray);
					newConfigs.add(newcfg);
				}
			}
			configGroups[numConsideredAttrs][numBits] = new IndexConfigGroup((IndexConfig[])newConfigs.toArray(new IndexConfig[]{}));
			newConfigs.clear();
		}
	}
	protected void buildAssistantConfigGroupsLessBit() {
		int prevAttrID = consideredAttrs[numConsideredAttrs-1];
		if (numConsideredAttrs == 1) 
		{
			for (int i=1; i<numBits; i++) 
			{
				int[] bitsAssigned = new int[numAttrs];
				for (int k=0; k<numAttrs; k++) 
				{
					bitsAssigned[k] = 0;
				}
				bitsAssigned[prevAttrID] = i;
				IndexConfig ic = new IndexConfig(bitsAssigned);
				IndexConfigGroup icg = new IndexConfigGroup(new IndexConfig[] {ic});
				configGroups[0][i] = icg;
			}
		} else {
			int[] bitsAssigned = new int[numAttrs];
			for (int k=0; k<numAttrs; k++) {
				bitsAssigned[k] = 0;
			}
			for (int j=numConsideredAttrs-1; j<numBits; j++) 
			{
				Vector newConfigs = new Vector();
				
				for (int k=numConsideredAttrs-1; k<j; k++) 
				{
					IndexConfig[] cfg = configGroups[numConsideredAttrs-2][k].getConfigs();
					int nConfigs = cfg.length;
					for (int l=0; l<nConfigs; l++) {
//						int[] numBitsArray = cfg[l].getNumBits();
						
						int[] newNumBitsArray = new int[numAttrs];						
						newNumBitsArray = cfg[l].getNumBits().clone();
						
//						for (int m=0; m<numAttrs; m++) {
//							newNumBitsArray[m] = numBitsArray[m];
//						}
						
						newNumBitsArray[prevAttrID] = j-k;
						IndexConfig newcfg = new IndexConfig(newNumBitsArray);
						newConfigs.add(newcfg);
					}
				}
				IndexConfigGroup newicg = new IndexConfigGroup((IndexConfig[])newConfigs.toArray(new IndexConfig[]{}));
				configGroups[numConsideredAttrs-1][j] = newicg;
				newConfigs.clear();
			}
		}
	}

	protected boolean arrangeCost(IndexConfig ic) {
		numConfigsChecked ++;
		int[] nBitsAssigned = ic.getNumBits();
		
		int numAttrsIncluded = 0;
		for (int i=0; i<numAttrs; i++) {
			if (nBitsAssigned[i] > 0) {
				numAttrsIncluded ++;
			}
		}
		
		if (numAttrsIncluded == 0)
		 return false;
		
		float totalCost = 0;
		for (int i=0; i<numQueryPatterns; i++) {
			QueryPattern qp = queryPatterns[i];
			if (qp.weight == 0) {
				continue;
			}
			int accumulatedBits = 0;
			int pid = qp.getPatternID();
			int numIndexedAttrs = 0;
			for (int j=0; j<numAttrs; j++) {
				int mask = (int)Math.pow(2, j);
				if ((pid & mask) == mask && nBitsAssigned[j]>0) {
					accumulatedBits += nBitsAssigned[j];
					numIndexedAttrs ++;
				}
			}
			float probeFactor = ((float)qp.weight)/(float)Math.pow(2, accumulatedBits);
			float hashFactor = qp.weight * numIndexedAttrs;
			float queryCost = numQueries * (compareCost * probeFactor * bucketSize + hashCost * hashFactor);
			totalCost += queryCost;
		}
		
		totalCost += numTuples * numAttrsIncluded * hashCost; 
		
		if (totalCost <= optimalConfig.getCost())
		{	
			ic.setCost(totalCost);			
			optimalConfig = ic;
			
			return true;
		} else {
			return false;
		}
	}
		
	protected boolean arrangeCost2(IndexConfig ic) 
	{
		numConfigsChecked ++;
		
//		System.out.println(ic);
		
		int[] nBitsAssigned = ic.getNumBits();
		int numAttrsIncluded = 0;
		for (int i=0; i<numAttrs; i++) 
		{
			if (nBitsAssigned[i] > 0) 
				numAttrsIncluded ++;
		}
				
		
		if (numAttrsIncluded <= 0)
			 return false;
		
		float totalCost = 0;
		for (int i=0; i<EvalQP.size(); i++) 
		{
			QueryPattern qp = queryPatterns[EvalQP.get(i)];
			
			int pid = qp.getPatternID();
			
			int accumulatedBits = 0;
			int numIndexedAttrs = 0;
			
			for (int j=0; j<numAttrs; j++) 
			{
				int mask = (int)Math.pow(2, j);
				if ((pid & mask) == mask && nBitsAssigned[j]>0) 
				{
					accumulatedBits += nBitsAssigned[j];
					numIndexedAttrs ++;
				}
			}
			
			float probeFactor = ((float)qp.weight)/(float)Math.pow(2, accumulatedBits);
			float hashFactor = qp.weight;
			float queryCost = numQueries * (compareCost * probeFactor * bucketSize + hashCost * hashFactor);
			totalCost += queryCost;
			
//			System.out.println("qp.weight="+qp.weight);
//			System.out.println("compareCost="+compareCost);
//			System.out.println("bucketSize="+bucketSize);
//			System.out.println("probeFactor="+probeFactor);
//			System.out.println("hashFactor="+hashFactor);
//			System.out.println("queryCost="+queryCost);
			
		}
		
		//totalCost += numTuples * numAttrsIncluded * hashCost; 
		totalCost += numTuples * hashCost; 
		
		//System.out.println("totalCost="+totalCost);
		
		
		if (totalCost <= optimalConfig.getCost())
		{
			
			ic.setCost(totalCost);
			optimalConfig = ic;
			return true;
		} 
		else 
		{
			return false;
		}

	}
}
