/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.STeM.Partition.IMPIndexHeuristic;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.Random;

public class RandomQueryPatternGenerator {
	int numAttrs;
	int numBits;
	int numQueries;
	double param; 
	String patternFileName;
	int numPatterns;
	int[] patternID;
	int[] frequency;
	Random ptnSelector;
	
	public RandomQueryPatternGenerator(int nAttrs, int nBits, int nQueries, double p, String ptnFileName) {
		numAttrs = nAttrs;
		numBits = nBits;
		numQueries = nQueries;
		param = p;
		patternFileName = ptnFileName;
		numPatterns = (int)Math.pow(2, numAttrs);
		patternID = new int[numPatterns];
		frequency = new int[numPatterns];
		for (int i=0; i<numPatterns; i++) {
			frequency[i] = 0;
		}
	}
	public void setNumAttrs(int nAttrs) {
		numAttrs = nAttrs;
	}
	
	public void setPatternFileName (String ptnFileName){
		patternFileName = ptnFileName;
	}
	protected void generatePatterns() {
		ptnSelector = new Random(System.currentTimeMillis());
		for (int i=0; i<numPatterns; i++) {
			patternID[i] = ptnSelector.nextInt(numPatterns);
		}
	}
	public void generateByNormalDist() {
		Random normalDist = new Random(System.currentTimeMillis());
		int shift = numPatterns/2;
		for (int i=0; i<numQueries; i++) {
			int pos = (int)(normalDist.nextGaussian() * param)+ shift;
			if (pos>=0 && pos<numPatterns) {
				frequency[pos]++; //asc order
				//frequency[patternID[pos]]++; //random order
			}
		}
	}
	public void writePatternStatToFile() {
		try {
			BufferedWriter writer = new BufferedWriter(new FileWriter(patternFileName));
			writer.write("numAttrs " + numAttrs + "\n");
			writer.write("numBits " + numBits + "\n");
			writer.write("numPatterns " + numPatterns + "\n");
			for (int i=0; i<numPatterns; i++) {
				writer.write("a " + i + " " + frequency[i] + "\n");
			}
			writer.flush();
			writer.close();
		} catch(Exception e) {
			e.printStackTrace();
		}
		
	}
	public static void main(String[] args) {
		String flag = args[0];
		int nAttrs = Integer.parseInt(args[1]);
		int nBits = Integer.parseInt(args[2]);
		int nQueries = Integer.parseInt(args[3]);
		double p = Double.parseDouble(args[4]);
		String ptnFileName = args[5];
		RandomQueryPatternGenerator gen = new RandomQueryPatternGenerator(nAttrs, nBits, nQueries, p, ptnFileName);
		gen.generatePatterns();
		if (flag.equalsIgnoreCase("normal")) {
			gen.generateByNormalDist();
		} else {
			System.out.println("Wrong generator type!");
		}
		gen.writePatternStatToFile();
		System.out.println("Done.");
		
//		Random ran = new Random(System.currentTimeMillis());
//		for (int i=0; i<1000; i++) {
//			int k = (int)(ran.nextGaussian() * 200 + 512);
//			System.out.println(k);
//		}
	}
}
