/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.STeM.Partition;
import java.util.Iterator;
import java.util.NoSuchElementException;
import edu.wpi.cs.dsrg.STeM.Partition.*;
//Karen Works

//This structure stores the queue of partition index info

public class PartitionListQueue<Partition> implements Iterable<edu.wpi.cs.dsrg.STeM.Partition.Partition> 
{
	    private int N;         // number of elements on queue
	    private Node first;    // beginning of queue
	    private Node last;     // end of queue

	    // helper linked list class
	    public class Node 
	    {
	        public edu.wpi.cs.dsrg.STeM.Partition.Partition item;
	        public Node next;
	    }

	    // create an empty queue
	    public PartitionListQueue() 
	    {
	        first = null;
	        last  = null;
	    }

	    // is the queue empty?
	    public boolean isEmpty() 
	    { 
	    	return first == null; 
	    }
	    
	    public int length()      
	    { 
	    	return N;             
	    }
	    
	    public int size()        
	    { 
	    	return N;             
	    }

	    // add an item to the queue
	    public void enqueue(edu.wpi.cs.dsrg.STeM.Partition.Partition item) 
	    {
	        Node x = new Node();
	        x.item = item;
	        if (isEmpty()) 
	        { 
	        	first = x;     
	        	last = x; 
	        }
	        else           
	        { 
	        	last.next = x; last = x; 
	        }
	        N++;
	    }

	    public void enqueueFront(edu.wpi.cs.dsrg.STeM.Partition.Partition item) 
	    {
	    
	        Node x = new Node();
	        if (isEmpty()) 
	        { 
	        	x.item = item;
	        	first = x;     
	        	last = x; 
	        }
	        else           
	        { 
	        	//move old first
		        x.item = first.item;
		        x.next = first.next;
		        
		        first.item = item;
	        	first.next = x; 
	        	if(x.next == null)
	        	{
	        	  last = x;	
	        	}	        		
	        }
	        N++;
	    }

///////////////////////////
	    // get last item
	    public edu.wpi.cs.dsrg.STeM.Partition.Partition getLast() 
	    {
	        if (isEmpty()) throw new RuntimeException("Queue underflow");
	        return last.item;
	    }
	    
	    // get first item
	    public edu.wpi.cs.dsrg.STeM.Partition.Partition getFirst() 
	    {
	        if (isEmpty()) throw new RuntimeException("Queue underflow");
	        return first.item;
	    }
	    
///////////////////////////
	    // get last node
	    public Node getLastNode() 
	    {
	        if (isEmpty()) throw new RuntimeException("Queue underflow");
	        return last;
	    }
	    
	    // get first node
	    public Node getFirstNode() 
	    {
	        if (isEmpty()) throw new RuntimeException("Queue underflow");
	        return first;
	    }
	    
///////////////////////////////	    
	    // get partition related to partition number
	    public edu.wpi.cs.dsrg.STeM.Partition.Partition getPartition(int PartitionNum) 
	    {
	        Node n;
	        
	       
	        n = first;
	        if(n!=null)
	        {	
		        //System.out.println("PartitionNumber="+p1.getPartitionNumber());
	
		        while(n.item.getPartitionNumber() != PartitionNum && n.next !=null)
		        {
		        	n = n.next;
			        //System.out.println("PartitionNumber="+p1.getPartitionNumber());
		        }
	        
	        
		        if (n.item.getPartitionNumber() == PartitionNum)	
		        	return n.item;
		        else
		        	return null;
	        }
	        else
	        	return null;
	        
	    }
	        
	    // get partition related to partition number
	    public Node getPartitionNode(int PartitionNum) 
	    {
	        Node n;
	        
	       
	        n = first;
	        if(n!=null)
	        {	
		        //System.out.println("PartitionNumber="+p1.getPartitionNumber());
	
		        while(n.item.getPartitionNumber() != PartitionNum && n.next !=null)
		        {
		        	n = n.next;
			        //System.out.println("PartitionNumber="+p1.getPartitionNumber());
		        }
	        
	        
		        if (n.item.getPartitionNumber() == PartitionNum)	
		        	return n;
		        else
		        	return null;
	        }
	        else
	        	return null;
	        
	    }
	    
	    // get partition related to partition number
	    public edu.wpi.cs.dsrg.STeM.Partition.Partition getPartitionFront(int PartitionNum) 
	    {
	    	edu.wpi.cs.dsrg.STeM.Partition.Partition p1;
	        Node n;
//	        System.out.println("Looking for PartitionNumber="+PartitionNum);
	        
	       
	        n = first;
	        if(n!=null)
	        {	
		        p1 = n.item;
	//	        System.out.println("PartitionNumber="+p1.getPartitionNumber());
	
		        while(p1.getPartitionNumber() < PartitionNum && n.next !=null)
		        {
		        	n = n.next;
		            p1 = n.item;
	//		        System.out.println("PartitionNumber="+p1.getPartitionNumber());
		        }
		        
		        
		        if (p1.getPartitionNumber() == PartitionNum)	
		        	return p1;
		        else
		        	return null;
	        }
	        else
	        	return null;
	    }
	    
	    // remove and return the least recently added item
	    public edu.wpi.cs.dsrg.STeM.Partition.Partition dequeue() 
	    {
	    	
	        if (isEmpty()) throw new RuntimeException("Queue underflow");
	        edu.wpi.cs.dsrg.STeM.Partition.Partition item = first.item;
	        //System.out.println("Dequeued PartitionNumber="+item.getPartitionNumber());
	        first = first.next;
	        N--;
	        return item;
	    }
	    
	    
///////////////////////////////
	    // remove and return the least recently added item
	    public String toString() 
	    {
	        String s = "";
	        for (Node x = first; x != null; x = x.next)
	        {
	            s += x.item + " ";
	        }
	        return s;
	    }
	    
////////////////////////////////////
	    public Iterator<edu.wpi.cs.dsrg.STeM.Partition.Partition> iterator()  
	    { 
	    	return new QueueIterator();  
	    }

	    // an iterator, doesn't implement remove() since it's optional
	    public class QueueIterator implements Iterator<edu.wpi.cs.dsrg.STeM.Partition.Partition> 
	    {
	        private Node current = first;

	        public boolean hasNext()  
	        { 
	        	return current != null;                     
	        }
	        
	        public void remove()      
	        { 
	        	throw new UnsupportedOperationException();  
	        }

	        public edu.wpi.cs.dsrg.STeM.Partition.Partition next() 
	        {
	            if (!hasNext()) throw new NoSuchElementException();
	            edu.wpi.cs.dsrg.STeM.Partition.Partition item = current.item;
	            current = current.next; 
	            return item;
	        }
	    }
	}





