/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.STeM.Partition;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Iterator;
import java.util.BitSet;

import edu.wpi.cs.dsrg.Mjoin.MJoinData;
import edu.wpi.cs.dsrg.eddyBatch.datastructure.*;
import edu.wpi.cs.dsrg.eddyBatch.operator.EddyPartitionOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.*;

public class STeMMultiStreamPartitionJoinProbeOperatorImp extends XATMultiQueueWindowStreamOperatorImp implements STeMMultiStreamPartitionJoinProbeOperator
{	
	
	public int windowSize;
	long startTime;
	long endTime;
	
	long numProcessedTuples;
	
	// is this an eddy op or not
	private boolean IsEddyOp = false;
	
	//STQM: is self-tuning?
	private boolean IsSelfTuning = false;
	long lastUpdateTime = -999; //last time statistics of this operator was updated
	
	// index position of STem and tuple that are being compared
	
	// data set associated with STeM
	private STeMPartitionData STeMStreamData;
	
	//Info about probe multi stream
	//private ArrayList<Integer> ProbeStreamIds;
	private HashMap<Integer,ArrayList<Integer>> NumProbeTupleIndex;
	
	//Info about stored stem data
	private int DataSTREAMId;
	private HashMap<Integer,ArrayList<Integer>> NumSTeMIndex;
	//indices per tup
	private ArrayList<ArrayList<Integer>> NumTupleIndexs;
	
	//operator bitmaps per STREAM
	ArrayList<BitSet> PreFilledPerStream;
		
	//STeM Data Index Array Pos
	private int STeMDataIndexArrayPos;
	
	//used for QMesh to look up operators
//	private OperatorLookUp OpLkUp;
	
	private int QMeshOperatorID;

	//used to determine when to create ruster
	int rusterSendOff;


	//number of results to batch together
	private int NumTupToProcessPerOp = 1000;
	

	private boolean bCountActiveProcTups = true;
	private HashMap <Integer,Integer> PartitionProcTupCount;
	
	//non IMP index
	HashMap<String, Integer> IndexPos; 	
	ArrayList<ArrayList<Integer>> HashKeyIndex;  

	//output file
	String outFile = "EDDY_RESULTS.txt";
	private boolean bTrackSearchRequestExecutionTime = false;

	public STeMMultiStreamPartitionJoinProbeOperatorImp(XATNode init) 
	{
		super(init);
	}
		
	public boolean run() throws Exception 
	{		
		int totalNumberTupledEnq = 0;
		PartitionProcTupCount = new HashMap <Integer,Integer>(); 
		
		NumTupleIndexs = new ArrayList<ArrayList<Integer>>();

		//input queue from STREAM
		XATQueueSchema inSchema;
		
		//only set bits for first result
		boolean bFirstResult = true;


		//for eddy bit set
		BitSet dbs = null;
		BitSet rbs = null;
		BitSet CompleteJoinAttrib = null;
		
		BitSet InvDone;
		EddyQueue q = null;
		
		ArrayList<Integer> HasIndex = null;
		
		//------------------------------------------------------------------------------
		// get input queue by operator id from the lookup table
		//------------------------------------------------------------------------------
		//XATQueue inputQueue = (XATQueue)this.OpLkUp.getOperatorQueue(this.getOperatorID());
		XATQueue inputQueue = (XATQueue)this.getInputQueue();
						
//		System.out.println("\nProcessing stream [" + DataSTREAMId + "]");
//////////		/** get the tuples */
//		System.out.println("# of tuples in the queue: " + inputQueue.getQueueSize());
		XATTuple[] tuples = inputQueue.dequeue(inputQueue.getQueueSize(), this);
		
		numProcessedTuples += tuples.length;
		
//		System.out.println("\t# of tuples in the STEM (to be probed) hash: " + STeMStreamData.TupleList.size());	
//		System.out.println("Dequeued: " + tuples.length + " tuples.");				
//		//update property for statistics monitoring
//		this.setProperty("NUMBER_OF_TUPLES_TO_DEQUEUE", ""+tuples.length); 		
							
		
		ArrayList outTuples = new ArrayList();
		int NumExpectedTuplesInRuster = 0;
		int NumReceivedTuplesAfterRToken = 0;
				
		int extraTupleCounter = 0;
		XATTuple[] tuplesUntilRToken = null;
		int i = -1;
		int tuplesToProcess = tuples.length;
		boolean bAlreadyDequeuedCutoffRusterTuples = false;

		// Eddy Batching
		int numInBatch = 0;
		
		//partition num
		int PartitionNum = 0;
		
		//index pos non Imp index
		int indxPos = -99;

		//num search requests completed.
		int numSearchRequestsCompleted = 0;
		
		ArrayList <Integer> PosIndices = new ArrayList();
		
		
//		System.out.println("Probing state "+DataSTREAMId);
//		System.out.println("------------------");
//		System.out.println("tuplesToProcess"+tuplesToProcess);
		//for(int i = 0; i < tuples.length;i++)
		while(i++ < tuplesToProcess)
		{									
			XATTuple ReceivedTuple = null;
			ArrayList<XATTuple> ProbingTuples = new ArrayList<XATTuple>();
			
			if (i < tuplesToProcess)
			{
//				System.out.println("Extracting tuple");
				ReceivedTuple = tuples[i];
			}
			else
			{	
				if (numInBatch > 0)
				{
					if(tuplesUntilRToken==null)
					{	
						//dequeue the tuple left in the Batch
						tuplesUntilRToken = inputQueue.dequeue(numInBatch, this);
						//System.out.println("Beyond i ");
						extraTupleCounter = 0;
					}	
				}
				else
					break;
				
			}
						
			EddyTuple et = null, net = null;
			EddyBatchTuple ebt = null;
			
			
			// if this is an eddy op
			// parse the probe tuple from the eddy tuple
			if (IsEddyOp)
			{				
				if (numInBatch==0) //first tuple must be an EddyBatchTuple
				{
					//only set bits for first result
					bFirstResult = true;

					// send out all old tuples in q					
					if(q !=null)
						if(q.length()>0)
						{	
//							System.out.println("Sending results to Eddy..."+q.length());
							EddyBatchTuple ebtOut =(EddyBatchTuple)q.dequeue();
							ebtOut.setNumberTuplesinBatch(q.length()+1);							
					
							this.getOutputQueue().enqueue((XATTuple)ebtOut, this);						
							
							while(q.length()>0)
							   this.getOutputQueue().enqueue((XATTuple)q.dequeue(), this);						
						}	
					
					//set batch queue to empty
					q = new EddyQueue();
					
					//get first batch tuple;					
				 	ebt = (EddyBatchTuple)tuples[i];				 	
				 	numInBatch = ebt.getNumberTuplesinBatch();
				 	//System.out.println("numInBatch="+numInBatch);
				 					 	
				 	et = new EddyTupleImp(ebt.getNumTuplesInAnswer(),  ebt.getPartitionNumber());				 	
					et.setAnswer(ebt.getAnswer());
					
					// ready and done bits from tuple set to probe
					// with bits already set for this probe op
//					System.out.println("Rbs="+ebt.getDoneBitSet().getBits().toString());
					for(int h = 0;h < this.EddyGetOpBitsPos().size();h++)
					{
						ebt.ClearReadyBit(this.EddyGetOpBitsPos().get(h));
						ebt.SetDoneBit(this.EddyGetOpBitsPos().get(h));
					}
			
					dbs = (BitSet)ebt.getDoneBitSet().getBits().clone();
					rbs = (BitSet)ebt.getReadyBitSet().getBits().clone();
					
//					// get inverse of done bits
//					// ie this represents ops not done
					InvDone = (BitSet)dbs.clone();
					InvDone.flip(1, dbs.size());			
				
					// and Inv of done bits with the rbs for the stream in the
					// solution
					// ie this represents all not done and ready
					BitSet rbsSolution = (BitSet)PreFilledPerStream.get(DataSTREAMId).clone();			
					rbsSolution.and(InvDone);
									
					// or the rbs from the solution with the normal rbs
					// ie this represents all not done by the added STREAM
					// and all that were not completed before this operator was called
					rbs.or(rbsSolution);
//					
//					System.out.println("Rbs="+rbs.toString());
					
				}
				else if (i < tuplesToProcess)
				{	
					et = (EddyTuple)tuples[i];
				}
				else
				{
					et = (EddyTuple)tuplesUntilRToken[extraTupleCounter];
					extraTupleCounter++;
//					System.out.println("processing last tuples from batch");
				}
				
				//partition number
				PartitionNum = et.getPartitionNumber();
				//System.out.println("PartitionNum = "+PartitionNum);
				
				
				String indexKey = "";
				
				//probing tuples
				NumTupleIndexs= new ArrayList<ArrayList<Integer>>();
				PosIndices = new ArrayList<Integer>();
				  
				for(int np = 0; np<et.getNumTuplesInAnswer();np++)
				{	
					if(et.getTuple(np) != null)
					{							
						//ProbeStreamIds.add(np);
						if(NumSTeMIndex.get(np) != null)
						{	
							NumTupleIndexs.add((ArrayList<Integer>)NumSTeMIndex.get(np).clone());
							ProbingTuples.add(et.getTuple(np));							
						
							if((indexKey.length()>0) && (IndexPos != null))
							{	
								//add index of combo of indices before
								PosIndices.add(IndexPos.get(indexKey));
								//add index for index just added 
								PosIndices.add(IndexPos.get(np));
								indexKey+="-";
							}	
							indexKey+=np;
							
						}	
					 }	
				  }
				
				if (IndexPos != null)
					  indxPos = IndexPos.get(indexKey);
				
			 	//for testing purposes only
			 	//System.out.println("indxPos = "+indxPos);
//			 	System.out.println("HashKeyIndex = "+HashKeyIndex.get(indxPos).toString());			 	
//			 	System.out.println("Num ProbingTuples = "+ProbingTuples.size());
//			 	System.out.println("Num in Batch = "+numInBatch);
//			 	System.out.println("tuplesToProcess = "+tuplesToProcess);
			 	
			 	numInBatch--;

			}
			
			
			//----------------------------------------------------------------------------
			// GENERAL PROBE - FOR BOTH EDDY AND QUERY MESH
			//----------------------------------------------------------------------------
			// search the data for matches			
			//System.out.println("probing");
			ArrayList <XATTuple> results = STeMStreamData.ProbeMulti(ProbingTuples,
																	indxPos,
																	NumTupleIndexs,
																	PartitionNum,
																	PosIndices,
																	et.getMinTimeStamp());				
			if (results == null)
			{
				numSearchRequestsCompleted++;							
				ReceivedTuple = null;
					
				//Count all active processing tuples
				if (bCountActiveProcTups)
				{	
					if (PartitionProcTupCount.get(PartitionNum) != null)
						PartitionProcTupCount.put(PartitionNum,PartitionProcTupCount.get(PartitionNum)-1);
					else
						PartitionProcTupCount.put(PartitionNum,-1);
				}
					
			}
			else
			{
					
				//Count all active processing tuples
				if (bCountActiveProcTups)
				{	
					if (PartitionProcTupCount.get(PartitionNum) != null)
						PartitionProcTupCount.put(PartitionNum,PartitionProcTupCount.get(PartitionNum)+(results.size()-1));
					else
						PartitionProcTupCount.put(PartitionNum,results.size()-1);
				}
					
			 	//for testing purposes only
				//System.out.println("\nProcessing stream [" + DataSTREAMId + "]");
			 	//System.out.println("******************");
			 	//System.out.println("indxPos = "+indxPos);
				//System.out.println(results.size()+" results");
				//----------------------------------------------------------------------
				// EDDY OPERATOR ANSWER
				//----------------------------------------------------------------------					
				if (IsEddyOp)
				{
												
					for(int k = 0; k < results.size(); k++)
					{
						if(bFirstResult)
						{
							//create Eddy Batch Tuple
							EddyBatchTuple netEbt = new EddyBatchTupleImp(et.getNumTuplesInAnswer(), NumSTeMIndex.size(),et.getNumTuplesInAnswer(), et.getPartitionNumber());
							netEbt.setAnswer(et.getAnswer());
							netEbt.addTuple(DataSTREAMId, results.get(k));
								
							netEbt.setDoneBitSet(dbs);
							netEbt.setReadyBitSet(rbs);
							
							q.enqueue(netEbt); 
							bFirstResult = false;
								
						}
						else
						{	
							// create new eedy tuple for each intermediate answer
							net  = new EddyTupleImp(et.getNumTuplesInAnswer(),et.getPartitionNumber());
							net.setAnswer(et.getAnswer());
							net.addTuple(DataSTREAMId, results.get(k));
							
							//increment counter
							totalNumberTupledEnq++;
								
							//need to add tuple to batch send
							q.enqueue(net);

							//queue too full?
							if(q.length() >= NumTupToProcessPerOp)
							{
//								System.out.println("Sending results to Eddy..."+q.length());
								EddyBatchTuple ebtOut =(EddyBatchTuple)q.dequeue();
								ebtOut.setNumberTuplesinBatch(q.length()+1);
								this.getOutputQueue().enqueue((XATTuple)ebtOut, this);						
									
								while(q.length()>0)
									this.getOutputQueue().enqueue((XATTuple)q.dequeue(), this);
								
								//set batch queue to empty
								q = new EddyQueue();
								bFirstResult = true;
							}								
						}	
					}//end iterating through results
					et = null;
				}					
			}//end result is not empty
			ProbingTuples = new ArrayList<XATTuple>();
		}//end white (iterating through all dequeued tuples)		

		//any batched tuples need to be sent?
		if (IsEddyOp)
		{	
			// send out all old tuples in q
			if(q != null)
			{
				if(q.length()>0)
				{					
//					System.out.println("Sending results to Eddy..."+q.length());
					EddyBatchTuple ebt =(EddyBatchTuple)q.dequeue();
					ebt.setNumberTuplesinBatch(q.length()+1);
					this.getOutputQueue().enqueue((XATTuple)ebt, this);						
					while(q.length()>0)
					{
						//System.out.println("Sending results to Eddy...");
						this.getOutputQueue().enqueue((XATTuple)q.dequeue(), this);						
					}
				}	
			}
		
			q = null;			
		}
		
//		if(bTrackSearchRequestExecutionTime)
//		{
//			long currentTime =System.currentTimeMillis();
//			WriteToFile(outFile, String.valueOf(currentTime), false);
//			WriteToFile(outFile, "", false);
//			WriteToFile(outFile, String.valueOf(numSearchRequestsCompleted), true);
//		}
		
		//send out active proc tup count
		if (bCountActiveProcTups)
		{			
			EddyActiveProcTuple eapt = new EddyActiveProcTupleImp(PartitionProcTupCount);
			this.getOutputQueue().enqueue((XATTuple)eapt, this);
					
//			if(!bTrackSearchRequestExecutionTime)
//			{
//				long currentTime =System.currentTimeMillis();
//				WriteToFile(outFile, String.valueOf(currentTime), false);				
//				WriteToFile(outFile, String.valueOf(numSearchRequestsCompleted), false);
//				WriteToFile(outFile, "", false);
//				WriteToFile(outFile, String.valueOf(currentTime), true);				
//			}
		}			
	
////////////////////////////////////////////		
//		System.out.println("Enqueued: " + totalNumberTupledEnq + " tuples.");
//		System.out.println("CountActiveProcTups");
//		for (i=0; i <= 5; i++)
//		{
//			if (PartitionProcTupCount.get(i) != null)					
//			{
//				System.out.println(i+"="+PartitionProcTupCount.get(i));
//			}
//		}
//		System.out.println("------------------");
///////////////////////////////////////////////		
		
//		endTime = System.currentTimeMillis();
//		long duration = endTime - startTime; //milliseconds		
//WriteToFile(outFile, numProcessedTuples + "\t" + String.valueOf(duration), true);
		
		PartitionProcTupCount = null;
		//System.gc();
		return true;
	}
////////////////////////////////////
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return false;
	}

	public boolean cleanAllAccumulatedTuples() 
	{
		return false;
	}
	
///////////////////////////////////	
	public boolean matchStates(XATStreamOperator op) 
	{
		return false;
	}

	public boolean hasUnmatchedState() 
	{
		return false;
	}

	public boolean recomputeUnmatchedState() 
	{
		return false;
	}

	public int setStateID(int startID) 
	{
		return 0;
	}
///////////////////////////////////

	public String toToolTipString() 
	{
		return "STeMJoinProbeOperator";
	}

	public boolean finalizeOperation() 
	{
		return false;
	}
///////////////////////////////////

	public void setSTeMStreamData(STeMPartitionData sd)
	{
		STeMStreamData = sd;
	}
	
	public void SetIsEddyOp(boolean b)
	{
		IsEddyOp = b;
	}
	
/////////////////////////////////////////////////	
	public void setNumTupToProcessPerOp(int n)
	{
		NumTupToProcessPerOp = n;
	}
	
/////////////////////////////////////////////////
	//STQM
	public void SetIsSelfTuning(boolean b)
	{
		IsSelfTuning = b;
	}
	
	//STQM
	public boolean GetIsSelfTuning()
	{
		return IsSelfTuning;
	}
	
//	///////////////////////////////////////////
//	//Info about probe
//	public void addProbeSTREAMId(int i)
//	{
//		if(ProbeSTREAMIds==null)
//			ProbeSTREAMIds = new ArrayList<Integer>(); 
//		ProbeSTREAMIds.add(i);
//	}
//
//	public ArrayList<Integer> getProbeSTREAMIds()
//	{
//		return ProbeSTREAMIds;
//	}
///////////////////////////////////////////////////
	public void SetNumTupleIndex(int Id ,String Index) 
	{
		
		if (NumProbeTupleIndex==null)			
			NumProbeTupleIndex = new HashMap<Integer,ArrayList<Integer>>();
		
		NumProbeTupleIndex.put(Id,new ArrayList<Integer>());
		
		ArrayList<Integer> Indices = NumProbeTupleIndex.get(Id);
			
		StringTokenizer st = new StringTokenizer(Index);
	    while (st.hasMoreElements()) 
	    {
	    	Indices.add(Integer.parseInt(st.nextElement().toString()));
	    }	
	}

	public ArrayList<Integer> getNumTupleIndex(int StreamId)
	{
		return NumProbeTupleIndex.get(StreamId);
	}

	///////////////////////////////////////////
	//Info about stored stem data
	
	public void SetDataSTREAMId(int i)
	{
		DataSTREAMId = i;
	}

	public int getDataSTREAMId()
	{
		return DataSTREAMId;
	}
	
//////////////////////////////////////////////
	public void SetSTeMDataIndexArrayPos(int i)
	{
		STeMDataIndexArrayPos = i;
	}

	public int getSTeMDataIndexArrayPos()
	{
		return STeMDataIndexArrayPos;
	}
//////////////////////////////////////////////////	
	public void SetNumSTeMIndex(int Id ,String Index) 
	{
		
		if (NumSTeMIndex==null)			
			NumSTeMIndex = new HashMap<Integer,ArrayList<Integer>>();
		
		NumSTeMIndex.put(Id,new ArrayList<Integer>());
		
		ArrayList<Integer> Indices = NumSTeMIndex.get(Id);
			
		StringTokenizer st = new StringTokenizer(Index);
	    while (st.hasMoreElements()) 
	    {
	    	Indices.add(Integer.parseInt(st.nextElement().toString()));
	    }	
	}

	public ArrayList<Integer> getNumSTeMIndex(int StreamId)
	{
		return NumSTeMIndex.get(StreamId);
	}
//////////////////////////////////////////////////
//	public void SetCompleteJoinAttribs(String val) 
//	{
//		CompleteJoinAttribs = new ArrayList();
//		StringTokenizer st = new StringTokenizer(val);
//	    while (st.hasMoreElements()) 
//	    {	    
//	    	CompleteJoinAttribs.add(Integer.parseInt(st.nextElement().toString()));
//	    }
	    
//		System.out.println("CompleteJoinAttribs=");
//		for (int nc = 0; nc<CompleteJoinAttribs.size();nc++)
//			 System.out.println(CompleteJoinAttribs.get(nc));
	    
//	}
//
//	public ArrayList<Integer> getCompleteJoinAttribs()
//	{
//		return CompleteJoinAttribs;
//	}
	
	
//////////////////////////////////////////////////////
	public void setPreFilledPerStream(ArrayList<BitSet> p)
	{
		PreFilledPerStream = p;
	}	
	
	// table for look up operators used by QMesh
//	public void setOperatorLookUp(OperatorLookUp z)
//	{
//		OpLkUp = z;
//	}
////////////////////////////////////////////////////////	
	public void setQMeshOperatorID(int id) 
	{
		QMeshOperatorID = id;
	}
	
	public int getQMeshOperatorID() 
	{
		return QMeshOperatorID;
	}
//////////////////////////////////////////////////////////	
	public void setSendOff(int so)
	{
		rusterSendOff = so;
	}	
	
	//used for experiments (otherwise comment it out)
	public static void WriteToFile(String fileLocation, String stringToWrite, boolean doNewLine)
	{
		try
		{
			//Create file 
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			if (doNewLine)
			{
				out.newLine();
			}
			else
				out.append("|");
			
			
			//Close the output stream
			out.flush();
			out.close();
			
			
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}
	
	//STQM
	public long GetLastUpdateTime()
	{
		return lastUpdateTime;
	}

	//STQM
	public void SetLastUpdateTime(long newUpdateTime)
	{
		lastUpdateTime = newUpdateTime;
	}

	//////////////////////////////////////
	//non IMP index
	public void setIndexPos(String val,int indx)
	{
		if(IndexPos==null)
			IndexPos = new HashMap<String, Integer>();
		
		IndexPos.put(val, indx);	    
	}

	public void setHashKeyIndex(String val)
	{
		if(HashKeyIndex==null)
			HashKeyIndex = new ArrayList<ArrayList<Integer>>();
		
		ArrayList<Integer> IndexAttribs = new ArrayList<Integer>();
		HashKeyIndex.add(IndexAttribs);
		
		StringTokenizer st = new StringTokenizer(val);
	    while (st.hasMoreElements()) 
	    {
	    	//System.out.println(st.nextElement());
	    	IndexAttribs.add(Integer.parseInt(st.nextElement().toString()));
	    }
		
	}

	public ArrayList<ArrayList<Integer>> getHashKeyIndex()
	{
		return HashKeyIndex;
	}
	
/////////////////////////////////////////////////
	public void SetoutFile(String val) 
	{
		outFile = val;
	}
	
	public void SetTrackSearchRequestExecutionTime(boolean b)
	{
		bTrackSearchRequestExecutionTime = b;
	}
	

}
