/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.STeM.Partition;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Random;
import java.util.Vector;
import java.util.Iterator;
import java.util.BitSet;
import java.util.Set;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.STeM.Partition.IMPIndex.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.TupleQueue.Node;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidArgumentException;
import edu.wpi.cs.dsrg.eddyBatch.datastructure.*;
import edu.wpi.cs.dsrg.eddyPartition.operator.*;
import edu.wpi.cs.dsrg.STeM.Partition.IMPIndexHeuristic.*;
import edu.wpi.cs.dsrg.STeM.Partition.IMPIndex.MultiHashKeyIndex;

//Karen Works

//This structure stores the data the STEM searches on 

public class STeMPartitionData 
{

	int MAX_WINDOW_SIZE = 250;
	// maximum number of tuples to store
	//need to ask what value should be??

	//links to tuples that are referenced
	private HashMap<Integer,ArrayList<MultiHashKeyIndex>> StoredTuples;
	private ArrayList<MultiHashKeyIndex> CompleteStoredTuples;
	private ArrayList<HashIndexData> CurrentHashIndexData;

///////////////////////////////////////////	
// if not using the single IMP index	
	// number of indexes this table is joined against
	int intNumIndexes; 
	//Array of positions in tuple that need to be indexed on
	ArrayList<ArrayList<Integer>> HashKeyIndex;  

	ArrayList<Integer> CurrentHashKeys;  
	ArrayList<Integer> NextHashKeys;  
///////////////////////////////////////////	
	
	// Array number of the data source
	int intSTREAMId; 
	
	EddyQueue CompleteTupleList;
	HashMap<Integer,EddyQueue> TupleList;
	// Current number of stored tuples
	int intCurrNumberofStoredTuples = 0; 

/*****************************************/	
	//partition data
	
	// partition queue
	PartitionListQueue PartitionList;
	
	//current Partition Number
	int currentPartitionNumber = 0;
	
	//earliest Partition Number
	int earliestPartitionNumber = 1;
	
	//max partition wibdow size
	int MAX_PARTITION_WINDOW_SIZE = 2;
	
	// maximum number of tuples in a partition
	int MAX_PARTITION_SIZE = 800;

	boolean bPartition = true;
	
/*****************************************/
	//IMP Index
	PriorityHashTreeIndex NextIMPIndex;
	PriorityHashTreeIndex IMPIndex;
	float pIMPIndexCost;
	QueryPattern[] IMPIndexqp;
	boolean bImpIndex = false;

	boolean bLimitedNumberOfIndices = false;
	int nNumLimitedIndices = 0;
	
/*****************************************/
//Index Assessment
	boolean bDoesAssessment = true;
	float[] FreqQueryPatterns;	
	boolean bFirstAssessment = true;
	
	double dComparisoncost;
	double dHashCost;
	
	//////////////////////////////////
	private ArrayList<Integer> AccessModuleIndexAssessment;  
	
	private HashMap<Integer, IndexAssessment> SRIA;	
	
	////////////////////////////////////
	private boolean bStoreActualSRIA = false;
	private HashMap<Integer, IndexAssessment> actualSRIA;
	private String strStoreActualIndicesFile;
	////////////////////////////////////
	private boolean bStoreComparisonTime = false;
	private String strStoreComparisonTimeFile;
	
	////////////////////////////////////
	private boolean bSRIA = true;
	private boolean bHH = false;
	
	private int nBucketId = 0;
	
	private double dHHErrorRate = 0.01;
	private double dCostFactorSignificance = 0.40;

	private boolean bHHH = true;
	private ArrayList<IndexAssessment> LatticeLeafNodes;
	private boolean bCombineNodesByPickingOne = true;
	private boolean bPickParentWithHighestFreq = false;
	
/*****************************************/
//Exhaustive Index Assessment
	private boolean bExhaustiveAssessment = false;
	private int nNumQueryPatterns = 0;
	private int nNumCols = 8;
	private int nTotalNumQueryRequests = 0;
	private int nSubTotalNumQueryRequests = 0;
	private int nTotalNumBuildRequests = 0;
	
/*****************************************/
////Continuous Index Assessment
	private boolean bContinuousAssessment = true;
	
/*****************************************/
//Index Assessment Sampling
	private boolean bNoSampling;
	/////////////////////////////////////////
	private boolean bSimpleRandomSampling;
	private int nSimpleRandomPercent;    //value between 0-100
	////////////////////////////////////////////////
	private boolean bStratifiedRandomSampling;
	private int nStratifiedRandomPercent; //value between 0-100 
	private int nStratifiedRandomWindow;
	private int nNumStratifiedRandomSampled = 0;
	private int nTotalStratifiedRandomToSample = 0;
	/////////////////////////////////////
	private boolean bSystematicSampling;
	private int nSystematicInterval; 
	
	//////////////////////////////
	private Random generator; //random num generator
	
	/////////////////////////////////////////
	//Deletion type
		//only set one type to true
	private boolean bWindowDeletion = false;
	private boolean bActProcTupsDeletion = true;
	
/*****************************************/
//Index Migration
	private boolean bNoMigration=false;
	private boolean bSymmetricMigration=true;
	private boolean bPartitionMigration=false;
	private boolean bCompleteMigration=false;

/*****************************************/
//Combined stats
	private boolean bCombinedStats=true;
	private int nNumCombinedStats;
	private double[] CombinedStatsFactor; 
	private Vector storedStats = new Vector();
/*****************************************/
//Index Selection trigger
	private boolean bNoIndexSelectionTrigger=true;
	private boolean bCostIncrIndexSelectionTrigger=false;
	private boolean bFreqIncrIndexSelectionTrigger=false;
	private boolean bCombinedIncrIndexSelectionTrigger=false;
	
	private double dDegreeCostIncr = 0.0;
	private double dDegreeFreqIncr =0.0;
	
/*****************************************/
//count migrations
	private boolean bCountMigrations=false;
	private String strCountMigration="";
	
/*****************************************/
    XATPunct EmptyPunct;
    int[] EmptyPunctTypeArray;
    long[] EmptyTimeStamp;

    /*****************************************/
	public void init(Integer STREAMId,ArrayList<ArrayList<Integer>> HashKeys )
	{
	
		nNumQueryPatterns = (int)(Math.pow(2.0,nNumCols*1.0))-1;
		
		//computeComparisoncost();		
		//computeHashcost();		
		dComparisoncost = 128.0;
		dHashCost = 43.0;
		
		//////////////////////////////
        EmptyPunct = new XATPunctImp(nNumCols);
        EmptyPunctTypeArray = new int[nNumCols];
        EmptyTimeStamp = new long[nNumCols];
        
        for (int j=0; j<nNumCols-1; j++) 
        {
        	EmptyPunct.setCell(j, new XATDataValueImp(new String("*")));
        	EmptyPunctTypeArray[j] = XATPunct.ALL_VALUE;
        	EmptyTimeStamp[j] = 0;
        }	
		//////////////////////////////
		
		//IMP Index
		if ( bImpIndex)
		{	
			intSTREAMId = STREAMId;
			IMPIndex = new PriorityHashTreeIndex();

			/////////////////////////////
			if(bSRIA)
				SRIA = new HashMap<Integer, IndexAssessment>();
			
			if(bStoreActualSRIA)
			  actualSRIA = new HashMap<Integer, IndexAssessment>();			

			else if(bHHH)
			{	
				SRIA = new HashMap<Integer, IndexAssessment>();
				LatticeLeafNodes = new ArrayList<IndexAssessment>();
			}
			else
				FreqQueryPatterns = new float[nNumQueryPatterns];

			IMPIndex.setWindow(MAX_PARTITION_WINDOW_SIZE*2);
			IMPIndex.setMaxNumTuples(MAX_PARTITION_WINDOW_SIZE*2);
						
			if(bDoesAssessment)
			{
				if(bContinuousAssessment)
					startExhaustiveAssessment();
			}		
			if (bPartition)
			{	
				//Partition List
				PartitionList = new PartitionListQueue();
				NextIMPIndex = null;
			}
			
		}
		else	//non IMP index
		{
			// set up required variables
			intNumIndexes = HashKeys.size();
			
			AccessModuleIndexAssessment = new ArrayList<Integer>();  
			CurrentHashKeys = new ArrayList<Integer>();  
			NextHashKeys = new ArrayList<Integer>();  

			//Partition List
			PartitionList = new PartitionListQueue();
						
			intSTREAMId = STREAMId;
			HashKeyIndex = HashKeys;

			CurrentHashIndexData = new ArrayList<HashIndexData>();

			if (bPartition)
			{	
				TupleList = new HashMap<Integer,EddyQueue>();
				TupleList.put(1,new EddyQueue());

				StoredTuples = new HashMap<Integer,ArrayList<MultiHashKeyIndex>>();
				StoredTuples.put(1,new ArrayList<MultiHashKeyIndex>(nNumLimitedIndices));

				for (int i =0;i<nNumLimitedIndices;i++)
				{
					CurrentHashKeys.add(i);
				}
				
								
			}
			else
			{	
				CompleteTupleList = new EddyQueue();
				CompleteStoredTuples  = new ArrayList<MultiHashKeyIndex>();
				
				for (int i =0;i<intNumIndexes;i++)
				{
					CompleteStoredTuples.add(i,new MultiHashKeyIndex());
				}
				
			}
			
			if (bPartition)
			{	
				//Partition List
				PartitionList = new PartitionListQueue();
				//add first partition
				currentPartitionNumber++;
				Partition p= new PartitionImp(currentPartitionNumber,null,CurrentHashKeys);		
				PartitionList.enqueue(p);
			}
									
			//to test freq of access patterns only
			FreqQueryPatterns = new float[nNumQueryPatterns];
		}

  		generator = new Random();
  		generator.setSeed(new Date().getTime());
	}
	
//////////////////////////////////////////////////	
	public void Insert(XATTuple tuple)
	{
		//incr Total Num Build Requests
		nTotalNumBuildRequests++;

		intCurrNumberofStoredTuples++;
    			
		if (!bImpIndex)
		{			
				//			//for testing only
//							System.out.println("**** Adding *****");
//							System.out.println("STREAM"+intSTREAMId);
//							System.out.println("tuple="+tuple.toString());
//							System.out.println("*********");
				
    		int tuplePartitionNumber = Integer.parseInt(tuple.getCell(tuple.getLength()-1).convertToString());
			
			if (bPartition)
			{
				if (tuplePartitionNumber > currentPartitionNumber)
				{	
					currentPartitionNumber = tuplePartitionNumber;
					if (NextHashKeys.size()>0)
					{	
					   CurrentHashKeys = (ArrayList<Integer>)(NextHashKeys.clone());  
					   NextHashKeys.clear();  
					}
					
				    Partition p= new PartitionImp(currentPartitionNumber,null,CurrentHashKeys);		
				    PartitionList.enqueue(p);	
				}
				
				
				//are we storing anything for this partition yet?
				if (StoredTuples.get(tuplePartitionNumber) == null)
				{
					
					StoredTuples.put(tuplePartitionNumber,new ArrayList<MultiHashKeyIndex>(nNumLimitedIndices));

					for (int i =0;i<nNumLimitedIndices;i++)
					{	
						int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};

						HashIndexData hid = CurrentHashIndexData.get(i);
						MultiHashKeyIndex hMap = new MultiHashKeyIndex();
						hMap.setGroupAttrPosArray(hid.GroupAttrPosArray);
						hMap.setIndexAttrPosArray(hid.IndexAttrPosArray);
						hMap.setHashFactors(hid.HashFactors);
						hMap.setWindow(MAX_PARTITION_WINDOW_SIZE);
						hMap.setMaxNumTuples(MAX_PARTITION_WINDOW_SIZE);
						
								
						StoredTuples.get(tuplePartitionNumber).add(i,hMap);
					}
					
					TupleList.put(tuplePartitionNumber,new EddyQueue());
				}
				//update partition				
				
			    // add the tuple to the list
			    TupleList.get(tuplePartitionNumber).enqueue(tuple);
			
			// for testing case where stem has only 1 index
			if (bLimitedNumberOfIndices)
			{
				
//				System.out.println("tuple =  "+tuple.toString());
				
				for (int j = 0;j<nNumLimitedIndices;j++)
				{	
					// store the tuple in the appropriate queue
					StoredTuples.get(tuplePartitionNumber).get(j).insertTuple(tuple);
					
				}	
			  }
			}
			else
			{	
			
				// add the tuple to the list
				CompleteTupleList.enqueue(tuple);
			
				// for testing case where stem has only 1 index
				if (bLimitedNumberOfIndices)
				{
				
//				System.out.println("tuple =  "+tuple.toString());
				
					for (int j = 0;j<nNumLimitedIndices;j++)
					{						
						// store the tuple in the appropriate queue
						CompleteStoredTuples.get(j).insertTuple(tuple);
					}	
				}
			}
		}
		
		else //IMP
		{
//			//for testing only
//			System.out.println("**** Adding *****");
//			System.out.println("STREAM"+intSTREAMId);
//			System.out.println("tuple="+tuple.toString());
//			System.out.println("*********");
						
			if (bPartition)
			{	
				//update partition
	    		int tuplePartitionNumber = Integer.parseInt(tuple.getCell(tuple.getLength()-1).convertToString());
				
				if (tuplePartitionNumber > currentPartitionNumber)
				{					
					if((NextIMPIndex == null) && !PartitionList.isEmpty())
					{	
						PriorityHashTreeIndex LastIMPIndex =((Partition)PartitionList.getLast()).getIMPIndex();
//						IMPIndex = LastIMPIndex;
						IMPIndex  = new PriorityHashTreeIndex();
						IMPIndex.setGroupAttrPosArray(LastIMPIndex.getGroupAttrPosArray().clone());
						IMPIndex.setIndexAttrPosArray(LastIMPIndex.getIndexAttrPosArray().clone());
						IMPIndex.setNumIndexEntries(LastIMPIndex.getNumIndexEntries().clone());
						
						
						HashFunction[] newHashFuncs = new HashFunction[LastIMPIndex.getNumIndexEntries().length];
						for (int j=0;j<LastIMPIndex.getNumIndexEntries().length;j++)
						{
							newHashFuncs[j] = new HashFunction(LastIMPIndex.getNumIndexEntries()[j]);			
						}
						IMPIndex.setHashFuncs(newHashFuncs);
					
						IMPIndex.setWindow(MAX_PARTITION_WINDOW_SIZE);
						IMPIndex.setMaxNumTuples(MAX_PARTITION_WINDOW_SIZE);
					}	
					else
					{	
						IMPIndex  = new PriorityHashTreeIndex();
						IMPIndex.setGroupAttrPosArray(NextIMPIndex.getGroupAttrPosArray());
						IMPIndex.setIndexAttrPosArray(NextIMPIndex.getIndexAttrPosArray());
						IMPIndex.setNumIndexEntries(NextIMPIndex.getNumIndexEntries());
						
						
						HashFunction[] newHashFuncs = new HashFunction[NextIMPIndex.getNumIndexEntries().length];
						for (int j=0;j<NextIMPIndex.getNumIndexEntries().length;j++)
						{
							newHashFuncs[j] = new HashFunction(NextIMPIndex.getNumIndexEntries()[j]);			
						}
						IMPIndex.setHashFuncs(newHashFuncs);
					
						IMPIndex.setWindow(MAX_PARTITION_WINDOW_SIZE);
						IMPIndex.setMaxNumTuples(MAX_PARTITION_WINDOW_SIZE);
						NextIMPIndex = null;						
					}
					
					currentPartitionNumber = tuplePartitionNumber; 
					Partition p= new PartitionImp(currentPartitionNumber,null,IMPIndex);		
					PartitionList.enqueue(p);	
				}
			}
			
			IMPIndex.insertTuple(tuple);
			
		}			
	}
	
//////////////////////////////////////////////
// not used for IMP Index
	private void PurgeTuple()
	{
	}

//////////////////////////////////////////////
		public void PurgeTuple(int nEPartitionNum)
		{
//		  System.out.println("nEPartitionNum = "+nEPartitionNum);
//		  System.out.println("earliestPartitionNumber = "+earliestPartitionNumber);
//		  System.out.println("currentPartitionNumber = "+currentPartitionNumber);

		  if (bImpIndex)
		  {
			if (intCurrNumberofStoredTuples == 0)
			    return;
			  
			if(earliestPartitionNumber==currentPartitionNumber)
				  return;
			
			if(earliestPartitionNumber<(nEPartitionNum-MAX_PARTITION_WINDOW_SIZE))
			{	
				int EPN = 0;
				Iterator<Partition> pIter = PartitionList.iterator();
				Partition p = (Partition)pIter.next();
				while (pIter.hasNext()&& p.getPartitionNumber()<(nEPartitionNum-MAX_PARTITION_WINDOW_SIZE)) 	
				{
//					System.out.println("Purging partition = "+ p.getPartitionNumber());
					
					if(p!=null)
					{	
						EPN =  p.getPartitionNumber();
						earliestPartitionNumber = EPN+1;
						
						PriorityHashTreeIndex Index = p.getIMPIndex();
						intCurrNumberofStoredTuples -= Index.clearState2();						
					}
					PartitionList.dequeue();
					
					pIter = PartitionList.iterator();
					p = (Partition)pIter.next();
					
				}
		      }
	      }
		  else //not IMP index
		  {
//			    if(true)
//			      return;
				if (bPartition)
				{	
			  
					if(earliestPartitionNumber==currentPartitionNumber)
						  return;
				    
				    if (intCurrNumberofStoredTuples == 0)
				    	return;
				    
					if(earliestPartitionNumber < (nEPartitionNum-MAX_PARTITION_WINDOW_SIZE))
					{	
	
							for (int EPN =earliestPartitionNumber;  EPN < (nEPartitionNum-MAX_PARTITION_WINDOW_SIZE);  EPN++) 	
							{
//								System.out.println("purging PN ="+EPN);
								if (PartitionList.getFirst().getPartitionNumber() == EPN)
								{	
								  PartitionList.dequeue();
								  earliestPartitionNumber++;
	
								  if(TupleList.get(EPN) != null)
								  {
									//remove all stored tuples
									while(TupleList.get(EPN).length()>0)
									{	
									    TupleList.get(EPN).dequeue();
										//reduce the num of stored tups
										intCurrNumberofStoredTuples--;
									}
								   }	
							
								   TupleList.remove(EPN);
								
								   //remove all index link
								   if(StoredTuples.get(EPN) != null)
								   {	   
									   for (int j =0;j<nNumLimitedIndices;j++)
									   {
										   intCurrNumberofStoredTuples -= StoredTuples.get(EPN).get(0).clearState2();
										   StoredTuples.get(EPN).remove(0);
										    
									   }
									
									   StoredTuples.get(EPN).clear();
									   StoredTuples.remove(EPN);
								   }	
									
							}
									   
						}
	//					System.out.println("End of purging....");
					}	
					}
					else
					{	
						if(earliestPartitionNumber==currentPartitionNumber)
							  return;
					    
					    if (intCurrNumberofStoredTuples == 0)
					    	return;
					    
						if(earliestPartitionNumber<(nEPartitionNum-MAX_PARTITION_WINDOW_SIZE+1))
						{	
							int EPN = 0;
								for (EPN =earliestPartitionNumber;  EPN < (nEPartitionNum-MAX_PARTITION_WINDOW_SIZE+1);  EPN++) 	
								{
										//remove all stored tuples
									    XATTuple t = (XATTuple)(CompleteTupleList.getFirstNode().item);
									    long PN = t.getCell(t.getLength()-1).convertToLong();
										while(PN==EPN)
										{												
										    XATTuple et = (XATTuple)CompleteTupleList.dequeue();
											for (int j =0;j<nNumLimitedIndices;j++)
											{
												// store the tuple in the appropriate queue
												intCurrNumberofStoredTuples -= CompleteStoredTuples.get(j).clearState2();
											}
											
										    t = (XATTuple)(CompleteTupleList.getFirstNode().item);
										    PN = t.getCell(t.getLength()-1).convertToLong();
										    
										}
								 }									
								
								earliestPartitionNumber = EPN;
							}																					   
					}
		//					System.out.println("End of purging....");
						
			}
				  
	}	
		
//////////////////////////////////////////////
		public void MigrateTuples()
		{
			CompleteStoredTuples = new ArrayList<MultiHashKeyIndex>();

			CurrentHashKeys = NextHashKeys;  
			NextHashKeys = null;
			
			Iterator<EddyQueue.Node> it = CompleteTupleList.iterator();
			while(it.hasNext())	
			{
				//remove all stored tuples
				XATTuple t = (XATTuple)(it.next().item);
				Insert(t);				
			}
		//	System.out.println("End of migrating....");
										  
	}	
	
////////////////////////////////////////////////
	
	// this function returns the STREAM is for this STeM Data
	public int getSTREAMId()
	{
		return intSTREAMId;
	}

//////////////////////////////////////////////////	
//	// this function probes the STEM data and returns results 
	public ArrayList <XATTuple> ProbeMulti(ArrayList<XATTuple> tups, int NumSTeMIndex, 
			ArrayList<ArrayList<Integer>> NumTupleIndexs, int PartitionNum, 
			ArrayList <Integer> PosIndices, long SearchTupMinTimeStamp)
	{
		//System.out.println("Stream id "+intSTREAMId+" processing" +tups.size()+" tupples");

//////////
// for testing purposes only
//	if (true)
//	{	
//		Integer QueryPatternId = 0;
//		for (int j=nNumCols-1; j>=0; j--) 
//		{
//			boolean bFound=false;
//			
//			for (int t = 0; t < NumTupleIndexs.size()&&!bFound; t++)
//				for (int y = 0; y < NumTupleIndexs.get(t).size()&&!bFound; y++)
//					if (NumTupleIndexs.get(t).get(y) == j)
//						bFound=true;
//						
//			if(bFound)
//				  QueryPatternId = QueryPatternId*2 + 1;
//			else
//			      QueryPatternId = QueryPatternId*2;
//			
//		}
//		
//		FreqQueryPatterns[QueryPatternId]++;
//		if (FreqQueryPatterns[QueryPatternId] % 100 == 0)
//		{
//			WriteToFile("C:\\users\\kworks\\QueryPatternFreq-"+intSTREAMId+".txt", String.valueOf(QueryPatternId), false);
//			WriteToFile("C:\\users\\kworks\\QueryPatternFreq-"+intSTREAMId+".txt", String.valueOf(FreqQueryPatterns[QueryPatternId]), true);
//			FreqQueryPatterns[QueryPatternId] = 0;
//			
//		}
//	}	
//////////
		
		ArrayList <XATTuple> results = new ArrayList<XATTuple>();
		
		long tupleMaxCounter =0;
		long tupleMinCounter = Long.MAX_VALUE;
		for(int h = 0;h<tups.size();h++)
		{	
		    long tupleCounter = Long.parseLong(tups.get(h).getCell(tups.get(h).getLength() - 2).convertToString());
		    if (tupleCounter > tupleMaxCounter)
		    	tupleMaxCounter = tupleCounter;
		    if (tupleCounter < tupleMinCounter)
		    	tupleMinCounter =tupleCounter;		    	
//				System.out.println("t = " + h);
//				System.out.println(tups.get(h).toString());
		}  	
		
//for synthetic data
//		tupleMaxCounter += 17;		
//		tupleMinCounter -= 17;
		
//for real data
//		tupleMaxCounter += 32;		
//		tupleMinCounter -= 32;
		tupleMaxCounter += 75;		
		tupleMinCounter -= 75;
		if(tupleMinCounter<=0)
			tupleMinCounter = 1;
		
//		System.out.println("tupleMaxCounter = " + tupleMaxCounter);                
//		System.out.println("tupleMinCounter = " + tupleMinCounter);

		XATPunct newPunct = new XATPunctImp(EmptyPunct.getArrayData().clone());
		long[] inputTimeStamp = tups.get(0).getTimeStamp();
		long[] newTimeStamp = EmptyTimeStamp.clone();
		int[] newPunctTypeArray = EmptyPunctTypeArray.clone();
		Integer QueryPatternId = 0;
				
        //sampling
//    	boolean bCollectStats = false;
//    	if(bExhaustiveAssessment)
//    	{	
////        	if (bNoSampling)
//        		bCollectStats = true;
////        	else if(bSimpleRandomSampling)
////        	{
////        		int nRan=generator.nextInt(100)+1;
////        		if (nRan<=nSimpleRandomPercent)
////            		bCollectStats = true;
////        	}
////        	else if(bSystematicSampling)
////        		if ((nTotalNumQueryRequests%nSystematicInterval) ==0)
////            		bCollectStats = true;
////        	else if(bStratifiedRandomSampling)
////        	{
////        		if ((nTotalNumQueryRequests%nStratifiedRandomWindow) ==0)
////        		{
////        			nNumStratifiedRandomSampled = 0;
////        			nTotalStratifiedRandomToSample = nStratifiedRandomWindow*(nStratifiedRandomPercent/100);
////        		}
////        		
////        		int nRan=generator.nextInt(nStratifiedRandomWindow - (nTotalNumQueryRequests - ((nTotalNumQueryRequests%nStratifiedRandomWindow)*nStratifiedRandomWindow)) +1);
////        		if (nRan<=nTotalStratifiedRandomToSample)
////        		{	
////        			bCollectStats = true;
////        			nTotalStratifiedRandomToSample++;
////        		}	
////        	}
//    	}
		QueryPatternId = 0;
		Iterator<ArrayList<Integer>> it = NumTupleIndexs.iterator();
		while (it.hasNext()) 
		{
			ArrayList<Integer> al = it.next();
//				System.out.println(al.toString());
			for (int h = 0;h<al.size();h++) 
			{
				int j  = al.get(h);
			        
				newPunct.setCell(j, tups.get(h).getCell(j));
				newTimeStamp[j] = inputTimeStamp[j];
				newPunctTypeArray[j] = XATPunct.SINGLE_VALUE;
					
				if(bImpIndex && bExhaustiveAssessment)
				{	
					QueryPatternId = QueryPatternId + ((Double)Math.pow(2,j)).intValue();
				}	
			}
		}
	  	newPunct.setTimeStamp(newTimeStamp);
		newPunct.setMinTimeStamp(tups.get(0).getMinTimeStamp());
		newPunct.setMaxTimeStamp(tups.get(0).getMaxTimeStamp());							
		newPunct.setPunctTypeArray(newPunctTypeArray);
		
		
		if (!bImpIndex)
		{			

			if (intCurrNumberofStoredTuples == 0)
		    	return null;
						
			
			if(bExhaustiveAssessment)
				if (AccessModuleIndexAssessment.get(NumSTeMIndex)!= null)
					AccessModuleIndexAssessment.set(NumSTeMIndex,AccessModuleIndexAssessment.get(NumSTeMIndex)+1);
				else
					AccessModuleIndexAssessment.add(NumSTeMIndex,1);					
			
			
			if (bPartition)
			{	
				int p = PartitionNum-MAX_PARTITION_WINDOW_SIZE+1;
			
				if (p < 0 )
					p = 1;
			
				boolean bCont = true;
				for (;bCont && p<=PartitionNum;p++) 
				{	
				
					if (PartitionList.getPartition(p) == null)
						continue;
				
					if (!PartitionList.getPartition(p).getHashKeys().contains(NumSTeMIndex))
					{
						//No hash index exists no need to do linear scan
						ArrayList<Integer> stIndex = HashKeyIndex.get(NumSTeMIndex);
				
						
						//is there another index that can help?
						int indx = 0;
						int nNumTupsInIndex = 0;
						boolean bC = true;
					
						//					System.out.println("stIndex="+stIndex.toString());
						//					System.out.println("PosIndices="+PosIndices.toString());
						for (int i = PosIndices.size(); bC && i>0 ;i--)
							if(PartitionList.getPartition(p).getHashKeys().contains(PosIndices.get(i-1)))
							{	
								bC = false;
								indx = PosIndices.get(i-1);
								//assumes one tup per field
								nNumTupsInIndex = HashKeyIndex.get(indx).size();
							}         		
        		
						if (!bC)
						{	
							if(StoredTuples.get(p) != null)
								// find matches to probe tuple key
								bCont = StoredTuples.get(p).get(indx).probe2(newPunct, tupleMaxCounter, tupleMinCounter, results);
						}		
						else
		    			// search entire list of stored tuples
								bCont = StoredTuples.get(p).get(indx).probe2(newPunct, tupleMaxCounter, tupleMinCounter, results);
			}		
			else //use index
			{				
    			int indx  = PartitionList.getPartition(p).getHashKeys().indexOf(NumSTeMIndex);    			
            	bCont = StoredTuples.get(p).get(indx).probe2(newPunct, tupleMaxCounter, tupleMinCounter, results);
			}	
        }
	}    
	else //no partitions
	{	
		
		if (!CurrentHashKeys.contains(NumSTeMIndex))
		{
			//No hash index exists no need to do linear scan
			ArrayList<Integer> stIndex = HashKeyIndex.get(NumSTeMIndex);
				
			//is there another index that can help?
			int indx = 0;
			int nNumTupsInIndex = 0;
			boolean bC = true;
			for (int i =PosIndices.size(); bC && i>0 ;i--)
			{	
				if((nNumLimitedIndices-1) >= PosIndices.get(i-1))
				{	
					bC = false;
					indx = PosIndices.get(i-1);
					nNumTupsInIndex = i;
				}         		
			}
		
			//where to search
			if (!bC)
			{
	            if(CompleteStoredTuples != null)
	    			// find matches to probe tuple key
	            	CompleteStoredTuples.get(indx).probe2(newPunct, tupleMaxCounter, tupleMinCounter, results);
            }		
			else
			{	
    			// search entire list of stored tuples
            	CompleteStoredTuples.get(indx).probe2(newPunct, tupleMaxCounter, tupleMinCounter, results);
			}
		}		
		else //use index
		{	
			int indx  = CurrentHashKeys.indexOf(NumSTeMIndex);
        	CompleteStoredTuples.get(indx).probe2(newPunct, tupleMaxCounter, tupleMinCounter, results);
		}
	}	
}
else	//IMP Index
{
			if (intCurrNumberofStoredTuples == 0)
		    	return null;

			PartitionListQueue.Node prtLst = null;
			int p = PartitionNum-MAX_PARTITION_WINDOW_SIZE+1;
			if (p <= 0)
				p = 1;
			
//			System.out.println("search tuplePartitionNumber="+tuplePartitionNumber);
			
			
			if(p!=earliestPartitionNumber)
              prtLst = PartitionList.getPartitionNode(p);
			else
			  prtLst = PartitionList.getFirstNode();
			
			if(prtLst == null)
	          return null;	  

			if(bExhaustiveAssessment)
			{    			
//				System.out.println("QueryPatternId = "+Integer.toBinaryString(QueryPatternId));
//				System.out.println();
    			
    			if (bSRIA || bHHH)
    			{	
    				if(SRIA.get(QueryPatternId) != null)
    					SRIA.get(QueryPatternId).nCountQueryPattern++;
    				else
    				{
    					IndexAssessment IA = new IndexAssessment(QueryPatternId,1,nBucketId-1);

//    					System.out.println("Stream id "+intSTREAMId);
//    					System.out.println("add QueryPatternId = "+Integer.toBinaryString(QueryPatternId));
    					        				
        				SRIA.put(QueryPatternId, IA);
        				        				
        				//add to list of leaf nodes
        				if(bHHH)
        				{
        				    int nPos = 0;
        					byte bIA = new Integer(IA.nQueryPattern).byteValue(); 
        				    while(nPos<LatticeLeafNodes.size())
        				     {
        				    	IndexAssessment IAn = (IndexAssessment)LatticeLeafNodes.get(nPos); 
            					byte bIAn = new Integer(IAn.nQueryPattern).byteValue();
            					
        				    	if((bIA & bIAn) == bIAn)
        				    	{
        				    		LatticeLeafNodes.remove(IAn);
        				    		IAn = null;
        				    	}
        				    	else
        				    		nPos++;
        				    } 
        				    
        				    //add the new index assessment to the list of leaf nodes
        					LatticeLeafNodes.add(IA);
        				}
    				}
    			}	
    			else
    			{
//					System.out.println("Stream id "+intSTREAMId);
//					System.out.println("add QueryPatternId = "+Integer.toBinaryString(QueryPatternId));
    				
    				FreqQueryPatterns[QueryPatternId]++;
    			}
    			
				nTotalNumQueryRequests++;
				
    			if (bHH || bHHH) 
    			{	
					if ((nTotalNumQueryRequests % (1/dHHErrorRate)) == 0)
					{
////						System.out.println("Start Prunning....");
        				if(bHHH)
        					PruneHHH();
        				else	
        					Prune();
					    nBucketId++;	
////						System.out.println("End Prunning....");
					}
				}
				
			}
			
///////////////////////
//KEW
			//partition number of probe request
    		//int tuplePartitionNumber = PartitionNum;
		
			
			//System.out.println("search tuplePartitionNumber="+tuplePartitionNumber);
			
			
			
//    	    while ((p<=tuplePartitionNumber) && (prtLst != null))
    		boolean bCont = true;	
    		for (;bCont && p<=PartitionNum && (prtLst != null) ;p++) 
            {
    	    	
//    			System.out.println("searching Partition  Number="+p);
    			bCont = prtLst.item.getIMPIndex().probe2(newPunct,tupleMaxCounter,tupleMinCounter,results);
//    	    	if (results.size()>0)
//    	    	{	
//    	    		System.out.println("searching Partition  Number="+p);	
//      	    	    System.out.println(results.size());
//    	    	}
                //nxt
                prtLst = prtLst.next;
//                if(prtLst != null)
//                  p  = prtLst.item.getPartitionNumber();
              
  		}
    	    
	}
		if (results !=  null)
		{		
			if (results.size() > 0)
			{	
				//System.out.println("returning "+results.size()+" results");
				return results;
			}	
			else
				return null;
		}	
		else
			return null;
	}
	/////////////////////////////////////////////

	
////////////////////////////////////////////
	
	public void SetMaxWindowSize(int val)
	{
		MAX_WINDOW_SIZE = val;
	}
//////////////////////////////////////////
	public void SetIMPIndex(int[] GroupPos, int[] IndexAttrbPos, int[] NumEntries)
	{
		IMPIndex  = new PriorityHashTreeIndex();
		IMPIndex.setGroupAttrPosArray(GroupPos);
		IMPIndex.setIndexAttrPosArray(IndexAttrbPos);
		IMPIndex.setNumIndexEntries(NumEntries);
		
		
		HashFunction[] newHashFuncs = new HashFunction[NumEntries.length];
		for (int j=0;j<NumEntries.length;j++)
		{
			newHashFuncs[j] = new HashFunction(NumEntries[j]);			
		}
		IMPIndex.setHashFuncs(newHashFuncs);
	
		IMPIndex.setWindow(MAX_PARTITION_WINDOW_SIZE);
		IMPIndex.setMaxNumTuples(MAX_PARTITION_WINDOW_SIZE);

		if (bPartition)
		{
			//Partition List
			if(PartitionList==null)
			  PartitionList = new PartitionListQueue();
			
//			//add first partition
			currentPartitionNumber=1;
			Partition p= new PartitionImp(currentPartitionNumber,null,IMPIndex);		
			PartitionList.enqueue(p);
			
		}	

	}
	///////////
	public void SetHashIndex(int[] GroupPos, int[] IndexAttrbPos, int[] NumEntries)
	{
		MultiHashKeyIndex currentMap = new MultiHashKeyIndex();
		currentMap.setGroupAttrPosArray(GroupPos);
		currentMap.setIndexAttrPosArray(IndexAttrbPos);
		currentMap.setHashFactors(NumEntries);
		currentMap.setWindow(MAX_PARTITION_WINDOW_SIZE);
		currentMap.setMaxNumTuples(MAX_PARTITION_WINDOW_SIZE);
		
		if (bPartition)
		  StoredTuples.get(1).add(currentMap);
		else
		  CompleteStoredTuples.add(currentMap);
			
		HashIndexData hid = new HashIndexData(GroupPos, IndexAttrbPos, NumEntries);		
		CurrentHashIndexData.add(hid);
	}

	///////////////////////////////////////////////////////////////////	
	public void UpdateIndex(int[] GroupPos, int[] IndexAttrbPos, int[] NumEntries,ExhaustiveEnum qbph,float cQueryCost,float cCreationCost)
	{


		//need to check to see if the index is different from current index or not
		//testing
//		System.out.println("IMPIndex");
		boolean bMatch = true;
		for (int j=0;j<IMPIndex.getNumIndexEntries().length&&bMatch;j++)
		{
//			System.out.print(IMPIndex.getNumIndexEntries()[j]+",");
			if(IMPIndex.getNumIndexEntries()[j]!=NumEntries[j])
				bMatch = false;
		}
//		System.out.println();	
		if(bMatch)
		{	
			for (int j=0;j<IMPIndex.getIndexAttrPosArray().length&&bMatch;j++)
			{
//				System.out.print(IMPIndex.getIndexAttrPosArray()[j]+",");
				
				if(IMPIndex.getIndexAttrPosArray()[j]!=IndexAttrbPos[j])
					bMatch = false;
			}
//			System.out.println();		
		}
		
		if(bMatch)
			return;
		
//		System.out.println("Updating Index");
		
		NextIMPIndex  = new PriorityHashTreeIndex();
		NextIMPIndex.setGroupAttrPosArray(GroupPos);
		NextIMPIndex.setIndexAttrPosArray(IndexAttrbPos);
		NextIMPIndex.setNumIndexEntries(NumEntries);
		
		
		HashFunction[] newHashFuncs = new HashFunction[NumEntries.length];
		for (int j=0;j<NumEntries.length;j++)
		{
			newHashFuncs[j] = new HashFunction(NumEntries[j]);			
		}
		NextIMPIndex.setHashFuncs(newHashFuncs);
	
		NextIMPIndex.setWindow(MAX_PARTITION_WINDOW_SIZE);
		NextIMPIndex.setMaxNumTuples(MAX_PARTITION_WINDOW_SIZE);

		if (bPartition)
			//Index Migration
			if(!bNoMigration && (qbph != null) && (PartitionList.length()>0))
			{				
				if (bSymmetricMigration)
					SymmetricMigration(NextIMPIndex,qbph, cQueryCost+cCreationCost);
				else if (bPartitionMigration)
					PartitionMigration(NextIMPIndex,qbph,cQueryCost,cCreationCost);					
				else if (bCompleteMigration)
					CompleteMigration(NextIMPIndex);					
			}


	}
/////////////////////////////////////////////////////
private void SymmetricMigration(PriorityHashTreeIndex IMPIndx,ExhaustiveEnum qbph, float cCost)
{
//	System.out.println("Symmetric Migration");
//	System.out.println("STREAM"+intSTREAMId);
	
	//migrated inps
	HashMap <PriorityHashTreeIndex,Boolean> MigratedIMP = new HashMap <PriorityHashTreeIndex,Boolean>();
	//loop through each partition
	for ( Iterator<Partition> pIter = PartitionList.iterator(); pIter.hasNext();) 	
	{
			Partition p = (Partition)pIter.next();
			PriorityHashTreeIndex indx = p.getIMPIndex();
		
//			System.out.println("evaluating partition="+p.getPartitionNumber());
			
			boolean bEvalMigration = true;
			if(MigratedIMP.get(indx) != null) 
				bEvalMigration = false;
			
			if(bEvalMigration) 	
			{	
				//need to compute just the cost to probe
				int[] numBits = new int[nNumCols];
				int[] AttrPosArray  = indx.getIndexAttrPosArray();
				int[] Bits  = indx.getNumIndexEntries();
				int apPos = 0;
				
				for (int i=0; (i<nNumCols) && (apPos<AttrPosArray.length); i++)
				{
					if(AttrPosArray[apPos] == i)
					{	
						numBits[i] =  Bits[apPos];
						apPos++;
					}	
					else
						numBits[i] =  0;
						
				}				
//KEW
//1) Need to validate costs
				IndexConfig ic = new IndexConfig(numBits);			
				float pCost = qbph.computeQueryCost(ic, nTotalNumQueryRequests, nTotalNumBuildRequests);

								
				//////////////////////////////////
				if(cCost < pCost) //need to migrate
				{
//				  System.out.println("Migrating partition="+p.getPartitionNumber());
				  MigratedIMP.put(indx, true);
				  MigrateAllIndex(p.getPartitionNumber());
				  
//				  //count migrations
//				  if (bCountMigrations)
//				  {	
//					  WriteToFile(strCountMigration, indx.getIndexAttrPosArray().toString(), false);
//					  WriteToFile(strCountMigration, NextIMPIndex.getIndexAttrPosArray().toString(), true);
//				  }

				}
				else
				{	
//				  System.out.println("NO Need to migrate partition="+p.getPartitionNumber());   	
				  MigratedIMP.put(indx, false);
				}	
				//////////////////////////////
		    }
			else if(MigratedIMP.get(indx) != null) 
				if (MigratedIMP.get(indx)) // IMP already migrated
				{
					p.setIMPIndex(NextIMPIndex);					
				}
//				else
//					  System.out.println("NO Need to migrate partition="+p.getPartitionNumber());
			
					
	}	

	MigratedIMP.clear();
}
/////////////////////////////////////////////////////
private float CalculateEstimatesUsage(Integer PartitionNumber)
{
	return	(float)((1.0 * (PartitionNumber-earliestPartitionNumber+1.0)) / (1.0*(currentPartitionNumber - earliestPartitionNumber+1)));
}

/////////////////////////////////////////////////////
private void PartitionMigration(PriorityHashTreeIndex IMPIndx,ExhaustiveEnum qbph,float cQueryCost,float cCreationCost)
{
//	System.out.println("Partition Migration");
//	System.out.println("STREAM "+intSTREAMId);
	
	//migrated inps
	HashMap <PriorityHashTreeIndex,Boolean> MigratedIMP = new HashMap <PriorityHashTreeIndex,Boolean>();
	HashMap <int[],Float> IMPCost = new HashMap <int[],Float>();
	
	for ( Iterator<Partition> pIter = PartitionList.iterator(); pIter.hasNext();) 	
	{
			Partition p = (Partition)pIter.next();
			PriorityHashTreeIndex indx = p.getIMPIndex();
						
//			System.out.println("evaluating partition="+p.getPartitionNumber());
//			System.out.println("indx="+indx.toString());
			
			boolean bEvalMigration = true;
			if(MigratedIMP.isEmpty()) 
				bEvalMigration = true;
			else if(MigratedIMP.get(indx) == null) 
				bEvalMigration = false;
			
			if(bEvalMigration) 
			{	
				//need to figure out per window
				//int numTupsPerWindow = MAX_PARTITION_WINDOW_SIZE * MAX_PARTITION_SIZE;
				
				float estUse = CalculateEstimatesUsage(p.getPartitionNumber());
		
				//need to compute just the cost to probe
				int[] numBits = new int[nNumCols];
				int[] AttrPosArray  = indx.getIndexAttrPosArray();
				int[] Bits  = indx.getNumIndexEntries();
				
				int apPos = 0;
				
				for (int i=0; (i<nNumCols)&&(apPos<AttrPosArray.length); i++)
				{
					if(AttrPosArray[apPos] == i)
					{	
						numBits[i] =  Bits[apPos];
						apPos++;
					}	
					else
						numBits[i] =  0;						
				}
				
				float pCost;
				if(IMPCost.get(numBits)==null)
				{	
				  IndexConfig ic = new IndexConfig(numBits);
//				qbph.computeNumConfigs();
				  pCost = qbph.computeQueryCost(ic, nTotalNumQueryRequests, nTotalNumBuildRequests);
				  IMPCost.put(numBits,new Float(pCost));
				  
				  pCost *= estUse;
				}
				else
					pCost =estUse*IMPCost.get(numBits);
				
//				System.out.println("pCost="+pCost);

				float cCost = estUse*cQueryCost + cCreationCost;
				
//				System.out.println("cCost="+cCost);
				
				//////////////////////////////////
				if(cCost < pCost) //need to migrate
				{
				  MigratedIMP.put(indx,true);				  
				  MigrateSinglePartitionIndex(p.getPartitionNumber());
				  
//					//count migrations
//					if (bCountMigrations)
//					{	
//						WriteToFile(strCountMigration, indx.getIndexAttrPosArray().toString(), false);
//						WriteToFile(strCountMigration, NextIMPIndex.getIndexAttrPosArray().toString(), true);
//					}
				}
//				else
//				{	
//				  System.out.println("NO Need to migrate partition="+p.getPartitionNumber());   	
//				}	
				//////////////////////////////
		    }
			else // IMP already migrated 
			{
//					System.out.println("Migrate partition="+p.getPartitionNumber());
					
//				//count migrations
//				if (bCountMigrations)
//				{	
//					WriteToFile(strCountMigration, indx.getIndexAttrPosArray().toString(), false);
//					WriteToFile(strCountMigration, NextIMPIndex.getIndexAttrPosArray().toString(), true);
//				}
				
				MigrateSinglePartitionIndex(p.getPartitionNumber());			  
			}
					
	}	
	
	MigratedIMP.clear();
}
/////////////////////////////////////////////////////
private void CompleteMigration(PriorityHashTreeIndex IMPIndx)
{
//	System.out.println("Complete Migration");
//	System.out.println("STREAM"+intSTREAMId);
	
	//loop through each partition
	for ( Iterator<Partition> pIter = PartitionList.iterator(); pIter.hasNext();) 	
	{
			Partition p = (Partition)pIter.next();
			PriorityHashTreeIndex indx = p.getIMPIndex();

			
			MigrateAllIndex(p.getPartitionNumber());
			
	}	

}
//////////////////////////////////////////////////////
private void MigrateAllIndex(int nPN)
{
	Partition p = PartitionList.getPartition(nPN);
		
	if(!bImpIndex)
	{	
//		//Old
//		ArrayList<Integer> OldHashKeyIndex = null;
//		OldHashKeyIndex = p.getHashKeys();
//		//Current
//		ArrayList<ArrayList<Integer>> CurrentHashKeyIndex = null;
//		CurrentHashKeyIndex =((Partition)PartitionList.getLast()).getHashKeys();
//
//		// find all tuples in partiton and delete them
//		EddyQueue.Node N = (EddyQueue.Node)p.getFirstTupleinPartition();
//		XATTuple Tup = (XATTuple)N.item;
//        Object strTupleCounter1 = Tup.getCell(Tup.getLength() - 1);
//        int TupPartitionNumber = Integer.parseInt(strTupleCounter1.toString());
//		
//		while (TupPartitionNumber == nPN)
//		{							
//			//re-index to new index
//			for (int j =0;j<CurrentHashKeyIndex.size();j++)
//			{					
//				String key = "";
//				for (int y = 0;y<CurrentHashKeyIndex.get(j).size(); y++)
//					key += Tup.getCell(CurrentHashKeyIndex.get(j).get(y)).convertToString()+"---";
//					 	
//				StoredTuples.get(j).remove(key);
//			}
//			
//			//delete old index
//			for (int j =0;j<OldHashKeyIndex.size();j++)
//			{					
//				String key = "";
//				for (int y = 0;y<OldHashKeyIndex.get(j).size(); y++)
//					key += Tup.getCell(OldHashKeyIndex.get(j).get(y)).convertToString()+"---";
//					 	
//				StoredTuples.get(j).remove(key);
//			}
//			
//			//need tuples to have prev & next
//			N = N.next;
//			if(N!=null)
//			{	
//				Tup = (EddyTuple)N.item;
//		        strTupleCounter1 = Tup.getCell(Tup.getLength() - 1);
//		        TupPartitionNumber = Integer.parseInt(strTupleCounter1.toString());
//			}
//		}
//		
//		//reset Index in PartitionList to new Index
//		p.setHashKeyIndex(CurrentHashKeyIndex);
	}
	else
	{	
		//Old
		PriorityHashTreeIndex OldIMPIndex = p.getIMPIndex();
		
		//testing
//		System.out.println("TotalNumTuples ="+OldIMPIndex.getTotalNumTuples());
		
		//Current
		PriorityHashTreeIndex CurrentIMPIndex = new PriorityHashTreeIndex();
		CurrentIMPIndex.setGroupAttrPosArray(NextIMPIndex.getGroupAttrPosArray().clone());
		CurrentIMPIndex.setIndexAttrPosArray(NextIMPIndex.getIndexAttrPosArray().clone());
		CurrentIMPIndex.setNumIndexEntries(NextIMPIndex.getNumIndexEntries().clone());
		
		
		HashFunction[] newHashFuncs = new HashFunction[NextIMPIndex.getNumIndexEntries().length];
		for (int j=0;j<NextIMPIndex.getNumIndexEntries().length;j++)
		{
			newHashFuncs[j] = new HashFunction(NextIMPIndex.getNumIndexEntries()[j]);			
		}
		CurrentIMPIndex.setHashFuncs(newHashFuncs);
	
		CurrentIMPIndex.setWindow(MAX_PARTITION_WINDOW_SIZE);
		CurrentIMPIndex.setMaxNumTuples(MAX_PARTITION_WINDOW_SIZE);

		//loop through all stored tuples
        XATPunct newPunct = new XATPunctImp(EmptyPunct.getArrayData().clone());
        int[] newPunctTypeArray = EmptyPunctTypeArray.clone();
        long[] newTimeStamp = EmptyTimeStamp.clone();

//        // Search for particular partition        
//        newPunct.setCell(nNumCols-1,new XATDataValueImp(new Integer(nPN)));
//		newPunctTypeArray[nNumCols-1] = XATPunct.SINGLE_VALUE;
//		newTimeStamp[nNumCols-1] = 0;
		
        newPunct.setPunctTypeArray(newPunctTypeArray);
		
        newPunct.setTimeStamp(newTimeStamp);
        newPunct.setMinTimeStamp(0);
        newPunct.setMaxTimeStamp(0);

        //find all tuples for this partition
        Vector result = OldIMPIndex.probe(newPunct);
        int numResultEntries = result.size();
        if (numResultEntries>0) 
        {
            for (int i=0; i<numResultEntries; i++) 
            {
                 TempHashEntry tempEntry = (TempHashEntry)result.get(i);

                 //System.out.println("migrating "+tempEntry.getTupleList().size()+" tuples");
                 if(tempEntry.getNumTuples()>0)	 
                   while (tempEntry.getTupleList().size()>0) 
                   {
                      XATTuple TupFound = (XATTuple)tempEntry.getTupleList().remove(0);
                      //index tuples using current index
                      if(TupFound!=null)
                        CurrentIMPIndex.insertTuple(TupFound);
                   }
            }
        }    


        OldIMPIndex.deleteTuple(newPunct);
        if (OldIMPIndex.getTotalNumTuples()==0)
        {
        	OldIMPIndex.delete();
        }	
        
//        //KEW Need to fix
//		//delete entire index?
//		OldIMPIndex.clearState();
		
		//reset Index in PartitionList to new Index
		p.setIMPIndex(CurrentIMPIndex);
	}	


}
//////////////////////////////////////////////////////
private void MigrateSinglePartitionIndex(int nPN)
{
	Partition p = PartitionList.getPartition(nPN);
		
	if(!bImpIndex)
	{	
	}
	else
	{	
		//Old
		PriorityHashTreeIndex OldIMPIndex = p.getIMPIndex();
		
		//Current
		PriorityHashTreeIndex CurrentIMPIndex = NextIMPIndex;

		//loop through all stored tuples
        XATPunct newPunct = new XATPunctImp(nNumCols);
        int[] newPunctTypeArray = new int[nNumCols];
        long[] newTimeStamp = new long[nNumCols];
        
        for (int j=0; j<nNumCols-1; j++) {
            newPunct.setCell(j, new XATDataValueImp("*"));
			newPunctTypeArray[j] = XATPunct.ALL_VALUE;
			newTimeStamp[j] = 0;
        }

// Search for particular partition        
        newPunct.setCell(nNumCols-1,new XATDataValueImp(new Integer(nPN)));
		newPunctTypeArray[nNumCols-1] = XATPunct.SINGLE_VALUE;
		newTimeStamp[nNumCols-1] = 0;        
        newPunct.setPunctTypeArray(newPunctTypeArray);
		
        newPunct.setTimeStamp(newTimeStamp);
        newPunct.setMinTimeStamp(0);
        newPunct.setMaxTimeStamp(0);

        //find all tuples for this partition
        Vector result = OldIMPIndex.probe(newPunct);
        int numResultEntries = result.size();
        if (numResultEntries>0) {
            for (int i=0; i<numResultEntries; i++) {
                 TempHashEntry tempEntry = (TempHashEntry)result.get(i);

//                 System.out.println("migrating "+tempEntry.getTupleList().size()+" tuples");
                 
                 if(tempEntry.getNumTuples()>0)	 
                     for (int j=0; j<tempEntry.getNumTuples(); j++) 
                     {
                    	 XATTuple TupFound = (XATTuple)tempEntry.getTupleList().get(j);
                    	 //index tuples using current index
                    	 CurrentIMPIndex.insertTuple(TupFound);
                     }
                 
            }
        }
        
        OldIMPIndex.deleteTuple(newPunct);
        if (OldIMPIndex.getTotalNumTuples()==0)
        {	
        	OldIMPIndex.delete();
        }	
        
		//reset Index in PartitionList to new Index
		p.setIMPIndex(CurrentIMPIndex);
	}	


}
//////////////////////////////////////////////////////
//sets the partition window size
public void setMaxPartitionSize(int val)
{
	MAX_PARTITION_SIZE = val;
}

public int getMaxPartitionSize()
{
	return MAX_PARTITION_SIZE;
}

//sets the partition window size
	public void setMaxPartitionWindowSize(int val)
	{
		MAX_PARTITION_WINDOW_SIZE = val;
	}

	public int getMaxPartitionWindowSize()
	{
		return MAX_PARTITION_WINDOW_SIZE;
	}
	
//////////////////////////////////////////////////////
	//is partitioned?
	public void setIsPartitioned(boolean val)
	{
		bPartition = val;
	}

	public boolean getIsPartitioned()
	{
		return bPartition;
	}

//////////////////////////////////////////////////////
	//use IMPIndex? No - hash index
	public void setUseIMPIndex(boolean val)
	{
		bImpIndex = val;
		bLimitedNumberOfIndices = !bImpIndex;
	}

	public boolean getUseIMPIndex()
	{
		return bImpIndex;
	}
	
//////////////////////////////////////////////////////
	//sets the Number of Limited Indices
	public void setNumLimitedIndices(int val)
	{
		nNumLimitedIndices = val;
	}

	public int getNumLimitedIndices()
	{
		return nNumLimitedIndices;
	}

//////////////////////////////////////////////////////
	//Exhaustive Assessment
	public void startExhaustiveAssessment()
	{
//		System.out.println("starting Exhaustive selection.");
		
		if (bExhaustiveAssessment)
			return;
		
		bExhaustiveAssessment = true;
				
		ClearFreqQueryPatterns();
		
		nTotalNumQueryRequests = 0;
		nTotalNumBuildRequests = 0;
	}


	public void endExhaustiveAssessment(int nNumBits)
	{
//		System.out.println("ending Exhaustive selection.");
		
		bExhaustiveAssessment = false;
		if (!bImpIndex)
		{
//			System.out.println("Running Exhaustive selection.");
			//NextHashKeys
			for (int i = 0; i<nNumLimitedIndices;i++)
			{
				Object obj = Collections.max(AccessModuleIndexAssessment);
				
				int index = AccessModuleIndexAssessment.indexOf(obj);
				NextHashKeys.add(index);
				AccessModuleIndexAssessment.set(index, 0);
				
				HashIndexData hid = CurrentHashIndexData.get(i);
				Iterator<Integer> it = HashKeyIndex.get(index).iterator();
				
				int p =0;
				int nIndxs = HashKeyIndex.get(index).size();
				int[] hf = new int[nIndxs];
				while(it.hasNext())
				{
					int indx = it.next();
					//index
					hid.IndexAttrPosArray[p] = indx;
				
					//num bits
					hf[p] =  nNumBits/nIndxs; 
					
				}
				hid.HashFactors = hf;
			}
			//	
//			System.out.println("End Exhaustive selection.");
			
		}	
		else
		{	
		
			QueryPattern[] qp = getFreqQueryPatterns(nNumBits);
			
			if (qp != null)
			{	
//				System.out.println("Running Exhaustive selection.");
				
				
				ExhaustiveEnum qbph = new ExhaustiveEnum();
				qbph.init((float)dComparisoncost, (float)dHashCost, (float)(1.0*nTotalNumBuildRequests), (float)(1.0*nTotalNumQueryRequests), (float)(MAX_PARTITION_WINDOW_SIZE),nNumBits,MAX_PARTITION_SIZE);
				qbph.init(nNumCols, nNumBits);
				qbph.computeNumConfigs();
				qbph.setQueryPatterns(qp);
				
	//			OWQCLeadingHeuristic qbph = new OWQCLeadingHeuristic();
	//			qbph.init((float)0.000056, (float)0.001392, (float)(1.0*nTotalNumBuildRequests), (float)(1.0*nTotalNumQueryRequests), (float)(1.0*nTotalNumQueryRequests),nNumBits);
	//			qbph.init(nNumCols,nNumBits);
	//			qbph.computeNumConfigs();
	//			qbph.setQueryPatterns(qp);
				
				
				boolean bRunSelection = bNoIndexSelectionTrigger;
				//////////////////////////////////////////////////////
				if(!bRunSelection)
				{	
					if(!bFirstAssessment &&( bCostIncrIndexSelectionTrigger || bCombinedIncrIndexSelectionTrigger))
					{	
						//need to compute just the cost to probe using current
						//IMP index
						
						int[] numBits = new int[nNumCols];
						int[] AttrPosArray  = IMPIndex.getIndexAttrPosArray();
						int[] Bits  = IMPIndex.getNumIndexEntries();
						int apPos = 0;
									
						for (int i=0; (i<nNumCols)&&(apPos<AttrPosArray.length); i++)
						{
							if(AttrPosArray[apPos] == i)
							{	
								numBits[i] =  Bits[apPos];
								apPos++;
							}	
							else
								numBits[i] =  0;						
						}
						IndexConfig ic = new IndexConfig(numBits);		
						float pCost = qbph.computeQueryCost(ic, nTotalNumQueryRequests, nTotalNumBuildRequests);
						if(Math.abs(1.0-(pCost/pIMPIndexCost)) <= dDegreeCostIncr)
							bRunSelection = true;
						else
							bRunSelection = false;
					}
				
					//////////////////////////////////////////////////////
					if(!bFirstAssessment &&( bFreqIncrIndexSelectionTrigger || (bCombinedIncrIndexSelectionTrigger && !bRunSelection)))				
					{	
						//freq diff
						double dDegFreqDiff = 0.0;
						for(int i=0;i<qp.length;i++)
						{
							if(qp[i].getWeight()>0)
							{	
								if (IMPIndexqp[i].getWeight()>0)
									dDegFreqDiff += Math.abs(1.0-(qp[i].getWeight()/IMPIndexqp[i].getWeight()));
								else
									dDegFreqDiff += 1.0;
							}	
						}
						if(dDegFreqDiff <= dDegreeFreqIncr)
							bRunSelection = true;
						else
							bRunSelection = false;
						
					}
				}
				
				//////////////////////////////////////////////////////
				if(bRunSelection || bFirstAssessment)
				{	
					bFirstAssessment = false;				
	//				System.out.println("Find optimal cost");
	
					qbph = new ExhaustiveEnum();
					qbph.init((float)dComparisoncost, (float)dHashCost, (float)(1.0*nTotalNumBuildRequests), (float)(1.0*nTotalNumQueryRequests),(float)1.0,nNumBits,MAX_PARTITION_SIZE);
					qbph.init(nNumCols, nNumBits);
					qbph.setQueryPatterns(qp);
					
					qbph.setVals(nTotalNumQueryRequests, nTotalNumBuildRequests,1);
					qbph.computeNumConfigs();
					//Find optimal cost.					
					//qbph.computeOptimalConfig();
					qbph.optimizedComputeOptimalConfig();
					
					IndexConfig ic = qbph.getOptimalConfig();
															
					
	//////			//for testing only
	//				pIMPIndexCost = qbph.computeQueryCost(ic, nTotalNumQueryRequests, nTotalNumBuildRequests);
	//				IMPIndexqp = qp;
	//				System.out.println("StreamId = "+intSTREAMId);				
					System.out.println("IndexConfig = "+ic.toString());
	//				float pCost = qbph.computeQueryCost(ic, nTotalNumQueryRequests, nTotalNumBuildRequests);
	//				System.out.println("Cost = "+pCost);
	
					//for testing
	//				qbph = new ExhaustiveEnum();
	//				qbph.init((float)dComparisoncost, (float)dHashCost, (float)(1.0*nTotalNumBuildRequests), (float)(1.0*nTotalNumQueryRequests), (float)(1.0),nNumBits);
	//				qbph.init(nNumCols, nNumBits);
	//				qbph.computeNumConfigs();
	//				qbph.setQueryPatterns(qp);
	//				qbph.computeOptimalConfig();
	//				//qbph.printAllCosts();
	//				ic = qbph.getOptimalConfig();
	//				System.out.println("StreamId = "+intSTREAMId);				
	//				System.out.println("IndexConfig = "+ic.toString());
	//				pCost = qbph.computeQueryCost(ic, nTotalNumQueryRequests, nTotalNumBuildRequests);
	//				System.out.println("Cost = "+pCost);
					
	
	//				if (bStoreActualSRIA)
	//				{	
	//					
	//					WriteToFile(strStoreActualIndicesFile, String.valueOf(intSTREAMId), false);
	//					
	//					WriteToFile(strStoreActualIndicesFile, ic.toString(), false);
	//					
	//					QueryPattern[] qp2 = getActualFreqQueryPatterns(nNumBits);
	//					//optimal index
	//					ExhaustiveEnum qbph2 = new ExhaustiveEnum();			
	//					qbph2.init((float)dComparisoncost, (float)dHashCost, (float)(1.0*nTotalNumBuildRequests), (float)(1.0*nTotalNumQueryRequests), (float)(1.0),nNumBits);
	//					qbph2.init(nNumCols, nNumBits);
	//					qbph2.setQueryPatterns(qp2);
	//					//Find optimal cost.			
	//					qbph2.optimizedComputeOptimalConfig();
	//					IndexConfig ic2 = qbph2.getOptimalConfig();
	//					
	//					WriteToFile(strStoreActualIndicesFile, ic2.toString(), true);
	//					
	//					qbph2 = null;
	//
	//				}	
						
		
					int[] NumBits = ic.getNumBits();
					int i= 0;
					for (int j = 0; j<NumBits.length;j++)
					{
						if (NumBits[j]>0)
						{
							i++;
						}
					}
						
					int[] GroupPos = new int[NumBits.length];
					int[] IndexAttrPos = new int[i];
					int[] NumEntries = new int[i];
						
					i= 0;
					boolean bOneFound = false;
					for (int j = 0; j<NumBits.length;j++)
					{
						GroupPos[j] = j;
						
						if (NumBits[j]>0)
						{	
							IndexAttrPos[i] = j;
							NumEntries[i] = NumBits[j];
							i++;
							bOneFound = true;
						}
					}
					
					if(bOneFound)
					{	
						float cQueryCost = qbph.computeQueryCost(ic, nTotalNumQueryRequests, nTotalNumBuildRequests)+ qbph.computetupleCost(nTotalNumBuildRequests);
						float cCreationCost = qbph.computetupleCost(nTotalNumBuildRequests);
						UpdateIndex(GroupPos, IndexAttrPos, NumEntries, qbph, cQueryCost,cCreationCost);
					}		
				}
				
//				System.out.println("Completing Exhaustive selection.");
			}
		}
		
		//clear
		ClearFreqQueryPatterns();		
		nTotalNumQueryRequests = 0;
		nTotalNumBuildRequests = 0;
	}
//////////////////////////////////////////////////////
	public void endExhaustiveAssessment(int nNumBits,String strTrackTimeToSearchOutputfile)
	{
		bExhaustiveAssessment = false;
		if (!bImpIndex)
		{
//			System.out.println("Running Exhaustive selection.");
			//NextHashKeys
			for (int i = 0; i<nNumLimitedIndices;i++)
			{
				Object obj = Collections.max(AccessModuleIndexAssessment);
				
				int index = AccessModuleIndexAssessment.indexOf(obj);
				NextHashKeys.add(index);
				AccessModuleIndexAssessment.set(index, 0);
				HashIndexData hid = CurrentHashIndexData.get(i);
				Iterator<Integer> it = HashKeyIndex.get(index).iterator();
				
				int p =0;
				int nIndxs = HashKeyIndex.get(index).size();
				int[] hf = new int[nIndxs];
				while(it.hasNext())
				{
					int indx = it.next();
					//index
					hid.IndexAttrPosArray[p] = indx;
				
					//num bits
					hf[p] =  nNumBits/nIndxs; 
					
				}
				hid.HashFactors = hf;
				
			}
			
//			if (!bPartition)
//			{	
//				if (!NextHashKeys.containsAll(CurrentHashKeys))
//				  MigrateTuples();
//				else
//				  NextHashKeys = null;
//			}
			//	
//			System.out.println("End Exhaustive selection.");			
		}	
		else
		{	
			QueryPattern[] qp = getFreqQueryPatterns(nNumBits);
			
			if (qp != null)
			{	
	//			System.out.println("Running Exhaustive selection.");
				
				
	//			WriteToFile(strTrackTimeToSearchOutputfile, String.valueOf(System.currentTimeMillis()), false);
				
				ExhaustiveEnum qbph = new ExhaustiveEnum();
				//OWQCLeadingHeuristic qbph = new OWQCLeadingHeuristic(); 
				qbph.init((float)dComparisoncost, (float)dHashCost, (float)(1.0*nTotalNumBuildRequests), (float)(1.0*nTotalNumQueryRequests), (float)(1.0),nNumBits,MAX_PARTITION_SIZE);
				qbph.init(nNumCols, nNumBits);
				qbph.setQueryPatterns(qp);
				
	//			OWQCLeadingHeuristic qbph = new OWQCLeadingHeuristic();
	//			qbph.init((float)0.000056, (float)0.001392, (float)(1.0*nTotalNumBuildRequests), (float)(1.0*nTotalNumQueryRequests), (float)(1.0*nTotalNumQueryRequests),nNumBits);
	//			qbph.init(nNumCols,nNumBits);
	////			qbph.computeNumConfigs();
	//			qbph.setQueryPatterns(qp);
				
				
				boolean bRunSelection = bNoIndexSelectionTrigger;
				//////////////////////////////////////////////////////
				if(!bRunSelection)
				{	
					if(!bFirstAssessment &&( bCostIncrIndexSelectionTrigger || bCombinedIncrIndexSelectionTrigger))
					{	
						//need to compute just the cost to probe using current
						//IMP index
						
						int[] numBits = new int[nNumCols];
						int[] AttrPosArray  = IMPIndex.getIndexAttrPosArray();
						int[] Bits  = IMPIndex.getNumIndexEntries();
						int apPos = 0;
						
						for (int i=0; (i<nNumCols)&&(apPos<AttrPosArray.length); i++)
						{
							if(AttrPosArray[apPos] == i)
							{	
								numBits[i] =  Bits[apPos];
								apPos++;
							}	
							else
								numBits[i] =  0;						
						}
						IndexConfig ic = new IndexConfig(numBits);		
						float pCost = qbph.computeQueryCost(ic, nTotalNumQueryRequests, nTotalNumBuildRequests);
						if(Math.abs(1.0-(pCost/pIMPIndexCost)) <= dDegreeCostIncr)
							bRunSelection = true;
						else
							bRunSelection = false;
					}
				
					//////////////////////////////////////////////////////
					if(!bFirstAssessment &&( bFreqIncrIndexSelectionTrigger || (bCombinedIncrIndexSelectionTrigger && !bRunSelection)))				
					{	
						//freq diff
						double dDegFreqDiff = 0.0;
						for(int i=0;i<qp.length;i++)
						{
							if(qp[i].getWeight()>0)
							{	
								if (IMPIndexqp[i].getWeight()>0)
									dDegFreqDiff += Math.abs(1.0-(qp[i].getWeight()/IMPIndexqp[i].getWeight()));
								else
									dDegFreqDiff += 1.0;
							}	
						}
						if(dDegFreqDiff <= dDegreeFreqIncr)
							bRunSelection = true;
						else
							bRunSelection = false;
						
					}
				}
				
				//////////////////////////////////////////////////////
				if(bRunSelection || bFirstAssessment)
				{	
					bFirstAssessment = false;				
	//				System.out.println("Find optimal cost");
	
					//Find optimal cost.	
					qbph.computeNumConfigs();
					//qbph.computeOptimalConfig();
					qbph.optimizedComputeOptimalConfig();
											
	//				System.out.println("Find optimal Config...");
					
					IndexConfig ic = qbph.getOptimalConfig();
					pIMPIndexCost = qbph.computeQueryCost(ic, nTotalNumQueryRequests, nTotalNumBuildRequests);
					IMPIndexqp = qp;
					
	
	//				if (bStoreActualSRIA)
	//				{	
	//					
	//					WriteToFile(strStoreActualIndicesFile, String.valueOf(intSTREAMId), false);
	//					
	//					WriteToFile(strStoreActualIndicesFile, ic.toString(), false);
	//					
	//					QueryPattern[] qp2 = getActualFreqQueryPatterns(nNumBits);
	//					//optimal index
	//					ExhaustiveEnum qbph2 = new ExhaustiveEnum();			
	//					qbph2.init((float)0.000056, (float)0.001392, (float)(1.0*nTotalNumBuildRequests), (float)(1.0*nTotalNumQueryRequests), (float)(1.0),nNumBits);
	//					qbph2.init(nNumCols, nNumBits);
	//					qbph2.setQueryPatterns(qp2);
	//					//Find optimal cost.			
	//					qbph2.optimizedComputeOptimalConfig();
	//					IndexConfig ic2 = qbph2.getOptimalConfig();
	//					
	//					WriteToFile(strStoreActualIndicesFile, ic2.toString(), true);
	//					
	//					qbph2 = null;
	//
	//				}	
	
	////				//for testing only
	////				System.out.println("StreamId = "+intSTREAMId);				
	//				System.out.println("IndexConfig = "+ic.toString());
					
	//				WriteToFile(strTrackTimeToSearchOutputfile, String.valueOf(System.currentTimeMillis()), true);
		
					int[] NumBits = ic.getNumBits();
					int i= 0;
					for (int j = 0; j<NumBits.length;j++)
					{
						if (NumBits[j]>0)
						{
							i++;
						}
					}
						
					int[] GroupPos = new int[NumBits.length];
					int[] IndexAttrPos = new int[i];
					int[] NumEntries = new int[i];
						
					i= 0;
					boolean bOneFound = false;
					for (int j = 0; j<NumBits.length;j++)
					{
						GroupPos[j] = j;
						if (NumBits[j]>0)
						{	
							IndexAttrPos[i] = j;
							NumEntries[i] = NumBits[j];
							i++;
							bOneFound = true;
						}
					}
					if(bOneFound)
					{	
						float cQueryCost = qbph.computeQueryCost(ic, nTotalNumQueryRequests, nTotalNumBuildRequests)+ qbph.computetupleCost(nTotalNumBuildRequests);
						float cCreationCost = qbph.computetupleCost(nTotalNumBuildRequests);
						UpdateIndex(GroupPos, IndexAttrPos, NumEntries, qbph, cQueryCost,cCreationCost);
					}		
				}	
				qbph = null;
			}
		}	
		//clear
		ClearFreqQueryPatterns();		
		nTotalNumQueryRequests = 0;
		nTotalNumBuildRequests = 0;
	    nSubTotalNumQueryRequests=0;
	}
	
//////////////////////////////////////////////////////
	//FreqQueryPatterns
	public QueryPattern[] getFreqQueryPatterns(int nNumBits)
	{
		QueryPattern[] qp = new QueryPattern[nNumQueryPatterns];
		boolean bOneResult = false;
		
//		System.out.println("=========");
//		System.out.println("FreqQueryPatterns");
//		System.out.println("Stream ID = "+intSTREAMId);
		
		if (bSRIA || bHHH)
		{				
			if (!bCombinedStats)
			{					
				for (int i = 0; i< nNumQueryPatterns;i++)
				{
					qp[i] = new QueryPattern();
					qp[i].setPatternID(i);
				}
				
				Iterator It = SRIA.keySet().iterator();
				
				ArrayList <Integer> keysAdded = new ArrayList<Integer>();
				
		        while (It.hasNext()) 
		        {
		            Integer Key = (Integer)(It.next());
					if (SRIA.get(Key).nCountQueryPattern > 0)
					{	
						bOneResult = true;
//						System.out.println("QueryPattern = "+Integer.toBinaryString(Key));
//						System.out.println("CountQueryPattern = "+SRIA.get(Key).nCountQueryPattern);
//						System.out.println("TotalNumQueryRequests = "+nTotalNumQueryRequests);
//						float freq2 =((float)SRIA.get(Key).nCountQueryPattern/(float)(nTotalNumQueryRequests*1.0)); 
//						System.out.println("Weight = "+freq2);
						
						float freq =((float)SRIA.get(Key).nCountQueryPattern/(float)(nTotalNumQueryRequests*1.0)); 
						if(bHHH)
						{
							if((freq-(freq/(2^nNumBits)))>=(dCostFactorSignificance-dHHErrorRate))
							{	
							  qp[Key].setWeight(freq);
//							  System.out.println("Added ");
							}  
							else
							{	
							  qp[Key].setWeight(0);
//							  System.out.println("Not Added ");
							} 
						}	
						else if(bHH)
						{	
							if((freq)>=(dCostFactorSignificance-dHHErrorRate))								
							  qp[Key].setWeight(freq);
//							System.out.println("Added ");						
						}	
						else 
							qp[Key].setWeight(freq);
							
					}				
				}
		        
			}
			else //SRIA combined stats
			{
//				System.out.println("SRIA combined stats");
				
				HashMap<Integer, IndexAssessment> combinedStats = new HashMap<Integer, IndexAssessment>();
				HashMap<Integer, IndexAssessment> stat = new HashMap<Integer, IndexAssessment>();
				
				//combine prev stats
				for(int i=0;i<storedStats.size();i++)
				{
//					System.out.println("SRIA combining "+i);

					stat = (HashMap)storedStats.get(i);
					
					//loop through stats
					Iterator It = stat.keySet().iterator();
			        while (It.hasNext()) 
			        {
			            Integer Key = (Integer)(It.next());
			            IndexAssessment IA = stat.get(Key);
			            
			            int count = (int)(IA.nCountQueryPattern * CombinedStatsFactor[nNumCombinedStats - storedStats.size()]);
			            
			            if(combinedStats.get(Key) != null)
			            	combinedStats.get(Key).nCountQueryPattern += count;
			            else
			            {
        					IndexAssessment newIA = new IndexAssessment(Key,count,0);        					
        					combinedStats.put(Key, newIA);			            	
			            }
			            	
			            nTotalNumQueryRequests += count;    
			        }
					
				}
				
				//add in current stats
				int nCount;
				for (int Key = 0; Key< nNumQueryPatterns;Key++)
				{
					qp[Key] = new QueryPattern();
					qp[Key].setPatternID(Key);

					if (SRIA.get(Key) != null)
					{	
						bOneResult = true;
						
						if(combinedStats.get(Key) != null)
						  nCount = SRIA.get(Key).nCountQueryPattern+combinedStats.get(Key).nCountQueryPattern;
						else
						  nCount = SRIA.get(Key).nCountQueryPattern;
						
						if (nCount >0)
						{	
//							System.out.println("QueryPattern = "+Integer.toBinaryString(Key));
//							System.out.println("CountQueryPattern = "+SRIA.get(Key).nCountQueryPattern);
//							System.out.println("TotalNumQueryRequests = "+dTotalNumQueryRequests);
//							System.out.println("Weight = "+nCount/dTotalNumQueryRequests);

							if(bHH && (((float)nCount/(float)(nTotalNumQueryRequests*1.0))>=(dCostFactorSignificance-dHHErrorRate)))
							{	
								qp[Key].setWeight((float)nCount/(float)(nTotalNumQueryRequests*1.0));
//								System.out.println("Added ");						
							}	
							else if(!bHH)
								qp[Key].setWeight((float)nCount/(float)(nTotalNumQueryRequests*1.0));
//							else
//								System.out.println("Not Added ");						
						}	
					}		
					else if(combinedStats.get(Key) != null)
					{	
						bOneResult = true;
//						System.out.println("QueryPattern = "+Integer.toBinaryString(Key));
//						System.out.println("CountQueryPattern = "+SRIA.get(Key).nCountQueryPattern);
//						System.out.println("TotalNumQueryRequests = "+nTotalNumQueryRequests);
//						System.out.println("Weight = "+SRIA.get(Key).nCountQueryPattern/dTotalNumQueryRequests);
						
						nCount = combinedStats.get(Key).nCountQueryPattern;
						
						if (nCount >0)
						{	
							if(bHH && (((float)nCount/(float)(nTotalNumQueryRequests*1.0))>=(dCostFactorSignificance-dHHErrorRate)))
							{	
								qp[Key].setWeight((float)nCount/(float)(nTotalNumQueryRequests*1.0));
//								System.out.println("Added ");						
							}	
							else if(!bHH)
								qp[Key].setWeight((float)nCount/(float)(nTotalNumQueryRequests*1.0));
//							else
//								System.out.println("Not Added ");						
						}	
					}		
					
				}
			
				
				//store current stats
				if(storedStats.size()==nNumCombinedStats)
				{
					storedStats.remove(0);
				}
				storedStats.add(SRIA);
			}
		}	
		else
		{	
			for (int i = 0; i< nNumQueryPatterns;i++)
			{
				qp[i] = new QueryPattern();
				qp[i].setPatternID(i);
				
				if (FreqQueryPatterns[i] > 0)
				{	
					bOneResult = true;
//					System.out.println("i = "+Integer.toBinaryString(i));
//					System.out.println("Weight = "+FreqQueryPatterns[i]/dTotalNumQueryRequests);
//					
					qp[i].setWeight((float)FreqQueryPatterns[i]/(float)(nTotalNumQueryRequests*1.0));				
					//reset assessment values???
					FreqQueryPatterns[i] = 0;
				}
				
			}
		}
		if(bOneResult)
			return qp;
		else
			return null;
			
	}
//////////////////////////////////////////////////////
	//FreqQueryPatterns
	public QueryPattern[] getActualFreqQueryPatterns(int nNumBits)
	{
		QueryPattern[] qp2 = new QueryPattern[nNumQueryPatterns];
		boolean bOneResult = false;
		
		
			
		for (int i = 0; i< nNumQueryPatterns;i++)
		{
			qp2[i] = new QueryPattern();
			qp2[i].setPatternID(i);
		}
				
		Iterator It = actualSRIA.keySet().iterator();
		while (It.hasNext()) 
		{
		   Integer Key = (Integer)(It.next());
		   if (actualSRIA.get(Key).nCountQueryPattern > 0)
		   {	
				bOneResult = true;
						
				qp2[Key].setWeight(actualSRIA.get(Key).nCountQueryPattern/(float)(nTotalNumQueryRequests*1.0));
							
			}				
		}
		
		if(bOneResult)
			return qp2;
		else
			return null;
			
	}
	
	//FreqQueryPatterns
	public void ClearFreqQueryPatterns()
	{

		if (!bImpIndex)
		{	
			AccessModuleIndexAssessment = new ArrayList<Integer>();
			for (int t = 0; t< HashKeyIndex.size();t++)
				AccessModuleIndexAssessment.add(0);
		}
		else
		{	
			//reset assessment values???
			if (bSRIA || bHHH)
			{	
	//			System.out.println("Clearing out SRIA");
				while (SRIA.size()>0)
				{	
					Iterator It = SRIA.keySet().iterator(); 
			        Integer Key = (Integer)(It.next());
			        
	//		        IndexAssessment IA = SRIA.get(Key);
	//		        IA = null;
			        
			        SRIA.remove(Key);	                
			        
				}    
				SRIA.clear();
				SRIA = new HashMap<Integer, IndexAssessment>();
				
				if(bStoreActualSRIA)
				{
					actualSRIA.clear();
	//				actualSRIA=null;	
					
					actualSRIA = new HashMap<Integer, IndexAssessment>();				
				}
				
				nBucketId = 1;
				
				if(bHHH)
				{	
					if(LatticeLeafNodes != null)
					{	
						Iterator itr = LatticeLeafNodes.iterator();
					    while(LatticeLeafNodes.size()>0)
					     {
	//				    	IndexAssessment IAn = (IndexAssessment)LatticeLeafNodes.get(0);
	//						IAn = null;
			                LatticeLeafNodes.remove(0);
					     }
					    
					    LatticeLeafNodes.clear();
	//				    LatticeLeafNodes=null;
					}  
					LatticeLeafNodes = new ArrayList<IndexAssessment>();
				}	
			}	
			else
			{
				if(FreqQueryPatterns==null)
				  FreqQueryPatterns = new float[nNumQueryPatterns];
	
				for (int i = 0; i< nNumQueryPatterns;i++)
					FreqQueryPatterns[i] = 0;		
			}
		}
		//System.gc();
	}
	
	//dTotalNumQueryRebquests
	public int getTotalNumQueryRequests()
	{
//		if (bHH || bHHH) 
//		{	
//			if (nSubTotalNumQueryRequests > ((nTotalNumQueryRequests / (1/dHHErrorRate))))
//			{
////				System.out.println("Start Prunning....");
//				if(bHHH)
//					PruneHHH();
//				else	
//					Prune();
//			    nBucketId++;	
////				System.out.println("End Prunning....");
//			    nSubTotalNumQueryRequests=0;
//			}
//		}
		
		return nTotalNumQueryRequests;
	}
	
///////////////////////////////////////////////////////////////////////////////////	
	//prune
	public void Prune()
	{
//		if(bStoreComparisonTime)
//		  WriteToFile(strStoreComparisonTimeFile, String.valueOf(System.currentTimeMillis()), false);
		
		Iterator It = SRIA.keySet().iterator();
		ArrayList<Integer> keyToRemove = new ArrayList<Integer>(); 
        while (It.hasNext()) 
        {
                Integer Key = (Integer)(It.next());
                IndexAssessment IA = SRIA.get(Key);
                if((IA.nCountQueryPattern+IA.nDelta)< nBucketId)
                {
//                	System.out.println("Removing Qp"+Key);
//                	System.out.println("Qp="+Integer.toBinaryString(Key)+" IA.nCountQueryPattern+IA.nDelta = "+(IA.nCountQueryPattern+IA.nDelta));
//    				System.out.println("TotalNumQueryRequests = "+dTotalNumQueryRequests);
//    				System.out.println("BucketId = "+nBucketId);
                	keyToRemove.add(Key);
                }
                
        }
        
        while(keyToRemove.size()>0)
        {
        	int key = keyToRemove.remove(0);
        	SRIA.remove(key);
        }
        
        
//		if(bStoreComparisonTime)
//			  WriteToFile(strStoreComparisonTimeFile, String.valueOf(System.currentTimeMillis()), true);
	}
	
	//prune HHH
	public void PruneHHH()
	{
//		if(bStoreComparisonTime)
//			  WriteToFile(strStoreComparisonTimeFile, String.valueOf(System.currentTimeMillis()), false);

		int i = 0;
		if ((LatticeLeafNodes != null) && (LatticeLeafNodes.size() > 0 ))
		{	
		    while(i < LatticeLeafNodes.size())
		     {
		    
		    	IndexAssessment IA = LatticeLeafNodes.get(i); 
//	            System.out.println("Processing QP ="+Integer.toBinaryString(IA.nQueryPattern));
//	            System.out.println("LatticeLeafNodes.size="+LatticeLeafNodes.size());
		    	
		    	if((IA.nCountQueryPattern+IA.nDelta)< nBucketId)
	            {
//	               System.out.println("Removing ");
//	               System.out.println("Qp="+Integer.toBinaryString(IA.nQueryPattern)+" IA.nCountQueryPattern+IA.nDelta = "+(IA.nCountQueryPattern+IA.nDelta));
//	    		   System.out.println("TotalNumQueryRequests = "+TotalNumQueryRequests);
//	    		   System.out.println("BucketId = "+nBucketId);
	    		   
				   if(IA.nCountQueryPattern!=0)
				   {   
					   String bs = Integer.toBinaryString(IA.nQueryPattern);
					   
					   BitSet IABS = new BitSet();
					   for (int s=bs.length();s>0;s--)
					   {   
						   
						   char ch = bs.charAt(s-1);
//			               System.out.println("setting s ="+s+" char="+ch);
			               
						   if(String.valueOf(ch).equalsIgnoreCase("1"))
							   IABS.set(bs.length()-s);
						   else
						       IABS.clear(bs.length()-s);
					   }
						  
//		               System.out.println("QueryPattern ="+IA.nQueryPattern);
//		               System.out.println("IABS ="+IABS.toString());
		    		   
		    		   //combine with parent
		    		   if (bCombineNodesByPickingOne)
		    		   {
		    			   //find first parent
		    			   int nNextSetPos = 0;
		    			   nNextSetPos = IABS.nextSetBit(nNextSetPos);
		    			   
		    			   boolean bParentFound = false;
						   BitSet testIABS = null;
						   ArrayList<IndexAssessment> PosParents = new ArrayList<IndexAssessment>();
						   
	    				   testIABS = (BitSet)IABS.clone();
	    				   
		    			   while(nNextSetPos >0)
		    			   {
//		    				   System.out.println("nNextSetPos="+nNextSetPos);
//				               System.out.println("testIABS ="+testIABS.toString());
		    				   
		    				   testIABS.clear(nNextSetPos);
		    				   
		    				   int pqp=0;
		    				   for (int q=testIABS.length()-1; q>=0; q--) 
		    				   {
		    				       if (testIABS.get(q)) 
		    				       {
		    				    	   pqp = pqp*2+1;
		    				       }
		    				       else if(pqp>0)
		    				    	   pqp = pqp*2;
		    				   }
		
		    				   if(SRIA.get(pqp) != null)
		    				   {
		    					   //parent exists
		    					   bParentFound = true;		    					   
		    					   PosParents.add(SRIA.get(pqp));
		    				   }
		    				   nNextSetPos = testIABS.nextSetBit(nNextSetPos);
		    			   }
		    			   
		    			   if (bParentFound)
		    			   {
		    				   //need to select random parent
		    				   int pp = generator.nextInt(PosParents.size());
		    				   
		    				   
		    				   IndexAssessment IAP  = PosParents.get(pp);
		    				   IAP.nCountQueryPattern += IA.nCountQueryPattern;
		    				   
//				               System.out.println("Parent QP ="+IAP.nQueryPattern);
		    				   
		    			   }
		    			   else //create a new parent none found
		    			   {
		    				 nNextSetPos = IABS.nextSetBit(0);
		    				 if(nNextSetPos != -1)
		    				 {	 
		     				   testIABS = (BitSet)IABS.clone();
		    				   testIABS.clear(nNextSetPos);
		    				   
		    				   int pqp=0;
		    				   for (int q=testIABS.length()-1; q>=0; q--) 
		    				   {
		    				       if (testIABS.get(q)) 
		    				       {
		    				    	   pqp = pqp*2+1;
		    				       }
		    				       else if(pqp>0)
		    				    	   pqp = pqp*2;
		    				   }
		    				   
		   					   IndexAssessment IANP = new IndexAssessment(pqp,IA.nCountQueryPattern,IA.nDelta);
//				               System.out.println("Parent QP ="+IANP.nQueryPattern);
							
		    				   SRIA.put(pqp, IANP);
		    				 }  
		    			   }
		    			   PosParents.clear();
		    			   PosParents=null;
		    		   }
		    		   else if(bPickParentWithHighestFreq)
		    		   {
		    			   //find first parent
		    			   int nNextSetPos = 0;
		    			   nNextSetPos = IABS.nextSetBit(nNextSetPos);
		    			   boolean bParentFound = false;
						   BitSet testIABS = (BitSet)IABS.clone();
						   IndexAssessment Parent = null;
		    			   while(nNextSetPos >0)
		    			   {
		    				   testIABS.clear(nNextSetPos);
		    				   
		    				   int pqp=0;
		    				   for (int q=testIABS.length()-1; q>=0; q--) 
		    				   {
		    				       if (testIABS.get(q)) 
		    				       {
		    				    	   pqp = pqp*2+1;
		    				       }
		    				       else if(pqp>0)
		    				    	   pqp = pqp*2;
		    				   }
		    				   
		
		    				   if(SRIA.get( pqp) != null)
		    				   {
		    					   if(Parent == null)
		    					   {   
		    						   //parent exists
		    						   bParentFound = true;
		    					   
		    						   Parent = SRIA.get(pqp);
		    					   } 
		    					   else if(Parent.nCountQueryPattern<SRIA.get(pqp).nCountQueryPattern)
		    					   {
		    						   Parent = SRIA.get(pqp);    						   
		    					   }
		    				   }
		    				   nNextSetPos = testIABS.nextSetBit(nNextSetPos);
		    			   }
		    			   
		    			   if (bParentFound)
		    			   {
		    				   //need to select random parent
		    				   Parent.nCountQueryPattern += IA.nCountQueryPattern;
		    				   
		    			   }
		    			   else //create a new parent none found
		    			   {
		      				 nNextSetPos = IABS.nextSetBit(0);
		    				 if(nNextSetPos != -1)
		    				 {	 
		     				   testIABS = (BitSet)IABS.clone();
		    				   testIABS.clear(nNextSetPos);
		    				   
		    				   int pqp=0;
		    				   for (int q=testIABS.length()-1; q>=0; q--) 
		    				   {
		    				       if (testIABS.get(q)) 
		    				       {
		    				    	   pqp = pqp*2+1;
		    				       }
		    				       else if(pqp>0)
		    				    	   pqp = pqp*2;
		    				   }
		    				   
		   					   IndexAssessment IANP = new IndexAssessment(pqp,IA.nCountQueryPattern,IA.nDelta);
							
		    				   SRIA.put(pqp, IANP);
		    				 }  
		    				   
		    			   }
			    		   testIABS = null;
		    		   }
		    		   
		    		   IABS = null;
	    		     //remove node
	    		     SRIA.remove(IA.nCountQueryPattern);
		             
	    		     LatticeLeafNodes.remove(i);
//	 	            System.out.println("LatticeLeafNodes.size="+LatticeLeafNodes.size());

				   }
	            }
	            else
	            	i++;
	        }	
        }	
//		if(bStoreComparisonTime)
//			  WriteToFile(strStoreComparisonTimeFile, String.valueOf(System.currentTimeMillis()), true);
	}
//////////////////////////////////////////////////////
	//uses NoSampling?
	public void setNoSampling(boolean val)
	{
		bNoSampling = val;
	}

	public boolean getNoSampling()
	{
		return bNoSampling;
	}
//////////////////////////////////////////////////////
	//uses SimpleRandomSampling?
	public void setSimpleRandomSampling(boolean val)
	{
		bSimpleRandomSampling = val;
	}

	public boolean getSimpleRandomSampling()
	{
		return bSimpleRandomSampling;
	}
//////////////////////////////////////////////////////
	//SimpleRandomPercent
	public void setSimpleRandomPercent(int val)
	{
		nSimpleRandomPercent = val;
	}

	public int getSimpleRandomPercent()
	{
		return nSimpleRandomPercent;
	}
	
//////////////////////////////////////////////////////
	//uses SimpleRandomSampling?
	public void setStratifiedRandomSampling(boolean val)
	{
		bStratifiedRandomSampling = val;
	}

	public boolean getStratifiedRandomSampling()
	{
		return bStratifiedRandomSampling;
	}
//////////////////////////////////////////////////////
	//StratifiedRandomPercent
	public void setStratifiedRandomPercent(int val)
	{
		nStratifiedRandomPercent = val;
	}

	public int getStratifiedRandomPercent()
	{
		return nStratifiedRandomPercent;
	}
//////////////////////////////////////////////////////
	//StratifiedRandomWindow
	public void setStratifiedRandomWindow(int val)
	{
		nStratifiedRandomWindow = val;
	}

	public int getStratifiedRandomWindow()
	{
		return nStratifiedRandomWindow;
	}
	
//////////////////////////////////////////////////////
	//uses SystematicSampling?
	public void setSystematicSampling(boolean val)
	{
		bSystematicSampling = val;
	}

	public boolean getSystematicSampling()
	{
		return bSystematicSampling;
	}
//////////////////////////////////////////////////////
	//SystematicInterval
	public void setSystematicInterval(int val)
	{
		nSystematicInterval = val;
	}
	public int getSystematicInterval()
	{
		return nSystematicInterval;
	}
//////////////////////////////////////////
	public void setNoMigration(boolean val)
	{
		bNoMigration=val;		
	}
	public boolean getNoMigration()
	{
		return bNoMigration;		
	}
	
	public void setSymmetricMigration(boolean val)	
	{
		bSymmetricMigration=val;		
	}
	public boolean getSymmetricMigration()	
	{
		return bSymmetricMigration;		
	}
	
	public void setPartitionMigration(boolean val)
	{
		bPartitionMigration=val;		
	}
	public boolean getPartitionMigration()
	{
		return bPartitionMigration;		
	}
	
	public void setCompleteMigration(boolean val)
	{
		bCompleteMigration=val;		
	}
	public boolean getCompleteMigration()
	{
		return bCompleteMigration;		
	}
/////////////////////////////////////////////////////////
//	SRIA
	public void setSRIA(boolean val)
	{
		bSRIA=val;		
	}
	public boolean getSRIA()
	{
		return bSRIA;		
	}
	
	public void setHH(boolean val)
	{
		bHH=val;		
	}
	public boolean getHH()
	{
		return bHH;		
	}
	
	public void setHHH(boolean val)
	{
		bHHH=val;		
	}
	public boolean getHHH()
	{
		return bHHH;		
	}
	
	public void setCombineNodesByPickingOne(boolean val)
	{
		bCombineNodesByPickingOne=val;		
	}
	public boolean getCombineNodesByPickingOne()
	{
		return bCombineNodesByPickingOne;		
	}

	public void setPickParentWithHighestFreq(boolean val)
	{
		bPickParentWithHighestFreq=val;		
	}
	public boolean getPickParentWithHighestFreq()
	{
		return bPickParentWithHighestFreq;		
	}
	
	public void setHHErrorRate(double val)
	{
		dHHErrorRate = val;
	}

	public double getHHErrorRate()
	{
		return dHHErrorRate;
	}
	
	public void setCostFactorSignificance(double val)
	{
		dCostFactorSignificance = val;
	}

	public double getCostFactorSignificance()
	{
		return dCostFactorSignificance;
	}
////////////////////////////////////////////
//Combined stats ?
	
	public void setCombinedStats(boolean val)
	{
		bCombinedStats=val;		
	}
	public boolean getCombinedStats()
	{
		return bCombinedStats;		
	}
	
	public void setNumCombinedStats(int val)
	{
		nNumCombinedStats = val;
	}

	public int getNumCombinedStats()
	{
		return nNumCombinedStats;
	}

	public void setCombinedStatsFactor(double[] val)
	{
		CombinedStatsFactor = val;
	}

	public double[] getCombinedStatsFactor()
	{
		return CombinedStatsFactor;
	}
	
/*****************************************/
//Index Selection trigger
	public void setNoIndexSelectionTrigger(boolean val)
	{
		bNoIndexSelectionTrigger=val;		
	}
	public boolean getNoIndexSelectionTrigger()
	{
		return bNoIndexSelectionTrigger;		
	}

	public void setCostIncrIndexSelectionTrigger(boolean val)
	{
		bCostIncrIndexSelectionTrigger=val;		
	}
	public boolean getCostIncrIndexSelectionTrigger()
	{
		return bCostIncrIndexSelectionTrigger;		
	}

	public void setFreqIncrIndexSelectionTrigger(boolean val)
	{
		bFreqIncrIndexSelectionTrigger=val;		
	}
	public boolean getFreqIncrIndexSelectionTrigger()
	{
		return bFreqIncrIndexSelectionTrigger;		
	}

	public void setCombinedIncrIndexSelectionTrigger(boolean val)
	{
		bCombinedIncrIndexSelectionTrigger=val;		
	}
	public boolean getCombinedIncrIndexSelectionTrigger()
	{
		return bCombinedIncrIndexSelectionTrigger;		
	}
		
	public void setDegreeCostIncr(double val)
	{
		dDegreeCostIncr = val;
	}

	public double getDegreeCostIncr()
	{
		return dDegreeCostIncr;
	}

	public void setDegreeFreqIncr(double val)
	{
		dDegreeFreqIncr = val;
	}

	public double getDegreeFreqIncr()
	{
		return dDegreeFreqIncr;
	}
//////////////////////////////////////
	public void setNumCols(int val)
	{
		nNumCols = val;
	}
	
////////////////////////////////
	public static void WriteToFile(String fileLocation, String stringToWrite, boolean doNewLine)
	{
		try
		{
			//Create file 
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			
			out.write(stringToWrite);
			
			if (doNewLine)
			{
				out.newLine();
			}
			else
				out.append("|");
			
			out.flush();			    
					//Close the output stream
			out.close();			    
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}
//////////////////////////////
	//TrackOptimalIndices
	public void setStoreActualSRIA(boolean val)
	{
		bStoreActualSRIA=val;		
	}
	public boolean getStoreActualSRIA()
	{
		return bStoreActualSRIA;		
	}
	
	//TrackOptimalIndicesOutputFile
	public void setStoreActualIndicesFile(String val)
	{
		strStoreActualIndicesFile=val;		
	}
	public String getStoreActualIndicesFile()
	{
		return strStoreActualIndicesFile;		
	}
//////////////////////////////////////////////
	public void setContinuousAssessment(boolean CA)
	{
		bContinuousAssessment = CA;
	}
//////////////////////////////////////////////
	public void setCountMigrations(boolean CM)
	{
		bCountMigrations = CM;
	}
	
	public void setCountMigrationsFile(String strFile)
	{
		strCountMigration = strFile;
	}
//////////////////////////////////////////////
	public void setStoreComparisonTime(boolean CM)
	{
		bStoreComparisonTime = CM;
	}
	
	public void setStoreComparisonTimeFile(String strFile)
	{
		strStoreComparisonTimeFile = strFile;
	}

//////////////////////////////////////////////	
	//Comparison cost
	public void computeComparisoncost()
	{	
//	System.out.println("Comparison cost");
//	Random generator;
	generator = new Random();
  	generator.setSeed(new Date().getTime());
	long totalTime = 0;
	long y;
	for (y = 0;y<1000;y++)
	{
		double val1,val2;
    	val1=((double)generator.nextInt(1000000000)+1.0)/1000000000.0;
    	val2=((double)generator.nextInt(1000000000)+1.0)/1000000000.0;
    	
//		System.out.println(y);
		
    	Date startTime = new Date();
		if (val1>val2) 
		{
			;
		}
        Date endTime = new Date();
        totalTime+=endTime.getTime() - startTime.getTime();
	}

	dComparisoncost =totalTime/((1.0)*(y-1));
//	System.out.println("comparisonCost ="+comparisonCost);
	}
//////////////////////////
//	Hash cost
public void computeHashcost()
{
	//System.out.println("Hash cost");
	Random generator;
	generator = new Random();
  	generator.setSeed(new Date().getTime());
	long totalTime = 0;
	Hashtable<Integer,String> h = new Hashtable<Integer,String>();
	
	long y;
	for (y = 0;y<1000;y++)
	{
		int val1;
    	val1=generator.nextInt(1000000000);
    	
		System.out.println(y);
		
    	Date startTime = new Date();
    	h.put(val1, "Hello World");
        Date endTime = new Date();
        totalTime+=endTime.getTime() - startTime.getTime();
	}

	dHashCost =totalTime/((1.0)*(y-1));
	//System.out.println("Hash Cost ="+comparisonCost);
  }
///////////////////////
}
