/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.STeM.Partition;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Iterator;
import java.util.BitSet;

import edu.wpi.cs.dsrg.Mjoin.MJoinData;
import edu.wpi.cs.dsrg.eddyBatch.datastructure.*;
import edu.wpi.cs.dsrg.eddyBatch.operator.EddyPartitionOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.*;

public class STeMPartitionJoinProbeOperatorImp extends XATMultiQueueWindowStreamOperatorImp implements STeMPartitionJoinProbeOperator
{	
	public int windowSize;
	String outFile = "C:\\KEW\\Research\\Eddy\\execution_stats\\RUNTIME_STEMPROBE_COST.txt";
	long startTime;
	long endTime;
	long numProcessedTuples;
	
	// is this an eddy op or not
	private boolean IsEddyOp = false;
	
	//STQM: is self-tuning?
	private boolean IsSelfTuning = false;
	long lastUpdateTime = -999; //last time statistics of this operator was updated
	
	// index position of STem and tuple that are being compared
	
	// data set associated with STeM
	private STeMPartitionData STeMStreamData;
	
	//Info about probe
	private int ProbeSTREAMId;
	private ArrayList<Integer> NumTupleIndex;
	
	//Info about stored stem data
	private int DataSTREAMId;
	private ArrayList<Integer> NumSTeMIndex;
	private int STeMDataIndexArrayPos = 0;
	
	//operator bitmaps per STREAM
	ArrayList<BitSet> PreFilledPerStream;
	
	//used for QMesh to look up operators
	private OperatorLookUp OpLkUp;
	
	private int QMeshOperatorID;

	//used to determine when to create ruster
	int rusterSendOff;

	//CompleteJoinAttrib
	private int CompleteJoinAttrib;
	
	//number of results to batch together
	private int NumTupToProcessPerOp = 500;

	private boolean bCountActiveProcTups = true;
	private HashMap <Integer,Integer> PartitionProcTupCount;


	public STeMPartitionJoinProbeOperatorImp(XATNode init) 
	{
		super(init);
	}
		
	public boolean run() throws Exception 
	{		
		int totalNumberTupledEnq = 0;
		startTime = System.currentTimeMillis();
		RoutingToken currentRToken = null;
		PartitionProcTupCount = new HashMap <Integer,Integer>(); 
		
		//input queue from STREAM
		XATQueueSchema inSchema;
		int inLength;
		
		//only set bits for first result
		boolean bFirstResult = true;


		//for eddy bit set
		BitSet dbs = null;
		BitSet rbs = null;
		BitSet InvDone;
		EddyQueue q = null;
		
//		//complete join Attrib
//		BitSet CompletedJoinAttribs = null;
		
		//------------------------------------------------------------------------------
		// get input queue by operator id from the lookup table
		//------------------------------------------------------------------------------
		//XATQueue inputQueue = (XATQueue)this.OpLkUp.getOperatorQueue(this.getOperatorID());
		XATQueue inputQueue = (XATQueue)this.getInputQueue();
		
		if ((inSchema = inputQueue.getSchema()) == null)
		{
			//System.out.println("input queue schema is NULL");
			return false;
		}
		
		inLength = inSchema.getLength();

		if (this.getOutputQueue().getSchema() == null) 
		{
			init();
		} 
		
		
		//int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		int howMany  = 5000; //rimma temp "hack"
		
		int len = howMany;
		
		XATTuple t = null;
				
		
//		System.out.println("\nProcessing stream [" + this.ProbeSTREAMId + "]");
//		System.out.println("joining to  stream [" + this.DataSTREAMId + "]");
//////		/** get the tuples */
//		System.out.println("# of tuples in the queue: " + inputQueue.getQueueSize());

		XATTuple[] tuples = inputQueue.dequeue(len, this);
		
		numProcessedTuples += tuples.length;
		
//		System.out.println("\t# of tuples in the STEM (to be probed) hash: " + STeMStreamData.TupleList.size());	
//		System.out.println("Dequeued: " + tuples.length + " tuples.");
				
		//update property for statistics monitoring
		this.setProperty("NUMBER_OF_TUPLES_TO_DEQUEUE", ""+tuples.length); 		
							
		//RoutingToken rt = null;
		boolean bSentRt  = false;			
		boolean bProbe  = false;
		boolean bWaitForRToken  = false;
		
		//ArrayList<Integer> CopyRusterOpStack = null;
		//ArrayList outTuples = new ArrayList();
		ArrayList outTuples = new ArrayList();
		int NumExpectedTuplesInRuster = 0;
		int NumReceivedTuplesAfterRToken = 0;
				
		int extraTupleCounter = 0;
		XATTuple[] tuplesUntilRToken = null;
		int i = -1;
		int tuplesToProcess = tuples.length;
		boolean bAlreadyDequeuedCutoffRusterTuples = false;

		// Eddy Batching
		int numInBatch = 0;
		int NumBits = 0;
		int NumOps = 0;
		int NumJoinAttrib = 0;
		
// for testing only		
//		String strOps = "";
		
		//partition Number
		int PartitionNum=0;
		
		//for(int i = 0; i < tuples.length;i++)
		
		while((i++ < tuplesToProcess) || (!IsEddyOp && (NumReceivedTuplesAfterRToken < NumExpectedTuplesInRuster)) ||(IsEddyOp &&(numInBatch > 0)))
		{									
			XATTuple ReceivedTuple = null;
			
//			System.out.println("processing i = "+i);
			
			if (i < tuplesToProcess)
			{
//				System.out.println("Extracting tuple");
				ReceivedTuple = tuples[i];
			}
			else
			{	
				if (IsEddyOp && (numInBatch > 0))
				{
					if(tuplesUntilRToken==null)
					{	
						//dequeue the tuple left in the Batch
						tuplesUntilRToken = inputQueue.dequeue(numInBatch, this);
						//System.out.println("Beyond i ");
						extraTupleCounter = 0;
					}	
				}
				else if (!IsEddyOp && !bAlreadyDequeuedCutoffRusterTuples)
				{
//					System.out.println("");
//					//System.out.println("i: " + i);
//					System.out.println("tuplesToProcess: " + tuplesToProcess);
//					//System.out.println("size of input queue: " + inputQueue.size());
//					System.out.println("NumReceivedTuplesAfterRToken: " + NumReceivedTuplesAfterRToken);
//					System.out.println("NumExpectedTuplesInRuster: " + NumExpectedTuplesInRuster);						
					
					int h = 0;
					if (inputQueue.getContents().size() > 0)
					{						
						String currTupleStr = inputQueue.getContents().get(h).getClass().toString();
						while(!currTupleStr.endsWith("RoutingTokenImp"))
						{
							h++;
							if (h == inputQueue.getContents().size())
							{
								break;
							}
							currTupleStr = inputQueue.getContents().get(h).getClass().toString();
						}
					}
					
					//int diffTupleToDequeueCount = NumExpectedTuplesInRuster - (NumReceivedTuplesAfterRToken);
					int actualTupleToDequeueCount = h;
					//System.out.println("diffTupleToDequeueCount: " + diffTupleToDequeueCount);
					//System.out.println("actualTupleToDequeueCount: " + actualTupleToDequeueCount);
					
					if (actualTupleToDequeueCount <= 0)
					{
						return true;
					}
					
					if (inputQueue.getContents().size() > 0)
					{
						if (inputQueue.getContents().get(0).getClass().toString().endsWith("RoutingTokenImp"))
						{
							return true;
						}
					}
					//dequeue the difference in the number of tuples (to make sure we have dequeued a complete ruster)
					//in this queue, we will dequeue tuples until the next r-token
					tuplesUntilRToken = inputQueue.dequeue(actualTupleToDequeueCount, this);

					//System.out.println("tuplesUntilRToken: " + tuplesUntilRToken.length);
										
										
					if (tuplesUntilRToken == null)
					{
						return true;
					}
					else if (tuplesUntilRToken.length == 0)
					{ 
						//done (assume queue is empty)
						return true;	
					}										
					
					bAlreadyDequeuedCutoffRusterTuples = true; //already dequeued tuples, so just go and finish their processing
				}
				else
					break;
				
				if (!IsEddyOp)
				{	

					//System.out.println("extraTupleCounter: "+extraTupleCounter);
					if (extraTupleCounter == tuplesUntilRToken.length)
					{
						return true;
					}
					
					ReceivedTuple = tuplesUntilRToken[extraTupleCounter++];
				}	
			}
						
			EddyTuple et = null, net = null;
			EddyBatchTuple ebt = null;
			QMeshTuple qmt = null, result = null;

			// if this is an eddy op
			// parse the probe tuple from the eddy tuple
			if (IsEddyOp)
			{
				
				if (numInBatch==0) //first tuple must be an EddyBatchTuple
				{
					//only set bits for first result
					bFirstResult = true;

					// send out all old tuples in q
					
					if(q !=null)
					{							
						if(q.length()>0)
						{	
//							System.out.println("Sending results to Eddy...");
							EddyBatchTuple ebtOut =(EddyBatchTuple)q.dequeue();
							ebtOut.setNumberTuplesinBatch(q.length()+1);
							this.getOutputQueue().enqueue((XATTuple)ebtOut, this);						
							
							while(q.length()>0)
							{
								this.getOutputQueue().enqueue((XATTuple)q.dequeue(), this);						
							}
						}	
						q = null;
					}
					
					//set batch queue to empty
					q = new EddyQueue();
						
				 	ebt = (EddyBatchTuple)tuples[i];				 	
				 	numInBatch = ebt.getNumberTuplesinBatch();
				 	
//					System.out.println("numInBatch = "+numInBatch);
				 	
				 	et = new EddyTupleImp(ebt.getNumTuplesInAnswer(),  ebt.getPartitionNumber());
				 	
					et.setAnswer(ebt.getAnswer());
					
					//compute done bits for Eddy batch					
					//send Tup Found to eddy for
					// further processing
				
					// ready and done bits from tuple set to probe
					// with bits already set for this probe op
					dbs = (BitSet)ebt.getDoneBitSet().getBits().clone();
					rbs = (BitSet)ebt.getReadyBitSet().getBits().clone();
					NumBits = ebt.getDoneBitSet().getSize();
					NumOps = ebt.getReadyBitSet().getSize();
					
//					NumJoinAttrib = ebt.GetCompletedJoinAttrib().getSize();
//					
//					CompletedJoinAttribs = (BitSet)ebt.GetCompletedJoinAttrib().getBits().clone();
//					CompletedJoinAttribs.set(CompleteJoinAttrib);
					
//					System.out.println("start Rbs="+rbs.toString());
//					System.out.println("start Dbs="+dbs.toString());
//					System.out.println("PreFilledPerStream="+PreFilledPerStream.get(DataSTREAMId).toString());
//					System.out.println("EddyGetOpBitsPos="+this.EddyGetOpBitsPos().toString());
					
					for(int h = 0;h < this.EddyGetOpBitsPos().size();h++)
					{
						rbs.clear(this.EddyGetOpBitsPos().get(h));
						dbs.set(this.EddyGetOpBitsPos().get(h));
					}
			
//					System.out.println("after current stream Rbs="+rbs.toString());
//					System.out.println("after current stream Dbs="+dbs.toString());
						
					// get inverse of done bits
					// ie this represents ops not done
					InvDone = (BitSet)dbs.clone();
					InvDone.flip(1, dbs.size());			
				
					// and Inv of done bits with the rbs for the stream in the
					// solution
					// ie this represents all not done and ready
					BitSet rbsSolution = (BitSet)PreFilledPerStream.get(DataSTREAMId).clone();			
					rbsSolution.and(InvDone);
									
					// or the rbs from the solution with the normal rbs
					// ie this represents all not done by the added STREAM
					// and all that were not completed before this operator was called
					rbs.or(rbsSolution);
//					System.out.println("Rbs="+rbs.toString());
					
// for testing only
//					strOps = "";
//					if (ebt.getstrOps() != null)
//						strOps = ebt.getstrOps();
//					strOps +=","+QMeshOperatorID;

				}
				else if (i < tuplesToProcess)
				{	
					et = (EddyTuple)tuples[i];
				}
				else
				{
					et = (EddyTuple)tuplesUntilRToken[extraTupleCounter];
					extraTupleCounter++;
//					System.out.println("processing last tuples from batch");
				}
				
			 	ReceivedTuple = et.getTuple(ProbeSTREAMId);
			 	PartitionNum = et.getPartitionNumber();
			 	bProbe  = true;
			 	
			 	//for testing purposes only
//			 	System.out.println("Num in Batch = "+numInBatch);
//			 	System.out.println("tuplesToProcess = "+tuplesToProcess);
			 	
			 	numInBatch--;

			 	//check to see if this is a join or just a validation of the join rule
			 	if (et.getTuple(DataSTREAMId) != null)
			 		bProbe  = false;
			 	else
			 		bProbe  = true;
			 		
			}
			//----------------------------------------------------------------------------
			// GENERAL PROBE - FOR BOTH EDDY AND QUERY MESH
			//----------------------------------------------------------------------------
			// search the data for matches
			if (!bProbe)
			{	//not probing just validating the answer
				boolean bValidResult = true;
				
//				System.out.println("No Probe");
				for (int ti = 0; ti<NumTupleIndex.size()&&bValidResult;ti++)
				{	
					if (!ReceivedTuple.getCell(NumTupleIndex.get(ti)).convertToString().equals(et.getTuple(DataSTREAMId).getCell(NumSTeMIndex.get(ti)).convertToString()))
						bValidResult = false;
				}
				if (bValidResult)
				{
					if(bFirstResult)
					{
						//create Eddy Batch Tuple
						EddyBatchTuple netEbt = new EddyBatchTupleImp(et.getNumTuplesInAnswer(), NumOps, NumBits, et.getPartitionNumber());
						netEbt.setAnswer(et.getAnswer());
						
						netEbt.setDoneBitSet(dbs);
						netEbt.setReadyBitSet(rbs);
						
//						netEbt.SetCompletedJoinAttrib(CompletedJoinAttribs);
						
						//increment counter
						totalNumberTupledEnq++;
						
						//num tuples in batch
						//netEbt.setNumberTuplesinBatch(results.size());
						
//						netEbt.setstrOps(strOps);
						
						q.enqueue(netEbt); 
						bFirstResult = false;
						
					}
					else
					{	
						// create new eedy tuple for each intermediate answer
						net  = new EddyTupleImp(et.getNumTuplesInAnswer(),et.getPartitionNumber());
						net.setAnswer(et.getAnswer());
						
						//increment counter
						totalNumberTupledEnq++;
						
						//need to add tuple to batch send
						q.enqueue(net);
						
						//old send
						//this.getOutputQueue().enqueue(net, this);
						
						//queue too full?
						if(q.length() >= NumTupToProcessPerOp)
						{
//							System.out.println("Sending results to Eddy..."+q.length());
							EddyBatchTuple ebtOut =(EddyBatchTuple)q.dequeue();
							ebtOut.setNumberTuplesinBatch(q.length()+1);
							this.getOutputQueue().enqueue((XATTuple)ebtOut, this);						
							
							while(q.length()>0)
							{
								this.getOutputQueue().enqueue((XATTuple)q.dequeue(), this);						
							}
							q = null;
							
							//set batch queue to empty
							q = new EddyQueue();
							bFirstResult = true;
						}
					
						
					}	
					et = null;
					
				}
				else //result not valid
				{
					//Count all active processing tuples
					if (bCountActiveProcTups)
					{	
						if (PartitionProcTupCount.get(PartitionNum) != null)
							PartitionProcTupCount.put(PartitionNum,PartitionProcTupCount.get(PartitionNum)-1);
						else
							PartitionProcTupCount.put(PartitionNum,-1);
					}					
					
				}
			}	
			else if (bProbe)
			{					
				//System.out.println("Probe");

/**	CHUAN LEI Probe method is not defined in STeMPartitionData.java, hence leaving else block as dead code*/
//				ArrayList <XATTuple> results = STeMStreamData.Probe(ReceivedTuple,
//																	STeMDataIndexArrayPos,
//																	NumTupleIndex,
//																	PartitionNum);				
				ArrayList <XATTuple> results = null;
				
				if (results == null)
				{			
					ReceivedTuple = null;
					//Count all active processing tuples
					if (bCountActiveProcTups)
					{	
//						System.out.println("Before PartitionProcTupCount.get(i) = "+PartitionProcTupCount.get(PartitionNum));
						if (PartitionProcTupCount.get(PartitionNum) != null)
							PartitionProcTupCount.put(PartitionNum,PartitionProcTupCount.get(PartitionNum)-1);
						else
							PartitionProcTupCount.put(PartitionNum,-1);
//						System.out.println("After PartitionProcTupCount.get(i) = "+PartitionProcTupCount.get(PartitionNum));
					}
				}
				else
				{					
					//Count all active processing tuples
					if (bCountActiveProcTups)
					{	    
							if (PartitionProcTupCount.get(PartitionNum) != null)
								PartitionProcTupCount.put(PartitionNum,(PartitionProcTupCount.get(PartitionNum)+(results.size()-1)));
							else
								PartitionProcTupCount.put(PartitionNum,(results.size()-1));
					}
										
//					System.out.println("\t  " + results.size()+" results");
					//----------------------------------------------------------------------
					// EDDY OPERATOR ANSWER
					//----------------------------------------------------------------------					
					if (IsEddyOp)
					{
						//System.out.println("Probe results found = "+results.size());
						
						for(int k = 0; k < results.size(); k++)
						{
							if(bFirstResult)
							{
								//create Eddy Batch Tuple
								EddyBatchTuple netEbt = new EddyBatchTupleImp(et.getNumTuplesInAnswer(), NumOps, NumBits, et.getPartitionNumber());
								
//								EddyBatchTuple netEbt = new EddyBatchTupleImp(et.getNumTuplesInAnswer(),NumOps, et.getNumTuplesInAnswer(), et.getPartitionNumber(),NumJoinAttrib);
								netEbt.setAnswer(et.getAnswer());
								netEbt.addTuple(DataSTREAMId, results.get(k));
								
								netEbt.setDoneBitSet(dbs);
								netEbt.setReadyBitSet(rbs);
								
//								netEbt.SetCompletedJoinAttrib(CompletedJoinAttribs);
								
								//increment counter
								totalNumberTupledEnq++;
								
								//num tuples in batch
								//netEbt.setNumberTuplesinBatch(results.size());
								
//								netEbt.setstrOps(strOps);
								
								q.enqueue(netEbt); 
								bFirstResult = false;
								
							}
							else
							{	
								// create new eddy tuple for each intermediate answer
								net  = new EddyTupleImp(et.getNumTuplesInAnswer(),et.getPartitionNumber());
								net.setAnswer(et.getAnswer());
								net.addTuple(DataSTREAMId, results.get(k));
								
								//increment counter
								totalNumberTupledEnq++;
								
								//need to add tuple to batch send
								q.enqueue(net);
								
								//old send
								//this.getOutputQueue().enqueue(net, this);
								
								//queue too full?
								if(q.length() >= NumTupToProcessPerOp)
								{
//									System.out.println("Sending results to Eddy..."+q.length());
									EddyBatchTuple ebtOut =(EddyBatchTuple)q.dequeue();
									ebtOut.setNumberTuplesinBatch(q.length()+1);
									this.getOutputQueue().enqueue((XATTuple)ebtOut, this);						
									
									while(q.length()>0)
									{
										this.getOutputQueue().enqueue((XATTuple)q.dequeue(), this);						
									}
									q = null;
									
									//set batch queue to empty
									q = new EddyQueue();
									bFirstResult = true;
								}
								
							}	
						}//end iterating through results
						et = null;
					}					
				}//end result is not empty								
			}//end if probe
					
		}//end white (iterating through all dequeued tuples)		

		//any batched tuples need to be sent?
		if (IsEddyOp)
		{	
			// send out all old tuples in q
			if(q != null)
			{
				if(q.length()>0)
				{					
//					System.out.println("Sending results to Eddy..."+q.length());
					EddyBatchTuple ebt =(EddyBatchTuple)q.dequeue();
					ebt.setNumberTuplesinBatch(q.length()+1);
					this.getOutputQueue().enqueue((XATTuple)ebt, this);						
					while(q.length()>0)
					{
						this.getOutputQueue().enqueue((XATTuple)q.dequeue(), this);						
					}
				}	
			}
			q = null;
			
		}
		
		
		//send out active proc tup count
		if (bCountActiveProcTups)
		{	
		
			EddyActiveProcTuple eapt = new EddyActiveProcTupleImp(PartitionProcTupCount);
			this.getOutputQueue().enqueue((XATTuple)eapt, this);						
			
		}
		
//		System.out.println("Enqueued: " + totalNumberTupledEnq + " tuples.");
//		System.out.println("CountActiveProcTups");
//		if(bNoProbeUsed)
//			System.out.println("NoProbeUsed");
//		for (i=0; i <= 5; i++)
//		{
//			if (PartitionProcTupCount.get(i) != null)					
//			{
//				System.out.println(i+"="+PartitionProcTupCount.get(i));
//			}
//		}
//		System.out.println("---------------");
		
		
		endTime = System.currentTimeMillis();
		long duration = endTime - startTime; //milliseconds
		
		//WriteToFile(outFile, numProcessedTuples + "\t" + String.valueOf(duration), true);
		PartitionProcTupCount = null;
		//System.gc();
		return true;
	}

	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return false;
	}

	public boolean cleanAllAccumulatedTuples() 
	{
		return false;
	}

	public boolean matchStates(XATStreamOperator op) 
	{
		return false;
	}

	public boolean hasUnmatchedState() 
	{
		return false;
	}

	public boolean recomputeUnmatchedState() 
	{
		return false;
	}

	public int setStateID(int startID) 
	{
		return 0;
	}

	public String toToolTipString() 
	{
		return "STeMJoinProbeOperator";
	}

	public boolean finalizeOperation() 
	{
		return false;
	}

	public void setSTeMStreamData(STeMPartitionData sd)
	{
		STeMStreamData = sd;
	}
	
	public void SetIsEddyOp(boolean b)
	{
		IsEddyOp = b;
	}

	//STQM
	public void SetIsSelfTuning(boolean b)
	{
		IsSelfTuning = b;
	}
	
	//STQM
	public boolean GetIsSelfTuning()
	{
		return IsSelfTuning;
	}
	
	///////////////////////////////////////////
	//Info about probe
	public void SetProbeSTREAMId(int i)
	{
		ProbeSTREAMId = i;
	}

	public int getProbeSTREAMId()
	{
		return ProbeSTREAMId;
	}

	public void SetNumTupleIndex(String val) 
	{
		NumTupleIndex = new ArrayList();
		StringTokenizer st = new StringTokenizer(val);
	    while (st.hasMoreElements()) 
	    {
	    	NumTupleIndex.add(Integer.parseInt(st.nextElement().toString()));
	    }	
	}

	public ArrayList<Integer> getNumTupleIndex()
	{
		return NumTupleIndex;
	}

	///////////////////////////////////////////
	//Info about stored stem data
	
	public void SetDataSTREAMId(int i)
	{
		DataSTREAMId = i;
	}

	public int getDataSTREAMId()
	{
		return DataSTREAMId;
	}
/////////////////////////////////////////
	public void SetSTeMDataIndexArrayPos(int i)
	{
		STeMDataIndexArrayPos = i;
	}

	public int getSTeMDataIndexArrayPos()
	{
		return STeMDataIndexArrayPos;
	}
////////////////////////////////////////	
	public void SetNumSTeMIndex(String val) 
	{
		NumSTeMIndex = new ArrayList();
		StringTokenizer st = new StringTokenizer(val);
	    while (st.hasMoreElements()) 
	    {	    
	    	NumSTeMIndex.add(Integer.parseInt(st.nextElement().toString()));
	    }	
	}

	public ArrayList<Integer> getNumSTeMIndex()
	{
		return NumSTeMIndex;
	}
/////////////////////////////////////////////////
	public void setPreFilledPerStream(ArrayList<BitSet> p)
	{
		PreFilledPerStream = p;
	}	
	
	// table for look up operators used by QMesh
	public void setOperatorLookUp(OperatorLookUp z)
	{
		OpLkUp = z;
	}
//////////////////////////////////////////////////////	
	public void setQMeshOperatorID(int id) 
	{
		QMeshOperatorID = id;
	}
	
	public int getQMeshOperatorID() 
	{
		return QMeshOperatorID;
	}
///////////////////////////////////////////////////////	
	public void setSendOff(int so)
	{
		rusterSendOff = so;
	}	
	
	//used for experiments (otherwise comment it out)
	public static void WriteToFile(String fileLocation, String stringToWrite, boolean doNewLine)
	{
		try
		{
			//Create file 
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			if (doNewLine)
			{
				out.newLine();
			}
			//Close the output stream
			out.close();			    
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}
//////////////////////////////////////////////////	
	//STQM
	public long GetLastUpdateTime()
	{
		return lastUpdateTime;
	}

	//STQM
	public void SetLastUpdateTime(long newUpdateTime)
	{
		lastUpdateTime = newUpdateTime;
	}
/////////////////////////////////////////////////
	public void SetCompleteJoinAttribs(int val) 
	{
	    CompleteJoinAttrib = val;
	}

	public int getCompleteJoinAttribs()
	{
		return CompleteJoinAttrib;
	}

}
