/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.STeM;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import edu.wpi.cs.dsrg.eddy.datastructure.EddyQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
//Karen Works

//This structure stores the data the STEM searches on 

public class STeMData 
{

	int MAX_WINDOW_SIZE = 500;
	// maximum number of tuples to store
	// need to ask what value should be??

	//links to tuples that are referenced
	private ArrayList<HashMap<String, EddyQueue>> StoredTuples;

	// number of indexes this table is joined against
	int intNumIndexes; 
	
	// Array number of the data source
	int intSTREAMId; 
	
	EddyQueue TupleList;
	//Array of positions in tuple that need to be indexed on
	ArrayList<ArrayList<Integer>> HashKeyIndex;  
	// Current number of stored tuples
	int intCurrNumberofStoredTuples = 0; 

	public void init(Integer STREAMId,ArrayList<ArrayList<Integer>> HashKeys)
	{
		// set up required variables
		intNumIndexes = HashKeys.size();
		StoredTuples = new ArrayList<HashMap<String, EddyQueue>>(intNumIndexes);
		for (int i =0;i<intNumIndexes;i++)
		{
			StoredTuples.add(i,new HashMap<String, EddyQueue>());
		}
		
		intSTREAMId = STREAMId;
		HashKeyIndex = HashKeys;
		TupleList = new EddyQueue();
	}
	
	public void Insert(XATTuple tuple)
	{
		// if adding this tuple will make the total#
		// > than the window purge the oldest tuple
		if ((intCurrNumberofStoredTuples + 1) > MAX_WINDOW_SIZE)
		{	
			PurgeTuple();
		}
		
		// add the tuple to the list
		TupleList.enqueue(tuple);
		
		// for each index
		for (int i = 0;i < intNumIndexes;i++)
		{
			// compute the key for this tuple
			String key = "";
			for (int y = 0;y<HashKeyIndex.get(i).size(); y++)
				key += tuple.getCell(HashKeyIndex.get(i).get(y)).convertToString()+"---";

			//for testing only			
			//System.out.println("adding  "+key+" to "+intSTREAMId);
			
			// store the tuple in the appropriate queue
			EddyQueue Eq = StoredTuples.get(i).get(key);
			if(Eq != null)
			{
				Eq.enqueue(tuple);	
			}
			else
			{
				EddyQueue<Object> eq = new EddyQueue<Object>();
				
				eq.enqueue(tuple);
				StoredTuples.get(i).put(key.toString(), eq);
			}
		}
		intCurrNumberofStoredTuples++;
	}
	
	private void PurgeTuple()
	{
		//while the number of tuples+1 is > the window
		while ((intCurrNumberofStoredTuples +1) > MAX_WINDOW_SIZE)
		{	
			// find the oldest tuple & remove it
			XATTuple DeleteTup = (XATTuple)TupleList.dequeue();
			
			// remove this tuple from all the index queues ad well
			for (int i =0;i<intNumIndexes;i++)
			{
				String key = "";
					for (int y = 0;y<HashKeyIndex.get(i).size(); y++)
						key += DeleteTup.getCell(HashKeyIndex.get(i).get(y)).convertToString()+"---";
					 	
				StoredTuples.get(i).remove(key);				
			}		
			intCurrNumberofStoredTuples--;
		}		
	}

	// this function returns the STREAM is for this STeM Data
	public int getSTREAMId()
	{
		return intSTREAMId;
	}

	// this function probes the STEM data and returns results 
	public ArrayList <XATTuple> Probe(XATTuple tup, int NumSTeMIndex, ArrayList<Integer> NumTupleIndex)
	{
		//int probeWindowSize = 100;
		int numTuplesIterated = 0;
		
		// validate indexes
		if ((NumSTeMIndex > intNumIndexes) || (NumSTeMIndex <0 ))
		{	
			System.out.println("Error STeM Probe - NumSTeMIndex out of range");
			return null;
		}
		
		ArrayList <XATTuple> results = new ArrayList<XATTuple>();
		
		// compute key of probe tuple
		String key = "";
		for (int y = 0; y < NumTupleIndex.size(); y++)
		{		
			int tupleIndex = NumTupleIndex.get(y);
			//System.out.println("tupleIndex = " + tupleIndex);
			
			if (tup == null)
			{
				System.out.println("tuple is NULL");				
			}
			
			Object tupleValue = tup.getCell(tupleIndex);
			
			key += tupleValue +"---";
			//System.out.println("probe key = " + key);
		}
		//for testing only			
		//System.out.println("looking for  "+key+" From "+tup.getStreamID()+"in "+intSTREAMId);

//		for testing only		
//		int numTuplesMatched = 0;
//		int numTuplesIteratedNonButNotIncluded = 0;
//		int numTuplesMadeIntoResults = 0;
		
		// find matches to probe tuple key
		EddyQueue Eq = StoredTuples.get(NumSTeMIndex).get(key.toString());		
				
		if(Eq != null) // if = null no tuples have been added yet
		{
			//System.out.println("Probing STEMDATA of size [" + Eq.size() + "]");
			//System.out.println("results found: "+Eq.size());
			
			//System.out.println("probe key: "+ key);			
			//System.out.println("probe tuple ts: "+tup.getCell(tup.getLength() - 1));
			
			// loop through tuple(s) found
			boolean bCont = true;
			
			for (Iterator <XATTuple>it = Eq.iterator(); it.hasNext() && bCont;) 
			{
				//below is a count-window limit to bound how many tuples to join with
				numTuplesIterated++;
				
				/*if (numTuplesIterated >= probeWindowSize)
				{
					break;
				}
				*/
				
		        XATTuple TupFound = it.next();
		        // if the tuples meet the window req return them

		        Object strTupleCounter1 = tup.getCell(tup.getLength() - 1);    		
				
		        if (strTupleCounter1.toString().equals(""))
				{
					results = null;
					break;
				}
				
	    		long tupleCounter1 = Long.parseLong(strTupleCounter1.toString());
	    		
	    		Object strTupleCounter2 = TupFound.getCell(TupFound.getLength() - 1);
	    		long tupleCounter2 = Long.parseLong(strTupleCounter2.toString());
	    		//rimma end -- 500 tuple window join
	    			    	
	    		//System.out.println("\t\tmatching key ts: " + tupleCounter2);
	    		
	    		//if (TupFound.getMinTimeStamp() <= tup.getMinTimeStamp())
	    		long diff = Math.abs(tupleCounter1 - tupleCounter2);
	    		if (diff <= 200)// && (TupFound.getMinTimeStamp() <= tup.getMinTimeStamp()))
		        {
		   			//for testing only			
//		   			System.out.println("Answer found!!");
		    	   
		        	results.add(TupFound);
//		        	numTuplesMadeIntoResults++;
		        }
		        else if (tupleCounter2 > tupleCounter1)
		        {
		        	
		   			bCont = false;
//		        	numTuplesIteratedNonButNotIncluded++;
		        }
		    }			
		}
		else
		{
			results = null;
		}
		
		Eq = null;
		
				
		if (results != null)
		{	
			if (results.size() > 0)
			{
				//System.out.println("results returned: " + results.size());
				return results;
			}
			else
			{
				//System.out.println("results returned: 0");
				return null;
			}
		}	
		else
			return null;
	}
	
	public void SetMaxWindowSize(int val)
	{
		MAX_WINDOW_SIZE = val;
	}
}
