/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.SyntheticDataGenerator;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Helper;

public class CombineTwoAttributeDomains 
{
	//-----------------------------------------------------------------------------
	
	public static void main(String args[])
	{
		String inputDirectory = "C:\\RIMMA\\Research\\QueryMesh\\distributions\\Poisson\\";
		String inputFile1 = inputDirectory + "StreamA.txt";
		String inputFile2 = inputDirectory + "StreamB.txt";
		
		String outputDirectory = inputDirectory;
		
		CombineTwoAttributeDomains ct = new CombineTwoAttributeDomains();
		
		Vector joinAttribueDomain1 = new Vector();
		Vector nonJoinAttribueDomain1 = new Vector();
		
		ct.processMultiValueFile(inputFile1, joinAttribueDomain1, nonJoinAttribueDomain1);
		
		Vector joinAttribueDomain2 = new Vector();
		ct.processSingleValueFile(inputFile2, joinAttribueDomain2);
		
		String currStreamContent = ct.combineValues(joinAttribueDomain1, joinAttribueDomain2, nonJoinAttribueDomain1);
		
		String fileToWriteTo = outputDirectory + "DATA.txt";
		
	    //write it out to file
	    Helper.WriteToFile(fileToWriteTo, currStreamContent);
	    
		System.out.println("Done.");
	}
	
	//-----------------------------------------------------------------------------
	
	public String combineValues(Vector nums1, Vector nums2, Vector others)
	{
		String combinedContent = "";
		for(int i = 0; i < nums1.size(); i++)
		{
			combinedContent += nums1.get(i).toString() + " " + nums2.get(i).toString() + " " + others.get(i);
		}
		return combinedContent;
	}
	
	//-----------------------------------------------------------------------------
	
	public void processMultiValueFile(String inputFile, Vector nums, Vector others)
	{	
	    String line;
	    int num = 0;

	    File file = new File(inputFile);
	    FileInputStream fis = null;
	    BufferedInputStream bis = null;
	    DataInputStream dis = null;
	    		       
	    try 
	    {	  	
	    	fis = new FileInputStream(file);		   
	    
	    	BufferedReader in = new BufferedReader(new InputStreamReader(fis));
	    	while ((line = in.readLine()) != null) 	            
	    	{	         
	    		if (!line.equalsIgnoreCase(""))	         
	    		{
	    			ArrayList arrCurrValues = new ArrayList();
	    			String [] vals = line.split(" ");
	    			String other = "";
	    			for (int p = 0; p < vals.length; p++)
	    			{
	    				if (p == 0)
	    				{
	    					num = Integer.parseInt(vals[p]);
	    				}
	    				else
	    				{
	    					other += vals[p] + " ";
	    				}	    				
	    			}
	    			
	    			other = other.substring(0, other.length() - 2) + System.getProperty("line.separator");
	    			nums.add(num);	  
	    			others.add(other);
	    		}	            
	    	}	        
	    } 
	    catch (IOException e) 	    
	    {	           
	    	System.err.println(e);      
	    	return;	        
	    }

	    return;
	}
	
	//-----------------------------------------------------------------------------
	
	public void processSingleValueFile(String inputFile, Vector nums)
	{	    
		int num = 0;
		String line;
		
	    File file = new File(inputFile);
	    FileInputStream fis = null;
	    BufferedInputStream bis = null;
	    DataInputStream dis = null;
	    		       
	    try 
	    {	  	
	    	fis = new FileInputStream(file);		   
	    
	    	BufferedReader in = new BufferedReader(new InputStreamReader(fis));
	    	while ((line = in.readLine()) != null) 	            
	    	{	         
	    		if (!line.equalsIgnoreCase(""))	         
	    		{
	    			num = Integer.parseInt(line);
	    			nums.add(num);	            	
	    		}	            
	    	}	        
	    } 
	    catch (IOException e) 	    
	    {	           
	    	System.err.println(e);      
	    	return;	        
	    }

	    return;
	}	

	//-----------------------------------------------------------------------------
	
}
