/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.SyntheticDataGenerator;


import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Random;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.SimpleOperatorInputSpecification;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.SimpleOperatorSpecification;
//Rimma V. Nehme

//This is the class creates the synthetic data pre the user defined xml
public class ConsumingDistributionStreamSyntheticDataGenerator implements java.io.Serializable 
{	
	
	//------------------------------------------------------------------------------------------
	
	//------------------------------------------------------------------------------------------
	
	public static void main(String args[])
	{
		String inputDirectory = "C:\\RIMMA\\Research\\QueryMesh\\distributions\\Poisson\\";
		String distributionInputFile = inputDirectory + "StreamA.txt";
		String outputDirectory = inputDirectory;
		
		//generate the files
		ConsumingDistributionStreamSyntheticDataGenerator sdg = new ConsumingDistributionStreamSyntheticDataGenerator(distributionInputFile, outputDirectory);
		
		System.out.println("Done generating data.");
	}
	
	//------------------------------------------------------------------------------------------
	
	public ConsumingDistributionStreamSyntheticDataGenerator(String distributionInputFile, String outputDirectory)
	{
		int counter = 0;
				
		//parse the xml file and get the dom object
		String currStreamContent = loadFile(distributionInputFile);

		String fileToWriteTo = outputDirectory + "DATA.txt";
		
	    //write it out to file
	    Helper.WriteToFile(fileToWriteTo, currStreamContent);
	}
	
	//------------------------------------------------------------------------------------------
	
	public String loadFile(String distributionInputFile)
	{	    
	    String line;
	    Vector nums = new Vector();
	    Vector numsUnsorted = new Vector();
	    double num, sum = 0.0;
	    double mean = 0.0;
	    double average_deviation = 0.0;
	    double standard_deviation = 0.0;
	    double variance = 0.0;
	    double skew = 0.0;
	    double kurtosis = 0.0;
	    double median = 0.0;
	    double min = 0.0;
	    double max = 0.0;
	    double deviation = 0.0;
	    int i, n, mid = 0;

	    File file = new File(distributionInputFile);
	    FileInputStream fis = null;
	    BufferedInputStream bis = null;
	    DataInputStream dis = null;
	    		       
	    try 
	    {	  	
	    	fis = new FileInputStream(file);		   
	    
	    	BufferedReader in = new BufferedReader(new InputStreamReader(fis));
	    	while ((line = in.readLine()) != null) 	            
	    	{	         
	    		if (!line.equalsIgnoreCase(""))	         
	    		{
	    			num = Double.parseDouble(line);
	    			sum += num;
	    			nums.add(new Double(num));	            	
	    		}	            
	    	}	        
	    } 
	    catch (IOException e) 	    
	    {	           
	    	System.err.println(e);      
	    	return "";	        
	    }

	    n = nums.size();
	    mean = sum / n;
	    for (i = 0; i < n; i++) 
	    {
	        deviation = ((Double)nums.get(i)).doubleValue() - mean;
	        average_deviation += Math.abs(deviation);
	        variance += Math.pow(deviation,2);
	        skew += Math.pow(deviation,3);
	        kurtosis += Math.pow(deviation,4);
	    }
	    
	    average_deviation /= n;
	    variance /= (n - 1);
	    standard_deviation = Math.sqrt(variance);
	    if (variance != 0.0) 
	    {
	        skew /= (n * variance * standard_deviation);
	        kurtosis = kurtosis/(n * variance * variance) - 3.0;
	    }
	    
	    numsUnsorted = (Vector)nums.clone(); //keep the copy of the unsorted
	    Collections.sort(nums);

	    mid = (n/2);
	    median = (n % 2 != 0) ?
	        ((Double)nums.get(mid)).doubleValue() :
	        (((Double)nums.get(mid)).doubleValue() +
	         ((Double)nums.get(mid-1)).doubleValue())/2;
	    	    
	    min = (Double)nums.get(0);
	    max = (Double)nums.get(n - 1);
	        
	    NumberFormat nf = NumberFormat.getInstance();
	    nf.setMaximumFractionDigits(13);
	    nf.setGroupingUsed(false);
	    nf.setMaximumFractionDigits(6);
	    nf.setMinimumFractionDigits(6);

	    
	    System.out.println("n:                  " + n);
	    System.out.println("min:                " + nf.format(min));
	    System.out.println("max:                " + nf.format(max));
	    System.out.println("median:             " + nf.format(median));
	    System.out.println("mean:               " + nf.format(mean));
	    System.out.println("average_deviation:  " + nf.format(average_deviation));
	    System.out.println("standard_deviation: " + nf.format(standard_deviation));
	    System.out.println("variance:           " + nf.format(variance));
	    System.out.println("skew:               " + nf.format(skew));
	    System.out.println("kurtosis:           " + nf.format(kurtosis));
	
	    //create stream content
	    String currStreamContent = createOutputString(numsUnsorted, min, max);
	    
	    return currStreamContent;
	}	

	//------------------------------------------------------------------------------------------
	
	private String createOutputString(Vector inputValues, double min, double max)
	{
		//max = 100;
		int col1NumDistinctValues = 5;		
		String [] col1CorrelatedValues = {"00000", "11111", "22222", "33333", "44444", "55555"};
		
		int col2NumDistinctValues = 10;
		String [] col2CorrelatedValues = {"00000", "aaaaa", "bbbbb", "ccccc", "ddddd", "eeeee",
										  "fffff", "ggggg", "hhhhh", "iiiii", "jjjjj"};
		
		String currStreamContent = ""; 
		for (int i = 0; i < inputValues.size(); i++)
		{
			double currValue = (Double)inputValues.get(i);			
			int col1Index = (int)(((currValue - min)/(max - min)) * col1NumDistinctValues);			
			if (col1Index > col1NumDistinctValues)
			{
				col1Index = col1NumDistinctValues;
			}
			String col1Value = col1CorrelatedValues[col1Index];
			
			int col2Index = (int)(((currValue - min)/(max - min)) * col2NumDistinctValues);
			if (col2Index > col2NumDistinctValues)
			{
				col2Index = col2NumDistinctValues;
			}
			String col2Value = col2CorrelatedValues[col2Index];
			
			int currValueInt = (int)currValue;
			currStreamContent += currValueInt + " " + col1Value + " " + col2Value + System.getProperty("line.separator");
		}
		return currStreamContent;
	}

	//------------------------------------------------------------------------------------------
}
