/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.SyntheticDataGenerator;


import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.SimpleOperatorInputSpecification;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.SimpleOperatorSpecification;
//Karen Works
//Nov 2007

//This is the class creates the synthetic data pre the user defined xml
public class SyntheticDataGenerator implements java.io.Serializable 
{	
	int WINDOW_SIZE = 100;
	
	//used to parse the stream xml file
	Document dom;
	
	//number of random Nums
	int NumRandomNum = 999999999; 
	
	//number of Streams
	int NumStreams; 

	//number of columns per stream
	ArrayList<Integer> NumCols;

	//tuple count per stream
	ArrayList<Integer> tupleCountPerStream;

	//use random numbers for filler per stream
	ArrayList<Boolean> useRandomNumbersPerStream;
	
	//partitions in streams		
	ArrayList<HashMap<Integer,ArrayList<StreamPartitionColumn>>> StreamPartitions;

    //Correlation of data between columns in a single stream
	ArrayList<HashMap<Integer,ArrayList<Correlation>>> ColsCorrelatedInStreams;

	
	public static void main(String args[])
	{
		if (args.length != 2)
		{
			String usage = "Usage:\n" +
			       		   "---xml input file (w/ generator parameters)\n" + 
			       		   "---data output directory (directory where generated data files will be placed).";
			
			System.out.println(usage);
			return;
		}
		
		//get the path of the xml file
		String xmlInputFile = args[0];		
		String outputDirectory = args[1];
		
		
		//generate the files
		SyntheticDataGenerator sdg = new SyntheticDataGenerator(xmlInputFile, outputDirectory);
		
		System.out.println("Done generating data.");
	}
	
	public SyntheticDataGenerator(String xmlInputFile, String outputDirectory)
	{
		int counter = 0;
		// initalize
		NumStreams = 0;
		NumCols = new ArrayList<Integer>();
		tupleCountPerStream = new ArrayList<Integer>();
		useRandomNumbersPerStream = new ArrayList<Boolean>();
		
		StreamPartitions = new ArrayList<HashMap<Integer,ArrayList<StreamPartitionColumn>>>();
		ColsCorrelatedInStreams = new ArrayList<HashMap<Integer,ArrayList<Correlation>>>();
		Correlation cor;
		
		//parse the xml file and get the dom object
		loadFile(xmlInputFile);

			    
		//seed for random number generator
		long seed = 985679429;
		
		//file path
		//String path = "C:\\Research\\QueryMesh\\testData\\";
		//String path = "D:\\Rimma\\Research\\QueryMesh\\data\\";
		String path = outputDirectory;
		
		//output file
		FileOutputStream out; 
		PrintStream p; 
		
		Random generator;
		

	    try
	    {
	      //create random numbr generator	
		  generator = new Random();
	      generator.setSeed(seed);
	      
	      //for each stream
	      for (int ns = 0;ns < NumStreams; ns++)
	      {	 
	    	 //reset counter for each tuple
	    	 counter = 100000;
	    	 
	    	 //create a new file 
	         //out = new FileOutputStream(path + "Data" + (ns) + ".txt",false);
	    	 
	         //append
	         out = new FileOutputStream(path + "Data" + (ns) + ".txt",true);
	         
	         p = new PrintStream( out );

	         //cols preset to equal values
	         HashMap<Integer,ArrayList<StreamPartitionColumn>> strmPartitions = StreamPartitions.get(ns);
	         
	         //Correlation of data between columns in a single stream
	         HashMap<Integer,ArrayList<Correlation>> cdbs = ColsCorrelatedInStreams.get(ns);
	         
	         //generate the output
	         ArrayList<String> output = new ArrayList<String>();
	         
	         //generate a set of WINDOW_SIZE possible values
	         for (int numTuples = 0; numTuples < WINDOW_SIZE; numTuples++)
	         {
	        	 String str = "";
	        	 int val = 0;
	        	 
	        	 //array to keep track of values generated
		         ArrayList<Integer> values = new ArrayList<Integer>();
		         
		         //for each column
	        	 for (int c = 0; c < NumCols.get(ns); c++)
	        	 {
	        		// does a correlation exist? 
			        if ( cdbs.get(c+1) != null)
		        	{	
			        	boolean bFound=false;
			        	//find the relation
			        	for (int numCor=0; numCor<cdbs.get(c+1).size()&&!bFound; numCor++)
			        	{	
				          cor = cdbs.get(c+1).get(numCor); 
  			        	  //is the column you are related to before it 
			        	  if( (c+1) > cor.ColTo)
			        	  {	
			        	     if (cor.partitionValue == values.get(cor.ColTo-1))
			        	     {
				        		val =cor.value;
			        		    cor.numTimesAppears--;
			        		    bFound=true;
			        	     }
			        	   }
			        	  else if (strmPartitions.get(cor.ColTo) != null)
			        	  {	  
				        	 if (strmPartitions.get(cor.ColTo).get(0) != null)
				        	  {	  
			        	        if(cor.partitionValue == strmPartitions.get(cor.ColTo).get(0).Value)
			        	        {
				        		  val =cor.value;
			        		      cor.numTimesAppears--;
			        		      bFound=true;			        		
				        	    }
				        	  }
			        	  }
			        	  //if this is the last time the correlation should appear set NumTimesAppears to 0
			        	  if (bFound && (cor.numTimesAppears == 0))
			        		  cdbs.get(c+1).remove(numCor);
			        	}
			        	
			        	//if you cannot find the correlation set to the random num or col
			        	if (!bFound)
			        	{	
			        		if (useRandomNumbersPerStream.get(ns))
			        			val =generator.nextInt(NumRandomNum);
			        		else
			        			val =c+1;
			        	}	
			        		
		        	}	
			        // is this col a partition?
			        else if ( strmPartitions.get(c+1) != null)
			        {	
			        	int n = strmPartitions.get(c+1).size();
	        			int col = generator.nextInt(n);
			        	
			        	//set val = to the partition
			        	StreamPartitionColumn sp = strmPartitions.get(c+1).get(col);			        	
		        		val = sp.Value;
		        		sp.Per--;
		        		
		        		//if it is the last time for this partition remove it from the list
		        		if (sp.Per==0)
		        			strmPartitions.get(c+1).remove(col);
		        		if (strmPartitions.get(c+1).size()==0)
		        			strmPartitions.remove(c+1);
		        		
			        }	
		        	//if you cannot find the correlation set to the random num or col
	        		else
		        	{	
		        		if (useRandomNumbersPerStream.get(ns))
		        			val =generator.nextInt(NumRandomNum);
		        		else
		        			val =c+1;
		        	}	

			       //add values to array 	
			        values.add(val);
			        
	        		str += val+"|"; 
	        	 }	        	
	        	 output.add(str);	        	 
	         }
	         
	         	         
	         //place output in file.
	        for (int numTuplesinFile = 0;numTuplesinFile<tupleCountPerStream.get(ns);numTuplesinFile=numTuplesinFile+WINDOW_SIZE)
	        {
	        	ArrayList<String> outputTemp = new ArrayList<String>();
	        	outputTemp = (ArrayList<String>)output.clone();
	        	
	        	// for each set of WINDOW_SIZE tuples
	        	// randomly pick from the preset arrya
	           for (int numTuples = 0;numTuples<WINDOW_SIZE;numTuples++)
	           {
	        	 //randomly pick a value  
	        	 int nextPos;
	        	 if (numTuples!=(WINDOW_SIZE - 1))
	        	 {	 
	        		 nextPos = generator.nextInt((WINDOW_SIZE - 1) - numTuples);
	        		 while(outputTemp.get(nextPos) == null)
	        		 {
	        			 nextPos = generator.nextInt((WINDOW_SIZE - 1) - numTuples);
	        		 
	        		 }
	        	 }
	        	 else
	        		 nextPos = 0;
	        	 
	        	 counter++;
	        	 
	        	 //add to output file
//	        	 System.out.println("counter="+counter);
	        	 p.println(outputTemp.get(nextPos) + counter);
	        	 outputTemp.remove(nextPos);
	           } 
	         }
	  
	    	  
	    	  System.gc();	    	  
		      
	    	  p.close();
		      out.close();  
	       }

	  	  	generator = null; 
	    }
	    catch (Exception e)
	    {
	          e.printStackTrace();
	    }

	}
	
	public void loadFile(String xmlInputFile)
	{
		//parse the xml file and get the dom object
		System.out.println("Parsing optimizer input file...");
		parseXmlFile(xmlInputFile);
		
		//parse the documents
		System.out.println("Extracting xml data from DOM...");
		parseDocument();	
	}
/////////////////////////////////////////////////////////////////
	// parse xml file
	private void parseXmlFile(String xmlInputFile)
	{
		//get the factory
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		
		try {
			
			//Using factory get an instance of document builder
			DocumentBuilder db = dbf.newDocumentBuilder();
			
			//parse using builder to get DOM representation of the XML file
			dom = db.parse(xmlInputFile);
			

		}
		catch(ParserConfigurationException pce) 
		{
			pce.printStackTrace();
		}
		catch(SAXException se) 
		{
			se.printStackTrace();
		}
		catch(IOException ioe) 
		{
			ioe.printStackTrace();
		}
	}

/////////////////////////////////////////////////////////////////
// parse Document
	private void parseDocument()
	{
		//get the root elememt
		Element docEle = dom.getDocumentElement();
		
		//----------------------------------------------------------------------
		// StreamS
		//----------------------------------------------------------------------		
		NodeList nl = docEle.getElementsByTagName("Stream");
		if(nl != null && nl.getLength() > 0) 
		{
			for(int i = 0 ; i < nl.getLength();i++) 
			{			
				//get the opInput element
				Element opInput = (Element)nl.item(i);
				
				//get id and attribute of the Input object
				int StreamId = Integer.parseInt(opInput.getAttribute("id"));
				int StreamNumCols = Integer.parseInt(opInput.getAttribute("numColumns"));
				int StreamNumTuplesPerFile = Integer.parseInt(opInput.getAttribute("numTuplesPerFile"));
				boolean UseRandomNumbers= Boolean.parseBoolean(opInput.getAttribute("useRandomNumbers"));
				
				
				NumStreams++;
				NumCols.add(StreamNumCols);
				tupleCountPerStream.add(StreamNumTuplesPerFile);
		    	StreamPartitions.add(new HashMap<Integer,ArrayList<StreamPartitionColumn>>());
		    	ColsCorrelatedInStreams.add(new HashMap<Integer,ArrayList<Correlation>>());
		    	useRandomNumbersPerStream.add(UseRandomNumbers);
		    	
				//partitions
				NodeList pl = opInput.getElementsByTagName("Partition");
				if(pl != null && pl.getLength() > 0) 
				{
					for(int j = 0 ; j < pl.getLength();j++)
					{				
						//get the opInput element
						Element opInputPartition = (Element)pl.item(j);
						
						int PartitionCols = Integer.parseInt(opInputPartition.getAttribute("column"));
						int PartitionPercent = Integer.parseInt(opInputPartition.getAttribute("percent"));
						int PartitionValue = Integer.parseInt(opInputPartition.getAttribute("value"));
						
						if (StreamPartitions.get(StreamId).get(PartitionCols)==null)
						  StreamPartitions.get(StreamId).put(PartitionCols, new ArrayList<StreamPartitionColumn>());
						
						StreamPartitionColumn spc = new StreamPartitionColumn(StreamId,PartitionCols,PartitionPercent,PartitionValue);
						StreamPartitions.get(StreamId).get(PartitionCols).add(spc);
					}
				}	

				//Correlation
				NodeList cl = opInput.getElementsByTagName("Correlation");
				if(cl != null && cl.getLength() > 0) 
				{
					for(int j = 0 ; j < cl.getLength();j++)
					{				
						//get the opInput element
						Element opInputCorrelation = (Element)cl.item(j);
						
						int sourceColumn = Integer.parseInt(opInputCorrelation.getAttribute("sourceColumn"));
						int sourceColumnPercent = Integer.parseInt(opInputCorrelation.getAttribute("sourceColumnPercent"));
						int sourceColumnValue = Integer.parseInt(opInputCorrelation.getAttribute("sourceColumnValue"));
						int destinationColumn = Integer.parseInt(opInputCorrelation.getAttribute("destinationColumn"));
						int destinationColumnValue = Integer.parseInt(opInputCorrelation.getAttribute("destinationColumnValue"));
						

						if (ColsCorrelatedInStreams.get(StreamId).get(sourceColumn)==null)
							ColsCorrelatedInStreams.get(StreamId).put(sourceColumn, new ArrayList<Correlation>());
						
						//get # of times value appears from the partition
						StreamPartitionColumn spc = null;
						boolean bNotFound = true;
						for(int h=0;h<StreamPartitions.get(StreamId).get(destinationColumn).size()&&bNotFound;h++)
						{	
							spc = StreamPartitions.get(StreamId).get(destinationColumn).get(h);
							if (spc.Value==destinationColumnValue)
								bNotFound = false;
						}
						
						int numTimes = (int)(spc.Per*(sourceColumnPercent*1.0)/100.0);
						Correlation cor = new Correlation(sourceColumn,destinationColumn,numTimes,sourceColumnValue,destinationColumnValue); 
						ColsCorrelatedInStreams.get(StreamId).get(sourceColumn).add(cor);
					}
				}	

			}
		}
		
	}
	
}
