/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.eddy.QueryPlan;

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNodeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDagImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATMemoryQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPersistentQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.Source;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.SourceImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.ProbabilisticMultiStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.ProbabilisticMultiStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.WBinaryJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.WBinaryJoinOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNodeImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.ConfigFileParser;
import edu.wpi.cs.dsrg.STeM.STeMData;
import edu.wpi.cs.dsrg.eddy.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator.QueryPlanGeneratorException;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator.QueryPlanGeneratorImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.client.*;
import java.util.ArrayList;

public class EddyTestQueryPlan extends QueryPlanGeneratorImp{
		public int NumStreams = 0;
		
		private ArrayList <Integer> STeMStreamId;
		private ArrayList <Integer> STeMNumIndexes;
		private ArrayList <Integer[]> STeMHashKeys;
		
		/**
		 * TestQueryPlan constructor comment.
		 */
		public EddyTestQueryPlan() {
			super();
			STeMStreamId = new ArrayList();
			STeMNumIndexes = new ArrayList();
			STeMHashKeys = new ArrayList();
		}
		/**
		 * Makes a single query plan.  Each QueryPlanGenerator implements
		 * this method in order to generate a query plan in their own manner
		 * 
		 * @return XATTree
		 * @exception QueryPlanGeneratorException 
		 */
		public XATTree makeQueryPlan() throws QueryPlanGeneratorException {
			return makeTestSimpleQuery();
		}		
		
		/**
		 * Insert the method's description here.
		 * Creation date: (2/1/2003 1:24:34 PM)
		 */
// Assumptions on DAG Creation
//KEW: Assume first input queues are STREAMS
//KEW: Assume last output queue  is the result queue
//KEW: Assumption is that STREAM Ids are set 1 to ....
		
//KEW: Assumption STeMStreamId is set to the an array of all the STREAMids that are members of a STEM
//KEW: Assumption STeMNumIndexes is set to the number of indexes on this STEM
//KEW: Assumption STeMHashKeys is set to the array of hash keys on this STEM

		
//KEW: Assumption Creation of a STeMOperator requires the set up of the
//NumSTeMIndex	Index that the STEM data is hashed on
//NumTupleIndex Index that the tuple is going to compare to
// Link to the data storage STeMData - this is set up in the execution controler

		public XATTree makeTestSimpleQuery() {
			
			//Need to set STREAM Id for each operator
			//one stream -> Eddy -> output
			//Assumption is that STREAM Ids are set 1 to ....
			NumStreams = 1;
			
//			XATDag dag = new XATDagImp();
			
//		    DagNode node1 = new DagNodeImp();
//		    EddyOperatorImp eddyop = new EddyOperatorImp(node1);
		    
//		    node1.setOperator(eddyop); 

		    //Stream A
		    
//		    DagNode sourceNode1 = new DagNodeImp();
//		    Source sourceOp1 = new SourceImp(sourceNode1);
//		    sourceOp1.setName("A");
//		    sourceNode1.setOperator(sourceOp1);
//		    sourceNode1.setParent(node1);
//		    node1.addOpChild(sourceNode1);
	
		   
		    /*
		    XATQueue outputQ = new XATMemoryQueueImp(eddyop);
		    Vector cols = new Vector();
		    cols.add("Col1");
		    XATQueueSchema schema = new XATQueueSchema(cols);
		    outputQ.setSchema(schema);

		    try {
		        eddyop.setOutputQueue(outputQ);
		        outputQ.register(this);
		    } catch (XATQueueException e) {
		        System.out.println(e);
		    }
*/
		    //eddyop.setDebug(false);
//		    eddyop.setDebug(true);
//		    dag.setRoot(node1);

//		    return dag;
		 
		    
			XATDag tree1 = new XATDagImp();
			DagNode A = new DagNodeImp();
			
			tree1.setRoot(A);
			
			//create dummy operators for each node
			EddyOperatorImp Aop = new EddyOperatorImp(A);

			return tree1; 		    
		}
		
		public ArrayList <Integer> getSTeMStreamId(){
			return STeMStreamId;
		}

		public ArrayList <Integer> getSTeMNumIndexes(){
			return STeMNumIndexes;
		}

		public ArrayList <Integer[]> getSTeMHashKeys(){
			return STeMHashKeys;
		}

		
}
