/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.eddy.datastructure;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Calendar;
import java.util.Date;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.eddy.operator.CBROperatorProfileEntry;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;

//Karen Works

//This structure stores the information associated with 
//an eddy tuple and eddy tuple operations.

public class EddyTupleImp extends XATTupleImp implements EddyTuple, Serializable
{
	// intermediate tuple that will be the answer
	private XATTuple[] answer = null;
	
	// ready bits & done bits
	private BitStructure ReadyBits;
	private BitStructure DoneBits;
	
	// number of tuples in answer
	private int NumTuplesInAnswer;
	
	// creation date
	private Date Creation;
	

	
	public EddyTupleImp(XATTuple t,int NumOps,int NumStreams,int TuplePos) 
	{ //throws XATTimeStampException {
		super();
		
		if (t != null) 
		{
			// set up answer
			answer = new XATTuple[NumStreams];
			answer[TuplePos] = t;         
			
			// set time
			this.setMaxTimeStamp(t.getMaxTimeStamp());
			this.setMinTimeStamp(t.getMinTimeStamp());
			
			NumTuplesInAnswer = NumStreams;
			
			// create ready & done bits
			ReadyBits = new BitStructure(NumOps);
			DoneBits = new BitStructure(NumOps);
			
			//set creation time
			Creation = Calendar.getInstance().getTime();
			
		} else {
			//hello
			System.out.println("tuple is null");
		}
	}

	public EddyTupleImp(XATTuple t,BitSet bs,int NumOps,int NumStreams,int TuplePos) 
	{
		//throws XATTimeStampException {
		super();
		if (t != null) 
		{
			answer = new XATTuple[NumStreams];
			answer[TuplePos] = t;         

			NumTuplesInAnswer = NumStreams;
			
			this.setMaxTimeStamp(t.getMaxTimeStamp());
			this.setMinTimeStamp(t.getMinTimeStamp());
			
			// create ready & done bits
			ReadyBits= new BitStructure();
			ReadyBits.CreateCopyFrom(bs);
			DoneBits = new BitStructure(NumOps);
			
			//set creation time
			Creation = Calendar.getInstance().getTime();			
			
		} else {
			//hello
			System.out.println("tuple is null");
		}
	}

	public EddyTupleImp(int NumStreams, int NumOps) 
	{ 
		//throws XATTimeStampException {
		super();
		answer = new XATTuple[NumStreams];

		NumTuplesInAnswer = NumStreams;
		
		// create ready & done bits
		ReadyBits= new BitStructure(NumOps);		
		DoneBits = new BitStructure(NumOps);
		
		//set creation time
		Creation = Calendar.getInstance().getTime();			
	}
	
	
	// returns the number of donebits set to complete
	public int NumDoneBitsComplete(){
		if (DoneBits != null)
		  return DoneBits.countNumCompletedBits();
		else
			return -1;
	}
	
	// returns true if all the done bits are set
	public boolean isDone(){
		if (DoneBits != null)
			return DoneBits.AllDone();
		else
			return false;
	}
	
	
	// sets the done bit in position pos to true
	public void SetDoneBit(int pos){
		if (DoneBits != null)
			DoneBits.set(pos);
		
	}
	
	// sets the ready bit in position pos to false
	public void ClearReadyBit(int pos)
	{
		if (ReadyBits != null)
		{
			ReadyBits.clear(pos);
		}
	}

	// returns the position of the first done bit 
	// not set to true
	public int NextNotDoneOp()
	{
		if (DoneBits != null)
			return DoneBits.FirstNotCompleted();
		else
			return -1;
	}

	// returns the position of the first ready bit 
	// set to true
	public int NextReadyOp()
	{
		if (ReadyBits != null)
			return ReadyBits.FirstReady();
		else
			return -1;
	}

	
	// returns the next operator with the highest# of tickets that is ready
	// to go
	public int NextReadyOpTicket(ArrayList<Double> OperatorTickets)
	{
		int nextOp = 0, nextPos;
		double highestFound=0;
		nextPos = ReadyBits.getBits().nextSetBit(0);


		while (nextPos != -1)
		{	
			if (highestFound <= OperatorTickets.get(nextPos-1))
			{	
				nextOp = nextPos;
				highestFound = OperatorTickets.get(nextPos-1);
			}
			nextPos++;
			nextPos = ReadyBits.getBits().nextSetBit(nextPos);
		}	
		return nextOp;
	}
	
	// returns the next operator with the highest# of tickets that is ready
	// to go
	public int NextReadyOpCBR(Hashtable operatorStreamClassifierAttribute, int streamId)
	{
		int nextOp = 0, nextPos;
		double smallestFound = 0;
		double currSelectivity = 0;
		nextPos = ReadyBits.getBits().nextSetBit(0);
		Hashtable hashProfile = (Hashtable)operatorStreamClassifierAttribute.get(nextPos);
		
		if (hashProfile == null)
		{
			return NextReadyOp(); //if profile is empty or not ready, pick the first ready operator then.
		}
		
		CBROperatorProfileEntry ope = (CBROperatorProfileEntry)hashProfile.get(streamId);
		
		if (ope == null)
		{
			return NextReadyOp(); //if profile is empty or not ready, pick the first ready operator then.
		}
		
		//selectivity depends on whether the operator has a classifying attribute or not
		//if yes, use a content-based selectivity
		//if no, use overall operator selectivty
		
		if (ope.getHasClassifierAttribute() == true)
		{
			//operator has classifier attribute: (1) look at the op content, (2) determine its content-based sel
			XATTupleImp currTuple = (XATTupleImp)this.answer[streamId];
			
			//get classifier attribute index
			int currClassifierAttributeIdx = ope.getClassifierAttributeIndex();
			
			//now retrieve the value of the classifying attribute in the current tuple
			double currClassifierAttributeVal = Double.parseDouble(currTuple.getCell(currClassifierAttributeIdx).toString());			
			double partition = currClassifierAttributeVal % 10000000;
			
			//get the selectivity of the current partition
			double partitionSelectivity = Double.parseDouble(ope.getHashClassifierAttributeDetailedSelectivities().get(partition).toString());
			currSelectivity = partitionSelectivity;
		}
		else
		{
			//operator does NOT have classifier attribute, use overall operator selectivity
			currSelectivity = ope.getOverallSelectivity();
		}		
		
		//get the profile for the current operator		
		while (nextPos != -1)
		{				
			if (smallestFound <= currSelectivity)
			{	
				nextOp = nextPos;
				smallestFound = currSelectivity;
			}
			nextPos++;
			nextPos = ReadyBits.getBits().nextSetBit(nextPos);
			hashProfile = (Hashtable)operatorStreamClassifierAttribute.get(nextPos);
			if (hashProfile == null)
			{
				continue;
			}
			ope = (CBROperatorProfileEntry)hashProfile.get(streamId);	
			currSelectivity = ope.getOverallSelectivity();
		}
		return nextOp;
	}
	
	// returns the creationdate
	public Date GetCreationDate(){
		return Creation;
	}
	
	// adds a tuple to position pos
	public void addTuple(int pos,XATTuple t)
	{
		answer[pos] = t;         	
	}
	
	// returns a tuple from position pos
	public XATTuple getTuple(int pos)
	{
		return answer[pos];         	
	}
	

	// returns the number of tuples in the answer
	public int getNumTuplesInAnswer()
	{
		return NumTuplesInAnswer;         	
	}
	
	// returns the DoneBits
	public BitStructure getDoneBitSet()
	{
		return DoneBits;
	}
	
	// sets the DoneBits
	public void setDoneBitSet(BitSet bs)
	{
		DoneBits.CreateCopyFrom(bs);
	}

	// returns the ReadyBits
	public BitStructure getReadyBitSet()
	{
		return ReadyBits;
	}
	
	// set the ReadyBits
	public void setReadyBitSet(BitSet bs)
	{
		ReadyBits.CreateCopyFrom(bs);
	}

	
	// returns the number of the Done Bits
	public int getSize()
	{
		return DoneBits.Size;
	}

	// copies the answer
	public XATTuple[] getAnswer()
	{
		return answer.clone();
	}

	// sets the answer
	public void setAnswer(XATTuple[] a)
	{
		answer = a;
	}
}

