/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.eddy.operator;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Coin;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import java.util.Vector;

import edu.wpi.cs.dsrg.STeM.STeMData;
import edu.wpi.cs.dsrg.STeM.STeMJoinProbeOperatorImp;
import edu.wpi.cs.dsrg.eddy.datastructure.*;
import java.util.BitSet;
import java.util.StringTokenizer;
//Karen Works

import org.w3c.dom.Node;

//This is the CBR-enabled eddy operator (i.e., eddy that
//can perform content-based routing) 

public class CBREddyOperatorImp extends XATMultiQueueWindowStreamOperatorImp implements EddyOperator
{
	//------------------------------------------------------------------------------------
	//                            BATCHED EXECUTION SETUP
	//------------------------------------------------------------------------------------
	
	String outFile = "C:\\Rimma\\Research\\QueryMesh\\execution_stats\\RUNTIME_CBR_EDDY_COST.txt";
	String outCBRSpecificFile = "C:\\Rimma\\Research\\QueryMesh\\execution_stats\\RUNTIME_CBR_COST.txt";
	
	//true --> batched execution, false --> tuple-by-tuple execution
	private boolean isBatchedExecution = true;
	
	//batch size
	private int batchSize = 100;
	
	//tuples that will be sent as a batch
	private int tuplesInFutureBatch = 0;
	
	//------------------------------------------------------------------------------------
	
	boolean DO_CBR_PROFILING = true;

	//------------------------------------------------------------------------------------
	
	private ExpressionStr expression;

	// number of operators
    private int NumOps = 0;
    
    //number of streams
    private int NumSTREAMS = 0;

    // number of tuples to process from pool
    private int NumTupToProcessPerOp = 500;
    private int NumTupToProcess = 2000;
    private int NumLevelsToProcess = 2;
    
	// Tuples that require processing
	private HashMap<Integer, EddyQueue> TuplePool;
	
	//operator bitmaps per STREAM
	private ArrayList <BitSet> PreFilledPerStream;
	
	// pointer to data stored for STEM
	private HashMap <Integer,STeMData> STeMStreamData;

	// array of operator's STREAM Ids
	private ArrayList <Integer> OperatorSTREAMIds;
	
	//router
	private String strRouter= "Simple";

	// array of operator's tickets for back pressure
	private ArrayList <Double> OperatorTickets;

	//are you going to flip flop strategies?
	boolean bFlip = true;

	//is looking at the top the current strategy
	boolean bTop = false;
	
	// stores window count for streams
	private HashMap <Integer,Integer> STREAMWindowCount;
	
	// for each operator stores for each input stream, its classifier attribute's values to be used in CBR routing 
	Hashtable operatorStreamClassifierAttribute = new Hashtable();
	
	//------------------------------------------------------------------------------------
	
	public CBREddyOperatorImp(XATNode init) 
	{
		super(init);		
	}

	//------------------------------------------------------------------------------------
	
	public void InitializeVariables()
	{
		//set up the PreFilledPerStream array
		PreFilledPerStream = new ArrayList(NumSTREAMS-1);
		STREAMWindowCount = new HashMap <Integer,Integer>(); 

		try
		{
			if (strRouter.compareTo("Ticket") == 0)
			{	
				OperatorTickets = new ArrayList(NumOps);
				//initialize tickets
				for (int O = 0; O < NumOps; O++)
				{
					OperatorTickets.add(0.0);
				}
			}
			
			//initializes basic bitsets for each streamid
			for (int i = 0; i <= NumSTREAMS - 1; i++)
			{					
				BitSet r = new BitSet(NumOps+1);
					
				for (int j=0; j < OperatorSTREAMIds.size();j++)
				{
					if (OperatorSTREAMIds.get(j) == i)				
						r.set(j+1, true);						
					else							
						r.clear(j+1);
				}//end for
					
				PreFilledPerStream.add(r);
			}
		}
		catch (Exception ex)
		{
			System.out.print("Error setting up EddyOperator!!");
		}  
	}
	
	//------------------------------------------------------------------------------------
	
	public boolean run() throws Exception 
	{		
		for (int i=0; i < this.getOutputQueueCount(); i++)
		{
			if(this.getOutputQueue(i).getSchema() == null)
			{
				init(i);
			}
		}

		//int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		int howMany  = 5000; //rimma temp "hack"
		
		int numHowManyProcessed = 0;	
		
		double startCBREddyTime = System.currentTimeMillis();
		
		//-------------------------------------------------------------------------
		// FIRST PROCESS THE TUPLES FROM OPERATORS INPUT QUEUES
		//-------------------------------------------------------------------------
		//Scan for Other Operators & STeM's returning Tuples.
		for (int j = NumSTREAMS; j <= (NumSTREAMS + NumOps - 1); j++) 
		{			
			XATTuple[] tuples = (this.getInputQueue(j)).dequeue(howMany, this);
			numHowManyProcessed += tuples.length;
			
			EddyTuple et;
			
			System.out.println("\nProcessing operator queue [" + j + "]");
			System.out.println("\t# of tuples in the queue: " + (this.getInputQueue(j)).getQueueSize());
			System.out.println("\tDequeued: " + tuples.length + " tuples.");
			System.out.println("");
						
			for(int k = 0; k < tuples.length;k++)
			{
				et = (EddyTuple)tuples[k];
							
				if (strRouter.equalsIgnoreCase("Ticket"))
				{	
					OperatorTickets.set(j - NumSTREAMS, OperatorTickets.get(j - NumSTREAMS) + 1);										
				}
				
				//--------------------------------------------------------------------
				// Tuple is ALL DONE
				//--------------------------------------------------------------------				
				// check to see if the tuple is done
				// ie all ops done bits set to true
				if (et.isDone() == true) 
				{					
					// if it is done, output result
					//Assume last output queue  is the result queue
					this.getOutputQueue(NumOps).enqueue(et, this);
					et = null;
				}
				else
				{	
					//--------------------------------------------------------------------
					// Tuple is NOT DONE
					//--------------------------------------------------------------------						
					// not done --> add to tuple pool for more processing
					int key = et.NumDoneBitsComplete();
					
					//System.out.println("# of ops done: " + key);
					
					EddyQueue Eq;
					
					if((Eq = TuplePool.get(key)) != null)
					{
						Eq.enqueue(et);	
					}
					else
					{
						EddyQueue<EddyTuple> eq = new EddyQueue<EddyTuple>();
						eq.enqueue(et);
						TuplePool.put(key, eq);
					}					
				}
			}//end for iterating over all tuples
		}//end for - retrieved tuples from operator queues and put them into the tuple pool		
						
		ArrayList<Boolean> OpFull = new ArrayList<Boolean>(NumOps);
		ArrayList<Integer> OpNumSent = new ArrayList<Integer>(NumOps);
		
		for (int g = 0; g < NumOps; g++)
		{	
			OpFull.add(false);
			OpNumSent.add(0);
			
			//for testing only
			if (TuplePool!= null)
			{				  
				if (TuplePool.get(g)!= null)
				{
					System.out.println(g+". at level (# of operators done): "+g+" there are "+TuplePool.get(g).size() + " tuples");
				}
				else
				{
					System.out.println(g+". TupleTool is null");
				}
			}  
		}	

		int NumOpsFull = 0;
		int NumSent = 0;
		int NumLevelsProcessed = 0;
		
		if(TuplePool != null)
		{	
			// look at each op to see if any ops have tuple returning 
			// to the eddy operator for processing
			if (bTop)
			{	
				System.out.println("\nUsing top strategy...");
				
				int numTuplesSent = 0;
				Hashtable hashTuplesSentByOp = new Hashtable();
				
				for (int k = NumOps - 1;((k >= 0) && (NumOpsFull <= NumOps) && (NumSent < NumTupToProcess))&&(NumLevelsProcessed <= NumLevelsToProcess);k--)
				{
					if (TuplePool.get(k)!= null)
					{
						if (TuplePool.get(k).size() != 0)
						{							
							System.out.println("Processing level [" + k + "]");
														
							tuplesInFutureBatch = 0; //when switching to a new level, reset the # of tuples in the batch
							
							NumLevelsProcessed++;
					
							// there are items in the queue to process
							EddyQueue Eq = TuplePool.get(k);							
							int count = 0;
							int NumInQ = Eq.length();

							int nextOp = 0;
							
							while((count != NumInQ) && (NumOpsFull < NumOps)&& (NumSent < NumTupToProcess))
							{
								EddyTuple TupletoProcess = (EddyTuple)Eq.dequeue();
								
								count++;
							
								//--------------------------------
								// CHECK IF BATCHED EXECUTION
								//--------------------------------
								if ((isBatchedExecution == true && tuplesInFutureBatch == 0) || //batched routing -- need to determine the route for the next batch
								    (isBatchedExecution == false)) //one-by-one routing  
								{								
									if (strRouter.compareTo("Simple") == 0)
									{
										nextOp = TupletoProcess.NextReadyOp();
									}
									else if (strRouter.compareTo("Ticket") == 0)
									{
										nextOp = TupletoProcess.NextReadyOpTicket(OperatorTickets);
									}
									else if (strRouter.compareTo("CBR") == 0)
									{
										//pick the next operator based on the content-based selectivity (if the operator has a classifier attribute)
										nextOp = TupletoProcess.NextReadyOpCBR(operatorStreamClassifierAttribute, TupletoProcess.getStreamID());
									}
									
									//System.out.println("The discovered next operator [" + nextOp + "]");
								}
								else
								/*{
									System.out.println("Skipping execution [# of tuples in batch: " + tuplesInFutureBatch + "]");
								}*/
								
								tuplesInFutureBatch++; //update the number of tuples in the current batch
								
								//determine if the batch is now full
								if (tuplesInFutureBatch == batchSize)
								{
									tuplesInFutureBatch = 0;
								}
								
								//add tuple to appropriate output queue
								if (!OpFull.get(nextOp - 1))
								{	
									if (OpNumSent.get(nextOp - 1) == NumTupToProcessPerOp)
									{	
										OpFull.set(nextOp-1, true);
										NumOpsFull++;
										Eq.enqueue(TupletoProcess);											
									}
									else
									{										
										//new tuples add to STEm data
										if (k == 0)
										{
											//if STeMData exists for STREAMId send tuple to STeMData
											if (STeMStreamData != null)
											{	
												STeMData sd = STeMStreamData.get(TupletoProcess.getStreamID());
												if (sd != null)
												{
													sd.Insert(TupletoProcess.getTuple(TupletoProcess.getStreamID()));
												}
												else
												{
													System.out.println("ERROR! want to catch this.");
												}
											}															
										}
										
										NumSent++;										
										
										OpNumSent.set(nextOp-1, OpNumSent.get(nextOp-1) + 1);
										this.getOutputQueue(nextOp-1).enqueue(TupletoProcess, this);
										
										if (hashTuplesSentByOp.get(nextOp-1) == null)
										{
											int countTuples = 1;
											hashTuplesSentByOp.put(nextOp-1, countTuples);
										}
										else
										{
											int countTuples = Integer.parseInt(hashTuplesSentByOp.get(nextOp-1).toString());
											countTuples++;
											hashTuplesSentByOp.put(nextOp-1, countTuples);
										}
										
										if (strRouter.compareTo("Ticket") == 0)
										{	
											OperatorTickets.set(nextOp-1, OperatorTickets.get(nextOp-1)+1);
										}										
									}
								}
								else
								{
									Eq.enqueue(TupletoProcess);
								}																
							}					
						}
					}
					
					System.out.println("---");
					 for (Enumeration e = hashTuplesSentByOp.keys(); e.hasMoreElements();) 
					 {
					        int key = Integer.parseInt(e.nextElement().toString());
					        int numSentToOp = Integer.parseInt(hashTuplesSentByOp.get(key).toString());
					        System.out.println("operator ["+key+"], "+this.getOutputQueue(key).getQueueCustomer()+" # of tuples sent [" + numSentToOp + "]");
					 }
					
					System.out.println("Done processing level: " + k +". Total # of tuples sent [" + NumSent +"]");					
				}					
			}//end if "top strategy"
			//look at the queue that is the most full	
			else
			{			    
				BitSet AlreadyProcessed = new BitSet(NumOps);
			    AlreadyProcessed.clear();
			    
				int Maxqueue = 0;
				int MaxqueueIndex = 0;
			    
				while ((NumOpsFull <= NumOps)&& (NumSent <= NumTupToProcess) && (NumLevelsProcessed <= NumLevelsToProcess))
				{
					Maxqueue = 0;
					MaxqueueIndex = 0;
					int i = 0;
					while (i<=NumOps)	
					{
						i = AlreadyProcessed.nextClearBit(i);						
						if ((TuplePool.get(i)!= null) && (!AlreadyProcessed.get(i)))
						{						  
							if (Maxqueue <TuplePool.get(i).size())
							{	
								Maxqueue = TuplePool.get(i).size();
								MaxqueueIndex = i;
							}	
						}
						else if (TuplePool.get(i)== null)
						{
							AlreadyProcessed.set(i);
						}
						i++;					
					}
					
					AlreadyProcessed.set(MaxqueueIndex);
					
					if (TuplePool.get(MaxqueueIndex)!= null)
					{
						NumLevelsProcessed++;
					
						// there are items in the queue to process
						EddyQueue Eq = TuplePool.get(MaxqueueIndex);
						int count = 0;
						int NumInQ = Eq.length();
						while((count != NumInQ) && (NumOpsFull<NumOps) && (NumSent<NumTupToProcess))
						{
							EddyTuple TupletoProcess = (EddyTuple)Eq.dequeue();
							count++;
							int nextOp = 0;
							if (strRouter.compareTo("Simple") == 0)
							{
								nextOp = TupletoProcess.NextReadyOp();
							}
							else if (strRouter.compareTo("Ticket") == 0)
							{
								nextOp = TupletoProcess.NextReadyOpTicket(OperatorTickets);
							}
							else if (strRouter.compareTo("CBR") == 0)
							{
								//pick the next operator based on the content-based selectivity (if the operator has a classifier attribute)
								nextOp = TupletoProcess.NextReadyOpCBR(operatorStreamClassifierAttribute, TupletoProcess.getStreamID());
							}
							
							//add tuple to appropriate output queue
							if (!OpFull.get(nextOp - 1))
							{	
								if (OpNumSent.get(nextOp - 1) == NumTupToProcessPerOp)
								{	
									OpFull.set(nextOp-1, true);
									NumOpsFull++;
									Eq.enqueue(TupletoProcess);	
								}
								else
								{
									NumSent++;
									OpNumSent.set(nextOp-1,OpNumSent.get(nextOp-1)+1);
									this.getOutputQueue(nextOp-1).enqueue(TupletoProcess, this);
									if (strRouter.compareTo("Ticket")==0)
									{	
										OperatorTickets.set(nextOp-1, OperatorTickets.get(nextOp-1)+1);
									}
								}
							}
							else
							{
								Eq.enqueue(TupletoProcess);
							}
						}						
					}
				}
			}		
			if(bFlip)	
			{
				bTop = !bTop;
			}
		}

		//-----------------------------------------------------------------------
		// SECOND PROCESS THE TUPLES FROM THE INPUT STREAMS
		//-------------------------------------------------------------------------
		ArrayList allStreamsSamples = new ArrayList();		
				
		/** get the tuples from STREAM*/
		for (int j = 0; j < (NumSTREAMS); j++) 
		{				
			System.out.println("\nProcessing stream queue [" + j + "]");
			System.out.println("\t# of tuples in the queue: " + (this.getInputQueue(j)).getQueueSize());
			
			XATTuple[] tuples = this.getInputQueue(j).dequeue(howMany, this);
			System.out.println("\tDequeued: " + tuples.length + " tuples.");
			System.out.println("");						
			
			//-----------------------------------------------------------------------
			// CBR: SAMPLE TUPLES FROM THE STREAM
			//-------------------------------------------------------------------------
			int sampleSize = 100;
			Coin flipCoin = new Coin();
			ArrayList  tupleSample = new ArrayList();
			
			for(int k = 0; k < tuples.length;k++)
			{
				flipCoin.flip();
				if (flipCoin.toString().equals("Heads"))
				{
					//sample the tuple
					tupleSample.add(tuples[k]);
				}
				
				if (tupleSample.size() == sampleSize)
				{
					break;
				}
			}
			allStreamsSamples.add(tupleSample);
						
			//-------------------------------------------------------------------------
			// END SAMPLE STREAM TUPLES
			//-------------------------------------------------------------------------
			
			EddyTuple et;
			// process all new tuples coming from STREAMS
			for(int k = 0; k < tuples.length;k++)
			{				
				et = new EddyTupleImp(tuples[k], PreFilledPerStream.get(j), NumOps, NumSTREAMS, j);
				et.setStreamID(j);
				// add new tuple to tuple pool
				
				EddyQueue Eq;
				if(TuplePool != null)
				{
					if((Eq = TuplePool.get(0)) != null)
					{
						Eq.enqueue(et);	
					}
					else
					{
						EddyQueue<EddyTuple> NewEq = new EddyQueue<EddyTuple>();
						NewEq.enqueue(et);
						TuplePool.put(0, NewEq);
					}
				}
				else
				{
					TuplePool = new HashMap();
					EddyQueue<EddyTuple> eq = new EddyQueue<EddyTuple>();
					eq.enqueue(et);
					TuplePool.put(0, eq);					
				}				
			}		
		}		

		//-------------------------------------------------------------------------
		// CBR-RELATED PROCESSING
		//-------------------------------------------------------------------------
		for (int j = NumSTREAMS; j <= (NumSTREAMS + NumOps - 1); j++) 
		{
			Hashtable operatorInputs = new Hashtable();
			
			ArrayList owners =  (ArrayList)(this.getInputQueue(j)).getQueueOwner();
			Object profileOperator = owners.get(0);
			
			if (profileOperator.getClass().toString().equalsIgnoreCase("class edu.wpi.cs.dsrg.STeM.STeMJoinProbeOperatorImp"))
			{				
				STeMJoinProbeOperatorImp realProfileOperator = (STeMJoinProbeOperatorImp)profileOperator;				
				int opId = realProfileOperator.getOperatorID();
			
				operatorStreamClassifierAttribute.put(opId, operatorInputs);
			}
		}
				
		ArrayList streamSchemas = new ArrayList();		
		
		//ArrayList classifierAttributeMatrix = new ArrayList(); //Classifier Attribute Matrix, CA[]		
		//ArrayList detailedSelectivitiesMatrix = new ArrayList();
		//ArrayList overallOperatorSelectivities = new ArrayList();
		
		int numPartitions = 5; //default value (but could be larger)

		//-------------------------------------------------------------------------
		// CBR: PRE-PROCESSING. DO A SIMPLE CHECK TO MAKE SURE THAT THE SERVER HAS
		// RECEIVED SCHEMA FROM ALL STREAM. IF THIS IS NOT THE CASE, SKIP THE PROFILING
		//-------------------------------------------------------------------------
	
		//-------------------------------------------------------------------------
		// GET THE SCHEMAS OF THE INPUT STREAMS
		//-------------------------------------------------------------------------
		for (int j = 0; j < (NumSTREAMS); j++) 
		{	
			XATQueueSchema currStreamSchema = this.getInputQueue(j).getSchema();	
			
			if (currStreamSchema == null)
			{
				DO_CBR_PROFILING = false;				
			}
			streamSchemas.add(currStreamSchema);
		}
		
		if (!DO_CBR_PROFILING)
		{
			System.out.println("Skipping CBR profile...");
		}
		
		//the schemas of the input streams are all received, so procede with profiling
		if (DO_CBR_PROFILING)
		{
			System.out.println("Running CBR profile...");
			
			double startCBRProfileTime = System.currentTimeMillis();
			
			//-------------------------------------------------------------------------
			// INITIALIZE TUPLES-IN AND TUPLES-OUT DATA STRUCTURES
			//-------------------------------------------------------------------------
			Hashtable hashOpTuplesInAndOut = new Hashtable();
				
			//initialize tuplesIn and tuplesOut data structures for each operator
			for (int j = NumSTREAMS; j <= (NumSTREAMS + NumOps - 1); j++) 
			{					
				ArrayList owners =  (ArrayList)(this.getInputQueue(j)).getQueueOwner();
				Object profileOperator = owners.get(0);
				if (profileOperator.getClass().toString().equalsIgnoreCase("class edu.wpi.cs.dsrg.STeM.STeMJoinProbeOperatorImp"))
				{			
					Hashtable tuplesIn = new Hashtable();
					Hashtable tuplesOut = new Hashtable();
					
					STeMJoinProbeOperatorImp realProfileOperator = (STeMJoinProbeOperatorImp)profileOperator;
					
					int opId = realProfileOperator.getOperatorID();
					
					int stream1Id = realProfileOperator.getProbeSTREAMId();
					int stream2Id = realProfileOperator.getDataSTREAMId();
							
					int numAttributesInStream1Schema = ((XATQueueSchema)streamSchemas.get(stream1Id)).getLength();
					int numAttributesInStream2Schema = ((XATQueueSchema)streamSchemas.get(stream2Id)).getLength();
	
					//stream 1
					ArrayList streamAttributesTuplesIn = new ArrayList();
					ArrayList streamAttributesTuplesOut = new ArrayList();		
					
					//the reason why we skip the last attribute is because it is monotonically increasing
					//and will never increasing and will never be a "classifier attribute" (see paper for more details)
					for(int v = 0; v < numAttributesInStream1Schema - 1; v++) 
					{
						Hashtable tuplesInEntry = new Hashtable(); //key is partition of the attribute
						Hashtable tuplesOutEntry = new Hashtable(); //key is partition of the attribute	
						
						streamAttributesTuplesIn.add(tuplesInEntry);
						streamAttributesTuplesOut.add(tuplesOutEntry);
					}
					tuplesIn.put(stream1Id, streamAttributesTuplesIn);
					tuplesOut.put(stream1Id, streamAttributesTuplesOut);
					ArrayList arrTuplesInAndOut = new ArrayList();
					arrTuplesInAndOut.add(tuplesIn);
					arrTuplesInAndOut.add(tuplesOut);
					hashOpTuplesInAndOut.put(opId, arrTuplesInAndOut);
					
					/*
					//stream 2			
					streamAttributesTuplesIn = new ArrayList();
					streamAttributesTuplesOut = new ArrayList();
					//similar as for stream 1, the reason why we skip the last attribute is because it is monotonically increasing
					//and will never increasing and will never be a "classifier attribute" (see paper for more details)
					for(int v = 0; v < numAttributesInStream2Schema - 1; v++)
					{
						Hashtable tuplesInEntry = new Hashtable(); //key is partition of the attribute
						Hashtable tuplesOutEntry = new Hashtable(); //key is partition of the attribute	
						
						streamAttributesTuplesIn.add(tuplesInEntry);
						streamAttributesTuplesOut.add(tuplesOutEntry);
					}
					tuplesIn.add(stream2Id, streamAttributesTuplesIn);
					tuplesOut.add(stream2Id, streamAttributesTuplesOut);
					*/				
				}
			}
						
			//-------------------------------------------------------------------------
			// CBR: PROFILE EACH OPERATOR
			//-------------------------------------------------------------------------
			for (int x = NumSTREAMS; x <= (NumSTREAMS + NumOps - 1); x++) 
			{						
				XATTuple[] tuples = (this.getInputQueue(x)).dequeue(howMany, this);
				
				System.out.println("Processing operator queue [" + x + "]");
				System.out.println("\t# of tuples in the queue: " + (this.getInputQueue(x)).getQueueSize());
				System.out.println("\tDequeued: " + tuples.length + " tuples.");
				
				ArrayList owners =  (ArrayList)(this.getInputQueue(x)).getQueueOwner();
				Object profileOperator = owners.get(0);
				if (profileOperator.getClass().toString().equalsIgnoreCase("class edu.wpi.cs.dsrg.STeM.STeMJoinProbeOperatorImp"))
				{				
					STeMJoinProbeOperatorImp realProfileOperator = (STeMJoinProbeOperatorImp)profileOperator;
					int opId = realProfileOperator.getOperatorID();
					
					int stream1Id = realProfileOperator.getProbeSTREAMId();
					int stream2Id = realProfileOperator.getDataSTREAMId();
					
					int stream1AttrId = realProfileOperator.getNumTupleIndex().get(0);
					int stream2AttrId = realProfileOperator.getSTeMDataIndexArrayPos();
									
					ArrayList stream1Sample = (ArrayList)allStreamsSamples.get(stream1Id);
					ArrayList stream2Sample = (ArrayList)allStreamsSamples.get(stream2Id);	
					
					//-------------------------------------------------------------------------
					// PARTITION ATTRIBUTE DOMAINS BASED ON PARTITIONING FUNCTION
					//-------------------------------------------------------------------------
					//-------------------------------------------------------------------------
					// SINCE WE KNOW ABOUT THE DATA, THE PARTITIONS LOOK LIKE THIS (FOR ALL ATTRIBUTE
					// EXCEPT THE LAST ONE)
					//-------------------------------------------------------------------------
					// 0 - 10000000 
					// 10000000 - 20000000
					// 20000000 - 30000000
					// 30000000 - 40000000
					// 40000000 - 50000000				
					
					//-------------------------------------------------------------------------
					// POPULATE TUPLES IN/OUT FOR THE CURRENT OPERATOR (FOR ALL ATTRIBUTE PARTITIONS)
					//-------------------------------------------------------------------------
					int tuplesInVal = 0;
					int tuplesOutVal = 0;
					for(int k = 0; k < stream1Sample.size(); k++)
					{
						tuplesInVal = 1;
						XATTuple sampleTuple = (XATTuple)stream1Sample.get(k);
						
						//probe the opposite stream's sample
						for(int m = 0; m < stream2Sample.size(); m++)
						{
							XATTuple oppositeStreamTuple = (XATTuple)stream2Sample.get(m);
							
							if (sampleTuple.getCell(stream1AttrId).equals(oppositeStreamTuple.getCell(stream2AttrId)))
							{
								tuplesOutVal++;
							}
						}
										
						ArrayList arrTuplesInAndOut = (ArrayList)hashOpTuplesInAndOut.get(opId);
						Hashtable tuplesIn = (Hashtable)arrTuplesInAndOut.get(0);
						Hashtable tuplesOut = (Hashtable)arrTuplesInAndOut.get(1);
						
						//get the appropriate ArrayList corresponding to all different attributes
						ArrayList stream1AttributesTuplesIn = (ArrayList)tuplesIn.get(stream1Id);
						ArrayList stream2AttributesTuplesIn = (ArrayList)tuplesIn.get(stream2Id);
						
						ArrayList stream1AttributesTuplesOut = (ArrayList)tuplesOut.get(stream1Id);
						ArrayList stream2AttributesTuplesOut = (ArrayList)tuplesOut.get(stream2Id);			
						
						//now update all the tuples in and tuples out values for all attributes and their partitions
						//-----------------------------------------------------------------
						// UPDATE VALUES FOR STREAM 1
						//-----------------------------------------------------------------
						XATDataValue[] stream1DataValues = sampleTuple.getArrayData();
												
						//iterate through each attribute in the schema
						for(int q = 0; q < stream1AttributesTuplesIn.size(); q++)
						{
							Hashtable tuplesInEntry = (Hashtable)stream1AttributesTuplesIn.get(q); 
							Hashtable tuplesOutEntry = (Hashtable)stream1AttributesTuplesOut.get(q); 
							
							//double partition = stream1DataValues[q].convertToDouble() % 10000000;
							int partition = 0;
							String keyPartition = ""; //the actual key that will be used in profile
							
							try
							{
								partition = (int) (stream1DataValues[q].convertToDouble() % 10);
								keyPartition = ""+partition;
							}
							catch(Exception ex)
							{
								keyPartition = stream1DataValues[q].toString(); //partition key for non-integer values (the actual value itself)
							}
							
							//now update all the tuples in and tuples out values for all attributes and their partitions
							
							int valIn = 0;
							if (tuplesInEntry.get(keyPartition) != null)
							{
								valIn = Integer.parseInt(tuplesInEntry.get(keyPartition).toString());
							}						
							valIn++;						
							tuplesInEntry.put(keyPartition, valIn);
							
							int valOut = 0;
							if (tuplesOutEntry.get(keyPartition) != null)
							{
								valOut = Integer.parseInt(tuplesOutEntry.get(keyPartition).toString());
							}						
							valOut += tuplesOutVal;						
							tuplesOutEntry.put(keyPartition, valOut);
						}					
					}
					
					//-------------------------------------------------------------------------
					// COMPUTE OVERALL SELECTIVITY FOR THE OPERATOR
					//-------------------------------------------------------------------------
					
					ArrayList arrTuplesInAndOut = (ArrayList)hashOpTuplesInAndOut.get(opId);
					Hashtable tuplesIn = (Hashtable)arrTuplesInAndOut.get(0);
					Hashtable tuplesOut = (Hashtable)arrTuplesInAndOut.get(1);
					
					ArrayList streamAttributesTuplesIn = (ArrayList)tuplesIn.get(stream1Id);
					ArrayList streamAttributesTuplesOut = (ArrayList)tuplesOut.get(stream1Id);
					
					int totalTuplesIn = 0;
					int totalTuplesOut = 0;								
					
					for (int a = 0; a < streamAttributesTuplesIn.size(); a++)
					{						
						Hashtable tuplesInEntry = (Hashtable)streamAttributesTuplesIn.get(a);					
						for (Enumeration e = tuplesInEntry.keys(); e.hasMoreElements();) 
						{				
							String key = e.nextElement().toString();
							
							if (tuplesInEntry.get(key) == null)
							{
								System.out.println("Want to catch this!");
							}
							else
							{
								String numTuples = tuplesInEntry.get(key).toString();								
								totalTuplesIn += Integer.parseInt(numTuples);
							}
						}
						
						Hashtable tuplesOutEntry = (Hashtable)streamAttributesTuplesOut.get(a);						
						for (Enumeration e = tuplesOutEntry.keys(); e.hasMoreElements();) 
						{				
							String key = e.nextElement().toString();
							totalTuplesOut += Integer.parseInt(tuplesOutEntry.get(key).toString());				   				
						}
					}
					
					double OverallSelectivity_R = 0;
					if (totalTuplesIn > 0)
					{
						OverallSelectivity_R = totalTuplesOut / totalTuplesIn;
					}
					//-------------------------------------------------------------------------
					// COMPUTE GAIN RATIO
					//-------------------------------------------------------------------------
					
					double Entropy_R = -OverallSelectivity_R * Math.log(OverallSelectivity_R)/Math.log(2) - (1 - OverallSelectivity_R) * Math.log(1 - OverallSelectivity_R)/Math.log(2);								
					
					//-------------------------------------------------------------------------
					// ITERATE THROUGH ALL ATTRIBUTES AND ALL ATTRIBUTE PARTITIONS FOR THE INPUT STREAM 
					// AND COMPUTE MACHINE LEARNING MEASURES (ATTRIBUTE + ITS PARTITIONS) TO DETERMINE THE CLASSIFIER ATTRIBUTE
					//-------------------------------------------------------------------------
									
					double maxGainRatio = 0.0;
					int currClassifierAttributeIdx = -1;
					
					ArrayList stream1AttributesTuplesIn = (ArrayList)tuplesIn.get(stream1Id);				
					Hashtable detailedSelectivities;
					Hashtable detailedSelectivitiesClassifierAttribute = new Hashtable();					
					
					for(int b = 0; b < stream1AttributesTuplesIn.size(); b++)
					{		
						detailedSelectivities = new Hashtable();
						double PartitionsContributions = 0;
						
						Hashtable tuplesInEntry = (Hashtable)streamAttributesTuplesIn.get(b);	
						Hashtable tuplesOutEntry = (Hashtable)streamAttributesTuplesOut.get(b);	
						
						numPartitions = tuplesInEntry.size();
						
						//-------------------------------------------------------------------------
						// INFO GAIN
						//-------------------------------------------------------------------------
						for (Enumeration e = tuplesOutEntry.keys(); e.hasMoreElements();) 
						{				
							String key = e.nextElement().toString();					
							int partitionTuplesIn = Integer.parseInt(tuplesInEntry.get(key).toString());	
							int partitionTuplesOut = Integer.parseInt(tuplesOutEntry.get(key).toString());	
							double OverallSelectivity_R_i = partitionTuplesOut / partitionTuplesIn;
							double Entropy_R_i = -OverallSelectivity_R_i * Math.log(OverallSelectivity_R_i)/Math.log(2) - (1 - OverallSelectivity_R_i) * Math.log(1 - OverallSelectivity_R_i)/Math.log(2);
							double fractionPerPartition = partitionTuplesIn / totalTuplesIn;
							PartitionsContributions += fractionPerPartition * Entropy_R_i;
						}
						
						double InfoGain_R_A = Entropy_R - PartitionsContributions;

						//-------------------------------------------------------------------------
						// GAIN RATIO
						//-------------------------------------------------------------------------
						//normalize info gain				
						double splitInfoPartitionsContributions = 0;
						
						for (Enumeration e = tuplesOutEntry.keys(); e.hasMoreElements();) 
						{				
							String key = e.nextElement().toString();					
							int partitionTuplesIn = Integer.parseInt(tuplesInEntry.get(key).toString());	
							int partitionTuplesOut = Integer.parseInt(tuplesOutEntry.get(key).toString());	
							
							double fractionPerPartition = partitionTuplesIn / totalTuplesIn;
							
							//store the detailed selectivity for the partition
							if (detailedSelectivities.get(key) == null)
							{
								detailedSelectivities.put(key, fractionPerPartition);
							}
							else
							{
								double currVal = Double.parseDouble(detailedSelectivities.get(key).toString());
								currVal += fractionPerPartition;
								detailedSelectivities.put(key, currVal);
							}
							
							double logFractionPerPartition = Math.log(fractionPerPartition)/Math.log(2);
							splitInfoPartitionsContributions += (fractionPerPartition * logFractionPerPartition);
						}
						
						double SplitInfo_R_A = -splitInfoPartitionsContributions;
						double GainRatio_R_A = InfoGain_R_A / SplitInfo_R_A;
						
						//if max gain ratio, update the classifier attribute for this operator
						if (GainRatio_R_A > maxGainRatio)
						{
							currClassifierAttributeIdx = b;
							maxGainRatio = GainRatio_R_A;
							detailedSelectivitiesClassifierAttribute = detailedSelectivities;
						}
					}//end if STeMJoinProbeOperatorImp		
					
					//prepare the detailed content-based selectivity table for the attribute that is now deemed as the "classifier" attribute
					//ArrayList tuples = (ArrayList)stream1AttributesTuplesIn.get(currClassifierAttributeIdx);
					
					//----------------------------------------------------------------------------------------
					// GET THE PROFILE ENTRY FOR THE CURRENT OPERATOR
					//----------------------------------------------------------------------------------------						
					//store the classifier attribute for the operator and its corresponding content-based selectivities				
					Hashtable operatorInputs = (Hashtable)operatorStreamClassifierAttribute.get(opId);
					boolean hasClassifierAttribute = false;
					
					//----------------------------------------------------------------------------------------
					// PREPARE THE PROFILE OF THE CURRENT OPERATOR
					//----------------------------------------------------------------------------------------					
					double overallOperatorSelectivity = OverallSelectivity_R;
					
					if (maxGainRatio > 0)
					{
						hasClassifierAttribute = true;						
					}
					else
					{
						hasClassifierAttribute = false;
					}
					
					//create profile entry for the operator
					CBROperatorProfileEntry ope = new CBROperatorProfileEntry(hasClassifierAttribute, currClassifierAttributeIdx, detailedSelectivitiesClassifierAttribute, overallOperatorSelectivity);
					
					//update the operator profile
					operatorInputs.put(stream1AttrId, ope);	
					
				}//if operator is a StemProbe			
			}//done iterating through all operators 
			
			System.out.println("Done CBR profiling.");
			
			
			double endCBRProfileTime = System.currentTimeMillis();
			double durationCBRProfile = endCBRProfileTime - startCBRProfileTime; //milliseconds
			
			WriteToFile(outCBRSpecificFile, "# of tuples: \t"+ numHowManyProcessed+ "\tTime to do CBR profile: " + "\t" + String.valueOf(durationCBRProfile));

		}//end if DO_PROFILE
		//-------------------------------------------------------------------------
		// END OF OPERATOR PROFILE
		//-------------------------------------------------------------------------		
		
		//System.out.println("# of tuples on level 0: " + TuplePool.get(0).size() + " tuples.");
		
		double endCBREddyTime = System.currentTimeMillis();
		double durationCBREddy = endCBREddyTime - startCBREddyTime; //milliseconds
		
		WriteToFile(outFile, "# of tuples: \t"+ numHowManyProcessed +  "\tTime to execute CBR EDDY: " + "\t" + String.valueOf(durationCBREddy));

		System.gc();
		return true;
	}
	
	//------------------------------------------------------------------------------------	
	
	private void ProfileOperator()
	{
		
	}
	
	//-------------------------------------------------------------------------------------
	
	//used for experiments (otherwise comment it out)
	public void WriteToFile(String fileLocation, String stringToWrite)
	{
		try
		{
			//Create file 
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			out.newLine();			
			//Close the output stream
			out.close();			    
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}

	//------------------------------------------------------------------------------------
	
	public void setWindowSize(String[] size) 
	{
		windowSize = Integer.parseInt(size[0]);
	}
	
	//------------------------------------------------------------------------------------
	
	public void init(int i) 
	{		
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) 
		{
			return;
		}
		
		XATQueueSchema outSchema = this.getOutputQueue(i).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) 
		{
			this.getOutputQueue(i).setSchema(inSchema);
		}
			
		computeOutputPositions();//no need to overload since all the output queue are the same
		numOutputColumns = outputPositions[0].length;
		setHasInit(true);				
	}

	//------------------------------------------------------------------------------------
	
	public String toToolTipString() 
	{
		return "EddyOperatorImp";
	}

	//------------------------------------------------------------------------------------
	
	public ExpressionStr getExpression() 
	{
		return expression;
	}
	
	//------------------------------------------------------------------------------------

	public void setExpression(ExpressionStr expression) 
	{
		this.expression = expression;
	}

	//------------------------------------------------------------------------------------
	
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return true;
	}

	//------------------------------------------------------------------------------------
	
	public boolean cleanAllAccumulatedTuples() 
	{
		return true;
	}

	//------------------------------------------------------------------------------------
	
	public boolean hasUnmatchedState() 
	{
		return false;
	}

	//------------------------------------------------------------------------------------
	
	public boolean matchStates(XATStreamOperator op) 
	{
		return false;
	}

	//------------------------------------------------------------------------------------
	
	public boolean recomputeUnmatchedState() 
	{
		return true;
	}

	//------------------------------------------------------------------------------------
	
	public int setStateID(int nextid) 
	{
		return nextid;
	}

	//------------------------------------------------------------------------------------
	
	public boolean finalizeOperation() 
	{
		return true;
	}

	public void setSTeMStreamData(HashMap<Integer,STeMData> sd)
	{
		STeMStreamData = sd;
	}

	//------------------------------------------------------------------------------------
	
	public void setNumStreams(int Ns)
	{
		NumSTREAMS = Ns;
	}
	
	//------------------------------------------------------------------------------------
	
	public void setNumOps(int NOps)
	{
		NumOps = NOps;	   
	}
	
	//------------------------------------------------------------------------------------
	
	public void setOpStringIds(String val)
	{
		OperatorSTREAMIds = new ArrayList();
		StringTokenizer st = new StringTokenizer(val);
	    while (st.hasMoreElements()) 
	    {
	    	//System.out.println(st.nextElement());
			OperatorSTREAMIds.add(Integer.parseInt(st.nextElement().toString()));
	    }
	    
	}

	//------------------------------------------------------------------------------------
	
	public void setRouter(String val)
	{
		strRouter = val;
	}

	//------------------------------------------------------------------------------------
	
	public void SetTuplePool(HashMap<Integer, EddyQueue> t)
	{
		TuplePool = t;
	}
	
	//------------------------------------------------------------------------------------
		
	public ArrayList<BitSet> getPreFilledPerStream()
	{
		return PreFilledPerStream;
	}	

	//------------------------------------------------------------------------------------
	
	public void setNumTupToProcessPerOp(int val)
	{
		NumTupToProcessPerOp = val;
	}

	//------------------------------------------------------------------------------------
	
	public void setNumTupToProcess(int val)
	{
		NumTupToProcess = val;
	}

	//------------------------------------------------------------------------------------
	
	public void setNumLevelsToProcess(int val)
	{
		NumLevelsToProcess = val;
	}

	//------------------------------------------------------------------------------------
	
	public void setFlipFlopTupleProcStrategy(boolean val)
	{
		bFlip = val;
	}

	//------------------------------------------------------------------------------------
	
	public void setUseTupleProcTopStrategy(boolean val)
	{
		bTop = val;
	}

	//------------------------------------------------------------------------------------
	
	public void AddSTREAMWindowCount(int streamId, int val)
	{
		STREAMWindowCount.put(streamId,val);
	}

	//------------------------------------------------------------------------------------
	
	public HashMap <Integer,Integer> getSTREAMWindowCount()
	{
		return STREAMWindowCount;
	}
	
	//------------------------------------------------------------------------------------
	
}


