/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.eddy.operator;

// Karen Works

// This is a dummy operator that does nothing but 
//displays the results

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.Vector;
import edu.wpi.cs.dsrg.eddy.datastructure.*;

public class DoNothingOperatorImp extends XATMultiQueueWindowStreamOperatorImp 
implements DoNothingOperator
{

	private ExpressionStr expression;

	public int windowSize;

	public DoNothingOperatorImp(XATNode init) 
	{
		super(init);
	}

	String outFile = "C:\\Rimma\\Research\\QueryMesh\\execution_stats\\EDDY_RESULTS.txt";

	public static void WriteToFile(String fileLocation, String stringToWrite, boolean doNewLine)
	{
		try
		{
			//Create file 
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			if (doNewLine)
			{
				out.newLine();
			}
			//Close the output stream
			out.close();			    
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}
	
	public boolean run() throws Exception 
	{		
		for (int i = 0; i < this.getOutputQueueCount();i++)
		{
			if(this.getOutputQueue(i).getSchema() == null)
			{
				init(i);
			}
		}
				
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		/** get the tuples */

		XATTuple[] tuples = this.getInputQueue().dequeue(howMany, this);

		WriteToFile(outFile, "Number of results: ", false);		
		WriteToFile(outFile, String.valueOf(tuples.length), true);
		
		for(int i = 0; i < tuples.length;i++)
		{	
			this.getOutputQueue().enqueue(tuples[i], this);	
			
			/*
			EddyTuple e = (EddyTuple)tuples[i];						
			for (int j = 0 ; j < e.getNumTuplesInAnswer();j++)
			{	
				if (e.getTuple(j)!= null)
				{	
					// for testing purposes only
					//System.out.println(j);
					//System.out.println(e.getTuple(j));
					WriteToFile(outFile, e.getTuple(j).toString(), false);
					//this.getOutputQueue().enqueue(e.getTuple(j), this);
				}		
			}
			WriteToFile(outFile, "", true);
			*/				
		}
		
		System.gc();
		return true;
	}

	public void setWindowSize(String[] size) 
	{
		windowSize = Integer.parseInt(size[0]);
	}

	// overload supper method

	public void init(int i) 
	{
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) 
		{
			return;
		}
		
		XATQueueSchema outSchema = this.getOutputQueue(i).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) 
		{
			this.getOutputQueue(i).setSchema(inSchema);
		}
		
		computeOutputPositions();//no need to overload since all the output queue are the same
		numOutputColumns = outputPositions[0].length;
		setHasInit(true);
	}

	public String toToolTipString() 
	{
		return "DoNothing (Eddy)";
	}

		
	public ExpressionStr getExpression() 
	{
		return expression;
	}

	public void setExpression(ExpressionStr expression) 
	{
		this.expression = expression;
	}

	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input
	 * queues. Inside method run() the numbers are computed based on the
	 * property "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler, while
	 * in cleanAccumulatedTuples, all tuples in input queues that have their max
	 * timesamp smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return true;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() 
	{
		return true;
	}

	/*
	 * check if this operator has any unmatched states. return: TRUE if this
	 * operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() 
	{
		return false;
	}

	public boolean matchStates(XATStreamOperator op) 
	{
		return false;
	}

	public boolean recomputeUnmatchedState() 
	{
		return true;
	}

	public int setStateID(int nextid) 
	{
		return nextid;
	}

	public boolean finalizeOperation() 
	{
		return true;
	}
}
