/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.eddyBatch.datastructure;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Calendar;
import java.util.Date;
import java.util.Random;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;

//Karen Works

//This structure stores the information associated with 
//an eddy tuple and eddy tuple operations.

public class EddyBatchTupleImp extends XATTupleImp implements EddyBatchTuple, Serializable
{
	// intermediate tuple that will be the answer
	private XATTuple[] answer = null;
	
	// ready bits & done bits
	private BitStructure ReadyBits;
	private BitStructure DoneBits;
	
//	//completed joins
//	private BitStructure CompletedJoinAttrib; 
	
	// number of tuples in answer
	private int NumTuplesInAnswer;
	
	// creation date
	private Date Creation;
	
	//Partition Number
	private int PartitionNumber;
	
	//Num Tuples inBatch
	private int NumTupBatch;
	
	// fore producing random numbers
	private Random generator;
		
	private double dPercNeverUsed = 0.05;
	
	public long SearchStart;
	
	
	public EddyBatchTupleImp(XATTuple t,int NumOps,int NumBits,int NumStreams,int TuplePos,int PartitionNum, long lSearchStart) 
	{ //throws XATTimeStampException {
		super();

		if (t != null) 
		{
			// set up answer
			answer = new XATTuple[NumStreams];
			answer[TuplePos] = t;         
	
			// set time
			this.setMaxTimeStamp(t.getMaxTimeStamp());
			this.setMinTimeStamp(t.getMinTimeStamp());
			
			NumTuplesInAnswer = NumStreams;
			
			// create ready & done bits
			ReadyBits = new BitStructure(NumOps);
			DoneBits = new BitStructure(NumBits);
			
//			//complete join
//			CompletedJoinAttrib = new BitStructure(NumJoinAttrib);
			
			//set creation time
			Creation = Calendar.getInstance().getTime();
			
			//partition number
			PartitionNumber = PartitionNum;
			
		    //create random number generator	
			generator = new Random();
		    generator.setSeed((new Date()).getTime());
		    
			SearchStart = lSearchStart;	
		}
	}

	public EddyBatchTupleImp(XATTuple t,BitSet bs,int NumOps,int NumBits,int NumStreams,int TuplePos,int PartitionNum, long lSearchStart) 
	{
		//throws XATTimeStampException {
		super();
		if (t != null) 
		{
			
			answer = new XATTuple[NumStreams];
			answer[TuplePos] = t;         

			NumTuplesInAnswer = NumStreams;
			
			this.setMaxTimeStamp(t.getMaxTimeStamp());
			this.setMinTimeStamp(t.getMinTimeStamp());
			
			// create ready & done bits
			ReadyBits= new BitStructure();
			ReadyBits.CreateCopyFrom(bs);
			DoneBits = new BitStructure(NumBits);
			
//			//complete join
//			CompletedJoinAttrib = new BitStructure(NumJoinAttrib);
			
			//set creation time
			Creation = Calendar.getInstance().getTime();			
			
			//partition number
			PartitionNumber = PartitionNum;
			
		    //create random number generator	
			generator = new Random();
		    generator.setSeed((new Date()).getTime());
		    
			SearchStart = lSearchStart;	

		}
		
	}

	public EddyBatchTupleImp(int NumStreams, int NumOps,int NumBits, int PartitionNum) 
	{ 
		//throws XATTimeStampException {
		super();
		
		answer = new XATTuple[NumStreams];

		NumTuplesInAnswer = NumStreams;
		
		// create ready & done bits
		ReadyBits= new BitStructure(NumOps);		
		DoneBits = new BitStructure(NumBits);
		
//		//complete join
//		CompletedJoinAttrib = new BitStructure(NumJoinAttrib);
		
		//set creation time
		Creation = Calendar.getInstance().getTime();
		
		//partition number
		PartitionNumber = PartitionNum;
		
	    //create random number generator	
		generator = new Random();
	    generator.setSeed((new Date()).getTime());		

		SearchStart = System.currentTimeMillis();	
	}
	
////////////////////////////////////////////////////////
//	// returns true if all the CompletedJoinAttrib are set
//	public boolean isComplete(){
//		if (DoneBits != null)
//			return DoneBits.AllDone();
//		else
//			return false;
//	}
//
//	// sets the CompletedJoinAttrib bit in position pos to false
//	public void ClearCompletedJoinAttribBit(int pos)
//	{
//		if (CompletedJoinAttrib != null)
//		{
//			CompletedJoinAttrib.clear(pos);
//		}
//	}
//	
//	// sets the CompletedJoinAttrib bit in position pos to true
//	public void SetCompletedJoinAttribBit(int pos){
//		if (CompletedJoinAttrib != null)
//			CompletedJoinAttrib.set(pos);
//		
//	}
//
//	// returns the CompletedJoinAttrib
//	public BitStructure GetCompletedJoinAttrib(){
//		return CompletedJoinAttrib;
//		
//	}
//
//	// sets the CompletedJoinAttrib
//	public void SetCompletedJoinAttrib(BitSet bs){
//			CompletedJoinAttrib.CreateCopyFrom(bs);		
//	}
//	
//////////////////////////////////////////////////////	
	// returns the number of donebits set to complete
	public int NumDoneBitsComplete(){
		if (DoneBits != null)
		  return DoneBits.countNumCompletedBits();
		else
			return -1;
	}

	// returns true if all the done bits are set
	public boolean isDone(){
		if (DoneBits != null)
			return DoneBits.AllDone();
		else
			return false;
	}
	
	// sets the Done bit in position pos to false
	public void ClearDoneBit(int pos)
	{
		if (DoneBits != null)
		{
			DoneBits.clear(pos);
		}
	}
	
	// sets the done bit in position pos to true
	public void SetDoneBit(int pos){
		if (DoneBits != null)
			DoneBits.set(pos);
		
	}
	
//////////////////////////////////////////////////////
	// sets the ready bit in position pos to false
	public void ClearReadyBit(int pos)
	{
		if (ReadyBits != null)
		{
			ReadyBits.clear(pos);
		}
	}
	
	// sets the ready bit in position pos to true
	public void SetReadyBit(int pos){
		if (ReadyBits != null)
			ReadyBits.set(pos);
		
	}

	
//////////////////////////////////////////////////////
	// returns the position of the first done bit 
	// not set to true
	public int NextNotDoneOp()
	{
		if (DoneBits != null)
			return DoneBits.FirstNotCompleted();
		else
			return -1;
	}

	// returns the position of the first ready bit 
	// set to true
	public int NextReadyOp()
	{
		if (ReadyBits != null)
			return ReadyBits.FirstReady();
		else
			return -1;
	}

//	// returns a random ready bit 
//	// set to true
//	public int NextRandomReadyOp()
//	{
//		if (ReadyBits != null)
//			return ReadyBits.RandomReady();
//		else
//			return -1;
//	}
	
	// returns the next operator with the highest# of tickets that is ready
	// to go
	public int NextReadyOpTicket(ArrayList<Double> OperatorTickets,ArrayList <ArrayList <Integer>> OperatorSTREAMIds)
	{
//		System.out.println("DoneBits????="+DoneBits.getSetBitsToString());
//		System.out.println("ReadyBits????="+ReadyBits.getSetBitsToString());
		
		int nextOp = 0, nextPos;
		double highestFound=0;
		nextPos = ReadyBits.getBits().nextSetBit(0);

		int nTotalTickets = 0;
		int nNumNeverUsed = 0;
		ArrayList<Integer> Ops = new ArrayList<Integer>();
		
		
		while (nextPos != -1)
		{	
			
			boolean bFound = false;
			int Pos;	
			Pos = DoneBits.getBits().nextSetBit(0);
						
			while ((Pos != -1) && !bFound)
			{
				
				if (OperatorSTREAMIds.get(nextPos-1).contains(Pos-1))
				  bFound = true;
				
				Pos = DoneBits.getBits().nextSetBit(Pos+1);				
			}
			
			if (bFound)
			{
			
				Ops.add(nextPos);
				if (OperatorTickets.get(nextPos-1)>0)
					nTotalTickets += OperatorTickets.get(nextPos-1);
				else
					nNumNeverUsed++;
			}	
//			System.out.println("Op = "+(nextPos-1)+" num tickets ="+OperatorTickets.get(nextPos-1));
				
			nextPos++;
			nextPos = ReadyBits.getBits().nextSetBit(nextPos);			
		}
		
	
		//let the number never called have at least dPercNeverUsed chance
		int nTicketNumNeverUsed = 0;
		if (nTotalTickets> 0)
		{	
			nTotalTickets = (int)(nTotalTickets/(1.0 - dPercNeverUsed * nNumNeverUsed));
			nTicketNumNeverUsed  = (int)(nTotalTickets * dPercNeverUsed); 
		}
		else
		{
			nTotalTickets = nNumNeverUsed;
			nTicketNumNeverUsed = 1;
		}
		
		int nNextTicketOp = nTotalTickets;
		if(Ops.size()>0)
		{	
		  nextPos = generator.nextInt(Ops.size());
		  nextPos=Ops.get(nextPos);
		}
		
//		System.out.println("nextPos = "+nextPos);
		
		while (nNextTicketOp > 0)
		{
			
//			System.out.println("nNextTicketOp = "+nNextTicketOp);
//			System.out.println("nextPos = "+nextPos);
			if (OperatorTickets.get(nextPos-1)>0)
				nNextTicketOp -= OperatorTickets.get(nextPos-1);
			else
				nNextTicketOp -= nTicketNumNeverUsed ;
			
			if (nNextTicketOp > 0)
			{	
				nextPos = generator.nextInt(Ops.size());
				nextPos=Ops.get(nextPos);
			}	
		}
		
		return nextPos;
	}
	
	// set the creationdate
	public void SetCreationDate(){
		Creation = Calendar.getInstance().getTime();
	}
	
	// returns the creationdate
	public Date GetCreationDate(){
		return Creation;
	}
	
	// adds a tuple to position pos
	public void addTuple(int pos,XATTuple t)
	{
		answer[pos] = t;         	
	}
	
	// returns a tuple from position pos
	public XATTuple getTuple(int pos)
	{
		return answer[pos];         	
	}
	

	// returns the number of tuples in the answer
	public int getNumTuplesInAnswer()
	{
		return NumTuplesInAnswer;         	
	}
/////////////////////////////////////////////////	
	// returns the DoneBits
	public BitStructure getDoneBitSet()
	{
		return DoneBits;
	}
	
	// sets the DoneBits
	public void setDoneBitSet(BitSet bs)
	{
		DoneBits.CreateCopyFrom(bs);
	}
//////////////////////////////////////////////////////
	// returns the ReadyBits
	public BitStructure getReadyBitSet()
	{
		return ReadyBits;
	}
	
	// set the ReadyBits
	public void setReadyBitSet(BitSet bs)
	{
		ReadyBits.CreateCopyFrom(bs);
	}

////////////////////////////////////////////////////	
	// returns the number of the Done Bits
	public int getSize()
	{
		return DoneBits.Size;
	}

	// copies the answer
	public XATTuple[] getAnswer()
	{
		return answer.clone();
	}

	// sets the answer
	public void setAnswer(XATTuple[] a)
	{
		answer = a;
	}
///////////////////////////////////////////////////////
	// get PartitionNumber
	public int getPartitionNumber()
	{
		return PartitionNumber;
	}

	// sets the PartitionNumber
	public void setPartitionNumber(int Pn)
	{
		PartitionNumber = Pn;
	}
	
/////////////////////////////////////////////////////
	// get NumberTuplesinBatch
	public int getNumberTuplesinBatch()
	{
		return NumTupBatch;
	}

	// sets the NumberTuplesinBatch
	public void setNumberTuplesinBatch(int n)
	{
		NumTupBatch = n;
	}
	
/////////////////////////////////////////////////////
////////////////////////////////////
	public long getSearchStart()
	{
		return SearchStart;
	}
	
	// clear the answer
	public void clearAnswer()
	{
		answer = null;
	}
	
}

