/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.eddyBatch.datastructure;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Calendar;
import java.util.Date;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;

//Karen Works

//This structure stores the information associated with 
//an eddy tuple and eddy tuple operations.

public class EddyTupleImp extends XATTupleImp implements EddyTuple, Serializable
{
	// intermediate tuple that will be the answer
	private XATTuple[] answer = null;
		
	// number of tuples in answer
	private int NumTuplesInAnswer;
	
	// creation date
	private Date Creation;
	
	//Partition Number
	private int PartitionNumber;
	
	public long SearchStart;
	
	public EddyTupleImp(XATTuple t,int NumStreams,int TuplePos,int PartitionNum,long lSearchStart) 
	{ //throws XATTimeStampException {
		super();
		
		if (t != null) 
		{
			// set up answer
			answer = new XATTuple[NumStreams];
			answer[TuplePos] = t;         
	
			// set time
			this.setMaxTimeStamp(t.getMaxTimeStamp());
			this.setMinTimeStamp(t.getMinTimeStamp());
			
			NumTuplesInAnswer = NumStreams;
			
			//set creation time
			Creation = Calendar.getInstance().getTime();
			
			//partition number
			PartitionNumber = PartitionNum;
			
			SearchStart = lSearchStart;
			
		} 
//		else {
//			//hello
//			System.out.println("tuple is null");
//		}
	}


	public EddyTupleImp(int NumStreams, int PartitionNum) 
	{ 
		//throws XATTimeStampException {
		super();
		
		answer = new XATTuple[NumStreams];
		
		NumTuplesInAnswer = NumStreams;
		
		//set creation time
		Creation = Calendar.getInstance().getTime();
		
		//partition number
		PartitionNumber = PartitionNum;
		
		SearchStart = System.currentTimeMillis();

	}
		
	// set the creationdate
	public void SetCreationDate(){
		Creation = Calendar.getInstance().getTime();
	}

	// returns the creationdate
	public Date GetCreationDate(){
		return Creation;
	}
	
	// adds a tuple to position pos
	public void addTuple(int pos,XATTuple t)
	{
		answer[pos] = t;         	
	}
	
	// returns a tuple from position pos
	public XATTuple getTuple(int pos)
	{
		return answer[pos];         	
	}
	

	// returns the number of tuples in the answer
	public int getNumTuplesInAnswer()
	{
		return NumTuplesInAnswer;         	
	}
	

	// copies the answer
	public XATTuple[] getAnswer()
	{
		return answer.clone();
	}
	
	// clear the answer
	public void clearAnswer()
	{
		answer = null;
	}

	// sets the answer
	public void setAnswer(XATTuple[] a)
	{
		answer = a;
	}

	// get PartitionNumber
	public int getPartitionNumber()
	{
		return PartitionNumber;
	}

	// sets the PartitionNumber
	public void setPartitionNumber(int Pn)
	{
		PartitionNumber = Pn;
	}

////////////////////////////////////
	public long getSearchStart()
	{
		return SearchStart;
	}
	
	
}

