/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.eddyBatch.operator;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.*;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import java.util.Vector;

import edu.wpi.cs.dsrg.STeM.Partition.STeMPartitionData;
import edu.wpi.cs.dsrg.STeM.Partition.Partition;
import edu.wpi.cs.dsrg.STeM.Partition.PartitionListQueue;
import edu.wpi.cs.dsrg.eddy.datastructure.EddyQueue;
import edu.wpi.cs.dsrg.eddyBatch.datastructure.*;

import java.util.BitSet;
import java.util.Iterator;
import java.util.StringTokenizer;
//Karen Works

import org.w3c.dom.Node;

import edu.wpi.cs.dsrg.STeM.Partition.IMPIndexHeuristic.*;

//This is the eddy operator 

public class EddyPartitionOperatorImp extends XATMultiQueueWindowStreamOperatorImp implements EddyPartitionOperator
{
	//public class EddyOperatorImp extends XATSingleQueueStreamOperatorImp implements EddyOperator{
	private ExpressionStr expression;
	//public int windowSize;
	
	public long beginingTime = -99;

	// number of operators
    private int NumOps = 0;
    
    //number of streams
    private int NumSTREAMS = 0;
    
	// number of bits
    private int NumBits = 0;

	// number of join atrib
    private int NumJoinAttribs = 0;
    
    // number of tuples to process from pool
    private int NumTupToProcessPerOp = 200;
    private int NumTupToProcess = 5000;
    private int NumLevelsToProcess = 2;
    
	// Tuples that require processing
	private HashMap<Integer, ArrayList<EddyQueue>> TuplePool;
	
	//operator bitmaps per STREAM
	private ArrayList <BitSet> PreFilledPerStream;
	
	// pointer to data stored for STEM
	private HashMap <Integer,STeMPartitionData> STeMStreamData;

//	// array of operator's STREAM Ids
	private ArrayList <ArrayList <Integer>> OperatorSTREAMIds;
	
	// array of operator's STEM Data STREAM Id
//	private ArrayList <Integer> OperatorSTEMDataSTREAMIds;
	
	//router
	private String strRouter= "Simple";

	// array of operator's tickets for back pressure
	private ArrayList <Double> OperatorTickets;

	//used for picking next tuple to run

	//are you going to flip flop strategies?
	boolean bFlip = true;
	//is looking at the top the current strategy
	boolean bTop = false;
	
	// stores window count for streams
	private HashMap <Integer,Integer> STREAMWindowCount;
	
	/*****************************************/	
	//partition data
	
	// maximum number of tuples in a partition
	int MAX_PARTITION_SIZE = 800;
	
	// maximum number of partitions in a window
	int MAX_PARTITION_WINDOW_SIZE = 2;
	
	//current Partition Number
	int currentPartitionNumber = 1;
	
	//is partitioned?
	boolean bIsPartitioned = true;
	
	//earliest Partition Number
	int earliestPartitionNumber = 1;
	
	//number of tuples stored in current partition
	int numTupCurrentPartitionNumber = 1;

	// stores the number of processing tuples by partition
	private boolean bCountActiveProcTups = true;
	private HashMap <Integer,Integer> PartitionProcTupCount;
	
	/*****************************************/
	boolean bSwitchIndex =true;
	
	//for static testing of index adaption only
	int nTotalNumberTupsProcessed = 0;

	/*****************************************/
	//use IMPIndex? No - hash index
	boolean bIMPIndex = true;

	//numbr of limited indices
	int nNumLimitedIndices = 0;
	
//	int nNumAttrib = 5;
	
	/*****************************************/
	//Index Assessment
	boolean bIndexAssessment;
		
	/*****************************************/
	//Exhaustive Index Assessment
	boolean bExhaustiveAssessment = true;
	
	/*****************************************/
	//Periodic Index Assessment
	boolean bPeriodicAssessment;
	int nPeriodicAssessmentWindow;
	int nPeriodicStreamId = 0;
	int nCurrentPeriodicAssessmentTupleCount = 0;
	
	/*****************************************/
	//Continuous Index Assessment
	boolean bContinuousAssessment;
	
	/*****************************************/
	//triggered Index Assessment
	boolean bTriggeredAssessment;
	int nTotalTuplesTriggeredAssessment = 0;
	int nTriggeredAssessmentWindow;
	

	//array of streams that need to assessed
	private ArrayList <Boolean> StreamAssessPos;
	
/*****************************************/
//Index Assessment Sampling
	boolean bNoSampling = true;
	/////////////////////////////////////////
	boolean bSimpleRandomSampling;
	int nSimpleRandomPercent;    //value between 0-100
	////////////////////////////////////////////////
	boolean bStratifiedRandomSampling;
	int nStratifiedRandomPercent; //value between 0-100 
	int nStratifiedRandomWindow;
	/////////////////////////////////////
	boolean bSystematicSampling;
	int nSystematicInterval; 
	
	
	/////////////////////////////////////////
	//Deletion type
	boolean bWindowDeletion = false;
	boolean bActProcTupsDeletion = true;
	
	
	private ArrayList<Integer> TriggeredStateAssessment;
	private ArrayList<Integer> CurrentTriggeredStateAssessment;
	private ArrayList<Integer> PrevTriggeredStateAssessment;
	//////////////////////////////////
/*****************************************/
//Index Migration
	private boolean bNoMigration=false;
	private boolean bSymmetricMigration=true;
	private boolean bPartitionMigration=false;
	private boolean bCompleteMigration=false;
/*****************************************/
//SRIA
	private boolean bSRIA;
	private boolean bHH;
	
	private double dHHErrorRate;
	private double dCostFactorSignificance;
	
//HHH
	private boolean bHHH;
	private boolean bCombineNodesByPickingOne;
	private boolean bPickParentWithHighestFreq;

/*****************************************/
//Combined stats ?
	private boolean bCombinedStats;
	private int nNumCombinedStats;
	private double[] CombinedStatsFactor;
	
	/*****************************************/
	//Index Selection trigger
//	private boolean bNoIndexSelectionTrigger=true;
//	private boolean bCostIncrIndexSelectionTrigger=false;
//	private boolean bFreqIncrIndexSelectionTrigger=false;
//	private boolean bCombinedIncrIndexSelectionTrigger=false;
		
	private double dDegreeCostIncr = 0.0;
	private double dDegreeFreqIncr =0.0;

	/*****************************************************/
	private boolean bTrackTimeToSearch = false;
	private String strTrackTimeToSearchOutputfile;
	
	/*****************************************************/
	private boolean bTrackOptimalIndices = false;
	private String strTrackOptimalIndicesOutputfile;
	
	/*****************************************/
	//count migrations
	private boolean bCountMigrations=false;
	private String strCountMigration="";

	/*****************************************/
	//ComparisonTime
	private boolean bStoreComparisonTime = false;
	private String strStoreComparisonTimeFile;
		
	//TrackCompletedPartitions
	boolean bTrackCompletedPartitions = false;
	String outFile = "EDDY_RESULTS.txt";
	
//Count tuples incoming from streams
	ArrayList<Integer> NumIncomingTuples;
	
	public EddyPartitionOperatorImp(XATNode init) 
	{
		super(init);
		
		STREAMWindowCount = new HashMap <Integer,Integer>();
	}

	public void InitializeVariables()
	{
		//set up the PreFilledPerStream array
		PreFilledPerStream = new ArrayList(NumSTREAMS-1);
	
		NumIncomingTuples = new ArrayList(NumSTREAMS);
		
		PartitionProcTupCount= new HashMap <Integer,Integer>();
	
		if(bIndexAssessment)
		{	
			StreamAssessPos  = new ArrayList(NumSTREAMS-1);
	
			if (bTriggeredAssessment)
			{	
				//FreqSTeMOp = new ArrayList(NumOps);
				TriggeredStateAssessment = new ArrayList<Integer>();
				
			}
		}
		
		try
		{
			//intialize tickets
			if (strRouter.compareTo("Ticket") == 0)
			{	
				OperatorTickets = new ArrayList(NumOps);
				//initialize tickets
				for (int O = 0;O<NumOps;O++)
				{
					OperatorTickets.add(0.0);
				}
			}
			
			//initializes basic bitsets for each streamid
			for (int i = 0; i <  NumSTREAMS ; i++)
			{
				
				BitSet r = new BitSet(NumOps+1);
				
				NumIncomingTuples.add(0);
				
				if(bIndexAssessment)
				  StreamAssessPos.add(true);
				
				for (int j=0; j < OperatorSTREAMIds.size();j++)
				{
					if (OperatorSTREAMIds.get(j).contains(i))				
						r.set(j+1, true);						
					else							
						r.clear(j+1);
					
				}//end for
//				
//				System.out.println("For stream "+i+" bit set = "+r.toString());
				
				PreFilledPerStream.add(r);
				
				if(bIndexAssessment)
				{	
				  if (bTriggeredAssessment)			
				  {
					TriggeredStateAssessment.add(0);
				  }
				}  
			}
			
		}
		catch (Exception ex)
		{
			System.out.print("Error setting up EddyOperator!!");
		}  
	}
	
	@SuppressWarnings("unchecked")
	public boolean run() throws Exception 
	{		
		for (int i=0; i < this.getOutputQueueCount(); i++)
		{
			if(this.getOutputQueue(i).getSchema() == null)
			{
				init(i);
			}
		}

///////////////////////////////////////////////
//for testing only
//		if (bCountActiveProcTups)
//		{	
//			boolean bFirst =true;
//			
////			System.out.println("earliestPartitionNumber="+earliestPartitionNumber);
////			System.out.println("currentPartitionNumber="+currentPartitionNumber);
//			
//			for (int i=earliestPartitionNumber; i <= currentPartitionNumber; i++)
//////			for (int i=earliestPartitionNumber; i <= 10; i++)
//			{
//				if (PartitionProcTupCount.get(i) != null)					
//				{
//					if(bFirst)
//					{
//						System.out.println("CountActiveProcTups");
//						bFirst=false;
//					}
//					System.out.println("Partition "+i+"="+PartitionProcTupCount.get(i));
//				}
//			}
//			if(!bFirst)
//				System.out.println();
//		}
///////////////////////////////////////////////
		
		//stream partition based
		//optimal operator per STREAM
		HashMap<Integer,ArrayList<EddyQueue>> EddyQueueByOp = null;
		HashMap<Integer,HashMap<String, EddyQueue>> EddyQueueByStreamByBitMap;
		//EarliestPartitionNumChgd?
		boolean bEarliestPartitionNumChgd = false;
		
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		//int howMany  = 500; //rimma temp "hack"
		
		//-------------------------------------------------------------------------
		// FIRST PROCESS THE TUPLES FROM OPERATORS INPUT QUEUES
		//-------------------------------------------------------------------------
		//Scan for Other Operators & STeM's returning Tuples.
		for (int j = NumSTREAMS; j <= (NumSTREAMS + NumOps - 1); j++) 
		{			
			howMany =this.getInputQueue(j).size();
			XATTuple[] tuples = (this.getInputQueue(j)).dequeue(howMany, this);
			EddyBatchTuple ebt = null;
			EddyTuple et = null;
			
//			if (tuples.length >0)
//			{
//				System.out.println("\nProcessing operator queue [" + (j- NumSTREAMS) + "]");
//				System.out.println("\t# of tuples in the queue: " + tuples.length + " tuples.");
//				System.out.println("");
//			}
			
			//Eddy Batching variables
			int NumTupInBatch = 0;
			boolean bBatchTup = true,bDone = false;
			EddyQueue Eq = new EddyQueue();
			int key = -1;
			
			//for each tuple in the stream from the op
			for(int k = 0; k < tuples.length;k++)
			{
			  //Count all active processing tuples
			  if (bCountActiveProcTups && tuples[k] instanceof EddyActiveProcTuple)
			  {
//				  System.out.println("recieving active proc tuple count");
				  EddyActiveProcTuple eat = (EddyActiveProcTupleImp)tuples[k];
				  HashMap <Integer,Integer> P = eat.GetPartitionProcTupCount();
				
//				  System.out.println("earliestPartitionNumber="+earliestPartitionNumber+" currentPartitionNumber = "+currentPartitionNumber);
				  
				  //update partition counts
					Iterator it = P.keySet().iterator();
					while (it.hasNext()) {
						Integer okey = Integer.valueOf(it.next().toString());
						if ((P.get(okey) != null) && (PartitionProcTupCount.get(okey) != null))					
						{
//							System.out.println();
//							System.out.println("before i="+okey+" Count = "+PartitionProcTupCount.get(okey));
							PartitionProcTupCount.put(okey,PartitionProcTupCount.get(okey)+P.get(okey));
//							System.out.println("after i="+okey+" Count = "+PartitionProcTupCount.get(okey));
							
						}
					  }				  
			  }
			  else
			  {		
				//test to see if this is the EddyBatch tuple
				if(NumTupInBatch == 0)
				{	
					ebt = (EddyBatchTuple)tuples[k];
					NumTupInBatch = ebt.getNumberTuplesinBatch();
					bBatchTup = true;
					bDone = false;
				}	
				else //otherwise it is a standard eddy tuple
				{	
					et = (EddyTuple)tuples[k];
					bBatchTup = false;
				}
				
				//keep count of the number in each batch
				NumTupInBatch--;
				
				//keep track of the tickets
				if (strRouter.equalsIgnoreCase("Ticket"))
				{	
					OperatorTickets.set(j - NumSTREAMS, OperatorTickets.get(j - NumSTREAMS) + 1);										
				}
				
				//do all testing on the batch tuple
				// every tuple in the batch follows the same path
				if(bBatchTup)
				{	

					
					//--------------------------------------------------------------------
					// Tuple is ALL DONE
					//--------------------------------------------------------------------				
					// check to see if the tuple is done
					// ie all ops done bits set to true
					if (ebt.isDone() == true) 
					{		
//						//System.out.println("Send to output");
						// if it is done, output result
						//Assume last output queue  is the result queue
						this.getOutputQueue(NumOps).enqueue(ebt, this);

						//Count all active processing tuples
						if (bCountActiveProcTups)
						{
//							if(ebt.getPartitionNumber()<=6)
//							{	
//								System.out.println("------Complete-------");
//								System.out.println("Before PartitionProcTupCount.get(i) = "+PartitionProcTupCount.get(ebt.getPartitionNumber()));
//								System.out.println("Num Complete = "+ebt.getNumberTuplesinBatch());
//								System.out.println("Partition = "+ebt.getPartitionNumber());
//							}
//							System.out.println();
//							System.out.println("before i="+ebt.getPartitionNumber()+" Count = "+PartitionProcTupCount.get(ebt.getPartitionNumber()));
							
							if (PartitionProcTupCount.get(ebt.getPartitionNumber()) != null)
								PartitionProcTupCount.put(ebt.getPartitionNumber(),PartitionProcTupCount.get(ebt.getPartitionNumber())-1);
							
//							System.out.println("after i="+ebt.getPartitionNumber()+" Count = "+PartitionProcTupCount.get(ebt.getPartitionNumber()));
						}
						
						ebt = null;
						bDone = true;						
					}
					else
					{
						//System.out.println("Adding Batch tuple to tuple pool");
						//--------------------------------------------------------------------
						// Tuple is NOT DONE
						//--------------------------------------------------------------------						
						// not done --> add to tuple pool for more processing
						key = ebt.NumDoneBitsComplete();
						
						//System.out.println("Key = "+key);
						
						//create new Eddy Queue
						Eq = new EddyQueue();
						Eq.enqueue(ebt);

						//create tuple pool and add batch
						if(TuplePool != null)
						{
							
							if(TuplePool.get(key) != null)
							{
								ArrayList<EddyQueue> e = TuplePool.get(key);
								e.add(Eq);
							}
							else
							{
								ArrayList<EddyQueue> e = new ArrayList<EddyQueue>(); 
								e.add(Eq);	
								TuplePool.put(key, e);
							}					
						}
						else
						{
							TuplePool = new HashMap();
							ArrayList<EddyQueue> e = new ArrayList<EddyQueue>(); 
							e.add(Eq);	
							TuplePool.put(key, e);
								
						}
					}									
				}
				else
				{ // eddy tuple follow path of the eddy batchtuple
					if (bDone)
					{
						if (et != null)
						{	
							//CountActiveProcTups
							if (bCountActiveProcTups)
							{
//								System.out.println();
//								System.out.println("before i="+et.getPartitionNumber()+" Count = "+PartitionProcTupCount.get(et.getPartitionNumber()));

								if (PartitionProcTupCount.get(et.getPartitionNumber()) != null)
									PartitionProcTupCount.put(et.getPartitionNumber(),PartitionProcTupCount.get(et.getPartitionNumber())-1);
	
//								System.out.println("after i="+et.getPartitionNumber()+" Count = "+PartitionProcTupCount.get(et.getPartitionNumber()));
							}
							
							// if it is done, output result
							//Assume last output queue  is the result queue
							this.getOutputQueue(NumOps).enqueue(et, this);
							
							et = null;						
						}
					}
					else
					{						
						//System.out.println("Adding tuple to EddyQueue");
						TuplePool.get(key).get(TuplePool.get(key).size()-1).enqueue(et);
					}
				}
			  }	
			}//end for iterating over all tuples			
		}//end for - retrieved tuples from operator queues and put them into the tuple pool		
		
//////////////////////////////////////////////
		//delete tuples from partitions
			if (PartitionProcTupCount.get(earliestPartitionNumber) != null)
				if ((PartitionProcTupCount.get(earliestPartitionNumber)==0) &&(earliestPartitionNumber != currentPartitionNumber)) 
				{			
					boolean bLoop = true;
					
					for (int i=earliestPartitionNumber; i < currentPartitionNumber && bLoop; i++)
					{
						if (PartitionProcTupCount.get(i) != null)
						{	
							if (PartitionProcTupCount.get(i)==0)
							{
//								//Partition completed
//								if(bTrackCompletedPartitions)
//								{	
//									if(	beginingTime == -99)
//									{	
//										beginingTime =System.currentTimeMillis();
//										WriteToFile(outFile, String.valueOf(beginingTime), false);
//										WriteToFile(outFile, String.valueOf(MAX_PARTITION_SIZE), false);
//										WriteToFile(outFile, "", false);
//										WriteToFile(outFile, String.valueOf(0), true);
//									}
//									else
//									{
//										long currentTime =System.currentTimeMillis();
//										WriteToFile(outFile, String.valueOf(currentTime), false);				
//										WriteToFile(outFile, String.valueOf(MAX_PARTITION_SIZE), false);
//										WriteToFile(outFile, "", false);
//										WriteToFile(outFile, String.valueOf(currentTime-beginingTime), true);
//										
//									}
//								}

								PartitionProcTupCount.remove(i);
								bEarliestPartitionNumChgd = true;												
								earliestPartitionNumber++;
								
								i++;
								
								while((PartitionProcTupCount.get(earliestPartitionNumber)== null) &&(earliestPartitionNumber <currentPartitionNumber))
								{	
//									//Partition completed
//									if(bTrackCompletedPartitions)
//									{	
//										if(	beginingTime == -99)
//										{	
//											beginingTime =System.currentTimeMillis();
//											WriteToFile(outFile, String.valueOf(beginingTime), false);
//											WriteToFile(outFile, String.valueOf(MAX_PARTITION_SIZE), false);
//											WriteToFile(outFile, "", false);
//											WriteToFile(outFile, String.valueOf(0), true);
//										}
//										else
//										{
//											long currentTime =System.currentTimeMillis();
//											WriteToFile(outFile, String.valueOf(currentTime), false);				
//											WriteToFile(outFile, String.valueOf(MAX_PARTITION_SIZE), false);
//											WriteToFile(outFile, "", false);
//											WriteToFile(outFile, String.valueOf(currentTime-beginingTime), true);
//											
//										}
//									}

									earliestPartitionNumber++;
									i++;
								}					
							}	
							else
								bLoop = false;
						}
						else
							bLoop = false;
				  }
				}	
		

//////////////////////////////////////////////
	//-------------------------------------------------------------------------
	// SECOND PROCESS THE TUPLES FROM THE INPUT STREAMS
	//-------------------------------------------------------------------------
	/** get the tuples from STREAM*/
	boolean bMore = true;
	boolean bF = true;
	while(bMore)
	{	
	    for (int j = 0; j < (NumSTREAMS); j++) 
	    {
		    //purge stored tuples
	    	if(bF)
				if(bActProcTupsDeletion && bEarliestPartitionNumChgd)
				{
//					 System.out.println("Purging"); 
					 STeMPartitionData sd = STeMStreamData.get(j);
					 if (sd != null)
						sd.PurgeTuple(earliestPartitionNumber);
				}

	    	//howMany  = 2500; //rimma temp "hack"			
		  if((this.getInputQueue(j)).getQueueSize() >= (MAX_PARTITION_SIZE/NumSTREAMS))			  
		    howMany = MAX_PARTITION_SIZE/NumSTREAMS;
		  else
		  {	  
			howMany =(this.getInputQueue(j)).getQueueSize();
			bMore = false;
		  }
		  
		
//		   System.out.println("\nProcessing stream queue [" + j + "]");
//		   System.out.println("\t# of tuples in the queue: " + (this.getInputQueue(j)).getQueueSize());
		
		   XATTuple[] tuples = this.getInputQueue(j).dequeue(howMany, this);
//		System.out.println("\tDequeued: " + tuples.length + " tuples.");
//		System.out.println("");
		
		EddyTuple et;
		EddyQueue Eq = new EddyQueue();
		// process all new tuples coming from STREAMS
		for(int k = 0; k < tuples.length;k++)
		{				
			/////////////////////////////////////
			//need new partition?
			if ((numTupCurrentPartitionNumber % MAX_PARTITION_SIZE) == 0)
			{			
				currentPartitionNumber++;
				numTupCurrentPartitionNumber = 0;
			}
			
			//update partition
			numTupCurrentPartitionNumber++;
			
//			System.out.println("currentPartitionNumber: " + currentPartitionNumber);
//			System.out.println("numTupCurrentPartitionNumber: " + numTupCurrentPartitionNumber);
							
			/////////////////////////////////////
			
			//add partition number to the tuple
			tuples[k].addCell(new XATDataValueImp(new Integer(currentPartitionNumber)));
			
			//updating tuple time counter
			NumIncomingTuples.set(j, NumIncomingTuples.get(j)+1);			 
			tuples[k].setCell(tuples[k].getLength() - 2, new XATDataValueImp(new Integer(NumIncomingTuples.get(j))));
			

			// create eddy batch tuple for first tuple
			if(k==0)
			{
				BitSet rbs = (BitSet)PreFilledPerStream.get(j).clone();
				
				EddyBatchTuple ebt = new EddyBatchTupleImp(tuples[k], rbs, NumOps,NumSTREAMS, NumSTREAMS, j, currentPartitionNumber,System.currentTimeMillis());
				ebt.setStreamID(j);
				//set op associated w/stream to done
				ebt.SetDoneBit(j+1);
				ebt.setNumberTuplesinBatch(tuples.length);
				Eq = new EddyQueue();
				Eq.enqueue(ebt);

				//Count all active processing tuples
				if (bCountActiveProcTups)
				{	
					if (PartitionProcTupCount.get(currentPartitionNumber) != null)
						PartitionProcTupCount.put(currentPartitionNumber,PartitionProcTupCount.get(currentPartitionNumber)+1);
					else
						PartitionProcTupCount.put(currentPartitionNumber,1);							
				}
				
			}
			else //otherwise create eddy tuple
			{	
//				System.out.println("tuples[k]: " + tuples[k]);

				et = new EddyTupleImp(tuples[k], NumSTREAMS, j, currentPartitionNumber,System.currentTimeMillis());
				et.setStreamID(j);
				Eq.enqueue(et);
				
				//Count all active processing tuples
				if (bCountActiveProcTups)
				{	
					if (PartitionProcTupCount.get(currentPartitionNumber) != null)
						PartitionProcTupCount.put(currentPartitionNumber,PartitionProcTupCount.get(currentPartitionNumber)+1);
					else
						PartitionProcTupCount.put(currentPartitionNumber,1);							
				}
				
			}				
		}
		
		//add eddy queue to tuple pool
		if(Eq.isEmpty()==false)
		{	
			if(TuplePool != null)
			{
				if(TuplePool.get(0) != null)
					TuplePool.get(0).add(0,Eq);
				else
				{
					ArrayList<EddyQueue> al = new ArrayList<EddyQueue>();
					al.add(Eq);
					TuplePool.put(0,al);					
				}
			}
			else
			{
				TuplePool = new HashMap();
				ArrayList<EddyQueue> al = new ArrayList<EddyQueue>();
				al.add(Eq);
				TuplePool.put(0, al);					
			}
		}	
		
	}	
	    bF = false;
}
		
	//set up processing arrays 	
		ArrayList<Boolean> OpFull = new ArrayList<Boolean>(NumOps);
		ArrayList<Integer> OpNumSent = new ArrayList<Integer>(NumOps);
		
		long lTotal=0;		
		for (int g = 0; g < NumOps; g++)
		{	
			OpFull.add(false);
			OpNumSent.add(0);
			
/////////////////////////////////////////////			
//			//for testing only
//			if (TuplePool!= null)				
//			{				  
//				if ((TuplePool.get(g)!= null) && (TuplePool.get(g).size()>0))
//				{
//					int NumTups = 0;
//					for (int j = 0; j<TuplePool.get(g).size();j++)
//					{
//						EddyQueue Eq = TuplePool.get(g).get(j);
//						NumTups += Eq.size();						
//					}
//					
//					//System.out.println(g+". at level (# of operators done): "+g+" there are "+TuplePool.get(g).size() + " entries in the arraylists of tuple batches size = "+NumTups);
//					System.out.println(g+" has "+NumTups);
//					lTotal+=NumTups;
//				}
//				else
//				{
//					System.out.println(g+". TuplePool is null");
//				}
//			}
//			if (g == NumOps-1)
//			  System.out.println(" Total num Proc Tuples = "+lTotal);
//
/////////////////////////////////////////////			
		}	

		//for testing only
			
			System.out.println("earliestPartitionNumber="+earliestPartitionNumber);
			System.out.println("currentPartitionNumber="+currentPartitionNumber);
			
			System.out.println();
			System.out.println("CountActiveProcTups");
			for (int i=earliestPartitionNumber; i <= currentPartitionNumber; i++)
////			for (int i=earliestPartitionNumber; i <= 10; i++)
			{
				if (PartitionProcTupCount.get(i) != null)					
				{
					System.out.println("Partition "+i+"="+PartitionProcTupCount.get(i));
				}
			}
			System.out.println();
		
		int NumOpsFull = 0;
		int NumSent = 0;
		int NumLevelsProcessed = 0;
		
		if(TuplePool != null)
		{	
			// look at each op to see if any ops have tuple returning 
			// to the eddy operator for processing
			if (bTop)
			{	
//				System.out.println("\nUsing top strategy...");
				
				int numTuplesSent = 0;
				Hashtable hashTuplesSentByOp = new Hashtable();
				
				//stream based partition
				EddyQueueByOp = new HashMap<Integer,ArrayList<EddyQueue>>();
				EddyQueueByStreamByBitMap = new HashMap<Integer,HashMap<String, EddyQueue>>();
				boolean bQuit = false;

				int nNumNewTupsProc = 0;
				//for each level in the tuple pool
				for (int k = (NumOps-1);(k >= 0) && (nNumNewTupsProc<=(MAX_PARTITION_SIZE)*4) && (NumOpsFull <= NumOps) && (NumSent < NumTupToProcess) && (NumLevelsProcessed <= NumLevelsToProcess) && !bQuit; k--)
				{
					if (TuplePool.get(k)!= null)
					{
						if (TuplePool.get(k).size() > 0)
						{
//							  System.out.println("Processing level [" + k + "]");
//							  System.out.println("NumOpsFull =" + NumOpsFull);
//							  System.out.println("NumLevelProcessed =" + NumLevelsProcessed);
//							  System.out.println("NumSent =" + NumSent);
//								for (int g = 0; g < NumOps; g++)
//								{
//									if (OpNumSent.get(g)>0)
//									  System.out.println("OP = "+g+" Num sent ="+OpNumSent.get(g)+" OpFull ="+OpFull.get(g));
//								}	
							
							NumLevelsProcessed++;
					
							ArrayList<EddyQueue> al = TuplePool.get(k);
							int NumItems = al.size();
							//process each EddyQueue at this level
							
							for (int j = NumItems-1; (j >= 0) &&  (NumOpsFull < NumOps) && (NumSent < NumTupToProcess) && (nNumNewTupsProc<=(MAX_PARTITION_SIZE)*4) && !bQuit; j--)
							{	
								// there are items in the queue to process
								EddyQueue Eq = al.remove(j);
								if(!Eq.isEmpty())
								{	
									int count = 0;
									int NumInQ = Eq.length();
									EddyBatchTuple BatchTupletoProcess;
									EddyTuple TupletoProcess;
									
									//remove the batch tuple
									BatchTupletoProcess = (EddyBatchTuple)Eq.dequeue();
									count++;
									
									//determine which op this tuple should go to
									int nextOp = 0;
									
									if (strRouter.compareTo("Simple") == 0) 
										nextOp = BatchTupletoProcess.NextReadyOp();									
									else if (strRouter.compareTo("Ticket") == 0)
										nextOp = BatchTupletoProcess.NextReadyOpTicket(OperatorTickets, OperatorSTREAMIds);
									
//									System.out.println("nextOp="+(nextOp));
									
									//create output queues for tuples in batches
									if (EddyQueueByStreamByBitMap.get(BatchTupletoProcess.getStreamID()) == null)
										EddyQueueByStreamByBitMap.put(BatchTupletoProcess.getStreamID(), new HashMap<String, EddyQueue>());
								
									EddyQueue q; 												
									if (EddyQueueByOp.get(nextOp) == null)
										EddyQueueByOp.put(nextOp, new ArrayList<EddyQueue>());
									
									//create temp queues to add tuples to
									if (EddyQueueByStreamByBitMap.get(BatchTupletoProcess.getStreamID()).get("Op = ["+nextOp+"]"+BatchTupletoProcess.getDoneBitSet().getSetBitsToString()) == null)
									{	
										q  = new EddyQueue();
										EddyQueueByStreamByBitMap.get(BatchTupletoProcess.getStreamID()).put("Op = ["+nextOp+"]"+BatchTupletoProcess.getDoneBitSet().getSetBitsToString(), q);
										EddyQueueByOp.get(nextOp).add(q);
									}
									else
									{	
										q = EddyQueueByStreamByBitMap.get(BatchTupletoProcess.getStreamID()).get("Op = ["+nextOp+"]"+BatchTupletoProcess.getDoneBitSet().getSetBitsToString());  
									}
									
									//add tuple to appropriate output queue
//									System.out.println("nextOp " + (nextOp-1));
									if (!OpFull.get(nextOp-1))
									{													
											//new tuples add to STEm data
											if (k == 0)
											{
												nNumNewTupsProc++;
												//System.out.println("StreamID="+BatchTupletoProcess.getStreamID());
												//if STeMData exists for STREAMId send tuple to STeMData
												STeMPartitionData sd = STeMStreamData.get(BatchTupletoProcess.getStreamID());
												sd.Insert(BatchTupletoProcess.getTuple(BatchTupletoProcess.getStreamID()));
													
												BatchTupletoProcess.SetCreationDate();
											}
												
											NumSent++;
												
											int n=1;
											//add tuple to queue to send
											if (EddyQueueByStreamByBitMap.get(BatchTupletoProcess.getStreamID()).get("Op = ["+nextOp+"]"+BatchTupletoProcess.getDoneBitSet().getSetBitsToString()).length()==0)
												EddyQueueByStreamByBitMap.get(BatchTupletoProcess.getStreamID()).get("Op = ["+nextOp+"]"+BatchTupletoProcess.getDoneBitSet().getSetBitsToString()).enqueue(BatchTupletoProcess);
											else
											{	
												EddyTuple et = new EddyTupleImp(NumSTREAMS, BatchTupletoProcess.getPartitionNumber());
												et.setAnswer(BatchTupletoProcess.getAnswer());
	
												EddyQueueByStreamByBitMap.get(BatchTupletoProcess.getStreamID()).get("Op = ["+nextOp+"]"+BatchTupletoProcess.getDoneBitSet().getSetBitsToString()).enqueue(et);													
											}
											
	
											// add each tuple in queue to 
											// the appropriate op queue
											while(Eq.isEmpty() == false) 											
											{							
												if (k==0)
												  nNumNewTupsProc++;
												EddyTuple et  = (EddyTuple)Eq.dequeue();
												EddyQueueByStreamByBitMap.get(BatchTupletoProcess.getStreamID()).get("Op = ["+nextOp+"]"+BatchTupletoProcess.getDoneBitSet().getSetBitsToString()).enqueue(et);
													
												NumSent++;																							
												n++;
													
												//new tuples add to STEm data
												if (k == 0)
												{												
													//if STeMData exists for STREAMId send tuple to STeMData
													if (STeMStreamData != null)
													{	
														//System.out.println("\nAdding tuple to STEM Data");
														STeMPartitionData sd = STeMStreamData.get(et.getStreamID());
														if (sd != null)
															sd.Insert(et.getTuple(et.getStreamID()));
														
														et.SetCreationDate();
													}
												}													
											}
												
											OpNumSent.set(nextOp-1, OpNumSent.get(nextOp-1) + n);
//											System.out.println("sending"+OpNumSent.get(nextOp-1));
																						
											if (hashTuplesSentByOp.get(nextOp-1) == null)
												hashTuplesSentByOp.put(nextOp-1, 1);
											else
											{
												int countTuples = Integer.parseInt(hashTuplesSentByOp.get(nextOp-1).toString()) +  n;
												hashTuplesSentByOp.put(nextOp-1, countTuples);
											}
												
											if (strRouter.compareTo("Ticket") == 0)
												OperatorTickets.set(nextOp-1, OperatorTickets.get(nextOp-1)+n);
											
											//Is op full?	
											if ((OpNumSent.get(nextOp-1) >= NumTupToProcessPerOp) && !OpFull.get(nextOp-1)) 
											{	
												OpFull.set(nextOp-1, true);
												NumOpsFull++;
											}
//										}
									}
									else
									{
										  if(!Eq.isEmpty())
										  {	
											  Eq.enqueueFront(BatchTupletoProcess);
											  al.add(Eq);
											  
											  if(k == 0)
												bQuit = true;
										  }
									}																
								}
								
//								if(al.size() <=0)
//									TuplePool.remove(k);

							}
						}
					}
				}					
				
				EddyQueueByStreamByBitMap = null;
			}//end if "top strategy"
			//look at the queue that is the most full	
		}
		
		
		//-------------------------------------------------------------------------
		// Loop through batched tuples and send them to ops
		//-------------------------------------------------------------------------
		OpFull = null;
		OpNumSent = null;
		
		if (EddyQueueByOp != null)
		{	
			for (int g = 1; g <= NumOps; g++)
			{	
				if (EddyQueueByOp.get(g) != null)
				{	
					ArrayList<EddyQueue> alQ  = ((ArrayList<EddyQueue>)EddyQueueByOp.remove(g));
					int numInList = alQ.size();
					EddyBatchTuple ebt = null;
					boolean bFirst = true;					
					for(int t = 0;t<numInList;t++)
					{
						EddyQueue q = alQ.remove(0);
						int len = q.length();
						bFirst = true;
						while(len>0)
						{	

							if(bFirst)
							{
								ebt = (EddyBatchTuple)q.dequeue();
								len = q.length()+1;
								
//								System.out.println("Sending "+len+" tuples to "+g+" Op");
								
								ebt.setNumberTuplesinBatch(len);
								this.getOutputQueue(g-1).enqueue(ebt, this);	
								
								/////////////////////////////////////////
								if (bTriggeredAssessment)
								{
									nTotalTuplesTriggeredAssessment+=len;
									TriggeredStateAssessment.set(g-1, TriggeredStateAssessment.get(g-1)+len); 
									
								}	
								/////////////////////////////////////////
								
								bFirst=false;
								len --;
							}
							else
							{	
								EddyTuple et = (EddyTuple)q.dequeue();
								this.getOutputQueue(g-1).enqueue(et, this);
//								System.out.println("Sending "+et.toString());
								
								len --;
							}
								
						}
					}
					alQ=null;
				}				
			}			
			EddyQueueByOp = null;
		}
		
		
		
//	}
//		//reset EarliestPartitionNumChgd
//		bEarliestPartitionNumChgd = false;

/////////////////////////////
//Index Assessment
		if(bIndexAssessment)
		{	
			// Triggered Assessment
			if(bTriggeredAssessment)
			{
				for (int s =0;s<NumSTREAMS;s++)
				{					
					if (s != nPeriodicStreamId)
					{	
						if (STeMStreamData.get(s).getTotalNumQueryRequests()>=nPeriodicAssessmentWindow)
						{	
							if(bTrackTimeToSearch)
								STeMStreamData.get(s).endExhaustiveAssessment(NumBits,strTrackTimeToSearchOutputfile);
							else
								STeMStreamData.get(s).endExhaustiveAssessment(NumBits);
						}
					}	
				}
				
				if (nTotalTuplesTriggeredAssessment >= nTriggeredAssessmentWindow)
				{	
					nTotalTuplesTriggeredAssessment = 0;
								
					if (CurrentTriggeredStateAssessment == null)
					{	
	  						//Order Streams by freq
						    CurrentTriggeredStateAssessment = (ArrayList<Integer>)TriggeredStateAssessment.clone();
							  for (int g = 1; g <= NumSTREAMS; g++)
							  {
								  TriggeredStateAssessment.set(g-1,0);
							  }
						    
						}
						else
						{	
						  PrevTriggeredStateAssessment = CurrentTriggeredStateAssessment;
						  CurrentTriggeredStateAssessment = (ArrayList<Integer>)TriggeredStateAssessment.clone();
						  
						  int nTotalPrevTriggeredStateAssessment = 0;
						  int nTotalCurrentTriggeredStateAssessment = 0;
						  
						  //clear variables for assessment
						  for (int g = 1; g <= NumSTREAMS; g++)
						  {
							
							nTotalPrevTriggeredStateAssessment+=PrevTriggeredStateAssessment.get(g-1);
							nTotalCurrentTriggeredStateAssessment+=CurrentTriggeredStateAssessment.get(g-1);
							
						  }
		
						  //compare current to previous
						  double dLargDiff = 0.0;
						  //1) general states
						  int nStateStreamId = -99;
						  
						  if (PrevTriggeredStateAssessment != null)
						  {
							  for (int g = 1; g <= NumSTREAMS; g++)
							  {
								  if((CurrentTriggeredStateAssessment.get(g-1) > 0) &&(nPeriodicStreamId != g ))
								  {  
								    double dDiff = ((CurrentTriggeredStateAssessment.get(g-1)*1.0)/(nTotalCurrentTriggeredStateAssessment*1.0))-((PrevTriggeredStateAssessment.get(g-1)*1.0)/(nTotalPrevTriggeredStateAssessment*1.0));
								  
								    if(dDiff > dLargDiff)
								    {
									  nStateStreamId = g-1;
									  dLargDiff = dDiff;
//									  System.out.println("Prev Freq State - STREAM ID of " + PrevTriggeredStateAssessment.get(g-1).StreamID + " = "+PrevTriggeredStateAssessment.get(g-1).nTotalSentByState);
//									  System.out.println("Cuurent Freq State - STREAM ID of " + CurrentTriggeredStateAssessment.get(g-1).StreamID + " = "+CurrentTriggeredStateAssessment.get(g-1).nTotalSentByState);		
								     }
								  }
								  
								  TriggeredStateAssessment.set(g-1,0);
							  }
							
						      PrevTriggeredStateAssessment = CurrentTriggeredStateAssessment;
						   }
											
						  if (nStateStreamId != -99)
						  {	  
						    STeMStreamData.get(nStateStreamId).startExhaustiveAssessment();						  
						    StreamAssessPos.set(nStateStreamId, false);
						    
//							System.out.println("Triggered startExhaustiveAssessment on "+nStateStreamId);

						  } 
						}
					}
			}
			
			//////////////////////////////////////////////
			// Periodic Assessment
			if(bPeriodicAssessment)
			{	
//				System.out.println("nCurrentPeriodicAssessmentTupleCount="+nCurrentPeriodicAssessmentTupleCount);
//				System.out.println("STeMStreamData.get(nPeriodicStreamId).getTotalNumQueryRequests()="+STeMStreamData.get(nPeriodicStreamId).getTotalNumQueryRequests());
				
				if(nCurrentPeriodicAssessmentTupleCount==0)
				{
					STeMStreamData.get(nPeriodicStreamId).startExhaustiveAssessment();
					nCurrentPeriodicAssessmentTupleCount=1;					
				}
				else if(STeMStreamData.get(nPeriodicStreamId).getTotalNumQueryRequests()>=nPeriodicAssessmentWindow)
				{	
					if(bTrackTimeToSearch)
						STeMStreamData.get(nPeriodicStreamId).endExhaustiveAssessment(NumBits,strTrackTimeToSearchOutputfile);
					else
						STeMStreamData.get(nPeriodicStreamId).endExhaustiveAssessment(NumBits);
					
					nPeriodicStreamId +=1;
					if (nPeriodicStreamId>=NumSTREAMS)
						nPeriodicStreamId =0;

					while(StreamAssessPos.get(nPeriodicStreamId) == false)
					{
						StreamAssessPos.set(nPeriodicStreamId,true);
						nPeriodicStreamId +=1;
						if (nPeriodicStreamId>=NumSTREAMS)
							nPeriodicStreamId =0;							
					}
										
					STeMStreamData.get(nPeriodicStreamId).startExhaustiveAssessment();
					nCurrentPeriodicAssessmentTupleCount++;
					
//					System.out.println("Periodic startExhaustiveAssessment on "+nPeriodicStreamId);

				}
				
			}
			
			//////////////////////////////////////////////
			// Continuous Assessment
			if (bContinuousAssessment)
			{
				for (int s =0;s<NumSTREAMS;s++)
				{						
					if(STeMStreamData.get(s).getTotalNumQueryRequests()>= nPeriodicAssessmentWindow)
					{							
						if(bTrackTimeToSearch)
							STeMStreamData.get(s).endExhaustiveAssessment(NumBits,strTrackTimeToSearchOutputfile);
						else
							STeMStreamData.get(s).endExhaustiveAssessment(NumBits);
						
						STeMStreamData.get(s).startExhaustiveAssessment();
					}
				  }	
				}
		 }
		//////////////////////////////////////////////

		
		//System.out.println("# of tuples on level 0: " + TuplePool.get(0).size() + " tuples.");
		return true;
	}

	public void setWindowSize(String[] size) 
	{
		windowSize = Integer.parseInt(size[0]);
	}
	
	public void init(int i) 
	{		
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) 
		{
			return;
		}
		
		XATQueueSchema outSchema = this.getOutputQueue(i).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) 
		{
			this.getOutputQueue(i).setSchema(inSchema);
		}
			
		computeOutputPositions();//no need to overload since all the output queue are the same
		numOutputColumns = outputPositions[0].length;
		setHasInit(true);				
	}

	public String toToolTipString() 
	{
		return "EddyOperatorImp";
	}
	
//////////////////////////////////////////////////////////
	public ExpressionStr getExpression() 
	{
		return expression;
	}

	public void setExpression(ExpressionStr expression) 
	{
		this.expression = expression;
	}
////////////////////////////////////////////////////////
	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input
	 * queues. Inside method run() the numbers are computed based on the
	 * property "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler, while
	 * in cleanAccumulatedTuples, all tuples in input queues that have their max
	 * timesamp smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return true;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() 
	{
		return true;
	}

	/*
	 * check if this operator has any unmatched states. return: TRUE if this
	 * operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() 
	{
		return false;
	}

	public boolean matchStates(XATStreamOperator op) 
	{
		return false;
	}

	public boolean recomputeUnmatchedState() 
	{
		return true;
	}

	public int setStateID(int nextid) 
	{
		return nextid;
	}

	public boolean finalizeOperation() 
	{
		return true;
	}
	
////////////////////////////
	public void setSTeMStreamData(HashMap<Integer,STeMPartitionData> sd)
	{
		STeMStreamData = sd;
	}
//////////////////////////////////////////////////////

	public void setNumStreams(int Ns)
	{
		NumSTREAMS = Ns;
	}
	
	public void setNumOps(int NOps)
	{
		NumOps = NOps;	   
	}
//////////////////////////////////////////////////////
	
	public void setNumBits(int NBits)
	{
		NumBits = NBits;	   
	}
//////////////////////////////////////////////////////
	
	public void setNumJoinAttribs(int NBits)
	{
		NumJoinAttribs = NBits;	   
	}
	
	public int getNumJoinAttribs()
	{
		return NumJoinAttribs;	   
	}
	
	public void setOpStringIds(int OpId,String val)
	{
		if(OperatorSTREAMIds==null)
			OperatorSTREAMIds = new ArrayList<ArrayList<Integer>>();
		if ((OperatorSTREAMIds.size()-1) < OpId)
			OperatorSTREAMIds.add(new ArrayList<Integer>());
		
		StringTokenizer st = new StringTokenizer(val);
	    while (st.hasMoreElements()) 
	    {
	    	//System.out.println(st.nextElement());
			OperatorSTREAMIds.get(OpId).add(Integer.parseInt(st.nextElement().toString()));
	    }
	    
	}
	
	public void setSTEMDataSTREAMIds(String val)
	{
//		if(OperatorSTEMDataSTREAMIds==null)
//			OperatorSTEMDataSTREAMIds = new ArrayList<Integer>();
//		StringTokenizer st = new StringTokenizer(val);
//	    while (st.hasMoreElements()) 
//	    {
//	    	//System.out.println(st.nextElement());
//	    	OperatorSTEMDataSTREAMIds.add(Integer.parseInt(st.nextElement().toString()));
//	    }
	    
	}
	
//////////////////////////////////////////////////////
	public void setRouter(String val)
	{
		strRouter = val;
	}

//////////////////////////////////////////////////////
		
	public ArrayList<BitSet> getPreFilledPerStream()
	{
		return PreFilledPerStream;
	}
	
//////////////////////////////////////////////////////

	public void setNumTupToProcessPerOp(int val)
	{
		NumTupToProcessPerOp = val;
	}
	
	public int getNumTupToProcessPerOp()
	{
		return NumTupToProcessPerOp;
	}

	public void setNumTupToProcess(int val)
	{
		NumTupToProcess = val;
	}

	public void setNumLevelsToProcess(int val)
	{
		NumLevelsToProcess = val;
	}
//////////////////////////////////////////////////////

	public void setFlipFlopTupleProcStrategy(boolean val)
	{
		bFlip = val;
	}

	public void setUseTupleProcTopStrategy(boolean val)
	{
		bTop = val;
	}
//////////////////////////////////////////////////////
	//stores the stream window size
	public void AddSTREAMWindowCount(int streamId, int val)
	{
		STREAMWindowCount.put(streamId,val);
	}

	public HashMap <Integer,Integer> getSTREAMWindowCount()
	{
		return STREAMWindowCount;
	}
	
	
//////////////////////////////////////////////////////
	public void setAdaptingIndice(boolean val)
	{
		bSwitchIndex = val;
	}

	public boolean getAdaptingIndice()
	{
		return bSwitchIndex;
	}
//////////////////////////////////////////////////////
	//sets the partition size
	public void setMaxPartitionSize(int val)
	{
		MAX_PARTITION_SIZE = val;
	}

	public int getMaxPartitionSize()
	{
		return MAX_PARTITION_SIZE;
	}
	
//////////////////////////////////////////////////////
	//sets the partition window size
	public void setMaxPartitionWindowSize(int val)
	{
		MAX_PARTITION_WINDOW_SIZE = val;
	}

	public int getMaxPartitionWindowSize()
	{
		return MAX_PARTITION_WINDOW_SIZE;
	}

	/*****************************************/
//////////////////////////////////////////////////////
	//is partitioned?
	public void setIsPartitioned(boolean val)
	{
		bIsPartitioned = val;
	}

	public boolean getIsPartitioned()
	{
		return bIsPartitioned;
	}

//////////////////////////////////////////////////////
	//use IMPIndex? No - hash index
	public void setUseIMPIndex(boolean val)
	{
		bIMPIndex = val;
	}

	public boolean getUseIMPIndex()
	{
		return bIMPIndex;
	}
//////////////////////////////////////////////////////
	//sets the Number of Limited Indices
	public void setNumLimitedIndices(int val)
	{
		nNumLimitedIndices = val;
	}

	public int getNumLimitedIndices()
	{
		return nNumLimitedIndices;
	}

//////////////////////////////////////////////////////
	//uses IndexAssessment?
	public void setIndexAssessment(boolean val)
	{
		bIndexAssessment = val;
	}

	public boolean getIndexAssessment()
	{
		return bIndexAssessment;
	}

//////////////////////////////////////////////////////
	//uses ExhaustiveAssessment?
	public void setExhaustiveAssessment(boolean val)
	{
		bExhaustiveAssessment = val;
	}

	public boolean getExhaustiveAssessment()
	{
		return bExhaustiveAssessment;
	}
	
//////////////////////////////////////////////////////
	//uses PeriodicAssessment?
	public void setPeriodicAssessment(boolean val)
	{
		bPeriodicAssessment = val;
	}

	public boolean getPeriodicAssessment()
	{
		return bPeriodicAssessment;
	}
	
//////////////////////////////////////////////////////
	//PeriodicAssessmentWindow
	public void setPeriodicAssessmentWindow(int val)
	{
		nPeriodicAssessmentWindow = val;
	}

	public int getPeriodicAssessmentWindow()
	{
		return nPeriodicAssessmentWindow;
	}
	
//////////////////////////////////////////////////////
	//uses ContinuousAssessment?
	public void setContinuousAssessment(boolean val)
	{
		bContinuousAssessment = val;
	}

	public boolean getContinuousAssessment()
	{
		return bContinuousAssessment;
	}
	
//////////////////////////////////////////////////////
	//uses TriggeredAssessment?
	public void setTriggeredAssessment(boolean val)
	{
		bTriggeredAssessment = val;
	}

	public boolean getTriggeredAssessment()
	{
		return bTriggeredAssessment;
	}
	
//////////////////////////////////////////////////////
	//TriggeredAssessmentWindow
	public void setTriggeredAssessmentWindow(int val)
	{
		nTriggeredAssessmentWindow = val;
	}

	public int getTriggeredAssessmentWindow()
	{
		return nTriggeredAssessmentWindow;
	}
	
//////////////////////////////////////////////////////
	//uses NoSampling?
	public void setNoSampling(boolean val)
	{
		bNoSampling = val;
	}

	public boolean getNoSampling()
	{
		return bNoSampling;
	}
//////////////////////////////////////////////////////
	//uses SimpleRandomSampling?
	public void setSimpleRandomSampling(boolean val)
	{
		bSimpleRandomSampling = val;
	}

	public boolean getSimpleRandomSampling()
	{
		return bSimpleRandomSampling;
	}
//////////////////////////////////////////////////////
	//SimpleRandomPercent
	public void setSimpleRandomPercent(int val)
	{
		nSimpleRandomPercent = val;
	}

	public int getSimpleRandomPercent()
	{
		return nSimpleRandomPercent;
	}
	
//////////////////////////////////////////////////////
	//uses SimpleRandomSampling?
	public void setStratifiedRandomSampling(boolean val)
	{
		bStratifiedRandomSampling = val;
	}

	public boolean getStratifiedRandomSampling()
	{
		return bStratifiedRandomSampling;
	}
//////////////////////////////////////////////////////
	//StratifiedRandomPercent
	public void setStratifiedRandomPercent(int val)
	{
		nStratifiedRandomPercent = val;
	}

	public int getStratifiedRandomPercent()
	{
		return nStratifiedRandomPercent;
	}
//////////////////////////////////////////////////////
	//StratifiedRandomWindow
	public void setStratifiedRandomWindow(int val)
	{
		nStratifiedRandomWindow = val;
	}

	public int getStratifiedRandomWindow()
	{
		return nStratifiedRandomWindow;
	}
	
//////////////////////////////////////////////////////
	//uses SystematicSampling?
	public void setSystematicSampling(boolean val)
	{
		bSystematicSampling = val;
	}

	public boolean getSystematicSampling()
	{
		return bSystematicSampling;
	}
//////////////////////////////////////////////////////
	//SystematicInterval
	public void setSystematicInterval(int val)
	{
		nSystematicInterval = val;
	}

	public int getSystematicInterval()
	{
		return nSystematicInterval;
	}
//////////////////////////////////////////
	public void setNoMigration(boolean val)
	{
		bNoMigration=val;		
	}
	public boolean getNoMigration()
	{
		return bNoMigration;		
	}
	
	public void setSymmetricMigration(boolean val)	
	{
		bSymmetricMigration=val;		
	}
	public boolean getSymmetricMigration()	
	{
		return bSymmetricMigration;		
	}
	
	public void setPartitionMigration(boolean val)
	{
		bPartitionMigration=val;		
	}
	public boolean getPartitionMigration()
	{
		return bPartitionMigration;		
	}
	
	public void setCompleteMigration(boolean val)
	{
		bCompleteMigration=val;		
	}
	public boolean getCompleteMigration()
	{
		return bCompleteMigration;		
	}
	
/////////////////////////////////////////////////////////
//	SRIA
	public void setSRIA(boolean val)
	{
		bSRIA=val;		
	}
	public boolean getSRIA()
	{
		return bSRIA;		
	}
	
	public void setHH(boolean val)
	{
		bHH=val;		
	}
	public boolean getHH()
	{
		return bHH;		
	}
	
	public void setHHH(boolean val)
	{
		bHHH=val;		
	}
	public boolean getHHH()
	{
		return bHHH;		
	}

	public void setCombineNodesByPickingOne(boolean val)
	{
		bCombineNodesByPickingOne=val;		
	}
	public boolean getCombineNodesByPickingOne()
	{
		return bCombineNodesByPickingOne;		
	}

	public void setPickParentWithHighestFreq(boolean val)
	{
		bPickParentWithHighestFreq=val;		
	}
	public boolean getPickParentWithHighestFreq()
	{
		return bPickParentWithHighestFreq;		
	}

	
	public void setHHErrorRate(double val)
	{
		dHHErrorRate = val;
	}

	public double getHHErrorRate()
	{
		return dHHErrorRate;
	}
	
	public void setCostFactorSignificance(double val)
	{
		dCostFactorSignificance = val;
	}

	public double getCostFactorSignificance()
	{
		return dCostFactorSignificance;
	}
////////////////////////////////////////////
//Combined stats ?
		
		public void setCombinedStats(boolean val)
		{
			bCombinedStats=val;		
		}
		public boolean getCombinedStats()
		{
			return bCombinedStats;		
		}
		
		public void setNumCombinedStats(int val)
		{
			nNumCombinedStats = val;
		}

		public int getNumCombinedStats()
		{
			return nNumCombinedStats;
		}

		public void setCombinedStatsFactor(double[] val)
		{
			CombinedStatsFactor = val;
		}

		public double[] getCombinedStatsFactor()
		{
			return CombinedStatsFactor;
		}
		/*****************************************/
//		//Index Selection trigger
//		public void setNoIndexSelectionTrigger(boolean val)
//		{
//			bNoIndexSelectionTrigger=val;		
//		}
//		public boolean getNoIndexSelectionTrigger()
//		{
//			return bNoIndexSelectionTrigger;		
//		}
//
//		public void setCostIncrIndexSelectionTrigger(boolean val)
//		{
//			bCostIncrIndexSelectionTrigger=val;		
//		}
//		public boolean getCostIncrIndexSelectionTrigger()
//		{
//			return bCostIncrIndexSelectionTrigger;		
//		}
//
//		public void setFreqIncrIndexSelectionTrigger(boolean val)
//		{
//			bFreqIncrIndexSelectionTrigger=val;		
//		}
//		public boolean getFreqIncrIndexSelectionTrigger()
//		{
//			return bFreqIncrIndexSelectionTrigger;		
//		}
//
//		public void setCombinedIncrIndexSelectionTrigger(boolean val)
//		{
//			bCombinedIncrIndexSelectionTrigger=val;		
//		}
//		public boolean getCombinedIncrIndexSelectionTrigger()
//		{
//			return bCombinedIncrIndexSelectionTrigger;		
//		}
			
		public void setDegreeCostIncr(double val)
		{
			dDegreeCostIncr = val;
		}

		public double getDegreeCostIncr()
		{
			return dDegreeCostIncr;
		}

		public void setDegreeFreqIncr(double val)
		{
			dDegreeFreqIncr = val;
		}

		public double getDegreeFreqIncr()
		{
			return dDegreeFreqIncr;
		}

//////////////////////////////
		//TrackTimeToSearch
		public void setTrackTimeToSearch(boolean val)
		{
			bTrackTimeToSearch=val;		
		}
		public boolean getTrackTimeToSearch()
		{
			return bTrackTimeToSearch;		
		}
		//TrackTimeToSearchOutputFile
		public void setTrackTimeToSearchOutputFile(String val)
		{
			strTrackTimeToSearchOutputfile=val;		
		}
		public String getTrackTimeToSearchOutputfile()
		{
			return strTrackTimeToSearchOutputfile;		
		}
		
//////////////////////////////
		//TrackOptimalIndices
		public void setTrackOptimalIndices(boolean val)
		{
			bTrackOptimalIndices=val;		
		}
		public boolean getTrackOptimalIndices()
		{
			return bTrackOptimalIndices;		
		}
		
		//TrackOptimalIndicesOutputFile
		public void setTrackOptimalIndicesOutputFile(String val)
		{
			strTrackOptimalIndicesOutputfile=val;		
		}
		public String getTrackOptimalIndicesOutputfile()
		{
			return strTrackOptimalIndicesOutputfile;		
		}
//////////////////////////////////////////////
		public void setCountMigrations(boolean CM)
		{
			bCountMigrations = CM;
		}
		
		public boolean getCountMigrations()
		{
			return bCountMigrations;
		}
		
		public void setCountMigrationsFile(String strFile)
		{
			strCountMigration = strFile;
		}
		public String getCountMigrationsFile()
		{
			return strCountMigration;
		}
//////////////////////////////////////////////

		public void setStoreComparisonTime(boolean CM)
		{
			bStoreComparisonTime = CM;
		}
		public boolean getStoreComparisonTime()
		{
			return bStoreComparisonTime;
		}
		///////////
		public void setStoreComparisonTimeFile(String strFile)
		{
			strStoreComparisonTimeFile = strFile;
		}
		public String getStoreComparisonTimeFile()
		{
			return strStoreComparisonTimeFile;
		}

///////////////////////////////////////////////////////////////////////////////////////////////
		public static void WriteToFile(String fileLocation, String stringToWrite, boolean doNewLine)
		{
			try
			{
				//Create file 
				FileWriter fstream = new FileWriter(fileLocation, true);
				BufferedWriter out = new BufferedWriter(fstream);
				
				out.write(stringToWrite);
				
				if (doNewLine)
				{
					out.newLine();
				}
				else
					out.append("|");
				
				out.flush();			    
						//Close the output stream
				out.close();			    
			}
			catch (Exception e)
			{
				//Catch exception if any
				System.err.println("Error: " + e.getMessage());
			}
		}
		
/////////////////////////////////////////////////
// TrackCompletedPartitions		
		public void SetoutFile(String val) 
		{
			outFile = val;
		}
		//
		public void setTrackCompletedPartitions(boolean CM)
		{
			bTrackCompletedPartitions = CM;
		}
		public boolean getTrackCompletedPartitions()
		{
			return bTrackCompletedPartitions;
		}

}


