/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.eddyPartition.operator;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import java.util.Vector;

import edu.wpi.cs.dsrg.STeM.Partition.STeMPartitionData;
import edu.wpi.cs.dsrg.eddy.datastructure.*;
import java.util.BitSet;
import java.util.StringTokenizer;
//Karen Works

import org.w3c.dom.Node;

//This is the eddy operator 

public class EddyPartitionOperatorImp extends XATMultiQueueWindowStreamOperatorImp implements EddyPartitionOperator
{
	//public class EddyOperatorImp extends XATSingleQueueStreamOperatorImp implements EddyOperator{
	private ExpressionStr expression;
	//public int windowSize;

	// number of operators
    private int NumOps = 0;
    //number of streams
    private int NumSTREAMS = 0;
	// number of bits in done bit set
    private int NumBits = 0;

    // number of tuples to process from pool
    private int NumTupToProcessPerOp = 500;
    private int NumTupToProcess = 2000;
    private int NumLevelsToProcess = 2;
    
	// Tuples that require processing
	private HashMap<Integer, EddyQueue> TuplePool;
	
	//operator bitmaps per STREAM
	private ArrayList <BitSet> PreFilledPerStream;
	
	// pointer to data stored for STEM
	private HashMap <Integer,STeMPartitionData> STeMStreamData;

	// array of operator's STREAM Ids
	private ArrayList <Integer> OperatorSTREAMIds;
	
	//router
	private String strRouter= "Simple";

	// array of operator's tickets for back pressure
	private ArrayList <Double> OperatorTickets;

	//used for picking next tuple to run

	//are you going to flip flop strategies?
	boolean bFlip = true;
	//is looking at the top the current strategy
	boolean bTop = false;
	
	// stores window count for streams
	private HashMap <Integer,Integer> STREAMWindowCount;
	
	public EddyPartitionOperatorImp(XATNode init) 
	{
		super(init);		
	}

	public void InitializeVariables()
	{
		//set up the PreFilledPerStream array
		PreFilledPerStream = new ArrayList(NumSTREAMS-1);
		STREAMWindowCount = new HashMap <Integer,Integer>(); 

		try
		{
			if (strRouter.compareTo("Ticket") == 0)
			{	
				OperatorTickets = new ArrayList(NumOps);
				//initialize tickets
				for (int O = 0;O<NumOps;O++)
				{
					OperatorTickets.add(0.0);
				}
			}
			
			//initializes basic bitsets for each streamid
			for (int i = 0; i <= NumSTREAMS - 1; i++)
			{					
				BitSet r = new BitSet(NumOps+1);
					
				for (int j=0; j < OperatorSTREAMIds.size();j++)
				{
					if (OperatorSTREAMIds.get(j) == i)				
						r.set(j+1, true);						
					else							
						r.clear(j+1);
				}//end for
					
				PreFilledPerStream.add(r);
			}
			
			//System.out.println("Hello");

		}
		catch (Exception ex)
		{
			System.out.print("Error setting up EddyOperator!!");
		}  
	}
	
	public boolean run() throws Exception 
	{		

		for (int i=0; i < this.getOutputQueueCount(); i++)
		{
			if(this.getOutputQueue(i).getSchema() == null)
			{
				init(i);
			}
		}

		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		//int howMany  = 500; //rimma temp "hack"
		
		//-------------------------------------------------------------------------
		// FIRST PROCESS THE TUPLES FROM OPERATORS INPUT QUEUES
		//-------------------------------------------------------------------------
		//Scan for Other Operators & STeM's returning Tuples.
		for (int j = NumSTREAMS; j <= (NumSTREAMS + NumOps - 1); j++) 
		{			
			XATTuple[] tuples = (this.getInputQueue(j)).dequeue(howMany, this);
			EddyTuple et;
			
//			System.out.println("\nProcessing operator queue [" + j + "]");
//			System.out.println("\t# of tuples in the queue: " + (this.getInputQueue(j)).getQueueSize());
//			System.out.println("\tDequeued: " + tuples.length + " tuples.");
//			System.out.println("");
						
			for(int k = 0; k < tuples.length;k++)
			{
				et = (EddyTuple)tuples[k];
							
				if (strRouter.equalsIgnoreCase("Ticket"))
				{	
					OperatorTickets.set(j - NumSTREAMS, OperatorTickets.get(j - NumSTREAMS) + 1);										
				}
				
				//--------------------------------------------------------------------
				// Tuple is ALL DONE
				//--------------------------------------------------------------------				
				// check to see if the tuple is done
				// ie all ops done bits set to true
				if (et.isDone() == true) 
				{					
					// if it is done, output result
					//Assume last output queue  is the result queue
					this.getOutputQueue(NumOps).enqueue(et, this);
					et = null;
				}
				else
				{	
					//--------------------------------------------------------------------
					// Tuple is NOT DONE
					//--------------------------------------------------------------------						
					// not done --> add to tuple pool for more processing
					int key = et.NumDoneBitsComplete();
					
					//System.out.println("# of ops done: " + key);
					
					EddyQueue Eq;
					
					if((Eq = TuplePool.get(key)) != null)
					{
						Eq.enqueue(et);	
					}
					else
					{
						EddyQueue<EddyTuple> eq = new EddyQueue<EddyTuple>();
						eq.enqueue(et);
						TuplePool.put(key, eq);
					}					
				}
			}//end for iterating over all tuples
		}//end for - retrieved tuples from operator queues and put them into the tuple pool		
		
		
		
		ArrayList<Boolean> OpFull = new ArrayList<Boolean>(NumOps);
		ArrayList<Integer> OpNumSent = new ArrayList<Integer>(NumOps);
		
		for (int g = 0; g < NumOps; g++)
		{	
			OpFull.add(false);
			OpNumSent.add(0);
			
//			//for testing only
//			if (TuplePool!= null)
//			{				  
//				if (TuplePool.get(g)!= null)
//				{
//					System.out.println(g+". at level (# of operators done): "+g+" there are "+TuplePool.get(g).size() + " tuples");
//				}
//				else
//				{
//					System.out.println(g+". TupleTool is null");
//				}
//			}  
		}	

		int NumOpsFull = 0;
		int NumSent = 0;
		int NumLevelsProcessed = 0;
		
		if(TuplePool != null)
		{	
			// look at each op to see if any ops have tuple returning 
			// to the eddy operator for processing
			if (bTop)
			{	
//				System.out.println("\nUsing top strategy...");
				
				int numTuplesSent = 0;
				Hashtable hashTuplesSentByOp = new Hashtable();
				
				for (int k = NumOps-1;((k >= 0) && (NumOpsFull <= NumOps) && (NumSent < NumTupToProcess))&&(NumLevelsProcessed <= NumLevelsToProcess);k--)
				{
					if (TuplePool.get(k)!= null)
					{
						if (TuplePool.get(k).size() != 0)
						{							
//							System.out.println("Processing level [" + k + "]");
							
							NumLevelsProcessed++;
					
							// there are items in the queue to process
							EddyQueue Eq = TuplePool.get(k);							
							int count = 0;
							int NumInQ = Eq.length();
							
							while((count != NumInQ) && (NumOpsFull < NumOps)&& (NumSent < NumTupToProcess))
							{
								EddyTuple TupletoProcess = (EddyTuple)Eq.dequeue();
								
//								System.out.println(TupletoProcess.getReadyBitSet().getSetBitsToString());
								
								count++;
								int nextOp = 0;
								if (strRouter.compareTo("Simple") == 0)
								{
									nextOp = TupletoProcess.NextReadyOp();
								}
								else if (strRouter.compareTo("Ticket") == 0)
								{
									nextOp = TupletoProcess.NextReadyOpTicket(OperatorTickets);
								}
						
//								System.out.println("nextOp="+nextOp);
								
								//add tuple to appropriate output queue
								if (!OpFull.get(nextOp-1))
								{	
									if (OpNumSent.get(nextOp-1) == NumTupToProcessPerOp)
									{	
										OpFull.set(nextOp-1, true);
										NumOpsFull++;
										Eq.enqueue(TupletoProcess);											
									}
									else
									{										
										//new tuples add to STEm data
										if (k == 0)
										{
											//if STeMData exists for STREAMId send tuple to STeMData
//											if (STeMStreamData != null)
//											{	
//												STeMPartitionData sd = STeMStreamData.get(TupletoProcess.getStreamID());
//												if (sd != null)
//												{
//													sd.Insert(TupletoProcess.getTuple(TupletoProcess.getStreamID()));
//												}
////												else
////												{
////													System.out.println("ERROR! want to catch this.");
////												}
//											}															
										}
										
										NumSent++;										
										
										OpNumSent.set(nextOp-1, OpNumSent.get(nextOp-1) + 1);
										this.getOutputQueue(nextOp-1).enqueue(TupletoProcess, this);
										
										if (hashTuplesSentByOp.get(nextOp-1) == null)
										{
											int countTuples = 1;
											hashTuplesSentByOp.put(nextOp-1, countTuples);
										}
										else
										{
											int countTuples = Integer.parseInt(hashTuplesSentByOp.get(nextOp-1).toString());
											countTuples++;
											hashTuplesSentByOp.put(nextOp-1, countTuples);
										}
										
										if (strRouter.compareTo("Ticket") == 0)
										{	
											OperatorTickets.set(nextOp-1, OperatorTickets.get(nextOp-1)+1);
										}										
									}
								}
								else
								{
									Eq.enqueue(TupletoProcess);
								}																
							}					
						}
					}
					
//					System.out.println("---");
//					 for (Enumeration e = hashTuplesSentByOp.keys(); e.hasMoreElements();) 
//					 {
//					        int key = Integer.parseInt(e.nextElement().toString());
//					        int numSentToOp = Integer.parseInt(hashTuplesSentByOp.get(key).toString());
//					        System.out.println("operator ["+key+"], "+this.getOutputQueue(key).getQueueCustomer()+" # of tuples sent [" + numSentToOp + "]");
//					 }
//					
//					System.out.println("Done processing level: " + k +". Total # of tuples sent [" + NumSent +"]");					
				}	
				
			   
				
			}//end if "top strategy"
			//look at the queue that is the most full	
			else
			{			    
				BitSet AlreadyProcessed = new BitSet(NumOps);
			    AlreadyProcessed.clear();
			    
				int Maxqueue = 0;
				int MaxqueueIndex = 0;
			    
				while ((NumOpsFull <= NumOps)&& (NumSent <= NumTupToProcess) && (NumLevelsProcessed <= NumLevelsToProcess))
				{
					Maxqueue = 0;
					MaxqueueIndex = 0;
					int i = 0;
					while (i<=NumOps)	
					{
						i = AlreadyProcessed.nextClearBit(i);						
						if ((TuplePool.get(i)!= null) && (!AlreadyProcessed.get(i)))
						{						  
							if (Maxqueue <TuplePool.get(i).size())
							{	
								Maxqueue = TuplePool.get(i).size();
								MaxqueueIndex = i;
							}	
						}
						else if (TuplePool.get(i)== null)
						{
							AlreadyProcessed.set(i);
						}
						i++;					
					}
					
					AlreadyProcessed.set(MaxqueueIndex);
					
					if (TuplePool.get(MaxqueueIndex)!= null)
					{
						NumLevelsProcessed++;
					
						// there are items in the queue to process
						EddyQueue Eq = TuplePool.get(MaxqueueIndex);
						int count = 0;
						int NumInQ = Eq.length();
						while((count != NumInQ) && (NumOpsFull<NumOps) && (NumSent<NumTupToProcess))
						{
							EddyTuple TupletoProcess = (EddyTuple)Eq.dequeue();
							count++;
							int nextOp = 0;
							if (strRouter.compareTo("Simple") == 0)
							{
								nextOp = TupletoProcess.NextReadyOp();
							}
							else if (strRouter.compareTo("Ticket") == 0)
							{
								nextOp = TupletoProcess.NextReadyOpTicket(OperatorTickets);
							}
							//add tuple to appropriate output queue
							if (!OpFull.get(nextOp-1))
							{	
								if (OpNumSent.get(nextOp-1) == NumTupToProcessPerOp)
								{	
									OpFull.set(nextOp-1, true);
									NumOpsFull++;
									Eq.enqueue(TupletoProcess);	
								}
								else
								{
									NumSent++;
									OpNumSent.set(nextOp-1,OpNumSent.get(nextOp-1)+1);
									this.getOutputQueue(nextOp-1).enqueue(TupletoProcess, this);
									if (strRouter.compareTo("Ticket")==0)
									{	
										OperatorTickets.set(nextOp-1, OperatorTickets.get(nextOp-1)+1);
									}
								}
							}
							else
							{
								Eq.enqueue(TupletoProcess);
							}
						}						
					}
				}
			}		
			if(bFlip)	
			{
				bTop = !bTop;
			}
		}

		//-------------------------------------------------------------------------
		// SECOND PROCESS THE TUPLES FROM THE INPUT STREAMS
		//-------------------------------------------------------------------------
		/** get the tuples from STREAM*/
		for (int j = 0; j < (NumSTREAMS); j++) 
		{
			//howMany  = 500; //rimma temp "hack"			
			
//			System.out.println("\nProcessing stream queue [" + j + "]");
//			System.out.println("\t# of tuples in the queue: " + (this.getInputQueue(j)).getQueueSize());
			
			XATTuple[] tuples = this.getInputQueue(j).dequeue(howMany, this);
//			System.out.println("\tDequeued: " + tuples.length + " tuples.");
//			System.out.println("");
			
			EddyTuple et;
			// process all new tuples coming from STREAMS
			for(int k = 0; k < tuples.length;k++)
			{
				BitSet rbs = (BitSet)PreFilledPerStream.get(j);
				for(int s = 0; s<NumSTREAMS;s++)
				{
					if(s!=j)
					{	
						BitSet InvRbs = (BitSet)PreFilledPerStream.get(s).clone();
						InvRbs.flip(1, rbs.length());						
						rbs.and(InvRbs);
					}
				}
				
				et = new EddyTupleImp(tuples[k], rbs, NumBits, NumSTREAMS, j);
				et.setStreamID(j);
				// add new tuple to tuple pool
				
				EddyQueue Eq;
				if(TuplePool != null)
				{
					if((Eq = TuplePool.get(0)) != null)
					{
						Eq.enqueue(et);	
					}
					else
					{
						EddyQueue<EddyTuple> NewEq = new EddyQueue<EddyTuple>();
						NewEq.enqueue(et);
						TuplePool.put(0, NewEq);
					}
				}
				else
				{
					TuplePool = new HashMap();
					EddyQueue<EddyTuple> eq = new EddyQueue<EddyTuple>();
					eq.enqueue(et);
					TuplePool.put(0, eq);					
				}
				
//				//if STeMData exists for STREAMId send tuple to STeMData
				if (STeMStreamData != null)
				{	
					STeMPartitionData sd = STeMStreamData.get(tuples[k].getStreamID());
					if (sd != null)
					{
						sd.Insert(tuples[k]);
					}
				}				
			}		
		}		
		
//		System.out.println("# of tuples on level 0: " + TuplePool.get(0).size() + " tuples.");
		//System.gc();
		return true;
	}

	public void setWindowSize(String[] size) 
	{
		windowSize = Integer.parseInt(size[0]);
	}
	
	public void init(int i) 
	{		
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) 
		{
			return;
		}
		
		XATQueueSchema outSchema = this.getOutputQueue(i).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) 
		{
			this.getOutputQueue(i).setSchema(inSchema);
		}
			
		computeOutputPositions();//no need to overload since all the output queue are the same
		numOutputColumns = outputPositions[0].length;
		setHasInit(true);				
	}

	public String toToolTipString() 
	{
		return "EddyOperatorImp";
	}

	public ExpressionStr getExpression() 
	{
		return expression;
	}

	public void setExpression(ExpressionStr expression) 
	{
		this.expression = expression;
	}

	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input
	 * queues. Inside method run() the numbers are computed based on the
	 * property "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler, while
	 * in cleanAccumulatedTuples, all tuples in input queues that have their max
	 * timesamp smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return true;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() 
	{
		return true;
	}

	/*
	 * check if this operator has any unmatched states. return: TRUE if this
	 * operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() 
	{
		return false;
	}

	public boolean matchStates(XATStreamOperator op) 
	{
		return false;
	}

	public boolean recomputeUnmatchedState() 
	{
		return true;
	}

	public int setStateID(int nextid) 
	{
		return nextid;
	}

	public boolean finalizeOperation() 
	{
		return true;
	}

	public void setSTeMStreamData(HashMap<Integer,STeMPartitionData> sd)
	{
		STeMStreamData = sd;
	}

	public void setNumStreams(int Ns)
	{
		NumSTREAMS = Ns;
	}
	
	public void setNumOps(int NOps)
	{
		NumOps = NOps;	   
	}
	
	public void setNumBits(int NBits)
	{
		NumBits = NBits;	   
	}
	
	public void setOpStringIds(String val)
	{
		OperatorSTREAMIds = new ArrayList();
		StringTokenizer st = new StringTokenizer(val);
	    while (st.hasMoreElements()) 
	    {
	    	//System.out.println(st.nextElement());
			OperatorSTREAMIds.add(Integer.parseInt(st.nextElement().toString()));
	    }
	    
	}

	public void setRouter(String val)
	{
		strRouter = val;
	}

	
	public void SetTuplePool(HashMap<Integer, EddyQueue> t)
	{
		TuplePool = t;
	}
		
	public ArrayList<BitSet> getPreFilledPerStream()
	{
		return PreFilledPerStream;
	}	

	public void setNumTupToProcessPerOp(int val)
	{
		NumTupToProcessPerOp = val;
	}

	public void setNumTupToProcess(int val)
	{
		NumTupToProcess = val;
	}

	public void setNumLevelsToProcess(int val)
	{
		NumLevelsToProcess = val;
	}

	public void setFlipFlopTupleProcStrategy(boolean val)
	{
		bFlip = val;
	}

	public void setUseTupleProcTopStrategy(boolean val)
	{
		bTop = val;
	}

	//stores the stream window size
	public void AddSTREAMWindowCount(int streamId, int val)
	{
		STREAMWindowCount.put(streamId,val);
	}

	public HashMap <Integer,Integer> getSTREAMWindowCount()
	{
		return STREAMWindowCount;
	}
}


