/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.calculators;

import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.manager.SpreadfireManager;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.AggregateStatisticsObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.SystemObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState;
/**
 * Insert the type's description here.
 * Creation date: (5/19/2003 11:55:02 AM)
 * @author: Bradley Momberger
 */
public class MemorySaturationCalculator extends AggregateStatisticsObject 
{
	/**
	 * MemorySaturationCalculator constructor comment.
	 */
	public MemorySaturationCalculator() 
	{
		super();	
	}
	/**
	 * MemorySaturationCalculator constructor comment.
	 * @param name java.lang.String
	 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
	 */
	public MemorySaturationCalculator(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) 
	{
		super(name, queryObject);
		
		//Runtime rt = Runtime.getRuntime();
		//systemOverhead = rt.totalMemory() - rt.freeMemory();
	}
	/**
	 * Calculates the new Value of the given property of this operator.  This
	 * should be overwritten by subclasses who can tailor this method to calculating 1
	 * specific property. <p>
	 * This method should update at least the local propertyValue and possibly the
	 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
	 *
	 * The procedure for calculating new values should be as follows:
	 *
	 * <ol>
	 * <li> Consult the statsTable for any propertyValues that are needed.
	 * <li> Perform the required calculations using the propertyValues from step 1
	 * <li> Update the local propertyValue.
	 * 
	 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
	 * </ol> 
	 */
	public void calculateNewValue() 
	{
	    //if (System.getProperty("MAX_TOTAL_STATE_SIZE") == null)
	    {
	        SpreadfireManager.setMaxStateSize();
	    }
		
		XATTree tree = (XATTree)this.getAssociatedQueryObject();
		XATNode[] nodes = tree.convertToArray();
		double mempct = 0;
		for (int i = 0; i < nodes.length; i++)
		{
			XATStreamOperator operator = (XATStreamOperator)nodes[i].getOperator();
			if (operator.getState() != null)
			{
			    int numstates = operator.getState().length;
			    for (int j = 0; j < numstates; j++)
			    {
			    	//double thisCount = operator.getStateSize();
			        if(((WOperatorState)operator.getState()[j]).getSchema() == null)
			            mempct += operator.getState()[j].getStateSize();
			        else
			        {
			            //System.out.println("Schema size is " + ((WOperatorState)operator.getState()[j]).getSchema().getLength());
			            mempct += operator.getState()[j].getStateSize() 
			    	    		* ((WOperatorState)operator.getState()[j]).getSchema().getLength();
			        }
			    }
			}
		} 	
	
		this.setPropertyValue("" + (mempct / Double.parseDouble(System.getProperty("MAX_TOTAL_STATE_SIZE"))));		
	}
}
