/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.calculators;

//import java.util.Iterator;

//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
 * This will look at all of the output queues of the associated query object 
 * (which has to be cast to a StreamOperator b/c XATOperators do not have queues)
 * and sum the TUPLE_COUNT properties 
 */
public class NumberOfAccumulatedTuplesInStreamQueues extends TuplesInQueueCalculator 
{
	/**
	 * NumberOfTuplesInOutoutQueues constructor comment.
	 */
	public NumberOfAccumulatedTuplesInStreamQueues() 
	{
		super();
	}
	/**
	 * NumberOfTuplesInOutoutQueues constructor comment.
	 * @param name java.lang.String
	 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
	 */
	public NumberOfAccumulatedTuplesInStreamQueues(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) 
	{
		super(name, queryObject);
	}
	
	/**
	 * Calculates the new Value of the given property of this operator.  This
	 * should be overwritten by subclasses who can tailor this method to calculating 1
	 * specific property. <p>
	 * This method should update at least the local propertyValue and possibly the
	 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
	 *
	 * The procedure for calculating new values should be as follows:
	 *
	 * <ol>
	 * <li> Consult the statsTable for any propertyValues that are needed.
	 * <li> Perform the required calculations using the propertyValues from step 1
	 * <li> Update the local propertyValue.
	 * 
	 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
	 * </ol> 
	 */
	public void calculateNewValue() 
	{
	    if (this.getAssociatedQueryObject() instanceof XATDag) 
	    {
	        XATDag dag = ((XATDag) this.getAssociatedQueryObject());
	        int tuplecount = 0;
	        XATNode[] nodes = dag.convertToArray();
	        for (int i=0; i<nodes.length; i++) 
	        {
	            XATStreamOperator operator = (XATStreamOperator)nodes[i].getOperator();
	            int inputsize = nodes[i].getChildrenSize();
	            for(int j=0; j< inputsize; j++) 
	            {
	            	// if the input is a stream
	            	//try {
	            	if(nodes[i].getChild(j) instanceof String)
	            		tuplecount += operator.getInputQueue(j).getQueueSize();
	            	//} catch (XATNode.OutOfBoundaryIndexException oobe) {
	            	//	break;
	            	//}
	            }
	        }
	        this.setPropertyValue("" + tuplecount);
	        return;
	    } 
	    
	    if (this.getAssociatedQueryObject() instanceof XATStreamOperator) 
	    {
			int tuplecount = 0;
			XATStreamOperator op = (XATStreamOperator) this.getAssociatedQueryObject();
			XATNode node = op.getNode();
			int inputsize = node.getChildrenSize();
			for (int i = 0; i < inputsize; i++) 
			{
				// if the input is a stream
				//try { 
					if (node.getChild(i) instanceof String)
						tuplecount += op.getInputQueue(i).getQueueSize();
				//} catch (XATNode.OutOfBoundaryIndexException oobe) {
				//	break;
				//}
			}
			this.setPropertyValue("" + tuplecount);
			return;
		}
	}
}