/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.calculators;

//import java.util.Iterator;

//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsObject;

/**
 * This will record the number of possible outputs for an operator.  Basically
 * this is equal to the numberOfTuplesToDequeueTotal is the operator only has
 * one child.  If the operator has multiple children, it is equal to the product
 * of the number of tuples outtputted.  The running total is kept here.  
 * @author: Brad Pielech
 */

public class NumberOfPossibleOutputs extends StatisticsObject 
{
	private long[] queueCounts = null;
	/**
	 * NumberOfPossibleOutputs constructor comment.
	 */
	public NumberOfPossibleOutputs() 
	{
		super();
	}
	
	/**
	 * NumberOfPossibleOutputs constructor comment.
	 * @param name java.lang.String
	 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
	 */
	public NumberOfPossibleOutputs(String name, XATQueryObject queryObject) 
	{
		super(name, queryObject);
	}
	
	/**
	 * Calculates the new Value of the given property of this operator.  This
	 * should be overwritten by subclasses who can tailor this method to calculating 1
	 * specific property. <p>
	 * This method should update at least the local propertyValue and possibly the
	 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
	 *
	 * The procedure for calculating new values should be as follows:
	 *
	 * <ol>
	 * <li> Consult the statsTable for any propertyValues that are needed.
	 * <li> Perform the required calculations using the propertyValues from step 1
	 * <li> Update the local propertyValue.
	 * 
	 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
	 * </ol> 
	 */
	public void calculateNewValue() 
	{
		XATStreamOperator op = (XATStreamOperator)this.getAssociatedQueryObject();
		this.setPropertyValue(op.getProperty("NUMBER_OF_POSSIBLE_OUTPUTS"));
	//	if(queueCounts == null){
	//		queueCounts = new long[op.getInputQueueCount()];
	//	}
		
		/*
		* iterate through the queues and figure out how many tuples this operator could have
		* outputted.  The statistic will contain how many tuples were available to the operator
		* before execution.  Therefore, we take this value and subtract the current available
		* tuples will yield how many the operator dequeued from the queue
		*
		* Then we will iterate through the queueCounts array and multiply the values by each
		* other to produce the maximum amount of tuples that could have been generated. 
		* For operators with only 1 queue, this should be equal to number of tuples dequeued total
		*/
	//	long total = 1;
	//	int count = 0;
	//	//if(op instanceof XATMultiQueueStreamOperator){
	//	//	int dummy = 0;
	//	//}
	//	for (Iterator i = op.getInputQueues(); i.hasNext(); ){
	//		XATQueue queue = (XATQueue)i.next();
	//		queueCounts[count] = Integer.parseInt(queue.getProperty("DEQUEUED_TUPLE_COUNT"));	
	////		int oldQueue = Integer.parseInt(stats.getValue("TUPLE_COUNT", queue));
	////		int newQueue = queue.getQueueSize(op);
	////		long difference = oldQueue - newQueue;
	////		queueCounts[count] = queueCounts[count] + difference;
	//		count++;
	//	}
	//	for (int i = 0; i < queueCounts.length; i++){
	//		//if(total == 0){
	//		//	total = queueCounts[i];
	//		//} else{
	//			total = total * queueCounts[i];
	//		//}
	//	}
	//	this.setPropertyValue("" + total);
	}
}
