/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.calculators;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsObject;

/**
 * the total number of tuples outputted NUMBER_OF_TUPLES_OUTPUTTED property plus the
 * existing aggregate value 
 */

public class NumberOfTuplesInEachState extends StatisticsObject 
{
	/**
	 * TimeTookToRunStatisticsObject constructor comment.
	 */
	public NumberOfTuplesInEachState() 
	{
		super();
	}
	
	/**
	 * TimeTookToRunStatisticsObject constructor comment.
	 * @param name java.lang.String
	 * @param operator edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator
	 */
	public NumberOfTuplesInEachState(String name, XATOperator operator) 
	{
		super(name, operator);
	}
	
	/**
	 * This is semantically the same as <tt>calculateNewValue()</tt>, but this
	 * provides a parameter which assists in the calculation.
	 * Calculates the new Value of the given property of this operator.  This
	 * should be overwritten by subclasses who can tailor this method to calculating 1
	 * specific property. <p>
	 * This method should update at least the local propertyValue and possibly the
	 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
	 *
	 * The procedure for calculating new values should be as follows:
	 *
	 * <ol>
	 * <li> Consult the statsTable for any propertyValues that are needed.
	 * <li> Perform the required calculations using the propertyValues from step 1
	 * <li> Update the local propertyValue.
	 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
	 * </ol> 
	 */
	public void calculateNewValue() 
	{
		OperatorState[] states = ((XATStreamOperator) this.getAssociatedQueryObject()).getState();
		
		if(states != null) 
		{
			this.setPropertyValue("");
			for(int i=0; i<states.length; i++) 
			{
				this.setPropertyValue(this.getPropertyValue() + states[i].getStateSize() + " ");
			}
		}
		else
		{
			this.setPropertyValue("null");
		}
	}
}
