/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.calculators;

import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
/**
 * This will look at all of the output queues of the associated query object 
 * (which has to be cast to a StreamOperator b/c XATOperators do not have queues)
 * and sum the TUPLE_COUNT properties 
 */
public class NumberOfTuplesInInputQueues extends TuplesInQueueCalculator 
{
	/**
	 * NumberOfTuplesInOutoutQueues constructor comment.
	 */
	public NumberOfTuplesInInputQueues() 
	{
		super();
	}
	/**
	 * NumberOfTuplesInOutoutQueues constructor comment.
	 * @param name java.lang.String
	 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
	 */
	public NumberOfTuplesInInputQueues(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) 
	{
		super(name, queryObject);
	}
	/**
	 * Calculates the new Value of the given property of this operator.  This
	 * should be overwritten by subclasses who can tailor this method to calculating 1
	 * specific property. <p>
	 * This method should update at least the local propertyValue and possibly the
	 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
	 *
	 * The procedure for calculating new values should be as follows:
	 *
	 * <ol>
	 * <li> Consult the statsTable for any propertyValues that are needed.
	 * <li> Perform the required calculations using the propertyValues from step 1
	 * <li> Update the local propertyValue.
	 * 
	 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
	 * </ol> 
	 */
	public void calculateNewValue() 
	{
		XATStreamOperator op = (XATStreamOperator)this.getAssociatedQueryObject();
		Iterator queues = op.getInputQueues();
		int count = 0;
		/** iterate through queues */
		for (;queues.hasNext();) 
		{
			XATQueue queue = (XATQueue)queues.next();
			
			//kew - start chg for QMesh
			if(!queue.isRegistered(op)) 
				queue.register(op);
			//kew - end chg for QMesh
			
			//if(!queue.isRegistered(op)) queue.register(op);
			if(!queue.isRegistered(op)) 
			{
			    System.out.println("Error in stats calculation of NumberOFTuplesInInputQueues!");
				System.out.println("op: " + op + " is not registered customer of queue: " + queue);
			}
			else 
			{
			    count += queue.getQueueSize(op);	
			    //System.out.println(count + " tuples in input queue sof op: " + op);
			}
			//queue.getQueueSize();
		}
		this.setPropertyValue("" + count);		
	}
}
