/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.calculators;

import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;

/**
 * Insert the type's description here.
 * Creation date: (5/19/2003 11:55:02 AM)
 * @author: Tim Sutherland
 */
public class OldestTupleInQueue extends edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsObject 
{
	/**
	 * OldestTupleInQueue constructor comment.
	 */
	public OldestTupleInQueue() 
	{
		super();
	}
	
	/**
	 * OldestTupleInQueue constructor comment.
	 * @param name java.lang.String
	 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
	 */
	public OldestTupleInQueue(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) 
	{
		super(name, queryObject);
	}
	
	/**
	 * Calculates the new Value of the given property of this operator.  This
	 * should be overwritten by subclasses who can tailor this method to calculating 1
	 * specific property. <p>
	 * This method should update at least the local propertyValue and possibly the
	 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
	 *
	 * The procedure for calculating new values should be as follows:
	 *
	 * <ol>
	 * <li> Consult the statsTable for any propertyValues that are needed.
	 * <li> Perform the required calculations using the propertyValues from step 1
	 * <li> Update the local propertyValue.
	 * 
	 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
	 * </ol> 
	 */
	public void calculateNewValue() 
	{
		XATStreamOperator op = (XATStreamOperator)this.getAssociatedQueryObject();
		Iterator queues = op.getInputQueues();
		long oldest = Long.MAX_VALUE;
		/** iterate through queues */
		for (;queues.hasNext();) 
		{
			/** iterate through each queue finding the oldest timestamp */
			XATQueue queue = (XATQueue)queues.next();
			if(queue.getQueueSize(op) > 0)
			{
				XATTuple tuple = queue.peek(op);
				oldest = Math.min(oldest,tuple.getMinTimeStamp());
	//			try{
	//				oldest = Math.min(oldest,tuple.getTimeStamp().getMinTimeStamp());
	//			} catch(XATTimeStampException e){
	//				//do nothing
	//			}
			}
			
			/*
			int queueSize = queue.getQueueSize();
			List l = queue.peekN(queueSize,op);
			Iterator i = l.iterator();
			for(;i.hasNext();){
				XATTuple tuple = (XATTuple) i.next();
				try{
				oldest = Math.min(oldest,tuple.getTimeStamp().getMinTimeStamp());
				}catch(XATTimeStampException e){ System.out.println(e.getMessage()); oldest = oldest;}	
			}
			*/	
		}
		// set the property now that we are finished
		this.setPropertyValue("" + oldest);		
	}
}
