/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Feb 22, 2005
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.calculators;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.SymmMwayHashJoinOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsObject;

import java.util.Enumeration;
import java.util.Hashtable;

/**
 * @author Bin Liu
 * To calculate the size of the symmhashjoinoperator, in terms of the number of tuples
 * in all the partitions of the operator
 */
public class PartitionSizeCalculator extends StatisticsObject 
{
	public PartitionSizeCalculator()
	{
		super();
	}
	
	public PartitionSizeCalculator(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) {
		super(name, queryObject);
	}

	public void calculateNewValue() 
	{
		/** 
		 * traverse tree and find the symmtrichashjoin operators  
		 */
		XATStreamOperator operator = (XATStreamOperator)this.getAssociatedQueryObject(); 
		if(operator instanceof SymmMwayHashJoinOperatorImp) 
		{
			Hashtable theStates = ((SymmMwayHashJoinOperatorImp)operator).getHashJoinStates();
			Enumeration keys = theStates.keys();
			double totalSize = 0;
			//double cur_unit_size = ((SymmMwayHashJoinOperatorImp)operator).getUnitTupleSize();
			double cur_base_unit_size = ((SymmMwayHashJoinOperatorImp)operator).getBaseUnitSize();
			while (keys.hasMoreElements()) 
			{ 
				Integer key = (Integer)keys.nextElement();
				//totalSize += ((HashStatePartition)theStates.get(key)).getTotalNumOfTuples()*cur_unit_size;
				totalSize += ((HashStatePartitionGroup)theStates.get(key)).getPartitionGroupSize()/cur_base_unit_size;
			}						
			this.setPropertyValue("" + (int)totalSize);
		}
	}	
}
