/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.calculators;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.AggregateStatisticsObject;
/**
 * Calculates the selectivity of an operator by dividing the total possible output tuples 
 * by the total tuples outputted.
 * If the operator is a single queue stream operator, the selectivity will be calculated by
 * using the total number of tuples dequeued (not possible outputs).  This is due to a timing
 * issue with the stream controller and execution where the number of possible outputs is
 * incorrect.  Using the total dequeued resolves this issue
 * 
 */
public class ReductionFactorCalculator extends AggregateStatisticsObject 
{
	/**
	 * SelectivityCalculator constructor comment.
	 */
	public ReductionFactorCalculator() 
	{
		super();	
	}
	/**
	 * SelectivityCalculator constructor comment.
	 * @param name java.lang.String
	 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
	 */
	public ReductionFactorCalculator(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) 
	{
		super(name, queryObject);	
	}
	/**
	 * Calculates the new Value of the given property of this operator.  This
	 * should be overwritten by subclasses who can tailor this method to calculating 1
	 * specific property. <p>
	 * This method should update at least the local propertyValue and possibly the
	 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
	 *
	 * The procedure for calculating new values should be as follows:
	 *
	 * <ol>
	 * <li> Consult the statsTable for any propertyValues that are needed.
	 * <li> Perform the required calculations using the propertyValues from step 1
	 * <li> Update the local propertyValue.
	 * 
	 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
	 * </ol> 
	 */
	public void calculateNewValue() 
	{
		Object oldCategory = stats.getCurrentCategory();
		stats.setCurrentCategory(this.getCategory());
		double outputted = 0;
		//if(this.getAssociatedQueryObject() instanceof XATSingleQueueStreamOperator){
		outputted = Double.parseDouble(stats.getValue("NUMBER_OF_TUPLES_OUTPUTTED", this.getAssociatedQueryObject())); 
		//}
		//else {
		//	outputted = Double.parseDouble(stats.getValue("NUMBER_OF_TUPLES_OUTPUTTED", this.getAssociatedQueryObject())); 
		//}
		double possible = 0;
		if(this.getAssociatedQueryObject() instanceof XATSingleQueueStreamOperator)
		{
			possible = Double.parseDouble(stats.getValue("NUMBER_OF_TUPLES_DEQUEUED", this.getAssociatedQueryObject()));
		} 
		else
		{
			possible = Double.parseDouble(stats.getValue("NUMBER_OF_POSSIBLE_OUTPUTS", this.getAssociatedQueryObject()));
		}
		try
		{
			if(possible == 0)
			{
				this.setPropertyValue("" + new Double(0.0));
				this.getAssociatedQueryObject().setProperty("REDUCTION_FACTOR", ""+Double.parseDouble(this.getPropertyValue()));		
			} 
			else
			{
				this.setPropertyValue("" + (1 - (outputted / possible)));
				this.getAssociatedQueryObject().setProperty("REDUCTION_FACTOR", ""+Double.parseDouble(this.getPropertyValue()));
			}
		}
		catch (Exception e)
		{
			e.printStackTrace();	
		}
		stats.setCurrentCategory(oldCategory);
	}
}
