/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.calculators;

import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsObject;
/**
 * Calculates the throughput of a tree. This is equal to the sum of the number of tuples the
 * roots have outputted
 * @author: Brad Pielech
 */
public class TreeThroughputCalculator extends StatisticsObject 
{
	/**
	 * OutputRateCalculator constructor comment.
	 */
	public TreeThroughputCalculator() 
	{
		super();
	}
	
	/**
	 * OutputRateCalculator constructor comment.
	 * @param name java.lang.String
	 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
	 */
	public TreeThroughputCalculator(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) 
	{
		super(name, queryObject);
	}
	/**
	 * This is semantically the same as <tt>calculateNewValue()</tt>, but this
	 * provides a parameter which assists in the calculation.
	 * Calculates the new Value of the given property of this operator.  This
	 * should be overwritten by subclasses who can tailor this method to calculating 1
	 * specific property. <p>
	 * This method should update at least the local propertyValue and possibly the
	 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
	 *
	 * The procedure for calculating new values should be as follows:
	 *
	 * <ol>
	 * <li> Consult the statsTable for any propertyValues that are needed.
	 * <li> Perform the required calculations using the propertyValues from step 1
	 * <li> Update the local propertyValue.
	 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
	 * </ol> 
	 */
	public void calculateNewValue() 
	{
	    Object oldCategory = stats.getCurrentCategory();
	    stats.setCurrentCategory(this.getCategory());
	    /** 
	     * set the tree's output rate to be that of its root
	     * if we have a DAG, then we will just average the average rates 
	    */
	    if (this.getAssociatedQueryObject() instanceof XATDag) 
	    {
	        XATDag dag = ((XATDag) this.getAssociatedQueryObject());
	        int throughput = 0;
	        for (Iterator i = dag.getRoots(); i.hasNext();) 
	        {
	            DagNode dagnode = (DagNode) i.next();
	            XATOperator operator = dagnode.getOperator();
	            throughput += Double.parseDouble(stats.getValue("NUMBER_OF_TUPLES_OUTPUTTED_TOTAL", operator));
	        }
	        //get the old throughput value and figure out how many we outputted since then 
	   
	        this.setPropertyValue("" + throughput);
	    } 
	    else 
	    {
	        //nope, dont have a dag	
	        XATOperator root =
	            ((XATTree) this.getAssociatedQueryObject()).getRoot().getOperator();
	
			double throughput = Double.parseDouble(stats.getValue("NUMBER_OF_TUPLES_OUTPUTTED_TOTAL", root));
	        this.setPropertyValue("" + throughput);
	    }
	    stats.setCurrentCategory(oldCategory);
	}
}
