/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.calculators;

import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsObject;
/**
 * This provides a convience method for summing the number of tuples in
 * output queues. It relies on the corresponding TUPLE_COUNT values that
 * are stored within the StatisticsGatherer
 */
public abstract class TuplesInQueueCalculator extends StatisticsObject 
{
	/**
	 * TuplesInQueueCalculator constructor comment.
	 */
	public TuplesInQueueCalculator() 
	{
		super();
	}
	/**
	 * TuplesInQueueCalculator constructor comment.
	 * @param name java.lang.String
	 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
	 */
	public TuplesInQueueCalculator(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) 
	{
		super(name, queryObject);
	}
	/**
	 * This is a convience method for the NumberOfTuplesInOutput and Input Queues
	 * Statistics Objects.  Rather than have both classes calculate the tuple count the same way,
	 * this method allows for the sharing of code.
	 * @param queues an Iteration of XATQueues
	 */
	public String calculateTupleCount(Iterator queues) 
	{	
		int count = 0;
		/** iterate through queues */
		for (;queues.hasNext();) 
		{
			XATQueue queue = (XATQueue)queues.next();
			count = count + queue.getQueueSize();		
		}	
		return "" + count;			
	}
}
