/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

/**
 * This class represents a Node in a DAG.
 * It need to extend from XATNode because
 * XATOperator only knows XATNode.
 * 
 * Creation date: (2003-1-9 22:03:39)
 * @author: JinHui Jian
 */

import diva.graph.modular.Node;
import diva.util.VisualObjectContainer;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

public interface DagNode extends XATNode, VisualObjectContainer{
/**
 * adds a new parent.  The parent is added to the right of the previous parents
 * @param parent edu.wpi.cs.dsrg.xmldb.xat.component.automata.common.DagNode
 */
void addParent(XATNode parent);


/**
 * Removes the specified parent
 * @param parent edu.wpi.cs.dsrg.xmldb.xat.component.automata.common.DagNode
 */
void deleteParent(XATNode parent);
void deleteAllParents();

/**
 * Returns the parent at the specified index
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.automata.common.DagNode
 * @param id int
 */
DagNode getParent(int id);

/**
 * Returns how many parents this DagNode has
 * @return int
 */
int getParentsSize();


/**
 * Returns the visual representation, a diva.graph.modular.Node, that is associated with this DagNode.
 * @return diva.graph.modular.Node
 */
Node getVisualObject();

/**
 * Set the specified parent 
 * @param id int
 * @param newParent the new XATNode
 */
public void setParent(int id, XATNode newParent);

/**
 * Set the visual representation, a diva.graph.modular.Node, associated with this DagNode.
 * @param node
 */
public void setVisualObject(Node node);

public DagNode makeNewCopy();

public boolean hasBeenVisited();
public void setHasBeenVisited(boolean temp);

}
