/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

/**
 * Insert the type's description here.
 * Creation date: (2003-1-9 22:13:21)
 * @author: JinHui Jian
 */
import java.util.Iterator;
import java.util.Set;
//import java.util.ArrayList;
import java.io.Serializable;
import java.lang.reflect.Constructor;

import diva.graph.modular.Node;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNodeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;

public class DagNodeImp extends XATNodeImp implements DagNode, Serializable {
	/**
	 * The diva visual representation of the DagNode.
	 */
	protected transient Node _visualObject = null;
	boolean visited = false; //used to mark the DagNode in all the traversal of the cyclic graph.
	/**
	 * DagNodeImp constructor comment.
	 */
	public DagNodeImp() {
		super();
	}
	/**
	 * adds a new parent.  The parent is added to the right of the previous parents
	 * @param parent edu.wpi.cs.dsrg.xmldb.xat.component.automata.common.DagNode
	 */
	public void addParent(XATNode newParent) {
		if (!parent.contains(newParent)) {
			parent.add(newParent);
		}
	}

	/**
	 * Removes the specified parent
	 * @param parent edu.wpi.cs.dsrg.xmldb.xat.component.automata.common.DagNode
	 */
	public void deleteParent(XATNode removeParent) {
		parent.remove(removeParent);
	}

	public void deleteAllParents() {
		parent.clear();
	}
	
	/**
	 * Returns the parent at the specified index
	 * @return edu.wpi.cs.dsrg.xmldb.xat.component.automata.common.DagNode
	 * @param id int
	 */
	public DagNode getParent(int id) {
		return (DagNode) parent.get(id);
	}
	/**
	 * Returns an Iterator of all of the parents
	 * @return Iterator of parents
	 */
	public Iterator getParents() {
		return parent.iterator();
	}
	/**
	 * Returns how many parents this DagNode has
	 * @return int
	 */
	public int getParentsSize() {
		return parent.size();
	}


	
	/**
	 * Returns the diva node which is the visual representation of this DagNode.
	 * @return diva.graph.modular.Node
	 */
	public Node getVisualObject(){
		return _visualObject;
	}
	
	/**
	 * Set the specified parent 
	 * @param id int
	 * @param newParent the new XATNode
	 */
	public void setParent(int id, XATNode newParent) {
		parent.set(id, newParent);
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (2003-2-4 14:23:01)
	 * @return java.util.Set
	 * @param newSchema java.util.Set
	 */
	public Set setSchema(Set newSchema) {
		Set old = this.schema;
		this.schema = newSchema;
		return old;
	}
	
	/**
	 * Sets this DagNode to the diva node which is its visual representation.
	 */
	public void setVisualObject(Node node){
		_visualObject = node;
	}
	
	/**
	 * Make a new copy of the current node. This is a deep copy, not bit by bit copy as the default clone in java.
	 * This is only called by the makeNewCopy make of the XATBox class.
	 * @return
	 */
	public DagNode makeNewCopy(){
		DagNode newnode = new DagNodeImp();
		
		newnode.setSchema(this.schema);
		newnode.setTreeID(this.treeID);
		//newnode.setMachine(this.machine);
		
		Constructor construct;
		Class[] c = new Class[2];
		c[0] = this.operator.getClass();
		c[1] = XATNode.class;
		try {
			construct = operator.getClass().getConstructor(c);
			newnode.setOperator((XATStreamOperator)construct.newInstance(new Object[] {this.operator, newnode }));
		} catch (Exception e) {
			e.printStackTrace();
		}
		//newnode.setOperator(this.operator.makeNewCopy());

		
		// we won't create new children and parent lists here since
		// these would be done when making new copies of an XATBox.
		
		return newnode;
	}
	
	public boolean hasBeenVisited(){
		return visited;
	}
	public void setHasBeenVisited(boolean temp){
		visited=temp;
	}
}
