/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

import java.io.Serializable;
/**
 * A timestamp for a range of columns in an XATTuple
 * Creation date: (5/22/2003 10:17:14 AM)
 * @author: Tim Sutherland
 */
public class TimeStamp implements Serializable, Cloneable {
	private int startColumn;
	private int endColumn;
	private Long timestamp;
/**
 * TimeStamp constructor comment.
 */
public TimeStamp() {
	timestamp = null;
	startColumn = 0;
	endColumn = 0;
}
/**
 * Another Constructor
 * Creation date: (5/22/2003 10:23:06 AM)
 * @param startColumn int
 * @param endColumn int
 */
public TimeStamp(int startColumn, int endColumn) {

	this.startColumn = startColumn;
	this.endColumn = endColumn;
	this.timestamp = null;
	
	}
/**
 * Another Constructor
 * Creation date: (5/22/2003 10:22:10 AM)
 * @param timestamp long
 */
public TimeStamp(long timestamp) {

	this.timestamp = new Long(timestamp);
	startColumn = 0;
	endColumn = 0;
	}
/**
 * Another Constructor
 * Creation date: (5/22/2003 10:24:16 AM)
 * @param timestamp long
 * @param startColumn int
 * @param endColumn int
 */
public TimeStamp(long timestamp, int startColumn, int endColumn) {

	this.timestamp = new Long(timestamp);
	this.startColumn = startColumn;
	this.endColumn = endColumn;
	
	}
/**
 * Returns the last column in the range of the timestamp.
 * Creation date: (5/23/2003 6:15:52 PM)
 * @return int
 */
public int getEndColumn() {
	return endColumn;
}
/**
 * Returns the first column in the range of the timestamp.
 * Creation date: (5/23/2003 6:15:52 PM)
 * @return int
 */
public int getStartColumn() {
	return startColumn;
}
/**
 * Returns the timestamp.
 * Creation date: (5/22/2003 10:25:40 AM)
 * @return long
 */
public long getTimeStamp() {
	return this.timestamp.longValue();
}
/**
 * Sets the end column of the range
 * Creation date: (5/23/2003 6:15:52 PM)
 * @param newEndColumn int
 */
public void setEndColumn(int newEndColumn) {
	endColumn = newEndColumn;
}
/**
 * Sets the start column of the range
 * Creation date: (5/23/2003 6:15:52 PM)
 * @param newStartColumn int
 */
public void setStartColumn(int newStartColumn) {
	startColumn = newStartColumn;
}
/**
 * Sets the timestamp
 * Creation date: (5/22/2003 10:26:12 AM)
 * @param timestamp long
 */
public void setTimeStamp(long timestamp) throws Exception{

	if(this.timestamp == null)
		this.timestamp = new Long(timestamp);
	else
		throw new Exception("The timestamp for this object has already been set.");
		
	}

/**
 * Insert the method's description here.
 * Creation date: (5/25/2003 7:31:44 PM)
 * @return int
 */
public int getRange() {
	return endColumn - startColumn + 1;
}

/**
 * Insert the method's description here.
 * Creation date: (5/25/2003 7:33:04 PM)
 * @return boolean
 * @param column int
 * @exception java.lang.Exception The exception description.
 */
public boolean inRange(int column){

	if(column >= startColumn && column <= endColumn)
		return true;
	else
		return false;
}

/**
 * Insert the method's description here.
 * Creation date: (5/27/2003 11:08:37 AM)
 * @return java.lang.String
 */
public String toString() {
	return "TimeStamp: " + timestamp + " Start Col: " + startColumn + " End Col: " + endColumn;
}

public Object clone() {
	TimeStamp ts = new TimeStamp();
	ts.setEndColumn(this.endColumn);
	ts.setStartColumn(this.startColumn);
	try {
		ts.setTimeStamp(this.getTimeStamp());
	} catch(Exception e) {
		e.printStackTrace(System.out);
	}
	return ts;
}
}
