/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

import java.lang.reflect.Array;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;
import java.io.Serializable;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperatorImp;
/**
 * Insert the type's description here.
 * Creation date: (1/31/03 4:40:07 PM)
 * @author: Brad Pielech
 */
public class XATDagImp extends edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp implements XATDag, Serializable {
/**
 * Extends XATTree to allow for multiple roots.  Several convience methods are provided
 */
public XATDagImp() {
	super();
	
	}
/**
 * Adds a new root to the Dag.  This will be added to the right of the others
 * @param newRoot edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode
 */
public void addRoot(DagNode newRoot) {	
	root.add(newRoot);	
}

public int size(){
	return (this.convertToArray()).length;
}
/*
 * cannot inherit from XATTree's appendChildLeaves method
 * in a dag, a leaf node may be reached more than one time, make sure no duplicates are stored
 * Creation date: (4/13/2003 3:17:40 PM)
 * @param v java.util.Vector
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */

protected void appendChildLeaves(Vector v, XATNode node) {
	if (((DagNode)node).hasBeenVisited()) return;
	((DagNode)node).setHasBeenVisited(true);
	XATNode[] children = node.getOpChildren();

	int i;
	int length = Array.getLength(children);
	if (length == 0) {
		if (!(v.contains(node)))
			v.add(node);
		return;
	}
	for (i = 0; i < length; i++) {
		appendChildLeaves(v, children[i]);
	}

}

/**
	* Converts this tree into an array, similar to getLeaves()
	* @return an array of XATNodes 
	*/
public edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode[] convertToArray() {

    //System.out.println("ConvertToArray from XATDag");
    
//    if(this.root.size() == 1) {
//        return super.convertToArray();
//    }
    
	Set resultSet = new HashSet();
  try {

	//Do breath first traverse
	Set curSet = new HashSet();//contains nodes on current visiting level
	Set nextSet = new HashSet();//contains nodes on next visiting level

		//traversal begins from root
	for (Iterator i= this.getRoots(); i.hasNext();)
		curSet.add((XATNode)i.next());
	
	do {
			//for each node in current level
		for (Iterator it = curSet.iterator(); it.hasNext();) {
		XATNode curNode = (XATNode) it.next();

		//visit this node
		resultSet.add(curNode);

		if (curNode.getOpChildrenSize() > 0) {
		  XATNode[] children = curNode.getOpChildren();

		  //push each child in next level
		  for (int i = 0; i < children.length; i++) {
			XATNode curChild = (XATNode) children[i];

			if (resultSet.contains(curChild)) continue;
//						//flag == true means all parents are visited
//			boolean flag = true;
//			//DagNode[] prevLevel = curChild.getParents();
//			//for (int j = 0; j < prevLevel.length; j++) {
//			for(Iterator prevIt = curChild.getParents(); prevIt.hasNext();){
//				XATNode curPrev = (XATNode)prevIt.next();
//				if (!resultSet.contains(curPrev)) {
//				flag = false;
//				break;
//			  }
//			}
//
//			if (flag)
			  nextSet.add(curChild);
		  }//push each child
		} //push next level

	  } //for

	  curSet = nextSet;
	  nextSet = new HashSet();
	} while (!curSet.isEmpty());

  } catch (Exception e) {
	e.printStackTrace();
  }

  return (XATNode[]) resultSet.toArray(new XATNode[]{});
}

/**
 * Creation date: (2002-12-20 15:57:08)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public static XATDag convertTreeToDag(XATTree xattree) {

	XATDag xatdag = new XATDagImp();
	
	//make logical plan by
	//Top-down visit
	xatdag.setRoot(visit(xattree.getRoot()));
  
  return xatdag;
}

/**
 * Traverse the tree and add the node one by one into DAGNode.
 *
 * Creation date: (2002-12-20 16:30:32)
 * @param op edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavUnnest
 */
private static DagNode visit(XATNode node) {

	DagNode curNode = new DagNodeImp();

	//pre-order Top-down visit
	XATStreamOperator curOp = (XATStreamOperator)node.getOperator().clone();
	curNode.setOperator(curOp);
	curOp.setNode(curNode);
	
	Iterator it = node.getStreamChildNames();
	
	while(it.hasNext()){
		curNode.addStreamChild((String)it.next());
	}
	
	XATNode[] children = node.getOpChildren();
	for(int i = 0; i < children.length; i++){
		// recursively visit the child node.
		DagNode childNode = visit(children[i]);
		// add the child to the parent.
		curNode.addOpChild(childNode);
		childNode.setParent(curNode);
	}

	return curNode;
}


/**
 * Override the getLeaves() in the XAT tree.
 * Creation date: (4/3/2003 1:22:08 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode[]
 */
public XATNode[] getLeaves() {
	this.clearHasBeenVisitedMark();
	Iterator roots = this.getRoots();
	Vector children = new Vector();
	while(roots.hasNext()){
		this.appendChildLeaves(children,(XATNode)roots.next());
	}
	return (XATNode[])(children.toArray(new XATNode[]{}));
}
/**
 * Insert the method's description here.
 * Creation date: (1/31/03 4:48:42 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode
 * @param i int
 */
public DagNode getNthRoot(int i) {
	return (DagNode)root.get(i);
}

public void deleteRoot(DagNode node) {
	root.remove(node);
}

/**
 * Returns true if node is a root node; false otherwise.
 * @return boolean
 */
public boolean isRoot(DagNode node){
	return root.contains(node);
}

/**
 * Insert the method's description here.
 * Creation date: (1/31/03 4:49:34 PM)
 * @param i int
 * @param newRoot edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode
 */
public void setNthRoot(int i, DagNode newRoot) {
	
	root.set(i, newRoot);
	
	}

public int getRootCount() {
	
	return root.size();
	
	}
public String toString() {
	this.clearHasBeenVisitedMark();
	String output = new String("");
	for (Iterator i = getRoots();i.hasNext();){
		output+=childrenString((XATNode)i.next(), "");
	}
    return output;
}
protected String childrenString(XATNode node, String indent) 
{
	if (((DagNode)node).hasBeenVisited()) return "";
	((DagNode)node).setHasBeenVisited(true);
    if (node == null)
        return indent + "\n";
    String sOutput = indent + node.toString() + "\n";
    String newIndent = indent + "  ";
    XATNode[] children = node.getOpChildren();
    for (int i = 0; i < children.length; i++) {
        sOutput = sOutput + childrenString(children[i], newIndent);
    }
    return sOutput;
}
//used for clear the hasBeenVisited boolean in cyclic graph traversal
public void clearHasBeenVisitedMark(){
	for (int i=0;i<root.size(); i++){
		clearMark((DagNode)root.get(i));
	}
}

private void clearMark(DagNode root){
	if (!root.hasBeenVisited()) return;
	((DagNode)root).setHasBeenVisited(false);// pre-order setting
	for (int i = 0; i < root.getChildrenSize(); i++) {
		Object name = root.getChild(i);
		if (name instanceof DagNode) {
			clearMark((DagNode) root.getChild(i));
		}
	}
	
}
}
