/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

/**
 * Implements XATQueue.
 * All methods are synchronized for multi-thread accessing.
 * 
 * Creation date: (4/27/2003 1:23:37 AM)
 * @author: Jinhui Jian
 */
import java.util.ArrayList;
//import java.util.Iterator;
import java.util.List;
//import java.util.Set;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;

/**
 * This class implements a queue where all tuples are always stored in main
 * memory. It is assumed that there will always be enough memory to process the
 * tuples, so no backup persistent storage is required.
 * 
 * Creation Date: (Jul 27, 2004 9:53:48 AM)
 */
public class XATMemoryQueueImp extends XATQueueImp 
{
	/**
	 * to indicate how many tuples have been cleaned up? debug code
	 * to test whether the java runtime gc() is useful?
	 */
	int have_been_cleaned_up = 0;
	
	public void printOutCursorInfo()
	{
		System.out.println("MINI CURSOR:" + this.minCursor + "CACH CURSOR:" + this.cachedCursor + "QID:" + this.queueNumber + "QOWN" + this.queueOwners.toString());
	}
	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:37:04 PM)
	 */
	public XATMemoryQueueImp() 
	{
		super();
	}
	
	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:36:59 PM)
	 * @param owner Object
	 */
	public XATMemoryQueueImp(Object owner) 
	{
		super(owner);
	}
	
	/**
	 * XATQueueImp constructor.
	 * 
	 * Creation Date: (Jun 16, 2004 2:36:53 PM)
	 * @param owner Object
	 * @param queueName String
	 */
	public XATMemoryQueueImp(Object owner, String queueName) 
	{
		super(owner, queueName);
	}

	/**
	 * Clean up tuples that have been dequeued by all customers.
	 *
	 * Creation date: (4/27/2003 3:10:56 AM)
	 */
	public synchronized void cleanTuples() 
	{
		// a short-cut to address the problem of multiple customers problem when distributed
		// query operator accross machines, in real situation, it has one customer only
		// need further debug to support. Probably in the connectionmanagerImp and connectionlistener -- Bin
		// short-cut code by Bin
		if (this.minCursor <= 0) 
		{
			this.minCursor = Integer.MAX_VALUE;
		}
		// end of short-cut code

		// if no more customers, clear the queue
		if(this.cachedCustomer == null)
		{
			this.contents.clear();
		}

		// otherwise, clean up tuples when necessary.
		// the number "5" is a tradeoff between speed and storage:
		// larger number requires less cleanup,
		// smaller number reduces the storage overhead for old tuples
		else if (this.cachedCursor > CLEANTHRESHOLD && this.minCursor > CLEANTHRESHOLD) 
		{			
			int min = Math.min(this.minCursor, this.cachedCursor); 
			
			// update the cached customer
			this.cachedCursor -= min;
			
			/* -- commented by Bin, for temporarily ignore the multiple customers case
			// update the rest  of the customers
			if (this.customerMap != null) {
				Set keySet = this.customerMap.keySet();
				for (Iterator it = keySet.iterator(); it.hasNext();) {
					Object curCustomer = it.next();
					XATQueueCursor oldCursor = (XATQueueCursor) this.customerMap.get(curCustomer);
					oldCursor.setCursor(oldCursor.getCursor() - min);					
				}

				// also update the minimal cursor
				this.minCursor -= min;
			}
			*/
			
			//debug, call java gc() very 2000 tuples have been cleanup?
			//if (min >0 ) {
			//	System.out.println("MIN=:" + min + "Threashold:" + CLEANTHRESHOLD);
			//	printOutCursorInfo();				
			//}
			
			have_been_cleaned_up += min;
			if (have_been_cleaned_up > 2000) 
			{
				//System.out.println("=====HAVE CLEANED UP MORE THAN 2000 TUPLES. TRY GC=====");
				/////((ArrayList)this.contents).trimToSize();
				//Runtime.getRuntime().gc();
				have_been_cleaned_up = 0;
			}
			
			// clean up tuples
			this.contents.subList(0, min).clear();
//			for (int i = 0; i < min; i++) {
//				this.contents.remove(0);
//			}
		}
	}
	
	/**
	 * Dequeues a number of tuples into an array.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:22 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	public synchronized XATTuple[] dequeue(int number, Object customer) 
	{		
		// make sure the customer can dequeue
		if (!this.isRegistered(customer)) 
		{
			this.register(customer);
		}

		// also make sure there are enough tuples to dequeue
		// if not, set to maximum possible to dequeue for this customer
		if (number > this.getQueueSize(customer)) 
		{
			number = this.getQueueSize(customer);
		}
		
		// get the tuples
		int cursor = this.getCursor(customer);
		XATTuple[] result = new XATTuple[number];
		for (int i = 0; i < number; i++) 
		{
			result[i] = (XATTuple)this.contents.get(cursor + i);	
		}
		
		this.setCursor(customer, cursor+number);
		if(customerMap.size() == 0) 
		{
			this.minCursor = cursor+number;
		}
		
		// clean up the queue
		this.cleanTuples();
		
		// update queue attributes
		//this.setCursor(customer, cursor + number);
		this.totalDequeued += result.length;
		
		return result;
	}

	/**
	 * Dequeues a tuple.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:31 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	public synchronized XATTuple dequeue(Object customer) throws XATQueueException 
	{
		// make sure the customer can dequeue
		if (!this.isRegistered(customer))
			this.register(customer);
			
		// also make sure there is a tuple to dequeue
		// if not, then cannot dequeue
		if (this.getQueueSize(customer) == 0) 
		{
			throw new XATQueueException("Not enough tuples to dequeue");
		}
		else 
		{		
			// get the tuple
			int cursor = this.getCursor(customer);
			XATTuple result = (XATTuple)this.contents.get(this.getCursor(customer));
			
			this.setCursor(customer, cursor+1);
			if(customerMap.size() == 0) 
			{
				this.minCursor = cursor+1;
			}
			
			// clean up the queue
			this.cleanTuples();
			
			// update queue attributes
			//this.setCursor(customer, this.getCursor(customer) + 1);
			this.totalDequeued++;
		
			return result;
		}
	}
	
	/**
	 * Dequeues a number of tuples into a List.
	 * 
	 * Creation Date: (Jun 16, 2004 2:34:18 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	public synchronized List dequeueN(int number, Object customer) 
	{		
		// make sure the customer can dequeue
		if (!this.isRegistered(customer)) 
		{
			this.register(customer);
		}

		// also make sure there are enough tuples to dequeue
		// if not, set to maximum possible to dequeue for this customer
		if (number > this.getQueueSize(customer)) 
		{
			number = this.getQueueSize(customer);
		}
		
		// get the tuples
		int cursor = this.getCursor(customer);
		List result = new ArrayList(number);
		for(int i = 0; i < number; i++) 
		{
			result.add(this.contents.get(cursor + i));
		}

		// clean up the queue
		this.setCursor(customer, cursor + number);
		if(customerMap.size() == 0) 
		{
			this.minCursor = cursor+1;
		}
		
		this.cleanTuples();

		// update queue attributes
		//this.setCursor(customer, cursor + number);
		this.totalDequeued += result.size();
		
		return result;
	}

	/**
	 * Enqueues a list of tuples.
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation date: (Jun 15, 2004 11:13:36 AM)
	 * @param newTuples List
	 * @param provider Object
	 */
	public synchronized void enqueue(List newTuples, Object provider) 
	{
		// add the tuples
		this.contents.addAll(newTuples);
		
		// update queue attributes
		this.totalEnqueued += newTuples.size();
		/*
		if (this.isFinalQueue) {
			for (int i=0; i<newTuples.size(); i++) {
				XATTuple tuple = (XATTuple)newTuples.get(i);
				if (!tuple.isPunctuation()) {
					this.appendNewResult(tuple.toString()+"\n");
				}
			}
			if (this.totalEnqueued%100 == 0) {
				this.appendNewResult("Produced "+totalEnqueued+" tuples\n");
			}
		} else {
			this.contents.addAll(newTuples);
		}*/
	}

	/**
	 * Enqueues a tuple. 
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation Date: (Jun 16, 2004 2:36:21 PM)
	 * @param newTuple XATTuple
	 * @param provider Object
	 */
	public synchronized void enqueue(XATTuple newTuple, Object provider) 
	{
		// add the tuple
		this.contents.add(newTuple);
		
		// update queue attributes
		this.totalEnqueued++;
		
		/*
		if (this.isFinalQueue) {
			if (!newTuple.isPunctuation()) {
				this.appendNewResult(newTuple.toString()+"\n");
				if (this.totalEnqueued%100 == 0) {
					this.appendNewResult("Produced "+totalEnqueued+" tuples\n");
				}
			}
		} else {
			this.contents.add(newTuple);
		}*/
	}

	/**
	 * Enqueues an array of tuples. 
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param tuples edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[] 
	 * @param provider java.lang.Object
	 */
	public synchronized void enqueue(XATTuple[] newTuples, Object provider) 
	{
		// add the tuples
		int numNewTuples = newTuples.length;
		for(int i = 0; i < numNewTuples; i++) 
		{
			this.contents.add(newTuples[i]);
		}
		
		// update queue attributes
		this.totalEnqueued += numNewTuples;
		/*
		if (this.isFinalQueue) {
			for(int i=0; i<newTuples.length; i++) {
				if (!newTuples[i].isPunctuation()) {
					this.appendNewResult(newTuples[i].toString()+"\n");
				}
			}
			if (this.totalEnqueued%100 == 0) {
				this.appendNewResult("Produced "+totalEnqueued+" tuples\n");
			}
		} else {
			for(int i = 0; i < newTuples.length; i++) {
				this.contents.add(newTuples[i]);
			}
		}*/
	}

	/**
	 * Returns a light weight query object of this queue.
	 * 
	 * Creation Date: (Jun 16, 2004 2:39:22 PM)
	 * @return LightWeightQueryObject
	 */
	public LightWeightQueryObject getLightWeightQueryObject() 
	{
		return new LightWeightQueryObject("XATMemoryQueue", this.getClass().getName(), true);
	}

	/**
	 * Gets the string representation of the queue's layout.
	 * 
	 * Creation Date: (Jun 29, 2004 4:14:05 PM)
	 * @return String
	 */
	public String getQueueLayout() 
	{
		return "|T->" + this.getQueueSize() + "|";
	}

	/**
	 * Returns the number of tuples that are in the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:40:12 PM)
	 * @return int
	 */
	public synchronized int getQueueSize() 
	{
		return this.contents.size();
	}

	/**
	 * position starts from 0
	 * @param position
	 * @param tuple
	 * @param provider
	 * used for inserting migration sync flag in stream queues.
	 */
	public synchronized void insertTuple(int position, XATTuple tuple, Object provider) 
	{
		this.contents.add(cachedCursor+position, tuple);
	}
	 
	/**
	 * Retrives a number of tuples from the queue. This is analagous to 
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:43:59 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	public synchronized XATTuple[] peek(int number, Object customer) 
	{	
		// make sure the customer can peek
		if (!this.isRegistered(customer)) 
		{
			this.register(customer);
		}

		// also make sure there are enough tuples to peek
		// if not, set to maximum possible to peek for this customer
		if (number > this.getQueueSize(customer)) 
		{
			number = this.getQueueSize(customer);
		}
		
		// get the tuples
		int cursor = this.getCursor(customer);
		XATTuple[] result = new XATTuple[number];
		for(int i = 0; i < number; i++) 
		{
			result[i] = (XATTuple)this.contents.get(cursor + i);
		}
		
		return result;
	}
	
	/**
	 * Retrieves a tuple from the queue. This is analagous to dequeue, but
	 * this will not increment or cleanup cursors.
	 *
	 * Creation Date: (Jun 16, 2004 4:45:49 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	public synchronized XATTuple peek(Object customer) throws XATQueueException 
	{
		// make sure the customer can peek
		if (!this.isRegistered(customer)) 
		{
			this.register(customer);
		}
	
		// also make sure there is a tuple to peek
		// if not, then cannot peek
		if (this.getQueueSize(customer) == 0) 
		{
			throw new XATQueueException("Not enough tuples to peek.");
		}
			
		// get the tuple
		XATTuple result = (XATTuple)this.contents.get(this.getCursor(customer));
		
		return result;
	}

	/**
	 * Retrieves the last tuple in the queue. This is for XATSubQueueImp.
	 * XATSubQueueImp needs to merge a "subtuple" into the last tuple 
	 * (which must also be a "subtuple") to form a "full tuple". This is 
	 * analagous to dequeue, but this will not increment or cleanup 
	 * cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:47:31 PM)
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	public synchronized XATTuple peekLast() throws XATQueueException 
	{	
		// make sure there is a tuple to peek
		// if not, throw an exception
		if (this.getQueueSize() == 0) 
		{
			throw new XATQueueException("Not enough tuples to peek.");
		}
		
		return (XATTuple)contents.get(this.getQueueSize() - 1);
	}
	
	/**
	 * Retrieves a number of tuples from the queue. This is analagous to
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:49:09 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	public synchronized List peekN(int number, Object customer) 
	{
		// make sure the customer can peek
		if (!this.isRegistered(customer)) 
		{
			this.register(customer);
		}
	
		// also make sure there are enough tuples to peek
		// if not, set to maximum possible to peek for this customer
		if (number > this.getQueueSize(customer)) 
		{
			number = this.getQueueSize(customer);
		}
		
		// get the tuples
		int cursor = this.getCursor(customer);
		List result = new ArrayList();
		for(int i = 0; i < number; i++) 
		{
			result.add(this.contents.get(cursor + i));
		}
		
		return result;
	}
	
	/**
	 * Sets the tuples of this queue to the new tuples.
	 * 
	 * Creation Date: (Jun 16, 2004 5:13:20 PM)
	 * @param newTuples List
	 */
	public synchronized void setTuples (List newTuples) 
	{
		this.contents = newTuples;
	}
	
	/**
	 * Returns the contents of the queue by calling toString() on each 
	 * tuple. Caller is the consumer of the queue you want to print out.
	 * Caller can be null if this is the top output queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:50:17 PM)
	 * @param caller Object
	 * @return String
	 */
	public synchronized String printContent(Object customer) 
	{
	    StringBuffer rs = new StringBuffer();
	
	    rs.append(this.getQueueSize());
	    rs.append(" tuples: \n");
	
	    if (customer != null) 
	    {
	    	int number = this.getQueueSize(customer);
	    	XATTuple[] result = this.peek(number, customer);
	        for (int i = 0; i < number; i++) 
	        {
	            rs.append(result[i].toString() + "\n");
	        }
	    }
	    else 
	    {
	    	customer = "new-customer";
	    	this.register(customer);
	    	int number = this.getQueueSize(customer);
	    	XATTuple[] result = this.peek(number, customer);
			for (int i = 0; i < number; i++) 
			{
				rs.append(result[i].toString() + "\n");
			}
	    }
	    
	    return rs.toString();
	}
	
	/**
	 * Skips the next tuple.
	 *
	 * Creation Date: (Jun 16, 2004 5:14:34 PM)
	 * @param customer Object
	 * @throws XATQueueException
	 */
	public synchronized void skip(Object customer) throws XATQueueException 
	{	
		// check validity
		if(this.getQueueSize(customer) == 0)
			throw new XATQueueException("Not enough tuples to skip!");
			
		// increase cursor
		this.setCursor(customer, this.getCursor(customer) + 1);		
	}
	
	/**
	 * Skips a number of tuples.
	 *
	 * Creation Date: (Jun 16, 2004 5:16:24 PM)
	 * @param number int
	 * @param customer Object
	 */
	public synchronized void skipN(int number, Object customer) throws XATQueueException 
	{
		// check validity
		if(number > this.getQueueSize(customer))
			throw new XATQueueException("Not enough tuples to dequeue!");
			
		// increase cursor
		this.setCursor(customer, this.getCursor(customer) + number);		
	}
	
	/**
	 * Parameter "number" starts from 0.
	 */
	public synchronized XATTuple peekNthTuple(int number, Object customer) 
	{		
		// make sure the customer can peek
		if (!this.isRegistered(customer)) 
		{
			this.register(customer);
		}
	
		// also make sure there are enough tuples to peek
		// if not, set to maximum possible to peek for this customer
		if (number > this.getQueueSize(customer)) 
		{
			number = this.getQueueSize(customer);
		}
		
		// get the tuples
		int cursor = this.getCursor(customer);

		// get the tuples
		XATTuple result = (XATTuple)this.contents.get(cursor + number);
		
		return result;
	}
	
	@Override
	public void enqueue(XATTupleNegativeImp newTuple, Object provider) 
	{
		// TODO Auto-generated method stub		
	}
	@Override
	public void enqueue(XATTupleNegativeImp[] newTuples, Object provider) 
	{
		// TODO Auto-generated method stub		
	}
}
