/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

import java.util.List;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;

public abstract class XATQueueNegativeImp extends XATQueueImp{

	public XATQueueNegativeImp() {
		super();
		// TODO Auto-generated constructor stub
	}
	

	/*------------- BEGIN ABSTRACT CLASSES ----------------------------------*/	
	/**
	 * Clean up tuples that have been dequeued by all customers.
	 *
	 * Creation date: (4/27/2003 3:10:56 AM)
	 */
	abstract public void cleanTuples();
	
	/**
	 * Dequeues a number of tuples into an array.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:22 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	abstract public XATTuple[] dequeue(int number, Object customer);

	/**
	 * Dequeues a tuple.
	 *
	 * Creation Date: (Jun 16, 2004 2:33:31 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	abstract public XATTuple dequeue(Object customer) throws XATQueueException;
	
	/**
	 * Dequeues a number of tuples into a List.
	 * 
	 * Creation Date: (Jun 16, 2004 2:34:18 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	abstract public List dequeueN(int number, Object customer);

	/**
	 * Enqueues a list of tuples.
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation date: (Jun 15, 2004 11:13:36 AM)
	 * @param newTuples List
	 * @param provider Object
	 */
	abstract public void enqueue(List newTuples, Object provider);

	/**
	 * Enqueues a tuple. 
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation Date: (Jun 16, 2004 2:36:21 PM)
	 * @param newTuple XATTuple
	 * @param provider Object
	 */
	abstract public void enqueue(XATTuple newTuple, Object provider);

	/**
	 * Enqueues an array of tuples. 
	 * Note: the provider is reserved for future usage.
	 *
	 * Creation date: (4/27/2003 1:23:37 AM)
	 * @param tuples edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[] 
	 * @param provider java.lang.Object
	 */
	abstract public void enqueue(XATTuple[] newTuples, Object provider);
	
	/**
	 * Returns a light weight query object of this queue.
	 * 
	 * Creation Date: (Jun 16, 2004 2:39:22 PM)
	 * @return LightWeightQueryObject
	 */
	abstract public LightWeightQueryObject getLightWeightQueryObject();

	/**
	 * Gets the string representation of the queue's layout.
	 * 
	 * Creation Date: (Jun 29, 2004 4:14:05 PM)
	 * @return String
	 */
	abstract public String getQueueLayout();
	
	/**
	 * Returns the total number of tuples that are in the queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:40:12 PM)
	 * @return int
	 */
	abstract public int getQueueSize();

	/**
	 * Retrives a number of tuples from the queue. This is analagous to 
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:43:59 PM)
	 * @param number int
	 * @param customer Object
	 * @return XATTuple[]
	 */
	abstract public XATTuple[] peek(int number, Object customer);
	
	/**
	 * Retrieves a tuple from the queue. This is analagous to dequeue, but
	 * this will not increment or cleanup cursors.
	 *
	 * Creation Date: (Jun 16, 2004 4:45:49 PM)
	 * @param customer Object
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	abstract public XATTuple peek(Object customer) throws XATQueueException;

	/**
	 * Retrieves the last tuple in the queue. This is for XATSubQueueImp.
	 * XATSubQueueImp needs to merge a "subtuple" into the last tuple 
	 * (which must also be a "subtuple") to form a "full tuple". This is 
	 * analagous to dequeue, but this will not increment or cleanup 
	 * cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:47:31 PM)
	 * @return XATTuple
	 * @throws XATQueueException
	 */
	abstract public XATTuple peekLast() throws XATQueueException;
	
	/**
	 * Retrieves a number of tuples from the queue. This is analagous to
	 * dequeue, but this will not increment or cleanup cursors.
	 * 
	 * Creation Date: (Jun 16, 2004 4:49:09 PM)
	 * @param number int
	 * @param customer Object
	 * @return List
	 */
	abstract public List peekN(int number, Object customer);

	abstract public XATTuple peekNthTuple(int number, Object customer);
	/**
	 * Returns the contents of the queue by calling toString() on each 
	 * tuple. Caller is the consumer of the queue you want to print out.
	 * Caller can be null if this is the top output queue.
	 * 
	 * Creation Date: (Jun 16, 2004 4:50:17 PM)
	 * @param caller Object
	 * @return String
	 */
	abstract public String printContent(Object caller);
		
	/**
	 * Sets the tuples of this queue to the new tuples.
	 * 
	 * Creation Date: (Jun 16, 2004 5:13:20 PM)
	 * @param newTuples List
	 */
	abstract public void setTuples (List newTuples);
	
	/**
	 * Skips the next tuple.
	 *
	 * Creation Date: (Jun 16, 2004 5:14:34 PM)
	 * @param customer Object
	 * @throws XATQueueException
	 */
	abstract public void skip(Object customer) throws XATQueueException;
	
	/**
	 * Skips a number of tuples.
	 *
	 * Creation Date: (Jun 16, 2004 5:16:24 PM)
	 * @param number int
	 * @param customer Object
	 */
	abstract public void skipN(int number, Object customer) throws XATQueueException;

	/*------------- END ABSTRACT CLASSES ------------------------------------*/

}
