/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

/**
 * Insert the type's description here.
 * Creation date: (2002-12-22 15:59:48)
 * @author: sk
 */
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Vector;
import java.io.Serializable;

@SuppressWarnings("serial")
public class XATQueueSchema implements Serializable {
	private Vector<String> colNames = null;
	// a vector of column names, these column stores whether some data exists or
	// not rather than the actual value
	// suited for a[b], only care whether there exists b rather than b's value
	private Vector<String> existentialColNames = null;
	private Vector<String> unsatisfiedExistentialColNames = null;
	// alias of the input
	private String name;

	/**
	 * QueueMetadata constructor comment.
	 */

	public XATQueueSchema() {
		super();
		colNames = new Vector<String>();
	}

	/**
	 * QueueMetadata constructor comment.
	 */

	public XATQueueSchema(Vector<String> names) {
		super();
		colNames = names;
	}

	/**
	 * Insert the method's description here. Creation date: (2002-12-24
	 * 14:08:13)
	 * 
	 * @param name
	 *            java.lang.String
	 */
	public void addColumn(String name) {
		colNames.add(name);
	}

	/**
	 * Insert the method's description here. Creation date: (2/19/2003 9:59:01
	 * PM)
	 * 
	 * @return java.util.Vector
	 */
	public Vector<String> getExistentialColNames() {
		return existentialColNames;
	}

	/**
	 * Insert the method's description here. Creation date: (1/23/2003 10:15:01
	 * PM)
	 * 
	 * @return int
	 */
	public int getLength() {
		return colNames.size();
	}

	/**
	 * Insert the method's description here. Creation date: (1/23/2003 5:44:25
	 * PM)
	 * 
	 * @return java.lang.String
	 */
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	/**
	 * Insert the method's description here. Creation date: (2002-12-22
	 * 16:01:12)
	 * 
	 * @return int
	 * @param colName
	 *            java.lang.String
	 */
	public int getPos(String colName) throws NoSuchElementException {
		int index = colNames.indexOf(colName);
		if (index < 0)
			throw new NoSuchElementException();
		return index;
	}

	/**
	 * Insert the method's description here. Creation date: (2/19/2003 9:03:28
	 * PM)
	 * 
	 * @return java.util.Vector
	 */
	public Vector<String> getUnsatisfiedExistentialColNames() {
		if (unsatisfiedExistentialColNames == null) {
			unsatisfiedExistentialColNames = existentialColNames;
		}

		return unsatisfiedExistentialColNames;
	}

	/**
	 * whether colName is a existential column Creation date: (2/19/2003 4:08:56
	 * PM)
	 * 
	 * @return boolean
	 * @param colName
	 *            java.lang.String
	 */
	public boolean isExistentialCol(String colName) {
		if (existentialColNames != null) {
			return existentialColNames.contains(colName);
		}
		return false;
	}

	/**
	 * remove the column from unsatisfiedexistentialcolnames Creation date:
	 * (2/19/2003 9:05:51 PM)
	 * 
	 * @param colName
	 *            java.lang.String
	 */
	public void removeFrUnsatisfied(String colName) {
		unsatisfiedExistentialColNames.remove(colName);
	}

	/**
	 * Insert the method's description here. Creation date: (2/19/2003 9:59:01
	 * PM)
	 * 
	 * @param newExistentialColNames
	 *            java.util.Vector
	 */
	public void setExistentialColNames(Vector<String> newExistentialColNames) {
		existentialColNames = newExistentialColNames;
	}

	/**
	 * Insert the method's description here. Creation date: (2002-12-25
	 * 14:53:14)
	 * 
	 * @return java.lang.String
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("{");
		for (Iterator<String> it = colNames.iterator(); it.hasNext();) {
			String name = (String) it.next();
			sb.append(this.getName());
			sb.append(".");
			sb.append(name);
			sb.append(", ");
		}
		sb.append("}");
		return sb.toString();
	}

	/**
	 * @return
	 */
	public Vector<String> getColNames() {
		return colNames;
	}

	/**
	 * @param vector
	 */
	public void setColNames(Vector<String> vector) {
		colNames = vector;
	}
}
