/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;




import java.util.concurrent.BlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.*;
import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;


import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;


public class XATSimpleQueueSecondImp extends XATQueueImp  
implements  java.io.Serializable {

    /**
     * Serialization ID. This class relies on default serialization
     * even for the items array, which is default-serialized, even if
     * it is empty. Otherwise it could not be declared final, which is
     * necessary here.
     */
    private static final long serialVersionUID = -817911632652898426L;

    /** The queued items  */
    private XATTuple[] items;
    /** items index for next take, poll or remove */
    private int takeIndex;
    /** items index for next put, offer, or add. */
    private int putIndex;
    /** Number of items in the queue */
    private int count;
    /** Max capacity of a simple queue*/
    private final static int maxCapacity=16384;
    
    /** Max capacity of a simple queue*/    
    private final static int defaultCapacity=1024;

    /*
     * Concurrency control uses the classic two-condition algorithm
     * found in any textbook.
     */

    /** Main lock guarding all access */
    private final ReentrantLock lock;
    /** Condition for waiting takes */
    private final Condition notEmpty;
    /** Condition for waiting puts */
    private final Condition notFull;
    
    
    
    /* Begin CAPE Queue Implemetation*/
    /**
     * Creates an <tt>ArrayBlockingQueue</tt> with the default (fixed at 1024)
     * capacity and default access policy.
     *
     */
    public XATSimpleQueueSecondImp(){
    	this(defaultCapacity,false);
    }
    
    /**
     * Creates an <tt>ArrayBlockingQueue</tt> with the default (fixed at 1024)
     * capacity and default access policy and its ownerObject, because this is a simple Q, owner usually is null 
     * or just an object which does nothing.
     *
     */
    public XATSimpleQueueSecondImp(Object owner){
    	this(defaultCapacity,false);
    	this.queueOwners.add(owner);
    }
    
    /**
     * Creates an <tt>ArrayBlockingQueue</tt> with the default (fixed at 1024)
     * capacity and default access policy and its ownerObject, because this is a simple Q, owner usually is null 
     * or just an object which does nothing.
     *
     */
    public XATSimpleQueueSecondImp(Object owner, String queueName){
    	this(defaultCapacity,false);
    	this.queueOwners.add(owner);
    	this.queueName= queueName;
    }
    
    public void addTuple(Object newTuple)
    {
    	 final ReentrantLock lock = this.lock;
         lock.lock();
         try {
        	 ensureCapacity(count+1);
        	 insert((XATTuple)newTuple);
         }
         finally
         {
        	 lock.unlock();
         }
    }
    
    public void addTuples(List newTuples)
    {
    	if (newTuples == null) throw new NullPointerException();
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            
        	ensureCapacity(count+newTuples.size());
            
            	// ArrayList is better to use for loop
            	if (newTuples instanceof ArrayList)
            	{
            	    for(int i=0;i<newTuples.size();i++)
            	    {
            	    	insert((XATTuple) newTuples.get(i));
            	    }
            		
            	}
            	// other lists usually better with iterator 
            	else 
            	{
            		Iterator itor=newTuples.iterator();
            		while(itor.hasNext())
            		{
            			insert((XATTuple) itor.next());
            		}
            	}
                
                
            
        } finally {
            lock.unlock();
        }
	this.totalEnqueued+=newTuples.size();
    }
    /* return every elements in the queue in a linkedlist
     * 
     * */
    
    public List getContents()
    {
    	List list=new LinkedList();
    	final ReentrantLock lock = this.lock;
        lock.lock();
        try {
    	Object temp[]=this.toArray();
    	for(int i=0;i<temp.length;i++)
    	{
    		list.add(temp[i]);
    	}
        }finally
        {
        	lock.unlock();
        }
    	return list;
    	
    }
    
    /*return the current takeIndex; this is a simple queue, so customer should be null
     *@param customer Object
     */
    public int getCursor(Object customer)
    {
    	return this.takeIndex;
    }
    
    /*return the current takeIndex; this is a simple queue, so customer should be null
     * this method is only here to support the interface
     *@param customer Object
     */
    
    public int getCustCursor(Object customer)
    {
    	return this.getCursor(customer);
    }
    
    /*return the queue size; this is a simple queue, so customer should be null
     * 
     *@param customer Object
     */
    
   
   /* Returns the contents of the queue in a linkedlist;
    * Note: left over from the old queue implementation in case other components use this method
    */
   public List getTuples()
   {
	   return this.getContents();
   }
   
   /**
    * this use will increaseCursor if the increment is less then the size of maxCapacity; 
    * @param number int
    */
   public void increaseCursor(int number)
   {
	   final ReentrantLock lock = this.lock;
       lock.lock();
       try {
    	   
    	   
        	   ensureCapacity(number+count);
        	   for(int i=0;i<number;i++)
        	   {
        		   inc(takeIndex);
        	   }
           
           
          
       } finally {
           lock.unlock();
       }
   }
   
   /**
    * this use will inert the tuple in the putIndex+position.  if the putIndex+position is less then the size of the queue; 
    * For simple queue provider should be null;
    * @param position int, XATTuple tuple, Object provider 
    */
   public void insertTuple(int position, XATTuple tuple, Object provider)
   {
	   final ReentrantLock lock = this.lock;
       lock.lock();
       try {
       	ensureCapacity(items.length+position);
           for(int i=0;i<position;i++)
           {
        	   inc(putIndex);
        	   
           }
           insert(tuple);
       } finally {
           lock.unlock();
       }
	        
	    
   }
   
   public void removeAllTuples()
   {
	   this.clear();
   }
   
   /**
    * shouldn't needed
    * @param  
    */
   public void cleanTuples()
   {
	   
   }
    
   public XATTuple[] dequeue(int number, Object customer)
   {
	   XATTuple[] result=null;
	   final ReentrantLock lock = this.lock;
       lock.lock();
      
        try {
	   if(number>count) number=count;
	   result= (XATTuple[]) new Object[number];
	   
	  
      
	        final XATTuple[] items = this.items;
	       for(int i=0;i<number;i++)
	       {
		       result[i]=items[takeIndex];
		       items[takeIndex]=null;
		       takeIndex=inc(takeIndex);
		       --count;
	       }
       }
       finally{
	       lock.unlock();
       }

	   this.totalDequeued+=number;
	   return result;
   }

   public XATTuple dequeue(Object customer)throws XATQueueException
   {
	   XATTuple t=null;
	    final ReentrantLock lock = this.lock;
       lock.lock();
       try
       {
	       if (count==0) throw new XATQueueException("Not enough tuples to dequeue");
        t=extract();
 
       }
       finally{
       lock.unlock();
       }
       return t;
   }

   public List dequeueN(int number, Object customer)
   {
   
	   XATTuple [] temp=this.dequeue(number, customer);
	   List result=new ArrayList(number);
	   for(int i=0;i<number;i++)
	   {
		   result.add(temp[i]);
	   }
	   return result;
   }

   public void enqueue(XATTuple[] newTuples, Object provider)
   {
	   List list=new ArrayList();
	   for(int i=0;i<newTuples.length;i++)
	   {
		   list.add(newTuples[i]);
	   }
	   addTuples(list);
   	
   }

   public void enqueue(List newTuples, Object provider)
   {
   addTuples(newTuples);
   }

   public void enqueue(XATTuple newTuple, Object provider)
   {
   offer(newTuple);
   this.totalEnqueued++;
   }

   public LightWeightQueryObject getLightWeighQueryObject()
   {
   return new LightWeightQueryObject("XATAnotherSimpleQueue", this.getClass().getName(),true);
   }

   public int getQueueSize()
   {
    return size();
   }

   public int getQueueSize(Object customer)
   {
   return size();
   }

   public XATTuple[] peek(int number, Object customer)
   {
	   final ReentrantLock lock = this.lock;
        lock.lock();
	XATTuple []result=null;
	int cursor=takeIndex;
        try {
            if (count == 0) return null;
	    else if(number>count) 
	    {
		    number=count;
		    result=(XATTuple[]) new Object[number];
		    final XATTuple[] items = this.items;
		    
		    for(int i=0;i<number;i++)
		    {
		    result[i]= items[takeIndex];
		    takeIndex=inc(takeIndex);
		    }
		   
		    takeIndex=cursor;
	    }
        } finally {
            lock.unlock();
        }

	return result;
   
   }

   public XATTuple peekNthTuple(int number, Object customer)
   {
	   final ReentrantLock lock = this.lock;
        lock.lock();
	XATTuple result=null;
	int  cursor=takeIndex;
        try {
            if (count == 0) return null;
	    else if(number>count) 
	    {
		    number=count;
		    
		    final XATTuple[] items = this.items;
		    
		    for(int i=0;i<number;i++)
		    {
		    takeIndex=inc(takeIndex);
		    }
		    result=extract();
		   
		    takeIndex=cursor;
	    }
        } finally {
            lock.unlock();
        }

	return result;
   
   }

   public XATTuple peek(Object customer) throws XATQueueException 
   {
	   XATTuple t=null;
   final ReentrantLock lock = this.lock;
        lock.lock();
        try {
        	if(count==0) {
            	throw new XATQueueException("Not enough tuples to peek");
            }
            else t= items[takeIndex];
        } finally {
            lock.unlock();
        }
        return t;
       
   }

     public List peekN(int number, Object customer)
     {
     List result=new ArrayList();
     XATTuple [] temp=peek(number, customer);
     for(int i=0;i<temp.length;i++)
     {
	     result.add(temp[i]);
     }
     return result;

     }
     
     @Override
      public XATTuple peekLast ()throws XATQueueException
   {
    	  XATTuple t=null;
   final ReentrantLock lock = this.lock;
        lock.lock();
        try {
        	if(count==0) {
            	throw new XATQueueException("Not enough tuples to peek");
            }
            else t= items[dec(putIndex)];
        } finally {
            lock.unlock();
        }
        return t;
   }

   public void setTuples(List newTuples)
   {
   clear();

   addTuples(newTuples);

   }

   public String printContent(Object customer)
   {
   StringBuffer rs=new StringBuffer();
   rs.append(this.size());
   rs.append("tuples: \n");
   XATTuple[] result=this.peek(this.size(), customer);
   for(int i=0;i<result.length;i++){
   rs.append(result[i].toString()+"\n");
  
   }
   return rs.toString();
   }

   public void skip(Object customer) throws XATQueueException
   {
    final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            if (count == 0)  throw new XATQueueException("Not enough tuples to skip!") ;
	    extract();
	    
        } finally {
            lock.unlock();
        }
   }

   public void skipN(int number, Object customer) throws XATQueueException
   {
    final ReentrantLock lock = this.lock;
        lock.lock();
        try {
		  if(number>count) throw new XATQueueException("Not enough tuples to skip!");
	          final XATTuple[] items = this.items;
	       for(int i=0;i<number;i++)
	       {
		       items[takeIndex]=null;
		       takeIndex=inc(takeIndex);
		       --count;
	       }
	
	}
	finally
	{
	lock.unlock();
	}

   }
 
    /*End CAPE Queue Implemetation*
     * The rests are Sun's codes
     */

   /**
     * Circularly decrement i.
     */
    final int dec(int i) {
        return (--i == -1)? items.length : i;
    }
   //other Sun's code
   /**
    * Increases the capacity of this <tt>SimpleQueue</tt> instance, if
    * necessary, to ensure that it can hold at least the number of elements
    * specified by the minimum capacity argument.
    * Call only when holding lock.
    * @param   minCapacity   the desired minimum capacity, must less than maxCapacity
    */
   private void ensureCapacity(int minCapacity) {
	assert minCapacity<maxCapacity;   
	int oldCapacity = items.length;
	if (minCapacity > oldCapacity) {
	    Object oldData[] = items;
	    int newCapacity = (oldCapacity * 3)/2 + 1;
   	    if (newCapacity < minCapacity)
		newCapacity = minCapacity;
           // minCapacity is usually close to size, so this is a win:
   	 
           items = Arrays.copyOf(items, newCapacity);       
	
     
     }
   }
   
   //End of other sun's code
    
    // Internal helper methods

    /**
     * Circularly increment i.
     */
    final int inc(int i) {
        return (++i == items.length)? 0 : i;
    }

    /**
     * Inserts element at current put position, advances, and signals.
     * Call only when holding lock.
     */
    private void insert(XATTuple x) {
        items[putIndex] = x;
        putIndex = inc(putIndex);
        ++count;
        notEmpty.signal();
    }

    /**
     * Extracts element at current take position, advances, and signals.
     * Call only when holding lock.
     */
    private XATTuple extract() {
        final XATTuple[] items = this.items;
        XATTuple x = items[takeIndex];
        items[takeIndex] = null;
        takeIndex = inc(takeIndex);
        --count;
        notFull.signal();
        return x;
    }

    /**
     * Utility for remove and iterator.remove: Delete item at position i.
     * Call only when holding lock.
     */
    void removeAt(int i) {
        final XATTuple[] items = this.items;
        // if removing front item, just advance
        if (i == takeIndex) {
            items[takeIndex] = null;
            takeIndex = inc(takeIndex);
        } else {
            // slide over all others up through putIndex.
            for (;;) {
                int nexti = inc(i);
                if (nexti != putIndex) {
                    items[i] = items[nexti];
                    i = nexti;
                } else {
                    items[i] = null;
                    putIndex = i;
                    break;
                }
            }
        }
        --count;
        notFull.signal();
    }

  
    
    
    /**
     * Creates an <tt>ArrayBlockingQueue</tt> with the given (fixed)
     * capacity and default access policy.
     *
     * @param capacity the capacity of this queue
     * @throws IllegalArgumentException if <tt>capacity</tt> is less than 1
     */
    public XATSimpleQueueSecondImp(int capacity) {
        this(capacity, false);
    }
    
    

    /**
     * Creates an <tt>ArrayBlockingQueue</tt> with the given (fixed)
     * capacity and the specified access policy.
     *
     * @param capacity the capacity of this queue
     * @param fair if <tt>true</tt> then queue accesses for threads blocked
     *        on insertion or removal, are processed in FIFO order;
     *        if <tt>false</tt> the access order is unspecified.
     * @throws IllegalArgumentException if <tt>capacity</tt> is less than 1
     */
    public XATSimpleQueueSecondImp(int capacity, boolean fair) {
        if (capacity <= 0)
            throw new IllegalArgumentException();
        this.items = (XATTuple[]) new Object[capacity];
        lock = new ReentrantLock(fair);
        notEmpty = lock.newCondition();
        notFull =  lock.newCondition();
    }

    /**
     * Creates an <tt>ArrayBlockingQueue</tt> with the given (fixed)
     * capacity, the specified access policy and initially containing the
     * elements of the given collection,
     * added in traversal order of the collection's iterator.
     *
     * @param capacity the capacity of this queue
     * @param fair if <tt>true</tt> then queue accesses for threads blocked
     *        on insertion or removal, are processed in FIFO order;
     *        if <tt>false</tt> the access order is unspecified.
     * @param c the collection of elements to initially contain
     * @throws IllegalArgumentException if <tt>capacity</tt> is less than
     *         <tt>c.size()</tt>, or less than 1.
     * @throws NullPointerException if the specified collection or any
     *         of its elements are null
     */
    public XATSimpleQueueSecondImp(int capacity, boolean fair,
                              Collection<? extends XATTuple> c) {
        this(capacity, fair);
        if (capacity < c.size())
            throw new IllegalArgumentException();

        for (Iterator<? extends XATTuple> it = c.iterator(); it.hasNext();)
            add(it.next());
    }

    /**
     * Inserts the specified element at the tail of this queue if it is
     * possible to do so immediately without exceeding the queue's capacity,
     * returning <tt>true</tt> upon success and throwing an
     * <tt>IllegalStateException</tt> if this queue is full.
     *
     * @param e the element to add
     * @return <tt>true</tt> (as specified by {@link Collection#add})
     * @throws IllegalStateException if this queue is full
     * @throws NullPointerException if the specified element is null
     */
    public boolean add(XATTuple e) {
    	   
    	        if (offer(e))
    	            return true;
    	        else
    	            throw new IllegalStateException("Queue full");
    	    
    }

    /**
     * Inserts the specified element at the tail of this queue if it is
     * possible to do so immediately without exceeding the queue's capacity,
     * returning <tt>true</tt> upon success and <tt>false</tt> if this queue
     * is full.  This method is generally preferable to method {@link #add},
     * which can fail to insert an element only by throwing an exception.
     *
     * @throws NullPointerException if the specified element is null
     */
    public boolean offer(XATTuple e) {
        if (e == null) throw new NullPointerException();
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
        	ensureCapacity(items.length+1);
            if (count == items.length)
                return false;
            else {
                insert(e);
                return true;
            }
        } finally {
            lock.unlock();
        }
    }

    /**
     * Inserts the specified element at the tail of this queue, waiting
     * for space to become available if the queue is full.
     *
     * @throws InterruptedException {@inheritDoc}
     * @throws NullPointerException {@inheritDoc}
     */
    public void put(XATTuple e) throws InterruptedException {
        if (e == null) throw new NullPointerException();
        final XATTuple[] items = this.items;
        final ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        try {
            try {
                while (count == items.length)
                    notFull.await();
            } catch (InterruptedException ie) {
                notFull.signal(); // propagate to non-interrupted thread
                throw ie;
            }
            insert(e);
        } finally {
            lock.unlock();
        }
    }

    /**
     * Inserts the specified element at the tail of this queue, waiting
     * up to the specified wait time for space to become available if
     * the queue is full.
     *
     * @throws InterruptedException {@inheritDoc}
     * @throws NullPointerException {@inheritDoc}
     */
    public boolean offer(XATTuple e, long timeout, TimeUnit unit)
        throws InterruptedException {

        if (e == null) throw new NullPointerException();
	long nanos = unit.toNanos(timeout);
        final ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        try {
            for (;;) {
                if (count != items.length) {
                    insert(e);
                    return true;
                }
                if (nanos <= 0)
                    return false;
                try {
                    nanos = notFull.awaitNanos(nanos);
                } catch (InterruptedException ie) {
                    notFull.signal(); // propagate to non-interrupted thread
                    throw ie;
                }
            }
        } finally {
            lock.unlock();
        }
    }

    public XATTuple poll() {
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            if (count == 0)
                return null;
            XATTuple x = extract();
            return x;
        } finally {
            lock.unlock();
        }
    }

    public XATTuple take() throws InterruptedException {
        final ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        try {
            try {
                while (count == 0)
                    notEmpty.await();
            } catch (InterruptedException ie) {
                notEmpty.signal(); // propagate to non-interrupted thread
                throw ie;
            }
            XATTuple x = extract();
            return x;
        } finally {
            lock.unlock();
        }
    }

    public XATTuple poll(long timeout, TimeUnit unit) throws InterruptedException {
	long nanos = unit.toNanos(timeout);
        final ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        try {
            for (;;) {
                if (count != 0) {
                    XATTuple x = extract();
                    return x;
                }
                if (nanos <= 0)
                    return null;
                try {
                    nanos = notEmpty.awaitNanos(nanos);
                } catch (InterruptedException ie) {
                    notEmpty.signal(); // propagate to non-interrupted thread
                    throw ie;
                }

            }
        } finally {
            lock.unlock();
        }
    }

    public XATTuple peek() {
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return (count == 0) ? null : items[takeIndex];
        } finally {
            lock.unlock();
        }
    }

    // this doc comment is overridden to remove the reference to collections
    // greater in size than Integer.MAX_VALUE
    /**
     * Returns the number of elements in this queue.
     *
     * @return the number of elements in this queue
     */
    public int size() {
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return count;
        } finally {
            lock.unlock();
        }
    }

    // this doc comment is a modified copy of the inherited doc comment,
    // without the reference to unlimited queues.
    /**
     * Returns the number of additional elements that this queue can ideally
     * (in the absence of memory or resource constraints) accept without
     * blocking. This is always equal to the initial capacity of this queue
     * less the current <tt>size</tt> of this queue.
     *
     * <p>Note that you <em>cannot</em> always tell if an attempt to insert
     * an element will succeed by inspecting <tt>remainingCapacity</tt>
     * because it may be the case that another thread is about to
     * insert or remove an element.
     */
    public int remainingCapacity() {
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return items.length - count;
        } finally {
            lock.unlock();
        }
    }

    /**
     * Removes a single instance of the specified element from this queue,
     * if it is present.  More formally, removes an element <tt>e</tt> such
     * that <tt>o.equals(e)</tt>, if this queue contains one or more such
     * elements.
     * Returns <tt>true</tt> if this queue contained the specified element
     * (or equivalently, if this queue changed as a result of the call).
     *
     * @param o element to be removed from this queue, if present
     * @return <tt>true</tt> if this queue changed as a result of the call
     */
    public boolean remove(Object o) {
        if (o == null) return false;
        final XATTuple[] items = this.items;
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            int i = takeIndex;
            int k = 0;
            for (;;) {
                if (k++ >= count)
                    return false;
                if (o.equals(items[i])) {
                    removeAt(i);
                    return true;
                }
                i = inc(i);
            }

        } finally {
            lock.unlock();
        }
    }

    /**
     * Returns <tt>true</tt> if this queue contains the specified element.
     * More formally, returns <tt>true</tt> if and only if this queue contains
     * at least one element <tt>e</tt> such that <tt>o.equals(e)</tt>.
     *
     * @param o object to be checked for containment in this queue
     * @return <tt>true</tt> if this queue contains the specified element
     */
    public boolean contains(Object o) {
        if (o == null) return false;
        final XATTuple[] items = this.items;
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            int i = takeIndex;
            int k = 0;
            while (k++ < count) {
                if (o.equals(items[i]))
                    return true;
                i = inc(i);
            }
            return false;
        } finally {
            lock.unlock();
        }
    }

    /**
     * Returns an array containing all of the elements in this queue, in
     * proper sequence.
     *
     * <p>The returned array will be "safe" in that no references to it are
     * maintained by this queue.  (In other words, this method must allocate
     * a new array).  The caller is thus free to modify the returned array.
     *
     * <p>This method acts as bridge between array-based and collection-based
     * APIs.
     *
     * @return an array containing all of the elements in this queue
     */
    public Object[] toArray() {
        final XATTuple[] items = this.items;
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] a = new Object[count];
            int k = 0;
            int i = takeIndex;
            while (k < count) {
                a[k++] = items[i];
                i = inc(i);
            }
            return a;
        } finally {
            lock.unlock();
        }
    }

    /**
     * Returns an array containing all of the elements in this queue, in
     * proper sequence; the runtime type of the returned array is that of
     * the specified array.  If the queue fits in the specified array, it
     * is returned therein.  Otherwise, a new array is allocated with the
     * runtime type of the specified array and the size of this queue.
     *
     * <p>If this queue fits in the specified array with room to spare
     * (i.e., the array has more elements than this queue), the element in
     * the array immediately following the end of the queue is set to
     * <tt>null</tt>.
     *
     * <p>Like the {@link #toArray()} method, this method acts as bridge between
     * array-based and collection-based APIs.  Further, this method allows
     * precise control over the runtime type of the output array, and may,
     * under certain circumstances, be used to save allocation costs.
     *
     * <p>Suppose <tt>x</tt> is a queue known to contain only strings.
     * The following code can be used to dump the queue into a newly
     * allocated array of <tt>String</tt>:
     *
     * <pre>
     *     String[] y = x.toArray(new String[0]);</pre>
     *
     * Note that <tt>toArray(new Object[0])</tt> is identical in function to
     * <tt>toArray()</tt>.
     *
     * @param a the array into which the elements of the queue are to
     *          be stored, if it is big enough; otherwise, a new array of the
     *          same runtime type is allocated for this purpose
     * @return an array containing all of the elements in this queue
     * @throws ArrayStoreException if the runtime type of the specified array
     *         is not a supertype of the runtime type of every element in
     *         this queue
     * @throws NullPointerException if the specified array is null
     */
    public <T> T[] toArray(T[] a) {
        final XATTuple[] items = this.items;
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            if (a.length < count)
                a = (T[])java.lang.reflect.Array.newInstance(
                    a.getClass().getComponentType(),
                    count
                    );

            int k = 0;
            int i = takeIndex;
            while (k < count) {
                a[k++] = (T)items[i];
                i = inc(i);
            }
            if (a.length > count)
                a[count] = null;
            return a;
        } finally {
            lock.unlock();
        }
    }

    public String toString() {
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return super.toString();
        } finally {
            lock.unlock();
        }
    }

    /**
     * Atomically removes all of the elements from this queue.
     * The queue will be empty after this call returns.
     */
    public void clear() {
        final XATTuple[] items = this.items;
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            int i = takeIndex;
            int k = count;
            while (k-- > 0) {
                items[i] = null;
                i = inc(i);
            }
            count = 0;
            putIndex = 0;
            takeIndex = 0;
            notFull.signalAll();
        } finally {
            lock.unlock();
        }
    }

    /**
     * @throws UnsupportedOperationException {@inheritDoc}
     * @throws ClassCastException            {@inheritDoc}
     * @throws NullPointerException          {@inheritDoc}
     * @throws IllegalArgumentException      {@inheritDoc}
     */
    public int drainTo(Collection<? super XATTuple> c) {
        if (c == null)
            throw new NullPointerException();
        if (c == this)
            throw new IllegalArgumentException();
        final XATTuple[] items = this.items;
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            int i = takeIndex;
            int n = 0;
            int max = count;
            while (n < max) {
                c.add(items[i]);
                items[i] = null;
                i = inc(i);
                ++n;
            }
            if (n > 0) {
                count = 0;
                putIndex = 0;
                takeIndex = 0;
                notFull.signalAll();
            }
            return n;
        } finally {
            lock.unlock();
        }
    }

    /**
     * @throws UnsupportedOperationException {@inheritDoc}
     * @throws ClassCastException            {@inheritDoc}
     * @throws NullPointerException          {@inheritDoc}
     * @throws IllegalArgumentException      {@inheritDoc}
     */
    public int drainTo(Collection<? super XATTuple> c, int maxElements) {
        if (c == null)
            throw new NullPointerException();
        if (c == this)
            throw new IllegalArgumentException();
        if (maxElements <= 0)
            return 0;
        final XATTuple[] items = this.items;
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            int i = takeIndex;
            int n = 0;
            int sz = count;
            int max = (maxElements < count)? maxElements : count;
            while (n < max) {
                c.add(items[i]);
                items[i] = null;
                i = inc(i);
                ++n;
            }
            if (n > 0) {
                count -= n;
                takeIndex = i;
                notFull.signalAll();
            }
            return n;
        } finally {
            lock.unlock();
        }
    }


    /**
     * Returns an iterator over the elements in this queue in proper sequence.
     * The returned <tt>Iterator</tt> is a "weakly consistent" iterator that
     * will never throw {@link ConcurrentModificationException},
     * and guarantees to traverse elements as they existed upon
     * construction of the iterator, and may (but is not guaranteed to)
     * reflect any modifications subsequent to construction.
     *
     * @return an iterator over the elements in this queue in proper sequence
     */
    public Iterator<XATTuple> iterator() {
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return new Itr();
        } finally {
            lock.unlock();
        }
    }

    /**
     * Iterator for ArrayBlockingQueue
     */
    private class Itr implements Iterator<XATTuple> {
        /**
         * Index of element to be returned by next,
         * or a negative number if no such.
         */
        private int nextIndex;

        /**
         * nextItem holds on to item fields because once we claim
         * that an element exists in hasNext(), we must return it in
         * the following next() call even if it was in the process of
         * being removed when hasNext() was called.
         */
        private XATTuple nextItem;

        /**
         * Index of element returned by most recent call to next.
         * Reset to -1 if this element is deleted by a call to remove.
         */
        private int lastRet;

        Itr() {
            lastRet = -1;
            if (count == 0)
                nextIndex = -1;
            else {
                nextIndex = takeIndex;
                nextItem = items[takeIndex];
            }
        }

        public boolean hasNext() {
            /*
             * No sync. We can return true by mistake here
             * only if this iterator passed across threads,
             * which we don't support anyway.
             */
            return nextIndex >= 0;
        }

        /**
         * Checks whether nextIndex is valid; if so setting nextItem.
         * Stops iterator when either hits putIndex or sees null item.
         */
        private void checkNext() {
            if (nextIndex == putIndex) {
                nextIndex = -1;
                nextItem = null;
            } else {
                nextItem = items[nextIndex];
                if (nextItem == null)
                    nextIndex = -1;
            }
        }

        public XATTuple next() {
            final ReentrantLock lock = XATSimpleQueueSecondImp.this.lock;
            lock.lock();
            try {
                if (nextIndex < 0)
                    throw new NoSuchElementException();
                lastRet = nextIndex;
                XATTuple x = nextItem;
                nextIndex = inc(nextIndex);
                checkNext();
                return x;
            } finally {
                lock.unlock();
            }
        }

        public void remove() {
            final ReentrantLock lock = XATSimpleQueueSecondImp.this.lock;
            lock.lock();
            try {
                int i = lastRet;
                if (i == -1)
                    throw new IllegalStateException();
                lastRet = -1;

                int ti = takeIndex;
                removeAt(i);
                // back up cursor (reset to front if was first element)
                nextIndex = (i == ti) ? takeIndex : i;
                checkNext();
            } finally {
                lock.unlock();
            }
        }
    }

	@Override
	public void enqueue(XATTupleNegativeImp newTuple, Object provider) {
		// TODO Auto-generated method stub
		
	}

	

	@Override
	public void enqueue(XATTupleNegativeImp[] newTuples, Object provider) {
		// TODO Auto-generated method stub
		
	}

	

	@Override
	public LightWeightQueryObject getLightWeightQueryObject() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public String getQueueLayout() {
		return "|T->" + this.getQueueSize() + "|";
	}

	

	
	

	
	
}



