/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

/**
 * XATTuple is the basic data unit of a XATQueue.
 *
 * Each tuple contains an order-set of cells,
 * each cell is an XATDataValue.
 * The order is maintained by the queue that contains the tuple.
 * 
 * Creation date: (4/26/2003 5:51:41 PM)
 * @author: Jinhui Jian
 */
import java.util.List;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
 
public interface XATTuple {
/**
 * Add one cell.
 *
 * Creation date: (4/29/2003 2:30:40 AM)
 * @param cell edu.wpi.cs.dsrg.xmldb.raindrop.common.XATData
 */
void addCell(XATDataValue cell);
void combineTimeStamps(XATTuple tuple1, XATTuple tuple2);
/**
 * Returns all cells in the tuple by an array.
 * 
 * Creation date: (4/26/2003 6:22:34 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue[]
 */
XATDataValue[] getArrayData();
/**
 * Returns the cell in the given position.
 * 
 * Creation date: (4/26/2003 6:22:00 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue
 * @param pos int
 */
XATDataValue getCell(int pos);
/**
 * Returns the posTH cell in the tuple.
 * 
 * Creation date: (4/26/2003 6:26:37 PM)
 * @return int
 */
int getLength();
/**
 * Returns true if the tuple is empty
 * Creation date: (5/16/2003 11:42:27 PM)
 * @return boolean
 */
boolean isEmpty();
/**
 * Depricated method. Do not use.
 * @return
 */
boolean isNew();
/**
 * Depricated method. Do not use.
 * 
 * @param isNew
 */
void setToNew(boolean isNew);
/**
 * Sets a cell at the given position.
 * The tuple must be large enough to include the given position,
 *
 * Creation date: (5/17/2003 2:00:13 PM)
 * @param pos int
 * @param data edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue
 */
void setCell(int pos, XATDataValue data);
/**
 * Set all cells with the given XATDataValue array.
 * 
 * Creation date: (4/26/2003 6:24:27 PM)
 * @param datas edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[]
 */
void setValues(XATDataValue[] datas);

void setUniTimeStamp(long unitime);
/**
 * Set all cells with the given XATDataValue array.
 * 
 * Creation date: (4/26/2003 6:24:27 PM)
 * @param datas edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[]
 */
void setValues(List datas);

/* The methods below this are for the storage manager's purposes only */
/**
 * Gets the max time stamp.
 * 
 * Creation Date: (Jun 7, 2004 1:33:13 PM)
 * @author Natasha Bogdanova
 * @return long
 */
long[] getTimeStamp();
long getNthTimeStamp(int n);
long getMaxTimeStamp();
long getMinTimeStamp();
void setNthTimeStamp(long time, int n);
void setMaxTimeStamp(long max);
void setMinTimeStamp(long min);

/**
 * Sets the punctuation ID.
 * 
 * Creation Date: (Jun 7, 2004 1:35:16 PM)
 * @author Natasha Bogdanova
 * @param pid long
 */
//void setPunctID(long pid);
/**
 * Sets the time stamp.
 * 
 * Creation Date: (Jun 7, 2004 1:38:42 PM)
 * @author Natasha Bogdanova
 * @param ts long
 */
void setTimeStamp(long[] ts);
/**
 * Returns the index of this tuple. 
 * Note: the tuple will have a valid index value only after
 * it has retreived from the disk.
 * 
 * Creation Date: (Jun 7, 2004 1:39:53 PM)
 * @author Natasha Bogdanova
 * @return int
 */
int getIndex();
/**
 * Sets the index of this tuple. Used by the storage manager
 * when it retreives tuples from the disk.
 * 
 * Creation Date: (Jun 7, 2004 1:41:06 PM)
 * @author Natasha Bogdanova
 * @param index int
 */

void setIndex(int index);

/**
 * The following two functions are basically for partitioned stream operators. each tuple contains
 * partition id information.
 * @return
 */
int getPartitionID();
void setPartitionID(int id);
boolean equals(XATTuple tp);
boolean equalKeys(Object[] inputKeys, int[] keyPos);
boolean equalKeys(XATDataValue[] punctContent, int[] punctType, int[] keysPos);
int getObjTypeID();
boolean isPunct();
boolean isSecPunct();
boolean isRoutingToken();
boolean isStqmPunct();
boolean isStreamingTagPunct();
public boolean isSingletonTuple();

/**
 * The following two functions are basically for pipelined stream operators. each tuple contains
 * stream id information, and boolean probing or not.
 * Added by Song Wang
 * @return
 */
void setTimeEnter(long l);
long getTimeEnter();
int getStreamID();
void setStreamID(int id);
boolean isProbe();
void setProbe(boolean probe);
public int getStateToJoin();
public void setStateToJoin(int i);
public boolean getEnding();
public void setEnding(boolean b);
public int getOpID();
public void setOpID(int i);
public boolean isDummy();
public void setDummy(boolean b);
public boolean isRealPurge();
public void setRealPurge(boolean b);
}
