/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

/**
 * Implements XATTuple.
 * 
 * Creation date: (5/16/2003 11:51:55 PM)
 * @author: JinHui Jian
 */
import java.util.List;
import java.io.Serializable;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
//import edu.wpi.cs.dsrg.xmldb.xat.raindrop.utils.DeepCopyMaker;

public class XATTupleImp implements XATTuple, Serializable, Cloneable {
	//dummy double array for space feeding
	//double[] dummyData = new double[]{0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
	//		0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,};
	
	/* It is solely for performance purpose that we use the oneCell/listCell
	   design. The motivation is that a lot of tuples have only one cell,
	   using a java collection to encapsulate them is expensive. In this case,
	   we use the oneCell to store the only cell. If there are more than one cell,
	   we set oneCell to null and use listCell.
	*/
	protected XATDataValue oneCell = null;
	protected XATDataValue[] arrayCell = null;
	protected long[] timeStamp = null;
	protected long maxTimeStamp;
	protected long minTimeStamp;
	
	//ID for XATTuple, for differentiating from XATPunct
	protected int objTypeID  = 0;
		
	//protected boolean ifLastTuple = false;
	protected boolean isNewTuple = false;
	protected boolean isSingleton = false;
	
	/* The following attributes are for the storage manager.*/
	
	/**Stores the index of the tuple for the purpose of the storage manager*/
	private int index  = -1;
	
	/* End storage manager attributes. */
	
	/* each XATTuple has partition id information, for partitioned operator only, add by Bin Oct. 2004 */
	private int partitionID = -1;
	/**
	 * XATTupleImp2 constructor comment.
	 */
	public XATTupleImp() {
		super();
	}
	/**
	 * XATTupleImp2 constructor comment.
	 */
	public XATTupleImp(XATDataValue[] array) {
		super();
		if (array == null || array.length == 0)
			return;
		if (array.length == 1)
			this.oneCell = array[0];
		else
			this.arrayCell = array;
	}
	/**
	 * XATTupleImp2 constructor comment.
	 */
	public XATTupleImp(int length) {
		super();
		//construct an array only if length > 1
		if (length > 1)
			this.arrayCell = new XATDataValue[length];
		else if (length == 1)
			this.oneCell = new XATDataValueImp();
	}
	/**
	 * XATTupleImp2 constructor comment.
	 */
	public XATTupleImp(List list) {
		super();

		if (list == null || list.isEmpty())
			return;

		if (list.size() == 1) {
			this.oneCell = (XATDataValue) list.get(0);
		} else {
			this.arrayCell = (XATDataValue[]) list.toArray(new XATDataValue[] {
			});
		}
	}
	/**
	 * Add one cell.
	 *
	 * Creation date: (5/16/2003 11:51:55 PM)
	 * @param cell edu.wpi.cs.dsrg.xmldb.raindrop.common.XATData
	 */
	public void addCell(XATDataValue cell) {

		if (this.oneCell == null) {
			//this is the first cell
			if (this.arrayCell == null) {
				this.oneCell = cell;
			}
			//otherwise add to listCell 
			else {
				XATDataValue[] temp =
					new XATDataValue[this.arrayCell.length + 1];
				for (int i = 0; i < this.arrayCell.length; i++)
					temp[i] = this.arrayCell[i];
				temp[this.arrayCell.length] = cell;
				this.arrayCell = temp;
			}
		}
		//otherwise it must be the second cell
		else {
			XATDataValue[] temp = new XATDataValue[2];
			temp[0] = this.oneCell;
			temp[1] = cell;

			this.arrayCell = temp;
			this.oneCell = null;
		}
	}
	
	public void combineTimeStamps(XATTuple tuple1, XATTuple tuple2) {
		long[] time1 = tuple1.getTimeStamp();
		long[] time2 = tuple2.getTimeStamp();

		this.timeStamp = new long[time1.length + time2.length];
		int i, j;
		for(i=0; i<time1.length; i++) {
			timeStamp[i] = time1[i];
		}
		for(j=0; j<time2.length; j++) {
			timeStamp[i+j] = time2[j];
		}
		long m1 = tuple1.getMinTimeStamp();
		long m2 = tuple2.getMinTimeStamp();
		this.minTimeStamp = (m1<m2)?m1:m2;
		m1 = tuple1.getMaxTimeStamp();
		m2 = tuple2.getMaxTimeStamp();
		this.maxTimeStamp = (m1>m2)?m1:m2;

		m1 = tuple1.getTimeEnter();
		m2 = tuple2.getTimeEnter();
		this.timeEnter = (m1>m2)?m1:m2;
		
		if(tuple1.isNew() && tuple2.isNew()) {
			this.isNewTuple = true;
		} else {
			this.isNewTuple = false;
		}
	}
	/**
	 * fill the tuple content into the holder from the first cell
	 * Creation date: (5/14/2003 5:59:49 PM)
	 * @param holder edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue[]
	 */
	public void fillDataInListTo(XATDataValue[] holder) {
		int size = getLength();

		for (int i = 0; i < size; i++) {
			holder[i] = this.getCell(i);
		}

	}
	/**
	 * given a holder, fill it from position begin 
	 * the filling content is part of the "data field" specified by list
	 * Creation date: (2/9/2003 11:47:11 AM)
	 * @param list java.util.Vector
	 */
	public void fillDataInListTo(
		XATDataValue[] holder,
		int[] list,
		int begin) {

		for (int i = 0; i < list.length; i++) {
			holder[begin + i] = (XATDataValue) this.getCell(list[i]);
		}

	}
	/**
	 * Returns all cells in the tuple by an array.
	 * 
	 * Creation date: (5/16/2003 11:51:55 PM)
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue[]
	 */
	public XATDataValue[] getArrayData() {
		//either have no cell or more than one cell
		if (this.oneCell == null) {
			//no cell
			if (this.arrayCell == null) {
				return new XATDataValue[] {
				};
			}
			//more than one cell
			else {
				return this.arrayCell;
			}
		}
		//otherwise there must be only one cell
		else {
			return new XATDataValue[] { this.oneCell };
		}
	}
	/**
	 * Returns the cell in the given position.
	 * 
	 * Creation date: (5/16/2003 11:51:55 PM)
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue
	 * @param pos int
	 */
	public XATDataValue getCell(int pos) {
		//either no cell or more than one cell
		if (this.oneCell == null) {
			//no cell
			if (this.arrayCell == null) {
				throw new RuntimeException("The cell in the given position does not exist!");
			}
			//more than one  cell
			else {
				return this.arrayCell[pos];
			}
		}
		//must be one cell
		else {
			if (pos != 0) {
				throw new RuntimeException("The cell in the given position does not exist!");
			} else {
				return this.oneCell;
			}
		}

	}
	/**
	 * Returns the nubmer of cells in the tuple.
	 * 
	 * Creation date: (5/16/2003 11:51:55 PM)
	 * @return int
	 */
	public int getLength() {
		//either no cell or more than one cell
		if (this.oneCell == null) {
			//no cell
			if (this.arrayCell == null) {
				return 0;
			}
			//more than one  cell
			else {
				return this.arrayCell.length;
			}
		}
		//must be one cell
		else {
			return 1;
		}

	}
	/**
	 * Insert the method's description here.
	 * Creation date: (5/16/2003 11:51:55 PM)
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp
	 */
//	public XATTimeStamp getTimeStamp() {
//		return this.timeStamp;
//	}
	/**
	 * Insert the method's description here.
	 * Creation date: (5/16/2003 11:51:55 PM)
	 * @return boolean
	 */
	public boolean isEmpty() {
		return this.oneCell == null && this.arrayCell == null;
	}

	public boolean isNew() {
		return this.isNewTuple;
	}

	public void setToNew(boolean isNew) {
		this.isNewTuple = isNew;
	}

	/**
	 * Sets a cell at the given position.
	 * The tuple must be large enough to include the given position,
	 *
	 * Creation date: (5/17/2003 1:55:57 PM)
	 * @param pos int
	 * @param data edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue
	 */
	public void setCell(int pos, XATDataValue data) {
		if (this.oneCell != null) {
			if (pos != 0)
				throw new RuntimeException("must initialize a tuple to certain size first!");
			this.oneCell = data;
		} else {
			this.arrayCell[pos] = data;
		}
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (5/16/2003 11:51:55 PM)
	 * @param ts edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp
	 */
//	public void setTimeStamp(XATTimeStamp ts) {
//		this.timeStamp = ts;
//	}
	/**
	 * Set all cells with the given XATDataValue array.
	 * 
	 * Creation date: (5/16/2003 11:51:55 PM)
	 * @param datas edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[]
	 */
	public void setValues(XATDataValue[] datas) {
		if (datas == null || datas.length == 0) {
			this.oneCell = null;
			this.arrayCell = null;
		} else {
			if (datas.length == 1) {
				this.oneCell = datas[0];
				this.arrayCell = null;
			} else {
				this.oneCell = null;
				this.arrayCell = datas;
			}
		}

	}
	
	public void setUniTimeStamp(long unitime) {
		if(oneCell != null) {
			timeStamp = new long[1];
			timeStamp[0] = unitime;
		}else if (arrayCell != null){
			int size = this.arrayCell.length;
			this.timeStamp = new long[size];
			for(int i=0; i<size; i++) {
				timeStamp[i] = unitime;
			}
		}
		this.minTimeStamp = unitime;
		this.maxTimeStamp = unitime;
		this.isSingleton = true;
	}
	
	
	/**
	 * Set all cells with the given XATDataValue array.
	 * 
	 * Creation date: (5/16/2003 11:51:55 PM)
	 * @param datas edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[]
	 */
	public void setValues(List list) {
		if (list == null || list.isEmpty()) {
			this.oneCell = null;
			this.arrayCell = null;
			return;
		}

		if (list.size() == 1) {
			this.oneCell = (XATDataValue) list.get(0);
			this.arrayCell = null;
		} else {
			this.oneCell = null;
			this.arrayCell = (XATDataValue[]) list.toArray(new XATDataValue[] {
			});
		}

	}
	/**
	 * Insert the method's description here.
	 * Creation date: (5/17/2003 1:54:59 AM)
	 * @return java.lang.String
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("\n ");
		for (int i = 0; i < this.getLength(); i++) {
			Object cellValue = this.getCell(i).getValue();
			if (cellValue != null) {
				sb.append(cellValue.toString());
			} else {
				sb.append(" ");
			}
			if (i < this.getLength() - 1)
				sb.append("  |  ");
		}
		sb.append("  |  " + this.minTimeStamp + "  |  " + this.maxTimeStamp);
		if (this.isNewTuple)
		    sb.append(" new ");
		else 
		    sb.append(" old ");
		
		if (this.isSingleton)
		    sb.append(" sg ");
		sb.append("\n");
        
		return sb.toString();
	}

	/**
	 * Gets the time stamp.
	 * 
	 * Creation Date: (Jun 7, 2004 1:33:13 PM)
	 * @author Natasha Bogdanova
	 * @return long
	 */
	public long[] getTimeStamp() {
		return timeStamp;
//		long maxTimeStamp = 0;
//		try {
//			maxTimeStamp = this.timeStamp.getMaxTimeStamp();
//		} catch (XATTimeStampException e) {
//			System.out.println(e.toString());
//		}
//		return maxTimeStamp;
	}

	/**
	 * Sets the punctuation ID.
	 * 
	 * Creation Date: (Jun 7, 2004 1:35:16 PM)
	 * @author Natasha Bogdanova
	 * @param pid long
	 */
//	public void setPunctID(long pid) {
//		this.punctID = pid;
//	}

	/**
	 * Sets the time stamp.
	 * 
	 * Creation Date: (Jun 7, 2004 1:38:42 PM)
	 * @author Natasha Bogdanova
	 * @param ts long
	 */
	public void setTimeStamp(long[] ts) {
		//this.timeStamp.setMaxTimeStamp(ts);
		this.timeStamp = ts;
	}
	
	/**
	 * Returns the index of this tuple. 
	 * Note: the tuple will have a valid index value only after
	 * it has retreived from the disk.
	 * 
	 * Creation Date: (Jun 7, 2004 1:39:53 PM)
	 * @author Natasha Bogdanova
	 * @return int
	 */
	public int getIndex() {
		return this.index;
	}
	
	/**
	 * Sets the index of this tuple. Used by the storage manager
	 * when it retreives tuples from the disk.
	 * 
	 * Creation Date: (Jun 7, 2004 1:41:06 PM)
	 * @author Natasha Bogdanova
	 * @param index int
	 */
	public void setIndex(int index){
		this.index = index;
	}
	public long getMaxTimeStamp() {
		return maxTimeStamp;
	}
	public long getMinTimeStamp() {
		return minTimeStamp;
	}
	public void setMaxTimeStamp(long max) {
		maxTimeStamp = max;
	}
	public void setMinTimeStamp(long min) {
		minTimeStamp = min;
	}
	public long getNthTimeStamp(int n) {
		if(n<timeStamp.length) {
			return timeStamp[n];
		}
		return -1;
	}
	public void setNthTimeStamp(long time, int n) {
		if(n < timeStamp.length) {
			timeStamp[n] = time;
		}
	}
	public static XATTuple combineTuples(XATTuple tuple1, XATTuple tuple2) {
		int numLeftColumns = tuple1.getLength();
		int numRightColumns = tuple2.getLength();
		
		XATDataValue[] leftValues = tuple1.getArrayData();
		XATDataValue[] rightValues = tuple2.getArrayData();
		long[] leftTimeStamp = tuple1.getTimeStamp();
		long[] rightTimeStamp = tuple2.getTimeStamp();
		long maxTime1 = tuple1.getMaxTimeStamp();
		long maxTime2 = tuple2.getMaxTimeStamp();
		long minTime1 = tuple1.getMinTimeStamp();
		long minTime2 = tuple2.getMinTimeStamp();
		
		XATTuple newTuple = new XATTupleImp(numLeftColumns+numRightColumns);
		XATDataValue[] newValues = new XATDataValueImp[numLeftColumns+numRightColumns];
		long[] newTimeStamp = new long[numLeftColumns+numRightColumns];
		int i=0, j=0;
		for (; i<numLeftColumns; i++) {
			newValues[i] = leftValues[i];
			newTimeStamp[i] = leftTimeStamp[i];
		}
		for (; j<numRightColumns; j++) {
			newValues[i+j] = rightValues[j];
			newTimeStamp[i+j] = rightTimeStamp[j];
		}
		newTuple.setValues(newValues);
		newTuple.setTimeStamp(newTimeStamp);
		newTuple.setMaxTimeStamp(maxTime1>maxTime2?maxTime1:maxTime2);
		newTuple.setMinTimeStamp(minTime1<minTime2?minTime1:minTime2);
		
		if(tuple1.isNew() && tuple2.isNew()) {
			newTuple.setToNew(true);
		} 
		
		/*
		if (tuple1.isPunctuation() && tuple2.isPunctuation()) {
			int[] newPunctTypes = new int[numLeftColumns+numRightColumns];
			int[] leftPunctTypes = tuple1.getPunctTypeArray();
			int[] rightPunctTypes = tuple2.getPunctTypeArray();
			i = 0;
			j = 0;
			for (; i<numLeftColumns; i++) {
				newPunctTypes[i] = leftPunctTypes[i];
			}
			for (; j<numRightColumns; j++) {
				newPunctTypes[i+j] = rightPunctTypes[j];
			}
			newTuple.setPunctTypeArray(newPunctTypes);
			newTuple.setAsPunctuation(true);
		} else {
			newTuple.setAsPunctuation(false);
		}*/
		return newTuple;
	}

	/*
	 * added by Bin @ Oct. 2004
	 * For partitioned operator, that each tuple contains partitioned id information
	 * Usually, these information will be set by a steamsplitoperator
	 */
	public int getPartitionID(){
		return partitionID;
	}
	public void setPartitionID(int id){
		partitionID = id;
	}
	// value by value comparision of two tuples
	public boolean equals(XATTuple tp){
		
		//System.out.println(this.arrayCell.length);
		//System.out.println(tp.getLength());
		if (this.getLength() != tp.getLength())
			return false;
		XATDataValue[] tpValue = tp.getArrayData();
		for (int i=0; i<tpValue.length; i++) {
			if (tp.getLength()==1)
				{if(!tpValue[i].equals(oneCell))
						return false;}
			else if (!tpValue[i].equals(arrayCell[i])){
				return false;
			}
		}
		return true;
	}
	//
	public boolean equalKeys(Object[] inputKeys, int[] keyPos) {
		int numKeysToCompare = inputKeys.length;
		if (this.arrayCell.length < numKeysToCompare)
			return false;
		for (int i=0; i<numKeysToCompare; i++) {
			if (!arrayCell[keyPos[i]].getValue().equals(inputKeys[i])) {
				return false;
			}
		}
		return true;
	}
	
	public boolean equalKeys(XATDataValue[] punctContent, int[] punctType, int[] keysPos) {
		int numKeysToCompare = keysPos.length;
		if (this.arrayCell.length < numKeysToCompare)
			return false;
		for (int i=0; i<numKeysToCompare; i++) {
			int index = keysPos[i];
			if (punctType[index] == XATPunct.SINGLE_VALUE 
					&& !arrayCell[index].getValue().equals(punctContent[index].getValue())) {
				return false;
			}
		}
		return true;
	}

	//Rimma: checks if it's a routing token
	public boolean isRoutingToken() 
	{
		return (objTypeID == 3);
	}
	
	//Rimma: checks if it's a security punctuation
	public boolean isSecPunct() 
	{
		return (objTypeID == 2);
	}
	
	public boolean isSingletonTuple() {
	    return this.isSingleton;
	}
	
	public boolean isPunct() {
		return (objTypeID == 1);
	}
	
	public int getObjTypeID() {
		return objTypeID;
	}

	/**
	 * The following two functions are basically for pipelined stream operators. each tuple contains
	 * stream id information, and boolean probing or not.
	 * Added by Song Wang
	 * @return
	 */
	private int streamID=0; 
	private boolean isProbing=false; 
	public int getStreamID(){
		return streamID;
	}
	public void setStreamID(int id){
		streamID=id;
	}
	public boolean isProbe(){
		return isProbing;
	}
	public void setProbe(boolean probe){
		isProbing=probe;
	}
	/**
	 * The following functions are basically for pipelined multiway join operators. 
	 * each intermediate result tuple contains the state number to be joined [0 to n-1]
	 * initialized as -1.
	 * The boolean ending means if it is an end tuple.
	 * The opID means the operator ID that generate this intermediate result
	 * For the original input stream tuples, all these three variables are meaningless.
	 * 
	 * NOTE: the streamID above is used to identify the stream that triggers the intermediate result.
	 * For example, if one a tuple from stream A is processed, then all the intermediate result has streamID = 0
	 * if the join order is A->C->B->E->D, then the (A,C) pair will have streamID=0, stateToJoin=1,
	 * (A,C,B) pair will have streamID=0, stateToJoin=4.
	 * If there is another join order of B->C->A->D->E, then the (B,C,A) pair will have streamID=1, stateToJoin=3
	 * Of cause the (A,C,B) and (B,C,A) tuple have the same schema and can't be distinguished by the tuple itself.
	 * 
	 * Actually, we can assume that the same prefix intermediate result will have the same further join ordering
	 * That is, the (A,C,B) and (B,C,A) tuples do not need to distinguish at all. The further join ordering
	 * must be the same for them. so right now, the streamID is used for the matrix entry only.
	 * 
	 * Added by Song Wang
	 * @return
	 */

	private int stateToJoin=-1;
	private boolean ending=false;
	private int opID=-1;
	private boolean dummy=false;
	private boolean realPurge = false;

	//system arrivial time, for calculation of response time
	private long timeEnter = 0;
	public void setTimeEnter(long l){
		this.timeEnter = l;
	}
	public long getTimeEnter(){
		return this.timeEnter;
	}
	
	public int getStateToJoin(){
		return stateToJoin;
	}
	public void setStateToJoin(int i){
		this.stateToJoin=i;
	}
	public boolean getEnding(){
		return ending;
	}
	public void setEnding(boolean b){
		this.ending= b;
	}
	public boolean isDummy(){
		return dummy;
	}
	public void setDummy(boolean b){
		this.dummy= b;
	}
	public int getOpID(){
		return opID;
	}
	public void setOpID(int i){
		this.opID=i;
	}
	
	public boolean isRealPurge() {
		return this.realPurge;
	}
	public void setRealPurge(boolean b) {
		this.realPurge=b;
		
	}
	/**
     * Clones the XATTuple.
     * @since   1.0        
     */

    public Object clone() {
        try {
            return super.clone();
        } catch (CloneNotSupportedException e) {
            System.err.println(e.getMessage());
            return null;//DeepCopyMaker.makeDeepCopy(this);
        }
    }
    
	//Rimma: checks if it's an STQM punctuation
	public boolean isStqmPunct() 
	{		
		return (objTypeID == 4);
	}
	
	//Rimma: checks if it's a streaming tag punctuation
	public boolean isStreamingTagPunct() 
	{		
		return (objTypeID == 5);
	}

}
