package edu.wpi.cs.dsrg.xmldb.xat.common.expression;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;


/**
 * This class defines a binary comparision expression.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 */

public class BinCOMPExpressionImp
    extends BinExpressionImp
    implements BinCOMPExpression {
	/**
	 * The left and right expression.
	 * @since 1.0
	 */   
    Expression eLeft, eRight;
    /**
     * The type of comparison.
     * @since 1.0
     */
    int iType;
/**
 * The default binary expression constructor.
 * @since 1.0
 */
public BinCOMPExpressionImp() {
    eLeft = null;
    eRight = null;
}
/**
 * The binary expression constructor.
 * @param left  The left part of binary expression.
 * @param right The right part of binary expression.
 * @since 1.0
 */
public BinCOMPExpressionImp(Expression left, Expression right) {
    eLeft = left;
    eRight = right;
}
/**
 * Execute the eval function for atomic values
 * @return boolean The comparison result.
 * @param left java.lang.Object The left part of expression.
 * @param right java.lang.Object The right part of expression.
 * @since 1.0
 */
 
public Boolean compAtomic(Object left, Object right) {
    try {
        long lLeft = (new Long(left.toString())).longValue();
        long lRight = (new Long(right.toString())).longValue();
        boolean bReturn;
        switch (iType) {
            case LT :
                bReturn = (lLeft < lRight);
                break;
            case GT :
                bReturn = (lLeft > lRight);
                break;
            case LEQ :
                bReturn = (lLeft <= lRight);
                break;
            case GEQ :
                bReturn = (lLeft >= lRight);
                break;
            case EQ :
                bReturn = (lLeft == lRight);
                break;
            case NEQ :
                bReturn = (lLeft != lRight);
                break;
            default :
                /* fail */
                return Boolean.FALSE;
        }
        return new Boolean(bReturn);
    } catch (NumberFormatException e) {
        try {
            double dLeft = (new Double(left.toString())).doubleValue();
            double dRight = (new Double(right.toString())).doubleValue();
            boolean bReturn;
            switch (iType) {
                case LT :
                    bReturn = (dLeft < dRight);
                    break;
                case GT :
                    bReturn = (dLeft > dRight);
                    break;
                case LEQ :
                    bReturn = (dLeft <= dRight);
                    break;
                case GEQ :
                    bReturn = (dLeft >= dRight);
                    break;
                case EQ :
                    bReturn = (dLeft == dRight);
                    break;
                case NEQ :
                    bReturn = (dLeft != dRight);
                    break;
                default :
                    /* fail */
                    return Boolean.FALSE;
            }
            return new Boolean(bReturn);
        } catch (NumberFormatException e2) {
            String sLeft = left.toString();
            String sRight = right.toString();
            boolean bReturn;
            switch (iType) {
                case LT :
                    bReturn = (sLeft.compareTo(sRight) < 0);
                    break;
                case GT :
                    bReturn = (sLeft.compareTo(sRight) > 0);
                    break;
                case LEQ :
                    bReturn = (sLeft.compareTo(sRight) <= 0);
                    break;
                case GEQ :
                    bReturn = (sLeft.compareTo(sRight) >= 0);
                    break;
                case EQ :
                    bReturn = (sLeft.equals(sRight));
                    break;
                case NEQ :
                    bReturn = (!sLeft.equals(sRight));
                    break;
                default :
                    /* fail */
                    return Boolean.FALSE;
            }

            return new Boolean(bReturn);
        }
    } catch (NullPointerException e) {
        return Boolean.FALSE;
    }
}
/**
 * Evaluate the expression of the XATOperator.
 * @return The evaluate result of expression.
 * @param xopOper The operator evaluated.
 * @param iPositions The position of tuple in the XATTable.
 * @since 1.0
 */

public Object eval(XATOperator xopOper, int[] iPositions) {

    // pre 1/21/03: this code is written incorrectly... this need to check for column names and such
    // 1/21/03: check for vectors in left or right hand side.
    // if any element matches from the left or right with the opposite side,
    //  return true, else return false.

    Object lValue = eLeft.eval(xopOper, iPositions);
   
    
    Vector lCollection = new Vector();

    Object rValue = eRight.eval(xopOper, iPositions);
    
    System.out.println("BinComp Evaluate expression \n");
    System.out.println("lValue: " + lValue  + "rValue: " + rValue);
    Vector rCollection = new Vector();

    if ((lValue == null) || (rValue == null))
        return new XATDataValueImp(Boolean.FALSE);

    //Not sure what following try to do
    //Modified to:
    //1. First change lValue and rValue to vector both
    //2. Iterate the two vectors to compare atomic items
    //3. ANY match, return TURE.
    //4. Not sure what convertToCollection() does! It seems it only handle vector, string and DocumentFragment
    //   So I just simply convert to string.
    //Rewriten by Song Wang

    lCollection = ((XATDataValue) lValue).convertToCollection();

    rCollection = ((XATDataValue) rValue).convertToCollection();

    for (int i = 0; i < lCollection.size(); i++) {
        for (int j = 0; j < rCollection.size(); j++) {
            if (compAtomic(lCollection.elementAt(i), rCollection.elementAt(j))
                .booleanValue())
            {
                System.out.println("Ccomparison result: TRUE");
            	return new XATDataValueImp(Boolean.TRUE);
            }
        }
    }
    System.out.println("Ccomparison result: FALSE");
    return new XATDataValueImp(Boolean.FALSE);
}
/**
 * Get the left part of expression.
 * @return Expression The left part of binary expression
 * @since 1.0
 */
 
public Expression getLeft() {
    return eLeft;
}
/**
 * Get the right part of expression.
 * @return Expression The right part of binary expression
 * @since 1.0
 */
public Expression getRight() {
    return eRight;
}
/**
 * Get the type of expression.
 * @return int The type of expression.
 * @since 1.0
 */
 
public int getType() {
    return iType;
}
/**
 * Set the left part of expression.
 * @param left The left part of a binary expression.
 * @since 1.0
 */
public void setLeft(Expression left) {
    eLeft = left;
}
/**
 * Set the right part of expression.
 * @param left The right part of a binary expression.
 * @since 1.0
 */
 
public void setRight(Expression right) {
    eRight = right;
}
/**
 * Set the type of the expression.
 * @param type The number used to represent the type of expression.
 * @since 1.0
 */
public void setType(int type) {
    iType = type;
}

/**
 * Output a string of this expression.
 * @return String the output result.
 * @since 1.0
 */
public String toString() {
    switch (iType) {
        case LT :
            return "(" + eLeft.toString() + " < " + eRight.toString() + ")";
        case GT :
            return "(" + eLeft.toString() + " > " + eRight.toString() + ")";
        case LEQ :
            return "(" + eLeft.toString() + " <= " + eRight.toString() + ")";
        case GEQ :
            return "(" + eLeft.toString() + " >= " + eRight.toString() + ")";
        case EQ :
            return "(" + eLeft.toString() + " == " + eRight.toString() + ")";
        case NEQ :
            return "(" + eLeft.toString() + " != " + eRight.toString() + ")";
        default :
            return "("
                + eLeft.toString()
                + " COMP("
                + Integer.toString(iType)
                + ") "
                + eRight.toString()
                + ")";
    }
}
}
