package edu.wpi.cs.dsrg.xmldb.xat.common.expression;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;

/**
 * This class defines a Uniary minus expression.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 */

public class UniMinusExpressionImp implements UniMinusExpression {

	/**
	 * The expression value.
	 * @since 1.0
	 */
    Expression eValue;
/**
 * The default UniMinusExpressionImp constructor.
 * @param exp The expression used to construct the UniMinusExpression.
 * @since 1.0
 */
public UniMinusExpressionImp(Expression exp) {
    eValue = exp;
}
/**
 * Evaluate the expression of the XATOperator.
 * @return The evaluate result of expression.
 * @param xopOper The operator evaluated.
 * @param iPositions The position of tuple in the XATTable.
 * @since 1.0
 */

public Object eval(XATOperator xopOper, int[] iPositions) {
    try {
        return new XATDataValueImp(
            new Double(- ((Number) eValue.eval(xopOper, iPositions)).doubleValue()));
    } catch (ClassCastException e2) {
        /* fail */
        return new XATDataValueImp();
    }
}
/**
 * Get the expression.
 * @return This uniary expression.
 * @since 1.0
 */ 
public Expression getExpression() {
    return eValue;
}
/**
 * get column names related with expression.
 * @return ColumnName[] The array containing all the related column names.
 * @since 1.0
 */
public ColumnName[] getRelatedColumnNames() {
    return eValue.getRelatedColumnNames();
}
/**
 * Set this uniary expression using given expression.
 * @param The given expression used to set this uniary expression.
 * @since 1.0
 */ 

public void setExpression(Expression exp) {
    eValue = exp;
}

/**
 * Output a string of this expression.
 * @return String the output result.
 * @since 1.0
 */
public String toString() {
    return "-" + eValue.toString();
}
}
