/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.migration.probablisticjoinoperator;

//import java.io.FileWriter;
//import java.io.IOException;
import java.util.*;
//import java.lang.reflect.Constructor;

//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDagImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.ProbabilisticMultiStreamOperatorImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATWBinaryJoinOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;

/**
 * Creation date: (11/4/2005 2:35:37 PM)
 * @author: Yali Zhu
 * 
 * This class represents a binary join with time-based window constraint. 
 * The selectivity of the join is generated by a random generator at a given interval.
 */
public class ChangingWBJoinOperatorImp extends ProbabilisticMultiStreamOperatorImp {

	/**
	 * copy constructor
	 * @param oldop
	 * @param newnode
	 */
	private double delay;

	//set to true and checked during processing only if isOldOperator == true
	private boolean isOldBoxLeaf = false;
	private boolean isOldBoxRoot = false;
	private boolean isOldOperator = false;
	private boolean isRoot = false;
	
//	private XATQueue lQueue;
//	private XATQueue oQueue;
	private java.util.Random rand;
//	private XATQueue rQueue;
	private double selectivity;
	private double oldselectivity;
	long timeWindow;
	
	// random selectivity assignment
	Random selectRand;
	// the trend of selectivity changing. "+" indicates increase while "-" indicates decrease.
	String selectTrend; 
	long selectInterval;
	long startTime;
	long trendStartTime;
	XATDag dag;
	
//	// stats ONLY collected inside the root operator
//	int stats_total_output_tuples = 0;
//	int stats_output_tuple_count = 0;
//	StringBuffer stats_string;
//	StringBuffer stats_output_tuples;
//	int[] totalTuplesDequeued;
//	long stats_start_time;
	
	public ChangingWBJoinOperatorImp(ChangingWBJoinOperatorImp oldop, XATNode newnode) {
		super(oldop, newnode);

		this.delay = oldop.delay;
		this.rand = oldop.rand;
		this.selectivity = oldop.selectivity;
		this.timeWindow = oldop.timeWindow;
	}
	/**
	 * DummyWindowJoinOperator constructor comment.
	 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
	 */
	public ChangingWBJoinOperatorImp(XATNode node) {
		super(node);
		selectivity = -1;
		delay = -1;
		rand = new java.util.Random();
		rand.setSeed(0);
		
		selectRand = new Random();
		//selectRand.setSeed(1);
		startTime = System.currentTimeMillis();
		trendStartTime = startTime;
		
		String statetype = this.getProperty("STATE_TYPE");
		try{			
			states = new WOperatorState[2];
			states[0] = (WOperatorState)Class.forName(statetype).newInstance();
			states[1] = (WOperatorState)Class.forName(statetype).newInstance();
			states[0].setOperator(this);
			states[1].setOperator(this);
		}catch(Exception e) {
			e.printStackTrace();	
		}
	}
	
	

/**
 * This will actually evaluate the join predicate.  It will return true if
 * the tuples are valid and should be outputted, false if they should be discarded
 * b/c they didnt meet the predicate
 * Creation date: (4/5/2003 6:59:19 PM)
 */
public boolean evalJoinPredicate() {
    	
	//set the selectivity we are supposed to have    
	if(selectivity == -1){   	
		selectivity = Double.parseDouble(this.getProperty("REAL_SELECTIVITY"));
	}
	
	//add delay to simulate predicate evaluation
//	for(int i=0; i<10; i++) {
//	}
	
    //return true if the next double is less than the selectivity,
    //otherwise return false
	
// 	try {
//		Thread.sleep(1);
//	} catch (InterruptedException e) {
//		e.printStackTrace();
//	}
	
    if (rand.nextDouble() < selectivity){
    	return true;
    }
    else{
    	return false;
    }
    
}
	
	public int execOperator(XATTuple[] tuples, WOperatorState state, boolean left){
		
		if(state.getStateSize() == 0){
			return 0;
		}
		int count = 0;
		XATTuple t1;
		XATTuple t2;
		XATTuple joinTuple;

		
		//figure out how many tuples should be generated based on this operator's selectivity
		long outputcount = Double.doubleToLongBits(tuples.length * state.getStateSize() * this.selectivity);  
		
		if(tuples == null || tuples.length < 1) {
			return 0;
		}
		state.purgeByTuple(tuples[0]);
		
		XATTuple[] stateTuples = state.getAllStateTuples();
		
		for(int i = 0; i < tuples.length; i++){
			t1 = tuples[i];
			
			// for distributed migration
			if(t1.getMaxTimeStamp() == -1) {
				this.processSyncTuple();
				continue;
			}
			
			for(int j = 0; j < stateTuples.length; j++){
				t2 = stateTuples[j];
				if(joinable(t1,t2)){
					if(evalJoinPredicate()){
						if(left){
							joinTuple = join(t1,t2);
						}
						else{
							joinTuple = join(t2,t1);
						}
						
						this.getOutputQueue().enqueue(joinTuple,this);
						count++;
					}		
				}
			}
		}
		
		return count;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/3/2003 5:00:07 AM)
	 * @return queryplanelements.WFIFOState
	 */
	public OperatorState getLeftState() {
		return states[0];
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/3/2003 5:00:50 AM)
	 * @return queryplanelements.WFIFOState
	 */
	public OperatorState getRightState() {
		return states[1];
	}
	
	public void setLeftState(OperatorState lstate) {
		this.states[0] = lstate;
	}
	
	public void setRightState(OperatorState rstate) {
		this.states[1] = rstate;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#getStateSize()
	 */
	public int getStateSize() {
		return states[0].getStateSize() + states[1].getStateSize();
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (3/15/2003 10:23:33 PM)
	 * @return int
	 */
	public long getTimeWindow() {
		return timeWindow;
	}

//	public void init() {
//		Iterator i = this.getInputQueues();
//		String schemaName = new String();
//		XATQueueSchema s;
//		XATQueueSchema outSchema = new XATQueueSchema();
//
//		while (i.hasNext()) {
//
//			s = ((XATQueue) i.next()).getSchema();
//			schemaName += "#" + s.getName();
//			
//			Iterator j = s.getColNames().iterator();
//			while(j.hasNext()){
//				outSchema.addColumn((String) j.next());
//			}
//		}
//
//		outSchema.setName(schemaName);
//		this.getOutputQueue().setSchema(outSchema);
//		
//		oQueue = this.getOutputQueue();
//		lQueue = this.getLeftInputQueue();
//		rQueue = this.getRightInputQueue();
//
//	}
/**
 * Insert the method's description here.
 * Creation date: (6/10/2003 2:26:12 PM)
 */

public XATTuple join(XATTuple lTuple, XATTuple rTuple) {

    /** join data together */
    //ArrayList data = (ArrayList) Arrays.asList(lTuple.getCurrentTuple());
    //data.addAll(Arrays.asList(rTuple.getCurrentTuple()));
    XATDataValue[] data1 = lTuple.getArrayData();
    XATDataValue[] data2 = rTuple.getArrayData();
    int size1 = data1.length;
    int size2 = data2.length;

    // we are not worried about schema here, just the join, so just make the new tuple.
    ArrayList newData = new ArrayList();
    for (int i = 0; i < size1; i++)
        newData.add(data1[i]);
    for (int i = 0; i < size2; i++)
        newData.add(data2[i]);

    XATTuple newTuple =
        new XATTupleImp( (XATDataValue[]) (newData.toArray(new XATDataValue[] {})) );

	newTuple.combineTimeStamps(lTuple, rTuple);

    return newTuple;
}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/3/2003 10:16:40 PM)
	 * @return boolean
	 */
	public boolean joinable(XATTuple tuple1, XATTuple tuple2) {
	
		/** If this is the last operator and if both tuples' are NEW, don't join */
		/*if (isOldBoxRoot)
		    if (tuple1.isNew() && tuple2.isNew())
		        return false; */
	
		/** if tuple1 and tuple2 is out of window size from each other, don't join */
		try {
			if (tuple1 == null || tuple1.getTimeStamp() == null || tuple2 == null || tuple2.getTimeStamp() == null)
				return false;
	
			if (!(tuple1.getMaxTimeStamp()
				- tuple2.getMinTimeStamp()
				< timeWindow))
				return false;
			if (!(tuple2.getMaxTimeStamp()
				- tuple1.getMinTimeStamp()
				< timeWindow))
				return false;
		} catch (Exception e) {
			System.out.println(e);
			return false;
		}
	
		return true;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() {
		
//		Iterator i = this.getInputQueues();
//
//		while (i.hasNext()) {
//
//			XATQueue q = (XATQueue) i.next();
//			if (q.getSchema() == null){
//				return false;
//			}
//		}
//
//		try{
//		if (this.getOutputQueue().getSchema() == null) {
//			System.out.println("init operator " + this);
//			init();
//		}
//		}catch(Exception e) {
//			System.out.println("\nCurrentOperator = " + this);
//			System.out.println("Parents are " + this.getNode().getParent());
//			System.out.println("Children are " + this.getNode().getChildren());
//		}
		

		int tuplesInLeft = this.getLeftInputQueue().getQueueSize(this);
		int tuplesInRight = this.getRightInputQueue().getQueueSize(this);
		int tuplesInLeftStateBefore = getLeftState().getStateSize();
		int tuplesInRightStateBefore = getRightState().getStateSize();
		int leftOutput = 0, rightOutput=0;
		
		//if(tuplesInLeft > 300){
		//	tuplesInLeft = 300;
		//}
		//if(tuplesInRight > 300){
		//	tuplesInRight = 300;
		//}
		
		//System.out.println("selectivity = " + this.selectivity);

//		//if time to change selectivity, randomly generate a new selectivity
//		if(System.currentTimeMillis() - startTime >= 20000) {
//			this.oldselectivity = this.selectivity;
//			this.selectivity = this.selectRand.nextDouble();
//			startTime = System.currentTimeMillis();
//			try{
//				this.setProperty("SELECTIVITY", ""+this.selectivity);
//				System.out.println("\n\nnew selectivity = " + this.selectivity + " for operator " + this.getObjectID());
//			}catch(Exception e) {
//				e.printStackTrace();
//			}
//		}
		
//		//if time to change selectivity, randomly generate a new selectivity
//		if(System.currentTimeMillis() - startTime >= 10000) {
//			this.oldselectivity = this.selectivity;
//			if (selectTrend.equals("+")) {
//			    this.selectivity += 0.1;
//			    if(System.currentTimeMillis() - trendStartTime >= 50000) {
//			        this.selectTrend = "-";
//			    	trendStartTime = System.currentTimeMillis();
//			    }
//			}
//			else {
//			    this.selectivity -= 0.1;
//			    if(this.selectivity <= 0) 
//			        this.selectivity = 0.01;
//			    if(System.currentTimeMillis() - trendStartTime >= 50000) {
//			        this.selectTrend = "+";
//			        trendStartTime = System.currentTimeMillis();
//			    }
//			}
//			//this.selectivity = this.selectRand.nextDouble();
//			startTime = System.currentTimeMillis();
//			
//			try{
//				this.setProperty("SELECTIVITY", ""+this.selectivity);
//				System.out.println("\n\nnew selectivity = " + this.selectivity + " for operator " + this.getObjectID());
//			}catch(Exception e) {
//				e.printStackTrace();
//			}
//		}
		
		//if time to change selectivity, choose a new selectivity
		if(System.currentTimeMillis() - startTime >= 30000) {
			this.oldselectivity = this.selectivity;
			if (this.selectivity == 0.02) {
			    this.selectivity = 0.2;
			}
			else if (this.selectivity == 0.2){
			    this.selectivity = 0.02;
			}
			startTime = System.currentTimeMillis();
			
			try{
				this.setProperty("SELECTIVITY", ""+this.selectivity);
				System.out.println("\n\nnew selectivity = " + this.selectivity + " for operator " + this.getObjectID());
			}catch(Exception e) {
				e.printStackTrace();
			}
		}
		
		//First join the tuples from the left queue with the right state
		if(tuplesInLeft > 0){
			XATTuple[] leftTuples = this.getLeftInputQueue().dequeue(tuplesInLeft,this);
			leftOutput = execOperator(leftTuples,(WOperatorState)getRightState(),true);	
			//Now add the tuples from the left queue to the left state
			((WOperatorState)getLeftState()).insertTuple(leftTuples);		
		}
		
		//Second join the tuples from the right queue with the left state		
		if(tuplesInRight > 0){
			XATTuple[] rightTuples = this.getRightInputQueue().dequeue(tuplesInRight,this);
			rightOutput = execOperator(rightTuples,(WOperatorState)getLeftState(),false);
			//Now add the tuples from the right queue to the right state.
			((WOperatorState)getRightState()).insertTuple(rightTuples);
		}

		int totalOutput = leftOutput + rightOutput;
		//System.out.println("# of total output tuples = " + totalOutput);

		if (totalOutput > 0) {
			if(this.isRoot) {
				dag.updateThroughputStats(totalOutput);
			}
		}
				//System.out.println("This is root operator, update stats");
				//stats_total_output_tuples += totalOutput;
				//System.out.println((System.currentTimeMillis() - dag.stats_start_time) + " # of total output tuples = " + dag.stats_total_output_tuples);
				//dag.stats_string.append((System.currentTimeMillis() - this.dag.stats_start_time)
				//			+ " \t " + this.dag.stats_total_output_tuples + "\n");

				//System.out.println((System.currentTimeMillis() - this.stats_start_time) + " # of total output tuples = " + this.stats_total_output_tuples);
				
				//for (int k = 0; k < output.length; k++) {
				//	this.stats_output_tuple_count++;
				//	this.stats_output_tuples.append(stats_output_tuple_count + "\t" + output[k]+ "\n");
				//}		
			//} 
			//else {
				//System.out.println("This is non-root operator, tuples outputted " + output.length);
			//}
			
			//if(output != null)
			//	System.out.println("# of output tuples " + output.length);
			//for (int k = 0; k < output.length; k++) {
			//	System.out.println("output tuple " + k + " " + output[k].toString());
			//}
		//}

		/* run successfully */
		return true;
	}
	
	/**
	 * Insert the method's description here.
	 * Creation date: (4/6/2003 7:46:53 AM)
	 */
	public void setAsBoxLeaf() {
		isOldBoxLeaf = true;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/6/2003 7:46:04 AM)
	 */
	public void setAsBoxRoot() {
		isOldBoxRoot = true;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/6/2003 8:29:45 AM)
	 */
	public void setAsOldOperator() {
		isOldOperator = true;
	}
/**
 * The Dummy operator has 2 parameters.  The SELECTIVITY
 * is a value between 0 and 1 and RESULT_DELAY is the
 * time it takes (in milleseconds) for this operator to place a tuple in its output queue
 * 
 */
protected void setDefaultProperties() {
	super.setDefaultProperties();
	addProperty("RESULT_DELAY","100");
	addProperty("REAL_SELECTIVITY",".5");	
	addProperty("STATE_TYPE", "edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WFIFOState");	
	addProperty("SELECT_RAND_SEED", "0");
}
/**
* sets the given property to the new value.  This method is responsible
* for validating the given property.  An exception is thrown if
* the value is not correct.
* Validation should be done in the following manner: 
* First the method should check if the object supports the given property by calling
* <tt>isValidPropertyName(String)<tt>.  If it is not a valid property name, then an exception should
* be thrown. <p>
*
* Once it is known that the object supports the property, the value must be validated.  Each implementor of
* XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
* For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
* has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
* to set a common type, it should be validated by the super class by calling <tt>super.setProperty(name, value)<tt>.  This
* will reduce duplicate code. <p>
*
* When it is determined the name / value pair is correct, the pair is updated using <tt>setNewPropertyValue()<tt> <p>
* <tt>XATStatisticsImp<tt> provides default validation by checking to see if the property's value is
* >=0.  propertyValues that are Strings will just be set.  If the property requires more strict constraints than this, 
* it is up to the subclass that created the property to validate it (as described above).
* @param name the name of the property to set
* @param value the new value of the property
* @throws Exception if the given property cannot be found or
* if the value is not valid for the given property.
*/
public void setProperty(String name, String value) throws java.lang.Exception {	
	//first see if the DummyOperator supports the property name	
	if(isValidPropertyName(name)){		
		if(name.equals("REAL_SELECTIVITY")){
			try {
				this.selectivity = Double.parseDouble(value);
				this.oldselectivity = selectivity;
				System.out.println("\n\nSetting op "+ this + " selectivity to be " + this.selectivity);
				if((selectivity >= 0) && (selectivity <= 2)){
					setNewPropertyValue(name, value);
					super.setProperty("SELECTIVITY", value);
					//System.out.println("Set the initial selectivity to: " + selectivity);
				} else{
					throw new NumberFormatException("The property named " + name + " requires a double value between 0 and 2 (inclusive)");
				}
			} catch (NumberFormatException e) {
				throw new NumberFormatException("The property named " + name + " requires a double value between 0 and 2 (inclusive)");
			}			
		}else{
			//the property name is valid, but its value is validated in the super class
			super.setProperty(name, value);
		}
	} else throw new Exception("The DummyOperator does not support the property named " + name);
	
	
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (3/15/2003 10:24:23 PM)
	 * @param newWindow int
	 */
	public void setTimeWindow(long newWindow) {
	
		//join operator has the same timeWindow with both its states
		timeWindow = newWindow;
		((WOperatorState)states[0]).setTimeWindow(newWindow);
		((WOperatorState)states[1]).setTimeWindow(newWindow);
		//System.out.println("The Time window has been set to: " + newWindow);
	}
	
	public void setSelectivity(double s) {
		this.selectivity = s;
		this.oldselectivity = s;
	}
	
	public void setSelectRandSeed(String seed) {
		this.selectRand.setSeed(Long.parseLong(seed));
		System.out.println("\n\nSetting select rand seed as " + seed);
	}
	
	public void setSelectTrend(String trend) {
	    this.selectTrend = trend;
	    System.out.println("\n\nSetting select trend as " + trend);
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		String rs = new String();

		String sName = this.getClass().getName();
		//find the last dot.
		int index = 0;
		for(index = sName.length()-1; index >=0; index--)
			if(sName.charAt(index) == '.')
				break;
		rs = sName.substring(index + 1) + " ";
		String id = this.getProperty("OBJECT_ID");
		if(id == "-1"){
			rs = rs + hashCode(); 
		} else{
			rs = rs + id;
		}	
		if(this.isActive()) rs += " ACTIVE ";
		else rs+= " INACTIVE ";

		/** now add the join predicate and window size info */
		rs += "\n";
		rs += "window size = "+ this.getTimeWindow()+"\n";
		rs += "state size = " + this.getStateSize()+"\n";

		return rs;
	}

	public void setAsRootOperator(boolean isroot) {
		this.isRoot = isroot;
	}
	
	public boolean isRootOperator() {
		return this.isRoot;
	}
	
	public void setDag(XATDag dag) {
		this.dag = dag; 
	}
	
	/**
	 * This method joins two WFIFOState of the current operator.
	 * 
	 * @return ArrayList of the joined tuples.
	 */
	public ArrayList joinFIFOStates() {
		ArrayList tuples = new ArrayList();
		
		int lstatesize = states[0].getStateSize();
		int rstatesize = states[1].getStateSize();
		
		WFIFOState lstate = (WFIFOState)states[0];
		WFIFOState rstate = (WFIFOState)states[1];
		
		XATTuple ltuple;
		XATTuple rtuple;
		double newstatesize = lstatesize * rstatesize * this.oldselectivity;
		double counter = 0;
		boolean done = false;
		for (int i = 0; i < lstatesize; i++) {
            if (!done) {
                ltuple = lstate.getStateTuple(i);
                for (int j = 0; j < rstatesize; j++) {
                    rtuple = rstate.getStateTuple(j);
                    //add delay to simulate predicate evaluation
                    //				for(int k=0; k<10; k++) {
                    //				}
                    //if((rand.nextDouble() < this.oldselectivity) &&
                    // (this.joinable(ltuple, rtuple))) {
                    if (counter < newstatesize) {
                        XATTuple result = this.join(ltuple, rtuple);
                        tuples.add(result);
                        counter++;
                    } else {
                        done = true;
                        break;
                    }
                }
            } else {
                break;
            }
        }
		
		System.out.println("lstate size = " + lstatesize + " rstate size = " + rstatesize + " old selectivity = " + oldselectivity);
		System.out.println("recomputed state tuples = " + tuples.size());
		
		return tuples;
	}
}
