/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.migration.probablisticjoinoperator;

import java.io.FileWriter;
import java.io.IOException;
import java.util.*;

//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.ProbabilisticMultiStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.mjoinoptimizer.JoinGraph;
//import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;

/**
 * Insert the type's description here.
 * Creation date: (10/21/2004 3:58:37 PM)
 * @author: Yali Zhu
 */
public class ProbabilisticWMJoinOperator extends ProbabilisticMultiStreamOperatorImp {

	/**
	 * copy constructor
	 * @param oldop
	 * @param newnode
	 */
	private double delay;
	
	private JoinGraph joinGraph = null;

	//set to true and checked during processing only if isOldOperator == true
	private boolean isOldBoxLeaf = false;
	private boolean isOldBoxRoot = false;
	private boolean isOldOperator = false;
	private boolean isRoot = false;
	
	
	// stats ONLY collected inside the root operator
	int stats_total_output_tuples = 0;
	int stats_output_tuple_count = 0;
	StringBuffer stats_string;
	StringBuffer stats_output_tuples;
	int[] totalTuplesDequeued;
	long stats_start_time;
	
	//SubWMJoinOperator rootSubOperator;
	
	// array of all suboperators, in the order from leaf to root.
	//SubWMJoinOperator allSubOperators[];
	
	// runningmode can be the following (defined in the xml query plan config file)
	// mjoin: a single mjoin suboperator
	// mtreewithsequence:  an mjoin tree (consists of mjoin suboperator) is generated based on pre-computed probe sequences
	// mtreewithoutsequence: an mjoin tree is generated without a pre-computed probe sequences
	// mtreemerge: an mjoin tree is generated by merging bottom up from the min-state binary join tree 
	private String runningMode;
	private boolean cartesionProductStatus = false;
	private boolean updateSelectivityStatus = false;
	
	private java.util.Random rand;
	long timeWindow;
	
	//boolean setQueueMapping = false;
	//QueueConnector[] queueconnectors;
	
	// for an m-join, the probeSequence is an m x (m-1) 2d array.
	private int[][] probeSequence;
	private double[][] probeSelectivityMatrix;
	
	//each input rates is in unit of tuples per millisecond.
	//private double input_rates[];
	
	//how many input stream this mjoin operates on. for binary join, this is set to be 2.
	int numOfInputs = 0;  
	
	// selectivityMatrix[i][j] stores the selectivity between input i and j.
	// selectivityMatrix[i][j] = selectivityMatrix[j][i]
	//
	// during join processing, if i<j, then i is treated as the left side and j at the right side of the joined tuple.
	// if i>j, i needs to be treated as the right side and j the left side.
	// this gurantee 1) the right predicate evaluation, and 2) the right order of columns no matter what the join order is.
	// this matrix is generated during parsing of a mjoin operator, by queryplangenerator.ClassVariableVisitor.
	private double[][] selectivityMatrix; 
	
	// arrivalRates[i] stores the arrival rate of the ith input.
	private double[] arrivalRates;
	
	public ProbabilisticWMJoinOperator(ProbabilisticWMJoinOperator oldop, XATNode newnode) {
		super(oldop, newnode);

		this.delay = oldop.delay;
		this.rand = oldop.rand;
		this.selectivityMatrix = oldop.selectivityMatrix;
		this.timeWindow = oldop.timeWindow;
		this.joinGraph = oldop.joinGraph;
	}
	
	/**
	 * DummyWindowJoinOperator constructor comment.
	 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
	 */
	public ProbabilisticWMJoinOperator(XATNode node) {
		super(node);
		delay = -1;
		rand = new java.util.Random();
		rand.setSeed(0);
	}
	
	/*
	 * The "numberofinputs" MUST be set before calling this method.
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#init()
	 */
	public void setupStates() {
		
		String statetype = this.getProperty("STATE_TYPE");
		try{			
			// set one state for each input
			states = new WOperatorState[numOfInputs];
			for(int i=0; i<numOfInputs; i++) {
				states[i] = (WOperatorState)Class.forName(statetype).newInstance();
				states[i].setOperator(this);
		        ((WOperatorState)states[i]).setTimeWindow(this.timeWindow);
			}
		}catch(Exception e) {
			e.printStackTrace();	
		}
	}
	
			
			//rootSubOperator = new SubWMJoinOperator(this, this.numOfInputs);
			//allSubOperators = new SubWMJoinOperator[1];
			//allSubOperators[0] = rootSubOperator;
			//this.computeProbeSequence();
			//rootSubOperator.setSubProbeSequence(this.probeSequence);
			//rootSubOperator.setSelectivities(this.selectivityMatrix);
			//rootSubOperator.setTimeWindow(this.timeWindow);
			//rootSubOperator.initStates();
		
			//return;
		//}

//		// if running as a min-state binary join tree
//		if(runningMode.compareToIgnoreCase("btree") == 0) {
//			generateMinStateBTree();
//			return;
//		}
//		
//		if(runningMode.compareToIgnoreCase("mtreewithsequence") == 0) {
//			this.computeProbeSequence();
//			generateMTreeWithSequence();
//			return;
//		}
//		
//		if(runningMode.compareToIgnoreCase("mtreewithoutsequence") == 0) {
//			generateMTreeWithoutSequence();
//			return;
//		}
//		
//		if(runningMode.compareToIgnoreCase("mtreemerge") == 0) {
//			return;
//		}
	//}
	
	/**
	 * compute the new probe sequence based on the selectivity and input rates
	 * this method is called in the classvariable parser if the running mode is set to "mjoin".
	 * we use the KBZ algorithm to compute the optimal sequence for acyclic join graph
	 * for cyclic join graph, the 
	 */
	public void computeProbeSequence() {

		// if running as a single mjoin operator
		// create a single mjoin suboperator
		if (runningMode.compareToIgnoreCase("mjoin") == 0) {

			this.probeSequence = new int[this.numOfInputs][this.numOfInputs - 1];
			this.joinGraph = new JoinGraph(this.selectivityMatrix,
					this.arrivalRates, this.updateSelectivityStatus);
			this.joinGraph.setTimeWindow(this.timeWindow);

			for (int i = 0; i < this.numOfInputs; i++) {
				this.probeSequence[i] = this.joinGraph.computeIthJoinOrder(i);
			}
			printProbeSequence();
			computeProbeSelectivityMatrix();
			printProbeSelectivityMatrix();
		}
	}
	
	/**
	 * Compute the correct selectivities given the computed probe sequences.
	 * Also set up the random generator for each selectivity;
	 * This selectivity matrix is used at runtime for looking up certain selectivity between two consecutive 
	 * stream pairs in a join sequence.
	 */
	private void computeProbeSelectivityMatrix() {
		this.probeSelectivityMatrix = new double[this.numOfInputs][this.numOfInputs - 1];
		//this.rand = new java.util.Random[this.numOfInputs][this.numOfInputs - 1];
		
		for(int i=0; i<this.numOfInputs; i++) {
			int index0 = this.probeSequence[i][0];
			probeSelectivityMatrix[i][0] = this.selectivityMatrix[i][index0];
			for(int j=1; j<this.numOfInputs-1; j++) {
				int indexj = this.probeSequence[i][j];
				probeSelectivityMatrix[i][j] = this.selectivityMatrix[i][indexj];
				for(int k=0; k<j; k++) {
					int indexk = this.probeSequence[i][k];
					probeSelectivityMatrix[i][j] *= this.selectivityMatrix[indexj][indexk];
				}
				
				//rand[i][j] = new Random();
				//rand[i][j].setSeed(0);
			}
		}
	}
	
	private void printProbeSequence() {
		for(int i=0; i<this.numOfInputs; i++) {
			System.out.print("join order for input "+i + ": ");
			for(int k=0; k<this.numOfInputs - 1; k++) {
				System.out.print(" -> " + this.probeSequence[i][k]);
			}
			System.out.println();
		}
	}
	
	private void printProbeSelectivityMatrix() {
		for(int i=0; i<this.numOfInputs; i++) {
			System.out.print("join selectivity for input "+i + ": ");
			for(int k=0; k<this.numOfInputs - 1; k++) {
				System.out.print(" -> " + this.probeSelectivityMatrix[i][k]);
			}
			System.out.println();
		}
	}
	
	private void printStatsToFile() {
		if (this.isRoot) {
			try {
				String fileName = new String(runningMode + "_" + System.currentTimeMillis()+".txt");
				FileWriter out = new FileWriter(fileName, true);
				out.write(this.stats_string.toString());
				out.close();
				
				//fileName = new String(runningMode + "_output_" + System.currentTimeMillis()+".txt");
				//out = new FileWriter(fileName, true);
				//out.write(this.stats_output_tuples.toString());
				//out.close();
			} catch (IOException e) {
				System.out.println(e.toString());
			}
		}
	}
	
	/**
	 * @param tuples
	 * @param state
	 * @param selectivity
	 * @param stateonright if the index of state is smaller (left) or larger (right) 
	 *        than the index of the queues the input tuples are dequeued from)
	 * @return list of joined tuple
	 */
	private XATTuple[] execOperator(XATTuple[] tuples, WOperatorState state, double selectivity, long overall_min){
		
		if(state.getStateSize() == 0){
			return null;
		}
		ArrayList joinedTuples = new ArrayList();
		
		XATTuple t1;
		XATTuple t2;
		XATTuple joinTuple;
		
		ArrayList stateTuples = state.getStateTuples();
		//System.out.println("number of state tuples = " + stateTuples.size());
		
		for(int i = 0; i < tuples.length; i++){
			t1 = tuples[i];
			if(overall_min != 0) 
				state.purgeByTuple(t1, overall_min);   // Purge the tuples we dont need
			int count = 0;
			int joincount = 0;
			
			for(int j = 0; j < stateTuples.size(); j++){
				t2 = (XATTuple)stateTuples.get(j);
				if(joinable(t1,t2)){
					count++;
				    if (rand.nextDouble() < selectivity){
						joinTuple = join(t1,t2);
						//this.getOutputQueue().enqueue(joinTuple,this);
						joinedTuples.add(joinTuple);
						joincount++;
						//System.out.println("join with a state tuple");
					}		
				}
			}
			//System.out.println("number of joinable tuples in state = " + count);
			//System.out.println("number of joined state tuples = " + joincount);
		}
		
		//System.out.println("total joined tuples in this probe = " + joinedTuples.size());
		XATTuple[] outputTuples = (XATTuple[])joinedTuples.toArray(new XATTuple[] {});
		joinedTuples = null;
		return outputTuples;

		//return joinedTuples;
	}
	
//	public int getNumOfInputs() {
//		return this.numOfInputs;
//	}
	
//	public OperatorState getIthState(int i) {
//		if(i>= 0 && i<this.numOfInputs)
//			return states[i];
//		else
//			return null;
//	}
	
//	/**
//	 * Insert the method's description here.
//	 * Creation date: (4/3/2003 5:00:07 AM)
//	 * @return queryplanelements.WFIFOState
//	 */
//	public OperatorState getLeftState() {
//		return states[0];
//	}
//	/**
//	 * Insert the method's description here.
//	 * Creation date: (4/3/2003 5:00:50 AM)
//	 * @return queryplanelements.WFIFOState
//	 */
//	public OperatorState getRightState() {
//		return states[1];
//	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#getStateSize()
	 */
	public int getStateSize() {
		int totalsize = 0;
		
		for(int i=0; i<states.length; i++)
			totalsize += states[i].getStateSize();
		
		return totalsize;
	}
	
	/**
	 * Insert the method's description here.
	 * Creation date: (3/15/2003 10:23:33 PM)
	 * @return int
	 */
	public long getTimeWindow() {
		return timeWindow;
	}
	
	public String getRunningMode() {
		return this.runningMode;
	}
	
	public boolean getUpdateSelectivityStatus() {
		return this.updateSelectivityStatus;
	}
	
	public double[][] getSelectivityMatrix() {
		return this.selectivityMatrix;
	}
	
	public double[] getArrivalRates() {
		return this.arrivalRates;
	}

//	public void init() {
//		Iterator i = this.getInputQueues();
//		String schemaName = new String();
//		XATQueueSchema s;
//		XATQueueSchema outSchema = new XATQueueSchema();
//
//		while (i.hasNext()) {
//
//			s = ((XATQueue) i.next()).getSchema();
//			schemaName += "#" + s.getName();
//			
//			Iterator j = s.getColNames().iterator();
//			while(j.hasNext()){
//				outSchema.addColumn((String) j.next());
//			}
//		}
//
//		outSchema.setName(schemaName);
//		this.getOutputQueue().setSchema(outSchema);
//		
//		oQueue = this.getOutputQueue();
//		lQueue = this.getLeftInputQueue();
//		rQueue = this.getRightInputQueue();
//
//	}
/**
 * Insert the method's description here.
 * Creation date: (6/10/2003 2:26:12 PM)
 */

public XATTuple join(XATTuple lTuple, XATTuple rTuple) {

    /** join data together */
    //ArrayList data = (ArrayList) Arrays.asList(lTuple.getCurrentTuple());
    //data.addAll(Arrays.asList(rTuple.getCurrentTuple()));
    XATDataValue[] data1 = lTuple.getArrayData();
    XATDataValue[] data2 = rTuple.getArrayData();
    int size1 = data1.length;
    int size2 = data2.length;

    // we are not worried about schema here, just the join, so just make the new tuple.
    ArrayList newData = new ArrayList();
    for (int i = 0; i < size1; i++)
        newData.add(data1[i]);
    for (int i = 0; i < size2; i++)
        newData.add(data2[i]);

    XATTuple newTuple =
        new XATTupleImp( (XATDataValue[]) (newData.toArray(new XATDataValue[] {})) );

	newTuple.combineTimeStamps(lTuple, rTuple);
//	lTuple = null;
//	rTuple = null;

    return newTuple;
}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/3/2003 10:16:40 PM)
	 * @return boolean
	 */
	public boolean joinable(XATTuple newtuple, XATTuple statetuple) {
	
		/** If this is the last operator and if both tuples' are NEW, don't join */
		/*if (isOldBoxRoot)
		    if (tuple1.isNew() && tuple2.isNew())
		        return false; */
	
		/** if tuple1 and tuple2 is out of window size from each other, don't join */
		try {
			//if (tuple1 == null || tuple1.getTimeStamp() == null || tuple2 == null || tuple2.getTimeStamp() == null)
			//	return false;
	
			if (newtuple.getMaxTimeStamp() - statetuple.getMinTimeStamp() > timeWindow)
				return false;
			
			if (statetuple.getMaxTimeStamp() - newtuple.getMinTimeStamp() > timeWindow)
				return false;
		} catch (Exception e) {
			System.out.println(e);
			return false;
		}
	
		return true;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() {
		
		//if output queue schema is not initialized yet, try to initialize it
		if(!this.hasInit) {
			//System.out.println("Initialize operator " + this);
			init();
		}
		
		//after the first try of initialization, if the output schema is still not initialized,
		//maybe due to one of the input schema not being initialized, then return right away.
		if(!this.hasInit) {
			//System.out.println("Operator cannot be intialized.");
			return false;
		}
		
		if(this.totalTuplesDequeued == null) {
			this.totalTuplesDequeued = new int[this.numOfInputs];
		}
		
		System.out.println("\n\nRun operator " + this);
		int[] tuplesToDequeue = new int[this.numOfInputs];
		long overall_min = Long.MAX_VALUE; // the overall min timestamp of the first tuple of input queues
		long min_time;
		boolean[] visited = new boolean[numOfInputs];
		
		for(int j=0; j<this.numOfInputs; j++) {
			visited[j] = false;
			tuplesToDequeue[j] = this.getInputQueue(j).getQueueSize(this);
			//totalTuplesDequeued[j] += tuplesToDequeue[j];
		}
		
		XATTuple[] output;
		double selectivity=0;
		//Random currentrand;
		int nextprobedstream;
		boolean updated = false;
		int howmany = 0;
		
		for(int index=0; index<numOfInputs; index++){
			if(tuplesToDequeue[index] > 0){
				visited[index] = true;
				//System.out.println("\n\nOperator " + this);
				//System.out.println("Input Queue " + index);
				//System.out.println("# of tuples to dequeue: " + tuplesToDequeue[index]);
		
				howmany = (tuplesToDequeue[index] <= (arrivalRates[index]/4)) ? tuplesToDequeue[index] : (int)(arrivalRates[index]/4);
				XATTuple[] tuples = this.getInputQueue(index).dequeue(howmany,this);
				totalTuplesDequeued[index] += howmany;
				nextprobedstream = this.probeSequence[index][0];
				//System.out.println("arrival rates = " + arrivalRates[index]);
				System.out.println("Dequeue " + howmany + " tuples from queue " + index);
				//System.out.println("Next probed state is " + nextprobedstream + " size = " + ((WOperatorState)states[nextprobedstream]).getStateSize());
				selectivity = probeSelectivityMatrix[index][0];
				//currentrand = this.rand[index][0];
				//System.out.println("selectivity is " + selectivity);
				
				// first do self purge on associated state
				//((WOperatorState)states[index]).purgeByTuple(tuples[tuples.length-1]);
				
				//now figure out the overall min timestamp of the first tuple in all other queues 
				//except the queues that has been visited (including the current queue) and the nextprobedstream.
				for(int i=0; i<this.numOfInputs; i++) {
					if(!visited[i] && (i != nextprobedstream)) {
						if (tuplesToDequeue[i] > 0) {
							min_time = this.getInputQueue(i).peek(this).getMinTimeStamp();
							if(min_time < overall_min) 
								overall_min = min_time;
						}
						else {
							overall_min = 0;
							break;
						}
					}
				}
				output = execOperator(tuples,(WOperatorState)states[nextprobedstream], selectivity, overall_min);
				
				// Now add the tuples from the left queue to the left state
				((WOperatorState)states[index]).insertTuple(tuples);
				tuples = null;
				
				//if(output != null)
				//	System.out.println("# of output tuples " + output.length);
				
				// start join sequence for this input stream
				for(int j=1; j<this.probeSequence[index].length; j++) {
					nextprobedstream = this.probeSequence[index][j];
					//System.out.println("Next probed state is " + nextprobedstream + " size = " + ((WOperatorState)states[nextprobedstream]).getStateSize());
					selectivity = probeSelectivityMatrix[index][j];
					//currentrand = this.rand[index][j];
					//System.out.println("selectivity is " + selectivity);
					
					//now figure out the overall min timestamp of the first tuple in all other queues 
					//except the queues that has been visited (including the current queue) and the nextprobedstream.
					for(int i=0; i<this.numOfInputs; i++) {
						if(!visited[i] && (i != nextprobedstream)) {
							if (tuplesToDequeue[i] > 0) {
								min_time = this.getInputQueue(i).peek(this).getMinTimeStamp();
								if(min_time < overall_min) 
									overall_min = min_time;
							}
							else {
								overall_min = 0;
								break;
							}
						}
					}
					
					if(output != null)
						output = execOperator(output,(WOperatorState)states[nextprobedstream], selectivity, overall_min);
				}
				
				//System.out.println("tuples inserted to state = " + tuples.length);
				//System.out.println("state size after insert = " + ((WOperatorState)states[index]).getStateSize());
				//tuples = null;
				
				//enqueue output tuples to output queue
				if (output != null && output.length > 0) {
					
					if(this.isRoot) {
						//System.out.println("This is root operator, update stats");
						this.stats_total_output_tuples += output.length;
						//System.out.println((System.currentTimeMillis() - this.stats_start_time) + " # of total output tuples = " + this.stats_total_output_tuples);
						
						//for (int k = 0; k < output.length; k++) {
						//	this.stats_output_tuple_count++;
						//	this.stats_output_tuples.append(stats_output_tuple_count + "\t" + output[k]+ "\n");
						//}
						
					} 
					else {
						//System.out.println("This is non-root operator, tuples outputted " + output.length);
						this.getOutputQueue().enqueue(output, this);
					}
						
					output = null;
					updated = true;
					//if(output != null)
					//	System.out.println("# of output tuples " + output.length);
					//for (int k = 0; k < output.length; k++) {
					//	System.out.println("output tuple " + k + " " + output[k].toString());
					//}
				}
			}
		}
		
		visited = null;
		tuplesToDequeue = null;
		
		if(this.isRoot && updated) {
			System.out.println((System.currentTimeMillis() - this.stats_start_time) + " # of total output tuples = " + this.stats_total_output_tuples);
			stats_string.append((System.currentTimeMillis() - this.stats_start_time)
								+ " \t " + this.stats_total_output_tuples + "\n");
		}
		
		/* run successfully */
		return true;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/6/2003 7:46:53 AM)
	 */
	public void setAsBoxLeaf() {
		isOldBoxLeaf = true;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/6/2003 7:46:04 AM)
	 */
	public void setAsBoxRoot() {
		isOldBoxRoot = true;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/6/2003 8:29:45 AM)
	 */
	public void setAsOldOperator() {
		isOldOperator = true;
	}
/**
 * The Dummy operator has 2 parameters.  The SELECTIVITY
 * is a value between 0 and 1 and RESULT_DELAY is the
 * time it takes (in milleseconds) for this operator to place a tuple in its output queue
 * 
 */
protected void setDefaultProperties() {
	super.setDefaultProperties();
	addProperty("RESULT_DELAY","100");
	addProperty("REAL_SELECTIVITY",".5");	
	addProperty("STATE_TYPE", "edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WFIFOState");	
	}

	public void setRunningMode(String mode){
		this.runningMode = mode;
	}

	public void setAsRootOperator() {
		this.isRoot = true;
		
		//initialize output stats string
		this.stats_string = new StringBuffer();
		stats_string.append("time \t output_tuples \n");
		stats_start_time = System.currentTimeMillis();
		
		this.stats_output_tuples = new StringBuffer();
		stats_output_tuples.append("count \t output_tuple \n");
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (3/15/2003 10:24:23 PM)
	 * @param newWindow int
	 */
	public void setTimeWindow(long newWindow) {
	
		//join operator has the same timeWindow with both its states
		timeWindow = newWindow;
		//((WOperatorState)states[0]).setTimeWindow(newWindow);
		//((WOperatorState)states[1]).setTimeWindow(newWindow);
		//System.out.println("The Time window has been set to: " + newWindow);
	}
	
	public void setCartesianProductStatus(boolean s) {
		this.cartesionProductStatus = s;
	}
	
	public void setUpdateSelectivityStatus(boolean s) {
		this.updateSelectivityStatus = s;
	}
	
	public void setNumOfInputs(int n) {
		this.numOfInputs = n;
	}
	
	public void setOneSelectivity(int i, int j, double s) {
		if(this.selectivityMatrix == null) {
			this.selectivityMatrix = new double[this.numOfInputs][this.numOfInputs];
		}
		
		this.selectivityMatrix[i][j] = s;
		this.selectivityMatrix[j][i] = s;
	}
	
	public void setSelectivityMatrix(double[][] s) {
		this.selectivityMatrix = s;
	}
	
	public void setArrivalRates(double[] a) {
		this.arrivalRates = a;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		String rs = new String();

		String sName = this.getClass().getName();
		//find the last dot.
		int index = 0;
		for(index = sName.length()-1; index >=0; index--)
			if(sName.charAt(index) == '.')
				break;
		rs = sName.substring(index + 1) + " ";
		String id = this.getProperty("OBJECT_ID");
		if(id == "-1"){
			rs = rs + hashCode(); 
		} else{
			rs = rs + id;
		}	
		if(this.isActive()) rs += " ACTIVE ";
		else rs+= " INACTIVE ";

		/** now add the join predicate and window size info */
		rs += "\n";
		rs += "window size = "+ this.getTimeWindow()+"\n";
		rs += "state size = " + this.getStateSize()+"\n";

		return rs;
	}
	
	public boolean finalizeOperation() {
		this.printStatsToFile();
		
		System.out.println("\n\nTotal numbers of tuples dequeued from each queue:");
		for(int i=0; i<this.numOfInputs; i++) {
			System.out.println("queue " + i + " = " + this.totalTuplesDequeued[i]);
		}
		
		return true;
	}

	/**
	 * @param tuples
	 * @param state
	 * @param selectivity
	 * @return list of joined tuple
	 */
	private ArrayList execOperator(ArrayList tuples, WOperatorState state, double selectivity, long overall_min){
		
		if(state.getStateSize() == 0){
			return null;
		}
		ArrayList joinedTuples = new ArrayList();
		
		XATTuple t1;
		XATTuple t2;
		XATTuple joinTuple;
		
		ArrayList stateTuples = state.getStateTuples();
		//System.out.println("number of state tuples = " + stateTuples.size());
		
		for(int i = 0; i < tuples.size(); i++){
			t1 = (XATTuple)tuples.get(i);
			state.purgeByTuple(t1, overall_min);   // Purge the tuples we dont need
			int count = 0;
			int joincount = 0;
			
			for(int j = 0; j < stateTuples.size(); j++){
				t2 = (XATTuple)stateTuples.get(j);
				if(joinable(t1,t2)){
					count++;
				    if (rand.nextDouble() < selectivity){
						joinTuple = join(t1,t2);
						this.getOutputQueue().enqueue(joinTuple,this);
						joinedTuples.add(joinTuple);
						joincount++;
						//System.out.println("join with a state tuple");
					}		
				} else 
					System.out.println("Not joinable");
				
			}
			//System.out.println("number of joinable tuples in state = " + count);
			//System.out.println("number of joined state tuples = " + joincount);
		}
		
		//System.out.println("total joined tuples in this probe = " + joinedTuples.size());
		//XATTuple[] outputTuples = (XATTuple[])joinedTuples.toArray(new XATTuple[] {});
		//joinedTuples = null;
		//return outputTuples;
		tuples = null;
		return joinedTuples;
	}
}
