/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Feb 20, 2005
 *
 * The operator that's inside a super WMJoin operator (ProbabilisticWMJoinOeprator). 
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.migration.probablisticjoinoperator;

//import java.util.Vector;

//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import java.io.Serializable;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState;

/**
 * @author Yali Zhu
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class SubWMJoinOperator implements Serializable{

	private ProbabilisticWMJoinOperator superOperator;
	
	private XATQueue inputQ[];
	
	private XATQueue outputQ;
	
	private SubWMJoinOperator children[] = null;
	
	private SubWMJoinOperator parent = null;
	
	private java.util.Random rand;
	
	long timeWindow;
	
	private WOperatorState states[];
	
	//how many input stream this mjoin operates on. for binary join, this is set to be 2.
	int numOfInputs = 0;  
	
	// for an m-join, the probeSequence is an m x (m-1) 2d array.
	// if the super operator is run in the operator mode, 
	// the probeSequence and selectivity matrix are both equal to the ones in the super operator.
	private int[][] probeSequence;
	
	// selectivities[i][j] stores the selectivity between input i and j.
	// selectivities[i][j] = selectivities[j][i]
	// selectivities[i][j] MAY NOT equal to selectivityMatrix[i][j], since the index of these two matrics have different meanings.
	// the index used in selectivityMatrix is in regards to the super operator.
	// the index used here in selectivities is in regards to this sub operator.
	private double[][] selectivities; 
	
	public SubWMJoinOperator(ProbabilisticWMJoinOperator superOp, int nway) {
		this.superOperator = superOp;
		this.numOfInputs = nway;
		this.rand = new java.util.Random();
		
		String statetype = superOp.getProperty("STATE_TYPE");
		try{			
			// set one state for each input
			states = new WOperatorState[numOfInputs];
			for(int i=0; i<numOfInputs; i++) {
				states[i] = (WOperatorState)Class.forName(statetype).newInstance();
				states[i].setOperator(this.superOperator);
			}
		}catch(Exception e) {
			e.printStackTrace();	
		}
	}
	
	public void run() {
		// get all tuples from each inputQ (vector), clear the vectors at the end of the run.
		// join follow the pre-computed join ordering
		// put result to output queue.
	}
	
	private boolean evalJoinPredicate(double selectivity) {
		if(selectivity == -1){
			System.out.println("Something is wrong. Trying to use a selectivity that is not initialized");
			return false;
		}
		
		//return true if the next double is less than the selectivity,
		//otherwise return false
		if (rand.nextDouble() < selectivity){
			return true;
		}
		else{
			return false;
		}
	}
	
	public int getNumberofTuplesInState() {
		int total = 0;
		for(int i=0; i<this.numOfInputs; i++) {
			total += states[i].getStateSize();
		}
		return total;
	}

	/**
	 * compute the new probe sequence based on the selectivity and input rates
	 * this method is called in the classvariable parser if the running mode is set to "mjoin".
	 * we use the KBZ algorithm to compute the optimal sequence for acyclic join graph
	 * for cyclic join graph, the 
	 */
	public void computeSubProbeSequence() {
		
	}
	
	public void setSubProbeSequence(int[][] probe) {
		this.probeSequence = probe;
	}
	
	public void setNumOfInputs(int n) {
		this.numOfInputs = n;
	}
	
	public void setInputQs(XATQueue[] inputQs) {
		this.inputQ = inputQs;
	}
	
	public void setOutputQ(XATQueue outputQ) {
		this.outputQ = outputQ;
	}
	
	public void setChildren(SubWMJoinOperator operators[]) {
		this.children = operators;
	}
	
	public SubWMJoinOperator[] getChildren() {
		return this.children;
	}
	
	public void setParent(SubWMJoinOperator operator) {
		this.parent = operator;
	}
	
	public void setSelectivities(double[][] s) {
		this.selectivities = s;
	}
	
	public void setTimeWindow(long w) {
		this.timeWindow = w;
	}
	
	public void initStates() {
		String statetype = superOperator.getProperty("STATE_TYPE");
		try{			
			// set one state for each input
			states = new WOperatorState[numOfInputs];
			for(int i=0; i<numOfInputs; i++) {
				states[i] = (WOperatorState)Class.forName(statetype).newInstance();
				states[i].setOperator(superOperator);
				states[i].setTimeWindow(this.timeWindow);
			}
		}catch(Exception e) {
			e.printStackTrace();	
		}
	}
}
