/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Oct 22, 2004
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.migration.realjoinoperator;

//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.StreamSplitOperatorImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.StreamSplitOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.SymmMwayHashJoinOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WEquiJoinStateImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WEquiJoinStateImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
//import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.DistributedExecutionController;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Enumeration;
//import java.util.Random;
//import java.io.*;

/**
 * @author Yali Zhu
 * Symmetric Mway Hash Join Operator 
 */
public class WTransitiveHashJoinOperator extends SymmMwayHashJoinOperatorImp {

	private int[] joinAttributePositions; // record the join attribute index (join condition) for each input queue
	private int[] joinInputQueueCells;	// the number of cells in each input queue
	private ArrayList joinExpressions;	//	store the join definitions parsed from xml file
	private int anotherEOFReceivedCounter = 0;
	private boolean fastpurgeDuringPTRedistribution = false;
	private int fastpurgeCountInterval = 10; //indicate the number of tuples passed before next tuple is used for fast purge.
	
	//following are added for parallel track load redistributing
	//private int EOFReceivedForPT = 0; //used to count EOF during parallel track load redistributing. 
	boolean isTopJoin = false;  //used by PT load redistribution to identify top most join.
	ArrayList partitionsToExpire = null;
	ArrayList partitionsToExpireCopy = null;
	int numberOfEOPartReceived = 0;
	boolean startCheckingEOPartitions;
	boolean isOnlyJoinInPlan = false;
	//boolean startCheckingNewTuples = false;
	//boolean receivingNewTuples = false;
	//int numOfChildJoin = 0;
	//int expirationReceived = 0;
	int printQueueInterval = 0;
	
	private long totalOutputTuples = 0;
	private long totalTupleDequeued = 0;
	private boolean collectstats = false;
	private boolean initialized = false;	// to indicate whether the initialization has been done, do not have to do it multiple times.
	private boolean completedFinalization = false; // if the finalization has been completed, then no need of re-do
	private boolean readyToClean = false; //for query tree with multiple join operators, the start of the cleanup stage has to wait the previous one has finished
	private int numberOfInstances = 0; // to indicate how many instances of this join (the number of machines that this join are deployed), 
	//it is set in the localadapter initialization time, based on the output queue numbers of previous_split_operator, it is used to in the finalization stage,
	//to compare the number with numberOfEOFReceived to inidcate the readyToClean
	
	//private int numberofpartitions;
	
	Hashtable hashJoinStates;	// hold the join states of the operator
	Hashtable statesPushedToDisk; // to indicate which states (ID) have been pushed into disk
	
	// for partition adaptations, protocol Step 3 -- DeactivatePartitions.
	ArrayList deactivatedParts = null; //contains partition IDs that will be deactivated
	// boolean noMoreDeactivatedPars = false; // indicate whether have partitions with IDs in deactiavtedParts
	boolean startChecking = false;
	int numberOfEOFReceived = 0; // to indicate whether we have received all EOFs, if it equals to the input queue size
	//int totalDiskTuples = 0;
	// Seems sometimes the tuples with partitionID == -1 get lost, thus the distributed adaptation will never go on,
	// to avoid this situation, the following sudoTimeOutCount is used to set the maximal times to check the ID==-1 tuples,
	// a enough long checking implies that all on-the-fly tuples have been received
	int sudoTimeOutCount = 0;
	
	//protocols to synchronize with the disk-adaptation and distributed adaptation
	//boolean noAdaptation = false;
	//boolean inAdaptation = false;
	
	//int inCheckingDiskAdaptCounter = 0;

	// this is used when a simulated memory usage cost model being used. To indicate how many tuples in
	// the all queues (input and output, might be some duplications since output of one operator could be the input 
	// of another operator) of all activte operators in the machine. it is set by the distributionexecutioncontroller
	int totalMachineQueueSize = 0;
	//double simThreshold = 10.0; // the disk adaptation threshold of simulated model
		
	//Apr-10-2005, to handle window-based join 
	long window_size = 0; //the size of the window of the join operator
	long invalidate_interval = 0; //the time interval that the join operator clean up the out of window states
	// note that for joins that no disk-based adaptation, this could be done each time before the probe. however, 
	// in case partital states have been pushed into disk, this could be inefficient		
	long last_time_invalidated = 0; // initially, it equals to the initialization time
	
	long totalSize; // to record the total state size of the operator, it is updated in the belowAverageMemoryUsage and used in getCurrentStatus
	String baseStateDirName = "states/diskPartitions_";
	String clientName = null; // only for the wpi-cluster that all machines in the cluster have the same virtual disk space?

	//double memoryThreshold = 0.8; // indicate the memory threshold that trigger the memory to disk dumping
	//int timesCalledForDiskDump = 0; // to control the time of merging disk partitions
	//int prePushPartitionSize=0; // to control whether we really need dump
	
//	public double getBaseUnitSize(){
//		return base_unit_size;
//	}

	// it is called in the localAdapter in the 
	public boolean isFinalizationComplete(){
		return completedFinalization;
	}
	// it is called in the localAdapter initialization time
	public void setNumberOfInstances(int n){
		numberOfInstances = n;
	}
	
//	public void setNumberOfPartitions(int n) {
//	    this.numberofpartitions = n;
//	}
	// it is called in the ClassVariableVisitor
	public void turnonStatisticsCollecter(){
		collectstats = true;
	}
//	//it is called in the localMultiJoinAdapter
//	public void setPreviousJoinAndSplit(WTransitiveHashJoinOperator j_ref, StreamSplitOperatorImp s_ref){
//		previous_join_reference = j_ref;
//		previous_split_reference = s_ref;
//	}
//	public WTransitiveHashJoinOperator getPreviousJoinReference(){
//		return previous_join_reference;
//	}
//	public StreamSplitOperatorImp getPreviousSplitReference(){
//		return previous_split_reference;
//	}
//	//it is called in the DistributedExecutionController::initExecution
//	public void setDistributedExectionControllerRef(DistributedExecutionController ref){
//		exec_controller_ref = ref;
//	}
	//for using by updatePreviousStats  
	public int[] getJoinAttributePosition(){
		return joinAttributePositions;
	}
	
	public void setWindowSize(long ws){
		window_size = ws;
	}
	public void setInvalidateInterval(long inInt){
		invalidate_interval = inInt;
	}
	
	public void setTotalMachineQueueSize(int qs){
		totalMachineQueueSize = qs;
	}
	
//	public boolean isInDiskAdaptation(){
//		return inAdaptation;
//	}
//	public void setInAdaptation(boolean val){
//		inAdaptation = val;
//	}	
//	public void stopDiskAdaptation(){
//		noAdaptation = true;
//	}
//	public void resumeDiskAdaptation(){
//		noAdaptation = false;
//	}
	public Hashtable getPartitionIDsPushedToDisk(){
		return statesPushedToDisk;
	}
	
	public void setClientName(String name){
		clientName = name;
	}
	
	// to start checking, set check == true, to stop checking, check == false, parts = null;
	public void setStartChecking(boolean check){
		startChecking = check;
		if(startChecking) {
		    anotherEOFReceivedCounter = 0;
		    numberOfEOFReceived = 0;
		}
	}
	
	public void setPartitionsToMove(ArrayList toMove){
		deactivatedParts = toMove;
	}
	
	public boolean getDeactivatedCompleteFlag(){
		if (numberOfEOFReceived >= this.getInputQueueCount()){
			//reset the numberOfEOFReceived,
		    //System.out.println("Now have " + numberOfEOFReceived + " deactivated inputs in " + this);
			numberOfEOFReceived = 0;			
			return true;
		}
		//System.out.println("Now have " + numberOfEOFReceived + " deactivated inputs in " + this);
		return false;
	}
	// get the partitions indicated by partsToGet from hashJoinStates, them remove the partitions moved from hashJoinStates
	// make sure the states been moved are all in memory, no disk-residented part are involved. I think this should be controlled
	// in diciding which partitions to be moved
	public ArrayList getPartitionsByID(ArrayList partsToGet){
	    //System.out.println("Execute getPartitionsByID in WTransitiveHashJoin");
		ArrayList ret = new ArrayList(partsToGet.size());
		for (int i=0; i<partsToGet.size(); i++){
			Integer idx = (Integer)partsToGet.get(i); 
			HashStatePartitionGroup hp = (HashStatePartitionGroup)hashJoinStates.get(idx);
			ret.add(hp);
			hashJoinStates.remove(idx);
		}
		return ret;
	}
	
	public int getStateSize() {
	    Enumeration e = this.hashJoinStates.elements();
	    int size = 0;
	    
	    while(e.hasMoreElements()) {
	        HashStatePartitionGroup hsg = (HashStatePartitionGroup)e.nextElement();
	        size += hsg.getTotalNumOfTuples();
	    }
	    return size;
	}
	
	public boolean setPartitionsByID(ArrayList ids, ArrayList parts) {
		for (int i=0; i<ids.size(); i++){
			Integer idx = (Integer)ids.get(i);
			HashStatePartitionGroup hp = (HashStatePartitionGroup)parts.get(i);
			hashJoinStates.put(idx,hp);
		}
		return true;
	}
	
	//set WEquiJoinStates by IDs and position.
	public boolean setStatesByID(ArrayList ids, int position, ArrayList parts) {
	    //System.out.println("\nSet new states: " + ids + " at position: " + position + " in operator: " + this);
		for (int i=0; i<ids.size(); i++){
			Integer idx = (Integer)ids.get(i);
			HashStatePartitionGroup hp = (HashStatePartitionGroup)hashJoinStates.get(idx);
			if(hp == null) {
			    //System.out.println("Create new partition group with ID: " + idx);
			    hp = new HashStatePartitionGroup(joinAttributePositions,joinInputQueueCells);
			}
			WEquiJoinStateImp newstate = (WEquiJoinStateImp)parts.get(i);
			int partitionsize = hp.getHashStates()[position].getStateSize();
			int groupsize = hp.getTotalNumOfTuples(); 
			int newgroupsize = groupsize - partitionsize + newstate.getStateSize();
			hp.setStatePartition(position, newstate);
			hp.setTotalNumOfTuples(newgroupsize);
			hashJoinStates.put(idx,hp);
			
			//System.out.println("Add state partition with ID: " + idx + " pos: " + position + " in " + this);	
			//System.out.println("old state size: " + partitionsize + " new state size: " + newstate.getStateSize() + " group size before: " + groupsize + " group size after: " + newgroupsize);
			
		}
		return true;
	}
	
	public Hashtable getHashJoinStates(){
		return hashJoinStates;
	}
	public void addJoinExpression(ExpressionStr exp){
		joinExpressions.add(exp);
	}
	
	public WTransitiveHashJoinOperator(XATNode node){
		super(node);
		hashJoinStates = new Hashtable();
		statesPushedToDisk = new Hashtable();
		joinExpressions = new ArrayList();
		// need some mechanisms to get compute the joinAttributePosistion & joinInputQueueCells
		addStatistic("JOIN_PARTITION_NUMBER");
		addStatistic("JOIN_PARTITION_SIZE");
	}
	
	public boolean run() throws Exception {
	    this.debugFlag = false;
	    int fastpurgecount = 0;
        long starttime = System.currentTimeMillis();
	    
	    //System.out.println("run " + this);
		// build input and output schemas
		for (int i=0; i<this.getInputQueueCount(); i++) {
			if (this.getInputQueue(i).getSchema() == null){
			    //System.out.println("Input queue: " + i + " is not initialized.");
				return false;
			}
		}
		
		if(this.getOutputQueue().getSchema() == null){
            System.out.println("init operator: " + this);
            
			init();
		}
		// build the joinAttributePosition and joinInputQueueCells
		if (!initialized) {			
			
			int queueSizes = this.getInputQueueCount();
			joinAttributePositions = new int[queueSizes];
			joinInputQueueCells = new int[queueSizes];			
			for (int i=0; i<queueSizes; i++){
				// note that length is based on starting point 1, i.e., 5, means 0 ~ 4 have contents
				joinInputQueueCells[i] = this.getInputQueue(i).getSchema().getLength();
			}
			// extract information from joinExpressions
			for (int i=0; i<joinExpressions.size(); i++){
				ExpressionStr exp = (ExpressionStr)joinExpressions.get(i);
				BinCOMPExpressionStr nExp = (BinCOMPExpressionStr)exp;
				TerminalExpressionStr lxp = (TerminalExpressionStr)nExp.getLeft();
				TerminalExpressionStr rxp = (TerminalExpressionStr)nExp.getRight();
				joinAttributePositions[lxp.getIQueuePosition()] = lxp.getIColumnPosition();
				joinAttributePositions[rxp.getIQueuePosition()] = rxp.getIColumnPosition();
			}
			// get the current running directory and compose the baseDirectory
			//File curDir = new File (".");
			//baseStateDirName = curDir.getCanonicalPath() + "\\" + baseStateDirName;
			if (clientName !=null) {				
				baseStateDirName += clientName + "_" + this.operatorID + "_";
			}
			System.out.println("BASE DIRECTORY:" + baseStateDirName);
			initialized = true;
			//System.out.println("INITIALIZATION:+++++++++++++++++++++++++++++++++++++++++"+"QUEUE OWNER:" 
			//		+ this.getOutputQueue().getQueueOwner().toString() + this.getOutputQueue().getQueueName()  + ":" + this.getOutputQueue().getQueueNumber());
			//kind of hard-coded, to unregister unnecessary customers of the output queue.
			//if (this.getOutputQueue().getQueueCustomer().size()>1) {
				//System.out.println("YES, UNREGISTER ONE CUSTOMER." + this.getNode().getParent().getOperator() + "BEFORE:" + this.getOutputQueue().getQueueCustomer().size());					
				//this.getOutputQueue().unregister(this.getNode().getParent().getOperator());
				//System.out.println("AFTER:" + this.getOutputQueue().getQueueCustomer().size());
			//}
			//last_time_invalidated = System.currentTimeMillis();
			//random_var = new Random(); // for deciding when to updating the cost statistics
			//to compute the unit tuple size
			//unit_tuple_size = this.getOutputQueue().getSchema().getLength() * 0.1; //oldversion
			//unit_tuple_size = this.getOutputQueue().getSchema().getLength()/6; //to make sure the base (3-way-join) is the 1 unit size
			//output_unit_size = this.getOutputQueue().getSchema().getLength();
			//intermediate_unit_rate = output_unit_size/base_unit_size; // you can set the intermediate_unit_rate ==1.0 to focus on the number of tuples only
		}
		
//		if(this.printQueueInterval >= 50) {
//		    this.printQueueInterval = 0;
//		    System.out.println("\nMax timestamp in queue: " + this);
//		    for(int i=0; i<this.getInputQueueCount(); i++) {
//		        System.out.print("queue number: " + i + " ");
//		        int q = this.getInputQueue(i).getQueueSize(this);
//		        if(q>0)
//		            System.out.println(this.getInputQueue(i).peekNthTuple(q-1, this).getMaxTimeStamp());
//		        System.out.println();
//		    }
//		} else {
//		    this.printQueueInterval ++;
//		}
		
		XATTuple newObj = null; // to hold the to_be_processed_tuple
		HashStatePartitionGroup newPartitionGroup = null; // to the partitions that will hold the tuple newObj
		XATTuple[] matchedTuples = null;	// hold the output of the probe
		Integer currentPartitionID = null; // hold the current tuple partition ID
		
		int tupleCount = 0;
	
	    if(this.debugFlag) 
	        System.out.println("\nRun " + this);
	    
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		if(this.debugFlag) 
		    System.out.println("tuples to dequeue: " + howMany);
		 
		if(howMany == 0) {
		    return false;
		}
		
		///debug usage, to check how many tuples in each input queue?
		/*
		String outStr = "";
		for (int dk = 0; dk <this.getInputQueueCount(); dk++){
			outStr += "queue " + dk;
			outStr += ":" + this.getInputQueue(dk).getQueueSize();
		}
		System.out.println(outStr);
		*/
		/// end of debug usage
		/// timeout mechanism for part ID == -1
		if (startChecking || startCheckingEOPartitions) {
			sudoTimeOutCount++;	
		}
		
		/// end of timeout mechanism for part ID == -1
		
		int emptyQueueCount = 0;
		int queueCount = this.getInputQueueCount();
		
		while (true) {
		    
			if (tupleCount >= howMany) {
			    //System.out.println("number of tuples dequeued: " + tupleCount);
			    this.totalTupleDequeued += tupleCount;
			    //System.out.println("throughput: " + totalOutputTuples);
			    //System.out.println("total tuples dequeues: " + totalTupleDequeued);
                //System.out.println("This run takes: " + (System.currentTimeMillis() - starttime));
				return true;
			}
			
			if(emptyQueueCount >= queueCount) {
			    //System.out.println("queues are all empty, return right away.");
			    return false;
			}
			emptyQueueCount = 0;
			
			try {
				//int queueSizes = this.getInputQueueCount();
			    
//				if(this.printQueueInterval > 50) {
//				    this.printQueueInterval = 0;
//				    //System.out.println("\nCurrent tuple from queue: " + i + ": " + newObj);
//				} else {
//				    this.printQueueInterval ++;
//				}
				
				// get one tuple from one input queue at a time and process the join, may not be that efficient? 
				// it could be improved to process batches of tuples at a time from each input queue?
				for (int i=0; i<queueCount; i++){
				    
				    if(this.getInputQueue(i).getQueueSize() <= 0) {
				        emptyQueueCount++;
				    }
//				    int tupleindex = 0;
//				    int queuesize = this.getInputQueue(i).getQueueSize();
//				    XATTuple[] tuplelist = this.getInputQueue(i).dequeue(queuesize, this);
//				    while(tupleindex < queuesize) {
					if (this.getInputQueue(i).getQueueSize()>0){
						
					    //if(this.debugFlag)
						//   System.out.println("QUEUE:" + i + "SIZE:" + this.getInputQueue(i).getQueueSize());
						
						newObj = this.getInputQueue(i).dequeue(this);		
//						if(this.printQueueInterval == 50) {
//						    //this.printQueueInterval = 0;
//						    System.out.println("\nCurrent tuple from queue " + i + ": " + newObj);
//						}
						
						
						//newObj = tuplelist[tupleindex];
						
//			            isTopJoin = false;
//			            this.partitionsToExpire = null;
//			            startCheckingNewTuples = false;
//			            this.numOfChildJoin = 0;
//			            this.expirationReceived = 0;
//			            EOFReceivedForPT = 0;
			            
						// check the partition id
						int partID = newObj.getPartitionID();
						if (partID == -1) {
						    //System.out.println("Receive tuple with partition ID = " + partID + " in " + this);
						    
						    //if moving state partition redistribution
							if (startChecking) {
								//yes, receive one eof tuple from split operator, we do not have to process it
							    //System.out.println("Receive EOF in " + this);
								numberOfEOFReceived ++;							
								sudoTimeOutCount = 0;
								if(numberOfEOFReceived >= this.getInputQueueCount()) {
							        startChecking = false;
							        sudoTimeOutCount = 0;
							        //System.out.println("Received EOF: " + numberOfEOFReceived + " so stop checking");
							        
									//pass on the EOF tuple if its parent is another join operator
								    if(this.getNode().getParent().getStreamOperator() instanceof WTransitiveHashJoinOperator) {
								        this.getOutputQueue().enqueue(newObj, this);
								        //System.out.println("Pass the EOF tuple to its parent.");
								    }								    
								}
							} 
//							//if parallel track partition redistribution
//							else if(startCheckingNewTuples) {
//								EOFReceivedForPT ++;							
//								sudoTimeOutCount = 0;
//								if(EOFReceivedForPT >= this.getInputQueueCount()) {
//								    startCheckingNewTuples = false;
//								    
//								    //from now on all tuples received are new tuples sent to both machines
//								    receivingNewTuples = true;
//								    
//								    //pass the EOF if this is not the top join
//								    if(!this.isTopJoin) {
//								        this.getOutputQueue().enqueue(newObj, this);
//								        System.out.println("Pass the EOF tuple to its parent");
//								    }								    
//								}
//							} 
							else {
								System.out.println("Wrong, receiving tuples in WTransitiveHashJoinOperator without partition ID.");
							}
						} else if(partID == -2) {
						    //indicating the end of parallel track partition redistribution
						    numberOfEOPartReceived++;
						    sudoTimeOutCount = 0;
						    //System.out.println("Receive EOPartition tuple in: " + this + " " + isTopJoin);
						    //System.out.println("EOPartitions have received: " + this.numberOfEOPartReceived);
						    if(numberOfEOPartReceived >= this.getInputQueueCount()) {
						        numberOfEOPartReceived = 0;
						        sudoTimeOutCount = 0;
						        this.startCheckingEOPartitions = false;
						        this.removeExpiredPartitions();
							    //pass on the EOPartition tuple if its parent is another join operator
							    if(this.getNode().getParent().getStreamOperator() instanceof WTransitiveHashJoinOperator) {
							        this.getOutputQueue().enqueue(newObj, this);
							        //System.out.println("Pass the EOPartition tuple to its parent.");
							    }		
						        if(this.isTopJoin) {
						            //this.removeExpiredPartitions();
						            isTopJoin = false;
						        }

						    }						    
						    
						} else if(partID == -10){
							//indicating the end of the streams, used to control the cleanup process of the multi-join operators in the query tree
							numberOfEOFReceived ++;
							//propagate the EOF tuple to all output queues of the operator
							//for(Iterator cursor= this.getOutputQueues(); cursor.hasNext();){
							//	((XATQueue)cursor.next()).enqueue(newObj,this);
							//}
							return true;							
						} else {
							//a timeout mechanism for possible losing tuples with ID == -1
							if (startChecking) {				
								if (sudoTimeOutCount>20){
									numberOfEOFReceived ++;
									sudoTimeOutCount = 0;
									System.out.println("Timeout mechanism activated in " + this + "!");
								}
							}
							//a timeout mechanism for possible losing tuples with ID == -2
							if (startCheckingEOPartitions) {				
								if (sudoTimeOutCount>20){
									numberOfEOPartReceived ++;
									sudoTimeOutCount = 0;
									System.out.println("Timeout mechanism activated in " + this + "!");
								}
							}
							
							//normal join process
							//if(this.debugFlag)
							//    System.out.println("Normal join process.");
							
							currentPartitionID = new Integer(partID);
							newPartitionGroup = (HashStatePartitionGroup)hashJoinStates.get(currentPartitionID);
							if (newPartitionGroup != null){
								// simply probe and insert the tuple
								if (window_size>0){
									// window-based join, use purge and probe function
								    if(partitionsToExpire != null && partitionsToExpire.contains(currentPartitionID)) {
								        matchedTuples = newPartitionGroup.purgeProbe(newObj,i,window_size, true, isTopJoin);
									    //System.out.println("new tuple in position: " + i + " in partition: " + currentPartitionID + " in " + this);
								    }
								    else
								        matchedTuples = newPartitionGroup.purgeProbe(newObj,i,window_size, false, false);
								} else {
									// non-window join
									matchedTuples = newPartitionGroup.probe(newObj,i);	
								}								
								if(matchedTuples != null)
								    ((HashStatePartitionGroup)newPartitionGroup).addTotalTuplesOutputed(matchedTuples.length);
								
								int curJoinAttrPos = joinAttributePositions[i];
								Object value = newObj.getCell(curJoinAttrPos).getValue();
								
								if(partitionsToExpire!= null && this.isOnlyJoinInPlan && newObj.isNew() && partitionsToExpire.contains(currentPartitionID)) {
								  //if in PT redistribution and this is the only join in the query plan,
								  //do not insert new tuples in to-be-expired partitions.
								} 
								else { 
								    newPartitionGroup.insertTuple(value, newObj, i);
								    if(newObj.isNew() && this.partitionsToExpire != null) {
								        //System.out.println("new tuple: " + newObj + " partition ID: " + currentPartitionID);
								        if(partitionsToExpire.contains(currentPartitionID)) {
								            //System.out.println("Before decrease: " + newPartitionGroup.getTotalNumOfTuples() + " in " + this);
								            newPartitionGroup.decreaseTotalTuplesByOne(i);
								            //System.out.println("Tuples in partition: " + currentPartitionID + " after decrease: " + newPartitionGroup.getTotalNumOfTuples() + " in " + this);
								        } 
								    
//								    	if(newObj.isSingletonTuple()) {
//								        	if(fastpurgecount >= 10) {
//								            	System.out.println("Fast purge use this tuple");
//								            	this.fastCrossPartitionGroupPurge(newObj, i);
//								        	} else
//								            	fastpurgecount++;
//								    	}
								    }	
								}
								
								//apply fast purge during parallel track partition redistribution
							    if(this.partitionsToExpireCopy != null) {
							        if(fastpurgecount >= this.fastpurgeCountInterval) {
							            fastpurgecount = 0;
							            //System.out.println("new tuple: " + newObj + " partition ID: " + currentPartitionID);
							            //System.out.println("Fast purge use this tuple in " + this);
							            this.fastCrossPartitionGroupPurge(newObj, i);
							        } else 
							            fastpurgecount++;
							    }

								//not sure whether need to reset the hashtable for the new classes, if it is a pure reference, then no, otherwise, yes
								//hashJoinStates.put(new Integer(partID),newPartition);
							} else {
								// create this partition and put them back into the hashJoinStates
								newPartitionGroup = new HashStatePartitionGroup(joinAttributePositions,joinInputQueueCells);
								newPartitionGroup.setPartitionID(currentPartitionID.intValue());
								int curJoinAttrPos = joinAttributePositions[i];
								Object value = newObj.getCell(curJoinAttrPos).getValue();
								newPartitionGroup.insertTuple(value,newObj,i);
								hashJoinStates.put(new Integer(partID), newPartitionGroup);
							}
							tupleCount ++;	
							// System.out.println("Partition ID:" + partID + "::" + newPartition.getTotalNumOfTuples());
							newPartitionGroup = null;
						}
						newObj = null;
					} //
					// output the result to the output queue

					if (matchedTuples != null){
						// System.out.println("Output:" + matchedTuples.length);
						// we set the default partition ID, in case that no split operator is defined after the join operator
						// This could happen when the join columns of different joins are the same. However, for join columns 
						// of different joins are not the same, that is, split operator is inserted between joins, this is not
						// needed at all, it can be safely commented out for a littler performance improvement. -- Bin Mar, 2005
						for (int dfp=0; dfp<matchedTuples.length; dfp++){
							matchedTuples[dfp].setPartitionID(currentPartitionID.intValue());
						}
						/// the above could be commented out for split operator is always defined after join
						
						this.getOutputQueue().enqueue(matchedTuples, this);
						this.totalOutputTuples += matchedTuples.length;
						matchedTuples = null;
						currentPartitionID = null;
					}
					
				}
				
//				if (window_size>0){
//					cleanInvalidatedStates();
//				}
				//String outStr = "SIZE:" + this.getOutputQueue().getQueueSize() + "JOIN QUTPUT QUEUE";
				//for (int pp = 0; pp<(this.getOutputQueue()).getQueueCustomer().size(); pp++){
				//	outStr += pp + ":" + (this.getOutputQueue()).getQueueCustomer().get(pp);
				//}
				//System.out.println(outStr);
				
				//if (startChecking) {
				//	noMoreDeactivatedPars = true;
				//}

			} catch(XATQueueException e1){
				return false;
			}			
		}
	}
	
	/*
	 * Compute the whole-minus, seems the arraylist.contain does not work well for this case.
	 * basically, remove tuples from whole that exist in minus
	 */
	public ArrayList computeDifference(ArrayList whole, ArrayList minus){
		ArrayList difference = new ArrayList();
		for (int i=0; i<whole.size(); i++){
			XATTuple curTP = (XATTuple)whole.get(i);
			if (!contains(minus,curTP)) {
				difference.add(curTP);
			}
		}
		return difference;
	}
	public boolean contains (ArrayList src, XATTuple tp){
		for (int i=0; i<src.size(); i++){
			XATTuple tmpTP = (XATTuple)src.get(i);
			// need to compare each cell whether it is the same
			if (tmpTP.equals(tp)){
				return true;
			}
		}
		return false;
	}
	
//	/* check the memory usage based on the runtime information, it may not reflect the actual system memory
//	 * available information, but it can be used as an indicator. especially the total amount of main memory one
//	 * JVM can be used is specified by parameter -mx
//	 * the flag simulated = true indicated that we use the simulated cost model instead the real main memory usage
//	 * */
//	public boolean belowAverageMemoryUsage(boolean simulated){
//		if (simulated){
//			// (1) check the state size of the operator
//			Enumeration keys = this.hashJoinStates.keys();			
//			totalSize = 0;
//			while (keys.hasMoreElements()) { 
//				Integer key = (Integer)keys.nextElement();
//				totalSize += ((HashStatePartitionGroup)hashJoinStates.get(key)).getTotalNumOfTuples();
//			}
//			double simulatedCost = (totalSize + totalMachineQueueSize)/1024.0;
//			if (simulatedCost>simThreshold) {
//				System.out.println("SIMULATED COST MODEL: Yes, it is over " + simThreshold);
//				return true;
//			}
//			return false;
//		} else {
//			Runtime runtime = Runtime.getRuntime();
//			long usedmemory = runtime.totalMemory()-runtime.freeMemory();
//			long maxmemory = runtime.maxMemory();
//			// add one short cut here, only half of the maxiaml memory is used in determine the 
//			// memoryusage percentage, this case, thus, to avoid the memory overflow of  the
//			// finalization stage 
//			double memoryusage = ((double)usedmemory*2)/(double)maxmemory;
//			if (memoryusage > memoryThreshold) {
//					System.out.println("REALMEMORY USAGE. Yes, it is over THRESHOLD " + memoryThreshold);
//					return true;
//			}
//			return false;
//		}		
//	} //end of checkMemoryUsage

	public boolean finalizeOperation(){
	    return true;
	}
//		//System.out.println("Yes, Come to the end." + this.operatorID);
//		if (completedFinalization) {
//			//System.out.println("Yes, Completed Finalization Stage of Operator:" + this.operatorID); 
//			return true;
//		} 
//		
//		boolean res = true;
//		if (res) {
//			// End of finalization, then send a EOF tuple with partition ID = -10 
//			XATTuple EOFTP = new XATTupleImp();
//			EOFTP.setPartitionID(-10);
//			this.getOutputQueue().enqueue(EOFTP, this);
//			completedFinalization = true;
//			//notifiy the query processor to modifiy the previous_join_reference
//		}
//		return res;
//	}
//	
//	/* in cleaning up the query tree with multiple join operators, we need to keep on processing
//	 * join with the data generated from the previous join operators, however, if the inputs are from
//	 * source streams or other sources, we simply discard them -- @ Apr 14, 2005
//	 * It is called by the finalizationOperation
//	*/
//	public boolean cleanUpRun(){
//		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
//		int tupleCount = 0;
//		if (numberOfEOFReceived >= numberOfInstances){
//			readyToClean = true;
//			return true;
//		}
//		// a timeout mechanism for possible losing tuples with ID==-1
//		//if (sudoTimeOutCount>20){
//		//	numberOfEOFReceived ++;
//		//	sudoTimeOutCount = 0;
//		//	System.out.println("Timeout mechanism activated for this adaptation!");
//		//}
//		while (true){
//			if (tupleCount>=howMany){
//				return true;				
//			}			
//			int queueSizes = this.getInputQueueCount();
//			XATTuple[] matchedTuples = null;
//			Integer currentPartitionID = null;
//			boolean allEmpty = true;
//			for (int i=0; i<queueSizes; i++){
//				if ((this.getInputQueue(i).isFinalQueue()) && (this.getInputQueue(i).size() >0)) {
//					allEmpty = false;
//					break;
//				}
//			}
//			if (allEmpty){
//				sudoTimeOutCount++;
//				return true;
//			}
//			try {
//				for (int i=0; i<queueSizes; i++){					
//					if (!this.getInputQueue(i).isFinalQueue()) {
//						// simply clean all the tuples in the queue, do not have to process them
//						this.getInputQueue(i).cleanTuples();
//					} else {
//						if (this.getInputQueue(i).getQueueSize()>0){
//							XATTuple newObj = this.getInputQueue(i).dequeue(this);						
//							// check the partition id
//							int partID = newObj.getPartitionID();
//							if (partID == -10) {
//								//yes, receive one eof tuple from split operator, we do not have to process it
//								numberOfEOFReceived ++;
//								sudoTimeOutCount = 0;
//							} else {
//								// normal join process
//								currentPartitionID = new Integer(partID);
//								HashStatePartition newPartition = (HashStatePartition)hashJoinStates.get(currentPartitionID);
//								if (newPartition != null){
//									// simply probe and insert the tuple
//									if (window_size>0){
//										// window-based join, use purge and probe function
//										matchedTuples = newPartition.purgeProbe(newObj,i,window_size);
//									} else {
//										// non-window join
//										matchedTuples = newPartition.probe(newObj,i);	
//									}								
//									int curJoinAttrPos = joinAttributePositions[i];
//									Object value = newObj.getCell(curJoinAttrPos).getValue();
//									newPartition.insertTuple(value,newObj,i);
//									//not sure whether need to reset the hashtable for the new classes, if it is a pure reference, then no, otherwise, yes
//									//hashJoinStates.put(new Integer(partID),newPartition);
//								} else {
//									// create this partition and put them back into the hashJoinStates
//									newPartition = new HashStatePartition(joinAttributePositions,joinInputQueueCells);
//									int curJoinAttrPos = joinAttributePositions[i];
//									Object value = newObj.getCell(curJoinAttrPos).getValue();
//									newPartition.insertTuple(value,newObj,i);
//									hashJoinStates.put(new Integer(partID),newPartition);
//								}
//								tupleCount ++;	
//								newPartition = null;
//							}
//							newObj = null;
//						} //
//						// output the result to the output queue
//						if (matchedTuples != null){
//							// System.out.println("Output:" + matchedTuples.length);
//							// we set the default partition ID, in case that no split operator is defined after the join operator
//							// This could happen when the join columns of different joins are the same. However, for join columns 
//							// of different joins are not the same, that is, split operator is inserted between joins, this is not
//							// needed at all, it can be safely commented out for a littler performance improvement. -- Bin Mar, 2005
//							for (int dfp=0; dfp<matchedTuples.length; dfp++){
//								matchedTuples[dfp].setPartitionID(currentPartitionID.intValue());
//							}
//							/// the above could be commented out for split operator is always defined after join						
//							this.getOutputQueue().enqueue(matchedTuples, this);
//							matchedTuples = null;
//							currentPartitionID = null;
//						}					
//					} 
//				}//end of for loop queueSizes
//			} catch(XATQueueException e1){
//				// no tuples in the queue? or other unexpected errors?
//				return true;
//			}			
//		}
//	}

	

	/* used for VLDB05-Demo in query tree case, the totalsize of the operator is set from localMultiJoinTreeAdapter
	 * for displaying into the GUI
	 */
	public void setTotalSize(long ts) {
		totalSize = ts;
	}
	/* overwrite the code in the XATStreamOperatorImp. For gui to display the operator
	 * information in the Operators tab -- Aug 10, 2005
	 */
	public String getCurrentStatus() {
		String statsString = "OID:" + this.operatorID + "State Size:" + totalSize
							+ "Total Partitions:" + hashJoinStates.size() + "\n";		
		/*
		if (statusBuffer.length() >0) {
			statsString += statusBuffer.toString();
			statusBuffer.delete(0,statusBuffer.length());
		}
		*/	
		return statsString;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAccumulatedTuples(long)
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAllAccumulatedTuples()
	 */
	public boolean cleanAllAccumulatedTuples() {
	    
	    try {
	        int totaltuples = 0;
	        for(int i=0; i<this.getInputQueueCount(); i++) {
	            totaltuples += this.getInputQueue(i).getQueueSize(this);
	        }
	        
	        this.setProperty("NUMBER_OF_TUPLES_TO_DEQUEUE", ""+totaltuples);
	        System.out.println("Clean accumulated tuples in: " + this);
	        System.out.println("Total number of accumulated tuples to clean: " + totaltuples);
	        return this.run();
	    	
	    }catch(Exception e) {
	    	    e.printStackTrace();
	    	    System.exit(-1);
	    }
	    return true;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#matchStates(edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator)
	 */
	public boolean matchStates(XATStreamOperator op) {
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#hasUnmatchedState()
	 */
	public boolean hasUnmatchedState() {
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#recomputeUnmatchedState()
	 */
	public boolean recomputeUnmatchedState() {
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#setStateID(int)
	 */
	public int setStateID(int startID) {
		return 0;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		return null;
	}
    /*
     * if it is window-based join, if the current time - last_invalidated_time > invalidate_interval
     * then, clean up the invalidated states.
     */
    public void cleanInvalidatedStates(){
    	if (System.currentTimeMillis() - last_time_invalidated > invalidate_interval){
    		Enumeration keys = hashJoinStates.keys();
    		while (keys.hasMoreElements()){
    			Integer key = (Integer)keys.nextElement();
    			//System.out.println("Purge Partition ID:" + key);
    			((HashStatePartitionGroup)hashJoinStates.get(key)).purgeState();
    		}			
    		last_time_invalidated = System.currentTimeMillis();
    	}
    }
    
	/**
	 * Recompute partition groups with the given IDs, return results for each partition group. 
	 * The join results from each partition group form a state partition, which is then added 
	 * to the parent operator during load balancing. Used in DistributedExecutionController.receivePartitionsInTree().
	 * @param IDs
	 * @return A list of partitions.
	 */
	public ArrayList joinPartitionsByIDs(ArrayList IDs) {
	    ArrayList newstates = new ArrayList();
	    
	    //join partitions in each partition group with unique partition ID.
	    for(int idindex=0; idindex<IDs.size(); idindex++) {
	        WEquiJoinStateImp newstate = new WEquiJoinStateImp();
	        XATTuple[] matchedTuples;
			Integer currentPartitionID = (Integer)IDs.get(idindex);
			HashStatePartitionGroup partitionGroup = (HashStatePartitionGroup)hashJoinStates.get(currentPartitionID);
			int totaljoinedtuples = 0;
			if (partitionGroup != null) {
			    //System.out.println("Join tuples in partitions group: " + currentPartitionID + " in " + this);
			    //get tuples in the first state of the partition group
			    WEquiJoinStateImp firststate = partitionGroup.getHashStates()[0];
			    int firststatesize = firststate.getStateSize();
			    Enumeration firstelements = firststate.getAllStateTuples().elements();
			    
			    while(firstelements.hasMoreElements()) {
			        ArrayList bucket = (ArrayList)firstelements.nextElement();
			        
			        for (int bindex = 0; bindex < bucket.size(); bindex++) {
                        XATTuple newTuple = (XATTuple) bucket.get(bindex);

                        // simply probe and insert the tuple
                        if (window_size > 0) {
                            // window-based join, use purge and probe function
                            matchedTuples = partitionGroup.purgeProbe(newTuple, 0, window_size, false, false);
                        } else {
                            // non-window join
                            matchedTuples = partitionGroup.probe(newTuple, 0);
                        }

                        if(matchedTuples != null) {
                            for (int rsize = 0; rsize < matchedTuples.length; rsize++) {
                                Object value = newTuple.getCell(0).getValue();
                                newstate.insertTuple(value, matchedTuples[rsize]);
                                totaljoinedtuples++;
                            }
                        }
                        matchedTuples = null;
                    }
			    }
			}
			//System.out.println("Computed new state partition: " + currentPartitionID + " with " + totaljoinedtuples + " tuples");
			newstate.setStateSize(totaljoinedtuples);
			newstates.add(newstate);
	    }
	    return newstates;
	}
	
	// Set a single state partition (not partition group) by ID and state position. 
	// The position of state corresponds to position of input queue.
	public boolean setSinglePartitionByID(int id, int position, WEquiJoinStateImp state) {
		HashStatePartitionGroup group = (HashStatePartitionGroup)hashJoinStates.get(new Integer(id));
		//if adding a new partition group (during load rebalancing)
		if(group == null) {
		    group = new HashStatePartitionGroup(joinAttributePositions,joinInputQueueCells);
		    //System.out.println("\nCreate new partition group with ID: " + id + " in " + this);
		}
		int partitionsize = group.getHashStates()[position].getStateSize();
		int groupsize = group.getTotalNumOfTuples(); 
		int newgroupsize = groupsize - partitionsize + state.getStateSize();
		group.setStatePartition(position, state);
		group.setTotalNumOfTuples(newgroupsize);
		hashJoinStates.put(new Integer(id), group);
		//System.out.println("Add state partition with ID: " + id + " pos: " + position + " in " + this);	
		//System.out.println("old state size: " + partitionsize + " new state size: " + state.getStateSize() + " group size before: " + groupsize + " group size after: " + newgroupsize);
		
		return true;
	}
	
	public void setIsTopJoin(boolean istop) {
	    this.isTopJoin = istop;
	}
	
	public void setPartitionsToExpire(ArrayList parts) {
	    System.out.println("Set partitionsToExpire in " + this);
	    this.partitionsToExpire = parts;
	    this.partitionsToExpireCopy = new ArrayList();
	    for(int i=0; i<parts.size(); i++)
	    this.partitionsToExpireCopy.add(parts.get(i));
	    //System.out.println("partitionsToExpireCopy = " + partitionsToExpireCopy);
	    //System.out.println("partitionsToExpire = " + partitionsToExpire);
	}
	
	public void removeExpiredPartitions() {
	    System.out.println("Remove expired partitions: " + this.partitionsToExpire + " in: " + this);
	    if(this.partitionsToExpire == null) {
	        System.out.println("PartitionsToExpire is null. Something is wrong!");
	        System.exit(-1);
	    }
	    for(int i=0; i<partitionsToExpire.size(); i++) {
	        Integer id = (Integer)partitionsToExpire.get(i);
	        //System.out.println("Remove partitions: " + id + " from " + this);
	        this.hashJoinStates.remove(id);
	    }
	    numberOfEOPartReceived = 0;
	    startCheckingEOPartitions = false;
	    partitionsToExpire = null;
	    partitionsToExpireCopy = null;
	}
	
	/**
	 * Used during parallel track partition redistribution to use a singleton tuples
	 * to purge tuples in to-be-expired partitions. This makes the process faster. 
	 * The singleton tuple tp do not fast-purge the to-be-expired partition is belongs to. 
	 * This is to prevent incorrect self-partition purge. 
	 * @param tp
	 */
	private synchronized void fastCrossPartitionGroupPurge(XATTuple tp, int excludingposition) {
	    int tuplepartid = tp.getPartitionID();
	    
	    if(this.partitionsToExpireCopy != null) {
	        for(int i=0; i<partitionsToExpireCopy.size(); i++) {
	            Integer partid = (Integer)partitionsToExpireCopy.get(i);
	            if(tuplepartid != partid.intValue()) {
	                //System.out.println("fast purge partition group: " + partid + "excludes position: " + excludingposition);
	                HashStatePartitionGroup group = (HashStatePartitionGroup)hashJoinStates.get(partid);
                    //System.out.println("group = " + group);
	                if(group!= null && group.getTotalNumOfTuples() > 0)
	                    group.fastPurge(tp, window_size, true, excludingposition);
	            }
	        }
	    }
	}
	
	/**
	 * Check if all to-be-expired partition groups are empty. 
	 * Only apply to the top join operator.
	 */ 
	public boolean getPTCompletedFlag() {
        if (this.allOldPartitionsExpired()) {
            //reset parameters
            //isTopJoin = false;
            partitionsToExpireCopy = null;
            //this.partitionsToExpire = null;
            //startCheckingNewTuples = false;
            //EOFReceivedForPT = 0;
            //receivingNewTuples = false;
            
            return true;
        }
        return false;
    }
	
	private synchronized boolean allOldPartitionsExpired() {
	    boolean isExpired = true;
	    
	    for(int i=0; i<partitionsToExpireCopy.size(); i++) {
	        HashStatePartitionGroup group = (HashStatePartitionGroup)this.hashJoinStates.get(partitionsToExpireCopy.get(i));
	        if(group == null || group.getTotalNumOfTuples() <= 0) {
	            //System.out.println("\npartition group: " + partitionsToExpireCopy.get(i) + " size: " + group.getTotalNumOfTuples() + " is expired!");
	            //System.out.println("Contents of this partition group:" + group.toString());
	            partitionsToExpireCopy.remove(i);
	            //System.out.println("partitionsToExpireCopy = " + partitionsToExpireCopy);
	            //System.out.println("partitionsToExpire = " + partitionsToExpire);
	            i--;
	        } else {
	            //System.out.println("\npartition group: " + partitionsToExpireCopy.get(i) + " size: " + group.getTotalNumOfTuples() + "- Not expired yet!");
	            //System.out.println("Contents of this partition group:" + group.toString());
	            isExpired = false;
	        }
	    }

	    return isExpired;
	}
	
	public void setStartCheckingEOPartitions(boolean value) {
	    this.startCheckingEOPartitions = value;
	    
	    if (!startCheckingEOPartitions) {
            isTopJoin = false;
            //partitionsToExpire = null;
        	numberOfEOPartReceived = 0;
        	this.partitionsToExpireCopy = null;
        }
	}
	
	public void turnonFastPurge() {
	    this.fastpurgeDuringPTRedistribution = true;
	}
	
	public void setFastPurgeCountInterval(int count) {
	    this.fastpurgeCountInterval = count;
	}
	
	public boolean isTopJoinDuringPT() {
	    return this.isTopJoin;
	}
	
	public void setIsOnlyJoinInPlan(boolean value) {
	    System.out.println("Set as the only join in plan: " + this);
	    this.isOnlyJoinInPlan = value;
	}
}
