package edu.wpi.cs.dsrg.xmldb.xat.common.operator;
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.Iterator;
import java.util.ArrayList;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATNodeDoesNotExistException;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataSource;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATTable;

/**
 * Super class for all the XAT operators. <P>
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0      
 */
public interface XATOperator extends XATQueryObject, Cloneable {

    /**
     * Clones the XATOperator.
     * @since   1.0              
     */
    public Object clone();
    /** 
    * Gets the node of this operator in order to get children or parent nodes.
    * @return XATNode
    * @since   1.0         
    */
    XATNode getNode();

    /**
     * Gets the output of this Operator.
     * @return the output XATTable of this operator.
     * @since   1.0          
     * @deprecated use XATOperator.getOutputDataSource, then cast to a table or a queue
     */
    XATTable getOutput();

    /**
     * Gets output column name.
     * @return output column name
     * @since   1.0          
     */
    String getOutputColumnName();

    /**
     * Get the output of this Operator.
     * @return the output DataSource of this operator.
     * @since   1.0          
     */
    XATDataSource getOutputDataSource();
    /**
     * Get all output data sources for this XATOperator
     * @return Iterator of data sources.
     * @since   1.0          
     */
    Iterator getOutputDataSources();

    /**
     * Get primary data source. The primary data source will be the first data source in all the operators. For example:
     * For all the Join the primary data source is source l. For all the Union operators is the first source.
     * For CO is the source S1.
     * @return XATDataSource of the primary data source.
     * @throws XATNodeDoesNotExistException    
     * @since   1.0           
     */
    XATDataSource getPrimaryDataSource() throws XATNodeDoesNotExistException;

    /**
     * Get primary data source. The primary data source will be the first data source in all the operators. For example:
     * For all the Join the primary data source is source l. For all the Union operators is the first source.
     * For CO is the source S1.
     * @return XATTable of the primary data source.
     * @throws XATNodeDoesNotExistException
     * @since   1.0          
     * @deprecated use XATOperator.getPrimaryDataSource, then cast to a table or a queue
     */
    XATTable getPrimarySource() throws XATNodeDoesNotExistException;

    /**
     * Sets the opertor's node.
     * @param newNode edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
     * @since   1.0          
     */
    void setNode(XATNode newNode);

    /**
     * Sets the ouput data source.
     * @param newoutput output data source
     */
    void setOutput(XATDataSource newoutput);

    /**
     * Sets output columns.
     * @param OutputColumnName java.lang.String
     * @since   1.0          
     */
    void setOutputColumnName(String OutputColumnName);
    
    /**
     * In case of finite data stream case or ..., we have to address how the operator handles before it
     * exits, i.e., read & process the left-tuples in input queue, combine disk residented states and 
     * geneate the results (cleanup stages). Details are operator specific,  
     * @author Bin Liu, Nov.23,2004
     */
    boolean finalizeOperation();

// KWORKS start - Eddy/QMesh

    public int GetStreamID();
    public void SetStreamID(int i);

// KWORKS end - Eddy/QMesh
}