/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Dec 23, 2005
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.clustersheddy;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.MotionStatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.ClassType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.QueryType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.UpdateType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GenericSpatioTemporalGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridCellIndex3D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.MotionGenericOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.MotionDiscreteResult;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MotionDiscreteState;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/**
 * @author Rimma V. Nehme
 *
 */
public class ClusterSheddyJoinOperatorImp extends XATMultiQueueWindowStreamOperatorImp implements ClusterSheddyJoinOperator 
{
	
   	//*************************************************************************	
	// input parameters specified for the operator through the config file
	// these parameters control the behavior of the operator (such as join and load shedding)	
	//Timestamp position (this is a timestamp of the location)
	int left_time_position;
	int right_time_position;	
	int range_query_size;	
   	//Join Properties
   	int delta_join;		
   	//Grid Size
   	int grid_x_dim;
   	int grid_y_dim;
   	int grid_t_dim;
   	//Clustering Properties
   	int threshold_distance_x;
   	int threshold_distance_y;
   	int threshold_speed;
   	int threshold_time;       	
   	boolean is_threshold_destination_enabled;
   	String threshold_destination_type;
   	int threshold_destination_approximate_distance;
   	//Load Shedding Properties
   	boolean is_load_shedding;
   	String load_shedding_which_clusters_policy;
   	String load_shedding_how_much_policy;
   	int delta_load_shed;
   	Coin flipCoin;
   	
   	static int [] objectsEncounteredArray = new int[10000];
   	static int [] queriesEncounteredArray = new int[10000];
   	
   	boolean DO_JOIN_BETWEEN = false;
   	boolean DID_JOIN_BETWEEN = false;
   	static int CURRENT_TIME = 0;
   	static int counterUntilJoinBetween = 20;
   	String resultsToString = "";
   	int resultTuples = 0;
   	
   	//*************************************************************************
	
	
			
	protected Time currentTime;
	
	private int totalTuples = 0; //tuples
	private int elementsLeftQ = 0;
	private int elementsRightQ = 0;
	private int totalNumberOfJoinResults = 0;	
	static int titleHelper = 1;
	
	int LEFT = 0;
	int RIGHT = 1;
	
	protected static int WINDOW_SIZE = 100; //window size = 100 tuples
	

	SingleClusterSheddyJoinState opState;	
	
	int leftWindowSize;
	int rightWindowSize;

	static int DELTA = 10;
	static int joinCounter = 0;

	int numTuplesInLeftWindow = 0;
	int numTuplesInRightWindow = 0;
		
	static XATTuple spTupleStart = null;
	static XATTuple spTupleEnd = null;
	

	public ClusterSheddyJoinOperatorImp(XATNode init) 
	{
		super(init);
		
		//set the similarity thresholds
		SimilarityImp.THRESHOLD_DISTANCE_X = this.getThresholdDistX();
		SimilarityImp.THRESHOLD_DISTANCE_Y = this.getThresholdDistY();
		SimilarityImp.THRESHOLD_SPEED = this.getThresholdSpeed();
		SimilarityImp.THRESHOLD_TIME = this.getThresholdTime();
		SimilarityImp.THRESHOLD_DESTINATION_ENABLED = this.getIsThresholdDestEnabled();
		if (SimilarityImp.THRESHOLD_DESTINATION_ENABLED == true)
		{
			SimilarityImp.THRESHOLD_DESTINATION_TYPE = SimilarityImp.translateDestThresholdType(this.getThresholdDestType());
			SimilarityImp.THRESHOLD_DESTINATION_VALUE = this.getThresholdDestApproxDist();
		}
		
		flipCoin = new Coin();
	}
	
	

	
	

	public boolean run() throws Exception 
	{

		if(this.getDebug())
		{
			System.out.println("Running: " + this.toToolTipString() + "\n");
		}


		
		//outputqueue for join output
		//modified by Song Wang to support multiple output queue
		for (int i=0;i<this.getOutputQueueCount();i++)
		{
			if(this.getOutputQueue(i).getSchema() == null)
			{
				init(i);
			}
		}	

		/*load shedding setup time*/
		double timeLoadSheddingSetupStart = 0;
		double timeLoadSheddingStartNs = 0;
		/*load shedding setup time*/
		double timeLoadSheddingSetupStop = 0;
		double timeLoadSheddingStartNsStop = 0;
		
		
		if (this.is_load_shedding)
		{
			System.out.println("Load Shedding enabled...");
			if(this.load_shedding_which_clusters_policy.startsWith("random_clusters") && (this.load_shedding_how_much_policy.equals("full_shedding")))
			{
				System.out.println("Random clusters policy...");
				//loop through the cluster grid and randomly discard clusters				
				for (int i=0; i < opState.getClusterGrid().getXDim(); i++)
				{
					//Loop through grid cells in the y-dimension
		 			for (int j=0; j < opState.getClusterGrid().getYDim(); j++)
					{
						//Loop through grid cells in the t-dimension
						for (int k=0; k < opState.getClusterGrid().getTDim(); k++)
						{
							if (opState.getClusterGrid().directory[i][j][k] != null)
							{	
								if (opState.getClusterGrid().directory[i][j][k].GetClusters() != null)
								{
									Object[] arrClusters = (Object[])opState.getClusterGrid().directory[i][j][k].GetClusters().entrySet().toArray();							
									
									//loop through clusters
									for (int q=0; q < arrClusters.length; q++)
									{															
										Map.Entry  entry1 =  (Map.Entry)arrClusters[q];		
										
										SCUBAGridEntry clusterGridEntry1 = (SCUBAGridEntry)entry1.getValue();
										
										//flip a coin
										flipCoin.flip();
										if (flipCoin.toString().equals("Heads"))
										{
											//on heads, drop the tuple
											opState.getClusterGrid().directory[i][j][k].RemoveMovingClusterEntry(clusterGridEntry1);
										}
									}
								}
							}
						}
					}
				}				
			}//end if random cluster policy is enabled
			
			if(this.load_shedding_which_clusters_policy.startsWith("uniform") && this.load_shedding_how_much_policy.startsWith("count_half"))
			{
				System.out.println("Uniform (partial -> half) policy...");
				//loop through the cluster grid and randomly discard clusters				
				for (int i=0; i < opState.getClusterGrid().getXDim(); i++)
				{
					//Loop through grid cells in the y-dimension
		 			for (int j=0; j < opState.getClusterGrid().getYDim(); j++)
					{
						//Loop through grid cells in the t-dimension
						for (int k=0; k < opState.getClusterGrid().getTDim(); k++)
						{
							if (opState.getClusterGrid().directory[i][j][k] != null)
							{	
								if (opState.getClusterGrid().directory[i][j][k].GetClusters() != null)
								{
									Object[] arrClusters = (Object[])opState.getClusterGrid().directory[i][j][k].GetClusters().entrySet().toArray();							
									
									//loop through clusters
									for (int q=0; q < arrClusters.length; q++)
									{															
										Map.Entry  entry1 =  (Map.Entry)arrClusters[q];		
										
										SCUBAGridEntry clusterGridEntry1 = (SCUBAGridEntry)entry1.getValue();

										MotionSCUBACluster currCluster = this.opState.getClusterStorage().getClusterByCID(clusterGridEntry1.GetID());																					    	
										double num_members_to_drop = Math.floor(currCluster.memberCount / 2);
											
										int oDropCounter = 0;
										int qDropCounter = 0;
										for (int m=0; m < num_members_to_drop; m++)
										{												
											if(m % 2 == 0)
											{
												MotionSCUBAClusterObjectMember currObjectMember = (MotionSCUBAClusterObjectMember)currCluster.objects.get(oDropCounter);
												currCluster.objects.remove(currObjectMember);
												oDropCounter++;
											}
											else
											{
												MotionSCUBAClusterQueryMember currQueryMember = (MotionSCUBAClusterQueryMember)currCluster.queries.get(qDropCounter);
												currCluster.queries.remove(currQueryMember);
												qDropCounter++;
											}										
										}										
									}
								}
							}
						}
					}					
				}//end if uniform cluster policy is enabled
			}
			
			if(this.load_shedding_which_clusters_policy.startsWith("selective") && this.load_shedding_how_much_policy.startsWith("volatility_membership_based"))
			{
				System.out.println("Selective (volatility-membership-based) policy...");
				timeLoadSheddingSetupStart = System.currentTimeMillis();
				timeLoadSheddingStartNs = System.nanoTime();
				
				int totalNumClusters = opState.getClusterStorage().storage.entrySet().size();
				
				Object[] arrClusters = (Object[])opState.getClusterStorage().storage.entrySet().toArray();
				Object[] arrClustersSorted = new Object[arrClusters.length]; 
				//sort the array by radius
				int min;
				MotionSCUBACluster temp;
				for (int q=0; q < arrClusters.length; q++)
				{
					min = q;
					for (int scan = q+1; scan < arrClusters.length; scan++)
					{
						Map.Entry  entry1 =  (Map.Entry)arrClusters[q];								
						MotionSCUBACluster cluster1 = (MotionSCUBACluster)entry1.getValue();
						arrClustersSorted[q] = cluster1;
						
						Map.Entry  entry2 =  (Map.Entry)arrClusters[scan];								
						MotionSCUBACluster cluster2 = (MotionSCUBACluster)entry2.getValue();
						arrClustersSorted[scan] = cluster2;
						
/** CHUAN LEI commented out as there is no volatilityIndex variable in MotionSCUBACluster class */
//			            if (cluster1.volatilityIndex < cluster2.volatilityIndex)
//			            {
//			               min = scan;
//			            }
			            
			            //Swap the values
				        temp = (MotionSCUBACluster) arrClustersSorted[min];
				        arrClustersSorted[min] = arrClustersSorted[q];
				        arrClustersSorted[q] = temp;
					}
				}
				
				
				
				//loop through the sorted clusters
				for (int q=0; q < arrClusters.length/2; q++)
				{															
					
					MotionSCUBACluster clusterStorageEntry = (MotionSCUBACluster)arrClustersSorted[q];
																							    	
					double num_members_to_drop = Math.floor(clusterStorageEntry.memberCount / 2);
						
					int oDropCounter = 0;
					int qDropCounter = 0;
					for (int m=0; m < num_members_to_drop; m++)
					{												
						if(m % 2 == 0)
						{
							MotionSCUBAClusterObjectMember currObjectMember = (MotionSCUBAClusterObjectMember)clusterStorageEntry.objects.get(oDropCounter);
							clusterStorageEntry.objects.remove(currObjectMember);
							oDropCounter++;
						}
						else
						{
							MotionSCUBAClusterQueryMember currQueryMember = (MotionSCUBAClusterQueryMember)clusterStorageEntry.queries.get(qDropCounter);
							clusterStorageEntry.queries.remove(currQueryMember);
							qDropCounter++;
						}
					}
				}
				/*load shedding setup time*/
				timeLoadSheddingSetupStop = System.currentTimeMillis();
				timeLoadSheddingStartNsStop = System.nanoTime();
			}//end if selective (size-based) cluster policy is enabled
			
			
			
			if(this.load_shedding_which_clusters_policy.startsWith("selective") && this.load_shedding_how_much_policy.startsWith("size_based"))
			{
				System.out.println("Selective (size-based) policy...");
				timeLoadSheddingSetupStart = System.currentTimeMillis();
				timeLoadSheddingStartNs = System.nanoTime();
				
				int totalNumClusters = opState.getClusterStorage().storage.entrySet().size();
				
				Object[] arrClusters = (Object[])opState.getClusterStorage().storage.entrySet().toArray();
				Object[] arrClustersSorted = new Object[arrClusters.length]; 
				//sort the array by radius
				int min;
				MotionSCUBACluster temp;
				for (int q=0; q < arrClusters.length; q++)
				{
					min = q;
					for (int scan = q+1; scan < arrClusters.length; scan++)
					{
						Map.Entry  entry1 =  (Map.Entry)arrClusters[q];								
						MotionSCUBACluster cluster1 = (MotionSCUBACluster)entry1.getValue();
						arrClustersSorted[q] = cluster1;
						
						Map.Entry  entry2 =  (Map.Entry)arrClusters[scan];								
						MotionSCUBACluster cluster2 = (MotionSCUBACluster)entry2.getValue();
						arrClustersSorted[scan] = cluster2;
			            if (cluster1.radius < cluster2.radius)
			            {
			               min = scan;
			            }
			            
			            //Swap the values
				        temp = (MotionSCUBACluster) arrClustersSorted[min];
				        arrClustersSorted[min] = arrClustersSorted[q];
				        arrClustersSorted[q] = temp;
					}
				}
				
				
				
				//loop through the sorted clusters
				for (int q=0; q < arrClusters.length/2; q++)
				{															
					
					MotionSCUBACluster clusterStorageEntry = (MotionSCUBACluster)arrClustersSorted[q];
																							    	
					double num_members_to_drop = Math.floor(clusterStorageEntry.memberCount / 2);
						
					int oDropCounter = 0;
					int qDropCounter = 0;
					for (int m=0; m < num_members_to_drop; m++)
					{												
						if(m % 2 == 0)
						{
							MotionSCUBAClusterObjectMember currObjectMember = (MotionSCUBAClusterObjectMember)clusterStorageEntry.objects.get(oDropCounter);
							clusterStorageEntry.objects.remove(currObjectMember);
							oDropCounter++;
						}
						else
						{
							MotionSCUBAClusterQueryMember currQueryMember = (MotionSCUBAClusterQueryMember)clusterStorageEntry.queries.get(qDropCounter);
							clusterStorageEntry.queries.remove(currQueryMember);
							qDropCounter++;
						}
					}
				}	
			}//end if selective (size-based) cluster policy is enabled
		

			if(this.load_shedding_which_clusters_policy.startsWith("random clusters") && this.load_shedding_how_much_policy.startsWith("count_half"))
			{
				System.out.println("Selective (random clusters) partial load shedding policy...");
				timeLoadSheddingSetupStart = System.currentTimeMillis();
				timeLoadSheddingStartNs = System.nanoTime();
				
				int totalNumClusters = opState.getClusterStorage().storage.entrySet().size();
				
				Object[] arrClusters = (Object[])opState.getClusterStorage().storage.entrySet().toArray();
				for (int q=0; q < arrClusters.length; q++)
				{
					
					//flip a coin
					flipCoin.flip();
					if (flipCoin.toString().equals("Heads"))
					{
						//on heads, drop the tuple, don't touch the cluster
						continue;						
					}
					
					Map.Entry  entry1 =  (Map.Entry)arrClusters[q];								
					MotionSCUBACluster currCluster = (MotionSCUBACluster)entry1.getValue();
																							    	
					double num_members_to_drop = Math.floor(currCluster.memberCount / 2);
						
					int oDropCounter = 0;
					int qDropCounter = 0;
					for (int m=0; m < num_members_to_drop; m++)
					{												
						if(m % 2 == 0)
						{
							MotionSCUBAClusterObjectMember currObjectMember = (MotionSCUBAClusterObjectMember)currCluster.objects.get(oDropCounter);
							currCluster.objects.remove(currObjectMember);
							oDropCounter++;
						}
						else
						{
							MotionSCUBAClusterQueryMember currQueryMember = (MotionSCUBAClusterQueryMember)currCluster.queries.get(qDropCounter);
							currCluster.queries.remove(currQueryMember);
							qDropCounter++;
						}
					}
				}	
		}//end if selective (random) with partial load shedding
			
			
		/*load shedding setup time*/
		timeLoadSheddingSetupStop = System.currentTimeMillis();
		timeLoadSheddingStartNsStop = System.nanoTime();
			
		}//end if load shedding is enabled
		
		
		
		
		/*start recording time*/
		double timeOverallStart = System.currentTimeMillis();
		double timeOverallStartNs = System.nanoTime();
		
		XATTuple newObj = null;
		int indication = LEFT;
		int tupleCount = 0;
	
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));	
				

		elementsLeftQ = this.getLeftInputQueue().getQueueSize();
		elementsRightQ = this.getRightInputQueue().getQueueSize();
				
		
		long leftTimeStamp = 0;
		long rightTimeStamp = 0;
		

		
		while (true) 
		{
			//System.out.println("tupleCount: " + tupleCount + "\n");
			if (tupleCount == howMany) /* done with processing */ 
			{
				break;
			}
			
			try 
			{
				//pick the next tuple to process
				if ((this.getLeftInputQueue().getQueueSize(this) == 0) && (this.getRightInputQueue().getQueueSize(this) > 0)) 
				{
					newObj = this.getRightInputQueue().dequeue(this);
					indication = RIGHT;
					
					
					if(this.getDebug())
					{
						//System.out.println("-RIGHT"+"\n");
					}
				} 
				else if ((this.getRightInputQueue().getQueueSize(this) == 0) && (this.getLeftInputQueue().getQueueSize(this) > 0))  
				{
					newObj = this.getLeftInputQueue().dequeue(this);
					indication = LEFT;
					if(this.getDebug())
					{
						//System.out.println("-LEFT"+"\n");
					}
				} 
				else 
				{
					leftTimeStamp = this.getLeftInputQueue().peek(this).getNthTimeStamp(this.left_time_position);
					rightTimeStamp = this.getRightInputQueue().peek(this).getNthTimeStamp(this.right_time_position);
					
					/*if (this.getDebug())
					{
						System.out.println("leftTimeStamp: " + leftTimeStamp +"\n");
						System.out.println("rightTimeStamp: " + rightTimeStamp +"\n");
					}*/
					
					if(leftTimeStamp <= rightTimeStamp) 
					{
						newObj = this.getLeftInputQueue().dequeue(this);
						indication = LEFT;
					} 
					else 
					{
						newObj = this.getRightInputQueue().dequeue(this);
						indication = RIGHT;
					}					
				}
			} 
			catch(XATQueueException e1) 
			{
				return true;
			}
	
			
			
			
			//processing a dequeued tuple/sp (from either of the queues)
			if (newObj != null)
			{		
				/*if (Integer.parseInt(newObj.getCell(4).convertToString()) > CURRENT_TIME)
				{
					CURRENT_TIME = Integer.parseInt(newObj.getCell(4).convertToString());
					DO_JOIN_BETWEEN = true;
				}*/
				
				if (this.is_load_shedding)
				{
					if(this.load_shedding_which_clusters_policy.startsWith("totally_random"))
					{
						flipCoin.flip();
						if (flipCoin.toString().equals("Heads"))
						{
							//on heads, drop the tuple
							continue;
						}
					}
				}
				
				if (--counterUntilJoinBetween == 0)
				{
					DO_JOIN_BETWEEN = true;
					counterUntilJoinBetween = 20;
				}
				
				//LEFT SIDE (moving objects)
				if (indication == LEFT) 
				{					
					totalTuples++;
					elementsLeftQ++;
																	
					//received a regular tuple
					tupleCount ++;						
					numTuplesInLeftWindow ++;
					
					//extract the values of the current update
					MotionSCUBAMovingObjectState mObject = new MotionSCUBAMovingObjectState(Integer.parseInt(newObj.getCell(1).toString()),//object id
																							EntityType.Object,//type = object
																							newObj.getCell(5).convertToDouble().longValue(),//x- coord
																							newObj.getCell(6).convertToDouble().longValue(),//y- coord
																							Integer.parseInt(newObj.getCell(4).convertToString()),//time when at (x,y) location
																							UpdateType.FromStringToEnum(newObj.getCell(0).convertToString()),//update type (e.g., new,disappear,point)
																							ClassType.FromIntToEnum(Integer.parseInt(newObj.getCell(3).convertToString())),//class type (e.g., pedestrian)
																							Integer.parseInt(newObj.getCell(8).convertToString()),//destination node x- coord
																							Integer.parseInt(newObj.getCell(9).convertToString()),//destination node y- coord
																							(int)Float.parseFloat(newObj.getCell(7).convertToString()));//speed
					
					//mark that the object has been encoutered
					objectsEncounteredArray[mObject.GetID()] = 1;
					
					
					//cluster the object
					int clusterID = clusterMovingObjectBasedOnUpdate(mObject);
					ArrayList results = doWithinClusterJoin(opState.getClusterStorage(), 
															clusterID, 
															EntityType.Object, 
															mObject.GetID(), 
															mObject.GetPosition());
					

					if (results.size() > 0)
					{
						XATTupleImp[] outputTuples = new XATTupleImp[results.size()];
						XATDataValueImp[] outputValue = new XATDataValueImp[1];
						//send the results to the output queue
						for (int i=0; i<results.size(); i++) 
						{
							QueryResult resultTuple = (QueryResult) results.get(i);
							outputValue[0] = new XATDataValueImp(resultTuple);
							outputTuples[i] = new XATTupleImp(outputValue);
							if(this.debugFlag)
							{
								Helper.StatsWriteLineToFile("ClusterSheddyResults.txt", resultTuple.toString());
							}
							resultsToString += resultTuple.toString() + "\n";
							resultTuples++;
						}					
						this.getOutputQueue().enqueue(outputTuples, this);						
					}
									
				}			 
				else 
				{					
					//RIGHT SIDE (moving queries)					
					totalTuples++;
					elementsRightQ++;
												
					//received a regular tuple
					tupleCount ++;						
					numTuplesInRightWindow ++;
					
				
					
					//extract the values of the current update
					MotionSCUBAMovingQueryState mQuery = new MotionSCUBAMovingQueryState(Integer.parseInt(newObj.getCell(1).toString()),//query id
																						  EntityType.Query,//type = query
																						  newObj.getCell(5).convertToDouble().longValue(),//x- coord
																						  newObj.getCell(6).convertToDouble().longValue(),//y- coord
																						  Integer.parseInt(newObj.getCell(4).convertToString()),//time when at (x,y) location
																						  UpdateType.FromStringToEnum(newObj.getCell(0).convertToString()),//update type (e.g., new,disappear,point)
																						  ClassType.FromIntToEnum(Integer.parseInt(newObj.getCell(3).convertToString())),//class type (e.g., pedestrian)
																						  QueryType.Range,
																						  this.getRangeQuerySize(),//query range (specified in the config file)
																						  Integer.parseInt(newObj.getCell(8).convertToString()),//destination node x- coord
																						  Integer.parseInt(newObj.getCell(9).convertToString()),//destination node y- coord
																						  (int)Float.parseFloat(newObj.getCell(7).convertToString()));//speed

					
					//mark that the query has been encoutered
					queriesEncounteredArray[mQuery.GetID()] = 1;
					
					//cluster the query
					int clusterID = clusterMovingQueryBasedOnUpdate(mQuery);
					ArrayList results = doWithinClusterJoin(opState.getClusterStorage(), 
															clusterID, 
															EntityType.Query, 
															mQuery.GetID(), 
															mQuery.GetPosition());

					if (results.size() > 0)
					{
						XATTupleImp[] outputTuples = new XATTupleImp[results.size()];
						XATDataValueImp[] outputValue = new XATDataValueImp[1];
						//	send the results to the output queue
						for (int i=0; i<results.size(); i++) 
						{
							QueryResult resultTuple = (QueryResult) results.get(i);
							outputValue[0] = new XATDataValueImp(resultTuple);
							outputTuples[i] = new XATTupleImp(outputValue);
							if(this.debugFlag)
							{
								Helper.StatsWriteLineToFile("ClusterSheddyResults.txt", resultTuple.toString());
							}
							resultsToString += resultTuple.toString() + "\n";
							resultTuples++;
						}					
						this.getOutputQueue().enqueue(outputTuples, this);						
					}

				}//end if object dequeued is NOT null
			
			}//end while
		
		
		
			if (DO_JOIN_BETWEEN)
			{
				DID_JOIN_BETWEEN = true;
				if (this.getDebug())
				{
					System.out.println("Executing Join-Between...");
				}
				
				//do join between clusters
				ArrayList results = doBetweenClusterJoin(this.getRangeQuerySize());
				if (results.size() > 0)
				{
					XATTupleImp[] outputTuples = new XATTupleImp[results.size()];
					XATDataValueImp[] outputValue = new XATDataValueImp[1];
					//	send the results to the output queue
					for (int i=0; i<results.size(); i++) 
					{
						QueryResult resultTuple = (QueryResult) results.get(i);
						outputValue[0] = new XATDataValueImp(resultTuple);
						outputTuples[i] = new XATTupleImp(outputValue);
						if(this.debugFlag)
						{
							Helper.StatsWriteLineToFile("ClusterSheddyResults.txt", resultTuple.toString());
						}
						resultsToString += resultTuple.toString() + "\n";
						resultTuples++;
					}					
					this.getOutputQueue().enqueue(outputTuples, this);						
				}
				
				if (this.getDebug())
				{
					System.out.println("Got "+ results.size() + " results from join-between.");
					Helper.StatsWriteLineToFile("ClusterSheddyResults.txt", "Got "+ results.size() + " results from join-between.");
				}
				
				if (this.getDebug())
				{
					System.out.println("Cluster Statistics...");
					System.out.println(opState.getClusterGrid().statisticsToString(this.opState.getClusterStorage()));
					System.out.println("");
					System.out.println(opState.getClusterGrid().graphicalStatisticsToString(this.opState.getClusterStorage()));
					System.out.println("# distinct objects processed " + this.getDistinctObjectsCount());
					System.out.println("# distinct queries processed " + this.getDistinctQueriesCount());
					System.out.println("# tuples processed at this run " + totalTuples);
					System.out.println("# tuples processed from left queue " + elementsLeftQ);
					System.out.println("# tuples processed from right queue " + elementsRightQ);
				}
				
				DO_JOIN_BETWEEN = false;
			}
		
		}
		
		double timeOverallEnd = System.currentTimeMillis();
		double timeOverallMilliseconds = (timeOverallEnd - timeOverallStart);//in ms		
		
		double timeOverallEndNs = System.nanoTime();
		double timeOverallNs = (timeOverallEndNs - timeOverallStartNs);//in ns	
		
		
		double timeOverallLoadSheddingSetupMs = timeLoadSheddingSetupStop - timeLoadSheddingSetupStart;
		double timeOverallLoadSheddingSetupNs = timeLoadSheddingStartNsStop - timeLoadSheddingStartNs;
		
		Helper.StatsWriteLineToFile("ClusterSheddyResults.txt", resultsToString);
		
		//if (DID_JOIN_BETWEEN == true)
		//{
			Helper.StatsWriteLineToFile("ClusterSheddyStatistics.txt", Double.toString(timeOverallNs) + "\t" + Double.toString(timeOverallMilliseconds) + "\t" + resultTuples  + "\t" + Double.toString(timeOverallLoadSheddingSetupMs) + "\t" + Double.toString(timeOverallLoadSheddingSetupNs));
			DID_JOIN_BETWEEN = false;
		//}
		resultsToString = "";
		resultTuples = 0;
		
		return true;
	}
	
	
	private int getDistinctObjectsCount()
	{
		int count = 0;
		for(int i=0; i<objectsEncounteredArray.length;i++)
		{
			if (objectsEncounteredArray[i]==1)
			{
				count++;
			}
		}
		return count;
	}
	
	
	private int getDistinctQueriesCount()
	{
		int count = 0;
		for(int i=0; i<queriesEncounteredArray.length;i++)
		{
			if (queriesEncounteredArray[i]==1)
			{
				count++;
			}
		}
		return count;
	}
	
	
	private ArrayList doBetweenClusterJoin(int querySize)
	{		
		int numOverlap = 0;
		int numJoinComparisons = 0;
		ArrayList miniResults = new ArrayList();
			
		//Loop through grid cells in the x-dimension
		for (int i=0; i < opState.getClusterGrid().getXDim(); i++)
		{
			//Loop through grid cells in the y-dimension
 			for (int j=0; j < opState.getClusterGrid().getYDim(); j++)
			{
				//Loop through grid cells in the t-dimension
				for (int k=0; k < opState.getClusterGrid().getTDim(); k++)
				{
					if (opState.getClusterGrid().directory[i][j][k] != null)
					{	
						if (opState.getClusterGrid().directory[i][j][k].GetClusters() != null)
						{
							Object[] arrClusters = (Object[])opState.getClusterGrid().directory[i][j][k].GetClusters().entrySet().toArray();							
							
							//loop through clusters
							for (int q=0; q < arrClusters.length; q++)
							{															
								for (int o=q; o < arrClusters.length; o++)
								{				
									Map.Entry  entry1 =  (Map.Entry)arrClusters[q];		
									Map.Entry  entry2 =  (Map.Entry)arrClusters[o];		
	
									SCUBAGridEntry clusterGridEntry1 = (SCUBAGridEntry)entry1.getValue();
									SCUBAGridEntry clusterGridEntry2 = (SCUBAGridEntry)entry2.getValue();
										
									//perform join between only if 2 clusters are different
									if (clusterGridEntry1.GetID() != clusterGridEntry2.GetID())
									{
										numJoinComparisons++;
										//do between-cluster join
										if (opState.getClusterGrid().DoBetweenJoin(clusterGridEntry1, clusterGridEntry2, new GridCellIndex3D(i,j,k), opState.getClusterStorage()))
										{
											numOverlap++;
											//clusters overlap, so perform a cluster-within join
											ArrayList joinResults = opState.getClusterGrid().clusterSheddyDoWithinJoin(clusterGridEntry1, clusterGridEntry2, new GridCellIndex3D(i,j,k), opState.getClusterStorage(), querySize);
									
											if (joinResults != null)
											{
												miniResults.addAll(joinResults);
											}									
										}
									}
								}
							}
						}
					}
				}
			}
		}	
		
		if(this.debugFlag)
		{
			Helper.StatsWriteLineToFile("ClusterSheddyResults.txt", "# of overlaps: "+numOverlap);
		}
				
		return miniResults;
	}
	
	
	public ArrayList doWithinClusterJoin(MotionSCUBAClusterStorage clusterStorage, int clusterId, int entityType, int entityId, PositionSpt2D location)
	{
		ArrayList withinClusterJoinResults = new ArrayList();
		//retrieve cluster
		MotionSCUBACluster cluster = clusterStorage.getClusterByCID(new Integer(clusterId));

		if (entityType == EntityType.Object)
		{			
			//join the object with queries
			for(int i=0; i < cluster.queries.size(); i++)
			{
				MotionSCUBAClusterQueryMember queryState = (MotionSCUBAClusterQueryMember)cluster.queries.get(i);
				PositionSpt2D queryPosition = MotionSCUBAPolarCoordinate.TranslateFromPolarToCartesian(queryState.coordinates, cluster.centerPoint);
			
				if (DoDiscreteJoin(location, queryPosition, this.getRangeQuerySize()) == true)
				{
					//object and query join
					QueryResult qResult = new QueryResult(queryState.id, entityId, (int)location.GetT(), "jw");
					withinClusterJoinResults.add(qResult);
				}
			}
		}
		else if (entityType == EntityType.Query)
		{
			//join the query with objects
			for(int i=0; i < cluster.objects.size(); i++)
			{
				MotionSCUBAClusterObjectMember objectState = (MotionSCUBAClusterObjectMember)cluster.objects.get(i);
				PositionSpt2D objectPosition = MotionSCUBAPolarCoordinate.TranslateFromPolarToCartesian(objectState.coordinates, cluster.centerPoint);
			
				if (DoDiscreteJoin(objectPosition, location, this.getRangeQuerySize()) == true)
				{
					//query and object join
					QueryResult qResult = new QueryResult(entityId, objectState.id, (int)location.GetT(), "jw");
					withinClusterJoinResults.add(qResult);
				}
			}
		}
	
		return withinClusterJoinResults;
	}
	
	
	public static boolean DoDiscreteJoin(PositionSpt2D objectPosition, //object
			   							 PositionSpt2D queryPosition,  //query
			   							 int queryBoundary)
	{	
		if ((objectPosition.GetX() < (queryPosition.GetX() + queryBoundary))&&
		(objectPosition.GetY() < (queryPosition.GetY() + queryBoundary))&&
		(objectPosition.GetX() > (queryPosition.GetX() - queryBoundary))&&
		(objectPosition.GetY() > (queryPosition.GetY() - queryBoundary))&&
		(objectPosition.GetT() == queryPosition.GetT()))
		{
			return true;
		}
		return false;
	}
	
	private int clusterMovingQueryBasedOnUpdate(MotionSCUBAMovingQueryState mQuery)
	{
		int clusterID = HelpClusteringImp.clusterMovingQuery(opState.getClusterGrid(), 
				   											 opState.getClusterStorage(), 
				   											 opState.getClusterHome(),
				   											 range_query_size,
				   											 threshold_distance_x,
				   											 threshold_distance_y,
				   											 threshold_speed,
				   											 threshold_time,       	
				   											 is_threshold_destination_enabled,
				   											 threshold_destination_type,
				   											 threshold_destination_approximate_distance,
				   											 mQuery);
		return clusterID;
	}
	
	private int clusterMovingObjectBasedOnUpdate(MotionSCUBAMovingObjectState mObject)
	{		
		//probe for a cluster (returns a cluster id that the object now belongs to)
		int clusterID =  HelpClusteringImp.clusterMovingObject(opState.getClusterGrid(), 
															   opState.getClusterStorage(), 
															   opState.getClusterHome(),
															   threshold_distance_x,
															   threshold_distance_y,
															   threshold_speed,
															   threshold_time,       	
															   is_threshold_destination_enabled,
															   threshold_destination_type,
															   threshold_destination_approximate_distance,
															   mObject);		
		return 	clusterID;	
	}
	
	
	
	private void resetStatisticsVariables()
	{
		totalTuples = 0; //tuples + sps
		elementsLeftQ = 0;
		elementsRightQ = 0;
		totalNumberOfJoinResults = 0;		
	}
	
	public void setState(SingleClusterSheddyJoinState specifiedState) 
	{
		this.opState = (SingleClusterSheddyJoinState)specifiedState;
		this.opState.createGridWithSpecifiedDimensions(this.getGridXDim(), this.getGridYDim(), this.getGridTDim());
	}

	

	public String toToolTipString() 
	{
		return "ClusterSheddy";
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#hasUnmatchedState()
	 */
	public boolean hasUnmatchedState() 
	{
		return false;
	}
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return true;
	}
	public boolean cleanAllAccumulatedTuples() 
	{
		return true;
	}
	public int setStateID(int startID) 
	{
		return 0;
	}
	public boolean recomputeUnmatchedState() 
	{
		return true;
	}
	public boolean matchStates(XATStreamOperator op) 
	{
		return true;
	}
	public boolean finalizeOperation()
	{
		return true;
	}
	
	
	public void setLeftWindowSize(int newSize)
	{
		leftWindowSize = newSize;
	}
	
	public void setRightWindowSize(int newSize)
	{
		rightWindowSize = newSize;
	}	
	public void setLeftTimestampPos(int lLocTimeAttrPos) 
	{
		left_time_position = lLocTimeAttrPos;		
	}
	public void setRightTimestampPos(int rLocTimeAttrPos) 
	{
		right_time_position = rLocTimeAttrPos;
	}
	public void setDeltaJoin(int deltaJoin) 
	{
		delta_join = deltaJoin;
	}	
	public void setGridXDim(int gridXDim) 
	{
		grid_x_dim = gridXDim;
	}	
	public void setGridYDim(int gridYDim) 
	{
		grid_y_dim = gridYDim;
	}
	public void setGridTDim(int gridTDim) 
	{
		grid_t_dim = gridTDim;
	}	
	public void setThresholdDistX(int thrDistX) 
	{
		threshold_distance_x = thrDistX;
	}	
	public void setThresholdDistY(int thrDistY) 
	{
		threshold_distance_y = thrDistY;
	}	
	public void setThresholdSpeed(int thrSpeed) 
	{
	   	threshold_speed = thrSpeed;	   			
	}	
	public void setThresholdTime(int thrTime) 
	{
		threshold_time = thrTime;
	}	
	public void setIsThresholdDestEnabled(boolean isEnabled) 
	{
		is_threshold_destination_enabled = isEnabled;
	}	
	public void setThresholdDestType(String thrDestType) 
	{
		threshold_destination_type = thrDestType;		
	}	
	public void setThresholdDestApproxDist(int thrDestApproxDist) 
	{
		threshold_destination_approximate_distance = thrDestApproxDist;		
	}	
	public void setIsLoadSheddingEnabled(boolean isEnabled) 
	{
		is_load_shedding = isEnabled;
	}	
	public void setLoadSheddingWhichClustersPolicy(String policyName) 
	{
		load_shedding_which_clusters_policy = policyName;		
	}	
	public void setLoadSheddingHowMuchPerClusterPolicy(String policyName) 
	{
		load_shedding_how_much_policy = policyName;		
	}	
	public void setDeltaLoadShed(int deltaLoadShed) 
	{
		delta_load_shed = deltaLoadShed;
	}
	public int getLeftWindowSize() 
	{
		return this.leftWindowSize;
	}
	public int getRightWindowSize() 
	{		
		return this.rightWindowSize;
	}
	public int getLeftTimestampPos() 
	{
		return this.left_time_position;
	}
	public int getRightTimestampPos() 
	{
		return this.right_time_position;
	}
	public int getDeltaJoin() 
	{
		return this.delta_join;
	}
	public int getGridXDim() 
	{
		return this.grid_x_dim;
	}
	public int getGridYDim() 
	{
		return this.grid_y_dim;
	}
	public int getGridTDim() 
	{
		return this.grid_t_dim;
	}
	public int getThresholdDistX() 
	{	
		return this.threshold_distance_x;
	}
	public int getThresholdDistY() 
	{	
		return this.threshold_distance_y;
	}
	public int getThresholdSpeed() 
	{
		return this.threshold_speed;
	}
	public int getThresholdTime() 
	{
		return this.threshold_time;
	}
	public boolean getIsThresholdDestEnabled() 
	{
		return this.is_threshold_destination_enabled;
	}
	public String getThresholdDestType() 
	{	
		return this.threshold_destination_type;
	}
	public int getThresholdDestApproxDist() 
	{	
		return this.threshold_destination_approximate_distance;
	}
	public boolean getIsLoadSheddingEnabled() 
	{
		return this.is_load_shedding;
	}
	public String getLoadSheddingWhichClustersPolicy() 
	{	
		return this.load_shedding_which_clusters_policy;
	}
	public String getLoadSheddingHowMuchPerClusterPolicy() 
	{	
		return this.load_shedding_how_much_policy;
	}
	public int getDeltaLoadShed() 
	{
		return this.delta_load_shed;
	}
	public XATQueue getNthInputQueue(int n) 
	{
		return null;
	}




	public void setRangeQuerySize(int rangeQuerySize) 
	{
		range_query_size = rangeQuerySize;		
	}




	public int getRangeQuerySize() 
	{		
		return range_query_size;
	}
}
