/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.clustersheddy;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.MotionSCUBAStatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.DimensionType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.FixedDiscrete3DGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GenericSpatioTemporalGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridCellIndex3D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridEntry;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingObjectState;

public class HelpClusteringImp 
{		
	public static int clusterMovingObject(MotionSCUBAClusterGrid clusterGrid,//hash table that contains cluster entries
										  MotionSCUBAClusterStorage clusterStorage,//table that stores clusters' properties 
										  MotionSCUBAClusterHome clusterHome,//mapping between objects and queris and their clusters
										  int threshold_distance_x,
										  int threshold_distance_y,
										  int threshold_speed,
										  int threshold_time,       	
										  boolean is_threshold_destination_enabled,
										  String threshold_destination_type,
										  int threshold_destination_approximate_distance,
										  MotionSCUBAMovingObjectState mObject)//moving object state 										  
										  	
	{
	
		boolean isClusterMember = false;
		boolean isOverlappingWithMultipleCells = false;
		int returnClusterID = -999;
		
		//probe to see if object previously belonged to any cluster
		Integer clusterID = (Integer)clusterHome.clusterHome.get("o" + mObject.GetID());		
		
		if (clusterID != null)
		{
			//object previously belonged to a cluster
			MotionSCUBACluster parentCluster = (MotionSCUBACluster)clusterStorage.storage.get(clusterID);
			
			//check to see if object still belongs to the parent cluster
			if (parentCluster.isClusterCandidate(mObject, 
												 threshold_distance_x,
					  							 threshold_distance_y,
					 							 threshold_speed,
					  							 threshold_time,       	
					  							 is_threshold_destination_enabled,
					 							 threshold_destination_type,
					 							 threshold_destination_approximate_distance))
			{
				boolean IsCenterMoved = false;
				boolean IsRadiusChanged = false;
				
				parentCluster.adjustOldClusterMemberObject(mObject,IsCenterMoved,IsRadiusChanged);		
				
		    	//re-adjust grid indicies for the cluster
				clusterGrid.adjustCluster(parentCluster,IsCenterMoved,IsRadiusChanged, isOverlappingWithMultipleCells);	
								
				returnClusterID = parentCluster.cid; 
				return returnClusterID;
			}
			else
			{
				boolean isEmptyCluster = false;
				//no longer belongs to this cluster
				//so remove it from the members list
				parentCluster.removeOldClusterMemberObject(mObject, isEmptyCluster);	
				clusterHome.clusterHome.remove("o" + mObject.GetID());
				
				if (isEmptyCluster)
				{
					clusterStorage.storage.remove(new Integer(parentCluster.cid));
				    clusterGrid.RemoveCluster(parentCluster);
					
				}
				
		    	//re-adjust grid indicies for the cluster
				//clusterGrid.AdjustCluster(parentCluster);	
			}
		}
		
		
		int xIndex = MotionSCUBAClusterGrid.GetCellIndexByLocation(mObject.GetPosition().GetX(), DimensionType.x, clusterGrid.getXDim());		
		int yIndex = MotionSCUBAClusterGrid.GetCellIndexByLocation(mObject.GetPosition().GetY(), DimensionType.y, clusterGrid.getYDim());
		int tIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)

		//check if cell exists, if it doesn't create it
		clusterGrid.CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex,tIndex);
		
		//get clusters that fall into the same grid cell as the current object update
		HashMap cellClusters = clusterGrid.GetClustersByCellIndex(xIndex,yIndex,tIndex);


		Iterator i = cellClusters.entrySet().iterator();
		while(i.hasNext())
		{						
			Map.Entry  entry =  (Map.Entry)i.next();			
			if (entry != null)
			{
				SCUBAGridEntry clusterGridEntry = (SCUBAGridEntry)entry.getValue();
			
				MotionSCUBACluster currCluster = clusterStorage.getClusterByCID(new Integer(clusterGridEntry.GetID()));
				if (currCluster.isClusterCandidate(mObject, 
						 						   threshold_distance_x,
						 						   threshold_distance_y,
						 						   threshold_speed,
						 						   threshold_time,       	
						 						   is_threshold_destination_enabled,
						 						   threshold_destination_type,
						 						   threshold_destination_approximate_distance))
				{
					boolean IsCenterMoved = false;
					boolean IsRadiusChanged = false;
				
					currCluster.AdoptNewClusterMemberObject(mObject, IsCenterMoved,IsRadiusChanged);				
					//save relationship in the ClusterHome table
					clusterHome.clusterHome.put("o"+mObject.GetID(), new Integer(currCluster.cid));
					isClusterMember = true;
				
					//re-adjust grid indicies for the cluster
					clusterGrid.adjustCluster(currCluster, IsCenterMoved,IsRadiusChanged, isOverlappingWithMultipleCells);
				
					returnClusterID = currCluster.cid; 
					break;
				}
			}
		}
		
		//not a member of any cluster, need to create a new cluster
		if (!isClusterMember)
		{
			//create new cluster
			MotionSCUBACluster newCluster = createNewClusterBasedOnMovingObject(mObject,clusterGrid);
			
			//save relationship in the ClusterHome table
			clusterHome.clusterHome.put("o"+mObject.GetID(), new Integer(newCluster.cid));
			
			//add new cluster's ID to the grid cell
			SCUBAGridEntry clusterGridEntry = new SCUBAGridEntry(newCluster.cid);	
			clusterGrid.directory[xIndex][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
			//save cluster in the Cluster Storage
			clusterStorage.addNewCluster(newCluster);
			returnClusterID = newCluster.cid; 
			
		}		
		return returnClusterID;
	}
	
	
	
	public static int clusterMovingQuery(MotionSCUBAClusterGrid clusterGrid,
				 						 MotionSCUBAClusterStorage clusterStorage,
				 						 MotionSCUBAClusterHome clusterHome,
				 						 int query_size,
				 						 int threshold_distance_x,
										 int threshold_distance_y,
										 int threshold_speed,
										 int threshold_time,       	
										 boolean is_threshold_destination_enabled,
										 String threshold_destination_type,
										 int threshold_destination_approximate_distance,
				 						 MotionSCUBAMovingQueryState mQuery)	
	{
				
		boolean isOverlappingWithMultipleCells = false;
		boolean isClusterMember = false;
		int returnClusterID = -1;
		
		//probe to see if query previously belonged to any cluster
		Integer clusterID = (Integer)clusterHome.clusterHome.get("q" + mQuery.GetID());		
		
		if (clusterID != null)
		{
			MotionSCUBACluster parentCluster = (MotionSCUBACluster)clusterStorage.storage.get(clusterID);			
			//check to see if query still belongs to the parent cluster
			
			if (parentCluster.isClusterCandidate(mQuery, 
					 							 threshold_distance_x,
					 							 threshold_distance_y,
					 							 threshold_speed,
					 							 threshold_time,       	
					 							 is_threshold_destination_enabled,
					 							 threshold_destination_type,
					 							 threshold_destination_approximate_distance))

			{
				boolean IsCenterMoved = false;
				boolean IsRadiusChanged = false;
		
				parentCluster.AdjustOldClusterMemberQuery(mQuery, IsCenterMoved, IsRadiusChanged);	
		
				//re-adjust grid indicies for the cluster
				clusterGrid.adjustCluster(parentCluster, IsCenterMoved, IsRadiusChanged, isOverlappingWithMultipleCells);					
		
				returnClusterID = parentCluster.cid; 
				return returnClusterID;
			}
			else
			{
				boolean isEmptyCluster = false;
				//no longer belongs to this cluster
				//so remove it from the members list
				parentCluster.removeOldClusterMemberQuery(mQuery,isEmptyCluster);
				clusterHome.clusterHome.remove("q" + mQuery.GetID());
				
				if (isEmptyCluster)
				{
					clusterStorage.storage.remove(new Integer(parentCluster.cid));
					clusterGrid.RemoveCluster(parentCluster);					
				}
				
				//re-adjust grid indicies for the cluster
				//clusterGrid.AdjustCluster(parentCluster);	
			}
		}
		
		int xIndex = MotionSCUBAClusterGrid.GetCellIndexByLocation(mQuery.GetPosition().GetX(), DimensionType.x, clusterGrid.getXDim());		
		int yIndex = MotionSCUBAClusterGrid.GetCellIndexByLocation(mQuery.GetPosition().GetY(), DimensionType.y, clusterGrid.getYDim());
		int tIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)

		
		clusterGrid.CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex,tIndex);
		
		//get clusters that fall into the same grid cell as the current query update
		HashMap cellClusters = clusterGrid.GetClustersByCellIndex(xIndex,yIndex,tIndex);
		
		
		Iterator i = cellClusters.entrySet().iterator();
		while(i.hasNext())
		{						
			Map.Entry  entry =  (Map.Entry)i.next();			
			if (entry != null)
			{
				SCUBAGridEntry clusterGridEntry = (SCUBAGridEntry)entry.getValue();				
				MotionSCUBACluster currCluster = clusterStorage.getClusterByCID(new Integer(clusterGridEntry.GetID()));
				if (currCluster.isClusterCandidate(mQuery, 
						 						   threshold_distance_x,
						 						   threshold_distance_y,
						 						   threshold_speed,
						 						   threshold_time,       	
						 						   is_threshold_destination_enabled,
						 						   threshold_destination_type,
						 						   threshold_destination_approximate_distance))
				{
					boolean IsCenterMoved = false;
					boolean IsRadiusChanged = false;
					
					currCluster.clusterSheddyAdoptNewClusterMemberQuery(mQuery, query_size, IsCenterMoved, IsRadiusChanged);
					
					//save relationship in the ClusterHome table
					clusterHome.clusterHome.put("q"+mQuery.GetID(), new Integer(currCluster.cid));
					isClusterMember = true;
					
					//re-adjust grid indicies for the cluster
					clusterGrid.adjustCluster(currCluster, IsCenterMoved, IsRadiusChanged, isOverlappingWithMultipleCells);	
					
					returnClusterID = currCluster.cid; 				
					break;
				}
			}
		}
		
		//not a member of any cluster, need to create a new cluster
		if (!isClusterMember)
		{
			//create new cluster
			MotionSCUBACluster newCluster = createNewClusterBasedOnMovingQuery(mQuery,query_size,clusterGrid);	
			
			//save relationship in the ClusterHome table
			clusterHome.clusterHome.put("q"+mQuery.GetID(), new Integer(newCluster.cid));
			
			//add new cluster's ID to the grid cell
			SCUBAGridEntry clusterGridEntry = new SCUBAGridEntry(newCluster.cid);	
			clusterGrid.directory[xIndex][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
			//save cluster in the Cluster Storage
			clusterStorage.addNewCluster(newCluster);
			
			returnClusterID = newCluster.cid;
			
		}		
			
		return returnClusterID;

	}

	
	
	
	
    public static MotionSCUBACluster createNewClusterBasedOnMovingObject(MotionSCUBAMovingObjectState mObject,MotionSCUBAClusterGrid clusterGrid)
    {
    	MotionSCUBACluster newCluster = new MotionSCUBACluster();
    	newCluster.centerPoint = mObject.GetPosition();
    	newCluster.nextNodeX = mObject.nextNodeX;
    	newCluster.nextNodeY = mObject.nextNodeY;
    	    	
    	//create polar coordinates for the current object
    	MotionSCUBAClusterObjectMember newClusterMember = new MotionSCUBAClusterObjectMember(mObject.GetID(), MotionSCUBAPolarCoordinate.TranslateFromCartesianToPolar(mObject.GetPosition()));
    	
    	newCluster.objects.add(newClusterMember);
    	newCluster.memberCount = 1; 
    	newCluster.speed = mObject.speed;
    	newCluster.radius = 0;
    	newCluster.furthestPosition = mObject.GetPosition();
    	newCluster.time = (int) mObject.GetPosition().GetT();
    	
    	//grid index of the centeroid of the cluster
    	newCluster.currentGridIndexX = MotionSCUBAClusterGrid.GetCellIndexByLocation(mObject.GetPosition().GetX(), DimensionType.x, clusterGrid.getXDim());		
    	newCluster.currentGridIndexY = MotionSCUBAClusterGrid.GetCellIndexByLocation(mObject.GetPosition().GetY(), DimensionType.y, clusterGrid.getYDim());
    	newCluster.currentGridIndexT = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)
    	
    	newCluster.gridCellsOverlap.add(new GridCellIndex3D(newCluster.currentGridIndexX,newCluster.currentGridIndexY,newCluster.currentGridIndexT));
    	return newCluster;
    }
    
    public static MotionSCUBACluster createNewClusterBasedOnMovingQuery(MotionSCUBAMovingQueryState mQuery, int querySize, MotionSCUBAClusterGrid clusterGrid)
    {
    	MotionSCUBACluster newCluster = new MotionSCUBACluster();
    	newCluster.centerPoint = mQuery.GetPosition();
    	newCluster.nextNodeX = mQuery.nextNodeX;
    	newCluster.nextNodeY = mQuery.nextNodeY;
    	    	
    	//create polar coordinates for the current object
    	MotionSCUBAClusterQueryMember newClusterMember = new MotionSCUBAClusterQueryMember(mQuery.GetID(), MotionSCUBAPolarCoordinate.TranslateFromCartesianToPolar(mQuery.GetPosition()), querySize);
    	
    	newCluster.queries.add(newClusterMember);
    	newCluster.memberCount = 1;
    	newCluster.speed = mQuery.speed;
    	newCluster.radius = querySize;
    	newCluster.furthestPosition = mQuery.GetPosition();
    	newCluster.time = (int) mQuery.GetPosition().GetT();
    	
    	//grid index of the centeroid of the cluster
    	newCluster.currentGridIndexX = MotionSCUBAClusterGrid.GetCellIndexByLocation(mQuery.GetPosition().GetX(), DimensionType.x, clusterGrid.getXDim());		
    	newCluster.currentGridIndexY = MotionSCUBAClusterGrid.GetCellIndexByLocation(mQuery.GetPosition().GetY(), DimensionType.y, clusterGrid.getYDim());
    	newCluster.currentGridIndexT = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)
    	newCluster.gridCellsOverlap.add(new GridCellIndex3D(newCluster.currentGridIndexX,newCluster.currentGridIndexY,newCluster.currentGridIndexT));

    	int x1 = MotionSCUBAClusterGrid.GetCellIndexByLocation(mQuery.GetPosition().GetX()+querySize, DimensionType.x, clusterGrid.getXDim());		    
    	int x2 = MotionSCUBAClusterGrid.GetCellIndexByLocation(mQuery.GetPosition().GetX()-querySize, DimensionType.x, clusterGrid.getXDim());		
    	int y1 = MotionSCUBAClusterGrid.GetCellIndexByLocation(mQuery.GetPosition().GetY()+querySize, DimensionType.y, clusterGrid.getYDim());
    	int y2 = MotionSCUBAClusterGrid.GetCellIndexByLocation(mQuery.GetPosition().GetY()-querySize, DimensionType.y, clusterGrid.getYDim());

    	//add new cluster's ID to the grid cell
		SCUBAGridEntry clusterGridEntry = new SCUBAGridEntry(newCluster.cid);	
    	if ((x1 < clusterGrid.getXDim()) && (y1 < clusterGrid.getYDim())){ 
		if ((x1 != newCluster.currentGridIndexX) && (y1 != newCluster.currentGridIndexY)) 
    	{
			clusterGrid.CheckCreateGridCellByIndexAndSpaceRange(x1,y1,0);	
			clusterGrid.directory[x1][y1][0].AddMovingClusterEntry(clusterGridEntry);		
			newCluster.gridCellsOverlap.add(new GridCellIndex3D(x1,y1,0));

    	}
    	else if ((x1 != newCluster.currentGridIndexX) && (y1 == newCluster.currentGridIndexY)) 
    	{
			clusterGrid.CheckCreateGridCellByIndexAndSpaceRange(x1,y1,0);	
			clusterGrid.directory[x1][y1][0].AddMovingClusterEntry(clusterGridEntry);
			newCluster.gridCellsOverlap.add(new GridCellIndex3D(x1,y1,0));
    	}
    	else if ((x1 == newCluster.currentGridIndexX) && (y1 != newCluster.currentGridIndexY)) 
    	{
			clusterGrid.CheckCreateGridCellByIndexAndSpaceRange(x1,y1,0);	
			clusterGrid.directory[x1][y1][0].AddMovingClusterEntry(clusterGridEntry);
			newCluster.gridCellsOverlap.add(new GridCellIndex3D(x1,y1,0));
    	}
    	}
    	
    	if ((x2 < clusterGrid.getXDim()) && (y2 < clusterGrid.getYDim())){ 
    	if ((x2 != newCluster.currentGridIndexX) && (y2 != newCluster.currentGridIndexY)) 
    	{
			clusterGrid.CheckCreateGridCellByIndexAndSpaceRange(x2,y2,0);	
    		clusterGrid.directory[x2][y2][0].AddMovingClusterEntry(clusterGridEntry);
			newCluster.gridCellsOverlap.add(new GridCellIndex3D(x2,y2,0));
    	}
    	else if ((x2 != newCluster.currentGridIndexX) && (y2 == newCluster.currentGridIndexY)) 
    	{
			clusterGrid.CheckCreateGridCellByIndexAndSpaceRange(x2,y2,0);	
			clusterGrid.directory[x2][y2][0].AddMovingClusterEntry(clusterGridEntry);
			newCluster.gridCellsOverlap.add(new GridCellIndex3D(x2,y2,0));
    	}
    	else if ((x2 == newCluster.currentGridIndexX) && (y2 != newCluster.currentGridIndexY)) 
    	{
			clusterGrid.CheckCreateGridCellByIndexAndSpaceRange(x2,y2,0);	
			clusterGrid.directory[x2][y2][0].AddMovingClusterEntry(clusterGridEntry);
			newCluster.gridCellsOverlap.add(new GridCellIndex3D(x2,y2,0));
    	}
    	}
    	return newCluster;
    }

}
