/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Dec 23, 2005
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.clustersheddy;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba.MotionSCUBAClusterGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba.MotionSCUBAClusterHome;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba.MotionSCUBAClusterStorage;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.INode;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.INodeIndex;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.ResultPackage;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.ResultPackageImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.SinglePWSingleEquiJoinState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.SinglePWSingleEquiJoinStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.TNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.TupleQueue.Node;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * @author Rimma V. Nehme
 *
 */


public class SingleClusterSheddyStateImp 
extends OperatorStateImp  implements SingleClusterSheddyJoinState 
{
	
	protected MotionSCUBAClusterHome objectsHash;
	protected MotionSCUBAClusterHome queriesHash;
	
	//stores mapping between objects and queries and the clusters
	//moving entities are stored as follows:
	//moving objects --> 'o'+o.id, cid (e.g., o11,123 mean object w/ id 11 belongs to cluster w/ id 123)
	//moving queries --> 'q'+q.id, cid (e.g., q25,123 mean query w/ id 25 belongs to cluster w/ id 123)
	protected MotionSCUBAClusterHome clusterHome; 
	
	//stores spatio-temporal properties of all clusters
	protected MotionSCUBAClusterStorage clusterStorage; 
	//grid (hash table) used for clustering (to speed up the process) and joining
	protected MotionSCUBAClusterGrid clusterGrid; 	
	
	public SingleClusterSheddyStateImp()
	{
		super();
		objectsHash = new MotionSCUBAClusterHome();
		queriesHash = new MotionSCUBAClusterHome();
		clusterHome = new MotionSCUBAClusterHome(); 
		clusterStorage = new MotionSCUBAClusterStorage(); 
		clusterGrid = new MotionSCUBAClusterGrid(1, 1, 1); 	
	}

	public void createGridWithSpecifiedDimensions(int xDim, int yDim, int tDim) 
	{
		clusterGrid = new MotionSCUBAClusterGrid(xDim, yDim, tDim); 			
	}
	
	public void copyStateContent(OperatorState newstate) {
		// TODO Auto-generated method stub
		
	}

	public void clearState() {
		// TODO Auto-generated method stub
		
	}

	public MotionSCUBAClusterHome getClusterHome() 
	{
		return this.clusterHome;
	}

	public MotionSCUBAClusterHome getObjectsHash() 
	{	
		return this.objectsHash;
	}

	public MotionSCUBAClusterHome getQueriesHash() 
	{	
		return this.queriesHash;
	}

	public MotionSCUBAClusterStorage getClusterStorage() 
	{	
		return this.clusterStorage;
	}

	public MotionSCUBAClusterGrid getClusterGrid() 
	{	
		return this.clusterGrid;
	}
	
					
}
