/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////

package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion;

/**
 * @author Rimma V. Nehme
 *
 */
public class MotionStatisticsGatherer implements java.io.Serializable
{
	

	public int executionCount;
	public int numberOfObjectsInTheSystem;
	public int numberOfQueriesInTheSystem;
	public int intervalTimeStart;
	public int intervalTimeEnd;
	
	public int numberOfIntersectionResults;
	public int numberOfMergedResults;
	public int numberOfObjectsEnteredSystem;
	public int numberOfObjectsLeftSystem;
	public int numberOfQueriesEnteredSystem;
	public int numberOfQueriesLeftSystem;
	public int numberOfObjectsLocationUpdates;
	public int numberOfQueriesLocationUpdates;
	public int numberOfObjectsBeingJoined;
	public int numberOfQueriesBeingJoined;

	public int tuplesObjectsDequeued;
	public int tuplesQueriesDequeued;
	public int tuplesObjectsDropped;
	public int tuplesObjectsKept;
	public int tuplesObjectsReturned;
	public int tuplesQueriesDropped;
	public int tuplesQueriesKept;
	public int tuplesQueriesReturned;
	
	public double timeFakeQObjectsStart;
	public double timeFakeQQueriesStart;
	public double timeFakeQObjectsEnd;
	public double timeFakeQQueriesEnd;
	
	public double timeOverallStart;
	public double timeOverallEnd;
	public double timeInitializationStart;
	public double timeInitializationEnd;
	public double timeObjectsUpdateStart;
	public double timeObjectsUpdateEnd;
	public double timeQueriesUpdateStart;
	public double timeQueriesUpdateEnd;
	public double timeJoinStart;
	public double timeJoinEnd;
	public double timeMergeStart;
	public double timeMergeEnd;
	public double timeTimelineStart;
	public double timeTimelineEnd;
	int numberIntersections;
	public long memoryBefore;
	public long memoryAfter;

	public static int totalTimeElapsed = 0;
	public static int totalTimeElapsedObjectsUpdate = 0;
	public static int totalTimeElapsedQueriesUpdate = 0;
	public static int totalTimeUpdate = 0;
	
	public static int totalNumberObjectsEnteredIntoHashTable = 0;
	public static int totalNumberQueriesEnteredIntoHashTable = 0;
	public static int totalNumberObjectsRemovedFromHashTable = 0;
	public static int totalNumberQueriesRemovedFromHashTable = 0;
	
	public static int totalNumberObjectsLocationUpdates = 0;
	public static int totalNumberQueriesLocationUpdates = 0;
	public static int numberObjectsLocationUpdatesNotParticipatingInJoin = 0;
	public static int numberQueriesLocationUpdatesNotParticipatingInJoin = 0;

	
	public void ResetStatistics()
	{
		totalTimeElapsed = 0;
		totalTimeElapsedObjectsUpdate = 0;
		totalTimeElapsedQueriesUpdate = 0;
		totalTimeUpdate = 0;
		totalNumberObjectsLocationUpdates = 0;
		totalNumberQueriesLocationUpdates = 0;
		
		totalNumberObjectsEnteredIntoHashTable = 0;
		totalNumberQueriesEnteredIntoHashTable = 0;
		totalNumberObjectsRemovedFromHashTable = 0;
		totalNumberQueriesRemovedFromHashTable = 0;
		numberObjectsLocationUpdatesNotParticipatingInJoin = 0;
		numberQueriesLocationUpdatesNotParticipatingInJoin = 0;
	}
	
	
	public static void giveGarbageCollectionAChance() 
	{
		System.gc();
		try 
		{
			Thread.sleep(20);
		} 
		catch (InterruptedException e) 
		{
			// ignore - should not happen
		}
	}

	public static long getUsedMemory() 
	{
		Runtime r = Runtime.getRuntime();
		return r.totalMemory() - r.freeMemory();
	}
	
	
	public String ContinuousStatisticsToString()
	{
		String rs="";

 	    rs="\n";
		rs = "\nExecution interval [" + executionCount + "]\n";
		rs += "Total time elapsed ["+(totalTimeElapsed)+"]\n";
		rs += "Corresponding time interval ["+intervalTimeStart+"-"+ intervalTimeEnd+"]\n";	
		rs += "Total time elapsed - objects update ["+(totalTimeElapsedObjectsUpdate)+"]\n";
		rs += "Total time elapsed - queries update ["+(totalTimeElapsedQueriesUpdate)+"]\n";
		rs += "Total time elapsed - update ["+(totalTimeElapsedObjectsUpdate + totalTimeElapsedQueriesUpdate)+"]\n";
		rs += "Total time elapsed - join ["+(timeJoinEnd - timeJoinStart)+"]\n";
		rs += "Total time elapsed - merge ["+(timeMergeEnd - timeMergeStart)+"]\n";
		rs += "Total time elapsed - timeline ["+(timeTimelineEnd - timeTimelineStart)+"]\n";
		rs += "Total Number of objects location updates["+totalNumberObjectsLocationUpdates+"]\n";
		rs += "Total Number of queries location updates["+totalNumberQueriesLocationUpdates+"]\n";
		
		rs += "Total Number of objects location updates NOT participating in JOIN ["+numberObjectsLocationUpdatesNotParticipatingInJoin+"]\n";
		rs += "Total Number of queries location updates NOT participating in JOIN ["+numberQueriesLocationUpdatesNotParticipatingInJoin+"]\n";
		
		
		rs += "Number of objects segments being joined ["+numberOfObjectsBeingJoined+"]\n";
		rs += "Number of queries segments being joined ["+numberOfQueriesBeingJoined+"]\n";
		
		
		rs += "Total number of intersection results - after join ["+numberOfIntersectionResults+"]\n";
		rs += "Total number of results - after merging ["+numberOfMergedResults+"]\n";
		rs += "Total number of objects in the system ["+(totalNumberObjectsEnteredIntoHashTable - totalNumberObjectsRemovedFromHashTable)+"]\n";
		rs += "Total number of queries in the system ["+(totalNumberQueriesEnteredIntoHashTable - totalNumberQueriesRemovedFromHashTable)+"]\n";		
		/*
		rs += "Number of objects entered the system ["+numberOfObjectsEnteredSystem+"]\n";
		rs += "Number of objects left the system ["+numberOfObjectsLeftSystem+"]\n";
		rs += "Number of queries entered the system ["+numberOfQueriesEnteredSystem+"]\n";
		rs += "Number of queries left the system ["+numberOfQueriesLeftSystem+"]\n";
		rs += "Number of location updates for objects ["+numberOfObjectsLocationUpdates+"]\n";
		rs += "Number of location updates for queries ["+numberOfQueriesLocationUpdates+"]\n";
		
		rs += "Tuples for objects dequeued ["+tuplesObjectsDequeued+"]\n";
		rs += "Tuples for queries dequeued ["+tuplesQueriesDequeued+"]\n";

		rs += "Tuples for objects kept ["+tuplesObjectsKept+"]\n";
		rs += "Tuples for queries kept ["+tuplesQueriesKept+"]\n";
		
		rs += "Tuples for objects dropped ["+tuplesObjectsDropped+"]\n";
		rs += "Tuples for queries dropped ["+tuplesQueriesDropped+"]\n";
		
		rs += "Tuples for objects returned ["+tuplesObjectsReturned+"]\n";
		rs += "Tuples for queries returned ["+tuplesQueriesReturned+"]\n";
	    */
		ResetStatistics();
		
		return rs;
	}
	
	
	public String ShortContinuousStatisticsToString()
	{
		String rs="";

 	    rs="\n";
		rs = "" + executionCount + "\t";
		rs += ""+(totalTimeElapsed)+"\t";
		rs += ""+intervalTimeStart+"-"+ intervalTimeEnd+"\t";	
		rs += ""+(totalTimeElapsedObjectsUpdate)+"\t";
		rs += ""+(totalTimeElapsedQueriesUpdate)+"\t";
		rs += ""+(totalTimeElapsedObjectsUpdate + totalTimeElapsedQueriesUpdate)+"\t";
		rs += ""+(timeJoinEnd - timeJoinStart)+"\t";
		rs += ""+(timeMergeEnd - timeMergeStart)+"\t";
		rs += ""+(timeTimelineEnd - timeTimelineStart)+"\t";
		rs += ""+totalNumberObjectsLocationUpdates+"\t";
		rs += ""+totalNumberQueriesLocationUpdates+"\t";
		
		rs += ""+numberObjectsLocationUpdatesNotParticipatingInJoin+"\t";
		rs += ""+numberQueriesLocationUpdatesNotParticipatingInJoin+"\t";
		
		
		rs += ""+numberOfObjectsBeingJoined+"\t";
		rs += ""+numberOfQueriesBeingJoined+"\t";
		
		
		rs += ""+numberOfIntersectionResults+"\t";
		rs += ""+numberOfMergedResults+"\t";
		rs += ""+(totalNumberObjectsEnteredIntoHashTable - totalNumberObjectsRemovedFromHashTable)+"\t";
		rs += ""+(totalNumberQueriesEnteredIntoHashTable - totalNumberQueriesRemovedFromHashTable)+"\t";
		rs += ""+(memoryBefore)+"\t";
		rs += ""+(memoryAfter)+"\t";
		
		rs += ""+ (timeFakeQObjectsEnd - timeFakeQObjectsStart)+"\t";
		rs += ""+ (timeFakeQQueriesEnd - timeFakeQQueriesStart)+"";
		
		/*
		rs += "Number of objects entered the system ["+numberOfObjectsEnteredSystem+"]\n";
		rs += "Number of objects left the system ["+numberOfObjectsLeftSystem+"]\n";
		rs += "Number of queries entered the system ["+numberOfQueriesEnteredSystem+"]\n";
		rs += "Number of queries left the system ["+numberOfQueriesLeftSystem+"]\n";
		rs += "Number of location updates for objects ["+numberOfObjectsLocationUpdates+"]\n";
		rs += "Number of location updates for queries ["+numberOfQueriesLocationUpdates+"]\n";
		
		rs += "Tuples for objects dequeued ["+tuplesObjectsDequeued+"]\n";
		rs += "Tuples for queries dequeued ["+tuplesQueriesDequeued+"]\n";

		rs += "Tuples for objects kept ["+tuplesObjectsKept+"]\n";
		rs += "Tuples for queries kept ["+tuplesQueriesKept+"]\n";
		
		rs += "Tuples for objects dropped ["+tuplesObjectsDropped+"]\n";
		rs += "Tuples for queries dropped ["+tuplesQueriesDropped+"]\n";
		
		rs += "Tuples for objects returned ["+tuplesObjectsReturned+"]\n";
		rs += "Tuples for queries returned ["+tuplesQueriesReturned+"]\n";
	    */
		ResetStatistics();
		
		return rs;
	}
	

	public String ShortDiscreteStatisticsToString()
	{
		String rs="\n";
		rs = "" + executionCount + "\t";
		rs += ""+(totalTimeElapsed)+"\t";
		rs += ""+intervalTimeStart+"-"+ intervalTimeEnd+"\t";
		rs += ""+(totalTimeElapsedObjectsUpdate)+"\t";
		rs += ""+(totalTimeElapsedQueriesUpdate)+"\t";
		rs += ""+(totalTimeElapsedObjectsUpdate + totalTimeElapsedQueriesUpdate)+"\t";
		rs += ""+(timeJoinEnd - timeJoinStart)+"\t";
		rs += ""+numberOfObjectsBeingJoined+"\t";
		rs += ""+numberOfQueriesBeingJoined+"\t";
		rs += ""+numberOfIntersectionResults+"\t";		
		
		
		rs += ""+totalNumberObjectsLocationUpdates+"\t";
		rs += ""+totalNumberQueriesLocationUpdates+"\t";
		
		rs += ""+numberObjectsLocationUpdatesNotParticipatingInJoin+"\t";
		rs += ""+numberQueriesLocationUpdatesNotParticipatingInJoin+"\t";
		
		
		rs += ""+numberOfObjectsBeingJoined+"\t";
		rs += ""+numberOfQueriesBeingJoined+"\t";
		
		rs += ""+(totalNumberObjectsEnteredIntoHashTable - totalNumberObjectsRemovedFromHashTable)+"\t";
		rs += ""+(totalNumberQueriesEnteredIntoHashTable - totalNumberQueriesRemovedFromHashTable)+"\t";
		rs += ""+(memoryBefore)+"\t";
		rs += ""+(memoryAfter)+"\t";
		
		rs += ""+ (timeFakeQObjectsEnd - timeFakeQObjectsStart)+"\t";
		rs += ""+ (timeFakeQQueriesEnd - timeFakeQQueriesStart)+"";
	
		
		/*rs += "Total number of objects in the system ["+numberOfObjectsInTheSystem+"]\n";
		rs += "Total number of queries in the system ["+numberOfQueriesInTheSystem+"]\n";		
		rs += "Number of objects entered the system ["+numberOfObjectsEnteredSystem+"]\n";
		rs += "Number of objects left the system ["+numberOfObjectsLeftSystem+"]\n";
		rs += "Number of queries entered the system ["+numberOfQueriesEnteredSystem+"]\n";
		rs += "Number of queries left the system ["+numberOfQueriesLeftSystem+"]\n";
		rs += "Number of location updates for objects ["+numberOfObjectsLocationUpdates+"]\n";
		rs += "Number of location updates for queries ["+numberOfQueriesLocationUpdates+"]\n";
		
		rs += "Tuples for objects dequeued ["+tuplesObjectsDequeued+"]\n";
		rs += "Tuples for queries dequeued ["+tuplesQueriesDequeued+"]\n";

		rs += "Tuples for objects kept ["+tuplesObjectsKept+"]\n";
		rs += "Tuples for queries kept ["+tuplesQueriesKept+"]\n";
		
		rs += "Tuples for objects dropped ["+tuplesObjectsDropped+"]\n";
		rs += "Tuples for queries dropped ["+tuplesQueriesDropped+"]\n";
		
		rs += "Tuples for objects returned ["+tuplesObjectsReturned+"]\n";
		rs += "Tuples for queries returned ["+tuplesQueriesReturned+"]\n";
		*/
		ResetStatistics();
		return rs;
	}
	
	public String DiscreteStatisticsToString()
	{
		String rs="\n";
		rs = "\nExecution interval [" + executionCount + "]\n";
		rs += "Total time elapsed ["+(totalTimeElapsed)+"]\n";
		rs += "Corresponding time interval ["+intervalTimeStart+"-"+ intervalTimeEnd+"]\n";
		rs += "Total time elapsed - objects update ["+(totalTimeElapsedObjectsUpdate)+"]\n";
		rs += "Total time elapsed - queries update ["+(totalTimeElapsedQueriesUpdate)+"]\n";
		rs += "Total time elapsed - update ["+(totalTimeElapsedObjectsUpdate + totalTimeElapsedQueriesUpdate)+"]\n";
		rs += "Total time elapsed - join ["+(timeJoinEnd - timeJoinStart)+"]\n";
		rs += "Number of objects being joined ["+numberOfObjectsBeingJoined+"]\n";
		rs += "Number of queries being joined ["+numberOfQueriesBeingJoined+"]\n";
		rs += "Total number of intersection results - after join ["+numberOfIntersectionResults+"]\n";		
		
		
		rs += "Total Number of objects location updates["+totalNumberObjectsLocationUpdates+"]\n";
		rs += "Total Number of queries location updates["+totalNumberQueriesLocationUpdates+"]\n";
		
		rs += "Total Number of objects location updates NOT participating in JOIN ["+numberObjectsLocationUpdatesNotParticipatingInJoin+"]\n";
		rs += "Total Number of queries location updates NOT participating in JOIN ["+numberQueriesLocationUpdatesNotParticipatingInJoin+"]\n";
		
		
		rs += "Number of objects tuples being joined ["+numberOfObjectsBeingJoined+"]\n";
		rs += "Number of queries tuples being joined ["+numberOfQueriesBeingJoined+"]\n";
		
		rs += "Total number of objects in the system ["+(totalNumberObjectsEnteredIntoHashTable - totalNumberObjectsRemovedFromHashTable)+"]\n";
		rs += "Total number of queries in the system ["+(totalNumberQueriesEnteredIntoHashTable - totalNumberQueriesRemovedFromHashTable)+"]\n";		
	
		
		/*rs += "Total number of objects in the system ["+numberOfObjectsInTheSystem+"]\n";
		rs += "Total number of queries in the system ["+numberOfQueriesInTheSystem+"]\n";		
		rs += "Number of objects entered the system ["+numberOfObjectsEnteredSystem+"]\n";
		rs += "Number of objects left the system ["+numberOfObjectsLeftSystem+"]\n";
		rs += "Number of queries entered the system ["+numberOfQueriesEnteredSystem+"]\n";
		rs += "Number of queries left the system ["+numberOfQueriesLeftSystem+"]\n";
		rs += "Number of location updates for objects ["+numberOfObjectsLocationUpdates+"]\n";
		rs += "Number of location updates for queries ["+numberOfQueriesLocationUpdates+"]\n";
		
		rs += "Tuples for objects dequeued ["+tuplesObjectsDequeued+"]\n";
		rs += "Tuples for queries dequeued ["+tuplesQueriesDequeued+"]\n";

		rs += "Tuples for objects kept ["+tuplesObjectsKept+"]\n";
		rs += "Tuples for queries kept ["+tuplesQueriesKept+"]\n";
		
		rs += "Tuples for objects dropped ["+tuplesObjectsDropped+"]\n";
		rs += "Tuples for queries dropped ["+tuplesQueriesDropped+"]\n";
		
		rs += "Tuples for objects returned ["+tuplesObjectsReturned+"]\n";
		rs += "Tuples for queries returned ["+tuplesQueriesReturned+"]\n";
		*/
		ResetStatistics();
		return rs;
	}
	
	public String ShortDiscreteStatisticsHeaderToString()
	{
		String rs="";
		rs = "Exec count\t";
		rs += "Total time elapsed\t";
		rs += "Time interval\t";
		rs += "Total time elapsed - objects update\t";
		rs += "Total time elapsed - queries update\t";
		rs += "Total time elapsed - update\t";
		rs += "Total time elapsed - join\t";
		rs += "# objects being joined\t";
		rs += "# queries being joined\t";
		rs += "# of intersection results - after join\t";		
		
		
		rs += "# objects location updates\t";
		rs += "# of queries location updates\t";
		
		rs += "# objects location updates NOT participating in JOIN\t";
		rs += "# queries location updates NOT participating in JOIN\t";
		
		
		rs += "# objects tuples being joined\t";
		rs += "# of queries tuples being joined\t";
		
		rs += "Total # objects in system\t";
		rs += "Total # queries in system\t";
		
		rs += "Memory before\t";
		rs += "Memory after\t";

		rs += "Time spent on objects fake queue\t";
		rs += "Time spent on queries fake queue";
		
		return rs;
	}
	
}