/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion;

/**
 * @author Rimma V. Nehme
 *
 */
import java.util.HashMap;
import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;

public class MultiMap extends HashMap 
{

    // suppose if someone puts an ArrayList value to a key?
    // so, use a derived ArrayList to store a multi-value instead.
    private class MultiMapList extends ArrayList {
    }

    public Object put(Object key, Object value) {
        Object existingValue = get(key);

        // no value yet: just put the single value
        if (existingValue == null) {
            super.put(key, value);
        }

        // already several values: add the value to the MultiMapList
        else if (existingValue instanceof MultiMapList) {
            ((List)existingValue).add(value);
            return value;
        }

        // already ONE value: create a MultiMapList and add both the
        // existing and the new value to the list
        else {
            MultiMapList list = new MultiMapList();
            list.add(get(key));
            list.add(value);
            super.put(key, list);
        }
        return value;
    }

    public int getValueCount(Object key) {
        Object value = get(key);

        if (value == null) return 0;
        if (value instanceof MultiMapList) return ((List)value).size();
        return 1;
    }

    public Iterator getValueIterator(Object key) {
        int count = getValueCount(key);

        if (count == 0) return new Iterator() {
            public boolean hasNext() { return false; }
            public Object next() { return null; }
            public void remove() { throw new UnsupportedOperationException(); }
        };

        if (count == 1) {
            final Object value = get(key);
            return new Iterator() {
                private boolean done = false;
                public boolean hasNext() { return !done; }
                public Object next() {
                    if (done) return null;
                    done = true;
                    return value;
                }
                public void remove() { throw new UnsupportedOperationException(); }
            };
        }

        return ((List)get(key)).iterator();
    }

    // test the class Multimap. Inserts four string values
    // to the key "key" and print the contents (to test the iterators)
  /*
    public static void main(String[] args) {
        MultiMap map = new MultiMap();

        System.out.println("key: #" + map.getValueCount("key") + ", "
            + valuesToString(map.getValueIterator("key")));

        map.put("key", "value1");

        System.out.println("key: #" + map.getValueCount("key") + ", "
            + valuesToString(map.getValueIterator("key")));

        map.put("key", "value2");

        System.out.println("key: #" + map.getValueCount("key") + ", "
            + valuesToString(map.getValueIterator("key")));

        map.put("key", "value3");

        System.out.println("key: #" + map.getValueCount("key") + ", "
            + valuesToString(map.getValueIterator("key")));

        map.put("key", "value4");

        System.out.println("key: #" + map.getValueCount("key") + ", "
            + valuesToString(map.getValueIterator("key")));
    }
*/
    

    
    private static String valuesToString(Iterator iterator) {
        if (!iterator.hasNext()) return "<no values";

        StringBuffer sb = new StringBuffer();
        while (iterator.hasNext()) {
            sb.append(iterator.next().toString()).append(", ");
        }

        return sb.toString();
    }
}

