/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion;

/**
 * @author Rimma V. Nehme
 */
class MyPointCS extends PositionSpt2D
{
    // bits[0] = left
    // bits[1] = right
    // bits[2] = below
    // bits[3] = above 
	// bits[4] = far
    // bits[5] = near 
    public boolean[] bits = new boolean[6];


    MyPointCS()
    {
    	bits[0] = bits[1] = bits[2] = bits[3] = bits[4] = bits[5] = false;
    }

    MyPointCS(long x, long y, long t)
    {
    	this.x = x;
    	this.y = y;
    	this.t = t;
    }

    MyPointCS(long x, long y, float t, MyPointCS winMin, MyPointCS winMax)
    {
    	bits[0] = bits[1] = bits[2] = bits[3] = bits[4] = bits[5] = false;
    	setPos(x, y, t, winMin, winMax);
    }

    public boolean inside()
    {
    	return !(bits[0] || bits[1] || bits[2] || bits[3] || bits[4] || bits[5]);
    }

    public void encode(MyPointCS winMin, MyPointCS winMax)
    {
    	if (x < winMin.x)
    		bits[0] = true;
    	else
    		bits[0] = false;

    	if (x > winMax.x)
    		bits[1] = true;
    	else
    		bits[1] = false;

    	if (y > winMax.y)
    		bits[2] = true;
    	else
    		bits[2] = false;

    	if (y < winMin.y)
    		bits[3] = true;
    	else
    		bits[3] = false;
    	
    	//3rd dimension
    	if (t > winMax.t)
    		bits[4] = true;
    	else
    		bits[4] = false;

    	if (t < winMin.t)
    		bits[5] = true;
    	else
    		bits[5] = false;
    }

    public void setPos(long x, long y, float t, MyPointCS winMin, MyPointCS winMax)
    {
    	this.x = x;
    	this.y = y;
    	this.t = t;
    	encode(winMin, winMax);
    }

}
