/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid;


import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.*;

/**
 * @author Rimma V. Nehme
 *
 */
public class FixedContinuous3DGrid extends GenericSpatioTemporalGrid
{
	//debug flag
	public static int DEBUG = 0; //1=true, 0=false
	
	private int xDim;
	private int yDim;
	private int tDim;
	
	public GenericSpatioTemporalContinuousGridCell[][][] directory;

	public String toString()
	{
		return statisticsToString();
	}
	
	public void Cleanup()
	{
		for (int i=0; i < xDim; i++)
		{
			for (int j=0; j < yDim; j++)
			{
				for (int k=0; k < tDim; k++)
				{

					if(directory[i][j][k] != null)
					{
					if (directory[i][j][k].GetMovingObjects().size() > 0) 
					{
						//remove all objects from the cell
						directory[i][j][k].GetMovingObjects().clear();
					}
					if (directory[i][j][k].GetMovingQueries().size() > 0)
					{
						//remove all queries from the cell
						directory[i][j][k].GetMovingQueries().clear();				 
					}
					}
				}
			}
		}
	}
	
	public String statisticsToString()
	{
		int totalNumberOfObjects = 0;
		int totalNumberOfQueries = 0;
		String outputString="*** Grid Statistics ***";
		
		for (int i=0; i < xDim; i++)
		{
			for (int j=0; j < yDim; j++)
			{
				for (int k=0; k < tDim; k++)
				{
					if (directory[i][j][k] != null)
					{
					totalNumberOfObjects += directory[i][j][k].GetMovingObjects().size();
					totalNumberOfQueries += directory[i][j][k].GetMovingQueries().size();
					
					if ((directory[i][j][k].GetMovingObjects().size() > 0) || 
					    (directory[i][j][k].GetMovingQueries().size() > 0))
					{
						outputString += "\ngrid["+i+"]["+j+"]["+k+"]: # objects = "+directory[i][j][k].GetMovingObjects().size()+", # queries = "+directory[i][j][k].GetMovingQueries().size();
						outputString +=  " [ Cell Dimension: i = "+ (i * GenericSpatioTemporalGrid.xDimensionSize)+ ", "+((i+1) * GenericSpatioTemporalGrid.xDimensionSize)+"; "+
														 "j = "+ (j * GenericSpatioTemporalGrid.yDimensionSize)+ ", "+((j+1) * GenericSpatioTemporalGrid.yDimensionSize)+"; "+														 
														 "k = " + (k * GenericSpatioTemporalGrid.tDimensionSize)+", "+((k+1) * GenericSpatioTemporalGrid.tDimensionSize)+";] ";
						 
					}
					}
				}
			}
		}
		
		outputString += "\nTotal # of object entries = "+totalNumberOfObjects+", Total # of queries entries = "+totalNumberOfQueries +"\n";
		return outputString;
	}
	
	public FixedContinuous3DGrid(int xDimension, int yDimension, int tDimension) 
	{
		xDim = xDimension;
		yDim = yDimension;
		tDim = tDimension;
		
		directory = new GenericSpatioTemporalContinuousGridCell[xDimension][yDimension][tDimension];
			
	}


	
	public void InsertObject(MovingObjectContinuousState mc, boolean deleteFromSystem, Time currentTime, int delta)	
	{
		boolean isInTheSameCell = true;
		int report_number = 1;
		int xStartIndex = GenericGrid.GetCellIndex(mc.GetStartPosition().GetX(), DimensionType.x);		
		int yStartIndex = GenericGrid.GetCellIndex(mc.GetStartPosition().GetY(), DimensionType.y);		
		int tStartIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)

		int xEndIndex = GenericGrid.GetCellIndex(mc.GetEndPosition().GetX(), DimensionType.x);		
		int yEndIndex = GenericGrid.GetCellIndex(mc.GetEndPosition().GetY(), DimensionType.y);		
		int tEndIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)
		int []sortedIndexX = new int[2];
		int []sortedIndexY = new int[2];
		
		
		CheckCreateGridCell(xStartIndex,yStartIndex,tStartIndex);
		CheckCreateGridCell(xEndIndex,yEndIndex,tEndIndex);
				
		
		if ((xStartIndex == xEndIndex) && (yStartIndex == yEndIndex) && (tStartIndex == tEndIndex))
		{
			isInTheSameCell = true;	
		}
		else
		{
			isInTheSameCell = false;
	
			
			//figure out the difference between the cells
			if (xStartIndex > xEndIndex)
			{
				sortedIndexX[0]=xEndIndex;
				sortedIndexX[1]=xStartIndex;
			}
			else
			{
				sortedIndexX[0]=xStartIndex;
				sortedIndexX[1]=xEndIndex;				
			}
	
			if (yStartIndex > yEndIndex)
			{
				sortedIndexY[0]=yEndIndex;
				sortedIndexY[1]=yStartIndex;
			}
			else
			{
				sortedIndexY[0]=yStartIndex;
				sortedIndexY[1]=yEndIndex;				
			}
		}

		if ((!GenericSpatioTemporalGrid.IsIndexWithinGridBoundary(xStartIndex, yStartIndex, tStartIndex)) || 
			(!GenericSpatioTemporalGrid.IsIndexWithinGridBoundary(xEndIndex, yEndIndex, tEndIndex)))
		{
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("DELETING "+EntityType.FromEnumToString(mc.GetEntityType())+" "+mc.GetID()+" from the sytems");
			}
			deleteFromSystem = true;
			return; //for now just return without entering the entity
		}
		
		GridContinuousEntry gceObject = new GridContinuousEntry(mc.GetID(), mc.GetStartPosition().GetT(), mc.GetEndPosition().GetT());
		

			if (isInTheSameCell)
			{
				directory[xStartIndex][yStartIndex][tStartIndex].AddMovingObjectEntry(gceObject);				
			}
			else
			{
				for(int i=sortedIndexX[0]; i<=sortedIndexX[1];i++)
				{
					for(int j=sortedIndexY[0]; j<=sortedIndexY[1];j++)
					{
						//intersect the trajectory between the start and the end positions 
						//and the intermediate cells and make entries in these cells												
						
						
						PositionSpt2D cellWinMin = GenericSpatioTemporalGrid.GetCellWindowMin(new GridCellIndex3D(i,j,0));
						cellWinMin.SetT(currentTime.GetTime());
						PositionSpt2D cellWinMax = GenericSpatioTemporalGrid.GetCellWindowMax(new GridCellIndex3D(i,j,0));
						cellWinMax.SetT(currentTime.GetTime()+delta);
						
					
						MotionContinuousState mcClipped = Clipper.clipLine3D(-1,
															  mc.GetID(),
															  cellWinMin,
															  cellWinMax,			   								
															  mc.GetStartPosition(),
															  mc.GetEndPosition());
						
						
						
						
						if (mcClipped != null)
						{

							if(DEBUG == 1)
							{
								//print grid statistics	
								//Helper.DebugWriteLine("\nResulting Line = "+mcClipped.toString());
							}
							
							CheckCreateGridCell(i,j,0);							
							directory[i][j][0].AddMovingObjectEntry(gceObject);
						}
						else
						{
							if(DEBUG == 1)
							{
								//print grid statistics	
								Helper.DebugWriteLine("\nClip resulted in NULL");
							}
						}
					}
				}
			}
			
	}
	

	public void InsertQuery(MovingQueryContinuousState mc, boolean deleteFromSystem, Time currentTime, int delta)	
	{
		boolean isInTheSameCell = true;
		int report_number = 1;
		//get the query boundary
		int queryBoundary = mc.GetQueryBoundary();
				
		int xStartIndex = GenericGrid.GetCellIndex(mc.GetStartPosition().GetX(), DimensionType.x);		
		int yStartIndex = GenericGrid.GetCellIndex(mc.GetStartPosition().GetY(), DimensionType.y);		
		int tStartIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)

		int xEndIndex = GenericGrid.GetCellIndex(mc.GetEndPosition().GetX(), DimensionType.x);		
		int yEndIndex = GenericGrid.GetCellIndex(mc.GetEndPosition().GetY(), DimensionType.y);		
		int tEndIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)
		int []sortedIndexX = new int[2];
		int []sortedIndexY = new int[2];
		
		//check if the cell is initialized
		if (directory[xStartIndex][yStartIndex][tStartIndex] == null)
		{			
			directory[xStartIndex][yStartIndex][tStartIndex] = new GenericSpatioTemporalContinuousGridCell(xStartIndex,
					yStartIndex,
					1,
					xStartIndex * GenericSpatioTemporalGrid.xDimensionSize,
					yStartIndex * GenericSpatioTemporalGrid.yDimensionSize,
					 1 * GenericSpatioTemporalGrid.tDimensionSize,
					 (xStartIndex+1) * GenericSpatioTemporalGrid.xDimensionSize,
					 (yStartIndex+1) * GenericSpatioTemporalGrid.yDimensionSize,
					 (1+1) * GenericSpatioTemporalGrid.tDimensionSize);
		}
		
		
		//check if the cell is initialized
		if (directory[xEndIndex][yEndIndex][tEndIndex] == null)
		{			
			directory[xStartIndex][yStartIndex][tEndIndex] = new GenericSpatioTemporalContinuousGridCell(xEndIndex,
					yEndIndex,
					1,
					xEndIndex * GenericSpatioTemporalGrid.xDimensionSize,
					yEndIndex * GenericSpatioTemporalGrid.yDimensionSize,
					 1 * GenericSpatioTemporalGrid.tDimensionSize,
					 (xEndIndex+1) * GenericSpatioTemporalGrid.xDimensionSize,
					 (yEndIndex+1) * GenericSpatioTemporalGrid.yDimensionSize,
					 (1+1) * GenericSpatioTemporalGrid.tDimensionSize);
		}
		
		
		if ((xStartIndex == xEndIndex) && (yStartIndex == yEndIndex) && (tStartIndex == tEndIndex))
		{
			isInTheSameCell = true;	
		}
		else
		{
			isInTheSameCell = false;
			
			//figure out the difference between the cells
			if (xStartIndex > xEndIndex)
			{
				sortedIndexX[0]=xEndIndex;
				sortedIndexX[1]=xStartIndex;
			}
			else
			{
				sortedIndexX[0]=xStartIndex;
				sortedIndexX[1]=xEndIndex;				
			}
	
			if (yStartIndex > yEndIndex)
			{
				sortedIndexY[0]=yEndIndex;
				sortedIndexY[1]=yStartIndex;
			}
			else
			{
				sortedIndexY[0]=yStartIndex;
				sortedIndexY[1]=yEndIndex;				
			}
		}

		if ((!GenericSpatioTemporalGrid.IsIndexWithinGridBoundary(xStartIndex, yStartIndex, tStartIndex)) || 
			(!GenericSpatioTemporalGrid.IsIndexWithinGridBoundary(xEndIndex, yEndIndex, tEndIndex)))
		{
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("DELETING "+EntityType.FromEnumToString(mc.GetEntityType())+" "+mc.GetID()+" from the sytems");
			}
			deleteFromSystem = true;
			return; //for now just return without entering the entity
		}
		
		
		GridContinuousEntry gceQuery = new GridContinuousEntry(mc.GetID(), mc.GetStartPosition().GetT(), mc.GetEndPosition().GetT());
		

			if (isInTheSameCell)
			{
				directory[xStartIndex][yStartIndex][tStartIndex].AddMovingQueryEntry(gceQuery);
				
				PositionSpt2D cellWinMin = GenericSpatioTemporalGrid.GetCellWindowMin(new GridCellIndex3D(xStartIndex,yStartIndex,0));
				cellWinMin.SetT(currentTime.GetTime());
				PositionSpt2D cellWinMax = GenericSpatioTemporalGrid.GetCellWindowMax(new GridCellIndex3D(xStartIndex,yStartIndex,0));
				cellWinMax.SetT(currentTime.GetTime()+delta);
				
				MakeEntriesInNeighboringCells(xStartIndex,yStartIndex,mc, cellWinMin, cellWinMax, gceQuery, queryBoundary);
			}
			else
			{
				for(int i=sortedIndexX[0]; i<=sortedIndexX[1];i++)
				{
					for(int j=sortedIndexY[0]; j<=sortedIndexY[1];j++)
					{
						//intersect the trajectory between the start and the end positions 
						//and the intermediate cells and make entries in these cells						
						
						PositionSpt2D cellWinMin = GenericSpatioTemporalGrid.GetCellWindowMin(new GridCellIndex3D(i,j,0));
						cellWinMin.SetT(currentTime.GetTime());
						PositionSpt2D cellWinMax = GenericSpatioTemporalGrid.GetCellWindowMax(new GridCellIndex3D(i,j,0));
						cellWinMax.SetT(currentTime.GetTime()+delta);
										
						MotionContinuousState mcClipped = Clipper.clipLine3D(
															  -1,
															  mc.GetID(),
															  cellWinMin,
															  cellWinMax,			   								
															  mc.GetStartPosition(),
															  mc.GetEndPosition());
						
						if (mcClipped != null)
						{
							directory[i][j][0].AddMovingQueryEntry(gceQuery);
				  		    MakeEntriesInNeighboringCells(i,j,mcClipped, cellWinMin, cellWinMax, gceQuery, queryBoundary);							
						}
						else
						{
							if(DEBUG == 1)
							{
								Helper.DebugWriteLine("\nClip resulted in NULL");
							}
						}
					}
				}
			}
		
	}
	
	
	/*dp:
	public void MakeEntriesInNeighboringCells(
			int i, //cell index in the x-dimension
			int j, //cell index in the y-dimension
			MotionContinuousState mcClipped, 
			PositionSpt2D cellWinMin, 
			PositionSpt2D cellWinMax,
			MovingQueryContinuousState newMovingQueryEntry,
			int queryBoundary)
	{
//		for start position of the trajectory (check the possible boundary
		//intersections with other cells
		if ((mcClipped.GetStartPosition().GetX() + queryBoundary) > cellWinMax.GetX())
		{
			//Helper.DebugWriteLine("start pos + boundary = [" +(mcClipped.GetStartPosition().GetX() + queryBoundary)+" go beyond cellWinMax = [" + cellWinMax.GetX()+"].");
			directory[i+1][j][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i+1][j+1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i+1][j-1][0].AddMovingQueryEntry(newMovingQueryEntry);	
		}
		
		if ((mcClipped.GetStartPosition().GetX() - queryBoundary) < cellWinMin.GetX())
		{
			directory[i-1][j][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i-1][j+1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i-1][j-1][0].AddMovingQueryEntry(newMovingQueryEntry);	
		}
		
		if ((mcClipped.GetStartPosition().GetY() + queryBoundary) > cellWinMax.GetY())
		{
			directory[i-1][j+1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i][j+1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i+1][j+1][0].AddMovingQueryEntry(newMovingQueryEntry);	
		}
		
		if ((mcClipped.GetStartPosition().GetY() - queryBoundary) < cellWinMin.GetY())
		{
			directory[i-1][j-1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i][j-1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i+1][j-1][0].AddMovingQueryEntry(newMovingQueryEntry);	
		}
		
		
		//for start position of the trajectory (check the possible boundary
		//intersections with other cells
		if ((mcClipped.GetEndPosition().GetX() + queryBoundary) > cellWinMax.GetX())
		{
			directory[i+1][j][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i+1][j+1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i+1][j-1][0].AddMovingQueryEntry(newMovingQueryEntry);	
		}
		
		if ((mcClipped.GetEndPosition().GetX() - queryBoundary) < cellWinMin.GetX())
		{
			directory[i-1][j][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i-1][j+1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i-1][j-1][0].AddMovingQueryEntry(newMovingQueryEntry);	
		}
		
		if ((mcClipped.GetEndPosition().GetY() + queryBoundary) > cellWinMax.GetY())
		{
			directory[i-1][j+1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i][j+1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i+1][j+1][0].AddMovingQueryEntry(newMovingQueryEntry);	
		}
		
		if ((mcClipped.GetEndPosition().GetY() - queryBoundary) < cellWinMin.GetY())
		{
			directory[i-1][j-1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i][j-1][0].AddMovingQueryEntry(newMovingQueryEntry);	
			directory[i+1][j-1][0].AddMovingQueryEntry(newMovingQueryEntry);	
		}

	}
	*/
	
	
	
	private void CheckCreateGridCell(int i, int j, int t)
	{
		if (directory[i][j][t] == null)
		{			
			directory[i][j][t] = new GenericSpatioTemporalContinuousGridCell(i,
					j,
					t,
					i * GenericSpatioTemporalGrid.xDimensionSize,
					j * GenericSpatioTemporalGrid.yDimensionSize,
					 t * GenericSpatioTemporalGrid.tDimensionSize,
					 (i+1) * GenericSpatioTemporalGrid.xDimensionSize,
					 (j+1) * GenericSpatioTemporalGrid.yDimensionSize,
					 (t+1) * GenericSpatioTemporalGrid.tDimensionSize);
		}
		
	}

	
	public void MakeEntriesInNeighboringCells(
			int i, //cell index in the x-dimension
			int j, //cell index in the y-dimension
			MotionContinuousState mcClipped, 
			PositionSpt2D cellWinMin, 
			PositionSpt2D cellWinMax,
			GridContinuousEntry gceQuery,
			int queryBoundary)
	{
		//for start position of the trajectory (check the possible boundary
		//intersections with other cells
		if ((mcClipped.GetStartPosition().GetX() + queryBoundary) > cellWinMax.GetX())
		{
			//Helper.DebugWriteLine("start pos + boundary = [" +(mcClipped.GetStartPosition().GetX() + queryBoundary)+" go beyond cellWinMax = [" + cellWinMax.GetX()+"].");
			CheckCreateGridCell(i+1,j,0);
			CheckCreateGridCell(i+1,j+1,0);
			CheckCreateGridCell(i+1,j-1,0);
			
			directory[i+1][j][0].AddMovingQueryEntry(gceQuery);	
			directory[i+1][j+1][0].AddMovingQueryEntry(gceQuery);	
			directory[i+1][j-1][0].AddMovingQueryEntry(gceQuery);	
		}
		
		if ((mcClipped.GetStartPosition().GetX() - queryBoundary) < cellWinMin.GetX())
		{
			CheckCreateGridCell(i-1,j,0);
			CheckCreateGridCell(i-1,j+1,0);
			CheckCreateGridCell(i-1,j-1,0);
			
			directory[i-1][j][0].AddMovingQueryEntry(gceQuery);	
			directory[i-1][j+1][0].AddMovingQueryEntry(gceQuery);	
			directory[i-1][j-1][0].AddMovingQueryEntry(gceQuery);	
		}
		
		if ((mcClipped.GetStartPosition().GetY() + queryBoundary) > cellWinMax.GetY())
		{
			CheckCreateGridCell(i-1,j+1,0);
			CheckCreateGridCell(i,j+1,0);
			CheckCreateGridCell(i+1,j+1,0);
			
			directory[i-1][j+1][0].AddMovingQueryEntry(gceQuery);	
			directory[i][j+1][0].AddMovingQueryEntry(gceQuery);	
			directory[i+1][j+1][0].AddMovingQueryEntry(gceQuery);	
		}
		
		if ((mcClipped.GetStartPosition().GetY() - queryBoundary) < cellWinMin.GetY())
		{
			CheckCreateGridCell(i-1,j-1,0);
			CheckCreateGridCell(i,j-1,0);
			CheckCreateGridCell(i+1,j-1,0);
			
			directory[i-1][j-1][0].AddMovingQueryEntry(gceQuery);	
			directory[i][j-1][0].AddMovingQueryEntry(gceQuery);	
			directory[i+1][j-1][0].AddMovingQueryEntry(gceQuery);	
		}
		
		
		//for end position of the trajectory (check the possible boundary
		//intersections with other cells
		if ((mcClipped.GetEndPosition().GetX() + queryBoundary) > cellWinMax.GetX())
		{
			CheckCreateGridCell(i+1,j,0);
			CheckCreateGridCell(i+1,j+1,0);
			CheckCreateGridCell(i+1,j-1,0);
			
			directory[i+1][j][0].AddMovingQueryEntry(gceQuery);	
			directory[i+1][j+1][0].AddMovingQueryEntry(gceQuery);	
			directory[i+1][j-1][0].AddMovingQueryEntry(gceQuery);	
		}
		
		if ((mcClipped.GetEndPosition().GetX() - queryBoundary) < cellWinMin.GetX())
		{
			CheckCreateGridCell(i-1,j,0);
			CheckCreateGridCell(i-1,j+1,0);
			CheckCreateGridCell(i-1,j-1,0);
			
			directory[i-1][j][0].AddMovingQueryEntry(gceQuery);	
			directory[i-1][j+1][0].AddMovingQueryEntry(gceQuery);	
			directory[i-1][j-1][0].AddMovingQueryEntry(gceQuery);	
		}
		
		if ((mcClipped.GetEndPosition().GetY() + queryBoundary) > cellWinMax.GetY())
		{
			CheckCreateGridCell(i-1,j+1,0);
			CheckCreateGridCell(i,j+1,0);
			CheckCreateGridCell(i+1,j+1,0);
			
			directory[i-1][j+1][0].AddMovingQueryEntry(gceQuery);	
			directory[i][j+1][0].AddMovingQueryEntry(gceQuery);	
			directory[i+1][j+1][0].AddMovingQueryEntry(gceQuery);	
		}
		
		if ((mcClipped.GetEndPosition().GetY() - queryBoundary) < cellWinMin.GetY())
		{
			CheckCreateGridCell(i-1,j-1,0);
			CheckCreateGridCell(i,j-1,0);
			CheckCreateGridCell(i+1,j-1,0);
			
			directory[i-1][j-1][0].AddMovingQueryEntry(gceQuery);	
			directory[i][j-1][0].AddMovingQueryEntry(gceQuery);	
			directory[i+1][j-1][0].AddMovingQueryEntry(gceQuery);	
		}

	}
	
	public MotionContinuousResult DoJoin(MovingQueryContinuousState queryTrajectory, 
							  			 MovingObjectContinuousState objectTrajectory, 
										 GridCellIndex3D gi)
	{
		
		int queryBoundary = queryTrajectory.GetQueryBoundary();
		
		MotionContinuousResult intersectionResult = null;
		
		PositionSpt2D queryMin = queryTrajectory.GetStartPosition();
		PositionSpt2D queryMax = queryTrajectory.GetEndPosition();
		

		PositionSpt2D qMinWindow = new PositionSpt2D(queryMin.GetX() - queryBoundary,
													 queryMin.GetY() - queryBoundary,
													 queryMin.GetT());
		PositionSpt2D qMaxWindow = new PositionSpt2D(queryMax.GetX() + queryBoundary,
				 								     queryMax.GetY() + queryBoundary,
													 queryMax.GetT());
		
		//checkMinsAndMaxForQuery(qMinWindow, qMaxWindow, queryBoundary);
		

		
		//queryMin.SetX(queryMin.GetX() - queryBoundary);
		//queryMin.SetY(queryMin.GetY() - queryBoundary);
		//queryMax.SetX(queryMax.GetX() + queryBoundary);
		//queryMax.SetY(queryMax.GetY() + queryBoundary);
		
		//Helper.DebugWriteLine("\nAdjusting boundaries of the query min.x = "+queryMin.GetX()+" now min.x = " + qMinWindow.GetX());
		//Helper.DebugWriteLine("Adjusting boundaries of the query min.y = "+queryMin.GetY()+" now min.y = " + qMinWindow.GetY());
		//Helper.DebugWriteLine("Adjusting boundaries of the query max.x = "+queryMax.GetX()+" now max.x = "+qMaxWindow.GetX());
		//Helper.DebugWriteLine("Adjusting boundaries of the query max.y = "+queryMax.GetY()+" now max.y = "+qMaxWindow.GetY());		
		
		//Helper.DebugWriteLine("\nClipping: query = "+queryTrajectory.GetID()+" with object = "+objectTrajectory.GetID()+"\nqueryMin = "+qMinWindow + "\nqueryMax = "+qMaxWindow+"\nobjectTrajectory.GetStartPosition() = " +objectTrajectory.GetStartPosition()+"\nobjectTrajectory.GetEndPosition() = " + objectTrajectory.GetEndPosition());
		
		//System.out.println("\nClipping: query = "+queryTrajectory.GetID()+" with object = "+objectTrajectory.GetID()+"\nqueryMin = "+qMinWindow + "\nqueryMax = "+qMaxWindow+"\nobjectTrajectory.GetStartPosition() = " +objectTrajectory.GetStartPosition()+"\nobjectTrajectory.GetEndPosition() = " + objectTrajectory.GetEndPosition());
		MotionContinuousState mcClippedLine = Clipper.clipLine3D(queryTrajectory.GetID(),
																	objectTrajectory.GetID(),
																	qMinWindow,
																	qMaxWindow,
																    objectTrajectory.GetStartPosition(),
																    objectTrajectory.GetEndPosition());
				
				
		if (mcClippedLine != null)
		{	
			
			intersectionResult = new MotionContinuousResult(queryTrajectory.GetID(),
															objectTrajectory.GetID(),
															QueryType.Range,
															mcClippedLine);
			//Helper.DebugWriteLine("Result of 3D intersection = "+intersectionResult.toString());
			//System.out.println("Result of 3D intersection = "+intersectionResult.toString());
		}
		else
		{
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("NO Result for 3D intersection!");
			}
			//System.out.println("NO Result for 3D intersection!");
		}
		return intersectionResult;
	}
	
	public void checkMinsAndMaxForQuery(PositionSpt2D p1, PositionSpt2D p2, int queryBoundary)
	 {
	 	PositionSpt2D min = new PositionSpt2D();
	 	PositionSpt2D max = new PositionSpt2D();
	 	
	 	if((p2.GetX() > p1.GetX()) && (p2.GetY() > p1.GetY()))
	 	{
	 		long xMin = p1.GetX() - queryBoundary; 
	 		long yMin = p1.GetY() - queryBoundary;
	 		
	 		long xMax = p2.GetX() + queryBoundary; 
	 		long yMax = p2.GetY() + queryBoundary;
	 		
	 		min = new PositionSpt2D(xMin, yMin, p1.GetT());
	 		max = new PositionSpt2D(xMax, yMax, p2.GetT());
	 		
	 		//re-assign min and max
	 		p1.SetX(min.GetX());
	 		p1.SetY(min.GetY());
	 		p1.SetT(min.GetT());	 		
	 		p2.SetX(max.GetX());
	 		p2.SetY(max.GetY());
	 		p2.SetT(max.GetT());
	 	}
	 	else if((p2.GetX() > p1.GetX()) && (p2.GetY() == p1.GetY()))
	 	{
	 		long xMin = p1.GetX() - queryBoundary; 
	 		long yMin = p1.GetY() - queryBoundary;
	 		
	 		long xMax = p2.GetX() + queryBoundary; 
	 		long yMax = p2.GetY() + queryBoundary;
	 		
	 		min = new PositionSpt2D(xMin, yMin, p1.GetT());
	 		max = new PositionSpt2D(xMax, yMax, p2.GetT());
	 		
	 		//re-assign min and max
	 		p1.SetX(min.GetX());
	 		p1.SetY(min.GetY());
	 		p1.SetT(min.GetT());	 		
	 		p2.SetX(max.GetX());
	 		p2.SetY(max.GetY());
	 		p2.SetT(max.GetT());
	 	}
	 	else if((p2.GetX() > p1.GetX()) && (p2.GetY() < p1.GetY()))
	 	{
	 		long xMin = p2.GetX() + queryBoundary; 
	 		long yMin = p2.GetY() - queryBoundary;
	 		
	 		long xMax = p1.GetX() - queryBoundary; 
	 		long yMax = p1.GetY() + queryBoundary;
	 		
	 		min = new PositionSpt2D(xMin, yMin, p2.GetT());
	 		max = new PositionSpt2D(xMax, yMax, p1.GetT());
	 		
	 		//re-assign min and max
	 		p1.SetX(min.GetX());
	 		p1.SetY(min.GetY());
	 		p1.SetT(min.GetT());	 		
	 		p2.SetX(max.GetX());
	 		p2.SetY(max.GetY());
	 		p2.SetT(max.GetT());
	 	}
	 	
	 	else if((p2.GetX() == p1.GetX()) && (p2.GetY() < p1.GetY()))
	 	{
	 		long xMin = p2.GetX() - queryBoundary; 
	 		long yMin = p2.GetY() - queryBoundary;
	 		
	 		long xMax = p1.GetX() + queryBoundary; 
	 		long yMax = p1.GetY() + queryBoundary;
	 		
	 		min = new PositionSpt2D(xMin, yMin, p2.GetT());
	 		max = new PositionSpt2D(xMax, yMax, p1.GetT());
	 		
	 		//re-assign min and max
	 		p1.SetX(min.GetX());
	 		p1.SetY(min.GetY());
	 		p1.SetT(min.GetT());	 		
	 		p2.SetX(max.GetX());
	 		p2.SetY(max.GetY());
	 		p2.SetT(max.GetT());
	 	}
	 	else if((p2.GetX() < p1.GetX()) && (p2.GetY() < p1.GetY()))
	 	{
	 		long xMin = p2.GetX() - queryBoundary; 
	 		long yMin = p2.GetY() - queryBoundary;
	 		
	 		long xMax = p1.GetX() + queryBoundary; 
	 		long yMax = p1.GetY() + queryBoundary;
	 		
	 		min = new PositionSpt2D(xMin, yMin, p2.GetT());
	 		max = new PositionSpt2D(xMax, yMax, p1.GetT());
	 		
	 		//re-assign min and max
	 		p1.SetX(min.GetX());
	 		p1.SetY(min.GetY());
	 		p1.SetT(min.GetT());	 		
	 		p2.SetX(max.GetX());
	 		p2.SetY(max.GetY());
	 		p2.SetT(max.GetT());
	 	}
	 	else if((p2.GetX() < p1.GetX()) && (p2.GetY() == p1.GetY()))
	 	{
	 		long xMin = p2.GetX() - queryBoundary; 
	 		long yMin = p2.GetY() - queryBoundary;
	 		
	 		long xMax = p1.GetX() + queryBoundary; 
	 		long yMax = p1.GetY() + queryBoundary;
	 		
	 		min = new PositionSpt2D(xMin, yMin, p2.GetT());
	 		max = new PositionSpt2D(xMax, yMax, p1.GetT());
	 		
	 		//re-assign min and max
	 		p1.SetX(min.GetX());
	 		p1.SetY(min.GetY());
	 		p1.SetT(min.GetT());	 		
	 		p2.SetX(max.GetX());
	 		p2.SetY(max.GetY());
	 		p2.SetT(max.GetT());
	 	}
	 	else if((p2.GetX() < p1.GetX()) && (p2.GetY() > p1.GetY()))
	 	{
	 		long xMin = p1.GetX() + queryBoundary; 
	 		long yMin = p1.GetY() - queryBoundary;
	 		
	 		long xMax = p2.GetX() - queryBoundary; 
	 		long yMax = p2.GetY() + queryBoundary;
	 		
	 		min = new PositionSpt2D(xMin, yMin, p1.GetT());
	 		max = new PositionSpt2D(xMax, yMax, p2.GetT());
	 		
	 		//re-assign min and max
	 		p1.SetX(min.GetX());
	 		p1.SetY(min.GetY());
	 		p1.SetT(min.GetT());	 		
	 		p2.SetX(max.GetX());
	 		p2.SetY(max.GetY());
	 		p2.SetT(max.GetT());
	 	}
	 	else if((p2.GetX() == p1.GetX()) && (p2.GetY() > p1.GetY()))
	 	{
	 		long xMin = p1.GetX() - queryBoundary; 
	 		long yMin = p1.GetY() - queryBoundary;
	 		
	 		long xMax = p2.GetX() + queryBoundary; 
	 		long yMax = p2.GetY() + queryBoundary;
	 		
	 		min = new PositionSpt2D(xMin, yMin, p1.GetT());
	 		max = new PositionSpt2D(xMax, yMax, p2.GetT());
	 		
	 		//re-assign min and max
	 		p1.SetX(min.GetX());
	 		p1.SetY(min.GetY());
	 		p1.SetT(min.GetT());	 		
	 		p2.SetX(max.GetX());
	 		p2.SetY(max.GetY());
	 		p2.SetT(max.GetT());
	 	}	 	
	 }
	 
	
}
