/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Clipper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.ClassType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.DimensionType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.QueryType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.range.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.MotionDiscreteResult;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MotionDiscretePositionUpdate;

/**
 * @author Rimma V. Nehme
 *
 */
public class FixedDiscrete3DGrid extends GenericSpatioTemporalGrid
{
	//debug flag
	public static int DEBUG = 1; //1=true, 0=false
	
	public GenericSpatioTemporalDiscreteGridCell[][][] directory;
	private int xDim;
	private int yDim;
	private int tDim;
	

	public String toString()
	{
		return statisticsToString();
	}
	
	public void Cleanup()
	{
		for (int i=0; i < xDim; i++)
		{
			for (int j=0; j < yDim; j++)
			{
				for (int k=0; k < tDim; k++)
				{
					if (directory[i][j][k] != null)
					{
					if (directory[i][j][k].GetMovingObjects().size() > 0) 
					{
						//remove all objects from the cell
						directory[i][j][k].GetMovingObjects().clear();
					}
					if (directory[i][j][k].GetMovingQueries().size() > 0)
					{
						//remove all queries from the cell
						directory[i][j][k].GetMovingQueries().clear();				 
					}
					}
				}
			}
		}
	}
	
	public String statisticsToString()
	{
		int totalNumberOfObjects = 0;
		int totalNumberOfQueries = 0;
		String outputString="*** Grid Statistics ***";
		
		for (int i=0; i < xDim; i++)
		{
			for (int j=0; j < yDim; j++)
			{
				for (int k=0; k < tDim; k++)
				{
					if (directory[i][j][k] != null)
					{
					totalNumberOfObjects += directory[i][j][k].GetMovingObjects().size();
					totalNumberOfQueries += directory[i][j][k].GetMovingQueries().size();
					
					if ((directory[i][j][k].GetMovingObjects().size() > 0) || 
					    (directory[i][j][k].GetMovingQueries().size() > 0))
					{
						outputString += "\ngrid["+i+"]["+j+"]["+k+"]: # objects = "+directory[i][j][k].GetMovingObjects().size()+", # queries = "+directory[i][j][k].GetMovingQueries().size();
						outputString +=  " [ Cell Dimension: i = "+ (i * GenericSpatioTemporalGrid.xDimensionSize)+ ", "+((i+1) * GenericSpatioTemporalGrid.xDimensionSize)+"; "+
														 "j = "+ (j * GenericSpatioTemporalGrid.yDimensionSize)+ ", "+((j+1) * GenericSpatioTemporalGrid.yDimensionSize)+"; "+														 
														 "k = " + (k * GenericSpatioTemporalGrid.tDimensionSize)+", "+((k+1) * GenericSpatioTemporalGrid.tDimensionSize)+";] ";
						 
					}
					}
				}
			}
		}
		
		outputString += "\nTotal # of objects entries = "+totalNumberOfObjects+", Total # of queries entries = "+totalNumberOfQueries +"\n";
		return outputString;
	}
	
	public FixedDiscrete3DGrid(int xDimension, int yDimension, int tDimension) 
	{
		xDim = xDimension;
		yDim = yDimension;
		tDim = tDimension;
		
		directory = new GenericSpatioTemporalDiscreteGridCell[xDimension][yDimension][tDimension];
			
	}


	public void CheckCreateGridCell(int i, int j, int k)
	{
		if (directory[i][j][k] == null)
		{			
			directory[i][j][k] = new GenericSpatioTemporalDiscreteGridCell(i,
				 	   j,
					   k,
					   i * GenericSpatioTemporalGrid.xDimensionSize,
					   j * GenericSpatioTemporalGrid.yDimensionSize,
					   k * GenericSpatioTemporalGrid.tDimensionSize,
					  (i+1) * GenericSpatioTemporalGrid.xDimensionSize,
					  (j+1) * GenericSpatioTemporalGrid.yDimensionSize,
					  (k+1) * GenericSpatioTemporalGrid.tDimensionSize);
			
		}
		
	}
	

	
	
	public void InsertObject(MovingObjectState mo, boolean deleteFromSystem, Time currentTime, int delta)	
	{
		int xIndex = GenericGrid.GetCellIndex(mo.GetPosition().GetX(), DimensionType.x);		
		int yIndex = GenericGrid.GetCellIndex(mo.GetPosition().GetY(), DimensionType.y);
		int tIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)

		CheckCreateGridCell(xIndex,yIndex,tIndex);
		
		if (!GenericSpatioTemporalGrid.IsIndexWithinGridBoundary(xIndex, yIndex, tIndex))
		{
			Helper.DebugWriteLine("DELETING "+EntityType.FromEnumToString(mo.GetEntityType())+" "+mo.GetID()+" from the sytems");
			deleteFromSystem = true;
			return; //for now just return without entering the entity
		}
		
		
		GridEntry geObject = new GridEntry(mo.GetID(), mo.GetPosition().GetT());		
		directory[xIndex][yIndex][tIndex].AddMovingObjectEntry(geObject);					
	}
	
	
	
	public void InsertQuery(MovingQueryState mq, boolean deleteFromSystem, Time currentTime, int delta)	
	{
		int xIndex = GenericGrid.GetCellIndex(mq.GetPosition().GetX(), DimensionType.x);		
		int yIndex = GenericGrid.GetCellIndex(mq.GetPosition().GetY(), DimensionType.y);		
		int tIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)

		CheckCreateGridCell(xIndex,yIndex,tIndex);
		
		if (!GenericSpatioTemporalGrid.IsIndexWithinGridBoundary(xIndex, yIndex, tIndex))
		{
			Helper.DebugWriteLine("DELETING "+EntityType.FromEnumToString(mq.GetEntityType())+" "+mq.GetID()+" from the systems");
			deleteFromSystem = true;
			return; //for now just return without entering the entity
		}
		
		GridEntry geQuery = new GridEntry(mq.GetID(), mq.GetPosition().GetT());
		directory[xIndex][yIndex][tIndex].AddMovingQueryEntry(geQuery);	
		
		
		//check if the boundary of the query might cause it to intersect
		//with other neightboring grid cells
		
		PositionSpt2D cellMin = GenericSpatioTemporalGrid.GetCellWindowMin(new GridCellIndex3D(xIndex,yIndex,tIndex));
		PositionSpt2D cellMax = GenericSpatioTemporalGrid.GetCellWindowMax(new GridCellIndex3D(xIndex,yIndex,tIndex));
		
		if ((mq.GetPosition().GetX() + mq.GetQueryBoundary()) > cellMax.GetX())
		{
			CheckCreateGridCell(xIndex+1,yIndex,tIndex);
			CheckCreateGridCell(xIndex+1,yIndex+1,tIndex);
			CheckCreateGridCell(xIndex+1,yIndex-1,tIndex);
			
			directory[xIndex+1][yIndex][tIndex].AddMovingQueryEntry(geQuery);	
			directory[xIndex+1][yIndex+1][tIndex].AddMovingQueryEntry(geQuery);	
			directory[xIndex+1][yIndex-1][tIndex].AddMovingQueryEntry(geQuery);	
		}
		
		if (mq.GetPosition().GetX() - mq.GetQueryBoundary() < cellMin.GetX())
		{
			CheckCreateGridCell(xIndex-1,yIndex,tIndex);
			CheckCreateGridCell(xIndex-1,yIndex+1,tIndex);
			CheckCreateGridCell(xIndex-1,yIndex-1,tIndex);
						
			directory[xIndex-1][yIndex][tIndex].AddMovingQueryEntry(geQuery);	
			directory[xIndex-1][yIndex+1][tIndex].AddMovingQueryEntry(geQuery);	
			directory[xIndex-1][yIndex-1][tIndex].AddMovingQueryEntry(geQuery);	
		}
		
		if (mq.GetPosition().GetY() + mq.GetQueryBoundary() > cellMax.GetY())
		{
			CheckCreateGridCell(xIndex-1,yIndex+1,tIndex);
			CheckCreateGridCell(xIndex,yIndex+1,tIndex);
			CheckCreateGridCell(xIndex+1,yIndex+1,tIndex);
			
			directory[xIndex-1][yIndex+1][tIndex].AddMovingQueryEntry(geQuery);	
			directory[xIndex][yIndex+1][tIndex].AddMovingQueryEntry(geQuery);	
			directory[xIndex+1][yIndex+1][tIndex].AddMovingQueryEntry(geQuery);	
		}
		
		if (mq.GetPosition().GetY() - mq.GetQueryBoundary() < cellMin.GetY())
		{
			CheckCreateGridCell(xIndex-1,yIndex-1,tIndex);
			CheckCreateGridCell(xIndex,yIndex-1,tIndex);
			CheckCreateGridCell(xIndex+1,yIndex-1,tIndex);
			
			directory[xIndex-1][yIndex-1][tIndex].AddMovingQueryEntry(geQuery);	
			directory[xIndex][yIndex-1][tIndex].AddMovingQueryEntry(geQuery);	
			directory[xIndex+1][yIndex-1][tIndex].AddMovingQueryEntry(geQuery);	
		}
		
	}


	public MotionDiscreteResult DoJoin(MovingQueryState queryState, 
							  		   MovingObjectState objectState, 
									   GridCellIndex3D gi)
	{
		MotionDiscreteResult intersectionResult = null;
		MotionDiscreteState discreteResultPosition = null;
		
		if ((objectState.GetPosition().GetX() < (queryState.GetPosition().GetX() + queryState.GetQueryBoundary()))&&
		    (objectState.GetPosition().GetY() < (queryState.GetPosition().GetY() + queryState.GetQueryBoundary()))&&
		    (objectState.GetPosition().GetX() > (queryState.GetPosition().GetX() - queryState.GetQueryBoundary()))&&
		    (objectState.GetPosition().GetY() > (queryState.GetPosition().GetY() - queryState.GetQueryBoundary()))&&
		    (objectState.GetPosition().GetT() == queryState.GetPosition().GetT()))
		{
			discreteResultPosition = new MotionDiscreteState(objectState.GetID(),
													EntityType.Object,
													objectState.GetPosition().GetX(),
													objectState.GetPosition().GetY(), 
													objectState.GetPosition().GetT());
			
			intersectionResult = new MotionDiscreteResult(queryState.GetID(),  
														  objectState.GetID(), 
														  QueryType.Range, 
														  discreteResultPosition);
		}			
		return intersectionResult;
	}
	
	
}

