/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid;
import java.util.*;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous.*;

/**
 * @author Rimma V. Nehme
 *
 */
public class GenericSpatioTemporalContinuousGridCell extends GenericSpatioTemporalGridCell
{
	private ArrayList movingObjects; //MovingObjectContinuousState[]
	private ArrayList movingQueries; //MovingQueryContinuousState[]
	
	public GenericSpatioTemporalContinuousGridCell(int newXIndex, int newYIndex, int newTIndex,
			  									   long newXMin, long newYMin, long newTMin,
			  									   long newXMax, long newYMax, long newTMax)
	{
		super(newXIndex, newYIndex, newTIndex, newXMin, newYMin, newTMin, newXMax, newYMax, newTMax);
		movingObjects = new ArrayList();
		movingQueries = new ArrayList();	
	}
	
	public ArrayList GetQueryUpdatesInTheCellByQueryIDAndGridIndex(int queryID)
	{
		ArrayList tmpQueryEntries = new ArrayList();
		tmpQueryEntries.clear();
		
		for (int i=0; i< this.movingObjects.size();i++)
		{
			MovingQueryContinuousState mQueryEntry = ((MovingQueryContinuousState)(this.movingQueries.get(i)));
			if (mQueryEntry.GetID() == queryID)
			{
				//get all location updates for that particular object in that particular grid cell
				tmpQueryEntries.add(mQueryEntry);
			}
		}
		return tmpQueryEntries;		
	}
	
	/*dp:
	boolean isContinuousStateAlreadyInTheCell(MotionContinuousState movingEntry)
	{
		if(movingEntry.GetEntityType() == EntityType.Object)
		{
			for (int i=0; i< this.movingObjects.size();i++)
			{
				MovingObjectContinuousState mObjectEntry = ((MovingObjectContinuousState)(this.movingObjects.get(i)));
				if ((mObjectEntry.GetID() == movingEntry.GetID()) && 
					(mObjectEntry.GetStartPosition().GetT() == movingEntry.GetStartPosition().GetT()) &&
					(mObjectEntry.GetEndPosition().GetT() == movingEntry.GetEndPosition().GetT()))
				{
					return true;
				}
			}			
		}
		else if(movingEntry.GetEntityType() == EntityType.Query)
		{
			for (int i=0; i< this.movingQueries.size();i++)
			{
				MovingQueryContinuousState mQueryEntry = ((MovingQueryContinuousState)(this.movingQueries.get(i)));
				if ((mQueryEntry.GetID() == movingEntry.GetID()) && 
				(mQueryEntry.GetStartPosition().GetT() == movingEntry.GetStartPosition().GetT()) &&
				(mQueryEntry.GetEndPosition().GetT() == movingEntry.GetEndPosition().GetT()))
				{
					return true;
				}
			}		
		}
		return false;
	}
	*/
	
	boolean isContinuousStateAlreadyInTheCell(GridContinuousEntry movingEntry, int entityType)
	{
		if(entityType == EntityType.Object)
		{
			for (int i=0; i< this.movingObjects.size();i++)
			{
				GridContinuousEntry mObjectEntry = ((GridContinuousEntry)(this.movingObjects.get(i)));
				if ((mObjectEntry.GetID() == movingEntry.GetID()) && 
					(mObjectEntry.GetTimeStart() == movingEntry.GetTimeStart()) &&
					(mObjectEntry.GetTimeEnd() == movingEntry.GetTimeEnd()))
				{
					return true;
				}
			}			
		}
		else if(entityType == EntityType.Query)
		{
			for (int i=0; i< this.movingQueries.size();i++)
			{
				GridContinuousEntry mQueryEntry = ((GridContinuousEntry)(this.movingQueries.get(i)));
				if ((mQueryEntry.GetID() == movingEntry.GetID()) && 
				(mQueryEntry.GetTimeStart() == movingEntry.GetTimeStart()) &&
				(mQueryEntry.GetTimeEnd() == movingEntry.GetTimeEnd()))
				{
					return true;
				}
			}		
		}
		return false;
	}
	
	public ArrayList GetObjectUpdatesInTheCellByObjectIDAndGridIndex(int objectID)
	{
		ArrayList tmpObjectEntries = new ArrayList();
		tmpObjectEntries.clear();
		
		for (int i=0; i< this.movingObjects.size();i++)
		{
			MovingObjectContinuousState mObjectEntry = ((MovingObjectContinuousState)(this.movingObjects.get(i)));
			if (mObjectEntry.GetID() == objectID)
			{
				//get all location updates for that particular object in that particular grid cell
				tmpObjectEntries.add(mObjectEntry);
			}
		}
		return tmpObjectEntries;		
	}
	
	/*dp:
	public void AddMovingObjectEntry(MotionContinuousState newMovingObjectEntry)
	{
		if (!isContinuousStateAlreadyInTheCell(newMovingObjectEntry))
		{
			movingObjects.add(newMovingObjectEntry);
		}
	}
	*/
	
	public void AddMovingObjectEntry(GridContinuousEntry movingObjectGridEntry)
	{
		if (!isContinuousStateAlreadyInTheCell(movingObjectGridEntry, EntityType.Object))
		{
			movingObjects.add(movingObjectGridEntry);
		}
	}
	
	/*dp:
	public void AddMovingQueryEntry(MotionContinuousState newMovingQueryEntry)
	{
		if (!isContinuousStateAlreadyInTheCell(newMovingQueryEntry))
		{
			movingQueries.add(newMovingQueryEntry);
		}
	}
	*/
	
	public void AddMovingQueryEntry(GridContinuousEntry movingQueryGridEntry)
	{
		if (!isContinuousStateAlreadyInTheCell(movingQueryGridEntry, EntityType.Query))
		{
			movingQueries.add(movingQueryGridEntry);
		}
	}
	
	public ArrayList GetMovingObjects()
	{
		return this.movingObjects;
	}
	
	public ArrayList GetMovingQueries()
	{
		return this.movingQueries;
	}

}
