/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.collections.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.*;

/**
 * @author Rimma V. Nehme
 *
 */
public class ResultsRecorder 
{

	//static String fileName = "D:\\Rimma\\WPI\\Thesis\\Debug\\Results.txt";
	//static String fileName = "C:\\Documents and Settings\\Rimma Nehme\\My Documents\\WPI\\Debug\\Results.txt";
	static String fileName = "Results.txt";
	
	public static void ResultsWrite(String textToWrite)
	{
	  try 
	  {
         BufferedWriter out = new BufferedWriter(new FileWriter(fileName, true));
         out.write(textToWrite);
         out.close();
	  } 
	  catch(IOException e) 
	  {
    
	  }
	}
	
	public static void ResultsWriteLine(String textToWrite)
	{
	  try 
	  {
         BufferedWriter out = new BufferedWriter(new FileWriter(fileName, true));
         out.write(textToWrite + "\n");
         out.close();
	  } 
	  catch(IOException e) 
	  {
    
	  }
	}
	
	
	public static void ContinuousSingleResultsWriteLine(MotionContinuousResult result)
	{
		 try 
		  {
		  	
	         BufferedWriter out = new BufferedWriter(new FileWriter(fileName, true));
	         
	         	MotionContinuousState resultValue = result.GetResultValue();
	         	out.write(result.GetQueryID()+" "+
	         			  QueryType.FromEnumToString(result.GetQueryType())+" "+
	         			  result.GetObjectID()+" "+         			 
	         			  EntityType.FromEnumToString(resultValue.GetEntityType())+" "+
						  resultValue.GetStartPosition().GetX()+" "+
						  resultValue.GetStartPosition().GetY()+" "+
						  resultValue.GetStartPosition().GetT()+" "+
						  resultValue.GetEndPosition().GetX()+" "+
						  resultValue.GetEndPosition().GetY()+" "+
						  resultValue.GetEndPosition().GetT()+" "+
						  FunctionType.FromEnumToString(resultValue.GetContinuityType())+"\n");					  

	         out.write("\n"); 
	         out.close();
		  } 
		  catch(IOException e) 
		  {
	    
		  }
		
	}
	
	public static void DiscreteSingleResultsWriteLine(MotionDiscreteResult result)
	{
		 try 
		  {	  		
	         BufferedWriter out = new BufferedWriter(new FileWriter(fileName, true));
	         out.write("\n");       
	       

	         	MotionDiscreteState resultValue = result.GetResultValue();
	         	out.write(result.GetQueryID()+" "+
	         			  QueryType.FromEnumToString(result.GetQueryType())+" "+
	         			  result.GetObjectID()+" "+         			 
	         			  EntityType.FromEnumToString(resultValue.GetEntityType())+" "+
						  resultValue.GetPosition().GetX()+" "+
						  resultValue.GetPosition().GetY()+" "+
						  resultValue.GetPosition().GetT()+"\n");					  
	   
	         
	         out.write("\n"); 
	         out.close();
		  } 
		  catch(IOException e) 
		  {
	    
		  }
		
	}
	
	
	public static void ContinuousResultsWriteLine(MotionContinuousResultsCollection continuousResultsCollection)
	{
	  try 
	  {
	  	
         BufferedWriter out = new BufferedWriter(new FileWriter(fileName, true));
         out.write("\n");
         /*
         out.write("QueryID"+" "+
   			       "QueryType()"+" "+
   			       "ObjectID"+" "+         			 
   			       "EntityType"+" "+
				   "x_start"+" "+
				   "y_start"+" "+
				   "t_start"+" "+
				   "x_end"+" "+
				   "y_end"+" "+
				   "t_end"+" "+
				   "ContinuityType"+"\n");
         */
         for (int i=0; i<continuousResultsCollection.GetContinuousResults().size();i++)
         {
         	MotionContinuousResult result = (MotionContinuousResult)continuousResultsCollection.GetContinuousResults().get(i);
         	MotionContinuousState resultValue = result.GetResultValue();
         	out.write(result.GetQueryID()+" "+
         			  QueryType.FromEnumToString(result.GetQueryType())+" "+
         			  result.GetObjectID()+" "+         			 
         			  EntityType.FromEnumToString(resultValue.GetEntityType())+" "+
					  resultValue.GetStartPosition().GetX()+" "+
					  resultValue.GetStartPosition().GetY()+" "+
					  resultValue.GetStartPosition().GetT()+" "+
					  resultValue.GetEndPosition().GetX()+" "+
					  resultValue.GetEndPosition().GetY()+" "+
					  resultValue.GetEndPosition().GetT()+" "+
					  FunctionType.FromEnumToString(resultValue.GetContinuityType())+"\n");					  
         }
         
         out.write("\n"); 
         out.close();
	  } 
	  catch(IOException e) 
	  {
    
	  }
	}
	
	public static void SCUBAResultsWriteLine(ArrayList scubaResultsCollection)
	{
		 
		  try 
		  {	  		
	         BufferedWriter out = new BufferedWriter(new FileWriter(fileName, true));
	         out.write("\n");       
	       
	         for (int i=0; i<scubaResultsCollection.size();i++)
	         {
	         	MotionDiscreteResult result = (MotionDiscreteResult)scubaResultsCollection.get(i);
	         	MotionDiscreteState resultValue = result.GetResultValue();
	         	out.write(result.GetQueryID()+" "+
	         			  QueryType.FromEnumToString(result.GetQueryType())+" "+
	         			  result.GetObjectID()+" "+         			 
	         			  EntityType.FromEnumToString(resultValue.GetEntityType())+" "+
						  resultValue.GetPosition().GetX()+" "+
						  resultValue.GetPosition().GetY()+" "+
						  resultValue.GetPosition().GetT()+"\n");					  
	         }
	         
	         out.write("\n"); 
	         out.close();
		  } 
		  catch(IOException e) 
		  {
	    
		  }
	}
	
	
	public static void DiscreteResultsWriteLine(MotionDiscreteResultsCollection discreteResultsCollection)
	{
	  
	  try 
	  {	  		
         BufferedWriter out = new BufferedWriter(new FileWriter(fileName, true));
         out.write("\n");       
       
         for (int i=0; i<discreteResultsCollection.GetDiscreteResults().size();i++)
         {
         	MotionDiscreteResult result = (MotionDiscreteResult)discreteResultsCollection.GetDiscreteResults().get(i);
         	MotionDiscreteState resultValue = result.GetResultValue();
         	out.write(result.GetQueryID()+" "+
         			  QueryType.FromEnumToString(result.GetQueryType())+" "+
         			  result.GetObjectID()+" "+         			 
         			  EntityType.FromEnumToString(resultValue.GetEntityType())+" "+
					  resultValue.GetPosition().GetX()+" "+
					  resultValue.GetPosition().GetY()+" "+
					  resultValue.GetPosition().GetT()+"\n");					  
         }
         
         out.write("\n"); 
         out.close();
	  } 
	  catch(IOException e) 
	  {
    
	  }
	  
	}
	
}
