/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
 * @author Rimma V. Nehme
 *
 */
public abstract class MotionGenericOperator 	
extends XATMultiQueueWindowStreamOperatorImp 
implements IMotionGenericOperatorInterface 
{
	protected static int executionIterationCount;
	protected static int delta;
	protected static int numberOfSupportedObjects;
	protected static int numberOfSupportedQueries;
	
	
	public MotionGenericOperator(XATNode node) 
	{
		super(node);			
	}

	public static int  GetExecutionIterationCount()
	{
		return executionIterationCount;
	}
	
	public static void  SetExecutionIterationCount(int incrementedCounter)
	{
		executionIterationCount = incrementedCounter;
	}
	
	
	public static int  GetDelta()
	{
		return delta;
	}
	
	public static void  SetDelta(int newDelta)
	{
		delta = newDelta;
	}
	
	public String toToolTipString() 
	{
		String rs = new String();

		String sName = this.getClass().getName();
		//find the last dot.
		int index = 0;
		for(index = sName.length()-1; index >=0; index--)
			if(sName.charAt(index) == '.')
				break;
		rs = sName.substring(index + 1) + " ";
		String id = this.getProperty("OBJECT_ID");
		if(id == "-1")
		{
			rs = rs + hashCode(); 
		} 
		else
		{
			rs = rs + id;
		}	
		if(this.isActive()) rs += " ACTIVE ";
		else rs+= " INACTIVE ";

		/** now add the join predicate and window size info */
		rs += "\n";
		rs += "window size = "+windowSize+"\n";

		return rs;
	}
	
	/* 
	 * process all accumulated tuples in input queues that have their maxtimestamp 
	 * smaller than the given maxtimestamp 
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		if ((this.getLeftInputQueue().getSchema() == null) || (this.getRightInputQueue().getSchema() == null))
		{
			return false;
		}
		if(this.getOutputQueue().getSchema() == null)
		{
			init();
		}
		
		/** figure out how many tuples to dequeue from left and right queue */
		XATQueue lQueue = this.getLeftInputQueue();
		XATQueue rQueue = this.getRightInputQueue();
		int size1 = lQueue.getQueueSize(this);
		int size2 = rQueue.getQueueSize(this);
		int lHowmany=size1; 
		int rHowmany=size2;

		if(lQueue.isOldBoxInput()) 
		{
			XATTuple[] tuples = lQueue.peek(size1, this);
			for (int i = 0; i < size1; i++) {
				if (((XATTuple) tuples[i]).getMaxTimeStamp() >= maxtimestamp) 
				{
					lHowmany = i;
					break;
				}
			}
		}
		
		if(rQueue.isOldBoxInput()) 
		{
			XATTuple[] tuples = rQueue.peek(size2, this);
			for (int i = 0; i < size2; i++) 
			{
				if (((XATTuple) tuples[i]).getMaxTimeStamp() >= maxtimestamp) 
				{
					rHowmany = i;
					break;
				}
			}
		}

		if((lHowmany + rHowmany) == 0) 
			return false;
			
		if (debugFlag) {
			System.out.println("dequeue " + lHowmany + "tuples from left queue");
			System.out.println("dequeue " + rHowmany + "tuples from right queue");
		}
			
		//execOperator(lHowmany, rHowmany);

		/* run successfully */
		return true;	
	}
	
	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() 
	{
		if ((this.getLeftInputQueue().getSchema() == null) || (this.getRightInputQueue().getSchema() == null))
		{
			return false;
		}
		if(this.getOutputQueue().getSchema() == null)
		{
			init();
		}
		
		/** figure out how many tuples to dequeue from left and right queue */
		XATQueue lQueue = this.getLeftInputQueue();
		XATQueue rQueue = this.getRightInputQueue();
		int lHowmany = lQueue.getQueueSize(this);
		int rHowmany = rQueue.getQueueSize(this);

		/** nothing to clean */
		if(lHowmany == 0 && rHowmany == 0) 
		{
			return false;
		}

		if (debugFlag) {
			System.out.println("dequeue " + lHowmany + "tuples from left queue");
			System.out.println("dequeue " + rHowmany + "tuples from right queue");
		}
			
		//execOperator(lHowmany, rHowmany);

		/* run successfully */
		return true;	
	}
	


	
	/*
	 * Assign new ID to each state inside this operator, starting from the given startID.
	 * 
	 * return the next available ID. For example, if startID is 2, and there are 2 states inside this operator,
	 * then these two states are assigned new IDs as 2 and 3, and 4 is then returned as the next available stateID.
	 */
	  public int setStateID(int startID) 
	  {
		if (states == null) 
		{
			return startID;
		}
	
		for(int i=0; i<states.length; i++) 
		{
			states[i].setStateID(startID);
			startID++;
		}
		return startID;
	  }
	  

	  public boolean matchStates(XATStreamOperator newboxoperator) 
	  {
	  	int thisSize = this.states.length;
	  	int thatSize = newboxoperator.getState().length;
	  	OperatorState[] thatstates = newboxoperator.getState();
	  	boolean hasmatchedstate = false;
	  	
	  	for(int i=0; i<thisSize; i++) 
	  	{
	  		for(int j=0; j<thatSize; j++) 
	  		{
	  			// found a matched state
	  			if(states[i].getStateID() == thatstates[j].getStateID()) 
	  			{
	  				thatstates[j].setMatched(true);
	  				thatstates[j].copyStateContent(states[i]);
	  				hasmatchedstate = true;
	  			}
	  		}
	  	}
	  	
	  	return hasmatchedstate;
	  }
	  // -- Bin@Nov.23,2004 for changing the XATMultiQueueStreamOperator interface
	public boolean finalizeOperation()
	{
		return false;
	}
	
	public abstract boolean run(); 
	
	

}
