/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.clusteroperator;
/**
 * This class represents a circle that will encompass a cluster of points
 *  @author Robert Contois
 *
 */
public class Circle
{
    private Points point = new Points();
    private double radius;
    
    /*
     * Constructor for the circle class
     * 
     * @param temp_x	temporary value for the x coordinate
     * @param temp_y	temporary value for the y coordinate
     * @param tenp_raduis	temporary value for the radius
     */
    public Circle(double temp_x, double temp_y, double temp_radius)
    {	
    	point.setX(temp_x);			//set the x coordinate
    	point.setY(temp_y);			//set the y coordinate
    	radius = temp_radius;		//set the radius for the circle
    }
    
    /*
     * Empty constructor for the circle class
     */
    public Circle()
    {	
    }
  
    /*
     * get the center point of the cluster
     * 
     * @return	the center of the cluster-circle in a point format
     */
    public Points getPoint()
    {
    	return point;
    }

    /*
     * get the radius of the cluster
     * 
     * @return	the radius of the cluster-circle
     */
    public double getRadius()
    {
    	return radius;
    }

    /*
     * set the radius of the circle. Nothing is return.
     * 
     * @param	temp_radius
     */
    public void setRadius(double temp_radius)
    {
    	radius = temp_radius;
    }

    /*
     * Takes in a point and checks to see if it is in the circle or not.
     * 
     * @param pt	The point to check to see if it is in the circle or not.
     * @return	true if the point is in the circle, false otherwise
     */
    public boolean inCircle(Points pt)
    {
    	double xCoord;		//difference of the x point between the center an point
    	double yCoord;		//difference of the y point between the center and point
    	double distance;	//distance between the two points
    	double difference;	//difference between the center of the 
    	boolean result;		//The result we are going ot return
	
    	//calcualte the difference between the center and the point we are checking
    	xCoord = Math.pow(pt.getX() - point.getX(),2d);
    	yCoord = Math.pow(pt.getY() - point.getY(),2d);
	
    	//find the distance between the two points
    	distance = Math.sqrt(xCoord + yCoord);
    	difference = radius - distance;
	
    	//because of our calcualtions around the circle we need a margin to check between
    	if(difference > -0.001 ||
			difference < 0.001)
	    {
    		result = true;
	    }
    	else
	    {
    		result = false;
	    }
    	return result;
    }
    
    /*
     * Gets the area of the current circle and return it
     * 
     * @return	the area of the circle as a double
     */
    public double getArea()
    {
    	return Math.PI * Math.pow( radius, 2);
    }
    
    /*
     * convert the class to a string
     * 
     * @return	the string representing the class
     */
    public String toString()
    {
    	return point.getX() + "," + point.getY() + " r" + radius;
    }

    /*
     * convert the class to a string
     * 
     * @param c	calculate the overlap between the input and this class
     * @return	the area of the overlapping circle
     */
    public double calculateOverlap(Circle c)
    {
        double centerDiff;	//the distance between the two center points
        double angle1;		//the first angle
        double angle2;		//the second angle
        double result;		//the resulting area

        //get the difference between the centers
        centerDiff = getDistance(point,c.getPoint());

        //find the angles created by the intersection of the circles and their center points
        angle1 = 2 * Math.acos((Math.pow(c.getRadius(),2) + Math.pow(centerDiff,2) - Math.pow(radius,2))/(2 * c.getRadius() *centerDiff));
        angle2 = 2 * Math.acos((Math.pow(radius,2) + Math.pow(centerDiff,2) - Math.pow(c.getRadius(),2))/(2 * radius * centerDiff));

        //check that the difference of the radius is valid
        if(centerDiff > (c.getRadius() + radius))
            {
        		//the circles do not overlap
                result = 0;
            }
        else if(((Math.pow(c.getRadius(),2) + Math.pow(centerDiff,2) - Math.pow(radius,2))/(2 * c.getRadius() * centerDiff) < -1) ||
        		(centerDiff == 0 && c.getRadius() < radius))
            {
                //circle completely in other
                result = Math.pow(c.getRadius(),2) * Math.PI;
            }
        else if (((Math.pow(c.getRadius(),2) + Math.pow(centerDiff,2) - Math.pow(radius,2))/(2 * c.getRadius() * centerDiff) > 1) ||
        		(centerDiff == 0 && c.getRadius() > radius))
        	{
        		//other case where 1 circle is in another circle
        		result = Math.pow(radius,2) * Math.PI;
            }
        else
            {
        		//the two circles just overlap
                result = ((angle1)*(Math.pow(c.getRadius(),2)) - Math.sin(angle1)*(Math.pow(c.getRadius(),2)) +
                          (angle2)*(Math.pow(radius,2)) - Math.sin(angle2)*(Math.pow(radius,2)))/2;
            }
        
        //return the result
        return result;
    }

    /*
     * calculate the distance between two points
     * 
     * @param pt1	The first point for distance checking
     * @param pt2	The second point for distance checking
     * @return	the distance between two points
     */
    private double getDistance(Points pt1, Points pt2)
    {
    	double xCoord;	//the roots of the difference between the coord set
    	double yCoord;	
    	double result;	//the resulting distance

    	xCoord = Math.pow(pt1.getX() - pt2.getX(),2d);
    	yCoord = Math.pow(pt1.getY() - pt2.getY(),2d);

    	//calculate the distance
    	result = Math.sqrt(xCoord + yCoord);
    	
    	return result;
    }
}




