/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.clusteroperator;
import java.util.Vector;
import java.math.*;


/**
 * This is the primary structure for clusters. It defines a circle that encompasses all of the points that are contained in a vector of points.
 * 
 * @author Robert Contois
 */
class Cluster
{
    private Vector pts;
    private Points far1;
    private Points far2;
	private Circle cluster;
    
    /*
     * Constructor for the Cluster class. Create a space in memory for the circle of the cluster
     * and the vector containing the points that make up the cluster
     */
	Cluster()
	{
		cluster = new Circle();
		pts = new Vector();
	}
	
    /*
     * calculate the distance between two points
     * 
     * @param pt1	The first point for distance checking
     * @param pt2	The second point for distance checking
     * @return	the distance between two points
     */
    private double getDistance(Points pt1, Points pt2)
    {
	double xCoord;
	double yCoord;
	double result;

	xCoord = Math.pow(pt1.getX() - pt2.getX(),2);
	yCoord = Math.pow(pt1.getY() - pt2.getY(),2);

	//	calculate the distance
	result = Math.sqrt(xCoord + yCoord);
	
	return result;
    }
    
    /*
     * given three point find the point directly in the center. This will be the center of a circle though all 3 points.
     * 
     * @param pt1	The first point on the circle
     * @param pt2	The second point on the circle
     * @param pt2	The second point on the circle
     * @return	the point in the center of the st of three points
     */
    private Points getCenter(Points pt1, Points pt2, Points pt3)
    {
    	double temp, bc, cd, det, cx, cy;

    	//matrix math to calculate the determinant
    	temp = Math.pow(pt2.getX(),2) + Math.pow(pt2.getY(),2);
    	bc = (Math.pow(pt1.getX(),2) + Math.pow(pt1.getY(),2) - temp) / 2;
    	cd = (temp - Math.pow(pt3.getX(),2) - Math.pow(pt3.getY(),2)) / 2;
    	det = ((pt1.getX() - pt2.getX()) * (pt2.getY() - pt3.getY())) - ((pt2.getX() - pt3.getX()) *  (pt1.getY() - pt2.getY()));
		det = 1/det;
		
		//calculate the x and y coord of the center point
		cx = ((bc * (pt2.getY() - pt3.getY())) - (cd * (pt1.getY() - pt2.getY()))) * det;
		cy = ((cd * (pt1.getX() - pt2.getX())) - (bc * (pt2.getX() - pt3.getX()))) * det;

		//create a point representing the center point
		Points result = new Points(cx,cy);

		//return the center point
		return result;
    }

    /*
     * Find the middle point between two points
     * 
     * @param pt1 	start point
     * @param pt2	end point
     *
     * @return		the point representing the midpoint of a segment between pt1 and pt2
     */
    private Points getMidpoint(Points pt1, Points pt2)
    {
    	Points result = new Points();
	
    	//get the x midpoint value
    	result.setX( ((pt1.getX() - pt2.getX()) / 2) + pt2.getX());
    	//get the y midpoint value
    	result.setY( ((pt1.getY() - pt2.getY()) / 2) + pt2.getY());
	
    	return result;
    }

        
    /*
     * add a point to the current cluster
     * 
     * @param x		the x coordinate of the point to add
     * @param y		the y coordinate of the point to add
     */
    public void addPoint(double x, double y)
    {
        pts.add(new Points(x,y));
    }


    public void calcCircle()
    {
    	if(pts.size() > 2)
    	{
    		double temp_far=0;
    		double temp;
    		//loop through the vector for the first point
    		for(int i=0; i<pts.size()-1; i++)
    		{
    			//loop through the vector for the second point
    			for(int j=i+1; j<pts.size(); j++)
    			{
    				//get the distance between the two prospective points
    				temp=getDistance((Points)(pts.get(i)), (Points)(pts.get(j)));
    				
    				//compare them with the greatest distance thus far
    				//we have a new furthest set
    				if(temp>temp_far)
    				{
    					temp_far=temp;
    					far1=(Points)(pts.get(i));
    					far2=(Points)(pts.get(j));
					}
    				//no change in our set
    				else
    				{
    					
    				}
    			}
    		}
    	
    		((Points)(cluster.getPoint())).setX(((Points)(getMidpoint(far1,far2))).getX());
    		((Points)(cluster.getPoint())).setY(((Points)(getMidpoint(far1,far2))).getY());
    		cluster.setRadius(getDistance(far2,cluster.getPoint()));
	
    		//loop through the vector for the first point
    		for(int i=0; i<pts.size(); i++)
    		{
    			if(cluster.inCircle((Points)(pts.get(i))) == true)
    			{
    				//the point is in the circle
    			}
    			else
    			{
    				//otherwise false, point not in circle
    				//System.out.println("Point not in the circle");
    				((Points)(cluster.getPoint())).setX(((Points)(getCenter(far1,far2,((Points)(pts.get(i)))))).getX());
    				((Points)(cluster.getPoint())).setY(((Points)(getCenter(far1,far2,((Points)(pts.get(i)))))).getY());
    				cluster.setRadius(getDistance(far1,(Points)(cluster.getPoint())));
    				i=0;
    			}
    		}
    	}
    	else
    	{
    	}
    }       

    /*
     * return the circle that defines the cluster
     * 
     * return the circle of the cluster
     */
    public Circle getCircle()
    {
    	return cluster;
    }
    
    /*
     * get the size of the array containing all the object points
     * 
     * @return the amount of objects in the array containing all the points
     */
    public int getSize()
    {
    	return pts.size();
    }
}
