/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.clusteroperator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.MultiMap;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.MotionGenericOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.MotionContinuousResult;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.MotionDiscreteResult;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.collections.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections.MovingObjectsIntervalStatesCollection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections.MovingQueriesIntervalStatesCollection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous.MotionContinuousState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous.MovingObjectContinuousState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous.MovingQueryContinuousState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingObjectState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingQueryState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.ClassType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.FunctionType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.QueryType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.UpdateType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.ResultsRecorder;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.range.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
 * @author Rimma V. Nehme
 * @author Robert Contois
 * 
 * This class defines the operator that returns the area covered by a cluster
 *
 */
public class MotionAreaClusterOperator extends MotionGenericClusterOperator
{
	//debug flag
	public static int DEBUG = 1; //1=true, 0=false
	boolean pushTimeIntervalForward = false;

	
	public MotionAreaClusterOperator(XATNode node) 
	{
		super(node);	
	}
	
	public boolean run() 
	{
		try
		{
			if(DEBUG == 1)
			{
				//Helper.DebugWriteLine("\nINSIDE CLUSTER OPERATOR!!!");
			}
			if ((this.getLeftInputQueue().getSchema() == null) || (this.getRightInputQueue().getSchema() == null))
			{
				return false;
			}
			if(this.getOutputQueue().getSchema() == null)
			{
				init();
			}
		
			if (pushTimeIntervalForward == true)
			{
				//increment execution count
				MotionGenericOperator.executionIterationCount++;
				//errorcommentMotionGenericOperator.numberOfSupportedObjects = 0;
				pushTimeIntervalForward = false;
			}
		
			if(DEBUG == 1)
			{
				//errorcommentHelper.DebugWriteLine("\nNumber of supported objects = "+ MotionGenericOperator.numberOfSupportedObjects);
			}
		
			//set the delta
			MotionGenericOperator.delta = 10;

			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Current execution count = " + MotionGenericOperator.executionIterationCount);
				Helper.DebugWriteLine("Current delta = " + MotionGenericOperator.delta);
			}
		
			//figure out how many tuples to dequeue from left and right queue
			XATQueue lQueue = this.getLeftInputQueue();
			XATQueue rQueue = this.getRightInputQueue();
			int size1 = lQueue.getQueueSize(this);
			int size2 = rQueue.getQueueSize(this);
			int lHowmany, rHowmany;

			int dequeueHowmany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));			
			if ((size1 + size2) != 0) 
			{
				lHowmany = (int) ((double) size1 / (size1 + size2) * dequeueHowmany);
				lHowmany = (size1 < lHowmany) ? size1 : lHowmany;
				rHowmany = (size2 < dequeueHowmany - lHowmany)? size2 : (dequeueHowmany - lHowmany);
					
				try 
				{
					lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
					rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
				} 
				catch(Exception e) 
				{
					System.out.println(e);
				}
			} 
			else 
			{
				try 
				{
					lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "0");
					rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "0");
				} 
				catch (Exception e) 
				{
				System.out.println(e);
				}
			/* nothing to run */
			return false;
			}
		
			if(DEBUG == 1)
			{
				//Helper.DebugWriteLine("Left queue tuples = " + lHowmany);
				//Helper.DebugWriteLine("Right queue tuples = " + rHowmany);
			}
			//end of checking how many results to dequeue
	
			executeMotionOperator(lHowmany, rHowmany);
		}
		catch(Exception e) 
		{
			e.printStackTrace();
			Helper.DebugWriteLine("OOH SOMETHING BAD HAPPENED");
		}
		/* run successfully */
		return true;
	}
	
	private XATTuple[] sortQueueArrayByTimeStamp(XATTuple[] tuples)
	{
		ArrayList tempTuples = new ArrayList();
		
		Arrays.sort(tuples, 
		new Comparator() 
		{
            public int compare(Object a, Object b) 
            {
            	int rc = 0; //return code
                
            	XATTuple tuplea = (XATTuple)a;
            	XATTuple tupleb = (XATTuple)b;
                
            	//check for tupleb, just incase if we are ordering odd number of tuples 
            	if ((tupleb != null) && (tuplea != null))
            	{            		
            		//--- Sort by timestamp, 
            		//  otherwise alphabetical ignoring case.
            		if (tuplea.getCell(4).convertToDouble().longValue() < tupleb.getCell(4).convertToDouble().longValue()) 
            		{
            			rc = -1;
            		} 
            		else if (tuplea.getCell(4).convertToDouble().longValue() > tupleb.getCell(4).convertToDouble().longValue()) 
            		{
            			rc = 1;
            		}
            	}
                return rc;
            }
		}
       );		
	   		
		//now remove duplicates from the array
		int i = 0;
		while (i < tuples.length)
		{
			int k=0;
			//add the object to the array list
			tempTuples.add(tuples[i]);

			
			if ((i + 1) < tuples.length)
			{
				String action1 = tuples[i].getCell(0).toString();
				String action2 = tuples[i+1].getCell(0).toString();
				int objectId1 = tuples[i].getCell(1).convertToDouble().intValue();
				int objectId2 = tuples[i+1].getCell(1).convertToDouble().intValue();
				int reportNum1 = tuples[i].getCell(2).convertToDouble().intValue();
				int reportNum2 = tuples[i+1].getCell(2).convertToDouble().intValue();
				long currentX1 = tuples[i].getCell(5).convertToDouble().intValue();
				long currentX2 = tuples[i+1].getCell(5).convertToDouble().intValue();
				long currentY1 = tuples[i].getCell(6).convertToDouble().intValue();
				long currentY2 = tuples[i+1].getCell(6).convertToDouble().intValue();
			
			while ((objectId1 == objectId2) &&
				   (reportNum1 == reportNum2) &&
				   (currentX1 == currentX2) &&
				   (currentY1 == currentY2) && (i+k < tuples.length))
				{				
					k++;
					if ((i + k + 1) < tuples.length)
					{
						if ((tuples[i+k] != null) && (tuples[i+k+1] != null))
						{
						objectId1 = tuples[i+k].getCell(1).convertToDouble().intValue();
						objectId2 = tuples[i+k+1].getCell(1).convertToDouble().intValue();
						reportNum1 = tuples[i+k].getCell(2).convertToDouble().intValue();
						reportNum2 = tuples[i+k+1].getCell(2).convertToDouble().intValue();
						currentX1 = tuples[i+k].getCell(5).convertToDouble().intValue();
						currentX2 = tuples[i+k+1].getCell(5).convertToDouble().intValue();
						currentY1 = tuples[i+k].getCell(6).convertToDouble().intValue();
						currentY2 = tuples[i+k+1].getCell(6).convertToDouble().intValue();
						}
					}					
				}
			}

			i = i + k + 1;			
		}
		
		tuples = (XATTuple[])(tempTuples.toArray(new XATTuple[] {}));		
	   //return sorted duplicate free tuples		
	   return tuples;
	}
	

	
	
	
	private void executeMotionOperator(int lHowmany, int rHowmany) 
	{		
		XATQueue oQueue = this.getOutputQueue();
		
		//get the input queues for objects and queries
		XATQueue linputQueue = this.getLeftInputQueue();
		XATQueue rinputQueue = this.getRightInputQueue();
		XATTuple[] tuplesMovingObjects = null;
		XATTuple[] tuplesMovingQueries = null;		
		
		XATTuple[] tuplesMovingObjectsToBeProcessed = null;
		XATTuple[] tuplesMovingQueriesToBeProcessed = null;		
		
	
		this.currentTime = new Time();
					
		//this is our first time through
		//we set the execution count for later so we can check the tuples we are going to process
		if (MotionGenericOperator.executionIterationCount == 0)
		{
			MotionGenericOperator.executionIterationCount = 1;
		}
		currentTime.SetTime((MotionGenericOperator.executionIterationCount - 1) * MotionGenericOperator.delta);  
		
		if(DEBUG == 1)
		{
			//Helper.DebugWriteLine("Current time = " + currentTime.GetTime());
			Helper.DebugWriteLine("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");
		}
				
		
		if(lHowmany != 0) 
		{
			//dequeue lHowmany tuples from the input queue to be processed
			tuplesMovingObjects = linputQueue.dequeue(lHowmany, this);
						
			//get only the tuples that belong to the CURRENT time interval, the rest put back to the queue
			ArrayList tuplesToKeep = new ArrayList();
			ArrayList tuplesToReturn = new ArrayList();
			ArrayList tuplesToDrop = new ArrayList();
			for(int i=0; i<tuplesMovingObjects.length; i++)
			{
				if ((tuplesMovingObjects[i].getCell(4).convertToDouble().longValue() >= currentTime.GetTime()) &&
					(tuplesMovingObjects[i].getCell(4).convertToDouble().longValue() < currentTime.GetTime() + MotionGenericOperator.delta))
				{
					//keep the tuple
					tuplesToKeep.add(tuplesMovingObjects[i]);					
				}
				else if (tuplesMovingObjects[i].getCell(4).convertToDouble().longValue() < currentTime.GetTime())
				{
					tuplesToDrop.add(tuplesMovingObjects[i]);					
				}
				else
				{
					//return the tuple back to the queue
					tuplesToReturn.add(tuplesMovingObjects[i]);
				}
			}
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Moving Objects:");
				Helper.DebugWriteLine("---------------------------");
				Helper.DebugWriteLine("# of tuples to keep = ["+ tuplesToKeep.size()+"]");
				Helper.DebugWriteLine("# of tuples to return = ["+ tuplesToReturn.size()+"]");
				Helper.DebugWriteLine("# of tuples to drop = ["+ tuplesToDrop.size()+"]");
			}
			
			if (tuplesToReturn.size() > 0)
			{
				//set the flag to increment evaluation count (need to move on to the next
				//time interval
				pushTimeIntervalForward = true;
			}

			//return the tuple back to the queue
			linputQueue.enqueue(tuplesToReturn, this);
			//drop the tuples
			tuplesToDrop = null;	
			
			tuplesMovingObjectsToBeProcessed = new XATTuple[tuplesToKeep.size()];
			tuplesMovingObjectsToBeProcessed = (XATTuple[])tuplesToKeep.toArray(tuplesMovingObjectsToBeProcessed);
						
			//there are new object updates that belong to this time interval
			if(tuplesToKeep.size() > 0)
			{			
				//sort the tuples dequeued by the timestamp (we assume they come ordered)
				tuplesMovingObjectsToBeProcessed = sortQueueArrayByTimeStamp(tuplesMovingObjectsToBeProcessed);
		
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("# of moving objects' updates = ["+ tuplesMovingObjectsToBeProcessed.length +"]");
				}
				
				//order our tuples into the arrary so they can be processed at delta
				recordUpdates(tuplesMovingObjectsToBeProcessed);
			}
			else
			{
				//if(DEBUG == 1)
				//{
				//	Helper.DebugWriteLine("# of moving objects' updates = [0]");
				//}
			}			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Location update phase for moving objects is done.");			
			}
		}
		
		if(pushTimeIntervalForward == true)
		{
			groupTuples();
			
			//tuples to be sent to the output queue
			XATTupleImp[] outputTuples = new XATTupleImp[tuplesToCluster.length + 1];
			//value of one tuple
			XATDataValueImp[] outputValue = new XATDataValueImp[5];
			int tupCount = 0;		//current count of tuples for the output queue
			int tupAmount = 0; 		//current total amount of tuples
			if(tuplesToCluster != null)
			{
				//loop through all the object positions for the time interval
				for (int objs=0; objs<1000; objs++)
				{
					//only process the clusters that are not null(points exist)
					//and that have more than 2 points in the circle
					if(tuplesToCluster[objs] != null)
					{
						if(tuplesToCluster[objs].getSize() > 2)
						{
							//calculate the circle for the cluster
							tuplesToCluster[objs].calcCircle();
								
							//place the values from our tuple into the output tuple
							outputValue[0] = new XATDataValueImp(new Long(MotionGenericOperator.delta + (int)(currentTime.GetTime())));
							outputValue[1] = new XATDataValueImp(new Long(objs));
							outputValue[2] = new XATDataValueImp(new Double(tuplesToCluster[objs].getCircle().getPoint().getX()));
							outputValue[3] = new XATDataValueImp(new Double(tuplesToCluster[objs].getCircle().getPoint().getY()));
							outputValue[4] = new XATDataValueImp(new Double(Math.pow(tuplesToCluster[objs].getCircle().getRadius(),2) * Math.PI));
							
							//add the output tuple to the output array
							outputTuples[tupCount] = new XATTupleImp(outputValue);
							tupAmount += tuplesToCluster[objs].getSize();
							
							tupCount++;		//increment our tuple count
						}
					}
				}
				//send the results to the output queue
				this.getOutputQueue().enqueue(outputTuples, this);
				
				tuplesToCluster = null;
				recentObjects = null;
			}
		}
		
		if(DEBUG == 1)
		{
			//Helper.DebugWriteLine("# of objects in the Objects table: "+this.objectsTable.GetAllObjects().size());
		}
	}	
}
