/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.clusteroperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.FixedContinuous3DGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GenericSpatioTemporalGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.collections.MotionContinuousResultsCollection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections.MovingObjectsStatesCollection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections.MovingQueriesStatesCollection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import java.util.Vector;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.*;
/**
 * @author Rimma V. Nehme
 * @author Robert Contois
 * 
 * This is the generic class that represents a cluster and has some common methods used by all cluster operators
 */
public abstract class MotionGenericClusterOperator extends MotionGenericOperator
{
	protected MovingObjectsStatesCollection objectsTable;
	protected MovingQueriesStatesCollection queriesTable;
	Time currentTime;
	Cluster[] tuplesToCluster = null;
	Vector recentObjects = null;
	
	
	public MotionGenericClusterOperator(XATNode node) 
	{
		super(node);	
	}
	
	public void initState(String statetype) 
	{		
		String statetypeMovingObjects = "edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections.MovingObjectsStatesCollection";
		String statetypeMovingQueries = "edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections.MovingQueriesStatesCollection";
		try
		{				
			states = new OperatorStateImp[2];
			states[0] = (MovingObjectsStatesCollection)Class.forName(statetypeMovingObjects).newInstance();
			states[1] = (MovingObjectsStatesCollection)Class.forName(statetypeMovingQueries).newInstance();
			
			states[0].setOperator(this);
			states[1].setOperator(this);	
			
			this.objectsTable = (MovingObjectsStatesCollection)this.states[0];		
			this.queriesTable = (MovingQueriesStatesCollection)this.states[1];
			
			currentTime = new Time();
		}
		catch(Exception e) 
		{
			e.printStackTrace();	
		}		
	}
	
	public OperatorState getLeftState() 
	{
		return states[0];
	}
	
	public int getLeftStateSize() 
	{
		return states[0].getStateSize();
	}

	public OperatorState getRightState() 
	{
		return states[1];
	}
	
	public int getRightStateSize() 
	{
		return states[1].getStateSize();
	}

	public int getStateSize() 
	{
		return (states[0].getStateSize() + states[1].getStateSize());
	}
	
	/*
	 * recompute the unmatched states inside this operator.
	 * Usually the unmatched states are computed based on the matched states inside one of the 
	 * children operator of this operator.
	 * 
	 * return: 
	 * TRUE -- if the states (inside child operator) that the recomputation of the unmatched states are based on are 
	 *      all matched, and the recomputation is successful.
	 * FALSE -- if any of the states (inside child operator) that are used to recompute the unmatched states
	 *          is unmatched, or the recomputation is unsucessful.  
	 */
	public boolean recomputeUnmatchedState() 
	{
		boolean returnvalue = true;
		
		return returnvalue;	
	}

	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() 
	{
		if (states == null) 
		{
			return false;
		}

		for (int i = 0; i < states.length; i++) 
		{
			if (!states[i].isMatched())
				return true;
		}
		return false;
	}

	/*
	 * takes the tuples from the execute operation that need to be processed
	 * during this execution and add them to the tuples to cluster array.
	 * Nothing is returned since an interal structure is being modified.
	 * 
	 * @param tuplesMovingObjectsToBeProcessed	an array of tuples that need to be added to our recent updates array
	 * 
	 */
	public void recordUpdates(XATTuple[] tuplesMovingObjectsToBeProcessed){
		int timedir;		//This is the current time from the tuple we are processing
		int objectdir;		//The current tuples object id
		double x_coord;		//the xcoord of the object
		double y_coord;		//the y coord of the object
		
		//create a vector to store only the most recent updates for an object
		if(recentObjects == null)
		{
			recentObjects = new Vector();
		}
		
		//loop through the objects we need to process
		for (int i=0; i < tuplesMovingObjectsToBeProcessed.length; i++)
		{
			//parse the tuple into its respective pieces
			timedir = (int)Double.parseDouble(tuplesMovingObjectsToBeProcessed[i].getCell(4).convertToDouble().toString());
			timedir = timedir - (int)(currentTime.GetTime());
			objectdir = (int)Double.parseDouble(tuplesMovingObjectsToBeProcessed[i].getCell(1).convertToDouble().toString());			
			x_coord = (double)Double.parseDouble(tuplesMovingObjectsToBeProcessed[i].getCell(5).convertToDouble().toString());
			y_coord = (double)Double.parseDouble(tuplesMovingObjectsToBeProcessed[i].getCell(6).convertToDouble().toString());

			//ensure the vector is large enough to set it, if not increase the size
			if(recentObjects.size() <= objectdir)
			{
				recentObjects.setSize(objectdir+1);
			}
			
			//check to see if we do not have an update
			if(recentObjects.get(objectdir) == null)
			{
				//no recent updates so we are going to add one
				recentObjects.set(objectdir, tuplesMovingObjectsToBeProcessed[i]);
			}
			else
			{
				//otherwise we need to check and ensure the one being added is the most recent
				int timeFromVector = (int)Double.parseDouble(((XATTuple)recentObjects.get(objectdir)).getCell(4).convertToDouble().toString());
				if(timeFromVector > timedir)
				{
					//add the newer update
					recentObjects.set(objectdir, tuplesMovingObjectsToBeProcessed[i]);
				}
			}
		}
	}
	
	/*
	 * Group the tuples based on their class type id into an array hashed on that id.
	 * There is no param or return because it modifies an interal structure.
	 */
	public void groupTuples(){
		int timedir;		//This is the current time from the tuple we are processing
		int objectdir;		//The current tuples object class
		double x_coord;		//the xcoord of the object
		double y_coord;		//the y coord of the object
		
		if(tuplesToCluster == null)
		{
			//setting the max number of groups
			tuplesToCluster = new Cluster[1000];
		}
		
		if(recentObjects != null)
		{
			//loop through the objects we need to process
			for (int i=0; i < recentObjects.size(); i++)
			{
				if(recentObjects.get(i) != null)
				{
					//parse the tuple into tis respective pieces
					timedir = (int)Double.parseDouble(((XATTuple)recentObjects.get(i)).getCell(4).convertToDouble().toString());
					timedir = timedir - (int)(currentTime.GetTime());
					objectdir = (int)Double.parseDouble(((XATTuple)recentObjects.get(i)).getCell(3).convertToDouble().toString());			
					x_coord = (double)Double.parseDouble(((XATTuple)recentObjects.get(i)).getCell(5).convertToDouble().toString());
					y_coord = (double)Double.parseDouble(((XATTuple)recentObjects.get(i)).getCell(6).convertToDouble().toString());

					if(tuplesToCluster[objectdir] == null)
					{
						tuplesToCluster[objectdir] = new Cluster();
					}
					//add the point to the cluster
					tuplesToCluster[objectdir].addPoint(x_coord, y_coord);
				}
			}

		}
	}
}
