/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.rangeoperator;
import java.util.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.hash.DoubleKeyHash;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.collections.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous.MotionContinuousState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;


/**
 * @author Rimma V. Nehme
 *
 *@version 1.0
 */
public class MotionContinuousRangeQueryOperator extends MotionRangeQueryOperator
{
	//debug flag
	/** Field DEBUG */
	public static int DEBUG = 0; //1=true, 0=false
	/** Field STATS_FLAG */
	public static int STATS_FLAG = 0; //1=verbose, 0=short
	/** Field RESULTS_FLAG */
	public static int RESULTS_FLAG = 1; //1=write answers, 0=don't write  answers
										//2=store and write files only when oid != qid
										//3=dump results one-by-one immediately after join
										//without any post-processing
	
	MotionContinuousResultsCollection results;
	FixedContinuous3DGrid grid;

	private static boolean isTheFirstIteration = true;
	boolean pushTimeIntervalForwardMovingObjects = false;
	boolean pushTimeIntervalForwardMovingQueries = false;
	MotionStatisticsGatherer motionOperatorStatistics;
	
	/** Field futureTuplesMovingObjects */
	public static Vector futureTuplesMovingObjects = new Vector();
	/** Field futureTuplesMovingQueries */
	public static Vector futureTuplesMovingQueries = new Vector();
	
	
	private class FutureTuplesTimeStampIndex
	{
		/** Field timestamp */
		public int timestamp;
		/** Field arrayIndex */
		public int arrayIndex;
		
		/**
		 * Method FutureTuplesTimeStampIndex
		 * @param newTimestamp
		 * @param newArrayIndex
		 */
		public FutureTuplesTimeStampIndex(int newTimestamp, int newArrayIndex)
		{
			timestamp = newTimestamp;
			arrayIndex = newArrayIndex;			
		}
	}
	
	/**
	 * Method MotionContinuousRangeQueryOperator
	 * @param node
	 */
	public MotionContinuousRangeQueryOperator(XATNode node) 
	{
		super(node);	
		
		if(results != null)
		{
			//clear the incremental results collection
			results.Clear();
		}
		
		//initialize the grid
		grid  = new FixedContinuous3DGrid(GenericSpatioTemporalGrid.GetXDimensionCellCount(),
										  GenericSpatioTemporalGrid.GetYDimensionCellCount(),
										  GenericSpatioTemporalGrid.GetTDimensionCellCount());
		
	}
	


	
	
	/**
	 * Method run
	 * @see MotionGenericOperator#run()
	 */
	public boolean run() 
	{
		//statistics
		motionOperatorStatistics = new MotionStatisticsGatherer();
		
		try
		{
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nINSIDE CONTINUOUS MOTION RANGE OPERATOR!!!");
			}
			
			//System.out.println("\nINSIDE CONTINUOUS MOTION RANGE OPERATOR!!!");
			
			if ((this.getLeftInputQueue().getSchema() == null) || (this.getRightInputQueue().getSchema() == null))
			{
				return false;
			}
			if(this.getOutputQueue().getSchema() == null)
			{
				init();
			}
		
			if ((pushTimeIntervalForwardMovingObjects == true) && (pushTimeIntervalForwardMovingQueries == true))		
			{
				//increment execution count
				MotionGenericOperator.executionIterationCount++;
				pushTimeIntervalForwardMovingObjects = false;
				pushTimeIntervalForwardMovingQueries = false;
			}
		
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nNumber of supported objects = "+ MotionGenericOperator.numberOfSupportedObjects);
				Helper.DebugWriteLine("Number of supported queries = "+ MotionGenericOperator.numberOfSupportedQueries +"\n");
				Helper.DebugWriteLine("Current execution count = " + MotionGenericOperator.executionIterationCount);
				Helper.DebugWriteLine("Current delta = " + MotionGenericOperator.delta);
			}
		
			//set the delta
			MotionGenericOperator.delta = 2;
		
			//initialize results structure
			results = new MotionContinuousResultsCollection();
			
			motionOperatorStatistics.memoryBefore = MotionStatisticsGatherer.getUsedMemory();
			//statistics
			motionOperatorStatistics.executionCount = MotionGenericOperator.executionIterationCount + 1;
			motionOperatorStatistics.timeOverallStart = System.currentTimeMillis();
		
			//run the operator
			executeMotionOperator();
			
			motionOperatorStatistics.timeOverallEnd= System.currentTimeMillis();
			//write the statistics out
			MotionStatisticsGatherer.totalTimeElapsed += (motionOperatorStatistics.timeOverallEnd - motionOperatorStatistics.timeOverallStart);
			
			//after memory
			motionOperatorStatistics.memoryBefore = MotionStatisticsGatherer.getUsedMemory();
			//write the statistics out (only if join occurred)
			if ((pushTimeIntervalForwardMovingObjects == true) && (pushTimeIntervalForwardMovingQueries == true))
			{	
				if (STATS_FLAG == 1)
				{
					//verbose statistics
					Helper.StatsWriteLine(motionOperatorStatistics.ContinuousStatisticsToString());
				}
				else
				{
					//short statistics
					Helper.StatsWriteLine(motionOperatorStatistics.ShortContinuousStatisticsToString());
				}
			}
		}
		catch(Exception e) 
		{
			Helper.DebugWriteLine("OOH SOMETHING BAD HAPPENED");
			Helper.DebugWriteLine(e.getMessage());
		}
		
		
		/* run successfully */
		return true;
	}


	private void trajectoryUpdatePhaseMovingObject(Time currentTime, 
			   									   int delta, 
												   MovingObjectState movingObjectUpdate)
	{		
		if (objectsHash.getByFirstKey(movingObjectUpdate.GetID()) == null )
		{
			numberOfSupportedObjects++;	
			MotionStatisticsGatherer.totalNumberObjectsEnteredIntoHashTable++;
		}
	
		objectsHash.add(movingObjectUpdate.GetID(), (int)movingObjectUpdate.GetPosition().GetT(), movingObjectUpdate);
		MotionStatisticsGatherer.totalNumberObjectsLocationUpdates++;
	}
	
	private void trajectoryUpdatePhaseMovingQuery(Time currentTime, 
			   									  int delta, 
												  MovingQueryState movingQueryUpdate)
	{
        if (queriesHash.getByFirstKey(movingQueryUpdate.GetID()) == null )
		{
			numberOfSupportedQueries++;	
			MotionStatisticsGatherer.totalNumberQueriesEnteredIntoHashTable++;
		}
		
		/*
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Inserting into hash table MOVING QUERY: "+movingQueryUpdate);			
		}
		*/
		
		queriesHash.add(movingQueryUpdate.GetID(), (int)movingQueryUpdate.GetPosition().GetT(), movingQueryUpdate);
		MotionStatisticsGatherer.totalNumberQueriesLocationUpdates++;
	}
	
	private void executeMotionOperator() 
	{		
		
		XATQueue oQueue = this.getOutputQueue();
		XATQueue linputQueue = this.getLeftInputQueue();
		XATQueue rinputQueue = this.getRightInputQueue();
		XATTuple[] tuplesMovingObjects = null;
		XATTuple[] tuplesMovingQueries = null;		
		
		//XATTuple[] tuplesMovingObjectsToBeProcessed = null;
		//XATTuple[] tuplesMovingQueriesToBeProcessed = null;		
		
		this.currentTime = new Time();
		int lHowmany, rHowmany;
		
		currentTime.SetTime(MotionGenericOperator.executionIterationCount * MotionGenericOperator.delta);  
		motionOperatorStatistics.intervalTimeStart = (int)currentTime.GetTime();
		motionOperatorStatistics.intervalTimeEnd = (int)(currentTime.GetTime()+MotionGenericOperator.delta);
		
		Helper.DebugWriteLine("\nCurrent active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");
		
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Current time = " + currentTime.GetTime());
			Helper.DebugWriteLine("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");
			Helper.DebugWriteLine("Starting Trajectory Phase...");
			Helper.DebugWriteLine("Updating trajectories of moving objects...");
		}
		
		//System.out.println("Current time = " + currentTime.GetTime());
		//System.out.println("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");

		
		lHowmany = linputQueue.getQueueSize();	
		rHowmany = rinputQueue.getQueueSize();
		
		if (lHowmany != 0)
		{
			//dequeue objects tuples
			tuplesMovingObjects = linputQueue.dequeue(lHowmany, this);			
		}
		
		if (rHowmany != 0)
		{
			//dequeue querie tuples
			tuplesMovingQueries = rinputQueue.dequeue(rHowmany, this);
		}
		
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("\nSize of left queue = " + lHowmany);
			Helper.DebugWriteLine("Size of right queue = " + rHowmany+"\n");
		}
		
		int veryFirstObjectUpdateTimestamp = -1;
		int veryFirstQueryUpdateTimestamp = -1;
		int veryFirstMostProgressiveTimestamp = -1;
		
		//adjust the starting timestamp (because the stream generator doesn't start streaming right from the beginning)
		//pick the most "progressive" (i.e., latest timestamp from either of the streams)
		if (isTheFirstIteration == true)
		{			
			if (tuplesMovingObjects.length > 0)
			{
				veryFirstObjectUpdateTimestamp = Integer.parseInt(tuplesMovingObjects[0].getCell(4).toString());
			}
			if (tuplesMovingQueries.length > 0)
			{
				veryFirstQueryUpdateTimestamp  = Integer.parseInt(tuplesMovingQueries[0].getCell(4).toString());
			}
			
			if (veryFirstObjectUpdateTimestamp >= veryFirstQueryUpdateTimestamp)
			{
				veryFirstMostProgressiveTimestamp = veryFirstObjectUpdateTimestamp;
			}
			else
			{
				veryFirstMostProgressiveTimestamp = veryFirstQueryUpdateTimestamp;
			}
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("The very first timestamp from objects stream = " + veryFirstObjectUpdateTimestamp);
				Helper.DebugWriteLine("The very first timestamp from queries stream = " + veryFirstQueryUpdateTimestamp);				
			}
			
			//System.out.println("The very first timestamp from objects stream = " + veryFirstObjectUpdateTimestamp);
			//System.out.println("The very first timestamp from queries stream = " + veryFirstQueryUpdateTimestamp);

			
			MotionGenericOperator.executionIterationCount = (int)(veryFirstMostProgressiveTimestamp/MotionGenericOperator.delta);
			currentTime.SetTime(MotionGenericOperator.executionIterationCount * MotionGenericOperator.delta);  					
			isTheFirstIteration = false;
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nAdjusting Time Interval...");
				Helper.DebugWriteLine("Current execution count = " + MotionGenericOperator.executionIterationCount);
				Helper.DebugWriteLine("Current time = " + currentTime.GetTime());
				Helper.DebugWriteLine("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");
			}			
		}//end of very first execution
		
		
		//statistics		
		motionOperatorStatistics.timeObjectsUpdateStart = System.currentTimeMillis();			
		
		//*** TRAJECTORY UPDATE PHASE ***//		
		if(lHowmany != 0) 
		{

			//get only the tuples that belong to the CURRENT time interval, the rest put back to the queue
			ArrayList tuplesToKeep = new ArrayList();
			//ArrayList tuplesToReturn = new ArrayList();
			ArrayList tuplesToDrop = new ArrayList();
			
			//if(DEBUG == 1)
			//{
				Helper.DebugWriteLine("First tuple on the queue (MO): " + tuplesMovingObjects[0].getCell(0)+" | "+ tuplesMovingObjects[0].getCell(1)+" | "+ tuplesMovingObjects[0].getCell(2)+" | "+ tuplesMovingObjects[0].getCell(3)+" | "+ tuplesMovingObjects[0].getCell(4)+" | "+ tuplesMovingObjects[0].getCell(5)+" | "+ tuplesMovingObjects[0].getCell(6)+" | "+ tuplesMovingObjects[0].getCell(7)+" | "+ tuplesMovingObjects[0].getCell(8));
				Helper.DebugWriteLine("Last tuple on the queue (MO): " + tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(0)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(1)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(2)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(3)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(4)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(5)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(6)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(7)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(8));
			//	Helper.DebugWriteLine(" *** ");
				Helper.DebugWriteLine("Left queue tuples = " + lHowmany);
			//}
			
			//System.out.println("First tuple on the queue (MO): " + tuplesMovingObjects[0].getCell(0)+" | "+ tuplesMovingObjects[0].getCell(1)+" | "+ tuplesMovingObjects[0].getCell(2)+" | "+ tuplesMovingObjects[0].getCell(3)+" | "+ tuplesMovingObjects[0].getCell(4)+" | "+ tuplesMovingObjects[0].getCell(5)+" | "+ tuplesMovingObjects[0].getCell(6)+" | "+ tuplesMovingObjects[0].getCell(7)+" | "+ tuplesMovingObjects[0].getCell(8));
			//System.out.println("Last tuple on the queue (MO): " + tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(0)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(1)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(2)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(3)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(4)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(5)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(6)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(7)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(8));

			
			for(int i=0; i<tuplesMovingObjects.length; i++)
			{											
				if ((Long.parseLong(tuplesMovingObjects[i].getCell(4).toString()) >= currentTime.GetTime()) &&
					(Long.parseLong(tuplesMovingObjects[i].getCell(4).toString()) < (currentTime.GetTime() + MotionGenericOperator.delta)))
				{
					tuplesToKeep.add(tuplesMovingObjects[i]);					
				}
				else if (Long.parseLong(tuplesMovingObjects[i].getCell(4).toString()) < currentTime.GetTime())
				{
					tuplesToDrop.add(tuplesMovingObjects[i]);					
				}
				else
				{								
					futureTuplesMovingObjects.add(tuplesMovingObjects[i]);
				}			
			}
			
			//Helper.DebugWriteLine(futureTuplesMovingObjects.toString());
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nMoving Objects:");
				Helper.DebugWriteLine("---------------------------");
				Helper.DebugWriteLine("# of tuples to keep = ["+ tuplesToKeep.size()+"]");
				Helper.DebugWriteLine("# of tuples to (return) - future tuples = ["+ futureTuplesMovingObjects.size()+"]");			
				Helper.DebugWriteLine("# of tuples to drop = ["+ tuplesToDrop.size()+"]");
			}

			motionOperatorStatistics.tuplesObjectsDequeued = lHowmany;
			motionOperatorStatistics.tuplesObjectsKept = tuplesToKeep.size();
			motionOperatorStatistics.tuplesObjectsReturned = futureTuplesMovingObjects.size();
			motionOperatorStatistics.tuplesObjectsDropped = tuplesToDrop.size();
			
			//drop the tuples
			tuplesToDrop = null;	
			
			
			if (futureTuplesMovingObjects.size() > 0)
			{
				XATTuple xt = (XATTuple)futureTuplesMovingObjects.get(futureTuplesMovingObjects.size() - 1);
				float lastTimestamp = Float.parseFloat(xt.getCell(4).toString());
				if (lastTimestamp >= (currentTime.GetTime() + MotionGenericOperator.delta))
				//if the last timestamp in the future tuples arraylist is greater 
				//than the current delta time interval, push the execution
				//set the flag to increment evaluation count (need to move on to the next time interval
				pushTimeIntervalForwardMovingObjects = true;
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("\n!!! SETTING PUSH FORWARD FOR MOVING OBJECTS TO TRUE !!!\n");
				}
			}

			
			//there are new object updates that belong to this time interval
			if(tuplesToKeep.size() > 0)
			{					
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("# of moving objects' updates = ["+ tuplesToKeep.size() +"]");
				}
				
				for (int i=0; i < tuplesToKeep.size(); i++)
				{							
					XATTuple tupleMovingObjectsToBeProcessed = (XATTuple)tuplesToKeep.get(i);
					MovingObjectState movingObjectUpdate = new MovingObjectState(
							Integer.parseInt(tupleMovingObjectsToBeProcessed.getCell(1).toString()),
							EntityType.Object,
							Long.parseLong(tupleMovingObjectsToBeProcessed.getCell(5).toString()),
							Long.parseLong(tupleMovingObjectsToBeProcessed.getCell(6).toString()),
							Integer.parseInt(tupleMovingObjectsToBeProcessed.getCell(4).toString()),
							UpdateType.FromStringToEnum(tupleMovingObjectsToBeProcessed.getCell(0).toString()),
							ClassType.FromIntToEnum(Integer.parseInt(tupleMovingObjectsToBeProcessed.getCell(3).toString())));
					
					
					//if (movingObjectUpdate.GetUpdateType() == UpdateType.DissappearPoint)
					//{
					//	//delete the object from the system
					//	objectsHash.removeByFirstKey(movingObjectUpdate.GetID());
					//	MotionStatisticsGatherer.totalNumberObjectsRemovedFromHashTable++;
					//}
					//else
					//{
						trajectoryUpdatePhaseMovingObject(currentTime, 
														  MotionGenericOperator.delta, 
													      movingObjectUpdate);
					//}
				}
			}	
			tuplesToKeep = null; //empty array of XAT tuples
		}
	
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Trajectory update for moving objects is done.");	
			Helper.DebugWriteLine("Updating trajectories of moving queries...");
		}
		
		//statistics
		motionOperatorStatistics.timeObjectsUpdateEnd = System.currentTimeMillis();
		MotionStatisticsGatherer.totalTimeElapsedObjectsUpdate += (motionOperatorStatistics.timeObjectsUpdateEnd - motionOperatorStatistics.timeObjectsUpdateStart);
		
		motionOperatorStatistics.timeQueriesUpdateStart = System.currentTimeMillis();
		

		//moving queries		
		if(rHowmany != 0) 
		{			
			//get only the tuples that belong to the CURRENT time interval, the rest put back to the queue
			ArrayList tuplesToKeep = new ArrayList();
			//ArrayList tuplesToReturn = new ArrayList();
			ArrayList tuplesToDrop = new ArrayList();

			//if(DEBUG == 1)
			//{
				Helper.DebugWriteLine("First tuple on the queue (MQ): " + tuplesMovingQueries[0].getCell(0)+" | "+ tuplesMovingQueries[0].getCell(1)+" | "+ tuplesMovingQueries[0].getCell(2)+" | "+ tuplesMovingQueries[0].getCell(3)+" | "+ tuplesMovingQueries[0].getCell(4)+" | "+ tuplesMovingQueries[0].getCell(5)+" | "+ tuplesMovingQueries[0].getCell(6)+" | "+ tuplesMovingQueries[0].getCell(7)+" | "+ tuplesMovingQueries[0].getCell(8));
				Helper.DebugWriteLine("Last tuple on the queue (MQ): " + tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(0)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(1)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(2)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(3)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(4)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(5)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(6)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(7)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(8));
				Helper.DebugWriteLine("Right queue tuples = " + rHowmany);
			//}
			
			//System.out.println("First tuple on the queue (MQ): " + tuplesMovingQueries[0].getCell(0)+" | "+ tuplesMovingQueries[0].getCell(1)+" | "+ tuplesMovingQueries[0].getCell(2)+" | "+ tuplesMovingQueries[0].getCell(3)+" | "+ tuplesMovingQueries[0].getCell(4)+" | "+ tuplesMovingQueries[0].getCell(5)+" | "+ tuplesMovingQueries[0].getCell(6)+" | "+ tuplesMovingQueries[0].getCell(7)+" | "+ tuplesMovingQueries[0].getCell(8));
			//System.out.println("Last tuple on the queue (MQ): " + tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(0)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(1)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(2)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(3)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(4)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(5)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(6)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(7)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(8));

			
			for(int i=0; i<tuplesMovingQueries.length; i++)
			{					
				if ((Long.parseLong(tuplesMovingQueries[i].getCell(4).toString()) >= currentTime.GetTime()) &&
				    (Long.parseLong(tuplesMovingQueries[i].getCell(4).toString()) < (currentTime.GetTime() + MotionGenericOperator.delta)))
				{
					tuplesToKeep.add(tuplesMovingQueries[i]);					
				}
				else if (Long.parseLong(tuplesMovingQueries[i].getCell(4).toString()) < currentTime.GetTime())
				{
					tuplesToDrop.add(tuplesMovingQueries[i]);					
				}			
				else
				{
					futureTuplesMovingQueries.add(tuplesMovingQueries[i]);
				}				
			}
			
			//Helper.DebugWriteLine(futureTuplesMovingQueries.toString());
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nMoving Queries:");
				Helper.DebugWriteLine("---------------------------");
				Helper.DebugWriteLine("# of tuples to keep = ["+ tuplesToKeep.size()+"]");
				Helper.DebugWriteLine("# of tuples to (return) - future tuples = ["+ futureTuplesMovingQueries.size()+"]");
				Helper.DebugWriteLine("# of tuples to drop = ["+ tuplesToDrop.size()+"]");
			}
			
			//statistics
			motionOperatorStatistics.tuplesQueriesDequeued = rHowmany;
			motionOperatorStatistics.tuplesQueriesKept = tuplesToKeep.size();
			motionOperatorStatistics.tuplesQueriesReturned = futureTuplesMovingQueries.size();
			motionOperatorStatistics.tuplesQueriesDropped = tuplesToDrop.size();
			
			//drop the tuples
			tuplesToDrop = null;
			
			
			
			if (futureTuplesMovingQueries.size() > 0)
			{
				XATTuple xt = (XATTuple)futureTuplesMovingQueries.get(futureTuplesMovingQueries.size() - 1);
				float lastTimestamp = Float.parseFloat(xt.getCell(4).toString());
				if (lastTimestamp >= (currentTime.GetTime() + MotionGenericOperator.delta))				
				{
					//set the flag to increment evaluation count (need to move on to the next time interval
					pushTimeIntervalForwardMovingQueries = true;
					if(DEBUG == 1)
					{
						Helper.DebugWriteLine("\n!!! SETTING PUSH FORWARD FOR MOVING QUERIES TO TRUE !!!\n");
					}
				}
			}
			
			
			//there are new object updates that belong to this time interval
			if(tuplesToKeep.size() > 0)
			{								
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("# of moving queries' updates = ["+ tuplesToKeep.size() +"]");
				}
						
				for (int i=0; i < tuplesToKeep.size(); i++)
				{
					XATTuple tuplesMovingQueryToBeProcessed = (XATTuple)tuplesToKeep.get(i);
					MovingQueryState movingQueryUpdate = new MovingQueryState(
							Integer.parseInt(tuplesMovingQueryToBeProcessed.getCell(1).toString()),
							EntityType.Query,
							Long.parseLong(tuplesMovingQueryToBeProcessed.getCell(5).toString()),
							Long.parseLong(tuplesMovingQueryToBeProcessed.getCell(6).toString()),
							Integer.parseInt(tuplesMovingQueryToBeProcessed.getCell(4).toString()),
							UpdateType.FromStringToEnum(tuplesMovingQueryToBeProcessed.getCell(0).toString()),
							ClassType.FromIntToEnum(Integer.parseInt(tuplesMovingQueryToBeProcessed.getCell(3).toString())),
							QueryType.Range,
							Integer.parseInt(tuplesMovingQueryToBeProcessed.getCell(10).toString()));					
					
					//if (movingQueryUpdate.GetUpdateType() == UpdateType.DissappearPoint)
					//{
						//delete the query from the system
					//	queriesHash.removeByFirstKey(movingQueryUpdate.GetID());
					//	MotionStatisticsGatherer.totalNumberQueriesRemovedFromHashTable++;
					//}
					//else
					//{
						trajectoryUpdatePhaseMovingQuery(currentTime, 
													 MotionGenericOperator.delta, 
													 movingQueryUpdate);
					//}
				}
				tuplesToKeep = null;
			}	
		
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Trajectory update for moving queries is done.");
			}	
			
			motionOperatorStatistics.timeQueriesUpdateEnd = System.currentTimeMillis();
			MotionStatisticsGatherer.totalTimeElapsedQueriesUpdate += (motionOperatorStatistics.timeQueriesUpdateEnd - motionOperatorStatistics.timeQueriesUpdateStart);
		}
		//*** TRAJECTORY UPDATE PHASE ***		
		
		//*** JOINING PHASE ***		
		if ((pushTimeIntervalForwardMovingObjects == true) && (pushTimeIntervalForwardMovingQueries == true))
		{		
			motionOperatorStatistics.timeFakeQObjectsStart = System.currentTimeMillis();
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nProcessing moving objects from the 'FAKE QUEUE'...");
				Helper.DebugWriteLine("Total entries on the objects 'FAKE QUEUE' = " + futureTuplesMovingObjects.size());
			}	
			

			if (futureTuplesMovingObjects.size() > 0)
			{
				int counter = 0;
				XATTuple tupleToBeProcessed = (XATTuple)futureTuplesMovingObjects.get(0);
				while (Float.parseFloat(tupleToBeProcessed.getCell(4).toString()) < (currentTime.GetTime() + MotionGenericOperator.delta))
				{	
					counter++;
					
					MovingObjectState movingObjectUpdate = new MovingObjectState(
								Integer.parseInt(tupleToBeProcessed.getCell(1).toString()),
								EntityType.Object,
								Long.parseLong(tupleToBeProcessed.getCell(5).toString()),
								Long.parseLong(tupleToBeProcessed.getCell(6).toString()),
								Integer.parseInt(tupleToBeProcessed.getCell(4).toString()),
								UpdateType.FromStringToEnum(tupleToBeProcessed.getCell(0).toString()),
								ClassType.FromIntToEnum(Integer.parseInt(tupleToBeProcessed.getCell(3).toString())));

	
					trajectoryUpdatePhaseMovingObject(currentTime, 
														  MotionGenericOperator.delta, 
														  movingObjectUpdate);									
													
					futureTuplesMovingObjects.remove(0);
					if (futureTuplesMovingObjects.size() == 0)
					{
						//fake queue is empty
						break;
					}
					tupleToBeProcessed = (XATTuple)futureTuplesMovingObjects.get(0);												
				}//end while
				
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Dequeued total "+counter+" objects updates from the 'FAKE QUEUE'...");
					Helper.DebugWriteLine("Now total entries on the objects 'FAKE QUEUE' = " + futureTuplesMovingObjects.size());
				}
				
			}//end if
			
			motionOperatorStatistics.timeFakeQObjectsEnd = System.currentTimeMillis();
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nProcessing moving queries from the 'FAKE QUEUE'...");
				Helper.DebugWriteLine("\nTotal entries on the queries 'FAKE QUEUE' = " + futureTuplesMovingQueries.size());
			}	
			
			motionOperatorStatistics.timeFakeQQueriesStart = System.currentTimeMillis();
						
			if (futureTuplesMovingQueries.size() > 0)
			{
				int counter = 0;
				XATTuple tupleToBeProcessed = (XATTuple)futureTuplesMovingQueries.get(0);
				while (Float.parseFloat(tupleToBeProcessed.getCell(4).toString()) < (currentTime.GetTime() + MotionGenericOperator.delta))
				{															
					counter++;
					MovingQueryState movingQueryUpdate = new MovingQueryState(
						Integer.parseInt(tupleToBeProcessed.getCell(1).toString()),
						EntityType.Query,
						Long.parseLong(tupleToBeProcessed.getCell(5).toString()),
						Long.parseLong(tupleToBeProcessed.getCell(6).toString()),
						Integer.parseInt(tupleToBeProcessed.getCell(4).toString()),
						UpdateType.FromStringToEnum(tupleToBeProcessed.getCell(0).toString()),
						ClassType.FromIntToEnum(Integer.parseInt(tupleToBeProcessed.getCell(3).toString())),
						QueryType.Range,
						Integer.parseInt(tupleToBeProcessed.getCell(10).toString()));		

						trajectoryUpdatePhaseMovingQuery(currentTime, 
								MotionGenericOperator.delta, 
								movingQueryUpdate);										
													
						futureTuplesMovingQueries.remove(0);
						if (futureTuplesMovingQueries.size() == 0)
						{
							//fake queue is empty
							break;
						}
						tupleToBeProcessed = (XATTuple)futureTuplesMovingQueries.get(0);												
				}//end while
				
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Dequeued total "+counter+" queries updates from the 'FAKE QUEUE'...");	
					Helper.DebugWriteLine("Now total entries on the objects 'FAKE QUEUE' = " + futureTuplesMovingObjects.size());
				}
			}//end if
			motionOperatorStatistics.timeFakeQQueriesEnd = System.currentTimeMillis();
						
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nStarting Joining Phase...\n");
			}	
			
			//System.out.println("STARTING JOIN...");
			
			//statistics
			motionOperatorStatistics.timeJoinStart = System.currentTimeMillis();			
			
			ArrayList miniClippedResults = joiningPhase(currentTime, MotionGenericOperator.delta);	

			//statistics
			motionOperatorStatistics.timeJoinEnd = System.currentTimeMillis();
			motionOperatorStatistics.numberOfIntersectionResults = miniClippedResults.size();			

			//System.out.println("DONE WITH JOIN...");
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Joining Phase is done");			
			}
		
			//*** MERGING PHASE ***
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("# of mini-clipped results = ["+miniClippedResults.size()+"]");
			}
		
			if (miniClippedResults.size() > 0)
			{	
				/*if(DEBUG == 1)
				{				
					for (int k=0; k < miniClippedResults.size();k++)
					{
						//MotionContinuousResult tmpRes = (MotionContinuousResult)miniClippedResults.get(k);					
						//Helper.DebugWriteLine("\nResult "+(k+1)+": \n"+tmpRes.toString());					
					}			
				}
				*/
				
				//System.out.println("STARTING MERGING...");
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Starting Merging Phase...");
				}
			
				motionOperatorStatistics.timeMergeStart = System.currentTimeMillis();
				MotionContinuousResultsCollection mergedResults = mergingPhase(currentTime, 
																			   MotionGenericOperator.delta, 
																			   miniClippedResults);			
				motionOperatorStatistics.timeMergeEnd = System.currentTimeMillis();
				motionOperatorStatistics.numberOfMergedResults  = mergedResults.GetContinuousResults().size();
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Merging Phase is done.");
					//Helper.DebugWriteLine("Merged Results Collection: " + mergedResults.toString());
				}
				//System.out.println("DONE WITH MERGING...");
				//*** MERGING PHASE ***
				
				//*** TIMELINE PHASE ***
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Starting Timeline Phase...");			
				}
			
				//System.out.println("STARTING TIMELINE...");
				motionOperatorStatistics.timeTimelineStart = System.currentTimeMillis();
				MotionContinuousResultsCollection sortedResults = timelinePhase(currentTime, 
																				MotionGenericOperator.delta, mergedResults);			
				motionOperatorStatistics.timeTimelineEnd = System.currentTimeMillis();
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Timeline Phase is done.");	
					//Helper.DebugWriteLine(sortedResults.toString());
				
					Helper.DebugWriteLine("# of big results (merged & sorted) = ["+sortedResults.GetContinuousResults().size()+"]");
					//ResultsRecorder.ResultsWriteLine("Current time"+this.currentTime.GetTime()+" # of big results (merged and sorted) = ["+sortedResults.GetContinuousResults().size()+"]");
				}
						
				//*** TIMELINE PHASE ***
				
				
				motionOperatorStatistics.timeOverallEnd = System.currentTimeMillis();
				//System.out.println("DONE WITH TIMELINE...");
				MotionStatisticsGatherer.totalTimeElapsed += (motionOperatorStatistics.timeOverallEnd - motionOperatorStatistics.timeOverallStart);
				
				//System.out.println("Writing results to file...");
				//flag to write out results is turned on
				if ((RESULTS_FLAG == 1) || (RESULTS_FLAG == 2) || (RESULTS_FLAG == 4)) 
				{				
					if (sortedResults.GetContinuousResults().size() > 0)
					{
						ResultsRecorder.ContinuousResultsWriteLine(sortedResults);
					}					
				}
				
				/*
				//System.out.println("Done writing results to file...");
				//write out the results to the file
				//print the results (for accounting purposes)
				if (sortedResults != null)
				{						
					//send results to the output queue
					int numberOfResults = sortedResults.GetContinuousResults().size();
					//tuples to be sent to the output queue
					XATTupleImp[] outputTuples = new XATTupleImp[numberOfResults];
					//value of one tuple
					XATDataValueImp[] outputValue = new XATDataValueImp[1];
					//array of results
			
					for(int i = 0; i < numberOfResults ; i++)
					{		
						MotionContinuousResult outputResult = (MotionContinuousResult)sortedResults.GetContinuousResults().get(i);
						outputValue[0] = new XATDataValueImp(outputResult);
						outputTuples[i] = new XATTupleImp(outputValue);
					}
			
					//send the results to the output queue
					this.getOutputQueue().enqueue(outputTuples, this);				
					outputTuples = null;
				}
				else
				{
					//no mini-clipped results, skipping the merging and timeline phase
					if(DEBUG == 1)
					{
						Helper.DebugWriteLine("No mini-clipped results, skipping the merging and timeline phase");				
					}
				}//end if pushExecutionForward
				*/
				sortedResults = null;
				
				motionOperatorStatistics.timeOverallStart = System.currentTimeMillis();
			
			}
						
			//clean the grid
			grid.Cleanup();
		
			//cleanup the Objects and Queries hash tables
			cleanupObjectsAndQueriesTables();
		
		}//end if push execution
			
		motionOperatorStatistics.numberOfObjectsInTheSystem = this.objectsHash.getNumberOfFirstKeyEntries();
		motionOperatorStatistics.numberOfQueriesInTheSystem = this.queriesHash.getNumberOfFirstKeyEntries();
		
		if(DEBUG == 1)
		{
			//print grid statistics		
			Helper.DebugWriteLine(grid.toString());
		}
		
		
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("# of objects in the Objects table: "+this.objectsHash.getNumberOfFirstKeyEntries());
			Helper.DebugWriteLine("# of queries in the Queries table: "+this.queriesHash.getNumberOfFirstKeyEntries());					
		}
		
		//System.out.println("DONE WITH OVERALL PERIODIC EXECUTION...");
		//call for garbage collection
		System.gc();
	 }

	
	private void cleanupObjectsAndQueriesTables()
	{
	    //Objects hash table		
		//objectsHash.forEachFirstKeyRemoveAllButLastTwoForSecondKey();
		//objectsHash.forEachFirstKeyRemoveAll();
		
				
		//Queires hash table
		//queriesHash.forEachFirstKeyRemoveAllButLastTwoForSecondKey();
		//queriesHash.forEachFirstKeyRemoveAll();
		
		
		
        //Objects hash table
		objectsHash.InternalHash.clear();
		
		
		//Queiries hash table
		queriesHash.InternalHash.clear();
	}
	
	
	private static MotionContinuousResultsCollection mergingPhase(Time T, 
														   int delta, 
														   ArrayList miniClippedResults)
	{	
		MotionContinuousResultsCollection results = new MotionContinuousResultsCollection();
		
		MultiMap  hmResults = new MultiMap();

		for (int i=0; i<miniClippedResults.size();i++)
		{			
			MotionContinuousResult mcr = (MotionContinuousResult)miniClippedResults.get(i);
			
			if (i > 0)
			{
				MotionContinuousResult mcrPrev = (MotionContinuousResult)miniClippedResults.get(i-1);
				if (mcrPrev.equals(mcr))
				{
					continue;
				}
			}
			
			Vector key = new Vector();
			key.add(Integer.toString(mcr.GetQueryID()));
			key.add(Integer.toString(mcr.GetObjectID()));
			hmResults.put(key.toString(), mcr);
		}
				
		Set set = hmResults.keySet();
		Iterator iter = set.iterator();
		int i=1;
		ArrayList tmp = new ArrayList();
		while(iter.hasNext())
		{
			Iterator iterator = hmResults.getValueIterator(iter.next());
			if (iterator.hasNext())
			{
				while (iterator.hasNext()) 
				{
					MotionContinuousResult mc = (MotionContinuousResult)iterator.next();
					tmp.add(mc);
				}

				//sort the array and merge it by timestamp
				Collections.sort(tmp, new Comparator() 
				{
					/**
					 * Method compare
					 * @see Comparator#compare(Object, Object)
					 */
					public int compare(Object a, Object b) 
					{
						int rc = 0; //return code
						                
					 	MotionContinuousResult s1 = (MotionContinuousResult)a;
					  	MotionContinuousResult s2 = (MotionContinuousResult)b;
						                
					    //--- Sort by timestamp, 
					    if (s1.GetResultValue().GetStartPosition().GetT() < s2.GetResultValue().GetStartPosition().GetT()) 
						{
						 	rc = -1;
					    } 
					    else if (s1.GetResultValue().GetStartPosition().GetT() > s2.GetResultValue().GetStartPosition().GetT()) 
						{
						    rc = 1;
						} 
						return rc;
					}
				});
				
				int k = 0;
				MotionContinuousResult res = null;
				MotionContinuousState resValue = null;
				int queryID = -1;
				int objectID = -1;
				float startTimestamp, intermTimestamp, endTimestamp;
				float p1StartT, p2StartT, p1EndT, p2EndT;
				
				if (tmp.size() > 0)
				{
					while(k < tmp.size())
					{
						MotionContinuousResult firstResult = (MotionContinuousResult)tmp.get(k);
						MotionContinuousResult currentResult = firstResult;				
						
						MotionContinuousState outputState = new MotionContinuousState(firstResult.GetResultValue().GetID(),
								firstResult.GetResultValue().GetEntityType(),
								firstResult.GetResultValue().GetStartPosition().GetX(),
								firstResult.GetResultValue().GetStartPosition().GetY(),
								firstResult.GetResultValue().GetStartPosition().GetT(),
								firstResult.GetResultValue().GetEndPosition().GetX(),
								firstResult.GetResultValue().GetEndPosition().GetY(),
								firstResult.GetResultValue().GetEndPosition().GetT(),
								firstResult.GetResultValue().GetContinuityType());						

						p1StartT = firstResult.GetResultValue().GetStartPosition().GetT();
						p1EndT = firstResult.GetResultValue().GetEndPosition().GetT();
						
						int b = k;
						
						if ((b+1) < tmp.size())
						{
							while(((++b) < tmp.size()))
							{		
								currentResult = (MotionContinuousResult)tmp.get(b);
								p2StartT = currentResult.GetResultValue().GetStartPosition().GetT();
								p2EndT = currentResult.GetResultValue().GetEndPosition().GetT();
								
								//two disjoint results
								if (p2StartT > p1EndT)
								{
									currentResult = (MotionContinuousResult)tmp.get(b - 1);
									break;
								}
								//second result embedded in the first
								else if((p2StartT >= p1StartT) && (p2EndT <= p1EndT))
								{
									//do nothing just keep the larger result
									//keep processing
								}
								//two intersecting results (share some time interval) - at the end time stamp
								else if ((p2StartT >= p1StartT) && (p2EndT > p1EndT))
								{
									outputState.SetEndPosition(currentResult.GetResultValue().GetEndPosition());
									p1EndT = p2EndT;
								}
								//two intersecting results (share some time interval) - at the start time stamp
								else if ((p2StartT < p1StartT) && (p2EndT <= p1EndT))
								{
									outputState.SetStartPosition(currentResult.GetResultValue().GetStartPosition());
									p1StartT = p2StartT;
								}
								//first result embedded in the second
								else if ((p2StartT < p1StartT) && (p2EndT > p1EndT))
								{
									outputState.SetStartPosition(currentResult.GetResultValue().GetStartPosition());
									outputState.SetEndPosition(currentResult.GetResultValue().GetEndPosition());
									p1StartT = p2StartT;
									p1EndT = p2EndT;
								}								
							}
						}
						else
						{
							k++;
						}						
						
						queryID = firstResult.GetQueryID();
						objectID = firstResult.GetObjectID();						
				
						//assign the merged result value
						resValue = outputState;					
				
						if (resValue != null)
						{
							res = new MotionContinuousResult(queryID, objectID, QueryType.Range, resValue);				
							results.Add(res);
						}				
				
						if (k < b)
						{
							k = b;
						}
					}//end outer while()
					
					//clear tmp arraylist for the next iteration
					tmp.clear();
				}//end if tmp.size() > 0
			}						
		}					
		return results;	
	}
	
	
	private static MotionContinuousResultsCollection timelinePhase(Time T, int delta, MotionContinuousResultsCollection mergedUnsortedResults)
	{
	
		//Sort the events by timestamp in the ascending order to imitate the events occurrence during time interval
		Collections.sort(mergedUnsortedResults.GetContinuousResults(), new Comparator() 
				{
				    /**
				     * Method compare
				     * @see Comparator#compare(Object, Object)
				     */
				    public int compare(Object a, Object b) 
				    {
				    	int rc = 0; //return code
				                
				    	MotionContinuousResult r1 = (MotionContinuousResult)a;
				    	MotionContinuousResult r2 = (MotionContinuousResult)b;
				                
				        //--- Sort by timestamp, 
				        if (r1.GetResultValue().GetStartPosition().GetT() < r2.GetResultValue().GetStartPosition().GetT()) 
				        {
				        	rc = -1;
				        } 
				        else if (r1.GetResultValue().GetStartPosition().GetT() > r2.GetResultValue().GetStartPosition().GetT()) 
				        {
				        	rc = 1;
				        } 
				        return rc;
				    }
				});
				
		//return now sorted (by timestamp)results
		return mergedUnsortedResults;				
	}	
	
	/*
	private static XATTuple[] sortQueueArrayByTimeStamp(XATTuple[] tuples)
	{
		ArrayList tempTuples = new ArrayList();
		
		Arrays.sort(tuples, 
		new Comparator() 
		{
            public int compare(Object a, Object b) 
            {
            	int rc = 0; //return code
                
            	XATTuple tuplea = (XATTuple)a;
            	XATTuple tupleb = (XATTuple)b;
                
            	//check for tupleb, just incase if we are ordering odd number of tuples 
            	if ((tupleb != null) && (tuplea != null))
            	{            		
            		//--- Sort by timestamp, 
            		//  otherwise alphabetical ignoring case.
            		if (tuplea.getCell(4).convertToDouble().longValue() < tupleb.getCell(4).convertToDouble().longValue()) 
            		{
            			rc = -1;
            		} 
            		else if (tuplea.getCell(4).convertToDouble().longValue() > tupleb.getCell(4).convertToDouble().longValue()) 
            		{
            			rc = 1;
            		}
            	}
                return rc;
            }
		}
       );		
	   		
		//now remove duplicates from the array
		int i = 0;
		while (i < tuples.length)
		{
			int k=0;
			//add the object to the array list
			tempTuples.add(tuples[i]);
			
			if ((i + 1) < tuples.length)
			{
				String action1 = tuples[i].getCell(0).toString();
				String action2 = tuples[i+1].getCell(0).toString();
				int objectId1 = tuples[i].getCell(1).convertToDouble().intValue();
				int objectId2 = tuples[i+1].getCell(1).convertToDouble().intValue();
				int reportNum1 = tuples[i].getCell(2).convertToDouble().intValue();
				int reportNum2 = tuples[i+1].getCell(2).convertToDouble().intValue();
				long currentX1 = tuples[i].getCell(5).convertToDouble().intValue();
				long currentX2 = tuples[i+1].getCell(5).convertToDouble().intValue();
				long currentY1 = tuples[i].getCell(6).convertToDouble().intValue();
				long currentY2 = tuples[i+1].getCell(6).convertToDouble().intValue();
			
			while ((objectId1 == objectId2)   &&
				   (reportNum1 == reportNum2) &&
				   (currentX1 == currentX2)   &&
				   (currentY1 == currentY2)   && (i+k < tuples.length))
				{				
					k++;
					if ((i + k + 1) < tuples.length)
					{
						if ((tuples[i+k] != null) && (tuples[i+k+1] != null))
						{
						objectId1  =  tuples[i+k].getCell(1).convertToDouble().intValue();
						objectId2  =  tuples[i+k+1].getCell(1).convertToDouble().intValue();
						reportNum1 =  tuples[i+k].getCell(2).convertToDouble().intValue();
						reportNum2 =  tuples[i+k+1].getCell(2).convertToDouble().intValue();
						currentX1  =  tuples[i+k].getCell(5).convertToDouble().intValue();
						currentX2  =  tuples[i+k+1].getCell(5).convertToDouble().intValue();
						currentY1  =  tuples[i+k].getCell(6).convertToDouble().intValue();
						currentY2  =  tuples[i+k+1].getCell(6).convertToDouble().intValue();
						}
					}					
				}
			}

			i = i + k + 1;			
		}
		
		tuples = (XATTuple[])(tempTuples.toArray(new XATTuple[] {}));		
	   //return sorted duplicate free tuples		
	   return tuples;
	}
	*/
	
	// Flushes all objects to the grid when the time interval expires
	//
	// @param       none
	// @return      none
	private void flushAllObjectsToGrid()
	{
		int totalObjectsFlushed = 0;
				
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Flushing Objects to grid...");
			Helper.DebugWriteLine("Total ["+objectsHash.getNumberOfFirstKeyEntries()+"] objects to flush");
		}
				
		int pointCount = 0;
		MovingObjectState movingObjectPoint1 = new MovingObjectState();
		MovingObjectState movingObjectPoint2 = new MovingObjectState();
		
		
		Iterator i = objectsHash.InternalHash.entrySet().iterator();
		while(i.hasNext())
		{			
			Map.Entry  entry =  (Map.Entry)i.next();			
			HashMap secondKeyInternalHash =(HashMap)entry.getValue();
		
			TreeMap treeSecondKeyInternalHash = new TreeMap(secondKeyInternalHash);		
			pointCount = 0;
			
			if (treeSecondKeyInternalHash.entrySet().size() >= 2)
			{
				Iterator i2 = treeSecondKeyInternalHash.entrySet().iterator();		
			
				while(i2.hasNext())
				{
					Map.Entry  entry2 =  (Map.Entry)i2.next();  	   
					if (pointCount == 0)
					{
						movingObjectPoint1 = (MovingObjectState)entry2.getValue();											
				   		pointCount = 1;
					}
					else if (pointCount == 1)
					{
						movingObjectPoint2 = (MovingObjectState)entry2.getValue();
						pointCount = 2;
					}
				   
					if (pointCount == 2)
					{
						MovingObjectContinuousState mc = new MovingObjectContinuousState(movingObjectPoint1.GetID(),
								movingObjectPoint1.GetEntityType(),
				   				movingObjectPoint1.GetPosition().GetX(),
				   				movingObjectPoint1.GetPosition().GetY(),
				   				movingObjectPoint1.GetPosition().GetT(),
				   				movingObjectPoint2.GetPosition().GetX(),
				   				movingObjectPoint2.GetPosition().GetY(),
				   				movingObjectPoint2.GetPosition().GetT(),
								FunctionType.Linear,
								movingObjectPoint1.GetObjectClass());

						/*
						if(DEBUG == 1)
						{
							Helper.DebugWriteLine("FLUSHING moving object:" + mc.toString());		
						}
						*/
						gridInsertObject(mc);
						totalObjectsFlushed++;
					    
						//set the 2nd point as the starting point for the next trajectory segment
						movingObjectPoint1 = movingObjectPoint2;
					       
						//reset the point count
						pointCount = 1;
					}				
				}
			}
			else
			{
				MotionStatisticsGatherer.numberObjectsLocationUpdatesNotParticipatingInJoin++;
			}
			
		}
		
		this.motionOperatorStatistics.numberOfObjectsBeingJoined = totalObjectsFlushed;

	}
	
	// Flushes all queries to the grid when the time interval expires
	//
	// @param       none
	// @return      none
	private void flushAllQueriesToGrid()
	{
		int totalQueriesFlushed = 0;
		
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Flushing Queries to grid...");
			Helper.DebugWriteLine("Total ["+queriesHash.getNumberOfFirstKeyEntries()+"] queries to flush");
		}
			
		int pointCount = 0;
		MovingQueryState movingQueryPoint1 = new MovingQueryState();
		MovingQueryState movingQueryPoint2 = new MovingQueryState();
		Map.Entry  entry, entry2 = null; 
		

		Iterator i = queriesHash.InternalHash.entrySet().iterator();
		while(i.hasNext())
		{			
			entry =  (Map.Entry)i.next();			
			HashMap secondKeyInternalHash =(HashMap)entry.getValue();
		
			TreeMap treeSecondKeyInternalHash = new TreeMap(secondKeyInternalHash);						
			pointCount = 0;
			
			if (treeSecondKeyInternalHash.entrySet().size() >= 2)
			{				
				Iterator i2 = treeSecondKeyInternalHash.entrySet().iterator();		
			
				while(i2.hasNext())
				{
					entry2 =  (Map.Entry)i2.next();  	   
			
					if (pointCount == 0)
					{
						movingQueryPoint1 = (MovingQueryState)entry2.getValue();
						pointCount = 1;
					}
					else if (pointCount == 1)
					{
						movingQueryPoint2 = (MovingQueryState)entry2.getValue();
						pointCount = 2;
					}
				
					if (pointCount == 2)
					{
						MovingQueryContinuousState mc = new MovingQueryContinuousState(movingQueryPoint1.GetID(),
								movingQueryPoint1.GetEntityType(),
								movingQueryPoint1.GetPosition().GetX(),
								movingQueryPoint1.GetPosition().GetY(),
								movingQueryPoint1.GetPosition().GetT(),
								movingQueryPoint2.GetPosition().GetX(),
								movingQueryPoint2.GetPosition().GetY(),
								movingQueryPoint2.GetPosition().GetT(),
								FunctionType.Linear,
								movingQueryPoint1.GetQueryClass(),
								movingQueryPoint1.GetQueryType(),
								movingQueryPoint1.GetQueryBoundary());			

						/*
						if(DEBUG == 1)
						{
							Helper.DebugWriteLine("FLUSHING moving query:" + mc.toString());		
						}
						*/
						gridInsertQuery(mc);
						totalQueriesFlushed++;
				    
						//set the 2nd point as the starting point for the next trajectory segment
						movingQueryPoint1 = movingQueryPoint2;
				       
						//reset the point count
						pointCount = 1;
					}
				}			
			}
			else
			{
				MotionStatisticsGatherer.numberQueriesLocationUpdatesNotParticipatingInJoin++;
			}
		}
		this.motionOperatorStatistics.numberOfQueriesBeingJoined = totalQueriesFlushed;
	}
	
	
	
	
	
	// Inserts a trajectory of a moving entity into a grid 
	// index.
	//
	// @param       
	// @return      none
	private void gridInsertObject(MovingObjectContinuousState mc)
	{
		boolean deleteFromSystem = false;
		
		grid.InsertObject(mc, deleteFromSystem, this.currentTime, MotionGenericOperator.delta);	
		
		if (deleteFromSystem == true)
		{
		    DeleteObjectFromSystemByID(mc.GetID());
			MotionGenericOperator.numberOfSupportedObjects--;
			MotionStatisticsGatherer.totalNumberObjectsRemovedFromHashTable++;
		}
	}
	
	// Inserts a trajectory of a moving entity into a grid 
	// index.
	//
	// @param       
	// @return      none
	private void gridInsertQuery(MovingQueryContinuousState mc)
	{
		boolean deleteFromSystem = false;
		
		grid.InsertQuery(mc, deleteFromSystem, this.currentTime, MotionGenericOperator.delta);	
		
		if (deleteFromSystem == true)
		{
		   //delete query
		   DeleteQueryFromSystemByID(mc.GetID());
		   MotionGenericOperator.numberOfSupportedQueries--;
		   MotionStatisticsGatherer.totalNumberQueriesRemovedFromHashTable++;
		}	
	}
	
	
	private void DeleteObjectFromSystemByID(int id)
	{
		objectsHash.removeByFirstKey(id);
	}
	
	private void DeleteQueryFromSystemByID(int id)
	{
		queriesHash.removeByFirstKey(id);
	}

	
	// Joins moving objects with moving queries. To be more specific, we go cell
	// by cell (in the grid) and join (i.e., calculate intersections) between
	// object trjectories (lines) and query trajectory volumes (cubes) and return
	// an array of mini-clipped results. (Mini-clipped results = line segments
	// at the granularity of grid cells that need to be merged to get the full
	// query answers.)
	// This is the third phase in our algorithm.
	//
	// @param       T    			MotionMovingEntityTimestamp (an operator-specific
	//								timestamp that starts counting off from the time
	//								the system started executing - 0..1..2..3..4, etc.)
	// @param       delta    		the length of the time interval 
	// @return      none
	// PHASE III - joining (i.e., calculating intersections between) moving objects and
	// moving queries.
	private ArrayList joiningPhase(Time T, int delta)
	{		
		ArrayList miniClippedResults = new ArrayList();
			
		//System.out.println("Flushing objects to grid...");
		flushAllObjectsToGrid();
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("\nDone flushing moving objects.");
			Helper.DebugWriteLine(grid.statisticsToString());
		}
		//System.out.println("Done flushing objects to grid...");
		//System.out.println("Flushing queries to grid...");
		flushAllQueriesToGrid();		
		if(DEBUG == 1)
		{			
			Helper.DebugWriteLine("\nDone flushing moving queries.\n");
			Helper.DebugWriteLine(grid.statisticsToString());
		}
		//System.out.println("Done flushing queries to grid...");
		//System.out.println(grid.statisticsToString());
		
		//clear the mini-clipped results ArrayList
		miniClippedResults.clear();
				
		//System.out.println("Doing cell-by-cell GRID JOIN...");
		
		//Loop through grid cells in the x-dimension
		for (int i=0; i < GenericSpatioTemporalGrid.GetXDimensionCellCount(); i++)
		{
			//Loop through grid cells in the y-dimension
 			for (int j=0; j < GenericSpatioTemporalGrid.GetYDimensionCellCount(); j++)
			{
				//Loop through grid cells in the t-dimension
				for (int k=0; k < GenericSpatioTemporalGrid.GetTDimensionCellCount(); k++)
				{
					if (grid.directory[i][j][k] != null) 
					{
					if (grid.directory[i][j][k].GetMovingQueries().size() > 0)
					{
						if(DEBUG == 1)
						{
							Helper.DebugWriteLine("Working with Cell: ["+i+"]["+j+"]["+k+"]" );
						}
						//System.out.println("Working with Cell: ["+i+"]["+j+"]["+k+"] #queries = " + grid.directory[i][j][k].GetMovingQueries().size() + " #objects = " + grid.directory[i][j][k].GetMovingObjects().size());
						//For each moving query MQ that has an entry in the grid cell G i,j,k
						for (int q=0; q < grid.directory[i][j][k].GetMovingQueries().size(); q++)
						{														
							//For each moving object MO that has an entry in the grid cell Gi,j,k  
							for (int o=0; o < grid.directory[i][j][k].GetMovingObjects().size(); o++)
							{	
								
								MovingQueryState mQ1 = null;
								MovingQueryState mQ2 = null;
								
								MovingObjectState mO1 = null;	
								MovingObjectState mO2 = null;	
			
								try{
							    GridContinuousEntry queryGridEntry = (GridContinuousEntry)grid.directory[i][j][k].GetMovingQueries().get(q);
								GridContinuousEntry objectGridEntry = (GridContinuousEntry)grid.directory[i][j][k].GetMovingObjects().get(o);


								 mQ1 = (MovingQueryState)queriesHash.get(queryGridEntry.GetID(), (int)queryGridEntry.GetTimeStart());
								 mQ2 = (MovingQueryState)queriesHash.get(queryGridEntry.GetID(), (int)queryGridEntry.GetTimeEnd());
								
								 mO1 = (MovingObjectState)objectsHash.get(objectGridEntry.GetID(), (int)objectGridEntry.GetTimeStart());	
								 mO2 = (MovingObjectState)objectsHash.get(objectGridEntry.GetID(), (int)objectGridEntry.GetTimeEnd());	
			
								 if ((mQ1 == null) || (mQ2 == null))
								 {
								 	if(DEBUG == 1)
									{
										Helper.DebugWriteLine("Null entry was retrieved for moving query with ID = " +queryGridEntry.GetID() +", start time = " +queryGridEntry.GetTimeStart() + ", end time = " + queryGridEntry.GetTimeEnd());
									}	
								 	//System.out.println("Null entry was retrieved for moving query with ID = " +queryGridEntry.GetID() +", start time = " +queryGridEntry.GetTimeStart() + ", end time = " + queryGridEntry.GetTimeEnd());
								 	continue;
								 }
								 
								 if ((mO1 == null) || (mO2 == null))
								 {
								 	if(DEBUG == 1)
									{
										Helper.DebugWriteLine("Null entry was retrieved for moving object with ID = " +objectGridEntry.GetID() +", start time = " +objectGridEntry.GetTimeStart() + ", end time = " + objectGridEntry.GetTimeEnd());										
									}
								 	//System.out.println("Null entry was retrieved for moving object with ID = " +objectGridEntry.GetID() +", start time = " +objectGridEntry.GetTimeStart() + ", end time = " + objectGridEntry.GetTimeEnd());
								 	continue;
								 }
								//create a continuous state out of two points
								MovingQueryContinuousState mQ = new MovingQueryContinuousState(mQ1.GetID(),
										mQ1.GetEntityType(),
										mQ1.GetPosition().GetX(),
										mQ1.GetPosition().GetY(),
										mQ1.GetPosition().GetT(),
										mQ2.GetPosition().GetX(),
										mQ2.GetPosition().GetY(),
										mQ2.GetPosition().GetT(),
										FunctionType.Linear,
										mQ1.GetQueryClass(),
										mQ1.GetQueryType(),
										mQ1.GetQueryBoundary());
								
								//create a continuous state out of two points
								MovingObjectContinuousState mO = new MovingObjectContinuousState(mO1.GetID(),
										mO1.GetEntityType(),
										mO1.GetPosition().GetX(),
										mO1.GetPosition().GetY(),
										mO1.GetPosition().GetT(),
										mO2.GetPosition().GetX(),
										mO2.GetPosition().GetY(),
										mO2.GetPosition().GetT(),
										FunctionType.Linear,
										mO1.GetObjectClass());


								
								//do the actual join
								MotionContinuousResult joinResult = grid.DoJoin(mQ, mO, new GridCellIndex3D(i,j,k));
							
								if (joinResult != null)
								{
									if (RESULTS_FLAG == 2) 
									{						
										if (mQ.GetID() != mO.GetID()) //store only the results where object id != query id
										{
											if (joinResult.GetResultValue().GetEndPosition().GetT() < joinResult.GetResultValue().GetStartPosition().GetT()) 
											{
												//switch the timestamped positions (smaller to bigger one)
												MotionContinuousState switchedContinuousState = new MotionContinuousState(joinResult.GetResultValue().GetID(),
														joinResult.GetResultValue().GetEntityType(),
														joinResult.GetResultValue().GetEndPosition().GetX(),
														joinResult.GetResultValue().GetEndPosition().GetY(),
														joinResult.GetResultValue().GetEndPosition().GetT(),
														joinResult.GetResultValue().GetStartPosition().GetX(),
														joinResult.GetResultValue().GetStartPosition().GetY(),
														joinResult.GetResultValue().GetStartPosition().GetT(),
														joinResult.GetResultValue().GetContinuityType());
																						
												MotionContinuousResult switchedJoinResult = new MotionContinuousResult(joinResult.GetQueryID(), 
														joinResult.GetObjectID(), 
														joinResult.GetQueryType(), 
														switchedContinuousState);
												
												miniClippedResults.add(switchedJoinResult);
											}
											else
											{									
												miniClippedResults.add(joinResult);
											}
										}
									}
									else if (RESULTS_FLAG == 3)
									{
										//write out the result and discard it
										ResultsRecorder.ContinuousSingleResultsWriteLine(joinResult);
										joinResult = null;
									}
									else if (RESULTS_FLAG == 4)
									{
										//store only a subset of results
										if (((mQ.GetID() % 10) == 0) && ((mO.GetID() % 10) == 0))
										{
											if (joinResult.GetResultValue().GetEndPosition().GetT() < joinResult.GetResultValue().GetStartPosition().GetT()) 
											{
												//switch the timestamped positions (smaller to bigger one)
												MotionContinuousState switchedContinuousState = new MotionContinuousState(joinResult.GetResultValue().GetID(),
														joinResult.GetResultValue().GetEntityType(),
														joinResult.GetResultValue().GetEndPosition().GetX(),
														joinResult.GetResultValue().GetEndPosition().GetY(),
														joinResult.GetResultValue().GetEndPosition().GetT(),
														joinResult.GetResultValue().GetStartPosition().GetX(),
														joinResult.GetResultValue().GetStartPosition().GetY(),
														joinResult.GetResultValue().GetStartPosition().GetT(),
														joinResult.GetResultValue().GetContinuityType());
																						
												MotionContinuousResult switchedJoinResult = new MotionContinuousResult(joinResult.GetQueryID(), 
														joinResult.GetObjectID(), 
														joinResult.GetQueryType(), 
														switchedContinuousState);
												
												miniClippedResults.add(switchedJoinResult);
											}
											else
											{									
												miniClippedResults.add(joinResult);
											}											
										}
									}
									else
									{	
									
									/*
									if(DEBUG == 1)
									{
										Helper.DebugWriteLine("Got a join result!" );
									}
									*/
									
									//DO SOME "DATA MASSAGING"
									//1) switch the timestamps (if it was a negative slope), we
									//might have received lower timestamp as the end timestamp
									
									if (joinResult.GetResultValue().GetEndPosition().GetT() < joinResult.GetResultValue().GetStartPosition().GetT()) 
									{
										//switch the timestamped positions (smaller to bigger one)
										MotionContinuousState switchedContinuousState = new MotionContinuousState(joinResult.GetResultValue().GetID(),
												joinResult.GetResultValue().GetEntityType(),
												joinResult.GetResultValue().GetEndPosition().GetX(),
												joinResult.GetResultValue().GetEndPosition().GetY(),
												joinResult.GetResultValue().GetEndPosition().GetT(),
												joinResult.GetResultValue().GetStartPosition().GetX(),
												joinResult.GetResultValue().GetStartPosition().GetY(),
												joinResult.GetResultValue().GetStartPosition().GetT(),
												joinResult.GetResultValue().GetContinuityType());
																				
										MotionContinuousResult switchedJoinResult = new MotionContinuousResult(joinResult.GetQueryID(), 
												joinResult.GetObjectID(), 
												joinResult.GetQueryType(), 
												switchedContinuousState);
										
										miniClippedResults.add(switchedJoinResult);
									}
									else
									{									
										miniClippedResults.add(joinResult);
									}
									}
								}
							}
							catch(Exception ex)
							{
								Helper.DebugWriteLine("Exception occurred" );									
							}
						}
					}
				}
				}
			}
		} 		
	}
	//System.out.println("Done doing cell-by-cell GRID JOIN...");
				
	return miniClippedResults;
}

}
