/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.rangeoperator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.MotionStatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.ClassType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.QueryType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.UpdateType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.hash.DoubleKeyHash;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.MemoryTestBench;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.ResultsRecorder;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.MotionGenericOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.MotionDiscreteResult;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.collections.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingObjectState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingQueryState;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
 * @author Rimma V. Nehme
 *
 *@version 1.0
 */
public class MotionDiscreteIncrementalFlushRangeQueryOperator extends MotionRangeQueryOperator
{
	//debug flag
	private MemoryTestBench memoryTestBench; 
	
	/** Field DEBUG */
	public static int DEBUG = 1; //1=true, 0=false
	/** Field STATS_FLAG */
	public static int STATS_FLAG = 0; //1=verbose, 0=short
	/** Field RESULTS_FLAG */
	public static int RESULTS_FLAG = 1; //1=return answers, 0=don't return  answers
										//2=store and write files only when oid != qid
										//3=dump results one-by-one immediately after join
										//without any post-processing
	
	
	private static boolean isTheFirstIteration = true;
	
	/** Field futureTuplesMovingObjects */
	public static Vector futureTuplesMovingObjects = new Vector();
	/** Field futureTuplesMovingQueries */
	public static Vector futureTuplesMovingQueries = new Vector();
	
	MotionDiscreteResultsCollection results;
	FixedDiscrete3DGrid grid;
	boolean pushTimeIntervalForwardMovingObjects = false;
	boolean pushTimeIntervalForwardMovingQueries = false;
	MotionStatisticsGatherer motionOperatorStatistics;

	
	private class FutureTuplesTimeStampIndex
	{
		/** Field timestamp */
		public int timestamp;
		/** Field arrayIndex */
		public int arrayIndex;
		
		/**
		 * Method FutureTuplesTimeStampIndex
		 * @param newTimestamp
		 * @param newArrayIndex
		 */
		public FutureTuplesTimeStampIndex(int newTimestamp, int newArrayIndex)
		{
			timestamp = newTimestamp;
			arrayIndex = newArrayIndex;			
		}
	}
	
	
	/**
	 * Method MotionDiscreteRangeQueryOperator
	 * @param node
	 */
	public MotionDiscreteIncrementalFlushRangeQueryOperator(XATNode node) 
	{
		super(node);	
		
		if(results != null)
		{
			//clear the incremental results collection
			results.Clear();
		}
		
		//initialize the grid
		grid  = new FixedDiscrete3DGrid(GenericSpatioTemporalGrid.GetXDimensionCellCount(),
										GenericSpatioTemporalGrid.GetYDimensionCellCount(),
										GenericSpatioTemporalGrid.GetTDimensionCellCount());
		
		memoryTestBench = new MemoryTestBench();
	}
	
	/**
	 * Method run
	 * @see MotionGenericOperator#run()
	 */
	public boolean run() 
	{
		//statistics
		motionOperatorStatistics = new MotionStatisticsGatherer();
		
		try
		{
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nINSIDE DISCRETE MOTION RANGE OPERATOR!!!");
			}
			
			//System.out.println("\nINSIDE DISCRETE MOTION RANGE OPERATOR!!!");
			
			
			if ((this.getLeftInputQueue().getSchema() == null) || (this.getRightInputQueue().getSchema() == null))
			{
				return false;
			}
			if(this.getOutputQueue().getSchema() == null)
			{
				init();
			}
		
			if ((pushTimeIntervalForwardMovingObjects == true) && (pushTimeIntervalForwardMovingQueries == true))
			{
				//increment execution count
				MotionGenericOperator.executionIterationCount++;
				pushTimeIntervalForwardMovingObjects = false;
				pushTimeIntervalForwardMovingQueries = false;
			}
		
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nNumber of supported objects = "+ MotionGenericOperator.numberOfSupportedObjects);
				Helper.DebugWriteLine("Number of supported queries = "+ MotionGenericOperator.numberOfSupportedQueries +"\n");
				Helper.DebugWriteLine("Current execution count = " + MotionGenericOperator.executionIterationCount);
				Helper.DebugWriteLine("Current delta = " + MotionGenericOperator.delta);
			}
		
			//set the delta
			MotionGenericOperator.delta = 1;

			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Current execution count = " + MotionGenericOperator.executionIterationCount);
				Helper.DebugWriteLine("Current delta = " + MotionGenericOperator.delta);
			}
		
			//initialize results structure
			results = new MotionDiscreteResultsCollection();
				
	
			motionOperatorStatistics.memoryBefore = MotionStatisticsGatherer.getUsedMemory();			
			//statistics			
			motionOperatorStatistics.executionCount = MotionGenericOperator.executionIterationCount + 1;
			motionOperatorStatistics.timeOverallStart = System.currentTimeMillis();
		
			//actual operator logic
			executeMotionOperator();
		

			MotionStatisticsGatherer.totalTimeElapsed += (motionOperatorStatistics.timeOverallEnd - motionOperatorStatistics.timeOverallStart);
			
			//memory after
			motionOperatorStatistics.memoryAfter = MotionStatisticsGatherer.getUsedMemory();
			
			//write the statistics out (only if join occurred)
			if ((pushTimeIntervalForwardMovingObjects == true) && (pushTimeIntervalForwardMovingQueries == true))
			{	
				if (STATS_FLAG == 1)
				{
					//verbose statistics
					Helper.StatsWriteLine(motionOperatorStatistics.DiscreteStatisticsToString());
				}
				else
				{
					//short statistics
					Helper.StatsWriteLine(motionOperatorStatistics.ShortDiscreteStatisticsToString());
				}
			}
		}
		catch(Exception e) 
		{
			Helper.DebugWriteLine("OOH SOMETHING BAD HAPPENED. Exception = " + e.toString());
			e.printStackTrace();
		}
		/* run successfully */
		return true;
	}

	/*
	private static XATTuple[] sortQueueArrayByTimeStamp(XATTuple[] tuples)
	{
		ArrayList tempTuples = new ArrayList();
		
		Arrays.sort(tuples, 
		new Comparator() 
		{

            public int compare(Object a, Object b) 
            {
            	int rc = 0; //return code
                
            	XATTuple tuplea = (XATTuple)a;
            	XATTuple tupleb = (XATTuple)b;
                
            	//check for tupleb, just incase if we are ordering odd number of tuples 
            	if ((tupleb != null) && (tuplea != null))
            	{            		
            		//--- Sort by timestamp, 
            		//  otherwise alphabetical ignoring case.
            		if (tuplea.getCell(4).convertToDouble().longValue() < tupleb.getCell(4).convertToDouble().longValue()) 
            		{
            			rc = -1;
            		} 
            		else if (tuplea.getCell(4).convertToDouble().longValue() > tupleb.getCell(4).convertToDouble().longValue()) 
            		{
            			rc = 1;
            		}
            	}
                return rc;
            }
		}
       );		
	   		
		//now remove duplicates from the array
		int i = 0;
		while (i < tuples.length)
		{
			int k=0;
			//add the object to the array list
			tempTuples.add(tuples[i]);
			
			if ((i + 1) < tuples.length)
			{
				String action1 = tuples[i].getCell(0).toString();
				String action2 = tuples[i+1].getCell(0).toString();
				int objectId1 = tuples[i].getCell(1).convertToDouble().intValue();
				int objectId2 = tuples[i+1].getCell(1).convertToDouble().intValue();
				int reportNum1 = tuples[i].getCell(2).convertToDouble().intValue();
				int reportNum2 = tuples[i+1].getCell(2).convertToDouble().intValue();
				long currentX1 = tuples[i].getCell(5).convertToDouble().intValue();
				long currentX2 = tuples[i+1].getCell(5).convertToDouble().intValue();
				long currentY1 = tuples[i].getCell(6).convertToDouble().intValue();
				long currentY2 = tuples[i+1].getCell(6).convertToDouble().intValue();
			
			while ((objectId1 == objectId2)   &&
				   (reportNum1 == reportNum2) &&
				   (currentX1 == currentX2)   &&
				   (currentY1 == currentY2)   &&   (i+k < tuples.length))
				{				
					k++;
					if ((i + k + 1) < tuples.length)
					{
						if ((tuples[i+k] != null) && (tuples[i+k+1] != null))
						{
							objectId1  = tuples[i+k].getCell(1).convertToDouble().intValue();
							objectId2  = tuples[i+k+1].getCell(1).convertToDouble().intValue();
							reportNum1 = tuples[i+k].getCell(2).convertToDouble().intValue();
							reportNum2 = tuples[i+k+1].getCell(2).convertToDouble().intValue();
							currentX1  = tuples[i+k].getCell(5).convertToDouble().intValue();
							currentX2  = tuples[i+k+1].getCell(5).convertToDouble().intValue();
							currentY1  = tuples[i+k].getCell(6).convertToDouble().intValue();
							currentY2  = tuples[i+k+1].getCell(6).convertToDouble().intValue();
						}
					}					
				}
			}

			i = i + k + 1;			
		}
		
		tuples = (XATTuple[])(tempTuples.toArray(new XATTuple[] {}));		
	   //return sorted duplicate free tuples		
	   return tuples;
	}
	
	*/
	
	private void discreteUpdatePhaseMovingObject(Time currentTime, 
			   									 int delta, 
												 MovingObjectState movingObjectUpdate)
	{
		if (objectsHash.getByFirstKey(movingObjectUpdate.GetID()) == null )
		{
			numberOfSupportedObjects++;	
			MotionStatisticsGatherer.totalNumberObjectsEnteredIntoHashTable++;
		}
		objectsHash.add(movingObjectUpdate.GetID(), (int)movingObjectUpdate.GetPosition().GetT(), movingObjectUpdate);
		MotionStatisticsGatherer.totalNumberObjectsLocationUpdates++;
	}
	
	
	private void discreteUpdatePhaseMovingQuery(Time currentTime, 
				  								int delta, 
												MovingQueryState movingQueryUpdate)
	{
	    if (queriesHash.getByFirstKey(movingQueryUpdate.GetID()) == null )
		{
			numberOfSupportedQueries++;		
			MotionStatisticsGatherer.totalNumberQueriesEnteredIntoHashTable++;
		}
		queriesHash.add(movingQueryUpdate.GetID(), (int)movingQueryUpdate.GetPosition().GetT(), movingQueryUpdate);
		MotionStatisticsGatherer.totalNumberQueriesLocationUpdates++;
	}
	
	private void executeMotionOperator() 
	{		
		XATQueue oQueue = this.getOutputQueue();
		XATQueue linputQueue = this.getLeftInputQueue();
		XATQueue rinputQueue = this.getRightInputQueue();
		XATTuple[] tuplesMovingObjects = null;
		XATTuple[] tuplesMovingQueries = null;		
		
		XATTuple[] tuplesMovingObjectsToBeProcessed = null;
		XATTuple[] tuplesMovingQueriesToBeProcessed = null;	
		int lHowmany; 
		int rHowmany;
		
		this.currentTime = new Time();
		
		currentTime.SetTime(MotionGenericOperator.executionIterationCount * MotionGenericOperator.delta);  
		motionOperatorStatistics.intervalTimeStart = (int)currentTime.GetTime();
		motionOperatorStatistics.intervalTimeEnd = (int)(currentTime.GetTime()+MotionGenericOperator.delta);
		
		Helper.DebugWriteLine("\nCurrent active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");
		
		
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Current time = " + currentTime.GetTime());
			Helper.DebugWriteLine("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");
			Helper.DebugWriteLine("Starting Location Updates Gathering Phase...");
			Helper.DebugWriteLine("Updating locations of moving objects...");
		}

		//System.out.println("Current time = " + currentTime.GetTime());
		//System.out.println("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");

		
		lHowmany = linputQueue.getQueueSize();	
		rHowmany = rinputQueue.getQueueSize();
		
		if (lHowmany != 0)
		{
			//dequeue objects tuples
			tuplesMovingObjects = /*sortQueueArrayByTimeStamp(*/linputQueue.dequeue(lHowmany, this)/*)*/;			
		}
		
		if (rHowmany != 0)
		{
			//dequeue querie tuples
			tuplesMovingQueries = /*sortQueueArrayByTimeStamp(*/rinputQueue.dequeue(rHowmany, this)/*)*/;
		}
		
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("\nSize of left queue = " + lHowmany);
			Helper.DebugWriteLine("Size of right queue = " + rHowmany+"\n");					
		}
		
		int veryFirstObjectUpdateTimestamp = -1;
		int veryFirstQueryUpdateTimestamp = -1;
		int veryFirstMostProgressiveTimestamp = -1;
		
		//adjust the starting timestamp (because the stream generator doesn't start streaming right from the beginning)
		//pick the most "progressive" (i.e., latest timestamp from either of the streams)
		if (isTheFirstIteration == true)
		{
			//statistics header
			Helper.StatsWriteLine(motionOperatorStatistics.ShortDiscreteStatisticsHeaderToString());
			
			
			if (tuplesMovingObjects.length > 0)
			{
				veryFirstObjectUpdateTimestamp = Integer.parseInt(tuplesMovingObjects[0].getCell(4).toString());
			}
			if (tuplesMovingQueries.length > 0)
			{
				veryFirstQueryUpdateTimestamp  = Integer.parseInt(tuplesMovingQueries[0].getCell(4).toString());
			}
			
			if (veryFirstObjectUpdateTimestamp >= veryFirstQueryUpdateTimestamp)
			{
				veryFirstMostProgressiveTimestamp = veryFirstObjectUpdateTimestamp;
			}
			else
			{
				veryFirstMostProgressiveTimestamp = veryFirstQueryUpdateTimestamp;
			}
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("The very first timestamp from objects stream = " + veryFirstObjectUpdateTimestamp);
				Helper.DebugWriteLine("The very first timestamp from queries stream = " + veryFirstQueryUpdateTimestamp);				
			}
			
			//System.out.println("The very first timestamp from objects stream = " + veryFirstObjectUpdateTimestamp);
			//System.out.println("The very first timestamp from queries stream = " + veryFirstQueryUpdateTimestamp);

			MotionGenericOperator.executionIterationCount = (int)(veryFirstMostProgressiveTimestamp/MotionGenericOperator.delta);
			currentTime.SetTime(MotionGenericOperator.executionIterationCount * MotionGenericOperator.delta);  					
			isTheFirstIteration = false;
			motionOperatorStatistics.intervalTimeStart = (int)currentTime.GetTime();
			motionOperatorStatistics.intervalTimeEnd = (int)(currentTime.GetTime()+MotionGenericOperator.delta);
			
						
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nAdjusting Time Interval...");
				Helper.DebugWriteLine("Current execution count = " + MotionGenericOperator.executionIterationCount);
				Helper.DebugWriteLine("Current time = " + currentTime.GetTime());
				Helper.DebugWriteLine("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");
			}
		}
				
		
		//statistics		
		motionOperatorStatistics.timeObjectsUpdateStart = System.currentTimeMillis();			
		
		//*** LOCATION UPDATE PHASE ***//		
		if(lHowmany != 0) 
		{			
			//get only the tuples that belong to the CURRENT time interval, the rest put back to the queue
			ArrayList tuplesToKeep = new ArrayList();
			//ArrayList tuplesToReturn = new ArrayList();
			ArrayList tuplesToDrop = new ArrayList();
			
			//if(DEBUG == 1)
			//{
				Helper.DebugWriteLine("First tuple on the queue (MO): " + tuplesMovingObjects[0].getCell(0)+" | "+ tuplesMovingObjects[0].getCell(1)+" | "+ tuplesMovingObjects[0].getCell(2)+" | "+ tuplesMovingObjects[0].getCell(3)+" | "+ tuplesMovingObjects[0].getCell(4)+" | "+ tuplesMovingObjects[0].getCell(5)+" | "+ tuplesMovingObjects[0].getCell(6)+" | "+ tuplesMovingObjects[0].getCell(7)+" | "+ tuplesMovingObjects[0].getCell(8));
				Helper.DebugWriteLine("Last tuple on the queue (MO): " + tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(0)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(1)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(2)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(3)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(4)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(5)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(6)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(7)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(8));
				Helper.DebugWriteLine("Left queue tuples = " + lHowmany);
			//	Helper.DebugWriteLine(" *** ");
			//}
			
			//System.out.println("First tuple on the queue (MO): " + tuplesMovingObjects[0].getCell(0)+" | "+ tuplesMovingObjects[0].getCell(1)+" | "+ tuplesMovingObjects[0].getCell(2)+" | "+ tuplesMovingObjects[0].getCell(3)+" | "+ tuplesMovingObjects[0].getCell(4)+" | "+ tuplesMovingObjects[0].getCell(5)+" | "+ tuplesMovingObjects[0].getCell(6)+" | "+ tuplesMovingObjects[0].getCell(7)+" | "+ tuplesMovingObjects[0].getCell(8));
			//System.out.println("Last tuple on the queue (MO): " + tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(0)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(1)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(2)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(3)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(4)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(5)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(6)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(7)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(8));

			
			
			for(int i=0; i<tuplesMovingObjects.length; i++)
			{
				//adjust the starting timestamp (because the stream generator doesn't start right from the beginning)
				if (isTheFirstIteration == true)
				{
					MotionGenericOperator.executionIterationCount = (int)(Integer.parseInt(tuplesMovingObjects[i].getCell(4).toString())/MotionGenericOperator.delta);
					currentTime.SetTime(MotionGenericOperator.executionIterationCount * MotionGenericOperator.delta);  					
					isTheFirstIteration = false;
					
					if(DEBUG == 1)
					{
						Helper.DebugWriteLine("\nAdjusting Time Interval...");
						Helper.DebugWriteLine("Current execution count = " + MotionGenericOperator.executionIterationCount);
						Helper.DebugWriteLine("Current time = " + currentTime.GetTime());
						Helper.DebugWriteLine("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");
					}
				}
				

				if ((Long.parseLong(tuplesMovingObjects[i].getCell(4).toString()) >= currentTime.GetTime()) &&
					(Long.parseLong(tuplesMovingObjects[i].getCell(4).toString()) < (currentTime.GetTime() + MotionGenericOperator.delta)))
				{
					//keep the tuple
					tuplesToKeep.add(tuplesMovingObjects[i]);					
				}
				else if (Long.parseLong(tuplesMovingObjects[i].getCell(4).toString()) < currentTime.GetTime())
				{
					//drop the tuples (they arrived too late)
					tuplesToDrop.add(tuplesMovingObjects[i]);					
				}
				else
				{
					futureTuplesMovingObjects.add(tuplesMovingObjects[i]);
				}
			}
			
//			Helper.DebugWriteLine(futureTuplesMovingObjects.toString());
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nMoving Objects:");
				Helper.DebugWriteLine("---------------------------");
				Helper.DebugWriteLine("# of tuples to keep = ["+ tuplesToKeep.size()+"]");
				Helper.DebugWriteLine("# of tuples to (return) - future tuples = ["+ futureTuplesMovingObjects.size()+"]");			
				Helper.DebugWriteLine("# of tuples to drop = ["+ tuplesToDrop.size()+"]");
			}

			motionOperatorStatistics.tuplesObjectsDequeued = lHowmany;
			motionOperatorStatistics.tuplesObjectsKept = tuplesToKeep.size();
			motionOperatorStatistics.tuplesObjectsReturned = futureTuplesMovingObjects.size();
			motionOperatorStatistics.tuplesObjectsDropped = tuplesToDrop.size();
			
			//drop the tuples
			tuplesToDrop = null;	
			
			
			if (futureTuplesMovingObjects.size() > 0)
			{
				XATTuple xt = (XATTuple)futureTuplesMovingObjects.get(futureTuplesMovingObjects.size() - 1);
				float lastTimestamp = Float.parseFloat(xt.getCell(4).toString());
				if (lastTimestamp >= (currentTime.GetTime() + MotionGenericOperator.delta))
				//if the last timestamp in the future tuples arraylist is greater 
				//than the current delta time interval, push the execution
				//set the flag to increment evaluation count (need to move on to the next time interval
				pushTimeIntervalForwardMovingObjects = true;
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("\n!!! SETTING PUSH FORWARD FOR MOVING OBJECTS TO TRUE !!!\n");
				}
			}
			
			
			if(tuplesToKeep.size() > 0)
			{					
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("# of moving objects' updates = ["+ tuplesToKeep.size() +"]");
				}
				
				for (int i=0; i < tuplesToKeep.size(); i++)
				{							
					XATTuple tupleMovingObjectsToBeProcessed = (XATTuple)tuplesToKeep.get(i);
					MovingObjectState movingObjectUpdate = new MovingObjectState(
							Integer.parseInt(tupleMovingObjectsToBeProcessed.getCell(1).toString()),
							EntityType.Object,
							Long.parseLong(tupleMovingObjectsToBeProcessed.getCell(5).toString()),
							Long.parseLong(tupleMovingObjectsToBeProcessed.getCell(6).toString()),
							Integer.parseInt(tupleMovingObjectsToBeProcessed.getCell(4).toString()),
							UpdateType.FromStringToEnum(tupleMovingObjectsToBeProcessed.getCell(0).toString()),
							ClassType.FromIntToEnum(Integer.parseInt(tupleMovingObjectsToBeProcessed.getCell(3).toString())));
					
					
						discreteUpdatePhaseMovingObject(currentTime, 
													MotionGenericOperator.delta, 
													movingObjectUpdate);
					
						//flush the object immediately to the grid
						flushObjectToGrid(movingObjectUpdate);
				}
				tuplesToKeep = null;
			}
			
					
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Location update phase for moving objects is done.");			
			}
		}
		
		//statistics
		motionOperatorStatistics.timeObjectsUpdateEnd = System.currentTimeMillis();
		MotionStatisticsGatherer.totalTimeElapsedObjectsUpdate += (motionOperatorStatistics.timeObjectsUpdateEnd - motionOperatorStatistics.timeObjectsUpdateStart);
		
		motionOperatorStatistics.timeQueriesUpdateStart = System.currentTimeMillis();
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Updating locations of moving queries...");
		}
			
		//moving queries		
		if(rHowmany != 0) 
		{			
			//get only the tuples that belong to the CURRENT time interval, the rest put back to the queue
			ArrayList tuplesToKeep = new ArrayList();
			//ArrayList tuplesToReturn = new ArrayList();
			ArrayList tuplesToDrop = new ArrayList();
			
			//if(DEBUG == 1)
			//{
			//	Helper.DebugWriteLine("Moving Query Timestamp = " + Long.parseLong(tuplesMovingQueries[0].getCell(4).toString()));				
				Helper.DebugWriteLine("First tuple on the queue (MQ): " + tuplesMovingQueries[0].getCell(0)+" | "+ tuplesMovingQueries[0].getCell(1)+" | "+ tuplesMovingQueries[0].getCell(2)+" | "+ tuplesMovingQueries[0].getCell(3)+" | "+ tuplesMovingQueries[0].getCell(4)+" | "+ tuplesMovingQueries[0].getCell(5)+" | "+ tuplesMovingQueries[0].getCell(6)+" | "+ tuplesMovingQueries[0].getCell(7)+" | "+ tuplesMovingQueries[0].getCell(8));
				Helper.DebugWriteLine("Last tuple on the queue (MQ): " + tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(0)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(1)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(2)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(3)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(4)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(5)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(6)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(7)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(8));
				Helper.DebugWriteLine("Right queue tuples = " + rHowmany);
			//	Helper.DebugWriteLine(" *** ");
			//}
			
			//System.out.println("First tuple on the queue (MQ): " + tuplesMovingQueries[0].getCell(0)+" | "+ tuplesMovingQueries[0].getCell(1)+" | "+ tuplesMovingQueries[0].getCell(2)+" | "+ tuplesMovingQueries[0].getCell(3)+" | "+ tuplesMovingQueries[0].getCell(4)+" | "+ tuplesMovingQueries[0].getCell(5)+" | "+ tuplesMovingQueries[0].getCell(6)+" | "+ tuplesMovingQueries[0].getCell(7)+" | "+ tuplesMovingQueries[0].getCell(8));
			//System.out.println("Last tuple on the queue (MQ): " + tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(0)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(1)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(2)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(3)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(4)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(5)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(6)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(7)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(8));

			
			for(int i=0; i<tuplesMovingQueries.length; i++)
			{
				if ((Long.parseLong(tuplesMovingQueries[i].getCell(4).toString()) >= currentTime.GetTime()) &&
				    (Long.parseLong(tuplesMovingQueries[i].getCell(4).toString()) < currentTime.GetTime() + MotionGenericOperator.delta))
				{
					//keep the tuple
					tuplesToKeep.add(tuplesMovingQueries[i]);					
				}
				else if (Long.parseLong(tuplesMovingQueries[i].getCell(4).toString()) < currentTime.GetTime())
				{
					//drop the tuples (they arrived too late)
					tuplesToDrop.add(tuplesMovingQueries[i]);					
				}			
				else
				{
					futureTuplesMovingQueries.add(tuplesMovingQueries[i]);
				}				
			}
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nMoving Queries:");
				Helper.DebugWriteLine("---------------------------");
				Helper.DebugWriteLine("# of tuples to keep = ["+ tuplesToKeep.size()+"]");
				Helper.DebugWriteLine("# of tuples to (return) - future tuples = ["+ futureTuplesMovingQueries.size()+"]");
				Helper.DebugWriteLine("# of tuples to drop = ["+ tuplesToDrop.size()+"]");
			}
			
			//statistics
			motionOperatorStatistics.tuplesQueriesDequeued = rHowmany;
			motionOperatorStatistics.tuplesQueriesKept = tuplesToKeep.size();
			motionOperatorStatistics.tuplesQueriesReturned = futureTuplesMovingQueries.size();
			motionOperatorStatistics.tuplesQueriesDropped = tuplesToDrop.size();
			
			//drop the tuples
			tuplesToDrop = null;
			
			
			
			if (futureTuplesMovingQueries.size() > 0)
			{
				XATTuple xt = (XATTuple)futureTuplesMovingQueries.get(futureTuplesMovingQueries.size() - 1);
				float lastTimestamp = Float.parseFloat(xt.getCell(4).toString());
				if (lastTimestamp >= (currentTime.GetTime() + MotionGenericOperator.delta))				
				{
					//set the flag to increment evaluation count (need to move on to the next time interval
					pushTimeIntervalForwardMovingQueries = true;
					if(DEBUG == 1)
					{
						Helper.DebugWriteLine("\n!!! SETTING PUSH FORWARD FOR MOVING QUERIES TO TRUE !!!\n");
					}
				}
			}
						
			
			//put tuples back on the queue
			//rinputQueue.enqueue(tuplesToReturn, this);
			
			
			if(tuplesToKeep.size() > 0)
			{					
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("# of moving queries' updates = ["+ tuplesToKeep.size() +"]");
				}
				
				for (int i=0; i < tuplesToKeep.size(); i++)
				{
					XATTuple tupleMovingQueryToBeProcessed = (XATTuple)tuplesToKeep.get(i);
					MovingQueryState movingQueryUpdate = new MovingQueryState(
							Integer.parseInt(tupleMovingQueryToBeProcessed.getCell(1).toString()),
							EntityType.Query,
							Long.parseLong(tupleMovingQueryToBeProcessed.getCell(5).toString()),
							Long.parseLong(tupleMovingQueryToBeProcessed.getCell(6).toString()),
							Integer.parseInt(tupleMovingQueryToBeProcessed.getCell(4).toString()),
							UpdateType.FromStringToEnum(tupleMovingQueryToBeProcessed.getCell(0).toString()),
							ClassType.FromIntToEnum(Integer.parseInt(tupleMovingQueryToBeProcessed.getCell(3).toString())),
							QueryType.Range,
							Integer.parseInt(tupleMovingQueryToBeProcessed.getCell(10).toString()));					

					
						discreteUpdatePhaseMovingQuery(currentTime, 
												   MotionGenericOperator.delta, 
												   movingQueryUpdate);
						
						//flush the query immediately to the grid
						flushQueryToGrid(movingQueryUpdate);
					
				}
				
				tuplesToKeep = null;
			}	
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Location update for moving queries is done.");
			}	
			
		}
		
		motionOperatorStatistics.timeQueriesUpdateEnd = System.currentTimeMillis();
		//*** LOCATION UPDATE PHASE ***		
		MotionStatisticsGatherer.totalTimeElapsedQueriesUpdate += (motionOperatorStatistics.timeQueriesUpdateEnd - motionOperatorStatistics.timeQueriesUpdateStart);
		
		MotionDiscreteResultsCollection miniResults = new MotionDiscreteResultsCollection();
		
		//*** JOINING PHASE ***		

		if ((pushTimeIntervalForwardMovingObjects == true) && (pushTimeIntervalForwardMovingQueries == true))
		{				
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nProcessing moving objects from the 'FAKE QUEUE'...");
				Helper.DebugWriteLine("Total entries on the objects 'FAKE QUEUE' = " + futureTuplesMovingObjects.size());
			}	
				
			motionOperatorStatistics.timeFakeQObjectsStart = System.currentTimeMillis();
			
			if (futureTuplesMovingObjects.size() > 0)
			{
				int counter = 0;
				XATTuple tupleToBeProcessed = (XATTuple)futureTuplesMovingObjects.get(0);
				while (Float.parseFloat(tupleToBeProcessed.getCell(4).toString()) < (currentTime.GetTime() + MotionGenericOperator.delta))
				{	
					counter++;
					
					MovingObjectState movingObjectUpdate = new MovingObjectState(
								Integer.parseInt(tupleToBeProcessed.getCell(1).toString()),
									EntityType.Object,
									Long.parseLong(tupleToBeProcessed.getCell(5).toString()),
									Long.parseLong(tupleToBeProcessed.getCell(6).toString()),
									Integer.parseInt(tupleToBeProcessed.getCell(4).toString()),
									UpdateType.FromStringToEnum(tupleToBeProcessed.getCell(0).toString()),
									ClassType.FromIntToEnum(Integer.parseInt(tupleToBeProcessed.getCell(3).toString())));

		
					discreteUpdatePhaseMovingObject(currentTime, 
													MotionGenericOperator.delta, 
													movingObjectUpdate);
					
					//flush the object immediately to the grid
					flushObjectToGrid(movingObjectUpdate);
					
					futureTuplesMovingObjects.remove(0);
			
					if (futureTuplesMovingObjects.size() == 0)
					{
						//fake queue is empty
						break;
					}
					tupleToBeProcessed = (XATTuple)futureTuplesMovingObjects.get(0);												
				}//end while
				
				motionOperatorStatistics.timeFakeQObjectsEnd = System.currentTimeMillis();
				
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Dequeued total "+counter+" objects updates from the 'FAKE QUEUE'...");
					Helper.DebugWriteLine("Now total entries on the objects 'FAKE QUEUE' = " + futureTuplesMovingObjects.size());
				}
			}//end if
			
			
			if(DEBUG == 1)
			{				
				Helper.DebugWriteLine("Done discrete updates for tuples from the 'fake' queue");				
				Helper.DebugWriteLine("Done removing tuples from the 'fake' queue");
				Helper.DebugWriteLine("\nProcessing moving queries from the 'fake queue'...");
			}	
			
			motionOperatorStatistics.timeFakeQQueriesStart = System.currentTimeMillis();
			if (futureTuplesMovingQueries.size() > 0)
			{
				int counter = 0;
				XATTuple tupleToBeProcessed = (XATTuple)futureTuplesMovingQueries.get(0);
				while (Float.parseFloat(tupleToBeProcessed.getCell(4).toString()) < (currentTime.GetTime() + MotionGenericOperator.delta))
				{															
					counter++;
					MovingQueryState movingQueryUpdate = new MovingQueryState(
						Integer.parseInt(tupleToBeProcessed.getCell(1).toString()),
						EntityType.Query,
						Long.parseLong(tupleToBeProcessed.getCell(5).toString()),
						Long.parseLong(tupleToBeProcessed.getCell(6).toString()),
						Integer.parseInt(tupleToBeProcessed.getCell(4).toString()),
						UpdateType.FromStringToEnum(tupleToBeProcessed.getCell(0).toString()),
						ClassType.FromIntToEnum(Integer.parseInt(tupleToBeProcessed.getCell(3).toString())),
						QueryType.Range,
						Integer.parseInt(tupleToBeProcessed.getCell(10).toString()));		
										
						discreteUpdatePhaseMovingQuery(currentTime, 
							MotionGenericOperator.delta, 
							movingQueryUpdate);		
						
						//flush the query immediately to the grid
						flushQueryToGrid(movingQueryUpdate);
													
						futureTuplesMovingQueries.remove(0);
						if (futureTuplesMovingQueries.size() == 0)
						{
							//fake queue is empty
							break;
						}
					tupleToBeProcessed = (XATTuple)futureTuplesMovingQueries.get(0);												
				}//end while
				
				motionOperatorStatistics.timeFakeQQueriesEnd = System.currentTimeMillis();
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Dequeued total "+counter+" queries updates from the 'FAKE QUEUE'...");	
					Helper.DebugWriteLine("Now total entries on the objects 'FAKE QUEUE' = " + futureTuplesMovingObjects.size());
				}
			}//end if
							
		
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nStarting Joining Phase...\n");
				Helper.DebugWriteLine(grid.statisticsToString());
			}	
		
			//System.out.println("STARTING JOIN...");
			//statistics
			motionOperatorStatistics.timeJoinStart = System.currentTimeMillis();
			
			miniResults = joiningPhase(currentTime, MotionGenericOperator.delta);	
			
			//statistics
			motionOperatorStatistics.timeJoinEnd = System.currentTimeMillis();
			//System.out.println("DONE WITH JOIN...");
			motionOperatorStatistics.numberOfIntersectionResults = miniResults.GetDiscreteResults().size();
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Joining Phase is done");
				Helper.DebugWriteLine("# of intersection results = ["+miniResults.GetDiscreteResults().size()+"]");				
			}
			
			//*** JOINING PHASE ***		
			
			motionOperatorStatistics.timeOverallEnd = System.currentTimeMillis();

			MotionStatisticsGatherer.totalTimeElapsed += (motionOperatorStatistics.timeOverallEnd - motionOperatorStatistics.timeOverallStart);
			
			
			/*
			if (miniResults.GetDiscreteResults().size() > 0)
			{		
				//System.out.println("Sending results...");
				int numberOfResults = miniResults.GetDiscreteResults().size();
				//tuples to be sent to the output queue
				XATTupleImp[] outputTuples = new XATTupleImp[numberOfResults];
				//value of one tuple
				XATDataValueImp[] outputValue = new XATDataValueImp[1];
				//array of results
			
				for(int i = 0; i < numberOfResults ; i++)
				{		
					MotionDiscreteResult outputResult = (MotionDiscreteResult)miniResults.GetDiscreteResults().get(i);
					outputValue[0] = new XATDataValueImp(outputResult);
					outputTuples[i] = new XATTupleImp(outputValue);
				}
			
				//send the results to the output queue
				this.getOutputQueue().enqueue(outputTuples, this);
							
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Clearing the results");					
				}
				//System.out.println("Done sending results...");								
			}
			*/
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Clearing the grid");					
			}
			
			//grid cleanup
			grid.Cleanup();
			
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Clearing Objects and Queries tables");					
			}
			
			//cleanup the Objects and Queries hash tables
			cleanupObjectsAndQueriesTables();
			
			if(DEBUG == 1)
			{
				//print grid statistics		
				Helper.DebugWriteLine(grid.toString());
			}
			

			motionOperatorStatistics.numberOfObjectsInTheSystem = this.objectsHash.getNumberOfFirstKeyEntries();
			motionOperatorStatistics.numberOfQueriesInTheSystem = this.queriesHash.getNumberOfFirstKeyEntries();
		
		}//end pushTimeIntervalForward

		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("# of objects in the Objects table: "+this.objectsHash.getNumberOfFirstKeyEntries());
			Helper.DebugWriteLine("# of queries in the Queries table: "+this.queriesHash.getNumberOfFirstKeyEntries());
		}
		
		//statistics (ended actual execution)
		motionOperatorStatistics.timeOverallEnd = System.currentTimeMillis();
		
		//System.out.println("DONE WITH CURRENT INTERVAL EXECUTION...");
		
		//print the results (for accounting purposes)
		//flag to write out results is turned on
		if ((RESULTS_FLAG == 1) || (RESULTS_FLAG == 2) || (RESULTS_FLAG == 4)) 
		{						
			if (miniResults.GetDiscreteResults().size() > 0)
			{
				ResultsRecorder.DiscreteResultsWriteLine(miniResults);
			}
		}

		motionOperatorStatistics.timeOverallStart = System.currentTimeMillis();		
		
		//System.out.println("Done writing results to a file...");
		
		//clear results collection
		miniResults.Clear();			

		
		//call for garbage collection
		//System.gc();	
	 }
	
	
	private void cleanupObjectsAndQueriesTables()
	{
		//Objects hash table
		objectsHash.InternalHash.clear();
		
		
		//Queiries hash table
		queriesHash.InternalHash.clear();
	}
	
		
	// Joins moving objects with moving queries. To be more specific, we go cell
	// by cell (in the grid) and join (i.e., calculate intersections) between
	// object trjectories (lines) and query trajectory volumes (cubes) and return
	// an array of mini-clipped results. (Mini-clipped results = line segments
	// at the granularity of grid cells that need to be merged to get the full
	// query answers.)
	// This is the third phase in our algorithm.
	//
	// @param       T    			MotionMovingEntityTimestamp (an operator-specific
	//								timestamp that starts counting off from the time
	//								the system started executing - 0..1..2..3..4, etc.)
	// @param       delta    		the length of the time interval 
	// @return      none
	// PHASE III - joining (i.e., calculating intersections between) moving objects and
	// moving queries.
	private MotionDiscreteResultsCollection joiningPhase(Time T, int delta)
	{
		int numJoinComparisons = 0;
		MotionDiscreteResultsCollection miniResults = new MotionDiscreteResultsCollection();
			
		//Helper.DebugWriteLine(grid.statisticsToString());
		
		//clear the mini-clipped results ArrayList
		miniResults.Clear();
				
		//Loop through grid cells in the x-dimension
		for (int i=0; i < GenericSpatioTemporalGrid.GetXDimensionCellCount(); i++)
		{
			//Loop through grid cells in the y-dimension
 			for (int j=0; j < GenericSpatioTemporalGrid.GetYDimensionCellCount(); j++)
			{
				//Loop through grid cells in the t-dimension
				for (int k=0; k < GenericSpatioTemporalGrid.GetTDimensionCellCount(); k++)
				{
					if (grid.directory[i][j][k] != null)
					{					
						if (grid.directory[i][j][k].GetMovingQueries().size() > 0)
						{
							//For each moving query MQ that has an entry in the grid cell G i,j,k
							for (int q=0; q < grid.directory[i][j][k].GetMovingQueries().size(); q++)
							{		
								//For each moving object MO that has an entry in the grid cell Gi,j,k  
								for (int o=0; o < grid.directory[i][j][k].GetMovingObjects().size(); o++)
								{									
									GridEntry queryGridEntry = (GridEntry)grid.directory[i][j][k].GetMovingQueries().get(q);
									GridEntry objectGridEntry = (GridEntry)grid.directory[i][j][k].GetMovingObjects().get(o);
																
									MovingQueryState mQ  = (MovingQueryState)queriesHash.get(queryGridEntry.GetID(), (int)queryGridEntry.GetTime());
									MovingObjectState mO = (MovingObjectState)objectsHash.get(objectGridEntry.GetID(), (int)objectGridEntry.GetTime());													
			
									if (mQ == null)
									{
										if(DEBUG == 1)
										{
											Helper.DebugWriteLine("NULL moving query: ID = "+queryGridEntry.GetID() + " timestamp = " + (int)queryGridEntry.GetTime());		
										}
										//let's try again (debug)
										mQ  = (MovingQueryState)queriesHash.get(queryGridEntry.GetID(), (int)queryGridEntry.GetTime());									
									}
								
									if (mO == null)
									{
										if(DEBUG == 1)
										{
											Helper.DebugWriteLine("NULL moving object: ID = "+objectGridEntry.GetID() + " timestamp = " + (int)objectGridEntry.GetTime());		
										}
										mO = (MovingObjectState)objectsHash.get(objectGridEntry.GetID(), (int)objectGridEntry.GetTime());																										
									}
								
									MotionDiscreteResult joinResult = null;
								
									//only join them if their timestamps match
									if (mQ.GetPosition().GetT() == mO.GetPosition().GetT())
									{								
										//do the actual join
										joinResult = grid.DoJoin(mQ, mO, new GridCellIndex3D(i,j,k));
										numJoinComparisons++;
									}
									
								
									if (joinResult != null)
									{
										if (RESULTS_FLAG == 2) 
										{						
											if (mQ.GetID() != mO.GetID()) //store only the results where object id != query id
											{
												miniResults.Add(joinResult);
											}
										}
										else if (RESULTS_FLAG == 3)
										{
											//write out the result and discard it
											ResultsRecorder.DiscreteSingleResultsWriteLine(joinResult);
											joinResult = null;
										}
										else if (RESULTS_FLAG == 4)
										{
											//store only a subset of results
											if (((mQ.GetID() % 10) == 0) && ((mO.GetID() % 10) == 0))
											{
												miniResults.Add(joinResult);
											}
										}
										else
										{									
											miniResults.Add(joinResult);
										}
									}
								}
							}
						}
						
						grid.directory[i][j][k] = null;
					}
				}
			}
		}	
		
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Total # of join comparisons is "+ numJoinComparisons);		
		}
		return miniResults;
	}
	
	
	//Flushes an object to the grid
	//
	// @param       none
	// @return      none
	private void flushObjectToGrid(MovingObjectState movingObject)
	{		
		if (movingObject != null)
		{
			gridInsertObject(movingObject);		
		    this.motionOperatorStatistics.numberOfObjectsBeingJoined++;
		}
	}//end flushObjectsDiscreteUpdatesToGrid

	
	// Flushes a query to the grid
	//
	// @param       none
	// @return      none
	private void flushQueryToGrid(MovingQueryState movingQuery)
	{
		if (movingQuery != null)
		{
			gridInsertQuery(movingQuery);				
			this.motionOperatorStatistics.numberOfQueriesBeingJoined++;
		}
	}

	private void gridInsertObject(MovingObjectState mo)
	{
		boolean deleteFromSystem = false;
		grid.InsertObject(mo, deleteFromSystem, this.currentTime, MotionGenericOperator.delta);
							
		if (deleteFromSystem == true)
		{
			//delete object
			DeleteObjectFromSystemByID(mo.GetID());	
			MotionGenericOperator.numberOfSupportedObjects--;
			MotionStatisticsGatherer.totalNumberObjectsRemovedFromHashTable++;
		}	
	}
	
	
	// Inserts a trajectory of a moving entity into a grid 
	// index.
	//
	// @param       
	// @return      none
	/*dp:
	private void gridInsertObject(MovingObjectState mo)
	{
		boolean deleteFromSystem = false;
		grid.InsertObject(mo, deleteFromSystem, this.currentTime, MotionGenericOperator.delta);
							
		if (deleteFromSystem == true)
		{
			//delete object
			DeleteObjectFromSystemByID(mo.GetID());		
		}		
	}
	*/
	

	
	
	private void gridInsertQuery(MovingQueryState mq)
	{
		boolean deleteFromSystem = false;
		grid.InsertQuery(mq, deleteFromSystem, this.currentTime, MotionGenericOperator.delta);
							
		if (deleteFromSystem == true)
		{
			//delete object
			DeleteQueryFromSystemByID(mq.GetID());	
			MotionGenericOperator.numberOfSupportedQueries--;
			MotionStatisticsGatherer.totalNumberQueriesRemovedFromHashTable++;
		}		
	}
	
	private void DeleteObjectFromSystemByID(int id)
	{	
		objectsHash.removeByFirstKey(id);		
	}
	
	private void DeleteQueryFromSystemByID(int id)
	{	
		queriesHash.removeByFirstKey(id);
	}	
	
}
