/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.rangeoperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.hash.DoubleKeyHash;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
 * @author Rimma V. Nehme
 *
 */
public abstract class MotionRangeQueryOperator extends MotionGenericOperator
{
	//protected MovingObjectsStatesCollection objectsTable;
	//protected MovingQueriesStatesCollection queriesTable;
	
	//protected MemoryEfficientHashtable objectsHash;
	//protected MemoryEfficientHashtable queriesHash;
	protected DoubleKeyHash objectsHash;
	protected DoubleKeyHash queriesHash;
	
	
	
	
	Time currentTime;
	
	public MotionRangeQueryOperator(XATNode node) 
	{
		super(node);	
		//objectsTable = new MovingObjectsStatesCollection();
		//queriesTable = new MovingQueriesStatesCollection();
		
		objectsHash = new DoubleKeyHash();
		queriesHash = new DoubleKeyHash();
		
		MotionGenericOperator.executionIterationCount = 1;
		MotionGenericOperator.delta = 4;			
	}
	
	public void initState(String statetype) 
	{		
		//String statetypeMovingObjects = "edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections.MovingObjectsStatesCollection";
		//String statetypeMovingQueries = "edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections.MovingQueriesStatesCollection";
		
		String statetypeMovingObjects = "edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.hash.DoubleKeyHash";
		String statetypeMovingQueries = "edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.hash.DoubleKeyHash";
		
		try
		{				
			states = new OperatorStateImp[2];
			//states[0] = (MovingObjectsStatesCollection)Class.forName(statetypeMovingObjects).newInstance();
			//states[1] = (MovingQueriesStatesCollection)Class.forName(statetypeMovingQueries).newInstance();
			
			states[0] = (DoubleKeyHash)Class.forName(statetypeMovingObjects).newInstance();
			states[1] = (DoubleKeyHash)Class.forName(statetypeMovingQueries).newInstance();
			
			
			states[0].setOperator(this);
			states[1].setOperator(this);	
			
			//this.objectsTable = (MovingObjectsStatesCollection)this.states[0];		
			//this.queriesTable = (MovingObjectsStatesCollection)this.states[1];
			
			this.objectsHash = (DoubleKeyHash)this.states[0];		
			this.queriesHash = (DoubleKeyHash)this.states[1];
			
			currentTime = new Time();
		}
		catch(Exception e) 
		{
			e.printStackTrace();	
		}		
	}
	
	public OperatorState getLeftState() 
	{
		return states[0];
	}
	
	public int getLeftStateSize() 
	{
		return states[0].getStateSize();
	}

	public OperatorState getRightState() 
	{
		return states[1];
	}
	
	public int getRightStateSize() 
	{
		return states[1].getStateSize();
	}

	public int getStateSize() 
	{
		return (states[0].getStateSize() + states[1].getStateSize());
	}
	
	/*
	 * recompute the unmatched states inside this operator.
	 * Usually the unmatched states are computed based on the matched states inside one of the 
	 * children operator of this operator.
	 * 
	 * return: 
	 * TRUE -- if the states (inside child operator) that the recomputation of the unmatched states are based on are 
	 *      all matched, and the recomputation is successful.
	 * FALSE -- if any of the states (inside child operator) that are used to recompute the unmatched states
	 *          is unmatched, or the recomputation is unsucessful.  
	 */
	public boolean recomputeUnmatchedState() 
	{
		boolean returnvalue = true;
		
		return returnvalue;	
	}

	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() 
	{
		if (states == null) 
		{
			return false;
		}

		for (int i = 0; i < states.length; i++) 
		{
			if (!states[i].isMatched())
				return true;
		}
		return false;
	}


}
