/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Jun 12, 2005
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;

/**
 * @author Rimma V. Nehme
 *
 */
public class DataPoint 
{
    private double mX,mY,mT;
    private String mObjName;
    private int mObjNumericId;
    private SCUBAKMeansCluster mCluster;
    private double mEuDt;
    public long mNextNodeX;  //destination node: x-position
    public long mNextNodeY;  //destination node: y-position
    public int mSpeed;
    public int mBoundary; //for queries only

    public DataPoint(double x, double y, double t, String name, int numericID, long nextNodeX, long nextNodeY, int Speed, int Boundary) 
    {
        this.mX = x;
        this.mY = y;
        this.mT = t;
        this.mObjName = name;
        this.mObjNumericId = numericID;
        mNextNodeX = nextNodeX;
        mNextNodeY = nextNodeY;
        mSpeed = Speed;
        this.mCluster = null;
        mBoundary = Boundary;
    }

    public void setCluster(SCUBAKMeansCluster cluster) 
    {
        this.mCluster = cluster;
        calcEuclideanDistance();
    }

    public void calcEuclideanDistance() 
    {     
    	//called when DP is added to a cluster or when a Centroid is recalculated.
        mEuDt = Math.sqrt(Math.pow((mX - mCluster.getCentroid().getCx()), 2) + Math.pow((mY - mCluster.getCentroid().getCy()), 2));
    }

    public double testEuclideanDistance(Centroid c) 
    {
        return Math.sqrt(Math.pow((mX - c.getCx()), 2) + Math.pow((mY - c.getCy()), 2));
    }

    public double getX() 
    {
        return mX;
    }

    public double getY() 
    {
        return mY;
    }
    
    public double getT() 
    {
        return mT;
    }

    public SCUBAKMeansCluster getCluster() 
    {
        return mCluster;
    }

    public double getCurrentEuDt() 
    {
        return mEuDt;
    }

    public String getObjName() 
    {
        return mObjName;
    }
    
    public int getObjNumericId() 
    {
        return this.mObjNumericId;
    }
    
    public long getNextNodeX() 
    {
        return mNextNodeX;
    }
    
    public long getNextNodeY() 
    {
        return mNextNodeY;
    }
    
    public int getSpeed() 
    {
    	return mSpeed;
    }
    
    public int getBoundary() 
    {
    	return mBoundary;
    }
    
    public PositionSpt2D get3DPosition()
    {
    	return new PositionSpt2D((long)this.getX(),(long)this.getY(),(float)this.getT());
    }
}
