/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on May 16, 2005
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GenericSpatioTemporalGridCell;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridEntry;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingObjectState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.range.MovingRangeQueryState;

/**
 * @author Rimma V. Nehme
 *
 */
public class GenericSpatioTemporalSCUBAGridCell extends GenericSpatioTemporalGridCell
{
	HashMap movingClusters; //MovingClusterID's

	public GenericSpatioTemporalSCUBAGridCell(int newXIndex, 
											  int newYIndex, 
											  int newTIndex,
											  long newXMin, long newYMin, long newTMin,
											  long newXMax, long newYMax, long newTMax)
	{
		super(newXIndex, newYIndex, newTIndex, newXMin, newYMin, newTMin, newXMax, newYMax, newTMax);
		movingClusters = new HashMap();
	}

	public void AddMovingClusterEntry(SCUBAGridEntry movingClusterGridEntry)
	{
		Integer key = new Integer(movingClusterGridEntry.GetID());
		if (movingClusters.get(key) == null)
		{
			movingClusters.put(new Integer(movingClusterGridEntry.GetID()),movingClusterGridEntry);
		}
	}
	
	public void RemoveMovingClusterEntry(SCUBAGridEntry movingClusterGridEntry)
	{
		movingClusters.remove(new Integer(movingClusterGridEntry.GetID()));	
	}
	
	public HashMap GetClusters()
	{
		return this.movingClusters;
	}
	
	public String  GetClustersStatistics(MotionSCUBAClusterStorage clusterStorage)
	{
		String outputString = "";
		Object[] arrClusters = (Object[])movingClusters.entrySet().toArray();							
				
		for (int i=0; i < arrClusters.length; i++)
		{															
			Map.Entry  entry =  (Map.Entry)arrClusters[i];		
			
			SCUBAGridEntry clusterGridEntry = (SCUBAGridEntry)entry.getValue();
			MotionSCUBACluster mCluster = clusterStorage.getClusterByCID(new Integer(clusterGridEntry.GetID()));
			outputString += "cluster id = " + mCluster.cid +", # members = "+mCluster.memberCount  +", grid index = ("+mCluster.currentGridIndexX+", "+mCluster.currentGridIndexY+", "+mCluster.currentGridIndexT+"), centroid position = ("+mCluster.centerPoint.GetX()+", "+mCluster.centerPoint.GetY()+", "+mCluster.centerPoint.GetT()+"), radius = "+mCluster.radius+", speed = "+mCluster.speed+", objects = ("+mCluster.getStringObjectIds()+"), queries = ("+mCluster.getStringQueriesIds()+"), dest_x = "+mCluster.nextNodeX+", dest_y = "+mCluster.nextNodeY+", time = "+mCluster.time+"\n";
		}	
		
		return outputString;
	}
}
