/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on May 14, 2005
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba;

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.ClassType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.DimensionType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.UpdateType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridCellIndex3D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.clusteroperator.Circle;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.clusteroperator.Points;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingObjectState;

/**
 * @author Rimma V. Nehme
 *
 */
public class MotionSCUBACluster 
{
	static private int nextClusterID = 0;
	
	
	public int cid; 		//cluster id
    public Vector objects;  //ids of objects that belong to the cluster
    public Vector queries;  //ids of queries that belong to the cluster
    public long nextNodeX;  //destination node: x-position
    public long nextNodeY;  //destination node: y-position
	public PositionSpt2D centerPoint; //spatio-temporal position of the focal point of the cluster      
    public int memberCount; //total number of members (objects + queries)
    public int speed;
    public int time;        //latest time update
    public float radius;
    public PositionSpt2D furthestPosition;
    public Vector gridCellsOverlap; //contains indicies that the cluster overlaps with
    
    //grid index of the center of the cluster
    public int currentGridIndexX;
    public int currentGridIndexY;
    public int currentGridIndexT;
    

    public MotionSCUBACluster()
    {
    	MotionSCUBANucleus nucleus = new MotionSCUBANucleus();
    	gridCellsOverlap = new Vector();    	
    	cid = nextClusterID;
    	objects = new Vector();
    	queries = new Vector();
    	nextNodeX = 0;
    	nextNodeY = 0;
    	centerPoint = new PositionSpt2D();
    	furthestPosition = new PositionSpt2D(); 
    	memberCount = 0;
    	speed = 0;
    	radius = 0;
    	
    	//default indicies for cluster upon creation
        currentGridIndexX = -1;
        currentGridIndexY = -1;
        currentGridIndexT = -1;
    	
    	nextClusterID++;
    }
    
    public String toString()
	{
		return "MotionSCUBACluster():\n"+
					"cid: "+ this.cid+"\n"+
					"objects: ["+this.objects.toString()+"]\n"+
					"queries: ["+this.queries.toString()+"]\n"+
					"nextNodeX: "+this.nextNodeX+"\n"+
					"nextNodeY: "+this.nextNodeY+"\n"+
					"centerPoint: "+this.centerPoint.toString()+"\n"+
					"radius: "+this.radius+"\n"+
					"memberCount: "+this.memberCount+"\n"+
					"speed: "+this.speed+"\n"+
					"time: "+this.time+"\n"+
					"currentGridIndexX: "+this.currentGridIndexX+"\n"+
					"currentGridIndexY: "+this.currentGridIndexY+"\n"+
					"currentGridIndexT: "+this.currentGridIndexT+"\n"+
					"grid indicies: " + gridCellsOverlap.toString()+"\n";
	}
    
    
    public boolean IsClusterCandidate(MotionSCUBAMovingObjectState mo)
	{    	
    	//(1) check destination
    	if ((mo.nextNodeX != this.nextNodeX) || (mo.nextNodeY != this.nextNodeY))
    	{
            return false;
    	}    	
    	
    	//(2) check distance from the centroid of the cluster
    	if (DistanceBetween2Points(this.centerPoint, mo.GetPosition()) > MotionSCUBAOperator.MotionSCUBADistanceThreshold)
    	{
    		return false;
    	}
    	//(3) check speed difference with the cluster
    	if (Math.abs(this.speed - mo.speed) > MotionSCUBAOperator.MotionSCUBASpeedThreshold)
    	{
    		return false;
    	}
    	
    	//otherwise
    	return true;
	}
    
    public boolean isClusterCandidate(MotionSCUBAMovingObjectState mo,
    	    						  int threshold_distance_x,
    		 						  int threshold_distance_y,
    		 						  int threshold_speed,
    		 						  int threshold_time,       	
    		 						  boolean is_threshold_destination_enabled,
    		 						  String threshold_destination_type,
    		 						  int threshold_destination_approximate_distance)
	{    
    	//first test the destination (if destination threshold is disabled, don't check for destination)
    	if (is_threshold_destination_enabled == true)
    	{
    		if (threshold_destination_type.startsWith("precise"))
    		{
    			//(1) check destination
    	    	if ((mo.nextNodeX != this.nextNodeX) || (mo.nextNodeY != this.nextNodeY))
    	    	{
    	            return false;
    	    	}  
    		}
    		else if (threshold_destination_type.startsWith("approximate"))
    		{
    			//(1) check approximate destination
    	    	if (((mo.nextNodeX < (this.nextNodeX + threshold_destination_approximate_distance)) && 
    	    		 (mo.nextNodeX > (this.nextNodeX - threshold_destination_approximate_distance)))&& 
    	    		((mo.nextNodeY < (this.nextNodeY + threshold_destination_approximate_distance)) && 
    	    	     (mo.nextNodeY > (this.nextNodeY - threshold_destination_approximate_distance))))
    	    	{
    	            //do nothing
    	    	}  
    	    	else
    	    	{
    	    		return false;
    	    	}
    		}    		
    	}
    	 	
    	
    	//(2) check distance from the centroid of the cluster
    	if ((Math.abs(this.centerPoint.GetX() - mo.GetPosition().GetX()) > threshold_distance_x) || (Math.abs(this.centerPoint.GetY()-mo.GetPosition().GetY()) > threshold_distance_y))
    	{
    		return false;
    	}
    	
    	//(3) check speed difference with the cluster
    	if (Math.abs(this.speed - mo.speed) > threshold_speed)
    	{
    		return false;
    	}
    	
    	//(4) check time difference with the cluster
    	if (Math.abs(this.time - mo.GetPosition().GetT()) > threshold_time)
    	{
    		return false;
    	}
    	
    	//otherwise
    	return true;
	}
    

	  
    
    private boolean clusterAlreadyContainsIndex(int indexX, int indexY, int indexT)
	{    	
    	for(int i=0; i<this.gridCellsOverlap.size();i++)
    	{
    		GridCellIndex3D gridIndex = (GridCellIndex3D)gridCellsOverlap.get(i);
    		if ((gridIndex.GetXIndex() == indexX) &&
			    (gridIndex.GetYIndex() == indexY) &&
			    (gridIndex.GetTIndex() == indexT))
    		{
    			return true;
    		}
    	}		
    	return false;
	}
    
    public void RecordGridIndexForCluster(int indexX, int indexY, int indexT)
    {
    	if (!clusterAlreadyContainsIndex(indexX, indexY,indexT))
    	{
    		this.gridCellsOverlap.add(new GridCellIndex3D(indexX,indexY,indexT));
    	}    	
    }
    
    public boolean IsClusterCandidate(MotionSCUBAMovingQueryState mq)
	{    	
    	//(1) check destination
    	if ((mq.nextNodeX != this.nextNodeX) || (mq.nextNodeY != this.nextNodeY))
    	{
            return false;
    	}
    	//(2) check distance from the centroid of the cluster
    	if (DistanceBetween2Points(this.centerPoint, mq.GetPosition()) > MotionSCUBAOperator.MotionSCUBADistanceThreshold)
    	{
    		return false;
    	}
    	//(3) check speed difference with the cluster
    	if (Math.abs(this.speed - mq.speed) > MotionSCUBAOperator.MotionSCUBASpeedThreshold)
    	{
    		return false;
    	}
    	
    	//otherwise
    	return true;
	}
    
    
    public boolean isClusterCandidate(MotionSCUBAMovingQueryState mq,
			  						  int threshold_distance_x,
			  						  int threshold_distance_y,
			  						  int threshold_speed,
			  						  int threshold_time,       	
			  						  boolean is_threshold_destination_enabled,
			  						  String threshold_destination_type,
			  						  int threshold_destination_approximate_distance)
    {    
		//first test the destination (if destination threshold is disabled, don't check for destination)
		if (is_threshold_destination_enabled == true)
		{
			if (threshold_destination_type.startsWith("precise"))
			{
				//(1) check destination
				if ((mq.nextNodeX != this.nextNodeX) || (mq.nextNodeY != this.nextNodeY))
				{
					return false;
				}  
			}
			else if (threshold_destination_type.startsWith("approximate"))
			{
				//(1) check approximate destination
				if (((mq.nextNodeX < (this.nextNodeX + threshold_destination_approximate_distance)) && 
				     (mq.nextNodeX > (this.nextNodeX - threshold_destination_approximate_distance)))&& 
				    ((mq.nextNodeY < (this.nextNodeY + threshold_destination_approximate_distance)) && 
			         (mq.nextNodeY > (this.nextNodeY - threshold_destination_approximate_distance))))
				{
						//do nothing
				}  
				else
				{
						return false;
				
				}    		
			}
    	}
		
		
		//(2) check distance from the centroid of the cluster
		if ((Math.abs(this.centerPoint.GetX() - mq.GetPosition().GetX()) > threshold_distance_x) || (Math.abs(this.centerPoint.GetY()-mq.GetPosition().GetY()) > threshold_distance_y))
		{
			return false;
		}
		
		//(3) check speed difference with the cluster
		if (Math.abs(this.speed - mq.speed) > threshold_speed)
		{
			return false;
		}
		
		//(4) check time difference with the cluster
		if (Math.abs(this.time - mq.GetPosition().GetT()) > threshold_time)
		{
			return false;
		}
		
		//otherwise
		return true;
    }
    
    
    public void adjustOldClusterMemberObject(MotionSCUBAMovingObjectState mo, boolean IsCenterMoved, boolean IsRadiusChanged)
    {
    	//(1) adjust centroid of the cluster
    	this.centerPoint = getNewCentroid(mo.GetPosition(), IsCenterMoved, IsRadiusChanged);
    	
    	//(2) update polar coordinates of the current object to the cluster    	
    	for (int i=0; i<this.objects.size(); i++)
    	{
    		MotionSCUBAClusterObjectMember currMember = (MotionSCUBAClusterObjectMember)this.objects.get(i);
    		if (currMember.id == mo.GetID())
    		{
    			this.objects.remove(currMember);
    			MotionSCUBAClusterObjectMember newClusterMember = new MotionSCUBAClusterObjectMember(mo.GetID(), MotionSCUBAPolarCoordinate.TranslateFromCartesianToPolar(mo.GetPosition()));    	    	
    			this.objects.add(newClusterMember);
    			break;
    		}
    	}
    	    	
    	//(4) adjust the speed
    	this.speed = (this.speed + mo.speed)/2;    
    	//(5) adjust the time
    	this.time = (int) mo.GetPosition().GetT();
    }
    
    void AddClusterMemberDataPoint(DataPoint dataPoint, boolean IsCenterMoved, boolean IsRadiusChanged)
    {
    	//(1) adjust centroid of the cluster
    	this.centerPoint = getNewCentroid(dataPoint.get3DPosition(), IsCenterMoved, IsRadiusChanged);
    	
    	if (dataPoint.getObjName().startsWith("o"))
    	{
    		MotionSCUBAClusterObjectMember newClusterMember = new MotionSCUBAClusterObjectMember(dataPoint.getObjNumericId(), MotionSCUBAPolarCoordinate.TranslateFromCartesianToPolar(dataPoint.get3DPosition()));    	    	
    		this.objects.add(newClusterMember);
    	}
    	else if (dataPoint.getObjName().startsWith("q"))
    	{
    		MotionSCUBAClusterQueryMember newClusterMember = new MotionSCUBAClusterQueryMember(dataPoint.getObjNumericId(), MotionSCUBAPolarCoordinate.TranslateFromCartesianToPolar(dataPoint.get3DPosition()),dataPoint.getBoundary());    	    	
    		this.queries.add(newClusterMember);    		
    	}
    	    	
    	//(4) adjust the speed
    	this.speed = (this.speed + dataPoint.getSpeed())/2;     
    	
    	//(5) adjust the time
    	this.time = (int) dataPoint.getT();
    }
    
    
	
    public void removeOldClusterMemberObject(MotionSCUBAMovingObjectState mo, boolean isEmptyCluster)
    {
    	//(1) adjust centroid of the cluster
    	//this.centerPoint = getNewCentroid(mo.GetPosition());
    	
    	//(2) remove the current object from the cluster    	
    	for (int i=0; i<this.objects.size(); i++)
    	{
    		MotionSCUBAClusterObjectMember currMember = (MotionSCUBAClusterObjectMember)this.objects.get(i);
    		if (currMember.id == mo.GetID())
    		{
    			this.objects.remove(currMember);
    			break;
    		}
    	} 
    	
    	this.memberCount--;
    	
    	if(memberCount == 0)
    	{
    		isEmptyCluster = true;
    	}    	
    }
    
    
    public void removeOldClusterMemberQuery(MotionSCUBAMovingQueryState mq, boolean isEmptyCluster)
    {
    	//(1) adjust centroid of the cluster
    	//this.centerPoint = getNewCentroid(mq.GetPosition());
    	
    	//(2) remove the current query from the cluster    	
    	for (int i=0; i<this.queries.size(); i++)
    	{
    		MotionSCUBAClusterQueryMember currMember = (MotionSCUBAClusterQueryMember)this.queries.get(i);
    		if (currMember.id == mq.GetID())
    		{
    			this.objects.remove(currMember);
    			break;
    		}
    	}
    	
    	this.memberCount--;
    	
    	if(memberCount == 0)
    	{
    		isEmptyCluster = true;
    	}
    }
    
    
    
    public void AdjustOldClusterMemberQuery(MotionSCUBAMovingQueryState mq, boolean IsCenterMoved, boolean IsRadiusChanged)
    {       	
    	//(1) adjust centroid of the cluster
    	this.centerPoint = getNewCentroid(mq.GetPosition(), IsCenterMoved, IsRadiusChanged);
    	
    	//(2) update polar coordinates of the current query to the cluster    	
    	for (int i=0; i<this.queries.size(); i++)
    	{
    		MotionSCUBAClusterQueryMember currMember = (MotionSCUBAClusterQueryMember)this.queries.get(i);
    		if (currMember.id == mq.GetID())
    		{
    			this.queries.remove(currMember);
    			MotionSCUBAClusterQueryMember newClusterMember = new MotionSCUBAClusterQueryMember(mq.GetID(), MotionSCUBAPolarCoordinate.TranslateFromCartesianToPolar(mq.GetPosition()), mq.GetQueryBoundary());    	    	
    			this.queries.add(newClusterMember);
    			break;
    		}
    	}
    	    	
    	//(4) adjust the speed
    	this.speed = (this.speed + mq.speed)/2;  
    	//(5) update the time
    	this.time = (int) mq.GetPosition().GetT();
    }
    
    
    public void AdoptNewClusterMemberObject(MotionSCUBAMovingObjectState mo, boolean IsCenterMoved, boolean IsRadiusChanged)
    {
    	//(1) adjust centroid of the cluster
    	this.centerPoint = getNewCentroid(mo.GetPosition(), IsCenterMoved, IsRadiusChanged);
    	
    	//(2) add polar coordinates of the current object to the cluster
    	MotionSCUBAClusterObjectMember newClusterMember = new MotionSCUBAClusterObjectMember(mo.GetID(), MotionSCUBAPolarCoordinate.TranslateFromCartesianToPolar(mo.GetPosition()));    	
    	this.objects.add(newClusterMember);

    	//(3) increment the count to the members
    	this.memberCount++;
    	
    	//(4) adjust the speed
    	this.speed = (this.speed + mo.speed)/2;  
    	//(5) update teh time
    	this.time = (int) mo.GetPosition().GetT();
    }
    
    public void AdoptNewClusterMemberQuery(MotionSCUBAMovingQueryState mq, boolean IsCenterMoved, boolean IsRadiusChanged)
    {
    	//(1) adjust centroid of the cluster
    	this.centerPoint = getNewCentroid(mq.GetPosition(), IsCenterMoved, IsRadiusChanged);
    	
    	//(2) add polar coordinates of the current query to the cluster
    	MotionSCUBAClusterQueryMember newClusterMember = new MotionSCUBAClusterQueryMember(mq.GetID(), MotionSCUBAPolarCoordinate.TranslateFromCartesianToPolar(mq.GetPosition()), mq.GetQueryBoundary());    	
    	this.queries.add(newClusterMember);
    	
    	//(3) increment the count to the members
    	this.memberCount++;
    	
    	//(4) adjust the speed
    	this.speed = (this.speed + mq.speed)/2;  
    	//(5) update the time
    	this.time = (int) mq.GetPosition().GetT();
    }
    
    public void clusterSheddyAdoptNewClusterMemberQuery(MotionSCUBAMovingQueryState mq, int querySize, boolean IsCenterMoved, boolean IsRadiusChanged)
    {
    	//(1) adjust centroid of the cluster
    	this.centerPoint = getNewCentroid(mq.GetPosition(), IsCenterMoved, IsRadiusChanged);
    	
    	//(2) check if the query size causes the radius of the cluster to change
    	if (this.radius < querySize)
    	{
    		this.radius = querySize;
    		IsRadiusChanged = true;
    	}
    	
    	//(3) add polar coordinates of the current query to the cluster
    	MotionSCUBAClusterQueryMember newClusterMember = new MotionSCUBAClusterQueryMember(mq.GetID(), MotionSCUBAPolarCoordinate.TranslateFromCartesianToPolar(mq.GetPosition()), mq.GetQueryBoundary());    	
    	this.queries.add(newClusterMember);
    	
    	//(4) increment the count to the members
    	this.memberCount++;
    	
    	//(5) adjust the speed
    	this.speed = (this.speed + mq.speed)/2;  
    	//(6) update the time
    	this.time = (int) mq.GetPosition().GetT();
    }
    
    
    private PositionSpt2D getMidpoint(PositionSpt2D pt1, PositionSpt2D pt2)
    {
    	PositionSpt2D result = new PositionSpt2D();
	
    	//get the x midpoint value
    	result.SetX( ((pt1.GetX() - pt2.GetX()) / 2) + pt2.GetX());
    	//get the y midpoint value
    	result.SetY( ((pt1.GetY() - pt2.GetY()) / 2) + pt2.GetY());
	
    	return result;
    }
     
    
    private PositionSpt2D getNewCentroid(PositionSpt2D newLocationUpdate, boolean IsCenterMoved, boolean IsRadiusChanged)
    {
    	//record current values for the center and the radius of the cluster to track the changes
    	PositionSpt2D oldCenter = this.centerPoint;
    	float oldRadius = this.radius;
    	    	
    	PositionSpt2D result = this.centerPoint;
    	
    	double currRadius = DistanceBetween2Points(this.centerPoint, this.furthestPosition);    	
    	double distanceBetweenCenterAndNewLocationUpdate = DistanceBetween2Points(this.centerPoint, newLocationUpdate);
    	
    	if (distanceBetweenCenterAndNewLocationUpdate > currRadius)
    	{   
    		if (currRadius != 0)
    		{    		
    			//adjust new centroid (center between 3 points)
    			result = CenterBetween3Points(this.centerPoint, this.furthestPosition, newLocationUpdate);
    		    			
    			//adjust radius
    			this.radius = (float)DistanceBetween2Points(result, newLocationUpdate);
    		
    			//adjust the furthest point for the cluster
    			this.furthestPosition = newLocationUpdate;
    			
    		}
    		else
    		{
    			//adjust radius
    			this.radius = (float)DistanceBetween2Points(this.centerPoint, newLocationUpdate);
    			
    			//adjust new centroid (center between 2 points)
    			result = getMidpoint(this.centerPoint, newLocationUpdate);    		    
    			
    			//adjust the furthest point for the cluster
    			this.furthestPosition = newLocationUpdate;
    		}
    	}
    	
    	//track that the center of the cluster has moved
    	if (Are2PositionsSame(oldCenter, result))
		{
    		IsCenterMoved = false;
		}
    	else
    	{
    		IsCenterMoved = true;
    	}
    	
    	//track that the radius of the cluster has changed
    	if (oldRadius == this.radius)
    	{
    		IsRadiusChanged = false;
    	}
    	else
    	{
    		IsRadiusChanged = true;
    	}
    	
    	return result;
    }
    
    private boolean Are2PositionsSame(PositionSpt2D p1, PositionSpt2D p2)
    {
    	if((p1.GetX() == p2.GetX()) && (p1.GetY() == p2.GetY()) && (p1.GetT() == p2.GetT()))
    	{
    		return true;
    	}
    	return false;
    }
      
    
    private double DistanceBetween2Points(PositionSpt2D p1, PositionSpt2D p2)
    {
    	return Math.sqrt(Math.pow((p1.GetX() - p2.GetX()),2)+Math.pow((p1.GetY() - p2.GetY()),2));    	
    }
    

    private PositionSpt2D CenterBetween3Points(PositionSpt2D p1, PositionSpt2D p2, PositionSpt2D p3)
    {
    	double temp, bc, cd, det, cx, cy;

    	//matrix math to calculate the determinant
    	temp = Math.pow(p2.GetX(),2) + Math.pow(p2.GetY(),2);
    	bc = (Math.pow(p1.GetX(),2) + Math.pow(p1.GetY(),2) - temp) / 2;
    	cd = (temp - Math.pow(p3.GetX(),2) - Math.pow(p3.GetY(),2)) / 2;
    	det = ((p1.GetX() - p2.GetX()) * (p2.GetY() - p3.GetY())) - ((p2.GetX() - p3.GetX()) *  (p1.GetY() - p2.GetY()));
		det = 1/det;
		
		//calculate the x and y coord of the center point
		cx = ((bc * (p2.GetY() - p3.GetY())) - (cd * (p1.GetY() - p2.GetY()))) * det;
		cy = ((cd * (p1.GetX() - p2.GetX())) - (bc * (p2.GetX() - p3.GetX()))) * det;

		float latestTimestamp = p1.GetT();
		if (p2.GetT() > latestTimestamp)
		{
			latestTimestamp = p2.GetT();
		}
		if (p3.GetT() > latestTimestamp)
		{
			latestTimestamp = p3.GetT();
		}
		//create a point representing the center point
		PositionSpt2D result = new PositionSpt2D((long)cx,(long)cy,latestTimestamp);

		//return the center point
		return result;
    }
    
    public String getStringObjectIds()
    {
    	String outputString = "";
    	for(int i=0; i < this.objects.size();i++)
    	{
    		MotionSCUBAClusterObjectMember obj = (MotionSCUBAClusterObjectMember)objects.get(i);
    		if (i < this.objects.size() - 1){
    		outputString += obj.id + ", ";
    		}else{
    			outputString += obj.id;
    		}
    	}
    	return outputString;
    }
    
    public String getStringQueriesIds()
    {
    	String outputString = "";
    	for(int i=0; i < this.queries.size();i++)
    	{
    		MotionSCUBAClusterQueryMember query = (MotionSCUBAClusterQueryMember)queries.get(i);
    		if (i < this.queries.size() - 1){
    		outputString += query.id + ", ";
    		}else{
    			outputString += query.id;
    		}
    	}
    	return outputString;
    }
}
