/**
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on May 15, 2005
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba;

import java.util.ArrayList;
import java.util.HashMap;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.clustersheddy.QueryResult;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.DimensionType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.QueryType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GenericSpatioTemporalGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridCellIndex3D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.MotionDiscreteResult;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MotionDiscreteState;

/**
 * @author Rimma V. Nehme
 */
public class MotionSCUBAClusterGrid extends GenericSpatioTemporalGrid
{
	//debug flag
	public static int DEBUG = 1; //1=true, 0=false
	
	//worcester parameters
	protected static long xDimensionSize    = 820000; //(108530000 - 107710000); //difference = 820000, previous: 110000000;
	protected static long yDimensionSize    = 800000; //(47800000 - 47200000);//difference = 800000, previous: 50000000;
	
	protected static long xDimensionStart    = 107710000; //(108530000 - 107710000); //difference = 820000, previous: 110000000;
	protected static long yDimensionStart    = 47200000; //(47800000 - 47200000);//difference = 800000, previous: 50000000;

	protected static long xDimensionEnd    = 108530000; //(108530000 - 107710000); //difference = 820000, previous: 110000000;
	protected static long yDimensionEnd    = 48000000; //(47800000 - 47200000);//difference = 800000, previous: 50000000;

	
	public GenericSpatioTemporalSCUBAGridCell[][][] directory;
	private int xDim;
	private int yDim;
	private int tDim;
	

	public String toString()
	{
		return statisticsToString(null);
	}
	
	public int getXDim()
	{
		return xDim;
	}
	public int getYDim()
	{
		return yDim;
	}
	public int getTDim()
	{
		return tDim;
	}
	
	public void Cleanup()
	{
		for (int i=0; i < xDim; i++)
		{
			for (int j=0; j < yDim; j++)
			{
				for (int k=0; k < tDim; k++)
				{
					if (directory[i][j][k] != null)
					{
						if (directory[i][j][k].GetClusters().size() > 0) 
						{
							//remove all clusters from the cell
							directory[i][j][k].GetClusters().clear();
						}
					}
				}
			}
		}
	}
	
	public String statisticsToString(MotionSCUBAClusterStorage clusterStorage)
	{
		int totalNumberOfClusters = 0;		
		String outputString="*** CLUSTER GRID STATISTICS ***";
		
		for (int i=0; i < xDim; i++)
		{
			for (int j=0; j < yDim; j++)
			{
				for (int k=0; k < tDim; k++)
				{
					if (directory[i][j][k] != null)
					{
						totalNumberOfClusters += directory[i][j][k].GetClusters().size();
									
						if (directory[i][j][k].GetClusters().size() > 0)
						{
							outputString += "\ngrid["+i+"]["+j+"]["+k+"]: # clusters = "+directory[i][j][k].GetClusters().size()+"";
							outputString +=  " [ Cell Dimension: i = ("+ directory[i][j][k].GetXMin()+ ", "+directory[i][j][k].GetXMax()+"); "+
														 "j = ("+ directory[i][j][k].GetYMin()+ ", "+directory[i][j][k].GetYMax()+"); "+														 
														 "k = (" + directory[i][j][k].GetTMin()+", "+directory[i][j][k].GetTMax()+");] ";
							if (clusterStorage != null)
							{
								outputString += "\n"+directory[i][j][k].GetClustersStatistics(clusterStorage);
							}						 
						}
					}
				}
			}
		}
		
		outputString += "\nTotal # of cluster entries = "+totalNumberOfClusters+"\n";
		return outputString;
	}
	
	
	public String graphicalStatisticsToString(MotionSCUBAClusterStorage clusterStorage)
	{
		int totalNumberOfClusters = 0;		
		String outputString="*** CLUSTER GRAPHICAL GRID PICTURE ***\n";
		
		for (int a=0; a<xDim;a++)
		{
			outputString += "--";
		}
		outputString += "\n|";
		
		for (int i=0; i < xDim; i++)
		{
			for (int j=0; j < yDim; j++)
			{
				for (int k=0; k < tDim; k++)
				{
					if (directory[i][j][k] != null)
					{
						totalNumberOfClusters += directory[i][j][k].GetClusters().size();
						
						if (directory[i][j][k].GetClusters().size() > 0)
						{
							outputString +=  directory[i][j][k].GetClusters().size() + "|";					 
						}
						else
						{
							outputString += "|";		
						}
					}
					else
					{
						outputString += "n|";		
					}
						
				}				
			}
			outputString += "\n|";		
		}
		
		for (int a=0; a<xDim;a++)
		{
			outputString += "---";
		}
		
		return outputString;
	}
	
	
	public MotionSCUBAClusterGrid(int xDimension, int yDimension, int tDimension) 
	{
		xDim = xDimension;
		yDim = yDimension;
		tDim = tDimension;
		
		directory = new GenericSpatioTemporalSCUBAGridCell[xDimension][yDimension][tDimension];
			
	}


	/*public void CheckCreateGridCell(int i, int j, int k)
	{
		if (directory[i][j][k] == null)
		{			
			directory[i][j][k] = new GenericSpatioTemporalSCUBAGridCell(i,
				 	   j,
					   k,
					   i * GenericSpatioTemporalGrid.xDimensionSize,
					   j * GenericSpatioTemporalGrid.yDimensionSize,
					   k * GenericSpatioTemporalGrid.tDimensionSize,
					  (i+1) * GenericSpatioTemporalGrid.xDimensionSize,
					  (j+1) * GenericSpatioTemporalGrid.yDimensionSize,
					  (k+1) * GenericSpatioTemporalGrid.tDimensionSize);
			
		}
		
	}
	*/
	
	
	public void CheckCreateGridCellByIndexAndSpaceRange(int i, int j, int k)
	{
		if (directory[i][j][k] == null)
		{	
			long xMin = (xDimensionStart + i * (xDimensionSize/this.getXDim()));
			long yMin = (yDimensionStart + j * (yDimensionSize/this.getYDim()));
			int tMin = 1;
			long xMax = (xDimensionStart + (i+1) * (xDimensionSize/this.getXDim()));
			long yMax = (yDimensionStart + (j+1) * (yDimensionSize/this.getYDim()));
			int tMax = 1;
		
				
			directory[i][j][k] = new GenericSpatioTemporalSCUBAGridCell(
					   i,
				 	   j,
					   k,
					   xMin,
					   yMin,
					   1,
					   xMax,
					   yMax,
					   1);

			//System.out.println("creating cell: " + i+", "+ j+", "+ k+", "+ xMin +", "+xMax+", "+yMin+", "+yMax );
		}
		
	}
	
	
	public HashMap GetClustersByCellIndex(int xIndex, int yIndex,int tIndex)
	{
		return directory[xIndex][yIndex][tIndex].GetClusters();
	}

	
	public static boolean IsIndexWithinGridBoundary(int xIndex, int yIndex, int tIndex)
	{			
		if ((xIndex >= MotionSCUBAOperator.MotionSCUBAClusterGridDimensionXCellCount) ||
			(yIndex >= MotionSCUBAOperator.MotionSCUBAClusterGridDimensionYCellCount) ||
			(tIndex >= MotionSCUBAOperator.MotionSCUBAClusterGridDimensionTCellCount))
		{			
			return false;
		}
		else
		{
			return true;
		}
	}
	
	
	
	public void InsertCluster(MotionSCUBACluster mc)	
	{
		int xIndex = MotionSCUBAClusterGrid.GetCellIndex(mc.centerPoint.GetX(), DimensionType.x);		
		int yIndex = MotionSCUBAClusterGrid.GetCellIndex(mc.centerPoint.GetY(), DimensionType.y);
		int tIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)

		if (!IsIndexWithinGridBoundary(xIndex, yIndex, tIndex))
		{
			Helper.DebugWriteLine("cluster " + mc + " must be deleted as it is outside of the boundaries of the grid");
			return; //for now just return without entering the entity
		}
		
		CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex,tIndex);		
		
		SCUBAGridEntry clusterGridEntry = new SCUBAGridEntry(mc.cid);		
		directory[xIndex][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
		
		//update grid index for the centroid of the cluster
		mc.currentGridIndexX = xIndex;
		mc.currentGridIndexY = yIndex;
		mc.currentGridIndexT = tIndex;
		
		//record the index for the cluster
		mc.RecordGridIndexForCluster(xIndex,yIndex,tIndex);
				
		//check if the boundary of the query might cause it to intersect
		//with other neightboring grid cells
		
		PositionSpt2D cellMin = MotionSCUBAClusterGrid.GetCellWindowMin(new GridCellIndex3D(xIndex,yIndex,tIndex));
		PositionSpt2D cellMax = MotionSCUBAClusterGrid.GetCellWindowMax(new GridCellIndex3D(xIndex,yIndex,tIndex));
		
		PositionSpt2D cellMidTop   = MotionSCUBAClusterGrid.GetCellTopMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
		PositionSpt2D cellMidBot   = MotionSCUBAClusterGrid.GetCellBottomMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
		PositionSpt2D cellMidRight = MotionSCUBAClusterGrid.GetCellRightMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
		PositionSpt2D cellMidLeft  = MotionSCUBAClusterGrid.GetCellLeftMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
		
		/*
		 * ---------MP
		 * |        |
		 * |        |
		 * |        |
		 * ----------
		 */
		if (((mc.centerPoint.GetX() + mc.radius) > cellMax.GetX()) && (mc.centerPoint.GetY() + mc.radius) > cellMax.GetY())
		{
			CheckCreateGridCellByIndexAndSpaceRange(xIndex+1,yIndex+1,tIndex);			
			directory[xIndex+1][yIndex+1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
			//record the index for the cluster
			mc.RecordGridIndexForCluster(xIndex+1,yIndex+1,tIndex);
		}
		
		/*
		 * ----------
		 * |        |
		 * |        |
		 * |        |
		 * MP--------
		 */
		if ((mc.centerPoint.GetX() - mc.radius < cellMin.GetX()) && (mc.centerPoint.GetY() - mc.radius < cellMin.GetY()))
		{
			CheckCreateGridCellByIndexAndSpaceRange(xIndex-1,yIndex-1,tIndex);
			directory[xIndex-1][yIndex-1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
			//record the index for the cluster
			mc.RecordGridIndexForCluster(xIndex-1,yIndex-1,tIndex);
		}
		
		
		/*
		 * ----MP----
		 * |        |
		 * |        |
		 * |        |
		 * ----------
		 */
		if (mc.centerPoint.GetY() + mc.radius > cellMidTop.GetY())
		{
			CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex+1,tIndex);
			directory[xIndex][yIndex+1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
			//record the index for the cluster
			mc.RecordGridIndexForCluster(xIndex,yIndex+1,tIndex);
		}
		
		/*
		 * ----------
		 * |        |
		 * |        |
		 * |        |
		 * ----MP----
		 */
		if (mc.centerPoint.GetY() - mc.radius < cellMidTop.GetY())
		{
			CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex-1,tIndex);
			directory[xIndex][yIndex-1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
			//record the index for the cluster
			mc.RecordGridIndexForCluster(xIndex,yIndex-1,tIndex);
		}
		
		
		/*
		 * ----------
		 * |        |
		 * |        MP
		 * |        |
		 * ----------
		 */
		if (mc.centerPoint.GetX() + mc.radius > cellMidRight.GetX())
		{
			CheckCreateGridCellByIndexAndSpaceRange(xIndex+1,yIndex,tIndex);
			directory[xIndex+1][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
			//record the index for the cluster
			mc.RecordGridIndexForCluster(xIndex+1,yIndex,tIndex);
		}
		
		/*
		 * ----------
		 * |        |
		 * MP       |
		 * |        |
		 * ----------
		 */
		if (mc.centerPoint.GetX() - mc.radius < cellMidLeft.GetX())
		{
			CheckCreateGridCellByIndexAndSpaceRange(xIndex-1,yIndex,tIndex);
			directory[xIndex-1][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
			//record the index for the cluster
			mc.RecordGridIndexForCluster(xIndex-1,yIndex,tIndex);
		}
	}
	

	public boolean DoBetweenJoin(SCUBAGridEntry cluster1Entry, 
							     SCUBAGridEntry cluster2Entry, 
								 GridCellIndex3D gi,
								 MotionSCUBAClusterStorage clusterStorage)
	{
		//retrieve clusters from ClusterStorage and join them
		MotionSCUBACluster cluster1 = clusterStorage.getClusterByCID(new Integer(cluster1Entry.GetID()));
		MotionSCUBACluster cluster2 = clusterStorage.getClusterByCID(new Integer(cluster2Entry.GetID()));
		
		//check if two clusters overlap

		//determine the distance between the cluster to ensure they overlap by some amount
		long x1 = cluster1.centerPoint.GetX();
		long y1 = cluster1.centerPoint.GetY();
		long x2 = cluster2.centerPoint.GetX();
		long y2 = cluster2.centerPoint.GetY();
		
		double distance = Math.sqrt( Math.pow(x2 - x1,2) + Math.pow(y2 - y1,2));
		
		//the clusters overlap
		if(distance <= (cluster1.radius + cluster2.radius))
		{
			return true;		
		}
		
		return false;
	}
	
	
	public static MotionDiscreteResult DoDiscreteJoin(MotionSCUBAClusterQueryMember queryState, 
											   PositionSpt2D queryClusterCenter,
											   MotionSCUBAClusterObjectMember objectState,
											   PositionSpt2D objectClusterCenter)
	{
		MotionDiscreteResult intersectionResult = null;
		MotionDiscreteState discreteResultPosition = null;
		
		PositionSpt2D queryPosition = MotionSCUBAPolarCoordinate.TranslateFromPolarToCartesian(queryState.coordinates, queryClusterCenter);
		PositionSpt2D objectPosition = MotionSCUBAPolarCoordinate.TranslateFromPolarToCartesian(objectState.coordinates, objectClusterCenter);

		if ((objectPosition.GetX() < (queryPosition.GetX() + queryState.queryBoundary))&&
			(objectPosition.GetY() < (queryPosition.GetY() + queryState.queryBoundary))&&
			(objectPosition.GetX() > (queryPosition.GetX() - queryState.queryBoundary))&&
			(objectPosition.GetY() > (queryPosition.GetY() - queryState.queryBoundary))&&
			(objectPosition.GetT() == queryPosition.GetT()))
		{
			discreteResultPosition = new MotionDiscreteState(objectState.id,
															 EntityType.Object,
															 objectPosition.GetX(),
															 objectPosition.GetY(), 
															 objectPosition.GetT());

			intersectionResult = new MotionDiscreteResult(queryState.id,  
						  								  objectState.id, 
														  QueryType.Range, 
														  discreteResultPosition);
		}			
		return intersectionResult;
	}
	
	public static QueryResult clusterSheddyDoDiscreteJoin(MotionSCUBAClusterQueryMember queryState, 
			   PositionSpt2D queryClusterCenter,
			   MotionSCUBAClusterObjectMember objectState,
			   PositionSpt2D objectClusterCenter,
			   int querySize)
	{
		QueryResult queryResult = null;
		
		PositionSpt2D queryPosition = MotionSCUBAPolarCoordinate.TranslateFromPolarToCartesian(queryState.coordinates, queryClusterCenter);
		PositionSpt2D objectPosition = MotionSCUBAPolarCoordinate.TranslateFromPolarToCartesian(objectState.coordinates, objectClusterCenter);
		
		if ((objectPosition.GetX() < (queryPosition.GetX() + querySize))&&
		(objectPosition.GetY() < (queryPosition.GetY() + querySize))&&
		(objectPosition.GetX() > (queryPosition.GetX() - querySize))&&
		(objectPosition.GetY() > (queryPosition.GetY() - querySize))&&
		(objectPosition.GetT() == queryPosition.GetT()))
		{
			queryResult = new QueryResult(queryState.id, objectState.id, (int) queryClusterCenter.GetT(), "jb+jw");			
		}			
		return queryResult;
	}

	
	
	
	
	public ArrayList DoWithinJoin(SCUBAGridEntry cluster1Entry, 
		     					  SCUBAGridEntry cluster2Entry, 
								  GridCellIndex3D gi,
								  MotionSCUBAClusterStorage clusterStorage)
	{
		 ArrayList withinClusterJoinResults = new ArrayList();
		//retrieve clusters from ClusterStorage and join them
		MotionSCUBACluster cluster1 = clusterStorage.getClusterByCID(new Integer(cluster1Entry.GetID()));
		MotionSCUBACluster cluster2 = clusterStorage.getClusterByCID(new Integer(cluster2Entry.GetID()));
		
		//do join between all the elements in that cluster
		for(int i=0; i < cluster1.queries.size(); i++)
		{
			for(int j=0; j < cluster2.objects.size(); j++)
			{
				MotionDiscreteResult joinResult = null;
				MotionSCUBAClusterQueryMember queryState = (MotionSCUBAClusterQueryMember)cluster1.queries.get(i);
				MotionSCUBAClusterObjectMember objectState = (MotionSCUBAClusterObjectMember)cluster2.objects.get(j);				
				joinResult = DoDiscreteJoin(queryState,cluster1.centerPoint,objectState,cluster2.centerPoint);
				
				//add the result to the collection
				if (joinResult != null)
				{
					withinClusterJoinResults.add(joinResult);
				}
			}			
		}
		
		for(int m=0; m < cluster2.queries.size(); m++)
		{
			for(int n=0; n < cluster1.objects.size(); n++)
			{
				MotionDiscreteResult joinResult = null;
				MotionSCUBAClusterQueryMember queryState = (MotionSCUBAClusterQueryMember)cluster2.queries.get(m);
				MotionSCUBAClusterObjectMember objectState = (MotionSCUBAClusterObjectMember)cluster1.objects.get(n);
				joinResult = DoDiscreteJoin(queryState, cluster2.centerPoint, objectState, cluster1.centerPoint);
				
				//add the result to the collection
				if (joinResult != null)
				{
					withinClusterJoinResults.add(joinResult);
				}
			}			
		}
	
		return withinClusterJoinResults;
	}
	
	
	public ArrayList clusterSheddyDoWithinJoin(SCUBAGridEntry cluster1Entry, 
			  								   SCUBAGridEntry cluster2Entry, 
			  								   GridCellIndex3D gi,
			  								   MotionSCUBAClusterStorage clusterStorage,
			  								   int querySize)
	{
		ArrayList withinClusterJoinResults = new ArrayList();
		//retrieve clusters from ClusterStorage and join them
		MotionSCUBACluster cluster1 = clusterStorage.getClusterByCID(new Integer(cluster1Entry.GetID()));
		MotionSCUBACluster cluster2 = clusterStorage.getClusterByCID(new Integer(cluster2Entry.GetID()));
		
		//do join between all the elements in that cluster
		for(int i=0; i < cluster1.queries.size(); i++)
		{
			for(int j=0; j < cluster2.objects.size(); j++)
			{
				QueryResult joinResult = null;
				MotionSCUBAClusterQueryMember queryState = (MotionSCUBAClusterQueryMember)cluster1.queries.get(i);
				MotionSCUBAClusterObjectMember objectState = (MotionSCUBAClusterObjectMember)cluster2.objects.get(j);				
				joinResult = clusterSheddyDoDiscreteJoin(queryState,cluster1.centerPoint,objectState,cluster2.centerPoint, querySize);				
				
				//add the result to the collection
				if (joinResult != null)
				{
					withinClusterJoinResults.add(joinResult);
				}
			}			
		}
		
		for(int m=0; m < cluster2.queries.size(); m++)
		{
			for(int n=0; n < cluster1.objects.size(); n++)
			{
				QueryResult joinResult = null;
				MotionSCUBAClusterQueryMember queryState = (MotionSCUBAClusterQueryMember)cluster2.queries.get(m);
				MotionSCUBAClusterObjectMember objectState = (MotionSCUBAClusterObjectMember)cluster1.objects.get(n);
				joinResult = clusterSheddyDoDiscreteJoin(queryState, cluster2.centerPoint, objectState, cluster1.centerPoint, querySize);
				
				//add the result to the collection
				if (joinResult != null)
				{
					withinClusterJoinResults.add(joinResult);
				}
			}			
		}
		
		return withinClusterJoinResults;
	}


	
	public static int getCellIndex(long position, int dimension)
	{
		int cellIndex = -1;
		if (dimension == DimensionType.x)
		{
			cellIndex = (int)((position-xDimensionStart)/ xDimensionSize);
		}
		else if (dimension == DimensionType.y)
		{
			cellIndex = (int)((position-yDimensionStart) / yDimensionSize);
		}
		else if (dimension == DimensionType.t)
		{
			cellIndex = 0;
		}
		return cellIndex;		
	}
	
	public static int GetCellIndexByLocation(long position, int dimension, int cellCount)
	{
		int cellIndex = -1;
		int cellSpatialRange = -1;
		
		if (dimension == DimensionType.x)
		{
			cellSpatialRange = (int)MotionSCUBAClusterGrid.xDimensionSize/cellCount;
			cellIndex = (int)((position - MotionSCUBAClusterGrid.xDimensionStart)/(cellSpatialRange));				
		}
		else if (dimension == DimensionType.y)
		{
			cellSpatialRange = (int)MotionSCUBAClusterGrid.yDimensionSize/cellCount;
			cellIndex = (int)((position - MotionSCUBAClusterGrid.yDimensionStart)/(cellSpatialRange));
		}
		else if (dimension == DimensionType.t)
		{
			cellIndex = 0;
		}
		return cellIndex;		
	}
	
	
	public static long GetXDimensionCellSize()
	{
		return (MotionSCUBAClusterGrid.xDimensionSize)/MotionSCUBAOperator.MotionSCUBAClusterGridDimensionXCellCount;
	}
	
	public static long GetYDimensionCellSize()
	{
		return (MotionSCUBAClusterGrid.yDimensionSize)/MotionSCUBAOperator.MotionSCUBAClusterGridDimensionYCellCount;
	}
	
	
	
	/*
	 * ----------
	 * |        |
	 * |        |
	 * |        |
	 * MP--------
	 */
	public static PositionSpt2D GetCellWindowMin(GridCellIndex3D index)
	{
		long xPos = (long)(index.GetXIndex() * MotionSCUBAClusterGrid.GetXDimensionCellSize());
		long yPos = (long)(index.GetYIndex() * MotionSCUBAClusterGrid.GetYDimensionCellSize());
		long tPos = 0;
		return new PositionSpt2D(xPos, yPos, tPos);
	}
	
	
	/*
	 * ---------MP
	 * |        |
	 * |        |
	 * |        |
	 * ----------
	 */
	public static PositionSpt2D GetCellWindowMax(GridCellIndex3D index)
	{
		long xPos = (long)((index.GetXIndex() + 1) * MotionSCUBAClusterGrid.GetXDimensionCellSize());
		long yPos = (long)((index.GetYIndex() + 1) * MotionSCUBAClusterGrid.GetYDimensionCellSize());
		long tPos = 0;
		return new PositionSpt2D(xPos, yPos, tPos);
	}
	
	
	/*
	 * ----MP----
	 * |        |
	 * |        |
	 * |        |
	 * ----------
	 */
	public static PositionSpt2D GetCellTopMidPoint(GridCellIndex3D index)
	{
		long xPos = (long)((index.GetXIndex() * MotionSCUBAClusterGrid.GetXDimensionCellSize()) + (MotionSCUBAClusterGrid.GetXDimensionCellSize() / 2));
		long yPos = (long)((index.GetYIndex() + 1) * MotionSCUBAClusterGrid.GetYDimensionCellSize());
		long tPos = 0;
		return new PositionSpt2D(xPos, yPos, tPos);
	}
	
	/*
	 * ----------
	 * |        |
	 * |        |
	 * |        |
	 * ----MP----
	 */
	public static PositionSpt2D GetCellBottomMidPoint(GridCellIndex3D index)
	{
		long xPos = (long)((index.GetXIndex() * MotionSCUBAClusterGrid.GetXDimensionCellSize()) + (MotionSCUBAClusterGrid.GetXDimensionCellSize() / 2));
		long yPos = (long)(index.GetYIndex() * MotionSCUBAClusterGrid.GetYDimensionCellSize());
		long tPos = 0;
		return new PositionSpt2D(xPos, yPos, tPos);
	}
	
	/*
	 * ----------
	 * |        |
	 * |        MP
	 * |        |
	 * ----------
	 */
	public static PositionSpt2D GetCellRightMidPoint(GridCellIndex3D index)
	{
		long xPos = (long)((index.GetXIndex() + 1) * MotionSCUBAClusterGrid.GetXDimensionCellSize());
		long yPos = (long)((index.GetYIndex() * MotionSCUBAClusterGrid.GetYDimensionCellSize()) + (MotionSCUBAClusterGrid.GetYDimensionCellSize() / 2));
		long tPos = 0;
		return new PositionSpt2D(xPos, yPos, tPos);
	}
	
	/*
	 * ----------
	 * |        |
	 * MP       |
	 * |        |
	 * ----------
	 */
	public static PositionSpt2D GetCellLeftMidPoint(GridCellIndex3D index)
	{
		long xPos = (long)(index.GetXIndex() * MotionSCUBAClusterGrid.GetXDimensionCellSize());
		long yPos = (long)((index.GetYIndex() * MotionSCUBAClusterGrid.GetYDimensionCellSize()) + (MotionSCUBAClusterGrid.GetYDimensionCellSize() / 2));
		long tPos = 0;
		return new PositionSpt2D(xPos, yPos, tPos);
	}
	
	public void RemoveCluster(MotionSCUBACluster mc)	
	{						
		SCUBAGridEntry clusterGridEntry = new SCUBAGridEntry(mc.cid);	
		
		for(int i=0; i<mc.gridCellsOverlap.size();i++)
		{
			GridCellIndex3D index = (GridCellIndex3D)mc.gridCellsOverlap.get(i);
			directory[index.GetXIndex()][index.GetYIndex()][index.GetTIndex()].RemoveMovingClusterEntry(clusterGridEntry);
			
			//if the cell doesn't have any other clusters, remove it from memory (to free up some space)
			CheckRemoveGridCell(index.GetXIndex(),index.GetYIndex(),index.GetTIndex());					
		}
	}
	
	public void CheckRemoveGridCell(int i, int j, int k)
	{
		if (directory[i][j][k].movingClusters.size() == 0)
		{	
			directory[i][j][k].movingClusters = null;
			directory[i][j][k] = null;			
		}		
	}
		
	
/*	
	public void AdjustCluster(MotionSCUBACluster mc, boolean IsCenterMoved, boolean IsRadiusChanged, boolean IsOverlappingWithMultipleCells)	
	{
		//adjust ONLY if the center of the cluster moved OR the radius of the cluster changed
		//otherwise do nothing
		if ((IsCenterMoved == true) || (IsRadiusChanged == true))
		{
			int xIndex = MotionSCUBAClusterGrid.GetCellIndex(mc.centerPoint.GetX(), DimensionType.x);		
			int yIndex = MotionSCUBAClusterGrid.GetCellIndex(mc.centerPoint.GetY(), DimensionType.y);
			int tIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)

			if (!IsIndexWithinGridBoundary(xIndex, yIndex, tIndex))
			{
				Helper.DebugWriteLine("cluster " + mc + " must be deleted as it is outside of the boundaries of the grid");
				return; //for now just return without entering the entity
			}
		
			CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex,tIndex);		
		
			SCUBAGridEntry clusterGridEntry = new SCUBAGridEntry(mc.cid);		
			directory[xIndex][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
		
			//update grid index for the centroid of the cluster
			mc.currentGridIndexX = xIndex;
			mc.currentGridIndexY = yIndex;
			mc.currentGridIndexT = tIndex;
		
			//record the index for the cluster
			mc.RecordGridIndexForCluster(xIndex,yIndex,tIndex);
				
			//check if the boundary of the query might cause it to intersect
			//with other neightboring grid cells
		
			PositionSpt2D cellMin = MotionSCUBAClusterGrid.GetCellWindowMin(new GridCellIndex3D(xIndex,yIndex,tIndex));
			PositionSpt2D cellMax = MotionSCUBAClusterGrid.GetCellWindowMax(new GridCellIndex3D(xIndex,yIndex,tIndex));
		
			PositionSpt2D cellMidTop   = MotionSCUBAClusterGrid.GetCellTopMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
			PositionSpt2D cellMidBot   = MotionSCUBAClusterGrid.GetCellBottomMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
			PositionSpt2D cellMidRight = MotionSCUBAClusterGrid.GetCellRightMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
			PositionSpt2D cellMidLeft  = MotionSCUBAClusterGrid.GetCellLeftMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
		
		/*
		 * ---------MP
		 * |        |
		 * |        |
		 * |        |
		 * ----------
		 */
			
		/*if (((mc.centerPoint.GetX() + mc.radius) > cellMax.GetX()) && (mc.centerPoint.GetY() + mc.radius) > cellMax.GetY())
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex+1,yIndex+1,tIndex);			
				directory[xIndex+1][yIndex+1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex+1,yIndex+1,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
		*/
		
		/*
		 * ----------
		 * |        |
		 * |        |
		 * |        |
		 * MP--------
		 */
		/*	if ((mc.centerPoint.GetX() - mc.radius < cellMin.GetX()) && (mc.centerPoint.GetY() - mc.radius < cellMin.GetY()))
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex-1,yIndex-1,tIndex);
				directory[xIndex-1][yIndex-1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex-1,yIndex-1,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
		*/
		
		/*
		 * ----MP----
		 * |        |
		 * |        |
		 * |        |
		 * ----------
		 */
		/*	if (mc.centerPoint.GetY() + mc.radius > cellMidTop.GetY())
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex+1,tIndex);
				directory[xIndex][yIndex+1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex,yIndex+1,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
			*/
		
		/*
		 * ----------
		 * |        |
		 * |        |
		 * |        |
		 * ----MP----
		 */
			/*if (mc.centerPoint.GetY() - mc.radius < cellMidTop.GetY())
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex-1,tIndex);
				directory[xIndex][yIndex-1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex,yIndex-1,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
		*/
		
		/*
		 * ----------
		 * |        |
		 * |        MP
		 * |        |
		 * ----------
		 */
		/*	if (mc.centerPoint.GetX() + mc.radius > cellMidRight.GetX())
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex+1,yIndex,tIndex);
				directory[xIndex+1][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex+1,yIndex,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
		*/
		/*
		 * ----------
		 * |        |
		 * MP       |
		 * |        |
		 * ----------
		 */
		/*	if (mc.centerPoint.GetX() - mc.radius < cellMidLeft.GetX())
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex-1,yIndex,tIndex);
				directory[xIndex-1][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex-1,yIndex,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
		}

	}
	
	*/
	
	
	public void adjustCluster(MotionSCUBACluster mc, boolean IsCenterMoved, boolean IsRadiusChanged, boolean IsOverlappingWithMultipleCells)	
	{
		//adjust ONLY if the center of the cluster moved OR the radius of the cluster changed
		//otherwise do nothing
		if ((IsCenterMoved == true) || (IsRadiusChanged == true))
		{
			int xIndex = MotionSCUBAClusterGrid.getCellIndex(mc.centerPoint.GetX(), DimensionType.x);		
			int yIndex = MotionSCUBAClusterGrid.GetCellIndex(mc.centerPoint.GetY(), DimensionType.y);
			int tIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)

			if (!IsIndexWithinGridBoundary(xIndex, yIndex, tIndex))
			{
				Helper.DebugWriteLine("cluster " + mc + " must be deleted as it is outside of the boundaries of the grid");
				return; //for now just return without entering the entity
			}
		
			CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex,tIndex);		
		
			SCUBAGridEntry clusterGridEntry = new SCUBAGridEntry(mc.cid);		
			directory[xIndex][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
		
			//update grid index for the centroid of the cluster
			mc.currentGridIndexX = xIndex;
			mc.currentGridIndexY = yIndex;
			mc.currentGridIndexT = tIndex;
		
			//record the index for the cluster
			mc.RecordGridIndexForCluster(xIndex,yIndex,tIndex);
				
			//check if the boundary of the query might cause it to intersect
			//with other neightboring grid cells
		
			PositionSpt2D cellMin = MotionSCUBAClusterGrid.GetCellWindowMin(new GridCellIndex3D(xIndex,yIndex,tIndex));
			PositionSpt2D cellMax = MotionSCUBAClusterGrid.GetCellWindowMax(new GridCellIndex3D(xIndex,yIndex,tIndex));
		
			PositionSpt2D cellMidTop   = MotionSCUBAClusterGrid.GetCellTopMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
			PositionSpt2D cellMidBot   = MotionSCUBAClusterGrid.GetCellBottomMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
			PositionSpt2D cellMidRight = MotionSCUBAClusterGrid.GetCellRightMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
			PositionSpt2D cellMidLeft  = MotionSCUBAClusterGrid.GetCellLeftMidPoint(new GridCellIndex3D(xIndex,yIndex,tIndex));
		
		/*
		 * ---------MP
		 * |        |
		 * |        |
		 * |        |
		 * ----------
		 */
			if (((mc.centerPoint.GetX() + mc.radius) > cellMax.GetX()) && (mc.centerPoint.GetY() + mc.radius) > cellMax.GetY())
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex+1,yIndex+1,tIndex);			
				directory[xIndex+1][yIndex+1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex+1,yIndex+1,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
		
		/*
		 * ----------
		 * |        |
		 * |        |
		 * |        |
		 * MP--------
		 */
			if ((mc.centerPoint.GetX() - mc.radius < cellMin.GetX()) && (mc.centerPoint.GetY() - mc.radius < cellMin.GetY()))
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex-1,yIndex-1,tIndex);
				directory[xIndex-1][yIndex-1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex-1,yIndex-1,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
		
		
		/*
		 * ----MP----
		 * |        |
		 * |        |
		 * |        |
		 * ----------
		 */
			if (mc.centerPoint.GetY() + mc.radius > cellMidTop.GetY())
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex+1,tIndex);
				directory[xIndex][yIndex+1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex,yIndex+1,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
		
		/*
		 * ----------
		 * |        |
		 * |        |
		 * |        |
		 * ----MP----
		 */
			if (mc.centerPoint.GetY() - mc.radius < cellMidTop.GetY())
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex-1,tIndex);
				directory[xIndex][yIndex-1][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex,yIndex-1,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
		
		
		/*
		 * ----------
		 * |        |
		 * |        MP
		 * |        |
		 * ----------
		 */
			if (mc.centerPoint.GetX() + mc.radius > cellMidRight.GetX())
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex+1,yIndex,tIndex);
				directory[xIndex+1][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex+1,yIndex,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
		
		/*
		 * ----------
		 * |        |
		 * MP       |
		 * |        |
		 * ----------
		 */
			if (mc.centerPoint.GetX() - mc.radius < cellMidLeft.GetX())
			{
				CheckCreateGridCellByIndexAndSpaceRange(xIndex-1,yIndex,tIndex);
				directory[xIndex-1][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
				//record the index for the cluster
				mc.RecordGridIndexForCluster(xIndex-1,yIndex,tIndex);
				
				IsOverlappingWithMultipleCells = true;
			}
		}

	}
	
}
