/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on May 11, 2005
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.MotionSCUBAStatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.MotionStatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.ClassType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.QueryType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.UpdateType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GenericSpatioTemporalGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridCellIndex3D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridEntry;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.MemoryTestBench;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.ResultsRecorder;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.MotionGenericOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.MotionDiscreteResult;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.collections.MotionDiscreteResultsCollection;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MotionDiscreteState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingObjectState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingQueryState;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * @author Rimma V. Nehme
 */
public class MotionSCUBADuplicateFreeRangeOperator extends MotionSCUBAOperator
{
	
	protected HashMap clusterJoinRecorder; //hash map for recording already
									//joined moving clusters (used for
									//duplicate elimination)
	
	//debug flag	
	public static int DEBUG = 1; //1=true, 0=false
	public static int STATS_FLAG = 0; //1=verbose, 0=short
	public static int RESULTS_FLAG = 1; //1=return answers, 0=don't return  answers
	public static int NO_DUPLICATES = 1; //1=turn on duplicate elimination, 
										 //0=turn off duplicate elimination
	public static int LOAD_SHEDDING_FACTOR = 1;
		
	private static boolean isTheFirstIteration = true;
	
	public static ArrayList futureTuplesMovingObjects = new ArrayList();
	public static ArrayList futureTuplesMovingQueries = new ArrayList();
	
	public static ArrayList futureTuplesMovingObjectsIndex = new ArrayList();
	public static ArrayList futureTuplesMovingQueriesIndex = new ArrayList();
	
	MotionDiscreteResultsCollection results;

	boolean pushTimeIntervalForwardMovingObjects = false;
	boolean pushTimeIntervalForwardMovingQueries = false;
	MotionSCUBAStatisticsGatherer motionSCUBAOperatorStatistics;

	
	private class FutureTuplesTimeStampIndex
	{
		public int timestamp;
		public int arrayIndex;
		
		public FutureTuplesTimeStampIndex(int newTimestamp, int newArrayIndex)
		{
			timestamp = newTimestamp;
			arrayIndex = newArrayIndex;			
		}
	}
	
	
	public MotionSCUBADuplicateFreeRangeOperator(XATNode node) 
	{
		super(node);	
		
		if(results != null)
		{
			//clear the incremental results collection
			results.Clear();
		}
		
		if (NO_DUPLICATES == 1)
		{
			//initialize the join recorder
			clusterJoinRecorder = new HashMap();
		}
	}
	
	private void recordJoinPair(int cid1, int cid2)
	{
		int smaller;
		int larger;
		
		//always sort the ids before recording them
		if (cid1 > cid2)
		{
			larger = cid1;
			smaller = cid2;
		}
		else if (cid2 > cid1)
		{
			larger = cid2;
			smaller = cid1;			
		}
		else
		{
			//they are equal
			larger = cid1;
			smaller = cid2;
		}
		
		//insert the pair into the join recorder
		this.clusterJoinRecorder.put(new String("c"+larger+"c"+smaller), "done");
	}
	
	private boolean isAlreadyJoined(int cid1, int cid2)
	{		
		int smaller;
		int larger;
		
		//sort ids
		if (cid1 > cid2)
		{
			larger = cid1;
			smaller = cid2;
		}
		else if (cid2 > cid1)
		{
			larger = cid2;
			smaller = cid1;			
		}
		else
		{
			//they are equal
			larger = cid1;
			smaller = cid2;
		}
		
		if (clusterJoinRecorder.get(new String("c"+larger+"c"+smaller)) == null)
		{
			return false;
		}
		else
		{
			return true;
		}
	}
	
	
	public boolean run() 
	{
		//statistics
		motionSCUBAOperatorStatistics = new MotionSCUBAStatisticsGatherer();
		
		try
		{
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nINSIDE SCUBA RANGE OPERATOR!!!");
			}
			
			//System.out.println("\nINSIDE SCUBA RANGE OPERATOR!!!");			
			
			if ((this.getLeftInputQueue().getSchema() == null) || (this.getRightInputQueue().getSchema() == null))
			{
				return false;
			}
			if(this.getOutputQueue().getSchema() == null)
			{
				init();
			}
		
			if ((pushTimeIntervalForwardMovingObjects == true) && (pushTimeIntervalForwardMovingQueries == true))
			{
				//increment execution count
				MotionGenericOperator.executionIterationCount++;
				pushTimeIntervalForwardMovingObjects = false;
				pushTimeIntervalForwardMovingQueries = false;
			}
		
			if(DEBUG == 1)
			{
				//Helper.DebugWriteLine("\nNumber of supported objects = "+ MotionGenericOperator.numberOfSupportedObjects);
				//Helper.DebugWriteLine("Number of supported queries = "+ MotionGenericOperator.numberOfSupportedQueries +"\n");
				Helper.DebugWriteLine("Current execution count = " + MotionGenericOperator.executionIterationCount);
				Helper.DebugWriteLine("Current delta = " + MotionGenericOperator.delta);
			}
		
			//set the delta
			MotionGenericOperator.delta = 1; //allow only location update per object and query per execution

			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Current execution count = " + MotionGenericOperator.executionIterationCount);
				Helper.DebugWriteLine("Current delta = " + MotionGenericOperator.delta);
			}
		
			//initialize results structure
			results = new MotionDiscreteResultsCollection();
			
			motionSCUBAOperatorStatistics.memoryBefore = MotionStatisticsGatherer.getUsedMemory();						
	
			//statistics
			motionSCUBAOperatorStatistics.executionCount = MotionGenericOperator.executionIterationCount + 1;
			motionSCUBAOperatorStatistics.timeOverallStart = System.currentTimeMillis();
		
			//actual operator logic
			executeMotionSCUBAOperator();		

			MotionStatisticsGatherer.totalTimeElapsed += (motionSCUBAOperatorStatistics.timeOverallEnd - motionSCUBAOperatorStatistics.timeOverallStart);
			
			//memory after
			motionSCUBAOperatorStatistics.memoryAfter = MotionStatisticsGatherer.getUsedMemory();
			
			
			//write the statistics out (only if join occurred)
			if ((pushTimeIntervalForwardMovingObjects == true) && (pushTimeIntervalForwardMovingQueries == true))
			{	
				if (STATS_FLAG == 1)
				{
					//verbose statistics
					Helper.StatsWriteLine(motionSCUBAOperatorStatistics.SCUBAStatisticsToString());
				}
				else
				{
					//short statistics
					Helper.StatsWriteLine(motionSCUBAOperatorStatistics.ShortSCUBAStatisticsToString());
				}
			}
		}
		catch(Exception e) 
		{
			Helper.DebugWriteLine("OOH SOMETHING BAD HAPPENED. Exception = " + e.toString());
			Helper.DebugWriteLine(e.getCause().toString());
			e.printStackTrace();
		}
		/* run successfully */
		return true;
	}
	
		
	private void clusterUpdateMovingObject(Time currentTime, 
			   							   int delta, 
										   MotionSCUBAMovingObjectState movingObjectUpdate)
	{
		//check if OID is in the Objects table
		if (objectsHash.get(new Integer(movingObjectUpdate.GetID())) == null )
		{
			numberOfSupportedObjects++;	
			MotionStatisticsGatherer.totalNumberObjectsEnteredIntoHashTable++;						
		}

		MotionStatisticsGatherer.totalNumberObjectsLocationUpdates++;
		objectsHash.put(new Integer(movingObjectUpdate.GetID()), movingObjectUpdate);

		boolean IsOverlappingWithMultipleCells = false;
		//probe for a cluster
		MotionSCUBAHelper.clusterMovingObject(clusterGrid, clusterStorage, clusterHome, movingObjectUpdate, currentTime, delta, motionSCUBAOperatorStatistics, IsOverlappingWithMultipleCells,LOAD_SHEDDING_FACTOR);
		
		if (IsOverlappingWithMultipleCells)
		{
			this.motionSCUBAOperatorStatistics.numberOfClustersOverlappingWithMoreThanOneGridCell++;
		}
	}
	
	
	private void clusterUpdateMovingQuery(Time currentTime, 
				  						  int delta, 
										  MotionSCUBAMovingQueryState movingQueryUpdate)
	{
	    if (queriesHash.get(new Integer(movingQueryUpdate.GetID())) == null )
		{
			numberOfSupportedQueries++;		
			MotionStatisticsGatherer.totalNumberQueriesEnteredIntoHashTable++;
		}
	    
		MotionStatisticsGatherer.totalNumberQueriesLocationUpdates++;
		queriesHash.put(new Integer(movingQueryUpdate.GetID()), movingQueryUpdate);

		boolean IsOverlappingWithMultipleCells = false;
		//probe for a cluster
		MotionSCUBAHelper.clusterMovingQuery(clusterGrid, clusterStorage, clusterHome, movingQueryUpdate, currentTime, delta, motionSCUBAOperatorStatistics,IsOverlappingWithMultipleCells,LOAD_SHEDDING_FACTOR);	
	
		if (IsOverlappingWithMultipleCells)
		{
			this.motionSCUBAOperatorStatistics.numberOfClustersOverlappingWithMoreThanOneGridCell++;
		}
	
	}
	
	private void executeMotionSCUBAOperator() 
	{		
		XATQueue oQueue = this.getOutputQueue();
		XATQueue linputQueue = this.getLeftInputQueue();
		XATQueue rinputQueue = this.getRightInputQueue();
		XATTuple[] tuplesMovingObjects = null;
		XATTuple[] tuplesMovingQueries = null;		
		
		XATTuple[] tuplesMovingObjectsToBeProcessed = null;
		XATTuple[] tuplesMovingQueriesToBeProcessed = null;	
		int lHowmany; 
		int rHowmany;		
		this.currentTime = new Time();
		
		currentTime.SetTime(MotionGenericOperator.executionIterationCount * MotionGenericOperator.delta);  
		motionSCUBAOperatorStatistics.intervalTimeStart = (int)currentTime.GetTime();
		motionSCUBAOperatorStatistics.intervalTimeEnd = (int)(currentTime.GetTime()+MotionGenericOperator.delta);
		motionSCUBAOperatorStatistics.numberOfClustersAtTheBeginningOfExecution = this.clusterStorage.storage.entrySet().size();
		
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Current time = " + currentTime.GetTime());
			Helper.DebugWriteLine("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");
			Helper.DebugWriteLine("Starting Location Updates Gathering Phase...");
			Helper.DebugWriteLine("Updating locations of moving objects...");
		}

		//System.out.println("Current time = " + currentTime.GetTime());
		//System.out.println("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");

		
		lHowmany = linputQueue.getQueueSize();	
		rHowmany = rinputQueue.getQueueSize();
		
		if (lHowmany != 0)
		{
			//dequeue objects tuples
			tuplesMovingObjects = linputQueue.dequeue(lHowmany, this);			
		}
		
		if (rHowmany != 0)
		{
			//dequeue querie tuples
			tuplesMovingQueries = rinputQueue.dequeue(rHowmany, this);
		}
		
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("\nSize of left queue = " + lHowmany);
			Helper.DebugWriteLine("Size of right queue = " + rHowmany+"\n");					
		}
		
		int veryFirstObjectUpdateTimestamp = -1;
		int veryFirstQueryUpdateTimestamp = -1;
		int veryFirstMostProgressiveTimestamp = -1;
		
		//adjust the starting timestamp (because the stream generator doesn't start streaming right from the beginning)
		//pick the most "progressive" (i.e., latest timestamp from either of the streams)
		if (isTheFirstIteration == true)
		{	
			
			//statistics header
			Helper.StatsWriteLine(motionSCUBAOperatorStatistics.ShortSCUBAStatisticsHeaderToString());
			
			
			if (tuplesMovingObjects.length > 0)
			{
				veryFirstObjectUpdateTimestamp = Integer.parseInt(tuplesMovingObjects[0].getCell(4).toString());
			}
			if (tuplesMovingQueries.length > 0)
			{
				veryFirstQueryUpdateTimestamp  = Integer.parseInt(tuplesMovingQueries[0].getCell(4).toString());
			}
			
			if (veryFirstObjectUpdateTimestamp >= veryFirstQueryUpdateTimestamp)
			{
				veryFirstMostProgressiveTimestamp = veryFirstObjectUpdateTimestamp;
			}
			else
			{
				veryFirstMostProgressiveTimestamp = veryFirstQueryUpdateTimestamp;
			}
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("The very first timestamp from objects stream = " + veryFirstObjectUpdateTimestamp);
				Helper.DebugWriteLine("The very first timestamp from queries stream = " + veryFirstQueryUpdateTimestamp);				
			}
			
			//System.out.println("The very first timestamp from objects stream = " + veryFirstObjectUpdateTimestamp);
			//System.out.println("The very first timestamp from queries stream = " + veryFirstQueryUpdateTimestamp);

			MotionGenericOperator.executionIterationCount = (int)(veryFirstMostProgressiveTimestamp/MotionGenericOperator.delta);
			currentTime.SetTime(MotionGenericOperator.executionIterationCount * MotionGenericOperator.delta);  					
			isTheFirstIteration = false;
			motionSCUBAOperatorStatistics.intervalTimeStart = (int)currentTime.GetTime();
			motionSCUBAOperatorStatistics.intervalTimeEnd = (int)(currentTime.GetTime()+MotionGenericOperator.delta);
			
						
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nAdjusting Time Interval...");
				Helper.DebugWriteLine("Current execution count = " + MotionGenericOperator.executionIterationCount);
				Helper.DebugWriteLine("Current time = " + currentTime.GetTime());
				Helper.DebugWriteLine("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");
			}
		}
				
		
		//statistics		
		motionSCUBAOperatorStatistics.timeObjectsUpdateStart = System.currentTimeMillis();			
		
		//*** LOCATION UPDATE PHASE ***//		
		if(lHowmany != 0) 
		{			
			//get only the tuples that belong to the CURRENT time interval, the rest put back to the queue
			ArrayList tuplesToKeep = new ArrayList();
			//ArrayList tuplesToReturn = new ArrayList();
			ArrayList tuplesToDrop = new ArrayList();
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("First tuple on the queue (MO): " + tuplesMovingObjects[0].getCell(0)+" | "+ tuplesMovingObjects[0].getCell(1)+" | "+ tuplesMovingObjects[0].getCell(2)+" | "+ tuplesMovingObjects[0].getCell(3)+" | "+ tuplesMovingObjects[0].getCell(4)+" | "+ tuplesMovingObjects[0].getCell(5)+" | "+ tuplesMovingObjects[0].getCell(6)+" | "+ tuplesMovingObjects[0].getCell(7)+" | "+ tuplesMovingObjects[0].getCell(8));
				Helper.DebugWriteLine("Last tuple on the queue (MO): " + tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(0)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(1)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(2)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(3)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(4)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(5)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(6)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(7)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(8));
				Helper.DebugWriteLine(" *** ");
			}
			
			//System.out.println("First tuple on the queue (MO): " + tuplesMovingObjects[0].getCell(0)+" | "+ tuplesMovingObjects[0].getCell(1)+" | "+ tuplesMovingObjects[0].getCell(2)+" | "+ tuplesMovingObjects[0].getCell(3)+" | "+ tuplesMovingObjects[0].getCell(4)+" | "+ tuplesMovingObjects[0].getCell(5)+" | "+ tuplesMovingObjects[0].getCell(6)+" | "+ tuplesMovingObjects[0].getCell(7)+" | "+ tuplesMovingObjects[0].getCell(8));
			//System.out.println("Last tuple on the queue (MO): " + tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(0)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(1)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(2)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(3)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(4)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(5)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(6)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(7)+" | "+ tuplesMovingObjects[tuplesMovingObjects.length-1].getCell(8));

			
			
			for(int i=0; i<tuplesMovingObjects.length; i++)
			{
				//adjust the starting timestamp (because the stream generator doesn't start right from the beginning)
				if (isTheFirstIteration == true)
				{
					MotionGenericOperator.executionIterationCount = (int)(Integer.parseInt(tuplesMovingObjects[i].getCell(4).toString())/MotionGenericOperator.delta);
					currentTime.SetTime(MotionGenericOperator.executionIterationCount * MotionGenericOperator.delta);  					
					isTheFirstIteration = false;
					
					if(DEBUG == 1)
					{
						Helper.DebugWriteLine("\nAdjusting Time Interval...");
						Helper.DebugWriteLine("Current execution count = " + MotionGenericOperator.executionIterationCount);
						Helper.DebugWriteLine("Current time = " + currentTime.GetTime());
						Helper.DebugWriteLine("Current active time interval is between [" + currentTime.GetTime()+" - "+(currentTime.GetTime()+MotionGenericOperator.delta)+"]");
					}
				}
				

				if ((Long.parseLong(tuplesMovingObjects[i].getCell(4).toString()) >= currentTime.GetTime()) &&
					(Long.parseLong(tuplesMovingObjects[i].getCell(4).toString()) < (currentTime.GetTime() + MotionGenericOperator.delta)))
				{
					//keep the tuple
					tuplesToKeep.add(tuplesMovingObjects[i]);					
				}
				else if (Long.parseLong(tuplesMovingObjects[i].getCell(4).toString()) < currentTime.GetTime())
				{
					//drop the tuples (they arrived too late)
					tuplesToDrop.add(tuplesMovingObjects[i]);					
				}
				else
				{
					//return the tuple back to the queue
					//tuplesToReturn.add(tuplesMovingObjects[i]);
					
					if(futureTuplesMovingObjectsIndex.size() == 0)
					{				
						//add a new timestamp as an index
						futureTuplesMovingObjectsIndex.add(new FutureTuplesTimeStampIndex(Integer.parseInt(tuplesMovingObjects[i].getCell(4).toString()), futureTuplesMovingObjects.size()));
					}
					else 
					{
						FutureTuplesTimeStampIndex ft = ((FutureTuplesTimeStampIndex)(futureTuplesMovingObjectsIndex.get(futureTuplesMovingObjectsIndex.size() - 1)));
						if (Integer.parseInt(tuplesMovingObjects[i].getCell(4).toString()) > ft.timestamp)
						{
							//add a new timestamp as an index
							futureTuplesMovingObjectsIndex.add(new FutureTuplesTimeStampIndex(Integer.parseInt(tuplesMovingObjects[i].getCell(4).toString()), futureTuplesMovingObjects.size()));
						}						
					}					
					futureTuplesMovingObjects.add(tuplesMovingObjects[i]);
				}
			}
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nMoving Objects:");
				Helper.DebugWriteLine("---------------------------");
				Helper.DebugWriteLine("# of tuples to keep = ["+ tuplesToKeep.size()+"]");
				Helper.DebugWriteLine("# of tuples to (return) - future tuples = ["+ futureTuplesMovingObjects.size()+"]");
				Helper.DebugWriteLine("# of time indicies for future tuples = ["+ futureTuplesMovingObjectsIndex.size()+"]");
				Helper.DebugWriteLine("# of tuples to drop = ["+ tuplesToDrop.size()+"]");
			}
			
			motionSCUBAOperatorStatistics.tuplesObjectsDequeued = lHowmany;
			motionSCUBAOperatorStatistics.tuplesObjectsKept = tuplesToKeep.size();
			motionSCUBAOperatorStatistics.tuplesObjectsReturned = futureTuplesMovingObjectsIndex.size();
			motionSCUBAOperatorStatistics.tuplesObjectsDropped = tuplesToDrop.size();			
			
			if (futureTuplesMovingObjectsIndex.size() > 0)
			{
				FutureTuplesTimeStampIndex ft = (FutureTuplesTimeStampIndex)futureTuplesMovingObjectsIndex.get(futureTuplesMovingObjectsIndex.size()-1);
			
				//if the last timestamp in the future tuples arraylist is greater 
				//than the current delta time interval, push the execution
				if (ft.timestamp >= (currentTime.GetTime() + MotionGenericOperator.delta))
				{
					//set the flag to increment evaluation count (need to move on to the next time interval
					pushTimeIntervalForwardMovingObjects = true;
					if(DEBUG == 1)
					{
						Helper.DebugWriteLine("\n!!! SETTING PUSH FORWARD FOR MOVING OBJECTS TO TRUE !!!\n");
					}
				}
			}

			//drop the tuples
			tuplesToDrop = null;	
						
			//return the tuple back to the queue
			//linputQueue.enqueue(tuplesToReturn, this);
			
			
			tuplesMovingObjectsToBeProcessed = new XATTuple[tuplesToKeep.size()];
			tuplesMovingObjectsToBeProcessed = (XATTuple[])tuplesToKeep.toArray(tuplesMovingObjectsToBeProcessed);
			tuplesToKeep = null;
			
			//there are new object updates that belong to this time interval
			if(tuplesMovingObjectsToBeProcessed.length > 0)
			{			
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("# of moving objects' updates = ["+ tuplesMovingObjectsToBeProcessed.length +"]");
				}
							
				for (int i=0; i < tuplesMovingObjectsToBeProcessed.length; i++)
				{						
					MotionSCUBAMovingObjectState movingObjectUpdate = new MotionSCUBAMovingObjectState(
							Integer.parseInt(tuplesMovingObjectsToBeProcessed[i].getCell(1).toString()),
							EntityType.Object,
							Long.parseLong(tuplesMovingObjectsToBeProcessed[i].getCell(5).toString()),
							Long.parseLong(tuplesMovingObjectsToBeProcessed[i].getCell(6).toString()),
							Integer.parseInt(tuplesMovingObjectsToBeProcessed[i].getCell(4).toString()),
							UpdateType.FromStringToEnum(tuplesMovingObjectsToBeProcessed[i].getCell(0).toString()),
							ClassType.FromIntToEnum(Integer.parseInt(tuplesMovingObjectsToBeProcessed[i].getCell(3).toString())),
							Long.parseLong(tuplesMovingObjectsToBeProcessed[i].getCell(8).toString()),
							Long.parseLong(tuplesMovingObjectsToBeProcessed[i].getCell(9).toString()),
							(int)Float.parseFloat(tuplesMovingObjectsToBeProcessed[i].getCell(7).toString()));

					
						clusterUpdateMovingObject(currentTime, 
												  MotionGenericOperator.delta, 
												  movingObjectUpdate);
				}
			}
					
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Location update phase for moving objects is done.");			
			}
		}
		
		//statistics
		motionSCUBAOperatorStatistics.timeObjectsUpdateEnd = System.currentTimeMillis();
		MotionStatisticsGatherer.totalTimeElapsedObjectsUpdate += (motionSCUBAOperatorStatistics.timeObjectsUpdateEnd - motionSCUBAOperatorStatistics.timeObjectsUpdateStart);
		
		motionSCUBAOperatorStatistics.timeQueriesUpdateStart = System.currentTimeMillis();
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Updating locations of moving queries...");
		}
			
		//moving queries		
		if(rHowmany != 0) 
		{			
			//get only the tuples that belong to the CURRENT time interval, the rest put back to the queue
			ArrayList tuplesToKeep = new ArrayList();
			//ArrayList tuplesToReturn = new ArrayList();
			ArrayList tuplesToDrop = new ArrayList();
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Moving Query Timestamp = " + Long.parseLong(tuplesMovingQueries[0].getCell(4).toString()));				
				Helper.DebugWriteLine("First tuple on the queue (MQ): " + tuplesMovingQueries[0].getCell(0)+" | "+ tuplesMovingQueries[0].getCell(1)+" | "+ tuplesMovingQueries[0].getCell(2)+" | "+ tuplesMovingQueries[0].getCell(3)+" | "+ tuplesMovingQueries[0].getCell(4)+" | "+ tuplesMovingQueries[0].getCell(5)+" | "+ tuplesMovingQueries[0].getCell(6)+" | "+ tuplesMovingQueries[0].getCell(7)+" | "+ tuplesMovingQueries[0].getCell(8));
				Helper.DebugWriteLine("Last tuple on the queue (MQ): " + tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(0)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(1)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(2)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(3)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(4)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(5)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(6)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(7)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(8));
				Helper.DebugWriteLine(" *** ");
			}
			
			//System.out.println("First tuple on the queue (MQ): " + tuplesMovingQueries[0].getCell(0)+" | "+ tuplesMovingQueries[0].getCell(1)+" | "+ tuplesMovingQueries[0].getCell(2)+" | "+ tuplesMovingQueries[0].getCell(3)+" | "+ tuplesMovingQueries[0].getCell(4)+" | "+ tuplesMovingQueries[0].getCell(5)+" | "+ tuplesMovingQueries[0].getCell(6)+" | "+ tuplesMovingQueries[0].getCell(7)+" | "+ tuplesMovingQueries[0].getCell(8));
			//System.out.println("Last tuple on the queue (MQ): " + tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(0)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(1)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(2)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(3)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(4)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(5)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(6)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(7)+" | "+ tuplesMovingQueries[tuplesMovingQueries.length-1].getCell(8));

			
			for(int i=0; i<tuplesMovingQueries.length; i++)
			{
				if ((Long.parseLong(tuplesMovingQueries[i].getCell(4).toString()) >= currentTime.GetTime()) &&
				    (Long.parseLong(tuplesMovingQueries[i].getCell(4).toString()) < currentTime.GetTime() + MotionGenericOperator.delta))
				{
					//keep the tuple
					tuplesToKeep.add(tuplesMovingQueries[i]);					
				}
				else if (Long.parseLong(tuplesMovingQueries[i].getCell(4).toString()) < currentTime.GetTime())
				{
					//drop the tuples (they arrived too late)
					tuplesToDrop.add(tuplesMovingQueries[i]);					
				}			
				else
				{
					if(futureTuplesMovingQueriesIndex.size() == 0)
					{				
						//add a new timestamp as an index
						futureTuplesMovingQueriesIndex.add(new FutureTuplesTimeStampIndex(Integer.parseInt(tuplesMovingQueries[i].getCell(4).toString()), futureTuplesMovingQueries.size()));
					}
					else 
					{
						FutureTuplesTimeStampIndex ft = ((FutureTuplesTimeStampIndex)(futureTuplesMovingQueriesIndex.get(futureTuplesMovingQueriesIndex.size() - 1)));
						if (Integer.parseInt(tuplesMovingQueries[i].getCell(4).toString()) > ft.timestamp)
						{
							//add a new timestamp as an index
							futureTuplesMovingQueriesIndex.add(new FutureTuplesTimeStampIndex(Integer.parseInt(tuplesMovingQueries[i].getCell(4).toString()), futureTuplesMovingQueries.size()));
						}						
					}
					
					futureTuplesMovingQueries.add(tuplesMovingQueries[i]);				
				}				
			}
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nMoving Queries:");
				Helper.DebugWriteLine("---------------------------");
				Helper.DebugWriteLine("# of tuples to keep = ["+ tuplesToKeep.size()+"]");
				Helper.DebugWriteLine("# of tuples to (return) - future tuples = ["+ futureTuplesMovingQueries.size()+"]");
				Helper.DebugWriteLine("# of time indicies for future tuples = ["+ futureTuplesMovingQueriesIndex.size()+"]");
				Helper.DebugWriteLine("# of tuples to drop = ["+ tuplesToDrop.size()+"]");
			}

			//statistics
			motionSCUBAOperatorStatistics.tuplesQueriesDequeued = rHowmany;
			motionSCUBAOperatorStatistics.tuplesQueriesKept = tuplesToKeep.size();
			motionSCUBAOperatorStatistics.tuplesQueriesReturned = futureTuplesMovingQueriesIndex.size();
			motionSCUBAOperatorStatistics.tuplesQueriesDropped = tuplesToDrop.size();
			
			
			if (futureTuplesMovingQueriesIndex.size() > 0)
			{
				FutureTuplesTimeStampIndex ft = (FutureTuplesTimeStampIndex)futureTuplesMovingQueriesIndex.get(futureTuplesMovingQueriesIndex.size()-1);
			
				//if the last timestamp in the future tuples arraylist is greater 
				//than the current delta time interval, push the execution
				if (ft.timestamp >= (currentTime.GetTime() + MotionGenericOperator.delta))
				{
					//set the flag to increment evaluation count (need to move on to the next time interval
					pushTimeIntervalForwardMovingQueries = true;
					if(DEBUG == 1)
					{
						Helper.DebugWriteLine("\n!!! SETTING PUSH FORWARD FOR MOVING QUERIES TO TRUE !!!\n");
					}
				}
			}
			
			
			//drop the tuples
			tuplesToDrop = null;
			
			//put tuples back on the queue
			//rinputQueue.enqueue(tuplesToReturn, this);
			
			
			tuplesMovingQueriesToBeProcessed = new XATTuple[tuplesToKeep.size()];
			tuplesMovingQueriesToBeProcessed = (XATTuple[])tuplesToKeep.toArray(tuplesMovingQueriesToBeProcessed);
			tuplesToKeep = null;
			
			//there are new object updates that belong to this time interval
			if(tuplesMovingQueriesToBeProcessed.length > 0)
			{								
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("# of moving queries' updates = ["+ tuplesMovingQueriesToBeProcessed.length +"]");
				}
				
				for (int i=0; i < tuplesMovingQueriesToBeProcessed.length; i++)
				{					
					MotionSCUBAMovingQueryState movingQueryUpdate = new MotionSCUBAMovingQueryState(
							Integer.parseInt(tuplesMovingQueriesToBeProcessed[i].getCell(1).toString()),
							EntityType.Query,
							Long.parseLong(tuplesMovingQueriesToBeProcessed[i].getCell(5).toString()),
							Long.parseLong(tuplesMovingQueriesToBeProcessed[i].getCell(6).toString()),
							Integer.parseInt(tuplesMovingQueriesToBeProcessed[i].getCell(4).toString()),
							UpdateType.FromStringToEnum(tuplesMovingQueriesToBeProcessed[i].getCell(0).toString()),
							ClassType.FromIntToEnum(Integer.parseInt(tuplesMovingQueriesToBeProcessed[i].getCell(3).toString())),
							QueryType.Range,
							Integer.parseInt(tuplesMovingQueriesToBeProcessed[i].getCell(10).toString()),
							Long.parseLong(tuplesMovingQueriesToBeProcessed[i].getCell(8).toString()),
							Long.parseLong(tuplesMovingQueriesToBeProcessed[i].getCell(9).toString()),
							(int)Float.parseFloat(tuplesMovingQueriesToBeProcessed[i].getCell(7).toString()));					
					
					clusterUpdateMovingQuery(currentTime, 
										     MotionGenericOperator.delta, 
											 movingQueryUpdate);
				}
			}	
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Location update for moving queries is done.");
			}	
			
		}
		
		motionSCUBAOperatorStatistics.timeQueriesUpdateEnd = System.currentTimeMillis();
		//*** LOCATION UPDATE PHASE ***		
		MotionStatisticsGatherer.totalTimeElapsedQueriesUpdate += (motionSCUBAOperatorStatistics.timeQueriesUpdateEnd - motionSCUBAOperatorStatistics.timeQueriesUpdateStart);
		
		ArrayList miniResults = new ArrayList();
		
		//*** JOINING PHASE ***		
		if ((pushTimeIntervalForwardMovingObjects == true) && (pushTimeIntervalForwardMovingQueries == true))
		{	

			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nProcessing moving objects from the 'FAKE QUEUE'...");
				Helper.DebugWriteLine("Total entries on the objects 'FAKE QUEUE' = " + futureTuplesMovingObjects.size());
			}	
				
			motionSCUBAOperatorStatistics.timeFakeQObjectsStart = System.currentTimeMillis();
			
			if (futureTuplesMovingObjects.size() > 0)
			{
				int counter = 0;
				XATTuple tupleToBeProcessed = (XATTuple)futureTuplesMovingObjects.get(0);
				while (Float.parseFloat(tupleToBeProcessed.getCell(4).toString()) < (currentTime.GetTime() + MotionGenericOperator.delta))
				{	
					counter++;
					
					MotionSCUBAMovingObjectState movingObjectUpdate = new MotionSCUBAMovingObjectState(
							Integer.parseInt(tupleToBeProcessed.getCell(1).toString()),
							EntityType.Object,
							Long.parseLong(tupleToBeProcessed.getCell(5).toString()),
							Long.parseLong(tupleToBeProcessed.getCell(6).toString()),
							Integer.parseInt(tupleToBeProcessed.getCell(4).toString()),
							UpdateType.FromStringToEnum(tupleToBeProcessed.getCell(0).toString()),
							ClassType.FromIntToEnum(Integer.parseInt(tupleToBeProcessed.getCell(3).toString())),
							Long.parseLong(tupleToBeProcessed.getCell(8).toString()),
							Long.parseLong(tupleToBeProcessed.getCell(9).toString()),
							(int)Float.parseFloat(tupleToBeProcessed.getCell(7).toString()));

				
					clusterUpdateMovingObject(currentTime, 
							  				  MotionGenericOperator.delta, 
											  movingObjectUpdate);	
					
					futureTuplesMovingObjects.remove(0);
			
					if (futureTuplesMovingObjects.size() == 0)
					{
						//fake queue is empty
						break;
					}
					tupleToBeProcessed = (XATTuple)futureTuplesMovingObjects.get(0);												
				}//end while
				
				motionSCUBAOperatorStatistics.timeFakeQObjectsEnd = System.currentTimeMillis();
				
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Dequeued total "+counter+" objects updates from the 'FAKE QUEUE'...");
					Helper.DebugWriteLine("Now total entries on the objects 'FAKE QUEUE' = " + futureTuplesMovingObjects.size());
				}
			}//end if
			
			
			if(DEBUG == 1)
			{				
				Helper.DebugWriteLine("Done discrete updates for tuples from the 'fake' queue");				
				Helper.DebugWriteLine("Done removing tuples from the 'fake' queue");
				Helper.DebugWriteLine("\nProcessing moving queries from the 'fake queue'...");
			}	
			
			motionSCUBAOperatorStatistics.timeFakeQQueriesStart = System.currentTimeMillis();
			if (futureTuplesMovingQueries.size() > 0)
			{
				int counter = 0;
				XATTuple tupleToBeProcessed = (XATTuple)futureTuplesMovingQueries.get(0);
				while (Float.parseFloat(tupleToBeProcessed.getCell(4).toString()) < (currentTime.GetTime() + MotionGenericOperator.delta))
				{															
					counter++;

					MotionSCUBAMovingQueryState movingQueryUpdate = new MotionSCUBAMovingQueryState(
							Integer.parseInt(tupleToBeProcessed.getCell(1).toString()),
							EntityType.Query,
							Long.parseLong(tupleToBeProcessed.getCell(5).toString()),
							Long.parseLong(tupleToBeProcessed.getCell(6).toString()),
							Integer.parseInt(tupleToBeProcessed.getCell(4).toString()),
							UpdateType.FromStringToEnum(tupleToBeProcessed.getCell(0).toString()),
							ClassType.FromIntToEnum(Integer.parseInt(tupleToBeProcessed.getCell(3).toString())),
							QueryType.Range,
							Integer.parseInt(tupleToBeProcessed.getCell(10).toString()),
							Long.parseLong(tupleToBeProcessed.getCell(8).toString()),
							Long.parseLong(tupleToBeProcessed.getCell(9).toString()),
							(int)Float.parseFloat(tupleToBeProcessed.getCell(7).toString()));					
					
					clusterUpdateMovingQuery(currentTime, 
											 MotionGenericOperator.delta, 
											 movingQueryUpdate);	
					
					
						futureTuplesMovingQueries.remove(0);
						if (futureTuplesMovingQueries.size() == 0)
						{
							//fake queue is empty
							break;
						}
					tupleToBeProcessed = (XATTuple)futureTuplesMovingQueries.get(0);												
				}//end while
				
				motionSCUBAOperatorStatistics.timeFakeQQueriesEnd = System.currentTimeMillis();
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Dequeued total "+counter+" queries updates from the 'FAKE QUEUE'...");	
					Helper.DebugWriteLine("Now total entries on the objects 'FAKE QUEUE' = " + futureTuplesMovingObjects.size());
				}
			}//end if
				
			
			//PRINT OUT ALL CLUSTERS (FOR DEBUGGING PURPOSES ONLY) - REMOVE THIS LATER
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine(clusterStorage.toString());
			}
			
						
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("\nSTARTING JOIN PHASE...\n");
			}	
		
			//System.out.println("STARTING JOIN...");
			//statistics
			motionSCUBAOperatorStatistics.timeJoinStart = System.currentTimeMillis();
			
			miniResults = SCUBAJoiningPhase(currentTime, MotionGenericOperator.delta);	
			
			//statistics
			motionSCUBAOperatorStatistics.timeJoinEnd = System.currentTimeMillis();
			//System.out.println("DONE WITH JOIN... total results = [" + miniResults.size()+"]");
			motionSCUBAOperatorStatistics.numberOfIntersectionResults = miniResults.size();
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Joining Phase is done");
				Helper.DebugWriteLine("# of intersection results = ["+miniResults.size()+"]");				
			}
			
			//*** END JOINING PHASE ***		
			
			/*
			if (miniResults.GetDiscreteResults().size() > 0)
			{		
				System.out.println("Sending results...");
				int numberOfResults = miniResults.GetDiscreteResults().size();
				//tuples to be sent to the output queue
				XATTupleImp[] outputTuples = new XATTupleImp[numberOfResults];
				//value of one tuple
				XATDataValueImp[] outputValue = new XATDataValueImp[1];
				//array of results
			
				for(int i = 0; i < numberOfResults ; i++)
				{		
					MotionDiscreteResult outputResult = (MotionDiscreteResult)miniResults.GetDiscreteResults().get(i);
					outputValue[0] = new XATDataValueImp(outputResult);
					outputTuples[i] = new XATTupleImp(outputValue);
				}
			
				//send the results to the output queue
				this.getOutputQueue().enqueue(outputTuples, this);
							
				if(DEBUG == 1)
				{
					Helper.DebugWriteLine("Clearing the results");					
				}
				System.out.println("Done sending results...");								
			}
			*/
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Clearing the grid");					
			}
			
			//cluster grid cleanup
			clusterGrid.Cleanup();
			
			//Helper.DebugWriteLine(clusterGrid.statisticsToString());
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Do cluster maintenance");					
			}
			
			if (NO_DUPLICATES == 1)
			{
				//clean up join recorder
				clusterJoinRecorder.clear();
			}
			
			motionSCUBAOperatorStatistics.timeClusterMaintenanceStart = System.currentTimeMillis();
			
			//do some cluster maintenance
			clusterMaintenance();
			
			motionSCUBAOperatorStatistics.timeClusterMaintenanceEnd = System.currentTimeMillis();
			
			if(DEBUG == 1)
			{
				Helper.DebugWriteLine("Clearing Objects and Queries tables");					
			}
			//cleanup the Objects and Queries hash tables
			cleanupObjectsAndQueriesTables();
			
			if(DEBUG == 1)
			{
				//print grid statistics		
				Helper.DebugWriteLine(clusterGrid.toString());
			}
			

			motionSCUBAOperatorStatistics.numberOfObjectsInTheSystem = this.objectsHash.keySet().size();
			motionSCUBAOperatorStatistics.numberOfQueriesInTheSystem = this.queriesHash.keySet().size();
		
		}//end pushTimeIntervalForward

		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("# of objects in the Objects table: "+this.objectsHash.keySet().size());
			Helper.DebugWriteLine("# of queries in the Queries table: "+this.queriesHash.keySet().size());
		}
		
		//statistics (ended actual execution)
		motionSCUBAOperatorStatistics.timeOverallEnd = System.currentTimeMillis();
		
		//System.out.println("DONE WITH CURRENT INTERVAL EXECUTION...");
		
		//print the results (for accounting purposes)
		//flag to write out results is turned on
		if (RESULTS_FLAG == 1)
		{						
			if (miniResults.size() > 0)
			{
				ResultsRecorder.SCUBAResultsWriteLine(miniResults);
				//System.out.println("Done writing results to a file...");
			}
		}
		
		
		
		//clear results collection
		miniResults.clear();		
		
		motionSCUBAOperatorStatistics.numberOfClustersAtTheEndOfExecution = this.clusterStorage.storage.entrySet().size();

		
		//call for garbage collection
		System.gc();	
	 }
	
	
	private void cleanupObjectsAndQueriesTables()
	{
		//Objects hash table
		objectsHash.clear();
		
		
		//Queiries hash table
		queriesHash.clear();
		
		this.motionSCUBAOperatorStatistics.numberOfObjectsInTheSystem = 0;
		
	}
	
	
	private void DissolveCluster(MotionSCUBACluster clusterToDissolve)
	{
		for(int i=0; i<clusterToDissolve.objects.size(); i++)
		{
			MotionSCUBAClusterObjectMember o = (MotionSCUBAClusterObjectMember)clusterToDissolve.objects.get(i);			
			clusterHome.clusterHome.remove("o"+o.id);
		}
		for(int j=0; j<clusterToDissolve.queries.size(); j++)
		{
			MotionSCUBAClusterQueryMember q = (MotionSCUBAClusterQueryMember)clusterToDissolve.queries.get(j);
			clusterHome.clusterHome.remove("q"+q.id);
		}

		
		//clusterStorage.storage.remove(new Integer(clusterToDissolve.cid));		
	}
	
	private void clusterMaintenance()
	{
		//if(DEBUG == 1)
		//{
		//	Helper.DebugWriteLine("Time started cluster maintenance: " + System.currentTimeMillis());		
		//}
		
		
		//get rid of all the clusters that might dissolve
		Iterator i = this.clusterStorage.storage.entrySet().iterator();
		while(i.hasNext())
		{			
			Map.Entry  entry =  (Map.Entry)i.next();			
			if (entry != null)
			{
				MotionSCUBACluster currCluster = (MotionSCUBACluster)entry.getValue();
			
				//approximately calculate where the cluster will be at the next time unit
				//if it passes the destination node, dissolve it
			
				double distanceUntilDestination = Math.sqrt(Math.pow(currCluster.nextNodeX-currCluster.centerPoint.GetX(), 2)+Math.pow(currCluster.nextNodeY-currCluster.centerPoint.GetY(),2));						
			
				if ((distanceUntilDestination > 0) && (currCluster.memberCount > 0))
				{
					float slope = (float)(currCluster.speed/distanceUntilDestination);
					float xDestdiff = currCluster.nextNodeX - currCluster.centerPoint.GetX();
					float yDestdiff = currCluster.nextNodeY - currCluster.centerPoint.GetY();
			
					long expectedX = (int)(slope * xDestdiff + currCluster.centerPoint.GetX());
					long expectedY = (int)(slope * yDestdiff + currCluster.centerPoint.GetY());
						
					currCluster.centerPoint.SetX(expectedX);
					currCluster.centerPoint.SetY(expectedY);
					currCluster.centerPoint.SetT(currCluster.centerPoint.GetT()+1);
			
					//the cluster "passed" the network node, so dissolve it
					if ((currCluster.centerPoint.GetX() >= currCluster.nextNodeX) ||
					    (currCluster.centerPoint.GetY() >= currCluster.nextNodeY))
					{
						//dissolve the cluster						
						//if(DEBUG == 1)
						//{
						//	//print grid statistics		
						//	Helper.DebugWriteLine("Dissolving cluster: " + currCluster);
						//}	
					
						//NOTE TO MYSELF: no need to remove the cluster from the grid
						//assumption is that the grid has already been cleaned up
						//and the clusters that do stay, get re-flushed after adjusting
						//their positions are adjusted
					
						//remove cluster from the grid 
						//GridRemoveCluster(currCluster);
					
						//dissolve the cluster
						DissolveCluster(currCluster);
						i.remove();
						this.motionSCUBAOperatorStatistics.numberOfClustersDissolved++;
					}
					else
					{
						//flush the cluster to the grid cell
						GridInsertCluster(currCluster);				
					}
				}
				else
				{
					//dissolving the cluster (it's already at the destination node) or there
					//are no members that belong to this cluster
					//if(DEBUG == 1)
					//{
					//	//print grid statistics		
					//	Helper.DebugWriteLine("Dissolving cluster: " + currCluster);
					//}
				
					//see comment above
					//remove cluster from the grid
					//GridRemoveCluster(currCluster);
				
					DissolveCluster(currCluster);
					i.remove();
					this.motionSCUBAOperatorStatistics.numberOfClustersDissolved++;
				}
			}
		}
		
		
		//if(DEBUG == 1)
		//{
		//	Helper.DebugWriteLine("Time ended cluster maintenance: " + motionSCUBAOperatorStatistics.timeClusterMaintenanceEnd);		
		//}
	}
		

	private ArrayList SCUBAJoiningPhase(Time T, int delta)
	{
		int numJoinComparisons = 0;
		double deTimeStart; //duplicate elimination time start
		double deTimeTotal = 0;   //duplicate elimination time end
		
		ArrayList miniResults = new ArrayList();

						
		//Helper.DebugWriteLine(clusterGrid.statisticsToString());
		
		//clear the mini-clipped results ArrayList
		miniResults.clear();					
		
		//Loop through grid cells in the x-dimension
		for (int i=0; i < MotionSCUBAOperator.MotionSCUBAClusterGridDimensionXCellCount; i++)
		{
			//Loop through grid cells in the y-dimension
 			for (int j=0; j < MotionSCUBAOperator.MotionSCUBAClusterGridDimensionYCellCount; j++)
			{
				//Loop through grid cells in the t-dimension
				for (int k=0; k < GenericSpatioTemporalGrid.GetTDimensionCellCount(); k++)
				{
					if (clusterGrid.directory[i][j][k] != null)
					{	
						if (clusterGrid.directory[i][j][k].GetClusters() != null)
						{
							Object[] arrClusters = (Object[])clusterGrid.directory[i][j][k].GetClusters().entrySet().toArray();							
							
							//For each moving query MQ that has an entry in the grid cell G i,j,k
							for (int q=0; q < arrClusters.length; q++)
							{															
								//For each moving object MO that has an entry in the grid cell Gi,j,k  
								for (int o=q; o < arrClusters.length; o++)
								{				
									Map.Entry  entry1 =  (Map.Entry)arrClusters[q];		
									Map.Entry  entry2 =  (Map.Entry)arrClusters[o];		
	
									SCUBAGridEntry clusterGridEntry1 = (SCUBAGridEntry)entry1.getValue();
									SCUBAGridEntry clusterGridEntry2 = (SCUBAGridEntry)entry2.getValue();
										
									if (NO_DUPLICATES == 1)
									{
										deTimeStart = System.currentTimeMillis();
										//check if need to be joined (this is done for duplicate elimination)
										//join ONLY if the clusters haven't already been joined
										if (!isAlreadyJoined(clusterGridEntry1.GetID(),clusterGridEntry2.GetID()))
										{
											deTimeTotal = deTimeTotal +(System.currentTimeMillis() - deTimeStart);
											//do between-cluster join
											numJoinComparisons++;
											if (clusterGrid.DoBetweenJoin(clusterGridEntry1, clusterGridEntry2, new GridCellIndex3D(i,j,k), clusterStorage))
											{
												//clusters overlap, so perform a cluster-within join
												ArrayList joinResults = clusterGrid.DoWithinJoin(clusterGridEntry1, clusterGridEntry2, new GridCellIndex3D(i,j,k), clusterStorage);
									
												if (joinResults != null)
												{
													miniResults.addAll(joinResults);
												}									
											}
										
											deTimeStart = System.currentTimeMillis();
											
											//record that already these clusters have been joined (again, done for duplicate elimination)
											recordJoinPair(clusterGridEntry1.GetID(),clusterGridEntry2.GetID());		
											
											deTimeTotal = deTimeTotal +(System.currentTimeMillis() - deTimeStart);
										}
									}
									else
									{
										//do regular between-cluster join (without duplicate elimination)
										if (clusterGrid.DoBetweenJoin(clusterGridEntry1, clusterGridEntry2, new GridCellIndex3D(i,j,k), clusterStorage))
										{
											//clusters overlap, so perform a cluster-within join
											ArrayList joinResults = clusterGrid.DoWithinJoin(clusterGridEntry1, clusterGridEntry2, new GridCellIndex3D(i,j,k), clusterStorage);
								
											if (joinResults != null)
											{
												miniResults.addAll(joinResults);
											}									
										}
									}
								}
							}
						}
					}
					
				}
			}
		}	
		
		if (NO_DUPLICATES == 1)
		{
			this.motionSCUBAOperatorStatistics.timeTotalDuplicateElimination = deTimeTotal;
		}
		
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Total # of join comparisons is "+ numJoinComparisons);		
		}
		
		return miniResults;
	}
	

	
	
	private void flushAllClustersToGrid()
	{		
		int totalClustersFlushed = 0;
	
		if(DEBUG == 1)
		{
			Helper.DebugWriteLine("Flushing CLUSTERS to grid...");
			Helper.DebugWriteLine("Total ["+clusterStorage.storage.keySet().size()+"] clusters to flush");
		}
		
		Iterator i = clusterStorage.storage.entrySet().iterator();
		while(i.hasNext())
		{			
			Map.Entry  entry =  (Map.Entry)i.next();			
			MotionSCUBACluster cluster =(MotionSCUBACluster)entry.getValue();
			
			//dissolve (and definately do not flush the cluster) that doesn't have any members
			if (cluster.memberCount == 0)
			{
				if(DEBUG == 1)
				{
					//print grid statistics		
					Helper.DebugWriteLine("Dissolving cluster: " + cluster);
				}
				
				//remove cluster from the grid
				GridRemoveCluster(cluster);
				//dissolve the grid
				DissolveCluster(cluster);
				
				clusterStorage.storage.remove(new Integer(cluster.cid));				
				this.motionSCUBAOperatorStatistics.numberOfClustersDissolved++;
			}
			else
			{
				GridInsertCluster(cluster);				
				totalClustersFlushed++;
			}
		}		
       this.motionSCUBAOperatorStatistics.numberOfClustersBeingJoined = totalClustersFlushed;

	}//end flushAllClustersToGrid

		
	private void GridInsertCluster(MotionSCUBACluster mc)
	{
		clusterGrid.InsertCluster(mc);								
	}
	
	private void GridRemoveCluster(MotionSCUBACluster mc)
	{
		clusterGrid.RemoveCluster(mc);								
	}
	
}
