/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on May 14, 2005
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.MotionSCUBAStatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.DimensionType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.FixedDiscrete3DGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GenericSpatioTemporalGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridCellIndex3D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridEntry;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingObjectState;

/**
 * @author Rimma V. Nehme
 *
 */
public class MotionSCUBAHelper 
{
		
	public static int clusterMovingObject(MotionSCUBAClusterGrid clusterGrid,
										  MotionSCUBAClusterStorage clusterStorage,
										  MotionSCUBAClusterHome clusterHome,
										  MotionSCUBAMovingObjectState mo, 
										  Time currentTime, 
										  int delta,
										  MotionSCUBAStatisticsGatherer scubaStatistics,
										  boolean IsOverlappingWithMultipleCells,
										  int loadSheddingFactor)	
	{
	
		boolean isClusterMember = false;
		int returnClusterID = -1;
		
		//probe to see if object previously belonged to any cluster
		Integer clusterID = (Integer)clusterHome.clusterHome.get("o" + mo.GetID());		
		
		if (clusterID != null)
		{
			MotionSCUBACluster parentCluster = (MotionSCUBACluster)clusterStorage.storage.get(clusterID);
			
			//check to see if object still belongs to the parent cluster
			if (parentCluster.IsClusterCandidate(mo))
			{
				boolean IsCenterMoved = false;
				boolean IsRadiusChanged = false;
				
				parentCluster.adjustOldClusterMemberObject(mo,IsCenterMoved,IsRadiusChanged);		
				scubaStatistics.numberOfObjectsStayedInParentCluster++;
				
		    	//re-adjust grid indicies for the cluster
				clusterGrid.adjustCluster(parentCluster,IsCenterMoved,IsRadiusChanged, IsOverlappingWithMultipleCells);	
								
				returnClusterID = parentCluster.cid; 
				return returnClusterID;
			}
			else
			{
				boolean isEmtpyCluster = false;
				//no longer belongs to this cluster
				//so remove it from the members list
				parentCluster.removeOldClusterMemberObject(mo, isEmtpyCluster);	
				scubaStatistics.numberOfObjectsLeftParentCluster++;
				clusterHome.clusterHome.remove("o" + mo.GetID());
				
				if (isEmtpyCluster)
				{
					clusterStorage.storage.remove(new Integer(parentCluster.cid));
				    clusterGrid.RemoveCluster(parentCluster);
					
				}
				
		    	//re-adjust grid indicies for the cluster
				//clusterGrid.AdjustCluster(parentCluster);	
			}
		}
		
		
		int xIndex = MotionSCUBAClusterGrid.GetCellIndex(mo.GetPosition().GetX(), DimensionType.x);		
		int yIndex = MotionSCUBAClusterGrid.GetCellIndex(mo.GetPosition().GetY(), DimensionType.y);
		int tIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)

		clusterGrid.CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex,tIndex);
		
		//get clusters that fall into the same grid cell as the current object update
		HashMap cellClusters = clusterGrid.GetClustersByCellIndex(xIndex,yIndex,tIndex);


		Iterator i = cellClusters.entrySet().iterator();
		while(i.hasNext())
		{						
			Map.Entry  entry =  (Map.Entry)i.next();			
			if (entry != null)
			{
				SCUBAGridEntry clusterGridEntry = (SCUBAGridEntry)entry.getValue();
			
				MotionSCUBACluster currCluster = clusterStorage.getClusterByCID(new Integer(clusterGridEntry.GetID()));
				if (currCluster.IsClusterCandidate(mo))
				{
					boolean IsCenterMoved = false;
					boolean IsRadiusChanged = false;
				
					currCluster.AdoptNewClusterMemberObject(mo, IsCenterMoved,IsRadiusChanged);				
					//save relationship in the ClusterHome table
					clusterHome.clusterHome.put("o"+mo.GetID(), new Integer(currCluster.cid));
					isClusterMember = true;
				
					//re-adjust grid indicies for the cluster
					clusterGrid.adjustCluster(currCluster, IsCenterMoved,IsRadiusChanged, IsOverlappingWithMultipleCells);
				
					returnClusterID = currCluster.cid; 
					break;
				}
			}
		}
		
		//not a member of any cluster, need to create a new cluster
		if (!isClusterMember)
		{
			//create new cluster
			MotionSCUBACluster newCluster = createNewClusterBasedOnMovingObject(mo);
			scubaStatistics.numberOfNewClustersCreated++;
			
			//save relationship in the ClusterHome table
			clusterHome.clusterHome.put("o"+mo.GetID(), new Integer(newCluster.cid));
			
			//add new cluster's ID to the grid cell
			SCUBAGridEntry clusterGridEntry = new SCUBAGridEntry(newCluster.cid);	
			clusterGrid.directory[xIndex][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);
			
			//save cluster in the Cluster Storage
			clusterStorage.addNewCluster(newCluster);
			returnClusterID = newCluster.cid; 
			
		}		
		return returnClusterID;
	}
	
	
	
	
	
	
	public static int clusterMovingQuery(MotionSCUBAClusterGrid clusterGrid,
			  								 MotionSCUBAClusterStorage clusterStorage,
			  								 MotionSCUBAClusterHome clusterHome,
											 MotionSCUBAMovingQueryState mq, 
											 Time currentTime, 
											 int delta,
											 MotionSCUBAStatisticsGatherer scubaStatistics,
											 boolean IsOverlappingWithMultipleCells,
											 int loadSheddingFactor)	
	{
		
		
		boolean isClusterMember = false;
		int returnClusterID = -1;
		
		//probe to see if query previously belonged to any cluster
		Integer clusterID = (Integer)clusterHome.clusterHome.get("q" + mq.GetID());		
		
		if (clusterID != null)
		{
			MotionSCUBACluster parentCluster = (MotionSCUBACluster)clusterStorage.storage.get(clusterID);			
			//check to see if query still belongs to the parent cluster
			if (parentCluster.IsClusterCandidate(mq))
			{
				boolean IsCenterMoved = false;
				boolean IsRadiusChanged = false;
				
				parentCluster.AdjustOldClusterMemberQuery(mq, IsCenterMoved, IsRadiusChanged);	
				scubaStatistics.numberOfQueriesStayedInParentCluster++;
				
		    	//re-adjust grid indicies for the cluster
				clusterGrid.adjustCluster(parentCluster, IsCenterMoved, IsRadiusChanged, IsOverlappingWithMultipleCells);					
				
				returnClusterID = parentCluster.cid; 
				return returnClusterID;
			}
			else
			{
				boolean isEmtpyCluster = false;
				//no longer belongs to this cluster
				//so remove it from the members list
				parentCluster.removeOldClusterMemberQuery(mq,isEmtpyCluster);
				scubaStatistics.numberOfQueriesLeftParentCluster++;
				clusterHome.clusterHome.remove("q" + mq.GetID());
				
				if (isEmtpyCluster)
				{
					clusterStorage.storage.remove(new Integer(parentCluster.cid));
				    clusterGrid.RemoveCluster(parentCluster);					
				}
				
		    	//re-adjust grid indicies for the cluster
				//clusterGrid.AdjustCluster(parentCluster);	
			}
		}
		
		int xIndex = MotionSCUBAClusterGrid.GetCellIndex(mq.GetPosition().GetX(), DimensionType.x);		
		int yIndex = MotionSCUBAClusterGrid.GetCellIndex(mq.GetPosition().GetY(), DimensionType.y);
		int tIndex = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)

		clusterGrid.CheckCreateGridCellByIndexAndSpaceRange(xIndex,yIndex,tIndex);

		//get clusters that fall into the same grid cell as the current query update
		HashMap cellClusters = clusterGrid.GetClustersByCellIndex(xIndex,yIndex,tIndex);
		

		Iterator i = cellClusters.entrySet().iterator();
		while(i.hasNext())
		{						
			Map.Entry  entry =  (Map.Entry)i.next();			
			if (entry != null)
			{
				SCUBAGridEntry clusterGridEntry = (SCUBAGridEntry)entry.getValue();				
				MotionSCUBACluster currCluster = clusterStorage.getClusterByCID(new Integer(clusterGridEntry.GetID()));
				if (currCluster.IsClusterCandidate(mq))
				{
					boolean IsCenterMoved = false;
					boolean IsRadiusChanged = false;
					
					currCluster.AdoptNewClusterMemberQuery(mq, IsCenterMoved, IsRadiusChanged);
					//save relationship in the ClusterHome table
					clusterHome.clusterHome.put("q"+mq.GetID(), new Integer(currCluster.cid));
					isClusterMember = true;
					
			    	//re-adjust grid indicies for the cluster
					clusterGrid.adjustCluster(currCluster, IsCenterMoved, IsRadiusChanged, IsOverlappingWithMultipleCells);	
					
					returnClusterID = currCluster.cid; 				
					break;
				}
			}
		}

		//not a member of any cluster, need to create a new cluster
		if (!isClusterMember)
		{
			//create new cluster
			MotionSCUBACluster newCluster = createNewClusterBasedOnMovingQuery(mq);	
			scubaStatistics.numberOfNewClustersCreated++;
			
			//save relationship in the ClusterHome table
			clusterHome.clusterHome.put("q"+mq.GetID(), new Integer(newCluster.cid));

			//add new cluster's ID to the grid cell
			SCUBAGridEntry clusterGridEntry = new SCUBAGridEntry(newCluster.cid);	
			clusterGrid.directory[xIndex][yIndex][tIndex].AddMovingClusterEntry(clusterGridEntry);

			//save cluster in the Cluster Storage
			clusterStorage.addNewCluster(newCluster);
			
			returnClusterID = newCluster.cid;
			
		}		
		return returnClusterID;
	}
	
	
	
	
    public static MotionSCUBACluster createNewClusterBasedOnMovingObject(MotionSCUBAMovingObjectState movingObjectUpdate)
    {
    	MotionSCUBACluster newCluster = new MotionSCUBACluster();
    	newCluster.centerPoint = movingObjectUpdate.GetPosition();
    	newCluster.nextNodeX = movingObjectUpdate.nextNodeX;
    	newCluster.nextNodeY = movingObjectUpdate.nextNodeY;
    	    	
    	//create polar coordinates for the current object
    	MotionSCUBAClusterObjectMember newClusterMember = new MotionSCUBAClusterObjectMember(movingObjectUpdate.GetID(), MotionSCUBAPolarCoordinate.TranslateFromCartesianToPolar(movingObjectUpdate.GetPosition()));
    	
    	newCluster.objects.add(newClusterMember);
    	newCluster.memberCount = 1; 
    	newCluster.speed = movingObjectUpdate.speed;
    	newCluster.radius = 0;
    	newCluster.furthestPosition = movingObjectUpdate.GetPosition();
    	
    	//grid index of the centeroid of the cluster
    	newCluster.currentGridIndexX = MotionSCUBAClusterGrid.GetCellIndex(movingObjectUpdate.GetPosition().GetX(), DimensionType.x);		
    	newCluster.currentGridIndexY = MotionSCUBAClusterGrid.GetCellIndex(movingObjectUpdate.GetPosition().GetY(), DimensionType.y);
    	newCluster.currentGridIndexT = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)
    	
    	newCluster.gridCellsOverlap.add(new GridCellIndex3D(newCluster.currentGridIndexX,newCluster.currentGridIndexY,newCluster.currentGridIndexT));
    	return newCluster;
    }
    
    public static MotionSCUBACluster createNewClusterBasedOnMovingQuery(MotionSCUBAMovingQueryState movingQueryUpdate)
    {
    	MotionSCUBACluster newCluster = new MotionSCUBACluster();
    	newCluster.centerPoint = movingQueryUpdate.GetPosition();
    	newCluster.nextNodeX = movingQueryUpdate.nextNodeX;
    	newCluster.nextNodeY = movingQueryUpdate.nextNodeY;
    	    	
    	//create polar coordinates for the current object
    	MotionSCUBAClusterQueryMember newClusterMember = new MotionSCUBAClusterQueryMember(movingQueryUpdate.GetID(), MotionSCUBAPolarCoordinate.TranslateFromCartesianToPolar(movingQueryUpdate.GetPosition()), movingQueryUpdate.GetQueryBoundary());
    	
    	newCluster.queries.add(newClusterMember);
    	newCluster.memberCount = 1;
    	newCluster.speed = movingQueryUpdate.speed;
    	newCluster.radius = 0;
    	newCluster.furthestPosition = movingQueryUpdate.GetPosition();
    	
    	//grid index of the centeroid of the cluster
    	newCluster.currentGridIndexX = MotionSCUBAClusterGrid.GetCellIndex(movingQueryUpdate.GetPosition().GetX(), DimensionType.x);		
    	newCluster.currentGridIndexY = MotionSCUBAClusterGrid.GetCellIndex(movingQueryUpdate.GetPosition().GetY(), DimensionType.y);
    	newCluster.currentGridIndexT = 0; //assume one cell in the temporal dimension (corresponding to the current time interval)
    	    
    	newCluster.gridCellsOverlap.add(new GridCellIndex3D(newCluster.currentGridIndexX,newCluster.currentGridIndexY,newCluster.currentGridIndexT));
    	return newCluster;
    }

}
