/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on May 14, 2005
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba;

import java.util.HashMap;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.FixedDiscrete3DGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.hash.DoubleKeyHash;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.MotionGenericOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * @author Rimma V. Nehme
 *
 */
public abstract class MotionSCUBAOperator extends MotionGenericOperator
{
	protected HashMap objectsHash;
	protected HashMap queriesHash;
	protected MotionSCUBAClusterHome clusterHome;
	protected MotionSCUBAClusterStorage clusterStorage;
	protected MotionSCUBAClusterGrid clusterGrid;
	public static int MotionSCUBAClusterGridDimensionXCellCount = 1000;
	public static int MotionSCUBAClusterGridDimensionYCellCount = 1000;
	public static int MotionSCUBAClusterGridDimensionTCellCount = 1;
	
	public static int MotionSCUBADistanceThreshold = 1000;
	public static int MotionSCUBASpeedThreshold = 500;
	
			
	protected Time currentTime;
	
	public MotionSCUBAOperator(XATNode node) 
	{
		super(node);	
		//initialize data structures
		
		objectsHash = new HashMap();
		queriesHash = new HashMap();
		clusterHome = new MotionSCUBAClusterHome();
		clusterStorage = new MotionSCUBAClusterStorage();
		clusterGrid = new MotionSCUBAClusterGrid(MotionSCUBAClusterGridDimensionXCellCount,
												 MotionSCUBAClusterGridDimensionYCellCount,
												 MotionSCUBAClusterGridDimensionTCellCount); // for now
		
		
		MotionGenericOperator.executionIterationCount = 1;
		MotionGenericOperator.delta = 1;			
	}
	
	public void initState(String statetype) 
	{		
		String statetypeMovingObjects = "edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.hash.DoubleKeyHash";
		String statetypeMovingQueries = "edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.hash.DoubleKeyHash";
		
		try
		{				
			states = new OperatorStateImp[2];
			states[0] = (DoubleKeyHash)Class.forName(statetypeMovingObjects).newInstance();
			states[1] = (DoubleKeyHash)Class.forName(statetypeMovingQueries).newInstance();			
			
			states[0].setOperator(this);
			states[1].setOperator(this);	
			
			this.objectsHash = (HashMap)this.states[0];		
			this.queriesHash = (HashMap)this.states[1];
			
			currentTime = new Time();
		}
		catch(Exception e) 
		{
			e.printStackTrace();	
		}		
	}
	
	public OperatorState getLeftState() 
	{
		return states[0];
	}
	
	public int getLeftStateSize() 
	{
		return states[0].getStateSize();
	}

	public OperatorState getRightState() 
	{
		return states[1];
	}
	
	public int getRightStateSize() 
	{
		return states[1].getStateSize();
	}

	public int getStateSize() 
	{
		return (states[0].getStateSize() + states[1].getStateSize());
	}
	
	/*
	 * recompute the unmatched states inside this operator.
	 * Usually the unmatched states are computed based on the matched states inside one of the 
	 * children operator of this operator.
	 * 
	 * return: 
	 * TRUE -- if the states (inside child operator) that the recomputation of the unmatched states are based on are 
	 *      all matched, and the recomputation is successful.
	 * FALSE -- if any of the states (inside child operator) that are used to recompute the unmatched states
	 *          is unmatched, or the recomputation is unsucessful.  
	 */
	public boolean recomputeUnmatchedState() 
	{
		boolean returnvalue = true;
		
		return returnvalue;	
	}

	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() 
	{
		if (states == null) 
		{
			return false;
		}

		for (int i = 0; i < states.length; i++) 
		{
			if (!states[i].isMatched())
				return true;
		}
		return false;
	}


}