/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on May 29, 2005
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;

/**
 * @author Rimma V. Nehme
  */
public class MotionSCUBAPolarCoordinate implements java.io.Serializable
{
	public float angle; //angle relative to the pole
	public float r;     //distance from the pole
	public float t;      //time (not a part of polar coordinate), just additional attribute

	
	public MotionSCUBAPolarCoordinate(float newAngle, float newR, float newT)
	{
		angle = newAngle;
		r = newR;		
		t = newT;
	}
	
	public static MotionSCUBAPolarCoordinate TranslateFromCartesianToPolar(PositionSpt2D cPos)
	{
		//r = ? (x2 + y2)
		float r = (float)Math.sqrt(Math.pow(cPos.GetX(), 2) + Math.pow(cPos.GetY(), 2));
		//? = atan( y / x )
		float angle = (float)Math.atan(cPos.GetY()/cPos.GetX());				
		return new MotionSCUBAPolarCoordinate(r, angle, cPos.GetT());
	}
	
	public static PositionSpt2D TranslateFromPolarToCartesian(MotionSCUBAPolarCoordinate pPos, PositionSpt2D centerPosition)
	{
		//x = r  cos(?)
		long x = (long)(pPos.r * Math.cos(pPos.angle));			
		//y = r  sin(?)
		long y = (long)(pPos.r * Math.sin(pPos.angle));	
				
		return new PositionSpt2D(centerPosition.GetX() + x,centerPosition.GetY() + y,pPos.t);
	}
	
	public String toString()
	{
		return "angle = "+this.angle+", r = " + this.r+", t = "+this.t;		
	}
}
