/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.QueryType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.ObjectFactory;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MotionDiscreteState;

/**
 * @author Rimma V. Nehme
 *
 */
public class MotionContinuousResult extends MotionGenericResult implements ObjectFactory
{
	protected MotionContinuousState value;
	
	public MotionContinuousResult(int newQueryID, int newObjectID, int newQueryType, MotionContinuousState newContinuousResult) 
	{
		super(newQueryID, newObjectID, newQueryType);
		this.value = newContinuousResult;
	}
	
	public MotionContinuousState GetResultValue()
	{
		return value;
	}
	

	public boolean equals(MotionContinuousResult resToCompare)
	{
		PositionSpt2D originalResult = this.GetResultValue().GetStartPosition();
		PositionSpt2D resultToCompare = resToCompare.GetResultValue().GetStartPosition();
		if ((originalResult.GetX() ==  resultToCompare.GetX()) &&
			(originalResult.GetY() ==  resultToCompare.GetY()) &&
			(originalResult.GetT() ==  resultToCompare.GetT()))
			{
				return true;
			}
		return false;
	}
	
	public String toString()
	{
		return "QueryID: "+this.queryID+"\nObjectID: "+this.objectID+"\nMotionContinuousResult():\n"+
				this.value.toString()+"\n";
	}


	public Object makeObject() 
	{
		//return a dummy object to approximate how much memory it will use up
		return new MotionContinuousResult(0, 0, QueryType.Range, new MotionContinuousState());
	}
}
