/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections;
import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.interval.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.*;

/**
 * @author Rimma V. Nehme
 *
 */
public class MovingObjectsIntervalStatesCollection extends MotionGenericIntervalState
{
	protected ArrayList objectIntervalUpdates; //MovingObjectState[]
	private PositionSpt2D intervalStartPosition;
	private PositionSpt2D intervalEndPosition;
	private MotionContinuousState latestTrajectory;

	
	public MotionContinuousState GetLatestTrajectory()
	{
		return latestTrajectory;
	}
	
	public MovingObjectsIntervalStatesCollection()
	{
		super(0, 0,new Time(0), new Time(0));
		objectIntervalUpdates = new ArrayList();
		intervalStartPosition = new PositionSpt2D();
		intervalEndPosition = new PositionSpt2D();
		latestTrajectory = null;
	}
	
	public MovingObjectsIntervalStatesCollection(int newID,
			  									 int newIntervalReportNumber,
												 Time newStartTime,
												 Time newEndTime)
	{
		super(newID, newIntervalReportNumber, newStartTime, newEndTime);
		objectIntervalUpdates = new ArrayList();
		intervalStartPosition = new PositionSpt2D();
		intervalEndPosition = new PositionSpt2D();
		latestTrajectory = null;
	}
	
	public void Cleanup(Time nextIntervalBeginTime, int delta)
	{
		//reset interval
		this.ResetInterval(nextIntervalBeginTime, delta);
		//
	}
	
	public void CleanupForBorderlinePositionPoints(Time nextIntervalBeginTime, int delta)
	{
		//reset interval
		this.ResetInterval(nextIntervalBeginTime, delta);
		//
	}
	
	public void ResetInterval(Time nextIntervalBeginTime, int delta)
	{
		//we reset everything, except for the latest trajectory
		this.ResetMotionGenericIntervalState(nextIntervalBeginTime, delta);
		objectIntervalUpdates.clear();
		intervalStartPosition = new PositionSpt2D();
		intervalEndPosition = new PositionSpt2D();
	}
	
	public boolean NewCleanup(Time nextIntervalBeginTime, int delta)
	{
		int id = -1; 
		int entityType = -1;
		int objectClass = -1; 
				 
		for (int i=0; i < objectIntervalUpdates.size();i++)
		{
			MovingObjectState mos = (MovingObjectState)objectIntervalUpdates.get(i);
			
			id = mos.GetID();
			entityType = mos.GetEntityType();
			objectClass = mos.GetObjectClass();
			
			if (mos.GetUpdateType() == UpdateType.DissappearPoint)
			{
				return true;
				
			}			
		}
		
		if (objectIntervalUpdates.size() > 0)
		{
			//if we got to here, that means all location points will not be used 
			//in the next time interval, so delete them
			objectIntervalUpdates.clear();
			this.startTime = nextIntervalBeginTime;
			this.endTime = new Time(nextIntervalBeginTime.GetTime() + delta);
			this.intervalReportNumber = 0;
			//this.intervalStartPosition
			//this.intervalEndPosition		
		
			//but now initialize the start and the end of the next time interval
			/*
			PositionSpt2D expPosStartNextTimeInterval = this.latestTrajectory.Calculate(nextIntervalBeginTime);
			PositionSpt2D expPosEndNextTimeInterval = this.latestTrajectory.Calculate(new Time(nextIntervalBeginTime.GetTime() + delta));
		
			MovingObjectState expStartStateNextTimeIntervalMovingObjectState = new MovingObjectState(id, entityType, expPosStartNextTimeInterval, UpdateType.CalculatedPoint, objectClass);
			MovingObjectState expEndStateNextTimeIntervalMovingObjectState = new MovingObjectState(id, entityType, expPosEndNextTimeInterval, UpdateType.CalculatedPoint, objectClass);
			
			objectIntervalUpdates.add(expStartStateNextTimeIntervalMovingObjectState);
			objectIntervalUpdates.add(expEndStateNextTimeIntervalMovingObjectState);
			*/
			//Helper.DebugWriteLine("Done # of object updates left: "+objectIntervalUpdates.size() + " next time interval = " + nextIntervalBeginTime.GetTime());
		}
		return false;
	}
	
	public boolean DiscreteNewCleanup(Time nextIntervalBeginTime, int delta)
	{
		for (int i=0; i < objectIntervalUpdates.size();i++)
		{
			MovingObjectState mos = (MovingObjectState)objectIntervalUpdates.get(i);
			if (mos.GetUpdateType() == UpdateType.DissappearPoint)
			{
				objectIntervalUpdates.clear();
				intervalReportNumber = 0;
				return true;
				
			}
			else
			{
				objectIntervalUpdates.clear();
				intervalReportNumber = 0;
			}
		}
		
		return false;
	}
	
	public void ResetInterval()
	{
		//we reset everything, except for the latest trajectory
		this.ResetMotionGenericIntervalState();
		objectIntervalUpdates.clear();
		intervalStartPosition = new PositionSpt2D();
		intervalEndPosition = new PositionSpt2D();
	}
	
	public PositionSpt2D GetIntervalStartPosition()
	{	
		return intervalStartPosition;
	}
	public PositionSpt2D GetIntervalEndPosition()
	{
		return intervalEndPosition;
	}
	
	public void Add(MovingObjectState newMovingObjectState, boolean updateTrajectory)
	{
		if (objectIntervalUpdates.size() == 0)
		{
			//Helper.DebugWriteLine("0 location updates so far");
			this.intervalStartPosition = newMovingObjectState.GetPosition();
			this.id = newMovingObjectState.GetID();
			this.startTime = new Time(newMovingObjectState.GetPosition().GetT());
			
			//set the end time and the end position to the same point
			this.intervalEndPosition = newMovingObjectState.GetPosition();
			this.endTime = new Time(newMovingObjectState.GetPosition().GetT());
			
			if (latestTrajectory == null)
			{
				//it's the very first location update... can't do anything about it 
				//Helper.DebugWriteLine("latest trajectory is null");
			}
			else
			{
				//Helper.DebugWriteLine("0 location updates BUT latest trajctory is NOT null");
				//this is the very first update for the next time interval
				//update trajectory accordingly 
				//(end of trajectory = new start of trajectory, new update = new end of trajectory)
	
				if (updateTrajectory == true)
				{
					PositionSpt2D tmpStart = latestTrajectory.GetEndPosition();
					PositionSpt2D tmpEnd  = newMovingObjectState.GetPosition();
				
					//update latest trajectory
					latestTrajectory = new MotionContinuousState(newMovingObjectState.GetID(),
							EntityType.Object, 
							tmpStart.GetX(),
							tmpStart.GetY(),
							tmpStart.GetT(),
							tmpEnd.GetX(),
							tmpEnd.GetY(),
							tmpEnd.GetT(),
							FunctionType.Linear);
				
					//Helper.DebugWriteLine("NEW latest trajctory is  "+latestTrajectory);
				}
			}
		}
		else
		{
			this.intervalEndPosition = newMovingObjectState.GetPosition();
			this.endTime = new Time(newMovingObjectState.GetPosition().GetT());
		}
		
		boolean isExist = false;
		
		//check if exactly the same update might have been already stored
		for (int n = 0; n < objectIntervalUpdates.size(); n++)
		{
			MovingObjectState mos = (MovingObjectState)objectIntervalUpdates.get(n);
			if ((mos.GetPosition().GetX() == newMovingObjectState.GetPosition().GetX()) &&
				(mos.GetPosition().GetY() == newMovingObjectState.GetPosition().GetY()) &&
				(mos.GetPosition().GetT() == newMovingObjectState.GetPosition().GetT()))
			{
				isExist = true;
				break;
			}
		}
		if (isExist == false)
		{
			objectIntervalUpdates.add(newMovingObjectState);
		
			//increment the counter
			this.intervalReportNumber++;
		}
		if (objectIntervalUpdates.size() > 1)
		{
			if (updateTrajectory == true)
			{
				MovingObjectState latestStart = (MovingObjectState)objectIntervalUpdates.get(intervalReportNumber-2);
				MovingObjectState latestEnd = (MovingObjectState)objectIntervalUpdates.get(intervalReportNumber-1); 
			
				//update latest trajectory
				latestTrajectory = new MotionContinuousState(newMovingObjectState.GetID(),
						EntityType.Object, 
						latestStart.GetPosition().GetX(),
						latestStart.GetPosition().GetY(),
						latestStart.GetPosition().GetT(),
						latestEnd.GetPosition().GetX(),
						latestEnd.GetPosition().GetY(),
						latestEnd.GetPosition().GetT(),
						FunctionType.Linear);
			}
		}
	}

	
	
	public void AddDiscrete(MovingObjectState newMovingObjectState)
	{
		if (objectIntervalUpdates.size() == 0)
		{
			//Helper.DebugWriteLine("0 location updates so far");
			this.intervalStartPosition = newMovingObjectState.GetPosition();	
			this.startTime = new Time(newMovingObjectState.GetPosition().GetT());
		}
		else
		{
			this.intervalEndPosition = newMovingObjectState.GetPosition();
			this.endTime = new Time(newMovingObjectState.GetPosition().GetT());
		}
		objectIntervalUpdates.add(newMovingObjectState);
		this.id = newMovingObjectState.GetID();
		
		//increment the counter
		this.intervalReportNumber++;				
	}

	
	
	public ArrayList GetAllObjectIntervalUpdates()
	{
		return objectIntervalUpdates;
	}
	
	public PositionSpt2D CalculateUsingLastTrajectory(Time cTime)
	{
		PositionSpt2D tmpPosS = this.latestTrajectory.GetStartPosition();
		PositionSpt2D tmpPosE = this.latestTrajectory.GetEndPosition();
		PositionSpt2D calculatedPosition = new PositionSpt2D();
		//calculate the slope
		try
		{
			long mx = (tmpPosE.GetX() - tmpPosS.GetX())/(long)(tmpPosE.GetT() - tmpPosS.GetT());
			calculatedPosition.SetX((long)(mx * (cTime.GetTime() - tmpPosS.GetT())) + tmpPosS.GetX());			

			long my = (tmpPosE.GetY() - tmpPosS.GetY())/(long)(tmpPosE.GetT() - tmpPosS.GetT());
			calculatedPosition.SetY((long)(my * (cTime.GetTime() - tmpPosS.GetT())) + tmpPosS.GetY());			
			
			//set the timestamp of the current position
			calculatedPosition.SetT(cTime.GetTime());
		}
		catch(ArithmeticException ex)
		{
			calculatedPosition.SetX(0);
			calculatedPosition.SetY(0);
			calculatedPosition.SetT(cTime.GetTime());
		}
		
		return calculatedPosition;
	}
	
	public void SetIntervalStartPosition(PositionSpt2D position)
	{
		intervalStartPosition = position;
	}
	
	public void SetIntervalEndPosition(PositionSpt2D position)
	{
		intervalEndPosition = position;
	}

}
