/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections;
import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;
/**
 * @author Rimma V. Nehme
 *
 */
public class MovingObjectsStatesCollection extends OperatorStateImp 
{
	
	
	protected ArrayList intervalAllObjectsStates;	//MovingObjectsIntervalStatesCollection

	
	
	public String GetTrajectoriesStatistics()
	{
		String strOutput = "";
		for (int i=0; i<intervalAllObjectsStates.size();i++)
		{
			MovingObjectsIntervalStatesCollection mTemp = (MovingObjectsIntervalStatesCollection)(intervalAllObjectsStates.get(i));
			
			strOutput += "\n"+mTemp.GetLatestTrajectory().toString();
		}
		return strOutput;
	}
	
	
	public void ClearData()
	{
		intervalAllObjectsStates.clear();
	}
	
	public void Cleanup(Time nextIntervalBeginTime, int delta)
	{
		//intervalAllObjectsStates.clear();
		for (int i=0; i<intervalAllObjectsStates.size(); i++)
		{
			MovingObjectsIntervalStatesCollection coll = (MovingObjectsIntervalStatesCollection)intervalAllObjectsStates.get(i);
			if (coll.NewCleanup(nextIntervalBeginTime, delta) == true)
			{
				//is cleanup returns true, that means the object sent
				//a message indicating that it will no longer send updates
				//so remove that object from the collection
				Helper.DebugWriteLine("DISSAPEARING object w/ ID = " + coll.GetID() + "\n");
				coll = null;
				intervalAllObjectsStates.remove(i);				
			}
		}
	}
	
	public void DiscreteCleanup(Time nextIntervalBeginTime, int delta)
	{
		//intervalAllObjectsStates.clear();
		for (int i=0; i<intervalAllObjectsStates.size(); i++)
		{
			MovingObjectsIntervalStatesCollection coll = (MovingObjectsIntervalStatesCollection)intervalAllObjectsStates.get(i);
			coll.objectIntervalUpdates.clear();
			
			/*
			if (coll.DiscreteNewCleanup(nextIntervalBeginTime, delta) == true)
			{
				//is cleanup returns true, that means the object sent
				//a message indicating that it will no longer send updates
				//so remove that object from the collection
				Helper.DebugWriteLine("DISSAPEARING object w/ ID = " + coll.GetID() + "\n");
				coll = null;
				intervalAllObjectsStates.remove(i);				
			}
			*/
		}
	}
	
	public boolean IsDisappearObject(int objectID)
	{
		MovingObjectsIntervalStatesCollection mic = GetObjectByID(objectID);
		for (int i=0; i<mic.GetAllObjectIntervalUpdates().size(); i++)
		{
			MovingObjectState mos = (MovingObjectState)mic.GetAllObjectIntervalUpdates().get(i);
			if (mos.GetUpdateType() == UpdateType.DissappearPoint)
			{
				return true;
			}
		}
		return false;
	}
	
	public MovingObjectsStatesCollection()
	{
		intervalAllObjectsStates = new ArrayList();
	}
	
	public void Add(MovingObjectsIntervalStatesCollection newIntervalStatesCollection)
	{
		intervalAllObjectsStates.add(newIntervalStatesCollection);
	}
	public ArrayList GetAllObjects()
	{
		return intervalAllObjectsStates;
	}
	public void copyStateContent(OperatorState newstate) 
	{
		
	}
	
	public MovingObjectsIntervalStatesCollection GetObjectByID(int newID)
	{
		for (int i=0; i<intervalAllObjectsStates.size();i++)
		{
			MovingObjectsIntervalStatesCollection mTemp = (MovingObjectsIntervalStatesCollection)(intervalAllObjectsStates.get(i));
			if (mTemp.GetID() == newID)
			{
				return mTemp;
			}
		}
		return null;
	}


	public void clearState() {
		// TODO Auto-generated method stub
		
	}
	
}
