/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections;

import java.util.ArrayList;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.FunctionType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.QueryType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.UpdateType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous.MotionContinuousState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingObjectState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.interval.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.*;

/**
 * @author Rimma V. Nehme
 *
 */
public class MovingQueriesIntervalStatesCollection extends MotionGenericIntervalState
{
	protected ArrayList queryIntervalUpdates; //MovingQueryState[]
	private PositionSpt2D intervalStartPosition;
	private PositionSpt2D intervalEndPosition;
	private MotionContinuousState latestTrajectory;
	
	public MovingQueriesIntervalStatesCollection()
	{
		super(0, 0,new Time(0), new Time(0));
		queryIntervalUpdates = new ArrayList();
		intervalStartPosition = new PositionSpt2D();
		intervalEndPosition = new PositionSpt2D();
		latestTrajectory = null;
	}
	
	public MotionContinuousState GetLatestTrajectory()
	{
		return latestTrajectory;
	}
	
	public MovingQueriesIntervalStatesCollection(int newID,
			  									 int newIntervalReportNumber,
												 Time newStartTime,
												 Time newEndTime)
	{
		super(newID, newIntervalReportNumber, newStartTime, newEndTime);
		queryIntervalUpdates = new ArrayList();
		intervalStartPosition = new PositionSpt2D();
		intervalEndPosition = new PositionSpt2D();
		latestTrajectory = null;
	}
	
	
	
	public boolean NewCleanup(Time nextIntervalBeginTime, int delta)
	{
		int id = -1; 
		int entityType = -1;
		int queryClass = -1; 
		int queryBoundary = -1; 
				 
		
		for (int i=0; i < queryIntervalUpdates.size();i++)
		{
			MovingQueryState mqs = (MovingQueryState)queryIntervalUpdates.get(i);
			
			id = mqs.GetID();
			entityType = mqs.GetEntityType();
			queryClass = mqs.GetQueryClass();
			queryBoundary = mqs.GetQueryBoundary();
			
			if (mqs.GetUpdateType() == UpdateType.DissappearPoint)
			{
				return true;				
			}			
		}
		
		if (queryIntervalUpdates.size() > 0)
		{
			//if we got to here, that means all location points will not be used 
			//in the next time interval, so delete them
			queryIntervalUpdates.clear();
		
			this.startTime = nextIntervalBeginTime;
			this.endTime = new Time(nextIntervalBeginTime.GetTime() + delta);
			this.intervalReportNumber = 0;
			//this.intervalStartPosition
			//this.intervalEndPosition	
			
			
			//but now initialize the start and the end of the next time interval
			/*
			PositionSpt2D expPosStartNextTimeInterval = this.latestTrajectory.Calculate(nextIntervalBeginTime);
			PositionSpt2D expPosEndNextTimeInterval = this.latestTrajectory.Calculate(new Time(nextIntervalBeginTime.GetTime() + delta));	

			MovingQueryState expStartStateNextTimeIntervalMovingQueryState = new MovingQueryState(id, entityType, expPosStartNextTimeInterval, UpdateType.CalculatedPoint, queryClass, QueryType.Range, queryBoundary);
			MovingQueryState expEndStateNextTimeIntervalMovingQueryState = new MovingQueryState(id, entityType, expPosEndNextTimeInterval, UpdateType.CalculatedPoint, queryClass, QueryType.Range, queryBoundary);
			
			queryIntervalUpdates.add(expStartStateNextTimeIntervalMovingQueryState);
			queryIntervalUpdates.add(expEndStateNextTimeIntervalMovingQueryState);
			*/					
			//Helper.DebugWriteLine("Done # of query updates left: "+queryIntervalUpdates.size() + " next time interval = " + nextIntervalBeginTime.GetTime());
		}
		return false;
	}
	
	
	
	public boolean DiscreteNewCleanup(Time nextIntervalBeginTime, int delta)
	{
		for (int i=0; i < queryIntervalUpdates.size();i++)
		{
			MovingQueryState mqs = (MovingQueryState)queryIntervalUpdates.get(i);
			if (mqs.GetUpdateType() == UpdateType.DissappearPoint)
			{
				queryIntervalUpdates.clear();
				intervalReportNumber = 0;
				return true;				
			}
			else
			{
				queryIntervalUpdates.clear();
				intervalReportNumber = 0;
			}
		}
		//Helper.DebugWriteLine("Done # of query updates left: "+queryIntervalUpdates.size() + " next time interval = " + nextIntervalBeginTime.GetTime());
		return false;
	}
	
	public void Cleanup(Time nextIntervalBeginTime, int delta)
	{
		//reset interval
		this.ResetInterval(nextIntervalBeginTime, delta);
	}
		
	
	
	public void ResetInterval(Time nextIntervalBeginTime, int delta)
	{
		//we reset everything, except for the latest trajectory
		this.ResetMotionGenericIntervalState(nextIntervalBeginTime, delta);
		queryIntervalUpdates.clear();
		intervalStartPosition = new PositionSpt2D();
		intervalEndPosition = new PositionSpt2D();
	}
	
	public PositionSpt2D GetIntervalStartPosition()
	{
		return new PositionSpt2D();
	}
	public PositionSpt2D GetIntervalEndPosition()
	{
		return new PositionSpt2D();
	}
	
	public void Add(MovingQueryState newMovingQueryState, boolean updateTrajectory)
	{
		if (queryIntervalUpdates.size() == 0)
		{
			this.intervalStartPosition = newMovingQueryState.GetPosition();

			this.id = newMovingQueryState.GetID();
			this.startTime = new Time(newMovingQueryState.GetPosition().GetT());
			
			//set the end time and the end position to the same point
			this.intervalEndPosition = newMovingQueryState.GetPosition();
			this.endTime = new Time(newMovingQueryState.GetPosition().GetT());
			
			
			if (latestTrajectory == null)
			{
				//it's the very first location update... can't do anything about it 
			}
			else
			{
				//this is the very first update for the next time interval
				//update trajectory accordingly 
				//(end of trajectory = new start of trajectory, new update = new end of trajectory)
				if (updateTrajectory == true)
				{
					PositionSpt2D tmpStart = latestTrajectory.GetEndPosition();
					PositionSpt2D tmpEnd  = newMovingQueryState.GetPosition();
				
					//update latest trajectory
					latestTrajectory = new MotionContinuousState(newMovingQueryState.GetID(),
															 EntityType.Query, 
															 tmpStart.GetX(),
															 tmpStart.GetY(),
															 tmpStart.GetT(),
															 tmpEnd.GetX(),
															 tmpEnd.GetY(),
															 tmpEnd.GetT(),
															 FunctionType.Linear);
				}
			}			
		}
		else
		{			
			this.intervalEndPosition = newMovingQueryState.GetPosition();
			this.endTime = new Time(newMovingQueryState.GetPosition().GetT());			
		}
		
		boolean isExist = false;
		
		//check if exactly the same update might have been already stored
		for (int n = 0; n < queryIntervalUpdates.size(); n++)
		{
			MovingQueryState mqs = (MovingQueryState)queryIntervalUpdates.get(n);
			if ((mqs.GetPosition().GetX() == newMovingQueryState.GetPosition().GetX()) &&
				(mqs.GetPosition().GetY() == newMovingQueryState.GetPosition().GetY()) &&
				(mqs.GetPosition().GetT() == newMovingQueryState.GetPosition().GetT()))
			{
				isExist = true;
				break;
			}
		}
		if (isExist == false)
		{
			queryIntervalUpdates.add(newMovingQueryState);
		
			//increment the counter
			this.intervalReportNumber++;
		}					
		
		if (queryIntervalUpdates.size() > 1)
		{
			if (updateTrajectory == true)
			{
			MovingQueryState latestStart = (MovingQueryState)queryIntervalUpdates.get(intervalReportNumber-2);
			MovingQueryState latestEnd = (MovingQueryState)queryIntervalUpdates.get(intervalReportNumber-1); 
						
			//update latest trajectory
			latestTrajectory = new MotionContinuousState(newMovingQueryState.GetID(),
														 EntityType.Query, 
														 latestStart.GetPosition().GetX(),
														 latestStart.GetPosition().GetY(),
														 latestStart.GetPosition().GetT(),
														 latestEnd.GetPosition().GetX(),
														 latestEnd.GetPosition().GetY(),
														 latestEnd.GetPosition().GetT(),
														 FunctionType.Linear);
			}
						
		}
	}
	
	
	public void AddDiscrete(MovingQueryState newMovingQueryState)
	{
		if (queryIntervalUpdates.size() == 0)
		{
			//Helper.DebugWriteLine("0 location updates so far");
			this.intervalStartPosition = newMovingQueryState.GetPosition();	
			this.startTime = new Time(newMovingQueryState.GetPosition().GetT());
		}
		else
		{
			this.intervalEndPosition = newMovingQueryState.GetPosition();
			this.endTime = new Time(newMovingQueryState.GetPosition().GetT());
		}
		queryIntervalUpdates.add(newMovingQueryState);
		this.id = newMovingQueryState.GetID();
		
		//increment the counter
		this.intervalReportNumber++;				
	}
	
	
	public ArrayList GetAllQueryIntervalUpdates()
	{
		return queryIntervalUpdates;
	}
	
	public PositionSpt2D CalculateUsingLastTrajectory(Time cTime)
	{		
		PositionSpt2D tmpPosS = this.latestTrajectory.GetStartPosition();
		PositionSpt2D tmpPosE = this.latestTrajectory.GetEndPosition();
		PositionSpt2D calculatedPosition = new PositionSpt2D();
		//calculate the slope
		try
		{
			long mx = (tmpPosE.GetX() - tmpPosS.GetX())/(long)(tmpPosE.GetT() - tmpPosS.GetT());
			calculatedPosition.SetX((long)(mx * (cTime.GetTime() - tmpPosS.GetT())) + tmpPosS.GetX());			

			long my = (tmpPosE.GetY() - tmpPosS.GetY())/(long)(tmpPosE.GetT() - tmpPosS.GetT());
			calculatedPosition.SetY((long)(my * (cTime.GetTime() - tmpPosS.GetT())) + tmpPosS.GetY());			
			
			//set the timestamp of the current position
			calculatedPosition.SetT(cTime.GetTime());
		}
		catch(ArithmeticException ex)
		{
			calculatedPosition.SetX(0);
			calculatedPosition.SetY(0);
			calculatedPosition.SetT(cTime.GetTime());
		}
		
		return calculatedPosition;
	}
	
	public void SetIntervalStartPosition(PositionSpt2D position)
	{
		intervalStartPosition = position;
	}
	
	public void SetIntervalEndPosition(PositionSpt2D position)
	{
		intervalEndPosition = position;
	}
}
