/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.collections;

import java.util.ArrayList;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.UpdateType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingObjectState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingQueryState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;

/**
 * @author Rimma V. Nehme
 *
 */
public class MovingQueriesStatesCollection extends OperatorStateImp 
{
	protected ArrayList intervalAllQueriesStates;	//MovingQueriesIntervalStatesCollection

	public MovingQueriesStatesCollection()
	{
		intervalAllQueriesStates = new ArrayList();
	}
	
	
	public boolean IsDisappearQuery(int queryID)
	{
		MovingQueriesIntervalStatesCollection mic = GetQueryByID(queryID);
		for (int i=0; i<mic.GetAllQueryIntervalUpdates().size(); i++)
		{
			MovingQueryState mqs = (MovingQueryState)mic.GetAllQueryIntervalUpdates().get(i);
			if (mqs.GetUpdateType() == UpdateType.DissappearPoint)
			{
				return true;
			}
		}
		return false;
	}
	
	public String GetTrajectoriesStatistics()
	{
		String strOutput = "";
		for (int i=0; i<intervalAllQueriesStates.size();i++)
		{
			MovingQueriesIntervalStatesCollection mTemp = (MovingQueriesIntervalStatesCollection)(intervalAllQueriesStates.get(i));
			
			strOutput += "\n"+mTemp.GetLatestTrajectory().toString();
		}
		return strOutput;
	}

	public void ClearData()
	{
		intervalAllQueriesStates.clear();
	}
	
	public void Cleanup(Time nextIntervalBeginTime, int delta)
	{
		for (int i=0; i<intervalAllQueriesStates.size(); i++)
		{
			MovingQueriesIntervalStatesCollection coll = (MovingQueriesIntervalStatesCollection)intervalAllQueriesStates.get(i);
			if (coll.NewCleanup(nextIntervalBeginTime, delta) == true)
			{
				//is cleanup returns true, that means the query sent
				//a message indicating that it will no longer send updates
				//so remove that object from the collection
				Helper.DebugWriteLine("DISSAPEARING query w/ ID = " + coll.GetID() + "n");
				coll = null;
				intervalAllQueriesStates.remove(i);				
			}
		}
	}
	
	public void DiscreteCleanup(Time nextIntervalBeginTime, int delta)
	{
		for (int i=0; i<intervalAllQueriesStates.size(); i++)
		{
			MovingQueriesIntervalStatesCollection coll = (MovingQueriesIntervalStatesCollection)intervalAllQueriesStates.get(i);
			coll.queryIntervalUpdates.clear();
			
			/*
			if (coll.DiscreteNewCleanup(nextIntervalBeginTime, delta) == true)
			{
				//is cleanup returns true, that means the query sent
				//a message indicating that it will no longer send updates
				//so remove that object from the collection
				Helper.DebugWriteLine("DISSAPEARING query w/ ID = " + coll.GetID() + "n");
				coll = null;
				intervalAllQueriesStates.remove(i);				
			}
			*/
		}
	}
		
	
	public void copyStateContent(OperatorState newstate) 
	{
		
	}	
	
	public void Add(MovingQueriesIntervalStatesCollection newIntervalStatesCollection)
	{
		intervalAllQueriesStates.add(newIntervalStatesCollection);
	}
	public ArrayList GetAllQueries()
	{
		return intervalAllQueriesStates;
	}
	
	public MovingQueriesIntervalStatesCollection GetQueryByID(int newID)
	{
		for (int i=0; i<intervalAllQueriesStates.size();i++)
		{
			MovingQueriesIntervalStatesCollection mTemp = (MovingQueriesIntervalStatesCollection)(intervalAllQueriesStates.get(i));
			if (mTemp.GetID() == newID)
			{
				return mTemp;
			}
		}
		return null;
	}


	public void clearState() {
		// TODO Auto-generated method stub
		
	}
}
