/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.interfaces.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.MotionGenericState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MotionDiscreteState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MovingObjectState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.*;

/**
 * @author Rimma V. Nehme
 *
 */
public class MotionContinuousState extends MotionGenericState implements IFunctionInterface
{
	protected PositionSpt2D startPosition;
	protected PositionSpt2D endPosition;
	protected int continuityType; //enum: FunctionType
	
	public MotionContinuousState() 
	{
		super(0,EntityType.NA);
		this.startPosition = new PositionSpt2D();
		this.endPosition = new PositionSpt2D();
		this.continuityType = FunctionType.NA;
	}
	
	public String toString()
	{
		return "MotionContinuousState():\n"+
					"ID: "+ this.id+"\n"+
					"Entity Type: "+EntityType.FromEnumToString(this.entityType)+"\n"+
					"Start Position: "+this.startPosition.toString()+"\n"+
					"End Position: "+this.endPosition.toString()+"\n";
	}
	
	public MotionContinuousState(int newID,
							     int newEntityType,
							     long newXStart,
							     long newYStart, 
							     float newTStart,
								 long newXEnd,
							     long newYEnd, 
							     float newTEnd,
								 int newContinuityType) 
	{
		super(newID, newEntityType);
		   
		this.startPosition = new PositionSpt2D(newXStart, newYStart, newTStart);
		this.endPosition = new PositionSpt2D(newXEnd, newYEnd, newTEnd);
		this.continuityType = newContinuityType;
	}
	
	public PositionSpt2D GetStartPosition()
	{
		return this.startPosition;
	}
	
	public PositionSpt2D GetEndPosition()
	{
		return this.endPosition;
	}
	
	public void SetStartPosition(PositionSpt2D newStartPosition)
	{
		startPosition = newStartPosition;
	}
	
	public void SetEndPosition(PositionSpt2D newEndPosition)
	{
		endPosition = newEndPosition;
	}
	
	
	public PositionSpt2D Calculate(Time t)
	{
		return new PositionSpt2D();
	}
	
	public int GetContinuityType()
	{
		return continuityType;
	}
	
	public PositionSpt2D Calculate(long t)
	{
		PositionSpt2D tmpPosS = this.startPosition;
		PositionSpt2D tmpPosE = this.endPosition;
		PositionSpt2D calculatedPosition = new PositionSpt2D();
		//calculate the slope
		try
		{
			long mx = (tmpPosE.GetX() - tmpPosS.GetX())/(long)(tmpPosE.GetT() - tmpPosS.GetT());
			calculatedPosition.SetX((mx * (long)(t - tmpPosS.GetT())) + tmpPosS.GetX());			

			long my = (tmpPosE.GetY() - tmpPosS.GetY())/(long)(tmpPosE.GetT() - tmpPosS.GetT());
			calculatedPosition.SetY((my * (long)(t - tmpPosS.GetT())) + tmpPosS.GetY());			
			
			//set the timestamp of the current position
			calculatedPosition.SetT(t);
		}
		catch(ArithmeticException ex)
		{
			calculatedPosition.SetX(0);
			calculatedPosition.SetY(0);
			calculatedPosition.SetT(t);
		}
		catch(Exception ex)
		{
			calculatedPosition.SetX(0);
			calculatedPosition.SetY(0);
			calculatedPosition.SetT(t);
		}
		
		return calculatedPosition;
	}
	
	public Time Calculate(PositionSpt2D p)
	{
		//TO IMPLEMENT
		return null;
	}

}
