/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.continuous;
import  edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MotionDiscreteState;

/**
 * @author Rimma V. Nehme
 *
 */
public class MovingObjectContinuousState extends MotionContinuousState
{
	protected int objectClass; //enum: ClassType
	
	public MovingObjectContinuousState() 
	{
		super();
		this.objectClass = ClassType.NA;
	}
	
	public MovingObjectContinuousState(int newID,
							     int newEntityType,
							     long newXStart,
							     long newYStart, 
							     float newTStart,
								 long newXEnd,
							     long newYEnd, 
							     float newTEnd,
								 int newContinuityType,
								 int newObjectClass) 
	{
		super(newID, newEntityType, newXStart, newYStart, newTStart, newXEnd, newYEnd, newTEnd, newContinuityType);		
		this.objectClass = newObjectClass;
	}
	
	public String toString()
	{
		return "MovingObjectContinuousState():\n"+
					"ID: "+ this.id+"\n"+
					"Entity Type: "+EntityType.FromEnumToString(this.entityType)+"\n"+
					"Start Position: "+this.startPosition.toString()+"\n"+
					"End Position: "+this.endPosition.toString()+"\n"+
					"Object Class: "+ClassType.FromEnumToString(this.objectClass)+"\n"+				
					"Continuity Type: "+FunctionType.FromEnumToString(this.continuityType)+"\n";
	}
	
	public int  GetObjectClass()
	{
		return objectClass;
	}
	
	public void  SetObjectClass(int newQueryClass)
	{
		objectClass = newQueryClass;
	}	
}
