/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.ObjectFactory;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.*;
/**
 * @author Rimma V. Nehme
 *
 */
public class MovingObjectState extends MotionDiscretePositionUpdate implements ObjectFactory
{
	protected int objectClass; //enum: ClassType

	public MovingObjectState() 
	{
		super(0, EntityType.NA, 0, 0, 0, UpdateType.NA);
		this.objectClass = ClassType.NA;
	}

	public MovingObjectState(int newID,
							 int newEntityType,							
							 long newX,
							 long newY, 
							 long newT,
							 int newUpdateType,
							 int newObjectClass) 
	{
		
		super(newID, newEntityType, newX, newY, newT, newUpdateType);
		this.objectClass = newObjectClass;
		
	}
	
	public MovingObjectState(int newID,
			 				 int newEntityType,							
							 PositionSpt2D newPosition,
							 int newUpdateType,
							 int newObjectClass) 
	{

		super(newID, newEntityType, newPosition.GetX(), newPosition.GetY(), newPosition.GetT(), newUpdateType);
		this.objectClass = newObjectClass;
	}
	
	public int  GetObjectClass()
	{
		return objectClass;
	}
	
	public void  SetObjectClass(int newObjectClass)
	{
		objectClass = newObjectClass;
	}
	
	public String toString()
	{
		return "MovingObjectState():\n"+
					"ID: "+ this.id+"\n"+
					"Entity Type: "+EntityType.FromEnumToString(this.entityType)+"\n"+
					"Object Class: "+ClassType.FromEnumToString(this.objectClass)+"\n"+
					"Update Type: "+UpdateType.FromEnumToString(this.updateType)+"\n"+
					"Position: "+this.position.toString()+"\n";
	}

	public Object makeObject() 
	{
		return new MovingObjectState();
	}
}
