/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.ObjectFactory;

/**
 * @author Rimma V. Nehme
 *
 */
public class MovingQueryState extends MotionDiscretePositionUpdate implements ObjectFactory
{
	protected int queryClass; //enum: ClassType
	protected int queryType;  //enum: QueryType
	protected int queryBoundary; //size of the query

	public MovingQueryState() 
	{
		super(0, EntityType.NA, 0, 0, 0, UpdateType.NA);
		this.queryClass = ClassType.NA;
		this.queryType = QueryType.NA;
		this.queryBoundary = 0;
	}

	public MovingQueryState (int newID,
							 int newEntityType,							
							 long newX,
							 long newY, 
							 float newT,
							 int newUpdateType,
							 int newQueryClass,
							 int newQueryType,
							 int newQueryBoundary) 
	{
		super(newID, newEntityType, newX, newY, newT, newUpdateType);
		this.queryClass = newQueryClass;
		this.queryType = newQueryType;
		this.queryBoundary = newQueryBoundary;
	}
	
	public MovingQueryState (int newID,
			 int newEntityType,							
			 PositionSpt2D newPosition,
			 int newUpdateType,
			 int newQueryClass,
			 int newQueryType,
			 int newQueryBoundary) 
	{
		super(newID, newEntityType, newPosition.GetX(), newPosition.GetY(), newPosition.GetT(), newUpdateType);
		this.queryClass = newQueryClass;
		this.queryType = newQueryType;
		this.queryBoundary = newQueryBoundary;
	}
	
	
	public int  GetQueryClass()
	{
		return queryClass;
	}
	
	public void  SetQueryClass(int newQueryClass)
	{
		queryClass = newQueryClass;
	}	
	
	public int  GetQueryType()
	{
		return queryType;
	}
	
	public int GetQueryBoundary()
	{
		return queryBoundary;
	}
	
	public void SetQueryBoundary(int newQueryBoundary)
	{
		queryBoundary = newQueryBoundary;
	}
	
	public void  SetQueryType(int newQueryType)
	{
		queryType = newQueryType;
	}	
	
	public String toString()
	{
		return  "MovingQueryState():\n"+
				"ID: "+ this.id+"\n"+
				"Entity Type: "+EntityType.FromEnumToString(this.entityType)+"\n"+
				"Query Type :"+QueryType.FromEnumToString(this.queryType)+"\n"+
				"Query Class :"+ClassType.FromEnumToString(this.queryClass)+"\n"+
				"Update Type :"+UpdateType.FromEnumToString(this.updateType)+"\n"+
				"Position: "+this.position.toString()+"\n";
	}


	public Object makeObject() 
	{
		return new MovingQueryState();
	}
}

