/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
//
//  Motion Data Processing 1.0 
//
//  Copyright 2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Rimma V. Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  me: <rimma3@WPI.EDU>.
//
/////////////////// WORCESTER POLYTECHNIC INSTITUTE ///////////////////////
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.interval;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.interfaces.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.*;
/**
 * @author Rimma V. Nehme
 *
 */
public abstract class MotionGenericIntervalState implements IIntervalStateInterface
{
	protected int id;
	protected int intervalReportNumber;
	protected Time startTime;
	protected Time endTime;
	
	public MotionGenericIntervalState() 
	{
		this.id = 0;
		this.intervalReportNumber = 0;
		this.startTime = new Time();
		this.endTime = new Time();
	}
	
	public MotionGenericIntervalState(int newID,
	   		  						  int newIntervalReportNumber,
									  Time newStartTime,
									  Time newEndTime)
	{
		this.id = newID;
		this.intervalReportNumber = newIntervalReportNumber;
		this.startTime = newStartTime;
		this.endTime = newEndTime;
	}
	
	
	protected void ResetMotionGenericIntervalState(Time nextIntervalBeginTime, int delta)
	{
		this.intervalReportNumber = 0;
		this.startTime = nextIntervalBeginTime;
		this.endTime = new Time(nextIntervalBeginTime.GetTime() + delta);		
	}
	
	protected void ResetMotionGenericIntervalState()
	{
		this.intervalReportNumber = 0;
		this.startTime = new Time();
		this.endTime = new Time();		
	}
	
	
	
	public int GetID()
	{
		return id;
	}
	
	public int GetIntervalReportNumber()
	{
		return intervalReportNumber;
	}
	
	public Time GetIntervalStartTime()
	{
		return startTime;
	}
	
	public Time GetIntervalEndTime()
	{
		return endTime;
	}
	
	public void SetIntervalReportNumber(int newIntervalReportNumber)
	{
		intervalReportNumber = newIntervalReportNumber;
	}
}
