/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Oct 14, 2004
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator;
import java.lang.reflect.Constructor;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.ArrayList;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATMemoryQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;

/**
 * @author Bin Liu
 * A split operator definition. Assume one input queue, multiple output queues. 
 * In general, one output queue corresponds to one remote machine. 
 */
public class StreamSplitOperatorImp extends XATMultiQueueStreamOperatorImp
		implements StreamSplitOperator {

	int numberOfPartitions = 0; // the number of partitions
	int partitionColumnIndex = 0; // indicate the column that the partition will be based on
	Hashtable mappingTable; // the mapping between the partition ids and the machines
	boolean initialized = false;	//indicates whether the initialized has been done?
	PartitionFunction partitionFunction;
	
	boolean partitioned_split = false; // to indicate whether the split operator is activated with join join operators
	
	XATMemoryQueueImp tmpqueue = null; // to hold the partition inputs having IDs that will be adapted
	boolean deactivated = false; // those two are used to indicate whether we need to put inputs to tmpqueue
	boolean reactivated = false; // or read inputs from the tmpqueue
	
	//for parallel track partition redistribution
	Hashtable tempMappingTable = null;  //hold partitions to be moved to different machine. 
	
	public boolean getPartitionSplitFlag(){
		return partitioned_split;
	}
	
	public int getStateSize() {
		return 0;
	}
	
	public boolean nullPartitionFunction(){
		if (partitionFunction == null)
			return true;
		else
			return false;
	}
	
	public synchronized void reactivatePartitions(ArrayList toAct, String ipAdd){
		for (int i=0; i<toAct.size(); i++){
			Integer idx = (Integer)toAct.get(i);
			mappingTable.put(idx,ipAdd);
		}
		reactivated = true;
	}
	
	public StreamSplitOperatorImp(XATNode node){
		super(node);
		// those are default properties defined in the query plan for this operator
		this.addProperty("PARTITION_ATTRIBUTE","0");
		this.addProperty("NUMBER_OF_PARTITION","0");
		this.addProperty("PARTITION_FUNCTION","edu.wpi.cs.dsrg.xmldb.xat.common.operator.complexstreamoperator.ModPartitionFunction");
		this.addProperty("PARTITIONED_SPLIT","false"); //default for the split operator, only activated in one machine
		mappingTable = new Hashtable();
	}
	
	public void setInitialProperties(){
		partitionColumnIndex = Integer.parseInt(this.getProperty("PARTITION_ATTRIBUTE"));
		numberOfPartitions = Integer.parseInt(this.getProperty("NUMBER_OF_PARTITION"));
		String ePartition = this.getProperty("PARTITION_FUNCTION");	
		String tmpP = this.getProperty("PARTITIONED_SPLIT");
		if (tmpP!=null){
			if (tmpP.equals("true")) {
				partitioned_split = true;
			}			
		}
	}
	
	/* 
	 * Major running functions of the operator, takes one tuple at a time and send them to the
	 * corresponding queues
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public synchronized boolean run() throws Exception {

		XATQueueSchema inSchema;
		int inLength;
		if ((inSchema = this.getInputQueue().getSchema()) == null){
				return false;
		}
		inLength = inSchema.getLength();
		
		if(this.getOutputQueue().getSchema() == null){
			init();
		}
		
		if (!initialized){
			// may need to handle exceptions due to converting numbers
			partitionColumnIndex = Integer.parseInt(this.getProperty("PARTITION_ATTRIBUTE"));
			numberOfPartitions = Integer.parseInt(this.getProperty("NUMBER_OF_PARTITION"));
			String ePartition = this.getProperty("PARTITION_FUNCTION");
			String tmpP = this.getProperty("PARTITIONED_SPLIT");
			if (tmpP!=null){
				if (tmpP.equals("true")) {
					partitioned_split = true;
				}			
			}
			Constructor[] c1 = Class.forName(ePartition).getConstructors();
			Constructor construct;
			for(int i = 0; i < c1.length; i++){
				construct = c1[i];
				Class[] parameters = construct.getParameterTypes();			
				if(parameters.length == 1){
					Class[] constructorArgumentTypes=null;
					Object[] constructorArgs=null;
					constructorArgumentTypes = new Class[] {Integer.class};
					constructorArgs = new Object[] {new Integer(numberOfPartitions)};
					partitionFunction = (PartitionFunction) construct.newInstance(constructorArgs);
				}
			}		
			if(partitionFunction == null){
				throw new Exception("The PartitionFunction could not be instantiated.");
			}			
			initialized = true;
		}
		
		if(this.mappingTable == null || this.mappingTable.size() == 0) {
		    //System.out.println("Partition mapping table is not set. Return right away.");
		    return false;
		}
//		else {
//		    System.out.println("The current mapping table is:\n" + this.mappingTable);
//		}
		
//		System.out.println("\nrun " + this);
		
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
//		System.out.println("NUMBER_OF_TUPLES_TO_DEQUEUE:" + howMany);
		int tupleCount = 0;
		
		//Runtime.getRuntime().gc();
		
		/** get the tuples **/
		while (true){
			XATTuple[] tuples = this.getInputQueue().dequeue(howMany, this);
//			System.out.println("Number of tuples dequeued: " + tuples.length);
			
			for (int i=0; i<tuples.length; i++){
				XATTuple curTuple = tuples[i];
				
				if (curTuple.getPartitionID()==-10) {
					//representing the end of input streams, send the tuple to all outputqueues
					for(Iterator cursor= this.getOutputQueues(); cursor.hasNext();){
						((XATQueue)cursor.next()).enqueue(curTuple,this);
					}
					return true;
				}
				
				// this might be memory inefficient?
				if (partitionColumnIndex>=curTuple.getLength()){
					System.out.println("Wrong!!!");
					return true;
				}
				int partitionID = partitionFunction.partitionTuple(curTuple,partitionColumnIndex);
				/*
				// manually find the partition ID				
				XATDataValue tuple_val = curTuple.getCell(partitionColumnIndex);
				String str_val = tuple_val.convertToString();
				int int_val = 0;
				try{
					int_val = Integer.parseInt(str_val);
				} catch (NumberFormatException ne){
					// do nothing
				}
				//simply return the mod 
				int partitionID = int_val%300;
				//end of manually compute partition ID
				*/
				tuples[i].setPartitionID(partitionID);
				// get the destination
				String ipAddress = (String)mappingTable.get(new Integer(partitionID));
				if (ipAddress == null){
					if (deactivated) {
						// put inputs temporary to tmpqueue
						tmpqueue.enqueue(tuples[i],this);
					} else {
						System.out.println("Something wrong. Can not find the destination of ID: " + partitionID + "while it is not deactivated.");
					}
				} else {
					//if in the process of parallel partition redistribution
					//send the tuple to another machine as well.
					if(tempMappingTable != null) {
					    String ipDuplicate = (String)tempMappingTable.get(new Integer(partitionID));
					    if(ipDuplicate != null) {
					        (this.getOutputQueueByName(ipDuplicate)).enqueue(tuples[i],this);
					        tuples[i].setToNew(true);
					    }
					}					
					// find the outputqueue and enqueue the tuple
					(this.getOutputQueueByName(ipAddress)).enqueue(tuples[i],this);
				}
				

			}
			
			// after reactivate, we need to check whether tmpqueue has inputs, if yes, also process inputs from tmpqueue
			if (reactivated) {
				if (tmpqueue != null) {
					if (tmpqueue.getQueueSize() >0) {
					    //System.out.println("Dequeue tuples from temporary buffer! howMany = " + howMany);
						//System.out.println("tuples been kept in temp queue: " + tmpqueue.getQueueSize());
						tuples = tmpqueue.dequeue(howMany, this);
						//System.out.println("tuples dequeued from temp queue: " + tuples.length);
						for (int i=0; i<tuples.length; i++){
							XATTuple curTuple = tuples[i];
							int partitionID = partitionFunction.partitionTuple(curTuple,partitionColumnIndex);
							tuples[i].setPartitionID(partitionID);
							// get the destination
							String ipAddress = (String)mappingTable.get(new Integer(partitionID));
							if (ipAddress == null){
								System.out.println("Something wrong. Can not find the destination of ID: " + partitionID + "while it is reactivated.");
							} else {
								(this.getOutputQueueByName(ipAddress)).enqueue(tuples[i],this);
							}
						} //end for						
					} else {
						// no more data in the tmpqueue
						reactivated = false;
					}
				}// end if tmpqueue!=null
			}
			// debug code, to check how many tuples in all output queuse
			//int totalOuts = 0;
			//for (Iterator it=this.getOutputQueues(); it.hasNext() ;) {
			//	totalOuts += ((XATQueue)it.next()).getQueueSize();
			//} 
			//System.out.println("SPLIT OUTPUT QUEUE SIZE: " + totalOuts);
			//end for debug code
			
			tuples = null;	
			return true;			
		} 
	} //end of run()

	/*
	 * This is called in the SymmMwayHashJoinOperator, when to trace the linearage of the output tuple
	 * in order to update the cost statistics
	 */
	public int partitionTuple(XATTuple tp, int column){
		return partitionFunction.partitionTuple(tp,column);
	}
	/*
	 * Several basic access functions 
	 */
	public void setNumberOfPartitions(int np){
		numberOfPartitions = np;
	}
	public int getNumberOfPartitions(){
		return numberOfPartitions;
	}
	public void setPartitionMappingTable(Hashtable mt){
		mappingTable = mt;
	}
	public Hashtable getPartitionMappingTable(){
		return mappingTable;
	}
	public void setPartitionColumnIndex(int pci){
		partitionColumnIndex = pci;
	}

	public synchronized boolean deactivatePartitions(ArrayList parts, String senderIPAddr){
		// (1) remove partition IDs in parts from the mappingTable
		for (int i=0; i<parts.size(); i++){
			Integer id = (Integer)parts.get(i);
			if (mappingTable.containsKey(id)){
				mappingTable.remove(id);
			}
		}
		// (2) create temporary queue to hold the inputs from these partitions
		if (tmpqueue == null) { // we do not have to create multiple times
			tmpqueue = new XATMemoryQueueImp(this); 
		}
		// (3) create a EOF tuple indicates the end of tuples with deactivated partition IDs
		// and send this tuple to the queue indicated by senderIPAddr
		XATTuple eofParts = new XATTupleImp();
		(this.getOutputQueueByName(senderIPAddr)).enqueue(eofParts,this);		
		// (4) set the deactivate flag
		deactivated = true;
		reactivated = false;
		
		return true;
	}
	/* 
	 * The outputQueue has more than one queues, return the one that has the same name
	 */
	public edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue getOutputQueueByName(String name){
		for (int i=0; i<outputQueue.size(); i++){
			if ( ((XATQueue)outputQueue.get(i)).getQueueName().equals(name) ){
				return (XATQueue)outputQueue.get(i);
			}
		}
		// if it is not found, just return null
		return null;
	}

	/*
	 * Remove all existing output queues
	 */
	public void removeQutputQueues(){
		outputQueue.clear();
	}
	/*
	 * Need own version of init() to set all output queues (include dynamic generated one) schema
	 */
	public void init() {
		Iterator i = this.getInputQueues();
		String schemaName = new String();
		XATQueueSchema s;
		XATQueueSchema outSchema = new XATQueueSchema();

		while (i.hasNext()) {

			s = ((XATQueue) i.next()).getSchema();
			schemaName += "#" + s.getName();
			
			Iterator j = s.getColNames().iterator();
			while(j.hasNext()){
				outSchema.addColumn((String) j.next());
			}
		}

		outSchema.setName(schemaName);
		Iterator it = this.getOutputQueues();
		
		while (it.hasNext()) {
			((XATQueue)it.next()).setSchema(outSchema);
		}
		//this.getOutputQueue().setSchema(outSchema);
		
		computeOutputPositions();
	}
	
	/*
	 * This function is used when this particular split instance is not activated in the machine, 
	 * while this operator is chosen when trace back the cost statistics in the symmmwayhashjoinoperator
	 * when multiple join operators exist in the query tree
	 */
	public void reloadPartitionFunction(){
		try{
			// may need to handle exceptions due to converting numbers
			partitionColumnIndex = Integer.parseInt(this.getProperty("PARTITION_ATTRIBUTE"));
			numberOfPartitions = Integer.parseInt(this.getProperty("NUMBER_OF_PARTITION"));
			String ePartition = this.getProperty("PARTITION_FUNCTION");
			String tmpP = this.getProperty("PARTITIONED_SPLIT");
			if (tmpP!=null){
				if (tmpP.equals("true")) {
					partitioned_split = true;
				}			
			}
			Constructor[] c1 = Class.forName(ePartition).getConstructors();
			Constructor construct;
			for(int i = 0; i < c1.length; i++){
				construct = c1[i];
				Class[] parameters = construct.getParameterTypes();			
				if(parameters.length == 1){
					Class[] constructorArgumentTypes=null;
					Object[] constructorArgs=null;
					constructorArgumentTypes = new Class[] {Integer.class};
					constructorArgs = new Object[] {new Integer(numberOfPartitions)};
					partitionFunction = (PartitionFunction) construct.newInstance(constructorArgs);
				}
			}		
			if(partitionFunction == null){
				throw new Exception("The PartitionFunction could not be instantiated.");
			}			
		} catch (Exception cnf) {
			System.out.println("class can not be found when reload the partition function?");
		}
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAccumulatedTuples(long)
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return false;
	}

	public boolean finalizeOperation(){
		return true;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAllAccumulatedTuples()
	 */
	public boolean cleanAllAccumulatedTuples() {
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#matchStates(edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator)
	 */
	public boolean matchStates(XATStreamOperator op) {
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#hasUnmatchedState()
	 */
	public boolean hasUnmatchedState() {
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#recomputeUnmatchedState()
	 */
	public boolean recomputeUnmatchedState() {
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#setStateID(int)
	 */
	public int setStateID(int startID) {
		return 0;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		return null;
	}
	
	public synchronized void sendDuplicatePartitions(ArrayList partitionsToDuplicate, String senderIPAddr, String receiverIPAddr) {
	    tempMappingTable = new Hashtable();
	    for(int i=0; i<partitionsToDuplicate.size(); i++) {
	        Object partid = partitionsToDuplicate.get(i);
	        tempMappingTable.put(partid, receiverIPAddr);
	    }
	    System.out.println("Create temp mapping table: \n" + tempMappingTable);
	    
		// create a EOF tuple indicates the end of old tuples (start of new tuples) 
	    // with duplicated partition IDs and send this tuple to the queue indicated by senderIPAddr
		//XATTuple eofParts = new XATTupleImp();
		//(this.getOutputQueueByName(senderIPAddr)).enqueue(eofParts,this);	
	}
	
	public synchronized void stopDuplicatePartitions(String senderIPAddr) {
	    //System.out.println("Stop duplicating partitions");
	    System.out.println("temp mapping table: \n" + tempMappingTable);
	    if(tempMappingTable != null) {
	        Enumeration e = tempMappingTable.keys();
            Integer partid = (Integer)e.nextElement();
            String ip = (String)tempMappingTable.get(partid);
            System.out.println("Receiver IP: " + ip);
            System.out.println("Replace: (" + partid + ", " + mappingTable.get(partid) + ") by: (" + partid + ", " + ip + ") in mappingtable");
            mappingTable.put(partid, ip);	
	        while(e.hasMoreElements()) {
	            partid = (Integer)e.nextElement();
	            System.out.println("Replace: (" + partid + ", " + mappingTable.get(partid) + ") by: (" + partid + ", " + ip + ") in mappingtable");
	            mappingTable.put(partid, ip);	            
	        }
	        tempMappingTable = null;
			// create a EOPARTS tuple with partitionID = -2 tuple indicates the end of sending tuples to sender machines during PT.
			XATTuple eofParts = new XATTupleImp();
			eofParts.setPartitionID(-2);
			(this.getOutputQueueByName(senderIPAddr)).enqueue(eofParts,this);	
	    } else {
	        System.out.println("Something is wrong! The temp mapping table should not be empty!");
	        System.exit(-1);
	    }
	}

}
