/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Oct 22, 2004
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator;

//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WEquiJoinStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.DistributedExecutionController;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Enumeration;
//import java.util.Iterator;
import java.util.Random;
import java.io.*;

/**
 * @author Bin Liu
 * Symmetric Mway Hash Join Operator 
 */
public class SymmMwayHashJoinOperatorImp extends XATMultiQueueStreamOperatorImp
		implements SymmMwayHashJoinOperator {

	private int[] joinAttributePositions; // record the join attribute index (join condition) for each input queue
	private int[] joinInputQueueCells;	// the number of cells in each input queue
	private ArrayList joinExpressions;	//	store the join definitions parsed from xml file
	
	private boolean initialized = false;	// to indicate whether the initialization has been done, do not have to do it multiple times.
	private boolean completedFinalization = false; // if the finalization has been completed, then no need of re-do
	private boolean readyToClean = false; //for query tree with multiple join operators, the start of the cleanup stage has to wait the previous one has finished
	private int numberOfInstances = 0; // to indicate how many instances of this join (the number of machines that this join are deployed), 
	//it is set in the localadapter initialization time, based on the output queue numbers of previous_split_operator, it is used to in the finalization stage,
	//to compare the number with numberOfEOFReceived to inidcate the readyToClean
	
	Hashtable hashJoinStates;	// hold the join states of the operator
	
	Hashtable statesPushedToDisk; // to indicate which states (ID) have been pushed into disk
	
	// for disk based process, put into the disk and read back from the disk.
	// for debug useage only, those information can be defined in the properties of query operator
	//long totalAvailableMemory = 10; // total available JVM memory, in terms of M
	String baseStateDirName = "states/diskPartitions_";
	String clientName = null; // only for the wpi-cluster that all machines in the cluster have the same virtual disk space?
	
	double memoryThreshold = 0.8; // indicate the memory threshold that trigger the memory to disk dumping
	int timesCalledForDiskDump = 0; // to control the time of merging disk partitions
	int prePushPartitionSize=0; // to control whether we really need dump
	
	// for partition adaptations, protocol Step 3 -- DeactivatePartitions.
	ArrayList deactivatedParts = null; //contains partition IDs that will be deactivated
	// boolean noMoreDeactivatedPars = false; // indicate whether have partitions with IDs in deactiavtedParts
	boolean startChecking = false;
	int numberOfEOFReceived = 0; // to indicate whether we have received all EOFs, if it equals to the input queue size
	int totalDiskTuples = 0;
	// Seems sometimes the tuples with partitionID == -1 get lost, thus the distributed adaptation will never go on,
	// to avoid this situation, the following sudoTimeOutCount is used to set the maximal times to check the ID==-1 tuples,
	// a enough long checking implies that all on-the-fly tuples have been received
	int sudoTimeOutCount = 0;
	
	//protocols to synchronize with the disk-adaptation and distributed adaptation
	boolean noAdaptation = false;
	boolean inAdaptation = false;
	
	int inCheckingDiskAdaptCounter = 0;

	// this is used when a simulated memory usage cost model being used. To indicate how many tuples in
	// the all queues (input and output, might be some duplications since output of one operator could be the input 
	// of another operator) of all activte operators in the machine. it is set by the distributionexecutioncontroller
	int totalMachineQueueSize = 0;
	double simThreshold = 10.0; // the disk adaptation threshold of simulated model
	
	//Apr-10-2005, to handle window-based join 
	long window_size = 0; //the size of the window of the join operator
	long invalidate_interval = 0; //the time interval that the join operator clean up the out of window states
	// note that for joins that no disk-based adaptation, this could be done each time before the probe. however, 
	// in case partital states have been pushed into disk, this could be inefficient		
	long last_time_invalidated = 0; // initially, it equals to the initialization time
	
	//Apr-12-2005, to handle the global statistics collect and update
	//only the last join operator before the application has the overall information about the linearage of the all join operators
	boolean statistics_collection = false;  // default is false,	
	SymmMwayHashJoinOperatorImp previous_join_reference = null;
	StreamSplitOperatorImp previous_split_reference = null;
	//it may not be a good design to have the execution controller reference in the operator
	//but seems this is the choice for now in order to trace and update the cost information in other machines
	DistributedExecutionController exec_controller_ref = null;
	Random random_var = null;	//to indicate that certain probability of doing updating statistics
	
	double base_unit_size = 2.0; // to indicate the one base unit size, usually it is the number of columns per input stream
	double output_unit_size = 2.0; // number of attributes in the output queue, used in the tracing the intermediate results
	double intermediate_unit_rate = 1.0; //defined as output_unit_size/base_unit_size, in updating, the number of intermediate results is multiple by this
	
	long totalSize; // to record the total state size of the operator, it is updated in the belowAverageMemoryUsage and used in getCurrentStatus
	//double unit_tuple_size = 1.0; // different join operator may have different number of schema size, to indicate the size of a tuple
	// it is initialized in the initialized part, based on the size of the schema?
	// the following is replaced by the base_unit_size 
	/*
	public double getUnitTupleSize(){
		return unit_tuple_size;
	}
	*/
	public double getBaseUnitSize(){
		return base_unit_size;
	}
	public void setReadyToClean(){
		readyToClean = true;
	}
	// it is called in the localAdapter in the 
	public boolean isFinalizationComplete(){
		return completedFinalization;
	}
	// it is called in the localAdapter initialization time
	public void setNumberOfInstances(int n){
		numberOfInstances = n;
	}
	// it is called in the ClassVariableVisitor
	public void turnonStatisticsCollecter(){
		statistics_collection = true;
	}
	//it is called in the localMultiJoinAdapter
	public void setPreviousJoinAndSplit(SymmMwayHashJoinOperatorImp j_ref, StreamSplitOperatorImp s_ref){
		previous_join_reference = j_ref;
		previous_split_reference = s_ref;
	}
	public SymmMwayHashJoinOperatorImp getPreviousJoinReference(){
		return previous_join_reference;
	}
	public StreamSplitOperatorImp getPreviousSplitReference(){
		return previous_split_reference;
	}
	//it is called in the DistributedExecutionController::initExecution
	public void setDistributedExectionControllerRef(DistributedExecutionController ref){
		exec_controller_ref = ref;
	}
	//for using by updatePreviousStats  
	public int[] getJoinAttributePosition(){
		return joinAttributePositions;
	}
	
	public void setWindowSize(long ws){
		window_size = ws;
	}
	public void setInvalidateInterval(long inInt){
		invalidate_interval = inInt;
	}
	
	public void setTotalMachineQueueSize(int qs){
		totalMachineQueueSize = qs;
	}
	
	public boolean isInDiskAdaptation(){
		return inAdaptation;
	}
	public void setInAdaptation(boolean val){
		inAdaptation = val;
	}	
	public void stopDiskAdaptation(){
		noAdaptation = true;
	}
	public void resumeDiskAdaptation(){
		noAdaptation = false;
	}
	public Hashtable getPartitionIDsPushedToDisk(){
		return statesPushedToDisk;
	}
	
	public void setClientName(String name){
		clientName = name;
	}
	
	// to start checking, set check == true, to stop checking, check == false, parts = null;
	public void setStartChecking(boolean check){
		startChecking = check;
	}
	public void setPartitionsToMove(ArrayList toMove){
		deactivatedParts = toMove;
	}
	
	public boolean getDeactivatedCompleteFlag(){
		if (numberOfEOFReceived >= this.getInputQueueCount()){
			//reset the numberOfEOFReceived,
			numberOfEOFReceived = 0;
			return true;
		}
		System.out.println("Now have " + numberOfEOFReceived + "deactivated inputs");
		return false;
	}
	// get the partitions indicated by partsToGet from hashJoinStates, them remove the partitions moved from hashJoinStates
	// make sure the states been moved are all in memory, no disk-residented part are involved. I think this should be controlled
	// in diciding which partitions to be moved
	public ArrayList getPartitionsByID(ArrayList partsToGet){
	    System.out.println("Execute getPartitionsByID in SymmMwayHashJoinOperatorImp");
		ArrayList ret = new ArrayList(partsToGet.size());
		for (int i=0; i<partsToGet.size(); i++){
			Integer idx = (Integer)partsToGet.get(i); 
			HashStatePartitionGroup hp = (HashStatePartitionGroup)hashJoinStates.get(idx);
			ret.add(hp);
			hashJoinStates.remove(idx);
		}
		return ret;
	}
	
	

	
	//set partition groups by ID.
	public boolean setPartitionsByID(ArrayList ids, ArrayList parts) {
		for (int i=0; i<ids.size(); i++){
			Integer idx = (Integer)ids.get(i);
			HashStatePartitionGroup hp = (HashStatePartitionGroup)parts.get(i);
			hashJoinStates.put(idx,hp);
		}
		return true;
	}
	
	// Set a single state partition (not partition group) by ID and state position. 
	// The position of state corresponds to position of input queue.
	public boolean setSinglePartitionByID(int id, int position, WEquiJoinStateImp state) {
		HashStatePartitionGroup group = (HashStatePartitionGroup)hashJoinStates.get(new Integer(id));
		//if adding a new partition group (during load rebalancing)
		if(group == null) {
		    group = new HashStatePartitionGroup(joinAttributePositions,joinInputQueueCells);
		    System.out.println("Create new partition group with ID: " + id + " in " + this);
		}
		
		group.setStatePartition(position, state);
		hashJoinStates.put(new Integer(id), group);
		System.out.println("Add a state partition with ID: " + id + " at position: " + position + " in " + this);	
		
		return true;
	}
	
	public Hashtable getHashJoinStates(){
		return hashJoinStates;
	}
	public void addJoinExpression(ExpressionStr exp){
		joinExpressions.add(exp);
	}
	
	public SymmMwayHashJoinOperatorImp(XATNode node){
		super(node);
		hashJoinStates = new Hashtable();
		statesPushedToDisk = new Hashtable();
		joinExpressions = new ArrayList();
		// need some mechanisms to get compute the joinAttributePosistion & joinInputQueueCells
		addStatistic("JOIN_PARTITION_NUMBER");
		addStatistic("JOIN_PARTITION_SIZE");
	}
	
	public boolean run() throws Exception {
		// build input and output schemas
		for (int i=0; i<this.getInputQueueCount(); i++){
			if (this.getInputQueue(i).getSchema() == null){
				return false;
			}
		}
		if(this.getOutputQueue().getSchema() == null){
			init();
		}
		// build the joinAttributePosition and joinInputQueueCells
		if (!initialized) {			
			
			int queueSizes = this.getInputQueueCount();
			joinAttributePositions = new int[queueSizes];
			joinInputQueueCells = new int[queueSizes];			
			for (int i=0; i<queueSizes; i++){
				// note that length is based on starting point 1, i.e., 5, means 0 ~ 4 have contents
				joinInputQueueCells[i] = this.getInputQueue(i).getSchema().getLength();
			}
			// extract information from joinExpressions
			for (int i=0; i<joinExpressions.size(); i++){
				ExpressionStr exp = (ExpressionStr)joinExpressions.get(i);
				BinCOMPExpressionStr nExp = (BinCOMPExpressionStr)exp;
				TerminalExpressionStr lxp = (TerminalExpressionStr)nExp.getLeft();
				TerminalExpressionStr rxp = (TerminalExpressionStr)nExp.getRight();
				joinAttributePositions[lxp.getIQueuePosition()] = lxp.getIColumnPosition();
				joinAttributePositions[rxp.getIQueuePosition()] = rxp.getIColumnPosition();
			}
			// get the current running directory and compose the baseDirectory
			//File curDir = new File (".");
			//baseStateDirName = curDir.getCanonicalPath() + "\\" + baseStateDirName;
			if (clientName !=null) {				
				baseStateDirName += clientName + "_" + this.operatorID + "_";
			}
			System.out.println("BASE DIRECTORY:" + baseStateDirName);
			initialized = true;
			//System.out.println("INITIALIZATION:+++++++++++++++++++++++++++++++++++++++++"+"QUEUE OWNER:" 
			//		+ this.getOutputQueue().getQueueOwner().toString() + this.getOutputQueue().getQueueName()  + ":" + this.getOutputQueue().getQueueNumber());
			//kind of hard-coded, to unregister unnecessary customers of the output queue.
			//if (this.getOutputQueue().getQueueCustomer().size()>1) {
				//System.out.println("YES, UNREGISTER ONE CUSTOMER." + this.getNode().getParent().getOperator() + "BEFORE:" + this.getOutputQueue().getQueueCustomer().size());					
				//this.getOutputQueue().unregister(this.getNode().getParent().getOperator());
				//System.out.println("AFTER:" + this.getOutputQueue().getQueueCustomer().size());
			//}
			last_time_invalidated = System.currentTimeMillis();
			random_var = new Random(); // for deciding when to updating the cost statistics
			//to compute the unit tuple size
			//unit_tuple_size = this.getOutputQueue().getSchema().getLength() * 0.1; //oldversion
			//unit_tuple_size = this.getOutputQueue().getSchema().getLength()/6; //to make sure the base (3-way-join) is the 1 unit size
			output_unit_size = this.getOutputQueue().getSchema().getLength();
			intermediate_unit_rate = output_unit_size/base_unit_size; // you can set the intermediate_unit_rate ==1.0 to focus on the number of tuples only
		}
		
		XATTuple newObj = null; // to hold the to_be_processed_tuple
		HashStatePartitionGroup newPartition = null; // to the partitions that will hold the tuple newObj
		XATTuple[] matchedTuples = null;	// hold the output of the probe
		Integer currentPartitionID = null; // hold the current tuple partition ID
		
		int tupleCount = 0;
	
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		///debug usage, to check how many tuples in each input queue?
		/*
		String outStr = "";
		for (int dk = 0; dk <this.getInputQueueCount(); dk++){
			outStr += "queue " + dk;
			outStr += ":" + this.getInputQueue(dk).getQueueSize();
		}
		System.out.println(outStr);
		*/
		/// end of debug usage
		/// timeout mechanism for part ID == -1
		if (startChecking) {
			sudoTimeOutCount++;	
		}
		/// end of timeout mechanism for part ID == -1
		
		while (true) {
			if (tupleCount >= howMany) {
				return true;
			}
			
			try {
				int queueSizes = this.getInputQueueCount();
				
				// get one tuple from one input queue at a time and process the join, may not be that efficient? 
				// it could be improved to process batches of tuples at a time from each input queue?
				for (int i=0; i<queueSizes; i++){
					if (this.getInputQueue(i).getQueueSize()>0){
						
						//System.out.println("QUEUE:" + i + "SIZE:" + this.getInputQueue(i).getQueueSize());
						newObj = this.getInputQueue(i).dequeue(this);						
						// check the partition id
						int partID = newObj.getPartitionID();
						if (partID == -1) {
							if (startChecking) {
								//yes, receive one eof tuple from split operator, we do not have to process it
								numberOfEOFReceived ++;
								sudoTimeOutCount = 0;
							} else {
								System.out.println("Wrong, receiving tuples in SymmMwayHashJoinOperator without partition ID.");
							}
						} else if(partID == -10){
							//indicating the end of the streams, used to control the cleanup process of the multi-join operators in the query tree
							numberOfEOFReceived ++;
							//propagate the EOF tuple to all output queues of the operator
							//for(Iterator cursor= this.getOutputQueues(); cursor.hasNext();){
							//	((XATQueue)cursor.next()).enqueue(newObj,this);
							//}
							return true;							
						} else {
							//a timeout mechanism for possible losing tuples with ID==-1
							if (startChecking) {				
								if (sudoTimeOutCount>20){
									numberOfEOFReceived ++;
									sudoTimeOutCount = 0;
									System.out.println("Timeout mechanism activated for this adaptation!");
								}
							}
							//normal join process
							currentPartitionID = new Integer(partID);
							newPartition = (HashStatePartitionGroup)hashJoinStates.get(currentPartitionID);
							if (newPartition != null){
								// simply probe and insert the tuple
								if (window_size>0){
									// window-based join, use purge and probe function
// CHUAN LEI - Incorrect function call
//									matchedTuples = newPartition.purgeProbe(newObj,i,window_size, false);
									matchedTuples = newPartition.purgeProbe(newObj, i, window_size, false, false);
								} else {
									// non-window join
									matchedTuples = newPartition.probe(newObj,i);	
								}								
								int curJoinAttrPos = joinAttributePositions[i];
								Object value = newObj.getCell(curJoinAttrPos).getValue();
								newPartition.insertTuple(value,newObj,i);
								//not sure whether need to reset the hashtable for the new classes, if it is a pure reference, then no, otherwise, yes
								//hashJoinStates.put(new Integer(partID),newPartition);
							} else {
								// create this partition and put them back into the hashJoinStates
								newPartition = new HashStatePartitionGroup(joinAttributePositions,joinInputQueueCells);
								int curJoinAttrPos = joinAttributePositions[i];
								Object value = newObj.getCell(curJoinAttrPos).getValue();
								newPartition.insertTuple(value,newObj,i);
								hashJoinStates.put(new Integer(partID),newPartition);
							}
							tupleCount ++;	
							// System.out.println("Partition ID:" + partID + "::" + newPartition.getTotalNumOfTuples());
							newPartition = null;
						}
						newObj = null;
					} //
					// output the result to the output queue

					if (matchedTuples != null){
						// System.out.println("Output:" + matchedTuples.length);
						// we set the default partition ID, in case that no split operator is defined after the join operator
						// This could happen when the join columns of different joins are the same. However, for join columns 
						// of different joins are not the same, that is, split operator is inserted between joins, this is not
						// needed at all, it can be safely commented out for a littler performance improvement. -- Bin Mar, 2005
						for (int dfp=0; dfp<matchedTuples.length; dfp++){
							matchedTuples[dfp].setPartitionID(currentPartitionID.intValue());
						}
						/// the above could be commented out for split operator is always defined after join
						
						this.getOutputQueue().enqueue(matchedTuples, this);
						
						int fg = random_var.nextInt(10); // [0,k), make sure only trace statistics with a certain probability
						if (fg==1) { //20% Sampling
						
							/*
							//update statistics locally within each join operator
							((HashStatePartition)hashJoinStates.get(currentPartitionID)).addTotalTuplesOutputed(matchedTuples.length);
							*/
						
							// for globally update statistics 
							if (statistics_collection){ //if it is the last join before the application, then simply update 
								//(1) simply update the partitions in the current join operator statistics
								((HashStatePartitionGroup)hashJoinStates.get(currentPartitionID)).addTotalTuplesOutputed(matchedTuples.length);
								//(2) call the updatePreviousOpStats to update the cost statistics recursively
								updatePreviousOpStats(matchedTuples,true,intermediate_unit_rate); //it is finaloutput							
							} else {
								//fg = random_var.nextInt(2);
								if (fg==1) { //20% Sampling, not sure we need to set the sampling the same with the global one
									//if it is not the final output, then update the projected tuple size, as the indicator of the intermediate result
									//(1) simply update the partitions in the current join operator statistics
									((HashStatePartitionGroup)hashJoinStates.get(currentPartitionID)).addTotalProjectedTupleNumber(matchedTuples.length,intermediate_unit_rate);
									//(2) call the updatePreviousOpStats to update the projected cost statistics recursively
									updatePreviousOpStats(matchedTuples,false,intermediate_unit_rate);	//intermediate results
								}
							}						
						}
						
						// based on the previous join and split operator, and recursively update the cost statistics
						matchedTuples = null;
						currentPartitionID = null;
					}
					
				} //end of for loop queueSizes
				
				if (window_size>0){
					cleanInvalidatedStates();
				}
				//String outStr = "SIZE:" + this.getOutputQueue().getQueueSize() + "JOIN QUTPUT QUEUE";
				//for (int pp = 0; pp<(this.getOutputQueue()).getQueueCustomer().size(); pp++){
				//	outStr += pp + ":" + (this.getOutputQueue()).getQueueCustomer().get(pp);
				//}
				//System.out.println(outStr);
				
				//if (startChecking) {
				//	noMoreDeactivatedPars = true;
				//}
				/**
				//debug for testing the available memory, true indicates simulated cost model, false indicates real memory usage model				
				if (belowAverageMemoryUsage(true)){
					
					//System.out.println("free memory vs. totalmemory below" + (1-memoryThreshold) + "!");					
					//we do not want to adapt if number of partitions is too few, note that this is only for real-memory usage adaptation. 
					//if (hashJoinStates.size()>=(prePushPartitionSize)) {
						// show the current memory usage
						Runtime runtime = Runtime.getRuntime();
						//System.out.println("free Memory:" +  runtime.freeMemory() + " Max:" + runtime.maxMemory() + "size:" + hashJoinStates.size());
						if (!noAdaptation) {
							prePushPartitionSize = hashJoinStates.size();
							
							inAdaptation = true; // indicate that current disk-based adaptation is on-going
							
							// (1) compute the partition IDs to be pushed
							ArrayList toBePushed = computePartitionsToPush(0.3,true);						
							// (2) dump the memory resident partitions into disk
							pushPartitionsToDisk(toBePushed);
							
							exec_controller_ref.printToGui("Spill 30% of Operator States\n"); // for VLDB05-Demo, single operator
							
							//dumpAllMemoryStates(); //100% dump, only for extreme case
							
							inAdaptation = false; // indicate the end of disk-based adaptation
							
							// show the current memory usage
							Runtime.getRuntime().gc();
						}
					//} else {
						// in certain case, partitions have been moved to other machines, then current hashJoinStates.size 
						// will never achieve the prePushPartitionSize. Thus, we set a counter to take out of this situation
					//	inCheckingDiskAdaptCounter ++;
					//	if (inCheckingDiskAdaptCounter>=3000){
							// this should be a large enough number to indicate the above situation
					//		inCheckingDiskAdaptCounter = 0;
							// for disk-based adaptation only, there is no decreasing of total partition number, then no need to reset the prePushPartitionSize
					//		prePushPartitionSize = hashJoinStates.size();
					//	}
					//}					
				}
				**/
			} catch(XATQueueException e1){
				// no tuples in the queue? or other unexpected errors?
				return true;
			}			
			
		} //end of while (true)
		
	} //end of run

	/*
	 * disk-based adaptation method, compute partitions to be pushed into disk 
	 * based on certain cost quantifying method 
	 * if max, then choose the partitions with the maximal cost first, instead choose the partitions
	 * with the minimal cost first
	 */
	public ArrayList computePartitionsToPush(double rate, boolean max){
		ArrayList ret = new ArrayList();
		Hashtable curCostTable = new Hashtable();
		
		Enumeration keys = this.hashJoinStates.keys();
		
		// first, compute the cost of each partitions in the operator, put them into costTables
		long totalSize = 0;
		while (keys.hasMoreElements()) { 
			Integer key = (Integer)keys.nextElement();
			int curSize = ((HashStatePartitionGroup)hashJoinStates.get(key)).getTotalNumOfTuples();
			int curOut = ((HashStatePartitionGroup)hashJoinStates.get(key)).getTotalTuplesOutputed();
			//we can plug in other cost caculation methods here
			double cost = 0.0;
			if (curOut >0){
				cost = (double)curSize/((double)curOut); 
			} else {
				cost = (double)curSize;
			}
			totalSize += curSize;
			curCostTable.put(key, new String(""+cost));
		}
		
		// then find the partition IDs having the largest cost so far, put them into the return list,
		// until it reaches the rate
		int retSize = 0;
		double curRate = (double)retSize/(double)totalSize;
		while (curRate <= rate) {
			// find the partition with maximal cost, 
			keys = curCostTable.keys();
			double curCanValue = 0;
			int curCanIndex = 0;
			
			if (keys.hasMoreElements()) {
				Integer index = (Integer)keys.nextElement(); 
				curCanIndex = index.intValue();
				curCanValue = Double.parseDouble((String)curCostTable.get(index));
			} else {
				break;
			}				
			while (keys.hasMoreElements()) {
				Integer key = (Integer)keys.nextElement();
				double value = Double.parseDouble((String)curCostTable.get(key));
				if (max) {
					// choose the max value as the candidate
					if (value > curCanValue) {
						curCanIndex = key.intValue();
						curCanValue = value;
					}
				} else {
					//choose the min value as the candidate
					if (value<curCanValue) {
						curCanIndex = key.intValue();
						curCanValue = value;
					}
				}
			} //end of while
			
			ret.add(new Integer(curCanIndex));
			curCostTable.remove(new Integer(curCanIndex));
			// compute the newrate
			retSize += ((HashStatePartitionGroup)hashJoinStates.get(new Integer(curCanIndex))).getTotalNumOfTuples();
			curRate = (double)(retSize)/(double)totalSize;
		}
		return ret;
	}
	
	/*
	 * push the partitions with IDs in toPush list into the disk.
	 */
	public void pushPartitionsToDisk(ArrayList toPush){
		for (int i=0; i<toPush.size(); i++){
			Integer keyID = (Integer)toPush.get(i);
			HashStatePartitionGroup statepartition = (HashStatePartitionGroup)hashJoinStates.get(keyID);

			Integer indicatorID = (Integer)statesPushedToDisk.get(keyID);
			// put the integer into the hashtable to indicate that the number of partitions have been pushed to disk
			// having the same partition ID. it is used when to retrieve the partitions from disk 
			if (indicatorID == null){
				indicatorID = new Integer(1); 
			} else {
				indicatorID = new Integer(indicatorID.intValue()+1);				
			}
			statesPushedToDisk.put(keyID,indicatorID);
			// write the partition into DISK
			if (writePartitionToDisk(statepartition,keyID, indicatorID)) {
				// yes, remove the hashtables from hashJoinStates
				hashJoinStates.remove(keyID);
			} else {
				// something wrong in write the partition to Disk, we can not remove the hash states from the hashtable
				// and we also need to retrive back the indicatorID
				if (indicatorID.intValue() == 1){
					statesPushedToDisk.remove(keyID);
				} else {
					statesPushedToDisk.put(keyID,new Integer(indicatorID.intValue()-1));	
				}								
			}						
		} //end for
		Runtime runtime = Runtime.getRuntime();
		runtime.gc();
		
		//exec_controller_ref.printToGui("Operator " + this.operatorID +  " Spill Operator States\n"); // for VLDB05-Demo, query tree case
	
	} //end pushPartitionsToDisk
	
	/* the following functions for debug use only, dump all memory states to disk, check the memory usage change
	 * and read tuples from disk back into to resume the process	 
	 */	
	/*
	 * debug usage method, push all states in hashJoinStates into disk 
	 */
	public void dumpAllMemoryStates(){
		Enumeration keyIDs = hashJoinStates.keys();
		while (keyIDs.hasMoreElements()){
			Integer keyID = (Integer)keyIDs.nextElement();
			HashStatePartitionGroup statepartition = (HashStatePartitionGroup)hashJoinStates.get(keyID);

			Integer indicatorID = (Integer)statesPushedToDisk.get(keyID);
			// put the integer into the hashtable to indicate that the number of partitions have been pushed to disk
			// having the same partition ID. it is used when to retrieve the partitions from disk 
			if (indicatorID == null){
				indicatorID = new Integer(1); 
			} else {
				indicatorID = new Integer(indicatorID.intValue()+1);				
			}
			statesPushedToDisk.put(keyID,indicatorID);
			// write the partition into DISK
			if (writePartitionToDisk(statepartition,keyID, indicatorID)) {
				// yes, remove the hashtables from hashJoinStates
				hashJoinStates.remove(keyID);
			} else {
				// something wrong in write the partition to Disk, we can not remove the hash states from the hashtable
				// and we also need to retrive back the indicatorID
				if (indicatorID.intValue() == 1){
					statesPushedToDisk.remove(keyID);
				} else {
					statesPushedToDisk.put(keyID,new Integer(indicatorID.intValue()-1));	
				}								
			}			
		} //end of while keyIDs
		Runtime runtime = Runtime.getRuntime();
		runtime.gc();
	} //end of dumpAllMemoryStates
	
	/* write the states to the Disk with the directory name: baseStateDirName + filename: part
	 * if file already exist, then simply append the states. since the append for ObjectOutputStream does not work, i
	 * instead create directory for each partition ID and have separate files for each time the partition with the same ID,
	 * it is indicator by the indicator 
	 * */
	public boolean writePartitionToDisk(HashStatePartitionGroup states, Integer keyID, Integer indicator){
		String dirName = baseStateDirName + keyID.intValue();
		String fileName = dirName + "/" + indicator.intValue() + ".par";
		//File file = new File(fileName);
		ObjectOutputStream objstream = null;
		FileOutputStream filestream = null; 
		try {
			if (indicator.intValue()==1) {
				//create directory first
				boolean success = (new File(dirName)).mkdir();
				if (!success) {
					System.out.println("Cannot create directory of DISK partitions." + dirName);
					return false;
				}
			}
			filestream = new FileOutputStream(fileName);
			BufferedOutputStream bufferstream = new BufferedOutputStream(filestream);
		    objstream = new ObjectOutputStream(bufferstream);
		    objstream.writeObject(states);
		} catch(IOException ex) {
			System.out.println("Cannot write partitions to Disk.");
			return false;
		} 
		try {
			if (objstream != null) {
				//flush and close "output" and its underlying streams
				objstream.flush();
				objstream.close();					
			}
			filestream.close();
		} catch (IOException ex ){
			System.out.println("Cannot close output stream.");
			return false;
		}
		
		return true;
	} //end of writePartitionToDisk
	
	/*
	 * read all states from files based on the information indicated by statesPushedToDisk
	 * Somehow we need to control the number of output tuples in order not to flood the downstream operators 
	 */
	public boolean readAllDiskStates(boolean finalized){
		Enumeration keyIDs = statesPushedToDisk.keys();
		while (keyIDs.hasMoreElements()){
			Integer keyID = (Integer)keyIDs.nextElement();
			if (readPartitionsFromDisk(keyID, finalized)) {
				statesPushedToDisk.remove(keyID);
			} else {
				System.out.println("Something wrong in read states from Disk?");
				// do nothing for now
			}
			if (this.getOutputQueue().getQueueSize() > 10000) {
				//If too many tuples accumulated in the output queue, then return 
				//false for this current process. Here 5000 is an example
				System.out.println("Has finalized so far: " + totalDiskTuples);
				return false;
			}
		} //end of while keyIDs
		System.out.println("End of Finalization: " + totalDiskTuples);
		return true;		
	} // end of readAllDiskStates
	
	/*
	 * read partitions from disk into hashJoinStates, the following two points need to be considered
	 * (1) if the file has more than one statepartitions, then we need to merge them into one large partitions, 
	 * output will be generated. The state file will be deleted
	 * (2) once the merge has  been done, need to check whether there are more states in the current hashJoinStates,
	 * if yes, then need to merge again and put the final states 
	 */
	public boolean readPartitionsFromDisk(Integer keyID, boolean finalized){

		int numberOfPartitions = ((Integer)statesPushedToDisk.get(keyID)).intValue();
		HashStatePartitionGroup newparts = null;
		    
		if (numberOfPartitions >1) {
			// multiple partitions, we need to read and merge these partitions, using inc. view maintenance algo 
			// output will be generated and sent through output queue directly
			newparts = readOnePartitionFromDisk(keyID,new Integer(1)); // here assuming the first partition is the base one		
			if (newparts==null){
				System.out.println("Wrong when reading the first disk partition of PartitionID:" + keyID.intValue());
				return false;
			}
			newparts.setPartitionID(keyID.intValue()); //set partition id
			
			HashStatePartitionGroup part2 = readOnePartitionFromDisk(keyID,new Integer(2));
			if (part2==null){
				System.out.println("Wrong when reading the second disk partition of PartitionID:" + keyID.intValue());
				return false;
			}
			newparts = mergeTwoPartitions(newparts, part2);
		    for (int i=3; i<=numberOfPartitions; i++){
		    	HashStatePartitionGroup parti = readOnePartitionFromDisk(keyID,new Integer(i));
				if (parti==null){
					System.out.println("Wrong when reading the i-th disk partition of PartitionID:" + keyID.intValue());
					return false;
				}		    	
		    	newparts = mergeTwoPartitions(newparts, parti);
		    }		    
		} else {
			// only one partition exist, simply read and return
			newparts = readOnePartitionFromDisk(keyID,new Integer(1)); // here assuming the first partition is the base one
			if (newparts==null){
				System.out.println("Wrong when reading the first disk partition of PartitionID:" + keyID.intValue());
				return false;
			}
			newparts.setPartitionID(keyID.intValue()); //set partition id
		}
		
		// delete the directory if keyID if all partitions have been sucessfully read
		String dirName = baseStateDirName + keyID.intValue();
		File dir = new File(dirName);
		deleteDirectory(dir);
		
		// process merge the disk partitions with memory partitions
		// check whether there is memory resident states with same partition ID, if yes, then merge them 
		HashStatePartitionGroup memPart = (HashStatePartitionGroup)hashJoinStates.get(keyID);
		if (memPart != null) {
			newparts = mergeTwoPartitions(newparts, memPart);
		}
		if (finalized) {
			// we do not have to put states in memory if it is finalized processing, instead we could 
			// even remove the memory resident one
			hashJoinStates.remove(keyID);
		} else {
			hashJoinStates.put(keyID,newparts);	
		}
		return true;
	}
	/*
	 * It read one partition from disk, it is specified by partitionID, and the number indicates which part of the partition
	 * called by  readPartitionsFromDisk()
	 */
	public HashStatePartitionGroup readOnePartitionFromDisk(Integer keyID, Integer indicator){
		HashStatePartitionGroup ret = null;
		String dirName = baseStateDirName + keyID.intValue();		
		String fileName = dirName + "/" + indicator.intValue() + ".par";
			
		ObjectInputStream objstream = null;
		FileInputStream filestream = null; 

		try {			
			filestream = new FileInputStream(fileName);
			BufferedInputStream bufferstream = new BufferedInputStream(filestream);
		    objstream = new ObjectInputStream(bufferstream);

		    ret = (HashStatePartitionGroup)objstream.readObject(); // here assuming the first partition is the base one
		    
		} catch(IOException ex) {
			System.out.println("Cannot read partitions from Disk." + ex.getMessage());
			return null;
		} catch(ClassNotFoundException exc) {
			System.out.println("Cannot read partitions from Disk:ClassNotFoundException.");
			return null;			
		}
		finally {
			try {
				if (objstream != null) {
					//flush and close "output" and its underlying streams
					objstream.close();
				}
				// File deletefile = new File(fileName);
				// deletefile.delete();				
			} catch (IOException ex ){
				System.out.println("Cannot close & delete file.");
				return null;
			}
		}
		return ret;
	}
	
	public boolean deleteDirectory(File path) {
		if( path.exists() ) {
			File[] files = path.listFiles();
			for(int i=0; i<files.length; i++) {
				if(files[i].isDirectory()) {
					deleteDirectory(files[i]);
				}
				else {
					files[i].delete();
				}
			}
		}
		return( path.delete() );
	}

	/*
	 * Merge two partitions with the same partition ID. this is used if multiple partitions with the same ID have
	 * been pushed to disk. It apply incremental view maintenance algorithm to compute the results from these two
	 * partitions that have not been generated
	 */
	public HashStatePartitionGroup mergeTwoPartitions(HashStatePartitionGroup basePart, HashStatePartitionGroup incPart){
		
		WEquiJoinStateImp[] incStates = incPart.getHashStates(); // all states of the inc partitions
		ArrayList allDeltaTuples = new ArrayList();
		ArrayList allIncTuples = new ArrayList();
		
		for (int i=0; i<incStates.length; i++){
			Hashtable hashtableOfStatesi = incStates[i].getAllStateTuples();
			Enumeration keys = hashtableOfStatesi.keys();
			while (keys.hasMoreElements()){
				Object key = keys.nextElement();
				ArrayList bucket = (ArrayList)hashtableOfStatesi.get(key);
				if (bucket!=null){
					for (int k=0; k<bucket.size(); k++){
						XATTuple currTp = (XATTuple)bucket.get(k);
						// probe the basePart, to compute the results due to basePart
						XATTuple[] theMatched = basePart.probe(currTp,i);
						if (theMatched !=null){
							for (int t=0; t<theMatched.length; t++){
								allDeltaTuples.add(theMatched[t]);
							}
							theMatched = null;
						}
						// then insert the tuple into basePart, to advance the basePart states
						basePart.insertTuple(key,currTp,i);
						///////////// now, at the same time, we need to calculate the tuples that can be generated from incPart only
						theMatched = incPart.probe(currTp,i);
						if (theMatched !=null){
							for (int t=0; t<theMatched.length; t++){
								allIncTuples.add(theMatched[t]);
							}
							theMatched = null;
						}
						////////////
					}
				}
			} //end while
		} /// end for incStates
		
		/// compute the difference allDeltaTuples - allIncTuples, the implementation may not be that efficient?
		allDeltaTuples = computeDifference(allDeltaTuples,allIncTuples);
		//for (int i=0; i<allIncTuples.size(); i++){
		//	XATTuple curTp = (XATTuple)allIncTuples.get(i);
		//	if (allDeltaTuples.contains(curTp)) {
		//		allDeltaTuples.remove(curTp);
		//	}
		//}
		
		//// output the delta changes due to partition merge
		if (allDeltaTuples.size()>0){
			XATTuple[] finalMatched = (XATTuple[])allDeltaTuples.toArray(new XATTuple[]{});
			if (finalMatched !=null){
				this.getOutputQueue().enqueue(finalMatched, this);
				totalDiskTuples += finalMatched.length;
				finalMatched = null;
			}			
		}
		// return the merged partitions
		
		return basePart;
		
	} //end of mergeTwoPartitions
	
	/*
	 * Compute the whole-minus, seems the arraylist.contain does not work well for this case.
	 * basically, remove tuples from whole that exist in minus
	 */
	public ArrayList computeDifference(ArrayList whole, ArrayList minus){
		ArrayList difference = new ArrayList();
		for (int i=0; i<whole.size(); i++){
			XATTuple curTP = (XATTuple)whole.get(i);
			if (!contains(minus,curTP)) {
				difference.add(curTP);
			}
		}
		return difference;
	}
	public boolean contains (ArrayList src, XATTuple tp){
		for (int i=0; i<src.size(); i++){
			XATTuple tmpTP = (XATTuple)src.get(i);
			// need to compare each cell whether it is the same
			if (tmpTP.equals(tp)){
				return true;
			}
		}
		return false;
	}
	
	/* check the memory usage based on the runtime information, it may not reflect the actual system memory
	 * available information, but it can be used as an indicator. especially the total amount of main memory one
	 * JVM can be used is specified by parameter -mx
	 * the flag simulated = true indicated that we use the simulated cost model instead the real main memory usage
	 * */
	public boolean belowAverageMemoryUsage(boolean simulated){
		if (simulated){
			// (1) check the state size of the operator
			Enumeration keys = this.hashJoinStates.keys();			
			totalSize = 0;
			while (keys.hasMoreElements()) { 
				Integer key = (Integer)keys.nextElement();
				totalSize += ((HashStatePartitionGroup)hashJoinStates.get(key)).getTotalNumOfTuples();
			}
			double simulatedCost = (totalSize + totalMachineQueueSize)/1024.0;
			if (simulatedCost>simThreshold) {
				System.out.println("SIMULATED COST MODEL: Yes, it is over " + simThreshold);
				return true;
			}
			return false;
		} else {
			Runtime runtime = Runtime.getRuntime();
			long usedmemory = runtime.totalMemory()-runtime.freeMemory();
			long maxmemory = runtime.maxMemory();
			// add one short cut here, only half of the maxiaml memory is used in determine the 
			// memoryusage percentage, this case, thus, to avoid the memory overflow of  the
			// finalization stage 
			double memoryusage = ((double)usedmemory*2)/(double)maxmemory;
			if (memoryusage > memoryThreshold) {
					System.out.println("REALMEMORY USAGE. Yes, it is over THRESHOLD " + memoryThreshold);
					return true;
			}
			return false;
		}		
	} //end of checkMemoryUsage
	// new version of the finalization-code -- Apr 14, 2005
	public boolean finalizeOperation(){
		//System.out.println("Yes, Come to the end." + this.operatorID);
		if (completedFinalization) {
			//System.out.println("Yes, Completed Finalization Stage of Operator:" + this.operatorID); 
			return true;
		} 
		if (previous_join_reference!=null){
			//this indicate there are joins below the current join operator have not been finalized.
			//thus, simply run the query operator
			return cleanUpRun();
		}
		if (!readyToClean){
			return cleanUpRun();
		}
		boolean res = readAllDiskStates(true);
		if (res) {
			// End of finalization, then send a EOF tuple with partition ID = -10 
			XATTuple EOFTP = new XATTupleImp();
			EOFTP.setPartitionID(-10);
			this.getOutputQueue().enqueue(EOFTP, this);
			completedFinalization = true;
			//notifiy the query processor to modifiy the previous_join_reference
			if (exec_controller_ref != null){
				exec_controller_ref.updateJoinReferences(this.operatorID);
			}
		}
		return res;
	}
	
	/* in cleaning up the query tree with multiple join operators, we need to keep on processing
	 * join with the data generated from the previous join operators, however, if the inputs are from
	 * source streams or other sources, we simply discard them -- @ Apr 14, 2005
	 * It is called by the finalizationOperation
	*/
	public boolean cleanUpRun(){
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		int tupleCount = 0;
		if (numberOfEOFReceived >= numberOfInstances){
			readyToClean = true;
			return true;
		}
		// a timeout mechanism for possible losing tuples with ID==-1
		//if (sudoTimeOutCount>20){
		//	numberOfEOFReceived ++;
		//	sudoTimeOutCount = 0;
		//	System.out.println("Timeout mechanism activated for this adaptation!");
		//}
		while (true){
			if (tupleCount>=howMany){
				return true;				
			}			
			int queueSizes = this.getInputQueueCount();
			XATTuple[] matchedTuples = null;
			Integer currentPartitionID = null;
			boolean allEmpty = true;
			for (int i=0; i<queueSizes; i++){
				if ((this.getInputQueue(i).isFinalQueue()) && (this.getInputQueue(i).size() >0)) {
					allEmpty = false;
					break;
				}
			}
			if (allEmpty){
				sudoTimeOutCount++;
				return true;
			}
			try {
				for (int i=0; i<queueSizes; i++){					
					if (!this.getInputQueue(i).isFinalQueue()) {
						// simply clean all the tuples in the queue, do not have to process them
						this.getInputQueue(i).cleanTuples();
					} else {
						if (this.getInputQueue(i).getQueueSize()>0){
							XATTuple newObj = this.getInputQueue(i).dequeue(this);						
							// check the partition id
							int partID = newObj.getPartitionID();
							if (partID == -10) {
								//yes, receive one eof tuple from split operator, we do not have to process it
								numberOfEOFReceived ++;
								sudoTimeOutCount = 0;
							} else {
								// normal join process
								currentPartitionID = new Integer(partID);
								HashStatePartitionGroup newPartition = (HashStatePartitionGroup)hashJoinStates.get(currentPartitionID);
								if (newPartition != null){
									// simply probe and insert the tuple
									if (window_size>0){
										// window-based join, use purge and probe function
// CHUAN LEI - Incorrect function call
//										matchedTuples = newPartition.purgeProbe(newObj,i,window_size, false);
										matchedTuples = newPartition.purgeProbe(newObj, i, window_size, false, false);
										
									} else {
										// non-window join
										matchedTuples = newPartition.probe(newObj,i);	
									}								
									int curJoinAttrPos = joinAttributePositions[i];
									Object value = newObj.getCell(curJoinAttrPos).getValue();
									newPartition.insertTuple(value,newObj,i);
									//not sure whether need to reset the hashtable for the new classes, if it is a pure reference, then no, otherwise, yes
									//hashJoinStates.put(new Integer(partID),newPartition);
								} else {
									// create this partition and put them back into the hashJoinStates
									newPartition = new HashStatePartitionGroup(joinAttributePositions,joinInputQueueCells);
									int curJoinAttrPos = joinAttributePositions[i];
									Object value = newObj.getCell(curJoinAttrPos).getValue();
									newPartition.insertTuple(value,newObj,i);
									hashJoinStates.put(new Integer(partID),newPartition);
								}
								tupleCount ++;	
								newPartition = null;
							}
							newObj = null;
						} //
						// output the result to the output queue
						if (matchedTuples != null){
							// System.out.println("Output:" + matchedTuples.length);
							// we set the default partition ID, in case that no split operator is defined after the join operator
							// This could happen when the join columns of different joins are the same. However, for join columns 
							// of different joins are not the same, that is, split operator is inserted between joins, this is not
							// needed at all, it can be safely commented out for a littler performance improvement. -- Bin Mar, 2005
							for (int dfp=0; dfp<matchedTuples.length; dfp++){
								matchedTuples[dfp].setPartitionID(currentPartitionID.intValue());
							}
							/// the above could be commented out for split operator is always defined after join						
							this.getOutputQueue().enqueue(matchedTuples, this);
							matchedTuples = null;
							currentPartitionID = null;
						}					
					} 
				}//end of for loop queueSizes
			} catch(XATQueueException e1){
				// no tuples in the queue? or other unexpected errors?
				return true;
			}			
		}
	}
	/*
	 * if it is window-based join, if the current time - last_invalidated_time > invalidate_interval
	 * then, clean up the invalidated states, if the current states have been pushed in disks, then
	 * these states also have to be pushed into disk, otherwise, they are deleted directly. 
	 */
	public void cleanInvalidatedStates(){
		if (System.currentTimeMillis() - last_time_invalidated > invalidate_interval){
			Enumeration keys = hashJoinStates.keys();
			while (keys.hasMoreElements()){
				Integer key = (Integer)keys.nextElement();
				if (statesPushedToDisk.get(key) == null){
					//directly purge the invalidated states,
					//System.out.println("Real Purge Partition ID:" + key);
					((HashStatePartitionGroup)hashJoinStates.get(key)).purgeState();
				} else {
					//push the invalidated states into the disk
					//(1) first get the indicator id
					Integer indicatorID = (Integer)statesPushedToDisk.get(key);
					// put the integer into the hashtable to indicate that the number of partitions have been pushed to disk
					// having the same partition ID. it is used when to retrieve the partitions from disk 
					if (indicatorID == null){
						indicatorID = new Integer(1); 
					} else {
						indicatorID = new Integer(indicatorID.intValue()+1);				
					}
					statesPushedToDisk.put(key,indicatorID);
					
					HashStatePartitionGroup in_hash_part = ((HashStatePartitionGroup)hashJoinStates.get(key)).getInvalidatedStates(joinAttributePositions,joinInputQueueCells);
					// (2) write the partition into DISK
					if (writePartitionToDisk(in_hash_part,key, indicatorID)) {
						// yes, purge the invalidated hashtables from hashJoinStates
						((HashStatePartitionGroup)hashJoinStates.get(key)).purgeState();				
					} else {
						// something wrong in write the partition to Disk, we can not remove the hash states from the hashtable
						// and we also need to retrive back the indicatorID
						if (indicatorID.intValue() == 1){
							statesPushedToDisk.remove(key);
						} else {
							statesPushedToDisk.put(key,new Integer(indicatorID.intValue()-1));	
						}								
					}						
					in_hash_part = null;
				} //end of pushing states and then purging
			} //end while keys			
			last_time_invalidated = System.currentTimeMillis();
		} //end if
	} //end of invalidateStates
	
	/*
	 * Recursively update the previous join operator, the number of output tuples have been generated
	 * here, the intermediate_unit_rate is used in tracing the intermediate results, that is, when finalOutput == false
	 * it is multipled by the number of intermediate results to estimate the memory usage
	 */
	public void updatePreviousOpStats(XATTuple[] outputTPs, boolean finalOutput, double inter_rate){
		SymmMwayHashJoinOperatorImp loop_join_ref = this.previous_join_reference;
		StreamSplitOperatorImp loop_split_ref = this.previous_split_reference;
		
		while ((loop_join_ref != null)&&(loop_split_ref !=null)){
			//it should be enough to always get the first join column, since partition function in the same join should be the same
			int[] cur_join_attrs = loop_join_ref.getJoinAttributePosition();
			int join_column = cur_join_attrs[0];
			//(1) Figure out the partitionIDs:number of Outputs correspondence
			Hashtable partID_outputN = new Hashtable();
			for (int i=0; i<outputTPs.length; i++){
				if (loop_split_ref.nullPartitionFunction()) {
					// if the split operator in not activated, then the partition function is null, we just reload the partition function
					loop_split_ref.reloadPartitionFunction();
				}
				int partitionID = loop_split_ref.partitionTuple(outputTPs[i],join_column);
				Integer pID = new Integer(partitionID);
				Integer curN = (Integer)partID_outputN.get(pID); 
				if (curN == null){
					partID_outputN.put(pID, new Integer(1));
				} else {
					partID_outputN.put(pID, new Integer(curN.intValue()+1));
				}
			}//end for outputTPs
			//(2) set or popagate the statistics information
			//(2.1) update the partitions that within the current operators, for those are not in the current machine
			//then lookup the partitionMap of the split operator to send out the update information
			Enumeration idKeys = partID_outputN.keys();
			while (idKeys.hasMoreElements()){
				Integer idKey = (Integer)idKeys.nextElement();				
				HashStatePartitionGroup part_ref = (HashStatePartitionGroup)loop_join_ref.getHashJoinStates().get(idKey);
				if (part_ref != null) {
					//update the cost information//
					if (finalOutput){
						part_ref.addTotalTuplesOutputed(((Integer)partID_outputN.get(idKey)).intValue());
					} else {
						part_ref.addTotalProjectedTupleNumber(((Integer)partID_outputN.get(idKey)).intValue(),inter_rate);
					}
					partID_outputN.remove(idKey);
				} else {
					//some cases, when the partitions have been pushed into the disk, then the part_ref is null, 
					//we then simply discard this cost information, we check the IDs that have been pushed 
					if (loop_join_ref.getPartitionIDsPushedToDisk().get(idKey)!=null){
						partID_outputN.remove(idKey);
					}
				}
			}
			//(2.2)now, all the IDs that left in the partID_outputN are those in the other machines
			//organize the IDs based on the machines, based on the partitionMap information
			idKeys = partID_outputN.keys();	
			if (idKeys.hasMoreElements()){
				Hashtable pMapTable = loop_split_ref.getPartitionMappingTable();
				Hashtable idMachineMap = new Hashtable();
				while (idKeys.hasMoreElements()){
					Integer idKey = (Integer)idKeys.nextElement();
					String machineIP = (String)pMapTable.get(idKey);
					ArrayList idList = (ArrayList)idMachineMap.get(machineIP);
					if (idList == null) {
						idList = new ArrayList();
					}
					idList.add(idKey);
					idMachineMap.put(machineIP,idList);
				}
				//call the corresponding distributedexecutioncontroller to send out the updated cost information
				if (exec_controller_ref != null){
					exec_controller_ref.propagatePartitionStats(idMachineMap,partID_outputN,loop_join_ref.operatorID,finalOutput,inter_rate);
				}				
			}
			//recursively call the if the current loop_join_ref has the next join operator connected to it
			loop_split_ref = loop_join_ref.getPreviousSplitReference();
			loop_join_ref = loop_join_ref.getPreviousJoinReference();
		} //end of while next_join_ref and next_split_ref
	} //end of updatePreviousOpStats

	/* used for VLDB05-Demo in query tree case, the totalsize of the operator is set from localMultiJoinTreeAdapter
	 * for displaying into the GUI
	 */
	public void setTotalSize(long ts) {
		totalSize = ts;
	}
	/* overwrite the code in the XATStreamOperatorImp. For gui to display the operator
	 * information in the Operators tab -- Aug 10, 2005
	 */
	public String getCurrentStatus() {
		String statsString = "OID:" + this.operatorID + "State Size:" + totalSize
							+ "Total Partitions:" + hashJoinStates.size() + "\n";		
		/*
		if (statusBuffer.length() >0) {
			statsString += statusBuffer.toString();
			statusBuffer.delete(0,statusBuffer.length());
		}
		*/	
		return statsString;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAccumulatedTuples(long)
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAllAccumulatedTuples()
	 */
	public boolean cleanAllAccumulatedTuples() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#matchStates(edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator)
	 */
	public boolean matchStates(XATStreamOperator op) {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#hasUnmatchedState()
	 */
	public boolean hasUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#recomputeUnmatchedState()
	 */
	public boolean recomputeUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#setStateID(int)
	 */
	public int setStateID(int startID) {
		// TODO Auto-generated method stub
		return 0;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		// TODO Auto-generated method stub
		return null;
	}
	
	//set WEquiJoinStates by IDs and position.
	public boolean setStatesByID(ArrayList ids, int position, ArrayList parts) {
		for (int i=0; i<ids.size(); i++){
			Integer idx = (Integer)ids.get(i);
			HashStatePartitionGroup hp = (HashStatePartitionGroup)parts.get(i);
			hp.setStatePartition(position, (WEquiJoinStateImp)parts.get(i));
			hashJoinStates.put(idx,hp);
		}
		return true;
	}
	
    /**
     * Recompute partition groups with the given IDs, return results for each partition group. 
     * The join results from each partition group form a state partition, which is then added 
     * to the parent operator during load balancing. Used in DistributedExecutionController.receivePartitionsInTree().
     * @param IDs
     * @return A list of partitions.
     */
    public ArrayList joinPartitionsByIDs(ArrayList IDs) {
        ArrayList newstates = new ArrayList();
        
        //join partitions in each partition group with unique partition ID.
        for(int idindex=0; idindex<IDs.size(); idindex++) {
            WEquiJoinStateImp newstate = new WEquiJoinStateImp();
            XATTuple[] matchedTuples;
    		Integer currentPartitionID = (Integer)IDs.get(idindex);
    		HashStatePartitionGroup partitionGroup = (HashStatePartitionGroup)hashJoinStates.get(currentPartitionID);
    		if (partitionGroup != null) {
    		    //get tuples in the first state of the partition group
    		    
    		    WEquiJoinStateImp firststate = partitionGroup.getHashStates()[0];
    		    int firststatesize = firststate.getStateSize();
    		    Enumeration firstelements = firststate.getAllStateTuples().elements();
    		    while(firstelements.hasMoreElements()) {
    		        XATTuple newTuple = (XATTuple)firstelements.nextElement();
    		        
    				// simply probe and insert the tuple
    				if (window_size>0){
    					// window-based join, use purge and probe function
// CHUAN LEI - Incorrect function call
//    					matchedTuples = partitionGroup.purgeProbe(newTuple, 0, window_size, false);
    					matchedTuples = partitionGroup.purgeProbe(newTuple, 0, window_size, false, false);    				
    				} else {
    					// non-window join
    					matchedTuples = partitionGroup.probe(newTuple, 0);	
    				}
    				
    				for(int rsize = 0; rsize<matchedTuples.length; rsize++) {
    				    Object value = newTuple.getCell(0).getValue();
    				    newstate.insertTuple(value, matchedTuples[rsize]);								
    				}
    				matchedTuples = null;
    		    }
    		}
    		newstates.add(newstate);
        }
        return newstates;
    }
}
