/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate;

import java.util.List;
import java.util.Vector;
import java.util.ArrayList;

//import streamgenerator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import java.io.Serializable;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;

/**
 * The state for NLJ algorithm pipelined sliding window join. Creation date:
 * (10/05/2005 5:34:51 PM)
 * 
 * @author: Song Wang
 */
public class PipeWFIFOState extends WOperatorState implements Serializable {

	//move some local variable out for fast execution.
	List purgedStates = new ArrayList();
	ArrayList joinedtuples = new ArrayList();
	List tuplepair = new Vector();
	List newvalues = new Vector();
	
	private long startWindow = 0;//use to keep the start time of the window

	private long endWindow = 0;//use to keep the end time of the window

	//private long timeWindow = 0;//use to keep the end-start time of the window
	
	public long previousDelayedPurgeTimestamp = 0;

	//input and output functions for the start and end window
	public void setStartWindow(long start) {
		startWindow = start;
	}

	public void setEndWindow(long end) {
		endWindow = end;
	}

	public long getStartWindow() {
		return startWindow;
	}

	public long getEndWindow() {
		return endWindow;
	}

//	public void setTimeWindow(long newWindow) {
//		timeWindow = newWindow;
//	}

	protected ArrayList statetuples = new ArrayList();//Array holding the states.

	public int getStateSize() {
		return statetuples.size();
	}

	public ArrayList getStateTuples() {
		return statetuples;
	}

	/**
	 * PipeWFIFOState constructor comment.
	 */
	public PipeWFIFOState() {
		super();
	}

	public PipeWFIFOState(XATStreamOperator op) {
		super();
		this.setOperator(op);
	}

	public void init(int stateID){
		this.stateID = stateID;
	}
	/**
	 * Insert of tuples into the statetuple arraylist.
	 */

	public void insertTuple(XATTuple newTuple) {
		statetuples.add(newTuple);
		totalNumTuples++;
	}

	public void insertTuple(XATTuple[] newTuples) {
		for (int i = 0; i < newTuples.length; i++) {
			statetuples.add(newTuples[i]);
		}
		totalNumTuples = totalNumTuples + newTuples.length;
	}

	public void insertTuple(ArrayList newTuples) {
		for (int i = 0; i < newTuples.size(); i++) {
			statetuples.add(newTuples.get(i));
		}
		totalNumTuples = totalNumTuples + newTuples.size();
	}

	/**
	 * @author Abhishek Mukherji
	 * Overwritten expire checking function using the endWindow and startWindow
	 * In case of cross-purge, we can only check that if the diff > endWindow.
	 * In case of self-purge, we need check that if the diff >
	 * endWindow-startWindow.
	 */
	public boolean timeStampExpiredSelf(long expireTimeStamp, XATTuple statetuple) {
		/**
		 * if tuple1 and tuple2 is out of window size from each other, return
		 * false
		 */
		//self-purge checking
		if (expireTimeStamp - statetuple.getMinTimeStamp() > endWindow
				- startWindow)
			return true;

		return false;
	}
	
	public boolean timeStampExpiredSelf(XATTuple newtuple, XATTuple statetuple) {
		/**
		 * if tuple1 and tuple2 is out of window size from each other, return
		 * false
		 */
		//self-purge checking
		if (newtuple.getMaxTimeStamp() - statetuple.getMinTimeStamp() > endWindow
				- startWindow)
			return true;

		return false;
	}

	/*** @author Abhishek Mukherji
	 * Purge Function with the purged ArrayList returned for self-purge only
	 * purge by timeStamp, used for hopping windows
	 */
	public List purgeByTimeStampSelf(long purgeIncrement) {
		//List purgedStates = new ArrayList();
		purgedStates.clear();
		
		int count = 0;

		/**
		 * purge a tuple if its out of window bound, stop when meet the first
		 * tuple in the state that cannot be purged here we assume that the
		 * state is ordered by the timestamps
		 */
		for (int i = 0; i < statetuples.size(); i++) {
			if (timeStampExpiredSelf(purgeIncrement, (XATTuple) statetuples.get(i))) {
				//System.out.println("purge a tuple from state");
				count++;
			} else {
				break;
			}
		}
		if (count > 0) {
			//System.out.println("i = " + i + " count = " + count);
			purgedStates.addAll(statetuples.subList(0, count));
			statetuples.subList(0, count).clear();
		}

		//if (count > 0) {
		//System.out.println("state size before purge = " +
		// statetuples.size());
		//statetuples.subList(0, count).clear();
		//System.out.println("number of tuples purged = " + count);
		//System.out.println("state size after purge = " +
		// statetuples.size());
		//}
		return purgedStates;
	}
	
	/**
	 * Purge Function with the purged ArrayList returned for self-purge only
	 */
	public List purgeByTimeStampPipeSelf(XATTuple newTuple) {
		//List purgedStates = new ArrayList();
		purgedStates.clear();
		
		if (newTuple == null) {
			return purgedStates;
		}

		int count = 0;

		/**
		 * purge a tuple if its out of window bound, stop when meet the first
		 * tuple in the state that cannot be purged here we assume that the
		 * state is ordered by the timestamps
		 */
		for (int i = 0; i < statetuples.size(); i++) {
			if (timeStampExpiredSelf(newTuple, (XATTuple) statetuples.get(i))) {
				//System.out.println("purge a tuple from state");
				count++;
			} else {
				break;
			}
		}
		if (count > 0) {
			//System.out.println("i = " + i + " count = " + count);
			purgedStates.addAll(statetuples.subList(0, count));
			statetuples.subList(0, count).clear();
		}

		//if (count > 0) {
		//System.out.println("state size before purge = " +
		// statetuples.size());
		//statetuples.subList(0, count).clear();
		//System.out.println("number of tuples purged = " + count);
		//System.out.println("state size after purge = " +
		// statetuples.size());
		//}
		return purgedStates;
	}
	
	/**
	 * negative tuple expiring a tuple
	 * @author Abhishek Mukherji
	 */
	public boolean NegativeTupleExpiredSelf(XATTuple negtuple, XATTuple statetuple) {
		/**
		 * if tuple1 and tuple2 match they are the same
		 */
		//self-purge checking
		if (negtuple.getMaxTimeStamp()== statetuple.getMaxTimeStamp() && negtuple.getMinTimeStamp()== statetuple.getMinTimeStamp() || statetuple.equals(negtuple))
			return true;

		return false;
	}	
	

	/**
	 * Overwritten expire checking function using the endWindow and startWindow
	 * In case of cross-purge, we can only check that if the diff > endWindow.
	 * The reason is that the diff of probing and been probed tuple stamps 
	 * will > endWindow-startWindow for 2nd and later pipejoins.
	 * In case of self-purge, we need check that if the diff >
	 * endWindow-startWindow.
	 */
	public boolean timeStampExpiredCross(XATTuple newtuple, XATTuple statetuple) {
		/**
		 * if tuple1 and tuple2 is out of window size from each other, return
		 * false
		 */
		
		if (newtuple.getMaxTimeStamp() - statetuple.getMaxTimeStamp() > endWindow)
			return true;

		return false;
	}

	/**
	 * Purge Function with the purged ArrayList returned for cross-purge only
	 */
	public List delayedPurgeByTuplePipeCross(XATTuple newTuple) {
		//ArrayList purgedStates = new ArrayList();
		//List purgedStates = new ArrayList();
		purgedStates.clear();
		
		if (newTuple == null) {
			return purgedStates;
		}

		int count = 0;
		int count2=0;
		/**
		 * purge a tuple if its out of window bound, stop when meet the first
		 * tuple in the state that cannot be purged here we assume that the
		 * state is ordered by the timestamps
		 */
		for (int i = 0; i < statetuples.size(); i++) {
			if (previousDelayPurged((XATTuple) statetuples.get(i))) {
				count2++;
				continue;
			}
			if (timeStampExpiredCross(newTuple, (XATTuple) statetuples.get(i))) {
				//System.out.println("purge a tuple from state");
				count++;
			} else {
				break;
			}
		}
		if (count > 0) {
			//System.out.println("i = " + i + " count = " + count);
			purgedStates.addAll(statetuples.subList(count2, count+count2));
			//statetuples.subList(0, count).clear(); this is different with the purgeByTuplePipeCross(XATTuple)
		}

		//if (count > 0) {
		//System.out.println("state size before purge = " +
		// statetuples.size());
		//statetuples.subList(0, count).clear();
		//System.out.println("number of tuples purged = " + count);
		//System.out.println("state size after purge = " +
		// statetuples.size());
		//}
		this.previousDelayedPurgeTimestamp = newTuple.getMinTimeStamp();
		return purgedStates;
	}
	private boolean previousDelayPurged(XATTuple statetuple){
		if (this.previousDelayedPurgeTimestamp - statetuple.getMaxTimeStamp() > endWindow) return true;
		return false;
	}
	/**
	 * Purge Function with the purged ArrayList returned for cross-purge only
	 */
	public List purgeByTuplePipeCross(XATTuple newTuple) {
		//ArrayList purgedStates = new ArrayList();
		//List purgedStates = new ArrayList();
		purgedStates.clear();
		
		if (newTuple == null) {
			return purgedStates;
		}

		int count = 0;

		/**
		 * purge a tuple if its out of window bound, stop when meet the first
		 * tuple in the state that cannot be purged here we assume that the
		 * state is ordered by the timestamps
		 */
		for (int i = 0; i < statetuples.size(); i++) {
			if (timeStampExpiredCross(newTuple, (XATTuple) statetuples.get(i))) {
				//System.out.println("purge a tuple from state");
				count++;
			} else {
				break;
			}
		}
		if (count > 0) {
			//System.out.println("i = " + i + " count = " + count);
			purgedStates.addAll(statetuples.subList(0, count));
			statetuples.subList(0, count).clear();
		}

		//if (count > 0) {
		//System.out.println("state size before purge = " +
		// statetuples.size());
		//statetuples.subList(0, count).clear();
		//System.out.println("number of tuples purged = " + count);
		//System.out.println("state size after purge = " +
		// statetuples.size());
		//}
		
		return purgedStates;
	}
	
	/**
	 * early purge due to the negative tuples
	 * @author Abhishek Mukherji
	 */
	
	public List purgeByNegativeTupleSelf(XATTupleNegativeImp negTuple) {
		//List purgedStates = new ArrayList();
		purgedStates.clear();
		
		if (negTuple == null) {
			return purgedStates;
		}

		int count = 0;

		/**
		 * purge a tuple if it matches the negative tuple
		 */
		for (int i = 0; i < statetuples.size(); i++) {
			if (NegativeTupleExpiredSelf(negTuple, (XATTuple) statetuples.get(i))) {
				statetuples.remove(i);
				
				//System.out.println("purge a tuple from state");
				count++;
			} else {
				break;
			}
		}
		

		//if (count > 0) {
		//System.out.println("state size before purge = " +
		// statetuples.size());
		//statetuples.subList(0, count).clear();
		//System.out.println("number of tuples purged = " + count);
		//System.out.println("state size after purge = " +
		// statetuples.size());
		//}
		return purgedStates;
	}

	/**
	 * Purge Function with the purged ArrayList returned for self-purge only
	 */
	public List purgeByTuplePipeSelf(XATTuple newTuple) {
		//List purgedStates = new ArrayList();
		purgedStates.clear();
		
		if (newTuple == null) {
			return purgedStates;
		}

		int count = 0;

		/**
		 * purge a tuple if its out of window bound, stop when meet the first
		 * tuple in the state that cannot be purged here we assume that the
		 * state is ordered by the timestamps
		 */
		for (int i = 0; i < statetuples.size(); i++) {
			if (timeStampExpiredSelf(newTuple, (XATTuple) statetuples.get(i))) {
				//System.out.println("purge a tuple from state");
				count++;
			} else {
				break;
			}
		}
		if (count > 0) {
			//System.out.println("i = " + i + " count = " + count);
			purgedStates.addAll(statetuples.subList(0, count));
			statetuples.subList(0, count).clear();
		}

		//if (count > 0) {
		//System.out.println("state size before purge = " +
		// statetuples.size());
		//statetuples.subList(0, count).clear();
		//System.out.println("number of tuples purged = " + count);
		//System.out.println("state size after purge = " +
		// statetuples.size());
		//}
		return purgedStates;
	}

	/**
	 * probe a tuple against the statetuples. It must be called after the
	 * purgeByTuplePipe() function with the same tuple. state is ordered by the
	 * timestamps
	 */
	public XATTuple[] probe(XATTuple tuple, ExpressionStr expression,
			int statepos) {

		int[][] outPositions = operator.getOutputPositions();

		int outputsize = operator.getOutputPositions(0).length
				+ operator.getOutputPositions(1).length;

		//ArrayList joinedtuples = new ArrayList();
		joinedtuples.clear();
		//List tuplepair = new Vector();
		tuplepair.clear();

		for (int i = 0; i < statetuples.size(); i++) {

			XATTuple nextstatetuple = (XATTuple) statetuples.get(i);

			/* join the two tuples if predicate is satisfied. */
			tuplepair.clear();

			if (statepos == 0) {
				tuplepair.add(nextstatetuple);
				tuplepair.add(tuple);

			} else {
				tuplepair.add(tuple);
				tuplepair.add(nextstatetuple);
			}

			//System.out.println("tuple pair: " + tuplepair.get(0) + " and " +
			//tuplepair.get(1));

			if (((Boolean) expression.eval(tuplepair, operator)).booleanValue()) {
				//List newvalues = new Vector();
				newvalues.clear();
				
				long[] newtimes = new long[outputsize];

				int timeindex = 0;
				for (int j = 0; j < 2; j++) {
					for (int k = 0; k < operator.getOutputPositions(j).length; k++) {
						newvalues.add(((XATTuple) tuplepair.get(j))
								.getCell(outPositions[j][k]));
						newtimes[timeindex] = ((XATTuple) tuplepair.get(j))
								.getNthTimeStamp(outPositions[j][k]);
						timeindex++;
					}
				}
				XATTuple joinedtuple = new XATTupleImp(newvalues);
				joinedtuple.setTimeStamp(newtimes);
				long m1 = tuple.getMinTimeStamp();
				long m2 = nextstatetuple.getMinTimeStamp();
				joinedtuple.setMinTimeStamp((m1 < m2) ? m1 : m2);
				m1 = tuple.getMaxTimeStamp();
				m2 = nextstatetuple.getMaxTimeStamp();
				joinedtuple.setMaxTimeStamp((m1 > m2) ? m1 : m2);
				joinedtuples.add(joinedtuple);
				if (tuple.isNew() && nextstatetuple.isNew()) {
					joinedtuple.setToNew(true);
				} else {
					joinedtuple.setToNew(false);
				}
			}
		}
		return (XATTuple[]) joinedtuples.toArray(new XATTuple[] {});
	}

	/**
	 * Used by multiple way join only. For binary join, should not use this
	 * 
	 * probe a tuple against the statetuples. It must be called after the
	 * purgeByTuplePipe() function with the same tuple. state is ordered by the
	 * timestamps
	 * 
	 * This is a fake probe in the sense that the result tuple construction is not correct
	 * in general. Here we simply attach the matched state tuple to the incoming tuple. 
	 * We ignore the schemas and consequent expression evaluation.
	 * Some schema mapping can be do later
	 *   
	 */
	public XATTuple[] probe(XATTuple tuple, ExpressionStr expression) {

//		int[][] outPositions = operator.getOutputPositions();
//
//		int outputsize = operator.getOutputPositions(0).length
//				+ operator.getOutputPositions(1).length;
//
		//ArrayList joinedtuples = new ArrayList();
		joinedtuples.clear();
		//List tuplepair = new Vector();
		tuplepair.clear();

		for (int i = 0; i < statetuples.size(); i++) {

			XATTuple nextstatetuple = (XATTuple) statetuples.get(i);

			/* join the two tuples if predicate is satisfied. */
			tuplepair.clear();

//			if (statepos == 0) {
				tuplepair.add(nextstatetuple);
				tuplepair.add(tuple);

//			} else {
//				tuplepair.add(tuple);
//				tuplepair.add(nextstatetuple);
//			}

			//System.out.println("tuple pair: " + tuplepair.get(0) + " and " +
			//tuplepair.get(1));

			if (joinable(tuple, nextstatetuple)&&((Boolean) expression.eval(tuplepair, operator)).booleanValue()) {
				//List newvalues = new Vector();
//				newvalues.clear();
//				
//				long[] newtimes = new long[outputsize];
//
//				int timeindex = 0;
//				for (int j = 0; j < 2; j++) {
//					for (int k = 0; k < operator.getOutputPositions(j).length; k++) {
//						newvalues.add(((XATTuple) tuplepair.get(j))
//								.getCell(outPositions[j][k]));
//						newtimes[timeindex] = ((XATTuple) tuplepair.get(j))
//								.getNthTimeStamp(outPositions[j][k]);
//						timeindex++;
//					}
//				}
				XATTuple joinedtuple = this.join(tuple, nextstatetuple);
//				joinedtuple.setTimeStamp(newtimes);
//				long m1 = tuple.getMinTimeStamp();
//				long m2 = nextstatetuple.getMinTimeStamp();
//				joinedtuple.setMinTimeStamp((m1 < m2) ? m1 : m2);
//				m1 = tuple.getMaxTimeStamp();
//				m2 = nextstatetuple.getMaxTimeStamp();
//				joinedtuple.setMaxTimeStamp((m1 > m2) ? m1 : m2);
				joinedtuples.add(joinedtuple);
//				if (tuple.isNew() && nextstatetuple.isNew()) {
//					joinedtuple.setToNew(true);
//				} else {
//					joinedtuple.setToNew(false);
//				}
			}
		}
		return (XATTuple[]) joinedtuples.toArray(new XATTuple[] {});
		
	}

	public boolean joinable(XATTuple newTuple, XATTuple stateTuple) {
		
		/** If this is the last operator and if both tuples' are NEW, don't join */
		/*if (isOldBoxRoot)
		    if (tuple1.isNew() && tuple2.isNew())
		        return false; */
	
		/** if tuple1 and tuple2 is out of window size from each other, don't join */
		try {
			if (newTuple == null || newTuple.getTimeStamp() == null || stateTuple == null || stateTuple.getTimeStamp() == null)
				return false;
	
			if (newTuple.getMaxTimeStamp()- stateTuple.getMinTimeStamp()< startWindow)
				return false;
			if (newTuple.getMaxTimeStamp()- stateTuple.getMinTimeStamp()>endWindow)
				return false;
		} catch (Exception e) {
			System.out.println(e);
			return false;
		}
	
		return true;
	}
	public static XATTuple join(XATTuple lTuple, XATTuple rTuple) {

	    /** join data together */
	    //ArrayList data = (ArrayList) Arrays.asList(lTuple.getCurrentTuple());
	    //data.addAll(Arrays.asList(rTuple.getCurrentTuple()));
	    XATDataValue[] data1 = lTuple.getArrayData();
	    XATDataValue[] data2 = rTuple.getArrayData();
	    int size1 = data1.length;
	    int size2 = data2.length;

	    // we are not worried about schema here, just the join, so just make the new tuple.
	    ArrayList newData = new ArrayList();
	    for (int i = 0; i < size1; i++)
	        newData.add(data1[i]);
	    for (int i = 0; i < size2; i++)
	        newData.add(data2[i]);

	    XATTuple newTuple =
	        new XATTupleImp( (XATDataValue[]) (newData.toArray(new XATDataValue[] {})) );

		newTuple.combineTimeStamps(lTuple, rTuple);
//		lTuple = null;
//		rTuple = null;

	    return newTuple;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState#checkAllOldTuplePurged()
	 */
	public boolean checkAllOldTuplePurged() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState#getAllStateTuples()
	 */
	public XATTuple[] getAllStateTuples() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState#getEarlistTimeStamp()
	 */
	public long[] getEarlistTimeStamp() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState#getEarlistTuple()
	 */
	public XATTuple getEarlistTuple() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState#getLatestTimeStamp()
	 */
	public long[] getLatestTimeStamp() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState#getLatestTuple()
	 */
	public XATTuple getLatestTuple() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState#purgeByTuple(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple)
	 */
	public void purgeByTuple(XATTuple newTuple) {
		// TODO Auto-generated method stub
		
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState#purgeByTuple(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple, long)
	 */
	public void purgeByTuple(XATTuple newTuple, long overall_min) {
		// TODO Auto-generated method stub
		
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState#probe(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[], edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr, int)
	 */
	public XATTuple[] probe(XATTuple[] tuples, ExpressionStr expression, int statepos) {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState#copyStateContent(edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState)
	 */
	public void copyStateContent(OperatorState newstate) {
		// TODO Auto-generated method stub
		
	}

	public void clearState() {
		// TODO Auto-generated method stub
		
	}
}