/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Nov 14, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate;

/**The ProbabilisticPipeWFIFOState is an extension to the the PipeWFIFOState with 
 * just the PROBE overridden to evaluate the join condition check probabilistically
 * by generating a random selectivity.
 * @author mukherab
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */

import java.util.*;
//import java.util.List;
//import java.util.Vector;
//import java.util.ArrayList;
//import java.util.Random;

//import streamgenerator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATPropertiesImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import java.io.Serializable;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

public class ProbabilisticPipeWFIFOState extends PipeWFIFOState implements Serializable {
	
	
	private java.util.Random rand = new java.util.Random();

	private double selectivity = 0;
	
	private int dummyLoop = 20;// to simulate different join cost
	
	/**
	 * ProbabilisticPipeWFIFOState constructor comment.
	 */
	public ProbabilisticPipeWFIFOState() {
		super();
	}
	
	public ProbabilisticPipeWFIFOState(XATNode node) {
		super();	
	}
		
	public ProbabilisticPipeWFIFOState(XATStreamOperator op) {
		super();
		this.setOperator(op);
	}

	
	/**
	 * the set and get methods for the selectivity of join operator.
	 * @return
	 */
		
	public void setSelectivity(double select)
	{
		selectivity = select;		
	}
	
	public double getSelectivity()
	{
		return selectivity;
	}
		

	public int getDummyLoop() {
		return dummyLoop;
	}

	public void setDummyLoop(int dummyLoop) {
		this.dummyLoop = dummyLoop;
	}

/**
	 * This will actually evaluate the join predicate.  It will return true if
	 * the tuples are valid and should be outputted, false if they should be discarded
	 * b/c they didnt meet the predicate.
	 */
	
	public boolean evalJoinPredicate() {
    	
	
    if (rand.nextDouble() < selectivity){
    	return true;
    }
    else{
    	return false;
    }
	}
	
    /**
	 * probe a tuple against the statetuples. It must be called after the
	 * purgeByTuplePipe() function with the same tuple. state is ordered by the
	 * timestamps. This is the PROBABILISTIC one and overrides the probe in PipeWFIFOState. 
	 * The join condition check is replaced by the evalJoinPredicate() method.
	 */
	public XATTuple[] probe(XATTuple tuple, ExpressionStr expression,
			int statepos) {

		int[][] outPositions = operator.getOutputPositions();

		int outputsize = operator.getOutputPositions(0).length
				+ operator.getOutputPositions(1).length;

		ArrayList joinedtuples = new ArrayList();

		List tuplepair = new Vector();

		for (int i = 0; i < statetuples.size(); i++) {

			XATTuple nextstatetuple = (XATTuple) statetuples.get(i);

			/* join the two tuples if predicate is satisfied. */
			tuplepair.clear();

			if (statepos == 0) {
				tuplepair.add(nextstatetuple);
				tuplepair.add(tuple);

			} else {
				tuplepair.add(tuple);
				tuplepair.add(nextstatetuple);
			}

			//System.out.println("tuple pair: " + tuplepair.get(0) + " and " +
			//tuplepair.get(1));

			if (evalJoinPredicate()) {
				List newvalues = new Vector();
				long[] newtimes = new long[outputsize];

				int timeindex = 0;
				for (int j = 0; j < 2; j++) {
					for (int k = 0; k < operator.getOutputPositions(j).length; k++) {
						newvalues.add(((XATTuple) tuplepair.get(j))
								.getCell(outPositions[j][k]));
						newtimes[timeindex] = ((XATTuple) tuplepair.get(j))
								.getNthTimeStamp(outPositions[j][k]);
						timeindex++;
					}
				}
				XATTuple joinedtuple = new XATTupleImp(newvalues);
				joinedtuple.setTimeStamp(newtimes);
				long m1 = tuple.getMinTimeStamp();
				long m2 = nextstatetuple.getMinTimeStamp();
				joinedtuple.setMinTimeStamp((m1 < m2) ? m1 : m2);
				m1 = tuple.getMaxTimeStamp();
				m2 = nextstatetuple.getMaxTimeStamp();
				joinedtuple.setMaxTimeStamp((m1 > m2) ? m1 : m2);
				joinedtuples.add(joinedtuple);
				if (tuple.isNew() && nextstatetuple.isNew()) {
					joinedtuple.setToNew(true);
				} else {
					joinedtuple.setToNew(false);
				}
			}
		}
		return (XATTuple[]) joinedtuples.toArray(new XATTuple[] {});
	}

	/**
	 * Used by multiple way join only. For binary join, should not use this
	 * 
	 * probe a tuple against the statetuples. It must be called after the
	 * purgeByTuplePipe() function with the same tuple. state is ordered by the
	 * timestamps
	 * 
	 * This is a fake probe in the sense that the result tuple construction is not correct
	 * in general. Here we simply attach the matched state tuple to the incoming tuple. 
	 * We ignore the schemas and consequent expression evaluation.
	 * Some schema mapping can be do later
	 *   
	 */
	public XATTuple[] probe(XATTuple tuple, ExpressionStr expression) {

//		int[][] outPositions = operator.getOutputPositions();
//
//		int outputsize = operator.getOutputPositions(0).length
//				+ operator.getOutputPositions(1).length;
//
		//ArrayList joinedtuples = new ArrayList();
		joinedtuples.clear();
		//List tuplepair = new Vector();
		tuplepair.clear();

		for (int i = 0; i < statetuples.size(); i++) {

			XATTuple nextstatetuple = (XATTuple) statetuples.get(i);

			/* join the two tuples if predicate is satisfied. */
			tuplepair.clear();

//			if (statepos == 0) {
				tuplepair.add(nextstatetuple);
				tuplepair.add(tuple);

//			} else {
//				tuplepair.add(tuple);
//				tuplepair.add(nextstatetuple);
//			}

			//System.out.println("tuple pair: " + tuplepair.get(0) + " and " +
			//tuplepair.get(1));

			//dummy loop
			double distance=0.0;
			for (int ii =0;ii<dummyLoop;ii++){
				//for (int iii =0; iii<1000; iii++){
				double doubleTemp = rand.nextDouble()-rand.nextDouble();
				distance += Math.pow(doubleTemp,2);
//				try {
//					Thread.sleep(dummyLoop);
//				} catch (Exception e) {
//					System.out.println(e);
//				}
				//}
			}

			if (joinable(tuple, nextstatetuple)&&this.evalJoinPredicate()) {
				
				//List newvalues = new Vector();
//				newvalues.clear();
//				
//				long[] newtimes = new long[outputsize];
//
//				int timeindex = 0;
//				for (int j = 0; j < 2; j++) {
//					for (int k = 0; k < operator.getOutputPositions(j).length; k++) {
//						newvalues.add(((XATTuple) tuplepair.get(j))
//								.getCell(outPositions[j][k]));
//						newtimes[timeindex] = ((XATTuple) tuplepair.get(j))
//								.getNthTimeStamp(outPositions[j][k]);
//						timeindex++;
//					}
//				}
				XATTuple joinedtuple = this.join(tuple, nextstatetuple);
//				joinedtuple.setTimeStamp(newtimes);
//				long m1 = tuple.getMinTimeStamp();
//				long m2 = nextstatetuple.getMinTimeStamp();
//				joinedtuple.setMinTimeStamp((m1 < m2) ? m1 : m2);
//				m1 = tuple.getMaxTimeStamp();
//				m2 = nextstatetuple.getMaxTimeStamp();
//				joinedtuple.setMaxTimeStamp((m1 > m2) ? m1 : m2);
				joinedtuples.add(joinedtuple);
//				if (tuple.isNew() && nextstatetuple.isNew()) {
//					joinedtuple.setToNew(true);
//				} else {
//					joinedtuple.setToNew(false);
//				}
			}
		}
		return (XATTuple[]) joinedtuples.toArray(new XATTuple[] {});
		
	}	
}