/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Oct 12, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import java.util.Vector;
/**
 * @author Song Wang
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class PipeStreamSelectOperatorImp extends XATSingleQueueStreamOperatorImp 
	{

	private ExpressionStr expression;
	private int queueID;//used to indicate with queue type is going to be filtered
						//this integer will be set by query plan generator
	
	public void setQueueID(int queueid){
		this.queueID = queueid;
	}
	public int getQueueID(){
		return this.queueID;
	}
	public PipeStreamSelectOperatorImp(XATNode init) {
		super(init);
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() throws Exception {
		XATTuple dummyTuple = null; //dummy xattuple to fill out the tupleArray for expression evaluation
		
		// Here we expect that the input queue have multiple schemas
		if (this.getInputQueue().getSchemas() == null) {
			return false;
		}

		// for the next pipelined window join
		//We just simply copy the input schemas to the output queue
		if (this.getOutputQueue()!=null && this.getOutputQueue().getSchema() == null) {
			this.getOutputQueue().setSchemas(
					this.getInputQueue().getSchemas());
		}
		
		/**
		 * first figure out how many tuples to dequeue, the selectivity, and 
		 * result delay  
		 */
		XATQueue inputQueue = this.getInputQueue();
		int size = inputQueue.getQueueSize(this);
		int howMany = 0;

		int dequeueHowmany = Integer.parseInt(this
				.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		howMany = (dequeueHowmany > size) ? size
				: dequeueHowmany;

		/** get the tuples */

		XATTuple[] tuples = this.getInputQueue().dequeue(howMany, this);
		
		/** fill out the tupleList for expression evaluation*/
		Vector tupleList = new Vector();
		for (int i=0;i < this.getInputQueue().getSchemas().size();i++)
			tupleList.add(dummyTuple);
		
		for (int i=0; i<tuples.length; i++) {
			
			// for distributed migration
			if(tuples[i].getMaxTimeStamp() == -1) {
				this.processSyncTuple();
				continue;
			}
			
			//for tuples from unfiltered queue schema, just skip
			if(tuples[i].getStreamID()!=this.queueID) {
				this.getOutputQueue().enqueue(tuples[i], this);				
				continue;
			}
			
			
			//tupleList.clear();
			tupleList.set(this.queueID,tuples[i]);
			Object out = expression.eval(tupleList, this);
			if(out instanceof Boolean) {
				if(((Boolean)out).booleanValue()) {
					//XATTuple newtuple = new XATTupleImp()
					this.getOutputQueue().enqueue(tuples[i], this);
				}
			}
		}
			
//			//the following code is wrong!!!
//			XATTuple newTuple = new XATTupleImp(projAttrPosArray.length);
//			for (int j=0; j<projAttrPosArray.length; j++) {
//				newTuple.setCell(j, tuples[i].getCell(projAttrPosArray[j]));
//			}
//			newTuple.setTimeStamp(tuples[i].getTimeStamp());
//			
//			this.getOutputQueue().enqueue(newTuple, this);
//		}

		return true;
	}

	public String toString() {
		return "Select";
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		// TODO Auto-generated method stub
		return "Select";
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamSelectOperator#getExpression()
	 */
	public ExpressionStr getExpression() {
		return expression;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamSelectOperator#setExpression(edu.wpi.cs.dsrg.xmldb.xat.common.streamexpression.ExpressionStr)
	 */
	public void setExpression(ExpressionStr expression) {
		this.expression = expression;
	}
	
	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input queues.
	 * Inside method run() the numbers are computed based on the property
	 * "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler,
	 * while in cleanAccumulatedTuples, all tuples in input queues that have their max timesamp
	 * smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() {
		return true;
	}
	
	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() {
		return false;
	}

	public boolean matchStates(XATStreamOperator op) {
		return false;
	}
	
	public boolean recomputeUnmatchedState() {
		return true;
	}
	
	public int setStateID(int nextid) {
		return nextid;
	}
	public boolean finalizeOperation(){
		return true;
	}

}