/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;

import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
//import java.lang.reflect.Constructor;
//import streamgenerator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;

/**
 * PipeWBinaryJoinOperatorBodyImp is a join operator for the body and tail of
 * the pipeline. Creation date: (10/05/2005 2:26:41 PM)
 * 
 * @author: Song Wang
 */
public class PipeWBinaryJoinOperatorBodyImp extends
		XATMultiQueueWindowStreamOperatorImp implements PipeWBinaryJoinOperator {

	//added only for statistic collection for exp. can be safely removed
	StringBuffer stats_string;

	
	protected ExpressionStr expression;

	/**
	 * Insert the method's description here. Creation date: (2/17/2003 4:48:47
	 * PM)
	 */
	public PipeWBinaryJoinOperatorBodyImp(XATNode node) {
		super(node);
	}

	/**
	 * overwrite the super class, because here the input queue has multiple schemas
	 * set the outputpositions for the i th output
	 * Creation date: (4/4/2003 5:17:32 PM)
	 * @param i int
	 * @param ithOutputPositions int[]
	 */
	public synchronized void setOutputPositions(int i, int[] ithOutputPositions) {
		if (outputPositions == null){
			outputPositions = new int[this.getInputQueue().getSchemas().size()][];
		}  
		outputPositions[i] = ithOutputPositions;
		}

	/**
	 * This function is used to initialize the two input states of the sliding
	 * window join It will be called by the plan generator in the
	 * visit(PipeWBinaryJoinOperatorHeadImp, XATNode). Creation date:
	 * (10/05/2005 4:48:47 PM)
	 */
	public void initState(String statetype) {

		try {
			//Constructor con = Class.forName(statetype).getConstructor();
			states = new WOperatorState[2];
			states[0] = (WOperatorState) Class.forName(statetype).newInstance();
			states[1] = (WOperatorState) Class.forName(statetype).newInstance();
			states[0].setOperator(this);
			states[1].setOperator(this);
			if (this.startPointArray[0] > 0)
				((PipeWFIFOState) states[0])
						.setStartWindow(this.startPointArray[0]);
			if (this.startPointArray[1] > 0)
				((PipeWFIFOState) states[1])
						.setStartWindow(this.startPointArray[1]);

			if (this.windowSizeArray[0] > 0) {
				((PipeWFIFOState) states[0])
						.setTimeWindow(this.windowSizeArray[0]);
				((PipeWFIFOState) states[0])
						.setEndWindow(this.startPointArray[0]
								+ this.windowSizeArray[0]);
			}
			if (this.windowSizeArray[1] > 0) {
				((PipeWFIFOState) states[1])
						.setTimeWindow(this.windowSizeArray[1]);
				((PipeWFIFOState) states[1])
						.setEndWindow(this.startPointArray[1]
								+ this.windowSizeArray[1]);
			}
			((PipeWFIFOState) states[0]).init(0);
			((PipeWFIFOState) states[1]).init(1);
			
			this.stats_string = new StringBuffer();
			//stats_string.append("StartTime\tEndTime\tOutputTuples\n");


		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * @return
	 */
	public ExpressionStr getExpression() {
		return expression;
	}

	public void printStatsToFile() {
			try {
				String fileName = new String("PipeJoinBody_" + System.currentTimeMillis()+".txt");
				FileWriter out = new FileWriter(fileName, true);
				out.write(this.stats_string.toString());
				out.close();
				
			} catch (IOException e) {
				System.out.println(e.toString());
			}
	}


	/**
	 * Insert the method's description here. Creation date: (4/3/2003 5:00:07
	 * AM)
	 * 
	 * @return queryplanelements.WFIFOState
	 */
	public OperatorState getLeftState() {
		return states[0];
	}

	public int getLeftStateSize() {
		return states[0].getStateSize();
	}

	/**
	 * Insert the method's description here. Creation date: (5/28/2003 12:11:23
	 * PM)
	 * 
	 * @return int
	 */
	//	public int getLPos() {
	//		return lPos;
	//	}
	/**
	 * Insert the method's description here. Creation date: (4/3/2003 5:00:50
	 * AM)
	 * 
	 * @return queryplanelements.WFIFOState
	 */
	public OperatorState getRightState() {
		return states[1];
	}

	public int getRightStateSize() {
		return states[1].getStateSize();
	}

	public int getStateSize() {
		return (states[0].getStateSize() + states[1].getStateSize());
	}

	void processLeftRegularTuple(XATTuple tuple) {
		//step 4: insert into state
		((PipeWFIFOState) states[0]).insertTuple(tuple);

	}
	int processLeftProbeTuple(XATTuple tuple) {
		int outputCount= 0;
		XATTuple[] joinedtuples = null;
		//step 1: cross purge
		List purgedTuples = ((PipeWFIFOState) states[1])
				.purgeByTuplePipeCross(tuple);
//		for (int i = 0; i < purgedTuples.size(); i++) {
//			((XATTuple) purgedTuples.get(i)).setStreamID(1);//0 for left, 1 for
//			// right
//		}
		//if (this.getOutputQueue(1)!=null) 
			this.getOutputQueue(this.getOutputQueueCount()-1).enqueue(purgedTuples,this);
		//step 2: probe the state
		joinedtuples = ((PipeWFIFOState) states[1]).probe(tuple,
				this.expression, 1);
		for (int i=0;i<this.getOutputQueueCount()-1;i++ ){
			this.getOutputQueue(i).enqueue(joinedtuples, this);
			outputCount+=joinedtuples.length;
		}
		//step 3: attach to outputqueue
//		if (this.getOutputQueue(1)!=null) {
//			XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) tuple).clone());
//			probeTuple.setProbe(true);
//			probeTuple.setStreamID(0);
			this.getOutputQueue(this.getOutputQueueCount()-1).enqueue(tuple,this);
//		}
			
			return outputCount;
	}

	void processRightRegularTuple(XATTuple tuple) {
		//step 4: insert into state
		((PipeWFIFOState) states[1]).insertTuple(tuple);
	}
	int processRightProbeTuple(XATTuple tuple) {
		int outputCount= 0;
		XATTuple[] joinedtuples = null;
		//step 1: cross purge
		List purgedTuples = ((PipeWFIFOState) states[0])
				.purgeByTuplePipeCross(tuple);
//		for (int i = 0; i < purgedTuples.size(); i++) {
//			((XATTuple) purgedTuples.get(i)).setStreamID(0);//0 for left, 1 for
//			// right
//		}
//		if (this.getOutputQueue(1)!=null) 
			this.getOutputQueue(this.getOutputQueueCount()-1).enqueue(purgedTuples,this);
		//step 2: probe the state
		joinedtuples = ((PipeWFIFOState) states[0]).probe(tuple,
				this.expression, 0);
		for (int i=0;i<this.getOutputQueueCount()-1;i++ ){
			this.getOutputQueue(i).enqueue(joinedtuples, this);
			outputCount+=joinedtuples.length;
		}
		//step 3: attach to outputqueue
//		if (this.getOutputQueue(1)!=null) {
//			XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) tuple).clone());
//			probeTuple.setProbe(true);
//			probeTuple.setStreamID(1);
			this.getOutputQueue(this.getOutputQueueCount()-1).enqueue(tuple,this);
//		}
			return outputCount;
	}

	/*
	 * overwritten init function to deal with the multiple schema in the input
	 * queue.
	 */
	public void init() {
		XATQueueSchema outSchema = this.getOutputQueue().getSchema();
		if (outSchema == null || outSchema.getLength() == 0) {
			String schemaName = new String();
			outSchema = new XATQueueSchema();
			//Here we expect only one inputQueue existing
			Iterator it = this.getInputQueue().getSchemas().iterator();
			while (it.hasNext()) {
				XATQueueSchema inSchema = (XATQueueSchema) it.next();
				schemaName += "#" + inSchema.getName();
				
				Iterator j = inSchema.getColNames().iterator();
				while(j.hasNext()){
					outSchema.addColumn((String) j.next());
				}
			}

			outSchema.setName(schemaName);
			this.getOutputQueue().setSchema(outSchema);
		}
		computeOutputPositions();

		this.numOutputColsPerStream = new int[outputPositions.length];
		this.numOutputColumns = 0;
		for (int i=0; i<outputPositions.length; i++) {
			int numCols = outputPositions[i].length;
			numOutputColsPerStream[i] = numCols;
			numOutputColumns += numCols;
		}
		this.setHasInit(true);
	}
	/*
	 * overwritten init function to deal with the multiple schema in the input
	 * queue.
	 */
	public void init(int outputQueueId) {
		XATQueueSchema outSchema = this.getOutputQueue(outputQueueId).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) {
			String schemaName = new String();
			outSchema = new XATQueueSchema();
			//Here we expect only one inputQueue existing
			Iterator it = this.getInputQueue().getSchemas().iterator();
			while (it.hasNext()) {
				XATQueueSchema inSchema = (XATQueueSchema) it.next();
				schemaName += "#" + inSchema.getName();
				
				Iterator j = inSchema.getColNames().iterator();
				while(j.hasNext()){
					outSchema.addColumn((String) j.next());
				}
				
			}

			outSchema.setName(schemaName);
			this.getOutputQueue(outputQueueId).setSchema(outSchema);
		}
		computeOutputPositions();

		this.numOutputColsPerStream = new int[outputPositions.length];
		this.numOutputColumns = 0;
		for (int i=0; i<outputPositions.length; i++) {
			int numCols = outputPositions[i].length;
			numOutputColsPerStream[i] = numCols;
			numOutputColumns += numCols;
		}
		this.setHasInit(true);
	}
	
	/**
	 * Overwritten function to deal with the multiple schema in the input queue
	 * compute the attribute mapping between input queuea(s) schema and output
	 * queue schema
	 */
	public void computeOutputPositions() {
		DagNode node = (DagNode)getNode(); 
		Vector outColNames = getOutputQueue().getSchema().getColNames();
		int outputColPos = 0;

		Iterator inputs = getInputQueue().getSchemas().iterator();
		//store all the input columns
		int j=0;
		while(inputs.hasNext()) {
			Vector outputPosVec = new Vector();
			Vector inputColNames = ((XATQueueSchema)inputs.next()).getColNames(); 
			for (Iterator i1 = outColNames.iterator(); i1.hasNext();){
				Object outputCol = i1.next();
				for (int k = 0; k<inputColNames.size(); k++){
					Object inputCol = inputColNames.get(k);
					if (outputCol.equals(inputCol)){
						outputPosVec.add(new Integer(k));
						break;
					}
				}
			}
			//convert vector to int[]
			int[] outputPosArray = new int[outputPosVec.size()];
			for (int count = 0; count < outputPosVec.size(); count++){
				outputPosArray[count] = ((Integer)outputPosVec.get(count)).intValue();
			}
			this.setOutputPositions(j, outputPosArray);
			j++;
		}
	}
	
	/**
	 * Insert the method's description here. return: -1 nothing in queue, no run
	 * 0 exception 1 run successfully
	 * 
	 * Creation date: (1/21/2003 10:57:01 AM)
	 */
	public boolean run() {
		XATTuple tuple = null;
		
		// Here we expect that the input queue have multiple schemas
		if (this.getInputQueue().getSchemas().isEmpty()) {
			return false;
		}

		//outputqueue for join output
		//could be multiple outputs. but the schema is the same.
		for (int i=0;i<this.getOutputQueueCount()-1;i++ )
			if (this.getOutputQueue(i).getSchema() == null) {
				init(i);
			}

		//Outputqueue for pipelined-join
		//We assume here that the plan generator will use the LAST outputqueue
		// for the next pipelined window join
		//We just simply copy the input schemas to the output queue
		if (this.getOutputQueue(this.getOutputQueueCount()-1)!=null && this.getOutputQueue(this.getOutputQueueCount()-1).getSchemas().isEmpty()) {
			this.getOutputQueue(this.getOutputQueueCount()-1).setSchemas(
					this.getInputQueue().getSchemas());
		}

		/** figure out how many tuples to dequeue from left and right queue */
		XATQueue inputQueue = this.getInputQueue();
		int size = inputQueue.getQueueSize(this);

		int dequeueHowmany = Integer.parseInt(this
				.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		//System.out.println(this + "tuple to dequeue " + dequeueHowmany);

		int total = (dequeueHowmany > size) ? size
				: dequeueHowmany;

		int totalOutput =0;
		//begin exp stat collect
		stats_string.append(System.currentTimeMillis()+"\t"); 

			for (int x = 0; x < total; x++) {
				tuple = inputQueue.dequeue(this);
				if (tuple.isProbe()){
					if (tuple.getStreamID()==0)
						totalOutput+=this.processLeftProbeTuple(tuple);
					else totalOutput+=this.processRightProbeTuple(tuple);
				} else {
					if (tuple.getStreamID()==0)
						this.processLeftRegularTuple(tuple);
					else this.processRightRegularTuple(tuple);
				}
			}
			stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

			try {
				inputQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + total);
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);
			}
			
			
		
		return true;
		}
	
	public void setDefaultProperties() {

		super.setDefaultProperties();
		addProperty("TIME_WINDOW", "0");
		addProperty("START_WINDOW", "0");
		addProperty("END_WINDOW", "0");
		// type of state
		addProperty("STATE_TYPE", "PIPEWFIFOSTATE");
	}

	/**
	 * @param expression
	 */
	public void setExpression(ExpressionStr expr) {
		this.expression = expr;
	}

	/**
	 * Insert the method's description here. Creation date: (2/14/2003 9:33:59
	 * AM)
	 * 
	 * @return java.lang.String
	 */
	public String toString() {
		String rs = new String();

		String sName = "StateSlicedJoin";//this.getClass().getName();
		//find the last dot.
		int index = 0;
		for (index = sName.length() - 1; index >= 0; index--)
			if (sName.charAt(index) == '.')
				break;
		rs = sName.substring(index + 1) + " ";
//		String id = this.getProperty("OBJECT_ID");
//		if (id == "-1") {
//			rs = rs + " " + this.hashCode();
//		} else {
//			rs = rs + id + " " + this.hashCode();
//		}
//
//		if (this.isActive())
//			rs += " ACTIVE ";
//		else
//			rs += " INACTIVE ";
		return rs;
	}

	public String toToolTipString() {
		String rs = new String();

		String sName = this.getClass().getName();
		//find the last dot.
		int index = 0;
		for (index = sName.length() - 1; index >= 0; index--)
			if (sName.charAt(index) == '.')
				break;
		rs = sName.substring(index + 1) + " ";
		String id = this.getProperty("OBJECT_ID");
		if (id == "-1") {
			rs = rs + hashCode();
		} else {
			rs = rs + id;
		}
//		if (this.isActive())
//			rs += " ACTIVE ";
//		else
//			rs += " INACTIVE ";
//
//		/** now add the join predicate and window size info */
//		rs += "\n";
//		rs += "predicate: " + expression + "\n";
//		rs += "window size = " + windowSize + "\n";

		return rs;
	}

	/*
	 * Assign new ID to each state inside this operator, starting from the given
	 * startID.
	 * 
	 * return the next available ID. For example, if startID is 2, and there are
	 * 2 states inside this operator, then these two states are assigned new IDs
	 * as 2 and 3, and 4 is then returned as the next available stateID.
	 */
	public int setStateID(int startID) {
		if (states == null) {
			return startID;
		}

		for (int i = 0; i < states.length; i++) {
			states[i].setStateID(startID);
			startID++;
		}
		return startID;
	}

	public boolean finalizeOperation() {
		this.printStatsToFile();
		
		System.out.println("\n\nStatistic Collection Done\n");
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAccumulatedTuples(long)
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		// TODO Auto-generated method stub
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAllAccumulatedTuples()
	 */
	public boolean cleanAllAccumulatedTuples() {
		// TODO Auto-generated method stub
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#matchStates(edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator)
	 */
	public boolean matchStates(XATStreamOperator op) {
		// TODO Auto-generated method stub
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#hasUnmatchedState()
	 */
	public boolean hasUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#recomputeUnmatchedState()
	 */
	public boolean recomputeUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

}
