/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;

import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
//import java.lang.reflect.Constructor;
//import streamgenerator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;

/**
 * PipeWBinaryJoinOperatorHeadImp is a join operator for the head of the
 * pipeline. Creation date: (10/05/2005 2:26:41 PM)
 * 
 * @author: Song Wang and Abhishek Mukherji
 */
public class PipeWBinaryJoinOperatorHeadImp
 extends
		XATMultiQueueWindowStreamOperatorImp implements PipeWBinaryJoinOperator {

	//added only for statistic collection for exp. can be safely removed
	protected StringBuffer stats_string;
	String stateType=null;
	private ExpressionStr expression;

	/**
	 * Insert the method's description here. Creation date: (2/17/2003 4:48:47
	 * PM)
	 */
	public PipeWBinaryJoinOperatorHeadImp(XATNode node) {
		super(node);
	}

	public String getStateType(){
		return this.stateType;
	}
	/**
	 * This function is used to initialize the two input states of the sliding
	 * window join It will be called by the plan generator in the
	 * visit(PipeWBinaryJoinOperatorHeadImp, XATNode). Creation date:
	 * (10/05/2005 4:48:47 PM)
	 */
	public void initState(String statetype) {
		this.stateType = statetype;
		try {
			//Constructor con = Class.forName(statetype).getConstructor();
			states = new WOperatorState[2];
			states[0] = (WOperatorState) Class.forName(statetype).newInstance();
			states[1] = (WOperatorState) Class.forName(statetype).newInstance();
			states[0].setOperator(this);
			states[1].setOperator(this);
			if (this.startPointArray[0] > 0)
				((PipeWFIFOState) states[0])
						.setStartWindow(this.startPointArray[0]);
			if (this.startPointArray[1] > 0)
				((PipeWFIFOState) states[1])
						.setStartWindow(this.startPointArray[1]);

			if (this.windowSizeArray[0] > 0) {
				((PipeWFIFOState) states[0])
						.setTimeWindow(this.windowSizeArray[0]);
				((PipeWFIFOState) states[0])
						.setEndWindow(this.startPointArray[0]
								+ this.windowSizeArray[0]);
			}
			if (this.windowSizeArray[1] > 0) {
				((PipeWFIFOState) states[1])
						.setTimeWindow(this.windowSizeArray[1]);
				((PipeWFIFOState) states[1])
						.setEndWindow(this.startPointArray[1]
								+ this.windowSizeArray[1]);
			}
			((PipeWFIFOState) states[0]).init(0);
			((PipeWFIFOState) states[1]).init(1);
			
			this.stats_string = new StringBuffer();
			//stats_string.append("StartTime\tEndTime\tOutputTuples\n");

		
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * @return
	 */
	public ExpressionStr getExpression() {
		return expression;
	}

	public void printStatsToFile() {
		try {
			String fileName = new String("PipeJoinHead_" + System.currentTimeMillis()+".txt");
			FileWriter out = new FileWriter(fileName, true);
			out.write(this.stats_string.toString());
			out.close();
			
		} catch (IOException e) {
			System.out.println(e.toString());
		}
}

	/**
	 * Insert the method's description here. Creation date: (4/3/2003 5:00:07
	 * AM)
	 * 
	 * @return queryplanelements.WFIFOState
	 */
	public OperatorState getLeftState() {
		return states[0];
	}

	public int getLeftStateSize() {
		return states[0].getStateSize();
	}

	/**
	 * Insert the method's description here. Creation date: (5/28/2003 12:11:23
	 * PM)
	 * 
	 * @return int
	 */
	//	public int getLPos() {
	//		return lPos;
	//	}
	/**
	 * Insert the method's description here. Creation date: (4/3/2003 5:00:50
	 * AM)
	 * 
	 * @return queryplanelements.WFIFOState
	 */
	public OperatorState getRightState() {
		return states[1];
	}

	public int getRightStateSize() {
		return states[1].getStateSize();
	}

	public int getStateSize() {
		return (states[0].getStateSize() + states[1].getStateSize());
	}

	private int processLeftTuple(XATTuple tuple,XATTuple probeTuple) {
		int ret=0;
		XATTuple[] joinedtuples = null;
		//step 1: cross purge
		List purgedTuples = ((PipeWFIFOState) states[1])
				.purgeByTuplePipeCross(tuple);
		for (int i = 0; i < purgedTuples.size(); i++) {
			((XATTuple) purgedTuples.get(i)).setStreamID(1);//0 for left, 1 for
			// right
		}
		this.getOutputQueue(this.getOutputQueueCount()-1).addTuples(purgedTuples);
		//step 2: probe the state
		joinedtuples = ((PipeWFIFOState) states[1]).probe(tuple,
				this.expression, 1);
		for (int i=0;i<this.getOutputQueueCount()-1;i++ ){
			this.getOutputQueue(i).enqueue(joinedtuples, this);
			ret+=joinedtuples.length ;
		}
		//step 3: attach to outputqueue
//		XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) tuple).clone());
//		probeTuple.setProbe(true);
//		probeTuple.setStreamID(0);
		this.getOutputQueue(this.getOutputQueueCount()-1).addTuple(probeTuple);
		//step 4: insert into state
		((PipeWFIFOState) states[0]).insertTuple(tuple);
		
		return ret;
	}

	private int processRightTuple(XATTuple tuple,XATTuple probeTuple) {
		int ret=0;
		XATTuple[] joinedtuples = null;
		//step 1: cross purge
		List purgedTuples = ((PipeWFIFOState) states[0])
				.purgeByTuplePipeCross(tuple);
		for (int i = 0; i < purgedTuples.size(); i++) {
			((XATTuple) purgedTuples.get(i)).setStreamID(0);//0 for left, 1 for
			// right
		}
		this.getOutputQueue(this.getOutputQueueCount()-1).addTuples(purgedTuples);
		//step 2: probe the state
		joinedtuples = ((PipeWFIFOState) states[0]).probe(tuple,
				this.expression, 0);
		for (int i=0;i<this.getOutputQueueCount()-1;i++ ){
			this.getOutputQueue(i).enqueue(joinedtuples, this);
			ret+=joinedtuples.length ;
		}
			
		//step 3: attach to outputqueue
//		XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) tuple).clone());
//		probeTuple.setProbe(true);
//		probeTuple.setStreamID(1);
		this.getOutputQueue(this.getOutputQueueCount()-1).addTuple(probeTuple);
		//step 4: insert into state
		((PipeWFIFOState) states[1]).insertTuple(tuple);
		
		return ret;
	}

	/**
	 * Insert the method's description here. return: -1 nothing in queue, no run
	 * 0 exception 1 run successfully
	 * 
	 * Creation date: (1/21/2003 10:57:01 AM)
	 */
	public boolean run() {
		
		XATTuple leftTuple = null;
		XATTuple[] leftTuples = null;
		XATTuple rightTuple = null;
		XATTuple[] rightTuples = null;
		
		ArrayList clonedLeftTupleList = new ArrayList();
		XATTuple[] clonedLeftTuples = null;
		ArrayList clonedRightTupleList = new ArrayList();
		XATTuple[] clonedRightTuples = null;
		

		if ((this.getLeftInputQueue().getSchema() == null)
				|| (this.getRightInputQueue().getSchema() == null)) {
			return false;
		}
		
		//outputqueue for join output
		//could be multiple outputs. but the schema is the same.
		for (int i=0;i<this.getOutputQueueCount()-1;i++ )
			if (this.getOutputQueue(i).getSchema() == null) {
				init(i);
			}
			
		//Outputqueue for pipelined-join
		//We assume here that the plan generator will use the LAST outputqueue
		// for the next pipelined window join
		//We just simply combine the input schemas to the output queue
		if	(this.getOutputQueue(this.getOutputQueueCount()-1).getSchemas().isEmpty()){
			this.getOutputQueue(this.getOutputQueueCount()-1).addSchema(
					this.getLeftInputQueue().getSchema());
			this.getOutputQueue(this.getOutputQueueCount()-1).addSchema(
					this.getRightInputQueue().getSchema());
		}

		/** figure out how many tuples to dequeue from left and right queue */
		XATQueue lQueue = this.getLeftInputQueue();
		XATQueue rQueue = this.getRightInputQueue();
		int size1 = lQueue.getQueueSize(this);
		int size2 = rQueue.getQueueSize(this);
		int lHowmany = 0, rHowmany = 0;

		int dequeueHowmany = Integer.parseInt(this
				.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		//System.out.println(this + "tuple to dequeue " + dequeueHowmany);

		int totalOutput =0;

		//left queue is empty. we only deal with tuples from right queue
		//here we assume the tuple from left and right will arrive in order
		if (size1 <= 0 && size2 > 0) {
			lHowmany = 0;
			rHowmany = (dequeueHowmany > size2) ? size2 : dequeueHowmany;
			try {
				lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
				rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);
			}
			
			rightTuples = rQueue.peek(rHowmany, this);
			for (int x = 0; x < rightTuples.length; x++) {
				XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) rightTuples[x]).clone());
				probeTuple.setProbe(true);
				probeTuple.setStreamID(1);
				clonedRightTupleList.add(probeTuple);
			}
			clonedRightTuples = (XATTuple[])clonedRightTupleList.toArray(new XATTuple[]{});
			rightTuples = null;
			
			//begin exp stat collect
			stats_string.append(System.currentTimeMillis()+"\t"); 

			rightTuples = rQueue.dequeue(rHowmany, this);
			for (int x = 0; x < rightTuples.length; x++) {
				totalOutput+=this.processRightTuple(rightTuples[x],clonedRightTuples[x]);
			}
			//end exp
			stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	
			
			return true;
		}

		//right queue is empty. we only deal with tuples from left queue
		//here we assume the tuple from left and right will arrive in order
		if (size2 <= 0 && size1 > 0) {
			rHowmany = 0;
			lHowmany = (dequeueHowmany > size1) ? size1 : dequeueHowmany;
			try {
				lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
				rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);
			}
			
			leftTuples = lQueue.peek(lHowmany, this);
			for (int x = 0; x < leftTuples.length; x++) {
				XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) leftTuples[x]).clone());
				probeTuple.setProbe(true);
				probeTuple.setStreamID(0);
				clonedLeftTupleList.add(probeTuple);
			}
			clonedLeftTuples = (XATTuple[])clonedLeftTupleList.toArray(new XATTuple[]{});
			leftTuples = null;

			//begin exp stat collect
			stats_string.append(System.currentTimeMillis()+"\t"); 

			leftTuples = lQueue.dequeue(lHowmany, this);
			for (int x = 0; x < leftTuples.length; x++) {
				totalOutput+=this.processLeftTuple(leftTuples[x],clonedLeftTuples[x]);
			}
			//end exp
			stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

			return true;
		}

		//Both queues are not empty.
		//here we assume the tuple from left and right will arrive in order
		if (size2 > 0 && size1 > 0) {
			lHowmany = 0;
			rHowmany = 0;
			int total = (dequeueHowmany > (size1 + size2)) ? (size1 + size2)
					: dequeueHowmany;

			rightTuples = rQueue.peek(size2, this);
			for (int x = 0; x < rightTuples.length; x++) {
				XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) rightTuples[x]).clone());
				probeTuple.setProbe(true);
				probeTuple.setStreamID(1);
				clonedRightTupleList.add(probeTuple);
			}
			clonedRightTuples = (XATTuple[])clonedRightTupleList.toArray(new XATTuple[]{});
			rightTuples = null;

			leftTuples = lQueue.peek(size1, this);
			for (int x = 0; x < leftTuples.length; x++) {
				XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) leftTuples[x]).clone());
				probeTuple.setProbe(true);
				probeTuple.setStreamID(0);
				clonedLeftTupleList.add(probeTuple);
			}
			clonedLeftTuples = (XATTuple[])clonedLeftTupleList.toArray(new XATTuple[]{});
			leftTuples = null;


			//begin exp stat collect
			stats_string.append(System.currentTimeMillis()+"\t"); 

			for (int x = 0; x < total; x++) {
				if (lHowmany < size1 && leftTuple == null) {
					leftTuple = lQueue.dequeue(this);
					lHowmany++;
				}
				if (rHowmany < size2 && rightTuple == null) {
					rightTuple = rQueue.dequeue(this);
					rHowmany++;
				}
				if (rightTuple!=null&&leftTuple == null
						|| rightTuple!=null&&leftTuple.getMaxTimeStamp()
								- rightTuple.getMaxTimeStamp() >= 0) {
					totalOutput+=this.processRightTuple(rightTuple,clonedRightTuples[rHowmany-1]);
					rightTuple = null;
				} else if (leftTuple!=null&&rightTuple == null
						|| leftTuple!=null&&rightTuple.getMaxTimeStamp()
								- leftTuple.getMaxTimeStamp() > 0) {
					totalOutput+=this.processLeftTuple(leftTuple,clonedLeftTuples[lHowmany-1]);
					leftTuple = null;
				}
			}
			stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	
			
			try {
				lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
				rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);
			}
			return true;
		}

		//Both queues are empty.
		return true;
	}

	public void setDefaultProperties() {

		super.setDefaultProperties();
		addProperty("TIME_WINDOW", "0");
		addProperty("START_WINDOW", "0");
		addProperty("END_WINDOW", "0");
		// type of state
		addProperty("STATE_TYPE", "PIPEWFIFOSTATE");
	}

	/**
	 * @param expression
	 */
	public void setExpression(ExpressionStr expr) {
		this.expression = expr;
	}

	/**
	 * Insert the method's description here. Creation date: (2/14/2003 9:33:59
	 * AM)
	 * 
	 * @return java.lang.String
	 */
	public String toString() {
		String rs = new String();

		String sName = this.getClass().getName();
		//find the last dot.
		int index = 0;
		for (index = sName.length() - 1; index >= 0; index--)
			if (sName.charAt(index) == '.')
				break;
		rs = sName.substring(index + 1) + " ";
		String id = this.getProperty("OBJECT_ID");
		if (id == "-1") {
			rs = rs + " " + this.hashCode();
		} else {
			rs = rs + id + " " + this.hashCode();
		}

		if (this.isActive())
			rs += " ACTIVE ";
		else
			rs += " INACTIVE ";
		return rs;
	}

	public String toToolTipString() {
		String rs = new String();

		String sName = this.getClass().getName();
		//find the last dot.
		int index = 0;
		for (index = sName.length() - 1; index >= 0; index--)
			if (sName.charAt(index) == '.')
				break;
		rs = sName.substring(index + 1) + " ";
		String id = this.getProperty("OBJECT_ID");
		if (id == "-1") {
			rs = rs + hashCode();
		} else {
			rs = rs + id;
		}
		if (this.isActive())
			rs += " ACTIVE ";
		else
			rs += " INACTIVE ";

		/** now add the join predicate and window size info */
		rs += "\n";
		rs += "predicate: " + expression + "\n";
		rs += "window size = " + windowSize + "\n";

		return rs;
	}

	/*
	 * Assign new ID to each state inside this operator, starting from the given
	 * startID.
	 * 
	 * return the next available ID. For example, if startID is 2, and there are
	 * 2 states inside this operator, then these two states are assigned new IDs
	 * as 2 and 3, and 4 is then returned as the next available stateID.
	 */
	public int setStateID(int startID) {
		if (states == null) {
			return startID;
		}

		for (int i = 0; i < states.length; i++) {
			states[i].setStateID(startID);
			startID++;
		}
		return startID;
	}

	public boolean finalizeOperation() {
		this.printStatsToFile();
		
		System.out.println("\n\nStatistic Collection Done\n");
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAccumulatedTuples(long)
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		// TODO Auto-generated method stub
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAllAccumulatedTuples()
	 */
	public boolean cleanAllAccumulatedTuples() {
		// TODO Auto-generated method stub
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#matchStates(edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator)
	 */
	public boolean matchStates(XATStreamOperator op) {
		// TODO Auto-generated method stub
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#hasUnmatchedState()
	 */
	public boolean hasUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#recomputeUnmatchedState()
	 */
	public boolean recomputeUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

}

