/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Oct 13, 2005
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator;

import java.io.FileWriter;
import java.io.IOException;
import java.util.*;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate.PipeWFIFOState;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * @author Song Wang
 *
 * Used for the last join operator in a pipe. 
 * The outputqueues should be null for this operaotr
 */
public class PipeWBinaryJoinOperatorTailImp extends
		PipeWBinaryJoinOperatorBodyImp {

	/**
	 * @param node
	 */
	public PipeWBinaryJoinOperatorTailImp(XATNode node) {
		super(node);
		
		// TODO Auto-generated constructor stub
	}

	int processLeftProbeTuple(XATTuple tuple) {
		int ret=0;
		XATTuple[] joinedtuples = null;
		//step 1: cross purge
		List purgedTuples = ((PipeWFIFOState) states[1])
				.purgeByTuplePipeCross(tuple);
//		for (int i = 0; i < purgedTuples.size(); i++) {
//			((XATTuple) purgedTuples.get(i)).setStreamID(1);//0 for left, 1 for
//			// right
//		}
		//if (this.getOutputQueue(1)!=null) this.getOutputQueue(1).enqueue(purgedTuples,this);
		//step 2: probe the state
		joinedtuples = ((PipeWFIFOState) states[1]).probe(tuple,
				this.expression, 1);
		for (int i=0;i<this.getOutputQueueCount();i++ ){
			this.getOutputQueue(i).enqueue(joinedtuples, this);
			ret+=joinedtuples.length;
		}
		//step 3: attach to outputqueue
//		if (this.getOutputQueue(1)!=null) {
//			XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) tuple).clone());
//			probeTuple.setProbe(true);
//			probeTuple.setStreamID(0);
//			this.getOutputQueue(1).enqueue(tuple,this);
//		}
		return ret;
	}

	int processRightProbeTuple(XATTuple tuple) {
		int ret=0;
		XATTuple[] joinedtuples = null;
		//step 1: cross purge
		List purgedTuples = ((PipeWFIFOState) states[0])
				.purgeByTuplePipeCross(tuple);
//		for (int i = 0; i < purgedTuples.size(); i++) {
//			((XATTuple) purgedTuples.get(i)).setStreamID(0);//0 for left, 1 for
//			// right
//		}
//		if (this.getOutputQueue(1)!=null) this.getOutputQueue(1).enqueue(purgedTuples,this);
		//step 2: probe the state
		joinedtuples = ((PipeWFIFOState) states[0]).probe(tuple,
				this.expression, 0);
		for (int i=0;i<this.getOutputQueueCount();i++ ){
			this.getOutputQueue(i).enqueue(joinedtuples, this);
			ret+=joinedtuples.length;
		}
		//step 3: attach to outputqueue
//		if (this.getOutputQueue(1)!=null) {
//			XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) tuple).clone());
//			probeTuple.setProbe(true);
//			probeTuple.setStreamID(1);
//			this.getOutputQueue(1).enqueue(tuple,this);
//		}
		return ret;
	}
	/**
	 * Insert the method's description here. return: -1 nothing in queue, no run
	 * 0 exception 1 run successfully
	 * overwrite the run() in parent class
	 * Creation date: (1/21/2003 10:57:01 AM)
	 */
	public boolean run() {
		XATTuple tuple = null;
		
		// Here we expect that the input queue have multiple schemas
		if (this.getInputQueue().getSchemas() == null) {
			return false;
		}

		//outputqueue for join output
		//could be multiple outputs. but the schema is the same.
		for (int i=0;i<this.getOutputQueueCount();i++ )
			if (this.getOutputQueue(i).getSchema() == null) {
				init(i);
			}


		/** figure out how many tuples to dequeue from left and right queue */
		XATQueue inputQueue = this.getInputQueue();
		int size = inputQueue.getQueueSize(this);
		int howmany = 0;

		int dequeueHowmany = Integer.parseInt(this
				.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		//System.out.println(this + "tuple to dequeue " + dequeueHowmany);

		int total = (dequeueHowmany > size) ? size
				: dequeueHowmany;

		int totalOutput =0;
		//begin exp stat collect
		stats_string.append(System.currentTimeMillis()+"\t"); 

			for (int x = 0; x < total; x++) {
				tuple = inputQueue.dequeue(this);
				if (tuple.isProbe()){
					if (tuple.getStreamID()==0)
						totalOutput+=this.processLeftProbeTuple(tuple);
					else totalOutput+=this.processRightProbeTuple(tuple);
				} else {
					if (tuple.getStreamID()==0)
						this.processLeftRegularTuple(tuple);
					else this.processRightRegularTuple(tuple);
				}
			}
			stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

			try {
				inputQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + howmany);
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);
			}
			return true;
		}
	public void printStatsToFile() {
		try {
			String fileName = new String("PipeJoinTail_" + System.currentTimeMillis()+".txt");
			FileWriter out = new FileWriter(fileName, true);
			out.write(this.stats_string.toString());
			out.close();
			
		} catch (IOException e) {
			System.out.println(e.toString());
		}
}
	
}
