/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator;

import java.util.Iterator;
import java.util.List;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate.PipeWFIFOState;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

public class PipeWMJoinOperatorBasicHeadImp extends PipeWMJoinOperatorBasicImp {
	int locked = -1;//show if current processing duration is done or not

	public PipeWMJoinOperatorBasicHeadImp(PipeWMJoinOperatorBasicImp oldop,
			XATNode newnode) {
		super(oldop, newnode);
		// TODO Auto-generated constructor stub
	}

	public PipeWMJoinOperatorBasicHeadImp(XATNode init) {
		super(init);
		// TODO Auto-generated constructor stub
	}

	public PipeWMJoinOperatorBasicHeadImp(XATNode init, int n) {
		super(init, n);
		// TODO Auto-generated constructor stub
	}
	public boolean run(){
		
		XATTuple[] tuples = new XATTuple[this.getInputQueueCount()];
		
		// Here we expect that the input queue have multiple schemas
		for (int i=0;i<this.getInputQueueCount()-1;i++){
			if (this.getInputQueue(i).getSchema()==null) {
				return false;
			}
		}

		//outputqueue for stream tuple propagation 
		//Note: we do not care about schema of intermediate result queue
		for (int i=0;i<this.getOutputQueueCount()-2;i++ )
			if (this.getOutputQueue(i).getSchema() == null) {
				init(i);
				this.getInputQueue(i).removeAllTuples();
			}

		//Outputqueue for pipelined-join
		//We assume here that the plan generator will use the LAST outputqueue
		// for the output of the pipelined window join
		if (this.getOutputQueue(this.getOutputQueueCount()-1)!=null && this.getOutputQueue(this.getOutputQueueCount()-1).getSchema()==null) {
			this.initResult(this.getOutputQueueCount()-1);
		}

		/** figure out how many tuples to dequeue from input queues */
		Iterator it= this.getInputQueues();
		int size =0;
		while (it.hasNext()) {
			XATQueue inputQueue = (XATQueue)it.next();
			size += inputQueue.getQueueSize(this);
		}
		int dequeueHowmany = Integer.parseInt(this
				.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		//System.out.println(this + "tuple to dequeue " + dequeueHowmany);

		int total = (dequeueHowmany > size) ? size
				: dequeueHowmany;

		int totalOutput =0;

		//begin exp stat collect
		long timeNow =System.currentTimeMillis()/1000;
		if (timeNow!=this.secondCounter) {
			//System.out.println(this.getOperatorID()+":"+this.getStateSize());
			//state_string.append(System.currentTimeMillis()+"\t"+this.getStateSize()+"\n");
			state_string.append(this.getStateSize()+"\n");
			this.secondCounter=timeNow;
		}

		XATTuple tempTuple = null;
		XATQueue tempQueue = null;
		for (int y=0;y<this.getInputQueueCount();y++){
			tempTuple = null;
			tempQueue = null;
			tempQueue = this.getInputQueue(y);
			if (tempQueue.getQueueSize()>0) {
					tempTuple = tempQueue.peek(this);

			}
			tuples[y]=tempTuple;
		}

		for (int x = 0; x < total;) {
			int z=-1;
			
			if (locked>0) {
				z = this.getInputQueueCount()-1;//process intermediate result queue only
				if  (this.getInputQueue(z).getQueueSize()<=0) return true;//wait for intermediate result
			}
			else z = findTupleWithMinTimestamp(tuples);
			
			if (z==-1) break;//total tuples have been processed. should not reach this.
			
			XATTuple tuple = this.getInputQueue(z).dequeue(this);
			
			tuple.setTimeEnter(System.currentTimeMillis());
//			tuple.setMaxTimeStamp(enterTime);
//			tuple.setMinTimeStamp(enterTime);
			
			x++;
			if (z<this.getInputQueueCount()-1) {//is stream tuple
				locked = this.numOfInputs-2;//e.g. a->B->C->D, wait for (a,b) and (a,b,c)
				XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) tuple).clone());
				probeTuple.setProbe(true);
				//as male tuple
				totalOutput += this.processMaleTuple(probeTuple,z);
				//as female tuple
				this.processFemaleTuple(tuple,z);
			} else {//is intermediate result tuple
				this.processIntermediateResult(tuple,z);
			}

			//fill the buffer, kind of merge-sorting of all the tuples in the input queues.
			tempTuple = null;
			tempQueue = null;
			tempQueue = this.getInputQueue(z);
			if (tempQueue.getQueueSize()>0) {
					tempTuple = tempQueue.peek(this);
			}
			tuples[z]=tempTuple;
			
		}
		//stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

//		try {
//			inputQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + total);
//			setNumberOfPossibleOutputs();
//		} catch (Exception e) {
//			System.out.println(e);
//		}
			
		return true;
	}

	int processMaleTuple(XATTuple tuple,int from) {
		int outputCount= 0;
		XATTuple[] joinedtuples = null;
		//step 1: cross purge of all the states
	    //propagate purged tuple to corresponding output queues
		for (int i = 0; i< this.numOfInputs; i++){
			List purgedTuples = ((PipeWFIFOState) states[i])
			.purgeByTuplePipeCross(tuple);
			this.getOutputQueue(i).enqueue(purgedTuples,this);
		}
//		for (int i = 0; i < purgedTuples.size(); i++) {
//			((XATTuple) purgedTuples.get(i)).setStreamID(1);//0 for left, 1 for
//			// right
//		}
		//if (this.getOutputQueue(1)!=null) 
		
		//step 2: attach to outputqueue
//		if (this.getOutputQueue(1)!=null) {
//			XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) tuple).clone());
//			probeTuple.setProbe(true);
//			probeTuple.setStreamID(0);
			this.getOutputQueue(from).enqueue(tuple,this);
//		}
			

		//step 3: probe the state
		int toBeProbe = this.joinOrderings[from][0];//the join probing state
		joinedtuples = ((PipeWFIFOState) states[toBeProbe]).probe(tuple,
				this.getExpression(0, 0));
		
		if (joinedtuples.length==0) {//make a dummy tuple
			XATTuple joinedtuple = PipeWFIFOState.join(tuple, tuple);
			joinedtuple.setDummy(true);
			joinedtuple.setOpID(this.operatorID);
			joinedtuple.setStreamID(from);
			joinedtuples = new XATTuple[1];
			joinedtuples[0] = joinedtuple;
		}

		
		//mark the intermediate result
		for (int i = 0; i < joinedtuples.length; i++) {
			joinedtuples[i].setStreamID(from);
			joinedtuples[i].setOpID(this.getOperatorID());
			joinedtuples[i].setStateToJoin(this.joinOrderings[from][1]);
		}
		
		//set the ending tuple
		joinedtuples[joinedtuples.length-1].setEnding(true);
		
		//by default, the last queue is join output queue
		//by default, the queue next to the last queue is the intermediate result queue
		//for (int i=0;i<this.getOutputQueueCount()-2;i++ ){
		this.getOutputQueue(this.getOutputQueueCount()-2).enqueue(joinedtuples, this);
		outputCount+=joinedtuples.length;
		//}
		return outputCount;
	}

	int processIntermediateResult(XATTuple tuple,int z){
		
		int outputCount= 0;
		XATTuple[] joinedtuples = new XATTuple[]{};
		boolean finalResult =false;
		int toBeProbe;
		
		if (tuple.getEnding()) locked--;
		
		//step 1: propagation if tuple.opID != operatorID
		if (tuple.getOpID()!=this.getOperatorID()) this.getOutputQueue(this.getOutputQueueCount()-2).enqueue(tuple,this);
		
		if (!tuple.isDummy()) {
	
			//step 2: probe the state
			toBeProbe = tuple.getStateToJoin();//the join probing state
			joinedtuples = ((PipeWFIFOState) states[toBeProbe]).probe(tuple,
					this.getExpression(0, 0));
		
			//check if or not the final join result
			int j=0;
			for (j=0;j<this.numOfInputs-2;j++) { 
				if (this.joinOrderings[tuple.getStreamID()][j]==toBeProbe)
					break;
			}
			if (j==this.numOfInputs-2){//final join output achieved
				finalResult = true;
			} else {
				toBeProbe =this.joinOrderings[tuple.getStreamID()][j+1];
			}

			//mark the intermediate result
			for (int i = 0; i < joinedtuples.length; i++) {
				joinedtuples[i].setStreamID(tuple.getStreamID());
				joinedtuples[i].setOpID(this.getOperatorID());
				if (!finalResult){//not final join output achieved
					joinedtuples[i].setStateToJoin(toBeProbe);
				}
			}
			
			outputCount+=joinedtuples.length;
			
			if (finalResult){
				this.getOutputQueue(this.getOutputQueueCount()-1).enqueue(joinedtuples, this);
				outputStat(joinedtuples);

				return outputCount;
			} else {//not final joined result
				//generate next arround ending tuple
				if (tuple.getOpID()==this.getOperatorID()&&tuple.getEnding()) {
					if (joinedtuples.length==0) {//make a dummy tuple
						XATTuple joinedtuple = PipeWFIFOState.join(tuple, tuple);
						joinedtuple.setDummy(true);
						joinedtuple.setOpID(this.operatorID);
						joinedtuple.setStreamID(tuple.getStreamID());
						joinedtuples = new XATTuple[1];
						joinedtuples[0] = joinedtuple;
					}
					//set the ending tuple
					joinedtuples[joinedtuples.length-1].setEnding(true);
					joinedtuples[joinedtuples.length-1].setStateToJoin(toBeProbe);
				}
				//not final result
				this.getOutputQueue(this.getOutputQueueCount()-2).enqueue(joinedtuples, this);
			}
		} else {//tuple is dummy
			//check if or not the final join result
			toBeProbe = tuple.getStateToJoin();//the join probing state
			
			int j=0;
			for (j=0;j<this.numOfInputs-2;j++) { 
				if (this.joinOrderings[tuple.getStreamID()][j]==toBeProbe)
					break;
			}
			if (j==this.numOfInputs-2){//final join output achieved
				finalResult = true;
			} else {
				toBeProbe =this.joinOrderings[tuple.getStreamID()][j+1];
			}

			if (finalResult){
				return outputCount;
			} else {//not final joined result
				//generate next arround ending tuple
				if (tuple.getOpID()==this.getOperatorID()&&tuple.getEnding()) {
					if (joinedtuples.length==0) {//make a dummy tuple
						XATTuple joinedtuple = PipeWFIFOState.join(tuple, tuple);
						joinedtuple.setDummy(true);
						joinedtuple.setOpID(this.operatorID);
						joinedtuple.setStreamID(tuple.getStreamID());
						joinedtuples = new XATTuple[1];
						joinedtuples[0] = joinedtuple;
					}
					//set the ending tuple
					joinedtuples[joinedtuples.length-1].setEnding(true);
					joinedtuples[joinedtuples.length-1].setStateToJoin(toBeProbe);
				}
				//not final result
				this.getOutputQueue(this.getOutputQueueCount()-2).enqueue(joinedtuples, this);
			}

		}
		
		return outputCount;
		
	}
	
	
	
}