/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Oct 22, 2004
 * @author Yali Zhu
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator;

import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate.PipeWFIFOState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

public class PipeWMJoinOperatorInterleavedImp	extends PipeWMJoinOperatorBasicImp implements PipeWMJoinOperator {
	int numOfInterleave = 0;
	
	public PipeWMJoinOperatorInterleavedImp(PipeWMJoinOperatorInterleavedImp oldop, XATNode newnode) {
		super(oldop, newnode);
	}
	
	public PipeWMJoinOperatorInterleavedImp(XATNode init) {
		super(init);
	}
	
	public PipeWMJoinOperatorInterleavedImp(XATNode init, int n) {
		super(init);
	}
	
	public int getNumOfInterleave(){
		return this.numOfInterleave;
	}

	public void setNumOfInterleave(int input) {
		this.numOfInterleave = input;
	}
	/**
	 * The following functions are basically for pipelined multiway join operators. 
	 * each intermediate result tuple contains the state number to be joined [0 to n-1]
	 * initialized as -1.
	 * The boolean ending means if it is an end tuple.
	 * The opID means the operator ID that generate this intermediate result
	 * For the original input stream tuples, all these three variables are meaningless.
	 * 
	 * NOTE: the streamID above is used to identify the stream that triggers the intermediate result.
	 * For example, if one a tuple from stream A is processed, then all the intermediate result has streamID = 0
	 * if the join order is A->C->B->E->D, then the (A,C) pair will have streamID=0, stateToJoin=1,
	 * (A,C,B) pair will have streamID=0, stateToJoin=4.
	 * If there is another join order of B->C->A->D->E, then the (B,C,A) pair will have streamID=1, stateToJoin=3
	 * Of cause the (A,C,B) and (B,C,A) tuple have the same schema and can't be distinguished by the tuple itself.
	 * 
	 * Actually, we can assume that the same prefix intermediate result will have the same further join ordering
	 * That is, the (A,C,B) and (B,C,A) tuples do not need to distinguish at all. The further join ordering
	 * must be the same for them. so right now, the streamID is used for the matrix entry only.
	 * 
	 * Added by Song Wang
	 * @return
	 */

	int processMaleTuple(XATTuple tuple,int from) {
		int outputCount= 0;
		XATTuple[] joinedtuples = null;
		//step 1: cross purge of all the states
	    //propagate purged tuple to corresponding output queues
		for (int i = 0; i< this.numOfInputs; i++){
			List purgedTuples = ((PipeWFIFOState) states[i])
			.delayedPurgeByTuplePipeCross(tuple);//.purgeByTuplePipeCross(tuple);
			this.getOutputQueue(i).enqueue(purgedTuples,this);
			
		}
//		for (int i = 0; i < purgedTuples.size(); i++) {
//			((XATTuple) purgedTuples.get(i)).setStreamID(1);//0 for left, 1 for
//			// right
//		}
		//if (this.getOutputQueue(1)!=null) 
		
		//step 2: attach to outputqueue
//		if (this.getOutputQueue(1)!=null) {
//			XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) tuple).clone());
//			probeTuple.setProbe(true);
//			probeTuple.setStreamID(0);
			this.getOutputQueue(from).enqueue(tuple,this);
//		}
			

		//step 3: probe the state
		int toBeProbe = this.joinOrderings[from][0];//the join probing state
		joinedtuples = ((PipeWFIFOState) states[toBeProbe]).probe(tuple,
				this.getExpression(0, 0));
		
		//mark the intermediate result
		for (int i = 0; i < joinedtuples.length; i++) {
			joinedtuples[i].setStreamID(from);
			joinedtuples[i].setOpID(this.getOperatorID());
			joinedtuples[i].setStateToJoin(this.joinOrderings[from][1]);
		}
			
		//by default, the last queue is join output queue
		//by default, the queue next to the last queue is the intermediate result queue
		//for (int i=0;i<this.getOutputQueueCount()-2;i++ ){
		this.getOutputQueue(this.getOutputQueueCount()-2).enqueue(joinedtuples, this);
		outputCount+=joinedtuples.length;
		//}
		return outputCount;
	}

	int processIntermediateResult(XATTuple tuple,int z){
		
		int outputCount= 0;
		XATTuple[] joinedtuples = new XATTuple[]{};
		boolean finalResult =false;
		
		//step 1: propagation if tuple.opID != operatorID
		if (tuple.getOpID()!=this.getOperatorID()) this.getOutputQueue(this.getOutputQueueCount()-2).enqueue(tuple,this);
		
		if (tuple.isDummy()) {
			return 0;
		}
		//finish the real purge since we now are using delayed state purge.
		//then do the real purge, propagate this purge tuple and no need to probe
		if (tuple.isRealPurge()){
			for (int i = 0; i< this.numOfInputs; i++){
				List purgedTuples = ((PipeWFIFOState) states[i])
				.purgeByTuplePipeCross(tuple);
			}
			this.getOutputQueue(this.getOutputQueueCount()-2).enqueue(tuple,this);
			//System.out.println("body: real purge");
			return 0;
		}

		//step 2: probe the state
		int toBeProbe = tuple.getStateToJoin();//the join probing state
		joinedtuples = ((PipeWFIFOState) states[toBeProbe]).probe(tuple,
				this.getExpression(0, 0));
	
		//mark the intermediate result
		for (int i = 0; i < joinedtuples.length; i++) {
			joinedtuples[i].setStreamID(tuple.getStreamID());
			joinedtuples[i].setOpID(this.getOperatorID());
			int j=0;
			for (j=0;j<this.numOfInputs-2;j++) { 
				if (this.joinOrderings[tuple.getStreamID()][j]==tuple.getStateToJoin())
					break;
			}
			if (j==this.numOfInputs-2){//final join output achieved
				finalResult = true;
			} else {
			joinedtuples[i].setStateToJoin(this.joinOrderings[tuple.getStreamID()][j+1]);
			}
		}
		if (finalResult){
			this.getOutputQueue(this.getOutputQueueCount()-1).enqueue(joinedtuples, this);
			outputStat(joinedtuples);

		} else {
			this.getOutputQueue(this.getOutputQueueCount()-2).enqueue(joinedtuples, this);
		}
		outputCount+=joinedtuples.length;
		
		return outputCount;
		
	}

}