/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;

import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
//import java.lang.reflect.Constructor;
//import streamgenerator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;

/**
 * PipeWBinaryJoinOperatorHeadImp is a join operator for the head of the
 * pipeline. Creation date: (10/05/2005 2:26:41 PM)
 * 
 * @author: Song Wang and Abhishek Mukherji
 */
public class RegularWBinaryJoinOperatorImp extends PipeWBinaryJoinOperatorHeadImp {

	/**
	 * Insert the method's description here. Creation date: (2/17/2003 4:48:47
	 * PM)
	 */
	public RegularWBinaryJoinOperatorImp(XATNode node) {
		super(node);
	}

	private int processLeftTuple(XATTuple tuple) {
		int ret=0;
		XATTuple[] joinedtuples = null;
		//step 1: cross purge
		List purgedTuples = ((PipeWFIFOState) states[1])
				.purgeByTuplePipeCross(tuple);
		//step 2: probe the state
		joinedtuples = ((PipeWFIFOState) states[1]).probe(tuple,
				this.getExpression(), 1);
		for (int i=0;i<this.getOutputQueueCount();i++ ){
			this.getOutputQueue(i).enqueue(joinedtuples, this);
			ret+=joinedtuples.length;
		}
		//step 3: attach to outputqueue
		//step 4: insert into state
		((PipeWFIFOState) states[0]).insertTuple(tuple);
		return ret;

	}

	private int processRightTuple(XATTuple tuple) {
		int ret=0;
		XATTuple[] joinedtuples = null;
		//step 1: cross purge
		List purgedTuples = ((PipeWFIFOState) states[0])
				.purgeByTuplePipeCross(tuple);
		//step 2: probe the state
		joinedtuples = ((PipeWFIFOState) states[0]).probe(tuple,
				this.getExpression(), 0);
		for (int i=0;i<this.getOutputQueueCount();i++ ){
			this.getOutputQueue(i).enqueue(joinedtuples, this);
			ret+=joinedtuples.length;
		}
		//step 3: attach to outputqueue
		//step 4: insert into state
		((PipeWFIFOState) states[1]).insertTuple(tuple);
		return ret;
	}

	/**
	 * Insert the method's description here. return: -1 nothing in queue, no run
	 * 0 exception 1 run successfully
	 * 
	 * Creation date: (1/21/2003 10:57:01 AM)
	 */
	public boolean run() {
		XATTuple leftTuple = null;
		XATTuple[] leftTuples = null;
		XATTuple rightTuple = null;
		XATTuple[] rightTuples = null;

		if ((this.getLeftInputQueue().getSchema() == null)
				|| (this.getRightInputQueue().getSchema() == null)) {
			return false;
		}
		
		//outputqueue for join output
		//could be multiple outputs. but the schema is the same.
		for (int i=0;i<this.getOutputQueueCount();i++ )
			if (this.getOutputQueue(i).getSchema() == null) {
				init(i);
			}
			
		/** figure out how many tuples to dequeue from left and right queue */
		XATQueue lQueue = this.getLeftInputQueue();
		XATQueue rQueue = this.getRightInputQueue();
		int size1 = lQueue.getQueueSize(this);
		int size2 = rQueue.getQueueSize(this);
		int lHowmany = 0, rHowmany = 0;

		int dequeueHowmany = Integer.parseInt(this
				.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		//System.out.println(this + "tuple to dequeue " + dequeueHowmany);

		int totalOutput =0;

		//left queue is empty. we only deal with tuples from right queue
		//here we assume the tuple from left and right will arrive in order
		if (size1 <= 0 && size2 > 0) {
			lHowmany = 0;
			rHowmany = (dequeueHowmany > size2) ? size2 : dequeueHowmany;
			try {
				lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
				rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);
			}

			//begin exp stat collect
			stats_string.append(System.currentTimeMillis()+"\t"); 

			rightTuples = rQueue.dequeue(rHowmany, this);
			for (int x = 0; x < rightTuples.length; x++) {
				totalOutput+=this.processRightTuple(rightTuples[x]);
			}
			stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

			return true;
		}

		//right queue is empty. we only deal with tuples from left queue
		//here we assume the tuple from left and right will arrive in order
		if (size2 <= 0 && size1 > 0) {
			rHowmany = 0;
			lHowmany = (dequeueHowmany > size1) ? size1 : dequeueHowmany;
			try {
				lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
				rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);
			}
			//begin exp stat collect
			stats_string.append(System.currentTimeMillis()+"\t"); 

			leftTuples = lQueue.dequeue(lHowmany, this);
			for (int x = 0; x < leftTuples.length; x++) {
				totalOutput+=this.processLeftTuple(leftTuples[x]);
			}
			stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

			return true;
		}

		//Both queues are not empty.
		//here we assume the tuple from left and right will arrive in order
		if (size2 > 0 && size1 > 0) {
			lHowmany = 0;
			rHowmany = 0;
			int total = (dequeueHowmany > (size1 + size2)) ? (size1 + size2)
					: dequeueHowmany;

			//begin exp stat collect
			stats_string.append(System.currentTimeMillis()+"\t"); 

			for (int x = 0; x < total; x++) {
				if (lHowmany < size1 && leftTuple == null) {
					leftTuple = lQueue.dequeue(this);
					lHowmany++;
				}
				if (rHowmany < size2 && rightTuple == null) {
					rightTuple = rQueue.dequeue(this);
					rHowmany++;
				}
				if (rightTuple!=null&&leftTuple == null
						|| rightTuple!=null&&leftTuple.getMaxTimeStamp()
								- rightTuple.getMaxTimeStamp() >= 0) {
					totalOutput+=this.processRightTuple(rightTuple);
					rightTuple = null;
				} else if (leftTuple!=null&&rightTuple == null
						|| leftTuple!=null&&rightTuple.getMaxTimeStamp()
								- leftTuple.getMaxTimeStamp() > 0) {
					totalOutput+=this.processLeftTuple(leftTuple);
					leftTuple = null;
				}
			}
			stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

			try {
				lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
				rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);
			}
			return true;
		}

		//Both queues are empty.
		return true;
	}
	public void printStatsToFile() {
		try {
			String fileName = new String("RegularJoin_" + System.currentTimeMillis()+".txt");
			FileWriter out = new FileWriter(fileName, true);
			out.write(this.stats_string.toString());
			out.close();
			
		} catch (IOException e) {
			System.out.println(e.toString());
		}
}


}

