/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;

import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
//import java.lang.reflect.Constructor;
//import streamgenerator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;

/**
 * PipeWBinaryJoinOperatorHeadImp is a join operator for the head of the
 * pipeline. Creation date: (10/05/2005 2:26:41 PM)
 * 
 * @author: Song Wang and Abhishek Mukherji
 */
public class RegularWMJoinOperatorCTRImp extends PipeWMJoinOperatorBasicImp {

	int numOfNode = 0;
	int numOfCopy = 0;
	private java.util.Random rand = new java.util.Random();

	/**
	 * Insert the method's description here. Creation date: (2/17/2003 4:48:47
	 * PM)
	 */
	public RegularWMJoinOperatorCTRImp(XATNode node) {
		super(node);
	}

	public RegularWMJoinOperatorCTRImp(RegularWMJoinOperatorCTRImp oldop, XATNode newnode) {
		super(oldop, newnode);
	}
	
	public RegularWMJoinOperatorCTRImp(XATNode init, int n) {
		super(init, n);
	}
	/*
	 * overwritten init function to deal with the multiple schema in the input
	 * queue. For the join output queue
	 */
	public void initResult(int outputQueueId) {
		XATQueueSchema outSchema = this.getOutputQueue(outputQueueId).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) {
			String schemaName = new String();
			outSchema = new XATQueueSchema();
			//Here we expect multiple inputQueues existing
			Iterator it =this.getInputQueues(); 
			while (it.hasNext()) {
				XATQueueSchema inSchema = ((XATQueue)it.next()).getSchema();
				if (inSchema!=null){
					schemaName += "#" + inSchema.getName();
					
					Iterator j = inSchema.getColNames().iterator();
					while(j.hasNext()){
						outSchema.addColumn((String) j.next());
					}
				}
			}

			outSchema.setName(schemaName);
			//by default, the last outputqueue is used to output final join result
			this.getOutputQueue(outputQueueId).setSchema(outSchema);
		}
		//commented out since one of the input queue is intermediate result without scheme.
//		computeOutputPositionsResult(outputQueueId);
//
//		this.numOutputColsPerStream = new int[outputPositions.length];
//		this.numOutputColumns = 0;
//		for (int i=0; i<outputPositions.length-1; i++) {
//			int numCols = outputPositions[i].length;
//			numOutputColsPerStream[i] = numCols;
//			numOutputColumns += numCols;
//		}
		this.setHasInit(true);
	}

	public boolean run(){
		XATTuple[] tuples = new XATTuple[this.getInputQueueCount()];
		
		// Here we expect that the input queue have multiple schemas
		// last input queue is the intermediate result queue
		for (int i=0;i<this.getInputQueueCount()-2;i++){
			if (this.getInputQueue(i).getSchema()==null) {
				return false;
			}
		}

		//outputqueue for stream tuple propagation 
		//Note: we do not care about schema of intermediate result queue
		//for (int i=0;i<this.getOutputQueueCount()-2;i++ )
		//	if (this.getOutputQueue(i).getSchema() == null) {
		//		init(i);
		//	}

		//Outputqueue for pipelined-join
		//We assume here that the plan generator will use the LAST outputqueue
		// for the output of the pipelined window join
		//NEW: for the regular MJoin CTR operator, there is two output queue, first one is intermediate result, last one is joined result
		if (this.getOutputQueue(this.getOutputQueueCount()-1)!=null && this.getOutputQueue(this.getOutputQueueCount()-1).getSchema()==null) {
			this.initResult(this.getOutputQueueCount()-1);
		}

		/** figure out how many tuples to dequeue from input queues */
		Iterator it= this.getInputQueues();
		int size =0;
		while (it.hasNext()) {
			XATQueue inputQueue = (XATQueue)it.next();
			size += inputQueue.getQueueSize(this);
		}
		//int dequeueHowmany = Integer.parseInt(this
		//		.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		int dequeueHowmany = 10;
		
		//System.out.println(this + "tuple to dequeue " + dequeueHowmany);

		int total = (dequeueHowmany > size) ? size
				: dequeueHowmany;

		int totalOutput =0;

		
		XATTuple tempTuple = null;
		XATQueue tempQueue = null;
		for (int y=0;y<this.getInputQueueCount();y++){
			tempTuple = null;
			tempQueue = null;
			tempQueue = this.getInputQueue(y);
			if (tempQueue.getQueueSize()>0) {
					tempTuple = tempQueue.dequeue(this);

			}
			tuples[y]=tempTuple;
		}

		for (int x = 0; x < total;) {
			
			int z = findTupleWithMinTimestamp(tuples);
			
			if (z==-1) break;//total tuples have been processed. should not reach this.
			
			//XATTuple tuple = this.getInputQueue(z).dequeue(this);
			XATTuple tuple = tuples[z];

			if (tuple.getTimeEnter()==0) tuple.setTimeEnter(System.currentTimeMillis());

			x++;
			if (z<this.getInputQueueCount()-1) {//is stream tuple
				
				//there is two possible situations
				//the sending out number from the CTR router is 
				//max(numOfNode/numOfCopy,numOfCopy)
				//when numOfNode/numOfCopy>numOfCopy, it means following case 1. otherwise case 2.
				//1. if numOfNode>numOfCopy^2, this means every tuple need probing but some for insertion.
				//2. if numOfNode<numOfCopy^2, this means every tuple need insertion but some for probing.
				//3. if numOfNode=numOfCopy^2, this means every tuple need insertion and for probing.
				
				if (rand.nextDouble()*(this.numOfNode-2)<=this.numOfCopy*this.numOfCopy) {
					this.processFemaleTuple(tuple,z);
				}
				
				if (rand.nextDouble()*this.numOfCopy*this.numOfCopy<=(this.numOfNode-2)) {
					this.processMaleTuple(tuple,z);
				}
//				if (tuple.isProbe()){//male tuple
//					totalOutput += this.processMaleTuple(tuple,z);
//				} else {//female tuple
//					this.processFemaleTuple(tuple,z);
//				}
			} else {//is intermediate result tuple
				//if is intermediate result, must probing. no droping.
				this.processIntermediateResult(tuple,z);
			}

			//fill the buffer, kind of merge-sorting of all the tuples in the input queues.
			tempTuple = null;
			tempQueue = null;
			tempQueue = this.getInputQueue(z);
			if (tempQueue.getQueueSize()>0) {
					tempTuple = tempQueue.dequeue(this);
			}
			tuples[z]=tempTuple;
			
		}
		//stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

//		try {
//			inputQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + total);
//			setNumberOfPossibleOutputs();
//		} catch (Exception e) {
//			System.out.println(e);
//		}
			
		//begin exp stat collect
		long timeNow =System.currentTimeMillis()/1000;
		if (timeNow!=this.secondCounter) {
			//System.out.println(this.getOperatorID()+":"+this.getStateSize());
			//state_string.append(System.currentTimeMillis()+"\t"+this.getStateSize()+"\n");
			state_string.append(this.getStateSize()+"\n");
			this.secondCounter=timeNow;
		}
		
		return true;
	}
	
	
	int processMaleTuple(XATTuple tuple,int from) {
		int outputCount= 0;
		XATTuple[] joinedtuples = null;
		//step 1: cross purge of all the states
	    //propagate purged tuple to corresponding output queues
		for (int i = 0; i< this.numOfInputs; i++){
			List purgedTuples = ((PipeWFIFOState) states[i])
			.purgeByTuplePipeCross(tuple);
			//this.getOutputQueue(i).enqueue(purgedTuples,this);
		}
//		for (int i = 0; i < purgedTuples.size(); i++) {
//			((XATTuple) purgedTuples.get(i)).setStreamID(1);//0 for left, 1 for
//			// right
//		}
		//if (this.getOutputQueue(1)!=null) 
		
		//step 2: attach to outputqueue
//		if (this.getOutputQueue(1)!=null) {
//			XATTuple probeTuple = (XATTupleImp) (((XATTupleImp) tuple).clone());
//			probeTuple.setProbe(true);
//			probeTuple.setStreamID(0);
			//this.getOutputQueue(from).enqueue(tuple,this);
//		}
			

		//step 3: probe the state
		int toBeProbe = this.joinOrderings[from][0];//the join probing state
		joinedtuples = ((PipeWFIFOState) states[toBeProbe]).probe(tuple,
				this.getExpression(0, 0));
		
		//mark the intermediate result
		for (int i = 0; i < joinedtuples.length; i++) {
			joinedtuples[i].setStreamID(from);
			joinedtuples[i].setOpID(this.getOperatorID());
			joinedtuples[i].setStateToJoin(this.joinOrderings[from][1]);
		}
			
		//by default, the last queue is join output queue
		//by default, the queue next to the last queue is the intermediate result queue
		//for (int i=0;i<this.getOutputQueueCount()-2;i++ ){
		this.getOutputQueue(this.getOutputQueueCount()-2).enqueue(joinedtuples, this);
		//for (int i = 0; i < joinedtuples.length; i++) {
		//	this.processIntermediateResult(joinedtuples[i], from);
		//}
		outputCount+=joinedtuples.length;
		//}
		return outputCount;
	}
	int processIntermediateResult(XATTuple tuple,int z){
		
		int outputCount= 0;
		XATTuple[] joinedtuples = new XATTuple[]{};
		boolean finalResult =false;
		
		//step 1: propagation if tuple.opID != operatorID
		//if (tuple.getOpID()!=this.getOperatorID()) this.getOutputQueue(this.getOutputQueueCount()-2).enqueue(tuple,this);
		
		//if (tuple.isDummy()) {
		//	return 0;
		//}

		//step 2: probe the state
		int toBeProbe = tuple.getStateToJoin();//the join probing state
		joinedtuples = ((PipeWFIFOState) states[toBeProbe]).probe(tuple,
				this.getExpression(0, 0));
	
		//mark the intermediate result
		for (int i = 0; i < joinedtuples.length; i++) {
			joinedtuples[i].setStreamID(tuple.getStreamID());
			joinedtuples[i].setOpID(this.getOperatorID());
			int j=0;
			for (j=0;j<this.numOfInputs-2;j++) { 
				if (this.joinOrderings[tuple.getStreamID()][j]==tuple.getStateToJoin())
					break;
			}
			if (j==this.numOfInputs-2){//final join output achieved
				finalResult = true;
			} else {
			joinedtuples[i].setStateToJoin(this.joinOrderings[tuple.getStreamID()][j+1]);
			}
		}
		if (finalResult){
			this.getOutputQueue(this.getOutputQueueCount()-1).enqueue(joinedtuples, this);
			outputStat(joinedtuples);
		} else {
			this.getOutputQueue(this.getOutputQueueCount()-2).enqueue(joinedtuples, this);
			//for (int ii = 0; ii < joinedtuples.length; ii++) {
			//	this.processIntermediateResult(joinedtuples[ii], z);
			//}
		}
		outputCount+=joinedtuples.length;
		
		return outputCount;
		
	}
	int findTupleWithMinTimestamp(XATTuple[] tuples){
		int ret = -1;
		XATTuple tempTuple = null;
		
		//make the intermediate result queue prior
		if (tuples[tuples.length-1]!=null) {
			ret = tuples.length-1;
			return ret;
		}
		for (int i=0;i<tuples.length;i++){
			if (tuples[i]==null) continue;
			if (tempTuple==null ||tuples[i].getMaxTimeStamp()<=tempTuple.getMaxTimeStamp()){
				ret =i;
				tempTuple = tuples[i];
			}
		}
		return ret;
	}

	public void setNumOfNode(int input) {
		// TODO Auto-generated method stub
		this.numOfNode = input;
	}

	public void setNumOfCopy(int input) {
		// TODO Auto-generated method stub
		this.numOfCopy = input;
	}

}
