/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Oct, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;
/**
 * @author Song Wang
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class StreamALTRouter extends XATMultiQueueWindowStreamOperatorImp 
	{
	//added only for statistic collection for exp. can be safely removed
	StringBuffer stats_string;

	//NOTE: to simplify, we assume the first input stream is the fixed master stream.
	//We also assume all the segment size are equal for all the streams
	int numOfInputs = 0;  
	int numOfNodes = 0;
	int segSize = 0;
	long initTime = 0;
	
	public StreamALTRouter(XATNode init) {
		
		super(init);
		this.stats_string = new StringBuffer();
		//stats_string.append("StartTime\tEndTime\tOutputTuples\n");

	}

	public int getNumOfInputs() {
		return numOfInputs;
	}
	
	public void setNumOfInputs(int n) {
		this.numOfInputs = n;
	}
	public int getNumOfNodes() {
		return numOfNodes;
	}
	
	public void setNumOfNodes(int n) {
		this.numOfNodes = n;
	}
	public int getSegSize() {
		return segSize;
	}
	
	public void setSegSize(int n) {
		this.segSize = n;
	}

	public int getOperatorID() {
		return this.operatorID;
	}
	
	public void setOperatorID(int n) {
		this.operatorID = n;
	}
	
	public void init(int inputQueueID, int outputQueueId) {
		XATQueueSchema outSchema = this.getOutputQueue(outputQueueId).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) {
			String schemaName = new String();
			outSchema = new XATQueueSchema();
			XATQueueSchema inSchema = this.getInputQueue(inputQueueID).getSchema();
			schemaName = inSchema.getName();
				
			Iterator j = inSchema.getColNames().iterator();
			while(j.hasNext()){
				outSchema.addColumn((String) j.next());
			}
			outSchema.setName(schemaName);
			this.getOutputQueue(outputQueueId).setSchema(outSchema);
		}
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() throws Exception {
		
		XATTuple[] tuples = new XATTuple[this.getInputQueueCount()];
		
		for (int i=0;i<this.getInputQueueCount();i++){
			if (this.getInputQueue(i).getSchema()==null) {
				return false;
			}
		}

		//outputqueue for stream tuple propagation 
		for (int i=0;i<this.getInputQueueCount();i++ )
			for (int j=0;j<this.numOfNodes-1;j++)
				if (this.getOutputQueue(i+this.numOfInputs*j).getSchema() == null) {
					init(i,i+this.numOfInputs*j);
					//this.getInputQueue(i).removeAllTuples();
				}

		/** figure out how many tuples to dequeue from input queues */
		Iterator it= this.getInputQueues();
		int size =0;
		while (it.hasNext()) {
			XATQueue inputQueue = (XATQueue)it.next();
			size += inputQueue.getQueueSize(this);
		}
		int dequeueHowmany = Integer.parseInt(this
				.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));


		int total = (dequeueHowmany > size) ? size
				: dequeueHowmany;

		int totalOutput =0;
		//begin exp stat collect
		//stats_string.append(System.currentTimeMillis()+"\t"); 

		XATTuple tempTuple = null;
		XATQueue tempQueue = null;
		for (int y=0;y<this.getInputQueueCount();y++){
			tempTuple = null;
			tempQueue = null;
			tempQueue = this.getInputQueue(y);
			if (tempQueue.getQueueSize()>0) {
					tempTuple = tempQueue.peek(this);

			}
			tuples[y]=tempTuple;
		}

		for (int x = 0; x < total;) {
			int sendTo=0;
			int sendToQueue=0;
			int z = findTupleWithMinTimestamp(tuples);
			
			if (z==-1) break;//total tuples have been processed. should not reach this.
			
			XATTuple tuple = this.getInputQueue(z).dequeue(this);
			
			if (initTime ==0) initTime = tuple.getMinTimeStamp();
			
			//if (tuple.getTimeEnter()==0) tuple.setTimeEnter(System.currentTimeMillis());
			//tuple.setMaxTimeStamp(timeEnter);
			//tuple.setMinTimeStamp(timeEnter);
			
			x++;

			//propagate the tuple to one of the output queue
			tuple.setToNew(true);//identify the new stream tuple
			
			//if the tuple comes from master stream
			if (z==0) {
				sendTo= segNum(tuple);
				sendToQueue = z+this.numOfInputs*sendTo;
				this.getOutputQueue(sendToQueue).enqueue(tuple, this);
				//System.out.println("Master:"+sendToQueue+":"+tuple);
			} else {
				int segModValue = segMod(tuple);
				int segNumValue = segNum(tuple);
				sendTo = segNumValue;
				sendToQueue = z+this.numOfInputs*sendTo;
				this.getOutputQueue(sendToQueue).enqueue(tuple, this);
				//System.out.println("Slave:"+z+":"+sendToQueue+":"+tuple);
				//send backward
				sendTo = segNumValue;
				int winMaster=(int) this.windowSizeArray[0];
				while (winMaster>=segModValue){
					sendTo--;
					sendToQueue = z+this.numOfInputs*segPositive(sendTo);
					this.getOutputQueue(sendToQueue).enqueue(tuple, this);
					//System.out.println("Slave backward:"+z+":"+sendToQueue+":"+tuple);
					winMaster-=segSize;
				}
				//send forward
				sendTo = segNumValue;
				int winSlave=(int) this.windowSizeArray[z];
				while (segSize-winSlave<=segModValue){
					sendTo++;
					sendToQueue = z+this.numOfInputs*segPositive(sendTo);
					this.getOutputQueue(sendToQueue).enqueue(tuple, this);
					//System.out.println("Slave forward:"+z+":"+sendToQueue+":"+tuple);
					winSlave-=segSize;
				}
			}
			
			//System.out.println("Time:" +tuple.getMaxTimeStamp()+"\t"+"windowPartitionSize:"+windowPartitionSize);
			//System.out.println("sendTo:" +sendTo+"\t"+"SendToQueue"+sendToQueue);
			//this.getOutputQueue(sendToQueue).enqueue(tuple, this);
			//System.out.println(sendToQueue+":"+tuple);	
			//fill the buffer, kind of merge-sorting of all the tuples in the input queues.
			tempTuple = null;
			tempQueue = null;
			tempQueue = this.getInputQueue(z);
			if (tempQueue.getQueueSize()>0) {
					tempTuple = tempQueue.peek(this);
			}
			tuples[z]=tempTuple;
			
		}
		//stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

//		try {
//			inputQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + total);
//			setNumberOfPossibleOutputs();
//		} catch (Exception e) {
//			System.out.println(e);
//		}
			
		return true;
	}

	private int segNum(XATTuple tuple){
		long diff =  (tuple.getMinTimeStamp()-initTime);
		int ret = (int)(diff%(segSize*(numOfNodes-1)))/segSize;
		return ret;
	}
		
	private int segMod(XATTuple tuple){
		long diff =  (tuple.getMinTimeStamp()-initTime);
		int ret = (int)(diff%(segSize*(numOfNodes-1)))%segSize;
		return ret;
	}
	
	private int segPositive (int seg){
		while (seg<0) {
			seg+=(numOfNodes-1);
		}
		return seg%(numOfNodes-1);
	}
	
	int findTupleWithMinTimestamp(XATTuple[] tuples){
		int ret = -1;
		XATTuple tempTuple = null;
		for (int i=0;i<tuples.length;i++){
			if (tuples[i]==null) continue;
			if (tempTuple==null ||tuples[i].getMaxTimeStamp()<tempTuple.getMaxTimeStamp()){
				ret =i;
				tempTuple = tuples[i];
			}
		}
		return ret;
	}
		

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		return "Select";
	}

	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input queues.
	 * Inside method run() the numbers are computed based on the property
	 * "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler,
	 * while in cleanAccumulatedTuples, all tuples in input queues that have their max timesamp
	 * smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() {
		return true;
	}
	
	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() {
		return false;
	}

	public boolean matchStates(XATStreamOperator op) {
		return false;
	}
	
	public boolean recomputeUnmatchedState() {
		return true;
	}
	
	public int setStateID(int nextid) {
		return nextid;
	}
	public boolean finalizeOperation(){
		this.printStatsToFile();
		
		System.out.println("\n\nStatistic Collection Done\n");
		return true;
	}
	public void printStatsToFile() {
		try {
			String fileName = new String("WindowSplit_" + System.currentTimeMillis()+".txt");
			FileWriter out = new FileWriter(fileName, true);
			out.write(this.stats_string.toString());
			out.close();
			
		} catch (IOException e) {
			System.out.println(e.toString());
		}
}


}