/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator;

import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

public class StreamCTRRouterNoInter extends StreamCTRRouter {
	private java.util.Random rand = new java.util.Random();
	
	public StreamCTRRouterNoInter(XATNode init) {
		super(init);
		// TODO Auto-generated constructor stub
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() throws Exception {
		
		XATTuple[] tuples = new XATTuple[this.getInputQueueCount()];
		
		//last one is NOT intermediate result
		for (int i=0;i<this.getInputQueueCount();i++){
			if (this.getInputQueue(i).getSchema()==null) {
				return false;
			}
		}

		//outputqueue for stream tuple propagation
		//input 0(A),1(B),2(C),
		//ouput 0(A),1(B),2(C),3(A),4(B),5(C)
		for (int i=0;i<this.getInputQueueCount();i++ )
			for (int j=0;j<this.numOfNodes-2;j++)
				if (this.getOutputQueue(i+(this.numOfInputs)*j).getSchema() == null) {
					init(i,i+(this.numOfInputs)*j);
					//this.getInputQueue(i).removeAllTuples();
				}

		/** figure out how many tuples to dequeue from input queues */
		Iterator it= this.getInputQueues();
		int size =0;
		while (it.hasNext()) {
			XATQueue inputQueue = (XATQueue)it.next();
			size += inputQueue.getQueueSize(this);
		}
		//int dequeueHowmany = Integer.parseInt(this
		//		.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		int dequeueHowmany = 10;

		int total = (dequeueHowmany > size) ? size
				: dequeueHowmany;

		int totalOutput =0;
		//begin exp stat collect
		//stats_string.append(System.currentTimeMillis()+"\t"); 

		XATTuple tempTuple = null;
		XATQueue tempQueue = null;
		for (int y=0;y<this.getInputQueueCount();y++){
			tempTuple = null;
			tempQueue = null;
			tempQueue = this.getInputQueue(y);
			if (tempQueue.getQueueSize()>0) {
					tempTuple = tempQueue.peek(this);

			}
			tuples[y]=tempTuple;
		}

		for (int x = 0; x < total;) {
			int sendTo=0;
			int sendToQueue=0;
			int z = findTupleWithMinTimestamp(tuples);
			
			if (z==-1) break;//total tuples have been processed. should not reach this.
			
			XATTuple tuple = this.getInputQueue(z).dequeue(this);
			
			if (initTime ==0) initTime = tuple.getMinTimeStamp();
			
			//if (tuple.getTimeEnter()==0) tuple.setTimeEnter(System.currentTimeMillis());
			//tuple.setMaxTimeStamp(timeEnter);
			//tuple.setMinTimeStamp(timeEnter);
			
			x++;

			//propagate the tuple to one of the output queue
			tuple.setToNew(true);//identify the new stream tuple
			
			int sending = (this.numOfCopy*this.numOfCopy > this.numOfNodes-2)?this.numOfCopy*this.numOfCopy:(this.numOfNodes -2);
			
			boolean sent = false;
			for (int kk = 0;kk<this.numOfNodes-2;kk++){
				if (rand.nextDouble()*(this.numOfNodes-2)*this.numOfCopy<=sending) {
					sendTo=kk;
					sendToQueue = z+(this.numOfInputs)*sendTo;
					this.getOutputQueue(sendToQueue).enqueue(tuple, this);
					//System.out.println("Sent:"+z+":"+sendToQueue+":"+tuple);
					sent = true;
				}
			}
			//at least send one copy out
			if (!sent) this.getOutputQueue(z).enqueue(tuple, this);
			
			//System.out.println("Time:" +tuple.getMaxTimeStamp()+"\t"+"windowPartitionSize:"+windowPartitionSize);
			//System.out.println("sendTo:" +sendTo+"\t"+"SendToQueue"+sendToQueue);
			//this.getOutputQueue(sendToQueue).enqueue(tuple, this);
			//System.out.println(sendToQueue+":"+tuple);	
			//fill the buffer, kind of merge-sorting of all the tuples in the input queues.
			tempTuple = null;
			tempQueue = null;
			tempQueue = this.getInputQueue(z);
			if (tempQueue.getQueueSize()>0) {
					tempTuple = tempQueue.peek(this);
			}
			tuples[z]=tempTuple;
			
		}
		//stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

//		try {
//			inputQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + total);
//			setNumberOfPossibleOutputs();
//		} catch (Exception e) {
//			System.out.println(e);
//		}
			
		return true;
	}

	int findTupleWithMinTimestamp(XATTuple[] tuples){
		int ret = -1;
		XATTuple tempTuple = null;

		for (int i=0;i<tuples.length;i++){
			if (tuples[i]==null) continue;
			if (tempTuple==null ||tuples[i].getMaxTimeStamp()<tempTuple.getMaxTimeStamp()){
				ret =i;
				tempTuple = tuples[i];
			}
		}
		return ret;
	}
		
	
	
}
