/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Oct, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Vector;
/**
 * @author Song Wang
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class StreamWindowSplitOperatorImp extends XATSingleQueueStreamOperatorImp 
	{
	//added only for statistic collection for exp. can be safely removed
	StringBuffer stats_string;

	//array holding all the outputqueues to the outside
	//the two following arraylist must be modified consistantly.
	//that is: queryWindowArray[i] and queryOutputQueueArray[i] must describe the same query
	//the objects in below arraylist are the window sizes.
	protected ArrayList queryWindowArray = new ArrayList();
	//the objects in below arraylist are the outputqueue objects.
	protected ArrayList queryOutputQueueArray = new ArrayList();
	
	
	//indicates if the arraylist is already sorted or not.
	private boolean sorted = false;
	
	public ArrayList getWindowArray(){
		return queryWindowArray;
	}
	public StreamWindowSplitOperatorImp(XATNode init) {
		
		super(init);
		this.stats_string = new StringBuffer();
		//stats_string.append("StartTime\tEndTime\tOutputTuples\n");

	}

	public void insertQuery(int outputQueueID, long windowSize){
		queryWindowArray.add(new Long(windowSize));
		queryOutputQueueArray.add(this.getOutputQueue(outputQueueID));
		sortQueryByWindowSizeIncrease(); 
		this.sorted =true;
	}
	
	private void swapQuery(int position1, int position2){
		Object tempWindowSize = queryWindowArray.get(position1);
		Object tempOutputQueue = queryOutputQueueArray.get(position1);
		
		queryWindowArray.set(position1,queryWindowArray.get(position2));
		queryOutputQueueArray.set(position1,queryOutputQueueArray.get(position2));
		
		queryWindowArray.set(position2,tempWindowSize);
		queryOutputQueueArray.set(position2,tempOutputQueue);
		
	}
	
	//compare the window size, 
	//return 1 if  queryWindowArray[position1]>queryWindowArray[position2]
	//return 0 if  queryWindowArray[position1]=queryWindowArray[position2]
	//return -1 if  queryWindowArray[position1]<queryWindowArray[position2]
	private int compareWindowSize(int position1, int position2){
		long window1 = ((Long)queryWindowArray.get(position1)).longValue();
		long window2 = ((Long)queryWindowArray.get(position2)).longValue();
		if ( window1>window2) return 1;
		else if ( window1==window2)return 0;
		else return -1;
	}

	
	//quick sort
	private void sortQueryByWindowSizeIncrease(){
		quickSort(0,queryWindowArray.size()-1);
	}
		
	
	private void quickSort(int left, int right){
		if(right>left){			
			int pivot = partition(left,right);
		
			quickSort(left,pivot-1);
			quickSort(pivot+1,right);
		}
	}
	
	private int partition(int left, int right){
		int pivot=left;
		int temp=left;
	
		for (int i=left+1; i <= right; i++){
			if(compareWindowSize(temp,i)==1){
				pivot++;
				swapQuery(pivot,i);
			}
		}
		swapQuery(pivot,left);
		return pivot;	
	}
	
	//find the first index that the windowSize > tupleDeltaWindow
	//need consider the duplicated windowSize in the Array.
	protected int findQuery(long tupleDeltaWindow){
		return findQueryLinear(tupleDeltaWindow);
	}
	
	
	private int findQueryLinear(long tupleDeltaWindow){
		
		if(((Long)queryWindowArray.get(0)).longValue()>tupleDeltaWindow)
			return 0;
		
		int position=queryWindowArray.size()-1;
		
		for(int i=0;i<queryWindowArray.size()-1 ;i++){
			if(((Long)queryWindowArray.get(i+1)).longValue()>tupleDeltaWindow &&((Long)queryWindowArray.get(i)).longValue()<= tupleDeltaWindow){
					position=i+1;
					break; //out of loop
				}
		}
		return position;
	}
	
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() throws Exception {
		/**
		 * first figure out how many tuples to dequeue, the selectivity, and 
		 * result delay  
		 */
		XATQueueSchema inSchema;
		int inLength;
		if ((inSchema = this.getInputQueue().getSchema()) == null){
				return false;
		}
		inLength = inSchema.getLength();
		
		//modified by Song Wang to support multiple output queue
		for (int i=0;i<this.getOutputQueueCount();i++){
			if(this.getOutputQueue(i).getSchema() == null){
				init(i);
			}
		}
		
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		int totalOutput =0;
		//begin exp stat collect
		stats_string.append(System.currentTimeMillis()+"\t"); 
		
		/** get the tuples */
		XATTuple[] tuples = this.getInputQueue().dequeue(howMany, this);
		/** figure out how many tuples to keep*/
		for (int i=0; i<tuples.length; i++) {
			
			// for distributed migration
//			if(tuples[i].getMaxTimeStamp() == -1) {
//				this.processSyncTuple();
//				continue;
//			}
			

			long tupleWindowDelta = tuples[i].getMaxTimeStamp() -tuples[i].getMinTimeStamp() ;

			
			for (int j=findQuery(tupleWindowDelta); j<queryWindowArray.size(); j++) {
				((XATQueue)queryOutputQueueArray.get(j)).enqueue(tuples[i], this);
				totalOutput++;
			}
			
		}
			
//			//the following code is wrong!!!
//			XATTuple newTuple = new XATTupleImp(projAttrPosArray.length);
//			for (int j=0; j<projAttrPosArray.length; j++) {
//				newTuple.setCell(j, tuples[i].getCell(projAttrPosArray[j]));
//			}
//			newTuple.setTimeStamp(tuples[i].getTimeStamp());
//			
//			this.getOutputQueue().enqueue(newTuple, this);
//		}
		stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

		return true;
	}
	//overload supper method
	public void init(int i) {
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) {
			return;
		}
		XATQueueSchema outSchema = this.getOutputQueue(i).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) {
			this.getOutputQueue(i).setSchema(inSchema);
		}
		computeOutputPositions();//no need to overload since all the output queue are the same
		numOutputColumns = outputPositions[0].length;
		setHasInit(true);
		if (!sorted){sortQueryByWindowSizeIncrease(); this.sorted =true;}
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		return "Select";
	}

	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input queues.
	 * Inside method run() the numbers are computed based on the property
	 * "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler,
	 * while in cleanAccumulatedTuples, all tuples in input queues that have their max timesamp
	 * smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() {
		return true;
	}
	
	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() {
		return false;
	}

	public boolean matchStates(XATStreamOperator op) {
		return false;
	}
	
	public boolean recomputeUnmatchedState() {
		return true;
	}
	
	public int setStateID(int nextid) {
		return nextid;
	}
	public boolean finalizeOperation(){
		this.printStatsToFile();
		
		System.out.println("\n\nStatistic Collection Done\n");
		return true;
	}
	public void printStatsToFile() {
		try {
			String fileName = new String("WindowSplit_" + System.currentTimeMillis()+".txt");
			FileWriter out = new FileWriter(fileName, true);
			out.write(this.stats_string.toString());
			out.close();
			
		} catch (IOException e) {
			System.out.println(e.toString());
		}
}


}
