/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on 2005-3-13
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate;

import java.util.Vector;
import java.util.Iterator;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;

/**
 * @author lisading
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class GroupbyMultiValueHashIndexImp extends GroupbyIndexImp {
	protected static int HASH_FACTOR = 128;
	
	public GroupbyMultiValueHashIndexImp() {
		super();
	}

	//may have collisions
	public Object hash(Object[] keys) {
		int hashValue = 0;
		int numKeys = keys.length;
		for (int i=0; i<numKeys; i++) {
			int hashKey = ((Number)keys[i]).intValue();
			hashValue += hashKey * (2^i) * HASH_FACTOR; 
		}
		return new Integer(hashValue);
	}

	public Object getEntry(Object[] keys, Object[] gbKeys, 
			String nextIndexType, boolean leafLevel) {
		Object hashKey = hash(keys);
		Object obj = hashtable.get(hashKey);
		if (obj == null) {
			if (leafLevel) {
				HashBucket bucket = new HashBucket(gbKeys);
				hashtable.put(hashKey, bucket);
				return bucket;
			} else {
				GroupbyIndex index = null;
				try {
					index = (GroupbyIndex)Class.forName(nextIndexType).newInstance();
				} catch (Exception e) {
					e.printStackTrace();
				}
				index.setSearchKeys(keys);
				hashtable.put(hashKey, index);
				return index;
			}
		} else {
			return obj;
		}
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.GroupbyIndex#probe(java.lang.Object)
	 */
	public Vector probe(XATTuple tuple, Object[] gbKeys, boolean leafLevel, 
			int[] indexAttrPosArray, Object[] ancKeys, String nextIndexType, 
			boolean nextLeafLevelFlag, long invalidTime) {
		Vector result = new Vector();
		int numIndexAttrs = indexAttrPosArray.length;
		Object[] keys = new Object[numIndexAttrs];
		for (int i=0; i<numIndexAttrs; i++) {
			keys[i] = tuple.getCell(indexAttrPosArray[i]).getValue();
		}
		Object hashKey = hash(keys);
		Object obj = hashtable.get(hashKey);
		if (obj == null) { //corresponding next-level index doesn't exist
			if (leafLevel) {
				HashBucket bucket = new HashBucket(gbKeys);
				bucket.add(tuple);
				hashtable.put(hashKey, bucket);
			} else {
				GroupbyIndex index = null;
				try {
					index = (GroupbyIndex)Class.forName(nextIndexType).newInstance();
				} catch (Exception e) {
					e.printStackTrace();
				}
				index.setAncesterKeys(ancKeys); //for migration purpose
				index.setSearchKeys(keys);
				hashtable.put(hashKey, index);
				result.add(index);
			}
		} else {
			if (leafLevel) { //obj is a HashBucket of XATTuple
				HashBucket bucket = (HashBucket)obj;
				int numTuplesInBucket = bucket.size();
				for (int i=0; i<numTuplesInBucket; i++) {
					if (((XATTuple)bucket.get(0)).getMinTimeStamp()<invalidTime) {
						bucket.remove(0);
					} else {
						break;
					}
				}
				bucket.add(tuple);
			} else {
				result.add(obj);
			}
		}
		return result;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.GroupbyIndex#generateGroups(edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue[], int[])
	 */
	public Vector generateGroups(XATDataValue[] punctContent, int[] punctType, 
			int[] groupAttrPosArray, boolean leafLevel, int[] indexAttrPosArray,
			long invalidTime) {
		//check whether wildcard exists
		int numIndexAttrs = indexAttrPosArray.length;
		Object[] keys = new Object[numIndexAttrs];
		boolean wildcardExist = false;
		for (int i=0; i<numIndexAttrs; i++) {
			int index = indexAttrPosArray[i];
			if (punctType[index] == XATPunct.ALL_VALUE) {
				wildcardExist = true;
				break;
			} else {
				keys[i] = punctContent[index].getValue();
			}
		}

		Vector result = new Vector();
		if (wildcardExist) { //punctuation pattern contains wildcard
			Iterator it = hashtable.values().iterator();
			if (leafLevel) {
				while (it.hasNext()) {
					HashBucket bucket = (HashBucket)it.next();
					int numEntries = bucket.size();
					int i=0;
					for (; i<numEntries; i++) {
						XATTuple entry = (XATTuple)bucket.get(0);
						if (entry.getMinTimeStamp()<invalidTime) {
							bucket.remove(0);
						} else {
							break;
						}
					}
					numEntries = bucket.size();
					i=0;
					for (; i<numEntries; i++) {
						XATTuple entry = (XATTuple)bucket.get(i);
						if (entry.equalKeys(punctContent, punctType, groupAttrPosArray)) {
							result.add(entry);
						}
					}
				}
			} else {
				while (it.hasNext()) {
					result.add(it.next());
				}
			}
		} else { //punctuation pattern doesn't contain wildcard
			Object hashKey = hash(keys);
			Object obj = hashtable.get(hashKey);
			if (obj != null) {
				if (leafLevel) {
					HashBucket bucket = (HashBucket) obj;
					int numEntries = bucket.size();
					int i=0;
					for (; i<numEntries; i++) {
						XATTuple entry = (XATTuple)bucket.get(0);
						if (entry.getMinTimeStamp()<invalidTime) {
							bucket.remove(0);
						} else {
							break;
						}
					}
					numEntries = bucket.size();
					i=0;
					for (; i<numEntries; i++) {
						XATTuple entry = (XATTuple)bucket.get(i);
						if (entry.equalKeys(punctContent, punctType, groupAttrPosArray)) {
							result.add(entry);
						}
					}
				} else {
					result.add(obj);
				}
			}
		} 
		return result;
	}
}
