/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Aug 11, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import java.util.ArrayList;
import java.util.Hashtable;

/**
 * @author lisading
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class PunctHashDistinctStateImp extends OperatorStateImp 
	implements PunctAwareState {
	
	int[] indexAttrPosArray;
	Hashtable hashtable = new Hashtable();
	
	public PunctHashDistinctStateImp() {
		super();
	}
	public PunctHashDistinctStateImp(PunctHashDistinctStateImp oldstate, XATStreamOperator newop) {
		super(oldstate, newop);
		this.isMatched = false;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.PunctAwareState#probe(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple, int)
	 */
	public XATTuple[] probe(Object key) {
		return null;
	}

	public XATTuple[] probe(XATTuple tuple) {	
		XATDataValue[] thisTupleValue = tuple.getArrayData();
		Object indexValue = thisTupleValue[indexAttrPosArray[0]].getValue();
		//Object key = hash(indexValue);
		Object bucket = hashtable.get(indexValue);
		if (bucket == null) {
			//update the state
			ArrayList bucketList = new ArrayList();
			bucketList.add(tuple);
			
			//update state size
			this.increStateSize();
			
			hashtable.put(indexValue, bucketList);
			//return the result
			XATTuple[] results = new XATTuple[]{tuple};
			return results;
		} else {
			ArrayList bucketList = (ArrayList) bucket;
			int numTuplesInBucket = bucketList.size();
			boolean matchedTuple = false;
			for (int i=0; i<numTuplesInBucket; i++) {
				XATTuple nextTuple = (XATTuple)bucketList.get(i);
				XATDataValue[] valuesToCompare = nextTuple.getArrayData();
				//if (! valuesToCompare[indexAttrPos].getValue().equals(indexValue)) {
				//	continue;
				//}
				boolean unmatchedAttr = false;
				for (int j=0; j<valuesToCompare.length; j++) {
					if (j == indexAttrPosArray[0]) {
						continue;
					}
					if (!valuesToCompare[j].equals(thisTupleValue[j])) {
						unmatchedAttr = true;
						break;
					}
				}
				if (!unmatchedAttr) {
					matchedTuple = true;
					break;
				}
			}
			if (!matchedTuple) {
				//update the state
				bucketList.add(tuple);
				//update state size
				this.increStateSize();
				
				hashtable.put(indexValue, bucketList);
				//return the result
				XATTuple[] results = new XATTuple[]{tuple};
				return results;
			}
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.PunctAwareState#purgeByPunct(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple)
	 */
	public int purge(Object key) {
		// TODO Auto-generated method stub
		return 0;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.PunctAwareState#purgeByPunct(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple, int)
	 */
	public int purge(XATPunct punct) {
		int type = punct.getPunctType(indexAttrPosArray[0]);
		int numTuplesPurged = 0;
		if (type == XATPunct.SINGLE_VALUE) {
			Object punctValue = punct.getCell(indexAttrPosArray[0]).getValue();
			//update state size
			Object bucket = hashtable.get(punctValue);
			if (bucket != null) {
				numTuplesPurged = ((ArrayList)bucket).size();
				this.decreStateSize(numTuplesPurged);
			}
			hashtable.remove(punctValue);
		} else if (type == XATPunct.SET_VALUE) {
			ArrayList punctList = (ArrayList) punct.getCell(indexAttrPosArray[0]);
			for (int i=0; i<punctList.size(); i++) {
				Object value = punctList.get(i);
				//update state size
				Object bucket = hashtable.get(value);
				if (bucket != null) {
					numTuplesPurged += ((ArrayList)bucket).size();
				}
				hashtable.remove(value);
			}
			this.decreStateSize(numTuplesPurged);
		} else if (type == XATPunct.RANGE_VALUE) {
				//to be finished
		} else if (type == XATPunct.ALL_VALUE || type == XATPunct.NULL_VALUE) {
		}
		return numTuplesPurged;
	}

	public Object[] invalidate(long latestProcessTupleTime) {
		return null;
	}
	public void copyStateContent(OperatorState newstate) {
	}

	public Object getAllStateTuples() {
		return hashtable;
	}
	public void setIndexAttrPosArray(int[] indexAttrPosArray) {
		int numAttrs = indexAttrPosArray.length;
		this.indexAttrPosArray = new int[numAttrs];
		for (int i=0; i<numAttrs; i++) {
			this.indexAttrPosArray[i] = indexAttrPosArray[i];
		}
	}
	public void clearState() {
		// TODO Auto-generated method stub
		
	}
}
