/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on 2005-8-17
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import java.util.Vector;
import java.util.Iterator;

/**
 * @author lisading
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class PunctHybridIndexGroupbyStateImp extends OperatorStateImp 
	implements PunctAwareState {

	protected GroupbyIndex firstLevelIndex;
	protected int numLevels; //# of levels of index structure
	protected String[] indexTypeArray; //index type, sequentially listed, from first level
	protected int numGroupAttrs; //# of grouping attributes
	protected int[] groupAttrPosArray; //all grouping attributes, in increasing order of their position
	protected int[][] indexAttrPosArray;
	protected long window;
	//used for state move migration 
	protected int[][] migrationAttrMatchArray;

	/**
	 * 
	 */
	public PunctHybridIndexGroupbyStateImp() {
	}
	
	/**
	 * @param gAttrPosArray
	 */
	public void setGroupAttrPosArray(int[] gAttrPosArray) {
		groupAttrPosArray = gAttrPosArray;
		numGroupAttrs = groupAttrPosArray.length;
	}
	/**
	 * @param iTypeArray
	 */
	public void setIndexTypeArray(String[] iTypeArray) {
		indexTypeArray = iTypeArray;
	}
	/**
	 * @param iAttrPosArray
	 */
	public void setIndexAttrPosArray(int[][] iAttrPosArray) {
		indexAttrPosArray = iAttrPosArray;
		numLevels = iAttrPosArray.length;
	}
	/**
	 * @param win
	 */
	public void setWindow(long win) {
		window = win;
	}
	public int getNumLevels() {
		return numLevels;
	}
	/**
	 * 
	 */
	public void init() {
		try {
			firstLevelIndex = (GroupbyIndexImp) Class.forName(indexTypeArray[0]).newInstance();
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.PunctAwareState#probe(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple)
	 */
	public XATTuple[] probe(XATTuple tuple) {
		//temp variables
		Object[] gbKeys = new Object[numGroupAttrs];
		for (int i=0; i<numGroupAttrs; i++) {
			gbKeys[i] = tuple.getCell(groupAttrPosArray[i]).getValue();
		}
		
		Vector v_index = new Vector();
		boolean leafLevel = false, nextLeafLevelFlag = false;
		String nextIndexType = null;
		long invalidTime = Math.max(tuple.getMinTimeStamp()-window, 0);

		v_index.add(firstLevelIndex);
		int numAncKeys = 0;
		Object[] ancKeys = null;
		for (int i=0; i<numLevels; i++) {
			if (i < numLevels - 1) {
				if (i == numLevels-2) {
					nextLeafLevelFlag = true;
				} else {
					nextLeafLevelFlag = false;
				}
				nextIndexType = indexTypeArray[i+1];
			} else {
				leafLevel = true;
			}
			
			Vector v_next_index = new Vector();
			int numIndex = v_index.size();
			for (int k=0; k<numIndex; k++) {
				GroupbyIndex index = (GroupbyIndex)v_index.get(k);
				Vector result = index.probe(tuple, gbKeys, leafLevel, 
						indexAttrPosArray[i], ancKeys, nextIndexType, 
						nextLeafLevelFlag, invalidTime);
				if (result.size() > 0) {
					v_next_index.addAll(result);
				}
			}
			v_index = v_next_index;
			
			//compute ancestor keys (for migration purpose), extra overhead
			if (i < (numLevels-1)) {
				int numIndexAttrs = indexAttrPosArray[i].length;
				Object[] nextAncKeys = new Object[numAncKeys + numIndexAttrs];
				int j=0;
				for (; j<numAncKeys; j++) {
					nextAncKeys[j] = ancKeys[j];
				}
				for (int k=0; k<numIndexAttrs; k++) {
					nextAncKeys[k+j] = tuple.getCell(indexAttrPosArray[i][k]).getValue();
				}
				ancKeys = nextAncKeys;
				numAncKeys += numIndexAttrs;
			}
		}
		v_index.clear();
		return null;
	}

	/**
	 * @param punct
	 * @return
	 */
	public Vector generateGroups(XATPunct punct) {
		XATDataValue[] punctContent = punct.getArrayData();
		int[] punctType = punct.getPunctTypeArray();
		long invalidTime = Math.max(punct.getMinTimeStamp() - window, 0);
		
		boolean leafLevel = false;
		Vector v_index = new Vector();
		v_index.add(firstLevelIndex);
		for (int i=0; i<numLevels; i++) {
			if (i == numLevels - 1) {
				leafLevel = true;
			}
			Vector v_next_index = new Vector();
			int numIndex = v_index.size();
			for (int j=0; j<numIndex; j++) {
				GroupbyIndex index = (GroupbyIndex)v_index.get(j);
				Vector result = index.generateGroups(punctContent, punctType, 
						groupAttrPosArray, leafLevel, indexAttrPosArray[i], invalidTime);
				if (result.size() > 0) {
					v_next_index.addAll(result);
				}
			}
			if (v_next_index.size() == 0) {
				return v_next_index;
			} else {
				v_index = v_next_index;
			}
		}
		return v_index;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.PunctAwareState#probe(java.lang.Object)
	 */
	public XATTuple[] probe(Object key) {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.PunctAwareState#purge(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct)
	 */
	public int purge(XATPunct punct) {
		// TODO Auto-generated method stub
		return 0;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.PunctAwareState#purge(java.lang.Object)
	 */
	public int purge(Object key) {
		// TODO Auto-generated method stub
		return 0;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.PunctAwareState#invalidate(long)
	 */
	public Object[] invalidate(long earliestProcessTupleTime) {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.PunctAwareState#getAllStateTuples()
	 */
	public Object getAllStateTuples() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.PunctAwareState#setIndexAttrPosArray(int[])
	 */
	public void setIndexAttrPosArray(int[] indexAttrPosArray) {
		// TODO Auto-generated method stub
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState#copyStateContent(edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState)
	 */
	public void copyStateContent(OperatorState newstate) {
		// TODO Auto-generated method stub
	}
	
	
	/////////////// methods used for migration ///////////////
	/**
	 * level - start from 0
	 * @return
	 */
	public Vector collectIndexObjects(int level) {
		Vector v_index = new Vector();
		v_index.add(firstLevelIndex);
		for (int i=0; i<level; i++) {
			Vector v_next_index = new Vector();
			int numIndex = v_index.size();
			for (int j=0; j<numIndex; j++) {
				GroupbyIndex index = (GroupbyIndex)v_index.get(j);
				Iterator it = index.iterator();
				while (it.hasNext()) {
					v_next_index.add(it.next());
				}
			}
			if (v_next_index.size() == 0) {
				return v_next_index;
			} else {
				v_index = v_next_index;
			}
		}
		return v_index;
	}

	/**
	 * @param bucketList
	 */
	public void insertTuples(Vector bucketList) {
		int numBuckets = bucketList.size();
		for (int i=0; i<numBuckets; i++) {
			HashBucket bucket = (HashBucket)bucketList.get(i);
			Object[] gbKeys = bucket.getKeys();
			int numTuples = bucket.size();
			Vector tupleList = bucket.getObjList();
			for (int j=0; j<numTuples; j++) {
				XATTuple tuple = (XATTuple)tupleList.get(j);
				GroupbyIndex index = firstLevelIndex;
				for (int l=0; l<numLevels; l++) {
					int numSearchKeys = indexAttrPosArray[l].length;
					Object[] keys = new Object[numSearchKeys];
					for (int k=0; k<numSearchKeys; k++) {
						keys[k] = tuple.getCell(indexAttrPosArray[l][k]).getValue();
					}
					if (l == numLevels - 1) {
						HashBucket bkt = (HashBucket)index.getEntry(keys, gbKeys, indexTypeArray[numLevels - 1], true);
						bkt.addTupleByTime(tuple); //guarantee order of timestamp
					} else {
						index = (GroupbyIndex)index.getEntry(keys, null, indexTypeArray[l+1], false);
					}
				} //end of iterating index levels
			} //end of iterating tuples in a bucket
		} //end of iterating buckets in bucket list
	}

	/**
	 * configuration of migrationAttrMatchArray: 
	 * if groupAttrPosArray is 1; 2; 3 and 
	 * indexAttrPosArray of new state is 1; 3; 2 
	 * then migrationAttrMatchArray should be 0; 2; 1
	 * which means attr used for level 0 in new state 
	 * is the 0'th grouping attribute
	 * 
	 * @param bucketList
	 */
	public void insertBuckets(Vector bucketList) {
		int numBuckets = bucketList.size();
		for (int i=0; i<numBuckets; i++) {
			HashBucket bucket = (HashBucket)bucketList.get(i);
			Object[] bucketKeys = bucket.getKeys();
			if (numLevels == 1) {
				int numSearchKeys = migrationAttrMatchArray[0].length;
				Object[] keys = new Object[numSearchKeys];
				for (int k=0; k<numSearchKeys; k++) {
					keys[k] = bucketKeys[migrationAttrMatchArray[0][k]];
				}
				Object hashKey = firstLevelIndex.hash(keys);
				firstLevelIndex.put(hashKey, bucket);
			} else {
				GroupbyIndex index = firstLevelIndex;
				Object[] keys = null;
				for (int j=0; j<numLevels-1; j++) {
					int numSearchKeys = migrationAttrMatchArray[j].length;
					keys = new Object[numSearchKeys];
					for (int k=0; k<numSearchKeys; k++) {
						keys[k] = bucketKeys[migrationAttrMatchArray[j][k]];
					}
					index = (GroupbyIndex)index.getEntry(keys, null, indexTypeArray[j+1], false);
				}
				int numSearchKeys = migrationAttrMatchArray[numLevels-1].length;
				keys = new Object[numSearchKeys];
				for (int k=0; k<numSearchKeys; k++) {
					keys[k] = bucketKeys[migrationAttrMatchArray[numLevels-1][k]];
				}
				Object hashKey = index.hash(keys);
				index.put(hashKey, bucket);
			}
		}
	}
	
	/**
	 * condition to use: >2 pure hierarchical index levels
	 * 
	 * configuration of migrationAttrMatchArray: 
	 * if indexAttrPosArray of old state is 0; 2; 1 
	 * and indexAttrPosArray of new state is 1; 2; 0,
	 * then migrationAttrMatchArray should be 2; 1; 0,
	 * which means attr used for level 0 in new state 
	 * was at level 2 in old state

	 * @param indexList
	 */
	public void insertIndexes(Vector indexList) {
		int numIndexes = indexList.size();
		for (int i=0; i<numIndexes; i++) {
			GroupbyIndex index = (GroupbyIndex)indexList.get(i);
			Object[] ancKeys = index.getAncesterKeys();
			int numAncKeys = ancKeys.length;
			Object[] searchKeys = index.getSearchKeys();
			int numSearchKeys = searchKeys.length;
			int numProbeKeys = numAncKeys + numSearchKeys;
			Object[] probeKeys = new Object[numProbeKeys];
			int j=0;
			for (; j<numAncKeys; j++) {
				probeKeys[j] = ancKeys[j];
			}
			for (int k=0; k<numSearchKeys; k++) {
				probeKeys[k+j] = searchKeys[k];
			}
				
			int levelNo = 0;
			GroupbyIndex nextIndex = firstLevelIndex;
			int numKeys = migrationAttrMatchArray[levelNo].length;
			Object[] keys = new Object[numKeys];
			for (int k=0; k<numKeys; k++) {
				keys[k] = probeKeys[migrationAttrMatchArray[levelNo][k]];
			}
			
			ancKeys = null;
			numAncKeys = 0;
			Object[] newAncKeys = null;
			int numAncKeysUsed = numKeys;
			while (numAncKeysUsed < numProbeKeys) {
				nextIndex = (GroupbyIndex)nextIndex.getEntry(keys, null,
						indexTypeArray[levelNo+1], false);
				nextIndex.setSearchKeys(keys);
				nextIndex.setAncesterKeys(ancKeys);

				newAncKeys = new Object[numAncKeys+numKeys];
				int l=0;
				for (; l<numAncKeys; l++) {
					newAncKeys[l] = ancKeys[l];
				}
				for (int k=0; k<numKeys; k++) {
					newAncKeys[k+l] = keys[k];
				}
				ancKeys = newAncKeys;
				numAncKeys += numKeys;

				levelNo ++;
				numKeys = migrationAttrMatchArray[levelNo].length;
				keys = new Object[numKeys];
				for (int k=0; k<numKeys; k++) {
					keys[k] = probeKeys[migrationAttrMatchArray[levelNo][k]];
				}
				numAncKeysUsed += numKeys;
			}
			index.setAncesterKeys(newAncKeys); //set new ancester key for index
			index.setSearchKeys(keys); //set new search key for index
			Object hashKey = nextIndex.hash(keys);
			nextIndex.put(hashKey, index);
		}
	}
			
	/**
	 * @param mAttrMatchArray
	 */
	public void setMigrationAttrMatchArray(int[][] mAttrMatchArray) {
		migrationAttrMatchArray = mAttrMatchArray;
	}

	public void clearState() {
		// TODO Auto-generated method stub
		
	}
}
