/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Aug 13, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import java.util.ArrayList;

/**
 * @author lisading
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class BinaryPWSingleEquiJoinOperatorImp 
	extends XATMultiQueueWindowStreamOperatorImp 
	implements BinaryPWSingleEquiJoinOperator {

	int LEFT = 0;
	int RIGHT = 1;
	
	protected boolean propagateFlag;
	protected int memoryLimit;

	int JOINANDDROPTUPLE = 1;
	int JOINANDINSERTTUPLE = 0;
	int DROPTUPLE = -1;
	
	int PROPAGATEPUNCT = 1;
	int DROPPUNCT = -1;
	int DONOTHING = 0;

	SinglePWSingleEquiJoinState leftState;
	SinglePWSingleEquiJoinState rightState;

	int leftJoinAttrPos;
	int rightJoinAttrPos;
	
	int invalidateThreshold;
	int purgeThreshold;

	long leftLatestProcessTupleTime;
	long rightLatestProcessTupleTime;

	int numNewTuplesSinceLastInvalidate = 0;
	int numLeftPunctsSinceLastPurge = 0;
	int numRightPunctsSinceLastPurge = 0;

	public BinaryPWSingleEquiJoinOperatorImp(XATNode init) {
		super(init);
	}

	public boolean run() throws Exception {
		if (!hasInit()) {
			init();
			if (!hasInit()) {
				return false;
			}
		}

		XATTuple newObj = null;
		int indication = LEFT;
		int tupleCount = 0;
	
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		while (true) {
			if (tupleCount == howMany) {
				return true;
			}
			try {
				//pick the next tuple to process
				if (this.getLeftInputQueue().getQueueSize(this) == 0) {
					newObj = this.getRightInputQueue().dequeue(this);
					indication = RIGHT;
				} else if (this.getRightInputQueue().getQueueSize(this) == 0) {
					newObj = this.getLeftInputQueue().dequeue(this);
					indication = LEFT;
				} else {
					long leftTimeStamp = 0;
					long rightTimeStamp = 0;
		
					leftTimeStamp =
						this.getLeftInputQueue().peek(this).getNthTimeStamp(this.leftJoinAttrPos);
					rightTimeStamp = 
						this.getRightInputQueue().peek(this).getNthTimeStamp(this.rightJoinAttrPos);
						
					if(leftTimeStamp <= rightTimeStamp) {
						newObj = this.getLeftInputQueue().dequeue(this);
						indication = LEFT;
					} else {
						newObj = this.getRightInputQueue().dequeue(this);
						indication = RIGHT;
					}					
				}
			} catch(XATQueueException e1) {
				return true;
			}
	
			if (indication == LEFT) {
				if (newObj.isPunct()) { //receive a new punctuation
					if (newObj.getCell(this.leftJoinAttrPos) == null) {
						continue;
					}
					boolean returnFlag1 = this.leftState.processPunct(newObj);
					if (returnFlag1) {
						Object punctVal = newObj.getCell(leftJoinAttrPos).getValue();
						if (this.propagateFlag) {
							propagateJoinValue(punctVal);
						}
						this.rightState.markAsPropagated(punctVal);
					}
					
					numLeftPunctsSinceLastPurge ++;
					if (numLeftPunctsSinceLastPurge == this.purgeThreshold) {
						ArrayList unpurgedPunctList = this.leftState.getUnpurgedPunctList();
						int cursor = 0;
						while (cursor<unpurgedPunctList.size()) {
							Object value = ((XATTuple)unpurgedPunctList.get(cursor)).getCell(leftJoinAttrPos).getValue();
							int returnFlag2 = this.rightState.purgeByPunct(value);
							if (returnFlag2 == this.PROPAGATEPUNCT) {
								if (this.propagateFlag) {
									propagateJoinValue(value);
								}
								this.leftState.getIndex().removeIndexNode(value);
								unpurgedPunctList.remove(cursor);
							} else if (returnFlag2 == this.DROPPUNCT) {
								this.leftState.getIndex().removeIndexNode(value);
								unpurgedPunctList.remove(cursor);
							} else {
								cursor ++;
							}
						}
						
						this.leftState.mergePunctListsAsPurged();
						numLeftPunctsSinceLastPurge = 0;
					} 
				} else {
					tupleCount ++;
					leftLatestProcessTupleTime = newObj.getNthTimeStamp(this.leftJoinAttrPos);
					numNewTuplesSinceLastInvalidate ++;
					if (numNewTuplesSinceLastInvalidate == this.invalidateThreshold) {
						Object[] propPunctVals = leftState.purgeByWindow(rightLatestProcessTupleTime-getWindowSize(0));
						
						if (propPunctVals!=null && this.propagateFlag) {
							for (int i=0; i<propPunctVals.length; i++) {
								this.propagateJoinValue(propPunctVals[i]);
							}
						}
						propPunctVals = rightState.purgeByWindow(leftLatestProcessTupleTime-getWindowSize(1));
						
						if (propPunctVals!=null && this.propagateFlag) {
							for (int i=0; i<propPunctVals.length; i++) {
								this.propagateJoinValue(propPunctVals[i]);
							}
						}
						numNewTuplesSinceLastInvalidate = 0;
					}
					Object value = newObj.getCell(leftJoinAttrPos).getValue();
					
					ResultPackage resultPackage = rightState.probe(value);
					if (resultPackage == null) {
						leftState.addTupleToIndex(value, newObj);
						continue;
					}
					int resultFlag = resultPackage.getDropFlag();
					if (resultFlag != this.DROPTUPLE) {
						XATTuple[] matchedTuples = resultPackage.getResultTuples();
						if (matchedTuples != null) {
							for (int i=0; i<matchedTuples.length; i++) {
								XATTuple resultTuple = this.generateResultTuple(
										new XATTuple[] {newObj, matchedTuples[i]});
								this.getOutputQueue().enqueue(resultTuple, this);
							}
						} 
						if (resultFlag == this.JOINANDINSERTTUPLE) {
							leftState.addTupleToIndex(value, newObj);
						}
					}
				}
			} else {
				if (newObj.isPunct()) { //receive a new punctuation
					if (newObj.getCell(this.rightJoinAttrPos) == null) {
						continue;
					}
					boolean returnFlag1 = this.rightState.processPunct(newObj);
					if (returnFlag1) {
						Object punctVal = newObj.getCell(rightJoinAttrPos).getValue();
						if (this.propagateFlag) {
							propagateJoinValue(punctVal);
						}
						this.leftState.markAsPropagated(punctVal);
					}
					
					numRightPunctsSinceLastPurge ++;
					if (numRightPunctsSinceLastPurge == this.purgeThreshold) {
						ArrayList unpurgedPunctList = this.rightState.getUnpurgedPunctList();
						int cursor = 0;
						while (cursor<unpurgedPunctList.size()) {
							Object value = ((XATTuple)unpurgedPunctList.get(cursor)).getCell(rightJoinAttrPos).getValue();
							int returnFlag2 = this.leftState.purgeByPunct(value);
							if (returnFlag2 == this.PROPAGATEPUNCT) {
								if (this.propagateFlag) {
									propagateJoinValue(value);
								}
								this.rightState.getIndex().removeIndexNode(value);
								unpurgedPunctList.remove(cursor);
							} else if (returnFlag2 == this.DROPPUNCT) {
								this.rightState.getIndex().removeIndexNode(value);
								unpurgedPunctList.remove(cursor);
							} else {
								cursor ++;
							}
						}
						
						this.rightState.mergePunctListsAsPurged();
						numRightPunctsSinceLastPurge = 0;
					} 
				} else {
					tupleCount ++;
					rightLatestProcessTupleTime = newObj.getNthTimeStamp(this.rightJoinAttrPos);
					numNewTuplesSinceLastInvalidate ++;
					if (numNewTuplesSinceLastInvalidate == this.invalidateThreshold) {
						Object[] propPunctVals = leftState.purgeByWindow(rightLatestProcessTupleTime-getWindowSize(0));
						
						if (propPunctVals!=null && this.propagateFlag) {
							for (int i=0; i<propPunctVals.length; i++) {
								this.propagateJoinValue(propPunctVals[i]);
							}
						}
						propPunctVals = rightState.purgeByWindow(leftLatestProcessTupleTime-getWindowSize(1));
						
						if (propPunctVals!=null && this.propagateFlag) {
							for (int i=0; i<propPunctVals.length; i++) {
								this.propagateJoinValue(propPunctVals[i]);
							}
						}
						numNewTuplesSinceLastInvalidate = 0;
					}
					Object value = newObj.getCell(rightJoinAttrPos).getValue();

					ResultPackage resultPackage = leftState.probe(value);
					if (resultPackage == null) {
						rightState.addTupleToIndex(value, newObj);
						continue;
					}
					int resultFlag = resultPackage.getDropFlag();
					if (resultFlag != this.DROPTUPLE) {
						XATTuple[] matchedTuples = resultPackage.getResultTuples();
						if (matchedTuples != null) {
							for (int i=0; i<matchedTuples.length; i++) {
								XATTuple resultTuple = this.generateResultTuple(
										new XATTuple[] {matchedTuples[i], newObj});
								this.getOutputQueue().enqueue(resultTuple, this);
							}
						} 
						
						if (resultFlag == this.JOINANDINSERTTUPLE) {
							rightState.addTupleToIndex(value, newObj);
						}
					}
				}
			}
		}
	}
	
	private void propagateJoinValue(Object value) {
		XATDataValue[] newPunctContent = new XATDataValueImp[numOutputColumns];
		int[] punctTypeArray = new int[numOutputColumns];
		int j=0, k=0;
		for (;j<numOutputColsPerStream[0]; j++) {
			if (outputPositions[0][j]==leftJoinAttrPos) {
				newPunctContent[j] = new XATDataValueImp(value);
				punctTypeArray[j] = XATPunct.SINGLE_VALUE;
			} else {
				newPunctContent[j] = new XATDataValueImp("*");
				punctTypeArray[j] = XATPunct.ALL_VALUE;
			}
		}
		for (; k<numOutputColsPerStream[1]; k++) {
			if (outputPositions[1][k]==rightJoinAttrPos) {
				newPunctContent[k+j] = new XATDataValueImp(value);
				punctTypeArray[k+j] = XATPunct.SINGLE_VALUE;
			} else {
				newPunctContent[k+j] = new XATDataValueImp("*");
				punctTypeArray[k+j] = XATPunct.ALL_VALUE;
			}
		}
		XATPunct propPunct = new XATPunctImp(newPunctContent);
		propPunct.setPunctTypeArray(punctTypeArray);
		this.getOutputQueue().enqueue(propPunct, this);
	} //?? timestamp of punctuation ??

	public void setState(OperatorState[] state) {
		this.states = new OperatorState[state.length];
		for(int i = 0; i < this.states.length;i++){
			this.states[i] = state[i];
		}
		this.states = state;
		this.leftState = (SinglePWSingleEquiJoinState) state[0];
		this.rightState = (SinglePWSingleEquiJoinState) state[1];
	}

	public int getLeftJoinAttrPos() {
		return this.leftJoinAttrPos;
	}
	public int getRightJoinAttrPos() {
		return this.rightJoinAttrPos;
	}
	public int getInvalidateThreshold() {
		return this.invalidateThreshold;
	}
	public int getPurgeThreshold() {
		return this.purgeThreshold;
	}
	public void setLeftJoinAttrPos(int lJoinAttrPos) {
		this.leftJoinAttrPos = lJoinAttrPos;
	}
	public void setRightJoinAttrPos(int rJoinAttrPos) {
		this.rightJoinAttrPos = rJoinAttrPos;
	}
	public void setInvalidateThreshold(int iThreshold) {
		this.invalidateThreshold = iThreshold;
	}
	public void setPurgeThreshold(int pThreshold) {
		this.purgeThreshold = pThreshold;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		return "PWJoin";
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#hasUnmatchedState()
	 */
	public boolean hasUnmatchedState() {
		return false;
	}
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}
	public boolean cleanAllAccumulatedTuples() {
		return true;
	}
	public int setStateID(int startID) {
		return 0;
	}
	public boolean recomputeUnmatchedState() {
		return true;
	}
	public boolean matchStates(XATStreamOperator op) {
		return true;
	}
	public boolean finalizeOperation(){
		return true;
	}
	public boolean getPropagateFlag() {
		return propagateFlag;
	}

	public void setPropagateFlag(boolean pFlag) {
		propagateFlag = pFlag;
	}
	
	public int getMemoryLimit() {
		return memoryLimit;
	}
	public void setMemoryLimit(int memLimit) {
		memoryLimit = memLimit;
	}
}
