/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Sep 6, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstreamoperator;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.Random;

/**
 * @author lisading
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class CorrelatedQueryStreamGenerator extends QueryStreamGenerator {
	
	public CorrelatedQueryStreamGenerator() {
		super();
	}
	
	public CorrelatedQueryStreamGenerator(int totalNumTuples, int numAttrs, 
			int[] startValue, int[] numDistinctValues, 
			double numTuplesBetweenQuery, String dataFileName, int[] seed) {
		
		super(totalNumTuples, numAttrs, startValue, numDistinctValues, 
				numTuplesBetweenQuery, dataFileName, seed);
	}

	public void generateUniformQueryData() {
		// assume the first attribute functionally determines the rest attributes
		int[][] oAttrs = new int[numDistinctValues[0]][numAttrs];
		
		for (int i=0; i<numDistinctValues[0]; i++) {
			oAttrs[i][0] = candidateValues[0][i];
			for (int j=1; j<numAttrs; j++) {
				int randomIndex = dataAttrValIndexGen[j].nextInt(numDistinctValues[j]);
				oAttrs[i][j] = candidateValues[j][randomIndex];
			}
		}

		Random qPatternRan = new Random(); //for choosing query patterns
		
		try {
			BufferedWriter writer = new BufferedWriter(new FileWriter(dataFileName));
			int numTuplesGenerated = 0;
			int numTuplesInCurrentBatch = -1, numQueriesInCurrentBatch = 1;
			if (numTuplesBetweenQuery >= 1) {
				numTuplesInCurrentBatch = (int)numTuplesBetweenQuery;
			} else if (numTuplesBetweenQuery == 0){
				numQueriesInCurrentBatch = 0;
			} else {
				numQueriesInCurrentBatch = (int)(1/numTuplesBetweenQuery);
			}

			boolean genQuery = false;
			while(numTuplesGenerated < totalNumTuples) {
				int dataValIndex = dataAttrValIndexGen[0].nextInt(numDistinctValues[0]);
				
				for (int i=0; i<numAttrs-1; i++) {
					writer.write(oAttrs[dataValIndex][i] + fieldDelim);
				}
				writer.write(oAttrs[dataValIndex][numAttrs-1] + recordDelim);
				writer.flush();
				numTuplesGenerated ++;

				if (numTuplesInCurrentBatch>0) {
					numTuplesInCurrentBatch --;
					if (numTuplesInCurrentBatch == 0) {
						genQuery = true;
						numTuplesInCurrentBatch = (int)numTuplesBetweenQuery;
					} else {
						continue;
					}
				} else {
					genQuery = true;
				}

				if (!genQuery) {
					continue;
				}
				//generate queries -- needs to be worked on
				for (int i=0; i<numQueriesInCurrentBatch; i++) {
					writer.write(punctIndicator);
					int qPatternIndex = qPatternRan.nextInt(numQueryPatterns);
					int[] qPattern = queryPatterns[qPatternIndex];
					int queryValIndex;
					boolean gotPattern = false;
					int[] attrValues = null;
					for (int j=0; j<numAttrs-1; j++) {
						if (qPattern[j] == 0) {
							writer.write(wildcard + fieldDelim);
						} else {
							if (gotPattern) {
								writer.write(attrValues[j] + fieldDelim);
							} else {
								queryValIndex = dataAttrValIndexGen[0].nextInt(numDistinctValues[0]);
								attrValues = oAttrs[queryValIndex];
								writer.write(attrValues[j] + fieldDelim);
								gotPattern = true;
							}
						}
					}
					if (qPattern[numAttrs-1] == 0) {
						writer.write(wildcard + recordDelim);
					} else {
						if (gotPattern) {
							writer.write(attrValues[numAttrs-1] + fieldDelim);
						} else {
							queryValIndex = dataAttrValIndexGen[0].nextInt(numDistinctValues[0]);
							attrValues = oAttrs[queryValIndex];
							writer.write(attrValues[numAttrs-1] + fieldDelim);
						}
					}
					writer.flush();
				}
				genQuery = false;
			}
			writer.close();
			System.out.println("Done.");
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

	public void generateZipfQueryData(double alpha) {
		// assume the first attribute functionally determines the rest attributes
		int[][] oAttrs = new int[numDistinctValues[0]][numAttrs];
		
		for (int i=0; i<numDistinctValues[0]; i++) {
			oAttrs[i][0] = candidateValues[0][i];
			for (int j=1; j<numAttrs; j++) {
				int randomIndex = dataAttrValIndexGen[j].nextInt(numDistinctValues[j]);
				oAttrs[i][j] = candidateValues[j][randomIndex];
			}
		}

/** CHUAN LEI - No ZipfDistribution found in projects*/
//		ZipfDistribution qPatternRan = new ZipfDistribution(0, numQueryPatterns-1, alpha, 1); //for choosing query patterns
		
		try {
			BufferedWriter writer = new BufferedWriter(new FileWriter(dataFileName));
			int numTuplesGenerated = 0;
			int numTuplesInCurrentBatch = -1, numQueriesInCurrentBatch = 1;
			if (numTuplesBetweenQuery >= 1) {
				numTuplesInCurrentBatch = (int)numTuplesBetweenQuery;
			} else if (numTuplesBetweenQuery == 0){
				numQueriesInCurrentBatch = 0;
			} else {
				numQueriesInCurrentBatch = (int)(1/numTuplesBetweenQuery);
			}

			boolean genQuery = false;
			while(numTuplesGenerated < totalNumTuples) {
				int dataValIndex = dataAttrValIndexGen[0].nextInt(numDistinctValues[0]);
				
				for (int i=0; i<numAttrs-1; i++) {
					writer.write(oAttrs[dataValIndex][i] + fieldDelim);
				}
				writer.write(oAttrs[dataValIndex][numAttrs-1] + recordDelim);
				writer.flush();
				numTuplesGenerated ++;

				if (numTuplesInCurrentBatch>0) {
					numTuplesInCurrentBatch --;
					if (numTuplesInCurrentBatch == 0) {
						genQuery = true;
						numTuplesInCurrentBatch = (int)numTuplesBetweenQuery;
					} else {
						continue;
					}
				} else {
					genQuery = true;
				}

				if (!genQuery) {
					continue;
				}
/** CHUAN LEI - Comment out as ZipfDistribution is not found*/
				//generate queries -- needs to be worked on
//				for (int i=0; i<numQueriesInCurrentBatch; i++) {
//					writer.write(punctIndicator);
//					int qPatternIndex = qPatternRan.random_variate();
//					int[] qPattern = queryPatterns[qPatternIndex];
//					int queryValIndex;
//					boolean gotPattern = false;
//					int[] attrValues = null;
//					for (int j=0; j<numAttrs-1; j++) {
//						if (qPattern[j] == 0) {
//							writer.write(wildcard + fieldDelim);
//						} else {
//							if (gotPattern) {
//								writer.write(attrValues[j] + fieldDelim);
//							} else {
//								queryValIndex = dataAttrValIndexGen[0].nextInt(numDistinctValues[0]);
//								attrValues = oAttrs[queryValIndex];
//								writer.write(attrValues[j] + fieldDelim);
//								gotPattern = true;
//							}
//						}
//					}
//					if (qPattern[numAttrs-1] == 0) {
//						writer.write(wildcard + recordDelim);
//					} else {
//						if (gotPattern) {
//							writer.write(attrValues[numAttrs-1] + fieldDelim);
//						} else {
//							queryValIndex = dataAttrValIndexGen[0].nextInt(numDistinctValues[0]);
//							attrValues = oAttrs[queryValIndex];
//							writer.write(attrValues[numAttrs-1] + fieldDelim);
//						}
//					}
//					writer.flush();
//				}
//				genQuery = false;
			}
			writer.close();
			System.out.println("Done.");
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

	public static void main(String[] args) {
		int total_num_tuples = Integer.parseInt(args[0]);
		int num_attrs = Integer.parseInt(args[1]);
		int cursor = 2;
		int[] start_value = new int[num_attrs];
		for (int i=0; i<num_attrs; i++) {
			start_value[i] = Integer.parseInt(args[cursor]);
			cursor ++;
		}
		int[] num_distinct_values = new int[num_attrs];
		for (int i=0; i<num_attrs; i++) {
			num_distinct_values[i] = Integer.parseInt(args[cursor]);
			cursor ++;
		}
		double num_tuples_betw_query = Double.parseDouble(args[cursor]);
		cursor ++;

		String data_file_name = args[cursor];
		cursor ++;
		
		int[] seed = new int[num_attrs];
		for (int i=0; i<num_attrs; i++) {
			seed[i] = Integer.parseInt(args[cursor]);
			cursor ++;
		}
		
		double zipfAlpha = Double.parseDouble(args[cursor]); //only used for generateZipfQueryData()
		
		CorrelatedQueryStreamGenerator cqsGen = new CorrelatedQueryStreamGenerator(
				total_num_tuples, num_attrs, start_value, 
				num_distinct_values, num_tuples_betw_query, 
				data_file_name, seed);
		
		cqsGen.getQueryPatternsFromFile("resources\\PGroupby\\data\\qpattern-4.txt");
		//cqsGen.generateUniformQueryData();
		cqsGen.generateZipfQueryData(zipfAlpha);
	}

}
