/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on 2005-1-22
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import java.util.ArrayList;
import java.util.Vector;

/**
 * @author lisading
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class MultiQueuePunctGenOperatorImp extends
		XATMultiQueueStreamOperatorImp implements MultiQueuePunctGenOperator {
	
	int baseStreamPos;
	int genStreamPos;
	int basePunctAttr;
	int genPunctAttr;
	long window;
	int mesgType;
	
	XATQueue baseStream;
	XATQueue genStream;
	long earliestPunctGenTime = Long.MAX_VALUE;
	ArrayList genPuncts = new ArrayList();
	
	public MultiQueuePunctGenOperatorImp(XATNode init) {
		super(init);
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() throws Exception {
		if (!hasInit()) {
			init();
			if (!hasInit()) {
				return false;
			}
		}
		
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		int tupleCount = 0;
		
		ArrayList baseTuples = new ArrayList();
		ArrayList genTuples = new ArrayList();

		while (tupleCount < howMany) {
			try {
				XATTuple tuple = null;
				//pick the next tuple to process
				if (baseStream.getQueueSize(this) == 0) {
					if (genStream.getQueueSize(this) == 0) {
						break;
					} else {
						genTuples.add(genStream.dequeue(this));
						tupleCount ++;
					}
				} else if (genStream.getQueueSize(this) == 0) {
					baseTuples.add(baseStream.dequeue(this));
					tupleCount ++;
				} else {
					long baseTimeStamp = 0;
					long genTimeStamp = 0;
					
					baseTimeStamp =
						baseStream.peek(this).getMinTimeStamp();
					genTimeStamp = 
						genStream.peek(this).getMinTimeStamp();
						
					if(baseTimeStamp <= genTimeStamp) {
						baseTuples.add(baseStream.dequeue(this));
						tupleCount ++;
					} else {
						genTuples.add(genStream.dequeue(this));
						tupleCount ++;
					}					
				}
			} catch(XATQueueException e1) {
				return true;
			}
		}
		if (baseTuples.size()>0) {
			this.insertPunct((XATTuple[]) baseTuples.toArray(new XATTuple[] {}));
			baseTuples.clear();
		} 
		
		int numGenTuples = genTuples.size();
		for (int i=0; i<numGenTuples; i++) {
			XATTuple nextTuple = (XATTuple)genTuples.get(i);
			if (!nextTuple.isPunct()) {
				if (nextTuple.getMinTimeStamp() > earliestPunctGenTime) {
					this.getOutputQueue().enqueue((XATPunct)genPuncts.get(0), this);
					genPuncts.remove(0);
					if (genPuncts.size() != 0) {
						earliestPunctGenTime = ((XATPunct)genPuncts.get(0)).getMinTimeStamp();
					} else {
						earliestPunctGenTime = Long.MAX_VALUE;
					}					
				} 
				this.getOutputQueue().enqueue(generateResultTuple(nextTuple), this);
			} else {
				this.getOutputQueue().enqueue(generateResultPunct((XATPunct)nextTuple), this);
			}
		}
		genTuples.clear();
		return true;
		
	}
	
	private XATTuple generateResultTuple(XATTuple tuple) {
		XATDataValue[] tupleContent = tuple.getArrayData();
		long[] tupleTime = tuple.getTimeStamp();
		long tupleMaxTime = tuple.getMaxTimeStamp();
		long tupleMinTime = tuple.getMinTimeStamp();

		XATDataValue[] resultTupleContent = new XATDataValueImp[numOutputColumns];
		long[] resultTupleTime = new long[numOutputColumns];
		
		int i = 0, j = 0;
		for (i=0; i<numOutputColumns; i++) {
			int pos = outputPositions[0][i];
			resultTupleContent[i] = tupleContent[pos];
			resultTupleTime[i] = tupleTime[pos];
		}
		XATTuple resultTuple = new XATTupleImp(resultTupleContent);
		resultTuple.setTimeStamp(resultTupleTime);
		resultTuple.setMaxTimeStamp(tupleMaxTime);
		resultTuple.setMinTimeStamp(tupleMinTime);
		return resultTuple;
	}
	
	private XATPunct generateResultPunct(XATPunct punct) {
		XATDataValue[] punctContent = punct.getArrayData();
		int[] punctTypeArray = punct.getPunctTypeArray();
		long[] punctTime = punct.getTimeStamp();
		long maxTime = punct.getMaxTimeStamp();
		long minTime = punct.getMinTimeStamp();

		XATDataValue[] resultPunctContent = new XATDataValueImp[numOutputColumns];
		int[] resultPunctTypeArray = new int[numOutputColumns];
		long[] resultPunctTime = new long[numOutputColumns];
		
		int i = 0, j = 0;
		for (i=0; i<numOutputColumns; i++) {
			int pos = outputPositions[0][i];
			resultPunctContent[i] = punctContent[pos];
			resultPunctTypeArray[i] = punctTypeArray[pos];
			resultPunctTime[i] = punctTime[pos];
		}
		XATPunct resultPunct = new XATPunctImp(resultPunctContent);
		resultPunct.setPunctTypeArray(resultPunctTypeArray);
		resultPunct.setTimeStamp(resultPunctTime);
		resultPunct.setMaxTimeStamp(maxTime);
		resultPunct.setMinTimeStamp(minTime);
		resultPunct.setMesgType(punct.getMesgType());
		return resultPunct;
	}

	public void insertPunct(XATTuple[] baseTuples) {
		int numTuples = baseTuples.length;
		ArrayList tempPunctList = new ArrayList();
		for (int i=0; i<numTuples; i++) {
			XATTuple tuple = baseTuples[i];
			if (tuple.getObjTypeID()==1) {
				continue;
			}
			XATDataValue[] tContent = tuple.getArrayData();
			XATDataValue[] pContent = new XATDataValue[numOutputColumns];
			int[] punctTypes = new int[numOutputColumns];
			for (int j=0; j<numOutputColumns; j++) {
				if (outputPositions[0][j] == genPunctAttr) {
					punctTypes[j] = XATPunct.SINGLE_VALUE;
					pContent[j] = tContent[j];
				} else {
					punctTypes[j] = XATPunct.ALL_VALUE;
					pContent[j] = new XATDataValueImp("*");
				}
			}
			XATPunct punct = new XATPunctImp(pContent);
			punct.setPunctTypeArray(punctTypes);
			punct.setUniTimeStamp(tuple.getMinTimeStamp()+window);
			punct.setMesgType(this.mesgType);
			tempPunctList.add(punct);
		}
		if (genPuncts.size() == 0) {
			earliestPunctGenTime = ((XATPunct)tempPunctList.get(0)).getMinTimeStamp();
		}
		genPuncts.addAll(tempPunctList);
	}

	public int getBaseStreamPos() {
		return baseStreamPos;
	}
	public int getGenStreamPos() {
		return genStreamPos;
	}
	public int getBasePunctAttr() {
		return basePunctAttr;
	}
	public int getGenPunctAttr() {
		return genPunctAttr;
	}
	public long getWindow() {
		return window;
	}
	public int getMesgType() {
		return mesgType;
	}
	public void setBaseStreamPos(int bStreamPos) {
		baseStreamPos = bStreamPos;
	}
	public void setGenStreamPos(int gStreamPos) {
		genStreamPos = gStreamPos;
	}
	public void setBasePunctAttr(int bPunctAttr) {
		basePunctAttr = bPunctAttr;
	}
	public void setGenPunctAttr(int gPunctAttr) {
		genPunctAttr = gPunctAttr;
	}
	public void setWindow(long win) {
		window = win;
	}
	public void setMesgType(int mesgType) {
		this.mesgType = mesgType;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAccumulatedTuples(long)
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAllAccumulatedTuples()
	 */
	public boolean cleanAllAccumulatedTuples() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#matchStates(edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator)
	 */
	public boolean matchStates(XATStreamOperator op) {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#hasUnmatchedState()
	 */
	public boolean hasUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#recomputeUnmatchedState()
	 */
	public boolean recomputeUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#setStateID(int)
	 */
	public int setStateID(int startID) {
		// TODO Auto-generated method stub
		return 0;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator#finalizeOperation()
	 */
	public boolean finalizeOperation() {
		// TODO Auto-generated method stub
		return false;
	}

	public void init() {
		baseStream = this.getInputQueue(baseStreamPos);
		genStream = this.getInputQueue(genStreamPos);
		XATQueueSchema baseSchema = baseStream.getSchema(),
		               genSchema = genStream.getSchema();
		if (baseSchema == null || baseSchema.getLength() == 0) {
			return;
		}
		
		if (genSchema == null || genSchema.getLength()==0) {
			return;
		}
		
		XATQueueSchema outSchema = this.getOutputQueue().getSchema();
		if (outSchema == null || outSchema.getLength()==0) {
			this.getOutputQueue().setSchema(genSchema);
		}
		Vector outputPosVec = new Vector();
		Vector outColNames = this.getOutputQueue().getSchema().getColNames();
		int[] outputPos = new int[outColNames.size()];
		Vector inColNames = genStream.getSchema().getColNames(); 
		
		for (int i=0; i<(outColNames.size()); i++){
			Object outputCol = outColNames.get(i);
			outputPos[i] = inColNames.indexOf(outputCol);
		}
		this.setOutputPositions(0, outputPos);
		numOutputColumns = outputPositions[0].length;
		setHasInit(true);
	}
}
