/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on May 20, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamDistinctOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.PunctHashDistinctStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;

/**
 * @author lisading
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class PunctuationStreamDistinctOperatorImp
	extends StreamDistinctOperatorImp
	implements PunctuationStreamDistinctOperator {
		
	protected boolean propagateFlag;
	protected int memoryLimit;
	
	PunctHashDistinctStateImp state;
	int indexAttrPos;

	/*
	public PunctuationStreamDistinctOperatorImp(PunctuationStreamDistinctOperatorImp oldop, XATNode newnode) {
		super(oldop, newnode);
		this.indexAttrPos = oldop.getIndexAttrPos();
		state = (PunctHashDistinctStateImp)this.states[0];
	}*/

	public PunctuationStreamDistinctOperatorImp(XATNode init) {
		super(init);
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() throws Exception {
		if (!hasInit()) {
			init();
			if (!hasInit()) {
				return false;
			}
		}
		XATQueue inputQueue = this.getInputQueue();
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		int tupleCount = 0;
		
		while (true) {
			if (tupleCount == howMany) {
				return true;
			}
			try {
				XATTuple tuple = inputQueue.dequeue(this);
				
				if (tuple.isPunct()) {
					state.purge(tuple);
					if (propagateFlag) {
						this.getOutputQueue().enqueue(tuple, this);
					}
				} else {
					tupleCount ++;
					XATTuple[] result = state.probe(tuple);
					this.getOutputQueue().enqueue(result, this);
				}
			} catch(Exception e) {
				break;
			}
		}
		return true;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		return "Distinct";
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamDistinctOperator#getIndexAttrPos()
	 */
	public int getIndexAttrPos() {
		return indexAttrPos;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamDistinctOperator#setIndexAttrPos(int)
	 */
	public void setIndexAttrPos(int iAttrPos) {
		this.indexAttrPos = iAttrPos;
	}

	public void setState(OperatorState[] state) {
		this.states = new OperatorState[state.length];
		for(int i = 0; i < this.states.length;i++){
			this.states[i] = state[i];
		}
		this.state = (PunctHashDistinctStateImp)state[0];
	}

	public boolean hasUnmatchedState() {
		return false;
	}
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}
	public boolean cleanAllAccumulatedTuples() {
		return true;
	}
	public int setStateID(int startID) {
		return 0;
	}
	public boolean recomputeUnmatchedState() {
		return true;
	}
	public boolean matchStates(XATStreamOperator op) {
		return true;
	}
	public boolean finalizeOperation(){
		return true;
	}
	public boolean getPropagateFlag() {
		return propagateFlag;
	}

	public void setPropagateFlag(boolean pFlag) {
		propagateFlag = pFlag;
	}
	
	public int getMemoryLimit() {
		return memoryLimit;
	}
	public void setMemoryLimit(int memLimit) {
		memoryLimit = memLimit;
	}

}
