/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on 2005-3-27
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstreamoperator;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.Random;
import java.util.StringTokenizer;

/**
 * @author lisading
 *
 * generate stream of tuples with multiple attributes
 * according to independent uniform distribution
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class QueryStreamGenerator {
	String punctIndicator = "##";
	String wildcard = "*";
	String fieldDelim = ", ";
	String recordDelim = "\n";

	int totalNumTuples;
	int numAttrs;
	int[] startValue;
	int[] numDistinctValues;
	double numTuplesBetweenQuery; //0 means no query
	String dataFileName;

	int[][] candidateValues;

	Random[] dataAttrValIndexGen;
	Random[] queryAttrValIndexGen;
	
	int[][] queryPatterns;
	int numQueryPatterns;

	public QueryStreamGenerator() {
		totalNumTuples = 10000;
		numAttrs = 1;
		startValue = new int[] {0};
		numDistinctValues = new int[] {100};
		numTuplesBetweenQuery = 20;
		dataFileName = "resources\\PGroupby\\Data\\testdata.dat";
		dataAttrValIndexGen = new Random[numAttrs];
		queryAttrValIndexGen = new Random[numAttrs];
		
		candidateValues = new int[1][];
		for (int i=0; i<numDistinctValues[0]; i++) {
			candidateValues[0][i] = startValue[0] + i;
			dataAttrValIndexGen[i] = new Random(1);
			queryAttrValIndexGen[i] = new Random(1);
		}
	}
	
	public QueryStreamGenerator(int totalNumTuples, int numAttrs, 
			int[] startValue, int[] numDistinctValues, 
			double numTuplesBetweenQuery, String dataFileName, int[] seed) {
		this.totalNumTuples = totalNumTuples;
		this.numAttrs = numAttrs;
		this.startValue = startValue;
		this.numDistinctValues = numDistinctValues;
		this.numTuplesBetweenQuery = numTuplesBetweenQuery;
		this.candidateValues = new int[numAttrs][];
		this.dataFileName = dataFileName;
		dataAttrValIndexGen = new Random[numAttrs];
		queryAttrValIndexGen = new Random[numAttrs];
		
		for (int i=0; i<numAttrs; i++) {
			candidateValues[i] = new int[numDistinctValues[i]];
			for (int j=0; j<numDistinctValues[i]; j++) {
				candidateValues[i][j] = startValue[i]+j;
			}
			
			dataAttrValIndexGen[i] = new Random(seed[i]);
			queryAttrValIndexGen[i] = new Random(seed[i]);
		}
	}
	
	public void generateUniformQueryData() {
		Random qPatternRan = new Random(); //for choosing query patterns
		
		try {
			BufferedWriter writer = new BufferedWriter(new FileWriter(dataFileName));
			int numTuplesGenerated = 0;
			int numTuplesInCurrentBatch = -1, numQueriesInCurrentBatch = 1;
			if (numTuplesBetweenQuery >= 1) {
				numTuplesInCurrentBatch = (int)numTuplesBetweenQuery;
			} else if (numTuplesBetweenQuery == 0){
				numQueriesInCurrentBatch = 0;
			} else {
				numQueriesInCurrentBatch = (int)(1/numTuplesBetweenQuery);
			}

			boolean genQuery = false;
			while(numTuplesGenerated < totalNumTuples) {
				int dataValIndex;
				for (int i=0; i<numAttrs-1; i++) {
					dataValIndex = dataAttrValIndexGen[i].nextInt(numDistinctValues[i]);
					writer.write(candidateValues[i][dataValIndex] + fieldDelim);
				}
				dataValIndex = dataAttrValIndexGen[numAttrs-1]
								.nextInt(numDistinctValues[numAttrs-1]);
				writer.write(candidateValues[numAttrs-1][dataValIndex] + recordDelim);
				writer.flush();
				numTuplesGenerated ++;

				if (numTuplesInCurrentBatch>0) {
					numTuplesInCurrentBatch --;
					if (numTuplesInCurrentBatch == 0) {
						genQuery = true;
						numTuplesInCurrentBatch = (int)numTuplesBetweenQuery;
					} else {
						continue;
					}
				} else {
					genQuery = true;
				}

				if (!genQuery) {
					continue;
				}
				//generate queries
				for (int i=0; i<numQueriesInCurrentBatch; i++) {
					writer.write(punctIndicator);
					int qPatternIndex = qPatternRan.nextInt(numQueryPatterns);
					int[] qPattern = queryPatterns[qPatternIndex];
					int queryValIndex;
					for (int j=0; j<numAttrs-1; j++) {
						if (qPattern[j] == 0) {
							writer.write(wildcard + fieldDelim);
						} else {
							queryValIndex = dataAttrValIndexGen[j].nextInt(numDistinctValues[j]);
							writer.write(candidateValues[j][queryValIndex] + fieldDelim);
						}
					}
					if (qPattern[numAttrs-1] == 0) {
						writer.write(wildcard + recordDelim);
					} else {
						queryValIndex = dataAttrValIndexGen[numAttrs-1]
											.nextInt(numDistinctValues[numAttrs-1]);
						writer.write(candidateValues[numAttrs-1][queryValIndex] + recordDelim);
					}
					writer.flush();
				}
				genQuery = false;
			}
			writer.close();
			System.out.println("Done.");
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

	public void generateZipfQueryData(double alpha) {
		
/** CHUAN LEI - Comment out as ZipfDistribution is not found*/
//		ZipfDistribution qPatternRan = new ZipfDistribution(0, numQueryPatterns-1, alpha, 1); //for choosing query patterns
		
		try {
			BufferedWriter writer = new BufferedWriter(new FileWriter(dataFileName));
			int numTuplesGenerated = 0;
			int numTuplesInCurrentBatch = -1, numQueriesInCurrentBatch = 1;
			if (numTuplesBetweenQuery >= 1) {
				numTuplesInCurrentBatch = (int)numTuplesBetweenQuery;
			} else {
				numQueriesInCurrentBatch = (int)(1/numTuplesBetweenQuery);
			}

			boolean genQuery = false;
			while(numTuplesGenerated < totalNumTuples) {
				int dataValIndex;
				for (int i=0; i<numAttrs-1; i++) {
					dataValIndex = dataAttrValIndexGen[i].nextInt(numDistinctValues[i]);
					writer.write(candidateValues[i][dataValIndex] + fieldDelim);
				}
				dataValIndex = dataAttrValIndexGen[numAttrs-1]
								.nextInt(numDistinctValues[numAttrs-1]);
				writer.write(candidateValues[numAttrs-1][dataValIndex] + recordDelim);
				writer.flush();
				numTuplesGenerated ++;

				if (numTuplesInCurrentBatch>0) {
					numTuplesInCurrentBatch --;
					if (numTuplesInCurrentBatch == 0) {
						genQuery = true;
						numTuplesInCurrentBatch = (int)numTuplesBetweenQuery;
					} else {
						continue;
					}
				} else {
					genQuery = true;
				}

				if (!genQuery) {
					continue;
				}
				//generate queries
				for (int i=0; i<numQueriesInCurrentBatch; i++) {
					writer.write(punctIndicator);
					
/** CHUAN LEI - Comment out as ZipfDistribution is not found*/
//					int qPatternIndex = qPatternRan.random_variate(); //the only line that is different
//					int[] qPattern = queryPatterns[qPatternIndex];
//					int queryValIndex;
//					for (int j=0; j<numAttrs-1; j++) {
//						if (qPattern[j] == 0) {
//							writer.write(wildcard + fieldDelim);
//						} else {
//							queryValIndex = dataAttrValIndexGen[j].nextInt(numDistinctValues[j]);
//							writer.write(candidateValues[j][queryValIndex] + fieldDelim);
//						}
//					}
//					if (qPattern[numAttrs-1] == 0) {
//						writer.write(wildcard + recordDelim);
//					} else {
//						queryValIndex = dataAttrValIndexGen[numAttrs-1]
//											.nextInt(numDistinctValues[numAttrs-1]);
//						writer.write(candidateValues[numAttrs-1][queryValIndex] + recordDelim);
//					}
					writer.flush();
				}
				genQuery = false;
			}
			writer.close();
			System.out.println("Done.");
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

	public void generateQueryPatterns() {
		//initialize seed patterns
		int[][] patterns = new int[2][numAttrs];
		patterns[0][0] = 0;
		patterns[1][0] = 1;
		for (int i=1; i<numAttrs; i++) {
			patterns[0][i] = 0;
			patterns[1][i] = 0;
		}

		for (int i=1; i<numAttrs; i++) {
			int numOldPatterns = patterns.length;
			int numNewPatterns = numOldPatterns * 2;
			int[][] newPatterns = new int[numNewPatterns][numAttrs];
			for (int j=0; j<numOldPatterns; j++) {
				newPatterns[j] = patterns[j];
				newPatterns[j + numOldPatterns] = new int[numAttrs];
				for (int k=0; k<numAttrs; k++) {
					newPatterns[j + numOldPatterns][k] = patterns[j][k];
				}
				newPatterns[j + numOldPatterns][i] = 1;
			}
			patterns = newPatterns;
		}
		queryPatterns = patterns;
		numQueryPatterns = queryPatterns.length;
	}
	
	public void getQueryPatternsFromFile(String inputFile) {
		try {
			BufferedReader reader = new BufferedReader(new FileReader(inputFile));
			String new_line = reader.readLine();
			StringTokenizer st = new StringTokenizer(new_line, ",");
			numQueryPatterns = Integer.parseInt(st.nextToken().trim());
			numAttrs = Integer.parseInt(st.nextToken().trim());
			queryPatterns = new int[numQueryPatterns][numAttrs];
			
			int cursor = 0;
			while((new_line = reader.readLine())!=null) {
				new_line = new_line.trim();
				st = new StringTokenizer(new_line, ",");
				int index = 0;
				while(st.hasMoreTokens()) {
					queryPatterns[cursor][index] = Integer.parseInt(st.nextToken().trim());
					index ++;
				}
				cursor ++;
			}
			reader.close();
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	public static void main(String[] args) {
		int total_num_tuples = Integer.parseInt(args[0]);
		int num_attrs = Integer.parseInt(args[1]);
		int cursor = 2;
		int[] start_value = new int[num_attrs];
		for (int i=0; i<num_attrs; i++) {
			start_value[i] = Integer.parseInt(args[cursor]);
			cursor ++;
		}
		int[] num_distinct_values = new int[num_attrs];
		for (int i=0; i<num_attrs; i++) {
			num_distinct_values[i] = Integer.parseInt(args[cursor]);
			cursor ++;
		}
		double num_tuples_betw_query = Double.parseDouble(args[cursor]);
		cursor ++;

		String data_file_name = args[cursor];
		cursor ++;
		
		int[] seed = new int[num_attrs];
		for (int i=0; i<num_attrs; i++) {
			seed[i] = Integer.parseInt(args[cursor]);
			cursor ++;
		}
		
		double zipfAlpha = Double.parseDouble(args[cursor]); //only used for generateZipfQueryData()
		
		QueryStreamGenerator qsGen = new QueryStreamGenerator(
				total_num_tuples, num_attrs, start_value, 
				num_distinct_values, num_tuples_betw_query, 
				data_file_name, seed);
		
		qsGen.getQueryPatternsFromFile("resources\\PGroupby\\data\\qpattern-4.txt");
		qsGen.generateUniformQueryData();
		//qsGen.generateZipfQueryData(zipfAlpha);
	}
	
}
