/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on 2005-1-27
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.PunctHashSingleAttributeGroupbyStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.GroupbyHashEntry;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.WStreamGroupbyOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.ArrayList;

/**
 * @author lisading
 *
 * The beginning of the first window is the timestamp of the first tuple received.
 */
public class WPunctuationStreamGroupbyOperatorImp 
	extends	WStreamGroupbyOperatorImp 
	implements PunctuationStreamOperator {

	protected boolean propagateFlag;
	protected int memoryLimit;

	protected PunctHashSingleAttributeGroupbyStateImp state;
	
	long currentWindowEnd;
	long nextWindowStart;
	Hashtable resultBuffer = new Hashtable();

	public WPunctuationStreamGroupbyOperatorImp(XATNode init) {
		super(init);
	}

	public boolean run() throws Exception {
		if (!hasInit()) {
			init();
			if (!hasInit()) {
				return false;
			}
			currentWindowEnd = this.windowSize;
			nextWindowStart = this.increment;
		}
		
		XATQueue inputQueue = this.getInputQueue();
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		//int howMany = 600;
		int tupleCount = 0;
		
		
		while (true) {
			if (tupleCount == howMany) {
				return true;
			}
			try {
				XATTuple tuple = inputQueue.dequeue(this);
								
				if (tuple.isPunct()) { //punctuation
					XATPunct punct = (XATPunct)tuple;
					long pStartTime = punct.getMinTimeStamp();
					long pEndTime = punct.getInvalidTime();
					Object punctValue = punct.getCell(groupAttrPosArray[0]).getValue();
					if (pStartTime<nextWindowStart && pEndTime>=currentWindowEnd) {
						GroupbyHashEntry hashentry = state.getSingleGroupEntry(punctValue);
						if (hashentry != null) {
							XATTuple[] result = (XATTuple[])hashentry.getTupleList().toArray(new XATTuple[]{});
							XATTuple resultTuple = generateResultForSingleGroup(result);
							this.getOutputQueue().enqueue(resultTuple, this);
							state.purge(punctValue);
						}
					} else if (pStartTime>nextWindowStart && pEndTime>=currentWindowEnd) {
						GroupbyHashEntry hashentry = state.getSingleGroupEntry(punctValue);
						if (hashentry != null) {
							XATTuple[] result = (XATTuple[])hashentry.getTupleList().toArray(new XATTuple[]{});
							XATTuple resultTuple = generateResultForSingleGroup(result);
							this.getOutputQueue().enqueue(resultTuple, this);
							
							//compute results for the future windows
							long nextStart = this.nextWindowStart;
							long nextEnd = nextStart+windowSize;
							int startIndex = 0;
							int endIndex = result.length-1;
							while (nextStart < pStartTime && nextEnd < pEndTime) {
								while(result[startIndex].getNthTimeStamp(groupAttrPosArray[0])<=nextStart
										&& startIndex <= endIndex) {
									startIndex ++;
								}
								if (startIndex > endIndex) {
									break;
								}
								XATTuple newResultTuple = this.generateResultForSingleGroup(result, startIndex, endIndex);
								Long key = new Long(nextStart);
								Object entry = resultBuffer.get(key);
								if (entry == null) {
									ArrayList tupleList = new ArrayList();
									tupleList.add(newResultTuple);
									resultBuffer.put(key, tupleList);
								} else {
									((ArrayList)entry).add(newResultTuple);
								}
								nextStart += increment;
							}
							state.invalidateSingleGroup(punctValue, nextStart);
						}
					}
				} else { //regular tuple
					tupleCount ++;
					if (tuple.getNthTimeStamp(groupAttrPosArray[0])>currentWindowEnd) {
						//output pre-compute result, if any
						Long key = new Long(currentWindowEnd-windowSize);
						Object preComputeResult = resultBuffer.get(key);
						if (preComputeResult!=null) {
							this.getOutputQueue().enqueue((ArrayList)preComputeResult, this);
						}
						//output result from current state
						generateResultForCurrentWindow();
						state.invalidate(currentWindowEnd-windowSize);
						currentWindowEnd += increment;
						nextWindowStart += increment;
					}
					state.probe(tuple);			
				}
			} catch(Exception e) {
				break;
			}
		}
		return true;
	}

	
	private void generateResultForCurrentWindow() {
		int numFunctions = aggFunctions.length;
		
		Hashtable hashtable = (Hashtable)state.getAllStateTuples();
		Enumeration en = hashtable.keys();
		while(en.hasMoreElements()) {
			ArrayList tupleList = ((GroupbyHashEntry) hashtable.get(en.nextElement())).getTupleList();
			
			XATDataValue[] resultValue = new XATDataValue[numOutputColumns];
			long[] resultTime = new long[numOutputColumns];
			XATTuple firstTupleInGroup = (XATTuple)tupleList.get(0);
			XATDataValue[] firstTupleContent = firstTupleInGroup.getArrayData();
			long[] firstTupleTime = firstTupleInGroup.getTimeStamp();
			long minTime = firstTupleInGroup.getMinTimeStamp();
			long maxTime = firstTupleInGroup.getMaxTimeStamp();
			
			int i=0;
			for (; i<numOutputColumns-numFunctions; i++) {
				int index = outputPositions[0][i];
				resultValue[i] = firstTupleContent[index];
				resultTime[i] = firstTupleTime[index];
			}
			for (int j=0; j<numFunctions; j++) {
				resultValue[j+i] = aggFunctions[j].evaluate((XATTuple[])tupleList.toArray(new XATTuple[]{}));
				resultTime[j+i] = minTime;
			}
			XATTuple resultTuple = new XATTupleImp(resultValue);
			resultTuple.setTimeStamp(resultTime);
			resultTuple.setMinTimeStamp(minTime);
			resultTuple.setMaxTimeStamp(maxTime);
			this.getOutputQueue().enqueue(resultTuple, this);
		}
	}
	
	public boolean getPropagateFlag() {
		return propagateFlag;
	}

	public void setPropagateFlag(boolean pFlag) {
		propagateFlag = pFlag;
	}
	
	public int getMemoryLimit() {
		return memoryLimit;
	}
	public void setMemoryLimit(int memLimit) {
		memoryLimit = memLimit;
	}
	public void setState(OperatorState[] state) {
		this.states = new OperatorState[state.length];
		for(int i = 0; i < this.states.length; i++){
			this.states[i] = state[i];
		}
		this.state = (PunctHashSingleAttributeGroupbyStateImp)state[0];
	}
}
