/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on April 6, 2005
 * based on edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.StreamSplitOperator;
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.spreadfire;
import java.lang.reflect.Constructor;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Date;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATMemoryQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATCommandMsg;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATCommandMsgImp;

/**
 * @author Bradley Momberger
 * @author bmombe1
 * 
 * based on the work of Bin Liu
 * A replication stream split operator definition. Assume one input queue, multiple output queues. 
 * In general, one output queue corresponds to one remote machine. 
 */
public class ReplicatedStreamSplitOperatorImp extends XATMultiQueueStreamOperatorImp
		implements ReplicatedStreamSplitOperator {

	Hashtable mappingTable; // the mapping between the partition ids and the machines
	boolean initialized = false;	//indicates whether the initialized has been done?
	//PartitionFunction partitionFunction;
	
	boolean partitioned_split = false; // to indicate whether the split operator is activated with join join operators
	
	//XATMemoryQueueImp tmpqueue = null; // to hold the partition inputs having IDs that will be adapted
	boolean deactivated = false; // those two are used to indicate whether we need to put inputs to tmpqueue
	boolean reactivated = false; // or read inputs from the tmpqueue
	
	public boolean getPartitionSplitFlag(){
		return partitioned_split;
	}
	
	public void reactivatePartitions(ArrayList toAct, String ipAdd){
		for (int i=0; i<toAct.size(); i++){
			Integer idx = (Integer)toAct.get(i);
			mappingTable.put(idx,ipAdd);
		}
		reactivated = true;
	}
	
	public ReplicatedStreamSplitOperatorImp(XATNode node){
		super(node);
		mappingTable = new Hashtable();
	}
	
	public void setInitialProperties(){

	}
	/* 
	 * Major running functions of the operator, takes one tuple at a time and send them to the
	 * corresponding queues
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public synchronized boolean run() throws Exception {

		XATQueueSchema inSchema;
		int inLength;
		if ((inSchema = this.getInputQueue().getSchema()) == null){
				return false;
		}
		inLength = inSchema.getLength();
		
		if(this.getOutputQueue().getSchema() == null){
			init();
		}
		
		if (!initialized){
			
			initialized = true;
		}
		
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		//System.out.println("NUMBER_OF_TUPLES_TO_DEQUEUE:" + howMany);
		int tupleCount = 0;
		
		//Runtime.getRuntime().gc();
		
		/** get the tuples **/
		while (true){
            Iterator outputs = this.getOutputQueues();
            boolean oneemptyqueue = false;
            boolean toofullqueue = false;
            while(outputs.hasNext())
            {
                //we'll try to keep the queue size low, so replicated operators
                //  will parse messages coming from here.
                XATQueue next = (XATQueue)outputs.next();
                if(next.getQueueSize() > 1000)
                    toofullqueue = true;
                if(next.getQueueSize() < 1)
                    oneemptyqueue = true;
            }
            if((!oneemptyqueue) && toofullqueue)
                return true;
            
            XATTuple[] tuples = this.getInputQueue().dequeue(howMany, this);
			ArrayList tlist = new ArrayList();
            //long starttime = System.currentTimeMillis();
			for (int i=0; i<tuples.length; i++){
				tlist.add(tuples[i]);
				
			}	
            //System.out.println("Split op took " + (System.currentTimeMillis() - starttime) + "ms to enqueue " + tlist.size() + " tuples");

            outputs = this.getOutputQueues();
			while (outputs.hasNext()) {
				XATMemoryQueueImp q = (XATMemoryQueueImp)outputs.next();
				// find the outputqueue and enqueue the tuple
				q.enqueue(tlist,this);
			}
			
			
			tuples = null;	
			return true;			
		} 
	} //end of run()

	/*
	 * Several basic access functions 
	 */
//	public void setNumberOfOutputs(int np){
//		this.getOutputQueueCount();
//	}
	public int getNumberOfOutputs(){
		return this.getOutputQueueCount();
	}
	public void setOutputTable(Hashtable mt){
		mappingTable = mt;
	}
//	public void setPartitionColumnIndex(int pci){
//		partitionColumnIndex = pci;
//	}

    public boolean hasTokenForState(int state)
    { return false; }
    
	public void generateTokenPassingMessage(String machine) {

		XATCommandMsg message = new XATCommandMsgImp();

		message.setUniTimeStamp((new Date()).getTime());
		message.setCommandTypeArray(new int[1]);
		message.setCommandArgsArray(new XATDataValue[1]);
		message.setCommand(0, XATCommandMsg.PASS_TOKEN, new XATDataValueImp (machine));
		
		Iterator it = this.getOutputQueues();
		while (it.hasNext())
		((XATMemoryQueueImp)it.next()).addTuple(message);
        System.out.println("Split operator " + this.getOperatorID() + " successfully sent token message.");
		
	}
	
	public void generateStreamTerminationMessage(String machine)
	{
		XATCommandMsg message = new XATCommandMsgImp();

		message.setUniTimeStamp((new Date()).getTime());
		message.setCommandTypeArray(new int[1]);
		message.setCommandArgsArray(new XATDataValue[1]);
		message.setCommand(0, XATCommandMsg.TERMINATE_STREAM, new XATDataValueImp (machine));

		Iterator it = this.getOutputQueues();
		while (it.hasNext())
		((XATMemoryQueueImp)it.next()).addTuple(message);
	}
	
	public synchronized void terminateStream(String machine, Hashtable connections)
	{
			generateStreamTerminationMessage(machine);
			connections.remove(machine);
	}


	/* 
	 * The outputQueue has more than one queues, return the one that has the same name
	 */
	public edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue getOutputQueueByName(String name){
		for (int i=0; i<outputQueue.size(); i++){
			if ( ((XATQueue)outputQueue.get(i)).getQueueName().equals(name) ){
				return (XATQueue)outputQueue.get(i);
			}
		}
		// if it is not found, just return null
		return null;
	}

    public boolean hasState()
    { return false; }
    
	/*
	 * Remove all existing output queues
	 */
	public void removeQutputQueues(){
		outputQueue.clear();
	}
	/*
	 * Need own version of init() to set all output queues (include dynamic generated one) schema
	 */
	public void init() {
		Iterator i = this.getInputQueues();
		String schemaName = new String();
		XATQueueSchema s;
		XATQueueSchema outSchema = new XATQueueSchema();

		while (i.hasNext()) {

			s = ((XATQueue) i.next()).getSchema();
			schemaName += "#" + s.getName();
			
			Iterator j = s.getColNames().iterator();
			while(j.hasNext()){
				outSchema.addColumn((String) j.next());
			}
		}

		outSchema.setName(schemaName);
		
		Iterator it = this.getOutputQueues();
		
		while (it.hasNext()) {
			XATQueue current = (XATQueue)it.next();
		    current.setSchema(outSchema);

		    XATCommandMsg schemamsg = new XATCommandMsgImp();
		    schemamsg.setCommandTypeArray(new int[]{(XATCommandMsg.SET_INPUT_SCHEMA)});
            java.util.ArrayList schemainfo = new java.util.ArrayList();
            schemainfo.add((Object)outSchema.getColNames());
            schemainfo.add((Object)outSchema.getName());
		    schemamsg.setCommandArgsArray(new XATDataValueImp[]{new XATDataValueImp(schemainfo)});
			current.enqueue(schemamsg, this);
            System.out.println("Split operator " + this.getOperatorID() + " successfully sent schema message.");
		}
		//this.getOutputQueue().setSchema(outSchema);
		
		computeOutputPositions();
		
		//Get the parent operator initialized to take on the storage token right now.
		//The "YOU" string is a special case for when we only have one machine active, because it's hard to know
		//  at init time the name of the machine that the packet is going to.
		//Usually the DM takes care of knowing which machine to send to, but right now it's just startup time.
		generateTokenPassingMessage("YOU");
		
		
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAccumulatedTuples(long)
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		// TODO Auto-generated method stub
		return false;
	}

	public boolean finalizeOperation(){
		return true;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAllAccumulatedTuples()
	 */
	public boolean cleanAllAccumulatedTuples() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#matchStates(edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator)
	 */
	public boolean matchStates(XATStreamOperator op) {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#hasUnmatchedState()
	 */
	public boolean hasUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#recomputeUnmatchedState()
	 */
	public boolean recomputeUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#setStateID(int)
	 */
	public int setStateID(int startID) {
		// TODO Auto-generated method stub
		return 0;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		// TODO Auto-generated method stub
		return null;
	}

	
	
}
