/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on 2005-4-24
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.spreadfire;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import java.util.Vector;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamSelectOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATCommandMsg;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.*;
import java.util.Calendar;
import java.util.GregorianCalendar;
/**
 * @author Mingzhu Wei
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class ReplicationAwareStreamSelectOperatorImp extends
		StreamSelectOperatorImp implements ReplicationAwareStreamSelectOperator {
	
	private ExpressionStr expression;
	boolean hastoken = false;
	boolean firstrun = true;
	
	String MachineName;
	
	
	public ReplicationAwareStreamSelectOperatorImp(XATNode init) {
		super(init);

		}
 
	
	public boolean run() throws Exception {

		if(firstrun)
		{
			MachineName = System.getProperty("NAME");
			if (MachineName.compareTo("Machine 1")  == 0)
				hastoken = true;
			else
				hastoken = false;
			firstrun = false;
		}
		
		/**
		 * first figure out how many tuples to dequeue, the selectivity, and 
		 * result delay  
		 */
		XATQueueSchema inSchema;
		int inLength;
		if ((inSchema = this.getInputQueue().getSchema()) == null){
				return false;
		}
		inLength = inSchema.getLength();
		
		if(this.getOutputQueue().getSchema() == null){
			init();
		}
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		/** get the tuples */

		XATTuple[] tuples = this.getInputQueue().dequeue(howMany, this);
		/** figure out how many tuples to keep*/
		Vector tupleList = new Vector();
		
		for (int i=0; i<tuples.length; i++) {
			//Judge whether the current tuple is a Commandtype:
			if (tuples[i].getObjTypeID()==208)
			{
				if ( ((XATCommandMsg)tuples[i]).getCommandType(0)==XATCommandMsg.PASS_TOKEN)
				{
					XATDataValue argsValue = ((XATCommandMsg)tuples[i]).getCommandArgs(0);
					String  tokenMachineName = argsValue.convertToString(); 
					if (this.MachineName.equals(tokenMachineName))
					{
						this.hastoken = true;
						GregorianCalendar Cal = new GregorianCalendar();
						System.out.println("I got the token at time " + Cal.get(Calendar.HOUR_OF_DAY) + ":" + Cal.get(Calendar.MINUTE) + ":" + Cal.get(Calendar.SECOND));
					}
					else 
						this.hastoken = false;
				}
			}
			else if (this.hastoken)
			{
			
				//if the operator has no token, then skip the 
				//tupleList.clear();
				tupleList.add(0,tuples[i]);
				Object out = expression.eval(tupleList, this);
				if(out instanceof Boolean) {
					if(((Boolean)out).booleanValue()) {
						//XATTuple newtuple = new XATTupleImp()
						this.getOutputQueue().enqueue(tuples[i], this);
					}
				}
			}
		}
		
			


		return true;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		// TODO Auto-generated method stub
		return "Select";
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamSelectOperator#getExpression()
	 */
	public ExpressionStr getExpression() {
		return expression;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamSelectOperator#setExpression(edu.wpi.cs.dsrg.xmldb.xat.common.streamexpression.ExpressionStr)
	 */
	public void setExpression(ExpressionStr expression) {
		this.expression = expression;
	}
	
	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input queues.
	 * Inside method run() the numbers are computed based on the property
	 * "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler,
	 * while in cleanAccumulatedTuples, all tuples in input queues that have their max timesamp
	 * smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() {
		return true;
	}
	
	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() {
		return false;
	}

	public boolean matchStates(XATStreamOperator op) {
		return false;
	}
	
	public boolean recomputeUnmatchedState() {
		return true;
	}
	
	public int setStateID(int nextid) {
		return nextid;
	}
	public boolean finalizeOperation(){
		return true;
	}

}
