/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Sep 14, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.spreadfire;

import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATCommandMsg;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * @author bmombe1
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class SpreadfireMergeOperator extends XATStreamOperatorImp implements SpreadfireSupportOperator {

	public class SpreadfireMergeTypeNotImplementedException extends Exception
	{}
	
	public class SpreadfireMergeTypeNotSupportedException extends Exception
	{}
	
	public static final int DISCARD_ALL_INPUT = -1;
	public static final int STORE_AND_WAIT = -2;
	public static final int UNION_TYPE = 0;
	public static final int SUM_TYPE = 1;
	public static final int INTERSECT_TYPE = 2;
	public static final int UNIQUE_TYPE = 3;
	
	protected int mergetype = 0;
	protected XATDataValueImp[] mergeoptions = null;
	/**
	 * @param node
	 */
	public SpreadfireMergeOperator(XATNode node) {
		super(node);
	}

	/**
	 * @param oldop
	 * @param newnode
	 */
	public SpreadfireMergeOperator(XATStreamOperatorImp oldop, XATNode newnode) {
		super(oldop, newnode);
	}

	/**
	 * @param oldop
	 * @param newnode
	 */
	public SpreadfireMergeOperator(XATStreamOperator oldop, XATNode newnode) {
		super(oldop, newnode);
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#init()
	 */
	public void init() {
		// TODO Auto-generated method stub
 		Iterator i = this.getInputQueues();
 		String schemaName = new String();
 		XATQueueSchema s;
 		XATQueueSchema outSchema = new XATQueueSchema();

 		if (i.hasNext()) {

 			s = ((XATQueue) i.next()).getSchema();
 			schemaName += "#" + s.getName();

 			Iterator j = s.getColNames().iterator();
 			while(j.hasNext()){
 				outSchema.addColumn((String) j.next());
 			}
 		}

 		outSchema.setName(schemaName);
 		this.getOutputQueue().setSchema(outSchema);

 		computeOutputPositions();
 		this.hasInit = true;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() throws Exception {
		XATQueueSchema inSchema;
		int inLength;
		
		if ((inSchema = this.getInputQueue().getSchema()) == null){
			return false;
		}
		inLength = inSchema.getLength();
		
		if(this.getOutputQueue().getSchema() == null){
			init();
		}
		
		Iterator it = this.getInputQueues();
		while(it.hasNext())
		{
			XATQueue next = (XATQueue)it.next();
			XATTuple[] tuples = next.peek(Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE")), this);
			int lasti = 0;
			for(int i = 0; i < tuples.length; i++)
			{
				if (tuples[i].getObjTypeID() == 208)
				{
					//If we find an XATCommandMsg, we want to process all the tuples
					// before it first, then the message, then the tuples after it.
					//This code should be backportable to the WJoin operator.
					processTuples((XATTuple[])next.dequeue(i - lasti, this));
					processXATCommandMsg((XATCommandMsg)next.dequeue(this));
					// if another message happens, we want to start dequeuing from the next tuple
					lasti = i + 1;
				}
			}
			
			processTuples((XATTuple[])next.dequeue(Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE")) - lasti, this));
		}
		return true;
	}

	protected void processTuples(XATTuple[] tuples)
	{
		try
		{
			switch(this.mergetype)
			{
			case UNION_TYPE:
				unionAllTuples(tuples);
				break;
				
			case SUM_TYPE:
				sumNumericTuples(tuples);
				break;
					
			case INTERSECT_TYPE:
				windowedIntersection(tuples);
				break;
						
			case UNIQUE_TYPE:
				findUniqueWindowed(tuples);
				break;
				
		    default:
		    	throw new SpreadfireMergeTypeNotSupportedException();
			}

		}
		catch(SpreadfireMergeTypeNotImplementedException e)
		{ e.printStackTrace(); }
		catch(SpreadfireMergeTypeNotSupportedException e)
		{ e.printStackTrace(); }
	}
	
	protected void processXATCommandMsg(XATCommandMsg msg)
	{
		//This function is a stub because we're currently only working with union.
		//When other functionality is introduced we will need to be able to switch 
		//  modes with this function.
	    
	    int[] commands = msg.getCommandTypeArray();
	    for(int i = 0; i < commands.length; i++)
	    {
	     if (commands[i] == XATCommandMsg.SET_MERGE_TYPE)   
	        this.mergetype = msg.getCommandArgs(0).convertToDouble().intValue();
	    }
	}
	
	protected void unionAllTuples(XATTuple[] tuples)
	{
				this.getOutputQueue().enqueue(tuples,this);	
	}
	
	protected void windowedIntersection(XATTuple[] tuples) throws SpreadfireMergeTypeNotImplementedException
	{
		throw new SpreadfireMergeTypeNotImplementedException();
	}
	
	protected void findUniqueWindowed(XATTuple[] tuples) throws SpreadfireMergeTypeNotImplementedException
	{
		throw new SpreadfireMergeTypeNotImplementedException();	
	}
	
	protected void sumNumericTuples(XATTuple[] tuples) throws SpreadfireMergeTypeNotImplementedException
	{
		throw new SpreadfireMergeTypeNotImplementedException();
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAccumulatedTuples(long)
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAllAccumulatedTuples()
	 */
	public boolean cleanAllAccumulatedTuples() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#matchStates(edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator)
	 */
	public boolean matchStates(XATStreamOperator op) {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#hasUnmatchedState()
	 */
	public boolean hasUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#recomputeUnmatchedState()
	 */
	public boolean recomputeUnmatchedState() {
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#setStateID(int)
	 */
	public int setStateID(int startID) {
		// TODO Auto-generated method stub
		return 0;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator#finalizeOperation()
	 */
	public boolean finalizeOperation() {
		// TODO Auto-generated method stub
		try { run(); } catch(Exception e) { e.printStackTrace();}
        return true;
	}

}
