/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.state;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import java.io.Serializable;
import java.util.ArrayList;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
//import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Feb 23, 2004
 * 
 * The operator state class represents internal state for StreamOperators
 * during execution. This state holds XATTuples for query processing. This state
 * must be serializable so if the operator is migrated, it's entire state can be sent.
 * 
 */
public abstract class OperatorStateImp implements OperatorState, Serializable {
	protected boolean debugFlag = false;
	//Used by moving-state-migration strategy
	protected boolean isMatched = false; 
	//If this state is not matched during moving state migration
	//it can be recomputed by joining the states in joiningsubstates.
	protected ArrayList joiningsubstates; 
	
	protected XATStreamOperator operator;
	protected int stateID; // used by moving state migration strategy to match states
	protected String stateSchemaName; //also used for across-machine partition-based load rebalancing using moving state.
	protected int totalNumTuples;

	public OperatorStateImp() {
		totalNumTuples = 0;
	}
	
	public OperatorStateImp(OperatorState oldstate, XATStreamOperator newop) {
		this((OperatorStateImp)oldstate, newop);
	}
	
	public OperatorStateImp(OperatorStateImp oldstate, XATStreamOperator newop) {
		//OperatorStateImp newstate = new OperatorStateImp();
		this.stateID = oldstate.stateID;
		this.totalNumTuples = 0;
		this.operator = newop;
	}
	
	public OperatorStateImp(XATStreamOperator op) {
		operator = op;
		totalNumTuples = 0;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState#decreStateSize()
	 */
	public void decreStateSize() {
		totalNumTuples --;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState#decreStateSize(int)
	 */
	public void decreStateSize(int decrement) {
		totalNumTuples -= decrement;
	}
	public boolean getDebugFlag() {
		return this.debugFlag;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState#getOperator()
	 */
	public XATStreamOperator getOperator() {
		return operator;
	}
	public int getStateID () {
		return stateID;
	}
	
	public String getStateSchemaName() {
	    return this.stateSchemaName;
	}
	
	public void setStateSchemaName(String name) {
	    this.stateSchemaName = name;
	}
	
	public void setStateSize(int size) {
	    this.totalNumTuples = size;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState#getStateSize()
	 */
	public int getStateSize() {
		return totalNumTuples;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState#increStateSize()
	 */
	public void increStateSize() {
		totalNumTuples ++;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState#increStateSize(int)
	 */
	public void increStateSize(int increment) {
		totalNumTuples += increment;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (5/25/2003 2:47:32 AM)
	 * @return boolean
	 */
	public boolean isMatched() {
		return isMatched;
	}
	
	public void setDebugFlag (boolean flag) {
		this.debugFlag = flag;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (2/25/2003 11:27:47 AM)
	 * @param newTuple streamgenerator.XATStreamTuple
	 */
	//public abstract int purgeByTuple(XATTuple newTuple, Object customer);
	/**
	 * Insert the method's description here.
	 * Creation date: (4/4/2003 3:09:55 PM)
	 * @return boolean
	 * @param debugFlag boolean
	 */
	//public boolean setDebug(boolean debugFlag) {
	//	this.debugFlag = debugFlag;
	//	return debugFlag;
	//}
	/**
	 * Insert the method's description here.
	 * Creation date: (5/25/2003 4:22:20 AM)
	 * @param matched boolean
	 */
	public void setMatched(boolean matched) {
	    isMatched = matched;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState#setOperator(edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator)
	 */
	public void setOperator(XATStreamOperator op) {
		operator = op;
	}
	
	public void setStateID (int id) {
		stateID = id;
	}
	
	public void addJoiningSubStates(int substateID) {
		if(this.joiningsubstates == null)
			this.joiningsubstates = new ArrayList();
		
		if(!this.joiningsubstates.contains(Integer.toString(substateID)))
			this.joiningsubstates.add(Integer.toString(substateID));
	}
	
	public ArrayList getJoiningSubStates() {
		return this.joiningsubstates;
	}
	
//	public void setStateContent(OperatorState os) {
//		
//	}
	
//	public void probe(XATTuple tuple) {
//		// sub-classes should overwrite this method.
//	}
//	public void probe(XATTuple tuple, ExpressionStr expression) {
//		//	sub-classes should overwrite this method.
//	}
//	public void probe(XATTuple[] tuples) {
//		//	sub-classes should overwrite this method.	
//	}
//	public void probe(XATTuple[] tuples, ExpressionStr expression) {
//		//	sub-classes should overwrite this method.
//	}
}
