/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Aug 24, 2004
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.state;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;

/**
 * @author lisading
 *
 */
public class WEquiJoinStateImp extends OperatorStateImp implements
		WEquiJoinState 
{
    
	Hashtable hashtable;
	int joinAttrPos;
	long windowSize;
	
	// Apr-2005, added by Bin, for handling purge operator states based on certain time interval
	// This is useful when certain states have been pushed into the disk, while the corresponding
	// memory resident states are invalidated, we cannot simply discard these states, instead, we put
	// them into the memory temporarily, and also push them into disk and merge them with the disk-resident
	// part at the clean up stage
	Hashtable invalidated_states;
	
	public WEquiJoinStateImp() 
	{
		super();
		hashtable = new Hashtable();
		invalidated_states = new Hashtable();
	}

	// this constructor is used to build a WEquiJoinState based on the hash table
	// it is used to wrap invalidated states
	public WEquiJoinStateImp(Hashtable states) 
	{
		super();
		hashtable = states;
	}
	
	/*
	 * Added by Bin @ Apr 09, 2005, simply delete the invalidated states
	 */
	public int purgeInvalidatedStates()
	{
		int purged_size = invalidated_states.size();
		invalidated_states.clear();
		return purged_size;
	}
	
	public WEquiJoinStateImp getInvalidatedStates()
	{
		WEquiJoinStateImp ret = new WEquiJoinStateImp(invalidated_states);
		return ret;
	}
	
	/*
	 * Added by Bin @ Apr 08, 2005. To invalidate the out-of-date states, and put them into the invalidated_states
	 * It is close to the purgeByWindow function below except recording the invalidated tuple for lazy purge 
	 */
	public int purgeByWindowSize(int joinPos, long wSize, long latestProcessTupleTime)
	{
		Enumeration en = hashtable.keys();
		int numTuplesPurged = 0;
		while(en.hasMoreElements()) 
		{
			Object key = en.nextElement();
			Object bucket = hashtable.get(key);
			if (bucket == null) 
			{
				continue;
			}
			ArrayList bucketList = (ArrayList)bucket;
			int cursor = 0;
			while (cursor<bucketList.size()) 
			{
				XATTuple tuple = (XATTuple)bucketList.get(cursor);
				if (tuple.getNthTimeStamp(joinPos)<latestProcessTupleTime - wSize) 
				{
					bucketList.remove(0);
					numTuplesPurged ++;
					// put the purged tuple into invdlidated_states;
					Object invalidated_bucket = invalidated_states.get(key);
					if (invalidated_bucket == null)
					{
						ArrayList invalidated_bucketList = new ArrayList();
						invalidated_bucketList.add(tuple);
						invalidated_states.put(key, invalidated_bucketList);						
					} 
					else 
					{
						((ArrayList)invalidated_bucket).add(tuple);
					}
				} 
				else 
				{
					cursor ++;
				}
			}
		}
		// this.decreStateSize(numTuplesPurged);	
		return numTuplesPurged;
	}
	
	/*
	 * Added by Bin @ Nov.21, 2004. To get all tuples stored in the hashtable. It is used when merging disk residented partitions
	 */	
	public WEquiJoinStateImp(WEquiJoinStateImp oldstate, XATStreamOperator newop) 
	{
		super(oldstate, newop);
		this.debugFlag = oldstate.debugFlag;
		this.isMatched = false;
		hashtable = new Hashtable();
		hashtable = oldstate.getAllStateTuples();
		joinAttrPos = oldstate.getJoinAttrPos();
		windowSize = oldstate.getWindowSize();
	}
	
	public int purgeByWindow(long wSize, long latestProcessTupleTime) 
	{
		Enumeration en = hashtable.keys();
		int numTuplesPurged = 0;
		while(en.hasMoreElements()) {
			Object key = en.nextElement();
			Object bucket = hashtable.get(key);
			if (bucket == null) 
			{
				continue;
			}
			ArrayList bucketList = (ArrayList)bucket;
			int cursor = 0;
			while (cursor<bucketList.size()) {
				XATTuple tuple = (XATTuple)bucketList.get(cursor);
				//if (tuple.getNthTimeStamp(joinAttrPos)<latestProcessTupleTime - windowSize) {
				if (tuple.getMinTimeStamp() < latestProcessTupleTime - wSize) 
				{
					bucketList.remove(cursor);
					numTuplesPurged ++;
				} 
				else 
				{
					cursor ++;
				}
			}
		}
		this.decreStateSize(numTuplesPurged);
		return numTuplesPurged;
	}
	
	public int purgeByWindow(long wSize, XATTuple purgetuple, boolean isSenderJoinDuringPT) 
	{
	    long latestProcessTupleTime = purgetuple.getMinTimeStamp();
	    
		Enumeration en = hashtable.keys();
		int numTuplesPurged = 0;
		//int realNumTuplesPurged = 0;
		while(en.hasMoreElements()) 
		{
			Object key = en.nextElement();
			Object bucket = hashtable.get(key);
			if (bucket == null) 
			{
				continue;
			}
			ArrayList bucketList = (ArrayList)bucket;
			int cursor = 0;
			//the following two record the start and end of consecutive purged tuples.
			int startpos = 0;
			int endpos = 0;
			//if the purging tuple is a singleton, purge more aggressively.
			if (purgetuple.isSingletonTuple()) 
			{
                while (cursor < bucketList.size()) 
                {
                    XATTuple tuple = (XATTuple) bucketList.get(cursor);
                    //notice that we now compare state tuple's mintimestamp 
                    //instead of maxtimestamp (as in the following else clause)
                    if (tuple.getMaxTimeStamp() < latestProcessTupleTime - wSize) 
                    {
                        //bucketList.remove(0);
                        if(endpos == cursor) 
                        {
                            endpos ++;
                            cursor ++;
                        } 
                        else 
                        {
                            bucketList.subList(startpos, endpos).clear();
                            startpos = cursor - (endpos-startpos);
                            cursor = startpos+1;
                            endpos = cursor;
                        }
                        
                        //realNumTuplesPurged++;
                        if (!isSenderJoinDuringPT || !tuple.isNew())
                            numTuplesPurged++;
                        
                    } 
                    else 
                    {
                        cursor++;
                    }
                }
                bucketList.subList(startpos, endpos).clear();
            } 
			else 
			{
                while (cursor < bucketList.size()) 
                {
                    XATTuple tuple = (XATTuple) bucketList.get(cursor);
                    //if (tuple.getNthTimeStamp(joinAttrPos)<latestProcessTupleTime - windowSize) {
                    if (tuple.getMaxTimeStamp() < latestProcessTupleTime - wSize) 
                    {
                        //bucketList.remove(0);
                        if(endpos == cursor) 
                        {
                            endpos ++;
                            cursor ++;
                        } 
                        else 
                        {
                            bucketList.subList(startpos, endpos).clear();
                            startpos = cursor - (endpos-startpos);
                            cursor = startpos+1;
                            endpos = cursor;
                        }
                        //realNumTuplesPurged++;
                        if (!isSenderJoinDuringPT || !tuple.isNew())
                            numTuplesPurged++;
                    } 
                    else 
                    {
                        cursor++;
                    }
                }
                bucketList.subList(startpos, endpos).clear();
			}
		}
		this.decreStateSize(numTuplesPurged);
		return numTuplesPurged;
	}
	
	public XATTuple[] probe(Object value) 
	{
		try
		{
			Object bucket = hashtable.get(value);
			if (bucket == null) 
			{
				return null;
			} 
			else 
			{
				return (XATTuple[])((ArrayList)bucket).toArray(new XATTuple[]{});
			}
		}
		catch(Exception ex)
		{
			System.out.println(ex.getMessage());
		}
		return null;
	}
	
	public void insertTuple(Object key, XATTuple tuple) 
	{
		Object bucket = this.hashtable.get(key);
		if (bucket == null) 
		{
			ArrayList bucketList = new ArrayList();
			bucketList.add(tuple);
			hashtable.put(key, bucketList);
		} 
		else 
		{
			((ArrayList)bucket).add(tuple);
		}
		this.increStateSize();
	}
	
	public Hashtable getAllStateTuples() 
	{
		return hashtable;
	}
	
	public int getJoinAttrPos() 
	{
		return joinAttrPos;
	}
	
	//get partition ID of tuples stored in this state
	//all tuples in this state should have the same partition ID during partitioned query processing. 
	public int getTuplePartitionID() 
	{
	    int id = -1;
		Enumeration en = hashtable.keys();
		while(en.hasMoreElements()) {
			Object key = en.nextElement();
			Object bucket = hashtable.get(key);
			if (bucket != null) 
			{
			    ArrayList bucketlist = (ArrayList)bucket;
			    if(bucketlist.size() > 0) 
			        id = ((XATTuple)bucketlist.get(0)).getPartitionID();
			    break;
			} 
		}
	    return id;
	}
	
	public void setJoinAttrPos(int jAttrPos) 
	{
		this.joinAttrPos = jAttrPos;
	}
	
	public long getWindowSize() 
	{
		return windowSize;
	}
	
	public void setWindowSize(long wSize) 
	{
		this.windowSize = wSize;
	}
	
	public void copyStateContent(OperatorState newstate) 
	{
	
	}

	public void clearState() 
	{
		hashtable.clear();
		this.setStateSize(0);
	}
	
	public void init(int stateID)
	{
		this.stateID = stateID;
	}
	
	public String toString() 
	{
	    StringBuffer s = new StringBuffer();
		Enumeration en = hashtable.keys();
		while(en.hasMoreElements()) 
		{
			Object key = en.nextElement();
			Object bucket = hashtable.get(key);
			if (bucket != null) 
			{
			    ArrayList bucketlist = (ArrayList)bucket;
			    s.append("\nbucket key: " + key + "\n" + bucketlist);
			} 
		}
		return s.toString();
	}

	public Iterator getkeys() {
		
		return hashtable.keySet().iterator();
	}

	public boolean removeTuple(XATTuple tuple) {
		
		
		
			Object bucket = hashtable.get(makeKey(tuple));
			if (bucket == null) 
			{
				return false;
			} 
			else 
			{
				ArrayList array=(ArrayList)bucket;
				for(int i=0;i<array.size();i++)
				{
					if(tuple.getMaxTimeStamp()==((XATTuple) array.get(i)).getMaxTimeStamp()
							&&tuple.getMinTimeStamp()==((XATTuple) array.get(i)).getMinTimeStamp())
					{
						array.remove(i);
						this.decreStateSize();
						return true;
					}
				}
			}
		
		
		return false;
	}
	
	public boolean removeTupleforValue(XATTuple tuple)
	{
		
		Object bucket = hashtable.get(makeKey(tuple));
		if (bucket == null) 
		{
			return false;
		} 
		else 
		{
			ArrayList array=(ArrayList)bucket;
			
			
				
					array.remove(0);
					this.decreStateSize();
					return true;
				
			
		}
		
	}
	
	private String makeKey(XATTuple newObj) {
		String key="";
		XATDataValue array[]=newObj.getArrayData();
		for(int i=0;i<array.length;i++)
		{
			key=key+array[i].toString();
		}
		return key;
	}

}
