/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.state;

import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleNegative;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleNegativeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate.PipeWFIFOState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;

public class WFIFOStateDeltaRecompute extends WOperatorState{

	List purgedStates = new ArrayList();
	List corrTuplesInState = new ArrayList();
	ArrayList joinedtuples = new ArrayList();
	List tuplepair = new Vector();
	List<XATDataValue> newvalues = new Vector<XATDataValue>();
	
	protected ArrayList statetuples = new ArrayList();//Array holding the states.
	private long startWindow = 0;//use to keep the start time of the window

	private long windowSize = 5000;//use to keep the end-start time of the window

	//input and output functions for the start and end window
	public void setStartWindow(long start) {
		startWindow = start;
	}

	public void setOperator(XATStreamOperator owner) {
	    operator = owner;
	}
	
	public long getStartWindow() {
		return startWindow;
	}

	public void setWindowSize(long newWindow) {
		windowSize = newWindow;
	}

	
	//protected ArrayList statetuples = new ArrayList();//Array holding the states.

	public int getStateSize() {
		return statetuples.size();
	}

	public ArrayList getStateTuples() {
		return statetuples;
	}

	/**
	 * WFIFOStateDeltaRecompute constructor comment.
	 */
	public WFIFOStateDeltaRecompute() {
		super();
	}

	public WFIFOStateDeltaRecompute(XATStreamOperator op) {
		super();
		this.setOperator(op);
	}

	
	public void init(int stateID){
		this.stateID = stateID;
	}
	/**
	 * Insert of tuples into the statetuple arraylist.
	 */

	public void insertTuple(XATTuple newTuple) {
		statetuples.add(newTuple);
		totalNumTuples++;
	}

	public void insertTuple(XATTuple[] newTuples) {
		for (int i = 0; i < newTuples.length; i++) {
			statetuples.add(newTuples[i]);
		}
		totalNumTuples = totalNumTuples + newTuples.length;
	}

	public void insertTuple(ArrayList newTuples) {
		for (int i = 0; i < newTuples.size(); i++) {
			statetuples.add(newTuples.get(i));
		}
		totalNumTuples = totalNumTuples + newTuples.size();
	}

	/**
	 * @author Abhishek Mukherji
	 * Overwritten expire checking function using the endWindow and startWindow
	 * In case of cross-purge, we can only check that if the diff > endWindow.
	 * In case of self-purge, we need check that if the diff >
	 * endWindow-startWindow.
	 */
	public boolean timeStampExpiredSelf(long expireTimeStamp, XATTuple statetuple) {
		/**
		 * if tuple1 and tuple2 is out of window size from each other, return
		 * false
		 */
		//self-purge checking
		if (expireTimeStamp - statetuple.getMaxTimeStamp() > this.windowSize)
			return true;

		return false;
	}
	/*** @author Abhishek Mukherji
	 * Purge Function with the purged ArrayList returned for self-purge only
	 * purge by timeStamp, used for hopping windows
	 */
	public List purgeByTimeStampSelf(long purgeIncrement) {
		
		purgedStates.clear();
		
		int count = 0;
		//System.out.println("state size before purge = " +statetuples.size());
		//System.out.println("state tuples size: "+ statetuples.size());
		for (int i = 0; i < statetuples.size(); i++) {
			long win = purgeIncrement - this.startWindow;
			
			if (((XATTuple) statetuples.get(i)).getMaxTimeStamp()-this.startWindow < purgeIncrement) {
				count++;
				//System.out.println("PURGED BY HOP: "+ statetuples.get(i));
				purgedStates.add((XATTuple) statetuples.get(i));
				statetuples.remove(i);
				
				//System.out.println("statetuple-startWindow:"+ win);
				//System.out.println("compared with: "+ purgeIncrement);
				
				
			}
		}
		//System.out.println("number of tuples purged = " + count);
		//System.out.println("state size after purge = " + statetuples.size());
		return purgedStates;
	}

	/**
	 * early purge due to the negative tuples
	 * @author Abhishek Mukherji
	 */
	
	public List purgeByNegativeTupleSelf(XATTupleNegative negTuple) {
		
		purgedStates.clear();
		
		if (negTuple == null) {
			return purgedStates;
		}
		int count = 0;

		/**
		 * purge a tuple if it matches the negative tuple
		 */
		for (int i = 0; i < statetuples.size(); i++) {
			
			if (FoundPositiveCorrToNegativeTupleSelf(negTuple, (XATTuple) statetuples.get(i))) {
				//System.out.println("purging due to negative tuple:"+statetuples.get(i));
				purgedStates.add((XATTuple) statetuples.get(i));
				statetuples.remove(i);
				count++;
			}
		}
		return purgedStates;
	}
	
	/**
	 * positive tuple corr to a negative tuple
	 * @author Abhishek Mukherji
	 */
	public boolean FoundPositiveCorrToNegativeTupleSelf(XATTuple negtuple, XATTuple statetuple) {
		/**
		 * if tuple1 and tuple2 match they are the same
		 */
		//self-purge checking
		if(((XATTupleNegative)statetuple).isNegativeFlag()){
		System.out.println("statetuple is negative: " + statetuple);
		}
		
		if (statetuple.equals(negtuple) && negtuple.getMaxTimeStamp()==((XATTuple)statetuple).getMaxTimeStamp() && negtuple.getMinTimeStamp()==((XATTuple)statetuple).getMinTimeStamp()){
			System.out.println("found corr positive tuple!: "+ statetuple + "to tuple: "+negtuple );
			return true;
		}

		return false;
	}
	
	/**
	 * NOT purge but simply return corr tuple(s) to the negative tuples
	 * @author Abhishek Mukherji
	 */
	
	public List CorrespondingTuplesToNegativeTuple(XATTupleNegative negTuple) {
		
		
		corrTuplesInState.clear();
		
		if (negTuple == null) {
			return corrTuplesInState;
		}
		int count = 0;

		/**
		 * collect a tuple if it matches the negative tuple
		 */
		for (int i = 0; i < statetuples.size(); i++) {
			
			if (FoundPositiveCorrToNegativeTupleSelf(negTuple, (XATTuple) statetuples.get(i))) {
				//System.out.println("purging due to negative tuple:"+statetuples.get(i));
				corrTuplesInState.add((XATTupleNegative) statetuples.get(i));
				count++;
			}
		}
		return corrTuplesInState;
	}
	/*added to purge out all the negative tuples from a state.
	 * Used for Delta results. Once the join for the window is computed the negative tuples need to be removed. 
	 * @author Abhishek Mukherji
	 */ 
	public int purgeAllNegativeTupleSelf() {
		//List purgedStates = new ArrayList();
		purgedStates.clear();
		
		int count = 0;

		for (int i = 0; i < statetuples.size(); i++) {
			
			if (((XATTupleNegative) statetuples.get(i)).isNegativeFlag()) {
			System.out.println("removing negative tuple");
				statetuples.remove(i);
				count++;
			}
		}
		//System.out.println("purged negative tuples: "+count);
		return count;
	}
	
	
	@Override
	public boolean checkAllOldTuplePurged() {
		// TODO Auto-generated method stub
		return false;
	}

	
	

	@Override
	public XATTuple[] getAllStateTuples() {
		// TODO Auto-generated method stub
		return (XATTuple[])statetuples.toArray(new XATTuple[]{});
	}

	@Override
	public long[] getEarlistTimeStamp() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public XATTuple getEarlistTuple() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public long[] getLatestTimeStamp() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public XATTuple getLatestTuple() {
		// TODO Auto-generated method stub
		return null;
	}
	
	public boolean timeStampExpiredSelf(XATTuple newtuple, XATTuple statetuple) {
		/**
		 * if tuple1 and tuple2 is out of window size from each other, return
		 * false
		 */
		//self-purge checking
		if (newtuple.getMaxTimeStamp() - statetuple.getMinTimeStamp() > this.windowSize)
			return true;

		return false;
	}

	/**
	 * Purge Function with the purged ArrayList returned for self-purge only
	 */
	public List purgeByTuplePipeSelf(XATTuple newTuple) {
		//List purgedStates = new ArrayList();
		purgedStates.clear();
		
		if (newTuple == null) {
			return purgedStates;
		}

		int count = 0;

		for (int i = 0; i < statetuples.size(); i++) {
			//if(!((XATTupleNegative)statetuples.get(i)).isNegativeFlag()){//only if statetuple is NOT NEGATIVE!
			if (timeStampExpiredSelf(newTuple, (XATTuple) statetuples.get(i))) {
				//System.out.println("purge a tuple from state " +(XATTuple) statetuples.get(i));
				purgedStates.add((XATTuple) statetuples.get(i));
				statetuples.remove(i);
				count++;
			} 
			//}
		}
		return purgedStates;
	}

	/**
	 * Overwritten expire checking function using the endWindow and startWindow
	 * In case of cross-purge, we can only check that if the diff > endWindow.
	 * The reason is that the diff of probing and been probed tuple stamps 
	 * will > endWindow-startWindow for 2nd and later pipejoins.
	 * In case of self-purge, we need check that if the diff >
	 * endWindow-startWindow.
	 */
	public boolean timeStampExpiredCross(XATTuple newtuple, XATTuple statetuple) {
		/**
		 * if tuple1 and tuple2 is out of window size from each other, return
		 * false
		 */
		
		if (newtuple.getMaxTimeStamp() - statetuple.getMinTimeStamp() > this.windowSize)
			return true;

		return false;
	}

	/**
	 * Purge Function with the purged ArrayList returned for cross-purge only
	 */
	public List purgeByTuplePipeCross(XATTuple newTuple) {
		//ArrayList purgedStates = new ArrayList();
		//List purgedStates = new ArrayList();
		purgedStates.clear();
		
		if (newTuple == null) {
			//System.out.println("new tuple is null!");
			return purgedStates;
		}

		int count = 0;

		for (int i = 0; i < statetuples.size(); i++) {
			if (timeStampExpiredCross(newTuple, (XATTuple) statetuples.get(i))) {
				//System.out.println("PurgedCross: "+statetuples.get(i));
				count++;
				purgedStates.add((XATTuple) statetuples.get(i));
				statetuples.remove(i);
				
			}
		}
		
		return purgedStates;
	}
	


	public XATTuple[] probe(XATTuple tuple, ExpressionStr expression, int statepos) {

		int[][] outPositions = operator.getOutputPositions();

		int outputsize = operator.getOutputPositions(0).length
				+ operator.getOutputPositions(1).length;

		//ArrayList joinedtuples = new ArrayList();
		joinedtuples.clear();
		//List tuplepair = new Vector();
		tuplepair.clear();

		//ArrayList subStatetuples;
		
		for (int i = 0; i < statetuples.size(); i++) {

			XATTuple nextstatetuple = (XATTuple) statetuples.get(i);
//System.out.println("probing tuple: "+ tuple);
//System.out.println("state tuple: "+nextstatetuple);
			/* join the two tuples if predicate is satisfied. */
			tuplepair.clear();

			if (statepos == 0) {
				tuplepair.add(nextstatetuple);
				tuplepair.add(tuple);

			} else {
				tuplepair.add(tuple);
				tuplepair.add(nextstatetuple);
			}

			//System.out.println("tuple pair: " + tuplepair.get(0) + " and " +
			//tuplepair.get(1));

			if (((Boolean) expression.eval(tuplepair, operator)).booleanValue() && nextstatetuple.getMaxTimeStamp()-tuple.getMaxTimeStamp()<this.windowSize && nextstatetuple.getMinTimeStamp()-tuple.getMinTimeStamp()<this.windowSize) {
				//List newvalues = new Vector();
				((XATTupleNegative)tuple).setAdded(true);
				((XATTupleNegative)nextstatetuple).setAdded(true);
			//	tuple.combineTimeStamps(tuple,nextstatetuple);
				newvalues.clear();
				
				long[] newtimes = new long[outputsize];
				
				int timeindex = 0;
				for (int j = 0; j < 2; j++) {
					for (int k = 0; k < operator.getOutputPositions(j).length; k++) {
						newvalues.add(((XATTuple) tuplepair.get(j))
								.getCell(outPositions[j][k]));
						//newtimes[timeindex] = ((XATTuple) tuplepair.get(j))	.getMaxTimeStamp();
						timeindex++;
					}
				}
				XATTupleNegative joinedtuple = new XATTupleNegativeImp(newvalues);
				//joinedtuple.setTimeStamp(newtimes);
				joinedtuple.setNegativeFlag(((XATTupleNegative)tuple).isNegativeFlag()||((XATTupleNegative)nextstatetuple).isNegativeFlag()?true:false);
				
				long m1 = tuple.getMinTimeStamp();
				long m2 = nextstatetuple.getMinTimeStamp();
				joinedtuple.setMinTimeStamp((m1 < m2) ? m1 : m2);
				m1 = tuple.getMaxTimeStamp();
				m2 = nextstatetuple.getMaxTimeStamp();
				joinedtuple.setMaxTimeStamp((m1 > m2) ? m1 : m2);
				
				if(joinedtuple.isNegativeFlag()){
					//System.out.println("joinedtuple is NEGATIVE"+ joinedtuple.toString());
					//if(((XATTupleNegative)tuple).isNegativeFlag())
						//System.out.println(tuple.toString());
					//else
						//System.out.println(nextstatetuple.toString());
				}
				joinedtuples.add(joinedtuple);
				if (tuple.isNew() && nextstatetuple.isNew()) {
					joinedtuple.setToNew(true);
				} else {
					joinedtuple.setToNew(false);
				}
			}
		}
		return (XATTuple[]) joinedtuples.toArray(new XATTuple[] {});
	}

	@Override
	public void purgeByTuple(XATTuple newTuple) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void purgeByTuple(XATTuple newTuple, long overall_min) {
		// TODO Auto-generated method stub
		
	}

	public void clearState() {
		// TODO Auto-generated method stub
		
	}

	public void copyStateContent(OperatorState newstate) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public XATTuple[] probe(XATTuple[] tuples, ExpressionStr expression, int statepos) {
		// TODO Auto-generated method stub
		return null;
	}
}
