/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.state;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;

import java.util.ArrayList;

//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObjectImp;

//import streamgenerator.*;
/**
 * Insert the type's description here.
 * Creation date: (2/17/2003 4:38:29 PM)
 * @author: Yali Zhu
 */
public abstract class WOperatorState extends OperatorStateImp {
	//protected boolean isMatched = false; //used by moving-state-migration strategy
	
	protected long timeWindow;
	protected XATQueueSchema stateSchema;
	//protected XATStreamOperator operator;
	protected boolean allOldPurged = false;
	
	public WOperatorState() {
		super();
	}
	
	public WOperatorState(WOperatorState oldstate, XATStreamOperator newop) {
		super(oldstate, newop);
		this.debugFlag = oldstate.debugFlag;
		this.isMatched = false;
		this.stateSchema = oldstate.stateSchema;
		this.timeWindow = oldstate.timeWindow;
	}
/**
 * Insert the method's description here.
 * Creation date: (4/10/2003 9:19:49 AM)
 * @return boolean
 */
public abstract boolean checkAllOldTuplePurged();
/**
 * Insert the method's description here.
 * Creation date: (6/3/2003 2:07:16 AM)
 * @return boolean
 */
public synchronized boolean getAllOldPurgedFlag() {
	return allOldPurged;
}

public abstract XATTuple[] getAllStateTuples();
public abstract ArrayList getStateTuples();
/**
 * Insert the method's description here.
 * Creation date: (6/1/2003 7:36:13 AM)
 * @return long
 */
public abstract long[] getEarlistTimeStamp();
/**
 * Insert the method's description here.
 *
 * Get tuple with earliest timestamp
 *
 * Creation date: (6/1/2003 7:36:13 AM)
 * @return long
 */
public abstract XATTuple getEarlistTuple();
/**
 * Insert the method's description here.
 * Creation date: (6/1/2003 7:36:35 AM)
 * @return long
 */
public abstract long[] getLatestTimeStamp();
/**
 * Insert the method's description here.
 *
 * Get tuple with latest timestamp
 *
 * Creation date: (6/1/2003 7:36:35 AM)
 * @return long
 */
public abstract XATTuple getLatestTuple();

/**
 * Insert the method's description here.
 * Creation date: (5/28/2003 6:07:56 AM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator
 */
public XATStreamOperator getOperator() {
	return operator;
}
/**
 * Insert the method's description here.
 * Creation date: (4/10/2003 9:00:32 AM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema
 */
public XATQueueSchema getSchema() {
	return stateSchema;
}

/**
 * Insert the method's description here.
 * Creation date: (6/23/2003 5:38:12 AM)
 * @return int
 */
//public abstract int getStateTupleCounts();
/**
 * Insert the method's description here.
 * Creation date: (3/15/2003 10:23:33 PM)
 * @return int
 */
public long getTimeWindow() {
	return timeWindow;
}
/**
 * Insert the method's description here.
 * Creation date: (2/25/2003 11:22:36 AM)
 */
public abstract void insertTuple(XATTuple newTuple);
public abstract void insertTuple(XATTuple[] newTuple);
public abstract void insertTuple(ArrayList newTuples);
/**
 * Insert the method's description here.
 * Creation date: (5/28/2003 6:10:14 AM)
 * @param operator edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator
 */
public void setOperator(XATStreamOperator owner) {
    operator = owner;
}
/**
 * Insert the method's description here.
 * Creation date: (4/10/2003 9:00:53 AM)
 * @param schema edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema
 */
public void setSchema(XATQueueSchema schema) {
    //stateQ.setSchema(schema);
    stateSchema = schema;
}
/**
 * Insert the method's description here.
 * Creation date: (5/25/2003 2:54:45 AM)
 * @param newWindow long
 */
public void setTimeWindow(long newWindow) {
    timeWindow = newWindow;
}

	/**
	 * Insert the method's description here. Creation date: (5/20/2003 2:06:45 AM)
	 * @return boolean
	 * @param t1  streamgenerator.XATStreamTuple
	 * @param t2  streamgenerator.XATStreamTuple
	 */
	public boolean timeStampExpired(XATTuple newtuple, XATTuple statetuple) {
		/**
		 * if tuple1 and tuple2 is out of window size from each other, return
		 * false
		 */
		// commented out and switch to self-purge
		if (newtuple.getMinTimeStamp() - statetuple.getMaxTimeStamp() > timeWindow)
			return true;
		//self-purge checking
		//if (newtuple.getMaxTimeStamp() - statetuple.getMinTimeStamp() > timeWindow)
		//	return true;
		//if (tuple2.getMaxTS() - tuple1.getMinTS() > timeWindow)
		//    return true;

		return false;
	}
	
	// Purge tuples based on the value of timestamp column
	public boolean purgeByTimeStamp(XATTuple newtuple, XATTuple statetuple)
	{
		
		//System.out.println("timeWindow: " + timeWindow);
		XATDataValue newTupleTsVal = newtuple.getCell(0);
		long newTupleTs = newTupleTsVal.convertToLong();
		XATDataValue stateTupleTsVal = statetuple.getCell(0);
		long stateTupleTs = stateTupleTsVal.convertToLong();
		if ((newTupleTs - stateTupleTs )> timeWindow)
		 return true;
		 
		return false; 
		
	}
	
	 

/**
 * Insert the method's description here. Creation date: (6/27/2003 12:37:29 PM)
 * 
 * @return java.lang.String
 */
public String toString() {
	return ("State " + stateSchema);
}

public abstract void purgeByTuple(XATTuple newTuple);

public abstract void purgeByTuple(XATTuple newTuple, long overall_min);
//public abstract void probe(XATTuple tuple);
//public abstract void probe(XATTuple tuple, ExpressionStr expression, int statepos);
//public abstract void probe(XATTuple[] tuples);
public abstract XATTuple[] probe(XATTuple[] tuples, ExpressionStr expression, int statepos);

/**
 * Insert the method's description here. Creation date: (5/20/2003 2:06:45 AM)
 * @return boolean
 * @param t1  streamgenerator.XATStreamTuple
 * @param t2  streamgenerator.XATStreamTuple
 */
public boolean timeStampExpired(XATTuple newtuple, XATTuple statetuple, long overall_min) {
	/**
	 * if tuple1 and tuple2 is out of window size from each other, return
	 * false
	 */
	// commented out and switch to self-purge
	long statetime = statetuple.getMaxTimeStamp();
	
	if ((overall_min - statetime > timeWindow) && (newtuple.getMinTimeStamp() - statetime > timeWindow))
		return true;
	
	//self-purge checking
	//if (newtuple.getMaxTimeStamp() - statetuple.getMinTimeStamp() > timeWindow)
	//	return true;
	//if (tuple2.getMaxTS() - tuple1.getMinTS() > timeWindow)
	//    return true;

	return false;
}
	
}
