/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on May 6, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;

/**
 * @author lisading
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class StreamFunctionAvg extends StreamFunctionImp {
	
	int numTuples = 0;
	
	public StreamFunctionAvg() {
		super();
		this.currentValue = 0.0;
		this.initialValue = 0.0;
		this.funcName = "AVG";
	}
	public double evaluateDouble(XATTuple[] tuples) {
		double result = 0.0;
		int length = tuples.length;
		
		//go through the tuples and sum all the result
		for (int i=0; i<length; i++) {
			Object objValue = tuples[i].getCell(attrPos).getValue();
			double value = 0.0;
			if (objValue instanceof Integer) {
				value = ((Integer)objValue).doubleValue();
			} else if (objValue instanceof Long) {
				value = ((Long)objValue).doubleValue();
			} else if (objValue instanceof Double) {
				value = ((Double)objValue).doubleValue();
			}
			result += value;
		}
		
		result = result / (double)length;
		return result;
	}

	public double evaluateDouble(XATTuple[] tuples, double currentVal, int currentNumTuples) {
		double result = 0.0;
		int length = tuples.length;
		
		//go through the tuples and sum all the result
		for (int i=0; i<length; i++) {
			Object objValue = tuples[i].getCell(attrPos).getValue();
			double value = 0.0;
			if (objValue instanceof Integer) {
				value = ((Integer)objValue).doubleValue();
			} else if (objValue instanceof Long) {
				value = ((Long)objValue).doubleValue();
			} else if (objValue instanceof Double) {
				value = ((Double)objValue).doubleValue();
			}
			result += value;
		}
		
		double newResult = (result + currentVal * currentNumTuples) 
			/ (double)(length + currentNumTuples);
		
		return newResult;
	}
	
	public double evaluateDouble(XATTuple[] tuples, int startIndex, int endIndex) {
		double result = 0.0;
		//go through the tuples and sum all the result
		for (int i=startIndex; i<=endIndex; i++) {
			Object objValue = tuples[i].getCell(attrPos).getValue();
			double value = 0.0;
			if (objValue instanceof Integer) {
				value = ((Integer)objValue).doubleValue();
			} else if (objValue instanceof Long) {
				value = ((Long)objValue).doubleValue();
			} else if (objValue instanceof Double) {
				value = ((Double)objValue).doubleValue();
			}
			result += value;
		}
		
		result = result / (double)(endIndex-startIndex+1);
		return result;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamFunction#evaluate(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[])
	 */
	public XATDataValue evaluate(XATTuple[] tuples) {
		double result = 0.0;
		int length = tuples.length;
		
		//go through the tuples and sum all the result
		for (int i=0; i<length; i++) {
			Object objValue = tuples[i].getCell(attrPos).getValue();
			double value = 0.0;
			if (objValue instanceof Integer) {
				value = ((Integer)objValue).doubleValue();
			} else if (objValue instanceof Long) {
				value = ((Long)objValue).doubleValue();
			} else if (objValue instanceof Double) {
				value = ((Double)objValue).doubleValue();
			}
			result += value;
		}
		
		result = result / (double)length;
		
		XATDataValue newValue = new XATDataValueImp(new Double(result));
		return newValue;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamFunction#evaluate(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[], double, int)
	 */
	public XATDataValue evaluate(XATTuple[] tuples, double currentVal, int currentNumTuples) {
		double result = 0.0;
		int length = tuples.length;
		
		//go through the tuples and sum all the result
		for (int i=0; i<length; i++) {
			Object objValue = tuples[i].getCell(attrPos).getValue();
			double value = 0.0;
			if (objValue instanceof Integer) {
				value = ((Integer)objValue).doubleValue();
			} else if (objValue instanceof Long) {
				value = ((Long)objValue).doubleValue();
			} else if (objValue instanceof Double) {
				value = ((Double)objValue).doubleValue();
			}
			result += value;
		}
		
		double newResult = (result + currentVal * currentNumTuples) 
			/ (double)(length + currentNumTuples);
		
		XATDataValue newValue = new XATDataValueImp(new Double(newResult));
		return newValue;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamFunction#evaluate(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[], int, int)
	 */
	public XATDataValue evaluate(XATTuple[] tuples, int startIndex, int endIndex) {
		double result = 0.0;
		//go through the tuples and sum all the result
		for (int i=startIndex; i<=endIndex; i++) {
			Object objValue = tuples[i].getCell(attrPos).getValue();
			double value = 0.0;
			if (objValue instanceof Integer) {
				value = ((Integer)objValue).doubleValue();
			} else if (objValue instanceof Long) {
				value = ((Long)objValue).doubleValue();
			} else if (objValue instanceof Double) {
				value = ((Double)objValue).doubleValue();
			}
			result += value;
		}
		
		result = result / (double)(endIndex-startIndex+1);
		
		XATDataValue newValue = new XATDataValueImp(new Double(result));
		return newValue;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamFunction#increment(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[])
	 */
	public void increment(XATTuple[] tuples) {
		double sum = 0.0;
		int length = tuples.length;
		numTuples += length;
		for (int i=0; i<length; i++) {
			sum += ((Double)tuples[i].getCell(attrPos).getValue()).doubleValue();
		}
		sum+=currentValue*numTuples;
		currentValue = sum / (double) numTuples;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamFunction#decrement(edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[])
	 */
	public void decrement(XATTuple[] tuples) {
		double sum = 0.0;
		int length = tuples.length;
		numTuples -= length;
		for (int i=0; i<length; i++) {
			sum += ((Double)tuples[i].getCell(attrPos).getValue()).doubleValue();
		}
		sum=currentValue*numTuples-sum;
		currentValue = sum / (double) numTuples;
	}

}
