/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on May 8, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import java.util.Vector;

/**
 * @author lisading
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class StreamGroupbyOperatorImp
	extends StreamAggregateOperatorImp
	implements StreamGroupbyOperator {

	protected int[] groupAttrPosArray;
	protected int numGroupAttrs;
	
	/**
	 * constructor
	 * @param init
	 */
	public StreamGroupbyOperatorImp(XATNode init) {
		super(init);
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() throws Exception {
		if (!hasInit()) {
			init();
			if (!hasInit()) {
				return false;
			}
		}
		// group-by execution logic
		return true;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamGroupbyOperator#getGroupAttrPos(int)
	 */
	public int getGroupAttrPos(int pos) {
		return this.groupAttrPosArray[pos];
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamGroupbyOperator#getGroupAttrPosArray()
	 */
	public int[] getGroupAttrPosArray() {
		return this.groupAttrPosArray;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamGroupbyOperator#setGroupAttrPos(int, int)
	 */
	public void setGroupAttrPos(int pos, int gAttrPos) {
		this.groupAttrPosArray[pos] = gAttrPos;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamGroupbyOperator#setGroupAttrPosArray(int[])
	 */
	public void setGroupAttrPosArray(int[] gAttrPosArray) {
		this.groupAttrPosArray = gAttrPosArray;
	}

	public int getNumGroupAttrs() {
		return numGroupAttrs;
	}
	
	public void setNumGroupAttrs(int nGroupAttrs) {
		numGroupAttrs = nGroupAttrs;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		return "Groupby";
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#init()
	 */
	public void init() {
		numAggFunctions = aggFunctions.length;
		numGroupAttrs = groupAttrPosArray.length;
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength()==0) {
			return;
		}
		XATQueueSchema outSchema = this.getOutputQueue().getSchema();
		if (outSchema == null || outSchema.getLength()==0) {
			outSchema = new XATQueueSchema();
			outSchema.setName(inSchema.getName());
			String[] colNames = (String[])inSchema.getColNames().toArray(new String[] {});
			for (int i=0; i<numGroupAttrs; i++) {
				outSchema.addColumn(colNames[groupAttrPosArray[i]]);
			}
			for (int i=0; i<numAggFunctions; i++) {
				StreamFunction func = aggFunctions[i];
				func.setAttrName(colNames[func.getAttrPos()]);
				outSchema.addColumn(func.getFuncName()+"_"+func.getAttrName());
			}

			this.getOutputQueue().setSchema(outSchema);
		}
		
		// compute output positions
		Vector outColNames = this.getOutputQueue().getSchema().getColNames();
		Vector inColNames = this.getInputQueue(0).getSchema().getColNames(); 
		numOutputColumns = outColNames.size();
		int[] outputPos = new int[numOutputColumns-numAggFunctions];
		
		for (int i=0; i<(numOutputColumns-numAggFunctions); i++){
			Object outputCol = outColNames.get(i);
			outputPos[i] = inColNames.indexOf(outputCol);
		}
		this.setOutputPositions(0, outputPos);
		setHasInit(true);
	}
	
	protected XATPunct generateResultPunct(XATPunct punct) {
		XATDataValue[] resultPunctContent = new XATDataValue[numOutputColumns];
		int[] resultPunctTypeArray = new int[numOutputColumns];
		long[] resultPunctTime = new long[numOutputColumns];
		XATDataValue[] punctContent = punct.getArrayData();
		int[] punctTypeArray = punct.getPunctTypeArray();
		long[] punctTime = punct.getTimeStamp();
		long minTime = punct.getMinTimeStamp();
		long maxTime = punct.getMaxTimeStamp();

		int i=0;
		for (; i<numOutputColumns-numAggFunctions; i++) {
			int index = outputPositions[0][i];
			resultPunctContent[i] = punctContent[index];
			resultPunctTypeArray[i] = punctTypeArray[index];
			resultPunctTime[i] = punctTime[index];
		}
		for (int j=0; j<numAggFunctions; j++) {
			resultPunctContent[j+i] = new XATDataValueImp("*");
			resultPunctTypeArray[j+i] = XATPunct.ALL_VALUE;
			resultPunctTime[j+i] = minTime;
		}
		XATPunct resultPunct = new XATPunctImp(resultPunctContent);
		resultPunct.setPunctTypeArray(resultPunctTypeArray);
		resultPunct.setTimeStamp(resultPunctTime);
		resultPunct.setMinTimeStamp(minTime);
		resultPunct.setMaxTimeStamp(maxTime);
		return resultPunct;
	}
	
	
	/**
	 * generate result tuple according to the schema of the output queue
	 * @param group the input tuples that belong to a group 
	 */
	protected XATTuple generateResultForSingleGroup(XATTuple[] group) {
		XATDataValue[] resultValue = new XATDataValue[numOutputColumns];
		long[] resultTime = new long[numOutputColumns];
		XATTuple firstTupleInGroup = group[0];
		XATDataValue[] firstTupleContent = firstTupleInGroup.getArrayData();
		long[] firstTupleTime = firstTupleInGroup.getTimeStamp();
		long minTime = firstTupleInGroup.getMinTimeStamp();
		long maxTime = firstTupleInGroup.getMaxTimeStamp();

		int i=0;
		for (; i<numOutputColumns-numAggFunctions; i++) {
			int index = outputPositions[0][i];
			resultValue[i] = firstTupleContent[index];
			resultTime[i] = firstTupleTime[index];
		}
		for (int j=0; j<numAggFunctions; j++) {
			resultValue[j+i] = aggFunctions[j].evaluate(group);
			resultTime[j+i] = minTime;
		}
		XATTuple resultTuple = new XATTupleImp(resultValue);
		resultTuple.setTimeStamp(resultTime);
		resultTuple.setMinTimeStamp(minTime);
		resultTuple.setMaxTimeStamp(maxTime);
		return resultTuple;
	}
}
