/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on 2004-3-25
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;

/**
 * @author lisading
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class StreamProjectOperatorImp
	extends XATSingleQueueStreamOperatorImp
	implements StreamProjectOperator {

	String[] retainAttrNames; //assumes no name conflict
	//	retainAttrPos[0] is the index of the input queue schema that matches the first column of the output queue schema. 
	//  int[] retainAttrPos;  

	public StreamProjectOperatorImp(StreamProjectOperatorImp oldop, XATNode newnode) {
		super(oldop, newnode);
		String[] rAttrNames = oldop.getRetainAttrNames();
		retainAttrNames = new String[rAttrNames.length];
		for (int i=0; i<rAttrNames.length; i++) {
			retainAttrNames[i] = rAttrNames[i];
		}
	}
	/**
	 * constructor
	 * @param init
	 */
	public StreamProjectOperatorImp(XATNode init) {
		super(init);
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamProjectOperator#getAttrNames()
	 */
	public String[] getRetainAttrNames() {
		return retainAttrNames;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamProjectOperator#setAttrName(int, java.lang.String)
	 */
	public void setRetainAttrName(int pos, String newName) {
		retainAttrNames[pos] = newName;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamProjectOperator#setAttrNames(java.lang.String[])
	 */
	public void setRetainAttrNames(String[] attrs) {
		retainAttrNames = attrs;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() throws Exception {
		if (!hasInit()) {
			init();
			if (!hasInit()) {
				return false;
			}
		}
				
		/*
		if(this.getOutputQueue().getSchema() == null){
			if (this.getInputQueue().getSchema() == null){
				return false;
			}
			init();
		}*/
		
		int numColumns = retainAttrNames.length;
		
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		//int howMany = 100;

		/** get the tuples */
		XATTuple[] tuples = this.getInputQueue().dequeue(howMany, this);

		for (int i=0; i<tuples.length; i++) {
			XATTuple tuple = tuples[i];
			
			// for distributed migration
			if(tuple.getMaxTimeStamp() == -1) {
				this.processSyncTuple();
				continue;
			}
			
			if (tuple.getObjTypeID()==0) {
				this.getOutputQueue().enqueue(this.generateResultTuple(tuple), this);
			} else {
				XATPunct newPunct = new XATPunctImp(numColumns);
				long[] inputTimeStamp = tuple.getTimeStamp();
				long[] newTimeStamp = new long[numColumns];

				for (int j=0; j<numColumns; j++) {
					newPunct.setCell(j, tuple.getCell(outputPositions[0][j]));
					newTimeStamp[j] = inputTimeStamp[outputPositions[0][j]];
				}
				newPunct.setTimeStamp(newTimeStamp);
				newPunct.setMinTimeStamp(tuple.getMinTimeStamp());
				newPunct.setMaxTimeStamp(tuple.getMaxTimeStamp());
				int[] punctTypeArray = ((XATPunct)tuple).getPunctTypeArray();
				int[] newPunctTypeArray = new int[numColumns];
				for (int j=0; j<numColumns; j++) {
					newPunctTypeArray[j] = punctTypeArray[outputPositions[0][j]];
				}
				newPunct.setPunctTypeArray(newPunctTypeArray);
				this.getOutputQueue().enqueue(newPunct, this);
			}
		}
		return true;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#init()
	 */
	public void init() {
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) {
			return;
		}
		
		XATQueueSchema outSchema = this.getOutputQueue().getSchema();
		if (outSchema == null || outSchema.getLength()==0) {
			outSchema = new XATQueueSchema();
			outSchema.setName(this.getInputQueue().getSchema().getName());
			for (int i=0; i<retainAttrNames.length; i++) {
				outSchema.addColumn(retainAttrNames[i]);
			}
			this.getOutputQueue().setSchema(outSchema);
		}
		this.computeOutputPositions();
		numOutputColumns = outputPositions[0].length;
		this.setHasInit(true);
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		// TODO Auto-generated method stub
		return "Project";
	}
	
	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input queues.
	 * Inside method run() the numbers are computed based on the property
	 * "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler,
	 * while in cleanAccumulatedTuples, all tuples in input queues that have their max timesamp
	 * smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return false;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() {
		return false;
	}
	
	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() {
		return false;
	}

	public boolean matchStates(XATStreamOperator op) {
		return false;
	}
	
	public boolean recomputeUnmatchedState() {
		return true;
	}
	
	public int setStateID(int nextid) {
		return nextid;
	}
	public boolean finalizeOperation(){
		return true;
	}


}
