/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Apr 12, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStrImp;

import java.util.Vector;
/**
 * @author Yali Zhu
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class StreamSelectOperatorImp extends XATSingleQueueStreamOperatorImp 
	implements StreamSelectOperator {

	private ExpressionStr expression;
	
	public StreamSelectOperatorImp(XATNode init) {
		super(init);
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() throws Exception {
		/**
		 * first figure out how many tuples to dequeue, the selectivity, and 
		 * result delay  
		 */
		XATQueueSchema inSchema;
		int inLength;
		if ((inSchema = this.getInputQueue().getSchema()) == null){
				return false;
		}
		inLength = inSchema.getLength();
		
		//modified by Song Wang to support multiple output queue
		for (int i=0;i<this.getOutputQueueCount();i++){
			if(this.getOutputQueue(i).getSchema() == null){
				init(i);
			}
		}
		
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		/** get the tuples */

		XATTuple[] tuples = this.getInputQueue().dequeue(howMany, this);
		/** figure out how many tuples to keep*/
		Vector tupleList = new Vector();
		for (int i=0; i<tuples.length; i++) {
			
			// for distributed migration
			if(tuples[i].getMaxTimeStamp() == -1) {
				this.processSyncTuple();
				continue;
			}
			
			//tupleList.clear();
			tupleList.add(0,tuples[i]);
			Object out = expression.eval(tupleList, this);
			if(out instanceof Boolean) {
				if(((Boolean)out).booleanValue()) {
					//XATTuple newtuple = new XATTupleImp()
					System.out.println("Starting iterating OutputQueue.");
					System.out.println("getOutputQueueCount:" + this.getOutputQueueCount() );
					for (int j=0;j<this.getOutputQueueCount();j++){
						this.getOutputQueue(j).enqueue(tuples[i], this);
						System.out.println("generate tuple:" + tuples[i].toString());
					}
				}
			}
			
			
		}
			
//			//the following code is wrong!!!
//			XATTuple newTuple = new XATTupleImp(projAttrPosArray.length);
//			for (int j=0; j<projAttrPosArray.length; j++) {
//				newTuple.setCell(j, tuples[i].getCell(projAttrPosArray[j]));
//			}
//			newTuple.setTimeStamp(tuples[i].getTimeStamp());
//			
//			this.getOutputQueue().enqueue(newTuple, this);
//		}

		return true;
	}
	//overload supper method
	public void init(int i) {
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) {
			return;
		}
		XATQueueSchema outSchema = this.getOutputQueue(i).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) {
			this.getOutputQueue(i).setSchema(inSchema);
		}
		System.out.println("\t run select Op init ");
		computeOutputPositions();//no need to overload since all the output queue are the same
		numOutputColumns = outputPositions[0].length;
		setHasInit(true);
	}

	public String toString() {
		return "Select";
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		// TODO Auto-generated method stub
		return "Select";
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamSelectOperator#getExpression()
	 */
	public ExpressionStr getExpression() {
		return expression;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamSelectOperator#setExpression(edu.wpi.cs.dsrg.xmldb.xat.common.streamexpression.ExpressionStr)
	 */
	public void setExpression(ExpressionStr expression) {
		this.expression = expression;
	}
	
	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input queues.
	 * Inside method run() the numbers are computed based on the property
	 * "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler,
	 * while in cleanAccumulatedTuples, all tuples in input queues that have their max timesamp
	 * smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() {
		return true;
	}
	
	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() {
		return false;
	}

	public boolean matchStates(XATStreamOperator op) {
		return false;
	}
	
	public boolean recomputeUnmatchedState() {
		return true;
	}
	
	public int setStateID(int nextid) {
		return nextid;
	}
	public boolean finalizeOperation(){
		return true;
	}

}