/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on May 20, 2004
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator;

import java.util.Iterator;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * @author lisading
 * Takes tuples from multiple input queues, output them into one outputqueue. 
 * Here,we assume that the imput schema of all inputs are the same. 
 */
public class StreamUnionOperatorImp extends StreamSetOperatorImp 
	implements XATMultiQueueStreamOperator {

	int totalEOFNumber = 0; // indicate how many EOFs should be received in the finalization() 
	int EOFReceived = 0; // how many have been received so far
	int totalreceived = 0; // total number of tuples received in the finalization stage
	int timeOutIndicator = 0;
	
	public StreamUnionOperatorImp(XATNode init) {
		super(init);
	}
	
 	/*
 	 * In this union operator, we do not have to combine schemas from multiple inputqueues, simply take the
 	 * first one from queue and use this as the output queue. It override the init() of XATMultiQueueStreamOperatorImp
 	 */
 	public void init() {

 		Iterator i = this.getInputQueues();
 		String schemaName = new String();
 		XATQueueSchema s;
 		XATQueueSchema outSchema = new XATQueueSchema();

 		if (i.hasNext()) {

 			s = ((XATQueue) i.next()).getSchema();
 			schemaName += "#" + s.getName();

 			Iterator j = s.getColNames().iterator();
 			while(j.hasNext()){
 				outSchema.addColumn((String) j.next());
 			}
 		}

 		outSchema.setName(schemaName);
 		this.getOutputQueue().setSchema(outSchema);

 		computeOutputPositions();
 		this.hasInit = true;
 	}

	/* 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#run()
	 */
	public boolean run() {
		//if (!hasInit()) {
		//	init();
		//	if (!hasInit()) {
		//		return false;
		//	}
		//}

		//System.out.println("-----");
		XATQueueSchema inSchema;
		int inLength;
		
		if ((inSchema = this.getInputQueue().getSchema()) == null){
			return false;
		}
		inLength = inSchema.getLength();
		
		if(this.getOutputQueue().getSchema() == null){
			init();
		}
		
//		System.out.println("\nrun " + this);
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"))/this.getInputQueueCount();
		//System.out.println("UNION OPERATOR - NUMBER_OF_TUPLES_TO_DEQUEUE:" + howMany);
		//System.out.println("Number of Input Queues"+this.getInputQueueCount());
		
		int tupleCount = 0;
		/** get the tuples */
		int k=0;
		for (Iterator i=this.getInputQueues(); i.hasNext() ;) {
			XATQueue theQueue = (XATQueue)i.next();
			XATTuple[] tuples = (theQueue).dequeue(howMany, this);		
			tupleCount += tuples.length;
			//System.out.println(k+" "+tuples.length);
			//XATTuple[] tuples = ((XATQueue)i.next()).dequeue(howMany, this);
			
			/*
			for(int z=0; z<tuples.length;z++)
			{
				System.out.println(tuples[z].toString());
			}
			*/
			
			//System.out.println("Size in Union:" + tuples.length);
			this.getOutputQueue().enqueue(tuples,this);
			k++;
		}		
		//System.out.println("+++UNION: IN TUPLES:" + this.getInputQueue().getQueueSize() +  "UNION OUTPUT QUEUE:" + this.getOutputQueue().getQueueSize());
//		System.out.println("Number of tuples dequeued: " + tupleCount);
		//System.out.println("-----");
		return true;
		
	} //end run()
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#getStateSize()
	 */
	public int getStateSize() {
		return 0;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#toToolTipString()
	 */
	public String toToolTipString() {
		// TODO Auto-generated method stub
		return null;
	}
	
	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input queues.
	 * Inside method run() the numbers are computed based on the property
	 * "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler,
	 * while in cleanAccumulatedTuples, all tuples in input queues that have their max timesamp
	 * smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() {
		return true;
	}
	
	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() {
		return false;
	}

	public boolean matchStates(XATStreamOperator op) {
		return false;
	}
	
	public boolean recomputeUnmatchedState() {
		return true;
	}
	
	public int setStateID(int nextid) {
		return nextid;
	}

	public void setTotalEOFNumer(int num){
		totalEOFNumber = num;
	}
	/**
	 * qqq
	 */
	public boolean finalizeOperation(){
		boolean done = false;
		this.active = false;

		while (!done){			
			for (Iterator i=this.getInputQueues(); i.hasNext() ;) {
				XATQueue cq = (XATQueue)i.next();
				System.out.println("Input Queue Size:" + cq.getQueueSize());
				while (cq.getQueueSize(this)>0) {
					XATTuple tp = cq.dequeue(this);
					if (tp.getPartitionID() != -10) {
						//only count the number of tuples have been received to save the memory space in the finalization 
						//this.getOutputQueue().enqueue(tp,this);
						totalreceived ++;
						tp = null;
					} else {
						EOFReceived ++;
						tp = null;
					}
				}
			}
			if (EOFReceived != totalEOFNumber) {
				System.out.println("UNION FINALIZATION:" + EOFReceived + "QUEUE COUNTER:" + this.getInputQueueCount() + "Total EOF:" + totalEOFNumber);
				try {
					Thread.sleep(500);
				} catch (Exception e) {
					e.printStackTrace();
				}
				timeOutIndicator ++;
				if (timeOutIndicator >= 2000) {
					return true;
				}
				return false;
			} else { 
				done = true;
			}
		}
		// 
		System.out.println("TOTAL RECEIVED AFTER TERMINATED: " + totalreceived);
		return true;
	}
}
