/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATMultiSourceOperator;
/**
 * Insert the type's description here.
 * Creation date: (1/6/2003 3:00:20 PM)
 * @author: Brad Pielech
 */
public interface XATMultiQueueStreamOperator extends XATMultiSourceOperator, XATStreamOperator {
/**
 * Adds another input queue to this operator.  This will be added to the
 * right of the others
 * @param queue edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 */
void addInputQueue(XATQueue queue);
/**
 * Adds another output queue to this operator.  This will be added to the
 * right of the others
 * @param queue edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 */
void addOutputQueue(XATQueue queue);
/**
 * get the left most input queue
 * this has the effect of calling super.getInputQueue()
 * Creation date: (1/6/2003 4:21:19 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 */
XATQueue getLeftInputQueue();
/**
 * get the left most output queue
 * this has the effect of calling super.getOutputQueue()
 * Creation date: (1/6/2003 4:21:19 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 */
XATQueue getLeftOutputQueue();
/**
 * Get the Nth input queue.  This is useful when there are more than 2 input
 * queues
 * Creation date: (1/6/2003 4:21:19 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 * @throws ArrayIndexOutOfBoundsException if the Nth queue does not exist
 */
XATQueue getNthInputQueue(int N) throws ArrayIndexOutOfBoundsException;
/**
 * Get the Nth output queue.  This is useful when there are more than 2 output
 * queues
 * Creation date: (1/6/2003 4:21:19 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 * @throws ArrayIndexOutOfBoundsException if the Nth queue does not exist
 */
XATQueue getNthOutputQueue(int N) throws ArrayIndexOutOfBoundsException;
/**
 * get the right most input queue.  This gets the right-most
 * input queue.  If an operator has more than 2 inputs, this will get the
 * farthest right input.  This has the same effect as calling getNthInput(2); 
 * Creation date: (1/6/2003 4:21:19 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 */
XATQueue getRightInputQueue();
/**
 * get the right most output queue.  This gets the right-most
 * output queue.  If an operator has more than 2 outputs, this will get the
 * farthest right output.  This has the same effect as calling getNthOutput(2); 
 * Creation date: (1/6/2003 4:21:19 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 */
XATQueue getRightOutputQueue();
/**
 * set the left most input queue
 * This first register's the Operator with the queue 
 * this has the effect of calling super.setInputQueue()
 * Creation date: (1/6/2003 4:21:19 PM)\
 * @param newQueue the new left-most input queue
 */
void setLeftInputQueue(XATQueue newQueue);
/**
 * set the left most output queue
 * this has the effect of calling super.setOutputQueue()
 * The operator must be the owner of the queue, otherwise an exception will be thrown.
 * Creation date: (1/6/2003 4:21:19 PM)
 * @param newQueue the new left-most input queue
 */
void setLeftOutputQueue(XATQueue newQueue);
/**
 * Set the Nth input queue.  This is useful when there are more than 2 input
 * queues
 * This first register's the Operator with the queue 
 * Creation date: (1/6/2003 4:21:19 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 * @param newQueue the new Nth input queue
 * @param int N the inputQueue to set
 */
void setNthInputQueue(XATQueue newQueue, int N);
/**
 * Set the Nth output queue.  This is useful when there are more than 2 output
 * queues
 * The operator must be the owner of the queue, otherwise an exception will be thrown.
 * Creation date: (1/6/2003 4:21:19 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 * @param newQueue the new Nth output queue
 * @param int N the outputQueue to set
 */
void setNthOuputQueue(XATQueue newQueue, int N);
/**
 * set the right most input queue 
 * Creation date: (1/6/2003 4:21:19 PM)\
 * @param newQueue the new right-most input queue
 */
void setRightInputQueue(XATQueue newQueue);
/**
 * set the right most output queue
 * this has the effect of calling super.setOutputQueue()
 * The operator must be the owner of the queue, otherwise an exception will be thrown.
 * Creation date: (1/6/2003 4:21:19 PM)
 * @param newQueue the new right-most input queue
 */
void setRightOutputQueue(XATQueue newQueue);
}
