/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataSource;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import java.util.Iterator;

/**
 * Insert the type's description here. Creation date: (1/6/2003 3:07:43 PM)
 * 
 * @author: Brad Pielech
 */
public abstract class XATMultiQueueStreamOperatorImp extends
		XATStreamOperatorImp implements XATMultiQueueStreamOperator {

	protected int numOutputColumns;
	protected int numOutputColsPerStream[];

	/**
	 * XATMultiSourceStreamOperatorImp constructor comment.
	 * 
	 * @param init
	 *            edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
	 */
	public XATMultiQueueStreamOperatorImp(XATNode init) {
		super(init);
	}

	/**
	 * copy constructor
	 * 
	 * @param oldop
	 * @param newnode
	 */
	public XATMultiQueueStreamOperatorImp(XATMultiQueueStreamOperatorImp oldop,
			XATNode newnode) {
		super(oldop, newnode);
	}

	public XATMultiQueueStreamOperatorImp(XATMultiQueueStreamOperator oldop,
			XATNode newnode) {
		super((XATMultiQueueStreamOperatorImp) oldop, newnode);
	}

	/**
	 * Adds another input queue to this operator. This will be added to the
	 * right of the others
	 * 
	 * @param queue
	 *            edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
	 */
	public synchronized void addInputQueue(XATQueue queue) {
		inputQueue.add(queue);
		queue.register(this);

		queue.setQueueNumber(inputQueue.indexOf(queue));
	}

	/**
	 * Adds another output queue to this operator. This will be added to the
	 * right of the others
	 * 
	 * @param queue
	 *            edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
	 */
	public synchronized void addOutputQueue(XATQueue queue) {
		outputQueue.add(queue);
	}

	/**
	 * Get a specific source of a certain number.
	 * 
	 * @return XATDataSource of that position.
	 * 
	 */
	public synchronized XATDataSource getDataSourceAt(int position) {
		return getNthInputQueue(position);
	}

	/**
	 * Return an array of XATDataSource objects that represent this operator's
	 * input queues
	 */
	public synchronized edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataSource[] getDataSources() {
		return (XATDataSource[]) inputQueue.toArray();
	}

	/**
	 * get the left most input queue this has the effect of calling
	 * super.getInputQueue() Creation date: (1/6/2003 4:27:25 PM)
	 * 
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
	 */
	public synchronized XATQueue getLeftInputQueue() {
		return (XATQueue) inputQueue.get(0);
	}

	/**
	 * get the left most output queue this has the effect of calling
	 * super.getOutputQueue() Creation date: (1/10/2003 2:12:56 PM)
	 * 
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
	 */
	public synchronized XATQueue getLeftOutputQueue() {
		return (XATQueue) outputQueue.get(0);
	}

	/**
	 * Get the Nth input queue. This is useful when there are more than 2 input
	 * queues Creation date: (1/6/2003 4:27:25 PM)
	 * 
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
	 * @throws ArrayIndexOutOfBoundsException
	 *             if the Nth queue does not exist
	 */
	public synchronized XATQueue getNthInputQueue(int n)
			throws java.lang.ArrayIndexOutOfBoundsException {
		return (XATQueue) inputQueue.get(n);
	}

	/**
	 * Get the Nth output queue. This is useful when there are more than 2
	 * output queues Creation date: (1/10/2003 2:12:56 PM)
	 * 
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
	 * @throws ArrayIndexOutOfBoundsException
	 *             if the Nth queue does not exist
	 */
	public synchronized XATQueue getNthOutputQueue(int n)
			throws ArrayIndexOutOfBoundsException {
		return (XATQueue) outputQueue.get(n);
	}

	/**
	 * get the right most input queue. This gets the right-most input queue. If
	 * an operator has more than 2 inputs, this will get the farthest right
	 * input. This has the same effect as calling getNthInput(2); Creation date:
	 * (1/6/2003 4:27:25 PM)
	 * 
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
	 */
	public synchronized XATQueue getRightInputQueue() {
		return (XATQueue) inputQueue.get(inputQueue.size() - 1);
	}

	/**
	 * get the right most output queue. This gets the right-most output queue.
	 * If an operator has more than 2 outputs, this will get the farthest right
	 * output. This has the same effect as calling getNthOutput(2); Creation
	 * date: (1/10/2003 2:12:56 PM)
	 * 
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
	 */
	public synchronized XATQueue getRightOutputQueue() {
		return (XATQueue) outputQueue.get(outputQueue.size() - 1);
	}

	/**
	 * Get a specific source of a certain number.
	 * 
	 * @return XATTable of that position.
	 * @deprecated use XATMultiSourceOperator.getDataSourceAt(int position)
	 */
	public synchronized edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATTable getSourceAt(
			int position) {
		return null;
	}

	/**
	 * Return a vector of XATTable objects got from its child nodes. You can get
	 * those XATTable objects from children nodes.
	 * 
	 * @return Vector of XATTable.
	 * @deprecated use XATMultiSourceOperator.getDataSources()
	 */
	public synchronized edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATTable[] getSources() {
		return null;
	}

	/**
	 * Insert the method's description here. Creation date: (4/4/2003 10:29:16
	 * PM)
	 * 
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple
	 * @param inputs
	 *            edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[]
	 */
	public synchronized XATTuple projectToDesiredOutput(XATTuple[] inputs) {

		XATDataValue[] resultData = new XATDataValue[getNode().getSchema()
				.size()];

		int[][] outputPositions = getOutputPositions();

		int fillPos = 0;

		for (int i = 0; i < outputPositions.length; i++) {

			((XATTupleImp) inputs[i]).fillDataInListTo(resultData,
					outputPositions[i], fillPos);

			// free the space
			inputs[i] = null;

			fillPos = fillPos + outputPositions[i].length;
		}

		XATTuple newTuple = new XATTupleImp(resultData);

		return newTuple;

	}

	/**
	 * Runs the actual operator.
	 * 
	 * Unlike in Rainbow, the Stream operators have the implementation tied to
	 * the operator because the StreamOperators represent physical operators
	 * instead of Logic. Creation date: (2/3/03 1:49:06 PM)
	 */
	public abstract boolean run() throws java.lang.Exception;

	/**
	 * set the left most input queue this has the effect of calling
	 * super.setInputQueue() Creation date: (1/6/2003 4:21:19 PM)\
	 * 
	 * @param newQueue
	 *            the new left-most input queue
	 */
	public synchronized void setLeftInputQueue(XATQueue newQueue) {
		/*
		 * throws XATQueueException { try { newQueue.register(this);
		 * if(inputQueue.size() == 0){ inputQueue.add(newQueue); } else{
		 * //inputQueue.set(0,newQueue); inputQueue.add(0, newQueue); }
		 * 
		 * } catch (XATQueueException e) { throw e; }
		 */
		if (inputQueue.size() == 0) {
			inputQueue.add(newQueue);
			newQueue.setQueueNumber(inputQueue.indexOf(newQueue));
			newQueue.register(this);
		} else {
			setNthInputQueue(newQueue, 0);
		}
	}

	/**
	 * set the left most output queue The operator must be the owner of the
	 * queue, otherwise an exception will be thrown. this has the effect of
	 * calling super.setOutputQueue() Creation date: (1/20/03 4:43:17 PM)
	 * 
	 * @param newQueue
	 *            the new left-most input queue
	 */
	public synchronized void setLeftOutputQueue(XATQueue newQueue) {
		/*
		 * throws XATQueueException { if (newQueue.isOwner(this)) { if
		 * (outputQueue.size() == 0) { outputQueue.add(newQueue); } else {
		 * outputQueue.set(0, newQueue); } } else { throw new XATQueueException(
		 * "Cannot set this to be the output queue because the operator does not own the queue"
		 * ); }
		 */
		setNthOuputQueue(newQueue, 0);

	}

	/**
	 * Set the Nth input queue. This is useful when there are more than 2 input
	 * queues This first register's the Operator with the queue Creation date:
	 * (1/20/03 4:43:17 PM)
	 * 
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
	 * @param newQueue
	 *            the new Nth input queue
	 * @param int N the inputQueue to set
	 */
	public synchronized void setNthInputQueue(XATQueue newQueue, int n) {
		try {
			newQueue.register(this);
			int initialSize = inputQueue.size();
			if (initialSize > n) {
				inputQueue.set(n, newQueue);
				newQueue.setQueueNumber(inputQueue.indexOf(newQueue));
			} else {
				// pad all the queues before it
				for (int i = 0; i < n - initialSize; i++) {
					inputQueue.add(null);
				}
				inputQueue.add(newQueue);
				newQueue.setQueueNumber(inputQueue.indexOf(newQueue));
				// inputQueue.set(n, newQueue);
			}

		} catch (XATQueueException e) {
			throw e;
		}
	}

	/**
	 * Set the Nth output queue. This is useful when there are more than 2
	 * output queues The operator must be the owner of the queue, otherwise an
	 * exception will be thrown. Creation date: (1/20/03 4:43:17 PM)
	 * 
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
	 * @param newQueue
	 *            the new Nth output queue
	 * @param int N the outputQueue to set
	 */
	public synchronized void setNthOuputQueue(XATQueue newQueue, int n) {
		if (outputQueue.size() > n) {
			outputQueue.set(n, newQueue);
		} else {
			outputQueue.add(newQueue);
		}

	}

	/**
	 * set the right most input queue This first register's the Operator with
	 * the queue Creation date: (1/20/03 4:43:17 PM)
	 * 
	 * @param newQueue
	 *            the new right-most input queue
	 */
	public synchronized void setRightInputQueue(XATQueue newQueue) {
		if (inputQueue.size() == 0) {
			inputQueue.add(newQueue);
			// this.addInputQueue(newQueue);
			newQueue.setQueueNumber(inputQueue.indexOf(newQueue));
			newQueue.register(this);
		} else
			setNthInputQueue(newQueue, inputQueue.size());
	}

	/**
	 * set the right most output queue this has the effect of calling
	 * super.setOutputQueue() The operator must be the owner of the queue,
	 * otherwise an exception will be thrown. Creation date: (1/20/03 4:43:17
	 * PM)
	 * 
	 * @param newQueue
	 *            the new right-most input queue
	 */
	public synchronized void setRightOutputQueue(XATQueue newQueue) {
		setNthOuputQueue(newQueue, outputQueue.size());

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator
	 * #hasState()
	 */
	public boolean hasState() {
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator
	 * #init()
	 */
	public void init() {
		Iterator it = this.getInputQueues();
		while (it.hasNext()) {
			XATQueueSchema inSchema = ((XATQueue) it.next()).getSchema();
			if (inSchema == null || inSchema.getLength() == 0) {
				System.out.println("NO input Schema!");
				return;
			} else {
				System.out.println("ALL Input Schema is GOOD!");
			}
		}

		XATQueueSchema outSchema = this.getOutputQueue().getSchema();
		if (outSchema == null || outSchema.getLength() == 0) {
			String schemaName = new String();
			outSchema = new XATQueueSchema();
			it = this.getInputQueues();
			while (it.hasNext()) {
				XATQueueSchema inSchema = ((XATQueue) it.next()).getSchema();
				schemaName += "#" + inSchema.getName();

				Iterator j = inSchema.getColNames().iterator();
				while (j.hasNext()) {
					outSchema.addColumn((String) j.next());
				}
			}

			outSchema.setName(schemaName);
			// System.out.println("output schema: "+ outSchema.toString());
			this.getOutputQueue().setSchema(outSchema);
		}
		computeOutputPositions();

		this.numOutputColsPerStream = new int[outputPositions.length];

		this.numOutputColumns = 0;
		for (int i = 0; i < outputPositions.length; i++) {
			int numCols = outputPositions[i].length;
			numOutputColsPerStream[i] = numCols;
			numOutputColumns += numCols;
		}
		this.setHasInit(true);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator
	 * #init()
	 */
	public void init(int outputQueueId) {
		Iterator it = this.getInputQueues();
		while (it.hasNext()) {
			XATQueueSchema inSchema = ((XATQueue) it.next()).getSchema();
			if (inSchema == null || inSchema.getLength() == 0) {
				return;
			}
		}

		XATQueueSchema outSchema = this.getOutputQueue(outputQueueId)
				.getSchema();
		if (outSchema == null || outSchema.getLength() == 0) {
			String schemaName = new String();
			outSchema = new XATQueueSchema();
			it = this.getInputQueues();
			while (it.hasNext()) {
				XATQueueSchema inSchema = ((XATQueue) it.next()).getSchema();
				schemaName += "#" + inSchema.getName();

				Iterator j = inSchema.getColNames().iterator();
				while (j.hasNext()) {
					outSchema.addColumn((String) j.next());
				}
			}

			outSchema.setName(schemaName);
			this.getOutputQueue(outputQueueId).setSchema(outSchema);
		}
		computeOutputPositions();

		this.numOutputColsPerStream = new int[outputPositions.length];
		this.numOutputColumns = 0;
		for (int i = 0; i < outputPositions.length; i++) {
			int numCols = outputPositions[i].length;
			numOutputColsPerStream[i] = numCols;
			numOutputColumns += numCols;
		}
		this.setHasInit(true);
	}

	/**
	 * generate result tuples according to the schema of the output queue
	 * 
	 * @param tuples
	 *            input tuples, tuples[i] is the tuple from the queue whose
	 *            position is i
	 * @return
	 */
	protected XATTuple generateResultTuple(XATTuple[] tuples) {
		XATDataValue[] resultTupleContent = new XATDataValueImp[numOutputColumns];
		long[] resultTupleTime = new long[numOutputColumns];
		long maxTimestamp = Long.MIN_VALUE;
		long minTimestamp = Long.MAX_VALUE;

		int numTuples = tuples.length;
		int startIndex = 0;
		for (int i = 0; i < numTuples; i++) {
			XATTuple tuple = tuples[i];
			XATDataValue[] tupleContent = tuple.getArrayData();
			long[] timestamp = tuple.getTimeStamp();
			long maxTime = tuple.getMaxTimeStamp();
			long minTime = tuple.getMinTimeStamp();

			int k = 0;
			for (; k < numOutputColsPerStream[i]; k++) {
				int pos = outputPositions[i][k];
				resultTupleContent[startIndex + k] = tupleContent[pos];
				resultTupleTime[startIndex + k] = timestamp[pos];
			}
			maxTimestamp = (maxTimestamp > maxTime ? maxTimestamp : maxTime);
			minTimestamp = (minTimestamp < minTime ? minTimestamp : minTime);
			startIndex += k;
		}

		XATTuple resultTuple = new XATTupleImp(resultTupleContent);
		resultTuple.setTimeStamp(resultTupleTime);
		resultTuple.setMaxTimeStamp(maxTimestamp);
		resultTuple.setMinTimeStamp(minTimestamp);
		return resultTuple;
	}
}
