/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
 * Insert the type's description here.
 * Creation date: (1/6/2003 3:07:43 PM)
 * @author: Brad Pielech
 */
public abstract class XATSingleQueueStreamOperatorImp 
	extends XATStreamOperatorImp implements XATSingleQueueStreamOperator {

	protected int numOutputColumns;
/**
 * The input queue of an operator is a pointer to the output queue
 * of the <emph>left</emph> child.  When an operator only has 1 child, this
 * points to the queue of its only child, but it is important to note this has the
 * implied semantic that it points to the left one.
 * @return XATQueue the queue that serves as input to this operator
 */

public XATSingleQueueStreamOperatorImp(XATNode init) {
	super(init);
}
public XATSingleQueueStreamOperatorImp(XATSingleQueueStreamOperatorImp oldop, XATNode newnode) {
	super(oldop, newnode);
}

public XATSingleQueueStreamOperatorImp(XATSingleQueueStreamOperator oldop, XATNode newnode) {
	super((XATSingleQueueStreamOperatorImp)oldop, newnode);
}
/**
 * Insert the method's description here.
 * Creation date: (2/17/2003 1:52:30 PM)
 * @return int[]
 */
public XATTuple projectToDesiredOutput(XATTuple input) {
	
    int[] colPos = getOutputPositions(0);

    XATDataValue[] resultData = new XATDataValue[colPos.length];

    ((XATTupleImp)input).fillDataInListTo(resultData, colPos, 0);

    //free the space for input tuples
    input = null;
    
    XATTuple newTuple = new XATTupleImp(resultData);

    return newTuple;
}
/**
 * Runs the actual operator.
 *
 * Unlike in Rainbow, the Stream operators have the implementation tied to the operator
 * because the StreamOperators represent physical operators instead of Logic.
 * Creation date: (1/22/03 4:18:24 PM)
 */
abstract public boolean run() throws java.lang.Exception;

/**
 * generate result tuples according to the schema of the output queue
 * @param tuple the input tuple
 * @return
 */
protected XATTuple generateResultTuple(XATTuple tuple) {
	XATDataValue[] resultContent = new XATDataValue[numOutputColumns];
	long[] resultTime = new long[numOutputColumns];
	XATDataValue[] tupleContent = tuple.getArrayData();
	long[] tupleTime = tuple.getTimeStamp();
	
	for (int i=0; i<numOutputColumns; i++) {
		resultContent[i] = tupleContent[outputPositions[0][i]];
		resultTime[i] = tupleTime[outputPositions[0][i]];
	}
	XATTuple resultTuple = new XATTupleImp(resultContent);
	resultTuple.setTimeStamp(tupleTime);
	resultTuple.setMinTimeStamp(tuple.getMinTimeStamp());
	resultTuple.setMaxTimeStamp(tuple.getMaxTimeStamp());
	return resultTuple;
}

/* (non-Javadoc)
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#init()
 */
public void init() {
	XATQueueSchema inSchema = this.getInputQueue().getSchema();
	if (inSchema == null || inSchema.getLength() == 0) {
		return;
	}
	XATQueueSchema outSchema = this.getOutputQueue().getSchema();
	if (outSchema == null || outSchema.getLength() == 0) {
		this.getOutputQueue().setSchema(inSchema);
	}
	computeOutputPositions();
	numOutputColumns = outputPositions[0].length;
	setHasInit(true);
}
}
