/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;
import java.lang.reflect.Constructor;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATMemoryQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPersistentQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATSimpleQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataSource;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
 * Insert the type's description here.
 * Creation date: (1/6/2003 3:02:21 PM)
 * @author: Brad Pielech
 */
public abstract class XATStreamOperatorImp extends XATOperatorImp implements XATStreamOperator 
{
    //	indicates the outputqueue id of the child nodes
	//  key:child_opid value:child output queue id.
	//  Added by Song Wang on Oct. 2005 to support multiple queue output.
	protected Hashtable outputQueueIdFromChildren = new Hashtable();
	protected int numberOfOutputQueue = 1; //indicates the total number of output queue

	protected boolean hasInit = false; //indicates whether the operator has been initialized
	protected boolean active;
	protected boolean isNewlyActivated = false;
	protected boolean isNewlyDeactivated = false;
	protected boolean debugFlag = false;
	protected ArrayList inputQueue;
	protected int operatorID = 0;  // useful for parsing initial query plan from xml file and for centralized query plan optimization
	protected int[][] outputPositions;
	protected ArrayList outputQueue;
	protected boolean running;
	protected OperatorState[] states = null;
	
	protected boolean isChanged = false;
	protected boolean isBoxOperator = false;

	//set to true and checked during processing only if isOldOperator == true
	protected boolean isNewBoxRoot = false;

	//set to true and checked during processing only if isOldOperator == true
	protected boolean isOldBoxLeaf = false;
	//added by Mingzhu to store the Name showing up in GUI window
	protected String opName = null;
	
	/* The migrationController should connect the new and old plan together, set migration flags, islastoperator,
	 * and get the current system time and set each old operator's migrationStartTime. 
	 * for new operators and operators that are not involved in migration, the migrationStartTime is useless, and 
	 * by default set to be -1.
	 *
	 */
	//private int migrationStartTime = -1;


	/* If isLastOperator is true, 
	 * that means the migration has started and haven't ended, so special processing need to be done at the last operator
	 * of the old query plan in order to prevent duplicate results.
	 *
	 * When isLastOperator is true, the isOldOperator must be true, and the migrationStartTime must be a positive integer.
	 */
	protected boolean isOldBoxRoot = false;


	/* isOldOperator = true means that migration has started and the operator belongs to "old" query plan. 
	 * the value is false by default.
	 * If isOldOperator = true, this operator should be removed when both states and both inputs does not contain
	 * tuples with isNew set to be true.
	 * when this happens, the operator will stop processing any new tuple and disconnect itself from the
	 * upper node.
	 *
	 * If isOldOperator == false, that means this operator either belongs to the new query plan, or does not involve
	 * in migration at all. In that case, the operator should be acting as usual.
	 */
	protected boolean isOldOperator = false;
	
	/* record the current status of the operator.
	 * this is used for the GUI to display the current status of the operator
	 */
	protected StringBuffer statusBuffer = new StringBuffer();
	
	/*
	 * add for distributed migration sync process.
	 */
	//protected boolean needToSync = false;
	//protected boolean isSyncLeaf = false;
	protected int numOfSyncedInputs = 0;
	boolean isSynced = false;
	boolean isSubPlanRoot = false;
	boolean isSubPlanOperator = false;
	
	//Added by Song Wang for support queue connection in a graph query plan
	public boolean connectQueuesDone = false;
	
/**
 * XATStreamOperatorImp constructor comment.
 * Creates the StreamOperator with 1 input queue.  The output queue
 * is set later
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public XATStreamOperatorImp(XATNode node) {
	super(node);
	inputQueue = new ArrayList();
	outputQueue = new ArrayList();
	//setOutputQueue(new XATMemoryQueueImp(this));
	
	String s = System.getProperty("DATA_MODEL");
	System.out.println("DATA_MODEL"+s);
	if(System.getProperty("DATA_MODEL").endsWith("XATMemoryQueueImp")) {
		setOutputQueue(new XATMemoryQueueImp(this));
	}
	else if (System.getProperty("DATA_MODEL").endsWith("XATPersistentQueueImp")) {
		setOutputQueue(new XATPersistentQueueImp(this));
	}
	else if (System.getProperty("DATA_MODEL").endsWith("XATSimpleQueueImp")) {
		setOutputQueue(new XATSimpleQueueImp(this));
	}
	else {
		System.out.println("!!! Error: Unknow type of DATA_MODEL !!!");
		System.exit(-1);
	} 
	active = false;
	running = false;
	states = null;
}

/**
 * copy constructure
 * copy everything except queues and tuples
 * used by the cloning stage at the biginning of query optimization.
 * @param oldop
 * @param newnode
 */
public XATStreamOperatorImp(XATStreamOperatorImp oldop, XATNode newnode) {
		this(newnode);
		
		//XATStreamOperatorImp newop = new XATStreamOperatorImp(newnode);
		this.active = oldop.active;
		this.columnsFiltered = oldop.columnsFiltered;
		this.debugFlag = oldop.debugFlag;
		this.operatorID = oldop.operatorID;
		this.outputColumnName = oldop.outputColumnName;
		//this.outputPositions = oldop.outputPositions;
		this.properties = oldop.properties;
		//this.statistics = (java.util.Properties)oldop.statistics.clone();
		this.statistics = oldop.statistics;
		if (oldop.states != null) {
			this.states = new OperatorState[oldop.states.length];
			
			//this.states = new OperatorState[oldop.states.length];		
			Constructor construct;
			Class[] c = new Class[2];
			c[1] = XATStreamOperator.class;
			try {
				for (int i = 0; i < oldop.states.length; i++) {
					c[0] = oldop.states[i].getClass();
					construct = oldop.states[i].getClass().getConstructor(c);
					this.states[i] = (OperatorState) construct.newInstance(new Object[] { oldop.states[i], this });
					//this.state[i] = oldop.state[i].makeNewCopy();
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		this.stats = oldop.stats;
		this.statsPresent = oldop.statsPresent;
}

public XATStreamOperatorImp(XATStreamOperator oldop, XATNode newnode) {
		this((XATStreamOperatorImp)oldop, newnode);
}

public synchronized void addInputQueue(XATQueue newInputQueue) throws XATQueueException {
	try {
		newInputQueue.register(this);
		inputQueue.add(newInputQueue);	
		newInputQueue.setQueueNumber(inputQueue.indexOf(newInputQueue));
		
	} catch (XATQueueException e) {
		throw e;
	}	
}

public synchronized void addInputQueue(XATQueue newInputQueue, int position) throws XATQueueException {
	try {
		newInputQueue.register(this);
		inputQueue.add(position, newInputQueue);	
		for(int i=position; i<inputQueue.size(); i++) {
			((XATQueue)inputQueue.get(i)).setQueueNumber(i);
			//newInputQueue.setQueueNumber(inputQueue.indexOf(newInputQueue));
		}
		
	} catch (XATQueueException e) {
		throw e;
	}	
}
//Added by Song Wang for support multiple outputQueues in one operator
public synchronized void addOutputQueue(XATQueue newOutputQueue, int position) throws XATQueueException {
	try {
		outputQueue.add(position, newOutputQueue);	
		for(int i=position; i<outputQueue.size(); i++) {
			((XATQueue)outputQueue.get(i)).setQueueNumber(i);
		}
		
	} catch (XATQueueException e) {
		throw e;
	}	
}
//Added by Song Wang for support multiple outputQueues in one operator
public synchronized void addOutputQueue(int position) throws XATQueueException {
	try {
		XATQueue newQueue=null;
		if(System.getProperty("DATA_MODEL").endsWith("XATMemoryQueueImp")) {
			newQueue = new XATMemoryQueueImp(this);
		}
		else if (System.getProperty("DATA_MODEL").endsWith("XATPersistentQueueImp")) {
			newQueue = new XATPersistentQueueImp(this);
		}
		else if (System.getProperty("DATA_MODEL").endsWith("XATSimpleQueueImp")) {
			newQueue = new XATSimpleQueueImp(this);
		}
		else {
			System.out.println("!!! Error: Unknow type of DATA_MODEL !!!");
			System.exit(-1);
		} 
		outputQueue.add(position, newQueue);	
		for(int i=position; i<outputQueue.size(); i++) {
			((XATQueue)outputQueue.get(i)).setQueueNumber(i);
		}
		
	} catch (XATQueueException e) {
		throw e;
	}	
}


//Added by Mingzhu Wei for support multiple outputQueues in one operator
//Note here we default use simpleQueue
public synchronized void addNewOutputQueue() throws XATQueueException {
	try {
		XATQueue newQueue=null;
		newQueue = new XATSimpleQueueImp(this);
		int i = outputQueue.size();
		//System.out.println("cur outQueue Size: " + i);
		outputQueue.add(i, newQueue);	
		((XATQueue)outputQueue.get(i)).setQueueNumber(i);
		numberOfOutputQueue++;
		
		
	} catch (XATQueueException e) {
		throw e;
	}	
}

/**
 * Clear tuples in input and output queues.
 */
public synchronized void clearAllTuples() {
	
	for(int i=0; i< inputQueue.size(); i++){
		XATQueue queue = (XATQueue)inputQueue.get(i);
		queue.removeAllTuples();
	}

	for(int i=0; i< outputQueue.size(); i++){
		XATQueue queue = (XATQueue)outputQueue.get(i);
		queue.removeAllTuples();
	}
}
	
	/**
	 * compute the attribute mapping between input queue(s) schema and output queue schema
	 */
	public void computeOutputPositions() {
		DagNode node = (DagNode)getNode(); 
		Vector outColNames = getOutputQueue().getSchema().getColNames();
		int outputColSize = outColNames.size();
		
		//added to pintout all the output COlNames
		//System.out.println("\n \t In XATSreamOperatorImp printout outputCol Name");
		for (Iterator itt = outColNames.iterator(); itt.hasNext();)
			
		{
			Object oCo1 = itt.next();
			//System.out.println("outputCOl Name:" + oCo1.toString());
			
			
		}
		int outputColPos = 0;

		Iterator inputs = getInputQueues();
		//store all the input columns
		int j=0;
		while(inputs.hasNext()) {
			Vector outputPosVec = new Vector();
			//Vector inputColNames = ((XATQueue)inputs.next()).getSchema().getColNames(); 
			XATQueue tempq = (XATQueue)inputs.next();
			if (tempq.getSchema()==null) continue;
			Vector inputColNames = tempq.getSchema().getColNames();
			for (Iterator i1 = outColNames.iterator(); i1.hasNext();){
				Object outputCol = i1.next();
				for (int k = 0; k<inputColNames.size(); k++){
					Object inputCol = inputColNames.get(k);
					//System.out.println("input column: "+inputCol.toString());
					//System.out.println("output column: "+outputCol.toString());
					if (outputCol.equals(inputCol)){
						outputPosVec.add(new Integer(k));
						break;
					}
				}
			}
			//convert vector to int[]
			int[] outputPosArray = new int[outputPosVec.size()];
			for (int count = 0; count < outputPosVec.size(); count++){
				outputPosArray[count] = ((Integer)outputPosVec.get(count)).intValue();
			}
			((XATStreamOperatorImp) node.getOperator()).setOutputPositions(j, outputPosArray);
			j++;
		}
	}
	
//	public boolean equals(Object op){
//		if(op instanceof XATStreamOperator) {
//			if(this.getProperty("OBJECT_ID").equals(((XATStreamOperator)op).getProperty("OBJECT_ID"))){
//				return true;
//			}
//		}
//		return false;
//	}
	
	public boolean equals(XATStreamOperator op){
		
		if(this.getProperty("OBJECT_ID").equals(op.getProperty("OBJECT_ID"))){
			return true;
		}
		return false;
	}

/**
 * The input queue of an operator is a pointer to the output queue
 * of the <emph>left</emph> child.  When an operator only has 1 child, this
 * points to the queue of its only child, but it is important to note this has the
 * implied semantic that it points to the left one.
 * @return XATQueue the queue that serves as input to this operator
 */
public synchronized edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue getInputQueue() 
{
	return (XATQueue)inputQueue.get(0); 
}

public boolean getDebug() 
{
	return this.debugFlag;
}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#getInputQueue(int)
	 */
	public XATQueue getInputQueue(int queueNum) 
	{
//		for(Iterator i = this.getInputQueues(); i.hasNext();){		
//			XATQueue next = (XATQueue) i.next();
//			if(next.getQueueNumber() == queueNum)
//				return next;
//		}
		
		if(this.inputQueue == null)
			return null;
		if(queueNum < this.inputQueue.size())
			return (XATQueue)this.inputQueue.get(queueNum);
		return null;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#getInputQueue(java.lang.String)
	 */
	public XATQueue getInputQueue(String queueName) {
		for(Iterator i = this.getInputQueues(); i.hasNext();){
			
			XATQueue next = (XATQueue) i.next();
			
			if(next.getQueueName().equals(queueName))
				return next;
			
		}
		
		return null;
	}
/**
 * Returns the number of inputqueues this stream operator has
 * @return int
 */
	public synchronized int getInputQueueCount() {
		return inputQueue.size(); 
	}
	public synchronized Hashtable getOutputQueueIdFromChildren() {
		return outputQueueIdFromChildren; 
	}
	public synchronized ArrayList getOutputQueueIdFromChild(int opid) {
		//int ret = 0;
		//if (outputQueueIdFromChildren.contains(new Integer(opid))) {
		ArrayList ret = (ArrayList)outputQueueIdFromChildren.get(new Integer(opid));
		//}
		return ret;
	}
	public synchronized void addOutputQueueIdFromChildren(int opid, int id) {
		//added to support multiple output queues from one identical child.
		//add an ArrayList instead.
		if (outputQueueIdFromChildren.get(new Integer(opid))!=null) {
			ArrayList tempList = (ArrayList)outputQueueIdFromChildren.get(new Integer(opid));
			tempList.add(new Integer(id));
			outputQueueIdFromChildren.put(new Integer(opid), tempList);				
		} else {
			ArrayList tempList = new ArrayList();
			tempList.add(new Integer(id));
			outputQueueIdFromChildren.put(new Integer(opid), tempList);				
		}
		return;
	}

    public void clearOutputQueueIdsFromChildren()
    {
        outputQueueIdFromChildren.clear();
    }

    
    
	public synchronized int getNumberOfOutputQueue() {
		return numberOfOutputQueue; 
	}
	public synchronized void setNumberOfOutputQueue(int value) {
		this.numberOfOutputQueue = value;
		return;
	}

public XATQueue getNthInputQueue(int n) {
	try {
		if(n >= this.inputQueue.size())
			throw new Exception("inquiring queue position out of input queue boundary");
		else
			return (XATQueue)inputQueue.get(n);
	} catch (Exception e) {
		e.printStackTrace();
		return null;
	}
}

public int getInputQueuePosition(XATQueue queue) {
	return inputQueue.indexOf(queue);
}

/**
 * Returns an iterator of all of the input queues of this operator
 * This is mostly a convience method where, if you dont know how many
 * input queues an operator has, but you want access to all of them. 
 * @return java.util.Iterator
 */
public synchronized java.util.Iterator getInputQueues() {
	return inputQueue.iterator();
}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject#getLightWeightQueryObject()
	 */
	public LightWeightQueryObject getLightWeightQueryObject() {
		return new LightWeightQueryObject("Op " + this.getProperty("OBJECT_ID"),Integer.parseInt(this.getProperty("OBJECT_ID")),this.getClass().getName(),this.isActive());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#getNthState(int)
	 */
	public OperatorState getNthState(int n) {
		return states[n];
	}
	
	public int getObjectID(){
		return Integer.parseInt(this.getProperty("OBJECT_ID"));
	}

public int getOperatorID(){
	return operatorID;
}
/**
* reimplemented the XATOperator's getOutputDataSource.
* Note that this casts the Queue to an XATDataSource which can be
* downcast to a queue 
*/
public synchronized XATDataSource getOutputDataSource() {
    return (XATQueue)outputQueue.get(0);
}
 /**
     * Get all output data sources for this XATOperator
     * @return Iterator of data sources.      
     */
public synchronized java.util.Iterator getOutputDataSources() {
	return getOutputQueues();
}
/**
 * the outputpositions records for each input queue, which columns are going to be output
 * outputPositions[i] specifies for input queue i, which columns are going to be output
 * OutputPosisions[i] stores an array, each integer in this array specifies the position of a column that is going to be output
 * e.g., outputPositions[2] = [3, 5]
 * it means for input queue 2, its 3 and 5th columns are in the output queue
 * Creation date: (4/4/2003 5:17:53 PM)
 * @return int[][]
 */
public synchronized int[][] getOutputPositions() {  

	if (outputPositions == null){ 
	  DagNode node = (DagNode)getNode(); 
	
    //Set schema = getNode().getSchema();
    Vector schema = getOutputQueue().getSchema().getColNames();
    
    int outputColPos = 0;

    //XATNode[] children = node.getChildren();

    //store all the input columns

    for (int j = 0; j < this.inputQueue.size(); j++) {

        Vector outputPosVec = new Vector();

        Vector childColNames =
             this.getInputQueue(j).getSchema().getColNames();

        for (Iterator i1 = schema.iterator(); i1.hasNext();){

	        Object outputCol = i1.next();

	        for (int k = 0; k<childColNames.size(); k++){
		        Object inputCol = childColNames.get(k);
		        if (outputCol.equals(inputCol)){
			        outputPosVec.add(new Integer(k));
			        break;
		        }
	        }
        }

        //convert vector to int[]
        int[] outputPosArray = new int[outputPosVec.size()];
        
        for (int count = 0; count < outputPosVec.size(); count++){
	        outputPosArray[count] = ((Integer)outputPosVec.get(count)).intValue();
        }

        ((XATStreamOperatorImp) node.getOperator()).setOutputPositions(j, outputPosArray);

    } //end of for j
	}//end of if
    return outputPositions;
}
/**
 * get the output positions for the ith input queue
 * Creation date: (4/4/2003 5:53:05 PM)
 * @return int[]
 * @param i int
 */  
public synchronized int[] getOutputPositions(int i) {   
	return getOutputPositions()[i];
}
/**
 * Each operator has only 1 queue in which it places its ouput.
 * @return XATQueue the queue that stores the output of this operator
 */
public synchronized edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue getOutputQueue() {

	if(outputQueue == null) 
		return null;
	return (XATQueue)outputQueue.get(0);
}

public synchronized XATQueue getOutputQueue(int position) {
	if(position < outputQueue.size())
		return (XATQueue)outputQueue.get(position);
	else
		return null;
}
/**
 * Returns the number of outputqueues this stream operator has
 * @return int
 */
public synchronized int getOutputQueueCount() {
	return outputQueue.size(); 
}
/**
 * Returns an iterator of all of the output queues of this operator
 * This is mostly a convience method where, if you dont know how many
 * output queues an operator has, but you want access to all of them. 
 * @return java.util.Iterator
 */
public synchronized java.util.Iterator getOutputQueues() {
	return outputQueue.iterator();
}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#getState()
	 */
	public OperatorState[] getState() {
		return states;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#hasState()
	 */
	public boolean hasState() {
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#init()
	 */
	public abstract void init();

/**
 * Returns whether the XATStreamOperator is active in
 * the query plan. When a query plan is distributed, it is
 * possible that only a partial number of operators in the
 * query plan are active. 
 * 
 */
public boolean isActive(){

	return this.active;	
}
public boolean hasInit() {
	return this.hasInit;
}
public boolean isNewlyActivated() {
	return this.isNewlyActivated;
}

public void setNewlyActivated(boolean is) {
	this.isNewlyActivated = is;
}

public boolean isNewlyDeactivated() {
	return this.isNewlyDeactivated;
}

public void setNewlyDeactivated(boolean is) {
	this.isNewlyDeactivated = is;
}

public synchronized boolean isChanged() {
	return this.isChanged;
}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#isRunning()
	 */
	public boolean isRunning() {
		return running;
	}
	
	public boolean isBoxOperator() {
		return isBoxOperator;
	}
	
	public boolean isOldBoxLeaf() {
		return isOldBoxLeaf;
	}
	
//	public abstract XATStreamOperatorImp makeNewCopy();

public void replaceInputQueue(XATQueue existingQueue, XATQueue newQueue) {
	newQueue.register(this);
	existingQueue.unregister(this);
	int i = inputQueue.indexOf(existingQueue);
	inputQueue.set(i, newQueue);
}

/**
 * Runs the actual operator.
 *
 * Unlike in Rainbow, the Stream operators have the implementation tied to the operator
 * because the StreamOperators represent physical operators instead of Logic.
 * Creation date: (1/22/03 4:18:24 PM)
 */
public abstract boolean run() throws java.lang.Exception;
/**
 * Sets whether or not the operator is active
 * 
 * @param active boolean
 * 
 */
public void setActive(boolean active){
	try{
		while(this.isRunning()) {
			Thread.sleep(500);
		}
	}catch (Exception e) {
		e.printStackTrace();
	}
	this.active = active;	
}
public void setHasInit(boolean hasInit) {
	this.hasInit = hasInit;
}
/**
 * Insert the method's description here.
 * Creation date: (6/26/2003 4:58:44 AM)
 */
public void setAsNewBoxRoot() {
	isNewBoxRoot = true;
}

/**
 * Insert the method's description here.
 * Creation date: (4/6/2003 7:46:53 AM)
 */
public void setAsOldBoxLeaf() {
	isOldBoxLeaf = true;
}

/**
 * Insert the method's description here.
 * Creation date: (4/6/2003 7:46:04 AM)
 */
public void setAsOldBoxRoot() {
	isOldBoxRoot = true;
}

/**
 * Insert the method's description here.
 * Creation date: (4/6/2003 8:29:45 AM)
 */
public void setAsOldOperator() {
	isOldOperator = true;
}



/**
 * set current operator to be an operator belong to a migration box.
 * this flag is set to be true by the run-time optimizer if this operator belongs to a box.
 * and the flag is reset to be false after the migration is done.
 * 
 * @param isboxop
 */
public void setBoxOperator(boolean isboxop) {
	this.isBoxOperator = isboxop;
}

public synchronized void setChanged(boolean ischanged) {
	this.isChanged = ischanged;
}

/**
 * set the debug flag of this operator.
 */
public void setDebug(boolean debug){
	debugFlag = debug;
}
/**
 * This will set the default properties that all
 * XATStreamOperators will have.
 * NUMBER_OF_TUPLES_TO_DEQUEUE is how many tuples an operator should dequeue
 * The granularities are relate to the input and ouput granularities
 */
protected void setDefaultProperties() {
	super.setDefaultProperties();
	addProperty("NUMBER_OF_TUPLES_TO_DEQUEUE","0");
	addProperty("OUTPUT_TUPLE_GRANULARITY", FULL_TUPLE);
	addProperty("INPUT_TUPLE_GRANULARITY", FULL_TUPLE);
	//whether it is invoked by child 
	addProperty("DATA_DRIVEN_INVOKING", "1");
	//whether it is invoked by group_by_end info by descendent
	addProperty("PARTITION_DELIMITATION_INVOKING", "0");
	//whether it is directly schedulable
	addProperty("DIRECTLY_SCHEDULABLE", "0");
	addProperty("INITIAL_GREEDY_PRIORITY", "1");
	//whether we turn on the quick response feature
	addProperty("QUICK_RESPONSE", "0");
	addProperty("NUMBER_OF_POSSIBLE_OUTPUTS", "0");
	addProperty("NUMBER_OF_TUPLES_OUTPUTTED", "0");
	// The property "PERFORMANCE_FACTOR" indicates the performance of current operator. 
	// A simple formula used in our system now is 
	// PERFORMANCE_FACOTR = reduction_factor (later we can add in operator cost).
	// This property is used by optimizer and rewrite rule to make optimization decisions, 
	// push down operator with higher PERFORMANCE_FACTOR.
	addProperty("PERFORMANCE_FACTOR", "0");
	addProperty("SELECTIVITY", "0");
	addProperty("REDUCTION_FACTOR", "0");
	/** 
	  * this is a unique ID assigned to each operator
	  * this will help to keep track of operators from
	  * one experiment to the next.  The idea is that
	  * if a query plan is generated the same way from 
	  * experiment to experiment, then its ID will be the same always.   
 	*/
	addProperty("OBJECT_ID", "-1");	
}
/**
 * This will set the default properties that all
 * XATStreamOperators will have.
 * NUMBER_OF_TUPLES_TO_DEQUEUE: how many tuples the operator should dequeue 
 * 		from its input queues in <emph> total <emph>
 * NUMBER_OF_TUPLES_OUTPUTTED: how many tuples the operator placed into its output buffer
 * 		during its last timeslice.
 * NUMBER_OF_TIMES_RUN: how many times has this operator's run method been called
 * LAST_TIME_RUN: when was the last time the operator was run
 * TIME_TOOK_TO_RUN: how long did the operator take to run last time
 * NUMBER_OF_TUPLES_IN_INPUT_QUEUE: how many tuples are in the operator's input queue's (combined)
 * SELECTIVITY: the ratio of tuples that were dequeued to the number of tuples outputted
 */
protected void setDefaultStatistics() 
{
	super.setDefaultStatistics();
	addStatistic("NUMBER_OF_TUPLES_DEQUEUED_TOTAL");
	addStatistic("NUMBER_OF_TUPLES_DEQUEUED");
	addStatistic("NUMBER_OF_POSSIBLE_OUTPUTS");
	addStatistic("NUMBER_OF_TUPLES_OUTPUTTED");
	addStatistic("NUMBER_OF_TUPLES_OUTPUTTED_TOTAL");
	addStatistic("NUMBER_OF_TIMES_RUN");
	addStatistic("LAST_TIME_RUN");
	addStatistic("TIME_TOOK_TO_RUN");
	addStatistic("TIME_TOOK_TO_RUN_TOTAL");
	addStatistic("NUMBER_OF_TUPLES_IN_INPUT_QUEUES");
	addStatistic("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES");
	addStatistic("NUMBER_OF_TUPLES_IN_STATE");
	addStatistic("NUMBER_OF_TUPLES_IN_EACH_STATE");
	addStatistic("SUM_OF_TUPLES_DEQUEUED_TOTAL");
	addStatistic("SELECTIVITY");
	addStatistic("REDUCTION_FACTOR");
	addStatistic("AVERAGE_TUPLE_PROCESSING_TIME");
	addStatistic("GREEDY_PRIORITY");
	addStatistic("AVERAGE_OUTPUT_RATE");
	addStatistic("TUPLES_IN_LEFT_STATE");
	addStatistic("TUPLES_IN_RIGHT_STATE");
	addStatistic("NUMBER_OF_TUPLES_IN_STREAM_QUEUES");
	addStatistic("THROUGHPUT");
	//addStatistic("OLDEST_TUPLE_IN_QUEUE");
	//can be removed for other operators -- Bin
	//addStatistic("JOIN_PARTITION_NUMBER");
}
/**
 * Sets the input queue of this operator. When the input queue is set,
 * the operator is first registered with the queue. If the operator may not register with the queue,
 * then the XATQueueException is thrown
 * @param newInputQueue edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 * @throws XATQueueException if the operator cannot register with the queue 
 *	(it is the queue owner or has already registered)
 */
public synchronized void setInputQueue(XATQueue newInputQueue) throws XATQueueException 
{
	try 
	{
		newInputQueue.register(this);
		newInputQueue.setQueueNumber(0);
		if(inputQueue.size() == 0)
		{
			inputQueue.add(newInputQueue);	
	    } 
		else
		{
			inputQueue.set(0,newInputQueue);	
	    }
		
	} 
	catch (XATQueueException e) 
	{
		throw e;
	}	
}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#setNthState(edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState, int)
	 */
	public void setNthState(OperatorState state, int n) 
	{
		this.states[n] = state;
	}

/**
 * This method sets the property "NUMBER_OF_POSSIBLE_OUTPUTS", which is the 
 * possible output tuples from current operator from each run.
 * For a single stream operator, this is simply the number of dequeued tuple.
 * For a multiple input stream operator, such as join, special method need to be applied 
 * in order to compute this property.
 * This property will be used by stats gatherer to compute the reduction factor of this operator.
 */
public void setNumberOfPossibleOutputs() 
{
	try{
		this.setProperty("NUMBER_OF_POSSIBLE_OUTPUTS", this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
	}catch(Exception e) {
		System.out.println(e);
	}
}

public void setOperatorID(int id) {
	operatorID = id;
}
/**
* Set the output table
* 
*/
public void setOutput(XATDataSource output) {
    throw new UnsupportedOperationException("setOutput is not supported by XATStreamOperator, please use setOutputQueue"); 
}
/**
 * set the outputpositions for the i th output
 * Creation date: (4/4/2003 5:17:32 PM)
 * @param i int
 * @param ithOutputPositions int[]
 */
public synchronized void setOutputPositions(int i, int[] ithOutputPositions) {
	if (outputPositions == null){
		outputPositions = new int[this.inputQueue.size()][];
	//	System.out.println("setting new output position");
	}  
	outputPositions[i] = ithOutputPositions;
//	System.out.println("setting ith output position as "+ ithOutputPositions.toString());
	}
/**
 * set the whole output positions (for every output)
 * Creation date: (4/4/2003 5:17:53 PM)
 * @param newOutputPositions int[][]
 */
public void setOutputPositions(int[][] newOutputPositions) {
	outputPositions = newOutputPositions;
}
/**
 * sets the output queue of this operator.  The operator must be the
 * owner of the queue, otherwise an exception will be thrown.
 * @param newOutputQueue edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue
 * @throws XATQueueException if this operator is not the owner of the queue
 */
//public void setOutputQueue(XATQueue newOutputQueue){
//    if (newOutputQueue.isOwner(this)) {
//	    if(outputQueue.size() == 0){
//			outputQueue.add(newOutputQueue);	
//	    } else{
//		    //first remove this operator as the owner of the existing queue
//		    ((XATQueue)outputQueue.get(0)).removeOwner(this);
//			outputQueue.set(0,newOutputQueue);	
//	    }           
//    } else {
//        throw new XATQueueException("Cannot set this to be the output queue because the operator does not own the queue");
//    }
//
//}

public void setOutputQueue(XATQueue newOutputQueue){
	outputQueue.clear();
	outputQueue.add(0, newOutputQueue);
	
	if(!newOutputQueue.isOwner(this)) {
		newOutputQueue.setQueueOwner(this);
	}
}

/**
* <pre>
* sets the given property to the new value.  This method is responsible
* for validating the given property.  An exception is thrown if
* the value is not correct.
* Validation should be done in the following manner: 
* First the method should check if the object supports the given property by calling
* isValidPropertyName(String).  If it is not a valid property name, then an exception should
* be thrown.
*
* Once it is known that the object supports the property, the value must be validated.  Each implementor of
* XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
* For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
* has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
* to set a common type, it should be validated by the super class by calling super.setProperty(name, value).  This
* will reduce duplicate code.  
* </pre>
* @param name the name of the property to set
* @param value the new value of the property
* @throws Exception if the given property cannot be found or
* if the value is not valid for the given property.
*/
public void setProperty(String name, String value) throws java.lang.Exception {

    //first see if the XATOperator supports the property name	
    if (isValidPropertyName(name)) {               
        if (name.equals("OUTPUT_TUPLE_GRANULARITY")
            || name.equals("INPUT_TUPLE_GRANULARITY")) {
            try {
                if ((value == XATStreamOperator.FULL_TUPLE)
                    || (value.equals(XATStreamOperator.SUB_TUPLE))) {
                    setNewPropertyValue(name, value); } else {
                    throw new Exception("The property named " + name + " requires a valid string ");
                        }
            } catch (Exception e) {
                throw new Exception("The property named " + name + " requires a valid string ");
                    }
        } else {
            super.setProperty(name, value);
       }
    } else
        throw new Exception(
            "The XATOperator does not support the property named " + name);
            }

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#setRunning(boolean)
	 */
	public void setRunning(boolean running) {
		this.running = running;

	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#setState(edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState[])
	 */
	public void setState(OperatorState[] state) {
		for(int i = 0; i < this.states.length;i++){
			this.states[i] = state[i];
		}
	}
	
	/*
	 * Assign new ID to each state inside this operator, starting from the given startID.
	 * 
	 * return the next available ID. For example, if startID is 2, and there are 2 states inside this operator,
	 * then these two states are assigned new IDs as 2 and 3, and 4 is then returned as the next available stateID.
	 */
	  public int setStateID(int startID) {
		if (states == null) {
			return startID;
		}
	
		for(int i=0; i<states.length; i++) {
			states[i].setStateID(startID);
			startID++;
		}
		return startID;
	  }
	  
//	public void setNeedToSync(boolean tosync) {
//		this.needToSync = tosync;
//	}
//	
//	public boolean getNeedToSync() {
//		return this.needToSync;
//	}
//	
//	public void setSyncLeaf(boolean sleaf) {
//		this.isSyncLeaf = sleaf;
//	}
//	
//	public boolean getIsSyncLeaf() {
//		return this.isSyncLeaf;
//	}
	  
	/**
	 * Make a new copy of the current operator. This is a deep copy, not bit by bit copy as the default clone in java.
	 * make exact copy of every fields except queues. 
	 * @return
	 */
//	public abstract XATStreamOperator makeNewCopy(XATNode newnode); 
//	{
		
//		XATStreamOperatorImp newop = new XATStreamOperatorImp(newnode);
//		newop.active = this.active;
//		newop.columnsFiltered = this.columnsFiltered;
//		newop.debugFlag = this.debugFlag;
//		newop.operatorID = this.operatorID;
//		newop.outputColumnName = this.outputColumnName;
//		newop.outputPositions = this.outputPositions;
//		
//		newop.state = new OperatorState[this.state.length];
//		for(int i=0; i<state.length; i++) {
//			newop.state[i] = this.state[i].makeNewCopy(newop);
//		}
//		
//		newop.stats = this.stats;
//		newop.statsPresent = this.statsPresent;
//	
//		return newop;
//	}
	
	public String toShortString(){
		return "Op " + this.getObjectID();
		//return "Op " + this.getOperatorID();
	}
/**
 * Insert the method's description here.
 * Creation date: (1/21/03 10:44:48 AM)
 * @return java.lang.String
 */
public String toString() {

	// added temporarily for vldb 2005 demo
	//return this.toShortString();
	
	String rs = new String();

	String sName = this.getClass().getName();
	//find the last dot.
	int index = 0;
	for(index = sName.length()-1; index >=0; index--)
		if(sName.charAt(index) == '.')
			break;
	rs = sName.substring(index + 1) + " ";
	rs = rs + this.getOperatorID();
	String id = this.getProperty("OBJECT_ID");
	if(id == "-1"){
		rs = rs + hashCode(); 
	} else{
		rs = rs + id;
	}
	
	if(this.isActive()) rs += " ACTIVE ";
	else rs+= " INACTIVE ";
	
	if(this.isSubPlanOperator) rs += " subplan op";
	
    return rs;
}

/**
 * append new status of the operator to the status buffer
 * @param newStatus
 */
public void appendNewStatus(String newStatus) {
	this.statusBuffer.append(newStatus+"\n");
}

/* (non-Javadoc)
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#clearStatus()
 */
public void clearStatus() {
	statusBuffer = new StringBuffer();
}

/* (non-Javadoc)
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#getCurrentStatus()
 */
public String getCurrentStatus() {
	return statusBuffer.toString();
}

public int getStateSize() {
	int totalNumTuples = 0;
	if (states == null) {
		return 0;
	}
	for (int i=0; i<states.length; i++) {
		totalNumTuples += states[i].getStateSize();
	}
	return totalNumTuples;
}

/**
 * this operator receives another sync tuple
 * sync tuple is an tuple with maxtimestamp set to be -1.
 * it is used during distributed migration process to synchronize 
 * the execution of changed operators.
 */
public void processSyncTuple() {
	this.numOfSyncedInputs++;
	
	if(this.numOfSyncedInputs == this.getInputQueueCount()) {
		// if not subplan root, propagate the sync tuple
		if (!this.isSubPlanRoot) {
			XATTuple newsynctuple = new XATTupleImp();
			newsynctuple.setMaxTimeStamp(-1);

			Iterator it = this.getOutputQueues();
			while (it.hasNext()) {
				((XATQueue) it.next()).enqueue(newsynctuple, this);
			}
		} else {
			this.isSynced = true;
		}
		
		this.numOfSyncedInputs = 0;
		
		// deactivate this operator right away
		if(isSubPlanOperator)
			this.active = false;
	}
}

public void setIsSubPlanOperator(boolean s) {
	this.isSubPlanOperator = s;
}

public boolean getIsSubPlanOperator() {
	return this.isSubPlanOperator;
}

public void setIsSubPlanRootOperator(boolean s) {
	this.isSubPlanRoot = s;
}

public void setIsSynced(boolean s) {
	this.isSynced = s;
}

public boolean getIsSynced() {
	return this.isSynced;
}

public void removeInputQueue(int position) {
	if(position < this.inputQueue.size())
		this.inputQueue.remove(position);
}

public void removeInputQueue(XATQueue queue) {
	this.inputQueue.remove(queue);
}

public void removeOutputQueue(XATQueue queue) {
	this.outputQueue.remove(queue);
	if(queue != null)
		queue.removeOwner(this);
}

public void removeOutputQueue(int position) {
	if(position < this.outputQueue.size())
		this.outputQueue.remove(position);
}


// Added by Mingzhu Wei @ June 2007
//set up the name for recognizing it in the GUI window
public void setOPName(String name)
{
	opName = name;
}

//Added by Mingzhu Wei @June 2007
public String getOPName()
{
	return opName;
}
/*
 * check if this operator has any unmatched states.
 * return: TRUE if this operator has unmatched states, FALSE otherwise.
 */
//public boolean hasUnmatchedState() {
//	if (states == null) {
//		return false;
//	}
//	
//	for(int i=0; i<states.length; i++) {
//		if(!states[i].isMatched())
//			return true;
//	}
//	return false;
//}

	
/*
 * Assign new ID to each state inside this operator, starting from the given startID.
 * 
 * return the next available ID. For example, if startID is 2, and there are 2 states inside this operator,
 * then these two states are assigned new IDs as 2 and 3, and 4 is then returned as the next available stateID.
 */
//public int setStateID(int startID) {
//	if (states == null) {
//		return startID;
//	}
//	
//	for(int i=0; i<states.length; i++) {
//		states[i].setStateID(startID);
//		startID++;
//	}
//	return startID;
//}


}