/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Aug 24, 2004
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WEquiJoinState;
import java.util.ArrayList;

/**
 * @author lisading
 *
 */
public class BinaryWSingleEquiJoinOperatorImp extends
		XATMultiQueueWindowStreamOperatorImp implements
		BinaryWSingleEquiJoinOperator 
{

	int LEFT = 0;
	int RIGHT = 1;

	WEquiJoinState leftState;
	WEquiJoinState rightState;
	
	int leftJoinAttrPos;
	int rightJoinAttrPos;

	int invalidateThreshold;
	
	long leftLatestProcessTupleTime;
	long rightLatestProcessTupleTime;

	int numNewTuplesSinceLastInvalidate = 0;

	int numOutputColumn;
	int numLeftColumn;
	int numRightColumn;

	public BinaryWSingleEquiJoinOperatorImp(BinaryWSingleEquiJoinOperatorImp oldop, XATNode newnode) 
	{
		super(oldop, newnode);
		this.windowSize = oldop.windowSize;
		this.leftJoinAttrPos = oldop.getLeftJoinAttrPos();
		this.rightJoinAttrPos = oldop.getRightJoinAttrPos();
		this.invalidateThreshold = oldop.getInvalidateThreshold();
	}
	
	/**
	 * Insert the method's description here.
	 * Creation date: (2/17/2003 4:48:47 PM)
	 */
	public BinaryWSingleEquiJoinOperatorImp(XATNode node) 
	{
		super(node);
	}

	public boolean run() throws Exception 
	{
		//System.out.println(this.toString());
		if(this.getOutputQueue().getSchema() == null)
		{
			init();
			if ((this.getLeftInputQueue().getSchema() == null) 
					|| (this.getRightInputQueue().getSchema() == null))
			{
				return false;
			}			
		}
		
		
		this.numOutputColumn = this.getOutputQueue().getSchema().getLength();
		this.numLeftColumn = this.getLeftInputQueue().getSchema().getLength();
		this.numRightColumn = this.getRightInputQueue().getSchema().getLength();

		XATTuple newObj = null;
		int indication = LEFT;
		int tupleCount = 0;
	
		//int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		int howMany  = 500; //rimma temp "hack"
		System.out.println("Will process... [" + howMany+"] tuples");
		
		int currIteration = 0;
		while (true) 
		{
			//System.out.println("processing... [" + (++currIteration)+"]");
			if (tupleCount == howMany) 
			{
				return true;
			}
			try 
			{
				//pick the next tuple to process
				if (this.getLeftInputQueue().getQueueSize(this) == 0) 
				{
					newObj = this.getRightInputQueue().dequeue(this);
					indication = RIGHT;
				} 
				else if (this.getRightInputQueue().getQueueSize(this) == 0) 
				{
					newObj = this.getLeftInputQueue().dequeue(this);
					indication = LEFT;
				} 
				else 
				{
					long leftTimeStamp = 0;
					long rightTimeStamp = 0;
		
					leftTimeStamp =
						this.getLeftInputQueue().peek(this).getNthTimeStamp(this.leftJoinAttrPos);
					rightTimeStamp = 
						this.getRightInputQueue().peek(this).getNthTimeStamp(this.rightJoinAttrPos);
						
					if(leftTimeStamp <= rightTimeStamp) 
					{
						newObj = this.getLeftInputQueue().dequeue(this);
						indication = LEFT;
					} 
					else 
					{
						newObj = this.getRightInputQueue().dequeue(this);
						indication = RIGHT;
					}					
				}
			} 
			catch(XATQueueException e1)
			{
				return true;
			}
	
			if (indication == LEFT) 
			{
				if (newObj.getObjTypeID()==1) 
				{ //receive a new punctuation
					continue;
				} 
				else 
				{
					tupleCount ++;
					
					leftLatestProcessTupleTime = newObj.getNthTimeStamp(this.leftJoinAttrPos);
					numNewTuplesSinceLastInvalidate ++;
					if (numNewTuplesSinceLastInvalidate == this.invalidateThreshold) 
					{
						int numLeftTuplesPurged = leftState.purgeByWindow(200, rightLatestProcessTupleTime);
						int numRightTuplesPurged = rightState.purgeByWindow(200, leftLatestProcessTupleTime);
						numNewTuplesSinceLastInvalidate = 0;
					}
					
					Object value = newObj.getCell(leftJoinAttrPos).getValue();
					
					if(value != null)
					{						
						XATTuple[] matchedTuples = rightState.probe(value);
						if (matchedTuples != null) 
						{
							ArrayList resultTupleArray = new ArrayList();
							XATDataValue[] thisTupleContent = newObj.getArrayData();
							long[] thisTupleTime = newObj.getTimeStamp();
							long thisTupleJoinAttrTime = thisTupleTime[this.leftJoinAttrPos];
							long thisTupleMaxTime = newObj.getMaxTimeStamp();
							long thisTupleMinTime = newObj.getMinTimeStamp();
							
							for (int i=0; i<matchedTuples.length; i++) 
							{
								XATTuple joinedTuple = matchedTuples[i];
								long[] joinedTupleTime = joinedTuple.getTimeStamp();
								long joinedTupleJoinAttrTime = joinedTupleTime[this.rightJoinAttrPos];
								/*
								if ((thisTupleJoinAttrTime-joinedTupleJoinAttrTime)> getWindowSize(this.RIGHT)) {
									continue;
								}*/
	
								XATDataValue[] joinedTupleContent = joinedTuple.getArrayData();
								int numOutputAttr = thisTupleContent.length + joinedTupleContent.length;
								XATDataValue[] resultTupleContent = new XATDataValueImp[numOutputAttr];
								long[] resultTupleTime = new long[numOutputAttr];
						
								int j=0, k=0;
								for (; j<thisTupleContent.length; j++) 
								{
									resultTupleContent[j] = thisTupleContent[j];
									resultTupleTime[j] = thisTupleTime[j];
								}
								for (; k<joinedTupleContent.length; k++) 
								{
									resultTupleContent[k+j] = joinedTupleContent[k];
									resultTupleTime[k+j] = joinedTupleTime[k];
								}
								XATTuple resultTuple = new XATTupleImp(resultTupleContent);
								resultTuple.setTimeStamp(resultTupleTime);
								long joinedTupleMaxTime = joinedTuple.getMaxTimeStamp();
								long joinedTupleMinTime = joinedTuple.getMinTimeStamp();
								resultTuple.setMaxTimeStamp(thisTupleMaxTime>joinedTupleMaxTime?thisTupleMaxTime:joinedTupleMaxTime);
								resultTuple.setMinTimeStamp(thisTupleMinTime<joinedTupleMinTime?thisTupleMinTime:joinedTupleMinTime);
								resultTupleArray.add(resultTuple);
							}
							if (resultTupleArray.size()>0) 
							{
								//System.out.println("Total enqueued " + resultTupleArray.size() + " results.");
								this.getOutputQueue().enqueue((XATTuple[]) resultTupleArray.toArray(new XATTuple[] {}), this);
							}
						}
						leftState.insertTuple(value, newObj);
					}//end value != null
				}
			} 
			else 
			{
				if (newObj.getObjTypeID() == 1) 
				{ 
					//receive a new punctuation
					continue;
				} 
				else 
				{
					tupleCount ++;
					
					rightLatestProcessTupleTime = newObj.getNthTimeStamp(this.rightJoinAttrPos);
					numNewTuplesSinceLastInvalidate ++;
					if (numNewTuplesSinceLastInvalidate == this.invalidateThreshold) 
					{
						int numLeftTuplesPurged = leftState.purgeByWindow(200, rightLatestProcessTupleTime);
						int numRightTuplesPurged = rightState.purgeByWindow(200, leftLatestProcessTupleTime);
						numNewTuplesSinceLastInvalidate = 0;
					}
					
					Object value = newObj.getCell(rightJoinAttrPos).getValue();
					if(value != null)
					{					
						XATTuple[] matchedTuples = leftState.probe(value);
						if (matchedTuples != null) 
						{
							ArrayList resultTupleArray = new ArrayList();
							XATDataValue[] thisTupleContent = newObj.getArrayData();
							long[] thisTupleTime = newObj.getTimeStamp();
							long thisTupleJoinAttrTime = thisTupleTime[this.rightJoinAttrPos];
							long thisTupleMaxTime = newObj.getMaxTimeStamp();
							long thisTupleMinTime = newObj.getMinTimeStamp();
	
							for (int i=0; i<matchedTuples.length; i++) 
							{
								XATTuple joinedTuple = matchedTuples[i];
	
								long[] joinedTupleTime = joinedTuple.getTimeStamp();
								long joinedTupleJoinAttrTime = joinedTupleTime[this.leftJoinAttrPos];
								/*
								if ((thisTupleJoinAttrTime-joinedTupleJoinAttrTime)> getWindowSize(this.LEFT)) {
									continue;
								}*/
	
								XATDataValue[] joinedTupleContent = joinedTuple.getArrayData();
								int numOutputAttr = thisTupleContent.length + joinedTupleContent.length;
								XATDataValue[] resultTupleContent = new XATDataValueImp[numOutputAttr];
								long[] resultTupleTime = new long[numOutputAttr];
						
								int j = 0, k = 0;
								for (; j < joinedTupleContent.length; j++) 
								{
									resultTupleContent[j] = joinedTupleContent[j];
									resultTupleTime[j] = joinedTupleTime[j];
								}
								
								for (; k < thisTupleContent.length; k++) 
								{
									resultTupleContent[k+j] = thisTupleContent[k];
									resultTupleTime[k+j] = thisTupleTime[k];
								}
								
								XATTuple resultTuple = new XATTupleImp(resultTupleContent);
								resultTuple.setTimeStamp(resultTupleTime);
								long joinedTupleMaxTime = joinedTuple.getMaxTimeStamp();
								long joinedTupleMinTime = joinedTuple.getMinTimeStamp();
								resultTuple.setMaxTimeStamp(thisTupleMaxTime>joinedTupleMaxTime?thisTupleMaxTime:joinedTupleMaxTime);
								resultTuple.setMinTimeStamp(thisTupleMinTime<joinedTupleMinTime?thisTupleMinTime:joinedTupleMinTime);
								resultTupleArray.add(resultTuple);
							}
							
							if (resultTupleArray.size() > 0) 
							{
								//System.out.println("Total enqueued " + resultTupleArray.size() + " results.");
								this.getOutputQueue().enqueue((XATTuple[]) resultTupleArray.toArray(new XATTuple[] {}), this);
							}
						}
						rightState.insertTuple(value, newObj);
					}//if value != null
				}
			}
		}
	}

	public void setState(OperatorState[] state) 
	{
		this.states = new OperatorState[state.length];
		for(int i = 0; i < this.states.length;i++)
		{
			this.states[i] = state[i];
		}
		this.states = state;
		this.leftState = (WEquiJoinState) state[0];
		this.rightState = (WEquiJoinState) state[1];
	}

	public int getLeftJoinAttrPos() 
	{
		return leftJoinAttrPos;
	}
	
	public int getRightJoinAttrPos() 
	{
		return rightJoinAttrPos;
	}
	
	public void setLeftJoinAttrPos(int lJoinAttrPos) 
	{
		this.leftJoinAttrPos = lJoinAttrPos;
	}
	
	public void setRightJoinAttrPos(int rJoinAttrPos) 
	{
		this.rightJoinAttrPos = rJoinAttrPos;
	}
	
	public int getInvalidateThreshold() 
	{
		return this.invalidateThreshold;
	}
	public void setInvalidateThreshold(int iThreshold) 
	{
		this.invalidateThreshold = iThreshold;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAccumulatedTuples(long)
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator#cleanAllAccumulatedTuples()
	 */
	public boolean cleanAllAccumulatedTuples() 
	{
		return false;
	}

	public boolean matchStates(XATStreamOperator op) 
	{
		return false;
	}

	public boolean hasUnmatchedState() 
	{
		return false;
	}

	public boolean recomputeUnmatchedState() 
	{
		return false;
	}

	public int setStateID(int startID) 
	{
		return 0;
	}

	public String toToolTipString() 
	{
		return null;
	}
	
	public boolean finalizeOperation()
	{
		return true;
	}

}
