/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
//import java.lang.reflect.Constructor;
//import streamgenerator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestreamoperator.PipeWBinaryJoinOperatorHeadImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;

/**
 * This operator computes join results in two modes- recompute and delta.
 * 
 * The switching mode mechanism is yet to be implemented
 *  
 * Creation date: (12/12/2006 2:26:41 PM)
 * 
 * @author: Abhishek Mukherji
 */
public class RegularWindowBinaryJoinDeltaOperatorImp extends PipeWBinaryJoinOperatorHeadImp {


	private TuplesList outputQueueCache = null;//recompute complete result

	//private TuplesList outputQueueCacheRecom = null;
	private TuplesList outputQueueCacheDelta= null; // delta result
	private boolean OutputModeFlag=true; //true means complete result, false means delta.
	private boolean InputModeFlag=false;  // true means complete result, false means delta.

	private long start = 0;


	public RegularWindowBinaryJoinDeltaOperatorImp(XATNode node) {
		super(node);
	}

	private int processLeftTuple(XATTuple tuple) {
		XATTuple[] joinedtuples = null;
		XATTuple[] prevJoinedtuples = null;
		//to track the output tuples previously joined and now purged by negative tuple 
		XATTuple[] prevJoinedNegtuples = null;

		int ret=0;
		if (this.outputQueueCache == null)
			this.outputQueueCache = new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

		if (this.outputQueueCacheDelta == null)
			this.outputQueueCacheDelta = new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

		//step 0: processing the negative tuple from the input queue
		if(!isInputModeFlag()){

			if(tuple instanceof XATTupleNegativeImp)
			{
				if(((XATTupleNegativeImp)tuple).isNegativeFlag()){
				//outputQueueCacheDelta.addToEnd(tuple);
				List purgedByNegativeTuple = ((PipeWFIFOState) states[0]).purgeByNegativeTupleSelf((XATTupleNegativeImp)tuple);

				ListIterator purgedByNegTuplesList = purgedByNegativeTuple.listIterator();

				//the output tuples corresponding to the purged tuples need to be removed.

				while (purgedByNegTuplesList.hasNext())
				{
					XATTuple t = (XATTuple) purgedByNegTuplesList.next();
					prevJoinedNegtuples = ((PipeWFIFOState) states[1]).probe(t,
							this.getExpression(), 0);
					for(int i =0;i<prevJoinedNegtuples.length;i++)
					{
						if(this.outputQueueCache.hasTuple(prevJoinedNegtuples[i]))
						{
							outputQueueCache.remove(prevJoinedNegtuples[i]);
							/*XATTupleNegativeImp tn = new XATTupleNegativeImp(prevJoinedNegtuples[i]);
							tn.setNegativeFlag(true);
							outputQueueCacheDelta.addToEnd(tn, true);*/
							System.out.println("\n  invalidating due to negative tuple " +prevJoinedNegtuples[i]);
						}
					}
				 }
					return 0;
				}
				 	
/*				if output requires delta, we need to put the negative tuple into the state
				if(!this.isOutputModeFlag())
				{
					((PipeWFIFOState) states[0]).insertTuple(tuple);
				}
				//return 0;*/
			    
			}
		}


		//step 1: cross purge
		List purgedTuples = ((PipeWFIFOState) states[1]).purgeByTuplePipeCross(tuple);

		ListIterator purgedTuplesList = purgedTuples.listIterator();

		//the output tuples corresponding to the purged tuples need to be removed.

		while (purgedTuplesList.hasNext())
		{
			XATTuple t = (XATTuple) purgedTuplesList.next();
			prevJoinedtuples = ((PipeWFIFOState) states[0]).probe(t,this.getExpression(), 0);
			for(int i =0;i<prevJoinedtuples.length;i++)
			{
				if(this.outputQueueCache.hasTuple(prevJoinedtuples[i]))
				{
					outputQueueCache.remove(prevJoinedtuples[i]);
					/*XATTupleNegativeImp tn = new XATTupleNegativeImp(prevJoinedtuples[i]);
					tn.setNegativeFlag(true);
					outputQueueCacheDelta.addToEnd(tn, true);*/
					System.out.println("\n expiring " +prevJoinedtuples[i]);
				}
			}
		}
		//step 2: probe the state
		joinedtuples = ((PipeWFIFOState) states[1]).probe(tuple,
				this.getExpression(), 1);
		//for (int i=0;i<this.getOutputQueueCount();i++ ){
		this.getOutputQueue().enqueue(joinedtuples, this);

		for(int i =0;i<joinedtuples.length;i++)
		{
			this.outputQueueCache.addToEnd(joinedtuples[i],true);
			/*XATTupleNegativeImp tn = new XATTupleNegativeImp(joinedtuples[i]);
			tn.setNegativeFlag(false);
			outputQueueCacheDelta.addToEnd(tn, true);*/
			this.outputQueueCacheDelta.addToEnd(joinedtuples[i],true);
		}

		
//		output at every increment (HOP)
		if(tuple.getMinTimeStamp()>this.start+this.windowSize)
		{
			TuplesList toEnqueue ;//= new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

			if (isOutputModeFlag()){
				toEnqueue = outputQueueCache;	
			}
			else
			{
				toEnqueue = outputQueueCacheDelta;
			}

			ListIterator cacheIt = toEnqueue.listIterator();
			while (cacheIt.hasNext())
			{
				XATTuple t = (XATTuple) cacheIt.next();
				this.getOutputQueue().enqueue(t, this);
			}		
			outputQueueCache.clear();
			outputQueueCacheDelta.clear();
		}

		ret+=joinedtuples.length;
		//}
		//step 3: attach to outputqueue
		//step 4: insert into state
		((PipeWFIFOState) states[0]).insertTuple(tuple);
		return ret;

	}

	private int processRightTuple(XATTuple tuple) {
		XATTuple[] joinedtuples = null;
		XATTuple[] prevJoinedtuples = null;
		//to track the output tuples previously joined and now purged by negative tuple 
		XATTuple[] prevJoinedNegtuples = null;

		int ret=0;


		if (this.outputQueueCache == null)
			this.outputQueueCache = new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

		if (this.outputQueueCacheDelta == null)
			this.outputQueueCacheDelta = new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

		//step 0: processing the negative tuple from the input queue
		if(!isInputModeFlag()){
			if(tuple instanceof XATTupleNegativeImp)
			{
				if(((XATTupleNegativeImp)tuple).isNegativeFlag()){
				//outputQueueCacheDelta.addToEnd(tuple);
				List purgedByNegativeTuple = ((PipeWFIFOState) states[1]).purgeByNegativeTupleSelf((XATTupleNegativeImp)tuple);

				ListIterator purgedByNegTuplesList = purgedByNegativeTuple.listIterator();

				//the output tuples corresponding to the purged tuples need to be removed.

				while (purgedByNegTuplesList.hasNext())
				{
					XATTuple t = (XATTuple) purgedByNegTuplesList.next();
					prevJoinedNegtuples = ((PipeWFIFOState) states[0]).probe(t,
							this.getExpression(), 0);
					for(int i =0;i<prevJoinedNegtuples.length;i++)
					{
						if(this.outputQueueCache.hasTuple(prevJoinedNegtuples[i]))
						{
							outputQueueCache.remove(prevJoinedNegtuples[i]);
				/*			XATTupleNegativeImp tn = new XATTupleNegativeImp(prevJoinedNegtuples[i]);
							tn.setNegativeFlag(true);
							outputQueueCacheDelta.addToEnd(tn, true);*/
							System.out.println("\n invalidating due to negative tuple" +prevJoinedNegtuples[i]);
						}
					}
				}
				return 0;
				}
			/*	//if output requires delta, we need to put the negative tuple into the state 
				if(!this.isOutputModeFlag())
				{
					((PipeWFIFOState) states[0]).insertTuple(tuple);
				}
				return 0;*/
			}
		}


		//step 1: cross purge
		List purgedTuples = ((PipeWFIFOState) states[0])
		.purgeByTuplePipeCross(tuple);


		ListIterator purgedTuplesList = purgedTuples.listIterator();
		//the output tuples corresponding to the purged tuples need to be removed.

		while (purgedTuplesList.hasNext())
		{
			XATTuple t = (XATTuple) purgedTuplesList.next();
			prevJoinedtuples = ((PipeWFIFOState) states[1]).probe(t,
					this.getExpression(), 0);
			for(int i =0;i<prevJoinedtuples.length;i++)
			{
				if(this.outputQueueCache.hasTuple(prevJoinedtuples[i]))
				{	
					outputQueueCache.remove(prevJoinedtuples[i]);
					/*XATTupleNegativeImp tn = new XATTupleNegativeImp(prevJoinedtuples[i]);
					tn.setNegativeFlag(true);
					outputQueueCacheDelta.addToEnd(tn, true);*/
					System.out.println("\n expiring " +prevJoinedtuples[i]);
				}
			}
		}

		//step 2: probe the state
		joinedtuples = ((PipeWFIFOState) states[0]).probe(tuple,
				this.getExpression(), 0);

		//if(this.outputQueueCacheDelta == null)
		//this.outputQueueCacheDelta = new TupleList();//this.getOutputQueue().getSchema(),"timestamp");
		for(int i =0;i<joinedtuples.length;i++)
		{
			this.outputQueueCache.addToEnd(joinedtuples[i],true);
			/*XATTupleNegativeImp tn = new XATTupleNegativeImp(joinedtuples[i]);
			tn.setNegativeFlag(true);
			outputQueueCacheDelta.addToEnd(tn, true);*/
			this.outputQueueCacheDelta.addToEnd(joinedtuples[i],true);

		}
//		output at every increment (HOP)
		if(tuple.getMinTimeStamp()>this.start+this.windowSize)
		{
			TuplesList toEnqueue ;//= new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

			if (isOutputModeFlag()){
				toEnqueue = outputQueueCache;	
			}
			else
			{
				toEnqueue = outputQueueCacheDelta;
			}

			ListIterator cacheIt = toEnqueue.listIterator();
			while (cacheIt.hasNext())
			{
				XATTuple t = (XATTuple) cacheIt.next();
				this.getOutputQueue().enqueue(t, this);
			}	
			outputQueueCache.clear();
			outputQueueCacheDelta.clear();
		}
		
		ret+=joinedtuples.length;
		//}
		//step 3: attach to outputqueue
		//step 4: insert into state
		((PipeWFIFOState) states[1]).insertTuple(tuple);
		return ret;
	}

	/**
	 * Insert the method's description here. return: -1 nothing in queue, no run
	 * 0 exception 1 run successfully
	 * 
	 * Creation date: (1/21/2003 10:57:01 AM)
	 */
	public boolean run() {

		
		XATTuple leftTuple = null;
		XATTuple[] leftTuples = null;
		XATTuple rightTuple = null;
		XATTuple[] rightTuples = null;
		long timestart = System.nanoTime();
		try {
			if ((this.getLeftInputQueue().getSchema() == null)
					|| (this.getRightInputQueue().getSchema() == null)) {
				return false;
			}

			/*//outputqueue for join output
		//could be multiple outputs. but the schema is the same.
		for (int i=0;i<this.getOutputQueueCount();i++ )
			if (this.getOutputQueue(i).getSchema() == null) {
				init(i);
			}*/


			if(this.getOutputQueue().getSchema() == null)
				init();
//			return false;



			/* Load the current time_window and increment lengths */
			this.increment = 1000;
			this.windowSize = 20000;

			//hopping through increment
			//left state
		//	List purgedByHopTuple = ((PipeWFIFOState) states[0]).purgeByTimeStampSelf(this.start+this.increment);

			
			
		
//			Hopping through right state
			//List purgedByHopTupleR = ((PipeWFIFOState) states[1]).purgeByTimeStampSelf(this.start+this.increment);

		
		/** figure out how many tuples to dequeue from left and right queue */
			XATQueue lQueue = this.getLeftInputQueue();
			XATQueue rQueue = this.getRightInputQueue();
			int size1 = lQueue.getQueueSize(this);
			int size2 = rQueue.getQueueSize(this);
			int lHowmany = 0, rHowmany = 0;

			int dequeueHowmany = Integer.parseInt(this
					.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

			//System.out.println(this + "tuple to dequeue " + dequeueHowmany);

			int totalOutput =0;

			//left queue is empty. we only deal with tuples from right queue
			//here we assume the tuple from left and right will arrive in order
			if (size1 <= 0 && size2 > 0) {
				lHowmany = 0;
				rHowmany = (dequeueHowmany > size2) ? size2 : dequeueHowmany;
				try {
					lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
					rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
					setNumberOfPossibleOutputs();
				} catch (Exception e) {
					System.out.println(e);
				}

				//begin exp stat collect
				stats_string.append(System.currentTimeMillis()+"\t"); 

				rightTuples = rQueue.dequeue(rHowmany, this);
				for (int x = 0; x < rightTuples.length; x++) {
					totalOutput+=this.processRightTuple(rightTuples[x]);
				}
				stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

				return true;
			}

			//right queue is empty. we only deal with tuples from left queue
			//here we assume the tuple from left and right will arrive in order
			if (size2 <= 0 && size1 > 0) {
				rHowmany = 0;
				lHowmany = (dequeueHowmany > size1) ? size1 : dequeueHowmany;
				try {
					lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
					rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
					setNumberOfPossibleOutputs();
				} catch (Exception e) {
					System.out.println(e);
				}
				//begin exp stat collect
				stats_string.append(System.currentTimeMillis()+"\t"); 

				leftTuples = lQueue.dequeue(lHowmany, this);
				for (int x = 0; x < leftTuples.length; x++) {
					totalOutput+=this.processLeftTuple(leftTuples[x]);
				}
				stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

				return true;
			}

			//Both queues are not empty.
			//here we assume the tuple from left and right will arrive in order
			if (size2 > 0 && size1 > 0) {
				lHowmany = 0;
				rHowmany = 0;
				int total = (dequeueHowmany > (size1 + size2)) ? (size1 + size2)
						: dequeueHowmany;

				//begin exp stat collect
				stats_string.append(System.currentTimeMillis()+"\t"); 

				for (int x = 0; x < total; x++) {
					if (lHowmany < size1 && leftTuple == null) {
						leftTuple = lQueue.dequeue(this);
						lHowmany++;
					}
					if (rHowmany < size2 && rightTuple == null) {
						rightTuple = rQueue.dequeue(this);
						rHowmany++;
					}
					if (rightTuple!=null&&leftTuple == null
							|| rightTuple!=null&&leftTuple.getMaxTimeStamp()
							- rightTuple.getMaxTimeStamp() >= 0) {
						totalOutput+=this.processRightTuple(rightTuple);
						rightTuple = null;
					} else if (leftTuple!=null&&rightTuple == null
							|| leftTuple!=null&&rightTuple.getMaxTimeStamp()
							- leftTuple.getMaxTimeStamp() > 0) {
						totalOutput+=this.processLeftTuple(leftTuple);

						leftTuple = null;
					}
				}
			}

			/* Change start to reflect the new beginning of the queue windows */ 
			start += this.increment;
			stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

			try{
				lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
				rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);


			}
		}catch(Exception e1) {
			e1.printStackTrace();
		}   
		//System.out.print("RunTime");
	    File file = new File ("runtimeJOperatorWithConvertion0.1k10.txt");
		FileWriter file_writer;
		try {
			file_writer = new FileWriter (file,true);
			file_writer.append(System.nanoTime()-timestart +"    \n");
			
			 file_writer.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
			
		return true;

	}

	public void printStatsToFile() {
		try {
			String fileName = new String("RegularJoinDelta_" + System.currentTimeMillis()+".txt");
			FileWriter out = new FileWriter(fileName, true);
			out.write(this.stats_string.toString());
			out.close();

		} catch (IOException e) {
			System.out.println(e.toString());
		}
	}

	public boolean isInputModeFlag() {
		return InputModeFlag;
	}

	public void setInputModeFlag(boolean inputModeFlag) {
		InputModeFlag = inputModeFlag;
	}

	public boolean isOutputModeFlag() {
		return OutputModeFlag;
	}

	public void setOutputModeFlag(boolean outputModeFlag) {
		OutputModeFlag = outputModeFlag;
	}


}

