/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import java.util.*;
//import java.lang.reflect.Constructor;
//import streamgenerator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
/**
 * Insert the type's description here.
 * Creation date: (2/17/2003 4:44:15 PM)
 * @author: Yali Zhu
 */
public class WBinaryJoinOperatorImp
	extends XATMultiQueueWindowStreamOperatorImp
	implements WBinaryJoinOperator {

	/** if this is oldBoxRoot, allOldOperators will be set to the right array. Otherwise, this is null */
	private XATStreamOperator[] allOldOperators = null;
	private ExpressionStr expression;
	ArrayList keptTuples = new ArrayList();


	/** The numOfChildren and numOfChildrenDone are here to decide if an operator inside an old box has done
	 *  during migration stage, meaning all the old tuple in states are purged. 
	 *  numOfChildren starts as the real num of children of this operator. Since we have only binary join now, this is
	 *  always set to be 2.
	 *  numOfChildrenDone always starts as 0, and increased by its child operator if that operator is done.
	 *
	 *  When numOfChildrenDone == numOfChildren, and the two states of this operator has no tuple gets processed after
	 *  migration starts, this operator is also done and will increase its parent's numOfChildrenDone.
	 */
	//private int numOfChildren = 0;    
	//private int numOfChildrenDone = 0;
	//private WFIFOState stateL;
	//private WFIFOState stateR;



	public WBinaryJoinOperatorImp(WBinaryJoinOperatorImp oldop, XATNode newnode) {
		super(oldop, newnode);
		this.expression = oldop.expression;
		//states = new WOperatorState[2];
		//this.inToOutColPos = oldop.inToOutColPos;
		this.windowSize = oldop.windowSize;
	}
	
	/**
	 * Insert the method's description here.
	 * Creation date: (2/17/2003 4:48:47 PM)
	 */
	public WBinaryJoinOperatorImp(XATNode node) {
		super(node);
	}
	
	public void initState(String statetype) {

		try{	
			//Constructor con = Class.forName(statetype).getConstructor();
			states = new WOperatorState[2];
			states[0] = (WOperatorState)Class.forName(statetype).newInstance();
			states[1] = (WOperatorState)Class.forName(statetype).newInstance();
			states[0].setOperator(this);
			states[1].setOperator(this);
			if(this.windowSize > 0) {
				((WOperatorState)states[0]).setTimeWindow(this.windowSize);
				((WOperatorState)states[1]).setTimeWindow(this.windowSize);
			}
		}catch(Exception e) {
			e.printStackTrace();	
		}
		
	}
	/**
	 * Clear all tuples inside this operator and all tuples insides its states
	 */
//	public void clearAllTuples() {
//		super.clearAllTuples();
//		
//		states[0].removeAllTuples();
//		states[1].removeAllTuples();
//	}
	/**
	 * This method is used at the end of Parallel Migration when old box is done.
	 * If this operator is a leaf operator of new box, the kept tuples
	 * for purpose of preserve partial order will now be pushed out and
	 * enqueued to output queue all at once.
	 * Creation date: (6/26/2003 4:31:05 AM)
	 */
    
    public void setDefaultStatistics()
    {
        super.setDefaultStatistics();
        //addStatistic("NUMBER_OF_TUPLES_OUTPUTTED_BY_JOIN");
        //specific.put("NUMBER_OF_PUNCTS_OUTPUTTED_BY_JOIN", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.NumberOfPunctsOutputtedByJoin");
        addStatistic("NUMBER_OF_TUPLES_IN_LEFT_STATE");
        addStatistic("NUMBER_OF_TUPLES_IN_RIGHT_STATE");

    }
    
	public void enqueueKeptTuples() {

		XATTuple[] tuples = (XATTuple[]) keptTuples.toArray(new XATTuple[] {
		});
		this.getOutputQueue().enqueue(tuples, this);
		System.out.println(
			"push out " + tuples.length + " tuples from newbox root operator");

		keptTuples.clear();
		this.isNewBoxRoot = false;
		return;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/5/2003 6:59:19 PM)
	 */
//	public boolean evalJoinPredicate(XATTuple t1, int t1Pos, XATTuple t2, int t2Pos) {
//
//		switch (relation) {
//			case BinCOMPExpressionStr.EQ :
//				//return (t1.getCell(t1Pos).equals(t2.getCell(t2Pos)));
//				return (
//					t1.getCell(t1Pos).convertToString().equals(
//						t2.getCell(t2Pos).convertToString()));
//		}
//		return false;
//	}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/3/2003 9:53:23 PM)
	 */
	private void execOperator(int lHowmany, int rHowmany) {

		XATQueue oQueue = this.getOutputQueue();
		XATQueue linputQueue = this.getLeftInputQueue();
		XATQueue rinputQueue = this.getRightInputQueue();
		XATTuple[] tuples;
		XATTuple[] joinedtuples;
		
		/* join tuples from left queue to right state */
		if(lHowmany != 0) {
			tuples = linputQueue.dequeue(lHowmany, this);
			if(isOldBoxLeaf && linputQueue.isOldBoxInput()) {
				for(int i=0; i<tuples.length; i++) {
					tuples[i].setToNew(true);
				}
			}
			if(states[1].getStateSize() != 0) {
				joinedtuples = ((WOperatorState)states[1]).probe(tuples, this.expression, 1);
				if(isOldBoxRoot) {
					for(int i=0; i<joinedtuples.length; i++) {
						joinedtuples[i].setToNew(false);
					}
				}
				if(!isNewBoxRoot)
					oQueue.enqueue(joinedtuples, this);
			}
			((WOperatorState)states[0]).insertTuple(tuples);

		}

		
		/*join tuples from right queue to left state */
		if(rHowmany != 0) {
			tuples = rinputQueue.dequeue(rHowmany, this);
			if(isOldBoxLeaf && rinputQueue.isOldBoxInput()) {
				for(int i=0; i<tuples.length; i++) {
					tuples[i].setToNew(true);
				}
			}
			if(states[0].getStateSize() != 0) {
				joinedtuples = ((WOperatorState)states[0]).probe(tuples, this.expression, 0);
			
				if(isOldBoxRoot) {
					for(int i=0; i<joinedtuples.length; i++) {
						joinedtuples[i].setToNew(false);
					}
				}
				if(!isNewBoxRoot)
					oQueue.enqueue(joinedtuples, this);
			}
			((WOperatorState)states[1]).insertTuple(tuples);
		}
	}
	
	/**
	 * Insert the method's description here.
	 * Creation date: (6/3/2003 2:07:16 AM)
	 * @return boolean
	 */
	public boolean getAllOldPurgedFlag() {
		return 
		(((WOperatorState)states[0]).checkAllOldTuplePurged() && ((WOperatorState)states[1]).checkAllOldTuplePurged());
	}
	
	/**
	 * @return
	 */
	public ExpressionStr getExpression() {
		return expression;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (4/3/2003 5:00:07 AM)
	 * @return queryplanelements.WFIFOState
	 */
	public OperatorState getLeftState() {
		return states[0];
	}
	
	public int getLeftStateSize() {
		return states[0].getStateSize();
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (5/28/2003 12:11:23 PM)
	 * @return int
	 */
//	public int getLPos() {
//		return lPos;
//	}
		
	/**
	 * Insert the method's description here.
	 * Creation date: (4/3/2003 5:00:50 AM)
	 * @return queryplanelements.WFIFOState
	 */
	public OperatorState getRightState() {
		return states[1];
	}
	
	public int getRightStateSize() {
		return states[1].getStateSize();
	}

	public int getStateSize() {
		return (states[0].getStateSize() + states[1].getStateSize());
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (3/15/2003 10:23:33 PM)
	 * @return int
	 */

	/**
	 * This reinit method is only used during the stage of optimization.
	 * This method assumes that all operator properties, including the predicates, window size, state schema 
	 * and input queue schema has been initialized at the execution init stage.
	 * After optimization, the state and output queue schema might be changed so this method just reset them 
	 * to the corrent schema according to the input queue schema.
	 */
//	public void reinit() {
//		super.init();
//	}
	
	/**
	 * Insert the method's description here.
	 * return: -1 nothing in queue, no run
	 *			0 exception
	 *			1 run successfully
	 * 
	 * Creation date: (1/21/2003 10:57:01 AM)
	 */
	public boolean run() {

		if ((this.getLeftInputQueue().getSchema() == null) || (this.getRightInputQueue().getSchema() == null)){
				return false;
		}
		if(this.getOutputQueue().getSchema() == null){
			init();
		}
		
		/** figure out how many tuples to dequeue from left and right queue */
		XATQueue lQueue = this.getLeftInputQueue();
		XATQueue rQueue = this.getRightInputQueue();
		int size1 = lQueue.getQueueSize(this);
		int size2 = rQueue.getQueueSize(this);
		int lHowmany, rHowmany;

		int dequeueHowmany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
			
		//System.out.println(this + "tuple to dequeue " + dequeueHowmany);

		if ((size1 + size2) != 0) {
			lHowmany =
				(int) ((double) size1 / (size1 + size2) * dequeueHowmany);
			lHowmany = (size1 < lHowmany) ? size1 : lHowmany;
			rHowmany =
				(size2 < dequeueHowmany - lHowmany)? size2 : (dequeueHowmany - lHowmany);
					
			try {
				lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
				rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
				setNumberOfPossibleOutputs();
			} catch(Exception e) {
				System.out.println(e);
			}
		} 
		else {
			try {
				lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "0");
				rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "0");
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);
			}
			/* nothing to run */
			return false;
		}

		/* if this is an old operator, dequeue all tuples from its input queue if the queue is not oldboxinputQ */
		/* this is for the purpose of finishing migration 1 quicker */
		if (this.isOldOperator) {
			lHowmany = size1;
			rHowmany = size2;
		}

		if (debugFlag) {
			System.out.println("dequeue " + lHowmany + "tuples from left queue");
			System.out.println("dequeue " + rHowmany + "tuples from right queue");
		}
			
		execOperator(lHowmany, rHowmany);

		/* if this is root operator of old box, check to see if both states have all old tuples purged */
		/* solely used by migration 1 */
		if (this.isOldOperator) {
			//if (this.isOldBoxLeaf) {
			if (lQueue.isOldBoxInput()) {
				if (!((WOperatorState)states[0]).getAllOldPurgedFlag())
					((WOperatorState)states[0]).checkAllOldTuplePurged();
			} else
				((WOperatorState)states[0]).checkAllOldTuplePurged();
			if (rQueue.isOldBoxInput()) {
				if (!((WOperatorState)states[1]).getAllOldPurgedFlag())
					((WOperatorState)states[1]).checkAllOldTuplePurged();
			} else
				((WOperatorState)states[1]).checkAllOldTuplePurged();
		}

		/* run successfully */
		return true;
	}

	/**
	 * Insert the method's description here.
	 *
	 * This method is currently NOT IN USE.
	 *
	 * Creation date: (5/25/2003 4:08:55 AM)
	 * @return int
	 */
	public int runForMigration1() {

		/** figure out how many tuples to dequeue from left and right queue */
		XATQueue lQueue = this.getLeftInputQueue();
		XATQueue rQueue = this.getRightInputQueue();

		int lHowmany = (lQueue.isOldBoxInput()) ? 0 : lQueue.getQueueSize(this);
		int rHowmany = (rQueue.isOldBoxInput()) ? 0 : rQueue.getQueueSize(this);

		if (debugFlag) {
			System.out.println(
				"dequeue " + lHowmany + "tuples from left queue");
			System.out.println(
				"dequeue " + rHowmany + "tuples from right queue");
		}

		try{
			lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
			rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
		} catch(Exception e) {
			System.out.println(e);
		}

		if ((lHowmany + rHowmany) == 0)
			/* nothing to run */
			return -1;
			
		execOperator(lHowmany, rHowmany);

		/* run successfully */
		return 1;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (5/25/2003 4:08:55 AM)
	 * @return int
	 */
	public int runForMigration2(long latestTimeStamp) {

		XATQueue lQueue = this.getLeftInputQueue();
		XATQueue rQueue = this.getRightInputQueue();

		int lsize = 0, rsize = 0;
		int lHowmany = 0;
		int rHowmany = 0;

		/** figure out howmany tuple to dequeue from left queue */
		if ((lsize = lQueue.getQueueSize(this)) != 0) {
			lHowmany = lsize;

			//if stream Queue, dequeue based on latestTimeStamp
			List ownerList = lQueue.getQueueOwner();
			int ownersize = ownerList.size();
			int ownerindex = 0;
			while (ownerindex < ownersize) {
				//if (lQueue.getQueueOwner() instanceof XATStream) {
				if (ownerList.get(ownerindex) instanceof XATStream) {
					XATTuple[] tuples = lQueue.peek(lsize, this);
					for (int i = 0; i < lsize; i++) {

						if (((XATTuple) tuples[i]).getMinTimeStamp() >= latestTimeStamp) {
							lHowmany = i;
							break;
						}
					}
				}
				break;
			}
			ownersize++;
		}
		
		/* figure out howmany tuple to dequeue from right queue */
		if ((rsize = rQueue.getQueueSize(this)) != 0) {
			rHowmany = rsize;

			if (rQueue.getQueueOwner() instanceof XATStream) {
				XATTuple[] tuples = rQueue.peek(rsize, this);
				for (int i = 0; i < rsize; i++) {

					if (tuples[i].getMinTimeStamp() >= latestTimeStamp) {
						rHowmany = i;
						break;
					}
				}
			}
		}
	
		try{
			lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
			rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
		} catch(Exception e) {
			System.out.println(e);
		}

		if ((lHowmany + rHowmany) == 0)
			/* nothing to run */
			return -1;

		if (debugFlag) {
			System.out.println(
				"dequeue " + lHowmany + "tuples from left queue");
			System.out.println(
				"dequeue " + rHowmany + "tuples from right queue");
		}

		execOperator(lHowmany, rHowmany);

		/* run successfully */
		return 1;
	}
	public void setDefaultProperties() {

		super.setDefaultProperties();
		addProperty("TIME_WINDOW","0");
		// type of state
		addProperty("STATE_TYPE", "WFIFOSTATE");
	}
	/**
	 * @param expression
	 */
	public void setExpression(ExpressionStr expr){
		this.expression = expr;
	}

	public void setNumberOfPossibleOutputs() {
		/** figure out how many tuples to dequeue from left and right queue */
		XATQueue lQueue = this.getLeftInputQueue();
		XATQueue rQueue = this.getRightInputQueue();

		int lHowmany = Integer.parseInt(lQueue.getProperty("DEQUEUED_TUPLE_COUNT"));
		int rHowmany = Integer.parseInt(rQueue.getProperty("DEQUEUED_TUPLE_COUNT"));

		int lStateSize = states[0].getStateSize();
		int rStateSize = states[1].getStateSize();
		
		int possibleOutputs = lHowmany * rStateSize + rHowmany * lStateSize + lHowmany * rHowmany;
		
		try{
			this.setProperty("NUMBER_OF_POSSIBLE_OUTPUTS", ""+possibleOutputs);
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (6/3/2003 1:40:53 AM)
	 * @param operators edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator[]
	 */
	public void setOldBoxOperators(XATStreamOperator[] operators) {

		allOldOperators = operators;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (3/15/2003 10:24:23 PM)
	 * @param newWindow int
	 */
	public void setWindowSize(long newWindow) {

		//join operator has the same timeWindow with both its states
		this.windowSize = newWindow;
		if(states != null) {
			((WOperatorState)states[0]).setTimeWindow(newWindow);
			((WOperatorState)states[1]).setTimeWindow(newWindow);
		}
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (2/14/2003 9:33:59 AM)
	 * @return java.lang.String
	 */
	public String toString() {
		String rs = new String();

		String sName = this.getClass().getName();
		//find the last dot.
		int index = 0;
		for(index = sName.length()-1; index >=0; index--)
			if(sName.charAt(index) == '.')
				break;
		rs = sName.substring(index + 1) + " ";
		String id = this.getProperty("OBJECT_ID");
		if(id == "-1"){
			rs = rs + " " + this.hashCode(); 
		} else{
			rs = rs + id + " " +this.hashCode();
		}
	
		if(this.isActive()) rs += " ACTIVE ";
		else rs+= " INACTIVE ";
		return rs;
	}
	
	public String toToolTipString() {
		String rs = new String();

		String sName = this.getClass().getName();
		//find the last dot.
		int index = 0;
		for(index = sName.length()-1; index >=0; index--)
			if(sName.charAt(index) == '.')
				break;
		rs = sName.substring(index + 1) + " ";
		String id = this.getProperty("OBJECT_ID");
		if(id == "-1"){
			rs = rs + hashCode(); 
		} else{
			rs = rs + id;
		}	
		if(this.isActive()) rs += " ACTIVE ";
		else rs+= " INACTIVE ";

		/** now add the join predicate and window size info */
		rs += "\n";
		rs += "predicate: " + expression +"\n";
		rs += "window size = "+windowSize+"\n";

		return rs;
	}
	
	/* 
	 * process all accumulated tuples in input queues that have their maxtimestamp 
	 * smaller than the given maxtimestamp 
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		if ((this.getLeftInputQueue().getSchema() == null) || (this.getRightInputQueue().getSchema() == null)){
				return false;
		}
		if(this.getOutputQueue().getSchema() == null){
			init();
		}
		
		/** figure out how many tuples to dequeue from left and right queue */
		XATQueue lQueue = this.getLeftInputQueue();
		XATQueue rQueue = this.getRightInputQueue();
		int size1 = lQueue.getQueueSize(this);
		int size2 = rQueue.getQueueSize(this);
		int lHowmany=size1; 
		int rHowmany=size2;

		if(lQueue.isOldBoxInput()) {
			XATTuple[] tuples = lQueue.peek(size1, this);
			for (int i = 0; i < size1; i++) {
				if (((XATTuple) tuples[i]).getMaxTimeStamp() >= maxtimestamp) {
					lHowmany = i;
					break;
				}
			}
		}
		
		if(rQueue.isOldBoxInput()) {
			XATTuple[] tuples = rQueue.peek(size2, this);
			for (int i = 0; i < size2; i++) {
				if (((XATTuple) tuples[i]).getMaxTimeStamp() >= maxtimestamp) {
					rHowmany = i;
					break;
				}
			}
		}

		if((lHowmany + rHowmany) == 0) 
			return false;
			
		if (debugFlag) {
			System.out.println("dequeue " + lHowmany + "tuples from left queue");
			System.out.println("dequeue " + rHowmany + "tuples from right queue");
		}
			
		execOperator(lHowmany, rHowmany);

		/* run successfully */
		return true;
	
	}
	
	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() {
		if ((this.getLeftInputQueue().getSchema() == null) || (this.getRightInputQueue().getSchema() == null)){
				return false;
		}
		if(this.getOutputQueue().getSchema() == null){
			init();
		}
		
		/** figure out how many tuples to dequeue from left and right queue */
		XATQueue lQueue = this.getLeftInputQueue();
		XATQueue rQueue = this.getRightInputQueue();
		int lHowmany = lQueue.getQueueSize(this);
		int rHowmany = rQueue.getQueueSize(this);

		/** nothing to clean */
		if(lHowmany == 0 && rHowmany == 0) {
			return false;
		}

		if (debugFlag) {
			System.out.println("dequeue " + lHowmany + "tuples from left queue");
			System.out.println("dequeue " + rHowmany + "tuples from right queue");
		}
			
		execOperator(lHowmany, rHowmany);

		/* run successfully */
		return true;	
	}
	
	/*
	 * recompute the unmatched states inside this operator.
	 * Usually the unmatched states are computed based on the matched states inside one of the 
	 * children operator of this operator.
	 * 
	 * return: 
	 * TRUE -- if the states (inside child operator) that the recomputation of the unmatched states are based on are 
	 *      all matched, and the recomputation is successful.
	 * FALSE -- if any of the states (inside child operator) that are used to recompute the unmatched states
	 *          is unmatched, or the recomputation is unsucessful.  
	 */
	public boolean recomputeUnmatchedState() {
		boolean returnvalue = true;
		
		try {
			
		/* if left state is unmatched */
		if (states[0].isMatched() == false) {
			
			if(this.getNode().getOpChildrenSize() < 1) 
				throw new Exception("operator with unmatched state must have at least one child!");
			
			XATStreamOperator childop = this.getNode().getOpChild(0).getStreamOperator();
			
			if(!childop.getOutputQueue().equals(this.getNthInputQueue(0)) || !(childop instanceof WBinaryJoinOperator)) {
				throw new Exception("Error recomputing unmatched state!");
			} 
			
			WOperatorState lstate = (WOperatorState)childop.getState()[0];
			WOperatorState rstate = (WOperatorState)childop.getState()[1];
			
			if(lstate.isMatched() && rstate.isMatched()) {
			
				XATTuple[] rstatetuples = rstate.getAllStateTuples();
				XATTuple[] joinedtuples;
			
				joinedtuples = lstate.probe(rstatetuples, this.expression, 0);
				((WOperatorState)states[0]).insertTuple(joinedtuples);
				states[0].setMatched(true);
			} else {
				returnvalue = false;
			} 			
		}
		
		/* if right state is unmatched */
		if (states[1].isMatched() == false) {
			
			if(this.getNode().getOpChildrenSize() < 1) 
				throw new Exception("operator with unmatched state must have at least one child!");
			
			XATStreamOperator childop = this.getNode().getOpChild(0).getStreamOperator();
			
			if(childop.getOutputQueue() != this.getNthInputQueue(1)) {
				
				childop = this.getNode().getOpChild(1).getStreamOperator();
				if(childop.getOutputQueue() != this.getRightInputQueue())
					throw new Exception("Error recomputing unmatched state!");
			}	
			
			if(!(childop instanceof WBinaryJoinOperator)) {
				throw new Exception("Error recomputing unmatched state!");
			} 
			
			WOperatorState lstate = (WOperatorState)childop.getState()[0];
			WOperatorState rstate = (WOperatorState)childop.getState()[1];
			
			if(lstate.isMatched() && rstate.isMatched()) {
			XATTuple[] rstatetuples = rstate.getAllStateTuples();
			XATTuple[] joinedtuples;
			
			joinedtuples = lstate.probe(rstatetuples, this.expression, 0);
			((WOperatorState)states[1]).insertTuple(joinedtuples);
			states[1].setMatched(true);
			} else {
				returnvalue = false;
			}
			
		}	
	} catch (Exception e) {
			System.out.println("Something wrong when recomputing unmatched states. ");
			e.printStackTrace();
			return false;
	}
	return returnvalue;
	
	}

	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() {
		if (states == null) {
			return false;
		}

		for (int i = 0; i < states.length; i++) {
			if (!states[i].isMatched())
				return true;
		}
		return false;
	}

	
	/*
	 * Assign new ID to each state inside this operator, starting from the given startID.
	 * 
	 * return the next available ID. For example, if startID is 2, and there are 2 states inside this operator,
	 * then these two states are assigned new IDs as 2 and 3, and 4 is then returned as the next available stateID.
	 */
	  public int setStateID(int startID) {
		if (states == null) {
			return startID;
		}
	
		for(int i=0; i<states.length; i++) {
			states[i].setStateID(startID);
			startID++;
		}
		return startID;
	  }
	  

	  public boolean matchStates(XATStreamOperator newboxoperator) {
	  	int thisSize = this.states.length;
	  	int thatSize = newboxoperator.getState().length;
	  	OperatorState[] thatstates = newboxoperator.getState();
	  	boolean hasmatchedstate = false;
	  	
	  	for(int i=0; i<thisSize; i++) {
	  		for(int j=0; j<thatSize; j++) {
	  			// found a matched state
	  			if(states[i].getStateID() == thatstates[j].getStateID()) {
	  				thatstates[j].setMatched(true);
	  				thatstates[j].copyStateContent(states[i]);
	  				hasmatchedstate = true;
	  			}
	  		}
	  	}
	  	
	  	return hasmatchedstate;
	  }
		public boolean finalizeOperation(){
			return true;
		}
	
}


