/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleNegative;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleNegativeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.pipestate.PipeWFIFOState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WFIFOStateDeltaRecompute;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.operator.Juggler.ComparableTupleValues.KeyInt;

public class WindowBinaryJoinOperatorD2C extends
XATMultiQueueWindowStreamOperatorImp  implements WindowBinaryDeltaRecomputeJoinOperator {

	private TuplesList outputQueueCache = null;//recompute complete result
	private TuplesList outputQueueCacheDelta= null; // delta result
	
	private boolean OutputModeFlag=true; //true means complete result, false means delta.
	private boolean InputModeFlag=false;  // true means complete result, false means delta.
	
	private long start = 0;
	
	
	protected StringBuffer stats_string;
	String stateType=null;
	private ExpressionStr expression;
	
	public WindowBinaryJoinOperatorD2C(XATNode init) {
		super(init);
		// TODO Auto-generated constructor stub
	}

	public ExpressionStr getExpression() {
		return this.expression;
	} 
	
	public synchronized void setOutputPositions(int i, int[] ithOutputPositions) {
		if (outputPositions == null){
			outputPositions = new int[this.getInputQueue().getSchemas().size()][];
		}  
		outputPositions[i] = ithOutputPositions;
		}
	/*
	 * overwritten init function to deal with the multiple schema in the input
	 * queue.
	 */
	public void init() {
		XATQueueSchema outSchema = this.getOutputQueue().getSchema();
		if (outSchema == null || outSchema.getLength() == 0) {
			String schemaName = new String();
			outSchema = new XATQueueSchema();
			//Here we expect 2 inputQueues existing
			this.getInputQueue().addSchema(this.getInputQueue(0).getSchema());
			this.getInputQueue().addSchema(this.getInputQueue(1).getSchema());
			Iterator it = this.getInputQueue().getSchemas().iterator();
			System.out.println("inputSchemas "+ this.getInputQueue().getSchemas().size());
			while (it.hasNext()) {
				XATQueueSchema inSchema = (XATQueueSchema) it.next();
				schemaName += "#" + inSchema.getName();
				
				Iterator j = inSchema.getColNames().iterator();
				while(j.hasNext()){
					outSchema.addColumn((String) j.next());
				}
			}

			outSchema.setName(schemaName);
			System.out.println("output schema= "+ outSchema.toString());
			this.getOutputQueue().setSchema(outSchema);
		}
		computeOutputPositions();

		this.numOutputColsPerStream = new int[outputPositions.length];
		this.numOutputColumns = 0;
		//System.out.println("outputpositions "+outputPositions.length);
		for (int i=0; i<outputPositions.length; i++) {
			int numCols = outputPositions[i].length;
			System.out.println("outputpositions "+outputPositions[i].length);
			numOutputColsPerStream[i] = numCols;
			numOutputColumns += numCols;
		}
		System.out.println("# output columns" + numOutputColumns);
		
		//System.out.println("number of output columns per stream"+ numOutputColsPerStream.toString());
		
		this.setHasInit(true);
	}
@Override
public boolean run() {

	
	XATTuple leftTuple = null;
	XATTuple[] leftTuples = null;
	XATTuple rightTuple = null;
	XATTuple[] rightTuples = null;
	long timestart = System.nanoTime();
	try {
		if ((this.getLeftInputQueue().getSchema() == null)
				|| (this.getRightInputQueue().getSchema() == null)) {
			return false;
		}

		System.out.println("entered join operator");
			if(this.getOutputQueue().getSchema() == null)
				init();
//				this.getOutputQueue().setSchemas(this.getInputQueue().getSchemas());
				
//		return false;



		/* Load the current time_window and increment lengths */
		this.increment = 1000;
		this.windowSize = 3000;

		//hopping through increment
		//left state
	//	List purgedByHopTuple = ((WFIFOStateDeltaRecompute) states[0]).purgeByTimeStampSelf(this.start+this.increment);

		
		
	
//		Hopping through right state
		//List purgedByHopTupleR = ((WFIFOStateDeltaRecompute) states[1]).purgeByTimeStampSelf(this.start+this.increment);

	
	/** figure out how many tuples to dequeue from left and right queue */
		XATQueue lQueue = this.getLeftInputQueue();
		XATQueue rQueue = this.getRightInputQueue();
		int size1 = lQueue.getQueueSize(this);
		int size2 = rQueue.getQueueSize(this);
		int lHowmany = 0, rHowmany = 0;

		int dequeueHowmany = Integer.parseInt(this
				.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		System.out.println(this + "tuple to dequeue " + dequeueHowmany);

		int totalOutput =0;

		//left queue is empty. we only deal with tuples from right queue
		//here we assume the tuple from left and right will arrive in order
		if (size1 <= 0 && size2 > 0) {
			lHowmany = 0;
			rHowmany = (dequeueHowmany > size2) ? size2 : dequeueHowmany;
			try {
				lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
				rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);
			}

			//begin exp stat collect
			stats_string.append(System.currentTimeMillis()+"\t"); 

			rightTuples = rQueue.dequeue(rHowmany, this);
			System.out.println("processing right tuples (nothing on the left!)");
			for (int x = 0; x < rightTuples.length; x++) {
				totalOutput+=this.processRightTuple(rightTuples[x]);
			}
			stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

			return true;
		}

		//right queue is empty. we only deal with tuples from left queue
		//here we assume the tuple from left and right will arrive in order
		if (size2 <= 0 && size1 > 0) {
			rHowmany = 0;
			lHowmany = (dequeueHowmany > size1) ? size1 : dequeueHowmany;
			try {
				lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
				rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
				setNumberOfPossibleOutputs();
			} catch (Exception e) {
				System.out.println(e);
			}
			//begin exp stat collect
			stats_string.append(System.currentTimeMillis()+"\t"); 

			leftTuples = lQueue.dequeue(lHowmany, this);
			System.out.println("processing left tuples (nothing on the right!)");
			for (int x = 0; x < leftTuples.length; x++) {
				totalOutput+=this.processLeftTuple(leftTuples[x]);
			}
			stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

			return true;
		}

		//Both queues are not empty.
		//here we assume the tuple from left and right will arrive in order
		if (size2 > 0 && size1 > 0) {
			lHowmany = 0;
			rHowmany = 0;
			int total = (dequeueHowmany > (size1 + size2)) ? (size1 + size2)
					: dequeueHowmany;

			//begin exp stat collect
			stats_string.append(System.currentTimeMillis()+"\t"); 

			for (int x = 0; x < total; x++) {
				if (lHowmany < size1 && leftTuple == null) {
					leftTuple = lQueue.dequeue(this);
					lHowmany++;
				}
				if (rHowmany < size2 && rightTuple == null) {
					rightTuple = rQueue.dequeue(this);
					rHowmany++;
				}
				if (rightTuple!=null&&leftTuple == null
						|| rightTuple!=null&&leftTuple.getMaxTimeStamp()
						- rightTuple.getMaxTimeStamp() >= 0) {
					System.out.println("processing right tuples first!");
					totalOutput+=this.processRightTuple(rightTuple);
					rightTuple = null;
				} else if (leftTuple!=null&&rightTuple == null
						|| leftTuple!=null&&rightTuple.getMaxTimeStamp()
						- leftTuple.getMaxTimeStamp() > 0) {
					System.out.println("processing left tuples second");
					totalOutput+=this.processLeftTuple(leftTuple);

					leftTuple = null;
				}
			}
		}

		/* Change start to reflect the new beginning of the queue windows */ 
		
		stats_string.append(System.currentTimeMillis()+"\t"+ totalOutput + "\n");	

		try{
			lQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + lHowmany);
			rQueue.setProperty("DEQUEUED_TUPLE_COUNT", "" + rHowmany);
			setNumberOfPossibleOutputs();
		} catch (Exception e) {
			System.out.println(e);


		}
	}catch(Exception e1) {
		e1.printStackTrace();
	}   
	//System.out.print("RunTime");
    File file = new File ("runtimeJOperatorWithConvertion0.1k10.txt");
	FileWriter file_writer;
	try {
		file_writer = new FileWriter (file,true);
		file_writer.append(System.nanoTime()-timestart +"    \n");
		
		 file_writer.close();
	} catch (IOException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
		
	return true;

}

private int processLeftTuple(XATTuple tupleL) {
	XATTuple[] joinedtuples = null;
	XATTupleNegative tuple = null;
	XATTuple[] prevJoinedNegTuples = null;
	
	
	
	int ret=0;
	if(!(tupleL instanceof XATTupleNegative)){
		tuple = new XATTupleNegativeImp(tupleL);
		System.out.println("LEFT tuple is normal XATTuple");
		tuple.setNegativeFlag(false);
		}
		else{
			System.out.println("LEFT tuple is a XATTupleNegative");
			tuple = (XATTupleNegative)tupleL;
		}
	
	List purgedTuples = ((WFIFOStateDeltaRecompute) states[0]).purgeByTuplePipeSelf(tuple);
	System.out.println("purged tuples: "+ purgedTuples.size());
    //XATTupleNegative tuple = new XATTupleNegativeImp(tupleL);//convert each tuple in the operator to negative tuple......
	if (this.outputQueueCache == null)
		this.outputQueueCache = new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

	if (this.outputQueueCacheDelta == null)
		this.outputQueueCacheDelta = new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

//	output at every increment (HOP)
	if(tuple.getMinTimeStamp()>this.start+this.windowSize)
	{
		long win = this.start+ this.windowSize;
		System.out.println("!!!HOPPING!!!as "+ tuple.getMinTimeStamp() +" > "+ win);
		XATTuple[] leftStateTuples = ((WFIFOStateDeltaRecompute) states[0]).getAllStateTuples();
		
		for(int i=0;i< leftStateTuples.length;i++){
		
		XATTuple leftStateTuple=	leftStateTuples[i];
		
		if((((XATTupleNegative)leftStateTuple).isNegativeFlag()&& ((XATTupleNegative)leftStateTuple).isAdded()) || !((XATTupleNegative)leftStateTuple).isNegativeFlag())//if negative tuple then only if its positive was added, else 
		{
			//System.out.println("Probing the right state with "+ leftStateTuple.toString());
			joinedtuples = ((WFIFOStateDeltaRecompute) states[1]).probe(leftStateTuple,this.getExpression(), 0);
			ret +=joinedtuples.length;
			System.out.println("Probing the right state with "+ leftStateTuple.toString() + "returned tuples = " + joinedtuples.length);
			//this.getOutputQueue().enqueue(joinedtuples,this);
			//this.outputQueueCache.addToEnd(t)
			for(int j =0;j<joinedtuples.length;j++)
		    {
			System.out.println("populating the outputqueuecache from LEFT PROBING RIGHT! WITH "+joinedtuples[j]);
			this.outputQueueCache.addToEnd(joinedtuples[j]);
			//XATTupleNegativeImp tn = new XATTupleNegativeImp(joinedtuples[i]);
			//tn.setNegativeFlag(true);
			//outputQueueCacheDelta.addToEnd(tn, true);
			//this.outputQueueCacheDelta.addToEnd(joinedtuples[j],true);
			//this.getOutputQueue().enqueue(joinedtuples[j]);
		}
		}
		}
		
		/*TuplesList toEnqueue ;//= new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

		if (isOutputModeFlag()){
			toEnqueue = outputQueueCache;	
			
		}
		else
		{
			toEnqueue = outputQueueCacheDelta;
			
		}*/
		
		ListIterator cacheIt = this.outputQueueCache.listIterator();
		System.out.println("# of joined results "+this.outputQueueCache.size());
		while (cacheIt.hasNext())
		{
			XATTuple t = (XATTuple) cacheIt.next();
			System.out.println(t.toString());
			this.getOutputQueue().enqueue(t, this);
			
			
		}
		
		System.out.println("outputQueue tuples "+this.getOutputQueue().size());
		//put dummy tuples to distinguish between hops
		XATTuple tDummy = new XATTupleImp();
		KeyInt data=new KeyInt(-1);  
	       tDummy.addCell( data);
		System.out.println("added dummy tuple to the output: LEFT");
	    this.getOutputQueue().enqueue(tDummy,this);

	    System.out.println("flushing output and state: LEFT");
	    int countNegTuplesL =((WFIFOStateDeltaRecompute)this.states[0]).purgeAllNegativeTupleSelf();
	    int countNegTuplesR =((WFIFOStateDeltaRecompute)this.states[1]).purgeAllNegativeTupleSelf();
	    List purgedNegativeTuplesL =((WFIFOStateDeltaRecompute)this.states[0]).purgeByTimeStampSelf(this.getIncrement());
	    List purgedNegativeTuplesR =((WFIFOStateDeltaRecompute)this.states[1]).purgeByTimeStampSelf(this.getIncrement());
	    
	    System.out.println("left tuples purged on HOP: "+ purgedNegativeTuplesL.size());
	    System.out.println("right tuples purged on HOP: "+ purgedNegativeTuplesR.size());
	    outputQueueCache.clear();
		outputQueueCacheDelta.clear();

		((WFIFOStateDeltaRecompute)this.states[0]).setStartWindow(this.start + this.increment);
		((WFIFOStateDeltaRecompute)this.states[1]).setStartWindow(this.start + this.increment);
		start += this.increment;
		
		//return ret;
	}
	if(!isInputModeFlag()){ 
		if(tuple instanceof XATTupleNegative) //step 0: processing the negative tuple from the input queue
		{
			if(((XATTupleNegative)tuple).isNegativeFlag()){
			//outputQueueCacheDelta.addToEnd(tuple);
			tuple.setToNew(true);
			System.out.println("negative tuple on Left!");
			List purgedByNegativeTuple = ((WFIFOStateDeltaRecompute) states[0]).purgeByNegativeTupleSelf((XATTupleNegative)tuple);

			ListIterator cacheIt = purgedByNegativeTuple.listIterator();
			while (cacheIt.hasNext())
			{
				System.out.println("found corr tuple to negative tuple: LEFT");
				XATTuple t = (XATTuple) cacheIt.next();
				if (((XATTupleNegative)t).isAdded()){//if the corresponding positive tuple has contributed to join result, set the negative tuple as added!
					
					((XATTupleNegative)tuple).setAdded(true);
					System.out.println("corr tuple was added!: LEFT");
					break;
				}
			}
			
			/*prevJoinedNegTuples = ((WFIFOStateDeltaRecompute) states[1]).probe(tuple,	this.getExpression(), 0);
			
			for (int i=0;i< prevJoinedNegTuples.length;i++){
				outputQueueCache.remove(prevJoinedNegTuples[i]);
				System.out.println("\n  invalidating due to negative tuple " +prevJoinedNegTuples[i]);
			   }*/
		    }
					    
			
			if(!isOutputModeFlag()||!((XATTupleNegative)tuple).isNegativeFlag()){ //either positive tuple or negative tuple in Delta complete mode
			((WFIFOStateDeltaRecompute) states[0]).insertTuple(tuple);
			System.out.println("inserting the tuple into the state: LEFT");
		    }
		  }
			return 0;
		}
	return ret;

}



private int processRightTuple(XATTuple tupleR) {
	XATTuple[] joinedtuples = null;
	XATTupleNegative tuple = null;
	XATTuple[] prevJoinedNegTuples = null;

	
	int ret=0;
//	convert each tuple in the operator to negative tuple......
	if(!(tupleR instanceof XATTupleNegative)){
		System.out.println("RIGHT: tuple is normal XATTuple");
	tuple = new XATTupleNegativeImp(tupleR);
	tuple.setNegativeFlag(false);
	}
	else{
		tuple = (XATTupleNegative)tupleR;
		System.out.println("RIGHT: tuple is negative tuple");
	}
	
	List purgedTuples = ((WFIFOStateDeltaRecompute) states[1]).purgeByTuplePipeSelf(tuple);
	System.out.println("purged tuples: "+ purgedTuples.size());
	
	if (this.outputQueueCache == null)
		this.outputQueueCache = new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

	if (this.outputQueueCacheDelta == null)
		this.outputQueueCacheDelta = new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

//	output at every increment (HOP)
	if(tuple.getMinTimeStamp()>this.start+this.windowSize)
	{
		long win = this.start+ this.windowSize;
		System.out.println("HOPPING!!! as "+tuple.getMinTimeStamp()+" > "+ win);
		XATTuple[] rightStateTuples = ((WFIFOStateDeltaRecompute) states[1]).getAllStateTuples();
		
		for(int i=0;i< rightStateTuples.length;i++){
		
		XATTuple rightStateTuple=	rightStateTuples[i];
		
		if((((XATTupleNegative)rightStateTuple).isNegativeFlag()&& ((XATTupleNegative)rightStateTuple).isAdded()) || !((XATTupleNegative)rightStateTuple).isNegativeFlag())//if negative tuple then only if its positive was added, else 
		{
			
		joinedtuples = ((WFIFOStateDeltaRecompute) states[0]).probe(rightStateTuple,
				this.getExpression(), 0);
		ret +=joinedtuples.length;
		System.out.println("probing left state with right tuple "+ rightStateTuple.toString() +"found joins "+ joinedtuples.length);
		
		//this.getOutputQueue().enqueue(joinedtuples,this);
		for(int j =0;j<joinedtuples.length;j++)
		{
			System.out.println("populating the outputqueuecache from RIGHT PROBING LEFT! WITH "+ joinedtuples[j]);
			this.outputQueueCache.addToEnd(joinedtuples[j]);
			//XATTupleNegativeImp tn = new XATTupleNegativeImp(joinedtuples[i]);
			//tn.setNegativeFlag(true);
			//outputQueueCacheDelta.addToEnd(tn, true);
		//	this.outputQueueCacheDelta.addToEnd(joinedtuples[j],true);
			//this.getOutputQueue().enqueue(joinedtuples[j], this);

		}
		}
		}
		/*TuplesList toEnqueue ;//= new TuplesList(this.getOutputQueue().getSchema(), "timestamp");

		if (isOutputModeFlag()){
			toEnqueue = outputQueueCache;	
		}
		else
		{
			toEnqueue = outputQueueCacheDelta;
			
		}*/

		ListIterator cacheIt = this.outputQueueCache.listIterator();
		System.out.println("# of joined results "+this.outputQueueCache.size());
		while (cacheIt.hasNext())
		{
			XATTuple t = (XATTuple) cacheIt.next();
			System.out.println(t.toString());
			this.getOutputQueue().enqueue(t, this);
			
			
		}
		//System.out.println("outputQueue tuples "+this.getOutputQueue().size());
		//put dummy tuples to distinguish between hops
		XATTuple tDummy = new XATTupleImp();
		KeyInt data=new KeyInt(-1);  
	       tDummy.addCell( data);
			System.out.println("added dummy tuple to the output: RIGHT");
	    this.getOutputQueue().enqueue(tDummy,this);
	    
	    int countNegTuplesL =((WFIFOStateDeltaRecompute)this.states[0]).purgeAllNegativeTupleSelf();
	    int countNegTuplesR =((WFIFOStateDeltaRecompute)this.states[1]).purgeAllNegativeTupleSelf();
	    List purgedNegativeTuplesL =((WFIFOStateDeltaRecompute)this.states[0]).purgeByTimeStampSelf(this.getIncrement());
	    List purgedNegativeTuplesR =((WFIFOStateDeltaRecompute)this.states[1]).purgeByTimeStampSelf(this.getIncrement());
	    System.out.println("left tuples purged on HOP: "+ purgedNegativeTuplesL.size());
	    System.out.println("right tuples purged on HOP: "+ purgedNegativeTuplesR.size());
	    outputQueueCache.clear();
		outputQueueCacheDelta.clear();
		System.out.println("flushing outputqueuecache: RIGHT");


		((WFIFOStateDeltaRecompute)this.states[0]).setStartWindow(this.start + this.increment);
		((WFIFOStateDeltaRecompute)this.states[1]).setStartWindow(this.start + this.increment);
		start += this.increment;
		
		//return ret;
	}
	if(!isInputModeFlag()){ //step 0: processing the negative tuple from the input queue
		if(tuple instanceof XATTupleNegative)
		{
			if(((XATTupleNegative)tuple).isNegativeFlag()&& isOutputModeFlag()){
			//outputQueueCacheDelta.addToEnd(tuple);
			tuple.setToNew(true);
			System.out.println("RIGHT: negative tuple");
			
			List purgedByNegativeTuple = ((WFIFOStateDeltaRecompute) states[1]).purgeByNegativeTupleSelf((XATTupleNegative)tuple);

			ListIterator cacheIt = purgedByNegativeTuple.listIterator();
			while (cacheIt.hasNext())
			{
				
				XATTuple t = (XATTuple) cacheIt.next();
				if (((XATTupleNegative)t).isAdded()){//if the corresponding positive tuple has contributed to join result, set the negative tuple as added!
					((XATTupleNegative)tuple).setAdded(true);
					System.out.println("RIGHT: found corr positive tuple that is ADDED");
					break;
				}
			}

			/*prevJoinedNegTuples = ((WFIFOStateDeltaRecompute) states[0]).probe(tuple,	this.getExpression(), 0);
			
			for (int i=0;i< prevJoinedNegTuples.length;i++){
				outputQueueCache.remove(prevJoinedNegTuples[i]);
				System.out.println("\n  invalidating due to negative tuple " +prevJoinedNegTuples[i]);
			   }*/
			}
			
			if(!isOutputModeFlag()||!((XATTupleNegative)tuple).isNegativeFlag()){ //either positive tuple or negative tuple in Delta complete mode
			((WFIFOStateDeltaRecompute) states[1]).insertTuple(tuple);
			System.out.println("inserting into RIGHT state");
			}
			
		  }
		
			return 0;
		}
	return ret;
}

public void printStatsToFile() {
	try {
		String fileName = new String("WBJoinD2C_" + System.currentTimeMillis()+".txt");
		FileWriter out = new FileWriter(fileName, true);
		out.write(this.stats_string.toString());
		out.close();

	} catch (IOException e) {
		System.out.println(e.toString());
	}
}

public boolean isInputModeFlag() {
	return InputModeFlag;
}

public void setInputModeFlag(boolean inputModeFlag) {
	InputModeFlag = inputModeFlag;
}

public boolean isOutputModeFlag() {
	return OutputModeFlag;
}

public void setOutputModeFlag(boolean outputModeFlag) {
	OutputModeFlag = outputModeFlag;
}

public boolean finalizeOperation() {
	// TODO Auto-generated method stub
	return false;
}

public boolean cleanAccumulatedTuples(long maxtimestamp) {
	// TODO Auto-generated method stub
	return false;
}

public boolean cleanAllAccumulatedTuples() {
	// TODO Auto-generated method stub
	return false;
}

public boolean hasUnmatchedState() {
	// TODO Auto-generated method stub
	return false;
}

public boolean matchStates(XATStreamOperator op) {
	// TODO Auto-generated method stub
	return false;
}

public boolean recomputeUnmatchedState() {
	// TODO Auto-generated method stub
	return false;
}

public String toToolTipString() {
	// TODO Auto-generated method stub
	return null;
}

public void enqueueKeptTuples() {
	// TODO Auto-generated method stub
	
}

public boolean getAllOldPurgedFlag() {
	// TODO Auto-generated method stub
	return false;
}

public OperatorState getLeftState() {
	// TODO Auto-generated method stub
	return ((WFIFOStateDeltaRecompute)this.states[0]);
}

public int getLeftStateSize() {
	// TODO Auto-generated method stub
	return ((WFIFOStateDeltaRecompute)this.states[0]).getStateSize();
}

public OperatorState getRightState() {
	// TODO Auto-generated method stub
	return ((WFIFOStateDeltaRecompute)this.states[1]);
}

public int getRightStateSize() {
	// TODO Auto-generated method stub
	return ((WFIFOStateDeltaRecompute)this.states[1]).getStateSize();
}


public int runForMigration1() {
	// TODO Auto-generated method stub
	return 0;
}

public int runForMigration2(long latestTimeStamp) {
	// TODO Auto-generated method stub
	return 0;
}

public void setOldBoxOperators(XATStreamOperator[] operators) {
	// TODO Auto-generated method stub
	
}
public void setExpression(ExpressionStr _expression){
	this.expression = _expression;
}

public void initState(String statetype) {
		this.stateType = statetype;
		try {
			//Constructor con = Class.forName(statetype).getConstructor();
			states = new WFIFOStateDeltaRecompute[2];
			states[0] = (WFIFOStateDeltaRecompute) Class.forName(statetype).newInstance();
			states[1] = (WFIFOStateDeltaRecompute) Class.forName(statetype).newInstance();
			states[0].setOperator(this);
			states[1].setOperator(this);
			if (this.startPointArray[0] > 0)
				((WFIFOStateDeltaRecompute) states[0])
						.setStartWindow(this.startPointArray[0]);
			if (this.startPointArray[1] > 0)
				((WFIFOStateDeltaRecompute) states[1])
						.setStartWindow(this.startPointArray[1]);

			if (this.windowSizeArray[0] > 0) {
				((WFIFOStateDeltaRecompute) states[0])
						.setTimeWindow(this.windowSizeArray[0]);
				
			}
			if (this.windowSizeArray[1] > 0) {
				((WFIFOStateDeltaRecompute) states[1])
						.setTimeWindow(this.windowSizeArray[1]);
				
			}
			((WFIFOStateDeltaRecompute) states[0]).init(0);
			((WFIFOStateDeltaRecompute) states[1]).init(1);
			
			this.stats_string = new StringBuffer();
			//stats_string.append("StartTime\tEndTime\tOutputTuples\n");

		
		} catch (Exception e) {
			e.printStackTrace();
		}
	
}
}