//--------------------------------------------------
// SampleStatsCollectorOperator.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.MotionStatisticsGatherer;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.ClassType;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.QueryType;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.UpdateType;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GenericSpatioTemporalGrid;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridCellIndex3D;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.MotionGenericOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.MotionDiscreteResult;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MotionDiscreteState;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.Hashtable;
//import java.io.IOException;
//import java.util.ArrayList;
//import java.util.Enumeration;
//import java.util.HashMap;
//import java.util.Map;

@SuppressWarnings({ "rawtypes", "serial" })
public class SampleStatsCollectorOperatorImp extends
		XATSingleQueueStreamOperatorImp implements SampleStatsCollectorOperator {

	// -----------------------------------------------------------------------------

	int totalTupleCounter = 0;
	int window_size;
	Coin flipCoin;

	static int counterLoops = 1;
	protected Time currentTime;
	private int totalTupleCounterSamples = 0; // tuples

	private int elementsInputQ = 0;
	int resultTuples;

	protected String samplingHeuristic;
	protected int numTuplesPerWindowToSample;

	protected static int numTuplesInSample;
	protected int numWindowsToSample;

	Hashtable histogramsTable = new Hashtable();
	String sampleDumpFile;
	Hashtable usedNotUsedAttributes = new Hashtable();

	// -----------------------------------------------------------------------------

	public SampleStatsCollectorOperatorImp(XATNode init) {
		super(init);
		flipCoin = new Coin();
	}

	// -----------------------------------------------------------------------------

	static int numWindowsAlreadySampled = 0;

	// -----------------------------------------------------------------------------

	public boolean run() throws Exception {
		// if sampled the desired number of windows, exit
		if (numTuplesInSample == numWindowsToSample
				* numTuplesPerWindowToSample)// (numWindowsAlreadySampled == numWindowsToSample)
		{
			System.out.println("Total # of windows sampled: "
					+ numWindowsAlreadySampled);
			System.out.println("Total # of tuples sampled: "
					+ numTuplesInSample);
			System.out.println("Done sampling.");
			System.in.read();
			System.exit(0);
		}

		if (this.getDebug()) {
			System.out.println("Running: " + this.toToolTipString() + "\n");
		}

		for (int i = 0; i < this.getOutputQueueCount(); i++) {
			if (this.getOutputQueue(i).getSchema() == null) {
				init();
			}
		}

		XATTuple inputTuple = null;
		int tupleCount = 0;
		int howMany = numTuplesPerWindowToSample;
		elementsInputQ = this.getInputQueue().getQueueSize();

		numWindowsAlreadySampled++;

		while (true) {
			if (tupleCount == howMany) /* done with processing */
			{
				break;
			}

			// pick the next tuple to process
			if (this.getInputQueue().getQueueSize(this) > 0) {
				inputTuple = this.getInputQueue().dequeue(this);

				if (samplingHeuristic.equalsIgnoreCase("simple_random")) {
					totalTupleCounter++;

					// --------------------------------------------------
					// SAMPLE BASED ON HEURISTIC
					// --------------------------------------------------

					flipCoin.flip();

					// on heads, sample the tuple, otherwise ignore it
					if (flipCoin.toString().equals("Heads")) {
						// write the tuple being sampled to a sample dump file
						// (this file will be fed to decision tree constructor)

						String currentTupleStr = CreateSampleTupleString(
								inputTuple, histogramsTable);
						// form a tab-spaced tuple string

						if (!currentTupleStr.equalsIgnoreCase("")) {
							WriteToSampleDumpFile(sampleDumpFile,
									currentTupleStr);
							totalTupleCounterSamples++;
							numTuplesInSample++;
							tupleCount++;
						}

					}// if heads was picked
				}// if sampling_heuristic == "simple_random"
			}// end if input size > 0

			// System.out.println("Total # of tuples sampled: "+
			// totalTupleCounterSamples);
			// System.out.println("Total # of tuples: "+ totalTupleCounter);

			counterLoops++;

			// if got the right number of tuples' samples, break
			if (totalTupleCounterSamples == numTuplesPerWindowToSample) {
				return true; // done sampling
			}

		}// end while(true)

		return true;
	}

	// -----------------------------------------------------------------------------

	public static String CreateSampleTupleString(XATTuple inputTuple,
			Hashtable histogramsTable) {
		String currentTupleStr = "";

		// add values to appropriate histograms
		XATDataValue[] arrValues = inputTuple.getArrayData();

		for (int r = 0; r < inputTuple.getLength(); r++) {
			// enumerate all the contents of the hashtable
			String key;
			BasicHist updateBasicHistogram;
			StringBasicHist updateStringBasicHistogram;

			key = "" + r;
			Object objectHistogram = histogramsTable.get(key);
			XATDataValueImp xatVal = (XATDataValueImp) arrValues[Integer
					.parseInt(key)];

			if (xatVal.getValue() == null) {
				continue;
			}

			if (!histogramsTable.containsKey(key)) {
				// output the actual value of the attribute
				currentTupleStr += inputTuple.getCell(r).toString()
						.replace(" ", "")
						+ "  "; // separator is a double-space
				continue;
			}

			// ---------------------------------------------------
			// check the type of the histogram
			// ---------------------------------------------------

			if (objectHistogram
					.getClass()
					.toString()
					.equals("class edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.BasicHist")) {
				// double type histogram
				updateBasicHistogram = (BasicHist) objectHistogram;

				// if an exception occurs, don't take the tuple for
				// the sample - data (especially real data, can be
				// sometimes dirty - and we decided not to address this
				// issue here
				try {
					System.out.println("for key [" + key.toString()
							+ "] value [" + xatVal + "]");
					String strVal = xatVal.convertToString();
					strVal = strVal.replace("%", "");
					double val = Double.parseDouble(strVal);
					// add the current sample value to the histogram
					updateBasicHistogram.add(val);

					// now retrieve the bucket value to output to the sample
					// dump
					String binNumber = "bin"
							+ updateBasicHistogram.getBinByValue(val) + ":"
							+ updateBasicHistogram.getBinRangesByValue(val);

					currentTupleStr += binNumber + "  ";
				} catch (Exception ex) {
					String msg = ex.getMessage();
					System.out.println("failed " + msg);
				}
			}

			// ------------------------------------------------------------
			// STRING HISTOGRAM
			// ------------------------------------------------------------
			else {
				// expect histogram type - class
				// edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.StringBasicHist
				// string type histogram
				updateStringBasicHistogram = (StringBasicHist) objectHistogram;

				String val = xatVal.convertToString();

				// add the current sample value to the histogram
				updateStringBasicHistogram.add(val);

				// now retrieve the bucket value to output to the sample dump
				String binNumber = "bin"
						+ updateStringBasicHistogram.getBinByValue(val);

				currentTupleStr += binNumber + "  ";
			}
		}// for tuple length

		return currentTupleStr;
	}

	// -----------------------------------------------------------------------------

	private void WriteToSampleDumpFile(String fileLocation, String stringToWrite) {
		try {
			// Create file
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			out.newLine();
			// Close the output stream
			out.close();
		} catch (Exception e) {
			// Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}

	// -----------------------------------------------------------------------------

	public String toToolTipString() {
		return "Training Tuple Set Creator Operator - SampleStatsCollectorOperatorImp";
	}

	// -----------------------------------------------------------------------------

	public boolean hasUnmatchedState() {
		return false;
	}

	// -----------------------------------------------------------------------------

	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}

	// -----------------------------------------------------------------------------

	public boolean cleanAllAccumulatedTuples() {
		return true;
	}

	// -----------------------------------------------------------------------------

	public int setStateID(int startID) {
		return 0;
	}

	// -----------------------------------------------------------------------------

	public void setHistogramsTable(Hashtable newHistogramsTable) {
		histogramsTable = newHistogramsTable;
	}

	// -----------------------------------------------------------------------------

	public void setSamplingHeuristic(String samplingHeuristicVal) {
		samplingHeuristic = samplingHeuristicVal;
	}

	// -----------------------------------------------------------------------------

	public void setNumSampleWindows(int numSampleWindowsVal) {
		numWindowsToSample = numSampleWindowsVal;
	}

	// -----------------------------------------------------------------------------

	public void setTuplesPerWindowToSample(int numTuplesPerWindowToSampleVal) {
		numTuplesPerWindowToSample = numTuplesPerWindowToSampleVal;
	}

	// -----------------------------------------------------------------------------

	public void setTupleSampleDumpFile(String sampleDumpFileVal) {
		sampleDumpFile = sampleDumpFileVal;
	}

	// -----------------------------------------------------------------------------

	public boolean recomputeUnmatchedState() {
		return true;
	}

	// -----------------------------------------------------------------------------

	public boolean matchStates(XATStreamOperator op) {
		return true;
	}

	// -----------------------------------------------------------------------------

	public boolean finalizeOperation() {
		return true;
	}

	// -----------------------------------------------------------------------------

	public void setWindowSize(int newSize) {
		window_size = newSize;
	}

	// -----------------------------------------------------------------------------

	public int getWindowSize() {
		return window_size;
	}

	public int getElementsInputQ() {
		return elementsInputQ;
	}

	public void setElementsInputQ(int elementsInputQ) {
		this.elementsInputQ = elementsInputQ;
	}

	// -----------------------------------------------------------------------------

	
}
