package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import edu.wpi.cs.dsrg.STeM.STeMData;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.RouteSignatureComparator;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.RouteSignatureStorage;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;

//Karen Works
//Rimma V. Nehme - introduced some enhancements
//This is the OnLine Classifier used in Query Mesh model

@SuppressWarnings("serial")
public class OnlineClassifierOperatorImp extends
		XATMultiQueueWindowStreamOperatorImp implements
		OnlineClassifierOperator {
	// -----------------------------------------------------------------------------

	// String outFile =
	// "C:\\Rimma\\Research\\QueryMesh\\execution_stats\\RUNTIME_CLASSIFIER_COST.txt";
	long startTime;
	long endTime;
	long numProcessedTuples;

	private ExpressionStr expression;

	public int windowSize;

	// number of operators
	private int NumOps = 0;
	// number of streams
	private int NumSTREAMS = 0;

	private OperatorLookUp OpLkUp;

	// pointer to data stored for STEM
	private HashMap<Integer, STeMData> STeMStreamData;

	// decision tree used to evaluate tuple & create ruster
	ArrayList<node> edt;

	// Rusters per STREAM
	ArrayList<HashMap<Integer, RusterNode>> RusterPerStream;

	// intermediate list of ruster & tuples
	// created prior to sending
	ArrayList<ArrayList<XATTuple>> rusterList;

	// used to determine when to create ruster
	int SendOff;

	// If the other operators contain at least the TupleCountThreshold then
	// no new tuples are send from the classifier to ops.
	// used to determine if any new tuples need to be processes or not
	int TupleCountThreshold;

	// used to store original DT
	ArrayList<ArrayList<DecisionTreeNode>> streamDT;

	// stores window count for streams
	private HashMap<Integer, Integer> STREAMWindowCount;

	// -----------------------------------------------------------------------------

	public OnlineClassifierOperatorImp(XATNode init) {
		super(init);
		STREAMWindowCount = new HashMap<Integer, Integer>();
	}

	// -----------------------------------------------------------------------------

	public void InitializeVariables() {}

	// -----------------------------------------------------------------------------

	public boolean run() throws Exception {
		startTime = System.currentTimeMillis();

		for (int i = 0; i < this.getOutputQueueCount(); i++) {
			if (this.getOutputQueue(i).getSchema() == null) {
				init(i);
			}
		}

		// int howMany =
		// Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		int howMany = 5000; // rimma temp "hack"
		numProcessedTuples = 0;
		int firstOp = 0;

		// check to see if the threshold for the total number of tuples has been
		// reached
		int TotalNumTupInOpQueues = 0;

		for (int y = 0; y < NumOps; y++) {
			if (OpLkUp.getOperatorQueue(y) != null) {
				TotalNumTupInOpQueues += OpLkUp.getOperatorQueue(y)
						.getQueueSize();
			}
		}

		if (TotalNumTupInOpQueues >= TupleCountThreshold) {
			System.out
					.println("Threshold reached! Will NOT run the classifier until query processor will relieve their load. Returning...");
			return true;
		}

		// Scan for tuples coming from STREAMS
		for (int j = 0; j < NumSTREAMS
				&& (TotalNumTupInOpQueues < TupleCountThreshold); j++) {
			// clean out the list of rusters for each stream
			ArrayList<ArrayList<XATTuple>> rusterList = new ArrayList<ArrayList<XATTuple>>(
					RusterPerStream.get(j).size());

			for (int i = 0; i < RusterPerStream.get(j).size(); i++) {
				rusterList.add(new ArrayList<XATTuple>());
			}

			// ---------------------------------------------------------------
			// get the root of the stream decision tree (by stream id)
			// ---------------------------------------------------------------
			node root = edt.get(j);

			// print the root of the decision tree for the stream (for debugging
			// purposes)
			System.out.print("Stream: id = " + j);
			/*
			 * if (root.DefaultPath != null) {
			 * System.out.println(", Default route = " + root.DefaultPath); }
			 * else { System.out.println(", Default route = NULL"); }
			 */

			// ---------------------------------------------------------------
			// dequeue tuples
			// ---------------------------------------------------------------
			XATTuple[] tuples = (this.getInputQueue(j)).dequeue(howMany, this);

			System.out
					.println(", Online classifier, dequeued " + tuples.length);

			numProcessedTuples += tuples.length;

			QMeshTuple qmt;
			ArrayList<Integer> os; // operator stack

			// ---------------------------------------------------------------
			// process dequeued tuples
			// ---------------------------------------------------------------
			for (int k = 0; k < tuples.length; k++) {
				// System.out.println("Classified " + k + " tuples so far.");
				tuples[k].setStreamID(j);

				qmt = new QMeshTupleImp(tuples[k], NumSTREAMS, j);
				TotalNumTupInOpQueues++;

				// add to stream data
				STeMData sd = STeMStreamData.get(tuples[k].getStreamID());
				sd.Insert(tuples[k]);

				// --------------------------------------------------------------------
				// CLASSIFY THE TUPLE
				// --------------------------------------------------------------------
				XATTuple currTuple = tuples[k];
				RusterNode classifiedRoute = classifyTuple(currTuple, root);

				// --------------------------------------------------------------------

				// send to output or next op?
				// System.out.println("stream = " + j);
				// System.out.println("classifiedRoute.RusterId = " +
				// classifiedRoute.RusterId);
				// System.out.println("rusterList size = " + rusterList.size());

				try {
					if (rusterList.get(classifiedRoute.RusterId).size() > 0) {
					}
				} catch (Exception ex) {
					classifiedRoute.RusterId = 0;
				}

				if (rusterList.get(classifiedRoute.RusterId).size() > 0) {
					// add tuple to existing ruster
					qmt = new QMeshTupleImp(tuples[k], NumSTREAMS, j);
					rusterList.get(classifiedRoute.RusterId).add(qmt);

					// if ruster reached maxsize send it off
					if (rusterList.get(classifiedRoute.RusterId).size() >= SendOff) {
						((RoutingToken) rusterList
								.get(classifiedRoute.RusterId).get(0))
								.SetNumTuples(rusterList.get(
										classifiedRoute.RusterId).size() - 1);
						OpLkUp.getOperatorQueue(classifiedRoute.os.get(0))
								.enqueue(
										rusterList
												.get(classifiedRoute.RusterId),
										this);
						rusterList.get(classifiedRoute.RusterId).clear();
					}
				} else {
					// create ruster first (ruster = r-token + group of tuples)
					rusterList.add(classifiedRoute.RusterId,
							new ArrayList<XATTuple>());

					String rTokenStreamId = "" + j;
					RoutingToken rt = new RoutingTokenImp(rTokenStreamId,
							classifiedRoute.GetRusterId());
					rt.setOpStackSize(classifiedRoute.os);

					// STQM - set the route id for this ruster
					/*
					 * rt.SetRouteId(classifiedRoute.GetRusterId());
					 * rt.setStreamID(j);
					 */
					rusterList.get(classifiedRoute.RusterId).add(rt);

					// add tuple
					qmt = new QMeshTupleImp(tuples[k], NumSTREAMS, j);
					rusterList.get(classifiedRoute.RusterId).add(qmt);
				}
			}// all tuples from the stream have been processed

			// send out any rusters from last stream
			if (rusterList != null) {
				for (int i = 0; i < rusterList.size(); i++) {
					if (rusterList.get(i).size() > 0) {
						RoutingToken rt = (RoutingToken) (rusterList.get(i)
								.get(0));
						int opId = rt.getTopOperator();
						ArrayList outTuples = rusterList.get(i);
						rt.SetNumTuples(outTuples.size() - 1);
						OpLkUp.getOperatorQueue(opId).enqueue(outTuples, this);
					}
				}
				rusterList = null;
			}
		}// done processing all streams
		System.gc();

		endTime = System.currentTimeMillis();
		long duration = endTime - startTime; // milliseconds

		System.out.println("numProcessedTuples: " + numProcessedTuples
				+ "\t time: " + String.valueOf(duration));
		// WriteToFile(outFile, numProcessedTuples + "\t" +
		// String.valueOf(duration), true);

		// update property for statistics monitoring
		this.setProperty("NUMBER_OF_TUPLES_TO_DEQUEUE", "" + numProcessedTuples);

		return true;
	}

	// -----------------------------------------------------------------------------

	private RusterNode classifyTuple(XATTuple currTuple, node root) {
		// ---------------------------------------------------------
		// classify tuple
		// ---------------------------------------------------------
		node e = new node(root.IdType, root.obj);
		ArrayList<Object> children = (ArrayList<Object>) e.obj.clone();
		RusterNode classifiedRoute = null;
		node n;

		boolean bFound = false;
		while (!bFound) {
			// ---------------------------------------------------------
			// if there are children, drill down
			// ---------------------------------------------------------
			if (children != null) {
				boolean bSearchNextLevel = false;

				for (int i = 0; i < children.size() && !bSearchNextLevel; i++) {
					if (e.IsEqualOperatorNode(i)) {
						EqualOperatorNode eqon = ((EqualOperatorNode) children.get(i));
						if ((n = eqon.FindMatch(currTuple)) != null) {
							// System.out.println("matched = clause"+tuples[k]);
							e = n;
							bSearchNextLevel = true;
						}
					} else if (e.IsRangeNode(i)) {
						RangeNode eqon = ((RangeNode) children.get(i));

						if ((n = eqon.FindMatch(currTuple)) != null) {
							e = n;
							bSearchNextLevel = true;
						}
					} else if (e.IsRuster(i)) {
						bFound = true;
						// get path
						classifiedRoute = ((RusterNode) children.get(i));
						bSearchNextLevel = true;
					}
				}// end for

				// no other levels were found that need to be searched
				if (!bSearchNextLevel) {
					bFound = true;
					// use default path

					ArrayList<Integer> oRoute = (ArrayList<Integer>) root.DefaultPath.os.clone();
					int RustId = root.DefaultPath.RusterId;
					// classifiedRoute = new RusterNode(root.DefaultPath.os,
					// root.DefaultPath.RusterId);

					classifiedRoute = new RusterNode(oRoute, RustId);
				} else if (!bFound) {
					children.clear();
					children = (ArrayList<Object>) e.obj.clone();
				}
			} else {
				// --------------------------------------------------------------------
				// children, use parent's node default route
				// --------------------------------------------------------------------
				bFound = true;
				// use default path
				classifiedRoute = new RusterNode(root.DefaultPath.os,
						root.DefaultPath.RusterId);
			}
		}
		return classifiedRoute;
	}

	// -----------------------------------------------------------------------------

	public void setWindowSize(String[] size) {
		windowSize = Integer.parseInt(size[0]);
	}

	// -----------------------------------------------------------------------------

	public void init(int i) {
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) {
			return;
		}

		XATQueueSchema outSchema = this.getOutputQueue(i).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) {
			this.getOutputQueue(i).setSchema(inSchema);
		}

		computeOutputPositions();// no need to overload since all the output queue are the same
		numOutputColumns = outputPositions[0].length;
		setHasInit(true);
	}

	// -----------------------------------------------------------------------------

	public String toToolTipString() {
		return "Online Classifier Operator (Query Mesh)";
	}

	// -----------------------------------------------------------------------------

	public ExpressionStr getExpression() {
		return expression;
	}

	// -----------------------------------------------------------------------------

	public void setExpression(ExpressionStr expression) {
		this.expression = expression;
	}

	// -----------------------------------------------------------------------------

	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input
	 * queues. Inside method run() the numbers are computed based on the
	 * property "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler, while
	 * in cleanAccumulatedTuples, all tuples in input queues that have their max
	 * timesamp smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}

	// -----------------------------------------------------------------------------

	public boolean cleanAllAccumulatedTuples() {
		return true;
	}

	// -----------------------------------------------------------------------------

	public boolean hasUnmatchedState() {
		return false;
	}

	// -----------------------------------------------------------------------------

	public boolean matchStates(XATStreamOperator op) {
		return false;
	}

	// -----------------------------------------------------------------------------

	public boolean recomputeUnmatchedState() {
		return true;
	}

	// -----------------------------------------------------------------------------

	public int setStateID(int nextid) {
		return nextid;
	}

	// -----------------------------------------------------------------------------

	public boolean finalizeOperation() {
		return true;
	}

	// -----------------------------------------------------------------------------

	// Sort of initialization
	public void setNumStreams(int Ns) {
		NumSTREAMS = Ns;

		RusterPerStream = new ArrayList<HashMap<Integer, RusterNode>>(Ns);
		streamDT = new ArrayList<ArrayList<DecisionTreeNode>>(Ns);

		for (int o = 0; o < Ns; o++) {
			RusterPerStream.add(new HashMap<Integer, RusterNode>());
			streamDT.add(new ArrayList<DecisionTreeNode>(Ns));
		}
	}

	// -----------------------------------------------------------------------------

	public int getNumStreams() {
		return NumSTREAMS;
	}

	// -----------------------------------------------------------------------------

	public void setNumOps(int NOps) {
		NumOps = NOps;
	}

	// -----------------------------------------------------------------------------

	public void setOperatorLookUp(OperatorLookUp z) {
		OpLkUp = z;
	}

	// -----------------------------------------------------------------------------

	public void setSTeMStreamData(HashMap<Integer, STeMData> sd) {
		STeMStreamData = sd;
	}

	// -----------------------------------------------------------------------------

	public void setExecutionDecisionTree(ArrayList<node> d) {
		edt = d;
	}

	// -----------------------------------------------------------------------------

	public void setSendOff(int so) {
		SendOff = so;
	}

	// -----------------------------------------------------------------------------

	public int getSendOff() {
		return SendOff;
	}

	// -----------------------------------------------------------------------------

	public void setTupleCountThreshold(int so) {
		TupleCountThreshold = so;
	}

	// -----------------------------------------------------------------------------

	public void setRusterPerStream(ArrayList<HashMap<Integer, RusterNode>> d) {
		RusterPerStream = d;
	}

	// -----------------------------------------------------------------------------

	public ArrayList<HashMap<Integer, RusterNode>> getRusterPerStream() {
		return RusterPerStream;
	}

	// -----------------------------------------------------------------------------

	public void AddRusterPerStream(int StreamId, int i, RusterNode d) {
		HashMap<Integer, RusterNode> hmp = RusterPerStream.get(StreamId);
		hmp.put(i, d);
	}

	// -----------------------------------------------------------------------------

	public void AddDTNodePerStream(int StreamId, int i, DecisionTreeNode d) {
		ArrayList<DecisionTreeNode> dt = streamDT.get(StreamId);
		while (dt.size() < (i + 1)) {
			dt.add(null);
		}
		dt.set(i, d);
	}

	// -----------------------------------------------------------------------------

	public ArrayList<ArrayList<DecisionTreeNode>> getStreamDT() {
		return streamDT;
	}

	// -----------------------------------------------------------------------------

	public void AddSTREAMWindowCount(int streamId, int val) {
		STREAMWindowCount.put(streamId, val);
	}

	// -----------------------------------------------------------------------------

	public HashMap<Integer, Integer> getSTREAMWindowCount() {
		return STREAMWindowCount;
	}

	// -----------------------------------------------------------------------------
	// STQM - signature is based on count of tuples
	public String GetCountBasedClassificationSignature(int streamId,
			ArrayList currentSample,
			Hashtable hashSTQMRouteSignaturesGlobalStorage,
			Hashtable<Integer, ArrayList> hashCurrentSampleClassificationTuples) {
		String currSignature = "";
		Hashtable<Integer, Integer> hashCurrentSampleClassification = new Hashtable<Integer, Integer>();// hash
																										// of
																										// the
																										// routes
																										// assigned
																										// to
																										// the
																										// current
																										// sample
		// get the decision tree by the stream id
		node root = edt.get(streamId);

		// probe the dt for the current sample and create the signature
		// --------------------------------------------------------------------
		// CLASSIFY THE TUPLES
		// --------------------------------------------------------------------
		for (int i = 0; i < currentSample.size(); i++) {
			XATTuple currTuple = (XATTuple) currentSample.get(i);
			RusterNode classifiedRoute = classifyTuple(currTuple, root);

			int key = classifiedRoute.RusterId;

			if (hashCurrentSampleClassification.containsKey(key)) {
				int count = Integer.parseInt(hashCurrentSampleClassification
						.get(key).toString());
				count++;
				hashCurrentSampleClassification.put(key, count);

				ArrayList tuplesToCurrentRoute = (ArrayList) hashCurrentSampleClassificationTuples
						.get(key);
				tuplesToCurrentRoute.add(currTuple);
			} else {
				int count = 1;
				hashCurrentSampleClassification.put(key, count);

				ArrayList tuplesToCurrentRoute = new ArrayList();
				tuplesToCurrentRoute.add(currTuple);
				hashCurrentSampleClassificationTuples.put(key,
						tuplesToCurrentRoute);
			}
		}

		// --------------------------------------------------------------------
		// CREATE SIGNATURES
		// --------------------------------------------------------------------
		String streamIdStr = "" + streamId;
		Hashtable hashStreamRoutesSignatures = (Hashtable) hashSTQMRouteSignaturesGlobalStorage
				.get(streamIdStr);
		ArrayList arrRoutes = new ArrayList();
		Enumeration eStreamRoutes = hashStreamRoutesSignatures.elements();
		while (eStreamRoutes.hasMoreElements()) {
			arrRoutes.add(eStreamRoutes.nextElement());
		}

		// sort the routes
		Collections.sort(arrRoutes, new RouteSignatureComparator());

		for (int i = 0; i < arrRoutes.size(); i++) {
			RouteSignatureStorage rss = (RouteSignatureStorage) arrRoutes
					.get(i);
			// RusterNode rn = (RusterNode)arrRoutes.get(i);
			int count = 0;
			if (hashCurrentSampleClassification.containsKey(Integer
					.parseInt(rss.GetRouteId()))) {
				count = Integer.parseInt(hashCurrentSampleClassification.get(
						Integer.parseInt(rss.GetRouteId())).toString());
				currSignature += rss.GetRouteId() + ":" + count + " ";
			} else {
				// set 0 for the count of number of tuples assigned to this
				// route in the current signature
				currSignature += rss.GetRouteId() + ":" + "0" + " ";
			}
		}
		if (currSignature.length() > 0) {
			currSignature = currSignature.substring(0,
					currSignature.length() - 1);
		}
		return currSignature;
	}

	// -----------------------------------------------------------------------------
	// STQM - signature is based on the attribute names and the splitting
	// criteria measure
	public String GetSplittingCriteriaClassificationSignature(int streamId,
			ArrayList currentSample, String strDecisionTreeHeader) {
		String strContent = "" + strDecisionTreeHeader;

		String currSignature = "";
		Hashtable<Integer, Integer> hashCurrentSampleClassification = new Hashtable<Integer, Integer>();// hash
																										// of
																										// the
																										// routes
																										// assigned
																										// to
																										// the
																										// current
																										// sample
		// get the decision tree by the stream id
		node root = edt.get(streamId);

		// probe the dt for the current sample and create the signature
		// --------------------------------------------------------------------
		// CLASSIFY THE TUPLES
		// --------------------------------------------------------------------
		for (int i = 0; i < currentSample.size(); i++) {
			XATTuple currTuple = (XATTuple) currentSample.get(i);
			RusterNode classifiedRoute = classifyTuple(currTuple, root);

			int key = classifiedRoute.RusterId;

			if (hashCurrentSampleClassification.containsKey(key)) {
				int count = Integer.parseInt(hashCurrentSampleClassification
						.get(key).toString());
				count++;
				hashCurrentSampleClassification.put(key, count);
			} else {
				int count = 1;
				hashCurrentSampleClassification.put(key, count);
			}
		}

		return currSignature;
	}

	// -----------------------------------------------------------------------------
	// STQM - signature is based on %-age of tuples
	public String GetPercentBasedClassificationSignature(int streamId,
			ArrayList currentSample,
			Hashtable hashSTQMRouteSignaturesGlobalStorage,
			Hashtable<Integer, ArrayList> hashCurrentSampleClassificationTuples) {
		String currSignature = "";
		// hash of the routes assigned to the current sample
		Hashtable<Integer, Integer> hashCurrentSampleClassification = new Hashtable<Integer, Integer>();
		// get the decision tree by the stream id
		node root = edt.get(streamId);

		// probe the dt for the current sample and create the signature
		// --------------------------------------------------------------------
		// CLASSIFY THE TUPLES
		// --------------------------------------------------------------------
		for (int i = 0; i < currentSample.size(); i++) {
			XATTuple currTuple = (XATTuple) currentSample.get(i);
			RusterNode classifiedRoute = classifyTuple(currTuple, root);

			int key = classifiedRoute.RusterId;

			if (hashCurrentSampleClassification.containsKey(key)) {
				int count = Integer.parseInt(hashCurrentSampleClassification
						.get(key).toString());
				count++;
				hashCurrentSampleClassification.put(key, count);

				ArrayList tuplesToCurrentRoute = (ArrayList) hashCurrentSampleClassificationTuples
						.get(key);
				tuplesToCurrentRoute.add(currTuple);
			} else {
				int count = 1;
				hashCurrentSampleClassification.put(key, count);

				ArrayList tuplesToCurrentRoute = new ArrayList();
				tuplesToCurrentRoute.add(currTuple);
				hashCurrentSampleClassificationTuples.put(key,
						tuplesToCurrentRoute);
			}
		}

		// --------------------------------------------------------------------
		// CREATE SIGNATURES
		// --------------------------------------------------------------------
		String streamIdStr = "" + streamId;
		Hashtable hashStreamRoutesSignatures = (Hashtable) hashSTQMRouteSignaturesGlobalStorage
				.get(streamIdStr);
		ArrayList arrRoutes = new ArrayList();
		Enumeration eStreamRoutes = hashStreamRoutesSignatures.elements();
		while (eStreamRoutes.hasMoreElements()) {
			arrRoutes.add(eStreamRoutes.nextElement());
		}

		// sort the routes
		Collections.sort(arrRoutes, new RouteSignatureComparator());

		for (int i = 0; i < arrRoutes.size(); i++) {
			RouteSignatureStorage rss = (RouteSignatureStorage) arrRoutes
					.get(i);
			// RusterNode rn = (RusterNode)arrRoutes.get(i);
			int count = 0;
			if (hashCurrentSampleClassification.containsKey(Integer
					.parseInt(rss.GetRouteId()))) {
				count = Integer.parseInt(hashCurrentSampleClassification.get(
						Integer.parseInt(rss.GetRouteId())).toString());

				// translate count of tuples to %
				int percent = (count / currentSample.size()) * 100; // percent
																	// of tuples
																	// assigned
																	// to a
																	// route
																	// from the
																	// current
																	// sample
				currSignature += rss.GetRouteId() + ":" + percent + " ";
			} else {
				// set 0 for the % of tuples assigned to this route in the
				// current signature
				currSignature += rss.GetRouteId() + ":" + "0" + " ";
			}
		}
		if (currSignature.length() > 0) {
			currSignature = currSignature.substring(0,
					currSignature.length() - 1);
		}
		return currSignature;
	}
	// -----------------------------------------------------------------------------
}
