package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution;

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

//Karen Works
//This is the eddy STREAM Select operator

@SuppressWarnings("serial")
public class QMeshStreamSelectOperatorImp extends
		XATMultiQueueWindowStreamOperatorImp implements
		QMeshStreamSelectOperator {

	private ExpressionStr expression;
	private OperatorLookUp OpLkUp;
	private int QMeshOperatorID;

	// Tuples that require processing
	// private HashMap<Integer, EddyQueue> TuplePool;

	public QMeshStreamSelectOperatorImp(XATNode init) {
		super(init);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator
	 * #run()
	 */
	@SuppressWarnings("unused")
	public boolean run() throws Exception {
		/**
		 * first figure out how many tuples to dequeue, the selectivity, and
		 * result delay
		 */
		XATQueueSchema inSchema;
		int inLength;
		if ((inSchema = this.getInputQueue().getSchema()) == null) {
			return false;
		}
		inLength = inSchema.getLength();

		// modified by Song Wang to support multiple output queue
		for (int i = 0; i < this.getOutputQueueCount(); i++) {
			if (this.getOutputQueue(i).getSchema() == null) {
				init(i);
			}
		}

		int howMany = Integer.parseInt(this
				.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));

		/** get the tuples */
		int len = howMany;
		XATTuple t = null;

		// ensure that we look at complete rusters
		if (this.getInputQueue().getQueueSize() > len) {
			t = this.getInputQueue().peekNthTuple(len, this);

			// make sure we look at entire ruster in one operator's turn to run
			while ((!t.isRoutingToken())
					&& (this.getInputQueue().getQueueSize() > (len)))
			// is this a routing puncuation?
			{
				len++;
				if (this.getInputQueue().getQueueSize() > len)
					t = this.getInputQueue().peekNthTuple(len, this);
				else
					break;
			}
		}

		XATTuple[] tuples = this.getInputQueue().dequeue(len, this);
		/** figure out how many tuples to keep */
		Vector<XATTuple> tupleList = new Vector<XATTuple>();

		RoutingToken rt = null;
		boolean bSentRt = false;
		for (int i = 0; i < tuples.length; i++) {

			if (tuples[i].isRoutingToken()) // is this a routing puncuation?
			{
				rt = ((RoutingToken) tuples[i]);
				rt.popTopOperator();
				bSentRt = false;
			} else {
				QMeshTuple qmt = ((QMeshTuple) tuples[i]);
				tupleList.add(0, qmt.getTuple(this.GetStreamID()));

				Object out = expression.eval(tupleList, this);
				if (out instanceof Boolean) {
					if (((Boolean) out).booleanValue()) {
						// Found Answer!!!

						// if there are no operators left send tuple to output
						if (rt.OpStackSize() == 0) {
							OpLkUp.getOutputQueue().enqueue(tuples[i], this);

						} else {
							int nextOp = rt.getTopOperator();

							if (!bSentRt) {
								OpLkUp.getOperatorQueue(nextOp).enqueue(rt,
										this);
								bSentRt = true;
							}
							OpLkUp.getOperatorQueue(nextOp).enqueue(tuples[i],
									this);
							// System.out.println(qmt.getTuple(this.GetStreamID()));
						}

					}
				}
			}
		}
		return true;
	}

	public void init(int i) {
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) {
			return;
		}
		XATQueueSchema outSchema = this.getOutputQueue(i).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) {
			this.getOutputQueue(i).setSchema(inSchema);
		}
		computeOutputPositions();// no need to overload since all the output queue are the same
		numOutputColumns = outputPositions[0].length;
		setHasInit(true);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator
	 * #toToolTipString()
	 */
	public String toToolTipString() {
		return "Select Operator (Query Mesh)";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamSelectOperator
	 * #getExpression()
	 */
	public ExpressionStr getExpression() {
		return expression;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamSelectOperator
	 * #
	 * setExpression(edu.wpi.cs.dsrg.xmldb.xat.common.streamexpression.ExpressionStr
	 * )
	 */
	public void setExpression(ExpressionStr expression) {
		this.expression = expression;
	}

	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input
	 * queues. Inside method run() the numbers are computed based on the
	 * property "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler, while
	 * in cleanAccumulatedTuples, all tuples in input queues that have their max
	 * timesamp smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) {
		return true;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() {
		return true;
	}

	/*
	 * check if this operator has any unmatched states. return: TRUE if this
	 * operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() {
		return false;
	}

	public boolean matchStates(XATStreamOperator op) {
		return false;
	}

	public boolean recomputeUnmatchedState() {
		return true;
	}

	public int setStateID(int nextid) {
		return nextid;
	}

	public boolean finalizeOperation() {
		return true;
	}

	public void setOperatorLookUp(OperatorLookUp z) {
		OpLkUp = z;
	}

	public void setQMeshOperatorID(int id) {
		QMeshOperatorID = id;
	}

	public int getQMeshOperatorID() {
		return QMeshOperatorID;
	}
}
