//--------------------------------------------------
// RoutingTokenImp.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution;

import java.util.ArrayList;
import java.util.List;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;

/**
 * @author Rimma V. Nehme
 * 
 */
@SuppressWarnings("serial")
public class RoutingTokenImp extends XATPunctImp implements RoutingToken {
	// -----------------------------------------------------------------------------

	// STQM - route identifier - needed to maintain the statistics for this
	// route and its actual runtime rusters
	private String routeId;

	// STQM - isExplorationRuster (ruster that was sent on a route other than
	// its best current route)
	private boolean isExplorationRuster = false;

	// STQM - isStatisticsRuster (ruster for which statistics are collected and
	// hence tuples are not really "discarded", just marked as discarded)
	private boolean isStatisticsRuster = false;

	protected ArrayList<Integer> operatorStack = null;

	// Number of tuples
	int NumTuples;

	// Stream ID
	private String rTokenStreamId;

	
	
	// -----------------------------------------------------------------------------

	public RoutingTokenImp() {
		super();
	}

	public RoutingTokenImp(String rTokenStreamIdVal, String routeIdVal) {
		super();
		this.objTypeID = 3; // routing token punctuation have object type id = 3
		rTokenStreamId = rTokenStreamIdVal;
		routeId = routeIdVal;
	}

	// -----------------------------------------------------------------------------

	@SuppressWarnings("unchecked")
	public RoutingTokenImp(String rTokenStreamIdVal, String routeIdVal,
			ArrayList<Integer> opStack) {
		super();
		this.objTypeID = 3; // routing token punctuation have object type id = 3
		rTokenStreamId = rTokenStreamIdVal;
		routeId = routeIdVal;
		operatorStack = (ArrayList<Integer>) opStack.clone();
	}

	// -----------------------------------------------------------------------------

	public RoutingTokenImp(String rTokenStreamIdVal, String routeIdVal,
			XATDataValue[] array) {
		super(array);
		this.objTypeID = 3; // routing token punctuation have object type id = 3
		rTokenStreamId = rTokenStreamIdVal;
		routeId = routeIdVal;
	}

	// -----------------------------------------------------------------------------

	public RoutingTokenImp(String rTokenStreamIdVal, String routeIdVal,
			int length) {
		super(length);
		this.objTypeID = 3; // routing token punctuation have object type id = 3
		rTokenStreamId = rTokenStreamIdVal;
		routeId = routeIdVal;
	}

	// -----------------------------------------------------------------------------

	@SuppressWarnings("rawtypes")
	public RoutingTokenImp(String rTokenStreamIdVal, String routeIdVal,
			List list) {
		super(list);
		this.objTypeID = 3; // routing token punctuation have object type id = 3
		rTokenStreamId = rTokenStreamIdVal;
		routeId = routeIdVal;
	}

	// -----------------------------------------------------------------------------

	public RoutingTokenImp CreateClone() {
		RoutingTokenImp rTokenClone = new RoutingTokenImp(this.rTokenStreamId,
				this.routeId, this.operatorStack);
		return rTokenClone;
	}

	// -----------------------------------------------------------------------------

	public void setRTokenPunctTypeArray(int[] pTypeArray,
			ArrayList<Integer> operatorStackVal) {
		punctTypeArray = pTypeArray;
		operatorStack = operatorStackVal;
	}

	// -----------------------------------------------------------------------------

	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("\nR-TOKEN: ");
		sb.append("\nStream Id: " + this.GetRTokenStreamId());
		sb.append("\nRoute Id: " + this.GetRouteId());
		if (this.isExplorationRuster) {
			sb.append("\nIs Exploration: true");
		} else {
			sb.append("\nIs Exploration: false");
		}
		sb.append("\n# of tuples in ruster: " + this.GetNumTuples());
		sb.append("\nOperator Stack: ");

		for (int i = 0; i < this.operatorStack.size(); i++) {
			int opID = this.operatorStack.get(i);
			sb.append(opID + ",");
		}
		return sb.toString().substring(0, sb.toString().length() - 1);
	}

	// -----------------------------------------------------------------------------

	public String getStackString() {
		StringBuffer sb = new StringBuffer();

		if (this.operatorStack.size() > 0) {
			for (int i = 0; i < this.operatorStack.size(); i++) {
				int opID = this.operatorStack.get(i);
				sb.append(opID + ",");
			}
			return sb.toString().substring(0, sb.toString().length() - 1);
		} else {
			return "";
		}
	}

	// -----------------------------------------------------------------------------

	public String toShortString() {
		StringBuffer sb = new StringBuffer();
		sb.append("\nR-TOKEN: \n");

		for (int i = 0; i < operatorStack.size(); i++) {
			sb.append("op_" + i + ":" + operatorStack.get(i) + "\n");
		}
		return sb.toString();
	}

	// -----------------------------------------------------------------------------

	public Integer getTopOperator() {
		// return operatorStack.get(operatorStack.size()-1);
		return operatorStack.get(0);
	}

	// -----------------------------------------------------------------------------

	public void popTopOperator() {
		if (operatorStack != null) {
			// operatorStack.remove(operatorStack.size()-1);
			operatorStack.remove(0);
		}
	}

	// -----------------------------------------------------------------------------

	public int OpStackSize() {
		if (operatorStack != null) {
			return operatorStack.size();
		} else {
			return 0;
		}
	}

	// -----------------------------------------------------------------------------

	// Should be named as setOpStack
	@SuppressWarnings("unchecked")
	public void setOpStackSize(ArrayList<Integer> os) {
		operatorStack = (ArrayList<Integer>) os.clone();
	}

	// -----------------------------------------------------------------------------

	public ArrayList<Integer> getOpStack() {
		return operatorStack;
	}

	// -----------------------------------------------------------------------------

	public int GetNumTuples() {
		return NumTuples;
	}

	// -----------------------------------------------------------------------------

	public String GetRTokenStreamId() {
		return rTokenStreamId;
	}

	// -----------------------------------------------------------------------------

	public void SetNumTuples(int t) {
		/*
		 * if (t == 1) { System.out.println("want to catch this"); }
		 */
		NumTuples = t;
	}

	// -----------------------------------------------------------------------------

	// STQM
	public String GetRouteId() {
		return routeId;
	}

	// -----------------------------------------------------------------------------

	public void SetRouteId(String routeIdVal) {
		routeId = routeIdVal;
	}

	// -----------------------------------------------------------------------------

	// STQM
	public boolean GetIsExplorationRuster() {
		return isExplorationRuster;
	}

	// -----------------------------------------------------------------------------

	public void SetIsExplorationRuster(boolean isExplorationRusterVal) {
		isExplorationRuster = isExplorationRusterVal;
	}

	// -----------------------------------------------------------------------------

	// STQM
	public boolean GetIsStatisticsRuster() {
		return isStatisticsRuster;
	}

	// -----------------------------------------------------------------------------

	public void SetIsStatisticsRuster(boolean isStatisticsRusterVal) {
		isStatisticsRuster = isStatisticsRusterVal;
	}
	// -----------------------------------------------------------------------------
}
