//--------------------------------------------------
// STQM.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2008
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.multiquery;

import java.io.*;
import java.text.*;
import java.util.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.sun.org.apache.xml.internal.serialize.OutputFormat; 
import com.sun.org.apache.xml.internal.serialize.XMLSerializer;

import edu.wpi.cs.dsrg.STeM.STeMJoinProbeOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.GlobalQueryMesh;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.STQM_Analyzer.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.*;

public class MQM extends Thread
{
	//-----------------------------------------------------------------------------
	
	static int countSTQMExecutions = 0;
	
	//-----------------------------------------------------------------------------
	
	private MQM_Monitor stqmMonitor;
	private MQM_Analyzer stqmAnalyzer;
	private MQM_Actuator stqmActuator;

	//-----------------------------------------------------------------------------
	
	public static String OPTIMIZER_INPUT_XML_FILE = "C:\\Users\\Rimma\\workspace\\Raindrop\\resources\\SelfTuningQueryMesh\\optimizer_input.xml"; //location of the xml file with necessary settings for QM optimizer
	public static String OPTIMIZER_TRAINING_SETS_DIR = "C:\\RIMMA\\Research\\QueryMesh\\training_sets\\"; //location of the xml file with necessary settings for QM optimizer
	public static String OPTIMIZER_STATISTICS_SAMPLES_DIR = "C:\\RIMMA\\Research\\QueryMesh\\statistics_samples\\"; //location of the xml file with necessary settings for QM optimizer
	public static String QM_SOLUTIONS_DIR = "C:\\RIMMA\\Research\\QueryMesh\\qm_solutions\\"; //location of directory that will contain generated QM solutions by QM optimizer for different training tuples/statistics (when concept drifts get detected)
	
	//-----------------------------------------------------------------------------
	
	public static String NEW_QM_XML;       //location of xml that contains materialized new QM computed when hybrid concept drift occurs
	public static String NEW_CLASSIFIER_XML;   //location of xml that contains materialized new CLASSIFIER computed when virtual concept drift occurs
	public static Hashtable NEW_ROUTES;             //location of xml that contains materialized new ROUTES computed when real concept drift occurs (key --> route id)
	
	MultiQueryQueryMeshExecutionController currExecutionController; //current execution controller
	
	//-----------------------------------------------------------------------------
	
	public MQM(MultiQueryQueryMeshExecutionController executionControllerVal)
	{
		currExecutionController = executionControllerVal;
		//initialize all components
		stqmMonitor = new MQM_Monitor();
		stqmAnalyzer = new MQM_Analyzer();
		stqmActuator = new MQM_Actuator(currExecutionController);
	}
    
	//-----------------------------------------------------------------------------
	
	 public void run()
	 {
		 //get the needed variables
		 Hashtable hashSTQMDataSamplesGlobalStorage = DistributedExecutionController.hashSTQMDataSamplesGlobalStorage; //global storage for all data samples		
		 Hashtable hashSTQMExecutionStatisticsSamplesGlobalStorage = DistributedExecutionController.hashSTQMExecutionStatisticsSamplesGlobalStorage; //global storage for all operators execution statistics		
		 Hashtable hashSTQMRouteSignaturesGlobalStorage = DistributedExecutionController.hashSTQMRouteSignaturesGlobalStorage; //global storage for all possible route signatures						
		 OnlineClassifierOperatorImp OnlineClassifierRef = DistributedExecutionController.OnlineClassifierRef;
					
		 //call the tuner
		 DoQMTuning(hashSTQMDataSamplesGlobalStorage,
					hashSTQMExecutionStatisticsSamplesGlobalStorage,
					hashSTQMRouteSignaturesGlobalStorage,
					OnlineClassifierRef);
		 		 
	 }
	//-----------------------------------------------------------------------------
	
	 public void DoQMTuning(Hashtable hashSTQMDataSamplesGlobalStorage, 
			 				Hashtable hashSTQMExecutionStatisticsSamplesGlobalStorage, 
			 				Hashtable hashSTQMRouteSignaturesGlobalStorage,
			 				OnlineClassifierOperatorImp OnlineClassifierRef)
	 {
		 //------------------------------------------------------------------
		 // QM TUNING PARAMETERS
		 //------------------------------------------------------------------
	
		 STQM_VCD_METHOD vcdMethod = STQM_VCD_METHOD.SIGNATURE_BASED;	  
		 STQM_RCD_METHOD rcdMethod = STQM_RCD_METHOD.NUMBER_OF_AFFECTED_ROUTES;
		 
		 //------------------------------------------------------------------
		 
		 Hashtable hashSamplesPerStream = new Hashtable();
		 Hashtable hashSamplesPerOperator = new Hashtable();
		 
		 System.out.println("");
		 System.out.println("--- STARTING STQM TUNER ---");
		 Helper.WriteToFile(Helper.LOG_FILE, System.getProperty("line.separator") + "----------------------------------------------------------------");
		 Helper.WriteToFile(Helper.LOG_FILE,  Helper.GetCurrentTimeString() + ": --- STARTING STQM TUNER --- ");
		 //------------------------------------------------------------------
		 // 1. GET ALL DATA SAMPLES
		 //------------------------------------------------------------------
		 Helper.WriteToFile(Helper.LOG_FILE,  Helper.GetCurrentTimeString() + ": --- Getting all data samples --- ");
		 
		 Enumeration e = hashSTQMDataSamplesGlobalStorage.keys();
		 while (e.hasMoreElements()) 
		 {
			int key = Integer.parseInt(e.nextElement().toString());				
		   	//System.out.println("Stream "+key+": "); //key here is a stream
		   	
		   	//retrieve data sample (which is a pointer to the sampling operator)
		   	DataSampleStorage ds = (DataSampleStorage)hashSTQMDataSamplesGlobalStorage.get(key);	
		   	TrainingTupleCollectorOperatorImp physOp = ds.GetOp();
		   	
		   	//get the old and the new training sets for the sampling operator for the stream
		   	ArrayList oldSample = physOp.GetOldTuplesSample();
		   	ArrayList newSample = physOp.GetNewTuplesSample();		   			   	
		   	
		   	StreamOldAndNewSamples ss = new StreamOldAndNewSamples();
		   	ss.SetNewSamples(newSample);
		   	ss.SetOldSamples(oldSample);
		   	
		   	hashSamplesPerStream.put(key, ss);
		}
		 //------------------------------------------------------------------
		 // 2. GET ALL EXECUTION STATISTICS
		 //------------------------------------------------------------------
		 Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Getting all execution statistics --- ");
		 
		 e = hashSTQMExecutionStatisticsSamplesGlobalStorage.keys();
		 while (e.hasMoreElements()) 
		 {
			int key = Integer.parseInt(e.nextElement().toString());	
			
			if (key == 0)
			{
				continue;
			}
			
		   	//System.out.println("Operator id "+key+": "); //key here is operator id
		   	
		   	ExecutionStatisticsSampleStorage es = (ExecutionStatisticsSampleStorage)hashSTQMExecutionStatisticsSamplesGlobalStorage.get(key); 
		   	STeMJoinProbeOperatorImp physOp = (STeMJoinProbeOperatorImp)es.GetOp();
		   	
		   	Hashtable hashNewSTQMRusterExploitationStatisticsLocalStorage = physOp.GetNewOperatorExploitationStatisticsPerRuster();
		   	hashSamplesPerOperator.put(key, hashNewSTQMRusterExploitationStatisticsLocalStorage);
		 }
		 
		 //------------------------------------------------------------------
		 // 3. TEST DATA FOR CONCEPT DRIFTS
		 //------------------------------------------------------------------
		 Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Testing data for concept drifts --- ");
		 Hashtable hashVirtualConceptDriftResults = new Hashtable();	
		 Hashtable hashRealConceptDriftResults = new Hashtable();
			
		 /*
		 stqmAnalyzer.DoQMAnalysis(hashSamplesPerStream, 		  
				                   hashSamplesPerOperator, 
					 			   hashSTQMRouteSignaturesGlobalStorage, 
					 			   OnlineClassifierRef,
					 			   vcdMethod, 
					 			   rcdMethod,
					 			   hashVirtualConceptDriftResults,
					 			   hashRealConceptDriftResults);
			 			
		 */
		 Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- "+countNumVirtualConceptDrifts(hashVirtualConceptDriftResults)+" (out of " +hashVirtualConceptDriftResults.size()+") virtual concept drifts --- ");
		 Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- "+countNumRealConceptDrifts(hashRealConceptDriftResults)+" (out of "+hashRealConceptDriftResults.size()+") real concept drifts --- ");
			 
		 //------------------------------------------------------------------
		 // 4. GIVEN THE DETECTED CONCEPT DRIFTS, PRODUCE RECOMMENDATIONS
		 //------------------------------------------------------------------		 		 
		 Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Producing recommendations based on concept drifts --- ");
		 /*
		 STQM_TUNING_RECOMMENDATIONS currentRecommendations = stqmAnalyzer.MakeQMRecommendations(hashSTQMDataSamplesGlobalStorage,
				 																				 hashVirtualConceptDriftResults, 
					 													 					     hashRealConceptDriftResults);

		 
		 Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Done producing recommendations --- ");
		 
		 if (currentRecommendations != STQM_TUNING_RECOMMENDATIONS.NONE)
		 {
			 DistributedExecutionController.bIsSTQMDoPhysicalMigration = true;
			 DistributedExecutionController.currentPhysicalRecommendations = currentRecommendations;
		 }
		 */
		 
		 DistributedExecutionController.stqmLastTime = System.currentTimeMillis();
		 DistributedExecutionController.bIsSTQMExecuting = false; //reset the flag indicating that the STQM is currently running
	 }

	//-----------------------------------------------------------------------------
	 
	 public void DoPhysicalMigration(Hashtable hashSTQMDataSamplesGlobalStorage, 
									 Hashtable hashSTQMExecutionStatisticsSamplesGlobalStorage, 
									 OnlineClassifierOperatorImp OnlineClassifierRef)
	 {
		 STQM_TUNING_RECOMMENDATIONS currentRecommendations = DistributedExecutionController.currentPhysicalRecommendations;
		 
		 //------------------------------------------------------------------
		 // 5. INVOKE THE ACTUATOR TO PERFORM THE PHYSICAL ADAPTATION
		 //------------------------------------------------------------------		 		 
		 if (currentRecommendations != STQM_TUNING_RECOMMENDATIONS.NONE)
		 {
			 Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Performing physical adaptation of QM --- ");
			 boolean isAdaptationSuccessfull = stqmActuator.ExecuteQMAdaptivity(currentRecommendations, 
					 															OnlineClassifierRef);
			 Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Finished adaptation. Was adaptation successful? [" + isAdaptationSuccessfull + "] --- ");
			 
			 //make the statistics used for latest adaptation to be the "old" statistics 
			 MakeNewStatisticsToBeOldStatistics(hashSTQMDataSamplesGlobalStorage, 
		 				 					    hashSTQMExecutionStatisticsSamplesGlobalStorage);
		 }
		 else
		 {
			 Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- NOT performing physical adaptation of QM --- ");
			 Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- REASON, recommendations = " +currentRecommendations+" --- ");
		 }
		 System.out.println("");
		 System.out.println("--- ENDED STQM TUNER --");
		 
		 //increment the counter
		 countSTQMExecutions++;
	 }
	 
	//-----------------------------------------------------------------------------
	
	private void MakeNewStatisticsToBeOldStatistics(Hashtable hashSTQMDataSamplesGlobalStorage, 
													Hashtable hashSTQMExecutionStatisticsSamplesGlobalStorage)
	{
		 //------------------------------------------------------------
		 // DATA SAMPLES
		 //------------------------------------------------------------		
		 Enumeration e = hashSTQMDataSamplesGlobalStorage.keys();
		 while (e.hasMoreElements()) 
		 {
			int key = Integer.parseInt(e.nextElement().toString());				

		   	DataSampleStorage ds = (DataSampleStorage)hashSTQMDataSamplesGlobalStorage.get(key);	
		   	TrainingTupleCollectorOperatorImp physOp = ds.GetOp();	
		   	
		   	physOp.MakeNewSampleToBeOldSample(); //set new sample to be the old sample
		 }//end while
		 
		 //------------------------------------------------------------
		 // EXECUTION STATISTICS
		 //------------------------------------------------------------
		 
		 e = hashSTQMExecutionStatisticsSamplesGlobalStorage.keys();
		 while (e.hasMoreElements()) 
		 {
			int key = Integer.parseInt(e.nextElement().toString());	// key = operator id
			
			if (key == 0)
			{
				continue;
			}

		   	ExecutionStatisticsSampleStorage es = (ExecutionStatisticsSampleStorage)hashSTQMExecutionStatisticsSamplesGlobalStorage.get(key); 
		   	STeMJoinProbeOperatorImp physOp = (STeMJoinProbeOperatorImp)es.GetOp();
		   	physOp.MakeNewStatisticsToBeOldStatistics();		   	
		 }//end while
	}
	 
	//----------------------------------------------------------------------------- 
	 
	private int countNumVirtualConceptDrifts(Hashtable hashVirtualConceptDriftResults)
	{
		int countVirtualDrifts = 0;
		//-----------------------------------------------------------------------
		// VIRTUTAL CONCEPT DRIFTS - do the count
		//-----------------------------------------------------------------------		
		Enumeration eConceptDrifts = hashVirtualConceptDriftResults.elements(); 		
		while (eConceptDrifts.hasMoreElements()) 
		{
			if (eConceptDrifts.nextElement().toString().equalsIgnoreCase("true"))
			{
				countVirtualDrifts++;
			}
		}
		return countVirtualDrifts;
	}
	
	//-----------------------------------------------------------------------------
	
	private int countNumRealConceptDrifts(Hashtable hashRealConceptDriftResults)
	{
		int countRealDrifts = 0;
		//-----------------------------------------------------------------------
		// REAL CONCEPT DRIFTS - do the count
		//-----------------------------------------------------------------------		
		Enumeration eConceptDrifts = hashRealConceptDriftResults.elements(); 		
		while (eConceptDrifts.hasMoreElements()) 
		{
			if (eConceptDrifts.nextElement().toString().equalsIgnoreCase("true"))
			{
				countRealDrifts++;
			}
		}
		return countRealDrifts;
	}
	
	//-----------------------------------------------------------------------------
	
	public static void UpdateOptimizerInputXmlFileWithLatestSettings(String statisticsSamplesSubDir, String trainingTuplesSamplesSubDir) 
	{
		//get the factory
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		
		try 
		{			
			//Using factory get an instance of document builder
			DocumentBuilder db = dbf.newDocumentBuilder();
			
			//parse using builder to get DOM representation of the XML file
			Document dom = db.parse(STQM.OPTIMIZER_INPUT_XML_FILE);			
		
			//get the root element
			Element docEle = dom.getDocumentElement();
			
			//get a nodelist of <employee> elements
			NodeList nl = docEle.getElementsByTagName("Input");
			if(nl != null && nl.getLength() > 0) 
			{
				for(int i = 0 ; i < nl.getLength();i++) 
				{					
					//get the Input element
					Element el = (Element)nl.item(i);
					
					//for each input need to modify two attributes:
					// (1) statistics_file - location of the file that will be used for stats computation by the optimizer
					// (2) training_tuples_file - location of the file that contains training tuples that will be used for classifier computation by the optimizer
					/* <Input id="0" name="Stream0" 
					 * statistics_file="C:\\Rimma\\Research\\QueryMesh\\statistics_samples\\SAMPLE_DS5_STREAM0.txt" 
					 * training_tuples_file="C:\\Rimma\\Research\\QueryMesh\\training_sets\\TRAINING_SET_DS5_STREAM0.txt">
					 */
					
					String currStreamId = el.getAttribute("id");

					el.setAttribute("statistics_file", STQM.OPTIMIZER_STATISTICS_SAMPLES_DIR + statisticsSamplesSubDir +"\\" + "SAMPLE_STREAM" + currStreamId + ".txt");
					el.setAttribute("training_tuples_file", STQM.OPTIMIZER_TRAINING_SETS_DIR + trainingTuplesSamplesSubDir +"\\" + "TRAINING_SET_STREAM" + currStreamId + ".txt");										
				}
			}
			
			//------------------------------------------------------------------
			// SERIALIZE UPDATED CONTENT BACK TO XML FILE
			//------------------------------------------------------------------
			OutputFormat format = new OutputFormat(dom);
			format.setIndenting(true);

			//to generate output to console use this serializer
			//XMLSerializer serializer = new XMLSerializer(System.out, format);


			//to generate a file output use fileoutputstream instead of system.out
			XMLSerializer serializer = new XMLSerializer(
			new FileOutputStream(new File(STQM.OPTIMIZER_INPUT_XML_FILE)), format);

			serializer.serialize(dom);//write the file back

		}
		catch(ParserConfigurationException pce) 
		{
			pce.printStackTrace();
		}
		catch(SAXException se) 
		{
			se.printStackTrace();
		}
		catch(IOException ioe) 
		{
			ioe.printStackTrace();
		}
	}
	
	//-----------------------------------------------------------------------------
	
	//returns sub-directory that contains all the statistics sample files that get created by this method
	public static String CreateNewOptimizerTrainingTuplesFiles(String subDir, Hashtable hashSTQMDataSamplesGlobalStorage)
	{        		
		String fullNameSubDir = STQM.OPTIMIZER_TRAINING_SETS_DIR + subDir + "\\";
		// Create one directory
	    boolean success = (new File(fullNameSubDir)).mkdir();
	    
	    //---------------------------------------------------------------------------
	    // ITERATE THROUGH EACH STREAM
  	    //---------------------------------------------------------------------------
	     //now iterate through the samples of data and create their respective files
		 Enumeration e = hashSTQMDataSamplesGlobalStorage.keys();
		 while (e.hasMoreElements()) 
		 {
			int key = Integer.parseInt(e.nextElement().toString());				
		   	//System.out.println("Stream "+key+": "); //key here is a stream
		   	
		   	//retrieve data sample (which is a pointer to the sampling operator)
		   	DataSampleStorage ds = (DataSampleStorage)hashSTQMDataSamplesGlobalStorage.get(key);	
		   	TrainingTupleCollectorOperatorImp physOp = ds.GetOp();
		   	
		   	//get the new training sets for the sampling operator for the stream
		   	ArrayList newSample = physOp.GetNewTuplesSample();
		   	
		   	String currStreamSamplesFileName = "TRAINING_SET_STREAM" + key + ".txt";
		   	
		   	Hashtable usedNotUsedAttributes = physOp.usedNotUsedAttributes; //get used/not used hashtable for the current stream
		   	Hashtable histogramsTable = physOp.histogramsTable; //get histograms hashtable for the current stream
		   	
		   	//create training tuples header (used by dt - it contains some metadata, like names of the attributes)
		   	String currAllTuplesStr = "";
		   	
		   	currAllTuplesStr += HistogramBuilderOperatorImp.CreateTrainingTupleFileHeader(physOp.GetDecisionTreeAttributes(), 
		   																                  usedNotUsedAttributes) + System.getProperty("line.separator");
		   	
		   	int numTrainingTuplesToOptimizer = 20; //RIMMA: (small hack) trying to limit this number
		   	//write the training tuples
		   	if (newSample.size() < numTrainingTuplesToOptimizer)
		   	{
		   		numTrainingTuplesToOptimizer = newSample.size(); //otherwise, use the size of the collected training tuples
		   	}
		   	
		   	for (int p = 0; p < numTrainingTuplesToOptimizer; p++)
		   	{
		   		XATTuple inputTuple = (XATTuple)newSample.get(p);		
		   		if (inputTuple == null)
		   		{
		   			continue;
		   		}
		   		String currentTupleStr = HistogramBuilderOperatorImp.CreateTrainingTupleString(inputTuple, usedNotUsedAttributes, histogramsTable);
		   		currAllTuplesStr += currentTupleStr + System.getProperty("line.separator");
		   	}
		    Helper.WriteToFile(fullNameSubDir + currStreamSamplesFileName, currAllTuplesStr);			   
		}
	    
		//return the name of the current sub-directory
	    return subDir;
	}
	
	//-----------------------------------------------------------------------------
	
	//returns sub-directory that contains all the statistics sample files that get created by this method
	public static String CreateNewOptimizerStatisticsSampleFiles(String subDir, Hashtable hashSTQMDataSamplesGlobalStorage)
	{
		String fullNameSubDir = STQM.OPTIMIZER_STATISTICS_SAMPLES_DIR + subDir + "\\";
		// Create one directory
	    boolean success = (new File(fullNameSubDir)).mkdir();
	    
	    //---------------------------------------------------------------------------
	    // ITERATE THROUGH EACH STREAM
  	    //---------------------------------------------------------------------------
	     //now iterate through the samples of data and create their respective files
		 Enumeration e = hashSTQMDataSamplesGlobalStorage.keys();
		 while (e.hasMoreElements()) 
		 {
			int key = Integer.parseInt(e.nextElement().toString());				
		   	//System.out.println("Stream "+key+": "); //key here is a stream
		   	
		   	//retrieve data sample (which is a pointer to the sampling operator)
		   	DataSampleStorage ds = (DataSampleStorage)hashSTQMDataSamplesGlobalStorage.get(key);	
		   	TrainingTupleCollectorOperatorImp physOp = ds.GetOp();
		   	
		   	//get the new training sets for the sampling operator for the stream
		   	ArrayList newSample = physOp.GetNewTuplesSample();
		   	
		   	String currStreamSamplesFileName = "SAMPLE_STREAM" + key + ".txt";
		   	Hashtable histogramsTable = physOp.histogramsTable; //get histograms hashtable for the current stream
		   	
		   	//write the sample tuples
		   	String currAllTuplesStr = "";
		   	for (int p = 0; p < newSample.size(); p++)
		   	{
		   		XATTuple inputTuple = (XATTuple)newSample.get(p);
		   		if (inputTuple != null)
		   		{
			   		String currentTupleStr = SampleStatsCollectorOperatorImp.CreateSampleTupleString(inputTuple, histogramsTable);
			   		currAllTuplesStr += currentTupleStr + System.getProperty("line.separator");
		   		}
		   	}
		    Helper.WriteToFile(fullNameSubDir + currStreamSamplesFileName, currAllTuplesStr);			   
		}
	    
		//return the name of the current sub-directory
	    return subDir;
	}
	
	//-----------------------------------------------------------------------------
}
