//--------------------------------------------------
// MQM_Actuator.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, Oct. 2008
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.multiquery;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Coin;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.DecisionTreeNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.OnlineClassifierOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.OnlineClassifierOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.RusterNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.node;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.STQM_Analyzer.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.DistributedExecutionController;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.MultiQueryQueryMeshExecutionController;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.SelfTuningQueryMeshExecutionController;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.*;


public class MQM_Actuator 
{
	//-----------------------------------------------------------------------------
	
	MultiQueryQueryMeshExecutionController currExecutionController; //current execution controller
	
	//-----------------------------------------------------------------------------
	
	//constructor
	public MQM_Actuator(MultiQueryQueryMeshExecutionController executionControllerVal)
	{
		System.out.println("Constructor call to STQM Actuator");
		currExecutionController = executionControllerVal;
	}
	
	//-----------------------------------------------------------------------------
	
	public boolean ExecuteQMAdaptivity(STQM_TUNING_RECOMMENDATIONS currentRecommendations,
									   OnlineClassifierOperatorImp OnlineClassifierRef)
	{
		boolean isAdaptationSuccessful = true;
				
		//--------------------------------------------------------------------------
		// NEW CLASSIFIER + NEW ROUTES (for hybrid concept drift)
		//--------------------------------------------------------------------------
		if (currentRecommendations == STQM_TUNING_RECOMMENDATIONS.NEW_CLASSIFIER_NEW_ROUTES)
		{
			Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Adapting entire QM (CLASSIFIER + ROUTES) --- ");

			//------------------------------------------------
			// MAKE NEW QM INSTANCE IN CAPE
			//------------------------------------------------
			CreateNewQMInstantiateInCAPE(OnlineClassifierRef);
		}
		//--------------------------------------------------------------------------
		// NEW CLASSIFIER + OLD ROUTES (for virtual concept drift)
		//--------------------------------------------------------------------------
		else if (currentRecommendations == STQM_TUNING_RECOMMENDATIONS.NEW_CLASSIFIER_OLD_ROUTES)
		{
			Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Adapting only CLASSIFIER --- ");

			//------------------------------------------------
			// MAKE NEW QM INSTANCE IN CAPE
			//------------------------------------------------
			CreateNewQMInstantiateInCAPE(OnlineClassifierRef);
		}
		//--------------------------------------------------------------------------
		// OLD CLASSIFIER + NEW ROUTES (for real concept drift)
		//--------------------------------------------------------------------------
		else if (currentRecommendations == STQM_TUNING_RECOMMENDATIONS.OLD_CLASSIFIER_NEW_ROUTES)
		{
			Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Adapting only ROUTES --- ");
			//------------------------------------------------
			// MAKE NEW QM INSTANCE IN CAPE
			//------------------------------------------------
			CreateNewQMInstantiateInCAPE(OnlineClassifierRef);
		}
		return isAdaptationSuccessful;
	}
	
	//-----------------------------------------------------------------------------
	
	private void CreateNewQMInstantiateInCAPE(OnlineClassifierOperatorImp OnlineClassifierRef)
	{
		//get the factory
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		
		try 
		{
			//process the xml of QM
			//Using factory get an instance of document builder
			DocumentBuilder db = dbf.newDocumentBuilder();
			
			//parse using builder to get DOM representation of the XML file
			Document dom = db.parse(STQM.NEW_QM_XML);			
		
			//get the root element
			Element docEle = dom.getDocumentElement();
			
			//get a nodelist of <employee> elements
			NodeList nl = docEle.getElementsByTagName("globalDecisionTree");			
			Node currNode = nl.item(0);			
			Node next = currNode;
			NamedNodeMap attrs = null;
			
			Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Parsing QM decision tree --- ");

			//clear Stream DT data structure
			((OnlineClassifierOperatorImp)OnlineClassifierRef).setNumStreams(((OnlineClassifierOperatorImp)OnlineClassifierRef).getNumStreams());
			//parse execution decision tree
			edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator.ClassVariableVisitor.ParseQMExecutionDecisionTree(((OnlineClassifierOperatorImp)OnlineClassifierRef), 
																											          next, 
																											          attrs);
			
			
			Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Parsing... Done --- ");
			
			//initialize all the necessary variables
			ArrayList<node> edt = null; //QM decision tree
			ArrayList<ArrayList<DecisionTreeNode>> StreamDT = null;
			ArrayList<HashMap<Integer,RusterNode>> RusterPerStream = null;
			node rootDt = null;
			
			//create the new classifier - execution DT
			int NumStreams = ((OnlineClassifierOperatorImp)OnlineClassifierRef).getNumStreams();				
			StreamDT = ((OnlineClassifierOperatorImp)OnlineClassifierRef).getStreamDT();
			RusterPerStream = ((OnlineClassifierOperatorImp)OnlineClassifierRef).getRusterPerStream();
			//execution DT
			edt = new ArrayList<node>(NumStreams);
			
			Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Creating new decision tree instance --- ");
			
			//clear route signatures (before creating new DT)
			DistributedExecutionController.hashSTQMRouteSignaturesGlobalStorage.clear();
			
			edt = currExecutionController.CreateExecutionDecisionTree(NumStreams, 
																	  edt, 
																	  rootDt,
																	  StreamDT,
																	  RusterPerStream);
			
			Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Creation... Done --- ");
			
			//update the classifier
			Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Updating online classifier with new decision tree instance --- ");
			((OnlineClassifierOperator)OnlineClassifierRef).setExecutionDecisionTree(edt);
			Helper.WriteToFile(Helper.LOG_FILE, Helper.GetCurrentTimeString() + ": --- Update of online classifier... Done --- ");
		}
		catch(ParserConfigurationException pce) 
		{
			pce.printStackTrace();
		}
		catch(SAXException se) 
		{
			se.printStackTrace();
		}
		catch(IOException ioe) 
		{
			ioe.printStackTrace();
		}
	}
	
	//-----------------------------------------------------------------------------
	
}
