//--------------------------------------------------
// STQM_Monitor.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2008
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.multiquery;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.BasicHist;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Coin;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.DecisionTreeID3;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.HistogramBuilderOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.StringBasicHist;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.OnlineClassifierOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.RusterNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.node;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.DistributedExecutionController;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.*;


public class MQM_Monitor 
{
	
    //-----------------------------------------------------------------------------
	
	//random sampling
	Coin flipCoin;
	Hashtable histogramsTable = new Hashtable();	
	//Hashtable usedNotUsedAttributes = new Hashtable();
	
    //-----------------------------------------------------------------------------
	
	//constructor
	public MQM_Monitor()
	{
		//System.out.println("Constructor call to STQM Monitor");
		flipCoin = new Coin();
	}
      	
    //-----------------------------------------------------------------------------
	
	public ArrayList DoClassifierDrivenSample(int streamId,
											  XATTuple[] tuples,         //tuples retrieved from the queue right now by the training tuple collector operator
											  ArrayList oldTuplesSample, //array list of old tuples samples (when QM was last updated)
											  int tuplesToSample,        //# of tuples needed in the sample
											  OnlineClassifierOperatorImp OnlineClassifierRef,
											  TrainingTupleCollectorOperatorImp physOp)
    {
		//--------------------------------------------------
		//              CLASSIFIER-DRIVEN SAMPLE
		//--------------------------------------------------	
    	XATTuple inputTuple = null;
    	ArrayList newTuplesSample = new ArrayList();	
    
		for (int p = 0; p < tuples.length; p++)
		{			
			inputTuple = tuples[p];
			newTuplesSample.add(inputTuple);
		}//end for	
			
		
	   	Hashtable usedNotUsedAttributes = physOp.usedNotUsedAttributes; //get used/not used hashtable for the current stream
	   	Hashtable histogramsTable = physOp.histogramsTable; //get histograms hashtable for the current stream
	   	
	   	
	   	//------------------------------------------
	   	// HANDLING OLD SAMPLES
	   	//------------------------------------------	   	
	   	//create training tuples header (used by dt - it contains some metadata, like names of the attributes)
	   	String currAllTuplesStr = "";
	   	
	   	currAllTuplesStr += HistogramBuilderOperatorImp.CreateTrainingTupleFileHeader(physOp.GetDecisionTreeAttributes(), 
	   																                  usedNotUsedAttributes) + System.getProperty("line.separator");
	   	
	   	//write the training tuples
	   	for (int p = 0; p < oldTuplesSample.size(); p++)
	   	{
	   		inputTuple = (XATTuple)oldTuplesSample.get(p);		   		
	   		String currentTupleStr = HistogramBuilderOperatorImp.CreateTrainingTupleString(inputTuple, usedNotUsedAttributes, histogramsTable);
	   		currAllTuplesStr += currentTupleStr + System.getProperty("line.separator");
	   	}
	   	
	   	//write tmp file for the decision tree
	   	String tmpFile = "C:\\QueryMesh\\tmp\\tmp.xml";
	   	Helper.WriteToFile(tmpFile, currAllTuplesStr);	
	   	
	   	String signatureOld = DecisionTreeID3.createDTSplitCriteriaSignature(tmpFile);
		
	   	//------------------------------------------
	   	// HANDLING NEW SAMPLES
	   	//------------------------------------------
	   	currAllTuplesStr = "";
	   	currAllTuplesStr += HistogramBuilderOperatorImp.CreateTrainingTupleFileHeader(physOp.GetDecisionTreeAttributes(), 
	                        usedNotUsedAttributes) + System.getProperty("line.separator");

	   	//write the training tuples
	   	for (int p = 0; p < oldTuplesSample.size(); p++)
	   	{
	   		inputTuple = (XATTuple)oldTuplesSample.get(p);		   		
	   		String currentTupleStr = HistogramBuilderOperatorImp.CreateTrainingTupleString(inputTuple, usedNotUsedAttributes, histogramsTable);
	   		currAllTuplesStr += currentTupleStr + System.getProperty("line.separator");
	   	}

	   	//write tmp file for the decision tree
	   	Helper.WriteToFile(tmpFile, currAllTuplesStr);	

	   	String signatureNew = DecisionTreeID3.createDTSplitCriteriaSignature(tmpFile);
		
	   	//check if new sample "deserves" to be analyzed further (based on the splitting criteria)
	   	
		return newTuplesSample;
    }//DoClusterDrivenSample    
	
    //-----------------------------------------------------------------------------
	
    public ArrayList DoRouteDrivenSample(int streamId,                    //current stream id
    								     ArrayList oldTuplesSample,       //array list of old tuples samples (when QM was last updated)
    									 XATTuple[] tuples,               //tuples retrieved from the queue right now by the training tuple collector operator
    									 int tuplesToSample,			  //# of tuples needed in the sample
    									 OnlineClassifierOperatorImp OnlineClassifierRef, //reference to the online classifier op
    									 Hashtable hashSTQMRouteSignaturesGlobalStorage)  //reference of the route signatures
    {
    	ArrayList newTuplesSample = new ArrayList();
    	
    	//--------------------------------------------------
    	//              ROUTE-DRIVEN SAMPLE
    	//--------------------------------------------------
    
    	int numTuplesToBeActuallySampled = 0;
    	
    	if (tuples.length > tuplesToSample)
    	{
    		numTuplesToBeActuallySampled = tuplesToSample;
    	}
    	else
    	{
    		numTuplesToBeActuallySampled = tuples.length;
    	}
    	    	
    	XATTuple inputTuple = null;    
    	ArrayList currentSample = new ArrayList();
    	
    	//add tuples from the current sample window to the arraylist  
		for (int p = 0; p < tuples.length; p++)
		{			
			inputTuple = tuples[p];			
			currentSample.add(inputTuple);			
		}//end for	

		//form signature for the old sample
		Hashtable<Integer, ArrayList> hashOldSampleClassificationTuples = new Hashtable<Integer, ArrayList>();;
		String oldSampleSignature = OnlineClassifierRef.GetPercentBasedClassificationSignature(streamId, 
																				   			   oldTuplesSample, 
				                                                                               hashSTQMRouteSignaturesGlobalStorage,
				                                                                               hashOldSampleClassificationTuples);
		
		//form signature for the new sample
		Hashtable<Integer, ArrayList> hashCurrentSampleClassificationTuples = new Hashtable<Integer, ArrayList>();
		String currentSampleSignature = OnlineClassifierRef.GetPercentBasedClassificationSignature(streamId, 
																					               currentSample, 
																					               hashSTQMRouteSignaturesGlobalStorage,
																					               hashCurrentSampleClassificationTuples);
		
		
		//iterate through signatures and perform sampling based on observed differences
		String [] arrOldSignature = oldSampleSignature.split(" ");
    	String [] arrNewSignature = currentSampleSignature.split(" ");

    	for (int i = 0; i < arrOldSignature.length; i++)
    	{  
    		String oldSignatureRouteId = arrOldSignature[i].toString().substring(0, (arrOldSignature[i].toString().length() - (arrOldSignature[i].toString().length() - arrOldSignature[i].toString().indexOf(":"))));
    		String oldSignatureRoutePercentage = arrOldSignature[i].toString().substring(arrOldSignature[i].toString().indexOf(":") + 1, arrOldSignature[i].toString().length());
    		
    		String newSignatureRouteId = arrNewSignature[i].toString().substring(0, (arrNewSignature[i].toString().length() - (arrNewSignature[i].toString().length() - arrNewSignature[i].toString().indexOf(":"))));
    		String newSignatureRoutePercentage = arrNewSignature[i].toString().substring(arrNewSignature[i].toString().indexOf(":") + 1, arrNewSignature[i].toString().length());
    		
    		//----------------------------------------------
    		// DETERMINE HOW TO SAMPLE
    		//----------------------------------------------
    		double numTuplesToSampleAssignedToCurrentRoute = 0;
    		//determine the # of samples that should be taken from current route-based tuples "partition"
    		if (Math.abs(Double.parseDouble(oldSignatureRoutePercentage) - Double.parseDouble(newSignatureRoutePercentage)) == 0)
    		{
    			//sample a fair proportion of tuples 
    			numTuplesToSampleAssignedToCurrentRoute = (numTuplesToBeActuallySampled / hashCurrentSampleClassificationTuples.size());
    		}
    		else
    		{
    			//sample a 'biased' proportion of tuples
    			int tmp = (numTuplesToBeActuallySampled / hashCurrentSampleClassificationTuples.size()) - 1;
    			double tmpPercentDiff = Math.abs(Double.parseDouble(oldSignatureRoutePercentage) - Double.parseDouble(newSignatureRoutePercentage)) / 100;
    			numTuplesToSampleAssignedToCurrentRoute = tmp + (tmpPercentDiff * tmp);
    		}
    		
    		//----------------------------------------------
    		// DO SAMPLE
    		//----------------------------------------------
    		ArrayList currentSubSample = new ArrayList(); //current sub-sample
    		ArrayList currentRouteTuples = (ArrayList)hashCurrentSampleClassificationTuples.get(Integer.parseInt(newSignatureRouteId));
    		int count = 0;
    		
    		if (currentRouteTuples == null)
    		{
    			continue;
    		}
    		
    		if (currentRouteTuples.size() < numTuplesToSampleAssignedToCurrentRoute)
    		{
    			numTuplesToSampleAssignedToCurrentRoute = currentRouteTuples.size();
    		}
    		
    		while (currentSubSample.size() != numTuplesToSampleAssignedToCurrentRoute)
    		{
    			//get the necessary # of tuples
    			currentSubSample.add(currentRouteTuples.get(count++));
    		}
    		newTuplesSample.addAll(currentSubSample);
    	}
		
		return newTuplesSample;
    }//DoRouteDrivenSample
    
    //-----------------------------------------------------------------------------
    
    public ArrayList DoRandomSample(XATTuple[] tuples, int tuplesToSample, Hashtable usedNotUsedAttributes)
    {
    	//--------------------------------------------------
    	//              RANDOM SAMPLE
    	//--------------------------------------------------	        
    	String currentTupleStr = "";
    	XATTuple inputTuple = null;
    	ArrayList oldTuplesSample = new ArrayList();
    	ArrayList newTuplesSample = new ArrayList();	
    
		for (int p = 0; p < tuples.length; p++)
		{			
			inputTuple = tuples[p];
			
			flipCoin.flip();
			if (flipCoin.toString().equals("Heads"))
			{			
				//BINGO! we will sample this tuple				
				newTuplesSample.add(inputTuple);										
			}//if heads was picked
			
		}//end for
		
		return newTuplesSample;
    }//end DoRandomSample
      
    //-----------------------------------------------------------------------------
    
    public String PrepareTupleString(XATTuple inputTuple, Hashtable usedNotUsedAttributes)
    {
		String currentTupleStr = "";
		String targetAttributeValue = "";
		//form a tab-spaced tuple string
		
		//on heads, sample the tuple, otherwise ignore it
		//add values to appropriate histograms					
		XATDataValue[] arrValues = inputTuple.getArrayData();
			
		for(int r = 0; r < inputTuple.getLength(); r++)
		{
			//-----------------------------------------------------------
			// ONLY INCLUDE THE ATTRIBUTE IF IT SHOULD BE USED IN THE
			// DECISION TREE
			//-----------------------------------------------------------
			String usedNotUsedThisAttribute = (String)usedNotUsedAttributes.get(r);
			if(usedNotUsedThisAttribute.startsWith("used"))
			{
				//enumerate all the contents of the hashtable
				String key;
				BasicHist updateBasicHistogram;
				StringBasicHist updateStringBasicHistogram;
							
				key = "" + r;
				Object objectHistogram = histogramsTable.get(key);
				XATDataValueImp xatVal = (XATDataValueImp)arrValues[Integer.parseInt(key)];
			   
			   if (xatVal.getValue() == null)
			   {
				   continue;
			   }	
				   
			   if (!histogramsTable.containsKey(key))
			   {								   
				   if(usedNotUsedThisAttribute.equals("used+target"))
					{
						//target attribute
						targetAttributeValue = inputTuple.getCell(r).toString().replace(" ", "") + "  ";
					}														
					else
					{							   									
					   //output the actual value of the attribute
					   currentTupleStr += inputTuple.getCell(r).toString().replace(" ", "") + "  ";					   
					   continue;
					}							  
			   }
				   							  
			   //---------------------------------------------------						
			   // check the type of the histogram
			   //---------------------------------------------------
		       if (objectHistogram != null)
		       {
				   if(objectHistogram.getClass().toString().equals("class edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.BasicHist"))
				   {
					   //double type histogram
					   updateBasicHistogram = (BasicHist)objectHistogram;	
					   
					   //if an exception occurs, don't take the tuple for
					   //the sample - data (especially real data, can be
					   //sometimes dirty - and we decided not to address this
					   //issue here
					   try
					   {
						   System.out.println("for key ["+key.toString()+"] value ["+ xatVal +"]");
						   String strVal = xatVal.convertToString();
						   strVal = strVal.replace("%", "");
						   double val = Double.parseDouble(strVal);
						   //add the current sample value to the histogram
						   updateBasicHistogram.add(val);
							   
						   //now retrieve the bucket value to output to the sample dump
						   String binNumber = "bin" + updateBasicHistogram.getBinByValue(val)+":"+updateBasicHistogram.getBinRangesByValue(val);
																   
						   if(usedNotUsedThisAttribute.equals("used+target"))									
						   {
							   //last column in the sample dump file
							   targetAttributeValue = binNumber + "  ";											
						   }														
						   else									
						   {							   											
							   currentTupleStr += binNumber + "  ";										
						   }									   
					   }									   
					   catch(Exception ex)									   
					   {										   
						   String msg = ex.getMessage();									
						   System.out.println("failed: " + msg);									  
					   }								   
				   }
					   
				   //------------------------------------------------------------							
				   // STRING HISTOGRAM
				   //------------------------------------------------------------
				   else								   
				   {									
					   //expect histogram type - class edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.StringBasicHist								
					   //string type histogram									  
					   updateStringBasicHistogram = (StringBasicHist)objectHistogram;	
						   									   
					   String val = xatVal.convertToString();
						   									   
					   //add the current sample value to the histogram									   
					   updateStringBasicHistogram.add(val);		
						   
					   //now retrieve the bucket value to output to the sample dump								
					   String binNumber = "bin" + updateStringBasicHistogram.getBinByValue(val);
						   									  
					   if(usedNotUsedThisAttribute.equals("used+target"))									   
					   {
						   //last column in the sample dump file									
						   targetAttributeValue = binNumber + "  ";									   
					   }									   
					   else									
					   {										
						   currentTupleStr += binNumber + "  ";									
					   }								   
				   }						       
		       }							
			}						
		}//for tuple length
		
		//add the target attribute value as the last column value				
		currentTupleStr += targetAttributeValue;
		
		return currentTupleStr;
    }//end prepareTuple
    
    //-----------------------------------------------------------------------------
    
}
