//--------------------------------------------------
// ExhaustiveSearch.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.DecisionTreeID3;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;

public class ExhaustiveSearch 
{	
	Hashtable CACHED_ROUTES = new Hashtable();
	private ArrayList minTrainingTuplePartitions;
	private LocalQueryMesh minQM;
	private double minQMCost = Double.MAX_VALUE;
	private LocalQueryMesh currentQM;
	private double currentQMCost;

	
	SimpleQueryPlanSpecification queryPlan;
	DecisionTreeID3 dt;
	Hashtable hashClassifierRoutesMapping = null;
	double totalQueryMeshCost = 0.0;
	int numTrainingTuples;
	S<String> s;
	//ArrayList allPartitions;
	
	public ExhaustiveSearch()
	{
		hashClassifierRoutesMapping = new Hashtable();
		minTrainingTuplePartitions = new ArrayList();
		
		//initialize cost as positive infinity
		minQMCost = Double.POSITIVE_INFINITY;
		currentQMCost = Double.POSITIVE_INFINITY;
	}
	
	
	
	public LocalQueryMesh runExhaustive(SimpleOperatorInputSpecification input, 
										ArrayList trainingTupleSet, 
										ArrayList partitionSpace, 
										List xmlOperators, 
										Hashtable xmlInputs)
	{
		//initialize query mesh for the current stream
		LocalQueryMesh streamQM = new LocalQueryMesh(input.getId());
		
		 DecisionTreeID3 currDT = null;
		 ArrayList currRoutes = new ArrayList();
		 
		 numTrainingTuples = trainingTupleSet.size();
		//---------------------------------------------------------------
	    // ITERATE THROUGH ALL POSSIBLE PARTITIONS
		//---------------------------------------------------------------
		for(int m = 0 ; m  < partitionSpace.size(); m++)
		{
			System.out.println("\n\n-----------------------------------------------------");
			System.out.println("done "+m+" out of "+partitionSpace.size());
			totalQueryMeshCost = 0.0;
    		S<S> currPartition = (S<S>)partitionSpace.get(m);        		
    		Object [] currSubsets = currPartition.toArray(); //all subsets in the current partition

			//---------------------------------------------------------------
		    // CREATE MAPPING BETWEEN TRAINING TUPLES AND SUBSETS IN THE CURRENT
    		// PARTITION
  		    //---------------------------------------------------------------
    		Hashtable trainingTuplesToSubsetsMapping = new Hashtable();
    		for(int p=1; p <= numTrainingTuples; p++)
    		{
    			String trainingTupleId = "" + p;
        		for(int i=0; i < currSubsets.length; i++)
        		{
        			if (currSubsets[i].toString().contains(trainingTupleId))
        			{
        				//create mapping between training tuple id and the subset 
        				String partitionName = "ruster"+currSubsets[i].toString();
        				trainingTuplesToSubsetsMapping.put(trainingTupleId, partitionName);
        				break;
        			}
        		}
    		}
		
    		//compute decision tree where target attribute is the partition of the tuple
    		String inputFileName = "C:\\QueryMesh\\tmp\\qmInput"+input.getId()+".txt";
    		String outputFileName = "C:\\QueryMesh\\tmp\\qmOutput"+input.getId()+".txt";
    		ArrayList trainingTuples = input.getTrainingTuples();
    		for (int i=0; i<trainingTuples.size(); i++)
    		{
    			String[] stringTokens = (String[])trainingTuples.get(i);
    			String stringToWrite = "";
    			for(int j=0; j < stringTokens.length; j++)
    			{   
    				if (j == stringTokens.length - 1)
    				{
    					if (i > 0)
    					{
        					//retrieve the ruster (partition) name for the current training tuple
    						String key = ""+i;
        					String rusterName = (String)trainingTuplesToSubsetsMapping.get(key);
        					//append the ruster as the last attribute (it will be a target attribute
        					//in the decision tree construction
        					stringToWrite += stringTokens[j] + "  "+ rusterName;
    					}
    					else
    					{
    						stringToWrite += stringTokens[j] + "  " + "RusterName";
    					}
    				}
    				else
    				{
    					stringToWrite += stringTokens[j] + "  ";
    				}
    			}        			
    			
    			if(i == 0)
				{
    				//delete sample dump (from previous runs)
					java.io.File file = new java.io.File (inputFileName);
			        file.delete ();
			    
			        DecisionTreeID3.WriteToDTDumpFile(inputFileName, stringToWrite);	    			
    				DecisionTreeID3.WriteToDTDumpFile(inputFileName, "//**************************************");
				}
    			else
    			{
    				DecisionTreeID3.WriteToDTDumpFile(inputFileName, stringToWrite);        			
    			}
    		}
		
		//reset mapping between DT leaves and the routes
		hashClassifierRoutesMapping = new Hashtable();
			
		//--------------------------------------------------------------------
		// COMPUTE DECISION TREE COST
		//--------------------------------------------------------------------
		dt = null;
		dt = DecisionTreeID3.optimizerCreateDT(input, inputFileName, outputFileName);
		double dtCost = dt.getDTCost(input.getStatisticsFileName());

		currDT = dt;
				 
		//--------------------------------------------------------------------
		// COMPUTE ROUTES' COST
		//--------------------------------------------------------------------
		
		ArrayList inputs;
		double selectivity;
		Hashtable hashPreparedInputs = new Hashtable();

		//List xmlOperators;
		//Hashtable xmlInputs;
		//iterate through hashtable of inputs and prepare OTHER (than the current
		//input) stream inputs

		Enumeration e = xmlInputs.elements();
		while (e.hasMoreElements()) 
		{  
			SimpleOperatorInputSpecification currInput = (SimpleOperatorInputSpecification)e.nextElement();  // do whatever you need with x
			if (currInput != input)
			{
				//only 'prepare' input if it is NOT the current input operator
				
				//compute cardinality of the input
				double currCardinality = currInput.getCardinality();
				OptimizerOpInput preparedInput = new OptimizerOpInput(currInput.getId(), currCardinality);
				hashPreparedInputs.put(currInput.getId(), preparedInput);
			}
		}
		
		//------------------------------------------------------------------
		// ITERATE THROUGH POSSIBLE RUSTERS OF THE CURRENT INPUT STREAM
		// AND COMPUTE THE ROUTES' COSTS
		//------------------------------------------------------------------
		
		double totalRoutesCost = 0;
		currRoutes = new ArrayList();
		
		for(int k = 0; k < currSubsets.length; k++)
		{
    		//now (for every leaf of the decision tree) of the current input 
    		//compute route cost
			String partitionName = "ruster"+currSubsets[k].toString();			
			
			QueryPlan currPlan;
			//if route is NOT cached
			if (!CACHED_ROUTES.containsKey(partitionName))
			{			
				int currCardinality = input.computeCardinalityOfPartition(partitionName, outputFileName);
				OptimizerOpInput preparedInput = new OptimizerOpInput(input.getId(), currCardinality);
				hashPreparedInputs.put(input.getId(), preparedInput);
				
	    		//------------------------------------------------------------------
	    		// ALL INPUTS ARE PREPARED
	    		//------------------------------------------------------------------
	
				ArrayList partitionTuples = input.getPartitionTuples((S)currSubsets[k]);
				
	    		//------------------------------------------------------------------
	    		// NOW PREPARE OPERATORS
	    		//------------------------------------------------------------------
				ArrayList allOperators = new ArrayList();
	    		for(int b = 0; b < xmlOperators.size(); b++)
	    		{
	    			SimpleOperatorSpecification currOp = (SimpleOperatorSpecification)xmlOperators.get(b);
	    			ArrayList currOpInputs = new ArrayList();
	    			
	    			boolean containsMeshInput = false; //true if input to operator is the stream for which we are computing query mesh
	    			
	    			//assert 2 inputs
	    			assert currOpInputs.size() == 2;
	    			
	    			//-------------------------------------------------------------
	    			// 1st OP INPUT
	    			//-------------------------------------------------------------
	    			boolean isPartitionBasedEstimation1 = false;
	    			boolean isPartitionBasedEstimation2 = false;
	    			String inputId1 = (String)currOp.getInputs().get(0);
	    			ArrayList op1PartitionTuples = new ArrayList();
	    			if (inputId1.equalsIgnoreCase(input.getId()))
	    			{
	    				isPartitionBasedEstimation1 = true;
	    				op1PartitionTuples = partitionTuples;
	    			}	        				
	    			//retrieve op by id
	    			OptimizerOpInput inp1 = (OptimizerOpInput)hashPreparedInputs.get(inputId1);
	    			SimpleOperatorInputSpecification input1 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId1);						
	    			int joinAttributeIdx1 = currOp.getJoinAttributeIdxByInputId(inputId1);
	    			currOpInputs.add(inp1);
	    			//-------------------------------------------------------------
	    			// 2nd OP INPUT
	    			//-------------------------------------------------------------	        			
	    			String inputId2 = (String)currOp.getInputs().get(1);
	    			ArrayList op2PartitionTuples = new ArrayList();
	    			if (inputId2.equalsIgnoreCase(input.getId()))
	    			{
	    				isPartitionBasedEstimation2 = true;
	    				op2PartitionTuples = partitionTuples;
	    			}	        				
	    			//retrieve op by id
	    			OptimizerOpInput inp2 = (OptimizerOpInput)hashPreparedInputs.get(inputId2);
	    			SimpleOperatorInputSpecification input2 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId2);
	    			int joinAttributeIdx2 = currOp.getJoinAttributeIdxByInputId(inputId2);	        			
	    			currOpInputs.add(inp2);
	    			
	    			//now estimate selectivity of the operator
	    				        			
	    			double currOpSelectivity = -999;
	    			
	    			
	    			JoinSelectivityEstimator selectivityEstimator = new JoinSelectivityEstimator(
																			        			input1,
																			        			isPartitionBasedEstimation1, 									
																			        			joinAttributeIdx1,
																			        			op1PartitionTuples,
																								input2, 
																								isPartitionBasedEstimation2,
																								joinAttributeIdx2,
																								op2PartitionTuples,
																								null);
	    			
	    			
	    			currOpSelectivity = selectivityEstimator.getSelectivity();	        				
	    			OptimizerOp op = new OptimizerOp(currOp.toString(), "join", currOpInputs, currOpSelectivity);
	    			
	    			allOperators.add(op);
	    		}//end of operator iterations
	    		
	    		queryPlan = new SimpleQueryPlanSpecification(allOperators);
	    		
	    		//get the plan
	    		currPlan = queryPlan.GetPlan(input.getId());    			    			    		
	    		//---------------------------------------------------------
	    		//             CACHE THE ROUTE FOR THE FUTURE
	    		//---------------------------------------------------------
	    		CACHED_ROUTES.put(partitionName, currPlan);
	    		
			}
			else
			{
				//retrieve route from cache
				currPlan = (QueryPlan)CACHED_ROUTES.get(partitionName);
			}
			
			double routeCost =  currPlan.getCost();
    		System.out.println("Cost :"+ routeCost);			 
     		
    		//add the plan to the curr routes list
    		currRoutes.add(new RouteSignature(currPlan, partitionName));    		
    		
    		totalRoutesCost += routeCost;
		}//end of subsest iterations
		
		totalQueryMeshCost = dtCost + totalRoutesCost;
		
		//-----------------------------------------------------------------
		//      UPDATE MIN SOLUTIONS BASED ON COSTS ESTIMATES
		//-----------------------------------------------------------------
		if (totalQueryMeshCost < minQMCost)
		{
			streamQM.setDTClassifier(currDT);
			streamQM.setRoutes(currRoutes);
			streamQM.setHashClassifierRoutesMapping(hashClassifierRoutesMapping);			
			minQM = streamQM;
			minQMCost = totalQueryMeshCost;
		}
		
		System.out.println("Curent query mesh cost: " + totalQueryMeshCost);
	}
		return minQM;
	}
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
public LocalQueryMesh runPartialExhaustive(SimpleOperatorInputSpecification input, 
									  ArrayList trainingTupleSet, 
			                          ArrayList partitionSpace, 
			                          List xmlOperators, 
			                          Hashtable xmlInputs,
			                          int iterationBound,//m -- the upper bound for the size of the full iteration subsets
			                          Hashtable CACHED_ROUTES_TO_RETURN)
{
	//---------------------------------------------------------------
	// ITERATION BOUND LOGIC IS AS FOLLOWS:
	// IF (# OF TRAINING TUPLES - PARTITION SIZE) < K
	// THIS PARTITION WILL BE ITERATED
	//---------------------------------------------------------------
		
	//initialize query mesh for the current stream
	LocalQueryMesh streamQM = new LocalQueryMesh(input.getId());
	
	DecisionTreeID3 currDT = null;
	ArrayList currRoutes = new ArrayList();
	numTrainingTuples = trainingTupleSet.size() - 1;
		
	//---------------------------------------------------------------
	// LIMIT THE PARTITIONS SET BY THE ITERATION BOUND
	//---------------------------------------------------------------
	ArrayList boundedPartitionSpace = new ArrayList();
	for (int p=0; p< partitionSpace.size();p++)
	{
		S<S> currPartition = (S<S>)partitionSpace.get(p);        		
		Object [] currSubsets = currPartition.toArray(); //all subsets in the current partition
		if ((numTrainingTuples - currSubsets.length) < iterationBound)
		{
			boundedPartitionSpace.add(currPartition);
		}
	}
	
	//---------------------------------------------------------------
	// ITERATE THROUGH ALL POSSIBLE PARTITIONS
	//---------------------------------------------------------------
	for(int m = 0 ; m  < boundedPartitionSpace.size(); m++)
	{
		totalQueryMeshCost = 0.0;
		S<S> currPartition = (S<S>)boundedPartitionSpace.get(m);        		
		Object [] currSubsets = currPartition.toArray(); //all subsets in the current partition
	
		//---------------------------------------------------------------
		// CREATE MAPPING BETWEEN TRAINING TUPLES AND SUBSETS IN THE CURRENT
		// PARTITION
		//---------------------------------------------------------------
		Hashtable trainingTuplesToSubsetsMapping = new Hashtable();
		for(int p=1; p <= numTrainingTuples; p++)
		{
			String trainingTupleId = "" + p;
	
			for(int i=0; i < currSubsets.length; i++)
			{
				if (currSubsets[i].toString().contains(trainingTupleId))
				{
					//create mapping between training tuple id and the subset 
					String partitionName = "ruster"+currSubsets[i].toString();
					trainingTuplesToSubsetsMapping.put(trainingTupleId, partitionName);
					break;
				}
			}
		}
	
		//compute decision tree where target attribute is the partition of the tuple
		String inputFileName = "C:\\QueryMesh\\tmp\\qmInput"+input.getId()+".txt";
		String outputFileName = "C:\\QueryMesh\\tmp\\qmOutput"+input.getId()+".txt";
		ArrayList trainingTuples = input.getTrainingTuples();
	
		for (int i=0; i<trainingTuples.size(); i++)
		{
			String[] stringTokens = (String[])trainingTuples.get(i);
			String stringToWrite = "";
			for(int j=0; j < stringTokens.length; j++)
			{   
				if (j == stringTokens.length - 1)
				{
					if (i > 0)
					{
						//retrieve the ruster (partition) name for the current training tuple
						String key = ""+i;
						String rusterName = (String)trainingTuplesToSubsetsMapping.get(key);
						//append the ruster as the last attribute (it will be a target attribute
						//in the decision tree construction
						stringToWrite += stringTokens[j] + "  "+ rusterName;
					}
					else
					{
						stringToWrite += stringTokens[j] + "  " + "RusterName";
					}
				}
				else
				{
					stringToWrite += stringTokens[j] + "  ";
				}
			}        			
	
		if(i == 0)
		{
			//delete sample dump (from previous runs)
			java.io.File file = new java.io.File (inputFileName);
			file.delete ();
			
			DecisionTreeID3.WriteToDTDumpFile(inputFileName, stringToWrite);	    			
			DecisionTreeID3.WriteToDTDumpFile(inputFileName, "//**************************************");
		}
		else
		{
			DecisionTreeID3.WriteToDTDumpFile(inputFileName, stringToWrite);        			
		}
	}
	
	//reset mapping between DT leaves and the routes
	hashClassifierRoutesMapping = new Hashtable();
	
	//--------------------------------------------------------------------
	// COMPUTE DECISION TREE COST
	//--------------------------------------------------------------------
	dt = null;
	dt = DecisionTreeID3.optimizerCreateDT(input, inputFileName, outputFileName);
	double dtCost = dt.getDTCost(input.getStatisticsFileName());
	
	currDT = dt;
	
	//--------------------------------------------------------------------
	// COMPUTE ROUTES' COST
	//--------------------------------------------------------------------
	
	ArrayList inputs;
	double selectivity;
	Hashtable hashPreparedInputs = new Hashtable();
	
	//List xmlOperators;
	//Hashtable xmlInputs;
	//iterate through hashtable of inputs and prepare OTHER (than the current
	//input) stream inputs
	
	Enumeration e = xmlInputs.elements();
	while (e.hasMoreElements()) 
	{  
		SimpleOperatorInputSpecification currInput = (SimpleOperatorInputSpecification)e.nextElement();  // do whatever you need with x
		if (currInput != input)
		{
			//only 'prepare' input if it is NOT the current input operator
	
			//compute cardinality of the input
			double currCardinality = currInput.getCardinality();
			OptimizerOpInput preparedInput = new OptimizerOpInput(currInput.getId(), currCardinality);
			hashPreparedInputs.put(currInput.getId(), preparedInput);
		}
	}
	
	//------------------------------------------------------------------
	// ITERATE THROUGH POSSIBLE RUSTERS OF THE CURRENT INPUT STREAM
	// AND COMPUTE THE ROUTES' COSTS
	//------------------------------------------------------------------
	currRoutes = new ArrayList();	
	double totalRoutesCost = 0;
	for(int k=0; k < currSubsets.length; k++)
	{
		//now (for every leaf of the decision tree) of the current input 
		//compute route cost
		String partitionName = "ruster"+currSubsets[k].toString();

		QueryPlan currPlan;
		//if route is NOT cached
		if (!CACHED_ROUTES.containsKey(partitionName))
		{	
			int currCardinality = input.computeCardinalityOfPartition(partitionName, outputFileName);
			OptimizerOpInput preparedInput = new OptimizerOpInput(input.getId(), currCardinality);
			hashPreparedInputs.put(input.getId(), preparedInput);
			
			//------------------------------------------------------------------
			// ALL INPUTS ARE PREPARED
			//------------------------------------------------------------------
			
			ArrayList partitionTuples = input.getPartitionTuples((S)currSubsets[k]);
					
			//------------------------------------------------------------------
			// NOW PREPARE OPERATORS
			//------------------------------------------------------------------
			ArrayList allOperators = new ArrayList();
			for(int b=0; b < xmlOperators.size(); b++)
			{
				SimpleOperatorSpecification currOp = (SimpleOperatorSpecification)xmlOperators.get(b);
				ArrayList currOpInputs = new ArrayList();
				
				boolean containsMeshInput = false; //true if input to operator is the stream for which we are computing query mesh
				
				//assert 2 inputs
				assert currOpInputs.size() == 2;
				
				//-------------------------------------------------------------
				// 1st OP INPUT
				//-------------------------------------------------------------
				boolean isPartitionBasedEstimation1 = false;
				boolean isPartitionBasedEstimation2 = false;
				String inputId1 = (String)currOp.getInputs().get(0);
				ArrayList op1PartitionTuples = new ArrayList();
				if (inputId1.equalsIgnoreCase(input.getId()))
				{
					isPartitionBasedEstimation1 = true;
					op1PartitionTuples = partitionTuples;
				}	        				
				//retrieve op by id
				OptimizerOpInput inp1 = (OptimizerOpInput)hashPreparedInputs.get(inputId1);
				SimpleOperatorInputSpecification input1 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId1);						
				int joinAttributeIdx1 = currOp.getJoinAttributeIdxByInputId(inputId1);
				currOpInputs.add(inp1);
				//-------------------------------------------------------------
				// 2nd OP INPUT
				//-------------------------------------------------------------	        			
				String inputId2 = (String)currOp.getInputs().get(1);
				ArrayList op2PartitionTuples = new ArrayList();
				if (inputId2.equalsIgnoreCase(input.getId()))
				{
					isPartitionBasedEstimation2 = true;
					op2PartitionTuples = partitionTuples;
				}	        				
				//retrieve op by id
				OptimizerOpInput inp2 = (OptimizerOpInput)hashPreparedInputs.get(inputId2);
				SimpleOperatorInputSpecification input2 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId2);
				int joinAttributeIdx2 = currOp.getJoinAttributeIdxByInputId(inputId2);	        			
				currOpInputs.add(inp2);
				
				//now estimate selectivity of the operator
							
				double currOpSelectivity = -999;
			
				JoinSelectivityEstimator selectivityEstimator = new JoinSelectivityEstimator(
													        			input1,
													        			isPartitionBasedEstimation1, 									
													        			joinAttributeIdx1,
													        			op1PartitionTuples,
																		input2, 
																		isPartitionBasedEstimation2,
																		joinAttributeIdx2,
																		op2PartitionTuples,
																		null);
		
				currOpSelectivity = selectivityEstimator.getSelectivity();	        				
				OptimizerOp op = new OptimizerOp(currOp.toString(), "join", currOpInputs, currOpSelectivity);
				
				allOperators.add(op);
			}//end of operator iterations
		
			queryPlan = new SimpleQueryPlanSpecification(allOperators);
			
			//get the plan
			currPlan = queryPlan.GetPlan(input.getId());    			    			    		
			//---------------------------------------------------------
			//             CACHE THE ROUTE FOR THE FUTURE
			//---------------------------------------------------------
			CACHED_ROUTES.put(partitionName, currPlan);		
		}
		else
		{
			//retrieve route from cache
			currPlan = (QueryPlan)CACHED_ROUTES.get(partitionName);
		}
		
		double routeCost =  currPlan.getCost();
		System.out.println("Cost :"+ routeCost);		
		
		//add the plan to the curr routes list
		currRoutes.add(new RouteSignature(currPlan, partitionName));    		
		
		totalRoutesCost += routeCost;
	}//end of subsest iterations
	
	totalQueryMeshCost = dtCost + totalRoutesCost;
	
	//-----------------------------------------------------------------
	//      UPDATE MIN SOLUTIONS BASED ON COSTS ESTIMATES
	//-----------------------------------------------------------------
	if (totalQueryMeshCost < minQMCost)
	{
		streamQM.setDTClassifier(currDT);
		ArrayList routes = (ArrayList)currRoutes.clone();
		streamQM.setRoutes(routes);
		streamQM.setHashClassifierRoutesMapping(hashClassifierRoutesMapping);			
		minQM = streamQM;
		minQMCost = totalQueryMeshCost;
	}
	
	System.out.println((m + 1) + "). Curent query mesh cost: " + totalQueryMeshCost);
	}
	
	//update the cache routes to return
	CACHED_ROUTES_TO_RETURN = CACHED_ROUTES;
	return streamQM;
	}//end run partial exhaustive

}
