//--------------------------------------------------
// IterativeImprovementSearch.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.StringTokenizer;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.DecisionTreeID3;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.QueryMeshOptimizer.QM_SEARCH_START_SOLUTION;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;


public class IterativeImprovementSearch 
{
	//-------------------------------------------------------------------------------------
	
	int NUM_ITERATIONS = 10; //serves as stop condition for II search
	
	//-------------------------------------------------------------------------------------
	
	Hashtable CACHED_ROUTES = new Hashtable();
	private ArrayList minTrainingTuplePartitions;
	private LocalQueryMesh minQM;
	private double minQMCost = Double.MAX_VALUE;
	private double currentQMCost;
		
	Hashtable hashClassifierRoutesMapping = null;
	DecisionTreeID3 currDT = null;
	ArrayList currRoutes = new ArrayList();
	
	//-------------------------------------------------------------------------------------
	
	public IterativeImprovementSearch()
	{
		hashClassifierRoutesMapping = new Hashtable();
		minTrainingTuplePartitions = new ArrayList();
		
		//initialize cost as positive infinity
		minQMCost = Double.POSITIVE_INFINITY;
		currentQMCost = Double.POSITIVE_INFINITY;
	}
	
	//-------------------------------------------------------------------------------------
	
	//------------------------------------------------------------------
	//  QUERY MESH ITERATIVE IMPROVEMENT
	//------------------------------------------------------------------
	/*
	 * Algorithm QM Iterative Improvement
	 1: minQM = null
	 2: while not stop condition do
	 3: QM = random solution
	 4: while not stop condition do
	 5: QM = random solution in NEIGHBORS(QM)
	 6: if (cost(QM ) < cost(QM)) then
	 7: QM = QM ;
	 8: end if
	 9: end while
	10: if (cost(QM) < cost(minQM) then
	11: minQM = QM;
	12: end if
	13: end while
	14: return minQM;
	 */

	//-------------------------------------------------------------------------------------
	
	SimpleQueryPlanSpecification queryPlan;
	DecisionTreeID3 dt;
	double dtCost = 0.0;
	double totalQueryMeshCost = 0.0;
	int numTrainingTuples;
	S<String> s;
	ArrayList allPartitions;
	
	//-------------------------------------------------------------------------------------
	
	public LocalQueryMesh runII(QM_SEARCH_START_SOLUTION startSolutionType,
								SimpleOperatorInputSpecification input, 
								ArrayList trainingTupleSet, 
								ArrayList partitionSpace, 
								List xmlOperators, 
								Hashtable xmlInputs,
								boolean isRandomStartSolution)
	{
				
		//initialize query mesh for the current stream
		LocalQueryMesh streamQM = new LocalQueryMesh(input.getId());
		
		//counts number of iterations by the algorithm
		int numIterations = 1;
	
		Hashtable hashTupleAttrTitles = new Hashtable();
		try
		{
			ArrayList partition = null;
			Object [] currSubsets = null;
			
			//array list into which training tuples will be placed per solution found
			ArrayList trainingTuplePartitions = new ArrayList();
			
			//num of trianing tuples
			numTrainingTuples = trainingTupleSet.size();
			
			BigInteger searchSpace = Bell.bigNumber(numTrainingTuples);		
			
			String currPartitionString = "";	
			Hashtable hashTrainingTupleMapping = new Hashtable();
					
			int count = 0;
			java.util.Enumeration en;			
			//----------------------------------------------------------
			// 1. PICK A START SOLUTION (BASED ON PARAMTER SPECIFIED BY USER)
			//----------------------------------------------------------
			switch (startSolutionType)
			{
				case RANDOM:
					
					//-------------------------START RANDOM START SOLUTION---------------------------------
					currSubsets = ComputeRandomStartSolution(hashTrainingTupleMapping);						
					//-------------------------END RANDOM START SOLUTION---------------------------------
					
					break;
				case EXTREME_1_ROUTE:
					
					//------------------------- EXTREME 1 START SOLUTION---------------------------------										
					currSubsets = ComputeExtreme1StartSolution(hashTrainingTupleMapping);					
					//------------------------- END EXTREME 1 START SOLUTION------------------------------------
					break;
									    
				case ROUTE_DRIVEN:
					
					//------------------------- ROUTE-DRIVEN START SOLUTION-------------------------------------					
					currSubsets = ComputeRouteDrivenStartSolution(hashTrainingTupleMapping);										
					//------------------------- END ROUTE-DRIVEN START SOLUTION---------------------------------
					break;
					
				case CONTENT_DRIVEN:
					
					//------------------------- CONTENT-DRIVEN START SOLUTION-----------------------------------					
					currSubsets = ComputeContentDrivenStartSolution(trainingTupleSet);					
					//------------------------- END CONTENT-DRIVEN START SOLUTION---------------------------------
					break;				
			}//end switch - done with start solution
									
			System.out.println();


			
			//-----------------------------------------------------------------
			//         2. DO DT AND ROUTES; PROCESSING
			//-----------------------------------------------------------------						
			
			System.out.println("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
			System.out.println((numIterations++) + ") Iteration\n");
			PrintCurrentPartitionSubsets(currSubsets);
			UpdatehashTrainingTupleMapping(currSubsets, 
										   hashTrainingTupleMapping);
			
			currentQMCost = ComputeDTAndRoutes(currSubsets, 
											   input, 
											   xmlOperators, 
											   xmlInputs,
											   hashTupleAttrTitles); //routes computed
					
			//-----------------------------------------------------------------
			//         3. IF CURRENT QM HAS THE SMALLES COST, STORE IT
			//-----------------------------------------------------------------						
			if (currentQMCost < minQMCost)
			{			
				streamQM.setDTClassifier(currDT);
				streamQM.setRoutes(currRoutes);
				streamQM.setHashClassifierRoutesMapping(hashClassifierRoutesMapping);
				streamQM.setCost(currentQMCost);
				minQMCost = currentQMCost;
				minQM = streamQM;
				
				PrintQM(minQM, "INTERMEDIATE(CHANGED)");
			}
			else
			{
				LocalQueryMesh discardedQM = new LocalQueryMesh(input.getId());	
				discardedQM.setDTClassifier(currDT);
				discardedQM.setRoutes(currRoutes);
				discardedQM.setHashClassifierRoutesMapping(hashClassifierRoutesMapping);	
				discardedQM.setCost(currentQMCost);
				PrintQM(discardedQM, "DISCARDED QM");
			}
		
			//-----------------------------------------------------------------
			//               4. RUN UNTIL STOP CONDITION
			//-----------------------------------------------------------------
			int stopCondition = NUM_ITERATIONS; //# iterations is the STOPPING CONDITION
			int i = 1;
		
			while (i < stopCondition)
			{
				System.out.println("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
				System.out.println((numIterations++) + ") Iteration");
				//-----------------------------------------------------------------
				//      5. PICK A RANDOM NEIGHBOR OF THE CURRENT SOLUTION
				//-----------------------------------------------------------------
				currSubsets = GetRandomNeighbor(hashTrainingTupleMapping);
				UpdatehashTrainingTupleMapping(currSubsets, 
						   				       hashTrainingTupleMapping);
				PrintCurrentPartitionSubsets(currSubsets);
				
				//-----------------------------------------------------------------
				//      6. ESTIMATE THE COST OF THE RANDOMLY PICKED NEIGHBOR
				//-----------------------------------------------------------------				

				currentQMCost = ComputeDTAndRoutes(currSubsets, 
												   input, 
												   xmlOperators, 
												   xmlInputs,
												   hashTupleAttrTitles); //routes computed
								
				//-----------------------------------------------------------------
				//      7. UPDATE MIN SOLUTIONS BASED ON COSTS ESTIMATES
				//-----------------------------------------------------------------
				if (currentQMCost < minQMCost)
				{
					streamQM.setDTClassifier(currDT);
					streamQM.setRoutes(currRoutes);
					streamQM.setHashClassifierRoutesMapping(hashClassifierRoutesMapping);	
					streamQM.setCost(currentQMCost);
					minQMCost = currentQMCost;
					minQM = streamQM;
					
					PrintQM(minQM, "INTERMEDIATE(CHANGED)");
				}
				else
				{
					LocalQueryMesh discardedQM = new LocalQueryMesh(input.getId());	
					discardedQM.setDTClassifier(currDT);
					discardedQM.setRoutes(currRoutes);
					discardedQM.setCost(currentQMCost);
					discardedQM.setHashClassifierRoutesMapping(hashClassifierRoutesMapping);
					
					PrintQM(discardedQM, "DISCARDED QM");
				}
				
				//increment num of iterations
				i++;
			}			
		}
		catch(Exception exp)
		{
			exp.printStackTrace();
		}
		
		//update optimizer's statistics "catalog"		
		QueryMeshOptimizer.hashNumIterationsPerStream.put(input.getInputName(), numIterations - 1);
		
		PrintQM(minQM, "FINAL");
		
		return minQM;
	}
	
	//-------------------------------------------------------------------------------------
	
	void PrintQM(LocalQueryMesh qm, String type)
	{
		System.out.println("");
		System.out.println("----------------------------------------");
		if (type.equalsIgnoreCase("FINAL"))
		{
			System.out.println("----        FINAL RESULT        -----");
		}
		else if (type.equalsIgnoreCase("INTERMEDIATE(CHANGED)"))
		{
			System.out.println("----  INTERMEDIATE RESULT (CHANGED) -----");
		}
		else if (type.equalsIgnoreCase("DISCARDED QM"))
		{
			System.out.println("----  DISCARDED QM  -----");
		}
		System.out.println("-------------------------");
		System.out.println("Chosen query mesh cost: [" + qm.getCost() +"]");
		System.out.println("Chosen query mesh: " + qm.toString());
		System.out.println("-------------------------");
	}
	
	//-------------------------------------------------------------------------------------
	
	private void UpdatehashTrainingTupleMapping(Object [] currSubsets, 
												Hashtable hashTrainingTupleMapping)
	{
		hashTrainingTupleMapping.clear();//clear mapping first
		
		for(int m = 0; m < currSubsets.length; m++)
		{
			S currPartition = new S();
			String pid = "" + currSubsets[m];
			if (hashTrainingTupleMapping.containsKey(pid))
			{
				currPartition = (S)hashTrainingTupleMapping.get(pid);
			}
			else
			{
				currPartition = new S();
			}
			currPartition = (S)currSubsets[m];
			hashTrainingTupleMapping.put(pid, currPartition);
		}
	}
	
	//-------------------------------------------------------------------------------------

	private void PrintCurrentPartitionSubsets(Object [] currSubsets)
	{
		String currPartitionString = currSubsets.length + " subsets: ";
		for(int m = 0; m < currSubsets.length; m++)
		{
			currPartitionString += currSubsets[m];
		}
		System.out.println(currPartitionString);
	}
	
	//-------------------------------------------------------------------------------------
	
	public double ComputeDTAndRoutes(Object [] currSubsets, 
									 SimpleOperatorInputSpecification input,
									 List xmlOperators, 
									 Hashtable xmlInputs,
									 Hashtable hashTupleAttrTitles)
	{
		//reset mapping between DT leaves and the routes
		hashClassifierRoutesMapping = new Hashtable();
		//clear routes collection
		currRoutes.clear();
		
		String tabs = "\t";
		//---------------------------------------------------------------
	    // CREATE MAPPING BETWEEN TRAINING TUPLES AND SUBSETS IN THE CURRENT PARTITION
		//---------------------------------------------------------------
		Hashtable trainingTuplesToSubsetsMapping = new Hashtable();
		for(int p = 0; p <= numTrainingTuples; p++)
		{
			String trainingTupleId = "" + p;
			//check to which subset the current training tuple belongs to
			for(int i = 0; i < currSubsets.length; i++)
			{
				if (currSubsets[i].toString().contains(trainingTupleId))
				{
					//create mapping between training tuple id and the subset 
					String partitionName = "ruster"+currSubsets[i].toString();
					trainingTuplesToSubsetsMapping.put(trainingTupleId, partitionName);
					break;
				}
			}
		}
		
		//-----------------------------------------------------------------
		//         2. ESTIMATE THE COST FOR THE QM SOLUTION
		//-----------------------------------------------------------------
		
		//compute decision tree where target attribute is the partition of the tuple
		String inputFileName = "C:\\QueryMesh\\tmp\\qmInput"+input.getId()+".txt";
		String dtInputFileName = "C:\\QueryMesh\\tmp\\qmDtInput"+input.getId()+".txt";
		String outputFileName = "C:\\QueryMesh\\tmp\\qmOutput"+input.getId()+".txt";

		String stringToWrite = "";

		
		//-----------------------------------------------------------------
		//                 PART 0: WRITE SUBSETS TO A LOG
		//-----------------------------------------------------------------
		
		ArrayList trainingTuples = input.getTrainingTuples();
		
		//write ruster files
		WriteRusterFiles(inputFileName,
					     trainingTuples,
					     trainingTuplesToSubsetsMapping);		
				
		//-----------------------------------------------------------------
		//                       PART I: ROUTES
		//-----------------------------------------------------------------
			
		ArrayList inputs;
		double selectivity;
		Hashtable hashPreparedInputs = new Hashtable();
	
		//--------------------------------------------------------------------
		// STORE OVERALL CARDINALITIES
		//--------------------------------------------------------------------
		Enumeration e = xmlInputs.elements();
		while (e.hasMoreElements()) 
		{  
			SimpleOperatorInputSpecification currInput = (SimpleOperatorInputSpecification)e.nextElement();  
			if (currInput != input)
			{
				//compute cardinality of the input
				double currCardinality = currInput.getCardinality();
				OptimizerOpInput preparedInput = new OptimizerOpInput(currInput.getId(), currCardinality);
				hashPreparedInputs.put(currInput.getId(), preparedInput);
			}
		}			
		
		//------------------------------------------------------------------
		// ITERATE THROUGH POSSIBLE RUSTERS OF THE CURRENT INPUT STREAM
		// AND COMPUTE THE ROUTES' COSTS
		//------------------------------------------------------------------
				
		double totalRoutesCost = 0;
		for(int k = 0; k < currSubsets.length; k++)
		{
	    	//now for the current input (for every subset in its partition) compute route cost
			String partitionName = "ruster" + currSubsets[k].toString();
				
			QueryPlan currPlan;
			//if route is NOT cached
			if (!CACHED_ROUTES.containsKey(partitionName))
			{				
				int currCardinality = input.computeCardinalityOfPartition(partitionName, inputFileName);
				OptimizerOpInput preparedInput = new OptimizerOpInput(input.getId(), currCardinality);
				hashPreparedInputs.put(input.getId(), preparedInput);
				
		    	//---------------------------
		    	// ALL INPUTS ARE PREPARED
		    	//---------------------------
		
				ArrayList partitionTuples = input.getPartitionTuples((S)currSubsets[k]);
					
		    	//---------------------------
		    	// NOW PREPARE OPERATORS
		    	//---------------------------
				ArrayList allOperators = new ArrayList();
				System.out.println("\nOperators' selectivities for current subset:");
					
		    	for(int b = 0; b < xmlOperators.size(); b++)
		    	{
		    		SimpleOperatorSpecification currOp = (SimpleOperatorSpecification)xmlOperators.get(b);
		    		ArrayList currOpInputs = new ArrayList();
		    			
		    		boolean containsMeshInput = false; //true if input to operator is the stream for which we are computing query mesh
		    			
		    		//assert 2 inputs
		    		assert currOpInputs.size() == 2;
		    			
		    		//-------------------------------------------------------------
		    		// 1st OP INPUT
		    		//-------------------------------------------------------------
		    		boolean isPartitionBasedEstimation1 = false;
		    		boolean isPartitionBasedEstimation2 = false;
		    		
		    		String inputId1 = (String)currOp.getInputs().get(0);
		    		ArrayList op1PartitionTuples = new ArrayList();
		    		if (inputId1.equalsIgnoreCase(input.getId()))
		    		{
		    			isPartitionBasedEstimation1 = true;
		    			op1PartitionTuples = partitionTuples;
		    		}	        				
		    		
		    		//retrieve op by id
		    		OptimizerOpInput inp1 = (OptimizerOpInput)hashPreparedInputs.get(inputId1);
		    		SimpleOperatorInputSpecification input1 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId1);						
		    		int joinAttributeIdx1 = currOp.getJoinAttributeIdxByInputId(inputId1);
		    		currOpInputs.add(inp1);
		    		
		    		//-------------------------------------------------------------
		    		// 2nd OP INPUT
		    		//-------------------------------------------------------------	        			
		    		String inputId2 = (String)currOp.getInputs().get(1);
		    		ArrayList op2PartitionTuples = new ArrayList();
		    		
		    		if (inputId2.equalsIgnoreCase(input.getId()))
		    		{
		    			isPartitionBasedEstimation2 = true;
		    			op2PartitionTuples = partitionTuples;
		    		}	        				
		    		
		    		//retrieve op by id
		    		OptimizerOpInput inp2 = (OptimizerOpInput)hashPreparedInputs.get(inputId2);
		    		SimpleOperatorInputSpecification input2 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId2);
		    		int joinAttributeIdx2 = currOp.getJoinAttributeIdxByInputId(inputId2);	        			
		    		currOpInputs.add(inp2);
		    			
		    		
		    		//now estimate selectivity of the operator		    				        			
		    		double currOpSelectivity = -999;		    			
		    			
		    		JoinSelectivityEstimator selectivityEstimator = new JoinSelectivityEstimator(input1,
																			        			 isPartitionBasedEstimation1, 									
																			        			 joinAttributeIdx1,
																			        			 op1PartitionTuples,
																								 input2, 
																								 isPartitionBasedEstimation2,																									 joinAttributeIdx2,
																								 op2PartitionTuples,
																								 hashTupleAttrTitles);
		    			
		    		currOpSelectivity = selectivityEstimator.getSelectivity();
		    		String selectivityDecriptor = selectivityEstimator.getDescriptor(); //describes current selectitivity (overall/subset);
		    		
		    		System.out.println(tabs + partitionName + ", " + 
		    				           currOp.getId() + 
		    				           "("+currOp.getInputs().get(0)+ "," +
		   					               currOp.getInputs().get(1)+ 
		   					           "): " + 
		   					           selectivityDecriptor + ": " + 
		   					           currOpSelectivity);
		    		
		   			OptimizerOp op = new OptimizerOp(currOp.toString(), "join", currOpInputs, currOpSelectivity);		    					 
		   			allOperators.add(op);		    		
		    	}//end of operator iterations
		    				    		
		    	queryPlan = new SimpleQueryPlanSpecification(allOperators);
		    		
		    	//get the plan
		    	currPlan = queryPlan.GetPlan(input.getId());    			    			    		
		    	
		    	//---------------------------------------------------------
		    	//       STORE ROUTE IN CACHE FOR THE FUTURE
		    	//---------------------------------------------------------
		    	CACHED_ROUTES.put(partitionName, currPlan);				    					
			}
			else
			{
				//---------------------------------------------------------
				// ROUTE IS CACHED - RETRIEVE IT FROM CACHE
				//---------------------------------------------------------

				currPlan = (QueryPlan)CACHED_ROUTES.get(partitionName);
			}
				
			double routeCost =  currPlan.getCost();
			
			System.out.println("***");
	    	System.out.println(tabs + partitionName + "\t\t\t\t\t\t"+ currPlan.toString() +" cost :"+ routeCost);		
	    		
	    	//add the plan to the curr routes list
	    	currRoutes.add(new RouteSignature(currPlan, partitionName));
	    		
	    	totalRoutesCost += routeCost;			
		}//end of subsest iterations
				

		//determine DISTINCT routes and store in the hash table mapping each subset with a DISTINCT route
		Hashtable hashTupleSubsetsWithTheirRoutes = new Hashtable();
		Hashtable hashDistinctRoutesWithTheirTupleSubsets = new Hashtable();
		
		for (int p = 0; p < currRoutes.size(); p++)
		{
			RouteSignature subsetRouteSignature = (RouteSignature)currRoutes.get(p);
			String subsetName = subsetRouteSignature.ruster;
			
			String routeStr = subsetRouteSignature.route.toString();
			ArrayList <String> subsetsWithSuchRoute;
			
			//store subset name --> its route
			hashTupleSubsetsWithTheirRoutes.put(subsetName, routeStr);
			
			//add subset under distinct routes
			if (!hashDistinctRoutesWithTheirTupleSubsets.containsKey(routeStr))
			{
				subsetsWithSuchRoute = new ArrayList();
				subsetsWithSuchRoute.add(subsetName);
			}
			else
			{
				subsetsWithSuchRoute = (ArrayList)hashDistinctRoutesWithTheirTupleSubsets.get(routeStr);
				subsetsWithSuchRoute.add(subsetName);
			}
		}
		
		if (hashDistinctRoutesWithTheirTupleSubsets.size() == 1)
		{
			//there is only ONE route --> no need for DT (it is empty)
			dt = null;
			dtCost = 0;
		}
		else
		{
			//compute DT, when there is more than ONE route
		
			//-----------------------------------------------------------------
			//                   PART II: DECISION TREE
			//-----------------------------------------------------------------
						
			//write route files (these input files will be used by DT induction algorithm)
			WriteRouteFiles(dtInputFileName,
						    trainingTuples,
						    trainingTuplesToSubsetsMapping,
						    hashTupleSubsetsWithTheirRoutes);		
					

			//--------------------------------------------------------------------
			// COMPUTE DECISION TREE COST
			//--------------------------------------------------------------------
			dt = null;
			dt = DecisionTreeID3.optimizerCreateDT(input, dtInputFileName, outputFileName);
			dtCost = dt.getDTCost(input.getStatisticsFileName());
		
			//set the current dt to the newly computed dt
			currDT = dt;
		}
		
		totalQueryMeshCost = dtCost + totalRoutesCost;				
		return totalQueryMeshCost;		
	}

	//-------------------------------------------------------------------------------------
	
	private Object [] GetRandomNeighbor(Hashtable hashTrainingTupleMapping)
	{
		Object [] currSubsets = null;
		
		if (hashTrainingTupleMapping.size() > 1)
		{
			System.out.println("------------------------------------------------------------------------------------>MERGE_SUBSETS");
			currSubsets = HybridSearch.GetRandomNeighborMergeTwoSubsets(hashTrainingTupleMapping);
		}
		else
		{
			System.out.println("------------------------------------------------------------------------------------>SPLIT_SUBSETS");
			currSubsets = HybridSearch.GetRandomNeighborSplitSubset(hashTrainingTupleMapping);
		}
		return currSubsets;
	}
	
	
	//-------------------------------------------------------------------------------------
	
	private Object [] ComputeExtreme1StartSolution(Hashtable hashTrainingTupleMapping)
	{
		int count;
		Object [] currSubsets;
		Enumeration en;
		
		System.out.println("*** EXTREME 1 START SOLUTION ***");
		
		S currPartition = new S();
		for(int m = 0; m < numTrainingTuples; m++)
		{						
			String pid = "" + "0";
			if (hashTrainingTupleMapping.containsKey(pid))
			{
				currPartition = (S)hashTrainingTupleMapping.get(pid);
			}
			else
			{
				currPartition = new S();
			}
			currPartition.add(m + 1);
			hashTrainingTupleMapping.put(pid, currPartition);
		}
							
		currSubsets = new Object[hashTrainingTupleMapping.size()];
		System.out.print("Current partition ");
		count = 0;
		en = hashTrainingTupleMapping.elements(); 
		while (en.hasMoreElements()) 
		{
			S currSubset = (S)en.nextElement();
			currSubsets[count++] = currSubset; 
			System.out.print(currSubset); 
		}

		return currSubsets;
	}

	//-------------------------------------------------------------------------------------
	
	private Object [] ComputeRouteDrivenStartSolution(Hashtable hashTrainingTupleMapping)
	{
		int count = 0;
		Object [] currSubsets;
		Enumeration en;
		
		System.out.println("*** ROUTE-DRIVEN START SOLUTION ***");
		
		//1. iterate though every training tuple and assign it to its own "subset" to imitate route-driven approach
		//the subsequent steps will do the following:
		//    (a) estimate selectitivy for every tuple		
		//    (b) compute route for every tuple		
		//    (c) group-by route-equivalent routes
		
		S currPartition = new S();
		for(int m = 0; m < numTrainingTuples; m++)
		{						
			String pid = "" + m;
			if (hashTrainingTupleMapping.containsKey(pid))
			{
				currPartition = (S)hashTrainingTupleMapping.get(pid);
			}
			else
			{
				currPartition = new S();
			}
			currPartition.add(m + 1);
			hashTrainingTupleMapping.put(pid, currPartition);
		}
							
		currSubsets = new Object[hashTrainingTupleMapping.size()];
		System.out.print("Current partition ");
		count = 0;
		en = hashTrainingTupleMapping.elements(); 
		while (en.hasMoreElements()) 
		{
			S currSubset = (S)en.nextElement();
			currSubsets[count++] = currSubset; 
			System.out.print(currSubset); 
		}
		
		return currSubsets;
	}
	
	//-------------------------------------------------------------------------------------
	
	private Object [] ComputeContentDrivenStartSolution(ArrayList trainingTupleSet)
	{
		int count = 0;
		Object [] currSubsets;
		Enumeration en;
		
		System.out.println("*** CONTENT-DRIVEN START SOLUTION ***");
		//start solution is picked based on content of the data 
		//the main idea - form "content clusters" based on attribute values

		int numAttributes = ((String[])trainingTupleSet.get(0)).length - 1;
		Hashtable [] attrContentClusters = new Hashtable [numAttributes];

		for (int a = 0; a < numAttributes; a++)
		{
			attrContentClusters[a] = new Hashtable();
		}
		
		//----------------------------------------------------------------
		// form hashtables by attributes (where training tuples are grouped by their content)
		//----------------------------------------------------------------
		for(int m = 0; m < trainingTupleSet.size(); m++)
		{												
			String[] tupleAttributes = (String[])trainingTupleSet.get(m);
				
		    //iterate through tokens and organize them based on content
			//skip the last attribute (b/c based on our data, we assume that
			//last column is the "count" column
			for (int t = 0; t < (tupleAttributes.length - 1); t++)
			{
				String attrValue = tupleAttributes[t];
					
				if (attrContentClusters[t].containsKey(attrValue))
				{
					//there is already a training tuple with this value
					ArrayList arrTrainingTupleIds = (ArrayList)attrContentClusters[t].get(attrValue);
					arrTrainingTupleIds.add(m); //m is the id for the training tuple 
				}
				else
				{
					//there is NO training tuple with this value
					//there is already a training tuple with this value
					ArrayList arrTrainingTupleIds = new ArrayList();
					arrTrainingTupleIds.add(m); //m is the id for the training tuple 
					attrContentClusters[t].put(attrValue, arrTrainingTupleIds);
				}
			}						
		}// end for m

		// 3. for simplicity (pick an attribute) based on which closeness of tuples is determined
		int contenAttrForGrouping = 0;
		
		Hashtable hashTrainingTuplesInContentGroupsByAttr = (Hashtable)attrContentClusters[contenAttrForGrouping];
		
		//number of groups in the partition
		currSubsets = new Object[hashTrainingTuplesInContentGroupsByAttr.size()];
		
		Enumeration eContentGroups = hashTrainingTuplesInContentGroupsByAttr.keys(); 					
		while (eContentGroups.hasMoreElements()) 
		{	
			S currSubset = new S();
			String keySharedValue = (String)eContentGroups.nextElement();//shared value by tuples in the group
			ArrayList trainingTupleIdsInGroup = (ArrayList)hashTrainingTuplesInContentGroupsByAttr.get(keySharedValue);
			for (int x = 0; x < trainingTupleIdsInGroup.size(); x++)
			{
				currSubset.add(trainingTupleIdsInGroup.get(x));
			}
			currSubsets[count++] = currSubset; 
			System.out.print(currSubset); 
		}//end wile		

		return currSubsets;
	}
	
	//-------------------------------------------------------------------------------------
	
	private Object [] ComputeRandomStartSolution(Hashtable hashTrainingTupleMapping)
	{
		int count = 0;
		Object [] currSubsets;
		Enumeration en;
		String currPartitionString = "";
		
		System.out.println("*** RANDOM START SOLUTION ***");
		//------------------------- RANDOM START SOLUTION---------------------------------
		//pick a start solution in the "middle" of the search space		
		BigInteger startSolution = Bell.bigNumber(numTrainingTuples).divide(new BigInteger("2"));		
		
		//now get the partitions of the solution
		int[] solutionPartitions = Bell.bigGrowthString(startSolution, numTrainingTuples);
		
		ArrayList partitions = new ArrayList();				
		
		for(int m = 0; m < solutionPartitions.length; m++)
		{
			S currPartition = new S();
			String pid = "" + solutionPartitions[m];
			if (hashTrainingTupleMapping.containsKey(pid))
			{
				currPartition = (S)hashTrainingTupleMapping.get(pid);
			}
			else
			{
				currPartition = new S();
			}
			currPartition.add(m + 1);
			hashTrainingTupleMapping.put(pid, currPartition);
			currPartitionString += solutionPartitions[m] + ",";
		}
		
		System.out.println("------------------------------------------------");
		System.out.println("[RANDOMLY SELECTED] current partition signature {" + currPartitionString + "}");
		//print partitions in a different (maybe more understandable) format
		
		currSubsets = new Object[hashTrainingTupleMapping.size()];
		System.out.print("Current partition ");
		count = 0;
		en = hashTrainingTupleMapping.elements(); 
		while (en.hasMoreElements()) 
		{
			S currSubset = (S)en.nextElement();
			currSubsets[count++] = currSubset; 
			System.out.print(currSubset); 
		}
		
		return currSubsets;
	}
	
	//-------------------------------------------------------------------------------------
	
	//writes log files containing training tuples and their subset (ruster) name 
	private void WriteRusterFiles(String inputFileName, 
								  ArrayList trainingTuples,
								  Hashtable trainingTuplesToSubsetsMapping)
	{
		String stringToWrite = "";
		
		//write a header in the input DT file
		int numAttributes = ((String[])trainingTuples.get(0)).length;	
		for (int i = 0; i < numAttributes; i++)
		{
			stringToWrite += "val" + i + "  ";
		}
		
		stringToWrite += "RusterName";
		
		//delete sample dump (from previous runs)
		java.io.File file = new java.io.File (inputFileName); file.delete ();
        
		DecisionTreeID3.WriteToDTDumpFile(inputFileName, stringToWrite);
		DecisionTreeID3.WriteToDTDumpFile(inputFileName, "//**************************************");
		
		
		//now actually write training tuples
		for (int i = 0; i < trainingTuples.size(); i++)
		{
			stringToWrite = "";
			String[] stringTokens = (String[])trainingTuples.get(i);
								
			for(int j = 0; j < stringTokens.length; j++)
			{   
				if (j == stringTokens.length - 1)
				{							    	
					//retrieve the ruster (partition) name for the current training tuple
					String key = ""+i;
					
					//retrieve subset (ruster name)
	    			String rusterName = (String)trainingTuplesToSubsetsMapping.get(key);
	    				    			
	    			//append the ruster as the last attribute (it will be a target attribute
	    			//in the decision tree construction
	    			stringToWrite += stringTokens[j] + "  "+ rusterName;						
				}
				else
				{
					stringToWrite += stringTokens[j] + "  ";
				}
			}
			
			DecisionTreeID3.WriteToDTDumpFile(inputFileName, stringToWrite);	
		}//end for

	}
	
	//-------------------------------------------------------------------------------------

	//writes log files containing training tuples and their route string (plan) 
	private void WriteRouteFiles(String dtInputFileName, 
			  					 ArrayList trainingTuples,
			  					 Hashtable trainingTuplesToSubsetsMapping,
			  					 Hashtable hashTupleSubsetsWithTheirRoutes)
	{
		String stringToWrite = ""; //reset string to write
		//write a header in the input DT file
		int numAttributes = ((String[])trainingTuples.get(0)).length;	
		
		for (int i = 0; i < numAttributes; i++)
		{
			stringToWrite += "val" + i + "  ";
		}
		
		stringToWrite += "RouteName";
		
		//delete sample dump (from previous runs)
		java.io.File file = new java.io.File (dtInputFileName); file.delete ();
        
		DecisionTreeID3.WriteToDTDumpFile(dtInputFileName, stringToWrite);
		DecisionTreeID3.WriteToDTDumpFile(dtInputFileName, "//**************************************");
		
		
		//now actually write training tuples
		for (int i = 0; i < trainingTuples.size(); i++)
		{
			stringToWrite = "";
			String[] stringTokens = (String[])trainingTuples.get(i);
								
			for(int j = 0; j < stringTokens.length; j++)
			{   
				if (j == stringTokens.length - 1)
				{							    	
					//retrieve the ruster (partition) name for the current training tuple
					String key = ""+i;
					
					//retrieve subset (ruster name)
	    			String rusterName = (String)trainingTuplesToSubsetsMapping.get(key);
	    			
	    			//retrieve the subsets route
	    			String routeName = (String)hashTupleSubsetsWithTheirRoutes.get(rusterName);
	    			
	    			//append the ruster as the last attribute (it will be a target attribute
	    			//in the decision tree construction
	    			stringToWrite += stringTokens[j] + "  "+ routeName;						
				}
				else
				{
					stringToWrite += stringTokens[j] + "  ";
				}
			}        							
			DecisionTreeID3.WriteToDTDumpFile(dtInputFileName, stringToWrite);			
		}//end for
	}
	
	//-------------------------------------------------------------------------------------
	
}
	
