/*
 * ChoicePoint.java
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;


import java.util.*;

/* Gli oggetti ChoicePoint servono per memorizzare i punti di scelta.
   Gli attributi di questa classe sono: l'intero cont che permette di 
   memorizzare il valore che deve essere assegnata al campo caseControl
   del constraint che ha aperto l'alternativa, l'oggetto statoVar, istanza 
   della classe varsState, che permette di memorizzare le variabili che al 
   momento della creazione del punto di scelta non sono inizializzate e 
   l'oggetto statoStore, istanza della classe StatoStore, che permette di 
   memorizzare lo stato dello store e il vincolo che ha aperto l'alternativa. */
/**
 * The <code>ChoicePoint</code> class implements an abstraction of "the choice" 
 * during a not deterministic resolution of constraints.
 * 
 *
 * @author  Elio Panegai
 * @author  Elisabetta Poleo
 * @version 1.3.0
 */
class ChoicePoint implements java.io.Serializable {

    /**
     * Index for next case of resolution
     */
    int cont;      

    /**
     * State of variables.
     */    
    VarState   varsState;   

    /**
     * State of constraints store.
     */     
    StoreState storeState; 


    /* Costructors */
    
    /**
     * Creates a new instance of ChoicePoint.
     */    
    ChoicePoint() {
	this(0, null, null);
    }

    /**
     * Creates a new instance of ChoicePoint with the specified variables and 
     * store state.
     *
     * @param The index for next case of resolution.
     * @param The object for the variables state.
     * @param The object for the constraints store state.
     */
    ChoicePoint(int i, VarState vs, StoreState ss) {
	this.cont       = i;
	this.varsState  = vs;
	this.storeState = ss;
    }

}
