/*
 * Constraint.java
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;

import java.util.*;

/**
 * The <code>Constraint</code> class implements an abstraction of a generic 
 * constraint.
 * 
 *
 * @author  Elio Panegai
 * @author  Elisabetta Poleo
 * @version 1.3.0
 */
public class Constraint implements java.io.Serializable {

    /* Members Vars */
    /**
     * Firt argument of constraint.
     */
    protected Object arg1;   
    
    /**
     * Second argument of constraint.
     */    
    protected Object arg2; 
    
    /**
     * Third argument of constraint.
     */    
    protected Object arg3;
    
    /**
     * Fourth argument of constraint.
     */    
    protected Object arg4 = null; 
    
    /**
     * Identifier for the constraint type. Default is 0.
     */
    protected int cons; 
    /* 0=eq, 1=neq, 2=in, 3=nin, 4=disj, 5=ndisj, 6=union, 7=nunion,
       8=ge, 81=gt, 9=le, 91=lt, 10=sum, 11=sub, 12=mul, 13=div,
       14=mod, 15=forall, 16=neq_union, 20=interv, 22=select */
    
    /**
     * Identify the non deterministic resolution step
     */    
    protected int caseControl = 0;
        
    /**
     * Identify (true) if the constraint is in solved form.
     */  
    protected boolean bool = false; 
    
    //protected int ctx = Solver.ctx;
    /**
     * Identifier for the contex of resolution. Actually not used.
     */     
    protected int ctx = 0;
    

    /* Costructors */

    /**
     * Constructs an empty constraint.
     */
    public Constraint() {
	this.arg1 = null;
	this.arg2 = null;
	this.arg3 = null;
	this.cons = 0;
    }  


    /* Costruttore per vincoli unari. Questo costruttore non viene mai 
       utilizzato per i vincoli predefiniti perch quest'ultimi sono
       tutti binari o ternari. Questo costruttore pu essere utilizzato
       dall'utente per definire nuovi vincoli unari. */
    /**
     * Constructs an unary constraint of the constraint type passed with the object argument.
     * @param k An integer identifier for the constraint type.
     * @param o The object argument of the constraints.
     */
    public Constraint(int k, Object o) {
        this();
        this.cons = k; 
	this.arg1 = o;	
    } 


    /* Costruttori per vincoli binari */
    /**
     * Constructs a binary constraint of the constraint type passed with the two objects argument.
     * @param k An integer identifier for the constraint type.
     * @param o1 The first object argument of the constraints.
     * @param o2 The second object argument of the constraints.
     */
    public Constraint(Object o1, int k, Object o2) {
        this(k,o1);
	this.arg2 = o2;
    }  


    /**
     * Constructs a binary constraint of the constraint type passed with the two objects argument.
     * @param k An integer identifier for the constraint type.
     * @param o1 The first object argument of the constraints.
     * @param o2 The second object argument of the constraints.
     */
    public Constraint(int k, Object o1, Object o2) {
        this(o1, k, o2);
    }  

    
    /**
     * Constructs a binary constraint of the constraint type passed..
     * @param k An integer identifier for the constraint type.
     * @param o1 The first object argument of the constraints.
     * @param i  integer for the second argument of the constraints.
     */
    public Constraint(int k, Object o1, int i) {
        this(k, o1, new Integer(i));
    }  


    /* Costruttori per vincoli ternari */
    /**
     * Constructs a ternary constraint of the constraint type passed.
     * @param k An integer identifier for the constraint type.
     * @param o1 The first object argument of the constraints.
     * @param o2 The second object argument of the constraints.
     * @param o3 The third object argument of the constraints.
     */    
    public Constraint(int k, Object o1, Object o2, Object o3) {
        this(k, o1, o2);
	this.arg3 = o3;
    }  

    /**
     * Constructs a ternary constraint of the constraint type passed.
     * @param k An integer identifier for the constraint type.
     * @param o1 The first object argument of the constraints.
     * @param o2 The second object argument of the constraints.
     * @param o3 integer for the third object argument of the constraints.
     */    
    public Constraint(int k, Object o1, Object o2, int o3) {
        this(k, o1, o2);	
	arg3 = new Integer(o3);
    }  


    /**
     * Constructs a constraint equals to the constraint passed.
     * @param s A constraint to be duplicated.
    */  
    public Constraint(Constraint s) {
	this.cons = s.cons;
	this.arg1 = s.arg1;
	this.arg2 = s.arg2;
	this.arg3 = s.arg3;
	this.arg4 = s.arg4;
    }
    
    
    /**
     * Constructs a constraint of the constraint type passed by name with one object argument.
     * @param name Name identifier for the constraint type.
     * @param o1 The first object argument of the constraints.
     */        
   public Constraint(String name, Object o1) {
        this(0, o1);
	int code = Enviroment.name_to_code(name);
        if (code == -1) code = Enviroment.addName(name);       
        this.cons = code; 
    }  

   
    public Constraint(String name, Object o1, Object o2) {
        this(0, o1, o2);
	int code = Enviroment.name_to_code(name);
        if (code == -1) code = Enviroment.addName(name);
        this.cons = code; 
    }  

    
    public Constraint(String name, Object o1, Object o2, Object o3) {
        this(0, o1, o2, o3);
	int code = Enviroment.name_to_code(name);
        if (code == -1) code = Enviroment.addName(name);
        this.cons = code; 
    }  
    
    
    /* I vincoli predefiniti sono binari o ternari ma per la loro risoluzione in
       alcuni casi  necessario poter introdurre un quarto parametro. */
    protected Constraint(int k, Object o1, Object o2, Object o3, Object o4) {
        this(k, o1, o2, o3);
	this.arg4 = o4;
    }    


    
    /**
     * Gets the identifier corresponding to the next branch of non deterministic resolution
     * of this constraint.
     * @return An integer corresponding the next case of non deterministic resolution.
     */
    public int getAlternative() {
         return caseControl;
    }    
    
    
    /**
     * Gets an ordered vector with only the not null members building the constraint.
     * 
     * @return an ordered vector with not null members of the constraint
     */
    public Vector getMembers() {
	Vector s = new Vector();
	if (this.arg1 != null) s.add(this.arg1);
        if (this.arg2 != null) s.add(this.arg2);
        if (this.arg3 != null) s.add(this.arg3);
        if (this.arg4 != null) s.add(this.arg4);	
	return s;
    }
    
        
    /**
     * Gets the specified argument of the constraints.
     * @param i index of the argument.
     * @return The object i-th argument of the constraint.
     */
    protected Object getArg(int i) {   
       if (i==1) return this.arg1;
       if (i==2) return this.arg2;
       if (i==3) return this.arg3;
       if (i==4) return this.arg4;
       else return null;
    }
    
    
    /**
     * Gets the integer code idetifying the constraint.
     * @return integer idetifying the constraint.
     */
    protected int getCode() {
         return this.cons;
    }

    
    /**
     * Gets the name idetifying the constraint.
     * @return A string idetifying the constraint.
     */
    public String getName() {
         return Enviroment.code_to_name(this.cons);
    }    


    /**
     * Sets the constraint as solved or not.
     * @param s the constraint solved form value.
     */
    public void setSolved(boolean s) {
        this.bool = s;
        return; 
    }    
    
    
    /**
     * Creates and returns a copy of this object.  The precise meaning of 
     * "copy" may depend on the class of the object.
     * @return a clone of this instance.
     */
    protected Constraint clone() {
        Constraint clonedObj = new Constraint();

        if (this.arg1 == null) clonedObj.arg1 = null;
        else if (this.arg1 instanceof Lvar) clonedObj.arg1 = (Lvar)((Lvar)this.arg1).clone();
        else if (this.arg1 instanceof Lst)  clonedObj.arg1 = (Lst)((Lst)this.arg1).clone(); 
        else if (this.arg1 instanceof Set)  clonedObj.arg1 = (Set)((Set)this.arg1).clone();
        else if (this.arg1 instanceof Int)  clonedObj.arg1 = (Int)((Int)this.arg1).clone(); 
        else clonedObj.arg1 = this.arg1; // XELIO maybe not correct!!!  (Cloning bit a bit)

        if (this.arg2 == null) clonedObj.arg1 = null;
        else if (this.arg2 instanceof Lvar) clonedObj.arg2 = (Lvar)((Lvar)this.arg2).clone();
        else if (this.arg2 instanceof Lst)  clonedObj.arg2 = (Lst)((Lst)this.arg2).clone();
        else if (this.arg2 instanceof Set)  clonedObj.arg2 = (Set)((Set)this.arg2).clone();
        else if (this.arg2 instanceof Int)  clonedObj.arg2 = (Int)((Int)this.arg2).clone(); 
        else clonedObj.arg2 = this.arg2;  // XELIO maybe not correct!!!  (Cloning bit a bit)

        if (this.arg3 == null) clonedObj.arg1 = null;
        else if (this.arg3 instanceof Lvar) clonedObj.arg3 = (Lvar)((Lvar)this.arg3).clone();
        else if (this.arg3 instanceof Lst)  clonedObj.arg3 = (Lst)((Lst)this.arg3).clone();
        else if (this.arg3 instanceof Set)  clonedObj.arg3 = (Set)((Set)this.arg3).clone();
        else if (this.arg3 instanceof Int)  clonedObj.arg3 = (Int)((Int)this.arg3).clone();
        else clonedObj.arg3 = this.arg3; // XELIO maybe not correct!!!  (Cloning bit a bit)

        if (this.arg4 == null) clonedObj.arg1 = null;
        else if (this.arg4 instanceof Lvar) clonedObj.arg4 = (Lvar)((Lvar)this.arg4).clone();
        else if (this.arg4 instanceof Lst)  clonedObj.arg4 = (Lst)((Lst)this.arg4).clone();
        else if (this.arg4 instanceof Set)  clonedObj.arg4 = (Set)((Set)this.arg4).clone();
        else if (this.arg4 instanceof Int)  clonedObj.arg4 = (Int)((Int)this.arg4).clone();         
        else clonedObj.arg4 = this.arg4; // XELIO maybe not correct!!!  (Cloning bit a bit)
 
        clonedObj.cons = this.cons; 
        clonedObj.caseControl = this.caseControl;
        clonedObj.bool = this.bool; 
        clonedObj.ctx  = this.ctx;
        return clonedObj;
    }
    
    
    /**
     * Creates and returns a bit a bit copy of this object.
     * @return a bit a bit clone of this instance.
     */
    protected Constraint cloneConstraint() {
	Constraint s = new Constraint();
	s.arg1 = this.arg1;
	s.arg2 = this.arg2;
	s.arg3 = this.arg3;
	s.arg4 = this.arg4;
	s.cons = this.cons;
	s.caseControl = this.caseControl;
	s.ctx  = this.ctx;
	s.bool = this.bool;
	return s;
    }
 
  
    /* Restituisce un nuovo vincolo ottenuto da quello su cui  invocato 
       sostituendo ogni variabile della lista Y con la variabile che si 
       trova nella stessa posizione nella lista Ynew. */    
    /**
     * Returns a new constraint equal to the invocator constraint but with items 
     * of the paramenter lists replaced. Every item of list y is replaced with the corrisponding 
     * item of list ynew.
     * @param Y list of item to be replaced
     * @param Ynew the replacing item list.
     * @return A new constraint with the specified list replacement.
     */       
    protected Constraint sostAll(Lst Y, Lst Ynew) {
	Constraint s = this.cloneConstraint();
	for(int i = 0; i < Y.size(); i++) {
	    s = s.sost((Lvar)Y.get(i), (Lvar)Ynew.get(i));
	}
	return s;
    }
  
 
    /* Restituisce un nuovo vincolo ottenuto da quello su cui  invocato 
       sostituendo la variabile Y con X. */
    /**
     * Returns a new constraint equal to the invocator constraint but with the paramenter 
     * varables replaced. Variable y is replaced with x.
     * @param Y The variable to be replaced
     * @param X The replacing variable.
     * @return A new constraint with the specified variable replacement.
     */
    protected Constraint sost(Lvar Y, Lvar X) {
	Constraint n = new Constraint();
	n.cons = cons;
	if(this.arg1 == Y) n.arg1 = X;
	else if(this.arg1 instanceof Set) n.arg1 = ((Set)this.arg1).sost(Y, X);
	else if(this.arg1 instanceof Lst) n.arg1 = ((Lst)this.arg1).sost(Y, X);
	else if(this.arg1 instanceof Constraint) n.arg1 = ((Constraint)this.arg1).sost(Y, X);
	else n.arg1 = arg1;
	if(this.arg2 == Y) n.arg2 = X;
	else if(this.arg2 instanceof Set) n.arg2 = ((Set)this.arg2).sost(Y, X);
	else if(this.arg2 instanceof Lst) n.arg2 = ((Lst)this.arg2).sost(Y, X);
	else if(this.arg2 instanceof Constraint) n.arg2 = ((Constraint)this.arg2).sost(Y, X);
	else n.arg2 = arg2;
	if(this.arg3 == null) n.arg3 = null;
	else if(this.arg3 == Y) n.arg3 = X;
	else if(this.arg3 instanceof Set) n.arg3 = ((Set)this.arg3).sost(Y, X);
	else if(this.arg3 instanceof Lst) n.arg3 = ((Lst)this.arg3).sost(Y, X);
	else if(this.arg3 instanceof Constraint) n.arg3 = ((Constraint)this.arg3).sost(Y, X);
	else n.arg3 = arg3;
	if(this.arg4 == null) n.arg4 = null;
	else if(this.arg4 instanceof Constraint) n.arg4 = ((Constraint)this.arg4).sost(Y, X);
	else if(this.arg4 instanceof Set) n.arg4 = ((Set)this.arg4).sost(Y, X);
	else if(this.arg4 instanceof Lst) n.arg4 = ((Lst)this.arg4).sost(Y, X);
	else if(this.arg4 instanceof Constraint) n.arg4 = ((Constraint)this.arg4).sost(Y, X);
	else n.arg4 = arg4;
	return n;
    } 


    /* Restituisce un nuovo vincolo ottenuto da quello su cui  invocato 
       sostituendo la lista L1 con L2. */
    /**
     * Returns a new constraint equal to the invocator constraint but with the paramenter 
     * lists replaced. List y is replaced with list x.
     * @param L1 list to be replaced
     * @param L2 the replacing list.
     * @return A new constraint with the specified list replacement.
     */    
    protected Constraint sost(Lst L1, Lst L2) {
	int i = 0;
	if(L1.size() > 0){
	    Constraint n = new Constraint();
	    n.cons = cons;
	    if(this.arg1 == L1.get(i)) n.arg1 = L2.get(i);
	    else if(this.arg1 instanceof Set) n.arg1 = ((Set)this.arg1).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg1 instanceof Lst) n.arg1 = ((Lst)this.arg1).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg1 instanceof Constraint) n.arg1 = ((Constraint)this.arg1).sost(L1,L2);
	    else n.arg1 = arg1;
	    if(this.arg2 == L1.get(i)) n.arg2 = L2.get(i);
	    else if(this.arg2 instanceof Set) n.arg2 = ((Set)this.arg2).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg2 instanceof Lst) n.arg2 = ((Lst)this.arg2).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg2 instanceof Constraint) n.arg2 = ((Constraint)this.arg2).sost(L1,L2);
	    else n.arg2 = arg2;
	    if(this.arg3 == null) n.arg3 = null;
	    else if(this.arg3 == L1.get(i)) n.arg3 = L2.get(i);
	    else if(this.arg3 instanceof Set) n.arg3 = ((Set)this.arg3).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg3 instanceof Lst) n.arg3 = ((Lst)this.arg3).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg3 instanceof Constraint) n.arg3 = ((Constraint)this.arg3).sost(L1,L2);
	    else n.arg3 = arg3;
	    if(this.arg4 == null) n.arg4 = null;
	    else if(this.arg4 instanceof Constraint) n.arg4 = ((Constraint)this.arg4).sost(L1,L2);
	    else if(this.arg4 instanceof Set) n.arg4 = ((Set)this.arg4).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg4 instanceof Lst) n.arg4 = ((Lst)this.arg4).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg4 instanceof Constraint) n.arg4 = ((Constraint)this.arg4).sost(L1,L2);
	    else n.arg4 = arg4;
	    i++;
	    return n.sost(i, L1, L2);
	}
	else return this;
    } 


    /* Restituisce un nuovo vincolo ottenuto da quello su cui  invocato 
       sostituendo il j-esimo elemento di L1 con il j-esimo elemento di L2
       per ogni j maggiore o uguale a i. */
    /**
     * Returns a new constraint equal to the invocator constraint but with the paramenter 
     * lists replaced from the i-throws object of the list to the tail. 
     * Items from i-throws position to the tail of the list L1 are replaced by corresponding items 
     * in the list L2.
     * @param i start index of replacing process.
     * @param L1 list to be replaced
     * @param L2 replacing list.
     * @return A new constraint with the specified list's items replacement.
     */           
    protected Constraint sost(int i, Lst L1, Lst L2) {
	if(L1.size() > i){
	    Constraint n = new Constraint();
	    n.cons = cons;
	    if(this.arg1 == L1.get(i)) n.arg1 = L2.get(i);
	    else if(this.arg1 instanceof Set) n.arg1 = ((Set)this.arg1).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg1 instanceof Lst) n.arg1 = ((Lst)this.arg1).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg1 instanceof Constraint) n.arg1 = ((Constraint)this.arg1).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else n.arg1 = arg1;
	    if(this.arg2 == L1.get(i)) n.arg2 = L2.get(i);
	    else if(this.arg2 instanceof Set) n.arg2 = ((Set)this.arg2).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg2 instanceof Lst) n.arg2 = ((Lst)this.arg2).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg2 instanceof Constraint) n.arg2 = ((Constraint)this.arg2).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else n.arg2 = arg2;
	    if(this.arg3 == null) n.arg3 = null;
	    else if(this.arg3 == L1.get(i)) n.arg3 = L2.get(i);
	    else if(this.arg3 instanceof Set) n.arg3 = ((Set)this.arg3).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg3 instanceof Lst) n.arg3 = ((Lst)this.arg3).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg3 instanceof Constraint) n.arg3 = ((Constraint)this.arg3).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else n.arg3 = arg3;
	    if(this.arg4 == null) n.arg4 = null;
	    else if(this.arg4 == L1.get(i)) n.arg4 = L2.get(i);
	    else if(this.arg4 instanceof Set) n.arg4 = ((Set)this.arg4).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg4 instanceof Lst) n.arg4 = ((Lst)this.arg4).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else if(this.arg4 instanceof Constraint) n.arg4 = ((Constraint)this.arg4).sost((Lvar)L1.get(i),(Lvar)L2.get(i));
	    else n.arg4 = arg4;
	    i++;
	    return n.sost(i, L1, L2);
	}
	else return this;
    } 


    /**
     * Creates and return a constraint conjunctiion  between the current 
     * constraint and the parameter constraint.
     * @param s a constraint to be and-joined with the current constraint.
     * @return A constraint conjunction.
     */
    public ConstraintsConjunction and(Constraint s) {
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(this);
	v.add(s);
	return v;
    }
    
    
    /**
     * Finds all the variables contained in the object parameter.
     * @param argN is the object where the variables are searched.
     * @param vettVars Vector modified by adding the variables found.
     */
    protected void findVars(Object argN, Vector vettVars) {
        if (argN != null) {
            if (argN instanceof Lvar) {
                // aggiunge Lvar
                vettVars.add((Lvar)argN);
                // nel caso in cui nell'Lvar sia memorizzato un set
                if (((Lvar)argN).value() instanceof Set){
                    //ricercare altre vars nel set
                    this.findVars((Set)((Lvar)argN).value(), vettVars);
                }
                // nel caso in cui nell'Lvar sia memorizzata una lista
                else if (((Lvar)argN).value() instanceof Lst){
                    //ricercare altre vars nella Lst
                    this.findVars((Lst)((Lvar)argN).value(), vettVars);
                }
            }
            if (argN instanceof Lst) {
                //ricercare altre vars nel Lst
                for (int i=0; i< ((Lst)argN).size_private(); i++ ) {
                   this.findVars(((Lst)argN).get(i), vettVars);
                }
            }
            if (argN instanceof Set) {
                //ricercare altre vars nel Set
                for (int i=0; i< ((Set)argN).size(); i++ ) {
                   this.findVars(((Set)argN).get(i), vettVars);
                }
            }
        }
        return;
    }
    
    
    public boolean equals(Constraint c){
        if (this.cons == c.cons) {
            if (((this.arg1 != null) && (c.arg1 != null) && (this.arg1.equals(c.arg1))) ||
                ((this.arg1 == null) && (c.arg1 == null)))
                if (((this.arg2 != null) && (c.arg2 != null) && (this.arg2.equals(c.arg2))) ||
                    ((this.arg2 == null) && (c.arg2 == null)))
                    if (((this.arg3 != null) && (c.arg3 != null) && (this.arg3.equals(c.arg3))) ||
                        ((this.arg3 == null) && (c.arg3 == null)))
                        if (((this.arg4 != null) && (c.arg4 != null) && (this.arg4.equals(c.arg4))) ||
                            ((this.arg4 == null) && (c.arg4 == null))) 
                             return true;                                
        }
        return false;                
    }
    
    
    public String toString() {
        String out = "Constraint type: " + cons; 
        /*  0=eq, 1=neq, 2=in, 3=nin, 4=disj, 5=ndisj, 6=union, 7=nunion,
            8=ge, 81=gt, 9=le, 91=lt, 10=sum, 11=sub, 12=mul, 13=div,
            14=mod, 15=forall, 16=neq_union, 20=interv, 22=select */
       if (this.arg1 != null) out += ", "+ "Arg1: "+ this.arg1.toString();
       else out += ", "+ "Arg1: null";
       if (this.arg1 != null) out += ", "+ "Arg2: "+ this.arg1.toString();
       else out += ", "+ "Arg2: null";
       if (this.arg1 != null) out += ", "+ "Arg3: "+ this.arg1.toString();
       else out += ", "+ "Arg3: null";
       if (this.arg1 != null) out += ", "+ "Arg4: "+ this.arg1.toString();
       else out += ", "+ "Arg4: null";                
       out += " Case control: "+ caseControl +", Resolved?: "+bool;
       return out;
    }       

}

