/*
 * Enviroment.java
 *
 * Created on 27 settembre 2004, 17.48
 */

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;

/**
 * The <code>Enviroment</code> class stores static values for default 
 * constraints used in the JSetL package.
 *
 * @author  Elio Panegai
 * @version 1.3
 */
public class Enviroment {
        
    
/**
 *  Initial value for constraints integer identifiers of JSetL.
 */    
    private static final int firstConstraintsCode    = 0;    
    
/**
 *  Initial value for the LibConstraints class constraints integer identifiers.
 */        
    private static final int firstLibConstraintsCode = 100;
    
/**
 *  Initial value for the NewConstraints class constraints integer identifiers.
 */         
    private static final int firstNewConstraintsCode = 500;
        
    
    static int constraintsCode;
    static int libConstraintsCode;
    static int newConstraintsCode;
        
    static String[] constraintNames = new String[firstNewConstraintsCode + firstLibConstraintsCode];
    
    /* JSetL defined constraints */
    static int eqCode;          // equality. Previous fixed code 0
    static int neqCode;         // inequality, previous fixed code 1
    static int neq_unionCode;   // union's case (6) e (7), previous fixed code 16
    static int inCode;          // in, previous fixed code 2
    static int ninCode;         // not in, previous fixed code 3
    static int selectCode;      // not determinist choice (similar to in), previous fixed code 22
    static int disjCode;        // disjunction, previous fixed code 4
    static int ndisjCode;       // not disjunction, previous fixed code 5
    static int unionCode;       // union, previous fixed code 6 
    static int nunionCode;      // not union, previous fixed code 7 
    static int geCode;          // greater or equal, previous fixed code 8
    static int gtCode;          // greater , previous fixed code 81
    static int leCode;          // less or equal, previous fixed code 9
    static int ltCode;          // less, previous fixed code 91
    static int sumCode;         // sum, previous fixed code 10
    static int subCode;         // difference, previous fixed code 11
    static int mulCode;         // product, previous fixed code 12
    static int divCode;         // quotient, previous fixed code 13
    static int modCode;         // module, previous fixed code 14
    static int forallCode;      // forall, previous fixed code 15
    static int intervCode;      // create intervals with variables bounds, previous fixed code 20
    static int integerCode;     // 
    static int nintegerCode;  // 
    static int cond_ifCode;     // not deterministic if, previous fixed code 30
    static int debugCode; 	// debugging, previous fixed code 77
        
//    static int concatCode; 
//    static int sizeCode; 

    static {                       
    /* generatin method's codes */ 
        constraintsCode     = firstConstraintsCode;
        libConstraintsCode  = firstLibConstraintsCode;        
        newConstraintsCode  = firstNewConstraintsCode;        
        
        /* JSetL methods codes */
        eqCode          = getMethodCode("constraintsCode");
        neqCode         = getMethodCode("constraintsCode");
        neq_unionCode   = getMethodCode("constraintsCode");
        inCode          = getMethodCode("constraintsCode");
        ninCode         = getMethodCode("constraintsCode");
        selectCode      = getMethodCode("constraintsCode");
        disjCode        = getMethodCode("constraintsCode");
        ndisjCode       = getMethodCode("constraintsCode");
        unionCode       = getMethodCode("constraintsCode");
        nunionCode      = getMethodCode("constraintsCode");
        geCode          = getMethodCode("constraintsCode");
        gtCode          = getMethodCode("constraintsCode");
        leCode          = getMethodCode("constraintsCode");
        ltCode          = getMethodCode("constraintsCode");
        sumCode         = getMethodCode("constraintsCode");
        subCode         = getMethodCode("constraintsCode");
        mulCode         = getMethodCode("constraintsCode");
        divCode         = getMethodCode("constraintsCode");
        modCode         = getMethodCode("constraintsCode");
        forallCode      = getMethodCode("constraintsCode");
        intervCode      = getMethodCode("constraintsCode");
        cond_ifCode     = getMethodCode("constraintsCode");
        debugCode       = getMethodCode("constraintsCode");       

        /* LibConstraints methods */
//        concatCode      = getMethodCode("libConstraintsCode"); 
//        sizeCode        = getMethodCode("libConstraintsCode");  
        
        constraintNames[eqCode]         = "eq";
        constraintNames[neqCode]        = "neq";
        constraintNames[neq_unionCode]  = "neq_union";
        constraintNames[inCode]         = "in";
        constraintNames[ninCode]        = "nin";
        constraintNames[selectCode]     = "select";
        constraintNames[disjCode]       = "disj";
        constraintNames[ndisjCode]      = "ndisj";
        constraintNames[unionCode]      = "union";
        constraintNames[nunionCode]     = "nunion";
        constraintNames[geCode]         = "ge";
        constraintNames[gtCode]         = "gt";
        constraintNames[leCode]         = "le";
        constraintNames[ltCode]         = "lt";
        constraintNames[sumCode]        = "sum";
        constraintNames[subCode]        = "sub";
        constraintNames[mulCode]        = "mul";
        constraintNames[divCode]        = "div";
        constraintNames[modCode]        = "mod";
        constraintNames[forallCode]     = "forall";
        constraintNames[intervCode]     = "interv";
        constraintNames[cond_ifCode]    = "cond_if";
        constraintNames[debugCode]      = "debug";       

        /* LibConstraints */
        // variabili definite in LibConstraints        
    }    
    
    
    /* Member Methods */
    /**
     * Get a new constraint code. Unique for all the classes in the package.
     *
     * @param       str a string that identfy the type of constraint.
     * @return      an integer, the code associated to the new method.
     * @exception   IllegalArgumentException if the string that identify the type
     *              of constraint is not correct.
     */
    protected static int getMethodCode(String str) throws IllegalArgumentException {
        /* increment politics, actually trivial */
        if (str.equals("constraintsCode")) return ++constraintsCode;
        else if (str.equals("libConstraintsCode")) return ++libConstraintsCode;
        else if (str.equals("newConstraintsCode")) return ++newConstraintsCode;
        else throw new IllegalArgumentException();
    }
    
     
    /**
     * Get a new constraint code for user-defined constraints. Unique for all 
     * the classes in the package.
     *
     * @return  an integer, the code associated to the new method.
     */    
    public static int getNewMethodCode() {
	return getMethodCode("newConstraintsCode");
    } 
    
    
    /**
     * Convert the integer constraint identifier to the associated String.
     * @param i an integer constraint idetifier.
     * @return The string corresponding to the integer constraint identifier.
     */
    public static String code_to_name(int i) {    
        return constraintNames[i];
    }


    /**
     * Convert the string constraint identifier to the associated (unique) integer.
     * @param name The constraint name.
     * @return The integer constraint identifier associated to the name passed.
     */
    public static int name_to_code(String name){
        return findName(name);
    }


    /**
     * Insert a new constraint name.
     * @param name The new constraint string identifier.
     * @return The integer identifier for the new constraint.
     */
    public static int addName(String name) { // throws IllegalArgumentException {
        int code = getMethodCode("newConstraintsCode");
//        if ((code > this.constraintNames.size()-1) ||
//            this.constraintNames[code] != "") throw new IllegalArgumentException();
//            else 
        constraintNames[code] = name;
        return code;
    }
          
    
    /**
     * Get the corresponding integer constraint identifier for the given string.
     */
    private static int findName(String name){
        for(int i = firstConstraintsCode; i < constraintsCode;i++) 
            if (name == constraintNames[i]) return i;
        for(int i = firstLibConstraintsCode; i < libConstraintsCode; i++) 
            if (name == constraintNames[i]) return i;
        for(int i = firstNewConstraintsCode; i < newConstraintsCode; i++) 
            if (name == constraintNames[i]) return i;
        return -1;
    }
    
}

