/*
 * Int.java
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;

import java.util.*;


/**
 * The <code>Int</code> class implements an abstraction of intervals.
 *
 * @author  Elio Panegai
 * @author  Elisabetta Poleo
 * @version 1.3.0
 */
public class Int {

    /* Member Vars */
    protected Integer sx;     // estremo sinistro dell'intervallo          
    protected Integer dx;     // estremo destro dell'intervallo   

    

    /* Costructors */
    protected Int() {
        this(Integer.MIN_VALUE/3, Integer.MAX_VALUE/3);
    }
    
    
    public Int(Integer a, Integer b) {
	this.sx = a;
	this.dx = b;
    }
    
    public Int(int a, int b) {
	this(new Integer(a), new Integer(b));
    }


    public Int(int a, Integer b) {
	this(new Integer(a), b);
    }


    public Int(Integer a, int b) {
        this(a, new Integer(b));
    }


    /* Member Methods */
    public boolean equals(Int I) {
	if(((Integer)this.sx).equals((Integer)I.sx) && 
	   ((Integer)this.dx).equals((Integer)I.dx)) return true;
	else return false;
    }


    public boolean isVoid() {
	if(((Integer)this.sx).intValue() > ((Integer)this.dx).intValue()) return true;
	else return false;
    }


    public Constraint eq(Object z)  { 
	Constraint s = new Constraint(this, Enviroment.eqCode, z); 
	return s;
    }     


    public Constraint neq(Object z) {
	Constraint s = new Constraint(this, Enviroment.neqCode, z); 
	return s;
    }   

 
    public void output() {
	System.out.print("Int("+this.sx+","+this.dx+")");
    } 


    public void print() {
	System.out.print("Int("+this.sx+","+this.dx+")");
    } 

    
    public String toString() {
	return "Int("+this.sx+","+this.dx+")";
    } 
    
    
    /**
     * Gets the value of the inferior end of the interval, Integer.MIN_VALUE if is not initialized.
     *
     * @return the value of the inferior end of the interval.
     */
    public int getInf() {
	if (this.sx == null) return Integer.MIN_VALUE; 
        else return ((Integer)this.sx).intValue();
    }
    
    
    /**
     * Gets the value of the superior end of the interval, Integer.MIN_VALUE if is not initialized.
     *
     * @return the value of the superior end of the interval.
     */    
    public int getSup() {
        if (this.dx == null) return Integer.MAX_VALUE; 
        else return ((Integer)this.dx).intValue();
    }

    
    /**
     * Checks if the interval have one of the admitted structure:
     *  inf == null & sup != null
     *  inf != null & sup == null
     *  inf <= sup 
     *
     * @return true if the interval have a proper structure, false otherwise.
     */
    public boolean checkInt(){
        if((this.sx == null && this.dx != null) ||
           (this.sx != null && this.dx == null) ||
           (this.getInf() == Integer.MIN_VALUE && this.dx != null) ||
           (this.sx != null && this.getSup() == Integer.MAX_VALUE) ||
           (this.getInf() <= this.getSup())) return true;
	else return false;
    }
    
    
    /**
     * Checks if the object is included in the interval.
     * @return true if the object is included in the interval, false otherwise.
     * @param o object to verify interval appartenence.
     */
    public boolean contains(Object o){        
         if((o instanceof Integer) && 
            (this.getInf() <= ((Integer)o).intValue()) && 
            (((Integer)o).intValue() <= this.getSup())) return true;
         return false;
    }    


    /**
     * Checks if the array of integer is an interval.
     * @return true if the array elements form an interval, false otherwise.
     * @param intArr the array to verify that be an interval.
     */
    public static boolean isInt(int[] intArr){
        int[] myArr = intArr.clone();
        Arrays.sort(myArr);

	for(int i = 0; i < myArr.length-1; i++) {
            if ( (myArr[i]+1) != myArr[i+1] ) return false;
	}
        return true;
    }  
    
    
    /**
     * Checks if the array of Object is an interval.
     *
     * @return true if the array elements form an interval, false otherwise.
     */
    private static boolean isInt(Object[] intArr){
        Object[] myArr = intArr.clone();
        Arrays.sort(myArr);
        
	for(int i = 0; i < myArr.length-1; i++) {
            if ( (((Integer)myArr[i]).intValue()+1) != ((Integer)myArr[i+1]).intValue()) return false;
	}
        return true;
    }
    
    
    /**
     * Get an interval builded with the elements of the array passed.
     * @return the interval builded with the elements of the array passed, null if is not possible build the interval.
     * @param arr the array of object for build an interval.
     */
    public static Int getIntFromObj(Object[] arr){
        if (isInt(arr)) {
            return new Int((Integer)arr[0], (Integer)arr[arr.length-1]);            
        }
        else return null;
    }
    
    
    /**
     * Get an interval builded with the elements of the array passed.
     * @return the interval builded with the elements of the array passed, null if is not possible build the interval.
     * @param arr the array of integer for build an interval.
     */
    public static Int getIntFromObj(int[] arr){
        if (isInt(arr)) {
            // se  un intervallo allora posso ordinare in modo definitivo l'array per determinare
            // il min e il max, chiaramente un getMin e getMax sarebbero pi veloci
            Arrays.sort(arr);
            return new Int(arr[0], arr[arr.length-1]);            
        }
        else return null;
    }
    
    
    public Int clone() {
        return new Int(this.sx.intValue(), this.dx.intValue());
    }
    
}
