/*
 * Lst.java
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;


import java.io.*;
import java.util.*;
import java.lang.Number.*;


/**
 * The <code>Lst</code> class implements an abstraction of logical list.
 *
 * @author  Elio Panegai
 * @author  Elisabetta Poleo
 * @version 1.3.0
 */
public class Lst implements java.io.Serializable {

    
    /* Member Vars */
    /**
     * Identify the list as initialized (true) or not (false).
     */
    protected boolean iniz = false; 
    
    /**
     * The (specified) elements in the list.
     */
    protected Vector lista;     
    
    /**
     * The unspecified rest of the list (a reference to another list).
     */
    protected Lst resto;      
    
    /**
     * Reference for binding equals list objects.
     */
    protected Lst equ = null;  
    
    /**
     * Unique identifier for the list.
     */
    protected int id;
    
    /**
     * Name of the list.
     */
    protected String  name = null;      
    
    /**
     * Vector of references to not initialized lists.
     */
    protected static Vector nonInizLst;      
    
    private   static int    counter;
    private   static String prefix;
    
    
    static {
        nonInizLst = new Vector();
        counter = 0;
        prefix = "Lst_";
    }
    
    /**
     * The empty list.
     */
    public static final Lst empty = new Lst("emptyLst", null, null);  

    
    /* Costructors */
    public Lst() {
	this.lista = null;
	this.resto = null;
	nonInizLst.add(this);
	this.id    = counter++;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n) {
        this();
	this.name = n;
    }

    
    private Vector wrapperLst() {
        Vector v   = new Vector();
	this.iniz  = true;
	this.resto = Lst.empty;  
        this.id    = counter++;
        return v;
    }
    
    
    public Lst(boolean[] arr) {                                      
	Vector v = this.wrapperLst();
	for (int i = 0; i < arr.length; i++) {
	    Boolean ii = new Boolean(arr[i]);
	    v.add(ii);
	}
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, boolean[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Boolean ii = new Boolean(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = n;
    }


    public Lst(byte[] arr) {                                      
        Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Byte ii = new Byte(arr[i]);
	    v.add(ii);
	}
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, byte[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Byte ii = new Byte(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = n;
    } 


    public Lst(short[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Short ii = new Short(arr[i]);
	    v.add(ii);
	}
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, short[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Short ii = new Short(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = n;
    }


    public Lst(int[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Integer ii = new Integer(arr[i]);
	    v.add(ii);
	}
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, int[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Integer ii = new Integer(arr[i]);
	    v.add(ii);
	}
	this.lista = v; 
	this.name  = n;
    } 


    public Lst(long[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Long ii = new Long(arr[i]);
	    v.add(ii);
	}
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, long[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Long ii = new Long(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = n;
    }

    public Lst(float[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Float ii = new Float(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, float[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Float ii = new Float(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = n;
    }


    public Lst(double[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Double ii = new Double(arr[i]);
	    v.add(ii);
	}
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, double[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Double ii = new Double(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = n;
    }


    public Lst(char[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Character c = new Character(arr[i]);
	    v.add(c);
	}
	this.lista = v;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, char[] arr) {                                      
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    Character c = new Character(arr[i]);
	    v.add(c);
	}
	this.lista = v;  
	this.name  = n;
    }


    public Lst(Object[] arr) {
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    v.add(arr[i]);  
	}
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, Object[] arr) {
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    v.add(arr[i]);  
	}
	this.lista = v;
	this.name  = n;
    }


    public Lst(Lst[] arr) {
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    v.add(arr[i]);  
	}
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, Lst[] arr) {
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    v.add(arr[i]);  
	}
	this.lista = v;
	this.name  = n;
    }


    public Lst(Lvar[] arr) {
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    v.add(arr[i]);  
	}
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, Lvar[] arr) {
	Vector v = this.wrapperLst();
	for(int i = 0; i < arr.length; i++) {
	    v.add(arr[i]);  
	}
	this.lista = v;
	this.name = n;
    }
  

    protected Lst(Vector v, Lst lst) {
	this.iniz  = true;
	this.lista = v;
	this.resto = lst;
        this.id    = counter++;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }

    protected Lst(String n, Vector v, Lst lst) {
        this(v, lst);
	this.name = n;
    }


    public Lst(Lst l) {
	this.iniz  = true;         
	this.lista = l.lista;
	this.resto = l.resto;
	this.equ  = l;
	this.id   = l.id; // counter++;
	this.name = prefix.concat((new Integer(counter++)).toString());
    }


    public Lst(String n, Lst l) {
        this(l);	        
	this.name = n;
    }


    public Lst(Lvar l) {
        Vector v = this.wrapperLst();
        v.add(l);
	this.lista = v;	
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, Lvar l) {
        this(l);	
	this.name = n;
    }


    public Lst(Set s) {
	Vector v = this.wrapperLst();
	v.add(s);	
	this.lista = v;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, Set s) {
        this(s);	
	this.name = n;
    }


    public Lst(int p, int q) {
	Vector v = this.wrapperLst();
	for(int i = p; i <= q; i++) {
	    Integer ii = new Integer(i); 
	    v.add(ii);
	}
	this.lista = v;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    public Lst(String n, int p, int q) {
        this(p, q);
	this.name = n;
    }
    
    
    
    /* Member Methods */  
        
    /**
     * Gets a list obtained by adding the Number paramenter as the head
     * element of the list.
     * @param n Number to insert at the head of the list.
     * @return A new Lst obtained from the invocationg list istance 
     * by adding the paramenter at the head.
     */
    public Lst ins1(Number n) { 
	if(this.equ == null) {
	    Vector v = new Vector(); 
	    v.add(n);
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1(n);
    }    
    

    /**
     * Gets a list obtained by adding the Object paramenter as the head
     * element of the list.
     * @param o Object to insert at the head of the list.
     * @return A new Lst obtained from the invocationg list istance 
     * by adding the paramenter at the head.
     */
    public Lst ins1(Object o) { 
	if(this.equ == null) {
	    Vector v = new Vector(); 
	    v.add(o);
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1(o);
    }
    

    /* I metodi ins1All restituiscono una Lst ottenuta da quella su cui  stato invocato il
       metodo aggiungendo in testa tutti gli elementi dell'array passato. */
    public Lst ins1All(boolean[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++) {
		Boolean o = new Boolean(arr[i]);
		v.add(o);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    }
  

    public Lst ins1All(byte[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++) {
		Byte o = new Byte(arr[i]);
		v.add(o);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    }


    public Lst ins1All(short[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++) {
		Short o = new Short(arr[i]);
		v.add(o);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    }


    public Lst ins1All(int[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++) {
		Integer o = new Integer(arr[i]);
		v.add(o);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    }


    public Lst ins1All(long[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++) {
		Long o = new Long(arr[i]);
		v.add(o);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    }


    public Lst ins1All(float[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++) {
		Float o = new Float(arr[i]);
		v.add(o);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    }


    public Lst ins1All(double[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++) {
		Double o = new Double(arr[i]);
		v.add(o);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    }


    public Lst ins1All(char[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++) {
		Character o = new Character(arr[i]);
		v.add(o);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    }    

 
    public Lst ins1All(Object[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();
	    for(int i = 0; i < arr.length; i++) {
		v.add(arr[i]);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    }


    public Lst ins1All(Lst[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();
	    for(int i = 0; i < arr.length; i++) {
		v.add(arr[i]);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    }


    public Lst ins1All(Lvar[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();
	    for(int i = 0; i < arr.length; i++) {
		v.add(arr[i]);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    } 


    public Lst ins1All(Set[] arr) { 
	if(this.equ == null) {
	    Vector v = new Vector();
	    for(int i = 0; i < arr.length; i++) {
		v.add(arr[i]);
	    }
	    Lst l = new Lst(v, this);
	    return l;  
	}
	else return this.equ.ins1All(arr);
    }


    /* I metodi insn restituiscono una Lst ottenuta da quella su cui  stato invocato il
       metodo aggiungendo in coda l'elemento passato. */    
    public Lst insn(Number elem) {
        if(this.equ == null) {
            Vector vet = this.toVector();
            vet.add(elem);
            Lst lst = new Lst(vet, empty);
            return lst;
        }
        else return this.equ.insn(elem);
    }


    public Lst insn(Object elem) { 
	if(this.equ == null) {
	    Vector vet = this.toVector();
	    vet.add(elem);
	    Lst lst = new Lst(vet, empty);
	    return lst;  
	}
	else return this.equ.insn(elem);
    }
    

    /* I metodi insnAll restituiscono una Lst ottenuta da quella su cui  stato invocato il
       metodo aggiungendo in coda tutti gli elementi dell'array passato. */
    public Lst insnAll(boolean[] arr) { 
	if(this.equ == null) {
	    Vector vet = this.toVector();
	    for(int i = 0; i < arr.length; i++) {
		Boolean b = new Boolean(arr[i]);
		vet.add(b);
	    }
	    Lst lst = new Lst(vet, empty);
	    return lst;  
	}
	else return this.equ.insnAll(arr);
    }


    public Lst insnAll(byte[] arr) { 
	if(this.equ == null) {
	    Vector vet = this.toVector();
	    for(int i = 0; i < arr.length; i++) {
		Byte b = new Byte(arr[i]);
		vet.add(b);
	    }
	    Lst lst = new Lst(vet, empty);
	    return lst;  
	}
	else return this.equ.insnAll(arr);
    }
    
    
    public Lst insnAll(short[] arr) { 
	if(this.equ == null) {
	    Vector vet = this.toVector();
	    for(int i = 0; i < arr.length; i++) {
		Short s = new Short(arr[i]); 
		vet.add(s);
	    }
	    Lst lst = new Lst(vet, empty);
	    return lst;  
	}
	else return this.equ.insnAll(arr);
    }


    public Lst insnAll(int[] arr) { 
	if(this.equ == null) {
	    Vector vet = this.toVector();
	    for(int i = 0; i < arr.length; i++) {
		Integer n = new Integer(arr[i]);
		vet.add(n);
	    }
	    Lst lst = new Lst(vet, empty);
	    return lst;  
	}
	else return this.equ.insnAll(arr);
    }


    public Lst insnAll(long[] arr) { 
	if(this.equ == null) {
	    Vector vet = this.toVector();
	    for(int i = 0; i < arr.length; i++) {
		Long l = new Long(arr[i]);
		vet.add(l);
	    }
	    Lst lst = new Lst(vet, empty);
	    return lst;  
	}
	else return this.equ.insnAll(arr);
    }


    public Lst insnAll(float[] arr) { 
	if(this.equ == null) {
	    Vector vet = this.toVector();
	    for(int i = 0; i < arr.length; i++) {
		Float f = new Float(arr[i]);
		vet.add(f);
	    }
	    Lst lst = new Lst(vet, empty);
	    return lst;  
	}
	else return this.equ.insnAll(arr);
    }


    public Lst insnAll(double[] arr) { 
	if(this.equ == null) {
	    Vector vet = this.toVector();  
	    for(int i = 0; i < arr.length; i++) {
		Double d = new Double(arr[i]);
		vet.add(d);
	    }
	    Lst lst = new Lst(vet, empty);
	    return lst;  
	}
	else return this.equ.insnAll(arr);
    }


    public Lst insnAll(char[] arr) { 
	if(this.equ == null) {
	    Vector vet = this.toVector();
	    for(int i = 0; i < arr.length; i++) {
		Character c = new Character(arr[i]);
		vet.add(c);
	    }
	    Lst lst = new Lst(vet, empty);
	    return lst;  
	}
	else return this.equ.insnAll(arr);
    }


    public Lst insnAll(Object[] arr) { 
	if(this.equ == null) {
	    Vector vet = this.toVector();
	    for(int i = 0; i < arr.length; i++) {
		vet.add(arr[i]);
	    }
	    Lst lst = new Lst(vet, empty);
	    return lst;  
	}
	else return this.equ.insnAll(arr);
    }

    
    /* Deterministic concat */
    
    /**
     * returns the list obtained by concatenating the invocating list and the parameter list.
     * @param l the list to append to this object
     * @return the concatenated list.
     */
    public Lst concat(Lst l) { 
	if(this.equ == null) {
	    Vector vetThis = this.toVector();
	    Vector vetl = l.toVector();
	    Vector vett = new Vector();
	    vett.addAll(vetThis);
	    vett.addAll(vetl);
	    Lst lst = new Lst(vett, empty);
	    return lst;  
	}
	else return this.equ.concat(l);
    }


    public Lst concat(Lvar l) { 
	if(this.equ == null) {
	    if(l.val instanceof Lst) return this.concat((Lst)l.val);
	    else throw new NotLstException();
	}
	else return this.equ.concat(l);
    }


    /* I metodi ext1 restituiscono una Lst ottenuta da quella su cui  stato invocato il
       metodo estraendo dalla lista il primo elemento, mentre i metodi extn restituiscono 
       una Lst ottenuta da quella su cui  stato invocato il metodo estraendo dalla lista 
       l'ultimo elemento. Se la lista su cui sono invocati i metodi  vuota, viene sollevata 
       un'eccezione EmptyLstException, se la lista non  inizializzata viene sollevata
       un'eccezione NotInitVarException e se la Lvar passata come parametro  inizializzata
       viene sollevata un'eccezione InitLvarException. In ogni caso la lista su cui viene 
       invocato il metodo non viene modificata.*/
    public Lst ext1(Lvar l) throws InitLvarException { 
	if(this.equ == null) {
	    if(!this.iniz) throw new NotInitVarException();  
	    else if(this == empty) throw new EmptyLstException();  
	    else if(!l.iniz) {
		Vector vet = this.toVector();
		if(vet.get(0) instanceof Lvar)l.equ = (Lvar)vet.get(0);
		else {
		    l.iniz = true;
		    l.val = vet.get(0);
		}
		vet.removeElementAt(0);
		Lst lst = new Lst(vet, this.ultimoResto());
		return lst;
	    }
	    else throw new InitLvarException();  
	}
	else return this.equ.ext1(l);
    }


    public Lst extn(Lvar l) throws InitLvarException { 
	if(this.equ == null) {
	    if(!this.iniz) throw new NotInitVarException();  
	    else if(this == empty) throw new EmptyLstException();  
	    else if(!l.iniz) {
		Vector vet = this.toVector();
		if(vet.get(vet.size() - 1) instanceof Lvar)l.equ = (Lvar)vet.get(vet.size() - 1);
		else {
		    l.iniz = true;
		    l.val = vet.get(vet.size()-1);
		}
		vet.removeElementAt(vet.size()-1);
		Lst lst = new Lst(vet, this.ultimoResto());
		return lst;
	    }
	    else throw new InitLvarException();  
	}
	else return this.equ.extn(l);
    }
  

    public Lst ext1() { 
	if(this.equ == null) {
	    if(!this.iniz) throw new NotInitVarException();  
	    else if(this == empty) throw new EmptyLstException();  
	    else {
		Vector vet = this.toVector();
		vet.removeElementAt(0);
		Lst lst = new Lst(vet, ultimoResto());
		return lst;
	    }
	}
	else return this.equ.ext1();
    }


    public Lst extn() { 
	if(this.equ == null) {
	    if(!this.iniz) throw new NotInitVarException();  
	    else if(this == empty) throw new EmptyLstException();  
	    else {
		Vector vet = this.toVector();
		vet.removeElementAt(vet.size()-1);
		Lst lst = new Lst(vet, this.ultimoResto());
		return lst;
	    }
	}
	else return this.equ.extn();
    }
  

    /* costruisce una lista contenente n Lvar non inizializzate */
    public static Lst mkLst(int n) {
	Vector lista = new Vector();   
	for(int i = 0; i < n; i++) {
	    Lvar lvar = new Lvar();
	    lista.add(lvar);    
	}
	Lst lst = new Lst(lista, empty);
	return lst;
    }


    /* restituisce l'oggetto che si trova in i-esima posizione */
    public Object get(int i) { 
	if(this.equ == null) {
	    Vector v = this.toVector();
	    return v.get(i);
	}
	else return this.equ.get(i);
    }


    /**
     * Returns the length of the list, that is the number of elements of the list.
     * @return The integer number of object in the list, 0 for the empty list. 
     * If the list is not initialized or unbounded a NotInitVarException is raised.
     * @throws JSetL.NotInitVarException if the list is not initialized.
     */
    public int size() throws NotInitVarException { 
	if(this.equ == null) {
	    if(((this.ultimoResto()).known()) && (this.iniz == true)) {
		Vector v = this.toVector();
		return v.size();
	    }
	    else throw new NotInitVarException();
	}
	else return this.equ.size();
    }


    /**
     * Returns the length of the list, that is the number of elements of the list.
     * If the list is not initialized or unbounded none exceptions are raised.
     * @return The integer number of object in the list, 0 for the empty list. 
     */
    protected int size_private() { 
	if(this.equ == null) {
	    Vector v = this.toVector();
	    return v.size();
	}
	else return this.equ.size();
    }
    
    
    /*  restituisce il nome della lista */
    public String setName() {
	return this.name;
    }

    /**
     * Gets the unique identifier  for the object.
     * @return an int identifier for the object.
     */
    protected int getId() {
	return this.id;
    }    
    
    /**
     * Sets the unique identifier for the object.
     */
    protected void setId() {
        this.id = counter++;
        return;
    }        
    
    
    public String getName() {
	return this.name;
    }


    /*  Il metodo output stampa il nome della Lst su cui  stato invocato e il suo valore. */
    public void output() {
	System.out.println();
	System.out.print(this.name + " = ");
	if(!this.iniz)System.out.print("unknown");
	else this.print();
	System.out.println();
	return;
    }


    /* Il metodo print, se la lista su cui  stato invocato  inizializzata, stampa il 
       suo valore, altrimenti stampa il suo nome preceduto da "_" per poter distinguere 
       i nomi dalle stringhe o caratteri usate come valore per le variabili.
       Per le liste e gli insiemi l'output  dato secondo le convenzioni sintattiche 
       del Prolog. */
    public void print() {
	if(this.equ == null) {
	    if(this == empty)System.out.print("[]");
	    else if(!this.iniz)System.out.print("_"+this.name);
	    else {
		System.out.print("[");
		Vector vett = this.toVector();
		for(int i = 0; i < vett.size(); i++) {
		    if(vett.get(i) instanceof Lvar)((Lvar)vett.get(i)).print();
		    else if(vett.get(i) instanceof Lst) {
			if(((Lst)vett.get(i)).iniz)((Lst)vett.get(i)).print();
			else System.out.print("_"+((Lst)vett.get(i)).name);
		    }
		    else if(vett.get(i) instanceof Set) {
			if(((Set)vett.get(i)).iniz)((Set)vett.get(i)).print();
			else System.out.print("_"+((Set)vett.get(i)).name);
		    }
		    else System.out.print(vett.get(i));
		    if(i < (vett.size() - 1))System.out.print(", ");
		}
		if(!this.ultimoResto().iniz)System.out.print("|_"+this.ultimoResto().name);
		System.out.print("]");
	    }
	}
	else this.equ.print();
	return;
    }


    public void setName(String n) {
	this.name = n;
	return;
    }


    /* Il metodo toVector() restituisce un Vector i cui elementi sono gli elementi di una Lst. 
       Esempio:
       int[] a = {1, 2, 3};
       Lvar x = new Lvar();
       Lst k = new Lst();
       Lst l = new Lst(x);
       Lst m = new Lst(l.add(4));
       Lst n = new Lst(m.append(a)); 
    
       n.toVector();   restituisce il Vector [1 2 3 4 x]
       n.ultimoResto();  restituisce la Lst k
    */
    public Vector toVector() {  
	if(this.equ == null) {
	    Vector v = new Vector();
	    if(lista != null && !lista.isEmpty()) v.addAll(lista);
	    if(resto != null && resto != empty) return resto.toVector(v);
	    else return v;
	}
	else return this.equ.toVector();
    }


    public Vector toVector(Vector vett) {  
	if(this.equ == null) {
	    if(lista != null && !lista.isEmpty()) vett.addAll(lista);
	    if(resto != null && resto != empty) return resto.toVector(vett);
	    else return vett;
	}
	else return this.equ.toVector(vett);
    }
  
 
    protected Lst ultimoResto() {  
	if(this.equ == null) {
	    if(resto == null)return this;
	    else if(resto == empty || (resto.lista == null && resto.resto == null)) return resto;
	    else return resto.ultimoResto();
	}
	else return this.equ.ultimoResto();
    } 


    /* restituisce il primo elemento della lista */
    protected Object first() { 
	if(equ == null) {
	    if(lista.size() > 0) return lista.firstElement();
	    else return resto.first();
	}
	else return this.equ.first();
    }

    /* estrae z dalla lista */
    protected Lst sub(Object z) { 
	if(equ == null) {
	    if(lista.contains(z)) {
		Vector v = new Vector();
		v.addAll(lista);
		v.remove(z);
		Lst s = new Lst(v, resto);      
		return s;
	    }
	    else {
		Lst ss = new Lst(lista, resto.sub(z));
		return ss;
	    }
	}
	else return this.equ.sub(z);
    }


    /* estrae il primo elemento */
    protected Lst subfirst() { 
	if(equ == null) {
	    Vector v = new Vector();
	    v.addAll(lista);
	    v.removeElementAt(0);
	    Lst s = new Lst(v, resto);      
	    return s;
	}
	else return this.equ.subfirst();
    }


    protected Lst sub() { 
	if(equ == null) {
	    if(lista.size() == 1) return resto;
	    else if(this.size() == 1) return this.ultimoResto();
	    else if(lista.size() == 0) return resto.sub();
	    else return this.subfirst();
	}
	else return this.equ.sub();
    }


    /*  restituisce true se z appartiene alla lista */
    public boolean occurs(Object z) {  
	if(this.equ == null) {
	    Vector v = this.toVector();
	    if(v.contains(z))return true;
	    else {
		for(int i = 0; i < v.size(); i++) {
		    if(v.get(i) instanceof Lst) {
			if(((Lst)v.get(i)).occurs(z))return true;
			else continue;
		    }
		    else if(v.get(i) instanceof Lvar && ((Lvar)v.get(i)).val instanceof Lst) {
			if(((Lst)((Lvar)v.get(i)).val).occurs(z))return true;
			else continue;
		    }
		    else continue;
		}
		return false;
	    }
	}
	else return this.equ.occurs(z);
    }

    
    /**
     * Creates and returns a copy of this object. The  method implements an
     * equivalence relation on non-null object and object references.
     * @return a clone of this instance.
     */  
   public Lst clone() {
       return this.clona();
   }
   
    
   public Lst clona() {
	if(this.equ == null) {
	    Lst clone = new Lst();
	    if (this == Lst.empty) return Lst.empty;
	    else if (!this.iniz) return clone;
	    else {
		Vector listaClone = new Vector();
		for(int i = 0; i < lista.size(); i++) {
		    if(lista.get(i) instanceof Lvar) listaClone.add(((Lvar)lista.get(i)).clona()); 
		    else if(lista.get(i) instanceof Lst) listaClone.add(((Lst)lista.get(i)).clona()); 
		    else if(lista.get(i) instanceof Set) listaClone.add(((Set)lista.get(i)).clona()); 
		    else listaClone.add(lista.get(i));
		}
		clone.lista = listaClone;
		clone.resto = this.resto.clona();
		clone.iniz  = this.iniz;
                clone.id    = this.id;
		clone.name  = this.name;
		clone.equ   = null;
		return clone;
	    }
	}
	else return this.equ.clona();
    } 

    
    /* DETERMINISTIC OPERATIONS ON Lsts */    

    
    /* I metodi inl restituiscono true se l'elemento appartiene alla lista e false altrimenti. */  
    
    /**
     * Verify if a list contains the parameter object.
     * @param i a boolean to search in the list.     
     * @return true if the parameter is on the list, false otherwise.
     */
    public boolean inl(boolean i) {
	if(this.equ == null) {
	    if(!this.iniz || this.ultimoResto() != empty) throw new NotInitVarException();
	    else {
		Boolean ii = new Boolean(i);
		return this.inl(ii);
	    }
	}
	else return this.equ.inl(i);
    }


    public boolean inl(byte i) {
	if(this.equ == null) {
	    if(!this.iniz || this.ultimoResto() != empty) throw new NotInitVarException();
	    else {
		Byte ii = new Byte(i);
		return this.inl(ii);
	    }
	}
	else return this.equ.inl(i);
    }


    public boolean inl(short i) {
	if(this.equ == null) {
	    if(!this.iniz || this.ultimoResto() != empty) throw new NotInitVarException();
	    else {
		Short ii = new Short(i);
		return this.inl(ii);
	    }
	}
	else return this.equ.inl(i);
    }


    public boolean inl(int i) {
	if(this.equ == null) {
	    if(!this.iniz || this.ultimoResto() != empty) throw new NotInitVarException();
	    else {
		Integer ii = new Integer(i);
		return this.inl(ii);
	    }
	}
	else return this.equ.inl(i);
    }


    public boolean inl(long i) {
	if(this.equ == null) { 
	    if(!this.iniz || this.ultimoResto() != empty) throw new NotInitVarException();
	    else {
		Long ii = new Long(i);
		return this.inl(ii);
	    }
	}
	else return this.equ.inl(i);
    }


    public boolean inl(float i) {
	if(this.equ == null) {
	    if(!this.iniz || this.ultimoResto() != empty) throw new NotInitVarException();
	    else {  
		Float ii = new Float(i);
		return this.inl(ii);
	    }
	}
	else return this.equ.inl(i);
    }


    public boolean inl(double i) {
	if(this.equ == null) {
	    if(!this.iniz || this.ultimoResto() != empty) throw new NotInitVarException();
	    else {
		Double ii = new Double(i);
		return this.inl(ii);
	    }
	}
	else return this.equ.inl(i);
    }


    public boolean inl(char c) {
	if(this.equ == null) {
	    if(!this.iniz || this.ultimoResto() != empty) throw new NotInitVarException();
	    else {
		Character cc = new Character(c);
		return this.inl(cc);
	    }
	}
	else return this.equ.inl(c);
    }

    
    public boolean inl(Object x) {        
        if(this.equ == null) {
             if(!this.iniz || this.ultimoResto() != empty) throw new NotInitVarException();                         
             Vector v = this.toVector();
             boolean found = false;
             int i = 0;
             while (!found && i< v.size()){                 
                 Object v_tmp = v.get(i);
                 if (v_tmp instanceof Lvar && x instanceof Lvar) found = ((Lvar)v_tmp).equals((Lvar)x);
                 else if (v_tmp instanceof Lst && x instanceof Lst) found = ((Lst)v_tmp).equals((Lst)x);
                 else if (v_tmp instanceof Set && x instanceof Set) found = ((Set)v_tmp).equals((Set)x);
                 else found = v_tmp.equals(x);
                 i++;
             }
             return found;
        }
        else return this.equ.inl(x);
    }


    // I metodi ninl restituiscono true se l'elemento non appartiene alla lista e false altrimenti.
    /**
     * Verify if a list does not contains the boolean parameter.
     * @param i a boolean to search in the list.     
     * @return true if the parameter is not on the list, false otherwise.
     */    
    public boolean ninl(boolean i) {
	Boolean ii = new Boolean(i);
	return !this.inl(ii);
    }


    public boolean ninl(byte i) {
	Byte ii = new Byte(i);
	return !this.inl(ii);
    }


    public boolean ninl(short i) {
	Short ii = new Short(i);
	return !this.inl(ii);
    }


    public boolean ninl(int i) {
	Integer ii = new Integer(i);
	return !this.inl(ii);
    }


    public boolean ninl(long i) {
	Long ii = new Long(i);
	return !this.inl(ii);
    }


    public boolean ninl(float i) {
	Float ii = new Float(i);
	return !this.inl(ii);
    }


    public boolean ninl(double i) {
	Double ii = new Double(i);
	return !this.inl(ii);
    }


    public boolean ninl(char c) {
	Character cc = new Character(c);
	return !this.inl(cc);
    }


    public boolean ninl(Object s) {
	return !this.inl(s);
    } 

 
    /**
     * Creates and returns a Set object whose elements are 
     * those of the list.
     * @return Set whose elements are the objects of the list.
     */
    public Set toSet() {
	if(this.equ == null) {
	    if(!this.iniz) {
		Set s = new Set();
		return s;
	    }
	    else if(this == Lst.empty || this.size() == 0) return Set.empty;
	    else {
		Set s = new Set();
		s.iniz  = this.iniz;
		s.lista = this.lista;
		s.resto = this.resto.toSet();
		return s;
	    }
	}
	else return this.equ.toSet();
    }

    
    /**
     * Creates and returns a Set object, with name, whose elements are 
     * those of the list.
     * @param n Desired name for the Set.
     * @return Set with the specified name, whose elements are the objects of the list.
     */
    public Set toSet(String n) {  
        Set s = toSet();
        s.setName(n);
        return s;
    }


    /**
     * Creates and return an equality constraint.
     * @param z an object that is the second equality operand.
     * @return A equality constraint between the current object and the parameter object.
     */
    public Constraint eq(Object z) { 
	Constraint s = new Constraint(this, Enviroment.eqCode, z); 
	return s;
    }


    /**
     * Creates and return an not equality constraint.
     * @param z an object that is the second not equality operand.
     * @return The not equality constraint between the current object and the parameter object.
     */
    public Constraint neq(Object z) { 
	Constraint s = new Constraint(this, Enviroment.neqCode, z); 
	return s;
    }


    /* costruisce e restituisce un vincolo di appartenenza */
    public Constraint in(Set set) { 
	Constraint s = new Constraint(this, Enviroment.inCode, set); 
	return s;
    }


    /* costruisce e restituisce un vincolo di non appartenenza */
    public Constraint nin(Set set)  { 
	Constraint s = new Constraint(this, Enviroment.ninCode, set); 
	return s;
    }


    public static Lst sost(Lvar y, Lvar x, Lst l) { 
	return l;
    }


    /**
     * Deterministic test that verify if this Lst has no components.
     * @return true if and only if this Lst has no components, that is, its size is zero; false otherwise. If the Lst is not initialized than a NotInitVarException exception is raised.
     * @throws JSetL.NotInitVarException if the list is not initialized.
     */     
    public boolean isEmpty() throws NotInitVarException {
	if(this.equ == null) {
            if (!this.iniz) throw new NotInitVarException();
            else if(this == empty) return true;
	    //else if(lista.size() == 0 && resto.isEmpty()) return true;
            else  if((lista != null && resto != null  && lista.size() == 0 && (resto.isEmpty())) ||
                     ((lista == null) && resto == null)) return true;                  
	    else return false;
	}
	else return this.equ.isEmpty();
    }             
    
    
    private boolean isEmptySL() {
        boolean res;
        try { res = this.isEmpty(); }
        /* if this set is not initialized then return false */
        catch (NotInitVarException e) { res = false; }
	return res;
    } 


    /**
     * Deterministic test that verify if this Lst is ground or not.
     *
     * @return true if this object is empty, completely specified and doesn't contains uninitialized or not assigned variables, false otherwise.
     */
    public boolean isGround() {
	if(this.equ == null) {
	    if (this == empty) return true;
	    else if (!iniz  || !ultimoResto().iniz) return false;
	    else {
		for (int i = 0; i < this.size(); i++) {
		    if (this.get(i) instanceof Lvar) { if (!((Lvar)this.get(i)).isGround()) return false; }
		    else if (this.get(i) instanceof Lst) { if (!((Lst)this.get(i)).isGround()) return false; }
                    else if (this.get(i) instanceof Set) { if (!((Set)this.get(i)).isGround()) return false; }
		}
		return true;
	    }
	}
	else return this.equ.isGround();
    }


    public boolean isAllUninitialized() {
       int res = 0; 
       for(int i = 0; i < this.size(); i++) {
           if(((Lvar)this.get(i)).isGround()) res++;
       }       
       if (res > 0) return false;
       else return true;
    }
        
    
    /**
     * Get the values of the list.
     * @return the value of the list.
     */    
    public Object getValue() { 
        return this.value();
    }    
    
    
    /**
     * Get the values of the list.
     * @return the value of the list.
     */
    protected Lst value() {
	if (this.equ == null) {
	    if (this == empty || !iniz) return this;
	    else {
		Vector newlista = new Vector();
		for(int i = 0; i < lista.size(); i++) {
		    if (lista.get(i) instanceof Lvar) {
			newlista.add(((Lvar)lista.get(i)).value());
		    }
		    else if(lista.get(i) instanceof Lst) {
			newlista.add(((Lst)lista.get(i)).value());
		    }
		    else if(this.get(i) instanceof Set) {
			newlista.add(((Set)lista.get(i)).value());
		    }
		    else newlista.add(lista.get(i));
		}
		return new Lst(newlista, resto.value());
	    }
	}
	else return this.equ.value();
    }


    /* true se la lista  inizializzata */
    /**
     * Verify is the list is initializized.
     * @return true if the list is initilalized, false otherwise.
     */
    public boolean known() {
	if(this.equ == null) return this.iniz;
	else return this.equ.known();
    }


    /* Il metodo sost restituisce una lista ottenuta da quella su cui  
       invocato il metodo sostituendo ogni occorrenza della variabile 
       y con la variabile x. */
    protected Lst sost(Lvar y, Lvar x) { 
	if(this.equ == null && y.equ == null && x.equ == null) {
	    if(!this.iniz || this.size() == 0)return this;
	    else {
		Vector v = this.toVector();
		Vector vnew = new Vector();
		for(int i = 0; i < v.size(); i++) {
		    if(v.get(i) instanceof Lvar) {
			if((Lvar)v.get(i) == y)vnew.add(x);
			else if(((Lvar)v.get(i)).val instanceof Set)vnew.add(((Set)((Lvar)v.get(i)).val).sost(y, x));
			else if(((Lvar)v.get(i)).val instanceof Lst)vnew.add(((Lst)((Lvar)v.get(i)).val).sost(y, x));
			else vnew.add((Lvar)v.get(i));
		    }
		    else if(v.get(i) instanceof Set) vnew.add(((Set)v.get(i)).sost(y, x));
		    else if(v.get(i) instanceof Lst) vnew.add(((Lst)v.get(i)).sost(y, x));
		    else vnew.add(v.get(i));
		}
		Lst snew = new Lst(vnew, this.ultimoResto().sost(y, x));
		return snew;
	    }
	}
	else if(this.equ != null)return this.equ.sost(y, x);
	else if(y.equ != null)return this.sost(y.equ, x);
	else return this.sost(y, x.equ);
    }


    public Object read() throws IOException, NotLstException, InstantiationException {
	if(iniz == false) {
	    InputStreamReader reader = new InputStreamReader(System.in);
	    BufferedReader input = new BufferedReader(reader);
	    String valore = input.readLine();
	    if(valore.charAt(0)== '[')
		return this.readLst(valore.substring(1, valore.lastIndexOf(']')));
	    else throw new NotLstException();
	}
	else throw new InstantiationException();
    }


    /* legge da tastiera il valore dell'Lvar */
    protected static Object sread(String valore) throws IOException {
	if(valore.charAt(0)== '"') {
	    if(valore.substring(1).indexOf('"') != -1) {
		return new String(valore.substring(1, valore.substring(1).indexOf('"')+1));
	    }
	    else {
		return new String(valore.substring(1));
	    } 
	}
	else if(valore.charAt(0)== '\'') {
	    return new Character(valore.charAt(1));
	}
	else if(valore.charAt(0) == '{') return Set.sreadSet(valore.substring(1, valore.lastIndexOf('}')));
	else if(valore.charAt(0)== '[') return sreadLst(valore.substring(1, valore.lastIndexOf(']')));
	else {
	    try {
		return new Integer(valore);
	    }
	    catch(NumberFormatException e) {
		try {
		    return new Float(valore);

		}
		catch(NumberFormatException f) {
		    throw new IOException();
		}     
	    }
	}
    }


    /* legge da tastiera il valore dell'Lvar */ 
    protected Object readLst(String st) throws IOException {
	Lst l = new Lst(Lst.empty);
	String s = new StringBuffer().append(", ").append(st).toString();
	int old_i = s.length()-1;
	for(int i = s.length()-1; i >= 0; i--) {
	    if(s.charAt(i) == '}') {
		int begins = Set.pareggiaSet(s, i);
		l = l.ins1(Set.sreadSet(s.substring(begins+1, i)));
		old_i = begins-2;
		i = begins-1;
	    }
	    else if(s.charAt(i) == ']') {
		int begin = Lst.pareggiaLst(s, i);
		l = l.ins1(sreadLst(s.substring(begin+1, i)));
		old_i = begin-2;
		i = begin-1;
	    }
	    else if(s.charAt(i) == ',') {
		l = l.ins1(sread(s.substring(i+1, old_i+1)));
		old_i = i-1;
	    }
	}
	iniz = true;
	lista = l.lista;
	resto = l.resto;
	return l;
    }


    /* legge da tastiera il valore dell'Lvar */
    protected static Object sreadLst(String st) throws IOException {
	Lst l = new Lst(Lst.empty);
	String s = new StringBuffer().append(", ").append(st).toString();
	int old_i = s.length()-1;
	for(int i = s.length()-1; i >= 0; i--) {
	    if(s.charAt(i) == '}') {
		int begins = Set.pareggiaSet(s, i);
		l = l.ins1(Set.sreadSet(s.substring(begins+1, i)));
		old_i = begins-2;
		i = begins-1;
	    }
	    else if(s.charAt(i) == ']') {
		int begin = Lst.pareggiaLst(s, i);
		l = l.ins1(Set.sreadSet(s.substring(begin+1, i)));
		old_i = begin-2;
		i = begin-1;
	    }
	    else if (s.charAt(i) ==',') {
		l = l.ins1(sread(s.substring(i+1, old_i+1)));
		old_i = i-1;
	    }
	}
	return l;

    }


    protected static int pareggiaLst(String s, int n) throws IOException {
	int j = -1;
	for(int i = n-1; i >= 0; i--) {
	    if(s.charAt(i) == ']') j = j-1;
	    else if(s.charAt(i) == '[') {
		if(j == -1)return i;
		else j=j+1;
	    }
	}
	throw new IOException();
    }
    
    /**
     * Deterministic test that verify if two lists are equals. The lists are equals when:
     * - they are the same object or reference to the same object, even if the object or the reference are not initialized,
     * - exactly in the same order, they contains equals objects or they contains reference to equals objects
     * @return true if the Lsts are equals, false otherwise.
     * @param s the list to be compared.
     */  
    public boolean equals (Lst l) {
        if (this.equ == null && l.equ == null) {
            if (this == l) return true;     // true, if the lists are the same reference. note that the result is true even when the lists are the same ref but they are not initialized 
            else if (!this.iniz && !l.iniz && this.id == l.id ) return true;
            else if (!this.iniz || !l.iniz) return false;  // false, if one or both are not initialized                    
            else if (this.isEmpty() && l.isEmpty()) return true; // true if the lists are both the empty lst
	    else {
                Vector t_lst = this.toVector();
                Vector s_lst = l.toVector();
                Object t_tmp = new Object();
                Object s_tmp = new Object();

                if (t_lst.size() != s_lst.size()) return false;
                boolean found = false;
   
                for (int i = 0; i < t_lst.size(); i++) {                    
                    t_tmp = t_lst.get(i);
                    s_tmp = s_lst.get(i);

                    if ((t_tmp instanceof Lvar) && (s_tmp instanceof Lvar)) found = ((Lvar)t_tmp).equals((Lvar)s_tmp);
                    else if ((t_tmp instanceof Lst) && (s_tmp instanceof Lst)) found = ((Lst)t_tmp).equals((Lst)s_tmp);
                    else if ((t_tmp instanceof Set) && (s_tmp instanceof Set)) found = ((Set)t_tmp).equals((Set)s_tmp);
                    else found = t_tmp.equals(s_tmp);
                    if (!found) return false;
                }
                // Check the List's rest                
                return ((Lst)this.ultimoResto()).equals(l.ultimoResto());
	    }            
        }
        else if (this.equ != null) return this.equ.equals(l);
        else return this.equals(l.equ);
    }


    public Lst normalizeLst() {
        if (this.equ == null) {
            if (!this.iniz) return this; // throw new NotInitVarException();
            else if (this.isEmpty()) return Lst.empty;            
            else {
                Lst nodup = new Lst(Lst.empty);
                Vector tmp_lst = this.toVector();            
                int i = 0;
                
                while (i < tmp_lst.size()){
                    Object t_tmp = tmp_lst.get(i);
                    if (t_tmp instanceof Set)  t_tmp = ((Set)t_tmp).normalizeSet();
                    if (t_tmp instanceof Lst)  t_tmp = ((Lst)t_tmp).normalizeLst();
                    if (t_tmp instanceof Lvar) t_tmp = ((Lvar)t_tmp).normalizeLvar();
                    nodup = nodup.insn(t_tmp);                   
                    i++;
                }
                return nodup;
            }            
        }
        else return this.equ.normalizeLst();
    }
    
    
    /**
     * Return a string corresponding to the object value.
     *
     * @return a string correspondig to the object value.
     */    
    public String toString() {
	if (this.equ == null) {
	    if (!this.iniz) return "List: "+this.name+", id: "+this.id+", is NOT initialized." ;
	    else {
                String outp = "List: " +this.name+ ", id: " +this.id+ ", elements: [" ;
                Vector vett = this.toVector();
		for(int i = 0; i < vett.size(); i++) {
		    if (vett.get(i) instanceof Lvar) outp += ((Lvar)vett.get(i)).toString();
		    else if (vett.get(i) instanceof Lst) outp += ((Lst)vett.get(i)).toString();
		    else if (vett.get(i) instanceof Set) outp += ((Set)vett.get(i)).toString();			
		    else outp += vett.get(i);
		    if (i < (vett.size() - 1)) outp += ", ";
		}
                outp += "], rest: [";
//		if (!this.ultimoResto().equals(Lst.empty)) outp += this.ultimoResto().toString();                
                if (this.ultimoResto() != null && !this.ultimoResto().iniz)
                    if (!this.ultimoResto().equals(Lst.empty)) outp += this.ultimoResto().toString();                
                outp += "] ";
                return outp;
            } 
	}
	else return this.equ.toString();
    }
    
}


