/*
 * Lvar.java
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;


import java.io.*;
import java.util.*;

/**
 * The <code>Lvar</code> class implements an abstraction of logical variables.
 *
 * @author  Elio Panegai
 * @author  Elisabetta Poleo
 * @version 1.3.0
 */
public class Lvar implements java.io.Serializable {    


    /**
     * Identify the logical variable as initialized (true) or not (false).
     */
    protected boolean iniz = false; 
    
    /**
     * Logical variable value, null if not initialized.
     */
    protected Object  val;      
    
    /**
     * Reference for binding equals logical variables.
     */
    protected Lvar    equ  = null;  

    /**
     * Unique identifier for the list.
     */
    protected int id;
    
    /**
     * Name of the logical variable.
     */
    protected String  name = null;  // nome della Lvar    
    
    /**
     * Vector of references to not initialized logical variables.
     */
    protected static Vector nonInizLvar; // vettore delle variabili non iniz.
    
    /**
     * Progressive logical variables count.
     */
    private   static int    counter;
    
    /**
     * Name prefix for logical variables without explicit name.
     */
    private   static String prefix;

    
    static {
        nonInizLvar = new Vector();
        counter = 0;
        prefix  = "Lvar_";
    }
        

    /**
     * Constructs a not initialized logical variable.
     */
    public Lvar() {     
	this.val = null;
	nonInizLvar.add(this);
        this.id = counter++;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    /**
     * Constructs a not initialized logical variable with a specified name.
     * @param n The name of the variable.
     */
    public Lvar(String n) {
        this();
	this.name = n;
    }


    /**
     * Constructs a logical variable with an integer value.
     * @param i The value of the variable.
     */
    public Lvar(int i) {
	this.iniz = true;
	this.val  = new Integer(i);
        this.id   = counter++;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    /**
     * Constructs a logical variable with an integer value and a name.
     * @param i The integer values of the variable.
     * @param n The name of the variable.
     */
    public Lvar(String n, int i) {
        this(i);
	this.name = n;
    }


    public Lvar(char ch) {
	this.iniz = true;
	this.val  = new Character(ch);
        this.id   = counter++;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    public Lvar(String n, char ch) {
        this(ch);
	this.name = n;
    }

        
    public Lvar(Number o) {
	this.iniz = true;
	this.val  = o;
        this.id   = counter++;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public Lvar(String n, Number o) {
        this(o);
        this.name = n;
    }

    
    public Lvar(Lvar l) {
	this.iniz = true;
	this.equ  = l;
        this.id   = l.id; // counter++;
	this.name = prefix.concat((new Integer(counter++)).toString());

    }


    public Lvar(String n, Lvar l) {
        this(l);
	this.name = n;
    }


    public Lvar(Lst l) {
	this.iniz = true;
	this.val  = l;
        this.id   = counter++;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    public Lvar(String n, Lst l) {
        this(l);
        this.name = n;
    }


    public Lvar(Set s) {
	this.iniz = true;
	this.val  = s;
        this.id   = counter++;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    public Lvar(String n, Set s) {
        this(s);
	this.name = n;
    }


    public Lvar(Object o) {
	this.iniz = true;
	this.val  = o;
        this.id   = counter++;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    public Lvar(String n, Object o) {
	this(o);
	this.name = n;
    }  


    /* Member Methods */ 
    /**
     * Gets the value of the Lvar object only if it is of type Lst, otherwise an 
     * JSetL.NotLstException is raised.
     * @return The value only if it is of type Lst.
     * @throws JSetL.NotLstException Exception raised if the Lvar's value is not of type Lst.
     */
    public Lst valLst() throws NotLstException {
	if(this.equ == null) {
	    if(this.val instanceof Lst) return ((Lst)this.val);
	    else throw new NotLstException();
	}
	else return this.equ.valLst();
    }


    /**
     * Gets the value of the Lvar object only if it is of type Set, otherwise an 
     * JSetL.NotSetException is raised.
     * @return The value only if it is of type Set.
     * @throws JSetL.NotSetException Exception raised if the Lvar's value is not of type Set.
     */
    public Set valSet() throws NotSetException {
	if(this.equ == null) {
	    if(this.val instanceof Set) return ((Set)this.val);
	    else throw new NotSetException();
	}
	else return this.equ.valSet();
    }


    
    /**
     * Note: All the not static methods of the classes Lst and Set are defined also 
     * for the class Lvar. This is due to the fact that the value of an Lvar 
     * can be a Lst or a Set, and in this case we must allows to execute the 
     * methods defined for the Lst and Set classes.
     * @param i The boolean value to inserting at the beginning of the list.
     * @return A new list with the new vector of values where the boolean passed is at the beginning.
     * @see Lst#ins1
     */          
    public Lst ins1(boolean i) {
	return ((Lst)this.valLst()).ins1(i);
    }


    public Lst ins1(byte i) {
	return ((Lst)this.valLst()).ins1(i);
    }


    public Lst ins1(short i) {
	return ((Lst)this.valLst()).ins1(i);
    }


    public Lst ins1(int i) {
	return ((Lst)this.valLst()).ins1(i);
    }


    public Lst ins1(long i) {
	return ((Lst)this.valLst()).ins1(i);
    }


    public Lst ins1(float i) {
	return ((Lst)this.valLst()).ins1(i);
    }


    public Lst ins1(double i) {
	return ((Lst)this.valLst()).ins1(i);
    }


    public Lst ins1(char c) {
	return ((Lst)this.valLst()).ins1(c);
    }

//    public Lst ins1(Number i) {
//	return ((Lst)this.valLst()).ins1(i);
//    }    
    

    public Lst ins1(Object o) {
	return ((Lst)this.valLst()).ins1(o);
    }


    public Lst ins1All(boolean[] i) {
	return ((Lst)this.valLst()).ins1All(i);
    }


    public Lst ins1All(byte[] i) {
	return ((Lst)this.valLst()).ins1All(i);
    }


    public Lst ins1All(short[] i) {
	return ((Lst)this.valLst()).ins1All(i);
    }


    public Lst ins1All(int[] i) {
	return ((Lst)this.valLst()).ins1All(i);
    }


    public Lst ins1All(long[] i) {
	return ((Lst)this.valLst()).ins1All(i);
    }


    public Lst ins1All(float[] i) {
	return ((Lst)this.valLst()).ins1All(i);
    }


    public Lst ins1All(double[] i) {
	return ((Lst)this.valLst()).ins1All(i);
    }


    public Lst ins1All(char[] c) {
	return ((Lst)this.valLst()).ins1All(c);
    }


    public Lst ins1All(Object[] o) { // pu anche essere un array di Lvar, Lst, Set
	return ((Lst)this.valLst()).ins1All(o);
    }


    public Lst insn(boolean i) {
	return ((Lst)this.valLst()).insn(i);
    }


    public Lst insn(byte i) {
	return ((Lst)this.valLst()).insn(i);
    }


    public Lst insn(short i) {
	return ((Lst)this.valLst()).insn(i);
    }


    public Lst insn(int i) {
	return ((Lst)this.valLst()).insn(i);
    }


    public Lst insn(long i) {
	return ((Lst)this.valLst()).insn(i);
    }


    public Lst insn(float i) {
	return ((Lst)this.valLst()).insn(i);
    }


    public Lst insn(double i) {
	return ((Lst)this.valLst()).insn(i);
    }


    public Lst insn(char c) {
	return ((Lst)this.valLst()).insn(c);
    }


    public Lst insn(Object o) {
	return ((Lst)this.valLst()).insn(o);
    }


    public Lst insnAll(boolean[] i) {
	return ((Lst)this.valLst()).insnAll(i);
    }


    public Lst insnAll(byte[] i) {
	return ((Lst)this.valLst()).insnAll(i);
    }


    public Lst insnAll(short[] i) {
	return ((Lst)this.valLst()).insnAll(i);
    }


    public Lst insnAll(int[] i) {
	return ((Lst)this.valLst()).insnAll(i);
    }


    public Lst insnAll(long[] i) {
	return ((Lst)this.valLst()).insnAll(i);
    }


    public Lst insnAll(float[] i) {
	return ((Lst)this.valLst()).insnAll(i);
    }


    public Lst insnAll(double[] i) {
	return ((Lst)this.valLst()).insnAll(i);
    }


    public Lst insnAll(char[] c) {
	return ((Lst)this.valLst()).insnAll(c);
    }


    public Lst insnAll(Object[] o) {
	return ((Lst)this.valLst()).insnAll(o);
    }


    public Lst concat(Lst l) { 
	return ((Lst)this.valLst()).concat(l);
    }


    public Set concat(Set s) { 
	return ((Set)this.valSet()).concat(s);
    }


    public Object concat(Lvar l) { 
	if(this.equ == null) {
	    if(this.val instanceof Lst) {
		if(l.val instanceof Lst)return ((Lst)this.val).concat((Lst)l.val);
		else throw new NotLstException();
	    }
	    else {
		if(this.val instanceof Set && l.val instanceof Set)
		    return ((Set)this.val).concat((Set)l.val);
		else throw new NotSetException();
	    }
	}
	else return this.equ.concat(l);
    }


    public Lst ext1(Lvar l) throws InitLvarException { 
	return ((Lst)this.valLst()).ext1(l);
    }


    public Lst extn(Lvar l) throws InitLvarException { 
	return ((Lst)this.valLst()).extn(l);
    }


    public Object get(int i) { 
	return ((Lst)this.valLst()).get(i);
    }


    public int size() { 
	return ((Lst)this.valLst()).size();
    }

    
    /* returns the number of objects in the list */
    protected int size_private() {
	return ((Lst)this.valLst()).size_private();
    }


    public boolean inl(boolean i) {
	return ((Lst)this.valLst()).inl(i);
    }


    public boolean inl(byte i) {
	return ((Lst)this.valLst()).inl(i);
    }


    public boolean inl(short i) {
	return ((Lst)this.valLst()).inl(i);
    }


    public boolean inl(int i) {
	return ((Lst)this.valLst()).inl(i);
    }


    public boolean inl(long i) {
	return ((Lst)this.valLst()).inl(i);
    }


    public boolean inl(float i) {
	return ((Lst)this.valLst()).inl(i);
    }


    public boolean inl(double i) {
	return ((Lst)this.valLst()).inl(i);
    }


    public boolean inl(char c) {
	return ((Lst)this.valLst()).inl(c);
    }


    public boolean inl(Object o) {
	return ((Lst)this.valLst()).inl(o);
    } 

 
    public boolean ninl(boolean i) {
	return ((Lst)this.valLst()).ninl(i);
    }


    public boolean ninl(byte i) {
	return ((Lst)this.valLst()).ninl(i);
    }


    public boolean ninl(short i) {
	return ((Lst)this.valLst()).ninl(i);
    }


    public boolean ninl(int i) {
	return ((Lst)this.valLst()).ninl(i);
    }


    public boolean ninl(long i) {
	return ((Lst)this.valLst()).ninl(i);
    }


    public boolean ninl(float i) {
	return ((Lst)this.valLst()).ninl(i);
    }


    public boolean ninl(double i) {
	return ((Lst)this.valLst()).ninl(i);
    }


    public boolean ninl(char c) {
	return ((Lst)this.valLst()).ninl(c);
    }


    public boolean ninl(Object o) {
	return ((Lst)this.valLst()).ninl(o);
    } 


    public Set toSet() {
	return ((Lst)this.valLst()).toSet();
    } 


    public Set toSet(String n) {  
	return ((Lst)this.valLst()).toSet(n);
    }


    public Set ins(boolean i) {
	return ((Set)this.valSet()).ins(i);
    }


    public Set ins(byte i) {
	return ((Set)this.valSet()).ins(i);
    }


    public Set ins(short i) {
	return ((Set)this.valSet()).ins(i);
    }


    public Set ins(int i) {
	return ((Set)this.valSet()).ins(i);
    }


    public Set ins(long i) {
	return ((Set)this.valSet()).ins(i);
    }


    public Set ins(float i) {
	return ((Set)this.valSet()).ins(i);
    }

    public Set ins(double i)
    {
	return ((Set)this.valSet()).ins(i);
    }

    public Set ins(char c)
    {
	return ((Set)this.valSet()).ins(c);
    }


    public Set ins(Object o) {
	return ((Set)this.valSet()).ins(o);
    }


    public Set insAll(boolean[] i) {
	return ((Set)this.valSet()).insAll(i);
    }

    public Set insAll(byte[] i) {
	return ((Set)this.valSet()).insAll(i);
    }

    public Set insAll(short[] i) {
	return ((Set)this.valSet()).insAll(i);
    }

    public Set insAll(int[] i) {
	return ((Set)this.valSet()).insAll(i);
    }

    public Set insAll(long[] i) {
	return ((Set)this.valSet()).insAll(i);
    }

    public Set insAll(float[] i) {
	return ((Set)this.valSet()).insAll(i);
    }

    public Set insAll(double[] i) {
	return ((Set)this.valSet()).insAll(i);
    }

    public Set insAll(char[] c) {
	return ((Set)this.valSet()).insAll(c);
    }


    public Set insAll(Object[] o) {
	return ((Set)this.valSet()).insAll(o);
    }


    /**
     * Tests if this Lvar has no components. It's a wrapper for the Lst.isEmpty() and Set.isEmpty().
     * @return true if and only if the object referenced by this Lvar has no components, that is, its size is zero; false otherwise. If this object is not initialized than a NotInitVarException exception is raised.
     * @throws JSetL.NotInitVarException the variable is not initialized.
     */     
    public boolean isEmpty() throws NotInitVarException {
	try {
	    return ((Set)this.valSet()).isEmpty();
	}  
	catch(NotSetException e) {
	    return ((Lst)this.valLst()).isEmpty();
	}
    }
	 
 
    public Vector toVector() {
	try {
	    return ((Set)this.valSet()).toVector();
	}  
	catch(NotSetException e) {
	    return ((Lst)this.valLst()).toVector();
	}
    }

	  
    public Vector toVector(Vector vett) {
	try {
	    return ((Set)this.valSet()).toVector(vett);
	}
	catch(NotSetException e) {
	    return ((Lst)this.valLst()).toVector(vett);
	}		
    }


    protected boolean occurs(Object z) {
	return ((Set)this.valSet()).occurs(z);
    }


    protected boolean contains(Object z) {
	return ((Set)this.valSet()).contains(z);
    }

    
    /* CONSTRAINTS */
    public Constraint eq(boolean i) {	
	return this.eq(new Boolean(i));
    }


    public Constraint eq(byte i) {	
	return this.eq(new Byte(i));
    }


    public Constraint eq(short i) {
	return this.eq(new Short(i));
    }


    public Constraint eq(int i) {
	return this.eq(new Integer(i));
    }


    public Constraint eq(long i) {
	return this.eq(new Long(i));
    }


    public Constraint eq(float i) {
	return this.eq(new Float(i));
    }


    public Constraint eq(double i) {
	return this.eq(new Double(i));
    }


    public Constraint eq(char i) {
	return this.eq(new Character(i));
    }

    
    public ConstraintsConjunction eq(LvarExpr a) {
	a.vet.add(this.eq(a.l));
	return a.vet;
    }
    

    public Constraint eq(Object z) {
	return new Constraint(this, Enviroment.eqCode, z);
    }


    public Constraint neq(boolean i) {
	return this.neq(new Boolean(i));
    }


    public Constraint neq(byte i) {
	return this.neq(new Byte(i));
    }


    public Constraint neq(short i) {
	return this.neq(new Short(i));
    }


    public Constraint neq(int i) {
	return this.neq(new Integer(i));
    }


    public Constraint neq(long i) {
	return this.neq(new Long(i));
    }


    public Constraint neq(float i) {
	return this.neq(new Float(i));
    }


    public Constraint neq(double i) {
	return this.neq(new Double(i));
    }


    public Constraint neq(char i) {
	return this.neq(new Character(i));
    }


    public Constraint neq(Object z) {
	return new Constraint(this, Enviroment.neqCode, z); 
    }


    public Constraint in(Set set) {
	return  new Constraint(this, Enviroment.inCode, set);
    }

     
    public Constraint in(MultiSet mset) {
	Constraint s = new Constraint(this, Enviroment.inCode, mset); 
	return s;
    }
     
     
    public Constraint in(Lvar lvar) {
	return new Constraint(this, Enviroment.inCode, lvar); 
    }


    public Constraint in(Int interv) {
	return new Constraint(this, Enviroment.inCode, interv); 
    }


    /* Select can't be utilized by the user, it is necessary for case (10) of unification */
    protected Constraint select(Set set, Set rest) {
	return new Constraint(Enviroment.selectCode, this, set, rest); 
    }


    public Constraint nin(Set set) {
	return new Constraint(this, Enviroment.ninCode, set); 
    }

    
    public Constraint nin(MultiSet mset) {
	Constraint s = new Constraint(this, Enviroment.ninCode, mset); 
	return s;
    }
    
    
    public Constraint nin(Lvar lvar) { 
	return new Constraint(this, Enviroment.ninCode, lvar); 
    }

    
    public Constraint nin(Int interv) {
	return new Constraint(this, Enviroment.ninCode, interv); 
    }

  
    public Constraint disj(Set set) {  
	return new Constraint(this, Enviroment.disjCode, set); 
    }   


    public Constraint disj(Lvar lvar) {
	return new Constraint(this, Enviroment.disjCode, lvar); 
    }

  
    public Constraint ndisj(Set set) {
	return new Constraint(this, Enviroment.ndisjCode, set);
    }    


    public Constraint ndisj(Lvar lvar) {
	return new Constraint(this, Enviroment.ndisjCode, lvar); 
    }


    public Constraint union(Set set1, Set set2) {
	return new Constraint(Enviroment.unionCode, set1, set2, this); 
    }    


    public Constraint union(Lvar l, Set set) {  
	return new Constraint(Enviroment.unionCode, l, set, this);
    }


    public Constraint union(Set set, Lvar l) {
	return new Constraint(Enviroment.unionCode, l, set, this);
    }


    public Constraint union(Lvar l1, Lvar l2) {
	return new Constraint(Enviroment.unionCode, l1, l2, this);
    }


    public Constraint nunion(Set set1, Set set2) {
	return new Constraint(Enviroment.nunionCode, set1, set2, this);
    }


    public Constraint nunion(Lvar l, Set set) {
	return new Constraint(Enviroment.nunionCode, l, set, this);
    }


    public Constraint nunion(Set set, Lvar l) {
	return new Constraint(Enviroment.nunionCode, l, set, this);
    }


    public Constraint nunion(Lvar l1, Lvar l2) {
	return new Constraint(Enviroment.nunionCode, l1, l2, this);
    }
    
    
    public ConstraintsConjunction inters(Set r, Set s) {  
        ConstraintsConjunction v = new ConstraintsConjunction();
        Set R = new Set();
	Set S = new Set();
	v.add(r.union(R, this));
	v.add(s.union(S, this));
	v.add(R.disj(S));
	return v;
    }    


    public ConstraintsConjunction inters(Lvar r, Set s) {  
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	v.add(r.union(R, this));
	v.add(s.union(S, this));
	v.add(R.disj(S));
	return v;
    }


    public ConstraintsConjunction inters(Set r, Lvar s) {
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	v.add(r.union(R, this));
	v.add(s.union(S, this));
	v.add(R.disj(S));
	return v;
    }    


    public ConstraintsConjunction inters(Lvar r, Lvar s) {
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	v.add(r.union(R, this));
	v.add(s.union(S, this));
	v.add(R.disj(S));
	return v;
    }    


    public ConstraintsConjunction ninters(Set r, Set s) {  
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	Set T = new Set();
	v.add(r.union(R, T));
	v.add(s.union(S, T));
	v.add(R.disj(S));
	v.add(this.neq(T));
	return v;
    }    


    public ConstraintsConjunction ninters(Lvar r, Set s) {  
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	Set T = new Set();
	v.add(r.union(R, T));
	v.add(s.union(S, T));
	v.add(R.disj(S));
	v.add(this.neq(T));
	return v;
    }


    public ConstraintsConjunction ninters(Set r, Lvar s) {
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	Set T = new Set();
	v.add(r.union(R, T));
	v.add(s.union(S, T));
	v.add(R.disj(S));
	v.add(this.neq(T));
	return v;
    }    


    public ConstraintsConjunction ninters(Lvar r, Lvar s) {
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	Set T = new Set();
	v.add(r.union(R, T));
	v.add(s.union(S, T));
	v.add(R.disj(S));
	v.add(this.neq(T));
	return v;
    }    


    public Constraint subset(Set s) {
	return s.union(this, s);
    }


    public Constraint subset(Lvar l) {
	return l.union(this, l);
    }


    public Constraint nsubset(Set s) {
	return s.nunion(this, s);
    }


    public Constraint nsubset(Lvar l) {
	return l.nunion(this, l);
    }

    
    public ConstraintsConjunction differ(Set r, Set s) {
	Set W = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(this, r));
	v.add(W.union(s, this));
	v.add(W.union(r, W));
	v.add(s.disj(this));
	return v;
    }    


    public ConstraintsConjunction differ(Lvar r, Set s) {
	Set W = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(this, r));
	v.add(W.union(s, this));
	v.add(W.union(r, W));
	v.add(s.disj(this));
	return v;
    }


    public ConstraintsConjunction differ(Set r, Lvar s) {
	Set W = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(this, r));
	v.add(W.union(s, this));
	v.add(W.union(r, W));
	v.add(s.disj(this));
	return v;
    }    


    public ConstraintsConjunction differ(Lvar r, Lvar s) {
	Set W = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(this, r));
	v.add(W.union(s, this));
	v.add(W.union(r, W));
	v.add(s.disj(this));
	return v;
    }    


    public ConstraintsConjunction ndiffer(Set r, Set s) {
	Set W = new Set();
	Set T = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(T, r));
	v.add(W.union(s, T));
	v.add(W.union(r, W));
	v.add(s.disj(T));
	v.add(T.neq(this));
	return v;
    }    


    public ConstraintsConjunction ndiffer(Lvar r, Set s) {  
	Set W = new Set();
	Set T = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(T, r));
	v.add(W.union(s, T));
	v.add(W.union(r, W));
	v.add(s.disj(T));
	v.add(T.neq(this));
	return v;
    }


    public ConstraintsConjunction ndiffer(Set r, Lvar s) {
	Set W = new Set();
	Set T = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(T, r));
	v.add(W.union(s, T));
	v.add(W.union(r, W));
	v.add(s.disj(T));
	v.add(T.neq(this));
	return v;
    }    


    public ConstraintsConjunction ndiffer(Lvar r, Lvar s) {
	Set W = new Set();
	Set T = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(T, r));
	v.add(W.union(s, T));
	v.add(W.union(r, W));
	v.add(s.disj(T));
	v.add(T.neq(this));
	return v;
    }    


    public Constraint ge(Lvar lvar) {
	return new Constraint(this, Enviroment.geCode, lvar);
    }


    public Constraint ge(int i) {
	return new Constraint(this, Enviroment.geCode, new Integer(i));
    }


    public Constraint gt(Lvar lvar) {
	Constraint s = new Constraint(this, Enviroment.gtCode, lvar);
	return s;
    }


    public Constraint gt(int i) {
	Integer ii = new Integer(i);
	Constraint s = new Constraint(this, Enviroment.gtCode, ii);
	return s;
    }


    public Constraint le(Lvar lvar) {
	Constraint s = new Constraint(this, Enviroment.leCode, lvar);
	return s;
    }


    public Constraint le(int i) {
	Integer ii = new Integer(i);
	Constraint s = new Constraint(this, Enviroment.leCode, ii);
	return s;
    }


    public Constraint lt(Lvar lvar) {
	Constraint s = new Constraint(this, Enviroment.ltCode, lvar);
	return s;
    }


    public Constraint lt(int i) {
	Integer ii = new Integer(i);
	Constraint s = new Constraint(this, Enviroment.ltCode, ii);
	return s;
    }


    public LvarExpr sum(LvarExpr a) {
	Lvar l = new Lvar();
	Constraint s = new Constraint(Enviroment.sumCode, l, this, a.l);
	a.vet.add(s);
	return a;
    }


    public LvarExpr sum(Object o) {
	Lvar l = new Lvar();
	Constraint s = new Constraint(Enviroment.sumCode, l, this, o);
	LvarExpr a = new LvarExpr(s, l);
	return a;
    }


    public LvarExpr sum(int o) {
	Lvar l = new Lvar();
	Integer oo = new Integer(o);
	Constraint s = new Constraint(Enviroment.sumCode, l, this, oo);
	LvarExpr a = new LvarExpr(s, l);
	return a;
    }


    public LvarExpr sub(LvarExpr a) {
	Lvar l = new Lvar();
	Constraint s = new Constraint(Enviroment.subCode, l, this, a.l);
	a.vet.add(s);
	return a;
    }


    public LvarExpr sub(Object o) {
	Lvar l = new Lvar();
	Constraint s = new Constraint(Enviroment.subCode, l, this, o);
	LvarExpr a = new LvarExpr(s, l);
	return a;
    }


    public LvarExpr sub(int o) {
	Lvar l = new Lvar();
	Integer oo = new Integer(o);
	Constraint s = new Constraint(Enviroment.subCode, l, this, oo);
	LvarExpr a = new LvarExpr(s, l);
	return a;
    }


    public LvarExpr mul(LvarExpr a) {
	Lvar l = new Lvar();
	Constraint s = new Constraint(Enviroment.mulCode, l, this, a.l);
	a.vet.add(s);
	return a;
    }


    public LvarExpr mul(Object o) {
	Lvar l = new Lvar();
	Constraint s = new Constraint(Enviroment.mulCode, l, this, o);
	LvarExpr a = new LvarExpr(s, l);
	return a;
    }


    public LvarExpr mul(int o) {
	Lvar l = new Lvar();
	Integer oo = new Integer(o);
	Constraint s = new Constraint(Enviroment.mulCode, l, this, oo);
	LvarExpr a = new LvarExpr(s, l);
	return a;
    }


    public LvarExpr div(LvarExpr a) {
	Lvar l = new Lvar();
	Constraint s = new Constraint(Enviroment.divCode, l, this, a.l);
	a.vet.add(s);
	return a;
    }


    public LvarExpr div(Object o) {
	Lvar l = new Lvar();
	Constraint s = new Constraint(Enviroment.divCode, l, this, o); 
	LvarExpr a = new LvarExpr(s, l);
	return a;
    }

    public LvarExpr div(int o) {
	Lvar l = new Lvar();
	Integer oo = new Integer(o);
	Constraint s = new Constraint(Enviroment.divCode, l, this, oo);
	LvarExpr a = new LvarExpr(s, l);
	return a;
    }


    public LvarExpr mod(LvarExpr a) {
	Lvar l = new Lvar();
	Constraint s = new Constraint(Enviroment.modCode, l, this, a.l);
	a.vet.add(s);
	return a;
    }


    public LvarExpr mod(Object o) { 
	Lvar l = new Lvar();
	Constraint s = new Constraint(Enviroment.modCode, l, this, o);
	LvarExpr a = new LvarExpr(s, l);
	return a;
    }


    public LvarExpr mod(int o) {
	Lvar l = new Lvar();
	Integer oo = new Integer(o);
	Constraint s = new Constraint(Enviroment.modCode, l, this, oo); 
	LvarExpr a = new LvarExpr(s, l);
	return a;
    }

 
    public ConstraintsConjunction less(Lvar L, Set T) throws Failure {
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(this.eq(T.ins(L))); 
	v.add(L.nin(T)); 
	return v;
    }


    public ConstraintsConjunction less(Lvar L, Lvar T) throws Failure {  
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(this.eq(T.ins(L))); 
	v.add(L.nin(T)); 
	return v;
    }    


    public static Constraint input(Lvar z) throws IOException {
	InputStreamReader input = new InputStreamReader(System.in);
	BufferedReader keyboardInput = new BufferedReader(input);   
	String r = keyboardInput.readLine();
	return z.eq(r);
    }


    public String getName() {
	return this.name;
    }


    public void setName(String n) {
	this.name = n;
	return;
    }


    /* stampa il nome e il valore della variabile */
    public void output() {
	System.out.println();
	System.out.print(this.name + " = ");
	if(!this.iniz)System.out.print("unknown");
	else this.print();
	System.out.println();
	return;
    }


    public void print() { // stampa il valore della Lvar
	if(this.equ == null) {
	    if(!this.iniz)System.out.print("_"+this.name);
	    else if(this.val instanceof Lst)((Lst)this.val).print();
	    else if(this.val instanceof Set)((Set)this.val).print();
            else if(this.val instanceof Int)((Int)this.val).output();	    
	    else System.out.print(this.val);
	}
	else this.equ.print();
	return;
    }
 
    
    /**
     * Gets the unique identifier  for the object.
     * @return an int identifier for the object.
     */
    protected int getId() {
	return this.id;
    }    
    

    /**
     * Sets the unique identifier for the object.
     */
    protected void setId() {
        this.id = counter++;
        return;
    }    
    
    
    /**
     * Get the values of the Lvar.
     * @return the value of the Lvar.
     */    
    public Object getValue() { // restituisce il valore della variabile
        if (this.equ == null) {
            if (this.val == null) return null;
            else return this.value();
        }
        else return this.equ.getValue();
    } 
    
    protected Object value() { // restituisce il valore della variabile
	if(equ == null) {
	    if(!iniz) return this;
	    else {
		if(val instanceof Set) return ((Set)val).value();     
		else if(val instanceof Lst)  return ((Lst)val).value();
                else if(val instanceof Lvar) return ((Lvar)val).value();
                else if(val instanceof MultiSet)return ((MultiSet)val).value();
		else return val;
	    }
	}
	else return equ.value();
    }


    /**
     * Gives true if the Lvar objects is a pure Lvar. Only chain references to Lvar objects are ammitted.
     *
     * @return false is there are references (from equalities) with objects of not class Lvar, true otherwise.
     */
    public boolean pureLvar() { 
	if (this.equ == null) {
	    if(!this.iniz) return true;
	    else {
		if((this.val instanceof Lst) || (this.val instanceof Set) || (this.val instanceof Int)) return false;		
		else return true;
	    }
	}
	else if (this.equ instanceof Lvar) return this.equ.pureLvar(); // Lvar chain is ok.
             else return false;
    }
    

    public String readString() throws IOException {
	if(iniz == false) {
	    InputStreamReader reader = new InputStreamReader(System.in);
	    BufferedReader input = new BufferedReader(reader);
	    String valore = input.readLine();
	    iniz = true;
	    val = valore; 
	    return valore;
	}
	else throw new InitLvarException();
    }


    public Character readChar() throws IOException {
	if(iniz == false) {
	    InputStreamReader reader = new InputStreamReader(System.in);
	    BufferedReader input = new BufferedReader(reader);
	    String valore = input.readLine();
	    Character charval = new Character(valore.charAt(0));
	    iniz = true;
	    val = charval;   
	    return charval;
	}
	else throw new InitLvarException();
    }


    public Object read() throws IOException {
	if(iniz == false) {
	    InputStreamReader reader = new InputStreamReader(System.in);
	    BufferedReader input = new BufferedReader(reader);
	    String valore = input.readLine();
	    if(valore.charAt(0)== '"') {
		if(valore.substring(1).indexOf('"') != -1) {
		    String newval2 = new String(valore.substring(1, valore.substring(1).indexOf('"')+1));
		    iniz = true;
		    val = newval2;  
		    return newval2;
		}
		else {
		    String newval3 = new String(valore.substring(1));
		    iniz = true;
		    val = newval3;  
		    return newval3;
		} 
	    }
	    else if(valore.charAt(0)== '\'') {
		Character charval1 = new Character(valore.charAt(1));
		iniz = true;
		val = charval1;
		return charval1;
	    }
	    else if(valore.charAt(0)== '{')
		return this.readSet(valore.substring(1, valore.lastIndexOf('}')));
	    else if(valore.charAt(0)== '[')
		return this.readLst(valore.substring(1, valore.lastIndexOf(']')));
	    else {
		try {
		    Integer valInt = new Integer(valore); 
		    iniz = true;
		    val = valInt;
		    return valInt;
		}
		catch(NumberFormatException e) {
		    try {
			Float valFloat = new Float(valore);
			iniz = true;
			val = valFloat; 
			return valFloat;

		    }
		    catch(NumberFormatException f) {
			throw new IOException();
		    }
		}
	    }
	}
	else throw new InitLvarException();
    }


    protected static Object sread(String valore)  throws IOException {
	if(valore.charAt(0)== '"') {
	    if(valore.substring(1).indexOf('"') != -1) {
		return new String(valore.substring(1, valore.substring(1).indexOf('"')+1));
	    }
	    else {
		return new String(valore.substring(1));
	    } 
	}
	else if(valore.charAt(0)== '\'') {
	    return new Character(valore.charAt(1));
	}
	else if(valore.charAt(0)== '{')
	    return Set.sreadSet(valore.substring(1, valore.lastIndexOf('}')));
	else if(valore.charAt(0)== '[')
	    return Lst.sreadLst(valore.substring(1, valore.lastIndexOf(']')));
	else {
	    try {
		return new Integer(valore);
	    }
	    catch(NumberFormatException e) {
		try {
		    return new Float(valore);

		}
		catch(NumberFormatException f) {
		    throw new IOException();
		}
	    }
	}
    }


    protected Object readSet(String st) throws IOException {
	Set set = new Set(Set.empty);
	String s = new StringBuffer().append(", ").append(st).toString();
	int old_i = s.length()-1;
	for(int i = s.length()-1; i >= 0; i--) {
	    if(s.charAt(i) == '}') {
		int begins = Set.pareggiaSet(s, i);
		set = set.ins(Set.sreadSet(s.substring(begins+1, i)));
		old_i = begins-2;
		i = begins-1;
	    }
	    else if(s.charAt(i) == ']') {
		int beginl = Lst.pareggiaLst(s, i);
		set = set.ins(Lst.sreadLst(s.substring(beginl+1, i)));
		old_i = beginl-2;
		i = beginl-1;
	    }
	    else if(s.charAt(i) == ',') {
		set = set.ins(sread(s.substring(i+1, old_i+1)));
		old_i = i-1;
	    }
	}
	iniz = true;
	val = set;
	return set;
    }


    protected Object readLst(String st) throws IOException {
	Lst l = new Lst(Lst.empty);
	String s = new StringBuffer().append(", ").append(st).toString();
	int old_i = s.length()-1;
	for(int i = s.length()-1; i >= 0; i--) {
	    if(s.charAt(i) == '}') {
		int begins = Set.pareggiaSet(s, i);
		l = l.ins1(Set.sreadSet(s.substring(begins+1, i)));
		old_i = begins-2;
		i = begins-1;
	    }
	    else if(s.charAt(i) == ']') {
		int begin = Lst.pareggiaLst(s, i);
		l = l.ins1(Lst.sreadLst(s.substring(begin+1, i)));
		old_i = begin-2;
		i = begin-1;
	    }
	    else if(s.charAt(i) == ',') {
		l = l.ins1(sread(s.substring(i+1, old_i+1)));
		old_i = i-1;
	    }
	}
	iniz = true;
	val = l;
	return l;
    }

    
    /**
     * Creates and returns a copy of this object. The  method implements an
     * equivalence relation on non-null object and object references.
     * @return a clone of this instance.
     */  
    public Lvar clone() {
        return this.clona();
    }


    /* restituisce un clone della variabile su cui  invocato */
    protected Lvar clona() {
	if(this.equ == null) {
	    Lvar lvarClone = new Lvar();
	    lvarClone.iniz = this.iniz;
	    lvarClone.id   = this.id;            
	    lvarClone.name = this.name;
	    lvarClone.equ  = null;
	    if (val instanceof Set) lvarClone.val = ((Set)val).clona();
	    else if (val instanceof Lst) lvarClone.val = ((Lst)val).clona();
            else if(val instanceof MultiSet)lvarClone.val = ((MultiSet)val).clona();
	    else lvarClone.val = val;
	    return lvarClone;
	}
	else return this.equ.clona();
    }
 

    /**
     * Deterministic test that verify if this object is ground or not.
     *
     * @return true if this object is empty, completely specified and doesn't contains uninitialized or not assigned variables, false otherwise.
     */  
    protected boolean isGround() {
	if (this.equ == null) {
	    if(!this.iniz) return false;
	    else {
		if (this.val instanceof Lst) return ((Lst)this.val).isGround();   
		else if (this.val instanceof Set) return ((Set)this.val).isGround();
                else if (this.val instanceof MultiSet) return ((MultiSet)this.val).isGround();
		else return true;
	    }
	}
	else return this.equ.isGround();
    }
    
    
    /**
     * Deterministic test that verify if the list is initialized.
     *
     * @return true if the list is initialized, false otherwise.
     */  
    public boolean known() {
	if(this.equ == null) {
	    if(!iniz) return false;
	    else {
		if(this.val instanceof Lst)return ((Lst)this.val).known();
		else if(this.val instanceof Set)return ((Set)this.val).known();
                else if(this.val instanceof MultiSet)return ((MultiSet)this.val).known();                
		else return true;
	    }
	}
	else return this.equ.known();
    }


    /**
     * Deterministic test that verify if two lvar are equals. The lvar are equals when:
     * - they are the same object or reference to the same object, even if the object or the reference are not initialized,
     * - they contains equals objects or they contains reference to equals object
     *
     * @return true if the Lvars are equals, false otherwise. 
     */      
    public boolean equals(Lvar L) {
	if (this.equ == null && L.equ == null) {
	    if (this == L) return true;
            else if (!this.iniz && !L.iniz && this.id == L.id ) return true;
	    else if (!this.iniz || !L.iniz) return false;
	    else {
		if (this.val instanceof Lst && L.val instanceof Lst) return ((Lst)this.val).equals((Lst)L.val);
		else if (this.val instanceof Set && L.val instanceof Set) return ((Set)this.val).equals((Set)L.val);
                else if (this.val instanceof MultiSet && L.val instanceof MultiSet) return ((MultiSet)this.val).equals((MultiSet)L.val);                
		else if (this.val.equals(L.val)) return true;
		else return false;
	    }
	}
	else if (this.equ != null) return this.equ.equals(L);
	else return this.equals(L.equ);
    }


    public boolean equals(Object o) {
        if(o instanceof Lvar) return this.equals((Lvar)o);
        else return o.equals(this);
    }
    
    
    public Lvar normalizeLvar() {
        if (this.equ == null) {
            if (!this.iniz) return this; // throw new NotInitVarException();  
            else if (this.val instanceof Set) this.val = ((Set)this.val).normalizeSet();
            else if (this.val instanceof Lst) this.val = ((Lst)this.val).normalizeLst();   
            return this;
        }
        else return this.equ.normalizeLvar();     
    }

    /**
     * Return a string corresponding to the object value.
     *
     * @return a string correspondig to the object value.
     */
    public String toString() {
	if (this.equ == null) {
	    if (!this.iniz) return "Lvar: "+this.name+", id: "+this.id+", is NOT initialized." ;
	    else {
                String outp = "Lvar: "+this.name+", id: "+this.id+", val: " ;
                if(this.val instanceof Lst) outp += ((Lst)this.val).toString();
                else if(this.val instanceof Set) outp += ((Set)this.val).toString(); 
                else if(this.val instanceof MultiSet) outp += ((MultiSet)this.val).toString();    
                else if(this.val instanceof Int) outp += ((Int)this.val).toString();              
                else outp += this.val.toString();     
                return outp; 
            } 
	}
	else return this.equ.toString();
    }
    
}

