/*
 * MultiSet.java
 *
 * Created on 23 novembre 2005, 18.00
 */

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;
import java.io.*;
import java.util.*;


/**
 * The <code>MultiSet</code> class.
 *
 * @author  Nadia Toledi
 * @author  Elio  Panegai
 * @version 1.3.0
 */
public class MultiSet implements java.io.Serializable {
    
    /**
     * Identify the set as initialized (true) or not (false).
     */
    protected boolean iniz = false;
    
    /**
     * The (specified) elements in the multiset.
     */
    protected Vector lista;
    
    /**
     * The unspecified rest of the multiset (a reference to another multiset).
     */
    protected MultiSet resto;
    
    /**
     * Reference for binding equal multisets.
     */
    protected MultiSet equ = null;
    
    
    /**
     * Unique identifier for the multiset.
     */
    protected int id;
    
    /**
     * Name of the multiset.
     */
    protected String name = null;
    
    
    /**
     * Name prefix for multisets without explicit name.
     */
    private   static String prefix;
    
    /**
     * Progressive multisets count.
     */
    private static int counter;
    
    /**
     * Vector of references to not initialized multisets.
     */
    protected static Vector nonInizMultiSet;
    
    
    /* init static vars */
    static {
        prefix     = "MultiSet_";
        counter    = 0;
        nonInizMultiSet = new Vector();
    }
    
    /**
     * The empty multiset.
     */
    public static final MultiSet  empty      = new MultiSet("emptyMultiSet", null, null);
    
    
    /** Costructors */
    public MultiSet() {
        this.lista = null;
        this.resto = null;
        this.id = counter++;
        nonInizMultiSet.add(this);
        this.name = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n) {
        this();
        this.name = n;
    }
    
    
    private Vector wrapperMultiSet() {
        Vector v   = new Vector();
        this.iniz  = true;
        this.resto = MultiSet.empty;
        this.id    = counter++;
        return v;
    }
    
    
    public MultiSet(boolean[] arr) {
        Vector v = this.wrapperMultiSet();
        for(int i = 0; i < arr.length; i++) {
            Boolean ii = new Boolean(arr[i]);
            v.add(ii);
        }
        this.lista = v;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, boolean[] arr) {
        this(arr);
        this.name = n;
    }
    
    
    public MultiSet(byte[] arr) {
        Vector v = this.wrapperMultiSet();
        for(int i = 0; i < arr.length; i++) {
            Byte ii = new Byte(arr[i]);
            v.add(ii);
        }
        this.lista = v;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, byte[] arr) {
        this(arr);
        this.name = n;
    }
    
    
    public MultiSet(short[] arr) {
        Vector v   = this.wrapperMultiSet();
        for(int i = 0; i < arr.length; i++) {
            Short ii = new Short(arr[i]);
            v.add(ii);
        }
        this.lista = v;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, short[] arr) {
        this(arr);
        this.name = n;
    }
    
    
    public MultiSet(int[] arr) {
        Vector v = this.wrapperMultiSet();
        for(int i = 0; i < arr.length; i++) {
            Integer ii = new Integer(arr[i]);
            v.add(ii);
        }
        this.lista = v;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, int[] arr) {
        this(arr);
        this.name = n;
    }
    
    
    public MultiSet(long[] arr) {
        Vector v = this.wrapperMultiSet();
        for(int i = 0; i < arr.length; i++) {
            Long ii = new Long(arr[i]);
            v.add(ii);
        }
        this.lista = v;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, long[] arr) {
        this(arr);
        this.name = n;
    }
    
    
    public MultiSet(float[] arr) {
        Vector v = this.wrapperMultiSet();
        for(int i = 0; i < arr.length; i++) {
            Float ii = new Float(arr[i]);
            v.add(ii);
        }
        this.lista = v;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, float[] arr) {
        this(arr);
        this.name = n;
    }
    
    
    public MultiSet(double[] arr) {
        Vector v = this.wrapperMultiSet();
        for(int i = 0; i < arr.length; i++) {
            Double ii = new Double(arr[i]);
            v.add(ii);
        }
        this.lista = v;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, double[] arr) {
        this(arr);
        this.name = n;
    }
    
    
    public MultiSet(char[] arr) {
        Vector v = this.wrapperMultiSet();
        for(int i = 0; i < arr.length; i++) {
            Character c = new Character(arr[i]);
            v.add(c);
        }
        this.lista = v;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, char[] arr) {
        this(arr);
        this.name = n;
    }
    
    
    public MultiSet(Object[] arr) {
        Vector v = this.wrapperMultiSet();
        for(int i = 0; i < arr.length; i++) {
            v.add(arr[i]);
        }
        this.lista = v;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, Object[] arr) {
        this(arr);
        this.name = n;
    }
    
    
    protected MultiSet(Vector v, MultiSet mset) {
        if(v == null && mset == null) {
            this.iniz  = true;
            this.lista = null;
            this.resto = null;
        }
        else {
            this.lista = v;
            this.resto = mset;
            if(this.lista.size() > 0) this.iniz = true;
            else this.iniz = mset.iniz;
        }
        if(!iniz) nonInizMultiSet.add(this);
        this.id = counter++;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    protected MultiSet(String n, Vector v, MultiSet mset) {
        this((Vector) v, (MultiSet) mset);
        this.name = n;
    }
    
    
    public MultiSet(MultiSet mset) {
        this.iniz  = true;
        this.lista = mset.lista;
        this.resto = mset.resto;
        this.equ   = mset;
        this.id    = mset.id; // counter++;
        this.name  = prefix.concat((new Integer(counter++)).toString());
    }
    
    
    public MultiSet(String n, MultiSet mset) {
        this(mset);
        this.name = n;
    }
    
    
    public MultiSet(Lvar l) {
        Vector v = this.wrapperMultiSet();
        v.add(l);
        this.lista = v;
        this.name = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, Lvar l)  {
        this(l);
        this.name = n;
    }
    
    
    public MultiSet(Lst l) {
        Vector v = this.wrapperMultiSet();
        v.add(l);
        this.lista = v;
        this.name = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, Lst l) {
        this(l);
        this.name = n;
    }
    
    
    public MultiSet(int p, int q)  {
        Vector v = this.wrapperMultiSet();
        for(int i = p; i <= q; i++) {
            Integer ii = new Integer(i);
            v.add(ii);
        }
        this.lista = v;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, int p, int q) {
        this(p, q);
        this.name = n;
    }
    
    
    public MultiSet(int p, Lvar q) {
        Vector v = this.wrapperMultiSet();
        
        int suplimit;
        if (q.getValue() == null) suplimit = p;
        else suplimit = ((Integer)q.getValue()).intValue();
        
        for(int i = p; i <= suplimit; i++) {
            Integer ii = new Integer(i);
            v.add(ii);
        }
        this.lista = v;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }
    
    
    public MultiSet(String n, int p, Lvar q) {
        this(p, q);
        this.name = n;
    }
    
    
    
    /* Member Methods */
    
    /**
     * Tests if this Multiset has no components.
     *
     * @return true if and only if this Multiset has no components, that is, its size is zero; false otherwise.
     * If the Multiset is not initialized than a NotInitVarException exception is raised.
     */
    public boolean isEmpty() {
        if(equ == null) {
            if(!this.iniz) throw new NotInitVarException();
            else if(this == MultiSet.empty) return true;
            else  if((this.lista != null && this.resto != null  && this.lista.size() == 0 && (this.resto.isEmpty())) ||
            ((this.lista == null) && this.resto == null)) return true;
            else return false;
        }
        else return this.equ.isEmpty();
    }
    
    
    /**
     * The same as isEmpty(), this method is used for
     *   a correct constraint resolution engine
     *
     * @return false if this object is not initialized.
     */
    protected boolean isEmptySL() {
        boolean res;
        try { res = this.isEmpty(); }
        /* if this set is not initialized then return false */
        catch (NotInitVarException e) { res = false; }
        return res;
    }
    
    /**
     * Function to insert a single element into a MultiSet. This method is a wrapper for the primitives types.
     *
     * @param o Object element
     * @return a Multiset obtained from the Muliset on which
     * this method has been invoked, adding the input data
     *
     */       
    public MultiSet ins(Number i) {
        if(this.equ == null) {
            Vector v = new Vector();
            v.add(i);
            MultiSet m = new MultiSet(v, this);
            return m;
        }
        else return this.equ.ins(i);
    }
    
    
//    public MultiSet ins(int i) {
//        if(this.equ == null) {
//            Vector v = new Vector();
//            Integer o = new Integer(i);
//            v.add(o);
//            MultiSet m = new MultiSet(v, this);
//            return m;
//        }
//        else return this.equ.ins(i);
//    }
//    
//    
//    public MultiSet ins(char i) {
//        if(this.equ == null) {
//            Vector v = new Vector();
//            Character o = new Character(i);
//            v.add(o);
//            MultiSet m = new MultiSet(v, this);
//            return m;
//        }
//        else return this.equ.ins(i);
//    }
    
    
    /**
     * Function to insert a single element into a MultiSet. This method is a wrapper for the general types:
     * for example to insert a MultiSet, Set, Lst or Lvar.
     *
     * @param o Object element
     * @return a Multiset obtained from the Muliset on which
     * this method has been invoked, adding the input data
     *
     */
    public MultiSet ins(Object o) {
        if(this.equ == null) {
            Vector v = new Vector();
            v.add(o);
            MultiSet m = new MultiSet(v, this);
            return m;
        }
        else return this.equ.ins(o);
    }
    
    
    /**
     * The insAll methods
     *
     * @param arr of boolean elements
     * @return a Multiset obtained from the Muliset on which
     * this method has been invoked, adding all the elements of the array as input
     */
    public MultiSet insAll(boolean[] arr) {
        if(this.equ == null) {
            Vector v = new Vector();
            for(int i = 0; i < arr.length; i++){
                Boolean o = new Boolean(arr[i]);
                v.add(o);
            }
            MultiSet l = new MultiSet(v, this);
            return l;
        }
        else return this.equ.insAll(arr);
    }
    
    
    public MultiSet insAll(byte[] arr) {
        if(this.equ == null){
            Vector v = new Vector();
            for(int i = 0; i < arr.length; i++){
                Byte o = new Byte(arr[i]);
                v.add(o);
            }
            MultiSet l = new MultiSet(v, this);
            return l;
        }
        else return this.equ.insAll(arr);
    }
    
    
    public MultiSet insAll(short[] arr) {
        if(this.equ == null) {
            Vector v = new Vector();
            for(int i = 0; i < arr.length; i++){
                Short o = new Short(arr[i]);
                v.add(o);
            }
            MultiSet l = new MultiSet(v, this);
            return l;
        }
        else return this.equ.insAll(arr);
    }
    
    
    public MultiSet insAll(int[] arr) {
        if(this.equ == null) {
            Vector v = new Vector();
            for(int i = 0; i < arr.length; i++){
                Integer o = new Integer(arr[i]);
                v.add(o);
            }
            MultiSet l = new MultiSet(v, this);
            return l;
        }
        else return this.equ.insAll(arr);
    }
    
    
    public MultiSet insAll(long[] arr) {
        if(this.equ == null) {
            Vector v = new Vector();
            for(int i = 0; i < arr.length; i++){
                Long o = new Long(arr[i]);
                v.add(o);
            }
            MultiSet l = new MultiSet(v, this);
            return l;
        }
        else return this.equ.insAll(arr);
    }
    
    
    public MultiSet insAll(float[] arr) {
        if(this.equ == null) {
            Vector v = new Vector();
            for(int i = 0; i < arr.length; i++){
                Float o = new Float(arr[i]);
                v.add(o);
            }
            MultiSet l = new MultiSet(v, this);
            return l;
        }
        else return this.equ.insAll(arr);
    }
    
    
    public MultiSet insAll(double[] arr) {
        if(this.equ == null) {
            Vector v = new Vector();
            for(int i = 0; i < arr.length; i++){
                Double o = new Double(arr[i]);
                v.add(o);
            }
            MultiSet l = new MultiSet(v, this);
            return l;
        }
        else return this.equ.insAll(arr);
    }
    
    
    public MultiSet insAll(char[] arr) {
        if(this.equ == null) {
            Vector v = new Vector();
            for(int i = 0; i < arr.length; i++){
                Character o = new Character(arr[i]);
                v.add(o);
            }
            MultiSet l = new MultiSet(v, this);
            return l;
        }
        else return this.equ.insAll(arr);
    }
    
    
    /**
     * The insAll method to insert an array of Objects: that can be MultiSets, Sets, Lsts and Lvars
     *
     * @param arr of Object
     * @return a Multiset obtained from the Muliset on which
     * this method has been invoked, adding all the elements of the array as input
     */
    public MultiSet insAll(Object[] arr) {
        if(this.equ == null) {
            Vector v = new Vector();
            for(int i = 0; i < arr.length; i++){
                v.add(arr[i]);
            }
            MultiSet l = new MultiSet(v, this);
            return l;
        }
        else return this.equ.insAll(arr);
    }
    
    /**
     * The insAll method to insert an array of Number to wrapper the primitives types.
     *
     * @param arr of Number
     * @return a Multiset obtained from the Muliset on which
     * this method has been invoked, adding all the elements of the array as input
     */
    
    public MultiSet insAll(Number[] arr) {
        if(this.equ == null) {
            Vector v = new Vector();
            for(int i = 0; i < arr.length; i++){
                v.add(arr[i]);
            }
            MultiSet l = new MultiSet(v, this);
            return l;
        }
        else return this.equ.insAll(arr);
    }
    
    
    /**
     * Function to obtain all the elements of a MultiSet
     *
     * @return a vector of MultiSet's elements
     */
    public Vector toVector() {
        if(this.equ == null) {
            Vector v = new Vector();
            if (this.iniz) {
                if(lista != null && !lista.isEmpty()) {
                    v.addAll(lista);
                    if(resto != null)return resto.toVector(v);
                    else return v;
                }
                else if(resto != null && !resto.isEmptySL()) return resto.toVector(v);
                else return v;
            }
            else return v;
        }
        else return this.equ.toVector();
    }
    
    
    /**
     * The same of toVector() but here we add the elements of the MultiSet to a Vector in input.
     *
     * @param vett vector where the elements are save
     * @return input vett added with the elements of the MultiSet
     */
    public Vector toVector(Vector vett) {
        if (this.equ == null) {
            if (this.iniz) {
                if (lista != null && !lista.isEmpty()) {
                    vett.addAll(lista);
                    if (resto != null) return resto.toVector(vett);
                    else return vett;
                }
                else if (resto != null && !resto.isEmptySL()) return resto.toVector(vett);
                else return vett;
            }
            else return vett;
        }
        else return this.equ.toVector(vett);
    }
    
    
    /**
     *
     *
    /**
     * Returns the size of the multiset, that is the determinable number of elements of the multiset.
     * If exist a rest, the count take care of specified elements in the rest of the multiset.     
     * The size() method is deterministic, but in the NewConstraint class there is an other definition
     * non-deterministic of this function.
     *    
     * @return The integer number of object in the set, 0 for the empty multiset.
     */    
    public int size() {
        if(equ == null) {
            if (!this.iniz || this.isEmptySL()) return 0;
            else if (this.lista.isEmpty()) return this.resto.size();
            else return this.lista.size() + this.resto.size();
        }
        else return this.equ.size();
    }
    
    
    /**
     * Get the tail of this MultiSet
     *
     * @return a MultiSet containing the tail of the initial MultiSet
     */    
    protected MultiSet ultimoResto() {
        if (this.equ == null) {
            if (!this.iniz) return this;
            else if (this.resto == MultiSet.empty || this.resto == null || !this.resto.iniz || this.resto.isEmptySL()) return this.resto;
            else return this.resto.ultimoResto();
        }
        else return this.equ.ultimoResto();
    }
    
    
    /**
     * Get the untail of the Multiset
     *
     * @return a MultiSet containig the untail of the initial MultiSet
     */    
    protected MultiSet notultimoResto(){
        if(this.equ == null){
            if(!this.iniz) return null;
            else if (this.resto == empty || this.resto == null || !this.resto.iniz || this.resto.isEmptySL()){
                return new MultiSet((Vector)this.lista, MultiSet.empty);
            }
            else {
                return new MultiSet((Vector)this.toVector(), MultiSet.empty);
            }
        }
        else return this.equ.notultimoResto();
    }
    
    
    /**
     * Get the first (sintactic) element of this Multiset
     *
     * @return an Object that is a Lvar, Set, Lst or MultiSet
     */    
    protected Object first() {
        if(this.equ == null) {
            if(!this.lista.isEmpty()) return this.lista.firstElement();
            else return this.resto.first();
        }
        else return this.equ.first();
    }
    
    
    /**
     * This method returns a Multiset obtained from the Multiset on which this function has been invoked,
     * removing the element z
     *
     * @param z Object to remove from the initial MultiSet
     * @return a new MultiSet removing param
     */    
    protected MultiSet sub(Object z) {
        if(this.equ == null) {
            if(this.lista.contains(z)) {
                Vector v = new Vector();
                v.addAll(this.lista);
                v.remove(z);
                if(v.size() == 0) return this.resto;
                else {
                    MultiSet s = new MultiSet(v, this.resto);
                    return s;
                }
            }
            else{
                MultiSet ss = new MultiSet(this.lista, this.resto.sub(z));
                return ss;
            }
        }
        else return this.equ.sub(z);
    }
    
    
    /**
     * This method returns a Multiset obtained from the Multiset on which this function has been invoked,
     * removing all the occurrences of the element z
     *
     * @param z the Object to remove from the initial MultiSet
     * @return a new MultiSet removing all istances of param
     */    
    protected MultiSet subAll(Object z) {
        if(this.equ == null) {
            if(this.lista.contains(z)) {
                Vector v = new Vector();
                v.addAll(this.lista);
                v.remove(z);
                if(v.size() == 0) return this.resto;
                else {
                    MultiSet s = new MultiSet(v, this.resto);
                    return s;
                }
            }
            else{
                MultiSet ss = new MultiSet(this.lista, this.resto.subAll(z));
                return ss;
            }
        }
        else return this.equ.subAll(z);
    }
    
    
    /**
     * This method returns a Multiset obtained from the Multiset on which this function has been invoked,
     * removing the first element. This method works only on the specified part of the MultiSet.
     *
     * @return a new MultiSet removing param
     */    
    protected MultiSet subfirst() {
        if(this.equ == null) {
            Vector v = new Vector();
            v.addAll(this.lista);
            v.removeElementAt(0);
            MultiSet s = new MultiSet(v, this.resto);
            return s;
        }
        else return this.equ.subfirst();
    }
    
    
    /**
     * This method returns a Multiset obtained from the Multiset on which this function has been invoked,
     * removing the first element, but consider the MultiSet entirely, the list of elements and the tail.
     *
     * @return a new MultiSet removing the first element
     */    
    protected MultiSet sub() {
        if(this.equ == null) {
            if (this.lista.size() == 1) return this.resto;
            else if (this.size() == 1) return this.ultimoResto();
            else if (this.lista.size() == 0) return this.resto.sub();
            else return this.subfirst();
        }
        else return this.equ.sub();
    }
    
    
    /**
     * This methods returns a Multiset that is equal to the connection of two Multisets
     *
     * @param m one of the initial MultiSets
     * @return a new MultiSet that is obtained from the concatenation of the two initial ones
     */
    public MultiSet concat(MultiSet m) {
        if(this.equ == null && m.equ == null) {
            Vector v = new Vector();
            Vector thisToVector = this.toVector();
            Vector MultisetToVector = m.toVector();
            v.addAll(thisToVector);
            v.addAll(MultisetToVector);
            MultiSet l = new MultiSet(v, empty);
            return l;
        }
        else if (this.equ == null) return this.concat(m.equ);
        else if (m.equ == null) return this.equ.concat(m);
        else return this.equ.concat(m.equ);
    }
    
    
    /**
     * This methods returns a Multiset that is equal to the connection of two Multisets
     *
     */
    protected MultiSet append(MultiSet m) {
        if (this.equ == null && m.equ == null) {
            if(this.isEmptySL() || !this.iniz) return m;
            else{
                MultiSet s = new MultiSet(this.lista, this.resto.append(m));
                return s;
            }
        }
        else if (this.equ == null) return this.append(m.equ);
        else if (m.equ == null) return this.equ.append(m);
        else return this.equ.append(m.equ);
    }
    
    
    /**
     * This methods returns a Multiset that is equal to the connection of two Multisets
     */    
    protected MultiSet appendGround(MultiSet m) {
        if (this.equ == null && m.equ == null) {
            if (this.isEmptySL()) return m;
            else{
                MultiSet s = new MultiSet(this.toVector(), m);
                return s;
            }
        }
        else if (this.equ == null) return this.append(m.equ);
        else return this.equ.append(m);
    }
    
    
    /**
     * This method verify if an element is in the MultiSet
     *
     * @return true if there is at least one occurrence of the object x to any level of this MultiSet, false otherwise.
     * @param x element to search in the initial MultiSet
     */
    protected boolean occurs(Object x) {
        if (this.equ == null) {
            if (!this.iniz || this == empty) return false;
            else {
                Vector v = this.toVector();
                if (v.contains(x)) return true;
                else {
                    for (int i = 0; i < v.size(); i++) {
                        if(v.get(i) instanceof MultiSet) {
                            if (((MultiSet)v.get(i)).occurs(x)) return true;
                        }
                        else if (v.get(i) instanceof Lvar && ((Lvar)v.get(i)).val instanceof MultiSet) {
                            if (((MultiSet)((Lvar)v.get(i)).val).occurs(x)) return true;
                        }
                    }
                    return false;
                }
            }
        }
        else return this.equ.occurs(x);
    }
    
    
    /* constraints */
    
    /**
     * Constraint of equality from a MultiSet and an Object
     *
     * @param z Object to unify with the invocation MultiSet
     * @return the constraint Multiset.eq(z)
     */
    public Constraint eq(Object z) {
        return new Constraint(this, Enviroment.eqCode, z);
    }
    
    
    /**
     * Constraint of disequality from a MultiSet and an Object
     *
     * @param z to compare with the invocation MultiSet
     * @return the constraint MultiSet.neq(z)
     */
    public Constraint neq(Object z) {
        return new Constraint(this, Enviroment.neqCode, z);
    }
    
    
    /**
     * Constraint of membership to the MultiSet
     *
     * @param multiset to search in the invocation element
     * @return the constraint MultiSet.in(MultiSet)
     */
    public Constraint in(MultiSet multiset) {
        return new Constraint(this, Enviroment.inCode, multiset);
    }
    
    /**
     * Constraint of membership
     *
     * @param lvar to search in the initial MultiSet
     * @return the constraint Lvar.in(MultiSet)
     */
    public Constraint in(Lvar lvar) {
        return new Constraint(this, Enviroment.inCode, lvar);
    }
    
    
    /**
     * Constraint of notmembership
     *
     * @param multiset to search in the invocation MultiSet
     * @return the constraint MultiSet.nin(MultiSet)
     */
    public Constraint nin(MultiSet multiset) {
        return new Constraint(this, Enviroment.ninCode, multiset);
    }
    
    
    /**
     * Constraint of notmembership
     *
     * @param lvar to search in the initial MultiSet
     * @return the constraint Lvar.nin(MultiSet)
     */
    public Constraint nin(Lvar lvar)  {
        return new Constraint(this, Enviroment.ninCode, lvar);
    }
    
    /**
     * Gets the name of the MultiSet
     *
     *@return the name of the Multiset
     */
    public String getName() {
        return this.name;
    }
    
    
    /**
     * Deterministic test that verify if this Set has no variables.
     *
     * @return true if this object is empty, completely specified and doesn't contains uninitialized or not assigned variables, false otherwise.
     */
    public boolean isGround() { // restituisce true se il Multiset  completamente noto
        if (this.equ == null) {
            if (this == empty) return true;
            else if(!iniz) return false;
            else if(!ultimoResto().iniz) return false;
            else {
                for(int i = 0; i < this.size(); i++){
                    if (this.get(i) instanceof Lvar) { if(!((Lvar)this.get(i)).isGround()) return false; }
                    else if (this.get(i) instanceof Lst) { if(!((Lst)this.get(i)).isGround()) return false; }
                    else if (this.get(i) instanceof Set) { if(!((Set)this.get(i)).isGround()) return false; }
                    else if (this.get(i) instanceof MultiSet) { if(!((MultiSet)this.get(i)).isGround()) return false;}
                }
                return true;
            }
        }
        else return this.equ.isGround();
    }
    
    
    
    /**
     * Deterministic test that verify if this Set has no variables.
     *
     * @return true if this object is empty, completely specified and doesn't contains uninitialized or not assigned variables, false otherwise.
     */
    public boolean isGrounded() { // restituisce true se il Multiset  completamente noto
        if (this.equ == null) {
            if (this == empty) return true;
            else if(!iniz) return false;
            //            else if(!ultimoResto().iniz) {
            //                this.output();
            //                ((MultiSet)ultimoResto()).output();
            //                System.out.println("XELIO caso 1");
            //                return false;
            //            }
            else if(ultimoResto().iniz && !ultimoResto().isGrounded()) {
                System.out.println("XELIO caso 2");
                return false;
            }
            else {
                for(int i = 0; i < this.size(); i++){
                    if (this.get(i) instanceof Lvar) {
                        if(!((Lvar)this.get(i)).isGround()) return false;
                    }
                    //                    else if (this.get(i) instanceof Lst) { if(!((Lst)this.get(i)).isGround()) return false; }
                    //                    else if (this.get(i) instanceof Set) { if(!((Set)this.get(i)).isGround()) return false; }
                    //                    else if (this.get(i) instanceof MultiSet) { if(!((MultiSet)this.get(i)).isGrounded()) return false;}
                }
                return true;
            }
        }
        else return this.equ.isGrounded();
    }
    
    
    /**
     * This method return a particular element in the initial MultiSet
     *
     * @param i position of the element to search
     * @return the element in the position i
     */
    public Object get(int i) {
        Vector v = this.toVector();
        return v.get(i);
    }
    
    
    /**
     * Tests if this Set is initialized.
     *
     * @return true if and only if the Set is initialized, false otherwise.
     */
    public boolean known() {
        if (this.equ == null) return this.iniz;
        else return this.equ.known();
    }
    
    
    /**
     * Prints the name of the Multiset on which this function has been invoked
     * and its value.
     */    
    public void output() {
        System.out.println();
        System.out.print(this.name + " = ");
        if (this.equ == null) {
            if (!this.iniz) System.out.print("unknown");
            else this.print();
        }
        else this.equ.print();
        System.out.println();
        return;
    }
    
    
    /**
     * The print() method prints the Multiset on which this function has been invoked, if this Multiset
     * is initialized, otherwise it prints its name preceded by "_", to distinguish name from string or character
     * that are used as value for the variables.
     * For Sets and Lists the output has the sintax of Prolog.
     */
    public void print() {
        if (this.equ == null) {
            if (this == MultiSet.empty) System.out.print("+{MultiSet.empty}+"); //print("+{}+");
            else if (!this.iniz) {
                System.out.print("_");
                printName(this);
            }
            else {
                System.out.print("+{");
                Vector vett = this.toVector();
                for(int i = 0; i < vett.size(); i++) {
                    if (vett.get(i) instanceof Lvar) ((Lvar)vett.get(i)).print();
                    else if (vett.get(i) instanceof Lst) {
                        if (((Lst)vett.get(i)).iniz) ((Lst)vett.get(i)).print();
                        else System.out.print("_"+((Lst)vett.get(i)).name);
                    }
                    else if (vett.get(i) instanceof MultiSet) {
                        if (((MultiSet )vett.get(i)).iniz) ((MultiSet )vett.get(i)).print();
                        else System.out.print("_"+((MultiSet )vett.get(i)).name);
                    }
                    else System.out.print(vett.get(i));
                    if (i < (vett.size() - 1)) System.out.print(",");
                }
                if (!this.ultimoResto().equals(MultiSet .empty)) {
                    System.out.print("|");
                    printName(this.ultimoResto());
                }
                System.out.print("}+");
            }
        }
        else this.equ.print();
        return;
    }
    
    
    public static void printName(MultiSet m) {
        if (m.equ == null) {
            System.out.print(m.name);
            return;
        }
        else printName(m.equ);
        return;
    }
    
    
    /**
     * Creates and returns a copy of this object. The  method implements an
     * equivalence relation on non-null object and object references.
     * @return a clone of this instance.
     */      
   public MultiSet clone() {
       return this.clona();
   }
   
   
    /**
     * This method create a perfect copy of a MultiSet
     *
     * @return a new MultiSet equal to the initial MultiSet
     */
    protected MultiSet clona() {
        if (this.equ == null) {
            MultiSet clone = new MultiSet();
            if(this == empty)return empty;
            else if(!this.iniz) return clone;
            else {
                Vector listaClone = new Vector();
                for (int i = 0; i < this.lista.size(); i++) {
                    if (lista.get(i) instanceof Lvar) listaClone.add(((Lvar)lista.get(i)).clona());
                    else if (lista.get(i) instanceof Lst) listaClone.add(((Lst)lista.get(i)).clona());
                    else if (lista.get(i) instanceof Set) listaClone.add(((Set)lista.get(i)).clona());
                    else if (lista.get(i) instanceof MultiSet) {
                        if (((MultiSet)lista.get(i)) == empty) listaClone.add(empty);
                        else listaClone.add(((MultiSet)lista.get(i)).clona());
                    }
                    else listaClone.add(lista.get(i));
                }
                clone.lista = listaClone;
                clone.resto = resto.clona();
                clone.iniz = iniz;
                clone.name = name;
                clone.equ = null;
                return clone;
            }
        }
        else return this.equ.clona();
    }
    
    /**
     * This method build a Multiset of n logical variables not initialized.
     *
     * @param n number of Lvar that the new MultiSet must contain
     * @return a MultiSet of n Lvar not initialized
     */
    
    public static MultiSet mkMultiset(int n) {
        Vector lista = new Vector();
        for (int i = 0; i < n; i++) {
            Lvar lvar = new Lvar();
            lista.add(lvar);
        }
        MultiSet m = new MultiSet(lista, empty);
        return m;
    }
    
    
    /**
     * The same of mkMultiSet(n) but we specify also a name for the new MultiSet
     *
     * @param name of the new MultiSet
     * @param n number of logical variables member of the new MultiSet
     * @return a new MultiSet
     */
    public static MultiSet mkMultiset(String name, int n) {
        Vector lista = new Vector();
        for (int i = 0; i < n; i++) {
            Lvar lvar = new Lvar();
            lista.add(lvar);
        }
        MultiSet m = new MultiSet(name, lista, empty);
        return m;
    }
    
    
    public int ariety(Object x) {
        int count = 0;
        if (this.equ == null) {
            Vector v = this.toVector();
            for (int i = 0; i < v.size(); i++) {
                if(v.get(i) instanceof MultiSet) {
                    ((MultiSet)v.get(i)).ariety(x);
                }
                else if (v.get(i) instanceof Lvar && ((Lvar)v.get(i)).val instanceof MultiSet) {
                    ((MultiSet)((Lvar)v.get(i)).val).ariety(x);
                }
                //                        else if (v.get(i) instanceof Set && ((Lvar)v.get(i)).val instanceof MultiSet){
                //                            ((MultiSet)((Lvar)v.get(i)).val).ariety(x);
                //                        }
                if(v.get(i).equals(x)){
                    count++;
                }
            }
            return count;
        }
        else return this.equ.ariety(x);
    }
    
    
    public int ariety(Number z) {        
        int count = 0;
        if (this.equ == null) {
            Vector v = this.toVector();
            for (int i = 0; i < v.size(); i++) {
                if(v.get(i) instanceof MultiSet) {
                    ((MultiSet)v.get(i)).ariety(z);
                }
                else if (v.get(i) instanceof Lvar && ((Lvar)v.get(i)).val instanceof MultiSet) {
                    ((MultiSet)((Lvar)v.get(i)).val).ariety(z);
                }
                //                        else if (v.get(i) instanceof Set && ((Lvar)v.get(i)).val instanceof MultiSet){
                //                            ((MultiSet)((Lvar)v.get(i)).val).ariety(x);
                //                        }
                if(v.get(i).equals(z)){
                    count++;
                }
            }
            return count += count;
        }
        else return this.equ.ariety(z);
    }
    
    
    /*
     * Questo metodo va corretto! Essendo direttamente derivato dai Set
     * non tiene conto del numero di elementi presenti nel multiset!
     * Per ora {a, b, b}.equals{a, b} = true
     *
     */
    /**
     * Deterministic test that verify if two MultiSet are equals. The MultiSet are equals when:
     * - they are the same object or reference to the same object, even if the object or the reference are not initialized,
     * - they contains equals objects or they contains reference to equals object
     *
     * @return true if the MultiSet are equals, false otherwise. 
     */          
    public boolean equals(MultiSet s) {
        if (this.equ == null && s.equ == null) {
            if (this == s) return true;    // true, if the sets are the same reference. note that the result is true even when the sets are the same ref but they are not initialized
            //            else if(!this.iniz && !s.iniz && this.id == s.id) return true;
            else if(!this.iniz || !s.iniz) return false;  // false, if one or both are not initialized
            else if (this.isEmpty() && s.isEmpty()) return true; // true if the sets are both the empty set
            else {
                Vector t_set = this.toVector();
                Vector s_set = s.toVector();
                Object t_tmp = new Object();
                Object s_tmp = new Object();
                boolean found;
                
                for (int i = 0; i < t_set.size(); i++) {
                    found = false;
                    int arieta_t_tmp = this.ariety(this.get(i));
                    //t_tmp = t_set.get(i);
                    for (int j = 0; j < s_set.size(); j++) {
                        int arieta_s_tmp = s.ariety(s.get(j));
                        //s_tmp = s_set.get(j);
                        if ((t_tmp instanceof Lvar) && (s_tmp instanceof Lvar) && (((Lvar)t_tmp).equals((Lvar)s_tmp))&& arieta_t_tmp == arieta_s_tmp)  found = true;
                        else if ((t_tmp instanceof Lst) && (s_tmp instanceof Lst) && (((Lst)t_tmp).equals((Lst)s_tmp))&& arieta_t_tmp == arieta_s_tmp) found = true;
                        else if ((t_tmp instanceof Set) && (s_tmp instanceof Set) && (((Set)t_tmp).equals((Set)s_tmp))&& arieta_t_tmp == arieta_s_tmp) found = true;
                        else if ((s_tmp instanceof MultiSet) && (t_tmp instanceof MultiSet) && (((MultiSet)s_tmp).equals((MultiSet)t_tmp))&& arieta_t_tmp == arieta_s_tmp) found = true;
                        else if (t_tmp.equals(s_tmp)) found = true;
                    }
                    if (!found) return false;
                }
                for (int i = 0; i < s_set.size(); i++) {
                    found = false;
                    s_tmp = s_set.get(i);
                    for (int j = 0; j < t_set.size(); j++) {
                        t_tmp = t_set.get(j);
                        if ((s_tmp instanceof Lvar) && (t_tmp instanceof Lvar) && (((Lvar)s_tmp).equals((Lvar)t_tmp))) found = true;
                        else if ((s_tmp instanceof Lst) && (t_tmp instanceof Lst) && (((Lst)s_tmp).equals((Lst)t_tmp))) found = true;
                        else if ((s_tmp instanceof Set) && (t_tmp instanceof Set) && (((Set)s_tmp).equals((Set)t_tmp))) found = true;
                        else if ((s_tmp instanceof MultiSet) && (t_tmp instanceof MultiSet) && (((MultiSet)s_tmp).equals((MultiSet)t_tmp))) found = true;
                        else if (s_tmp.equals(t_tmp)) found = true;
                    }
                    if (!found) return false;
                }
                
                return this.ultimoResto().equals(s.ultimoResto());
            }
        }
        else if (this.equ != null) return this.equ.equals(s);
        else return this.equals(s.equ);
    }
    
    
    /**
     * Checks appartenence to a Multiset.
     *
     *@return true if and only if z is an element of throws Multiset
     */
    protected boolean contains(Object z) {
        if(this.equ == null){
            if(this.iniz){
                if(this == MultiSet.empty) return false;
                else{
                    Vector v = this.toVector();
                    if(v.contains(z)) return true;
                    else return false;
                }
            }
            else return false;
        }
        else return this.equ.contains(z);
    }
    
    
    /**
     * This method is used to obtain the Id of the initial MultiSet
     *
     * @return the Id
     */
    public int getId() {
        return this.id;
    }
    
    
    public void setId() {
        this.id = counter++;
        return;
    }
    
    
    /**
     * Get the values of the multiset.
     * @return the value of the multiset.
     */    
    public Object getValue() { 
        return this.value();
    }    
    
    protected MultiSet value() {
        if(this.equ == null) {
            if (this == empty || !iniz) return this;
            else {
                Vector newlista = new Vector();
                for (int i = 0; i < lista.size(); i++) {
                    if (lista.get(i) instanceof Lvar) {
                        newlista.add(((Lvar)lista.get(i)).value());
                    }
                    else if (lista.get(i) instanceof Lst) {
                        newlista.add(((Lst)lista.get(i)).value());
                    }
                    else if (this.get(i) instanceof Set) {
                        newlista.add(((Set)lista.get(i)).value());
                    }
                    else if (this.get(i) instanceof MultiSet) {
                        newlista.add(((MultiSet)lista.get(i)).value());
                    }
                    else newlista.add(lista.get(i));
                }
                return new MultiSet(newlista,resto.value());
            }
        }
        else return this.equ.value();
    }
    
    
    /**
     * This method is used to write correctly a MultiSet, also if there is nested MultiSets
     *
     * @return correct MultiSet
     */
    public MultiSet normalizeMultiSet() {
        if (this.equ == null) {
            if (!this.iniz) throw new NotInitVarException();
            else if (this.isEmpty()) return MultiSet.empty;
            else {
                MultiSet nodup = new MultiSet(MultiSet.empty);
                Vector t_set = this.toVector();
                Vector indexofduplicated = new Vector();
                int i = 0;
                
                while (i < t_set.size()){
                    Integer Iint = new Integer(i);
                    Object t_tmp = t_set.get(i);    // t_tmp = t_set.get(i);
                    if (t_tmp instanceof MultiSet) t_tmp = ((MultiSet)t_tmp).normalizeMultiSet();
                    if (t_tmp instanceof Set)  t_tmp = ((Set)t_tmp).normalizeSet();
                    if (t_tmp instanceof Lst)  t_tmp = ((Lst)t_tmp).normalizeLst();
                    if (t_tmp instanceof Lvar) t_tmp = ((Lvar)t_tmp).normalizeLvar();
                    
                    if ((indexofduplicated == null) ||
                    (indexofduplicated != null && !indexofduplicated.contains(Iint))) {
                        nodup = nodup.ins(t_tmp);
                        
                        if (i+1 <= t_set.size()){
                            boolean found;
                            for (int j=i+1; j< t_set.size(); j++) {
                                Integer Jint = new Integer(j);
                                found = false;
                                Object t_tmp2 = t_set.get(j);
                                
                                if ((t_tmp instanceof Lvar) && (t_tmp2 instanceof Lvar) && (((Lvar)t_tmp).equals((Lvar)t_tmp2)))  found = true;
                                else if ((t_tmp instanceof Lst) && (t_tmp2 instanceof Lst) && (((Lst)t_tmp).equals((Lst)t_tmp2))) found = true;
                                else if ((t_tmp instanceof Set) && (t_tmp2 instanceof Set) && (((Set)t_tmp).equals((Set)t_tmp2))) found = true;
                                else if ((t_tmp instanceof MultiSet) && (t_tmp2 instanceof MultiSet) && (((MultiSet)t_tmp).equals((MultiSet)t_tmp2))) found = true;
                                else if (t_tmp.equals(t_tmp2)) found = true;
                                
                                if (found) indexofduplicated.add(Jint);
                            }
                        }
                    }
                    i++;
                }
                return nodup;
            }
        }
        else return this.equ.normalizeMultiSet();
    }
    
}

