/*
 * Set.java
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;

import java.io.*;
import java.util.*;


/**
 * The <code>Set</code> class.
 *
 * @author  Elio Panegai
 * @author  Elisabetta Poleo
 * @version 1.3.0
 */
public class Set implements java.io.Serializable {

    /**
     * Identify the set as initialized (true) or not (false).
     */
    protected boolean iniz = false;  
    
    /**
     * The (specified) elements in the set.
     */    
    protected Vector lista;       
    
    /**
     * The unspecified rest of the set (a reference to another set).
     */    
    protected Set resto;        
    
    /**
     * Reference for binding equal sets.
     */    
    protected Set equ = null;   
    
    protected boolean isInt = false; 
    protected Int inter = null;
    
    /**
     * Unique identifier for the set.
     */
    protected int id; 
    
    /**
     * Name of the set.
     */    
    protected String name = null; 
    
    /**
     * Name prefix for sets without explicit name.
     */    
    private   static String prefix;   
    
    /**
     * Progressive sets count.
     */    
    private static int counter;   
    
    /**
     * Vector of references to not initialized sets.
     */    
    protected static Vector nonInizSet;      

    /* init static vars */
    static { 
        prefix     = "Set_";
        counter    = 0;   
        nonInizSet = new Vector(); 
    }
    
    /**
     * The empty set.
     */      
    public static final Set empty = new Set("emptySet", null, null); // insieme vuoto    

    
    /* Costructors */
    public Set() {
	this.lista = null;
	this.resto = null;
        this.id = counter++;
	nonInizSet.add(this);
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n) {
        this();
	this.name = n;
    }

    
    private Vector wrapperSet() {
        Vector v   = new Vector();
	this.iniz  = true;
	this.resto = Set.empty;  
        this.id    = counter++;
        return v;
    }


    public Set(boolean[] arr) {
	Vector v = this.wrapperSet();
	for(int i = 0; i < arr.length; i++) {
	    Boolean ii = new Boolean(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, boolean[] arr) {
        this(arr);
	this.name = n;
    }


    public Set(byte[] arr) {
	Vector v = this.wrapperSet();
	for(int i = 0; i < arr.length; i++) {
	    Byte ii = new Byte(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, byte[] arr) {
        this(arr);
	this.name = n;
    }


    public Set(short[] arr) {
	Vector v   = this.wrapperSet();
	for(int i = 0; i < arr.length; i++) {
	    Short ii = new Short(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, short[] arr) {       
        this(arr);
	this.name = n;
    }


    public Set(int[] arr) {                  
	Vector v = this.wrapperSet();
	for(int i = 0; i < arr.length; i++) {
	    Integer ii = new Integer(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, int[] arr) {
        this(arr);
	this.name = n;
    }
        

    public Set(long[] arr) {
	Vector v = this.wrapperSet();
	for(int i = 0; i < arr.length; i++) {
	    Long ii = new Long(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, long[] arr) { 
        this(arr);
	this.name = n;
    }


    public Set(float[] arr) {
	Vector v = this.wrapperSet();
	for(int i = 0; i < arr.length; i++) {
	    Float ii = new Float(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, float[] arr) {
        this(arr);
	this.name = n;
    }


    public Set(double[] arr) {
	Vector v = this.wrapperSet();
	for(int i = 0; i < arr.length; i++) {
	    Double ii = new Double(arr[i]);
	    v.add(ii);
	}
	this.lista = v;  
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, double[] arr) {
        this(arr);
	this.name = n;
    }


    public Set(char[] arr) {
	Vector v = this.wrapperSet();
	for(int i = 0; i < arr.length; i++) {
	    Character c = new Character(arr[i]);
	    v.add(c);
	}
	this.lista = v;  
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, char[] arr) {       
        this(arr);
	this.name = n;
    }


    public Set(Object[] arr) {
	Vector v = this.wrapperSet();
	for(int i = 0; i < arr.length; i++) {
	    v.add(arr[i]);  
	}
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, Object[] arr) {
        this(arr);
	this.name = n;
    }


    protected Set(Vector v, Set s) {
	if(v == null && s == null) {
	    this.iniz  = true;
	    this.lista = null;
	    this.resto = null;
	}
	else {
	    this.lista = v;
	    this.resto = s;
	    if(this.lista.size() > 0) this.iniz = true;
	    else this.iniz = s.iniz;
	}
	if(!iniz) nonInizSet.add(this);
	this.id = counter++;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    protected Set(String n, Vector v, Set s) {
        this((Vector) v, (Set) s);
	this.name = n;
    }


    public Set(Set s) {
	this.iniz  = true;         
	this.lista = s.lista;
	this.resto = s.resto;
	this.equ   = s;
	this.id    = s.id; // counter++;
	this.name  = prefix.concat((new Integer(counter++)).toString());
    }


    public Set(String n, Set s) {
        this(s);
	this.name = n;
    }


    public Set(Lvar l) {
	Vector v = this.wrapperSet();
	v.add(l);
	this.lista = v;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, Lvar l)  {
        this(l);
	this.name = n;
    }
    

    public Set(Lst l) {
	Vector v = this.wrapperSet();
	v.add(l);
	this.lista = v;
	this.name = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, Lst l) {
        this(l);
	this.name = n;
    }
    

    public Set(Int I) {
	this.iniz  = true;
	this.isInt = true;
	this.inter = I;
	this.lista = null;
	this.resto = null;
        this.id    = counter++;
        this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, Int I) {
        this(I);
	this.name = n;
    }


    public Set(int p, int q)  {
	Vector v = this.wrapperSet();
	for(int i = p; i <= q; i++) {
	    Integer ii = new Integer(i);
	    v.add(ii);
	}	
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, int p, int q) {
        this(p, q);
	this.name = n;
    }


    public Set(int p, Lvar q) {
	Vector v = this.wrapperSet();    
        
        int suplimit;
        if (q.getValue() == null) suplimit = p;
        else suplimit = ((Integer)q.getValue()).intValue();
            
	for(int i = p; i <= suplimit; i++) {
	    Integer ii = new Integer(i);
	    v.add(ii);
	}
	this.lista = v;
	this.name  = prefix.concat((new Integer(this.id)).toString());
    }


    public Set(String n, int p, Lvar q) {
        this(p, q);
	this.name = n;
    }
     


    /* Member Methods */


    /* I metodi ins restituiscono un Set ottenuto da quello su cui  stato invocato 
       il metodo aggiungendo il parametro passato. */      
    /**
     * Insertion of a single element in a set.
     * @param n the element to insert.
     * @return returns a set obtained by adding n to the set
     */
    public Set ins(Number n) { 
	if(this.equ == null) {
	    Vector v = new Vector(); 
	    v.add(n);
	    Set s = new Set(v,this);
	    return s;
	}
	else return this.equ.ins(n);
    }

    
    /**
     * Insertion of a single object in a set.
     * @param o the object to insert.
     * @return returns a set obtained by adding o to the set
     */
    public Set ins(Object o) {
	if(this.equ == null) {
	    Vector v = new Vector(); 
	    v.add(o);
	    Set s = new Set(v,this);
	    return s;
	}
	else return this.equ.ins(o);
    }



    /* I metodi insAll restituiscono un Set ottenuto da quello su cui  stato invocato 
       il metodo aggiungendo gli elementi dell'array passato come parametro. */
    public Set insAll(boolean[] arr) {
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++){
		Boolean o = new Boolean(arr[i]);
		v.add(o);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }


    public Set insAll(byte[] arr) {
	if(this.equ == null){
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++){
		Byte o = new Byte(arr[i]);
		v.add(o);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }


    public Set insAll(short[] arr) {
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++){
		Short o = new Short(arr[i]);
		v.add(o);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }


    public Set insAll(int[] arr) {
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++){
		Integer o = new Integer(arr[i]);
		v.add(o);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }


    public Set insAll(long[] arr) {
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++){
		Long o = new Long(arr[i]);
		v.add(o);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }


    public Set insAll(float[] arr) {
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++){
		Float o = new Float(arr[i]);
		v.add(o);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }


    public Set insAll(double[] arr) {
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++){
		Double o = new Double(arr[i]);
		v.add(o);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }


    public Set insAll(char[] arr) {
	if(this.equ == null) {
	    Vector v = new Vector();  
	    for(int i = 0; i < arr.length; i++){
		Character o = new Character(arr[i]);
		v.add(o);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }


    /**
     * with arr array of objects of type t, returns a set obtained by adding all elements of arr to the set.
     * @param arr the array of objects to insert. 
     * @return set obtained by adding all elements of arr to the set.
     */
    public Set insAll(Object[] arr) {
	if(this.equ == null) {
	    Vector v = new Vector();
	    for(int i = 0; i < arr.length; i++){
		v.add(arr[i]);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }

    
    /* Encapsulated by public Set insAll(Object[] arr)
    public Set insAll(Lvar[] arr) {
	if(equ == null) {
	    Vector v = new Vector();
	    for(int i = 0; i < arr.length; i++){
		v.add(arr[i]);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }


    public Set insAll(Lst[] arr) {
	if(equ == null) {
	    Vector v = new Vector();
	    for(int i = 0; i < arr.length; i++){
		v.add(arr[i]);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }


    public Set insAll(Set[] arr) {
	if(equ == null) {
	    Vector v = new Vector();
	    for(int i = 0; i < arr.length; i++){
		v.add(arr[i]);
	    }
	    Set l = new Set(v,this);
	    return l;
	}
	else return this.equ.insAll(arr);
    }
    */

        
    /**
     * Tests if this Set has no components.
     *
     * @return true if and only if this Set has no components, that is, its size is zero; false otherwise. If the Set is not initialized than a NotInitVarException exception is raised.
     */         
    public boolean isEmpty() {
	if(this.equ == null) {
	    if(this.isInt) return ((Int)this.inter).isVoid();            
	    else if(!this.iniz) throw new NotInitVarException();
	    else if(this == Set.empty) return true;
	    //else if(lista.size() == 0 && resto.isEmpty()) return true;
            else  if((this.lista != null && this.resto != null  && this.lista.size() == 0 && (this.resto.isEmpty())) ||
                     ((this.lista == null) && this.resto == null)) return true;                
	    else return false;
	}
	else return this.equ.isEmpty();
    }
    
    
    /* The same as isEmpty(), but return false if this object is not initialized. This method is used for
       a correct constraint resolution engine */
    protected boolean isEmptySL() {
	boolean res;
        try { res = this.isEmpty(); }
        /* if this set is not initialized then return false */
        catch (NotInitVarException e) { res = false; }
	return res;
    }

    
    /**
     * Deterministic test that verify if two Sets are equals. The Sets are equals when:
     * - they are the same object or reference to the same object, even if the object or the reference are not initialized,
     * - they contains equals objects or they contains reference to equals objects (order does not matter!)
     *
     * @return true if the Sets are equals, false otherwise. 
     */  
    public boolean equals (Set s) {
        if (this.equ == null && s.equ == null) {
            if (this == s) return true;    // true, if the sets are the same reference. note that the result is true even when the sets are the same ref but they are not initialized 
            else if (!this.iniz && !s.iniz && this.id == s.id ) return true;
            else if(!this.iniz || !s.iniz) return false;  // false, if one or both are not initialized                    
            else if (this.isEmpty() && s.isEmpty()) return true; // true if the sets are both the empty set
	    else {
                Vector t_set = this.toVector();
                Vector s_set = s.toVector();
                Object t_tmp = new Object();
                Object s_tmp = new Object();
                boolean found;
                
                for (int i = 0; i < t_set.size(); i++) {
                    found = false;
                    t_tmp = t_set.get(i);
                    for (int j = 0; j < s_set.size(); j++) {                        
                        s_tmp = s_set.get(j);
                        if ((t_tmp instanceof Lvar) && (s_tmp instanceof Lvar) && (((Lvar)t_tmp).equals((Lvar)s_tmp)))  found = true;
                        else if ((t_tmp instanceof Lst) && (s_tmp instanceof Lst) && (((Lst)t_tmp).equals((Lst)s_tmp))) found = true;
                        else if ((t_tmp instanceof Set) && (s_tmp instanceof Set) && (((Set)t_tmp).equals((Set)s_tmp))) found = true;
                        else if (t_tmp.equals(s_tmp)) found = true;
                    }
                    if (!found) return false;
                }
                for (int i = 0; i < s_set.size(); i++) {
                    found = false;
                    s_tmp = s_set.get(i);
                    for (int j = 0; j < t_set.size(); j++) {                        
                        t_tmp = t_set.get(j);
                        if ((s_tmp instanceof Lvar) && (t_tmp instanceof Lvar) && (((Lvar)s_tmp).equals((Lvar)t_tmp))) found = true;
                        else if ((s_tmp instanceof Lst) && (t_tmp instanceof Lst) && (((Lst)s_tmp).equals((Lst)t_tmp))) found = true;
                        else if ((s_tmp instanceof Set) && (t_tmp instanceof Set) && (((Set)s_tmp).equals((Set)t_tmp))) found = true;
                        else if (s_tmp.equals(t_tmp)) found = true;
                    }
                    if (!found) return false;
                }                
                // Check the Set's rest
                return this.ultimoResto().equals(s.ultimoResto());
	    }            
        }
        else if (this.equ != null) return this.equ.equals(s);
        else return this.equals(s.equ);
    }
      

    /* Il metodo toVector() restituisce un Vector i cui elementi sono gli 
       elementi di un Set. Esempio:
       int[] a = {1, 2, 3};
       Lvar x = new Lvar();
       Set k = new Set();
       Set l = new Set(x);
       Set m = new Set(l.add(4));
       Set n = new Set(m.append(a)); 
    
       n.toVector();   restituisce il Vector [1 2 3 4 x]
       n.ultimoResto();  restituisce la Lst k
    */        
    public Vector toVector() {
	if(this.equ == null) {
	    Vector v = new Vector();
	    if (this.iniz) {
		if(this.lista != null && !this.lista.isEmpty()) {
		    v.addAll(this.lista);
		    if(this.resto != null)return this.resto.toVector(v);
		    else return v;
		}
		else if(this.resto != null && !this.resto.isEmptySL()) return this.resto.toVector(v);
		else return v;
	    }
	    else return v;
	}
	else return this.equ.toVector();
    }


    public Vector toVector(Vector vett) {
	if (this.equ == null) {
	    if (this.iniz) {
		if (this.lista != null && !this.lista.isEmpty()) {
		    vett.addAll(this.lista);
		    if (this.resto != null) return this.resto.toVector(vett);
		    else return vett;
		}
		else if (this.resto != null && !this.resto.isEmptySL()) return this.resto.toVector(vett);
		else return vett;
	    }
	    else return vett;
	}
	else return this.equ.toVector(vett);
    }
  

    /* Il metodo size() restituisce un intero uguale al numero di elementi del set.
       Questa size  deterministica, ma nella classe NewConstraint si pu trovare
       una definizione non deterministica della size(). */   
    /**
     * Returns the size of the set, that is the determinable number of elements of the set.
     * If exist a rest, the count take care of specified elements in the rest of the set.     
     * @return The integer number of object in the set, 0 for the empty set.          
     */    
    public int size() {
	if(this.equ == null) {
            if (!this.iniz) return 0;
            Set tmp = this.normalizeSet();
	    if (!tmp.iniz || tmp.isEmptySL()) return 0;
	    else if (tmp.lista.isEmpty()) return tmp.resto.size();
	    else return tmp.lista.size() + tmp.resto.size();
	}
	else return this.equ.size();
    }

    
    /* Get the tail of this Set */
    protected Set ultimoResto() {
	if (this.equ == null) {
	    if (!this.iniz) return this;
	    else if (this.isInt) return this;
	    else if (this.resto == empty || this.resto == null || !this.resto.iniz || this.resto.isEmptySL()) return this.resto;
	    else return this.resto.ultimoResto();    
	}
	else return this.equ.ultimoResto();
    } 

     
    /* Get the first (sintactic) element of this Set */
    public Object first() { 
	if(this.equ == null) {
	    if(!this.lista.isEmpty()) return this.lista.firstElement();
	    else return this.resto.first();
	}
	else return this.equ.first();
    }


    /* estrae z dall'insieme */
    protected Set sub(Object z) { 
	if(this.equ == null) {
	    if(this.lista.contains(z)) {
		Vector v = new Vector();
		v.addAll(this.lista);
		v.remove(z);
		if(v.size() == 0) return this.resto;
		else {
		    Set s = new Set(v, this.resto);      
		    return s;
		}
	    }
	    else{
		Set ss = new Set(this.lista, this.resto.sub(z));
		return ss;
	    }
	}
	else return this.equ.sub(z);
    }


    protected Set subfirst() { 
	if(this.equ == null) {
	    Vector v = new Vector();
	    v.addAll(this.lista);
	    v.removeElementAt(0);
	    Set s = new Set(v, this.resto);      
	    return s;
	}
	else return this.equ.subfirst();
    }


    /* estrae il primo elemento sell'insieme */
    public Set sub() { 
	if(this.equ == null) {
	    if (this.lista.size() == 1) return this.resto;
	    else if (this.size() == 1)  return this.ultimoResto();
	    else if (this.lista.size() == 0) return this.resto.sub();
	    else return this.subfirst();
	}
	else return this.equ.sub();
    }


    /* restituisce un set uguale alla concatenazione dei due set */
    public Set concat(Set set) {
	if(this.equ == null && set.equ == null) {
	    Vector v = new Vector();
	    Vector thisToVector = this.toVector();
	    Vector setToVector = set.toVector();
	    v.addAll(thisToVector);
	    v.addAll(setToVector);
	    Set l = new Set(v, empty);
	    return l;
	}
	else if (this.equ == null) return this.concat(set.equ);
	else if (set.equ  == null) return this.equ.concat(set);
	else return this.equ.concat(set.equ);
    }


    /* Restituisce un set uguale alla concatenazione dei due set */
    protected Set append(Set set) {
	if (this.equ == null && set.equ == null) {
	    if(this.isEmptySL() || !this.iniz) return set;
	    else{
		Set s = new Set(this.lista, this.resto.append(set));
		return s;
	    }
	}
	else if (this.equ == null) return this.append(set.equ);
	else if (set.equ  == null) return this.equ.append(set);
	else return this.equ.append(set.equ);
    }


    /* Restituisce un set uguale alla concatenazione dei due set */
    protected Set appendGround(Set set) {
	if (this.equ == null && set.equ == null) {
	    if (this.isEmptySL()) return set;
	    else{
		Set s = new Set(this.toVector(), set);
		return s;
	    }
	}
	else if (this.equ == null) return this.append(set.equ);
	else return this.equ.append(set);
    }

       
    /**
     *
     * @return true if there is at least one occurrence of the object x to any level of this Set, false otherwise.
     */
    protected boolean occurs (Object x) {
        if (this.equ == null) {
            if (!this.iniz || this == empty) return false;
            else {
                Vector v = this.toVector();
		if (v.contains(x)) return true;
		else {
                    for (int i = 0; i < v.size(); i++) {
                        if(v.get(i) instanceof Set) {
                            if (((Set)v.get(i)).occurs(x)) return true;
			}
                        else if (v.get(i) instanceof Lvar && ((Lvar)v.get(i)).val instanceof Set) {
                            if (((Set)((Lvar)v.get(i)).val).occurs(x)) return true;
			}
                    }
                    return false;
                }
            }
	}
	else return this.equ.occurs(x);
    }
  

    /* true se z  un elemento dell'insieme */
    protected boolean contains (Object z) {
	if(this.equ == null){
	    if(this.iniz){
		if(this == empty) return false;
		else{
		    Vector v = this.toVector();
		    if(v.contains(z)) return true;
		    else return false;
		}
	    }
	    else return false;
	}
	else return this.equ.contains(z);
    }

    

    /* Constraints */

    public Constraint eq (Object z) { 
	return new Constraint(this, Enviroment.eqCode, z); 
    }


    public Constraint neq (Object z) {
	return new Constraint(this, Enviroment.neqCode, z);
    }   


    public Constraint in (Set set) {
	return new Constraint(this, Enviroment.inCode, set);
    }

    public Constraint in (Lvar lvar) {
	return new Constraint(this, Enviroment.inCode, lvar);
    }


    public Constraint nin (Set set) { 
	return new Constraint(this, Enviroment.ninCode, set);
    }


    public Constraint nin (Lvar lvar) {
	return new Constraint(this, Enviroment.ninCode, lvar);
    }
     
  
    public Constraint disj (Set set) { 
	return new Constraint(this, Enviroment.disjCode, set);
    }    


    public Constraint disj (Lvar lvar) {
	return new Constraint(this, Enviroment.disjCode, lvar);
    }


    public Constraint ndisj (Set set) {  
	return new Constraint(this, Enviroment.ndisjCode, set); 
    } 


    public Constraint ndisj (Lvar lvar) {
	return new Constraint(this, Enviroment.ndisjCode, lvar); 
    }


    public Constraint union(Set set1, Set set2)  { 
	return new Constraint(Enviroment.unionCode, set1, set2, this); 
    }    


    public Constraint union(Lvar l, Set set) { 
	return new Constraint(Enviroment.unionCode, l, set, this); 
    }

    public Constraint union(Set set, Lvar l) {
	return new Constraint(Enviroment.unionCode, l, set, this); 
    }


    public Constraint union(Lvar l1, Lvar l2) {
	return new Constraint(Enviroment.unionCode, l1, l2, this); 
    }


    public Constraint nunion(Set set1, Set set2) { 
	return new Constraint(Enviroment.nunionCode, set1, set2, this); 
    }    


    public Constraint nunion(Lvar l, Set set) {  
	return new Constraint(Enviroment.nunionCode, l, set, this); 
    }

    public Constraint nunion(Set set, Lvar l) {  
	return new Constraint(Enviroment.nunionCode, l, set, this); 
    }


    public Constraint nunion(Lvar l1, Lvar l2) { 
	return new Constraint(Enviroment.nunionCode, l1, l2, this); 
    }

 
    public ConstraintsConjunction inters(Set r, Set s) {
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	v.add(r.union(R,this));
	v.add(s.union(S,this));
	v.add(R.disj(S));
	return v;
    }


    public ConstraintsConjunction inters(Lvar r, Set s) {  
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	v.add(r.union(R,this));
	v.add(s.union(S,this));
	v.add(R.disj(S));
	return v;
    }

    public ConstraintsConjunction inters(Set r, Lvar s) {  
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	v.add(r.union(R,this));
	v.add(s.union(S,this));
	v.add(R.disj(S));
	return v;
    }


    public ConstraintsConjunction inters(Lvar r,Lvar s) {  
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	v.add(r.union(R,this));
	v.add(s.union(S,this));
	v.add(R.disj(S));
	return v;
    }


    public ConstraintsConjunction ninters(Set r, Set s) {
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	Set T = new Set();
	v.add(r.union(R,T));
	v.add(s.union(S,T));
	v.add(R.disj(S));
	v.add(this.neq(T));
	return v;
    }

    public ConstraintsConjunction ninters(Lvar r, Set s) {  
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	Set T = new Set();
	v.add(r.union(R,T));
	v.add(s.union(S,T));
	v.add(R.disj(S));
	v.add(this.neq(T));
	return v;
    }

    public ConstraintsConjunction ninters(Set r, Lvar s) {  
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	Set T = new Set();
	v.add(r.union(R,T));
	v.add(s.union(S,T));
	v.add(R.disj(S));
	v.add(this.neq(T));
	return v;
    }


    public ConstraintsConjunction ninters(Lvar r, Lvar s) {  
	ConstraintsConjunction v = new ConstraintsConjunction();
	Set R = new Set();
	Set S = new Set();
	Set T = new Set();
	v.add(r.union(R,T));
	v.add(s.union(S,T));
	v.add(R.disj(S));
	v.add(this.neq(T));
	return v;
    }


    public Constraint subset (Set s) {
	return s.union(this,s);
    }

    public Constraint subset (Lvar l) {
	return l.union(this,l);
    }


    public Constraint nsubset(Set s) {
	return s.nunion(this,s);
    }

    
    public Constraint nsubset(Lvar l) {
	return l.nunion(this,l);
    }
    
    
    /* Realizza un intervallo con estremi variabili e sensibile al backtracking.
       Supponiamo ad esempio di voler creare l insieme degli interi compresi
       tra X e Y con X e Y Lvar. Si tratta cio di un intervallo i cui estremi possono
       essere modificati in caso di backtracking, quindi anche l'intervallo deve
       essere sensibile al backtracking. */
    public Constraint interv(int n, Lvar X) throws Failure {
	Integer nn = new Integer(n);
	Constraint s = new Constraint(Enviroment.intervCode, this, nn, X); 
	return s;
    }
 

    public Constraint interv(Lvar X,int n) {
	Integer nn = new Integer(n);
	Constraint s = new Constraint(Enviroment.intervCode, this, X, nn); 
	return s;
    }


    public Constraint interv(Lvar X, Lvar Y) {	 
	Constraint s = new Constraint(Enviroment.intervCode, this, X, Y); 
	return s;
    }    

 
    public ConstraintsConjunction differ(Set r, Set s) { 
	Set W = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(this,r));
	v.add(W.union(s,this));
	v.add(W.union(r,W));
	v.add(s.disj(this));
	return v;
    }


    public ConstraintsConjunction differ(Lvar r, Set s) {  
	Set W = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(this,r));
	v.add(W.union(s,this));
	v.add(W.union(r,W));
	v.add(s.disj(this));
	return v;
    }

    
    public ConstraintsConjunction differ(Set r,Lvar s) {  
	Set W = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(this,r));
	v.add(W.union(s,this));
	v.add(W.union(r,W));
	v.add(s.disj(this));
	return v;
    }


    public ConstraintsConjunction differ(Lvar r, Lvar s) {  
	Set W = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(this,r));
	v.add(W.union(s,this));
	v.add(W.union(r,W));
	v.add(s.disj(this));
	return v;
    }


    public ConstraintsConjunction ndiffer(Set r, Set s) {
	Set W = new Set();
	Set T = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(T,r));
	v.add(W.union(s,T));
	v.add(W.union(r,W));
	v.add(s.disj(T));
	v.add(T.neq(this));
	return v;
    }


    public ConstraintsConjunction ndiffer(Lvar r, Set s) {  
	Set W = new Set();
	Set T = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(T,r));
	v.add(W.union(s,T));
	v.add(W.union(r,W));
	v.add(s.disj(T));
	v.add(T.neq(this));
	return v;
    }

    
    public ConstraintsConjunction ndiffer(Set r, Lvar s) {  
	Set W = new Set();
	Set T = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(T,r));
	v.add(W.union(s,T));
	v.add(W.union(r,W));
	v.add(s.disj(T));
	v.add(T.neq(this));
	return v;
    }


    public ConstraintsConjunction ndiffer(Lvar r, Lvar s) {  
	Set W = new Set();
	Set T = new Set();
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(r.union(T,r));
	v.add(W.union(s,T));
	v.add(W.union(r,W));
	v.add(s.disj(T));
	v.add(T.neq(this));
	return v;
    }


    public ConstraintsConjunction less(Lvar L, Set T) throws Failure {  
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(this.eq(T.ins(L))); 
	v.add(L.nin(T)); 
	return v;
    }    


    public ConstraintsConjunction less1(Lvar L, Set T) throws Failure {  
	if (this.isGround()) {
	    ConstraintsConjunction v = new ConstraintsConjunction();
	    v.add(L.eq(this.first())); 
	    v.add(T.eq(this.sub()));
	    return v;
	}
	else {
	    ConstraintsConjunction v = new ConstraintsConjunction();
	    v.add(this.eq(T.ins(L))); 
	    v.add(L.nin(T)); 
	    return v;
	}
    }    


    public ConstraintsConjunction less(Lvar L, Lvar T) throws Failure {
	ConstraintsConjunction v = new ConstraintsConjunction();
	v.add(this.eq(T.ins(L))); 
	v.add(L.nin(T)); 
	return v;
    }    


    /* Stampa il nome del Set su cui  stato invocato e il suo valore. */
    public void output() {     
	System.out.println();
	System.out.print(this.name + " = ");
	if (this.equ == null) {
	    if (!this.iniz) System.out.print("unknown");
	    else this.print();
	}
	else this.equ.print();
	System.out.println();
	return;
    }

    
    /* Il metodo print, se l'insieme su cui  stato invocato  inizializzato, stampa il 
       suo valore, altrimenti stampa il suo nome preceduto da "_" per poter distinguere 
       i nomi dalle stringhe o caratteri usate come valore per le variabili.
       Per le liste e gli insiemi l'output  dato secondo le convenzioni sintattiche 
       del Prolog. */
    public void print() {
	if (this.equ == null) {
	    if (this == empty) System.out.print("{}");
	    else if (!this.iniz) {
		System.out.print("_");
		printName(this);
	    }
	    else if (this.isInt) {
		((Int)this.inter).print();
		return;
	    }
	    else {
		System.out.print("{");
		Vector vett = this.toVector();
		for(int i = 0; i < vett.size(); i++) {
		    if (vett.get(i) instanceof Lvar) ((Lvar)vett.get(i)).print();
		    else if (vett.get(i) instanceof Lst) {
			if (((Lst)vett.get(i)).iniz) ((Lst)vett.get(i)).print();
			else System.out.print("_"+((Lst)vett.get(i)).name);
		    }
		    else if (vett.get(i) instanceof Set) {
			if (((Set)vett.get(i)).iniz) ((Set)vett.get(i)).print();
			else System.out.print("_"+((Set)vett.get(i)).name);
		    }
		    else System.out.print(vett.get(i));
		    if (i < (vett.size() - 1)) System.out.print(",");
		}
		if (!this.ultimoResto().equals(Set.empty)) {
		    System.out.print("|");
		    if (this.ultimoResto().isInt) ((Int)this.ultimoResto().inter).print();
		    else printName(this.ultimoResto());
		}
		System.out.print("}");
	    }
	} 
	else this.equ.print();
	return;
    }


    public static void printName(Set z) {
	if (z.equ == null) {
	    System.out.print(z.name);
	    return;
	}
	else printName(z.equ);
	return;
    }
 
    
    public void setName(String n) {
	this.name = n;
	return;
    }
    
    
    public String getName() {
	return this.name;
    }
 
 
    /**
     * Creates and returns a copy of this object. The  method implements an
     * equivalence relation on non-null object and object references.
     * @return a clone of this instance.
     */      
    public Set clone() {
        return this.clona();
    }
    
    
    protected Set clona() {
	if (this.equ == null) {
	    Set clone = new Set();
	    if (this == Set.empty) return Set.empty;
	    else if (!this.iniz) return clone;
	    else if (this.isInt) {
		clone.iniz  = true;
		clone.isInt = true;
		clone.inter = this.inter;
		return clone;
	    }
	    else {
		Vector listaClone = new Vector();
		for (int i = 0; i < this.lista.size(); i++) {
		    if (lista.get(i) instanceof Lvar) listaClone.add(((Lvar)lista.get(i)).clona()); 
		    else if (lista.get(i) instanceof Lst) listaClone.add(((Lst)lista.get(i)).clona()); 
		    else if (lista.get(i) instanceof Set) {
			if (((Set)lista.get(i)) == empty) listaClone.add(empty);
			else listaClone.add(((Set)lista.get(i)).clona()); 
		    }
		    else listaClone.add(lista.get(i));
		}
		clone.lista = listaClone;
		clone.resto = this.resto.clona();
		clone.iniz  = this.iniz;
                clone.id    = this.id;
		clone.name  = this.name;
		clone.equ   = null;
		return clone;
	    }
	}
	else return this.equ.clona();
    } 

    
    /* Costruisce un Set di n variabili logiche non inizializzate */
    public static Set mkset(int n) {
	Vector lista = new Vector();   
	for (int i = 0; i < n; i++) {
	    Lvar lvar = new Lvar();
	    lista.add(lvar);    
	}
	Set set = new Set(lista, empty);
	return set;
    }

    
    public static Set mkset(String name, int n) {
	Vector lista = new Vector();   
	for (int i = 0; i < n; i++) {
	    Lvar lvar = new Lvar();
	    lista.add(lvar);    
	}
	Set set = new Set(name,lista, empty);
	return set;
    }
    
  
    public Object get(int i) {
	Vector v = this.toVector();
	return v.get(i);
    }

    /**
     * Gets the unique identifier  for the object.
     * @return an int identifier for the object.
     */   
    protected int getId() {
	return this.id;
    }    
    
    /**
     * Sets the unique identifier for the object.
     */  
    protected void setId() {
        this.id = counter++;
        return;
    }    

        
    /**
     * Deterministic test that verify if this Set has no variables.
     *
     * @return true if this object is empty, completely specified and doesn't contains uninitialized or not assigned variables, false otherwise.
     */  
    public boolean isGround() { // restituisce true se il Set  completamente noto
        if (this.equ == null) {
            if (this == empty) return true;
            else if(!iniz) return false;
            else if(this.isInt) return true;
            else if(!ultimoResto().iniz) return false;
            else {
                for(int i = 0; i < this.size(); i++){
                    if (this.get(i) instanceof Lvar) { if(!((Lvar)this.get(i)).isGround()) return false; }
                    else if (this.get(i) instanceof Lst) { if(!((Lst)this.get(i)).isGround()) return false; }
                    else if (this.get(i) instanceof Set) { if(!((Set)this.get(i)).isGround()) return false; }
                }
                return true;
            }
        }
        else return this.equ.isGround();
    }
    

    /**
     * Tests if this Set is initialized.
     *
     * @return true if and only if the Set is initialized, false otherwise.
     */
    public boolean known() {
	if (this.equ == null) return this.iniz;
	else return this.equ.known();
    }     


    /**
     * Replaces all the occurrences of a varible with another.
     *
     * @return A new Set objects where the occurrences of the first (parameter) variable are replcaed by the second (parameter) variable.
     */
    protected Set sost(Lvar y, Lvar x) {        // Replace y with x
	if (this.equ == null && y.equ == null && x.equ == null) {
	    if (!this.iniz || this.isEmptySL() )return this;
	    else{
		Vector v = this.toVector();
		Vector vnew = new Vector();
		for (int i = 0; i < v.size(); i++) {
		    if (v.get(i) instanceof Lvar) {
			if ((Lvar)v.get(i) == y) vnew.add(x);
			else if (((Lvar)v.get(i)).val instanceof Set) vnew.add(((Set)((Lvar)v.get(i)).val).sost(y,x));
			else if (((Lvar)v.get(i)).val instanceof Lst) vnew.add(((Lst)((Lvar)v.get(i)).val).sost(y,x));
			else vnew.add((Lvar)v.get(i));
		    }
		    else if(v.get(i) instanceof Set) vnew.add(((Set)v.get(i)).sost(y,x));
		    else if(v.get(i) instanceof Lst) vnew.add(((Lst)v.get(i)).sost(y,x));
		    else vnew.add(v.get(i));
		}
		Set snew = new Set(vnew,this.ultimoResto().sost(y,x));
		return snew;
	    }
	}
	else if (this.equ != null) return this.equ.sost(y,x);
	else if (y.equ != null) return this.sost(y.equ,x);
	else return this.sost(y,x.equ);
    }
    

    /**
     * Get the values of the set.
     * @return the value of the set.
     */    
    public Object getValue() { 
        return this.value();
    }       
    
    
    protected Set value() {
	if(this.equ == null) {
	    if (this == empty || !iniz) return this;
	    else if(this.isInt) return this;
	    else {
		Vector newlista = new Vector();
		for (int i = 0; i < lista.size(); i++) {
		    if (lista.get(i) instanceof Lvar) {
			newlista.add(((Lvar)lista.get(i)).value());
		    }
		    else if (lista.get(i) instanceof Lst) {
			newlista.add(((Lst)lista.get(i)).value());
		    }
		    else if (this.get(i) instanceof Set) {
			newlista.add(((Set)lista.get(i)).value());
		    }
		    else newlista.add(lista.get(i));
		}
		return new Set(newlista,resto.value());
	    }
	}
	else return this.equ.value();
    }


    public Object read() throws IOException, NotSetException, InstantiationError {
	if(iniz == false) {
	    InputStreamReader reader = new InputStreamReader(System.in);
	    BufferedReader input = new BufferedReader(reader);
	    String valore = input.readLine();
	    if(valore.charAt(0)== '{') return this.readSet(valore.substring(1,valore.lastIndexOf('}')));
	    else throw new NotSetException();
	}
	else throw new InstantiationError();
    }

    
    protected static Object sread(String valore) throws IOException {
	if (valore.charAt(0)== '"') {
	    if (valore.substring(1).indexOf('"') != -1) {
		return new String(valore.substring(1,valore.substring(1).indexOf('"')+1));
	    }
	    else {
		return new String(valore.substring(1));
	    } 
	}
	else if (valore.charAt(0)== '\'') return new Character(valore.charAt(1));
	else if (valore.charAt(0)== '{') return sreadSet(valore.substring(1,valore.lastIndexOf('}')));
	else if(valore.charAt(0)== '[') return Lst.sreadLst(valore.substring(1,valore.lastIndexOf(']')));
	else{ 
	    try { return new Integer(valore); }
	    catch(NumberFormatException e) {
		try{ return new Float(valore); }
		catch(NumberFormatException f) { throw new IOException(); }     
	    }
	}
    }


    protected Object readSet(String st) throws IOException {
	Set set = new Set(Set.empty);
	String s = new StringBuffer().append(",").append(st).toString();
	int old_i = s.length()-1;
	for (int i = s.length()-1; i >= 0; i--) {
	    if (s.charAt(i) == '}') {
		int begins = Set.pareggiaSet(s,i);
		set = set.ins(sreadSet(s.substring(begins+1,i)));
		old_i = begins-2;
		i = begins-1;
	    }
	    else if (s.charAt(i) == ']') {
		int beginl = Lst.pareggiaLst(s,i);
		set = set.ins(Lst.sreadLst(s.substring(beginl+1,i)));
		old_i = beginl-2;
		i = beginl-1;
	    }
	    else if (s.charAt(i) == ',') {
		set = set.ins(sread(s.substring(i+1,old_i+1)));
		old_i = i-1;
	    }
	}
	this.iniz = true;
	this.lista = set.lista;
	this.resto = set.resto;
	return set;
    }

    protected static Set sreadSet(String st) throws IOException {
	Set set = new Set(Set.empty);
	String s = new StringBuffer().append(",").append(st).toString();
	int old_i = s.length()-1;
	for (int i = s.length()-1; i >= 0; i--) {
	    if (s.charAt(i) == '}') {
		int begins = Set.pareggiaSet(s,i);
		set = set.ins(sreadSet(s.substring(begins+1,i)));
		old_i = begins-2;
		i = begins-1;
	    }
	    else if (s.charAt(i) == ']') {
		int beginl = Lst.pareggiaLst(s,i);
		set = set.ins(Lst.sreadLst(s.substring(beginl+1,i)));
		old_i = beginl-2;
		i = beginl-1;
	    }
	    else if (s.charAt(i) == ',') {
		set = set.ins(sread(s.substring(i+1,old_i+1)));
		old_i = i-1;
	    }
	}
	return set;
    }


    protected static int pareggiaSet(String s, int n) throws IOException {
	int j = -1;
	for (int i = n-1; i >= 0; i--) {
	    if (s.charAt(i) == '}') j=j-1;
	    else if (s.charAt(i) == '{') {
		if (j == -1) return i;
		else j=j+1;
	    }
	}
	throw new IOException();
    }
    
    
    public Set normalizeSet() {
        if (this.equ == null) {
            if (!this.iniz) throw new NotInitVarException();
            else if (this.isEmpty()) return Set.empty;
            else {
                Set nodup = new Set(Set.empty);
                Vector t_set = this.toVector();
                //Object t_tmp = new Object();
                
                Vector indexofduplicated = new Vector();
                int i = 0;

                while (i < t_set.size()){
                    Integer Iint = new Integer(i);
                    Object t_tmp = t_set.get(i);    // t_tmp = t_set.get(i);                      
                    if (t_tmp instanceof Set)  t_tmp = ((Set)t_tmp).normalizeSet();
                    if (t_tmp instanceof Lst)  t_tmp = ((Lst)t_tmp).normalizeLst();
                    if (t_tmp instanceof Lvar) t_tmp = ((Lvar)t_tmp).normalizeLvar();
                    
                    if ((indexofduplicated == null) ||
                        (indexofduplicated != null && !indexofduplicated.contains(Iint))) {                        
                        nodup = nodup.ins(t_tmp);
                        
                        if (i+1 <= t_set.size()){
                            // Object t_tmp2 = new Object();
                            boolean found;
                            for (int j=i+1; j< t_set.size(); j++) {
                                Integer Jint = new Integer(j);
                                found = false;
                                Object t_tmp2 = t_set.get(j); // t_tmp2 = t_set.get(j);
                                                               
                                if ((t_tmp instanceof Lvar) && (t_tmp2 instanceof Lvar) && (((Lvar)t_tmp).equals((Lvar)t_tmp2)))  found = true;
                                else if ((t_tmp instanceof Lst) && (t_tmp2 instanceof Lst) && (((Lst)t_tmp).equals((Lst)t_tmp2))) found = true;
                                else if ((t_tmp instanceof Set) && (t_tmp2 instanceof Set) && (((Set)t_tmp).equals((Set)t_tmp2))) found = true;
                                else if (t_tmp.equals(t_tmp2)) found = true;
                                
                                if (found) indexofduplicated.add(Jint);
                            }
                        }
                    }
                    i++;
                }
                return nodup;
            }            
        }
        else return this.equ.normalizeSet();
    }
    
    
    /**
     * Return a string corresponding to the object value.
     *
     * @return a string correspondig to the object value.
     */
    public String toString() {
	if (this.equ == null) {
	    if (!this.iniz) return "Set: "+this.name+", id: "+this.id+", is NOT initialized." ;
	    else {
                String outp = "Set: "+this.name+", id: "+this.id+", elements: {" ;
                Vector vett = this.toVector();
		for(int i = 0; i < vett.size(); i++) {
		    if (vett.get(i) instanceof Lvar) outp += ((Lvar)vett.get(i)).toString();
		    else if (vett.get(i) instanceof Lst) outp += ((Lst)vett.get(i)).toString();
		    else if (vett.get(i) instanceof Set) outp += ((Set)vett.get(i)).toString();			
		    else outp += vett.get(i);
		    if (i < (vett.size() - 1)) outp += ", ";
		}
                outp += "}, rest: {";
                if (this.ultimoResto() != null && !this.ultimoResto().iniz)
                    if (!this.ultimoResto().equals(Set.empty)) outp += this.ultimoResto().toString();
                outp += "} ";
                return outp;                
		// isInt 
                // inter    
            } 
//                if (this.isInt) return this;
//	    else if (this.resto == empty || this.resto == null || !this.resto.iniz || this.resto.isEmptySL()) return this.resto;
//	    else return this.resto.ultimoResto();    
	}
	else return this.equ.toString();
    }
    
//    public String toString() {
//        Set tmp = (this.clona()).normalizeSet();
//        Vector v = tmp.toVector();
//        
//        return (String)this.name+ v.toString();
//    }

}

