package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;

import java.util.*;

/**
 * The <code>SolverClass</code> class implements the abstraction of the solver.
 * Now the solver is istantiable!
 * 
 * @author  Elio Panegai
 * @version 1.3.0
 */
public class SolverClass { 


    /**
     * Checks if there are variations on the constraint store;
     * true if no variations occurred, false otherwise.
     */
    protected boolean storeInvariato = true;
    
    /**
     * Used to solve only precise constraint on the constraint store.
     */
    protected int ctx = 0;     // ctx = context.  
    
    /**
     * Size of the constraint store.
     */
    protected int storeSize = 0; 
    
    /**
     * Computation time.
     */
    private static Date date;  


    /**
     * The constraint store.
     */
    protected ConstraintStore store = new ConstraintStore(this);  
    
    /**
     * A backtracking object exlusively by this solver.
     */
    protected Backtracking B = new Backtracking(this);
    
    /**
     * The store state of this solver.
     */
    protected StoreState SS = new StoreState(this);  
    
    /**
     * Due to this instance the NewConstrains object can deal with solver objects.
     */
    protected NewConstraintsClass N;
    
    /**
     * Due to this instance the LibConstrains object can deal with solver objects.
     */
    protected LibConstraints LC = new LibConstraints(this); 
    
    /**
     * The rules for rewriting constraints. Due to this instance the 
     * RewritingConstraintsRules object can deal with solver objects such 
     * as Backtracking.
     */
    protected RewritingConstraintsRules C = new RewritingConstraintsRules(this);
    
    
    static {
        date = new Date();
    }
  
    
    /* Member Methods */

/*
    public void set_ctx(int n) {
	this.ctx = n;
	return;
    }


    public int get_ctx() {
	return this.ctx;
    }
*/
    
    public void setNewConstraints(NewConstraintsClass x){
        this.N = x; 
    }

    public void add(Constraint s) {
	this.store.add(s);   // Add the constraint s at the end of the store
	return;
    }


    /* Questo metodo add aggiunge un vettore di vincoli in coda allo store,
       ad esempio l'espressione: Solver.add(X.eq(Y).and(Z.neq(X)).and(x.in(S)));
       introduce nello store i tre vincoli contenuti nel vettore passato come 
       parametro. Infatti X.eq(Y).and(Z.neq(X)).and(x.in(S))  un vettore di 
       tre vincoli, cio di tre oggetti di tipo Constraint. */  
    /**
     * Adds a constraint conjunction in tail to the store.
     * @param storeVector The constraints to add to the store.
     */
     public void add(ConstraintsConjunction storeVector) {	 
    	for(int i = 0; i < storeVector.size(); i++) {
	    this.store.add((Constraint)storeVector.get(i));
	}
	return;
    }
        

    /**
     * Add the constraint at the i-th position of the store.
     * @param i A position in the constraint store.
     * @param s The constraint to added to the store.
     */
    public void add(int i, Constraint s) {	
	this.store.add(i, s); // Add the constraint s at the i-th position of the store
	return;
    }

    
    public void addChoicePoint(Constraint s) {
	B.addChoicePoint(s);   
	return;
    }


    /**
     * Searches for the first occurence of the given argument.
     * @param s The constraint to research.     
     * @return the index of the first occurrence of the constraint searched.
     * Returns -1 if the constraint is not found.
     */
    public int indexOf(Constraint s) {	
	return this.store.indexOf(s);   
    }


    public int size() {
	return this.store.size();
    }


    public Object get(int x) {
	return this.store.get(x);
    }


    public void removeAllElements() {
	this.store.removeAllElements();
	return;
    }


    /* Il metodo solve scorre lo store e risolve i vincoli che non sono ancora in forma
       risolta. Se un vincolo  inconsistente viene sollevata un'eccezione Fail dal
       metodo fail; se c' un fallimento il catch della solve() cattura l'eccezione e
       interrompe la solve passando il controllo al metodo gestisciFallimento() della classe                      
       Backtracking che innesca il backtracking.
       Quando la solve arriva al termine dello store, se lo store  rimasto invariato,
       termina altrimenti ricomincia la risoluzione dei vincoli.*/
    /**
     * Solves (using backtracking) the constraints that are not in solved form. 
     * Note: A Fail exception can be raised if constraints are inconsistents, this 
     * exception is catched inside the method and starts the backtracking procedure.
     * @throws JSetL.Failure If doesn't exist any solution for the constraints.
     */
    public void solve() throws Failure {  
        if (this.store.existForall()) {
            /* Check for quantified vars. Quantified forall's variables do not have to be used 
               in other constraints. This vars sholud appear newly in other foralls but always 
               as quantified.
             */                     
            this.store.getConstraintStoreVars();
            if(this.store.ctrlConstraintStoreVars()) throw new IllegalVariableUse();           
        }
	if(B.alternatives.size() > 0) {
	    B.updateStoreAlternative(); 
	    B.updateVarAlternative();
	    /* I metodi updateStoreAlternative() e updateVarAlternative() permettono di
	       aggiungere ai punti di scelta creati da una solve precedente i vincoli introdotti
	       successivamente alla loro creazione. In questo modo in caso di backtracking non
	       vengono persi i vincoli introdotti dopo la creazione del punto di scelta al quale
	       il risolutore  tornato durante il backtracking. */
	}
	do {
	    try { this.ris(); }
	    catch(Fail f){ B.gestisciFallimento(); }
	} while(!this.storeInvariato);

	/* Il campo booleano storeInvariato permette di stabilire se il constraint store 
	    gi in forma risolta o se  necessario ricominciare a scorrerlo per continuare
	   a risolvere i vincoli che non sono ancora in forma risolta.
	   il valore di storeInvariato  true quando una passata della solve non produce
	   alcuna modifica sul constraint store. */
	// check_union();

	/* Al termine della risoluzione del constraint store, il metodo check_union() 
	   ricerca eventuali vincoli che rientrino nei casi (6) e (7) della procedura
	   di riscrittura della union (cfr. "Sets and Constraint logic programming"). */
	this.storeSize = this.store.size();
	return;
    }

    
    /**
     * Solves (using backtracking) the constraints that are't in solved form. 
     * Before the resolution starts, the constraint parameter will be added to the store.
     * @param cc A Constraint or a ConstraintsConjunction that will be added 
     * to the constraint store.
     * @throws JSetL.Failure If doesn't exist any solution for the constraints.
     * @throws JSetL.NotDefinedConstraint If there is a mismatching constraints identifier in the constraint store.
     * @see SolverClass#solve
     */
    public void solve(Object cc) throws Failure, NotDefinedConstraint {        
        if (cc instanceof ConstraintsConjunction) this.add((ConstraintsConjunction)cc); 
        else if (cc instanceof Constraint) this.add((Constraint)cc);
        else throw new NotDefinedConstraint();
        
        this.finalSolve();
        return;
   }

    
    /**
     * Solves the constraints that are't in solved form.
     * @return True if the constraints are in solved form.
     * @throws JSetL.Failure If doesn't exist any solution for the constraints.
     */
    public boolean boolSolve() {
        VarState   tmpVar   = this.B.getVarState();
        StoreState tmpStore = this.B.getStoreState();
	try {
            SolverClass temporarySolver = new SolverClass();
            temporarySolver.add(this.store.getStore());    
            
	    temporarySolver.solve();            
            this.store = temporarySolver.store;   // semantically: this = temporarySolver;
	    return true;
	}
	catch(Failure e) { 
            this.B.restoreVarState(tmpVar); 
            this.B.restoreStoreState(tmpStore.statoStore); 
	    return false; 
	}
    }
    
    
    /**
     * Resolves the constraints that are't in solved form.
     * Before the resolution starts, the constraint parameter will be added to the store.
     * @param cc A Constraint or a ConstraintsConjunction 
     * that will be added to the constraint store.
     * @return True if the constraints are in solved form.
     * @throws JSetL.NotDefinedConstraint If there is a mismatching constraints identifier in the constraint store.
     * @see SolverClass#solve
     * @see SolverClass#finalSolve
     */
    public boolean boolSolve(Object cc) throws NotDefinedConstraint {  
        VarState   tmpVar   = this.B.getVarState();
        StoreState tmpStore = this.B.getStoreState();
        try {   
            SolverClass temporarySolver = new SolverClass();
            temporarySolver.add(this.store.getStore());            
            
            if (cc instanceof ConstraintsConjunction) temporarySolver.add((ConstraintsConjunction)cc); 
            else if (cc instanceof Constraint) temporarySolver.add((Constraint)cc);
            else throw new NotDefinedConstraint();
        
            temporarySolver.solve();
            this.store = temporarySolver.store;   // semantically: this = temporarySolver;
            return true;
        }
        catch (Failure f){
            this.B.restoreVarState(tmpVar); 
            this.B.restoreStoreState(tmpStore.statoStore); 
            return false;
        }
    }
    

    /**
     * Solves the constraints that are't in solved form.
     * @return True if the constraints are in solved form.
     * @throws JSetL.Failure If doesn't exist any solution for the constraints.
     */
    public boolean boolSolve1() {
        VarState   tmpVar   = this.B.getVarState();
        StoreState tmpStore = this.B.getStoreState();
	try {
            SolverClass temporarySolver = new SolverClass();
            temporarySolver.add(this.store.getStore());    
            
	    temporarySolver.solve1();            
            this.store = temporarySolver.store;   // semantically: this = temporarySolver;
	    return true;
	}
	catch(Failure e) { 
            this.B.restoreVarState(tmpVar); 
            this.B.restoreStoreState(tmpStore.statoStore); 
	    return false; 
	}
    }    

    
    /* Il metodo solve scorre lo store e risolve i vincoli che non sono ancora in forma
       risolta. Se un vincolo  inconsistente viene sollevata un'eccezione Fail dal
       metodo fail; se c' un fallimento il catch della solve() cattura l'eccezione e
       interrompe la solve passando il controllo al metodo gestisciFallimento() della 
       classe Backtracking che innesca il backtracking.
       Quando la solve arriva al termine dello store, se lo store  rimasto invariato,
       termina altrimenti ricomincia la risoluzione dei vincoli. */
    public void finalSolve() throws Failure {
        if (this.store.existForall()) {
            /* Controlla che le vars quantificate presenti nel forall non vengano utilizzate
               in altri vincoli. Queste vars possono comparire nuovamente in altri forall
               ma sempre come vars quantificate. */
            this.store.getConstraintStoreVars();
            if(this.store.ctrlConstraintStoreVars()) throw new IllegalVariableUse();           
        }
	if(B.alternatives.size() > 0) {
	    B.updateStoreAlternative();
	    B.updateVarAlternative();
	    /* I metodi updateStoreAlternative() e updateVarAlternative() permettono di
	       aggiungere ai punti di scelta creati da una solve precedente i vincoli introdotti
	       successivamente alla loro creazione. In questo modo in caso di backtracking non
	       vengono persi i vincoli introdotti dopo la creazione del punto di scelta al quale
	       il risolutore  tornato durante il backtracking. */
	}
	do {
	    try{ this.ris(); }
	    catch(Fail f){ B.gestisciFallimento(); }
	} while(!this.storeInvariato);

	/* Il campo booleano storeInvariato permette di stabilire se il constraint store 
	    gi in forma risolta o se  necessario ricominciare a scorrerlo per continuare
	   a risolvere i vincoli che non sono ancora in forma risolta.
	   il valore di storeInvariato  true quando una passata della solve non produce
	   alcuna modifica sul constraint store. */
	check_union();
	check_unsolved_arit_cons();
	/* Al termine della risoluzione del constraint store, il metodo check_union() 
	   ricerca eventuali vincoli che rientrino nei casi (6) e (7) della procedura
	   di riscrittura della union (cfr. "Sets and Constraint logic programming"). */
	this.storeSize = this.store.size();
	return;
    }


    /* Il metodo solve1() richiama il metodo solve per determinare una possibile 
       soluzione, ma poi elimina tutte le Backtracking.alternative rimaste inesplorate, cio 
       realizza un "cut" stile Prolog. */
    public void solve1() throws Failure {
	this.solve();
	B.alternatives.removeAllElements(); 
	return;
    }

 
    /**
     * Finds all possible solutions for an Lvar that satisfies the constraints store.
     * @return The Set of all possibles values for the logical variables.
     * @param ob The logical variable of which we search the values.
     * @throws JSetL.Failure If doesn't exist any solution for the constraints.
     * @see SolverClass#solve
     * @see SolverClass#finalSolve
     */    
    public Set setof(Lvar ob) throws Failure {
        return setofSL(ob);        
    }
    
    
    /**
     * Finds all possible solutions for a Lst that satisfies the constraints store.
     * @param ob The list of which we search the values.
     * @throws JSetL.Failure If doesn't exist any solution for the constraints.
     * @return The Set of all possibles values for the list.
     * @see SolverClass#solve
     * @see SolverClass#finalSolve
     */
    public Set setof(Lst ob) throws Failure {
        return setofSL(ob);
    }
    
    
    /**
     * Finds all possible solutions for a Set that satisfies the constraints store.
     * @param ob The set of which we search the values.
     * @throws JSetL.Failure If doesn't exist any solution for the constraints.
     * @return The Set of all possibles values for the set.
     * @see SolverClass#solve
     * @see SolverClass#finalSolve
     */
    public Set setof(Set ob) throws Failure {
        return setofSL(ob);
    }   
    
    /**
     * Finds all possible solutions for a MultiSet that satisfies the constraints store.
     * @param ob The multiset of which we search the values.
     * @throws JSetL.Failure If doesn't exist any solution for the constraints.
     * @return The MultiSet of all possibles values for the set.
     * @see SolverClass#solve
     * @see SolverClass#finalSolve
     */    
    public MultiSet setof(MultiSet ob) throws Failure {
        return multisetofSL(ob);
    }     
    
    
    public Set setof(Object ob, Object cc) throws Failure, NotDefinedConstraint {        
        if (cc instanceof ConstraintsConjunction) this.add((ConstraintsConjunction)cc); 
        else if (cc instanceof Constraint)        this.add((Constraint)cc);
        else throw new NotDefinedConstraint();        
        return this.setofSL(ob);        
   }  
    
    
    /* This method calls the solveSetof() method that is similar in the implementation to the solve() method
       but, when it has found a solution, it store it and forces a failure with gestisciFallimentoSetof(),
       to find another choice point and another new solution.
       While all Backtracking.alternative aren't explored the computation goes on. */
    private Set setofSL(Object ob) throws Failure {
        Vector allValues = new Vector();
        this.finalSolve();
        Object newO;
        if (ob instanceof Lvar) newO = ((Lvar)ob).clona();
        else if (ob instanceof Lst) newO = ((Lst)ob).clona();
        else newO = ((Set)ob).clona();         
        allValues.add(newO);
        
        while (B.alternatives.size() != 0) {
	    B.gestisciFallimentoSetof();
	    if (solveSetof()) {
                Object newOcp;
                if (ob instanceof Lvar) newOcp = ((Lvar)ob).clona();
                else if (ob instanceof Lst) newOcp = ((Lst)ob).clona();
                else newOcp = ((Set)ob).clona();                               
		allValues.add(newOcp);
	    }
	}
	return new Set(allValues, Set.empty);        
    }    

    /* like  Set setofSL(Object ob) but for MultiSet */
    private MultiSet multisetofSL(Object ob) throws Failure {
        Vector allValues = new Vector();
        finalSolve();
        Object newO;
        if (ob instanceof Lvar) newO = ((Lvar)ob).clona();
        else if (ob instanceof Lst) newO = ((Lst)ob).clona();
        else if (ob instanceof MultiSet) newO = ((MultiSet)ob).clona();
        else newO = ((Set)ob).clona();              // if (ob instanceof Set)
        
        allValues.add(newO);
        
        while (B.alternatives.size() != 0) {
	    B.gestisciFallimentoSetof();
	    if (solveSetof()) {
                Object newOcp;
                if (ob instanceof Lvar) newOcp = ((Lvar)ob).clona();
                else if (ob instanceof Lst) newOcp = ((Lst)ob).clona();
                else if (ob instanceof MultiSet) newOcp = ((MultiSet)ob).clona();
                else newOcp = ((Set)ob).clona();    // if (ob instanceof Set)
		allValues.add(newOcp);
	    }
	}
	return new MultiSet(allValues, MultiSet.empty);        
    }
    

    /**
     * The nextSolution method, in case of multiple solutions,
     * allows to find a new solution each time it is called.
     * @throws JSetL.Failure If doesn't exist any solution for the constraints.
     * @return true if exist a next solution, false otherwise.
     * @see SolverClass#solve
     * @see SolverClass#finalSolve
     * @see SolverClass#boolSolve
     */      
    public boolean nextSolution() throws Failure {
	if (B.alternatives.size() > 0) {
	    B.gestisciFallimentoSetof();
	    return this.solveSetof();
	}
	else return false;
    }

    
    private boolean solveSetof() throws Failure {
        if (this.store.existForall()) {
            /* Controlla che le vars quantificate presenti nel forall non vengano utilizzate
               in altri vincoli. Queste vars possono comparire nuovamente in altri forall
               ma sempre come vars quantificate. */
            this.store.getConstraintStoreVars();
            if(this.store.ctrlConstraintStoreVars()) throw new IllegalVariableUse();           
        }
	if (B.alternatives.size() > 0) {
	    B.updateStoreAlternative();    
	    B.updateVarAlternative();
	}
	do {
	    try { this.ris(); }
	    catch(Fail f) {
		if(B.alternatives.size() > 0) B.gestisciFallimentoSetof();
		else return false;
	    }
	} while(!this.storeInvariato);
        
	//check_union();
	this.storeSize = this.store.size();
	return true;
    }


    private void ris() throws Failure, Fail {
	this.storeInvariato = true;
	int i = 0;
	do {
	    C.risElem(i);   
	    i++;
	} while(i < store.size());
	return;
    }


    protected void check_union() throws Failure {
	int i = 0;
	do {
	    this.check_union(i);
	    i++;
	} while(i < store.size());
	return;
    }
  

    protected void check_union(int k) throws Failure {
	Constraint s = (Constraint)this.store.get(k);
	if(!s.bool && s.cons == Enviroment.unionCode) check_neq(s);
	return;
    }


    protected void check_neq(Constraint s) throws Failure {
	int i = 0;
	do {
	    check_neq(s,i);
	    i++;
	} while(i < store.size());
	return;
    }


    protected void check_neq(Constraint s_union, int k) throws Failure {
	Constraint s = (Constraint)this.store.get(k);
	if(!s.bool && s.cons == Enviroment.neqCode) check_union_neq(s_union, s);
	return;
    }


    protected void check_union_neq(Constraint s_union, Constraint s_neq) throws Failure {
	if(s_union.arg3 == s_neq.arg1) {
	    s_neq.cons = Enviroment.neq_unionCode;
	    s_neq.caseControl = 0;
	}
	else if(s_union.arg3 == s_neq.arg2) {
	    s_neq.cons = Enviroment.neq_unionCode;
	    s_neq.arg2 = s_neq.arg1;
	    s_neq.arg1 = s_union.arg3;
	    s_neq.caseControl = 0;
	}
	else if(s_union.arg1 == s_neq.arg1) {
	    s_neq.cons = Enviroment.neq_unionCode;
	    s_neq.caseControl = 0;
	}
	else if(s_union.arg1 == s_neq.arg2) {
	    s_neq.cons = Enviroment.neq_unionCode;
	    s_neq.arg2 = s_neq.arg1;
	    s_neq.arg1 = s_union.arg1;
	    s_neq.caseControl = 0;
	}
	else if(s_union.arg2 == s_neq.arg1) {
	    s_neq.cons = Enviroment.neq_unionCode;
	    s_neq.caseControl = 0;
	}
	else if(s_union.arg2 == s_neq.arg2) {
	    s_neq.cons = Enviroment.neq_unionCode;
	    s_neq.arg2 = s_neq.arg1;
	    s_neq.arg1 = s_union.arg2;
	    s_neq.caseControl = 0;
	}
	do {
	    try{ ris(); }
	    catch(Fail f){ B.gestisciFallimento(); }
	} while(!storeInvariato);
	return;
    }

    
    protected void check_unsolved_arit_cons() throws Failure {
	for(int i = 0; i < this.store.size(); i++) {
	    Constraint s = (Constraint)store.get(i);
	    if(!s.bool) {
		if(s.cons == Enviroment.geCode || 
                   s.cons == Enviroment.gtCode || 
                   s.cons == Enviroment.leCode || 
                   s.cons == Enviroment.ltCode ||
		   s.cons == Enviroment.sumCode || 
                   s.cons == Enviroment.subCode || 
                   s.cons == Enviroment.mulCode || 
                   s.cons == Enviroment.divCode || 
                   s.cons == Enviroment.modCode)
		    throw new Uninitialized_Variable_in_arithmetical_expression();
		else continue;
	    }
	}
	return;
    }

    
    /* Al campo bool del Constraint fallito viene assegnato il valore true 
       in modo che quel vincolo venga eliminato logicamente.          
       La ragione di questa scelta pu essere chiarita dal seguente semplice 
       esempio. Supponiamo che venga introdotto e risolto un vincolo che non 
       ha nessuna soluzione, in questo caso verr sollevata un'eccezione Failure.
       Supponiamo che questa eccezione venga catturata e venga gestita 
       semplicemente dando in output un messaggio che indica che non  stata 
       trovata nessuna soluzione. Supponiamo che in seguito venga introdotto un
       altro vincolo che invece ammette soluzioni e che venga richiamata la solve.
       Se il vincolo inconsistente non avesse il campo bool uguale a true, il 
       risolutore cercherebbe una soluzione soddisfacente anche quel vincolo e 
       si avrebbe di nuovo un fallimento. In questo modo invece il vincolo che ha
       generato l'eccezione Failure gestita in precedenza non verr pi considerato
       e verr determinata una soluzione soddisfacente i vincoli introdotti 
       successivamente (cfr. esempio SimmList). */
    public void fail(Constraint st) throws Fail {
	st.bool = true;
	throw new Fail();
    }

    
    public void failure() throws Failure {
	throw new Failure();
    }

    
    /* L'espressione Solver.add(allDifferent(s)) permette di aggiungere al constraint store 
       i vincoli di disuguaglianza tra tutti gli elementi di s, dove s pu essere una Lst
       un Set o una Lvar inizializzata con Lst o Set. */    
    public void allDifferent(Object s) {
        if (s instanceof Lvar) this.add(allDiff((Lvar)s));
        else if (s instanceof Lst) this.add(allDiff((Lst)s));
        else if (s instanceof Set) this.add(allDiff((Set)s));        
        return;
    }
    
    
    protected ConstraintsConjunction allDiff(Lst s) {
	Vector v = s.toVector();
	ConstraintsConjunction storeVector = new ConstraintsConjunction();
	for(int i = 0; i < v.size()-1; i++) {
	    for(int j = i+1; j < v.size(); j++) {
		Constraint st = new Constraint(v.get(i), Enviroment.neqCode, v.get(j));
		storeVector.add(st);
	    }
	}
	return storeVector;
    }


    protected ConstraintsConjunction allDiff(Set s) {
	if(s.equ == null) {
	    Vector v = s.toVector();
	    ConstraintsConjunction storeVector = new ConstraintsConjunction();
	    for(int i = 0; i < v.size()-1; i++) {
		for(int j = i+1; j < v.size(); j++) {
		    Constraint st = new Constraint(v.get(i), Enviroment.neqCode, v.get(j));
		    storeVector.add(st);
		}
	    }
	    return storeVector;
	}
	else return allDiff(s.equ);
    }


    protected ConstraintsConjunction allDiff(Lvar l) {  
	if(l.equ == null){
	    if(l.val instanceof Set) return allDiff((Set)l.val);
	    else if(l.val instanceof Lst) return allDiff((Lst)l.val);
	    else throw new NotSetException();
	}
	else return allDiff(l.equ);
    }
 

    /* Metodi per l'annidamento di forall. il vincolo st 
       viene inserito per ogni variabile Yi (della lista Y) 
       appartenente all'insieme  Si (della lista S). */
    public Constraint forall(Lst Y, Lst S, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, st); 
	add(s);
	return s;
    }


    public Constraint forall(Lvar[] Y, Set[] S, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, new Lst(Y), new Lst(S), st); 
	add(s);
	return s;
    }


    public Constraint forall(Lst Y, Lst S, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, vst); 
	add(s);
	return s;
    }


    public Constraint forall(Lvar[] Y, Set[] S, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, new Lst(Y), new Lst(S), vst); 
	add(s);
	return s;
    }

 
    public Constraint forall(Lst Y, Lst S, Lvar l, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, new Lst(l), st); 
	add(s);
	return s;
    }
    

    public Constraint forall(Lst Y, Lst S, Lst L, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, L, st); 
	add(s);
	return s;
    }


    public Constraint forall(Lvar[] Y, Set[] S, Lvar[] L, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, new Lst(Y), new Lst(S), new Lst(L), st); 
	add(s);
	return s;
    }


    public Constraint forall(Lst Y, Lst S, Lvar l, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, new Lst(l), vst); 
	add(s);
	return s;
    }
    
    
    public Constraint forall(Lst Y, Lst S, Lst L, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, L, vst); 
	add(s);
	return s;
    }


    public Constraint forall(Lvar[] Y, Set[] S, Lvar[] L, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, new Lst(Y), new Lst(S), new Lst(L), vst); 
	add(s);
	return s;
    }



    /* L'espressione Solver.forall(Y,S,st) introduce nel constraint store il 
       vincolo st per ogni Y appartenente all'insieme S. */
    public void forall(Lvar Y, Set S, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, Lst.empty, st); 
	add(s);          
	return;
    }
 

    /* L'espressione Solver.forall(Y,S,L,st) introduce nel constraint store il 
       vincolo st per ogni Y appartenente all'insieme S, la lista L contiene
       le variabili esistenziali di st. Queste variabilli sono diverse per ogni 
       vincolo introdotto. */
    public void forall(Lvar Y, Set S, Lvar l, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, new Lst(l), st); 
	add(s);
	return;
    }


    public void forall(Lvar Y, Set S, Lvar[] lvar_ar, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, new Lst(lvar_ar), st); 
	add(s);
	return;         
    }


    public void forall(Lvar Y, Set S, Lst L, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, L, st); 
	add(s);
	return;         
    }



    /*  L'espressione Solver.forall(Y,S,vst) introduce nel constraint store i
	vincoli contenuti nel vettore vst per ogni Y appartenente all'insieme S. */        
    public void forall(Lvar Y, Set S, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, Lst.empty, vst); 
	add(s);
	return;
    }


    /* L'espressione Solver.forall(Y,S,L,vst) introduce nel constraint store i 
       vincoli contenuti nel vettore vst per ogni Y appartenente all'insieme S, 
       La lista L contiene le variabili esistenziali di vst. Queste variabilli          
       sono diverse per ogni vincolo introdotto. */
    public void forall(Lvar Y, Set S, Lvar l, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, new Lst(l), vst); 
	add(s);
	return;
    }


    public void forall(Lvar Y, Set S, Lvar[] lvar_ar, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, new Lst(lvar_ar), vst); 
	add(s);
	return;
    }


    public void forall(Lvar Y, Set S, Lst L, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, S, L, vst); 
	add(s);
	return;
    }


    /* L'espressione Solver.forall(Y,I,st) introduce nel constraint store il
       vincolo st per ogni Y appartenente all'intervallo I. */
    public void forall(Lvar Y, Int I, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, I, Lst.empty, st); 
	add(s);
	return;
    }


    /* L'espressione Solver.forall(Y,S,L,st) introduce nel constraint store il
       vincolo st per ogni Y appartenente all'intervallo I.
       La lista L contiene le variabili esistenziali di vst. Queste variabilli          
       sono diverse per ogni vincolo introdotto. */
    public void forall(Lvar Y, Int I, Lvar l, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, I,new Lst(l), st); 
	add(s);
	return;
    }


    public void forall(Lvar Y, Int I,Lvar[] lvar_ar,Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, I, new Lst(lvar_ar), st); 
	add(s);
	return;
    }


    public void forall(Lvar Y, Int I, Lst L, Constraint st) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, I, L, st); 
	add(s);
	return;
    }


    public void forall(Lvar Y, Int I, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, I, Lst.empty, vst); 
	add(s);
	return;
    }


    public void forall(Lvar Y, Int I, Lvar l, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, I, new Lst(l), vst); 
	add(s);
	return;
    }


    public void forall(Lvar Y, Int I, Lvar[] lvar_ar, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, I, new Lst(lvar_ar), vst); 
	add(s);
	return;
    }


    public void forall(Lvar Y, Int I, Lst L, ConstraintsConjunction vst) {
	Constraint s = new Constraint(Enviroment.forallCode, Y, I, L, vst); 
	add(s);
	return;
    }


    /* Il metodo showStore() permette di visualizzare i vincoli memorizzati 
       nello store che non sono ancora in forma risolta. */
    public void showStore() {
	boolean empty_store = true;
	System.out.print("\n\nStore: ");
	for(int i = 0; i < this.store.size(); i++) {
	    Constraint s = (Constraint)store.get(i);
	    if(!s.bool) {
		empty_store = false;
		C.showStoreElem(s);
		System.out.print("  ");
	    }
	}
	if(empty_store == true) System.out.print("empty");
	System.out.println();
	return;
    }


    public void append(Lst L1, Lst L2, Lst L3) {
	if(L1.equ == null && L2.equ == null && L3.equ == null) {
	    if(L1 == Lst.empty || L1.size_private() == 0)add(L3.eq(L2));
	    else {
		Lvar X = new Lvar();
		Lst L1new = L1.ext1(X);
		Lst L3new = new Lst();
		append(L1new,L2,L3new);
		add(L3new.ins1(X).eq(L3));
	    }
	    return;  
	}
	else {
	    if(L1.equ != null) { 
		append(L1.equ,L2,L3);
		return;
	    }
	    else if(L2.equ != null) {
		append(L1,L2.equ,L3);
		return;
	    }
	    else {
		append(L1,L2,L3.equ);
		return;
	    }
	}
    }

    /* Rimuove tutti i vincoli memorizzati nello store */
    public void clearStore() {
	this.store.removeAllElements();
	return;
    }
    
    
    // return the number of choice point stored in the alternatives stack
    public int numberOfAlternatives() {
	return this.B.getNumOfAlternative();   
    }


    /* Restituisce il tempo di computazione */
    public static double getTime() {
	Date d1 = new Date();
	return (double)(d1.getTime()-date.getTime())/1000;
    }


    public static void printTime() {
	System.out.print("\n\ntime = " + getTime() + " sec");
	return;
    }
        
}

