/*
 * VarState.java
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;

import java.util.*;

/**
 * The <code>VarState</code> class keep track of the uninitialized vars, used 
 * during choice point creation.
 *
 * @author  Elio Panegai
 * @author  Elisabetta Poleo
 * @version 1.3.0
 */
public class VarState implements java.io.Serializable {

    
    /**
     * References to uninitialized logical variables.
     */
    Vector lvarNotInit; 
    
    /**
     * References to uninitialized sets.
     */
    Vector setNotInit;  
    
    /**
     * References to uninitialized lists.
     */
    Vector lstNotInit;      

    /**
     * References to uninitialized Multiset.
     */    
    Vector multisetNotInit; 
    
    /**
     * Constructs a new empty VarState.
     */
    public VarState() {
	this.lvarNotInit     = null;
	this.setNotInit      = null;
	this.lstNotInit      = null;
        this.multisetNotInit = null; 
    }

    
    /**
     * Construct a new instance of VarState containing the specified variables.
     * @param lvars a vector of all Lvar objects not initialized involved in the resolution.
     * @param sets a vector of all Set objects not initialized involved in the resolution.
     * @param lists a vector of all Lst objects not initialized involved in the resolution.
     */
    public VarState(Vector lvars, Vector sets, Vector lists, Vector multisets) {
	this.lvarNotInit = lvars;
	this.setNotInit  = sets;
	this.lstNotInit  = lists;
        this.multisetNotInit = multisets;
    }

    
    /**
     * Clone (bit a bit) the current object.
     *
     * @return a clone of the object.
     */
    protected VarState cloneVarState() {
	VarState vs = new VarState();        
	vs.lvarNotInit = (Vector)((Vector)this.lvarNotInit).clone();
	vs.setNotInit  = (Vector)((Vector)this.setNotInit).clone();
	vs.lstNotInit  = (Vector)((Vector)this.lstNotInit).clone();
        vs.multisetNotInit = (Vector)((Vector)this.multisetNotInit).clone();
	return vs;
    }    
    
}
