//--------------------------------------------------
// JoinSelectivityEstimator.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;

public class JoinSelectivityEstimator implements java.io.Serializable
{
	
	//------------------------------------------------------------------------------------------------------
	
	SimpleOperatorInputSpecification input1; //input stream
	boolean isPartitionBasedEstimation1;     //boolean if it's an overall stream statistics or partition-based									
	int joinAttributeIdx1;                   //join attribute of the current stream
	ArrayList input1PartitionVal;
	SimpleOperatorInputSpecification input2; 
	boolean isPartitionBasedEstimation2;
	int joinAttributeIdx2;
	ArrayList input2PartitionVal;
	Hashtable hashTupleAttrTitles;
	
	//cached sample tuples
	ArrayList tuples = new ArrayList();

	//------------------------------------------------------------------------------------------------------
	
	public JoinSelectivityEstimator(SimpleOperatorInputSpecification input1Val, //input stream
									boolean isPartitionBasedEstimation1Val,     //boolean if it's an overall stream statistics or partition-based									
									int joinAttributeIdx1Val,                   //join attribute of the current stream
									ArrayList input1PartitionValues,
									SimpleOperatorInputSpecification input2Val, 
									boolean isPartitionBasedEstimation2Val,
									int joinAttributeIdx2Val,
									ArrayList input2PartitionValues,
									Hashtable hashTupleAttrTitlesVal)
	{
		input1 = input1Val; 
		isPartitionBasedEstimation1 = isPartitionBasedEstimation1Val;									
		joinAttributeIdx1 = joinAttributeIdx1Val;
		input1PartitionVal = input1PartitionValues;
		input2 = input2Val; 
		isPartitionBasedEstimation2 = isPartitionBasedEstimation2Val;
		joinAttributeIdx2 = joinAttributeIdx2Val;
		input2PartitionVal = input2PartitionValues;
		hashTupleAttrTitles = hashTupleAttrTitlesVal;
	}
	
	//------------------------------------------------------------------------------------------------------
	
	public String getDescriptor()
	{
        String NEW_LINE = System.getProperty("line.separator");
		String str = "";		
		str += "("+input1.getId()+"."+joinAttributeIdx1 + "-->" + ((isPartitionBasedEstimation1 == true) ? "subset-stats " : "overall-stats")+", ";
		str += input2.getId()+"."+joinAttributeIdx2 + "-->" + ((isPartitionBasedEstimation2 == true) ? "subset-stats " : "overall-stats") + ")";
		return str;
	}
	
	//------------------------------------------------------------------------------------------------------
	
	public double getSelectivity()
	{
		double selectivity = 0;
		double joinCardinality = 0;
		
		try
		{
			//read in statistics for the 1st input
			ArrayList inputTuples1 = readInInputSample(input1.getStatisticsFileName());
			ArrayList inputTuples2 = readInInputSample(input2.getStatisticsFileName());
								
			if ((isPartitionBasedEstimation1) && (isPartitionBasedEstimation2))
			{
				System.out.println("RIMMA FIX THIS");
			}
			else if (isPartitionBasedEstimation1)
			{
				//-----------------------------------------------
				// PARTITION - STREAM-BASED SELECTIVITY
				//-----------------------------------------------
				double currValueJoinCardinality = 0;
				
				for(int i = 0; i < input1PartitionVal.size(); i++)
				{
					String [] tokensOuter = (String [])input1PartitionVal.get(i);				
					
					for(int j = 0; j < inputTuples2.size(); j++)
					{
						String innerTupleStr = (String)inputTuples2.get(j);
						String [] tokensInner = innerTupleStr.split("\\|");
						
						if (tokensOuter[/*transl_attr_idx*/joinAttributeIdx1].equalsIgnoreCase(tokensInner[joinAttributeIdx2])) 
						{
							currValueJoinCardinality++;
						}
					}
				}	
				
				if (currValueJoinCardinality != 0)
				{
					joinCardinality += currValueJoinCardinality/(inputTuples1.size() + inputTuples2.size());
				}
			}
			else if (isPartitionBasedEstimation2)
			{										
				//-----------------------------------------------
				// PARTITION - STREAM-BASED SELECTIVITY
				//-----------------------------------------------
				double currValueJoinCardinality = 0;
				
				for(int i = 0; i < input2PartitionVal.size(); i++)
				{				
					String [] tokensOuter = (String [])input2PartitionVal.get(i);
				
					String attr_name = input2.getAttrNameByIndex(joinAttributeIdx2);	
					String str_transl_attr_idx = (String)hashTupleAttrTitles.get(attr_name);									
					
					for(int j = 0; j < inputTuples1.size(); j++)
					{
						String innerTupleStr = (String)inputTuples1.get(j);
						String [] tokensInner = innerTupleStr.split("\\|");
												
						if (tokensOuter[/*transl_attr_idx*/joinAttributeIdx2].equalsIgnoreCase(tokensInner[joinAttributeIdx1])) 
						{
							currValueJoinCardinality++;
						}
					}
				}	
				
				if (currValueJoinCardinality != 0)
				{
					joinCardinality += currValueJoinCardinality/(inputTuples1.size() + inputTuples2.size());
				}
			}
			else
			{
				//-----------------------------------------------
				// OVERALL STREAM-BASED SELECTIVITY
				//-----------------------------------------------
			
				double currValueJoinCardinality = 0;
				
				for(int i = 0; i < inputTuples1.size(); i++)
				{				
					String outerTupleStr = (String)inputTuples1.get(i);
					String [] tokensOuter = outerTupleStr.split("\\|");
									
					for(int j = 0; j < inputTuples2.size(); j++)
					{
						String innerTupleStr = (String)inputTuples2.get(j);
						String [] tokensInner = innerTupleStr.split("\\|");
																							
						if (tokensOuter[joinAttributeIdx1].equalsIgnoreCase(tokensInner[joinAttributeIdx2])) 
						{
							currValueJoinCardinality++;
						}
					}//end for (itarating inner tuples)
				}//end for	(iterating outer tuples)
				
				if (currValueJoinCardinality != 0)
				{
					joinCardinality += currValueJoinCardinality /(inputTuples1.size() + inputTuples2.size());
				}
			}//end else
		}
		catch(Exception ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
		}
		selectivity = joinCardinality;
		return selectivity;
	}	
	
	//------------------------------------------------------------------------------------------------------
	
	private ArrayList readInInputSample(String statisticsFile)
	{
		//if input file was not previously read, read it and cache it
		//otherwise, read the input from the cache
		
		if (tuples.size() == 0)
		{
			FileInputStream fin;		
			try
			{
				FileReader input = new FileReader(statisticsFile);
				/* Filter FileReader through a Buffered read to read a line at a time */
				BufferedReader bufRead = new BufferedReader(input);
				String line;    // String that holds current file line
				
				// Read first line
				line = bufRead.readLine();
				
				if (line != null)
				{
					if (!line.equalsIgnoreCase(""))
					{
						tuples.add(line);
					}
				}
	
			    //Read through file one line at time. Traverse the DT
			    while (line != null)
			    {
				    line = bufRead.readLine();
				    
				    if (line == null) continue;
				    if (line.startsWith("//")) continue;
				    if (line.equals("")) continue;
				    
				    tuples.add(line);
			    }		    
			    // Close our input stream
			    bufRead.close();
	
			}
			// Catches any error conditions
			catch (IOException e)
			{
				System.err.println ("!!! Unable to read from file [" + statisticsFile + "]!!!");
			}
		}
		return tuples;
	}
	
	//------------------------------------------------------------------------------------------------------
	
}
