//--------------------------------------------------
//LocalQueryMesh.java
//Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;

import java.util.ArrayList;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.DecisionTreeID3;

public class LocalQueryMesh implements java.io.Serializable
{	
	//----------------------------------------------------------------------------------------------------
	
	String streamId;
	private Hashtable hashClassifierRoutesMapping;
	private Object classifier;
	private ArrayList<RouteSignature> routes; 
	private RouteSignature defaultRoute;
	double cost;
	
	//----------------------------------------------------------------------------------------------------
	
	public LocalQueryMesh(String streamIdVal)
	{
		streamId = streamIdVal;
		hashClassifierRoutesMapping = new Hashtable();
		classifier = new ArrayList();
		routes = new ArrayList();
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public LocalQueryMesh(String streamIdVal, ArrayList classifierVal, ArrayList routesVal)
	{
		streamId = streamIdVal;
		hashClassifierRoutesMapping = new Hashtable();
		classifier = (ArrayList)classifierVal;
		routes = (ArrayList)routesVal;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public String getStreamId()
	{
		return streamId;		
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public Object getClassifier()
	{
		return classifier;		
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public Hashtable getHashClassifierRoutesMapping()
	{
		return hashClassifierRoutesMapping;		
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public ArrayList getRoutes()
	{
		return routes;		
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public void setDTClassifier(DecisionTreeID3 dt)
	{
		classifier = dt;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public void setHashClassifierRoutesMapping(Hashtable ht)
	{
		hashClassifierRoutesMapping = ht;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public void setRoutes(ArrayList executionRoutes)
	{
		routes.clear();
		routes.addAll(executionRoutes);
	}

	//----------------------------------------------------------------------------------------------------
	
	public RouteSignature getDefaultRoute()
	{
		return defaultRoute;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public void setDefaultRoute(RouteSignature defaultRouteSignature)
	{
		defaultRoute = defaultRouteSignature;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public void clearRoutes()
	{
		routes.clear();
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public double getCost()
	{
		return cost;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public void setCost(double costVal)
	{
		cost = costVal;
	}
	
	//----------------------------------------------------------------------------------------------------
	  
	public String toString() 
	{	
		StringBuilder result = new StringBuilder();	 
		String NEW_LINE = System.getProperty("line.separator");
 
		result.append(NEW_LINE + "LocalQueryMesh:"+ NEW_LINE);
		result.append("streamId: " + streamId + NEW_LINE);	       
		result.append("defaultRoute: " + defaultRoute + NEW_LINE);
		for (int i = 0; i < routes.size(); i++)
		{
			RouteSignature route = (RouteSignature)routes.get(i);
			result.append("route: " + route.ruster +": " + route.route + NEW_LINE);
		}
		return result.toString();    	    
	}
	
	//----------------------------------------------------------------------------------------------------
}
