//--------------------------------------------------
// QueryMeshOptimizer.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;


import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.DecisionTreeID3;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.translation.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.translation.ClassifierInputSignature.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.MJoinDynamicAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes.JoinGraphImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.Stream;

public class QueryMeshOptimizer implements java.io.Serializable
{
	//-----------------------------------------------------------------------------
	
	private static String WORKING_DIRECTORY = "C:\\QueryMesh\\tmp\\";
	private static String STATISTICS_OUTPUT_FILE = "C:\\QueryMesh\\stats\\stats.txt";
	private static int nextChildQueueIdVal = 0;
	private static String globalChildOpIdVal = "0";
	private static QM_TYPE CURRENT_QM_TYPE = QM_TYPE.STATIC; //default = static QM
	
	public static Hashtable hashNumIterationsPerStream = new Hashtable(); //number of query mesh configs that were enumerated by the optimizer per stream

	//-----------------------------------------------------------------------------
	
    public enum QM_SEARCH_START_SOLUTION 
    {
    	EXTREME_1_ROUTE,  //QM where all tuples are assigned one route
    	//EXTREME_N_ROUTES, //QM where each tuple has an individual route //rimma: "route-driven" case takes care of this
    	RANDOM,           //randomly chosen QM
    	CONTENT_DRIVEN,   //tuples, are first partitioned based on similarity of tuple content, then QM is formed
    	ROUTE_DRIVEN     //routes are computed, then tuples are "grouped-by" similar route
    }
	
	//-----------------------------------------------------------------------------
    
    public enum QM_SEARCH_STRATEGY 
    {
    	ITERATIVE_IMPROVEMENT,
    	SIMULATED_ANNEALING
    }
    
	//-----------------------------------------------------------------------------
    
	//------------------------------------------------------------------------
	// STQM: QM can be either 'static' or 'adaptive'
	//------------------------------------------------------------------------
    public enum QM_TYPE 
    {
    	STATIC,
    	ADAPTIVE
    }
   
	//-----------------------------------------------------------------------------
    
	//-----------------------------------------------------------------------------
	// PARSED FROM XML
	//-----------------------------------------------------------------------------
	static List xmlOperators;
	static Hashtable xmlInputs;
	
	//STQM: training set collectors for adaptive QM
	static Hashtable xmlTrainingTupleCollectorInputs;
	Document dom;	

	//-----------------------------------------------------------------------------
	
	private SimpleQueryPlanSpecification queryPlan;

	//-----------------------------------------------------------------------------
	
	public QueryMeshOptimizer(String xmlInputFile)
	{
		xmlInputs = new Hashtable();
		xmlOperators = new ArrayList();
		xmlTrainingTupleCollectorInputs = new Hashtable();
		
		//--------------------------------------------------------------------
		// PROCESS XML
		//--------------------------------------------------------------------		
		//parse the xml file and get the dom object
		System.out.println("Parsing optimizer input file...");
		parseXmlFile(xmlInputFile);
		
		//parse the documents
		System.out.println("Extracting xml data from DOM...");
		parseDocument();
		
		//Iterate through the list and print the data
		printData();
		//--------------------------------------------------------------------
		// DONE WITH XML PROCESSING
		//--------------------------------------------------------------------		
	}
	
	//-----------------------------------------------------------------------------
		
	private void printData()
	{		
		System.out.println("No of xml operators [" + xmlOperators.size() + "]");
		System.out.println("\n---------------------------OPERATOR DETAILS-----------------------");
		
		Iterator it = xmlOperators.iterator();
		while(it.hasNext()) 
		{
			SimpleOperatorSpecification currOpt = (SimpleOperatorSpecification)it.next();
			System.out.println(currOpt.toString());
		}
		System.out.println("---------------------------END OPERATOR DETAILS-----------------------");
		
		System.out.println("No of xml inputs '" + xmlInputs.size() + "'.");
		System.out.println("\n----------------------------------INPUT DETAILS-----------------------");
		Enumeration e = xmlInputs.elements(); 
		while (e.hasMoreElements()) 
		{  
			Object x = e.nextElement();  // do whatever you need with x 
			System.out.println(x.toString());
		}
		System.out.println("----------------------------------END INPUT DETAILS-----------------------");
	}
	
	//-----------------------------------------------------------------------------
	
	private void parseXmlFile(String xmlInputFile)
	{
		//get the factory
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		
		try 
		{			
			//Using factory get an instance of document builder
			DocumentBuilder db = dbf.newDocumentBuilder();
			
			//parse using builder to get DOM representation of the XML file
			dom = db.parse(xmlInputFile);
		}
		catch(ParserConfigurationException pce) 
		{
			pce.printStackTrace();
		}
		catch(SAXException se) 
		{
			se.printStackTrace();
		}
		catch(IOException ioe) 
		{
			ioe.printStackTrace();
		}
	}

	//-----------------------------------------------------------------------------
	
	private void parseDocument()
	{
		//get the root elememt
		Element docEle = dom.getDocumentElement();

		//----------------------------------------------------------------------
		// STQM: TYPE OF QUERY MESH (was added after adaptive functionality has been added)
		//----------------------------------------------------------------------		
		//get <QueryMesh type="adaptive"></QueryMesh> node
		NodeList nQM = docEle.getElementsByTagName("QueryMesh");
		if(nQM != null && nQM.getLength() == 1) 
		{
			//get the QueryMesh element
			Element el = (Element)nQM.item(0);
				
			//get the QueryMesh type
			String qmType = el.getAttribute("type");
			
			if (qmType.equalsIgnoreCase("adaptive"))
			{
				CURRENT_QM_TYPE = QM_TYPE.ADAPTIVE;
			}
			else
			{
				CURRENT_QM_TYPE = QM_TYPE.STATIC;
			}
		}
				
		//----------------------------------------------------------------------
		// INPUTS
		//----------------------------------------------------------------------		
		//get a nodelist of <Input> elements
		NodeList nl = docEle.getElementsByTagName("Input");
		if(nl != null && nl.getLength() > 0) 
		{
			for(int i = 0 ; i < nl.getLength();i++) 
			{				
				//get the employee element
				Element el = (Element)nl.item(i);
				
				//get the Input object
				SimpleOperatorInputSpecification e = getXmlOperatorInput(el);
				
				TrainingSetCollectorOperatorInputSpecification et = getXmlTrainingSetCollectorOperatorInput(el);
				
				//add input to list
				xmlInputs.put(e.getId(), e);

				//add training tuple collector to list
				xmlTrainingTupleCollectorInputs.put(e.getId(), et);
			}
		}
		
		//----------------------------------------------------------------------
		// OPERATORS
		//----------------------------------------------------------------------
		//get a nodelist of <Operator> elements
		nl = docEle.getElementsByTagName("Operator");
		if(nl != null && nl.getLength() > 0) 
		{
			for(int i = 0 ; i < nl.getLength();i++) 
			{				
				//get the employee element
				Element el = (Element)nl.item(i);
				
				//get the Operator object
				SimpleOperatorSpecification e = getXmlOperator(el);
				
				//add it to list
				xmlOperators.add(e);
			}
		}
	}

	//-----------------------------------------------------------------------------
	
	private SimpleOperatorSpecification getXmlOperator(Element opEl) 
	{		
		String id = opEl.getAttribute("id");
		String type = opEl.getAttribute("type");
		
		ArrayList inputs = new ArrayList();
		Hashtable hashJoinAttributes = new Hashtable();

		//-------------------------------------------------------
		// retrieve operator inputs
		//-------------------------------------------------------
		//get a nodelist of <Input> elements
		NodeList nl = opEl.getElementsByTagName("OpInput");
		if(nl != null && nl.getLength() > 0) 
		{
			for(int i = 0 ; i < nl.getLength();i++) 
			{				
				//get the opInput element
				Element opInput = (Element)nl.item(i);
				
				//get id and attribute of the Input object
				String opInputId = opInput.getAttribute("id");
				String opAttributeIdx = opInput.getAttribute("attr_idx");
				
				//retrieve input object from the hashtable
				SimpleOperatorInputSpecification opInputObj = (SimpleOperatorInputSpecification)xmlInputs.get(opInputId);
				
				assert opInputObj != null;
				
				//add input id to the input list
				inputs.add(opInputId);

				//add attr index
				hashJoinAttributes.put(opInputId, opAttributeIdx);
			}
		}
		
		//Create a new SimpleOperatorSpecification with the value read from the xml nodes
		SimpleOperatorSpecification e = new SimpleOperatorSpecification(id,type,inputs,hashJoinAttributes);
		
		return e;
	}
	
	//-----------------------------------------------------------------------------
	
	private String getTextValue(Element ele, String tagName) 
	{
		String textVal = null;
		NodeList nl = ele.getElementsByTagName(tagName);
		if(nl != null && nl.getLength() > 0) 
		{
			Element el = (Element)nl.item(0);
			textVal = el.getFirstChild().getNodeValue();
		}
		return textVal;
	}
	
	//-----------------------------------------------------------------------------

	private int getIntValue(Element ele, String tagName) 
	{
		//in production application you would catch the exception
		return Integer.parseInt(getTextValue(ele,tagName));
	}
	
	//-----------------------------------------------------------------------------
	
	private SimpleOperatorInputSpecification getXmlOperatorInput(Element opEl) 
	{				
		String id = opEl.getAttribute("id");
		String name = opEl.getAttribute("name");
		String trainingTuplesFile = opEl.getAttribute("training_tuples_file");
		String statsFile = opEl.getAttribute("statistics_file");
		
		//parse the schema of the input stream
		NodeList elNodeSchema = opEl.getElementsByTagName("Schema");
		
		assert elNodeSchema.getLength() == 1;		
		Element elSchema = (Element)elNodeSchema.item(0);

		Hashtable schemaAttributes = new Hashtable();
		//get a nodelist of <attribute> elements
		NodeList nl = elSchema.getElementsByTagName("attribute");
		if(nl != null && nl.getLength() > 0) 
		{
			for(int i = 0 ; i < nl.getLength();i++) 
			{				
				//get the employee element
				Element el = (Element)nl.item(i);
				
				String attr_name = el.getAttribute("name");
				
				schemaAttributes.put(attr_name, ""+i); //store attribute_name, attribute_index in the schema in the hashtable
			}
		}
		
		//Create a new Operator Input specification with the value read from the xml nodes
		SimpleOperatorInputSpecification e = new SimpleOperatorInputSpecification(id, name, trainingTuplesFile, statsFile, schemaAttributes);
		
		return e;
	}

	//-----------------------------------------------------------------------------
		
	private TrainingSetCollectorOperatorInputSpecification getXmlTrainingSetCollectorOperatorInput(Element opEl) 
	{			
		/*
		 * <Input id="8" name="Stream8" statistics_file="C:\\Rimma\\Research\\QueryMesh\\statistics_samples\\SAMPLE_DS5_STREAM8.txt" training_tuples_file="C:\\Rimma\\Research\\QueryMesh\\training_sets\\TRAINING_SET_DS5_STREAM8.txt">
				...
				<TrainingSetCollection>
					<attribute name="val0" type="int" use="true" is_target="true"/>
					<attribute name="val1" type="int" use="true" is_target="false"/>
					<attribute name="val2" type="int" use="true" is_target="false"/>
					<attribute name="val3" type="int" use="true" is_target="false"/>
				</TrainingSetCollection>					
			</Input>
		 * */
		
		String id = opEl.getAttribute("id");
		String name = opEl.getAttribute("name");
		
		//parse the schema of the input stream
		NodeList elNodeSchema = opEl.getElementsByTagName("Schema");
		
		assert elNodeSchema.getLength() == 1;		
		Element elSchema = (Element)elNodeSchema.item(0);

		ArrayList trainingSetAttributes = new ArrayList();
		//get a nodelist of <attribute> elements
		NodeList nl = elSchema.getElementsByTagName("attribute");
		if(nl != null && nl.getLength() > 0) 
		{
			for(int i = 0 ; i < nl.getLength();i++) 
			{				
				//get the employee element
				Element el = (Element)nl.item(i);
				
				String attr_name = el.getAttribute("name");
				String attr_type = el.getAttribute("type");
				String attr_b_use = el.getAttribute("use");
				String attr_b_istarget = el.getAttribute("is_target");
				
				AttributeProperties ap = new AttributeProperties(attr_name,attr_type,attr_b_use,attr_b_istarget);
				trainingSetAttributes.add(ap); //store attribute_name, attribute_index in the schema in the hashtable
			}
		}
		
		TrainingSetCollectorOperatorInputSpecification e = new TrainingSetCollectorOperatorInputSpecification(id, name, trainingSetAttributes);
		
		return e;
	}
	
	//-----------------------------------------------------------------------------
	
	public Hashtable GetAllInputs()
	{	
		return xmlInputs;
	}	
	
	//-----------------------------------------------------------------------------
	   
	public SimpleQueryPlanSpecification GetQueryPlanSpecification()
	{	
		return queryPlan;	 
	}
   
	//-----------------------------------------------------------------------------
	
	public static void main(String args[])
	{
		//default start solution for QM search
		QM_SEARCH_START_SOLUTION startSolution = QM_SEARCH_START_SOLUTION.CONTENT_DRIVEN;
		
		//indicates if user inputs the type of start solution
		boolean isUserSpecifiedStartSolution = false;
		
		System.out.println("Executing Query Mesh optimizer...");
		try
		{
			double timeStart = System.currentTimeMillis();

			//--------------------------------------------------------------------
			// READ IN QUERY MESH OPTIMIZER INPUT FILE
			// BASED ON WHICH QUERY MESH WILL BE CONSTRUCTED
			//--------------------------------------------------------------------
			
			int num = args.length;
			if (num == 7) 
			{
				isUserSpecifiedStartSolution = true;
				String inputStartSolution = args[6];
				if (inputStartSolution.equalsIgnoreCase("random"))
				{
					startSolution = QM_SEARCH_START_SOLUTION.RANDOM;
				}
				else if (inputStartSolution.equalsIgnoreCase("extreme1"))
				{
					startSolution = QM_SEARCH_START_SOLUTION.EXTREME_1_ROUTE;
				}
				/*else if (inputStartSolution.equalsIgnoreCase("extremeN"))
				{
					startSolution = QM_SEARCH_START_SOLUTION.EXTREME_N_ROUTES;
				}*/
				else if (inputStartSolution.equalsIgnoreCase("content"))
				{
					startSolution = QM_SEARCH_START_SOLUTION.CONTENT_DRIVEN;
				}
				else if (inputStartSolution.equalsIgnoreCase("route"))
				{
					startSolution = QM_SEARCH_START_SOLUTION.ROUTE_DRIVEN;
				}
				
			}

			//if 6 parameters were specified, the last one is the statistics file
			if (num == 6)
			{
				STATISTICS_OUTPUT_FILE = args[5];
			}
			
			if (num < 5) 
			{
				System.out.println("USAGE:");
	      		System.out.println("\tYou need to specify the name query mesh optimizer input file at the command line " );
	      		System.out.println("\tYou need to specify the full enumeration (m) parameter at the command line " );	      		
	      		System.out.println("\tYou need to specify the search heuristic (SA, II) at the command line " );
	      		System.out.println("\tYou need to specify the location of the QM dump file (file where QM content will be put) " );
	      		System.out.println("\tYou need to specify the location of the QM query plan xml file " );
	      		System.out.println("");
	      		System.out.println("Optional:");
	      		System.out.println("\tYou can specify a start solution 'hint' to QM optimizer, e.g:" );
	      		System.out.println("\t\t'random' - random start" );
	      		System.out.println("\t\t'extreme1' - one route for all" );
	      		System.out.println("\t\t'extremeN - each has individual route" );
	      		System.out.println("\t\t'content' - content-driven" );
	      		System.out.println("\t\t'route' - route-driven" );
	      		return;
			}
			
			
			String inputfileName = args[0];
			String inputSearchHeuristic = args[2];
			QM_SEARCH_STRATEGY searchHeuristic = QM_SEARCH_STRATEGY.ITERATIVE_IMPROVEMENT;// default initialization
			
			if (inputSearchHeuristic.toLowerCase().equals("sa"))
			{
				searchHeuristic = QM_SEARCH_STRATEGY.SIMULATED_ANNEALING;
				System.out.println("Heuristic: simulated annealing");
			}
			else if (inputSearchHeuristic.toLowerCase().equals("ii"))
			{
				searchHeuristic = QM_SEARCH_STRATEGY.ITERATIVE_IMPROVEMENT;
				System.out.println("Heuristic: iterated improvement");
			}
					
			//--------------------------------------------------------------------
			// INITIALIZE QUERY MESH OPTIMIZER BASED ON THE INPUT FILE
			//--------------------------------------------------------------------	
			int m = Integer.parseInt(args[1]); //full enumeration subset size
			System.out.println("Upperbound of subset size for full enumeration: " + m);
			
			QueryMeshOptimizer qmOptimizer = new QueryMeshOptimizer(inputfileName);	
					
						
			//--------------------------------------------------------------------
			// PRODUCE GLOBAL QUERY MESH
			//--------------------------------------------------------------------	
			GlobalQueryMesh bestQM = qmOptimizer.produceGlobalQueryMesh(searchHeuristic, 
																		m, 
																		startSolution);			
			//--------------------------------------------------------------------
			// OPTIMIZER IS DONE
			//--------------------------------------------------------------------	
			
			double timeEnd = System.currentTimeMillis();	
			double timeDuration = timeEnd - timeStart;
			double timeDurationSec = timeDuration / 1000;
			
			System.out.println("\n\n\n*****************************************************************");
			System.out.println("*****************************************************************");
			System.out.println("Optimizer is done. Query mesh is produced.");
			System.out.println("Time duration [" + timeDuration + "] msecs, or [" + timeDurationSec + "] secs");
			
			//write the search time to the stats
			Helper.StatsWriteLineToFile(STATISTICS_OUTPUT_FILE, GetCurrentDateTime() +" --> "+ Double.toString(timeDuration) + "\n");
			System.out.println("Stored time in stats file");
			
			//--------------------------------------------------------------------
			// DUMP QM (FOUND BY THE OPTIMIZER) TO THE FILE
			// QM EXECUTOR WILL INSTANTIAT THE RUN-TIME QM BASED
			// ON THIS FILE "DUMP"
			//--------------------------------------------------------------------	
			//String qmDumpFile = args[3];
			//String strQM = getGlobalQMString(bestQM);
			//WriteToDumpFile(qmDumpFile, strQM);

			System.out.print("Preparing data for config file...");
			
			//--------------------------------------------------------------------
			// CONVERT QM CONTENT (FOUND BY THE OPTIMIZER) TO THE PLAN XML PLAN
			//--------------------------------------------------------------------	
			
			String outputXmlPlanDirectory = args[4];
						
			DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
	        java.util.Date date = new java.util.Date();
	        String datetime = dateFormat.format(date);
	        datetime = datetime.replaceAll(":", ".");
	        datetime = datetime.replaceAll(" ", ".");
	        datetime = datetime.replaceAll("/", ".");
	        //date.getYear()+"." +date.getMonth()+"." + date.getDate() + "." + date.getHours() + "." + date.getMinutes() + "." + date.getSeconds();
	        String qmFileTimestamp = datetime;
	        
	        String outputXmlPlan = outputXmlPlanDirectory + "qm_query_plan_" + qmFileTimestamp + ".xml";
	        
			//write the contents of just computed QM to xml file
			qmOptimizer.MaterializeQMSolutionIntoXmlFile(bestQM, outputXmlPlan);
			
			System.out.println("Done.");
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		
		System.out.println("Query Mesh optimizer is all done.");
		
		GetOptimizerStatistics();
	}
	
	//-----------------------------------------------------------------------------
	
	public void MaterializeQMSolutionIntoXmlFile(GlobalQueryMesh bestQM, String outputXmlPlan)
	{
		//--------------------------------------------------------------------
		// COMPUTE DISTINCT STEMS
		//--------------------------------------------------------------------
		double timePrepareStarted = System.currentTimeMillis();
		
		ArrayList allLogicalToPhysicalOpCombinations = new ArrayList();
		Hashtable hashDistinctStems = new Hashtable();
		ArrayList operatorsDetails = new ArrayList();
		int childInputCount = 1;
		for(int k = 0; k < xmlOperators.size(); k++)
		{
			SimpleOperatorSpecification opSpec = (SimpleOperatorSpecification)xmlOperators.get(k);
			String input1ID = (String)opSpec.getInputs().get(0);
			String input2ID = (String)opSpec.getInputs().get(1);
			
			String probeAttrIdxInput1 = ""+opSpec.getJoinAttributeIdxByInputId(input1ID);
			String probeAttrIdxInput2 = ""+opSpec.getJoinAttributeIdxByInputId(input2ID);
			
			String stem1ProbeKey = input1ID + "," +probeAttrIdxInput1 + "|" + input2ID + "," +probeAttrIdxInput2;
			String stem2ProbeKey = input2ID + "," +probeAttrIdxInput2 + "|" + input1ID + "," +probeAttrIdxInput1;
			//-----------------------------------------------------------------------------
			// ADD SIGNATURES FOR BOTH STEMS
			//-----------------------------------------------------------------------------
			// EXAMPLE
			//STEM Operator- store A.a1 probe B.b3-
			//      DataSTREAMId="1"
		    //      NumSTeMIndex="1"
			//
			//      ProbeSTREAMId="5"
			//      NumTupleIndex="3"
			//
			// STEM Operator - store B.b3 probe A.a1
			//      DataSTREAMId="5"
			//      NumSTeMIndex="3"
			//
			//      ProbeSTREAMId="1"
			//      NumTupleIndex="1"
			
			//--------------------------------------------------------------------
			// "GLOBAL CHILD" FOR ALL STEMS --> IS THE ONLINE CLASSIFIER OPERATOR
			//--------------------------------------------------------------------
			String opIdVal1 = globalChildOpIdVal;    //default
			String queueIdVal1 = ""+(nextChildQueueIdVal); //default
			nextChildQueueIdVal++;
			ChildOperatorSignature globalChildSignature1 = new ChildOperatorSignature(opIdVal1, queueIdVal1);

			ArrayList childrenOps1 = new ArrayList();
			childrenOps1.add(globalChildSignature1);
			//---------------------------------------------------------------------------------
			//STEM-1
			//---------------------------------------------------------------------------------
			if (!hashDistinctStems.containsKey(stem1ProbeKey))
			{		
				hashDistinctStems.put(stem1ProbeKey, input1ID+"|"+childInputCount);
				String stem1Id = ""+childInputCount;
				STEMOperatorSignature stemOp = new STEMOperatorSignature(stem1Id,            //opIdVal
																		 input1ID,           //streamIdVal
																		 input1ID,           //probeStreamIdVal
																		 probeAttrIdxInput1, //numTupleIndexVal
																		 input2ID,           //stemStreamIdVal
																		 probeAttrIdxInput2, //numSTeMIndexVal				
																		 childrenOps1);      //childrenOpsVal
				childInputCount++;
				operatorsDetails.add(stemOp);
				
				//add the logical to physical op combination
				String logicalOpIdVal = opSpec.getId();
				String streamIdVal = input1ID;
				String physicalOpIdVal = stem1Id;
				LogicalOpToPhysicalOpSignature logToPhysCombination = 
					new LogicalOpToPhysicalOpSignature(logicalOpIdVal, streamIdVal, physicalOpIdVal);
				allLogicalToPhysicalOpCombinations.add(logToPhysCombination);
			}					
			//---------------------------------------------------------------------------------
			//STEM-2
			//---------------------------------------------------------------------------------

			String opIdVal2 = globalChildOpIdVal;    //default
			String queueIdVal2 = ""+(nextChildQueueIdVal); //default
			nextChildQueueIdVal++;
			ChildOperatorSignature globalChildSignature2 = new ChildOperatorSignature(opIdVal2, queueIdVal2);

			ArrayList childrenOps2 = new ArrayList();
			childrenOps2.add(globalChildSignature2);
			
			if (!hashDistinctStems.containsKey(stem2ProbeKey))
			{					
				hashDistinctStems.put(stem2ProbeKey, input2ID+"|"+childInputCount);
				String stem2Id = ""+childInputCount;
				STEMOperatorSignature stemOp = new STEMOperatorSignature(stem2Id,
																		 input2ID,
																		 input2ID,
																		 probeAttrIdxInput2,
																		 input1ID,
																		 probeAttrIdxInput1,
																		 childrenOps2);
				childInputCount++;
				operatorsDetails.add(stemOp);
				
				//add the logical to physical op combination
				String logicalOpIdVal = opSpec.getId();
				String streamIdVal = input2ID;
				String physicalOpIdVal = stem2Id;
				LogicalOpToPhysicalOpSignature logToPhysCombination = 
					new LogicalOpToPhysicalOpSignature(logicalOpIdVal, streamIdVal, physicalOpIdVal);
				allLogicalToPhysicalOpCombinations.add(logToPhysCombination);
			}
		}
		
		//--------------------------------------------------------------------
		// ITERATE THROUGH DISTINCT STEMS TO ADD THEM AS CHILDREN OF THE ROOT OP
		//--------------------------------------------------------------------	
		ArrayList rootOpChildren = new ArrayList();    //children of the root operator
		ArrayList classifierParents = new ArrayList(); //parents of the classifier operator
		Enumeration eHashDistinctStems = hashDistinctStems.keys(); 
		
		while (eHashDistinctStems.hasMoreElements()) 
		{				
			String key = (String)eHashDistinctStems.nextElement();
			String opInputIdAndStemId = (String)hashDistinctStems.get(key);
			
			//extract stem id
			char[] charsOpInputIdAndStemId = opInputIdAndStemId.toCharArray();
			int u = 0; 
			while(charsOpInputIdAndStemId[u++] != '|') {}
			
			String childOpIdStr = "";
			for (int p = u; p < charsOpInputIdAndStemId.length; p++)
			{
				childOpIdStr += charsOpInputIdAndStemId[p];
			}
			
			//String [] tokens = opInputIdAndStemId.split("|");
			String childOpId = childOpIdStr;//tokens[tokens.length-1];
							
			ChildOperatorSignature childObj = new ChildOperatorSignature(childOpId, ""+childInputCount);
			rootOpChildren.add(childObj);
			
			ParentOperatorSignature parentObj = new ParentOperatorSignature(childOpId);				
			classifierParents.add(parentObj);
		}
		
		String rootOpId = "" + (hashDistinctStems.size()+1);
		
		
		//--------------------------------------------------------------------
		// PREPARE DATA FOR ONLINE CLASSIFIER CONFIG
		//--------------------------------------------------------------------
		ArrayList localQMSignatures = PrepareDataForOnlineClassifierConfig(bestQM);
		
		//--------------------------------------------------------------------
		// ONE - MORE THING, CONVERT ARRAY MAPPING LOGICAL OP TO PHYSICAL TO
		// A HASHTABLE FOR FASTER LOOKUP
		//--------------------------------------------------------------------
		Hashtable hashAllLogicalToPhysicalOpCombinations = new Hashtable();
		for (int b = 0; b < allLogicalToPhysicalOpCombinations.size();b++)
		{
			LogicalOpToPhysicalOpSignature logToPhysCombination = 
				(LogicalOpToPhysicalOpSignature)allLogicalToPhysicalOpCombinations.get(b);
			String key = logToPhysCombination.GetLogicalOpId()+"|" + logToPhysCombination.GetStreamId(); 
			hashAllLogicalToPhysicalOpCombinations.put(key, logToPhysCombination);
		}
		
		//--------------------------------------------------------------------
		// LAST BUT NOT THE LEAST - CREATE A MAPPING (LOGICAL OP ID + STREAM ID)
		// THIS WILL BE HELPFUL IN ASSIGNING PHYSICAL TO LOGICAL OPERATORS 
		//--------------------------------------------------------------------
		Hashtable hashLogicalOperatorsWithInputsCombinations = new Hashtable();
		Iterator it = xmlOperators.iterator();
		while(it.hasNext()) 
		{
			SimpleOperatorSpecification currOpt = (SimpleOperatorSpecification)it.next();
			String input1 = (String)currOpt.getInputs().get(0);
			String input2 = (String)currOpt.getInputs().get(1);
			ArrayList inputStreams = new ArrayList();
			inputStreams.add(input1);
			inputStreams.add(input2);
			//String insertStr1 = "op:" + currOpt.getId() +"|str:" +  input1;
			//String insertStr2 = "op:" + currOpt.getId() +"|str:" +  input2;
			hashLogicalOperatorsWithInputsCombinations.put(currOpt.getId(), inputStreams);
		}
		System.out.println("Done.");

		//--------------------------------------------------------------------
		// STQM: TRAINING SET COLLECTOR OPERATORS
		//--------------------------------------------------------------------
		ArrayList trainingSetCollectorOperatorsDetails = new ArrayList();
		
		Enumeration eXmlTrainingTupleCollectorInputs = xmlTrainingTupleCollectorInputs.keys(); 
		
		int startIndexTrainingSetCollectorOpIndex = operatorsDetails.size() + 2;
		int startIndexTrainingSetCollectorOpIndexPersistent = startIndexTrainingSetCollectorOpIndex;
			
		while (eXmlTrainingTupleCollectorInputs.hasMoreElements()) 
		{
			String key = (String)eXmlTrainingTupleCollectorInputs.nextElement();
			TrainingSetCollectorOperatorInputSpecification inputObj = (TrainingSetCollectorOperatorInputSpecification)xmlTrainingTupleCollectorInputs.get(key);
			TrainingSetCollectorOperatorSignature physTrainingSetCollectorObj = 
			new TrainingSetCollectorOperatorSignature(Integer.toString(startIndexTrainingSetCollectorOpIndex),
													  "0", //rimma: operator id = "0" is reserved for online classifier operator
													  inputObj.GetInputName(),   //stream name
													  inputObj.GetAttributes()); //stream attributes
			trainingSetCollectorOperatorsDetails.add(physTrainingSetCollectorObj);
			startIndexTrainingSetCollectorOpIndex++;
		}
					
		//--------------------------------------------------------------------
		// CREATE CLASSIFIER OP SIGNATURE
		//--------------------------------------------------------------------
		ArrayList classifierChildren = new ArrayList();	
		if ((xmlTrainingTupleCollectorInputs == null) || (xmlTrainingTupleCollectorInputs.size() == 0))
		{						
			Enumeration eXmlInputs = xmlInputs.keys(); 			
			while (eXmlInputs.hasMoreElements()) 
			{
				String key = (String)eXmlInputs.nextElement();
				SimpleOperatorInputSpecification inputObj = (SimpleOperatorInputSpecification)xmlInputs.get(key);
				ClassifierInputSignature childStream = new ClassifierInputSignature(INPUT_TYPE.STREAM, inputObj.getId(), inputObj.getInputName(), null);
				classifierChildren.add(childStream);
			}
		}
		else
		{
			//--------------------------------------------------------------------
			// STQM: training tuple set collectors preceding input streams before
			// they stream into the Online Classifier operator
			//--------------------------------------------------------------------		
			for(int i = 0; i < trainingSetCollectorOperatorsDetails.size(); i++)
			{
				TrainingSetCollectorOperatorSignature inputObj = (TrainingSetCollectorOperatorSignature)trainingSetCollectorOperatorsDetails.get(i);
				ClassifierInputSignature childStream = new ClassifierInputSignature(INPUT_TYPE.OPERATOR, null, inputObj.GetInputStreamName(), inputObj.GetOpId());
				classifierChildren.add(childStream);
			}
		}
		
		OnlineClassifierOperatorSignature classifierSignature = new OnlineClassifierOperatorSignature(globalChildOpIdVal,													  
												  				""+xmlInputs.size(),
												  				""+rootOpChildren.size(),
												  				classifierParents,
												  				classifierChildren);
		
		//--------------------------------------------------------------------
		// GENERATE THE ACTUAL QUERY MESH - QUERY PLAN
		//--------------------------------------------------------------------		
		double timePrepareEnded = System.currentTimeMillis();
		double timePrepareDuration = timePrepareEnded - timePrepareStarted;
		double timePreparationDurationSec = timePrepareDuration / 1000;
		System.out.println("Preparation duration time ["+timePrepareDuration+"]...or ["+timePreparationDurationSec+"] sec");
		
		System.out.println("Generating Query Mesh config plan...");
		
		OptimizerToExecutorTranslator.GenerateConfigFile(CURRENT_QM_TYPE, //STQM functionality
														(ArrayList) xmlOperators,
				 										 outputXmlPlan, 
														 rootOpId, 
														 rootOpChildren,
														 operatorsDetails,			
														 classifierSignature,
														 localQMSignatures,
														 hashAllLogicalToPhysicalOpCombinations,
														 hashLogicalOperatorsWithInputsCombinations,
														 trainingSetCollectorOperatorsDetails);
	}
	
	//-----------------------------------------------------------------------------
	
	public static void GetOptimizerStatistics()
	{		
		Enumeration en = hashNumIterationsPerStream.keys();
		while(en.hasMoreElements()) 
		{
			String key = (String)en.nextElement();			
			int numIterations = Integer.valueOf(hashNumIterationsPerStream.get(key).toString());
			
			System.out.println(key + " --> " + numIterations + " iterations");			
			Helper.StatsWriteLineToFile(STATISTICS_OUTPUT_FILE, key + " --> " + numIterations + " iterations" + System.getProperty("line.separator"));
		}
	}
	
	//-----------------------------------------------------------------------------
	
	public static String GetCurrentDateTime() 
	{
		String DATE_FORMAT_NOW = "yyyy-MM-dd HH:mm:ss";
	    Calendar cal = Calendar.getInstance();
	    SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_NOW);
	    return sdf.format(cal.getTime());
	}
	
	//-----------------------------------------------------------------------------
	
	private static ArrayList PrepareDataForOnlineClassifierConfig(GlobalQueryMesh globalQM)
	{
		String strGlobalQM = "";
		String NEW_LINE = System.getProperty("line.separator");
		ArrayList localQMSignatures = new ArrayList();
		
		for(int i = 0; i < globalQM.localQueryMeshes.size(); i++)
		{			
			LocalQueryMesh localQM = (LocalQueryMesh)globalQM.localQueryMeshes.get(i);
			//--------------------------------------------------------------------
			// TRAVERSE LOCAL CLASSIFIER
			//--------------------------------------------------------------------
			Object localClassifier = localQM.getClassifier();
			DecisionTreeID3 dtClassifier = null;
			
			assert localClassifier != null;
			
			if(localClassifier.getClass().toString().equals("class edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.DecisionTreeID3"))
			{
				dtClassifier = (DecisionTreeID3) localClassifier;
			}
			
			assert dtClassifier != null;
									
			int nestedLevel = 1;			

			//--------------------------------------------------------------------
			// DT NODES SIGNATURES - START
			//--------------------------------------------------------------------			
			ArrayList allLocalDTNodes = new ArrayList();
			String rootNodeId = "0";
			dtClassifier.NodeID = 1;
			
			try
			{			
				ArrayList childrenNodeIds = dtClassifier.prepareNodeSignatures(rootNodeId, dtClassifier.GetRoot(), allLocalDTNodes);
			
				//create root node			
				DecisionTreeNodeSignature nodeRoot = 
					new DecisionTreeNodeSignature("",
												   rootNodeId, 
												   "root",
												   "",
												   "",//attrIdx,
												   "",//attrName,
												   "",//isBucketized,
												   "",//nodeOperation,
												   "",//nodeOperationTest);
												   childrenNodeIds);
				
				
				allLocalDTNodes.add(nodeRoot);
				
			}
			catch(Exception ex)
			{
				System.out.println(ex.getMessage());
			}
			//--------------------------------------------------------------------
			// DT NODES SIGNATURES - END
			//--------------------------------------------------------------------
					
			
			//--------------------------------------------------------------------
			// TRAVERSE LOCAL ROUTES
			//--------------------------------------------------------------------			
			Hashtable <String,PhysicalRouteSignature> allLocalRoutes = new Hashtable();
			Hashtable hashRusterRouteIdMapping = new Hashtable();
			
			Hashtable hashCachePhysicalOrderOfPlans = new Hashtable();
			Hashtable hashCachePhysicalOrderStrSignatureOfPlans = new Hashtable();
			String routeStrSignature = "";
			
			for(int j = 0; j < localQM.getRoutes().size()+1; j++)
			{				
				RouteSignature routeSignature;
				QueryPlan currRoute;
				String rusterName = "";
				
				if (j == localQM.getRoutes().size())
				{
					//process default route
					routeSignature = localQM.getDefaultRoute();
					currRoute = (QueryPlan)routeSignature.getQueryPlan();
					rusterName = "default_route";
				}
				else
				{
					routeSignature = (RouteSignature)localQM.getRoutes().get(j);
					currRoute = (QueryPlan)routeSignature.getQueryPlan();
					rusterName = routeSignature.getRuster();
				}
				
				String routeId = ""+(j+1);
				strGlobalQM += "<Route id=\""+routeId+"\" value=\""+currRoute.toString()+"\" ruster=\""+rusterName+"\">" + NEW_LINE;								
				
				Hashtable physicalOpOrder = new Hashtable();
				
				if (!hashCachePhysicalOrderOfPlans.containsKey(currRoute.toString()))
				{
					//strGlobalQM += currRoute.toXmlString();			
					Hashtable hashUnwindPlan = new Hashtable();
					UnwindPlan(hashUnwindPlan, currRoute.getRoot());
					Hashtable hashOpOrder = new Hashtable();
					
					//--------------------------------------------------------------------
					// ITERATVE HASH AND CREATE XML PLAN
					//--------------------------------------------------------------------
					Set keys = hashUnwindPlan.keySet();
					Object [] strKeys = (Object [])keys.toArray();
					for (int q =0; q < keys.size(); q++)
					{
						String key = (String)strKeys[q];
						OperatorSignature opSignature = (OperatorSignature)hashUnwindPlan.get(key);
						String order = "";
						if (key.length() == 4)
						{
							order = "1";
						}
						else if (key.length() == 7)
						{
							order = "2";
						}
						else if (key.length() == 10)
						{
							order = "3";
						}
						else if (key.length() == 13)
						{
							order = "4";
						}
						else if (key.length() == 16)
						{
							order = "5";
						}
						else if (key.length() == 19)
						{
							order = "6";
						}
						else if (key.length() == 22)
						{
							order = "7";
						}
						else if (key.length() == 25)
						{
							order = "8";
						}
						else if (key.length() == 28)
						{
							order = "9";
						}
						
						hashOpOrder.put(order, opSignature);
					}							
					
					
					for(int q = 1; q <= hashOpOrder.size(); q++)
					{
						OperatorSignature opSignature = (OperatorSignature)hashOpOrder.get(""+q+"");
						
						strGlobalQM += "<Operator>" + NEW_LINE;
						strGlobalQM += "<Input>" + opSignature.leftChild +"</Input>"+NEW_LINE;
						strGlobalQM += "<Input>" + opSignature.rightChild +"</Input>"+NEW_LINE;					
						strGlobalQM += "</Operator>" + NEW_LINE;
						String order = ""+q;
						String opID = GetOpIdBasedOnInputs(opSignature.leftChild, opSignature.rightChild);
						if (opID == "")
						{
							System.out.println();
						}
						physicalOpOrder.put(order, opID);
						routeStrSignature += order+"->"+opID+"|";
					}
					strGlobalQM += "</Route>" + NEW_LINE;
	
					hashCachePhysicalOrderOfPlans.put(currRoute.toString(), physicalOpOrder);
					hashCachePhysicalOrderStrSignatureOfPlans.put(currRoute.toString(), routeStrSignature);
				}//end if there is no physical order of the plan in local cache
				else
				{
					physicalOpOrder = (Hashtable)hashCachePhysicalOrderOfPlans.get(currRoute.toString());
					routeStrSignature = (String)hashCachePhysicalOrderStrSignatureOfPlans.get(currRoute.toString());
				}
								
				boolean isDefaultRoute = false;
				if (rusterName.equalsIgnoreCase("default_route"))
				{
					isDefaultRoute = true;
				}
				else
				{
					isDefaultRoute = false;
				}
				
				PhysicalRouteSignature physRouteSignature = new PhysicalRouteSignature(routeId, 
																					   rusterName, 
																					   currRoute.toString(),
																					   physicalOpOrder, 
																					   isDefaultRoute);
				
				if (!allLocalRoutes.containsKey(routeStrSignature))				
				{
					allLocalRoutes.put(routeStrSignature, physRouteSignature);
					hashRusterRouteIdMapping.put(rusterName, routeId);   
				}
				else
				{					
					PhysicalRouteSignature alreadyExistingRoute = (PhysicalRouteSignature)allLocalRoutes.get(routeStrSignature);
					String alreadyExistingRouteId = alreadyExistingRoute.GetRouteId();
					//store the mapping to already existing route
					hashRusterRouteIdMapping.put(rusterName, alreadyExistingRouteId);
					
					if (isDefaultRoute == true)
					{
						alreadyExistingRoute.SetDefaultRoute(isDefaultRoute);
					}
				}
				routeStrSignature = "";
			}//iteration of routes

			//--------------------------------------------------------------------
			// ASSEMBLE LOCAL QM SIGNATURE
			//--------------------------------------------------------------------	
			LocalQMSignature localQMSignature = new LocalQMSignature(localQM.getStreamId(), 
																	 allLocalDTNodes, 
																	 allLocalRoutes,
																	 hashRusterRouteIdMapping);
			localQMSignatures.add(localQMSignature);
			
		}//iteration of all local query meshes
		
		return localQMSignatures;
	}
	
	//-----------------------------------------------------------------------------
	
	private static String GetOpIdBasedOnInputs(String leftChild, String rightChild)
	{
		//NOTE: because we assume left-linear query plans, the right child is
		//the id of the stream
		String opId = "";
		//--------------------------------------------------------------------
		// TRAVERSE OPERATORS AND MATCH AGAINST THEIR INPUTS (ASSUME LEFT-LINEAR TREES)
		//--------------------------------------------------------------------
		Iterator it = xmlOperators.iterator();
		while(it.hasNext()) 
		{
			SimpleOperatorSpecification currOpt = (SimpleOperatorSpecification)it.next();
			//System.out.println(currOpt.toString());
			String input1 = (String)currOpt.getInputs().get(0);
			String input2 = (String)currOpt.getInputs().get(1);
			
			if (rightChild.length() > 2)
			{
				String temp = leftChild;
				leftChild = rightChild;
				rightChild = temp;
			}
			
			if (input1.equals(rightChild))
			{
				if (leftChild.contains(input2))
				{
					//match found
					opId = currOpt.getId();
					return opId;
				}
			}
			else
				if (input2.equals(rightChild))
				{
					if (leftChild.contains(input1))
					{
						//match found
						opId = currOpt.getId();
						return opId;
					}
				}
		}
		return opId;
	}
	
	//-----------------------------------------------------------------------------
	
	private static void GetTotalNumberOfPhysicalOperators(GlobalQueryMesh globalQM, Hashtable hashPhysOperators)
	{
		//--------------------------------------------------------------------
		// ITERATE THROUGH LOCAL QUERY MESHES
		//--------------------------------------------------------------------
		for(int i=0; i < globalQM.localQueryMeshes.size(); i++)
		{
			LocalQueryMesh localQM = (LocalQueryMesh)globalQM.localQueryMeshes.get(i);
			
			for(int j=0; j < localQM.getRoutes().size(); j++)
			{
					
				RouteSignature routeSignature = (RouteSignature)localQM.getRoutes().get(j);
				QueryPlan currRoute = (QueryPlan)routeSignature.getQueryPlan();
				String rusterName = routeSignature.getRuster();
				//strGlobalQM += "<Route id=\""+(j + 1)+"\" value=\""+currRoute.toString()+"\" ruster=\""+rusterName+"\">" + NEW_LINE;								
				
				//strGlobalQM += currRoute.toXmlString();			
				Hashtable hashUnwindPlan = new Hashtable();
				UnwindPlan(hashUnwindPlan, currRoute.getRoot());
				Hashtable hashOpOrder = new Hashtable();
				
				//--------------------------------------------------------------------
				// ITERATVE HASH AND CREATE XML PLAN
				//--------------------------------------------------------------------
				Set keys = hashUnwindPlan.keySet();
				Object [] strKeys = (Object [])keys.toArray();
				for (int q =0; q < keys.size(); q++)
				{
					String key = (String)strKeys[q];
					OperatorSignature opSignature = (OperatorSignature)hashUnwindPlan.get(key);
					String order = "";
					if (key.length() == 4)
					{
						order = "1";
					}
					else if (key.length() == 7)
					{
						order = "2";
					}
					else if (key.length() == 10)
					{
						order = "3";
					}
					else if (key.length() == 13)
					{
						order = "4";
					}
					else if (key.length() == 16)
					{
						order = "5";
					}
					else if (key.length() == 19)
					{
						order = "6";
					}
					else if (key.length() == 22)
					{
						order = "7";
					}
					else if (key.length() == 25)
					{
						order = "8";
					}
					else if (key.length() == 28)
					{
						order = "9";
					}
					else if (key.length() == 31)
					{
						order = "10";
					}
					hashOpOrder.put(order, opSignature);
				}
				
				for(int q = 1; q <= hashOpOrder.size(); q++)
				{
					OperatorSignature opSignature = (OperatorSignature)hashOpOrder.get(""+q+"");
					
					//strGlobalQM += "<Operator>" + NEW_LINE;
					//strGlobalQM += "<Input>" + opSignature.leftChild +"</Input>"+NEW_LINE;
					//strGlobalQM += "<Input>" + opSignature.rightChild +"</Input>"+NEW_LINE;					
					//strGlobalQM += "</Operator>" + NEW_LINE;
				}
				//strGlobalQM += "</Route>" + NEW_LINE;
			}
		}
	}
	
	//-----------------------------------------------------------------------------
	
	public static void WriteToDumpFile(String fileLocation, String stringToWrite)
	{
		try
		{
			//Create file 
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			out.newLine();
			//Close the output stream
			out.close();			    
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}
	
	//-----------------------------------------------------------------------------

	public static String getGlobalQMString(GlobalQueryMesh globalQM)
	{
		String strGlobalQM = "";
		assert globalQM != null;
		String NEW_LINE = System.getProperty("line.separator");
		
		strGlobalQM += "<GlobalQueryMesh>" + NEW_LINE;
		
		//--------------------------------------------------------------------
		// ITERATE THROUGH LOCAL QUERY MESHES
		//--------------------------------------------------------------------
		for(int i=0; i < globalQM.localQueryMeshes.size(); i++)
		{
			strGlobalQM += "<!-- Local Query Mesh "+(i+1)+" -->" + NEW_LINE;
			
			strGlobalQM += "<LocalQueryMesh id=\""+(i+1)+"\">" + NEW_LINE;
			LocalQueryMesh localQM = (LocalQueryMesh)globalQM.localQueryMeshes.get(i);
			//--------------------------------------------------------------------
			// TRAVERSE LOCAL CLASSIFIER
			//--------------------------------------------------------------------
			Object localClassifier = localQM.getClassifier();
			DecisionTreeID3 dtClassifier = null;
			
			assert localClassifier != null;
			
			if(localClassifier.getClass().toString().equals("class edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.DecisionTreeID3"))
			{
				dtClassifier = (DecisionTreeID3) localClassifier;
			}
			
			assert dtClassifier != null;
									
			strGlobalQM += "<!---------------------Decision Tree "+(i+1)+"-------------------------------->" + NEW_LINE;
			//strGlobalQM += dtClassifier.getTreeStr(dtClassifier.GetRoot(), "");
			int nestedLevel = 1;			
			strGlobalQM += "<DecisionTree>" + NEW_LINE;
			strGlobalQM += dtClassifier.printXmlTree(dtClassifier.GetRoot(), "" + nestedLevel);
			strGlobalQM += "</DecisionTree>" + NEW_LINE;
			//--------------------------------------------------------------------
			// TRAVERSE LOCAL ROUTES
			//--------------------------------------------------------------------
			for(int j=0; j < localQM.getRoutes().size(); j++)
			{
				strGlobalQM += "<!---------------------Route "+(j+1)+"----------------------------->" + NEW_LINE;
				
				RouteSignature routeSignature = (RouteSignature)localQM.getRoutes().get(j);
				QueryPlan currRoute = (QueryPlan)routeSignature.getQueryPlan();
				String rusterName = routeSignature.getRuster();
				strGlobalQM += "<Route id=\""+(j + 1)+"\" value=\""+currRoute.toString()+"\" ruster=\""+rusterName+"\">" + NEW_LINE;								
				
				//strGlobalQM += currRoute.toXmlString();			
				Hashtable hashUnwindPlan = new Hashtable();
				UnwindPlan(hashUnwindPlan, currRoute.getRoot());
				Hashtable hashOpOrder = new Hashtable();
				
				//--------------------------------------------------------------------
				// ITERATVE HASH AND CREATE XML PLAN
				//--------------------------------------------------------------------
				Set keys = hashUnwindPlan.keySet();
				Object [] strKeys = (Object [])keys.toArray();
				for (int q =0; q < keys.size(); q++)
				{
					String key = (String)strKeys[q];
					OperatorSignature opSignature = (OperatorSignature)hashUnwindPlan.get(key);
					String order = "";
					if (key.length() == 4)
					{
						order = "1";
					}
					else if (key.length() == 7)
					{
						order = "2";
					}
					else if (key.length() == 10)
					{
						order = "3";
					}
					else if (key.length() == 13)
					{
						order = "4";
					}
					else if (key.length() == 16)
					{
						order = "5";
					}
					else if (key.length() == 19)
					{
						order = "6";
					}
					else if (key.length() == 22)
					{
						order = "7";
					}
					else if (key.length() == 25)
					{
						order = "8";
					}
					else if (key.length() == 28)
					{
						order = "9";
					}
					else if (key.length() == 31)
					{
						order = "10";
					}
					hashOpOrder.put(order, opSignature);
				}
				
				for(int q=1; q <= hashOpOrder.size(); q++)
				{
					OperatorSignature opSignature = (OperatorSignature)hashOpOrder.get(""+q+"");
					
					strGlobalQM += "<Operator>" + NEW_LINE;
					strGlobalQM += "<Input>" + opSignature.leftChild +"</Input>"+NEW_LINE;
					strGlobalQM += "<Input>" + opSignature.rightChild +"</Input>"+NEW_LINE;					
					strGlobalQM += "</Operator>" + NEW_LINE;
				}
				strGlobalQM += "</Route>" + NEW_LINE;
			}
		}		
		strGlobalQM += "</GlobalQueryMesh>" + NEW_LINE;
		return strGlobalQM;
	}

	//-----------------------------------------------------------------------------
	/*
	public static String UnwindPlan(Hashtable hashPlan, Node node)
	{		
		if (node.chidrenCount() == 0)
		{
			return node.getLabel();
		}
		else
		{
			return UnwindPlan(hashPlan, (Node)node.getChild(0));
		}
	}
	*/
	//-----------------------------------------------------------------------------
	
	public static void UnwindPlan(Hashtable hashPlan, Node plan)
	{			
		if(plan.getChild(0) != null)
		{
			if(plan.getChild(0) instanceof BinaryJoinNodeImpl)
			{
				UnwindPlan(hashPlan,((Node) plan.getChild(0)));
			}
			else if(plan.getChild(0) instanceof Node)
			{
				UnwindPlan(hashPlan,((Node) plan.getChild(0)));
			}
		}
		if(plan.getChild(1) != null)
		{
			if(plan.getChild(1) instanceof BinaryJoinNodeImpl)
			{
				UnwindPlan(hashPlan, ((Node) plan.getChild(1)));
			}
			else if(plan.getChild(1) instanceof Node)
			{
				UnwindPlan(hashPlan, ((Node) plan.getChild(1)));
			}
		}		
		
		String leftChildLabel = "";
		if (plan.getChild(0) != null)
		{
			leftChildLabel = plan.getChild(0).toString();
		}
		
		String rightChildLabel = "";
		if (plan.getChild(1) != null)
		{
			rightChildLabel = plan.getChild(1).toString();
		}
		
		OperatorSignature opSignature = new OperatorSignature(leftChildLabel, rightChildLabel);
		String key = "("+leftChildLabel + rightChildLabel+")";
		if (!key.equalsIgnoreCase("()"))
		{
			hashPlan.put(key,opSignature);
		}
		if (key.startsWith("&"))
		{
			key = rightChildLabel;
		}
		
		return;
	}
	
	//-----------------------------------------------------------------------------
	
	//--------------------------------------------------------------------
	// PRODUCES QUERY MESH PER INPUT STREAM
	//--------------------------------------------------------------------	
	public LocalQueryMesh produceLocalQueryMesh(SimpleOperatorInputSpecification input, 
												QM_SEARCH_STRATEGY heuristic,
									  			int mVal,
									  			QM_SEARCH_START_SOLUTION startSolution)
	{
		String tabs = "\t";
		//initialize query mesh for the current stream
		LocalQueryMesh streamQM = new LocalQueryMesh(input.getId());
				
		DecisionTreeID3 dt;
		double totalQueryMeshCost = 0.0;
		int numTrainingTuples;
		S<String> s;
		ArrayList allPartitions;
		
		int m; 
		HybridSearch hs; //search algorithm that embodies other search heuristics
		
		//--------------------------------------------------------------------------
    	numTrainingTuples = input.getNumTrainingTuples();
		System.out.println(tabs + "# of training tuples: " + numTrainingTuples);
		
		//initialize tuple space - from 1...to...n, encodes training tuples
		//that we are working with. 1 - first training tuple in the set,
		//10 - 10th training tuple, and so on
		s = new S<String>();
		
		for(int i = 1; i <= numTrainingTuples; i++)
		{
			s.add("" + i);
		}
		//---------------------------------------------------------------
	    // ITERATE BASED ON THE NUMBER OF SUBSETS IN A PARTITION... from n to 1
		// where n = # training tuples
		//---------------------------------------------------------------
		
		allPartitions = new ArrayList();
		m = mVal;
		if (m != 0)
		{
    		for(int i = numTrainingTuples; i >= 0; i--)
    		{
    			ArrayList currSubsets = s.partitionsBySubsetCount(i);
    			System.out.println(tabs + "partitions by k  ("+currSubsets.size()+"): "+ currSubsets);
    			allPartitions.addAll(currSubsets);
    		}
		}
		        		
		hs = new HybridSearch();
		streamQM = hs.runSearchQM(heuristic, 
								  startSolution, 
								  input, 
								  input.getTrainingTuples(), 
								  allPartitions, 
								  xmlOperators, 
								  xmlInputs, 
								  m);
		//--------------------------------------------------------------------------
				
		return streamQM;
	}	

	//-----------------------------------------------------------------------------
	
	//-----------------------------------------------------------------------------
	// DELETE FILES FROM TMP DIRECTORY
	//-----------------------------------------------------------------------------
	public void cleanWorkingDirectory()
	{
		File dir = new File(WORKING_DIRECTORY);
		String[] info = dir.list();
		for (int i = 0; i < info.length; i++) 
		{
			File n = new File(WORKING_DIRECTORY + info[i]);
			System.out.println("Removed: " + n.getPath());
			try
			{
				n.delete();
			}
			catch(Exception ex)
			{
				System.err.println("Couldn't remove: " + n.getPath());
				System.err.println(ex.getMessage());
			}
		}
	}
	
	//-----------------------------------------------------------------------------
	
	//-----------------------------------------------------------------------------
	// PRODUCES QUERY MESH FOR ALL INPUT STREAMS
	//-----------------------------------------------------------------------------
	public GlobalQueryMesh produceGlobalQueryMesh(QM_SEARCH_STRATEGY heuristic, 
												  int m, 
												  QM_SEARCH_START_SOLUTION startSolution)
	{	
		GlobalQueryMesh bestQM = new GlobalQueryMesh();
		LocalQueryMesh streamQM;
		
		//clean working directory before starting query mesh search
		System.out.println("Cleaning up working directory...");
		cleanWorkingDirectory();
	
		System.out.println("Optimizer inputs:");
		System.out.println("\tsearch strategy: " + heuristic.toString());
		System.out.println("\tvalue of m: " + m);
		System.out.println("\tstart solution: " + startSolution);
		
		
		Hashtable allInputs = GetAllInputs();		
		Enumeration e = allInputs.elements(); 
		
		System.out.println("\nCreating local query meshes...");
		int qmCount = 0;
		while (e.hasMoreElements()) 
		{
			qmCount++;
			System.out.println("\n\n*************************************************************************************************");
			System.out.println("*************************************************************************************************");
			SimpleOperatorInputSpecification currInput = (SimpleOperatorInputSpecification)e.nextElement();  // do whatever you need with x
			System.out.println("\n--------------------QUERY MESH ["+qmCount+"], Stream ["+currInput.getId()+"]------------------------");
			 
			//--------------------------------------------------------------------
			// CALL LOCAL QUERY MESH SEARCH FOR EACH STREAM
			//--------------------------------------------------------------------	
			streamQM = produceLocalQueryMesh(currInput, 
											 heuristic, 
											 m, 
											 startSolution);	 
			bestQM.addQM(streamQM);
		}//end while (iterating all stream inputs)
		
		System.out.println("\n\n************************* DONE *************************************************************");
		
		return bestQM;
	}	
	
	//-----------------------------------------------------------------------------
}


