//--------------------------------------------------
// S (Set).java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;


public class S<T> extends HashSet<T> {
    
    private static final long serialVersionUID = -8875179328198507416L;
 
    public S(Collection<T> collection) { super(collection); }
    
    public S(T arg) { this.add(arg); }
    
    public S(T ... args) { super(Arrays.asList(args)); }
    
    
private S<T> unionHere(T arg) {
        
        add(arg);
        
        return this;
    }
    
    public S<T> union(T arg) { 
        
        return new S<T>(this).unionHere(arg);
    }
    
    public S<T> union(Collection<T> collection) {
        
        S<T> copy= new S<T>(this);
        
        copy.addAll(collection);
        
        return copy;
    }
    
    public S<T> difference(T arg) {
        
        S<T> copy= new S<T>(this);
        
        copy.remove(arg);
        
        return copy;
    }
    
    public S<T> difference(Collection<T> collection) {
        
        S<T> copy= new S<T>(this);
        
        copy.removeAll(collection);
        
        return copy;
    }

    
    public S<T> symDifference(Collection<T> collection) {
        
        return difference(collection).union(new S<T>(collection).difference(this));
    }
    
    public S<T> intersection(Collection<T> collection) {
        
        S<T> copy= new S<T>(this);
        
        copy.retainAll(collection);
        
        return copy;
    }

    
public S<S<T>> combinations() 
{
        
        S<S<T>> result= new S<S<T>>(new S<T>());
        
        for (T current : this) 
        {
            S<S<T>> comb = difference(current).combinations();
            result.addAll(comb);
            for (S<T> set : comb)
                result.unionHere(set.union(current));
        }
        
        return result;
    }


public S<S<S<T>>> partitions() {
    
    S<S<S<T>>> result= new S<S<S<T>>>();

    if (size() < 2)
        return result.unionHere(new S<S<T>>(this));
        
    T current= iterator().next();
    S<S<S<T>>> part= difference(current).partitions();

    for (S<S<T>> set : part) 
    {
        result.unionHere(set.union(new S<T>(current)));
        
        for (S<T> elem : set) 
        {
            S<S<T>> copy= set.difference(elem);
            result.unionHere(copy.unionHere(elem.union(current  )));           
        }
    }
    
    return result;
}

public ArrayList partitionsBySubsetCount(int subsetCount) 
{
	ArrayList prunedResult= new ArrayList();
    S<S<S<T>>> result= new S<S<S<T>>>();

    //if (size() < 2)
    //    return result.unionHere(new S<S<T>>(this));
        
    T current= iterator().next();
    S<S<S<T>>> part= difference(current).partitions();

    for (S<S<T>> set : part) 
    {
        result.unionHere(set.union(new S<T>(current)));
        
        for (S<T> elem : set) 
        {
            S<S<T>> copy= set.difference(elem);
            result.unionHere(copy.unionHere(elem.union(current  )));           
        }
    }
    
    //-------------------------------------------------------------------
    //now return only the partitions with the subset count requested
    //-------------------------------------------------------------------
    
    Iterator iter = result.iterator();			
	while (iter.hasNext())
	{		
		Object key = iter.next();				
		//System.out.print("\t");	
		S<S> inner = (S<S>)key;
		
		if (inner.size() == subsetCount)
		{
			prunedResult.add(inner);
		}
		
		/*
			Iterator iterInner = inner.iterator();			
			while (iterInner.hasNext())
			{
				Object innerKey = iterInner.next();	
				System.out.println("inner"+ innerKey);
			}	
		}
		*/
	}
    
    return prunedResult;
}



}
    
 
 