//--------------------------------------------------
// SimpleOperatorInputSpecification.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;

public class SimpleOperatorInputSpecification implements java.io.Serializable
{
	//----------------------------------------------------------------------------------------------------
	
	private String id;
	private String name;
	private String trainingSetFile;	
	private String statisticsFile;	
	private int overallCardinality = -999;
	private Hashtable schemaAttributes;
	private Hashtable reverseSchemaAttributes;
	
	//----------------------------------------------------------------
	// STATISTICS DATA VALUES
	//----------------------------------------------------------------
	private int numTrainingTuples;
	private ArrayList trainingTuples;
	
	//----------------------------------------------------------------------------------------------------
	
    public SimpleOperatorInputSpecification(String idVal, 
    										String nameVal,
    										String trainingSetFileVal,
    										String statisticsFileVal,
    										Hashtable schemaAttributesVal)
    {
    	id = idVal;
    	name = nameVal;
    	trainingSetFile = trainingSetFileVal;
    	statisticsFile = statisticsFileVal;
    	schemaAttributes = schemaAttributesVal;
    	reverseSchemaAttributes = doReverseSchemaAttributeMapping(schemaAttributes);    	
    	populateStatistics();
    }

  //----------------------------------------------------------------------------------------------------
    
    private Hashtable doReverseSchemaAttributeMapping(Hashtable attributeIdxHash)
    {
    	Hashtable reverseAttributeIdxHash = new Hashtable();
    	
    	java.util.Enumeration en = attributeIdxHash.keys(); 
		while (en.hasMoreElements()) 
		{
			String currAttributeKey = (String)en.nextElement();
			String attrName = (String)attributeIdxHash.get(currAttributeKey);
			reverseAttributeIdxHash.put(attrName, currAttributeKey);
		}
    	
    	return reverseAttributeIdxHash;
    }
    
    //----------------------------------------------------------------------------------------------------
    
    public String getAttrNameByIndex(int attr_idx)
    {
    	return (String)reverseSchemaAttributes.get(""+attr_idx);
    }
    
   //----------------------------------------------------------------------------------------------------
    
    public String toString() 
    {
        StringBuilder result = new StringBuilder();
        String NEW_LINE = System.getProperty("line.separator");

        result.append("-----------operator input ----------------"+ NEW_LINE);
        result.append("Input Id: " + id + NEW_LINE);
        result.append("Input Name: " + name + NEW_LINE);
        result.append("Input Training Set File: " + trainingSetFile + NEW_LINE );
        result.append("Input Statistics File: " + statisticsFile + NEW_LINE);
        
        result.append("Input Stream Schema: " + NEW_LINE);
        
        java.util.Enumeration en = reverseSchemaAttributes.keys(); 
        
		while (en.hasMoreElements()) 
		{
			String currAttributeKey = (String)en.nextElement();
			String attrName = (String)reverseSchemaAttributes.get(currAttributeKey);
			
			result.append(currAttributeKey +". "+ attrName + NEW_LINE);			
		}
    	        
        result.append("-----------end operator input-------------"+ NEW_LINE);
        return result.toString();
     }

    //----------------------------------------------------------------------------------------------------
    
	public String getId()
	{
		return id;		
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public Hashtable getSchemaMapping()
	{
		return schemaAttributes;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public String getInputName()
	{
		return name;		
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public String getStatisticsFileName()
	{
		return statisticsFile;		
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public int getNumTrainingTuples()
	{
		return numTrainingTuples;		
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public ArrayList getTrainingTuples()
	{
		return trainingTuples;		
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public double getCardinality()
	{
		double card = 0;
		
		//-------------------------------------------------------------------
		// OPEN SAMPLE DUMP FILE AND COMPUTE CARDINALITY
		//-------------------------------------------------------------------
		if (overallCardinality == -999)
		{
			overallCardinality = computeCardinalityFromStatsFile();
		}
		card = overallCardinality;
	    return card;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public int computeCardinalityOfPartition(String partitionName, String workingFile)
	{
		int cardinality = 0;
		
		FileInputStream fin;		
		int count = 0;
		try
		{
			FileReader input = new FileReader(workingFile);
			BufferedReader bufRead = new BufferedReader(input);
			String line;    // String that holds current file line
			
			// Read first line
			line = bufRead.readLine();
			if (line != null)
			{
				while(!line.startsWith("//")) 					
				{
					 line = bufRead.readLine();					
				}
			}

		    //Read through file one line at time. Traverse the DT
		    while (line != null)
		    {
			    line = bufRead.readLine();
			    
			    if (line == null) continue;
			    if (line.startsWith("//")) continue;

			    String [] tokens = line.split("  ");
			    
			    //last column contains the name of the partition
			    String currTuplePartition = tokens[tokens.length-1];
			    
			    if (currTuplePartition.equalsIgnoreCase(partitionName))
			    {
			    	//found a match, increment the count
			    	count++;			    	
			    }	     		
		    }		    
		    // Close our input stream
		    bufRead.close();
		}
		// Catches any error conditions
		catch (IOException e)
		{
			System.err.println ("!!! Unable to read from file [" + workingFile + "]!!!");
		}
		cardinality = count;

		return cardinality;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public int computeCardinalityFromStatsFile()
	{
		int cardinality = 0;
		
		FileInputStream fin;		
		int count = 0;
		try
		{
			FileReader input = new FileReader(statisticsFile);
			/* Filter FileReader through a Buffered read to read a line at a time */
			BufferedReader bufRead = new BufferedReader(input);
			String line;    // String that holds current file line
			
			// Read first line
			line = bufRead.readLine();
			if (line != null)
			{
				if (!line.equalsIgnoreCase(""))
				{
					count++;
				}
			}
			
		    while (line != null)
		    {
			    line = bufRead.readLine();
			    
			    if (line == null) continue;
			    if (line.startsWith("//")) continue;
			    if (line.equalsIgnoreCase("")) continue;
			    count++;
		    }		    
		    // Close our input stream
		    bufRead.close();

		}
		// Catches any error conditions
		catch (IOException e)
		{
			System.err.println ("!!! Unable to read from file [" + statisticsFile + "]!!!");
		}
		cardinality = count;
		
		return cardinality;
	}
	
	//----------------------------------------------------------------------------------------------------
	
	private void populateStatistics()
	{
		int card = 0;
		trainingTuples = new ArrayList();
		
		//-------------------------------------------------------------------
		// OPEN SAMPLE DUMP FILE AND COMPUTE CARDINALITY
		//-------------------------------------------------------------------
		try 
		{
	        BufferedReader in = new BufferedReader(new FileReader(trainingSetFile));
	        String str;
	        while ((str = in.readLine()) != null) 
	        {
				if (str == null) break;
				if (str.startsWith("//")) continue;
				if (str.equals("")) continue;
				if (str.equals(" ")) continue;
				
	        	//-----------------------------------------------------------
	        	// TOKENIZE STRING
	        	//-----------------------------------------------------------
	        	String [] tokens = str.split("\\|");
	        	int numAttributes = tokens.length;
				if (numAttributes < 1) 
				{
					System.err.println( "Read line: [" + str + "]");
					System.err.println( "Could not obtain attributes in the line");
					System.err.println( "Expecting at least one input attribute and one output attribute");
					continue;
				}

				String[] attributes = new String[numAttributes];
				for (int i=0; i < numAttributes; i++) 
				{
					attributes[i] = tokens[i];
				}

	        	//-----------------------------------------------------------
	        	// ADD TRAINING TUPLE TO THE COLLECTION
	        	//-----------------------------------------------------------
				trainingTuples.add(attributes);
	        	
	        	//-----------------------------------------------------------
	        	// INCREASE CARDINALITY
	        	//-----------------------------------------------------------
	        	card++;
	        }
	        in.close();	        
	    } 
		catch (IOException e) 
	    {
			e.printStackTrace();			
	    }
		
		//--------------------------------------------------------------------
		numTrainingTuples = card; 
	}
	
	//----------------------------------------------------------------------------------------------------
	
	public ArrayList getPartitionTuples(S partitionTuplesIdxs)
	{
		ArrayList partitionTuples = new ArrayList();
		ArrayList trainingTuples = this.getTrainingTuples();
		
		for(int p = 0; p < trainingTuples.size(); p++)
		{
			if (partitionTuplesIdxs.contains(p))
			{
				partitionTuples.add(trainingTuples.get(p));
			}
		}
		
		return partitionTuples;
	}
	
	//----------------------------------------------------------------------------------------------------
	
}
