//--------------------------------------------------
// SimpleOperatorSpecification.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;

public class SimpleOperatorSpecification implements java.io.Serializable
{
	private String id;
	private String operatorType;
	private ArrayList<SimpleOperatorInputSpecification> inputs;
	private double operatorSelectivity;
	private Hashtable hashJoinAttributes; //hash tables where for every input id, it contains the index of the join attribute
	
    public SimpleOperatorSpecification(String idVal,
    								   String operatorTypeVal,     								   
    								   ArrayList inputsVal,
    								   Hashtable hashJoinAttributesVal)
    {
    	id = idVal;
    	operatorType = operatorTypeVal;
    	inputs = inputsVal;
    	hashJoinAttributes = hashJoinAttributesVal;
    }
    
    public String toString() 
    {
        StringBuilder result = new StringBuilder();
        String NEW_LINE = System.getProperty("line.separator");

        //result.append(this.getClass().getName() + " Object {" + NEW_LINE);
        result.append("---------------- operator----------------" + NEW_LINE);
        result.append("Op Id: " + id + NEW_LINE);
        result.append("Op Type: " + operatorType + NEW_LINE);
        result.append("# inputs: " + inputs.size() + NEW_LINE );        
        result.append("-------------end operator----------------"+ NEW_LINE);
        return result.toString();
     }
    
	public String getId()
	{
		return id;		
	}
	
	public double computeOveralStreamSelectivity()
	{
		return operatorSelectivity;		
	}
	
	public double computeOpSelectivityBasedOnPartition()
	{
		double partitionBasedOperatorSelectivity = 0;
		return partitionBasedOperatorSelectivity;		
	}
	
	public ArrayList getInputs()
	{
		return inputs;		
	}
		
	public int getJoinAttributeIdxByInputId(String inputId)
	{
		if (hashJoinAttributes.containsKey(inputId))
		{
			Integer joinAttrIdx = Integer.parseInt((String)hashJoinAttributes.get(inputId));	
			return joinAttrIdx;
		}
		return -1; 
	}
}
