//--------------------------------------------------
// SimpleQueryPlanSpecification.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;

import java.util.ArrayList;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.MJoinDynamicAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.bGreedyAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.bGreedyAlgorithmForward;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes.JoinGraphImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.Stream;

public class SimpleQueryPlanSpecification implements java.io.Serializable
{

	//----------------------------------------------------------------------------------------------------
	
	private ArrayList operators;
	private JoinGraphImpl graph;

	//----------------------------------------------------------------------------------------------------
	
	public SimpleQueryPlanSpecification(ArrayList operatorsVal)
	{	
		operators = operatorsVal;	 
	}	    
	
	//----------------------------------------------------------------------------------------------------
		
	public ArrayList getOperators()
	{
		return operators;			
	}
	
	//----------------------------------------------------------------------------------------------------
		 
	private Stream SetupJoinGraph(String startPointName)
	{				 
		graph = new JoinGraphImpl();
		Stream startPoint = null;
		Hashtable duplCheck = new Hashtable();
		 
		for (int i = 0; i < operators.size(); i++)
		{		
			ArrayList tmpStreams = new ArrayList();		
			OptimizerOp op = (OptimizerOp)operators.get(i);
			 			 
			for(int j = 0;j < op.getInputs().size(); j++)
			{			
				OptimizerOpInput input = (OptimizerOpInput)op.getInputs().get(j);				 
			
				Stream s = new Stream(input.getInputName());			
				if(startPointName.equalsIgnoreCase(input.getInputName()))
				{

					startPoint = s;
				}
				 
				s.setInputRate(input.getCardinality());
				 
				if (!duplCheck.containsKey(input.getInputName()))
				{				
					graph.addVertex(s);				
					duplCheck.put(input.getInputName(), input);				 
				}				 
				tmpStreams.add(s);			 
			}
			try
			{			
				//add the edge			
				graph.addEdge((Stream)tmpStreams.get(0),
						 	  (Stream)tmpStreams.get(1),
						 	   op.getOpSelectivity());
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}
		 }	
		 return startPoint;
	 }
	 
	//----------------------------------------------------------------------------------------------------
	
	 public QueryPlan GetPlan(String startInputName)
	 {
		 try
		 {
			 System.out.println();
			 
			 Stream startVetix = SetupJoinGraph(startInputName);
			 //QueryPlan plan = MJoinDynamicAlgorithm.order(graph, startVetix);
			 QueryPlan plan = bGreedyAlgorithmForward.order(graph, startVetix);
			 //plan.printXml();						 
			 return plan;
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		
		return null;
	 }
	 
	//----------------------------------------------------------------------------------------------------
	 
}
