//--------------------------------------------------
// SimulatedAnnealingSearch.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2007
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.DecisionTreeID3;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.QueryMeshOptimizer.QM_SEARCH_START_SOLUTION;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;


public class SimulatedAnnealingSearch 
{
	Hashtable CACHED_ROUTES = new Hashtable();
	private ArrayList minTrainingTuplePartitions;
	private LocalQueryMesh minQM;
	private double minQMCost = Double.MAX_VALUE;
	private double currentQMCost;
			
	DecisionTreeID3 currDT = null;
	ArrayList currRoutes = new ArrayList();
	Hashtable hashClassifierRoutesMapping = null;
	
	
	public SimulatedAnnealingSearch()
	{
		hashClassifierRoutesMapping = new Hashtable();
		minTrainingTuplePartitions = new ArrayList();
		
		//initialize cost as positive infinity
		minQMCost = Double.POSITIVE_INFINITY;
		currentQMCost = Double.POSITIVE_INFINITY;
	}
	
	//------------------------------------------------------------------
	//  QUERY MESH SIMULATED ANNEALING
	//------------------------------------------------------------------
	/*
	 1. Start-up:
		Pick initial query mesh solution QM randomly
		Choose an initial (high) temperature t > 0
		Choose a value for p, the rate of cooling parameter
	 2.	Choose a random neighbour of QM and call it QM
	 3.	Calculate the difference in query meshes costs: 
	 4. delta = cost(QM)-cost(QM) // Decide if accept the new query mesh or not 
	 5.	if delta <= 0 then
	 6.	  QM = QM  //QM is better than QM or the same as QM
	 7.	else
	 8.	  QM = QM with probability e^delta t
	 9.	end if
	 10.	if stopping condition is met then
	 11.   exit with QM a final solution
	 12. else
	 13.   reduce the temperature by setting temp=rho*temp, and go to Step 2
	 14. end if	 
	 *  
	 */
	
	SimpleQueryPlanSpecification queryPlan;
	DecisionTreeID3 dt;
	double totalQueryMeshCost = 0.0;
	int numTrainingTuples;
	S<String> s;
	ArrayList allPartitions;
	
	
	//-----------------------------------------------------------------
	//                     CHOOSE INITIAL TEMPERATURE
	//-----------------------------------------------------------------
	double temperature = 10;
	
	//-----------------------------------------------------------------
	//                     CHOOSE RATE OF COOLING
	//-----------------------------------------------------------------
	double coolingParameter = 0.5;

	
	public LocalQueryMesh runSA(QM_SEARCH_START_SOLUTION startSolutionType,
								SimpleOperatorInputSpecification input, 
								ArrayList trainingTupleSet, 
								ArrayList partitionSpace, 
								List xmlOperators, 
								Hashtable xmlInputs,
								boolean isRandomStartSolution)
	{
		int numIterations = 1;
		java.util.Enumeration enl;
		//initialize query mesh for the current stream
		LocalQueryMesh streamQM = new LocalQueryMesh(input.getId());
				
		Hashtable hashTupleAttrTitles = new Hashtable();
		try
		{
			ArrayList partition = null;
			Object [] currSubsets = null;
			
			//array list into which training tuples will be placed per solution found
			ArrayList trainingTuplePartitions = new ArrayList();
			
			//num of trianing tuples
			numTrainingTuples = trainingTupleSet.size() - 1; //subtracting 1 because we expect a title as the very 1st element
			
			BigInteger searchSpace = Bell.bigNumber(numTrainingTuples);		
			
			//verify that the #of partitions matches the bell number
			//assert searchSpace.intValue() == partitionSpace.size();
		
			String currPartitionString = "";	
			Hashtable hashTrainingTupleMapping = new Hashtable();
					
			int count = 0;
			java.util.Enumeration en;
			S currSubset;
			
			//get the titles of the tuple attributes				
			String[] tokensTrainingTupleSchema = (String [])trainingTupleSet.get(0);
			
			String temp = tokensTrainingTupleSchema[0];
			tokensTrainingTupleSchema = temp.split("  ");
			
			for(int p = 0; p<tokensTrainingTupleSchema.length; p++)
			{
				hashTupleAttrTitles.put(tokensTrainingTupleSchema[p], ""+ p);
			}
			
			//----------------------------------------------------------
			// 1. PICK A STARTING SOLUTION (BASED ON TYPE SPECIFIED BY USER)
			//----------------------------------------------------------
			switch (startSolutionType)
			{
				case RANDOM:
					//------------------------- RANDOM START SOLUTION---------------------------------
					//pick a start solution in the "middle" of the search space		
					BigInteger startSolution = Bell.bigNumber(numTrainingTuples).divide(new BigInteger("2"));		
					
					//now get the partitions of the solution
					int[] solutionPartitions = Bell.bigGrowthString(startSolution, numTrainingTuples);
					
					ArrayList partitions = new ArrayList();				
					
					for(int m = 0; m < solutionPartitions.length; m++)
					{
						S currPartition = new S();
						String pid = "" + solutionPartitions[m];
						if (hashTrainingTupleMapping.containsKey(pid))
						{
							currPartition = (S)hashTrainingTupleMapping.get(pid);
						}
						else
						{
							currPartition = new S();
						}
						currPartition.add(m + 1);
						hashTrainingTupleMapping.put(pid, currPartition);
						currPartitionString += solutionPartitions[m] + ",";
					}
					
					System.out.println("------------------------------------------------");
					System.out.println("Current partition {" + currPartitionString + "}");
					//print partitions in a different (maybe more understandable) format
					
					currSubsets = new Object[hashTrainingTupleMapping.size()];
					System.out.print("Current partition ");
					count = 0;
					en = hashTrainingTupleMapping.elements(); 
					while (en.hasMoreElements()) 
					{
						currSubset = (S)en.nextElement();
						currSubsets[count++] = currSubset; 
						System.out.print(currSubset); 
					}
					//-------------------------END RANDOM START SOLUTION---------------------------------
					
					break;
				case EXTREME_1_ROUTE:
					//------------------------- EXTREME 1 START SOLUTION---------------------------------
										
					currSubsets = new Object[1];
					System.out.print("Current partition ");
					count = 0;
					
					currSubset = new S();
					for(int m = 1; m < trainingTupleSet.size(); m++)
					{
						currSubset.add(m);						
					}// end for m

					currSubsets[0] = currSubset; 
					
					//------------------------- END EXTREME 1 START SOLUTION------------------------------
					break;
				case ROUTE_DRIVEN:										
					//--------------------------------------------------------------------
					// INFORMATION NEEDED FOR COMPUTING ROUTES' COST
					//--------------------------------------------------------------------					
					ArrayList inputs;
					double selectivity;
					Hashtable hashPreparedInputs = new Hashtable();
			
					Enumeration e = xmlInputs.elements();
					while (e.hasMoreElements()) 
					{  
						SimpleOperatorInputSpecification currInput = (SimpleOperatorInputSpecification)e.nextElement();  // do whatever you need with x
						if (currInput != input)
						{
							//only 'prepare' input if it is NOT the current input operator
							
							//compute cardinality of the input
							double currCardinality = currInput.getCardinality();
							OptimizerOpInput preparedInput = new OptimizerOpInput(currInput.getId(), currCardinality);
							hashPreparedInputs.put(currInput.getId(), preparedInput);
						}
					}

					//------------------------- ROUTE-DRIVEN START SOLUTION---------------------------------					
					//1. first start with individual tuples, just like in the extreme-N solution
					currSubsets = new Object[trainingTupleSet.size()];
					System.out.print("Current partition ");
					count = 0;
										
					for(int m = 0; m < trainingTupleSet.size(); m++)					
					{
						currSubset = new S();
						currSubset.add(m);
						currSubsets[count++] = currSubset;	
					}// end for m
										
					String tabs = "\t";
					
					//2. then compute routes for each individual tuple	
					Hashtable hashTrainingTupleRoutes = new Hashtable();
					
					for(int k = 0; k < currSubsets.length; k++)
					{			    						
						QueryPlan currPlan;
						//if route is NOT cached					
						String subsetName = currSubsets[k].toString();
						subsetName = subsetName.replace("[", "");
						subsetName = subsetName.replace("]", "");
						
						int currCardinality = ((S)currSubsets[k]).size();
						OptimizerOpInput preparedInput = new OptimizerOpInput(input.getId(), currCardinality);
						hashPreparedInputs.put(input.getId(), preparedInput);
						
				    	//------------------------------------------------------------------
				    	// ALL INPUTS ARE PREPARED
				    	//------------------------------------------------------------------
				
						ArrayList partitionTuples = input.getPartitionTuples((S)currSubsets[k]);
						System.out.println("");
				    	//------------------------------------------------------------------
				    	// NOW PREPARE OPERATORS
				    	//------------------------------------------------------------------
						ArrayList allOperators = new ArrayList();
				    	for(int b = 0; b < xmlOperators.size(); b++)
				    	{
				    		SimpleOperatorSpecification currOp = (SimpleOperatorSpecification)xmlOperators.get(b);
				    		ArrayList currOpInputs = new ArrayList();
				    			
				    		boolean containsMeshInput = false; //true if input to operator is the stream for which we are computing query mesh
				    			
				    		//assert 2 inputs
				    		assert currOpInputs.size() == 2;
				    			
				    		//-------------------------------------------------------------
				    		// 1st OP INPUT
				    		//-------------------------------------------------------------
				    		boolean isPartitionBasedEstimation1 = false;
				    		boolean isPartitionBasedEstimation2 = false;
				    		String inputId1 = (String)currOp.getInputs().get(0);
				    		ArrayList op1PartitionTuples = new ArrayList();
				    		if (inputId1.equalsIgnoreCase(input.getId()))
				    		{
				    			isPartitionBasedEstimation1 = true;
				    			op1PartitionTuples = partitionTuples;
				    		}	        				
				    
				    		//retrieve op by id
				    		OptimizerOpInput inp1 = (OptimizerOpInput)hashPreparedInputs.get(inputId1);
				    		SimpleOperatorInputSpecification input1 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId1);						
				    		int joinAttributeIdx1 = currOp.getJoinAttributeIdxByInputId(inputId1);
				    		currOpInputs.add(inp1);
				    		
				    		//-------------------------------------------------------------
				    		// 2nd OP INPUT
				    		//-------------------------------------------------------------	        			
				    		String inputId2 = (String)currOp.getInputs().get(1);
				    		ArrayList op2PartitionTuples = new ArrayList();
				    		if (inputId2.equalsIgnoreCase(input.getId()))
				    		{
				    			isPartitionBasedEstimation2 = true;
				    			op2PartitionTuples = partitionTuples;
				    		}	        				
				    		
				    		//retrieve op by id
				    		OptimizerOpInput inp2 = (OptimizerOpInput)hashPreparedInputs.get(inputId2);
				    		SimpleOperatorInputSpecification input2 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId2);
				    		int joinAttributeIdx2 = currOp.getJoinAttributeIdxByInputId(inputId2);	        			
				    		currOpInputs.add(inp2);
				    			
				    		//now estimate selectivity of the operator
				    				        			
				    		double currOpSelectivity = -999;
				    			
				    		JoinSelectivityEstimator selectivityEstimator = new JoinSelectivityEstimator(input1,
																						        		 isPartitionBasedEstimation1, 									
																						        		 joinAttributeIdx1,
																						        		 op1PartitionTuples,
																										 input2, 
																										 isPartitionBasedEstimation2,
																										 joinAttributeIdx2,
																										 op2PartitionTuples,
																										 hashTupleAttrTitles);
				    			
				    			
				    		currOpSelectivity = selectivityEstimator.getSelectivity();
				    		System.out.print(tabs + "Op id = " + currOp.getId() + ", w/ inputs: " + inputId1 + ", " + inputId2 + "; ");
				    		if ((isPartitionBasedEstimation1 == true) || (isPartitionBasedEstimation2 == true))
				    		{
				    			System.out.println(tabs + "Partition-based selectivity: " + currOpSelectivity);
				    		}
				    		else
				    		{
				    			System.out.println(tabs + "Stream-based selectivity: " + currOpSelectivity);
				    		}
				    		
				    		OptimizerOp op = new OptimizerOp(currOp.toString(), "join", currOpInputs, currOpSelectivity);
				    			
				    		allOperators.add(op);				    		
				    	}//end of operator iterations
				    		
				    	queryPlan = new SimpleQueryPlanSpecification(allOperators);
				
				    	//get the plan
				    	System.out.print(tabs + "Current stream: " + input.getId() + ": ");
				    	currPlan = queryPlan.GetPlan(input.getId());    			    			    		
				    	//---------------------------------------------------------
				    	//             CACHE THE ROUTE FOR THE FUTURE
				    	//---------------------------------------------------------
				    	if(hashTrainingTupleRoutes.containsKey(currPlan.toString()))
				    	{
				    		ArrayList currTuplesWithSameRoute = (ArrayList)hashTrainingTupleRoutes.get(currPlan.toString());
				    		currTuplesWithSameRoute.add(subsetName);
				    		hashTrainingTupleRoutes.put(currPlan.toString(), currTuplesWithSameRoute);
				    	}
				    	else
				    	{
				    		//first time this route was encountered
				    		ArrayList currTuplesWithSameRoute = new ArrayList();
				    		currTuplesWithSameRoute.add(subsetName);
				    		hashTrainingTupleRoutes.put(currPlan.toString(), currTuplesWithSameRoute);				    		
				    	}				    					    	
					}
					//3. group-by tuples by identical routes
					currSubsets = new Object[hashTrainingTupleRoutes.size()];
					count = 0;
					enl = hashTrainingTupleRoutes.elements(); 					
					while (enl.hasMoreElements()) 
					{
						ArrayList currTuplesWithSameRoute = (ArrayList)enl.nextElement();
						S tupleSetWithSameRoute = new S();
						for(int y = 0; y < currTuplesWithSameRoute.size(); y++)
						{
							tupleSetWithSameRoute.add(currTuplesWithSameRoute.get(y));
						}
						currSubsets[count++] = tupleSetWithSameRoute; 
					}					
					//------------------------- END ROUTE-DRIVEN START SOLUTION---------------------------------
					break;
				case CONTENT_DRIVEN:
					//------------------------- CONTENT-DRIVEN START SOLUTION---------------------------------
					//start solution is picked based on content of the data 
					//(default approach) to selecting start solution 
				
					//1. extract join attribute indexes in the schema of the current stream
					ArrayList currentStreamJoinIdxs = new ArrayList();
					Iterator it = xmlOperators.iterator();
					while(it.hasNext()) 
					{
						SimpleOperatorSpecification currOpt = (SimpleOperatorSpecification)it.next();
						//System.out.println(currOpt.toString());
						int attr_idx = currOpt.getJoinAttributeIdxByInputId(input.getId());
						
						if (attr_idx != -1)
						{
							currentStreamJoinIdxs.add(""+attr_idx);
						}						
					}
					
					//get the titles of the tuple attributes				
					/*String[] tokensTrainingTupleSchema = (String [])trainingTupleSet.get(0);
					for(int p = 0; p<tokensTrainingTupleSchema.length; p++)
					{
						hashTupleAttrTitles.put(tokensTrainingTupleSchema[p], ""+ p);
					}
					*/
					
					//2. group tuples by distinct values of the join attributes
					ArrayList arrHashesJoinAttributes = new ArrayList();
					ArrayList arrHashesTrainingTupleIndexes = new ArrayList();
					
					//retrieve the mapping between the tuples in the training set 
					//and the actual schema of the tuples
					
					for(int p = 0; p < currentStreamJoinIdxs.size(); p++)
					{
						int attr_idx = Integer.parseInt((String)currentStreamJoinIdxs.get(p));
						Hashtable hashJoinAttrGroups = new Hashtable();
						Hashtable hashTrainTuplesIdxsByGroup = new Hashtable();
																
						String attr_name = input.getAttrNameByIndex(attr_idx);
						
						String str_transl_attr_idx = (String)hashTupleAttrTitles.get(attr_name);
						int transl_attr_idx = Integer.parseInt(str_transl_attr_idx);
						
						for(int m = 1; m < trainingTupleSet.size(); m++)
						{
							String[] tokens = (String [])trainingTupleSet.get(m);
							
							String tempTokens = tokens[0];
							tokens = tempTokens.split("  ");
							
							if (!hashJoinAttrGroups.containsKey(tokens[transl_attr_idx]))
							{
								hashJoinAttrGroups.put(tokens[transl_attr_idx], "1");							
								S attrTrainTuplesIdxsByGroup = new S();
								attrTrainTuplesIdxsByGroup.add(m);
								hashTrainTuplesIdxsByGroup.put(tokens[transl_attr_idx], attrTrainTuplesIdxsByGroup);
							}
							else
							{
								String strCounter = (String)hashJoinAttrGroups.get(tokens[transl_attr_idx]);
								int counter = Integer.parseInt(strCounter);
								counter++;
								//update the counter
								hashJoinAttrGroups.put(tokens[transl_attr_idx], ""+counter);
								S attrTrainTuplesIdxsByGroup = (S)hashTrainTuplesIdxsByGroup.get(tokens[transl_attr_idx]);
								attrTrainTuplesIdxsByGroup.add(m);
							}
						}// end for m
						arrHashesJoinAttributes.add(hashJoinAttrGroups);
						arrHashesTrainingTupleIndexes.add(hashTrainTuplesIdxsByGroup);
					}//end for p
					System.out.println("------------------------------------------------");
					
					Hashtable hashTraingTuplesIdxsByGroup = (Hashtable)arrHashesTrainingTupleIndexes.get(0);
					en = hashTraingTuplesIdxsByGroup.elements(); 
					while (en.hasMoreElements()) 
					{
						currSubset = (S)en.nextElement();
						System.out.println(currSubset);
					}
					
					currSubsets = new Object[hashTraingTuplesIdxsByGroup.size()];
					System.out.print("Current partition ");
					count = 0;
					enl = hashTraingTuplesIdxsByGroup.elements(); 
					while (enl.hasMoreElements()) 
					{
						currSubset = (S)enl.nextElement();
						currSubsets[count++] = currSubset; 
						System.out.print(currSubset); 
					}
					//------------------------- END CONTENT-DRIVEN START SOLUTION---------------------------------
					break;				
			}//end switch - done with start solution
									
			System.out.println();
			System.out.println("------------------------------------------------");
	
			
			for(int m = 0; m < currSubsets.length; m++)
			{
				S currPartition = new S();
				String pid = "" + currSubsets[m];
				if (hashTrainingTupleMapping.containsKey(pid))
				{
					currPartition = (S)hashTrainingTupleMapping.get(pid);
				}
				else
				{
					currPartition = new S();
				}
				currPartition.add(m + 1);
				hashTrainingTupleMapping.put(pid, currPartition);
				currPartitionString += currSubsets[m] + ",";
			}
			
			
			//-----------------------------------------------------------------
			//         2. DO DT AND ROUTES; PROCESSING
			//-----------------------------------------------------------------						
						
			System.out.println("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
			System.out.println((numIterations++) + ") Iteration");
			
			currentQMCost = ComputeDTAndRoutes(currSubsets, 
											   input, 
											   xmlOperators, 
											   xmlInputs,											   
											   hashTupleAttrTitles);
			
		
			//-----------------------------------------------------------------
			//         3. IF CURRENT QM HAS THE SMALLES COST, STORE IT
			//-----------------------------------------------------------------						
			if (currentQMCost < minQMCost)
			{
				streamQM.setDTClassifier(currDT);
				streamQM.setRoutes(currRoutes);
				streamQM.setHashClassifierRoutesMapping(hashClassifierRoutesMapping);				
				minQMCost = currentQMCost;
				minQM = streamQM;
			}
		
			//-----------------------------------------------------------------
			//               4. RUN UNTIL STOP CONDITION
			//-----------------------------------------------------------------
			int stopCondition = 10; //10 times the cost of min QM has not changed is the STOPPING CONDITION
			int sameCount = 0;
		
			 while (sameCount < stopCondition) 
			 {							 
				System.out.println("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
				System.out.println((numIterations++) + ") Iteration");
				//-----------------------------------------------------------------
				//      5. PICK A RANDOM NEIGHBOR OF THE CURRENT SOLUTION
				//-----------------------------------------------------------------
				//currSubsets = GetRandomNeighbor(hashTrainingTupleMapping);
							
				//-----------------------------------------------------------------
				//      6. ESTIMATE THE COST OF THE RANDOMLY PICKED NEIGHBOR
				//-----------------------------------------------------------------
				
				currDT = null;
				currRoutes = new ArrayList();
				
				currentQMCost = ComputeDTAndRoutes(currSubsets, 
												   input, 
												   xmlOperators, 
												   xmlInputs,  
												   hashTupleAttrTitles);
				//-----------------------------------------------------------------
				//      7. CALCULATE DIFFERENCE IN QUERY MESHES COSTS
				//-----------------------------------------------------------------
				double delta = currentQMCost - minQMCost;
				
				//-----------------------------------------------------------------
				//      8. UPDATE MIN COST QUERY MESH
				//-----------------------------------------------------------------				
				if (delta < 0)
				{
					minQMCost = currentQMCost;
					minQM = streamQM;
					sameCount = 0; //reset the same count if solution is smaller in cost
				}
				else
				{
					 // QM = QM with probability e^(-delta/temperature)										
				    if (anneal(delta))
					{
						minQMCost = currentQMCost;
						minQM = streamQM;
						sameCount = 0; //reset the same count if solution is smaller in cost
					}
				    else
				    {
				    	//increment same count
				    	sameCount++;
				    }
				}
				//-----------------------------------------------------------------
				//      9. REDUCE TEMPERATURE BY COOLING RATE 
				//-----------------------------------------------------------------				
				
				temperature = coolingParameter * temperature;				
			}	
			 
			streamQM.setDTClassifier(currDT);
			streamQM.setRoutes(currRoutes);
		}
		catch(Exception exp)
		{
			exp.printStackTrace();
		}
		
		QueryMeshOptimizer.hashNumIterationsPerStream.put(input.getInputName(), numIterations - 1);
		
		return streamQM;
	}

			
	  
	  //---------------------------------------------------------------
	  // Called to determine if annealing should take place.
	  //---------------------------------------------------------------
	  public boolean anneal(double d)
	  {	    
		  if (temperature < 1.0E-4) 	   
		  {
			  if (d > 0.0)	       
				  return true;	      
			  else	        
				  return false;	    
		  }
	    
		  if (Math.random() < Math.exp(d / temperature))
			  return true;	    
		  else	      
			  return false;	  
	  }
	  
	  public String getRusterName(String trainingTupleId, Hashtable trainingTuplesToSubsetsMapping)
	  {
		  String tabs = "\t";
		  String rusterName = "";
		  boolean foundRuster = false;
		  
		  Enumeration e = trainingTuplesToSubsetsMapping.elements();
		  while (e.hasMoreElements()) 
		  {  			  
			  String currRusterName = (String)e.nextElement();
			  String name = currRusterName;
			  currRusterName = currRusterName.replace("ruster", "");
			  currRusterName = currRusterName.replace("[", "");
			  currRusterName = currRusterName.replace("]", "");
			  String [] tokens = currRusterName.split(",");
			  for (int p = 0; p < tokens.length; p++)
			  {
				  String element = tokens[p].replace(" ", "");
				  if (element.equalsIgnoreCase(trainingTupleId))
				  {
					  rusterName = name;
					  return rusterName;
				  }
			  }
		  }
		  return rusterName;
	  }
		
		public double ComputeDTAndRoutes(Object [] currSubsets, 
										 SimpleOperatorInputSpecification input,
										 List xmlOperators, 
										 Hashtable xmlInputs,
										 Hashtable hashTupleAttrTitles)
		{

			String tabs = "\t";
			//reset mapping between DT leaves and the routes
			hashClassifierRoutesMapping = new Hashtable();
			
			//---------------------------------------------------------------
		    // CREATE MAPPING BETWEEN TRAINING TUPLES AND SUBSETS IN THE CURRENT
			// PARTITION
			//---------------------------------------------------------------
			Hashtable trainingTuplesToSubsetsMapping = new Hashtable();
			try
			{
				for(int p = 1; p <= numTrainingTuples; p++)
				{
					String trainingTupleId = "" + p;
					for(int i = 0; i < currSubsets.length; i++)
					{
						//System.out.println("value of i: " + i);
						String currRusterName = currSubsets[i].toString();
												
						currRusterName = currRusterName.replace("[", "");
						currRusterName = currRusterName.replace("]", "");
						String [] tokens = currRusterName.split(",");
						
						boolean found = false;
						for (int m = 0; m < tokens.length; m++)
					    {
						     String element = tokens[m].replace(" ", "");
							 if (element.equalsIgnoreCase(trainingTupleId))
							 {								 
								 //create mapping between training tuple id and the subset 
								 String rusterName = "ruster"+currSubsets[i].toString();
								 trainingTuplesToSubsetsMapping.put(trainingTupleId, rusterName);
								 found = true;
								 break;								  
							  }
						 }

						if (found)
						{
							break;
						}
					}
				}
			}
			catch(Exception ex)
			{
				ex.printStackTrace();
			}
			//-----------------------------------------------------------------
			//         2. ESTIMATE THE COST FOR THE START SOLUTION
			//-----------------------------------------------------------------
			
				//compute decision tree where target attribute is the partition of the tuple
				String inputFileName = "C:\\QueryMesh\\tmp\\qmInput"+input.getId()+".txt";
				String outputFileName = "C:\\QueryMesh\\tmp\\qmOutput"+input.getId()+".txt";
				ArrayList trainingTuples = input.getTrainingTuples();
				for (int i=0; i < trainingTuples.size(); i++)
				{
					String[] stringTokens = (String[])trainingTuples.get(i);
					String stringToWrite = "";
					for(int j=0; j < stringTokens.length; j++)
					{   
						if (j == stringTokens.length - 1)
						{
							if (i > 0)
							{								
		    					//retrieve the ruster (partition) name for the current training tuple
								String key = ""+i;
								
		    					String rusterName = (String)trainingTuplesToSubsetsMapping.get(key);
		    					//String rusterName = getRusterName(key, trainingTuplesToSubsetsMapping);
		    					//append the ruster as the last attribute (it will be a target attribute
		    					//in the decision tree construction
//		    					stringToWrite += stringTokens[j] + "  "+ rusterName;
		    					stringToWrite += stringTokens[j] + rusterName;
							}
							else
							{
								stringToWrite += stringTokens[j] + "  " + "RusterName";
//								stringToWrite += stringTokens[j] +"RusterName";
							}
						}
						else
						{
							stringToWrite += stringTokens[j] + "  ";
						}
					}        			
					
					if(i == 0)
					{
						//delete sample dump (from previous runs)
						java.io.File file = new java.io.File (inputFileName);
				        file.delete ();
				        DecisionTreeID3.WriteToDTDumpFile(inputFileName, stringToWrite);						
				        DecisionTreeID3.WriteToDTDumpFile(inputFileName, "//**************************************");
					}
					else
					{
						DecisionTreeID3.WriteToDTDumpFile(inputFileName, stringToWrite);						
					}
				}//end for
			
				//--------------------------------------------------------------------
				// COMPUTE DECISION TREE COST
				//--------------------------------------------------------------------				
				dt = null;
				dt = DecisionTreeID3.optimizerCreateDT(input, inputFileName, outputFileName);
				double dtCost = dt.getDTCost(input.getStatisticsFileName());
		
				//set the current dt to the newly computed dt
				currDT = dt;
				
				//--------------------------------------------------------------------
				// COMPUTE ROUTES' COST
				//--------------------------------------------------------------------
				
				ArrayList inputs;
				double selectivity;
				Hashtable hashPreparedInputs = new Hashtable();
		
				Enumeration e = xmlInputs.elements();
				while (e.hasMoreElements()) 
				{  
					SimpleOperatorInputSpecification currInput = (SimpleOperatorInputSpecification)e.nextElement();  // do whatever you need with x
					if (currInput != input)
					{
						//only 'prepare' input if it is NOT the current input operator
						
						//compute cardinality of the input
						double currCardinality = currInput.getCardinality();
						OptimizerOpInput preparedInput = new OptimizerOpInput(currInput.getId(), currCardinality);
						hashPreparedInputs.put(currInput.getId(), preparedInput);
					}
				}
				
				//------------------------------------------------------------------
				// ITERATE THROUGH POSSIBLE RUSTERS OF THE CURRENT INPUT STREAM
				// AND COMPUTE THE ROUTES' COSTS
				//------------------------------------------------------------------
				
				double totalRoutesCost = 0;
				ArrayList rt = new ArrayList();
				for(int k = 0; k < currSubsets.length; k++)
				{
		    		//now (for every leaf of the decision tree) of the current input 
		    		//compute route cost
					String partitionName = "ruster"+currSubsets[k].toString();
					System.out.println();
					System.out.print(tabs + partitionName + ": "); 
					
					QueryPlan currPlan;
					//if route is NOT cached					
					
					if (!CACHED_ROUTES.containsKey(partitionName))
					{						
						//int currCardinality = input.computeCardinalityOfPartition(partitionName, outputFileName);
						int currCardinality = ((S)currSubsets[k]).size();
						OptimizerOpInput preparedInput = new OptimizerOpInput(input.getId(), currCardinality);
						hashPreparedInputs.put(input.getId(), preparedInput);
					
			    		//------------------------------------------------------------------
			    		// ALL INPUTS ARE PREPARED
			    		//------------------------------------------------------------------
			
						ArrayList partitionTuples = input.getPartitionTuples((S)currSubsets[k]);
						System.out.println("");
			    		//------------------------------------------------------------------
			    		// NOW PREPARE OPERATORS
			    		//------------------------------------------------------------------
						ArrayList allOperators = new ArrayList();
			    		for(int b = 0; b < xmlOperators.size(); b++)
			    		{
			    			SimpleOperatorSpecification currOp = (SimpleOperatorSpecification)xmlOperators.get(b);
			    			ArrayList currOpInputs = new ArrayList();
			    			
			    			boolean containsMeshInput = false; //true if input to operator is the stream for which we are computing query mesh
			    			
			    			//assert 2 inputs
			    			assert currOpInputs.size() == 2;
			    			
			    			//-------------------------------------------------------------
			    			// 1st OP INPUT
			    			//-------------------------------------------------------------
			    			boolean isPartitionBasedEstimation1 = false;
			    			boolean isPartitionBasedEstimation2 = false;
			    			String inputId1 = (String)currOp.getInputs().get(0);
			    			ArrayList op1PartitionTuples = new ArrayList();
			    			if (inputId1.equalsIgnoreCase(input.getId()))
			    			{
			    				isPartitionBasedEstimation1 = true;
			    				op1PartitionTuples = partitionTuples;
			    			}	        				
			    			//retrieve op by id
			    			OptimizerOpInput inp1 = (OptimizerOpInput)hashPreparedInputs.get(inputId1);
			    			SimpleOperatorInputSpecification input1 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId1);						
			    			int joinAttributeIdx1 = currOp.getJoinAttributeIdxByInputId(inputId1);
			    			currOpInputs.add(inp1);
			    			//-------------------------------------------------------------
			    			// 2nd OP INPUT
			    			//-------------------------------------------------------------	        			
			    			String inputId2 = (String)currOp.getInputs().get(1);
			    			ArrayList op2PartitionTuples = new ArrayList();
			    			if (inputId2.equalsIgnoreCase(input.getId()))
			    			{
			    				isPartitionBasedEstimation2 = true;
			    				op2PartitionTuples = partitionTuples;
			    			}	        				
			    			//retrieve op by id
			    			OptimizerOpInput inp2 = (OptimizerOpInput)hashPreparedInputs.get(inputId2);
			    			SimpleOperatorInputSpecification input2 = (SimpleOperatorInputSpecification)xmlInputs.get(inputId2);
			    			int joinAttributeIdx2 = currOp.getJoinAttributeIdxByInputId(inputId2);	        			
			    			currOpInputs.add(inp2);
			    			
			    			//now estimate selectivity of the operator
			    				        			
			    			double currOpSelectivity = -999;
			    			
			    			
			    			JoinSelectivityEstimator selectivityEstimator = new JoinSelectivityEstimator(
																					        			input1,
																					        			isPartitionBasedEstimation1, 									
																					        			joinAttributeIdx1,
																					        			op1PartitionTuples,
																										input2, 
																										isPartitionBasedEstimation2,
																										joinAttributeIdx2,
																										op2PartitionTuples,
																										hashTupleAttrTitles);
			    			
			    			
			    			currOpSelectivity = selectivityEstimator.getSelectivity();
			    			System.out.print(tabs + "Op id = " + currOp.getId() + ", w/ inputs: " + inputId1 + ", " + inputId2 + "; ");
			    			if ((isPartitionBasedEstimation1 == true) || (isPartitionBasedEstimation2 == true))
			    			{
			    				System.out.println(tabs + "Partition-based selectivity: " + currOpSelectivity);
			    			}
			    			else
			    			{
			    				System.out.println(tabs + "Stream-based selectivity: " + currOpSelectivity);
			    			}
			    			OptimizerOp op = new OptimizerOp(currOp.toString(), "join", currOpInputs, currOpSelectivity);
			    			
			    			allOperators.add(op);
			    		}//end of operator iterations
			    		
			    		queryPlan = new SimpleQueryPlanSpecification(allOperators);
			
			    		//get the plan
			    		System.out.print(tabs + "Current stream: " + input.getId() + ": ");
			    		currPlan = queryPlan.GetPlan(input.getId());    			    			    		
			    		//---------------------------------------------------------
			    		//             CACHE THE ROUTE FOR THE FUTURE
			    		//---------------------------------------------------------
			    		CACHED_ROUTES.put(partitionName, currPlan);
			    		
					}
					else
					{
						//retrieve route from cache
						currPlan = (QueryPlan)CACHED_ROUTES.get(partitionName);							
					}
										
					double routeCost =  currPlan.getCost();					
					System.out.println(", Route cost:"+ routeCost);		
			    	
		    		//add the plan to the curr routes list
		    		rt.add(new RouteSignature(currPlan, partitionName));
		    		
		    		totalRoutesCost += routeCost;
				}//end of subsest iterations
				
				currRoutes = rt;
				totalQueryMeshCost = dtCost + totalRoutesCost;
				System.out.println(tabs + "Curent query mesh cost: " + totalQueryMeshCost);
				return totalQueryMeshCost;
			}

		
		private Object [] GetRandomNeighbor(Hashtable hashTrainingTupleMapping)
		{
			Object [] currSubsets = null;
			currSubsets = HybridSearch.GetRandomNeighborMergeTwoSubsets(hashTrainingTupleMapping);
			
			return currSubsets;
		}
					
}
