//--------------------------------------------------
// TrainingTupleCollectorOperatorImp.java
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2008
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.MotionStatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.PositionSpt2D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.Time;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.ClassType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.EntityType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.QueryType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.enums.UpdateType;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GenericSpatioTemporalGrid;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.grid.GridCellIndex3D;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.MotionGenericOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.operators.scuba.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.results.MotionDiscreteResult;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.states.discrete.MotionDiscreteState;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.BasicHist;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Coin;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.DecisionTreeHeader;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.HistogramBuilderOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.StringBasicHist;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.DistributedExecutionController;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

public class TrainingTupleCollectorOperatorImp extends XATSingleQueueStreamOperatorImp implements HistogramBuilderOperator 
{
	//-------------------------------------------------------------------------------------

   	protected Time currentTime;			
	protected String samplingHeuristic;	
	protected int countTuplesToSample; 
	String sampleDumpFile;
	
	public Hashtable histogramsTable = new Hashtable();	
	public Hashtable usedNotUsedAttributes = new Hashtable();
	
	String attributesHeader = "";
	long lastUpdateTime = -999;
	boolean bDoSample = true; //set to false when sample is collected, but not yet been used
	protected ArrayList oldTuplesSample = new ArrayList();
	protected ArrayList newTuplesSample = new ArrayList();	
	
	ArrayList decisionTreeAttributes; //array list that contains the names of the attributes in the stream
	
	//String outFile = "C:\\Rimma\\Research\\QueryMesh\\execution_stats\\DATA_MONITORING.txt";
	
	//-------------------------------------------------------------------------------------
	
	public TrainingTupleCollectorOperatorImp(XATNode init) 
	{
		super(init);		
	}
	
	//-------------------------------------------------------------------------------------
		
	static int numWindowsAlreadySampled = 0;
	
	//-------------------------------------------------------------------------------------
	
	public boolean run() throws Exception 
	{	
		int tuplesDQ = 0;
		int tuplesEQ = 0;
		
		if(this.getDebug())
		{
			System.out.println("Running: " + this.toToolTipString() + "\n");
		}
				
		for (int i = 0; i < this.getOutputQueueCount();i++)
		{
			if(this.getOutputQueue(i).getSchema() == null)
			{
				init();
			}
		}	

		//check if should do sampling		
		if (newTuplesSample.size() == countTuplesToSample) 
		{	
			System.out.println("Skipping sampling, just forwarding the tuples...");
			bDoSample = false;			
		}
		
		double timeSamplingStarted = System.currentTimeMillis();
		
		XATTuple inputTuple = null;
		int len = 10000;
		System.out.println("input stream id (queue customer): " + this.getInputQueue().getQueueName());
		System.out.println("total # of tuples in the queue: " + this.getInputQueue().size());
		XATTuple[] tuples = this.getInputQueue().dequeue(len, this);				
		System.out.println("# of tuples dequeued: " + tuples.length);		
		XATTuple[] sampleWindow = new XATTuple[tuples.length];
		
		//------------------------------------------------------
		// PROPAGATE TUPLES
		//------------------------------------------------------
		int countPropagated = 0;
		for (int p = 0; p < tuples.length; p++)
		{			
			inputTuple = tuples[p];
			
			if (inputTuple.isStqmPunct())
			{
				//stqm punctuations are "triggers" for distribution change in the stream - we will not propogate them further
				boolean shouldRunStqm = true;
				System.out.println("Received STQM punctuation");
			}
			else
			{
				sampleWindow[p] = inputTuple;
				//propagate the tuple to the parent operator
				this.getOutputQueue().enqueue(inputTuple, this);	
				countPropagated++;
			}
		}
		System.out.println("# of tuples propagated: " + tuples.length);				
		System.out.println("# of tuples to sample: " + countTuplesToSample);	
		System.out.println("---- BEFORE SAMPLING ----");
		System.out.println("oldTuplesSample = " + oldTuplesSample.size());
		System.out.println("newTuplesSample = " + newTuplesSample.size());
		
		//------------------------------------------------------
		// DO SAMPLE BASED ON HEURISTIC
		//------------------------------------------------------	
		STQM_Monitor stqmMonitor = new STQM_Monitor();
		System.out.println("Sampling method: " + samplingHeuristic);
		ArrayList currRunNewTuplesSample = new ArrayList();
		
		if (samplingHeuristic.equalsIgnoreCase("classifier_driven"))
		{		
			String strDecisionTreeHeader = CreateTrainingTupleFileHeader(decisionTreeAttributes, usedNotUsedAttributes);
			//classifier-driven sampling (see paper for details)
			String currentStreamName = this.getInputQueue().getQueueName();
			int currentStreamId = Integer.parseInt(currentStreamName.replace("Stream", ""));
			currRunNewTuplesSample = stqmMonitor.DoClassifierDrivenSample(currentStreamId,
																		  sampleWindow,
																		  oldTuplesSample,
																		  countTuplesToSample,
																		  DistributedExecutionController.OnlineClassifierRef,
																		  this);				
		}
		else if (samplingHeuristic.equalsIgnoreCase("route_driven"))
		{
			String currentStreamName = this.getInputQueue().getQueueName();
			int currentStreamId = Integer.parseInt(currentStreamName.replace("Stream", ""));
			currRunNewTuplesSample = stqmMonitor.DoRouteDrivenSample(currentStreamId,
																	 oldTuplesSample,
																	 sampleWindow, 
																	 countTuplesToSample,
																	 DistributedExecutionController.OnlineClassifierRef,
																	 DistributedExecutionController.hashSTQMRouteSignaturesGlobalStorage);

		}
		else if (samplingHeuristic.equalsIgnoreCase("simple_random"))
		{
			String currentStreamName = this.getInputQueue().getQueueName();
			int currentStreamId = Integer.parseInt(currentStreamName.replace("Stream", ""));
			currRunNewTuplesSample = stqmMonitor.DoRandomSample(sampleWindow, countTuplesToSample, usedNotUsedAttributes);			
		}

		lastUpdateTime = System.currentTimeMillis();
		newTuplesSample.addAll(currRunNewTuplesSample); //add the current sample to the overall new sample
		
		//------------------------------------------------------------------------
		// SHRINK THE SAMPLE SIZE IF NECESSARY
		//------------------------------------------------------------------------
		//check the current size of the new sample (remove the tuples from the beginning of the arr) if the size is larger than needed
		if (newTuplesSample.size() > countTuplesToSample)
		{
			int countNumTuplesToRemoveFromBeginningOfSample = newTuplesSample.size() - countTuplesToSample;
			for (int d = 0; d < countNumTuplesToRemoveFromBeginningOfSample; d++)
			{
				newTuplesSample.remove(0);
			}
		}		
		
		double timeSamplingEnded = System.currentTimeMillis();
		double timeSamplingDuration = timeSamplingEnded - timeSamplingStarted;
		double timePerTuple = timeSamplingDuration / tuples.length;
		String outString = "DATA MONITORING OVERHEAD: # tuples: \t" + tuples.length + "\t duration: \t" + timeSamplingDuration + "\t per tuple overhead: \t" + timePerTuple;
		//WriteToFile(outFile, outString);
		
		System.out.println("---- AFTER SAMPLING ----");
		System.out.println("oldTuplesSample = " + oldTuplesSample.size());
		System.out.println("newTuplesSample = " + newTuplesSample.size());
		return true;
	}

	//-------------------------------------------------------------------------------------
	
	public void MakeNewSampleToBeOldSample()
	{
		oldTuplesSample = (ArrayList)newTuplesSample.clone();
		newTuplesSample.clear();		
	}
	
	//-------------------------------------------------------------------------------------
	
	private void WriteToFile(String fileLocation, String stringToWrite)
	{
		try
		{
			//Create file 
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			out.newLine();
			//Close the output stream
			out.close();			    
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}
	 
	
	//-------------------------------------------------------------------------------------
	
	public String toToolTipString() 
	{
		return "Training Tuple Set Creator Operator";
	}
	
	//-------------------------------------------------------------------------------------
	
	public boolean hasUnmatchedState() 
	{
		return false;
	}
	
	//-------------------------------------------------------------------------------------
	
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return true;
	}
	
	//-------------------------------------------------------------------------------------
	
	public boolean cleanAllAccumulatedTuples() 
	{
		return true;
	}
	
	//-------------------------------------------------------------------------------------
	
	public int setStateID(int startID) 
	{
		return 0;
	}	
	
	//-------------------------------------------------------------------------------------
	
	public void setHistogramsTable(Hashtable newHistogramsTable) 
	{
		histogramsTable = newHistogramsTable;
	}	
	
	//-------------------------------------------------------------------------------------
	
	public void setSamplingHeuristic(String samplingHeuristicVal) 
	{
		samplingHeuristic = samplingHeuristicVal;
	}

	//-------------------------------------------------------------------------------------
	
	public void setTuplesToSample(int numTuplesToSampleVal) 
	{		
		countTuplesToSample = numTuplesToSampleVal;
	}
	
	//-------------------------------------------------------------------------------------
	
	public void setTupleSampleDumpFile(String sampleDumpFileVal) 
	{
		sampleDumpFile = sampleDumpFileVal;
	}

	//-------------------------------------------------------------------------------------
	
	public ArrayList GetDecisionTreeAttributes()
	{
		return decisionTreeAttributes;
	}
	
	//-------------------------------------------------------------------------------------
	
	public void SetDecisionTreeAttributes(ArrayList decisionTreeAttributesVal)
	{
		decisionTreeAttributes = decisionTreeAttributesVal;
	}
	
	//-----------------------------------------------------------------------------	
	
	public void setTupleSampleDumpFileHeader() 
	{
		String strSampleDumpFileHeader = CreateTrainingTupleFileHeader(decisionTreeAttributes, usedNotUsedAttributes);
		
		//WriteToSampleDumpFile(sampleDumpFile, strSampleDumpFileHeader);
		//WriteToSampleDumpFile(sampleDumpFile,"//**************************************");	
	}
	
	//-----------------------------------------------------------------------------
	
	public static String CreateTrainingTupleFileHeader(ArrayList decisionTreeAttributes, Hashtable usedNotUsedAttributes)
	{	
		String strSampleDumpFileHeader = "";
		String strTargetColumnName = "";
		
		for(int i = 0; i < decisionTreeAttributes.size();i++)
		{
			DecisionTreeHeader dth = (DecisionTreeHeader)decisionTreeAttributes.get(i);
			
			if(dth.useAttr)
			{				
				if(dth.isTarget)
				{
					//attribute is used, and it is THE TARGET
					usedNotUsedAttributes.put(i, "used+target");
					strTargetColumnName = dth.attrName;
				}
				else
				{
					//attribute is used, but is not the target
					usedNotUsedAttributes.put(i, "used");
					
					strSampleDumpFileHeader += dth.attrName;
					strSampleDumpFileHeader += "  ";					
				}				
			}
			else
			{
				//strSampleDumpFileHeader += "not_used_"+dth.attrName;	
				usedNotUsedAttributes.put(i, "not_used");
			}			
		}//end for
		
		//add the target column name
		strSampleDumpFileHeader += strTargetColumnName + System.getProperty("line.separator");;		
		strSampleDumpFileHeader += "//**************************************";
		
		return strSampleDumpFileHeader;
	}

	//-------------------------------------------------------------------------------------
			
	public boolean recomputeUnmatchedState() 
	{
		return true;
	}
	
	//-------------------------------------------------------------------------------------
	
	public boolean matchStates(XATStreamOperator op) 
	{
		return true;
	}
	
	//-------------------------------------------------------------------------------------
	
	public boolean finalizeOperation()
	{
		return true;
	}		

	//-------------------------------------------------------------------------------------
	
	public long GetLastUpdateTime()
	{
		return lastUpdateTime;
	}
	
	//-------------------------------------------------------------------------------------
	
	public void SetLastUpdateTime(long newUpdateTime)
	{
		lastUpdateTime = newUpdateTime;
	}

	//-------------------------------------------------------------------------------------

	public int getWindowSize() 
	{
		// TODO Auto-generated method stub
		return 0;
	}

	//-------------------------------------------------------------------------------------

	public void setWindowSize(int newSize) 
	{
		// TODO Auto-generated method stub	
	}
	
	//-------------------------------------------------------------------------------------
	
	//STQM
	public ArrayList GetOldTuplesSample()
	{
		return oldTuplesSample;
	}

	//-------------------------------------------------------------------------------------
	
	//STQM
	public void SetOldTuplesSample(ArrayList oldTuplesSampleVal)
	{
		oldTuplesSample = oldTuplesSampleVal;
	}
	
	//-------------------------------------------------------------------------------------
	
	//STQM
	public ArrayList GetNewTuplesSample()
	{
		return newTuplesSample;
	}

	//-------------------------------------------------------------------------------------
	
	//STQM
	public void SetNewTuplesSample(ArrayList newTuplesSampleVal)
	{
		newTuplesSample = newTuplesSampleVal;
	}
	
	//-------------------------------------------------------------------------------------
	
}
