package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.translation;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.AttributeProperties;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.SimpleOperatorSpecification;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.QueryMeshOptimizer.QM_TYPE;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.translation.ClassifierInputSignature.INPUT_TYPE;

public class OptimizerToExecutorTranslator 
{
	
	String generatedConfigFileName = "";
	static String NEW_LINE = System.getProperty("line.separator");
	static Hashtable onlineClassifierParentQueues = new Hashtable();

	public static void GenerateConfigFile(QM_TYPE qmType,
										  ArrayList xmlOperators,
			 							  String outputFileLocation,
										  String rootOpId,
										  ArrayList rootOpChildren,
										  ArrayList operatorsDetails,
										  OnlineClassifierOperatorSignature classifierSignature,
										  ArrayList localQMSignatures,
										  Hashtable hashAllLogicalToPhysicalOpCombinations,
										  Hashtable hashLogicalOperatorsWithInputsCombinations,
										  ArrayList trainingSetCollectorOperatorsDetails /*STQM*/)
	{
		String configStr = "";
		String qmTypeStr = "";
		
		//--------------------------------------------------------------------------------------
		// STQM: added after adaptive functionality to QM has been added
		//--------------------------------------------------------------------------------------
		if (qmType == QM_TYPE.ADAPTIVE)
		{
			qmTypeStr = "adaptive";
		}
		else
		{
			qmTypeStr = "static";
		}
		
		configStr += "<!-- Autogenerated QMesh Query Plan -->" + NEW_LINE;		
		configStr += "<queryplan qmtype=\""+qmTypeStr+"\">" + NEW_LINE;
		
		//-------------------------------------------------------------------------------------
		// DUMMY OPERATOR
		//-------------------------------------------------------------------------------------
		configStr += generateDummyOperatorString(rootOpId, rootOpChildren); //generate dummy operator
		System.out.println("Dummy operator...Done");
		//-------------------------------------------------------------------------------------
		// STEM OPERATOR
		//-------------------------------------------------------------------------------------
		for (int i=0; i < operatorsDetails.size();i++)
		{
			STEMOperatorSignature stemOp = (STEMOperatorSignature)operatorsDetails.get(i); 
			String opId = stemOp.GetOpId();
			String streamId = stemOp.GetStreamId();
			String probeStreamId = stemOp.GetProbeStreamId();
			String numTupleIndex = stemOp.GetNumTupleIndex();
			String stemStreamId = stemOp.GetStemStreamId();
			String numSTeMIndex = stemOp.GetNumSTEMIndex();			
			ArrayList childrenOps = stemOp.GetChildren();

			configStr += generateQMSTEMOperatorString(opId, 
													  streamId, 
													  probeStreamId,
													  numTupleIndex,
													  stemStreamId,
													  numSTeMIndex,
													  rootOpId,
													  childrenOps); //generate QM operator
		}
		System.out.println("STEM operators...Done");
		//-------------------------------------------------------------------------------------
		// STQM: TRAINING SET COLLECTOR OPERATORS
		//-------------------------------------------------------------------------------------
		for (int i = 0; i < trainingSetCollectorOperatorsDetails.size();i++)
		{
			TrainingSetCollectorOperatorSignature trainingOp = (TrainingSetCollectorOperatorSignature)trainingSetCollectorOperatorsDetails.get(i); 
			String opId = trainingOp.GetOpId();
			String parentOpId = trainingOp.GetParentOpId();
			String inputStreamName = trainingOp.GetInputStreamName();		
			ArrayList attributeCollection = trainingOp.GetAttributeCollection();

			//-------------------------------------------------------------------------------------
			// STEM OPERATOR
			//-------------------------------------------------------------------------------------
			configStr += generateTrainingSetCollectorOperator(opId, 
 															  parentOpId, 
															  inputStreamName,
															  attributeCollection); //generate QM operator			
		}
		System.out.println("Training set operators...Done");
		//-------------------------------------------------------------------------------------
		// ONLINE CLASSIFIER OPERATOR
		//-------------------------------------------------------------------------------------
		configStr += generateQMClassifierOperatorString(xmlOperators,
														classifierSignature.GetOpId(), 
														classifierSignature.GetNumStreams(), 
														classifierSignature.GetNumOperators(),				
														classifierSignature.GetParentOperators(),
														classifierSignature.GetInputStreams(),
														localQMSignatures,
														hashAllLogicalToPhysicalOpCombinations,
														hashLogicalOperatorsWithInputsCombinations); //generate classifier
	
		
		configStr += "</queryplan>" + NEW_LINE;
		System.out.println("Online classifier...Done");
		WriteToFile(outputFileLocation,configStr);
	}
	
	
	private static String generateQMClassifierOperatorString(ArrayList xmlOperators,
															 String opIdVal, 
															 String numStreams, 
															 String numOperators,				
														 	 ArrayList parentOps,
															 ArrayList inputStreams,
															 ArrayList localQMSignatures,
															 Hashtable hashAllLogicalToPhysicalOpCombinations,
															 Hashtable hashLogicalOperatorsWithInputsCombinations)
	{
		String configStr = "";
	    int numSendOff = 100;
		configStr += "<operator root=\"false\" id=\""+opIdVal+"\" className=\"edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.OnlineClassifierOperatorImp\" numberOfOutputQueue = \""+parentOps.size()+"\">" + NEW_LINE;
		configStr += "<classVariables>" + NEW_LINE;		
		configStr += "<variable name=\"Num_Streams\" value=\""+numStreams+"\" />" + NEW_LINE; 
		configStr += "<variable name=\"Num_Operators\" value=\""+numOperators+"\" /> " + NEW_LINE;		
		configStr += "<variable name=\"Num_SendOff\" value=\""+numSendOff+"\" />" + NEW_LINE; 
		configStr += "<variable name=\"TupleCountThreshold\" value=\"5000\" />" + NEW_LINE;
		
		configStr += "<!-- variables needed for each stream -->" + NEW_LINE;
		
		//sort the child streams
		Object [] inputStreamsArray = inputStreams.toArray();
		
		Arrays.sort(inputStreamsArray, new InputStreamSignatureComparer());
		for(int d = 0; d < inputStreamsArray.length; d++)
		{
			ClassifierInputSignature child = (ClassifierInputSignature)inputStreamsArray[d];
			String streamName = child.GetStreamName();
			String streamId = child.GetStreamId();
			String operatorId = child.GetOperatorId();
	
			if (child.GetType() == INPUT_TYPE.STREAM)
			{
				configStr += "<variable name = \""+streamName+"\"  QueueId =\""+d+"\" window_type=\"CountBased\" window_size=\"5000\" />" + NEW_LINE;
			}
		}
		
		
		configStr += "<globalDecisionTree>" + NEW_LINE;
		for(int i = 0; i < localQMSignatures.size();i++)
		{						
			LocalQMSignature localQMSignature = (LocalQMSignature)localQMSignatures.get(i);
			Hashtable hashRusterRouteIdMapping = localQMSignature.GetHashRusterRouteIdMapping();
			configStr += "<localQM id=\""+i+"\" stream_id=\""+localQMSignature.GetStreamId()+"\">" + NEW_LINE;
			//----------------------------------------------------------------------
			// DECISION TREE CONTENT
			//----------------------------------------------------------------------					
			//check number of routes, if 1, no need to output the DT
			double timeDtTranslationStarted = System.currentTimeMillis();
			if (localQMSignature.GetRoutes().size() > 1)
			{							
				configStr += "<localDecisionTree id=\""+i+"\" stream_id=\""+localQMSignature.GetStreamId()+"\" is_empty=\"false\" >" + NEW_LINE;
				ArrayList dtNodes = localQMSignature.GetDecisionTreeNodes();
					
				System.out.println("Stream ID: " + localQMSignature.GetStreamId() + ", Total # of DT nodes: " + dtNodes.size());
				for (int j = 0; j < dtNodes.size();j++)
				{					
					DecisionTreeNodeSignature dtNode = (DecisionTreeNodeSignature)dtNodes.get(j);
					
					String attrIdx = "";
					String attrName = ""; 
					String isBucketized = "";
					String nodeOperation = "";
					String nodeOperationTest = "";
					String rusterName = "";
					String routeId = "";
					String bucketValueStart = "";
					String bucketValueEnd = "";
					
					if (dtNode.GetNodeType().equalsIgnoreCase("internal"))
					{
						attrIdx = dtNode.GetAttrIdxVal();
						attrName = dtNode.GetAttrName(); 
						isBucketized = dtNode.GetIsBucketized();
						nodeOperation = dtNode.GetNodeOperation();
						nodeOperationTest = dtNode.GetNodeOperationTest();					
					}
					else if (dtNode.GetNodeType().equalsIgnoreCase("leaf"))
					{
						//rusterName = nodeOperationTest = dtNode.GetNodeOperationTest();
						//get the route id based on the leaf ruster name						
						//routeId = (String)hashRusterRouteIdMapping.get(rusterName);
						
						String routeName = nodeOperationTest = dtNode.GetNodeOperationTest();
						routeId = localQMSignature.GetRouteIdBySignature(routeName);
					}
					
					if (dtNode.GetNodeType().equalsIgnoreCase("leaf"))
					{
						nodeOperationTest = "";
					}
						
					configStr += "<DTnode id=\""+dtNode.GetNodeId() +
								 "\" type=\""+dtNode.GetNodeType() +
								 "\" attr_idx=\"" + attrIdx +
								 "\" attr_name=\"" + attrName +
								 "\" is_bucket_value=\"" + isBucketized +
								 "\" bucket_range_start=\"" + bucketValueStart +
								 "\" bucket_range_end=\"" + bucketValueEnd +
								 "\" operation =\"" + nodeOperation +
								 "\" test_value =\"" + nodeOperationTest +
								 "\" route_id =\"" + routeId +
								 "\">" + NEW_LINE;
					//----------------------------------------------------------------------
					// NODE PARENTS
					//----------------------------------------------------------------------
					if (dtNode.GetNodeType().equalsIgnoreCase("root"))
					{
						configStr += "<parents/>" + NEW_LINE;
					}
					else
					{
						configStr += "<parents>" + NEW_LINE;
						configStr += "<parent node_id=\""+dtNode.GetParentNodeId()+"\"/>" + NEW_LINE;
						configStr += "</parents>" + NEW_LINE;
					}
					//----------------------------------------------------------------------
					// NODE CHILDREN
					//----------------------------------------------------------------------				
					if (dtNode.GetChildrenNodesIds() != null)
					{
						if (dtNode.GetChildrenNodesIds().size() > 0)
						{
							configStr += "<children>" + NEW_LINE;
							for(int p = 0; p < dtNode.GetChildrenNodesIds().size(); p++)
							{
								String child_id = (String)dtNode.GetChildrenNodesIds().get(p);
								configStr += "<child node_id=\""+child_id+"\"/>" + NEW_LINE;
							}
							configStr += "</children>" + NEW_LINE;
						}
						else
						{
							configStr += "<children/>" + NEW_LINE;
						}
					}
					else
					{
						configStr += "<children/>" + NEW_LINE;
					}
					configStr += "</DTnode>" + NEW_LINE;
				}
				configStr += "</localDecisionTree>" + NEW_LINE;
			}
			else
			{				
				configStr += "<localDecisionTree id=\""+i+"\" stream_id=\"" + localQMSignature.GetStreamId()+"\" is_empty=\"true\" />" + NEW_LINE;
			}
			
			double timeDtTranslationEnded = System.currentTimeMillis();
			double timeDtTranslationDuration = timeDtTranslationEnded - timeDtTranslationStarted;
			double timeDtTranslationDurationSec = timeDtTranslationDuration / 1000;
			Helper.WriteToFile(Helper.LOG_FILE,  Helper.GetCurrentTimeString() + " --- Local DT ["+localQMSignature.GetDecisionTreeNodes().size()+"] nodes, translation time ["+timeDtTranslationDuration+"] msec or ["+timeDtTranslationDurationSec+"] sec --- ");
			 
			
			//----------------------------------------------------------------------
			// ROUTES CONTENT
			//----------------------------------------------------------------------
			
			double timeRouteTranslationStarted = System.currentTimeMillis();
			
			configStr += "<allRoutes>" + NEW_LINE;
			Hashtable allRoutes = (Hashtable)localQMSignature.GetRoutes();						
			
			Enumeration eRoutes = allRoutes.elements(); 
			ArrayList streamsInSchema = new ArrayList();
			
			while (eRoutes.hasMoreElements()) 
			{
				PhysicalRouteSignature physRouteSignature = (PhysicalRouteSignature)eRoutes.nextElement();
				
				String strTraditionalPlan = "";
				String strRouteOrderedOperators = "";
				//now produce the actual order of physical operators in that route
				ArrayList operatorsNotFilled = new ArrayList();
				int numOperatorsAssigned = 0;
				String last_done_op_id = "";
				String first_done_op_id = "";
				String last_done_op_order = "";
				String last_done_op_input = "";
				
				for(int k = 0; k < physRouteSignature.GetPhysicalOperatorOrder().size();k++)
				{ 
					String probeKey = ""+(k + 1);
					//logical op id
					String currOpId = (String)physRouteSignature.GetPhysicalOperatorOrder().get(probeKey);
					ArrayList opInputStreams = GetInputsByOpId(xmlOperators,currOpId);
					strTraditionalPlan += opInputStreams +",";
					//map logical op id to physical op id										
					String keyLogPhysProbee = currOpId+"|" + localQMSignature.GetStreamId();
										
					LogicalOpToPhysicalOpSignature logToPhysCombination = (LogicalOpToPhysicalOpSignature)hashAllLogicalToPhysicalOpCombinations.get(keyLogPhysProbee);
					
					if (logToPhysCombination != null)
					{
						//add stream id to the schema list 						
						ArrayList opInputs = (ArrayList)hashLogicalOperatorsWithInputsCombinations.get(currOpId);
						for(int g = 0; g < opInputs.size(); g++)
						{
							if (!IsIDInArrayList(streamsInSchema, (String)opInputs.get(g)))
							{
								streamsInSchema.add(opInputs.get(g));
							}
						}
						
						//append to the route phys op id
						strRouteOrderedOperators += logToPhysCombination.GetPhysicalOpId() + "|";						
						
						/*
						if(numOperatorsAssigned==0)
						{
							first_done_op_id = currOpId; 
						}
						numOperatorsAssigned++;
						last_done_op_id = currOpId;
						last_done_op_order = probeKey;
						last_done_op_input = localQMSignature.GetStreamId();
						*/
					}
					else
					{
						operatorsNotFilled.add(currOpId);
						
						ArrayList tmpStreamsInSchema = streamsInSchema;
						//iterate through all streams in the schema and see if 
						for (int b = 0; b < tmpStreamsInSchema.size(); b++)
						{					
							String currStream = (String)tmpStreamsInSchema.get(b);						
							String keyLogPhysProbe = currOpId+"|" + currStream;
							
							LogicalOpToPhysicalOpSignature logToPhysCombinationAttempt = (LogicalOpToPhysicalOpSignature)hashAllLogicalToPhysicalOpCombinations.get(keyLogPhysProbe);
							
							if (logToPhysCombinationAttempt != null)
							{
								//add stream id to the schema list 						
								ArrayList opInputs = (ArrayList)hashLogicalOperatorsWithInputsCombinations.get(currOpId);
								for(int g = 0; g < opInputs.size(); g++)
								{
									if (!IsIDInArrayList(streamsInSchema, (String)opInputs.get(g)))
									{
										streamsInSchema.add(opInputs.get(g));
									}
								}
								
								//append to the route phys op id
								strRouteOrderedOperators += logToPhysCombinationAttempt.GetPhysicalOpId() + "|";
								break;
							}
						}//end for all operators in the schema
					}
				}//end iterating through a particular route
								
				//---------------------------------------------------------------------------------
				// NOW ADD THE OPERATORS THAT HAVE NOT BEEN PREVIOUSLY ADDED
				//---------------------------------------------------------------------------------
				
				/*
				int b = 0;
				while (operatorsNotFilled.size() != 0)
				{																		
					String currOpId = (String)operatorsNotFilled.get(b);
					
					for (int f = 0; f < streamsInSchema.size(); f++)
					{
						String availableInputStream = (String)streamsInSchema.get(f);
												
						String keyLogPhysProbee = currOpId + "|" + availableInputStream;
						LogicalOpToPhysicalOpSignature logToPhysCombination = (LogicalOpToPhysicalOpSignature)hashAllLogicalToPhysicalOpCombinations.get(keyLogPhysProbee);
								
						if (logToPhysCombination != null)
						{
							//append to the route phys op id
							strRouteOrderedOperators += logToPhysCombination.GetPhysicalOpId() + "|";						
							//last_done_op_id = currOpId;	
							//last_done_op_input = nextOpInput;
							opIndexToRemove = b;
							//remove operator from the list
							operatorsNotFilled.remove(opIndexToRemove);
							b = 0; //reset the b counter
							
							//add inputs to the schema list
							ArrayList opInputs = (ArrayList)hashLogicalOperatorsWithInputsCombinations.get(currOpId);
							for(int g = 0; g < opInputs.size(); g++)
							{
								if (!IsIDInArrayList(streamsInSchema, (String)opInputs.get(g)))
								{
									streamsInSchema.add(opInputs.get(g));
								}
							}
							break;
						}						
					}//end for b
					
				}//end while all "not previously added" operators have been added to the route 					
				*/
					
				strRouteOrderedOperators = strRouteOrderedOperators.substring(0,strRouteOrderedOperators.length()-1);
				strTraditionalPlan = strTraditionalPlan.substring(0,strTraditionalPlan.length()-1);	
				
				String strRusterName = physRouteSignature.GetRusterName();				
					
				//if the number of routes at the end is only one, we know
				//that other routes were no different from the default route
				if (localQMSignature.GetRoutes().size() == 1)
				{
					strRusterName = "default_route";
				}
					
				if (!physRouteSignature.IsDefaultRoute())
				{
					//default route
					//ruster_name=\""+strRusterName+"\"
					configStr += "<route id=\""+physRouteSignature.routeId+"\" is_default=\"false\" path=\""+strRouteOrderedOperators+"\" logical_plan=\""+strTraditionalPlan+"\" />"+ NEW_LINE;
				}
				else
				{					
					//ruster_name=\"\"
					configStr += "<route id=\""+physRouteSignature.routeId+"\" is_default=\"true\" path=\""+strRouteOrderedOperators+"\" logical_plan=\""+strTraditionalPlan+"\" />"+ NEW_LINE;
				}													
			}//end while iterating through all routes (while loop)			
			
			configStr += "</allRoutes>" + NEW_LINE;
			
			double timeRouteTranslationEnded = System.currentTimeMillis();
			double timeRouteTranslationDuration = timeRouteTranslationEnded - timeRouteTranslationStarted;
			double timeRouteTranslationDurationSec = timeRouteTranslationDuration / 1000;
			Helper.WriteToFile(Helper.LOG_FILE,  Helper.GetCurrentTimeString() + " --- Local ROUTES ["+localQMSignature.GetRoutes().size()+"] translation time ["+timeRouteTranslationDuration+"] msec or ["+timeRouteTranslationDurationSec+"] sec --- ");
			 
						
			//----------------------------------------------------------------------
			configStr += "</localQM>" + NEW_LINE;
			
			System.out.println("Local classifier for stream ["+i+"]...Done");
		}
		configStr += "</globalDecisionTree>" + NEW_LINE;
		configStr += "</classVariables>" + NEW_LINE;
		
		configStr += "<properties/>" + NEW_LINE;
		configStr += "<schema/>" + NEW_LINE;
		configStr += "<parents>" + NEW_LINE;
		
		//sort the parent operators
		Object [] parentOpArray = parentOps.toArray();
		Arrays.sort(parentOpArray, new ParentOperatorSignatureComparer());
		for(int d = parentOpArray.length - 1; d >= 0; d--)
		{		
			ParentOperatorSignature parent = (ParentOperatorSignature)parentOpArray[d];
			String parentOpId = parent.GetOpId();
			String parentOpQueueId = (String)onlineClassifierParentQueues.get(parentOpId);
			configStr += "<parent id = \""+parentOpId+"\" queueId = \""+parentOpQueueId+"\"/>" + NEW_LINE;
		}
		System.out.println("Parent signatures...Done");
		configStr += "</parents>" + NEW_LINE;
		configStr += "<children>" + NEW_LINE;
				
		//sort the child streams
		//Object [] inputStreamsArray = inputStreams.toArray();
		//Arrays.sort(inputStreamsArray, new InputStreamSignatureComparer());

		if (((ClassifierInputSignature)inputStreamsArray[0]).GetType() == INPUT_TYPE.STREAM)
		{
			//----------------------------------------------------------
			// CHILD IS A STREAM
			//----------------------------------------------------------			
			for(int d = 0; d < inputStreamsArray.length; d++)
			{
				ClassifierInputSignature child = (ClassifierInputSignature)inputStreamsArray[d];			
				String streamName = child.GetStreamName();
				String streamId = child.GetStreamId();
				configStr += "<child type=\"stream\" id=\""+streamId +"\" name = \""+streamName+"\" />" + NEW_LINE;
			}
		}
		else
		{
			//----------------------------------------------------------
			// CHILD IS AN OPERATOR
			//----------------------------------------------------------
			for(int d = inputStreamsArray.length - 1; d >= 0; d--)
			{
				ClassifierInputSignature child = (ClassifierInputSignature)inputStreamsArray[d];
				String operatorId = child.GetOperatorId();
				String streamName = child.GetStreamName();
				String streamId = streamName.substring(6); //remove the "Stream" prefix
				configStr += "<child type=\"operator\" id=\""+operatorId +"\" stream_id =\""+streamId+"\" />" + NEW_LINE;
			}
		}
		
		
		configStr += "</children>" + NEW_LINE;
		configStr += "</operator>" + NEW_LINE;
		return configStr;
	}

	 private static ArrayList GetInputsByOpId(ArrayList xmlOperators, String id)
	{
		 ArrayList opIntputs = new ArrayList();
		 for (int i=0; i<xmlOperators.size(); i++)
		 {
			 SimpleOperatorSpecification op = (SimpleOperatorSpecification)xmlOperators.get(i);
			 if (op.getId().equalsIgnoreCase(id))
			 {
				 opIntputs.addAll(op.getInputs());
				 return opIntputs;
			 }
		 }	
		 return opIntputs;
	}
	 
	private static boolean IsIDInArrayList(ArrayList opInputs, String id)
	{
		for (int k=0; k<opInputs.size(); k++)
		{
			if (opInputs.get(k).equals(id))
			{
				return true;
			}
		}
		return false;
	}
	
	private static String getNextOperatorBasedOnIgnoreInput(Hashtable hashAllLogicalToPhysicalOpCombinations, 
												 String currOpId, 
												 String inputToIgnore)
	{			
		String nextOpInput = "";
		Enumeration eKeys = hashAllLogicalToPhysicalOpCombinations.keys(); 		
		while (eKeys.hasMoreElements()) 
		{
			String currKey = (String)eKeys.nextElement();
			//split the key into op id and the input stream
			String[] tokens = currKey.split("|");
			String extractedOpId = tokens[tokens.length-3];
			String extractedInputId = tokens[tokens.length-1];
			
			if (extractedOpId.equalsIgnoreCase(currOpId))
			{
				if (!extractedInputId.equalsIgnoreCase(inputToIgnore))
				{
					nextOpInput = extractedInputId;
					return nextOpInput;
				}
			}
		}
		return nextOpInput;
	}

	
	private static String getNextOperatorBasedOnConfineInput(Hashtable hashAllLogicalToPhysicalOpCombinations, 
													      String currOpId, 
													      String inputToConfineTo)
	{			
		String nextOpInput = "";
		Enumeration eKeys = hashAllLogicalToPhysicalOpCombinations.keys(); 		
		while (eKeys.hasMoreElements()) 
		{
			String currKey = (String)eKeys.nextElement();
			//split the key into op id and the input stream
			String[] tokens = currKey.split("|");
			String extractedOpId = tokens[tokens.length-3];
			String extractedInputId = tokens[tokens.length-1];
			
			if (extractedOpId.equalsIgnoreCase(currOpId))
			{
				if (!extractedInputId.equalsIgnoreCase(inputToConfineTo))
				{
					nextOpInput = extractedInputId;
					return nextOpInput;
				}
			}
		}
		return nextOpInput;
	}

	private static String generateQMSTEMOperatorString(String opIdVal, 
												String streamIdVal, 
												String probeStreamId,
												String numTupleIndex,
												String stemStreamId,
												String numSTeMIndex,
												String rootOpId,
												ArrayList childrenOps)
	{
		String configStr = "";
		configStr += "<operator root = \"false\" id = \""+opIdVal+"\" className = \"edu.wpi.cs.dsrg.STeM.STeMJoinProbeOperatorImp\">" + NEW_LINE;
		configStr += "<classVariables>" + NEW_LINE;
		configStr += "<variable name=\"QMeshOperatorID\" value=\""+opIdVal+"\" />" + NEW_LINE;
		configStr += "<variable name=\"IsEddyOp\" value=\"false\" />" + NEW_LINE; 
		configStr += "<variable name=\"StreamID\" value=\""+streamIdVal+"\" /> " + NEW_LINE;
		configStr += "<!-- probe data  --> " + NEW_LINE;		
		configStr += "<variable name=\"ProbeSTREAMId\" value=\""+probeStreamId+"\" />" + NEW_LINE;
		configStr += "<variable name=\"NumTupleIndex\" value=\""+numTupleIndex+"\"/>" + NEW_LINE;				
		configStr += "<!-- stored data  --> " + NEW_LINE;				
		configStr += "<variable name=\"STeMSTREAMId\" value=\""+stemStreamId+"\" />" + NEW_LINE;
		configStr += "<variable name=\"NumSTeMIndex\" value=\""+numSTeMIndex+"\"/>" + NEW_LINE;		
		configStr += "<expressions>" + NEW_LINE;
		configStr += "</expressions>" + NEW_LINE;
		configStr += "</classVariables>" + NEW_LINE;
		configStr += "<properties>" + NEW_LINE;
		configStr += "</properties>" + NEW_LINE;
		configStr += "<schema/>" + NEW_LINE;
		configStr += "<parents>" + NEW_LINE;
		configStr += "<parent id = \""+rootOpId+"\"/>" + NEW_LINE;    	
		configStr += "</parents>" + NEW_LINE;
		configStr += "<children>" + NEW_LINE;

		if (childrenOps != null)
		{
			for(int i=0; i< childrenOps.size();i++)
			{
				ChildOperatorSignature child = (ChildOperatorSignature)childrenOps.get(i);
				String childOpId = child.GetOpId();
				String childQueueId = child.GetQueueId();
				configStr += "<child type=\"operator\" id = \""+childOpId+"\" queueId=\""+childQueueId+"\"/>" + NEW_LINE;
				
				onlineClassifierParentQueues.put(opIdVal, childQueueId);
			}
		}
		configStr += "</children>" + NEW_LINE;
		configStr += "</operator>" + NEW_LINE;
		return configStr;
	}
	
	
	//--------------------------------------------------------------------------------------------
	// STQM Functionality
	//--------------------------------------------------------------------------------------------
	private static String generateTrainingSetCollectorOperator(String opId,
															   String parentOpId,
															   String inputStreamName,
															   ArrayList attributeCollection)
	{
		String streamId = inputStreamName.substring(6); //remove the "Stream" prefix
		String configStr = "";
		configStr += "<operator root = \"false\" id = \""+opId+"\" className = \"edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.TrainingTupleCollectorOperatorImp\">" + NEW_LINE;
		configStr += "<classVariables>" + NEW_LINE;   
		configStr += "<!-- Generic Properties -->" + NEW_LINE;		
		configStr += "<variable name=\"window_size\" value=\"100\"/>" + NEW_LINE;	
		configStr += "<variable name=\"operator_state\" value=\"edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.SingleStreamSamplingState\"/>" + NEW_LINE;	
     	
		configStr += "<!-- Variable used to determine the sampling strategy -->" + NEW_LINE;			     	
		configStr += "<variable name=\"num_sample_windows\" value=\"" + "10" + "\"/>" + NEW_LINE;
		configStr += "<variable name=\"sample_windows_size\" value=\"100\"/>" + NEW_LINE;
		configStr += "<variable name=\"tuples_per_window_to_sample\" value=\"10\"/>" + NEW_LINE;
		configStr += "<variable name=\"sampling_heuristic\" value=\"simple_random\"/>" + NEW_LINE;	 
		
		configStr += "<!-- Variable used to specify location of the sample DUMP -->" + NEW_LINE;			
		configStr += "<!-- This file will be used in the construction of the decistion tree -->" + NEW_LINE;			
		configStr += "<variable name=\"sample_dump_file\" value=\"C:\\QueryMesh\\training_sets\\TRAINING_SET_DS5_STREAM8.txt\"/>" + NEW_LINE;	  					
		configStr += "<!-- Decision tree header contains the names of the attributes to be used -->" + NEW_LINE;	
		configStr += "<!-- in the decision tree. You can just copy these attributes directly from -->" + NEW_LINE;	
		configStr += "<!-- the stream specification (but you will need to add an additional parameter to ignore/to use)-->" + NEW_LINE;	
		configStr += "<!-- tuple attribute in the decision tree)-->" + NEW_LINE;			
		configStr += "<decisionTreeHeader>" + NEW_LINE;	
		
		for (int i = 0; i < attributeCollection.size(); i++)
		{
			AttributeProperties tts = (AttributeProperties)attributeCollection.get(i);
			configStr += "<attribute name=\""+tts.GetAttrName()+"\" type=\""+tts.GetAttrType()+"\" use=\""+tts.GetBUseAttr()+"\" is_target=\""+tts.GetBIsTarget()+"\"/>" + NEW_LINE;
		}			
		configStr += "</decisionTreeHeader>" + NEW_LINE;				
		
		configStr += "<!-- There are 2 types of histograms: (a) number-based, (b) string-based -->" + NEW_LINE;	
		configStr += "<!-- parameters for the number-based histogram: # of buckets, min and max values -->" + NEW_LINE;	
		configStr += "<!-- parameters for the string-based histogram: nothing, each new string will be assigned its own bucket -->" + NEW_LINE;	
		
		configStr += "<histograms>" + NEW_LINE;		  	
		configStr += "</histograms>" + NEW_LINE;	
		
		configStr += "</classVariables>" + NEW_LINE;	
		
		configStr += "<properties>" + NEW_LINE;	 
		configStr += "</properties>" + NEW_LINE;	
		
		configStr += "<schema/>" + NEW_LINE;	
		
		configStr += "<parents>" + NEW_LINE;	
		configStr += "<parent id = \"" + parentOpId + "\"/>" + NEW_LINE;	 
		configStr += "</parents>" + NEW_LINE;	
		
		configStr += "<children>" + NEW_LINE;	
		configStr += "<child type=\"stream\" name=\""+inputStreamName+"\" queueId=\"0\"/>" + NEW_LINE;	
		configStr += "</children>" + NEW_LINE;	
		configStr += "</operator>" + NEW_LINE;			
		return configStr;
	}


					
	private static String generateDummyOperatorString(String opIdVal, ArrayList childrenOps)
	{
		String configStr = "";
		configStr += "<operator root = \"true\" id = \""+opIdVal+"\" className = \"edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.DoNothingOperatorImp\" numberOfOutputQueue = \"1\">" + NEW_LINE;
		configStr += "<classVariables>" + NEW_LINE;
		configStr += "</classVariables>" + NEW_LINE;
		configStr += "<properties>" + NEW_LINE;
		configStr += "</properties>" + NEW_LINE;    
		configStr += "<schema/>" + NEW_LINE;
		configStr += "<parents>" + NEW_LINE;
		configStr += "</parents>" + NEW_LINE;
		configStr += "<children>" + NEW_LINE;
		
	    //sort the child operators
		Object [] childrenOpsArray = childrenOps.toArray();
		Arrays.sort(childrenOpsArray, new ChildOperatorSignatureComparer());
		for(int i=0; i < childrenOpsArray.length; i++)
		{
			ChildOperatorSignature child = (ChildOperatorSignature)childrenOpsArray[i];
			String childOpId = child.GetOpId();
			String childQueueId = child.GetQueueId();
			//configStr += "<child type=\"operator\" id = \""+childOpId+"\" queueId=\""+childQueueId+"\"/>" + NEW_LINE;
			configStr += "<child type=\"operator\" id = \""+childOpId+"\" />" + NEW_LINE;
		}					
		configStr += "</children>" + NEW_LINE;
		configStr += "</operator>" + NEW_LINE;
		return configStr;					
	}
					
	public static void WriteToFile(String fileLocation, String stringToWrite)
	{
		try
		{
			//Create file 
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			out.newLine();
			//Close the output stream
			out.close();			    
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}
}