/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security;

/**
 * @author Rimma V. Nehme
 *
 */
public class AccessControlModel implements java.io.Serializable
{
	//------------------------------------------------------------------------------------
	
	//discretionary access control
	public static final int DAC_MODEL = 1;
	//mandatory access control
	public static final int MAC_MODEL = 2;
	//flat role-based access control
	public static final int FLAT_RBAC_MODEL = 3;
	//flat purpose-based access control
	public static final int FLAT_PBAC_MODEL = 4;
	
	protected int type = 0;	
	public String value = "";
	protected String source =""; //source for this access control model (where it came from)

	//------------------------------------------------------------------------------------
	
	public AccessControlModel(int newType, String newValue, String newSource)
	{
		type = newType;
		value = newValue;		
		source = newSource;
	}
	
	//------------------------------------------------------------------------------------
	
	public int getACMType()
	{
		return type;
	}
	
	//------------------------------------------------------------------------------------
	
	public String getACMTypeString()
	{
		String typeString = "";
		switch(type)
		{
			case DAC_MODEL:
				typeString = "dac";
				break;
				
			case MAC_MODEL:
				typeString = "mac";
				break;		
				
			case FLAT_RBAC_MODEL:
				typeString = "frbac";
				break;	
				
			case FLAT_PBAC_MODEL:
				typeString = "fpbac";
				break;				
		}
		return typeString;
	}
	
	//------------------------------------------------------------------------------------
	
	public void setACMType(int newType)
	{
		type = newType;
	}
	
	//------------------------------------------------------------------------------------
	
	public String getACMValue()
	{
		return value;
	}
	
	//------------------------------------------------------------------------------------
	
	public void setACMValue(String newValue)
	{
		value = newValue;
	}
	
	//------------------------------------------------------------------------------------
	
	public String getACMSource()
	{
		return source;
	}
	
	//------------------------------------------------------------------------------------
	
	public void setACMSource(String newSource)
	{
		source = newSource;
	}

	//------------------------------------------------------------------------------------
	
	public String  GetTypeString(int type)
	{
		String strType = "";
		switch(type)
		{
		case 1:
			strType = "DAC_MODEL";
			break;
		case 2:
			strType = "MAC_MODEL";
			break;
		case 3:
			strType = "FLAT_RBAC_MODEL";
			break;
		case 4:
			strType = "FLAT_PBAC_MODEL";
			break;			
		}//end switch
		return strType;
	}
	
	//------------------------------------------------------------------------------------
	
    public String toString() 
    {
        StringBuilder result = new StringBuilder();
        String NEW_LINE = System.getProperty("line.separator");

        result.append(NEW_LINE + "AccessControlModel:"+ NEW_LINE);
        result.append("type (int): " + type + NEW_LINE);
        result.append("type (string): " + GetTypeString(type) + NEW_LINE);
        result.append("value: " + value + NEW_LINE);
        result.append("source: " + source + NEW_LINE);
        return result.toString();    
    }
	
  //----------------------------------------------------------------------------------------------------
    
}
