/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security;

import java.io.Serializable;
import java.util.Hashtable;
import java.util.Vector;


public class AccessControlPolicyStore implements  Serializable 
{
	//------------------------------------------------------------------------------------
	
	private int MAX_POLICY_COUNT = 200;
	private Object[] policyTable;
	private int current_policy_count = 0;
	private int current_key = 0;
	
	//------------------------------------------------------------------------------------
	
	public AccessControlPolicyStore()
	{
		policyTable = new Object[MAX_POLICY_COUNT]; //allow max 200 different policies 
	}
	
	//------------------------------------------------------------------------------------
	
	public int addNewPolicy(Object policy)
	{
		if (current_key == MAX_POLICY_COUNT)
		{
			current_key = 0;
		}
		
		while (policyTable[current_key] != null && current_key < this.MAX_POLICY_COUNT)
		{
			current_key++;
		}
		
		policyTable[current_key] = policy;
		current_policy_count++;
		current_key++;
		
		//return the index where the policy was inserted
		return current_key-1;
	}
	
	//------------------------------------------------------------------------------------
	
	public Object retrieveExistingPolicy(int index)
	{
		return policyTable[index];
	}
	
	//------------------------------------------------------------------------------------

	void removeExistingPolicy(int key)
	{
		policyTable[key] = null;
		current_policy_count--;
	}

	//------------------------------------------------------------------------------------
	
	boolean isAlreadyInPolicyStore(Object policy)
	{
		for (int i=0; i<MAX_POLICY_COUNT;i++)
		{
			if (policyTable[i] != null)
			{				
				AccessControlModel [] spAcmModels = (AccessControlModel [])policyTable[i];		
				AccessControlModel [] policyToCheck = (AccessControlModel [])policy;
				
				String strAcmModel = spAcmModels[0].getACMValue().toUpperCase();
				String strPolicyToCheck = policyToCheck[0].getACMValue().toUpperCase();
					
				if (strAcmModel.equals(strPolicyToCheck))
				{			     
					return true;		
				}								
			}			
		}
		return false;
	}
	
	//------------------------------------------------------------------------------------
	
	public int getCurrentPolicyCount()
	{
		return current_policy_count;
	}
	
	//------------------------------------------------------------------------------------
	
	public int getCurrentKey()
	{
		return current_key;
	}
	
	//------------------------------------------------------------------------------------
	
}
