/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstreamoperator.BinaryPWSingleEquiJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;

/**
 * @author Rimma V. Nehme
 *
 */
public class BinaryIndexSCJoinOperatorImp extends XATMultiQueueWindowStreamOperatorImp implements BinaryPWSingleEquiSCJoinOperator 
{
	
	//-----------------------------------------------------------------------------
	
	private int totalTuples = 0; //tuples
	private int totalSps = 0; //sps
	private int totalOnlyTuplesLeftQ = 0;
	private int totalOnlySpsLeftQ = 0;
	private int totalOnlyTuplesRightQ = 0;
	private int totalOnlySpsRightQ = 0;
	private int elementsLeftQ = 0;
	private int elementsRightQ = 0;
	private int totalNumberOfJoinResults = 0;
	
	static int titleHelper = 1;
	
	double timeJoinStart = 0;
	double timeJoinEnd = 0;
	double timeJoin = 0;
	
	double timeSPMaintenanceStart = 0;
	double timeSPMaintenanceEnd = 0;
	double timeSPMaintenance = 0;
	
	double timeTupleMaintenanceStart = 0;
	double timeTupleMaintenanceEnd = 0;
	double timeTupleMaintenance = 0;
		
	int LEFT = 0;
	int RIGHT = 1;
	
	protected static int WINDOW_SIZE = 100; //window size = 100 tuples
		
	protected boolean propagateFlag;
	protected int memoryLimit;


	SinglePWSingleEquiSCJoinState leftState;
	SinglePWSingleEquiSCJoinState rightState;

	int leftJoinAttrPos;
	int rightJoinAttrPos;	
	
	int leftWindowSize;
	int rightWindowSize;

	int numTuplesInLeftWindow = 0;
	int numTuplesInRightWindow = 0;
	
	int numSecurityPunctuationsInLeftWindow = 0;
	int numSecurityPunctuationsInRightWindow = 0;
	
	static XATTuple spTupleStart = null;
	static XATTuple spTupleEnd = null;
	
	int SP_UNKNOWN = -999;
	int SP_ARRIVED_INITIALIZED = 1;
	int SP_TUPLE_START_INITIALIZED = -1;
	int SP_TUPLE_END_INITIALIZED = 0;
	
	int currentLeftSPStatus = SP_UNKNOWN;
	int currentRightSPStatus = SP_UNKNOWN;

	//-----------------------------------------------------------------------------
	
	public BinaryIndexSCJoinOperatorImp(XATNode init) 
	{
		super(init);
		
		if (leftWindowSize == 0)
		{
			leftWindowSize = WINDOW_SIZE;
		}
		
		if (rightWindowSize == 0)
		{
			rightWindowSize = WINDOW_SIZE;
		}
	}
	
	//-----------------------------------------------------------------------------	

	public boolean run() throws Exception 
	{
		if (titleHelper == 1) 
		{
			Helper.StatsWriteLineToFile("SCJoinstats.txt", "inside");
		}

		if(this.getDebug())
		{
			System.out.println("Running: " + this.toToolTipString() + "\n");
		}
		
		//outputqueue for join output
		//could be multiple outputs. but the schema is the same.
		for (int i=0;i<this.getOutputQueueCount()-1;i++ )
		{
			if (this.getOutputQueue(i).getSchema() == null) 
			{
				init(i);
			}
		}
		
		//Outputqueue for pipelined-join
		//We assume here that the plan generator will use the LAST outputqueue
		// for the next pipelined window join
		//We just simply copy the input schemas to the output queue
		if (this.getOutputQueue(this.getOutputQueueCount()-1)!=null && this.getOutputQueue(this.getOutputQueueCount()-1).getSchemas().isEmpty())
		{
			this.getOutputQueue(this.getOutputQueueCount()-1).setSchemas(
					this.getInputQueue().getSchemas());
		}
		
		
		XATTuple newObj = null;
		int indication = LEFT;
		int tupleCount = 0;
	
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));	

		elementsLeftQ = this.getLeftInputQueue().getQueueSize();
		elementsRightQ = this.getRightInputQueue().getQueueSize();
		
		/*start recording time*/
		//double timeOverallStart = System.currentTimeMillis();
		double timeOverallStart = System.nanoTime();
		
		while (true) 
		{		
			if (tupleCount == howMany) /* done with processing */ 
			{
				break;
			}
			
			try 
			{
				//pick the next tuple to process
				if ((this.getLeftInputQueue().getQueueSize(this) == 0) && (this.getRightInputQueue().getQueueSize(this) > 0)) 
				{
					newObj = this.getRightInputQueue().dequeue(this);
					indication = RIGHT;
				} 
				else if ((this.getRightInputQueue().getQueueSize(this) == 0) && (this.getLeftInputQueue().getQueueSize(this) > 0))  
				{
					newObj = this.getLeftInputQueue().dequeue(this);
					indication = LEFT;
				} 
				else 
				{					
					long leftTimeStamp = 0;
					long rightTimeStamp = 0;
		
					leftTimeStamp = this.getLeftInputQueue().peek(this).getNthTimeStamp(this.leftJoinAttrPos);
					rightTimeStamp = this.getRightInputQueue().peek(this).getNthTimeStamp(this.rightJoinAttrPos);
					
					if(leftTimeStamp <= rightTimeStamp) 
					{
						newObj = this.getLeftInputQueue().dequeue(this);
						indication = LEFT;
					} 
					else 
					{
						newObj = this.getRightInputQueue().dequeue(this);
						indication = RIGHT;
					}					
				}
			} 
			catch(XATQueueException e1) 
			{
				return true;
			}
	
			//processing a dequeued tuple/sp (from either of the queues)
			if (newObj != null)
			{			
				//LEFT SIDE
				if (indication == LEFT) 
				{
					if (newObj.isSecPunct()) 
					{	
						totalSps++;
						totalOnlySpsLeftQ++;
												
						//receive a new security punctuation
						XATSecPunctImp sp = (XATSecPunctImp)newObj;					

						timeSPMaintenanceStart = System.nanoTime();
						//add current sp to the state
						leftState.addSPToState(sp);
						timeSPMaintenanceEnd = System.nanoTime();
						timeSPMaintenance += (timeSPMaintenanceEnd - timeSPMaintenanceStart);
						
						numSecurityPunctuationsInLeftWindow++;
						
						if ((currentLeftSPStatus == SP_TUPLE_START_INITIALIZED) || (currentLeftSPStatus == SP_UNKNOWN))
						{							
							//set the status of the sp
							currentLeftSPStatus = SP_ARRIVED_INITIALIZED;
						}
					} 					
					else 
					{						
						totalTuples++;
						totalOnlyTuplesLeftQ++;
						
						if (leftState.getActiveSP() == null)
						{
							//without associated sp, we will not process the data
							continue;
						}
						
						//received a regular tuple
						tupleCount ++;						
						numTuplesInLeftWindow ++;
						
						if (numTuplesInLeftWindow > leftWindowSize)
						{
							int difference = Math.abs(leftWindowSize - numTuplesInLeftWindow);
							timeTupleMaintenanceStart = System.nanoTime();
							leftState.removeTuplesFromState(difference);
							timeTupleMaintenanceEnd = System.nanoTime();
							timeTupleMaintenance += (timeTupleMaintenanceEnd - timeTupleMaintenanceStart);
							numTuplesInLeftWindow = numTuplesInLeftWindow - difference;
						}
					
						timeTupleMaintenanceStart = System.nanoTime();
						//add tuple to the state
						leftState.addTupleToState(newObj);
						timeTupleMaintenanceEnd = System.nanoTime();
						timeTupleMaintenance += (timeTupleMaintenanceEnd - timeTupleMaintenanceStart);
						
						if (currentLeftSPStatus == SP_ARRIVED_INITIALIZED)
						{							
							leftState.getActiveSP().setStartTuple(leftState.getCurrentWindowTuplesQueue().getLastNode());
							leftState.getActiveSP().setEndTuple(leftState.getCurrentWindowTuplesQueue().getLastNode());
							currentLeftSPStatus = SP_TUPLE_START_INITIALIZED;
						}
						else if (currentLeftSPStatus == SP_TUPLE_START_INITIALIZED)
						{
							leftState.getActiveSP().setEndTuple(leftState.getCurrentWindowTuplesQueue().getLastNode());
						}
						
						Object value = newObj.getCell(leftJoinAttrPos).getValue();		
					
						timeJoinStart = System.nanoTime();
						ResultPackage resultPackage = rightState.probeWithSP(rightJoinAttrPos, value, leftState.getActiveSP().getSP());
						timeJoinEnd = System.nanoTime();
						timeJoin += (timeJoinEnd - timeJoinStart);
						
						if (resultPackage == null) 
						{							
							continue;
						}
						
						XATTuple[] matchedTuples = resultPackage.getResultTuples();
						if (matchedTuples != null) 
						{
							for (int i=0; i<matchedTuples.length; i++) 
							{
								//XATTuple resultTuple = this.generateResultTuple(new XATTuple[] {newObj, matchedTuples[i]});
								XATTuple resultTuple =  matchedTuples[i];
								this.getOutputQueue().enqueue(resultTuple, this);
								totalNumberOfJoinResults++;
							}
						}						
					}
				}			 
				else 
				{
					
					//RIGHT SIDE
					if (newObj.isSecPunct()) 
					{ 
						totalSps++;
						totalOnlySpsRightQ++;
						
						XATSecPunctImp sp = (XATSecPunctImp)newObj;	
						//add current sp to the state
						
						timeSPMaintenanceStart = System.nanoTime();
						rightState.addSPToState(sp);
						timeSPMaintenanceEnd = System.nanoTime();
						timeSPMaintenance += (timeSPMaintenanceEnd - timeSPMaintenanceStart);
						
						numSecurityPunctuationsInRightWindow++;		
						
						if ((currentRightSPStatus == SP_TUPLE_START_INITIALIZED) || (currentRightSPStatus == SP_UNKNOWN))
						{							
							//set the status of the sp
							currentRightSPStatus = SP_ARRIVED_INITIALIZED;
						}
					} 					
					else 
					{
						totalTuples++;
						totalOnlyTuplesRightQ++;
						
						if (rightState.getActiveSP() == null)
						{
							//without associated sp, we will not process the data
							continue;
						}
						
						//received a regular tuple
						tupleCount ++;						
						numTuplesInRightWindow ++;
						
						if (numTuplesInRightWindow > rightWindowSize)
						{
							int difference = Math.abs(rightWindowSize - numTuplesInRightWindow);
							timeTupleMaintenanceStart = System.nanoTime();
							rightState.removeTuplesFromState(difference);
							timeTupleMaintenanceEnd = System.nanoTime();
							timeTupleMaintenance += (timeTupleMaintenanceEnd - timeTupleMaintenanceStart);
							numTuplesInRightWindow = numTuplesInRightWindow - difference;
						} 
					
						timeTupleMaintenanceStart = System.nanoTime();
						//add tuple to the right state
						rightState.addTupleToState(newObj);
						timeTupleMaintenanceEnd = System.nanoTime();
						timeTupleMaintenance += (timeTupleMaintenanceEnd - timeTupleMaintenanceStart);
						
						
						if (currentRightSPStatus == SP_ARRIVED_INITIALIZED)
						{							
							rightState.getActiveSP().setStartTuple(rightState.getCurrentWindowTuplesQueue().getLastNode());
							rightState.getActiveSP().setEndTuple(rightState.getCurrentWindowTuplesQueue().getLastNode());
							currentRightSPStatus = SP_TUPLE_START_INITIALIZED;
						}
						else if (currentRightSPStatus == this.SP_TUPLE_START_INITIALIZED)
						{
							rightState.getActiveSP().setEndTuple(rightState.getCurrentWindowTuplesQueue().getLastNode());
						}
												
						Object value = newObj.getCell(rightJoinAttrPos).getValue();
						timeJoinStart = System.nanoTime();
						ResultPackage resultPackage = leftState.probeWithSP(leftJoinAttrPos, value, rightState.getActiveSP().getSP());
						timeJoinEnd = System.nanoTime();
						timeJoin += (timeJoinEnd - timeJoinStart);
						
						if (resultPackage == null) 
						{
							continue;
						}
						
						XATTuple[] matchedTuples = resultPackage.getResultTuples();
						if (matchedTuples != null) 
						{							
							for (int i=0; i<matchedTuples.length; i++) 
							{
								//XATTuple resultTuple = this.generateResultTuple(new XATTuple[] {newObj, matchedTuples[i]});
								XATTuple resultTuple =  matchedTuples[i];
								this.getOutputQueue().enqueue(resultTuple, this);
								totalNumberOfJoinResults++;
							}
						}																	
					}//end else received a regular tuple
				}//end else (processing right side)
			}//end if object dequeued is NOT null
		
		}//end while
		
		double timeOverallEnd = System.nanoTime();
		double timeOverallMilliseconds = (timeOverallEnd - timeOverallStart)/(1000000);//in ms
		double timeOverallNanoseconds = (timeOverallEnd - timeOverallStart);//in ns
		
		if (titleHelper == 1)
		{
			String strTitle =  "Time Overall(nanoseconds)" + "\t"+"Time Overall(milliseconds)" + "\t"+"Time To Join(milliseconds)" + "\t"+"Time SP Maintenance (milliseconds)"+ "\t"+"Time Tuple Maintenance (milliseconds)"+ "\t"+ "Number of both SPs and Tuples" + "\t" +  "Number of Tuples Only" + "\t" + "Number of Sps Only"+"\t" + "# of Tuples in Left Queue" + "\t" + "# of Sps in Left Queue" + "\t"+ "# of Tuples in Right Queue" +"\t" +"# of Sps in Right Queue" + "\t" + "Total # of Join Results";
			titleHelper++;
			Helper.StatsWriteLineToFile("SCJoinstats.txt", strTitle);
		}
		double timeSPMaintenanceMs = timeSPMaintenance /(1000000);//in ms
		double timeTupleMaintenanceMs = timeTupleMaintenance /(1000000);//in ms
		double timeJoinMs = timeJoin/(1000000);//in ms
		
		String strStats =  timeOverallNanoseconds + "\t"+timeOverallMilliseconds +"\t" +timeJoinMs+"\t" +timeSPMaintenanceMs+"\t" +timeTupleMaintenanceMs+"\t"+  (totalTuples + totalSps) + "\t" + totalTuples+ "\t"+totalSps + "\t"+ totalOnlyTuplesLeftQ + "\t"+ totalOnlySpsLeftQ+"\t" +totalOnlyTuplesRightQ + "\t" +totalOnlySpsRightQ + "\t"+ totalNumberOfJoinResults;
			
		Helper.StatsWriteLineToFile("SCJoinstats.txt", strStats);
		resetStatisticsVariables();
		
		return true;
	}
	
	//-----------------------------------------------------------------------------
	
	private void resetStatisticsVariables()
	{
		totalTuples = 0; //tuples + sps
		totalSps = 0; //tuples + sps
		totalOnlyTuplesLeftQ = 0;
		totalOnlySpsLeftQ = 0;
		totalOnlyTuplesRightQ = 0;
		totalOnlySpsRightQ = 0;
		elementsLeftQ = 0;
		elementsRightQ = 0;
		totalNumberOfJoinResults = 0;		
		timeJoin = 0;
		timeSPMaintenance = 0;
		timeTupleMaintenance = 0;
	}
	
	//-----------------------------------------------------------------------------
	
	public void setState(OperatorState[] state) 
	{
		this.states = new OperatorState[state.length];
		for(int i = 0; i < this.states.length;i++)
		{
			this.states[i] = state[i];
		}
		this.states = state;
		this.leftState = (SinglePWSingleEquiSCJoinState) state[0];
		this.rightState = (SinglePWSingleEquiSCJoinState) state[1];
	}
	
	//-----------------------------------------------------------------------------

	public int getLeftJoinAttrPos() 
	{
		return this.leftJoinAttrPos;
	}
	
	//-----------------------------------------------------------------------------
	
	public int getRightJoinAttrPos() 
	{
		return this.rightJoinAttrPos;
	}
	
	//-----------------------------------------------------------------------------
	
	public void setLeftJoinAttrPos(int lJoinAttrPos) 
	{
		this.leftJoinAttrPos = lJoinAttrPos;
	}
	
	//-----------------------------------------------------------------------------
	
	public void setRightJoinAttrPos(int rJoinAttrPos) 
	{
		this.rightJoinAttrPos = rJoinAttrPos;
	}	
	
	//-----------------------------------------------------------------------------

	public String toToolTipString() 
	{
		return "Index SCJoin";
	}
	
	//-----------------------------------------------------------------------------
	
	public boolean hasUnmatchedState() 
	{
		return false;
	}
	
	//-----------------------------------------------------------------------------
	
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return true;
	}
	
	//-----------------------------------------------------------------------------
	
	public boolean cleanAllAccumulatedTuples() 
	{
		return true;
	}
	
	//-----------------------------------------------------------------------------
	
	public int setStateID(int startID) 
	{
		return 0;
	}
	
	//-----------------------------------------------------------------------------
	
	public boolean recomputeUnmatchedState() 
	{
		return true;
	}
	
	//-----------------------------------------------------------------------------
	
	public boolean matchStates(XATStreamOperator op) 
	{
		return true;
	}
	
	//-----------------------------------------------------------------------------
	
	public boolean finalizeOperation()
	{
		return true;
	}
	
	//-----------------------------------------------------------------------------
	
	public boolean getPropagateFlag() 
	{
		return propagateFlag;
	}
	
	//-----------------------------------------------------------------------------
	
	public void setPropagateFlag(boolean pFlag) 
	{
		propagateFlag = pFlag;
	}	
	
	//-----------------------------------------------------------------------------
	
	public int getMemoryLimit() 
	{
		return memoryLimit;
	}
	
	//-----------------------------------------------------------------------------
	
	public void setMemoryLimit(int memLimit) 
	{
		memoryLimit = memLimit;
	}
	
	//-----------------------------------------------------------------------------
	
	public void setLeftWindowSize(int newSize)
	{
		leftWindowSize = newSize;
	}	
	
	//-----------------------------------------------------------------------------
	
	public void setRightWindowSize(int newSize)
	{
		rightWindowSize = newSize;
	}
	
	//-----------------------------------------------------------------------------
	
	public int getPurgeThreshold() 
	{
		return 0;
	}
	
	//-----------------------------------------------------------------------------
	
	public XATQueue getNthInputQueue(int n) 
	{
		return null;
	}
	
	//-----------------------------------------------------------------------------
	
	public void init() 
	{
		XATQueueSchema outSchema = this.getOutputQueue().getSchema();
		if (outSchema == null || outSchema.getLength() == 0) 
		{
			String schemaName = new String();
			outSchema = new XATQueueSchema();
			//Here we expect only one inputQueue existing
			Iterator it = this.getInputQueue().getSchemas().iterator();
			while (it.hasNext()) 
			{
				XATQueueSchema inSchema = (XATQueueSchema) it.next();
				schemaName += "#" + inSchema.getName();
				
				Iterator j = inSchema.getColNames().iterator();
				while(j.hasNext()){
					outSchema.addColumn((String) j.next());
				}
			}
			outSchema.setName(schemaName);
			this.getOutputQueue().setSchema(outSchema);
		}
		computeOutputPositions();

		this.numOutputColsPerStream = new int[outputPositions.length];
		this.numOutputColumns = 0;
		for (int i=0; i<outputPositions.length; i++) {
			int numCols = outputPositions[i].length;
			numOutputColsPerStream[i] = numCols;
			numOutputColumns += numCols;
		}
		this.setHasInit(true);
	}

	//-----------------------------------------------------------------------------
	
	public void init(int outputQueueId) 
	{
		XATQueueSchema outSchema = this.getOutputQueue(outputQueueId).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) 
		{
			String schemaName = new String();
			outSchema = new XATQueueSchema();
			//Here we expect only one inputQueue existing
			Iterator it = this.getInputQueue().getSchemas().iterator();
			while (it.hasNext()) {
				XATQueueSchema inSchema = (XATQueueSchema) it.next();
				schemaName += "#" + inSchema.getName();
				
				Iterator j = inSchema.getColNames().iterator();
				while(j.hasNext()){
					outSchema.addColumn((String) j.next());
				}				
			}
			outSchema.setName(schemaName);
			this.getOutputQueue(outputQueueId).setSchema(outSchema);
		}
		computeOutputPositions();

		this.numOutputColsPerStream = new int[outputPositions.length];
		this.numOutputColumns = 0;
		for (int i=0; i<outputPositions.length; i++) 
		{
			int numCols = outputPositions[i].length;
			numOutputColsPerStream[i] = numCols;
			numOutputColumns += numCols;
		}
		this.setHasInit(true);
	}
	
	//-----------------------------------------------------------------------------
	
	public synchronized void setOutputPositions(int i, int[] ithOutputPositions) 
	{
		if (outputPositions == null)
		{
			outputPositions = new int[this.getInputQueue().getSchemas().size()][];
		}  
		outputPositions[i] = ithOutputPositions;
	}
	
	//-----------------------------------------------------------------------------
	
	public void computeOutputPositions() 
	{
		DagNode node = (DagNode)getNode(); 
		Vector outColNames = getOutputQueue().getSchema().getColNames();
		int outputColPos = 0;

		Iterator inputs = getInputQueue().getSchemas().iterator();
		//store all the input columns
		int j=0;
		while(inputs.hasNext()) 
		{
			Vector outputPosVec = new Vector();
			Vector inputColNames = ((XATQueueSchema)inputs.next()).getColNames(); 
			for (Iterator i1 = outColNames.iterator(); i1.hasNext();)
			{
				Object outputCol = i1.next();
				for (int k = 0; k<inputColNames.size(); k++)
				{
					Object inputCol = inputColNames.get(k);
					if (outputCol.equals(inputCol))
					{
						outputPosVec.add(new Integer(k));
						break;
					}
				}
			}
			
			//convert vector to int[]
			int[] outputPosArray = new int[outputPosVec.size()];
			for (int count = 0; count < outputPosVec.size(); count++){
				outputPosArray[count] = ((Integer)outputPosVec.get(count)).intValue();
			}
			this.setOutputPositions(j, outputPosArray);
			j++;
		}
	}
	
	//-----------------------------------------------------------------------------
	
}
