/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.INode;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.INodeIndex;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.ResultPackage;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.ResultPackageImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.SinglePWSingleEquiJoinState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.SinglePWSingleEquiJoinStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.TNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.TupleQueue.Node;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * @author Rimma V. Nehme
 *
 */



public class IndexSinglePWSingleEquiSCJoinStateImp 
extends SinglePWSingleEquiJoinStateImp  implements SinglePWSingleEquiSCJoinState 
{
	protected static int FILTER_PROBE = 1;  //first filter, based on sp, then probe
	
	//protected static int CURRENT_JOIN_APPROACH = PROBE_FILTER;
	protected static int CURRENT_JOIN_APPROACH = FILTER_PROBE;
	
	protected ArrayList currentWindowSPs = new ArrayList();	
	protected TupleQueue currentWindowTuples = new TupleQueue();
	protected SpToTuplesMapping activeSP;
	protected SpNode lastSpNode;

	protected RNode[] r_node_array = new RNode[100]; //max number of supported roles is assumed to be 100 (for now)
	
	protected int currentWindowTuplesCount;
	protected int currentWindowSPsCount;
	
	
	public IndexSinglePWSingleEquiSCJoinStateImp()
	{
		super();
		currentWindowTuplesCount = 0;
		currentWindowSPsCount = 0;
	}	
	
	private class SpNode
	{
		int tuple_counter; // counter for the number of tuples with sp policy
		SpToTuplesMapping sp_and_tuples; //includes physical sp and the pointers to tuples that it applies (in the tuple queue)
		SpINode sp_i_node;
		SpNode next_sp_node;
	}
	
	private class SpINode
	{
		SpNode sp_node; //pointer to sp_node in the window
		SpINode next_sp_i_node; //pointer to the next sp_i_node
	}
		
	private class RNode
	{		
		SpINode r_head;
		SpINode r_tail;		
		int count; //count of sp_i_nodes 
		
		public RNode()
		{
			r_head = new SpINode();
			r_tail = new SpINode();		
			count = 0;			
		}
	}
	
	public class TupleToSpNodeMapping
	{
		 protected SpNode spNode;        		 // sp of the tuple
		 protected XATTuple tuple;           // the tuple
		 
		 
		 public TupleToSpNodeMapping(SpNode valueSpNode, XATTuple valueTuple)
		 {
			 spNode = valueSpNode;
			 tuple = valueTuple;
		 }
		 
		 public SpNode getSPNode()
		 {
			 return spNode;
		 }
		 
		 public XATTuple getTuple()
		 {
			 return tuple;
		 }
	}
	
	public void addSPToState(XATSecPunctImp sp) 
	{
		activeSP = new SpToTuplesMapping(sp);
		//---------------------
		//create new sp_node
		//---------------------
		SpNode sp_node = new SpNode();
		sp_node.sp_and_tuples = activeSP; //assign physical sp
		sp_node.tuple_counter = -1; //number of tuples with this policy is yet unknown
		
		//store pointer to the last sp_node
		lastSpNode = sp_node;
		//-------------------
		//create sp_i_node
		//-------------------
		SpINode sp_i_node = new SpINode();
		sp_i_node.sp_node = sp_node;
		
		//------------------------------------------		
		//link sp_i_node to the appropriate r_nodes
		//------------------------------------------
		for(int j=0; j < sp.acmModels.length;j++)
		{
			AccessControlModel acm = sp.acmModels[j];
 			int r_node_index =Integer.parseInt(acm.value.substring(2, acm.value.length() - 1)); //remove the R prefix from the role definition

			//-----------------------------
			//make an entry for the r-node
			//-----------------------------
			if(r_node_array[r_node_index] == null)
			{
				r_node_array[r_node_index] = new RNode();
			}
			else
			{
				//System.out.print("already in r-node array\n");
			}
			r_node_array[r_node_index].r_tail.next_sp_i_node = sp_i_node;
			r_node_array[r_node_index].r_tail = sp_i_node;
			if (r_node_array[r_node_index].count == 0)
			{
				r_node_array[r_node_index].r_head = sp_i_node;
			}
			r_node_array[r_node_index].count++;
			
		}//end for
		
		currentWindowSPsCount++;
	}
	
	
	public void addTupleToState(XATTuple tuple) 
	{
		//buffer sp only if its policy has been received
		if (lastSpNode != null)
		{
			currentWindowTuples.enqueue(new TupleToSpNodeMapping(lastSpNode, tuple));	
			this.activeSP.setEndTuple(currentWindowTuples.getLastNode());
			if(this.activeSP.getStartTuple() == null)
			{
				this.activeSP.setStartTuple(currentWindowTuples.getLastNode());
			}
			currentWindowTuplesCount++;
		}
	}
	
	public void removeTuplesFromState(int count) 
	{
		for(int i=0; i < count-1; i++)
		{
			currentWindowTuples.dequeue();		
			currentWindowTuplesCount--;
		}
	}
	
	
	public SpToTuplesMapping getActiveSP() 
	{
		return activeSP;		
	}
	
	
	public ResultPackage probeWithSP(int currentStateTupleJoinPosition, Object value, XATSecPunctImp valueSP) 
	{
		ResultPackage rPackage = null;	
		ArrayList resultTupleList = new ArrayList();
		
		//should always be true
		if (CURRENT_JOIN_APPROACH == FILTER_PROBE)
		{
			for(int j=0; j < valueSP.acmModels.length;j++)
			{			
				AccessControlModel acm = valueSP.acmModels[j];
	 			int r_node_index =Integer.parseInt(acm.value.substring(2, acm.value.length() - 1)); //remove the R prefix from the role definition
	
				//-----------------------------
				//find an r-node entry
				//-----------------------------
				if(r_node_array[r_node_index] == null)
				{
					//nothing to join with, return immediately
					continue;
				}
				else
				{
					if (r_node_array[r_node_index].count == 0)
					{
						//nothing to join with, return immediately
						continue;
					}
					
					//scan through all sp_i_nodes for the r_node					
					SpINode currentSpInode = r_node_array[r_node_index].r_head;
					for(int p=0; p < r_node_array[r_node_index].count; p++)
					{
						//access the sp_node 
						SpNode currentSpNode = currentSpInode.sp_node;
						
						//check its starting role (we may be able to skip it thanks to our "skipping" rule)
												
						AccessControlModel currentAcm = currentSpNode.sp_and_tuples.getSP().acmModels[0];
			 			int probe_r_node_index = Integer.parseInt(currentAcm.value.substring(2, currentAcm.value.length() - 1)); //remove the R prefix from the role definition
						
			 			if (probe_r_node_index < r_node_index)
			 			{
			 				//skip it (must have already joined)
			 				continue;
			 			}
						                                              
						//join current tuple with compatible tuples
			 			Node currentTupleNode = currentSpNode.sp_and_tuples.getStartTuple();
			 			Node lastTupleNode = currentSpNode.sp_and_tuples.getEndTuple();
			 					 			
			 			//scan and joing with the tuples
						if(currentTupleNode != null)
						{
							while(currentTupleNode != lastTupleNode.next)
							{	
								TupleToSpNodeMapping mapping = (TupleToSpNodeMapping)currentTupleNode.item;
								XATTuple tpThis = (XATTuple)mapping.getTuple();
								if (tpThis.getCell(currentStateTupleJoinPosition).toString().equalsIgnoreCase(value.toString()))
								{
									resultTupleList.add(tpThis);				
								}
								currentTupleNode = currentTupleNode.next;
								if (currentTupleNode == null)
								{
									break;
								}
							}//end while	
						}//end if current node != null						
					}
				}											
			}//end for			
		}
		
		//System.out.println(" resultTupleList size... "+ resultTupleList.size()+"\n");
		if (resultTupleList.size() != 0) 
		{
			rPackage = new ResultPackageImp();
			rPackage.setResultTuples((XATTuple[]) resultTupleList.toArray(new XATTuple[] {}));
		}
		
		return rPackage;
		
	}
	
	public TupleQueue getCurrentWindowTuplesQueue()
	{
		return this.currentWindowTuples;
	}
	
	public int getCurrentWindowTuplesCount()
	{
		return this.currentWindowTuplesCount;
	}
	
	public int getCurrentWindowSPsCount()
	{
		return this.currentWindowSPsCount;
	}
				
}
