/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamProjectOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;

/**
 * @author Rimma V. Nehme
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class SPAware_StreamProjectOperatorImp
	extends XATSingleQueueStreamOperatorImp
	implements StreamProjectOperator 
{

	private int totalTuples = 0; //tuples + sps
	private int tuplesPassed = 0;
	private int tuplesDropped = 0;
	private int numSecPunct = 0;
	static int titleHelper = 1;
	
	String[] retainAttrNames; //assumes no name conflict
	//	retainAttrPos[0] is the index of the input queue schema that matches the first column of the output queue schema. 
	//  int[] retainAttrPos;  

	public SPAware_StreamProjectOperatorImp(SPAware_StreamProjectOperatorImp oldop, XATNode newnode) 
	{
		super(oldop, newnode);
		String[] rAttrNames = oldop.getRetainAttrNames();
		retainAttrNames = new String[rAttrNames.length];
		for (int i=0; i<rAttrNames.length; i++) 
		{
			retainAttrNames[i] = rAttrNames[i];
		}
	}
	
	/**
	 * constructor
	 * @param init
	 */
	public SPAware_StreamProjectOperatorImp(XATNode init) 
	{
		super(init);
	}
	
	public String[] getRetainAttrNames() 
	{
		return retainAttrNames;
	}
	
	public void setRetainAttrName(int pos, String newName) 
	{
		retainAttrNames[pos] = newName;
	}
	
	public void setRetainAttrNames(String[] attrs) {
		retainAttrNames = attrs;
	}
	
	//overload super method
	public void init(int i) 
	{
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) 
		{
			return;
		}
		XATQueueSchema outSchema = this.getOutputQueue(i).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) 
		{
			this.getOutputQueue(i).setSchema(inSchema);
		}
		computeOutputPositions();//no need to overload since all the output queue are the same
		numOutputColumns = outputPositions[0].length;
		setHasInit(true);
	}
		
	public boolean run() throws Exception 
	{
		if (!hasInit()) 
		{
			init();
			if (!hasInit()) 
			{
				return false;
			}
		}
		
		//modified by Song Wang to support multiple output queue
		for (int i=0;i<this.getOutputQueueCount();i++)
		{
			if(this.getOutputQueue(i).getSchema() == null)
			{
				init(i);
			}
		}
		
		
		int numColumns = retainAttrNames.length;		
		int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		//int howMany = 100;

		/** get the tuples */
		XATTuple[] tuples = this.getInputQueue().dequeue(howMany, this);

		/*start recording time*/
		//double timeOverallStart = System.currentTimeMillis();
		double timeOverallStart = System.nanoTime();
		
		for (int i = 0; i < tuples.length; i++) 
		{
			totalTuples++;
			XATTuple tuple = tuples[i];
			
			if (tuple.isSecPunct()) 
			{
				this.getOutputQueue().enqueue(tuple, this);
				numSecPunct++;
				continue;
			}
			
			// for distributed migration
			if(tuple.getMaxTimeStamp() == -1) 
			{
				this.processSyncTuple();
				continue;
			}
			
			if (tuple.getObjTypeID()==0) 
			{
				this.getOutputQueue().enqueue(this.generateResultTuple(tuple), this);
				tuplesPassed++;
			}
			else 
			{
				totalTuples++;
				XATPunct newPunct = new XATPunctImp(numColumns);
				long[] inputTimeStamp = tuple.getTimeStamp();
				long[] newTimeStamp = new long[numColumns];

				for (int j=0; j<numColumns; j++) 
				{
					newPunct.setCell(j, tuple.getCell(outputPositions[0][j]));
					newTimeStamp[j] = inputTimeStamp[outputPositions[0][j]];
				}
				newPunct.setTimeStamp(newTimeStamp);
				newPunct.setMinTimeStamp(tuple.getMinTimeStamp());
				newPunct.setMaxTimeStamp(tuple.getMaxTimeStamp());
				int[] punctTypeArray = ((XATPunct)tuple).getPunctTypeArray();
				int[] newPunctTypeArray = new int[numColumns];
				for (int j=0; j<numColumns; j++) 
				{
					newPunctTypeArray[j] = punctTypeArray[outputPositions[0][j]];
				}
				newPunct.setPunctTypeArray(newPunctTypeArray);
				this.getOutputQueue().enqueue(newPunct, this);
			}
		}
		
		//double timeOverallEnd = System.currentTimeMillis();
		double timeOverallEnd = System.nanoTime();
		double timeOverallMilliseconds = (timeOverallEnd - timeOverallStart)/(1000000);//in ms
		double timeOverallNanoseconds = (timeOverallEnd - timeOverallStart);//in ns
		
		if (titleHelper == 1)
		{
			String strTitle =  "Time Overall(nanoseconds)" + "\t"+"Time Overall(milliseconds)" + "\t"+ "Number of SPs" + "\t" +  "Number of Tuples Forwarded" + "\t" + "Number of Tuples Dropped"  + "\t" + "Total Number of Elements Processed (tuples + SPs)";
			titleHelper++;
			//Helper.StatsWriteLineToFile("SPAware_Project_Stats.txt", strTitle);
		}
		
		//String strStats =  timeOverallNanoseconds + "\t"+timeOverallMilliseconds + "\t"+ numSecPunct + "\t" +  tuplesPassed + "\t" + tuplesDropped  + "\t" + totalTuples;			
		//Helper.StatsWriteLineToFile("SPAware_Project_Stats.txt", strStats);
		
		String strStats = "SP-AWARE PROJECT: timeOverallMilliseconds = " + timeOverallMilliseconds + ", numSecPunct = "+ numSecPunct + ", tuplesPassed = " +  tuplesPassed + ", tuplesDropped = " + tuplesDropped  + ", totalTuples = " + totalTuples;		
		System.out.println(strStats + "\n");
		
		resetStatisticsVariables();
		return true;
	}
	
	
	private void resetStatisticsVariables()
	{
		numSecPunct = 0; 
		tuplesPassed = 0; 
		tuplesDropped = 0;
		totalTuples = 0;
	}

	public void init() 
	{
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) 
		{
			return;
		}
		
		XATQueueSchema outSchema = this.getOutputQueue().getSchema();
		if (outSchema == null || outSchema.getLength()==0) 
		{
			outSchema = new XATQueueSchema();
			outSchema.setName(this.getInputQueue().getSchema().getName());
			for (int i=0; i < retainAttrNames.length; i++) 
			{
				outSchema.addColumn(retainAttrNames[i]);
			}
			this.getOutputQueue().setSchema(outSchema);
		}
		this.computeOutputPositions();
		numOutputColumns = outputPositions[0].length;
		this.setHasInit(true);
	}

	public String toToolTipString() 
	{
		return "Security-Aware Project";
	}
	
	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input queues.
	 * Inside method run() the numbers are computed based on the property
	 * "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler,
	 * while in cleanAccumulatedTuples, all tuples in input queues that have their max timesamp
	 * smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return false;
	}

	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() 
	{
		return false;
	}
	
	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() 
	{
		return false;
	}

	public boolean matchStates(XATStreamOperator op) 
	{
		return false;
	}
	
	public boolean recomputeUnmatchedState() 
	{
		return true;
	}
	
	public int setStateID(int nextid) 
	{
		return nextid;
	}
	public boolean finalizeOperation()
	{
		return true;
	}
}
