/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.INode;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.INodeIndex;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.ResultPackage;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.ResultPackageImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.SinglePWSingleEquiJoinState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.SinglePWSingleEquiJoinStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.TNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.TupleQueue.Node;
//import edu.wpi.cs.dsrg.xmldb.xat.common.tagging.TagPunctuationImp;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * @author Rimma V. Nehme
 *
 */

public class SinglePWSingleEquiSCJoinStateImp 
extends SinglePWSingleEquiJoinStateImp  implements SinglePWSingleEquiSCJoinState 
{
	protected static int PROBE_FILTER = -1; //first join, then filter based on the sp
	protected static int FILTER_PROBE = 1;  //first filter, based on sp, then probe
	
	
	//protected static int CURRENT_JOIN_APPROACH = PROBE_FILTER;
	protected static int CURRENT_JOIN_APPROACH = FILTER_PROBE;
	
	protected ArrayList currentWindowSPs = new ArrayList();	
	protected TupleQueue currentWindowTuples = new TupleQueue();

	protected SpToTuplesMapping activeSP;	
	
	protected int currentWindowTuplesCount;
	protected int currentWindowSPsCount;
	
	public SinglePWSingleEquiSCJoinStateImp()
	{
		super();
		currentWindowTuplesCount = 0;
		currentWindowSPsCount = 0;
	}
	
	public void addSPToState(XATSecPunctImp sp) 
	{
		activeSP = new SpToTuplesMapping(sp);
		currentWindowSPs.add(activeSP);
		currentWindowSPsCount++;
	}
		
	public void addTupleToState(XATTuple tuple) 
	{
		currentWindowTuples.enqueue(new TupleToSpMapping(this.activeSP.getSP(),tuple));		
		currentWindowTuplesCount++;
	}
	
	public void removeTuplesFromState(int count) 
	{
		for(int i=0; i<count-1;i++)
		{
			currentWindowTuples.dequeue();		
			currentWindowTuplesCount--;
		}
	}
	
	
	public SpToTuplesMapping getActiveSP() 
	{
		return activeSP;		
	}
	
	
	public ResultPackage probeWithSP(int currentStateTupleJoinPosition, Object value, XATSecPunctImp valueSP) 
	{
		ResultPackage rPackage = null;	
		ArrayList resultTupleList = new ArrayList();
		
		
		if (CURRENT_JOIN_APPROACH == PROBE_FILTER)
		{		
			Node currentNode = currentWindowTuples.getFirstNode();
			Node lastNode = currentWindowTuples.getLastNode();
			
			if (currentNode == null)
				return rPackage;
	
			//System.out.println("Probe: " + value + "\n");
	
			if(currentNode != null)
			{
			while(currentNode != lastNode.next)
			{	
				TupleToSpMapping mapping = (TupleToSpMapping)currentNode.item;
				XATTuple tpThis = (XATTuple)mapping.getTuple();
				//System.out.println(tpThis.getCell(currentStateTupleJoinPosition) + " vs "+value + "\n");
				if ((tpThis.getCell(currentStateTupleJoinPosition).toString() == value.toString()) ||
				    (tpThis.getCell(currentStateTupleJoinPosition).toString().startsWith(value.toString())))
				{
					//System.out.println(tpThis.getCell(currentStateTupleJoinPosition) + " =  "+value + "\n");
					
					//got a join result, now let's check the security settings to see if the two tuples can be joined
					//check the match
					boolean doMatch = false;
					for(int j=0; j<mapping.getSP().acmModels.length;j++)
					{
						AccessControlModel acm1 = mapping.getSP().acmModels[j];
						for(int k=0; k<valueSP.acmModels.length;k++)
						{
							AccessControlModel acm2 = valueSP.acmModels[k];							
							if ((acm1.type == acm2.type) && (acm1.value.toString().equalsIgnoreCase(acm2.value.toString())))
							{
								doMatch = true;
								break;
							}
						}
						if (doMatch == true)
						{
							break;
						}
					}
					
					if (doMatch == true)
					{
						//sps match, so record the join result
						resultTupleList.add(tpThis);
					}
				}
				currentNode = currentNode.next;
				if (currentNode == null)
				{
					break;
				}
			}//end while
			}
		}
		else if (CURRENT_JOIN_APPROACH == FILTER_PROBE)
		{
			for(int i=0; i<currentWindowSPs.size();i++)
			{
				//check which security punctuations match to the sp of the current value
				//when sps match, then perform the join with the tuples that the sp corresponds to
				SpToTuplesMapping currentSPMapping = (SpToTuplesMapping)currentWindowSPs.get(i);
				XATSecPunctImp currentSP = currentSPMapping.getSP();
								
				//check the match
				for(int j=0; j<currentSP.acmModels.length;j++)
				{
					AccessControlModel acm1 = currentSP.acmModels[j];
					for(int k=0; k<valueSP.acmModels.length;k++)
					{
						AccessControlModel acm2 = valueSP.acmModels[k];
						
						if ((acm1.type == acm2.type) && (acm1.value.toString().equalsIgnoreCase(acm2.value.toString())))
						{
							//System.out.println("match in sps found");
							
							//perform the join
							Node currentNode = currentSPMapping.getStartTuple();
							Node lastNode = currentSPMapping.getEndTuple();
							
							if(currentNode != null)
							{
								while(currentNode != lastNode.next)
								{	
									TupleToSpMapping mapping = (TupleToSpMapping)currentNode.item;
									XATTuple tpThis = (XATTuple)mapping.getTuple();
									//System.out.println(tpThis.getCell(currentStateTupleJoinPosition) + " vs "+value + "\n");
									if ((tpThis.getCell(currentStateTupleJoinPosition).toString() == value.toString()) ||
									    (tpThis.getCell(currentStateTupleJoinPosition).toString().startsWith(value.toString())))
									{
										resultTupleList.add(tpThis);				
									}
									currentNode = currentNode.next;
									if (currentNode == null)
									{
										break;
									}
								}//end while	
							}//end if current node != null
						}//end if 2 sps match
					}
				}				
			}
		}
		
		
		
		//System.out.println(" resultTupleList size... "+ resultTupleList.size()+"\n");
		if (resultTupleList.size() != 0) 
		{
			rPackage = new ResultPackageImp();
			rPackage.setResultTuples((XATTuple[]) resultTupleList.toArray(new XATTuple[] {}));
		}
		

		return rPackage;
		
	}
	
	public TupleQueue getCurrentWindowTuplesQueue()
	{
		return this.currentWindowTuples;
	}
	
	public int getCurrentWindowTuplesCount()
	{
		return this.currentWindowTuplesCount;
	}
	
	public int getCurrentWindowSPsCount()
	{
		return this.currentWindowSPsCount;
	}
				
}
