/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Jan 7, 2006
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.security;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.XATSecPunctImp.SEC_PUNCT_ENFORCEMENT_TYPE;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.XATSecPunctImp.SEC_PUNCT_TYPE;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.fence.FenceOperatorSpecification;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;
/**
 * @author Rimma V. Nehme
 *
 */
public class StreamSecurityShieldPlusOperatorImp 
	extends XATSingleQueueStreamOperatorImp 
	implements StreamSecurityShieldOperator 
	
{

	//------------------------------------------------------------------------------------
	
	protected boolean propagateFlagFurther = true;
	protected Hashtable accessControlModelsTable = new Hashtable();
	protected Hashtable apOperatorsTable = new Hashtable();
	private static XATSecPunctImp activeSP = null; 
	private static boolean matchFound = false;
	private int totalTuples = 0; //tuples + sps
	private int tuplesPassed = 0;
	private int tuplesDropped = 0;
	private int numDataSecPunct = 0;
	static int titleHelper = 1;	
	private ExpressionStr securityExpression;
	
	//------------------------------------------------------------------------------------

	Hashtable<String, ArrayList> hashBufferDspsForStreams = new Hashtable<String, ArrayList>(); 
	Hashtable<String, ArrayList> hashBufferQspsForStreams = new Hashtable<String, ArrayList>(); 
		
	Hashtable<String, Hashtable> hashRolesInDspQspIntersectionForStreams = new Hashtable<String,Hashtable>();
	
	//------------------------------------------------------------------------------------
	
	public StreamSecurityShieldPlusOperatorImp(XATNode init) 
	{
		super(init);
		tuplesPassed = 0;
		tuplesDropped = 0;
		numDataSecPunct = 0;
		totalTuples = 0;
	}

	//------------------------------------------------------------------------------------
	
	public boolean run() throws Exception 
	{	
		/*start recording time*/
		//double timeOverallStart = System.currentTimeMillis();
		double timeOverallStart = System.nanoTime();
		
		/**
		 * first figure out how many tuples to dequeue, the selectivity, and 
		 * result delay  
		 */
		XATQueueSchema inSchema;
		int inLength;
		
		if ((inSchema = this.getInputQueue().getSchema()) == null)
		{
			return false;
		}
	
		inLength = inSchema.getLength();
		
		//modified by Song Wang to support multiple output queue
		for (int i = 0; i < this.getOutputQueueCount(); i++)
		{
			if(this.getOutputQueue(i).getSchema() == null)
			{
				init(i);
			}
		}
		
		for (int i = 0; i < this.inputQueue.size(); i++) 
		{	
			//------------------------------------------------------------------------					
			//int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
			int howMany  = 30; //rimma temp "hack"
			
			//for qsp stream, process a qsp at a time
			if (i == (this.inputQueue.size() - 1))
			{
				howMany = 1; //qsp stream
			}
			
			XATTuple[] tuples = (this.getInputQueue(i)).dequeue(howMany, this);
			System.out.println("Stream ["+i+"] dequeued [" + tuples.length + "] tuples");
			
			int numResultTuples = 0; //number of results tuples for this stream
			
			for (int p = 0; p < tuples.length; p++)
			{
				XATTuple currTuple = (XATTuple)tuples[p];
				//------------------------------
				// SECURITY PUNCTUATION
				//------------------------------
				if (currTuple.isSecPunct()) 				
				{
					String streamId = "" + i;
					String qspStreamId = "" + (this.inputQueue.size() - 1); //the query sp stream id (used for retrieving appropriate buffer)
																			//to intersect sps
					SFA_processSecPunct(currTuple, streamId, qspStreamId);
				}
				//------------------------------
				// REGULAR TUPLE
				//------------------------------
				else
				{
					if (SFA_processTuple(currTuple, i))
					{
						numResultTuples++;
					}
				}
			}//end for (processing of all tuples for the current stream)
			
			System.out.println("Stream ["+i+"] enqueued [" + numResultTuples + "] tuples\n");
		}
				
		//double timeOverallEnd = System.currentTimeMillis();
		double timeOverallEnd = System.nanoTime();
		double timeOverallMilliseconds = (timeOverallEnd - timeOverallStart)/(1000000);//in ms
		double timeOverallNanoseconds = (timeOverallEnd - timeOverallStart);//in ns
		
		if (titleHelper == 1)
		{
			String strTitle =  "Time Overall(nanoseconds)" + "\t"+"Time Overall(milliseconds)" + "\t"+ "Number of SPs" + "\t" +  "Number of Tuples Forwarded" + "\t" + "Number of Tuples Dropped"  + "\t" + "Total Number of Elements Processed (tuples + SPs)";
			titleHelper++;
			//Helper.StatsWriteLineToFile("SSstats.txt", strTitle);
		}
		
		//String strStats =  timeOverallNanoseconds + "\t"+timeOverallMilliseconds + "\t"+ numSecPunct + "\t" +  tuplesPassed + "\t" + tuplesDropped  + "\t" + totalTuples;			
		//Helper.StatsWriteLineToFile("SSstats.txt", strStats);
		
		String strStats = "SECURITY SHIELD PLUS: timeOverallMilliseconds = " + timeOverallMilliseconds + ", numDataSecPunct = "+ numDataSecPunct + ", tuplesPassed = " +  tuplesPassed + ", tuplesDropped = " + tuplesDropped  + ", totalTuples = " + totalTuples;				
		System.out.println(strStats + "\n");
		
		resetStatisticsVariables();
		
		//garbage collection
		System.gc();
		
		return true;
	}
	
	//------------------------------------------------------------------------------------
	
	private void resetStatisticsVariables()
	{
		numDataSecPunct = 0; 
		tuplesPassed = 0; 
		tuplesDropped = 0;
		totalTuples = 0;
	}
	
	//------------------------------------------------------------------------------------
	
	//overload super method
	public void init(int i) 
	{
		XATQueueSchema inSchema = this.getInputQueue().getSchema();
		if (inSchema == null || inSchema.getLength() == 0) 
		{
			return;
		}
		XATQueueSchema outSchema = this.getOutputQueue(i).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) 
		{
			this.getOutputQueue(i).setSchema(inSchema);
		}
		computeOutputPositions();//no need to overload since all the output queue are the same
		numOutputColumns = outputPositions[0].length;
		setHasInit(true);
	}

	//------------------------------------------------------------------------------------
	
	public String toToolTipString() 
	{
		return "Security Shield Plus";
	}
	
	//------------------------------------------------------------------------------------

	public ExpressionStr getExpression() 
	{
		return securityExpression;
	}
	
	//------------------------------------------------------------------------------------
	
	public void setExpression(ExpressionStr expression) 
	{
		this.securityExpression = expression;
	}
	
	//------------------------------------------------------------------------------------
	
	public void setPropagateSPFurtherFlag(String value) 
	{
		if(value.toLowerCase().contains("true"))
		{
			this.propagateFlagFurther = true;
		}
		else if(value.toLowerCase().contains("false"))
		{ 
			this.propagateFlagFurther = false;
		}
		else
		{
			//by default - propagate the security punctuations further
			this.propagateFlagFurther = true;
		}
	}
	
	//------------------------------------------------------------------------------------
	
	/**
	 * The method cleanAccumulatedTuples() is similar to the method run(),
	 * except the way of figuring out how many tuples to dequeue from input queues.
	 * Inside method run() the numbers are computed based on the property
	 * "NUMBER_OF_TUPLES_TO_DEQUEUE" set by the system scheduler,
	 * while in cleanAccumulatedTuples, all tuples in input queues that have their max timesamp
	 * smaller than the given maxtimestamp should be dequened.
	 */
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return true;
	}

	//------------------------------------------------------------------------------------
	
	/*
	 * process all accumulated tuples in this operators' input queues.
	 */
	public boolean cleanAllAccumulatedTuples() 
	{
		return true;
	}
	
	//------------------------------------------------------------------------------------
	
	/*
	 * check if this operator has any unmatched states.
	 * return: TRUE if this operator has unmatched states, FALSE otherwise.
	 */
	public boolean hasUnmatchedState() 
	{
		return false;
	}

	//------------------------------------------------------------------------------------
	
	public boolean matchStates(XATStreamOperator op) 
	{
		return false;
	}
	
	//------------------------------------------------------------------------------------
	
	public boolean recomputeUnmatchedState() 
	{
		return true;
	}
	
	//------------------------------------------------------------------------------------
	
	public int setStateID(int nextid) 
	{
		return nextid;
	}
	
	//------------------------------------------------------------------------------------
	
	public boolean finalizeOperation()
	{
		return true;
	}
	
	//------------------------------------------------------------------------------------
	
	//set associated processing operators
	public void setAssociatedProcessingOperators(Hashtable newApOperatorsTable)
	{
		apOperatorsTable = newApOperatorsTable;
	}
	
	//------------------------------------------------------------------------------------
	
	//set associated access control models
	public void setAssociatedAccessControlModels(Hashtable newAccessControlModels)
	{
		accessControlModelsTable = newAccessControlModels;
	}
	
	//------------------------------------------------------------------------------------
	
	public String toShortString() 
	{
		StringBuffer sb = new StringBuffer();
		sb.append("\nSS state: \n");
		sb.append("propagate flag further: "+this.propagateFlagFurther+"\n");
		
		for (Iterator it = accessControlModelsTable.entrySet().iterator(); it.hasNext(); )
		{
            Map.Entry e = (Map.Entry) it.next();
            sb.append( e.getKey() + "=> " + "\n");

            Vector acmVector = (Vector)e.getValue();
            for(int i=0; i < acmVector.size();i++)
            {            	            
	            StreamSecurityShieldOperatorACMSettings acm = (StreamSecurityShieldOperatorACMSettings)acmVector.get(i);
	            sb.append("\t"+i +":\n");
	            sb.append("\t\tname: " + acm.name + "\n");
	            sb.append("\t\ttype: " + acm.type + "\n");
	            sb.append("\t\tvalue: " + acm.value + "\n");
            }
            sb.append("\n");
        }
		return sb.toString();
	}
	
	//------------------------------------------------------------------------------------
	
	private void SFA_processSecPunct(XATTuple currTuple, 
			   					     String streamId, 
			   					     String qspStreamId)
	{
		//make sure it is a sec punct
		if (currTuple.isSecPunct())
		{
			XATSecPunctImp currentSp = (XATSecPunctImp)currTuple;					

			//-------------------------------
			// SP TYPE
			//-------------------------------

			//dsp
			if (currentSp.getSecPunctType() == SEC_PUNCT_TYPE.DATA_SP)
			{
				//remove (old) dsp from buffer (dsp)
				emptySpFromBufferByStream(streamId, currentSp.getSecPunctType());

				//add (new) dsp to buffer(dsp)
				addSpToBufferByStream(streamId, currentSp);

				//perform intersection with opposite buffer
				Hashtable hashRolesInDspQspIntersectionForStream = intersectDspWithQsp(currentSp, qspStreamId);

				//store the intersection for the stream
				hashRolesInDspQspIntersectionForStreams.put(streamId, hashRolesInDspQspIntersectionForStream);
			}

			//qsp
			else if (currentSp.getSecPunctType() == SEC_PUNCT_TYPE.QUERY_SP)
			{
				//remove (old) qsp from buffer (qsp)
				emptySpFromBufferByStream(streamId, currentSp.getSecPunctType());

				//add qsp to buffer(qsp)
				addSpToBufferByStream(streamId, currentSp);

				//perform intersection with opposite buffer
				Hashtable hashRolesInDspQspIntersectionForStream = intersectQspWithDsp(currentSp, streamId);

				//store the intersection with all data streams (dsps)
				hashRolesInDspQspIntersectionForStreams = hashRolesInDspQspIntersectionForStream;				
			}
		}
	}

	//--------------------------------------------------------------------------------------
	
	private void emptySpFromBufferByStream(String streamId, SEC_PUNCT_TYPE spType)
	{
		ArrayList arrBuffer = null;
		
		if (spType == SEC_PUNCT_TYPE.DATA_SP)
		{
			arrBuffer = this.hashBufferDspsForStreams.get(streamId);
		}
		else if (spType == SEC_PUNCT_TYPE.QUERY_SP)
		{
			arrBuffer = this.hashBufferQspsForStreams.get(streamId);
		}
		
		if (arrBuffer != null)
		{		
			if (arrBuffer.size() > 0)
			{
				arrBuffer.clear();
			}
		}
	}
	
    //--------------------------------------------------------------------------------------
	
	private void addSpToBufferByStream(String streamId, XATSecPunctImp sp)
	{
		ArrayList arrBuffer = null;
		
		if (sp.getSecPunctType() == SEC_PUNCT_TYPE.DATA_SP)
		{
			arrBuffer = this.hashBufferDspsForStreams.get(streamId);
			if (arrBuffer == null)
			{
				arrBuffer = new ArrayList();
				hashBufferDspsForStreams.put(streamId, arrBuffer);
			}
		}
		else if (sp.getSecPunctType() == SEC_PUNCT_TYPE.QUERY_SP)
		{
			arrBuffer = this.hashBufferQspsForStreams.get(streamId);
			if (arrBuffer == null)
			{
				arrBuffer = new ArrayList();
				hashBufferQspsForStreams.put(streamId, arrBuffer);
			}
		}
		
		arrBuffer.add(sp); //add sp
	}
		
	//------------------------------------------------------------------------------------

	private Hashtable intersectDspWithQsp(XATSecPunctImp dsp, String qspStreamId)
	{
		Hashtable hashRolesInIntersection = new Hashtable();
		
		AccessControlModel [] dspAcmModels = dsp.getAccessControlModels();
		
		ArrayList arrBufferQsps = this.hashBufferQspsForStreams.get(qspStreamId);
		
		if (arrBufferQsps != null)
		{
			//iterate through the opposite (qsp) buffer
			for (int i = 0; i < arrBufferQsps.size(); i++)
			{
				XATSecPunctImp qsp = (XATSecPunctImp)arrBufferQsps.get(i);
				AccessControlModel [] qspAcmModels = qsp.getAccessControlModels();
				
				//intersect the roles
				hashRolesInIntersection = intersectRolesInAcm(dspAcmModels,
															  qspAcmModels,
															  hashRolesInIntersection);
			}
		}
		return hashRolesInIntersection;
	}
	
	//------------------------------------------------------------------------------------

	private Hashtable intersectQspWithDsp(XATSecPunctImp qsp, String streamId)
	{
		Hashtable hashRolesInDspQspIntersectionForStream = new Hashtable();		
		
		AccessControlModel [] qspAcmModels = qsp.getAccessControlModels();
		
		int streamsCount = Integer.parseInt(streamId);
		//streamId passed here is the stream id of qsps (which is assumed 
		//to be the last stream in the streams config specification
		for (int k = 0; k < streamsCount; k++)
		{
			Hashtable hashRolesInIntersection = new Hashtable();
			String dataStreamId = "" + k;
			ArrayList arrBufferQsps = this.hashBufferDspsForStreams.get(dataStreamId);
			
			//iterate through the opposite (qsp) buffer
			for (int i = 0; i < arrBufferQsps.size(); i++)
			{
				XATSecPunctImp dsp = (XATSecPunctImp)arrBufferQsps.get(i);
				AccessControlModel [] dspAcmModels = dsp.getAccessControlModels();
				
				//intersect the roles
				hashRolesInIntersection = intersectRolesInAcm(dspAcmModels,
															  qspAcmModels,  
															  hashRolesInIntersection);
			}
			
			hashRolesInDspQspIntersectionForStream.put(dataStreamId, hashRolesInIntersection);
		}
		return hashRolesInDspQspIntersectionForStream;
	}
	
	//------------------------------------------------------------------------------------
	
	Hashtable intersectRolesInAcm(AccessControlModel[] arrOneAcm, 
								  AccessControlModel[] arrTwoAcm,
								  Hashtable hashRolesInIntersection)
	{
		Hashtable hashNewRolesInIntersection = new Hashtable();
		
		for (int i = 0; i < arrOneAcm.length; i++)
		{
			AccessControlModel oneAcm = arrOneAcm[i];
			String oneRoles = oneAcm.getACMValue();
			oneRoles = oneRoles.replace("{", "");
			oneRoles = oneRoles.replace("}", "");			
			String [] oneRoleArr = oneRoles.split(",");
			
			for (int j = 0; j < arrTwoAcm.length; j++)
			{
				AccessControlModel twoAcm = arrTwoAcm[j];
				
				String twoRoles = twoAcm.getACMValue();
				twoRoles = twoRoles.replace("{", "");
				twoRoles = twoRoles.replace("}", "");			
				String [] twoRoleArr = twoRoles.split(",");
				
				for (int p = 0; p < twoRoleArr.length; p++)
				{
					for (int q = 0; q < oneRoleArr.length; q++)
					{
						//match is found
						if (twoRoleArr[p].equalsIgnoreCase(oneRoleArr[q]))
						{							
							//intersect the two role sets
							String intersectingRoles = twoRoleArr[p]; 
								
							//check if the roles in the intersection are already stored
							if (!hashRolesInIntersection.containsKey(intersectingRoles))
							{
								//store it in the hash table of intersecting roles
								hashRolesInIntersection.put(intersectingRoles, intersectingRoles);
							}
						}
					}//end for iteration oneRoles
				}//end for iteration twoRoles
			}
		}
		
		hashNewRolesInIntersection = hashRolesInIntersection;		
		return hashNewRolesInIntersection;
	}
		
	//------------------------------------------------------------------------------------

	private boolean SFA_processTuple(XATTuple currTuple, int streamId)
	{
		//make sure it is not a sec punct
		if (!currTuple.isSecPunct())
		{			
			//check if access is allowed 
			if (evaluateSSPlusOperator(streamId))
			{				
				//enqueue tuple (after all security and query predicates have been evaluated)
				this.getOutputQueue().enqueue(currTuple, this);		
				return true;
			}
			else
			{
				//no access to the tuple is allowed
				currTuple = null;
			}//end if access is not allowed
		}//end if check if it is a security punctuation
		
		return false;
	}
	
	//------------------------------------------------------------------------------------	

	//true = means propagate the tuple, false = means filter tuple (no access is allowed)
	boolean evaluateSSPlusOperator(int streamId)
	{
		boolean canAccessTuple = false; //denial-by-default
		
		//retrieve the intersection (between roles of the data and the query for the current stream)
		//and see of the intersection is non-empty
		
		//store the intersection for the stream
		Hashtable  hashRolesInDspQspIntersectionForStream = hashRolesInDspQspIntersectionForStreams.get(""+streamId);
		if (hashRolesInDspQspIntersectionForStream == null)
		{
			canAccessTuple = false;
		}
		else
		{
			if (hashRolesInDspQspIntersectionForStream.size() > 0)
			{
				canAccessTuple = true;
			}
			else
			{
				canAccessTuple = false;
			}
		}
		return canAccessTuple;
	}
	
	//------------------------------------------------------------------------------------
		
}
