/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on 2005-12-24
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.security;

import java.util.Hashtable;
import java.util.List;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;

/**
 * @author Rimma V. Nehme
 *
 */
public class XATSecPunctImp extends XATPunctImp implements XATSecPunct 
{	 
	
	//-----------------------------------------------------------------------------

	public enum SEC_PUNCT_ENFORCEMENT_TYPE 
	{
		IMMEDIATE_ENFORCEMENT,
        DEFERRED_ENFORCEMENT,
        UNDEFINED
	}
	
	//-----------------------------------------------------------------------------
	
	public enum SEC_PUNCT_TYPE 
	{
		DATA_SP,
        QUERY_SP,
        UNDEFINED     
	}
	
	//-----------------------------------------------------------------------------
	
	protected SEC_PUNCT_TYPE secPunctType = SEC_PUNCT_TYPE.UNDEFINED;
	protected AccessControlModel [] acmModels = null;
	protected String sign = "";
	protected boolean isImmutable = false;
	protected SEC_PUNCT_ENFORCEMENT_TYPE secEnforcementType = SEC_PUNCT_ENFORCEMENT_TYPE.UNDEFINED;
	protected int effectWindow = 0;
	protected String spTimestamp = "";
	
	//-----------------------------------------------------------------------------
	
	public XATSecPunctImp() 
	{
		super();		
		this.objTypeID = 2; //security punctuations have object type id = 2
	}
	
	//-----------------------------------------------------------------------------
	
	public XATSecPunctImp(XATDataValue[] array) 
	{
		super(array);
		this.objTypeID = 2; //security punctuations have object type id = 2
	}
	
	//-----------------------------------------------------------------------------
	
	public XATSecPunctImp(int length) 
	{
		super(length);
		this.objTypeID = 2; //security punctuations have object type id = 2
	}
	
	//-----------------------------------------------------------------------------
	
	public XATSecPunctImp(List list) 
	{
		super(list);
		this.objTypeID = 2; //security punctuations have object type id = 2
	}

	//-----------------------------------------------------------------------------

	public void setSecPunctTypeArray(int[] pTypeArray, 
									 AccessControlModel [] spAcmModels, 
									 String spSign, 
									 boolean spIsImmutable) 
	{
		punctTypeArray = pTypeArray;
		acmModels = spAcmModels;
		sign = spSign;
		isImmutable = spIsImmutable;
	}
	
	//-----------------------------------------------------------------------------

	public void setSecPunctTypeArray(int[] pTypeArray, 
								     AccessControlModel [] spAcmModels, 
								     String spSign,
								     String spType,
								     String spEnforcement,
								     String spTimestampVal) 
	{
		punctTypeArray = pTypeArray;
		acmModels = spAcmModels;
		sign = spSign;
		isImmutable = false; //by default all sps are mutable from now on
		
		if (spEnforcement.equals("D"))
		{
			secEnforcementType = SEC_PUNCT_ENFORCEMENT_TYPE.DEFERRED_ENFORCEMENT;
		}
		else if (spEnforcement.equals("I"))
		{
			secEnforcementType = SEC_PUNCT_ENFORCEMENT_TYPE.IMMEDIATE_ENFORCEMENT;
		}
		spTimestamp = spTimestampVal;
		
		
		if (spType.equalsIgnoreCase("dsp"))
		{			
			secPunctType = SEC_PUNCT_TYPE.DATA_SP;
		}
		else if (spType.equalsIgnoreCase("qsp"))
		{			
			secPunctType = SEC_PUNCT_TYPE.QUERY_SP;
		}
	}
	
	//-----------------------------------------------------------------------------

	public String toString() 
	{
		StringBuffer sb = new StringBuffer();
		sb.append("\nSP: ");
		
		for (int i = 0; i < this.getLength(); i++) 
		{
			Object cellValue = this.getCell(i).getValue();
			if (cellValue != null) 
			{
				sb.append(cellValue.toString());
			} 
			else 
			{
				sb.append(" ");
			}
			if (i < this.getLength() - 1)
				sb.append("  |  ");
		}
		
		if (acmModels != null)
		{
			for (int i = 0; i < this.acmModels.length; i++) 
			{
				sb.append("  |  source: "+ this.acmModels[i].getACMSource()+" " + this.acmModels[i].getACMTypeString() + ":" + this.acmModels[i].getACMValue());
			}
		}
		
		sb.append("| auth: " + this.sign);
		
		sb.append("  |  " + this.minTimeStamp + "  |  " + this.maxTimeStamp);
		return sb.toString();
	}
	
	//-----------------------------------------------------------------------------
	
	public String toShortString() 
	{
		StringBuffer sb = new StringBuffer();
		sb.append("\nSP: \n");
						
		if (acmModels != null)
		{
			
			sb.append("sign: " + this.sign + "\n");
			if(this.isImmutable == true)
			{
				sb.append("is immutable: " + "true" + "\n");}
			else
			{
				sb.append("is immutable: " + "false" + "\n");
			}	
			for (int i = 0; i < this.acmModels.length; i++) 
			{
				sb.append("\t" + "source: "+ this.acmModels[i].getACMSource()+"\n");
				sb.append("\t" + this.acmModels[0].getACMTypeString()+": ");
				sb.append(this.acmModels[i].getACMValue() + "\n");
			}
		}		
		return sb.toString();
	}
	
	//-----------------------------------------------------------------------------
	
	public AccessControlModel [] getAccessControlModels()
	{
		return acmModels;
	}
	
	//-----------------------------------------------------------------------------
	
	public String getSign()
	{
		return sign;
	}
	
	//-----------------------------------------------------------------------------
	
	public SEC_PUNCT_TYPE getSecPunctType()
	{
		return secPunctType;
	}
	
	//-----------------------------------------------------------------------------

	public SEC_PUNCT_ENFORCEMENT_TYPE getSecPunctEnforcementType()
	{
		return secEnforcementType;
	}
	
	//-----------------------------------------------------------------------------
	
	public int getSecPunctEffectWindow()
	{
		return effectWindow;
	}
	
	//-----------------------------------------------------------------------------
	
	//returns true, if sp's DDP is all as such <*|*|*|*|...|*>
	public boolean isDDPNonSpecific()
	{
		XATDataValue [] ddpElements = (XATDataValue[])this.getArrayData();
		
		for (int i = 0; i < ddpElements.length; i++)
		{
			XATDataValue currDataValue = (XATDataValue)ddpElements[i];
			if (!currDataValue.convertToString().equalsIgnoreCase("*"))
			{
				return false;
			}
		}
		return true; // all fields contain '*'
	}
	
	//-----------------------------------------------------------------------------
	
	public boolean isDDPContainAtLeastOneEmpty()
	{
		XATDataValue [] ddpElements = (XATDataValue[])this.getArrayData();
		
		for (int i = 0; i < ddpElements.length; i++)
		{
			XATDataValue currDataValue = (XATDataValue)ddpElements[i];
			if (currDataValue.convertToString().equalsIgnoreCase("-"))
			{
				return true; // at least one field disallows access NO MATTER WHAT
			}
		}
		return false; 
	}
	
	//-----------------------------------------------------------------------------
	
	public Hashtable extractSecurityPredicate()
	{
		Hashtable <String,String>hashSecPredicatesByAttributeIndex = new Hashtable(); //key --> attribute index, object --> filter condition 
																				  //(security predicate)
		
		XATDataValue [] ddpElements = (XATDataValue[])this.getArrayData();
		
		for (int i = 0; i < ddpElements.length; i++)
		{
			XATDataValue currDataValue = (XATDataValue)ddpElements[i];
			if (!currDataValue.convertToString().equalsIgnoreCase("*"))
			{
				hashSecPredicatesByAttributeIndex.put(""+i, currDataValue.convertToString());
			}
		}
		return hashSecPredicatesByAttributeIndex; 
	}
	
	//-----------------------------------------------------------------------------
	
}
