/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security.fence;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.S;

public class FenceInputSpecification implements java.io.Serializable
{
	//------------------------------------------------------------------------------------
	
	private String id;
	private String name;
	private int cardinality = 0;
	private Hashtable schemaAttributes;
	private Hashtable reverseSchemaAttributes;

	//------------------------------------------------------------------------------------
	
    public FenceInputSpecification(String idVal, 
    							   String nameVal,    								   										
    							   Hashtable schemaAttributesVal)
    {
    	id = idVal;
    	name = nameVal;
    	schemaAttributes = schemaAttributesVal;
    	reverseSchemaAttributes = doReverseSchemaAttributeMapping(schemaAttributes);
    }

    //------------------------------------------------------------------------------------
    
    private Hashtable doReverseSchemaAttributeMapping(Hashtable attributeIdxHash)
    {
    	Hashtable reverseAttributeIdxHash = new Hashtable();
    	
    	java.util.Enumeration en = attributeIdxHash.keys(); 
		while (en.hasMoreElements()) 
		{
			String currAttributeKey = (String)en.nextElement();
			String attrName = (String)attributeIdxHash.get(currAttributeKey);
			reverseAttributeIdxHash.put(attrName, currAttributeKey);
		}
    	
    	return reverseAttributeIdxHash;
    }
    
    //------------------------------------------------------------------------------------
    
    public String getAttrNameByIndex(int attr_idx)
    {
    	return (String)reverseSchemaAttributes.get(""+attr_idx);
    }
    
    //------------------------------------------------------------------------------------
    
    public String toString() 
    {
        StringBuilder result = new StringBuilder();
        String NEW_LINE = System.getProperty("line.separator");

        result.append("-----------operator input ----------------"+ NEW_LINE);
        //result.append(this.getClass().getName() + " Object {" + NEW_LINE);
        result.append("Input Id: " + id + NEW_LINE);
        result.append("Input Name: " + name + NEW_LINE);       
        result.append("Input Stream Schema: " + NEW_LINE);
        
        java.util.Enumeration en = reverseSchemaAttributes.keys(); 
        
		while (en.hasMoreElements()) 
		{
			String currAttributeKey = (String)en.nextElement();
			String attrName = (String)reverseSchemaAttributes.get(currAttributeKey);
			
			result.append(currAttributeKey +". "+ attrName + NEW_LINE);			
		}
    	        
        result.append("-----------end operator input-------------"+ NEW_LINE);
        return result.toString();     
    }

    //------------------------------------------------------------------------------------

    public String getId()
	{
		return id;		
	}
	
    //------------------------------------------------------------------------------------
    
	public Hashtable getSchemaMapping()
	{
		return schemaAttributes;
	}
	
	//------------------------------------------------------------------------------------
	
	public String getInputName()
	{
		return name;		
	}			
	
	//------------------------------------------------------------------------------------
	
	public int getCardinality()
	{
		return cardinality;		
	}
		
	//------------------------------------------------------------------------------------
		
}
