/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security.fence;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;

public class FenceOneWayJoinOperator implements java.io.Serializable
{
	//------------------------------------------------------------------------------------
	
	private String id;
	private String operatorType;
	private ArrayList inputs;
	private double operatorSelectivity;
	private Hashtable hashJoinAttributes; //hash tables where for every input id, it contains the index of the join attribute
	
	//------------------------------------------------------------------------------------
	
    public FenceOneWayJoinOperator(String idVal,
    								  String operatorTypeVal,     								   
    								  ArrayList inputsVal,
    								  Hashtable hashJoinAttributesVal)
    {
    	id = idVal;
    	operatorType = operatorTypeVal;
    	inputs = inputsVal;
    	hashJoinAttributes = hashJoinAttributesVal;
    }
      
    //------------------------------------------------------------------------------------
    
    public String toString() 
    {
        StringBuilder result = new StringBuilder();
        String NEW_LINE = System.getProperty("line.separator");

        //result.append(this.getClass().getName() + " Object {" + NEW_LINE);
        result.append("---------------- operator----------------" + NEW_LINE);
        result.append("Op Id: " + id + NEW_LINE);
        result.append("Op Type: " + operatorType + NEW_LINE);
        result.append("# inputs: " + inputs.size() + NEW_LINE );        
        result.append("-------------end operator----------------"+ NEW_LINE);
        return result.toString();   
    }
    
    //------------------------------------------------------------------------------------
    
	public String getId()
	{
		return id;		
	}
	
	//------------------------------------------------------------------------------------
	
	public double computeOveralStreamSelectivity()
	{
		return operatorSelectivity;		
	}
	
	//------------------------------------------------------------------------------------
	
	public double computeOpSelectivityBasedOnPartition()
	{
		double partitionBasedOperatorSelectivity = 0;
		return partitionBasedOperatorSelectivity;		
	}
	
	//------------------------------------------------------------------------------------
	
	public ArrayList getInputs()
	{
		return inputs;		
	}
	
	//------------------------------------------------------------------------------------
	
	public int getJoinAttributeIdxByInputId(String inputId)
	{
		if (hashJoinAttributes.containsKey(inputId))
		{
			Integer joinAttrIdx = Integer.parseInt((String)hashJoinAttributes.get(inputId));	
			return joinAttrIdx;
		}
		return -1; 
	}
	
	//------------------------------------------------------------------------------------
	
}
