/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security.fence;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.XATSecPunctImp;

public class FenceOperatorSpecification implements java.io.Serializable
{
	//------------------------------------------------------------------------------------
	
	private int WINDOW_SIZE; //window of tuples that will be kept in the buffer (count-based)
	private String id;
	private String operatorType;
	private ArrayList inputs;
	private double operatorSelectivity;
	private Hashtable <String,String> hashJoinAttributes; //hash tables where for every input id (stream id), it contains the index of the join attribute		
	private Hashtable <String,ArrayList> hashBuffers; //hash table where for every inputs id (stream id), it contains ArrayList of current tuples in the window
	
	private Hashtable <String,Hashtable> hashSecurityPredicatesByInputId; //hash table used by QRA to add security predicates to the actual query predicates
																	   //key--> stream id (input id), object --> security predicate (hash where key is attribute index and object is the string with DDP value)
	//------------------------------------------------------------------------------------
	
    public FenceOperatorSpecification(int WINDOW_SIZE,
    								  String idVal,
    								  String operatorTypeVal,     								   
    								  ArrayList inputsVal,
    								  Hashtable hashJoinAttributesVal)
    {
    	id = idVal;
    	operatorType = operatorTypeVal;
    	inputs = inputsVal;
    	hashJoinAttributes = hashJoinAttributesVal;
    	hashBuffers = new Hashtable();
    }
      
    //------------------------------------------------------------------------------------
    
    public String toString() 
    {
        StringBuilder result = new StringBuilder();
        String NEW_LINE = System.getProperty("line.separator");

        //result.append(this.getClass().getName() + " Object {" + NEW_LINE);
        result.append("---------------- operator----------------" + NEW_LINE);
        result.append("Op Id: " + id + NEW_LINE);
        result.append("Op Type: " + operatorType + NEW_LINE);
        result.append("# inputs: " + inputs.size() + NEW_LINE); 
        for (int i = 0; i < inputs.size(); i++)
        {
        	String inputId = (String)inputs.get(i);
        	result.append("\t input = [" + inputId +"] attribute id = [" + hashJoinAttributes.get(inputId) + "]" + NEW_LINE);
        }
        result.append("-------------end operator----------------"+ NEW_LINE);
        return result.toString();   
    }
    
    //------------------------------------------------------------------------------------
    
	public String getId()
	{
		return id;		
	}
	
	//------------------------------------------------------------------------------------
	
	public double computeOveralStreamSelectivity()
	{
		return operatorSelectivity;		
	}
	
	//------------------------------------------------------------------------------------
	
	public double computeOpSelectivityBasedOnPartition()
	{
		double partitionBasedOperatorSelectivity = 0;
		return partitionBasedOperatorSelectivity;		
	}
	
	//------------------------------------------------------------------------------------
	
	public ArrayList getInputs()
	{
		return inputs;		
	}
	
	//------------------------------------------------------------------------------------
	
	public String getType()
	{
		return operatorType;		
	}
	
	//------------------------------------------------------------------------------------
	
	public int getJoinAttributeIdxByInputId(String inputId)
	{
		if (hashJoinAttributes.containsKey(inputId))
		{
			Integer joinAttrIdx = Integer.parseInt((String)hashJoinAttributes.get(inputId));	
			return joinAttrIdx;
		}
		return -1; 
	}
	
	//------------------------------------------------------------------------------------
	
	public ArrayList <XATTuple> ProcessTuple(XATTuple currTuple, String currStreamId)
	{
		ArrayList <XATTuple> results = null;
		
		if (this.operatorType.equalsIgnoreCase("join"))
		{
			//determine the (CURRENT TUPLE) join attributes by probing hashJoinAttributes
			int joinAttrIdx = getJoinAttributeIdxByInputId(currStreamId);
			storeTupleInLocalState(currStreamId, currTuple);
			results = probeOppositeState(currStreamId, currTuple);			
		}
		return results;
	}
	
	//------------------------------------------------------------------------------------
	
	private ArrayList <XATTuple> probeOppositeState(String currStreamId, XATTuple currTuple)
	{
		ArrayList results = null;
		
		//1. retrieve opposite stream id (to probe against)
		String oppositeStreamId = getJoinOppositeInputId(currStreamId);
		//2. probe the buffer of that opposite stream
		
		//3. return results (if any)
		return results;
	}
	
	//------------------------------------------------------------------------------------
	
	private void storeTupleInLocalState(String inputId, XATTuple currTuple)
	{
		ArrayList<XATTuple> arrTuplesInBuffer = (ArrayList)hashBuffers.get(inputId);
		if (arrTuplesInBuffer.size() < WINDOW_SIZE)
		{
			//if the buffer size is smaller than the window size, then just append the tuple
			arrTuplesInBuffer.add(currTuple);
		}
		else
		{
			//else purge tuples from the current window (so that the size is equal to WINDOW_SIZE)
			while(arrTuplesInBuffer.size() != (WINDOW_SIZE - 1))
			{
				//purge the tuples from the head
				arrTuplesInBuffer.remove(0);
			}
		}
	}
	
	//------------------------------------------------------------------------------------
	
	private String getJoinOppositeInputId(String currStreamId)
	{
		String strOppositeStreamId = "-999";
		Enumeration e = hashBuffers.keys(); 
		while (e.hasMoreElements()) 
		{  
			String strInputId = (String)e.nextElement();  //key = stream id (participating in join) 
			if (currStreamId != strInputId)
			{
				strOppositeStreamId = strInputId;
				break;
			}
		}
		return strOppositeStreamId;
	}
	
	//------------------------------------------------------------------------------------
	
	private ArrayList <XATTuple> joinWithOppositeStateTuples(String currStreamId, XATTuple currTuple)
	{
		ArrayList results = null;
		
		//1. retrieve opposite stream id (to probe against)
		String oppositeStreamId = getJoinOppositeInputId(currStreamId);
		//2. probe the buffer of that opposite stream
		
		//3. return results (if any)
		return results;
	}
	
	//------------------------------------------------------------------------------------
	
	public void RemoveSecurityPredicate(String currStreamId)
	{
		hashSecurityPredicatesByInputId.remove(currStreamId);
	}
	
	//------------------------------------------------------------------------------------
	
	public void AttachSecurityPredicate(XATSecPunctImp currentSp, String currStreamId)
	{
		Hashtable <String,String>hashSecPredicatesByAttributeIndex = currentSp.extractSecurityPredicate();
		hashSecurityPredicatesByInputId.put(currStreamId, hashSecPredicatesByAttributeIndex);
	}
	

	
	//------------------------------------------------------------------------------------
	
}
