/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security.fence;

import java.util.ArrayList;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.MJoinDynamicAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes.JoinGraphImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.Stream;

public class FenceQueryPlanSpecification implements java.io.Serializable
{
	//------------------------------------------------------------------------------------
	
	private ArrayList operators;
	private JoinGraphImpl graph;

	//------------------------------------------------------------------------------------
	
	public FenceQueryPlanSpecification(ArrayList operatorsVal)
	{	
		operators = operatorsVal;	 
	}	    
			 
	//------------------------------------------------------------------------------------
	
	public ArrayList getOperators()	 
	{		
		return operators;			 
	}
	
	//------------------------------------------------------------------------------------
	
	private Stream SetupJoinGraph(String startPointName)
	{			
		graph = new JoinGraphImpl();
		Stream startPoint = null;
		
		Hashtable duplCheck = new Hashtable();
		for (int i=0; i<operators.size(); i++)
		{		
			ArrayList tmpStreams = new ArrayList();		
			FenceOptimizerOp op = (FenceOptimizerOp)operators.get(i);
			 
			for(int j = 0; j < op.getInputs().size();j++)
			{			
				FenceOptimizerOpInput input = (FenceOptimizerOpInput)op.getInputs().get(j);				 
				 				 
				Stream s = new Stream(input.getInputName());
				if(startPointName.equalsIgnoreCase(input.getInputName()))
				{				
					startPoint = s;				 
				}
				 
				s.setInputRate(input.getCardinality());
				if (!duplCheck.containsKey(input.getInputName()))				
				{					 
					graph.addVertex(s);				
					duplCheck.put(input.getInputName(), input);				 
				}				 
				tmpStreams.add(s);			 
			}
			try
			{
				 //add the edge
				 graph.addEdge((Stream)tmpStreams.get(0),
						 	   (Stream)tmpStreams.get(1),
						 	   op.getOpSelectivity());
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}		
		}			 
		return startPoint;	 
	}
	
	//------------------------------------------------------------------------------------
	 	
	public QueryPlan GetPlan(String startInputName)
	{	
		try		
		{			 
			Stream startVetix = SetupJoinGraph(startInputName);			 
			QueryPlan plan = MJoinDynamicAlgorithm.order(graph, startVetix);
			//plan.printXml();						 			 
			return plan;
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		
		return null;	 
	}
	
	//------------------------------------------------------------------------------------
}
