/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security.fence;

import edu.wpi.cs.dsrg.eddy.datastructure.EddyTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstate.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstreamoperator.BinaryPWSingleEquiJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.AttributeProperties;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.OperatorSignature;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.SimpleOperatorInputSpecification;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.SimpleOperatorSpecification;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.TrainingSetCollectorOperatorInputSpecification;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.QueryMeshOptimizer.QM_TYPE;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.AccessControlModel;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.XATSecPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.XATSecPunctImp.SEC_PUNCT_ENFORCEMENT_TYPE;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.XATSecPunctImp.SEC_PUNCT_TYPE;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * @author Rimma V. Nehme
 */

//------------------------------------------------------------------------------------
@SuppressWarnings("serial")
public class FenceQuerySpaceOperatorImp extends XATMultiQueueWindowStreamOperatorImp implements FenceQuerySpaceOperator 
{
	//------------------------------------------------------------------------------------
	
    public enum FENCE_QUERY_PROCESSING_METHOD 
    {
    	QUERY_REWRITE_APPROACH,
    	SECURITY_FILTER_APPROACH
    }
    
  //------------------------------------------------------------------------------------
    
    FENCE_QUERY_PROCESSING_METHOD fenceQPApproach = FENCE_QUERY_PROCESSING_METHOD.SECURITY_FILTER_APPROACH;  //change this var to control how queries are executed   
    String fenceInputXmlFile = "";
    
	//------------------------------------------------------------------------------------
	//--------------------
	// FENCE OP VARIABLES
	//--------------------
	int Wg = 5; //default window of tuples to keep for immediate semantics
    static List <FenceOperatorSpecification> xmlOperators;
	static Hashtable xmlInputs;
	
	RoundRobin roundRobinScheduler; //local round-robin scheduler
	
	static Hashtable<String, FenceOperatorSpecification> hashFenceOperators = new Hashtable<String, FenceOperatorSpecification>(); 
	static Hashtable<String, Hashtable> hashFencePlansForStreams = new Hashtable<String, Hashtable>(); 

	Document dom;	
	
	//------------------------------------------------------------------------------------
	Hashtable<String, ArrayList> hashBufferDspsForStreams = new Hashtable<String, ArrayList>(); 
	Hashtable<String, ArrayList> hashBufferQspsForStreams = new Hashtable<String, ArrayList>(); 
		
	Hashtable<String, Hashtable> hashRolesInDspQspIntersectionForStreams = new Hashtable<String,Hashtable>();
	
	//------------------------------------------------------------------------------------
	
	protected static int numOperatorExecuted = 0;
	
	private int totalTuples = 0; //tuples
	private int totalSps = 0;    //sps

	protected static int JOIN_OP_WINDOW_SIZE = 100; //window size = 100 tuples
		
	//------------------------------------------------------------------------------------
	
	public FenceQuerySpaceOperatorImp(XATNode init) 
	{
		super(init);		
	}
	
	//------------------------------------------------------------------------------------	

	public boolean run() throws Exception 
	{
		if (numOperatorExecuted == 0)
		{
			this.setupTheWorkspace(fenceInputXmlFile, JOIN_OP_WINDOW_SIZE);
		}
		numOperatorExecuted++;
		
		System.out.println("\n-------------------["+numOperatorExecuted+"] time operator is executed-------------------");
		
		//input queue from STREAM
		XATQueueSchema inSchema;
		int inLength;
		
		XATQueue inputQueue = (XATQueue)this.getInputQueue();
		
		if ((inSchema = inputQueue.getSchema()) == null)
		{
			System.out.println("input queue schema is NULL");
			return false;
		}
		
		inLength = inSchema.getLength();

		if (this.getOutputQueue().getSchema() == null) 
		{
			init();
		} 
		
		System.out.println();
		
		//------------------------------------------------------------------------
		// SECURITY FILTER QUERY PROCESSING APPROACH
		//------------------------------------------------------------------------		
		if (fenceQPApproach == FENCE_QUERY_PROCESSING_METHOD.SECURITY_FILTER_APPROACH)
		{					
			System.out.println("\nSA-CQP = [SECURITY_FILTER_APPROACH]");
			
			//-------------------------------------------------------------------------
			// FIRST PROCESS THE TUPLES FROM OPERATORS INPUT QUEUES
			//-------------------------------------------------------------------------
			
			for (int i = 0; i < this.inputQueue.size(); i++) 
			{	
				//------------------------------------------------------------------------					
				//int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
				int howMany  = 30; //rimma temp "hack"
				
				//for qsp stream, process a qsp at a time
				if (i == (this.inputQueue.size() - 1))
				{
					howMany = 1; //qsp stream
				}
				
				XATTuple[] tuples = (this.getInputQueue(i)).dequeue(howMany, this);
				System.out.println("Stream ["+i+"] dequeued [" + tuples.length + "] tuples");
				
				int numResultTuples = 0; //number of results tuples for this stream
				
				for (int p = 0; p < tuples.length; p++)
				{
					XATTuple currTuple = (XATTuple)tuples[p];
					//------------------------------
					// SECURITY PUNCTUATION
					//------------------------------
					if (currTuple.isSecPunct()) 				
					{
						String streamId = "" + i;
						String qspStreamId = "" + (this.inputQueue.size() - 1); //the query sp stream id (used for retrieving appropriate buffer)
																				//to intersect sps
						SFA_processSecPunctAtTheVeryBeginning(currTuple, streamId, qspStreamId);
					}
					//------------------------------
					// REGULAR TUPLE
					//------------------------------
					else
					{
						if (SFA_processTuple(currTuple, i))
						{
							numResultTuples++;
						}
					}
				}//end for (processing of all tuples for the current stream)
				
				System.out.println("Stream ["+i+"] enqueued [" + numResultTuples + "] tuples\n");
			}
			
			//garbage collection
			System.gc();
			
			/*
			//-------------------------------------
			// SCHEDULE OPS IN ROUND-ROBIN FASHION
			//-------------------------------------
			for (int k = 0; k < this.hashFenceOperators.size(); k++)
			{
				int nextOpRunId = roundRobinScheduler.next();			
				FenceOperatorSpecification currFenceOpToRun = (FenceOperatorSpecification)hashFenceOperators.get(""+nextOpRunId);
				
				runOperator(currFenceOpToRun);
			}
			*/
		}
		//------------------------------------------------------------------------
		// REWRTITE QUERY PROCESSING APPROACH
		//------------------------------------------------------------------------		
		else if (fenceQPApproach == FENCE_QUERY_PROCESSING_METHOD.QUERY_REWRITE_APPROACH)
		{
			System.out.println("SA-CQP = [QUERY_REWRITE_APPROACH]");
						
			//-------------------------------------------------------------------------
			// FIRST PROCESS THE TUPLES FROM OPERATORS INPUT QUEUES
			//-------------------------------------------------------------------------
			
			for (int i = 0; i < this.inputQueue.size(); i++) 
			{	
				//------------------------------------------------------------------------					
				//int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
				int howMany  = 30; //rimma temp "hack"
				
				//for qsp stream, process a qsp at a time
				if (i == (this.inputQueue.size() - 1))
				{
					howMany = 1; //qsp stream
				}
				
				XATTuple[] tuples = (this.getInputQueue(i)).dequeue(howMany, this);
				System.out.println("Stream ["+i+"] dequeued [" + tuples.length + "] tuples");
				
				int numResultTuples = 0; //number of results tuples for this stream
				
				for (int p = 0; p < tuples.length; p++)
				{
					XATTuple currTuple = (XATTuple)tuples[p];
					//------------------------------
					// SECURITY PUNCTUATION
					//------------------------------
					if (currTuple.isSecPunct()) 				
					{
						String streamId = "" + i;
						String qspStreamId = "" + (this.inputQueue.size() - 1);
						QRA_processSecPunctAtTheVeryBeginning(currTuple, streamId, qspStreamId);
					}
					//------------------------------
					// REGULAR TUPLE
					//------------------------------
					else
					{
						if (QRA_processTuple(currTuple, i))
						{
							numResultTuples++;
						}
					}
				}//end for (processing of all tuples for the current stream)
				
				System.out.println("Stream ["+i+"] enqueued [" + numResultTuples + "] tuples");
			}
			
			//garbage collection
			System.gc();
		}
				
		resetStatisticsVariables();
		
		System.out.println("Done running FENCE SPACE operator-----------------------------------------------------------------");
		return true;
	}
	
	//------------------------------------------------------------------------------------
	
	private void resetStatisticsVariables()
	{
		totalTuples = 0; //tuples + sps
		totalSps = 0; //tuples + sps
	}	

	//------------------------------------------------------------------------------------
	
	public String toToolTipString() 
	{
		return "FENCE QUERY SPACE";
	}
	
	//------------------------------------------------------------------------------------

	public boolean hasUnmatchedState() 
	{
		return false;
	}
	
	//------------------------------------------------------------------------------------
	
	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return true;
	}
	
	//------------------------------------------------------------------------------------
	
	public boolean cleanAllAccumulatedTuples() 
	{
		return true;
	}
	
	//------------------------------------------------------------------------------------
	
	public int setStateID(int startID) 
	{
		return 0;
	}
	
	//------------------------------------------------------------------------------------
	
	public boolean recomputeUnmatchedState() 
	{
		return true;
	}
	
	//------------------------------------------------------------------------------------
	
	public boolean matchStates(XATStreamOperator op) 
	{
		return true;
	}
	
	//------------------------------------------------------------------------------------
	
	public boolean finalizeOperation()
	{
		return true;
	}
	
	//------------------------------------------------------------------------------------
	
	public int getPurgeThreshold() 
	{
		// TODO Auto-generated method stub
		return 0;
	}		

	//------------------------------------------------------------------------------------
	
	public void init(int outputQueueId) 
	{
		XATQueueSchema outSchema = this.getOutputQueue(outputQueueId).getSchema();
		if (outSchema == null || outSchema.getLength() == 0) 
		{
			String schemaName = new String();
			outSchema = new XATQueueSchema();
			//Here we expect only one inputQueue existing
			Iterator it = this.getInputQueue().getSchemas().iterator();
			while (it.hasNext()) 
			{
				XATQueueSchema inSchema = (XATQueueSchema) it.next();
				schemaName += "#" + inSchema.getName();
				
				Iterator j = inSchema.getColNames().iterator();
				while(j.hasNext())
				{
					outSchema.addColumn((String) j.next());
				}				
			}
			outSchema.setName(schemaName);
			this.getOutputQueue(outputQueueId).setSchema(outSchema);
		}
		computeOutputPositions();

		this.numOutputColsPerStream = new int[outputPositions.length];
		this.numOutputColumns = 0;
		for (int i = 0; i < outputPositions.length; i++) 
		{
			int numCols = outputPositions[i].length;
			numOutputColsPerStream[i] = numCols;
			numOutputColumns += numCols;
		}
		this.setHasInit(true);
	}
	
	//------------------------------------------------------------------------------------
	
	public int getMemoryLimit() 
	{
		// TODO Auto-generated method stub
		return 0;
	}

	//------------------------------------------------------------------------------------
	
	public boolean getPropagateFlag() 
	{
		// TODO Auto-generated method stub
		return false;
	}

	//------------------------------------------------------------------------------------
	
	public void setMemoryLimit(int memLimit) 
	{
		// TODO Auto-generated method stub	
	}
	
	//------------------------------------------------------------------------------------

	public void setPropagateFlag(boolean flag) 
	{
		// TODO Auto-generated method stub		
	}
	
	//------------------------------------------------------------------------------------
		
	public void initializeFenceSpaceOperator(String xmlInputFile)
	{
		fenceInputXmlFile = xmlInputFile;
	}
	
	//------------------------------------------------------------------------------------
	
	private void setupTheWorkspace(String xmlInputFile, int WINDOW_SIZE)
	{
		xmlOperators = new ArrayList();
		xmlInputs = new Hashtable();
		
		//----------------
		// PROCESS XML
		//----------------		
		//parse the xml file and get the dom object
		System.out.println("Parsing fence space input file...");
		parseXmlFile(xmlInputFile);
		
		//parse the documents
		System.out.println("Extracting xml data from DOM...");
		parseDocument(WINDOW_SIZE);
		
		//Iterate through the list and print the data
		printData();		
		
		//initialize hash table of operators from just parsed xml operators and inputs specifications
		setHashFenceOperators();
		
		//create plans for the input streams
		createInitialPlansForEachStream();
	}	
	
	//------------------------------------------------------------------------------------	
	
	public void setImmediateWindowSemantics(String fenceImmediateWindowSemantics)
	{
		Wg = Integer.parseInt(fenceImmediateWindowSemantics);
	}
	
	//------------------------------------------------------------------------------------
	
	private void setHashFenceOperators()
	{
		List items = new ArrayList();	
		for (int i = 0; i < this.xmlOperators.size(); i++)
		{
			FenceOperatorSpecification currFenceOp = (FenceOperatorSpecification)xmlOperators.get(i);
			hashFenceOperators.put(currFenceOp.getId(), currFenceOp);
			
			//add to array list (array of operator ids - that will be locally scheduled by scheduler)
			items.add(new Robin(Integer.parseInt(currFenceOp.getId())));
		}
		
		//initialize the round robin scheduler
		roundRobinScheduler = new RoundRobin(items);	
	}
	//------------------------------------------------------------------------------------	
	
	private void parseXmlFile(String xmlInputFile)
	{
		//get the factory
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		
		try 
		{			
			//Using factory get an instance of document builder
			DocumentBuilder db = dbf.newDocumentBuilder();
			
			//parse using builder to get DOM representation of the XML file
			dom = db.parse(xmlInputFile);
		}
		catch(ParserConfigurationException pce) 
		{
			pce.printStackTrace();
		}
		catch(SAXException se) 
		{
			se.printStackTrace();
		}
		catch(IOException ioe) 
		{
			ioe.printStackTrace();
		}
	}
	
	//------------------------------------------------------------------------------------
	
	private void parseDocument(int WINDOW_SIZE)
	{
		//get the root elememt
		Element docEle = dom.getDocumentElement();
	
		//----------
		// INPUTS
		//----------		
		//get a nodelist of <Input> elements
		NodeList nl = docEle.getElementsByTagName("Input");
		if(nl != null && nl.getLength() > 0) 
		{
			for(int i = 0 ; i < nl.getLength();i++) 
			{				
				//get the employee element
				Element el = (Element)nl.item(i);
				
				//get the Input object
				FenceInputSpecification e = getXmlOperatorInput(el);
								
				//add input to list
				xmlInputs.put(e.getId(), e);
			}
		}
		
		//-------------
		// OPERATORS
		//-------------
		//get a nodelist of <Operator> elements
		nl = docEle.getElementsByTagName("Operator");
		if(nl != null && nl.getLength() > 0) 
		{
			for(int i = 0 ; i < nl.getLength();i++) 
			{				
				//get the employee element
				Element el = (Element)nl.item(i);
				
				//get the Operator object
				FenceOperatorSpecification e = getXmlOperator(el, WINDOW_SIZE);
				
				//add it to list
				xmlOperators.add(e);
			}
		}
	}

	//------------------------------------------------------------------------------------
	
	private FenceOperatorSpecification getXmlOperator(Element opEl, 
													  int WINDOW_SIZE) 
	{		
		String id = opEl.getAttribute("id");
		String type = opEl.getAttribute("type");
		
		ArrayList inputs = new ArrayList();
		Hashtable hashJoinAttributes = new Hashtable();

		//-------------------------------------------------------
		// retrieve operator inputs
		//-------------------------------------------------------
		//get a nodelist of <Input> elements
		NodeList nl = opEl.getElementsByTagName("OpInput");
		if(nl != null && nl.getLength() > 0) 
		{
			for(int i = 0 ; i < nl.getLength();i++) 
			{				
				//get the opInput element
				Element opInput = (Element)nl.item(i);
				
				//get id and attribute of the Input object
				String opInputId = opInput.getAttribute("id");
				String opAttributeIdx = opInput.getAttribute("attr_idx");
				
				//retrieve input object from the hashtable
				FenceInputSpecification opInputObj = (FenceInputSpecification)xmlInputs.get(opInputId);
				
				assert opInputObj != null;
				
				//add input id to the input list
				inputs.add(opInputId);

				//add attr index
				hashJoinAttributes.put(opInputId, opAttributeIdx);
			}
		}
		
		//Create a new SimpleOperatorSpecification with the value read from the xml nodes
		FenceOperatorSpecification e = new FenceOperatorSpecification(WINDOW_SIZE,
																	  id,
																	  type,
																	  inputs,
																	  hashJoinAttributes);
		
		return e;
	}
	
	//------------------------------------------------------------------------------------
	
	private String getTextValue(Element ele, String tagName) 
	{
		String textVal = null;
		NodeList nl = ele.getElementsByTagName(tagName);
		if(nl != null && nl.getLength() > 0) 
		{
			Element el = (Element)nl.item(0);
			textVal = el.getFirstChild().getNodeValue();
		}
		return textVal;
	}
	
	//------------------------------------------------------------------------------------

	private int getIntValue(Element ele, String tagName) 
	{
		//in production application you would catch the exception
		return Integer.parseInt(getTextValue(ele,tagName));
	}
	
	//------------------------------------------------------------------------------------
	
	private FenceInputSpecification getXmlOperatorInput(Element opEl) 
	{				
		String id = opEl.getAttribute("id");
		String name = opEl.getAttribute("name");
	
		//parse the schema of the input stream
		NodeList elNodeSchema = opEl.getElementsByTagName("Schema");
		
		assert elNodeSchema.getLength() == 1;		
		Element elSchema = (Element)elNodeSchema.item(0);

		Hashtable schemaAttributes = new Hashtable();
		//get a nodelist of <attribute> elements
		NodeList nl = elSchema.getElementsByTagName("attribute");
		if(nl != null && nl.getLength() > 0) 
		{
			for(int i = 0 ; i < nl.getLength();i++) 
			{				
				//get the employee element
				Element el = (Element)nl.item(i);
				
				String attr_name = el.getAttribute("name");
				
				schemaAttributes.put(attr_name, ""+i); //store attribute_name, attribute_index in the schema in the hashtable
			}
		}
		
		//Create a new Operator Input specification with the value read from the xml nodes
		FenceInputSpecification e = new FenceInputSpecification(id, 
																name, 
																schemaAttributes);
		
		return e;
	}
	
	//------------------------------------------------------------------------------------
		
	private void printData()
	{		
		System.out.println("No of xml operators [" + xmlOperators.size() + "]");
		System.out.println("\n---------------------------OPERATOR DETAILS-----------------------");
		
		Iterator it = xmlOperators.iterator();
		while(it.hasNext()) 
		{
			FenceOperatorSpecification currOpt = (FenceOperatorSpecification)it.next();
			System.out.println(currOpt.toString());
		}
		System.out.println("---------------------------END OPERATOR DETAILS-----------------------");
		
		System.out.println("No of xml inputs '" + xmlInputs.size() + "'.");
		System.out.println("\n----------------------------------INPUT DETAILS-----------------------");
		Enumeration e = xmlInputs.elements(); 
		while (e.hasMoreElements()) 
		{  
			Object x = e.nextElement();  // do whatever you need with x 
			System.out.println(x.toString());
		}
		System.out.println("----------------------------------END INPUT DETAILS-----------------------");
	}
		
	//------------------------------------------------------------------------------------
	
	private void createInitialPlansForEachStream()
	{
		//------------------------
		// NOW PREPARE OPERATORS
		//------------------------
		ArrayList allOperators = new ArrayList();
		Enumeration e = this.hashFenceOperators.keys();
	    	
	    //iterate through Hashtable keys Enumeration
	    while(e.hasMoreElements())
	    {
	    	String key = (String)e.nextElement();
	    	
	    	FenceOperatorSpecification currOp = (FenceOperatorSpecification)hashFenceOperators.get(key);
	    	
	    	//set up operator inputs for the optimizer
	    	ArrayList currOpInputs = currOp.getInputs();
	    	ArrayList currOpInputsForOptimizer  = new ArrayList();
	    	for (int k = 0; k < currOpInputs.size(); k++)
	    	{
	    		String opInputId = (String)currOpInputs.get(k);
	    		String nameVal = opInputId;
	    		double cardinalityVal = 0;	    		
	    		FenceOptimizerOpInput optimizerOpInput = new FenceOptimizerOpInput(nameVal, cardinalityVal);
	    		currOpInputsForOptimizer.add(optimizerOpInput);
	    	}	    	
	    	
	    	//initialize operator for optimizer
	    	FenceOptimizerOp currOpForOptimizer = new FenceOptimizerOp(currOp.getId(),
	    															   currOp.getType(),
	    															   currOpInputsForOptimizer,
	    															   0); //selectivity
	    	allOperators.add(currOpForOptimizer);
	    }
					
		FenceQueryPlanSpecification queryPlan = new FenceQueryPlanSpecification(allOperators);
		
		for (int i = 0; i < this.xmlInputs.size(); i++)
		{
			String streamId = "" + i;
			FenceInputSpecification input = (FenceInputSpecification)xmlInputs.get(streamId);
			
			assert(input != null);
			
			//get the plan
			QueryPlan currPlan = queryPlan.GetPlan(input.getId());
			Hashtable <String, String>hashPhysicalOpOrder = getOrderedListOfOperators(currPlan);			
			this.hashFencePlansForStreams.put(input.getId(), hashPhysicalOpOrder); //store
		}		
		System.out.println("Done creating all initial plans.");
	}
	
	//------------------------------------------------------------------------------------
	
	private Hashtable getOrderedListOfOperators(QueryPlan currPlan)
	{	
		Hashtable hashUnwindPlan = new Hashtable();
		UnwindPlan(hashUnwindPlan, currPlan);
		Hashtable hashOpOrder = new Hashtable();
		
		//--------------------------------------------------------------------
		// ITERATVE HASH AND CREATE XML PLAN
		//--------------------------------------------------------------------
		Set keys = hashUnwindPlan.keySet();
		Object [] strKeys = (Object [])keys.toArray();
		for (int q = 0; q < keys.size(); q++)
		{
			String key = (String)strKeys[q];
			FenceOperatorSignature fenceOpSignature = (FenceOperatorSignature)hashUnwindPlan.get(key);
			String order = "";
			if (key.length() == 4)
			{
				order = "1";
			}
			else if (key.length() == 7)
			{
				order = "2";
			}
			else if (key.length() == 10)
			{
				order = "3";
			}
			else if (key.length() == 13)
			{
				order = "4";
			}
			else if (key.length() == 16)
			{
				order = "5";
			}
			else if (key.length() == 19)
			{
				order = "6";
			}
			else if (key.length() == 22)
			{
				order = "7";
			}
			else if (key.length() == 25)
			{
				order = "8";
			}
			else if (key.length() == 28)
			{
				order = "9";
			}
			
			hashOpOrder.put(order, fenceOpSignature);
		}
		
		Hashtable hashPhysicalOpOrder = new Hashtable();
		Hashtable hashCachePhysicalOrderOfPlans = new Hashtable();
		
		for(int q = 1; q <= hashOpOrder.size(); q++)
		{
			FenceOperatorSignature opSignature = (FenceOperatorSignature)hashOpOrder.get(""+q+"");
			
			String order = ""+q;
			String opID = GetOpIdBasedOnInputs(opSignature.leftChild, opSignature.rightChild);
			hashPhysicalOpOrder.put(order, opID);
		}

		//hashCachePhysicalOrderOfPlans.put(currPlan.toString(), hashPhysicalOpOrder);		
		return hashPhysicalOpOrder; //return hash table with physical op order
	}

	//-----------------------------------------------------------------------------

	/*
	private static String translateFromBinaryOperatorsToUnaryOperators()
	{
		
	}
	*/
	
	//-----------------------------------------------------------------------------

	private static String GetOpIdBasedOnInputs(String leftChild, String rightChild)
	{
		//NOTE: because we assume left-linear query plans, the right child is
		//the id of the stream
		String opId = "";
		//--------------------------------------------------------------------
		// TRAVERSE OPERATORS AND MATCH AGAINST THEIR INPUTS (ASSUME LEFT-LINEAR TREES)
		//--------------------------------------------------------------------
		Iterator it = xmlOperators.iterator();
		while(it.hasNext()) 
		{
			FenceOperatorSpecification currOpt = (FenceOperatorSpecification)it.next();
			//System.out.println(currOpt.toString());
			String input1 = (String)currOpt.getInputs().get(0);
			String input2 = (String)currOpt.getInputs().get(1);
			
			if (rightChild.length() > 2)
			{
				String temp = leftChild;
				leftChild = rightChild;
				rightChild = temp;
			}
			
			if (input1.equals(rightChild))
			{
				if (leftChild.contains(input2))
				{
					//match found
					opId = currOpt.getId();
					return opId;
				}
			}
			else
				if (input2.equals(rightChild))
				{
					if (leftChild.contains(input1))
					{
						//match found
						opId = currOpt.getId();
						return opId;
					}
				}
		}
		return opId;
	}
	
	//------------------------------------------------------------------------------------	
	
	public static String UnwindPlan(Hashtable hashPlan, Node node)
	{		
		return node.getLabel();
	}
	
	//------------------------------------------------------------------------------------
	
	public static void UnwindPlan(Hashtable hashPlan, QueryPlan plan)
	{			
		if(plan.getRoot().getChild(0) != null)
		{
			if(plan.getRoot().getChild(0) instanceof Plan)
			{
				UnwindPlan(hashPlan, ((QueryPlan) plan.getRoot().getChild(0)));
			}
			else if(plan.getRoot().getChild(0) instanceof Node)
			{
				UnwindPlan(hashPlan,((Node) plan.getRoot().getChild(0)));
			}
		}
		if(plan.getRoot().getChild(1) != null)
		{
			if(plan.getRoot().getChild(1) instanceof Plan)
			{
				UnwindPlan(hashPlan, ((QueryPlan) plan.getRoot().getChild(1)));
			}
			else if(plan.getRoot().getChild(1) instanceof Node)
			{
				UnwindPlan(hashPlan, ((Node) plan.getRoot().getChild(1)));
			}
		}		
		
		String leftChildLabel = "";
		if (plan.getRoot().getChild(0) != null)
		{
			leftChildLabel = plan.getRoot().getChild(0).toString();
		}
		
		String rightChildLabel = "";
		if (plan.getRoot().getChild(1) != null)
		{
			rightChildLabel = plan.getRoot().getChild(1).toString();
		}
		
		FenceOperatorSignature fenceOpSignature = new FenceOperatorSignature(leftChildLabel, rightChildLabel);
		String key = "("+leftChildLabel + rightChildLabel+")";
		if (!key.equalsIgnoreCase("()"))
		{
			hashPlan.put(key,fenceOpSignature);
		}
		if (key.startsWith("&"))
		{
			key = rightChildLabel;
		}
		return;
	}
	
	//------------------------------------------------------------------------------------

	private void runOperator(FenceOperatorSpecification currFenceOpToRun)
	{
		
	}
	
	//------------------------------------------------------------------------------------
	
	private void SFA_processSecPunctAtTheVeryBeginning(XATTuple currTuple, 
												   String streamId, 
												   String qspStreamId)
	{
		//make sure it is a sec punct
		if (currTuple.isSecPunct())
		{
			XATSecPunctImp currentSp = (XATSecPunctImp)currTuple;					
				
			//-------------------------------
			// SP TYPE
			//-------------------------------
			//dsp
			if (currentSp.getSecPunctType() == SEC_PUNCT_TYPE.DATA_SP)
			{
				//remove (old) dsp from buffer (dsp)
				emptySpFromBufferByStream(streamId, currentSp.getSecPunctType());
				
				//add (new) dsp to buffer(dsp)
				addSpToBufferByStream(streamId, currentSp);
				
				//perform intersection with opposite buffer
				Hashtable hashRolesInDspQspIntersectionForStream = intersectDspWithQsp(currentSp, qspStreamId);
				
				//store the intersection for the stream
				hashRolesInDspQspIntersectionForStreams.put(streamId, hashRolesInDspQspIntersectionForStream);
			}
			//qsp
			else if (currentSp.getSecPunctType() == SEC_PUNCT_TYPE.QUERY_SP)
			{
				//remove (old) qsp from buffer (qsp)
				emptySpFromBufferByStream(streamId, currentSp.getSecPunctType());
				
				//add qsp to buffer(qsp)
				addSpToBufferByStream(streamId, currentSp);
				
				//perform intersection with opposite buffer
				Hashtable hashRolesInDspQspIntersectionForStream = intersectQspWithDsp(currentSp, streamId);
				
				//store the intersection with all data streams (dsps)
				hashRolesInDspQspIntersectionForStreams = hashRolesInDspQspIntersectionForStream;				
			}
			
			//-------------------------------
			// ENFORCEMENT
			//-------------------------------
			//immediate enforcement
			if (currentSp.getSecPunctEnforcementType() == SEC_PUNCT_ENFORCEMENT_TYPE.IMMEDIATE_ENFORCEMENT)
			{
				
			}
			//deferred enforcement
			else if (currentSp.getSecPunctEnforcementType() == SEC_PUNCT_ENFORCEMENT_TYPE.DEFERRED_ENFORCEMENT)
			{
				//do nothing
			}
						
		}
	}
	
	//------------------------------------------------------------------------------------

	private void emptySpFromBufferByStream(String streamId, SEC_PUNCT_TYPE spType)
	{
		ArrayList arrBuffer = null;
		
		if (spType == SEC_PUNCT_TYPE.DATA_SP)
		{
			arrBuffer = this.hashBufferDspsForStreams.get(streamId);
		}
		else if (spType == SEC_PUNCT_TYPE.QUERY_SP)
		{
			arrBuffer = this.hashBufferQspsForStreams.get(streamId);
		}
		
		if (arrBuffer != null)
		{		
			if (arrBuffer.size() > 0)
			{
				arrBuffer.clear();
			}
		}
	}

	//------------------------------------------------------------------------------------

	private void addSpToBufferByStream(String streamId, XATSecPunctImp sp)
	{
		ArrayList arrBuffer = null;
		
		if (sp.getSecPunctType() == SEC_PUNCT_TYPE.DATA_SP)
		{
			arrBuffer = this.hashBufferDspsForStreams.get(streamId);
			if (arrBuffer == null)
			{
				arrBuffer = new ArrayList();
				hashBufferDspsForStreams.put(streamId, arrBuffer);
			}
		}
		else if (sp.getSecPunctType() == SEC_PUNCT_TYPE.QUERY_SP)
		{
			arrBuffer = this.hashBufferQspsForStreams.get(streamId);
			if (arrBuffer == null)
			{
				arrBuffer = new ArrayList();
				hashBufferQspsForStreams.put(streamId, arrBuffer);
			}
		}
		
		arrBuffer.add(sp); //add sp
	}
	
	//------------------------------------------------------------------------------------

	private Hashtable intersectDspWithQsp(XATSecPunctImp dsp, String qspStreamId)
	{
		Hashtable hashRolesInIntersection = new Hashtable();
		
		AccessControlModel [] dspAcmModels = dsp.getAccessControlModels();
		
		ArrayList arrBufferQsps = this.hashBufferQspsForStreams.get(qspStreamId);
		
		if (arrBufferQsps != null)
		{
			//iterate through the opposite (qsp) buffer
			for (int i = 0; i < arrBufferQsps.size(); i++)
			{
				XATSecPunctImp qsp = (XATSecPunctImp)arrBufferQsps.get(i);
				AccessControlModel [] qspAcmModels = qsp.getAccessControlModels();
				
				//intersect the roles
				hashRolesInIntersection = intersectRolesInAcm(dspAcmModels,
															  qspAcmModels,
															  hashRolesInIntersection);
			}
		}
		return hashRolesInIntersection;
	}
	
	//------------------------------------------------------------------------------------

	private Hashtable intersectQspWithDsp(XATSecPunctImp qsp, String streamId)
	{
		Hashtable hashRolesInDspQspIntersectionForStream = new Hashtable();		
		
		AccessControlModel [] qspAcmModels = qsp.getAccessControlModels();
		
		int streamsCount = Integer.parseInt(streamId);
		//streamId passed here is the stream id of qsps (which is assumed 
		//to be the last stream in the streams config specification
		for (int k = 0; k < streamsCount; k++)
		{
			Hashtable hashRolesInIntersection = new Hashtable();
			String dataStreamId = "" + k;
			ArrayList arrBufferQsps = this.hashBufferDspsForStreams.get(dataStreamId);
			
			//iterate through the opposite (qsp) buffer
			for (int i = 0; i < arrBufferQsps.size(); i++)
			{
				XATSecPunctImp dsp = (XATSecPunctImp)arrBufferQsps.get(i);
				AccessControlModel [] dspAcmModels = dsp.getAccessControlModels();
				
				//intersect the roles
				hashRolesInIntersection = intersectRolesInAcm(dspAcmModels,
															  qspAcmModels,  
															  hashRolesInIntersection);
			}
			
			hashRolesInDspQspIntersectionForStream.put(dataStreamId, hashRolesInIntersection);
		}
		return hashRolesInDspQspIntersectionForStream;
	}
	
	//------------------------------------------------------------------------------------
	
	Hashtable intersectRolesInAcm(AccessControlModel[] arrOneAcm, 
								  AccessControlModel[] arrTwoAcm,
								  Hashtable hashRolesInIntersection)
	{
		Hashtable hashNewRolesInIntersection = new Hashtable();
		
		for (int i = 0; i < arrOneAcm.length; i++)
		{
			AccessControlModel oneAcm = arrOneAcm[i];
			String oneRoles = oneAcm.getACMValue();
			oneRoles = oneRoles.replace("{", "");
			oneRoles = oneRoles.replace("}", "");			
			String [] oneRoleArr = oneRoles.split(",");
			
			for (int j = 0; j < arrTwoAcm.length; j++)
			{
				AccessControlModel twoAcm = arrTwoAcm[j];
				
				String twoRoles = twoAcm.getACMValue();
				twoRoles = twoRoles.replace("{", "");
				twoRoles = twoRoles.replace("}", "");			
				String [] twoRoleArr = twoRoles.split(",");
				
				for (int p = 0; p < twoRoleArr.length; p++)
				{
					for (int q = 0; q < oneRoleArr.length; q++)
					{
						//match is found
						if (twoRoleArr[p].equalsIgnoreCase(oneRoleArr[q]))
						{							
							//intersect the two role sets
							String intersectingRoles = twoRoleArr[p]; 
								
							//check if the roles in the intersection are already stored
							if (!hashRolesInIntersection.containsKey(intersectingRoles))
							{
								//store it in the hash table of intersecting roles
								hashRolesInIntersection.put(intersectingRoles, intersectingRoles);
							}
						}
					}//end for iteration oneRoles
				}//end for iteration twoRoles
			}
		}
		
		hashNewRolesInIntersection = hashRolesInIntersection;
		
		return hashNewRolesInIntersection;
	}
		
	//------------------------------------------------------------------------------------

	private boolean SFA_processTuple(XATTuple currTuple, int streamId)
	{
		//make sure it is not a sec punct
		if (!currTuple.isSecPunct())
		{			
			//check if access is allowed 
			if (evaluateSSPlusOperator(streamId))
			{
				//retrieve the "query plan" for the stream with id "i"
				Hashtable hashPhysicalOpOrder = (Hashtable)this.hashFencePlansForStreams.get("" + streamId);
				
				for (int k = 1; k <= hashPhysicalOpOrder.size(); k++)
				{
					//get the next operator id to run
					String opIdToRunNextStr = (String)hashPhysicalOpOrder.get(""+k);

					//retrieve operator detailed specification to determine how to run it
					FenceOperatorSpecification currOpToRun = (FenceOperatorSpecification)hashFenceOperators.get(opIdToRunNextStr);
					
					System.out.println(currOpToRun.toString());
					String currStreamId = "" + streamId;
					ArrayList <XATTuple> resultTuples = currOpToRun.ProcessTuple(currTuple, currStreamId);
				}
				
				//enqueue tuple (after all security and query predicates have been evaluated)
				this.getOutputQueue().enqueue(currTuple, this);		
				return true;
			}
			else
			{
				//no access to the tuple is allowed
				currTuple = null;
			}//end if access is not allowed
		}//end if check if it is a security punctuation
		
		return false;
	}
	
	//------------------------------------------------------------------------------------
	//true = means propagate the tuple, false = means filter tuple (no access is allowed)
	boolean evaluateSSPlusOperator(int streamId)
	{
		boolean canAccessTuple = false; //denial-by-default
		
		//retrieve the intersection (between roles of the data and the query for the current stream)
		//and see of the intersection is non-empty
		
		//store the intersection for the stream
		Hashtable  hashRolesInDspQspIntersectionForStream = hashRolesInDspQspIntersectionForStreams.get(""+streamId);
		if (hashRolesInDspQspIntersectionForStream == null)
		{
			canAccessTuple = false;
		}
		else
		{
			if (hashRolesInDspQspIntersectionForStream.size() > 0)
			{
				canAccessTuple = true;
			}
			else
			{
				canAccessTuple = false;
			}
		}
		return canAccessTuple;
	}
	
	//------------------------------------------------------------------------------------
		
	private boolean QRA_processTuple(XATTuple currTuple, int streamId)
	{
		//make sure it is not a sec punct
		if (!currTuple.isSecPunct())
		{			
			//check if access is allowed 
			if (evaluateSSPlusOperator(streamId))
			{
				//retrieve the "query plan" for the stream with id "i"
				Hashtable hashPhysicalOpOrder = (Hashtable)this.hashFencePlansForStreams.get(streamId);
				
				for (int k = 0; k < hashPhysicalOpOrder.size(); k++)
				{
					//get the next operator id to run
					String opIdToRunNext = (String)hashPhysicalOpOrder.get(""+k);
					
					//retrieve operator detailed specification to determine how to run it
					FenceOperatorSpecification currOpToRun = (FenceOperatorSpecification)hashFenceOperators.get(opIdToRunNext);
					
					System.out.println("--- Running FENCE operator: " + currOpToRun + " ---");
					currOpToRun.ProcessTuple(currTuple, ""+streamId);
				}
				
				//enqueue tuple (after all security and query predicates have been evaluated)
				this.getOutputQueue().enqueue(currTuple, this);		
				return true;
			}
			else
			{
				//no access to the tuple is allowed
				currTuple = null;
			}//end if access is not allowed
		}//end if check if it is a security punctuation
		
		return false;
	}
	
	//------------------------------------------------------------------------------------
	
	private void QRA_processSecPunctAtTheVeryBeginning(XATTuple currTuple, 
												       String streamId, 
												       String qspStreamId)
	{
		//make sure it is a sec punct
		if (currTuple.isSecPunct())
		{
			XATSecPunctImp currentSp = (XATSecPunctImp)currTuple;					
				
			//-------------------------------
			// SP TYPE
			//-------------------------------
			//dsp
			if (currentSp.getSecPunctType() == SEC_PUNCT_TYPE.DATA_SP)
			{
				//remove (old) dsp from buffer (dsp)
				emptySpFromBufferByStream(streamId, currentSp.getSecPunctType());
				
				//add (new) dsp to buffer(dsp)
				addSpToBufferByStream(streamId, currentSp);
				
				//perform intersection with opposite buffer
				Hashtable hashRolesInDspQspIntersectionForStream = intersectDspWithQsp(currentSp, qspStreamId);
				
				//store the intersection for the stream
				hashRolesInDspQspIntersectionForStreams.put(streamId, hashRolesInDspQspIntersectionForStream);
			}
			//qsp
			else if (currentSp.getSecPunctType() == SEC_PUNCT_TYPE.QUERY_SP)
			{
				//remove (old) qsp from buffer (qsp)
				emptySpFromBufferByStream(streamId, currentSp.getSecPunctType());
				
				//add qsp to buffer(qsp)
				addSpToBufferByStream(streamId, currentSp);
				
				//perform intersection with opposite buffer
				Hashtable hashRolesInDspQspIntersectionForStream = intersectQspWithDsp(currentSp, streamId);
				
				//store the intersection with all data streams (dsps)
				hashRolesInDspQspIntersectionForStreams = hashRolesInDspQspIntersectionForStream;				
			}
			
			//-------------------------------
			// ENFORCEMENT
			//-------------------------------
			//immediate enforcement
			if (currentSp.getSecPunctEnforcementType() == SEC_PUNCT_ENFORCEMENT_TYPE.IMMEDIATE_ENFORCEMENT)
			{
				
			}
			//deferred enforcement
			else if (currentSp.getSecPunctEnforcementType() == SEC_PUNCT_ENFORCEMENT_TYPE.DEFERRED_ENFORCEMENT)
			{
				//do query rewriting based on the arrived sp
			}
						
		}
	}
	
	//------------------------------------------------------------------------------------
	
	private void QRA_RewriteQuery(XATTuple currentTuple, 
		       					  String currStreamId, 
		       					  String qspStreamId)
	{
		XATSecPunctImp currentSp = (XATSecPunctImp)currentTuple;
		
		// 1. check to see if data description part (DDP) is 'NON-SPECIFIC' in this sp or not
		// by 'NON-SPECIFIC', we mean it applies to all values (no matter what their value is)
		
		//-----------------------------------
		// DDP IS NON-SPECIFIC
		//-----------------------------------
		if (currentSp.isDDPNonSpecific())
		{
			//----------------------------------------------------------------------
			// (a) intersect with 'opposite buffer'
			Hashtable hashRolesInDspQspIntersectionForStream = new Hashtable();
			
			//qsp
			if (currentSp.getSecPunctType() == SEC_PUNCT_TYPE.QUERY_SP)
			{			
				//perform intersection with opposite buffer
				hashRolesInDspQspIntersectionForStream = intersectQspWithDsp(currentSp, currStreamId);								
			}
			else //dsp
			if (currentSp.getSecPunctType() == SEC_PUNCT_TYPE.DATA_SP)
			{
				//perform intersection with opposite buffer
				hashRolesInDspQspIntersectionForStream = intersectQspWithDsp(currentSp, currStreamId);					
			}
			
			// (b) if intersection is empty, add drop operator
			if (hashRolesInDspQspIntersectionForStream.size() == 0)
			{
				QRA_CreateDropOperator(currStreamId);
			}
			// (c) else do nothing (if the intersection is non-empty) that means access is allowed
			// to all data, and it is already being processed by the plan, so do nothing 
		}
		else
		{
			//-----------------------------------
			// DDP IS SPECIFIC - NEED TO REWRTITE THE PLAN
			//-----------------------------------
			
			// 0. match by stream
			Hashtable hashOperatorsInPlanByStream = hashFencePlansForStreams.get(currStreamId);
			boolean bSecurityPredicateAttached = false;
			
			Enumeration e = hashOperatorsInPlanByStream.keys(); 
			while (e.hasMoreElements()) 
			{  
				String opOrder = (String)e.nextElement();  
				String opId = (String)hashOperatorsInPlanByStream.get(opOrder);
				
				//retrieve the actual operator properties
				FenceOperatorSpecification actualOp = (FenceOperatorSpecification)hashFenceOperators.get(opId);

				//check op predicate for the match with current sp
				
				//1. check the stream id
				boolean matchesStream = false; //first check if current sp matches by stream (can only attach to operato
											   //if it operates on the same input stream as the sp that has arrived
				for (int k = 0; k < actualOp.getInputs().size(); k++)
				{
					String inputId = (String)actualOp.getInputs().get(k);
					if (currStreamId == inputId)
					{
						matchesStream = true; //match found by stream
						break;
					}
				}//iterating through operators and finding the best match to attach security predicate to
				
				if (matchesStream)
				{
					//attach the "security predicate"
					actualOp.RemoveSecurityPredicate(currStreamId);
					actualOp.AttachSecurityPredicate(currentSp, currStreamId);
					bSecurityPredicateAttached = true;					
				}
				
				if (!bSecurityPredicateAttached)
				{
					//if security predicate was NOT attached, create a new operator based on security filter condition
					QRA_CreateFilterOperator(currStreamId, "");
				}
				
				//System.out.println(x.toString());
			}//while iterating through operators to determine the best match to attach the "security predicate" to			
		}
 
	}
	
	//------------------------------------------------------------------------------------

	private void QRA_CreateDropOperator(String currStreamId)
	{
		
	}
	
	//------------------------------------------------------------------------------------
	
	private void QRA_CreateFilterOperator(String currStreamId, String filterCondition)
	{
		
	}
	
	//------------------------------------------------------------------------------------

}
