/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.security.fence;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.motion.helpers.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATSingleQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.AccessControlModel;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.StreamSecurityShieldOperatorACMSettings;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.XATSecPunctImp;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;
/**
 * @author Rimma V. Nehme
 *
 */
public class FenceSecurityShieldOperator implements java.io.Serializable	
{

	//------------------------------------------------------------------------------------
	
	protected boolean propagateFlagFurther = true;
	protected Hashtable accessControlModelsTable = new Hashtable();
	protected Hashtable apOperatorsTable = new Hashtable();
	private static XATSecPunctImp activeSP = null; 
	private static boolean matchFound = false;
	private int totalTuples = 0; //tuples + sps
	private int tuplesPassed = 0;
	private int tuplesDropped = 0;
	private int numSecPunct = 0;
	static int titleHelper = 1;
	private ExpressionStr securityExpression;
	
	//------------------------------------------------------------------------------------
	
	public FenceSecurityShieldOperator(XATNode init) 
	{
		tuplesPassed = 0;
		tuplesDropped = 0;
		numSecPunct = 0;
		totalTuples = 0;
	}

	//------------------------------------------------------------------------------------
	
	public boolean run(XATTuple[] tuples) throws Exception 
	{		
		
		/** figure out how many tuples to keep*/
		Vector tupleList = new Vector();
		for (int i=0; i<tuples.length; i++) 
		{
			// for distributed migration
			if(tuples[i].getMaxTimeStamp() == -1) 
			{
				continue;
			}
			
			if (tuples[i].isSecPunct()) 
			{
				//record the number of security punctuations (for statistics purposes)
				numSecPunct++;
				totalTuples++;
				
				//activate current security punctuation
				activeSP = (XATSecPunctImp)tuples[i];					
							
				//check if the current active SP 'matches' the SS operator state
				AccessControlModel [] spAcmModels = activeSP.getAccessControlModels();
				//reset the match flag
				matchFound = false;
				for(int m=0; m < spAcmModels.length; m++)
				{
					String strAcmModel = spAcmModels[m].getACMTypeString().toUpperCase();
					
					//unparse values for values corresponding to access control models
					String valueList = spAcmModels[m].value.substring(1, spAcmModels[m].value.length() - 1);
					valueList = valueList.trim();
									
					//check if the current acm is in the accessControlModelsTable of the SS operator
					Vector acmElements = (Vector)accessControlModelsTable.get(new String(strAcmModel));
					
					if (acmElements != null)
					{
						//System.out.println("NOT null value");
						for(int n = 0; n < acmElements.size();n++)
						{
							StreamSecurityShieldOperatorACMSettings ssSettings = (StreamSecurityShieldOperatorACMSettings)acmElements.get(n);

							if (valueList.equals(ssSettings.value))
							{
								matchFound = true;
								if(this.propagateFlagFurther == true)
								{
									//this.getOutputQueue(0).enqueue(tuples[i], this);
								}
								break;
							}
							else
							{
								//drop the security punctuation (all of the tuples corresponding to it will be dropped as well)
								//if even at the very last setting, the security punctuation still has NOT passed any of the settings
								//otherwise, we are still giving a "chance" for the current SP to match some access control model
								if (n == (acmElements.size()-1))
								{
									tuples[i] = null;
								}
							}
						}//end for
					}

					
					if(matchFound == true)
					{
						break; //at least one match found, so allow the tuples to pass through 
					}
				}//end for
				
				if (matchFound == false)
				{
					/*if(debug)
					{				        
				        //out.append("Access control settings don't match the SS state - DROP the upcoming tuples"+"\n");
						//out.append(activeSP.toShortString()+"\n");
						//out.append(this.toShortString());
					}*/
					
				}
				else
				{
					/*if(debug)
					{				        
						//out.append("Access control settings do match the SS state - PASS ON the upcoming tuples"+"\n");
						//out.append(activeSP.toShortString()+"\n");
						//out.append(this.toShortString());
					}*/
				}
			}//end if security punctuation 
			else 				
			{
				//handle regular tuples
				if (matchFound == true)
				{
					//this.getOutputQueue(0).enqueue(tuples[i], this);
					tuplesPassed++;	
					totalTuples++;
				}
				else
				{
					//free up space
					tuples[i] = null;
					tuplesDropped++;
					totalTuples++;
				}
			}
		}				
		resetStatisticsVariables();
		
		return true;
	}
	
	//------------------------------------------------------------------------------------
	
	private void resetStatisticsVariables()
	{
		numSecPunct = 0; 
		tuplesPassed = 0; 
		tuplesDropped = 0;
		totalTuples = 0;
	}		
	
	//------------------------------------------------------------------------------------

	public String toToolTipString() 
	{
		return "SS";
	}

	//------------------------------------------------------------------------------------
	
	public ExpressionStr getExpression() 
	{
		return securityExpression;
	}
	
	//------------------------------------------------------------------------------------
	
	public void setExpression(ExpressionStr expression) 
	{
		this.securityExpression = expression;
	}
	
	//------------------------------------------------------------------------------------
	
	public void setPropagateSPFurtherFlag(String value) 
	{
		if(value.toLowerCase().contains("true"))
		{
			this.propagateFlagFurther = true;
		}
		else if(value.toLowerCase().contains("false"))
		{ 
			this.propagateFlagFurther = false;
		}
		else
		{
			//by default - propagate the security punctuations further
			this.propagateFlagFurther = true;
		}
	}
	
	//------------------------------------------------------------------------------------
	
	//set associated processing operators
	public void setAssociatedProcessingOperators(Hashtable newApOperatorsTable)
	{
		apOperatorsTable = newApOperatorsTable;
	}
	
	//------------------------------------------------------------------------------------
	
	//set associated access control models
	public void setAssociatedAccessControlModels(Hashtable newAccessControlModels)
	{
		accessControlModelsTable = newAccessControlModels;
	}
	
	//------------------------------------------------------------------------------------
	
	public String toShortString() 
	{
		StringBuffer sb = new StringBuffer();
		sb.append("\nSS state: \n");
		sb.append("propagate flag further: "+this.propagateFlagFurther+"\n");
		
		for (Iterator it = accessControlModelsTable.entrySet().iterator(); it.hasNext(); )
		{
            Map.Entry e = (Map.Entry) it.next();
            sb.append( e.getKey() + "=> " + "\n");

            Vector acmVector = (Vector)e.getValue();
            for(int i=0; i < acmVector.size();i++)
            {            	            
	            StreamSecurityShieldOperatorACMSettings acm = (StreamSecurityShieldOperatorACMSettings)acmVector.get(i);
	            sb.append("\t"+i +":\n");
	            sb.append("\t\tname: " + acm.name + "\n");
	            sb.append("\t\ttype: " + acm.type + "\n");
	            sb.append("\t\tvalue: " + acm.value + "\n");
            }
            sb.append("\n");
        }
		return sb.toString();
	}
	
	//------------------------------------------------------------------------------------
}
