package edu.wpi.cs.dsrg.xmldb.xat.common.treestructure;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Vector;
import java.lang.StringBuffer;

import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTreeNodeHasNoParentException;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObjectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.raindrop.utils.DeepCopyMaker;
import edu.wpi.cs.dsrg.xmldb.xat.raindrop.utils.XATLeaf;

/** 
 * This class implemented the XATTree.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 */

public class XATTreeImp extends XATQueryObjectImp implements XATTree {
    /**
     * The array containing all the roots of XATTree.
     * @since 1.0
     */
    protected ArrayList root;
    
	// used only by optimizer and migrator.
	// right now multiple processeors share the same tree object reference
	// at any time only one process can optimize the part of the query tree that's running on its own machine.
    protected boolean isLocked = false; 
	
	// this flag set to be "true" is a signal to the GUI that the tree has been changed by the query optimizer.
	// the GUI needs to refresh the tree and reset this flag to false.
	protected boolean isOptimized = false;
//	protected int optimizeAttempt = 0;
	protected int optimizeCounter = 0;
//	protected int changeSchedulerAttempt = 0;
	protected int changeSchedulerCounter = 0;
	protected boolean isSchedulerChanged = false;
	protected String currentScheduler = new String();
	protected ArrayList availableSchedulerCategories = new ArrayList();

    // recording the activities of all operators in the tree
    StringBuffer operatorActivityBuffer = new StringBuffer();
    StringBuffer migrationActivityBuffer = new StringBuffer();

	// stats ONLY collected inside the root operator
	public int stats_total_output_tuples = 0;
	public int stats_output_tuple_count = 0;
	public StringBuffer stats_string=null;
	public StringBuffer stats_output_tuples=null;
	public int[] totalTuplesDequeued=null;
	public long stats_start_time=0;

	
	/**
	 * these methods are added for vldb demo
	 * parse adaptation status to GUI
	 */
	
	public synchronized void addSchedulerCategory(String newcategory) {
		availableSchedulerCategories.add(newcategory);
	}
	
	public synchronized ArrayList getAvailableSchedulerCategories() {
		return availableSchedulerCategories;
	}
	
	public synchronized void setOptimized(boolean ischanged) {
		this.isOptimized = ischanged;
	}
    
	public synchronized void setSchedulerChanged(boolean ischanged) {
		this.isSchedulerChanged = ischanged;
	}
	
	public synchronized boolean isOptimized() {
		return this.isOptimized;
	}
    
	public synchronized boolean isSchedulerChanged() {
		return this.isSchedulerChanged;
	}
	
	public synchronized String getCurrentScheduler() {
		return this.currentScheduler;
	}
	
	public synchronized void setCurrentScheduler(String scheduler) {
		this.currentScheduler = scheduler;
	}
	
	public void addOptimizedCounter() {
		optimizeCounter++;
	}
	public void addSchedulerChangedCounter() {
		changeSchedulerCounter++;
	}
	
	public int getOptimizedCounter() {
		return this.optimizeCounter;
	}
	
	public int getSchedulerChangedCounter() {
		return this.changeSchedulerCounter;
	}
	
	public String getMigrationActivity() {
		return this.migrationActivityBuffer.toString();
	}
    
	public void appendMigrationActivity(String newActivity) {
		//System.out.println("New activity added: " + newActivity);
		this.migrationActivityBuffer.append(newActivity);
	}
    
	public void clearMigrationActivity() {
		this.migrationActivityBuffer = new StringBuffer();
	}

	

    /**
     * The default XATTreeImp constructor.
     * @since 1.0
     */
    public XATTreeImp() {
        root = new ArrayList();
        
		//initialize output stats string
		this.stats_string = new StringBuffer();
		stats_string.append("time \t output_tuples \n");
		stats_start_time = System.currentTimeMillis();
		
		this.stats_output_tuples = new StringBuffer();
		stats_output_tuples.append("count \t output_tuple \n");
    }
    /**
     * Add a new parent node for a given DOMPatternNode node. 
     * The parent node is updated to have a link to the child node.
     * The child node is updates to have a link to the parent node.
     * @return void
     * @param parent The parent node it attached to.
     * @param child The child node attached.
     * @since 1.0
     */
    public void addNode(XATNode newNode, XATNode parentNode) {
        if (newNode == null || parentNode == null)
            return;
        newNode.setParent(parentNode);
        parentNode.addOpChild(newNode);
    }

    /**
     * For all the leaves, add the subtree to it.
     * Note: We clone the subtree, and then added it to the tree.
     * Hence, further modifications will not by automatically updated.
     * Make sure create the subtree first, then added to the tree.
     * @return void
     * @param childTree The XATTree attached.
     * @since 1.0
     */
    public void addXATTree(XATTree subtree) {

        if (subtree == null) // null subtree
            return;

        if (subtree.getRoot() == null) // empty subtree
            return;

        if (subtree.getLeaves().length == 0) // empty subtree.
            return;

        if (this.getRoot() == null) { //empty main tree
            // copy the subtree as the main tree.
            this.setRoot(((XATTree) subtree.clone()).getRoot());
            return;
        }

        // get all the leaves.
        XATNode[] leaves = this.getLeaves();

        if (leaves.length == 0) { // there is no node in the tree
            this.setRoot(((XATTree) subtree.clone()).getRoot());
        } else {
            // iteratre through the leaves.
            for (int i = 0; i < leaves.length; i++) {
                // clone the subtree.
                XATTree clone = (XATTree) subtree.clone();

                // append the subtree.
                leaves[i].addOpChild(clone.getRoot());

                // set the root's parent is the leaf.
                clone.getRoot().setParent(leaves[i]);
            }
        }
    }

    /**
     * Append all the decendent nodes of given XAT Node into a vector.
     * @return void
     * @param v    The vector contains all decendent nodes.
     * @param node The node searching for.
     * @since 1.0
     */
    protected void appendChildLeaves(Vector v, XATNode node) {
        if (node == null)
            return;

        XATNode[] children = node.getOpChildren();

        int i;
        int length = Array.getLength(children);
        if (length == 0) {
            v.add(node);
            return;
        }
        for (i = 0; i < length; i++) {
            appendChildLeaves(v, children[i]);
        }
    }

    /**
     * Append the decendent nodes within given depth of XATTree.
     * @return void
     * @param v    The vector contains all decendent nodes.
     * @param node The node searching for.
     * @param depth The depth of XATTree, within it the decendent nodes are computed.
     * @since 1.0
     */
    private void appendChildLeavesWithDepth(Vector v, XATNode node, int depth) {
        XATNode[] children = node.getOpChildren();

        int i;
        int length = Array.getLength(children);
        if (length == 0) {
            XATLeaf leaf = new XATLeaf(node, depth);
            v.addElement(leaf);
            return;
        }

        depth = depth + 1;
        for (i = 0; i < length; i++) {
            appendChildLeavesWithDepth(v, children[i], depth);
        }
    }
    /**
     * Append all the distinct decendent nodes of given XAT Node into a vector.
     * @return void
     * @param v    The vector contains all decendent nodes.
     * @param node The node searching for.
     * @since 1.0
     */
    protected void appendNodes(Vector v, XATNode node) {
        XATNode[] children = node.getOpChildren();
        /** only add if it doesnt already contain it */
        if (!v.contains(node)) {
            v.add(node);
            System.out.println("add node " + node);
        }
        for (int i = 0; i < children.length; i++) {
            appendNodes(v, children[i]);
        }
        return;
    }

    /**
     * Append a node to all the leaves of the current XATTree.
     * Remember further updates to this leaf will not be updated
     * automatically!
     * @return void
     * @param leaf The node searching for.
     * @since 1.0    
     */
    public void appendXATNode(XATNode leaf) {
        if (this.getRoot() == null) { // if it's a empty tree.
            this.setRoot((XATNode) leaf.clone());
            return;
        }

        // for all the leaves
        XATNode[] leaves = this.getLeaves();

        if (leaves.length == 0) { // if the tree is empty.
            this.setRoot((XATNode) leaf.clone());
        } else {
            for (int i = 0; i < leaves.length; i++) {
                XATNode clone = (XATNode) leaf.clone();
                leaves[i].addOpChild(clone);
                clone.setParent(leaves[i]);
            }
        }
    }

    /**
     * Convert all the children of the given XATNode into a string with indent format.
     * @return String The converted String.
     * @param node The XATNode searching for.
     * @param indent The indent format.
     * @since 1.0    
     */
    protected String childrenString(XATNode node, String indent) {
        if (node == null)
            return indent + "\n";
        String sOutput = indent + node.toString() + "\n";
        String newIndent = indent + "  ";
        XATNode[] children = node.getOpChildren();
        for (int i = 0; i < children.length; i++) {
            sOutput = sOutput + childrenString(children[i], newIndent);
        }
        return sOutput;
    }
    /**
     * Clone the XATTree.
     * @return Object The result 
     * @since 1.0
     */
    public Object clone() {

        // deep clone the XATTree. The original clone algorithm will
        // break the integrity between Node and its contained operators.

        try {
            return DeepCopyMaker.makeDeepCopy(this);
        } catch (Exception e) {
            // Deep Clone Failed.
            if (isDebug())
                e.printStackTrace(System.err);
            return new XATTreeImp();
        }
    }
    /**
     * Converts this tree into an array, similar to getLeaves()
     * @return XATNode[] An array of XATNodes 
     * @since 1.0
     */
    public edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode[] convertToArray() {
        //System.out.println("ConvertToArray from XATTree");
        Vector nodeList = new Vector();
        if (getRoot() != null)
            appendNodes(nodeList, getRoot());
        
//        XATNode[] nodeArray = new XATNode[nodeList.size()];
//        
//        for(int i=0; i<nodeList.size(); i++) 
//            nodeArray[i] = (XATNode)nodeList.get(i);
//        
//        return nodeArray;
        return (XATNode[]) nodeList.toArray(new XATNode[] {});
    }
	/**
	 * Converts this tree into an Vector, similar to getLeaves()
	 * @return Vector An Vector of XATNodes 
	 * @since 1.0
	 */
	public Vector convertToVector() {
		Vector nodeList = new Vector();
		if (getRoot() != null)
			appendNodes(nodeList, getRoot());
		return nodeList;
	}
    /**
     * Copy the specified tree as current tree.
     * @return void
     * @param tree The new XATTree.
     * @since 1.0
     */
    public void copy(XATTreeImp tree) {
        if (tree != null) {
            
            this.root = tree.root;
        }
    }
    /**
     * Count number of nodes in the subtree of a given root.
     * @return int The number of XAT nodes in this XATTree.
     * @param root The root of XATTree counted.
     * @since 1.0
     */
    int count(XATNode root) {
        // if there is no such node, sure, no nodes then.
        if (root == null)
            return 0;

        int result = 1; // one root.

        for (int i = 0; i < root.getOpChildren().length; i++) {
            result += count(root.getOpChildren()[i]);
        }

        return result;
    }

    /**
     * Delete node. It will affect number of nodes. All the children nodes of the
     * to-be-deleted node will be the children of the parent node of the to-be-delete node.
     * If the to-be-deleted node is the root node of the XATTree, the exception XATTreeNodeHasNoParentException
     * will the throwed. Instead, the root node should be changed by setRoot() method.
     * @return void
     * @param node the to-be-deleted child node.
     * @since 1.0     
     */
    public void deleteNode(XATNode node) throws XATTreeNodeHasNoParentException {
        // if this node is the root node, then if the children of the current node
        // is only one, then its child becomes the root node.
        XATNode parent = node.getParent();
        XATNode[] children = node.getOpChildren();
        if (parent == null && children.length != 1) {
            throw new XATTreeNodeHasNoParentException();
        }

        if (this.getRoot() == node) { // if this is the root node.
            if (parent != null) {
                // if there is a parent,
                // That's the case for FOR and Groupby operator, 
                // their inner XTreFunctions might link to their 
                // owner Node.
                // But, their owner node has no idea of this kind
                // of node. When we delete such node, we don't want
                // to add the children of this Node to its parent node,
                // because its parent node has no idea of this children.

                // remove this node.
                // let this node's parent knows about its FOR/GB father.
                children[0].setParent(parent);
                this.setRoot(children[0]);
            } else { // there is no parent of this root node.
                //make the only child as the root.
                children[0].setParent(null);
                this.setRoot(children[0]);
            }
        } else { // this is not the root node.
            if (parent != null) {
                int i;
                int length = Array.getLength(children);
                /** remember to keep the correct order after you delete this node.*/
                // get the index of the node.
                int index = parent.getIndexOfOpChild(node);
                // add all the children of this node to the current position
                // of the deleted node.
                for (i = length - 1; i >= 0; i--) {
                    children[i].setParent(parent);
                    parent.addOpChild(children[i], index);
                }
                // delete this node.
                parent.deleteOpChild(node);
                // disconnet this node from its parent.
                node.setParent(null);
            } else { // ERROR! It shouldn't be here.
	            if ( isDebug()){
                System.err.println(
                    "ERROR! This node {" + node + "} is not root, but its parent is null!");
	            }
            }

        }

    }
    /**
     * Get all the leaf node of given XATTree, which is rooted at given root node.
     * <pre>
     * This method is used to find all of the Source operators in an XATTree
     * Looks at all of the root's children and adds them to the vector
     * if that nodes' operator is an XML source .  The function is then
     * recurvsively called on each of root's children until there are no more nodes 
     * to look at.  
     * </pre>
     * @return Vector of all the leaf nodes including the subtree of FOR and GB.
     * @param root The root of a tree to iterate over
     * @see getLeaves();
     * @since 1.0     
     */
    public Vector getAllLeafNodes(XATNode root) {
        /**
         * XIN: Also need to look into subtrees of FOR operator.
         * XIN: Also, in terms of terminates, how about just terminate at the 
         * end of the tree when do the push down?
         */

        // vector containing all the leaf operators.
        Vector sourceSet = new Vector();
        Vector childrenVector = new Vector();
        XATNode[] children = root.getOpChildren();
        for (int i = 0; i < children.length; i++) {
            childrenVector.add(children[i]);
        }
        XATNode currentNode;
        int x;


        //loop recursively until there are no children of a node
        if (childrenVector.size() == 0) {
            //XIN: add the current node.
            sourceSet.addElement(root);
            return sourceSet;

        } else {
            for (x = 0; x < childrenVector.size(); x++) {
                currentNode = (XATNode) childrenVector.elementAt(x);
                sourceSet.addAll(getAllLeafNodes(currentNode));
            }
        }
        return sourceSet;
    }

    /**
     * Get all the leaves of this XATTree.
     * @return XATNode[] An array of XATNodes those are the leaf nodes.
     * @since 1.0
     */
    public XATNode[] getLeaves() {
        Vector nodeList = new Vector();
        appendChildLeaves(nodeList, getRoot());
        return (XATNode[]) nodeList.toArray(new XATNode[] {
        });
    }
    /**
     * Get all the leaves of this XATTree along with their depth.
     * @return XATLeaf[] An array of XATNodes those are the leaf nodes.
     * @since 1.0
     */
    public XATLeaf[] getLeavesWithDepth() {
        Vector nodeList = new Vector();
        int depth = 0;
        /* hideous, hideous kludge... java is too pedantic */
        appendChildLeavesWithDepth(nodeList, getRoot(), depth);
        return (XATLeaf[]) nodeList.toArray(new XATLeaf[] {
        });
    }
    /** 
     * Get the root node of this XATTree.
     * @return XATNode The root of XATTree.
     * @since 1.0
     */
    public XATNode getRoot() {
        return (root.size() > 0) ? (XATNode) root.get(0) : null;
    }
    /**
     * Assuming the XATTree is a forest sometime, this method get all the roots of XAT.
     * @return java.util.Iterator All the roots of given XAT.
     * @since 1.0     
     */
    public java.util.Iterator getRoots() {
        return root.iterator();
    }
    /**
     * Insert a new node as the root of this tree.
     * @return void
     * @param root The root will be inserted in.
     * @since 1.0     
     */
    public void insertRoot(XATNode newroot) {
        XATNode oldroot = getRoot();
        if (oldroot != null) {

            newroot.addOpChild(oldroot);

            // if the new root has no parent, but the original root has a parent,
            // then remember that parent in the new root.
            // that parent might point to the outer tree.
            if (newroot.getParent() == null && oldroot.getParent() != null) {
                newroot.setParent(oldroot.getParent());
            }

            oldroot.setParent(newroot);
        }
        // otherwise, directly set this new root as the root.
        this.setRoot(newroot);
    }
    /**
    * Get the value of the propertity "DEBUG_XATTreeImp".
    * @return boolean True: With debug information. False: No debug information.
    * @since 1.0
    */
    public boolean isDebug() {
        return System.getProperty("DEBUG_XATTreeImp") != null;
    }
    
    public synchronized boolean isLocked() {
    	return this.isLocked;
    }
    
    
    
    public void moveNode(XATNode child, XATNode newParent) {
        XATNode parent = child.getParent();
        if (parent != null) {
            parent.deleteOpChild(child);
        }
        child.setParent(newParent);
        if (newParent != null) {
            newParent.addOpChild(child);
        }
    }
    /**
     * Move node. It will affect three nodes, the original parent node A, the child
     * node B and the new parent node C. The A's children is updated,
     * The B's parent is updated, and C's children is updated.
     * <p>
     * The child node will always move to the end of the children of the parent node.
     * @return void
     * @param child The to-be-moved child node.
     * @param new_parent The new parent node.
     * @since 1.0     
     */
    public void moveNode(XATNode child, XATNode new_parent, int position) {
        XATNode parent = child.getParent();
        if (parent != null) {
            parent.deleteOpChild(child);
        }
        child.setParent(new_parent);
        if (new_parent != null) {
            new_parent.addOpChild(child, position);
        }
    }
    /**
     * Replace oldNode in Tree with newNode.
     * @return void
     * @param oldNode The old XATNode will be replaced.
     * @param newNode The new XATNode used to replace the old one.
     * @since 1.0          
     */
    public void replaceNode(XATNode oldNode, XATNode newNode) {

        // check if the oldNode or newNode is valid.
        if (oldNode == null || newNode == null) {
            // Error.
            if (isDebug()) {
                System.out.println(
                    "ERROR: XATTreeImp:replaceNode() oldNode: "
                        + oldNode
                        + " or newNode: "
                        + newNode
                        + " is NULL.");
            }
            return;
        }

        // check if the oldNode is source.
        if (oldNode.getParent() == this.getRoot()) {
            // it's the root node.
            this.setRoot(newNode);
        } else
            if (oldNode.getParent() == null) {
                // it's already a disconnected node.
                // send out error message.
                if (isDebug()) {
                    System.out.println(
                        "ERROR: XATTreeImp:replaceNode() Try to replace a unrelated node: " + oldNode);
                }
                return;
            } else {
                // it's not a root.
                // get its parent.
                XATNode parentNode = oldNode.getParent();

                // get the index of this oldNode in the parentNode.
                int index = parentNode.getIndexOfOpChild(oldNode);

                // set the newNode to take place of the oldNode.
                parentNode.setOpChild(index, newNode);

                // set parent of the newNode to the new parent.
                newNode.setParent(parentNode);

                // disconnect the oldNode.
                oldNode.setParent(null);

                // done.
            }
    }

    /**
     * Called by the constructor.  this will set the default properties
     * that the ALL query objects will support. Each subclass should override
     * this method and add their own unique properties, but should also call
     * super.setDefaultProperties() so that the common properties may be set.
     * @return void
     * @since 1.0
     */
    protected void setDefaultProperties() {
        addProperty("AVERAGE_TUPLES_IN_QUEUES_WEIGHT", ".3");
        addProperty("TREE_ID","-1");
		addProperty("LAST_LOCAL_MIGRATION_TIME", "0");
		addProperty("LAST_PARTITION_ADAPTATION_TIME", "0");
    }

    /**
     * Called by the constructor.  this will set the default properties
     * that the ALL query objects will support. Each subclass should override
     * this method and add their own unique properties, but should also call
     * super.setDefaultProperties() so that the common properties may be set.
     */
    protected void setDefaultStatistics() 
    {
        addStatistic("OUTPUT_RATE");
        addStatistic("TIMESLICE_OUTPUT_RATE");
        addStatistic("TOTAL_TUPLES_IN_QUEUES");
        addStatistic("THROUGHPUT");
        addStatistic("NUMBER_OF_TIMES_OPERATORS_WERE_RUN");
        addStatistic("OPERATOR_UTILIZATION");
        addStatistic("AVERAGE_TUPLE_DELAY");
        addStatistic("TOTAL_TUPLES_IN_STATES");
        addStatistic("NUMBER_OF_TUPLES_IN_STREAM_QUEUES");
        addStatistic("MEMORY_SATURATION");
		addStatistic("LAST_LOCAL_MIGRATION_TIME");
		addStatistic("LAST_PARTITION_ADAPTATION_TIME");
    }
    /**
     * Set root node.
     * @return XATNode The original root node.
     * @param new_root The new root node.
     * @since 1.0               
     */
    public XATNode setRoot(XATNode newRoot) {
        XATNode tmp = getRoot();
        root.clear();
        root.add(newRoot);
        return tmp;
    }
    /**
     * Give number of nodes in this tree.
     * @return int The computed size of the tree.
     * @since 1.0               
     */
    public int size() {
        if (root == null)
            return 0;
        int result = 0;
        //iterate through the arraylist and count for each tree.
        XATNode[] roots = (XATNode[]) root.toArray(new XATNode[] {
        });
        for (int i = 0; i < roots.length; i++)
            result += count(roots[i]);
        return result;
    }
    
    public String toString() {
        return childrenString(getRoot(), "");
    }
    /**
     * This going to validate the parent-child
     * relationship.
     * It's a top down traversal.
     * And, if it find a children't parent is not set properly
     * in the tree, a warning message will be shown, and also
     * the parent link will be fixed.
     * @return boolean Tree: This is a valid XATTree. False: otherwise.
     * @since 1.0          
     */
    public boolean validate() {
        boolean valid = validateNode(this.getRoot());

        if (isDebug()) {
            if (valid)
            	if ( isDebug()){
                System.err.println("This Tree is Valid");
            	}
            else{
	            if ( isDebug()){
                System.err.println("!!!!WARNING: This tree used to be Invalid, now fixed!!!!");
	            }
            }
        }

        return valid;

    }
    /**
     * This going to validate the parent-child
     * relationship.
     * It's a top down traversal.
     * And, if it find a children't parent is not set properly
     * in the tree, a warning message will be shown, and also
     * the parent link will be fixed.
     * Creation date: (3/8/2003 3:15:56 PM)
     * @return boolean
     */
    boolean validateNode(XATNode curNode) {

        // assume it's valid.
        boolean valid = true;

        // validate its operator's integrity.
        if (curNode.getOperator() != null) {
            if (curNode.getOperator().getNode() != curNode) {
                curNode.getOperator().setNode(curNode);
                if ( isDebug()){
                System.err.println(
                    "REPAIRED: Broken relationship between node ["
                        + curNode
                        + "] and its operator.");
                }
            }
        }

        // get all its children.
        XATNode[] children = curNode.getOpChildren();

        // iterate through the children.
        for (int i = 0; i < children.length; i++) {
            if (children[i].getParent() != curNode) {
                valid = false;
                children[i].setParent(curNode);
                if ( isDebug()){
                System.err.println(
                    "REPAIRED: Broken relationship between parent ["
                        + curNode
                        + "] child ["
                        + children[i]
                        + "].");}
            }
            // recursively visit its children.
            valid = validateNode(children[i]);
        }

        return valid;
    }
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject#getLightWeightQueryObject()
	 */
	public LightWeightQueryObject getLightWeightQueryObject() {
		return new LightWeightQueryObject("Tree " + this.getProperty("TREE_ID"),this.getClass().getName(),true);
	}
	
	//methods for demo purpose
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree#getOperatorActivity()
	 */
	public String getOperatorActivity() {
		return this.operatorActivityBuffer.toString();
	}
    
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree#appendOperatorActivity(java.lang.String)
	 */
	public void appendOperatorActivity(String newActivity) {
		//System.out.println("New activity added: " + newActivity);
		this.operatorActivityBuffer.append(newActivity);
	}
    
    /* (non-Javadoc)
     * @see edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree#clearOperatorActivity()
     */
    public void clearOperatorActivity() {
    	this.operatorActivityBuffer = new StringBuffer();
    }
    
    public synchronized void setLocked(boolean islocked) {
    	this.isLocked = islocked;
    }
    
    public void updateThroughputStats(int newOutputTuples) {
    	this.stats_total_output_tuples += newOutputTuples;
    	long time = System.currentTimeMillis() - stats_start_time; 
		stats_string.append(time + " \t " + stats_total_output_tuples + "\n");
		//System.out.println(time + " # of total output tuples = " + stats_total_output_tuples);
    }
    
//	public boolean finalizeOperation() {
//		this.printStatsToFile();
//		
//		//System.out.println("\n\nTotal numbers of tuples dequeued from each queue:");
//		//for(int i=0; i<2; i++) {
//		//	System.out.println("queue " + i + " = " + this.totalTuplesDequeued[i]);
//		//}
//		
//		return true;
//	}
	
	public void printStatsToFile() {
		try {
			String fileName = new String("Tree_Throughput" + System.currentTimeMillis()+".txt");
			FileWriter out = new FileWriter(fileName, true);
			out.write(this.stats_string.toString());
			out.close();
			
			//fileName = new String(runningMode + "_output_" + System.currentTimeMillis()+".txt");
			//out = new FileWriter(fileName, true);
			//out.write(this.stats_output_tuples.toString());
			//out.close();
		} catch (IOException e) {
			System.out.println(e.toString());
		}
	}

	public void appendParent(Vector v, Vector children) {
		// TODO Auto-generated method stub
		
	}

	public Vector getAllLeaves() {
		// TODO Auto-generated method stub
		return null;
	}

	public XATNode[] getBottomUpOrderNodes() {
		// TODO Auto-generated method stub
		return null;
	}
	
}