/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern;

/**
 * The visitor provide the additional functionality to an existing data structure. This data structure should
 * be open enough to make the visitor make the job down. <p>
 * Usually, the Visitable class should provide public methods to access its data.
 */
public interface Visitor
{
    /**
     * In the visit() function, you don't need to provide any kind of traversal strategy. That should be provided by the
     * accept() function in the Visitable classes. <p> It will only provide the additional functionality except
     * the traversing provided by the accept() function. <p> The visit function should NEVER call obj.accept()!!!! Because
     * that interfere with the traverse strategy in the obj.accept(). <p>
     * If there is an inheritance between the visited objects, it will only match the highest class.
     * @return Object 
     * @param object The object visited.
     * @since 1.0
     */
    Object visit(Object object);
}
