/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.xatnode;

import java.util.ArrayList;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import java.util.Iterator;
import java.util.Set;

import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATNodeDoesNotExistException;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.Visitable;
//import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;

/**
 * The class defines the node of XATTree.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public interface XATNode extends Visitable {
	
	/**
	* Add one XATNode child to this node. This should be called only by the XATTree.addNode().
	* Remember that children are orded sensitive.
	* Changed by M. Jbantova
	* @return int The number of XATNodes-children of current node. Streams not counted.
	* @param new_child The new XATNode child added.
	* @since 1.0
	*  //@deprecated
	*/

    int addOpChild(XATNode new_child);
    
    /**
    * Adds an XATNode-child to the children vector. XATNode position is interepreted as 
    * relative to the number of the rest of the XATNodes in the children vector. 
    * Remember: position is not an absolute number. Counting of XATNodes starts at zero.
    * For example, position 2 means second XATNode in the children vector, 
    * not the second element in the children vector, and position will be set to 1.
    * Changed by M. Jbantova to refelct order sensitive structure of children.
    * @param new_child The new XATNode child.
    * @return position The relative position of new added XATNode.
    * @since 1.0
    *  //@deprecated
    */
    int addOpChild(XATNode new_child, int position);
    /**
     * Clear the schema of this XATNode.
     * @return Set The old schema of this XATNode.
     * @since 1.0
     */
    Set clearSchema();
    /**
     * Deep clone this XATNode.
     * @return Object The cloned result.
     * @since 1.0
     */
    public Object clone();
    /**
    * Delete an XATTnode-child from this node.
    * Changed by M. Jbantova.
    * @return int The number of XATNode children. Number of Streams-children not included.
    * @param child The to-be-deleted child node.
    * @since 1.0
    *  //@deprecated
    */
    
    int deleteOpChild(XATNode child);


    /**
    * Get a specific XATNode child of this node by its relative position to the rest
    * of the XATNodes in the children vector. XATNode position is interepreted as 
    * relative to the number of the rest of the XATNodes in the children vector. 
    * Remember: position is not an absolute number. Counting of XATNodes starts at zero.
    * For example, position 1 means second XATNode in the children vector, 
    * not the second element in the children vector.
    * Changed by M. Jbantova to refelct order sensitive structure of children.
    * @return XATNode The child node at specified relative position.
    * @param position The relative position of the child node.
    * @throws XATNodeDoesNotExistException
    * @since 1.0
    */
    XATNode getOpChild(int position) throws XATNodeDoesNotExistException;
    
    /** 
    * Get all the XATNodes children of this node.
    * The returned array does not reflect the positions of the XATNOdes in the children vector.
    * Changed by M. Jbantova to reflect order sensitive structure of children.
    * @return XATNode[] The array of all the XATNodes children.
    * @since 1.0
    * */
    XATNode[] getOpChildren();


    /** Returns the number of XATNodes children of current node.
    *  Streams not counted.
    *  Changed by M. Jbantova 
    * @return int Number of XATNodes chilren of this node.
    */
    int getOpChildrenSize();
    
    
    /**
    * Returns the relative index of the specified XATNode of a child node.
    * The returned index of the XATNode should be interpreted as relative 
    * to the number of the rest of the XATNodes in the children vector. 
    * Remember: index is not an absolute number. Counting of XATNodes starts at zero.
    * For example, index of 2 means third XATNode in the children vector 
    * not the third element in the children vector since the children vector contains also the Streams
    * children of the current node.
    * Changed by M. Jbantova to refelct order sensitive structure of children.
    * @return int. The relative index of given node. -1 if node not found.
    * @param node The XATNode computing index for.
    * @since 1.0
    *  //@deprecated
    */
    int getIndexOfOpChild(XATNode node);
    /** 
     * Get the operator in this Node.
     * @return XATOperator The operator of this node.
     * @since 1.0
     */

    XATOperator getOperator();
   
    /** 
     * Get the parent node of this node.
     * @return XATNode The parent of this node.
     * @since 1.0
     */

    XATNode getParent();
    
    /**
     * Returns the number of parents for this node.
     * 
     * @return The number of parents
     */
    int getParentSize();
    /**
     * Get an Iterator of all of the parents.  
     * The XATNode only has one parent, thus this iterator will only contain one element. 
     * However, subclasses may have multiple parents, thus nothing is lost by placing this method here.
     * @return Iterator Including all the parents.
     * @since 1.0
     */
    java.util.Iterator getParents();
    
    /**
     * Get the arraylist of parent.
     * @return ArrayList of parent.
     */
	public ArrayList getParentArray();
	
	public void deleteParent(XATNode node);
	
    /**
     * Get the schema of this XATNode.
     * @return Set The schema of this XATNode.
     * @since 1.0
     */

    Set getSchema();
    
	/**
	 * Returns the stream operator for this node
	 */
	XATStreamOperator getStreamOperator();
    
    
	/**
	* Returns the relative index of the specified XATNode of child node.
	* The returned index of the XATNode should be interpreted as relative 
	* to the number of the rest of the XATNodes in the children vector. 
	* Remember: index is not an absolute number. Counting of XATNodes starts at zero.
	* For example, index of 2 means third XATNode in the children vector 
	* not the third element in the children vector since the children vector contains also the Streams
	* children of the current node.
	* If not found, return -1.
	* Changed by M. Jbantova to refelct order sensitive structure of children.
	* @return int. The relative index of this XATNode child.
	* @param node The relative reference of the child node searching for.
	* @since 1.0
	*  //@deprecated
	*/
  
    int locateOpChild(XATNode child);
    
    /**
    * Set a specific child to be a new child node.
    * Position is interpreted relative to the rest of the XATNodes in the children vector.
    * Counting of XATNodes starts at zero.
    * Changed by M. Jbantova to refelct order sensitive structure of children.
    * @return XATNode The original child node.
    * @param position The relative position of the child node.
    * @param new_child The new child node.
    * @since 1.0
    *  //@deprecated
    */
    XATNode setOpChild(int position, XATNode new_child)
        throws XATNodeDoesNotExistException;
   
   
    /**
    * Set all the XATNodes children  of this node to new values.
    * The relative positions of XATNodes in the children vector are preserved.
    * Changed by M. Jbantova to refelct order sensitive structure of children.
    * @return XATNode[] The array containing all the old XATNode children.
    * @param new_children The new children nodes.
    * @since 1.0
    *  //@deprecated
    */
    XATNode[] setOpChildren(XATNode[] new_children);
    
    /**
     * Set the operator in this Node.
     * @return XATOperator The old operator in this node.
     * @param new_operator The new operator in this node.
     * @since 1.0
     */
    XATOperator setOperator(XATOperator new_operator);
    /**
     * Set the parent node of this node.
     * @return XATNode The original parent of this node.
     * @param new_parent The new parent node of this node.
     * @since 1.0
     */
    XATNode setParent(XATNode new_parent);
    /**
     * Set the schema for current node.
     * @return Set The old schema of this node.
     * @param new_schema The new schema for this node.
     * @since 1.0
     */
    Set setSchema(Set new_schema);
    
    /**
    * Add one Stream child to this node. 
    * Remember the children are order sensitive.
    * Changed by M. Jbantova
    * @param streamName The new Stream child added.
    * @since 1.0
    *  //@deprecated
    */
	void addStreamChild(String streamName);
	
	/**
	* Returns how many Streams children this node has. XATNodes-children not included.
	* Changed by M. Jbantova.
	* @return int The number of Streams.
	* @since 1.0
	*  //@deprecated
	*/
	int getStreamChildSize();
	

	/**
	* Returns the name of a Stream child of this node by its relative position to the rest
	* of the Streams in the children vector. Stream position is interepreted as 
	* relative to the number of the rest of the Streams in the children vector. 
	* Remember: position is not an absolute number. Counting of Streams starts at zero.
	* For example, position 1 means SECOND Stream in the children vector, 
	* not the second element in the children vector.
	* Changed by M. Jbantova to reflect order sensitive structure of children.
	* @return String The Stream child at specified relative position.
	* @param position The relative position of the Stream child.
	* @throws XATNodeDoesNotExistException
	* @since 1.0
	*  //@deprecated
	*/
	String getStreamChildName(int id) throws XATNodeDoesNotExistException;
	
	/**
	* Set all the Streams children  of this node to new values.
	* The relative positions of Streams in the children vector are preserved.
	* Changed by M. Jbantova to reflect order sensitive structure of children.
	* @param streamNames An ArrayList of strings with the new Stream names.
	* @since 1.0
	*  //@deprecated
	*/
	public void setStreamChildNames(ArrayList streamnames);
	
	/** Returns an iterator of the streams the current node is connected to.
	* Changed by M. Jbantova to reflect order sensitive structure of children.
	* @return Iterator Streams of current node. 
	*  //@deprecated
	*/

	Iterator getStreamChildNames();
	
	/**
	* Deletes specific Stream child of this node. 
	* Remember the children are order sensitive.
	* Changed by M. Jbantova
	* @param streamName The Stream child to be deleted.
	* @since 1.0
	*  //@deprecated
	*/
	void deleteStreamChild(String streamName);
	/**
	 * Sets which tree the node belongs to.
	 * @param id The id for the tree
	 */
	void setTreeID(String id);
	/**
	 * Make a new copy of the current node. This is a deep copy, not bit by bit copy as the default clone in java.
	 * @return
	 */
	//public XATNode makeNewCopy();



	/**
	 * Gets the ID of the tree that this node belongs to.
	 * @return The ID of the tree that this node belongs to.
	 */
	String getTreeID();
	
	/**Adds a new child to the current node. A child is defined as either stream
	* or an operator (XATNode). Children are added in the order specified in the queryplan file.
	* @author jbantova
	* @param new_child
	* @return number of all children of current node
	*/
	
	public int appendChild(Object new_child);
	
	/**Inserts a new child for the current node at the specified position. A child is defined as either a stream
	* or an operator (XATNode). Children are added in the order specified in the queryplan file.
	* If a child is to be entered at a position beyond the size of the current children vector an OutOfBoundaryIndexException is thrown. 
	* @author jbantova
	* @param new_child
	* @param position
	* @return number of all children of current node
	*/
	public int insertChild(Object new_child,int position); //throws OutOfBoundaryIndexException;
	
	/** Deletes specified child.
	* @author jbantova
	* @param new_child
	* @return number of children
	*/
	public int deleteChild(Object new_child);
	
	/**Streams are stored as Strings and operator-children as XATNodes. instanceof operator can be used
	* to distinquish between both types of children.
	* If the specified position is beyond the size of the current children vector
	* an OutOfBoundaryIndexException is thrown. 
	* @author jbantova
	* @param position
	* @return child at specified position as an Object
	*/
	public Object getChild (int position); //throws OutOfBoundaryIndexException;
	
	
	/** Children are either streams or XATNodes (operators).
	 * @author jbantova
	 * @return an array of all the children of current node
	 */
	public Object[] getChildren ();
	
	/**
	 * @author jbantova
	 * @return number of all children of current node
	 **/
	public int getChildrenSize();
	
	
	/**
	* If child not found returns -1.
	* Child can be eather a Stream name or an XATNode.
	* @author jbantova
	* @param Object Child to be searched for.
	* @return int The position of specified child
	*/
	public int getIndexOfChild(Object child);
	
	
	/**
	* Replaces the child of the current node at the specified position with the new_child.
	* If the position specified is bigger than the current size of the children array,
	* an OutOfBoundaryIndexException is thrown.
	* @param position
	* @param new_child
	* @return Object Original Object (String or an XATNode)
	* @throws XATNodeDoesNotExistException
	*/
	public Object replaceChild(int position, Object new_child) throws OutOfBoundaryIndexException;
	
	/**
	* Set all the children nodes of this node to the new values.
	* The number of the new_children does not have to equal the number of the
	* old children.
	* @param new_children The new children.
	* @return Object[] The array containing all the old children.
	*/
	Object[] replaceChildren(Object[] new_children);
	
	public class OutOfBoundaryIndexException extends Exception {
		OutOfBoundaryIndexException() {}

	    OutOfBoundaryIndexException(String msg)
	    {
	        super(msg);
	    }

	}
	
		
//	public int getNodeID();
//	public void setNodeID(int nodeID); 
}
