package edu.wpi.cs.dsrg.xmldb.xat.common.xatnode;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//Rainbow 1.0 - XQuery Processing.
//
//Copyright 2003, All Rights Reserved.
//
//Permission to use, copy, modify and distribute this software and
//its documentation for educational, research and non-profit
//purposes, without fee, and without a written agreement is hereby
//granted, provided that this copyright notice appear in all copies.
//
//Permission to incorporate this software into commercial products
//may be obtained directly from the authors.
//
//This software is provided "as is" without express or implied warranty.
//In no event will the authors be held liable for any damages arising
//from the use of this software.
//
//Authors:
//Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//Computer Science Department
//Worcester Polytechnic Institute
//Worcester, MA 01609 USA
//
//http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//Send bug reports, bug fixes, enhancements, requests, or comments to
//the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

//import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;
//import java.lang.reflect.Constructor;

import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATNodeDoesNotExistException;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATTableTupleNoMatchException;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode.OutOfBoundaryIndexException;
//import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.raindrop.utils.DeepCopyMaker;

/**
 * The class defines the node of XATTree.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

/**
 * @author jbantova
 * 
 * TODO To change the template for this generated type comment go to Window -
 * Preferences - Java - Code Style - Code Templates
 */
public class XATNodeImp implements XATNode, Cloneable {

	/**
	 * The operator connected with this XATNode.
	 * 
	 * @since 1.0
	 */
	protected XATOperator operator;

	/**
	 * The parent nodes of this XATNode.
	 * 
	 * @since 1.0
	 */
	protected ArrayList parent;

	/**
	 * The child nodes of this XATNode. Will store XATNodes and Streams to
	 * refelct the order in which they were enetered in the queryplan.xml file
	 * from November 2004.
	 * 
	 * @since 1.0
	 */
	protected Vector children;

	/**
	 * The required schema of this XATNode.
	 * 
	 * @since 1.0
	 */
	protected Set schema = null;

	/**
	 * Array will no longer be used. Streams will be stored in the children
	 * vector.
	 * 
	 * @authormjba
	 * @since novemebr 2004
	 */

	protected String treeID;

	protected int nodeID;

	/**
	 * The default constructor of XATNodeImp.
	 * 
	 * @since 1.0
	 */
	public XATNodeImp() {
		operator = null;
		parent = null;
		children = new Vector();
		parent = new ArrayList();
		treeID = "";
	}

	//public int getNodeID() {
	//return nodeID;
	//}
	//
	//public void setNodeID(int newID) {
	//nodeID = newID;
	//}

	/**
	 * Add one XATNode child to this node. This should be called only by the
	 * XATTree.addNode(). Remember that children are orded sensitive. Changed by
	 * M. Jbantova
	 * 
	 * @return int The number of XATNodes-children of current node. Streams not
	 *         counted.
	 * @param new_child
	 *            The new XATNode child added.
	 * @since 1.0 //@deprecated
	 */

	public int addOpChild(XATNode new_child) {
		if ((new_child != null) && !children.contains(new_child))
			children.addElement(new_child);
		int i = 0;
		int count = 0;
		while (i < children.size()) {
			if (children.elementAt(i) instanceof XATNode) {
				count++;
			}
			i++;
		}
		//debugging
		//System.out.println("Number of XATNodes hildren is: " + count);
		return count;
	}

	/**
	 * Adds an XATNode-child to the children vector. XATNode position is
	 * interepreted as relative to the number of the rest of the XATNodes in the
	 * children vector. Remember: position is not an absolute number. Counting
	 * of XATNodes starts at zero. For example, position 2 means second XATNode
	 * in the children vector, not the second element in the children vector,
	 * and position will be set to 1. Changed by M. Jbantova to refelct order
	 * sensitive structure of children.
	 * 
	 * @param new_child
	 *            The new XATNode child.
	 * @return position The relative position of new added XATNode.
	 * @since 1.0 //@deprecated
	 */

	public int addOpChild(XATNode new_child, int position) {
		//int count=0;
		int count = -1;
		int i = 0;
		while (i < children.size() && count < position) {
			if (children.elementAt(i) instanceof XATNode) {
				count++;
				if (count == position)
					break;
			}
			i++;
		}
		if ((new_child != null) && !children.contains(new_child))
			children.insertElementAt(new_child, i);
		count = 0;
		i = 0;
		while (i < children.size()) {
			if (children.elementAt(i) instanceof XATNode) {
				count++;
			}
			i++;
		}
		return count;
	}

	/**
	 * Clear the schema of this XATNode.
	 * 
	 * @return Set The old schema of this XATNode.
	 * @since 1.0
	 */
	public Set clearSchema() {
		Set oldSchema = schema;
		schema = null;
		return oldSchema;
	}

	/**
	 * Deep clone this XATNode.
	 * 
	 * @return Object The cloned result.
	 * @since 1.0
	 */
	public Object clone() {
		XATNode result = new XATNodeImp();
		try {
			result.setOperator((XATOperator) DeepCopyMaker
					.makeDeepCopy(operator));
			result.setSchema((Set) DeepCopyMaker.makeDeepCopy(schema));
		} catch (Exception e) {
			System.out.println("Deep clone failed!");
			e.printStackTrace(System.err);
		}
		result.getOperator().setNode(result);
		return result;
	}

	/**
	 * Delete an XATTnode-child from this node. Changed by M. Jbantova.
	 * 
	 * @return int The number of XATNode children. Number of Streams-children
	 *         not included.
	 * @param child
	 *            The to-be-deleted child node.
	 * @since 1.0 //@deprecated
	 */
	public int deleteOpChild(XATNode child) {
		children.removeElement(child);
		int count = 0;
		for (int i = 0; i < children.size(); i++) {
			if (children.elementAt(i) instanceof XATNode) {
				count++;
			}
		}
		return count;
	}

	/**
	 * Get a specific XATNode child of this node by its relative position to the
	 * rest of the XATNodes in the children vector. XATNode position is
	 * interepreted as relative to the number of the rest of the XATNodes in the
	 * children vector. Remember: position is not an absolute number. Counting
	 * of XATNodes starts at zero. For example, position 1 means second XATNode
	 * in the children vector, not the second element in the children vector.
	 * Changed by M. Jbantova to refelct order sensitive structure of children.
	 * 
	 * @return XATNode The child node at specified relative position.
	 * @param position
	 *            The relative position of the child node.
	 * @throws XATNodeDoesNotExistException
	 * @since 1.0
	 */

	public XATNode getOpChild(int position) throws XATNodeDoesNotExistException {
		try {
			/*
			 * if(position>=children.size()){ throw new
			 * XATNodeDoesNotExistException(); }
			 */
			//int count=0;
			int count = -1;
			int index = 0;
			while (position < children.size() && count < position
					&& index < children.size()) {
				if (children.elementAt(index) instanceof XATNode) {
					count++;
					if (count == position)
						break;
				}
				index++;
			}

			if (count == position)
				return (XATNode) children.elementAt(index);
			else
				throw new XATNodeDoesNotExistException();
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new XATNodeDoesNotExistException();
		}
	}

	/**
	 * Get all the XATNodes children of this node. The returned array does not
	 * reflect the positions of the XATNOdes in the children vector. Changed by
	 * M. Jbantova to reflect order sensitive structure of children.
	 * 
	 * @return XATNode[] The array of all the XATNodes children.
	 * @since 1.0
	 */

	public XATNode[] getOpChildren() {
		/* hideous, hideous kludge... java is too pedantic */
		Vector xatnode = new Vector();
		for (int i = 0; i < children.size(); i++) {
			if (children.elementAt(i) instanceof XATNode) {
				xatnode.add(children.elementAt(i));
			}
		}
		return (XATNode[]) xatnode.toArray(new XATNode[] {});
	}

	/**
	 * Returns the number of XATNodes children of current node. Streams not
	 * counted. Chnaged by M. Jbantova
	 * 
	 * @return int Number of XATNodes chilren of this node.
	 */
	public int getOpChildrenSize() {
		int count = 0;
		for (int i = 0; i < children.size(); i++) {
			if (children.elementAt(i) instanceof XATNode) {
				count++;
			}
		}
		return count;
	}

	/**
	 * Returns the relative index of the specified XATNode of a child node. The
	 * returned index of the XATNode should be interpreted as relative to the
	 * number of the rest of the XATNodes in the children vector. Remember:
	 * index is not an absolute number. Counting of XATNodes starts at zero. For
	 * example, index of 2 means third XATNode in the children vector not the
	 * third element in the children vector since the children vector contains
	 * also the Streams children of the current node. Changed by M. Jbantova to
	 * refelct order sensitive structure of children.
	 * 
	 * @return int. The relative index of given node. -1 if node not found.
	 * @param node
	 *            The XATNode computing index for.
	 * @since 1.0 //@deprecated
	 */

	public int getIndexOfOpChild(XATNode node) {
		int count = 0;
		for (int i = 0; i < children.size(); i++) {
			if (children.elementAt(i).equals(node)) {
				return count;
			}
			if (children.elementAt(i) instanceof XATNode) {
				count++;
			}
		}
		return -1;
	}

	/**
	 * Get the operator in this Node.
	 * 
	 * @return XATOperator The operator of this node.
	 * @since 1.0
	 */
	public XATOperator getOperator() {
		return operator;
	}

	/**
	 * Get the parent node of this node.
	 * 
	 * @return XATNode The parent of this node.
	 * @since 1.0
	 */
	public XATNode getParent() {
		return (parent.size() > 0) ? (XATNode) parent.get(0) : null;
	}

	/**
	 * Get an Iterator of all of the parents. The XATNode only has one parent,
	 * thus this iterator will only contain one element. However, subclasses may
	 * have multiple parents, thus nothing is lost by placing this method here.
	 * 
	 * @return Iterator Including all the parents.
	 * @since 1.0
	 */
	public java.util.Iterator getParents() {
		return parent.iterator();
	}
	
	public ArrayList getParentArray() {
		return this.parent;
	}
	
	/**
	 * delete the parent at the position
	 * @param position
	 */
	public void deleteParent(XATNode node) {
		parent.remove(node);
	}

	/**
	 * Get the schema of this XATNode.
	 * 
	 * @return Set The schema of this XATNode.
	 * @since 1.0
	 */
	public Set getSchema() {
		return schema;
	}

	/**
	 * Returns the relative index of the specified XATNode of child node. The
	 * returned index of the XATNode should be interpreted as relative to the
	 * number of the rest of the XATNodes in the children vector. Remember:
	 * index is not an absolute number. Counting of XATNodes starts at zero. For
	 * example, index of 2 means third XATNode in the children vector not the
	 * third element in the children vector since the children vector contains
	 * also the Streams children of the current node. If not found, return -1.
	 * Changed by M. Jbantova to refelct order sensitive structure of children.
	 * 
	 * @return int. The relative index of this XATNode child.
	 * @param node
	 *            The relative reference of the child node searching for.
	 * @since 1.0 //@deprecated
	 */

	public int locateOpChild(XATNode child) {
		if (children.contains(child)) {
			int index = children.indexOf(child);
			int count = 0;
			for (int i = 0; i <= index; i++) {
				if (children.elementAt(i) instanceof XATNode) {
					count++;
				}
			}
			return count;
		} else
			return -1;
	}

	/**
	 * Set a specific child to be a new child node. Position is interpreted
	 * relative to the rest of the XATNodes in the children vector. Counting of
	 * XATNodes starts at zero. Changed by M. Jbantova to refelct order
	 * sensitive structure of children.
	 * 
	 * @return XATNode The original child node.
	 * @param position
	 *            The relative position of the child node.
	 * @param new_child
	 *            The new child node.
	 * @since 1.0 //@deprecated
	 */

	public XATNode setOpChild(int position, XATNode new_child)
			throws XATNodeDoesNotExistException {
		try {
			XATNode tmp = null;
			int count = -1;
			int index = 0;
			while (count < position && index < children.size()) {
				if (children.elementAt(index) instanceof XATNode) {
					count++;
					if (count == position)
						break;
				}

				index++;
			}
			if (count == position && index < children.size()) {
				tmp = (XATNode) children.elementAt(index);
				children.setElementAt(new_child, index);
			} else {
				children.addElement(new_child);
			}
			return tmp;
		} catch (NullPointerException e) {
			throw new XATNodeDoesNotExistException();
		}
	}

	/**
	 * Set all the XATNodes children of this node to new values. The relative
	 * positions of XATNodes in the children vector are preserved. Changed by M.
	 * Jbantova to refelct order sensitive structure of children.
	 * 
	 * @return XATNode[] The array containing all the old XATNode children.
	 * @param new_children
	 *            The new children nodes.
	 * @since 1.0 //@deprecated
	 */

	public XATNode[] setOpChildren(XATNode[] new_children) {

		Vector tmp = children;
		int i;
		Vector xatNode = new Vector();
		Vector streamsNames = new Vector();
		Vector xatPositions = new Vector();

		for (i = 0; i < children.size(); i++) {
			if (children.elementAt(i) instanceof XATNode) {
				xatNode.add(children.elementAt(i));
				Integer k = new Integer(i);
				xatPositions.addElement(k);
			}
			if (children.elementAt(i) instanceof String) {
				streamsNames.addElement(children.elementAt(i));
			}
		}

		if (new_children == null) {
			children.clear();

			//children vector will contain only the streams
			for (i = 0; i < streamsNames.size(); i++) {
				children.add(streamsNames.elementAt(i));
			}
			return null;
		}

		int length = new_children.length;

		for (i = 0; i < length; i++) {
			Integer k = (Integer) xatPositions.get(i);
			int index = k.intValue();
			children.setElementAt(new_children[i], index);
		}
		return (XATNode[]) xatNode.toArray(new XATNode[] {});
	}

	/**
	 * Set the operator in this Node.
	 * 
	 * @return XATOperator The old operator in this node.
	 * @param new_operator
	 *            The new operator in this node.
	 * @since 1.0
	 */
	public XATOperator setOperator(XATOperator new_operator) {
		XATOperator tmp = operator;
		operator = new_operator;
		return tmp;
	}

	/**
	 * Set the parent node of this node.
	 * 
	 * @return XATNode The original parent of this node.
	 * @param new_parent
	 *            The new parent node of this node.
	 * @since 1.0
	 */
	public XATNode setParent(XATNode new_parent) {

		if (new_parent == null) {
			parent.clear();
			return null;
		}

		XATNode tmp = getParent();
		if (parent.size() > 0) {
			parent.set(0, new_parent);
		} else {
			parent.add(new_parent);
		}
		return tmp;

	}

	/**
	 * Set the schema for current node.
	 * 
	 * @return Set The old schema of this node.
	 * @param new_schema
	 *            The new schema for this node.
	 * @since 1.0
	 */
	public Set setSchema(Set new_schema) {

		if (new_schema == null)
			return schema;

		Set tmp = schema;
		schema = new_schema;
		// also make the output xat table in this schema.
		XATTable output = new XATTableImp();
		Object[] names = schema.toArray();
		try {
			for (int i = 0; i < names.length; i++)
				output.appendColumn((String) names[i]);
		} catch (XATTableTupleNoMatchException nme) {
			nme.printStackTrace(System.err);
		}

		if (this.getOperator() != null)
			this.getOperator().setOutput(output);

		return tmp;
	}

	/**
	 * Set all the Streams children of this node to new values. The relative
	 * positions of Streams in the children vector are preserved. Changed by M.
	 * Jbantova to refelct order sensitive structure of children.
	 * 
	 * @param streamNames
	 *            An ArrayList of strings with the new Stream names.
	 * @since 1.0 //@deprecated
	 */

	public void setStreamChildNames(ArrayList streamNames) {

		Vector tmp = children;
		int i;
		/** stores temporarily the XATNodes children of the current node */
		Vector xatNode = new Vector();
		/** stores temp the streams connected to the current node */
		Vector streamsNames = new Vector();
		/** stores temp the stream positions in the children vector */
		Vector streamPositions = new Vector();

		for (i = 0; i < children.size(); i++) {
			if (children.elementAt(i) instanceof XATNode) {
				xatNode.add(children.elementAt(i));

			}
			if (children.elementAt(i) instanceof String) {
				streamsNames.addElement(children.elementAt(i));
				Integer k = new Integer(i);
				streamPositions.addElement(k);
			}
		}

		if (streamNames == null) {
			children.clear();

			//children vector will contain only the XATNodes
			for (i = 0; i < xatNode.size(); i++) {
				children.add(xatNode.elementAt(i));
			}

		}

		int length = streamNames.size();
		for (i = 0; i < length; i++) {
			Integer k = (Integer) streamPositions.get(i);
			int index = k.intValue();
			children.setElementAt(streamNames.get(i), index);
		}

	}

	/**
	 * Print current node information
	 * 
	 * @return String The String representation fot this node.
	 * @since 1.0
	 */
	public String toString() {
		try {
			return this.hashCode() + ":" + operator.toString();
		} catch (NullPointerException e) {

			return "NOP";
		}

	}

	/**
	 * Add one Stream child to this node. Remember the children are order
	 * sensitive. Changed by M. Jbantova
	 * 
	 * @param streamName
	 *            The new Stream child added.
	 * @since 1.0 //@deprecated
	 */

	public void addStreamChild(String streamName) {
		children.add(streamName);

	}

	/**
	 * Deletes specific Stream child of this node. Remember the children are
	 * order sensitive. Changed by M. Jbantova
	 * 
	 * @param streamName
	 *            The Stream child to be deleted.
	 * @since 1.0 //@deprecated
	 */

	public void deleteStreamChild(String streamName) {
		children.remove(streamName);

	}

	/**
	 * Returns the name of a Stream child of this node by its relative position
	 * to the rest of the Streams in the children vector. Stream position is
	 * interepreted as relative to the number of the rest of the Streams in the
	 * children vector. Remember: position is not an absolute number. Counting
	 * of Streams starts at zero. For example, position 1 means SECOND Stream in
	 * the children vector, not the second element in the children vector.
	 * Changed by M. Jbantova to refelct order sensitive structure of children.
	 * 
	 * @return String The Stream child at specified relative position.
	 * @param position
	 *            The relative position of the Stream child.
	 * @throws XATNodeDoesNotExistException
	 * @since 1.0 //@deprecated
	 */

	public String getStreamChildName(int id)
			throws XATNodeDoesNotExistException {
		try {
			/*
			 * if(position>=children.size()){ throw new
			 * XATNodeDoesNotExistException(); }
			 */
			int count = -1;
			int index = 0;
			while (id < children.size() && index < children.size()) {
				if (children.elementAt(index) instanceof String) {
					count++;
					if (count == id)
						break;
				}
				index++;
			}

			if (count == id) {
				//debugging
				//System.out.println("Found Stream is: " +
				// children.elementAt(index));
				return (String) children.elementAt(index);
			} else
				throw new XATNodeDoesNotExistException();
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new XATNodeDoesNotExistException();
		}
	}

	/**
	 * Returns how many Streams children this node has. XATNodes-children not
	 * included. Changed by M. Jbantova.
	 * 
	 * @return int The number of Streams.
	 * @since 1.0 //@deprecated
	 */

	public int getStreamChildSize() {
		int count = 0;
		for (int i = 0; i < children.size(); i++) {
			if (children.elementAt(i) instanceof String) {
				count++;
			}
		}
		return count;
	}

	/**
	 * Returns an iterator of the streams the current node is connected to.
	 * Changed by M. Jbantova to reflect order sensistive structure of children.
	 * 
	 * @return Iterator Streams of current node. //@deprecated
	 */

	public Iterator getStreamChildNames() {
		Vector streamsNames = new Vector();
		for (int i = 0; i < children.size(); i++) {
			if (children.elementAt(i) instanceof String)
				streamsNames.add(children.elementAt(i));
		}
		return streamsNames.iterator();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode#getParentSize()
	 */
	public int getParentSize() {
		return parent.size();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode#getTreeID()
	 */
	public String getTreeID() {

		return this.treeID;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode#setTreeID(java.lang.String)
	 */
	public void setTreeID(String id) {
		this.treeID = id;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode#getStreamOperator()
	 */
	public XATStreamOperator getStreamOperator() {
		return (XATStreamOperator) operator;
	}

	/**
	 * Adds a new child to the current node. A child is defined as either stream
	 * or an operator (XATNode). Children are added in the order specified in
	 * the queryplan file.
	 * 
	 * @author jbantova
	 * @param new_child
	 * @return number of all children of current node
	 */

	public int appendChild(Object new_child) {
		if ((new_child != null) && !children.contains(new_child))
			children.addElement(new_child);
		return children.size();
	}

	/**
	 * Inserts a new child for the current node at the specified position. A
	 * child is defined as either a stream or an operator (XATNode). Children
	 * are added in the order specified in the queryplan file. If a child is to
	 * be entered at a position beyond the size of the current children vector
	 * an OutOfBoundaryIndexException is thrown.
	 * 
	 * @author jbantova
	 * @param new_child
	 * @param position
	 * @return number of all children of current node
	 */

	public int insertChild(Object new_child, int position) {
		//			throws OutOfBoundaryIndexException {

		//		if (children.size() == 0 && position == 0) {
		//			children.add(position, new_child);
		//		}
		//		
		//		try {
		//			if (position > children.size()) {
		//				throw new OutOfBoundaryIndexException(
		//						"Specified index is beyond the size of current vector");
		//			}
		//		} catch (OutOfBoundaryIndexException e) {
		//			System.err.println();
		//			e.printStackTrace();
		//			throw new OutOfBoundaryIndexException();
		//		}

		//if (position <= children.size() && !children.contains(new_child)) {
		if (!children.contains(new_child)) {
			children.add(position, new_child);
		}
		return children.size();
	}

	/**
	 * Deletes specified child.
	 * 
	 * @author jbantova
	 * @param new_child
	 * @return number of children
	 */
	public int deleteChild(Object new_child) {
		children.remove(new_child);
		return children.size();
	}

	/**
	 * Streams are stored as Strings and operator-children as XATNodes.
	 * instanceof operator can be used to distinquish between both types of
	 * children. If the specified position is beyond the size of the current
	 * children vector an OutOfBoundaryIndexException is thrown.
	 * 
	 * @author jbantova
	 * @param position
	 * @return child at specified position as an Object
	 */
	public Object getChild(int position) {
		//throws OutOfBoundaryIndexException{
		//try{
		//    if(position<children.size())
		return (Object) children.elementAt(position);
		//     else
		//   throw new OutOfBoundaryIndexException("Specified position is beyond
		// the size of the current vector.");
		//   } catch (OutOfBoundaryIndexException e) {
		//   	e.printStackTrace();
		//    throw new OutOfBoundaryIndexException();
		//  }
	}

	/**
	 * Children are either streams or XATNodes (operators).
	 * 
	 * @author jbantova
	 * @return an array of all the children of current node
	 */
	public Object[] getChildren() {
		return (Object[]) children.toArray(new Object[] {});

	}

	/**
	 * @author jbantova
	 * @return number of all children of current node
	 */
	public int getChildrenSize() {
		return children.size();
	}

	/**
	 * If child not found returns -1. Child can be eather a Stream name or an
	 * XATNode.
	 * 
	 * @author jbantova
	 * @param Object
	 *            Child to be searched for.
	 * @return int The position of specified child
	 */

	public int getIndexOfChild(Object child) {
		if (children.contains(child)) {
			return children.indexOf(child);
		} else
			return -1;
	}

	/**
	 * Replaces the child of the current node at the specified position with the
	 * new_child. If the position specified is bigger than the current size of
	 * the children array, an OutOfBoundaryIndexException is thrown.
	 * 
	 * @param position
	 * @param new_child
	 * @return Object Original Object (String or an XATNode)
	 * @throws XATNodeDoesNotExistException
	 */

	public Object replaceChild(int position, Object new_child)
			throws OutOfBoundaryIndexException {
		Object tmp = null;
		if (children.size() >= position) {
			try {
				throw new OutOfBoundaryIndexException(
						"Out of boundary Exception");
			} catch (OutOfBoundaryIndexException e) {
				e.printStackTrace();
				throw new OutOfBoundaryIndexException();
			}
		}
		if (children.size() < position) {
			tmp = children.elementAt(position);
			children.setElementAt(new_child, position);
		}
		return tmp;
	}

	/**
	 * Set all the children nodes of this node to the new values. The number of
	 * the new_children does not have to equal the number of the old children.
	 * 
	 * @param new_children
	 *            The new children.
	 * @return Object[] The array containing all the old children.
	 */
	public Object[] replaceChildren(Object[] new_children) {
		Vector tmp = children;
		int i;
		if (new_children == null) {
			children.clear();
			return null;
		}
		int length = new_children.length;
		children = new Vector(length);
		for (i = 0; i < length; i++) {
			children.addElement(new_children[i]);
		}
		return (Object[]) children.toArray(new Object[] {});
	}

	/**
	 * Make a new copy of the current node. This is a deep copy, not bit by bit
	 * copy as the default clone in java. This is only called by the makeNewCopy
	 * make of the XATBox class.
	 * 
	 * @return
	 */

	//public XATNode makeNewCopy(){
	//XATNode newnode = new XATNodeImp();
	//
	//newnode.setStreamNames(this.streamNames);
	//newnode.setSchema(this.schema);
	//newnode.setTreeID(this.treeID);
	//Constructor construct;
	//Class[] c = new Class[2];
	//c[0] = this.operator.getClass();
	//c[1] = this.getClass();
	//try {
	//construct = operator.getClass().getConstructor(c);
	//newnode.setOperator((XATStreamOperator)construct.newInstance(new Object[] { this.operator, this }));
	//} catch (Exception e) {
	//e.printStackTrace();
	//}
	////newnode.setOperator(this.operator.makeNewCopy());
	//
	//
	//// we won't create new children and parent lists here since
	//// these would be done when making new copies of an XATBox.
	//
	//return newnode;
	//}
	//public XATNode makeNewCopy(){
	//XATNode newnode = new XATNodeImp();
	//
	//newnode.setStreamNames(this.streamNames);
	//newnode.setSchema(this.schema);
	//newnode.setTreeID(this.treeID);
	//Constructor construct;
	//Class[] c = new Class[2];
	//c[0] = this.operator.getClass();
	//c[1] = XATNode.class;
	//try {
	//construct = operator.getClass().getConstructor(c);
	//newnode.setOperator((XATStreamOperator)construct.newInstance(new Object[] {this.operator, this }));
	//} catch (Exception e) {
	//e.printStackTrace();
	//}
	////newnode.setOperator(this.operator.makeNewCopy());
	//
	//
	//// we won't create new children and parent lists here since
	//// these would be done when making new copies of an XATBox.
	//
	//return newnode;
	//}
}