/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.common.xmlstreamexpression;

import java.util.List;
import java.util.Vector;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import edu.wpi.cs.dsrg.xmldb.xat.common.expression.BinCOMPExpression;
import edu.wpi.cs.dsrg.xmldb.xat.common.expression.BinCOMPExpressionImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.expression.Expression;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.raindrop.utils.XMLNodeManipulator;


/**
 * Title:
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author 
 * @version 1.0.0.0 
 */

public class BinCOMPExpressionStrImp extends ExpressionStrImp implements BinCOMPExpressionStr {
  ExpressionStr eLeft,eRight;
  int iType;

/**
 * Insert the method's description here.
 * Creation date: (2/4/2003 12:56:05 PM)
 * @param expression edu.wpi.cs.dsrg.xmldb.xat.common.expression.BinCOMPExpression
 */
public BinCOMPExpressionStrImp(Expression expr) {
	
	ExpressionStrImp expressionStrImp = new ExpressionStrImp();
	
	eLeft = expressionStrImp.createExpressionStr(((BinCOMPExpression)expr).getLeft());

	eRight = expressionStrImp.createExpressionStr(((BinCOMPExpression)expr).getRight());

	iType = ((BinCOMPExpressionImp)expr).getType();

	untrueExpression = (BinCOMPExpressionStrImp)this.clone();    
	
	
	}
  public BinCOMPExpressionStrImp(ExpressionStr left, ExpressionStr right, int type)
  {
	eLeft = left;
	eRight = right;
	iType = type;
	untrueExpression = (BinCOMPExpressionStrImp)this.clone();    
   
  }
/**
 * Insert the method's description here.
 * Creation date: (1/23/2003 10:08:21 PM)
 * @return java.lang.Boolean
 * @param left java.lang.Object
 * @param right java.lang.Object
 */
public Boolean compAtomic(Object left, Object right) {
	try {
		String sLeft = null;
		String sRight = null;

		if ((left instanceof String) || (right instanceof String)) {
			if (left instanceof Element)
				sLeft = (new XATDataValueImp(left)).toString();
			else //if a dom node, actually its string value is compared
				if (left instanceof XATDataValue){
					if (((XATDataValueImp)left).isNode()){
					Node node = ((XATDataValueImp) left).convertToDOM();
					sLeft = XMLNodeManipulator.getStringValue(node);
					}else
					//must be an atomic value
					sLeft = ((XATDataValueImp)left).toString();
				}
				else
					sLeft = left.toString().trim();

			if (right instanceof Element)
				sRight = (new XATDataValueImp(left)).toString();
			else
				if (right instanceof XATDataValue){
					Node node = ((XATDataValueImp)right).convertToDOM();
					sRight = XMLNodeManipulator.getStringValue(node);
				}
				else
					sRight = right.toString().trim();

			if (iType == EQ) {
				return new Boolean(sLeft.equals(sRight));
			} else if (iType==NEQ)
				return new Boolean(!sLeft.equals(sRight));
		}

		long lLeft = ((Long) left).longValue();
		long lRight = ((Long) right).longValue();
		boolean bReturn;
		switch (iType) {
			case LT :
				bReturn = (lLeft < lRight);
				break;
			case GT :
				bReturn = (lLeft > lRight);
				break;
			case LEQ :
				bReturn = (lLeft <= lRight);
				break;
			case GEQ :
				bReturn = (lLeft >= lRight);
				break;
			case EQ :
				bReturn = (lLeft == lRight);
				break;
			case NEQ :
				bReturn = (lLeft != lRight);
				break;
			default :
				/* fail */
				return Boolean.FALSE;
		}
		return new Boolean(bReturn);
	} catch (ClassCastException e) {
		try {
			double dLeft = ((Number) left).doubleValue();
			double dRight = ((Number) right).doubleValue();
			boolean bReturn;
			switch (iType) {
				case LT :
					bReturn = (dLeft < dRight);
					break;
				case GT :
					bReturn = (dLeft > dRight);
					break;
				case LEQ :
					bReturn = (dLeft <= dRight);
					break;
				case GEQ :
					bReturn = (dLeft >= dRight);
					break;
				case EQ :
					bReturn = (dLeft == dRight);
					break;
				case NEQ :
					bReturn = (dLeft != dRight);
					break;
				default :
					/* fail */
					return Boolean.FALSE;
			}
			return new Boolean(bReturn);
		} catch (ClassCastException e2) {
			String sLeft = left.toString();
			String sRight = right.toString();
			boolean bReturn;
			switch (iType) {
				case LT :
					bReturn = (sLeft.compareTo(sRight) < 0);
					break;
				case GT :
					bReturn = (sLeft.compareTo(sRight) > 0);
					break;
				case LEQ :
					bReturn = (sLeft.compareTo(sRight) <= 0);
					break;
				case GEQ :
					bReturn = (sLeft.compareTo(sRight) >= 0);
					break;
				case EQ :
					bReturn = (sLeft.equals(sRight));
					break;
				case NEQ :
					bReturn = (!sLeft.equals(sRight));
					break;
				default :
					/* fail */
					return Boolean.FALSE;
			}

			return new Boolean(bReturn);
		}
	} catch (NullPointerException e) {
		return Boolean.FALSE;
	}
}
 /**
	* It will evaluate itself and return the value as an object.
	* Remember, it could be like: Integer, Float, Double, String, Boolean.
	*/
public java.lang.Object eval(edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator xopOper, int[] iPositions) {
	return null;
}
  // this code is written incorrectly... this need to check for column names and such
  public Object eval(List tuples, XATStreamOperator op)
  {
	Object lValue = eLeft.eval(tuples, op);
	Vector lCollection = new Vector();

	Object rValue = eRight.eval(tuples, op);
	Vector rCollection = new Vector();

	boolean left = false;

	if (lValue instanceof XATDataValue) {
		lCollection = ((XATDataValue) lValue).convertToCollection();
		left = true;
	} else
		if (lValue instanceof Vector) {
			lCollection = (Vector) lValue;
			left = true;
		} else
			if (rValue instanceof XATDataValue) {
				rCollection = ((XATDataValue) rValue).convertToCollection();
			} else
				if (rValue instanceof Vector) {
					rCollection = (Vector) rValue;
				} else { // it's a atomic value.
					Boolean result = compAtomic(lValue, rValue);
					if (result == Boolean.TRUE){
						setUntrueExpression(null);
					}
					else{
						setUntrueExpression(this);
					}
					return result;
				}

	if (left) {
		// then, let's handle the vectors.
		for (int i = 0; i < lCollection.size(); i++) {
			Boolean comp = compAtomic(lCollection.elementAt(i), rValue);
			if (comp.equals(Boolean.TRUE)){
				setUntrueExpression(null);
				return comp;
			}

		}
	} else {
		for (int i = 0; i < rCollection.size(); i++) {
			Boolean comp = compAtomic(lValue, rCollection.elementAt(i));
			if (comp.equals(Boolean.TRUE)){
				setUntrueExpression(null);
				return comp;
			}

		}
	}

	setUntrueExpression(this);
	return Boolean.FALSE;
  }
	public ExpressionStr getLeft()
	{
	  return eLeft;
	}
	public ExpressionStr getRight()
	{
	  return eRight;
	}
/**
 * Insert the method's description here.
 * Creation date: (2/17/2003 4:25:52 PM)
 * @return java.util.Vector
 */
public java.util.Vector getTerminalColumnExprs() {
	Vector results = new Vector();

	results.addAll(eLeft.getTerminalColumnExprs());
	results.addAll(eRight.getTerminalColumnExprs());
	
	return results;
}
/**
 * Insert the method's description here.
 * Creation date: (2/17/2003 4:25:52 PM)
 * @return java.util.Vector
 */
public java.util.Vector getTerminalColumnNames() {
	if (terminalColumnNames == null) {
		terminalColumnNames = new Vector();

		terminalColumnNames.addAll(eLeft.getTerminalColumnNames());
		terminalColumnNames.addAll(eRight.getTerminalColumnNames());
	}

	return terminalColumnNames;

}
  public int getType()
  {
	return iType;
  }
	public void setLeft(ExpressionStr left)
	{
	  eLeft = left;
	}
	public void setRight(ExpressionStr right)
	{
	  eRight = right;
	}
  public void setType(int type)
  {
	iType = type;
  }
/**
 * Insert the method's description here.
 * Creation date: (1/26/2003 8:51:26 PM)
 * @param newUntrueExpression edu.wpi.cs.dsrg.xmldb.xat.common.streamexpression.ExpressionStr
 */
public void setUntrueExpression(ExpressionStr newUntrueExpression) {
	untrueExpression = newUntrueExpression;
}

	public String toString()
	{
	switch (iType)
	{
	case LT:
		return "(" + eLeft.toString() + " < " + eRight.toString() + ")";
	case GT:
		return "(" + eLeft.toString() + " > " + eRight.toString() + ")";
	case LEQ:
		return "(" + eLeft.toString() + " <= " + eRight.toString() + ")";
	case GEQ:
		return "(" + eLeft.toString() + " >= " + eRight.toString() + ")";
	case EQ:  
		return "(" + eLeft.toString() + " == " + eRight.toString() + ")";
	case NEQ:
		return "(" + eLeft.toString() + " != " + eRight.toString() + ")";
	default:
		return "(" + eLeft.toString() + " COMP(" + Integer.toString(iType) + ") " + eRight.toString() + ")";
	}
	}
}
