/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.xmldb.xat.component.application;

import java.net.*;
import java.io.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import javax.swing.*;

public class RaindropApplication 
{
	private static String USAGE = "Usage: RaindropApplication port [-o sfc]\n s- screen \n f- file";
	
	/**
	 * This is the application that will receive the answer of a query
	 * from a root operator. It can display the query result on the screen
	 * and/or put the result to a file.
	 */
	public RaindropApplication() 
	{
		super();
	}

	public static void main(String[] args) 
	{
		int port = -1;
		boolean toFile = false;
		boolean toScreen = false;
		boolean count = true;

		try 
		{
			if (args.length > 4 || args.length == 0) 
			{
				System.out.println(USAGE);
				return;
			}

			port = Integer.parseInt(args[0]);

			if (port <= 0 || port > 20000) 
			{
				System.out.println
				(
					"Error: Please specifiy a port between 1 and 20000");
				return;
			}

			if (args.length > 1) 
			{
				if (args[1].equals("-o")) 
				{
					if (args[2].equals("s")) 
					{
						toFile = false;
						toScreen = true;
					}
					if (args[2].equals("f")) 
					{
						toFile = true;
						toScreen = false;
					}
					if (args[2].equals("sf") || args[2].equals("fs")) 
					{
						toFile = true;
						toScreen = true;
					}
				}
			}
			
			ServerSocket socket = new ServerSocket(port);
			System.out.println("Starting Application on port: " + port);
			if (toFile)
			{
				System.out.println("Will write query results to file.");
			}
			if (toScreen)
			{
				System.out.println("Will write query results to screen.");
			}

			int i = 1;
			while(true) 
			{
			//for (;;) {
				Socket inc = socket.accept();
                socket.setPerformancePreferences(1, 0, 2);
                socket.setReceiveBufferSize(524288);
                socket.setReuseAddress(true);
				System.out.println("Now start Receiving Query Results");
				Thread t = new QueryAnswerHandler(inc, toScreen, toFile, count);
				t.setName("QueryAnswerHandler" + i);
				t.setDaemon(true);
                if (args.length > 3) 
                {
                	System.out.println("Setting identifier with parameter [" + args[3] + "]");
                    ((QueryAnswerHandler)t).setIdentifier(args[3]);
                }
                else 
                {
                	System.out.println("Setting identifier with parameter [" + args[0] + "]");
                    ((QueryAnswerHandler)t).setIdentifier(args[0]);
                }
				t.start();
				i++;
			}
		} 
		catch (IOException ioe) 
		{
			System.out.println("Could not start the Query Receiver: " + ioe.getMessage());
		}
	}
}

/**
 * 
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 8, 2003
 * 
 * The PacketReceiverHandler class is responsible for
 * maintaining a connection between the local execution engine
 * and a remote engine. The PacketReceiver will spawn one of these threads
 * whenever a Packet from a new location is sent.
 * 
 */
class QueryAnswerHandler extends Thread 
{
	private Socket incoming;
	private String identifier;
	private String queryName;
	private boolean toScreen;
	private boolean toFile;
	private JFrame frame;
	private JTextArea text;
	private JScrollPane pane;
	private JPanel panel;
	private boolean count;
	private int tupleCount;

	public QueryAnswerHandler(Socket i, boolean toScreen, boolean toFile, boolean count) 
	{
		this.incoming = i;
		this.toScreen = toScreen;
		this.toFile = toFile;
		text = new JTextArea();
		text.setAutoscrolls(true);
		text.setEditable(false);
		text.setFocusable(true);
		this.count = count;
		tupleCount = 0;
		
	}
	
	public void setIdentifier(String id)
	{
		this.identifier = id;
	}

	public void run() 
	{
		// 	Listen to incoming packets and place them into the Packet Pool.

		File outFile = null;
		FileWriter outWriter = null;
		long lastTime = System.currentTimeMillis();
        long startTime = System.currentTimeMillis();
		
		try 
		{
		//	boolean enable = false;
			boolean enable = true;
			boolean haveSchema = false;
			XATTuple t;
			ObjectInputStream in = new ObjectInputStream(incoming.getInputStream());
				
			//If we are going to write to file, create the file
			if (toFile) 
			{
				outFile = new File(this.identifier+"_result.csv");
				outWriter = new FileWriter(outFile);
			}
			
			// If we are going to print to screen, create the screen.
			if (toScreen) 
			{
				//frame = new JFrame(queryName + " result.");
				frame = new JFrame("Query Result.");
				pane = new JScrollPane();
				panel = new JPanel();
				frame.setContentPane(panel);

				pane.setBounds(6, 4, 700, 300);
				pane.setViewportView(text);
				frame.setSize(750, 400);
				frame.setVisible(true);
				frame.setResizable(true);
				frame.setFocusable(true);
				text.append("Query Answer:\n");
				text.setBounds(0, 0, 700, 305);
				text.setFocusable(true);
				frame.getContentPane().add(pane);
				frame.requestFocus();

			}
			
			while (enable) 
			{			
				if(this.count)
				{
					if(System.currentTimeMillis() - lastTime > 15000)
					{
						lastTime = System.currentTimeMillis();
						System.out.println("Time: " + (System.currentTimeMillis() - startTime) + " received a total of: " +tupleCount+" tuples.");
					}
				}
				

				//Receive the new Packets	
                try 
                {
                
                	while(incoming.getInputStream().available() == 0)
                    {
                        try 
                        {
                        	Thread.sleep(100);
                        }
                        catch (InterruptedException ie) 
                        {
                        	
                        } 
                    }
					Object o = in.readObject();
					NetworkPacket np = (NetworkPacket) o;
					Packet p = null;

					for (int j = 0; j < np.size(); j++) 
					{
						p = np.getPacket(j);
	
						if(p.getControlMessage().equals("schema") && !haveSchema)
						{
							System.out.println("Received The Schema: " + p.getObject().toString());
							if (toScreen) 
							{
								text.append(p.getObject().toString() + "\n");
								text.requestFocus();
							}
							
							if (toFile) 
							{
								outWriter.write(p.getObject().toString() + "\n");
								outWriter.flush();
							}
							haveSchema = true;					
						}
						else if(p.getControlMessage().equals("schema"))
						{
							//do nothing
						}
						else if (p.getControlMessage().compareTo("newquery") == 0) 
						{
							identifier = p.getIdentifier();
							queryName = (String) p.getObject();
						}
						//Take the tuple and show the result
						else 
						{
							t = (XATTuple) p.getObject();
							tupleCount++;
							if (toScreen) 
							{
								text.append(t.toString() + "\n");
								text.requestFocus();
							}
							if (toFile) 
							{
								outWriter.write(t.toString() + "\n");
								outWriter.flush();
							}
						}

						if (p.getControlMessage().compareTo("end") == 0) 
						{
							enable = false;
						}
					}                	
                } 
                catch (EOFException e) 
                {
                    e.printStackTrace();
                    in = new ObjectInputStream(incoming.getInputStream());
                    continue;
                } 
			}		
		}
		catch(StreamCorruptedException sce)
		{
			sce.printStackTrace();
		}
		catch (SocketException se) 
		{
		    System.out.println("Time: " + (System.currentTimeMillis() - startTime) + "received a total of: " +tupleCount+" tuples.");
			System.out.println
			(
				"Connection Failed: Cannot get stream data from "
					+ incoming.getInetAddress()
					+ ":"
					+ incoming.getPort());
		} 
		catch (Exception e) 
		{
			e.printStackTrace();
		}
	}
}
