/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.xmldb.xat.component.application;

import java.net.*;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.io.*;

import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;

import javax.swing.*;


import org.jdesktop.jdic.browser.WebBrowser;
import org.jdesktop.jdic.browser.WebBrowserEvent;
import org.jdesktop.jdic.browser.WebBrowserListener;

public class VRMLApplication {

	private static String USAGE = "Usage: VRMLApplication port [-o sfc]\n s- screen \n f- file";
	
	/**
	 * This is the application that will receive the answer of a query
	 * from a root operator. It can display the query result on the screen
	 * and/or put the result to a file.
	 */
	public VRMLApplication() {
		super();
	}

	public static void main(String[] args) {

		   JFrame frame = new JFrame("FireStream ICDE 2007 Demo - Istanbul");
	        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	        
	        final WebBrowser webBrowser = new WebBrowser();
	       
	        //Use below code to check the status of the navigation process,
	        //or register a listener for the notification events.
	        webBrowser.addWebBrowserListener(
	            new WebBrowserListener() {
	            boolean isFirstPage = true;
							public void initializationCompleted(WebBrowserEvent event){;}            
	            public void downloadStarted(WebBrowserEvent event) {;}
	            public void downloadCompleted(WebBrowserEvent event) {;}
	            public void downloadProgress(WebBrowserEvent event) {;}
	            public void downloadError(WebBrowserEvent event) {;}
	            public void documentCompleted(WebBrowserEvent event) {
	                // Uncomment below code to test getContent()/setContent()/
	                // executeScript() APIs.
	                // As the setContent() call will invoke this event, which falls
	                // into a loop, so check if this event is fired by the first
	                // loaded page.
	                /*
	                if (isFirstPage) {
	                    testDOMAPI(webBrowser);
	                    isFirstPage = false;
	                }
	                */
	            }
	            public void titleChange(WebBrowserEvent event) {;}  
	            public void statusTextChange(WebBrowserEvent event) {;}        
	        });

	        try {
	        	
	            webBrowser.setURL(new URL("file:///C:/Documents and Settings/venky/workspace/Raindrop/resources/firestream/VRML/Nightclub/Demo-NightClub.htm"));
	        	//webBrowser.setURL(new URL("file://./resources/firestream/VRML/Nightclub/Demo-NightClub.htm"));
	            
	            // Below Chinese website tests unicode support.
	            //webBrowser.setURL(new URL("http://www.google.com/intl/zh-CN/"));
	            
	            // Print out debug messages in the command line.
	            //webBrowser.setDebug(true);
	        } catch (MalformedURLException e) {
	            System.out.println(e.getMessage());
	            return;
	        }               
	        webBrowser.setPreferredSize(new Dimension(500, 500));
	        //JPanel panel = new JPanel();
	        //panel.setLayout(new BorderLayout());
	        //panel.add(webBrowser, BorderLayout.CENTER);
	        
	        frame.setPreferredSize(new Dimension(400, 350));
	        frame.getContentPane().add(webBrowser, BorderLayout.CENTER);
	        frame.pack();
	        frame.setVisible(true);        
	    
        
		int port = -1;
		boolean toFile = false;
		boolean toScreen = false;
		boolean count = true;

		try {

			if (args.length > 4 || args.length == 0) {
				System.out.println(USAGE);
				return;
			}

			port = Integer.parseInt(args[0]);

			if (port <= 0 || port > 20000) {
				System.out.println(
					"Error: Please specifiy a port between 1 and 20000");
				return;
			}
			
			ServerSocket socket = new ServerSocket(port);
			
			
			int i = 1;
			while(true) {
				
				Socket inc = socket.accept();
				System.out.println("Now start Receiving Query Results to Display in VRML");
				
				Thread t = new VRMLResultDisplayHandler(inc, webBrowser);
				t.setName("VRMLResultDisplayHandler" + i);
				t.setDaemon(true);
				
				t.start();
				i++;
				
				//if(webBrowser.isInitialized())
				//{
				//	webBrowser.executeScript("javascript:var a = document.getElementsByName(\"Cortona\").item(0).Engine.Nodes(\"TCA\").Fields('setTCTemperature')= \"1,2200\"");
				//}
								
			}
		} catch (IOException ioe) {
			System.out.println(
				"Could not start the Query Receiver: " + ioe.getMessage());
		}
	}
}

class VRMLResultDisplayHandler extends Thread {
	private Socket incoming;
	private WebBrowser webBrowser;
	
	private String identifier;
	private String queryName;
	private boolean toScreen;
	private boolean toFile;
	private JFrame frame;
	private JTextArea text;
	private JScrollPane pane;
	private JPanel panel;
	private boolean count;
	private int tupleCount;

	public VRMLResultDisplayHandler(Socket i, WebBrowser _webBrowser) {
		this.incoming = i;
		this.webBrowser = _webBrowser;
	}
	
	public static float round(float val, int places) {
    	long factor = (long)Math.pow(10,places);
    	
    	val = val * factor;
    	
    	// Round to the nearest integer.
    	long tmp = Math.round(val);
    	
    	// Shift the decimal the correct number of places
    	// back to the left.
    	return (float)tmp / factor;
    }
	
	public int getInstrumentID(String _name)
	{
		String id = new String("-1");
		if(_name.trim().indexOf("DAN") ==0){
			id =_name.replaceFirst("DAN","");
		}else if(_name.trim().indexOf("DAE") ==0)
		{
			id =_name.replaceFirst("DAE","");
		}else if(_name.trim().indexOf("DAS") ==0)
		{
			id =_name.replaceFirst("DAS","");
		}else if(_name.trim().indexOf("SC") ==0)
		{
			id =_name.replaceFirst("SC","");
		}else if(_name.trim().indexOf("DFN") ==0)
		{
			id =_name.replaceFirst("DFN", "");
		}else if(_name.trim().indexOf("DFF") ==0)
		{
			id = _name.replaceFirst("DFF","");
		}
		
		return Integer.parseInt(id);
	}


	public String getInstrumentName(String _name)
	{
		String mappedName = new String(" ");
		if(_name.trim().indexOf("DAN") ==0){
			mappedName ="TCSLA";
		}else if(_name.trim().indexOf("DAE") ==0)
		{
			mappedName ="TCA";
		}else if(_name.trim().indexOf("DAS") ==0)
		{
			mappedName ="TCSRA";
		}else if(_name.trim().indexOf("SC") ==0)
		{
			mappedName ="TCB";
		}else if(_name.trim().indexOf("DFN") ==0)
		{
			mappedName ="TCC";
		}else if(_name.trim().indexOf("DFF") ==0)
		{
			mappedName = "TCD";
		}
		
		return mappedName;
	}

	public void run() {
		// 	Listen to incoming packets and place them into the Packet Pool.
		String instrumentName;
		int instrumentID;
		float reading;
		
        long startTime = System.currentTimeMillis();
		try {
			boolean enable = true;
			boolean haveSchema = false;
			XATTuple t;
			ObjectInputStream in =
				new ObjectInputStream(incoming.getInputStream());
			
			while(true)
			{
				//Receive the new Packets	
				try {
					Object o = in.readObject();
	
					NetworkPacket np = (NetworkPacket) o;
					Packet p = null;
					
					for (int j = 0; j < np.size(); j++) {
						
						p = np.getPacket(j);
						
						if(p.getControlMessage().equals("schema") && !haveSchema){
							System.out.println("Received The Schema: " + p.getObject().toString());
							haveSchema = true;
						}
						else if(p.getControlMessage().equals("schema")){
							//do nothing
						}
						else if (p.getControlMessage().compareTo("newquery") == 0) {
							identifier = p.getIdentifier();
							queryName = (String) p.getObject();
						}
						//Take the tuple and show the result
						else {
							t = (XATTuple) p.getObject();
							
							tupleCount++;
							
							System.out.println(t.getCell(1).toString()+" "+t.getCell(2)+" "+t.getCell(3).toString());
							instrumentName = getInstrumentName(t.getCell(1).toString());
							instrumentID = getInstrumentID(t.getCell(1).toString());
							reading = Math.round(Float.parseFloat(t.getCell(2).toString()));
							
							
							
							if(webBrowser.isInitialized())
							{
								webBrowser.executeScript("javascript:var a = document.getElementsByName(\"Cortona\").item(0).Engine.Nodes(\""+instrumentName+"\").Fields('setTCTemperature')= \""+instrumentID+","+reading+"\"");
							}	
						}
						
					}
				} catch (EOFException e) {
					System.out.println("Error");
					e.printStackTrace();
					in = new ObjectInputStream(incoming.getInputStream());
				} 
			}
		}catch(StreamCorruptedException sce){
			sce.printStackTrace();
		}catch (SocketException se) {
		    System.out.println("Time: " + (System.currentTimeMillis() - startTime) + "received a total of: " +tupleCount+" tuples.");
			System.out.println(
				"Connection Failed: Cannot get stream data from "
					+ incoming.getInetAddress()
					+ ":"
					+ incoming.getPort());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
