/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration;

import java.io.Serializable;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATPropertiesImp;
import java.net.*;
/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Feb 10, 2004
 * 
 * Represents an Application that is listening for the output of the
 * query plan.
 * 
 */
public class Application extends XATPropertiesImp implements Serializable {

	/**
	 * 
	 */
	public Application() {
		setDefaultProperties();
	}
	public String getIPAddress() {
		return this.getProperty("HOST_ADDRESS");
	}

	public String getMachineName() {
		return this.getProperty("NAME");
	}
	public int getPort() {
		return Integer.parseInt(this.getProperty("PORT"));
	}
	public int getRootID() {
		return Integer.parseInt(this.getProperty("ROOTID"));
	}
	public void setRootID(String rootID) {
		try {
			this.setProperty("ROOTID",rootID);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void setApplicationName(String name) {
		try {
			this.setProperty("NAME", name);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATPropertiesImp#setDefaultProperties()
	 */
	protected void setDefaultProperties() {

		super.setDefaultProperties();
		addProperty("NAME", "localhost");
		addProperty("HOST_ADDRESS", "127.0.0.1");
		addProperty("PORT", "16001");
		addProperty("CONNECTS_TO", "1");
		
		addProperty("ROOTID","0");//added by Song Wang for support of multiple applications in one query plan
	}

	public void setIPAddress(String address) {
		try {
			InetAddress[] ia = InetAddress.getAllByName(address);
			String r = ia[0].getHostAddress().toString();
			this.setProperty("HOST_ADDRESS", r);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void setPort(int port) {
		try {
			this.setProperty("PORT", Integer.toString(port));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public String toString(){
		return "Application: " + this.getProperty("HOST_ADDRESS") + ":" + this.getProperty("PORT") + " to Query Plan " + this.getProperty("CONNECTS_TO")+" to root operator(0:defalt value for single root query plan)"+ this.getProperty( "ROOTID");
	}

}
