/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration;

import java.io.Serializable;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATPropertiesImp;
import java.util.*;
import java.net.*;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Oct 23, 2003
 * 
 * Insert comments here.
 */
public class Machine extends XATPropertiesImp implements Serializable {

	ArrayList preferences;
	ArrayList algorithms;

	/**
	 * 
	 */
	public Machine() {
		super();
		preferences = new ArrayList();
		algorithms = new ArrayList();
		setDefaultProperties();
	}
	
	protected void setDefaultProperties(){
		super.setDefaultProperties();
		addProperty("NAME","localhost");
		addProperty("HOST_ADDRESS","127.0.0.1");
		addProperty("PORT","8000");
		addProperty("TUPLE_RECEIVER_PORT","9000");
		addProperty("CONNECTION_LISTENER_PORT","10000");
		addProperty("ADAPTIVE_HEURISTIC","none");
		addProperty("LOCAL_PARTITION_ADAPTER","none");
		addProperty("DEBUG","false");
		addProperty("UPDATE_OPERATOR_PROPERTY_FREQUENCY", "1000");
		addProperty("UPDATE_TREE_PROPERTY_FREQUENCY", "1000");
		addProperty("STATUS_CHECK_FREQUENCY", "1000");
		addProperty("STATS_TABLE_STATS", "false");
		addProperty("STARVATION_CHECK", "30");
		addProperty("YIELD_PROBABILITY", "0.9");
		addProperty("NO_WORK_SLEEP_TIME", "100");
		addProperty("WORKLOAD_RATIO", "1");
		addProperty("WORKLOAD_THRESHOLD", "50");
		addProperty("SCHEDULER_TYPE", "Regular"); //Rimma: can be Regular, Eddy, QueryMesh
		addProperty("MIGRATION_STRATEGY", "off");
		addProperty("MIGRATION_INTERVAL", "5000");
		addProperty("TOTAL_MIGRATION_TIMES", "1");
		addProperty("GUI","off");
		addProperty("GUI_TYPE","DISTRIBUTION_MANAGER");
		addProperty("PROJECT","off");
	}
	
	public String getMachineName(){ return this.getProperty("NAME");}
	public String getIPAddress(){ return this.getProperty("HOST_ADDRESS");}
	public int getPort() {return Integer.parseInt(this.getProperty("PORT"));}
	public int getTupleReceiverPort() {return Integer.parseInt(this.getProperty("TUPLE_RECEIVER_PORT"));}	
	public int getConnectionListenerPort() {return Integer.parseInt(this.getProperty("CONNECTION_LISTENER_PORT"));}
	public void setMachineName(String name) {
		try{
			this.setProperty("NAME",name);
		}catch(Exception e){ 
			e.printStackTrace();
		}
	}
	
	public void setPort(int port){
		try{
			this.setProperty("PORT",Integer.toString(port));
		}catch(Exception e){ 
			e.printStackTrace();
		}
	}
	
	public void setIPAddress(String address){
		try{
			InetAddress[] ia = InetAddress.getAllByName(address);
			String r = ia[0].getHostAddress().toString();
			this.setProperty("HOST_ADDRESS",r);
		}catch(Exception e){ 
			e.printStackTrace();
		}
	}
	
	public Iterator getSchedulingPreferences(){
		return preferences.iterator();
	}
	
	public Iterator getSchedulingAlgorithms(){
		return algorithms.iterator();
	}
	public void addSchedulingAlgorithm(XATSchedulingAlgorithm sa){
		algorithms.add(sa);
	}
	public void addSchedulingPreference(XATPreference p){
		preferences.add(p);
	}
	
	public String toString(){
		String ret = new String();
		
		ret+= "Machine: " + this.getMachineName() + "\n";
		ret+= "\t Properties: \t\t\t\t";
		Enumeration e = this.getProperties();
		while(e.hasMoreElements()){
			String name = (String) e.nextElement();
			String prop = this.getProperty(name);
			ret+= name + "=" + prop + ", ";
		}	
		ret+= "\n";
		ret+= "\t Scheduling Preferences: \t";
		for(Iterator i = preferences.iterator();i.hasNext();){
			XATPreference p = (XATPreference) i.next();
			ret += p.toString() + "\t";	
		}
		ret += "\n";
		ret+= "\t Scheduling Algorithms: \t";
		for(Iterator i = algorithms.iterator();i.hasNext();){
			XATSchedulingAlgorithm s = (XATSchedulingAlgorithm) i.next();
			ret += s.toString() + ", ";	
		}
		ret += "\n";
		return ret;
	}
	
	public String toShortString(){
		
		String ret = new String();
		ret+= this.getMachineName() + ": "+ this.getIPAddress();
		return ret;
	}
	
	public boolean equals(Machine m){
		
		if( this.getIPAddress().equals(m.getIPAddress()) &&
			(this.getPort() == m.getPort()) )
				return true;
				
	
		return false;
		
	}
}

