/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.probablisticjoinoperator.WMJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATPropertiesImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator.QueryPlanGenerator;
import edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.mjoinoptimizer.*;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Oct 27, 2003
 * 
 * Insert comments here.
 */
public class QueryPlan extends XATPropertiesImp implements Serializable{
	
	private XATDag queryPlan;
	private QueryPlanGenerator qpg;
	private int treeID;
	private MultiJoinOptimizer optimizer;

	public QueryPlan(int treeID) {
		super();
		queryPlan = null;
		qpg = null;
		this.treeID = treeID;
		setDefaultProperties();
	}
	
	public void setDefaultProperties(){
		super.setDefaultProperties();
		addProperty("FILE_NAME","none");
		addProperty("GENERATOR","none");
		addProperty("STREAMS","none");
		addProperty("QUERY_ID", "1");
		addProperty("QUERY_FILE_NAME", "none");
		addProperty("edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator","none");
		
	}
	
	public void setQueryPlanGenerator(QueryPlanGenerator g) throws Exception{
		
		if(qpg != null) throw new Exception("The Queryplan generator has already been set!");
		this.qpg = g;
	}
	
	public void generateQueryPlan(){
		try{
			//queryPlan = XATDagImp.convertTreeToDag(qpg.makeQueryPlan());	
			queryPlan = (XATDag)qpg.makeQueryPlan();
			
			// query plan initial optimization
			// if the query only contains an mjoin operator, invoke the mjoin optimizer.
			XATNode[] nodes = queryPlan.convertToArray();
			for(int z=0;z<nodes.length;z++){
				
			};
			if(nodes.length == 1 && nodes[0].getStreamOperator() instanceof WMJoinOperator) {
				WMJoinOperator op = (WMJoinOperator)nodes[0].getStreamOperator();
				if(op.getOptimizationMethod().equalsIgnoreCase("exhaustive-search")) {
					this.optimizer = new ExhaustiveSearchOptimizer(queryPlan);
					optimizer.setQueueModel(qpg.getDataModel());					
					optimizer.initialOptimization();
					
					// for static optimization evaluation
					queryPlan = optimizer.optimize(queryPlan);
					System.exit(0);
				}
				
				else if(op.getOptimizationMethod().equalsIgnoreCase("mtree-from-mjoin")) {
					this.optimizer = new FromMJoinOptimizer(queryPlan);
					optimizer.setQueueModel(qpg.getDataModel());
					optimizer.initialOptimization();
					
					// for static optimization evaluation
					queryPlan = optimizer.optimize(queryPlan);
					System.exit(0);
				}
				
				else if(op.getOptimizationMethod().equalsIgnoreCase("mtree-from-btree")) {
					this.optimizer = new FromBTreeOptimizer(queryPlan);
					optimizer.setQueueModel(qpg.getDataModel());
					optimizer.initialOptimization();
					
					// for static optimization evaluation
					queryPlan = optimizer.optimize(queryPlan);
					System.exit(0);
				}
			}
			
			queryPlan.setProperty("TREE_ID",""+treeID);
			XATNode[] temp = queryPlan.convertToArray();
			for(int i=0; i< temp.length;i++){
				temp[i].setTreeID(""+treeID);
			}
		}catch(Exception e){ 
			e.printStackTrace(); 
			System.exit(-1);
		}

	}
	
	/*
	 * Give unique Id to all operators in a query plan.
	 * Modified by Yali Zhu @ Aug 21, 2005.
	 */
	public int numberQueryPlanOperators(int startingNumber){
		int opcount = startingNumber;
		XATNode[] nodes = queryPlan.convertToArray();
		
		try{
			for(int i = 0; i < nodes.length ; i++){
				XATStreamOperator op = ((XATStreamOperator)nodes[i].getOperator());
				
				// assign id to the operator
				//op.setProperty("OBJECT_ID",Integer.toString(opcount));
				//New version. Modified by Yali Zhu on March 31, 2006. Set OBJECT_ID to be the same as operator ID. 
				//Operator ID is the same as defined in the query plan xml config file. This avoid ID confusion.
				op.setProperty("OBJECT_ID", ""+op.getOperatorID());
				opcount++;
			}
		}catch(Exception e){
			e.printStackTrace();
		}
		return opcount;	
	}
	
	/*
	 * Give unique Id to all states in a query plan.
	 * Modified by Yali Zhu @ Aug 21, 2005.
	 */
	public int numberQueryPlanStates(int startingNumber){
		int statecount = startingNumber;
		XATNode[] nodes = queryPlan.convertToArray();
		
		try{
			for(int i = 0; i < nodes.length ; i++){
				XATStreamOperator op = ((XATStreamOperator)nodes[i].getOperator());
				
				// assign id to its states
				statecount = op.setStateID(statecount);
			}
		}catch(Exception e){
			e.printStackTrace();
		}
		return statecount;	
	}
	
	/*
	 * Give unique Id to all queues in a query plan.
	 * Modified by Yali Zhu @ Aug 21, 2005.
	 */
	public int numberQueryPlanQueues(int startingNumber){
		int queuecount = startingNumber;
		XATNode[] nodes = queryPlan.convertToArray();
		
		try{
			for(int i = 0; i < nodes.length ; i++){
				XATStreamOperator op = ((XATStreamOperator)nodes[i].getOperator());
				
				// assign id to its input queues
				Iterator it = op.getInputQueues();
				XATQueue queue;
				while(it.hasNext()) {
					queue = (XATQueue)it.next();
					if(queue.getQueueID() < 0) {
						queue.setQueueID(queuecount);
						queuecount++;
					}
				}
				
				// if root node, assign id to its output queue as well. 
				//modified by Song Wang for supporting multiple root DAG query plan
				for (Iterator j = queryPlan.getRoots(); j.hasNext();){
					if(((XATNode)j.next()).equals(nodes[i])) {
						op = ((XATStreamOperator)nodes[i].getOperator());
						it = op.getOutputQueues();
						while(it.hasNext()) {
							queue = (XATQueue)it.next();
							if(queue.getQueueID() < 0) {
								queue.setQueueID(queuecount);
								queuecount++;
							}
						}				
					}
				}
			}
		}catch(Exception e){
			e.printStackTrace();
		}
		return queuecount;	
	}
	
	
	public XATDag getQueryPlan() throws Exception{
		if(queryPlan == null) throw new Exception("The query plan has not been created yet!");
		return queryPlan;
	}
	

	
	public String toString(){
		String ret = new String();
		
		ret+= "Query Plan: " + "\n";
		ret+= "\t Properties: \t\t\t\t";
		Enumeration e = this.getProperties();
		while(e.hasMoreElements()){
			String name = (String) e.nextElement();
			String prop = this.getProperty(name);
			ret+= name + "=" + prop + ", ";
		}
		ret+="\n";
		ret+="\t XATTree: \n";
		ret += queryPlan.toString();
		return ret;
		
	}
	
	
}
