/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Nov 30, 2004
 * This is used to activate the split operator. Two extra things need to be done besides the typical
 * Activate used by other operators: (1) Create dynamic queues in the operator if the parent (join)
 * is distributed to more than one machines. (2) Set the partitionMaps into the split operator. 
 * Note that the connection operations could be done in the same fashion, for example, use SendData, SendStream
 * etc.
 * The real process of activating the operator is done by the ConnectionListener in each query processor 
 */
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection;

import java.io.Serializable;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.*;

/**
 * @author Bin Liu
 */
public class ActivatePartition implements Serializable, Connection {

	private String operatorID;
	private int queueNum;
	private Hashtable partitionMaps; // the partitionMaps for this operator only
	private Machine[] machines;
	private Machine curMachine;
	
	// constructor function
	public ActivatePartition(String operatorID, int queueNum, Hashtable maps, Machine[] ms, Machine m){
		this.operatorID = operatorID;
		this.queueNum = queueNum;
		this.partitionMaps = maps;
		this.machines = ms;
		this.curMachine = m;
	}
	
	/* 
	 * The function name might be confusing, but i use this since it is from the Connection interface 
	 * In this case, it returns the operator ID that is activated.
	 */
	public String getFromOperator() throws ConnectionException {
		return this.operatorID;
	}

	/* 
	 * Again, the function name might be confusing, but i use this since it is from the Connection interface 
	 * In this case, it returns the queue ID that is activated.
	 */
	public int getFromQueueNumber() throws ConnectionException {
		return this.queueNum;
	}

	/* 
	 * The followings are basic gets functions.
	 */
	public String getClassName() {
		return this.getClass().getName();
	}	
	public Hashtable getPartitionMaps(){
		return this.partitionMaps;
	}
	public Machine[] getMachines(){
		return machines;
	}
	public String toString(){
		return "Activating the operator (ID): " + this.operatorID;
	}
	public Machine getCurrentMachine(){
		return curMachine;
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getFromIPAddress()
	 */
	public String getFromIPAddress() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getFromPort()
	 */
	public int getFromPort() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getStreamLength()
	 */
	public int getStreamLength() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getStreamName()
	 */
	public String getStreamName() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToIPAddress()
	 */
	public String getToIPAddress() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToOperator()
	 */
	public String getToOperator() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToPort()
	 */
	public int getToPort() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToQueueNumber()
	 */
	public int getToQueueNumber() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

} //end class
