/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Mar 2, 2004
 * 
 * This class is an interface for a Connection between query objects. Tuples will 
 * be passed amongst this connection.
 * 
 * There are several possibilities for a connection.
 * 
 * - Local Machine
 * - Stream to Machine
 * - Machine to Machine
 * - Machine to Application
 * 
 * Because of this we keep track of the local and remote operator that
 * we will be connecting to.
 * 
 * From Location
 * --------------
 * 	IP Address
 *  Port
 *  Operator id
 *  Output Queue number
 * 
 * To Location
 * --------------
 * 	IP Address
 * 	Port
 *  Operator id
 * 	Input Queue number
 * 
 */
public interface Connection {


/** 
 * Returns the name of the connection class.
 * @return
 */
public String getClassName();

/**
 * Returns the IP address where the data will come from.
 * 
 * @return String -ipAddress
 */
public String getFromIPAddress() throws ConnectionException;

/**
 * Returns the ID for the operator where the data will come from.
 * 
 * @return String operatorID
 */

public String getFromOperator() throws ConnectionException;

/**
 * Returns the port where the data will come from.
 * 
 * @return int port
 */
public int getFromPort() throws ConnectionException;


/**
 * Returns the queue number from where the data will come from.
 * 
 * @return int queueNumber
 */
public int getFromQueueNumber() throws ConnectionException;

/**
 * Returns the amount of time for which the stream should be run.
 * 
 * @return
 * @throws ConnectionException
 */
public int getStreamLength() throws ConnectionException;

/**
 * Returns the stream name that will connect to the "to" operator
 * 
 * @return the Stream Name
 * @throws ConnectionException
 */
public String getStreamName() throws ConnectionException;

/**
 * Returns the IP address where the data will go to.
 * 
 * @return String -ipAddress
 */
public String getToIPAddress() throws ConnectionException;

/**
 * Returns the ID for the operator to where the data will go.
 * 
 * @return String operatorID
 */
public String getToOperator() throws ConnectionException;

/**
 * Returns the port where the data will go to.
 * 
 * @return int port
 */
public int getToPort() throws ConnectionException;

/**
 * Returns the queue number to where the data will go.
 * 
 * @return int queueNumber
 */
public int getToQueueNumber() throws ConnectionException;

/**
 * Returns the String representation of the connection
 * @return
 */
public String toString();
}
