/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection;
import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ConnectException;
//import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.Enumeration;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.spreadfire.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.SubPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.ExecutionSynced;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.GetSyncTimeStamp;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.MigrationProtocol;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.PlanChanged;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.RequestedState;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendOperatorsToFillStates;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendStateToMachine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendSubPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendSyncFlag;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.StatesFilled;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SyncTimestampObtained;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.PacketPool;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.DistributedExecutionController;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.StreamSplitOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATMemoryQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPersistentQueueImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.adprotocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.protocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
//import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.SymmMwayHashJoinOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 8, 2003
 * 
 * The connection listner is responsible for running on the Raindrop Client
 * and listens for connection requests from the Distribution Manger. These
 * requests are then processed by the listener and the query plan is updated
 * to reflect these changes. There are currently 3 types of connections:
 * 
 *  - Stream to Machine
 *  - Machine to Machine
 *  - Machine to Application
 * 
 */
public class ConnectionListener extends Thread implements Runnable {

	/**Stores the reference to the port number*/
	private int port = -1;
	private Socket incoming;
	private ConnectionTable inConnections;
	private ConnectionTable outConnections;
	private String clientname;
	private boolean keepGoing;
	private ArrayList connectionHandlers;
	private DistributedExecutionController dExec;
	
	/**
	 * Creates a new Package receiver for the execution controller.
	 */
	public ConnectionListener(int port, String clientname, ConnectionTable in, ConnectionTable out, DistributedExecutionController exec)
		throws IOException {
		super();
		this.port = port;
		this.inConnections = in;
		this.outConnections = out;
		this.clientname = clientname;
		keepGoing = true;
		connectionHandlers = new ArrayList(10);
		this.dExec = exec;
	}

	/* Runs the Connection Listener
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		try {
			ServerSocket socket = new ServerSocket(this.port);
			System.out.println("Starting up Connection Listener on port: " + port);
			int i = 1;
			while (keepGoing) {
				Socket inc = socket.accept();
				ConnectionHandler newHandler = new ConnectionHandler(inc, inConnections, outConnections, i, dExec);
				newHandler.setName("ConnectionHandler" + i);
				newHandler.setDaemon(true);
				
				int recounter = Integer.parseInt(System.getProperty("REDISTACTUAL"));
				recounter++;
				System.setProperty("REDISTACTUAL", ""+recounter);
				System.setProperty("ISREDIST", "true");

				newHandler.start();
				connectionHandlers.add(newHandler);
				i++;
			}

		} catch (IOException ioe) {
			System.out.println("Could not start the Connection Listener on port " + this.port + ioe.getMessage());
		}
	}

	public void requestStop() {
		System.out.println("Trying to interrupt the threads");

		Iterator i = connectionHandlers.iterator();

		while (i.hasNext()) {
			ConnectionHandler c = (ConnectionHandler) i.next();
			if (c != null) {
				c.requestStop();
			}
		}

		keepGoing = false;
		interrupt();

	}
}

/**
 * 
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 8, 2003
 * 
 * The PacketReceiverHandler class is responsible for
 * maintaining a connection between the local execution engine
 * and a remote engine. The PacketReceiver will spawn one of these threads
 * whenever a Packet from a new location is sent.
 * 
 */
class ConnectionHandler extends Thread {

	private Socket incoming;
	private PacketPool pool;
	private int handlerNumber;
	private boolean keepGoing;
	private ConnectionTable inConnections;
	private ConnectionTable outConnections;
	private StatisticsGatherer stats;
	private DistributedExecutionController dExec;
	
	public ConnectionHandler(
		Socket i,
		ConnectionTable in,
		ConnectionTable out,
		int handlerNumber, DistributedExecutionController exec) {
		this.incoming = i;
		this.handlerNumber = handlerNumber;
		this.keepGoing = true;
		this.inConnections = in;
		this.outConnections = out;
		stats = StatisticsGatherer.getStatisticsGather();
		this.dExec = exec;
		
	}

	public synchronized void run() {
	
		try {

			ObjectInputStream in = new ObjectInputStream(incoming.getInputStream());
			XATStreamOperator operator = null;
			String stream;
			String operatorNumber;
			
			boolean enable = true;
			while (enable) {

				//Receive the new Packets
				Object o = in.readObject();
				Packet cp = (Packet) o;
				Connection c = null; 
				//System.out.println(cp.toString());
				Object curO = cp.getObject();  
				
				// -- Bin @ Dec.10,2004 to handle the commands from distribution manager 
				// that exclusively for adapting partitions
				if (curO instanceof Protocol) {
					System.out.println("\nReceiving Partition Adaptation Protocol.");
					if (curO instanceof ComputePartitionsToMove) {
						// (1) first get the operator
						operatorNumber = ((ComputePartitionsToMove)curO).getOperatorID();
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
								operator = temp;
								break;
							}
						}
						// (2) call dExec to get partition IDs to move and 
						// send PartitionsToMove message to Distribution Manager via statistics-gathering connection
						System.out.println("Compute partitions to move from the tree.");
						dExec.inGlobalAdaptation = true;
						dExec.computePartitionsToMoveInTree(operator,((ComputePartitionsToMove)curO).getRate());		
						//System.out.println("Compute partitions to move from operator " + operator);
					} else if (curO instanceof DeactivatePartitions) {
						// note that the current operator ID is the parent node of the partition operator 
						// that this message will work on. (1) get the operator indicated by the message
						String parentNumber = ((DeactivatePartitions)curO).getOperatorID();
						XATStreamOperator parentOper = null;
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(parentNumber)) {
								parentOper = temp;
								break;
							}
						}
						// (2) find all the children node, and deactivate the partitions if they are active in current machine
						Object[] children = (parentOper.getNode()).getChildren();
						String senderIPAddr = ((DeactivatePartitions)curO).getSenderMachine().getIPAddress();
						dExec.inGlobalAdaptation = true;
						//dExec.deactivatePartitions(parentOper, children,((DeactivatePartitions)curO).getPartitionsToMove(),senderIPAddr);
						dExec.deactivatePartitionsInTree(((DeactivatePartitions)curO).getPartitionsToMove(),senderIPAddr);
					} else if (curO instanceof SendPartitions){
						// (1) we first get the operator, 
						operatorNumber = ((SendPartitions)curO).getOperatorID();
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
								operator = temp;
								break;
							}
						}
						// (2) get partitions states and send them to the receiver (via ReceivePartitions message)
						//dExec.sendPartitions(operator,((SendPartitions)curO).getPartsToSend(),((SendPartitions)curO).getReceiverMachine());
						dExec.sendPartitionsInTree(((SendPartitions)curO).getPartsToSend(),((SendPartitions)curO).getReceiverMachine());
					} else if (curO instanceof ReceivePartitions) {
						// System.out.println("Yes, receiving partitio ns: " + ((ReceivePartitions)curO).getPartitionsIDs().size());
						// (1) get the operator
						operatorNumber = ((ReceivePartitions)curO).getOperatorID();				    
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
					        XATStreamOperator temp = ((XATStreamOperator) j.next());
					        if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
					            operator = temp;
					            break;
					        }
					    }
					    // (2) set the partitions to the operator, and notifiy the DM after it is finished
					    System.out.println("Receive protocol to receive moved partitions");
					    dExec.receivePartitions(operator,((ReceivePartitions)curO).getPartitionsIDs(),((ReceivePartitions)curO).getPartitions());
					} else if (curO instanceof ReceivePartitionsInTree) {
						// System.out.println("Yes, receiving partitions: " + ((ReceivePartitions)curO).getPartitionsIDs().size());
					    System.out.println("Receive protocol to receive moved partitions in Tree");
						dExec.receivePartitionsInTree(((ReceivePartitionsInTree)curO).getPartitionsIDs(), ((ReceivePartitionsInTree)curO).getPartitions());
					} else if (curO instanceof ReactivatePartitions){
						String parentNumber = ((ReactivatePartitions)curO).getOperatorID();
						XATStreamOperator parentOper = null;
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(parentNumber)) {
								parentOper = temp;
								break;
							}
						}
						// (2) find all the children node, and reactivate the partitions if they are active in current machine
						Object[] children = (parentOper.getNode()).getChildren();
						System.out.println("Receive protocol to reactivate partitions");
//					    if(dExec.inLocalMigration) {
//					        try {
//					            Thread.sleep(1000);
//					        } catch(Exception e) {
//					            e.printStackTrace();
//					        }
//					    }
						
						//dExec.reactivatePartitions(children,((ReactivatePartitions)curO).getPartitionsToActivate(),((ReactivatePartitions)curO).getToMachine());
						dExec.reactivatePartitionsInTree(((ReactivatePartitions)curO).getPartitionsToActivate(),((ReactivatePartitions)curO).getToMachine());
						dExec.inGlobalAdaptation = false;
					} else if (curO instanceof ForcePartitionsToPush){
						// (1) first get the operator
						operatorNumber = ((ForcePartitionsToPush)curO).getOperatorID();
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
								operator = temp;
								break;
							}
						}
						// (2) call dExec to get partition IDs to push and the actual operations 
						System.out.println("Yes, receive the forcePartitionsToPush message.");
						dExec.forcePartitionsToPush(operator,((ForcePartitionsToPush)curO).getRate());												
					} else if (curO instanceof SendDuplicatePartitions) {
					    dExec.inGlobalAdaptation = true;
					    ArrayList partitions = ((SendDuplicatePartitions)curO).getPartitionsToDuplicate();
					    Machine fromM = ((SendDuplicatePartitions)curO).getFromMachine();
					    Machine toM = ((SendDuplicatePartitions)curO).getToMachine();
					    dExec.sendDuplicatePartitions(partitions, fromM, toM);
					} else if (curO instanceof StopDuplicatePartitions) {
					    ArrayList partitions = ((StopDuplicatePartitions)curO).getPartitionsToStopDuplicate();
					    Machine fromM = ((StopDuplicatePartitions)curO).getFromMachine();
					    Machine toM = ((StopDuplicatePartitions)curO).getToMachine();
					    dExec.stopDuplicatePartitions(partitions, fromM, toM);
					    dExec.inGlobalAdaptation = false;
					}
				} 
				
				// Yali's Globle Plan Migration Protocols.
				else if (curO instanceof MigrationProtocol) {
					
					/** need to get sync timestamp from this machine */
					if(curO instanceof GetSyncTimeStamp) {
						
						ArrayList leafOperators = new ArrayList();
						ArrayList activeLeafOperators = new ArrayList();
						ArrayList oldSubplanOperators = new ArrayList();
						XATStreamOperator oldSubplanRootOperator = null;
						
						this.dExec.getTree()[0].appendMigrationActivity("\n\n\n ******* New Migration Process Activated by DM ********");
						
						System.out.println("\n ~~~ MigrationProtocol ~~~ : local sync timestamp requested by DM");
						ArrayList leafOperatorIDs = ((GetSyncTimeStamp)curO).getLeafOperatorIDs();
						ArrayList subplanOperatorIDs = ((GetSyncTimeStamp)curO).getOperatorIDs();
						int oldrootid = ((GetSyncTimeStamp)curO).getOldSubplanRootOpID();
						this.dExec.getTree()[0].appendMigrationActivity("\n Op " + subplanOperatorIDs.get(0) + " and Op " + subplanOperatorIDs.get(1) + " are to be switched");
						this.dExec.getTree()[0].appendMigrationActivity("\n\n ~~~ MigrationProtocol ~~~ : received local sync timestamp request from DM");

						this.dExec.getTree()[0].appendMigrationActivity("\n peek input queues of leaf operators: " + leafOperatorIDs);
						
						XATTree[] trees = dExec.getTree();
						for(int i=0; i<trees.length; i++) {
							XATTree querytree = trees[i];
							XATNode[] nodes = querytree.convertToArray();

							for(int j=0; j<nodes.length; j++) {
								XATStreamOperator op = nodes[j].getStreamOperator();
								String id = Integer.toString(op.getObjectID());
								
								// figure out leaf operators that need to be synced
								if(leafOperatorIDs.contains(id)) {
									leafOperators.add(op);
									if(op.isActive())
										activeLeafOperators.add(op);
								}
								
								// figure out operators to be replaced by subplan 
								if(subplanOperatorIDs.contains(id)) {
									oldSubplanOperators.add(op);	
									op.setIsSubPlanOperator(true);
								}
								
								// figure out the root operator of the old subplan active on this machine
								if((oldrootid != -1) && (oldrootid == Integer.parseInt(id))) {
									oldSubplanRootOperator = op;
									op.setIsSubPlanRootOperator(true);
								}
							}			
						}
						
						//System.out.println("leafOperators: "+leafOperators);
						//System.out.println("ActiveLeafOperators: " + activeLeafOperators);
						//System.out.println("subplanOperatorIDs: " + subplanOperatorIDs);
						//System.out.println("oldSubplanOperators: " + oldSubplanOperators);
						
						this.dExec.getMigrationkeeper().setLeafOperators(leafOperators);
						this.dExec.migrationkeeper.setActiveLeafOperators(activeLeafOperators);
						this.dExec.migrationkeeper.setOldSubplanOperators(oldSubplanOperators);
						this.dExec.migrationkeeper.setOldSubplanRootOperator(oldSubplanRootOperator);
						
						this.dExec.getTree()[0].appendMigrationActivity("\n active leaf operators to peek: ");

						// inactivate leaf operators
						for(int i=0; i<activeLeafOperators.size(); i++) {
							this.dExec.getTree()[0].appendMigrationActivity(" " + ((XATStreamOperator)activeLeafOperators.get(i)).getObjectID());
							((XATStreamOperator)activeLeafOperators.get(i)).setActive(false);

						}
						
						XATStreamOperator op;
						long currentmaxtime = 0;
						long temptime;
						
						// get max timestamp accumulated in leaf operator queues
						for(int i=0; i<activeLeafOperators.size(); i++) {
							op = (XATStreamOperator)activeLeafOperators.get(i);						
							Object[] cnodes = op.getNode().getChildren();
							
							for(int k=0; k<cnodes.length; k++) {
								// is a stream
								if(!(cnodes[k] instanceof XATNode) && (op.getInputQueue(k).getQueueSize() != 0)) {
									System.out.println("peek input queue " + k + " of operator " + op);
									this.dExec.getTree()[0].appendMigrationActivity("\npeek input queue " + k + " of operator " + op.getObjectID());
									
									temptime = op.getInputQueue(k).peek(op).getMaxTimeStamp();
									
									if(temptime > currentmaxtime)
										currentmaxtime = temptime;
								}
							}
							//op.setSyncLeaf(true);
						}
						
						// now send the maxtimestamp back to distribution manager.
						System.out.println("Now send the sync time back to DM: " + currentmaxtime);
						this.dExec.getTree()[0].appendMigrationActivity("\n Now send the sync time back to DM:  " + currentmaxtime);

						SyncTimestampObtained st = new SyncTimestampObtained(currentmaxtime);
					    try{
					    	if (dExec.sStat == null){
							    dExec.sStat = new Socket(dExec.getConfig().getManagerIPAddress(),23000);
						    }		
					    	if(dExec.outStat == null){
								dExec.outStat = new ObjectOutputStream(dExec.sStat.getOutputStream());			
					    	}
					    	//dExec.outStat.reset();
					    	dExec.outStat.writeObject(st);
					    	dExec.outStat.flush();
					    } catch(Exception e){
						    e.printStackTrace();
					    }
					    continue;
					}
					
					if(curO instanceof SendSyncFlag) {
						
						ArrayList activeLeafOperators = this.dExec.migrationkeeper.getActiveLeafOperators();
						XATStreamOperator oldSubplanRootOperator = this.dExec.migrationkeeper.getOldSubplanRootOperator();
						// add a sync flag to stream queues connecting to contributing leaf operators.
						long synctime = ((SendSyncFlag)curO).getSyncTimestamp();
						System.out.println("\n~~~ MigrationProtocol ~~~ : receive system sync timestamp: " + synctime);
						this.dExec.getTree()[0].appendMigrationActivity("\n\n ~~~ MigrationProtocol ~~~ : receive system sync timestamp: " + synctime);
	
						
/*						// insert a sync tuple to stream queues of leaf operator
						// the tuple's maxtimestamp is set to be -1, indicating that this is a special sync operator
						XATStreamOperator op;
						XATTuple synctuple;
						
						for(int i=0; i<activeLeafOperators.size(); i++) {
							op = (XATStreamOperator)activeLeafOperators.get(i);						
							Object[] cnodes = op.getNode().getChildren();
							XATQueue queue;
							for(int k=0; k<cnodes.length; k++) {
								
								// if is a stream
								if(!(cnodes[k] instanceof XATNode)) {
									int count = 0;
									queue = op.getInputQueue(k);
									
									// insert sync tuple to the correct position
									while(count<queue.getQueueSize(op)) {
										if(queue.peekNthTuple(count, op).getMaxTimeStamp() >= synctime) {
											synctuple = new XATTupleImp();
											synctuple.setMaxTimeStamp(-1);
											queue.insertTuple(count, synctuple, null);
											break;
										}
									}
								}
							}
							//op.setSyncLeaf(true);
						}
						
						// now activate the leaf operators
						for(int i=0; i<activeLeafOperators.size(); i++) {
							((XATStreamOperator)activeLeafOperators.get(i)).setActive(true);
						}
						
						// if the old root operator of the subplan is active on this machine
						// wait until it receives sync tuple from all its inputs (issynced flag been set to true)
						// then send a allexecutionsynced message back to the distribution manager.
						// if the old root is not active on this machine, this message processing is done.
						 */
						
						ArrayList oldops = dExec.getMigrationkeeper().getOldSubplanOperators();
						for(int i=0; i<oldops.size(); i++) {
							//((XATStreamOperator)oldops.get(i)).setIsSynced(true);
							((XATStreamOperator)oldops.get(i)).setActive(false);
						}
						
						if(oldSubplanRootOperator != null) {
							oldSubplanRootOperator.setIsSynced(true);
							// wait until execution is synced
							while(!oldSubplanRootOperator.getIsSynced()) {
								Thread.sleep(500);
							}
							
							// send all-synced message back to the distribution manager
							System.out.println("Now send the sync ack back to DM.");
							this.dExec.getTree()[0].appendMigrationActivity("\n Now send the sync ack back to DM.");
							
							
							ExecutionSynced es = new ExecutionSynced();
						    try{
						    	if (dExec.sStat == null){
								    dExec.sStat = new Socket(dExec.getConfig().getManagerIPAddress(),23000);
							    }		
						    	if(dExec.outStat == null){
									dExec.outStat = new ObjectOutputStream(dExec.sStat.getOutputStream());			
						    	}
						    	//dExec.outStat.reset();
						    	dExec.outStat.writeObject(es);
						    	dExec.outStat.flush();
						    } catch(Exception e){
							    e.printStackTrace();
						    }
							
							oldSubplanRootOperator.setIsSynced(false);
						}
	
						continue;
					}
					
					if(curO instanceof SendSubPlan) {
						ArrayList oldstates = new ArrayList();
						SubPlan subplan = new SubPlan();
						ArrayList oldSubplanOperators = this.dExec.migrationkeeper.getOldSubplanOperators();
						this.dExec.getCurrentScheduler().isLocked = true;
						try{
							while(this.dExec.treeLocked) {
								System.out.println("Connection Listerner: tree is locked.");
								Thread.sleep(1000);
							}
						}catch(Exception e) {
							e.printStackTrace();
						}
						
						this.dExec.treeLocked = true;
						
						for(int i=0; i<oldSubplanOperators.size(); i++) {
							//((XATStreamOperator)oldops.get(i)).setIsSynced(true);
							((XATStreamOperator)oldSubplanOperators.get(i)).setActive(false);
						}

						// Now activate the leaf operators
						// This is necessary in case the query plan is already synced when the migration starts
						// so that the SendSyncFlag won't be sent by the distributed manager.
						ArrayList activeleafops = this.dExec.getMigrationkeeper().getActiveLeafOperators();
						for(int i=0; i<activeleafops.size(); i++) {
							((XATStreamOperator)activeleafops.get(i)).setActive(true);
						}
						
						subplan = ((SendSubPlan)curO).getSubPlan();
						System.out.println("\n~~~ MigrationProtocol ~~~ : receive subplan from DM: " + subplan);
						this.dExec.getTree()[0].appendMigrationActivity("\n\n ~~~ MigrationProtocol ~~~ : received subplan from DM: " + subplan);

						
						// first keep the references to all states in the old subplan
						for(int i=0; i<oldSubplanOperators.size(); i++){
							XATStreamOperator op = (XATStreamOperator)oldSubplanOperators.get(i);
							OperatorState[] states = op.getState();
							for(int j=0; j<states.length; j++) {
								oldstates.add(states[j]);
							}
							stats.removeQueryObject(op);
						}
						
						this.dExec.migrationkeeper.setSubPlan(subplan);
						this.dExec.getTree()[0].appendMigrationActivity("\n disconnecting old subplan... done.");
						this.dExec.migrationkeeper.setOldStates(oldstates);
						this.dExec.getTree()[0].appendMigrationActivity("\n connecting new subplan... done.");
						
						// disconnect old subplan and connect new subplan
						//System.out.println("\n\nOriginal query plan:\n" + this.dExec.getTree()[0]);
						//System.out.println("\n OldSubplanoperators: " + oldSubplanOperators);
						ArrayList disconnectedNodes = this.dExec.disconnectOldSubPlan(oldSubplanOperators);	
						//System.out.println("\n\nAfter disconnect old subplan:\n" + this.dExec.getTree()[0]);
						
						//System.out.println("\n subplan: " + subplan);
						//System.out.println("\n disconnectedNodes: " + disconnectedNodes);
						this.dExec.connectNewSubPlan(subplan, disconnectedNodes);
						
						//System.out.println("\n\nAfter connect new subplan:\n" + this.dExec.getTree()[0]);
						
						ArrayList subplanoperators = subplan.getOperators();
						for(int i=0; i<subplanoperators.size(); i++) {
							((XATStreamOperator)subplanoperators.get(i)).setIsSubPlanOperator(false);
							((XATStreamOperator)subplanoperators.get(i)).setIsSubPlanRootOperator(false);
						}
						
						//System.out.println("\n\nAfter reset operator flag:\n" + this.dExec.getTree()[0]);
						
						
						
						// send planchanged message back to the distribution manager
						System.out.println("Now send the planchanged ack back to DM.");
						this.dExec.getTree()[0].appendMigrationActivity("\n Now send the planchanged ack back to DM.");
						this.dExec.treeLocked = false;
						PlanChanged pc = new PlanChanged();
					    try{
					    	if (dExec.sStat == null){
							    dExec.sStat = new Socket(dExec.getConfig().getManagerIPAddress(),23000);
						    }		
					    	if(dExec.outStat == null){
								dExec.outStat = new ObjectOutputStream(dExec.sStat.getOutputStream());			
					    	}
					    	//dExec.outStat.reset();
					    	dExec.outStat.writeObject(pc);
					    	dExec.outStat.flush();
					    } catch(Exception e){
						    e.printStackTrace();
					    }
						continue;
					}
					if(curO instanceof SendOperatorsToFillStates) {
						ArrayList statesToFill = new ArrayList();
						SubPlan subplan = this.dExec.migrationkeeper.getSubPlan();
						ArrayList oldstates = this.dExec.migrationkeeper.getOldStates();
						
						ArrayList operatorIDs = ((SendOperatorsToFillStates)curO).getOperatorIDs();
						Hashtable statelocations = subplan.getStateLocationTable();
						ArrayList subplanoperators = subplan.getOperators();
						System.out.println("\n\nMigrationProtocol: receive operators to fill states from DM: " + operatorIDs);
						this.dExec.getTree()[0].appendMigrationActivity("\n\n ~~~ MigrationProtocol ~~~ : receive operators to fill states from DM: " + operatorIDs);
						this.dExec.getTree()[0].appendMigrationActivity("\n filling states ... done.");
						
						//oldstates
						
						// figure out the states to copy or recompute on this machine
						for(int i=0; i<subplanoperators.size(); i++) {
							XATStreamOperator op = (XATStreamOperator)subplanoperators.get(i);
							if(operatorIDs.contains(Integer.toString(op.getObjectID()))) {
								OperatorState[] states = op.getState();
								for(int k=0; k<states.length; k++)
									statesToFill.add(states[k]);
							}
						}
						this.dExec.migrationkeeper.setStatesToFill(statesToFill);
						// TODO need to fix bugs later
						//this.dExec.fillStates(oldstates, statelocations, statesToFill);
						
						// now update stats table and scheduler
						// TODO right now only support RoundRobin Scheduler
						XATTree[] trees = this.dExec.getTree();
						this.dExec.getCurrentScheduler().setTrees(trees);
						for(int tsize=0; tsize<trees.length; tsize++) {
							trees[tsize].setOptimized(true);
						}
					
						this.dExec.resetLightWeightStats();
						
						this.dExec.migrationkeeper = null;
						
						this.dExec.getCurrentScheduler().isLocked = false;
						
						//leafOperators.clear();
						//activeLeafOperators.clear();
						//oldSubplanOperators.clear();
						//oldSubplanRootOperator = null;
						
						//reset all states' isMatchedFlag to be false
						
						// send StatesFilled message back to the distribution manager
						StatesFilled sf = new StatesFilled();
						
					    try{
					    	if (dExec.sStat == null){
							    dExec.sStat = new Socket(dExec.getConfig().getManagerIPAddress(),23000);
						    }		
					    	if(dExec.outStat == null){
								dExec.outStat = new ObjectOutputStream(dExec.sStat.getOutputStream());			
					    	}
					    	
					    	//dExec.outStat.reset();
					    	dExec.outStat.writeObject(sf);
					    	dExec.outStat.flush();
					    } catch(Exception e){
						    e.printStackTrace();
					    }
					    System.out.println("\nLocal migration is done.");
						this.dExec.getTree()[0].appendMigrationActivity("\nLocal migration is done.");
						this.dExec.getTree()[0].appendMigrationActivity("\nSend migration-done ack to DM.");
						
						continue;
					}
					
					if (curO instanceof SendStateToMachine) {
						Machine tomachine = ((SendStateToMachine) curO).getRequestingMachine();
						ArrayList oldstates = this.dExec.migrationkeeper.getOldStates();
						
						int stateID = ((SendStateToMachine) curO).getStateID();

						for (int i = 0; i < oldstates.size(); i++) {
							OperatorState state = (OperatorState) oldstates.get(i);
							if (state.getStateID() == stateID) {
								try {
									Socket ms = new Socket(tomachine.getIPAddress(), tomachine.getConnectionListenerPort());
									ObjectOutputStream mout = new ObjectOutputStream(ms.getOutputStream());
									RequestedState rs = new RequestedState(state);

									Packet pk = new Packet("migrationprotocol",rs);
									mout.writeObject(pk);
									mout.close();
									ms.close();
									System.out.println("Sending requested state " + stateID + "to machine " + tomachine);

								} catch (Exception e) {
									e.printStackTrace();
								}
								break;
							}
						}
					}
					
					if(curO instanceof RequestedState) {
						OperatorState receivedstate = ((RequestedState)curO).getRequestedState();
						System.out.println("\n\nReceiving requested state " + receivedstate.getStateID() + "from remote machine ");
						ArrayList oldstates = this.dExec.migrationkeeper.getOldStates();
						ArrayList statesToFill = this.dExec.migrationkeeper.getStatesToFill();
						
						for(int i=0; i<statesToFill.size(); i++) {
							OperatorState state = (OperatorState)statesToFill.get(i);
							if(state.getStateID() == receivedstate.getStateID()) {
								state.copyStateContent(receivedstate);
								oldstates.add(state);
								statesToFill.remove(i);
								break;
							}
						}
						this.dExec.stateReceived = true;
					}
					
//					if(curO instanceof MoveStates) {
//						continue;
//					}
					
//					if(curO instanceof RecomputeStates) {
//						continue;
//					}
					
				}
				/// Begin Brad M.'s protocol for passing storage tokens.
				else if(curO instanceof ReplicationStyleProtocol) {
					if(curO instanceof PassStorageToken) {

						String opNumber = ((PassStorageToken)curO).getOperatorID();
						Iterator j = stats.getAllKnownOperators();
                        
                        System.out.println("Now passing token for state " + ((PassStorageToken)curO).getStateNumber() + "for operator " + opNumber);
                         
                        
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getOperatorID() == Integer.parseInt(opNumber)) {
									String str = ((PassStorageToken)curO).getTargetMachineID();
									((ReplicatedStreamSplitOperator)temp).generateTokenPassingMessage(str);
									System.err.println("Token passing message requested");
								break;
							}
						}
					}
					else if(curO instanceof RemoveMachineFromOutput)
					{
						//TODO: Add code here to handle this particular thing.
						// We want to remove a target machine from an operator's output queue,
						//  assuming multiple machines in the queue.
						String opNumber = ((RemoveMachineFromOutput)curO).getOperatorID();
						Iterator j = stats.getAllKnownOperators();
						
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (Integer.toString(temp.getOperatorID()).equals(opNumber)) {
								//String machineName = ((RemoveMachineFromOutput)curO).getMachineName();
								
								Object connections = outConnections.getConnectionObject(temp);
								
								if (connections instanceof Hashtable) //  (if it's not a Hashtable, then we're not properly partitioned)
								{
									//note that the HashCode here is the IP address and port, separated by ':'

									//TODO: figure out some way to get the machine name.
									if (temp instanceof ReplicatedStreamSplitOperator) 
									{
										((ReplicatedStreamSplitOperatorImp)temp).terminateStream(((RemoveMachineFromOutput)curO).getHashCode(), (Hashtable)connections);
									}

                                    outConnections.addConnection(temp, (Hashtable)connections);

                                }
									
							}
						}
					}
					else if(curO instanceof AddMachineToOutput) 
					{
						String opNumber = ((AddMachineToOutput)curO).getOperatorID();
						Iterator j = stats.getAllKnownOperators();
						
                        System.out.println("Adding a machine to output of operator " + opNumber);
                        
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (Integer.toString(temp.getOperatorID()).equals(opNumber)) {
								//String machineName = ((RemoveMachineFromOutput)curO).getMachineName();

                                //Make sure that the tuple distributor no longer thinks that this queue
                                //  is local-only.
                                temp.getOutputQueue().setQueueName("SPREADFIRE");
                                
								Object connections = outConnections.getConnectionObject(temp);
                               
								if (connections == null || connections instanceof Hashtable) //  (if it's not a Hashtable, then we're not properly partitioned)
								{
                                    if (connections == null) connections = new Hashtable();
									//note that the HashCode here is the IP address and port, separated by ':'
									Socket s = new Socket (((AddMachineToOutput)curO).getTargetIP(), ((AddMachineToOutput)curO).getTargetPort());
									ObjectOutputStream oos = new ObjectOutputStream(s.getOutputStream());
									((Hashtable)connections).put(((AddMachineToOutput)curO).getHashCode(), oos);
                                    
                                    outConnections.addConnection(temp, (Hashtable)connections);

								}
                                else throw new Exception();
							}
						}
                        System.out.println("Finished adding machine to output");
					}
					else if(curO instanceof RemoveMachineFromInput)
					{
						String opNumber = ((RemoveMachineFromInput)curO).getOperatorID();
						Iterator j = stats.getAllKnownOperators();
						
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (Integer.toString(temp.getOperatorID()).equals(opNumber)) {
								//String machineName = ((RemoveMachineFromOutput)curO).getMachineName();
								
								Object connections = inConnections.getConnectionObject(temp);
								
                                if (connections instanceof Hashtable) //  (if it's not a Hashtable, then we're not properly partitioned)
								{
									//note that the HashCode here is the IP address and port, separated by ':'
									((Hashtable)connections).remove(((RemoveMachineFromOutput)curO).getHashCode());

                                    inConnections.addConnection(temp, (Hashtable)connections);
								}
									
							}
						}						
					}
					else if(curO instanceof AddMachineToInput)
					{
						String opNumber = ((AddMachineToInput)curO).getOperatorID();
						Iterator j = stats.getAllKnownOperators();
						
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (Integer.toString(temp.getOperatorID()).equals(opNumber)) {
								//String machineName = ((RemoveMachineFromOutput)curO).getMachineName();
								
								Object connections = inConnections.getConnectionObject(temp);
								
								if (connections == null || connections instanceof Hashtable) //  (if it's not a Hashtable, then we're not properly partitioned)
								{
                                    if (connections == null) connections = new Hashtable();
									//note that the HashCode here is the IP address and port, separated by ':'
									Socket s = new Socket (((AddMachineToInput)curO).getTargetIP(), ((AddMachineToInput)curO).getTargetPort());
									ObjectInputStream ois = new ObjectInputStream(s.getInputStream());
									((Hashtable)connections).put(((AddMachineToInput)curO).getHashCode(), ois);

                                    inConnections.addConnection(temp, (Hashtable)connections);
								}
									
							}
						}	
					}
					else if(curO instanceof SpreadfireEmergencyShutdown)
					{
						//IF we have to stop immediately due to too much memory usage,
						// let's go out gracefully and with all the results intact.
						this.dExec.terminateExecutionNow();
						System.out.println("Memory saturation on a node has become too great.  The system is shutting down.");
					}
				} else {

					// the original code that handles the connection messages -- Bin
					// Manage the connection and add it to the hashtable 
					// whether its incoming or outgoing					
					//*** STREAM TO MACHINE CONNECTION
					c = (Connection) cp.getObject();
					if (c instanceof ReceiveStream) {
						stream = c.getStreamName();
						operatorNumber = c.getToOperator();

						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
								operator = temp;
								break;
							}
						}
		
						OperatorQueue oq = new OperatorQueue(operator,c.getToQueueNumber());

						// Add the operator to the table if it bridges a network
						// connection. This will apply to Stream->Machine, 
						// Machine -> Application and Machine->Remote Machine
						if (operator != null) {
							inConnections.addConnection(stream, oq);
							System.out.println(inConnections);
							operator.setActive(true);
						}
					} //endif
					//*** LOCAL MACHINE CONNECTION
					else if (c instanceof Activate) {
						// nothing to do here but active the operator(s)
						// the queues are already connected
						
						operatorNumber = c.getFromOperator();
						System.out.println("Activate operator " + operatorNumber);

						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
								operator = temp;
								break;
							}
						}
						//Activate the operator
						//if (!operator.isActive()) {
						operator.setActive(true);
						operator.setNewlyActivated(true);
						// only in wpi-cluster, we need to set the join operator the clientName to differentiate the disk partitions
						if (operator instanceof SymmMwayHashJoinOperatorImp){
							((SymmMwayHashJoinOperatorImp)operator).setClientName(dExec.getClientName());
						}
	                    //}
					} //endif
					// *** Set partitionMaps and create dynamic queues for split operator
					// -- Bin Nov.30, 2004
					else if (c instanceof ActivatePartition) {
						operatorNumber = c.getFromOperator();
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
								operator = temp;
								break;
							}
						}
						//now we know that operator is StreamSplitOperatorImp, and c is ActivatePartition, cast them into appropriate classes
						ActivatePartition tmpc = (ActivatePartition)c;
						//(1) set the partitionMaps
						System.out.println("Received partition map for operator " + operator);
						System.out.println("The partition map is: \n" + tmpc.getPartitionMaps());
						((StreamSplitOperatorImp)operator).setPartitionMappingTable(tmpc.getPartitionMaps());
						
						//(2) create dynamic queues
						XATQueue curQue = operator.getOutputQueue();
						Machine[] allMachines = tmpc.getMachines();
						Machine curM = tmpc.getCurrentMachine();
						
						//(2.1) maybe we need to first remove the current output queue, then
						//(2.2) create dynamic queues correspondingly based on current queue type
						if (curQue instanceof XATMemoryQueueImp) {
							// we remove any existing queues, however, we need to reconnect the queue that has the same machine address
							// with its parent (join) operator
							((StreamSplitOperatorImp)operator).removeQutputQueues();
							for (int im=0; im<allMachines.length; im++){
								XATMemoryQueueImp nqueue = new XATMemoryQueueImp(operator);
								nqueue.setQueueNumber(curQue.getQueueNumber());
								//String nname = curQue.getQueueNumber() + ":" + curMachines[im].getIPAddress();
								String nname = allMachines[im].getIPAddress();
								nqueue.setQueueName(nname);
								((StreamSplitOperatorImp)operator).addOutputQueue(nqueue);
								if (nname.equals(curM.getIPAddress())) {
									// yes, need to set the queue as the local inputqueue of the parent node
									XATStreamOperator parentOp = (XATStreamOperator)((operator.getNode()).getParent()).getOperator();
									parentOp.replaceInputQueue(curQue,nqueue);
									System.out.println("Debug");
								}
							}
						} else if (curQue instanceof XATPersistentQueueImp) {
							// we remove any existing queues, however, we need to reconnect the queue that has the same machine address
							// with its parent (join) operator
							((StreamSplitOperatorImp)operator).removeQutputQueues();
							for (int im=0; im<allMachines.length; im++){
								XATPersistentQueueImp nqueue = new XATPersistentQueueImp(operator);
								nqueue.setQueueNumber(curQue.getQueueNumber());
								//String nname = curQue.getQueueNumber() + ":" + curMachines[im].getIPAddress();
								String nname = allMachines[im].getIPAddress();
								nqueue.setQueueName(nname);
								((StreamSplitOperatorImp)operator).addOutputQueue(nqueue);
								if (nname.equals(curM.getIPAddress())) {
									// yes, need to set the queue as the local inputqueue of the parent node
									XATStreamOperator parentOp = (XATStreamOperator)((operator.getNode()).getParent()).getOperator();
									parentOp.replaceInputQueue(curQue,nqueue);
								}							
							}						
						} else {
							System.out.println("Unknow queue type?");
						}
					} //end if
					// *** SEND DATA
					// *** -- Bin @ Dec.1,2004, to set the outConnections of the partitioned operator
					else if (c instanceof SendData) {
						//First get the operator to connect to the application
						operatorNumber = c.getFromOperator();
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
								operator = temp;
								break;
							}
						}

						if ( ((SendData)c).getPartitionedFlag() ) {
							// yes, handle the outConnections of the partitioned operator
							// create all output streams except the one in the same ip address with the current machine
							// all these output streams are kept in a hashtable keyed on ip address
							Hashtable streams = new Hashtable();
							Machine[] machines = ((SendData)c).getAllMachines();
							for (int i=0; i<machines.length; i++){
								Machine curM = machines[i];
								if ((curM.getIPAddress()).equals(c.getToIPAddress())
								        && curM.getPort() == (c.getToPort())){
									// we do not need to create the output stream if it is in the same machine
								} else {
									Socket s = new Socket(curM.getIPAddress(),curM.getTupleReceiverPort());
									ObjectOutputStream outputStream = new ObjectOutputStream(s.getOutputStream());
									streams.put(curM.getIPAddress(),outputStream);
									System.out.println("SendData from op " + operatorNumber + " to machine " + curM);
								}
							} //end for machines.length
							//then, add all these streams to the outConnections
							outConnections.addConnection(operator, streams);						
						} else {
							//in case the operator is not a partitioned operator
							// Now create the socket
							Socket s = new Socket(c.getToIPAddress(), c.getToPort());
							ObjectOutputStream outputStream = new ObjectOutputStream(s.getOutputStream());

							outConnections.addConnection(operator, outputStream);	

						} //end if getPartitionedFlag
						//System.out.println("Out-Connections:" + outConnections.toString());
						//System.out.println("In-Connections:" + inConnections.toString());
					} //endif
					//*** RECEIVING FROM MACHINE CONNECTION
					else if (c instanceof ReceiveData) {
						
						String operatorFrom = c.getFromOperator();
						String operatorTo = c.getToOperator();
						
						System.out.println("ReceiveData from op " + operatorFrom + " to op " + operatorTo);
						
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorTo)) {
								operator = temp;
								break;
							}
						}

						OperatorQueue oq = new OperatorQueue(operator,c.getToQueueNumber());
						// Add the operator to the table if it bridges a network
						// connection. This will apply to Stream->Machine, 
						// Machine -> Application and Machine->Remote Machine
						if (operator != null) {
							inConnections.addConnection(operatorFrom, oq);
						}
						//System.out.println("Out-Connections:" + outConnections.toString());
						//System.out.println("In-Connections:" + inConnections.toString());
					} //endif
					else if(c instanceof LightWeightStats){
						LightWeightStats lws = (LightWeightStats) c;
						dExec.setLightweightStats(lws.getStats());
					}//endif
					else if (c instanceof DeActivate) {
						// nothing to do here but active the operator(s)
						// the queues are already connected

						operatorNumber = c.getFromOperator();

						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
								operator = temp;
								break;
							}
						}
						//De-activate the operator
						//if (operator.isActive()) {
						operator.setActive(false);
						//}						
						//operator.setNewlyDeactivated(true);
						
					} //endif
					else if (c instanceof StopSend) {

						//First get the operator to connect to the application
						operatorNumber = c.getFromOperator();

						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
								operator = temp;
								break;
							}
						}

						if(operator != null){
							outConnections.removeConnection(operator);
							System.out.println(outConnections);
						}

					} //endif
					else if (c instanceof SendState) {
						OperatorState[] opstate;
						
						//First get the operator to connect to the application
						operatorNumber = c.getFromOperator();

						// Now, where to send it. 
						String ipAddress = c.getToIPAddress();
						int port = c.getToPort();
						
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
								operator = temp;
								break;
							}
						}

						if(operator.hasState()){
							opstate = operator.getState();
							// toOperator and fromOperator has the same Object_ID in this situation
							Connection d = new edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.State(operatorNumber, opstate);
							Packet p = new Packet("state",d);
							Socket s = new Socket(ipAddress,port);
							ObjectOutputStream os = new ObjectOutputStream(s.getOutputStream());
							os.writeObject(p);
							os.close();
							s.close();
						
						}
					} //endif
					else if (c instanceof edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.State) {

						//First get the operator to connect to the application
						operatorNumber = c.getToOperator();
						// @ Java 1.5 has its own State class, need to specify the whole path to differ Connection State -- Bin May, 2005
						edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.State tempState = (edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.State) c;
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getProperty("OBJECT_ID").equals(operatorNumber)) {
								operator = temp;
								break;
							}
						}

						if(operator != null){
							operator.setState(tempState.getState());
						}

					} //endif
					else if (c instanceof UpdatePartitionStatistics){
						//to update the cost statistics across machines -- Bin @ Apr 12, 2005//
						int operatorID = ((UpdatePartitionStatistics)c).getOperatorID();
						boolean finalOutput = ((UpdatePartitionStatistics)c).isFinalOutput();
						double inter_rate = ((UpdatePartitionStatistics)c).getInterRate();
						
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if (temp.getOperatorID() == operatorID) {
								operator = temp;
								break;
							}
						}
						if (operator instanceof SymmMwayHashJoinOperatorImp){
							ArrayList ids = ((UpdatePartitionStatistics)c).getIDsToBeUpdated();
							Hashtable values = ((UpdatePartitionStatistics)c).getStatsValues();
							Hashtable hashStates = ((SymmMwayHashJoinOperatorImp)operator).getHashJoinStates();
							for (int kk=0; kk<ids.size(); kk++){
								Integer partID = (Integer)ids.get(kk);
								Integer outN = (Integer)values.get(partID);
								if (hashStates.get(partID) == null){
									//it is possible that this partition is pushed into disks//
								} else {
									if (finalOutput) { //update the cost of outputs
										((HashStatePartitionGroup)hashStates.get(partID)).addTotalTuplesOutputed(outN.intValue());	
									} else { // update the cost of intermediate results
										((HashStatePartitionGroup)hashStates.get(partID)).addTotalProjectedTupleNumber(outN.intValue(),inter_rate);
									}										
								}
							}							
						} else {
							System.out.println("Must be something wrong. No Symm Join Operator is required to propaged statistics?");
						}
					} //endif
					else if (c instanceof PropagatePartitionMaps){ 
						// to propagate the partitionMaps in the splitopeators that have not activated in this processor -- Bin Apr 17, 2005
						Hashtable pMaps = ((PropagatePartitionMaps)c).getPartitionMaps();
						Iterator j = stats.getAllKnownOperators();
						for (; j.hasNext();) {
							XATStreamOperator temp = ((XATStreamOperator) j.next());
							if ((temp instanceof StreamSplitOperatorImp) && (!temp.isActive())){
								//get the partitonMap from c
								Enumeration pmKeys = pMaps.keys();
								while (pmKeys.hasMoreElements()){
									StreamSplitOperatorImp key = (StreamSplitOperatorImp)pmKeys.nextElement();
									if (key.getOperatorID() == temp.getOperatorID()) {
										((StreamSplitOperatorImp)temp).setPartitionMappingTable((Hashtable)pMaps.get(key));
									}
								}
							}
						}
					} //end if
					else if (c instanceof ConnectionSeperator){
						
					}
					else{
						System.out.println("The Connection: " + c + " is not supported by the Connection Listener.");
					}
					
				}//end if instanceof Connection
			} //end of while enable
		} catch (ConnectException ce) {
			System.out.println("==========ConnectException=========");
			ce.printStackTrace();			
			System.out.println("===================================");
		} 
		catch (EOFException eofe) {
			//
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void requestStop() {
		keepGoing = false;
	}
}
