/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection;

import java.io.ObjectOutputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.io.Serializable;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 12, 2003
 * 
 * The connection table class is similar to a hash table but stores
 * specific information about the types of connections
 * in Raindrop There are two types of connections that are monitored:
 * 
 * 1. Incoming
 * 2. Outgoing
 * 
 * Incoming connections know the Operator ID from which a tuple came, and
 * the Operator for which that tuple is supposed to go.
 * 
 * Outgoing connections know all of the operators that are supposed
 * to output across the network and the socket connection to physically send
 * the tuples.
 * 
 */
public class ConnectionTable implements Serializable {

	private Hashtable table;
	/**
	 * Default Constructor
	 */
	public ConnectionTable() {
		this.table = new Hashtable();
	}
	
	//following two are added by song wang to support multipe queue output destination of one operator
	//for inConnection, the operatorID+queueNum will be used as the String key
	//i.e. it will use the original addConnection(String tuplename,OperatorQueue oq)
	//for outConnection, the original used addConnection(XATStreamOperator operator, ObjectOutputStream os)
	//will be replaced by the following addConnection(OperatorQueue oq, ObjectOutputStream os)
	//We use the 
	//
	//NOTE that multiple entries of the OperatorQueue can share one physical ObjectOutputStream,
	//in case they have the same destination. This is not done for now. Now no sharing is done 
	
	public synchronized void addConnection(OperatorQueue oq, ObjectOutputStream os){
		table.put(oq,os);
	}
	public synchronized Object getConnectionObject(OperatorQueue oq){
		return table.get(oq);
	}
	
	
	// add new addConnection and getConnection functions to hadle partitioned operator
	public synchronized void addConnection(XATStreamOperator operator, Hashtable ht){
		table.put(operator,ht);
	}
	//since we do not know whether the connections of an operator is a outputstream or a set of outputstreams 
	public synchronized Object getConnectionObject(XATStreamOperator operator){
		return table.get(operator);
	}
	
	// These are all simple get and set methods.
	public synchronized void addConnection(String tuplename,OperatorQueue oq){
		table.put(tuplename,oq);
	}
	
	public synchronized void addConnection(XATStreamOperator operator, ObjectOutputStream os){
		table.put(operator,os);
	}
	
	public synchronized int connectionCount(){
		return this.table.size();
	}
	
	public synchronized OperatorQueue getConnection(String tuplename){
		return (OperatorQueue) table.get(tuplename);
	}
	
	public synchronized ObjectOutputStream getConnection(XATStreamOperator operator){
		return (ObjectOutputStream) table.get(operator);
	}
	
	public synchronized Enumeration getKeys(){
		return this.table.keys();
	}
	
	public synchronized Enumeration getValues(){
		return this.table.elements();
	}
	
	public synchronized void removeConnection(String tuplename){
		this.table.remove(tuplename);
	}
	
	public synchronized void removeConnection(XATStreamOperator operator){
		this.table.remove(operator);
	}
	
	public synchronized void clearTable(){
		this.table.clear();
	}
	
	public String toString(){
		String ret = new String();
		
		Enumeration e= table.keys();
		
		ret+="    Connection Table     \n";
		ret+="-------------------------\n";
		
		while(e.hasMoreElements()){
			Object o = e.nextElement();
			ret+= o +  "=>" + table.get(o) +"\n";
		}
		
		return ret;
	}
}