/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection;

import java.io.ObjectOutputStream;
import java.net.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Mar 30, 2004
 * 
 * This class will give repeated connection requests to the Client to which it is connected.
 * This class will server to be a test for how many connection requests a client can handle
 * without significantly reducing the performance of the system.
 * 
 */
public class ConnectionTester 
{
	private static String USAGE = "Usage: java ConnectionTester <ip address> <port> <messages per second>";
	
	public static void main(String[] args) 
	{
		try 
		{
			if(args.length != 3)
			{
				System.out.println(USAGE);
				return;				
			}	
			
			String ipAddress = args[0];
			int port = Integer.parseInt(args[1]);
			int messagesPerSecond = Integer.parseInt(args[2]);
			System.out.println("Connecting to: " + ipAddress + ":" + port);
			System.out.println("Sending " + messagesPerSecond + " messages per second.");
			Socket s = new Socket(ipAddress, port);
			ObjectOutputStream out = new ObjectOutputStream(s.getOutputStream());
			while(true)
			{				
				for(int i = 0; i < messagesPerSecond; i++)
				{
					Connection c = new Activate("1");
					Packet p = new Packet("tester",c);
					out.writeObject(p);
				}
				
				Thread.sleep(1000);				
			}
			
		} 
		catch (NumberFormatException e) 
		{
			System.out.println(USAGE);
			System.exit(1);
		} 
		catch (UnknownHostException uhe)
		{
			System.out.println(uhe.getMessage());
		} 
		catch (Exception e)
		{
			e.printStackTrace();
		}			
	}
}
