/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////



package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection;

import java.io.Serializable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.*;
/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Mar 2, 2004
 * 
 * Insert comments here.
 */
public class SendData implements Serializable, Connection {

	private String operatorID;
	private String ipAddress;
	private int port;
	private int queueNum;
	
	// -- Bin Nov.30, 2004
	private boolean partitioned = false; // to indicate whether a partitioned operator
	private Machine[] machines = null;

	public SendData(String operatorID, int queueNum, Machine[] m, Machine curm) {
		this.operatorID = operatorID;
		this.queueNum = queueNum;
		this.port = curm.getTupleReceiverPort();
		this.ipAddress = curm.getIPAddress();
		partitioned = true;
		machines = m;
	}

	public boolean getPartitionedFlag(){
		return partitioned;
	}
	
	public Machine[] getAllMachines(){
		return machines;
	}
	
	// the followings are original methods
	public SendData(String operatorID, int queueNum, Machine m) {
		this.operatorID = operatorID;
		this.queueNum = queueNum;
		this.port = m.getTupleReceiverPort();
		this.ipAddress = m.getIPAddress();
		
	}
	
	public SendData(String operatorID, int queueNum, String ipAddress, int port) {
		this.operatorID = operatorID;
		this.queueNum = queueNum;
		this.port = port;
		this.ipAddress =ipAddress;		
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getClassName()
	 */
	public String getClassName() {
		return this.getClass().getName();
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getFromIPAddress()
	 */
	public String getFromIPAddress() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getFromOperator()
	 */
	public String getFromOperator() throws ConnectionException {
		return this.operatorID;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getFromPort()
	 */
	public int getFromPort() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getFromQueueNumber()
	 */
	public int getFromQueueNumber() throws ConnectionException {
		return this.queueNum;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getStreamLength()
	 */
	public int getStreamLength() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getStreamName()
	 */
	public String getStreamName() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToIPAddress()
	 */
	public String getToIPAddress() throws ConnectionException {
		return this.ipAddress;			
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToOperator()
	 */
	public String getToOperator() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToPort()
	 */
	public int getToPort() throws ConnectionException {
		return this.port;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToQueueNumber()
	 */
	public int getToQueueNumber() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	public String toString(){
		return "Sending data to: " + this.ipAddress +":" + this.port;
	}
}
