/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection;

import java.io.Serializable;
import java.util.Hashtable;
import java.util.ArrayList;


/**
 * @author Bin Liu
 * When handling query tree with multiple SymmMwayHashJoins, we need to update the cost statistics
 * (the number of output that this join operator contributed), these partitions could be in different 
 * machines, we thus require the DistributedExecutionController to spread these information
 */
public class UpdatePartitionStatistics implements Serializable, Connection{

	int operatorID = -1;
	ArrayList idsToBeUpdated = null;
	Hashtable statsValues = null;
	boolean finalOutput = false;
	double inter_rate = 1.0;
	
	public UpdatePartitionStatistics(int opID, ArrayList ids, Hashtable values,boolean finalOut,double i_rate){
		operatorID = opID;
		idsToBeUpdated = ids;
		statsValues = values;
		finalOutput = finalOut;
		inter_rate = i_rate;
	}
	public double getInterRate(){
		return inter_rate;
	}
	public boolean isFinalOutput(){
		return finalOutput;
	}
	
	public int getOperatorID(){
		return operatorID;
	}
	public ArrayList getIDsToBeUpdated(){
		return idsToBeUpdated;
	}
	public Hashtable getStatsValues(){
		return statsValues;
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getClassName()
	 */
	public String getClassName() {
		return this.getClass().getName();
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getFromIPAddress()
	 */
	public String getFromIPAddress() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getFromOperator()
	 */
	public String getFromOperator() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getFromPort()
	 */
	public int getFromPort() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getFromQueueNumber()
	 */
	public int getFromQueueNumber() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getStreamLength()
	 */
	public int getStreamLength() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getStreamName()
	 */
	public String getStreamName() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToIPAddress()
	 */
	public String getToIPAddress() throws ConnectionException {
		throw new ConnectionException(this.getClassName());		
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToOperator()
	 */
	public String getToOperator() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToPort()
	 */
	public int getToPort() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection#getToQueueNumber()
	 */
	public int getToQueueNumber() throws ConnectionException {
		throw new ConnectionException(this.getClassName());
	}

}
