/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel;

import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Mar 16, 2004
 * 
 * Insert comments here.
 */
public class AverageTupleExecutionTime extends CostModel {

	/**
	 * Default Constructor
	 * @param st -The statistics table from which to calculate the cost from
	 */
	public AverageTupleExecutionTime(StatisticsTable st){
		this.st = st;
		this.requiredStats = new ArrayList();
		setStatistics();
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.CostModel#requiredStatistics()
	 */
	public Iterator requiredStatistics() {
		return this.requiredStats.iterator();	
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.CostModel#setStatistics()
	 */
	protected void setStatistics() {
		super.setStatistics();
		this.requiredStats.add("AVERAGE_TUPLE_PROCESSING_TIME");
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.CostModel#calculateCost()
	 */
	public CostTable calculateCost(StatisticsTable st, DistributionTable dt) {
		this.dt = dt;
		this.st = st;
		CostTable ct = new CostTable(st.size());
		
		try{
		
		for(Enumeration e = st.keys();e.hasMoreElements();){
			Machine m = (Machine) e.nextElement();
			LightWeightStatistic lws = st.get(m);
			double time = 0;
			
			//First, calculate the raw average load on each machine.
			for(Enumeration en = lws.getKeys(); en.hasMoreElements();){
				LightWeightQueryObject xqo = (LightWeightQueryObject) en.nextElement();
				if(xqo.isActive()){
					time+= Double.parseDouble(lws.get(xqo,"AVERAGE_TUPLE_PROCESSING_TIME"));
					ct.putRawCost(m,time);
				}
				
			}
			
			//Second, calculate the normalized load.
			ct = this.linearLormalize(ct,MAXIMIZE);
			
			//	Finally calculate the shared average load for each operator on the machine.

			ArrayList nodes = dt.getOperatorsOnMachine(m);
			
			for(Iterator i = nodes.iterator(); i.hasNext();){
				XATNode node = (XATNode) i.next();
				LightWeightQueryObject xqo = lws.getQueryObject(node);
				
				double thisTime = Double.parseDouble(lws.get(xqo,"AVERAGE_TUPLE_PROCESSING_TIME"));
				double cost;
				if(time > 0) {
					cost = thisTime/time;
				} else {
					cost = 0;
				}

				NodeCost nc = new NodeCost(node,cost);
				ct.put(m,nc);
			}
		}
		
		}catch (Exception e){
			e.printStackTrace();
		}
		
		return ct;
	}

}
