/*
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
 * Created on Mar 13, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel;


import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.*;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * The abstract class cost model represents the cost model calculation for a machine. 
 * Given certain statistics about a machine, a concrete CostModel class will
 * calculate the "cost" or fullness of the server based on the particular model of the class.
 * 
 */
public abstract class CostModel {

	public final int MAXIMIZE = 1;
	public final int MINIMIZE = 0;
	
	protected ArrayList requiredStats;
	protected StatisticsTable st;
	protected DistributionTable dt;
	
	/**
	 * Returns an iterator of statistics that are required for the cost model.
	 * @return Iterator -statistics required by the cost model.
	 * 
	 */
	public abstract Iterator requiredStatistics();
	public abstract CostTable calculateCost(StatisticsTable st, DistributionTable dt);
	
	/** All machines need to collect memory usage.
	 *  this can be used to detect system-wide overflow and apply cross-machine migration 
	 *  or global decided data spill as necessary. */
	protected void setStatistics(){
		this.requiredStats.add("USED_MEMORY");
		this.requiredStats.add("MAX_MEMORY");
	}
	
	public CostTable linearLormalize(CostTable ct, int quantifier){
		CostTable c = ct;
		boolean maximize = false;
		double maxCost = 0, minCost = 0;
		
		if(quantifier == MAXIMIZE)
			maximize = true;
			
		if(maximize){
			maxCost = c.maxRawCost();
			double normalizedCost;
			Enumeration e = ct.getMachines();
			while(e.hasMoreElements()){
				Machine m = (Machine) e.nextElement();
				double rawCost = c.getRawCost(m);
				if(maxCost == 0){
					normalizedCost = 0;
				}
				else {
					normalizedCost = rawCost/maxCost;
				}
				
				c.put(m,normalizedCost);
			}
		}
		else{
		
			minCost = c.minRawCost();
		}
			
			
		
		
		
		return c;
		
	}
}
