/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel;


import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.raindrop.utils.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.*;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Mar 16, 2004
 * 
 * An extension of the hashtable to hold each machines "cost" as 
 * determined by the cost model.
 * 
 * Key: Machine
 * Value: Calculated cost by the concrete CostModel
 * 
 */
public class CostTable {

	private Hashtable machineCost;
	private Hashtable nodeCost;
	private Hashtable rawCost;
	
	/**
	 *  Constructs a new Cost table
	 */
	public CostTable(Machine[] m) {
		machineCost = new Hashtable(m.length);
		nodeCost = new Hashtable(m.length);
		rawCost = new Hashtable(m.length);
		
		for(int i = 0; i < m.length; i++){
			machineCost.put(m[i],new Double(0));
			rawCost.put(m[i],new Double(0));
			nodeCost.put(m[i],new ArrayList(5));
		}
	}
	
	public CostTable(int size) {
		machineCost = new Hashtable(size);
		nodeCost = new Hashtable(size);
		rawCost = new Hashtable(size);
	}
	
	public Enumeration getMachines(){
		if(rawCost.size() > 0)
			return rawCost.keys();
		else
			return machineCost.keys();
	}
	
	public void put(Machine m, double cost){
		machineCost.put(m,new Double(cost));
	}
	
	public void put(Machine m, NodeCost nc){
		
		ArrayList queryObjects;
		
		if((queryObjects = this.getNodeCost(m)) == null)
			queryObjects = new ArrayList();
			
		queryObjects.add(nc);
		nodeCost.put(m,queryObjects);
	}
	
	public void put(Machine m, NodeCost nc, int index){
		
		ArrayList queryObjects;
		
		if((queryObjects = this.getNodeCost(m)) == null) {
			queryObjects = new ArrayList();
			queryObjects.add(nc);
			return;
		}
			
		queryObjects.add(index, nc);
		nodeCost.put(m,queryObjects);
	}
	
	public void putRawCost(Machine m, double cost){
		rawCost.put(m, new Double(cost));
	}
	
	public double getMachineCost(Machine m){
		if(machineCost == null)
			return 0;
		
		if(machineCost.get(m) == null)
			return 0;
		
		Double d = (Double) machineCost.get(m);
		return d.doubleValue();
	}
	
	public double getRawCost(Machine m){
		Double d = (Double) rawCost.get(m);
		return d.doubleValue();
	}
	
	public ArrayList getNodeCost(Machine m){
		ArrayList a  = (ArrayList) nodeCost.get(m);
		return a;
	}
	
	public String toString(){
		int maxLength = 10;
		
		String ret = new String();
		ret+="-----------------------------------------------------\n";
		ret+="|                    Cost Table                     |\n";
		ret+="-----------------------------------------------------\n";
		ret+="| Machine    | Norm. Cost | Cost       | Nodes      |\n";
		ret+="-----------------------------------------------------\n";
		for(Enumeration e = machineCost.keys(); e.hasMoreElements();){
			
			Machine m = (Machine) e.nextElement();
			double d =  this.getMachineCost(m);
			double d2 = this.getRawCost(m);
			ArrayList a = this.getNodeCost(m);
			String machineName = m.getMachineName();
			String doubleCost = Double.toString(d);
			String rawCost = Double.toString(d2);
			
			if(machineName.length() > maxLength)
				machineName = machineName.substring(0,maxLength);
			if(machineName.length() < maxLength){
				machineName = machineName.concat(BlankString.createWhiteSpace(maxLength-machineName.length()));
			}
			
			if(doubleCost.length() > maxLength)
				doubleCost = doubleCost.substring(0,maxLength);	
			if(doubleCost.length() < maxLength){
				doubleCost = doubleCost.concat(BlankString.createWhiteSpace(maxLength-doubleCost.length()));
			}
			
			if(rawCost.length() > maxLength)
				rawCost = rawCost.substring(0,maxLength);	
			if(rawCost.length() < maxLength){
				rawCost = rawCost.concat(BlankString.createWhiteSpace(maxLength-rawCost.length()));
			}
				
			ret += "| " + machineName + " | " + doubleCost + " | " + rawCost + " | ";
			
			String nodeCost=new String("");
			
			//System.out.println("a.size = " + a.size());
			
			if(a != null && a.size() > 0) {
				Iterator i = a.iterator();
				NodeCost nc = (NodeCost) i.next();
				nodeCost = nc.toString();
			
				if(nodeCost.length() > maxLength)
					nodeCost = nodeCost.substring(0,maxLength);
				if(nodeCost.length() < maxLength){
					nodeCost = nodeCost.concat(BlankString.createWhiteSpace(maxLength-nodeCost.length()));
				}
			
			ret+= nodeCost + " |\n";
			while(i.hasNext()){
				nc = (NodeCost) i.next();
				nodeCost = nc.toString();
			
				if(nodeCost.length() > maxLength)
					nodeCost = nodeCost.substring(0,maxLength);
				if(nodeCost.length() < maxLength){
					nodeCost = nodeCost.concat(BlankString.createWhiteSpace(maxLength-nodeCost.length()));
				}
								
				ret+= "|                                      | " + nodeCost + " |\n";
			}
			}
			ret+="-----------------------------------------------------\n";	
		}
		
		return ret;
	}
	
	public double maxRawCost(){
		double max = Double.MIN_VALUE;
		Collection c = rawCost.values();
		Iterator i = c.iterator();
		
		while(i.hasNext()){
			double temp = ((Double)i.next()).doubleValue();
			if(temp > max){
				max = temp;
			}
		}
		return max;
	}
	
	public double maxCost(){
		//double max = Double.MIN_VALUE;
		double max = 0;
		Collection c = machineCost.values();
		Iterator i = c.iterator();
		
		while(i.hasNext()){
			double temp = ((Double)i.next()).doubleValue();
			if(temp > max){
				max = temp;
			}
		}
		return max;
	}
	
	public Machine maxMachine(){
		Machine m = null;
		//double max = Double.MIN_VALUE;
		double max = 0;
		Enumeration e = this.machineCost.keys();
	
		while(e.hasMoreElements()){
			Machine n = (Machine) e.nextElement();
			double temp = this.getMachineCost(n);
			if(temp > max){
				max = temp;
				m = n;
			}
		}
		return m;
	}
	
	public Machine NthMaxMachine(int nth){
		Machine m = null;
		ArrayList usedUp = new ArrayList();
		double max = Double.MAX_VALUE;
		//double nextLowest = Double.MIN_VALUE;
		double nextLowest = 0.0;
		
		for(int i = nth; i > 0;i--){
			Enumeration en = this.machineCost.keys();
			while(en.hasMoreElements()){
				Machine n = (Machine) en.nextElement();
				double temp = this.getMachineCost(n);
				if(temp <= max && temp > nextLowest && !usedUp.contains(n)){
					nextLowest = temp;
					m = n;
				}
			}
			
			usedUp.add(m);
			max = nextLowest;
			nextLowest = Double.MIN_VALUE;
		}
		
		return m;
	}
	public double minRawCost(){
		double min = Double.MAX_VALUE;
		Collection c = rawCost.values();
		Iterator i = c.iterator();
		
		while(i.hasNext()){
			double temp = ((Double)i.next()).doubleValue();
			if(temp < min){
				min = temp;
			}
		}
		return min;
	}
	// to return the cost next to the minimal
	public double minRawCostExcept(double min){
		double Smin = Double.MAX_VALUE;
		Collection c = rawCost.values();
		Iterator i = c.iterator();
		
		while(i.hasNext()){
			double temp = ((Double)i.next()).doubleValue();
			if ((temp < Smin) && (temp !=min)){
				Smin = temp;
			}
		}
		return Smin;
	}
	
	public double minCostExcept(double min){
		double Smin = Double.MAX_VALUE;
		Collection c = machineCost.values();
		Iterator i = c.iterator();
		
		while(i.hasNext()){
			double temp = ((Double)i.next()).doubleValue();
			if((temp < Smin) && (temp!=min)){
				Smin = temp;
			}
		}
		return Smin;
	}

	public double minCost(){
		double min = Double.MAX_VALUE;
		Collection c = machineCost.values();
		Iterator i = c.iterator();
		
		while(i.hasNext()){
			double temp = ((Double)i.next()).doubleValue();
			if(temp < min){
				min = temp;
			}
		}
		return min;
	}
	
	public Machine minMachine(){
		Machine m = null;
		double min = Double.MAX_VALUE;
		Enumeration e = this.machineCost.keys();
	
		while(e.hasMoreElements()){
			Machine n = (Machine) e.nextElement();
			double temp = this.getMachineCost(n);
			if(temp < min){
				min = temp;
				m = n;
			}
		}
		return m;
	}
	
	public int size(){
		return this.machineCost.size();
	}
	
	/*
	 * Removes a particular XATNode from the machine in the table
	 */
	public void remove(Machine m, XATNode node){
		ArrayList queryObjects;
//		ArrayList finalObjects = new ArrayList();
		
		if((queryObjects = this.getNodeCost(m)) == null)
			return;
			
//		Iterator i = queryObjects.iterator();
//		while(i.hasNext()){
//			NodeCost nc = (NodeCost) i.next();
//			if(!node.equals(nc.getNode()))
//				finalObjects.add(nc);
//		}

		for(int i=0; i<queryObjects.size(); i++) {
			NodeCost nc = (NodeCost) queryObjects.get(i);			
			if(node.equals(nc.getNode())) {
				queryObjects.remove(i);
//				System.out.println("Remove node "+nc);
				i--;
			}
		}
		
		//nodeCost.put(m,finalObjects);
	}
	
	/**
	 * Converts the cost table into a distribution table.
	 * @return
	 */
	public DistributionTable convertToDistributionTable(){
		DistributionTable dt = new DistributionTable();
		
		Enumeration e = this.getMachines();
		while(e.hasMoreElements()){
			Machine m = (Machine) e.nextElement();
			Iterator i = this.getNodeCost(m).iterator();
			while(i.hasNext()){
				NodeCost nc = (NodeCost) i.next();
				XATNode node = nc.getNode();
				dt.add(node,m);
			}
		}
		return dt;
	}
}
