/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel;

import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
/**
 * @author Yali Zhu
 * @author Yaliz@cs.wpi.edu
 * @version 1.0
 * 
 * Aug 20, 2005
 * 
 * Calculate the percentage of used resource on each machine.
 * In the first version, this resource usage is represented by percentage of memory usage on each machine.
 * This percentage is calculated as (used memory)/(max memory). 
 */
public class ResourceUsedPct extends CostModel {

	/**
	 * Default Constructor
	 * @param st -The statistics table from which to calculate the cost from
	 */
	public ResourceUsedPct(StatisticsTable st){
		this.st = st;
		this.requiredStats = new ArrayList();
		setStatistics();
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.CostModel#requiredStatistics()
	 */
	public Iterator requiredStatistics() {
		return this.requiredStats.iterator();	
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.CostModel#setStatistics()
	 */
	protected void setStatistics() {
		super.setStatistics();
		this.requiredStats.add("USED_MEMORY_PERCENTAGE");
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.CostModel#calculateCost()
	 */
	public CostTable calculateCost(StatisticsTable st, DistributionTable dt) {
		this.dt = dt;
		this.st = st;
		CostTable ct = new CostTable(st.size());
		
		for(Enumeration e = st.keys();e.hasMoreElements();){
			Machine m = (Machine) e.nextElement();
			LightWeightStatistic lws = st.get(m);
			double mem = 0;
			
			//First, calculate the raw load on each machine.
			for(Enumeration en = lws.getKeys(); en.hasMoreElements();){
				LightWeightQueryObject xqo = (LightWeightQueryObject) en.nextElement();
				mem = Double.parseDouble(lws.get(xqo,"USED_MEMORY_PERCENTAGE"));
				ct.putRawCost(m,mem);
				
			}
			
			/**
			//Second, calculate the normalized load.
			ct = this.linearLormalize(ct,MAXIMIZE);
			
			//	Finally calculate the shared load for each operator on the machine.

			ArrayList nodes = dt.getOperatorsOnMachine(m);
			int count = nodes.size();
			
			double avgCost = 0;
			if(count > 0)
				avgCost = ct.getMachineCost(m)/count;
				
				System.out.println(avgCost + " = " + ct.getMachineCost(m) +"/"+count + "(" +m.getMachineName()+")");
				
			for(Iterator i = nodes.iterator(); i.hasNext();){
				XATNode node = (XATNode) i.next();
				NodeCost nc = new NodeCost(node,avgCost);
				ct.put(m,nc);
			} */
		}
		
		return ct;
	}

}
