/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel;

import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.*;
/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Mar 16, 2004
 * 
 * Represents a table for Lightweight statistics
 * 
 * Key: Machine
 * Value: LightWeightStatistics for the machine
 * 
 */
public class StatisticsTable {

	private Hashtable table;
	
	/**
	 *  Constructs a new Cost table
	 */
	public StatisticsTable(Machine[] m) {
		table = new Hashtable(m.length);
		
		for(int i = 0; i < m.length; i++){
			table.put(m[i],new LightWeightStatistic());
		}
	}
	
	public synchronized void put(Machine m, LightWeightStatistic lws){
		table.put(m,lws);
	}
	
	public synchronized LightWeightStatistic get(Machine m){
		return (LightWeightStatistic) table.get(m);
	}
	
	public String toString(){
		String ret = new String();
		ret+= "\n           Statistic Table            \n";
		ret+= "--------------------------------------\n";
		for(Enumeration e = table.keys(); e.hasMoreElements();){
			
			Machine m = (Machine) e.nextElement();
			LightWeightStatistic l = (LightWeightStatistic) table.get(m);
			ret += m.getMachineName() + " | " + l.toString() +"\n";
			
		}
		return ret;
	}
	
	public int size(){
		return table.size();
	}
	
	public Enumeration keys(){
		return this.table.keys();
	}
}
