/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;

//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;

/**
 * @author Bin Liu
 * Based on the total available memory and used memory to decide how much percentage should be
 * moved from one machine to another
 */
public class TotalMemoryUsed extends CostModel {

	/**
	 * Default Constructor
	 * @param st -The statistics table from which to calculate the cost from
	 */
	public TotalMemoryUsed(StatisticsTable st){
		this.st = st;
		this.requiredStats = new ArrayList();
		setStatistics();
	}

	public Iterator requiredStatistics() {
		return this.requiredStats.iterator();
	}

	protected void setStatistics() {
		super.setStatistics();
		
		//these two stats are already added in super class CostModel
		this.requiredStats.add("USED_MEMORY");
		this.requiredStats.add("USED_MEMORY_PERCENTAGE");
		this.requiredStats.add("MAX_MEMORY");		
		//this.requiredStats.add("NUMBER_OF_TUPLES_OUTPUTTED_TOTAL");		
	}
	/*
	 * Here, simply return the rawcost of each machine in terms of memory usage
	 */
	public CostTable calculateCost(StatisticsTable st, DistributionTable dt) {
		this.dt = dt;
		this.st = st;
		CostTable ct = new CostTable(st.size());
		
		for(Enumeration e = st.keys();e.hasMoreElements();){
			Machine m = (Machine) e.nextElement();
			LightWeightStatistic lws = st.get(m);
			double mem = 0;
			//double maxmem = 0;
			
			String memS = lws.get("USED_MEMORY");
			if (memS !=null) {
				mem = Double.parseDouble(memS);
			}
			ct.putRawCost(m,mem);
			
			System.out.println("Machine:" + m.getMachineName() + " UsedMem:" + mem);
			
			memS = lws.get("USED_MEMORY_PERCENTAGE");
			if (memS !=null) {
				mem = Double.parseDouble(memS);
			}
			ct.put(m, mem);
			System.out.println("Machine:" + m.getMachineName() + " UsedMemPct:" + mem);
			
			memS = lws.get("MAX_MEMORY");
			if (memS !=null) {
				mem = Double.parseDouble(memS);
			}
			System.out.println("Machine:" + m.getMachineName() + " MaxMem:" + mem);
			
			/*
			//First, calculate the raw load on each machine.
			for(Enumeration en = lws.getKeys(); en.hasMoreElements();){
				LightWeightQueryObject xqo = (LightWeightQueryObject) en.nextElement();
				mem = Double.parseDouble(lws.get(xqo,"USED_MEMORY"));
				//maxmem = Double.parseDouble(lws.get(xqo,"MAX_MEMORY"));
				ct.putRawCost(m,mem);
				//ct.put(m,maxmem);
			}
			*/
		}
		return ct;
	}

}
