/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;

/**
 * @author Yali Zhu
 * The total number of tuples in input queues and in operator states. 
 */
public class TotalNumberOfTuples extends CostModel {	
	
	public TotalNumberOfTuples(StatisticsTable st){
		this.st = st;
		this.requiredStats = new ArrayList();
		setStatistics();
	}
	
	public Iterator requiredStatistics() {
		return this.requiredStats.iterator();
	}

	protected void setStatistics() {
		super.setStatistics();
		
		//these two stats are already added in super class CostModel
		this.requiredStats.add("USED_MEMORY"); // as a reference to show how much real memory has been used
		this.requiredStats.add("MAX_MEMORY");
		
		//this.requiredStats.add("JOIN_PARTITION_NUMBER");
		//this.requiredStats.add("JOIN_PARTITION_SIZE");
		this.requiredStats.add("NUMBER_OF_TUPLES_IN_STATE");
		this.requiredStats.add("NUMBER_OF_TUPLES_IN_INPUT_QUEUES");
		//this.requiredStats.add("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES");
	}
	
	/* 
	 * 
	 */
	public CostTable calculateCost(StatisticsTable st, DistributionTable dt) {
		
		this.dt = dt;
		this.st = st;
		CostTable ct = new CostTable(st.size());
		
		for(Enumeration e = st.keys();e.hasMoreElements();){
			Machine m = (Machine) e.nextElement();
			LightWeightStatistic lws = st.get(m);
			double mem = 0;
			double maxm = 0;
			double outputNum = 0;
			int partNum = 0;
			int partSize = 0;
			
			// no longer been used by a simulated cost model, only for display usage
			String memS = lws.get("USED_MEMORY");
			String maxS = lws.get("MAX_MEMORY");
			if (memS !=null && maxS != null) {
				mem = Double.parseDouble(memS);
				maxm = Double.parseDouble(maxS);
				double ratio = 0;
				if(maxm != 0) {
				    ratio = mem/maxm;
				    ct.put(m, ratio);
				}
				System.out.println("\n" + m.getMachineName() + " UsedM: " + mem + " MaxM: " + maxm + " Ratio: " + ratio);
			}
			
			// we need to count all tuples in input_queue and output_queue of all operators
			int totalInQueue = 0;
			int totalOutQueue = 0;
			int totalState = 0;
			for(Enumeration en = lws.getKeys(); en.hasMoreElements();){
				LightWeightQueryObject xqo = (LightWeightQueryObject) en.nextElement();
				String inStr = lws.get(xqo,"NUMBER_OF_TUPLES_IN_INPUT_QUEUES");
				//only need to count the output queue of the join operator
				//String outStr = lws.get(xqo,"NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES");
				String stateStr = lws.get(xqo, "NUMBER_OF_TUPLES_IN_STATE");
				
				if (inStr != null){
					totalInQueue += Integer.parseInt(inStr);
				}
				//if (outStr != null){
				//	totalOutQueue += Integer.parseInt(outStr);
				//}
				if (stateStr != null) {
				    totalState += Integer.parseInt(stateStr);
				}
			}

			if (totalOutQueue>15000){
				// might be transitant increase of queue size, here, we cut it to 15,000
				// the effect on the main memory usage is not linear
				totalOutQueue = 15000 + ((totalOutQueue-15000)/10);
			}
			//int queSum = totalInQueue+totalOutQueue;
			//double rate = (queSum*1.0 + totalState)/1024.0;
			//double total = queSum*1.0 + totalState;
			//double total = totalInQueue + totalState;
			double total = totalState;
			ct.putRawCost(m, total);			
			//System.out.println("Machine:" + m.getMachineName() + " RealUM: " + mem + " SimUM: " + rate + " PART:SIZE:QN " + partNum + ":" + partSize + ":" + queSum);
			//System.out.println(m.getMachineName() + " SimUM: " + rate + " PART:SIZE:QN " + partNum + ":" + partSize + ":" + queSum);
			System.out.println(m.getMachineName() + " SimUM: " + total + " QueueSize: " + totalInQueue + " StateSize: " + totalState);
		}
		return ct;
	}

}
