/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.manager;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.lang.reflect.Constructor;
import java.net.ConnectException;
import java.net.SocketException;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.Timer;
import jxl.Workbook;
import jxl.write.Label;
import jxl.write.WritableSheet;
import jxl.write.WritableWorkbook;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Application;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.DistributionConfiguration;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Stream;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.LightWeightStats;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.CostModel;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.CostTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.ResourceUsedPct;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.StatisticsTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.DistributedMigrationController;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.DistributedOptimizer;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.NodePairToSwitch;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.SubPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.GetSyncTimeStamp;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendOperatorsToFillStates;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendSubPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendSyncFlag;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SwitchJoinOperatorPlanChange;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.ConnectionManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.ConnectionManagerImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionValidator;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.XATDesignPattern;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.RoundRobinPartitionDistribution;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.redistribution.BalanceMemoryByPartitions;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.redistribution.XATRedistributionPolicy;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.adprotocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.protocol.*;
//import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
//import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUIQueryEngine;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Oct 28, 2003
 * 
 * The DistributionManager is responsible for all of the distribution
 * for Query Execution. It knows all of the query plans, machines to use,
 * streams to connect to, and applications that are listening for results.
 * 
 * The manager is responsible for deciding how to distribute the query plan
 * and tell the machines who does work, and where to send
 * the information. These are the main components:
 * 
 * --Distribution Creator
 * --Connection Sender
 * --Query Plan Validator
 * @TODO fill this in
 */
public class DistributionManager implements Runnable {
	
	private DistributionConfiguration config;
	long lastTimeAdapted;
	boolean inRedistributing = false;
	private DistributionTable distTable;
	
	// Yali @Aug. 20, 2005 Add reference to distributedmigrationcontroller
	DistributedMigrationController migrationController;
	ArrayList timestampreceived = new ArrayList();
	boolean allExecutionSynced = false;
	int planChangedFlagReceived = 0;
	int statesFilledFlagReceived = 0;
	
    boolean done = false;
    
	/* -- Bin @Dec.11,2004
	 * The global information for adaptation protocols, since these values will be set by other statistics-receiver
	 * All these are used in this.adaptPartitions(...)
	 */
	boolean gotPartitionsToMove = false; // indicate whether get the PartitionsToMove message
	ArrayList partitionsToMove = null; // it is set along the getPartitionsToMove message
	
	// for Steps 3 & 4 Protocol: DeactivatePartitions and Deactivated
	int totalNumOfDeactivated = 0;
	
	// for received message
	boolean received = false;
	
	//for load redistribution using parallel track redistribution strategy.
	int totalNumOfDuplicated = 0;
	boolean oldPartitionsExpired = false;
	
	//for coordinate between local migration and other global adaptation
	int localMigrationFlag = 0;
	
	
	public void addTimeStampReceived(long time) {
		this.timestampreceived.add(Long.toString(time));
	}
	
	public void setAllExecutionSynced(boolean s) {
		this.allExecutionSynced = s;
	}
	
	public void setReceived(boolean flag){
		received = flag;
	}

	public synchronized void setDeactivatedMachine(Machine mac){
		totalNumOfDeactivated ++;
	}
	
	public void setOldPartitionsExpired(boolean isExpired) {
	    this.oldPartitionsExpired = isExpired;
	}
	
	public void setPartitionsToMove(boolean flag, ArrayList toMove){
		gotPartitionsToMove = flag;
		partitionsToMove = toMove;
	}
	/**
	 * Default Constructor.
	 * 
	 * @param DistributionConfiguration - The configuration object.
	 */
	public DistributionManager(DistributionConfiguration config) {
		this.config = config;
		init();
	}

	/**
	 *  
	 * Initializes the Manager.
	 *
	 */

	private void init() {

		ArrayList cantConnect = new ArrayList();
		
		migrationController = new DistributedMigrationController(this);

		// Send the configuration information to all of the other servers
		System.out.println("Sending configuration information to " + config.machineCount() + " machine(s).");
		for (Iterator i = config.getMachines(); i.hasNext();) {
			Machine m = (Machine) i.next();
			String ipAddress = m.getIPAddress();
			int port = m.getPort();
			System.out.println("Connecting to " + ipAddress + ":" + port);
			try {

				Socket socket = new Socket(ipAddress, port);
				ObjectOutputStream out = new ObjectOutputStream(socket.getOutputStream());

				out.writeObject(config);
				System.out.println("Successfully sent distribution information to:" + m.getMachineName() + " @ " + m.getIPAddress() + ":" + m.getPort());
				
				// wait to receive settings back from initial GUI
				
				socket.close();

			} catch (ConnectException ce) {
                
				System.out.println("Cannot connect to " + ipAddress + ":" + port + "\n" + ce.getMessage());

				// add this machine to be removed after the loop
				// we cant remove them here because we would get
				// a concurrent modification exception.
				cantConnect.add(m);
			} catch (Exception e) {
				e.printStackTrace();
			}		
		}
		// Remove the machines we couldnt connect to
		if (cantConnect.size() > 0)
			System.out.println("Could not connect to " + cantConnect.size() + " machine(s)!");
		for (int i = 0; i < cantConnect.size(); i++) {
			Machine r = (Machine) cantConnect.get(i);
			config.removeMachine(r.getIPAddress(), r.getPort());
			System.out.println("Removed: " + r.getIPAddress() + ":" + r.getPort());
		}
		
		lastTimeAdapted = System.currentTimeMillis();
		/*// start GUI if necessary
		if (config.getProperty("GUI").equals("on")) {
			CAPE_GUI gui = new CAPE_GUIQueryEngine();
			Thread guiThread = new Thread();
			guiThread = new Thread((CAPE_GUIQueryEngine)gui);
			guiThread.setName("GUI");
			guiThread.setDaemon(true);
			guiThread.start();
		}*/
	}

	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	public void run() {

		if (config.machineCount() == 0) {
			System.out.println("We have no machines to do work. Now Exiting.");
			return;
		}

		try {
			System.out.println("Starting Distribution Manager.");
			
			//Initialize
			//Threads
			ResultReceiver rr;
			StatisticsReceiver sr;
			
			//Tables
			Hashtable appTable = new Hashtable(4);
			Hashtable serverTable = new Hashtable(4);
			//DistributionTable table;
			CostTable costTable = new CostTable(config.getMachineArray());
			StatisticsTable statisticsTable = new StatisticsTable(config.getMachineArray());
			
			//Properties
			int reDistDelay = Integer.parseInt(config.getProperty("REDISBRIBUTION_TIME"));
			int executionTime = Integer.parseInt(config.getProperty("EXECUTION_DURATION"));
			int distDelay = Integer.parseInt(config.getProperty("DISTRIBUTION_DELAY"));
			String distributionPatten = config.getProperty("DISTRIBUTION_PATTERN");
			String costModelString = config.getProperty("WORKLOAD_COST_MODEL");
			int streamTime = Integer.parseInt(config.getProperty("STREAM_DURATION"));
			double reDistPct = Double.parseDouble(config.getProperty("REDISTRIBUTION_PERCENT")) / 100.0;
			int maxStateSize = Integer.parseInt(config.getProperty("STATE_SIZE_THRESHOLD"));
			double differenceThreshold = Double.parseDouble(config.getProperty("DIFFERENCE_THRESHOLD"));
			int requiredMinTupleSize = Integer.parseInt(config.getProperty("REQUIRED_MIN_THRESHOLD"));
			String scope = config.getProperty("REDISTRIBUTION_SCOPE");
			String redistributionStrategy = config.getProperty("REDISTRIBUTION_STRATEGY");
			String redistributionPolicy = config.getProperty("REDISTRIBUTION_POLICY");
			
			long startTime = System.currentTimeMillis();
			long currentTime = startTime;
			executionTime -= distDelay;
			boolean global = false;
			ArrayList lightStats = new ArrayList();
			
			//Distribution Objects
			ConnectionManager cm;
			XATDesignPattern pattern = null;
			DistributionValidator validate;
			XATRedistributionPolicy rp = null;
			CostModel costModel = null;
			
			// for partitioned operator processing -- Bin @ Nov. 29, 2004
			boolean partitionedProcessing = false; // indicate whether it is a partitioned processing
			Hashtable partitionMaps = null; // hold partitionMappings (hashtables) of each split operator
			
			//Spreadsheet Objects
			Date d;
			WritableWorkbook workbook;
			WritableSheet[] sheets;
			
			// for dynamic distributed optimization and migration
			DistributedOptimizer optimizer = new DistributedOptimizer(this);
			Random rand = new Random();
			
//			if(maxStateSize < 0){
//				//there is no maximum threshold for state size
//				maxStateSize = Integer.MAX_VALUE;
//			}
			
			if(scope.equals("global")){
				global = true;
			}
			
			// Start the result receiver
			
			rr = new ResultReceiver(22000, config.machineCount());
			
			Thread tT = new Thread(rr);
			tT.setName("Result Receiver");
			tT.start();
			
			// Start the statistics receiver
			
			sr = new StatisticsReceiver(23000, statisticsTable, config.getMachineArray(),this, migrationController);
			
			Thread sT = new Thread(sr);
			sT.setName("Statistics Receiver");
			sT.start();
			
			// Setup the cost model
			Constructor[] conf = Class.forName(costModelString).getConstructors();
			Constructor cmodelConstructor;

			for(int i = 0; i < conf.length; i++){
				cmodelConstructor = conf[i];
				Class[] parameters = cmodelConstructor.getParameterTypes();
				
				if(parameters.length == 1){
					costModel = (CostModel) cmodelConstructor.newInstance(new Object[] {statisticsTable});
				}
			}
			
			if(costModel == null){
				throw new Exception("The Cost Model could not be instantiated.");
			}
			
			//Find out what statistics to send back to the machines
			Iterator iter = costModel.requiredStatistics();
			while(iter.hasNext()){
				String temp = (String) iter.next();
				lightStats.add(temp);
			}
			
			// Added for detecting system-wide resource overflow.
			// Currently this overflow is decided to exist if all machine's memory percentage 
			// (used memory /max memory) usage is above a certain threshold. 
			// Yali Zhu   August 19, 2005.
			CostModel resourceModel = new ResourceUsedPct(statisticsTable);
			if(!lightStats.contains("USED_MEMORY_PERCENTAGE")) {
				lightStats.add("USED_MEMORY_PERCENTAGE");
			}
			

			// Sleep while waiting to distribute
			// this is done to allow the clients sufficient time to start up.
			Thread.sleep(distDelay);
			
			// Send the statistics	
			LightWeightStats lw = new LightWeightStats(lightStats);
			Packet p = new Packet("lws",lw);
			System.out.print("Sending statistics to machines...");
			for(Iterator it = config.getMachines(); it.hasNext();){
				Machine m = (Machine) it.next();
				Socket s = null;
                long sockettimeout = System.currentTimeMillis() + 15000;
                try {
                s = new Socket(m.getIPAddress(),m.getConnectionListenerPort());
                } catch (SocketException se) {
                    if(sockettimeout < System.currentTimeMillis()) {
                        System.err.println("System could not open socket to " + m.getMachineName());
                        System.exit(-53);
                    }
                    Thread.sleep(1000);
                } while (s == null);
//				Socket s = null;
//				System.out.println ("Attempting to start a socket at " + m.getIPAddress() + m.getConnectionListenerPort());
//				do
//				{
//					try
//					{
//						s = new Socket(m.getIPAddress(),m.getConnectionListenerPort());
//					}
//					catch(Exception ex)
//						{System.err.println("creation of Socket s failed.  Had to do it again");
//						Thread.sleep(100);
//					}
//	 		}while(s == null || s.isClosed());
				 
				ObjectOutputStream out = new ObjectOutputStream(s.getOutputStream());
				out.writeObject(p);
				out.close();
				s.close();
			}
			System.out.print("done.\n");
			
			// Put all of the available applications in the table
			// with the key being the queryplan to which it is to connect
			
			Application[] a = config.getApplicationArray();
			for(int i = 0; i < a.length; i++){
				//appTable.put(a[i].getProperty("CONNECTS_TO"),a[i]);
				//modified by Song Wang on Oct. 2005 to support multiple applications for one query plan
				//basically the change is: the appTable will be a hash table with ArrayList as content.
				if (appTable.get(a[i].getProperty("CONNECTS_TO"))==null){//1st app. for the query plan
					ArrayList appList = new ArrayList();
					appList.add(a[i]);
					appTable.put(a[i].getProperty("CONNECTS_TO"),appList);
				} else {//2nd or more app. for the query plan
					ArrayList appList = (ArrayList)appTable.get(a[i].getProperty("CONNECTS_TO"));
					appList.add(a[i]);
				}
			}
			
			//Find out what all the servers are by streamname
			//put them in another table
			Stream[] s = config.getStreamArray();
			for(int i = 0; i < s.length; i++){
				serverTable.put(s[i].getName(),s[i].getServer());
			}	

			//Create the distribution pattern
			Constructor[] con = Class.forName(distributionPatten).getConstructors();
			Constructor patternConstructor;

			for(int i = 0; i < con.length; i++){
				patternConstructor = con[i];
				Class[] parameters = patternConstructor.getParameterTypes();
				
				if(parameters.length == 2){
					pattern = (XATDesignPattern) patternConstructor.newInstance(new Object[] {config.getMachineArray(), config.getQueryPlanArray()});
				}
			}
			
			if(pattern == null){
				throw new Exception("The Distribution Pattern could not be instantiated.");
			}
			
			//DISTRIBUTE THE OPERATORS
            // distTable is now defined in the class.
			//DistributionTable distTable = pattern.distribute();
            distTable = pattern.distribute();
            
			// based on the table, check whether the distribution is a partitioned distribution, 
			// for example, one join operator has been distributed to multiple machines -- Bin @ Nov.29, 2004 
			partitionedProcessing = checkPartitionedProcessing(distTable);
            System.out.println("partitioned processing = " + partitionedProcessing);
			//partitionedProcessing = true;  // replace the above statement by this line for testing purpose.
            System.out.println(distTable.toString());
			Machine separateMachine = null; // in case of separate union machine, we also need to spearate the

											// the union machine from the adaptation
			Timer tpt = null; //Token-passing timer for Brad M.'s first experiment.
			
			if (partitionedProcessing) {

				boolean replicationstyle = false;
				ArrayList splitIDs = new ArrayList();
				//Decide whether partitioning is Bin's style or Brad M.'s style.
				Enumeration eta = distTable.getNodeEnumeration();
				while (eta.hasMoreElements())
				{
					XATNode nextelem = (XATNode)eta.nextElement();
					if(nextelem.getOperator().getClass().getName().indexOf("Replicat") >= 0
							|| nextelem.getOperator().getClass().getName().indexOf("Spreadfire") >= 0)
						replicationstyle = true;
					if(nextelem.getOperator().getClass().getName().indexOf("Split") >= 0)
						splitIDs.add(nextelem);
				}

				if (replicationstyle) {
					//handle Brad M.'s code, including a timer to fire token messages for experimental purposes.
					//tpt = new Timer(true);  //true as a timer arg is meant to set as daemon.
						// Daemon timers don't hold up the parent class, but rather terminate when asked.
					//tpt.scheduleAtFixedRate((TimerTask)new TimedStorageTokenPassing(table, splitIDs), 30000, 30000);

					
					cm = new ConnectionManagerImp(distTable,serverTable,appTable,streamTime);
					cm.connect();
				} else {
				
					// for a partitioned processing, first get the partitionMaps from distribution pattern
					// a short cut in case we need skewed initial distribution, the true or false indicate whether separate union?
					partitionMaps = ((RoundRobinPartitionDistribution)pattern).skewedMapPartitions(0, 0.45, false);
					// need to be called only if the third parameter above is true
					separateMachine = ((RoundRobinPartitionDistribution)pattern).getUnionMachine();
				
					//partitionMaps = pattern.mapPartitions(); // normal round robin initial partition distribution
					System.out.println("size:" + partitionMaps.size());
					
					cm = new ConnectionManagerImp(distTable,partitionMaps,serverTable,appTable,streamTime);
					cm.connectPartitions();
					// for without using trace method to update cost statistics for multiple joins, this propagatePartitionMaps
					// method could be safely commented out -- Bin Apr 17, 2005
					cm.propagatePartitionMaps(config.getMachineArray());
				}	
			} else {
				// the original process developed by d-cape
				//Validate the distribution
				validate = new DistributionValidator(config.getQueryPlanArray());			
				if (validate.validate(distTable)) {
					//Connect the operators to the machines
					cm = new ConnectionManagerImp(distTable,serverTable,appTable,streamTime);
					cm.connect();
				}
				else{
					System.out.println("This is an illegal distribution.");
					return;
				}				
			}
						
			System.out.println("Completed Initial distribution.");
			
			// Create the redistribution policy
			Constructor[] cons = Class.forName(redistributionPolicy).getConstructors();
			Constructor reDistConstructor;
			
			for(int i = 0; i < cons.length; i++){
				reDistConstructor = cons[i];
				Class[] parameters = reDistConstructor.getParameterTypes();
				
				if(parameters.length == 0){
					rp = (XATRedistributionPolicy) reDistConstructor.newInstance(new Object[] {});
				}
			}
			
			if(rp == null){
				throw new Exception("The Redistribution Policy could not be instantiated.");
			} else {
			    if (redistributionPolicy.endsWith("BalanceMemoryByPartitions")) {
			       ((BalanceMemoryByPartitions)rp).setDifferenceThreshold(differenceThreshold);
			       ((BalanceMemoryByPartitions)rp).setMinRequiredMemorySize(requiredMinTupleSize);		        
			    }
			}
			
			//System.out.println(table);
			
			int count = 0;
			Machine[] machineArray = config.getMachineArray();
			int time_to_debug = 0; //only for debug use
			//if it is too close to the end, then we do not perform adaptation
			boolean noMoreAdaptation = false; 
			int time_to_wait = 6; //the default time to wait (in terms of reDistDelay)
			int migration_count=0;
			
			//Thread.sleep(10000);
			
		try{	
			do{
				System.out.println(count);
				double max,min;
				System.out.print("Sleeping ..\n");
				Thread.sleep(reDistDelay);
				System.out.println("DM has run for: " + (System.currentTimeMillis() - startTime));
				System.out.println("\n\nRedistributing..");
				migration_count++;
				
				//System.out.println(statisticsTable);
				
				//*** Redistributing partitions among machines *****
				if (partitionedProcessing) {
				    System.out.println("Partition-level Adaptation");
					costTable = costModel.calculateCost(statisticsTable, distTable);
					
					if (((System.currentTimeMillis() - lastTimeAdapted) > reDistDelay)
							&& (!inRedistributing) && (!noMoreAdaptation)) {
					    
					    //if one of the processor is in local migration, 
					    //wait for one more round to apply load balancing.
					    if(this.localMigrationFlag > 0) {
					        System.out.println("At least one process is in local migration");
					        System.out.println("Wait for one more round to apply partition redistribution");
					        continue;
					    }
					    
						//redistribute only after certain time, old adaptation
						//has been completed
					    System.out.println("\nRedistribute Partitions");
						ArrayList reDistPara = rp.redistribute(distTable, costTable, statisticsTable, config.getMachineArray(), separateMachine);
						if (reDistPara != null) {
							if (reDistPara.size() == 3) {
								//it is a push to disk adaptation
								System.out.println("A ForcePartitionsToPush adaptation.");
								pushPartitions(reDistPara);
								time_to_wait = 6;
							} else {
								//it is distributed adaptation across machines
							    if(redistributionStrategy.equals("paralleltrack"))
							        adaptPartitionsUseParallelTrack(reDistPara);
							    else
							        adaptPartitions(reDistPara);
								time_to_wait = 6;
							}
							lastTimeAdapted = System.currentTimeMillis();
							time_to_debug++;
						} else {
							//do nothing
						}
					}
					
					// skip the operator-level redistribution.
					continue;
				}
				
				//****/
				/*//debug code for distribution-adaptation				
				while (time_to_debug <=1) {
					if (time_to_debug>0){
						System.out.println("yes, second adaptation?");
					}
					ArrayList reDistPara = rp.redistribute(table,costTable,config.getMachineArray());
					if (reDistPara != null) {
						adaptPartitions(reDistPara);
						time_to_debug++;
					} else {
						break;
					}
				}
				*/
				//System.out.println("Tables before recalculate");
				//System.out.println(statisticsTable);
				//System.out.println(table);
	

				
				//double resourcePctThreshold = 0.7;
				//costTable = resourceModel.calculateCost(statisticsTable, distTable);
				//min = costTable.minCost();
				//if(min >= resourcePctThreshold) {
				System.out.println("Try Global Migration");
				double nextrand = 0;
				nextrand = rand.nextDouble();
				//System.out.println("\n\n nextrand = " + nextrand);
				//System.out.println("migration count = " + migration_count);
				if(nextrand > 0 && migration_count == 0) {
					migration_count ++;
					System.out.println("System resource overflow detected");
					globalPlanMigration(cm, optimizer);
					continue;
				}				
				
				/** The following is the operator-based adaptation -- Bin */
				System.out.println("Operator-level Adaptation");
				
/**
 * 	COMMENTED OUT BY CHUAN LEI
 */			
//				costTable = costModel.calculateCost(statisticsTable, distTable);
//				//System.out.println(costTable);
//				max = costTable.maxCost();
//				min = costTable.minCost();
//				
//				
//				System.out.println(statisticsTable);
//				System.out.println("Cost table after recalculate");
//				System.out.println(costTable);
//				
//				//System.out.println("reDistPct = " + reDistPct);
//				//System.out.println("max = "+max);
//				//System.out.println("min = "+min);
//				
//	
//				/** Tim's operator-level redistribution */
//				if(reDistPct < max - min){
//					//System.out.println(statisticsTable);
//					//System.out.println(costTable);
//					System.out.println("The reDist is big enough to redistirbute");
//					
//					//Comment out for now.
//					//Until testing is done.
//					DistributionTable newDistTable = rp.redistribute(distTable,costTable,null,config.getMachineArray());
//					
//					//DistributionTable newTable = (DistributionTable) table.clone();
//					//XATNode nodeToMove = newTable.getNode("2");
//					//Machine tempMachine = machineArray[count];
//					//newTable.add(nodeToMove,tempMachine);
//					
//					//System.out.println("New Distribution:");
//					//System.out.println(newDistTable);
//			
//					//table = (DistributionTable) newTable.clone();
//					cm.reconnect(newDistTable);	
//					//System.out.println("new table after reconnect:");
//					//System.out.println(newTable);
//					distTable = newDistTable;
//					//System.out.println("Distribution table after reconnect:");
//					//System.out.println(table);
//				
//				} else {
//					//if(reDistPct >= 1)	
//					//System.out.println("Redistribution is disabled. ");
//				}
				//****************/
				
				System.out.print("done.\n");
								
				currentTime = System.currentTimeMillis();
				
				if(count >= machineArray.length-1){
					count = 0;
				}
				else{
					count++;
				}
				if (currentTime - startTime > executionTime - reDistDelay){
					noMoreAdaptation = true;
				}
			} while (currentTime - startTime < executionTime && !done);
		} catch (InterruptedException ie) 
			{
				//Note that the purpose of the catch block is just to
				// stop the main loop and do finalization in case 
				// immediate shutdown is needed.
			}
			System.out.println("Waiting for results from " + config.machineCount() + " machines.");
			
			while(!rr.isDone())
					Thread.sleep(2000);
			
			//Request that the Result Receiver stop.
			rr.requestStop();
					
			System.out.print("Received all of the results files, now creating master file...");
			d = new java.util.Date();
			
			
			// CREATE THE EXCEL WORKBOOK
			workbook = Workbook.createWorkbook(new File("Result-" + d.getMonth() + "-" + d.getDate() + "-" + d.getHours() + "-" + d.getMinutes() + ".xls"));
			sheets = new WritableSheet[rr.getNumFileNames() +2];
			
			//Add the Parameter worksheet
			sheets[0] = workbook.createSheet("Experiment Parameters",0);
			
			StringTokenizer st = new StringTokenizer(config.toString(),"\n");
			int i=0;
			while(st.hasMoreTokens()){
				String temp = st.nextToken();
				Label label = new Label(0, i, temp); 
				sheets[0].addCell(label); 
				i++;
			}
						
			int j=1;
			int k=0;
			int l=0;
			
			//Add one worksheet for each machine.
			
			for(Iterator it = rr.getFileNames();it.hasNext();){
				String name = (String) it.next();
				String line;
				sheets[j] = workbook.createSheet(name,j);
			
				//Now add in the data
				
				BufferedReader br = new BufferedReader(new FileReader(name));
				k=0;
				while((line = br.readLine()) != null){
					StringTokenizer tr = new StringTokenizer(line,",");
					
					l =0;
					while(tr.hasMoreTokens()){
						String token = tr.nextToken();
						Label label = new Label(l,k,token);
						sheets[j].addCell(label);
						l++;
					}
					k++;
				}
				
				j++;
			}
			//now make the experiment worksheet.
			sheets[j] = workbook.createSheet("Charts",j);

			workbook.write(); 
			workbook.close(); 
			
			// END EXCEL WORKBOOK
			
			System.out.print("done.\n");
			System.exit(0);
			return;
			

		} catch (Exception e) {
			e.printStackTrace();
		}

	} //end run
		
	/* -- Bin @ Nov.29, 2004
	 * check whether the distribution table generated is a partitioned processing
	 * If we have one operator assigned to more than one machines, return true, else return false
	 */
	public boolean checkPartitionedProcessing(DistributionTable table){
		Enumeration nodeKeys = table.getNodeEnumeration();
		while (nodeKeys.hasMoreElements()){
			XATNode curNode = (XATNode)nodeKeys.nextElement();
			Object machines = table.getObjectByNode(curNode);
			if (machines.getClass().isArray()){
				return true;
			}
		}		
		return false;
	} //end checkPartitionedProcessing
	
	public DistributionConfiguration getConfig() {
		return this.config;
	}
	
	/* -- Bin @ Jan 2, 2005
	 * the parameters in pushPara, (0) machine, (1) operatorID, (2) rate
	 */
	public boolean pushPartitions(ArrayList pushPara){
		try {
			Machine mm = (Machine)pushPara.get(0);
			Socket ms = new Socket(mm.getIPAddress(),mm.getConnectionListenerPort());					 
			ObjectOutputStream mout = new ObjectOutputStream(ms.getOutputStream());
			ForcePartitionsToPush fpt = new ForcePartitionsToPush((String)pushPara.get(1),Double.parseDouble((String)pushPara.get(2)));
		
			Packet pk = new Packet("protocol",fpt);					
			mout.writeObject(pk);
			mout.close();
			ms.close();
			System.out.println("Sending out the forcePartitionsToPush message.");
		} catch (Exception e){
			e.printStackTrace();
			return false;
		}
		return true;
	}
	
	/* -- Bin @ Dec.11, 2004
	 * This function contains the main logic (protocols) of adapting partitions across participant machines
	 * The protocols include, (1) ComputePartitionsToMove, (2)PartitionsToMove, (3)DeactivatePartitions, (4)Deactivated
	 * (5)SendPartitions, (6)ReceivePartitions, (7)Received, and (8)ReactivatePartitions
	 * reDistPara includes the basic information about this adaptation, sender, receiver, operator and percentage
	 * (0) sender Machine, (1) receiver Machine, (2) operator ID, (3) rate
	 */
	public boolean adaptPartitions(ArrayList reDistPara){
		inRedistributing = true;
        long msstarttime = System.currentTimeMillis();
		try {
			// (1), get the sender Machine information, and send a ComputePartitionsToMove message
			Machine fm = (Machine)reDistPara.get(0);
			Machine tm = (Machine)reDistPara.get(1);
			Socket fs = new Socket(fm.getIPAddress(),fm.getConnectionListenerPort());					 
			ObjectOutputStream fout = new ObjectOutputStream(fs.getOutputStream());
			ComputePartitionsToMove cpt = new ComputePartitionsToMove((String)reDistPara.get(2),Double.parseDouble((String)reDistPara.get(3)));
		
			Packet pk = new Packet("protocol",cpt);					
			fout.writeObject(pk);
			fout.close();
			fs.close();
			
			// (2) wait until the gotPartitionsToMove has been set
			while (!gotPartitionsToMove){
				Thread.sleep(1500);
			}			
			System.out.println("yes, we got PartitionsToMove message back to DM.");
			
			// (3) send DeactivatePartitions message to all machines			
			for(Iterator it = config.getMachines(); it.hasNext();){
				Machine m3 = (Machine) it.next();
				Socket s3 = new Socket(m3.getIPAddress(),m3.getConnectionListenerPort());
				 
				ObjectOutputStream out3 = new ObjectOutputStream(s3.getOutputStream());
				DeactivatePartitions dp = new DeactivatePartitions((String)reDistPara.get(2),partitionsToMove,fm);
				Packet pk3 = new Packet("protocol",dp);	
				out3.writeObject(pk3);
				out3.close();
				s3.close();
			}
			
			// (4) wait until the Deactivated messages from all machines, a short-cut, only use the count
			while (totalNumOfDeactivated < config.machineCount()) {
				Thread.sleep(1500);
			}			
			System.out.println("yes, we got all Deactivated message back to DM.");
			
			// (5) then send SendPartitions message to the sender machine
			SendPartitions sp = new SendPartitions((String)reDistPara.get(2),partitionsToMove,fm,tm);
			fs = new Socket(fm.getIPAddress(),fm.getConnectionListenerPort());					 
			fout = new ObjectOutputStream(fs.getOutputStream());
			
			Packet pk5 = new Packet("protocol",sp);
			fout.writeObject(pk5);
			fout.close();
			fs.close();
			System.out.println("send out the SendPartitions Message?");
			// (6) wait until the Received message sent from tm 
			while (!received) {
				Thread.sleep(1500);
			}
			System.out.println("yes, we got all Received message back to DM.");
			
			// (7), ok, the final step is to reactivate the partitions in all partition operators
			for(Iterator it = config.getMachines(); it.hasNext();){
				Machine m7 = (Machine) it.next();
				Socket s7 = new Socket(m7.getIPAddress(),m7.getConnectionListenerPort());
				 
				ObjectOutputStream out7 = new ObjectOutputStream(s7.getOutputStream());
				ReactivatePartitions rp = new ReactivatePartitions((String)reDistPara.get(2),partitionsToMove,tm);
				
				Packet pk7 = new Packet("protocol",rp);	
				
				out7.writeObject(pk7);
				out7.close();
				s7.close();
			}
			// debug use, print out the partitions IDs that have been moved
			for (int i=0; i<partitionsToMove.size(); i++){
				System.out.print("P:" + (Integer)partitionsToMove.get(i) + ", ");
			}
			// clear all set flags for the next adaptation.
			gotPartitionsToMove = false;
			partitionsToMove = null;
			totalNumOfDeactivated = 0;
			received = false;
			
		} catch (Exception e) {
			System.out.println(e.getMessage());
			return false;
		}
		inRedistributing = false;
        System.out.println("\nMS global balancing takes " + (System.currentTimeMillis() - msstarttime));
		return true;
	} //end of adaptPartitions
	
	public DistributionTable getDistributionTable() {
		return distTable;
	}

	public ArrayList findOperatorsToSynchronize(SubPlan subplan) {
		// figure out old leaf operators that need to be synchronized
		ArrayList ops = new ArrayList();
		ArrayList leafops = new ArrayList();
		boolean isleaf = false;
		
		ArrayList rootops = subplan.getRootOperators();
		for(int i =0; i<rootops.size(); i++) {
			ops.add(rootops.get(i));
		}
		//ops.add(subplan.getRootOperators());
		for(int i=0; i<ops.size(); i++) {
			XATStreamOperator currentop = (XATStreamOperator)ops.get(i);
			Object[] cnodes = currentop.getNode().getChildren();
			for(int j=0; j<cnodes.length; j++) {
				if(cnodes[j] instanceof XATNode)
					ops.add(((XATNode)cnodes[j]).getStreamOperator());
				else
					isleaf = true;
			}
			if(isleaf) {
				if(!leafops.contains(currentop))
					leafops.add(currentop);
				isleaf = false;
			}		
		}
		System.out.println("leaf ops: " + leafops);
		return leafops;
	}
	
	public ArrayList findOperatorsToSynchronize(NodePairToSwitch nodepair) {
		// figure out old leaf operators that need to be synchronized
		ArrayList ops = new ArrayList();
		ArrayList leafops = new ArrayList();
		boolean isleaf = false;
		
		ops.add(nodepair.getParentNode().getStreamOperator());

		for(int i=0; i<ops.size(); i++) {
			XATStreamOperator currentop = (XATStreamOperator)ops.get(i);
			Object[] cnodes = currentop.getNode().getChildren();
			for(int j=0; j<cnodes.length; j++) {
				if(cnodes[j] instanceof XATNode)
					ops.add(((XATNode)cnodes[j]).getStreamOperator());
				else
					isleaf = true;
			}
			if(isleaf) {
				if(!leafops.contains(currentop))
					leafops.add(currentop);
				isleaf = false;
			}		
		}
		System.out.println("leaf ops: " + leafops);
		return leafops;
	}
	
	private ArrayList requireSyncTimestamp(DistributionTable disttable, ArrayList leafops, SubPlan subplan) {
		ArrayList leafids = new ArrayList();
		ArrayList leafActiveMachines = new ArrayList();
		Machine machine;
		
		for(int i=0; i<leafops.size(); i++) {
			XATStreamOperator currentop = (XATStreamOperator)leafops.get(i);
			machine = disttable.getMachine(currentop.getNode());
			if(!leafActiveMachines.contains(machine))
				leafActiveMachines.add(machine);
			leafids.add(Integer.toString(currentop.getObjectID()));
		}
//			// find leaf operators on the same machine
//			for(int j=i+1; j<leafops.size(); j++) {
//				currentop = (XATStreamOperator)leafops.get(j);
//				if(disttable.getMachine(currentop.getNode()).equals(machine)) {
//					leafids.add(Integer.toString(currentop.getObjectID()));
//					leafops.remove(j);
//					j--;
//				}
//			}
			
		for(int i=0; i< leafActiveMachines.size(); i++) {
			machine = (Machine)leafActiveMachines.get(i);
			// send sync message to all leaf active machines
			try {
				//System.out.println("DM migration: send sync time request to machine " + machine);
				Socket ms = new Socket(machine.getIPAddress(), machine.getConnectionListenerPort());
				ObjectOutputStream mout = new ObjectOutputStream(ms.getOutputStream());
				GetSyncTimeStamp synctime;
				// if the old root operator of the subplan is active on this machine
				// send the root operator id to that machine
				if(disttable.getMachine(subplan.getOldRootOperator().getNode()).equals(machine)) {
					synctime = new GetSyncTimeStamp(subplan.getOperatorIDs(), leafids, subplan.getOldRootOperatorID());
				} else {
					synctime = new GetSyncTimeStamp(subplan.getOperatorIDs(), leafids, -1);
				}
				Packet pk = new Packet("migrationprotocol", synctime);
				mout.writeObject(pk);
				mout.close();
				ms.close();
				System.out.println("Sending out the sync time request to machine " + machine.getMachineName());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return leafActiveMachines;
	}
	
	private ArrayList requireSyncTimestamp(DistributionTable disttable, ArrayList leafops, NodePairToSwitch nodepair) {
		ArrayList leafids = new ArrayList();
		ArrayList leafActiveMachines = new ArrayList();
		Machine machine;
		
		for(int i=0; i<leafops.size(); i++) {
			XATStreamOperator currentop = (XATStreamOperator)leafops.get(i);
			machine = disttable.getMachine(currentop.getNode());
			if(!leafActiveMachines.contains(machine))
				leafActiveMachines.add(machine);
			leafids.add(Integer.toString(currentop.getObjectID()));
		}
			
		for(int i=0; i< leafActiveMachines.size(); i++) {
			machine = (Machine)leafActiveMachines.get(i);
			// send sync message to all leaf active machines
			try {
				//System.out.println("DM migration: send sync time request to machine " + machine);
				Socket ms = new Socket(machine.getIPAddress(), machine.getConnectionListenerPort());
				ObjectOutputStream mout = new ObjectOutputStream(ms.getOutputStream());
				GetSyncTimeStamp synctime;
				// if the old root operator of the subplan is active on this machine
				// send the root operator id to that machine.
				// Since the disttable is already updated, the two machines the two nodes are on have been switched,
				// so we need the machine that the child node (the new top join) is on, which is the machine the old
				// top operator is on.
				if(nodepair.getParentOldMachine().equals(machine)) {
					synctime = new GetSyncTimeStamp(nodepair.getOperatorIDs(), leafids, nodepair.getParentOperatorID());
				} else {
					synctime = new GetSyncTimeStamp(nodepair.getOperatorIDs(), leafids, -1);
				}
				Packet pk = new Packet("migrationprotocol", synctime);
				mout.writeObject(pk);
				mout.close();
				ms.close();
				System.out.println("Sending out the sync time request to machine " + machine.getMachineName());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return leafActiveMachines;
	}
	
	private long computeSyncTimestamp(ArrayList timestamps) {
		long maxtimestamp;
		long time;
		
		maxtimestamp = Long.parseLong((String)timestamps.get(0));
		for(int i=1; i<timestamps.size(); i++) {
			time = Long.parseLong((String)timestamps.get(i));
			
			if(time > maxtimestamp)
				maxtimestamp = time;
		}
		return maxtimestamp;
	}
	
	private void sendSyncFlag(ArrayList leafMachines, long synctimestamp) {
		
		Machine machine;
		
		for(int i=0; i<leafMachines.size(); i++) {
			// send sync message to one machine
			machine = (Machine)leafMachines.get(i);
			try {
				Socket ms = new Socket(machine.getIPAddress(), machine.getConnectionListenerPort());
				ObjectOutputStream mout = new ObjectOutputStream(ms.getOutputStream());
				SendSyncFlag syncflag = new SendSyncFlag(synctimestamp);

				Packet pk = new Packet("migrationprotocol", syncflag);
				mout.writeObject(pk);
				mout.close();
				ms.close();
				System.out.println("Sending out the sync time to machine " + machine.getMachineName());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}
	
	private void sendSubplanToAllMachines(DistributionTable disttable, SubPlan subplan) {
		
		// first update the statelocation hashtable in subplan to replace operator by machine
		Hashtable stateLocations = subplan.getStateLocationTable();
		Enumeration en = stateLocations.keys();
		XATStreamOperator op;
		Machine machine;
		String stateID;
		while(en.hasMoreElements()) {
			stateID = (String)en.nextElement();
			op = (XATStreamOperator)stateLocations.get(stateID);
			machine = disttable.getMachine(op.getNode());
			stateLocations.put(stateID, machine);
		}
		
		Iterator it = this.config.getMachines();
		
		while(it.hasNext()) {		
			//send the subplan to one machine
			machine = (Machine)it.next();
			try {
				Socket ms = new Socket(machine.getIPAddress(), machine.getConnectionListenerPort());
				ObjectOutputStream mout = new ObjectOutputStream(ms.getOutputStream());
				SendSubPlan sendplan = new SendSubPlan(subplan);

				Packet pk = new Packet("migrationprotocol", sendplan);
				mout.writeObject(pk);
				mout.close();
				ms.close();
				System.out.println("Sending the new subplan to machine " + machine.getMachineName());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}
	
	private void sendPlanChangeToAllMachines(DistributionTable disttable, NodePairToSwitch nodepair) {
		
		Iterator it = this.config.getMachines();
		Machine machine;
		int parentOpID = nodepair.getParentOperatorID();
		int childOpID = nodepair.getChildOperatorID();
		//Machine parentActiveMachine = nodepair.getChildOldMachine();
		//Machine childActiveMachine = nodepair.getParentOldMachine();
		boolean isLeft = nodepair.getIsLeftChild();
		
		while(it.hasNext()) {		
			//send the planchange to one machine
			machine = (Machine)it.next();
			try {
				Socket ms = new Socket(machine.getIPAddress(), machine.getConnectionListenerPort());
				ObjectOutputStream mout = new ObjectOutputStream(ms.getOutputStream());
				int oldActiveOpID = -1;
				int newActiveOpID = -1;
				
				if(machine == nodepair.getChildOldMachine()) {
				    oldActiveOpID = childOpID;
				    newActiveOpID = parentOpID;
				} else if(machine == nodepair.getParentOldMachine()) {
				    oldActiveOpID = parentOpID;
				    newActiveOpID = childOpID;
				}
				
				SwitchJoinOperatorPlanChange sendchange = new SwitchJoinOperatorPlanChange(parentOpID, childOpID, oldActiveOpID, newActiveOpID, isLeft);
				Packet pk = new Packet("migrationprotocol", sendchange);
				mout.writeObject(pk);
				mout.close();
				ms.close();
				System.out.println("Sending plan change to machine " + machine.getMachineName());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Send to a machine the set of operator IDs to be activated on that machine after the migration is over.
	 * This indirectly tells the machine what operator states need to be copied, moved or recomputed. 
	 * Each machine will figure that out based on existing state list, statelocations sent in a subplan and this
	 * set of operator IDs.
	 * @param subplan
	 */
	public int sendOperatorIDsToBeActivated(DistributionTable disttable, SubPlan subplan) {
		int machinecount = 0;
		
		ArrayList operators = subplan.getOperators();
		boolean[] visited = new boolean[operators.size()];
		for(int i=0; i<visited.length; i++)
			visited[i] = false;
		
		ArrayList activeopid = new ArrayList();
		XATStreamOperator op;
		for (int i = 0; i < operators.size(); i++) {
			if (visited[i] != true) {
				activeopid.clear();
				op = (XATStreamOperator) operators.get(i);
				Machine machine = disttable.getMachine(op.getNode());
				activeopid.add(Integer.toString(op.getObjectID()));

				for (int j = i + 1; j < operators.size(); j++) {
					op = (XATStreamOperator) operators.get(j);
					if ((visited[j] != true) && (disttable.getMachine(op.getNode()).equals(machine))) {
						activeopid.add(Integer.toString(op.getObjectID()));
						visited[j] = true;
						//operators.remove(j);
						//j--;
					}
				}
				//send the operatorid set to current machine
				try {
					machinecount++;
					Socket ms = new Socket(machine.getIPAddress(), machine.getConnectionListenerPort());
					ObjectOutputStream mout = new ObjectOutputStream(ms.getOutputStream());
					SendOperatorsToFillStates ofs = new SendOperatorsToFillStates(activeopid);

					Packet pk = new Packet("migrationprotocol", ofs);
					mout.writeObject(pk);
					mout.close();
					ms.close();
					System.out.println("Sending operator to fill" + activeopid + " to machine " + machine.getMachineName());
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
		return machinecount;
	}
	
	
	public void terminateExecutionNow()
	{
		//this will tell the DM to stop the main loop and do graceful shutdown.
		//  it will also tell all the QPs to do the same.
		
		Thread[] allthreads = new Thread[Thread.currentThread().getThreadGroup().activeCount()];
		Thread.enumerate(allthreads);
		for (int i = 0; i < allthreads.length; i++)
		{
			if(allthreads[i].getName().equals("Distribution Manager"))
					allthreads[i].interrupt();
		}
		
		System.out.println("Now stopping execution.");
		
		Machine[] m = this.getConfig().getMachineArray();
		for (int i = 0; i < m.length ; i++)
		{
			try {
			Socket s = new Socket(m[i].getIPAddress(), m[i].getConnectionListenerPort());
			ObjectOutputStream ooe = new ObjectOutputStream(s.getOutputStream());
			Packet p = new Packet("shut down", new SpreadfireEmergencyShutdown());
			ooe.writeObject(p);
			s.close();
			System.out.println("Sent shutdown request to machine: " + m[i].getProperty("NAME"));
			
			} catch (java.net.UnknownHostException uhe) { uhe.printStackTrace(); }
			  catch (IOException ioe) { ioe.printStackTrace(); }
		}
        
        done = true;	
	}
	
	/** Add for distributed migration across machines **/
	/** If resources on all machines are over certain resource threshold 
	 *  (this means simply moving workload from one machine to another is no longer applicable), 
	 *  distribution manager would try to apply run-time query plan optimization and migration
	 *  across machines to decrease the total resources required by the system. 
	 *  Added by Yali Zhu
	 *  v1.0 August, 17, 2005: assumes that all machines have the same available resources so only one
	 *                         threshold is employed. also assume current cost model is to measure a 
	 *                         type of system resources, such as memory. */
	private void globalPlanMigration(ConnectionManager cm, DistributedOptimizer optimizer) {
        try {
            /** Apply optimization heuristics and get the new subplan */
            // Assume the system only executes a single query plan
            QueryPlan[] plans = config.getQueryPlanArray();
            NodePairToSwitch nodepair;
            nodepair = optimizer.optimizeQueryPlan(plans[0], this.distTable);
            //System.out.println("\n\nDM migration step 1: generates subplan: "
            // + subplan);
            // If plan not changed, skip the migration process.
            if (nodepair == null) {
                System.out.println("Plan not changed, skip the migration process");
                return;
            }

            /** step 1: Get synchronize timestamps from contributing machines */
            System.out.println("\n\n");
            System.out.println("DM migration step 1: send sync time request to leaf machines");
            //System.out.println("***** Subplan is: *******\n" + subplan);
            ArrayList leafops = findOperatorsToSynchronize(nodepair);
            ArrayList activeLeafMachines = requireSyncTimestamp(distTable, leafops, nodepair);
            while (this.timestampreceived.size() < activeLeafMachines.size()) {
                //System.out.println("DM waiting: received synctime ack from "
                // + this.timestampreceived.size() + " machines");
                Thread.sleep(1000);
            }
            System.out.println("DM migration step 2: received synctime ack from ALL machines");
            long synctime = this.computeSyncTimestamp(timestampreceived);
            System.out.println("The computed synctime is " + synctime);
            if (synctime == 0)
                System.out.println("The execution is already synced. Skip step 3.");
            //this.timestampreceived.clear();

            /**
             * step 3: Send sync message to contributing machines and wait for
             * all-synced ack
             */
            if (synctime != 0) {
                System.out.println("\n\n");
                System.out.println("DM migration step 3: send sync message to machines");
                //System.out.println("***** Subplan is: *******\n" + subplan);
                this.sendSyncFlag(activeLeafMachines, synctime);
                while (this.allExecutionSynced == false) {
                    //System.out.println("DM waiting: need to receive allsync
                    // ack from root operator");
                    Thread.sleep(1000);
                }
                this.allExecutionSynced = false;
                System.out.println("DM migration step 4: received allsync ack from root operator");
            }

            /**
             * step 5: Send subplan to all machines and wait for planchanged ack
             * from all machines
             */
            System.out.println("\n\n");
            System.out.println("DM migration step 5: Send subplan to ALL machines.");
            //System.out.println("***** Subplan is: *******\n" + subplan);
            //sendSubplanToAllMachines(distTable, subplan);
            
            sendPlanChangeToAllMachines(distTable, nodepair);
            while (this.planChangedFlagReceived != this.config.machineCount()) {
                //System.out.println("DM waiting: received planchanged ack from
                // " + this.planChangedFlagReceived + " machines");
                Thread.sleep(1000);
            }
            this.planChangedFlagReceived = 0;
            System.out.println("DM migration step 6: received planchanged ack from ALL machines");

//            /**
//             * step 7: Send movingstates message to all machines and wait ack
//             * indicating it's done
//             */
//            System.out.println("\n\n");
//            System.out.println("DM migration step 7: Send operatorIDs to fill to machines.");
//            //System.out.println("***** Subplan is: *******\n" + subplan);
//            int machinecount = sendOperatorIDsToBeActivated(distTable, nodepair);
//            //System.out.println("***** Subplan is: *******\n" + subplan);
//            while (this.statesFilledFlagReceived != machinecount) {
//                //System.out.println("DM waiting: received statefilled ack from
//                // " + this.statesFilledFlagReceived + " machines");
//                Thread.sleep(1000);
//            }
//            this.statesFilledFlagReceived = 0;
//            System.out.println("DM migration step 8: received statefilled ack from ALL machines");

           
            /** 
             * step 7: reactivate operators in subplan according to distribution table 
             */
            System.out.println("\n\nDM migration step 7: Reactivate and connect changed plan.");
            //System.out.println("***** Subplan is: *******\n" + subplan);
            //cm.connectSubPlan(subplan);
            
            ArrayList subplanoperators = new ArrayList();
            subplanoperators.add(nodepair.getParentNode().getStreamOperator());
            subplanoperators.add(nodepair.getChildNode().getStreamOperator());
            cm.connectSubPlan(subplanoperators);

            //System.out.println("\n\nThe query tree after migration is: " + this.config.getQueryPlanArray()[0].getQueryPlan());

//            ArrayList subplanoperators = subplan.getOperators();
//            for (int i = 0; i < subplanoperators.size(); i++) {
//                ((XATStreamOperator) subplanoperators.get(i)).setIsSubPlanOperator(false);
//                ((XATStreamOperator) subplanoperators.get(i)).setIsSubPlanRootOperator(false);
//            }

            //System.out.println("\n\nThe query tree after migration is: " + this.config.getQueryPlanArray()[0].getQueryPlan());
            System.out.println("\n\n Migration process is done.");

            /** Reset all class variables if necessary */
            timestampreceived.clear();
            allExecutionSynced = false;

            // subplan operators reset to non subplanoperators or rootoperators.
            //subplan = null;
            nodepair = null;

        } catch (Exception e) {
            e.printStackTrace();
            System.exit(-1);
        }
    }
	
	
	/* -- Yali @ April.2, 2006
	 * This function contains the main logic (protocols) of adapting partitions across participant machines
	 * using parallel track distribution strategy.
	 * The join operator in the query plan should be WTransitiveHashJoinOperator.
	 * reDistPara includes the basic information about this adaptation, sender, receiver, operator and percentage
	 * (0) sender Machine, (1) receiver Machine, (2) operator ID, (3) rate
	 */
	public boolean adaptPartitionsUseParallelTrack(ArrayList reDistPara){
		inRedistributing = true;
        long ptstarttime = System.currentTimeMillis();
		try {
			// (1), get the sender Machine information, and send a ComputePartitionsToMove message
			Machine fm = (Machine)reDistPara.get(0);
			Machine tm = (Machine)reDistPara.get(1);
			Socket fs = new Socket(fm.getIPAddress(),fm.getConnectionListenerPort());					 
			ObjectOutputStream fout = new ObjectOutputStream(fs.getOutputStream());
			ComputePartitionsToMove cpt = new ComputePartitionsToMove((String)reDistPara.get(2),Double.parseDouble((String)reDistPara.get(3)));
		
			Packet pk = new Packet("protocol",cpt);					
			fout.writeObject(pk);
			fout.close();
			fs.close();
			
			// (2) wait until the gotPartitionsToMove has been set
			while (!gotPartitionsToMove){
				Thread.sleep(1000);
			}			
			System.out.println("yes, we got PartitionsToMove message back to DM.");
			
			// (3) send SendDuplicatePartitions message to all machines to reach all split operators.			
			for(Iterator it = config.getMachines(); it.hasNext();){
				Machine m3 = (Machine) it.next();
				Socket s3 = new Socket(m3.getIPAddress(),m3.getConnectionListenerPort());
				 
				ObjectOutputStream out3 = new ObjectOutputStream(s3.getOutputStream());
				SendDuplicatePartitions dp = new SendDuplicatePartitions(partitionsToMove, fm, tm);
				Packet pk3 = new Packet("protocol",dp);	
				out3.writeObject(pk3);
				out3.close();
				s3.close();
			}
			
			// (4) then wait for sender machine to send back "done" message.
			//     this means all to-be-moved partitions are expired 
			while (!oldPartitionsExpired) {
				Thread.sleep(1000);
			}			
			System.out.println("yes, we got oldTupleExpired message back to DM.");
			
			// (5), ok, the final step is to stop sending duplicate to-be-moved partitions in all split operators
			for(Iterator it = config.getMachines(); it.hasNext();){
				Machine m7 = (Machine) it.next();
				Socket s7 = new Socket(m7.getIPAddress(),m7.getConnectionListenerPort());
				 
				ObjectOutputStream out7 = new ObjectOutputStream(s7.getOutputStream());
				StopDuplicatePartitions sp = new StopDuplicatePartitions(partitionsToMove, fm, tm);
				
				Packet pk7 = new Packet("protocol", sp);	
				
				out7.writeObject(pk7);
				out7.close();
				s7.close();
			}
			// debug use, print out the partitions IDs that have been moved
			for (int i=0; i<partitionsToMove.size(); i++){
				System.out.print("P:" + (Integer)partitionsToMove.get(i) + ", ");
			}
			// clear all set flags for the next adaptation.
			gotPartitionsToMove = false;
			partitionsToMove = null;
			oldPartitionsExpired = false;
			
		} catch (Exception e) {
			System.out.println(e.getMessage());
			return false;
		}
		inRedistributing = false;
        System.out.println("\nPT global balancing takes " + (System.currentTimeMillis() - ptstarttime));
		return true;
	} //end of adaptPartitions
	
	public synchronized void increaseLocalMigrationFlag() {
	    this.localMigrationFlag ++;
	    System.out.println("localMigrationFlag = " + localMigrationFlag);
	}
	
	public synchronized void decreaseLocalMigrationFlag() {
	    this.localMigrationFlag --;
	    System.out.println("localMigrationFlag = " + localMigrationFlag);
	}
}