/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.manager;
import java.io.BufferedInputStream;
import java.io.EOFException;
import java.io.FileWriter;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.Iterator;
/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 8, 2003
 * 
 * Insert comments here.
 */
public class ResultReceiver extends Thread implements Runnable {

	/**Stores the reference to the port number*/
	private int port = -1;
	/** The packet pool we will add all incoming tuples to. */
	
	private ArrayList p;
	private ArrayList fileNames;
	private boolean keepGoing;
	private boolean done = false;
	private int totalNumMachines;

	/**
	 * Creates a new Package receiver for the execution controller.
	 */
	public ResultReceiver(int port, int machineCount)
		throws IOException {
		super();
		this.port = port;
		this.p = new ArrayList();
		this.fileNames = new ArrayList();
		keepGoing = true;
		done = false;
		totalNumMachines = machineCount;
	}
	
	public Iterator getFileNames(){
		return fileNames.iterator();
	}
	
	public void addFileName(String filename){
		fileNames.add(filename);
	}
	
	public int getNumFileNames(){
		return fileNames.size();
	}

	/* Runs the Packet Receiver
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		try {
			ServerSocket socket = new ServerSocket(this.port);
			System.out.println("Starting ResultReceiver on port: " + port);
			int i = 0;
			while (keepGoing) {
				Socket inc = socket.accept();
				ResultReceiverHandler tmp =  new ResultReceiverHandler(inc, i, this);
				tmp.setName("ResultReceiverHandler" + i);
				tmp.setDaemon(true);
				tmp.start();
				p.add(tmp);
				i++;
			}
			
		} catch (IOException ioe) {
			System.out.println(
				"Could not start the Result Receiver: " + ioe.getMessage());
		}
	}

	public void requestStop() {

		keepGoing = false;
		interrupt();

	}
	/**
	 * @return
	 */
	public boolean isDone() {
		
		if(p.size() != this.totalNumMachines){
			return false;
		}
		else{
			for(Iterator it = p.iterator(); it.hasNext();){
				ResultReceiverHandler tmp = (ResultReceiverHandler) it.next();
				if(tmp.isRunning()){
					return false;
				}
			}

			return true;
		}
		
		
	}

}
/**
 * 
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 8, 2003
 * 
 * The PacketReceiverHandler class is responsible for
 * maintaining a connection between the local execution engine
 * and a remote engine. The PacketReceiver will spawn one of these threads
 * whenever a Packet from a new location is sent.
 * 
 */
class ResultReceiverHandler extends Thread {

	private Socket incoming;
	
	private int handlerNumber;
	private boolean running;
	private ResultReceiver parent;
	
	public ResultReceiverHandler(Socket i,int handlerNumber, ResultReceiver rr) {
		this.incoming = i;
		this.handlerNumber = handlerNumber;
		this.setPriority(MAX_PRIORITY);
		this.running = true;
		parent = rr;
		
	}
	
	
	public boolean isRunning(){
		return running;
	}

	public synchronized void run() {
		// 	Listen to incoming packets and place them into the Packet Pool.
	
		try {
			String filename = "";
			String temp;
			BufferedInputStream in = new BufferedInputStream(incoming.getInputStream());
			char[] fileread = new char[1];
			char read;
			int r;
			
			// First get the file name
			
			do{
		
				r = in.read();
				fileread[0] = (char) r;		
				temp = new String(fileread);
				
				
				if(!temp.equals("@"))
					filename = filename.concat(temp);
				
			}while(fileread[0] != '@');
			
			filename = filename.concat(".csv");
			parent.addFileName(filename);
			FileWriter fw = new FileWriter(filename);
			// Now get the rest of the file
			r=0;
			while(r != -1){
				
				r = in.read();
				read = (char) r;
				fw.write(read);
				
			}
			
			System.out.println("Received results from " + filename);
			fw.flush();
			fw.close();
			in.close();
			incoming.close();
			this.running = false;
			this.interrupt();
			return;
			
		}catch (EOFException eofe){
			try{
			
			 incoming.close();
			 
			 
			}catch(Exception e){
				System.out.println(e.getMessage());
			}
			 
		}
		catch (SocketException se) {
			System.out.println(
				"Connection Failed: Cannot get stream data from "
					+ incoming.getInetAddress()
					+ ":"
					+ incoming.getPort());
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		
	}

}
