/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.manager;

import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
//import java.io.ObjectOutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.StatisticsTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.adprotocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.protocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.manager.DistributionManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.DistributedMigrationController;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.manager.SpreadfireManager;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 8, 2003
 * 
 * Insert comments here.
 */
public class StatisticsReceiver extends Thread implements Runnable {

	/**Stores the reference to the port number*/
	private int port = -1;
	/** The packet pool we will add all incoming tuples to. */

	private ArrayList p;
	private boolean keepGoing;
	private boolean done = false;
	private StatisticsTable table;
	private Machine[] m;
	
	// -- Bin @ Dec.11,2004, Add the reference to the Distribution Manager for handling adaptation
	// protocols received from DistributedExecutionController via this statistics-receiver
	
	private DistributionManager dm;
	
	// -- Yali @ Aug. 20, 2005. Add reference to distributed migration controller.
	private DistributedMigrationController dmc;
	/**
	 * Creates a new Package receiver for the execution controller.
	 */
	public StatisticsReceiver(int port, StatisticsTable sTable, Machine[] m, DistributionManager d, DistributedMigrationController dmc) throws IOException {
		super();
		this.port = port;
		this.p = new ArrayList();
		keepGoing = true;
		this.m = m;
		table = sTable;
		done = false;
		this.dm = d;
		this.dmc = dmc;
	}

	/* Runs the Packet Receiver
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		try {
			ServerSocket socket = new ServerSocket(this.port);
			System.out.println("Starting StatisticsReceiver on port: " + port);
			int i = 0;
			while (keepGoing) {
				Socket inc = socket.accept();
				StatisticReceiverHandler tmp = new StatisticReceiverHandler(inc, i, m, table, dm, dmc);
				tmp.setName("StatisticsReceiverHandler" + i);
				tmp.setDaemon(true);
				tmp.start();
				p.add(tmp);
				i++;
			}

		} catch (IOException ioe) {
			System.out.println(
				"Could not start the Statistics Receiver: " + ioe.getMessage());
		}
	}

	public void requestStop() {

		for(Iterator i = p.iterator();i.hasNext();){
			StatisticReceiverHandler stop = (StatisticReceiverHandler) i.next();
			stop.requestStop();
		}

		keepGoing = false;
		interrupt();

	}

}
/**
 * 
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 8, 2003
 * 
 * The PacketReceiverHandler class is responsible for
 * maintaining a connection between the local execution engine
 * and a remote engine. The PacketReceiver will spawn one of these threads
 * whenever a Packet from a new location is sent.
 * 
 */
class StatisticReceiverHandler extends Thread {

	private Socket incoming;

	private int handlerNumber;
	private boolean enable;
	private StatisticsReceiver parent;
	private String machineName;
	private Machine[] m;
	private Machine thisMachine;
	private StatisticsTable table;
	private DistributionManager ddm;
	private DistributedMigrationController migrationController;
	
	public StatisticReceiverHandler(Socket i, int handlerNumber, Machine[] m, StatisticsTable table, DistributionManager d, DistributedMigrationController mc) {
		this.incoming = i;
		this.handlerNumber = handlerNumber;
		this.enable = true;
		this.machineName = new String("empty");
		this.table = table;
		this.m = m;
		this.ddm = d;
		this.migrationController = mc;
	}
	
	public void requestStop(){
		this.enable = false;
	}

	public synchronized void run() {

		try {

			ObjectInputStream in = new ObjectInputStream(incoming.getInputStream());

			while (enable) {

				Object o = in.readObject();

				if (o instanceof LightWeightStatistic) {
					LightWeightStatistic lws = (LightWeightStatistic) o;
					//System.out.println("\n\nGet light stats from machine " + this.machineName);
					//System.out.println("Light stats received are:\n" + lws);
					
					if(this.machineName.equals("empty")){
						this.machineName = lws.getMachineName();
						for(int i=0;i<m.length;i++){
							if(m[i].getMachineName().equals(this.machineName))
								thisMachine = m[i];
						}
					}	
					table.put(thisMachine,lws);
					System.out.println("Received lightWeightStats from machine " + this.machineName);
				//end of statistics gathering	
				} 
				
				// partition-level adaptation protocols
				else if (o instanceof Protocol) {
					if (o instanceof PartitionsToMove) {
						// yes, it is PartitionsToMove message
						ddm.setPartitionsToMove(true,((PartitionsToMove)o).getPartitionsToMove());
					} else if (o instanceof Deactivated){
						// it is a Deactivated message
						ddm.setDeactivatedMachine(((Deactivated)o).getDeactivatedMachine());					
					} else if (o instanceof Received) {
						ddm.setReceived(true);
					} else if (o instanceof OldPartitionsExpired) {
					    ddm.setOldPartitionsExpired(true);
					} else if (o instanceof LocalMigrationStarts) {
					    System.out.println("Received localMigrationStarts message from " + ((LocalMigrationStarts)o).getMachineName());
					    ddm.increaseLocalMigrationFlag();
					} else if (o instanceof LocalMigrationFinishes) {
					    System.out.println("Received localMigrationFinishes message from " + ((LocalMigrationFinishes)o).getMachineName());
					    ddm.decreaseLocalMigrationFlag();
					}

				} else if (o instanceof ReplicationStyleProtocol) {
					//Brad M.'s protocols begin here
					if (o instanceof ReplicatedOperatorAssistance)
					{
					    System.out.println("Request to assist operator " + ((ReplicatedOperatorAssistance)o).getOperatorID() + ". Now granting.");
						SpreadfireManager.passToken(ddm.getDistributionTable(), table, m, ((ReplicatedOperatorAssistance)o).getOperatorID(), ((ReplicatedOperatorAssistance)o).getStateNumber());
					}
					else if (o instanceof DeactivateEmptyStateOperator)
					{
						//new function
						SpreadfireManager.deactivateOperator(ddm.getDistributionTable(), m, (DeactivateEmptyStateOperator)o);
					}
					else if (o instanceof SpreadfireEmergencyShutdown)
					{
					    
					    System.out.println("Emergency Shutdown request received.  Now Granting...");
						ddm.terminateExecutionNow();
					}				    
				}
				// operator-level across-machine plan migration protocols
				// process messages sent back from machine to distribution manager 
				// (usually ack messages indicating that certain task has been done by that machine).
				else if (o instanceof MigrationProtocol) {
					
					// sync timestamp received from one machine
					if (o instanceof SyncTimestampObtained) {
						long synctime = ((SyncTimestampObtained)o).getMaxTimestamp();
						System.out.println("DM stats: received sync timestamp from one machine: "+synctime);
						this.ddm.addTimeStampReceived(synctime);
						
					} 
					// all-synced ack received from the root operator of the subplan to be replaced.
					else if (o instanceof ExecutionSynced) {
						this.ddm.allExecutionSynced = true;
						System.out.println("DM stats: received all sync ack from root op.");
						
					} 
					else if (o instanceof PlanChanged) {
						this.ddm.planChangedFlagReceived++;
						System.out.println("DM stats: received planChanged ack from one machine");
						
					} 
					else if(o instanceof StatesFilled) {
						this.ddm.statesFilledFlagReceived++;
						System.out.println("DM stats: received statesfilled ack from one machine");
					}
					else if (o instanceof AllStatesMoved) {
						
					} 
					else if (o instanceof AllStatesRecomputed) {
						
					}
					
				}
				
			} //end of while
			
			return;

		} catch (EOFException eofe) {
			try {

				incoming.close();

			} catch (Exception e) {
				System.out.println(e.getMessage());
			}

		} catch (SocketException se) {
			System.out.println("Connection Failed: Cannot get stream data from "+ incoming.getInetAddress()+ ":"+ incoming.getPort());
		} catch (java.io.StreamCorruptedException se) {
            System.err.println("StreamCorruptedException: " + se.getLocalizedMessage());
        
		    se.printStackTrace();
            if (se.getCause() != null)
                se.getCause().printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

}
