/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller;

//import java.util.ArrayList;
//import java.util.ArrayList;
//import java.util.Enumeration;
//import java.util.Hashtable;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.realjoinoperator.WBJoinOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WEquiJoinStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.ProbabilisticWindowJoinOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.manager.DistributionManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;

/**
 * @author Yali Zhu
 *
 */
public class DistributedOptimizer {
	
	DistributionManager dm;
	public DistributedOptimizer(DistributionManager dm) {
		this.dm = dm;
	}
	
	public NodePairToSwitch optimizeQueryPlan(QueryPlan qp, DistributionTable dt) {
	    //ArrayList pair = new ArrayList();
	    //OperatorPairToSwitch nodepair = new OperatorPairToSwitch();
	    
		int startQID;
		XATDag queryplan = null;
		
		try{
			queryplan = qp.getQueryPlan();	
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(-1);
		}
		
		XATNode[] nodes = queryplan.convertToArray();
		
		//for WBJoinOperator 
		for(int i=0; i<nodes.length; i++) {
		    if(nodes[i].getOperator() instanceof XATMultiQueueStreamOperatorImp) {
		        XATNode[] cnodes = nodes[i].getOpChildren();
		
		        //find the first pair and switch the two joins
		        for(int j=0; j<cnodes.length; j++) {
		            if(cnodes[j].getOperator() instanceof XATMultiQueueStreamOperatorImp) {
		                Machine highmachine = dt.getMachine(nodes[i]);
		                Machine lowmachine = dt.getMachine(cnodes[j]);
				    
		                if(highmachine != lowmachine) {
		                    System.out.println("Found a node pair on different machines to switch!");
		                    NodePairToSwitch nodepair = new NodePairToSwitch(nodes[i], cnodes[j], highmachine, lowmachine);
		                    
		                    if(j == 0) {
		                        nodepair.setIsLeftChild(true);
                                System.out.println("Switch " + nodes[i] + " with left child " + cnodes[j]);
		                    } else {
		                        nodepair.setIsLeftChild(false);
                                System.out.println("Switch " + nodes[i] + " with right child " + cnodes[j]);
		                    }
		                    
		                    //switch the two operators
		                    WBJoinOperator upperop = (WBJoinOperator)nodes[i].getStreamOperator();
		                    WBJoinOperator lowerop = (WBJoinOperator)nodes[i].getStreamOperator();
		                    this.SwitchJoins(upperop, lowerop, queryplan);
				        
		                    //update distribution table, switch the machines the two nodes are distributed to.
		                    dt.remove(nodes[i]);
		                    dt.remove(cnodes[j]);
		                    dt.add(nodes[i], lowmachine);
		                    dt.add(cnodes[j], highmachine);
		                    return nodepair;
		                }
		            }
		        }
		    }
		}
	    return null;
	}
	
	public SubPlan optimizeQueryPlan(QueryPlan qp) {
		
		int startQID;
		XATDag queryplan = null;
		
		try{
			queryplan = qp.getQueryPlan();	
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(-1);
		}

		// switch join operators with lower selectivity
		// In the first demo version, find one change (across-machine) and 
		// the optimization process is then finished.
		SubPlan subplan = ApplySwitchJoinHeuristics(queryplan);
		return subplan;
	}
	
	private SubPlan ApplySwitchJoinHeuristics(XATDag dag) {
		SubPlan subplan;
		
		// for demo: find two consecutive joins with the largest selectivity difference,
		// and the lower join's selectivity is larger than the higher join's selectivity.
		XATNode[] nodes = dag.convertToArray();
		int upperjoinindex = 0;
		int lowerjoincindex = 0;
		double diff = 0;
		boolean switchfound = false;
		java.util.Random rand = new java.util.Random();
		
		for(int i=0; i<nodes.length; i++) {
			//for ProbabilisticWindowJoinOperator
			if(nodes[i].getOperator() instanceof ProbabilisticWindowJoinOperatorImp) {
				XATNode[] cnodes = nodes[i].getOpChildren();
			
				for(int j=0; j<cnodes.length; j++) {
					if(cnodes[j].getOperator() instanceof ProbabilisticWindowJoinOperatorImp) {
						double s1 = Double.parseDouble(nodes[i].getOperator().getProperty("SELECTIVITY"));
						double s2 = Double.parseDouble(cnodes[j].getOperator().getProperty("SELECTIVITY"));
						//if((s2 -s1) > diff) {
						//	upperjoinindex = i;
						//	lowerjoincindex = j;
						//  diff = s2 - s1;
						//}
						// used for demo, randomly choose a pair to switch
						double d = rand.nextDouble();
						if(d > diff) {
							upperjoinindex = i;
							lowerjoincindex = j;
							diff = d;
							switchfound = true;
						}
					}
				}
			}
 		}
		
		if(switchfound) {
			System.out.println("Found a join pair that can be switched: upperOp " + upperjoinindex + " lowerOp: " + lowerjoincindex);
			XATStreamOperator upperjoin = (XATStreamOperator)nodes[upperjoinindex].getOperator();
			XATNode[] cnodes = nodes[upperjoinindex].getOpChildren();
			XATStreamOperator lowerjoin = (XATStreamOperator)cnodes[lowerjoincindex].getOperator();
			
			// add output queues to subplan
			subplan = new SubPlan();
			int outqid = upperjoin.getOutputQueue().getQueueID();
			subplan.addOutputQID(Integer.toString(outqid));
			
			// add input queues connecting to lower operator to subplan
			Iterator inq = lowerjoin.getInputQueues();
			while(inq.hasNext()) {
				int id = ((XATQueue)inq.next()).getQueueID();
				subplan.addInputQID(Integer.toString(id));
			}
			// add input queues connecting to upper operator to subplan
			outqid = lowerjoin.getOutputQueue().getQueueID();
			inq = upperjoin.getInputQueues();
			while(inq.hasNext()) {
				int id = ((XATQueue)inq.next()).getQueueID();
				if(id != outqid)
					subplan.addInputQID(Integer.toString(id));
			}
			
			
			subplan.addOperator(upperjoin);
			subplan.addOperator(lowerjoin);
			subplan.addRootOperator(lowerjoin);
			subplan.setOldRootOperator(upperjoin);
			subplan.setOldRootOperatorID(upperjoin.getObjectID());
			upperjoin.setActive(false);
			lowerjoin.setActive(false);
			upperjoin.setIsSubPlanOperator(true);
			lowerjoin.setIsSubPlanOperator(true);
			
			// switch the join pair
			SwitchJoins((ProbabilisticWindowJoinOperatorImp)upperjoin, (ProbabilisticWindowJoinOperatorImp)lowerjoin, dag, subplan);
		}
		else {
			System.out.println("Query plan is unchanged.");
			return null;
		}
		
		return subplan;
	}
	
	/**
	 * @param upperjoin
	 * @param lowerjoin
	 * @param dag
	 * This method switches two probablistic window join operators. 
	 * It assumes that the join operator stores one state for each input queue.
	 * Also assumes the join is a binary join.
	 */
	private void SwitchJoins(ProbabilisticWindowJoinOperatorImp upperjoin, ProbabilisticWindowJoinOperatorImp lowerjoin, XATDag dag, SubPlan subplan) {
		// two things to each for each node: parent, children
		// four things to change for each operator: input queue, output queue and 
		// state schema for window join operator (run init) again to achieve this goal,
		// this will take care of other things that need to be re-initialized.
		
		// indicates if lowerNode is the right child of upperNode.
		boolean isRightChild = true; 
		
		// first change the input and output queues of involved operators
		if(upperjoin.getNthInputQueue(0).equals(lowerjoin.getOutputQueue())) {
			isRightChild = false; 
		}
		System.out.println("lower is upper's right child: "+isRightChild);
			
		// reserve useful queue references
		XATQueue lowerLeftInputQueue = lowerjoin.getLeftInputQueue();
		XATQueue lowerRightInputQueue = lowerjoin.getRightInputQueue();
		XATQueue upperLeftInputQueue = upperjoin.getLeftInputQueue();
		XATQueue upperRightInputQueue = upperjoin.getRightInputQueue();
		XATQueue lowerOutputQueue = lowerjoin.getOutputQueue();
		XATQueue upperOutputQueue = upperjoin.getOutputQueue();
		
		// reserve useful node references
		XATNode lowerNode = lowerjoin.getNode();
		XATNode upperNode = upperjoin.getNode();
		
		Object lowerLeftChild = lowerNode.getChild(0);
		Object lowerRightChild = lowerNode.getChild(1);
		
		/* first take care of queue connnections */
		if(isRightChild) {
			
			upperRightInputQueue.unregister(upperjoin);

			lowerjoin.setOutputQueue(upperOutputQueue);
			lowerjoin.replaceInputQueue(lowerLeftInputQueue, upperRightInputQueue);
			upperjoin.replaceInputQueue(upperRightInputQueue, lowerLeftInputQueue);
			upperjoin.setOutputQueue(lowerOutputQueue);
			//lowerOutputQueue.unregister(upperjoin);
		}
		else {
			upperLeftInputQueue.unregister(upperjoin);
			//lowerRightInputQueue.unregister(lowerjoin);
			
			lowerjoin.setOutputQueue(upperOutputQueue);
			lowerjoin.replaceInputQueue(lowerRightInputQueue, upperLeftInputQueue);
			upperjoin.replaceInputQueue(upperLeftInputQueue, lowerRightInputQueue);
			upperjoin.setOutputQueue(lowerOutputQueue);
		}

		/* then take care of the parent-child relation of involved XATNodes */
		XATNode upperParentNode = upperNode.getParent();
		
		if (isRightChild) {
			lowerNode.setParent(upperParentNode);
			lowerNode.addOpChild(upperNode, 0);
			lowerNode.deleteChild(lowerLeftChild);

			upperNode.deleteOpChild(lowerNode);
			upperNode.appendChild(lowerLeftChild);
			upperNode.setParent(lowerNode);

			if (lowerLeftChild instanceof XATNode) {
				((XATNode)lowerLeftChild).setParent(upperNode);
			}

			// take care of stateID
			int maxstateid = this.dm.getConfig().getCurrentMaxStateID();
			upperjoin.getRightState().setStateID(lowerjoin.getLeftState().getStateID());
		    maxstateid++;
		    subplan.setStateLocation(Integer.toString(lowerjoin.getLeftState().getStateID()), lowerjoin);
			lowerjoin.getLeftState().setStateID(maxstateid);
			maxstateid++;
			this.dm.getConfig().setCurrentMaxStateID(maxstateid);
			
			upperjoin.getRightState().setMatched(true);
			upperjoin.getLeftState().setMatched(true);
			lowerjoin.getRightState().setMatched(true);
			lowerjoin.getLeftState().setMatched(false);
			lowerjoin.getLeftState().addJoiningSubStates(upperjoin.getLeftState().getStateID());
			lowerjoin.getLeftState().addJoiningSubStates(upperjoin.getRightState().getStateID());
		}
		else {
			lowerNode.setParent(upperParentNode);
			lowerNode.addOpChild(upperNode);
			lowerNode.deleteChild(lowerRightChild);
				
			upperNode.deleteOpChild(lowerNode);
		    upperNode.insertChild(lowerRightChild, 0);
			upperNode.setParent(lowerNode);
			
			if(lowerRightChild instanceof XATNode) {
				((XATNode)lowerRightChild).setParent(upperNode);
			}
			
			// take care of stateID -- no states in probablisticjoin operators.
			int maxstateid = this.dm.getConfig().getCurrentMaxStateID();
			upperjoin.getLeftState().setStateID(lowerjoin.getRightState().getStateID());
			maxstateid++;
			subplan.setStateLocation(Integer.toString(lowerjoin.getRightState().getStateID()), lowerjoin);
			lowerjoin.getRightState().setStateID(maxstateid);
			maxstateid++;
			this.dm.getConfig().setCurrentMaxStateID(maxstateid);
			
			upperjoin.getRightState().setMatched(true);
			upperjoin.getLeftState().setMatched(true);
			lowerjoin.getLeftState().setMatched(true);
			lowerjoin.getRightState().setMatched(false);
			lowerjoin.getRightState().addJoiningSubStates(upperjoin.getLeftState().getStateID());
			lowerjoin.getRightState().addJoiningSubStates(upperjoin.getRightState().getStateID());
		}
			
		if (upperParentNode != null) {
			int position = upperParentNode.getIndexOfOpChild(upperNode);
			upperParentNode.addOpChild(lowerNode, position);
			upperParentNode.deleteOpChild(upperNode);
			//upperNode.setParent(lowerNode);
		} else {
			dag.setRoot(lowerNode);
		}

		// as the last step. reinitialize these two operators, 
		// starting from the new lower op (orig upper op), MUST FOLLOW THIS ORDER!
		upperjoin.init();
		lowerjoin.init();
		return;
	}
	
	/**
	 * @param upperjoin
	 * @param lowerjoin
	 * @param dag
	 * This method switches two window binary join operators. 
	 * It assumes that the join operator stores one state for each input queue.
	 * Also assumes the join is a binary join.
	 */
	private boolean SwitchJoins(WBJoinOperator upperJoin, WBJoinOperator lowerJoin, XATDag dag) {

		boolean isRightChild = true; // indicates if lowerNode is the right child of upperNode.
		
		/** first change the input and output queues of involved operators */
		if(upperJoin.getNthInputQueue(0).equals(lowerJoin.getOutputQueue())) {
			isRightChild = false; 
			System.out.println("lower is upper's left child.");
		} else {
		    System.out.println("lower is upper's right child");
		}
			
		XATQueue lowerLeftInputQueue = lowerJoin.getLeftInputQueue();
		XATQueue lowerRightInputQueue = lowerJoin.getRightInputQueue();
		XATQueue upperLeftInputQueue = upperJoin.getLeftInputQueue();
		XATQueue upperRightInputQueue = upperJoin.getRightInputQueue();
		XATQueue lowerOutputQueue = lowerJoin.getOutputQueue();
		XATQueue upperOutputQueue = upperJoin.getOutputQueue();
		
		
		XATNode lowerLeftChild;
		XATNode lowerRightChild;
		
		if(lowerJoin.getNode().getOpChildrenSize() == 0) {
			lowerLeftChild = null;
			lowerRightChild = null;
		} else if(lowerJoin.getNode().getOpChildrenSize() == 1)  {
			if (lowerJoin.getNthInputQueue(0).equals(lowerJoin.getNode().getOpChild(0).getStreamOperator().getOutputQueue())) {

				lowerLeftChild = lowerJoin.getNode().getOpChild(0);
				lowerRightChild = null;
			}
			else {
				lowerLeftChild = null;
				lowerRightChild = lowerJoin.getNode().getOpChild(0);
			}
		} else if (lowerJoin.getNode().getOpChildrenSize() == 2) {
			lowerLeftChild = lowerJoin.getNode().getOpChild(0);
			lowerRightChild = lowerJoin.getNode().getOpChild(1);
		} else {
			System.out.println("Binary join can only have at most two children. ");
			return false;
		}
		
		/* first take care of queue connnections */
		if(isRightChild) {
			
			upperRightInputQueue.unregister(upperJoin);

			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerLeftInputQueue, upperRightInputQueue);
			upperJoin.replaceInputQueue(upperRightInputQueue, lowerLeftInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
			lowerOutputQueue.unregister(upperJoin);
		}
		else {
			upperLeftInputQueue.unregister(upperJoin);
			//lowerRightInputQueue.unregister(lowerJoin);
			
			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerRightInputQueue, upperLeftInputQueue);
			upperJoin.replaceInputQueue(upperLeftInputQueue, lowerRightInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
		}

		/* then take care of the parent-child relation of involved XATNodes */
		XATNode upperParentNode = upperJoin.getNode().getParent();
		
		if(isRightChild) {
				lowerJoin.getNode().setParent(upperParentNode);
				lowerJoin.getNode().addOpChild(upperJoin.getNode(), 0);
				lowerJoin.getNode().deleteOpChild(lowerLeftChild);
				
				upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
				upperJoin.getNode().addOpChild(lowerLeftChild);
				upperJoin.getNode().setParent(lowerJoin.getNode());
				
				if(lowerLeftChild != null) {
					lowerLeftChild.setParent(upperJoin.getNode());
				}		
		}
		else {
			lowerJoin.getNode().setParent(upperParentNode);
			lowerJoin.getNode().addOpChild(upperJoin.getNode());
			lowerJoin.getNode().deleteOpChild(lowerRightChild);
				
			upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
			upperJoin.getNode().addOpChild(lowerRightChild, 0);
			upperJoin.getNode().setParent(lowerJoin.getNode());
			
			if(lowerRightChild != null) {
				lowerRightChild.setParent(upperJoin.getNode());
			}	
		}
			
		if (upperParentNode != null) {
			int position = upperParentNode.getIndexOfOpChild(upperJoin.getNode());
			upperParentNode.addOpChild(lowerJoin.getNode(), position);
			upperParentNode.deleteOpChild(upperJoin.getNode());
			upperJoin.getNode().setParent(lowerJoin.getNode());
		} else {
			dag.setRoot(lowerJoin.getNode());
		}
		

		// as the last step. reinitialize these two operators, 
		// starting from the new lower op (orig upper op), MUST FOLLOW THIS ORDER!
		upperJoin.init();
		lowerJoin.init();
		
		String lname = upperJoin.getInputQueue(0).getSchema().getName();
		String rname = upperJoin.getInputQueue(1).getSchema().getName();
		String newname = "#" + lname + "#" + rname;
		XATQueueSchema outSchema = upperJoin.getOutputQueue().getSchema();
		outSchema.setName(newname);
		
//		System.out.println("Input queue schema for operator: " + upperJoin);
//		System.out.print("{");
//		for(int i=0; i<upperJoin.getInputQueueCount(); i++) {
//		    System.out.print(upperJoin.getInputQueue(i).getSchema().getName() + ", ");
//		}
//		System.out.println("}");
//		
//		System.out.println("Input queue schema for operator: " + lowerJoin);
//		System.out.print("{");
//		for(int i=0; i<lowerJoin.getInputQueueCount(); i++) {
//		    System.out.print(lowerJoin.getInputQueue(i).getSchema().getName() + ", ");
//		}
//		System.out.println("}");
		
		return true;
	}
	
}
