/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet;

import java.io.Serializable;
/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Jan 30, 2004
 * 
 * The network packet class is simply an array of Packet objects to send
 * across the network.
 */
public class NetworkPacket implements Serializable {
	private int cursor;

	private Packet[] packets;
	private int size;
	
	/**
	 * Default constructor, a network packet of size 1000 will be created
	 */
	public NetworkPacket() {
		this.size = 100;
		this.cursor = 0;
		packets = new Packet[size];
	}
	
	/**
	 * Default constructor
	 * 
	 * @param size -The number of Packets this object can hold.
	 */
	
	public NetworkPacket(int size){
		
		this.size = size;
		this.cursor = 0;
		packets = new Packet[this.size];
	}
	
	/**
	 * Adds a packet to the Network Packet.
	 * 
	 * @param p - The packet to add.
	 * @throws ArrayIndexOutOfBoundsException
	 */
	public void addPacket(Packet p) throws ArrayIndexOutOfBoundsException{
		
		packets[cursor] = p;
		this.cursor++;
	}
	
	/**
	 * Gets a packet from the Network Packet from the specific location.
	 * 
	 * @param location -The location of the packet we wish to get.
	 * @return The packet from the location.
	 */
	public Packet getPacket(int location){
		return packets[location];
	}
	
	
	/**
	 * Returns the size of the network packet.
	 * @return The size of the network packet.
	 */
	public int size(){
		return this.cursor;
	}
	
	public int totalSize(){
		return this.size;
	}
	
}
