/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet;

import java.util.ArrayList;

/**
 * @author
 * 
 * This class holds the packets that are receieved but not yet enqueued
 */

public class PacketPool {
	/**Stores the reference to the array list storing the packets*/
	ArrayList packets = null;
	
	/**
	 * Constructs a packet pool
	 */
	public PacketPool() {
		this.packets = new ArrayList();		
	}
	
	/**
	 * Adds the packet to the packet pool
	 * @param p
	 */
	public synchronized void addPacket(Packet p)
	{
		packets.add(p);
		if(packets.size() == 1)
			notify();
	}
	
	/**
	 * Removes the 1st packet from the packet pool and returns it
	 * @return Packet
	 * @throws InterruptedException
	 */
	public synchronized Packet removePacket() throws InterruptedException
	{
		if(packets.size() == 0)
		{
			wait();
		}
		return((Packet)packets.remove(0));
	}
	
	public synchronized int size(){
		return packets.size();
	}
	
	public String toString() {
		return this.packets.toString();
	}
}
