/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet;

import java.io.NotSerializableException;
import java.io.Serializable;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.Hashtable;
import java.io.ObjectOutputStream;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.spreadfire.SpreadfireOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ConnectionTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.StreamSplitOperatorImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATMemoryQueueImp;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 15, 2003
 */
public class TupleDistributor implements Runnable, Serializable {

	private ConnectionTable table;
	private boolean enable;
	private Machine machine; // to hold the current machine that the TupleDistributor is running
    int MAX_NP_SIZE = 2000;
    
	/**
	 * @param port
	 * @param receivePool
	 * @param tree
	 * @throws IOException
	 */
	public TupleDistributor(ConnectionTable ct, Machine m){
		this.table = ct;	
		this.enable = true;	
		this.machine = m;
	}
	
	public void run(){

		XATStreamOperator op = null;
		Packet p = null;
		NetworkPacket np = null;
		XATTuple[] tuples;
		Enumeration e;
		int queueSize;
		int dQ;
		XATQueueSchema xqs;
		
		while(enable){
		
			try{
				// Scan through the operators that are supposed to output and
				// take the tuples out of their queues, turn them into packets, and
				// put them into the tuple pool				
				
				// Get the current keys in the table
				// Remember that the table will change so we will need to get
				// the keys after every iteration to get the changes.
				
				e = table.getKeys();
				
				while(e.hasMoreElements()){
				
					//Try to fetch the value, if it cannot be fetched
					//It may have been removed by another object, simply disregard it.
					try{
						op = (XATStreamOperator) e.nextElement();
						if(op.isActive()) {
						// -- Bin @ Dec.1, 2004, to handle partitioned operator						
						Object str = table.getConnectionObject(op);
						if (str instanceof Hashtable) {
							Hashtable streams = (Hashtable)str;
							// we now know that this must be a StreamSplitOperatorImp, then cast the operator
							// distribute all tuples in the queue except the one that goes to the same machine 
							Iterator iq = op.getOutputQueues();
							while (iq.hasNext()) {
								XATQueue q = (XATQueue)iq.next();
								if (q.getQueueName().equals(machine.getIPAddress())){
									// in the same machine, we do not have to send it via network
								} else {
									queueSize = q.getQueueSize(this);									
									if (!q.isRegistered(this)) {
										q.register(this);
									}
									
									if(queueSize > 0){	
										dQ = (int) ((queueSize<MAX_NP_SIZE)?queueSize:MAX_NP_SIZE);
										tuples = q.dequeue(dQ, this);
										np = new NetworkPacket(dQ+1);
										
										// Put the schema in., some odd errors may happen, xqs is null?
										// then set schema of q as op.getOutputQueue().getSchema										
										xqs = q.getSchema();
										if (xqs==null){
											xqs = op.getOutputQueue().getSchema();
											if (xqs!=null) {
												q.setSchema(xqs);
												System.out.println("NULL SCHEMA AGAIN!!!!, Now Reset.");
											}
										}
										p = new Packet(op.getProperty("OBJECT_ID"),xqs,"schema");
										np.addPacket(p);
										
										// Add each tuple to a network packet
                                        long starttime = System.currentTimeMillis();
										for(int i=0; i< tuples.length; i++){
											p = new Packet(op.getProperty("OBJECT_ID"),tuples[i],"");
											np.addPacket(p);
										}
                                        //System.out.println("\nMaking an np of size " + tuples.length + " for " + op.getProperty("OBJECT_ID") + " takes " + (System.currentTimeMillis() - starttime) + "ms");
                                        starttime = System.currentTimeMillis();
										//output the packet
										if (((XATStreamOperator)q.getQueueOwner().get(0)) instanceof SpreadfireOperator)
										{
											Enumeration en = streams.elements();
											while (en.hasMoreElements())
											{
												ObjectOutputStream currentStream = (ObjectOutputStream)en.nextElement();
												currentStream.reset();
												currentStream.writeObject(np);
											}
										}
										else
										{
											((ObjectOutputStream)streams.get(q.getQueueName())).reset();
											((ObjectOutputStream)streams.get(q.getQueueName())).writeObject(np);
										}
										
                                        //System.out.println("Sending the np of size " + tuples.length + " takes " + (System.currentTimeMillis() - starttime) + "ms");
                                        Thread.yield();
                                        //try{
                                        //    Thread.sleep(1);
                                        //} catch(Exception te) {
                                        //    te.printStackTrace();
                                        //}
                                       
										tuples = null; np =null;
									//	System.gc();
									}																														
								} //end
							} //end while iq has next
							
							streams = null;
							// *** end of partitioned operator
						} else {
							
							if(op.getOutputQueue() != null) {
							queueSize = op.getOutputQueue().getQueueSize(this);
							
							// Register to remove stuff from the output queue
							if(!op.getOutputQueue().isRegistered(this)){
								op.getOutputQueue().register(this);
							}
							
							if(queueSize > 0){
								
								//dQ = (int) (queueSize);
                                dQ = (int) ((queueSize<MAX_NP_SIZE)?queueSize:MAX_NP_SIZE);
                                
								//System.out.println(">>>>>>>>>>>>>>>>>>>>>>Tuples Get Dequeued:" + dQ + ">>>>>>>>>>>>");
								//((XATMemoryQueueImp)op.getOutputQueue()).printOutCursorInfo();
								//System.out.println(">>>>>>>>>>>>This" + this.toString());
								
								tuples = op.getOutputQueue().dequeue(dQ, this);
								np = new NetworkPacket(dQ+1);
								
								// Put the schema in.
								
								xqs = op.getOutputQueue().getSchema();
								p = new Packet(op.getProperty("OBJECT_ID"),xqs,"schema");
								np.addPacket(p);
								
								// Add each tuple to a network packet
                                long starttime = System.currentTimeMillis();
								for(int i=0; i< tuples.length; i++){
									p = new Packet(op.getProperty("OBJECT_ID"),tuples[i],"");
									np.addPacket(p);
								}
                                //System.out.println("\nMaking an np of size " + tuples.length + " for " + op.getProperty("OBJECT_ID") + " takes " + (System.currentTimeMillis() - starttime) + "ms");                               

                                starttime = System.currentTimeMillis();
								table.getConnection(op).reset();
								table.getConnection(op).writeObject(np);                               
                                //System.out.println("Sending the np of size " + tuples.length + " takes " + (System.currentTimeMillis() - starttime) + "ms");
                                Thread.yield();
                                
                                //try{
                                //    Thread.sleep(1);
                                //} catch(Exception te) {
                                //    te.printStackTrace();
                                //}
							}							
							}
						} //end if str Hashtable/OutputStream
						} // end if op.isActive true
						
					} catch (NullPointerException ne) {
						System.out.println("The key " + op.toString() + " no longer exists in the output table.");
						ne.printStackTrace();
						continue;
					} catch (IllegalArgumentException iag) {
						System.out.println("============= IllegalArgumentException ============");
						System.out.println(iag.getLocalizedMessage());
						System.out.println(iag.getMessage());
						System.out.println(op.toString());
						System.out.println(op.getOutputQueue().getSchema());
						iag.printStackTrace();
						System.out.println("====================================");
					}
				}
			} 
			catch(XATQueueException xqe)
			{
				System.out.println("============= XATQueueException ============");
				System.out.println(op);
				xqe.printStackTrace();
				System.out.println(xqe.getStackTrace());
				System.out.println("====================================");
			} 
			catch(NotSerializableException nse)
			{
				System.out.println("===== NotSerializableException =====");
				System.out.println(nse.getMessage());
				nse.printStackTrace();
				System.out.println(nse.getStackTrace());
				System.out.println("====================================");				
			} 
			catch(Exception ex)
			{
				System.out.println("============= Exception ============");
			 	ex.printStackTrace();
			 	System.out.println("Message: " + ex.getMessage());
			 	System.out.println("Stack Trace: " + ex.getStackTrace().toString());
				System.out.println("====================================");
			}
		}
	}
	
	public void requestStop(){
		this.enable = false;
	}
}
