/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.*;
//import java.io.FileWriter;
/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 15, 2003
 * 
 * The packet feeder takes tuples from the packet pool and places them
 * at the input queues of the operator that is to process it.
 * 
 */
public class TupleFeeder implements Runnable {

	private PacketPool pool;
	private ConnectionTable table;
	private boolean enable;
	
	/**
	 * @param port
	 * @param receivePool
	 * @param tree
	 * @throws IOException
	 */
	public TupleFeeder(PacketPool receivePool, ConnectionTable ct){
		this.pool = receivePool;
		this.table = ct;	
		this.enable = true;	
	}
	
	public synchronized void run()
	{
		int count = 0;
		XATStreamOperator op;
		String from;
		//File f;
		//FileWriter outWriter = null;
		
		//try{
		
		//f = new File("fed_into_input_queues.txt");
		//outWriter = new FileWriter(f);
		//outWriter.write("start it up...\n");
		//}catch (Exception e){}
		
		while(enable)
		{
			Packet p = null;
			OperatorQueue oq;
			try{
				//while(pool.size() < 200)
				//	this.wait();
				
				// get the packet
				p = pool.removePacket();
				//outWriter.write(p+"\n");
				// find out where this packet needs to go.
				from = p.getIdentifier(); 

				// Op will be the operator to which this tuple should go.
				oq = table.getConnection(from);
				op = null;
				
				// TODO this is temp solution for demo, need to fix later
				if(oq != null)
					op = oq.getOperator();				
				
				if(op != null)
				{				
					if(p.getObject() instanceof XATQueueSchema) 
					{
					    if(p.getIdentifier().matches(".*[a-zA-Z].*"))
					    {
					    	System.out.println("Received XATQueueSchema with signature " + p.getIdentifier());
					    }
						op.getInputQueue(oq.getQueueNum()).setSchema((XATQueueSchema)p.getObject());
					} 
					else if(p.getObject() instanceof ReplaceInConnectionEntry) 
					{
                        ReplaceInConnectionEntry entry = (ReplaceInConnectionEntry)p.getObject();
                        String oldop = entry.getOldOperatorID();
                        String newop = entry.getNewOperatorID();
                        OperatorQueue newoq = table.getConnection(oldop);
                        if(newoq != null) 
                        {
                            table.removeConnection(oldop);
                            table.addConnection(newop, newoq);
                            System.out.println("\n\nReplace in-connection entry from: " + oldop + " to " + newop);
                        }        
                        System.out.println("\n\nIn-connection Table: \n" + table);
                    }
					else
					{
						// Add the tuple to the input queue
						// Make sure to add it to the right queue!!!
						/*
						try {
							FileWriter out = new FileWriter("tuples.txt", true);
							out.write("Putting tuple \n    " + ((XATTuple)p.getObject()).toString() + "\nin queue \n    " + op.getInputQueue(oq.getQueueNum()) + "\n");
							out.close();
						} catch (Exception ex) {
							ex.printStackTrace();
						}*/
						
						// TODO this is temp solution for demo, need to fix later
						if(op.isActive() && op.getInputQueue(oq.getQueueNum()) != null) 
						{
						    try 
						    { 
                                Integer.parseInt(from);
                                if (((XATTuple)p.getObject()).getObjTypeID() == 208)        
                                {
                                    System.out.println("Received a command for operator " + op.getOperatorID() + " which will now be queued behind " + op.getInputQueue(oq.getQueueNum()).getQueueSize(op) + " tuples.");
                                }                                
						    }                             
						    catch (NumberFormatException nfe) 
						    {
                                 //nfe.printStackTrace();
                             }
						    
                             XATTuple nexttuple = (XATTuple) p.getObject();
                             op.getInputQueue(oq.getQueueNum()).enqueue(nexttuple, null);
                             
                             count++;
                             
                             //if(nexttuple.isNew() && nexttuple.isSingletonTuple()) {
//                             if(nexttuple.isSingletonTuple()) {
//                                 System.out.println("Pool size: " +  pool.size());
//                                 System.out.println("Feed tuple to queue " + oq.getQueueNum() + " of " + op + ": "+ nexttuple);                                
//                             }
                            
						} else {
							//otherwise put the packet back into the pool for future processing
							pool.addPacket(p);
						}
					}
				}
			} catch(NullPointerException npe){
				System.out.println(npe.getMessage());
				npe.printStackTrace();
			}
			catch(Exception e){
				e.printStackTrace();
			}
		}
		
		System.out.println("# of tuples received: " + count);
		
		return;
	}
	
	public void requestStop(){
		this.enable = false;
	}
	
}
