/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet;
import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
//import java.io.*;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.ArrayList;
/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 8, 2003
 * 
 * Insert comments here.
 */
public class TupleReceiver extends Thread implements Runnable {

	/**Stores the reference to the port number*/
	private int port = -1;
	/** The packet pool we will add all incoming tuples to. */
	private PacketPool receivePool = null;
	private ArrayList tree;
	private ArrayList p;
	private boolean keepGoing;

	/**
	 * Creates a new Package receiver for the execution controller.
	 */
	public TupleReceiver(int port, PacketPool receivePool, ArrayList tree) throws IOException {
		super();
		this.port = port;
		this.receivePool = receivePool;
		this.tree = tree;
		this.p = new ArrayList();
		keepGoing = true;
	}

	/* Runs the Packet Receiver
	 * @see java.lang.Runnable#run()
	 */
	public synchronized void run() 
	{
		try 
		{
			ServerSocket socket = new ServerSocket(this.port);
			System.out.println("Starting TupleReceiver on port: " + port);
			int i = 0;
			while (keepGoing) 
			{
				Socket inc = socket.accept();
				
				PacketReceiverHandler tmp =  new PacketReceiverHandler(inc, receivePool, i);
				tmp.setName("PacketReceiverHandler" + i);
				tmp.setDaemon(true);
				tmp.start();
				p.add(tmp);
				System.out.println("P-ArrayList Size: " + p.size() + "Buffer Size:" + inc.getReceiveBufferSize());
				i++;
			}
		} 
		catch (IOException ioe) 
		{
			System.out.println("Could not start the Tuple Receiver: " + ioe.getMessage());
		}
	}

	public void requestStop() 
	{
		System.out.println("Trying to interrupt the threads");

		//Request that each sub-thread is stopped.
		for (int j = 0; j < p.size(); j++) 
		{
			
			PacketReceiverHandler stopHand = (PacketReceiverHandler) p.get(j);
			
			stopHand.requestStop();
		}

		keepGoing = false;
		interrupt();

	}
}
/**
 * 
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 8, 2003
 * 
 * The PacketReceiverHandler class is responsible for
 * maintaining a connection between the local execution engine
 * and a remote engine. The PacketReceiver will spawn one of these threads
 * whenever a Packet from a new location is sent.
 * 
 */
class PacketReceiverHandler extends Thread {

	private Socket incoming;
	private PacketPool pool;
	private int handlerNumber;
	private boolean keepGoing;

	public PacketReceiverHandler(Socket i, PacketPool p, int handlerNumber) {
		this.incoming = i;
		this.pool = p;
		this.handlerNumber = handlerNumber;
		this.setPriority(MAX_PRIORITY);
		this.keepGoing = true;
	}

	public synchronized void run() 
	{
		// 	Listen to incoming packets and place them into the Packet Pool.
		Packet p = null;
		Object o = null;
		NetworkPacket np = null;

		try 
		{		
			//File f = new File("received_from_stream_" + handlerNumber + ".txt");
			//FileWriter outWriter = new FileWriter(f);
			ObjectInputStream in = new ObjectInputStream(incoming.getInputStream());
			
			int counter = 0;
			while (keepGoing) 
			{
				//Receive the new Packets				
				o = in.readObject();				
				np = (NetworkPacket) o;

				for(int j= 0; j < np.size();j++)
				{					
					p = np.getPacket(j);					
					pool.addPacket(p);
					counter++;
					//outWriter.write(p.toString() + "\n");
				}
				//if(np.size() > 0)
				//	System.out.println("Received a packet with" + np.size() + " tuples." + " From: " + (np.getPacket(0)).getIdentifier());
				o = null;
				np = null;
				p = null;
				
				//in.reset();
				//in = null;
								
				if (counter >= 1000) 
				{
					counter = 0;
					System.out.println("MORE THAN 1000 TUPLES HAVE BEEN RECEIVED. CALL RUNTIME GC");
					Runtime.getRuntime().gc();
										
					try 
					{
						Thread.sleep(5000);
					} 
					catch (Exception e)
					{
						//nothing
					}							
				}		
			}
			
			incoming.close();
		}
		catch (EOFException eofe)
		{
			try
			{			
				incoming.close();
				this.requestStop();			 
			}
			catch(Exception e)
			{
				System.out.println(e.getMessage());
			}
			 
		}
		catch (SocketException se) 
		{
			System.out.println("Connection Failed: Cannot get stream data from " + incoming.getInetAddress() + ":" + incoming.getPort());
		} 
		catch (Exception e) 
		{
			System.out.println("BAD!");
			e.printStackTrace();
		}				
	}

	public void requestStop() 
	{
		keepGoing = false;
	}
}
