/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet;

import java.io.File;
import java.io.FileWriter;
import java.io.ObjectOutputStream;
import java.net.Socket;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.*;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 15, 2003
 * 
 * The packet sender takes tuples from the packet pool and sends them to other
 * machines in the network.
 * 
 */
public class TupleSender implements Runnable {

	private PacketPool pool;
	private ConnectionTable table;
	private boolean enable;
	
	/**
	 * @param port
	 * @param receivePool
	 * @param tree
	 * @throws IOException
	 */
	public TupleSender(PacketPool sendPool, ConnectionTable ct){
		this.pool = sendPool;
		this.table = ct;	
		this.enable = true;	
	}
	
	public void run(){
		
		XATStreamOperator op;
		Socket sendTo;
		ObjectOutputStream outputStream =null;
		File f;
		FileWriter outWriter = null;
		
		// find out where this packet needs to go.
		
		try{
		
			sendTo = new Socket("127.0.0.1",16001);	
			outputStream = new ObjectOutputStream(sendTo.getOutputStream());
			Packet p = new Packet("newquery","My First Query","newquery");
			outputStream.writeObject(p);
			f = new File("sending_to_application.txt");
			outWriter = new FileWriter(f);
			
		}catch(Exception e){e.printStackTrace();}
		
		while(enable || pool.size() > 0){
		
			try{
		
				// get the packet
				Packet p = pool.removePacket();
		
				if(outputStream != null){
					outputStream.writeObject(p);	
					outWriter.write(p.toString()+"\n");
				}
				
		
			} catch(Exception e){
				e.printStackTrace();
			}
		}
	}
	
	public void requestStop(){
		this.enable = false;
	}
}
