/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation;

import java.util.ArrayList;
import java.util.Hashtable;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
/**
 * @author Bin Liu
 * This is an abstract class for the adaptation decisions. An concrete class will be created
 * for a specific adaptation class. The basic functionality is to decide which partitions to
 * be adapted
 */
public abstract class Adapter {
	// hold the query tree in the query processor
	protected ArrayList queryTree = null;
	protected int totalInandOutQueueSize = 0;
	
	public abstract ArrayList computePartitionsToMoveInTree(XATTree tree, double rate, boolean maximal);
	public abstract ArrayList computePartitionsToMove(XATStreamOperator oper, double rate, Hashtable bePushed,boolean maximal);
	public abstract ArrayList computePartitionsToPush(XATStreamOperator oper, double rate);
	public abstract void localAdapt();
	public abstract void tearDownJoinReferences(int operatorID);
	public abstract boolean isCleanUpCompleted();
	public abstract void setFinalizedQueues();
	
	public void updateQueueSizeInfo(int totalSize) {
		totalInandOutQueueSize = totalSize;
	}
}
