/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Random;

import edu.wpi.cs.dsrg.xmldb.xat.common.migration.realjoinoperator.WTransitiveHashJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.StreamSplitOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;

/**
 * @author Yali Zhu
 * creation time: Feb 22, 2006 5:16:44 PM
 */

public class LocalMemoryAdapter extends Adapter {

    Random rand;
    public LocalMemoryAdapter(ArrayList tr){
    	queryTree = tr;
    	rand = new Random();
    	// do nothing
    }
    
    /**
     * Compute partitions to move to another machine during runtime load balancing.
     * Different from the another method with the same name, this one selects partitions from ALL operators
     * in the query plan tree, not just one operator. Partitions with the same partition ID from ALL operators
     * are being selected to move to another machine. This is useful when the query plan contains more than one 
     * stateful operator. 
     * Assume no disk-based adaptation applied.
     * @return the selected partition
     */
	public ArrayList computePartitionsToMoveInTree(XATTree tree, double rate, boolean maximal) {
	    //assume only run one query plan (tree) in the system.
	    
	    //selected partitionIDs
	    ArrayList selected = new ArrayList();
	    ArrayList joinOps = new ArrayList();
	    XATNode[] nodes = tree.convertToArray();
   		Hashtable costTable = new Hashtable();
	    int totalPartitionIDs = 0;
	    
	    for(int i=0; i<nodes.length; i++) {
	        if(nodes[i].getStreamOperator() instanceof WTransitiveHashJoinOperator) {
	            joinOps.add(nodes[i].getStreamOperator());
	        }
	        
	        if(nodes[i].getStreamOperator() instanceof StreamSplitOperatorImp) {
	            totalPartitionIDs = ((StreamSplitOperatorImp)nodes[i].getStreamOperator()).getPartitionMappingTable().size();
	            System.out.println("Total number of partitions is " + totalPartitionIDs);
	        }
	    }
	      
	    // Calculate the total num of tuples of all partitions with each partition ID in the query plan.  
	    for(int pnum=0; pnum<totalPartitionIDs; pnum++) {
	        int partitionSize = 0;
	        Integer pInt = new Integer(pnum);
	        for(int opnum=0; opnum<joinOps.size(); opnum++) {
	            WTransitiveHashJoinOperator op = (WTransitiveHashJoinOperator)joinOps.get(opnum);
	            HashStatePartitionGroup partitiongroup = (HashStatePartitionGroup)op.getHashJoinStates().get(pInt);
	            if(partitiongroup != null)
	                partitionSize += partitiongroup.getTotalNumOfTuples();
	        }
	        System.out.println("Partition group size for ID " + pnum + " is " + partitionSize);
	        costTable.put(new Integer(pnum), new Integer(partitionSize));
	    }
	    
	    //calculate total partitionSize in the tree for all partition IDs
	    StatisticsGatherer stats = StatisticsGatherer.getStatisticsGather();
	    stats.updateValue("TOTAL_TUPLES_IN_STATES", tree);
	    int totalStateTuples = Integer.parseInt(stats.getValue("TOTAL_TUPLES_IN_STATES", tree));
	    System.out.println("total state tuples: " + totalStateTuples);
	    
	    System.out.println("Partition IDs and Partition Sizes:\n" + costTable);
	    
	    //each time select the smallest or largest partition (depends on value of "maximal") in the tree
	    double curRate = 0;
	    Enumeration keys;
	    int retSize = 0;
	    while (curRate < (rate - 20)) {
			// find the partition with maximal cost
			keys = costTable.keys();
			double curMaxValue = 0;
			int curMaxIndex = -1;
			if (maximal){
				curMaxValue = Double.MIN_VALUE;
			} else {
				curMaxValue = Double.MAX_VALUE;
			}
			
			if (!keys.hasMoreElements()) 
				break;
			
			int costvalue = 0;
			while (keys.hasMoreElements()) {
				Integer key = (Integer)keys.nextElement();
				costvalue = ((Integer)costTable.get(key)).intValue();
				// apparently, we do not want to choose partitions with cost equals to ZERO
				if (costvalue !=0) {
					if (maximal){
						if (costvalue > curMaxValue) {
							curMaxIndex = key.intValue();
							curMaxValue = costvalue;
						}						
					} else {
						if (costvalue < curMaxValue) {
							curMaxIndex = key.intValue();
							curMaxValue = costvalue;
						}												
					}						
				}
			}
			
			if (curMaxIndex != -1){
				selected.add(new Integer(curMaxIndex));
				costTable.remove(new Integer(curMaxIndex));
				retSize += curMaxValue;					
				System.out.println("Select partition ID: " + curMaxIndex + " size: " + curMaxValue);
			} else {
				// also means no right candidates are available, return what ever we have so far
				break;
			}
			//curRate = (double)(retSize)/(double)totalStateTuples;
			curRate = (double)(retSize);
			System.out.println("Current tuples: " + curRate + " Target tuples: " + rate);
		}
	    System.out.println("selected partitions: " + selected);
	    return selected;
    }
	
     
    /* 
     * Sort all partitions based on the totalTuples, return the partition IDs
     * for which the size of these partitions/size of all partitions reaches the rate.
     * The maximal, true or false value indicate whether to choose largest or smallest cost as candidates. 
     * TODO double check if the following method is correct. It is originally copied from LocalOutputRateAdaptor and is modified.
     */
    public ArrayList computePartitionsToMove(XATStreamOperator oper, double rate, Hashtable bePushed, boolean maximal) {
    	
    	ArrayList ret = new ArrayList();
    	
    	if (oper instanceof WTransitiveHashJoinOperator) {
    	    WTransitiveHashJoinOperator joinOper = (WTransitiveHashJoinOperator)oper;
    		Hashtable states = joinOper.getHashJoinStates();
    		Hashtable costTables = new Hashtable();
    		Enumeration keys = states.keys();
    		
    		// first, compute the cost of each partitions in the operator, put them into costTables
    		long totalSize = 0;
    		while (keys.hasMoreElements()) { 
    			Integer key = (Integer)keys.nextElement();
    			int curSize = ((HashStatePartitionGroup)states.get(key)).getTotalNumOfTuples(); 
    			int cost = 0;
    			if (curSize !=0){
    				cost = curSize; 
    			}
    			totalSize += curSize;
    			costTables.put(key, new String(""+cost));
    		}
    		
    		// then find the partition IDs having the largest cost so far, put them into the return list,
    		// until it reaches the rate
    		int retSize = 0;
    		double curRate = (double)retSize/(double)totalSize;
    		while (curRate <= rate) {
    			// find the partition with maximal cost
    			keys = costTables.keys();
    			double curMaxValue = 0;
    			int curMaxIndex = -1;
    			if (maximal){
    				curMaxValue = Double.MIN_VALUE;
    			} else {
    				curMaxValue = Double.MAX_VALUE;
    			}
    			
    			if (keys.hasMoreElements()) {
    				//Integer index = (Integer)keys.nextElement(); 
    				//curMaxIndex = index.intValue();
    				//curMaxValue = ((Integer)costTables.get(index)).intValue();
    			} else {
    				break;
    			}				
    			while (keys.hasMoreElements()) {
    				Integer key = (Integer)keys.nextElement();
    				double value = Double.parseDouble((String)costTables.get(key));
    				// apparently, we do not want to choose partitions with cost equals to ZERO
    				if (value !=0) {
    					if (maximal){
    						if ((value > curMaxValue) && (bePushed.get(key)==null)) {
    							curMaxIndex = key.intValue();
    							curMaxValue = value;
    						}						
    					} else {
    						if ((value < curMaxValue) && (bePushed.get(key)==null)) {
    							curMaxIndex = key.intValue();
    							curMaxValue = value;
    						}												
    					}						
    				}
    			}
    			
    			if (curMaxIndex != -1){
    				ret.add(new Integer(curMaxIndex));
    				costTables.remove(new Integer(curMaxIndex));
    				// compute the newrate
    				retSize += ((HashStatePartitionGroup)states.get(new Integer(curMaxIndex))).getTotalNumOfTuples();					
    			} else {
    				// also means no right candidates are available, return what ever we have so far
    				break;
    			}
    			curRate = (double)(retSize)/(double)totalSize;
    		} // end while
    	}		
    	return ret;
    	
    } //end computePartitionsToMove

    /* 
     * The method computePartitionsToPush() is not supported when LocalMemoryAdapter is applied.
     * This local adapter just moves partitions among machines. It does not push partitions to disk.
     */ 
    public ArrayList computePartitionsToPush(XATStreamOperator oper, double rate) {
        return null;
    }

    public void localAdapt() {
        // TODO Auto-generated method stub

    }

    /* (non-Javadoc)
     * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation.Adapter#tearDownJoinReferences(int)
     */
    public void tearDownJoinReferences(int operatorID) {
        // TODO Auto-generated method stub

    }

    /* (non-Javadoc)
     * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation.Adapter#isCleanUpCompleted()
     */
    public boolean isCleanUpCompleted() {
        // TODO Auto-generated method stub
        return true;
    }

    /* (non-Javadoc)
     * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation.Adapter#setFinalizedQueues()
     */
    public void setFinalizedQueues() {
        // TODO Auto-generated method stub

    }



}
