/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamUnionOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
//import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode.OutOfBoundaryIndexException;

/**
 * @author Bin Liu
 * The class that handles local adaptation decisions for query tree. It has
 * multiple symm-mway hash joins. Here, the adaptation decision need to be made
 * for which join operator and how much need to be adapted
 */
public class LocalMultiJoinTreeAdapter extends Adapter {
	
	double memoryThreshold = 10.0;
	ArrayList join_op_refs; //to store all the join operator instances, 0-top, 1-the next, and so on
	ArrayList split_op_refs; //to store the corresponding split operator instances, 0-the split of the top, ...
	XATStreamOperator last_op_ref = null; //the reference to last operator of the query plan, accessed in the isCleanUpCompleted?
	
	public LocalMultiJoinTreeAdapter(ArrayList tr){
		queryTree = tr;
		//to figure out the relationship of all these join operators, since the push and cleanup
		//depends on the child-parent relationship of the join operators, and also the statistics updates
		join_op_refs = new ArrayList();
		split_op_refs = new ArrayList();
		
		//we only support one query tree for now
		//get the order information about the join operators and the corresponding splits
		XATNode[] nodes = ((XATTree)queryTree.get(0)).convertToArray();
	    for (int i = 0; i < nodes.length; i++){
	    	if(nodes[i].getOperator() instanceof SymmMwayHashJoinOperatorImp){
				join_op_refs.add(nodes[i].getOperator());
				XATNode[] chNodes = nodes[i].getOpChildren();
				// since we only need the corresponding partitionMap Information, it does not matter which split operators is 
				// stored if multiple of them are available
				for (int j=0; j<chNodes.length; j++){
					if (chNodes[j].getOperator() instanceof StreamSplitOperatorImp) {
						split_op_refs.add(chNodes[j].getOperator());
						break;
					}
				}
	    	}	    	
	  	}
	    //set the corresponding join operator the previous join and split operator
	    //these information is used to trace the linearage information when update the cost statistics
	    for (int i=0; i<join_op_refs.size()-1; i++){
	    	SymmMwayHashJoinOperatorImp next_join_ref = (SymmMwayHashJoinOperatorImp)join_op_refs.get(i+1);
	    	StreamSplitOperatorImp next_split_ref = (StreamSplitOperatorImp)split_op_refs.get(i+1);
	    	((SymmMwayHashJoinOperatorImp)join_op_refs.get(i)).setPreviousJoinAndSplit(next_join_ref, next_split_ref);
	    }
		System.out.println("Done with LocalMultiJoinAdapter Initialization.");
	}
	/*
	 * based on the join_op_refs, to update the previous_join_reference, previous_split_reference
	 * (1) find the operator with operatorID, (2) find the next operator of the operatorID, 
	 * (3) set the previous reference of the next operator null. Called at DistributedExecutionController 
	 */
	public void tearDownJoinReferences(int operatorID){
		for (int i=1; i<join_op_refs.size(); i++){
			if (((SymmMwayHashJoinOperatorImp)join_op_refs.get(i)).getOperatorID() ==operatorID){
				((SymmMwayHashJoinOperatorImp)join_op_refs.get(i-1)).setPreviousJoinAndSplit(null, null);
			}
		}
	}
	
    /**
     * Compute partitions to move to another machine during runtime load balancing.
     * Different from the another method with the same name, this one selects partitions from ALL operators
     * in the query plan tree, not just one operator. Partitions with the same partition ID from ALL operators
     * are being selected to move to another machine. This is useful when the query plan contains more than one 
     * stateful operator. 
     * @return the selected partition
     */
	public ArrayList computePartitionsToMoveInTree(XATTree tree, double rate, boolean maximal) {
	    return null;
    }
	
	public ArrayList computePartitionsToMove(XATStreamOperator oper,
			double rate, Hashtable bePushed, boolean maximal) {
		return null;
	}

	public ArrayList computePartitionsToPush(XATStreamOperator oper, double rate) {
		return null;
	}

	/*
	 * The main function that handles the local adaptations
	 */
	public void localAdapt() {
		
		if (computeSystemLoad(false) > memoryThreshold) {
			System.out.println("Start Disk-Based Adaptation ...");
			
			// consider the partitions in a global manner
			Hashtable idsToBePushed = computePartitionsToPush(0.3,true);
			
			/**
			// consider the partitions in sequence, i.e., from Bottom->Top (True), from Top->Bottom (False) 
			Hashtable idsToBePushed = computePartitionsToPushSeq(0.3,true, true);
			*/
			System.out.println("Now, Get IDs to be Pushed ... and Start Pushing to Disk ...");
			Enumeration curOPKeys = idsToBePushed.keys(); // keys in terms of the cursor in join_op_refs
			while (curOPKeys.hasMoreElements()){
				Integer curKey = (Integer)curOPKeys.nextElement();
				ArrayList bePushed = (ArrayList)idsToBePushed.get(curKey);
				//// refer to the join_reference to push the partitions ////
				((SymmMwayHashJoinOperatorImp)join_op_refs.get(curKey.intValue())).pushPartitionsToDisk(bePushed);
			}			
		}		
	} //end of localAdapt

	public boolean isCleanUpCompleted(){
		for (int i=0; i<join_op_refs.size(); i++){
			if (!((SymmMwayHashJoinOperatorImp)join_op_refs.get(i)).isFinalizationComplete()) {
				return false;
			}
		}
		// now, all the joins have been cleaned, we need to wait until nothing in the output queue of
		// the last operator  
		if (last_op_ref == null){
			XATNode[] nodes = ((XATTree)queryTree.get(0)).convertToArray();
		    for (int i = 0; i < nodes.length; i++){
		    	if(((XATStreamOperator)nodes[i].getOperator()).isActive()){
		    		last_op_ref = (XATStreamOperator)nodes[i].getOperator();
		    		break;
		    	}
		    }			
		}
		// here, the overall system stops only if the last operator is the union, then we can safely stop, other wise,
		// the particular instance may in a dead loop
	    if (last_op_ref.getOutputQueue().size()>0)
	    	return false;
	    
	    if (last_op_ref instanceof StreamUnionOperatorImp) {
	    	return true;
	    } else {
	    	try {
	    		Thread.sleep(100000);
	    	} catch (Exception ee){
	    		//do nothing
	    	}
	    	return true;
	    	//return false; //potential dead loop for other instances without the active union operator
	    }
	}
	/*
	 * Mark outputqueue of join operators, split operator with the children
	 * Here, true means the data in the queue needs to be processed, while false represents that the data in the queue could be discarded
	 */
	public void setFinalizedQueues(){
		XATNode[] nodes = ((XATTree)queryTree.get(0)).convertToArray();
	    for (int i = 0; i < nodes.length; i++){
	    	for (int j=0; j<nodes[i].getChildrenSize(); j++){
	    		//try {
	    			if (nodes[i].getChild(j) instanceof String){
	    				/// it is a stream child, disconnected the queue ///
	    				((XATStreamOperator)nodes[i].getOperator()).getInputQueue(j).setAsFinalQueue(false);
	    			} else {
	    				((XATStreamOperator)nodes[i].getOperator()).getInputQueue(j).setAsFinalQueue(true);
	    			}
	    		//}catch (OutOfBoundaryIndexException oe){
	    		//	System.out.println("OutofBoundaryException in setFinalizeQueues");
	    		//}
	    	}
	    }
	    //set the bottom join operator readyToClean initially
	    for (int i=0; i<join_op_refs.size(); i++){
	    	if (((SymmMwayHashJoinOperatorImp)join_op_refs.get(i)).getPreviousJoinReference() == null){
	    		((SymmMwayHashJoinOperatorImp)join_op_refs.get(i)).setReadyToClean();
	    	}
	    }
	    //set the number of join instances in each join based on the outputqueue number of the split operator
	    //this makes sure that the above join operator will wait all the previous join operators (from multiple machines) complete 
	    for (int i=0; i<join_op_refs.size(); i++){
	    	int insNum = ((StreamSplitOperatorImp)split_op_refs.get(i)).getOutputQueueCount();
	    	((SymmMwayHashJoinOperatorImp)join_op_refs.get(i)).setNumberOfInstances(insNum);
	    }

	    
	} //end of setFinalizedQueues
	///// several aux. functions //////
	/*
	 * Globally check all the operators in the query tree, based on the values of Ps/Po
	 * (other cost models could be applied), choose the partitions until the chosen partitions has
	 * reached the specified rate, max = true represents choose maximal Ps/Po to push 
	 */
	public Hashtable computePartitionsToPush(double rate, boolean max){
		Hashtable bePushed = new Hashtable(); //the partition IDs will be pushed
		// (1) first compute the cost of each partitions
		Hashtable overallCostTable = new Hashtable(); //the cost of all 
		double totalSize = 0.0; //project with the size with the unit_tuple_size
		for (int cursor=0; cursor<join_op_refs.size(); cursor++){
			Hashtable curHashStates = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(cursor)).getHashJoinStates();
			Enumeration curKeys = curHashStates.keys();
			Hashtable curCostTable = new Hashtable();
			double cur_base_unit_size = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(cursor)).getBaseUnitSize();
			//double cur_unit_size = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(cursor)).getUnitTupleSize();
			/// if we make the cur_unit_size == 1, we then ignore the effect of the cur_unit_size ///
			while (curKeys.hasMoreElements()){
				Integer key = (Integer)curKeys.nextElement();
				//int curSize = ((HashStatePartition)curHashStates.get(key)).getTotalNumOfTuples();
				double curSize = ( ((HashStatePartitionGroup)curHashStates.get(key)).getPartitionGroupSize() )/cur_base_unit_size;
				int curOut = ((HashStatePartitionGroup)curHashStates.get(key)).getTotalTuplesOutputed();
				int projectedSize = ((HashStatePartitionGroup)curHashStates.get(key)).getTotalProjectedTupleNumber(); // intermediate results cost factors
				// if the projected cost has not been collected, then it is always zero, no effects on the cost
				//we can plug in other cost caculation methods here
				double cost = 0.0;
				double skewed_factor = 5.0; //depends on the samping rate in the statisitics collection in the SymmMwayHashJoinOperator
				if (curOut >0){
					//cost = ((double)curSize*cur_unit_size + projectedSize*5.0)/((double)curOut*skewed_factor);
					cost = (curSize + projectedSize*skewed_factor)/((double)curOut*skewed_factor); 
				} else {
					cost = curSize + projectedSize*skewed_factor;
				}
				// the totalSize will not affected by the projectedSize
				//totalSize += curSize*cur_unit_size; // we can change the cost model of the total state size later as required
				totalSize += curSize; // we can change the cost model of the total state size later as required
				curCostTable.put(key, new String(""+cost));
			}
			//Integer oID = new Integer(((SymmMwayHashJoinOperatorImp)join_op_refs.get(cursor)).getOperatorID());
			//Integer cursorID = cursor
			overallCostTable.put(new Integer(cursor),curCostTable);
		}
		
		// (2) now choose the partitions with the maximal Po/Ps or minimal Po/Ps value as specified		
		double retSize = 0;
		double curRate = (double)retSize/(double)totalSize;
		Enumeration cursorKeys;
		while (curRate <= rate) {

			double curCanValue = 0;
			Integer curCanIndex = null;
			Integer curCursorIndex = null;
			// (1) get the first index and value
			if (max) { //largest value as candidate
				curCanValue = Double.MIN_VALUE;
			} else { //smallest value as candidate
				curCanValue = Double.MAX_VALUE;
			}
			
			// find the partition with maximal cost,
			cursorKeys = overallCostTable.keys();
			while (cursorKeys.hasMoreElements()){
				Integer cursorKey = (Integer)cursorKeys.nextElement();
				Hashtable curOPCost = (Hashtable)overallCostTable.get(cursorKey);
				Enumeration keys = curOPCost.keys();
				while (keys.hasMoreElements()) {
					Integer key = (Integer)keys.nextElement();
					double value = Double.parseDouble((String)curOPCost.get(key));
					if (max) {
						if (value > curCanValue) { // choose the max value as the candidate
							curCanIndex = key;
							curCanValue = value;
							curCursorIndex = cursorKey;
						}
					} else { //choose the min value as the candidate
						if (value<curCanValue) {
							curCanIndex = key;
							curCanValue = value;
							curCursorIndex = cursorKey;
						}
					}
				} //end of while inner				
			} //end of while outer
			
			//if the curCanIndex != -1, record the
			if (curCursorIndex != null){
				//yes, we found it,
				ArrayList retOP = (ArrayList)bePushed.get(curCursorIndex);
				if (retOP == null){
					retOP = new ArrayList();
				}
				retOP.add(curCanIndex);
				bePushed.put(curCursorIndex,retOP);
				((Hashtable)overallCostTable.get(curCursorIndex)).remove(curCanIndex);
				// find the join_reference with the operator ID
				Hashtable tmpStates = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(curCursorIndex.intValue())).getHashJoinStates();						
				//double cur_unit_size = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(curCursorIndex.intValue())).getUnitTupleSize();
				double cur_base_unit_size = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(curCursorIndex.intValue())).getBaseUnitSize();
				//retSize += ((HashStatePartition)tmpStates.get(curCanIndex)).getTotalNumOfTuples()*cur_unit_size;
				retSize += ((HashStatePartitionGroup)tmpStates.get(curCanIndex)).getPartitionGroupSize()/cur_base_unit_size;
				// compute the newrate
				curRate = (double)(retSize)/(double)totalSize;				
			} else {
				break;
			}
		} //end of while curRate<=rate		
		return bePushed;
	} //end of computePartitionsToPush

	/*
	 * In this method, we always choose the partitions to be pushed based on the sequence of join operators,
	 * either from top->bottom or from bottom->top. It is defined by the boolean order: true: bottom->top, while fase
	 * from top->bottom. Usually this methods are used with the local cost models, but it is not restricted though
	 */
	public Hashtable computePartitionsToPushSeq(double rate, boolean max, boolean order){
		Hashtable bePushed = new Hashtable(); //the partition IDs will be pushed
		// (1) first compute the cost of each partitions
		Hashtable overallCostTable = new Hashtable(); //the cost of all 
		double totalSize = 0.0; //project with the size with the unit_tuple_size
		for (int cursor=0; cursor<join_op_refs.size(); cursor++){
			Hashtable curHashStates = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(cursor)).getHashJoinStates();
			Enumeration curKeys = curHashStates.keys();
			Hashtable curCostTable = new Hashtable();
			//double cur_unit_size = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(cursor)).getUnitTupleSize();
			double cur_base_unit_size = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(cursor)).getBaseUnitSize();
			/// if we make the cur_unit_size == 1, we then ignore the effect of the cur_unit_size ///
			while (curKeys.hasMoreElements()){
				Integer key = (Integer)curKeys.nextElement();
				//int curSize = ((HashStatePartition)curHashStates.get(key)).getTotalNumOfTuples();
				double curSize = ((HashStatePartitionGroup)curHashStates.get(key)).getPartitionGroupSize()/cur_base_unit_size;
				int curOut = ((HashStatePartitionGroup)curHashStates.get(key)).getTotalTuplesOutputed();
				int projectedSize = ((HashStatePartitionGroup)curHashStates.get(key)).getTotalProjectedTupleNumber(); // intermediate results cost factors
				// if the projected cost has not been collected, then it is always zero, no effects on the cost
				//we can plug in other cost caculation methods here
				double cost = 0.0;
				double skewed_factor = 5.0; //this is no longer been used in this method since projectedSize ==0
				if (curOut >0){
					//cost = ((double)curSize*cur_unit_size + projectedSize)/((double)curOut*skewed_factor);
					cost = ((double)curSize + projectedSize*skewed_factor)/((double)curOut*skewed_factor);
				} else {
					//cost = (double)(curSize*cur_unit_size + projectedSize);
					cost = (double)(curSize + projectedSize*skewed_factor);
				}
				// the totalSize will not affected by the projectedSize
				//totalSize += curSize*cur_unit_size; // we can change the  cost model of the total state size later as required
				totalSize += curSize; // we can change the  cost model of the total state size later as required
				curCostTable.put(key, new String(""+cost));
			}
			//Integer oID = new Integer(((SymmMwayHashJoinOperatorImp)join_op_refs.get(cursor)).getOperatorID());
			//Integer cursorID = cursor
			overallCostTable.put(new Integer(cursor),curCostTable);
		}
		
		// (2) now choose the partitions with the maximal Po/Ps or minimal Po/Ps value as specified		
		double retSize = 0;
		double curRate = (double)retSize/(double)totalSize;
		Enumeration cursorKeys;
		while (curRate <= rate) {

			double curCanValue = 0;
			Integer curCanIndex = null;
			Integer curCursorIndex = null;
			// (1) get the first index and value
			if (max) { //largest value as candidate
				curCanValue = Double.MIN_VALUE;
			} else { //smallest value as candidate
				curCanValue = Double.MAX_VALUE;
			}
			
			// find the target operator cursor key
			Integer curCursorKey = null;
			if (order) {
				// from Bottom->Top
				for (int ss=join_op_refs.size()-1;ss>=0;ss--){
					if (!((Hashtable)overallCostTable.get(new Integer(ss))).isEmpty()){
						curCursorKey = new Integer(ss);
						break;
					}
				}				
			} else {
				// from Top->Bottom
				for (int ss=0; ss<join_op_refs.size(); ss++){
					if (!((Hashtable)overallCostTable.get(new Integer(ss))).isEmpty()){
						curCursorKey = new Integer(ss);
						break;
					}					
				}				
			}
			if (curCursorKey == null){
				// no partitoins could be selected, simply return
				break;
			}

			Hashtable curOPCost = (Hashtable)overallCostTable.get(curCursorKey);
			Enumeration keys = curOPCost.keys();
			while (keys.hasMoreElements()) {
				Integer key = (Integer)keys.nextElement();
				double value = Double.parseDouble((String)curOPCost.get(key));
				if (max) {
					if (value > curCanValue) { // choose the max value as the candidate
						curCanIndex = key;
						curCanValue = value;
						curCursorIndex = curCursorKey;
					}
				} else { //choose the min value as the candidate
					if (value<curCanValue) {
						curCanIndex = key;
						curCanValue = value;
						curCursorIndex = curCursorKey;
					}
				}
			} //end of while inner				
			
			//if the curCanIndex != -1, record the
			if (curCursorIndex != null){
				//yes, we found it,
				ArrayList retOP = (ArrayList)bePushed.get(curCursorIndex);
				if (retOP == null){
					retOP = new ArrayList();
				}
				retOP.add(curCanIndex);
				bePushed.put(curCursorIndex,retOP);
				((Hashtable)overallCostTable.get(curCursorIndex)).remove(curCanIndex);
				// find the join_reference with the operator ID
				Hashtable tmpStates = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(curCursorIndex.intValue())).getHashJoinStates();
				//double cur_unit_size = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(curCursorIndex.intValue())).getUnitTupleSize();
				double cur_base_unit_size = ((SymmMwayHashJoinOperatorImp)join_op_refs.get(curCursorIndex.intValue())).getBaseUnitSize();
				retSize += ((HashStatePartitionGroup)tmpStates.get(curCanIndex)).getPartitionGroupSize()/cur_base_unit_size;
				// compute the newrate
				curRate = (double)(retSize)/(double)totalSize;				
			} else {
				break;
			}
		} //end of while curRate<=rate		
		return bePushed;
	} //end of computePartitionsToPush
	
	
	/*
	 * Compute the current load of the query processor. The load is composed by the size of the
	 * data in all queues, and the size of the states in all join operators.
	 * Note that the size of all queues are already computed and set in updateSimCostofSymmJoin to 
	 * all join operators
	 * If the flag is the true memory load, the we simply return the real memory usage
	 */
	public double computeSystemLoad(boolean real_load){
	
		if (real_load){ // actual main memory usage
			Runtime runtime = Runtime.getRuntime();
			long usedmemory = runtime.totalMemory()-runtime.freeMemory();
			long maxmemory = runtime.maxMemory();
			// add one short cut here, only half of the maxiaml memory is used in determine the 
			// memoryusage percentage, this case, thus, to avoid the memory overflow of  the
			// finalization stage 
			// double memoryusage = ((double)usedmemory*2)/(double)maxmemory;			
			//System.out.println("REAL-MEMORY USAGE: " + usedmemory);
			return usedmemory;			

		} else { // simulated cost model, use tuples in the queue and join states
			// (1) get the total state size of all join operaors in the query tree
			double totalStateSize = 0; //total state size projected with the unit_tuple_size
			for (int i = 0; i < queryTree.size(); i++) {			
				XATTree currentTree = (XATTree) queryTree.get(i);
				XATNode[] curNode = currentTree.convertToArray();
				for (int j=0; j<curNode.length; j++){
					if ((curNode[j]).getStreamOperator().isActive()){
						if (curNode[j].getStreamOperator() instanceof SymmMwayHashJoinOperatorImp){
							long localSize = 0; // for VLDB05-Demo to compute the operator size of the operator
							SymmMwayHashJoinOperatorImp tmpJOp = (SymmMwayHashJoinOperatorImp)curNode[j].getStreamOperator();
							Enumeration keys = tmpJOp.getHashJoinStates().keys();
							//double cur_unit_size = tmpJOp.getUnitTupleSize();
							double cur_base_unit_size = tmpJOp.getBaseUnitSize();
							while (keys.hasMoreElements()) { 
								Integer key = (Integer)keys.nextElement();
								//totalStateSize += ((HashStatePartition)tmpJOp.getHashJoinStates().get(key)).getTotalNumOfTuples()*cur_unit_size;
								double tmp = ((HashStatePartitionGroup)tmpJOp.getHashJoinStates().get(key)).getPartitionGroupSize();
								totalStateSize += tmp/cur_base_unit_size;
								localSize += (long)tmp; // for VLDB05-Demo
							}
							((SymmMwayHashJoinOperatorImp)curNode[j].getStreamOperator()).setTotalSize(localSize); //for VLDB05-Demo
						}
					}
				}
			}
			// (2) add the totalStateSize with the totalQueueSize
			int totalQueueSize = this.totalInandOutQueueSize;
			if (totalQueueSize > 15000){
				// might be transitant increase of queue size, here, we cut it to 15,000
				// the effect on the main memory usage is not linear
				totalQueueSize = 15000 + ((totalQueueSize-15000)/10);
			}
			double simulatedCost = (totalStateSize + totalQueueSize)/1024.0;
			//System.out.println("SIMULATED-MEMORY USAGE: " + simulatedCost);
			return simulatedCost;			
		}
	} //end of computeSystemLoad
	
}
