/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Enumeration;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamUnionOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.StreamSplitOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.SymmMwayHashJoinOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
/**
 * @author Bin Liu
 * This class computes the partitions to move/push based on the output rate. Basically,
 * the idea is to have partitions with high output rate moved to other machines, while 
 * push partitions with lower output rate to disk
 */

public class LocalOutputRateAdapter extends Adapter {

	XATStreamOperator last_op_ref = null; //the reference to last operator of the query plan, accessed in the isCleanUpCompleted?
	SymmMwayHashJoinOperatorImp join_op_ref = null; //the reference to the join operator of the query plan, accessed in the isCleanUpCompleted?
	
	public LocalOutputRateAdapter(ArrayList tr){
		queryTree = tr;
		// do nothing
	}
	
	public void localAdapt(){
		return;
	}

	public void tearDownJoinReferences(int operatorID){
		return;
	}
	
	public boolean isCleanUpCompleted(){
		XATNode[] nodes = ((XATTree)queryTree.get(0)).convertToArray();
		
		if (join_op_ref == null){
			for (int i = 0; i < nodes.length; i++){
				if (nodes[i].getOperator() instanceof SymmMwayHashJoinOperatorImp) {
					join_op_ref = (SymmMwayHashJoinOperatorImp)nodes[i].getOperator();
					break;
				}
			}
		}

		if (!join_op_ref.isFinalizationComplete()) {
			return false;
		}	    		

		// now, all the joins have been cleaned, we need to wait until nothing in the output queue of
		// the last operator  
		if (last_op_ref == null){
		    for (int i = 0; i < nodes.length; i++){
		    	if(((XATStreamOperator)nodes[i].getOperator()).isActive()){
		    		last_op_ref = (XATStreamOperator)nodes[i].getOperator();
		    		break;
		    	}
		    }			
		}
		// here, the overall system stops only if the last operator is the union, then we can safely stop, other wise,
		// the particular instance may in a dead loop
	    if (last_op_ref.getOutputQueue().size()>0)
	    	return false;
	    
	    if (last_op_ref instanceof StreamUnionOperatorImp) {
	    	return true;
	    } else {
	    	try {
	    		Thread.sleep(10000);
	    	} catch (Exception ee){
	    		//do nothing
	    	}
	    	return true;
	    	//return false; //potential dead loop for other instances without the active union operator
	    }
	}
	
	public void setFinalizedQueues(){
		XATNode[] nodes = ((XATTree)queryTree.get(0)).convertToArray();
		int numberofInstances = 0;
		int joinIdentify = -1;
		for (int i = 0; i < nodes.length; i++){
	    	for (int j=0; j<nodes[i].getChildrenSize(); j++){
	    		//try {
	    			if (nodes[i].getChild(j) instanceof String){
	    				/// it is a stream child, disconnected the queue ///
	    				((XATStreamOperator)nodes[i].getOperator()).getInputQueue(j).setAsFinalQueue(false);
	    			} else {
	    				((XATStreamOperator)nodes[i].getOperator()).getInputQueue(j).setAsFinalQueue(true);
	    			}
	    		//}catch (OutOfBoundaryIndexException oe){
	    		//	System.out.println("OutofBoundaryException in setFinalizeQueues");
	    		//}
	    	}
		    //set the join operator readyToClean initially
	    	if (nodes[i].getOperator() instanceof SymmMwayHashJoinOperatorImp) {
	    		((SymmMwayHashJoinOperatorImp)nodes[i].getOperator()).setReadyToClean();
	    		joinIdentify = i;
	    	}
	    	//if we find split operator, get the numberofInstances
	    	if (nodes[i].getOperator() instanceof StreamSplitOperatorImp) {
	    		numberofInstances = ((StreamSplitOperatorImp)nodes[i].getOperator()).getOutputQueueCount();
	    	}
	    }		
	    //set the number of join instances in each join based on the outputqueue number of the split operator
	    //this makes sure that the above join operator will wait all the previous join operators (from multiple machines) complete 
	    if ((joinIdentify != -1) && (numberofInstances >0)){
	    	((SymmMwayHashJoinOperatorImp)nodes[joinIdentify].getOperator()).setNumberOfInstances(numberofInstances);
	    }		
		return;
	}
	
    /**
     * Compute partitions to move to another machine during runtime load balancing.
     * Different from the another method with the same name, this one selects partitions from ALL operators
     * in the query plan tree, not just one operator. Partitions with the same partition ID from ALL operators
     * are being selected to move to another machine. This is useful when the query plan contains more than one 
     * stateful operator. 
     * @return the selected partition
     */
	public ArrayList computePartitionsToMoveInTree(XATTree tree, double rate, boolean maximal) {
	    return null;
    }
	/* 
	 * Sort all partitions based on the totalTuples + totalTuplesOutputed, return the partition IDs
	 * that the size of these partitions/size of all partitions reaches the rate
	 * The maximal, true or false value indicate whether to choose largest or smallest cost as candidates 
	 */
	public ArrayList computePartitionsToMove(XATStreamOperator oper, double rate, Hashtable bePushed, boolean maximal) {
		
		ArrayList ret = new ArrayList();
		
		if (oper instanceof SymmMwayHashJoinOperatorImp) {
			SymmMwayHashJoinOperatorImp joinOper = (SymmMwayHashJoinOperatorImp)oper;
			Hashtable states = joinOper.getHashJoinStates();
			Hashtable costTables = new Hashtable();
			Enumeration keys = states.keys();
			
			// first, compute the cost of each partitions in the operator, put them into costTables
			long totalSize = 0;
			while (keys.hasMoreElements()) { 
				Integer key = (Integer)keys.nextElement();
				int curSize = ((HashStatePartitionGroup)states.get(key)).getTotalNumOfTuples(); 
				double cost = 0;
				if (curSize !=0){
				    double totaloutput = ((double)((HashStatePartitionGroup)states.get(key)).getTotalTuplesOutputed());
				    if(totaloutput > 0)
				        cost = ((double)curSize) / totaloutput;
					System.out.println("partition " + key + " size: " + curSize + " cost: " + cost);
				}
				totalSize += curSize;
				costTables.put(key, new String(""+cost));
			}
			
			// then find the partition IDs having the largest cost so far, put them into the return list,
			// until it reaches the rate
			int retSize = 0;
			double curRate = (double)retSize/(double)totalSize;
			while (curRate <= rate) {
				// find the partition with maximal cost
				keys = costTables.keys();
				double curMaxValue = 0;
				int curMaxIndex = -1;
				if (maximal){
					curMaxValue = Double.MIN_VALUE;
				} else {
					curMaxValue = Double.MAX_VALUE;
				}
				
				if (keys.hasMoreElements()) {
					//Integer index = (Integer)keys.nextElement(); 
					//curMaxIndex = index.intValue();
					//curMaxValue = ((Integer)costTables.get(index)).intValue();
				} else {
					break;
				}				
				while (keys.hasMoreElements()) {
					Integer key = (Integer)keys.nextElement();
					double value = Double.parseDouble((String)costTables.get(key));
					// apparently, we do not want to choose partitions with cost equals to ZERO
					if (value !=0) {
						if (maximal){
							if ((value > curMaxValue) && (bePushed.get(key)==null)) {
								curMaxIndex = key.intValue();
								curMaxValue = value;
							}						
						} else {
							if ((value < curMaxValue) && (bePushed.get(key)==null)) {
								curMaxIndex = key.intValue();
								curMaxValue = value;
							}												
						}						
					}
				}
				
				if (curMaxIndex != -1){
					ret.add(new Integer(curMaxIndex));
					System.out.println("choose partition " + curMaxIndex);
					costTables.remove(new Integer(curMaxIndex));
					// compute the newrate
					retSize += ((HashStatePartitionGroup)states.get(new Integer(curMaxIndex))).getTotalNumOfTuples();					
				} else {
					// also means no right candidates are available, return what ever we have so far
					break;
				}
				curRate = (double)(retSize)/(double)totalSize;
			} // end while
		}		
		return ret;
		
	} //end computePartitionsToMove

	/* In case of disk-based adaptation,
	 * sort all partitions based on the size of partitions/tuples outputed, choose the largest one until reaches
	 * the rate
	 */
	public ArrayList computePartitionsToPush(XATStreamOperator oper, double rate) {

		ArrayList ret = new ArrayList();
		
		if (oper instanceof SymmMwayHashJoinOperatorImp) {
			SymmMwayHashJoinOperatorImp joinOper = (SymmMwayHashJoinOperatorImp)oper;
			Hashtable states = joinOper.getHashJoinStates();
			Hashtable costTables = new Hashtable();
			Enumeration keys = states.keys();
			
			// first, compute the cost of each partitions in the operator, put them into costTables
			long totalSize = 0;
			while (keys.hasMoreElements()) { 
				Integer key = (Integer)keys.nextElement();
				int curSize = ((HashStatePartitionGroup)states.get(key)).getTotalNumOfTuples(); 
				double cost = ((double)curSize)/((HashStatePartitionGroup)states.get(key)).getTotalTuplesOutputed();
				totalSize += curSize;
				costTables.put(key, new String(""+cost));
			}
			
			// then find the partition IDs having the largest cost so far, put them into the return list,
			// until it reaches the rate
			int retSize = 0;
			double curRate = (double)retSize/(double)totalSize;
			while (curRate <= rate) {
				// find the partition with maximal cost
				keys = costTables.keys();
				double curMaxValue = 0;
				int curMaxIndex = 0;
				if (!(keys.hasMoreElements())) {
					// if no more entries in the cost table, then return whatever we have
					break;
				}				
				while (keys.hasMoreElements()) {
					Integer key = (Integer)keys.nextElement();
					double value = Double.parseDouble((String)costTables.get(key)); 
					if (value > curMaxValue) {
						curMaxIndex = key.intValue();
						curMaxValue = value;
					}
				}
				
				ret.add(new Integer(curMaxIndex));
				costTables.remove(new Integer(curMaxIndex));
				// compute the newrate
				retSize += ((HashStatePartitionGroup)states.get(new Integer(curMaxIndex))).getTotalNumOfTuples();
				curRate = (double)(retSize)/(double)totalSize;
			} // end while
			System.out.println("Total State Size:" + states.size() + " Pushed Size:" + ret.size());
		}		
		return ret;

	} //end of computePartitionsToPush

} //end of localoutputrateadapter
