/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
/*
 * -- Bin. Modified on Nov 30, 2004
 * Basically, functions are added to handle the partitioned distribution case. (1) add new constructor
 * to hold the partitionMaps of all split operator (2) new version of connect() 
 */

package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern;

import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.ConnectException;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Application;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Server;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Activate;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.DeActivate;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ReceiveData;
//import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ReceiveStream;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.SendData;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.SendState;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.SendStream;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.StopSend;
//import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.SubPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.StreamSplitOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ActivatePartition;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.PropagatePartitionMaps;
/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Feb 17, 2004
 * 
 * Insert comments here.
 */
public class ConnectionManagerImp implements ConnectionManager {

	private DistributionTable currentTable;
	private Hashtable serverTable;
	private Hashtable appTable;
	private int streamTime;
	private Hashtable partitionMaps; // added to hold partitionMaps of split operator
	
	/**
	 * original version of constructor
	 */
	public ConnectionManagerImp(DistributionTable table, Hashtable servers, Hashtable applications, int st) {
		currentTable = table;
		serverTable = servers;
		appTable = applications;
		streamTime = st;
	}
	/* -- Bin Nov.30, 2004
	 * used when partitioned-parallelism distribution has been applied
	 */
	public ConnectionManagerImp(DistributionTable table, Hashtable maps, Hashtable servers, Hashtable applications, int st) {
		currentTable = table;
		partitionMaps = maps;
		serverTable = servers;
		appTable = applications;
		streamTime = st;
	}
	
	/*
	 * It sends propagatePartitionMaps message to all query processors. It is used when query tree has multiple join operators
	 * while we need trace back to update the cost statistics of the join operators. This is needed since the split operator may 
	 * only be activated in one query processor, thus, the partitionMap and the partitionFunction may not be set for those have not
	 * been activated. -- Bin Apr 17, 2005
	 */
	public void propagatePartitionMaps(Machine[] ms){	

		PropagatePartitionMaps pMaps = new PropagatePartitionMaps(partitionMaps);		
		Packet p = new Packet("propagate",pMaps);
		
		try{
			for (int i=0; i<ms.length; i++){
			
				Socket s = new Socket(ms[i].getIPAddress(),ms[i].getConnectionListenerPort());			 
				ObjectOutputStream out = new ObjectOutputStream(s.getOutputStream());

				out.writeObject(p);
				out.close();
				s.close();
			}
		} catch (IOException ioe){
			System.out.println("Exceptions in propagatePartitionMaps.");
		}
	}
	
	/* -- Bin Nov.30, 2004
	 * used when partitioned-parallelism distribution table has been applied. it is close to the version connect(), 
	 * however, the major difference is (1) to handle Machine[] in the distribution table, and (2) use the new 
	 * ActivatePartition to set partitionMaps and create Dynamic queues for the split operator in the query plan 
	 */
	public boolean connectPartitions(){
		Socket currSocket = null;
		ObjectOutputStream currOut;
		
		try {
			Enumeration e = currentTable.getNodeEnumeration();
			
			while(e.hasMoreElements()){
				XATNode currentNode = (XATNode) e.nextElement();
				// need to handle the case if the node is distributed to more than one machines
				Object m = currentTable.getObjectByNode(currentNode);
				Machine[] machines;
				if (m.getClass().isArray()) {
					machines = (Machine[])m; 
				} else {
					machines = new Machine[1];
					machines[0] = (Machine)m;
				}
				for (int k=0; k<machines.length; k++) {
					Machine curMachine = machines[k];
					currSocket = new Socket(curMachine.getIPAddress(), curMachine.getConnectionListenerPort());
					System.out.println("Current operator is: " + currentNode.getOperator().getProperty("OBJECT_ID")); //
					currOut = new ObjectOutputStream(currSocket.getOutputStream());
					
					//See if it connects to a stream
					if(currentNode.getStreamChildSize() > 0){
						//Get all of the streams it is connected to					
						System.out.println("Streamsize of current node is: " + currentNode.getStreamChildSize());
						System.out.println("Children Operators of current node are: " + currentNode.getOpChildrenSize());
					
						Iterator i = currentNode.getStreamChildNames();
									
						while(i.hasNext()){
							String streamName = (String) i.next();
			
							Server server = (Server) this.serverTable.get(streamName);				
							int queueNum = 0;
	                    
							System.out.println("Current operator is: " + ((XATStreamOperator)currentNode.getOperator()));
							System.out.println("Current operator has: " + ((XATStreamOperator)currentNode.getOperator()).getInputQueueCount() + " input queues");
							System.out.println("Input stream is: " + streamName);
						
							XATQueue q = ((XATStreamOperator)currentNode.getOperator()).getInputQueue(streamName);
							System.out.println(q.getQueueNumber());
						
							queueNum = q.getQueueNumber();
						
							//Create the connection object
							Connection c = new SendStream(streamName,streamTime,curMachine,currentNode.getOperator().getProperty("OBJECT_ID"),queueNum);
							Packet p = new Packet("connect",c);
						
							//Send the packet to the stream generator
							Socket sgSocket = new Socket(server.getIPAddress(),server.getPort());
							ObjectOutputStream sgOut = new ObjectOutputStream(sgSocket.getOutputStream());
							sgOut.writeObject(p);
							sgOut.close();
							sgSocket.close();
						} // end while stream child size
					} //end if streamchild
					
					// Now activate the node on the machine					
					Connection c = new Activate(currentNode.getOperator().getProperty("OBJECT_ID"));
					Packet p = new Packet("local",c);					
					currOut.writeObject(p);

					// Now look and see where the parent is and connect it to the parent.					
					Iterator it = currentNode.getParents();
			
					while(it.hasNext()){
					
						//Remember there may be more than 1 parent
						XATNode parentNode = (XATNode) it.next();
						int queueNum = ((XATStreamOperator) currentNode.getOperator()).getOutputQueue().getQueueNumber();
						
						//check the deployment of the parent node, if the parent node is distributed to more than one machines, then
						//it means we need to set the partitionMaps and create dynamic queues of the currentNode
						//Changed: for a query tree, we need to check to make sure the currentNode is a split operator, since multiple split 
						//operators may activated in multiple machines
						Object pm = currentTable.getObjectByNode(parentNode);
						Machine[] pmachines;
						String curOPName = currentNode.getOperator().getClass().getName();
						
						if ( pm.getClass().isArray()  ) {
						    if (curOPName.endsWith("StreamSplitOperatorImp")) {
								//(1) the currentNode is a split operator, first set the partitionMaps and create dynamic queues to the currentNode
								pmachines = (Machine[])pm;
								Hashtable pMaps = (Hashtable)partitionMaps.get((StreamSplitOperatorImp) currentNode.getOperator());
								System.out.println("CM:Send partition mapping table to operator " + currentNode.getOperator());
								System.out.println("CM:The partition table is: \n" + pMaps);
								Connection apc = new ActivatePartition(currentNode.getOperator().getProperty("OBJECT_ID"),queueNum,pMaps,pmachines,curMachine);
								Packet papc = new Packet("machines",apc);
								currOut.writeObject(papc);
								
								//(2) set the outConnections (SendData) of currentNode
								Connection sdc = new SendData(currentNode.getOperator().getProperty("OBJECT_ID"),queueNum,pmachines,curMachine);
								Packet psdc = new Packet("machines",sdc);
								currOut.writeObject(psdc);
								
								//(3) set the inConnection (ReceiveData) of parentNode
								for (int im=0; im<pmachines.length; im++){
									Machine curParentMachine = pmachines[im];
									if (curParentMachine.equals(curMachine)){
										//The parent is on the same machine as the child, so this will be connected later? or we may also need do something
									} else {
										Connection rdc = new ReceiveData(currentNode.getOperator().getProperty("OBJECT_ID"),parentNode.getOperator().getProperty("OBJECT_ID"),queueNum);
										Packet prdc = new Packet("machines",rdc);
														
										//Send it to the parents machine
										Socket parentSocket = new Socket(curParentMachine.getIPAddress(),curParentMachine.getConnectionListenerPort());
										ObjectOutputStream parentOut = new ObjectOutputStream(parentSocket.getOutputStream());
										parentOut.writeObject(prdc);
										parentOut.close();
										parentSocket.close();															
									}
								}								
							} else {
								// this is the case when a partition query tree where the currentnode is join, while parent node is split
								// we assume this split operator is always deployed in the same with the join operator
								// basically do thing here
							}
							
						} else {
							//the normal cases that the currentNode is not being partitioned
							Machine parentMachine = (Machine)pm; 
							if (curMachine.equals(parentMachine)){
								//The parent is on the same machine as the child, so this will be connected later.
							} else {
								//The parent is on another machine so tell our machine how to connect it, and then
								//let the other machine know that it will be getting tuples from the child.
								//@TODO RIGHT NOW WE ONLY SUPPORT TREES, HERE IS WHERE THE CHANGE WILL HAVE TO BE MADE TO SUPPORT DAG's
															
								Connection fc = new SendData(currentNode.getOperator().getProperty("OBJECT_ID"),queueNum,parentMachine);
								Packet np = new Packet("machines",fc);
								
								//Send it to the childs machine
								currOut.writeObject(np);
							
								//Now tell the parent about the tuples it will be receiving
								Connection tc = new ReceiveData(currentNode.getOperator().getProperty("OBJECT_ID"),parentNode.getOperator().getProperty("OBJECT_ID"),queueNum);
								Packet mp = new Packet("machines",tc);
												
								//Send it to the parents machine
								Socket parentSocket = new Socket(parentMachine.getIPAddress(),parentMachine.getConnectionListenerPort());
								ObjectOutputStream parentOut = new ObjectOutputStream(parentSocket.getOutputStream());
								parentOut.writeObject(mp);
								parentOut.close();
								parentSocket.close();								
							} 													
						} //end if pm.getClass().isArray()						
					} // end while loop of parents nodes
					
					//If the iterator is null, there are no parents so connect it to an application
					if(currentNode.getParentSize() == 0){					
						String currID = currentNode.getTreeID();
						Application currApp = new Application();
						//Application currApp = (Application) appTable.get(currID);
						//modified by Song Wang for multiple application support in one query plan
						ArrayList currAppList = (ArrayList) appTable.get(currID);
						if (currAppList.size()==1 && ((Application)currAppList.get(0)).getRootID()==0){
							//default case: one application for the query plan only. 
							currApp = (Application)currAppList.get(0);
						} else {
							//new case: multiple applications for one query plan. check the rootID field
							for (int i = 0; i < currAppList.size();i++ ) {
								if (((Application)currAppList.get(i)).getRootID()==((XATStreamOperator)currentNode.getOperator()).getOperatorID()){
									currApp = (Application)currAppList.get(i);
									break;
								}
							}
						}
						Connection ac = new SendData(currentNode.getOperator().getProperty("OBJECT_ID"),0,currApp.getIPAddress(),currApp.getPort());
						Packet ap = new Packet("connect",ac);
						//Send this info the the stream generator
						currOut.writeObject(ap);
					}
				
					currOut.close();
					currSocket.close();
										
				} //end of for machines.length
			} //end of while all nodes in the currentTable
			
		} catch(ConnectException ce){
			System.out.println("=======ConnectException=========");
			System.out.println("Message: " + ce.getMessage());
			System.out.println("Cause " + ce.getCause());
			System.out.println(currSocket);
			ce.printStackTrace();
			System.out.println("=========================");
		}catch(Exception e){
			System.out.println("=======Exception=========");
			e.printStackTrace();
			System.out.println("=========================");
		}
		
		return true;
	} //end of connectPartitions
	/* 
	 * original version, no partitioned parallelism has been involved.
	 */
	public boolean connect() {
		
		Socket currSocket = null;
		ObjectOutputStream currOut;
		
		try{
		
			Enumeration e = currentTable.getNodeEnumeration();		
			System.out.println("distribution table is: " + currentTable.toString());
		
			//Connect all of the nodes
			while(e.hasMoreElements()){
				//Thread.sleep(100);
				XATNode currentNode = (XATNode) e.nextElement();

			
				//See if it connects to a stream
				if(currentNode.getStreamChildSize() > 0){
					//Get all of the streams it is connected to
				
					System.out.println("Streamsize of current node is: " + currentNode.getStreamChildSize());
					System.out.println("Children Operators of current node are: " + currentNode.getOpChildrenSize());
					//System.out.println("Child OP of current node is: " + ((XATNode)currentNode.getOpChild(0)).getOperator().getProperty("OBJECT_ID"));
				
					Iterator i = currentNode.getStreamChildNames();
								
					while(i.hasNext()){
						String streamName = (String) i.next();
		
						Server server = (Server) this.serverTable.get(streamName);				
						int queueNum = 0;
                    
						System.out.println("Current operator has : " + ((XATStreamOperator)currentNode.getOperator()).getInputQueueCount() + " input queues");
						System.out.println("Inputqueue for operator: " + ((XATStreamOperator)currentNode.getOperator()));
						System.out.println("for stream: " + streamName + " is " );
					
						XATQueue q = ((XATStreamOperator)currentNode.getOperator()).getInputQueue(streamName);
						System.out.println(q.getQueueNumber());
					
						queueNum = q.getQueueNumber();
						//System.out.println(q.getQueueNumber());
					
						//Create the connection object
						Connection c = new SendStream(streamName,streamTime,currentTable.getMachine(currentNode),currentNode.getOperator().getProperty("OBJECT_ID"),queueNum);
						Packet p = new Packet("connect",c);
					
						//Send the packet to the stream generator
						Socket sgSocket = new Socket(server.getIPAddress(),server.getPort());
						ObjectOutputStream sgOut = new ObjectOutputStream(sgSocket.getOutputStream());
						sgOut.writeObject(p);
						sgOut.close();
						sgSocket.close();
					} // end while stream child size
				} //end if streamchild
			
				// Now activate the node on the machine
				Machine[] currentMachines = currentTable.getMachineArray(currentNode);
				for (int icm = 0; icm < currentMachines.length; icm++)
				{
					currSocket = new Socket(currentMachines[icm].getIPAddress(),currentMachines[icm].getConnectionListenerPort());
				
					System.out.println("Current operator is: " + currentNode.getOperator().getProperty("OBJECT_ID"));
				
					//mjba keep trying to connect until an open socket is returned
					//while(currSocket==null){
				
					//do{
					//	try{
					//		currSocket = new Socket(currentTable.getMachine(currentNode).getIPAddress(),currentTable.getMachine(currentNode).getConnectionListenerPort());
					//	} catch(Exception ex) {
					//		System.err.println("Couldn't create currSocket.  Trying again...");
					//		Thread.sleep(100);
					//	};
					//} while (currSocket == null || currSocket.isClosed());
					//System.out.println("Socket number is:" + currSocket.getLocalPort() + "TOSTRING:" + currSocket.toString());

					currOut = new ObjectOutputStream(currSocket.getOutputStream());
					
				Connection c = new Activate(currentNode.getOperator().getProperty("OBJECT_ID"));
				Packet p = new Packet("local",c);
			
				currOut.writeObject(p);
							
				// Now look and see where the parent is and connect it to the parent.
			
				Iterator it = currentNode.getParents();
	
				while(it.hasNext()){
			
					//Remember there may be more than 1 parent
					XATNode parentNode = (XATNode) it.next();

					if(currentMachines.length == 1 && currentTable.getMachineArray(parentNode).length == 1 && currentMachines[icm].equals(currentTable.getMachineArray(parentNode)[0])){
						//The parent is on the same machine as the child, so this will be connected later.
					} else {
						//The parent is on another machine so tell our machine how to connect it, and then
						//let the other machine know that it will be getting tuples from the child.
					
						//@TODO RIGHT NOW WE ONLY SUPPORT TREES, HERE IS WHERE THE CHANGE WILL HAVE TO BE MADE TO SUPPORT DAG's

						//Code changed on 4/26/2005 by Brad M. to allow for single parent on multiple machines.
						int queueNum = ((XATStreamOperator) currentNode.getOperator()).getOutputQueue().getQueueNumber();

						//System.out.println ("ConnectionListener.connect(): THE NAME OF THE CURRENT QUEUE IS" + ((XATStreamOperator) currentNode.getOperator()).getOutputQueue().getQueueName());
						
						Connection fc = null;
//						if (currentTable.getMachineArray(parentNode).length == 1) {
//							fc = new SendData(currentNode.getOperator().getProperty("OBJECT_ID"),queueNum,currentTable.getMachine(parentNode));
//							Packet np = new Packet("machines",fc);
//							// Send it to the childs machine
//							currOut.writeObject(np);
//
//							//Now tell the parent about the tuples it will be receiving
//							Connection tc = new ReceiveData(currentNode.getOperator().getProperty("OBJECT_ID"),parentNode.getOperator().getProperty("OBJECT_ID"),queueNum);
//							Packet mp = new Packet("machines",tc);
//										
//							//Send it to the parents machine
//							Socket parentSocket = new Socket(currentTable.getMachine(parentNode).getIPAddress(),currentTable.getMachine(parentNode).getConnectionListenerPort());
//							ObjectOutputStream parentOut = new ObjectOutputStream(parentSocket.getOutputStream());
//							parentOut.writeObject(mp);
//							parentOut.close();
//							parentSocket.close();						
//
//						} else {
							//comement out for the VLDB05-Demo -- Bin
							//fc = new SendData(currentNode.getOperator().getProperty("OBJECT_ID"),queueNum,currentTable.getMachines(parentNode),currentMachines[icm]);
							fc = new SendData(currentNode.getOperator().getProperty("OBJECT_ID"),queueNum,currentTable.getMachine(parentNode)); //fixed for VLDB2005
							Packet np = new Packet("machines",fc);
							// Send it to the childs machine
							currOut.writeObject(np);
							
							Machine[] pmachines = currentTable.getMachines(parentNode);
							for (int im=0; im < pmachines.length; im++){
								Machine curParentMachine = pmachines[im];
								if (curParentMachine.equals(currentMachines[icm])){
									//The parent is on the same machine as the child, so this will be connected later? or we may also need do something
								} else {
									((XATStreamOperator) currentNode.getOperator()).getOutputQueue().setQueueName(curParentMachine.getIPAddress());
									Connection rdc = new ReceiveData(currentNode.getOperator().getProperty("OBJECT_ID"),parentNode.getOperator().getProperty("OBJECT_ID"),queueNum);
									Packet prdc = new Packet("machines",rdc);
													
									//Send it to the parents machine
									Socket parentSocket = new Socket(curParentMachine.getIPAddress(),curParentMachine.getConnectionListenerPort());
									ObjectOutputStream parentOut = new ObjectOutputStream(parentSocket.getOutputStream());
									parentOut.writeObject(prdc);
									parentOut.close();
									parentSocket.close();	
								}
							}
//						}

					}
				
				}

				//If the iterator is null, there are no parents so connect it to an application
				if(currentNode.getParentSize() == 0){
				
					String currID = currentNode.getTreeID();
					Application currApp = new Application();
					//Application currApp = (Application) appTable.get(currID);
					//modified by Song Wang for multiple application support in one query plan
					ArrayList currAppList = (ArrayList) appTable.get(currID);
					if (currAppList.size()==1 && ((Application)currAppList.get(0)).getRootID()==0){
						//default case: one application for the query plan only. 
						currApp = (Application)currAppList.get(0);
					} else {
						//new case: multiple applications for one query plan. check the rootID field
						for (int i = 0; i < currAppList.size();i++ ) {
							if (((Application)currAppList.get(i)).getRootID()==((XATStreamOperator)currentNode.getOperator()).getOperatorID()){
								currApp = (Application)currAppList.get(i);
								break;
							}
						}
					}

					Connection ac = new SendData(currentNode.getOperator().getProperty("OBJECT_ID"),0,currApp.getIPAddress(),currApp.getPort());
					Packet ap = new Packet("connect",ac);
					//Send this info the the stream generator

					currOut.writeObject(ap);
				}
			
				currOut.close();
				currSocket.close();
			}
			} //finished with all of the nodes 
		
		}catch(ConnectException ce){
			System.out.println("=======ConnectException=========");
			System.out.println("Message: " + ce.getMessage());
			System.out.println("Cause " + ce.getCause());
			System.out.println(currSocket);
			ce.printStackTrace();
			System.out.println("=========================");
		}catch(Exception e){
			System.out.println("=======Exception=========");
			e.printStackTrace();
			System.out.println("=========================");
		}
		
		//PHEW DONE
		return true;
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.ConnectionManager#reconnect(edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable)
	 */
	public boolean reconnect(DistributionTable newTable) {
		
		try {
			DistributionTable diffTable = currentTable.difference(newTable);
			
			System.out.println("Difference:");
			System.out.println(diffTable);
			
			//For every node we will move, do the following steps:
			// NOTATION: (N) = New Machine, (O) = Old Machine, OQ = Output Queue
			//1. Connect the XATNode(N)'s OQ to Parent Node
			//2. Connect XATNode(O)'s Childrens OQ's to XATNode(N)
			//3. Deactivate XATNode(O) when not running
			//4. Send XATNode(O)'s state to XATNode(N)
			//5. Stop XATNode(0)'s oq's when they are empty
			//6. Activate XATNode(N)
			for(Enumeration e = diffTable.getNodeEnumeration(); e.hasMoreElements();){
			
				// Get the node and other essential information
				XATNode currNode = (XATNode) e.nextElement();
				XATNode parentNode = currNode.getParent();
				XATNode children[] = currNode.getOpChildren();
				System.out.println("currentNode" + currNode);
				System.out.println("parentNode" + parentNode);
				Machine parentMachine = currentTable.getMachine(parentNode);
				Machine currentMachine = currentTable.getMachine(currNode);
				Machine newMachine = newTable.getMachine(currNode);
				String opID = currNode.getStreamOperator().getProperty("OBJECT_ID");
				String parID = parentNode.getStreamOperator().getProperty("OBJECT_ID");
						
				System.out.println("Moving " + currNode + " from " + currentMachine.getMachineName() + " to " + newMachine.getMachineName());
				
				//1.
				//Only need to do this step if the node and parent are on different machines
				if(parentNode != null && !parentMachine.equals(newMachine)){
				
					int queueNum = currNode.getStreamOperator().getOutputQueue().getQueueNumber();

					Connection c = new SendData(opID,queueNum,parentMachine);
					Connection d = new ReceiveData(opID,parID,queueNum);
					Packet p = new Packet("reconnection",c);
					Packet q = new Packet("reconnection",d);
					Socket firstStep = new Socket(newMachine.getIPAddress(), newMachine.getConnectionListenerPort());
					ObjectOutputStream firstOut = new ObjectOutputStream(firstStep.getOutputStream());
					firstOut.writeObject(p);
					firstOut.close();
					firstStep.close();
					
					firstStep = new Socket(parentMachine.getIPAddress(),parentMachine.getConnectionListenerPort());
					firstOut = new ObjectOutputStream(firstStep.getOutputStream());
					firstOut.writeObject(q);
					firstOut.close();
					firstStep.close();
				}
				
				//2.
				//Now reconnect the OQ of each child
				for(int i = 0; i < children.length; i++){
					Machine childMachine = currentTable.getMachine(children[i]);
					String childID = children[i].getStreamOperator().getProperty("OBJECT_ID");
					//If the child and parent are on different machines
					//we have to connect them, just like Step 1.
					if(!childMachine.equals(newMachine)){
						int queueNum = children[i].getStreamOperator().getOutputQueue().getQueueNumber();
						
						Connection c = new SendData(childID,queueNum,newMachine);
						Connection d = new ReceiveData(childID,opID,queueNum);
						Packet p = new Packet("reconnection",c);
						Packet q = new Packet("reconnection",d);
					
						Socket secondStep = new Socket(childMachine.getIPAddress(), childMachine.getConnectionListenerPort());
						ObjectOutputStream firstOut = new ObjectOutputStream(secondStep.getOutputStream());
						firstOut.writeObject(p);
						firstOut.close();
						secondStep.close();
					
						secondStep = new Socket(newMachine.getIPAddress(),newMachine.getConnectionListenerPort());
						firstOut = new ObjectOutputStream(secondStep.getOutputStream());
						firstOut.writeObject(q);
						firstOut.close();
						secondStep.close();
					}
					//else remove a the connection if it is on the same machine.
					else{
						
						Connection c = new StopSend(childID);
						Packet p = new Packet("reconnection", c);
						Socket secondStep = new Socket(childMachine.getIPAddress(), childMachine.getConnectionListenerPort());
						ObjectOutputStream firstOut = new ObjectOutputStream(secondStep.getOutputStream());
						firstOut.writeObject(p);
						firstOut.close();
						secondStep.close();
					}
				}
				
				//3.
				// Now deactivate the operator on the old machine
				Connection c = new DeActivate(opID);
				Packet p = new Packet("reconnection",c);
				Socket thirdStep = new Socket(currentMachine.getIPAddress(),currentMachine.getConnectionListenerPort());
				ObjectOutputStream thirdOut = new ObjectOutputStream(thirdStep.getOutputStream());
				thirdOut.writeObject(p);
				thirdOut.close();
				thirdStep.close();
				
				//4.
				//Now send the state from old machine to new machine
				c = new SendState(opID,newMachine);
				p = new Packet("reconnection",c);
				Socket fourthStep = new Socket(currentMachine.getIPAddress(),currentMachine.getConnectionListenerPort());
				ObjectOutputStream fourthOut = new ObjectOutputStream(fourthStep.getOutputStream());
				fourthOut.writeObject(p);
				fourthOut.close();
				fourthStep.close();
				
				//5.
				//Stop sending data from the old queue when it is empty. 
				c = new StopSend(opID);
				p = new Packet("reconnection",c);
				Socket fifthStep = new Socket(currentMachine.getIPAddress(),currentMachine.getConnectionListenerPort());
				ObjectOutputStream fifthOut = new ObjectOutputStream(fifthStep.getOutputStream());
				fifthOut.writeObject(p);
				fifthOut.close();
				fifthStep.close();
				
				//6. 
				//Activate the operator on the new machine
				c = new Activate(opID);
				p = new Packet("reconnection",c);
				Socket sixthStep = new Socket(newMachine.getIPAddress(),newMachine.getConnectionListenerPort());
				ObjectOutputStream sixthOut = new ObjectOutputStream(sixthStep.getOutputStream());
				sixthOut.writeObject(p);
				sixthOut.close();
				sixthStep.close();
				
				//Finally update the currentTable to reflect the change
				currentTable.add(currNode,newMachine);
			}
			
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return false;
	}
	
	/* 
	 * connect a subplan
	 * assume that subplan is not directly connected to a stream.
	 */
	public boolean connectSubPlan(ArrayList subplanoperators) {
		
		Socket currSocket = null;
		ObjectOutputStream currOut;
		//ArrayList subplanoperators = subplan.getOperators();
		
		try{	
			System.out.println("distribution table is: " + currentTable.toString());
			System.out.println("connecting "+subplanoperators.size()+"operators");
			//Connect all of the nodes
			for(int count=0; count<subplanoperators.size(); count++) { 
				
				//Thread.sleep(100);
				XATNode currentNode = ((XATStreamOperator)subplanoperators.get(count)).getNode();
						
				//See if it connects to a stream
				if(currentNode.getStreamChildSize() > 0){
					//Get all of the streams it is connected to
				
					System.out.println("Streamsize of current node is: " + currentNode.getStreamChildSize());
					System.out.println("Children Operators of current node are: " + currentNode.getOpChildrenSize());
					//System.out.println("Child OP of current node is: " + ((XATNode)currentNode.getOpChild(0)).getOperator().getProperty("OBJECT_ID"));
				
					Iterator i = currentNode.getStreamChildNames();
								
					while(i.hasNext()){
						String streamName = (String) i.next();
		
						Server server = (Server) this.serverTable.get(streamName);				
						int queueNum = 0;
                    
						System.out.println("Current operator has : " + ((XATStreamOperator)currentNode.getOperator()).getInputQueueCount() + " input queues");
						System.out.println("Input queue for operator: " + ((XATStreamOperator)currentNode.getOperator()));
						System.out.println("for stream: " + streamName + " is " );
					
						XATQueue q = ((XATStreamOperator)currentNode.getOperator()).getInputQueue(streamName);
						System.out.println(q.getQueueNumber());
					
						queueNum = q.getQueueNumber();
						//System.out.println(q.getQueueNumber());
					
						//Create the connection object
						Connection c = new SendStream(streamName,streamTime,currentTable.getMachine(currentNode),currentNode.getOperator().getProperty("OBJECT_ID"),queueNum);
						Packet p = new Packet("connect",c);
					
						//Send the packet to the stream generator
						Socket sgSocket = new Socket(server.getIPAddress(),server.getPort());
						ObjectOutputStream sgOut = new ObjectOutputStream(sgSocket.getOutputStream());
						sgOut.writeObject(p);
						sgOut.close();
						sgSocket.close();
					} // end while stream child size
				} //end if streamchild
				
				// now activate the node on the machine
				Machine[] currentMachines = currentTable.getMachineArray(currentNode);
				for (int icm = 0; icm < currentMachines.length; icm++)
				{
					currSocket = new Socket(currentMachines[icm].getIPAddress(),currentMachines[icm].getConnectionListenerPort());
				
					System.out.println("Current operator is: " + currentNode.getOperator().getProperty("OBJECT_ID"));
				
					currOut = new ObjectOutputStream(currSocket.getOutputStream());
					
					Connection c = new Activate(currentNode.getOperator().getProperty("OBJECT_ID"));
					Packet p = new Packet("local",c);
			
					currOut.writeObject(p);
					
					// connects to its children
					Object[] childNodes = currentNode.getChildren();
					for(int cnum =0; cnum < childNodes.length; cnum++) {
						
						//Remember there may be more than 1 child
						if(childNodes[cnum] instanceof XATNode) {
							XATNode childNode = (XATNode) childNodes[cnum];

							if(currentMachines.length == 1 && currentTable.getMachineArray(childNode).length == 1 && currentMachines[icm].equals(currentTable.getMachine(childNode))){
								//The parent is on the same machine as the child, so this will be connected later.
								System.out.println("child on the same machine ");
							} else {
								System.out.println("child on different machine");
							
 								//Code changed on 4/26/2005 by Brad M. to allow for single parent on multiple machines.
 								int queueNum = cnum;
 					
 								Connection fc = null;
 								//comement out for the VLDB05-Demo -- Bin
 								//fc = new SendData(currentNode.getOperator().getProperty("OBJECT_ID"),queueNum,currentTable.getMachines(parentNode),currentMachines[icm]);
 								fc = new SendData(childNode.getOperator().getProperty("OBJECT_ID"),queueNum,currentTable.getMachine(currentNode));
 								Packet np = new Packet("machines",fc);
 								// Send it to the childs machine
 					
 								Socket childSocket = new Socket(currentTable.getMachine(childNode).getIPAddress(),currentTable.getMachine(childNode).getConnectionListenerPort());
 								ObjectOutputStream childOut = new ObjectOutputStream(childSocket.getOutputStream());
 								childOut.writeObject(np);
 								childOut.close();
 								childSocket.close();
 								
 								Connection rc = new ReceiveData(childNode.getOperator().getProperty("OBJECT_ID"), currentNode.getOperator().getProperty("OBJECT_ID"), queueNum);
 								Packet  rcp = new Packet("machines",rc);
 								currOut.writeObject(rcp);
 							}
 						}
					}

					// Now look and see where the parent is and connect it to
					// the parent.
					Iterator it = currentNode.getParents();
					while(it.hasNext()){
			
						//Remember there may be more than 1 parent
						XATNode parentNode = (XATNode) it.next();

						if(currentMachines.length == 1 && currentTable.getMachineArray(parentNode).length == 1 && currentMachines[icm].equals(currentTable.getMachine(parentNode))){
						//The parent is on the same machine as the child, so this will be connected later.
							// so this will be connected later.
							System.out.println("Parent on the same machine ");
						} else {
							System.out.println("Parent on differnt machine ");
							
							//Code changed on 4/26/2005 by Brad M. to allow for
							// single parent on multiple machines.
							int queueNum = ((XATStreamOperator) currentNode.getOperator()).getOutputQueue().getQueueNumber();

							Connection fc = null;
							//comement out for the VLDB05-Demo -- Bin
							//fc = new
							// SendData(currentNode.getOperator().getProperty("OBJECT_ID"),queueNum,currentTable.getMachines(parentNode),currentMachines[icm]);
							fc = new SendData(currentNode.getOperator().getProperty("OBJECT_ID"), queueNum, currentTable.getMachine(parentNode)); //fixed for VLDB2005
							Packet np = new Packet("machines", fc);
							// Send it to the parent machine
							currOut.writeObject(np);
							
							Machine[] pmachines = currentTable.getMachines(parentNode);
							for (int im=0; im < pmachines.length; im++){
								Machine curParentMachine = pmachines[im];
								if (curParentMachine.equals(currentMachines[icm])){
									//The parent is on the same machine as the
									// child, so this will be connected later?
									// or we may also need do something
								} else {
									((XATStreamOperator) currentNode.getOperator()).getOutputQueue().setQueueName(curParentMachine.getIPAddress());
									Connection rdc = new ReceiveData(currentNode.getOperator().getProperty("OBJECT_ID"),parentNode.getOperator().getProperty("OBJECT_ID"),queueNum);
									Packet prdc = new Packet("machines",rdc);
													
									//Send it to the parents machine
									Socket parentSocket = new Socket(curParentMachine.getIPAddress(),curParentMachine.getConnectionListenerPort());
									ObjectOutputStream parentOut = new ObjectOutputStream(parentSocket.getOutputStream());
									parentOut.writeObject(prdc);
									parentOut.close();
									parentSocket.close();	
								}
							}
						}
					}


					//If the iterator is null, there are no parents so connect it to an application
					if(currentNode.getParentSize() == 0){
						String currID = currentNode.getTreeID();
						Application currApp = (Application) appTable.get(currID);
						Connection ac = new SendData(currentNode.getOperator().getProperty("OBJECT_ID"),0,currApp.getIPAddress(),currApp.getPort());
						Packet ap = new Packet("connect",ac);
						
						//Send this info the the stream generator
						currOut.writeObject(ap);
					}
			
					currOut.close();
					currSocket.close();
					}
			} //finished with all of the nodes 
		
		}catch(ConnectException ce){
			System.out.println("=======ConnectException=========");
			System.out.println("Message: " + ce.getMessage());
			System.out.println("Cause " + ce.getCause());
			System.out.println(currSocket);
			ce.printStackTrace();
			System.out.println("=========================");
		}catch(Exception e){
			System.out.println("=======Exception=========");
			e.printStackTrace();
			System.out.println("=========================");
		}
		
		//PHEW DONE
		return true;
	}
}
