/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern;


//import java.io.BufferedWriter;
//import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.raindrop.utils.BlankString;
/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Feb 17, 2004
 * 
 * Insert comments here.
 */
public class DistributionTable {

//	private static Hashtable ht;
	private Hashtable ht;	
	/**
	 * 
	 */
	public DistributionTable() {
		ht = new Hashtable();
	}
	
	public Enumeration getNodeEnumeration(){
		return ht.keys();
	}
	
	public boolean contains(XATNode node){
		if(ht.containsKey(node)) return true;
		return false;	
	}
	
	public ArrayList getOperatorsOnMachine(Machine m){
		ArrayList ret = new ArrayList();
		Enumeration e = this.getNodeEnumeration();
		while(e.hasMoreElements()){
			XATNode n = (XATNode) e.nextElement();
			Machine[] ma = this.getMachines(n);
			for (int i=0; i<ma.length; i++){
				if(ma[i].equals(m))
					ret.add(n);				
			}
		}		
		return ret;
	}
	
	public XATNode[] getNodes(){
		
		XATNode[] nodes;
		Enumeration e = ht.keys();
		nodes = new XATNode[ht.size()];
		int counter = 0;
		while(e.hasMoreElements()){
			
			nodes[counter] = (XATNode) e.nextElement();
			counter++;
		}
		
		return nodes;
	}
	
	//-- Bin @ Dec.1,2004
	public Machine[] getMachines(XATNode node){
		Machine[] retMs;
		if (ht == null) {
			return null;
		} else {
			Object ret = ht.get(node);
			if (ret == null)
				return null;
			if (ret.getClass().isArray()){				
				return (Machine[])ret;
			} else {
				Machine[] rets = new Machine[1];
				rets[0] = (Machine)ret;
				return rets;
			}
		}		
	}
	
	public Machine getMachine(XATNode node){
		if (ht == null) {
			return null;
		} else {
			Object ret = ht.get(node);
			if (ret == null)
				return null;
			if (ret.getClass().isArray()){
				System.out.println("The node is distributed to more than one machine, please use getMachineArray to get all machines.");
				return null;
			} else {
				return (Machine)ret;
			}
			//return (Machine) ht.get(node);
		}
	}
	/* -- Bin @ Nov.29,2004
	 * In case of one operator is assigned to more than one machines
	 * then we do not know whether it is an Machine or Machine[], then we simply return Object type,
	 * the process is done by caller
	 */
	public Object getObjectByNode(XATNode node) {
		if ((ht == null) || (node==null)) {
			return null;
		} else {
			return ht.get(node);
		}		
	}
	/* -- Bin @ Nov.29,2004
	 * In case of one operator is assigned to more than one machines 
	 */
	public Machine[] getMachineArray(XATNode node){
		Machine[] retArray = null;
		if (ht == null) {
			return null;
		} else {
			Object ret = ht.get(node);
			if (ret == null)
				return null;
			if (ret.getClass().isArray()){
				retArray = (Machine[])ht.get(node);				
			} else {
				retArray = new Machine[1];
				retArray[0] = (Machine)ret;
			}
			return retArray;
		}
	}
	
	/* -- Bin @ Nov.29,2004
	 * In case of one operator is assigned to more than one machines
	 */
	public void add(XATNode node, Machine[] m){
		if (node == null) { 
			System.out.println("Null Node");
			return;
		}
		
		if (m == null) {
			System.out.println("Null Machine");
			return;
		}		 
		ht.put(node,m);
	}
	
	public void add(XATNode node, Machine m){
		if (node == null) { 
			System.out.println("Null Node");
			return;
		}
		
		if (m == null) {
			System.out.println("Null Machine");
			return;
		}
		 
		ht.put(node,m);
	}
	
	public void remove(XATNode node){
		ht.remove(node);
	}
	
	// this one might need to be adjusted in case of partitioned processing -- Bin @ Nov.29, 2004	
	public DistributionTable difference(DistributionTable newTable){
		
		DistributionTable differenceTable = new DistributionTable();
		
		Enumeration e1 = this.getNodeEnumeration();
		Enumeration e2 = newTable.getNodeEnumeration();
		
		while(e2.hasMoreElements()){
			
			XATNode node = (XATNode) e2.nextElement();
			
			if(this.contains(node)){
				//If the node is in the first table
				//compare to see if the machines are the same, if they arent
				// add it to the difference table
				
				if( !this.getMachine(node).equals(newTable.getMachine(node))){
					differenceTable.add(node, newTable.getMachine(node));
				}				
			}
			else{
				//The node is in the newer table but not the original
				differenceTable.add(node, newTable.getMachine(node));					
			}
		}
		
		return differenceTable;		
	}
	
	public String toString(){
		int maxLength = 10;
		String ret = new String("\n");
		ret+="--------------------------\n";
		ret+="|   Distribution Table   |\n";
		ret+="--------------------------\n";
		ret+="| Operator   |  Machine  |\n";
		ret+="--------------------------\n";
		for(Enumeration e = ht.keys(); e.hasMoreElements();){
			
			XATNode node = (XATNode) e.nextElement();
			Object m = ht.get(node);
			Machine[] ms = null;
			if (m.getClass().isArray()){
				ms = (Machine[])m;
			} else {
				ms = new Machine[1];
				ms[0] = (Machine)m;
			}
			for (int i=0; i<ms.length; i++) {
				Machine mx = ms[i];
				String operatorName = ((XATStreamOperator)node.getOperator()).toShortString();
				String machineName = mx.getMachineName();
				
				if(operatorName.length() > maxLength)
					operatorName = operatorName.substring(0,maxLength);
				if(operatorName.length() < maxLength){
					operatorName = operatorName.concat(BlankString.createWhiteSpace(maxLength-operatorName.length()));
				}
				
				if(machineName.length() > maxLength)
					machineName = machineName.substring(0,maxLength);
				if(machineName.length() < maxLength){
					machineName = machineName.concat(BlankString.createWhiteSpace(maxLength-machineName.length()));
				}				
				ret +="| " + operatorName + " | " + machineName +"|\n";				
			}
			//Machine m = (Machine) ht.get(node);			
		}
		ret+="--------------------------\n";
		return ret;
	}

	public XATNode getNode(String operatorNumber){
		
		Enumeration e = this.getNodeEnumeration();
		
		while(e.hasMoreElements()){
			
			XATNode temp = (XATNode) e.nextElement();
			if(temp.getStreamOperator().getProperty("OBJECT_ID").equals(operatorNumber))
				return temp;
		}
		
		return null;
	}
	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public Object clone() throws CloneNotSupportedException {
		DistributionTable ret = new DistributionTable();
		
		Enumeration e = this.getNodeEnumeration();
		while(e.hasMoreElements()){
			XATNode temp = (XATNode) e.nextElement();
			Machine m = (Machine) this.getMachine(temp);
			
			ret.add(temp,m);
		}
		
		return ret;
	}

}
