/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern;
import java.util.*;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.QueryPlan;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Feb 10, 2004
 * 
 * Insert comments here.
 */
public class GroupingDistribution extends DesignPattern {

	public GroupingDistribution(Machine[] m, QueryPlan[] qp){
		super(m,qp);
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.XATDesignPattern#distribute()
	 */
	public DistributionTable distribute() {
	System.out.println("Distributing Query Plan");
	
	DistributionTable table = new DistributionTable();
	
	try{
		int machineNum = 0;
		int machineSize = machines.length;
		int querySize = queryPlans.length;
		
		// Now decide how many machines a query can have
		int machinePerQuery[] = new int[querySize];
		
		int avgMachinePerQuery = machineSize / querySize;
		
		if(machineSize == 1){
			avgMachinePerQuery = 1;
		}
		
		//Give each query plan a portion of machines to use.
		for(int i = 0; i < machinePerQuery.length; i++){
			machinePerQuery[i] = avgMachinePerQuery;
		}
		//Give the extra machines out now
		int machinesLeft = machineSize - avgMachinePerQuery * querySize;
		
		for(int i = machinesLeft; i > 0; i--){
			machinePerQuery[i] += 1;
		}
		
		int machineCount = 0;

		for(int i = 0 ; i < this.queryPlans.length ; i++){
			XATNode[] nodes = queryPlans[i].getQueryPlan().convertToArray();
			int machinesICanUse = machinePerQuery[i];
			int nodesPerMachine = nodes.length / machinesICanUse;
			ArrayList nodesLeft = new ArrayList();
			for(int j = 0; j < nodes.length;j++){
				nodesLeft.add(nodes[j]);
			}
			if(machineSize == 1){
				machineCount = 0;
			}
			int extraNodes = nodes.length - (nodesPerMachine * machinesICanUse);
			
			while(nodesLeft.size() > 0){
				XATNode currNode = (XATNode) nodesLeft.get(0);
				int keep = 0;
				
				for(int l = 0; l < nodesLeft.size(); l++){
					currNode = (XATNode) nodesLeft.get(l);
					keep = l;
					if(currNode.getOpChildrenSize() == 0){
						break;
					}
				}
				 
				currNode = (XATNode) nodesLeft.remove(keep);
				
				XATNode nextNode;
				boolean failedOnce, failedTwice;
				
				for(int k = 0 ; k < nodesPerMachine + extraNodes;){
					k++;
					failedOnce = true; failedTwice = true;
					//System.out.println("Adding: " + currNode);
					table.add(currNode,this.machines[machineCount]);
					if(nodesLeft.size() > 0 && k < (nodesPerMachine + extraNodes)){
						if(currNode.getParentSize() > 0){
							nextNode = currNode.getParent();
							if(nodesLeft.contains(nextNode)){
								nodesLeft.remove(nextNode);
								currNode = nextNode;
								failedOnce = false;
							}
						}
						if(currNode.getOpChildrenSize() > 0 && failedOnce){
								nextNode = currNode.getOpChild(0);
								if(nodesLeft.contains(nextNode)){
									nodesLeft.remove(nextNode);
									currNode = nextNode;
									failedTwice = false;
								}
								if(failedTwice && currNode.getOpChildrenSize() > 1){
									nextNode = currNode.getOpChild(1);
									if(nodesLeft.contains(nextNode)){
										nodesLeft.remove(nextNode);
										currNode = nextNode;
										failedTwice = false;
									}
								}
						}
						if(failedOnce && failedTwice){
							nextNode = (XATNode) nodesLeft.remove(0);
							currNode = nextNode;
						}
					}
				}
				extraNodes = 0;
				machineCount++;
				
			}
			
			
	
		}
	
	}catch(Exception e){
		e.printStackTrace();
	}
		System.out.println("Here is what I have:");
		System.out.println(table);
		return table;
	}

}
