/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern;
import java.util.*;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.QueryPlan;

/**
 * @author Song Wnag
 * 
 * March, 2007
 * 
 * This pattern is used for pipelined multi-way join only.
 * The distribution is simply done by the operator ID
 * It requires the same number of machines as the pipelined sliced joins.
 */
public class OperatorIDDistribution extends DesignPattern {

	public OperatorIDDistribution(Machine[] m, QueryPlan[] qp){
		super(m,qp);
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.XATDesignPattern#distribute()
	 */
	public DistributionTable distribute() {
	System.out.println("Operator ID-based Distributing Query Plan");
	
	DistributionTable table = new DistributionTable();
	
	try{
		for(int i = 0 ; i < this.queryPlans.length ; i++){
			XATNode[] nodes = queryPlans[i].getQueryPlan().convertToArray();
			for(int j = 0; j < nodes.length;j++){
				XATNode currNode = (XATNode) nodes[j];
				//System.out.println("Adding: " + currNode);
				int machineNum = ((XATStreamOperator)currNode.getOperator()).getOperatorID()/10;
				table.add(currNode,this.machines[machineNum]);
			}
		}
	
	}catch(Exception e){
		e.printStackTrace();
	}
		System.out.println("Here is what I have:");
		System.out.println(table);
		return table;
	}

}
