/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
/*
 * Created on Nov 29, 2004
 * This class performs simple round roubin distribution for non-join query operators in the query plan
 * to available machines. For the join operator, it will be distributed all available machines
 * The output of the distribution contains, DistributionTable (operator<->machines), 
 * PartitionMap(split operator<->partitionMaps). Here partitionMaps includes information about
 * the partition IDs and machines.
 * Note that Dynamic queues will be created when the split operators are activated later
 */
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern;

import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.StreamSplitOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamUnionOperatorImp;

import java.util.Hashtable;
/**
 * @author Bin Liu
 */
public class RoundRobinPartitionDistribution extends DesignPattern {

	// hold the paritionmapping of each split operator in the query plan, partitions <-> ipaddress of the machine for each split operator
	Hashtable partitionMaps = null;
	DistributionTable table = null;
	// in case of separate union operator, we need to return the union machine,
	// this will be used when adapt the states since it only adapts among join operators
	Machine unionMachine = null; 
	
	public Machine getUnionMachine(){
		return unionMachine;
	}
	
	public RoundRobinPartitionDistribution(Machine[] m, QueryPlan[] qp){
		super(m,qp);
		partitionMaps = new Hashtable();
		table = new DistributionTable();
	}
	
	/* 
	 * Round roubin distribution of non-join operators and spread join operators to all machines
	 * Added at
	 */
	public DistributionTable distribute() {
		System.out.println("Roundrobin Partition Distributing Query Plan");				 
		// first, do the distribution
		try{
			int machineNum = 0;
			int machineSize = machines.length;
			for(int i = 0 ; i < this.queryPlans.length ; i++){
				XATNode[] nodes = ((XATTree)queryPlans[i].getQueryPlan()).convertToArray();
                System.out.println("\nQuery plan nodes: ");
                
                for(int k=0; k<nodes.length; k++) {
                    System.out.println(nodes[k]);
                }
				
				for(int j=0; j< nodes.length;j++){
					XATNode curNode = nodes[j];
					String test = curNode.getOperator().getClass().getName();
					if (test.endsWith("SymmMwayHashJoinOperatorImp")){
						// if it is a join operator, then spread the operator to all machines
						//for (int k=0; k<machineSize; k++){
						//	table.add(curNode,this.machines[k]);
						//}
						table.add(curNode,this.machines);
                        System.out.println("allocate all machines to op " + curNode);
					} else if(test.endsWith("WTransitiveHashJoinOperator")) {
					    table.add(curNode,this.machines);
                        System.out.println("allocate all machines to op " + curNode);
					} else if (test.indexOf("ReplicationAware") >= 0
							&& test.indexOf("Split") < 0
							&& test.indexOf("Merge") < 0) {
						Machine[] singleElementArray = new Machine[1];
						singleElementArray[0] = machines[machineNum];
						table.add(curNode,singleElementArray);
						machineNum++;
						if(machineNum >= machineSize)
							machineNum=0;
					} else if (test.endsWith("StreamSplitOperatorImp")){
						// if it is a split operator, and the partitioned_split flag is true, then 
						// distributed the operator into all machine, otherwise, distributed round roubin
						// first, to set the initialProperties of the operator
						if(test.indexOf("Replicated") >= 0 || test.indexOf("Spreadfire") >= 0) {
							table.add(curNode,this.machines[machineNum]);
							machineNum++;
							if(machineNum >= machineSize)
								machineNum=0;
						} else {
						((StreamSplitOperatorImp)curNode.getOperator()).setInitialProperties();
						if (((StreamSplitOperatorImp)curNode.getOperator()).getPartitionSplitFlag()){
							table.add(curNode,this.machines); 
						} else { // for other split operators, then the same
							table.add(curNode,this.machines[machineNum]);
                            System.out.println("allocate machine " + machineNum + " to op " + curNode);
							machineNum++;
							if(machineNum >= machineSize)
								machineNum=0;
						}
						}						
					} else {
						table.add(curNode,this.machines[machineNum]);
                        System.out.println("allocate machine " + machineNum + " to op " + curNode);
						machineNum++;
						if(machineNum >= machineSize)
							machineNum=0;
					}					
				} // end for nodes.length
			} //end for queryplans.length		
		}catch(Exception e){
			e.printStackTrace();
		} 		
		return table;
	}
	
	/*
	 * then, create initial partition maps for the split operator, 
	 * also use a simple round roubin of partition IDs over available machines 
	 */
	public Hashtable mapPartitions(){
		try{
			int machineNum = 0;
			int machineSize = machines.length;
			
			for(int i = 0 ; i < this.queryPlans.length ; i++){
				XATNode[] nodes = queryPlans[i].getQueryPlan().convertToArray();
				
				for(int j=0; j< nodes.length;j++){
					XATNode curNode = nodes[j];
					//begin the initial mapping ID <-> machine IP address
					String test = curNode.getOperator().getClass().getName();
					machineNum = 0;
					if (test.endsWith("StreamSplitOperatorImp")){
						StreamSplitOperatorImp curSplitNode = (StreamSplitOperatorImp)curNode.getOperator();
						curSplitNode.setInitialProperties();
						Hashtable partitionMapping = new Hashtable();
						int totalNumParts = curSplitNode.getNumberOfPartitions();
						for (int k=0; k<=totalNumParts; k++){
							partitionMapping.put(new Integer(k),this.machines[machineNum].getIPAddress());
							machineNum++;
							if(machineNum >= machineSize)
								machineNum=0;							
						}
						partitionMaps.put(curSplitNode,partitionMapping);
					} 				
				} // end for nodes.length
			} //end for queryplans.length
			//System.out.println("Debug");
		}catch(Exception e){
			e.printStackTrace();
		}		
		return partitionMaps;
	}
	/*
	 * create an skewed initial partition maps for the split operator, 
	 * first, put aound the ratio of partitions to a specied machine, then for the others,
	 * use simple round robin distribution 
	 * To have the union operator separated out from the rest join operators, simply assign the 
	 * machine has the union operator no partitions
	 */
	public Hashtable skewedMapPartitions(int skewedMachineID, double rate, boolean separateUnion){
		try{
			int machineNum = 0;
			if (separateUnion){
				// first find the machine id that has the union assigned if we separate the union from join
				if (table!=null) {				
					XATNode[] nodes = table.getNodes();
					for (int n=0; n<nodes.length; n++){
						if (nodes[n].getOperator() instanceof StreamUnionOperatorImp) {
							unionMachine = table.getMachine(nodes[n]);
							break;
						}
					}
				}
			} // end if separateUnion
			
			int machineSize = machines.length;
			
			for(int i = 0 ; i < this.queryPlans.length ; i++){
				XATNode[] nodes = queryPlans[i].getQueryPlan().convertToArray();
				
				for(int j=0; j< nodes.length;j++){
					XATNode curNode = nodes[j];
					//begin the initial mapping ID <-> machine IP address
					String test = curNode.getOperator().getClass().getName();
					machineNum = 0;
					if (test.endsWith("StreamSplitOperatorImp")){
						StreamSplitOperatorImp curSplitNode = (StreamSplitOperatorImp)curNode.getOperator();
						curSplitNode.setInitialProperties();
						Hashtable partitionMapping = new Hashtable();
						int totalNumParts = curSplitNode.getNumberOfPartitions();
						int skewedParts = (int)(totalNumParts * rate);
						for (int k=0; k<skewedParts; k++) {
							partitionMapping.put(new Integer(k),this.machines[skewedMachineID].getIPAddress());
						}
						for (int k=skewedParts; k<totalNumParts; k++){
							if ((separateUnion) && (unionMachine!=null)) {
								if ((this.machines[machineNum].getIPAddress()).equals(unionMachine.getIPAddress())) {
									machineNum++;
									if(machineNum >= machineSize)
										machineNum=0;							
								} 
							}
							partitionMapping.put(new Integer(k),this.machines[machineNum].getIPAddress());								
							machineNum++;
							
							if(machineNum >= machineSize)
								machineNum=0;							
						}
						partitionMaps.put(curSplitNode,partitionMapping);
					} 				
				} // end for nodes.length
			} //end for queryplans.length
			//System.out.println("Debug");
		}catch(Exception e){
			e.printStackTrace();
		}		
		return partitionMaps;
	}

}
