/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern;
import java.util.*;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.QueryPlan;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Feb 10, 2004
 * 
 * Insert comments here.
 */
public class WeightedGroupingDistribution extends DesignPattern {

	public WeightedGroupingDistribution(Machine[] m, QueryPlan[] qp){
		super(m,qp);
	}

	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.XATDesignPattern#distribute()
	 */
	public DistributionTable distribute() {
	System.out.println("Distributing Query Plan");
	
	DistributionTable table = new DistributionTable();
	
	try{
		int machineNum = 0;
		int machineSize = machines.length;
		int querySize = queryPlans.length;
		
		// Now decide how many machines a query can have
		int machinePerQuery[] = new int[querySize];
		
		int avgMachinePerQuery = machineSize / querySize;
		
		if(machineSize == 1){
			avgMachinePerQuery = 1;
		}
		
		//Give each query plan a portion of machines to use.
		for(int i = 0; i < machinePerQuery.length; i++){
			machinePerQuery[i] = avgMachinePerQuery;
		}
		//Give the extra machines out now
		int machinesLeft = machineSize - avgMachinePerQuery * querySize;
		
		for(int i = machinesLeft; i > 0; i--){
			machinePerQuery[i] += 1;
		}
		
		int machineCount = 0;

		for(int i = 0 ; i < this.queryPlans.length ; i++){
			XATNode[] nodes = queryPlans[i].getQueryPlan().convertToArray();
			int machinesICanUse = machinePerQuery[i];
			int totalNodeWeight=0;
			for (int m = 0; m < nodes.length; m++)
			{
					totalNodeWeight += weightOfOperator(nodes[m]);
			}
			
			int weightPerMachine = totalNodeWeight/machinesICanUse;

			int nodesPerMachine = nodes.length / machinesICanUse;
			ArrayList nodesLeft = new ArrayList();
			for(int j = 0; j < nodes.length;j++){
				nodesLeft.add(nodes[j]);
			}

			if(machineSize == 1){
				machineCount = 0;
			}
			int extraNodes = nodes.length - (nodesPerMachine * machinesICanUse);
			int extraWeight = totalNodeWeight - (weightPerMachine * machinesICanUse);
			
			while(nodesLeft.size() > 0){
				XATNode currNode = (XATNode) nodesLeft.get(0);
				int currMachineWeight = 0;
				
				while (currMachineWeight < weightPerMachine)
				{
					int remainingWeight = weightPerMachine - currMachineWeight;
					int[] weightlist = new int[nodesLeft.size()];
					for(int l = 0; l < nodesLeft.size(); l++){
						weightlist[l] = subtreeWeight(nodesLeft, (XATNode)nodesLeft.get(l));
					}

				
					int nearestweight = 0;
					int bestcandidate = 0;
					for(int l = 0; l < nodesLeft.size(); l++){
						if (remainingWeight - weightlist[l] <= remainingWeight - nearestweight
							&& remainingWeight - weightlist[l] + extraWeight >= 0)
						{
							bestcandidate = l;
							nearestweight = weightlist[l];
						}
					}
					
					//System.out.println("Bestcandidate=" + bestcandidate + " arraysize=" + nodesLeft.size());
					XATNode[] nodesToAdd = getSubtreeAsArray(nodesLeft, (XATNode)nodesLeft.get(bestcandidate));
					for(int m = 0; m < nodesToAdd.length; m++)
					{
						System.out.println("Now adding node to Machine " + (machineCount + 1));
						table.add(nodesToAdd[m],this.machines[machineCount]);
						nodesLeft.remove(nodesToAdd[m]);
						if (weightlist[bestcandidate] > weightPerMachine) 
							extraWeight -= weightlist[bestcandidate] - weightPerMachine + currMachineWeight;
					}

					
					currMachineWeight += weightlist[bestcandidate];
					System.out.println("For machine " + (machineCount + 1) + " current weight is " + currMachineWeight);
				}
				extraNodes = 0;
				if (machineCount < machinesICanUse - 1)
					machineCount++;
				else
				{
					while(nodesLeft.size() > 0)
					{
						table.add((XATNode)nodesLeft.remove(0),this.machines[machineCount]);
					}
				}
				
			}			
			
		}	
	
	
	}catch(Exception e){
		e.printStackTrace();
	}
		System.out.println("Here is what I have:");
		System.out.println(table);
		return table;
	}

	private int weightOfOperator(XATNode theOp)
	{
		String opName = theOp.getOperator().getClass().toString();

		if (opName.indexOf("Join") > -1) return 2;
		if (opName.indexOf("MultiStream") > -1) return 2;
		if (opName.indexOf("Aggregate") > -1) return 2;
		if (opName.indexOf("Distinct") > -1) return 2;
		if (opName.indexOf("Group") > -1) return 2;
		if (opName.indexOf("Intersection") > -1) return 2;
		if (opName.indexOf("Avg") > -1) return 2;
		if (opName.indexOf("Count") > -1) return 2;
		//we determined that Union was a stateless operator, so if that changes, change it here.

		return 1;
	}
	
	private int subtreeWeight(ArrayList undistributedNodes, XATNode theOp)
	{
		if(undistributedNodes.contains(theOp))
		{
			if (theOp.getOpChildrenSize() > 1) 
				return weightOfOperator(theOp)+ subtreeWeight(undistributedNodes, theOp.getOpChild(0))+ subtreeWeight(undistributedNodes, theOp.getOpChild(1));
			else if (theOp.getOpChildrenSize() == 1)
				return weightOfOperator(theOp) + subtreeWeight(undistributedNodes, theOp.getOpChild(0));
			else return weightOfOperator(theOp);
		}
		else return 0;
	}

	private XATNode[] getSubtreeAsArray(ArrayList undistributedNodes, XATNode rootNode)
	{
		if(undistributedNodes.contains(rootNode))
		{
			if (rootNode.getOpChildrenSize() > 1)
			{
				XATNode[] leftSubTree = getSubtreeAsArray(undistributedNodes, rootNode.getOpChild(0));
				XATNode[] rightSubTree = getSubtreeAsArray(undistributedNodes, rootNode.getOpChild(1));
				XATNode[] returnArray = new XATNode[leftSubTree.length + rightSubTree.length + 1];
				returnArray[0] = rootNode;
				for(int i = 1; i<= leftSubTree.length; i++)
					returnArray[i] = leftSubTree[i-1];
				for(int i = leftSubTree.length + 1; i <= leftSubTree.length+rightSubTree.length; i++)
					returnArray[i] = rightSubTree[i - leftSubTree.length -1];
				return returnArray;
			}
			else if (rootNode.getOpChildrenSize() == 1)
			{
				XATNode[] subTree = getSubtreeAsArray(undistributedNodes, rootNode.getOpChild(0));
				XATNode[] returnArray = new XATNode[subTree.length + 1];
				returnArray[0] = rootNode;
				for(int i = 1; i<= subTree.length; i++)
					returnArray[i] = subTree[i-1];
				return returnArray;
			}
			else
			{
				XATNode[] returnArray = new XATNode[1];
				returnArray[0] = rootNode;
				return returnArray;
			}
		}
		else return new XATNode[0];
	}
	
}
