/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.redistribution;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;
import java.util.*;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 */
public class Balance extends XATRedistributionPolicy {

	public Balance(){
		super();
	}
	
	// this should not be used if it is operator-based adaptation, only for partition adaptation only
	public ArrayList redistribute(DistributionTable currTable, CostTable currCost, StatisticsTable staTable, Machine[] elegibleMachines, Machine um) {
		return null;
	}
	
	public DistributionTable redistribute(
		DistributionTable currTable,
		CostTable currCost,
		XATNode[] ineligibleNodes,
		Machine[] ineligibleMachines) {
		boolean canBalance = true;
		
		if(ineligibleNodes != null){
			for(int i = 0; i < ineligibleNodes.length ; i++){
				this.inEligibleNodes.add(ineligibleNodes[i]);
			}
		}
		try {
			int count = 1;
			while(count < currCost.size()){
				//Take the heaviest node and move some of its work
				// to the lowest node.
				Machine maxMachine = currCost.NthMaxMachine(count);
				Machine minMachine = currCost.minMachine();
				double maxCost = currCost.getMachineCost(maxMachine);
				double minCost = currCost.minCost();
				count++;
				
				System.out.println("The heaviest loaded machine is: " + maxMachine.getMachineName());
				System.out.println("The lightest loaded machine is: " + minMachine.getMachineName());
				
				double diffInCost = (currCost.getMachineCost(maxMachine) - currCost.getMachineCost(minMachine)) / 2;
				System.out.println("The avg diff in cost is: " + diffInCost);
				// now find the operator that has this difference
				// and move it over!
				
				ArrayList maxNodes = currCost.getNodeCost(maxMachine);
				Iterator i = maxNodes.iterator();
				double closest = Double.MAX_VALUE;
				XATNode closestNode = null;
				
				while(i.hasNext()){
					NodeCost nc = (NodeCost) i.next();
					if(Math.abs(nc.getCost() - diffInCost) <= Math.abs(closest - diffInCost)){
						if(isEligible(nc.getNode())){
							closestNode = nc.getNode();
							closest = nc.getCost();
						}
					}	
				}
				
				//System.out.println("The closest node is: " + closestNode);
				
				//if we can find something small enough to move, then do it
//				if(closest <= (maxCost - minCost) / 2 && closest != 0){
//				if(closest != 0 && closestNode.getParent() != null){
				if(closestNode.getParent() != null){
					System.out.println("Find a node to move " + closestNode +". Now update the cost table");
								
					//Update the invidual node costs on the min machine
				
					double newCost = currCost.getMachineCost(minMachine) + closest;
					double oldCost = currCost.getMachineCost(minMachine);
					double totalNewCost = 0;
					ArrayList nodeCosts = currCost.getNodeCost(minMachine);
					if (nodeCosts != null) {
						for (int x = 0; x < nodeCosts.size(); x++) {
							NodeCost nc = (NodeCost) nodeCosts.get(x);
							double oldNodeCost = nc.getCost();
							XATNode tempNode = nc.getNode();
							double newNodeCost =
								oldNodeCost * oldCost / newCost;
							currCost.remove(minMachine, tempNode);
							currCost.put(
								minMachine,
								new NodeCost(tempNode, newNodeCost), x);
							totalNewCost += newNodeCost;
							
						}
//							Iterator it = nodeCosts.iterator();
//						while (it.hasNext()) {
//							NodeCost nc = (NodeCost) it.next();
//							double oldNodeCost = nc.getCost();
//							XATNode tempNode = nc.getNode();
//							double newNodeCost =
//								oldNodeCost * oldCost / newCost;
//							currCost.remove(minMachine, tempNode);
//							currCost.put(
//								minMachine,
//								new NodeCost(tempNode, newNodeCost));
//							totalNewCost += newNodeCost;
//						}
					}
				
					//move the node
					currCost.remove(maxMachine,closestNode);
					currCost.put(minMachine,new NodeCost(closestNode,1-totalNewCost));
				
					currCost.put(maxMachine,currCost.getMachineCost(maxMachine) - closest);
					currCost.put(minMachine,currCost.getMachineCost(minMachine) + closest);
				
					//System.out.println(currCost);
					
//					count = 1;
//					count++;					
				}
			}
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		System.out.println("Cost table after redistribute. ");
		System.out.println(currCost);
		return currCost.convertToDistributionTable();
	}

}
