/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.redistribution;

import java.util.ArrayList;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.CostTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.StatisticsTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;
//import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.realjoinoperator.WTransitiveHashJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.SymmMwayHashJoinOperatorImp;
/**
 * @author Bin Liu
 * This class decides which operator and the sender machine, the receiver machine, and the percentage
 * to be moved between two machines.
 */
public class BalanceMemoryByPartitions extends XATRedistributionPolicy {

	//double threshold = 0.5;
	//initial maximal available memory of query processor, 
	//it will be set when it is available from the statistics table
	double maxMemorySize = 10.0;  //MB
	double minRequiredMemorySize = 1000; //simulated memory usage (tuples_number)
	double differenceThreshold = 0.8;
	
	public BalanceMemoryByPartitions(){
		super();
		//Runtime runtime = Runtime.getRuntime();
		//maxMemorySize = (runtime.maxMemory()/(1024*1024));
		//maxMemorySize = 200.0; // a short-cut for now
	}
	
	public void setMinRequiredMemorySize(double m) {
	    this.minRequiredMemorySize = m;
	    System.out.println("set minRequiredMemorySize: " + m);
	}
	
	public void setDifferenceThreshold(double t) {
	    this.differenceThreshold = t;
	    System.out.println("Set differenceThreshold: " + t);
	}
	
	/*
	 * The memory usage should meet certain thresthold. Here i define a fixed number as the total
	 * available memory, since the total memory got from statistics is not 'real', it increases
	 * based on the available buffers until reaches maximal JVM memory.
	 * (This needs to be improved to reflect the different maximal JVM memory of different machines)
	 *  
	 * Simply get the machine with highest rawcost as sender, the machine with lowest rawcost as
	 * receiver, the rate is defined as (max-min)/2max,
	 */
	public ArrayList redistribute(DistributionTable currTable, CostTable currCost, StatisticsTable staTable, Machine[] elegibleMachines, Machine unionMachine){
		ArrayList ret = new ArrayList(5);
		
		Machine maxM=null;
		Machine minM=null;
		
		double max = currCost.maxRawCost();
		double min = currCost.minRawCost();
		
		if (unionMachine!=null) {
			//need to separte the cost of the union machine cost if the union is the current minimal
			double unionCost = currCost.getRawCost(unionMachine);
			if (min==unionCost){
				min = currCost.minRawCostExcept(unionCost);
			}
			if (max == unionCost) {
				return null;
			}
		} 
		double rate = 0.0;
		int operID = 0;

		// do not redistribute if it is below threshold
		if (max < minRequiredMemorySize) {
		    System.out.println("max memory usage is less than required memory size: " + this.minRequiredMemorySize);
			return null;
		}
		
		// do not redistribute if it is below threshold
		//if (((2*max)/maxMemorySize)<threshold) {
		//	return null;
		//}
		
		// too little difference between maximal and minimal, no point of redistributing
		if ((min/max)> differenceThreshold) {		
		    System.out.println("Little difference between max and min, no point of redistributing");
			return null;
		}
		
		// compute the rate
		//rate = (max-min)/(2*max);
		rate = (max-min)/2;
		
		// if it is too much data to be moved one time, i.e., >10M, reduce the rate
		//if ((rate*max)>10.0) {
		//	System.out.println("Too much data to be moved, Max:" + max + " rate:" + rate + "! reduce to 10M instead.");
		//	rate = 10.0/max;
		//}
				
		// find the sender and receiver machines
		for (int i=0; i<elegibleMachines.length; i++){
			if (max==currCost.getRawCost(elegibleMachines[i])) {
				maxM = elegibleMachines[i];
				continue;
			}
			if (min==currCost.getRawCost(elegibleMachines[i])) {
				minM = elegibleMachines[i];
				continue;
			}			
		}
		
		
		// the following is for the real main memory cost model only.
		// if the receiver machine is too full, i.e., more than 80%, then it is not a good idea to adapt
		// or we can try to find another machine? now we simply stop adaptation
		// For the simulated cost model, here the max_memory is still the real maximal JVM memory  
		
		double minMemPct = currCost.getMachineCost(minM);
		double maxMemPct = currCost.getMachineCost(maxM);
//		if(maxMemPct < minMemPct) {
//		    System.out.println("\nminMemPct: " + minMemPct + " is larger than maxMemPct: " + maxMemPct);
//		    System.out.println("Do not adapt: Simulated memory does not conform to the real memory usage.");
//		    return null;
//		}
		if(minMemPct > 0.8) {
		    System.out.println("\nDo not adapt: The receiver machine " + minM.getMachineName() + " is more than 80% full.");
		    return null;
		}
		    
//		LightWeightStatistic lws = staTable.get(minM);
//		String staVal =  lws.get("MAX_MEMORY");
//		double maxMinMachineMem = 0;
//		if (staVal != null){
//			maxMinMachineMem = Double.parseDouble(staVal);
//		}
//				
//		
//		if ((min/maxMinMachineMem)>0.8) {
//			return null;
//		} else {
//			System.out.println("min/minMachineMaxMem = " + (min/maxMinMachineMem) + "minMaxMem:" + maxMinMachineMem);
//		}
		// end of real main memory cost model
		
/*		// the following is for the simulated cost mode, kind of hard-coded for the maxMemorySize
		if (((min + (max-min)/2)/maxMemorySize) >0.8) {
			return null;
		} else {
			System.out.println("min/minMachineMaxMem = " + ((min + (max-min)/2)/maxMemorySize) + "minMaxMem:" + maxMemorySize);
		}
*/		
		// now, we have the first maxAvailable memory, then set the maxMemorySize if it is 0.0
		//if (maxMemorySize==0.0) {
		//	maxMemorySize = maxMinMachineMem;
			//check the thresthold again, return null if is it not met
		//	if ((max/maxMemorySize)<threshold) {
		//		return null;
		//	}			
		//}
		
		// find the first join operator
		XATNode[] nodes = currTable.getNodes();
		for (int i=0; i<nodes.length; i++){
			XATStreamOperator currOp = (XATStreamOperator)(nodes[i]).getOperator();
			if (currOp instanceof SymmMwayHashJoinOperatorImp){
				// yes, this is the operator we need,
				operID = currOp.getOperatorID();
				break;
			}
			if (currOp instanceof WTransitiveHashJoinOperator){
				// yes, this is the operator we need,
				operID = currOp.getOperatorID();
				break;
			}
		}
		/*
		///debug usage, always, maxM=230, minM = 33
		for (int i=0; i<elegibleMachines.length; i++){
			if ((elegibleMachines[i].getIPAddress()).equals("130.215.27.230")) {
				maxM = elegibleMachines[i];
				continue;
			}
			if ((elegibleMachines[i].getIPAddress()).equals("130.215.28.33")) {
				minM = elegibleMachines[i];
				continue;
			}			
		}
		*/
		// (0) sender Machine, (1) receiver Machine, (2) operator ID, (3) rate
		ret.add(maxM);
		ret.add(minM);
		ret.add(new String(""+ operID));
		ret.add(new String("" + rate));
		System.out.println("Move " + rate + " " + " tuples "+ operID + " from " + maxM.getMachineName() + " to " + minM.getMachineName());
		
		// increase thresthold for the next adaptation, debug code
		// threshold += 0.1;
		
		return ret;
	} //
	
	/* 
	 * This method is not been used if the redistribution is at partition level,
	 * It is here since we use the existing XATRedistributionPolicy
	 */
	public DistributionTable redistribute(DistributionTable currTable,
			CostTable currCost, XATNode[] elegibleNodes,
			Machine[] elegibleMachines) {
		return null;
	}

}
