/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.redistribution;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;
import java.util.*;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 */
public class MinimizeNetworkConnection extends XATRedistributionPolicy {

	private boolean init = false;
	private Hashtable machineRates = new Hashtable();

	public MinimizeNetworkConnection(){
		super();
	}
	
	// this should not be used if it is operator-based adaptation, only for partition adaptation only
	public ArrayList redistribute(DistributionTable currTable, CostTable currCost, StatisticsTable staTable, Machine[] elegibleMachines, Machine um) {
		return null;
	}
	
	public DistributionTable redistribute(
		DistributionTable currTable,
		CostTable currCost,
		XATNode[] ineligibleNodes,
		Machine[] ineligibleMachines) {
			
			DistributionTable newTable = null;
			
			if(!init){
				init(currCost);
			}
			
			System.out.println("Starting Redistribution...");
			
			try {
					
				
				if(ineligibleNodes != null){
					for(int i = 0; i < ineligibleNodes.length ; i++){
						this.inEligibleNodes.add(ineligibleNodes[i]);
					}
				}
				
				newTable = (DistributionTable) currTable.clone();
				
				// Take the machine with the lowest output rate, and see if you can move some of the nodes
				Machine minMachine = currCost.minMachine();
				ArrayList nodes = currTable.getOperatorsOnMachine(minMachine);
				XATNode[] fragment = findSmallestFragment(nodes);
				
				if(fragment == null){
					System.out.println("The fragment is empty");
				}
				else{
					Machine parMachine = null;
					Machine childMachine = null;
					
					for(int i = 0; i < fragment.length; i++){
						//System.out.println("The Fragment is: " + fragment[i]);
						newTable.remove(fragment[i]);
						parMachine = currTable.getMachine(fragment[i].getParent());
						//parMachine = currTable.getMachine(fragment[i]);
						childMachine = currTable.getMachine(fragment[i].getOpChild(0));
						int opsOnParMachine = newTable.getOperatorsOnMachine(parMachine).size();
						int opsOnChildMachine = newTable.getOperatorsOnMachine(childMachine).size();
						
						if(opsOnParMachine >= opsOnChildMachine){
							newTable.add(fragment[i],childMachine);
						}
						else{
							newTable.add(fragment[i],parMachine);
						}	
					}
				}
				
				// Take the machine with the highest output rate, to try to alleviate it a bit.
				Machine maxMachine = currCost.maxMachine();
				nodes = currTable.getOperatorsOnMachine(maxMachine);
				fragment = findSmallestFragment(nodes);
				
				if(fragment == null){
					System.out.println("The fragment is empty");
				}
				else{
					Machine parMachine = null;
					Machine childMachine = null;
					
					for(int i = 0; i < fragment.length; i++){
						//System.out.println("The Fragment is: " + fragment[i]);
						newTable.remove(fragment[i]);
						parMachine = currTable.getMachine(fragment[i].getParent());
						//parMachine = currTable.getMachine(fragment[i]);
						childMachine = currTable.getMachine(fragment[i].getOpChild(0));
						int opsOnParMachine = newTable.getOperatorsOnMachine(parMachine).size();
						int opsOnChildMachine = newTable.getOperatorsOnMachine(childMachine).size();
						
						if(opsOnParMachine >= opsOnChildMachine){
							newTable.add(fragment[i],childMachine);
						}
						else{
							newTable.add(fragment[i],parMachine);
						}	
					}
				}
				
				//Now go through each machine and see if the output rate has fallen more than 10%
				// if it has, try to balance it out a bit..
				Enumeration e = currCost.getMachines();
				
				while(e.hasMoreElements()){
					Machine m = (Machine) e.nextElement();
					double currentCost = currCost.getMachineCost(m);
					double cachedCost = ((Double) this.machineRates.get(m)).doubleValue();
					
					if(((cachedCost - currentCost) / cachedCost) > .10){
						this.machineRates.put(m,new Double(currentCost));
	
						nodes = currTable.getOperatorsOnMachine(m);
						fragment = findSmallestFragment(nodes);
				
						if(fragment == null){
							System.out.println("The fragment is empty");
						}
						else{
							Machine parMachine = null;
							Machine childMachine = null;
					
							for(int i = 0; i < fragment.length; i++){
								//System.out.println("The Fragment is: " + fragment[i]);
								newTable.remove(fragment[i]);
								parMachine = currTable.getMachine(fragment[i].getParent());
								childMachine = currTable.getMachine(fragment[i].getOpChild(0));
								int opsOnParMachine = newTable.getOperatorsOnMachine(parMachine).size();
								int opsOnChildMachine = newTable.getOperatorsOnMachine(childMachine).size();
						
								if(opsOnParMachine > opsOnChildMachine){
									newTable.add(fragment[i],childMachine);
								}
								else{
									newTable.add(fragment[i],parMachine);
								}	
							}
						}					
					}
				}
				
			} catch (CloneNotSupportedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			return newTable;
				
	}

	private XATNode[] findSmallestFragment(ArrayList nodes){
		//System.out.println("Finding Smallest Fragment from: " + nodes.toString());
		
		//boolean smallFragment = true;
		ArrayList temp = new ArrayList();
		//XATNode[] temp = new XATNode[1];
		
		for(int i = 0; i < nodes.size(); i++){
			//temp[0] = (XATNode) nodes.get(i);
			XATNode tempnode = (XATNode) nodes.get(i);
			//smallFragment = true;
			//System.out.println("Trying: "+ temp[0]);
			XATNode parent = tempnode.getParent();
			XATNode[] children = tempnode.getOpChildren();
			
			// comment out for vldb 2005 demo to relax the redistributing condition.
//			if(nodes.contains(parent)){
//				System.out.println(temp[0] + " is not a small fragment because the parent is on the same machine.");
//				smallFragment = false;
//			}
//			
//			for(int j = 0; j < children.length; j++){
//				if(nodes.contains(children[j])){
//					System.out.println(temp[0] + " is not a small fragment because a child is on the same machine.");
//					smallFragment = false;
//				}
//			}
			
			if(tempnode.getParentSize() == 0){
				//smallFragment = false;
				//System.out.println(tempnode + " is not a small fragment because the node is a root");
				continue;
			}
			
			if( tempnode.getStreamChildSize() > 0){
				//smallFragment = false;
				//System.out.println(tempnode + " is not a small fragment because the node is connected to a stream.");
				continue;
			}
			temp.add(tempnode);
		}
		if(temp.size() > 0){
			//System.out.println("Returning " + temp);
			return (XATNode[])temp.toArray(new XATNode[] {});
		}
		System.out.println("Returning null");
		return null;	
	}
	
	private void init(CostTable currCost){
		Enumeration e = currCost.getMachines();
		while(e.hasMoreElements()){
			Machine m = (Machine) e.nextElement();
			double cost = currCost.getRawCost(m);
			this.machineRates.put(m,new Double(cost));
		}
	}
}
