/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.redistribution;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;

import java.util.*;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 */
public class SwitchSameOperators extends XATRedistributionPolicy {

	private boolean init = false;
	private Hashtable machineRates = new Hashtable();
	//Change this line to the operators you want to switch each time.
	private static final String[] OPSTOROTATE = {"2","3","4","5"};
	private Machine[] machines = new Machine[2];
	
	public SwitchSameOperators(){
		super();
	}

	// this should not be used if it is operator-based adaptation, only for partition adaptation only
	public ArrayList redistribute(DistributionTable currTable, CostTable currCost, StatisticsTable staTable, Machine[] elegibleMachines, Machine um) {
		return null;
	}
	
	public DistributionTable redistribute(
		DistributionTable currTable,
		CostTable currCost,
		XATNode[] ineligibleNodes,
		Machine[] ineligibleMachines) {
			
			DistributionTable newTable = null;
			
			try {
				
				newTable = (DistributionTable) currTable.clone();
				
				// Assume we are rotating between two machines
				Enumeration e = currCost.getMachines();
				int count = 0;
				while(e.hasMoreElements()){
					Machine tempM = (Machine) e.nextElement();
					machines[count] = tempM;
					count++;
				}
				
				for(int i = 0; i < OPSTOROTATE.length; i++){
					String opNumber = OPSTOROTATE[i];
					System.out.println("Rotating "+ opNumber);
					XATStreamOperator operator = null;
					
					Enumeration f = currTable.getNodeEnumeration();
					for (; f.hasMoreElements();) {
						XATStreamOperator temp = ((XATNode) f.nextElement()).getStreamOperator();
						if (temp
							.getProperty("OBJECT_ID")
							.equals(opNumber)) {
							operator = temp;
							break;
						}
					}
					
					Machine oldMachine = currTable.getMachine(operator.getNode());
					Machine newMachine = getNextMachine(oldMachine);
					newTable.add(operator.getNode(),newMachine);

				}
				
				
				
			} catch (CloneNotSupportedException e) {
				e.printStackTrace();
			}
			return newTable;	
			
	}

	private Machine getNextMachine(Machine m){
		if(m.equals(machines[0])){
			return machines[1];
		}
		else{
			return machines[0];
		}
	}

}
