/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.redistribution;

import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 */
public abstract class XATRedistributionPolicy {

	protected ArrayList inEligibleNodes = new ArrayList();
	
	public abstract DistributionTable redistribute(DistributionTable currTable, CostTable currCost, XATNode[] elegibleNodes, Machine[] elegibleMachines);
	public abstract ArrayList redistribute(DistributionTable currTable, CostTable currCost, StatisticsTable staTable, Machine[] elegibleMachines, Machine um);
	
	/**
	 * Returns whether or not the given node can be moved.
	 * 
	 * @param node
	 * @return
	 */
	protected boolean isEligible(XATNode node){
		
		//if(node.getChildrenSize() == 0)
		//	return false;
			
		if(node.getStreamChildSize() > 0) {
			System.out.println("node " + node +" connects to streams, cannot move. ");
			return false;
		}
		
		if(node.getParent() == null) {
			System.out.println("Root node "+node+" cannot move");
		}
		
		if(inEligibleNodes.contains(node))
			return false;
		
		return true;
	}
}
