/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.manager;

import java.io.ObjectOutputStream;
import java.io.IOException;
import java.net.Socket;
import java.text.DecimalFormat;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.DistributionConfiguration;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Connection;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.DeActivate;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.Activate;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ReceiveData;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.SendData;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.adprotocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.protocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.DeActivate;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.manager.DistributionManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.DistributedExecutionController;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.SystemObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsReference;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.costmodel.StatisticsTable;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.spreadfire.ReplicatedStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WOperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
/**
 * @author bmombe1
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class SpreadfireManager {

	public static double HIGH_THRESHOLD = 0.0;
	public static double LOW_THRESHOLD = 0.0;
	
    /**
     * Sets the values for high and low thresholds based on the values contained in the config file, if available.
     * <P>
     * Runs at query processor
     * 
     * @return void
     * @param config  the system's DistributionConfiguration
     * @since 1.7
     * 
     */  
	public static void setThresholds(DistributionConfiguration config)
	{
	    if(config.getProperty("SPREADFIRE_HIGH_THRESHOLD") == null
	            || Double.parseDouble(config.getProperty("SPREADFIRE_HIGH_THRESHOLD")) <= 0.0) 
	        HIGH_THRESHOLD = 0.7;
	    else
	        HIGH_THRESHOLD = Double.parseDouble(config.getProperty("SPREADFIRE_HIGH_THRESHOLD"));
	    
	    if(config.getProperty("SPREADFIRE_LOW_THRESHOLD") == null
	            || Double.parseDouble(config.getProperty("SPREADFIRE_LOW_THRESHOLD")) <= 0.0)
	        LOW_THRESHOLD = 0.3;
	    else
	        LOW_THRESHOLD = Double.parseDouble(config.getProperty("SPREADFIRE_LOW_THRESHOLD"));
	}
	
    
    /**
     * Sets the maximum number of cells that can be used for all of the active states on the local machine.
     * This function is used by the memory saturation calculator.
     * 
     * @return the value applied to the system property MAX_TOTAL_STATE_SIZE
     * @param void
     * @since 1.7
     * 
     */
	public static long setMaxStateSize()
	{
        boolean DEBUG_SIMPLE_CALC = false;
        boolean DEBUG_COUNT_TUPLES = false;
        boolean DEBUG_COUNT_CELLS = true;
	    //Being very sloppy with memory, we're going to allow 128 bytes per cell.
	    //Now we calculate the allowable number of tuples per state based on the
	    //  following formula:
	    // (Runtime max memory - input queue allocation) 
	    //     / 256 bytes (liberal allocation for each tuple) 
	    //     * operator input selectivity / total plan selectivity
	    
		//First thing we'll do here is check for a large saturation.
		//If we go above the threshold we'll pick an op and get rid of a token.
	    double statetotal = 0.0;
        double queuetotal = 0.0;
        
        if(DEBUG_COUNT_TUPLES)
        {
            StatisticsGatherer stats = StatisticsGatherer.getStatisticsGather();
            queuetotal = Double.parseDouble(stats.getValue("TOTAL_TUPLES_IN_QUEUES",
                                                        stats.getQueryObject("TOTAL_TUPLES_IN_QUEUES", 
                                                                             StatisticsReference.LAST)));
            
            statetotal = Double.parseDouble(stats.getValue("TOTAL_TUPLES_IN_STATES", 
                                                       stats.getQueryObject("TOTAL_TUPLES_IN_STATES", 
                                                                            StatisticsReference.LAST)));
        }
        
        if(DEBUG_COUNT_CELLS)
        {
            Iterator it = StatisticsGatherer.getStatisticsGather().getAllKnownOperators();
            while (it.hasNext())
            {
                XATStreamOperator op = (XATStreamOperator)it.next();
                if(op instanceof ReplicatedStreamOperator)
                {
                    for(int i = 0; i < ((ReplicatedStreamOperator)op).getState().length; i++)
                    {
                        WOperatorState s = (WOperatorState)((ReplicatedStreamOperator)op).getNthState(i);
                         XATTuple t =  s.getEarlistTuple();
                         if(t != null)statetotal += s.getStateSize() * t.getLength();
                    }
                    
                    for(int i = 0; i < ((ReplicatedStreamOperator)op).getInputQueueCount(); i++)
                    {
                        XATQueue q = op.getInputQueue(i);
                        if (q.getSchema() != null)
                            queuetotal += q.getQueueSize() * q.getSchema().getLength();
                    }
                }
            }
        }
//        "NUMBER_OF_TUPLES_IN_INPUT_QUEUES"
//        "NUMBER_OF_TUPLES_IN_EACH_STATE"
//        "NUMBER_OF_TUPLES_IN_LEFT_STATE"
//        "NUMBER_OF_TUPLES_IN_RIGHT_STATE"
//        
        double maxstatememory;
        DistributedExecutionController exec = DistributedExecutionController.getDistributedExecutionController();
        if(DEBUG_SIMPLE_CALC)
            maxstatememory = (Runtime.getRuntime().maxMemory() 
                                - exec.getSystemOverhead())
                              / 2;
        else
            maxstatememory = (Runtime.getRuntime().maxMemory() 
                                - exec.getSystemOverhead())
                                * statetotal / (statetotal + queuetotal);
                                //* Runtime.getRuntime().maxMemory()
                                //  / (Runtime.getRuntime().maxMemory() - Runtime.getRuntime().freeMemory() - exec.getSystemOverhead());

        //int numstates = 0;
	    //figure out selectivities.
	    //XATNode[] allnodes = exec.getCurrentScheduler().getAllXATNodes();
	   // double totalselectivity = 0.0;
	   // for(int i = 0; i < allnodes.length; i++)
	   // {
	       // if(allnodes[i].getStreamOperator().isActive())
	      //  {
	      //      java.util.Hashtable h = exec.lws.get(allnodes[i]
	      //                                             .getStreamOperator()
	      //                                             .getLightWeightQueryObject());
	            
	      //      double currentselectivity = Double.parseDouble((String)h.get("SELECTIVITY"));
	            
	      //      if (allnodes[i].getStreamOperator().getClass().getName().indexOf("Replicat") >= 0
	      //              && allnodes[i].getStreamOperator().getClass().getName().indexOf("Split") < 0
	     //               && allnodes[i].getStreamOperator().getClass().getName().indexOf("Merge") < 0)
	      //          totalselectivity += currentselectivity;
	      //  }
	   //     numstates += allnodes[i].getStreamOperator().getStateSize();
	   // }
	   // for(int i = 0; i < allnodes.length; i++)
	  //  {
	  //      try {
	  //      allnodes[i].getStreamOperator().setProperty("MAX_STATE_SIZE", "" + (maxstatememory / numstates));
	  //      }catch(Exception e) {}
	  //  }
	    System.setProperty("MAX_TOTAL_STATE_SIZE", "" + maxstatememory / 1024.0);
	    return (long)maxstatememory / 1024;
	}
	
	//passToken (procedure)
	// params: DistributionTable,
	//         StatisticsTable, 
	//         array of all Machines in plan, 
	//         operator ID, and 
	//         ordinal number of the state in the operator.
	//Runs at Distribution Manager
    /**
     * pass a tuple storage token away from its current residence
     * <P>
     * Runs At DistributionManager
     * @return void
     * @param ddt       the system's distribution table
     * @param table     the system's statistics table
     * @param m         array of all machines
     * @param opID      operatorID of the operator in question
     * @param statenum  the sequential order number of the state in question
     * @since 1.2
     */
	public static void passToken(DistributionTable ddt, StatisticsTable table, Machine[] m, int opID, int statenum)
	{
		// Steps to achieving this process here.
		// First, set up a thingy.
		//   Let's assume that it's a hashtable.
		//   Index with op-id and state number all in a string.
		//   The keyed value will be the machine.
		// Then, read off the thingy.
		// Then, decide what to do with the thingy.
		// Then tell the whatsit to do a flurble.
		

		// At this point, we're going to be looking for a machine to take the token.
		// Here's the order that we'll be expecting to do it:
		//#1: look for a machine that's below "low" threshold in memory saturation.  
		//  If there are multiple hits, find the one closest to the threshold so that 
		//  we can eliminate unnecessary nodes.
		//#2: add a new node, the one with the lowest saturation.
		//#3: look for an active machine that's below "high" threshold.  
		//  If there are multiple hits, find the one with
		//  the least saturation.
		//#4: if all else fails, we've hit limit and should abort the execution.
		
		//Let's do #1:
		double currentlow = 0.0;
		Machine lowestmachine = null;
		//Since this is active ops only, we'll get the machine array from the DT:
        String opIDStr = Integer.toString(opID);
		Machine[] activeMachines = ddt.getMachineArray(ddt.getNode(opIDStr));
		for(int i = 0; i < activeMachines.length; i++)
		{
			LightWeightStatistic mls = table.get(activeMachines[i]);
			double currentpct = Double.parseDouble(mls.get("MEMORY_SATURATION"));
			if(currentpct > currentlow && currentpct < LOW_THRESHOLD)
			{
				currentlow = currentpct;
				lowestmachine = activeMachines[i];
			}
		}
		
		//Now #2 if #1 fails
		if(lowestmachine == null)
		{
			Machine[] inactiveMachines = new Machine[m.length - activeMachines.length];
			int k = 0;
			for(int i = 0; i < m.length; i++)
			{
				boolean foundMachine = false;
				for(int j = 0; j < activeMachines.length; j++)
				{
					if(m[i].equals(activeMachines[j]))
						foundMachine = true;
				}
				if(!foundMachine) inactiveMachines[k++] = m[i];
			}

			currentlow = 1.0;
            System.out.println("Saturation for nonparticipating machines:");
			for(int i = 0; i < inactiveMachines.length; i++)
			{
				LightWeightStatistic mls = table.get(inactiveMachines[i]);
                try {
                    double currentpct = Double.parseDouble(mls.get("MEMORY_SATURATION"));
                    if (Double.isNaN(currentpct)) currentpct = 0.0;
                    System.out.println(inactiveMachines[i].getMachineName() + ": " + currentpct);
                    
                    if(currentpct < currentlow)
                    {
                        currentlow = currentpct;
                        lowestmachine = inactiveMachines[i];
                    }
                }
                catch (NullPointerException npe)
                {
                    System.err.println("Getting MEMORY_SATURATION failed for machine " + inactiveMachines[i].getMachineName());
                }
			}
			if(lowestmachine != null)
			{
				//Here we'll activate the operator, but passing the token will come a little later.
				activateOperator(ddt, m, ddt.getNode(Integer.toString(opID)), lowestmachine);
			}
		}
		
		//Now #3 if #2 fails
		if(lowestmachine == null)
		{
			currentlow = 1.0;
			for(int i = 0; i < activeMachines.length; i++)
			{
				LightWeightStatistic mls = table.get(activeMachines[i]);
				double currentpct = Double.parseDouble(mls.get("MEMORY_SATURATION"));
				if(currentpct < currentlow)
				{
					currentlow = currentpct;
					lowestmachine = activeMachines[i];
				}
			}		
		}
        
        //Now we have to figure out some stats about the appropriate split operator.
        XATNode childNode = ddt.getNode(opIDStr).getOpChild(statenum);
		XATStreamOperator childOperator = childNode.getStreamOperator();
        int childID = childOperator.getObjectID();
        Machine childMachine = ddt.getMachine(childNode);
        
        
		ReplicationStyleProtocol p = new PassStorageToken(childID,
				statenum,
				lowestmachine.getMachineName());
        
		try {
			Socket s = new Socket(childMachine.getIPAddress(), childMachine.getConnectionListenerPort());
		
			new ObjectOutputStream(s.getOutputStream()).writeObject(new Packet("SPREADFIRE", p));
			s.close();
		}
		catch(java.net.UnknownHostException uhe) { uhe.printStackTrace(); }
		catch(java.io.IOException ioe) {ioe.printStackTrace();}
	}
	
    /**
     * Shuts down an empty-state operator partition
     * <P>
     * Runs at distribution manager.
     * 
     * @return void
     * @param ddt the system's distribution table
     * @param m an array with all machines
     * @param mesg a DeactivateEmptyStateOperator message from the Distribution Manager
     * @since 1.2
     */
	public static void deactivateOperator(DistributionTable ddt, Machine[] m, DeactivateEmptyStateOperator mesg)
	{
		//First, remove the machine from the node's hosts in the distribution table
		Machine[] supported = ddt.getMachineArray(ddt.getNode(Integer.toString(mesg.getOperatorID())));
		Machine[] newsupported = new Machine[supported.length - 1];
		int curr = 0;
		Machine currentMachine = null;
		for (int i = 0; i < supported.length; i++)
		{
			if (supported[i].getMachineName().equals(mesg.getMachineName()))
			{
				currentMachine = supported[i];
			}
			else
			{
				newsupported[curr++] = supported[i];
			}
		}
		ddt.add(ddt.getNode(Integer.toString(mesg.getOperatorID())), newsupported);

		//Second, remove its inputs.
		Object[] children = ddt.getNode(Integer.toString(mesg.getOperatorID())).getChildren();
		for(int i = 0; i < children.length; i++)
		{
			if (children[i] instanceof XATNode)
			{
				ReplicationStyleProtocol c = new RemoveMachineFromOutput(Integer.toString(((XATNode)children[i]).getStreamOperator().getObjectID()), currentMachine);
				Packet p = new Packet("remove input", c);
				Machine[] childhoused = ddt.getMachines((XATNode)children[i]);
				for(int j = 0; j < childhoused.length; j++)
				{
					try {
					Socket stepfour = new Socket(childhoused[j].getIPAddress(), childhoused[j].getConnectionListenerPort());
					ObjectOutputStream fourout = new ObjectOutputStream(stepfour.getOutputStream());
					fourout.writeObject(p);
					fourout.close();
					} catch (IOException ioe) { ioe.printStackTrace(); }
				}
			}
		}

		//Third, deactivate the node on the target machine.
		try {
			Connection c = new DeActivate(Integer.toString(ddt.getNode(Integer.toString(mesg.getOperatorID())).getStreamOperator().getObjectID()));
			Packet p = new Packet("reconnection",c);
			Socket thirdStep = new Socket(currentMachine.getIPAddress(),currentMachine.getConnectionListenerPort());
			ObjectOutputStream deacOut = new ObjectOutputStream(thirdStep.getOutputStream());
			deacOut.writeObject(p);
			deacOut.close();
			thirdStep.close();
		} catch (IOException ioe) { ioe.printStackTrace(); }
		
		//fourth, remove its outputs
		Iterator parents = ddt.getNode(Integer.toString(mesg.getOperatorID())).getParents();
		while (parents.hasNext())
		{
			Object parent = parents.next();
			if (parent instanceof XATNode)
			{
				XATStreamOperator parentOp = ((XATNode)parent).getStreamOperator();
				ReplicationStyleProtocol c = new RemoveMachineFromInput(Integer.toString(parentOp.getObjectID()), currentMachine);
				Packet p = new Packet("remove output", c);
				Machine[] parenthoused = ddt.getMachines((XATNode)parent);
				for(int j = 0; j < parenthoused.length; j++)
				{
					try {
					Socket stepfour = new Socket(parenthoused[j].getIPAddress(), parenthoused[j].getConnectionListenerPort());
					ObjectOutputStream fourout = new ObjectOutputStream(stepfour.getOutputStream());
					fourout.writeObject(p);
					fourout.close();
					} catch (IOException ioe) { ioe.printStackTrace(); }
				}
			}
		}
	}
	
    /**
     * Adds a new operator partition to the set of partitions for an operator in a query plan
     * <P>
     * Runs at distribution manager
     * 
     * @return void
     * @param ddt the system's distribution table
     * @param m the array of all machines
     * @param opNode the node in the query plan tree containing the operator in question
     * @param newM the machine to be added to the set of operator partitions
     * @since 1.2
     */
	public static void activateOperator(DistributionTable ddt, Machine[] m, XATNode opNode, Machine newM)
	{
		//First, add the machine to the node's hosts in the distribution table
		Machine[] supported = ddt.getMachineArray(opNode);
		Machine[] newsupported = new Machine[supported.length + 1];
		int curr = 0;
		//Machine currentMachine = null;
		for (int i = 0; i < supported.length; i++)
		{
			newsupported[i] = supported[i];
		}
		newsupported[newsupported.length - 1] = newM;
		
		ddt.add(opNode, newsupported);

		//Second, activate the node on the target machine.
		try {
			Connection c = new Activate(Integer.toString(opNode.getStreamOperator().getObjectID()));
			Packet p = new Packet("reconnection",c);
			Socket thirdStep = new Socket(newM.getIPAddress(),newM.getConnectionListenerPort());
			ObjectOutputStream acOut = new ObjectOutputStream(thirdStep.getOutputStream());
			acOut.writeObject(p);
			acOut.close();
			thirdStep.close();
		} catch (IOException ioe) { ioe.printStackTrace(); }

		//Third, add its output
		Iterator parents = opNode.getParents();
		while (parents.hasNext())
		{
            //this first part here is for opening up the input for the parent operator.
			Object parent = parents.next();
            Machine[] parenthoused = ddt.getMachines((XATNode)parent);
			if (parent instanceof XATNode && !ddt.getMachine((XATNode)parent).equals(newM))
			{
				XATStreamOperator parentOp = ((XATNode)parent).getStreamOperator();
				ReplicationStyleProtocol c = new AddMachineToInput(Integer.toString(parentOp.getObjectID()), newM);
				Packet p = new Packet("add input", c);
				
				for(int j = 0; j < parenthoused.length; j++)
				{
					try {
					Socket stepfour = new Socket(parenthoused[j].getIPAddress(), parenthoused[j].getConnectionListenerPort());
					ObjectOutputStream fourout = new ObjectOutputStream(stepfour.getOutputStream());
					fourout.writeObject(p);
					fourout.close();
					} catch (IOException ioe) { ioe.printStackTrace(); }
				}
			}
            
            //Then we instruct the partitioned operator to send data out to the parent
            try {
                Socket stepfourparttwo = new Socket(newM.getIPAddress(), newM.getConnectionListenerPort());
                ObjectOutputStream fourtwoout = new ObjectOutputStream(stepfourparttwo.getOutputStream());
                for(int i = 0; i < parenthoused.length; i++)
                {
                    Packet p = new Packet("send to parent", new SendData(opNode.getStreamOperator().getProperty("OBJECT_ID"), 0, parenthoused[i]));
                    fourtwoout.writeObject(p);
                }
                fourtwoout.close();
            } catch (IOException ioe) { ioe.printStackTrace(); }
		}
		
		//Fourth, start up its inputs.
		Object[] children = opNode.getChildren();
		for(int i = 0; i < children.length; i++)
		{
            Machine[] childhoused = ddt.getMachines((XATNode)children[i]);
			if (children[i] instanceof XATNode && !ddt.getMachine((XATNode)children[i]).equals(newM))
			{
			    ReplicationStyleProtocol c = new AddMachineToOutput(Integer.toString(((XATNode)children[i]).getStreamOperator().getObjectID()), newM);
				Packet p = new Packet("add input", c);
				for(int j = 0; j < childhoused.length; j++)
				{
					try {
					Socket stepfour = new Socket(childhoused[j].getIPAddress(), childhoused[j].getConnectionListenerPort());
					ObjectOutputStream fourout = new ObjectOutputStream(stepfour.getOutputStream());
					fourout.writeObject(p);
					fourout.close();
					} catch (IOException ioe) { ioe.printStackTrace(); }
				}
			}
            
            //Then we instruct the partitioned operator to receive data from the child
            try {
                Socket stepfourparttwo = new Socket(newM.getIPAddress(), newM.getConnectionListenerPort());
                ObjectOutputStream fourtwoout = new ObjectOutputStream(stepfourparttwo.getOutputStream());
                Packet p = new Packet ("receive from child", new ReceiveData(((XATNode)children[i]).getStreamOperator().getProperty("OBJECT_ID"), 
                                                    opNode.getStreamOperator().getProperty("OBJECT_ID"), i));
                 fourtwoout.writeObject(p);

                fourtwoout.close();
            } catch (IOException ioe) { ioe.printStackTrace(); }
            
		}
	}
	
    /**
     * Calculate new saturation values, request any necessary assistance, and return the new values.
     * <P>
     * Runs at query processor
     * 
     * @return the newly calculated saturation value [0-1]
     * @param exec the system's execution controller
     * @param currentScheduler the system's scheduler for operators
     * @param operator the-last run operator
     * @param systemOverhead the amount of memory used before execution started
     * @param lastMemoryUsage the saturation value previously reported by this function
     * @since 1.2
     */
	public static double doSpreadfirePostExecution(DistributedExecutionController exec,
												 XATScheduler currentScheduler,
												 XATStreamOperator operator,
												 long systemOverhead, 
												 double lastMemoryUsage)
	{
	    System.gc();
        
	    if(HIGH_THRESHOLD <= 0.0 || LOW_THRESHOLD <= 0.0)
	        setThresholds(exec.getConfig());
	    
	    if(System.getProperty("MAX_TOTAL_STATE_SIZE") == null)
	        setMaxStateSize();
	    
		/**double currentMemoryUsage = (double)(Runtime.getRuntime().totalMemory() 
											- Runtime.getRuntime().freeMemory() 
											- systemOverhead) 
				 					/ (double)(Runtime.getRuntime().totalMemory() 
				 							- systemOverhead);
				 							*/
	    //OperatorState[] operatorstates = operator.getState();
	   // int maxstatesize = 0;
	    //for(int i = 0; i < operatorstates.length; i++)
	    //{
	    //    if(operatorstates[i].getStateSize() > maxstatesize)
	   //         maxstatesize = operatorstates[i].getStateSize();
	   // }
        //double currentMemoryUsage = (double)maxstatesize 
		//	/ Double.parseDouble(operator.getProperty("MAX_STATE_SIZE"));
	    double currentMemoryUsage = Double.parseDouble(
	            					     StatisticsGatherer.getStatisticsGather()
	            					                       .getValue("MEMORY_SATURATION",
	            					                               currentScheduler.getTree()
	            					));
  
	    
	     //	/ Double.parseDouble(System.getProperty("MAX_TOTAL_STATE_SIZE"));
	    
	     
	    //Zeroth thing to do is check for emergency conditions, i.e. 99% memory used.
		if (currentMemoryUsage >= 0.99)
		{
			try {
			Socket newSStat = new Socket (exec.sStat.getInetAddress(), exec.sStat.getPort()); 
			ObjectOutputStream oos = new ObjectOutputStream(newSStat.getOutputStream());
			oos.writeObject(new SpreadfireEmergencyShutdown());
			oos.flush();
			oos.close();
			
			} catch (IOException e) {}
			
			//May as well not wait for the DM to tell us to stop.  
			// It can happen multiple times without negative effect.
			// 2005.11.1 or maybe this isn't a good idea.
			//  Something is causing a StreamCorruptedException
			// No, no, what happened was that StreamCorruptedExceptions occur when you make 
			//  multiple ObjectOutputStreams on the same socket.  That's been fixed.  Just
			//  make a new socket with the same details and you're good.

			//exec.terminateExecutionNow();
			//DEBUG LINE BELOW:

			
			return 1.0;
		}
		
		if (currentMemoryUsage >= SpreadfireManager.HIGH_THRESHOLD && currentMemoryUsage >= lastMemoryUsage + 0.01)
  	  		SpreadfireManager.chooseOpForAssistance(exec, currentScheduler);
  	  	lastMemoryUsage = currentMemoryUsage;

  	  	
  	  	//Second thing we'll do here is check for an empty state.
  	  	// That is, we'll see if the last used op has zero tuples in all of its states.
  	  	// We can deactivate the operator if that's the case.
  	  	OperatorState[] opStates = operator.getState();
  	  	boolean allstatesempty = true;
  	  	for(int i = 0; i < opStates.length; i++)
  	  	{
  	  		if(opStates[i].getStateSize() > 0)
  	  			allstatesempty = false;
  	  	}
  	  	if (allstatesempty)
  	  	{
  	  		// Now we have to generate a packet to send to the StatisticsReceiver.
  			try {
  				ObjectOutputStream conduit = new ObjectOutputStream(exec.getSocket().getOutputStream());
  				conduit.writeObject(new DeactivateEmptyStateOperator(operator.getObjectID(),
  																		System.getProperty("NAME")));
  			} catch(IOException ioe) {
  				
  			}
  	  	}
  	  	System.out.println ("Current saturation is " + new DecimalFormat("#.#####").format(lastMemoryUsage) 
  	  	                                             + " with memory usage at " 
                                                     + new DecimalFormat("#.#####").format(Double.parseDouble(StatisticsGatherer.getStatisticsGather()
                                                         .getValue("USED_MEMORY_PERCENTAGE",
                                                                   StatisticsGatherer.getStatisticsGather()
                                                                                      .getQueryObject("USED_MEMORY_PERCENTAGE", 
                                                                                                      StatisticsReference.LAST)))));
        //.out.println(" (" + (((double)Runtime.getRuntime().totalMemory()-(double)Runtime.getRuntime().freeMemory())/(double)Runtime.getRuntime().totalMemory()) + " calculated)");
  	  	return lastMemoryUsage;
	}
	
    /**
     * Send a request to the distribution manager to assign the storage token for one of the operator's states
     * to another machine
     * <P>
     * Runs at query processor
     * @return void
     * @param dec the system's execution controller
     * @param op the operator in need of assistance
     * @param stateNum the position of the largest active state in the operator (counting from zero)
     * @since 1.2
     */
	public static void requestReplicationAssistance(DistributedExecutionController dec, 
													ReplicatedStreamOperator op, 
													int statenum)
	{
		try {
		    Socket tempsocket = new Socket(dec.getSocket().getInetAddress(), dec.getSocket().getPort());
			ObjectOutputStream conduit = new ObjectOutputStream(tempsocket.getOutputStream());
			conduit.writeObject(new ReplicatedOperatorAssistance(op, dec.getClientName(), statenum));
			conduit.flush();conduit.close();
		} catch(IOException ioe) {
			
		}
	}
	
    /**
     * Find out which operator on this machine is contributing the most to memory overload, then call requestReplicationAssistance
     * <P>
     * Runs at query processor
     * @return void
     * @param dec the system's execution controller
     * @param currentScheduler the system's operator scheduler
     * @since 1.2
     */
	public static void chooseOpForAssistance(DistributedExecutionController dec, 
			                                 XATScheduler currentScheduler)
	{
		XATNode[] allops = currentScheduler.getAllXATNodes();
		int nodes = allops.length;
		XATStreamOperator largeststateop = null;
		long largeststatesize = 0;
        int statewithlargest = -1;  
		for (int i = 0; i < nodes; i++)
		{
			XATStreamOperator currentop = allops[i].getStreamOperator(); 
			if(currentop.isActive() && currentop.hasState() && currentop instanceof ReplicatedStreamOperator)
			{
				//calculate the total number of fields in states (schema size * tuples)
				long fieldcount = 0;
				OperatorState[] currentstates = currentop.getState();
			
				if (currentstates.length == currentop.getInputQueueCount())
				{
					for (int j = 0; j < currentstates.length; j++)
					{
						try {
						if (((ReplicatedStreamOperator)currentop).hasTokenForState(j))
                        {    
                            int schemalength = currentop.getInputQueue(j).getSchema().getLength();
							fieldcount = schemalength * currentstates[j].getStateSize();

                            if (fieldcount > largeststatesize)
                            {
                                largeststatesize = fieldcount;
                                largeststateop = currentop;
                                statewithlargest = j;
                            }
                        }
                        } catch (ArrayIndexOutOfBoundsException oobe) {
							break;
						}
					}
				}
				else
				{
					//We can't make an assumption about how the input queues map to the states
					// so we'll estimate by taking the average schema size and multiplying that by the tuple count.
					double averageschemasize = 0.0;
					int queuecount = currentop.getInputQueueCount();
					for (int j = 0; j < queuecount; j++)
						averageschemasize += currentop.getInputQueue(j).getSchema().getLength();

					averageschemasize /= (double)queuecount;
					
					for (int j = 0; j < currentstates.length; j++)
					{
                        if(((ReplicatedStreamOperator)currentop).hasTokenForState(j))
                        {
                            fieldcount = (long)(averageschemasize * (double)currentstates[j].getStateSize());
                            if (fieldcount > largeststatesize)
                            {
                                largeststatesize = fieldcount;
                                largeststateop = currentop;
                                statewithlargest = j;
                            }
                        }
                    }
				}
				

			}
		}
		
		//TODO: getting NPE here which indicates that it's not finding an op.
		//  I'm thinking that maybe it's asking for assistance with no tuples.
		//  That, of course, doesn't make any sense.
//		OperatorState[] whichtomove = largeststateop.getState();
//		int largestlocalstatesize = 0;
//
//		for (int j = 0; j < whichtomove.length; j++)
//		{
//			if (whichtomove[j].getStateSize() > largestlocalstatesize
//					&& ((ReplicatedStreamOperator)largeststateop).hasTokenForState(j))
//			{
//				statewithlargest = j;
//				largestlocalstatesize = whichtomove[j].getStateSize();
//			}
//		}
        if (largeststateop != null)
            SpreadfireManager.requestReplicationAssistance(dec, (ReplicatedStreamOperator)largeststateop, statewithlargest);
	}
	
	
	//This function is not used.  It would only be used if there were a per-operator
	//  allocation of max state.
    /**
     * When measuring memory saturation per-state instead of overall, this function chooses the state in most need of assistance.
     * <P>
     * Runs at query processor
     * @return void
     * @param dec the system's execution controller
     * @param currentScheduler the system's operator cheduler
     * @param op the operator from which to pluck a growing state
     * @since 1.6
     * @deprecated not used under current memory model
     */
	public static void chooseStateForAssistance(DistributedExecutionController dec, 
            XATScheduler currentScheduler, XATStreamOperator op)
	{
	    XATStreamOperator largeststateop = null;
	    long largeststatesize = -1;
	    int statewithlargest = -1;
        OperatorState[] currentstates = op.getState();
        if(op.isActive() && op.hasState() && op instanceof ReplicatedStreamOperator)
        {
            //calculate the total number of fields in states (schema size * tuples)

            for (int j = 0; j < currentstates.length; j++)
            {
                try {
                    if (((ReplicatedStreamOperator)op).hasTokenForState(j)
                         && currentstates[j].getStateSize() > largeststatesize)
                        {
                            largeststatesize = currentstates[j].getStateSize();
                            statewithlargest = j;
                        }                
                    } catch (ArrayIndexOutOfBoundsException oobe) {
                        break;
                    }
            }
        }
        else 
        {
            //Error condition.  We send a nonreplicated operator in
            System.err.println("Error in chooseStateForAssistance:  nonReplicated operator used");
            return;
        }

        SpreadfireManager.requestReplicationAssistance(dec, (ReplicatedStreamOperator)op, statewithlargest);
	}
	
}
