/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.protocol;

import java.io.ObjectOutputStream;
import java.net.Socket;
//import java.net.Packet;
import java.util.TimerTask;
import java.util.ArrayList;

import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.pattern.DistributionTable;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.protocol.PassStorageToken;

/**
 * @author Administrator
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class TimedStorageTokenPassing extends TimerTask {

	private int _lastMachine;
	private int _lastOperator;
	
	private ArrayList _allSplitOps;
	private ArrayList machinestonotify;
	private DistributionTable _dt;
	
	public TimedStorageTokenPassing()
	{
		super();
	}
	
	public TimedStorageTokenPassing(DistributionTable dt, ArrayList splitOps)
	{
		_allSplitOps = splitOps;
		_dt = dt;

		_lastMachine = 1;
		_lastOperator = 0;
	}
	
	public void run()
	{
		//Execution steops outlined.
		// 1. enumerate all of the split operators.
		// 2a. For each split operator find its operator ID  
		// 2b.   and the proper machine it's on.
		// 2b.   and the order the split operator resides on among it's siblings
		// 3. Pick the next machine to host the token.
		// 4. Package this material and send.
	
		int splitOperatorID;
		
		_lastOperator++;
		if (_lastOperator >= _allSplitOps.size()) {
			_lastOperator = 0;
			_lastMachine++;
			if (_lastMachine >= _dt.getMachineArray(((XATNode)_allSplitOps.get(_lastOperator)).getParent()).length) {
				_lastMachine = 0;
			}
		}
		
		XATNode currentOp = (XATNode)_allSplitOps.get(_lastOperator);
		int currentID = Integer.parseInt(currentOp.getOperator().getProperty("OBJECT_ID")); 
		Machine machineToContact = _dt.getMachine(currentOp);
		String targetMachineID = _dt.getMachineArray(currentOp.getParent())[_lastMachine].getProperty("NAME");
		XATNode parentOp = currentOp.getParent();
		Object[] sibsArray = parentOp.getChildren();
		int orderOfCurrent = -1;
		for (int i = sibsArray.length - 1; i >= 0; i--)
		{
			if (sibsArray[i].equals(currentOp)) orderOfCurrent = i;
		}
			
		Packet p = new Packet("passtoken", new PassStorageToken(currentID, orderOfCurrent, targetMachineID));
		
		//Socket s = new Socket(m.getIPAddress(),m.getConnectionListenerPort());
		Socket s = null;
		try {
			int trialcount = 0;
			do {
				try {
					s = new Socket(machineToContact.getIPAddress(),machineToContact.getConnectionListenerPort());
				}
				catch(Exception ex) {
					Thread.sleep(100);
					trialcount++;
				}
			} while((s == null || s.isClosed()) && trialcount < 6);
		 
			ObjectOutputStream out = new ObjectOutputStream(s.getOutputStream());
			out.writeObject(p);
			out.close();
			s.close();
		} catch (Exception ex) {}
	}
	
}
