/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATPreference;
/**
 * Responsible for parsing the xml configuration file that will be used
 * by the ExecutionController
 */
public class ConfigFileParser {
	private org.w3c.dom.Document document;
/**
 * ConfigFileParser constructor comment.
 */
public ConfigFileParser(String fileName) throws Exception{
    super();
    try {
        String key = "javax.xml.parsers.DocumentBuilderFactory";
        String value = "org.apache.xerces.jaxp.DocumentBuilderFactoryImpl";
        //set the DocumentBuilderFactory to use
        java.lang.System.setProperty(key, value);
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        //DocumentBuilderFactory factory = new DocumentBuilderFactoryImpl();
        DocumentBuilder builder = factory.newDocumentBuilder();
        document = builder.parse(fileName);
    } catch (Exception e) {
		throw e;
    }

}
/**
 * Returns an Iterator of the name / value pairs of properties
 * for a given element name 
 * @return Iterator
 * @param elementName java.lang.String
 * @param attributeName java.lang.String[]
 */
private Iterator getAttributes(String elementName) throws IOException {
    NodeList nodes = (NodeList) document.getElementsByTagName(elementName);
    Node node = nodes.item(0);
    ArrayList list = new ArrayList();
    if (node != null) {
        NodeList propertiesList = node.getChildNodes();
        //System.out.println(statsPropertiesList.getLength());

        Node nextNode = node.getFirstChild();
        /** traverses sibling to sibling and add properties found to the vector*/
        while (nextNode != null) {
            if (nextNode instanceof Element) {
                /**
                * check the nodeName, should = property
                * if it doesnt, then do nothing
                */

                if (nextNode.getNodeName().equals("property")) {
                    /** now we've found the properties element*/
                    NamedNodeMap attrs = nextNode.getAttributes();
                    String name = ((Node) attrs.getNamedItem("name")).getNodeValue();
                    String value = ((Node) attrs.getNamedItem("value")).getNodeValue();
                    list.add(new NameValuePair(name, value));
                }
            }
            nextNode = nextNode.getNextSibling();
        }
    }

    return list.iterator();
}
/**
 * Get the properties related to the columns to printout
 * @return Iterator
 */
public Iterator getColumnNamesForOutput() throws IOException {
	return getAttributes("outputColumnNames");
}
/**
 * Return all statistics to gather every time after execution relating to an operator
 * @return Iterator
 */
public Iterator getEveryTimeOperatorStatistics() throws IOException {
	return getAttributes("everyTimeOperator");
}
/**
 * Return all of the properties of the ExecutionAdapter
 * @return Iterator
 */
public Iterator getExecutionAdapterProperties() throws IOException {
	return getAttributes("ExecutionAdapter");
}
/**
 * Return all of the properties of the ExecutionController
 * @return Iterator
 */
public Iterator getExecutionControllerProperties() throws IOException {
	return getAttributes("ExecutionController");
}
/**
 * Return all properties relating to the execution
 * @return Iterator
 */
public Iterator getExperimentProperties() throws IOException {
	return getAttributes("experiment");
}
/**
 * Get the properties related to the output format
 * @return Iterator
 */
public Iterator getOutputFormatProperties() throws IOException {
	return getAttributes("outputFormat");
}
/**
 * Return all statistics to gather Periodically after execution relating to an operator
 * @return Iterator
 */
public Iterator getPeriodicOperatorStatistics() throws IOException {
	return getAttributes("periodicOperator");
}
/**
 * Return all statistics to gather Periodically after execution relating to the XATTree
 * @return Iterator
 */
public Iterator getPeriodicTreeStatistics() throws IOException {
	return getAttributes("periodicTree");
}
/**
 * Get the user's preferences.  Unfortunately, the preference
 * requires 3 attributes, therefore the rest of the parsing code
 * cannot be reused.  
 * @return Iterator
 */
public Iterator getPreferences() throws Exception {
	String elementName = "preferences";
	NodeList nodes = (NodeList) document.getElementsByTagName(elementName);
    Node node = nodes.item(0);
    /** iterate through the properties of the statisticsToGather element*/
    NodeList preferenceList = node.getChildNodes();
    //System.out.println(statsPropertiesList.getLength());
    ArrayList list = new ArrayList();
    Node nextNode = node.getFirstChild();
    double totalWeight = 0;
    /** traverses sibling to sibling and add properties found to the vector*/
    while (nextNode != null) {
        if (nextNode instanceof Element) {
            /**
            * check the nodeName, should = property
            * if it doesnt, then do nothing
            */

            if (nextNode.getNodeName().equals("preference")) {                
            	/** now we've found the properties element*/
            	NamedNodeMap attrs = nextNode.getAttributes();
            	String statistic = ((Node) attrs.getNamedItem("statistic")).getNodeValue();
            	String quantifier = ((Node) attrs.getNamedItem("quantifier")).getNodeValue();
            	double weight = Double.parseDouble(((Node) attrs.getNamedItem("weight")).getNodeValue());
            	XATPreference p = new XATPreference();
            	p.setStatistic(statistic);
            	if(quantifier.equalsIgnoreCase("min")){
					p.setQuantifier(XATPreference.MIN);
            	} else if(quantifier.equalsIgnoreCase("max")){
	            	p.setQuantifier(XATPreference.MAX);
            	} else{
					throw new IllegalArgumentException("The quanitifier must be min or max");
            	}            	
            	p.setWeight(weight);
            	totalWeight += weight;
            	list.add(p);
            }            
        }
        nextNode = nextNode.getNextSibling();
    }
    if(totalWeight != 1.0){
		throw new IllegalArgumentException("The sum of the weights must be exactly 1");
    }
	return list.iterator();
}
/**
 * Return all properties relating to available machines in the system.
 * @return Iterator
 */
public Iterator getMachineProperties() throws IOException {
	return getAttributes("machine");
}
/**
 * Return all properties relating to the query plan generator
 * @return Iterator
 */
public Iterator getQueryPlanGeneratorProperties() throws IOException {
	return getAttributes("queryPlanGenerator");
}
/**
 * Return all of the scheduling algorihms
 * @return Iterator
 */
public Iterator getSchedulingAlgorithms() throws IOException {
	return getAttributes("algorithms");
}
/**
 * Return all properties relating to the scheduling algorihms
 * @return Iterator
 */
public Iterator getSchedulingProperties() throws IOException {
	return getAttributes("scheduling");
}
/**
 * Returns all of the system properties
 * @return Iterator
 */
public Iterator getSystemProperties() throws IOException {
	return getAttributes("system");
}
/**
 * Returns all of the tree properties
 * @return Iterator
 */
public Iterator getTreeProperties() throws IOException {
	return getAttributes("treeProperties");
}
}
