/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.NotSerializableException;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.net.Socket;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import edu.wpi.cs.dsrg.STeM.STeMJoinProbeOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.Helper;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.DataSampleStorage;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.ExecutionStatisticsSampleStorage;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.STQM;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.StreamOldAndNewSamples;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.TrainingTupleCollectorOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.STQM_Analyzer.STQM_TUNING_RECOMMENDATIONS;

//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
//import edu.wpi.cs.dsrg.xmldb.xat.common.migration.probablisticjoinoperator.ChangingWBJoinOperatorImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.migration.realjoinoperator.WTransitiveHashJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.realjoinoperator.WTransitiveHashJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.spreadfire.ReplicatedStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WEquiJoinStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamUnionOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.DistributionConfiguration;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ConnectionListener;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ConnectionTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.MigrationKeeper;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.SubPlan;
//import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendOperatorsToFillStates;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendStateToMachine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.PacketPool;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleDistributor;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleFeeder;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleReceiver;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleSender;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.manager.SpreadfireManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.FireStreamGUI;
//import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUIDistributionManager;
//import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUIDistributionManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUIQueryEngine;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATExecutionAdapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerException;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsPrinter;
import edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.heuristics.Optimizer;
//import edu.wpi.cs.dsrg.xmldb.xat.component.queuemanager.XATQueueManager;
//import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.RoundRobinScheduler;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation.Adapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation.LocalMemoryAdapter;
//import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.protocol.ReplicatedOperatorAssistance;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.adprotocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamUnionOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.UpdatePartitionStatistics;
/**
 * This controls everything about execution including asking the scheduler
 * which operator to run next and for how long, running the operator, updating
 * the statistics for the operator, replanning, and changing the scheduling algorithm.  
 *  
 * @author: Timothy M. Sutherland
 *
 * Modification History (put your comments here, especially if they are major ones)
 * 
 * Rimma V. Nehme (rnehme@cs.purdue.edu)
 * Date: 02/10/2009
 * Comment: Made some code pretty-print
 *          Added STQM pieces that were previously in a separate code base  
 */

public class DistributedExecutionController extends ExecutionController implements Runnable 
{

	//---------------------------------------------------------------------------------------------------
	
    private static DistributedExecutionController _this;
    protected ArrayList tree; // an arraylist of XATTree's
	protected StatisticsPrinter printer;
	protected TupleReceiver packetReceiver;
	protected ConnectionListener connectionListener;
	
	protected XATScheduler currentScheduler;
	protected Machine machine;
	/** array of statistics to periodically gatherer about operators */
	protected String[] periodicOperatorStatistics;
	/** array of statistics to periodically gatherer about trees */
	protected String[] periodicTreeStatistics;
	/** array of statistics to gatherer about operators everytime the operator is run*/
	protected String[] everyTimeOperatorStatistics;
	
	protected String[] everyTimeSystemStatistics;
	
	public boolean inLocalMigration = false;
	public boolean inGlobalAdaptation = false;
	long startTime;  //indicate the time the system starts running.
	
	//KEW for testing only
	boolean bTesting = false;
	
	/** 
	 * this defines the last time that certain properties about this operator were updated.
	 * In the ExecController, we do not want to update every property everytime an operator
	 * has run, but would prefer to do it periodically.  This property will keep track of the
	 * last time those properties were updated. The key is the operator, the value is the lastTime
	 * it was updated. 
	*/
	protected java.util.Hashtable lastTimeOperatorUpdated;
	/** 
	 * this defines the last time that the status of execution was checked.  That is,
	 * when did we last determine how well execution is performing.  This value is
	 * tied to the STATUS_CHECK_FREQUENCY property.  
	*/
	protected long lastTimeStatusChecked;
	/** 
	 * this defines the last time that the tree statistics were updated.  
	*/
	protected long lastTimeTreesUpdated;
	/** 
	 * this defines the last time that tuples were uploaded to the application.  
	*/
	protected boolean debug;
	/** this adapter manages all of the scheduling algorithms and decides which one to run*/
	protected XATExecutionAdapter adapter;
	protected java.util.Hashtable schedulerCount;
	protected boolean debug_scheduling_notice;
	protected DistributionConfiguration config;
	protected String clientName;
	protected PacketPool inPool;
	protected PacketPool outPool;
	protected ConnectionTable inConnections;
	protected ConnectionTable outConnections;
	protected TupleFeeder tupleFeeder;
	protected TupleDistributor tupleDistributor;
	protected TupleSender tupleSender;
	protected static CAPE_GUI gui;
	protected XATDag[] treeArray;
	protected SystemObject system;
	protected ArrayList lightStats;
	protected Optimizer optimizer;
	protected String[] distInfo = new String[4];
	//whether to continue running the execution controller
    protected boolean keepGoing = true;
    protected boolean doneFinal = false;
	// local partition adapter
	protected Adapter localAdapter = null;
	// Make these two global to allow access from other methods through statistics gathering connection
    public Socket sStat = null; 
	public ObjectOutputStream outStat = null;
	// to synchronize the disk-based adaptation and distributed-adaptation
	boolean isComputingPartitionsToMove = false;
	//Added by Bin @ Apr 22, 2005. For recording the inter query processor communication due to updating statisitics
	Hashtable interConnectionTable = null;
    long systemOverhead = -1;
	//added by Yali Zhu @ Aug. 22, 2005. Used for distributed migration moving state among machines.
	public boolean stateReceived = false;
	public LightWeightStatistic lws;
	public MigrationKeeper migrationkeeper = new MigrationKeeper();
	public boolean treeLocked; // can be locked by migration process
	
	
	//---------------------------------------------------------------------------------------------------
	// STQM
	//---------------------------------------------------------------------------------------------------    		
	public static Hashtable hashSTQMDataSamplesGlobalStorage = new Hashtable(); //global storage for all data samples
	public static Hashtable hashSTQMExecutionStatisticsSamplesGlobalStorage = new Hashtable(); //global storage for all operators execution statistics
	public static Hashtable hashSTQMRouteSignaturesGlobalStorage = new Hashtable(); //global storage for all possible route signatures
	
	public static OnlineClassifierOperatorImp OnlineClassifierRef = null;	
	
	String qmOptimizerInputFileName = "resources\\SelfTuningQueryMesh\\Dataset5\\optimizer_input.xml";
	public STQM stqmManager;
	public static long stqmCurrentTime;
	public static long stqmLastTime;
	public static boolean bIsSTQMExecuting = false;
	public static boolean bIsSTQMDoPhysicalMigration = false;
	public static STQM_TUNING_RECOMMENDATIONS currentPhysicalRecommendations = STQM_TUNING_RECOMMENDATIONS.NONE;
	
	
	//int stqmFrequency = 30000; //tune every 30 seconds
	int stqmFrequency = 60000; //tune every 1 minute
	//int stqmFrequency = 180000; //tune every 3 minutess
	//int stqmFrequency = 300000; //tune every 5 minutes
	//int stqmFrequency = 600000; //tune every 10 minutes	
	boolean DO_STQM = false;
	
	//---------------------------------------------------------------------------------------------------
		
	public DistributedExecutionController(DistributionConfiguration dc, String clientName) throws Exception 
	{	
		super();
		    
		//make controller statically available
	    _this = this;
	    
	    config = dc;			
		this.clientName = clientName;
		System.out.println("My Name is: " + clientName);
		System.setProperty("NAME", clientName);
		tree = new ArrayList();
		system = new SystemObject();
		System.out.println("before calling init");
		initExecution();	
		System.out.println("after calling init");
		this.inPool = new PacketPool();
		this.outPool = new PacketPool();
		this.inConnections = new ConnectionTable();
		this.outConnections = new ConnectionTable();
		
		packetReceiver = new TupleReceiver(machine.getTupleReceiverPort(),inPool,tree);
		connectionListener = new ConnectionListener(machine.getConnectionListenerPort(),clientName, inConnections, outConnections, this);
		tupleFeeder = new TupleFeeder(inPool,inConnections);
		tupleDistributor = new TupleDistributor(outConnections,machine);
		tupleSender = new TupleSender(outPool,outConnections);
		
		interConnectionTable = new Hashtable();
		
		//STQM manager that implements QM self-tuning (disabled by default)
		if (DO_STQM)
		{
			stqmManager = new STQM((SelfTuningQueryMeshExecutionController)this);	
			stqmLastTime = System.currentTimeMillis();
		}
	}

	//---------------------------------------------------------------------------------------------------

	public String getClientName()
	{
		return this.clientName;
	}
	
	//---------------------------------------------------------------------------------------------------

	public ConnectionTable getInConnectionTable() 
	{    
		return this.inConnections;
	}
	
	//---------------------------------------------------------------------------------------------------
	
	public ConnectionTable getOutConnectionTable() 
	{    
		return this.outConnections;
	}

	//---------------------------------------------------------------------------------------------------
	
	/**
     * If the time since the status was checked is greater than STATUS_CHECK_FREQUENCY,
     * then this will check the status of execution, otherwise it will just return.  
     * 
     */
	
	protected void checkSchedulerStatus() 
	{	
		long frequency = Long.parseLong(this.getProperty("STATUS_CHECK_FREQUENCY"));
		//if we are not forcing a change and if it isnt time to do so, we won't check the exec status
	
		if((System.currentTimeMillis() - lastTimeStatusChecked) < frequency)
		{		
			/** not time to check the status of execution */		
			return;			
		} 
		else
		{				
			/** 
			 * ok, time to check the status of execution 
			 * first we will update the tree and operator statistics b/c these contain a lot of
			 * the metrics that will evaluate the execution status
			 */
				
			XATTree[] treeArray = new XATTree[tree.size()];
		
			for(int i = 0; i < tree.size();i++)
			{
				treeArray[i] = (XATTree) tree.get(i);
			}
				
			for(int count=0; count < tree.size(); count++)
			{			
				XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray(); 
				
				for (int j = 0; j < nodes.length; j++) 
				{     		   
					for (int i = 0; i < periodicOperatorStatistics.length; i++) 
					{      	      
						stats.updateValue(periodicOperatorStatistics[i], nodes[j].getOperator());       	   	
					}
     	   
					lastTimeOperatorUpdated.put(nodes[j].getOperator(), new Long(System.currentTimeMillis()));    		
				}			
			}									
		
			for (int i = 0; i < periodicTreeStatistics.length; i++)
			{			
				stats.updateValue(periodicTreeStatistics[i], treeArray);		
			}
		
			lastTimeTreesUpdated = System.currentTimeMillis();
		
			if(debug)
			{			
				System.out.println("*****Time to check Execution Status*****");		
			}

			currentScheduler = adapter.getNextScheduler();
			stats.setCurrentCategory(adapter.getCategory());

			if(debug)
			{			
				System.out.println("***** Next Scheduler is: " + adapter.getCategory().toString() + " *****");		
			}
		
			printer.setCategory(stats.getCurrentCategory());		
			lastTimeStatusChecked = System.currentTimeMillis();		
			updateSchedulerCount();	
		}	
	}
	
	//---------------------------------------------------------------------------------------------------
	
	/**
	 * Returns the current scheduling algorithm
	 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler
	 */
	
	public edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler getCurrentScheduler() 
	{	
		return currentScheduler;
	}
	
	//---------------------------------------------------------------------------------------------------
	
	/**
	 * Returns the handle to the statistics printer
	 * @return edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsPrinter
	 */
	
	public StatisticsPrinter getPrinter() 
	{		
		return printer;
	}
	
	//---------------------------------------------------------------------------------------------------

	public void setLightweightStats(ArrayList stats)
	{
		this.lightStats = stats;
	}

	//---------------------------------------------------------------------------------------------------
	
	/**
	 * Returns the handle to the packet receiver
	 * @return edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsPrinter
	 */

	public TupleReceiver getTupleReceiver() 
	{		
		return packetReceiver;
	}
	
	//---------------------------------------------------------------------------------------------------
	
	/**
	 * Returns the handle to the packet receiver
	 * @return edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsPrinter
	 */

	public ConnectionListener getConnectionListener() 
	{		
		return connectionListener;
	}
	
	//---------------------------------------------------------------------------------------------------

	public TupleFeeder getTupleFeeder()
	{	
		return tupleFeeder;
	}
	
	//---------------------------------------------------------------------------------------------------

	public TupleDistributor getTupleDistributor()
	{	
		return tupleDistributor;
	}
	
	//---------------------------------------------------------------------------------------------------

	public TupleSender getTupleSender()
	{
		return tupleSender;
	}
	
	//---------------------------------------------------------------------------------------------------

	public CAPE_GUI getGUI()
	{	
		if (gui == null) 
		{
			//gui = new CAPE_GUIDistributionManager(treeArray, this.clientName, this.system);
			//System.out.println("This is the message before the GUI constructor is called.");
			gui = new CAPE_GUIQueryEngine(treeArray, this.clientName, this.system, this.distInfo);
		}
		return gui;
	}
	
	//---------------------------------------------------------------------------------------------------

	/**
	 * Insert the method's description here.
	 * Creation date: (4/6/2003 8:46:07 PM)
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
	 */
	
	public edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree[] getTree() 
	{
		return (XATTree[])tree.toArray(new XATTree[]{});
	}
	
	//---------------------------------------------------------------------------------------------------

	/**
	 * Accesses the ExecutionController's socket to the distribution manager.
	 * Creation date: (9/21/2005 8:46:07 PM)
	 * @return java.net.Socket
	 */
	
	public Socket getSocket() 
	{
		return sStat;
	}
	
	//---------------------------------------------------------------------------------------------------

	/**
	 * Initializes various parts of the execution including
	 * query plan, input, schedulers, and statistics table
	 */
	
	protected void initExecution() throws Exception 
	{	
		/* init query plan optimizer */
		optimizer = new Optimizer();
	
		/** next set up the ExecutionController properties */
		Enumeration e = config.getProperties();
		System.out.println("config file is: " + config.toString());
	
		while(e.hasMoreElements())
		{
			String name = (String) e.nextElement();
			String value = config.getProperty(name);
			if(this.isValidStatistic(name))
			{
				this.setProperty(name, value);
			}
		}

		//set up distribution information
		//System.out.println("Outside while loop");
		distInfo = new String[4];
		distInfo[0] = "Initial distribution algorithm: " + config.getProperty("DISTRIBUTION_PATTERN").substring(config.getProperty("DISTRIBUTION_PATTERN").lastIndexOf(".") + 1);
		distInfo[1] = "Redistribution algorithm: " + config.getProperty("REDISTRIBUTION_POLICY").substring(config.getProperty("REDISTRIBUTION_POLICY").lastIndexOf(".") + 1);
		distInfo[2] = "Redistribution interval: " + (Integer.parseInt(config.getProperty("REDISBRIBUTION_TIME")) / 1000);
		distInfo[3] = "Redistribution percentage: " + config.getProperty("REDISTRIBUTION_PERCENT") + "%";
	
		// Get the machine properties
		machine = config.getMachine(clientName);
		Enumeration en = machine.getProperties();
	
		while(en.hasMoreElements())
		{
			String name = (String) en.nextElement();
			String property = machine.getProperty(name);
			if(this.isValidPropertyName(name))
			{
				this.setProperty(name, property);
			}
		}

		// Get the query plans	
		int c = 0;
		int numQueryPlans = 0;
	
		for(Iterator i= config.getQueryPlans(); i.hasNext();)
		{
			QueryPlan qp = (QueryPlan) i.next();
			System.out.println("queryplan to add is: " + qp.getProperty("QUERY_ID"));
			tree.add(qp.getQueryPlan());
			c++;
			numQueryPlans++;	
		}
	
		System.out.println("tree size is: " + tree.size());    
	
		debug = (new Boolean(this.getProperty("EXECUTION_DEBUG"))).booleanValue();
		debug_scheduling_notice =  (new Boolean(this.getProperty("SCHEDULING_NOTIFICATION_DEBUG"))).booleanValue();
		lastTimeOperatorUpdated = new Hashtable();
		lastTimeStatusChecked = System.currentTimeMillis();
		lastTimeTreesUpdated = System.currentTimeMillis();
		schedulerCount = new Hashtable();
	
		stats.setProperty("TIMER", this.getProperty("STATS_TABLE_STATS"));	
	
		XATDag[] statsTree = new XATDag[tree.size()];
	
		for(int i = 0; i < tree.size(); i++)
		{
			statsTree[i] = (XATDag) tree.get(i);
		}

		stats.initStatsTable(statsTree);  
		stats.addNewQueryObject(system);       
         	             		   
		if(debug)
		{
			System.out.println(tree);
		}
	    
		/**
          init the last time some of its properties were updated
          set the last updated time to 0, that way the properties will be updated the first time.    
          This helps with experiments 
        */
        
		for(int count=0; count < tree.size(); count++)
		{		
			XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
    	
			for (int i = 0; i < nodes.length; i++)
			{    		
				if(nodes[i].getOperator() instanceof XATStreamOperator)
				{				
					XATStreamOperator operator = (XATStreamOperator)nodes[i].getOperator();				
					lastTimeOperatorUpdated.put(operator, new Long(0));    		
				}  		
			}    
		}
        
		//set up schedulers
		String className = this.getProperty("ADAPTIVE_HEURISTIC"); 
		treeArray = new XATDag[tree.size()];
	
		for(int i = 0; i < tree.size();i++)
		{
			treeArray[i] = (XATDag) tree.get(i);	
		}

		Constructor[] c1 = Class.forName(className).getConstructors();
		Constructor construct;
    
		for(int i = 0; i < c1.length; i++)
		{		
			construct = c1[i];
			Class[] parameters = construct.getParameterTypes();
		
			System.out.println(parameters[0].getName());
		
			if(parameters.length == 2)
			{			
				if(parameters[0].getName().equals("edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine"))
				{				
					adapter = (XATExecutionAdapter)construct.newInstance(new Object[] { machine, treeArray});		
				}					
			}	
		}
				
		if(adapter == null)
		{		
			throw new Exception("The Adapter could not be instantiated.");	
		}
		
		//start at scheduler 0 
		currentScheduler = adapter.getCurrentScheduler();
		updateSchedulerCount();
    
		//set up some of the statistics to gather
		initStatsToGather();

		//set up the printer
		initStatisticsPrinter();

		//set the starting category in the statistics gatherer to be the first scheduler
		stats.setCurrentCategory(adapter.getCategory());
		printer.setCategory(stats.getCurrentCategory());	
	
		//set up local partition adapter -- Bin @ Dec.10, 2004
		className = this.getProperty("LOCAL_PARTITION_ADAPTER");
	
		if (!className.equals("none")) 
		{		
			c1 = Class.forName(className).getConstructors();
			construct = c1[0]; // might go wrong if multiple constructor exists
			//localAdapter = (Adapter)construct.newInstance(null);		
			localAdapter = (Adapter)construct.newInstance(new Object[]{tree});			
		}
		
		//set the distributed execution controller reference to the symmmwayhashjoinoperator
		//--Bin @ Apr 12, 2005, to handle the tracing of cost information in different machines
		//this part of the code could be safely commented when run other queries/without of tracing cost statisitics
		for(int count=0; count < tree.size(); count++)
		{
			XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
			for (int i = 0; i < nodes.length; i++)
			{
				if(nodes[i].getOperator() instanceof SymmMwayHashJoinOperatorImp)
				{
					((SymmMwayHashJoinOperatorImp)nodes[i].getOperator()).setDistributedExectionControllerRef(this);
				}
			}
		}
	
		System.out.println("Initialization...done.");
	}
	
	//---------------------------------------------------------------------------------------------------
	
	/**
	 * Propagate the cost update information to the partition IDs in different machines
	 * @param machIDs
	 * @param idStats
	 * @param operatorID
	 */
	
	public void propagatePartitionStats(Hashtable machIDs, Hashtable idStats, int operatorID, boolean finalOutput,double inter_rate)
	{	
		Enumeration machineIPs = machIDs.keys();
	
		if (interConnectionTable == null)
		{		
			interConnectionTable = new Hashtable();
		}
			
		while (machineIPs.hasMoreElements())
		{
			String machineIP = (String)machineIPs.nextElement();
			//get the machine and its corresponding linstener port?
			Machine destM = config.getMachineByIP(machineIP);
		
			if (destM !=null)
			{
				Socket ds = null;
				ObjectOutputStream fout = null;			
			
				try 
				{				
					fout = (ObjectOutputStream)interConnectionTable.get(machineIP);				
					if (fout == null) 
					{
						ds = new Socket(destM.getIPAddress(),destM.getConnectionListenerPort());					 
						fout = new ObjectOutputStream(ds.getOutputStream());									
					}
				
					ArrayList idsToUpdate = (ArrayList)machIDs.get(machineIP);
					UpdatePartitionStatistics ust = new UpdatePartitionStatistics(operatorID,idsToUpdate,idStats,finalOutput,inter_rate);		
					Packet pk = new Packet("statistics update",ust);
					fout.reset();
					fout.writeObject(pk);
					fout.flush();				

					//then record back the fout into the interConnectionTable
					interConnectionTable.put(destM.getIPAddress(),fout);
				} 
				catch (Exception e) 
				{				
					System.out.println("From PropagatePartitionStats: " + e.getMessage());				
					e.printStackTrace();
					// return;			
				}
				/*					
				   finally 
				   {
					   try 
					   {
					       if (fout!=null) 
					       {
						       fout.close();
					   }
					   if (ds != null) 
					   {
						   ds.close();
					   }
					   fout = null;
					   ds = null;
				   } 
				   catch (Exception e1)
				   {
					   System.out.println("From PropagatePartitionStats: E1: " + e1.getMessage());
				   }			   
			    */		
			}	
		}
	} //end of propagatePartitionStats

	//---------------------------------------------------------------------------------------------------
	
	/**
	 * In response the ForcePartitionsToPush message, in case the distribution manager finds that the output
	 * rate of machines are not balanced
	 * @param oper
	 * @param rate
	 * @return
	 */
	
	public boolean forcePartitionsToPush(XATStreamOperator oper, double rate)
	{	
		// (1) compute partitions to Move by calling local partition adapter
		ArrayList toPush = null;  
		if (localAdapter != null) 
		{
			//(1.1)First stop the opeator's self disk-based adaptation, it will be resumed after this forceAdaptation
			((SymmMwayHashJoinOperatorImp)oper).stopDiskAdaptation();
			//(1.2) wait until the current disk-based adaptation to finish if there is one on-going
		
			while (((SymmMwayHashJoinOperatorImp)oper).isInDiskAdaptation())
			{
				try 
				{				
					System.out.println("On-going Disk-based adaptation while starts Distributed-adaptation, wait.");
					Thread.sleep(1000);
				} 
				catch (Exception e) 
				{				
					// nothing here	
					e.printStackTrace();
				}		
			}

			//(1.3) compute the partitions to be pushed  
			toPush = localAdapter.computePartitionsToPush(oper,rate);	
		} 
		else 
		{		
			System.out.println("No local partition adapter has been defined.");		
			return false;	
		}
		
		// set inDisk adaptation
		((SymmMwayHashJoinOperatorImp)oper).setInAdaptation(true);
		((SymmMwayHashJoinOperatorImp)oper).pushPartitionsToDisk(toPush);
		((SymmMwayHashJoinOperatorImp)oper).setInAdaptation(false);
		
		((SymmMwayHashJoinOperatorImp)oper).resumeDiskAdaptation();
		return true;
	}
	
	//---------------------------------------------------------------------------------------------------
	
	/**
	 * In response to the ComputePartitionsToMove (Step 1 Protocol) message received in ConnectionListener
	 * @param oper
	 * @param rate
	 * @return
	 */
	
	public boolean computePartitionsToMove(XATStreamOperator oper, double rate)
	{        
		// (1) compute partitions to Move by calling local partition adapter
		ArrayList toMove = null;  
	
		if (localAdapter != null) 
		{	    
			if(localAdapter instanceof LocalMemoryAdapter) 
			{	        
				System.out.println(oper);
				toMove = localAdapter.computePartitionsToMove(oper,rate,((SymmMwayHashJoinOperatorImp)oper).getPartitionIDsPushedToDisk(),false);
			}	    
			else 
			{
				//(1.1)First stop the opeator's disk-based adaptation, it will be resumed in the sendPartitions protocol
				((SymmMwayHashJoinOperatorImp)oper).stopDiskAdaptation();
				//(1.2)wait until the current disk-based adaptation to finish if there is one on-going
				while (((SymmMwayHashJoinOperatorImp)oper).isInDiskAdaptation())
				{			
					try 
					{
						System.out.println("On-going Disk-based adaptation while starts Distributed-adaptation, wait.");
						Thread.sleep(1000);
					} 
					catch (Exception e) 
					{
						// nothing here
						e.printStackTrace();
					}
				}
				// we prefer to choose false partitions to move		
				toMove = localAdapter.computePartitionsToMove(oper,rate,((SymmMwayHashJoinOperatorImp)oper).getPartitionIDsPushedToDisk(),false);	    
			}	    
		} 
		else 
		{
			System.out.println("No local partition adapter has been defined.");		
			return false;
		}

		// (2) sendPartitionsToMove message to Distribution Manager via statistics-gathering connection
		try
		{    	
			if (sStat == null || sStat.isClosed())
			{		    
				sStat = new Socket(config.getManagerIPAddress(),23000);
				outStat = new ObjectOutputStream(sStat.getOutputStream());	
			}		

			//if(outStat == null){
			//	outStat = new ObjectOutputStream(sStat.getOutputStream());			
			//}
	
	      	//System.out.println("Create new outStat");
	    	//System.out.println("Create new sStat");
	    	    
			PartitionsToMove ptm = new PartitionsToMove(toMove);
			System.out.println("Computed partition IDs to move " + toMove);
			outStat.writeObject(ptm);
			outStat.close();
			sStat.close();

			//System.out.println("Close outStat");
			//System.out.println("Close sStat");    
		} 
		catch(Exception e)
		{
			e.printStackTrace();
		}	
		return true;	
	} //end computePartitionsToMove

	//---------------------------------------------------------------------------------------------------

	public boolean computePartitionsToMoveInTree(XATStreamOperator oper, double rate)
	{    
		try 
		{        
			((XATTree)this.tree.get(0)).setProperty("LAST_PARTITION_ADAPTATION_TIME", ""+(System.currentTimeMillis() - startTime));
		}
		catch(Exception e) 
		{
			e.printStackTrace();    
		}
    
		// (1) compute partitions to Move by calling local partition adapter
		ArrayList toMove = null;  
	
		if (localAdapter != null) 
		{
			if (localAdapter instanceof LocalMemoryAdapter) 
			{
				//System.out.println(oper);
				toMove = localAdapter.computePartitionsToMoveInTree((XATTree)this.tree.get(0), rate, false);
			}	    
			else 
			{
				//(1.1)First stop the opeator's disk-based adaptation, it will be resumed in the sendPartitions protocol
				((SymmMwayHashJoinOperatorImp)oper).stopDiskAdaptation();

				//(1.2)wait until the current disk-based adaptation to finish if there is one on-going
				while (((SymmMwayHashJoinOperatorImp)oper).isInDiskAdaptation())
				{
					try 
					{
						System.out.println("On-going Disk-based adaptation while starts Distributed-adaptation, wait.");
						Thread.sleep(1000);
					} 
					catch (Exception e) 
					{
						// nothing here
					}
				}

				// we prefer to choose false partitions to move		
				toMove = localAdapter.computePartitionsToMove(oper,rate,((SymmMwayHashJoinOperatorImp)oper).getPartitionIDsPushedToDisk(),false);
			}	    
		} 
		else 
		{
			System.out.println("No local partition adapter has been defined.");
			return false;
		}
	
		// (2) sendPartitionsToMove message to Distribution Manager via statistics-gathering connection
		try
		{
			if (sStat == null || sStat.isClosed())
			{
				sStat = new Socket(config.getManagerIPAddress(),23000);
				outStat = new ObjectOutputStream(sStat.getOutputStream());	
			}		

			//if(outStat == null){
			//	outStat = new ObjectOutputStream(sStat.getOutputStream());			
			//}
	
			//System.out.println("Create new outStat");
			//System.out.println("Create new sStat");
	    	
			PartitionsToMove ptm = new PartitionsToMove(toMove);
			System.out.println("Computed partition IDs to move " + toMove);
			outStat.writeObject(ptm);
			outStat.close();
			sStat.close();

			//System.out.println("Close outStat");
			//System.out.println("Close sStat");    
		} 
		catch(Exception e)
		{
			e.printStackTrace();
		}
		return true;	
	} //end computePartitionsToMoveInTree

	//---------------------------------------------------------------------------------------------------
	
	/**
	 * In response to the DeactivatedPartitions message (step 3)
	 * @param children
	 * @param partitionsToStop
	 * @return
	 */

	public boolean deactivatePartitions(XATStreamOperator parentOP, Object[] children, ArrayList partitionsToStop, String senderIPAddr)
	{
		// (1) if this is the sender machine, then set the join operator (parent operator) 
		// to start checking the final EOF tuple. If this is the sender, the parentOP must be activated.
		if ((this.machine.getIPAddress()).equals(senderIPAddr))
		{
			// here we assume the parentOP is a SymmMwayHashJoinOperator
			((SymmMwayHashJoinOperatorImp)parentOP).setStartChecking(true);
		}

		// (2) find the activated split operator, remove the partitions from the partition Map, create temporary queue etc.
		for (int i = 0; i < children.length; i++)
		{
			XATStreamOperator curOp = ((XATNode)children[i]).getStreamOperator();
			if ((curOp instanceof StreamSplitOperatorImp) && (curOp.isActive())) 
			{
				((StreamSplitOperatorImp)curOp).deactivatePartitions(partitionsToStop,senderIPAddr);
				System.out.println("Deactivate partitions " + partitionsToStop + " in " + curOp);
			}
		}

		// (3) send Deactivated message to Distribution Manager via statistics-gathering connection
		Deactivated dd = new Deactivated("deactivated", this.machine);
    
		try
		{
			if (sStat == null || sStat.isClosed())
			{
				sStat = new Socket(config.getManagerIPAddress(),23000);
				outStat = new ObjectOutputStream(sStat.getOutputStream());	

				//System.out.println("Create new outStat");
				//System.out.println("Create new sStat");
			}		
			//if(outStat == null){					
			//}
    	
			outStat.writeObject(dd);
			outStat.close();
			sStat.close();
			
			//System.out.println("Close outStat");
			//System.out.println("Close sStat");    
		} 
		catch(Exception e)
		{
			e.printStackTrace();
		}	
		return true;
	} // end of deactivatePartitions

	//---------------------------------------------------------------------------------------------------
	
	public boolean deactivatePartitionsInTree(ArrayList partitionsToStop, String senderIPAddr)
	{
		boolean isSender = this.machine.getIPAddress().equals(senderIPAddr);    
		XATNode[] nodes = ((XATTree)tree.get(0)).convertToArray();
		XATStreamOperator curOp;
    
		for(int i = 0; i < nodes.length; i++) 
		{        
			curOp = nodes[i].getStreamOperator();
			// (1) if this is the sender machine, then set join operators to start checking the EOF from split operator. 
        
			if(isSender && curOp.isActive() &&  curOp instanceof SymmMwayHashJoinOperatorImp)         
				((SymmMwayHashJoinOperatorImp)curOp).setStartChecking(true);
        
			// (2) find the activated split operator, remove the partitions from the partition Map, create temporary queue etc.
			if (curOp instanceof StreamSplitOperatorImp && curOp.isActive()) 
			{			
				((StreamSplitOperatorImp)curOp).deactivatePartitions(partitionsToStop,senderIPAddr);			
				System.out.println("Deactivate partitions " + partitionsToStop + " in " + curOp);        
			}    
		}
		
		// (3) send Deactivated message to Distribution Manager via statistics-gathering connection
		Deactivated dd = new Deactivated("deactivated", this.machine);
    
		try
		{    	
			if (sStat == null || sStat.isClosed())
			{		    
				sStat = new Socket(config.getManagerIPAddress(),23000);
				outStat = new ObjectOutputStream(sStat.getOutputStream());	

				//System.out.println("Create new outStat");
				//System.out.println("Create new sStat");	    
			}		
			
			outStat.writeObject(dd);
			outStat.close();
			sStat.close();
			//System.out.println("Close outStat");
			//System.out.println("Close sStat");    
		} 
		catch(Exception e)
		{
			e.printStackTrace();
		}
		return true;
	}

	//---------------------------------------------------------------------------------------------------

	/**
	 * In response to SendPartitons message
	 * @param sender
	 * @param partitionsToSend
	 * @param toMachine
	 * @return
	 */
	
	public boolean sendPartitions(XATStreamOperator sender, ArrayList partitionsToSend, Machine toMachine)
	{	
		// (1) check whether all on-the-fly inputs from partitions in partitionsToSend have been received
		// assume the sender is a m-way hash join operator
	
		try 
		{
			SymmMwayHashJoinOperatorImp sendOp = (SymmMwayHashJoinOperatorImp)sender;
			
			//sendOp.setStartChecking(true);
			sendOp.setPartitionsToMove(partitionsToSend);
			System.out.println("Begin to check no more stoped tuples?");
		
			while (!sendOp.getDeactivatedCompleteFlag()) 
			{			
				Thread.sleep(1500);
			}
		
			System.out.println("Yes, no more stoped tuples coming");
			// wait until on-the-fly inputs have been processed
			sendOp.setStartChecking(false);
			sendOp.setPartitionsToMove(null);
		
			// (2) get the real join partitions from sendOp, note that the corresponding IDs are indicated by partitionsToSend
			ArrayList pToMove = sendOp.getPartitionsByID(partitionsToSend);
		
			// (3) send the partitions to receiver via ReceiverPartitions message
			Socket fs = new Socket(toMachine.getIPAddress(),toMachine.getConnectionListenerPort());					 
			ObjectOutputStream fout = new ObjectOutputStream(fs.getOutputStream());
				
			ReceivePartitions rp = new ReceivePartitions(sendOp.getOperatorID(),partitionsToSend,pToMove);
			Packet pk5 = new Packet("protocol",rp);
			fout.writeObject(pk5);
			fout.close();
			fs.close();			
		} 
		catch (Exception e) 
		{		
			e.printStackTrace();
		}
	
		// now resume the disk-based adaptation of this operator
		((SymmMwayHashJoinOperatorImp)sender).resumeDiskAdaptation();
		return true;
	}
	
	//---------------------------------------------------------------------------------------------------

	public boolean sendPartitionsInTree(ArrayList partitionsToSend, Machine toMachine)
	{
		XATNode[] nodes = ((XATTree)tree.get(0)).convertToArray();
		ArrayList partitionsToMove = new ArrayList();
		ArrayList joinOperators = new ArrayList();
		System.out.println("Partitions to send: " + partitionsToSend);
    
		// (1) check whether all on-the-fly inputs from partitions in partitionsToSend have been received
		// assume the sender is a m-way hash join operator
		try 
		{
			//traverse the tree from bottom-up		
			for(int i=nodes.length-1; i>=0; i--) 
			{		    
				if(nodes[i].getStreamOperator() instanceof SymmMwayHashJoinOperatorImp) 
				{
					SymmMwayHashJoinOperatorImp sendOp = (SymmMwayHashJoinOperatorImp)nodes[i].getStreamOperator();
					joinOperators.add(sendOp);
					sendOp.setPartitionsToMove(partitionsToSend);				
					System.out.println("Begin to check no more stoped tuples?");
				
					//wait until on-the-fly inputs have been processed
					while (!sendOp.getDeactivatedCompleteFlag()) 
					{
						Thread.sleep(1000);
					}
					System.out.println("Yes, no more stoped tuples coming");
				
					//sendOp.setStartChecking(false);
					//sendOp.setPartitionsToMove(null);
					// (2) get the real join partitions from sendOp, note that the corresponding IDs are indicated by partitionsToSend
					ArrayList pToMove = sendOp.getPartitionsByID(partitionsToSend);
								
					for(int k = 0; k < pToMove.size(); k++) 
					{
						HashStatePartitionGroup pgroup = (HashStatePartitionGroup)pToMove.get(k);
                    
						if(pgroup != null) 
						{
							WEquiJoinStateImp[] states = pgroup.getHashStates();
				    
							if(states.length != sendOp.getInputQueueCount()) 
							{
								System.out.println("Error! Number of states do not equal to number of inputs for " + sendOp);
								System.exit(-1);
							}
                        
							Iterator it = sendOp.getInputQueues();
							int index = 0;
                        
							while (it.hasNext()) 
							{
								String schemaname = ((XATQueue)it.next()).getSchema().getName();
                        
								System.out.println("For operator: " + sendOp);
								System.out.println("set states " + index + " to have schema " + schemaname);
								//System.out.println("states " + index + " contain tuples: " + states[index]);
								states[index].setStateSchemaName(schemaname);
								partitionsToMove.add(states[index]);
								index++;                        
							}                    
						}				
					}			

					// now resume the disk-based adaptation of this operator
					//sendOp.resumeDiskAdaptation();		    
				}		
			}

			//setup sockets connections.
			Socket fs = new Socket(toMachine.getIPAddress(),toMachine.getConnectionListenerPort());					 
			ObjectOutputStream fout = new ObjectOutputStream(fs.getOutputStream());
		
			// (3) send the partitions to receiver via ReceiverPartitions message
			ReceivePartitionsInTree rp = new ReceivePartitionsInTree(partitionsToSend, partitionsToMove);
			Packet pk5 = new Packet("protocol", rp);
			fout.writeObject(pk5);
			fout.close();
			fs.close();
			partitionsToMove = null;
		
			for(int index = 0; index < joinOperators.size(); index++) 
			{
				WTransitiveHashJoinOperator operator = (WTransitiveHashJoinOperator)joinOperators.get(index);
				operator.setStartChecking(false);
				operator.setPartitionsToMove(null);
			}
		} 
		catch (Exception e) 
		{
			e.printStackTrace();
		}	
		return true;
	}
	
	//---------------------------------------------------------------------------------------------------

	public boolean receivePartitions(XATStreamOperator receiver, ArrayList ids, ArrayList parts) 
	{    
		try 
		{		
			// here we assume the operator is SymmMwayHashJoinOperator		
			SymmMwayHashJoinOperatorImp recOp = (SymmMwayHashJoinOperatorImp)receiver;
			System.out.println("Receiving Partition IDs:" + ids.toString());
			recOp.setPartitionsByID(ids, parts);		
		
			// notifiy the DM that the partitions are received and installed
			if (sStat == null || sStat.isClosed())
			{
				sStat = new Socket(config.getManagerIPAddress(),23000);
				outStat = new ObjectOutputStream(sStat.getOutputStream());	

				//System.out.println("Create new outStat");
				//System.out.println("Create new sStat" + sStat);
				//Socket testsocket = new Socket(config.getManagerIPAddress(),23000);
				//System.out.println("Create new testsStat" + testsocket);
			}	
			
			//if(outStat == null){
			//	outStat = new ObjectOutputStream(sStat.getOutputStream());			
			//}
    	
			Received rd = new Received("received");
			outStat.writeObject(rd);					
		} 
		catch (Exception e) 
		{
			e.printStackTrace();
		}
		return true;
	}

	//---------------------------------------------------------------------------------------------------
	
	/**
     * Receive all partitions from sender machine. Some partitions have matching
     * partitions in receiver, they can be inserted to the state right away.
     * Other partitions do not have matching partitions. This means the query
     * plan has been optimized either on local machine or on the sender machine.
     * In this case, some received partitions may need to be discarded, and some
     * local partitions need to be computed, as in the moving state strategy of
     * query plan migration. Received state partitions (not partition groups)
     * each has a stateschemaname, which is used to match with each operator's
     * input queues and determine which partition goes to which state. If a
     * partition is not matched with any input queue schema, this partition
     * belong to a state that does not exist in current machine. The unmatched
     * state on current machine will need to be recomputed.
     * 
     * @param ids
     * @param parts
     * @return
     */
    public boolean receivePartitionsInTree(ArrayList ids, ArrayList parts) 
    {
        long receiveStartTime = System.currentTimeMillis();
        System.out.println("Received partitions from sender");
        System.out.println("Receiving Partition IDs:" + ids.toString());
        System.out.println("Current tree is: " + (XATTree)this.tree.get(0));        
        XATNode[] nodes = ((XATTree) tree.get(0)).convertToArray();

        //First organize received hash partitions.
        //Create a hashtable with keys as the stateschemaname, and the item is another hashtable.
        //The second hashtable use partitionid as key and the corresponding item is single state partition.
        Hashtable namehash = new Hashtable();
        for (int k = 0; k < parts.size(); k++) 
        {
            WEquiJoinStateImp state = (WEquiJoinStateImp) parts.get(k);
            int id = state.getTuplePartitionID();
            
            if(id >= 0) 
            {
                String stateschemaname = state.getStateSchemaName();
                Object nh = namehash.get(stateschemaname);
                Hashtable idhash;
                if (nh == null)
                    idhash = new Hashtable();
                else
                    idhash = (Hashtable) nh;

                idhash.put(new Integer(id), state);
                namehash.put(stateschemaname, idhash);
            }
        }
        
        //System.out.println("The received partitions are organized as: " + namehash);
        parts = null;

        //now put received partitions in the right operator at the right position
        //record operator and position that is unmatched in a hashtable.
        ArrayList unmatchedOps = new ArrayList();
        ArrayList unmatchedPositions = new ArrayList();
        for (int i = nodes.length - 1; i >= 0; i--) 
        {
            if (nodes[i].getStreamOperator() instanceof SymmMwayHashJoinOperatorImp) 
            {
                SymmMwayHashJoinOperatorImp recOp = (SymmMwayHashJoinOperatorImp) nodes[i].getStreamOperator();
                //System.out.println("\nTry to match received partitions to states in: " + recOp);
                Iterator it = recOp.getInputQueues();
                int position = 0;
                while (it.hasNext()) 
                {
                    String queuename = ((XATQueue) it.next()).getSchema().getName();
                    //System.out.println("The queue schema is:" + queuename);
                    
                    //now search received partitions for a match of schemanames.
                    Hashtable ht = (Hashtable)namehash.get(queuename);
                    //found matching partitions
                    if (ht != null) 
                    {
                        //System.out.println("Found a match mapping to input queue: " + queuename);
                        Hashtable idtopart = (Hashtable)ht;
                        Enumeration keys = idtopart.keys();
                        
                        while(keys.hasMoreElements()) 
                        {
                            Integer id = (Integer)keys.nextElement();
                            WEquiJoinStateImp state = (WEquiJoinStateImp)idtopart.get(id);
                            recOp.setSinglePartitionByID(id.intValue(), position, state);
                        }
                    } 
                    else 
                    {
                        unmatchedOps.add(recOp);
                        unmatchedPositions.add(new Integer(position));
                        //System.out.println("Find unmatched! op: " + recOp + " pos: " + position + " name: " + queuename);
                    }
                    position++;         
                }
            }
        }

        //now recompute unmatched partitions on current machine if there is any
        while (unmatchedOps.size() > 0) 
        {
            System.out.println("\nRecompute unmatched states!");
            for (int i = 0; i < unmatchedOps.size(); i++) 
            {
                SymmMwayHashJoinOperatorImp curOp = (SymmMwayHashJoinOperatorImp) unmatchedOps.get(i);
                int position = ((Integer)unmatchedPositions.get(i)).intValue();

                System.out.println("Unmatched operator has cop:" + curOp.getNode().getOpChild(position));
                XATStreamOperator tempop = curOp.getNode().getOpChild(position).getStreamOperator();
                SymmMwayHashJoinOperatorImp cop = (SymmMwayHashJoinOperatorImp) tempop;
                // if child op also has unmatched partition, do not recompute now.
                if (unmatchedOps.contains(cop))
                    continue;

                // if partitions in child op have all been matched, recompute
                // partition for the parent op.
                ArrayList newstatepartitions = cop.joinPartitionsByIDs(ids);
                curOp.setStatesByID(ids, position, newstatepartitions);
                unmatchedOps.remove(i);
                unmatchedPositions.remove(i);
                i--;
            }
        }
        
        //notifiy the DM that the partitions are received and installed
        try
        {
            if (sStat == null || sStat.isClosed()) 
            {
                sStat = new Socket(config.getManagerIPAddress(), 23000);
                outStat = new ObjectOutputStream(sStat.getOutputStream());
            }
            Received rd = new Received("received");
            outStat.writeObject(rd);
        } 
        catch (Exception e) 
        {
            e.printStackTrace();
        }
        unmatchedOps = null;
        unmatchedPositions = null;
        namehash = null;
        nodes = null;
        
        System.out.println("Receiving partitions takes: " + (System.currentTimeMillis() - receiveStartTime));
        return true;
    }
    
    //---------------------------------------------------------------------------------------------------

    public boolean reactivatePartitions(Object[] children, ArrayList partitionsToAct, Machine toM)
    {
    	for (int i = 0; i < children.length; i++)
    	{
    		XATStreamOperator curOp = ((XATNode)children[i]).getStreamOperator();
		
    		if ((curOp instanceof StreamSplitOperatorImp) && (curOp.isActive())) 
    		{
    			StreamSplitOperatorImp splitop = (StreamSplitOperatorImp)curOp;
    			splitop.reactivatePartitions(partitionsToAct, toM.getIPAddress());
    			System.out.println(splitop + "partition table: \n" + splitop.getPartitionMappingTable());
    		}
    	}

    	System.out.println("Yes, Reactivated.\n\n");
    	return true;
    }
    
   //---------------------------------------------------------------------------------------------------
    
    public boolean reactivatePartitionsInTree(ArrayList partitionsToAct, Machine toM)
    {
    	XATNode[] nodes = ((XATTree)tree.get(0)).convertToArray();
    	for (int i = 0; i < nodes.length; i++)
    	{
    		XATStreamOperator curOp = nodes[i].getStreamOperator();
    		if ((curOp instanceof StreamSplitOperatorImp) && (curOp.isActive())) 
    		{
    			StreamSplitOperatorImp splitop = (StreamSplitOperatorImp)curOp;
    			splitop.reactivatePartitions(partitionsToAct, toM.getIPAddress());
    			//System.out.println(splitop + "partition table: \n" + splitop.getPartitionMappingTable());
    		}
    	}
    	System.out.println("Yes, Reactivated.\n\n");
    	return true;
    }
    
    //---------------------------------------------------------------------------------------------------

	/*
	 * For using the simulated cost model, the join operator can not based on the main memory usage, which could
	 * be got in the operator itself. In the simulated case, the cost information is set by the execution controller
	 * to the join operator. The number of tuples in the input_queue and output_queue of all operators in the machine
	 * is set to the join operator. -- Bin
	 */

    public void updateSimCostofSymmJoin(XATStreamOperator oper) 
    {
    	if (oper instanceof SymmMwayHashJoinOperatorImp) 
    	{
    		// compute the all tuples in all queues of all operators in the machine?
    		int totalInInQs = 0;
    		int totalInOutQs = 0; 
		
    		for (int i = 0; i < tree.size(); i++) 
    		{			
    			XATTree currentTree = (XATTree) tree.get(i);
    			XATNode[] curNode = currentTree.convertToArray();
			
    			for (int j = 0; j < curNode.length; j++)
    			{
    				if ((curNode[j]).getStreamOperator().isActive())
    				{					
    					totalInInQs += Integer.parseInt(stats.getValue("NUMBER_OF_TUPLES_IN_INPUT_QUEUES",(curNode[j]).getStreamOperator()));
    					//totalInOutQs += Integer.parseInt(stats.getValue("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES",(curNode[j]).getStreamOperator()));
    				}				
    			}
    		}

    		//only count the outputqueue size of the join operator
    		totalInOutQs = Integer.parseInt(stats.getValue("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES",oper));
		
    		// it might be temporarliy overload, we do not want to show this, the effect on the main memory usage is not linear
    		if (totalInOutQs > 15000)
    		{
    			totalInOutQs = 15000 + ((totalInOutQs-15000)/10);
    		} 
    		//end of overload_control
		
    		int total = totalInInQs + totalInOutQs;
    		// note that in case that the local adapter makes the decisions of the adaptation, then 
    		// we do not need to set the queuesize information into the join operator, instead, we 
    		// only need to set the information to the localAdapter. I leave the duplicated code to 
    		// make it compatible to the old code that developed for handling only one join operator case
    		// that the local adaptation decision is made by the join operator itself. -- Bin, Apr, 2005
    		((SymmMwayHashJoinOperatorImp)oper).setTotalMachineQueueSize(total);
    		//System.out.println("INQUEUES:OUTQUEUES" + totalInInQs + ":" + totalInOutQs);

    		if (localAdapter != null)
    		{
    			localAdapter.updateQueueSizeInfo(total);
    		}
    	}	
    	return;
    }
  
    //---------------------------------------------------------------------------------------------------
    
    /*
     * When finializing the query tree with multiple SymmMwayHashJoinOperators, we have to clean up them one by one
     * with first cleaning the bottom join operator. Here, we check the previous_join_reference of each join to indicate
     * whether the cleanup of the operator has been done. We then change the references when the bottom join operator completes.
     * -- Bin Apr 15, 2005
     */
    
    public void updateJoinReferences(int opID)
    {
    	if (localAdapter!= null)
    	{
    		localAdapter.tearDownJoinReferences(opID);
    	}
    }
    
	//-----------------------------------------------------------------------------
	
	//STQM
	private boolean CanDoQMSelfTuning(Hashtable hashSTQMDataSamplesGlobalStorage)
	{
		boolean canDoQMSelfTuning = false;
		int numStreamsReceivedSchema = 0;
		int numExpectedStreams = 0;	
		
		//-----------------------------------------------------------------------------
		// Perform the following checks to ensure that STQM can be executed
		// 1. All input streams' schemes have been received
		// 2. Data samples for all streams have been present
		// 3. All operators have some statistics present
		//-----------------------------------------------------------------------------
		
		//-----------------------------------------------------------------------------
		// CHECK INPUT STREAMS' SCHEMAS HAVE BEEN RECEIVED
		//-----------------------------------------------------------------------------
		XATNode[] nodes = ((XATTree) tree.get(0)).convertToArray();		 
		for (int i = nodes.length - 1; i >= 0; i--) 		     
		{		            
			if (nodes[i].getStreamOperator() instanceof OnlineClassifierOperatorImp) 		 
			{		                
				OnlineClassifierOperatorImp recOp = (OnlineClassifierOperatorImp) nodes[i].getStreamOperator();		     		          			
				numExpectedStreams = recOp.getNumStreams();				
			}
			else if (nodes[i].getStreamOperator() instanceof TrainingTupleCollectorOperatorImp) 		 
			{
				TrainingTupleCollectorOperatorImp recOp = (TrainingTupleCollectorOperatorImp) nodes[i].getStreamOperator();		     
		        
				//this is a "leaf" operator, check to see if it has received stream schema
				if (recOp.getInputQueue(0).getSchema() != null)
				{
					numStreamsReceivedSchema++;
				}
			}
		}
		
		//compare number expected streams vs. number of streams from which received the schema
		if (numStreamsReceivedSchema == numExpectedStreams)
		{
			canDoQMSelfTuning = true;
		}
		else
		{
			canDoQMSelfTuning = false;
			return canDoQMSelfTuning;
		}
				
		Helper.WriteToFile(Helper.LOG_FILE,  "\t"+Helper.GetCurrentTimeString() + ": --- CanDoQMSelfTuning: All schemas were received --- ");
		 
		 Enumeration e = hashSTQMDataSamplesGlobalStorage.keys();
		 while (e.hasMoreElements()) 
		 {
			int key = Integer.parseInt(e.nextElement().toString());				
		   	//System.out.println("Stream "+key+": "); //key here is a stream
		   	
		   	//retrieve data sample (which is a pointer to the sampling operator)
		   	DataSampleStorage ds = (DataSampleStorage)hashSTQMDataSamplesGlobalStorage.get(key);	
		   	TrainingTupleCollectorOperatorImp physOp = ds.GetOp();
		   	
		   	//get the new training sets for the sampling operator for the stream
		   	ArrayList newSample = physOp.GetNewTuplesSample();
		   	
		   	if (newSample.size() == 0)
		   	{
		   		return false;
		   	}		   			   
		}
		
		/*
		//------------------------------------------------------------------
		 // 2. CHECK ALL DATA SAMPLES ARE PRESENT
		 //------------------------------------------------------------------		 
		 Enumeration e = hashSTQMDataSamplesGlobalStorage.keys();
		 while (e.hasMoreElements()) 
		 {
			int key = Integer.parseInt(e.nextElement().toString());				
		   	//System.out.println("Stream "+key+": "); //key here is a stream
		   	
		   	//retrieve data sample (which is a pointer to the sampling operator)
		   	DataSampleStorage ds = (DataSampleStorage)hashSTQMDataSamplesGlobalStorage.get(key);	
		   	TrainingTupleCollectorOperatorImp physOp = ds.GetOp();
		   	
		   	//get the old and the new training sets for the sampling operator for the stream
		   	ArrayList oldSample = physOp.GetOldTuplesSample();
		   	ArrayList newSample = physOp.GetNewTuplesSample();
		   	
		   	StreamOldAndNewSamples ss = new StreamOldAndNewSamples();
		   	ss.SetNewSamples(newSample);
		   	ss.SetOldSamples(oldSample);
		   	
		   	if ((ss.GetOldSamples().size() == 0) || (ss.GetNewSamples().size() == 0))
		   	{
				canDoQMSelfTuning = false;
				Helper.WriteToFile(Helper.LOG_FILE,  "\t"+Helper.GetCurrentTimeString() + ": --- CanDoQMSelfTuning: Stream "+key +" NOT all data samples are present --- ");
				return canDoQMSelfTuning;
		   	}
		   	else
		   	{
		   		Helper.WriteToFile(Helper.LOG_FILE,  "\t"+Helper.GetCurrentTimeString() + ": --- CanDoQMSelfTuning: Stream "+key +" All data samples are present --- ");
		   	}
		   	//hashSamplesPerStream.put(key, ss);
		}
		 
		 //------------------------------------------------------------------
		 // 3. CHECK ALL EXECUTION STATISTICS ARE PRESENT
		 //------------------------------------------------------------------
		 
		 e = hashSTQMExecutionStatisticsSamplesGlobalStorage.keys();
		 while (e.hasMoreElements()) 
		 {
			int key = Integer.parseInt(e.nextElement().toString());	// key = operator id
			
			if (key == 0)
			{
				continue;
			}
			
		   	//System.out.println("Operator id "+key+": "); //key here is operator id
		   	
		   	ExecutionStatisticsSampleStorage es = (ExecutionStatisticsSampleStorage)hashSTQMExecutionStatisticsSamplesGlobalStorage.get(key); 
		   	STeMJoinProbeOperatorImp physOp = (STeMJoinProbeOperatorImp)es.GetOp();
		   	
		   	Hashtable hashOpStatsPerRuster = physOp.GetNewOperatorExploitationStatisticsPerRuster();
		   	
		   	if (hashOpStatsPerRuster.size() == 0)
		   	{
				canDoQMSelfTuning = false;
				Helper.WriteToFile(Helper.LOG_FILE,  "\t"+Helper.GetCurrentTimeString() + ": --- CanDoQMSelfTuning: Operator "+key +" No execution statistics are present --- ");
				return canDoQMSelfTuning;
		   	}
		   	else
		   	{
		   		Helper.WriteToFile(Helper.LOG_FILE,  "\t"+Helper.GetCurrentTimeString() + ": --- CanDoQMSelfTuning: Stream "+key +" All execution statistics are present --- ");
		   	}
		   	//hashSamplesPerOperator.put(key, hashOpStatsPerRuster);
		 }
		 */
		
		return canDoQMSelfTuning;
	}
	
	//-----------------------------------------------------------------------------
	
    
  //---------------------------------------------------------------------------------------------------
            
   /*
    * Initializes the statistics printer by call the config file parser and setting all of the
    * properties that are returned
    * 
    */
    
    protected void initStatisticsPrinter() throws Exception 
    {
    	printer = new StatisticsPrinter();
    
    	for (Iterator i = config.getPrinterOutputFormat(); i.hasNext();) 
    	{        
    		NameValuePair pair = (NameValuePair) i.next();
    		printer.setProperty(pair.getName(), pair.getValue());        
    	}

    	// set machine name to the printer
    	printer.setProperty("MACHINE_NAME", machine.getMachineName());
   
    	ArrayList columnList = new ArrayList();   
    	for (Iterator i = config.getPrinterColumnNames();i.hasNext();) 
    	{
    		columnList.add(i.next());   	
    	}   
    	printer.setColumnHeaders(columnList);
    }

    //---------------------------------------------------------------------------------------------------

    /**
     * Calls config file parser and gathers all of the statistics that 
     * should be gatherered for an operator both every time it is run and perodically
     *
     * This will also do the same for tree statistics
     */

    protected void initStatsToGather() throws Exception 
    {
    	/** call populate statistics to do the real work */
    	everyTimeOperatorStatistics = populateStatistics(config.getEveryTimeOperator());
    	periodicOperatorStatistics = populateStatistics(config.getPeriodicOperator());
    	periodicTreeStatistics = populateStatistics(config.getPeriodicTree());
    	everyTimeSystemStatistics = populateStatistics(config.getEveryTimeSystem());	
    }
    
    //---------------------------------------------------------------------------------------------------

    /**
     * This method first checks if the query plan need to be optimized or not. If yes, it then locates the 
     * part of the query plan that need to be optimized and apply rewriting rules to the sub-plan. 
     * It then creates the old and new boxes and passes them to the migration controller to actually finish the migration.
     * For moving state, it's straight-forward, but for parallel states, the old sub-query plan has to be removed
     * after it is done purging all the old tuples. 
     *
     */

    public void localPlanOptimization() 
    {	
    	boolean treechanged = false;

    	for (int i = 0; i < tree.size(); i++) 
    	{		
    		XATTree currentTree = (XATTree) tree.get(i);
		
    		/** if current tree is locked by another processor, continue to the next tree */		
    		if(currentTree.isLocked())			
    			continue;
			
    		//long migrationStartTime = System.currentTimeMillis();
    		currentTree.setLocked(true);
		
    		System.out.println("Start Optimizing the query plan");
				
    		// if the tree is changed, return true, otherwise return false
    		//if(optimizer.optimizeQueryTree(currentTree)) {
/*		
		if(optimizer.efficientOptimization(currentTree, this)) {
			treechanged = true;	
			currentTree.addOptimizedCounter();
			currentTree.setOptimized(true);
			//System.out.println("Migration takes: " + (System.currentTimeMillis() - migrationStartTime));
			System.out.println("Tree is changed.");	
			this.currentScheduler.setTree(currentTree);
			//System.out.println("The new tree is: \n" + currentTree);
/*			try {
				
//				String className = this.getProperty("ADAPTIVE_HEURISTIC");
//				if (className.compareTo("edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.NeverRotateAdapter") == 0)
//					neverRotateScheduler = true;					

				XATStreamOperator[] oldOperators = optimizer.getOldBox().getAllOperators();	
				for (int j = 0; j < oldOperators.length; j++) {

					stats.removeQueryObject(oldOperators[j]);
					lastTimeOperatorUpdated.remove(oldOperators[j]);
					
					Iterator oldQueues = oldOperators[j].getInputQueues();				
					while (oldQueues.hasNext()) {
						stats.removeQueryObject((XATQueue) oldQueues.next());
					}
				}
				
				XATStreamOperator[] newOperators = optimizer.getNewBox().getAllOperators();		
				stats.addNewQueryObject(newOperators);

				for (int k = 0; k < newOperators.length; k++) {

					lastTimeOperatorUpdated.put(newOperators[k], new Long(0));
					
					Iterator newQueues = newOperators[k].getInputQueues();
					while (newQueues.hasNext()) {
						stats.addNewQueryObject((XATQueue) newQueues.next());
					}
				}
//				if(gui instanceof NonDistributedGUI) {
//					((NonDistributedGUI)gui).redrawTree();
//				}
			} catch (Exception e) {
				System.out.println("Exception caught during query plan optimization: " + e);
			}
					
		} else
			System.out.println("The current query plan is not changed.");
*/			
		currentTree.setLocked(false);
	}
	
	//reset trees in scheduler
/*	if(treechanged) {
		Enumeration e = adapter.getSchedulerObjects();
		while(e.hasMoreElements()) {
			((XATScheduler)e.nextElement()).setTrees((XATTree[])tree.toArray(new XATTree[] {}));
		}
	}*/

    }

    //---------------------------------------------------------------------------------------------------

    /**
     * This will update the statistics for operators and take care of any additional
     * book keeping before the operator is allowed to run.
     * @param decision XATSchedulerDecision
     * @return boolean true if the operator should continue to run, false if it should not 
     */
    
    protected boolean performPreExecTasks(XATSchedulerDecision decision) throws Exception 
    {		
    	try
    	{
    		while(treeLocked) 
    		{
    			Thread.sleep(1000);
    			System.out.println("perform pre exec: tree is locked");
    		}	
    	}
    	catch(Exception e) 
    	{
    		e.printStackTrace();
    	}
	
    	this.treeLocked = true;
    	boolean returnValue = true;
    
	    /**
	     * if the workload is 0, then we dont want to run the operator,
	     * but rather note that we detected starvation.  If
	     * we get too many starvations in a row, we will force a change of algorithm
	     * If that doesnt work, the execution will terminate.
	     * Note: some algorithms have built in starvation checks, therefore, this
	     * checking is mearly a little overhead.  However, other algorithms are
	     * inherently prone to starvation and thus, this check is necessary, otherwise
	     * we dont get anywhere.  
	     */	
        
    	if (decision.getNumberOfTuples() < 1) 
    	{
    		// There is no work to do.
    		this.treeLocked = false;
    		return false;    
    	} 
    	else 
    	{
    		XATStreamOperator operator = (XATStreamOperator) decision.getNextOperator();
	
    		if (debug) 
    		{
    			System.out.println("Next operator to run: " + operator);
    			System.out.println("Dequeuing this many tuples: " + decision.getNumberOfTuples() + "\n");
    		}
    		
    		/** set the operator's property */
    		operator.setProperty("NUMBER_OF_TUPLES_TO_DEQUEUE", "" + decision.getNumberOfTuples());
    		stats.updateValue("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES", operator);
        
    		/** 
    		 * figure out how many tuples are available for this operator
    		 * and update the TUPLE_COUNT statistic.  This will be used
    		 * later to figure out the number of possible outputs.
    		 */
    		
    		for (Iterator i = operator.getInputQueues(); i.hasNext();) 
    		{
    			XATQueue queue = (XATQueue) i.next();
    			int work = queue.getQueueSize(operator);
    			stats.updateValue("TUPLE_COUNT", queue, "" + work);
    		}
    		returnValue = true;
    	}
    
    	this.treeLocked = false;
    	return returnValue;
    }
    
    //---------------------------------------------------------------------------------------------------
    
	/**
	 * Takes an iterator of NameValuePairs and adds the names to the given String array.
	 * This is used for taking the names of the statisticsToGather and putting them into the appropriate array 
	 * @param iterator Iterator
	 * @return array java.lang.String[]
	 */
    
    protected String[] populateStatistics(Iterator iterator) 
    {
    	ArrayList list = new ArrayList();
    	for (; iterator.hasNext();) 
    	{
    		list.add(iterator.next());
    	}
    
    	//initialize the Array size
    	String[] array = new String[list.size()];
    
    	int index = 0;
    	//initialize the properties in the array
    	
    	for (Iterator i = list.iterator(); i.hasNext();) 
    	{
    		NameValuePair pair = (NameValuePair) i.next();
    		array[index] = pair.getName();
    		index++;
    	}
    	return array;
    }
    
    //---------------------------------------------------------------------------------------------------

    /*
     * Runs the actual execution
     *
     */
    
    public void run() 
    {
    	System.out.println("Starting a new Distributed Execution Controller.");
    	//Initially try to run the thread...
    
    	try 
    	{
    
    	} 
    	catch (Exception e) 
    	{
    		System.out.println(e.toString());
    		return;    
    	}    
    	// Here is where we start execution.

    	XATSchedulerDecision d = null;
    	int totalTuples;
    	int memCount = 0;
    	boolean notInterrupted = true;
	    startTime = System.currentTimeMillis();
	    long currentTime = System.currentTimeMillis();
	    boolean timeUp = false;
	    boolean runSuccess = false;
	    long lastRedistTime = currentTime;
		long lastMigrationTime = currentTime;
		int migrationcount = 0;
		int totalMigrationCount = Integer.parseInt(this.getProperty("TOTAL_MIGRATION_TIMES"));
	    //Socket sStat = null; // -- Make these two global to allow access from other methods
		//ObjectOutputStream outStat = null;
		DecimalFormat df = new DecimalFormat("0.00#");
		XATStreamOperator operator;
		int oldOutput;
		String timeBefore;
		int reDistTime = Integer.parseInt(config.getProperty("REDISBRIBUTION_TIME"));
		int runTime = Integer.parseInt(config.getProperty("EXECUTION_DURATION"));
		int local_migration_interval = Integer.parseInt(this.getProperty("MIGRATION_INTERVAL"));
		Runtime rt = Runtime.getRuntime();
		String timeAfter;
		String timeRun;
		LightWeightStatistic lws;
		//XATQueueManager qm = XATQueueManager.getQueueManager();
		rt.gc();
	
	    try
	    {    	
	    	Thread.sleep(2000);
		    sStat = new Socket(config.getManagerIPAddress(),23000);
			
	    }
	    catch(Exception e)
	    {
		    e.printStackTrace();
	    }
	    //Find out how long to run the machine.
	
	    System.out.println("Will process data for " + runTime + " milliseconds.");
	    
	    //Grab the initial memory profile, while states and queues are empty.
	    System.gc();
	    systemOverhead = rt.totalMemory() - rt.freeMemory();
	    double lastMemoryUsage = 0.0;
	    try 
	    {
	    	//int cc = 0;
	    	while (keepGoing && !timeUp) 
	    	{
	    		//System.out.println("Starting while loop");
	    		try 
	    		{
	    			currentTime = System.currentTimeMillis();
	    			
	    			//if(currentTime - startTime > runTime && inGlobalAdaptation == false){
	    			if(currentTime - startTime > runTime)
	    			{
	    				timeUp = true;
	    				System.out.println("Time is up!");
	    			}
				
	    			if(currentTime - lastRedistTime > reDistTime)
	    			{
	    			    System.out.println("\nSystem has run for: " + (currentTime - startTime) +  " miliseconds");
	    				if (sStat == null) {
	    				    sStat = new Socket(config.getManagerIPAddress(),23000);
	    				}
	    				
	    				if (bTesting)
	    				{	
		    				System.out.println("Current Stats:");
		                    System.out.println("-----------------------------------------------------");
		                    System.out.println(stats.getValue("TOTAL_TUPLES_IN_QUEUES", (XATTree)this.tree.get(0)) + " tuples in input queues");
		                    System.out.println(stats.getValue("NUMBER_OF_TIMES_OPERATORS_WERE_RUN", (XATTree)this.tree.get(0)) + " operator executions");
	    				}    
		                stats.updateValue("THROUGHPUT", (XATTree)this.tree.get(0));
		                stats.updateValue("OUTPUT_RATE", (XATTree)this.tree.get(0));
		                    
		    			if (bTesting)
		    			{	
		                    System.out.println(stats.getValue("THROUGHPUT", (XATTree)this.tree.get(0)) + " throughput");
		                    System.out.println(stats.getValue("OUTPUT_RATE", (XATTree)this.tree.get(0)) + " output rate");
	    				}
	    				
	                    Iterator opit = stats.getAllKnownOperators();
	                    while(opit.hasNext())
	                    {       
	        				if (bTesting)
	        				{	
	        					System.out.println("new op");
	        				}
	        				
	                        XATStreamOperator op = (XATStreamOperator)opit.next();
	                        
	        				if (bTesting)
	        				{	
	        					System.out.println(op);
	        				}
	                        
	                        if(op.isActive() && op.getClass().getName().indexOf("Join") >-1)
	                        {
	            				if (bTesting)
	            					System.out.println("join " + op.getOperatorID() + ": " + stats.getValue("NUMBER_OF_TUPLES_OUTPUTTED", op) + " tuples output");
	                           OperatorState[] states = op.getState();
	                           int i = 0;
	                           /*
	                           for (OperatorState state : states)
	                           {
	                               System.out.println("  : " + state.getStateSize() + " tuples in state " + (i++));
	                           }
	                           */
	                           i = 0;
	                           Iterator iqit = op.getInputQueues();
	                           while (iqit.hasNext())
	                           {
	                               XATQueue queue = (XATQueue)iqit.next();
	                               if (bTesting)
	                            	   System.out.println("  : " + queue.getQueueSize() + " tuples in queue " + (i++));
	                           }
	                        }
	                        else
	                        {
	                        	 int i = 0;
	                        	 Iterator iqit = op.getInputQueues();
	                             while (iqit.hasNext())
	                             {
	                 				if (bTesting)
	                 					System.out.println("non-join " + op.getOperatorID() + ": " + stats.getValue("NUMBER_OF_TUPLES_OUTPUTTED", op) + " tuples output");
	                                 XATQueue queue = (XATQueue)iqit.next();
	                 				if (bTesting)
	                 					System.out.println("  : " + queue.getQueueSize() + " tuples in queue " + (i++));
	                             }
	                        }
	                    }
	                    
	    				if(sStat.isClosed()) 
	    				{
	    				    sStat = new Socket(config.getManagerIPAddress(),23000);
	    				    outStat = new ObjectOutputStream(sStat.getOutputStream());
	    				    //System.out.println("Create new socket sStat");
	    				    //System.out.println("Create new socket outState");
	    				    //sStat.connect(sStat.getRemoteSocketAddress());
	    				}
	    				if(outStat == null)
	    				{
	    					outStat = new ObjectOutputStream(sStat.getOutputStream());
	    				}
	
	    				lastRedistTime = currentTime;
	    				if(lightStats != null)
	    				{    					
	        				if (bTesting)
	        					System.out.println("lightStats: ["+ lightStats + "]");
	    					lws = new LightWeightStatistic(this.clientName, lightStats);
	    					outStat.writeObject(lws);
	        				if (bTesting)
	        					System.out.println("Send lightWeightStats to DM");
	    				}
					
	    				//System.out.println("-------------Memory Information " + memCount++ +" --------------");
	    				//System.out.println("Tuples in input pool: " + this.inPool.size());
	    				//System.out.println(this.stats.displaySize());
	    				//System.out.println("Free memory: " + df.format(((double)rt.freeMemory()/1024.0/1024.0)) + "MB");
	    				//System.out.println("Total memory: " + df.format(((double)rt.totalMemory()/1024.0/1024.0)) + "MB");
	    				//System.out.println("Max memory: " + df.format(((double)rt.maxMemory()/1024.0/1024.0)) + "MB");
	    				//System.gc();
	    				//System.out.println("Free memory: " + df.format(((double)rt.freeMemory()/1024.0/1024.0)) + "MB");
	    				//System.out.println("Total memory: " + df.format(((double)rt.totalMemory()/1024.0/1024.0)) + "MB");
	    				//System.out.println("Max memory: " + df.format(((double)rt.maxMemory()/1024.0/1024.0)) + "MB");	
	    				//System.out.println("----------------------------------------------");
	    			}
	    			
	    			
	    			/** have the scheduler make the next decision */
	    			while(currentScheduler.isLocked) 
	    			{
	    				if (bTesting)
	    					System.out.println("Sleeping...");
	    				Thread.sleep(1000);
	    			}
	    			
	    			//System.out.println("Scheduler makes the next decision.");
	    			d = currentScheduler.makeNextDecision();
	    			//System.out.println("current scheduler made the next decision: " + d);
	    			//cc++;
	    			/** 
	    			 perform some pre exec tasks like configuring operator, calculating statistics 
	    			 and whatnot.  If this method returns false, then we should not run the operator
	    			 for one reason or another (it could have no work to do or there could be
	    			 starvation or something)
	    			 */
	            
	    			if (performPreExecTasks(d)) 
	    			{	    				
	    				operator = d.getNextOperator();                
	    				if (bTesting)
	    					System.out.println("Run op: " + operator);
	    				oldOutput = Integer.parseInt(stats.getValue("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES", operator));
	                
	    				timeBefore = "" + System.currentTimeMillis();
	                
	    				//----------------------
	    				//run the operator
	    				//----------------------
	    				//System.out.println("===== RUNNING: " + operator);
	    				//System.out.println(((RoundRobinScheduler)currentScheduler).getAllOperators());  
	    				//System.out.println("\n Index: " + ((RoundRobinScheduler)currentScheduler).getIndex());
	    				
	    				operator.setRunning(true);
	    				if (bTesting)
	    					System.out.println("Setting operator status to running.");
	                    long startoptime = System.currentTimeMillis();
	                    
	    				if (bTesting)
	    					System.out.println("Executing operator...");
	    					    				
	    				runSuccess = operator.run();
	    				if (bTesting)
	    					System.out.println("Done executing operator...");
	    				 
	    				if (bTesting)
	    					System.out.println("\n*** RESULT OF OPERATOR ["+operator.getObjectID()+"] RUN: [" + runSuccess + "] ***");
	    				
	                    //if(runSuccess) {
	                        //System.out.println("Run: " + operator);
	                        //System.out.println("tuples to dequeue: " + operator.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
	                        //System.out.println("This run takes: " + (System.currentTimeMillis() - startoptime));
	                    //}
	    				// to update the simulated cost information of the symmhashjoinoperator -- Bin @ Feb 22, 2005
	    				updateSimCostofSymmJoin(operator); // non-symmhashjoinoperator query plan could safely remove this
	    				
	    				// to preform local adaptation -- Bin @ Apr 10,2005
	    				// for partitioned symm join query trees
	    				if (operator instanceof SymmMwayHashJoinOperatorImp){
	        				if (localAdapter != null)
	        					localAdapter.localAdapt();    					
	    				}
	    				
	    				// append operator activities
	    				String currentStatus = operator.getCurrentStatus();
	    				if (currentStatus.length()>0) {
	    					XATTree currentTree = (XATTree)this.tree.get(0);
	    					currentTree.appendOperatorActivity(currentStatus);
	    					operator.clearStatus();
	    				}
	    				//System.out.println(currentTree.getOperatorActivity());
	    				//currentTree.clearOperatorActivity();
	                
	    				operator.setRunning(false);
	    				//-------------------------
	    				//Operator running complete
	    				//-------------------------
	    				//System.out.println("===========================");
	                
	    				//System.out.println("------------- Queue Manager Information -------");
	    				//System.out.println("Queues Managed by QM: " + operator.getInputQueue().getQueueManager().getNumQueues());
	    				//System.out.println("Queues: ");
	    				//operator.getInputQueue().getQueueManager().printQueues();
	    				//System.out.println("Total Tuples: " + operator.getInputQueue().getQueueManager().getTotalTuples());
	    				//System.out.println("Total Tuples on Disk: " + operator.getInputQueue().getQueueManager().getTotalOnDisk());
	    				//System.out.println("----------------------------------------------");
	                
	    				timeAfter = "" + System.currentTimeMillis();
	    				timeRun = "" + (Long.parseLong(timeAfter) - Long.parseLong(timeBefore));
	
	    				//now update the stats related to the operator and how it performed
	    				if(runSuccess)
	    				{
	    					
	              	  		updatePostExecStats(operator,timeAfter,timeRun,oldOutput);
	        				if (bTesting)
	        					System.out.println("Done updating statistics.");
	              	  	
	              	  		/* check if it's time to change the query plans */
	              	  		currentTime = System.currentTimeMillis();
	        				if (bTesting)
	        					System.out.println("Checking if adapation needs to be done.");
	              	  	    if(!this.inGlobalAdaptation && (currentTime - lastMigrationTime > local_migration_interval)) 
	              	  	    {
	            				if (bTesting)
	            				{	
		              	  	         System.out.println("Yes, adapation needs to be done.");
		         	  	    		//System.out.println("\n\nQuery executed for " + (currentTime - lastMigrationTime));
		              	  	        System.out.println("\n\nQuery executed for " + (currentTime - startTime));
		          	  	    		System.out.println("local_migration_interval = " + local_migration_interval);
	            				}
		          	  	    		
		          	  	    	lastMigrationTime = System.currentTimeMillis();
	          	  	    		
	          	  	    	
	              	  	    	/* invoke runtime optimizer and migrator if the flag is set to be on */
	              	  	    	if(migrationcount < totalMigrationCount && this.getProperty("MIGRATION_STRATEGY").toLowerCase().compareTo("on") == 0) {
	              	  	    	    
	              	  	    	    migrationcount++;
	              	  	    	    ((XATTree)this.tree.get(0)).setProperty("LAST_LOCAL_MIGRATION_TIME", ""+(lastMigrationTime - startTime));
	              	  	    	    //System.out.print("set tree: " + this.tree.get(0) + " local migration time: " + (lastMigrationTime - startTime) + "\n");
	              	  	    	    
	              	                //sends notification to distributed manager
	              	  	    	    LocalMigrationStarts lms = new LocalMigrationStarts("local migration starts", this.machine.getMachineName());
	              	  	    	    //System.out.println("sStat: " + sStat);
	              	  	    	    //System.out.println("outStat: " + outStat);
	              	                if (sStat == null || sStat.isClosed()) {
	              	                    sStat = new Socket(config.getManagerIPAddress(), 23000);
	              	                    outStat = new ObjectOutputStream(sStat.getOutputStream());
	              	                }
	              	                if(outStat == null)
	              	                    outStat = new ObjectOutputStream(sStat.getOutputStream());
	              	                //System.out.println("sStat: " + sStat);
	            	  	    	    //System.out.println("outStat: " + outStat);
	              	                outStat.writeObject(lms);
	
	              	                //perform the local migration
	              	  				localPlanOptimization();
	
	              	  				//sends another notification to distributed manager
	              	  				LocalMigrationFinishes lmf = new LocalMigrationFinishes("local migration finishes", this.machine.getMachineName());
	              	                if (sStat == null || sStat.isClosed()) {
	              	                    sStat = new Socket(config.getManagerIPAddress(), 23000);
	              	                    outStat = new ObjectOutputStream(sStat.getOutputStream());
	              	                }
	              	                outStat.writeObject(lmf);
	              	                outStat.close();
	              	                sStat.close();
	              	                
	                				if (bTesting)
	                					System.out.println("Local migration takes: " + (System.currentTimeMillis() - lastMigrationTime));
	              	  			}
	              	  		}	
	              	  		
	              	  		if ((operator.getClass().getName().indexOf("Replicat") >= 0
	          	  				|| operator.getClass().getName().indexOf("Spreadfire") >= 0)
	          	  				&& operator.getClass().getName().indexOf("Split") < 0
	          	  				&& operator.getClass().getName().indexOf("Merge") < 0)
	              	  		{
	             	  			lastMemoryUsage = SpreadfireManager.doSpreadfirePostExecution(this, 
	             	  														this.currentScheduler, 
																			operator,
																			systemOverhead, 
																			lastMemoryUsage);
	              	  		}
	              	  		
	              	  		/* check if the distribution manager has started a cross-machine global migration process */
	              	  		//if(System.getProperty("START_GLOBAL_MIGRATION").toLowerCase().compareTo("true") != 0) {
	              	  		//	System.out.println("start global migration on machine" + this.machine.getMachineName());
	              	  		//}
	    				} 
	    				else 
	    				{
	    					//System.out.println("WARNING: OPERATOR DID NOT RUN:  " + operator.toString());
	    				}
	    				//check the status, but don't force a change. 
	    				//This is where we may change scheduling algorithms.
	    				
	    				if (bTesting)
	    					System.out.println("Checking scheduler status.");
	    				checkSchedulerStatus(); 
	    				if (bTesting)
	    					System.out.println("Done. checking scheduler status.");
	    			} 
	//    			else 
	//    			{
	//    				//don't run the operator 
	//    				System.out.println("WARNING: For some reason did not run the operator.");			
	//    			}
	    		}
	    		catch(XATSchedulerException xse)
	    		{ 
	    			//System.out.println("XATSchedulerException: " + xse.getMessage());
	    			// If we have no work to do, we will sleep for a bit
	    			// and then try again. This will primarily happen when we
	    			// are setting up the distribution and nobody is connected yet.
	    			try
	    			{
	    				Thread.sleep(1000);
	    			}
	    			catch (Exception f)
	    			{
	    				f.printStackTrace();     	
	    			}        	
	    		} 
	    		catch (NotSerializableException nse)
	    		{
	    			System.out.println("NotSerializableException: " + nse.getMessage());
	    			FileWriter fw;
	    			PrintWriter p;
	    			try{
	    				fw = new FileWriter("nse-client.txt");
	    				p = new PrintWriter(fw);
	    				System.out.println("---Not Serializable Exception---");
	    				nse.printStackTrace(p);
	    				System.out.println(nse.getMessage());
	    				System.out.println("--------------------------------");
	    			}catch(Exception e){}
	    		} 
	    		catch(OutOfMemoryError oome)
	    		{
	    			System.out.println("Out of memory. Sending whatever statistics we can.");
	    			timeUp = true;//to output the current statistics
	    			Runtime.getRuntime().gc();        	
	    		} 
	    		catch (Exception e) 
	    		{
	    			System.out.println(e.getMessage());
	    			e.printStackTrace();
	    			System.exit(1);
	    		} 	
	    		
	    		//System.out.println("Restarting while loop...");
	    	} // end of while
	    } 
	    catch (Exception ex) 
	    {
	    	System.out.println("Exception: " + ex.getMessage());
			try {
				FileWriter out = new FileWriter("execution_exceptions.txt", true);
				out.write(ex.getStackTrace().toString());
				out.close();
			} 
			catch (IOException e) 
			{
				System.out.println(e);
				e.printStackTrace();
			}
	    }
        
	    /** Added by Yali @ June 10, 2005, for centralized optimization of mjoin. */
	    currentTime = System.currentTimeMillis();
	    System.out.println("The processor has run for " + (currentTime - startTime));
	    // stats.updateValue("NUMBER_OF_TUPLES_IN_STREAM_QUEUES", (XATTree)this.tree.get(0));  
	    // System.out.println("Accumulated tuples in stream queues = " + stats.getValue("NUMBER_OF_TUPLES_IN_STREAM_QUEUES", (XATTree)this.tree.get(0))); 
	  
	    // now begin to process the final operations that need to be performed in each operator before exit -- Bin@Nov.23,2004
	    // please comments out this part if you do not have finalize process in your operator
	    long beforeFinalization = System.currentTimeMillis();
	    /** New version of finalization code, for query tree with multiple symmmwayhashjoin operators -- Bin @ Apr 14, 2005*/
	    //boolean doneFinal = false;
	    //setup the finalization queue, new input data in queues other than finalization queues will be directly discarded
	    //use the flag isFinalized in the XATQueueImp to indicated that
	    if (localAdapter !=null)
	    {
	    	//mark queues that will be used in the finalization stage and the first finalization join operator
	    	localAdapter.setFinalizedQueues();
	    }
    
	    while (localAdapter != null && !doneFinal)
	    {    	
	    	try
	    	{
	    		d = currentScheduler.makeNextDecision();    
	    		performPreExecTasks(d);
	    		operator = d.getNextOperator();                
	    		oldOutput = Integer.parseInt(stats.getValue("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES", operator));
			
	    		operator.setRunning(true);
    		
	    		if (operator instanceof SymmMwayHashJoinOperatorImp) 
	    		{
	    			System.out.println("Finalize operator " + operator);
	    			runSuccess = operator.finalizeOperation();	    		
	    		}
	    		else 
	    		{
	    			System.out.println("Run operator " + operator);
	    			operator.setDebug(true);
	    			runSuccess = operator.run();
	    		}
    						        
	    		operator.setRunning(false);
   
	    		if (localAdapter != null)
	    		{
	    			doneFinal = localAdapter.isCleanUpCompleted();
	    			System.out.println("Local adapter cleaned up:" + doneFinal);
	    		}
	    		
	    		//TODO:  Cease to treat non-adaptive runs of D-CAPE like chopped liver.
	    		//  put in something here to trueify doneFinal where there's no adapter.		
	    	}
	    	catch (XATSchedulerException xse)
	    	{			
	    		System.out.println("XATSchedulerException in Finalization Phase.");		
	    	} 
	    	catch (Exception e) 
	    	{
	    		//System.out.println("Exception in performPreExecTasks of Finalization Phase.");
	    		e.printStackTrace();
	    	}
	    }
    
	    if (bTesting)	
	    	System.out.println("\nClean up all operators");
	    
	    /** old version of finalization code, the new version is provided above -- Bin */
	    XATStreamOperator finalUnion = null;
	    boolean doneFinal = false;
    
	    while (!doneFinal)
	    {
	    	doneFinal = true;
    	
	    	for(int count = 0; count < tree.size(); count++)
	    	{
	    		XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
	    		for (int j = 0; j < nodes.length; j++) 
	    		{
	    			/// since not all operators have the final cleanup stages, then here need to check the operator type to decide whether need to call
	    			/// the final cleanup operations. It could be improved to have all operators having final cleanup operations
	    			/// ideally, this clean up should be performed in order. a short-cut here is to have the union operator postponed to the end
	    			XATStreamOperator curOP = (XATStreamOperator)nodes[j].getOperator();
     	    	
	    			if (curOP.isActive()) 
	    			{
	    				if (curOP instanceof StreamUnionOperatorImp)
	    				{
	    					// postpone the operation
	    					finalUnion = curOP;
	    				} 
	    				else 
	    				{
	    					// set the doneFinal false whenever there are unfinished finalizeOperation()
	    					if (!(nodes[j].getOperator()).finalizeOperation()) 
	    					{
	    						doneFinal = false;	     	 	    	
	    					}     	 	 	    	 	    		
	    				} 	    		
	    			}
	    		}
	    	}
    	
			//    	System.out.println("Clean up union operator");
			//    	if (finalUnion != null) {
			//    		((StreamUnionOperatorImp)finalUnion).setTotalEOFNumer(config.machineCount());
			//    		if (!finalUnion.finalizeOperation()){
			//    			doneFinal = false;	
			//    		}    		
			//    	}

    	
	    	try 
	    	{    		
	    		//System.out.println("ye, garbage collection.");
	    		//Runtime.getRuntime().gc();     	
	    	} 
	    	catch (Exception e)
	    	{    		
	    	}    	
    	
	    	//if all the finalizeOperation return true, the the finalization finishes    
	    } //end of finalization process
		
	    if (bTesting)
	    {	
	    	System.out.println("Finalization Time:" + (System.currentTimeMillis()-beforeFinalization));
	    	System.out.println("Completed Execution, will now  print and send statistics.");
	    	System.out.print("Printing Statistics....");
	    }
	    
	    /** Added by Yali @ June 10, 2005, for centralized optimization of mjoin. */
	    // stats.updateValue("NUMBER_OF_TUPLES_IN_STREAM_QUEUES", (XATTree)this.tree.get(0));  
	    // System.out.println("Accumulated tuples in stream queues = " + stats.getValue("NUMBER_OF_TUPLES_IN_STREAM_QUEUES", (XATTree)this.tree.get(0))); 
    
	    /** 
	    * now the we are done running, update the periodic properties one last time
	    * to ensure the up to date value is printed out
	    */

	    for(int count = 0; count < tree.size(); count++)
	    {
	    	XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
	    	for (int j = 0; j < nodes.length; j++) 
	    	{
	    		for (int i = 0; i < periodicOperatorStatistics.length; i++) 
	    		{
	    			stats.updateValue(periodicOperatorStatistics[i], nodes[j].getOperator());
	    		}
	    	}
	    }
    
	    for (int i = 0; i < periodicTreeStatistics.length; i++) 
	    {
	    	for(int j = 0; j < tree.size();j++)
	    	{
	    		stats.updateValue(periodicTreeStatistics[i], (XATTree) tree.get(j));
	    	}
	    }

	    /** stop the printing because execution is done */
	    printer.interrupt();
    
	    for (Enumeration e = stats.getCategories(); e.hasMoreElements();) 
	    {
	    	Object o = e.nextElement();
	    	if (!new String("GENERIC").equals(o)) 
	    	{
	    		printer.setCategory(o);
	    		printer.gatherStatistics();
	    	}
	    }
    
	    printer.print();
	    if (bTesting)
	    {
	    	System.out.print("done.\n");
	    }
	    String resultFilename = printer.getFileName(false);
    
	    if (bTesting)
	    {
	    	System.out.println("Printing results stats to " + resultFilename);
	    }
   
	    if (debug) 
	    {
	    	for (Enumeration e = stats.getProperties(); e.hasMoreElements();) 
	    	{
	    		String next = (String) e.nextElement();
	    		System.out.println(next + ": " + stats.getProperty(next));
	    	}
	    }
    
	
	    if (bTesting)
	    	System.out.println("Sending result stats to Distribution Manager...");
		
	    try
	    {	
	    	Socket s = new Socket(this.config.getManagerIPAddress(), 22000);

	    	BufferedInputStream in = new BufferedInputStream(new FileInputStream(resultFilename));
	    	BufferedOutputStream out = new BufferedOutputStream(s.getOutputStream());
	
	    	out.write(this.clientName.getBytes());
	    	out.write((int) '@');
	
	    	char ch;
	    	do
	    	{
	    		ch = (char) in.read();
	    		out.write(ch);				
	    	} while((int)ch != 65535);
		
	    	if(outStat != null)			
	    		outStat.close();
		
	    	if(sStat != null)			
	    		sStat.close();
		
	    	out.write(-1);
	    	out.close();
	    	s.close();
	
	    	System.out.print("done.\n");
    
			//Stop the tuple feeder
			//tupleFeeder.requestStop();
			//Stop the tuple sender
			//tupleSender.requestStop();
	    
			//Stop Distributing
			//this.outConnections.clearTable();
			//tupleDistributor.requestStop();    	
	    }
	    catch(Exception e)
	    {
	    	e.printStackTrace();
	    }    
	    return;
    }
	
    //---------------------------------------------------------------------------------------------------
    
    /**
    * called by the constructor.  this will set the default properties
    * that the ALL query objects will support. Each subclass should override
    * this method and add their own unique properties, but should also call
    * super.setDefaultProperties() so that the common properties may be set.
    */   
    
	protected void setDefaultProperties()
	{
		super.setDefaultProperties();
		/** 
		 * this defines how often some of the properties should be updated
		 * some properties, such as operator output rate do not need to be updated
		 * every time an operator is run, but rather, they can be updated periodically
		 * This will define how often (in seconds) these properties should be updated.  
		 *
		 * The tree frequency serves the same purpose, except for tree properties
		 */
		addProperty("UPDATE_OPERATOR_PROPERTY_FREQUENCY","5000");
		addProperty("UPDATE_TREE_PROPERTY_FREQUENCY","5000");
		/** 
		 * similar to the previous property, this defines how often
		 * the ExecController should check the execution status and determine if
		 * the scheduling algorithm should change.  This value should be set sufficently high
		 * such that any strategy has a chance to work (or not work).
		*/		
		addProperty("STATUS_CHECK_FREQUENCY","10000");
		addProperty("STATS_TABLE_STATS","false");
		addProperty("EXECUTION_DEBUG","false");
		//true if you want notification of the next scheduling algorithm and statistics
		//about which algorithm ran to be printed
		addProperty("SCHEDULING_NOTIFICATION_DEBUG","false");		
		/** the default adapter will always stick with the first algorithm */
		addProperty("ADAPTIVE_HEURISTIC","edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.NeverRotateAdapter");
		/** how many times will an operator be run (in a row) without any work before the algorithm is changed */
		addProperty("STARVATION_CHECK","3");
		/** how long, in milleseconds, that the query should be executed */
		addProperty("EXECUTION_DURATION","10000");
		addProperty("NO_WORK_SLEEP_TIME","100");
		addProperty("MIGRATION_STRATEGY", "off");
		addProperty("MIGRATION_INTERVAL", "5000");
		addProperty("TOTAL_MIGRATION_TIMES", "1");
		//-- Bin
		addProperty("LOCAL_PARTITION_ADAPTER", "none");	
	}
	
	//---------------------------------------------------------------------------------------------------

	/**
	 * Insert the method's description here.
	 * Creation date: (1/27/03 1:26:53 PM)
	 * @return java.lang.String
	 */
	
	public String toString() 
	{
		String rs = new String();

		String sName = this.getClass().getName();
	
		//find the last dot.
		int index = 0;
		for(index = sName.length()-1; index >=0; index--)		
			if(sName.charAt(index) == '.')			
				break;
	
		rs = sName.substring(index + 1) + " " + hashCode(); // + "  data:";	
		return rs;
	}
	
	//---------------------------------------------------------------------------------------------------
	
	/**
	 * This will update the statistics for operators. This is broken into 2 phases,
	 * the first phase will update those properties that should be updated every time.  The
	 * rest define those that are updated periodically 
	 * @param operator edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator
	 */

	//write to file
	static boolean DO_PRINT_STATS = true;
	static int STATS_UPDATE_COUNTER = 1;
	//static String  STATS_OUTPUT_FILENAME = "C:\\KEW\\Research\\Eddy\\execution_stats\\runtime_stats.txt";
	static String  STATS_OUTPUT_FILENAME = "C:\\QueryMesh\\experiments\\runtime_stats_"+System.currentTimeMillis()+".txt";

	//---------------------------------------------------------------------------------------------------
	
	public static void WriteToFile(String fileLocation, String stringToWrite, boolean newLine)
	{	
		try
		{
			//Create file 
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			
			if (newLine)
			{
				out.newLine();
			}
			//Close the output stream
			out.close();			    
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}

	//---------------------------------------------------------------------------------------------------
	static double timeLastPosted = System.currentTimeMillis();
	protected void updatePostExecStats(XATStreamOperator operator,String timeAfter, String timeRun, int oldOutput) 
	{
		
		double timeNow = System.currentTimeMillis();
		double timeDuration = timeNow - timeLastPosted;
		
		String opStats = "";
		if (DO_PRINT_STATS)
		{
			// write the header for the statistics only once - the very first time
			if (STATS_UPDATE_COUNTER == 1)
			{
				STATS_UPDATE_COUNTER++;
				String statsTitle = "OPERATOR ID" + "\t";
			    // every time operator stats
				for (int i = 0; i < everyTimeOperatorStatistics.length; i++) 
			    {
					statsTitle += everyTimeOperatorStatistics[i] + "\t";	        
			    }
				//every time system stats
			    for(int i = 0; i < everyTimeSystemStatistics.length; i++)
			    {
			    	statsTitle += everyTimeSystemStatistics[i] + "\t";	    	
			    }
				//periodic operator stats
			    for(int i = 0; i < periodicOperatorStatistics.length; i++)
			    {
			    	statsTitle += periodicOperatorStatistics[i] + "\t";	    	
			    }	
				//periodic tree stats
			    for(int i = 0; i < periodicTreeStatistics.length; i++)
			    {
			    	statsTitle += periodicTreeStatistics[i] + "\t";	    	
			    }		
			    statsTitle = statsTitle.substring(0, statsTitle.length() - 1);
			    statsTitle += "\n";
			    
				WriteToFile(STATS_OUTPUT_FILENAME, statsTitle, true);
			}
		}
	
		if (bTesting)
		{
			System.out.println("----------------------------------------");
			System.out.println("UPDATING POST EXEC STATS FOR [" + operator.toString() + "]");	
			System.out.println("----------------------------------------");
		}
		
		if (DO_PRINT_STATS)
	    {
			opStats += operator.toString() + "\t";		
	    }
		
		try
		{
			while(treeLocked) 
			{
				Thread.sleep(1000);
			}
		}
		catch(Exception e) 
		{
			e.printStackTrace();
		}
		this.treeLocked = true;
	
		stats.updateValue("LAST_TIME_RUN", operator, timeAfter);
		stats.updateValue("TIME_TOOK_TO_RUN", operator, timeRun);
		stats.updateValue("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES", operator);
		stats.updateValue("NUMBER_OF_TUPLES_OUTPUTTED", operator, "" + oldOutput);
			
		if (bTesting)
		{	
			System.out.println("----------------------------------------");
			System.out.println("---  Every time OPERATOR statistics  ---");
			System.out.println("----------------------------------------");
		}
		
		for (int i = 0; i < everyTimeOperatorStatistics.length; i++) 
	    {
	        stats.updateValue(everyTimeOperatorStatistics[i], operator);
	        String statsLine = everyTimeOperatorStatistics[i] +":"+ stats.getValue(everyTimeOperatorStatistics[i], operator);
	        String statsValue = stats.getValue(everyTimeOperatorStatistics[i], operator);
			if (bTesting)
				System.out.println(statsLine);
	        
	        if (DO_PRINT_STATS)
	        {
	        	opStats += statsValue + "\t";        	
	        }
	    }
		
		if (bTesting)
		{	
			System.out.println("----------------------------------------");
			System.out.println("---  Every time SYSTEM statistics  ---");
			System.out.println("----------------------------------------");
		}	
		
		for(int i = 0; i < everyTimeSystemStatistics.length; i++)
	    {
	    	stats.updateValue(everyTimeSystemStatistics[i], system);
	    	String statsValue = stats.getValue(everyTimeSystemStatistics[i], system);
	    	String statsLine = everyTimeSystemStatistics[i] +":"+ stats.getValue(everyTimeSystemStatistics[i], system);
			if (bTesting)
				System.out.println(statsLine);
	    	
	        if (DO_PRINT_STATS)
	        {
	        	opStats += statsValue + "\t";
	        }
	    }
	    
		/**
	     * can't really generalize the next 2 properties
	     *
	     * also update the NUMBER_OF_TUPLES_IN_INPUT_QUEUES property for the operator's
	     * parent(s)
	     */
		
	    try 
	    {
		    XATNode parent = operator.getNode().getParent();
		    if(parent != null)
		    {
			    XATOperator parentOp = parent.getOperator();
		        stats.updateValue("NUMBER_OF_TUPLES_IN_INPUT_QUEUES", parentOp);
	        }
	             
	    } 
	    catch (Exception e) 
	    {
	        //no parent so nothing to worry about 
	    }
   
	    try     
	    {        
	    	/** update the number of tuples in output queues for all the children operators */
	    	for (int i = 0; i < operator.getNode().getChildren().length; i++) 
	    	{
	    		try 
	    		{
	    			if(operator.getNode().getChild(i) instanceof XATNode)
	    			{
	    				XATOperator child = ((XATNode)operator.getNode().getChild(i)).getOperator();
	    				stats.updateValue("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES", child);
	    			}
	    		}
	        	//catch (XATNode.OutOfBoundaryIndexException oobe) {
	        	//	break;
	        	//}        	
	    		catch (Exception e) 
	    		{
	    			break;	    		
	    		}		
	    	}     
	    } 
	    catch 
	    (	    		
	    		edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATNodeDoesNotExistException e) 
	    {	    	
        //no child so nothing to worry about     
	    }

	    /** now determine if we should update the periodic properties for the operator*/
	    long frequency = Long.parseLong(this.getProperty("UPDATE_OPERATOR_PROPERTY_FREQUENCY"));
	    long lastTime = ((Long) lastTimeOperatorUpdated.get(operator)).longValue();
	    long currentTime = System.currentTimeMillis();
    
    
	    /** if it has been long enough, then update these statistics */
	    //if ((currentTime - lastTime) > frequency) 
	    //{
	    	if (bTesting)
	    	{	
		    	System.out.println("----------------------------------------");
		    	System.out.println("---  Periodic  OPERATOR statistics  ---");
		    	System.out.println("----------------------------------------");
	    	}
	    	
		    for (int i = 0; i < periodicOperatorStatistics.length; i++) 
	        {
	            stats.updateValue(periodicOperatorStatistics[i], operator);
	            String statsValue = stats.getValue(periodicOperatorStatistics[i], operator);
	            String statsLine = periodicOperatorStatistics[i] +":"+ stats.getValue(periodicOperatorStatistics[i], operator);
				if (bTesting)
					System.out.println(statsLine);
	            
	            if (DO_PRINT_STATS)
	            {
	            	opStats += statsValue + "\t";
	            }
	        }
	    //}
		    
	    /** update the operator's lastTimeUpdated so we stay consistent */
	    lastTimeOperatorUpdated.put(operator, new Long(currentTime));
	
	    frequency = Long.parseLong(this.getProperty("UPDATE_TREE_PROPERTY_FREQUENCY"));
	    /** determine if we should update the periodic properties for the tree*/
	    //if ((currentTime - lastTimeTreesUpdated) > frequency) 
	    //{
			XATTree[] treeArray = new XATTree[tree.size()];
			for(int i = 0; i < tree.size();i++)
			{
				treeArray[i] = (XATTree) tree.get(i);
			}
			
			if (bTesting)
			{	
		    	System.out.println("----------------------------------------");
		    	System.out.println("---  Periodic TREE statistics  ---");
		    	System.out.println("----------------------------------------");
			}	
			
		    for (int i = 0; i < periodicTreeStatistics.length; i++) 
	        {
	            stats.updateValue(periodicTreeStatistics[i], treeArray);
	            
	            for(int count=0; count < treeArray.length;count++)
	            {
	            	String statsValue = stats.getValue(periodicTreeStatistics[i], (XATQueryObject)treeArray[count]);
	            	String statsLine = periodicTreeStatistics[i] +":"+ stats.getValue(periodicTreeStatistics[i], (XATQueryObject)treeArray[count]); 
					if (bTesting)
						System.out.println(statsLine);            
	                
	                if (DO_PRINT_STATS)
	                {
	                	opStats += statsValue + "\t";
	                }
	        	}         
	        }
	        lastTimeTreesUpdated = System.currentTimeMillis();
	    //}
	    
	    this.treeLocked = false;
	    
	    //add new line in stats file
	    if (DO_PRINT_STATS)
	    {
	    	System.out.println("");
			
	    	//if ((timeDuration >= 60000)) //every 60 seconds (1 minute), write the update to file
			//{
				WriteToFile(STATS_OUTPUT_FILENAME, opStats + "\t" + System.currentTimeMillis(), true);
				timeLastPosted = timeNow;
			//}    	
	    }
	}

	//---------------------------------------------------------------------------------------------------

	/**
	 * Increments the count for the number of times each algorithm was run
	 */
	protected void updateSchedulerCount() 
	{
		//if we havent seen the currentScheduler yet, then add it
		//to the scheduler count
		if(!schedulerCount.containsKey(currentScheduler))
		{
			schedulerCount.put(currentScheduler, new Integer(0));
		}
		int count = ((Integer)(schedulerCount.get(currentScheduler))).intValue() + 1;
		schedulerCount.put(currentScheduler, new Integer(count));	
	}
	
	//---------------------------------------------------------------------------------------------------
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject#getLightWeightQueryObject()
	 */
	public LightWeightQueryObject getLightWeightQueryObject() 
	{
		return new LightWeightQueryObject("ExecutionController",this.getClass().getName(),true);
	}
	
	//---------------------------------------------------------------------------------------------------
	
	/**
	 * Added by Yali Zhu @ Aug. 23, 2005.
	 * Disconnect the part of the query plan that is going to be replaced by the new subplan
	 * This will change the inconnection and outconnection tables of the machines that 
	 * the operators inside the old subplan are connecting to and from.
	 * @param newsubplan
	 * @return ArrayList which keeps all XATNodes outside the subplan been disconnected directly from the oldSubPlan.
	 */
	
	public ArrayList disconnectOldSubPlan(ArrayList oldSubPlanOperators)
	{
		ArrayList disconnectedNodes = new ArrayList();
		XATStreamOperator op;
		
		for(int i=0; i<oldSubPlanOperators.size(); i++) 
		{
			op = (XATStreamOperator)oldSubPlanOperators.get(i);
			this.stats.removeQueryObject(op);
			
			//disconnect all out connections (outputqueues + parentnodes)
			ArrayList parentnodes = op.getNode().getParentArray();
			for (int index = 0; index < parentnodes.size(); index++) 
			{
				// if this parent is outside subplan
				XATNode parentnode = (XATNode) parentnodes.get(index);
				if (parentnode.getStreamOperator().getIsSubPlanOperator() == false) 
				{
					disconnectedNodes.add(parentnode);

					parentnode.deleteChild(op.getNode());
					parentnode.getStreamOperator().removeInputQueue(op.getOutputQueue(index));
					op.getNode().deleteParent(parentnode);
					op.removeOutputQueue(index);
				}
			}
			//update inconnectiontable and outconnectiontable
			this.outConnections.removeConnection(op);
			this.inConnections.removeConnection(Integer.toString(op.getObjectID()));
			
			//disconnect all in connections (intputqueues + childnodes)
			Object[] childnodes = op.getNode().getChildren();		
			for(int index=0; index<childnodes.length; index++) 
			{
				if(childnodes[index] instanceof XATNode) 
				{
					XATNode childnode = (XATNode)childnodes[index];
					if(childnode.getStreamOperator().getIsSubPlanOperator() == false) 
					{
						disconnectedNodes.add(childnode);
						
						childnode.deleteParent(op.getNode());
						childnode.getStreamOperator().removeOutputQueue(op.getInputQueue(index));
						op.getNode().deleteChild(childnode);
						op.removeInputQueue(index);
					}
					// assume a node has only one parent
					// @TODO need to relax later
					this.outConnections.removeConnection(childnode.getStreamOperator());
					this.inConnections.removeConnection(Integer.toString(childnode.getStreamOperator().getObjectID()));
				}
			}
		}
		
		// release the memory of old subplan operators
		oldSubPlanOperators = null;
		
		return disconnectedNodes;
	}
	
	//---------------------------------------------------------------------------------------------------
	
	/**
	 * plug in new subplan
	 *
	 * Added by Yali Zhu @ Aug. 23, 2005.
	 * @param newsubplan
	 */
	
	public void connectNewSubPlan(SubPlan newsubplan, ArrayList disconnectedNodes) 
	{
		ArrayList newops = newsubplan.getOperators(); 
		
		for(int i=0; i<newops.size(); i++) 
		{
			XATStreamOperator op = (XATStreamOperator)newops.get(i);
			this.stats.addNewQueryObject(op);
			
			//reconnet parents
			ArrayList parents = op.getNode().getParentArray();
			for(int j = 0; j<parents.size(); j++)  
			{
				XATNode parentnode = (XATNode)parents.get(j);
				if(parentnode.getStreamOperator().getIsSubPlanOperator() == false) 
				{					
					//find the corresponding parent node in the query tree
					for(int k = 0; k<disconnectedNodes.size(); k++) 
					{
						XATNode dnode = (XATNode)disconnectedNodes.get(k);
						if(dnode.getStreamOperator().getObjectID() == parentnode.getStreamOperator().getObjectID()) 
						{							
							int queueindex = parentnode.getIndexOfChild(op.getNode());
							int opindex = parentnode.getIndexOfOpChild(op.getNode());
							
							op.getNode().deleteParent(parentnode);
							parentnode.deleteChild(op.getNode());
							parentnode.getStreamOperator().removeInputQueue(op.getOutputQueue());
							
							op.getNode().setParent(dnode);
							dnode.addOpChild(op.getNode(), opindex);
							dnode.getStreamOperator().addInputQueue(op.getOutputQueue(), queueindex);
						}	
					}
				}
			}
			
			//reconnet children
			XATNode[] children = op.getNode().getOpChildren();
			for(int j = 0; j < children.length; j++) 
			{
				XATNode childnode = (XATNode)children[j];
				if(childnode.getStreamOperator().getIsSubPlanOperator() == false) 
				{					
					//find the corresponding child node in the query tree
					for(int k=0; k<disconnectedNodes.size(); k++) 
					{
						XATNode dnode = (XATNode)disconnectedNodes.get(k);
						if(dnode.getStreamOperator().getObjectID() == childnode.getStreamOperator().getObjectID()) 
						{							
							int queueindex = op.getNode().getIndexOfChild(childnode);
							int opindex = op.getNode().getIndexOfOpChild(childnode);
							
							op.getNode().deleteChild(childnode);
							childnode.deleteParent(op.getNode());
							childnode.getStreamOperator().removeOutputQueue(op.getNthInputQueue(queueindex));
							
							op.getNode().addOpChild(dnode, opindex);
							dnode.setParent(op.getNode());
							dnode.getStreamOperator().setOutputQueue(op.getNthInputQueue(queueindex));
						}
					}
				}
			}
		}
		
		disconnectedNodes = null;
	}
	
	//---------------------------------------------------------------------------------------------------
	
	public void fillStates(ArrayList oldstates, Hashtable statelocations, ArrayList statesToFill) 
	{
		//now fill the states one by one
		while(statesToFill.size() > 0) 
		{
			OperatorState state = (OperatorState)statesToFill.get(0);
			fillOneState(state, oldstates, statelocations, statesToFill);
			//statesToFill.remove(state);
		}
	}
	
	//---------------------------------------------------------------------------------------------------
	
	private void fillOneState(OperatorState state, ArrayList oldstates, Hashtable statelocations, ArrayList statesToFill) 
	{
		boolean found = false;
		if(state.isMatched()) 
		{
			//if a state exist in oldStates, copy it from local machine
			for(int j=0; j<oldstates.size(); j++) 
			{
				if(((OperatorState)oldstates.get(j)).getStateID() == state.getStateID()) 
				{
					state.copyStateContent((OperatorState)oldstates.get(j));
					statesToFill.remove(state);
					found = true;
					break;
				}
			}
			
			//otherwise it must exist on another machine
			if(!found) 
			{
				Machine remotemachine = (Machine)statelocations.get(Integer.toString(state.getStateID()));
				
				//send request to another machine to ask for that state
				try 
				{
					Socket ms = new Socket(remotemachine.getIPAddress(), remotemachine.getConnectionListenerPort());
					ObjectOutputStream mout = new ObjectOutputStream(ms.getOutputStream());
					SendStateToMachine stm = new SendStateToMachine(this.machine, state.getStateID());

					Packet pk = new Packet("migrationprotocol", stm);
					mout.writeObject(pk);
					mout.close();
					ms.close();
    				if (bTesting)
    					System.out.println("Sending request of state to machine " + remotemachine);

					while (!this.stateReceived) 
					{
						Thread.sleep(200);
					}
				} 
				catch (Exception e) 
				{
					e.printStackTrace();
				}

				this.stateReceived = false;			
			}	
		} 
		else if(!state.isMatched()) 
		{
			ArrayList substateIDs = state.getJoiningSubStates();
			OperatorState[] substates = new OperatorState[substateIDs.size()];
			boolean justfilled = false;
			for(int k=0; k<substateIDs.size(); k++) 
			{
				justfilled = false;
				String substateID = (String)substateIDs.get(k);
				
				//if this substate needs to be filled itself, fill it first
				for(int index = 0; index < statesToFill.size(); index++) 
				{
					OperatorState tofillstate = (OperatorState)statesToFill.get(index);
					if(tofillstate.getStateID() == Integer.parseInt(substateID)) 
					{
						fillOneState(tofillstate, oldstates, statelocations, statesToFill);
						substates[k] = tofillstate;
						justfilled = true;
					} 
				}
				
				//otherwise it must exist in oldstates
				if(!justfilled) 
				{
					for(int c = 0; c < oldstates.size(); c++) 
					{
						OperatorState oldstate = (OperatorState)oldstates.get(c);
						if(oldstate.getStateID() == Integer.parseInt(substateID)) 
						{
							substates[k] = oldstate;
						}
					}
				}
			}
			
			//now recompute this unmatched state by joining all its substates
			//TODO need to complete later
			state.setMatched(true);
			statesToFill.remove(state);
			oldstates.add(state);
		}
	}
	
	//---------------------------------------------------------------------------------------------------
	
	public synchronized void resetLightWeightStats() 
	{
		this.lws = new LightWeightStatistic(this.clientName,lightStats);
	}
	
	//---------------------------------------------------------------------------------------------------
	
	public MigrationKeeper getMigrationkeeper() 
	{
		if(this.migrationkeeper == null) 
		{
			this.migrationkeeper = new MigrationKeeper();
			return this.migrationkeeper;
		}
		return this.migrationkeeper;
	}
	
	//---------------------------------------------------------------------------------------------------
	
	public DistributionConfiguration getConfig()
	{ 
		return config; 		
	}
    
	//---------------------------------------------------------------------------------------------------
	
    public long getSystemOverhead()
    {
        return systemOverhead;
    }
    
    //---------------------------------------------------------------------------------------------------
    
    public static DistributedExecutionController getDistributedExecutionController()
    {
        return _this;
    }

    //---------------------------------------------------------------------------------------------------
    
	public void terminateExecutionNow()
	{
		this.keepGoing = false;
		this.doneFinal = true;
	}
	
	//---------------------------------------------------------------------------------------------------
	
	public boolean sendDuplicatePartitions(ArrayList partitionsToDuplicate, Machine fromM, Machine toM) 
	{
 
		//	    try {
		//	        ((XATTree)this.tree.get(0)).setProperty("LAST_PARTITION_ADAPTATION_TIME", ""+(System.currentTimeMillis() - startTime));
		//	    }catch(Exception e) {
		//	        e.printStackTrace();
		//	    }
	    
	    String senderIPAddr = fromM.getIPAddress();
        String receiverIPAddr = toM.getIPAddress();
        boolean isSender = this.machine.getIPAddress().equals(senderIPAddr);
        WTransitiveHashJoinOperator topjoin = null;
        //ArrayList joinOperators = new ArrayList();

        XATNode[] nodes = ((XATTree) tree.get(0)).convertToArray();
        XATStreamOperator curOp;
        try 
        {
            for (int i = 0; i < nodes.length; i++) 
            {
                curOp = nodes[i].getStreamOperator();
                // (1) if this is the sender machine, then set join operators to
                // start checking the EOF from split operator.
                if (isSender && curOp.isActive() && curOp instanceof WTransitiveHashJoinOperator) 
                {
                    //((WTransitiveHashJoinOperator) curOp).setStartCheckingEOPartitions(true);
//                  joinOperators.add(curOp);                                     
                  ((WTransitiveHashJoinOperator) curOp).setPartitionsToExpire(partitionsToDuplicate);
                    if (curOp.getNode().getParent().getStreamOperator() instanceof StreamUnionOperatorImp) 
                    {
                        topjoin = (WTransitiveHashJoinOperator) curOp;
                        topjoin.setIsTopJoin(true);
                        //topjoin.setPartitionsToExpire(partitionsToDuplicate);
        				if (bTesting)
        					System.out.println("Set as top join operator: " + curOp);
                    }
                }
                // (2) find the activated split operator, remove the partitions
                // from the partition Map, create temporary queue etc.
                if (curOp instanceof StreamSplitOperatorImp && curOp.isActive()) 
                {
                    ((StreamSplitOperatorImp) curOp).sendDuplicatePartitions(partitionsToDuplicate, senderIPAddr, receiverIPAddr);
                    System.out.println("Duplicate partitions " + partitionsToDuplicate + " in " + curOp);
                    System.out.println("Partitions sent to both: " + fromM.getMachineName() + " and " + toM.getMachineName());
                }
            }

            // (3) wait until adaptation is over and send oldPartitionsExpired message to DM
            if (isSender) 
            {
                long ptstarttime = System.currentTimeMillis();
                while (!topjoin.getPTCompletedFlag()) 
                {
                    Thread.sleep(1000);
                    System.out.println("\nCheck isPTCompleted Flag in " + topjoin);
                    System.out.println("System has run for: " + (System.currentTimeMillis() - this.startTime));
                }
                
                ptstarttime = System.currentTimeMillis() - ptstarttime;
                System.out.println("Time takes to expire all old partitions: " + ptstarttime);
                //System.out.println("old partitions in top join: " + topjoin + " has expired!");
                //                for(int index=0; index<joinOperators.size(); index++)
                //                    ((WTransitiveHashJoinOperator)joinOperators.get(index)).setStartCheckingNewTuples(false);

                OldPartitionsExpired dd = new OldPartitionsExpired("expired");
                if (sStat == null || sStat.isClosed()) 
                {
                    sStat = new Socket(config.getManagerIPAddress(), 23000);
                    outStat = new ObjectOutputStream(sStat.getOutputStream());
                }
                outStat.writeObject(dd);
                outStat.close();
                sStat.close();
            }
            
        } 
        catch (Exception e) 
        {
            e.printStackTrace();
        }

        return true;
    }
	
	//---------------------------------------------------------------------------------------------------
	
	public void stopDuplicatePartitions(ArrayList partitions, Machine fromM, Machine toM) 
	{
        String senderIPAddr = fromM.getIPAddress();
        //String receiverIPAddr = toM.getIPAddress();
        boolean isSender = this.machine.getIPAddress().equals(senderIPAddr);
        WTransitiveHashJoinOperator topjoin = null;
        
        XATNode[] nodes = ((XATTree) tree.get(0)).convertToArray();
        XATStreamOperator curOp;
        for (int i = 0; i < nodes.length; i++) 
        {
            curOp = nodes[i].getStreamOperator();
            if(curOp instanceof WTransitiveHashJoinOperator) 
            {            
            	((WTransitiveHashJoinOperator) curOp).setStartCheckingEOPartitions(true);
               if (((WTransitiveHashJoinOperator)curOp).isTopJoinDuringPT()) 
               {
                   topjoin = (WTransitiveHashJoinOperator)curOp;
               }
            }
        }

        for (int i = 0; i < nodes.length; i++) 
        {
            curOp = nodes[i].getStreamOperator();
            if (curOp instanceof StreamSplitOperatorImp && curOp.isActive()) 
            {
                System.out.println("Stop sending duplicate partitions " + partitions + " in " + curOp);
                ((StreamSplitOperatorImp) curOp).stopDuplicatePartitions(senderIPAddr);               
            }
        }
        
        try
        {
            System.out.println("top join is: " + topjoin);
            while(topjoin!= null && topjoin.isTopJoinDuringPT()) 
            {
                Thread.sleep(100);
            }
        }
        catch (Exception e) 
        {
            e.printStackTrace();
        }
	}
	
	//---------------------------------------------------------------------------------------------------

}
