/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
/*
 * Created on Apr 13, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller;

//import edu.wpi.cs.dsrg.xmldb.xat.component.migrationcontroller.*;
//import edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.rewritingrules.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.heuristics.Optimizer;
//import edu.wpi.cs.dsrg.xmldb.xat.component.gui.NonDistributedGUI;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.NotSerializableException;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.net.Socket;
import java.text.DecimalFormat;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATBox;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.DistributionConfiguration;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATExecutionAdapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerException;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;

/**
 * @author Yali Zhu
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class DistributedMigrationExecutionController extends DistributedExecutionController {
	
	private int migrationAlgorithm = -1; // -1: no migration; 0- parallel track; 1- moving state	
	private boolean neverRotateScheduler = true;
	//private RewriteRule rewriter = new PushDownJoinRule(); 
	/**
	 * Using the constructor in DistributedExecutionController.
	 */
	public DistributedMigrationExecutionController(DistributionConfiguration dc, String clientName) throws Exception {
		super(dc, clientName);
		//initExecution();
	}
	/**
	 * Initializes various parts of the execution including
	 * query plan, input, schedulers, and statistics table
	 */
	protected void initExecution() throws Exception {
				
		/** next set up the ExecutionController properties */
		Enumeration e = config.getProperties();
		while(e.hasMoreElements()){
			String name = (String) e.nextElement();
			String value = config.getProperty(name);
			if(this.isValidStatistic(name))
				this.setProperty(name, value);
		}
	
		// Get the machine properties
		machine = config.getMachine(clientName);
		
		Enumeration en = machine.getProperties();
		while(en.hasMoreElements()){
			String name = (String) en.nextElement();
			String property = machine.getProperty(name);
			if(this.isValidPropertyName(name))
				this.setProperty(name, property);
		}
		
		// Get the query plans	
		int c = 0;
		for(Iterator i= config.getQueryPlans(); i.hasNext();){
			QueryPlan qp = (QueryPlan) i.next();
			tree.add(qp.getQueryPlan());
			c++;
		}
	          
		//set up the rest of the class variables
		if(config.getProperty("MIGRATION_ALGORITHM").compareToIgnoreCase("pt") == 0) {
			migrationAlgorithm = 0;
		}
		else if(config.getProperty("MIGRATION_ALGORITHM").compareToIgnoreCase("ms") == 0) {
			migrationAlgorithm = 1;
		}
		debug = (new Boolean(this.getProperty("EXECUTION_DEBUG"))).booleanValue();
		debug_scheduling_notice =  (new Boolean(this.getProperty("SCHEDULING_NOTIFICATION_DEBUG"))).booleanValue();
		lastTimeOperatorUpdated = new Hashtable();
		lastTimeStatusChecked = System.currentTimeMillis();
		lastTimeTreesUpdated = System.currentTimeMillis();
		schedulerCount = new Hashtable();
		
		stats.setProperty("TIMER", this.getProperty("STATS_TABLE_STATS"));	
		
		XATTree[] statsTree = new XATTree[tree.size()];
		for(int i = 0; i < tree.size(); i++){
			statsTree[i] = (XATTree) tree.get(i);
		}
	
	    stats.initStatsTable(statsTree);  
	    stats.addNewQueryObject(system);   
	    
	         	             		
	    if(debug){
		    System.out.println(tree);
		}
		
	    /**
	      init the last time some of its properties were updated
	      set the last updated time to 0, that way the properties will be updated the first time.    
	      This helps with experiments 
	    */
	    
	    for(int count=0; count < tree.size(); count++){
			XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
	    	for (int i = 0; i < nodes.length; i++){
	    		if(nodes[i].getOperator() instanceof XATStreamOperator){
					XATStreamOperator operator = (XATStreamOperator)nodes[i].getOperator();
					lastTimeOperatorUpdated.put(operator, new Long(0));
	    		}
	  		}
	    }
	    
	    //set up schedulers
	    String className = this.getProperty("ADAPTIVE_HEURISTIC"); 
		treeArray = new XATDag[tree.size()];
		for(int i = 0; i < tree.size();i++){
			treeArray[i] = (XATDag) tree.get(i);
		}
	
	    Constructor[] c1 = Class.forName(className).getConstructors();
	    Constructor construct;
	
	    
		for(int i = 0; i < c1.length; i++){
			construct = c1[i];
			Class[] parameters = construct.getParameterTypes();
			System.out.println(parameters[0].getName());
			if(parameters.length == 2){
				if(parameters[0].getName().equals("edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine")){
					adapter = (XATExecutionAdapter)construct.newInstance(new Object[] { machine, treeArray});
				}
				
			}
		}
				
		if(adapter == null){
			throw new Exception("The Adapter could not be instantiated.");
		}
	    //start at scheduler 0 
	    currentScheduler = adapter.getCurrentScheduler();
	    updateSchedulerCount();
	    
		//set up some of the statistics to gather
		initStatsToGather();
	    //set up the printer
	    initStatisticsPrinter();
	
	    //set the starting category in the statistics gatherer to be the first scheduler
		stats.setCurrentCategory(adapter.getCategory());
		printer.setCategory(stats.getCurrentCategory());	
		System.out.println("Initialization...done.");
	}
	/**
	 * This method first checks if the query plan need to be optimized or not. If yes, it then locates the 
	 * part of the query plan that need to be optimized and apply rewriting rules to the sub-plan. 
	 * It then creates the old and new boxes and passes them to the migration controller to actually finish the migration.
	 * For moving state, it's straight-forward, but for parallel states, the old sub-query plan has to be removed
	 * after it is done purging all the old tuples. 
	 *
	 */
	public void optimizeQueryPlan() {
		
		// method to check potential optimization spots, figuring out the sub-query plans.
		// Method to apply rewriting rules to the sub-query plan.
		// For now assume the sub-query plan contains the whole query plan.
		// Need to be modified later.

		//first update the PERFORMANCE_FACOTR property in each existing operators
		
		for (int i = 0; i < tree.size(); i++) {
			XATTree currentTree = (XATTree) tree.get(i);
			long startTime = System.currentTimeMillis();
			XATNode[] oldNodes = currentTree.convertToArray();
			
//			XATTree newTree = (XATTree) currentTree.clone();
//			System.out.println("Tree Clone takes " + (System.currentTimeMillis() - startTime));
				
			// clear tuples in newTree
			XATNode[] newNodes = currentTree.convertToArray();
			for (int index = 0; index < newNodes.length; index++)
				((XATStreamOperator) newNodes[index].getOperator()).clearAllTuples();
				
			System.out.println("Start Optimizing the query plan");
			Optimizer optimizer = new Optimizer();
				
			// if the tree is changed, return true, otherwise return false
			//if(optimizer.ApplyHeuristics()) {
				
				/** setup oldbox */
				startTime = System.currentTimeMillis();
				/** create a new adaptor due to the change to the tree. */
				try {
					String className = this.getProperty("ADAPTIVE_HEURISTIC");
					if (className.compareTo("edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.NeverRotateAdapter") == 0)
						neverRotateScheduler = true;
					
					//start at scheduler 0 
					currentScheduler = adapter.getCurrentScheduler();
					currentScheduler.setTrees((XATTree[])tree.toArray(new XATTree[] {}));
					for (int j = 0; j < oldNodes.length; j++) {
						XATStreamOperator oldOp = (XATStreamOperator) oldNodes[j].getOperator();
						stats.removeQueryObject(oldNodes[j].getOperator());
						lastTimeOperatorUpdated.remove(oldNodes[j].getOperator());
						Iterator oldQueues = oldOp.getInputQueues();
						while (oldQueues.hasNext()) {
							stats.addNewQueryObject((XATQueue) oldQueues.next());
						}
					}
					newNodes = currentTree.convertToArray();
					for (int k = 0; k < newNodes.length; k++) {
						XATStreamOperator newOp = (XATStreamOperator) newNodes[k].getOperator();
						stats.addNewQueryObject(newOp);
						lastTimeOperatorUpdated.put(newNodes[k].getOperator(), new Long(0));
						Iterator newQueues = newOp.getInputQueues();
						while (newQueues.hasNext()) {
							stats.addNewQueryObject((XATQueue) newQueues.next());
						}
					}
					
					
//					if(gui instanceof NonDistributedGUI) {
//						((NonDistributedGUI)gui).redrawTree();
//					}
				} catch (Exception e) {
					System.out.println(
						"Exception caught during query plan optimization: "
							+ e);
				}
//			} else
//				System.out.println("The current query plan is not changed.");
		}
	}
	
	/**
	 * Runs the actual execution
	 *
	 */
	public void run() {
		System.out.println("Starting a new Distributed Execution Controller.");
		//Initially try to run the thread...
	    try {
	    } catch (Exception e) {
	        e.printStackTrace();
	        return;
	    }    
	    // Here is where we start execution.
	    boolean keepGoing = true;
	    XATSchedulerDecision d = null;
		long optimizeInterval = Long.parseLong(config.getProperty("OPTIMIZATION_INTERVAL"));
	    int totalTuples;
	    int memCount = 0;
	    boolean notInterrupted = true;
	    long startTime = System.currentTimeMillis();
	    long currentTime = System.currentTimeMillis();
	    boolean timeUp = false;
	    boolean runSuccess = false;
	    long lastRedistTime = currentTime;
		long lastTimeOptimized = startTime;
	    Socket sStat = null;
		ObjectOutputStream outStat = null;
		DecimalFormat df = new DecimalFormat("0.00#");
		XATStreamOperator operator;
		int oldOutput;
		String timeBefore;
		int reDistTime = Integer.parseInt(config.getProperty("REDISBRIBUTION_TIME"));
		int runTime = Integer.parseInt(config.getProperty("EXECUTION_DURATION"));
		Runtime rt = Runtime.getRuntime();
		String timeAfter;
		String timeRun;
		LightWeightStatistic lws;
		rt.gc();
		
		//set up the rest of the class variables
		if(config.getProperty("MIGRATION_ALGORITHM").compareToIgnoreCase("pt") == 0) {
			migrationAlgorithm = 0;
		}
		else if(config.getProperty("MIGRATION_ALGORITHM").compareToIgnoreCase("ms") == 0) {
			migrationAlgorithm = 1;
		}
		
	    try{
	    	Thread.sleep(2000);
		    sStat = new Socket(config.getManagerIPAddress(),23000);
			
	    }catch(Exception e){
		    e.printStackTrace();
	    }
	    //Find out how long to run the machine.
	
	    System.out.println("Will process data for " + runTime + " milliseconds.");
	    while (keepGoing && !timeUp) {
	        try {
				currentTime = System.currentTimeMillis();
				if(currentTime - startTime > runTime){
					timeUp = true;
				}
				
				if(currentTime - lastRedistTime > reDistTime){
					if(outStat == null){
						outStat = new ObjectOutputStream(sStat.getOutputStream());
					}
	
					lastRedistTime = currentTime;
					if(lightStats != null){
						lws = new LightWeightStatistic(this.clientName,lightStats);
						outStat.writeObject(lws);
					}
					
					System.out.println("-------------Memory Information " + memCount++ +" --------------");
					System.out.println("Tuples in input pool: " + this.inPool.size());
					System.out.println(this.stats.displaySize());
					System.out.println("Free memory: " + df.format(((double)rt.freeMemory()/1024.0/1024.0)) + "MB");
					System.out.println("Total memory: " + df.format(((double)rt.totalMemory()/1024.0/1024.0)) + "MB");
					System.out.println("Max memory: " + df.format(((double)rt.maxMemory()/1024.0/1024.0)) + "MB");
					System.gc();
					System.out.println("Free memory: " + df.format(((double)rt.freeMemory()/1024.0/1024.0)) + "MB");
					System.out.println("Total memory: " + df.format(((double)rt.totalMemory()/1024.0/1024.0)) + "MB");
					System.out.println("Max memory: " + df.format(((double)rt.maxMemory()/1024.0/1024.0)) + "MB");	
					System.out.println("----------------------------------------------");
				}
	            /** have the scheduler make the next decision */
	            d = currentScheduler.makeNextDecision();
	        	
	            /** 
	              perform some pre exec tasks like configuring operator, calculating statistics 
	              and whatnot.  If this method returns false, then we should not run the operator
	              for one reason or another (it could have no work to do or there could be
	              starvation or something)
	            */
	            if (performPreExecTasks(d)) {
	                operator = d.getNextOperator();                
	                oldOutput =
	                    Integer.parseInt(stats.getValue("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES", operator));
	                
	                timeBefore = "" + System.currentTimeMillis();
	                
	                //run the operator
	                //System.out.println("===== RUNNING " + operator);
	              
	                operator.setRunning(true);
	             
	                runSuccess = operator.run();
				
	                operator.setRunning(false);
	                //System.out.println("===========================");
	                
	                timeAfter = "" + System.currentTimeMillis();
	                timeRun = "" + (Long.parseLong(timeAfter) - Long.parseLong(timeBefore));
	
	                //now update the stats related to the operator and how it performed
	                if(runSuccess){
	              	  updatePostExecStats(operator,timeAfter,timeRun,oldOutput);
					  //System.out.println("Run "+operator);
	                }
	                //check the status, but don't force a change. 
	                //This is where we may change scheduling algorithms.
	                checkSchedulerStatus(); 
	            } else {
	                //don't run the operator 
	            }
				//check if it's time to try to optimize
				if((migrationAlgorithm != -1) && (System.currentTimeMillis() - lastTimeOptimized) > optimizeInterval) {
					System.out.println("\nExec has run for "+(System.currentTimeMillis() - lastTimeOptimized));
					System.out.println("Now try to optimize query plan.");
					lastTimeOptimized = System.currentTimeMillis();
					long opstarttime = lastTimeOptimized;
					optimizeQueryPlan();
					System.out.println("Optimization takes " + (System.currentTimeMillis() - opstarttime));
				}
	        }catch(XATSchedulerException xse){ 
	        	// If we have no work to do, we will sleep for a bit
	        	// and then try again. This will primarily happen when we
	        	// are setting up the distribution and nobody is connected yet.
	        	try{
					Thread.sleep(1000);
	        	}   
	        	catch (Exception f){
	        		f.printStackTrace();     	
	        	}
	        	
	        }
	        catch (NotSerializableException nse){
	        	FileWriter fw;
	        	PrintWriter p;
	        	try{
				fw = new FileWriter("nse-client.txt");
	        	p = new PrintWriter(fw);
	        	System.out.println("---Not Serializable Exception---");
	        	nse.printStackTrace(p);
				System.out.println(nse.getMessage());
	        	System.out.println("--------------------------------");
				}catch(Exception e){}
	        }
	        catch(OutOfMemoryError oome){
	        	System.out.println("Out of memory. Sending whatever statistics we can.");
	        	Runtime.getRuntime().gc();
	        	
	        }
	        catch (Exception e) {
	        	System.out.println(e.getMessage());
	        	e.printStackTrace();
	        	System.exit(1);
	        } 
	    } // end of while
	   
	    System.out.println("Completed Execution, will now  print and send statistics.");
	    System.out.print("Printing Statistics....");
	 
	    /** 
	    * now the we are done running, update the periodic properties one last time
	    * to ensure the up to date value is printed out
	    */
		for(int count=0; count < tree.size(); count++){
			XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
	 	    for (int j = 0; j < nodes.length; j++) {
	     	    for (int i = 0; i < periodicOperatorStatistics.length; i++) {
	            stats.updateValue(periodicOperatorStatistics[i], nodes[j].getOperator());
	       		}
	   		}
		}
	    for (int i = 0; i < periodicTreeStatistics.length; i++) {
	    	for(int j=0; j < tree.size();j++){
				stats.updateValue(periodicTreeStatistics[i], (XATTree) tree.get(j));
	    	}
	    }
	
	    /** stop the printing because execution is done */
	    printer.interrupt();
	    for (Enumeration e = stats.getCategories(); e.hasMoreElements();) {
	        Object o = e.nextElement();
	        if (!new String("GENERIC").equals(o)) {
	            printer.setCategory(o);
	            printer.gatherStatistics();
	        }
	    }
	    
	    printer.print();
	    System.out.print("done.\n");
	    String resultFilename = printer.getFileName(false);
	    
	    System.out.println("Printing results stats to " + resultFilename);
	    if (debug) {
	        for (Enumeration e = stats.getProperties(); e.hasMoreElements();) {
	            String next = (String) e.nextElement();
	            System.out.println(next + ": " + stats.getProperty(next));
	        }
	    }
	    
		System.out.println("Sending result stats to Distribution Manager...");
		
		try{
		
		Socket s = new Socket(this.config.getManagerIPAddress(),22000);
	
		BufferedInputStream in = new BufferedInputStream(new FileInputStream(resultFilename));
		BufferedOutputStream out = new BufferedOutputStream(s.getOutputStream());
		
		out.write(this.clientName.getBytes());
		out.write((int) '@');
		
		char ch;
		do{
			ch = (char) in.read();
			out.write(ch);
			
		}while((int)ch != 65535);
		outStat.close();
		sStat.close();
		out.write(-1);
		out.close();
		s.close();
		
	    System.out.print("done.\n");
	    
	    //Stop the tuple feeder
	    //tupleFeeder.requestStop();
	    //Stop the tuple sender
	    //tupleSender.requestStop();
	    
	    //Stop Distributing
	    //this.outConnections.clearTable();
	    //tupleDistributor.requestStop();
	    
		}catch(Exception e){
			e.printStackTrace();
		}
	    return;
	}

}
