/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller;
import java.util.HashMap;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.NotSerializableException;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.net.Socket;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.realjoinoperator.WTransitiveHashJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WEquiJoinStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamUnionOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.DistributionConfiguration;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ConnectionListener;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ConnectionTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.MigrationKeeper;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.SubPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendStateToMachine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.PacketPool;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleDistributor;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleFeeder;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleReceiver;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleSender;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.manager.SpreadfireManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.FireStreamGUI;
import edu.wpi.cs.dsrg.STeM.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUIQueryEngine;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATExecutionAdapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerException;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsPrinter;
import edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.heuristics.Optimizer;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation.Adapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation.LocalMemoryAdapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.adprotocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.UpdatePartitionStatistics;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.*;

import edu.wpi.cs.dsrg.eddy.datastructure.EddyQueue;
import edu.wpi.cs.dsrg.eddy.operator.*;
import edu.wpi.cs.dsrg.STeM.STeMData;

public class EddyExecutionController extends DistributedExecutionController 
{
	
	public EddyExecutionController(DistributionConfiguration dc,String clientName) throws Exception 
	{
		super(dc, clientName);	
	}

	/**
	 * Override the Initialization
	 */
	protected void initExecution() throws Exception 
	{				
		/* init query plan optimizer */
		optimizer = new Optimizer();
				
		/** next set up the ExecutionController properties */
		Enumeration e = config.getProperties();
		System.out.println("config file is: " + config.toString());
		while(e.hasMoreElements())
		{
			String name = (String) e.nextElement();
			String value = config.getProperty(name);
			if(this.isValidStatistic(name))
			{
				this.setProperty(name, value);
			}
			
		}

			//set up distribution information
			//System.out.println("Outside while loop");
			distInfo = new String[4];
			distInfo[0] = "Initial distribution algorithm: " + config.getProperty("DISTRIBUTION_PATTERN").substring(config.getProperty("DISTRIBUTION_PATTERN").lastIndexOf(".") + 1);
			distInfo[1] = "Redistribution algorithm: " + config.getProperty("REDISTRIBUTION_POLICY").substring(config.getProperty("REDISTRIBUTION_POLICY").lastIndexOf(".") + 1);
			distInfo[2] = "Redistribution interval: " + (Integer.parseInt(config.getProperty("REDISBRIBUTION_TIME")) / 1000);
			distInfo[3] = "Redistribution percentage: " + config.getProperty("REDISTRIBUTION_PERCENT") + "%";
				
			// Get the machine properties
			machine = config.getMachine(clientName);
			Enumeration en = machine.getProperties();
			while(en.hasMoreElements()){
				String name = (String) en.nextElement();
				String property = machine.getProperty(name);
				if(this.isValidPropertyName(name))
					this.setProperty(name, property);
			}
		
		// Get the query plans	
		int c = 0;
		int numQueryPlans = 0;
				for(Iterator i= config.getQueryPlans(); i.hasNext();){
					QueryPlan qp = (QueryPlan) i.next();
					System.out.println("queryplan to add is: " + qp.getProperty("QUERY_ID"));
					tree.add(qp.getQueryPlan());
					c++;
					numQueryPlans++;
				}
				System.out.println("tree size is: " + tree.size());    
		
				debug = (new Boolean(this.getProperty("EXECUTION_DEBUG"))).booleanValue();
				debug_scheduling_notice =  (new Boolean(this.getProperty("SCHEDULING_NOTIFICATION_DEBUG"))).booleanValue();
				lastTimeOperatorUpdated = new Hashtable();
				lastTimeStatusChecked = System.currentTimeMillis();
				lastTimeTreesUpdated = System.currentTimeMillis();
				schedulerCount = new Hashtable();
				
				stats.setProperty("TIMER", this.getProperty("STATS_TABLE_STATS"));	
				
				XATDag[] statsTree = new XATDag[tree.size()];
				for(int i = 0; i < tree.size(); i++){
					statsTree[i] = (XATDag) tree.get(i);
				}

			    stats.initStatsTable(statsTree);  
			    stats.addNewQueryObject(system);       
			         	             		
			    if(debug){
				    System.out.println(tree);
				}
				
			    /**
			      init the last time some of its properties were updated
			      set the last updated time to 0, that way the properties will be updated the first time.    
			      This helps with experiments 
			    */
			    
			    for(int count=0; count < tree.size(); count++){
					XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
			    	for (int i = 0; i < nodes.length; i++){
			    		if(nodes[i].getOperator() instanceof XATStreamOperator){
							XATStreamOperator operator = (XATStreamOperator)nodes[i].getOperator();
							lastTimeOperatorUpdated.put(operator, new Long(0));
			    		}
			  		}
			    }
			    
			    //set up schedulers
			    String className = this.getProperty("ADAPTIVE_HEURISTIC"); 
				treeArray = new XATDag[tree.size()];
				for(int i = 0; i < tree.size();i++){
					treeArray[i] = (XATDag) tree.get(i);
				}

			    Constructor[] c1 = Class.forName(className).getConstructors();
			    Constructor construct;
			    
				for(int i = 0; i < c1.length; i++){
					construct = c1[i];
					Class[] parameters = construct.getParameterTypes();
					System.out.println(parameters[0].getName());
					if(parameters.length == 2){
						if(parameters[0].getName().equals("edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine"))
						{
							adapter = (XATExecutionAdapter)construct.newInstance(new Object[] { machine, treeArray});
						}			
					}
				}
						
				if(adapter == null){
					throw new Exception("The Adapter could not be instantiated.");
				}
			    //start at scheduler 0 
			    currentScheduler = adapter.getCurrentScheduler();
			    
			    updateSchedulerCount();
			    
				//set up some of the statistics to gather
				initStatsToGather();
			    //set up the printer
			    initStatisticsPrinter();

			    //set the starting category in the statistics gatherer to be the first scheduler
				stats.setCurrentCategory(adapter.getCategory());
				printer.setCategory(stats.getCurrentCategory());	
				
				//set up local partition adapter -- Bin @ Dec.10, 2004
				className = this.getProperty("LOCAL_PARTITION_ADAPTER");
				if (!className.equals("none")) {
					c1 = Class.forName(className).getConstructors();
					construct = c1[0]; // might go wrong if multiple constructor exists
					//localAdapter = (Adapter)construct.newInstance(null);		
					localAdapter = (Adapter)construct.newInstance(new Object[]{tree});		
				}
				//set the distributed execution controller reference to the symmmwayhashjoinoperator
				//--Bin @ Apr 12, 2005, to handle the tracing of cost information in different machines
			    //this part of the code could be safely commented when run other queries/without of tracing cost statisitics
				for(int count=0; count < tree.size(); count++){
					XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
			    	for (int i = 0; i < nodes.length; i++){
			    		if(nodes[i].getOperator() instanceof SymmMwayHashJoinOperatorImp){
							//((SymmMwayHashJoinOperatorImp)nodes[i].getOperator()).setDistributedExectionControllerRef(this);
			    		}
			  		}
			    }

//		//Start KWORKS - Eddy
//		// Tuples that require processing
//		HashMap<Integer, EddyQueue> TuplePool;
//		TuplePool = new HashMap<Integer, EddyQueue>();
//
//		
//
//		for (int i = 0; i < nodes.length; i++){
//    		if(nodes[i].getOperator() instanceof EddyOperatorImp){
//    			((EddyOperatorImp)nodes[i].getOperator()).SetTuplePool(TuplePool);
//    		}
//    		else if(nodes[i].getOperator() instanceof EddyStreamSelectOperatorImp){
//    			((EddyStreamSelectOperatorImp)nodes[i].getOperator()).SetTuplePool(TuplePool);
//    		}
//  	}

		XATNode[] nodes = ((XATTree)tree.get(0)).convertToArray();
		// Create STEM data structures required for processing
				
			ArrayList<Integer> STeMStreamIds;
			HashMap<Integer,ArrayList<ArrayList<Integer>>> STeMHashKeys;

			STeMHashKeys = new HashMap(); //HashKeys
			STeMStreamIds = new ArrayList(); //STREAM Ids
			ArrayList<BitSet> PreFilledPerStream=null;
			HashMap <Integer,Integer> STREAMWindowCount = null; 
			boolean bHasSTeM = false;
	
			nodes = ((XATTree)tree.get(0)).convertToArray();
			    
				for (int i = 0; i < nodes.length; i++){
			    		if(nodes[i].getOperator() instanceof STeMJoinProbeOperatorImp){
			    			bHasSTeM = true;
			    			//look up data
			    			ArrayList<ArrayList<Integer>> indexes = null;
			    			if (STeMHashKeys.get(((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId()) == null) {
			    				indexes = new ArrayList();
			    				indexes.add(0,new ArrayList<Integer>());
			    				indexes.set(0,((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getNumSTeMIndex());
			    				STeMHashKeys.put(((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId(),indexes);
			    				
			    				STeMStreamIds.add((((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId()));
			    				((STeMJoinProbeOperatorImp)nodes[i].getOperator()).SetSTeMDataIndexArrayPos(STeMStreamIds.size()-1);
			    			}
			    			else
			    			{
			    				indexes = STeMHashKeys.get(((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId());
			    				indexes.add(new ArrayList<Integer>());

			    				indexes.set(indexes.size()-1,((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getNumSTeMIndex());			    				
			    			}
			    			// set index number for operator
			    			((STeMJoinProbeOperatorImp)nodes[i].getOperator()).SetSTeMDataIndexArrayPos(indexes.size()-1);

			    		}
			    		else if(nodes[i].getOperator() instanceof EddyOperatorImp){
			    			((EddyOperatorImp)nodes[i].getOperator()).InitializeVariables();
			    			PreFilledPerStream = ((EddyOperatorImp)nodes[i].getOperator()).getPreFilledPerStream();
			    			STREAMWindowCount= ((EddyOperatorImp)nodes[i].getOperator()).getSTREAMWindowCount();

			    		}

			  	}

				
				HashMap<Integer,STeMData> STeMStreamData = null;


				if (STeMStreamIds.size()> 0)
				{	
					STeMStreamData = new HashMap(STeMStreamIds.size());
					for (int k = 0;k<STeMStreamIds.size();k++)
					{
						bHasSTeM = true;
						STeMData sd = new STeMData();
						sd.init(STeMStreamIds.get(k),STeMHashKeys.get(STeMStreamIds.get(k)));
						STeMStreamData.put(STeMStreamIds.get(k),sd);
						
						//set window size??
						if (STREAMWindowCount != null)
						{
							int windowSize = 0;
							if (STREAMWindowCount.get(STeMStreamIds.get(k)) != null)
							{
								windowSize =STREAMWindowCount.get(STeMStreamIds.get(k));
								sd.SetMaxWindowSize(windowSize);
							}							
						}
						
						
					}
				}


				
				//Create Eddy DataStructures required for processing
							    
				for (int i = 0; i < nodes.length; i++){
			    		if(nodes[i].getOperator() instanceof EddyOperatorImp){
			    			if (bHasSTeM == true)
							  ((EddyOperatorImp)nodes[i].getOperator()).setSTeMStreamData(STeMStreamData);
			    		}
			    		else if(nodes[i].getOperator() instanceof STeMJoinProbeOperatorImp){
			    			//probe
			    			STeMData sd = STeMStreamData.get(((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId());
			    			if (sd != null){
			    			  ((STeMJoinProbeOperatorImp)nodes[i].getOperator()).setSTeMStreamData(sd);
			    			}
			    			((STeMJoinProbeOperatorImp)nodes[i].getOperator()).setPreFilledPerStream(PreFilledPerStream);
			    			
			    		}
			  	}
				//End KWORKS - Eddy
				
				System.out.println("Initialization...done.");
			}

			


	}

