/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller;
import java.util.HashMap;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.NotSerializableException;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.net.Socket;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import org.w3c.dom.Node;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.realjoinoperator.WTransitiveHashJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WEquiJoinStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamUnionOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.DistributionConfiguration;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ConnectionListener;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ConnectionTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.MigrationKeeper;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.SubPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendStateToMachine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.PacketPool;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleDistributor;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleFeeder;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleReceiver;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleSender;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.manager.SpreadfireManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.FireStreamGUI;
import edu.wpi.cs.dsrg.STeM.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUIQueryEngine;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATExecutionAdapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerException;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsPrinter;
import edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.heuristics.Optimizer;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation.Adapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation.LocalMemoryAdapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.adprotocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.UpdatePartitionStatistics;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.*;

import edu.wpi.cs.dsrg.eddyBatch.datastructure.EddyQueue;
import edu.wpi.cs.dsrg.eddyBatch.operator.*;
import edu.wpi.cs.dsrg.STeM.Partition.STeMPartitionJoinProbeOperatorImp;
import edu.wpi.cs.dsrg.STeM.Partition.STeMMultiStreamPartitionJoinProbeOperatorImp;
import edu.wpi.cs.dsrg.STeM.Partition.STeMPartitionData;


public class EddyPartitionExecutionControllerOld extends DistributedExecutionController 
{
	
	public EddyPartitionExecutionControllerOld(DistributionConfiguration dc,String clientName) throws Exception 
	{
		super(dc, clientName);	
	}

	/**
	 * Override the Initialization
	 */
	protected void initExecution() throws Exception 
	{				
		/* init query plan optimizer */
		optimizer = new Optimizer();
				
		/** next set up the ExecutionController properties */
		Enumeration e = config.getProperties();
		System.out.println("config file is: " + config.toString());
		while(e.hasMoreElements())
		{
			String name = (String) e.nextElement();
			String value = config.getProperty(name);
			if(this.isValidStatistic(name))
			{
				this.setProperty(name, value);
			}
			
		}

			//set up distribution information
			//System.out.println("Outside while loop");
			distInfo = new String[4];
			distInfo[0] = "Initial distribution algorithm: " + config.getProperty("DISTRIBUTION_PATTERN").substring(config.getProperty("DISTRIBUTION_PATTERN").lastIndexOf(".") + 1);
			distInfo[1] = "Redistribution algorithm: " + config.getProperty("REDISTRIBUTION_POLICY").substring(config.getProperty("REDISTRIBUTION_POLICY").lastIndexOf(".") + 1);
			distInfo[2] = "Redistribution interval: " + (Integer.parseInt(config.getProperty("REDISBRIBUTION_TIME")) / 1000);
			distInfo[3] = "Redistribution percentage: " + config.getProperty("REDISTRIBUTION_PERCENT") + "%";
				
			// Get the machine properties
			machine = config.getMachine(clientName);
			Enumeration en = machine.getProperties();
			while(en.hasMoreElements()){
				String name = (String) en.nextElement();
				String property = machine.getProperty(name);
				if(this.isValidPropertyName(name))
					this.setProperty(name, property);
			}
		
		// Get the query plans	
		int c = 0;
		int numQueryPlans = 0;
				for(Iterator i= config.getQueryPlans(); i.hasNext();){
					QueryPlan qp = (QueryPlan) i.next();
					System.out.println("queryplan to add is: " + qp.getProperty("QUERY_ID"));
					tree.add(qp.getQueryPlan());
					c++;
					numQueryPlans++;
				}
				System.out.println("tree size is: " + tree.size());    
		
				debug = (new Boolean(this.getProperty("EXECUTION_DEBUG"))).booleanValue();
				debug_scheduling_notice =  (new Boolean(this.getProperty("SCHEDULING_NOTIFICATION_DEBUG"))).booleanValue();
				lastTimeOperatorUpdated = new Hashtable();
				lastTimeStatusChecked = System.currentTimeMillis();
				lastTimeTreesUpdated = System.currentTimeMillis();
				schedulerCount = new Hashtable();
				
				stats.setProperty("TIMER", this.getProperty("STATS_TABLE_STATS"));	
				
				XATDag[] statsTree = new XATDag[tree.size()];
				for(int i = 0; i < tree.size(); i++){
					statsTree[i] = (XATDag) tree.get(i);
				}

			    stats.initStatsTable(statsTree);  
			    stats.addNewQueryObject(system);       
			         	             		
			    if(debug){
				    System.out.println(tree);
				}
				
			    /**
			      init the last time some of its properties were updated
			      set the last updated time to 0, that way the properties will be updated the first time.    
			      This helps with experiments 
			    */
			    
			    for(int count=0; count < tree.size(); count++){
					XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
			    	for (int i = 0; i < nodes.length; i++){
			    		if(nodes[i].getOperator() instanceof XATStreamOperator){
							XATStreamOperator operator = (XATStreamOperator)nodes[i].getOperator();
							lastTimeOperatorUpdated.put(operator, new Long(0));
			    		}
			  		}
			    }
			    
			    //set up schedulers
			    String className = this.getProperty("ADAPTIVE_HEURISTIC"); 
				treeArray = new XATDag[tree.size()];
				for(int i = 0; i < tree.size();i++){
					treeArray[i] = (XATDag) tree.get(i);
				}

			    Constructor[] c1 = Class.forName(className).getConstructors();
			    Constructor construct;
			    
				for(int i = 0; i < c1.length; i++){
					construct = c1[i];
					Class[] parameters = construct.getParameterTypes();
					System.out.println(parameters[0].getName());
					if(parameters.length == 2){
						if(parameters[0].getName().equals("edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine"))
						{
							adapter = (XATExecutionAdapter)construct.newInstance(new Object[] { machine, treeArray});
						}			
					}
				}
						
				if(adapter == null){
					throw new Exception("The Adapter could not be instantiated.");
				}
			    //start at scheduler 0 
			    currentScheduler = adapter.getCurrentScheduler();
			    
			    updateSchedulerCount();
			    
				//set up some of the statistics to gather
				initStatsToGather();
			    //set up the printer
			    initStatisticsPrinter();

			    //set the starting category in the statistics gatherer to be the first scheduler
				stats.setCurrentCategory(adapter.getCategory());
				printer.setCategory(stats.getCurrentCategory());	
				
				//set up local partition adapter -- Bin @ Dec.10, 2004
				className = this.getProperty("LOCAL_PARTITION_ADAPTER");
				if (!className.equals("none")) {
					c1 = Class.forName(className).getConstructors();
					construct = c1[0]; // might go wrong if multiple constructor exists
					//localAdapter = (Adapter)construct.newInstance(null);		
					localAdapter = (Adapter)construct.newInstance(new Object[]{tree});		
				}
				//set the distributed execution controller reference to the symmmwayhashjoinoperator
				//--Bin @ Apr 12, 2005, to handle the tracing of cost information in different machines
			    //this part of the code could be safely commented when run other queries/without of tracing cost statisitics
				for(int count=0; count < tree.size(); count++){
					XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
			    	for (int i = 0; i < nodes.length; i++){
			    		if(nodes[i].getOperator() instanceof SymmMwayHashJoinOperatorImp){
							//((SymmMwayHashJoinOperatorImp)nodes[i].getOperator()).setDistributedExectionControllerRef(this);
			    		}
			  		}
			    }


//		//Start KWORKS - Eddy

		XATNode[] nodes = ((XATTree)tree.get(0)).convertToArray();
		// Create STEM data structures required for processing
				
			ArrayList<Integer> STeMStreamIds;
			HashMap<Integer,ArrayList<ArrayList<Integer>>> STeMHashKeys;

			STeMHashKeys = new HashMap(); //HashKeys
			STeMStreamIds = new ArrayList(); //STREAM Ids
			
			ArrayList<BitSet> PreFilledPerStream=null;
			HashMap <Integer,Integer> STREAMWindowCount = null;
			
			//partition window size
			int PartitionSize = 0;
			int PartitionWindowSize = 0;
			
			//using IMP index or partitions
			boolean bIMPIndex = false;
			boolean bPartitioned = false;
			
			boolean bHasSTeM = false;
	
			//number of limited indices
			int nNumLimitedIndices = 0;

			//sampling?
			boolean bNoSampling = false;
			boolean bSimpleRandomSampling = false;
			int nSimpleRandomPercent = 0;
			boolean bStratifiedRandomSampling = false;
			int nStratifiedRandomPercent = 0;
			int nStratifiedRandomWindow = 0;
			boolean bSystematicSampling = false;
			int nSystematicInterval = 0;
			
			//index migration?
			boolean bNoMigration=false;
			boolean bSymmetricMigration=false;
			boolean bPartitionMigration=false;
			boolean bCompleteMigration=false;
			
			////////////////////////////////////
			//SRIA
			boolean bSRIA = true;
			boolean bHH = true;
			boolean bHHH = true;
			double dHHErrorRate = 0.01;
			double dCostFactorSignificance = 0.40;

			//Combined Stats
			boolean bCombinedStats=true;
			int nNumCombinedStats=2;
			double[] CombinedStatsFactor = {0.3,0.5};
			
			//Index Selection trigger			
			boolean bNoIndexSelectionTrigger=true;
			boolean bCostIncrIndexSelectionTrigger=false;
			boolean bFreqIncrIndexSelectionTrigger=false;
			boolean bCombinedIncrIndexSelectionTrigger=false;
			
			double dDegreeCostIncr = 0.0;
			double dDegreeFreqIncr =0.0;
			
			//ContinuousAssessment
			boolean bContinuousAssessment = false;
			
			//testing
			boolean bStoreActualSRIA = false;
			String strStoreActualIndicesFile = "";
			
			//count migrations
			boolean bCountMigrations=false;
			String strCountMigration="";
			
			//ComparisonTime
			boolean bStoreComparisonTime = false;
			String strStoreComparisonTimeFile="";

			//num tuples to proc per Op
			int nNumTupToProcessPerOp = 0;
			
			int nNumCols = 0;
			
			nodes = ((XATTree)tree.get(0)).convertToArray();
			    
				for (int i = 0; i < nodes.length; i++){
//			    		if(nodes[i].getOperator() instanceof STeMPartitionJoinProbeOperatorImp){
//			    			bHasSTeM = true;
//			    			//look up data
//			    			ArrayList<ArrayList<Integer>> indexes = null;
//			    			if (STeMHashKeys.get(((STeMPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId()) == null) {
//			    				indexes = new ArrayList();
//			    				indexes.add(0,new ArrayList<Integer>());
//			    				indexes.set(0,((STeMPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getNumSTeMIndex());
//			    				STeMHashKeys.put(((STeMPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId(),indexes);
//			    				
//			    				STeMStreamIds.add((((STeMPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId()));
//			    				((STeMPartitionJoinProbeOperatorImp)nodes[i].getOperator()).SetSTeMDataIndexArrayPos(STeMStreamIds.size()-1);
//			    			}
//			    			else
//			    			{
//			    				indexes = STeMHashKeys.get(((STeMPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId());
//			    				indexes.add(new ArrayList<Integer>());
//
//			    				indexes.set(indexes.size()-1,((STeMPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getNumSTeMIndex());			    				
//			    			}
//			    			// set index number for operator
//			    			((STeMPartitionJoinProbeOperatorImp)nodes[i].getOperator()).SetSTeMDataIndexArrayPos(indexes.size()-1);
//
//			    		}
			    		if(nodes[i].getOperator() instanceof STeMMultiStreamPartitionJoinProbeOperatorImp){
			    			bHasSTeM = true;
			    			//look up data
			    			ArrayList<ArrayList<Integer>> indexes = ((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getHashKeyIndex();
			    			if ((STeMHashKeys.get(((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId()) == null)&&(indexes!= null)) 
			    			{
			    				STeMHashKeys.put(((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId(),indexes);
			    				
			    				STeMStreamIds.add((((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId()));
			    				((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).SetSTeMDataIndexArrayPos(STeMStreamIds.size()-1);
			    				
				    			// set index number for operator
				    			((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).SetSTeMDataIndexArrayPos(indexes.size()-1);

			    			}
			    			else
			    			{	
			    				STeMStreamIds.add((((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId()));
			    				((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).SetSTeMDataIndexArrayPos(STeMStreamIds.size()-1);
			    			}	

			    		}
			    		
			    		else if(nodes[i].getOperator() instanceof EddyPartitionOperatorImp){
			    			((EddyPartitionOperatorImp)nodes[i].getOperator()).InitializeVariables();
			    			PreFilledPerStream = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getPreFilledPerStream();
			    			STREAMWindowCount = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getSTREAMWindowCount();
			    			PartitionWindowSize = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getMaxPartitionWindowSize();
			    			PartitionSize = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getMaxPartitionSize();
			    			bIMPIndex = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getUseIMPIndex();
			    			bPartitioned = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getIsPartitioned();
			    			nNumLimitedIndices = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getNumLimitedIndices();
			    			
			    			//sampling?
			    			bNoSampling= ((EddyPartitionOperatorImp)nodes[i].getOperator()).getNoSampling();
			    			bSimpleRandomSampling= ((EddyPartitionOperatorImp)nodes[i].getOperator()).getSimpleRandomSampling();
			    			nSimpleRandomPercent= ((EddyPartitionOperatorImp)nodes[i].getOperator()).getSimpleRandomPercent();
			    			bStratifiedRandomSampling= ((EddyPartitionOperatorImp)nodes[i].getOperator()).getStratifiedRandomSampling();
			    			nStratifiedRandomPercent= ((EddyPartitionOperatorImp)nodes[i].getOperator()).getStratifiedRandomPercent();
			    			nStratifiedRandomWindow= ((EddyPartitionOperatorImp)nodes[i].getOperator()).getStratifiedRandomWindow();
			    			bSystematicSampling= ((EddyPartitionOperatorImp)nodes[i].getOperator()).getSystematicSampling();
			    			nSystematicInterval= ((EddyPartitionOperatorImp)nodes[i].getOperator()).getSystematicInterval();

			    			//index migration
			    			bNoMigration=((EddyPartitionOperatorImp)nodes[i].getOperator()).getNoMigration();
			    			bSymmetricMigration=((EddyPartitionOperatorImp)nodes[i].getOperator()).getSymmetricMigration();
			    			bPartitionMigration=((EddyPartitionOperatorImp)nodes[i].getOperator()).getPartitionMigration();
			    			bCompleteMigration=((EddyPartitionOperatorImp)nodes[i].getOperator()).getCompleteMigration();

			    			//SRIA
			    			bSRIA=((EddyPartitionOperatorImp)nodes[i].getOperator()).getSRIA();
			    			bHH = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getHH();
			    			bHHH = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getHHH();
			    			dHHErrorRate = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getHHErrorRate(); 
			    			dCostFactorSignificance = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getCostFactorSignificance();

			    			//Combined Stats
			    			bCombinedStats=((EddyPartitionOperatorImp)nodes[i].getOperator()).getCombinedStats();
			    			nNumCombinedStats=((EddyPartitionOperatorImp)nodes[i].getOperator()).getNumCombinedStats();
			    			CombinedStatsFactor=((EddyPartitionOperatorImp)nodes[i].getOperator()).getCombinedStatsFactor();

			    			//Index Selection trigger			
//			    			bNoIndexSelectionTrigger=((EddyPartitionOperatorImp)nodes[i].getOperator()).getNoIndexSelectionTrigger();
//			    			bCostIncrIndexSelectionTrigger=((EddyPartitionOperatorImp)nodes[i].getOperator()).getCostIncrIndexSelectionTrigger();
//			    			bFreqIncrIndexSelectionTrigger=((EddyPartitionOperatorImp)nodes[i].getOperator()).getFreqIncrIndexSelectionTrigger();
//			    			bCombinedIncrIndexSelectionTrigger=((EddyPartitionOperatorImp)nodes[i].getOperator()).getCombinedIncrIndexSelectionTrigger();
			    			
			    			dDegreeCostIncr =((EddyPartitionOperatorImp)nodes[i].getOperator()).getDegreeCostIncr();
			    			dDegreeFreqIncr =((EddyPartitionOperatorImp)nodes[i].getOperator()).getDegreeFreqIncr();

			    			//testing
			    			bStoreActualSRIA =((EddyPartitionOperatorImp)nodes[i].getOperator()).getTrackOptimalIndices();
			    			strStoreActualIndicesFile=((EddyPartitionOperatorImp)nodes[i].getOperator()).getTrackOptimalIndicesOutputfile();

			    			//CountMigrations
			    			bCountMigrations =((EddyPartitionOperatorImp)nodes[i].getOperator()).getCountMigrations();
			    			strCountMigration=((EddyPartitionOperatorImp)nodes[i].getOperator()).getCountMigrationsFile();

			    			//StoreComparisonTime
			    			bStoreComparisonTime = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getStoreComparisonTime();
			    			strStoreComparisonTimeFile =((EddyPartitionOperatorImp)nodes[i].getOperator()).getStoreComparisonTimeFile();
			    			
			    			//ContinuousAssessment
			    			bContinuousAssessment=((EddyPartitionOperatorImp)nodes[i].getOperator()).getContinuousAssessment();
			    		
			    			nNumTupToProcessPerOp = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getNumTupToProcessPerOp();
			    			nNumCols = ((EddyPartitionOperatorImp)nodes[i].getOperator()).getNumJoinAttribs();
			    		}

			  	}

				
				HashMap<Integer,STeMPartitionData> STeMStreamData = null;


				if (STeMStreamIds.size()> 0)
				{	
					STeMStreamData = new HashMap(STeMStreamIds.size());
					for (int k = 0;k<STeMStreamIds.size();k++)
					{
						bHasSTeM = true;
						STeMPartitionData sd = new STeMPartitionData();
						
						//set partition window size
						sd.setMaxPartitionSize(PartitionSize);
						sd.setMaxPartitionWindowSize(PartitionWindowSize);						
						sd.setIsPartitioned(bPartitioned);
						sd.setUseIMPIndex(bIMPIndex);
						sd.setNumLimitedIndices(nNumLimitedIndices);

						
		    			//num attribs
		    			sd.setNumCols(nNumCols);
		    			
		    			//testing parameters
		    			sd.setStoreActualSRIA(bStoreActualSRIA);
		    			sd.setStoreActualIndicesFile(strStoreActualIndicesFile);		    			
		    			
		    			//ContinuousAssessment
		    			sd.setContinuousAssessment(bContinuousAssessment);

						//initialize state
		    			sd.init(STeMStreamIds.get(k),STeMHashKeys.get(STeMStreamIds.get(k)));
						
						//sampling?
						sd.setNoSampling(bNoSampling);
						sd.setSimpleRandomSampling(bSimpleRandomSampling);
						sd.setSimpleRandomPercent(nSimpleRandomPercent);
						sd.setStratifiedRandomSampling(bStratifiedRandomSampling);
						sd.setStratifiedRandomPercent(nStratifiedRandomPercent);
						sd.setStratifiedRandomWindow(nStratifiedRandomWindow);
						sd.setSystematicSampling(bSystematicSampling);
						sd.setSystematicInterval(nSystematicInterval);
						
						//index migration
						sd.setNoMigration(bNoMigration);
						sd.setSymmetricMigration(bSymmetricMigration);
						sd.setPartitionMigration(bPartitionMigration);
						sd.setCompleteMigration(bCompleteMigration);

		    			//SRIA
						sd.setSRIA(bSRIA);
		    			sd.setHH(bHH);
		    			sd.setHHH(bHHH);
		    			sd.setHHErrorRate(dHHErrorRate);
		    			sd.setCostFactorSignificance(dCostFactorSignificance);

		    			
		    			//Combined Stats
		    			sd.setCombinedStats(bCombinedStats);
		    			sd.setNumCombinedStats(nNumCombinedStats);
		    			sd.setCombinedStatsFactor(CombinedStatsFactor);
						
		    			//Index Selection trigger			
		    			sd.setNoIndexSelectionTrigger(bNoIndexSelectionTrigger);
		    			sd.setCostIncrIndexSelectionTrigger(bCostIncrIndexSelectionTrigger);
		    			sd.setFreqIncrIndexSelectionTrigger(bFreqIncrIndexSelectionTrigger);
		    			sd.setCombinedIncrIndexSelectionTrigger(bCombinedIncrIndexSelectionTrigger);
		    			
		    			sd.setDegreeCostIncr(dDegreeCostIncr);
		    			sd.setDegreeFreqIncr(dDegreeFreqIncr);

		    			//CountMigrations
		    			sd.setCountMigrations(bCountMigrations);
		    			sd.setCountMigrationsFile(strCountMigration);

		    			//StoreComparisonTime
		    			sd.setStoreComparisonTime(bStoreComparisonTime);
		    			sd.setStoreComparisonTimeFile(strStoreComparisonTimeFile);
		    			
		    			STeMStreamData.put(STeMStreamIds.get(k),sd);
						
						//set window size??
						if (STREAMWindowCount != null)
						{
							int windowSize = 0;
							if (STREAMWindowCount.get(STeMStreamIds.get(k)) != null)
							{
								windowSize =STREAMWindowCount.get(STeMStreamIds.get(k));
								sd.SetMaxWindowSize(windowSize);
							}							
						}

						
						
//						//preset starting index
//for 3 streams						
//						if(k==0)
//						{	
//							int[] GroupPos = new int[]{0, 2, 4};
//							int[] IndexAttrPos = new int[]{0, 2, 4};
//							int[] NumEntries = new int[]{8, 6, 6};
//							
//							sd.SetIMPIndex(GroupPos, IndexAttrPos, NumEntries);							
//						}
//						else if(k==1)
//						{	
//							int[] GroupPos = new int[]{0, 1, 4};
//							int[] IndexAttrPos = new int[]{0, 1, 4};
//							int[] NumEntries = new int[]{8, 6, 6};
//							
//							sd.SetIMPIndex(GroupPos, IndexAttrPos, NumEntries);							
//						}						
//						else if(k==2)
//						{	
//							int[] GroupPos = new int[]{1, 2, 4};
//							int[] IndexAttrPos = new int[]{1, 2, 4};
//							int[] NumEntries = new int[]{8, 6, 6};
//							
//							sd.SetIMPIndex(GroupPos, IndexAttrPos, NumEntries);							
//						}

//for 4 streams
					if(bIMPIndex)
					{	
						if(k==0)
						{	
							int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
							int[] IndexAttrPos = new int[]{2,3,6};
//							int[] NumEntries = new int[]{18, 6};
							int[] NumEntries = new int[]{4,1,3};
							
//							int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							int[] IndexAttrPos = new int[]{0, 2, 5};
//							int[] NumEntries = new int[]{21, 21, 22};
							
//							int[] NumEntries = new int[]{8, 4, 4};
//							int[] NumEntries = new int[]{8, 12, 12};
							
							sd.SetIMPIndex(GroupPos, IndexAttrPos, NumEntries);							
						}
						else if(k==1)
						{
							int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
							int[] IndexAttrPos = new int[]{2, 3, 6};
							int[] NumEntries = new int[]{2, 1,5};
//							int[] NumEntries = new int[]{18, 6};
							
//							int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							int[] IndexAttrPos = new int[]{0, 1, 4};
//							int[] NumEntries = new int[]{21, 21, 22};
							
//							int[] NumEntries = new int[]{8, 4, 4};
//							int[] NumEntries = new int[]{8, 12, 12};
							
							sd.SetIMPIndex(GroupPos, IndexAttrPos, NumEntries);							
						}						
						else if(k==2)
						{	
							int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
							int[] IndexAttrPos = new int[]{2, 3, 6};
							int[] NumEntries = new int[]{1, 6, 1};
//							int[] NumEntries = new int[]{6, 18};
							
//							int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							int[] IndexAttrPos = new int[]{2, 1, 3};
//							int[] NumEntries = new int[]{21, 21, 22};
							
//							int[] NumEntries = new int[]{8, 4, 4};
//							int[] NumEntries = new int[]{8, 4, 4};
//							int[] NumEntries = new int[]{8, 12, 12};
							
							sd.SetIMPIndex(GroupPos, IndexAttrPos, NumEntries);							
						}
						else if(k==3)
						{	
							int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
							int[] IndexAttrPos = new int[]{2,3};
							int[] NumEntries = new int[]{4, 4};
//							int[] NumEntries = new int[]{6, 18};
							
//							int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							int[] IndexAttrPos = new int[]{5, 4, 3};
//							int[] NumEntries = new int[]{21, 21, 22};
							
//							int[] NumEntries = new int[]{8, 4, 4};
//							int[] NumEntries = new int[]{8, 12, 12};
							
							sd.SetIMPIndex(GroupPos, IndexAttrPos, NumEntries);							
						}
////							
					}	
					else
					{
						if(k==0)
						{
							if (nNumLimitedIndices>=1)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{0, 2, 5};
//							  int[] hFactors = new int[]{21, 21, 22};
							  
//							  real data
							  int[] IndexAttrPos = new int[]{2, 3};
							  int[] hFactors = new int[]{4, 4};
							  
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=2)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};

//							  real data
							  int[] IndexAttrPos = new int[]{2,3,6};							  
							  int[] hFactors = new int[]{2,2,4};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=3)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
							  
//							  real data
							  int[] IndexAttrPos = new int[]{2};							  
							  int[] hFactors = new int[]{8};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=4)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{2, 5};
//							  int[] hFactors = new int[]{32, 32};
							
//							  real data
							  int[] IndexAttrPos = new int[]{2, 6};
							  int[] hFactors = new int[]{4, 4};
							  
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=5)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{0};
//							  int[] hFactors = new int[]{64};
							
//							  real data
							  int[] IndexAttrPos = new int[]{3,6};
							  int[] hFactors = new int[]{4,4};
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=6)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{2};
//							  int[] hFactors = new int[]{64};
							
//							  real data
							  int[] IndexAttrPos = new int[]{3};
							  int[] hFactors = new int[]{8};
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=7)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{5};
//							  int[] hFactors = new int[]{64};
							
//							  real data
							  int[] IndexAttrPos = new int[]{6};
							  int[] hFactors = new int[]{8};
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
						}
						else if(k==1)
						{
							if (nNumLimitedIndices>=1)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
							  
//							  real data
							  int[] IndexAttrPos = new int[]{2, 3,6};							  
							  int[] hFactors = new int[]{4, 2,2};

//							  synthetic data
//							  int[] IndexAttrPos = new int[]{0, 1, 4};
//							  int[] hFactors = new int[]{21, 21, 22};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=2)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
							  
//							  real data
							  int[] IndexAttrPos = new int[]{2,6};							  
							  int[] hFactors = new int[]{4,4};

//							  synthetic data
//							  int[] IndexAttrPos = new int[]{0,  4};
//							  int[] hFactors = new int[]{32, 32};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=3)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
							  
//							  real data
							  int[] IndexAttrPos = new int[]{2};
							  int[] hFactors = new int[]{8};

//							  synthetic data
//							  int[] IndexAttrPos = new int[]{1,  4};
//							  int[] hFactors = new int[]{32, 32};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=4)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
								int[] IndexAttrPos = new int[]{2, 6};
								int[] hFactors = new int[]{4,4};
							  
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{0, 1};
//							  int[] hFactors = new int[]{32, 32};
//							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=5)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
								int[] IndexAttrPos = new int[]{3, 6};
								int[] hFactors = new int[]{4,4};
							  
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{0};
//							  int[] hFactors = new int[]{64};
//							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=6)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
								int[] IndexAttrPos = new int[]{3};
								int[] hFactors = new int[]{8};
							  
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{4};
//							  int[] hFactors = new int[]{64};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=7)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
								int[] IndexAttrPos = new int[]{6};
								int[] hFactors = new int[]{8};

//							  synthetic data
//							  int[] IndexAttrPos = new int[]{1};
//							  int[] hFactors = new int[]{64};
//							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
						}						
						else if(k==2)
						{	
							if (nNumLimitedIndices>=1)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
//							  real data
							  int[] IndexAttrPos = new int[]{2};							  
							  int[] hFactors = new int[]{8};
							  
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{1, 2, 3};
//							  int[] hFactors = new int[]{21, 21, 22};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=2)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
//							  real data
							  int[] IndexAttrPos = new int[]{2, 3};							 							  
							  int[] hFactors = new int[]{4, 4};
							  

//							  synthetic data
//							  int[] IndexAttrPos = new int[]{1,  2};
//							  int[] hFactors = new int[]{32, 32};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=3)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
//							  real data
							  int[] IndexAttrPos = new int[]{2,3,6};
							  int[] hFactors = new int[]{2,2,4};

//							  synthetic data
//							  int[] IndexAttrPos = new int[]{1,  3};
//							  int[] hFactors = new int[]{32, 32};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=4)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
								int[] IndexAttrPos = new int[]{2, 6};
								int[] hFactors = new int[]{4, 4};
							  
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{2, 3};
//							  int[] hFactors = new int[]{32, 32};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=5)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
								int[] IndexAttrPos = new int[]{3, 6};
								int[] hFactors = new int[]{4,4};
							  
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{3};
//							  int[] hFactors = new int[]{64};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=6)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
								int[] IndexAttrPos = new int[]{3};
								int[] hFactors = new int[]{8};
							  
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{2};
//							  int[] hFactors = new int[]{64};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=7)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
								int[] IndexAttrPos = new int[]{6};
								int[] hFactors = new int[]{8};
								
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{1};
//							  int[] hFactors = new int[]{64};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
						}
						else if(k==3)
						{	
							
							if (nNumLimitedIndices>=1)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
//							  real data
							  int[] IndexAttrPos = new int[]{2,6};							  
							  int[] hFactors = new int[]{4,4};
							  
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{3, 4, 5};
//							  int[] hFactors = new int[]{21, 21, 22};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=2)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
//							  real data
							  int[] IndexAttrPos = new int[]{2, 3, 6};							  
							  int[] hFactors = new int[]{4, 2, 2};
							  
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{4,  5};
//							  int[] hFactors = new int[]{32, 32};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=3)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8};
//							  real data
							  int[] IndexAttrPos = new int[]{3,6};
							  int[] hFactors = new int[]{4,4};
							  
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{5};
//							  int[] hFactors = new int[]{64};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=4)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
							  int[] IndexAttrPos = new int[]{2, 6};
							  int[] hFactors = new int[]{4, 4};
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{4};
//							  int[] hFactors = new int[]{64};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=5)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
							  int[] IndexAttrPos = new int[]{3, 6};
							  int[] hFactors = new int[]{4, 4};
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{3, 4};
//							  int[] hFactors = new int[]{32, 32};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=6)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
								int[] IndexAttrPos = new int[]{3};
								int[] hFactors = new int[]{8};
							  
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{3, 5};
//							  int[] hFactors = new int[]{32, 32};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
							
							if (nNumLimitedIndices>=7)
							{	
							  int[] GroupPos = new int[]{0, 1, 2, 3, 4, 5, 6, 7};
//							  real data
								int[] IndexAttrPos = new int[]{6};
								int[] hFactors = new int[]{8};
//
//							  synthetic data
//							  int[] IndexAttrPos = new int[]{3};
//							  int[] hFactors = new int[]{64};
							
							  sd.SetHashIndex(GroupPos, IndexAttrPos, hFactors);
							}
						}
						
					}
					

				  }
										
				}


				
				//Create Eddy DataStructures required for processing
							    
				for (int i = 0; i < nodes.length; i++){
			    		if(nodes[i].getOperator() instanceof EddyPartitionOperatorImp){
			    			if (bHasSTeM == true)
							  ((EddyPartitionOperatorImp)nodes[i].getOperator()).setSTeMStreamData(STeMStreamData);
			    		}
			    		else if(nodes[i].getOperator() instanceof STeMPartitionJoinProbeOperatorImp){
			    			//probe
			    			STeMPartitionData sd = STeMStreamData.get(((STeMPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId());
			    			if (sd != null){
			    			  ((STeMPartitionJoinProbeOperatorImp)nodes[i].getOperator()).setSTeMStreamData(sd);
			    			}
			    			((STeMPartitionJoinProbeOperatorImp)nodes[i].getOperator()).setPreFilledPerStream(PreFilledPerStream);
			    			
			    		}
			    		else if(nodes[i].getOperator() instanceof STeMMultiStreamPartitionJoinProbeOperatorImp){
			    			//probe
			    			STeMPartitionData sd = STeMStreamData.get(((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId());
			    			if (sd != null){
			    			  ((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).setSTeMStreamData(sd);
			    			}
			    			
							//num to proc per op
			    			((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).setNumTupToProcessPerOp(nNumTupToProcessPerOp);

			    			((STeMMultiStreamPartitionJoinProbeOperatorImp)nodes[i].getOperator()).setPreFilledPerStream(PreFilledPerStream);
			    			
			    		}
			  	}
				//End KWORKS - Eddy
				
				System.out.println("Initialization...done.");
			}

			


	}

