/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller;
import java.util.HashMap;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.NotSerializableException;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.net.Socket;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.realjoinoperator.WTransitiveHashJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WEquiJoinStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamUnionOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.xmlstreamexpression.BinCOMPExpressionStrImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xmlstreamexpression.TerminalExpressionStrImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.DistributionConfiguration;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ConnectionListener;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ConnectionTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.MigrationKeeper;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationcontroller.SubPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.migrationprotocol.SendStateToMachine;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.PacketPool;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleDistributor;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleFeeder;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleReceiver;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleSender;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.spreadfire.manager.SpreadfireManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.FireStreamGUI;
import edu.wpi.cs.dsrg.STeM.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUIQueryEngine;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATExecutionAdapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision;
import edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerException;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.LightWeightStatistic;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsPrinter;
import edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.heuristics.Optimizer;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation.Adapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.partitionadaptation.LocalMemoryAdapter;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.adprotocol.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.UpdatePartitionStatistics;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.*;


import edu.wpi.cs.dsrg.eddy.datastructure.EddyQueue;
import edu.wpi.cs.dsrg.eddy.operator.*;
import edu.wpi.cs.dsrg.STeM.STeMData;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.TerminalConstantImp;

public class QueryMeshExecutionController extends DistributedExecutionController 
{	
	/*************************************************/
	public class DTChild 
	{
		//node n;
		int id;
		
		public DTChild(int i /*, node nd*/)
		{
			id = i;
			//n = nd;
		}
	}
	/*************************************************/
	
	public QueryMeshExecutionController(DistributionConfiguration dc,String clientName) throws Exception 
	{
		super(dc, clientName);	
	}

	
	public String getGlobalDTString()
	{
		String str = "";		
		return str;
	}
	
	/**
	 * Override the Initialization
	 */	
	protected void initExecution() throws Exception 
	{			
		/* init query plan optimizer */
		optimizer = new Optimizer();
				
		/** next set up the ExecutionController properties */
		Enumeration e = config.getProperties();
		System.out.println("config file is: " + config.toString());
		while(e.hasMoreElements())
		{
			String name = (String) e.nextElement();
			String value = config.getProperty(name);
			if(this.isValidStatistic(name))
			{
				this.setProperty(name, value);
			}
		}

		distInfo = new String[4];
		distInfo[0] = "Initial distribution algorithm: " + config.getProperty("DISTRIBUTION_PATTERN").substring(config.getProperty("DISTRIBUTION_PATTERN").lastIndexOf(".") + 1);
		distInfo[1] = "Redistribution algorithm: " + config.getProperty("REDISTRIBUTION_POLICY").substring(config.getProperty("REDISTRIBUTION_POLICY").lastIndexOf(".") + 1);
		distInfo[2] = "Redistribution interval: " + (Integer.parseInt(config.getProperty("REDISBRIBUTION_TIME")) / 1000);
		distInfo[3] = "Redistribution percentage: " + config.getProperty("REDISTRIBUTION_PERCENT") + "%";
				
		// Get the machine properties
		machine = config.getMachine(clientName);
		Enumeration en = machine.getProperties();
		while(en.hasMoreElements())
		{
			String name = (String) en.nextElement();
			String property = machine.getProperty(name);
			if(this.isValidPropertyName(name))	
			{
				this.setProperty(name, property);
			}
		}
		
		// Get the query plans	
		int c = 0;
		int numQueryPlans = 0;
				
		for(Iterator i= config.getQueryPlans(); i.hasNext();)
		{
			QueryPlan qp = (QueryPlan) i.next();
			System.out.println("queryplan to add is: " + qp.getProperty("QUERY_ID"));
			tree.add(qp.getQueryPlan());
			c++;
			numQueryPlans++;			
		}
				
		System.out.println("tree size is: " + tree.size());    
		
		debug = (new Boolean(this.getProperty("EXECUTION_DEBUG"))).booleanValue();
		debug_scheduling_notice =  (new Boolean(this.getProperty("SCHEDULING_NOTIFICATION_DEBUG"))).booleanValue();
		lastTimeOperatorUpdated = new Hashtable();
		lastTimeStatusChecked = System.currentTimeMillis();
		lastTimeTreesUpdated = System.currentTimeMillis();
		schedulerCount = new Hashtable();
				
		stats.setProperty("TIMER", this.getProperty("STATS_TABLE_STATS"));	
				
		XATDag[] statsTree = new XATDag[tree.size()];
		for(int i = 0; i < tree.size(); i++)
		{
			statsTree[i] = (XATDag) tree.get(i);
		}

		stats.initStatsTable(statsTree);  
		stats.addNewQueryObject(system);       
			         	             		
		if(debug)
		{
		    System.out.println(tree);
		}
				
		/**
		 init the last time some of its properties were updated
		 set the last updated time to 0, that way the properties will be updated the first time.    
		 This helps with experiments 
		*/
			    			
		for(int count=0; count < tree.size(); count++)
		{
			XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
			for (int i = 0; i < nodes.length; i++)
			{  		
				if(nodes[i].getOperator() instanceof XATStreamOperator)
				{
					XATStreamOperator operator = (XATStreamOperator)nodes[i].getOperator();
					lastTimeOperatorUpdated.put(operator, new Long(0));			    		
				}			  		
			}			    
		}
			    			   
		//set up schedulers
		String className = this.getProperty("ADAPTIVE_HEURISTIC"); 
		treeArray = new XATDag[tree.size()];
		for(int i = 0; i < tree.size();i++)
		{
			treeArray[i] = (XATDag) tree.get(i);
		}
		
		Constructor[] c1 = Class.forName(className).getConstructors();
	    Constructor construct;
			    
		for(int i = 0; i < c1.length; i++)
		{
			construct = c1[i];
			Class[] parameters = construct.getParameterTypes();
			System.out.println(parameters[0].getName());
			if(parameters.length == 2)
			{
				if(parameters[0].getName().equals("edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.Machine"))
				{
					adapter = (XATExecutionAdapter)construct.newInstance(new Object[] { machine, treeArray});
				}							
			}				
		}						
				
		if(adapter == null)
		{		
			throw new Exception("The Adapter could not be instantiated.");
		}
		
		//start at scheduler 0 
		currentScheduler = adapter.getCurrentScheduler();
		updateSchedulerCount();
			    
		//set up some of the statistics to gather
		initStatsToGather();
		
		//set up the printer
		initStatisticsPrinter();
			   
		//set the starting category in the statistics gatherer to be the first scheduler
		stats.setCurrentCategory(adapter.getCategory());
		printer.setCategory(stats.getCurrentCategory());	
				
		//set up local partition adapter -- Bin @ Dec.10, 2004
		className = this.getProperty("LOCAL_PARTITION_ADAPTER");
		if (!className.equals("none")) 
		{
			c1 = Class.forName(className).getConstructors();
			construct = c1[0]; // might go wrong if multiple constructor exists
			//localAdapter = (Adapter)construct.newInstance(null);		
			localAdapter = (Adapter)construct.newInstance(new Object[]{tree});		
		}
		
		//set the distributed execution controller reference to the symmmwayhashjoinoperator
		//--Bin @ Apr 12, 2005, to handle the tracing of cost information in different machines
		//this part of the code could be safely commented when run other queries/without of tracing cost statisitics
		for(int count=0; count < tree.size(); count++)
		{
			XATNode[] nodes = ((XATTree)tree.get(count)).convertToArray();
		  	for (int i = 0; i < nodes.length; i++)
		  	{
				if(nodes[i].getOperator() instanceof SymmMwayHashJoinOperatorImp)
				{
					//((SymmMwayHashJoinOperatorImp)nodes[i].getOperator()).setDistributedExectionControllerRef(this);
			    }			  	
		  	}			   
		}

		//execution decision tree
		int NumStreams = 0;
		node rootDt;
		int attributeId;				
		ArrayList<node> edt = null;
		ArrayList<ArrayList<DecisionTreeNode>> StreamDT = null;
		ArrayList<HashMap<Integer,RusterNode>> RusterPerStream = null;
		
		XATNode[] nodes = ((XATTree)tree.get(0)).convertToArray();
	
		// set up op lookup
		//# ops = num nodes - classifier & dummy
		OperatorLookUp OpLkUp = new OperatorLookUpImp(nodes.length);
		XATQueue inputQ;
		
		ArrayList<Integer> STeMStreamIds;
		HashMap<Integer,ArrayList<ArrayList<Integer>>> STeMHashKeys;

		STeMHashKeys = new HashMap(); //HashKeys
		STeMStreamIds = new ArrayList(); //STREAM Ids
		ArrayList<BitSet> PreFilledPerStream = null;
		
		HashMap <Integer,Integer> STREAMWindowCount = null;
		
		boolean bHasSTeM = false;
		int NumSendOff = 0;
	   		
		// KEW		
		for (int i = 0; i < nodes.length; i++)
		{
    		if(nodes[i].getOperator() instanceof OnlineClassifierOperatorImp)
    		{
				NumStreams = ((OnlineClassifierOperatorImp)nodes[i].getOperator()).getNumStreams();
				StreamDT = ((OnlineClassifierOperatorImp)nodes[i].getOperator()).getStreamDT();
				RusterPerStream = ((OnlineClassifierOperatorImp)nodes[i].getOperator()).getRusterPerStream();
				//execution DT
				edt = new ArrayList<node>(NumStreams);
    			STREAMWindowCount = ((OnlineClassifierOperatorImp)nodes[i].getOperator()).getSTREAMWindowCount();
    			NumSendOff	= ((OnlineClassifierOperatorImp)nodes[i].getOperator()).getSendOff();
    		}    		
    		else if(nodes[i].getOperator() instanceof edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.DoNothingOperatorImp)
    		{
    			inputQ= ((XATMultiQueueStreamOperatorImp)nodes[i].getOperator()).getInputQueue(0);
    			OpLkUp.setOutputQueue(inputQ);    			
    		}    		
    		else if(nodes[i].getOperator() instanceof QMeshStreamSelectOperatorImp)
    		{
    			inputQ= ((XATMultiQueueStreamOperatorImp)nodes[i].getOperator()).getInputQueue(0);
    			OpLkUp.setOperatorQueue(inputQ,((QMeshStreamSelectOperatorImp)nodes[i].getOperator()).getQMeshOperatorID());
    		}
    		else if(nodes[i].getOperator() instanceof STeMJoinProbeOperatorImp)
    		{
    			inputQ = ((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getInputQueue();
    			OpLkUp.setOperatorQueue(inputQ,((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getQMeshOperatorID());
    			
    			// set up STEM
    			bHasSTeM = true;
    			//look up data
    			ArrayList<ArrayList<Integer>> indexes = null;
    			if (STeMHashKeys.get(((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId()) == null) 
    			{
    				indexes = new ArrayList();
    				indexes.add(0,new ArrayList<Integer>());
    				indexes.set(0,((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getNumSTeMIndex());
    				STeMHashKeys.put(((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId(),indexes);
    				
    				STeMStreamIds.add((((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId()));
    				((STeMJoinProbeOperatorImp)nodes[i].getOperator()).SetSTeMDataIndexArrayPos(STeMStreamIds.size()-1);
    			}
    			else
    			{
    				indexes = STeMHashKeys.get(((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId());
    				indexes.add(new ArrayList<Integer>());

    				indexes.set(indexes.size()-1,((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getNumSTeMIndex());			    				
    			}
    			// set index number for operator
    			((STeMJoinProbeOperatorImp)nodes[i].getOperator()).SetSTeMDataIndexArrayPos(indexes.size()-1);
    		}
		}

		//create execution decision tree		
		for (int j = 0; j < NumStreams; j++)
		{
			edt.add(new node());
			rootDt  = edt.get(j);

			//dt for each stream
			if (StreamDT.get(j).size() == 0)
			{
				//only one path default path
				rootDt.DefaultPath = RusterPerStream.get(j).get(0);				
			}
			else
			{
				// need to parse array to create execution dt
				//First look at children of root
				DecisionTreeNode root = StreamDT.get(j).get(0);

				ArrayList<DTChild> childrenToExpand = new ArrayList<DTChild>();
				
				//edt.add(new node());
				rootDt  = edt.get(j);
				rootDt.DefaultPath = RusterPerStream.get(j).get(0);

				node n = rootDt; // parent node
				node newNode = null;
				node childNode = null;
				int DefaultRouteId = 0;
				
				//-----------------------------------------------------------------
				// array list of ids of children to explore
				//-----------------------------------------------------------------
				for (int ch = 0;ch < root.children.size();ch++)
				{
					childrenToExpand.add(new DTChild(root.children.get(ch)));
				}
				
				//-----------------------------------------------------------------
				// keep on exploring until the child list is empty
				//-----------------------------------------------------------------				
				while (childrenToExpand.size() > 0)
				{
					DTChild exploreChild = childrenToExpand.get(0);
					childrenToExpand.remove(0);
					
					int ch = exploreChild.id;
					//n = exploreChild.n;
					//for each child
					
					//-----------------------------------------------------------------
					// get the actual decision tree node
					//-----------------------------------------------------------------	
					DecisionTreeNode child = StreamDT.get(j).get(ch);
					
					//-----------------------------------------------------------------
					// BUCKET CHECK e.g., value must be in the range [0...100]
					//-----------------------------------------------------------------	
					if 	(child.is_bucket_value) 
					{
						RangeNode rn = null;
						//does this already have a range node at this level for this attrib?
						for (int u = 0; u < n.obj.size(); u++)
						{
							if(n.IsRangeNode(u))
							{
								rn = (RangeNode)n.obj.get(u);
								if (rn.getAttributeId() == child.attr_idx)
								{
									break;
								}
								else
								{
									rn = null;
								}
							}
						}
												
						if (rn == null)
						{						   
							//add new						   
							childNode = new node();
						    childNode.parent= n;
						   
						    rn = new RangeNode(child.attr_idx, child.bucket_range_start, child.bucket_range_end, childNode);							
						    n.addObject(n.RangeNodeId, rn);
						}
						else
						{
							//node exists							
							childNode = new node();
							childNode.parent= n;
							
							rn.AddRangeValue(child.bucket_range_start, child.bucket_range_end, childNode);
						}
						
						for (int f = 0; f < child.children.size(); f++)
						{
							childrenToExpand.add(new DTChild(child.children.get(f)));
						}					  
					}	
					//-----------------------------------------------------------------
					// EQUALITY CHECK node e.g., x == 100
					//-----------------------------------------------------------------	
					//Equal
					else if (child.operation.equalsIgnoreCase("EQ")) 
					{						 
						EqualOperatorNode eqn = null;
						//does this already have a range node at this level for this attrib?
						for (int u =0; u < n.obj.size(); u++)
						{						
							if(n.IsEqualOperatorNode(u))
							{						
								eqn = ((EqualOperatorNode)n.obj.get(u));
								if (eqn.getAttributeId() == child.attr_idx)
								{								  
									break;
								}
								else
								{								 
									eqn = null;
								}
							}
						 }		
						 if (eqn == null)
						 {							
							 childNode = new node();
							 childNode.parent = n;
							
							 //add new
							 eqn = new EqualOperatorNode(child.attr_idx);
							 eqn.addNode(child.test_value, childNode);
							
							 n.addObject(n.EqualOperatorNodeId, eqn);
						 }
						 else
						 {
							 childNode = new node();
							
							 childNode.parent = n;
							 //node exists
							 eqn.addNode(child.test_value, childNode);
						 }
						 
						 for (int f=0;f<child.children.size();f++)
						 {
							 childrenToExpand.add(new DTChild(child.children.get(f)));
						 }
					}						 
					//ruster
					else
					{						
						RusterNode rn = null;
						if (child.route_id == -1)
						{
							//do nothing, will use default route for this case
						}
						else if(((RusterNode)RusterPerStream.get(j).get(child.route_id)).IsDefault())
						{	
							rn = RusterPerStream.get(j).get(0);
							DefaultRouteId = child.route_id;
						}
						else
						{
							rn = RusterPerStream.get(j).get(child.route_id);		
						}

						//newNode = new node();
						//newNode.addObject(newNode.RusterId,rn);
						//newNode.parent = n;

						n.addObject(n.RusterId, rn);																					
					}				 
				}				
				 
				//need to sort range array
				//RusterPerStream.get(j).remove(DefaultRouteId);				
			}
				
			System.out.println("The runtime DT is ready");			
		}// end for //iteration through individual decision trees

		//set up STEM data
		HashMap<Integer,STeMData> STeMStreamData = null;

		if (STeMStreamIds.size() > 0)
		{	
			STeMStreamData = new HashMap(STeMStreamIds.size());
			for (int k = 0;k < STeMStreamIds.size();k++)
			{
				bHasSTeM = true;
				STeMData sd = new STeMData();
				sd.init(STeMStreamIds.get(k),STeMHashKeys.get(STeMStreamIds.get(k)));
				STeMStreamData.put(STeMStreamIds.get(k),sd);
				
				//set window size??
				if (STREAMWindowCount != null)
				{
					int windowSize = 0;
					if (STREAMWindowCount.get(STeMStreamIds.get(k)) != null)
					{
						windowSize =STREAMWindowCount.get(STeMStreamIds.get(k));
						sd.SetMaxWindowSize(windowSize);
					}							
				}				
			}//end for
		}//end if
		
		
		//initialize variables
		for (int i = 0; i < nodes.length; i++)
		{
    		if(nodes[i].getOperator() instanceof OnlineClassifierOperatorImp)
    		{
    			((OnlineClassifierOperator)nodes[i].getOperator()).setOperatorLookUp(OpLkUp);
    			((OnlineClassifierOperator)nodes[i].getOperator()).setExecutionDecisionTree(edt);
    			//((OnlineClassifierOperator)nodes[i].getOperator()).setRusterPerStream(RusterPerStream);
    			if (bHasSTeM == true)
    			{					  
    				((OnlineClassifierOperator)nodes[i].getOperator()).setSTeMStreamData(STeMStreamData);
    			}
    		}
    		else if(nodes[i].getOperator() instanceof QMeshStreamSelectOperatorImp)
    		{
    			((QMeshStreamSelectOperator)nodes[i].getOperator()).setOperatorLookUp(OpLkUp);    			
    		}
    		else if(nodes[i].getOperator() instanceof STeMJoinProbeOperatorImp)
    		{
    			((STeMJoinProbeOperatorImp)nodes[i].getOperator()).setOperatorLookUp(OpLkUp);
    			
	    		//probe
	    		STeMData sd = STeMStreamData.get(((STeMJoinProbeOperatorImp)nodes[i].getOperator()).getDataSTREAMId());
	    		if (sd != null)
	    			((STeMJoinProbeOperatorImp)nodes[i].getOperator()).setSTeMStreamData(sd);
	    	
	    		((STeMJoinProbeOperatorImp)nodes[i].getOperator()).setPreFilledPerStream(PreFilledPerStream);
	    		((STeMJoinProbeOperatorImp)nodes[i].getOperator()).setSendOff(NumSendOff);	    			    			
	    	}
    	}

		//End KWORKS - QMESH
					
		System.out.println("Initialization...done.");
	}	
}

//Start KWORKS - QMESH

				//start create hard coded execution decision tree(s)
//				int NumStreams = 2;
//				node rootDt;
//				int attributeId;
//				
//				ArrayList<node> edt = new ArrayList<node>(NumStreams);
//				ArrayList<HashMap<Integer,RusterNode>> RusterPerStream = new ArrayList<HashMap<Integer,RusterNode>>(NumStreams); 
//hardcoded path
//	//////////////////////////////
//	//stream A
//				edt.add(new node());
//				rootDt  = edt.get(0);
//				
//				RusterPerStream.add(new HashMap<Integer,RusterNode>());
//				
//				ArrayList<Integer> os;
//				EqualOperatorNode eon;
//				RangeNode rn;
//				
//				//////////////////////////////
//				// default path
//				os = new ArrayList<Integer>(4);
//				os.add(2);
//				os.add(4);
//				os.add(3);
//				os.add(1);
//				
//				
//				rootDt.DefaultPath = new RusterNode(os,0);
//				RusterPerStream.get(0).put(0, rootDt.DefaultPath);
//
//				//////////////////////////////
//				// second path
//				os = new ArrayList<Integer>(4);
//				os.add(3);
//				os.add(2);
//				os.add(4);
//				os.add(1);
//				
//				RusterNode Path2 =  new RusterNode(os,1);
//				
//				node nPath2 = new node();
//				nPath2.addObject(nPath2.RusterId,Path2);
//				nPath2.parent = rootDt;
//
//				RusterPerStream.get(0).put(1, Path2);
//
//				//////////////////////////////
//				// third path
//				os = new ArrayList<Integer>(4);
//				os.add(3);
//				os.add(4);
//				os.add(2);
//				os.add(1);
//				
//				RusterNode Path3 =  new RusterNode(os,2);
//				
//				node nPath3 = new node();
//				nPath3.addObject(nPath3.RusterId,Path3);
//				nPath3.parent = rootDt;
//				RusterPerStream.get(0).put(2, Path3);
//
//				//////////////////////////////
//				// fourth path
//				os = new ArrayList<Integer>(4);
//				os.add(4);
//				os.add(3);
//				os.add(2);
//				os.add(1);
//				
//				RusterNode Path4 =  new RusterNode(os,3);
//				
//				node nPath4 = new node();
//				nPath4.addObject(nPath4.RusterId,Path4);
//				nPath4.parent = rootDt;
//				
//				RusterPerStream.get(0).put(3, Path4);
//				//////////////////////////////
//				// add equal node
//				
//				attributeId = 3; //moteid
//				eon = new EqualOperatorNode(attributeId);
//
//				attributeId = 7; //Voltage
//				EqualOperatorNode eonVoltage = new EqualOperatorNode(attributeId);
//				eonVoltage.addNode("2.69",nPath3);
//				eonVoltage.addNode("2.68",nPath4);
//				
//				node nodeVoltage = new node();
//				nodeVoltage.addObject(nodeVoltage.EqualOperatorNodeId,eonVoltage);
//				nodeVoltage.parent = rootDt;
//				
//				eon.addNode("44",nPath2);
//				eon.addNode("48",nodeVoltage);
//				eon.addNode("53",nodeVoltage);
//				
//				
//				rootDt.addObject(rootDt.EqualOperatorNodeId,eon);
//				
//	//////////////////////////////
//	//stream B
//				edt.add(new node());
//				RusterPerStream.add(new HashMap<Integer,RusterNode>());
//				
//				rootDt  = edt.get(1);
//				
//				//////////////////////////////
//				// default path
//				os = new ArrayList<Integer>(3);
//				os.add(2);
//				os.add(3);
//				os.add(0);
//				os.add(4);
//				
//				rootDt.DefaultPath = new RusterNode(os,0);
//				RusterPerStream.get(1).put(0, rootDt.DefaultPath);
//
//				node n = new node();
//				n.addObject(n.RusterId,rootDt.DefaultPath);
//				n.parent = rootDt;
//
//				//////////////////////////////
//				// second path B Stream
//				os = new ArrayList<Integer>(4);
//				os.add(2);
//				os.add(0);
//				os.add(3);
//				os.add(4);
//				
//				RusterNode Path2B =  new RusterNode(os,1);
//				RusterPerStream.get(1).put(1, Path2B);
//
//				node nPath2B = new node();
//				nPath2B.addObject(nPath2B.RusterId,Path2B);
//				nPath2B.parent = rootDt;
//
//				//////////////////////////////
//				// third path B Stream
//				os = new ArrayList<Integer>(4);
//				os.add(3);
//				os.add(0);
//				os.add(2);
//				os.add(4);
//				
//				RusterNode Path3B =  new RusterNode(os,2);
//				RusterPerStream.get(1).put(2, Path3B);
//
//				node nPath3B = new node();
//				nPath3B.addObject(nPath3B.RusterId,Path3B);
//				nPath3B.parent = rootDt;
//
//				//////////////////////////////
//				// add range node
//				 rn = new RangeNode(5,27,32,nPath2B);
//				 rn.AddRangeValue(35, 40, nPath2B);
//				 rn.AddRangeValue(42, 47, nPath3B);
//				 
//				 node nRangeNode = new node();
//				 nRangeNode.addObject(nRangeNode.RangeNodeId,rn);
//				 
//				//////////////////////////////
//				// add equal node
//				
//				attributeId = 3; //moteid
//				eon = new EqualOperatorNode(attributeId);
//				eon.addNode("34",nRangeNode);
//				eon.addNode("38",n);
//				eon.addNode("43",nRangeNode);
//				
//				
//				
//				rootDt.addObject(n.EqualOperatorNodeId,eon);
//
//				nRangeNode.parent = n;
//				
								
//end - create hard coded execution decision tree(s)
				