/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui;


import java.awt.Toolkit;
import java.awt.Dimension;
import java.awt.Point;
import javax.swing.JFrame;
import javax.swing.JPanel;

import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.SystemObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils.PanelMaker;


public class FireStreamGUI extends CAPE_GUIQueryEngine implements Runnable{
	
	private String _GUIName = new String("FireStream");
	
	/** Constructor initializes the <code>CAPE_GUI</code>.
	 * Note that no trees have been loaded at this point. */
	public FireStreamGUI(){
		super();
		this._pm = new PanelMaker(this);
		initGUI();
	}
	
	public FireStreamGUI(XATTree[] trees, String clientName){
		super(trees, clientName);
		initGUI();
	}
		
	public FireStreamGUI(XATTree[] trees, String clientName, SystemObject systemObject, String[] distInfo){
		super(trees,clientName, systemObject, distInfo);
		this._XATTrees = trees;
		this._clientName = clientName;
		this._systemObject = systemObject;
		this._distInfo = (String[])distInfo.clone();

		this._pm = new PanelMaker(this);
		initGUI();
	}
	public String getName(){
		return _GUIName;
	}
	
	public void init(){
		//Set the look and feel
		JFrame.setDefaultLookAndFeelDecorated(false);
		
		//Create and setup window
		JFrame frame = new JFrame("CAPE: " + this._clientName);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		//Create and setup content pane
		FireStreamGUI gui = this;
		gui.setOpaque(true);
		frame.setJMenuBar(gui.getJMenuBar());
		frame.setContentPane(gui);
		
		//Display window
		frame.pack();
		Toolkit tk = Toolkit.getDefaultToolkit();
		frame.setSize(new Dimension(new Double(tk.getScreenSize().getWidth()*0.90).intValue(), 
			new Double(tk.getScreenSize().getHeight()*0.90).intValue()));
		frame.setLocation(new Point(new Double(tk.getScreenSize().getWidth()*0.05).intValue(),
			new Double(tk.getScreenSize().getHeight()*0.04).intValue()));
		frame.setVisible(true);
		frame.toFront();
		
		//Draw the trees
		getSplitPane().setDividerLocation(0.60);
		try{
			for(int i=0; i<getXATTreeCount(); i++){
				getGraphManager().drawModel(getModelMaker().makeModel(getXATTree(i)),getXATTree(i));
			}
		}catch(Exception e){System.err.println("No trees have been loaded.");}
	
		
		getStatisticStorage().initialize();
		makeUpdateQueuesTimer().start();
		makeUpdateNodesTimer().start();
		makeUpdateStatisticTableTimer();
		makeUpdateTreesTimer().start();
	}
	
	/** Main method. */
	public static void main(String[] args){
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			CAPE_GUIQueryEngine gui = new CAPE_GUIQueryEngine();
			
			public void run() {
				gui.init();
			}
		 });
	}
	
	/** Run method. */
	public void run(){
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			 public void run() {
				 init();
			 }
		 });	
	}
}
