/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.initialgui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.Dimension;
import java.awt.Color;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JFormattedTextField;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JTextField;
import javax.swing.JTextArea;
import javax.swing.JTable;
import javax.swing.JScrollPane;
//import javax.swing.JScrollBar;
//import java.awt.Dimension;
import javax.swing.ScrollPaneConstants;
import javax.swing.SpringLayout;
import javax.swing.text.NumberFormatter;
//import javax.swing.BoxLayout;

import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils.SpringUtilities;

/**
 * This is the initial GUI that pops up BEFORE the system starts running.
 *
 * Creation Date: (Aug 18, 2004 3:55:07 PM)
 * @author Natasha Bogdanova
 */
public class CAPEInitialGUI extends CAPE_GUI implements Runnable {

	private String dir;

	public CAPEInitialGUI(String dir) {
		//super();
		this.dir = dir;
		System.out.println("batch file dir: "+dir);
		//System.setProperty("user.dir", dir);
	}

	public String getName() {return "";}
	public void init() {
		// create and setup window
		JFrame frame = new JFrame("CAPE Configuration");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		// create and setup content pane
		CAPEInitialGUI gui = this;
		gui.setOpaque(true);
		frame.setContentPane(gui);

		// insert content
		frame.getContentPane().add(getConfigOptionsPanel(), BorderLayout.CENTER);

		// display window
		frame.pack();
		Toolkit tk = Toolkit.getDefaultToolkit();
		int width = 400;
		int height = 720;
		//frame.setSize(new Dimension(width, height));
		frame.setLocation(new Point(
				new Double(tk.getScreenSize().getWidth()/2 - width/2).intValue(),
				new Double(tk.getScreenSize().getHeight()/2 - height/2).intValue()));
//		frame.pack();
		frame.setVisible(true);
		frame.toFront();
	}

	public JPanel getConfigOptionsPanel() {
		JPanel panel = new JPanel();
		panel.setLayout(new SpringLayout());

		// stream file selection panel
		JPanel streamFileSelectPanel = new JPanel();
		streamFileSelectPanel.setBorder(
			BorderFactory.createTitledBorder(
				BorderFactory.createEmptyBorder(),
				"Select Stream Configuration File"));
		// stream file text field
		JTextField streamFileTextField = new JTextField();
		streamFileTextField.setColumns(35);
		streamFileSelectPanel.add(streamFileTextField);
		// stream file OK button
		//JButton streamOKButton = new JButton("OK");
		//streamOKButton.setActionCommand("STREAM_FILE_OK");
		//streamFileSelectPanel.add(streamOKButton);
		// stream file browse button
		JButton streamBrowseButton = new JButton("Browse");
		streamBrowseButton.setActionCommand("STREAM_FILE_BROWSE");
		//streamBrowseButton.addActionListener(new CAPEInitialGUIListener(streamFileTextField));
		streamFileSelectPanel.add(streamBrowseButton);

		// configuration file selection panel
		JPanel configFileSelectPanel = new JPanel();
		configFileSelectPanel.setBorder(
			BorderFactory.createTitledBorder(
				BorderFactory.createEmptyBorder(),
				"Select System Configuration File"));
		// configuration file text field
		JTextField configFileTextField = new JTextField();
		configFileTextField.setColumns(35);
		configFileSelectPanel.add(configFileTextField);
		// configuration file OK button
		//JButton configOKButton = new JButton("OK");
		//configOKButton.setActionCommand("CONFIG_FILE_OK");
		//configFileSelectPanel.add(configOKButton);
		// configuration file browse button
		JButton configBrowseButton = new JButton("Browse");
		configBrowseButton.setActionCommand("CONFIG_FILE_BROWSE");
		//configBrowseButton.addActionListener(new CAPEInitialGUIListener(configFileTextField));
		configFileSelectPanel.add(configBrowseButton);

		//stream schema display panel
		JPanel streamSchemaDisplayPanel = new JPanel();
		streamSchemaDisplayPanel.setBorder(
				BorderFactory.createTitledBorder(
					BorderFactory.createEmptyBorder(),
					"Stream Schema"));
//		streamSchemaDisplayPanel.setLayout(new BorderLayout());
//		streamSchemaDisplayPanel.setPreferredSize(new Dimension(50, 138));

		String[][] tableData = {
				{"          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          "},
				{"          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          "},
				{"          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          "},
				{"          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          "},
				{"          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          ", "          "}
		};

		String[] columnNames = {"Stream", "Column 1", "Column 2", "Column 3", "Column 4", "Column 5", "Column 6", "Column 7", "Column 8", "Column 9", "Column 10", "Column 11", "Column 12"};

		JTable streamSchemaTable = new JTable(tableData, columnNames);

		//JTable streamSchemaTable = new JTable(5, 13);
		streamSchemaTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		JScrollPane streamSchemaPane = new JScrollPane(streamSchemaTable, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		//streamSchemaPane.setSize(100, 80);
		//streamSchemaPane.repaint();
		streamSchemaPane.setPreferredSize(new Dimension(500, 115));
		streamSchemaDisplayPanel.add(streamSchemaPane, BorderLayout.CENTER);
//		streamSchemaPane.setLocation(100, 0);

		//query text display panel
		JPanel queryTextDisplayPanel = new JPanel();
		queryTextDisplayPanel.setBorder(
				BorderFactory.createTitledBorder(
					BorderFactory.createEmptyBorder(),
					"CAPE-QL Query"));
		//query text area
		JTextArea queryTextArea = new JTextArea(7, 37);
		queryTextDisplayPanel.add(queryTextArea);

		// slider panel
		JPanel sliderPanel = new JPanel();
		sliderPanel.setBorder(
			BorderFactory.createTitledBorder("Adaptation Configuration"));
		SpringLayout sliderLayout = new SpringLayout();
		sliderPanel.setLayout(sliderLayout);
		int sliderMin = 0;
		int sliderMax = 60;
		int sliderInit = (sliderMax - sliderMin) / 2;
		java.text.NumberFormat numberFormat =
			java.text.NumberFormat.getIntegerInstance();
		NumberFormatter formatter = new NumberFormatter(numberFormat);

		// scheduler label
		JLabel schedLabel = new JLabel("Scheduler Interval:");
		schedLabel.setForeground(Color.BLUE);
		sliderPanel.add(schedLabel);
		// scheduler slider
		JSlider schedSlider = new JSlider(JSlider.HORIZONTAL, sliderMin, sliderMax, sliderInit);
		schedSlider.setMajorTickSpacing(10);
		schedSlider.setMinorTickSpacing(2);
		schedSlider.setPaintTicks(true);
		schedSlider.setPaintLabels(true);
		schedSlider.setPaintTrack(true);
		schedLabel.setLabelFor(schedSlider);
		// scheduler text field
		JFormattedTextField schedTextField = new JFormattedTextField(formatter);
		schedTextField.setValue(new Integer(sliderInit));
		schedTextField.setColumns(2); //get some space
		// text field label
		JLabel schedSecLabel = new JLabel("sec");
		// set actions
		schedSlider.addChangeListener(new CAPEInitialGUIListenerC(schedTextField));
		schedTextField.addPropertyChangeListener(new CAPEInitialGUIListenerPC(schedSlider));
		sliderPanel.add(schedSlider);
		sliderPanel.add(schedTextField);
		sliderPanel.add(schedSecLabel);

		// migration panel
		//JPanel migPanel = new JPanel();
		// migration label
		JLabel migLabel = new JLabel("Migration Interval:");
		migLabel.setForeground(Color.BLUE);
		sliderPanel.add(migLabel);
		// migration slider
		JSlider migSlider = new JSlider(JSlider.HORIZONTAL, sliderMin, sliderMax, sliderInit);
		migSlider.setMajorTickSpacing(10);
		migSlider.setMinorTickSpacing(2);
		migSlider.setPaintTicks(true);
		migSlider.setPaintLabels(true);
		migSlider.setPaintTrack(true);
		migLabel.setLabelFor(migSlider);
		// migration text field
		formatter.setMinimum(new Integer(sliderMin));
		formatter.setMaximum(new Integer(sliderMax));
		JFormattedTextField migTextField = new JFormattedTextField(formatter);
		migTextField.setValue(new Integer(sliderInit));
		migTextField.setColumns(2); //get some space
		// text field label
		JLabel migSecLabel = new JLabel("sec");
		//SpringLayout.Constraints migTextFieldCons = sliderLayout.getConstraints(migTextField);
		//migTextFieldCons.setWidth(Spring.constant(1));
		//migTextFieldCons.setHeight(Spring.constant(1));
		// set actions
		migSlider.addChangeListener(new CAPEInitialGUIListenerC(migTextField));
		migTextField.addPropertyChangeListener(new CAPEInitialGUIListenerPC(migSlider));
		sliderPanel.add(migSlider);
		sliderPanel.add(migTextField);
		sliderPanel.add(migSecLabel);

		// distribution manager panel
		//JPanel distPanel = new JPanel();
		// distribution manager label
		JLabel distLabel = new JLabel("Distribution Interval:");
		distLabel.setForeground(Color.BLUE);
		sliderPanel.add(distLabel);
		// distribution manager slider
		JSlider distSlider = new JSlider(JSlider.HORIZONTAL, sliderMin, sliderMax, sliderInit);
		distSlider.setMajorTickSpacing(10);
		distSlider.setMinorTickSpacing(2);
		distSlider.setPaintTicks(true);
		distSlider.setPaintLabels(true);
		distSlider.setPaintTrack(true);
		distLabel.setLabelFor(distSlider);
		// distribution manager text field
		JFormattedTextField distTextField = new JFormattedTextField(formatter);
		distTextField.setValue(new Integer(sliderInit));
		distTextField.setColumns(2); //get some space
		// text field label
		JLabel distSecLabel = new JLabel("sec");
		// set actions
		distSlider.addChangeListener(new CAPEInitialGUIListenerC(distTextField));
		distTextField.addPropertyChangeListener(new CAPEInitialGUIListenerPC(distSlider));
		sliderPanel.add(distSlider);
		sliderPanel.add(distTextField);
		sliderPanel.add(distSecLabel);

		// adjust the slider panel layout
		SpringUtilities.makeCompactGrid(sliderPanel,
										sliderPanel.getComponentCount()/4, 4,	//rows, cols
										5, 5,	//initX, initY
										5, 10);	//xPad, yPad

		// button panel
		JPanel buttonPanel = new JPanel();
		// start button
		JTextField[] startArgs = {streamFileTextField, configFileTextField};
		JSlider[] sliders = {migSlider, distSlider, schedSlider};
		JButton startButton = new JButton("Start");
		startButton.setForeground(Color.RED);
		startButton.setEnabled(false);
		startButton.setActionCommand("START_SYSTEM");
		startButton.addActionListener(new CAPESystemActionListener(startArgs, sliders, this.dir));
		// create listeners for the text fields that will determine when to make
		// the start button active
		JTextField[] depFields = {streamFileTextField, configFileTextField};
		Component[] parents2 = {startButton, streamFileTextField, streamSchemaPane};
		Component[] parents1 = {startButton, configFileTextField, queryTextArea};
		streamBrowseButton.addActionListener(new CAPEInitialGUIListener(parents2, depFields));
		configBrowseButton.addActionListener(new CAPEInitialGUIListener(parents1, depFields));
		//streamOKButton.addActionListener(new CAPEInitialGUIListener(startButton, depFields));
		//configOKButton.addActionListener(new CAPEInitialGUIListener(startButton, depFields));
		buttonPanel.add(startButton);
		// stop button
		JButton stopButton = new JButton("Stop");
		stopButton.setForeground(Color.BLUE);
		stopButton.setActionCommand("STOP_SYSTEM");
		stopButton.addActionListener(new CAPESystemActionListener());
		stopButton.setEnabled(true);
		buttonPanel.add(stopButton);

		panel.add(streamFileSelectPanel);
		panel.add(configFileSelectPanel);
		panel.add(streamSchemaDisplayPanel);
		panel.add(queryTextDisplayPanel);
		panel.add(sliderPanel);
		panel.add(buttonPanel);

		// adjust the layout
		SpringUtilities.makeCompactGrid(panel,
										panel.getComponentCount(), 1,	//rows, cols
										5, 5,	//initX, initY
										5, 5);	//xPad, yPad

		return panel;
	}

	public void run() {
		init();
	}

	public static void main(String[] args) {
		if(args.length != 1) {
			System.out.println("Usage: java CAPEInitialGUI [projectFolderDir (for example: C:/eclipse/Raindrop)] ");
			return;
		}
		Thread guiThread = new Thread(new CAPEInitialGUI(args[0]));
		guiThread.setName("CAPE Initial GUI");
		guiThread.setDaemon(true);
		guiThread.start();
	}
}
