/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.initialgui;

import javax.xml.parsers.*;

import org.xml.sax.*;
import org.w3c.dom.*;

import java.io.*;
import java.util.ArrayList;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JTextField;
import javax.swing.JTextArea;
import javax.swing.JTable;
import javax.swing.JScrollPane;


/**
 * The general action listener for the initial CAPE GUI.
 * 
 * Creation Date: (Aug 18, 2004 3:43:35 PM)
 * @author Natasha Bogdanova
 */
public class CAPEInitialGUIListener implements ActionListener {
	
	private Component[] parents;
	private Component[] depComponents;
	
	/**
	 * The main <code>CAPEInitialGUIListener</code> constructor.
	 */
	public CAPEInitialGUIListener() {
		super();
	}
	
	/**
	 * A constructor for <code>CAPEInitialGUIListener</code>s whose actions 
	 * require a parent component.
	 * 
	 * @param parent Component
	 */
	public CAPEInitialGUIListener(Component parent) {
		super();
		this.parents = new Component[1];
		this.parents[0] = parent;
	}
	
	/**
	 * A constructor for <code>CAPEInitialGUIListener</code>s whose actions 
	 * require a parent component and depend on other components.
	 * 
	 * @param parent Component
	 * @param depComponents Component[]
	 */
	public CAPEInitialGUIListener(Component parent, Component[] depComponents) {
		super();
		this.parents = new Component[1];
		this.parents[0] = parent;
		this.depComponents = depComponents;
	}
	
	/**
	 * A constructor for <code>CAPEInitialGUIListener</code>s whose actions 
	 * require several parent components and depend on other components.
	 * 
	 * @param parent Component[]
	 * @param depComponents Component[]
	 */
	public CAPEInitialGUIListener(Component[] parents, Component[] depComponents) {
		super();
		this.parents = parents;
		this.depComponents = depComponents;
	}

	/**
	 * Responds to general actions performed in the <code>CAPEInitialGUI</code>.
	 * 
	 * @param ActionEvent the event performed
	 */
	public void actionPerformed(ActionEvent event) {
		// get the event's command string
		String action = event.getActionCommand();
		
		if (action == "STREAM_FILE_BROWSE") {
			// open the file browser
			JFileChooser fileBrowser = new JFileChooser();
			fileBrowser.setMultiSelectionEnabled(false);
			fileBrowser.setFileSelectionMode(JFileChooser.FILES_ONLY);
			fileBrowser.setApproveButtonText("OK");
			fileBrowser.setControlButtonsAreShown(true);
			fileBrowser.setDialogTitle("Select Stream Configuration File");
			fileBrowser.setCurrentDirectory(new File(System.getProperty("user.dir")));
			fileBrowser.setFileFilter(new XMLFileFilter());
			fileBrowser.showOpenDialog(this.parents[1]);
			
			// put the selected file in the stream file text field
			if (fileBrowser.getSelectedFile() != null) {
				((JTextField)this.parents[1]).setText(fileBrowser.getSelectedFile().getAbsolutePath());
				// set system property for future browsers
				System.setProperty("user.dir", fileBrowser.getSelectedFile().getPath());
				try {
					ArrayList streamSchemaArray = this.retrieveStreamSchema(fileBrowser.getSelectedFile());
					JScrollPane streamSchemaPane = (JScrollPane)this.parents[2];
					JTable streamSchemaTable = (JTable)streamSchemaPane.getViewport().getView();
					for (int i=0; i<streamSchemaArray.size(); i++) {
						ArrayList streamSchema = (ArrayList)streamSchemaArray.get(i);
						for (int j=0; j<streamSchema.size(); j++) {
							streamSchemaTable.setValueAt((String)streamSchema.get(j), i, j);
						}
					}
				} catch(Exception e) {
						e.printStackTrace();
				}
			}
						
			// enable Start button if all selections are valid
			boolean allValid = true;
			for (int i = 0; i < this.depComponents.length; i++) {
				if (((JTextField)depComponents[i]).getText() == null ||
					!(new File(((JTextField)depComponents[i]).getText()).exists())) {
					allValid = false;
					break;
				}
			}
			if (allValid) {
				((JButton)this.parents[0]).setEnabled(true);
			}
		} else if (action == "CONFIG_FILE_BROWSE") {
			// open the file browser
			JFileChooser fileBrowser = new JFileChooser();
			fileBrowser.setMultiSelectionEnabled(false);
			fileBrowser.setFileSelectionMode(JFileChooser.FILES_ONLY);
			fileBrowser.setApproveButtonText("OK");
			fileBrowser.setControlButtonsAreShown(true);
			fileBrowser.setDialogTitle("Select System Configuration File");
			fileBrowser.setCurrentDirectory(new File(System.getProperty("user.dir")));
			fileBrowser.setFileFilter(new XMLFileFilter());
			fileBrowser.showOpenDialog(this.parents[1]);
				
			// put the selected file in the system configuration file text field
			if (fileBrowser.getSelectedFile() != null) {
				((JTextField)this.parents[1]).setText(fileBrowser.getSelectedFile().getAbsolutePath());
				// set system property for future browsers
				System.setProperty("user.dir", fileBrowser.getSelectedFile().getPath());
				try {
					((JTextArea) this.parents[2]).setText(this.retrieveQueryText(fileBrowser.getSelectedFile()));
				} catch(Exception e) {
					e.printStackTrace();
				}
			}
			
			// enable Start button if all selections are valid
			boolean allValid = true;
			for (int i = 0; i < this.depComponents.length; i++) {
				if (((JTextField)depComponents[i]).getText() == null ||
					!(new File(((JTextField)depComponents[i]).getText()).exists())) {
					allValid = false;
					break;
				}
			}
			if (allValid) {
				((JButton)this.parents[0]).setEnabled(true);
			}
		} 
		/*
		else if (action == "STREAM_FILE_OK" ||
				   action == "CONFIG_FILE_OK") {
			// enable Start button if all selections are valid
			boolean allValid = true;
			for (int i = 0; i < this.depComponents.length; i++) {
				if (((JTextField)depComponents[i]).getText() == null ||
					!(new File(((JTextField)depComponents[i]).getText()).exists())) {
					allValid = false;
					break;
				}
			}
			if (allValid) {
				((JButton)this.parents[0]).setEnabled(true);
			}
		}*/
	}

	private String retrieveQueryText(File configFile) throws ParserConfigurationException, SAXException, IOException {
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();

		DocumentBuilder db = dbf.newDocumentBuilder();

		Document doc = db.parse(configFile);

		NodeList queryPlanNodes = doc.getElementsByTagName("QueryPlan");
		NodeList queryPlanPropertyList = queryPlanNodes.item(0).getChildNodes();
		
		String planTextFileName = "";
		
		for (int i = 0; i < queryPlanPropertyList.getLength(); i++){
			Node next = queryPlanPropertyList.item(i);
			if(next.getNodeType() == Node.ELEMENT_NODE){
				NamedNodeMap attrs = next.getAttributes();
				String name = ((Node) attrs.getNamedItem("name")).getNodeValue();
				if (name.equals("QUERY_FILE_NAME")) {
					planTextFileName = ((Node) attrs.getNamedItem("value")).getNodeValue();
					break;
				} 
			}
		}
		
		if (planTextFileName.length() == 0) {
			return "";
		}
		
		StringBuffer queryTextBuffer = new StringBuffer();
		BufferedInputStream in = new BufferedInputStream(new FileInputStream(planTextFileName));
		char ch = (char) in.read();
		while((int)ch != 65535) {
			queryTextBuffer.append(ch);
			ch = (char) in.read();
		}	
		return queryTextBuffer.toString();
	}
	
	private ArrayList retrieveStreamSchema(File streamConfigFile) throws ParserConfigurationException, SAXException, IOException {
		DocumentBuilderFactory dbf =
			DocumentBuilderFactory.newInstance();
		DocumentBuilder db = dbf.newDocumentBuilder();
		Document doc = db.parse(streamConfigFile);
		NodeList schemaList = doc.getElementsByTagName("schema");
		
		ArrayList streamSchemaArray = new ArrayList();
		for (int i=0; i<schemaList.getLength(); i++) {
			NodeList schemaPropertyList = schemaList.item(i).getChildNodes();
			ArrayList singleStreamSchemaArray = new ArrayList();
			for (int j = 0; j < schemaPropertyList.getLength(); j++){
				Node next = schemaPropertyList.item(j);
				if(next.getNodeType() == Node.ELEMENT_NODE) {
					NamedNodeMap attrs = next.getAttributes();
					String colName = ((Node) attrs.getNamedItem("name")).getNodeValue();
					if (next.getNodeName() == "attribute"){
						String colType = ((Node) attrs.getNamedItem("type")).getNodeValue();
						singleStreamSchemaArray.add(colName+" ("+colType+")");
					} else {
						singleStreamSchemaArray.add(colName);
					}
				}
			}
			streamSchemaArray.add(singleStreamSchemaArray);
		}
		return streamSchemaArray;		
	}

}


	
