/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.initialgui;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.rmi.ConnectException;
import java.util.ArrayList;

import javax.swing.JSlider;
import javax.swing.JTextField;

import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.NameValuePair;

/**
 * The system action listener for the initial CAPE GUI.
 * 
 * Creation Date: (Aug 18, 2004 4:43:12 PM)
 * @author Natasha Bogdanova
 */
public class CAPESystemActionListener implements ActionListener {
	
	/** a list of all the running processes **/
	private static ArrayList processes;
	/** the current working directory **/
	private String dir = "C:/eclipse/workspace/Raindrop";
	//private String dir = "C:/Program Files/eclipse-2.1.3/workspace/Raindrop";
	/** the text fields which were passed in for this listener **/
	private JTextField[] args;
	/** the sliders that specify different configuration intervals **/
	private JSlider[] sliders;

	/**
	 * The main <code>CAPESystemActionListener</code> constructor.
	 */
	public CAPESystemActionListener() {
		super();
		if (processes == null)
			processes = new ArrayList();
	}
	
	/**
	 * A constructor for the <code>CAPESystemActionListener</code>s whose 
	 * actions require some arguments.
	 * 
	 * @param args JTextField[]
	 */
	public CAPESystemActionListener(JTextField[] args, JSlider[] sliders, String newdir) {
		super();
		this.dir = newdir;
		if (processes == null)
			processes = new ArrayList();
		this.args = args;
		this.sliders = sliders;
	}

	/**
	 * Responds to actions performed that require the system to do something.
	 * 
	 * @param ActionEvent the event performed
	 */
	public void actionPerformed(ActionEvent event) {
		// get the event's command string
		String action = event.getActionCommand();
		
		if (action == "START_SYSTEM") {
			for (int i = 0; i < this.args.length; i++) {
				System.out.println(args[i].getText());
			}
			try {				
		  		// get the files containing the commands to run
	            File streamGeneratorFile = new File(dir + "/resources/GUI/StreamGenerator.bat");
				File applicationFile = new File(dir + "/resources/GUI/Application.bat");
				File queryProcessorFile = new File(dir + "/resources/GUI/QueryProcessor.bat");
				File testRunFile = new File(dir + "/resources/GUI/TestRun.bat");
				
				// check file validity
				if (!streamGeneratorFile.exists()) {
					System.out.println("ERROR: Stream Generator file does not exist.");
					System.out.println(" ---> \"" + streamGeneratorFile.getAbsolutePath() + "\"");
					return;
				} else if (!applicationFile.exists()) {
					System.out.println("ERROR: Application file does not exist.");
					System.out.println(" ---> \"" + applicationFile.getAbsolutePath() + "\"");
					return;
				} else if (!queryProcessorFile.exists()) {
					System.out.println("ERROR: Query Processor file does not exist.");
					System.out.println(" ---> \"" + queryProcessorFile.getAbsolutePath() + "\"");
					return;
				} else if (!testRunFile.exists()) {
					System.out.println("ERROR: Test Run file does not exist.");
					System.out.println(" ---> \"" + testRunFile.getAbsolutePath() + "\"");
					return;
				}
				
				// add passed in arguments to streamGeneratorFile and testRunFile
				BufferedReader br = new BufferedReader(new FileReader(streamGeneratorFile));
				String line = br.readLine();				
				line = line + " \"" + args[0].getText() + "\"";
				File streamGeneratorTempFile = new File(dir + "/resources/GUI/StreamGeneratorTemp.bat");
				streamGeneratorTempFile.createNewFile();
				BufferedWriter bw = new BufferedWriter(new FileWriter(streamGeneratorTempFile));
				bw.write(line);
				bw.close();
				
				br = new BufferedReader(new FileReader(testRunFile));
				line = br.readLine();
				line = line + " \"" + args[1].getText() + "\"";
				File testRunTempFile = new File(dir + "/resources/GUI/TestRunTemp.bat");
				testRunTempFile.createNewFile();
				bw = new BufferedWriter(new FileWriter(testRunTempFile));
				bw.write(line);
				bw.close();
				
				// start the processes running
		   		Runtime rt = Runtime.getRuntime();
		   		Process streamGeneratorProc = rt.exec(streamGeneratorTempFile.getAbsolutePath());
		   		Process applicationProc = rt.exec(applicationFile.getAbsolutePath());
		   		Process queryProcessorProc = rt.exec(queryProcessorFile.getAbsolutePath());
		   		Process testRunProc = rt.exec(testRunTempFile.getAbsolutePath());
		   		
		   		// add the processes to the currently running list
		   		processes.add(streamGeneratorProc);
				processes.add(applicationProc);
				processes.add(queryProcessorProc);
				processes.add(testRunProc);
				
				// delete temporary files
				streamGeneratorTempFile.deleteOnExit();
				testRunTempFile.deleteOnExit();
				
				// connect to the server process to send configuration information
				// TODO: change this so it is not hardcoded ---
				InetAddress addr = InetAddress.getLocalHost();
				//Get IP Address
				byte[] ipAddr = addr.getAddress();
				//Convert to dot representation
				String ipAddrStr = "";
				for (int i = 0; i < ipAddr.length; i++) {
					if (i > 0) {
						ipAddrStr += ".";
					}
					ipAddrStr += ipAddr[i] & 0xFF;
				}
				int port = 8005;
				//---
				
				//System.out.println("IP: " + ipAddrStr + ", Port: " + port);
				System.out.println("Connecting to server...");
				try {
					ServerSocket startup = new ServerSocket(port);
					Socket outgoing = startup.accept();
					
					ObjectOutputStream out =
						new ObjectOutputStream(outgoing.getOutputStream());
					System.out.println("Successfully connected to server.");
					// send the slider values to the server in an array of NameValuePairs
					// since values are in seconds, convert to milliseconds first
					NameValuePair[] sliderVals = new NameValuePair[3];
					sliderVals[0] = 
						new NameValuePair("MIGRATION_INTERVAL",
										  Integer.toString(sliders[0].getValue()*1000));
					sliderVals[1] = 
						new NameValuePair("REDISBRIBUTION_TIME",
										  Integer.toString(sliders[1].getValue()*1000));
					sliderVals[2] = 
						new NameValuePair("RESCHEDULING_INTERVAL",
										  Integer.toString(sliders[2].getValue()*1000));
					
					System.out.println("Sending slider information to server...");
					out.writeObject(sliderVals);
						
					System.out.println("Successfully sent slider information to server.");
					outgoing.close();

				} catch (ConnectException ce) {
					System.out.println("Cannot connect to server");
				} catch (Exception e) {
					e.printStackTrace();
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else if (action == "STOP_SYSTEM") {
			// NOTE: this does not work yet
			for (int i = 0; i < processes.size(); i++) {
				System.out.println("Attempting to destroy: " + ((Process)processes.get(i)).toString());
				((Process)processes.get(i)).destroy();
			}
		}
	}

}