/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.listeners;

import java.awt.Cursor;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.storage.OptionsStorage;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils.CAPETab;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils.FramesManager;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils.JTabbedPaneCAPE;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils.PlotFromFile;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils.PlotFromFileQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils.StatsPlotLive;

/** The <code>CAPE_GUI</code>'s main listener for events. */
public class CAPEListener implements ActionListener
{
	//------------------------------------------------------------------------------
	
	/** The <code>CAPE_GUI</code> */	
	private CAPE_GUI _gui;
	
	//------------------------------------------------------------------------------
	
	/** Constructor associates the given <code>gui</code> with this listener. */
	public CAPEListener(CAPE_GUI gui)
	{
		_gui = gui;
	}
	
	//------------------------------------------------------------------------------
	
	/** Handles all events passed to this listener. */
	public void actionPerformed(ActionEvent e) 
	{
		//needed to get current screen resolution
		Toolkit tk = Toolkit.getDefaultToolkit();
		
		//--------------- Open What to Plot Panel ---------------//
		if(e.getActionCommand() == "WHAT_TO_PLOT")
		{
			String c = e.getActionCommand();
			FramesManager fm = _gui.getFramesManager();
			
			//If the frame is already open, then focus it.
			if(fm.isOpenFrame(c))
			{
				fm.toFront(c);
			}
			//Otherwise, create a new frame and display it IF trees have been loaded.
			else
			{
				try
				{
					//if no trees have been loaded, then display error window.
					if(_gui.getXATTreeCount() == 0)
					{
						JOptionPane.showMessageDialog(null, "There are no statistics to plot.",
							"Plot Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
					}
					else
					{
						//Otherwise, create a new frame and display it
						JFrame newFrame = fm.makeNewFrame("WHAT_TO_PLOT", "Select Statistics to Plot", new Point(
							new Double(tk.getScreenSize().getWidth()*0.5-100).intValue(), 
							new Double(tk.getScreenSize().getHeight()*0.5-100).intValue()));
					}
				}
				catch(Exception ex){System.err.println("No trees have been loaded.");
					JOptionPane.showMessageDialog(null, "There are no statistics available.",
						"Display Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);}
			}
						
			//--------------- What to Plot Panel 'Ok' Button ---------------//
		}
		else if(e.getActionCommand() == "WHAT_TO_PLOT_OK")
		{
			ArrayList radios = (ArrayList)_gui.getOptionsStorage().getOptions("WHAT_TO_PLOT");
			for(int i=0; i<radios.size(); i++)
			{
				JRadioButton button = (JRadioButton)radios.get(i);
				if(button.isSelected())
				{
					if(button.getActionCommand().equals("PLOT_TREE_STATS"))
					{
						FramesManager fm = _gui.getFramesManager();
						if(fm.isOpenFrame("PLOT_TREE_STATS"))
						{
							fm.toFront("PLOT_TREE_STATS");
						}
						else
						{
							JFrame newFrame = fm.makeNewFrame("PLOT_TREE_STATS", "Plot Tree Statistics",
								new Point(new Double(tk.getScreenSize().getWidth()/2-325).intValue(), 
								new Double(tk.getScreenSize().getHeight()/2-150).intValue()));
						}
					}
					else if(button.getActionCommand().equals("PLOT_QUEUE_STATS"))
					{
						FramesManager fm = _gui.getFramesManager();
						if(fm.isOpenFrame("PLOT_QUEUE_STATS"))
						{
							fm.toFront("PLOT_QUEUE_STATS");
						}
						else
						{
							JFrame newFrame = fm.makeNewFrame("PLOT_QUEUE_STATS", "Plot Queue Statistics",
								new Point(new Double(tk.getScreenSize().getWidth()/2-275).intValue(), 
								new Double(tk.getScreenSize().getHeight()/2-200).intValue()));
						}						
					}
					else if(button.getActionCommand().equals("PLOT_OPERATOR_STATS"))
					{
						FramesManager fm = _gui.getFramesManager();
						if(fm.isOpenFrame("PLOT_OPERATOR_STATS"))
						{
							fm.toFront("PLOT_OPERATOR_STATS");
						}
						else
						{
							JFrame newFrame = fm.makeNewFrame("PLOT_OPERATOR_STATS", "Plot Operator Statistics",
								new Point(new Double(tk.getScreenSize().getWidth()/2-300).intValue(), 
								new Double(tk.getScreenSize().getHeight()/2-250).intValue()));
						}						
					}
					else if(button.getActionCommand().equals("PLOT_SYSTEM_STATS"))
					{
						FramesManager fm = _gui.getFramesManager();
						if(fm.isOpenFrame("PLOT_SYSTEM_STATS"))
						{
							fm.toFront("PLOT_SYSTEM_STATS");
						}
						else
						{
							JFrame newFrame = fm.makeNewFrame("PLOT_SYSTEM_STATS", "Plot System Statistics",
								new Point(new Double(tk.getScreenSize().getWidth()/2-325).intValue(), 
								new Double(tk.getScreenSize().getHeight()/2-150).intValue()));						
						}
					} 
					else if(button.getActionCommand().equals("PLOT_FILE_STATS"))
					{
						FramesManager fm = _gui.getFramesManager();
						// reset the options
						_gui.getOptionsStorage().setOptions("PLOT_FILE_STATS_OPTIONS", null);

						if(fm.isOpenFrame("PLOT_FILE_STATS"))
						{
							fm.toFront("PLOT_FILE_STATS");
						}
						else
						{
							JFrame newFrame = fm.makeNewFrame("PLOT_FILE_STATS", "Plot Statistics from File",
								new Point(new Double(tk.getScreenSize().getWidth()/2-300).intValue(), 
								new Double(tk.getScreenSize().getHeight()/2-250).intValue()));
						}
					}
				}
			}
			_gui.getFramesManager().removeOpenFrame("WHAT_TO_PLOT");
								
		//--------------- What to Plot Panel 'Cancel' Button --------------//
		}
		else if(e.getActionCommand() == "WHAT_TO_PLOT_CANCEL")
		{
			_gui.getFramesManager().removeOpenFrame("WHAT_TO_PLOT");
						
			//--------------- Open Plot Tree Statistics Panel ----------------//
		}
		else if(e.getActionCommand() == "PLOT_TREE_STATS")
		{
			FramesManager fm = _gui.getFramesManager();
			//if the frame is already open, then focus to it
			if(fm.isOpenFrame("PLOT_TREE_STATS"))
			{
				fm.toFront("PLOT_TREE_STATS");
			}
			else
			{ //Otherwise, create a new frame and display it IF trees have been loaded.
				//If no trees have been loaded, then display error window.
				if(_gui.getXATTreeCount() == 0)
				{
					JOptionPane.showMessageDialog(null, "There are no statistics to plot.",
						"Plot Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
				}
				else
				{//Otherwise, create a new frame and display it
					JFrame newFrame = fm.makeNewFrame("PLOT_TREE_STATS", "Plot Tree Statistics",
						new Point(new Double(tk.getScreenSize().getWidth()/2-320).intValue(), 
						new Double(tk.getScreenSize().getHeight()/2-150).intValue()));
				}
			}	
		
		//--------------- Plot Tree Statistics 'Ok' Button ---------------//
		}
		else if(e.getActionCommand() == "PLOT_TREE_STATS_OK")
		{
			//if no graphs are open, then give error window
			if(_gui.getGraphManager().getGraphTabbedPane().getTabCount() == 0)
			{
				JOptionPane.showMessageDialog(null, "There are no trees available to plot.",
					"Plot Tree Statistics Information", JOptionPane.INFORMATION_MESSAGE);
			}
			else
			{
				try
				{
					FramesManager fm = _gui.getFramesManager();
					JFrame plotOptionsFrame = fm.getFrame("PLOT_TREE_STATS");
			
					//get the user's choices
					ArrayList treeStats = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_TREE_STATS");
					ArrayList trees = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_TREES");
					JTextField freqTextField = (JTextField)_gui.getOptionsStorage().getOptions("PLOT_TREE_STATS_FREQUENCY");
				
					//stores user-selected choices
					ArrayList selectedTreeStats = new ArrayList();
					ArrayList selectedTrees = new ArrayList();
					for(int i=0; i<treeStats.size(); i++)
					{
						if(((JCheckBox)treeStats.get(i)).isSelected())
						{
							selectedTreeStats.add(((JCheckBox)treeStats.get(i)).getText());
					}}
					for(int i=0; i<trees.size(); i++)
					{
						if(((JCheckBox)trees.get(i)).isSelected())
						{
							selectedTrees.add(((JCheckBox)trees.get(i)).getText());
					}}
					
					//for each selected tree, get the actual tree from storage
					ArrayList actualTrees = new ArrayList();
					try{
						for(int i=0; i<selectedTrees.size(); i++){
							String treeID = ((String)selectedTrees.get(i)).substring(1);
							actualTrees.add((XATTree)_gui.getTreeStorage().getTree(treeID));
					}}catch(Exception q){
						JOptionPane.showMessageDialog(plotOptionsFrame, 
						"The XATTree with the given objectID does not exist.", 
						"XATTree Error", JOptionPane.ERROR_MESSAGE);}
				
					//make sure frequency has NO decimals
					if(_gui.checkFreqFormat(freqTextField)){
						plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
						StatsPlotLive plot = new StatsPlotLive(selectedTreeStats, actualTrees, freqTextField.getText());
						JFrame plotFrame = new JFrame("Plot Tree Statistics");
						if(fm.getOpenPlotsCount() > 4){
							fm.resetOpenPlotsCount();
						}
						plotFrame.setLocation(new Double(tk.getScreenSize().getWidth()/3).intValue()-50 + fm.getOpenPlotsCount()*22,
							 fm.getOpenPlotsCount()*22);
						fm.incrementOpenPlotsCount();
						plotFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
						plotFrame.getContentPane().add(plot);
						plotFrame.pack();
						plotFrame.setVisible(true);
						plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
					}else{
						JOptionPane.showMessageDialog(plotOptionsFrame, "Update frequency must be an integer.", 
							"Update Frequency Error", JOptionPane.ERROR_MESSAGE);
						plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
					}
					_gui.getFramesManager().removeOpenFrame("PLOT_TREE_STATS");
				}catch(Exception exp){
					exp.printStackTrace();
					JOptionPane.showMessageDialog(null, "Update frequency must be an integer.", 
						"Update Frequency Error", JOptionPane.ERROR_MESSAGE);
				}
			}
			
		//--------------- Plot Tree Statistics 'Cancel' Button ---------------//
		}
		else if(e.getActionCommand() == "PLOT_TREE_STATS_CANCEL")
		{
			_gui.getFramesManager().removeOpenFrame("PLOT_TREE_STATS");
					
		//--------------- Select All Available Tree Statistics (Plot) ---------------//
		}
		else if(e.getActionCommand() == "PLOT_TREE_STATS_SELECT_ALL")
		{
			ArrayList plotStatBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_TREE_STATS");
			for(int i=0; i<plotStatBoxes.size(); i++)
			{
				((JCheckBox)plotStatBoxes.get(i)).setSelected(true);
			}
					
		//--------------- Deselect All Available Tree Statistics (Plot) ---------------//
		}
		else if(e.getActionCommand() == "PLOT_TREE_STATS_DESELECT_ALL")
		{
			ArrayList plotStatBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_TREE_STATS");
			for(int i=0; i<plotStatBoxes.size(); i++)
			{
				((JCheckBox)plotStatBoxes.get(i)).setSelected(false);
			}
		//--------------- Open Plot File Statistics Panel ----------------//
		// Added by Natasha Bogdanova
		} 
		else if (e.getActionCommand() == "PLOT_FILE_STATS") 
		{
			FramesManager fm = _gui.getFramesManager();
			//if the frame is already open, then focus to it
			if (fm.isOpenFrame("PLOT_FILE_STATS")) 
			{
				fm.toFront("PLOT_FILE_STATS");
			} 
			else 
			{ 
				//Otherwise, create a new frame and display it.
				JFrame newFrame = fm.makeNewFrame("PLOT_FILE_STATS", "Plot Statistics from File",
					new Point(new Double(tk.getScreenSize().getWidth()/2-320).intValue(), 
							  new Double(tk.getScreenSize().getHeight()/2-150).intValue()));
			}
			//--------------- Plot File Statistics 'Browse...' Button ---------------//
		// Added by Natasha Bogdanova
		} 
		else if (e.getActionCommand() == "PLOT_FILE_STATS_BROWSE") 
		{
			FramesManager fm = _gui.getFramesManager();
			OptionsStorage os = _gui.getOptionsStorage();
			// open file browser
			JFileChooser fileBrowser = (JFileChooser)os.getOptions("FILE_BROWSER");
			fileBrowser.setMultiSelectionEnabled(true);
			fileBrowser.setFileSelectionMode(JFileChooser.FILES_ONLY);
			fileBrowser.setApproveButtonText("Select");
			fileBrowser.setControlButtonsAreShown(true);
			fileBrowser.setDialogTitle("Select Statistics File(s)");			
			fileBrowser.showOpenDialog(fm.getFrame("PLOT_FILE_STATS"));
			
			// add the new selections to the text area
			JScrollPane scrollPane = (JScrollPane)os.getOptions("PLOT_FILE_STATS_SCROLL_PANE");
			JTextArea filesTextArea = (JTextArea)scrollPane.getViewport().getView();
			String file = new String();
			for (int i = 0, pos = 0; i < fileBrowser.getSelectedFiles().length; i++, pos++) 
			{
				file = fileBrowser.getSelectedFiles()[i].getPath();
				filesTextArea.insert(file, pos);
				pos += file.length();
				filesTextArea.insert("\n", pos);
			}
			
		//--------------- Plot File Statistics 'Ok' Button ---------------//
		// Added by Natasha Bogdanova
		} else if (e.getActionCommand() == "PLOT_FILE_STATS_OK") 
		{		
			FramesManager fm = _gui.getFramesManager();
			fm.makeNewFrame("PLOT_FILE_STATS_OPTIONS", "Select Statistics to Plot for Each File",
							new Point(new Double(tk.getScreenSize().getWidth()/2-350).intValue(), 
			 				new Double(tk.getScreenSize().getHeight()/2-180).intValue()));
			_gui.getFramesManager().removeOpenFrame("PLOT_FILE_STATS");
			
	
		//--------------- Plot File Statistics 'Cancel' Button ---------------//
		// Added by Natasha Bogdanova
		} else if (e.getActionCommand() == "PLOT_FILE_STATS_CANCEL") 
		{
			_gui.getFramesManager().removeOpenFrame("PLOT_FILE_STATS");
			
		//--------------- Select All Available File Statistics (Plot) ---------------//
		}
		else if(e.getActionCommand() == "PLOT_FILE_STATS_SELECT_ALL")
		{
			ArrayList plotStatBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_FILE_STATS_BOXES");
			for(int i=0; i < plotStatBoxes.size(); i++){
				((JCheckBox)plotStatBoxes.get(i)).setSelected(true);
			}
				
		//--------------- Deselect All Available File Statistics (Plot) ---------------//
		}else if(e.getActionCommand() == "PLOT_FILE_STATS_DESELECT_ALL"){
			ArrayList plotStatBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_FILE_STATS_BOXES");
			for(int i=0; i<plotStatBoxes.size(); i++){
				((JCheckBox)plotStatBoxes.get(i)).setSelected(false);
			}
		//--------------- Plot File Statistics Option Select Box ---------------//
		// Added by Natasha Bogdanova
		} 
		else if (e.getActionCommand() == "PLOT_FILE_SELECT_TYPE") 
		{
			//--------------- Plot File Statistics Options 'Ok' Button ---------------//
			// Added by Natasha Bogdanova
		} 
		else if (e.getActionCommand() == "PLOT_FILE_OPTION_STATS_OK") 
		{
			JFrame plotFileFrame = new JFrame();				
			plotFileFrame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			FramesManager fm = _gui.getFramesManager();
			OptionsStorage os = _gui.getOptionsStorage();
			JFrame plotOptionsFrame = fm.getFrame("PLOT_FILE_STATS_OPTIONS");
			
			// get the user's choices
			ArrayList fileStats = (ArrayList)os.getOptions("PLOT_FILE_STATS_BOXES");
			Hashtable queryObjects = (Hashtable)os.getOptions("PLOT_FILE_STATS_QOBJECTS");
			JTextField freqTextField = (JTextField)os.getOptions("PLOT_FILE_STATS_FREQUENCY");
				
			// store user-selected choices
			ArrayList selectedFileStats = new ArrayList();
			ArrayList selectedQueryObjects = new ArrayList();
			for (int i = 0; i < fileStats.size(); i++) 
			{
				if (((JCheckBox)fileStats.get(i)).isSelected()) 
				{
					selectedFileStats.add(((JCheckBox)fileStats.get(i)).getText());
				}
			}
			for (Enumeration en = queryObjects.keys(); en.hasMoreElements();) 
			{
				PlotFromFileQueryObject qo = (PlotFromFileQueryObject)en.nextElement();
				JCheckBox qoBox = (JCheckBox)queryObjects.get(qo);
				if (qoBox.isSelected()) 
				{
					selectedQueryObjects.add(qo);
				}
			}
				
			// make sure frequency has NO decimals
			if (_gui.checkFreqFormat(freqTextField)) 
			{
				plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				StatsPlotLive plot = new StatsPlotLive(selectedFileStats, selectedQueryObjects, freqTextField.getText());
				JFrame plotFrame = new JFrame("Plot Tree Statistics");
				if(fm.getOpenPlotsCount() > 4)
				{
					fm.resetOpenPlotsCount();
				}
				plotFrame.setLocation(new Double(tk.getScreenSize().getWidth()/3).intValue()-50 + fm.getOpenPlotsCount()*22,
					 fm.getOpenPlotsCount()*22);
				fm.incrementOpenPlotsCount();
				plotFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
				plotFrame.getContentPane().add(plot);
				plotFrame.pack();
				plotFrame.setVisible(true);
				plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
			}
			else 
			{
				JOptionPane.showMessageDialog(plotOptionsFrame, "Update frequency must be an integer.", 
					"Update Frequency Error", JOptionPane.ERROR_MESSAGE);
				plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
			}
			_gui.getFramesManager().removeOpenFrame("PLOT_FILE_STATS_OPTIONS");

		//--------------- Plot File Statistics 'Cancel' Button ---------------//
		// Added by Natasha Bogdanova
		} 
		else if (e.getActionCommand() == "PLOT_FILE_OPTION_STATS_CANCEL") 
		{
			_gui.getFramesManager().removeOpenFrame("PLOT_FILE_STATS_OPTIONS");
			
		//--------------- Open Plot Queue Statistics Panel ----------------//
		}
		else if(e.getActionCommand() == "PLOT_QUEUE_STATS")
		{
			FramesManager fm = _gui.getFramesManager();
			//if the frame is already open, then focus to it
			if(fm.isOpenFrame("PLOT_QUEUE_STATS"))
			{
				fm.toFront("PLOT_QUEUE_STATS");
			}
			else
			{ //Otherwise, create a new frame and display it IF trees have been loaded.
				//If no trees have been loaded, then display error window.
				if(_gui.getXATTreeCount() == 0){
					JOptionPane.showMessageDialog(null, "There are no statistics to plot.",
						"Plot Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
				}else{//Otherwise, create a new frame and display it
					JFrame newFrame = fm.makeNewFrame("PLOT_QUEUE_STATS", "Plot Queue Statistics",
						new Point(new Double(tk.getScreenSize().getWidth()/2-275).intValue(), 
						new Double(tk.getScreenSize().getHeight()/2-200).intValue()));
				}
			}	
				
		//--------------- Plot Queue Statistics 'Ok' Button ---------------//
		}
		else if(e.getActionCommand() == "PLOT_QUEUE_STATS_OK")
		{
			//if no graphs are open, then give error window
			if(_gui.getGraphManager().getGraphTabbedPane().getTabCount() == 0)
			{
				JOptionPane.showMessageDialog(null, "There are no queues available to plot.",
					"Plot Queue Statistics Information", JOptionPane.INFORMATION_MESSAGE);
			}
			else
			{
				try{FramesManager fm = _gui.getFramesManager();
					JFrame plotOptionsFrame = fm.getFrame("PLOT_QUEUE_STATS");
				
					//get the user's choices
					ArrayList queueStats = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_QUEUE_STATS");
					ArrayList queues = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_QUEUES");
					JTextField freqTextField = (JTextField)_gui.getOptionsStorage().getOptions("PLOT_QUEUE_STATS_FREQUENCY");
					
					//stores user-selected choices
					ArrayList selectedQueueStats = new ArrayList();
					ArrayList selectedQueues = new ArrayList();
					for(int i=0; i<queueStats.size(); i++){
						if(((JCheckBox)queueStats.get(i)).isSelected()){
							selectedQueueStats.add(((JCheckBox)queueStats.get(i)).getText());
					}}
					for(int i=0; i<queues.size(); i++){
						if(((JCheckBox)queues.get(i)).isSelected()){
							selectedQueues.add(((JCheckBox)queues.get(i)).getText());
					}}
					
					//for each selected queue, get the actual queue from storage
					ArrayList actualQs = new ArrayList();
					boolean okToPlot = true;
					try{
						for(int i=0; i<selectedQueues.size(); i++){
							String qID = (String)selectedQueues.get(i);
							actualQs.add((XATQueue)_gui.getTreeStorage().getQueue(qID));
					}}catch(Exception q){
						JOptionPane.showMessageDialog(plotOptionsFrame, 
						"The XATQueue with the given objectID does not exist.", 
						"XATQueue Error", JOptionPane.ERROR_MESSAGE);
						okToPlot = false;}
					
					//make sure frequency has NO decimals
					if(okToPlot){
						if(_gui.checkFreqFormat(freqTextField)){
							plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
							StatsPlotLive plot = new StatsPlotLive(selectedQueueStats, actualQs, freqTextField.getText());
							JFrame plotFrame = new JFrame("Plot Queue Statistics");
							if(fm.getOpenPlotsCount() > 4){
								fm.resetOpenPlotsCount();
							}
							plotFrame.setLocation(new Double(tk.getScreenSize().getWidth()/3).intValue()-50 + fm.getOpenPlotsCount()*22,
								fm.getOpenPlotsCount()*22);
							fm.incrementOpenPlotsCount();
							plotFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
							plotFrame.getContentPane().add(plot);
							plotFrame.pack();
							plotFrame.setVisible(true);
							plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
						}else{
							JOptionPane.showMessageDialog(plotOptionsFrame, "Update frequency must be an integer.", 
								"Update Frequency Error", JOptionPane.ERROR_MESSAGE);
							plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
						}
					}
					_gui.getFramesManager().removeOpenFrame("PLOT_QUEUE_STATS");
				}catch(Exception exp){exp.printStackTrace();}
			}
			
		//--------------- Plot Queue Statistics 'Cancel' Button ---------------//
		}
		else if(e.getActionCommand() == "PLOT_QUEUE_STATS_CANCEL")
		{
			_gui.getFramesManager().removeOpenFrame("PLOT_QUEUE_STATS");
			
		//--------------- Select All Available Queue Statistics (Plot) ---------------//
		}
		else if(e.getActionCommand() == "PLOT_QUEUE_STATS_SELECT_ALL")
		{
			ArrayList plotStatBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_QUEUE_STATS");
			for(int i=0; i<plotStatBoxes.size(); i++)
			{
				((JCheckBox)plotStatBoxes.get(i)).setSelected(true);
			}
					
		//--------------- Deselect All Available Queue Statistics (Plot) ---------------//
		}
		else if(e.getActionCommand() == "PLOT_QUEUE_STATS_DESELECT_ALL")
		{
			ArrayList plotStatBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_QUEUE_STATS");
			for(int i=0; i<plotStatBoxes.size(); i++)
			{
				((JCheckBox)plotStatBoxes.get(i)).setSelected(false);
			}
			
		//--------------- Open Plot Operator Statistics Panel ----------------//
		}
		else if(e.getActionCommand() == "PLOT_OPERATOR_STATS")
		{
			FramesManager fm = _gui.getFramesManager();
			//if the frame is already open, then focus to it
			if(fm.isOpenFrame("PLOT_OPERATOR_STATS"))
			{
				fm.toFront("PLOT_OPERATOR_STATS");
			}
			else
			{ //Otherwise, create a new frame and display it IF trees have been loaded.
				//If no trees have been loaded, then display error window.
				if(_gui.getXATTreeCount() == 0){
					JOptionPane.showMessageDialog(null, "There are no statistics to plot.",
						"Plot Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
				}else{//Otherwise, create a new frame and display it
					JFrame newFrame = fm.makeNewFrame("PLOT_OPERATOR_STATS", "Plot Operator Statistics",
						new Point(new Double(tk.getScreenSize().getWidth()/2-300).intValue(), 
						new Double(tk.getScreenSize().getHeight()/2-250).intValue()));
				}
			}
			
		//--------------- Plot Operator Statistics 'Ok' Button ---------------//
		}
		else if(e.getActionCommand() == "PLOT_OPERATOR_STATS_OK")
		{
			//if no graphs are open, then give error window
			if(_gui.getGraphManager().getGraphTabbedPane().getTabCount() == 0){
				JOptionPane.showMessageDialog(null, "There are no operators available to plot.",
					"Plot Operator Statistics Information", JOptionPane.INFORMATION_MESSAGE);
			}else{
				try{FramesManager fm = _gui.getFramesManager();
					JFrame plotOptionsFrame = fm.getFrame("PLOT_OPERATOR_STATS");
				
					//get the user's choices
					ArrayList opStats = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_OPERATOR_STATS");
					ArrayList ops = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_OPERATORS");
					JTextField freqTextField = (JTextField)_gui.getOptionsStorage().getOptions("PLOT_OPERATOR_STATS_FREQUENCY");
					
					//stores user-selected choices
					ArrayList selectedOpStats = new ArrayList();
					ArrayList selectedOps = new ArrayList();
					for(int i=0; i<opStats.size(); i++){
						if(((JCheckBox)opStats.get(i)).isSelected()){
							selectedOpStats.add(((JCheckBox)opStats.get(i)).getText());
					}}
					for(int i=0; i<ops.size(); i++){
						if(((JCheckBox)ops.get(i)).isSelected()){
							selectedOps.add(((JCheckBox)ops.get(i)).getText());
					}}
					
					//for each selected operator, get the actual operator from storage
					ArrayList actualOps = new ArrayList();
					try{
						for(int i=0; i<selectedOps.size(); i++){
							String opID = (String)selectedOps.get(i);
							actualOps.add((XATStreamOperator)_gui.getTreeStorage().getOperator(opID));
					}}catch(Exception q){
						JOptionPane.showMessageDialog(plotOptionsFrame, 
						"The XATStreamOperator with the given objectID does not exist.", 
						"XATStreamOperator Error", JOptionPane.ERROR_MESSAGE);}
			
					//make sure frequency has NO decimals
					if(_gui.checkFreqFormat(freqTextField)){
						plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
						StatsPlotLive plot = new StatsPlotLive(selectedOpStats, actualOps, freqTextField.getText());
						JFrame plotFrame = new JFrame("Plot Operator Statistics");
						if(fm.getOpenPlotsCount() > 4){
							fm.resetOpenPlotsCount();
						}
						plotFrame.setLocation(new Double(tk.getScreenSize().getWidth()/3).intValue()-50 + fm.getOpenPlotsCount()*22,
							fm.getOpenPlotsCount()*22);
						fm.incrementOpenPlotsCount();
						plotFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
						plotFrame.getContentPane().add(plot);
						plotFrame.pack();
						plotFrame.setVisible(true);
						plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
					}else{
						JOptionPane.showMessageDialog(plotOptionsFrame, "Update frequency must be an integer.", 
							"Update Frequency Error", JOptionPane.ERROR_MESSAGE);
						plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
					}
					_gui.getFramesManager().removeOpenFrame("PLOT_OPERATOR_STATS");
				}catch(Exception exp){
					exp.printStackTrace();
					JOptionPane.showMessageDialog(null, "Update frequency must be an integer.", 
						"Update Frequency Error", JOptionPane.ERROR_MESSAGE);
				}
			}
				
		//--------------- Plot Operator Statistics 'Cancel' Button ---------------//
		}else if(e.getActionCommand() == "PLOT_OPERATOR_STATS_CANCEL"){
			_gui.getFramesManager().removeOpenFrame("PLOT_OPERATOR_STATS");
			
		//--------------- Select All Available Operator Statistics (Plot) ---------------//
		}else if(e.getActionCommand() == "PLOT_OPERATOR_STATS_SELECT_ALL"){
			ArrayList plotStatBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_OPERATOR_STATS");
			for(int i=0; i<plotStatBoxes.size(); i++){
				((JCheckBox)plotStatBoxes.get(i)).setSelected(true);
			}
					
		//--------------- Deselect All Available Operator Statistics (Plot) ---------------//
		}else if(e.getActionCommand() == "PLOT_OPERATOR_STATS_DESELECT_ALL"){
			ArrayList plotStatBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_OPERATOR_STATS");
			for(int i=0; i<plotStatBoxes.size(); i++){
				((JCheckBox)plotStatBoxes.get(i)).setSelected(false);
			}
			
		//--------------- Open Plot System Statistics Panel ----------------//
		}else if(e.getActionCommand() == "PLOT_SYSTEM_STATS"){
			FramesManager fm = _gui.getFramesManager();
			//if the frame is already open, then focus to it
			if(fm.isOpenFrame("PLOT_SYSTEM_STATS")){
				fm.toFront("PLOT_SYSTEM_STATS");
			}else{ //Otherwise, create a new frame and display it IF trees have been loaded.
				//If no trees have been loaded, then display error window.
				if(_gui.getXATTreeCount() == 0){
					JOptionPane.showMessageDialog(null, "There are no statistics to plot.",
						"Plot Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
				}else{//Otherwise, create a new frame and display it
					JFrame newFrame = fm.makeNewFrame("PLOT_SYSTEM_STATS", "Plot System Statistics",
						new Point(new Double(tk.getScreenSize().getWidth()/2-325).intValue(), 
						new Double(tk.getScreenSize().getHeight()/2-150).intValue()));
				}
			}
						
		//--------------- Plot System Statistics 'Ok' Button ---------------//
		}else if(e.getActionCommand() == "PLOT_SYSTEM_STATS_OK"){
			//if no graphs are open, then give error window
			if(_gui.getGraphManager().getGraphTabbedPane().getTabCount() == 0){
				JOptionPane.showMessageDialog(null, "There are no system statistics available to plot.",
					"Plot System Statistics Information", JOptionPane.INFORMATION_MESSAGE);
			}else{
				try{FramesManager fm = _gui.getFramesManager();
					JFrame plotOptionsFrame = fm.getFrame("PLOT_SYSTEM_STATS");
				
					//get the user's choices
					ArrayList sysStats = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_SYSTEM_STATS");
					JTextField freqTextField = (JTextField)_gui.getOptionsStorage().getOptions("PLOT_SYSTEM_STATS_FREQUENCY");
					
					//stores user-selected choices
					ArrayList selectedSysStats = new ArrayList();
					for(int i=0; i<sysStats.size(); i++){
						if(((JCheckBox)sysStats.get(i)).isSelected()){
							selectedSysStats.add(((JCheckBox)sysStats.get(i)).getText());
					}}
				
					//make sure frequency has NO decimals
					if(_gui.checkFreqFormat(freqTextField)){
						plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
						StatsPlotLive plot = new StatsPlotLive(selectedSysStats, _gui.getSystemObject(), freqTextField.getText());
						JFrame plotFrame = new JFrame("Plot System Statistics");
						if(fm.getOpenPlotsCount() > 4){
							fm.resetOpenPlotsCount();
						}
						plotFrame.setLocation(new Double(tk.getScreenSize().getWidth()/3).intValue()-50 + fm.getOpenPlotsCount()*22,
							fm.getOpenPlotsCount()*22);
						fm.incrementOpenPlotsCount();
						plotFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
						plotFrame.getContentPane().add(plot);
						plotFrame.pack();
						plotFrame.setVisible(true);
						plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
					}else{
						JOptionPane.showMessageDialog(plotOptionsFrame, "Update frequency must be an integer.", 
							"Update Frequency Error", JOptionPane.ERROR_MESSAGE);
						plotOptionsFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
					}
					_gui.getFramesManager().removeOpenFrame("PLOT_SYSTEM_STATS");
				}catch(Exception exp){
					exp.printStackTrace();
					JOptionPane.showMessageDialog(null, "Update frequency must be an integer.", 
						"Update Frequency Error", JOptionPane.ERROR_MESSAGE);
				}
			}
			
		//--------------- Plot System Statistics 'Cancel' Button ---------------//
		}else if(e.getActionCommand() == "PLOT_SYSTEM_STATS_CANCEL"){
			_gui.getFramesManager().removeOpenFrame("PLOT_SYSTEM_STATS");
			
		//--------------- Select All Available System Statistics (Plot) ---------------//
		}else if(e.getActionCommand() == "PLOT_SYSTEM_STATS_SELECT_ALL"){
			ArrayList plotStatBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_SYSTEM_STATS");
			for(int i=0; i<plotStatBoxes.size(); i++){
				((JCheckBox)plotStatBoxes.get(i)).setSelected(true);
			}
					
		//--------------- Deselect All Available System Statistics (Plot) ---------------//
		}else if(e.getActionCommand() == "PLOT_SYSTEM_STATS_DESELECT_ALL"){
			ArrayList plotStatBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("PLOT_SELECT_SYSTEM_STATS");
			for(int i=0; i<plotStatBoxes.size(); i++){
				((JCheckBox)plotStatBoxes.get(i)).setSelected(false);
			}		
					
		//--------------- Plot No Trees Error 'Ok' Button ----------------//
		}else if(e.getActionCommand() == "PLOT_TREE_STATS_OK_CLOSE"){
			_gui.getFramesManager().removeOpenFrame("PLOT_TREE_STATS");
			
		//--------------- Plot No Trees Error 'Ok' Button ----------------//
		}else if(e.getActionCommand() == "PLOT_QUEUE_STATS_OK_CLOSE"){
			_gui.getFramesManager().removeOpenFrame("PLOT_QUEUE_STATS");
			
		//--------------- Plot No Trees Error 'Ok' Button ----------------//
		}else if(e.getActionCommand() == "PLOT_OPERATOR_STATS_OK_CLOSE"){
			_gui.getFramesManager().removeOpenFrame("PLOT_OPERATOR_STATS");
			
		//--------------- Plot No Trees Error 'Ok' Button ----------------//
		}else if(e.getActionCommand() == "PLOT_SYSTEM_STATS_OK_CLOSE"){
			_gui.getFramesManager().removeOpenFrame("PLOT_SYSTEM_STATS");
		
		//--------------- Open Select Tree Statistics Panel ----------------//
		}else if(e.getActionCommand() == "SELECT_TREE_STATS"){
			String c = e.getActionCommand();
			FramesManager fm = _gui.getFramesManager();
			//If the frame is already open, then focus it.
			if(fm.isOpenFrame(c)){
				fm.toFront(c);
			}else{ //Otherwise, create a new frame and display it IF trees have been loaded.
				//If no trees have been loaded, then display error window.
				if(_gui.getXATTreeCount() == 0){
					JOptionPane.showMessageDialog(null, "There are no statistics available.",
								"Select Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
				}else{//Otherwise, create a new frame and display it
					JFrame newFrame = fm.makeNewFrame("SELECT_TREE_STATS", "Select Tree Statistics", new Point(
						new Double(tk.getScreenSize().getWidth()/2-155).intValue(), 
						new Double(tk.getScreenSize().getHeight()/2-150).intValue()));
				}
			}

		//--------------- Select All Available Tree Statistics ---------------//
		}else if(e.getActionCommand() == "TREE_STATS_SELECT_ALL"){
			ArrayList statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_TREE_STATS");
			for(int i=0; i<statBoxes.size(); i++){
				((JCheckBox)statBoxes.get(i)).setSelected(true);
			}					
					
		//--------------- Deselect All Available Tree Statistics ---------------//
		}else if(e.getActionCommand() == "TREE_STATS_DESELECT_ALL"){
			ArrayList statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_TREE_STATS");
			for(int i=0; i<statBoxes.size(); i++){
				((JCheckBox)statBoxes.get(i)).setSelected(false);
			}
			
		//--------------- Select Tree Statistics Ok Button ---------------//
		}else if(e.getActionCommand() == "TREE_STATS_OK"){
			_gui.getFramesManager().removeOpenFrame("SELECT_TREE_STATS");
			
		//--------------- Display Selected Tree Statistics in Table ----------------//
		}
		else if(e.getActionCommand() == "DISPLAY_TREE_STATS")
		{
			try
			{
				//if there are no trees loaded, then display error
				if(_gui.getXATTreeCount() == 0){
					JOptionPane.showMessageDialog(null, "There are no statistics available.",
						"Display Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
				}else{//otherwise, display tree statistics table
					JTabbedPaneCAPE statsTabbedPane = _gui.getGraphManager().getStatsTabbedPane();
					JTabbedPaneCAPE graphTabbedPane = _gui.getGraphManager().getGraphTabbedPane();
				
					//Construct the tab title to compare to the open tabs
					String title = graphTabbedPane.getTitleAt(graphTabbedPane.getSelectedIndex());
					boolean makeNewTab = true;
			
					//if the tab is already open, then focus to that tab
					for(int i=0; i<statsTabbedPane.getTabCount(); i++){
						if(statsTabbedPane.getTitleAt(i).equals(title)){
							statsTabbedPane.setSelectedIndex(i);
							makeNewTab = false;
							break;
						}
					}
					if(makeNewTab){
						XATTree tree = (XATTree)((CAPETab)_gui.getGraphManager().getGraphTabbedPane().getIconAt(
							_gui.getGraphManager().getGraphTabbedPane().getSelectedIndex())).getXATQueryObject();
						JScrollPane tablePane = _gui.getTableMaker().makeTable(tree);
						String treeTitle = new String("T"+tree.getProperty("TREE_ID"));
						statsTabbedPane.addTab(treeTitle, tree, tablePane, "STATS");
						statsTabbedPane.setSelectedIndex(statsTabbedPane.getTabCount()-1);
						_gui.getUpdateStatisticTableTimer().start();
					}
				}
			}catch(Exception exp){System.err.println("No trees have been loaded.");
				JOptionPane.showMessageDialog(null, "There are no statistics available.",
					"Display Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);}
		
		//--------------- Open Select Queue Statistics Panel ---------------//
		}else if(e.getActionCommand() == "SELECT_QUEUE_STATS")
		{
			String c = e.getActionCommand();
			FramesManager fm = _gui.getFramesManager();
			//If the frame is already open, then focus it.
			if(fm.isOpenFrame(c)){
				fm.toFront(c);
			}else{ //Otherwise, create a new frame and display it IF trees have been loaded.
				//if no trees have been loaded, then display error window.
				if(_gui.getXATTreeCount() == 0){
					JOptionPane.showMessageDialog(null, "There are no statistics available.",
						"Select Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
				}else{//Otherwise, create a new frame and display it
					JFrame newFrame = fm.makeNewFrame("SELECT_QUEUE_STATS", "Select Queue Statistics", new Point(
						new Double(tk.getScreenSize().getWidth()/2-135).intValue(), 
						new Double(tk.getScreenSize().getHeight()/2-100).intValue()));
				}
			}										

		//--------------- Select All Available Queue Statistics ---------------//
		}else if(e.getActionCommand() == "QUEUE_STATS_SELECT_ALL"){
			ArrayList statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_QUEUE_STATS");
			for(int i=0; i<statBoxes.size(); i++){
				((JCheckBox)statBoxes.get(i)).setSelected(true);
			}					
					
		//--------------- Deselect All Available Queue Statistics ---------------//
		}else if(e.getActionCommand() == "QUEUE_STATS_DESELECT_ALL"){
			ArrayList statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_QUEUE_STATS");
			for(int i=0; i<statBoxes.size(); i++){
				((JCheckBox)statBoxes.get(i)).setSelected(false);
			}
			
		//--------------- Select Queue Statistics Ok Button ---------------//
		}else if(e.getActionCommand() == "QUEUE_STATS_OK"){
			_gui.getFramesManager().removeOpenFrame("SELECT_QUEUE_STATS");
				
		//--------------- Open Select Operator Statistics Panel ---------------//
		}else if(e.getActionCommand() == "SELECT_OPERATOR_STATS"){
			String c = e.getActionCommand();
			FramesManager fm = _gui.getFramesManager();
			//If the frame is already open, then focus it.
			if(fm.isOpenFrame(c)){
				fm.toFront(c);
			}else{ //Otherwise, create a new frame and display it IF trees have been loaded.
					//If no trees have been loaded, then display error window.
					if(_gui.getXATTreeCount() == 0){
						JOptionPane.showMessageDialog(null, "There are no statistics available.",
									"Select Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
					}else{//Otherwise, create a new frame and display it
						JFrame newFrame = fm.makeNewFrame("SELECT_OPERATOR_STATS", "Select Operator Statistics", new Point(
							new Double(tk.getScreenSize().getWidth()/2-140).intValue(), 
							new Double(tk.getScreenSize().getHeight()/2-265).intValue()));
					}
			}
					
		//--------------- Select All Available Operator Statistics ---------------//
		}else if(e.getActionCommand() == "OPERATOR_STATS_SELECT_ALL"){
			ArrayList statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_OPERATOR_STATS");
			for(int i=0; i<statBoxes.size(); i++){
				((JCheckBox)statBoxes.get(i)).setSelected(true);
			}					
					
		//--------------- Deselect All Available Operator Statistics ---------------//
		}else if(e.getActionCommand() == "OPERATOR_STATS_DESELECT_ALL"){
			ArrayList statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_OPERATOR_STATS");
			for(int i=0; i<statBoxes.size(); i++){
				((JCheckBox)statBoxes.get(i)).setSelected(false);
			}
			
		//--------------- Select Operator Statistics Ok Button ---------------//
		}else if(e.getActionCommand() == "OPERATOR_STATS_OK"){
			_gui.getFramesManager().removeOpenFrame("SELECT_OPERATOR_STATS");
				
		//--------------- Open Select System Statistics Panel ---------------//
		}else if(e.getActionCommand() == "SELECT_SYSTEM_STATS"){
			String c = e.getActionCommand();
			FramesManager fm = _gui.getFramesManager();
			//If the frame is already open, then focus it.
			if(fm.isOpenFrame(c)){
				fm.toFront(c);
			}else{ //Otherwise, create a new frame and display it IF trees have been loaded.
					//If no trees have been loaded, then display error window.
					if(_gui.getXATTreeCount() == 0){
						JOptionPane.showMessageDialog(null, "There are no statistics available.",
									"Select Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
					}else{//Otherwise, create a new frame and display it
						JFrame newFrame = fm.makeNewFrame("SELECT_SYSTEM_STATS", "Select System Statistics", new Point(
							new Double(tk.getScreenSize().getWidth()/2-155).intValue(), 
							new Double(tk.getScreenSize().getHeight()/2-150).intValue()));
					}
			}	

		//--------------- Select All Available System Statistics ---------------//
		}else if(e.getActionCommand() == "SYSTEM_STATS_SELECT_ALL"){
			ArrayList statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_SYSTEM_STATS");
			for(int i=0; i<statBoxes.size(); i++){
				((JCheckBox)statBoxes.get(i)).setSelected(true);
			}					
					
		//--------------- Deselect All Available System Statistics ---------------//
		}else if(e.getActionCommand() == "SYSTEM_STATS_DESELECT_ALL"){
			ArrayList statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_SYSTEM_STATS");
			for(int i=0; i<statBoxes.size(); i++){
				((JCheckBox)statBoxes.get(i)).setSelected(false);
			}
			
		//--------------- Select System Statistics Ok Button ---------------//
		}else if(e.getActionCommand() == "SYSTEM_STATS_OK"){
			_gui.getFramesManager().removeOpenFrame("SELECT_SYSTEM_STATS");
			
		//--------------- Display Selected System Statistics in Table ---------------//
		}else if(e.getActionCommand() == "DISPLAY_SYSTEM_STATS"){
			try{
				//if there are no trees loaded, then display error
				if(_gui.getXATTreeCount() == 0){
					JOptionPane.showMessageDialog(null, "There are no statistics available.",
						"Display Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
				}else{//otherwise, display system statistics table
					JTabbedPaneCAPE statsTabbedPane = _gui.getGraphManager().getStatsTabbedPane();
					JTabbedPaneCAPE graphTabbedPane = _gui.getGraphManager().getGraphTabbedPane();
				
					boolean makeNewTab = true;
			
					//if the tab is already open, then focus to that tab
					for(int i=0; i<statsTabbedPane.getTabCount(); i++){
						if(statsTabbedPane.getTitleAt(i).equals("System")){
							statsTabbedPane.setSelectedIndex(i);
							makeNewTab = false;
							break;
						}
					}
					if(makeNewTab){
						JScrollPane tablePane = _gui.getTableMaker().makeTable(_gui.getSystemObject());
						statsTabbedPane.addTab("System", _gui.getSystemObject(), tablePane, "STATS");
						statsTabbedPane.setSelectedIndex(statsTabbedPane.getTabCount()-1);
						_gui.getUpdateStatisticTableTimer().start();
					}
				}
			}catch(Exception exp2){System.err.println("No trees have been loaded.");
				JOptionPane.showMessageDialog(null, "There are no statistics available.",
						"Display Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);}
		//--------------- Open Customize Graph Panel ---------------//
		}else if(e.getActionCommand() == "CUSTOMIZE"){
			String c = e.getActionCommand();
			FramesManager fm = _gui.getFramesManager();
			//If the frame is already open, then focus it.
			if(fm.isOpenFrame(c)){
				fm.toFront(c);
			}else{ //Otherwise, create a new frame and display it IF trees have been loaded.
				//if no trees have been loaded, display error window
				if(_gui.getXATTreeCount() == 0){
					JOptionPane.showMessageDialog(null, "There are no graphs available to customize.",
						"Customize Graph Information", JOptionPane.INFORMATION_MESSAGE);
				}else{//otherwise, create a new frame and display it
				fm.makeNewFrame("CUSTOMIZE", "Customize Graph", new Point(
					new Double(tk.getScreenSize().getWidth()/2-110).intValue(), 
					new Double(tk.getScreenSize().getHeight()/2-100).intValue()));
				}
			}
		//--------------- Apply Customization ---------------//
		}else if(e.getActionCommand() == "CUSTOMIZE_APPLY"){
			//if no graphs are open, then give error window
			if(_gui.getGraphManager().getGraphTabbedPane().getTabCount() == 0){
				JOptionPane.showMessageDialog(null, "There are no graphs available to customize.",
					"Customize Graph Information", JOptionPane.INFORMATION_MESSAGE);
			}else{
				_gui.applyCustomization();
			}
					
		//--------------- Customize Ok Button ---------------//
		}else if(e.getActionCommand() == "CUSTOMIZE_OK"){
			//if no graphs are open, then give error window
			if(_gui.getGraphManager().getGraphTabbedPane().getTabCount() == 0){
				JOptionPane.showMessageDialog(null, "There are no graphs available to customize.",
					"Customize Graph Information", JOptionPane.INFORMATION_MESSAGE);
			}else{
				_gui.applyCustomization();
				_gui.getFramesManager().removeOpenFrame("CUSTOMIZE");
			}
			
		//--------------- Customize Graph Cancel Button ---------------//
		}
		else if(e.getActionCommand() == "CUSTOMIZE_CANCEL")
		{
			_gui.getFramesManager().removeOpenFrame("CUSTOMIZE");
			
		//--------------- Open the About Panel ---------------//
		}
		else if(e.getActionCommand() == "ABOUT_CAPE")
		{
			String c = e.getActionCommand();
			FramesManager fm = _gui.getFramesManager();
			//If the frame is already open, then focus it.
			if(fm.isOpenFrame(c))
			{
				fm.toFront(c);
			}
			else
			{ 
				//Otherwise, create a new frame and display it
				fm.makeNewFrame("ABOUT_CAPE", "About CAPE", new Point(
					new Double(tk.getScreenSize().getWidth()/2-225).intValue(), 
					new Double(tk.getScreenSize().getHeight()/2-200).intValue()));
			}
			
		//--------------- Open Help Panel --------------------//
		}
		else if(e.getActionCommand() == "HELP"){
			String c = e.getActionCommand();
			FramesManager fm = _gui.getFramesManager();
			//If the frame is already open, then focus it.
			if(fm.isOpenFrame(c)){
				fm.toFront(c);
			}else{ //Otherwise, create a new frame and display it
				fm.makeNewFrame("HELP", "CAPE Help", new Point(
					new Double(tk.getScreenSize().getWidth()/2-150).intValue(), 
					new Double(tk.getScreenSize().getHeight()/2-100).intValue()));
			}
			
		//--------------- Start Statistics Timer ---------------//
		}else if(e.getActionCommand() == "START_STATISTICS_TIMER"){
			_gui.getUpdateStatisticTableTimer().start();
			
		//--------------- Stop Statistics Timer ---------------//
		}else if(e.getActionCommand() == "STOP_STATISTICS_TIMER"){
			_gui.getUpdateStatisticTableTimer().stop();
			
		//--------------- Exit Rainstorm ---------------//
		}else if(e.getActionCommand() == "EXIT"){
			System.exit(0);
		}
		else if(e.getActionCommand() == "PLOT_FILE_STATS_DEMO"){
			JFrame demoFrame = new JFrame();				
			demoFrame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			FramesManager fm = _gui.getFramesManager();
			/* ------------------ */
			// set demo properties
			String frequency = "1000";
			PlotFromFile f = new PlotFromFile("stats/test.csv", _gui);
			f.getNextStats();
			PlotFromFileQueryObject obj = (PlotFromFileQueryObject)f.getQueryObjects().get(0);
			ArrayList selectedStats = new ArrayList();
			//selectedStats.add(f.getOptions()[0]);
			//selectedStats.add(f.getOptions()[6]);
			ArrayList qobjs = new ArrayList();
			qobjs.add(f.getQueryObjects().get(2));
			qobjs.add(f.getQueryObjects().get(7));
			/* ------------------ */
			// create plot
			StatsPlotLive plot = new StatsPlotLive(selectedStats, qobjs, frequency);
			/* ------------------ */
			// other stuff
			JFrame plotFrame = new JFrame("Plot Queue Statistics");
			if(fm.getOpenPlotsCount() > 4){
				fm.resetOpenPlotsCount();
			}
			plotFrame.setLocation(new Double(tk.getScreenSize().getWidth()/3).intValue()-50 + fm.getOpenPlotsCount()*22,
				fm.getOpenPlotsCount()*22);
			fm.incrementOpenPlotsCount();
			plotFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
			plotFrame.getContentPane().add(plot);
			plotFrame.pack();
			plotFrame.setVisible(true);
			demoFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));			
		}
		else if(e.getActionCommand() == "DEVICE_STREAMSHIELD")
		{
			String c = e.getActionCommand();
			FramesManager fm = _gui.getFramesManager();
			//If the frame is already open, then focus it.
			if(fm.isOpenFrame(c))
			{
				fm.toFront(c);
			}
			else
			{ 
				//Otherwise, create a new frame and display it
				fm.makeNewFrame("DEVICE_STREAMSHIELD", 
								"StreamShield: Device (For Setting Security Preferences)", 
								new Point(new Double(tk.getScreenSize().getWidth()/2-225).intValue(), 
								new Double(tk.getScreenSize().getHeight()/2-200).intValue()));
			}
		}
		else if(e.getActionCommand() == "CLIENT_STREAMSHIELD")
		{
			String c = e.getActionCommand();
			FramesManager fm = _gui.getFramesManager();
			//If the frame is already open, then focus it.
			if(fm.isOpenFrame(c))
			{
				fm.toFront(c);
			}
			else
			{ 
				//Otherwise, create a new frame and display it
				fm.makeNewFrame("CLIENT_STREAMSHIELD", 
								"Query Mesh Application: Moving Objects (With Various Security Preferences)", 
								new Point(new Double(tk.getScreenSize().getWidth()/2-225).intValue(), 
								new Double(tk.getScreenSize().getHeight()/2-200).intValue()));
			}
		}
		else if(e.getActionCommand() == "SERVER_STREAMSHIELD")
		{
			String c = e.getActionCommand();
			FramesManager fm = _gui.getFramesManager();
			//If the frame is already open, then focus it.
			if(fm.isOpenFrame(c))
			{
				fm.toFront(c);
			}
			else
			{ 
				//Otherwise, create a new frame and display it
				fm.makeNewFrame("SERVER_STREAMSHIELD", 
								"StreamShield: Server Policy Control", 
								new Point(new Double(tk.getScreenSize().getWidth()/2-225).intValue(), 
								new Double(tk.getScreenSize().getHeight()/2-200).intValue()));
			}
		}
				
	}
}