/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.listeners;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Ellipse2D;
import java.util.Hashtable;
import java.util.Iterator;

import diva.graph.JGraph;
import diva.graph.NodeRenderer;
import diva.graph.basic.BasicEdgeRenderer;
import diva.graph.basic.BasicNodeRenderer;
import diva.graph.modular.Edge;
import diva.graph.modular.Node;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.partitionstreamoperator.StreamSplitOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamUnionOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;

import java.awt.geom.Ellipse2D;

/** The listener for the update node color timer.
 * Updates the node color to:
 * <ul>
 * <li>gray - if the node is inactive
 * <li>red - if the node is active and
 * <li>green - if the node is running.
 * </ul>
 */
public class UpdateNodeColorListener implements ActionListener{
	
	/** The CAPE_GUI */
	private CAPE_GUI _gui;
	
	private static Hashtable nodeToCounter = new Hashtable();
	
	/** Constructor associates the given <code>gui</code> with this listener. */
	public UpdateNodeColorListener(CAPE_GUI gui){
		_gui = gui;
	}
		
	public void actionPerformed(ActionEvent evt){

		try{//for each graph

			for(Iterator it = _gui.getGraphManager().getGraphs(); it.hasNext();){
				JGraph graph = (JGraph)it.next();
				//for each node
				for(Iterator nodes = graph.getGraphPane().getGraphModel().getNodes(); nodes.hasNext();){
					Node node = (Node)nodes.next();
					
					//get the semantic object(dag node) of this node
					DagNode dagNode = (DagNode)node.getSemanticObject();
										
					//get the node renderer for this node
					BasicNodeRenderer nrend = (BasicNodeRenderer)graph.getGraphPane()
						.getGraphController().getNodeController(node).getNodeRenderer();
						
					//if the node represents an input/output stream(s) node, then paint blue
					if(((XATStreamOperator)dagNode.getOperator()) == null){
						nrend.setNodeFill(Color.blue);
					}//otherwise the node is an actual node
					else{
						// if the node was recently migrated, then paint it yellow
						if (((XATStreamOperator)dagNode.getOperator()).isChanged()) {
							//nrend.setNodeStroke(new BasicStroke(3));
							nrend.setNodeFill(Color.yellow);
							//node.setLabelFigureMachineID(new LabelFigure("M"));
							((XATStreamOperator)dagNode.getOperator()).setChanged(false);
						} 
						else if(((XATStreamOperator)dagNode.getOperator()).isNewlyActivated()) {
							nrend.setNodeFill(Color.ORANGE);
							((XATStreamOperator)dagNode.getOperator()).setNewlyActivated(false);							
						}
						else {
							//if the node is running, then paint green
							if(((XATStreamOperator)dagNode.getOperator()).isRunning()){
								nrend.setNodeFill(Color.green);
							}//if the node is active, then paint red
							else if(((XATStreamOperator)dagNode.getOperator()).isActive()){
								XATStreamOperator tmpOP = (XATStreamOperator)dagNode.getOperator(); 
								if (tmpOP instanceof StreamSplitOperator) {
									nrend.setNodeFill(Color.yellow);
								} else if (tmpOP instanceof StreamUnionOperatorImp) {
									nrend.setNodeFill(Color.yellow);
								} else {
									nrend.setNodeFill(Color.red);
								}
							}//otherwise the node is inactive and paint gray
							else{nrend.setNodeFill(Color.gray);}
							//node.setLabelFigureMachineID(new LabelFigure(""));
						}
					}

					graph.getGraphPane().getGraphController().getNodeController(node).setNodeRenderer(nrend);
					graph.getGraphPane().getGraphController().getNodeController(node).drawNode(node);
				}//for each edge
				for(Iterator edges = graph.getGraphPane().getGraphModel().getEdges(); edges.hasNext();){
					Edge edge = (Edge)edges.next();
					
					//get the edge renderer for this edge
					BasicEdgeRenderer erend = (BasicEdgeRenderer)graph.getGraphPane().getGraphController().
						getEdgeController(edge).getEdgeRenderer();
					
					//erend.setLabelTextSize(10);
					if(((XATQueue)edge.getSemanticObject()).getSchema() == null){
						erend.setLabelTextColor(Color.gray);
					}
					else{erend.setLabelTextColor(Color.black);}
					
					graph.getGraphPane().getGraphController().getEdgeController(edge).setEdgeRenderer(erend);	
					graph.getGraphPane().getGraphController().getEdgeController(edge).drawEdge(edge);
				}
				_gui.getGraphManager().redrawGraph(graph);
			}
		}
		catch(Exception e)
		{
			System.err.println("UpdateNodeColorListener: Error updating node color. Message: " + e.getMessage());
		}
		
	}
}