/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.listeners;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Iterator;
import java.util.List;

import diva.canvas.toolbox.LabelFigure;
import diva.graph.JGraph;
import diva.graph.modular.Edge;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;

/** The listener for the update queues timer.
 * Update all queues every <code>UPDATE_QUEUES_SECONDS</code> milliseconds. 
 * 
 * <p>Future Work: Modify Diva's repaint to properly repaint the changed region only rather
 * than redrawing the entire graph. */
public class UpdateQueuesListener implements ActionListener{
	
	/** The CAPE_GUI */
	private CAPE_GUI _gui;
	
	/** Constructor associates the given <code>gui</code> with this listener.
	 * @param gui gui to associate with this listener */
	public UpdateQueuesListener(CAPE_GUI gui){
		_gui = gui;
	}
	
	public void actionPerformed(ActionEvent evt){
		try{ //for each graph
			for(Iterator it = _gui.getGraphManager().getGraphs(); it.hasNext();){
				JGraph graph = (JGraph)it.next();
				//for each edge
				for(Iterator edges = graph.getGraphPane().getGraphModel().getEdges(); edges.hasNext();){
					Edge edge = (Edge) edges.next();
					XATQueue queue = (XATQueue)edge.getSemanticObject();
					int queueNum = queue.getQueueNumber();
					int lastDQed = queue.size();
					List owners = queue.getQueueOwner();
					
					//if this queue is an input stream then make the label:
					//size  Example: 249
					LabelFigure label = new LabelFigure();
					if(queue.getQueueName().length() > 0){
						//label = new LabelFigure(queue.getQueueName()+"_O"+
						//							((XATStreamOperator)owners.get(0)).getProperty("OBJECT_ID")
						//							+"_Q"+queueNum+":"+lastDQed);
						label = new LabelFigure(Integer.toString(lastDQed));
					}
					else{
						//label = new LabelFigure("O"+((XATStreamOperator)owners.get(0))
						//							.getProperty("OBJECT_ID")+"_Q"+queueNum+":"+lastDQed);
						label = new LabelFigure(Integer.toString(lastDQed));
					}
					edge.setLabelFigure(label);

				}
			}
		}catch(Exception e){
			e.printStackTrace();
		}
	}
}