/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
//--------------------------------------------------
// ListViewCanvas.java
//  
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2009
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.component.gui.security;

import java.awt.Dimension;
import java.awt.Canvas;
import java.util.ArrayList;
import java.awt.Graphics;
import java.awt.Event;
import java.awt.Image;
import java.awt.Color;
import java.awt.Font;
import java.awt.Cursor;
import java.awt.FontMetrics;

public class ListViewCanvas extends Canvas implements CommonSettings
{
	//-----------------------------------------------------------------------------
	
	/***********Global Variable Declarations****************/
	Dimension offDimension,dimension;
    Image offImage;
    Graphics offGraphics;  	
    ServerStreamShieldMainForm streamShieldMainForm;    
    ArrayList ListArray;
    int G_ILoop,XOffset,YOffset;
    MessageObject messageobject;
    ScrollView scrollview;   
    FontMetrics fontmetrics; 
    int CanvasType, TotalWidth, TotalHeight;
    protected String SelectedCPQuery;
    
  //-----------------------------------------------------------------------------
    
	/**********Constructor Of Image Canvas *************/
	ListViewCanvas(ServerStreamShieldMainForm Parent, 
			       int canvastype)
	{
		streamShieldMainForm = Parent;		
		dimension = size();
		ListArray = new ArrayList();
		SelectedCPQuery = "";							
		CanvasType = canvastype;	
		setFont(streamShieldMainForm.getFont());		
		fontmetrics = streamShieldMainForm.getFontMetrics(streamShieldMainForm.getFont());	
	}
	
	//-----------------------------------------------------------------------------
	
	protected void AddListItemToMessageObject(String ListItem)
	{	
		int m_startY = DEFAULT_LIST_CANVAS_POSITION;	
		if(ListArray.size() > 0)
		{
			messageobject = (MessageObject) ListArray.get(ListArray.size()-1);
			m_startY = 	messageobject.StartY + DEFAULT_LIST_CANVAS_INCREMENT;
		}
		messageobject = new MessageObject();
		messageobject.Message = ListItem;
		messageobject.StartY  = m_startY;
		messageobject.Selected = false;
		messageobject.Width	  = fontmetrics.stringWidth(ListItem)+DEFAULT_LIST_CANVAS_INCREMENT;
		ListArray.add(messageobject);
		TotalWidth = Math.max(TotalWidth,messageobject.Width);
		scrollview.setValues(TotalWidth, m_startY + DEFAULT_LIST_CANVAS_HEIGHT);
		scrollview.setScrollPos(1,1);
		scrollview.setScrollSteps(2,1,DEFAULT_SCROLLING_HEIGHT);
		repaint();					
	}	
	
	//-----------------------------------------------------------------------------
	
	protected void AddItemsToMessageObject()
	{
		
		String[] cpQueries = TapPanel.readCPQueriesDirectory();
		int NumCPQueries = cpQueries.length;
		
		
		for(G_ILoop = 0; G_ILoop < NumCPQueries; G_ILoop++)
		{
			String ListItem = cpQueries[G_ILoop];
			AddListItemToMessageObject(ListItem);
		}			
	}
	
	//-----------------------------------------------------------------------------
	
	/****** Function To Clear All the Item From ListArray *********/
	protected void ClearAll()
	{
		ListArray.clear();
		TotalWidth = 0;
		TotalHeight = 0;
		scrollview.setValues(TotalWidth, TotalHeight);	
	}
	
	//-----------------------------------------------------------------------------
	
	/*******Function To Get the Index of Give Message from List Array *********/
	private int GetIndexOf(String Message)
	{
		int m_listSize = ListArray.size();
		
		for(G_ILoop = 0 ; G_ILoop < m_listSize; G_ILoop++)
		{
			messageobject = (MessageObject) ListArray.get(G_ILoop);
			if(messageobject.Message.equals(Message))			
				return G_ILoop;			
		}
		
		return -1;			
	}
	
	//-----------------------------------------------------------------------------
	
	protected void SelectCPQuery(String SelectedCPQuery)
	{
		int m_listIndex = GetIndexOf(SelectedCPQuery);
		if (m_listIndex >= 0)
		{
			messageobject = (MessageObject) ListArray.get(m_listIndex);
			ListArray.set(m_listIndex, messageobject);			
		}	
	}	
	
	//-----------------------------------------------------------------------------
	
	protected void SendDirectMessage()
	{
		streamShieldMainForm.MainCommandCanvas.AddMessageToMessageObject("Invalid User Selection!",MESSAGE_TYPE_ADMIN);
	    return;
	}	
	
	//-----------------------------------------------------------------------------
	
	/********** Function To Remove the Given Item From the List Array ********/
	protected void RemoveListItem(String ListItem)
	{
		int ListIndex = GetIndexOf(ListItem);
		if( ListIndex >= 0)
		{
			messageobject = (MessageObject) ListArray.get(ListIndex);
			int m_StartY = messageobject.StartY;
			ListArray.remove(ListIndex);		
			int m_listSize = ListArray.size();
			int m_nextStartY;
			for(G_ILoop = ListIndex; G_ILoop < m_listSize; G_ILoop++)
			{
				messageobject = (MessageObject) ListArray.get(G_ILoop);
				m_nextStartY = messageobject.StartY;
				messageobject.StartY = m_StartY;
				m_StartY = m_nextStartY;	
			} 				
		}
		repaint();
	}
	
	//-----------------------------------------------------------------------------
	
	private void PaintFrame(Graphics graphics)
	{
		int m_listArraySize = ListArray.size();		
		
		for(G_ILoop = 0; G_ILoop < m_listArraySize; G_ILoop++)
		{			
			messageobject = (MessageObject) ListArray.get(G_ILoop);			
			if((messageobject.StartY + messageobject.Height) >= YOffset)
			{				
				PaintListItemIntoCanvas(graphics,messageobject);	
			}
		}
	}
	
	//-----------------------------------------------------------------------------
	
	private void PaintListItemIntoCanvas(Graphics graphics, MessageObject messageObject)
	{
		int m_StartY = messageObject.StartY - YOffset;
		int m_imageIndex = ROOM_CANVAS_ICON;
		switch (CanvasType)
		{
			case CP_QUERY_CANVAS:
				{
					m_imageIndex = USER_CANVAS_IGNORE_ICON;
					/*					
						m_imageIndex = USER_CANVAS_NORMAL_ICON;
					*/
					break;	
				}	
		}
		
		graphics.drawImage(streamShieldMainForm.emotiSpIconArray[m_imageIndex], 
						   5 - XOffset,
						   m_StartY,
						   DEFAULT_LIST_CANVAS_HEIGHT,
						   DEFAULT_LIST_CANVAS_HEIGHT,
						   this);		
				
		if(messageobject.Selected == true)
		{
			graphics.setColor(Color.blue);
			graphics.fillRect(5 - XOffset + DEFAULT_LIST_CANVAS_HEIGHT,
							  m_StartY,
							  TotalWidth,
							  DEFAULT_LIST_CANVAS_INCREMENT);
			
			graphics.setColor(Color.white);		
			graphics.drawString(messageObject.Message,
								5 - XOffset + DEFAULT_LIST_CANVAS_INCREMENT,
								m_StartY + DEFAULT_LIST_CANVAS_HEIGHT);	
		}
		else
		{
			graphics.setColor(Color.white);
			graphics.fillRect(5 - XOffset + DEFAULT_LIST_CANVAS_HEIGHT,
					          m_StartY,
					          TotalWidth,
					          DEFAULT_LIST_CANVAS_INCREMENT);
			graphics.setColor(Color.black);		
			graphics.drawString(messageObject.Message,
					            5 - XOffset + DEFAULT_LIST_CANVAS_INCREMENT,
					            m_StartY + DEFAULT_LIST_CANVAS_HEIGHT);		
		}				
	}
	
	//-----------------------------------------------------------------------------
	
	public boolean handleEvent(Event event)
    {
        if(event.id == 1001 && event.arg == scrollview)
        {
            if(event.modifiers == 1)
                XOffset = event.key;
            else
                YOffset = event.key;            
            repaint();
            return true;
        } 
        else
        {
            return super.handleEvent(event);
        }			
    }
	
	//-----------------------------------------------------------------------------
  
    public boolean 	mouseDown(Event event, int i, int j)
	{		
		int CurrentY = j + YOffset;
		int m_listArraySize = ListArray.size();
		boolean SelectedFlag = false;
		
		streamShieldMainForm.tappanel.TxtNewCpQuery.setText("");
		
		for(G_ILoop = 0; G_ILoop <  m_listArraySize; G_ILoop++)
		{
			messageobject = (MessageObject) ListArray.get(G_ILoop);
			if((CurrentY >= messageobject.StartY) && (CurrentY <= (messageobject.StartY+DEFAULT_LIST_CANVAS_HEIGHT)))
			{				
				messageobject.Selected = true;
				SelectedCPQuery = messageobject.Message;
				SelectedFlag = true;
	
				if(CanvasType == CP_QUERY_CANVAS)
				{
					
				}			
			}
			else
			{
				messageobject.Selected=false;												
			}			
		}		
		repaint();
		
		if ((!SelectedFlag))
			SelectedCPQuery="";
		
		
		if((event.clickCount == 2) && 
		   (CanvasType == CP_QUERY_CANVAS) && 
		   (!(SelectedCPQuery.equals(""))) && 
		   (!(SelectedCPQuery.equals(CommonSettings.STREAMSHIELD_USER_NAME))))
		{
			CreatePrivateWindow();	
		}
		
		return true;
	}
    
  //-----------------------------------------------------------------------------
	
	private void CreatePrivateWindow()
	{
		/*
		//**** Chk whether ignored user *********
		if(!(IsIgnoredUser(SelectedUser)))	
		{
			boolean PrivateFlag = false;
			for(G_ILoop = 0; G_ILoop < streamShieldMainForm.PrivateWindowCount;G_ILoop++)
			{
				if(streamShieldMainForm.privatewindow[G_ILoop].UserName.equals(SelectedUser))
				{
					streamShieldMainForm.privatewindow[G_ILoop].show();
					streamShieldMainForm.privatewindow[G_ILoop].requestFocus();
					PrivateFlag = true;
					break;										
				}
			}	
			
			if(!(PrivateFlag))
			{	
				if(streamShieldMainForm.PrivateWindowCount >= MAX_PRIVATE_WINDOW)
				{
					streamShieldMainForm.MainCommandCanvas.AddMessageToMessageObject("You are Exceeding private window limit! So you may lose some message from your friends!",MESSAGE_TYPE_ADMIN);	
				}
				else							
				{
					
					streamShieldMainForm.privatewindow[streamShieldMainForm.PrivateWindowCount++] = new PrivateChat(streamShieldMainForm,SelectedUser);				
					streamShieldMainForm.privatewindow[streamShieldMainForm.PrivateWindowCount-1].show();
					streamShieldMainForm.privatewindow[streamShieldMainForm.PrivateWindowCount-1].requestFocus();
																
				}
			}
			
		}
		*/			
	}
	
	//-----------------------------------------------------------------------------
	
	public void paint(Graphics graphics)
	{			
		/*************Double Buffering**************/		
		dimension = size();

		/*********** Create the offscreen graphics context**************/
		if ((offGraphics == null) || (dimension.width != offDimension.width)|| (dimension.height != offDimension.height)) 
		{			
	    	offDimension = dimension;
	    	offImage = createImage(dimension.width, dimension.height);
	    	offGraphics = offImage.getGraphics();	    		    		    		    	
		}

		/********* Erase the previous image*********/
		offGraphics.setColor(Color.white);
		offGraphics.fillRect(0, 0, dimension.width, dimension.height);	

		/*************** Paint the frame into the image*****************/
		PaintFrame(offGraphics);

		/****************** Paint the image onto the screen*************/
		graphics.drawImage(offImage, 0, 0, null);
	}
	
	//-----------------------------------------------------------------------------
	
	public void update(Graphics graphics)	
	{
		paint(graphics);
	}
	
	//-----------------------------------------------------------------------------

}