/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
//--------------------------------------------------
// TapPanel.java
//  
// Author: Rimma V. Nehme, rnehme@purdue.edu, 2009
//--------------------------------------------------

package edu.wpi.cs.dsrg.xmldb.xat.component.gui.security;

import java.awt.*;
import java.awt.event.*;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.File;
import java.io.FilenameFilter;
public class TapPanel extends Panel implements CommonSettings,ActionListener
{
	
	//-----------------------------------------------------------------------------
	
	/*********Global Variable Declarations***********/	
	ServerStreamShieldMainForm streamShieldMainForm;
	
	protected TextArea TxtNewCpQuery;
	
	ScrollView StreamsScrollView, 
			   ImageScrollView,
			   RoomScrollView;
	
	protected StreamsCanvas StreamCanvas;
	protected EmotionCanvas EmotiSpCanvas;
	protected ListViewCanvas UserCanvas, CPQueryCanvas;
	
	Button CmdCreateNewCpQuery;
	
	//-----------------------------------------------------------------------------
	
	private static String cPQueriesFolderLocation = "C:\\Users\\Rimma\\workspace\\Raindrop\\resources\\security\\DEMO\\CPQueries\\";	
	int NumCpQueries = 0;
	
	//-----------------------------------------------------------------------------
	
	TapPanel(ServerStreamShieldMainForm parent, int NumStreams)
	{			  
		/***********Initialize the Components***********/
		streamShieldMainForm = parent;
	  
		Panel Tappanel = new Panel(new BorderLayout());
		CardLayout cardlayout = new CardLayout();
		Panel MainPanel = new Panel(cardlayout);
	  
		//----------------------------------------------------
		// STREAMS
		//----------------------------------------------------
	  
		/********Image Panel Coding ***********/
		Panel StreamPanel = new Panel(new BorderLayout());
	  
		StreamCanvas = new StreamsCanvas(streamShieldMainForm);
		StreamsScrollView = new ScrollView(StreamCanvas,
										   true,
										   true,
										   TAPPANEL_CANVAS_WIDTH,
										   TAPPANEL_CANVAS_HEIGHT,
										   SCROLL_BAR_SIZE);
		
		StreamCanvas.scrollview = StreamsScrollView;
	  
		/**********Add Icons into MessageObject *********/
		StreamCanvas.AddIconsToMessageObject(NumStreams);
		StreamPanel.add("Center", StreamsScrollView);
		
		
		//----------------------------------------------------
		// CP-QUERIES
		//----------------------------------------------------
		/********Room Panel Coding Starts***********/
		Panel RoomPanel = new Panel(new BorderLayout());
		CPQueryCanvas = new ListViewCanvas(streamShieldMainForm, CP_QUERY_CANVAS);
	  
		RoomScrollView = new ScrollView(CPQueryCanvas,
										true,
										true,
										TAPPANEL_CANVAS_WIDTH,
										TAPPANEL_CANVAS_HEIGHT,
										SCROLL_BAR_SIZE);
		
		CPQueryCanvas.scrollview = RoomScrollView;	 
		CPQueryCanvas.AddItemsToMessageObject();
		NumCpQueries = TapPanel.readCPQueriesDirectory().length;
		RoomPanel.add("Center", RoomScrollView);	  
	  
		Panel RoomButtonPanel = new Panel(new BorderLayout());
		Panel RoomCountPanel = new Panel(new BorderLayout());
		Label LblCaption = new Label("NEW CP-QUERY:", 1);
		
		RoomCountPanel.add("North", LblCaption);
		//TxtUserCount = new TextField();
		//TxtUserCount.setEditable(false);
		
		TxtNewCpQuery = new TextArea(5,5);
		TxtNewCpQuery.setEditable(true);
		//TxtNewCpQuery.setSize(TAPPANEL_CANVAS_WIDTH, 500);		
		
		RoomCountPanel.add("Center", TxtNewCpQuery);	  	  
		RoomButtonPanel.add("Center", RoomCountPanel);
	  
		CmdCreateNewCpQuery = new CustomButton(streamShieldMainForm, 
											   "Create New CP-Query");
		CmdCreateNewCpQuery.addActionListener(this);
		RoomButtonPanel.add("South", CmdCreateNewCpQuery);
	  
		RoomPanel.add("South", RoomButtonPanel);
	  
		//----------------------------------------------------
		// EMOTI-SPS
		//----------------------------------------------------
		/********Image Panel Coding ***********/
		Panel ImagePanel = new Panel(new BorderLayout());
	  
		EmotiSpCanvas = new EmotionCanvas(streamShieldMainForm);
		ImageScrollView = new ScrollView(EmotiSpCanvas,
										 true,
										 true,
										 TAPPANEL_CANVAS_WIDTH,
										 TAPPANEL_CANVAS_HEIGHT,
										 SCROLL_BAR_SIZE);
		
		EmotiSpCanvas.scrollview = ImageScrollView;
	  
		/**********Add Icons into MessageObject *********/
		EmotiSpCanvas.AddIconsToMessageObject();
		ImagePanel.add("Center", ImageScrollView);
	  
		//----------------------------------------------------
		// ADD SOME PANELS
		//----------------------------------------------------	
		/*********Add All the Panel in to Main Panel*********/
		
		MainPanel.add("StreamPanel", StreamPanel);
		MainPanel.add("RoomPanel", RoomPanel);
		MainPanel.add("ImagePanel", ImagePanel);
		
		cardlayout.show(MainPanel, "StreamPanel");
		
		BorderPanel borderpanel = new BorderPanel(this,
												  streamShieldMainForm,
												  cardlayout,
												  MainPanel,
												  TAPPANEL_WIDTH,
												  TAPPANEL_HEIGHT);
	  
		borderpanel.addTab("STREAMS", "StreamPanel");
		borderpanel.addTab("CP-QUERIES", "RoomPanel");
		borderpanel.addTab("EMOTI-SPS", "ImagePanel");
	  
		Tappanel.add(borderpanel);
		
		add("Center", Tappanel);	  		    	    	  
	}
	
	//-----------------------------------------------------------------------------
	
	/***********Action Listener coding **********/
	public void actionPerformed(ActionEvent evt)
	{
		if(evt.getSource().equals(CmdCreateNewCpQuery))
		{
			//******** Change Room Coding *********
			//streamShieldMainForm.ChangeRoom();	
			
			//-------------------------------------
			// CREATE NEW CP-QUERY
			//-------------------------------------
			String cpQueryContent = this.TxtNewCpQuery.getText();
			
			//write cp-query to file			
			WriteCPQuery(cPQueriesFolderLocation, NumCpQueries++, cpQueryContent);
			
			//clear the used text field
			this.TxtNewCpQuery.setText("");
			
			//refresh CP-Query list view					
			this.CPQueryCanvas.ClearAll();
			this.CPQueryCanvas.AddItemsToMessageObject();			
		}									
	}
	
	//-----------------------------------------------------------------------------
	
	private void WriteCPQuery(String directory, 
							  int counter,
							  String cpQueryContent)
	{		   
		try
		{		
			// Create file 
			String newFile = directory + "\\CpQuery_" + counter + ".txt";
		    FileWriter fstream = new FileWriter(newFile);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(cpQueryContent);
			
			//Close the output stream
			out.close();			    
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());			
		}
	}

	//-----------------------------------------------------------------------------
	  
	public static String[] readCPQueriesDirectory() 
	{
		File dir = new File(cPQueriesFolderLocation);
	    
	    String[] children = dir.list();
	    
	    return children;
	    
	    /*
	    if (children == null) 
	    {
	        // Either dir does not exist or is not a directory
	    } 
	    else 
	    {
	        for (int i = 0; i < children.length; i++) 
	        {
	            //Get filename of file or directory
	            String filename = children[i];
	        }
	    }
	    */
	}	

	//-----------------------------------------------------------------------------
	
}