/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.storage;

import java.util.HashMap;

/** The <code>OptionsStorage</code> class stores all of the user's 
 * selected/deselected options in a <code>HashMap</code>. The map's key
 * is a unique name for the options and the value is the option(s)
 * (i.e. <code>JCheckBox</code>es).
 * Examples of options are:
 * <ul><li>Which tree, queue, operator and system statistic to display in the
 * statistic table(s),
 * <li>Which tree(s), queue(s) or operator(s) to plot selected statistics for, or
 * <li>Which graph customizations to apply to the graph in the selected tab.
 * </ul>
 * 
 * @author Maylene Waltz
 */
public class OptionsStorage{
	
	/** The key a unique identifier naming the option(s) and the value
	 * is the actual options (i.e. <code>JCheckBox</code>es). */
	private HashMap _options = new HashMap();
	
	/** Constructor. */
	public OptionsStorage(){
	}
	
	/** Returns the options associated with the given <code>key</code>.
	 * @param key the unique name for the options to retrieve
	 * @return the options associated with the given <code>key</code> */
	public Object getOptions(String key){
		return _options.get(key);
	}
	
	/** Puts the given <code>key</code> and the given <code>options</code>
	 * into the <code>OptionsStorage HashMap</code>.
	 * @param key the unique name of the options to store
	 * @param options the actual options to store */
	public void setOptions(String key, Object options){
		_options.put(key, options);
	}
}