/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.storage;

import java.sql.Timestamp;
import java.text.SimpleDateFormat;

/** A <code>StatisticValue</code> is a pair of strings.
 * The first string is the statistic value and the second
 * string is the time at which the value was stored in
 * <code>StatisticStorage</code>.
 * <p>
 * The time is in the format of <code>hh:mm:ss:SSS</code> (i.e. </code>17:24:55:937</code>) where:
 * <ul>
 * <li><code>h</code> is the hour (1-24) in the day
 * <li><code>m</code> is the minute in the hour
 * <li><code>s</code> is the second in the minute
 * <li><code>S</code> is the millisecond in the second
 * </ul> 
 * 
 * @author Maylene Waltz
 */
public class StatisticValue{
	
	/** The format for the timestamp. */
	final SimpleDateFormat TIME = new SimpleDateFormat("k:mm:ss:SSS");
	
	/** The statistic value. */
	private String _value;
	
	/** The time in the format of h:mm:ss:SSS (i.e. 17:24:55:937) where:
	 * <ul>
	 * <li>h is the hour (1-24) in the day
	 * <li>m is the minute in the hour
	 * <li>s is the second in the minute
	 * <li>S is the millisecond in the second
	 * </ul> */
	private String _time;
	
	/** Constructor that makes a new statistic value pair with the 
	 * given <code>value</code> and time in the format of h:mm:ss:SSS (i.e. 17:24:55:937) where:
	 * <ul>
	 * <li>h is the hour (1-24) in the day 
	 * <li>m is the minute in the hour
	 * <li>s is the second in the minute
	 * <li>S is the millisecond in the second
	 * </ul>
	 * @param value the statistic value
	 * @param ts the timestamp of when the value was stored in <code>StatisticStorage</code> */
	public StatisticValue(String value, Timestamp ts){
			_value = value;
			_time = TIME.format(ts);
	}
	
	/** Returns the time when this statistic value was stored in <code>StatisticStorage</code>. */
	public String getTime(){
		return _time;
	}
	
	/** Returns the value of this statistic value pair. */
	public String getValue(){
		return _value;
	}
}