/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils;

import java.awt.Shape;
import java.awt.Color;
import java.awt.geom.Rectangle2D;

/** Utility functions for the CAPE_GUI. */
public class CAPEUtils{
	
	/** Determines the text color based on the selected index
	 * of 'Node Text Color' combo box in the customize panel.
	 * @param index the selected index of the 'Node Text Color' combo box
	 * @return the color at the selected index of the 'Node Text Color' combo box */
	static public Color determineColor(int index){
		Color color = new Color(0,0,0);		
		switch(index){
			case 0: color = Color.BLACK; break;
			case 1: color = Color.WHITE; break;
			case 2: color = Color.DARK_GRAY; break;
			case 3: color = Color.GRAY; break;
			case 4: color = Color.RED; break;
			case 5: color = Color.ORANGE; break;
			case 6: color = Color.YELLOW; break;
			case 7: color = Color.GREEN; break;
			case 8: color = Color.CYAN; break;
			case 9: color = Color.BLUE; break;
			default: return color;
		}
		return color;
	}
	
	/** Determines the node size based on the selected index
	 * of the 'Node Size' combo box in the customize panel.
	 * @param index the selected index of the 'Node Size' combo box
	 * @return a rectangle relative in size to the selected index of the 'Node Size' combo box */
	static public Shape determineNodeSize(int index){
		Rectangle2D size = new Rectangle2D.Double(0.0,0.0,10.0,10.0); //default size
		switch(index){
			case 0: size.setRect(0.0,0.0,10.0,10.0); break; //Extra Small
			case 1: size.setRect(0.0,0.0,20.0,20.0); break; //Small
			case 2: size.setRect(0.0,0.0,30.0,30.0); break; //Medium
			case 3: size.setRect(0.0,0.0,40.0,40.0); break; //Large
			case 4: size.setRect(0.0,0.0,50.0,50.0); break; //Extra Large
			default: return size;
		}
		return size;
	}
	
	/** Determines the text size based on the selected index
	 * of the 'Node Text Size' combo boxe in the customize panel.
	 * @param index the selected index of the 'Node Text Size' combo box
	 * @return the text size relative to the selected index of the 'Node Text Size' combo box */
	static public int determineTextSize(int index){
		int size = 12; //default size
		switch(index){
			case 0: size = 12; break; //Extra Small
			case 1: size = 16; break; //Small
			case 2: size = 22; break; //Medium
			case 3: size = 28; break; //Large
			case 4: size = 36; break; //Extra Large
			default: return size;
		}	
		return size;
	}
}