/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils;

import java.awt.BorderLayout;
import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.JPanel;
import javax.swing.JTabbedPane;

import diva.graph.GraphController;
import diva.graph.GraphModel;
import diva.graph.GraphPane;
import diva.graph.JGraph;
import diva.graph.basic.BasicGraphController;
import diva.graph.basic.BasicLayoutTarget;
import diva.graph.layout.GlobalLayout;
import diva.graph.layout.LevelLayout;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;

public class GraphManager
{
	
	//------------------------------------------------------------------------------
	
	private XATQueryObject queryObject;

	/** The CAPE_GUI */
	private CAPE_GUI _gui;
	
	/** An <code>ArrayList</code> of the open<code>JGraph</code>s. */
	private ArrayList _graphs;
	
	/** The tabbed pane to display the trees as graphs. Each tab contains a graph of a tree. */
	private JTabbedPaneCAPE _graphTabbedPane;
	
	/** The tabbed pane to display statistic tables. */
	private JTabbedPaneCAPE _statsTabbedPane;

	private GraphModel model;
		
	//------------------------------------------------------------------------------
	
	/** Constructor initializes the <code>ArrayList</code> that holds the open <code>JGraph</code>s and
	 * associates the given <code>gui</code> with this <code>GraphManager</code>
	 * @param gui the <code>CAPE_GUI</code> to associate with this <code>GraphManager</code> */
	public GraphManager(CAPE_GUI gui)
	{
		_gui = gui;
		_graphs = new ArrayList();
	}
	
	//------------------------------------------------------------------------------
	
	/** Add the given <code>graph</code> to the <code>ArrayList</code> of open graphs.
	 * @param the graph to add to known open graphs */
	public void addGraph(JGraph graph)
	{
		_graphs.add(graph);
	}
	
	//------------------------------------------------------------------------------
	
	/** Draws the given <code>model</code> as a new <code>JGraph</code> and places the graph
	 * in the graph tabbed pane. The given <code>queryObject</code> is associated with the
	 * tab of this tree.
	 * @param model the graph model of the tree
	 * @param queryObject the query object to associate with the tab of this tree */
	public void drawModel(GraphModel model, XATQueryObject queryObject)
	{
		this.model = model;
		this.queryObject = queryObject;
		JGraph graph = new JGraph(new GraphPane(new BasicGraphController(_gui), model));	
		graph.getGraphPane().getGraphController().setStatsTabbedPane(_statsTabbedPane);
		_graphs.add(graph);
		
		JPanel panel = new JPanel();
		panel.setLayout(new BorderLayout());
		panel.add(graph);
		String title = new String("StreamShield Graph");
		if(queryObject instanceof XATTree) title = new String("Query (Tree Id: "+((XATTree)queryObject).getProperty("TREE_ID")+ ")");
		_graphTabbedPane.addTab(title, queryObject, panel, "QUERY");
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns the <code>JGraph</code> located at the given <code>index</code>.
	 * @return the graph at the given <code>index</code> */
	public JGraph getGraph(int index)
	{
		return (JGraph)_graphs.get(index);
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns an <code>Iterator</code> of all the open graphs.
	 * @return an <code>Iterator</code> of all the open graphs */
	public Iterator getGraphs()
	{
		return _graphs.iterator();
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns the tabbed pane that holds the open graphs.
	 * @return the tabbed pane that holds the tree graphs */
	public JTabbedPaneCAPE getGraphTabbedPane()
	{
		return _graphTabbedPane;
	}
	
	//------------------------------------------------------------------------------
	
	/** Returns the tabbed pane that holds the statistic tables.
	 * @return the tabbed pane that holds the statistic tables */
	public JTabbedPaneCAPE getStatsTabbedPane()
	{
		return _statsTabbedPane;
	}
	
	//------------------------------------------------------------------------------
	
	/** Creates an empty <code>JTabbedPaneCAPE</code>.
	 * @return an empty tabbed pane */
	protected java.awt.Component initializeTabbedPaneRaindrop()
	{
		return new JTabbedPaneCAPE(JTabbedPane.TOP, JTabbedPane.WRAP_TAB_LAYOUT, _gui);
	}
	
	//------------------------------------------------------------------------------
	
	/** Redraws the given <code>graph</code> using the <code>LevelLayout</code> in the diva package. 
	 * @param graph the graph to be redrawn */
	public void redrawGraph(JGraph graph)
	{
		//System.out.println("~~~~~~~~~~~~~ Redraw ~~~~~~~~~~~~");
		/*if (_gui instanceof CAPE_GUIDistributionManager) {	
			_graphs.remove(graph);
			// recreate the graph with proper distribution settings
			for (Iterator nodes = this.model.getNodes(); nodes.hasNext();) {
				Node graphNode = (Node)nodes.next();
				DagNode n = (DagNode)graphNode.getSemanticObject();
				LabelFigure machineID = new LabelFigure(((CAPE_GUIDistributionManager)_gui).getMachineName(n));
				graphNode.setLabelFigureMachineID(machineID);
			}
			
			// draw the new model
			JGraph newGraph = new JGraph(new GraphPane(new BasicGraphController(_gui), model));
			newGraph.getGraphPane().getGraphController().setStatsTabbedPane(_statsTabbedPane);
			_graphs.add(newGraph);
			JPanel panel = new JPanel();
			panel.setLayout(new BorderLayout());
			panel.add(newGraph);
			String title = new String("");
			if(queryObject instanceof XATTree) title = new String("T"+((XATTree)queryObject).getProperty("TREE_ID"));
			_graphTabbedPane.addTab(title, queryObject, panel, "QUERY");		


			// remove the old graph tab
			//_graphTabbedPane.removeAll();
			// draw new model
			//this.drawModel(model, queryObject);
		}
		else {*/
			GraphController graphController = graph.getGraphPane().getGraphController();
			BasicLayoutTarget target = new BasicLayoutTarget(graphController);
			GlobalLayout graphLayout = new LevelLayout(target);
			graphLayout.layout(graph.getGraphPane().getGraphModel().getRoot());
		//}
	}
	
	//------------------------------------------------------------------------------
	
	/** Redraws the graph at the given <code>index</code>.
	 * @param the index of the graph to be redrawn */
	public void redrawGraph(int index)
	{
		redrawGraph((JGraph)_graphs.get(index));
	}
	
	//------------------------------------------------------------------------------
	
	/** Removes the given <code>graph</code> from the <code>ArrayList</code> of open graphs.
	 * @param the graph to be removed from the known open graphs */
	public void removeGraph(JGraph graph)
	{
		_graphs.remove(graph);
	}
	
	//------------------------------------------------------------------------------
	
	/** Removes the graph at the given <code>index</code>.
	 * @param the index of the graph to be removed from the known open graphs */
	public void removeGraph(int index)
	{
		_graphs.remove(index);
		_graphTabbedPane.remove(index);
	}
	
	//------------------------------------------------------------------------------
		
	/** Sets the graph tabbed pane to the given <code>graphTabbedPane</code>.
	 * @param a tabbed pane to be used to display tree graph(s) */
	public void setGraphTabbedPane(JTabbedPaneCAPE graphTabbedPane)
	{
		_graphTabbedPane = graphTabbedPane;
	}
	
	//------------------------------------------------------------------------------
	
	/** Sets the statistics tabbed pane to the given <code>statsTabbedPane</code>.
	 * @param a tabbed pane to be used to display statistic table(s) */
	public void setStatsTabbedPane(JTabbedPaneCAPE statsTabbedPane)
	{
		_statsTabbedPane = statsTabbedPane;
	}
	
	//------------------------------------------------------------------------------
	
	public int getGraphIndex(GraphModel m) throws Exception 
	{
		int index = 0;
		for (Iterator it = getGraphs(); it.hasNext(); index++) 
		{
			JGraph g = (JGraph)it.next();
			GraphModel model= g.getGraphPane().getGraphModel();
			if (model.getProperty(model.getRoot(), "TREE_ID") == m.getProperty(m.getRoot(), "TREE_ID")) 
			{
				return index;
			}
		}
		throw new Exception("No graph exists for this model.");
	}
	
	//------------------------------------------------------------------------------
	
}