/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils;

import javax.swing.ImageIcon;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;

import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.listeners.CAPEListener;

/** The <code>JMenuBar</code> specific to the CAPE. */
public class JMenuBarCAPE extends JMenuBar
{	
	/** The <code>CAPE_GUI</code>. */
	private CAPE_GUI _gui;
	
	/** Constructor associates the given <code>gui</code> with this menu bar. */
	public JMenuBarCAPE(CAPE_GUI gui)
	{
		_gui = gui;
	}
	
	/** Creates the menu bar for this <code>CAPE_GUI</code>.
	 * @return the menu bar for this <code>CAPE_GUI</code> */
	public JMenuBar createJMenuBarCAPE()
	{
		JMenuBar menuBar = new JMenuBar();
		JMenu menu, subMenu;
		JMenuItem menuItem;
		
		//create menu icons
		java.net.URL exitURL = CAPE_GUI.class.getResource("images/exit.png");
		ImageIcon exit = new ImageIcon(exitURL, "Exit");
		java.net.URL customURL = CAPE_GUI.class.getResource("images/customize.png");
		ImageIcon custom = new ImageIcon(customURL, "Customize Graph");
		java.net.URL plotURL = CAPE_GUI.class.getResource("images/plot.png");
		ImageIcon plot = new ImageIcon(plotURL, "Plot Statistics");
		java.net.URL treeURL = CAPE_GUI.class.getResource("images/tree.png");
		ImageIcon tree = new ImageIcon(treeURL, "Tree Statistics");
		java.net.URL sysURL = CAPE_GUI.class.getResource("images/system.png");
		ImageIcon sys = new ImageIcon(sysURL, "System Statistics");
		java.net.URL opURL = CAPE_GUI.class.getResource("images/operator.png");
		ImageIcon op = new ImageIcon(opURL, "Operator Statistics");
		java.net.URL qURL = CAPE_GUI.class.getResource("images/queue.png");
		ImageIcon q = new ImageIcon(qURL, "Queue Statistics");
		java.net.URL fURL = CAPE_GUI.class.getResource("images/open_file.png");
		ImageIcon f = new ImageIcon(fURL, "File Statistics");
	
		
		//----------------------------------------------------------------
		//StreamShield menu icons
		java.net.URL redURL = CAPE_GUI.class.getResource("images/security/red_dot.gif");
		ImageIcon redIcon = new ImageIcon(redURL, "Data Security Punctuation");
		
		java.net.URL blueURL = CAPE_GUI.class.getResource("images/security/blue_dot.gif");
		ImageIcon blueIcon = new ImageIcon(blueURL, "Query Security Punctuation");
		
		java.net.URL yellowURL = CAPE_GUI.class.getResource("images/security/yellow_dot.gif");
		ImageIcon yellowIcon = new ImageIcon(yellowURL, "Device Security Punctuation");
		
		//----------------------------------------------------------------
		
		//Create File Menu
		menu = new JMenu("File");
		menu.setMnemonic('F');
		menuBar.add(menu);
		
		menuItem = new JMenuItem("Exit", exit);
		menuItem.setMnemonic('X');
		menuItem.setActionCommand("EXIT");
		menuItem.addActionListener(new CAPEListener(_gui));
		menu.add(menuItem);
		
		//Create View Menu
		menu = new JMenu("View");
		menu.setMnemonic('V');
		menuBar.add(menu);
		
		menuItem = new JMenuItem("Customize Graph...", custom);
		menuItem.setActionCommand("CUSTOMIZE");
		menuItem.addActionListener(new CAPEListener(_gui));
		menu.add(menuItem);
		
		menu = new JMenu("Tools");
		menu.setMnemonic('T');
		menuBar.add(menu);
				
		// Table Statistics -->
		subMenu = new JMenu("Table Statistics");	
		menuItem = new JMenuItem("Tree Statistics", tree);
		menuItem.setActionCommand("DISPLAY_TREE_STATS");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		
		menuItem = new JMenuItem("System Statistics", sys);
		menuItem.setActionCommand("DISPLAY_SYSTEM_STATS");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		menu.add(subMenu);
		
		// Plot Statistics -->
		subMenu = new JMenu("Plot Statistics");		
		menuItem = new JMenuItem("Tree", tree);
		menuItem.setActionCommand("PLOT_TREE_STATS");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		
		
		menuItem = new JMenuItem("Queue", q);
		menuItem.setActionCommand("PLOT_QUEUE_STATS");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		
		menuItem = new JMenuItem("Operator", op);
		menuItem.setActionCommand("PLOT_OPERATOR_STATS");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		
		menuItem = new JMenuItem("System", sys);
		menuItem.setActionCommand("PLOT_SYSTEM_STATS");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		
		menuItem = new JMenuItem("From File", f);
		menuItem.setActionCommand("PLOT_FILE_STATS");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		menu.add(subMenu);
		
		// Select Statistics to Display -->
		subMenu = new JMenu("Select Statistics for Table(s)         ");
		menuItem = new JMenuItem("Tree", tree);
		menuItem.setActionCommand("SELECT_TREE_STATS");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		
		menuItem = new JMenuItem("Queue", q);
		menuItem.setActionCommand("SELECT_QUEUE_STATS");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		
		menuItem = new JMenuItem("Operator", op);
		menuItem.setActionCommand("SELECT_OPERATOR_STATS");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		
		menuItem = new JMenuItem("System", sys);
		menuItem.setActionCommand("SELECT_SYSTEM_STATS");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		menu.add(subMenu);
		
		//-------------------------------------------------------------------
		//Rimma: StreamShield DEMO		
		menu = new JMenu("Security");
		menu.setMnemonic('S');
		menuBar.add(menu);
	
		subMenu = new JMenu("StreamShield");
		
        //Device StreamShield Window
		subMenu = new JMenu("StreamShield");	
		menuItem = new JMenuItem("Device StreamShield", yellowIcon);
		menuItem.setActionCommand("DEVICE_STREAMSHIELD");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		
		//Client StreamShield Window
	
		menuItem = new JMenuItem("Application StreamShield", redIcon);
		menuItem.setActionCommand("CLIENT_STREAMSHIELD");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		
        //Server StreamShield Window
		menuItem = new JMenuItem("Server StreamShield", blueIcon);
		menuItem.setActionCommand("SERVER_STREAMSHIELD");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);
		menu.add(subMenu);
		
		// Query Security Punctuations -->
		subMenu = new JMenu("Security-Aware Query Plan");	
		menuItem = new JMenuItem("Create QSP", redIcon);
		menuItem.setActionCommand("QP_STREAMSHIELD");
		menuItem.addActionListener(new CAPEListener(_gui));
		subMenu.add(menuItem);		
		
		//-------------------------------------------------------------------		
		menu = new JMenu("Help");
		menu.setMnemonic('H');
		menuBar.add(menu);
		
		menuItem = new JMenuItem("Help");
		menuItem.setActionCommand("HELP");
		menuItem.addActionListener(new CAPEListener(_gui));
		menu.add(menuItem);
		
		menuItem = new JMenuItem("About CAPE...");
		menuItem.setActionCommand("ABOUT_CAPE");
		menuItem.addActionListener(new CAPEListener(_gui));
		menu.add(menuItem);
		
		return menuBar;
	}
}