/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.io.File;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
//import javax.swing.JViewport;
import javax.swing.SpringLayout;
import javax.swing.SwingConstants;
import javax.swing.border.TitledBorder;
import java.awt.Font;
//import javax.swing.BoxLayout;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.listeners.CAPEListener;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.security.DeviceStreamShieldMainForm;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.security.Stream;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.security.ServerStreamShieldMainForm;
//import edu.wpi.cs.dsrg.xmldb.xat.component.gui.security.Boston.SplineInterpolatorTest;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.storage.OptionsStorage;

/** The <code>PanelMaker</code> makes all of the panels for the <code>CAPE_GUI</code>.
 * Any selectable/deselected components, such as statistic check boxes, are stored
 * in <code>OptionsStorage</code>. All panels must have a unique name.
 *
 * @author Maylene Waltz
 */
public class PanelMaker
{

	/** The <code>CAPE_GUI</code>. */
	private CAPE_GUI _gui;

	/** Stores all of the panels ever-made.
	 * This avoids remaking panels. If the panel had previously been made,
	 * simply retrieve it.
	 * All panels must have a unique name. */
	private HashMap _panels = new HashMap();

	private ObjectInputStream distIn;

	/** Constructor associates the given <code>gui</code> with this <code>PanelMaker</code>. */
	public PanelMaker(CAPE_GUI gui)
	{
		_gui = gui;
	}

	/** Returns the panel with the given <code>panelName</code>.
	 * @param panelName the name of the panel that would like to be retrieved.
	 * @return the panel with the given <code>panelName</code>.
	 * Returns <code>null</code> if no panel exists with the given <code>panelName</code>. */
	public JPanel getPanel(String panelName){
		return (JPanel)_panels.get(panelName);
	}

	/** Puts the given <code>panel</code> with the given <code>panelName</code>
	 * into the <code>HashMap</code> of made panels.
	 * @param panelName the name of the panel. Must be a unique identifier.
	 * @param panel the panel to be placed in the <code>HashMap</code> of made panels */
	public void setPanel(String panelName, JPanel panel){
		_panels.put(panelName, panel);
	}

	/** Makes the panel for the given <code>panelName</code>.
	 * If the panel has previously been made, then the panel is
	 * retrieved from panel storage and returned.
	 * @param panelName the name of the panel to be made
	 * @return the actual <code>JPanel</code> associated with the given <code>panelName</code> */
	public java.awt.Component makePanel(String panelName)
	{
		//If the panel has already been made once, return it.
		if(_panels.containsKey(panelName) && panelName != "PLOT_FILE_STATS_OPTIONS")
		{
			return (JPanel)_panels.get(panelName);
		}
		//Otherwise, make a new panel and return it.
		else
		{
			if(panelName.equals("ABOUT_CAPE"))
			{
				JPanel panel = (JPanel)makeAboutCAPEPanel();
				setPanel(panelName, panel);
				return panel;
			}
			else if(panelName.equals("HELP"))
			{
				JPanel panel = (JPanel)makeHelpPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("CUSTOMIZE")){
				JPanel panel = (JPanel)makeCustomizePanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("GRAPH")){
				JPanel panel = (JPanel)makeGraphPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("PLOT_TREE_STATS")){
				JPanel panel = (JPanel)makePlotTreeStatsPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("PLOT_QUEUE_STATS")){
				JPanel panel = (JPanel)makePlotQueueStatsPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("PLOT_OPERATOR_STATS")){
				JPanel panel = (JPanel)makePlotOperatorStatsPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("PLOT_SYSTEM_STATS")){
				JPanel panel = (JPanel)makePlotSystemStatsPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("PLOT_FILE_STATS")) {
				JPanel  panel = (JPanel)makePlotFileStatsPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("PLOT_FILE_STATS_OPTIONS")) {
				JPanel  panel = (JPanel)makePlotFileStatsOptionsPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("SELECT_TREE_STATS")){
				JPanel panel = (JPanel)makeSelectTreeStatsPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("SELECT_QUEUE_STATS")){
				JPanel panel = (JPanel)makeSelectQueueStatsPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("SELECT_OPERATOR_STATS")){
				JPanel panel = (JPanel)makeSelectOperatorStatsPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("SELECT_SYSTEM_STATS")){
				JPanel panel = (JPanel)makeSelectSystemStatsPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("STATS")){
				JPanel panel = (JPanel)makeStatsPanel();
				setPanel(panelName, panel);
				return panel;
			}else if(panelName.equals("WHAT_TO_PLOT")){
				JPanel panel = (JPanel)makeWhatToPlotPanel();
				setPanel(panelName, panel);
				return panel;
			} else if(panelName.equals("USERINPUT")) {
				JPanel panel = (JPanel)makeUserInputPanel();
				setPanel(panelName, panel);
				return panel;
			} 
			else if (panelName.equals("CURRENT_CONFIGURATION")) 
			{
				JPanel panel = (JPanel)makeCurrentConfigurationPanel();
				setPanel(panelName, panel);
				return panel;
			}
			else if (panelName.equals("DEVICE_STREAMSHIELD")) 
			{
				//-------------------------------
				// DEVICE STREAMSHIELD
				//-------------------------------
				XATTree tree = _gui.getXATTree(0);
				XATNode [] leaves = tree.getLeaves();
				Stream [] streams = new Stream[leaves.length];
				
				int NumStreams = leaves.length;
				for(int s = 0; s < NumStreams; s++)
				{
					streams[s] = new Stream(leaves[s].toString());
				}

				JPanel panel = (JPanel)makeDeviceStreamShieldPanel(streams);
				panelName = "Device StreamShield";
				setPanel(panelName, panel);
				return panel;
			}
			else if (panelName.equals("CLIENT_STREAMSHIELD")) 
			{
				//-------------------------------
				// CLIENT STREAMSHIELD
				//-------------------------------
				XATTree tree = _gui.getXATTree(0);
				XATNode [] leaves = tree.getLeaves();
				Stream [] streams = new Stream[leaves.length];
				
				int NumStreams = leaves.length;
				for(int s = 0; s < NumStreams; s++)
				{
					streams[s] = new Stream(leaves[s].toString());
				}

				JPanel panel = (JPanel)makeClientStreamShieldPanel(streams);
				panelName = "Application (for Query Mesh Framework)";
				setPanel(panelName, panel);
				return panel;
			}
			else if (panelName.equals("SERVER_STREAMSHIELD")) 
			{
				//-------------------------------
				// SERVER STREAMSHIELD
				//-------------------------------
				XATTree tree = _gui.getXATTree(0);
				XATNode [] leaves = tree.getLeaves();
				Stream [] streams = new Stream[leaves.length];
				
				int NumStreams = leaves.length;
				for(int s = 0; s < NumStreams; s++)
				{
					streams[s] = new Stream(leaves[s].toString());
				}

				JPanel panel = (JPanel)makeServerStreamShieldPanel(streams);
				panelName = "Server StreamShield";
				setPanel(panelName, panel);
				return panel;
			}
			
			else 
				return new JPanel();
		}
	}

	/**
	 * @return
	 */
	private JPanel makeCurrentConfigurationPanel() {
		JPanel p = new JPanel();
		p.setOpaque(true);
		p.setLayout(new BorderLayout());
		p.setMaximumSize(new Dimension(700, 100));
		//p.setPreferredSize(new Dimension(700,170));
		p.setBorder(BorderFactory.createEmptyBorder(0,2,2,2));

		// create and set up tabbed pane
		JTabbedPaneCAPE tabs = (JTabbedPaneCAPE)_gui.getGraphManager().initializeTabbedPaneRaindrop();
		//tabs.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
		//tabs.setTabPlacement(JTabbedPane.TOP);


		// create and set up main panel
		JPanel currPanel = new JPanel();
		currPanel.setName("CURRENT_CONFIGURATION");
		currPanel.setOpaque(true);
		currPanel.setLayout(new GridLayout(1, 3));
		//currPanel.setLayout(new BorderLayout());
		currPanel.setPreferredSize(new Dimension(700,100));
		currPanel.setBorder(BorderFactory.createEmptyBorder(0,2,2,2));


		// create operator tab
		JPanel opPanel = new JPanel();
		opPanel.setName("Operators");
		opPanel.setLayout(new BorderLayout());
		opPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createLineBorder(Color.GRAY),
				"Operators", TitledBorder.LEFT, TitledBorder.TOP,
				new Font("Dialog", Font.BOLD, 12), Color.BLUE));
		// ****** PUT OPERATORS CONTENT LAYOUT HERE ******
		/* Example: */

		/* added by Luping*/
		JTextArea opStatusText = new JTextArea();
		//opStatusText.setLineWrap(true);
		opStatusText.setOpaque(false);
		JScrollPane opStatusPane = new JScrollPane(opStatusText, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		opStatusPane.setOpaque(false);
		opPanel.add(opStatusPane, BorderLayout.CENTER);
		//opStatusPane.setLocation(100, 0);

		/* commented by Luping
		JLabel label = new JLabel(); // this can contain any string
		opPanel.add(label);
		JButton button = new JButton(); // probably don't need this, but it's a standard button
		opPanel.add(button);
		String[] listContents = {"one", "two", "three", "four", "five"}; // this would be all available options, such as all schedulers
		JList list = new JList(listContents);
		list.setSelectedIndex(0); // the selected options, such as current scheduler
		list.setVisibleRowCount(5); // how many rows of options to display
		opPanel.add(list);
		*/
		// add operator tab
		//tabs.add(opPanel.getName(), opPanel);
//		currPanel.add(opPanel.getName(), opPanel);

		//create scheduler tab
		JPanel schedPanel = new JPanel();
		schedPanel.setName("Scheduler Adaptation");
		schedPanel.setLayout(new BorderLayout());
		schedPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createLineBorder(Color.GRAY),
				"Scheduler Adaptation", TitledBorder.LEFT, TitledBorder.TOP,
				new Font("Dialog", Font.BOLD, 12), Color.BLUE));

		JLabel schedCounter = new JLabel();
		schedCounter.setText("Counter: "+0);
		schedCounter.setForeground(Color.BLUE);
		schedPanel.add(schedCounter, BorderLayout.NORTH);

		// create scheduler list
/*		JList schedList = new JList();
		schedList.setPreferredSize(new Dimension(100, 100));
		schedList.setBorder(BorderFactory.createEmptyBorder());
		//schedList.setVisibleRowCount(2);
		schedPanel.add(schedList, BorderLayout.CENTER);
		//schedList.setLocation(1000, 10000);

		// add plan optimization tab
		//tabs.add(optPanel.getName(), optPanel);
		currPanel.add(schedPanel.getName(), schedPanel);*/


		// create plan optimization tab
		JPanel optPanel = new JPanel();
		optPanel.setName("Plan Migration");
		optPanel.setLayout(new BorderLayout());
		optPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createLineBorder(Color.GRAY),
				"Plan Migration", TitledBorder.LEFT, TitledBorder.TOP,
				new Font("Dialog", Font.BOLD, 12), Color.BLUE));

		// create migration counter
		//JLabel migLabel = new JLabel("Plan Migration", SwingConstants.LEFT); // this can contain any string
		//migLabel.setForeground(Color.GRAY);
		//optPanel.add(migLabel);
/*		JLabel migCounter = new JLabel();
		migCounter.setText("Counter: "+0);
		migCounter.setForeground(Color.BLUE);
		optPanel.add(migCounter, BorderLayout.NORTH);*/
		//migCounter.setLocation(100, 0);

//		JTextArea optTextArea = new JTextArea(15, 15);
		JTextArea optTextArea = new JTextArea();
//		optTextArea.setLineWrap(true);
		optTextArea.setOpaque(false);
		JScrollPane optStatusPane = new JScrollPane(optTextArea, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		optStatusPane.setOpaque(false);
		optPanel.add(optStatusPane, BorderLayout.CENTER);
		//optStatusPane.setLocation(100, 10000);

		currPanel.add(optPanel.getName(), optPanel);

		// create scheduler counter
		//JLabel schedLabel = new JLabel("Scheduler Adaptation", SwingConstants.LEFT); // this can contain any string

		//schedLabel.setSize(new Dimension(3000,1000));
		//schedLabel.setBackground(Color.gray);
		//optPanel.add(schedLabel);
		// create plan optimization tab


		// create distribution manager tab
		JPanel distPanel = new JPanel();
		distPanel.setName("Distribution");
		distPanel.setLayout(new BorderLayout());
		distPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createLineBorder(Color.GRAY),
				"Distribution", TitledBorder.LEFT, TitledBorder.TOP,
				new Font("Dialog", Font.BOLD, 12), Color.BLUE));
		// ****** PUT DISTRIBUTION CONTENT LAYOUT HERE ******

		JTextArea distTextArea = new JTextArea();
//		distTextArea.setLineWrap(true);
		distTextArea.setOpaque(false);
		JScrollPane distStatusPane = new JScrollPane(distTextArea, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		distStatusPane.setOpaque(false);
		distPanel.add(distStatusPane, BorderLayout.CENTER);

		/*
		distPanel.setLayout(new BorderLayout());
		JLabel distBodyText = new JLabel();
		distBodyText.setName("Redistribution Log");
		distBodyText.setText("<HTML>");
		JScrollPane distScroller = new JScrollPane(distBodyText, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		distPanel.add(distScroller, BorderLayout.CENTER);
		distScroller.setLocation(100, 0);
		*/
		//distScroller.setPreferredSize(new Dimension(distPanel.getWidth(), distPanel.getHeight()));

		/* commented temporarily
		int i;
		for (i = 0; i < _gui.getDistInfo().length; i++ )
		{
			distBodyText.setText(distBodyText.getText() + _gui.getDistInfo()[i] + "<BR>\n");
		}
		distBodyText.setText(distBodyText.getText() + "<BR>\n");
		*/

		//System.getProperty("REDISTRIBUTION_PERCENT");
		// add distribution manager tab
		//tabs.add(distPanel.getName(), distPanel);
		currPanel.add(distPanel.getName(), distPanel);

		/*
		// create output pane
		JPanel resultDisplayPanel = new JPanel();
		resultDisplayPanel.setName("Query Result");
		resultDisplayPanel.setLayout(new BorderLayout());
		resultDisplayPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createLineBorder(Color.GRAY),
				"Query Result", TitledBorder.LEFT, TitledBorder.TOP,
				new Font("Dialog", Font.BOLD, 12), Color.BLUE));

		JTextArea resultDisplayText = new JTextArea();
		resultDisplayText.setOpaque(false);
		JScrollPane resultDisplayPane = new JScrollPane(resultDisplayText, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		resultDisplayPane.setOpaque(false);
		resultDisplayPanel.add(resultDisplayPane, BorderLayout.CENTER);
		resultDisplayText.setLocation(100, 0);
		currPanel.add(resultDisplayPanel.getName(), resultDisplayPanel);
		*/

		// add tabbed pane to the main panel
		//currPanel.add(tabs);
		//JPanel panel = new JPanel();
		//JLabel label = new JLabel();
		//panel.add(label);
		//JScrollPane sp = new JScrollPane(panel);
		//sp.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		//sp.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
		//currPanel.add(sp);

		//create the panel to hold the tabbed pane
		//try{
		//	_gui.getGraphManager().setStatsTabbedPane((JTabbedPaneCAPE)_gui.getGraphManager().initializeTabbedPaneRaindrop());
		//	currPanel.add(_gui.getGraphManager().getStatsTabbedPane(), BorderLayout.CENTER);
		//}catch(Exception e){System.err.println("PanelMaker: Error creating STATS panel.");}
		tabs.addTab("Adaptation Status", currPanel);
		_gui.getGraphManager().setStatsTabbedPane(tabs);
		//_gui.getGraphManager().getStatsTabbedPane().addTab("Current Configuration", currPanel);
		p.add(_gui.getGraphManager().getStatsTabbedPane(), BorderLayout.CENTER);

		return p;
	}

	/** Creates the About CAPE panel.
	 * @return the panel with information about CAPE */
	private java.awt.Component makeAboutCAPEPanel(){
		JPanel aboutPanel = new JPanel();
		aboutPanel.setLayout(new BorderLayout());
		aboutPanel.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));

		try{
			//make the text
			JLabel about = new JLabel("<html>" +
			"<p align=\"center\"><font size=\"5\" color=\"#990000\">CAPE</font><br>" +
			"<font size=\"3\">Constraint-exploiting Adaptive Processing Engine<br><hr></font>" +
			"<table width=\"450\"><tr>" +
			"<td align=\"left\" valign=\"top\" width=\"50%\">Advisors:<ul><li>" +
				"Prof: Elke A. Rundensteiner</li><li>" +
				"Prof: George T. Heineman</li></ul>" +
			"Alumni:<ul>" +
			"<li> Timothy Sutherland</li>" +
			"<li> Bradford Pielech</li>" +
			"<li> Jinhui Jian</li>" +
			"<li> Nishant Mehta</li></ul></td>" +
			"<td align=\"left\" valign=\"top\" width=\"50%\">Graduate Students:<ul>" +
			"<li> Luping Ding</li>" +
			"<li> Yali Zhu</li>" +
			"<li>Mariana Jbantova</li>" +
			"<li> Malav Shah</li></ul>" +
			"Undergraduate Students:<ul>" +
			"<li> Maylene Waltz</li>" +
			"<li>Natasha Bogdanova</li></ul></td></tr></table><hr></html>", JLabel.CENTER);

			//add the WPI Logo
			//java.net.URL imageURL = PanelMaker.class.getResource("images/WPI.gif");
			//ImageIcon wpi = new ImageIcon(imageURL, "WPI Logo");
			//JLabel logo = new JLabel(wpi, JLabel.CENTER);

			//more text
			JLabel aboutCont = new JLabel("<html><center><font size=\"3\">" +
			"Computer Science Department<br>Worcester Polytechnic Institute<br>" +
			"http://davis.wpi.edu/dsrg/CAPE/</font></center></html>", JLabel.CENTER);

			aboutPanel.add(about, BorderLayout.PAGE_START);
			//aboutPanel.add(logo, BorderLayout.CENTER);
			aboutPanel.add(aboutCont, BorderLayout.PAGE_END);
		}catch(Exception e){System.err.println("PanelMaker: Error creating ABOUT_CAPE panel.");}

		return aboutPanel;
	}

	/** Creates the CAPE Help panel.
	 * @return the help panel */
	private java.awt.Component makeHelpPanel(){
		JPanel helpPanel = new JPanel();

		JLabel text = new JLabel("<html>" +
		"<p align=\"center\"><font size=\"5\" color=\"#990000\">CAPE</font><br>" +
		"<font size=\"3\">Constraint-exploiting Adaptive Processing Engine<br><hr></font>" +
		"<br><center>Click nodes and edges to open statistic tables.</center>", JLabel.CENTER);

		helpPanel.add(text);

		return helpPanel;
	}

	/** Creates the panel to customize node text color and size, node size,
	 * and edge text size.
	 * @return the customize panel */
	private java.awt.Component makeCustomizePanel(){
		JPanel customizePanel = new JPanel();
		customizePanel.setOpaque(true);
		customizePanel.setLayout(new BorderLayout());
		customizePanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
		try{
			JPanel springPanel = new JPanel();
			springPanel.setOpaque(false);
			springPanel.setLayout(new SpringLayout());
			springPanel.setBorder(BorderFactory.createTitledBorder("Options"));

			//create the labels and menus to customize node/edge attributes
			String[] labels = {"Node Text Color: ","Node Text Size: ","Node Size: ","Edge Text Size: "};
			String[] colors = {"Black","White","Dark Gray","Gray","Red","Orange","Yellow","Green","Cyan","Blue"};
			String[] sizes = {"Extra Small","Small","Medium","Large","Extra Large"};

			ArrayList comboBoxes = new ArrayList(labels.length);
			int numPairs = labels.length;

			for(int i=0; i < numPairs; i++){
				if(labels[i].endsWith("Color: ")){
					JLabel label = new JLabel(labels[i], JLabel.TRAILING);
					springPanel.add(label);
					JComboBox comboBox = new JComboBox(colors);
					comboBox.setSelectedIndex(0); //BLACK
					label.setLabelFor(comboBox);
					springPanel.add(comboBox);
					comboBoxes.add(comboBox);
				}else if(labels[i].endsWith("Node Size: ")) {
					JLabel label = new JLabel(labels[i], JLabel.TRAILING);
					springPanel.add(label);
					JComboBox comboBox = new JComboBox(sizes);
					comboBox.setSelectedIndex(0); //Extra Small
					label.setLabelFor(comboBox);
					springPanel.add(comboBox);
					comboBoxes.add(comboBox);
				}else {
					JLabel label = new JLabel(labels[i], JLabel.TRAILING);
					springPanel.add(label);
					JComboBox comboBox = new JComboBox(sizes);
					comboBox.setSelectedIndex(0); //Extra Small
					label.setLabelFor(comboBox);
					springPanel.add(comboBox);
					comboBoxes.add(comboBox);
				}
			}
			//store these options in OptionStorage
			_gui.getOptionsStorage().setOptions("CUSTOMIZE", comboBoxes);

			SpringUtilities.makeCompactGrid(springPanel, numPairs, 2, 6, 6, 6, 6);

			//create buttons and add to the bottom of customizePanel
			JPanel buttonPanel = new JPanel();
			buttonPanel.setOpaque(false);
			buttonPanel.setLayout(new FlowLayout());

			JButton applyButton = new JButton("Apply");
			applyButton.setToolTipText("Apply");
			applyButton.addActionListener(new CAPEListener(_gui));
			applyButton.setActionCommand("CUSTOMIZE_APPLY");
			applyButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton okButton = new JButton("Ok");
			okButton.setToolTipText("Ok");
			okButton.addActionListener(new CAPEListener(_gui));
			okButton.setActionCommand("CUSTOMIZE_OK");
			okButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton cancelButton = new JButton("Cancel");
			cancelButton.setToolTipText("Cancel");
			cancelButton.addActionListener(new CAPEListener(_gui));
			cancelButton.setActionCommand("CUSTOMIZE_CANCEL");
			cancelButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			buttonPanel.add(applyButton);
			buttonPanel.add(okButton);
			buttonPanel.add(cancelButton);
			buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));

			//add panels to the customizePanel
			customizePanel.add(springPanel, BorderLayout.PAGE_START);
			customizePanel.add(buttonPanel, BorderLayout.PAGE_END);
		}catch(Exception e){System.err.println("PanelMaker: Error creating CUSTOMIZE panel.");}

		return customizePanel;
	}

	/** Creates the graph panel and initializes an empty tabbed pane
	 * to hold the algebraic trees.
	 * @return the graph panel that displays the algebraic trees */
	private java.awt.Component makeGraphPanel(){
		JPanel graphPanel = new JPanel();
		graphPanel.setOpaque(false);
		graphPanel.setLayout(new BorderLayout());
		graphPanel.setPreferredSize(new Dimension(700,300));
		graphPanel.setBorder(BorderFactory.createEmptyBorder(0,2,2,2));

		try{
			_gui.getGraphManager().setGraphTabbedPane((JTabbedPaneCAPE)_gui.getGraphManager().initializeTabbedPaneRaindrop());
			graphPanel.add(_gui.getGraphManager().getGraphTabbedPane());
		}catch(Exception e){System.err.println("PanelMaker: Error creating GRAPH panel.");}

		return graphPanel;
	}

	/** Creates the panel to:
	 * <ul><li>select which tree statistics to plot,
	 * <li>select which trees to plot statistics for, and
	 * <li>specify the update frequency.
	 * </ul>
	 * @return the plot options panel for tree statistics */
	private java.awt.Component makePlotTreeStatsPanel(){
		try{
			if(_gui.getXATTreeCount() <= 0){
				JOptionPane.showMessageDialog(null, "Please load a query plan.",
					"Display Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
					return null;
			}else{
				OptionsStorage os = _gui.getOptionsStorage();
				JPanel plotPanel = new JPanel();//main content panel
				plotPanel.setOpaque(true);
				plotPanel.setLayout(new BorderLayout());
				plotPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));

				//--------------- Left Panel (statistics) ---------------//
				//Create the stats with check boxes panel
				JPanel statsPanel = new JPanel();
				statsPanel.setOpaque(false);
				statsPanel.setLayout(new BoxLayout(statsPanel, BoxLayout.PAGE_AXIS));
				ArrayList plotStatBoxes = new ArrayList();
				for(Enumeration e = _gui.getXATTree(0).getStatistics(); e.hasMoreElements();){
					JCheckBox box = new JCheckBox();
					String property = new String(e.nextElement().toString());
					box.setText(property);
					box.setHorizontalTextPosition(SwingConstants.LEADING);
					box.setAlignmentX(JCheckBox.RIGHT_ALIGNMENT);
					box.setAlignmentY(JCheckBox.TOP_ALIGNMENT);
					statsPanel.add(box);
					box.setSelected(false);
					plotStatBoxes.add(box);
				}
				//Store these options in OptionStorage
				os.setOptions("PLOT_SELECT_TREE_STATS", plotStatBoxes);

				//Create buttons for select stats panel
				JPanel statsButtonPanel = new JPanel();
				statsButtonPanel.setOpaque(false);
				statsButtonPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));

				JButton selectAllButton = new JButton("Select All");
				selectAllButton.setToolTipText("Select All Statistics");
				selectAllButton.addActionListener(new CAPEListener(_gui));
				selectAllButton.setActionCommand("PLOT_TREE_STATS_SELECT_ALL");
				selectAllButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				JButton deselectAllButton = new JButton("Deselect All");
				deselectAllButton.setToolTipText("Deselect All Statistics");
				deselectAllButton.addActionListener(new CAPEListener(_gui));
				deselectAllButton.setActionCommand("PLOT_TREE_STATS_DESELECT_ALL");
				deselectAllButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				statsButtonPanel.add(selectAllButton);
				statsButtonPanel.add(deselectAllButton);

				JPanel leftPanel = new JPanel(new BorderLayout());
				leftPanel.setOpaque(false);
				leftPanel.setBorder(BorderFactory.createTitledBorder("Statistics"));
				leftPanel.add(statsPanel, BorderLayout.PAGE_START);
				leftPanel.add(statsButtonPanel, BorderLayout.PAGE_END);

				//--------------- Right Panel (operators and update frequency) ---------------//
				//create operators panel
				JPanel opPanel = new JPanel();
				opPanel.setLayout(new BoxLayout(opPanel, BoxLayout.PAGE_AXIS));
				opPanel.setOpaque(false);
				opPanel.setBorder(BorderFactory.createTitledBorder("Trees"));
				ArrayList plotTreesBoxes = new ArrayList();

				//for each key representing a tree
				for(Iterator it = _gui.getTreeStorage().getKeys().iterator(); it.hasNext();){
					XATTree tree = (XATTree)it.next();
					String label = new String("T"+tree.getProperty("TREE_ID"));
					JCheckBox box = new JCheckBox();
					box.setText(label);
					box.setHorizontalTextPosition(SwingConstants.TRAILING);
					box.setAlignmentX(JCheckBox.LEFT_ALIGNMENT);
					box.setAlignmentY(JCheckBox.TOP_ALIGNMENT);
					box.setSelected(false);
					opPanel.add(box);
					plotTreesBoxes.add(box);
				}

				//Store these options in OptionStorage
				os.setOptions("PLOT_SELECT_TREES", plotTreesBoxes);

				//create update frequency panel
				JPanel freqPanel = new JPanel(new GridBagLayout());
				GridBagConstraints c = new GridBagConstraints();
				freqPanel.setOpaque(true);
				freqPanel.setBorder(BorderFactory.createTitledBorder("Update Frequency"));
				freqPanel.setMaximumSize(new Dimension(100,80));

				JLabel label0 = new JLabel(	"<html>The update frequency identifies how often to display<br>" +
											"the latest statistic value(s) on the plot.</html>");
				JLabel label1 = new JLabel("Enter update frequency in milliseconds:");
				label0.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
				label1.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));

				JTextField freqTextField = new JTextField("1000", 4);
				freqTextField.setBorder(BorderFactory.createLineBorder(Color.gray));
				freqTextField.setHorizontalAlignment(JTextField.CENTER);

				//Store this option in OptionStorage
				os.setOptions("PLOT_TREE_STATS_FREQUENCY", freqTextField);

				c.gridx = 0;
				c.gridy = 0;
				c.gridwidth = 2;
				freqPanel.add(label0, c);
				c.gridy = 1;
				c.gridwidth = 1;
				freqPanel.add(label1, c);
				c.gridx = 1;
				freqPanel.add(freqTextField, c);

				JPanel rightPanel = new JPanel(new BorderLayout());
				rightPanel.setOpaque(true);
				rightPanel.add(opPanel, BorderLayout.CENTER);
				rightPanel.add(freqPanel, BorderLayout.PAGE_END);

				//--------------- Button/Bottom Panel ---------------//
				JPanel buttonPanel = new JPanel();
				buttonPanel.setOpaque(false);
				buttonPanel.setLayout(new FlowLayout(FlowLayout.TRAILING));
				buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));

				JButton plotButton = new JButton("Plot");
				plotButton.setToolTipText("Plot");
				plotButton.addActionListener(new CAPEListener(_gui));
				plotButton.setActionCommand("PLOT_TREE_STATS_OK");
				plotButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				JButton cancelButton = new JButton("Cancel");
				cancelButton.setToolTipText("Cancel");
				cancelButton.addActionListener(new CAPEListener(_gui));
				cancelButton.setActionCommand("PLOT_TREE_STATS_CANCEL");
				cancelButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				buttonPanel.add(plotButton);
				buttonPanel.add(cancelButton);

				plotPanel.add(leftPanel, BorderLayout.LINE_START); //statistics
				plotPanel.add(rightPanel, BorderLayout.LINE_END); //operator and update frequency
				plotPanel.add(buttonPanel, BorderLayout.PAGE_END);

				return plotPanel;
			}
		}catch(Exception e){System.err.println("PanelMaker: Error creating PLOT_TREE_STATS panel.");}
		//never gets here
		return new JPanel();
	}

	/** Creates the panel to:
	 * <ul><li>select which queue statistics to plot,
	 * <li>select which queues to plot statistics for, and
	 * <li>specify the update frequency.
	 * </ul>
	 * @return the plot options panel for queue statistics */
	private java.awt.Component makePlotQueueStatsPanel(){
		try{
			if(_gui.getXATTreeCount() <= 0){
				JOptionPane.showMessageDialog(null, "Please load a query plan.",
					"Display Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
					return null;
			}else{
				OptionsStorage os = _gui.getOptionsStorage();
				JPanel plotPanel = new JPanel();//main content panel
				plotPanel.setOpaque(true);
				plotPanel.setLayout(new BorderLayout());
				plotPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));

				//--------------- Left Panel (statistics) ---------------//
				//Create the stats with check boxes panel
				JPanel statsPanel = new JPanel();
				statsPanel.setOpaque(false);
				statsPanel.setLayout(new BoxLayout(statsPanel, BoxLayout.PAGE_AXIS));
				ArrayList plotStatBoxes = new ArrayList();
				for(Enumeration e = ((XATStreamOperator)_gui.getXATTree(0).getRoot().getOperator())
						.getInputQueue().getStatistics(); e.hasMoreElements();){
					JCheckBox box = new JCheckBox();
					String property = new String(e.nextElement().toString());
					box.setText(property);
					box.setHorizontalTextPosition(SwingConstants.LEADING);
					box.setAlignmentX(JCheckBox.RIGHT_ALIGNMENT);
					box.setAlignmentY(JCheckBox.TOP_ALIGNMENT);
					statsPanel.add(box);
					box.setSelected(false);
					plotStatBoxes.add(box);
				}
				//Store these options in OptionStorage
				os.setOptions("PLOT_SELECT_QUEUE_STATS", plotStatBoxes);

				//Create buttons for select stats panel
				JPanel statsButtonPanel = new JPanel();
				statsButtonPanel.setOpaque(false);
				statsButtonPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));

				JButton selectAllButton = new JButton("Select All");
				selectAllButton.setToolTipText("Select All Statistics");
				selectAllButton.addActionListener(new CAPEListener(_gui));
				selectAllButton.setActionCommand("PLOT_QUEUE_STATS_SELECT_ALL");
				selectAllButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				JButton deselectAllButton = new JButton("Deselect All");
				deselectAllButton.setToolTipText("Deselect All Statistics");
				deselectAllButton.addActionListener(new CAPEListener(_gui));
				deselectAllButton.setActionCommand("PLOT_QUEUE_STATS_DESELECT_ALL");
				deselectAllButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				statsButtonPanel.add(selectAllButton);
				statsButtonPanel.add(deselectAllButton);

				JPanel leftPanel = new JPanel(new BorderLayout());
				leftPanel.setOpaque(false);
				leftPanel.setBorder(BorderFactory.createTitledBorder("Statistics"));
				leftPanel.add(statsPanel, BorderLayout.PAGE_START);
				leftPanel.add(statsButtonPanel, BorderLayout.PAGE_END);

				//--------------- Right Panel (queues and update frequency) ---------------//
				//create queues panel
				JPanel qPanel = new JPanel(new BorderLayout());
				qPanel.setOpaque(false);
				qPanel.setBorder(BorderFactory.createTitledBorder("Queues"));
				JTabbedPane qPane = new JTabbedPane(JTabbedPane.TOP, JTabbedPane.WRAP_TAB_LAYOUT);
				ArrayList plotQueueBoxes = new ArrayList();

				//for each key representing a tree
				for(Iterator it = _gui.getTreeStorage().getKeys().iterator(); it.hasNext();){
					XATTree tree = (XATTree)it.next();
					String title = tree.getProperty("TREE_ID");
					JPanel treePanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
					treePanel.setPreferredSize(new Dimension(100,100));
					ArrayList queues = (ArrayList)_gui.getTreeStorage().getQueues(tree);

					//for each key's value which is an ArrayList of queues
					for(int i=0; i<queues.size(); i++){
						XATQueue queue = (XATQueue)queues.get(i);
						JCheckBox box = new JCheckBox();
						for(Iterator owners = queue.getQueueOwner().iterator(); owners.hasNext();){
							String[] ownerStringSplit = owners.next().toString().split("\\p{Space}");
							String ownerID = ownerStringSplit[1];
							String label = "T"+title+"_O"+ownerStringSplit[1]+"_Q"+String.valueOf(queue.getQueueNumber());
							box.setText(label);
							box.setHorizontalTextPosition(SwingConstants.TRAILING);
							box.setAlignmentX(JCheckBox.LEFT_ALIGNMENT);
							box.setAlignmentY(JCheckBox.TOP_ALIGNMENT);
							box.setSelected(false);
							treePanel.add(box);
							plotQueueBoxes.add(box);
						}
					}
					treePanel.setMinimumSize(new Dimension(100,80));
					treePanel.setPreferredSize(new Dimension(100,175));
					JScrollPane scrollPane = new JScrollPane(treePanel);
					scrollPane.setWheelScrollingEnabled(true);
					qPane.addTab("T"+title, scrollPane);
				}
				qPanel.add(qPane, BorderLayout.CENTER);

				//Store these options in OptionStorage
				os.setOptions("PLOT_SELECT_QUEUES", plotQueueBoxes);

				//create update frequency panel
				JPanel freqPanel = new JPanel(new GridBagLayout());
				GridBagConstraints c = new GridBagConstraints();
				freqPanel.setOpaque(true);
				freqPanel.setBorder(BorderFactory.createTitledBorder("Update Frequency"));
				freqPanel.setMaximumSize(new Dimension(100,80));

				JLabel label0 = new JLabel(	"<html>The update frequency identifies how often to display<br>" +
											"the latest statistic value(s) on the plot.</html>");
				JLabel label1 = new JLabel("Enter update frequency in milliseconds:");
				label0.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
				label1.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));

				JTextField freqTextField = new JTextField("1000", 4);
				freqTextField.setBorder(BorderFactory.createLineBorder(Color.gray));
				freqTextField.setHorizontalAlignment(JTextField.CENTER);

				//Store this option in OptionStorage
				os.setOptions("PLOT_QUEUE_STATS_FREQUENCY", freqTextField);

				c.gridx = 0;
				c.gridy = 0;
				c.gridwidth = 2;
				freqPanel.add(label0, c);
				c.gridy = 1;
				c.gridwidth = 1;
				freqPanel.add(label1, c);
				c.gridx = 1;
				freqPanel.add(freqTextField, c);

				JPanel rightPanel = new JPanel(new BorderLayout());
				rightPanel.setOpaque(true);
				rightPanel.add(qPanel, BorderLayout.CENTER);
				rightPanel.add(freqPanel, BorderLayout.PAGE_END);

				//--------------- Button/Bottom Panel ---------------//
				JPanel buttonPanel = new JPanel();
				buttonPanel.setOpaque(false);
				buttonPanel.setLayout(new FlowLayout(FlowLayout.TRAILING));
				buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));

				JButton plotButton = new JButton("Plot");
				plotButton.setToolTipText("Plot");
				plotButton.addActionListener(new CAPEListener(_gui));
				plotButton.setActionCommand("PLOT_QUEUE_STATS_OK");
				plotButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				JButton cancelButton = new JButton("Cancel");
				cancelButton.setToolTipText("Cancel");
				cancelButton.addActionListener(new CAPEListener(_gui));
				cancelButton.setActionCommand("PLOT_QUEUE_STATS_CANCEL");
				cancelButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				buttonPanel.add(plotButton);
				buttonPanel.add(cancelButton);

				plotPanel.add(leftPanel, BorderLayout.LINE_START); //statistics
				plotPanel.add(rightPanel, BorderLayout.LINE_END); //operator and update frequency
				plotPanel.add(buttonPanel, BorderLayout.PAGE_END);

				return plotPanel;
			}
		}catch(Exception e){System.err.println("PanelMaker: Error creating PLOT_QUEUE_STATS panel.");}
		//never gets here
		return new JPanel();
	}

	/** Creates the panel to:
	 * <ul><li>select which operator statistics to plot,
	 * <li>select which operators to plot statistics for, and
	 * <li>specify the update frequency.
	 * </ul>
	 * @return the plot options panel for operator statistics */
	private java.awt.Component makePlotOperatorStatsPanel(){
		try{
			if(_gui.getXATTreeCount() <= 0){
				JOptionPane.showMessageDialog(null, "Please load a query plan.",
					"Display Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
					return null;
			}else{
				OptionsStorage os = _gui.getOptionsStorage();
				JPanel plotPanel = new JPanel();//main content panel
				plotPanel.setOpaque(true);
				plotPanel.setLayout(new BorderLayout());
				plotPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));

				//--------------- Left Panel (statistics) ---------------//
				//Create the stats with check boxes panel
				JPanel statsPanel = new JPanel();
				statsPanel.setOpaque(false);
				statsPanel.setLayout(new BoxLayout(statsPanel, BoxLayout.PAGE_AXIS));
				ArrayList plotStatBoxes = new ArrayList();
				for(Enumeration e = _gui.getXATTree(0).getRoot().getOperator().getStatistics(); e.hasMoreElements();){
					JCheckBox box = new JCheckBox();
					String property = new String(e.nextElement().toString());
					box.setText(property);
					box.setHorizontalTextPosition(SwingConstants.LEADING);
					box.setAlignmentX(JCheckBox.RIGHT_ALIGNMENT);
					box.setAlignmentY(JCheckBox.TOP_ALIGNMENT);
					statsPanel.add(box);
					box.setSelected(false);
					plotStatBoxes.add(box);
				}
				//Store these options in OptionStorage
				os.setOptions("PLOT_SELECT_OPERATOR_STATS", plotStatBoxes);

				//Create buttons for select stats panel
				JPanel statsButtonPanel = new JPanel();
				statsButtonPanel.setOpaque(false);
				statsButtonPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));

				JButton selectAllButton = new JButton("Select All");
				selectAllButton.setToolTipText("Select All Statistics");
				selectAllButton.addActionListener(new CAPEListener(_gui));
				selectAllButton.setActionCommand("PLOT_OPERATOR_STATS_SELECT_ALL");
				selectAllButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				JButton deselectAllButton = new JButton("Deselect All");
				deselectAllButton.setToolTipText("Deselect All Statistics");
				deselectAllButton.addActionListener(new CAPEListener(_gui));
				deselectAllButton.setActionCommand("PLOT_OPERATOR_STATS_DESELECT_ALL");
				deselectAllButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				statsButtonPanel.add(selectAllButton);
				statsButtonPanel.add(deselectAllButton);

				JPanel leftPanel = new JPanel(new BorderLayout());
				leftPanel.setOpaque(false);
				leftPanel.setBorder(BorderFactory.createTitledBorder("Statistics"));
				leftPanel.add(statsPanel, BorderLayout.PAGE_START);
				leftPanel.add(statsButtonPanel, BorderLayout.PAGE_END);

				//--------------- Right Panel (operators and update frequency) ---------------//
				//create operators panel
				JPanel opPanel = new JPanel(new BorderLayout());
				opPanel.setOpaque(false);
				opPanel.setBorder(BorderFactory.createTitledBorder("Operators"));
				JTabbedPane opPane = new JTabbedPane(JTabbedPane.TOP, JTabbedPane.WRAP_TAB_LAYOUT);
				ArrayList plotOperatorBoxes = new ArrayList();

				//for each key representing a tree
				for(Iterator it = _gui.getTreeStorage().getKeys().iterator(); it.hasNext();){
					XATTree tree = (XATTree)it.next();
					String title = tree.getProperty("TREE_ID");
					JPanel treePanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
					treePanel.setPreferredSize(new Dimension(100,100));
					ArrayList operators = (ArrayList)_gui.getTreeStorage().getOperators(tree);

					//for each key's value which is an ArrayList of operators
					for(int i=0; i<operators.size(); i++){
						JCheckBox box = new JCheckBox();
						String label = ((XATQueryObject)operators.get(i)).getProperty("OBJECT_ID");
						box.setText(label);
						box.setHorizontalTextPosition(SwingConstants.TRAILING);
						box.setAlignmentX(JCheckBox.LEFT_ALIGNMENT);
						box.setAlignmentY(JCheckBox.TOP_ALIGNMENT);
						box.setSelected(false);
						treePanel.add(box);
						plotOperatorBoxes.add(box);
					}
					opPane.addTab("T"+title, treePanel);
				}
				opPanel.add(opPane, BorderLayout.CENTER);

				//Store these options in OptionStorage
				os.setOptions("PLOT_SELECT_OPERATORS", plotOperatorBoxes);

				//create update frequency panel
				JPanel freqPanel = new JPanel(new GridBagLayout());
				GridBagConstraints c = new GridBagConstraints();
				freqPanel.setOpaque(true);
				freqPanel.setBorder(BorderFactory.createTitledBorder("Update Frequency"));
				freqPanel.setMaximumSize(new Dimension(100,80));

				JLabel label0 = new JLabel(	"<html>The update frequency identifies how often to display<br>" +
											"the latest statistic value(s) on the plot.</html>");
				JLabel label1 = new JLabel("Enter update frequency in milliseconds:");
				label0.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
				label1.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));

				JTextField freqTextField = new JTextField("1000", 4);
				freqTextField.setBorder(BorderFactory.createLineBorder(Color.gray));
				freqTextField.setHorizontalAlignment(JTextField.CENTER);

				//Store this option in OptionStorage
				os.setOptions("PLOT_OPERATOR_STATS_FREQUENCY", freqTextField);

				c.gridx = 0;
				c.gridy = 0;
				c.gridwidth = 2;
				freqPanel.add(label0, c);
				c.gridy = 1;
				c.gridwidth = 1;
				freqPanel.add(label1, c);
				c.gridx = 1;
				freqPanel.add(freqTextField, c);

				JPanel rightPanel = new JPanel(new BorderLayout());
				rightPanel.setOpaque(true);
				rightPanel.add(opPanel, BorderLayout.CENTER);
				rightPanel.add(freqPanel, BorderLayout.PAGE_END);

				//--------------- Button/Bottom Panel ---------------//
				JPanel buttonPanel = new JPanel();
				buttonPanel.setOpaque(false);
				buttonPanel.setLayout(new FlowLayout(FlowLayout.TRAILING));
				buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));

				JButton plotButton = new JButton("Plot");
				plotButton.setToolTipText("Plot");
				plotButton.addActionListener(new CAPEListener(_gui));
				plotButton.setActionCommand("PLOT_OPERATOR_STATS_OK");
				plotButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				JButton cancelButton = new JButton("Cancel");
				cancelButton.setToolTipText("Cancel");
				cancelButton.addActionListener(new CAPEListener(_gui));
				cancelButton.setActionCommand("PLOT_OPERATOR_STATS_CANCEL");
				cancelButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				buttonPanel.add(plotButton);
				buttonPanel.add(cancelButton);

				plotPanel.add(leftPanel, BorderLayout.LINE_START); //statistics
				plotPanel.add(rightPanel, BorderLayout.LINE_END); //operator and update frequency
				plotPanel.add(buttonPanel, BorderLayout.PAGE_END);

				return plotPanel;
			}
		}catch(Exception e){System.err.println("PanelMaker: Error creating PLOT_OPERATOR_STATS panel.");}
		//never gets here
		return new JPanel();
	}

	/** Creates the panel to select which system statistics to plot
	 * and to specify the update frequency.
	 * @return the plot options panel for system statistics */
	private java.awt.Component makePlotSystemStatsPanel(){
		try{
			if(_gui.getSystemObject() == null){
				JOptionPane.showMessageDialog(null, "No System Object Available.",
					"Display Statistic(s) Information", JOptionPane.INFORMATION_MESSAGE);
					return null;
			}else{
				OptionsStorage os = _gui.getOptionsStorage();
				JPanel plotPanel = new JPanel();//main content panel
				plotPanel.setOpaque(true);
				plotPanel.setLayout(new BorderLayout());
				plotPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));

				//--------------- Left Panel (statistics) ---------------//
				//Create the stats with check boxes panel
				JPanel statsPanel = new JPanel();
				statsPanel.setOpaque(false);
				statsPanel.setLayout(new BoxLayout(statsPanel, BoxLayout.PAGE_AXIS));
				ArrayList plotStatBoxes = new ArrayList();
				for(Enumeration e = _gui.getSystemObject().getStatistics(); e.hasMoreElements();){
					JCheckBox box = new JCheckBox();
					String property = new String(e.nextElement().toString());
					box.setText(property);
					box.setHorizontalTextPosition(SwingConstants.LEADING);
					box.setAlignmentX(JCheckBox.RIGHT_ALIGNMENT);
					box.setAlignmentY(JCheckBox.TOP_ALIGNMENT);
					statsPanel.add(box);
					box.setSelected(false);
					plotStatBoxes.add(box);
				}
				//Store these options in OptionStorage
				os.setOptions("PLOT_SELECT_SYSTEM_STATS", plotStatBoxes);

				//Create buttons for select stats panel
				JPanel statsButtonPanel = new JPanel();
				statsButtonPanel.setOpaque(false);
				statsButtonPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));

				JButton selectAllButton = new JButton("Select All");
				selectAllButton.setToolTipText("Select All Statistics");
				selectAllButton.addActionListener(new CAPEListener(_gui));
				selectAllButton.setActionCommand("PLOT_SYSTEM_STATS_SELECT_ALL");
				selectAllButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				JButton deselectAllButton = new JButton("Deselect All");
				deselectAllButton.setToolTipText("Deselect All Statistics");
				deselectAllButton.addActionListener(new CAPEListener(_gui));
				deselectAllButton.setActionCommand("PLOT_SYSTEM_STATS_DESELECT_ALL");
				deselectAllButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				statsButtonPanel.add(selectAllButton);
				statsButtonPanel.add(deselectAllButton);

				JPanel leftPanel = new JPanel(new BorderLayout());
				leftPanel.setOpaque(false);
				leftPanel.setBorder(BorderFactory.createTitledBorder("Statistics"));
				leftPanel.add(statsPanel, BorderLayout.PAGE_START);
				leftPanel.add(statsButtonPanel, BorderLayout.PAGE_END);

				//--------------- Right Panel (operators and update frequency) ---------------//
				//create operators panel
				JPanel opPanel = new JPanel(new BorderLayout());
				opPanel.setOpaque(false);
				opPanel.setBorder(BorderFactory.createTitledBorder("System"));
				JTabbedPane opPane = new JTabbedPane(JTabbedPane.TOP, JTabbedPane.WRAP_TAB_LAYOUT);
				ArrayList plotOperatorBoxes = new ArrayList();

				//create update frequency panel
				JPanel freqPanel = new JPanel(new GridBagLayout());
				GridBagConstraints c = new GridBagConstraints();
				freqPanel.setOpaque(true);
				freqPanel.setBorder(BorderFactory.createTitledBorder("Update Frequency"));
				freqPanel.setMaximumSize(new Dimension(100,80));

				JLabel label0 = new JLabel(	"<html>The update frequency identifies how often to display<br>" +
											"the latest statistic value(s) on the plot.</html>");
				JLabel label1 = new JLabel("Enter update frequency in milliseconds:");
				label0.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
				label1.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));

				JTextField freqTextField = new JTextField("1000", 4);
				freqTextField.setBorder(BorderFactory.createLineBorder(Color.gray));
				freqTextField.setHorizontalAlignment(JTextField.CENTER);

				//Store this option in OptionStorage
				os.setOptions("PLOT_SYSTEM_STATS_FREQUENCY", freqTextField);

				c.gridx = 0;
				c.gridy = 0;
				c.gridwidth = 2;
				freqPanel.add(label0, c);
				c.gridy = 1;
				c.gridwidth = 1;
				freqPanel.add(label1, c);
				c.gridx = 1;
				freqPanel.add(freqTextField, c);

				JPanel rightPanel = new JPanel(new BorderLayout());
				rightPanel.setOpaque(true);
				rightPanel.add(opPanel, BorderLayout.CENTER);
				rightPanel.add(freqPanel, BorderLayout.PAGE_END);

				//--------------- Button/Bottom Panel ---------------//
				JPanel buttonPanel = new JPanel();
				buttonPanel.setOpaque(false);
				buttonPanel.setLayout(new FlowLayout(FlowLayout.TRAILING));
				buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));

				JButton plotButton = new JButton("Plot");
				plotButton.setToolTipText("Plot");
				plotButton.addActionListener(new CAPEListener(_gui));
				plotButton.setActionCommand("PLOT_SYSTEM_STATS_OK");
				plotButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				JButton cancelButton = new JButton("Cancel");
				cancelButton.setToolTipText("Cancel");
				cancelButton.addActionListener(new CAPEListener(_gui));
				cancelButton.setActionCommand("PLOT_SYSTEM_STATS_CANCEL");
				cancelButton.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(3,10,3,10)));

				buttonPanel.add(plotButton);
				buttonPanel.add(cancelButton);

				plotPanel.add(leftPanel, BorderLayout.LINE_START); //statistics
				plotPanel.add(rightPanel, BorderLayout.LINE_END); //operator and update frequency
				plotPanel.add(buttonPanel, BorderLayout.PAGE_END);

				return plotPanel;
			}
		}catch(Exception e){System.err.println("PanelMaker: Error creating PLOT_SYSTEM_STATS panel.");}
		//never gets here
		return new JPanel();
	}
	//-------------------------------------------------------------------------
	/**
	 * Creates the panel for the user to select which file
	 * contains statistics to show in the tabbed pane.
	 * @return java.awt.Component */
	private java.awt.Component makePlotFileStatsPanel(){
		//System.out.println("******** GOES IN NEW PANEL **********");
		OptionsStorage os = _gui.getOptionsStorage();
		JPanel selectFilePanel = new JPanel();
		selectFilePanel.setOpaque(true);
		selectFilePanel.setLayout(new BorderLayout());
		selectFilePanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));

		// open file browser

		JFileChooser fileBrowser = new JFileChooser();
		fileBrowser.setMultiSelectionEnabled(true);
		fileBrowser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		fileBrowser.setApproveButtonText("Select");
		fileBrowser.setControlButtonsAreShown(true);
		fileBrowser.setDialogTitle("Select Statistics File(s)");
		fileBrowser.setCurrentDirectory(new File("."));
		fileBrowser.setFileFilter(new CSVFileFilter());
		//os.setOptions("FILE_BROWSER", null);
		os.setOptions("FILE_BROWSER", fileBrowser);
		fileBrowser.showOpenDialog(selectFilePanel);

		try {

			//create the file selection panel
			JPanel filePanel = new JPanel();
			filePanel.setOpaque(false);
			filePanel.setLayout(new BoxLayout(filePanel, BoxLayout.PAGE_AXIS));
			filePanel.setBorder(BorderFactory.createTitledBorder("Plot From File(s)"));

			JTextArea filesTextArea = new JTextArea(5, 40);
			filesTextArea.setLineWrap(false);
			JScrollPane scrollPane = new JScrollPane(filesTextArea,
								JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
								JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
			String file = new String();
			for (int i = 0, pos = 0; i < fileBrowser.getSelectedFiles().length; i++, pos++) {
				file = fileBrowser.getSelectedFiles()[i].getPath();
				filesTextArea.insert(file, pos);
				pos += file.length();
				filesTextArea.insert("\n", pos);
			}

			// store this option in OptionStorage
			os.setOptions("PLOT_FILE_STATS_SCROLL_PANE", null);
			os.setOptions("PLOT_FILE_STATS_SCROLL_PANE", scrollPane);

			filePanel.add(scrollPane);
			//create buttons and add to the bottom of customizePanel
			JPanel buttonPanel = new JPanel();
			buttonPanel.setOpaque(false);
			buttonPanel.setLayout(new FlowLayout(FlowLayout.TRAILING));
			buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));

			JButton browseButton = new JButton("Browse...");
			browseButton.setToolTipText("Browse...");
			browseButton.addActionListener(new CAPEListener(_gui));
			browseButton.setActionCommand("PLOT_FILE_STATS_BROWSE");
			browseButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton okButton = new JButton("Ok");
			okButton.setToolTipText("Ok");
			okButton.addActionListener(new CAPEListener(_gui));
			okButton.setActionCommand("PLOT_FILE_STATS_OK");
			okButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton cancelButton = new JButton("Cancel");
			cancelButton.setToolTipText("Cancel");
			cancelButton.addActionListener(new CAPEListener(_gui));
			cancelButton.setActionCommand("PLOT_FILE_STATS_CANCEL");
			cancelButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			buttonPanel.add(browseButton);
			buttonPanel.add(okButton);
			buttonPanel.add(cancelButton);

			selectFilePanel.add(filePanel, BorderLayout.PAGE_START);
			selectFilePanel.add(buttonPanel, BorderLayout.PAGE_END);

		} catch (Exception e) {
			System.err.println("PanelMaker: Error creating SELECT_FILE_STATS panel.");
		}

		return selectFilePanel;
	}
	//-------------------------------------------------------------------------
	/**
	 * Creates the panel for the user to select the statistics from
	 * the user-specified file that should be plotted.
	 * @return java.awt.Component */
	private java.awt.Component makePlotFileStatsOptionsPanel() {
		OptionsStorage os = _gui.getOptionsStorage();
		JPanel plotPanel = new JPanel(); //main content panel
		plotPanel.setOpaque(true);
		plotPanel.setLayout(new BorderLayout());
		plotPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
		plotPanel.setSize(400, 300);

		// get the user's file choice
		JScrollPane scrollPane = (JScrollPane)os.getOptions("PLOT_FILE_STATS_SCROLL_PANE");
		JTextArea filesTextArea = (JTextArea)scrollPane.getViewport().getView();
		// get the file names into an array, create PlotFromFile's, and store them
		String[] files = filesTextArea.getText().split("\n");
		ArrayList plotFiles = new ArrayList();
		for (int i = 0; i < files.length; i++) {
			PlotFromFile plotFile = new PlotFromFile(files[i], _gui);
			plotFiles.add(plotFile);
		}
		os.setOptions("PLOT_FILE_STATS_FILES", plotFiles);
		//System.out.println(((PlotFromFile)plotFiles.get(0)).getName());
		// create the options selection pane
		//--------------- Left Panel (statistics) ---------------//
		// create the stats with check boxes pane
		JPanel statsPanel = new JPanel();
		statsPanel.setOpaque(false);
		statsPanel.setLayout(new BoxLayout(statsPanel, BoxLayout.PAGE_AXIS));
		// get a file to work with; note: assuming that all files will have the
		// same available stats
		PlotFromFile file = (PlotFromFile)plotFiles.get(0);
		// initialize the statistics
		file.initStats();
		// now, get the available options
		ArrayList options = ((PlotFromFileQueryObject)file.getQueryObjects().get(0)).getAvailableOptions();
		// this will store the checkboxes
		ArrayList plotStatBoxes = new ArrayList();
		// create check box for each option
		for(int i = 0; i < options.size(); i++){
			JCheckBox box = new JCheckBox();
			String property = (String)options.get(i);
			box.setText(property);
			box.setHorizontalTextPosition(SwingConstants.LEADING);
			box.setAlignmentX(JCheckBox.RIGHT_ALIGNMENT);
			box.setAlignmentY(JCheckBox.TOP_ALIGNMENT);
			statsPanel.add(box);
			box.setSelected(false);
			plotStatBoxes.add(box);
		}
		// store the options boxes
		os.setOptions("PLOT_FILE_STATS_BOXES", plotStatBoxes);

		// create buttons for select stats panel
		JPanel statsButtonPanel = new JPanel();
		statsButtonPanel.setOpaque(false);
		statsButtonPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));

		JButton selectAllButton = new JButton("Select All");
		selectAllButton.setToolTipText("Select All Statistics");
		selectAllButton.addActionListener(new CAPEListener(_gui));
		selectAllButton.setActionCommand("PLOT_FILE_STATS_SELECT_ALL");
		selectAllButton.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createRaisedBevelBorder(),
			BorderFactory.createEmptyBorder(3,10,3,10)));

		JButton deselectAllButton = new JButton("Deselect All");
		deselectAllButton.setToolTipText("Deselect All Statistics");
		deselectAllButton.addActionListener(new CAPEListener(_gui));
		deselectAllButton.setActionCommand("PLOT_FILE_STATS_DESELECT_ALL");
		deselectAllButton.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createRaisedBevelBorder(),
			BorderFactory.createEmptyBorder(3,10,3,10)));

		statsButtonPanel.add(selectAllButton);
		statsButtonPanel.add(deselectAllButton);

		JPanel leftPanel = new JPanel(new BorderLayout());
		leftPanel.setOpaque(false);
		leftPanel.setBorder(BorderFactory.createTitledBorder("Statistics"));
		leftPanel.add(statsPanel, BorderLayout.PAGE_START);
		leftPanel.add(statsButtonPanel, BorderLayout.PAGE_END);

		//--------------- Right Panel (query objects and update frequency) ---------------//
		// create query objects panel
		JPanel qoPanel = new JPanel(new BorderLayout());
		qoPanel.setOpaque(false);
		qoPanel.setBorder(BorderFactory.createTitledBorder("Query Objects"));
		JTabbedPane qoPane = new JTabbedPane(JTabbedPane.TOP, JTabbedPane.WRAP_TAB_LAYOUT);
		Hashtable plotQOBoxes = new Hashtable();
		// create a tab for each file
		for (int i = 0; i < plotFiles.size(); i++) {
			PlotFromFile plotFile = (PlotFromFile)plotFiles.get(i);
			plotFile.initStats();
			String title = plotFile.getProperty("FILE_ID");
			JPanel filePanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
			filePanel.setPreferredSize(new Dimension(100,100));
			ArrayList queryObjects = plotFile.getQueryObjects();
			ArrayList fileBoxes = new ArrayList();

			// create a checkbox for each query object
			for (int j = 0; j < queryObjects.size(); j++) {
				JCheckBox box = new JCheckBox();
				String label = ((PlotFromFileQueryObject)queryObjects.get(j)).getProperty("OBJECT_ID");
				box.setText(label);
				box.setToolTipText(((PlotFromFileQueryObject)queryObjects.get(j)).getObjectString());
				box.setHorizontalTextPosition(SwingConstants.TRAILING);
				box.setAlignmentX(JCheckBox.LEFT_ALIGNMENT);
				box.setAlignmentY(JCheckBox.TOP_ALIGNMENT);
				box.setSelected(false);
				filePanel.add(box);
				fileBoxes.add(box);
				plotQOBoxes.put(queryObjects.get(j), box);
			}
			qoPane.addTab(title, null, filePanel, plotFile.getShortName());
		}
		qoPanel.add(qoPane, BorderLayout.CENTER);
		// store these options in OptionStorage
		os.setOptions("PLOT_FILE_STATS_QOBJECTS", plotQOBoxes);

		// create update frequency panel
		JPanel freqPanel = new JPanel(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		freqPanel.setOpaque(true);
		freqPanel.setBorder(BorderFactory.createTitledBorder("Update Frequency"));
		freqPanel.setMaximumSize(new Dimension(100,80));

		JLabel label0 = new JLabel(	"<html>The update frequency identifies how often to display<br>" +
									"the latest statistic value(s) on the plot.</html>");
		JLabel label1 = new JLabel("Enter update frequency in milliseconds:");
		label0.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
		label1.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));

		JTextField freqTextField = new JTextField("1000", 4);
		freqTextField.setBorder(BorderFactory.createLineBorder(Color.gray));
		freqTextField.setHorizontalAlignment(JTextField.CENTER);

		// store this option in OptionStorage
		os.setOptions("PLOT_FILE_STATS_FREQUENCY", freqTextField);

		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth = 2;
		freqPanel.add(label0, c);
		c.gridy = 1;
		c.gridwidth = 1;
		freqPanel.add(label1, c);
		c.gridx = 1;
		freqPanel.add(freqTextField, c);

		JPanel rightPanel = new JPanel(new BorderLayout());
		rightPanel.setOpaque(true);
		rightPanel.add(qoPanel, BorderLayout.CENTER);
		rightPanel.add(freqPanel, BorderLayout.PAGE_END);

		//--------------- Button/Bottom Panel ---------------//
		JPanel buttonPanel = new JPanel();
		buttonPanel.setOpaque(false);
		buttonPanel.setLayout(new FlowLayout(FlowLayout.TRAILING));
		buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));


		JButton plotButton = new JButton("Plot");
		plotButton.setToolTipText("Plot");
		plotButton.addActionListener(new CAPEListener(_gui));
		plotButton.setActionCommand("PLOT_FILE_OPTION_STATS_OK");
		plotButton.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createRaisedBevelBorder(),
			BorderFactory.createEmptyBorder(3,10,3,10)));

		JButton cancelButton = new JButton("Cancel");
		cancelButton.setToolTipText("Cancel");
		cancelButton.addActionListener(new CAPEListener(_gui));
		cancelButton.setActionCommand("PLOT_FILE_OPTION_STATS_CANCEL");
		cancelButton.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createRaisedBevelBorder(),
			BorderFactory.createEmptyBorder(3,10,3,10)));

		//buttonPanel.add(type);
		buttonPanel.add(plotButton);
		buttonPanel.add(cancelButton);

		plotPanel.add(leftPanel, BorderLayout.LINE_START);
		plotPanel.add(rightPanel, BorderLayout.LINE_END);
		plotPanel.add(buttonPanel, BorderLayout.PAGE_END);

		return plotPanel;
	}

	/** Creates the panel for the user to select which tree
	 * statistics to show in the tabbed pane.
	 * @return the panel with selectable/deselectable tree statistics */
	private java.awt.Component makeSelectTreeStatsPanel(){
		JPanel selectStatsPanel = new JPanel();
		selectStatsPanel.setOpaque(true);
		selectStatsPanel.setLayout(new BorderLayout());
		selectStatsPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
		try{
			//create the stats with check boxes panel
			JPanel statsPanel = new JPanel();
			statsPanel.setOpaque(false);
			statsPanel.setLayout(new BoxLayout(statsPanel, BoxLayout.PAGE_AXIS));
			statsPanel.setBorder(BorderFactory.createTitledBorder("Tree Statistics"));
			ArrayList statBoxes = new ArrayList();
			for(Enumeration e = _gui.getXATTree(0).getStatistics(); e.hasMoreElements();){
				JCheckBox box = new JCheckBox();
				String property = new String(e.nextElement().toString());
				box.setText(property);
				box.setHorizontalTextPosition(SwingConstants.LEADING);
				box.setAlignmentX(JCheckBox.RIGHT_ALIGNMENT);
				statsPanel.add(box);
				box.setSelected(true);
				statBoxes.add(box);
			}
			//store these options in OptionStorage
			_gui.getOptionsStorage().setOptions("SELECT_TREE_STATS", statBoxes);

			//create buttons and add to the bottom of customizePanel
			JPanel buttonPanel = new JPanel();
			buttonPanel.setOpaque(false);
			buttonPanel.setLayout(new FlowLayout(FlowLayout.TRAILING));
			buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));

			JButton selectAllButton = new JButton("Select All");
			selectAllButton.setToolTipText("Select All Statistics");
			selectAllButton.addActionListener(new CAPEListener(_gui));
			selectAllButton.setActionCommand("TREE_STATS_SELECT_ALL");
			selectAllButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton deselectAllButton = new JButton("Deselect All");
			deselectAllButton.setToolTipText("Deselect All Statistics");
			deselectAllButton.addActionListener(new CAPEListener(_gui));
			deselectAllButton.setActionCommand("TREE_STATS_DESELECT_ALL");
			deselectAllButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton okButton = new JButton("Ok");
			okButton.setToolTipText("Ok");
			okButton.addActionListener(new CAPEListener(_gui));
			okButton.setActionCommand("TREE_STATS_OK");
			okButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			buttonPanel.add(selectAllButton);
			buttonPanel.add(deselectAllButton);
			buttonPanel.add(okButton);

			selectStatsPanel.add(statsPanel, BorderLayout.PAGE_START);
			selectStatsPanel.add(buttonPanel, BorderLayout.PAGE_END);
		}catch(Exception e){System.err.println("PanelMaker: Error creating SELECT_TREE_STATS panel.");}

		return selectStatsPanel;
	}

	/** Creates the panel for the user to select which queue
	 * statistics to show in the tabbed pane.
	 * @return the panel with selectable/deselectable queue statistics */
	private java.awt.Component makeSelectQueueStatsPanel(){
		JPanel selectStatsPanel = new JPanel();
		selectStatsPanel.setOpaque(true);
		selectStatsPanel.setLayout(new BorderLayout());
		selectStatsPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
		try{
			//create the stats with check boxes panel
			JPanel statsPanel = new JPanel();
			statsPanel.setOpaque(false);
			statsPanel.setLayout(new BoxLayout(statsPanel, BoxLayout.PAGE_AXIS));
			statsPanel.setBorder(BorderFactory.createTitledBorder("Queue Statistics"));
			ArrayList statBoxes = new ArrayList();
			for(Enumeration e = ((XATStreamOperator)_gui.getXATTree(0).getRoot().getOperator())
					.getInputQueue().getStatistics(); e.hasMoreElements();){
				JCheckBox box = new JCheckBox();
				String property = new String(e.nextElement().toString());
				box.setText(property);
				box.setHorizontalTextPosition(SwingConstants.LEADING);
				box.setAlignmentX(JCheckBox.RIGHT_ALIGNMENT);
				statsPanel.add(box);
				box.setSelected(true);
				statBoxes.add(box);
			}
			//store these options in OptionStorage
			_gui.getOptionsStorage().setOptions("SELECT_QUEUE_STATS", statBoxes);

			//create buttons and add to the bottom of customizePanel
			JPanel buttonPanel = new JPanel();
			buttonPanel.setOpaque(false);
			buttonPanel.setLayout(new FlowLayout(FlowLayout.TRAILING));
			buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));

			JButton selectAllButton = new JButton("Select All");
			selectAllButton.setToolTipText("Select All Statistics");
			selectAllButton.addActionListener(new CAPEListener(_gui));
			selectAllButton.setActionCommand("QUEUE_STATS_SELECT_ALL");
			selectAllButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton deselectAllButton = new JButton("Deselect All");
			deselectAllButton.setToolTipText("Deselect All Statistics");
			deselectAllButton.addActionListener(new CAPEListener(_gui));
			deselectAllButton.setActionCommand("QUEUE_STATS_DESELECT_ALL");
			deselectAllButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton okButton = new JButton("Ok");
			okButton.setToolTipText("Ok");
			okButton.addActionListener(new CAPEListener(_gui));
			okButton.setActionCommand("QUEUE_STATS_OK");
			okButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			buttonPanel.add(selectAllButton);
			buttonPanel.add(deselectAllButton);
			buttonPanel.add(okButton);

			selectStatsPanel.add(statsPanel, BorderLayout.PAGE_START);
			selectStatsPanel.add(buttonPanel, BorderLayout.PAGE_END);
		}catch(Exception e){System.err.println("PanelMaker: Error creating SELECT_QUEUE_STATS panel.");}

		return selectStatsPanel;
	}

	/** Creates the panel for the user to select which operator
	 * statistics to show in the tabbed pane when a node is right-clicked.
	 * There must be atleast one tree open in order to view selectable statistics.
	 * @TODO Make sure OK action is ok when there are no stats.
	 * @return java.awt.Component */
	private java.awt.Component makeSelectOperatorStatsPanel(){
		JPanel selectStatsPanel = new JPanel();
		selectStatsPanel.setOpaque(true);
		selectStatsPanel.setLayout(new BorderLayout());
		selectStatsPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
		try{
			//create the stats with check boxes panel
			JPanel statsPanel = new JPanel();
			statsPanel.setOpaque(false);
			statsPanel.setLayout(new BoxLayout(statsPanel, BoxLayout.PAGE_AXIS));
			statsPanel.setBorder(BorderFactory.createTitledBorder("Statistics"));
			ArrayList statBoxes = new ArrayList();
			for(Enumeration e = _gui.getXATTree(0).getRoot().getOperator().getStatistics(); e.hasMoreElements();){
				JCheckBox box = new JCheckBox();
				String property = new String(e.nextElement().toString());
				box.setText(property);
				box.setHorizontalTextPosition(SwingConstants.LEADING);
				box.setAlignmentX(JCheckBox.RIGHT_ALIGNMENT);
				statsPanel.add(box);
				box.setSelected(true);
				statBoxes.add(box);
			}
			//store these options in OptionStorage
			_gui.getOptionsStorage().setOptions("SELECT_OPERATOR_STATS", statBoxes);

			//create buttons and add to the bottom of customizePanel
			JPanel buttonPanel = new JPanel();
			buttonPanel.setOpaque(false);
			buttonPanel.setLayout(new FlowLayout(FlowLayout.TRAILING));
			buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));

			JButton selectAllButton = new JButton("Select All");
			selectAllButton.setToolTipText("Select All Statistics");
			selectAllButton.addActionListener(new CAPEListener(_gui));
			selectAllButton.setActionCommand("OPERATOR_STATS_SELECT_ALL");
			selectAllButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton deselectAllButton = new JButton("Deselect All");
			deselectAllButton.setToolTipText("Deselect All Statistics");
			deselectAllButton.addActionListener(new CAPEListener(_gui));
			deselectAllButton.setActionCommand("OPERATOR_STATS_DESELECT_ALL");
			deselectAllButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton okButton = new JButton("Ok");
			okButton.setToolTipText("Ok");
			okButton.addActionListener(new CAPEListener(_gui));
			okButton.setActionCommand("OPERATOR_STATS_OK");
			okButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			buttonPanel.add(selectAllButton);
			buttonPanel.add(deselectAllButton);
			buttonPanel.add(okButton);

			selectStatsPanel.add(statsPanel, BorderLayout.PAGE_START);
			selectStatsPanel.add(buttonPanel, BorderLayout.PAGE_END);
		}catch(Exception e){System.err.println("PanelMaker: Error creating SELECT_OPERATOR_STATS panel.");}

		return selectStatsPanel;
	}

	/** Creates the panel for the user to select which system
	 * statistics to show in the tabbed pane.
	 * @return java.awt.Component */
	private java.awt.Component makeSelectSystemStatsPanel(){
		JPanel selectStatsPanel = new JPanel();
		selectStatsPanel.setOpaque(true);
		selectStatsPanel.setLayout(new BorderLayout());
		selectStatsPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));

		try{
			//create the stats with check boxes panel
			JPanel statsPanel = new JPanel();
			statsPanel.setOpaque(false);
			statsPanel.setLayout(new BoxLayout(statsPanel, BoxLayout.PAGE_AXIS));
			statsPanel.setBorder(BorderFactory.createTitledBorder("System Statistics"));
			ArrayList statBoxes = new ArrayList();
			for(Enumeration e = _gui.getSystemObject().getStatistics(); e.hasMoreElements();){
				JCheckBox box = new JCheckBox();
				String property = new String(e.nextElement().toString());
				box.setText(property);
				box.setHorizontalTextPosition(SwingConstants.LEADING);
				box.setAlignmentX(JCheckBox.RIGHT_ALIGNMENT);
				statsPanel.add(box);
				box.setSelected(true);
				statBoxes.add(box);
			}
			//store these options in OptionStorage
			_gui.getOptionsStorage().setOptions("SELECT_SYSTEM_STATS", statBoxes);

			//create buttons and add to the bottom of customizePanel
			JPanel buttonPanel = new JPanel();
			buttonPanel.setOpaque(false);
			buttonPanel.setLayout(new FlowLayout(FlowLayout.TRAILING));
			buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));

			JButton selectAllButton = new JButton("Select All");
			selectAllButton.setToolTipText("Select All Statistics");
			selectAllButton.addActionListener(new CAPEListener(_gui));
			selectAllButton.setActionCommand("SYSTEM_STATS_SELECT_ALL");
			selectAllButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton deselectAllButton = new JButton("Deselect All");
			deselectAllButton.setToolTipText("Deselect All Statistics");
			deselectAllButton.addActionListener(new CAPEListener(_gui));
			deselectAllButton.setActionCommand("SYSTEM_STATS_DESELECT_ALL");
			deselectAllButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton okButton = new JButton("Ok");
			okButton.setToolTipText("Ok");
			okButton.addActionListener(new CAPEListener(_gui));
			okButton.setActionCommand("SYSTEM_STATS_OK");
			okButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			buttonPanel.add(selectAllButton);
			buttonPanel.add(deselectAllButton);
			buttonPanel.add(okButton);

			selectStatsPanel.add(statsPanel, BorderLayout.PAGE_START);
			selectStatsPanel.add(buttonPanel, BorderLayout.PAGE_END);
		}catch(Exception e){System.err.println("PanelMaker: Error creating SELECT_SYSTEM_STATS panel.");}
			return selectStatsPanel;
	}

	/** Creates the statistics panel and initializes an empty tabbed pane
	 * to hold the statistic table(s).
	 * @return the statistics panel to display statistic tables */
	private java.awt.Component makeStatsPanel(){
		JPanel statsPanel = new JPanel();
		statsPanel.setOpaque(true);
		statsPanel.setLayout(new BorderLayout());
		statsPanel.setPreferredSize(new Dimension(700,170));
		statsPanel.setBorder(BorderFactory.createEmptyBorder(0,2,2,2));

		//create the panel to hold the tabbed pane
		try{
			_gui.getGraphManager().setStatsTabbedPane((JTabbedPaneCAPE)_gui.getGraphManager().initializeTabbedPaneRaindrop());
			statsPanel.add(_gui.getGraphManager().getStatsTabbedPane(), BorderLayout.CENTER);
		}catch(Exception e){System.err.println("PanelMaker: Error creating STATS panel.");}

		//create the panel to hold the buttons
		JPanel buttonPanel = new JPanel();
		buttonPanel.setOpaque(false);
		buttonPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
		buttonPanel.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createEmptyBorder(3,0,3,0),
			BorderFactory.createEtchedBorder()));

		JButton start = new JButton("Start");
		start.setToolTipText("Start Refreshing Statistics");
		start.addActionListener(new CAPEListener(_gui));
		start.setActionCommand("START_STATISTICS_TIMER");
		start.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createRaisedBevelBorder(),
			BorderFactory.createEmptyBorder(3,10,3,10)));

		JButton stop = new JButton("Stop");
		stop.setToolTipText("Stop Refreshing Statistics");
		stop.addActionListener(new CAPEListener(_gui));
		stop.setActionCommand("STOP_STATISTICS_TIMER");
		stop.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createRaisedBevelBorder(),
			BorderFactory.createEmptyBorder(3,10,3,10)));

		buttonPanel.add(start);
		buttonPanel.add(stop);

		statsPanel.add(buttonPanel, BorderLayout.PAGE_START);
		return statsPanel;
	}

	/** Creates the panel to select which statistics to plot:
	 * <ul>
	 * <li>Tree
	 * <li>Queue
	 * <li>Operator
	 * <li>System
	 * </ul>
	 * @return the panel with radio buttons to select which statistics to plot */
	private java.awt.Component makeWhatToPlotPanel(){
		JPanel mainPanel = new JPanel();
		mainPanel.setOpaque(true);
		mainPanel.setLayout(new BorderLayout());
		mainPanel.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
		mainPanel.setPreferredSize(new Dimension(200,185));

		try{
			//create panel with radio buttons to select which statistics to plot
			JPanel selectPanel = new JPanel();
			selectPanel.setOpaque(false);
			selectPanel.setLayout(new BoxLayout(selectPanel, BoxLayout.PAGE_AXIS));
			selectPanel.setBorder(BorderFactory.createTitledBorder("Select Statistics to Plot"));
			ArrayList selectBoxes = new ArrayList();//used for OptionsStorage

			JRadioButton tree = new JRadioButton("Tree", true);
			tree.setActionCommand("PLOT_TREE_STATS");
			selectPanel.add(tree);
			selectBoxes.add(tree);

			JRadioButton queue = new JRadioButton("Queue", false);
			queue.setActionCommand("PLOT_QUEUE_STATS");
			selectPanel.add(queue);
			selectBoxes.add(queue);

			JRadioButton operator = new JRadioButton("Operator", false);
			operator.setActionCommand("PLOT_OPERATOR_STATS");
			selectPanel.add(operator);
			selectBoxes.add(operator);

			JRadioButton system = new JRadioButton("System", false);
			system.setActionCommand("PLOT_SYSTEM_STATS");
			system.setSelected(true);
			selectPanel.add(system);
			selectBoxes.add(system);

			JRadioButton file = new JRadioButton("From File", false);
			file.setActionCommand("PLOT_FILE_STATS");
			selectPanel.add(file);
			selectBoxes.add(file);

			ButtonGroup group = new ButtonGroup();
			group.add(tree);
			group.add(queue);
			group.add(operator);
			group.add(system);
			group.add(file);

			//store these options in OptionStorage
			_gui.getOptionsStorage().setOptions("WHAT_TO_PLOT", selectBoxes);

			//create buttons and add to the bottom of the panel
			JPanel buttonPanel = new JPanel();
			buttonPanel.setOpaque(false);
			buttonPanel.setLayout(new FlowLayout(FlowLayout.TRAILING));
			buttonPanel.setBorder(BorderFactory.createEmptyBorder(3,0,0,0));

			JButton okButton = new JButton("Ok");
			okButton.setToolTipText("Plot Selected Statistics");
			okButton.addActionListener(new CAPEListener(_gui));
			okButton.setActionCommand("WHAT_TO_PLOT_OK");
			okButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			JButton cancelButton = new JButton("Cancel");
			cancelButton.setToolTipText("Cancel");
			cancelButton.addActionListener(new CAPEListener(_gui));
			cancelButton.setActionCommand("WHAT_TO_PLOT_CANCEL");
			cancelButton.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createRaisedBevelBorder(),
				BorderFactory.createEmptyBorder(3,10,3,10)));

			buttonPanel.add(okButton);
			buttonPanel.add(cancelButton);

			mainPanel.add(selectPanel, BorderLayout.PAGE_START);
			mainPanel.add(buttonPanel, BorderLayout.PAGE_END);
		}catch(Exception e){System.err.println("PanelMaker: Error creating WHAT_TO_PLOT panel.");}

		return mainPanel;
	}

	/** Creates the user input panel with tabs for:
	 * - operator
	 * - scheduler
	 * - migrator
	 * - distributor
	 * @return the user input panel to accept user-defined changes to the run */
	private java.awt.Component makeUserInputPanel(){
		JPanel userInputPanel = new JPanel();
		userInputPanel.setOpaque(true);
		userInputPanel.setLayout(new BorderLayout());
		userInputPanel.setPreferredSize(new Dimension(200,300));
		userInputPanel.setBorder(BorderFactory.createEmptyBorder(0,2,2,2));

		// create a tabbed panel for each type of configuration
		JTabbedPane tabsPane = new JTabbedPane(JTabbedPane.TOP);
		java.net.URL schedURL = CAPE_GUI.class.getResource("images/scheduler_small.png");
		ImageIcon sched = new ImageIcon(schedURL, "Scheduler");
		java.net.URL opURL = CAPE_GUI.class.getResource("images/operator_small.png");
		ImageIcon op = new ImageIcon(opURL, "Operator");
		java.net.URL migURL = CAPE_GUI.class.getResource("images/migration2.png");
		ImageIcon mig = new ImageIcon(migURL, "Migration");
		java.net.URL distURL = CAPE_GUI.class.getResource("images/distributor_small.png");
		ImageIcon dist = new ImageIcon(distURL, "Distribution");

		// make the scheduler panel
		JPanel schedPanel = new JPanel();
		TitledBorder schedTitle =
			BorderFactory.createTitledBorder(
				BorderFactory.createLineBorder(Color.gray),
				"Scheduler Options");
		schedPanel.setBorder(schedTitle);
		// currently running...
		//String distributionPatten = config.getProperty("DISTRIBUTION_PATTERN");
		// create the radio buttons
		JRadioButton adaptModeButton = new JRadioButton("Adapt Mode");
		adaptModeButton.setActionCommand("ADAPT_MODE");
		adaptModeButton.setSelected(true);
		JRadioButton selectModeButton = new JRadioButton("Select Algorithm");
		selectModeButton.setActionCommand("SELECT_MODE");
		// add radio buttons to button group
		ButtonGroup schedModeButtons = new ButtonGroup();
		schedModeButtons.add(adaptModeButton);
		schedModeButtons.add(selectModeButton);
		JPanel schedModeRadioPanel = new JPanel(new GridLayout(0, 1));
		schedModeRadioPanel.add(adaptModeButton);
		schedModeRadioPanel.add(selectModeButton);
		// add button group to scheduler panel
		schedPanel.add(schedModeRadioPanel);
		/*
		// create the drop-down menu for algorithms
		String[] algorithms = {"Chain", "First In, First Out", "Greedy Priority", "Most Tuples In Queue", "Probabilistic Greedy", "Push Tuples Through", "Round Robin"};
		config.getSchedulingAlgorithms()
		JComboBox petList = new JComboBox(petStrings);
		petList.setSelectedIndex(4);
		petList.addActionListener(this);
		*/

		// make the operator panel
		JPanel opPanel = new JPanel();
		TitledBorder opTitle =
			BorderFactory.createTitledBorder(
				BorderFactory.createLineBorder(Color.gray),
				"Operator Options");
		opPanel.setBorder(opTitle);

		// make the migration panel
		JPanel migPanel = new JPanel();
		TitledBorder migTitle =
			BorderFactory.createTitledBorder(
				BorderFactory.createLineBorder(Color.gray),
				"Migration Options");
		migPanel.setBorder(migTitle);

		// make the distribution panel
		JPanel distPanel = new JPanel();
		TitledBorder distTitle =
			BorderFactory.createTitledBorder(
				BorderFactory.createLineBorder(Color.gray),
				"Distribution Options");
		distPanel.setBorder(distTitle);

//		tabsPane.addTab(null, sched, schedPanel);
		tabsPane.addTab(null, op, opPanel);
		tabsPane.addTab(null, mig, migPanel);
		tabsPane.addTab(null, dist, distPanel);

		/*
		//create the panel to hold the tabbed pane
		try{
			_gui.getGraphManager().setStatsTabbedPane((JTabbedPaneCAPE)_gui.getGraphManager().initializeTabbedPaneRaindrop());
			statsPanel.add(_gui.getGraphManager().getStatsTabbedPane(), BorderLayout.CENTER);
		}catch(Exception e){System.err.println("PanelMaker: Error creating STATS panel.");}
		*/

		/*
		//create the panel to hold the buttons
		JPanel buttonPanel = new JPanel();
		buttonPanel.setOpaque(false);
		buttonPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
		buttonPanel.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createEmptyBorder(3,0,3,0),
			BorderFactory.createEtchedBorder()));

		JButton start = new JButton("Start");
		start.setToolTipText("Start Refreshing Statistics");
		start.addActionListener(new CAPEListener(_gui));
		start.setActionCommand("START_STATISTICS_TIMER");
		start.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createRaisedBevelBorder(),
			BorderFactory.createEmptyBorder(3,10,3,10)));

		JButton stop = new JButton("Stop");
		stop.setToolTipText("Stop Refreshing Statistics");
		stop.addActionListener(new CAPEListener(_gui));
		stop.setActionCommand("STOP_STATISTICS_TIMER");
		stop.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createRaisedBevelBorder(),
			BorderFactory.createEmptyBorder(3,10,3,10)));

		buttonPanel.add(start);
		buttonPanel.add(stop);
		*/

		/* statsPanel.add(buttonPanel, BorderLayout.PAGE_START); */
		userInputPanel.add(tabsPane, BorderLayout.PAGE_START);

		return userInputPanel;
	}

	//----------------------------------------------------------------------------------
	
	private java.awt.Component makeDeviceStreamShieldPanel(Stream [] streams)
	{
		DeviceStreamShieldMainForm mainFrame = 
		new edu.wpi.cs.dsrg.xmldb.xat.component.gui.security.DeviceStreamShieldMainForm();		
		return mainFrame;
	}
	
	//----------------------------------------------------------------------------------
	
	private java.awt.Component makeServerStreamShieldPanel(Stream [] streams)
	{
		ServerStreamShieldMainForm mainFrame = new edu.wpi.cs.dsrg.xmldb.xat.component.gui.security.ServerStreamShieldMainForm(streams);		
		return mainFrame;
	}

	//----------------------------------------------------------------------------------
	
	private java.awt.Component makeClientStreamShieldPanel(Stream [] streams)
	{
/** CHUAN LEI - Hotfix as *.Boston.SplinInterpolatorTest.java has been removed.*/
		return null;
//		SplineInterpolatorTest mainFrame = new edu.wpi.cs.dsrg.xmldb.xat.component.gui.security.Boston.SplineInterpolatorTest();
//		return mainFrame;
	}
}