/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
/*
 * Created on Aug 3, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObjectImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.storage.OptionsStorage;

/**
 * @author natasha
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class PlotFromFile extends XATQueryObjectImp {
	
	/** the current timestamp **/
	protected int currTimestamp;
	/** the current distribution pattern **/
	private String distPattern;
	/** the actual file that is being read from **/
	private File file;
	/** the ID's for all the files **/
	private static Hashtable fileIDs = null;
	/** a reference to the current gui **/
	protected CAPE_GUI gui;
	/** the next ID that should be used when a new file is created **/
	private static Integer nextID = null;
	/** the previous timestamp, used to normalize values **/
	protected int prevTimestamp;
	/** the available propeties for this file **/
	private Hashtable properties;
	/** a list of query ojects that are available for this file **/
	private ArrayList queryObjects;
	/** the reader used to read from the file **/
	private BufferedReader reader;
	/** a flag indicating whether the header has already been read **/
	private boolean readHeader;
		
	/**
	 * Constructor for PlotFromFile. Initializes the file and creates a 
	 * BufferedReader which will be used to get statistics.
	 * 
	 * Creation Date: (Aug 5, 2004 11:56:54 AM)
	 * @param filename String
	 * @param gui CAPE_GUI
	 */
	public PlotFromFile(String filename, CAPE_GUI guiInit) {
		// create the file
		this.file = new File(filename);
		// create the reader for this file
		try {
			this.reader = new BufferedReader(new FileReader(this.file));
		} catch (Exception e) {
			e.printStackTrace();
		}
		// set the ID for this file
		if (nextID == null) {
			nextID = new Integer(0);
		}
		if (fileIDs == null) {
			fileIDs = new Hashtable();
		}
		fileIDs.put(this, nextID);
		// increment the next file ID
		nextID = new Integer(nextID.intValue() + 1);
		// initialize the properties for this file
		this.setProperties();
		// set the gui
		this.gui = guiInit;	
		this.readHeader = false;
	}
	
	/**
	 * Returns the current distribution pattern for the statistics we are
	 * reading.
	 * 
	 * Creation Date: (Aug 5, 2004 12:00:12 PM)
	 * @return String
	 */
	public String getDistPattern() {
		return this.distPattern;
	}

	/**
	 * Gets the LightWeightQueryObject representation of this file.
	 * 
	 * @return LightWeightQueryObject
	 */
	public LightWeightQueryObject getLightWeightQueryObject() {
		return new LightWeightQueryObject(this.getName(), ((Integer)fileIDs.get(this)).intValue(), this.getClass().getName(), true);
	}
	
	/**
	 * Returns the full name of the file.
	 * 
	 * Creation Date: (Aug 5, 2004 11:57:36 AM)
	 * @return String
	 */
	public String getName() {
		return this.file.getName();
	}
	
	/**
	 * Gets the next set of statistics from the file.
	 * 
	 * Creation Date: (Aug 5, 2004 12:00:35 PM)
	 * @return ArrayList
	 */
	public void getNextStats() {
		try {
			// get the options
			OptionsStorage os = this.gui.getOptionsStorage();
			ArrayList options = (ArrayList)os.getOptions("PLOT_FILE_STATS_OPTIONS"); 
			// read the first two lines; this is to check for end of file
			String firstLine = this.reader.readLine();
			String secondLine = this.reader.readLine();
			// now, get the statistics
			if (!(firstLine == null) && !(secondLine == null)) {
				int index = firstLine.indexOf(",");
				// first line is the distribution pattern
				if (index != -1) this.distPattern = firstLine.substring(0, index);
				else this.distPattern = firstLine;
				// second line is the timestamp
				index = secondLine.indexOf(",");
				this.prevTimestamp = this.currTimestamp;
				if (index != -1) this.currTimestamp = Integer.parseInt(secondLine.substring(11, index));
				else this.currTimestamp = Integer.parseInt(secondLine.substring(11));
				// third line might be header
				if (options == null) {
					options = new ArrayList();
					String[] tempOptions = this.reader.readLine().substring(12).split(",");
					// convert tempOptions to an ArrayList
					for (int i = 0; i < tempOptions.length; i++) {
						options.add(tempOptions[i]);
					}
					os.setOptions("PLOT_FILE_STATS_OPTIONS", options);
					readHeader = true;
				}
				else if (!readHeader) {
					// skip the header
					reader.readLine();
					readHeader = true;
				}
				// now read in stats and query objects, if necessary
				String nextLine = this.reader.readLine();	
				if (this.queryObjects == null) {
					this.queryObjects = new ArrayList();
					while (nextLine.split(",").length > 1) { 
						String[] newStats = new String[options.size()];					
						String[] nextLineInfo = nextLine.split(",");						
						// get the new stats
						for (int i = 1; i < nextLineInfo.length; i++) {
							newStats[i-1] = nextLineInfo[i];
						}
						// add a new query object
						PlotFromFileQueryObject qo = 
							new PlotFromFileQueryObject(this, nextLineInfo[0].split(" "));
						this.queryObjects.add(qo);
						// set the query object's stats
						qo.setStats(newStats);
						nextLine = this.reader.readLine();				
					}
				}
				else {
					for (int i = 0; i < this.queryObjects.size(); i++) {
						String[] newStats = new String[options.size()];					
						String[] nextLineInfo = nextLine.split(",");
						// get the new stats
						for (int j = 1; j < nextLineInfo.length; j++) {
							newStats[j-1] = nextLineInfo[j];
						}
						// set the query object's stats
						PlotFromFileQueryObject qo = (PlotFromFileQueryObject)this.queryObjects.get(i);
						qo.setStats(newStats);
						nextLine = this.reader.readLine();
					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	* Returns an enumeration of property names supported by this query object.
	* 
	* @return Enumeration     
	*/
	public Enumeration getProperties() {
		return this.properties.elements();
	}
	
	/**
	* Retrieves the value of the property with the given name.
	* 
	* @param name String
	* @return String
	* @throws IllegalArgumentException
	*/
	public String getProperty (String name) throws IllegalArgumentException {
		if (this.isValidPropertyName(name)) {
			return (String)this.properties.get(name);
		}
		else {
			throw new IllegalArgumentException();
		}		
	}

	/**
	 * Gets this file's query objects.
	 * 
	 * @return ArrayList
	 */
	public ArrayList getQueryObjects() {
		return this.queryObjects;
	}

	/**
	 * Returns the short file name, i.e. just the filename, without the path.
	 * 
	 * Creation Date: (Aug 5, 2004 11:59:16 AM)
	 * @return String
	 */
	public String getShortName() {
		return this.file.getName().substring(
			this.file.getName().lastIndexOf(File.separator)+1, 
			this.file.getName().length());
	}
	
	/**
	 * Returns the current timestamp for the statistics we are reading.
	 * 
	 * Creation Date: (Aug 5, 2004 11:59:45 AM)
	 * @return int
	 */
	public int getTimestamp() {
		return this.currTimestamp;
	}
	
	/**
	 * Gets the closest timestamp to the value passed in.
	 * 
	 * @param val int
	 * @return int
	 */
	public int getTimestamp(int val) {
		if (val - this.prevTimestamp < this.currTimestamp - val) {
			return this.prevTimestamp;
		}
		else {
			return this.currTimestamp;
		}
	}
	
	/**
	 * Initializes the statistics for this file, if necessary.
	 */
	public void initStats() {
		OptionsStorage os = this.gui.getOptionsStorage();
		//ArrayList options = (ArrayList)os.getOptions("PLOT_FILE_STATS_OPTIONS");
		if (this.queryObjects == null /*|| options == null*/) this.getNextStats();
	}

	/**
	 * Checks if the given property is valid for this QueryObject.
	 * 
	 * @param nameToCheck String
	 * @return boolean
	 */
	public boolean isValidPropertyName(String nameToCheck) {
		return this.properties.containsKey(nameToCheck);
	}
		
	/**
	 * Sets the properties for this PlotFromFile object.
	 * 
	 * Creation Date: (Aug 6, 2004 8:46:09 AM)
	 */
	private void setProperties() {
		this.properties = new Hashtable();
		this.properties.put("FILE_ID", "F" + fileIDs.get(this).toString());
	}
	
	/**
	* Sets the given property to the new value.
	* 
	* @param name String
	* @param value String
	* @throws Exception      
	*/
	public void setProperty(String name, String value) throws Exception {
		if (!this.isValidPropertyName(name)) {
			throw new Exception("Invalid property name.");
		}
		else {
			this.properties.put(name, value);
		}
	}
    
	/**
	 * This method is just used for testing this class.
	 *  
	 * @param args String[]
	 */
	public static void main(String[] args) {
		
		Hashtable hash = new Hashtable();
		for (int i = 0; i < 10; i++) {
			hash.put(String.valueOf(i), new Integer(i*100));
		}
		for (Enumeration e = hash.elements(); e.hasMoreElements();) {
			System.out.println(e.nextElement());
		}
		//String[] test = {"one", "two", "three"};
		//System.out.println(test);
		/*
		PlotFromFile f = new PlotFromFile("stats/test.csv");
		for (int n = 0; n < 10; n++) {
			f.getNextStats();
			for (int i = 0; i < f.getQueryObjects().size(); i++) {
				//System.out.println(((PlotFromFileQueryObject)f.getQueryObjects().get(i)).toString() + "\n");
				try {
					System.out.println(
						((PlotFromFileQueryObject)f.getQueryObjects().get(i)).getStatistic(f.getOptions()[1]));
				} catch (Exception e) {
					System.out.println(e.toString());
				}
			}
		}*/
	}
}
