package edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.swing.JTextField;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObjectImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.exceptions.InvalidOptionException;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.storage.OptionsStorage;

/**
 * @author natasha
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class PlotFromFileQueryObject extends XATQueryObjectImp {

	/** meta data about the query object, as read off from file **/
	protected String[] object;
	/** the current stats for this object, as read off from file **/
	protected Hashtable currentStats;
	/** the previous set of stats for this object, used for interpolating the values **/
	protected Hashtable previousStats;
	/** the parent file of this object **/
	protected PlotFromFile parent;
	/** the available propeties for this query object **/
	private Hashtable properties;
	/** the last x value of this query object **/
	private Point2D.Double prevPoint;
	
	/**
	 * The constructor takes in info about the object.
	 * 
	 * @param objectInit String[]
	 */
	public PlotFromFileQueryObject(PlotFromFile parentInit, String[] objectInit) {
		this.parent = parentInit;
		this.object = objectInit;
		this.setProperties();
		this.prevPoint = new Point2D.Double(0.0,0.0);
	}
	
	/**
	 * Gets the available options for this query object. Note: an invalid
	 * option is one that begins with a "-".
	 * 
	 * @return String[]
	 */
	public ArrayList getAvailableOptions() {
		// get all the options from the OptionsStorage
		this.parent.initStats();
		OptionsStorage os = parent.gui.getOptionsStorage();
		ArrayList options = (ArrayList)os.getOptions("PLOT_FILE_STATS_OPTIONS");
		// get the available options for this object
		ArrayList availableOptions = new ArrayList();
		if (this.currentStats == null) parent.getNextStats();
		for (int i = 0; i < options.size(); i++) {
			if (this.isValidOption((String)options.get(i))) {
				availableOptions.add(options.get(i));
			}
		}
		return availableOptions;
	}
	
	/**
	 * Gets the LightWeightQueryObject representation of this file.
	 * 
	 * @return LightWeightQueryObject
	 */
	public LightWeightQueryObject getLightWeightQueryObject() {
		return new LightWeightQueryObject(this.getProperty("OBJECT_ID"), 0, this.getClass().getName(), true);
	}
	
	/**
	 * Gets the next point. Uses linear interpolation to get points inbetween
	 * recorded values.
	 * 
	 * @return Point
	 */
	public Point2D.Double getNextPoint(String option) {
		try {
			// get the update frequency from OptionsStorage
			OptionsStorage os = this.parent.gui.getOptionsStorage();
			double frequency = Double.parseDouble(((JTextField)os.getOptions("PLOT_FILE_STATS_FREQUENCY")).getText());
			// set the next x value
			double nextX = this.prevPoint.x + frequency;
			// update statistics if necessary
			if (nextX > this.parent.currTimestamp) this.parent.getNextStats();
			
			// set the current x and y values
			double currX = this.parent.currTimestamp;
			double currY = this.getStatistic(option);
			
			// get the previous x and y values
			double prevX = prevPoint.x;
			double prevY = prevPoint.y;
			
			// now, calculate the nextY value
			double nextY = (currY - prevY)/(currX - prevX) * nextX;
			
			// update the prevPoint value
			this.prevPoint = new Point2D.Double(nextX, nextY);
			
			return prevPoint;
		} catch (InvalidOptionException ioe) {
			ioe.printStackTrace();
			return prevPoint;
		}
	}
	
	/**
	 * Gets this object's meta data, represented as an array of strings.
	 * 
	 * @return String[]
	 */	
	public String[] getObject() {
		return this.object;
	}
	
	/**
	 * Gets this object's meta data, represented as a single string.
	 * 
	 * @return String
	 */
	public String getObjectString() {
		// create the string, with each value in the array separated by a space
		String str = new String();
		for (int i = 0; i < this.object.length; i++) {
			str += this.object[i] + " ";
		}
		return str;
	}
	
	/**
	 * Returns the parent file of this query object.
	 * 
	 * @return PlotFromFile
	 */
	public PlotFromFile getParent() {
		return this.parent;
	}
	
	/**
	* Returns an enumeration of property names supported by this query object.
	* 
	* @return Enumeration     
	*/
	public Enumeration getProperties() {
		return this.properties.elements();
	}
	
	/**
	* Retrieves the value of the property with the given name.
	* 
	* @param name String
	* @return String
	* @throws IllegalArgumentException
	*/
	public String getProperty (String name) throws IllegalArgumentException {
		if (this.isValidPropertyName(name)) {
			return (String)this.properties.get(name);
		}
		else {
			throw new IllegalArgumentException();
		}		
	}
	
	/**
	 * If the statistic is valid, returns the integer value for this statistic;
	 * otherwise, throws an exception.
	 * 
	 * @param option String
	 * @return int
	 */
	public double getStatistic (String option) throws InvalidOptionException {
		// check if the option is valid; if not, throw exception
		// note: this assumes all statistics can be parsed to a double value
		if (this.isValidOption(option)) {
			return Double.parseDouble((String)currentStats.get(option));
		}
		else {
			throw new InvalidOptionException("Invalid option.");
		}
	}
	
	/**
	 * Checks whether the given option is valid for this query object.
	 * 
	 * @param option String
	 * @return boolean
	 */
	public boolean isValidOption(String option) {
		if (currentStats != null &&
			currentStats.get(option) != null && 
			!((String)currentStats.get(option)).startsWith("-")) {
			return true;
		}
		else return false;
	}
	
	/**
	 * Checks if the given property is valid for this QueryObject.
	 * 
	 * @param nameToCheck String
	 * @return boolean
	 */
	public boolean isValidPropertyName(String nameToCheck) {
		return this.properties.containsKey(nameToCheck);
	}
		
	/**
	 * Sets the properties for this PlotFromFile object.
	 * 
	 * Creation Date: (Aug 6, 2004 8:46:09 AM)
	 */
	private void setProperties() {
		this.properties = new Hashtable();
		String id = new String();
		// if it's an operator, use the number as an ID
		if (this.object.length > 2) id = "O" + this.object[1];
		// otherwise, just use the name of the object
		else id = this.getObjectString();
		this.properties.put("OBJECT_ID", id);
		this.properties.put("OBJECT_LEGEND_ID", 
							this.parent.getProperty("FILE_ID") + "_" + id);
	}
	
	/**
	* Sets the given property to the new value.
	* 
	* @param name String
	* @param value String
	* @throws Exception      
	*/
	public void setProperty(String name, String value) throws Exception {
		if (!this.isValidPropertyName(name)) {
			throw new Exception("Invalid property name.");
		}
		else {
			this.properties.put(name, value);
		}
	}
		
	/**
	 * Sets the most recent statistics for this query object.
	 * 
	 * @param stats String[]
	 */
	public void setStats(String[] stats) {
		// get the available options from the OptionsStorage
		this.parent.initStats();
		OptionsStorage os = parent.gui.getOptionsStorage();
		ArrayList options = (ArrayList)os.getOptions("PLOT_FILE_STATS_OPTIONS");
		// set the current stats as the previous
		if (this.currentStats == null) this.currentStats = new Hashtable();
		else {
			if (this.previousStats == null) this.previousStats = new Hashtable();
			this.previousStats = this.currentStats;
		}
		// put the new stats into the current stats table		
		for (int i = 0; i < options.size(); i++) {
			currentStats.put(options.get(i), stats[i]);
		}
	}
	
	/**
	 * Returns the string value of this query object and its current statistics.
	 * 
	 * @return String
	 */
	public String toString() {
		String str = this.getObjectString() + " [";
		if (this.currentStats != null) {
			try {
				// get the options
				this.parent.initStats();
				OptionsStorage os = parent.gui.getOptionsStorage();
				ArrayList options = (ArrayList)os.getOptions("PLOT_FILE_STATS_OPTIONS");
				// add the statistics
				for (int i = 0; i < options.size()-1; i++) {
					str += this.getStatistic((String)options.get(i)) + ", ";
				}
				str += this.getStatistic((String)options.get(options.size()-1));
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		str += "]\n";
		
		return str;
	}
}
