/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.migrationcontroller;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATBox;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
/**
 * Insert the type's description here.
 * Creation date: (4/20/2003 12:11:58 PM)
 * @author: Yali Zhu
 */
public class ExecController extends XATPropertiesImp implements Runnable{
	//private XATTree tree;
	private XATDag dag;
	
	private double dequeueFactor;
	private int dequeueHowmany;
	private boolean debug = false;
	private double yield;
	private StatsCollector statsCollector;
	private Thread streamThread = null;
	private ArrayList operators = new ArrayList();
	private boolean suspendFlag;
/**
 * Insert the method's description here.
 * Creation date: (5/21/2003 3:44:17 AM)
 */
public ExecController(XATDag dag) {

    super();
    this.dag = dag;

    XATNode[] nodes = dag.convertToArray();
    int size = nodes.length;

    statsCollector = StatsCollector.getStatsCollector();
    statsCollector.setTree(dag);

    operators.clear();
    for (int i = 0; i < size; i++)
        operators.add((XATStreamOperator)nodes[i].getOperator());
}
/**
 * Insert the method's description here.
 * Creation date: (4/22/2003 12:04:33 PM)
 */
public void addOperatorsToRun(
    XATStreamOperator[] newOperators,
    XATStreamOperator[] newRootOperators) {

    statsCollector.addOperatorsToCollect(newRootOperators);
    int size = newOperators.length;
    XATStreamOperator currentOperator;

    try {
        for (int i = 0; i < size; i++) {
            if (!operators.contains(newOperators[i])) {
                currentOperator = newOperators[i];
                currentOperator.setProperty("DEQUEUE_FACTOR", "" + dequeueFactor);
                currentOperator.setProperty("NUMBER_OF_TUPLES_TO_DEQUEUE", "" + dequeueHowmany);
                //currentOperator.setDebug(debug);
                /** if it's a window binary join operator, initialize it based on join predicate */
                /*if (currentOperator instanceof WBinaryJoinOperator) {
                    ((WBinaryJoinOperator) currentOperator).setTimeWindow(
                        Long.parseLong(getProperty("GLOBAL_TIME_WINDOW")));
                }*/
                operators.add(currentOperator);
            }
        }
    } catch (Exception e) {
        System.out.println(e);
    }
}
/**
 * Insert the method's description here.
 * Creation date: (5/20/2003 6:08:04 AM)
 */
public void addQueuesToCollect(XATQueue[] newQueues) {

    statsCollector.addQueuesToCollect(newQueues);
}
/**
 * Insert the method's description here.
 * Creation date: (6/23/2003 5:32:30 AM)
 */
public void addStatesToCollect(WOperatorState[] newStates) {

    statsCollector.addStatesToCollect(newStates);
}
/**
 * Insert the method's description here.
 * Creation date: (6/26/2003 11:26:00 PM)
 */
public synchronized void execResume() {
	suspendFlag = false;
	notify();
}
/**
 * Insert the method's description here.
 * Creation date: (6/26/2003 11:25:49 PM)
 */
public void execSuspend() {
    suspendFlag = true;
}
/**
 * Insert the method's description here.
 * Creation date: (4/22/2003 12:05:09 PM)
 */
public void removeOperatorsToRun(XATStreamOperator[] oldOperators, XATStreamOperator[] oldRootOperators) {

	statsCollector.removeOperatorsToCollect(oldRootOperators);
    int size = oldOperators.length;

    for (int i = 0; i < size; i++) {
        if (operators.contains(oldOperators[i]))
            operators.remove(oldOperators[i]);
    }
}
/**
 * Insert the method's description here.
 * Creation date: (4/22/2003 12:48:31 PM)
 */
public void removeQueuesToCollect(XATQueue[] oldQueues) {

    statsCollector.removeQueuesToCollect(oldQueues);
}
/**
 * Insert the method's description here.
 * Creation date: (6/23/2003 5:44:52 AM)
 * @param oldStates queryplanelements.WState[]
 */
public void removeStatesToCollect(WOperatorState[] oldStates) {
    statsCollector.removeStatesToCollect(oldStates);
}
/**
 * Insert the method's description here.
 * Creation date: (4/20/2003 12:13:16 PM)
 */
public void roundrobinSchedule() {

    Random rand = new Random();
    int finishCount = 0;
    XATStreamOperator currentOperator;

    /** initialize all stream operators' dequeue factor and debug flag */
    int size = operators.size();
    try {
        for (int i = 0; i < size; i++) {
            currentOperator = (XATStreamOperator) operators.get(i);
            currentOperator.setProperty("DEQUEUE_FACTOR", "" + dequeueFactor);
            currentOperator.setProperty("NUMBER_OF_TUPLES_TO_DEQUEUE", "" + dequeueHowmany);
            //currentOperator.setDebug(debug);

            /** if it's a window binary join operator, initialize it based on join predicate */
            /*if (currentOperator instanceof WBinaryJoinOperator) {
                ((WBinaryJoinOperator) currentOperator).setTimeWindow(
                    Long.parseLong(getProperty("GLOBAL_TIME_WINDOW")));
            }*/
        }
    } catch (Exception e) {
        System.out.println(e);
    }

    /** start schedule each operators in order of positions in operators list */
    while (finishCount < operators.size()) {
        finishCount = 0;
        try {
            for (int i = 0; i < operators.size(); i++) {
                currentOperator = (XATStreamOperator) operators.get(i);
                if (debug)
                    System.out.println("Run operator " + currentOperator);
                /* if nothing in operator's input queue */
                if (!((WBinaryJoinOperator)currentOperator).run()) {
                    finishCount++;
                    if (debug)
                        System.out.println("Nothing to run for operator " + currentOperator);
                }

                Thread.sleep(10);
                synchronized (this) {
                    while (suspendFlag) {
                        System.out.println("Exec Thread Suspended!");
                        System.setProperty("ProceedFlag", "true");
                        wait();
                        System.out.println("Exec Thread Resumed!");
                    }
                }

                if (finishCount == operators.size() && streamThread.isAlive()) {
                    finishCount = 0;
                    //Thread.sleep(10);
                }
            }
        } catch (Exception e) {
            System.out.println(e);
            return;
        }
    }
    //if (debug)
    System.out.println("** Execution controller finished.");
    return;
}
/**
 * Insert the method's description here.
 * Creation date: (4/20/2003 2:19:54 PM)
 */
public void run() {
    if (getProperty("DEBUG") == "true")
        debug = true;
    yield = Double.parseDouble(getProperty("YIELD_RATE"));
    dequeueFactor = Double.parseDouble(getProperty("DEQUEUE_FACTOR"));
    dequeueHowmany = Integer.parseInt(getProperty("DEQUEUE_HOWMANY"));
    //System.out.println("Window Size = " + Long.parseLong(getProperty("GLOBAL_TIME_WINDOW")));

      /* if streamThread is not assigned, stop execution and return */
    if(streamThread == null)
    {    
	    System.out.println("Haven't set stream thread for ExecController!");
	    return;
    }
        
    try {
        statsCollector.setProperty("DEBUG", getProperty("DEBUG"));
    } catch (Exception e) {
        System.out.println(e);
        return;
    }
    //statsCollector = StatsCollector.getStatsCollector();

    Thread statsT = new Thread(statsCollector);
    statsT.setName("Stats Collector");
    statsT.start();
    roundrobinSchedule();
    statsT.interrupt();
}
/**
 * Insert the method's description here.
 * Creation date: (5/20/2003 4:00:52 AM)
 */
public void setDefaultProperties() {
    addProperty("DEBUG", "false");
    addProperty("YIELD_RATE", "0.2");
    addProperty("DEQUEUE_FACTOR", "0.2");
    addProperty("DEQUEUE_HOWMANY", "30");
   //addProperty("GLOBAL_TIME_WINDOW", "1000");
}
/**
 * Insert the method's description here.
 * Creation date: (6/2/2003 2:59:48 AM)
 * @param metadata java.lang.StringBuffer
 */
public void setMetaData(StringBuffer metadata) {
    //statsCollector.addMetaData(metadata);
    statsCollector.setMetaData(metadata);
}
/**
 * Insert the method's description here.
 * Creation date: (5/24/2003 10:06:46 AM)
 * @param streamT java.lang.Thread
 */
public void setStreamThread(Thread streamT) {
    streamThread = streamT;
}
/**
 * Insert the method's description here.
 * Creation date: (4/20/2003 12:25:01 PM)
 */
public void setTree(XATDag dag) {
    this.dag = dag;
    XATStreamOperator currentOperator;

    statsCollector.setTree(dag);
    XATNode[] nodes = dag.convertToArray();
    int size = nodes.length;

    operators.clear();
    try {
        for (int i = 0; i < size; i++) {
            currentOperator = (XATStreamOperator) nodes[i].getOperator();
            currentOperator.setProperty("DEQUEUE_FACTOR", "" + dequeueFactor);
            currentOperator.setProperty("NUMBER_OF_TUPLES_TO_DEQUEUE", "" + dequeueHowmany);
            currentOperator.setDebug(debug);
            operators.add(currentOperator);
        }
    } catch (Exception e) {
        System.out.println(e);
    }
}
}
