/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.migrationcontroller;

//import streamgenerator.XATStream;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import java.io.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.client.Stream;

/**
 * Insert the type's description here.
 * Creation date: (4/20/2003 1:10:29 PM)
 * @author: Yali Zhu
 */
public class StatsCollector extends XATPropertiesImp implements Runnable {
	private static StatsCollector stats;

	/** The list of operators that we need to collect output tuple counts from.
	 *  Generally speaking, it's the root operator or the last operator that's not utility computation operator 
	 *  of the tree
	 */
	private ArrayList operatorsToCollect = new ArrayList(); // list of root operators for current tree
	private ArrayList outputQueuesToCollect = new ArrayList(); // list of output queues for current tree
	private ArrayList queuesToCollect = new ArrayList();    // list of intermediate Qs for current tree, including stream Qs
	private ArrayList statesToCollect = new ArrayList();
	private StringBuffer outputString = new StringBuffer();
	private StringBuffer metaData = new StringBuffer();
	private boolean debug = false;
	/** if some operators are removed after execController starts, we should still keep their output tuple counts */
	private int countOffset = 0;
	private XATDag dag;
/**
 * Insert the method's description here.
 * Creation date: (5/20/2003 4:03:09 AM)
 */
private StatsCollector() {
    super();
}
/**
 * Insert the method's description here.
 * Creation date: (4/22/2003 12:46:16 PM)
 */
public synchronized void addOperatorsToCollect(XATStreamOperator[] newOperators) {
    int size = newOperators.length;

    for (int i = 0; i < size; i++) {
        if (!operatorsToCollect.contains(newOperators[i]))
            operatorsToCollect.add(newOperators[i]);
    }
}
/**
 * Insert the method's description here.
 * Creation date: (5/20/2003 6:08:04 AM)
 */
public synchronized void addQueuesToCollect(XATQueue[] newQueues) {
    int size = newQueues.length;

    for (int i = 0; i < size; i++) {
        if (!queuesToCollect.contains(newQueues[i]))
            queuesToCollect.add(newQueues[i]);
    }
}
/**
 * Insert the method's description here.
 * Creation date: (6/23/2003 5:32:30 AM)
 */
public synchronized void addStatesToCollect(WOperatorState[] newStates) {
    int size = newStates.length;

    for (int i = 0; i < size; i++) {
        if (!statesToCollect.contains(newStates[i]))
            statesToCollect.add(newStates[i]);
    }
}
/**
 * Insert the method's description here.
 * Creation date: (4/20/2003 1:12:54 PM)
 * @param newInterQueueSize int
 */
private synchronized int collectInterQueueSize() {

    int count = 0;
    for (int i = 0; i < queuesToCollect.size(); i++)
        count += ((XATQueue) queuesToCollect.get(i)).getQueueSize();
    return count;
}
/**
 * Insert the method's description here.
 * Creation date: (4/20/2003 12:28:09 PM)
 * @param newOutputTupleCount java.util.ArrayList
 */
private synchronized int collectOutputTupleCounts() {
    //outputTupleCounts = newOutputTupleCounts;
    //int count = countOffset;
    /*for (int i = 0; i < operatorsToCollect.size(); i++)
        count += ((XATStreamOperator)operatorsToCollect.get(i)).getOutputTupleCount(); */

    int count = 0;
    for (int i = 0; i < outputQueuesToCollect.size(); i++)
        count += ((XATQueue)outputQueuesToCollect.get(i)).getQueueSize();
    return count;

}
/**
 * Insert the method's description here.
 * Creation date: (6/23/2003 5:43:03 AM)
 */
private synchronized int collectStateSize() {

    int count = 0;
    for (int i = 0; i < statesToCollect.size(); i++)
        count += ((WOperatorState) statesToCollect.get(i)).getStateSize();
    return count;
}
/**
 * Insert the method's description here.
 * Creation date: (5/20/2003 6:03:29 AM)
 * @return java.lang.StringBuffer
 */
public StringBuffer getOutputString() {
	return outputString;
}
/**
 * Insert the method's description here.
 * Creation date: (5/20/2003 4:03:40 AM)
 * @return queryplanelements.StatsCollector
 */
public static StatsCollector getStatsCollector() {
    if (stats == null) {
        stats = new StatsCollector();
    }
    return stats;
}
/**
 * Insert the method's description here.
 * Creation date: (4/20/2003 1:43:50 PM)
 */
public void printStats() {

	long currentTime = System.currentTimeMillis();
	String dir = "C:\\research\\implementation\\Experiments\\";
	String fileName = dir + "result" + String.valueOf(currentTime) + ".txt";
	try {
        FileWriter out = new FileWriter(fileName, true);
        /** print the outputString*/
        out.write(metaData.toString());
        out.write(outputString.toString());
        out.close();
        /** now erase the string so we dont print it out again later*/
        outputString = new StringBuffer();
    } catch (IOException e) {
        System.out.println(e.toString());
    }
}
/**
 * Insert the method's description here.
 * Creation date: (4/22/2003 12:48:31 PM)
 */
public synchronized void removeOperatorsToCollect(XATStreamOperator[] oldOperators) {

    int size = oldOperators.length;

    for (int i = 0; i < size; i++) {
        if (operatorsToCollect.contains(oldOperators[i])) {
            countOffset += Integer.parseInt(((XATStreamOperator) oldOperators[i]).getProperty("NUMBER_OF_TUPLES_OUTPUTTED"));
            //getOutputTupleCount();
            operatorsToCollect.remove(oldOperators[i]);
        }
    }
}
/**
 * Insert the method's description here.
 * Creation date: (4/22/2003 12:48:31 PM)
 */
public synchronized void removeQueuesToCollect(XATQueue[] oldQueues) {

    int size = oldQueues.length;

    for (int i = 0; i < size; i++) {
        if (queuesToCollect.contains(oldQueues[i])) {
            queuesToCollect.remove(oldQueues[i]);
        }
    }
}
/**
 * Insert the method's description here.
 * Creation date: (6/23/2003 5:44:52 AM)
 * @param oldStates queryplanelements.WState[]
 */
public synchronized void removeStatesToCollect(WOperatorState[] oldStates) {

    int size = oldStates.length;

    for (int i = 0; i < size; i++) {
        if (statesToCollect.contains(oldStates[i])) {
            statesToCollect.remove(oldStates[i]);
        }
    }
}
/**
 * Insert the method's description here.
 * Creation date: (4/20/2003 1:13:27 PM)
 */
public void run() {

    /** stats to gather */
    long currentTime;
    int outputTupleCount = 0;
    int interQueueSize = 0;
    int stateSize = 0;

    if (getProperty("DEBUG") == "true")
        debug = true;
    long updateInterval = Long.parseLong(getProperty("INTERVAL"));
    long startTime = System.currentTimeMillis();

    outputString.append("Time");
    outputString.append('\t');
    outputString.append("OutputTupleCount");
    outputString.append('\t');
    outputString.append("InterQueueSize");
    outputString.append('\t');
    outputString.append("StateSize\n");
    //outputString.append("\n\n\n');

    while (true) {
        /** collect stats */
        outputTupleCount = collectOutputTupleCounts();
        interQueueSize = collectInterQueueSize();
        stateSize = collectStateSize();

        currentTime = System.currentTimeMillis() - startTime;
        if (debug)
            System.out.println(
                "Stats thread running ---- Time: "
                    + currentTime
                    + " Output Tuple Count: "
                    + outputTupleCount
                    + "\n");

        outputString.append(currentTime);
        outputString.append('\t');
        outputString.append(outputTupleCount);
        outputString.append('\t');
        outputString.append(interQueueSize);
        outputString.append('\t');
        outputString.append(stateSize);
        outputString.append('\n');

        try {
            Thread.sleep(updateInterval);
        } catch (InterruptedException e) {
            /** thread is interrupted by execution controller thread and should be stopped. */
            System.out.println("** Stats Collector is ended ");

            /* collect last stat before finish the thread */
            outputTupleCount = collectOutputTupleCounts();
            interQueueSize = collectInterQueueSize();
            stateSize = collectStateSize();

            currentTime = System.currentTimeMillis() - startTime;
            outputString.append(currentTime);
            outputString.append('\t');
            outputString.append(outputTupleCount);
            outputString.append('\t');
            outputString.append(interQueueSize);
            outputString.append('\t');
            outputString.append(stateSize);
            outputString.append('\n');

            printStats();
            return;
        }
    }
}
/**
 * Insert the method's description here.
 * Creation date: (5/20/2003 4:00:52 AM)
 */
public void setDefaultProperties() {
    addProperty("DEBUG", "false");
    addProperty("INTERVAL", "2000");
}
/**
 * Insert the method's description here.
 * Creation date: (6/2/2003 3:00:13 AM)
 * @param metadata java.lang.StringBuffer
 */
public void setMetaData(StringBuffer metadata) {
    //outputString.append(metadata);
    this.metaData = metadata;
}
/**
 * Insert the method's description here.
 * Creation date: (4/21/2003 11:40:55 PM)
 * @param tree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public void setTree(XATDag dag) {

    this.dag = dag;

    /** get root operators to collect output tuple counts */
    /*Iterator rootIt = dag.getRoots();
    operatorsToCollect.clear();
    while (rootIt.hasNext()) {
        operatorsToCollect.add(
            (XATStreamOperator) ((XATNode) (rootIt.next())).getOperator());
    }*/
    //operatorsToCollect = (XATStreamOperator[]) rootOperators.toArray(new XATStreamOperator[] {});

    /** get output queue */
    outputQueuesToCollect.clear();
    Iterator rootIt = dag.getRoots();
    outputQueuesToCollect.clear();
    while (rootIt.hasNext())
        outputQueuesToCollect.add(
            ((XATStreamOperator) ((XATNode) (rootIt.next())).getOperator()).getOutputQueue());

    /** get all intermediate queues (excluding input and output queues), and get all states */
//    DagNode[] allnodes = XATBox.convertDagToArray(dag);
	XATNode[] allnodes = dag.convertToArray();
    queuesToCollect.clear();
    statesToCollect.clear();
    XATStreamOperator currentOp;
    XATQueue currentQ;
    WOperatorState currentState;

    int count = allnodes.length;
    for (int i = 0; i < count; i++) {
        currentOp = (XATStreamOperator) allnodes[i].getOperator();

        //get states
        if (currentOp instanceof WBinaryJoinOperator) {
            currentState = (WOperatorState)((WBinaryJoinOperator) currentOp).getLeftState();
            if (!statesToCollect.contains(currentState))
                statesToCollect.add(currentState);
            currentState = (WOperatorState)((WBinaryJoinOperator) currentOp).getRightState();
            if (!statesToCollect.contains(currentState))
                statesToCollect.add(currentState);
        }

        //get intermediate queues
        Iterator it = currentOp.getInputQueues();
        while (it.hasNext()) {
            currentQ = (XATQueue) it.next();
            if (!(currentQ.getQueueOwner() instanceof Stream)) {
                if (!queuesToCollect.contains(currentQ))
                    queuesToCollect.add(currentQ);
            }
        }
    }
}
}
