/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.migrationcontroller;

//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.WBinaryJoinOperator;

import java.util.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
//import queryplanelements.*;
/**
 * Insert the type's description here.
 * Creation date: (3/21/2003 5:51:58 PM)
 * @author: Yali Zhu
 */
public class XATBox extends XATQueryObjectImp {

	/** 
	 * A box is identified by both its root node and an array of all nodes inside the box.
	 * Each box is like a XATTree, with just a SINGLE box root node and output queue.
	 * And each box can have multiple input queues and multiple leaf nodes.
	 */
    private XATQueue[] inputQs;
    private XATQueue outputQ;
    private XATQueue[] interQs;

    //private DagNode rootNode;
    //private DagNode[] leafNodes;
    private DagNode[] allNodes = null;
    private XATStreamOperator[] allOperators;
    private XATStreamOperator[] leafOperators;
    private XATStreamOperator rootOperator;
    private OperatorState[] allStates;
    
    private int operatorCounter = 0;  // the largest operator ID, all IDs in range [0, operatorCounter]. 
    private int stateCounter = 0;     // the largets state ID, all IDs in range [0, stateCounter].

    //private edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree tree;
    //private XATDag dag;
    //private XATOperator[] rootOperators;
/**
 * Insert the method's description here.
 * Creation date: (4/3/2003 11:35:14 PM)
 */
public XATBox(DagNode root, DagNode[] allnodes) {
    super();
    //rootNode = root;
    rootOperator = (XATStreamOperator) root.getOperator();
    allNodes = allnodes;
    //ArrayList operators = new ArrayList();
    allOperators = new XATStreamOperator[allNodes.length];

    for (int i = 0; i < allNodes.length; i++) {
    	allOperators[i] = allNodes[i].getStreamOperator();
    }
//        if (!(allNodes[i] instanceof Source)) {
//            if (allNodes[i].getOperator() instanceof XATStreamOperator)
//                operators.add((XATStreamOperator) allNodes[i].getOperator());
//        }
    //allOperators = (XATStreamOperator[])operators.toArray(new XATStreamOperator[] {});
    setup();
}

public XATBox(DagNode root, XATDag dag){
	this(root, convertDagToArray(dag));	
}

public int getOperatorCounter() {
	return operatorCounter;
}
public void setOperatorCounter(int c) {
	operatorCounter = c;
}

public void setRootOperator(XATStreamOperator newroot) {
	rootOperator = newroot;
}

public int getStateCounter() {
	return stateCounter;
}
public void setStateCounter(int c) {
	stateCounter = c;
}

public static DagNode[] convertDagToArray(XATDag dag) {
	Vector nodeList = new Vector();
	Iterator i = dag.getRoots();

	while (i.hasNext()) 
		appendNodes(nodeList, (DagNode)i.next());
		//nodeList.add((XATNode)i.next());

	return (DagNode[]) nodeList.toArray(new DagNode[] {});
}

protected static void appendNodes(Vector v, DagNode node) {
	XATNode[] children = node.getOpChildren();
	/** only add if it doesnt already contain it */
	if (!v.contains(node)) {
		v.add(node);
	}
	for (int i = 0; i < children.length; i++) {
		appendNodes(v, (DagNode)children[i]);
	}
	return;
}

/**
 * Insert the method's description here.
 * Creation date: (4/6/2003 6:57:53 AM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode[]
 */
public DagNode[] getAllNodes() {
	return allNodes;
}
/**
 * Insert the method's description here.
 * Creation date: (4/16/2003 10:56:26 AM)
 */
public XATStreamOperator[] getAllOperators() {

    return allOperators;
}
/**
 * Insert the method's description here.
 * Creation date: (4/3/2003 10:52:42 PM)
 */
public OperatorState[] getAllStates() {
    return allStates;
}
/**
 * Insert the method's description here.
 * Creation date: (4/10/2003 10:49:33 AM)
 * @return queryplanelements.WBinaryJoinOperator
 */
public XATStreamOperator[] getBoxLeafOperators() {

	return leafOperators;
}
/**
 * Insert the method's description here.
 * Creation date: (4/11/2003 2:45:00 AM)
 * @return queryplanelements.WBinaryJoinOperator
 */
public XATStreamOperator getBoxRootOperator() {

    return rootOperator;

}
/**
 * Insert the method's description here.
 * Creation date: (3/21/2003 5:57:13 PM)
 */
public XATQueue[] getInputQs() {

    return inputQs;
}
/**
 * Insert the method's description here.
 * Creation date: (5/20/2003 10:11:42 AM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue[]
 */
public XATQueue[] getInterQs() {
	return interQs;
}
/**
 * Insert the method's description here.
 * Creation date: (3/21/2003 5:57:48 PM)
 */
public XATQueue getOutputQ() {

    return outputQ;
}
/**
 * Insert the method's description here.
 *
 * This method is to check if a queue inside the current box is a box input Q.
 * The definition of boxInputQs is the Q that goes into box (custormer inside box),
 * but originally from outside the box (owner outside the box).
 *
 * Creation date: (4/16/2003 10:51:37 AM)
 * @return boolean
 */
public boolean isBoxInputQ(XATQueue queue) {
	int size = allOperators.length;

	for(int i=0; i<size; i++)
	{
		/* if this queue is the output queue of any operator in the box, return false */
		if(queue.isOwner(allOperators[i]))
			return false;
	}
	
	return true;
}
/**
 * Insert the method's description here.
 * Creation date: (5/21/2003 11:12:47 AM)
 * @return boolean
 * @param operator edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator
 */
private boolean isBoxOperator(XATStreamOperator operator) {
    int size = allOperators.length;

    /** check if this operator is one of the box operators */
    for (int i = 0; i < size; i++) {
        if (allOperators[i].equals(operator))
            return true;
    }

    return false;
}
/**
 * Insert the method's description here.
 * Creation date: (4/16/2003 1:20:46 PM)
 */
public void setup() {

    /** compute outputQs */
    outputQ = rootOperator.getOutputQueue();

    /** compute inputQs, interQs and all states*/
    int size = allOperators.length;
    XATQueue queue;
    ArrayList inputQList = new ArrayList();
    ArrayList interQList = new ArrayList();
    ArrayList stateList = new ArrayList();
    ArrayList leafList = new ArrayList();
    boolean boxLeafOp = false;
    for (int index = 0; index < size; index++) {

//        if (allOperators[index] instanceof WBinaryJoinOperator) {
//            stateList.add(((WBinaryJoinOperator) allOperators[index]).getLeftState());
//            stateList.add(((WBinaryJoinOperator) allOperators[index]).getRightState());
//        }
		if(allOperators[index].hasState()) {
			OperatorState[] states = allOperators[index].getState();
			
			for(int i=0; i<states.length; i++ ) {
				stateList.add(states[i]);
			}
		}

        Iterator i = allOperators[index].getInputQueues();

        while (i.hasNext()) {
            inputQList.add(i.next());
        }
    }
    
    for (int index=0; index<size; index++) {
    	
    	Iterator i = allOperators[index].getOutputQueues();
    	while(i.hasNext()) {
    		queue = (XATQueue)i.next();
    		if(inputQList.contains(queue)) {
    			interQList.add(queue);
    			inputQList.remove(queue);
    		}
    	}
    }
    
    for(int i=0; i<inputQList.size(); i++) {
    	List customers = ((XATQueue)inputQList.get(i)).getQueueCustomer();
    	for(int j=0; j<customers.size(); j++) {
    		XATStreamOperator customer = (XATStreamOperator)customers.get(j);
    		for(int k=0; k<allOperators.length; k++) {
    			if(customer.equals(allOperators[k]) && !leafList.contains(customer)) {
    				leafList.add(customer);
    				break;
    			}
    		}
    	}
    }

    inputQs = (XATQueue[]) inputQList.toArray(new XATQueue[] {});
    interQs = (XATQueue[]) interQList.toArray(new XATQueue[] {});
    allStates = (OperatorState[]) stateList.toArray(new OperatorState[] {});
    leafOperators = (XATStreamOperator []) leafList.toArray(new XATStreamOperator[] {});
}
/* (non-Javadoc)
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject#getLightWeightQueryObject()
 */
public LightWeightQueryObject getLightWeightQueryObject() {
	return new LightWeightQueryObject("XATBox",this.getClass().getName(),true);
}

/**
 * make a new copy of current box. Method invoked by run time plan optimizer.
 * @return
 */
public XATBox makeNewCopy(){
	int size = allNodes.length;
	DagNode[] newnodes = new DagNode[size];
	for(int i=0; i<size; i++) {
		newnodes[i] = (DagNode)allNodes[i].makeNewCopy();
//		double reductionfactor = Double.parseDouble(allNodes[i].getStreamOperator().getProperty("REDUCTION_FACTOR"));
//		try {
//			newnodes[i].getStreamOperator().setProperty("REDUCTION_FACTOR", ""+reductionfactor);
//		} catch (Exception e) {
//			e.printStackTrace();
//		}
	}
	
	// root node is assumed to be the first node in nodes array
	connectNewNodes(newnodes);	
	
	int[] l = new int[newnodes.length];
	for(int i=0; i<l.length; i++) {
		l[i] = 0;
	}
	boolean done = false;
	boolean isready = false;
	
	// propogate input queue schema
	while(!done) {
		done = true;
		for(int i=newnodes.length-1; i>=0; i--) {
			
			if(l[i] == 0) {
				isready = true;
				Iterator inputs = newnodes[i].getStreamOperator().getInputQueues();
				while(inputs.hasNext()) {
					if(((XATQueue)inputs.next()).getSchema() == null) {
						done = false;
						isready = false;
						break;
					}
				}
			
				if(isready) {
					newnodes[i].getStreamOperator().init();
					l[i] = 1;
				}
			}
		}
	}
	
	XATBox newBox = new XATBox(newnodes[0], newnodes);
	newBox.operatorCounter = this.operatorCounter;
	newBox.stateCounter = this.stateCounter;
	
	return newBox;
}


/**
 * connect the new nodes and creates new queues along the way
 * invoked by makeNewCopy() method.
 * @param newrootnode
 */
private void connectNewNodes(DagNode[] newnodes) {
	int position;
	XATQueue newQ;
	
	// first set box input Q IDs
	XATQueue[] oldInputQs = this.inputQs;
	for(int i=0; i<oldInputQs.length; i++) {
			oldInputQs[i].setBoxInputQID(i+1);
	}
	
	//first set the output queue for the new root operator
	XATStreamOperator rootop = newnodes[0].getStreamOperator();
	if (System.getProperty("DATA_MODEL").endsWith("XATMemoryQueueImp")) {
		newQ = new XATMemoryQueueImp(rootop);
		rootop.setOutputQueue(newQ);
	} else if (
		System.getProperty("DATA_MODEL").endsWith("XATPersistentQueueImp")) {
		newQ = new XATPersistentQueueImp(rootop);
		rootop.setOutputQueue(newQ);
	} else {
		System.out.println("!!! Error: Unknown type of queue DATA_MODEL !!!");
		System.exit(-1);
	}
	
	// then set the parents and children relations among new nodes
	// according to the same relation among old nodes, set interQs along the way.
	for (int index = 0; index < newnodes.length; index++) {
		
		XATNode[] oldnodes = allNodes[index].getOpChildren();

		for (int i = 0; i < oldnodes.length; i++) {
			XATStreamOperator operator = (XATStreamOperator) oldnodes[i].getOperator();
			
			if(operator.isBoxOperator()) {
				
				newnodes[index].addOpChild(newnodes[operator.getOperatorID()]);
				newnodes[operator.getOperatorID()].addParent(newnodes[index]);
			
				XATStreamOperator childOp = (XATStreamOperator) newnodes[operator.getOperatorID()].getOperator();
				XATStreamOperator parentOp = (XATStreamOperator) newnodes[index].getOperator();
			
				if (System.getProperty("DATA_MODEL").endsWith("XATMemoryQueueImp")) {
					newQ = new XATMemoryQueueImp(childOp);
					childOp.setOutputQueue(newQ);
					parentOp.addInputQueue(newQ);
				} else if (
					System.getProperty("DATA_MODEL").endsWith("XATPersistentQueueImp")) {
					newQ = new XATPersistentQueueImp(childOp);
					childOp.setOutputQueue(newQ);
					parentOp.addInputQueue(newQ);
				} else {
					System.out.println("!!! Error: Unknow type of queue DATA_MODEL !!!");
					System.exit(-1);
				}
			}
		}
	}
	
	// now generate all input queues of the box
	for (int k=0; k< this.inputQs.length; k++) {
		List customers = inputQs[k].getQueueCustomer();
		for(int i=0; i<customers.size(); i++) {
			XATStreamOperator tempOp = (XATStreamOperator)customers.get(i);
			
			if(tempOp.isBoxOperator()) {
				
				if (System.getProperty("DATA_MODEL").endsWith("XATMemoryQueueImp")) {
					newQ = new XATMemoryQueueImp();				
					position = tempOp.getInputQueuePosition(inputQs[k]);
					newnodes[tempOp.getOperatorID()].getStreamOperator().addInputQueue(newQ, position);
					newQ.setBoxInputQID(inputQs[k].getBoxInputQID());
					newQ.setSchema(inputQs[k].getSchema());
					
				} else if ( System.getProperty("DATA_MODEL").endsWith("XATPersistentQueueImp")) {
					newQ = new XATPersistentQueueImp();
					position = tempOp.getInputQueuePosition(inputQs[k]);
					newnodes[tempOp.getOperatorID()].getStreamOperator().addInputQueue(newQ, position);
					newQ.setBoxInputQID(inputQs[k].getBoxInputQID());
					newQ.setSchema(inputQs[k].getSchema());
					
				} else {
					System.out.println("!!! Error: Unknow type of queue DATA_MODEL !!!");
					System.exit(-1);
				}
			}
		}
	}
}


///*
// * return the index of node in nodearray.
// * if not exist, return -1.
// */
//private int getNodeIndex(XATNode[] nodearray, XATNode node) {
//	for(int i=0; i<nodearray.length; i++) {
//		if(nodearray[i].equals(node))
//			return i;
//	}
//	
//	return -1;
//}


	/**
 	* This method assigns a new ID to each operator and state in the tree.
 	* The state IDs are used for matching states during MS migration.
 	* The operator IDs are used when applying rewriting rules so the right stats can be retrieved from the stats table.
 	*/
	public void assignNewIDs() {
		//XATStreamOperator[] operators = oldbox.getAllOperators();
		int total = allOperators.length;
		operatorCounter = 0;
		stateCounter = 0;
		
		for(int i=0; i<total; i++) {
			//allNodes[i].setNodeID(operatorcounter);
			allOperators[i].setOperatorID(operatorCounter);
			operatorCounter++;
			if(allOperators[i].hasState()) {
				OperatorState[] states = allOperators[i].getState();
				for(int j=0; j<states.length; j++) {
					states[j].setStateID(stateCounter);
					stateCounter++;
				}
			}
		}
		
		return;
	}

	public void deleteNode(XATNode node) {
		int newsize = allNodes.length - 1;
		int i, j;
		DagNode[] newnodes = new DagNode[newsize];
		XATStreamOperator[] newop = new XATStreamOperator[newsize];
		
		for(i=0, j=0; i<newsize+1; ) {
			if(allNodes[j].equals(node)) {
				j++;
				continue;
			}
			newnodes[i] = allNodes[j];
			newop[i] = allOperators[j];
			i++;
			j++;
		}
	}
}
