/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.heuristics;
/*
 * Created on Apr 19, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */

import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp;
import java.util.ArrayList;
//import java.io.OutputStream;
//import java.io.FileOutputStream;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.migrationcontroller.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.WBinaryJoinOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;

/**
 * @author Yali Zhu
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class Optimizer {

	private XATTree tree;
	private XATBox oldbox;
	private XATBox newbox;
	private ArrayList HeuristicsArray;
	private MigrationController migrationcontrol = new MigrationController();
	
	//private static String outputString = new String();
	//private static XATTree tree = new XATTreeImp();
	
	public Optimizer() {
		super();
		HeuristicsArray = new ArrayList();
		//HeuristicsArray.add(new PushdownSelect());
		HeuristicsArray.add(new SwitchJoinByStateSizeHeuristics());
	}
	
	/**
	 * 
	 */
//	public Optimizer(XATTree currentTree) {
//		super();
//		//tree = currentTree;
//		HeuristicsArray = new ArrayList();
//		//HeuristicsArray.add(new PushdownSelect());
//		HeuristicsArray.add(new SwitchJoin());
//	}

	/**
	 * This method figures out the sub-plan that needs to be optimized. The sub-plan can be as small as an operator,
	 * or as large as the whole query plan passed in as the parameter. 
	 * This is the first step of plan optimization.
	 * 
	 * As the first version, we can assume that the sub-plan is always equal to the whole plan, so this method just take
	 * a query plan tree and generate a box object from it. 
	 * TODO Later we need to add the functionality to figure out the sub-plan which may be part of the complete query plan.
	 * This is an important step for localized plan optimization and migration, and the size of the box can be determined
	 * by several system parameters, such as available system resources, migration overhead, etc.
	 * 
	 * @param oldtree
	 * @return
	 */
	private boolean constructOldBox(XATTree oldtree) {
		
		// first check if the old plan needs to be re-optimized
		// if not, returns null indicating so

//		if(checkHuristics(oldtree) == false) {
//			return null;
//		}  

		this.tree = oldtree;
		
		oldbox = new XATBox((DagNode)oldtree.getRoot(), (XATDag)oldtree);
		oldbox.assignNewIDs();
		
		XATStreamOperator[] allboxoperators = oldbox.getAllOperators();
		for(int i=0; i<allboxoperators.length; i++) {
			allboxoperators[i].setBoxOperator(true);
		}

		if(checkHeuristics(oldbox) == false) {
			return false;
		} 
		return true;
	}
	
	/**
	 * delete a node from the new box
	 * likely being called when applying rewriting rules, such as merging two operators.
	 */
	public void deleteNewNode(XATNode newnode) {
		XATNode[] nodes = newbox.getAllNodes();
		
	}
	/**
	 * TODO
	 * This method construct the new box based on the old box.
	 * This is the second step of plan optimization.
	 * 
	 * It first clones the old box to get the initial new box.
	 * It then apply heuristics and rewriting rules to this initial new box and get the final new box.
	 * 
	 * @param oldtree
	 * @return
	 */
	private void constructNewBox(XATBox oldbox) {
		newbox = oldbox.makeNewCopy();
		
		XATStreamOperator[] allboxoperators = newbox.getAllOperators();
		for(int i=0; i<allboxoperators.length; i++) {
			allboxoperators[i].setBoxOperator(true);
		}
		
		applyHeuristics(newbox);
		
		/* need to re-setup the new box after plan rewriting. */
		newbox.setup();
		//return newbox;
	}

	public XATBox getNewBox() {
		return newbox;
	}
	
	public XATBox getOldBox() {
		return oldbox;
	}
	
    /**
     * TODO
     * The following method calls the run-time plan migrator to migrate from the old box to the new box, using the 
     * specified migration strategy.
     * 
     * @param oldbox
     * @param newbox
     * @param migrationstrategy
     */
	private void migrate(String migrationstrategy) throws Exception{
	
		// when this method is called, the new and old boxes should have already been set up correctly.
		if(oldbox == null || newbox == null)
			throw new Exception("Error -- either old or new box in optimizer is not initialized!");
		else {
			// do migration
			migrationcontrol.setTree(this.tree);
			migrationcontrol.setOldBox(oldbox);
			migrationcontrol.setNewBox(newbox);
			migrationcontrol.runMSMigration();
		}
		   
		
		// after migration is done	
		XATStreamOperator[] allboxoperators = newbox.getAllOperators();	
		for(int i=0; i<allboxoperators.length; i++) {
			allboxoperators[i].setBoxOperator(false);
		}
		
		allboxoperators = oldbox.getAllOperators();	
		for(int i=0; i<allboxoperators.length; i++) {
			allboxoperators[i].setBoxOperator(false);
		}
	}
	
	/**
	 * 
	 * @param newbox
	 * @return
	 */
	private boolean applyHeuristics(XATBox newbox) {
		try {

			for (int i = 0; i < this.HeuristicsArray.size(); i++) {
				((Heuristic) HeuristicsArray.get(i)).ApplyHeuristic(newbox, tree);
			}

		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return true;
	}
	
	/**
	 * 
	 * @param oldbox
	 * @return
	 */
	private boolean checkHeuristics(XATBox oldbox) {
		
		for (int i = 0; i < this.HeuristicsArray.size(); i++) {
			if(((Heuristic) HeuristicsArray.get(i)).CheckHeuristic(oldbox, tree))
				return true;
		}

		return false;
	}
	
	/**
	 * 
	 * @param oldtree
	 * @return
	 */
	public boolean optimizeQueryTree(XATTree oldtree) {
		this.tree = oldtree;
		
		try{
			tree.appendMigrationActivity("*** Start Optimizing plan *** \n");
			tree.appendMigrationActivity("Construct old migration box.\n");
			if(constructOldBox(oldtree)) {
				tree.appendMigrationActivity("Construct new migration box.\n");
				constructNewBox(this.oldbox);
				//applyHeuristics(this.newbox);
				tree.appendMigrationActivity("Migrate from old to new boxes.\n");
				migrate("MS");
				tree.appendMigrationActivity("Migration done. \n");
				tree.appendMigrationActivity("*** Plan is changed. *** \n\n");	
				return true;
			}
			tree.appendMigrationActivity("*** Plan is NOT changed. *** \n\n");
			return false;
			} catch (Exception e) {
				e.printStackTrace();
				return false;
			}
	}
	
	
	
//	public static void addToOutputString(String st){
//		outputString += st;
//	}
//	
//	public static XATTree getTree() {
//		return tree;
//	}
//	
//	public static void setTreeRoot(XATNode newRoot) {
//		tree.setRoot(newRoot);
//	}
	
	/*
	 * Author: Yali Zhu   11/8/2005
	 * This is different version of the local plan optimizer.
	 * It is more efficient during the migration process.
	 * Moving state migration strategy is assumed.
	 * The migrator does not need to make a new copy of the current query plan.
	 * Instead the missing states are recomputed when two operators are being switched. 
	 * (see the changes in switchoperator rewriting rule).
	 */
	public boolean efficientOptimization(XATTree tree) {
		boolean changed = false;
		
		for (int i = 0; i < this.HeuristicsArray.size(); i++) {
			changed = ((Heuristic) HeuristicsArray.get(i)).ApplyHeuristic(null, tree);
		}
		
		return changed;
	}
	
	public boolean mjoinOptimization(XATTree tree) {
		
		double threshold = 0.8;
		
		//figure out memory usage percentage
		Runtime rt = Runtime.getRuntime();
		double freeMem = ((double) rt.freeMemory()) / 1048576.0;
		double totalMem = ((double) rt.totalMemory()) / 1048576.0;
		double memPct = 1.0 - freeMem/totalMem;
		
		//System.out.println("Calculate pct memory usgage: freemem = " + freeMem + "\n" + "totalmem = " + totalMem + "\n" + "mempct = " + memPct);
		
		//optimize if memory used is close to threshold
		if(memPct > threshold) {
			
		}
		
		return true;
	}
}
