/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
/**
 * Author			:	Sicilia Wijaya	
 * Project			:	Project 3 - Query Plan Optimization in a Streaming Query Engine
 * Course			:	CS4432 D04
 * Date				: 	05/01/04
 * How to run		:	Run the batch file Project3 Startup.bat and Project3 Run.bat
 * Assumption		: 	- Select has 1 child and Binary join has 2 children
 * 						- For join operator, the left predicate comes from the left input
 * 						  queue and vice versa
 * 						- PushSelect is always called before SwitchJoin, therefore,
 * 						  all the Select operators are always at the bottom of the tree 
 * Known bugs		:	-
 * Design Rationale	:	see README
 */

/*
 * Created on May 7, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.heuristics;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.rewritingrules.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.migrationcontroller.XATBox;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;

/**
 * @author Yali Zhu
 * @author Sicilia Wijaya
 *
 * This program pushes all the select operator down the query tree
 */
public class PushdownSelectHeuristics extends Heuristic{

	public PushdownSelectHeuristics() {
		super();
	}
	/**
	 * Go through each node of the tree and apply the push down select heuristics
	 * 
	 * Creation date: 05/01/04
	 * @author Sicilia Wijaya
	 * @param tree to be traversed
	 * @return boolean whether heuristics was applied
	 */
	public boolean ApplyHeuristic(XATBox newbox, XATTree oldtree) {

		boolean done = false;		
		this.box = newbox;
		this.tree = oldtree;

		if(newbox.getAllOperators().length == 0)
			return false;
			
		//if any pushing of select operator happens, go up the tree
		//and see if there is any conflict with the grandparents or parents
		while(!done) {
			done = pushSelect(newbox.getAllOperators());
		}
		
		return true;
	}
	
	
	/**
	 * This method checks a node and see if any rewriting should be done
	 * 
	 * @author Sicilia Wijaya
	 * @param node parent
	 * @return whether any pushing down of select is done
	 */
	public boolean pushSelect (XATStreamOperator[] alloperators)
	{
		//boolean changed = false;
		
		try{
			for(int i=0; i<alloperators.length; i++) {
			
				if(alloperators[i] instanceof StreamSelectOperator) {
					System.out.println ("Found a select operator");
					XATNode inode = alloperators[i].getNode();
				
					// if already at leaf level, continue to the next operator in the box
					if(inode.getOpChildrenSize() == 0) {
						continue;
					}
				
					if(inode.getOpChildrenSize() != 1) { 
						throw new Exception("Select can only have one child!");
					}
				
					XATStreamOperator coperator = inode.getOpChild(0).getStreamOperator();
				
					//if there is a select operator below current select operator,
					// and if the lower select is already at the box leaf level, merge these two selects.
					if(coperator.isBoxOperator()) {
						if(coperator instanceof StreamSelectOperator) {
				
							if(coperator.getNode().getOpChildrenSize() != 1) { 
								throw new Exception("Select can only have one child!");
							}

							if(!coperator.getNode().getOpChild(0).getStreamOperator().isBoxOperator()) {
								MergeOperatorRules.Merge((StreamSelectOperator)alloperators[i], (StreamSelectOperator)coperator, this.box);
								return true;
							}
							else {
/** CHUAN LEI - Couldn't find DistributedExecutionController, so pass 'null' as the last parameter into the following method calls*/
								SwitchOperatorRules.SwitchOperators((StreamSelectOperator)alloperators[i], (StreamSelectOperator)coperator, this.box, null);
								return true;
							}
						}
						// otherwise switch current select operator with whatever box operator below it.
						else {
/** CHUAN LEI - Couldn't find DistributedExecutionController, so pass 'null' as the last parameter into the following method calls*/
							SwitchOperatorRules.Switch(inode.getStreamOperator(), coperator, this.box.getStateCounter(), this.box, null);
							return true;
						}			
					}
				}
			}
		} catch(Exception e) {
			e.printStackTrace();	
		}
		
		return false;
	}
	
	/**
	 * check if the given box have any select operator that can be pushed down.
	 * 
	 * the method returns TRUE right away if it finds the first chance of pushing down select.
	 * it returns FALSE is none of such chance is found.
	 */
	public boolean CheckHeuristic(XATBox box, XATTree tree) {
		
		XATStreamOperator[] operators = box.getAllOperators();
		
		for(int i=0; i<operators.length; i++) {
			if(operators[i] instanceof StreamSelectOperator) {
				
				XATNode[] childNodes = operators[i].getNode().getOpChildren();
				
				for(int j=0; j<childNodes.length; j++) {
					if(childNodes[j].getStreamOperator().isBoxOperator())
						return true;
				}
				
			}
		}
		
		return false;
	}
	
}
