/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.heuristics;
/**
 * Author			:	Sicilia Wijaya	& Yali Zhu
 * Project			:	Project 3 - Query Plan Optimization in a Streaming Query Engine
 * Course			:	CS4432 D04
 * Date				: 	05/01/04
 * How to run		:	Run the batch file Project3 Startup.bat and Project3 Run.bat
 * Assumption		: 	- Select has 1 child and Binary join has 2 children
 * 						- For join operator, the left predicate comes from the left input
 * 						  queue and vice versa
 * 						- PushSelect is always called before SwitchJoin, therefore,
 * 						  all the Select operators are always at the bottom of the tree 
 * Known bugs		:	-
 * Design Rationale	:	see README
 */

/*
 * Created on May 7, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp;
//import edu.wpi.cs.cs4432.project3.rewritingrules.SwitchOperatorRules;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.migrationcontroller.XATBox;
import edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.rewritingrules.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;


/**
 * @author Yali Zhu
 * @author Sicilia Wijaya
 * 
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class SwitchJoinBySelectivityHeuristics extends Heuristic{
	
	public SwitchJoinBySelectivityHeuristics() {
		super();
	}
	/**
	 * Go through each node of the tree and apply the switch join heuristics
	 * 
	 * Creation date: 05/01/04
	 * @author Sicilia Wijaya
	 * @param tree to be traversed
	 * @return boolean whether heuristics was applied
	 */
	public boolean ApplyHeuristic(XATBox newbox, XATTree oldtree) {
		
		this.box = newbox;
		this.tree = oldtree;

		XATStreamOperator[] operators;
		
		//If efficient migration is used
		if(box == null) {
			XATNode[] nodes = oldtree.convertToArray();
			if(nodes.length == 0)
				return false;
			
			operators = new XATStreamOperator[nodes.length];
			for(int i=0; i<nodes.length; i++) {
				operators[i] = nodes[i].getStreamOperator();
			}	
		} 
		//If old migration strategy is used (need to make a copy of current plan), get the operator from the box
		else {
			operators = box.getAllOperators();
		}
		
		// check if there's anything in the box		
		if(operators.length == 0)
			return false;
		
		boolean changed = false;
		boolean done = false;
		while(!done) {
			if(switchJoins(operators) == true)
				changed = true;
			else
				done = true;
		}
		return changed;
	}
	
	/**
	 * This method compare a node and if it is a join operator that has join operators
	 * as its children, switch that node with the child that has the highest selectivity
	 * 
	 * @author Sicilia Wijaya
	 * @param node parent
	 * @return whether any switching of joins is done
	 */
	public boolean switchJoins(XATStreamOperator[] operators)
	{
		boolean changed = false;
		
		try{
			for(int i=0; i<operators.length; i++) {
			
				if(operators[i] instanceof XATMultiQueueStreamOperator) {
//					System.out.println ("Found a join operator");
					XATNode inode = operators[i].getNode();
				
					// if already at leaf level, continue to the next operator in the box
					if(inode.getOpChildrenSize() == 0) {
						continue;
					}
								
					XATNode[] cnodes = inode.getOpChildren();
					double currentfactor = Double.parseDouble(inode.getStreamOperator().getProperty("SELECTIVITY"));
					double maxfactor = currentfactor;
					int maxposition = 0;
					
					System.out.println("upper SELECTIVITY = " + currentfactor);
					for(int j=0; j<cnodes.length; j++) {
						XATStreamOperator tempop = cnodes[j].getStreamOperator();

						if(tempop.getClass() == inode.getStreamOperator().getClass()) {
							//System.out.println(tempop.getClass());
							//System.out.println(inode.getStreamOperator().getClass());
							double factor = Double.parseDouble(cnodes[j].getStreamOperator().getProperty("SELECTIVITY"));
							System.out.println("lower SELECTIVITY = " + factor);
							
							if(factor - maxfactor > 0) {
								maxfactor = factor;
								maxposition = j;
								
							}
						}
					}
					if (maxfactor != currentfactor) {
						if(inode.equals(tree.getRoot())) {
							tree.setRoot(cnodes[maxposition]);
						}
/** CHUAN LEI - Couldn't find DistributedExecutionController, so pass 'null' as the last parameter into the following method calls*/						
						if(this.box == null) {
							SwitchOperatorRules.Switch(
									inode.getStreamOperator(), cnodes[maxposition].getStreamOperator(), 0, null, null);
						} else {
							SwitchOperatorRules.Switch(
									inode.getStreamOperator(), cnodes[maxposition].getStreamOperator(), this.box.getStateCounter(), this.box, null);
						}
						inode.getStreamOperator().setChanged(true);
						cnodes[maxposition].getStreamOperator().setChanged(true);
//						this.tree.appendMigrationActivity(
//							"Switch node " + inode.getStreamOperator().getObjectID()
//						  + " and node " + cnodes[minposition].getStreamOperator().getObjectID());
						System.out.println("switch " + inode + " " + cnodes[maxposition]);
						return true;
						//return false;
					}
				}
			}
		} catch(Exception e) {
			e.printStackTrace();	
		}
		
		return changed;
	}
	
	/**
	 * check if the given box have any two consecutive join operators that can be switched.
	 * 
	 * the method returns TRUE right away if it finds the first chance of switch two join operators.
	 * it returns FALSE is none of such chance can be found.
	 */
	public boolean CheckHeuristic(XATBox box, XATTree tree) {
		
		XATStreamOperator[] operators = box.getAllOperators();
		
		for(int i=0; i<operators.length; i++) {
			if((operators[i] instanceof XATMultiQueueStreamOperator) && (operators[i].isActive())) {
				
				XATNode[] childNodes = operators[i].getNode().getOpChildren();
				double upperSelectivity = Double.parseDouble(operators[i].getProperty("SELECTIVITY"));
				System.out.println("upper SELECTIVITY = " + upperSelectivity);
				
				for (int j = 0; j < childNodes.length; j++) {
					System.out.println("lower SELECTIVITY = "
								+ Double.parseDouble(childNodes[j].getStreamOperator().getProperty(
										"SELECTIVITY")));
					if ((childNodes[j].getStreamOperator().getClass() == operators[i].getClass())
						&& (childNodes[j].getStreamOperator().isActive())
						&& (childNodes[j].getStreamOperator().isBoxOperator())
//						&& (Double.parseDouble(childNodes[j].getStreamOperator().getProperty("SELECTIVITY")) > upperSel ectivity))
						&& (Double.parseDouble(childNodes[j].getStreamOperator().getProperty("SELECTIVITY"))) - upperSelectivity > 0)

						return true;
				}				
			}
		}
		
		return false;
	}
}
