/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.mjoinoptimizer;

import java.util.ArrayList;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDagImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;

/**
 * @author Yali Zhu
 * 
 * Initial optimization: apply exhaustive-search optimization to the multiple joins.
 * The search space contains all mjoin tree space, including left-deep, right-deep and bushy trees.
 */
public class ExhaustiveSearchOptimizer extends MultiJoinOptimizer {

	//hashtable of DPEntries, key is pstring, value is the dpentry with the pstring.
	Hashtable DPEntries = new Hashtable();
	int duplicates = 0;
	int entrycount = 0;
	double qualifiedmemory = 0;
	double qualifiedcpu = 0;
	int hit = 0;
	
	//hashtable of inputs, key is the input name (assume single letter), value is the index of the input starting from 0;
	//Hashtable indextable;
	char[] inputs;
	
	public ExhaustiveSearchOptimizer(XATDag dag) {
		super(dag);
	}
	
	/**
	 * find qualified query plan based on current join graph
	 */
	public XATDag optimize(XATDag olddag) {
		String pstring = exhaustiveSearch();
		System.out.println("done!");
		if(pstring != null) {
			System.out.println("find qualified query plan " + pstring);
			return this.convertPStringToTree(pstring);
		} else{
			return olddag;
		}
	}
	
	/**
	 * Use dynamic programming to exhaustively find a qualified query plan.
	 * @return the first qualified plan (represented by pstring) found by the algorithm.
	 */
	public String exhaustiveSearch() {
		
		this.DPEntries.clear();
		duplicates = 0;
		entrycount = 0;
		qualifiedmemory = 0;
		qualifiedcpu = 0;
		hit = 0;
		
		//init table inputs, keyarray and entry hashtable
		inputs = new char[this.joinGraph.numNodes];
		ArrayList keyarray = new ArrayList();
		
		char initchar = 'A';
		for(int i=0; i<inputs.length; i++) {
			inputs[i] = initchar;
			keyarray.add(""+initchar);
			
			DPJoinEntry entry = new DPJoinEntry(1, ""+initchar);
			entry.setCPUCost(0);
			entry.setMemoryCost(0);
			entry.setOutputRate(this.joinGraph.arrivalRates[i]);
			DPEntries.put(""+initchar, entry);
			entrycount++;
			initchar++;
		}
		
		int keyposition = 0; //index of key in keyarray
		boolean done = false;
		//int duplicates = 0;  //for debug purposes
		//char lastchar = 'A' + this.joinGraph.numNodes -1;
		
		//Dynamic Programming
		while(keyposition < keyarray.size()) {
			String currentkey = (String)keyarray.get(keyposition);
			char lastinput = this.getLargestLetter(currentkey);
			int startingIndex = lastinput - 'A' + 1; 
			int lastIndex = this.joinGraph.numNodes-1;
			int streamsize = ((DPJoinEntry)DPEntries.get(currentkey)).streamsize+1;
			
			//each time add one more letter	
			for(int i=startingIndex; i<=lastIndex; i++) {
				char nextinputletter = inputs[i];
				for(int k = 0; k < currentkey.length(); k++) {
					String newpstring = null;
					
					//create a new join
					if(currentkey.charAt(k) == '(') {	
						newpstring = this.createPStringWithNewJoin(currentkey, nextinputletter, k);
					} 
					//create a new join with a input stream
					else if(Character.isLetter(currentkey.charAt(k))) {
						newpstring = this.createPStringWithNewLeafJoin(currentkey, nextinputletter, k);
					}
					//merge the input with an existing join to form a large mjoin
					else if(currentkey.charAt(k) == ')') {
						newpstring = this.createPStringWithOneLargerJoin(currentkey, nextinputletter, k);
					}
					
					if(newpstring != null) {
						//System.out.println("calculate " +newpstring);
						DPJoinEntry newentry = new DPJoinEntry(streamsize, newpstring);
						ArrayList inputpstrings = this.getRootInputPString(newpstring);
						
						int inputsize = inputpstrings.size();
						double[] inputrates = new double[inputsize];					
						int[] inputstreamsizes = new int[inputsize];
						
						DPJoinEntry[] inputentries = new DPJoinEntry[inputsize];
						double treememorycost = 0;
						double treecpucost = 0;
						
						
						try{
						for(int m=0; m<inputsize; m++) {
							inputentries[m] = (DPJoinEntry)DPEntries.get((String)inputpstrings.get(m));
							inputrates[m] = inputentries[m].getOutputRate();
							inputstreamsizes[m] = inputentries[m].getStreamSize();
							treememorycost += inputentries[m].treememorycost;
							treecpucost += inputentries[m].treecpucost;
						}}catch(NullPointerException e) {
							//System.out.println(e);
							//System.out.println("invalid entry");
							continue;
						}
						
						newentry.setInputSize(inputsize);
						newentry.setInputRates(inputrates);
						newentry.setInputStreamSizes(inputstreamsizes);
						newentry.timeWindow = this.joinGraph.timeWindow;
						
						//calculate selectivities
						double[][] selectivities = new double[inputsize][inputsize];
						
						//init selectivities
						for(int x=0; x<inputsize; x++) {
							for(int y=0; y<inputsize; y++) {
								selectivities[x][y] = 1;
							}
						}
						//compute selectivities based on selectivity matrix stored in joinGraph.
						for(int x=0; x<inputsize; x++) {
							for(int y=x+1; y<inputsize; y++) {
								String xpstring = (String)inputpstrings.get(x);
								String ypstring = (String)inputpstrings.get(y);
								double selectxy = 1;
								for(int m=0; m<xpstring.length(); m++) {
									char mchar = xpstring.charAt(m);
									if(!Character.isLetter(mchar)) {
										continue;
									}
									for(int n=0; n<ypstring.length(); n++) {
										char nchar = ypstring.charAt(n);
										if(!Character.isLetter(nchar)) {
											continue;
										}
										int mint = mchar - 'A';
										int nint = nchar - 'A';
										selectxy *= joinGraph.selectivities[mint][nint];
									}
								}
								selectivities[x][y] = selectxy;
								selectivities[y][x] = selectxy;
							}
						}
						
						//calculate memory and cpu of the join orders represented by the new string
						newentry.selectivities = selectivities;
						newentry.calculateProbeOrders();
						newentry.calculateOutputRate();
						newentry.calculateMemoryCost();
						newentry.calculateCPUCost();
						newentry.treememorycost = treememorycost + newentry.memoryCost;
						newentry.treecpucost = treecpucost + newentry.CPUCost;
						
						//if not both costs are over system threshold, keep the partial result in hashtable.
						//if(!((newentry.memoryCost > this.memorythreshold) && (newentry.CPUCost > this.CPUthreshold))) {					
							//for debug purpose, the newpstring should have have existed in the hashtable.
							if(DPEntries.containsKey(newpstring)) {
								duplicates++;
							} else{
							DPEntries.put(newpstring, newentry);
							keyarray.add(newpstring);
							entrycount++;
							}
						//} 
					    //else {
							//System.out.println("\ninvalid entry!!!");
						//}
						
						//if this is a qualified plan, return right away
						if(newentry.streamsize == this.joinGraph.numNodes) {
							//System.out.println("dp done.");
							if((newentry.treememorycost < this.memorythreshold) && (newentry.treecpucost < this.CPUthreshold)) {
								System.out.println("find qualified plan: " + newentry.pstring);
								System.out.println("memorycost = " + newentry.treememorycost + " cpucost = " + newentry.treecpucost);
								qualifiedmemory = newentry.treememorycost;
								qualifiedcpu = newentry.treecpucost;
								hit = 1;
								return newpstring;
							}
						}
					}
				}
			}
			//go to the next key in keyarray.
			keyposition++;
		}
		
		return null;
	}
	
	/**
	 * @param s The orginal string
	 * @param c The char represent the stream name to be inserted in the string
	 * @param pos The position of the corresponding open parenthesis '('. pos starts from 0.
	 * @return A new string with c combined with its previous input to create a new string.
	 *         all chars before pos are the same as in the original string except one more '(' is inserted, 
	 *         all chars from and after pos are shifted to the right.
	 * Example: input (AB)  C  0
	 *          output(C(AB))
	 */
	public String createPStringWithNewJoin(String s, char c, int pos) {
		
		StringBuffer newstring = new StringBuffer();
		
		for(int i=0; i<pos; i++) {
			newstring.append(s.charAt(i));
		}
		char[] newchars = new char[]{'(', c};
		newstring.append(newchars);
		
		int level = 0;
		boolean inserted = false;
		for(int i=pos; i<s.length(); i++) {
			newstring.append(s.charAt(i));
			
			if(!inserted) {
				if(s.charAt(i) == '(') {
					level++;
					continue;
				}
			
				if(s.charAt(i) == ')') {
					level--;
					if(level == 0) {
						newstring.append(')');
						inserted = true;
					}
				}
			}
		}
		
		return newstring.toString();
	}
	
	/**
	 * @param s The orginal string
	 * @param c The char represent the stream name to be inserted in the string
	 * @param pos The position of the stream that needs to be merged with c to form a new leaf join operator.
	 * @return
	 * Example: input (AB)  C  1
	 *          output((CA)B)
	 */
	public String createPStringWithNewLeafJoin(String s, char c, int pos) {
		StringBuffer newstring = new StringBuffer();
		
		//if current position is not a letter
		if(!Character.isLetter(s.charAt(pos))) {
			return null;
		}
		
		for(int i=0; i<pos; i++) {
			newstring.append(s.charAt(i));
		}
		char[] newchars = new char[]{'(', c, s.charAt(pos), ')'};
		newstring.append(newchars);
		
		for(int i=pos+1; i<s.length(); i++) {
			newstring.append(s.charAt(i));
		}		
		
		return newstring.toString();
	}
	
	/**
	 * @param s The orginal string
	 * @param c The char represent the stream name to be inserted in the string
	 * @param pos The position of the corresponding close parenthesis ')'. pos starts from 0.
	 * @return
	 * Example: input (AB)  C  3
	 *          output((ABC)
	 */
	public String createPStringWithOneLargerJoin(String s, char c, int pos) {
		
//		if(Character.isLetter(s.charAt(pos-1))) {
//			return null;
//		}
		
		StringBuffer newstring = new StringBuffer();
		
		for(int i=0; i<pos; i++) {
			newstring.append(s.charAt(i));
		}
		newstring.append(c);
		
		for(int i=pos; i<s.length(); i++) {
			newstring.append(s.charAt(i));
		}
		
		return newstring.toString();
	}
	
	/**
	 * @param s
	 * @return The memory (first) and cpu (second) costs of the query plan represented by the input string s.
	 */
	public double[] calculateCosts(String s) {
		double[] costs = new double[2];
		
		ArrayList rootPStrings = this.getRootInputPString(s);
		ArrayList rootstrings = new ArrayList();
		
		//get the input string with only stream names
		for(int i=0; i<rootPStrings.size(); i++) {
			String pstring = (String)rootPStrings.get(i);
			StringBuffer sbuffer = new StringBuffer();
			for(int k=0; k<pstring.length(); k++) {
				if(Character.isLetter(pstring.charAt(k))) {
					sbuffer.append(pstring.charAt(k));
				}
			}
			rootstrings.add(sbuffer.toString());
		}
		
		double[][] selectivities = joinGraph.selectivities;
		double[] inputrates = joinGraph.arrivalRates;
		
		//calculate input rates
		
		//calculate selectivities
	
		//calculate join orderings
		
		return costs;
	}
	
	/**
	 * @param s The query plan is represented by input pstring s.
	 * @return  Get the input pstring to the root operator.
	 * Example: input (A(BC)D)
	 *          the output arraylist contains 3 pstring: {A, (BC), D}
	 */
	public ArrayList getRootInputPString(String s) {
		//int leftPLevel = 2; // p is parenthesis.
		ArrayList pstringlist = new ArrayList();
		
		for(int i=1; i<s.length(); i++) {
			
			if(Character.isLetter(s.charAt(i))) {
				pstringlist.add("" + s.charAt(i));
			}
			
			if(s.charAt(i) == '(') {
				StringBuffer sbuffer = new StringBuffer();
				int k;
				int level = 0;
				for(k=i; k<s.length(); k++) {
					if(s.charAt(k) == '(') {
						level++;
					}
					if(s.charAt(k) == ')') {
						level--;						
						if(level == 0) {
							sbuffer.append(')');
							pstringlist.add(sbuffer.toString());
							break;
						}
					}				
					sbuffer.append(s.charAt(k));
				}
				i = k;
			}
		}
		return pstringlist;
	}
	
	/**
	 * @param s Input string.
	 * @return  The last letter in the input string.
	 * Example: input=(AB(CD)E), output=E
	 * return 0 if the string does not contain a letter.
	 */
	public char getLargestLetter(String s) {
		int l = s.length();
		char largestchar = 'A';
		
		for(int i=0; i<s.length(); i++) {
			char c = s.charAt(i);
			if(Character.isLetter(c) && (c > largestchar)) {
				largestchar = c;
			}
		}
		return largestchar;
	}
	
	/**
	 * @param pstring
	 * @return The number of stream names contained in the input string.
	 * Example: input ((AB)C)
	 *          output 3
	 */
	public int getStreamSize(String pstring) {
		int size = 0;
		
		for(int i=0; i<pstring.length(); i++) {
			if(Character.isLetter(pstring.charAt(i))) {
				size++;
			}
		}		
		return size;
	}
	
	//TODO
	private XATDag convertPStringToTree(String pstring) {
		XATDag dag = new XATDagImp();
		
		return dag;
	}
}
