/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.mjoinoptimizer;

import java.util.ArrayList;
import java.util.Hashtable;
//import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDagImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.ComplexOperator.WMJoinOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNodeImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTreeImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.ComplexOperator.WMJoinOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;


//import java.util.ArrayList;
//import java.lang.reflect.Constructor;
//import java.util.List;

/**
 * @author Yali Zhu
 * 
 * Insert comments here.
 */
public class FromBTreeOptimizer extends MultiJoinOptimizer {
	
	//create hashtable from entries for easy access
	//key = entryID, value is the entry itself
	//0.75 is the default load factor of a hashtable in java.
	//int initialcapacity = entries.size() * 4/3; 
	Hashtable entryHashTable = new Hashtable();
	double qualifiedmemory = 0;
	double qualifiedcpu = 0;
	int hit = 0;
	
	public FromBTreeOptimizer(XATDag dag) {
		super(dag);
	}
	
	/**
	 * find qualified query plan based on current join graph
	 */
	public XATDag optimize(XATDag olddag) {

		qualifiedmemory = 0;
		qualifiedcpu = 0;
		hit = 0;
		
		this.treeIsChanged = false;
		
		// if the current tree is qualified, return it right away
		// get memory and cpu cost of current query plan
		// return tree;
		

		//generate minstateBTree and check if it is a qualified plan
		ArrayList joinentries = this.generateMinStateBTreeEntries();
		double btree_memorycost = this.computeMemoryCost(joinentries);
		double btree_cpucost = this.computeCPUCost(joinentries);
		this.treeIsChanged = true;
		System.out.println("btree: memory cost = " + btree_memorycost + " cpu cost = " + btree_cpucost);

		if(btree_memorycost <= this.memorythreshold && btree_cpucost <= this.CPUthreshold) {
			//XATDag newtree = this.convertEntryListToRealQueryTree(joinentries);
			XATDag newtree = new XATDagImp();  // for static experiment only.
			System.out.println("find qualified btree plan:"); 
			//System.out.println("entries: " + joinentries);
			System.out.println("memorycost = " + btree_memorycost + " cpucost = " + btree_cpucost);
			System.out.println("done!");
			qualifiedmemory = btree_memorycost;
			qualifiedcpu = btree_cpucost;
			hit = 1;
			return newtree;
		}
		
//		//generate mjoin and check if it is a qualified plan
//		int[][] joinOrders = this.joinGraph.generateJoinOrders();
//		double[][] joinSelectivities = this.joinGraph.computeProbeSelectivityMatrix(joinOrders);
//		double mjoin_memorycost = this.computeMJoinMemoryCost();
//		double mjoin_cpucost = this.computeMJoinCPUCost(joinOrders, joinSelectivities);
//		System.out.println("mjoin: memory cost = " + mjoin_memorycost + " cpu cost = " + mjoin_cpucost);
//		
//		if(mjoin_memorycost <= this.memorythreshold && mjoin_cpucost <= this.CPUthreshold) {
//			//XATDag newtree = this.convertJoinOrderToQueryTree(joinOrders, joinSelectivities);
//			XATDag newtree = new XATDagImp();  // for static experiment only.
//			System.out.println("find qualified plan:"); 
//			System.out.println("memorycost = " + mjoin_memorycost + " cpucost = " + mjoin_cpucost);
//			System.out.println("done!");
//			return newtree;
//		}
		
		//if neither of the two is qualified plan, generate mjointree
//		if(mjoin_memorycost < this.memorythreshold) {
			boolean foundqualifiedplan = generateMTreeEntriesFromBTree(joinentries, btree_memorycost, btree_cpucost);
			if(foundqualifiedplan) {
				//XATDag newtree = this.convertEntryListToRealQueryTree(joinentries);
				XATDag newtree = new XATDagImp();  // for static experiment only.
				//System.out.println("find qualified mjointree plan:"); 
				System.out.println("done!");
				return newtree;
			} else{
				 //TODO otherwise return mjoin.
				 //XATDag newtree = this.convertJoinOrderToQueryTree(joinOrders, joinSelectivities);
				 //return newtree;
				
				//for static experiments
				System.out.println("done!");
				return null; 
			}
//		} 
		
		//return olddag;
		
		//for static experiments
		//System.out.println("done!");
		//return null; 
	}
		
//		for (int i=0; i<nodes.length; i++) {
//			if(nodes[i].getOperator() instanceof WMJoinOperator) {
//				WMJoinOperator currentOperator = (WMJoinOperator)nodes[i].getOperator();
//				String runningMode = currentOperator.getRunningMode();
//				this.timewindow = currentOperator.getTimeWindow(); // in milliseconds
//				
//				XATNode btreeroot = generateMinStateBTree(currentOperator, nodes[i]);
//				((WMJoinOperator)btreeroot.getOperator()).setAsRootOperator();
//				((WMJoinOperator)btreeroot.getOperator()).setRunningMode("btree");
//				System.out.println("The generated min-state btree is:\n" + this.getBTreeString(btreeroot, " "));
//					
//				if(currentOperator == queryPlan.getRoot().getOperator()) {
//					queryPlan.setRoot(btreeroot);
//				} else {
//					XATNode parentNode = currentOperator.getNode().getParent();
//					btreeroot.setParent(parentNode);
//					int cindex = parentNode.getIndexOfChild(currentOperator.getNode());
//					try {
//						parentNode.replaceChild(cindex, btreeroot);
//					}catch(Exception e){
//						e.printStackTrace();
//						System.exit(-1);
//					}
//				}
//			}	
//				else if (runningMode.equalsIgnoreCase("mjoin")) {
//					if(currentOperator.getNode().getParent() == null) {
//						currentOperator.setAsRootOperator();
//					}
//		}
//	}
	
	/**
	 * use the top-down algorithm to generate the mjoin tree
	 * require the computeProbeSequence() to be called first.
	 * return true if a qualified mjoin tree is found, otherwise return false.
	 * 
	 * ASSUMPTION: entries are ordered by entryID, and the smallest entry id in the list is this.JoinGraph.numNodes+1.
	 */
	private boolean generateMTreeEntriesFromBTree(ArrayList entries, double memoryCost, double cpuCost) {
		boolean done = false;
		int largestEntryID = ((MJoinEntry)entries.get(entries.size()-1)).entryID;
		
		double oldMemoryCost = memoryCost;
		double oldCPUCost = cpuCost;
		double newmemorysaved = 0;
		double newcpusaved = 0;
		

		entryHashTable.clear();
		for(int i=0; i<entries.size(); i++) {
			MJoinEntry entry = (MJoinEntry)entries.get(i);
			entryHashTable.put(""+entry.entryID, entry);
		}
		
		while(!done) {
			//if only one entry left, it is an mjoin operator, stop the optimization and return false.
			if(entries.size() == 1)
				return false;
			
			// the join entry within which the intermediate result is to be removed by merging operators.
			MJoinEntry stateOldEntry = null;
			MJoinEntry mergeOldEntry = null;
			//int InputIndex = -1;
			//double largestState = 0;
			MJoinEntry newInsertedEntry = null;
			//double memorySaved, CPUSaved;
			double largestCPUToMemRatio = -Double.MAX_VALUE;
			//int removedcounter = 0;
			
			//go through the entry list and find intermediate state to merge
			for(int i=0; i<entries.size(); i++) {
				
				MJoinEntry currententry = (MJoinEntry)entries.get(i);
//				if(currententry.removed) {
//					//removedcounter++;
//					continue;
//				}
				
				//go through current entry's inputs
				for(int k=0; k<currententry.childrenIDs.length; k++) {
					
					//if the input is another join node (corresponding state is intermediate state)
					if(currententry.childrenIDs[k] >= this.joinGraph.numNodes) {
						int centryid = currententry.childrenIDs[k];
						
						//find the entry with the entryid
						MJoinEntry mergeEntry;
						//mergeEntry = (MJoinEntry)entries.get(centryid - this.joinGraph.numNodes);
						mergeEntry = (MJoinEntry)entryHashTable.get(""+centryid);
						
						largestEntryID++;
						MJoinEntry newentry = createNewEntryByMerging(currententry, mergeEntry, k, largestEntryID);
						if(newentry == null) {
							continue;
						}
						
						double memorySaved = currententry.getMemoryCost() + mergeEntry.getMemoryCost() - newentry.getMemoryCost();
						double cpuSaved = currententry.getCPUCost() + mergeEntry.getCPUCost() - newentry.getCPUCost();
						
						//if the new plan is qualified, return right away.
						if((oldMemoryCost - memorySaved < this.memorythreshold) && (oldCPUCost - cpuSaved < this.CPUthreshold)) {
							mergeEntries(entries, currententry, mergeEntry, newentry);
							System.out.println("Find a qualified mjointree plan.");
							System.out.println("memory saved = " + memorySaved + " cpu increased = " + (-cpuSaved));
							qualifiedmemory = oldMemoryCost - memorySaved;
							qualifiedcpu = oldCPUCost - cpuSaved;
							hit = 1;
							return true;
						}
						
						double ratio = cpuSaved/memorySaved;
						
						//if it is the best move so far, record it.
						if(ratio > largestCPUToMemRatio) {
							largestCPUToMemRatio = ratio;
							stateOldEntry = currententry;
							mergeOldEntry = mergeEntry;
							newInsertedEntry = newentry;
							newmemorysaved = memorySaved;
							newcpusaved = cpuSaved;
							
							//currententry = null;
							mergeEntry = null;
							newentry = null;
						}				
					}
				}
				currententry = null;
			}

			if(newInsertedEntry != null) {
				//System.out.println("merge two entries: " + stateOldEntry +" " + mergeOldEntry);
				//System.out.println("memory saved = " + newmemorysaved + " cpu increased = " + (-newcpusaved));
				mergeEntries(entries, stateOldEntry, mergeOldEntry, newInsertedEntry);
			} else {
				done = true;
			}
		}
		return false;
	}
	
	/**
	 * newentry is the merged entry of stateentry and mergeentry.
	 * After this method, the entry list is updated to reflect the merge.
	 * @param entries
	 * @param stateentry
	 * @param mergeentry
	 * @param newentry: The merged entry of stateentry and mergeentry.
	 */
	private void mergeEntries(ArrayList entries, MJoinEntry stateOldEntry, MJoinEntry mergeOldEntry, MJoinEntry newEntry) {
		entries.remove(stateOldEntry);
		entries.remove(mergeOldEntry);
		//stateOldEntry.removed = true;
		//mergeOldEntry.removed = true;
		entries.add(newEntry);
		entryHashTable.put(""+newEntry.entryID, newEntry);
		
		//update parentid of children
		for(int n=0; n<newEntry.childrenIDs.length; n++) {
			int temp = newEntry.childrenIDs[n];
			if(temp > this.joinGraph.numNodes) {
				//MJoinEntry centry = (MJoinEntry)entries.get(temp - joinGraph.numNodes);
				MJoinEntry centry = (MJoinEntry)entryHashTable.get(""+temp);
				centry.parentID = newEntry.entryID;
			}
		}
		//update childid of parent
		int temp = newEntry.parentID; 
		if(temp >=0) {
			//MJoinEntry pentry = (MJoinEntry)entries.get(temp - joinGraph.numNodes);
			MJoinEntry pentry = (MJoinEntry)entryHashTable.get(""+temp);
			for(int n=0; n<pentry.childrenIDs.length; n++) {
				if(pentry.childrenIDs[n] == stateOldEntry.entryID) {
					pentry.childrenIDs[n] = newEntry.entryID;
				}
			}
		}
	}
	
	/**
	 * @param stateentry: the entry that has the intermediate state to be removed
	 * @param mergeEntry: the child entry of stateentry that contributes to the to-be-removed intermediate state.
	 * @param stateInputIndex: the index of the intermediate state in stateentry.
	 * @param newEntryID: the entry id of the new entry.
	 * @return  the new entry as a result of merging the input two entries.
	 */
	private MJoinEntry createNewEntryByMerging(MJoinEntry stateentry, MJoinEntry mergeEntry, int stateInputIndex, int newEntryID) {
		//create new entry by merging current entry and the child entry
		//the intermediate state is removed.
		int newsize = stateentry.inputSize + mergeEntry.inputSize - 1;
		int[] inputindex = new int[newsize];
		int[] inputstreamsize = new int[newsize];
		double[] inputrates = new double[newsize];
		double[][] selectivities = new double[newsize][newsize];
		
		for(int x=0; x < stateInputIndex; x++) {
			inputindex[x] = stateentry.childrenIDs[x];
			inputstreamsize[x] = stateentry.childrenstreamsizes[x];
			inputrates[x] = stateentry.inputRates[x];
		}
					
		for(int x=0; x<mergeEntry.inputSize; x++) {
			inputindex[x + stateInputIndex] = mergeEntry.childrenIDs[x];
			inputstreamsize[x + stateInputIndex] = mergeEntry.childrenstreamsizes[x];
			inputrates[x + stateInputIndex] = mergeEntry.inputRates[x];
		}
					
		for(int x=stateInputIndex+1; x<stateentry.inputSize; x++) {
			inputindex[x+mergeEntry.inputSize-1] = stateentry.childrenIDs[x];
			inputstreamsize[x+mergeEntry.inputSize-1] = stateentry.childrenstreamsizes[x];
			inputrates[x+mergeEntry.inputSize-1] = stateentry.inputRates[x];
		}
		
		for(int i=0; i<inputindex.length; i++) {
			selectivities[i][i] = 1;
			for(int j=i+1; j<inputindex.length; j++) {
				Object temp;
				if((temp = this.allselectivities.get(""+inputindex[i] +" " +inputindex[j])) != null) {
					selectivities[i][j] = Double.parseDouble((String)temp);
					selectivities[j][i] = Double.parseDouble((String)temp);
				} else if ((temp = this.allselectivities.get(""+inputindex[j] +" " +inputindex[i])) != null) {
					selectivities[i][j] = Double.parseDouble((String)temp);
					selectivities[j][i] = Double.parseDouble((String)temp);
				} else {
					//System.out.println("\nThe selectivity is not found in the hashtable!\n\n");
					return null;
				}
			}
		}
		
		MJoinEntry newentry = new MJoinEntry(newEntryID, inputindex, inputstreamsize, this.joinGraph.timeWindow, null);
		
		//recompute cpu and memory costs of the new entry list
		newentry.setInputRates(inputrates);
		newentry.setSelectivities(selectivities);
		newentry.calculateProbeOrders();
		newentry.calculateOutputRate();
		newentry.calculateCPUCost();
		newentry.calculateMemoryCost();
		newentry.parentID = stateentry.parentID;
		
		return newentry;
	}

}
