/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.mjoinoptimizer;

import java.util.ArrayList;
//import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDagImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.ComplexOperator.WMJoinOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;

/**
 * @author Yali Zhu
 * 
 * Insert comments here.
 */
public class FromMJoinOptimizer extends MultiJoinOptimizer {

	double qualifiedmemory = 0;
	double qualifiedcpu = 0;
	int hit = 0;
	
	public FromMJoinOptimizer(XATDag dag) {
		super(dag);
	}
	
	// use top-down algorithm (from mjoin to mtree) to generate the mjoin tree
	// this method does not need a pre-computed probe sequence.
	private boolean generateMTreeEntriesFromMJoin(ArrayList entries, double memoryCost, double cpuCost) {

		//int rootEntryIndex = 0;
		double totalMemoryCost = memoryCost;
		double totalCPUCost = cpuCost;
		
		// first generate edge list from selectivity matrix
		ArrayList edgelist = new ArrayList();
		double[][] selectivities = this.joinGraph.getSelectivities();
		double[] rates = this.joinGraph.getArrivalRates();
		long timewindow = this.joinGraph.getTimeWindow();
		
		
		// if use cartesian product
		if (this.joinGraph.getUseCartesianProductStatus()) {
			for(int i=0; i<selectivities.length; i++) {
				for (int j=i+1; j<selectivities.length; j++) {
						Edge edge = new Edge(i, j, rates[i], rates[j], selectivities[i][j]);
						edgelist.add(edge);
				}
			}
		} else {
			for (int i = 0; i < selectivities.length; i++) {
				for (int j=i+1; j<selectivities.length; j++) {
					if (selectivities[i][j] < 1) {
						Edge edge = new Edge(i, j, rates[i], rates[j], selectivities[i][j]);
						edgelist.add(edge);
					}
				}
			}
		}

		// New join entry ID starts from next number larger than the largest input id.
		int minJoinID = rates.length;
		int joinID = minJoinID;
		boolean done = false;
		
		while ((edgelist.size() > 1) && !done) {
			
			double memoryIncreased, cpuSaved;
			Edge pickedEdge = null;
			MJoinEntry newCEntry = null;
			MJoinEntry newPEntry = null;
			double newmemoryincreased = 0;
			double newcpusaved = 0;
			double largestCPUToMemoryRatio = - Double.MAX_VALUE;

			//root join entry is always the first entry of entries list.
			MJoinEntry rootJoinEntry = (MJoinEntry)entries.get(0);
			double[] rootInputRates = rootJoinEntry.inputRates;
			double[][] rootSelectivities = rootJoinEntry.selectivities;
			int rootInputSize = rootJoinEntry.inputSize;
			int[] rootInputIDs = rootJoinEntry.childrenIDs;
			int[] rootstreamsizes = rootJoinEntry.childrenstreamsizes;
			
			for(int i=0; i<edgelist.size(); i++) {
				
				//pick one edge and create two separate join entries representing the operator breaking.
				Edge currentedge = (Edge)edgelist.get(i);
			
				//create new child join entry
				double[][] selectivitymatrix = new double[2][2];
				double[] arrivalrates = new double[2];
				selectivitymatrix[0][1] = currentedge.selectivity;
				selectivitymatrix[1][0] = currentedge.selectivity;
				selectivitymatrix[0][0] = 1;
				selectivitymatrix[1][1] = 1;
				arrivalrates[0] = currentedge.rate1;
				arrivalrates[1] = currentedge.rate2;
			
				int[] inputIDs = new int[2];
				int[] inputstreamsizes = new int[2];
				inputIDs[0] = currentedge.v1;
				inputIDs[1] = currentedge.v2;
				inputstreamsizes[0] = currentedge.streamsize1;
				inputstreamsizes[1] = currentedge.streamsize2;
				MJoinEntry newChildJoinEntry = new MJoinEntry(joinID, inputIDs, inputstreamsizes, timewindow, null);
				joinID++;
				newChildJoinEntry.setInputRates(arrivalrates);
				newChildJoinEntry.setSelectivities(selectivitymatrix);
				newChildJoinEntry.setTimeWindow(this.joinGraph.timeWindow);
				newChildJoinEntry.calculateProbeOrders();
				newChildJoinEntry.calculateCPUCost();
				newChildJoinEntry.calculateMemoryCost();
				newChildJoinEntry.calculateOutputRate();
				newChildJoinEntry.parentID = 0; //parent is the new parent entry with entryID=0
				
				//create new parent join entry
				int inputsize = rootJoinEntry.inputSize - 1;
				selectivitymatrix = new double[inputsize][inputsize];
				arrivalrates = new double[inputsize];
				inputIDs = new int[inputsize];
				inputstreamsizes = new int[inputsize];
				selectivitymatrix = new double[inputsize][inputsize];
				
				int counter = 0;
				int rowcounter = 0;
				int columncounter = 0;
				for(int k=0; k<rootInputSize; k++) {
					if((rootInputIDs[k] != currentedge.v1) && (rootInputIDs[k] != currentedge.v2)) {
						arrivalrates[counter] = rootInputRates[k];
						inputIDs[counter] = rootInputIDs[k];
						inputstreamsizes[counter] = rootstreamsizes[k];
						counter++;
						double newselectivity = 1;
						columncounter = 0;
						for(int c=0; c<rootInputSize; c++) {
							if((rootInputIDs[c] != currentedge.v1) && (rootInputIDs[c] != currentedge.v2)) {
								selectivitymatrix[rowcounter][columncounter] = rootSelectivities[k][c];
								columncounter++;
							} else{
								newselectivity *= rootSelectivities[k][c];
							}						
						}
						selectivitymatrix[rowcounter][columncounter] = newselectivity;
						selectivitymatrix[columncounter][rowcounter] = newselectivity;
						selectivitymatrix[rowcounter][rowcounter] = 1;
						rowcounter++;
					}
				}
				arrivalrates[inputsize-1] = newChildJoinEntry.getOutputRate();
				inputIDs[inputsize-1] = newChildJoinEntry.entryID;
				inputstreamsizes[inputsize-1] = newChildJoinEntry.numOfStreams;
				for(int c=0; c<inputsize-1; c++) {
					selectivitymatrix[inputsize-1][c] = selectivitymatrix[c][inputsize-1]; 
				}
				selectivitymatrix[inputsize-1][inputsize-1] = 1;
				
				//new parent entryID = 0
				MJoinEntry newParentJoinEntry = new MJoinEntry(0, inputIDs, inputstreamsizes, timewindow, null);
				newParentJoinEntry.setInputRates(arrivalrates);
				newParentJoinEntry.setSelectivities(selectivitymatrix);
				newParentJoinEntry.setTimeWindow(this.joinGraph.timeWindow);
				newParentJoinEntry.calculateProbeOrders();
				newParentJoinEntry.calculateCPUCost();
				newParentJoinEntry.calculateMemoryCost();
				newParentJoinEntry.calculateOutputRate();

				memoryIncreased = newChildJoinEntry.getMemoryCost() + newParentJoinEntry.getMemoryCost() - rootJoinEntry.getMemoryCost();
				cpuSaved = rootJoinEntry.getCPUCost() - newChildJoinEntry.getCPUCost() - newParentJoinEntry.getCPUCost();
				double ratio = cpuSaved/memoryIncreased;
				
				//if the improve is the largest so far, record it
				if(ratio > largestCPUToMemoryRatio) {
					largestCPUToMemoryRatio = ratio;
					pickedEdge = currentedge;
					newCEntry = newChildJoinEntry;
					newPEntry = newParentJoinEntry;
					newmemoryincreased = memoryIncreased;
					newcpusaved = cpuSaved;
				}
				
				//if the new plan is qualified, return right away.
				if((totalMemoryCost + memoryIncreased < this.memorythreshold) && (totalCPUCost - cpuSaved < this.CPUthreshold)) {
					pickedEdge = currentedge;
					newCEntry = newChildJoinEntry;
					newPEntry = newParentJoinEntry;
					//totalMemoryCost = totalMemoryCost + memoryIncreased;
					//totalCPUCost = totalCPUCost - cpuSaved;
					done = true;
					//System.out.println("\nFound a qualified mjointree plan");
					//System.out.println("memorycost = " + (totalMemoryCost + memoryIncreased));
					//System.out.println("cpucost = " + (totalCPUCost - cpuSaved));
 					break;
				}
			} //END of FOR
			
			//System.out.println("\n picked edge " + pickedEdge);
			//System.out.println("memory increased = " + newmemoryincreased);
			//System.out.println("cpu saved = " + newcpusaved);
			
			//now update the edgelist with the pickededge
			for (int x = 0; x < edgelist.size(); x++) {
				Edge nextedge = (Edge) edgelist.get(x);
				if(nextedge == pickedEdge) {
					edgelist.remove(nextedge);
					x--;
					continue;
				}
				if (nextedge.v1 == pickedEdge.v1 || nextedge.v1 == pickedEdge.v2) {
					nextedge.v1 = newCEntry.entryID;
					nextedge.streamsize1 = pickedEdge.streamsize1 + pickedEdge.streamsize2;
					nextedge.rate1 = pickedEdge.weight * timewindow / 1000;
					nextedge.updateWeight();
					continue;
				}
				if (nextedge.v2 == pickedEdge.v1 || nextedge.v2 == pickedEdge.v2) {
					nextedge.v2 = newCEntry.entryID;
					nextedge.streamsize2 = pickedEdge.streamsize1 + pickedEdge.streamsize2;
					nextedge.rate2 = pickedEdge.weight * timewindow / 1000;
					nextedge.updateWeight();
				}
			}

			//now combine edges with the same vertices
			for (int x = 0; x < edgelist.size(); x++) {
				Edge edge1 = (Edge) edgelist.get(x);
				for (int y = x + 1; y < edgelist.size(); y++) {
					Edge edge2 = (Edge) edgelist.get(y);

					if ((edge1.v1 == edge2.v1 && edge1.v2 == edge2.v2)
							|| (edge1.v1 == edge2.v2 && edge1.v2 == edge2.v1)) {
						edge1.selectivity = edge1.selectivity * edge2.selectivity;
						edge1.updateWeight();
						edgelist.remove(y);
						break;
					}
				}
			}
			
			//update the join entry list
			entries.remove(0);
			entries.add(0, newPEntry);
			entries.add(newCEntry);
			
			//update memory cost and cpu cost
			memoryIncreased = newCEntry.getMemoryCost() + newPEntry.getMemoryCost() - rootJoinEntry.getMemoryCost();
			cpuSaved = rootJoinEntry.getCPUCost() - newCEntry.getCPUCost() - newPEntry.getCPUCost();
			totalMemoryCost = totalMemoryCost + memoryIncreased;
			totalCPUCost = totalCPUCost - cpuSaved;
			
			// if the input to this new join is the intermediate output from
			// another join, mark the parent/child relationship in each input join entry
			if ((pickedEdge.v1 > minJoinID) || (pickedEdge.v2 > minJoinID)) {
				for (int x = 0; x < entries.size(); x++) {
					MJoinEntry entry = (MJoinEntry) entries.get(x);
					if (entry.entryID == pickedEdge.v1 || entry.entryID == pickedEdge.v2) {
						entry.parentID = newCEntry.entryID;
					}
				}
			}
		} //END OF WHILE
		
		//if the new plan is qualified, return true, otherwise return false.
		if((totalMemoryCost < this.memorythreshold) && (totalCPUCost < this.CPUthreshold)) {
			System.out.println("\nFound a qualified mjointree plan");
			System.out.println("memorycost = " + totalMemoryCost);
			System.out.println("cpucost = " + totalCPUCost);
			qualifiedmemory = totalMemoryCost;
			qualifiedcpu = totalCPUCost;
			hit = 1;
			return true;
		}
		else 
			return false;
	}

	/**
	 * find qualified query plan based on current join graph
	 */
	public XATDag optimize(XATDag olddag) {
		this.treeIsChanged = false;
		qualifiedmemory = 0;
		qualifiedcpu = 0;
		hit = 0;
		
		// if the current tree is qualified, return it right away
		// get memory and cpu cost of current query plan
		// return tree;
		
	
		//generate mjoin and check if it is a qualified plan
		int[][] joinOrders = this.joinGraph.generateJoinOrders();
		double[][] joinSelectivities = this.joinGraph.computeProbeSelectivityMatrix(joinOrders);
		double mjoin_memorycost = this.computeMJoinMemoryCost();
		double mjoin_cpucost = this.computeMJoinCPUCost(joinOrders, joinSelectivities);
		//this.printProbeOrders(joinOrders);
		//this.printProbeSelectivities(joinSelectivities);
		System.out.println("mjoin: memory cost = " + mjoin_memorycost + " cpu cost = " + mjoin_cpucost);
		
		if(mjoin_memorycost <= this.memorythreshold && mjoin_cpucost <= this.CPUthreshold) {
			//XATDag newtree = this.convertJoinOrderToQueryTree(joinOrders, joinSelectivities);
			XATDag newtree = new XATDagImp();  // for static experiment only.
			System.out.println("find a qualified mjoin plan:"); 
			//this.printProbeOrders(joinOrders);
			//this.printProbeSelectivities(joinSelectivities);
			System.out.println("memorycost = " + mjoin_memorycost + " cpucost = " + mjoin_cpucost);
			System.out.println("done!");
			qualifiedmemory = mjoin_memorycost;
			qualifiedcpu = mjoin_cpucost;
			hit = 1;
			return newtree;
		}
		
//		//generate minstateBTree and check if it is a qualified plan
//		ArrayList joinentries = this.generateMinStateBTreeEntries();
//		double btree_memorycost = this.computeMemoryCost(joinentries);
//		double btree_cpucost = this.computeCPUCost(joinentries);
//		this.treeIsChanged = true;
//		System.out.println("btree: memory cost = " + btree_memorycost + " cpu cost = " + btree_cpucost);
//
//		if(btree_memorycost <= this.memorythreshold && btree_cpucost <= this.CPUthreshold) {
//			//XATDag newtree = this.convertEntryListToRealQueryTree(joinentries);
//			XATDag newtree = new XATDagImp();  // for static experiment only.
//			System.out.println("find qualified plan:"); 
//			System.out.println("memorycost = " + btree_memorycost + " cpucost = " + btree_cpucost);
//			System.out.println("done!");
//			return newtree;
//		}
		
		//if neither of the two is qualified plan, generate mjointree
		ArrayList entries = new ArrayList();
		int[] inputindex = new int[this.joinGraph.numNodes];
		int[] inputstreamsize = new int[this.joinGraph.numNodes];
		for(int n=0; n<this.joinGraph.numNodes; n++) {
			inputindex[n] = n;
			inputstreamsize[n] = 1;
		}
		MJoinEntry mjoinentry = new MJoinEntry(this.joinGraph.numNodes, inputindex, inputstreamsize, this.joinGraph.timeWindow, null);
		mjoinentry.setSelectivities(this.joinGraph.selectivities);
		mjoinentry.setInputRates(this.joinGraph.arrivalRates);
		mjoinentry.calculateOutputRate();
		mjoinentry.setProbeOrders(joinOrders);
		mjoinentry.setProbeSelectivities(joinSelectivities);
		mjoinentry.setMemoryCost(mjoin_memorycost);
		mjoinentry.setCPUCost(mjoin_cpucost);
		entries.add(mjoinentry);
		
		if(mjoin_memorycost < this.memorythreshold) {
			boolean foundqualifiedplan = generateMTreeEntriesFromMJoin(entries, mjoin_memorycost, mjoin_cpucost);
			if(foundqualifiedplan) {
				//XATDag newtree = this.convertEntryListToRealQueryTree(joinentries);
				XATDag newtree = new XATDagImp();  // for static experiment only.
				//System.out.println("find qualified mjointree plan:"); 
				System.out.println("done!");
				return newtree;
			} else{
				//TODO otherwise return mjoin
				//XATDag newtree = this.convertJoinOrderToQueryTree(joinOrders, joinSelectivities);
				//return newtree;
				
				//for static experiments
				System.out.println("done!");
				return null; 
			}
		} 
		//return olddag;
		
		//for static experiments
		System.out.println("done!");
		return null; 
	}
}
