/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.mjoinoptimizer;

import java.io.Serializable;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.migration.probablisticjoinoperator.RootedTreeNode;

//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;

/**
 * @author Yali Zhu
 *
 */
public class JoinGraph implements Serializable{

	int numEdges = 0;
	int numNodes = 0;
	long timeWindow = 0;
	
	//boolean updateSelectivityStatus = false;
	boolean useCartesianProductStatus = false;
	
	double[][] selectivities;
	double[] arrivalRates;
	
	Vector edges = new Vector();
	//Vertex[] vertics = null;
	Vector minSpanningEdges = new Vector();
	
	// joinMatrix[i][j] = 1 means the vertics i and j is connected by an edge in the min-spanning tree
	// joinMatrix[i][j] = 0 means the vertics i and j is not connected in the min-spanning tree.
	int[][] minJoinMatrix;
	
	public JoinGraph(double[][] selectivities, double[] arrivalRates, boolean cartesian) {

		this.numNodes = arrivalRates.length;
		this.selectivities = selectivities;
		this.arrivalRates = arrivalRates;
		
		//Vector edges = new Vector();
		//this.vertics = new Vertex[numVertics];
		
		//this.updateSelectivityStatus = updateSelect;
		this.useCartesianProductStatus = cartesian;
		
		// if use cartesian product
		if (cartesian) {
			for(int i=0; i<numNodes; i++) {
				for (int j=0; j<i; j++) {
					if(i != j) {
						Edge edge = new Edge(i, j, arrivalRates[i], arrivalRates[j], selectivities[i][j]);
						edges.add(edge);
					}
				}
			}
		} else {
			for (int i = 0; i < numNodes; i++) {
				for (int j = 0; j < i; j++) {
					if (selectivities[i][j] <= 1) {
						Edge edge = new Edge(i, j, arrivalRates[i], arrivalRates[j], selectivities[i][j]);
						edges.add(edge);
					}
				}
			}

		}
		this.numEdges = edges.size();
		//origGraph = (Edge[]) edges.toArray(new Edge[] {});
		
		// generate min spanning tree of the original graph
		computeMinSpanningTree();
	}
	
	//public void setUpdateSelectivityStatus(boolean s) {
	//	this.updateSelectivityStatus = s;
	//}
	
	public void setUseCartesianProductStatus(boolean s) {
		this.useCartesianProductStatus = s;
	}
	
	public void setNumNodes(int num) {
		this.numNodes = num;
	}
	
	public void setSelectivities(double[][] selectivities) {
		if(this.selectivities == null)
			this.selectivities = new double[numNodes][numNodes];
		this.selectivities = selectivities;
	}
	
	public void setArrivalRates(double[] rates) {
		
		if(arrivalRates == null)
			arrivalRates = new double[rates.length];
		this.arrivalRates = rates;
	}
	
	public void setTimeWindow(long window) {
		this.timeWindow = window;
	}
	
	private void computeMinSpanningTree() {
		
		int[] flags = new int[this.numNodes];
		int flagindex = 0;
		
		for(int i=0; i<numNodes; i++) {
			flags[i] = 0;
		}
		
		// each time pick the smallest weighted edge from the edges
		// if the two vertex has flag 0, or have different flags, then put the edge to minspanningedge list,
		// then increase the flagindex, and set the flagindex of all vertex connecting to these two vertex
		// to the current flagindex
		
		double minweight = 0;
		int minindex = 0;
		
		for(int i=0; i < edges.size(); i++) {
			minweight = ((Edge)edges.get(0)).getWeight();
			minindex = 0;
			for(int k=1; k < edges.size(); k++) {
				if(((Edge)edges.get(k)).getWeight() < minweight) {
					minindex = k;
					minweight = ((Edge)edges.get(k)).getWeight();
				}
			}
			Edge minEdge = (Edge)edges.remove(minindex);
			
			//if both vertex are not yet covered
			if((flags[minEdge.getv1()] == 0) && (flags[minEdge.getv2()] == 0)) {
				this.minSpanningEdges.add(minEdge);
				flagindex++;
				flags[minEdge.getv1()] = flagindex;
				flags[minEdge.getv2()] = flagindex;
				continue;
			}
			
			//if one is covered, one is not
			if((flags[minEdge.getv1()] == 0) && (flags[minEdge.getv2()] != 0)) {
				this.minSpanningEdges.add(minEdge);
				flags[minEdge.getv1()] = flags[minEdge.getv2()];
				continue;
			}
			
			//if one is covered, one is not
			if((flags[minEdge.getv1()] != 0) && (flags[minEdge.getv2()] == 0)) {
				this.minSpanningEdges.add(minEdge);
				flags[minEdge.getv2()] = flags[minEdge.getv1()];
				continue;
			}
			
			//if both are covered but have different flagindex
			if(flags[minEdge.getv1()] != flags[minEdge.getv2()]) {
				this.minSpanningEdges.add(minEdge);
				flagindex++;
				int flag1 = flags[minEdge.getv1()];
				int flag2 = flags[minEdge.getv2()];
				for(int m=0; m<numNodes; m++) {
					if(flags[m] == flag1)
						flags[m] = flagindex;
					else if (flags[m] == flag2)
						flags[m] = flagindex;
				}
				continue;
			}
			
			i--;
		}
		
		setupMinJoinMatrix();
	}
	
	// setup join matrix based on min-spanning edges
	private void setupMinJoinMatrix() {
		minJoinMatrix = new int[this.numNodes][this.numNodes];
		int v1, v2;
		//init join matrix
		for(int i=0; i<this.numNodes; i++) 
			for(int k=0; k<this.numNodes; k++)
				minJoinMatrix[i][k] = 0;
		
		for(int i=0; i<this.minSpanningEdges.size(); i++) {
			v1 = ((Edge)this.minSpanningEdges.get(i)).getv1();
			v2 = ((Edge)this.minSpanningEdges.get(i)).getv2();
			minJoinMatrix[v1][v2] = 1;
			minJoinMatrix[v2][v1] = 1;
		}
	}
	
	/**
	 * Use the KBZ algorithm to compute the join ordering (probe sequence) for the ith vetex in the graph
	 * for a graph with n vertices, they are named from 0 ~ n-1
	 * 
	 * the order is computed based on the min-spanning tree
	 * @param i
	 * @return an array of n-1 integers, represents the order of joining for the ith vertex (input).
	 */
	public int[] computeIthJoinOrder(int index) {
		int[] order = new int[numNodes-1];
		int[] visited = new int[numNodes];
		
		for(int i=0; i<visited.length; i++) 
			visited[i] = 0;
		
		visited[index] = 1;
		RootedTreeNode rootNode = generateRootedTree(index, visited);
		
		applyIKAlgorithm(rootNode);
		
		
		for(int i=0; i<this.numNodes-1; i++) {
			RootedTreeNode cnode = (RootedTreeNode)rootNode.getChildren().get(0);
			order[i] = cnode.getNodeIndex();
			rootNode = cnode;
		}
		
		return order;
	}
	
	
	
	/**
	 * generate a rooted tree with the root denotes to the vertex with index i
	 */
	private RootedTreeNode generateRootedTree(int index, int[] visited) {
		RootedTreeNode root = new RootedTreeNode(index);
//		root.setCValue(0);
//		root.setTValue(this.arrivalRates[index]);
		
		for(int i=0; i<this.numNodes; i++) {
			if((this.minJoinMatrix[index][i] == 1) && (visited[i] == 0)){
				visited[i] = 1;
				RootedTreeNode child = generateRootedTree(i, visited);
				double selectivity = this.selectivities[index][i];
				double rate = this.arrivalRates[i];
				double cvalue = selectivity * rate * timeWindow/1000;
				double tvalue = selectivity * rate * timeWindow/1000;
				double rank = (tvalue - 1)/cvalue;
				child.setCValue(cvalue);
				child.setTValue(tvalue);
				child.setRank(rank);
				child.setParent(root);
				root.addChild(child);			
			}
		}
		return root;
	}
	
	/**
	 * Apply the IK algorithm to find the optimal join ordering for a tree rooted at rootNode.
	 * @param rootNode
	 */
	private void applyIKAlgorithm(RootedTreeNode rootNode) {
		//orderBranches(rootNode);
		mergeBranches(rootNode);
	}
	
	// order each children branch of the given rootNode.
//	private void orderBranches(RootedTreeNode rootNode) {
//		
//		RootedTreeNode[] cnodes = new RootedTreeNode[rootNode.getChildren().size()];
//		
//		for(int i=0; i<rootNode.getChildren().size(); i++) {
//			cnodes[i] = (RootedTreeNode)rootNode.getChildren().get(i);
//			orderOneBranch(cnodes[i]);
//		}
//	}
	
	private void mergeBranches(RootedTreeNode rootNode) {
		
		RootedTreeNode[] cnodes = new RootedTreeNode[rootNode.getChildren().size()];
		
		for(int i=0; i<rootNode.getChildren().size(); i++) {
			cnodes[i] = (RootedTreeNode)rootNode.getChildren().get(i);
			orderOneBranch(cnodes[i]);
		}
		
		RootedTreeNode currentNode = rootNode;
		while(!allSorted(cnodes)) {
			int index = getIndexWithLowestRank(cnodes);
			currentNode.getChildren().clear();
			currentNode.addChild(cnodes[index]);
			cnodes[index].setParent(currentNode);
			currentNode = cnodes[index];
			
			//RootedTreeNode nextnode = cnodes[index];
			
			if(currentNode.getChildren().size() == 0) {
				cnodes[index] = null;
			} 
			
			if(currentNode.getChildren().size() > 1) {
				System.out.println("Error!! Something is wrong. The node cannot have more than one children node!");
			}
			
			RootedTreeNode tempnode;
			while(currentNode.getChildren().size() == 1) {
				tempnode = (RootedTreeNode) currentNode.getChildren().get(0);
				if (!tempnode.isTailNode()) {
					//currentNode = nextnode;
					cnodes[index] = tempnode;
					break;
				} 
				currentNode = tempnode;
			}		
			if(currentNode.isTailNode() && (currentNode.getChildren().size() == 0))
				//currentNode = nextnode;
				cnodes[index] = null;
		}
	}
	
	private void orderOneBranch(RootedTreeNode startnode) {
		
		if(startnode.getChildren().size() == 0) 
			return;
		
		if(startnode.getChildren().size() > 1) {
			//orderBranches(startnode);
			mergeBranches(startnode);
		}
		
		RootedTreeNode cnode = (RootedTreeNode)startnode.getChildren().get(0);
		if(cnode.getRank() < startnode.getRank()) {
			cnode.setAsTailNode();
			//need to recompute the rank of startnode and cnode! their rank should be set to the same value
			double cvalue = startnode.getCValue() + startnode.getTValue() * cnode.getCValue();
			double tvalue = startnode.getTValue() * cnode.getTValue();
			double rank = (tvalue - 1)/cvalue;
			startnode.setCValue(cvalue);
			startnode.setTValue(tvalue);
			startnode.setRank(rank);
			cnode.setCValue(cvalue);
			cnode.setTValue(tvalue);
			cnode.setRank(rank);
		}
		
		orderOneBranch(cnode);
		
	}
	
	private boolean allSorted(RootedTreeNode[] nodes) {
		for(int i=0; i<nodes.length; i++) {
			if(nodes[i] != null)
				return false;
		}		
		return true;
	}
	
	private int getIndexWithLowestRank(RootedTreeNode[] nodes) {
		int minindex = -1;
		double minrank = Double.MAX_VALUE;
		
		for(int i=0; i<nodes.length; i++) {
			if(nodes[i] != null) {
				if(nodes[i].getRank() < minrank) {
					minindex = i;
					minrank = nodes[i].getRank();
				}
			}
		}	
		return minindex;
	}
	
	public double[][] getSelectivities() {
		return this.selectivities;
	}
	
	public double[] getArrivalRates() {
		return this.arrivalRates;
	}
	
	public long getTimeWindow() {
		return this.timeWindow;
	}
	
	public boolean getUseCartesianProductStatus() {
		return this.useCartesianProductStatus;
	}

	/**
	 * Compute the correct selectivities given the computed probe sequences.
	 * Also set up the random generator for each selectivity;
	 * This selectivity matrix is used at runtime for looking up certain selectivity between two consecutive 
	 * stream pairs in a join sequence.
	 */
	public double[][] computeProbeSelectivityMatrix(int[][] probeOrders) {
		double[][] probeSelectivityMatrix = new double[this.numNodes][this.numNodes - 1];
		//this.rand = new java.util.Random[this.numOfInputs][this.numOfInputs - 1];
		
		for(int i=0; i<this.numNodes; i++) {
			int index0 = probeOrders[i][0];
			probeSelectivityMatrix[i][0] = this.selectivities[i][index0];
			for(int j=1; j<this.numNodes-1; j++) {
				int indexj = probeOrders[i][j];
				probeSelectivityMatrix[i][j] = this.selectivities[i][indexj];
				for(int k=0; k<j; k++) {
					int indexk = probeOrders[i][k];
					probeSelectivityMatrix[i][j] *= this.selectivities[indexj][indexk];
				}
			}
		}
		return probeSelectivityMatrix;
	}
	
	public int[][] generateJoinOrders() {
		int[][] joinorders = new int[numNodes][numNodes-1];
	
		for(int i=0; i<numNodes; i++) {
			joinorders[i] = this.computeIthJoinOrder(i);
		}
		return joinorders;
	}
}
