/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.mjoinoptimizer;

//import java.util.ArrayList;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNodeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.probablisticjoinoperator.WMJoinOperator;


/**
 * @author Yali Zhu
 *
 */
public class MJoinEntry {
	
	private static final double COST_TUPLE_INSERT = 0.0002;
	private static final double COST_TUPLE_DELETE = 0.0002;
	private static final double COST_TUPLE_JOIN = 0.0022;
	
	public int inputSize;
	public int numOfStreams; //how many streams does this join entry contain
	public int entryID = -1;
	public int[] childrenIDs;
	//if the inputs are (AB) and (ABC), the childrensizes = {2, 3}.
	//used when computing costs.
	public int[] childrenstreamsizes;
	public int totalsize; //sum of all childrensize.
	
	public double[] inputRates;
	public double[][] selectivities;
	private int[][] probeOrders;
	private double[][] probeSelectivities;
	private JoinGraph joinGraph = null;
	private long timeWindow;
	
	//calculated using cost models
	private double outputRate;
	private double memoryCost;
	private double CPUCost;
	
	// if A join B, the entry name is AB.
	//String entryName;
	
//	public int joinID = -1;	
	public int parentID = -1;
	//public int lchildID = -1;
	//public int rchildID = -1;
	
	//indicate if this operator has been connected to the btree.
	public boolean connected = false;
	public WMJoinOperator operator;
	public boolean removed = false;
	
	public MJoinEntry(int entryid, int[] inputindex, int[] streamsizes, long window, WMJoinOperator operator) {
		this.entryID = entryid;
		this.inputSize = inputindex.length;
		this.childrenIDs = inputindex;
		this.timeWindow = window;
		this.operator = operator;
		this.childrenstreamsizes = streamsizes;
		this.numOfStreams = 0;
		for(int i=0; i<streamsizes.length; i++) {
			numOfStreams += streamsizes[i];
		}
		
		//this.calculateOutputRate();
		
		connected = false;
	}
	
	public void createEntryOperator() {
		DagNode newnode = new DagNodeImp();
		this.operator = new WMJoinOperator(newnode);
		
		operator.setRunningMode("mjoin");
		operator.setNumOfInputs(this.inputSize);
		operator.setSelectivityMatrix(selectivities);
		operator.setArrivalRates(inputRates);
		operator.setTimeWindow(this.timeWindow);
		operator.setupStates();
		operator.computeProbeSequence();
	}

	public void calculateProbeOrders() {
		this.probeOrders = new int[this.inputSize][this.inputSize - 1];
		this.joinGraph = null;
		this.joinGraph = new JoinGraph(this.selectivities, this.inputRates, true);
		this.joinGraph.setTimeWindow(this.timeWindow);
		
		//this.joinGraph.setArrivalRates(this.arrivalRates);
		//this.joinGraph.setSelectivities(this.selectivityMatrix);
		
		for (int i = 0; i < this.inputSize; i++) {
			this.probeOrders[i] = joinGraph.computeIthJoinOrder(i);
		}
		
		//printProbeOrders();
		calculateProbeSelectivities();
		//printProbeSelectivities();
	}
	
	/**
	 * Compute the correct selectivities given the computed probe sequences.
	 * Also set up the random generator for each selectivity;
	 * This selectivity matrix is used at runtime for looking up certain selectivity between two consecutive 
	 * stream pairs in a join sequence.
	 */
	private void calculateProbeSelectivities() {
		this.probeSelectivities = new double[this.inputSize][this.inputSize - 1];
		
		for(int i=0; i<this.inputSize; i++) {
			int index0 = this.probeOrders[i][0];
			probeSelectivities[i][0] = this.selectivities[i][index0];
			for(int j=1; j<this.inputSize-1; j++) {
				int indexj = this.probeOrders[i][j];
				probeSelectivities[i][j] = this.selectivities[i][indexj];
				for(int k=0; k<j; k++) {
					int indexk = this.probeOrders[i][k];
					probeSelectivities[i][j] *= this.selectivities[indexj][indexk];
				}
			}
		}
	}
	
	private void printProbeOrders() {
		for(int i=0; i<this.inputSize; i++) {
			System.out.print("join order for input "+i + ": ");
			for(int k=0; k<this.inputSize - 1; k++) {
				System.out.print(" -> " + this.probeOrders[i][k]);
			}
			System.out.println();
		}
	}
	
	private void printProbeSelectivities() {
		for(int i=0; i<this.inputSize; i++) {
			System.out.print("join selectivity for input "+i + ": ");
			for(int k=0; k<this.inputSize - 1; k++) {
				System.out.print(" -> " + this.probeSelectivities[i][k]);
			}
			System.out.println();
		}
		
		System.out.print("children input size: ");
		for(int i=0; i<this.inputSize; i++) {
			System.out.print(this.childrenstreamsizes[i] + " ");
		}
		System.out.println();
	}
	
	public void calculateMemoryCost() {
		this.memoryCost = 0;
		
		for(int i=0; i<inputRates.length; i++)
			memoryCost += inputRates[i]* this.childrenstreamsizes[i] * this.timeWindow/1000;
			//memoryCost += inputRates[i]* this.timeWindow/1000;
	}
	
	public void calculateCPUCost() {
		this.CPUCost = 0;
		
		for(int i=0; i<this.inputSize; i++) {
			// first the cost of inserting and deleting tuples to/from states
			this.CPUCost += this.inputRates[i]*this.childrenstreamsizes[i] * (COST_TUPLE_INSERT + COST_TUPLE_DELETE);
			
			double inter_results = this.inputRates[i] * this.childrenstreamsizes[i];
			double temp;
			// then the cost of probing each of the remaining states
			for(int k=0; k<(this.inputSize-1); k++) {
				int probedinput = this.probeOrders[i][k];
				double statesize = this.inputRates[probedinput] * this.timeWindow/1000;
				temp = inter_results * statesize * this.probeSelectivities[i][k];
				this.CPUCost += temp * (COST_TUPLE_JOIN);
				inter_results = temp;
			}
		}
	}
	
	public void calculateOutputRate() {
		outputRate = 1;
		for(int i=0; i<inputRates.length; i++)
			outputRate *= inputRates[i] * timeWindow/1000;
		
		for(int i=0; i<this.selectivities.length; i++) {
			for(int j=i+1; j<this.selectivities.length; j++) {
				outputRate *= selectivities[i][j];
			}
		}
		
		outputRate = outputRate *1000 / timeWindow;
	}
	
	public void setConnected() {
		this.connected = true;
	}
	
	public boolean isConnected() {
		return this.connected;
	}
	
	public boolean containIndex(int i) {
		for(int k=0; k<inputSize; k++) {
			if(childrenIDs[k] == i) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * @param cid --  ID of the child entry
	 * @return The index of the input child ID in the childID array.
	 *         If the input id does not exist in the childID array, return -1.    
	 */
	public int IndexOfChildID(int cid) {
		for(int k=0; k<inputSize; k++) {
			if(childrenIDs[k] == cid) {
				return k;
			}
		}
		return -1;
	}

	//	public int getv1() {
	//		return v1;
	//	}
	//	
	//	public int getv2() {
	//		return v2;
	//	}
	//	
	//	public double getrate1() {
	//		return rate1;
	//	}
	//	
	//	public double getrate2() {
	//		return rate2;
	//	}
	//	
	//	public double getselectivity() {
	//		return selectivity;
	//	}
	//	
	//	public double getWeight() {
	//		return this.weight;
	//	}
		
		
		public String toString() {
			String rs = new String();
			
			rs += "\nJoinEntry: ";
			rs += "ID=" + entryID;
			rs += " children IDs: ";
			for(int i=0; i<childrenIDs.length; i++) {
				rs += childrenIDs[i] + " ";
			}
			
			rs += " memcost: " + this.memoryCost;
			rs += " cpucost: " + this.CPUCost;


			return rs;
		}
		
		public double getMemoryCost() {
			return this.memoryCost;
		}
		
		public double getCPUCost() {
			return this.CPUCost;
		}
		
		public double getOutputRate() {
			return this.outputRate;
		}
		
		/**
		 * @param maxStreamEntryID
		 * @return Index of the input that connects to the largest intermediate state in this join entry.
		 *         If no intermediate state is present in this entry, return -1.
		 */
		public int getIndexOfFastestInput(int maxStreamEntryID) {
			int index = -1;
			double max_state = 0;
			
			for(int i=0; i<childrenIDs.length; i++) {
				if ((childrenIDs[i] > maxStreamEntryID) && (inputRates[i] > max_state)) {
					index = i;
					max_state = inputRates[i];
				}
			}
			
			return index;
		}
		
		public void setSelectivities(double[][] s) {
			this.selectivities = s;
		}
		
		public void setInputRates(double[] r) {
			this.inputRates = r;
			//this.calculateOutputRate();
		}
		
		public void setTimeWindow(long w) {
			this.timeWindow = w;
		}
		
		public void setProbeOrders(int[][] o) {
			this.probeOrders = o;
		}
		
		public void setProbeSelectivities(double[][] s) {
			this.probeSelectivities = s;
		}
		
		public void setCPUCost(double c) {
			this.CPUCost = c;
		}
		
		public void setMemoryCost(double m) {
			this.memoryCost = m;
		}
}
