/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.mjoinoptimizer;

import java.util.Hashtable;
import java.util.Random;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDag;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATDagImp;

public class Tester2 {
	public static void main(String args[]) {
		Random raterand = new Random();
		Random selectrand = new Random();
		StringBuffer avgstats = new StringBuffer();
		//StringBuffer runtime = new StringBuffer(); 
		StringBuffer costs = new StringBuffer();
		
		int raterange = 100;
		int selectdividend = 1;
		int inputrange = 20;
		
		double mavgtime = 0;    // average runningtime for each n.
		double bavgtime = 0;
		double dpavgtime = 0;
		double mavgmemory = 0;  //average memory cost of qualified plan for each n.
		double bavgmemory = 0;
		double dpavgmemory = 0;
		double mavgcpu = 0;     //average cpu cost of qualified plan for each n.
		double bavgcpu = 0;
		double dpavgcpu = 0;
		int mhit = 0;
		int bhit = 0;
		int dphit = 0;
		
		Tester2 experiment = new Tester2();
		
		FromMJoinOptimizer moptimizer = new FromMJoinOptimizer(null);
		FromBTreeOptimizer boptimizer = new FromBTreeOptimizer(null);
		ExhaustiveSearchOptimizer dpoptimizer = new ExhaustiveSearchOptimizer(null);

		raterand.setSeed(1);
		selectrand.setSeed(0);
		long runtime = 0;
		
		mavgtime = 0;    
		bavgtime = 0;
		mavgmemory = 0;  
		bavgmemory = 0;
		mavgcpu = 0;    
		bavgcpu = 0;
		mhit = 0;
		bhit = 0;
		
		dpavgtime = 0;
		dpavgmemory = 0;
		dpavgcpu = 0;
		dphit = 0;
		
		int n = 5;
		JoinGraph g = experiment.generateRandomJoinGraph(n, raterand, selectrand, raterange, selectdividend);
		experiment.printJoinGraph(g);
		boptimizer.setJoinGraph(g);
		boptimizer.setQueueModel("XATSimpleQueueImp");
		boptimizer.optimize(null);
		
	}
	/**
	 * Given the size of input streams, this method generate a join graph
	 * with randomly generated input stream rates and selectivities.
	 * @param n -- size of input streams.
	 * @return the randomly generates join graph with n input streams.
	 */
	private JoinGraph generateRandomJoinGraph(int n, Random raterand, Random selectrand, int raterange, int selectdividend) {
		double[] rates = new double[n];
		double[][] selectivities = new double[n][n];
		

		//generate rates in the range of [1, 100].
		for(int i=0; i<n; i++) {
			rates[i] = raterand.nextDouble()*raterange;
		}
		
		//generate selectivities in the range of [0, 1]. 
		for(int i=0; i<n; i++) {
			selectivities[i][i] = 1;
			for(int j=0; j<i; j++) {
				selectivities[i][j] = selectrand.nextDouble()/selectdividend;
				selectivities[j][i] = selectivities[i][j];
			}
		}
		JoinGraph joingraph = new JoinGraph(selectivities, rates, true);
		joingraph.setTimeWindow(1000);
		
		
		return joingraph;
	}
	
	private void printJoinGraph(JoinGraph g) {
		double[] rates = g.getArrivalRates();
		double[][] selectivities = g.getSelectivities();
		
		int n = g.numNodes;
		
		System.out.println("** new join graph n = " + n + "**");
		System.out.println("rates: ");
		
		for(int i=0; i<n; i++) {
			System.out.println(i+"  "+rates[i]);
		}
		System.out.println();
		
		System.out.println("selectivities: ");
		//generate selectivities in the range of [0, 1]. 
		for(int i=0; i<n; i++) {
			for(int j=0; j<n; j++) {
				System.out.println("["+i+","+j+"]"+selectivities[i][j] + " ");
			}
			System.out.println();
		}
		System.out.println("*****************************\n");
	}
	
}
