/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.optimizer.rewritingrules;

//import java.util.ArrayList;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
//simport java.util.Iterator;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.migration.probablisticjoinoperator.ChangingWBJoinOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.realjoinoperator.WTransitiveHashJoinOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.HashStatePartitionGroup;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WEquiJoinStateImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.WFIFOState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.StreamSelectOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.ProbabilisticWindowJoinOperatorImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.predicate.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.WBinaryJoinOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
//import edu.wpi.cs.cs4432.project3.heuristics.Optimizer;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.ConnectionTable;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection.OperatorQueue;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.NetworkPacket;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.ReplaceInConnectionEntry;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleDistributor;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.TupleFeeder;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.DistributedExecutionController;
import edu.wpi.cs.dsrg.xmldb.xat.component.migrationcontroller.XATBox;
import edu.wpi.cs.dsrg.xmldb.xat.common.relationalstreamexpression.ExpressionStr;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstreamoperator.BinaryPunctuationStreamSingleEquiJoinOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.punctuationstreamoperator.BinaryPWSingleEquiJoinOperator;

/**
 * @author Yali Zhu
 * 
 * This rewriting rule switch the position of two xatoperators (maybe their associated xatnodes as well) in a DAG
 * query plan tree. The Switch methods are static, and needs two parameters. The operator passed in as the first 
 * parameter is always the upper operator and the second parameter is always the lower operator. 
 * The appropriate Switch method will be invoked based on the parameters passed in.  
 * 
 */
public class SwitchOperatorRules {

	public SwitchOperatorRules() {
		super();
	}
	
	public static boolean Switch(XATStreamOperator upperop, XATStreamOperator lowerop, int maxstateid, XATBox newbox, DistributedExecutionController ec) { 
		
		try {
		    System.out.println("Switch operators:\n" + upperop + " & " + lowerop);
			if ((upperop instanceof WTransitiveHashJoinOperator) && (lowerop instanceof WTransitiveHashJoinOperator)) {

				return SwitchOperators(
					(WTransitiveHashJoinOperator)upperop,
					(WTransitiveHashJoinOperator)lowerop, maxstateid, newbox, ec);
			}
			
			if ((upperop instanceof ChangingWBJoinOperatorImp) && (lowerop instanceof ChangingWBJoinOperatorImp)) {

				return SwitchOperators(
					(ChangingWBJoinOperatorImp)upperop,
					(ChangingWBJoinOperatorImp)lowerop, maxstateid, newbox, ec);
			}
			
			if ((upperop instanceof WBinaryJoinOperator) && (lowerop instanceof WBinaryJoinOperator)) {

				return SwitchOperators(
					(WBinaryJoinOperator) upperop,
					(WBinaryJoinOperator) lowerop, maxstateid, newbox, ec);
			}
			
			if ((upperop instanceof ProbabilisticWindowJoinOperatorImp) && (lowerop instanceof ProbabilisticWindowJoinOperatorImp)) {
				return SwitchOperators(
					(ProbabilisticWindowJoinOperatorImp) upperop,
					(ProbabilisticWindowJoinOperatorImp) lowerop, maxstateid, newbox, ec);
			}
			
			if ((upperop instanceof BinaryPunctuationStreamSingleEquiJoinOperator) && (lowerop instanceof BinaryPunctuationStreamSingleEquiJoinOperator)) {
				return SwitchOperators(
					(BinaryPunctuationStreamSingleEquiJoinOperator) upperop,
					(BinaryPunctuationStreamSingleEquiJoinOperator) lowerop, maxstateid, newbox, ec);
			}
			
//			if ((upperop instanceof BinaryPWSingleEquiJoinOperator) && (lowerop instanceof BinaryPWSingleEquiJoinOperator)) {
//				return Switch(
//					(BinaryPWSingleEquiJoinOperator) upperop,
//					(BinaryPWSingleEquiJoinOperator) lowerop,
//					newbox);
//			}

			if ((upperop instanceof StreamSelectOperator) && (lowerop instanceof StreamSelectOperator)) {
				return SwitchOperators (
					(StreamSelectOperator) upperop,
					(StreamSelectOperator) lowerop, 
					newbox, ec);
			}

			if ((upperop instanceof WBinaryJoinOperator) && (lowerop instanceof WBinaryJoinOperator)) {
				return SwitchOperators(
					(WBinaryJoinOperator) upperop,
					(WBinaryJoinOperator) lowerop, maxstateid, newbox, ec);
			} else {
				throw new Exception("Unsupported operator type for switch operator rules.");
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return false;
	}
	
	/**
	 * Switch two select operators if the upper selectivity is lower than the lower selectivity
	 * 
	 * Creation Date: 05/01/04
	 * @author Sicilia Wijaya
	 * @param upperSelect 
	 * @param lowerSelect 
	 * @return boolean whether switching is performed
	 */
	public static boolean SwitchOperators(StreamSelectOperator upperSelect, StreamSelectOperator lowerSelect, XATBox newbox, DistributedExecutionController ec) {
		
		//First of all, set the parent of the lowerSelect to be the parent of upperSelect
		if (upperSelect.getNode().getParent() != null)
		{
			XATNode parent = upperSelect.getNode().getParent();
			int index = parent.getIndexOfOpChild(upperSelect.getNode());
			
			parent.setOpChild(index, lowerSelect.getNode());
			lowerSelect.getNode().setParent(parent);
			parent.getStreamOperator().replaceInputQueue(upperSelect.getOutputQueue(), lowerSelect.getOutputQueue());
			
		}
		else
		{
			//if upperSelect has no parent, i.e. was a root, set lowerSelect as the new root of the tree
			lowerSelect.getNode().setParent(null);
			if(newbox != null)
				newbox.setRootOperator(lowerSelect);
			//Optimizer.getTree().setRoot(lowerSelect.getNode());
		}
		
		/* if the input queue schema of upper and lower select are not the same
		 *  that means schema cleanup has minimize the output schema of lowerSelect
		 * in that case we have to reset the column position of the expressions in upperSelect
		 */
		if (upperSelect.getInputQueue().getSchema().getColNames()
			!= lowerSelect.getInputQueue().getSchema().getColNames()) {
			// @TODO do something about the expression.
		}
		
		//Secondly, let upperSelect adopt the children of lowerSelect
		if (lowerSelect.getNode().getOpChildrenSize() > 0)
		{
			XATNode [] children = lowerSelect.getNode().getOpChildren();
			for (int i=0; i < children.length; i++)
			{
				//XATNode child = children[i];
				upperSelect.getNode().addOpChild(children[i]);
				children[i].setParent(upperSelect.getNode());
				upperSelect.replaceInputQueue(lowerSelect.getOutputQueue(), children[i].getStreamOperator().getOutputQueue());
				//delete the child from lowerSelect
				lowerSelect.getNode().deleteOpChild(children[i]);
				
			}
		}
		else
			upperSelect.setInputQueue(lowerSelect.getInputQueue());
			
		//Finally, set the parent and children relationship between
		//upperSelect and lowerSelect
		upperSelect.getNode().deleteOpChild(lowerSelect.getNode());
		upperSelect.getNode().setParent(lowerSelect.getNode());
		lowerSelect.getNode().setOpChild(0,upperSelect.getNode());
		lowerSelect.setInputQueue(upperSelect.getOutputQueue());
		
		return true;
	}
	/**
	 * Switch a select and join operator
	 * 
	 * Creation Date: 05/01/04
	 * @author Sicilia Wijaya
	 * @param upperSelect
	 * @param lowerJoin
	 * @return boolean whether switching is performed
	 */
	public static boolean SwitchOperators(StreamSelectOperator upperSelect, WBinaryJoinOperator lowerJoin, XATBox newbox, DistributedExecutionController ec) {

		//First of all, see if we can actually move upperSelect to either the left
		//or right branch of lowerJoin
		Vector leftColNames = lowerJoin.getNthInputQueue(0).getSchema().getColNames();
		Vector rightColNames = lowerJoin.getNthInputQueue(1).getSchema().getColNames();
		Vector allColNames = new Vector();
		allColNames.addAll(leftColNames);
		allColNames.addAll(rightColNames);
		
		if (upperSelect.getInputQueue().getSchema().getColNames() != allColNames) {
//			@TODO do something about the expression.
		}
		
		int[] leftColumnPositions = new int[leftColNames.size()];
		int[] rightColumnPositions = new int[rightColNames.size()];
		for(int i=0; i<leftColumnPositions.length; i++) {
			leftColumnPositions[i] = i;
		}
		for(int i=0; i<rightColumnPositions.length; i++) {
			rightColumnPositions[i] = i+leftColumnPositions.length;
		}
			

		//First of all, let the parent of upperSelect adopt lowerJoin in place of upperSelect
		if (upperSelect.getNode().getParent() != null)
		{
			XATNode parent = upperSelect.getNode().getParent();
			int index = parent.getIndexOfOpChild(upperSelect.getNode());
			parent.setOpChild(index, lowerJoin.getNode());
			lowerJoin.getNode().setParent(parent);
			parent.getStreamOperator().replaceInputQueue(upperSelect.getOutputQueue(), lowerJoin.getOutputQueue());
			
			//delete upperSelect from the parent's children
			//parent.deleteChild (upperSelect.getNode());
		}
		else
		{
			lowerJoin.getNode().setParent(null);
			//Optimizer.getTree().setRoot(lowerJoin.getNode());
			if(newbox != null)
				newbox.setRootOperator(lowerJoin);
		}
		
		/*
		 * Right now only push select to ONE side of the join
		 * @TODO add the ability to break upperSelect operator if necessary
		 *       and push them to both input sides of the join. 
		 */
		ExpressionStr leftSubExpression = upperSelect.getExpression().getRelatedSubExpression(0, leftColumnPositions);

		// if push to the left
		if(leftSubExpression.equals(upperSelect.getExpression())) {
		
			/* if this is a leaf join operator*/
			if(lowerJoin.getNode().getOpChildrenSize() == 0) {
				lowerJoin.getNode().addOpChild(upperSelect.getNode());
			}
			
			/* if the join operator has one child, see which side it is attached to this join */
			if(lowerJoin.getNode().getOpChildrenSize() > 0) {
				XATNode child = lowerJoin.getNode().getOpChild(0);
				
				// if the left-most child is indeed connected to the left most queue
				if (lowerJoin.getNthInputQueue(0).getSchema().getColNames()
					== child.getStreamOperator().getOutputQueue().getSchema().getColNames()) {
						lowerJoin.getNode().setOpChild(0, upperSelect.getNode());
						child.setParent(upperSelect.getNode());
						upperSelect.getNode().addOpChild(child);
				}	
			}
			
			upperSelect.getNode().setParent(lowerJoin.getNode());
			upperSelect.getNode().deleteOpChild(lowerJoin.getNode());
			upperSelect.replaceInputQueue(lowerJoin.getOutputQueue(), lowerJoin.getNthInputQueue(0));
			
			// reset upperselect's output schema and lowerjoin's input schema
			upperSelect.getOutputQueue().setSchema(upperSelect.getInputQueue().getSchema());
			lowerJoin.replaceInputQueue(lowerJoin.getNthInputQueue(0), upperSelect.getOutputQueue());
			lowerJoin.getNthInputQueue(0).setSchema(upperSelect.getOutputQueue().getSchema());
			//lowerJoin.reinit();
			
			// as the last step. reinitialize these two operators, 
			// starting from the new lower op (orig upper op), MUST FOLLOW THIS ORDER!
			upperSelect.init();
			lowerJoin.init();
			return true;
		}
		
		// if push to the right
		ExpressionStr rightSubExpression = upperSelect.getExpression().getRelatedSubExpression(0, rightColumnPositions);		
		if(rightSubExpression.equals(upperSelect.getExpression())) {
				
			/* if this is a leaf join operator*/
			if(lowerJoin.getNode().getOpChildrenSize() == 0) {
				lowerJoin.getNode().addOpChild(upperSelect.getNode());
			}
			
			/* if the join operator has more than one child, see which side it is attached to this join */
			if(lowerJoin.getNode().getOpChildrenSize() > 0) {
				XATNode[] children = lowerJoin.getNode().getOpChildren();
				int i;
				for(i=0; i<children.length; i++) {
					// if the left-most child is indeed connected to the left most queue
					if (lowerJoin.getNthInputQueue(1).getSchema().getColNames()
						== children[i].getStreamOperator().getOutputQueue().getSchema().getColNames()) {
							
							lowerJoin.getNode().setOpChild(i, upperSelect.getNode());
							children[i].setParent(upperSelect.getNode());
							upperSelect.getNode().addOpChild(children[i]);
							break;
					}	
				}
				
				// if none of the children's output queue schema matches the second input queue of lower join
				if(i == children.length) {
					lowerJoin.getNode().addOpChild(upperSelect.getNode());
				}
			} 
			
			upperSelect.getNode().setParent(lowerJoin.getNode());
			upperSelect.getNode().deleteOpChild(lowerJoin.getNode());
			upperSelect.replaceInputQueue(lowerJoin.getOutputQueue(), lowerJoin.getNthInputQueue(0));
			
			// reset upperselect's output schema and lowerjoin's input schema
			upperSelect.getOutputQueue().setSchema(upperSelect.getInputQueue().getSchema());
			lowerJoin.replaceInputQueue(lowerJoin.getNthInputQueue(0), upperSelect.getOutputQueue());
			lowerJoin.getNthInputQueue(0).setSchema(upperSelect.getOutputQueue().getSchema());
			lowerJoin.init();
			
			return true;
		}

		return false;
	}
	/**
	 * Switch two join operators
	 * 
	 * Creation Date: 05/01/04
	 * @param upperJoin
	 * @param lowerJoin
	 * @return boolean whether switching is performed
	 */
	public static boolean SwitchOperators(WBinaryJoinOperator upperJoin, WBinaryJoinOperator lowerJoin, int maxstateID, XATBox newbox, DistributedExecutionController ec) 
	{
		// two things to each for each node: parent, children
		// four things to change for each operator: input queue, output queue and 
		// state schema for window join operator (run init) again to achieve this goal,
		// this will take care of other things that need to be re-initialized.
		boolean isRightChild = true; // indicates if lowerNode is the right child of upperNode.
		int maxstateid = newbox.getStateCounter();
		
		/** first change the input and output queues of involved operators */
		if(upperJoin.getNthInputQueue(0).equals(lowerJoin.getOutputQueue())) {
			isRightChild = false; 
		}
		System.out.println("lower is upper's right child: "+isRightChild);
		
		// first check expressions
		// if this switch resulting in potential cartesian products, don't switch
		if(isRightChild) {
			
			if (!upperJoin.getExpression().containColumnPositions(1, lowerJoin.getOutputPositions(0))) {
				// potential cartesian product, don't switch!
				return false;
			}
		} else {

			if (!upperJoin.getExpression().containColumnPositions(0, lowerJoin.getOutputPositions(1))) {
				// potential cartesian product, don't switch!
				return false;
			}		
		}	
			
		XATQueue lowerLeftInputQueue = lowerJoin.getLeftInputQueue();
		XATQueue lowerRightInputQueue = lowerJoin.getRightInputQueue();
		XATQueue upperLeftInputQueue = upperJoin.getLeftInputQueue();
		XATQueue upperRightInputQueue = upperJoin.getRightInputQueue();
		XATQueue lowerOutputQueue = lowerJoin.getOutputQueue();
		XATQueue upperOutputQueue = upperJoin.getOutputQueue();
		
		
		XATNode lowerLeftChild;
		XATNode lowerRightChild;
		
		if(lowerJoin.getNode().getOpChildrenSize() == 0) {
			lowerLeftChild = null;
			lowerRightChild = null;
		} else if(lowerJoin.getNode().getOpChildrenSize() == 1)  {
			if (lowerJoin.getNthInputQueue(0).equals(lowerJoin.getNode().getOpChild(0).getStreamOperator().getOutputQueue())) {

				lowerLeftChild = lowerJoin.getNode().getOpChild(0);
				lowerRightChild = null;
			}
			else {
				lowerLeftChild = null;
				lowerRightChild = lowerJoin.getNode().getOpChild(0);
			}
		} else if (lowerJoin.getNode().getOpChildrenSize() == 2) {
			lowerLeftChild = lowerJoin.getNode().getOpChild(0);
			lowerRightChild = lowerJoin.getNode().getOpChild(1);
		} else {
			System.out.println("Binary join can only have at most two children. ");
			return false;
		}
		
		/* first take car of queue connnections */
		if(isRightChild) {
			
			//upperOutputQueue.removeOwner(upperJoin);
			//lowerLeftInputQueue.unregister(lowerJoin);
			
			upperRightInputQueue.unregister(upperJoin);

			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerLeftInputQueue, upperRightInputQueue);
			upperJoin.replaceInputQueue(upperRightInputQueue, lowerLeftInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
			lowerOutputQueue.unregister(upperJoin);
		}
		else {
			upperLeftInputQueue.unregister(upperJoin);
			//lowerRightInputQueue.unregister(lowerJoin);
			
			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerRightInputQueue, upperLeftInputQueue);
			upperJoin.replaceInputQueue(upperLeftInputQueue, lowerRightInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
		}

		/** then take care of the parent-child relation of involved XATNodes */
		XATNode upperParentNode = upperJoin.getNode().getParent();
		
		if(isRightChild) {
				lowerJoin.getNode().setParent(upperParentNode);
				lowerJoin.getNode().addOpChild(upperJoin.getNode(), 0);
				lowerJoin.getNode().deleteOpChild(lowerLeftChild);
				
				upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
				upperJoin.getNode().addOpChild(lowerLeftChild);
				upperJoin.getNode().setParent(lowerJoin.getNode());
				
				if(lowerLeftChild != null) {
					lowerLeftChild.setParent(upperJoin.getNode());
				}
				
				// take care of stateID
				upperJoin.getRightState().setStateID(lowerJoin.getLeftState().getStateID());
				maxstateid++;
				lowerJoin.getLeftState().setStateID(maxstateid);				
		}
		else {
			lowerJoin.getNode().setParent(upperParentNode);
			lowerJoin.getNode().addOpChild(upperJoin.getNode());
			lowerJoin.getNode().deleteOpChild(lowerRightChild);
				
			upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
			upperJoin.getNode().addOpChild(lowerRightChild, 0);
			upperJoin.getNode().setParent(lowerJoin.getNode());
			
			if(lowerRightChild != null) {
				lowerRightChild.setParent(upperJoin.getNode());
			}
			
			// take care of stateID
			upperJoin.getLeftState().setStateID(lowerJoin.getRightState().getStateID());
			maxstateid++;
			lowerJoin.getRightState().setStateID(maxstateid);		
				
		}
			
		if (upperParentNode != null) {
			int position = upperParentNode.getIndexOfOpChild(upperJoin.getNode());
			upperParentNode.addOpChild(lowerJoin.getNode(), position);
			upperParentNode.deleteOpChild(upperJoin.getNode());
			upperJoin.getNode().setParent(lowerJoin.getNode());
		} 
		
		if ((upperParentNode == null) || (!upperParentNode.getStreamOperator().isBoxOperator())){
		//	lowerJoin.getNode().setParent(null);
		//	upperJoin.getNode().setParent(lowerJoin.getNode());
			if(newbox != null)
				newbox.setRootOperator(lowerJoin);
		}
		
		// as the last step. reinitialize these two operators, 
		// starting from the new lower op (orig upper op), MUST FOLLOW THIS ORDER!
		upperJoin.init();
		lowerJoin.init();
		return true;
	}
		
		
	public static boolean SwitchOperators(ProbabilisticWindowJoinOperatorImp upperJoin, ProbabilisticWindowJoinOperatorImp lowerJoin, int maxstateID, XATBox newbox, DistributedExecutionController ec) 
	{
		// two things to each for each node: parent, children
		// four things to change for each operator: input queue, output queue and 
		// state schema for window join operator (run init) again to achieve this goal,
		// this will take care of other things that need to be re-initialized.
		boolean isRightChild = true; // indicates if lowerNode is the right child of upperNode.
//		int maxstateid = newbox.getStateCounter();
		
		/** first change the input and output queues of involved operators */
		if(upperJoin.getNthInputQueue(0).equals(lowerJoin.getOutputQueue())) {
			isRightChild = false; 
		}
		System.out.println("lower is upper's right child: "+isRightChild);
			
		XATQueue lowerLeftInputQueue = lowerJoin.getLeftInputQueue();
		XATQueue lowerRightInputQueue = lowerJoin.getRightInputQueue();
		XATQueue upperLeftInputQueue = upperJoin.getLeftInputQueue();
		XATQueue upperRightInputQueue = upperJoin.getRightInputQueue();
		XATQueue lowerOutputQueue = lowerJoin.getOutputQueue();
		XATQueue upperOutputQueue = upperJoin.getOutputQueue();
		
		
		XATNode lowerLeftChild;
		XATNode lowerRightChild;
		
		if(lowerJoin.getNode().getOpChildrenSize() == 0) {
			lowerLeftChild = null;
			lowerRightChild = null;
		} else if(lowerJoin.getNode().getOpChildrenSize() == 1)  {
			if (lowerJoin.getNthInputQueue(0).equals(lowerJoin.getNode().getOpChild(0).getStreamOperator().getOutputQueue())) {

				lowerLeftChild = lowerJoin.getNode().getOpChild(0);
				lowerRightChild = null;
			}
			else {
				lowerLeftChild = null;
				lowerRightChild = lowerJoin.getNode().getOpChild(0);
			}
		} else if (lowerJoin.getNode().getOpChildrenSize() == 2) {
			lowerLeftChild = lowerJoin.getNode().getOpChild(0);
			lowerRightChild = lowerJoin.getNode().getOpChild(1);
		} else {
			System.out.println("Binary join can only have at most two children. ");
			return false;
		}
		
		/* first take car of queue connnections */
		if(isRightChild) {
			
			upperRightInputQueue.unregister(upperJoin);

			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerLeftInputQueue, upperRightInputQueue);
			upperJoin.replaceInputQueue(upperRightInputQueue, lowerLeftInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
			lowerOutputQueue.unregister(upperJoin);
		}
		else {
			upperLeftInputQueue.unregister(upperJoin);
			//lowerRightInputQueue.unregister(lowerJoin);
			
			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerRightInputQueue, upperLeftInputQueue);
			upperJoin.replaceInputQueue(upperLeftInputQueue, lowerRightInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
		}

		/** then take care of the parent-child relation of involved XATNodes */
		XATNode upperParentNode = upperJoin.getNode().getParent();
		
		if(isRightChild) {
				lowerJoin.getNode().setParent(upperParentNode);
				lowerJoin.getNode().addOpChild(upperJoin.getNode(), 0);
				lowerJoin.getNode().deleteOpChild(lowerLeftChild);
				
				upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
				upperJoin.getNode().addOpChild(lowerLeftChild);
				upperJoin.getNode().setParent(lowerJoin.getNode());
				
				if(lowerLeftChild != null) {
					lowerLeftChild.setParent(upperJoin.getNode());
				}
				
				// take care of stateID
//				upperJoin.getRightState().setStateID(lowerJoin.getLeftState().getStateID());
//				maxstateid++;
//				lowerJoin.getLeftState().setStateID(maxstateid);				
		}
		else {
			lowerJoin.getNode().setParent(upperParentNode);
			lowerJoin.getNode().addOpChild(upperJoin.getNode());
			lowerJoin.getNode().deleteOpChild(lowerRightChild);
				
			upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
			upperJoin.getNode().addOpChild(lowerRightChild, 0);
			upperJoin.getNode().setParent(lowerJoin.getNode());
			
			if(lowerRightChild != null) {
				lowerRightChild.setParent(upperJoin.getNode());
			}
			
			// take care of stateID -- no states in probablisticjoin operators.
//			upperJoin.getLeftState().setStateID(lowerJoin.getRightState().getStateID());
//			maxstateid++;
//			lowerJoin.getRightState().setStateID(maxstateid);		
				
		}
			
		if (upperParentNode != null) {
			int position = upperParentNode.getIndexOfOpChild(upperJoin.getNode());
			upperParentNode.addOpChild(lowerJoin.getNode(), position);
			upperParentNode.deleteOpChild(upperJoin.getNode());
			upperJoin.getNode().setParent(lowerJoin.getNode());
		} 
		
		if ((upperParentNode == null) || (!upperParentNode.getStreamOperator().isBoxOperator())){
			if(newbox != null)
				newbox.setRootOperator(lowerJoin);
		}

		// as the last step. reinitialize these two operators, 
		// starting from the new lower op (orig upper op), MUST FOLLOW THIS ORDER!
		upperJoin.init();
		lowerJoin.init();
		return true;
	}
	
	
	/** If used by distributed optimizer, set newBox to be null */
	public static boolean SwitchOperators(BinaryPunctuationStreamSingleEquiJoinOperator upperJoin, BinaryPunctuationStreamSingleEquiJoinOperator lowerJoin, int maxstateID, XATBox newbox, DistributedExecutionController ec) 
	{
		// two things to each for each node: parent, children
		// four things to change for each operator: input queue, output queue and 
		// state schema for window join operator (run init) again to achieve this goal,
		// this will take care of other things that need to be re-initialized.
		boolean isRightChild = true; // indicates if lowerNode is the right child of upperNode.
		int maxstateid = 0;
		if(newbox != null)
			maxstateid = newbox.getStateCounter();
		
		/** first change the input and output queues of involved operators */
		if(upperJoin.getNthInputQueue(0).equals(lowerJoin.getOutputQueue())) {
			isRightChild = false; 
		}
		System.out.println("lower is upper's right child: "+isRightChild);
		
		// first check join attr position, recompute the position if necessary
		int upperLeftJoinPos = upperJoin.getLeftJoinAttrPos();
		int upperRightJoinPos = upperJoin.getRightJoinAttrPos();
		int lowerLeftJoinPos = lowerJoin.getLeftJoinAttrPos();
		int lowerRightJoinPos = lowerJoin.getRightJoinAttrPos();
			
		int upperLeftColSize = upperJoin.getLeftInputQueue().getSchema().getLength();
		int upperRightColSize = upperJoin.getRightInputQueue().getSchema().getLength();
		int lowerLeftColSize = lowerJoin.getLeftInputQueue().getSchema().getLength();
		int lowerRightColSize = lowerJoin.getRightInputQueue().getSchema().getLength();
		
		if(isRightChild) {
			
			if(upperRightJoinPos >= lowerLeftColSize) {
				System.out.println("Potential cartesian product after swith, so don't switch!");
				return false;
			}
			
			lowerJoin.setLeftJoinAttrPos(upperLeftColSize + lowerLeftJoinPos);
		} else {
			
			if(upperLeftJoinPos < lowerLeftColSize) {
				System.out.println("Potential cartesian product after swith, so don't switch!");
				return false;
			}
			
			upperJoin.setLeftJoinAttrPos(upperLeftJoinPos - lowerLeftColSize);		
		}	
			
		XATQueue lowerLeftInputQueue = lowerJoin.getLeftInputQueue();
		XATQueue lowerRightInputQueue = lowerJoin.getRightInputQueue();
		XATQueue upperLeftInputQueue = upperJoin.getLeftInputQueue();
		XATQueue upperRightInputQueue = upperJoin.getRightInputQueue();
		XATQueue lowerOutputQueue = lowerJoin.getOutputQueue();
		XATQueue upperOutputQueue = upperJoin.getOutputQueue();
		
		
		XATNode lowerLeftChild;
		XATNode lowerRightChild;
		
		if(lowerJoin.getNode().getOpChildrenSize() == 0) {
			lowerLeftChild = null;
			lowerRightChild = null;
		} else if(lowerJoin.getNode().getOpChildrenSize() == 1)  {
			if (lowerJoin.getNthInputQueue(0).equals(lowerJoin.getNode().getOpChild(0).getStreamOperator().getOutputQueue())) {

				lowerLeftChild = lowerJoin.getNode().getOpChild(0);
				lowerRightChild = null;
			}
			else {
				lowerLeftChild = null;
				lowerRightChild = lowerJoin.getNode().getOpChild(0);
			}
		} else if (lowerJoin.getNode().getOpChildrenSize() == 2) {
			lowerLeftChild = lowerJoin.getNode().getOpChild(0);
			lowerRightChild = lowerJoin.getNode().getOpChild(1);
		} else {
			System.out.println("Binary join can only have at most two children. ");
			return false;
		}
		
		/* first take car of queue connnections */
		if(isRightChild) {
			
			//upperOutputQueue.removeOwner(upperJoin);
			//lowerLeftInputQueue.unregister(lowerJoin);
			
			upperRightInputQueue.unregister(upperJoin);

			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerLeftInputQueue, upperRightInputQueue);
			upperJoin.replaceInputQueue(upperRightInputQueue, lowerLeftInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
			lowerOutputQueue.unregister(upperJoin);
		}
		else {
			upperLeftInputQueue.unregister(upperJoin);
			//lowerRightInputQueue.unregister(lowerJoin);
			
			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerRightInputQueue, upperLeftInputQueue);
			upperJoin.replaceInputQueue(upperLeftInputQueue, lowerRightInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
		}

		/** then take care of the parent-child relation of involved XATNodes */
		XATNode upperParentNode = upperJoin.getNode().getParent();
		
		if(isRightChild) {
				lowerJoin.getNode().setParent(upperParentNode);
				lowerJoin.getNode().addOpChild(upperJoin.getNode(), 0);
				lowerJoin.getNode().deleteOpChild(lowerLeftChild);
				
				upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
				upperJoin.getNode().addOpChild(lowerLeftChild);
				upperJoin.getNode().setParent(lowerJoin.getNode());
				
				if(lowerLeftChild != null) {
					lowerLeftChild.setParent(upperJoin.getNode());
				}
				
				// take care of stateID
				upperJoin.getRightState().setStateID(lowerJoin.getLeftState().getStateID());
				maxstateid++;
				lowerJoin.getLeftState().setStateID(maxstateid);				
		}
		else {
			lowerJoin.getNode().setParent(upperParentNode);
			lowerJoin.getNode().addOpChild(upperJoin.getNode());
			lowerJoin.getNode().deleteOpChild(lowerRightChild);
				
			upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
			upperJoin.getNode().addOpChild(lowerRightChild, 0);
			upperJoin.getNode().setParent(lowerJoin.getNode());
			
			if(lowerRightChild != null) {
				lowerRightChild.setParent(upperJoin.getNode());
			}
			
			// take care of stateID
			upperJoin.getLeftState().setStateID(lowerJoin.getRightState().getStateID());
			maxstateid++;
			lowerJoin.getRightState().setStateID(maxstateid);		
				
		}
			
		if (upperParentNode != null) {
			int position = upperParentNode.getIndexOfOpChild(upperJoin.getNode());
			upperParentNode.addOpChild(lowerJoin.getNode(), position);
			upperParentNode.deleteOpChild(upperJoin.getNode());
			upperJoin.getNode().setParent(lowerJoin.getNode());
		} 
		
		if ((upperParentNode == null) || (!upperParentNode.getStreamOperator().isBoxOperator())){
		//	lowerJoin.getNode().setParent(null);
		//	upperJoin.getNode().setParent(lowerJoin.getNode());
			if(newbox != null)
				newbox.setRootOperator(lowerJoin);
		}
		
		// as the last step. reinitialize these two operators, 
		// starting from the new lower op (orig upper op), MUST FOLLOW THIS ORDER!
		upperJoin.init();
		lowerJoin.init();
		return true;
	}
	
	public static boolean SwitchOperators(ChangingWBJoinOperatorImp upperJoin, ChangingWBJoinOperatorImp lowerJoin, int maxstateID, XATBox newbox, DistributedExecutionController ec) 
	{
		// two things to each for each node: parent, children
		// two things to change for each operator: input queue, output queue. 
		// changingWBJoinOperatorImp does not follow the schema restriction.

		boolean isRightChild = true; // indicates if lowerNode is the right child of upperNode.
		//int maxstateid = newbox.getStateCounter();
		
		/** first change the input and output queues of involved operators */
		if(upperJoin.getNthInputQueue(0).equals(lowerJoin.getOutputQueue())) {
			isRightChild = false; 
		}
		System.out.println("lower is upper's right child: "+isRightChild);
			
		XATQueue lowerLeftInputQueue = lowerJoin.getLeftInputQueue();
		XATQueue lowerRightInputQueue = lowerJoin.getRightInputQueue();
		XATQueue upperLeftInputQueue = upperJoin.getLeftInputQueue();
		XATQueue upperRightInputQueue = upperJoin.getRightInputQueue();
		XATQueue lowerOutputQueue = lowerJoin.getOutputQueue();
		XATQueue upperOutputQueue = upperJoin.getOutputQueue();
		
		
		XATNode lowerLeftChild;
		XATNode lowerRightChild;
		
		if(lowerJoin.getNode().getOpChildrenSize() == 0) {
			lowerLeftChild = null;
			lowerRightChild = null;
		} else if(lowerJoin.getNode().getOpChildrenSize() == 1)  {
			if (lowerJoin.getNthInputQueue(0).equals(lowerJoin.getNode().getOpChild(0).getStreamOperator().getOutputQueue())) {

				lowerLeftChild = lowerJoin.getNode().getOpChild(0);
				lowerRightChild = null;
			}
			else {
				lowerLeftChild = null;
				lowerRightChild = lowerJoin.getNode().getOpChild(0);
			}
		} else if (lowerJoin.getNode().getOpChildrenSize() == 2) {
			lowerLeftChild = lowerJoin.getNode().getOpChild(0);
			lowerRightChild = lowerJoin.getNode().getOpChild(1);
		} else {
			System.out.println("Binary join can only have at most two children. ");
			return false;
		}
		
		/* first take car of queue connnections */
		if(isRightChild) {
			
			upperRightInputQueue.unregister(upperJoin);

			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerLeftInputQueue, upperRightInputQueue);
			upperJoin.replaceInputQueue(upperRightInputQueue, lowerLeftInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
			lowerOutputQueue.unregister(upperJoin);
		}
		else {
			upperLeftInputQueue.unregister(upperJoin);
			//lowerRightInputQueue.unregister(lowerJoin);
			
			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerRightInputQueue, upperLeftInputQueue);
			upperJoin.replaceInputQueue(upperLeftInputQueue, lowerRightInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
		}

		/** then take care of the parent-child relation of involved XATNodes */
		XATNode upperParentNode = upperJoin.getNode().getParent();
		
		if(isRightChild) {
				lowerJoin.getNode().setParent(upperParentNode);
				lowerJoin.getNode().addOpChild(upperJoin.getNode(), 0);
				lowerJoin.getNode().deleteOpChild(lowerLeftChild);
				
				upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
				upperJoin.getNode().addOpChild(lowerLeftChild);
				upperJoin.getNode().setParent(lowerJoin.getNode());
				
				if(lowerLeftChild != null) {
					lowerLeftChild.setParent(upperJoin.getNode());
				}
				
				// take care of stateID
				//upperJoin.getRightState().setStateID(lowerJoin.getLeftState().getStateID());
				//maxstateid++;
				//lowerJoin.getLeftState().setStateID(maxstateid);				
		}
		else {
			lowerJoin.getNode().setParent(upperParentNode);
			lowerJoin.getNode().addOpChild(upperJoin.getNode());
			lowerJoin.getNode().deleteOpChild(lowerRightChild);
				
			upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
			upperJoin.getNode().addOpChild(lowerRightChild, 0);
			upperJoin.getNode().setParent(lowerJoin.getNode());
			
			if(lowerRightChild != null) {
				lowerRightChild.setParent(upperJoin.getNode());
			}
			
			// take care of stateID 
			//upperJoin.getLeftState().setStateID(lowerJoin.getRightState().getStateID());
			//maxstateid++;
			//lowerJoin.getRightState().setStateID(maxstateid);		
				
		}
			
		if (upperParentNode != null) {
			int position = upperParentNode.getIndexOfOpChild(upperJoin.getNode());
			upperParentNode.addOpChild(lowerJoin.getNode(), position);
			upperParentNode.deleteOpChild(upperJoin.getNode());
			upperJoin.getNode().setParent(lowerJoin.getNode());
		} 
		
		
		//if ((upperParentNode == null) || (!upperParentNode.getStreamOperator().isBoxOperator())){
		//	if(newbox != null)
		//		newbox.setRootOperator(lowerJoin);
		//}
		
		//Now recompute the unmatched state
		if(isRightChild) {
		    System.out.println("Is right child.");
		    System.out.println("old upper left state size = " + upperJoin.getLeftState().getStateSize());
		    System.out.println("old upper right state size = " + upperJoin.getRightState().getStateSize());
		    System.out.println("old lower left state size = " + lowerJoin.getLeftState().getStateSize());
		    System.out.println("old lower right state size = " + lowerJoin.getRightState().getStateSize());
		    
			OperatorState lowerLeftState = lowerJoin.getLeftState();
			upperJoin.getRightState().clearState();
			upperJoin.setRightState(lowerLeftState);
			
			
			WFIFOState newstate = new WFIFOState();
			ArrayList newstatetuples = upperJoin.joinFIFOStates();
			newstate.insertTuple(newstatetuples);
			newstate.insertTuple(newstatetuples);
			//newstate.insertTuple(newstatetuples);
			lowerJoin.setLeftState(newstate);
			
		    System.out.println("new upper left state size = " + upperJoin.getLeftState().getStateSize());
		    System.out.println("new upper right state size = " + upperJoin.getRightState().getStateSize());
		    System.out.println("new lower left state size = " + lowerJoin.getLeftState().getStateSize());
		    System.out.println("new lower right state size = " + lowerJoin.getRightState().getStateSize());
		
			
		} else {
		    System.out.println("Is left child.");
		    System.out.println("old upper left state size = " + upperJoin.getLeftState().getStateSize());
		    System.out.println("old upper right state size = " + upperJoin.getRightState().getStateSize());
		    System.out.println("old lower left state size = " + lowerJoin.getLeftState().getStateSize());
		    System.out.println("old lower right state size = " + lowerJoin.getRightState().getStateSize());
		    
			OperatorState lowerRightState = lowerJoin.getRightState();
			upperJoin.getLeftState().clearState();
			upperJoin.setLeftState(lowerRightState);
			
			WFIFOState newstate = new WFIFOState();
			ArrayList newstatetuples = upperJoin.joinFIFOStates();
			newstate.insertTuple(newstatetuples);
			newstate.insertTuple(newstatetuples);
			//newstate.insertTuple(newstatetuples);
			lowerJoin.setRightState(newstate);
			
		    System.out.println("new upper left state size = " + upperJoin.getLeftState().getStateSize());
		    System.out.println("new upper right state size = " + upperJoin.getRightState().getStateSize());
		    System.out.println("new lower left state size = " + lowerJoin.getLeftState().getStateSize());
		    System.out.println("new lower right state size = " + lowerJoin.getRightState().getStateSize());
	
		}
		
		if(upperJoin.isRootOperator()) {
			upperJoin.setAsRootOperator(false);
			lowerJoin.setAsRootOperator(true);
		}

		// as the last step. reinitialize these two operators, 
		// starting from the new lower op (orig upper op), MUST FOLLOW THIS ORDER!
		//upperJoin.init();
		//lowerJoin.init();
		return true;
	}
	
	/**
	 * This method assumes that the root operator is the UnionOperaotr, and the leaf operators are all split operators.
	 * The joins are in the middle of the query plan.
	 * It also assume that both upperjoin and lowerjoin are binary joins (with two inputs).
	 * @param upperJoin
	 * @param lowerJoin
	 * @param maxstateID
	 * @param newbox
	 * @return
	 */
	public static boolean SwitchOperators(WTransitiveHashJoinOperator upperJoin, WTransitiveHashJoinOperator lowerJoin, int maxstateID, XATBox newbox, DistributedExecutionController ec) 
	{

		boolean isRightChild = true; // indicates if lowerNode is the right child of upperNode.
        ConnectionTable intable = ec.getInConnectionTable();
        ConnectionTable outtable = ec.getOutConnectionTable();
        TupleDistributor distributor = ec.getTupleDistributor();
        TupleFeeder feeder = ec.getTupleFeeder();
        
        System.out.println("Connection tables before switch.");
        System.out.println("in-connections: \n" + intable);
        System.out.println("out-connections: \n" + outtable);
		
		/** first change the input and output queues of involved operators */
		if(upperJoin.getNthInputQueue(0).equals(lowerJoin.getOutputQueue())) {
			isRightChild = false; 
			System.out.println("lower is upper's left child.");
		} else {
		    System.out.println("lower is upper's right child");
		}
			
		XATQueue lowerLeftInputQueue = lowerJoin.getLeftInputQueue();
		XATQueue lowerRightInputQueue = lowerJoin.getRightInputQueue();
		XATQueue upperLeftInputQueue = upperJoin.getLeftInputQueue();
		XATQueue upperRightInputQueue = upperJoin.getRightInputQueue();
		XATQueue lowerOutputQueue = lowerJoin.getOutputQueue();
		XATQueue upperOutputQueue = upperJoin.getOutputQueue();
		
		
		XATNode lowerLeftChild;
		XATNode lowerRightChild;
		
		if(lowerJoin.getNode().getOpChildrenSize() == 0) {
			lowerLeftChild = null;
			lowerRightChild = null;
		} else if(lowerJoin.getNode().getOpChildrenSize() == 1)  {
			if (lowerJoin.getNthInputQueue(0).equals(lowerJoin.getNode().getOpChild(0).getStreamOperator().getOutputQueue())) {

				lowerLeftChild = lowerJoin.getNode().getOpChild(0);
				lowerRightChild = null;
			}
			else {
				lowerLeftChild = null;
				lowerRightChild = lowerJoin.getNode().getOpChild(0);
			}
		} else if (lowerJoin.getNode().getOpChildrenSize() == 2) {
			lowerLeftChild = lowerJoin.getNode().getOpChild(0);
			lowerRightChild = lowerJoin.getNode().getOpChild(1);
		} else {
			System.out.println("Binary join can only have at most two children. ");
			return false;
		}
		
		/* first take care of queue connnections */
		if(isRightChild) {
			
			upperRightInputQueue.unregister(upperJoin);

			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerLeftInputQueue, upperRightInputQueue);
			upperJoin.replaceInputQueue(upperRightInputQueue, lowerLeftInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
			lowerOutputQueue.unregister(upperJoin);
		}
		else {
			upperLeftInputQueue.unregister(upperJoin);
			//lowerRightInputQueue.unregister(lowerJoin);
			
			lowerJoin.setOutputQueue(upperOutputQueue);
			lowerJoin.replaceInputQueue(lowerRightInputQueue, upperLeftInputQueue);
			upperJoin.replaceInputQueue(upperLeftInputQueue, lowerRightInputQueue);
			upperJoin.setOutputQueue(lowerOutputQueue);
		}

        /* now update outconnection tables */
        Object ob = outtable.getConnection(upperJoin);
        if(ob != null) {
            if(ob instanceof ObjectOutputStream) {
                ObjectOutputStream os = (ObjectOutputStream)ob;
                
                ReplaceInConnectionEntry entry = new ReplaceInConnectionEntry(upperJoin.getProperty("OBJECT_ID"), lowerJoin.getProperty("OBJECT_ID"));
                Packet p = new Packet(upperJoin.getProperty("OBJECT_ID"), entry); 
                NetworkPacket np = new NetworkPacket(1);
                np.addPacket(p);
                boolean sent = false;
                while(!sent) {
                    try{
                        os.reset();
                        os.writeObject(np);
                        sent = true;
                        System.out.println("Send replace-inconenction-entry packet.");
                    } catch (Exception e) {
                        System.out.println("Out stream is not ready.");
                        try{
                            System.out.println("Wait for socket to be available.");
                            Thread.sleep(100);
                        }catch(Exception se) {
                            se.printStackTrace();
                        }
                    }
                }                
                outtable.removeConnection(upperJoin);
                outtable.addConnection(lowerJoin, os);
                
            } else if(ob instanceof Hashtable) {
                Hashtable ht = (Hashtable)ob;
                outtable.removeConnection(upperJoin);
                outtable.addConnection(lowerJoin, ht);
            }
        }
        
		/** then take care of the parent-child relation of involved XATNodes */
		XATNode upperParentNode = upperJoin.getNode().getParent();
		
		if(isRightChild) {
				lowerJoin.getNode().setParent(upperParentNode);
				lowerJoin.getNode().addOpChild(upperJoin.getNode(), 0);
				lowerJoin.getNode().deleteOpChild(lowerLeftChild);
				
				upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
				upperJoin.getNode().addOpChild(lowerLeftChild);
				upperJoin.getNode().setParent(lowerJoin.getNode());
				
				if(lowerLeftChild != null) {
					lowerLeftChild.setParent(upperJoin.getNode());
                    
                    //update inconnection table
                    String opid = lowerLeftChild.getOperator().getProperty("OBJECT_ID");
                    OperatorQueue oq = intable.getConnection(opid);
                    System.out.println("old operator queue: " + oq);
                    if(oq != null) {
                        oq.setOperator(upperJoin);
                        oq.setQueueNum(1);
                    }
                    System.out.println("new operator queue: " + oq);
				}
				
				// take care of stateID
				//upperJoin.getRightState().setStateID(lowerJoin.getLeftState().getStateID());
				//maxstateid++;
				//lowerJoin.getLeftState().setStateID(maxstateid);				
		}
		else {
			lowerJoin.getNode().setParent(upperParentNode);
			lowerJoin.getNode().addOpChild(upperJoin.getNode());
			lowerJoin.getNode().deleteOpChild(lowerRightChild);
				
			upperJoin.getNode().deleteOpChild(lowerJoin.getNode());
			upperJoin.getNode().addOpChild(lowerRightChild, 0);
			upperJoin.getNode().setParent(lowerJoin.getNode());
			
			if(lowerRightChild != null) {
				lowerRightChild.setParent(upperJoin.getNode());
                
                //update inconnection table
                String opid = lowerRightChild.getOperator().getProperty("OBJECT_ID");
                OperatorQueue oq = intable.getConnection(opid);
                System.out.println("old operator queue: " + oq);
                if(oq != null) {
                    oq.setOperator(upperJoin);
                    oq.setQueueNum(0);
                }
                System.out.println("new operator queue: " + oq);
			}
			
			// take care of stateID 
			//upperJoin.getLeftState().setStateID(lowerJoin.getRightState().getStateID());
			//maxstateid++;
			//lowerJoin.getRightState().setStateID(maxstateid);		
		}
			
		if (upperParentNode != null) {
			int position = upperParentNode.getIndexOfOpChild(upperJoin.getNode());
			upperParentNode.addOpChild(lowerJoin.getNode(), position);
			upperParentNode.deleteOpChild(upperJoin.getNode());
			upperJoin.getNode().setParent(lowerJoin.getNode());
		} 
		
		
		//if ((upperParentNode == null) || (!upperParentNode.getStreamOperator().isBoxOperator())){
		//	if(newbox != null)
		//		newbox.setRootOperator(lowerJoin);
		//}
		
		//Now reuse matched state and recompute unmatched state
		if(isRightChild) {
		    //System.out.println("Is right child.");
		    Hashtable upperstates = upperJoin.getHashJoinStates();
		    Hashtable lowerstates = lowerJoin.getHashJoinStates();
		    ArrayList IDs = new ArrayList();
		    
		    //System.out.println("old upper join state size = " + upperstates.values().size());
		    //System.out.println("old lower join state size = " + lowerstates.values().size());
		    
		    //get the larger set of partition IDs
		    Enumeration e;
		    if(upperstates.size() >= lowerstates.size()) {
		        e = upperstates.keys();
		    } else {
		        e = lowerstates.keys();
		    }
		    while(e.hasMoreElements()) {
		        Integer it = (Integer)e.nextElement();
		        IDs.add(it);
		    }
		    
		    //move each left state of lower join to right state of upperjoin
		    for(int index=0; index<IDs.size(); index++) {
		        Integer id = (Integer)IDs.get(index);
		        HashStatePartitionGroup hpg = (HashStatePartitionGroup)lowerstates.get(id);
		        if(hpg != null) {
		            WEquiJoinStateImp lowerleftstate = ((HashStatePartitionGroup)lowerstates.get(id)).getHashStates()[0];
		            HashStatePartitionGroup upperjoingroup = (HashStatePartitionGroup)upperstates.get(id);
		            upperjoingroup.setStatePartition(1, lowerleftstate);
		            upperstates.put(id, upperjoingroup);
		            //System.out.println("Copy part: " + id + " pos: 0 from lowerjoin to upperjoin");
		        }
		    }
		    
		    //recompute unmatched state: compute left state of lowerjoin by joining partitions in upperjoin
		    ArrayList newstates = upperJoin.joinPartitionsByIDs(IDs);
		    //System.out.println("\nnew states are:" + newstates);
		    
		    for(int idindex=0; idindex<IDs.size(); idindex++) {
		        WEquiJoinStateImp newstate = (WEquiJoinStateImp)newstates.get(idindex);
		        lowerJoin.setSinglePartitionByID(((Integer)IDs.get(idindex)).intValue(), 0, newstate);
		    }
		    
		    //System.out.println("new upper join state size = " + upperstates.values().size());
		    //System.out.println("new lower join state size = " + lowerstates.values().size());
			
		} else {
		    //System.out.println("Is left child.");
		    Hashtable upperstates = upperJoin.getHashJoinStates();
		    Hashtable lowerstates = lowerJoin.getHashJoinStates();
		    ArrayList IDs = new ArrayList();
		    
		    //System.out.println("old upper join state size = " + upperstates.values().size());
		    //System.out.println("old lower join state size = " + lowerstates.values().size());
		    
		    //get the larger set of partition IDs
		    Enumeration e;
		    if(upperstates.size() >= lowerstates.size()) {
		        e = upperstates.keys();
		    } else {
		        e = lowerstates.keys();
		    }
		    while(e.hasMoreElements()) {
		        Integer it = (Integer)e.nextElement();
		        IDs.add(it);
		    }
		    
		    //move each right state of lower join to left state of upperjoin
		    for(int index=0; index<IDs.size(); index++) {
		        Integer id = (Integer)IDs.get(index);
		        HashStatePartitionGroup hpg = (HashStatePartitionGroup)lowerstates.get(id);
		        if(hpg != null) {
		            WEquiJoinStateImp lowerrightstate = hpg.getHashStates()[1];
		            HashStatePartitionGroup upperjoingroup = (HashStatePartitionGroup)upperstates.get(id);
                    //System.out.println("upperjoingroup = " + upperjoingroup);
                    //System.out.println("lowerrightstate = " + lowerrightstate);
                    if(upperjoingroup != null) {
                        upperjoingroup.setStatePartition(0, lowerrightstate);
                        upperstates.put(id, upperjoingroup);
                    }
		            //System.out.println("Copy part: " + id + " pos: 1 from lowerjoin to upperjoin");
		        } else {
		            //System.out.println("Lower join does not have partition group: " + id + " Just skip it.");
		            //System.out.println("The content of this partition in upper join is: ");
		            //System.out.println(upperstates.get(id));
		        }
		    }
		    
		    //recompute unmatched state: compute right state of lowerjoin by joining partitions in upperjoin
		    ArrayList newstates = upperJoin.joinPartitionsByIDs(IDs);
		    //System.out.println("\nnew states are:" + newstates);
		    for(int idindex=0; idindex<IDs.size(); idindex++) {
		        WEquiJoinStateImp newstate = (WEquiJoinStateImp)newstates.get(idindex);
		        lowerJoin.setSinglePartitionByID(((Integer)IDs.get(idindex)).intValue(), 1, newstate);
		    }
		    
		    //System.out.println("new upper join state size = " + upperstates.values().size());
		    //System.out.println("new lower join state size = " + lowerstates.values().size());
		}

		// as the last step. reinitialize these two operators, 
		// starting from the new lower op (orig upper op), MUST FOLLOW THIS ORDER!

		upperJoin.init();
		lowerJoin.init();
		
		String lname = upperJoin.getInputQueue(0).getSchema().getName();
		String rname = upperJoin.getInputQueue(1).getSchema().getName();
		String newname = "#" + lname + "#" + rname;
		XATQueueSchema outSchema = upperJoin.getOutputQueue().getSchema();
		outSchema.setName(newname);
		
//		System.out.println("Input queue schema for operator: " + upperJoin);
//		System.out.print("{");
//		for(int i=0; i<upperJoin.getInputQueueCount(); i++) {
//		    System.out.print(upperJoin.getInputQueue(i).getSchema().getName() + ", ");
//		}
//		System.out.println("}");
//		
//		System.out.println("Input queue schema for operator: " + lowerJoin);
//		System.out.print("{");
//		for(int i=0; i<lowerJoin.getInputQueueCount(); i++) {
//		    System.out.print(lowerJoin.getInputQueue(i).getSchema().getName() + ", ");
//		}
//		System.out.println("}");		
        

        
        System.out.println("Connection tables after switch.");
        System.out.println("in-connections: \n" + intable);
        System.out.println("out-connections: \n" + outtable);
        
		return true;
	}
	
}
		
		
