/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.classes.SourceImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel.BinaryJoin.SymmetricNLJWithNoTouchCost;

/**
 *Given a Join Graph with n Streams S1, S2, ..., Sn 
 * This algorithm is used to generate the optimal join ordering represented by bushy trees 
 * 
 *
 * @author Venkatesh Raghavan 
 * @email venky@cs.wpi.edu
 * @version 1.0, Summer 2007
 * 
 */
public class BJTreeDynamicProgramming {

/**
  *Boolean variable which can be set to true if we want to print all the plans generated in each iteration and their costs
  *
  * @author Venkatesh Raghavan 
  * @email venky@cs.wpi.edu
  * @version 1.0, Summer 2007
  * 
  */		

	public static boolean printOption = false;

/**
  * For example, if the left source is A and right source is E
  * The join graph had no edge between A and E. so it will return false
  * 
  * @param _leftSources, _rightSources
  * @param _graph
  * @return Boolean value
  */	
	
	
	public static boolean hasEdgeInJoinGraph(Vector<Source> _leftSources, Vector<Source> _rightSources, JoinGraph _graph)
	{
		for(int j=0;j<_leftSources.size();j++)
		{
			if(Utils.hasEdge(_graph, _rightSources, _leftSources.elementAt(j)))
			{
				return true;
			}
		}
		return false;
	}

/**
  
  *	For example, if the left source is A and the right plan is (AB), A is an element of the right plan.
  *	So the function will return true
  *	@param _leftSources
  *	@param _rightPlan
  *	@return boolean value	
  */ 	
	
	
	public static boolean haveSimilarSource(Vector<Source> _leftSources, Plan _rightPlan)
	{
		for(int j=0;j<_leftSources.size();j++)
		{
			if(_rightPlan.visited((SourceImpl)_leftSources.elementAt(j)))
			{
				return true;
			}
		}
		return false;
	}

/** 
  * 
  * Prints all the plans of a particular size along with their costs present in the corresponding hash table
  * @param _index
  * @return all plans of that size with their costs
  */
	
	public static void printHashtable(int _index)
	{
		System.out.println("######## HASH TABLE ### "+_index+" ######");
		if(tables[_index] !=null)
		{
			Enumeration planEnum = tables[_index].elements();
			Plan nextPlan;
			while(planEnum.hasMoreElements())
			{
				nextPlan = (Plan) planEnum.nextElement();
				nextPlan.printPretty();
				System.out.println(" With CPU Cost: "+nextPlan.getCPUCost());
			}
		}
		System.out.println("##############################");
	}
	public static Hashtable tables[];

/**
  * This algorithms starts bottom-up.
  *  
  * @param _index
  * @param _graph
  * @return
  */

	public static void iterate(JoinGraph _graph, int _index)
	{

		Enumeration rightEnum, leftEnum;
		Plan leftPlan, rightPlan, contestPlan, finalPlan;
		QueryPlan newPlan;
		BinaryJoinNodeImpl joinNode;
		Vector<Source> leftSources;
		String newPlanName;
		if(_index == _graph.getVerticesCount())
		{	
			return;
		}
		else
		{
			if (printOption)
			{
				System.out.println("Now consider on join pair of size "+_index);
				System.out.println("===================================");
			}
			/* declare a different hash table for plans of different sizes */
			tables[_index] = new Hashtable();

			int lhsIndex, rhsIndex;
			//double newIndex = new Double(_index)/2;
			//for(int i = 0; i <= Math.floor(_index/2);i++)
			//System.out.println(Math.ceil(newIndex/2));

			for(int i = 0; i <= (Math.ceil(new Double(_index)/2)-1);i++)
			{
				lhsIndex = i; 
				// Note size = 0 means all singleton sources A, B, C.
				// And size = 1 means pairs of size 2 i.e. AB, AC, AD, BC, etc.
				rhsIndex = _index -1  - i;



				leftEnum = tables[lhsIndex].elements();

				while(leftEnum.hasMoreElements())
				{
					if (printOption)
					{
						System.out.println("-----------------------------------");
						System.out.print("Left Plan is");
					}
					leftPlan = (Plan) leftEnum.nextElement();
					if (printOption)
					{	
						leftPlan.printPretty();
						System.out.println("");
					}
					leftSources = leftPlan.getSources();
					rightEnum = tables[rhsIndex].elements();
					while(rightEnum.hasMoreElements())
					{
						if (printOption)
						{
							System.out.println(" - - - - - - - - - - - - - - -");
						}
						rightPlan = (Plan) rightEnum.nextElement();
						if (printOption)
						{
							System.out.print(" Right Plan is");
							rightPlan.printPretty();
							System.out.println("");
						}

						// Now check if the left sources are in the right. If so do not consider the combination.
						if(!BJTreeDynamicProgramming.haveSimilarSource(leftSources, rightPlan))
						{
							if(BJTreeDynamicProgramming.hasEdgeInJoinGraph(leftSources, rightPlan.getSources(), _graph))
							{
								if (printOption)
								{
									System.out.print("  Has Edge");
								}
								joinNode = new BinaryJoinNodeImpl();
								joinNode.addLeftChild(leftPlan);
								joinNode.addRightChild(rightPlan);
								newPlan = new QueryPlan(joinNode); 
								newPlanName = Utils.sortedName(newPlan);

								SymmetricNLJWithNoTouchCost.evaluateCost(newPlan,_graph); 

								if (printOption)
								{
									System.out.print("  New Plan is");
									newPlan.printPretty();
									System.out.println(" With Cost :"+newPlan.getCPUCost());
								}

								if(tables[_index].get(newPlanName) != null)
								{

									contestPlan = (Plan) tables[_index].get(newPlanName);
									// I am only considering CPU as limiting cost
									if (printOption)
									{
										System.out.print("  A Plan Exist : ");
										contestPlan.printPretty();
										System.out.println(" With Cost: "+ contestPlan.getCPUCost());
									}

									if(contestPlan.getCPUCost() > newPlan.getCPUCost())
									{
										if (printOption)
										{
											System.out.println("    New Plan Cheaper Therfore Use That Instead");
										}
										tables[_index].remove(newPlan);
										tables[_index].put(newPlanName, newPlan);
									}else
									{
										if (printOption)
										{
											System.out.println("    New Plan >= In Expensive");
										}
									}
								}else
								{
									if (printOption)
									{
										System.out.println("   Plan Does Not Exist, Therefore insert into hash table");
									}
									tables[_index].put(newPlanName,newPlan);

								}
							}else
							{
								if (printOption)
								{
									System.out.println("   There is no edge from any source on LHS to any source in RHS. Therefore, Ignore");
								}
							}
						}else
						{
							if (printOption)
							{
								System.out.println("  Ignore Since They Have Similar Sources");
							}
						}
						if (printOption)
						{
							System.out.println(" - - - - - - - - - - - - - - -");
						}
					}
					if (printOption)
					{
						System.out.println("-----------------------------------");
					}
				}
			}
		}
		if (printOption)
		{
			System.out.println("===================================");
			BJTreeDynamicProgramming.printHashtable(_index);
		}
		BJTreeDynamicProgramming.iterate(_graph, _index+1);
	}

/**
  * For a given join graph JG with streams S1, S2, ... Sn
  * Generate the optimal join order of for processing tuples from Si.
  * @param _graph
  * @return all plans of size 1 
  */

	public static QueryPlan order(JoinGraph _graph)
	{

		int numberOfSource = _graph.getVerticesCount();
		tables = new Hashtable[numberOfSource];
		int index =0;
		
		tables[index] = new Hashtable();
		Enumeration sourceEnum = _graph.getVertices();
		QueryPlan nextPlan;
		while(sourceEnum.hasMoreElements())
		{
			nextPlan = new QueryPlan((Source) sourceEnum.nextElement());
			SymmetricNLJWithNoTouchCost.evaluateCost(nextPlan,_graph);
			tables[index].put(Utils.sortedName(nextPlan),nextPlan);
		}	
		
		if (printOption)
		{
		BJTreeDynamicProgramming.printHashtable(index);
		}
		BJTreeDynamicProgramming.iterate(_graph, index+1);
		
		if(tables[numberOfSource-1] != null)
		{
			Enumeration planEnum = tables[_graph.getVerticesCount()-1].elements();
			while(planEnum.hasMoreElements())
			{
				return (QueryPlan) planEnum.nextElement();
			}
		}
		
		return null;
	}
}
