/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms;

import java.io.BufferedWriter;
import java.util.Enumeration;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.experiments.Experiment_1;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Edge;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Vertex;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.CombinedRankedNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.RankedRootedTreeNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.RankedRootedTreeNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.RootedTreeNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.classes.SourceImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.utils.Util;

public class IKAlgorithm {
	
	public static boolean isChain(RootedTreeNode _current){
		if(_current.chidrenCount() > 1)
		{
			return false;
		}else if(_current.chidrenCount() == 0)
		{
			return true; 
		}else
		{
			return IKAlgorithm.isChain((RootedTreeNode) _current.getChild(0));
		}
	}
	
	public static boolean isSortedChain(RankedRootedTreeNode _current)
	{
		RankedRootedTreeNode child;
		if(IKAlgorithm.isChain(_current))
		{
			if(_current.chidrenCount() ==0)
			{
				return true;
			}else
			{
				child = (RankedRootedTreeNode) _current.getChild(0);
				if(_current.getRank() < child.getRank())
				{
					return IKAlgorithm.isSortedChain(child);
				}else
				{
					return false;
				}
				
			}
		}else
		{
			//TODO: FISHY MAY BE HAVE AN EXCEPTION
			return false;
		}
	}
	public static int cheapestRankedNode(Vector<RootedTreeNode> _set)
	{
		double cheapestRank = Double.MAX_VALUE;
		int index =0;
		for(int i=0; i<_set.size();i++)
		{
			if(((RankedRootedTreeNode)_set.get(i)).getRank() < cheapestRank)
			{
				index =i;
				cheapestRank = ((RankedRootedTreeNode)_set.get(i)).getRank();
			}
		}

		return index;
	}
	
	public static RankedRootedTreeNode orderWedge(Vector<RootedTreeNode> _set,RankedRootedTreeNode _parent)
	{
		RankedRootedTreeNode nextCheapest;
		int index;
		if(_set.size() ==1)
		{
			_set.get(0).removeAllChildren();
			_set.get(0).setParentNode(_parent);
			_parent.removeAllChildren();
			_parent.addChild(_set.get(0));
			return _parent;
		}else
		{
			index = IKAlgorithm.cheapestRankedNode(_set);
			nextCheapest = (RankedRootedTreeNode)_set.get(index);
			_set.remove(index);
			nextCheapest = IKAlgorithm.orderWedge(_set,nextCheapest);
			nextCheapest.setParentNode(_parent);
			_parent.removeAllChildren();
			_parent.addChild(nextCheapest);
			return _parent;
		}
	}
	
	public static RankedRootedTreeNode orderChain(RankedRootedTreeNode _current)
	{
		RankedRootedTreeNode child;
		if(!IKAlgorithm.isSortedChain(_current))
		{
			if(_current.chidrenCount() ==0)
			{
				_current.removeAllChildren();
				return _current;
			}else
			{
				child = (RankedRootedTreeNode)_current.getChild(0);
				if(child.getRank() < _current.getRank())
				{
					CombinedRankedNodeImpl combinedNode = new CombinedRankedNodeImpl(_current,child);
					if(child.getChild(0) != null)
					{
						combinedNode.addChild(child.getChild(0));
					}
					child.removeAllChildren();
					if(!IKAlgorithm.isSortedChain(combinedNode))
					{
						return IKAlgorithm.orderChain(combinedNode);
					}else return combinedNode;
					
				}else
				{
					child = IKAlgorithm.orderChain(child);
					_current.removeAllChildren();
					_current.addChild(child);
					if(!IKAlgorithm.isSortedChain(_current))
					{
						return IKAlgorithm.orderChain(_current);
					}else return _current;
				}
			}
		}else return _current;
	}
	public static RankedRootedTreeNode unMergeChain(RankedRootedTreeNode _current)
	{
		if(IKAlgorithm.isChain(_current))
		{
			if(_current instanceof CombinedRankedNodeImpl)
			{
				RankedRootedTreeNode ancestor = (RankedRootedTreeNode)_current.getParent();
				RankedRootedTreeNode head = ((CombinedRankedNodeImpl)_current).getHead();
				RankedRootedTreeNode tail = ((CombinedRankedNodeImpl)_current).getTail();
				if(ancestor !=null)
				{
					ancestor.removeAllChildren();
					ancestor.addChild(head);
				}
				if(tail !=null)
				{
					if(_current.getChild(0) !=null)
					{
						tail.addChild(_current.getChild(0));
					}
				}
				return IKAlgorithm.unMergeChain(head);
			}else
			{
				if(_current.chidrenCount() >0)
				{
					RankedRootedTreeNode child = IKAlgorithm.unMergeChain((RankedRootedTreeNode)_current.getChild(0));
					_current.removeAllChildren();
					_current.addChild(child);
					child.setParentNode(_current);
					return _current;
				}else
				{
					return _current;
				}
			}
		}else
		{
			return null;
		}
	}
	public static RankedRootedTreeNode order(RankedRootedTreeNode _current)
	{
		if(IKAlgorithm.isChain((RootedTreeNode)_current))
		{
			return IKAlgorithm.orderChain(_current);
		}
		
		if(_current.chidrenCount() ==0)
		{
			return _current;
		}
		
		Vector<RootedTreeNode> children = new Vector<RootedTreeNode>();
		for(int i=0; i <_current.chidrenCount();i++)
		{
			if(IKAlgorithm.isChain((RootedTreeNode)_current.getChild(i)))
			{
				children.add(IKAlgorithm.order((RankedRootedTreeNode)_current.getChild(i)));
			}else
			{
				children.add(IKAlgorithm.order((RankedRootedTreeNode) _current.getChild(i)));
			}
		}
		if(children.size() ==1)
		{
			_current.removeAllChildren();
			_current.addChild(children.get(0));
			children.get(0).setParentNode(_current);
		}else
		{
			Vector<RootedTreeNode> set = new Vector<RootedTreeNode>();
			set.addAll(children);
			for(int k=0; k<children.size();k++)
			{
				set.addAll(children.get(k).getDescendants());
			}
			_current = IKAlgorithm.orderWedge(set, _current);
		}
		return IKAlgorithm.order(_current);
	}
	
	public static RankedRootedTreeNode generateRootedTree(JoinGraph _acyclicJoinGraph, Vertex _start, RankedRootedTreeNode _current, RankedRootedTreeNode _parent)
	{
		RankedRootedTreeNodeImpl newNode;
		
		double rank, inputRate, windowSize, selectivity;
		double cValue=0, tValue=0;
		try
		{
			//TODO: Fishy design
			Source currentSource = (Source) _current.getVertex();
			
			if(_parent == null)
			{
				cValue =0;
				tValue =currentSource.getValue();
				rank =0;
			}else
			{
				if(currentSource != null)
				{
					inputRate = currentSource.getCardinality();
					windowSize = currentSource.getWindowSize();
					selectivity = _acyclicJoinGraph.getEdge(_parent,_current).getValue();
					
					cValue = inputRate * selectivity * windowSize;
					tValue = cValue;
					
					rank = (tValue - 1)/cValue;
					
				}else
				{
					rank = 0;
				}
			}
			_current.setCValue(cValue);
			_current.setTValue(tValue);
			_current.setRank(rank);

			Vector<Vertex> children;
			
			children = _acyclicJoinGraph.getNeighbors(_current.getVertex());
			Vertex nextVertex;
			for(int i=0; i<children.size();i++)
			{
				nextVertex = children.elementAt(i);
				if(_parent == null || nextVertex.getLabel().compareTo(_parent.getLabel())!=0 )
				{
					newNode = new RankedRootedTreeNodeImpl(nextVertex);
					newNode.setParentNode(_current);
					newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.generateRootedTree(_acyclicJoinGraph,_start,newNode,_current);
					_current.addChild(newNode);
				}
			}
			return _current;
			
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		return null;
	}

	

	public static RankedRootedTreeNode generateOptimalOrdering(JoinGraph _acyclicJoinGraph, Source _startStream)
	{
		
		RankedRootedTreeNodeImpl newNode = new RankedRootedTreeNodeImpl(_startStream);
		// Generate Rooted Tree Node.
		newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.generateRootedTree(_acyclicJoinGraph,_startStream,newNode,null);
		

		
		newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.order(newNode);
		

		
		newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.unMergeChain(newNode);
		
		
		return newNode;
	}
	
	public static RankedRootedTreeNode generateOptimalMJoinOrdering(JoinGraph _graph, Source _startStream, BufferedWriter _out)
	{
		JoinGraph  acyclicJoinGraph= MinimumSpanningTree.kruskalAlgorithm(_graph);
		Experiment_1.printSelectivity(acyclicJoinGraph, _out);

		return IKAlgorithm.generateOptimalOrdering(acyclicJoinGraph,_startStream);
		
		/*
		Experiment_1.printSelectivity(acyclicJoinGraph, _out);
		
		RankedRootedTreeNodeImpl newNode = new RankedRootedTreeNodeImpl(_startStream);
		// Generate Rooted Tree Node.
		newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.generateRootedTree(acyclicJoinGraph,_startStream,newNode,null);
		
		//System.out.println("****************");
		newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.order(newNode);
		
		//System.out.println("----------------");
		//System.out.println("UnMerged OUTPUT");
		//System.out.println("----------------");
		//print(newNode);
		//System.out.println("****************");
		
		newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.unMergeChain(newNode);
		
		//System.out.println("----------------");
		//System.out.println(" Final OUTPUT");
		//System.out.println("----------------");
		//print(newNode);
		//System.out.println("****************");
		
		return newNode;
		*/
	}

	
	public static Vector<RankedRootedTreeNode> generateOptimalMJoin(JoinGraph _graph)
	{
		Vector<RankedRootedTreeNode> result = new Vector<RankedRootedTreeNode>();
		JoinGraph acyclicJoinGraph = MinimumSpanningTree.kruskalAlgorithm(_graph);
		Enumeration enums = acyclicJoinGraph.getVertices();
		
		Source startStream;
		RankedRootedTreeNodeImpl newNode;
		while(enums.hasMoreElements())
		{
			startStream = (Source) enums.nextElement();
			newNode = new RankedRootedTreeNodeImpl(startStream);
			newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.generateRootedTree(acyclicJoinGraph,startStream,newNode,null);
			newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.order(newNode);	
			newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.unMergeChain(newNode);
			result.add(newNode);
		}
		return result;
	}
	
}
