/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.classes.SourceImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel.BinaryJoin.SymmetricNLJWithNoTouchCost;


public class JTreeDynamicProgramming {
	public static Hashtable[] subPlans;

	public static void order(JoinGraph _graph)
	{

		int numberOfSource = _graph.getVerticesCount();
		subPlans = new Hashtable[numberOfSource];

		int index =0;

		subPlans[index] = new Hashtable<String,Plan>();
		Enumeration sourceEnum = _graph.getVertices();
		QueryPlan nextPlan;

		while(sourceEnum.hasMoreElements())
		{
			nextPlan = new QueryPlan((Source) sourceEnum.nextElement());
			subPlans[index].put(Utils.sortedName(nextPlan),nextPlan);
		}	

	}	

	public static void recurssive(int index, JoinGraph graph )
	{
		// 0. Generate partitions by bell number methodology.
		Hashtable PT = Utils.generateBellPartitions(index+1); 
		
		// 1. Generate all unique subpairs of size = index
		
		// 2. Call iterate function.
		// 3.0 if(index= graph.getVeriticesCount()-1 then exit
		// 3.1 else increment index and call itself.
	}
	
	public static void generateSubPlans(Hashtable[] subPlans, int index, JoinGraph _graph)
	{
		Enumeration sourceEnum = _graph.getVertices();
		Enumeration subsetEnum = subPlans[index - 1].elements();
		Vector<Source> nextSubSet;
		while (subsetEnum.hasMoreElements())
		{
			nextSubSet = (Vector<Source>) subsetEnum.nextElement();
			for (int i=0; i<subPlans[index-1].size(); i++)
			{
				
			}
		}
	}
	
	public static void iterate(ArrayList<Integer> PT, Vector<Source> subset, int index, JoinGraph _graph, ArrayList<Plan> mapping)
	{
		// 0. Get from current 
		Vector<Source> currentVisited = null;
		for(int i=0; i <index ;i++)
		{
			currentVisited.add((Source)mapping.get(i));
		}

		// 1. Partition Subset
		int planSize = PT.get(index);
		Vector<Source> notVisitedSources = new Vector<Source>();
		notVisitedSources = JTreeDynamicProgramming.removeVisitedNodes(subset, currentVisited);

		// 2. Remove all visited nodes
		JTreeDynamicProgramming.removeVisitedNodes(subset, currentVisited);

		Vector<Plan> finalSubPlan;
		// 3. Now get all plans of size "planSize" of "nonVisitedSources"
		finalSubPlan = JTreeDynamicProgramming.getSubPlans(planSize, subPlans, notVisitedSources);

		Enumeration<Plan> plans = finalSubPlan.elements();

		index=0;
		Plan nextPlan;
		while (plans.hasMoreElements())
		{
			nextPlan = plans.nextElement();
			mapping.set(index, nextPlan);

			if(index == subset.size()-1)
			{
				// 1. Convert mapping into an actual query plan
				// Note: # of children equals the partition size.
				// 1.1. If # of children == 2 then BinaryNode i.e create a binary node and add the two children to it
				// 1.2. If MJoinNode then create a mjoin node and add the chidren 
				// 2. Add the new query plan to the hash table of size "index"

			}else
			{
				iterate(PT,subset,index+1,_graph,mapping);
			}	
		}
	}

	public static Vector<Plan> getSubPlans(int planSize, Hashtable[] subPlans, Vector<Source>nonVisitedSources)
	{
		Vector<Plan> finalSubPlans = new Vector<Plan>();

		Enumeration<Plan> subPlanEnum = subPlans[planSize].elements();

		boolean add = true;
		Plan nextPlan;
		Source nextSource;
		while (subPlanEnum.hasMoreElements())
		{
			nextPlan = subPlanEnum.nextElement();

			for (int i=0; i<nonVisitedSources.size() && add==true; i++)
			{
				nextSource = nonVisitedSources.get(i);
				if(!nextPlan.isSourcePresent(nextSource))
				{
					add = false;
				}
			}

			if(add==true)
			{
				finalSubPlans.add(nextPlan);
			}

			add =true;
		}
		return finalSubPlans;
	}

	public static Vector<Source> removeVisitedNodes(Vector<Source>subset, Vector<Source>currentVisited)
	{
		Vector<Source> nonVisitedSources = new Vector<Source>();
		for (int i=0; i< subset.size(); i++)
		{
			Object o = subset.get(i);
			if (!(currentVisited.contains(o)))
				nonVisitedSources.add(subset.get(i));
		}
		return nonVisitedSources;
	}	



}



