/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms;

import java.util.Enumeration;
import java.util.Vector;

import com.sun.org.apache.xalan.internal.xsltc.runtime.Hashtable;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.Stream;

/**
 * This is an improvement of MJoinDynamicAlgorithm.
 * @author Venkatesh Raghavan 
 * @email venky@cs.wpi.edu
 * @version 1.0, Summer 2007
 * 
 */
public class MJoinDynamicCheaperAlgorithm extends MJoinDynamicAlgorithm{
	
	/**
	 * This algorithms starts bottom-up. Starts with plans with size =1.
	 * In MJoin the start stream is predetermined based on the pipeline one in processing.
	 * Instead of generating the plans and then pruning (as in MJoinDynamicAlgorithm), this approach prunes while adding.
	 * @param _plans
	 * @param _graph
	 * @return
	 */
	public static QueryPlan iterate(Vector<QueryPlan> _plans, JoinGraph _graph)
	{
		QueryPlan nextPlan, otherPlan;
		String name;

		Vector<QueryPlan> newPlans = new Vector<QueryPlan>();
		int count =1;
		int maxCount = _graph.getVerticesCount()-1;
		while(count <= maxCount)
		{
			if(_plans!=null && _plans.size() >0)
			{
				// Increase the size by 1.
				// That is, if you currently have plans of size 2, then now look at plans of size 3.
				// The iteration is truncated by the number of streams in teh Join Graph.
				for(int i=0; i<_plans.size(); i++)
				{
					newPlans = MJoinDynamicAlgorithm.generatePairs(_plans.elementAt(i), _graph);
					if(newPlans != null && newPlans.size() !=0)
					{
						// Prune plans that have identical leafs
						// For example, 
						//  PlanA: ((S1 JOIN S3) JOIN S2)
						//  PlanB: ((S1 JOIN S2) JOIN S3)
						// Both will have the name sources {S1, S2, S3} 
						for(int j=0; j<newPlans.size();j++)
						{
							nextPlan = newPlans.get(j);
							nextPlan.setCost(costEst.generateCost(nextPlan,_graph));
							name = Utils.sortedName(nextPlan);
							if(plansTable.get(name) == null)
							{
								plansTable.put(name, nextPlan);
							}else
							{
								otherPlan = (QueryPlan)plansTable.get(name);

								if(otherPlan.getCost() > nextPlan.getCost())
								{
									plansTable.put(name, nextPlan);
								}
							}
						}
					}
				}

				_plans.removeAllElements();
				Enumeration enums = plansTable.keys();
				while(enums.hasMoreElements())
				{
					_plans.add((QueryPlan)plansTable.get(enums.nextElement()));
				}
				plansTable.clear();
			}
			
			count++;
		}
		
		return _plans.get(0);
	}
	
	/**
	 * For a given join graph JG with streams S1, S2, ... Sn
	 * And, start stream Si.
	 * Generate the optimal join order of for processing tuples from Si.
	 * @param _graph
	 * @param _start
	 * @return
	 */
	public static QueryPlan order(JoinGraph _graph, Stream _start)
	{
		MJoinDynamicCheaperAlgorithm.plansTable = new Hashtable();
		int numberOfSource = _graph.getVerticesCount();
		// Start the whole process by creating the first plan with just the start stream.
		QueryPlan startPlan = new QueryPlan(_start);
		Vector<QueryPlan> ps = new Vector<QueryPlan>();
		ps.add(new QueryPlan(_start));
		return MJoinDynamicCheaperAlgorithm.iterate(ps,_graph);
	}
}
