/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms;

import java.util.Enumeration;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.Stream;

public class Selinger {
	
	public static QueryPlan order(JoinGraph _graph)
	{

		Vector<QueryPlan> ps = new Vector<QueryPlan>();
		if(_graph.getVerticesCount() >0)
		{
			Enumeration sourceEnum = _graph.getVertices();
			while(sourceEnum.hasMoreElements())
			{
				ps.add(new QueryPlan((Source)sourceEnum.nextElement()));
			}
			//return null;
			return Selinger.iterate(ps, _graph);
		}
		else return null;
	}
	
	public static Vector<QueryPlan> generatePairs(QueryPlan _input, JoinGraph _graph)
	{
		Vector<QueryPlan> plans = new Vector<QueryPlan>();
		/*
		 * Here we want to do the following:
		 * 1. We would want to generate all the streams that are not involved in the given plan.
		 *    For example, in our example above, (S1 JOIN S3) does not include the streams S2, S4 and S5.
		 * 2. Since the common heuristic is to not consider cartesian product.
		 *    In our example since there is no edge in the join graph S1-S4 or S3-S4,
		 *    therefore we only consider S2 and S5.
		 * Here the result of not visited node is a set {S2,S5}    
		 */
		Vector<Source> remainingSources = Utils.notVisitedSources(_graph, _input);
		BinaryJoinNodeImpl joinNode;
		QueryPlan nQueryPlan;
		 
		   
		/*
		 * Now, having the set {S2, S5} we can now generate plans:
		*/
		if(remainingSources !=null && remainingSources.size() >0)
		{
			 /*
			  ** --- ((S1 JOIN S3) JOIN S2)
			  * --- ((S1 JOIN S3) JOIN S5)
			  * To do this:
			  */
			
			for(int i=0; i<remainingSources.size();i++)
			{
					
				// 1. create an new binary node
				joinNode = new BinaryJoinNodeImpl();
				// 2. we clone the original plan
				// 3. Add the clone as its left child
				joinNode.addLeftChild(_input.clone());
				// 4. Add the next candidate in the set as the right child
				joinNode.addRightChild(remainingSources.elementAt(i));
				nQueryPlan = new QueryPlan(joinNode);
				// 5. Add the plan to the list
				plans.add(nQueryPlan);
				
			}
			
		}
		// Return the list of plans
		return plans;
	}
	
	
	public static QueryPlan iterate(Vector<QueryPlan> _plans, JoinGraph _graph)
	{
	 /*
			Vector<Source> remSources = MJoinDynamicAlgorithm.notVisitedSources(_graph, _plans.elementAt(0));
	
			if(remSources !=null && remSources.size() >0)
			{
				for(int j=0; j<remSources.size();j++)
				{
					_plans.add(new QueryPlan (remSources.elementAt(j)));
				}
			
		}
		*/
		Vector<QueryPlan> newPlans = new Vector<QueryPlan>();
		int count =1;
		int maxCount = _graph.getVerticesCount()-1;
		while(count <= maxCount)
		{
			if(_plans!=null && _plans.size() >0)
			{
				newPlans = new Vector<QueryPlan>();
				
				for(int i=0; i<_plans.size(); i++)
				{
					newPlans.addAll(Selinger.generatePairs(_plans.elementAt(i), _graph));
				}
				_plans = MJoinDynamicAlgorithm.prune(newPlans,_graph);
				
				Utils.print(_plans);
			}
			count++;
		}
		return _plans.get(0);
				
	}	
}
	
