/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.heuristic.ReverseHeuristic;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes.EdgeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel.CostEstimator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Vertex;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.classes.SourceImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;

public class bGreedyAlgorithm 
{
	static Hashtable hashNeighbors = new Hashtable();
	
	public static QueryPlan order(JoinGraph _graph, Stream _startStream)
	{
		hashNeighbors = new Hashtable();
		
		CostEstimator cE = new CostEstimator();
		QueryPlan newPlan = new QueryPlan();
		Node rootNode = bGreedyAlgorithm.greedy(null, null,_graph,_startStream,1);
		newPlan = new QueryPlan(rootNode);
		newPlan.setCost(cE.generateCost(newPlan, _graph));
		return newPlan;
	}
	
	private static void PopulateNeighborHash(JoinGraph _graph, Stream _start)
	{
		//rimma start: check for connection in the graph
		Vertex currVertex = _graph.getVertex(_start.getLabel());
		Vector currEdges = _graph.getEdges(currVertex);
		for (int v = 0; v < currEdges.size(); v++)
		{
			EdgeImpl empl = (EdgeImpl)currEdges.get(v);
			if (!empl.getVertexA().getLabel().equalsIgnoreCase(_start.getLabel()))
			{
				String neighbor = empl.getVertexA().getLabel();
				if (!hashNeighbors.containsKey(neighbor))
				{
					hashNeighbors.put(neighbor, neighbor);
				}
			}
			if (!empl.getVertexB().getLabel().equalsIgnoreCase(_start.getLabel()))
			{
				String neighbor = empl.getVertexB().getLabel();
				if (!hashNeighbors.containsKey(neighbor))
				{
					hashNeighbors.put(neighbor, neighbor);
				}
			}
		}			
	    //rimma end
	}
	
	private static Stream findCheapestCandidate(JoinGraph _graph, Streams _set, Stream _start)
	{
		Iterator sIter;
		Stream cheapest = null;
		Stream nextStream;
		double cost= Integer.MAX_VALUE;
		double tempCost;
		
		if(_set != null)
		{
			//PopulateNeighborHash(_graph, _start);
			
			if(_set.size() == 1)
			{
				return _set.get(0);
			}
			
			if(_set.size() == 2)
			{
				int index = _set.getIndexOf(_start);
				if(index ==0)
				{
					return _set.get(1);
				}
				else
				{
					return _set.get(0);
				}
			}
			else
			{
				sIter = _set.iterator();
				while(sIter.hasNext())
				{
					nextStream = (Stream) sIter.next();
					if(nextStream.getLabel().compareTo(_start.getLabel()) != 0)
					{
						//check that the next stream is "connected" to the node
						//if (hashNeighbors.containsKey(nextStream.getLabel()))
						//{						
							tempCost = ReverseHeuristic.calculateHeuristic(nextStream, _graph, _set);
							//System.out.println(nextStream.getLabel()+" Heuristic "+ tempCost);
							if(Double.compare(tempCost, cost) < 0)
							{
								cost = tempCost;
								cheapest = nextStream; 
							}
						//}
					}
				}
				//System.out.println(cheapest.getLabel());
				
				//PopulateNeighborHash(_graph, cheapest); //expand neighbor list based on newly added (cheapest) operator
				return cheapest;
			}
		}
		
		return null;
	}
	
	private static Streams getUnVisitedStreams(QueryPlan _plan, JoinGraph _graph, Stream _start)
	{	
		Streams temp = new Streams();
		Object next;
		Enumeration sourceEnum = _graph.getVertices();
		while(sourceEnum.hasMoreElements())
		{
			next = sourceEnum.nextElement();
			if(next instanceof Stream)
			{
				if(_plan == null)
				{
					temp.add((Stream) next);
				}
				else
				{
					if(!_plan.visited((SourceImpl) next))
					{
						temp.add((Stream) next);
					}
				}
			}
		}
		return temp;

	}
	
	private static Streams getUnVisitedStreamsBasedOnJoinGraph(QueryPlan _plan, JoinGraph _graph, Stream _start)
	{	
		Streams temp = new Streams();
		Stream tmpStream;
		Object next;
		
		Vector candidates = _graph.getEdges(_start);
		
		for (int i = 0; i < candidates.size(); i++)
		{
			EdgeImpl e = (EdgeImpl)candidates.get(i);
			if (e.getVertexA().getLabel().equalsIgnoreCase(_start.getLabel()))
			{
				tmpStream = new Stream(e.getVertexB().getLabel());
			}
			else 
			{
				tmpStream = new Stream(e.getVertexA().getLabel());
			}
			
			if(_plan == null)
			{
				temp.add(tmpStream);
			}
			else
			{
				if(!_plan.visited((SourceImpl) tmpStream))
				{
					temp.add((Stream) tmpStream);
				}
			}
		}		
	
		return temp;
	}
	
	private static Node greedy(QueryPlan _plan, Node node, JoinGraph _graph, Stream _startStream, int _index)
	{	
		BinaryJoinNodeImpl joinNode = new BinaryJoinNodeImpl();
		Node rightChild;
		Streams candidates; 
		
		if(_index > 0)
		{
			if(_index < _graph.getVerticesCount())
			{
				candidates = bGreedyAlgorithm.getUnVisitedStreams(_plan, _graph, _startStream);
								
				rightChild = findCheapestCandidate(_graph, candidates, _startStream);
				
				joinNode.addChild(1, rightChild);
				
				if(node == null)
				{
					_plan = new QueryPlan(joinNode);
					
				}
				else
				{
					node.addChild(0,joinNode);
				}
			
				bGreedyAlgorithm.greedy(_plan,joinNode,_graph,_startStream,_index+1);
				return joinNode;
			}else
			{
				node.addChild(0,_startStream);
				return node;
			}
		}else
		{
			return null;
		}
	}
	
}
