/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms;

import java.util.Enumeration;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel.CostEstimator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.classes.SourceImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.*;


public class fGreedyAlgorithm {
	public static QueryPlan order(JoinGraph _graph, Stream _startStream)
	{
		CostEstimator cE = new CostEstimator();
		QueryPlan newPlan = new QueryPlan((Node)_startStream);
		newPlan = fGreedyAlgorithm.greedy(newPlan,_graph,1);
		newPlan.setCost(cE.generateCost(newPlan, _graph));
		return newPlan;
	}

	/*
	private static Streams getUnVisitedStreams(QueryPlan _plan, JoinGraph _graph)
	{	
		Streams temp = new Streams();
		Object next;
		Enumeration sourceEnum = _graph.getVertices();
		while(sourceEnum.hasMoreElements())
		{
			next = sourceEnum.nextElement();
			if(next instanceof Stream)
			{
				if(!_plan.visited((SourceImpl) next))
				{
					temp.add((Stream) next);
				}	
			}
		}
		return temp;

	}
	*/
	
	private static QueryPlan greedy(QueryPlan _plan, JoinGraph _graph, int _index)
	{
		QueryPlan tempPlan;
		BinaryJoinNodeImpl joinNode = new BinaryJoinNodeImpl();
		double currentCheapestCost = Float.MAX_VALUE;
		
		double tempCost;
		Source cheapestStream = null;
		Source tempStream = null;
	
		
		CostEstimator costEstimator = new CostEstimator();
		if(_index != _graph.getVerticesCount())
		{
			//Streams unVisitedStreams = fGreedyAlgorithm.getUnVisitedStreams(_plan, _graph);
			
			Vector<Source> unVisitedStreams = Utils.notVisitedSources(_graph,_plan);
			
			if(unVisitedStreams!=null && unVisitedStreams.size() >0 )
			{
				for(int i=0; i<unVisitedStreams.size();i++)
				{
					tempStream = (Source) unVisitedStreams.get(i);
					joinNode.addChild(0, _plan);
					joinNode.addChild(1, tempStream);
					tempPlan = new QueryPlan(joinNode);
					tempCost = costEstimator.generateCost(tempPlan, _graph);
					if(tempCost < currentCheapestCost)
					{
						currentCheapestCost = tempCost;
						cheapestStream = tempStream;
					}
				}
			}else
			{
				return _plan;
			}
			joinNode.addChild(0, _plan);
			joinNode.addChild(1, cheapestStream);
			tempPlan = new QueryPlan(joinNode);
			
			return  fGreedyAlgorithm.greedy(tempPlan,_graph,_index+1);
		}else
		{
			return _plan;
		}
	}
}
