/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.heuristic;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Vertex;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.*;

public class ReverseHeuristic {

	public static double calculateHeuristic(Stream  _candidate, JoinGraph _graph, Streams _set)
	{
		double heuristic =1;
		int _candidateIndex;
		if((_set != null) && (_set.size() >0))
		{
			_candidateIndex = _set.getIndexOf(_candidate);
			if(_candidateIndex >=0 )
			{
				for(int i=0; i < _set.size(); i++)
				{
					//System.out.println("HERE");
					if(i!=_candidateIndex)
					{
						heuristic *= _set.get(i).getInputRate();

						for(int j=i+1; j < _set.size(); j++)
						{
							if(j!=_candidateIndex)
							{
								try
								{
									if(_graph.hasEdge((Vertex) _set.get(i), (Vertex) _set.get(j)))
									{
										heuristic *= _graph.getEdge((Vertex) _set.get(i), (Vertex) _set.get(j)).getValue();
									}
								}catch(Exception e)
								{
									e.printStackTrace();
								}
							}
						}
					}
				}
				return heuristic;
			}
		}

		return -1;
	}
	
	
	public static double calculateForwardHeuristic(Streams _setStreamsAlreadyInPlan, 
			 									   Stream _candidate, 
			 									   JoinGraph _graph, 
			 									   Streams _set)
	{
		double heuristic = Integer.MAX_VALUE;
		
		try
		{
			//find the min edge
			
			for (int i = 0; i < _setStreamsAlreadyInPlan.size(); i++)
			{
				Stream currStream = _setStreamsAlreadyInPlan.get(i);
				
				Vertex vA = (Vertex)currStream;
				Vertex vB = (Vertex)_candidate;
				
				if(_graph.hasEdge(vA, vB))
				{
					double currCost = _graph.getEdge(vA, vB).getValue();
					
					if (currCost < heuristic)
					{
						heuristic = currCost;
					}
				}
			}			
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		
		return heuristic;
	}

}