/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.experiments;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Enumeration;

import com.sun.org.apache.xalan.internal.xsltc.runtime.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.IKAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.MJoinDynamicAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.MJoinDynamicCheaperAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.MinimumSpanningTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.bGreedyAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.fGreedyAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.generators.JoinGraphGenerator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.generators.StarJoinGenerator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.generators.StreamGenerator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes.JoinGraphImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel.CostEstimator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Edge;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.RankedRootedTreeNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.Stream;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.utils.Timer;

public class AcyclicJoin {

	public static String generateHeadling()
	{
		String headline = new String();
		headline+= "ID"+",";
		headline+= "Dynamic Plan"+",";
		headline+= "Dynamic Plan Cost"+",";
		headline+= "Time"+",";
		headline+= "IK Plan"+",";
		headline+= "IK Plan Cost"+",";
		headline+= "Time"+",";
		headline+= "F-Greedy Plan"+",";
		headline+= "F-Greedy Plan Cost" +",";
		headline+= "Time"+",";
		headline+= "B-Greedy Plan"+",";
		headline+= "B-Greedy Plan Cost";
		headline+= "Time"+",";
		headline+= "IK/Optimal"+",";
		headline+= "F-Greedy/Optimal"+",";
		headline+= "B-Greedy/Optimal"+",";
		headline+= "FAB-Greedy/Optimal"+",";
		headline+= "ALL/Optimal"+",";
		headline+= "\n";
		return headline;
	}

	public static void runCyclic()
	{
		try
		{
			// Dynamic Programming 

			myTimer.start();
			optPlan = MJoinDynamicCheaperAlgorithm.order(graph,firstStream);
			myTimer.stop();

			MJoinDynamicCheaperAlgorithm.plansTable.clear();
			System.gc();

			cyclicWriter.write(optPlan.toString()+","+optPlan.getCost()+","+Long.toString(myTimer.duration())+",");

			//IK-Algorithm

			myTimer.start();
			newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.generateOptimalMJoinOrdering(graph,firstStream,mstWriter);
			myTimer.stop();

			ikPlan = Utils.generateQueryPlan(null,newNode,graph);

			ikPlan.setCost(cE.generateCost(ikPlan,graph));

			cyclicWriter.write(ikPlan.toString()+","+ikPlan.getCost()+","+Long.toString(myTimer.duration())+",");

			//F-Greedy

			myTimer.start();
			fGreedyPlan = fGreedyAlgorithm.order(graph,firstStream);
			myTimer.stop();

			cyclicWriter.write(fGreedyPlan.toString()+","+fGreedyPlan.getCost()+","+Long.toString(myTimer.duration())+",");

			//B-Greedy

			myTimer.start();
			bGreedyPlan = bGreedyAlgorithm.order(graph,firstStream);
			myTimer.stop();

			cyclicWriter.write(bGreedyPlan.toString()+","+bGreedyPlan.getCost()+","+Long.toString(myTimer.duration())+",");

			cyclicWriter.write(ikPlan.getCost()/optPlan.getCost()+","+ fGreedyPlan.getCost()/optPlan.getCost()+","+ bGreedyPlan.getCost()/optPlan.getCost()+","+Math.min(bGreedyPlan.getCost(), fGreedyPlan.getCost())/optPlan.getCost()+","+Math.min(ikPlan.getCost(), Math.min(bGreedyPlan.getCost(), fGreedyPlan.getCost()))/optPlan.getCost());

			System.gc();
			
			if(ikPlan.getCost() == optPlan.getCost())
			{
				optIK++;
			}

			if(bGreedyPlan.getCost() == optPlan.getCost())
			{
				optBGreedy++;
			}

			if(fGreedyPlan.getCost() == optPlan.getCost())
			{
				optFGreedy++;
			}

			if( bGreedyPlan.getCost() == optPlan.getCost() || fGreedyPlan.getCost() == optPlan.getCost())
			{
				optFABGreedy++;
			}

			if( bGreedyPlan.getCost() == optPlan.getCost() || fGreedyPlan.getCost() == optPlan.getCost() || ikPlan.getCost() == optPlan.getCost())
			{
				optIKFBGreedy++;
			}

			cyclicWriter.write("\n");

		}catch (IOException e) 
		{
		}

	}
	public static void runAcyclic()
	{
		try
		{
			
			
			Utils.writeSelectivity(graph, treeWriter);
			
			// Dynamic Programming 

			myTimer.start();
			optPlan = MJoinDynamicCheaperAlgorithm.order(graph,firstStream);
			myTimer.stop();

			MJoinDynamicCheaperAlgorithm.plansTable.clear();
			
			System.gc();

			acyclicWriter.write(optPlan.toString()+","+optPlan.getCost()+","+Long.toString(myTimer.duration())+",");

			//IK-Algorithm

			myTimer.start();
			newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.generateOptimalOrdering(graph,firstStream);
			myTimer.stop();

			ikPlan = Utils.generateQueryPlan(null,newNode,graph);

			ikPlan.setCost(cE.generateCost(ikPlan,graph));

			acyclicWriter.write(ikPlan.toString()+","+ikPlan.getCost()+","+Long.toString(myTimer.duration())+",");

			//F-Greedy

			myTimer.start();
			fGreedyPlan = fGreedyAlgorithm.order(graph,firstStream);
			myTimer.stop();

			acyclicWriter.write(fGreedyPlan.toString()+","+fGreedyPlan.getCost()+","+Long.toString(myTimer.duration())+",");

			//B-Greedy

			myTimer.start();
			bGreedyPlan = bGreedyAlgorithm.order(graph,firstStream);
			myTimer.stop();

			acyclicWriter.write(bGreedyPlan.toString()+","+bGreedyPlan.getCost()+","+Long.toString(myTimer.duration())+",");

			acyclicWriter.write(ikPlan.getCost()/optPlan.getCost()+","+ fGreedyPlan.getCost()/optPlan.getCost()+","+ bGreedyPlan.getCost()/optPlan.getCost()+","+Math.min(bGreedyPlan.getCost(), fGreedyPlan.getCost())/optPlan.getCost()+","+Math.min(ikPlan.getCost(), Math.min(bGreedyPlan.getCost(), fGreedyPlan.getCost()))/optPlan.getCost());
			
			System.gc();

			if(ikPlan.getCost() == optPlan.getCost())
			{
				acyclicOptIK++;
			}

			if(bGreedyPlan.getCost() == optPlan.getCost())
			{
				acyclicOptBGreedy++;
			}

			if(fGreedyPlan.getCost() == optPlan.getCost())
			{
				acyclicOptFGreedy++;
			}

			if( bGreedyPlan.getCost() == optPlan.getCost() || fGreedyPlan.getCost() == optPlan.getCost())
			{
				acyclicOptFABGreedy++;
			}

			if( bGreedyPlan.getCost() == optPlan.getCost() || fGreedyPlan.getCost() == optPlan.getCost() || ikPlan.getCost() == optPlan.getCost())
			{
				acyclicOptIKFBGreedy++;
			}

			acyclicWriter.write("\n");

		}catch (IOException e) 
		{
		}

	}
	
	public static int minN, maxN, maxCount;

	public static Timer myTimer;
	public static CostEstimator cE;

	public static Stream firstStream;

	public static String folderName;
	
	public static BufferedWriter selectivityWriter, mstWriter, cyclicWriter, acyclicWriter, summaryWriter, treeWriter;
	public static String selectivityFName, mstFName, cyclicFName, acyclicFName, summaryFName, treeFName;

	public static RankedRootedTreeNodeImpl newNode;

	public static QueryPlan ikPlan, optPlan, bGreedyPlan, fGreedyPlan;

	public static JoinGraphImpl graph;
	
	public static int optIK, optBGreedy, optFGreedy, optFABGreedy, optIKFBGreedy;

	public static int acyclicOptIK, acyclicOptBGreedy, acyclicOptFGreedy, acyclicOptFABGreedy, acyclicOptIKFBGreedy;
	
	/**
	 * 
	 * @param args Discussed Below:
	 *        args[0] -min N (Number of Streams)
	 *        args[1] -max N (Number of Streams)
	 *        args[2] -iterations
	 *        args[3] -folder location
	 */

	public static void main(String args[])
	{
		myTimer = new Timer();
		cE  = new CostEstimator();

		if(args[0] != null && args[1] != null && args[2] != null)
		{
			minN = Integer.parseInt(args[0]);
			maxN = Integer.parseInt(args[1]);
			maxCount = Integer.parseInt(args[2]);

			if(args[3] != null)
			{
				folderName = args[3];
			}
			else folderName = ""; 

			String headline = Run1.generateHeadling();
			int counter;

			graph = new JoinGraphImpl();

			for(int i=minN; i<=maxN; i++)
			{
				counter =1;

				// Generate the file name

				selectivityFName = folderName+"/NUM_"+i+"_selectivity.txt";
				mstFName = folderName+"/NUM_"+i+"_MST.txt";
				cyclicFName = folderName+"/NUM_"+i+"_Cyclic.csv";
				acyclicFName = folderName+"/NUM_"+i+"_Acyclic.csv";
				summaryFName = folderName+"/NUM_"+i+"_Summary.txt";
				treeFName = folderName+"/NUM_"+i+"_Tree.txt";
				
				// Initialize Counts.
				optIK = 0;
				optBGreedy =0;
				optFGreedy =0;
				optFABGreedy =0;
				optIKFBGreedy =0;
				
				acyclicOptIK = 0;
				acyclicOptBGreedy =0;
				acyclicOptFGreedy =0;
				acyclicOptFABGreedy =0;
				acyclicOptIKFBGreedy =0;

				try
				{
					// Open files.

					selectivityWriter = new BufferedWriter(new FileWriter(selectivityFName));
					mstWriter = new BufferedWriter(new FileWriter(mstFName));
					cyclicWriter =  new BufferedWriter(new FileWriter(cyclicFName));
					acyclicWriter =  new BufferedWriter(new FileWriter(acyclicFName));
					summaryWriter = new BufferedWriter(new FileWriter(summaryFName));
					treeWriter = new BufferedWriter(new FileWriter(treeFName));
					
					acyclicWriter.write(headline);
					cyclicWriter.write(headline);
					
					while(counter <= maxCount)
					{
						// Generate Streams
						graph = (JoinGraphImpl) StreamGenerator.generate(graph,i);
						
						// Generate Selectivities
						graph = (JoinGraphImpl) JoinGraphGenerator.generate(graph);
						
						// Identify the first stream

						firstStream = null;

						Enumeration enums = graph.getVertices();
						Object nextObject;

						while(enums.hasMoreElements())
						{
							nextObject = enums.nextElement();
							if(((Stream) nextObject).getLabel().compareTo("1")==0)
							{
								firstStream = (Stream) nextObject;
							}
						}

						
						
						Utils.writeIterationInformation(selectivityWriter, counter);
						Utils.writeIterationInformation(mstWriter,counter);
						Utils.writeIterationInformation(treeWriter,counter);
						
						Utils.writeSelectivity(graph,selectivityWriter);
						
						
						cyclicWriter.write(counter+",");
						acyclicWriter.write(counter+",");

						graph= (JoinGraphImpl) MinimumSpanningTree.kruskalAlgorithm(graph);
						
						runAcyclic();
						
						graph =  (JoinGraphImpl) edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.generators.Utils.generateCompleteGraph(graph);
						
						runCyclic();
						
						counter++;
					}
					summaryWriter.write("Number of streams: "+ graph.getVerticesCount()+"\n");
					summaryWriter.write("Total Number of Runs: "+maxCount+"\n");
					summaryWriter.write("####################\n\n");
					summaryWriter.write("       Acyclic: Without Cartesian \n");
					summaryWriter.write("####################\n\n");
					summaryWriter.write("IK Algorithm Optimal: "+optIK+"\n");
					summaryWriter.write("F-Greedy Algorithm Optimal: "+optFGreedy+"\n");
					summaryWriter.write("B-Greedy Algorithm Optimal: "+optBGreedy+"\n");
					summaryWriter.write("F-Greedy V B-Greedy Algorithm Optimal: "+optFABGreedy+"\n");
					summaryWriter.write("F-Greedy V B-Greedy V IK Algorithm Optimal: "+optIKFBGreedy+"\n");
					
					summaryWriter.write("####################\n\n");
					summaryWriter.write("       Acyclic: Cartesian Product \n");
					summaryWriter.write("####################\n\n");
					summaryWriter.write("IK Algorithm Optimal: "+acyclicOptIK+"\n");
					summaryWriter.write("F-Greedy Algorithm Optimal: "+acyclicOptFGreedy+"\n");
					summaryWriter.write("B-Greedy Algorithm Optimal: "+acyclicOptBGreedy+"\n");
					summaryWriter.write("F-Greedy V B-Greedy Algorithm Optimal: "+acyclicOptFABGreedy+"\n");
					summaryWriter.write("F-Greedy V B-Greedy V IK Algorithm Optimal: "+acyclicOptIKFBGreedy+"\n");
					

					selectivityWriter.close();
					mstWriter.close();
					cyclicWriter.close();
					acyclicWriter.close();
					summaryWriter.close();
					treeWriter.close();

				}catch (IOException e) 
				{
				}		
			}

		}
	}
}
