/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.experiments;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Enumeration;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.IKAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.MJoinDynamicAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.MinimumSpanningTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.bGreedyAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.fGreedyAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.generators.JoinGraphGenerator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.generators.StreamGenerator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes.JoinGraphImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel.CostEstimator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Edge;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.RankedRootedTreeNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.Stream;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.utils.Timer;

public class Experiment2 {

	public static String generateHeadling()
	{
		String headline = new String();
		headline+= "ID"+",";
		headline+= "Dynamic Plan"+",";
		headline+= "Dynamic Plan Cost"+",";
		headline+= "Time"+",";
		headline+= "IK Plan"+",";
		headline+= "IK Plan Cost"+",";
		headline+= "Time"+",";
		headline+= "F-Greedy Plan"+",";
		headline+= "F-Greedy Plan Cost" +",";
		headline+= "Time"+",";
		headline+= "B-Greedy Plan"+",";
		headline+= "B-Greedy Plan Cost";
		headline+= "Time"+",";
		headline+= "\n";
		return headline;
	}
	/**
	 * Generates a query Plan out of a Linear tree;
	 * @param _parent
	 * @param _currentNode
	 * @param _graph
	 * @return
	 */
	public static QueryPlan generateQueryPlan(QueryPlan _parent, Node _currentNode, JoinGraph _graph)
	{
		QueryPlan plan;
		BinaryJoinNodeImpl joinNode;
		if(_parent == null)
		{
			plan = new QueryPlan((Source)_graph.getVertex(_currentNode.getLabel()));
		}else
		{
			joinNode = new BinaryJoinNodeImpl();
			joinNode.addLeftChild(_parent);
			joinNode.addRightChild((Source)_graph.getVertex(_currentNode.getLabel()));

			plan = new QueryPlan(joinNode);
		}

		if(_currentNode.chidrenCount() ==1)
		{
			plan = Experiment_1.generateQueryPlan(plan, (Node)_currentNode.getChild(0), _graph);
		}

		return plan;
	}

	public static void printSelectivity(JoinGraph _graph, BufferedWriter _out)
	{
		try
		{
			Enumeration enums = _graph.getVertices();
			Object nextObject;
			_out.write("Streams: \n");
			while(enums.hasMoreElements())
			{
				nextObject = enums.nextElement();
				_out.write(((Stream) nextObject).toString()+" ("+((Stream) nextObject).getCardinality()+") \n");
			}

			Enumeration edges = _graph.getEdges();
			Edge nextEdge;
			_out.write("Join Selectivities: \n");
			while(edges.hasMoreElements())
			{
				nextEdge = (Edge) edges.nextElement();
				_out.write(nextEdge.toString()+"\n");
			}
		}catch (IOException e) 
		{
		}
	}

	public static void main(String args[])
	{
		Timer myTimer = new Timer();

		CostEstimator cE = new CostEstimator();

		RankedRootedTreeNodeImpl newNode;

		QueryPlan ikPlan, optPlan, bGreedyPlan, fGreedyPlan;

		int counter;
		int maxCount =1;

		JoinGraphImpl graph = new JoinGraphImpl();

		for(int i=7; i<=7; i++)
		{
			counter =1;

			JoinGraph  acyclicJoinGraph;
			while(counter <= maxCount)
			{

				graph = (JoinGraphImpl) StreamGenerator.generate(graph,i);
				graph = (JoinGraphImpl) JoinGraphGenerator.generate(graph);
				acyclicJoinGraph= MinimumSpanningTree.kruskalAlgorithm(graph);

				Stream firstStream = null;

				Enumeration enums = acyclicJoinGraph.getVertices();
				Object nextObject;


				while(enums.hasMoreElements())
				{
					nextObject = enums.nextElement();
					if(((Stream) nextObject).getLabel().compareTo("1")==0)
					{
						firstStream = (Stream) nextObject;
					}
				}

				acyclicJoinGraph.print();


				// Dynamic Programming 

				myTimer.start();
				optPlan = MJoinDynamicAlgorithm.order(acyclicJoinGraph,firstStream);
				myTimer.stop();

				System.out.println("Dynamic: "+optPlan.toString()+","+optPlan.getCost()+","+Long.toString(myTimer.duration())+",");

				//IK-Algorithm

				myTimer.start();
				newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.generateOptimalOrdering(acyclicJoinGraph,firstStream);
				myTimer.stop();

				ikPlan = Experiment_1.generateQueryPlan(null,newNode,acyclicJoinGraph);

				ikPlan.setCost(cE.generateCost(ikPlan,acyclicJoinGraph));

				System.out.println("IK: "+ikPlan.toString()+","+ikPlan.getCost()+","+Long.toString(myTimer.duration())+",");

				//F-Greedy

				myTimer.start();
				fGreedyPlan = fGreedyAlgorithm.order(acyclicJoinGraph,firstStream);
				myTimer.stop();

				System.out.println("F-Greedy: "+fGreedyPlan.toString()+","+fGreedyPlan.getCost()+","+Long.toString(myTimer.duration())+",");

				//B-Greedy

				myTimer.start();
				bGreedyPlan = bGreedyAlgorithm.order(acyclicJoinGraph,firstStream);
				myTimer.stop();

				System.out.println("B-Greedy: "+bGreedyPlan.toString()+","+bGreedyPlan.getCost()+","+Long.toString(myTimer.duration())+",");

				counter++;
			}

		}
	}
}
