/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.experiments;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Enumeration;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.IKAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.MJoinDynamicAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.MinimumSpanningTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.bGreedyAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.algorithms.fGreedyAlgorithm;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.generators.JoinGraphGenerator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.generators.StreamGenerator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes.JoinGraphImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel.CostEstimator;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Edge;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.RankedRootedTreeNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.streams.Stream;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.utils.Timer;

public class Experiment_1 {
	
	public static String generateHeadling()
	{
		String headline = new String();
		headline+= "ID"+",";
		headline+= "Dynamic Plan"+",";
		headline+= "Dynamic Plan Cost"+",";
		headline+= "Time"+",";
		headline+= "IK Plan"+",";
		headline+= "IK Plan Cost"+",";
		headline+= "Time"+",";
		headline+= "F-Greedy Plan"+",";
		headline+= "F-Greedy Plan Cost" +",";
		headline+= "Time"+",";
		headline+= "B-Greedy Plan"+",";
		headline+= "B-Greedy Plan Cost";
		headline+= "Time"+",";
		headline+= "\n";
		return headline;
	}
	/**
	 * Generates a query Plan out of a Linear tree;
	 * @param _parent
	 * @param _currentNode
	 * @param _graph
	 * @return
	 */
	public static QueryPlan generateQueryPlan(QueryPlan _parent, Node _currentNode, JoinGraph _graph)
	{
		QueryPlan plan;
		BinaryJoinNodeImpl joinNode;
		if(_parent == null)
		{
			plan = new QueryPlan((Source)_graph.getVertex(_currentNode.getLabel()));
		}else
		{
			joinNode = new BinaryJoinNodeImpl();
			joinNode.addLeftChild(_parent);
			joinNode.addRightChild((Source)_graph.getVertex(_currentNode.getLabel()));
			
			plan = new QueryPlan(joinNode);
		}
			
		if(_currentNode.chidrenCount() ==1)
		{
			plan = Experiment_1.generateQueryPlan(plan, (Node)_currentNode.getChild(0), _graph);
		}
		
		return plan;
	}
	
	public static void printSelectivity(JoinGraph _graph, BufferedWriter _out)
	{
		try
		{
			Enumeration enums = _graph.getVertices();
			Object nextObject;
			_out.write("Streams: \n");
			while(enums.hasMoreElements())
			{
				nextObject = enums.nextElement();
				_out.write(((Stream) nextObject).toString()+" ("+((Stream) nextObject).getCardinality()+") \n");
			}
			
			Enumeration edges = _graph.getEdges();
			Edge nextEdge;
			_out.write("Join Selectivities: \n");
			while(edges.hasMoreElements())
			{
				nextEdge = (Edge) edges.nextElement();
				_out.write(nextEdge.toString()+"\n");
			}
		}catch (IOException e) 
		{
		}
	}
	
	public static void main(String args[])
	{
		Timer myTimer = new Timer();
		
		CostEstimator cE = new CostEstimator();
		
		BufferedWriter selectivityWriter, mstWriter, outputWriter, summaryWriter;
		
		String selectivityFName, mstFName, outputFName, summaryFName;
		
		int optIK, optBGreedy, optFGreedy, optFABGreedy, optIKFBGreedy;
		
		int bLik, bLf; 
		int fLik, fLb;
		int ikLb, ikLf, ikLfab;
		int fabLik;
		
		String fName;
		
		RankedRootedTreeNodeImpl newNode;
		
		QueryPlan ikPlan, optPlan, bGreedyPlan, fGreedyPlan;
		
		String headline = Experiment_1.generateHeadling();
		
		int counter;
		int maxCount =500;
		
		JoinGraphImpl graph = new JoinGraphImpl();
		
		for(int i=17; i<=20; i++)
		{
			counter =1;
			selectivityFName = "Experiments/NUM_"+i+"_selectivity.txt";
			mstFName = "Experiments/NUM_"+i+"_MST.txt";
			outputFName = "Experiments/NUM_"+i+"_ALL.csv";
			summaryFName = "Experiments/NUM_"+i+"_SUMMARY.txt";
			
			try
			{
				optIK = 0;
				optBGreedy =0;
				optFGreedy =0;
				optFABGreedy =0;
				optIKFBGreedy =0;
				
				bLik =0;
				bLf = 0;
				
				fLik = 0;
				fLb =0;
				
				ikLb =0;
				ikLf =0;
				ikLfab =0;
				
				fabLik =0;
				
				selectivityWriter = new BufferedWriter(new FileWriter(selectivityFName));
				mstWriter = new BufferedWriter(new FileWriter(mstFName));
				outputWriter =  new BufferedWriter(new FileWriter(outputFName));
				summaryWriter = new BufferedWriter(new FileWriter(summaryFName));
				outputWriter.write(headline);
				while(counter < maxCount)
				{
					
					graph = (JoinGraphImpl) StreamGenerator.generate(graph,i);
					graph = (JoinGraphImpl) JoinGraphGenerator.generate(graph);
					
					Stream firstStream = null;
					
					Enumeration enums = graph.getVertices();
					Object nextObject;
					
					
					while(enums.hasMoreElements())
					{
						nextObject = enums.nextElement();
						if(((Stream) nextObject).getLabel().compareTo("1")==0)
						{
							firstStream = (Stream) nextObject;
						}
					}
					
					selectivityWriter.write("-----------------------------\n");
					selectivityWriter.write(" ID: "+counter+"\n");
					selectivityWriter.write("-----------------------------\n");
					
					mstWriter.write("-----------------------------\n");
					mstWriter.write(" ID: "+counter+"\n");
					mstWriter.write("-----------------------------\n");
					
					Experiment_1.printSelectivity(graph,selectivityWriter);
					
					outputWriter.write(counter+",");
					
					// Dynamic Programing 
					
					myTimer.start();
					optPlan = MJoinDynamicAlgorithm.order(graph,firstStream);
					myTimer.stop();
					
					outputWriter.write(optPlan.toString()+","+optPlan.getCost()+","+Long.toString(myTimer.duration())+",");
					
					//IK-Algorithm
					
					myTimer.start();
					newNode = (RankedRootedTreeNodeImpl) IKAlgorithm.generateOptimalMJoinOrdering(graph,firstStream,mstWriter);
					myTimer.stop();
					
					ikPlan = Experiment_1.generateQueryPlan(null,newNode,graph);
					
					ikPlan.setCost(cE.generateCost(ikPlan,graph));
					
					outputWriter.write(ikPlan.toString()+","+ikPlan.getCost()+","+Long.toString(myTimer.duration())+",");
					
					//F-Greedy
					
					myTimer.start();
					fGreedyPlan = fGreedyAlgorithm.order(graph,firstStream);
					myTimer.stop();
					
					outputWriter.write(fGreedyPlan.toString()+","+fGreedyPlan.getCost()+","+Long.toString(myTimer.duration())+",");
					
					//B-Greedy
					
					myTimer.start();
					bGreedyPlan = bGreedyAlgorithm.order(graph,firstStream);
					myTimer.stop();
					
					outputWriter.write(bGreedyPlan.toString()+","+bGreedyPlan.getCost()+","+Long.toString(myTimer.duration())+",");
					
					//System.out.println(ikPlan.toString()); 
					outputWriter.write("\n");
					
					if(ikPlan.getCost() == optPlan.getCost())
					{
						optIK++;
					}
					
					if(bGreedyPlan.getCost() == optPlan.getCost())
					{
						optBGreedy++;
					}
					
					if(fGreedyPlan.getCost() == optPlan.getCost())
					{
						optFGreedy++;
					}
					
					if( bGreedyPlan.getCost() == optPlan.getCost() || fGreedyPlan.getCost() == optPlan.getCost())
					{
						optFABGreedy++;
					}
					
					if( bGreedyPlan.getCost() == optPlan.getCost() || fGreedyPlan.getCost() == optPlan.getCost() || ikPlan.getCost() == optPlan.getCost())
					{
						optIKFBGreedy++;
					}
					
					if(bGreedyPlan.getCost() < fGreedyPlan.getCost())
					{
						bLf++;
					}
					
					if(bGreedyPlan.getCost() < ikPlan.getCost())
					{
						bLik++;
					}
					
					if(fGreedyPlan.getCost() < fGreedyPlan.getCost())
					{
						fLb++;
					}
					
					if(fGreedyPlan.getCost() < ikPlan.getCost())
					{
						fLik++;
					}
					
					if(ikPlan.getCost() < bGreedyPlan.getCost())
					{
						ikLf++;
					}
					
					if(ikPlan.getCost() < fGreedyPlan.getCost())
					{
						ikLb++;
					}
					
					if(ikPlan.getCost() < fGreedyPlan.getCost() && ikPlan.getCost() < bGreedyPlan.getCost())
					{
						ikLfab++;
					}
					
					if((bGreedyPlan.getCost() < ikPlan.getCost()) || (fGreedyPlan.getCost() < ikPlan.getCost()))
					{
						fabLik++;
					}
					counter++;
				}
				
				summaryWriter.write("Number of streams: "+graph.getVerticesCount()+"\n");
				summaryWriter.write("####################\n\n");
				summaryWriter.write(" Total Number of Runs: "+counter+"\n");
				summaryWriter.write("IK Algorithm Optimal: "+optIK+"\n");
				summaryWriter.write("F-Greedy Algorithm Optimal: "+optFGreedy+"\n");
				summaryWriter.write("B-Greedy Algorithm Optimal: "+optBGreedy+"\n");
				summaryWriter.write("F-Greedy V B-Greedy Algorithm Optimal: "+optFABGreedy+"\n");
				summaryWriter.write("F-Greedy V B-Greedy V IK Algorithm Optimal: "+optIKFBGreedy+"\n");
				
				summaryWriter.write("\n\n");
				summaryWriter.write("F-Greedy < B-Greedy Algorithm: "+fLb+"\n");
				summaryWriter.write("F-Greedy < IK Algorithm: "+fLik+"\n");
				
				summaryWriter.write("\n\n");
				summaryWriter.write("B-Greedy < F-Greedy Algorithm: "+bLf+"\n");
				summaryWriter.write("B-Greedy < IK Algorithm: "+bLik+"\n");
				
				summaryWriter.write("\n\n");
				summaryWriter.write("IK Algorithm < F-Greedy: "+ikLf+"\n");
				summaryWriter.write("IK Algorithm < B-Greedy: "+ikLb+"\n");
				summaryWriter.write("IK Algorithm < BEST(B-Greedy,F-Greedy): "+ikLfab+"\n");
				
				summaryWriter.write("\n\n");
				summaryWriter.write("BEST(F-Greedy,B-Greedy) < IK Algorithm: "+fabLik+"\n");
				
				selectivityWriter.close();
				mstWriter.close();
				outputWriter.close();
				summaryWriter.close();
				
			}catch (IOException e) 
			{
			}	
		}
	}
}
