/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.classes.exceptions.JoinGraphException;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Edge;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Vertex;

public class JoinGraphImpl implements JoinGraph{

	Hashtable edges;
	Hashtable vertices;
	
	public JoinGraphImpl()
	{
		edges = new Hashtable();
		vertices = new Hashtable();
	}
	
	public Vertex getVertex(String _label)
	{
		return (Vertex) this.vertices.get(_label);
	}
	private Object fetchEdge(Vertex _start, Vertex _end)
	{	
		Object edgeObject = edges.get(_start.getLabel()+":"+_end.getLabel());
		if(edgeObject != null)
		{
			return edgeObject;
		}else
		{
			return edges.get(_end.getLabel()+":"+_start.getLabel());
		}
	}
	
	public void addVertex(Vertex _node)
	{
		this.vertices.put(_node.getLabel(), _node);
	}
	
	public void removeAllEdges()
	{
		this.edges = new Hashtable();
	}
	public void addEdge(Vertex _start, Vertex _end, double _value) throws JoinGraphException
	{
		
		// Starting and Ending Vertices are the same then throw an excpetion
		if(_start.getLabel().compareTo(_end.getLabel()) ==0)
		{
			JoinGraphException e = new JoinGraphException(" Start Vertex and End Vertex have same label");
			throw e;
		}else
		{
			//System.out.println("Adding Edge: "+ _start.getLabel()+":"+_end.getLabel());
			
			EdgeImpl newEdge = new EdgeImpl(_start, _end, _value);	
			String key = _start.getLabel()+":"+_end.getLabel();
			if(edges.get(key) == null)
			{
				edges.put(key, newEdge); 
			}else
			{
				System.out.println("ERROR:"+_start.getLabel()+":"+_end.getLabel());
			}
			
			/*
			Enumeration enums =  edges.keys();
			int count =0;
			while(enums.hasMoreElements())
			{
				enums.nextElement();
				count++;
			}
			System.out.println(count);
			*/
		}
	}
	
	public boolean hasEdge(Vertex _start, Vertex _end)
	{
		Edge edge = (Edge) this.fetchEdge(_start,_end);
		if(edge== null)
		{
			edge = (Edge) this.fetchEdge(_end,_start);
			if(edge == null)
			{
				return false;
			}else return true;
			
		}else return true;
		
	}
	public Edge getEdge(Vertex _start, Vertex _end) throws JoinGraphException
	{
		Edge edge = (Edge) this.fetchEdge(_start,_end);
		if(edge== null)
		{
			edge = (Edge) this.fetchEdge(_end,_start);
			if(edge == null)
			{
				JoinGraphException e = new JoinGraphException("No edge between Start Vertex"+ _start.getLabel()+"and End "+ _end.getLabel()+" Vertex");
				throw e;
			}else return edge;
			
		}else return edge;
		
	}
	
	public Enumeration getVertices()
	{
		return this.vertices.elements();
	}
	public Enumeration getEdges()
	{
		return this.edges.elements();
	}
	
	public int getVerticesCount()
	{
		return this.vertices.size();
	}
	
	public void print()
	{
		System.out.println("\n**** Join Graph ****");
		System.out.println(" Vertices:"); 
		Enumeration vertexEnum = this.getVertices();
		Vertex tempVertex;
		while(vertexEnum.hasMoreElements())
		{
			tempVertex = (Vertex) vertexEnum.nextElement();
			tempVertex.print();
		}
		System.out.println(" Edges:");
		
		Enumeration edgeEnum = this.getEdges();
		Edge tempEdge;
		while(edgeEnum.hasMoreElements())
		{
			tempEdge = (Edge) edgeEnum.nextElement();
			tempEdge.print();
		}
		System.out.println("**** ****");
	}
	
	
	public Vector<Edge> getEdges(Vertex _node)
	{
		
		Vector<Edge> result = new Vector<Edge>();
		Enumeration allEdges = this.edges.elements();
		Edge nextEdge;
		while(allEdges.hasMoreElements())
		{
			nextEdge = (Edge) allEdges.nextElement();
			//nextEdge.print();
			if((nextEdge.getVertexA().getLabel().compareTo(_node.getLabel())==0) || (nextEdge.getVertexB().getLabel().compareTo(_node.getLabel())==0))
			{
				result.add(nextEdge);
			}
		}
		return result;
	}
	
	public Vector<Vertex> getNeighbors(Vertex _current)
	{
		Vector<Edge> connectedEdges = this.getEdges(_current);
		Vector<Vertex> result = new Vector<Vertex>();
		String currentLabel = _current.getLabel();
		Vertex a, b;
		
		if(connectedEdges != null)
		{
			for (int i =0; i< connectedEdges.size(); i++)
			{
				a = connectedEdges.get(i).getVertexA();
				b = connectedEdges.get(i).getVertexB();
				if(a.getLabel().compareTo(currentLabel) !=0)
				{
					result.add(a);
				}else
				{
					result.add(b);
				}
			}
			return result;
		}
		return null;
	}
	public void removeAllVertex()
	{
		this.removeAllEdges();
		this.vertices = new Hashtable();
	}
	public void removeEdge(Edge edge) 
	{	Vertex _start=edge.getVertexA();
		Vertex _end =edge.getVertexB();

		Edge edgeTemp = (Edge) this.fetchEdge(_start,_end);
		if(edgeTemp== null)
		{
			edgeTemp = (Edge) this.fetchEdge(_end,_start);
			if(edge == null)
			{
				System.out.println("no edge between"+_start);
			}
			else 
			{
				this.edges.remove(_end.getLabel()+":"+_start.getLabel());
			}
		}else
		{
			this.edges.remove(_start.getLabel()+":"+_end.getLabel());
		}
	}
	
	public Vector removeVertex(Vertex vertex)
	{
		Vector edges=this.getEdges(vertex);
		for(int i=0;i<edges.size();i++)
		{
			this.removeEdge((Edge) edges.get(i));
		}
		this.vertices.remove(vertex.getLabel());
		return edges;
	}
}
