/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel.BinaryJoin;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.BinaryJoinNodeImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes.QueryPlan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.classes.SourceImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel.Utils;

public class SymmetricNLJWithNoTouchCost {
	public static float insertionCost = 0; // Hashing Cost
	public static float deleteCost =0; // Purging Cost
	public static float joinCost = 1; // Probing Cost
	public static float windowSize = 1;
	
	public static void evaluateCost(Plan _currentPlan, JoinGraph _graph)
	{
		Node rootNode = _currentPlan.getRoot(); 
		double inputRate, tuplesDueToLeftProbes, tuplesDueToRightProbes;
		double leftState, rightState, selectivity; 
		double  cardinality = 0.0,cpuCost =0.0, memoryCost =0.0, probes = 0.0;
		
		if(rootNode instanceof SourceImpl)
		{
			inputRate = ((Source) rootNode).getCardinality();
			
			cardinality = inputRate * SymmetricNLJWithNoTouchCost.windowSize;
			cpuCost =  inputRate * (SymmetricNLJWithNoTouchCost.deleteCost + SymmetricNLJWithNoTouchCost.insertionCost);
			memoryCost = 0.0;
			probes = inputRate; 
		}else if(rootNode instanceof Plan)
		{
			SymmetricNLJWithNoTouchCost.evaluateCost((Plan)_currentPlan.getRoot(), _graph);
			
			cardinality = ((Plan)_currentPlan.getRoot()).getCardinality() * SymmetricNLJWithNoTouchCost.windowSize;
			cpuCost =  ((Plan)_currentPlan.getRoot()).getCPUCost();
			memoryCost = ((Plan)_currentPlan.getRoot()).getMemoryCost();
			probes = ((Plan)_currentPlan.getRoot()).getCardinality();
		}else if(rootNode instanceof BinaryJoinNodeImpl)
		{
			Plan lPlan, rPlan;
			
			if( (((BinaryJoinNodeImpl) rootNode).getLeftChild() != null) && (((BinaryJoinNodeImpl) rootNode).getRightChild() != null))
			{
				if(((BinaryJoinNodeImpl) rootNode).getLeftChild() instanceof Plan)
				{
					lPlan = (Plan) ((BinaryJoinNodeImpl) rootNode).getLeftChild();
					SymmetricNLJWithNoTouchCost.evaluateCost(lPlan,_graph);
				}else
				{
					lPlan = new QueryPlan((Node) ((BinaryJoinNodeImpl) rootNode).getLeftChild());
					SymmetricNLJWithNoTouchCost.evaluateCost(lPlan,_graph);
				}
				
				if(((BinaryJoinNodeImpl) rootNode).getRightChild() instanceof Plan)
				{
					rPlan = (Plan) ((BinaryJoinNodeImpl) rootNode).getRightChild();
					SymmetricNLJWithNoTouchCost.evaluateCost(rPlan,_graph);
				}else
				{
					rPlan = new QueryPlan((Node) ((BinaryJoinNodeImpl) rootNode).getRightChild());
					SymmetricNLJWithNoTouchCost.evaluateCost(rPlan,_graph);
				}
					
				selectivity = Utils.binaryJoinSelectivity(lPlan.getSources(), rPlan.getSources(), _graph);
				
				leftState =  lPlan.getCardinality();
				rightState =  rPlan.getCardinality(); 
				
				cardinality = leftState * rightState * selectivity;
				
				tuplesDueToLeftProbes = lPlan.getProbes() * rightState * selectivity;
				
				tuplesDueToRightProbes = rPlan.getProbes() * leftState * selectivity;
				
				cpuCost = rPlan.getCPUCost() + lPlan.getCPUCost() + (tuplesDueToLeftProbes * SymmetricNLJWithNoTouchCost.joinCost) + (tuplesDueToRightProbes * SymmetricNLJWithNoTouchCost.joinCost);
				
				memoryCost = rPlan.getMemoryCost() + lPlan.getMemoryCost() + leftState +rightState;
				
				probes = tuplesDueToLeftProbes  + tuplesDueToRightProbes;
			}
		}
		
		_currentPlan.setCardinality(cardinality);
		_currentPlan.setCPUCost(cpuCost);
		_currentPlan.setMemoryCost(memoryCost);
		_currentPlan.setProbes(probes);
		return;
	}

}
