/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.costmodel;

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Edge;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.JoinGraph;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.classes.SourceImpl;

public class Utils {
	/**
	 * This function returns the cummulative product of the join selectivities.
	 * For example, A Join B, the selectivity is just \sigma(AB), while
	 * the cummulative selectivity of plan ((A Join B) Join C) is \sigma(AC) * \sigma(BC).
	 * Note that the \sigma(AB) is already incorporated in the cost of A Join B. Also, as the cost
	 * is generated bottom-up, we don't consider \sigma(AB). 
	 * @param leftSources sources in the left hand side of the join operator.
	 * @param rightSources sources in the right hand side of the join operator.
	 * @param _graph is the join graph that specifies the individual selectivities between
	 * the various sources.
	 * @return
	 */
	public static double binaryJoinSelectivity(Vector leftSources, Vector rightSources, JoinGraph _graph)
	{
		//System.out.println("Selectivity");
		double selectivity = 1;
		if(leftSources != null && rightSources!= null && leftSources.size() >0 && rightSources.size() >0)
		{
			SourceImpl left, right;
			Edge tempEdge;
			for(int i=0; i< leftSources.size(); i++)
			{
				left = (SourceImpl) leftSources.get(i);
				for(int j=0; j<rightSources.size();j++)
				{
					right = (SourceImpl) rightSources.get(j);
					try
					{
						if(_graph.hasEdge(left,right))
						{
							tempEdge = _graph.getEdge(left, right);
							if(tempEdge != null)
							{
								//System.out.print(tempEdge.getSelectivity() +" * ");
								selectivity *= tempEdge.getValue();
							}
						}
					}catch(Exception e)
					{
						e.printStackTrace();
					}
				}
			}
			
			//System.out.println("\n"+selectivity);
			return selectivity;
		}else
		{
			return 0.0;
		}
	}
}
