/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes;

import java.io.Serializable;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.BinaryJoinNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;

/**
 * Implementation of a binary join operator.
 * @author Venkatesh Raghavan (venky@cs.wpi.edu)
 * @version 1.0, July 2007
 */
public class BinaryJoinNodeImpl implements BinaryJoinNode, Serializable{
	/**
	 * Label of the node.
	 */
	private String label = "BINARY JOIN";
	/**
	 * Left child of the join operator.
	 */
	private Object leftChild;
	/**
	 * Right child of the operator.
	 */
	private Object rightChild;
	
	public BinaryJoinNodeImpl()
	{
	}
	
	public double getValue()
	{
		return 0.0;
	}
	
	public void setValue(double _value)
	{
		// DO nothing
	}
	/**
	 * 
	 */
	public String getLabel()
	{
		return this.label;
	}
	
	public int chidrenCount(){
		if(leftChild == null && rightChild ==null)
		{
			return 0;
		}else if(leftChild != null && rightChild != null)
		{
			return 2;
		}else
		{
			return 1;
		}
	}
	
	public Object getChild(int _index)
	{
		if(_index == 0)
		{
			return this.leftChild;
		}else if (_index == 1)
		{
			return this.rightChild;
		}else
		{
			return null;
		}
	}
	
	public void addChild(Object _child)
	{
		if(this.leftChild == null)
		{
			this.leftChild = _child;
		}else
		{
			this.rightChild = _child;
		}
	}
	
	public void addChild(int _index, Object _child)
	{
		if(_index == 0)
		{
			this.leftChild = _child;
		}else if (_index == 1)
		{
			this.rightChild = _child;
		}
	}

	public void print()
	{
		System.out.println(this.label);
		if(this.leftChild != null)
		{
			if(this.leftChild instanceof Plan)
			{
				((Plan) this.leftChild).print();
			}else if(this.leftChild instanceof Node)
			{
				((Node) this.leftChild).print();
			}
		}
		if(this.rightChild != null)
		{
			if(this.rightChild instanceof Plan)
			{
				((Plan) this.rightChild).print();
			}else if(this.rightChild instanceof Node)
			{
				((Node) this.rightChild).print();
			}
		}
	}
	
	public String toString()
	{
		String result = new String();
		result += "(";
		
		if(this.leftChild != null)
		{
			if(this.leftChild instanceof Plan)
			{
				result += ((Plan) this.leftChild).toString();
			}
			else if(this.leftChild instanceof Node)
			{
				result += ((Node) this.leftChild).toString();
			}
		}
		if(this.rightChild != null)
		{
			if(this.rightChild instanceof Plan)
			{
				result += ((Plan) this.rightChild).toString();
			}else if(this.rightChild instanceof Node)
			{
				result += ((Node) this.rightChild).toString();
			}
		}
		result +=")";
		
		return result;
	}
	
	public String toXmlString()
	{
		String result = new String();
		result += "<Operator>";
		
		if(this.leftChild != null)
		{
			if(this.leftChild instanceof Plan)
			{
				result += ((QueryPlan) this.leftChild).toXmlString();
			}
			else if(this.leftChild instanceof Node)
			{
				result += ((Node) this.leftChild).toXmlString();
			}
		}
		if(this.rightChild != null)
		{
			if(this.rightChild instanceof Plan)
			{
				result += ((QueryPlan) this.rightChild).toXmlString();
			}
			else if(this.rightChild instanceof Node)
			{
				result += ((Node) this.rightChild).toXmlString();
			}
		}
		result +="</Operator>";
		
		return result;
	}
	
	public void printPretty(){
		System.out.print("(");
		if(this.leftChild != null)
		{
			if(this.leftChild instanceof Plan)
			{
				((Plan) this.leftChild).printPretty();
			}else if(this.leftChild instanceof Node)
			{
				((Node) this.leftChild).printPretty();
			}
		}
		if(this.rightChild != null)
		{
			if(this.rightChild instanceof Plan)
			{
				((Plan) this.rightChild).printPretty();
			}else if(this.rightChild instanceof Node)
			{
				((Node) this.rightChild).printPretty();
			}
		}
		System.out.print(")");
	}
	
	public Object getLeftChild()
	{
		return this.leftChild;
	}
	
	public void addLeftChild(Object _child)
	{
		this.leftChild = _child;
	}
	
	public Object getRightChild()
	{
		return this.rightChild;
	}
	
	public void addRightChild(Object _child)
	{
		this.rightChild = _child;
	}
	
	public Object clone()
	{
		BinaryJoinNodeImpl clone = new BinaryJoinNodeImpl();
		for(int i=0;i <this.chidrenCount();i++)
		{
			if(this.getChild(i) instanceof Node)
			{
				clone.addChild(i,((Node) this.getChild(i)).clone());
			}else if(this.getChild(i) instanceof Plan)
			{
				clone.addChild(i,((Plan) this.getChild(i)).clone());
			}
		}
		return clone;
	}
	
	public void removeAllChildren()
	{
		this.leftChild = null;
		this.rightChild = null;
	}
}
