/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes;

import java.util.Hashtable;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.MJoinNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;

/**
 * 
 * @author Venkatesh Raghavan (venky@cs.wpi.edu)
 *
 */
public class MJoinNodeImpl implements MJoinNode{

	private Vector<Plan> probingSequence;
	private Vector<Object> chidren;
	/**
	 * Label of the node.
	 */
	private String label = "MJOIN";
	
	public MJoinNodeImpl()
	{
		this.chidren = new Vector<Object>();
		this.probingSequence = new Vector<Plan>();
	}
	
	public int chidrenCount()
	{
		return chidren.size();
	}
	
	public Object getChild(int _index)
	{
		if(this.chidren == null)
		{
			return null;
		}else
		{
			if(_index >= this.chidrenCount())
			{
				// TODO Raise Exception;
				return null;
			}else
			{
				return this.chidren.elementAt(_index);
			}
		}
	}
	
	public void addChild(Object _child)
	{
		if(this.chidren == null)
		{
			chidren = new Vector<Object>();
			this.chidren.add(_child);
		}else
		{
			this.chidren.add(_child);
		}
	}
	
	public void addChild(int _index, Object _child)
	{
		if(this.chidren == null)
		{
			chidren = new Vector<Object>();
			this.chidren.add(_index,_child);
		}else
		{
			this.chidren.add(_index,_child);
		}
	}
	
	public void removeAllChildren()
	{
		this.chidren.removeAllElements();
	}
	
	public String toXmlString()
	{
		return this.label;
	}
	
	public Vector<Object> getChildren()
	{
		return this.chidren;
	}
	
	public Object clone()
	{
		MJoinNodeImpl clone = new MJoinNodeImpl();
		for(int i=0;i <this.chidrenCount();i++)
		{
			if(this.getChild(i) instanceof Node)
			{
				clone.addChild(i,((Node) this.getChild(i)).clone());
			}else if(this.getChild(i) instanceof Plan)
			{
				clone.addChild(i,((Plan) this.getChild(i)).clone());
			}
		}
		return clone;
	}
	
	public String toString()
	{
		String result = new String();
		result += "[";
		for(int i=0;i <this.chidrenCount();i++)
		{
			result += this.chidren.elementAt(i).toString();
		}
		result +="]";
		
		return result;
	}
	
	public void printPretty(){
		System.out.print("[");
		for(int i=0;i <this.chidrenCount();i++)
		{
			if(this.chidren.elementAt(i) instanceof Plan)
			{
				((Plan) this.chidren.elementAt(i)).printPretty();
			}else if(this.chidren.elementAt(i) instanceof Node)
			{
				((Node) this.chidren.elementAt(i)).printPretty();
			}
		}
		System.out.print("]");
	}	
	
	public void print()
	{
		System.out.println(this.label);
		for(int i=0;i <this.chidrenCount();i++)
		{
			if(this.chidren.elementAt(i) instanceof Plan)
			{
				((Plan) this.chidren.elementAt(i)).print();
			}else if(this.chidren.elementAt(i) instanceof Node)
			{
				((Node) this.chidren.elementAt(i)).print();
			}
		}
	
	}
	
	public String getLabel()
	{
		return this.label;
	}
	
	public double getValue()
	{
		return 0.0;
	}
	
	public void setValue(double _value)
	{
		// DO nothing
	}
	
	
	public Plan getProbingSequence(Object _node)
	{
		
		if(this.probingSequence == null)
		{
			return null;
		}else
		{
			int index = this.chidren.indexOf(_node);
			
			if(index > -1)
			{
				return this.probingSequence.elementAt(index);
			}
		}
		return null;
	}
	public void setProbingSequence(Object _node, Plan _plan)
	{
		if(this.probingSequence == null)
		{
			this.probingSequence = new Vector<Plan>();
			if(this.chidren == null)
			{
				this.chidren = new Vector<Object>();
				this.chidren.add(_node);
				this.probingSequence.add(_plan);
			}else
			{
				int index = this.chidren.indexOf(_node);
				
				if(index > -1)
				{
					this.probingSequence.add(index, _plan);
				}else
				{
					this.chidren.add(_node);
					index = this.chidren.indexOf(_node);
					this.probingSequence.add(_plan);
				}
			}
			
		}
		
	}
}
