/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes;

import java.io.Serializable;
import java.util.Hashtable;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Plan;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.classes.SourceImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.Node;

public class QueryPlan implements Plan,  Serializable{
	private Node root;
	private double cpuCost = 0.0;
	private double memoryCost = 0.0;
	private double cardinality = 0.0;
	private double probes = 0.0;
	private double planCost = -1;
	public QueryPlan()
	{
		
	}
	public QueryPlan(Node _root)
	{
		this.root = _root;
	}
	public Node getRoot()
	{
		return this.root;
	}
	public double getCost()
	{
		return this.planCost;
	}
	public void setCost(double _value)
	{
		this.planCost = _value;
	}
	
	public void printPretty()
	{
		this.root.printPretty();	
	}
	
	public void printXml()
	{
		this.root.printPretty();
	}	
	
	public void print()
	{
		this.root.print();
	}
	
	private boolean find(SourceImpl _source)
	{
		if(root instanceof SourceImpl)
		{
			if(root.getLabel().compareTo(_source.getLabel()) ==0)
			{
				return true;
			}else{
				return false;
			}
		}else if(root instanceof Plan)
		{
			return ((Plan)root).visited(_source);
		}else if(root instanceof BinaryJoinNodeImpl)
		{
			Plan tempPlan = null;
			if(root.getChild(0) instanceof Plan)
			{
				tempPlan = (Plan) root.getChild(0);
			}else if(root.getChild(0) instanceof Node)
			{
				tempPlan = new QueryPlan((Node) root.getChild(0));
			}
			if(tempPlan!=null  && tempPlan.visited(_source))
			{
				return true;
			}else
			{
				tempPlan = null;
				if(root.getChild(1) instanceof Plan)
				{
					tempPlan = (Plan) root.getChild(1);
				}else if(root.getChild(1) instanceof Node)
				{
					tempPlan = new QueryPlan((Node) root.getChild(1));
				}
				
				if(tempPlan!=null  && tempPlan.visited(_source))
				{
					return true;
				}else return false;
				
				/*
				tempPlan = new QueryPlan((Node) root.getChild(1));
				return tempPlan.visited(_source);
				*/
				
			}
		}else
		{
			return false;
		}
	}
	
	public boolean visited(SourceImpl _source)
	{
		return this.find(_source);
	}
	public Vector getSources()
	{
		Vector result = new Vector();
		if(root instanceof SourceImpl)
		{
			result.add(root);
			return result;
		}else if(root instanceof Plan)
		{
			return ((Plan) root).getSources();
		}else if(root instanceof BinaryJoinNodeImpl)
		{
			Plan tempPlan = null;
			for(int i=0; i < root.chidrenCount();i++)
			{
				if(root.getChild(i) instanceof Plan)
				{
					tempPlan = (Plan) root.getChild(i);
				}else if(root.getChild(i) instanceof Node)
				{
					tempPlan = new QueryPlan((Node) root.getChild(i));
				}
				if(tempPlan != null)
				{
					result.addAll(tempPlan.getSources());
				}
			}
			return result;
		}else
		{
			return null;
		}
	}
	
	public Object clone()
	{
		QueryPlan clone = new QueryPlan((Node) this.root.clone());
		return clone;
	}
	
	public String toString()
	{
		return  this.root.toString();
	}
	
	public String toXmlString()
	{
		int nestedLevel = 1;
		return this.root.toXmlString();
	}
	

	public void setCardinality(double _cardinality)
	{
		this.cardinality = _cardinality;
	}
	public double getCardinality()
	{
		return this.cardinality;
	}
	public void setCPUCost(double _cpu)
	{
		this.cpuCost = _cpu;
	}
	public double getCPUCost()
	{
		return this.cpuCost;
	}
	public void setMemoryCost(double _memory)
	{
		this.memoryCost = _memory;
	}
	
	public double getMemoryCost()
	{
		return this.memoryCost;
	}
	
	public void setProbes(double _probes)
	{
		this.probes = _probes;
	}
	public double getProbes()
	{
		return this.probes;
	}

	public boolean isSourcePresent(Source target)
	{
		boolean found = false;
		Vector<Source> sources =  this.getSources(); 
		for(int i=0; i<sources.size();i++)
		{
			if(target.getLabel().compareTo(sources.elementAt(i).getLabel()) ==0)
			{
				return true;
			}
		}
		return found;
	}
}
