/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.classes;

import java.util.Enumeration;
import java.util.Vector;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.classes.SourceImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Vertex;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces.RootedTreeNode;

public class RootedTreeNodeImpl implements RootedTreeNode{
	private RootedTreeNode parent;
	private Vector children;
	private Vertex source;
	public RootedTreeNodeImpl(Vertex _source)
	{
		children = new Vector();
		source = _source;
	}
	
	public String getLabel()
	{
		if(source !=null)
		{
			return this.source.getLabel();
		}
		return null;
	}
	
	public int chidrenCount()
	{
		if(this.children !=null)
		{
			return this.children.size();
		}
		else return 0;
	}
	
	public Object getChild(int _index)
	{
		if(this.children !=null)
		{
			if(this.children.size() > _index)
			{
				return this.children.elementAt(_index);
			}
		}
		return null;
	}
	
	public void addChild(int _index, Object _child)
	{
		if(this.children ==null)
		{
			this.children = new Vector();
		}
		this.children.add(_index,_child);
	}
	
	public void addChild(Object _child)
	{
		if(this.children ==null)
		{
			this.children = new Vector();
		}
		this.children.add(_child);
	}
	
	public void print()
	{
		if(source !=null)
		{
			System.out.println(this.source.getLabel());
		}else System.out.println(" NO LABEL ");
	}
	
	public void printPretty()
	{
		if(source !=null)
		{
			System.out.print("["+this.source.getLabel()+"]");
		}else System.out.print(" NO LABEL ");
	}
	
	public Object clone()
	{
		//TODO
		return null;
	}
	
	public void setParentNode(RootedTreeNode _parent)
	{
		this.parent = _parent;
	}
	public RootedTreeNode getParent()
	{
		return this.parent;
	}
	public Enumeration getChildren()
	{
		return this.children.elements();
	}
	
	public double getValue()
	{
		// TODO:
		return 0.0;
	}
	
	public void setValue(double _value)
	{
		// TODO:
	}
	public Vertex getVertex()
	{
		return this.source;
	}
	
	public Vector<RootedTreeNode> getDescendants()
	{
		Vector<RootedTreeNode> result = new Vector<RootedTreeNode>();
		RootedTreeNode nextChild;
		for(int i=0; i<this.chidrenCount();i++)
		{
			nextChild = (RootedTreeNode) this.getChild(i);
			result.add(nextChild);
			result.addAll(nextChild.getDescendants());
		}
		return result;
	}
	
	public void removeAllChildren()
	{
		this.children = new Vector();
	}

	public String toXmlString() {
		// TODO Auto-generated method stub
		return null;
	}
}
