/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.interfaces.Vertex;

/**
 * A Node interface to define the function required by tree.
 * @author Venkatesh Raghavan (venky@cs.wpi.edu)
 * @version 1.0 July 2007
 */

public interface Node extends Vertex{
	/**
	 * Number of children for a given node
	 * @return integer numebr of the number of children.
	 */
	public int chidrenCount();

	/**
	 * Returns a child with a specified index.
	 * @param _index an index of the child.
	 * @return child (which can either be a Node or a Plan) at the specified index.
	 */
	public Object getChild(int _index);
	/**
	 * Add a child to a node.
	 * @param _child Child to be added.
	 */
	public void addChild(Object _child);
	/**
	 * Add a childe to the node at a specified index.
	 * @param _index an index of the child.
	 * @param _child which can be either a node or a plan.
	 */
	public void addChild(int _index, Object _child);
	/**
	 * Returns a clone
	 * @return
	 */
	public Object clone();
	
	public void removeAllChildren();
	
	public String toXmlString();
}
