/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.graph.tree.interfaces;

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.classes.SourceImpl;
import edu.wpi.cs.dsrg.xmldb.xat.component.queryoptimizer.source.interfaces.Source;
/**
 * Plan interface is used to define functions that are common for all 
 * query plans.
 * @author Venkatesh Ragahvan (venky@cs.wpi.edu)
 * @version 1.0, July 2007
 */
public interface Plan{
	
	
	/**
	 * Returns the root of the plan.
	 * @return this can eithe be a node or a subplan.
	 */
	public Node getRoot();
	/**
	 * Returns the cost of the plan.
	 * @return cost of the plan.
	 */
	public double getCost();
	/**
	 * Assign cost to the plan.
	 * @param _value cost of the plan.
	 */
	public void setCost(double _value);
	/**
	 * Prints the plan at iteratively call its children to print.
	 */
	public void print();
	
	/**
	 * All in one line
	 *	 */
	public void printPretty();
	/**
	 * Return true if the plan has this source
	 * @param _source
	 * @return
	 */
	public boolean visited(SourceImpl _source);
	
	/**
	 * Returns the sources in the plan.
	 * @return a vector of sources in the plan.
	 */
	public Vector getSources();
	
	/**
	 * Returns a clonable object
	 * @return
	 */
	public Object clone();	
	
	public boolean isSourcePresent(Source target);
	
	public String toString();
	
	public void setCardinality(double _cardinality);
	public double getCardinality();
	public void setCPUCost(double _cpu);
	public double getCPUCost();
	public void setMemoryCost(double _memory);
	public double getMemoryCost();
	public void setProbes(double _probes);
	public double getProbes();
}
