/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator;

//import java.io.FileNotFoundException;
//import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.Hashtable;
//import java.util.Enumeration;

//import javax.xml.parsers.DocumentBuilder;
//import javax.xml.parsers.DocumentBuilderFactory;

//import org.w3c.dom.Attr;
//import org.w3c.dom.DOMException;
//import org.w3c.dom.Document;
//import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

//import edu.wpi.cs.dsrg.xmldb.xat.common.CQLParser.CQLParser;
//import edu.wpi.cs.dsrg.xmldb.xat.common.CQLParser.ParseException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNodeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATNodeDoesNotExistException;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

/**
 * Generate an XAT (or dag) from an XML file.  It is probably easiest to
 * work out the query plan on paper and then encode it in the XML.  
 * There are not many assumptions in the format of the XML file, but here they are:
 * <ol>
 * <li> The xml file does not encode both Stream and NonStream operators.  While
 * care has been taken to ensure this will work with both of them, it is not
 * logically correct to inter-twine them in one plan
 * <li> The XML file should not remove any elements even if they are unused.  That is,
 * the operator element should still have the children element even if it is a leaf.  Otherwise,
 * the xml parser will complain
 * </ol>
 *
 * More to come 
 * @author: Brad Pielech
 */
public class CQLQueryPlanGenerator extends QueryPlanGeneratorImp  {
	private Hashtable operatorTable;
	private Hashtable operatorOutputSchemaTable;
	private ClassVariableVisitor variableVisitor;
/**
 * FromXMLFileQueryPlanGenerator constructor comment.
 */
protected CQLQueryPlanGenerator() {
	super();	
	operatorTable = new Hashtable();
	operatorOutputSchemaTable = new Hashtable();
	variableVisitor = new ClassVariableVisitor();
	
}

private void addChildren(Node children, DagNode node) {
	
	NodeList childList = children.getChildNodes();
	System.out.println("adds children");
	for (int i = 0; i < childList.getLength(); i++){
		Node next = childList.item(i);
		if(next.getNodeType() == Node.ELEMENT_NODE){
			NamedNodeMap attrs = next.getAttributes();
			Node idN=attrs.getNamedItem("type");
			String iD=idN.getNodeValue();
			if(iD.equals("operator")){
				System.out.println("adds operator to xmlOC");
				DagNode child;
				Node idO=attrs.getNamedItem("id");
				System.out.println("Operator id is:" + idO);
				String id=idO.getNodeValue();
			   if(operatorTable.containsKey(id)){
				child = (DagNode)operatorTable.get(id);	
			} else{
				child = new DagNodeImp();
				operatorTable.put(id,child);
			}
			node.appendChild(child);
			child.addParent(node);
		}
		if(iD.equals("stream")){
			System.out.println("adds stream to xmlOC");
			Node idS=attrs.getNamedItem("name");
			String name=idS.getNodeValue();
			System.out.println("Stream name is " + name);
			node.appendChild(name);
			
		}
	}
	}
}




/**
 * This method will call the visit method in variableVisitor.  It is too
 * hard to generalize the creation of variables, so each operator
 * type is handled in its own visit method.  The visit method will
 * pass the operator and the Node (DOM, not XAT node) and each
 * visit method can be customized per operator type.  
 * @param children org.w3c.dom.Node
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode
 */
private void addClassVariables(Node classVariables, DagNode node) {
	variableVisitor.visit(node.getOperator(), classVariables);	
	}

private void assignSchema(Node schema, DagNode node, String opID) {
	XATQueueSchema outSchema = new XATQueueSchema();

	NodeList nodeList = schema.getChildNodes();
	
	// if schema is not defined in config file (nodelist.getLenth() == 0),
	// the schema of the output queue is set by combining input queue schemas 
	// in the run() method of each operator.
	if(nodeList.getLength() > 0) {
		for (int i = 0; i < nodeList.getLength(); i++){
			Node next = nodeList.item(i);
			if(next.getNodeType() == Node.ELEMENT_NODE){
				NamedNodeMap attrs = next.getAttributes();
				String name = ((Node) attrs.getNamedItem("name")).getNodeValue();
				if (name.equals("col_name")) {
					outSchema.addColumn(((Node) attrs.getNamedItem("value")).getNodeValue());
				} else {
					System.out.println("Wrong configuration!");
				}
			}
		}
		operatorOutputSchemaTable.put(opID, outSchema);
	}
}

/**
 * Parses the parents element and connects parents.  If the parent
 * does not exist in the hash table, then just create a new DagNode with
 * the corresponding ID.  The operator will be connected later
 * Creation date: (2/1/03 12:53:59 PM)
 * @param parents org.w3c.dom.Node
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode
 */
private void addParents(Node parents, DagNode node) {
	
	NodeList parentList = parents.getChildNodes();
	for (int i = 0; i < parentList.getLength(); i++){
		Node next = parentList.item(i);
		if(next.getNodeType() == Node.ELEMENT_NODE){
			NamedNodeMap attrs = next.getAttributes();
			Node idN = attrs.getNamedItem("id");
			String id = idN.getNodeValue();
			DagNode parent;
			if(operatorTable.containsKey(id)){
				parent = (DagNode)operatorTable.get(id);	
			} else{
				parent = new DagNodeImp();
				operatorTable.put(id,parent);
			}
			node.addParent(parent);
			//parent.addOpChild(node);
			parent.appendChild(node);
		}
	}
	
}
/**
 * Adds the properties of the DagNode as specified in the XML document
 * This is pretty straight forward- parse the properties element,
 * get the name and value attributes, then set the property
 * @param children org.w3c.dom.Node
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode
 */
private void addProperties(Node properties, DagNode node) throws Exception {
	
	NodeList propertiesList = properties.getChildNodes();
	for (int i = 0; i < propertiesList.getLength(); i++){
		Node next = propertiesList.item(i);
		if(next.getNodeType() == Node.ELEMENT_NODE){
			NamedNodeMap attrs = next.getAttributes();
            String name = ((Node) attrs.getNamedItem("name")).getNodeValue();
            String value = ((Node) attrs.getNamedItem("value")).getNodeValue();
            node.getOperator().setProperty(name, value);
		}
	}
	
}
/**
 * If we are working with XATStreamOperators, we can connect all of the
 * operators together with this recursive method 
 * @param XATStreamOperator root.  
 */

/*private void connectQueues(XATStreamOperator root) throws Exception {

	//Make input queues for the streams if any are going to be connected
	for(int j=0; j<root.getNode().getStreamChildSize();j++){
		String sName = root.getNode().getStreamChildName(j);
		System.out.println("Stream whose queue to connect is: " + sName);
		if(root instanceof XATStreamOperator){
			Constructor[] c1 = Class.forName(this.datamodel).getConstructors();
			Class[] cl;
			Constructor construct;
			int count = 0;
			
			do{
			
			construct = c1[count];
			cl = construct.getParameterTypes();
			count++;
			}while(cl.length != 1);
			
			XATQueue q = (XATQueue) construct.newInstance(new Object[] {root});	
			q.setQueueName(sName);
			System.out.println("add input queue for " + sName);
			root.addInputQueue(q);	
			System.out.println("Input queue nimber is: " + root.getInputQueuePosition(q));
		}
	} 

    XATNode[] children = root.getNode().getOpChildren();
    for (int i = 0; i < children.length; i++) {
	    try{
        XATNode node = children[i];
        XATStreamOperator child = (XATStreamOperator) node.getOperator();
        XATQueue queue = ((XATStreamOperator) child).getOutputQueue();
        *//** set the input queue.  If we are a multi queue operator, then add the queue *//*
        if(root instanceof XATMultiQueueStreamOperator){
			((XATMultiQueueStreamOperator) root).addInputQueue(queue);
        } else { 
         ((XATStreamOperator) root).setInputQueue(queue);
        }

        connectQueues(child);
    	} catch (XATNodeDoesNotExistException e){
			*//**
			 this is thrown when an operator does not have
			 a child.  This shouldnt happen because we only iterate
			 through the available children, but this catch will prevent
			 the error from being propagated higher
			*//*
    	}
	}

    return;
}
*/


//mjba
private void connectQueues(XATStreamOperator root) throws Exception {

	//Make input queues for the streams if any are going to be connected
	for(int j=0; j<root.getNode().getChildrenSize();j++){
		System.out.println("children size is : " + root.getNode().getChildrenSize());
		System.out.println("j in for loop is:" + j);
		Object name = root.getNode().getChild(j);
		if(name instanceof String){
			System.out.println("name is stream: " + name);
			if(root instanceof XATStreamOperator){
				Constructor[] c1 = Class.forName(this.datamodel).getConstructors();
				Class[] cl;
				Constructor construct;
				int count = 0;
				
				do{				
					construct = c1[count];
					cl = construct.getParameterTypes();
					count++;
				} while(cl.length != 1);
				
				XATQueue q = (XATQueue) construct.newInstance(new Object[] {root});	
				String sName= (String) name;
				q.setQueueName(sName);
				System.out.println("queue name is " + sName + " which is (name): " + name);
				root.addInputQueue(q);		
			}
		} 
		if(name instanceof XATNode){
			System.out.println("name is XATNode: " + name);
			try{
				XATNode node = (XATNode) name;
				XATStreamOperator child = (XATStreamOperator) node.getOperator();
				XATQueue queue = ((XATStreamOperator) child).getOutputQueue();
				/** set the input queue.  If we are a multi queue operator, then add the queue */
				if(root instanceof XATMultiQueueStreamOperator){
					((XATMultiQueueStreamOperator) root).addInputQueue(queue);
				} else { 
					((XATStreamOperator) root).setInputQueue(queue);
				}
				System.out.println("connects operators" + name);
				connectQueues(child);
			} catch (XATNodeDoesNotExistException e){
			/**
			 this is thrown when an operator does not have
			 a child.  This shouldnt happen because we only iterate
			 through the available children, but this catch will prevent
			 the error from being propagated higher
			*/
			}
		}
	}

    return;
}



/**
 * This function calls the CQLParser and generates a query
 * from the filename listed in the FILE_NAME property.
 * @return XATTree
 * @exception QueryPlanGeneratorException 
 */
public XATTree makeQueryPlan() throws QueryPlanGeneratorException
{
//	String fileName = this.getProperty("FILE_NAME");
//	try {
//		return CQLParser.makeQueryPlan(fileName);
//	} catch (FileNotFoundException e) {
//		e.printStackTrace();
//	} catch (ParseException e) {
//		e.printStackTrace();
//	}
//	
	return null;
}


/**
 * Sets the properties and default values that are available for this
 * particular QueryPlanGenerator
 */
protected void setDefaultProperties() {
	super.setDefaultProperties();    
    /**
    * This is the name of the xml file to parse 
    */
    addProperty("FILE_NAME", "resources\\CQLParser\\input.txt"); 
    addProperty("QUERY_FILE_NAME", "query.txt");
}
/**
* sets the given property to the new value.  This method is responsible
* for validating the given property.  An exception is thrown if
* the value is not correct.
* Validation should be done in the following manner: 
* First the method should check if the object supports the given property by calling
* <tt>isValidPropertyName(String)<tt>.  If it is not a valid property name, then an exception should
* be thrown. <p>
*
* Once it is known that the object supports the property, the value must be validated.  Each implementor of
* XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
* For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
* has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
* to set a common type, it should be validated by the super class by calling <tt>super.setProperty(name, value)<tt>.  This
* will reduce duplicate code. <p>
*
* When it is determined the name / value pair is correct, the pair is updated using <tt>setNewPropertyValue()<tt> <p>
* <tt>XATStatisticsImp<tt> provides default validation by checking to see if the property's value is
* >=0.  propertyValues that are Strings will just be set.  If the property requires more strict constraints than this, 
* it is up to the subclass that created the property to validate it (as described above).
* @param name the name of the property to set
* @param value the new value of the property
* @throws Exception if the given property cannot be found or
* if the value is not valid for the given property.
*/
public void setProperty(java.lang.String name, java.lang.String value)
    throws java.lang.Exception {
    super.setProperty(name, value);
}
}
