/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Iterator;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNodeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.XATNodeDoesNotExistException;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.migration.probablisticjoinoperator.ChangingWBJoinOperatorImp;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.CEP.StreamCopyOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.spreadfire.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
//import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNodeImp;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperatorImp;

/**
 * Generate an XAT (or dag) from an XML file.  It is probably easiest to
 * work out the query plan on paper and then encode it in the XML.  
 * There are not many assumptions in the format of the XML file, but here they are:
 * <ol>
 * <li> The xml file does not encode both Stream and NonStream operators.  While
 * care has been taken to ensure this will work with both of them, it is not
 * logically correct to inter-twine them in one plan
 * <li> The XML file should not remove any elements even if they are unused.  That is,
 * the operator element should still have the children element even if it is a leaf.  Otherwise,
 * the xml parser will complain
 * </ol>
 *
 * More to come 
 * @author: Brad Pielech
 */
public class DistributedFromXMLFileQueryPlanGenerator extends QueryPlanGeneratorImp  {
	private Hashtable operatorTable;
	private Hashtable operatorOutputSchemaTable;
	private ClassVariableVisitor variableVisitor;
	
	private boolean hasSupportOperatorsAlready = false;
/**
 * FromXMLFileQueryPlanGenerator constructor comment.
 */
public DistributedFromXMLFileQueryPlanGenerator() {
	super();	
	operatorTable = new Hashtable();
	operatorOutputSchemaTable = new Hashtable();
	variableVisitor = new ClassVariableVisitor();
	
}
/**
 * This will connect the children of the given node to those
 * that were specificed in the XML document.  It iterates through
 * the Node, finds all elements with the id attribute, checks
 * the operator table to see if this dagNode exists.  If it does,
 * they are connected.  If the operator does not exist, then it is created
 * and the parent / child relationship is established.  The operator of the
 * other dag node will be added later 
 * @param children org.w3c.dom.Node
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode
 *//*
private void addChildren(Node children, DagNode node) {
	
	NodeList childList = children.getChildNodes();
	for (int i = 0; i < childList.getLength(); i++){
		Node next = childList.item(i);
		if(next.getNodeType() == Node.ELEMENT_NODE){
			NamedNodeMap attrs = next.getAttributes();
			Node idN = attrs.getNamedItem("id");
			String id = idN.getNodeValue();
			DagNode child;
			if(operatorTable.containsKey(id)){
				child = (DagNode)operatorTable.get(id);	
			} else{
				child = new DagNodeImp();
				operatorTable.put(id,child);
			}
			node.addOpChild(child);
			child.addParent(node);
		}
	}	
	}*/

	/**
	 * This will connect the children of the given node to those
	 * that were specificed in the XML document.  It iterates through
	 * the Node, finds all elements with the id attribute, checks
	 * the operator table to see if this dagNode exists.  If it does,
	 * they are connected.  If the operator does not exist, then it is created
	 * and the parent / child relationship is established.  The operator of the
	 * other dag node will be added later 
	 * @param children org.w3c.dom.Node
	 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode
	 *//*
	private void addStreams(Node children, DagNode node) {
	
		NodeList childList = children.getChildNodes();
		for (int i = 0; i < childList.getLength(); i++){
			Node next = childList.item(i);
			if(next.getNodeType() == Node.ELEMENT_NODE){
				NamedNodeMap attrs = next.getAttributes();
				Node idN = attrs.getNamedItem("name");
				String name = idN.getNodeValue();
				node.addStreamChild(name);
			}
		}	
	}
	


*/
//mjba 
private void addChildren(Node children, DagNode node) {
	
	NodeList childList = children.getChildNodes();
//	System.out.println("adds children");
	for (int i = 0; i < childList.getLength(); i++){
		Node next = childList.item(i);
		if(next.getNodeType() == Node.ELEMENT_NODE){
			NamedNodeMap attrs = next.getAttributes();
			Node idN=attrs.getNamedItem("type");
			String iD=idN.getNodeValue();
			if(iD.equals("operator")){
//				System.out.println("adds operator to xmlOC");
				DagNode child;
				Node idO=attrs.getNamedItem("id");
//				System.out.println("Operator id is:" + idO);
				String id=idO.getNodeValue();
			   if(operatorTable.containsKey(id)){
				child = (DagNode)operatorTable.get(id);	
			} else{
				child = new DagNodeImp();
				operatorTable.put(id,child);
			}

			/*
			 * Added by Song Wang, Oct. 2005, to support multiple
			 * outputqueue
			 */
			Attr idQ = (Attr) attrs.getNamedItem("queueId");
			if (idQ != null) {
				int value = Integer.parseInt(idQ.getValue());
				((XATStreamOperator) node.getStreamOperator())
						.addOutputQueueIdFromChildren(Integer.parseInt(id), value);
			} else {
				//by default, the child's 0 outputqueue will be
				// connected to the parent
				((XATStreamOperator) node.getStreamOperator())
						.addOutputQueueIdFromChildren(Integer.parseInt(id),0);

			}

			node.appendChild(child);
			//child.addParent(node);
		}
		if(iD.equals("stream")){
//			System.out.println("adds stream to xmlOC");
			Node idS=attrs.getNamedItem("name");
			String name=idS.getNodeValue();
//			System.out.println("Appending stream named " + name + " to children of operator " + node.getOperator().toString());
			node.appendChild(name);
			
		}
	}
	}
}




/**
 * This method will call the visit method in variableVisitor.  It is too
 * hard to generalize the creation of variables, so each operator
 * type is handled in its own visit method.  The visit method will
 * pass the operator and the Node (DOM, not XAT node) and each
 * visit method can be customized per operator type.  
 * @param children org.w3c.dom.Node
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode
 */
private void addClassVariables(Node classVariables, DagNode node) {
	variableVisitor.visit(node.getOperator(), classVariables);	
	}

private void assignSchema(Node schema, DagNode node, String opID) {
	XATQueueSchema outSchema = new XATQueueSchema();

	NodeList nodeList = schema.getChildNodes();
	
	// if schema is not defined in config file (nodelist.getLenth() == 0),
	// the schema of the output queue is set by combining input queue schemas 
	// in the run() method of each operator.
	if(nodeList.getLength() > 0) {
		for (int i = 0; i < nodeList.getLength(); i++){
			Node next = nodeList.item(i);
			if(next.getNodeType() == Node.ELEMENT_NODE){
				NamedNodeMap attrs = next.getAttributes();
				String name = ((Node) attrs.getNamedItem("name")).getNodeValue();
				if (name.equals("col_name")) {
					outSchema.addColumn(((Node) attrs.getNamedItem("value")).getNodeValue());
				} else {
					System.out.println("Wrong configuration!");
				}
			}
		}
		operatorOutputSchemaTable.put(opID, outSchema);
	}
}

/**
 * Parses the parents element and connects parents.  If the parent
 * does not exist in the hash table, then just create a new DagNode with
 * the corresponding ID.  The operator will be connected later
 * Creation date: (2/1/03 12:53:59 PM)
 * @param parents org.w3c.dom.Node
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode
 */
private void addParents(Node parents, DagNode node) {
	

	NodeList parentList = parents.getChildNodes();
	for (int i = 0; i < parentList.getLength(); i++){
		Node next = parentList.item(i);
		if(next.getNodeType() == Node.ELEMENT_NODE){
			NamedNodeMap attrs = next.getAttributes();
			Node idN = attrs.getNamedItem("id");
			String id = idN.getNodeValue();
			DagNode parent;
			if(operatorTable.containsKey(id)){
				parent = (DagNode)operatorTable.get(id);	
			} else{
				parent = new DagNodeImp();
				operatorTable.put(id,parent);
			}
			node.addParent(parent);
			//parent.addOpChild(node);
			//parent.appendChild(node);
		}
	}
	// Defaultly, each node only has 1 parent
	// added by Mingzhu, when there are multiple parent nodes, add parentLenth-1 outputqueue
	//to make the number of outputQueue equal to the number of parents 
	XATStreamOperator op = node.getStreamOperator();
	
	
	
//	if (op instanceof StreamCopyOperatorImp)
//	{
//		//System.out.println("current number of outqueue: " + op.getNumberOfOutputQueue());
//		System.out.println("copy op "+ op.toString() + " ID " + op.getOperatorID());
//		System.out.println("See if copyoperator has correct parents and output queue");
//		System.out.println("copy's parent number " + parentList.getLength());		
//		// Iterate each parent, if parent is a node, increase the nubmer of outptut Queue
//		for (int j = 0; j < parentList.getLength(); j++){
//			Node next = parentList.item(j);
//			if(next.getNodeType() == Node.ELEMENT_NODE){
//				NamedNodeMap attrs = next.getAttributes();
//				Node idN = attrs.getNamedItem("id");
//				String id = idN.getNodeValue();
//				System.out.println(" parent node Id " + id.toString());
//				DagNode parent;
//				if(operatorTable.containsKey(id))
//					parent = (DagNode)operatorTable.get(id);
//				op.addNewOutputQueue();
//			// update the nubmer of outputQueue
//				//System.out.println("current number of outqueue: " + op.getNumberOfOutputQueue());
//				}
//			
//		}
//		System.out.println("COPY output queues number: " + op.getNumberOfOutputQueue());
//	
//	}
	
}
/**
 * Adds the properties of the DagNode as specified in the XML document
 * This is pretty straight forward- parse the properties element,
 * get the name and value attributes, then set the property
 * @param children org.w3c.dom.Node
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode
 */
private void addProperties(Node properties, DagNode node) throws Exception {
	
	NodeList propertiesList = properties.getChildNodes();
	for (int i = 0; i < propertiesList.getLength(); i++){
		Node next = propertiesList.item(i);
		if(next.getNodeType() == Node.ELEMENT_NODE){
			NamedNodeMap attrs = next.getAttributes();
            String name = ((Node) attrs.getNamedItem("name")).getNodeValue();
            String value = ((Node) attrs.getNamedItem("value")).getNodeValue();
            node.getOperator().setProperty(name, value);
		}
	}
	
}
/**
 * If we are working with XATStreamOperators, we can connect all of the
 * operators together with this recursive method 
 * @param XATStreamOperator root.  
 */

/*private void connectQueues(XATStreamOperator root) throws Exception {

	//Make input queues for the streams if any are going to be connected
	for(int j=0; j<root.getNode().getStreamChildSize();j++){
		String sName = root.getNode().getStreamChildName(j);
		System.out.println("Stream whose queue to connect is: " + sName);
		if(root instanceof XATStreamOperator){
			Constructor[] c1 = Class.forName(this.datamodel).getConstructors();
			Class[] cl;
			Constructor construct;
			int count = 0;
			
			do{
			
			construct = c1[count];
			cl = construct.getParameterTypes();
			count++;
			}while(cl.length != 1);
			
			XATQueue q = (XATQueue) construct.newInstance(new Object[] {root});	
			q.setQueueName(sName);
			System.out.println("add input queue for " + sName);
			root.addInputQueue(q);	
			System.out.println("Input queue nimber is: " + root.getInputQueuePosition(q));
		}
	} 

    XATNode[] children = root.getNode().getOpChildren();
    for (int i = 0; i < children.length; i++) {
	    try{
        XATNode node = children[i];
        XATStreamOperator child = (XATStreamOperator) node.getOperator();
        XATQueue queue = ((XATStreamOperator) child).getOutputQueue();
        *//** set the input queue.  If we are a multi queue operator, then add the queue *//*
        if(root instanceof XATMultiQueueStreamOperator){
			((XATMultiQueueStreamOperator) root).addInputQueue(queue);
        } else { 
         ((XATStreamOperator) root).setInputQueue(queue);
        }

        connectQueues(child);
    	} catch (XATNodeDoesNotExistException e){
			*//**
			 this is thrown when an operator does not have
			 a child.  This shouldnt happen because we only iterate
			 through the available children, but this catch will prevent
			 the error from being propagated higher
			*//*
    	}
	}

    return;
}
*/


//mjba
private void connectQueues(XATStreamOperator root) throws Exception {
	//added by song wang for suport of cyclic graph traversal
	if (((DagNode)root.getNode()).hasBeenVisited()) return;
	((DagNode)root.getNode()).setHasBeenVisited(true);// pre-order setting

		//Added by Song Wang for support of queue connection in a graph query
		// plan
		if (((XATStreamOperatorImp) root).connectQueuesDone) {
			//already done with the subtree rooted at the root node, just
			// return
			return;
		}

	//Make input queues for the streams if any are going to be connected
	for(int j=0; j<root.getNode().getChildrenSize();j++){
//		System.out.println("children size is : " + root.getNode().getChildrenSize());
//		System.out.println("j in for loop is:" + j);
		Object name = root.getNode().getChild(j);
		if(name instanceof String){
//			System.out.println("name is stream: " + name);
			if(root instanceof XATStreamOperator){
				Constructor[] c1 = Class.forName(this.datamodel).getConstructors();
				Class[] cl;
				Constructor construct;
				int count = 0;
				
				do{				
					construct = c1[count];
					cl = construct.getParameterTypes();
					count++;
				} while(cl.length != 1);
				
				XATQueue q = (XATQueue) construct.newInstance(new Object[] {root});	
				String sName= (String) name;
				q.setQueueName(sName);
				System.out.println("queue name is " + sName + " which is (name): " + name);
				root.addInputQueue(q);		
			}
		} 
		if(name instanceof XATNode){
			System.out.println("name is XATNode: " + name);
			try{
				XATNode node = (XATNode) name;
				XATStreamOperator child = (XATStreamOperator) node.getOperator();
				for (int i=0;i<root.getOutputQueueIdFromChild(child.getOperatorID()).size();i++){
					XATQueue queue = ((XATStreamOperator) child)
							.getOutputQueue(((Integer)root.getOutputQueueIdFromChild(child.getOperatorID()).get(i)).intValue());

				/** set the input queue.  If we are a multi queue operator, then add the queue */
				if(root instanceof XATMultiQueueStreamOperator){
					((XATMultiQueueStreamOperator) root).addInputQueue(queue);
				} else { 
					((XATStreamOperator) root).setInputQueue(queue);
				}
				System.out.println("connects operators" + name);
				}
				connectQueues(child);
			} catch (XATNodeDoesNotExistException e){
			/**
			 this is thrown when an operator does not have
			 a child.  This shouldnt happen because we only iterate
			 through the available children, but this catch will prevent
			 the error from being propagated higher
			*/
			}
		}
	}

	//Added by Song Wang for support of queue connection in a graph query
	// plan
	//showing that this node and the subtree has been visited.
	((XATStreamOperatorImp) root).connectQueuesDone = true;
    return;
}


/**
 * Traverses the XML document and finds an element by a given name with the given sibling.
 * If the first element encountered does not have the same name as the parameter name, an exception is thrown
 * @return org.w3c.dom.Node
 * @param name java.lang.String
 * @param sibling org.w3c.dom.Node
 */
private Node getNodeByName(String name, Node sibling) throws IOException{
	while (sibling != null) {
                /**
                * check the nodeName, should = edges
                * if it doesnt, then throw exception
                */
                if (sibling instanceof Element) {
//                	System.out.println("sibling.getNodeName() = " + sibling.getNodeName());
//                	System.out.println("name = " + name);
                    if(!sibling.getNodeName().equals(name)) {
                        throw new IOException("The XML document does not conform to the default Schema");
                    }
                    /** now we've found the edges element*/
                    return sibling;
                }
                sibling = sibling.getNextSibling();
            }
	throw new IOException("The element was not found");
}
/**
 * Parses an XML document and returns an XATTree data structure.
 * This will first parse the file, find all operator elements, then create
 * operators one by one.  Then the data models (table or queue) will be connected
 *
 * Dags will be created if the dag attribute of the query plan element is set to true.  Otherwise,
 * XATNodes will be created.
 * @return XATTree
 * @exception QueryPlanGeneratorException 
 */
public XATTree makeQueryPlan() throws QueryPlanGeneratorException {
    try {
        String key = "javax.xml.parsers.DocumentBuilderFactory";
        String value = "org.apache.xerces.jaxp.DocumentBuilderFactoryImpl";
        //set the DocumentBuilderFactory to use
        java.lang.System.setProperty(key, value);
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        //DocumentBuilderFactory factory = new DocumentBuilderFactoryImpl();
        DocumentBuilder builder = factory.newDocumentBuilder();
        Document document = builder.parse(this.getProperty("FILE_NAME"));
		
        NodeList operators = document.getElementsByTagName("operator");
        for (int i = 0; i < operators.getLength(); i++) {
            parseOperator(operators.item(i));
        }
        
        doSpreadfirePlanMigration();

		//Added by Song for support of graph query plan
		//set all the sign to false
        this.getDag().clearHasBeenVisitedMark();
		for (Iterator i = getDag().getRoots(); i.hasNext();) {
			clearConnectQueueSign((XATNode) i.next());
		}

		/** 
         if we have XATStreamOperators, then we will connect the input and output
         queues together with this method.
        */
        this.getDag().clearHasBeenVisitedMark();
		for (Iterator i = getDag().getRoots(); i.hasNext();) {
			XATOperator operator = ((XATNode) i.next()).getOperator();
			if (operator instanceof XATStreamOperator) {
				connectQueues((XATStreamOperator) operator);
			}
			for (Iterator j = ((XATStreamOperator) operator)
					.getOutputQueues(); j.hasNext();)
				((XATQueue) j.next()).setAsFinalQueue(true);
		}

        //Added by Song for support of graph query plan
		//set all the sign to false
        this.getDag().clearHasBeenVisitedMark();
		for (Iterator i = getDag().getRoots(); i.hasNext();) {
			clearConnectQueueSign((XATNode) i.next());
		}
        this.getDag().clearHasBeenVisitedMark();

		// for each operator, assign schema to its output queue
		Enumeration en = operatorOutputSchemaTable.keys();
		while (en.hasMoreElements()) {
			Object opID = en.nextElement();
			XATQueueSchema outSchema = (XATQueueSchema)operatorOutputSchemaTable.get(opID);
			((XATStreamOperator)
					((DagNode)operatorTable.get(opID)).getStreamOperator())
					.getOutputQueue().setSchema(outSchema);
		}
		
		
//		XATNode[] nodes = getDag().convertToArray();
//		for (int i=nodes.length-1; i>=0; i--) {
//			if (nodes[i].getOperator() instanceof XATStreamOperator) {
//				((XATStreamOperator)nodes[i].getOperator()).init();
//				System.out.println("Init operator "+ nodes[i].getOperator());
//			}
//		}
        return getDag();
    } catch (Exception e) {
    	e.printStackTrace();
        throw new QueryPlanGeneratorException(e.toString());
    }
}
	/**
	 * Clear the connectQueueDone sign in the XATStreamOperatorImp
	 * 
	 * @param root
	 */
	private void clearConnectQueueSign(XATNode root) {
		if (((DagNode)root).hasBeenVisited()) return;
		((DagNode)root).setHasBeenVisited(true);// pre-order setting
		((XATStreamOperatorImp) (root.getOperator())).connectQueuesDone = false;
		for (int i = 0; i < root.getChildrenSize(); i++) {
			Object name = root.getChild(i);
			if (name instanceof XATNode) {
				clearConnectQueueSign((XATNode) root.getChild(i));
			}
		}
		
	}

/**
 * Takes an element as a parameter, parses the element, creates
 * an XATNode and operator, connects parents and children (if possible)
 * and then stores the operator in the operatorTable
 * @param operatorElement org.w3c.dom.Node
 */
//@SuppressWarnings("unchecked")
private void parseOperator(Node operatorElement) throws Exception {
		
    /** first make sure we were given an element */
    if (operatorElement.getNodeType() != Node.ELEMENT_NODE) {
        throw new DOMException(
            DOMException.INVALID_ACCESS_ERR,
            "The parameter was not a DOM Element");
    }
    /** next make sure we were given an operator element */
    if (!operatorElement.getNodeName().equalsIgnoreCase("operator")) {
        throw new DOMException(
            DOMException.INVALID_ACCESS_ERR,
            "The parameter was not an operator Element");
    }
    NamedNodeMap nodeAttributes = operatorElement.getAttributes();

    Attr nodeIdentifier = (Attr) nodeAttributes.getNamedItem("id");
    if (nodeIdentifier == null) {
        throw new DOMException(
            DOMException.SYNTAX_ERR,
            "The ID attribute was not set properly");
    }
        
    String id = nodeIdentifier.getValue();
    DagNode dagNode;
    /** 
     * first check the hashtable to see if we have a DagNode matching this operator's ID 
     * if it was, grab it and we will use it, otherwise, create a new one.
     */
    if (operatorTable.containsKey(id)) {
        dagNode = (DagNode) operatorTable.get(id);
    } else {
        dagNode = new DagNodeImp();
        operatorTable.put(id, dagNode);
    }

    /** see if this operator is a root or not */
    nodeIdentifier = (Attr) nodeAttributes.getNamedItem("root");
    Boolean isRoot = new Boolean(nodeIdentifier.getValue());
    if (isRoot.booleanValue()) {
        getDag().addRoot(dagNode);
    }

    /** first get the className attribute */
    nodeIdentifier = (Attr) nodeAttributes.getNamedItem("className");
    if (nodeIdentifier == null) {
        throw new DOMException(
            DOMException.SYNTAX_ERR,
            "The className attribute was not set properly");
    }
    String className = nodeIdentifier.getValue();

    //BEGIN SPREADFIRE EXTENSION
    if (className.indexOf("spreadfire") >= 0 
    		&& (className.indexOf("Split") >= 0
    				|| className.indexOf("Merge") >= 0))
    	this.hasSupportOperatorsAlready = true;
    

    //END SPREADFIRE EXTENSION
    
    //Class[] c1 = new Class[] { DagNode.class};
    //Constructor construct = Class.forName(className).getConstructor(c1);
   /* Constructor[] c1 = Class.forName(className).getConstructors();
    Constructor construct = c1[0];
    XATOperator operator = (XATOperator) construct.newInstance(new Object[] { dagNode });

    ((XATStreamOperator)operator).setOperatorID(Integer.parseInt(id));
*/
    
    //Constructor construct = Class.forName(className).getConstructor(c1);
    Constructor[] c1 = Class.forName(className).getConstructors();
    Constructor construct;// = c1[0];
    System.out.print("Constructor length "+ c1.length);
    
    //mjba to fix the problem with calling the wrong constructor
    boolean right_constructor=false;
    int temp=0;
    XATOperator operator=null;
    while (!right_constructor&& temp<c1.length){
    	construct=c1[temp];
    	try{
    	   operator = (XATOperator) construct.newInstance(new Object[] { dagNode });
    	   right_constructor=true;		   
    	}catch(IllegalArgumentException e){
    		temp++;
    		
    	}
    }
    //XATOperator operator = (XATOperator) construct.newInstance(new Object[] { dagNode });
    
    if((operator instanceof ChangingWBJoinOperatorImp)) {
    	if(isRoot.booleanValue()) {
    		((ChangingWBJoinOperatorImp)operator).setAsRootOperator(true);
        }
    	((ChangingWBJoinOperatorImp)operator).setDag(getDag());
    }
    
    
    dagNode.setOperator(operator);
	((XATStreamOperator) operator).setOperatorID(Integer.parseInt(id));
	/**
	 * Second get the numberOfOutputQueue attribute This attibute is
	 * optional, by default, the numberOfOutputQueue of an operator is 1
	 * Added by Song Wang on Oct, 2005 to support multiple outputQueues for
	 * an operator
	 *  
	 */
	nodeIdentifier = (Attr) nodeAttributes
			.getNamedItem("numberOfOutputQueue");
	if (nodeIdentifier != null) {
		int value = Integer.parseInt(nodeIdentifier.getValue());
		((XATStreamOperator) operator).setNumberOfOutputQueue(value);
		if (value > 1) {
			for (int i = 1; i < value; i++)
				//by default the outputqueue at position 0 has been
				// initialized
				//the adding of outputqueue begins from position 1.
				((XATStreamOperator) operator).addOutputQueue(i);
		}
	}

	/** 
	 set the data model for this operator.
	 This is a small hack, it assumes it is a stream operator.  This will
	 have to be altered if it is an XATTable
	*/
    
    /** now get the classVariables, properties, parents, and children of the operator */
    Node nextNode = operatorElement.getFirstChild(),
        classVariables = null,
        properties = null,
		schema = null,
        parents = null,
        children = null;
        //streams = null;
         
    
    classVariables = getNodeByName("classVariables", nextNode);
    nextNode = classVariables.getNextSibling();
    properties = getNodeByName("properties", nextNode);
    nextNode = properties.getNextSibling();
    schema = getNodeByName("schema", nextNode);
    nextNode = schema.getNextSibling();
    parents = getNodeByName("parents", nextNode);
    nextNode = parents.getNextSibling();
    children = getNodeByName("children", nextNode);
    
    
    
    //nextNode = children.getNextSibling();
	//streams = getNodeByName("streams",nextNode);
    

    	addChildren(children, dagNode);
    	addParents(parents, dagNode);
    	assignSchema(schema, dagNode, id);
    	addProperties(properties, dagNode);
    	addClassVariables(classVariables, dagNode);
    	//addStreams(streams,dagNode);

    
    return;
}

public void doSpreadfirePlanMigration()
{

    if(this.hasSupportOperatorsAlready) return;
    //BEGIN SPREADFIRE EXTENSION
    Hashtable newtable = (Hashtable)this.operatorTable.clone();
    Enumeration en = newtable.elements();
    while(en.hasMoreElements())
    {
        DagNodeImp dagNode = (DagNodeImp)en.nextElement();
        boolean isReplicatedOperator = false;
        String className = dagNode.getStreamOperator().getClass().getName();
        if (className.indexOf("spreadfire") >= 0 
            && className.indexOf("Split") < 0
            && className.indexOf("Merge") < 0
            && !this.hasSupportOperatorsAlready)
        isReplicatedOperator = true;
    
        DagNode mergeParentNode = null;
        DagNode[] splitChildrenNodes = null;
        if(isReplicatedOperator)
        {
        //Here we make the Three Musketeers, or rather,
        // the merge parent (Porthos)
        // the actual operator (Athos)
        // the split children (Aramis...ses)
            mergeParentNode = new DagNodeImp();
            mergeParentNode.setOperator(new SpreadfireMergeOperator(mergeParentNode));
        
            String parentID = null;
            do
            {
                parentID = Long.toString(Math.round(Math.random() * 65536.0));
            } while (operatorTable.containsKey(parentID));
            //DEBUG
            System.out.println("Query plan generator: inserting generated parent with ID " + parentID);
            //END DEBUG
            mergeParentNode.getStreamOperator().setOperatorID(Integer.parseInt(parentID));
            operatorTable.put(parentID, mergeParentNode);

            mergeParentNode.setOpChild(0, dagNode);
            //dagNode.setParent(mergeParentNode);
        
        //assignSchema(schema, mergeParentNode, parentID);        
            boolean isRoot = (dagNode.getParentsSize() == 0);
            
            XATNode tempParent = dagNode.getParent();
            dagNode.setParent(mergeParentNode);
            mergeParentNode.setParent(tempParent);
        
            if(tempParent != null)
            {
                tempParent.setOpChild(tempParent.getIndexOfChild(dagNode), mergeParentNode);
       
                ((XATStreamOperator) tempParent.getStreamOperator())
                .addOutputQueueIdFromChildren(Integer.parseInt(parentID),0);
            }
        
            ((XATStreamOperator) mergeParentNode.getStreamOperator())
            .addOutputQueueIdFromChildren(dagNode.getStreamOperator().getOperatorID(),0);
        
            if (tempParent == null)
            {
            //If the replicated operator is the root, we have to change the plan
            //  so its new parent is a root instead.
            //getDag().setRoot(mergeParentNode);
                Iterator it = getDag().getRoots();
                DagNode root;
                int i = 0;
                while(it.hasNext())
                {
                    root = (DagNode)it.next();
                    if(root.getStreamOperator().getOperatorID() == dagNode.getStreamOperator().getOperatorID())
                    {
                        getDag().setNthRoot(i, mergeParentNode);
                    }
                    else i++;
                }
                
            }
        
         System.out.println("node " + dagNode.getStreamOperator().getOperatorID() + " has "+ dagNode.getChildrenSize() + " children"); 
        splitChildrenNodes = new DagNode[dagNode.getChildrenSize()];
        for(int i = 0; i < splitChildrenNodes.length; i++)
        {
            String currentChildID = null;
            do
            {
                currentChildID = Long.toString(Math.round(Math.random() * 65536.0));
            } while (operatorTable.containsKey(currentChildID));
            //DEBUG
            System.out.println("Query plan generator: inserting generated child with ID " + currentChildID);
            //END DEBUG
            splitChildrenNodes[i] = new DagNodeImp();
            operatorTable.put(currentChildID, splitChildrenNodes[i]);
            
            ((XATStreamOperator) dagNode.getStreamOperator())
                .addOutputQueueIdFromChildren(Integer.parseInt(currentChildID),0);
            
            splitChildrenNodes[i].setOperator(new ReplicatedStreamSplitOperatorImp(splitChildrenNodes[i]));
            splitChildrenNodes[i].getStreamOperator().setOperatorID(Integer.parseInt(currentChildID));
            splitChildrenNodes[i].setParent(dagNode);
        }
        
        //here we may have a problem with addChildren.  it's expecting the full 
        //  Children node in the XML file, but we want to split it up.
        /**@TODO: figure out a fix for children of the split children*/
        // Proposed solution for fix:  get all real children of the op first,
        //  -Then assign each real child to a split op instead.

        for(int i = 0; i < dagNode.getChildrenSize(); i++)
        {
            if (dagNode.getChild(i) instanceof XATNode)
            {
                splitChildrenNodes[i].addOpChild((DagNode)dagNode.getChild(i));
                ((DagNode)dagNode.getChild(i)).setParent(splitChildrenNodes[i]);
                splitChildrenNodes[i].getStreamOperator().clearOutputQueueIdsFromChildren();
                ((XATStreamOperator) splitChildrenNodes[i].getStreamOperator())
                .addOutputQueueIdFromChildren(((DagNode)dagNode.getChild(i)).getStreamOperator().getOperatorID(),0);
 
            }
            else
                splitChildrenNodes[i].addStreamChild((String)dagNode.getChild(i));
        }
        dagNode.replaceChildren(splitChildrenNodes);
        System.currentTimeMillis();
    }
    }
    //END SPREADFIRE EXTENSION
    
}

/**
 * Sets the properties and default values that are available for this
 * particular QueryPlanGenerator
 */
protected void setDefaultProperties() {
	super.setDefaultProperties();    
    /**
    * This is the name of the xml file to parse 
    */
    addProperty("FILE_NAME", "queryplan.xml"); 
    addProperty("QUERY_FILE_NAME", "query.txt");
}
/**
* sets the given property to the new value.  This method is responsible
* for validating the given property.  An exception is thrown if
* the value is not correct.
* Validation should be done in the following manner: 
* First the method should check if the object supports the given property by calling
* <tt>isValidPropertyName(String)<tt>.  If it is not a valid property name, then an exception should
* be thrown. <p>
*
* Once it is known that the object supports the property, the value must be validated.  Each implementor of
* XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
* For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
* has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
* to set a common type, it should be validated by the super class by calling <tt>super.setProperty(name, value)<tt>.  This
* will reduce duplicate code. <p>
*
* When it is determined the name / value pair is correct, the pair is updated using <tt>setNewPropertyValue()<tt> <p>
* <tt>XATStatisticsImp<tt> provides default validation by checking to see if the property's value is
* >=0.  propertyValues that are Strings will just be set.  If the property requires more strict constraints than this, 
* it is up to the subclass that created the property to validate it (as described above).
* @param name the name of the property to set
* @param value the new value of the property
* @throws Exception if the given property cannot be found or
* if the value is not valid for the given property.
*/
public void setProperty(java.lang.String name, java.lang.String value)
    throws java.lang.Exception {
    super.setProperty(name, value);
}
}
