/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator;

import java.util.Enumeration;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
/**
 * Provides a way to create random graphs that have a certain set of properties.
 * In order to specify the type of random graphs to generate, the RandomGraphGenManager
 * should be queried.  
 * @author: Brad Pielech
 */
public interface QueryPlanGenerator extends XATQueryObject {

	public final static java.lang.String factoryID = "edu.wpi.cs.dsrg.xmldb.xat.component.queryplangenerator";
	/**
	* <pre>
	* creates N query plans according to the properties supported by the QueryPlanGenerator
	* <b>pre</b>:
	* the properties of the query plan have been initialized accordingly
	* <b>post</b>:
	* The Enumeration returned will consist of randomly generated IQueryPlans
	* </pre>
	* @return Enumeration of IQueryPlans
	* @param int howMany how many IQueryPlans to generate
	* @exception QueryPlanGeneratorException if either a property of the generator was not initialized or
	* if there is a conflict in values ex: max < min
	*/	
	Enumeration makeNQueryPlans(int howMany) throws QueryPlanGeneratorException;
	/**
	* <pre>
	* creates N query plans according to the properties supported by the QueryPlanGenerator
	* <b>pre</b>:
	* the properties of the query plan have been initialized accordingly
	* <b>post</b>:
	* The Enumeration returned will consist of randomly generated IQueryPlans
	* </pre>
	* @return Enumeration of IQueryPlans
	* @param int howMany how many IQueryPlans to generate
	* @exception QueryPlanGeneratorException if either a property of the generator was not initialized or
	* if there is a conflict in values ex: max < min
	*/	
	public abstract XATTree makeQueryPlan() throws QueryPlanGeneratorException;
	public void setDataModel(String model);
	public String getDataModel();
}
