/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.queuemanager;

/**
 * This class is simply a place-holder inside a queue, which represents
 * the place where tuples are stored on disk rather than in main memory.
 * The actual disk file is managed by the QueueManager.
 * 
 * Creation Date: (Jun 10, 2004 4:34:51 PM)
 * @author Natasha Bogdanova
 */
public class PlaceHolder {
	
	private int numTuples;
	
	/**
	 * PlaceHolder constructor.
	 * 
	 * Creation Date: (Jun 10, 2004 4:36:30 PM)
	 */
	public PlaceHolder() {
		numTuples = 0;
	}
	
	/**
	 * PlaceHolder constructor.
	 * 
	 * Creation Date: (Jun 10, 2004 4:38:23 PM)
	 * @param numTuplesInit int
	 */
	public PlaceHolder(int numTuplesInit) {
		numTuples = numTuplesInit;
	}
	
	/**
	 * Returns the number of tuples stored on disk for this queue.
	 * 
	 * Creation Date: (Jun 10, 2004 4:40:19 PM)
	 * @return int
	 */
	public int getNumTuples() {
		return this.numTuples;
	}
	
	/**
	 * Sets the number of tuples stored on disk. This is only for easy
	 * access to the information and does not have any actual effect on 
	 * how many tuples are stored on disk. All physical storage is handled
	 * by the QueueManager.
	 * 
	 * Creation Date: (Jun 10, 2004 4:41:37 PM)
	 * @param num int
	 */
	public void setNumTuples(int num) {
		this.numTuples = num;
	}
}
