/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;

/**
 * The Chain class represents a set of operators that are grouped together
 * by a common slope built from a <tt> ProgressChart </tt>.
 * Creation date: (4/15/2003 4:20:50 PM)
 * @author: Tim Sutherland
 * @version: 1.1.1
 */
public class Chain {

/** A vector of Operators in the Chain */	
private Vector operator;
/** The slope of the Chain Object */
private double slope;
/** Default Constructor. Creates an Empty Chain */
public Chain(){

	operator = new Vector();
	slope = 0;

}
/** Constructor. Creates a chain with the given array
 * of <tt>XATStreamOperators </tt>
 */
public Chain(XATStreamOperator[] xso){

	for(int i = 0 ; i < xso.length ; i++){
		operator.addElement(xso[i]);
	}
	slope = 0;

}
/**
 * Adds an operator to the Chain
 * Creation date: (4/17/2003 10:38:02 AM)
 * @param XATStreamOperator
 */
public void addOperator(XATStreamOperator xso) {
	operator.addElement(xso);
	}
/**
 * Returns the number of operators in the chain
 * Creation date: (4/20/2003 12:04:50 PM)
 * @return int- the number of operators
 */
public int getNumberOfOperators() {
	return operator.size();
}
/** Gets the operator at the specified location
 * @param int the location of the operator in the chain
 * @return XATStreamOperator- the operator at the specified location
 */
 
public XATStreamOperator getOperator(int location){

	return (XATStreamOperator) operator.elementAt(location);
}
/**
 * Returns the slope of the chain
 * Creation date: (4/15/2003 7:03:47 PM)
 * @return double- the slope of the chain
 */
public double getSlope() {
	return slope;
}
/**
 * Sets the slope of the chain
 * Creation date: (4/15/2003 7:03:47 PM)
 * @param newSlope double the new slope of the chain
 */
public void setSlope(double newSlope) {
	slope = newSlope;
}
/**
 * Returns a representation of the Chain as a string.
 * Creation date: (4/22/2003 6:11:42 PM)
 * @return java.lang.String- a representation of the Chain
 */
public String toString() {

	String s = "Chain: \nSlope- " + slope + "\nOperators- \n";

	String t = new String();

	for(int i = 0 ; i < getNumberOfOperators(); i++){

		t += getOperator(i).toString() + "\n";

	}
	return s + t;
}
}
