/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsReference;
import edu.wpi.cs.dsrg.eddy.operator.*;
/**
 * Schedules the operator with the most tuples in its input queues.  This scheduler
 * may not have the best throughput, and may be prone to starvation, but the intermediate
 * queues sizes should be smaller 
 * @author: Brad Pielech
 */
public class EddyScheduler extends XATScheduler {
/**
 * MostTuplesInQueueScheduler constructor comment.
 */
	private int index;
public EddyScheduler() {
	super();
}
/**
 * Insert the method's description here.
 * Creation date: (2/22/03 1:51:50 PM)
 */
private void init() {
	
	XATTree[] trees = getTrees();
	XATNode[] temp;
	XATNode EddyOp = null;
	int c = 0;
	int size=0;
	int eddySplitPos = 5;
	int firstEddyPos = 0;
	int Eddypt = 0;
	for(int i=0; i < trees.length; i++){
		size += getTree(i).size();
	}
	nodes = new XATNode[size+(int)(size/eddySplitPos)-1];

	Eddypt = (int)(size/((int)(size / eddySplitPos)));
	firstEddyPos =Eddypt;
		
	for(int i=0; i < trees.length; i++){
		temp = getTree(i).convertToArray();
		for(int j=0; j < temp.length;j++){
			if (Eddypt==c)
			{
				Eddypt+=Eddypt;
				c++;
			}	
			if (temp[j].getOperator() instanceof EddyOperatorImp)
				EddyOp = temp[j];
			else
			{	
				nodes[c] = temp[j];
				c++;
			}	
		}
	}
	Eddypt = (int)(size/((int)(size / eddySplitPos)));
	while(Eddypt <= (size+(int)(size/eddySplitPos)-1))
	{
		if (Eddypt == (size+(int)(size/eddySplitPos)-1))
			nodes[Eddypt-1]=EddyOp;
		else	
			nodes[Eddypt]=EddyOp;
		Eddypt+=Eddypt;		
	}
	//index = 0;
	// Set index this way so schduler starts from leaf node, instead of root node.
	index = firstEddyPos;
	}
/**
 * make the next scheduling decision.   
 * If the operator is chosen but has workload of 0, then
 * we will choose another. If we have checked all operators and there is still no work, then
 * we will break.  This prevents
 * starvation, that is, when the scheduler decides to run an operator with no tuples in its
 * input buffer.  
 * Creation date: (1/22/03 4:05:38 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision
 * @throws XATSchedulerException when there is no work to do
 */
public synchronized XATSchedulerDecision makeNextDecision() throws XATSchedulerException {
    XATSchedulerDecision decision = new XATSchedulerDecision();
    XATStreamOperator operator;
    int totalTuples = 0;
    int count = 0;

	//operator = (XATStreamOperator)nodes[index].getOperator();
	
	//while (!operator.isActive()) {
	//	index--; count++;
	//	if(index < 0){
	//	  index = nodes.length-1;
	//	}
	//	operator = (XATStreamOperator)nodes[index].getOperator();

		// Throw an exception if we have gone through all of the
		// operators and all are inactive.
	//	if(count > nodes.length) throw new XATSchedulerException();
		
	//}
	
	do{
		try{
			while (nodes == null || nodes.length <= index) {
				Thread.sleep(1000);
			}
		}catch(Exception e) {
			e.printStackTrace();
			System.exit(-1);
		}
		operator = (XATStreamOperator)nodes[index].getOperator();
			
		//index++; count++;
		//if(index >= nodes.length){
		//	index = 0;
		//}	
		index--; count++;
		if(index < 0){
			index = nodes.length-1;;
		}	
				
		// Throw an exception if we have gone through all of the
		// operators and all are inactive.
		//System.out.println("All operators are inactive");
		if(count > nodes.length) throw new XATSchedulerException();
				
	} while(!operator.isActive());

    int workload = calculateWorkLoad(operator);
     	        
	decision.setNextOperator(operator);		
    decision.setNumberOfTuples(workload);
        
    return decision;		  
}

public int NumberOfActivateOperators(){
    XATStreamOperator operator;
    int count = 0;

    for (int i=0; i<nodes.length; i++){
    	operator = (XATStreamOperator)nodes[i].getOperator();
    	if (operator.isActive()) {
    		count++;
    	}
    }
    return count;
}

public String getAllOperators(){
	String ret = "";
	XATStreamOperator operator;
    for (int i=0; i<nodes.length; i++){
    	operator = (XATStreamOperator)nodes[i].getOperator();
    	ret += i + ":" + operator + "\n";
    }	
    return ret;
}

public int getIndex(){
	return index;
}

/**
 * Insert the method's description here.
 * Creation date: (2/22/03 1:46:59 PM)
 * @param newTree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public void setTree(edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree newTree) {
	super.setTree(newTree);
	init();
}

public void setTrees(edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree[] newTree) {
	super.setTrees(newTree);
	init();
}
/**
 * Insert the method's description here.
 * Creation date: (2/26/2003 6:22:23 PM)
 * @return java.lang.String
 */
public String toString() {
	String rs = new String();

	String sName = this.getClass().getName();
	//find the last dot.
	int index = 0;
	for(index = sName.length()-1; index >=0; index--)
		if(sName.charAt(index) == '.')
			break;
	rs = sName.substring(index + 1) + " " + hashCode(); // + "  data:";	
    return rs;
}
}
