/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import java.util.ArrayList;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;

/**
 * This is a GreedyScheduler that will run the operator with the highest
 * GREEDY_PRIORITY.  This priority is based on the ratio between (1 - SELECTIVITY) and
 * AVERAGE_TUPLE_PROCESSING_TIME.  This method may be prone to starvation because the operator
 * with the highest priority may not have any work to do.  
 */
public class GreedyPriorityScheduler extends XATScheduler {
/**
 * GreedyPriorityScheduler constructor comment.
 */
public GreedyPriorityScheduler() {
	super();
}
/**
 * This is called by the Execution Controller.  This method decides
 * which operator to run next and how many tuples that operator should dequeue.
 * This will just pick the operator with the highest priority.  If that operator
 * does not have any work to do, then it will pick the 2nd and so on.  If nothing
 * has any work to do, then throw the exception 
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision
 */
public XATSchedulerDecision makeNextDecision() throws XATSchedulerException {

    XATSchedulerDecision decision = new XATSchedulerDecision();
    ArrayList operatorsChecked = new ArrayList();
    int totalTuples = 0;
    int length = this.getTree().convertToArray().length;
    for (int i = length - 1; i >= 0; i--) {
        /** get the next highest priority operator*/
        XATStreamOperator operator =
            (XATStreamOperator) StatisticsGatherer.getStatisticsGather().getQueryObject(
                "GREEDY_PRIORITY", i);
        int workload = calculateWorkLoad(operator);
        if (workload > 0 && operator.isActive()) {
            decision.setNextOperator(operator);
            decision.setNumberOfTuples(workload);
            return decision;
        }
    }

    // if we looked at all operators in the for loop, then there is nothing left to do 
	throw new XATSchedulerException("Looked at all operators and there is no work left to do");
          
}
}
