/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsReference;
/**
 * Schedules the operator with the most tuples in its input queues.  This scheduler
 * may not have the best throughput, and may be prone to starvation, but the intermediate
 * queues sizes should be smaller 
 * @author: Brad Pielech
 */
public class MostTuplesInQueueScheduler extends XATScheduler {
/**
 * MostTuplesInQueueScheduler constructor comment.
 */
public MostTuplesInQueueScheduler() {
	super();
}
/**
 * This is called by the Execution Controller.  This method decides
 * which operator to run next and how many tuples that operator should dequeue.
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision
 */
public XATSchedulerDecision makeNextDecision() throws XATSchedulerException {
    XATSchedulerDecision decision = new XATSchedulerDecision();
    int length = this.getTree().convertToArray().length;
    int operatorsChecked = 0;
    while (true) {
        XATStreamOperator operator =
            (XATStreamOperator) StatisticsGatherer.getStatisticsGather().getQueryObject(
                "NUMBER_OF_TUPLES_IN_INPUT_QUEUES",
                StatisticsReference.HIGHEST, true);
        operatorsChecked++;
        int workload = calculateWorkLoad(operator);
        if (workload > 0 && operator.isActive()) {
            decision.setNextOperator(operator);
            decision.setNumberOfTuples(workload);
            return decision;
        } else {
            /** 
             * if there is no work, then check to see if we have looked at all operators
             * if we have, then break
             */
            if (operatorsChecked >= length) {
	            int temp = 0;
                throw new XATSchedulerException("No more work to do");
            }
        }
    }
}
}
