/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.ConfigFileParser;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.configuration.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
/**
 * This is the default execution adapter.  It will always stick with
 * the same initial algorithm regardless of how well it performs.  
 * @author: Brad Pielech
 */
public class NeverRotateAdapter extends XATExecutionAdapter {
/**
 * NeverRotateAdapter constructor comment.
 * @param parser edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.ConfigFileParser
 * @param tree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public NeverRotateAdapter(ConfigFileParser parser, edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree tree) {
	super(parser, tree);
}

public NeverRotateAdapter(Machine config, XATTree[] tree){
	super(config, tree);
}
/**
 * Periodically the execution controller will request the Adapter to
 * change the algorithm b/c the previous one was not working well (ex: starvation).
 * Therefore, this method is similar to getNextScheduler, but getNext could result
 * in the same scheduler being used as the previous.  This method will guarentee a different
 * algorithm <p>
 * This can create a problem if there is only 1 algorithm available.  Therefore, it throws
 * an IllegalStateException if a switch is not possible. 
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler
 * @exception IllegalStateException if there is only 1 scheduler and a change is not possible
 */
public XATScheduler forceChange() throws java.lang.IllegalStateException {
	throw new IllegalStateException("Cannot switch algorithms because this strategy does not support changing");
}
/**
 * Returns the index of the Scheduling algorithm that should be used next.
 * Each subclass will override this method b/c this is where the actual logic of switching 
 * takes place.  This method updates the currentCategory class variable and then
 * return the next scheduler using getCurrentScheduler().  
 * This is how schedulers are referenced internally.
 * @return XATScheduler the next algorithm to use.  
 */
public XATScheduler getNextScheduler() {	
	return getCurrentScheduler();
}

/* (non-Javadoc)
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject#getLightWeightQueryObject()
 */
public LightWeightQueryObject getLightWeightQueryObject() {
	return new LightWeightQueryObject("NeverRotate",1,this.getClass().getName(), true);
}
}
