/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import java.util.ArrayList;
import java.util.Random;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsReference;
/**
 * This is a GreedyScheduler that will run each operator with a probibility equal to its 
 * GREEDY_PRIORITY.  This priority is based on the ratio between (1 - SELECTIVITY) and
 * AVERAGE_TUPLE_PROCESSING_TIME.  This method may be prone to starvation because the operator
 * with the chosen may not have any work to do.  
 */
public class ProbabilisticGreedyScheduler extends XATScheduler {
	private java.util.Random rand;
/**
 * GreedyPriorityScheduler constructor comment.
 */
public ProbabilisticGreedyScheduler() {
	super();
	rand = new Random();
}
/**
 * This is called by the Execution Controller.  This method decides
 * which operator to run next and how many tuples that operator should dequeue.
 * This will just pick the operator with the highest priority.  Starvation may
 * occur, but it is up to the Execution Controller to detect this and change the
 * algorithm 
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision
 */
public XATSchedulerDecision makeNextDecision() throws XATSchedulerException {

    XATSchedulerDecision decision = new XATSchedulerDecision();
    ArrayList operatorsChecked = new ArrayList();
    int totalTuples = 0;
    XATNode[] nodes = this.getTree().convertToArray();
	StatisticsGatherer stats = StatisticsGatherer.getStatisticsGather();
    double count = 0;
    //sum the scores
    try {    	    
    for (int i = 0; i < nodes.length; i++) {
        XATStreamOperator next = (XATStreamOperator) nodes[i].getOperator();
        count += Double.parseDouble(stats.getValue("GREEDY_PRIORITY", next, StatisticsReference.AVERAGE));
    }
    } catch (Exception e) {
    	System.out.println(e.toString());
    }
	//System.out.println("Priority Count: " + count);
    //get the random variable between 0 and count
    double nextValue = rand.nextDouble() * count;
    count = 0;
    XATStreamOperator operator = null;
    boolean keepGoing = true;
    boolean activeOperator = false;
    
    int i = 0;    
    
    while(!activeOperator){
    
    	//now walk over the scores again and see which one makes the sum greater than our target
    	while(keepGoing){
        	operator = (XATStreamOperator) nodes[i].getOperator();
        	count += Double.parseDouble(stats.getValue("GREEDY_PRIORITY", operator, StatisticsReference.AVERAGE));
        	if (count >= nextValue) {
            	keepGoing = false;
        	}
        	i++;
    	}
    activeOperator = operator.isActive();

    }
	int workload = calculateWorkLoad(operator);
	/** 
	 * with greedy, if there is no work for an operator to do,
	 * that is ok b/c greedy is inherently prone to starvation 
	 */
	decision.setNextOperator(operator);
	decision.setNumberOfTuples(workload); 
    return decision;
}
}
