/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;

/**
 * Represents a point on a progress chart
 * Creation date: (4/15/2003 4:20:28 PM)
 * @author: Tim Sutherland
 */
public class ProgressChartPoint {

/** The operator associated with this point */
private XATStreamOperator operator;
/** The selectivity of the operator for this object */
private double selectivity;
/** The APT time for the operator */
private double avgProcessTime;

/** The selectivity of this point relative to the points before it */
private double relativeSelectivity;
/** The processing time of this point relative to other points before it */
private double relativeAverageProcessTime;
/**
 * Default Constructor
 */
public ProgressChartPoint() {
	this.operator = null;
	this.selectivity = 0;
	this.avgProcessTime = 0;
	
}
/**
 * Returns the AvgProcessTime of the operator at this point.
 * Creation date: (4/15/2003 6:18:25 PM)
 * @return double
 */
public double getAvgProcessTime() {
	return avgProcessTime;
}
/**
 * Returns the operator at this point.
 * Creation date: (4/15/2003 6:18:25 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public XATStreamOperator getOperator() {
	return operator;
}
/**
 * Returns the relative processing time of this point.
 * Creation date: (4/17/2003 10:43:22 AM)
 * @return double
 */
public double getRelativeAverageProcessTime() {
	return relativeAverageProcessTime;
}
/**
 * Returns the relative selectivity of this point.
 * Creation date: (4/17/2003 10:43:22 AM)
 * @return double
 */
public double getRelativeSelectivity() {
	return relativeSelectivity;
}
/**
 * Returns the selectivity of the operator at this point
 * Creation date: (4/15/2003 6:18:25 PM)
 * @return double
 */
public double getSelectivity() {
	return selectivity;
}
/**
 * sets the AvgProcessTime of this point
 * Creation date: (4/15/2003 6:18:25 PM)
 * @param newAvgProcessTime double
 */
public void setAvgProcessTime(double newAvgProcessTime) {
	avgProcessTime = newAvgProcessTime;
}
/**
 * Sets the operator of this point
 * Creation date: (4/15/2003 6:18:25 PM)
 * @param newNode edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public void setOperator(XATStreamOperator newOperator) {
	this.operator = newOperator;
}
/**
 * Sets the Relative process time of this point
 * Creation date: (4/17/2003 10:43:22 AM)
 * @param newRelativeAverageProcessTime double
 */
public void setRelativeAverageProcessTime(double newRelativeAverageProcessTime) {
	relativeAverageProcessTime = newRelativeAverageProcessTime;
}
/**
 * Sets the relative selectivity of this point
 * Creation date: (4/17/2003 10:43:22 AM)
 * @param newRelativeSelectivity double
 */
public void setRelativeSelectivity(double newRelativeSelectivity) {
	relativeSelectivity = newRelativeSelectivity;
}
/**
 * Sets the selectivity based on the operator for this point.
 * Creation date: (4/15/2003 6:18:25 PM)
 * @param newSelectivity double
 */
public void setSelectivity(double newSelectivity) {
	selectivity = newSelectivity;
}
/**
 * Returns the point as a string
 * Creation date: (4/22/2003 5:39:01 PM)
 * @return java.lang.String
 */
public String toString() {
	return
		"Sel: " + this.getSelectivity() + "ATP: " + this.getAvgProcessTime() + "Rel-ATP: " + this.getRelativeAverageProcessTime() + " Rel-Sel: " +
			this.getRelativeSelectivity() + " Operator: " + this.getOperator().toString();
}
}
