/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

/**
 * Insert the type's description here.
 * Creation date: (6/23/2003 11:32:43 AM)
 * @author: Tim Sutherland
 */
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;

public class RoundRobinToChainAdapter extends XATExecutionAdapter {
/** The time in milliseconds for the round robin algorithm to run. */	
private int roundRobinRunTime = 0;
	private long startTime;
/**
 * RoundRobinToChainAdapter constructor comment.
 * @param parser edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.ConfigFileParser
 * @param inTree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public RoundRobinToChainAdapter(edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.ConfigFileParser parser, edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree inTree) {
	super(parser, inTree);
    startTime = System.currentTimeMillis();
	roundRobinRunTime = (new Integer(this.getProperty("RoundRobinRunTime")).intValue());
	System.out.println("Will run Round Robin for " + roundRobinRunTime + " milliseconds, and then go to chain.");
	
	}
/**
 * Periodically the execution controller will request the Adapter to
 * change the algorithm b/c the previous one was not working well (ex: starvation).
 * Therefore, this method is similar to getNextScheduler, but getNext could result
 * in the same scheduler being used as the previous.  This method will guarentee a different
 * algorithm <p>
 * This can create a problem if there is only 1 algorithm available.  Therefore, it throws
 * an IllegalStateException if a switch is not possible. 
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler
 * @exception IllegalStateException if there is only 1 scheduler and a change is not possible
 */
public XATScheduler forceChange() throws IllegalStateException {
	throw new IllegalStateException("Cannot switch algorithms because this strategy does not support changing");
}
/**
 * Returns the index of the Scheduling algorithm that should be used next.
 * Each subclass will override this method b/c this is where the actual logic of switching 
 * takes place.  This method updates the currentCategory class variable and then
 * return the next scheduler using getCurrentScheduler().  
 * This is how schedulers are referenced internally.
 * @return XATScheduler the next algorithm to use.  
 */
public XATScheduler getNextScheduler() {

	XATScheduler choice;

	// If the system time is less than the time we want to run round robin, run round robin.
	// Otherwise run chain.

	// This adapter will eventually be implemented such that you can choose any TWO algorithms, but
	// for simplicities sake, we will just use round robin and chain.

	if(System.currentTimeMillis() - startTime < roundRobinRunTime)
		choice = (XATScheduler) schedulers.get("RoundRobin");
	else
		choice = (XATScheduler) schedulers.get("Chain");
	
	return getCurrentScheduler();
}
	/**
    * called by the constructor.  this will set the default properties
    * that the ALL query objects will support. Each subclass should override
    * this method and add their own unique properties, but should also call
    * super.setDefaultProperties() so that the common properties may be set.
    */   
	protected void setDefaultProperties(){
		super.setDefaultProperties();
		//the weight of the previous values in calculating average
		addProperty("RoundRobinRunTime", "0");	
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject#getLightWeightQueryObject()
	 */
	public LightWeightQueryObject getLightWeightQueryObject() {
		return new LightWeightQueryObject("RRToChain",1,this.getClass().getName(), true);
	}
}
