/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import java.io.*;
/**
 * An instance of this class will encode a user's preference.  There
 * are 3 parts to a preference: the statistic, the quantifier (min or max) and
 * the weight.  The weights of all preferences in the system should sum to 1, if
 * they do not, an error will occur and execution will stop.  
 * @author: Brad Pielech
 */
public class XATPreference implements Serializable{
	/** constant for maximizing */
	public static int MAX = 1;
	/** constant for minimizing */
	public static int MIN = -1;
	/** the statistic related to this preference */
	private String statistic;
	/** 
	 * the quantifer for this preference.  This is either 1 (maximize)
	 * or -1 (minimize)
	*/
	private int quantifier;
	/** the relative weight assigned to this preference */
	private double weight;
	
/**
 * XATPreference constructor comment.
 */
public XATPreference() {
	super();
}
/**
 * Insert the method's description here.
 * Creation date: (3/29/2003 10:39:27 PM)
 * @return int
 */
public int getQuantifier() {
	return quantifier;
}
/**
 * Insert the method's description here.
 * Creation date: (3/29/2003 10:39:27 PM)
 * @return java.lang.String
 */
public java.lang.String getStatistic() {
	return statistic;
}
/**
 * Insert the method's description here.
 * Creation date: (3/29/2003 10:39:27 PM)
 * @return double
 */
public double getWeight() {
	return weight;
}
/**
 * Set the new quantifier, this should be 1 or -1.
 * @param newQuantifier int
 */
public void setQuantifier(int newQuantifier) {
	if((newQuantifier == XATPreference.MAX) || (newQuantifier == XATPreference.MIN)){
		quantifier = newQuantifier;
		return ;
	} else{
		throw new IllegalArgumentException("The quantifier should be 1 or -1");
	}
	
}
/**
 * set the statistic associated with this preference.  An error
 * is thrown if the statistic name is not supported by the statistics gatherer.
 * @param newStatistic java.lang.String
 */
public void setStatistic(java.lang.String newStatistic) {
	//@TODO fix this to work with distribution
	//if(!StatisticsGatherer.getStatisticsGather().isStatisticSupported(newStatistic)){
	//	throw new IllegalArgumentException("The statistic " + newStatistic + " is not supported by the Stats Gatherer");
	//}
	statistic = newStatistic;
}
/**
 * set the weight, this must be between 0 and 1
 * @param newWeight double
 */
public void setWeight(double newWeight) {
	if((weight < 0) || (weight > 1)){
		throw new IllegalArgumentException("The weight must be between 0 and 1 inclusive");
	}
	weight = newWeight;
}
/**
 * Insert the method's description here.
 * Creation date: (3/30/2003 4:15:42 PM)
 * @return java.lang.String
 */
public String toString() {
	String s = "";
	s = "XATPreference: ";
	s = s + "\t Statistic: " + statistic;
	s = s + "\t Quantifier: " + quantifier;
	s = s + "\t Weight: " + weight;
	return s;
}
}
