/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer;

/**
 * Statistics that wish to keep an aggregate value (total), mean, min and max, but report
 * their values in increments (such as the statistic from the last running) as opposed to
 * the overall total, should derive from this class instead of AggregateStatisticsObject.
 * @author: Brad Pielech
 */
public abstract class AggregateDifference extends AggregateStatisticsObject {
/**
 * AggregateDifference constructor comment.
 */
public AggregateDifference() {
	super();
}
/**
 * AggregateDifference constructor comment.
 * @param name java.lang.String
 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
 */
public AggregateDifference(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) {
	super(name, queryObject);
}
/**
 * Sets the property value.  For AggregateStatisticsObjects, this method
 * will calculate the new weighed mean and set the value to what was
 * passed in.  This assumes the newly passed in value represents the
 * total property value and not the difference from when it was calculated
 * previously. <p>
 *
 * For example, the THROUGHPUT calculator will report
 * how many tuples the tree has outputted in total, not just how many
 * were outputted since the last time it ran.  OTOH, the NUMBER_OF_TUPLES_OUTPUTTED
 * calculator will report how many tuples have been outputted since it last ran.  Therefore,
 * that calculator should derive from AggregateDifference
 * @param newPropertyValue java.lang.String
 */ 
public void setPropertyValue(String newPropertyValue) {
	double thisValue = Double.parseDouble(newPropertyValue);
	double oldValue = 0;
	String oldValueS = getPropertyValue();
	if(oldValueS != null){ 
		oldValue = Double.parseDouble(oldValueS);
	}
	double total = thisValue + oldValue;	
    super.setPropertyValue("" + total);
    
}
}
