/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer;

//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
//import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
//import java.util.*;
/**
 * This is the StatisticsObject used for any property that is incremented by
 * some value.  Calling <tt>calculateNewValue()</tt> will increment by 1.  Calling
 * <tt>calculateNewValue(String newValue)</tt> will increment by whatever int newValue encodes. <p>
 *
 * If you need to be able to decrement, you can pass "-1" If you need to increment based on
 * some other property, then you have to make a seperate StatisticsObject for that purpose. 
 * 
 */
public class IncrementValueByN extends StatisticsObject {
/**
 * TupleCountStatisticsObject constructor comment.
 */
public IncrementValueByN() {
	super();
}
/**
 * TupleCountStatisticsObject constructor comment.
 * @param name java.lang.String
 * @param queryObject edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
 */
public IncrementValueByN(String name, edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject queryObject) {
	super(name, queryObject);
}
/**
 * Calculates the new Value of the given property of this operator.  This
 * should be overwritten by subclasses who can tailor this method to calculating 1
 * specific property. <p>
 * This method should update at least the local propertyValue and possibly the
 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
 *
 * The procedure for calculating new values should be as follows:
 *
 * <ol>
 * <li> Consult the associatedOperator or statsTable for any propertyValues that are needed.
 * <li> Perform the required calculations using the propertyValues from step 1
 * <li> Update the local propertyValue.
 * <li> Update the property of the operator iff <tt>getCategory() == StatisticsGatherer.OVERALL<tt>
 * and the property is an aggregate value.  Otherwise, leave the operator property alone. This
 * way the operator property's value's represent the OVERALL aggregate values.
 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
 * </ol> 
 */
public void calculateNewValue() {
	
	calculateNewValue("1");
}
/**
 * This is semantically the same as <tt>calculateNewValue()</tt>, but this
 * provides a parameter which assists in the calculation.
 * Calculates the new Value of the given property of this operator.  This
 * should be overwritten by subclasses who can tailor this method to calculating 1
 * specific property. <p>
 * This method should update at least the local propertyValue and possibly the
 * value with the operator The entire StatisticsGatherer is available for assistance.   <p>
 *
 * The procedure for calculating new values should be as follows:
 *
 * <ol>
 * <li> Consult the statsTable for any propertyValues that are needed.
 * <li> Perform the required calculations using the propertyValues from step 1
 * <li> Update the local propertyValue.
 * <li> Update the lastTimeUpdated value by using <tt>lastTimeUpdated = System.currentTimeMillis();</tt>
 * </ol> 
 */
public void calculateNewValue(String newValue) {
	
	int count = Integer.parseInt(this.getPropertyValue());
	int add = Integer.parseInt(newValue); 
	this.setPropertyValue("" + (count + add));	
}
}
