/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;
/*
 * Created on May 10, 2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */

import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.lang.reflect.Field;


/**
 * @author Luping Ding and Nishant Mehta
 * Represents a class that is a collection of hetrogeneous types
 */
public class ClassSymbolEntry extends ComplexSymbolEntry {
	/**Stores the reference to the Class represented by this Entry*/	
	Class theClass;
	/**Stores the reference to field array, these contain field objects
	 * for the current symbol entry
	 */
	
	Field[] fields;
	
	/**
	 * @param t
	 * @param typeName
	 * @param varName
	 */
	public ClassSymbolEntry(Object t, String typeName, String varName) {
		super(t, typeName, varName);
		theClass = t.getClass();
		fields = theClass.getDeclaredFields();
		for(int i=0;i<fields.length;i++)
		{
				fields[i].setAccessible(true);	
		}
	}

	/**
	 * @see edu.wpi.cs.dsrg.conjoin.storagemanager.SymbolEntry#write(java.lang.Object, java.io.ObjectOutputStream)
	 */
	public void write(Object t, ObjectOutputStream os) {
		if (t != null) {
			//The length of the fields and the length of the
			//child array list has to be same
			//System.out.println("In write of Class Symbol Entry");
			Object obj;
			
			for(int i=0,j=0;i<fields.length;i++) {
				try {				
					obj = fields[i].get(t);
					if(obj != null)
					{				
						((SymbolEntry) (childSymbolEntry.get(j))).write(obj,os);				
						j++;
					}				
				} catch (IllegalArgumentException e) {			
					e.printStackTrace();
					System.out.println("The Value of i "+i);
					System.out.println("The name of the class "+t.getClass().getName());
					System.out.println("The name of the field we are trying to access "+fields[i].getName());
					System.out.println("The type of the symbol entry "+typeName);
					System.exit(0);
				} catch (IllegalAccessException e) {				
					e.printStackTrace();
				}
			}
		}
	}

	/**
	 * @see edu.wpi.cs.dsrg.conjoin.storagemanager.SymbolEntry#read(java.io.ObjectInputStream)
	 */
	public Object read(ObjectInputStream oi) {
		Object obj = null;
		//System.out.println("In read of Class Symbol Entry");		
		try {
			obj = theClass.newInstance();
			//Note the number of fields should be the same as the number
			//of child elements
			int i = 0;
			for(int j=0; j < fields.length; j++)
			{
				if(fields[j].getName().compareTo(((SymbolEntry) this.childSymbolEntry.get(i)).varName) == 0)
				{
					Object readObject = ((SymbolEntry) this.childSymbolEntry.get(i)).read(oi);
					i++;
					fields[j].set(obj, readObject);
				}
			}
			
		} catch (InstantiationException e) {			
			e.printStackTrace();
		} catch (IllegalAccessException e) {			
			e.printStackTrace();
		}
		return(obj);
	}
}
