/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;
/*
 * Created on May 10, 2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */

import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;

/**
 * Note this class represents a complex symbol entry this
 * entry represents a complex type this type is nothing but a 
 * collection of simple types or complex types or both
 * @author Luping Ding and Nishant Mehta
 */
public abstract class ComplexSymbolEntry extends SymbolEntry {
	
	/**Stores the list of the child symbol entries*/
	protected ArrayList childSymbolEntry = null;
	 
	/**
	 * Constructs a new complex symobol entry
	 * @param t
	 * @param typeName
	 * @param varName
	 */	
	public ComplexSymbolEntry(Object t, String typeName, String varName) {
		super(t, typeName, varName);
		this.childSymbolEntry = new ArrayList();
	}

	/** 
	 * @see edu.wpi.cs.dsrg.conjoin.storagemanager.SymbolEntry#write(java.lang.Object, java.io.ObjectOutputStream)
	 */
	public abstract void write(Object t, ObjectOutputStream os);

	/**
	 * @see edu.wpi.cs.dsrg.conjoin.storagemanager.SymbolEntry#read(java.io.ObjectInputStream)
	 */
	public abstract Object read(ObjectInputStream oi);
	
	/**
	 * Adds a child to the complex Symbol entry
	 * Note each complex entry can be a parent of one or more 
	 * symbol entries. Where each of those symbol entries can be 
	 * a complex or simple symbol entry	 
	 * @param child
	 */
	public void addChild(SymbolEntry child)
	{
		this.childSymbolEntry.add(child);
	}	
	
	/**
	 * @see edu.wpi.cs.dsrg.conjoin.storagemanager.SymbolEntry#returnString()
	 */
	public String returnString()
	{
		String temp = this.varName+ "\n";
		
		for(int i=0;i<this.childSymbolEntry.size();i++)
		{
			temp = temp + ((SymbolEntry) this.childSymbolEntry.get(i)).returnString()+"\n";
		}
		return(temp);
	}

}
