/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;

/**
 * @author Nishant Mehta and Abhishekh Kumar
 * A class that stores the list of deleted index entries
 * It is sorrted list of index entries Sorted on the value of 
 * position
 */
public class DeletedIndexEntries {
	/*Pointer to head of the link list*/
	
	protected Node head = null;	
	/**
	 * Constructor for DeletedIndexEntries.
	 */
	public DeletedIndexEntries() {
		head = null;	
	}
	
	/**
	 * Adds a new entry into the link liat
	 * @param newEntry
	 */
	public void addIndexEntry(IndexEntry newEntry)
	{
		Node temp = null;
		Node probe = this.head;
		if(probe == null)
		{
			head = new Node(newEntry);
		}
		else
		{
			while (probe != null) {
			if (probe.getIndexEntry() != null) {
					if(probe.getIndexEntry().getPosition() < newEntry.getPosition()) {
					temp = probe;
					probe = probe.next;
					}
					else
					break;
			}//end of if	
			
		  }//end of while
		  //Inserting into the 1st Position
		  if(probe == head)
		  {
		  	//This is making the current head pointing to the new node		  	 
		  	head = new Node(newEntry,head);			  	
		  }		  		  
		  else if(probe == null)
		  {
		  	//Inserting a new node that is the last node of the list
			temp.setNext(new Node(newEntry));		  	
		  }
		  else
		  {
		  	//Inbetween node of the list
		  	temp.setNext(new Node(newEntry,probe));		  			  		
		  }
		}//end of else
	}	
	
	/**
	 * Returns no of Index Entries in the list having a position value greater than
	 * the indexEntry used to probe the list
	 * @param probeIndex
	 * @return int
	 */
	public int noOfEntriesBefore(IndexEntry probeIndex)
	{
		Node  temp = head;
		int counter = 0;
		while(temp != null)
		{
			if(probeIndex.getPosition() > temp.getIndexEntry().getPosition())
			{
			counter = counter + 1;	
			temp = temp.next;
			}
			else
			break;
		}
		return(counter);
	}

	
	/**
	 * @see java.lang.Object#finalize()
	 */
	public void finalize()
	{
		Node temp;
		while(head != null)
		{
			temp = head	;	
			temp = null;
			head = head.getNext();
		}
	}
	
	public int noOfNodes()
	{
		int counter = 0;
		Node temp = head;
		while(temp != null)
		{
			counter = counter + 1;
			temp= temp.next;
		}
		return(counter);
	}	
	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		Node temp = head;
		String value = "";
		
		while(temp != null)
		{
			value = value + temp.getIndexEntry().getPosition()+ "  ";
			temp = temp.getNext();			
		}
		return(value);						
	}
	

}


/**
 * @author Nishant Mehta and Abhishekh Kumar
 *
 *Node of the link list
 */
class Node
{
	/*The data item in the node*/
	protected IndexEntry indexEntry;
	/*Pointer to the next Node*/
	protected Node next; 
	/**
	 * Method Node.
	 * @param ie
	 */
	public Node(IndexEntry ie)
	{
		this.indexEntry = ie;
		this.next = null;		
	}
	
	/**
	 * Method Node.
	 * @param ie
	 * @param ptrNext
	 */
	public Node(IndexEntry ie,Node ptrNext)
	{
		this.indexEntry = ie;
		this.next = ptrNext;
	}
	/**
	 * Returns the next.
	 * @return Node
	 */
	public Node getNext() {
		return next;
	}

	/**
	 * Sets the next.
	 * @param next The next to set
	 */
	public void setNext(Node next) {
		this.next = next;
	}

	/**
	 * Returns the indexEntry.
	 * @return IndexEntry
	 */
	public IndexEntry getIndexEntry() {
		return indexEntry;
	}

}	
	