/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;


/** 
 * Used to create various types of objects used by the StorageManager module
 * Uses the metainformation objects to create records from tuples
 * Also used to create tuple objects from record objects 
 * Also converts records to byte arrays
 * Used by the DiskFile extensively to construct records and return
 * records to the storage manager. The storage manager can then use this
 * class to construct tuple arrays
 * Note this is a singleton
 * @author Nishant Mehta and Abhishek Kumar
 */
public class Factory {
	/* Reference to the factory object*/	 
	//protected Factory f = null;	
	/*Reference to the dummy tuple object passed by the storage manager*/
	protected XATTuple dummyInstance = null;		
	/*Stores the length of data array*/
	int lengthOfArray = 0;
	
	protected SymbolEntry[] symbolArray = null;
	/*Stores the reference to the root of the symbol entry tree*/
	protected SymbolEntry root = null;
	
	/**
	 * Method Factory.
	 * Expects as parameter the root of the metainformation tree
	 * @param mi
	 * @param dummyInstance
	 */
	public  Factory(XATTuple dummyInstance)	
	{
				
		this.dummyInstance = dummyInstance;			
		MetaInformation mi = new MetaInformation(dummyInstance, "dummyInstance");
		root = mi.getSymbolEntryTree();
		//root =
		//			MetaInformation
		//				.getMetaInformation(dummyInstance, "dummyInstance")
		//				.getSymbolEntryTree();
			
	}
	
	
	/**
	 * @see java.lang.Object#finalize()
	 * USed to set all the instance objects to null
	 */
	public void finalize()
	{
		//f = null;					
	}
	
	
	
	/**
	 * Makes a record array from the byteArray given by input
	 * Used by the disk file to convert bytesArrays to recordArrays 
	 * @param byteArray
	 * @return Record[]
	 */
	public Record[] makeRecord(byte[] arrayByte)
	 {
		return(null);
	}
	
	
	/**
	 * Makes tuples from records. 
	 * Using the metainformation it retrieves the data from each record
	 * @param arrayRecord
	 * @return IndexedTuple[]
	 * 
	 */
	public XATTupleImp[] makeTuples(Record[] arrayRecord)
	{
		//start working from here
		
		XATTupleImp[] indexedTuples = new XATTupleImp[arrayRecord.length];

		try {
			for (int i = 0; i < arrayRecord.length; i++) {
				//System.out.println("The length of the byte array read "+arrayRecord[i].getByteArray().length);				
				ByteArrayInputStream bais =
					new ByteArrayInputStream(arrayRecord[i].getByteArray());
				
				ObjectInputStream oi = new ObjectInputStream(bais);
				//System.out.println("---Started Retreiving XATTuple---");
				//System.out.println("---End of  Retreiving XATTuple---");
				//reads the complete XATTuple from the object input stream
				
				XATTupleImp t = (XATTupleImp) root.read(oi);
				
				t.setIndex(arrayRecord[i].getIndex());					
				indexedTuples[i] = t;
				bais = null;
				oi = null;
			}
		} catch (IOException e) {
			System.out.println("Exception " + e);
			e.printStackTrace();
		}
		catch(Exception eo)
		{
			eo.printStackTrace();
		}
	
		return(indexedTuples);	
	}
		
	
	/**
	 * Makes Records from tuples. 
	 * Coverts tuples tuples to records. This method has to 
	 * store the byte represenation of the XATTuple.
	 * @param arrayTuple
	 * @return Record[]
	 * 
	 */
	public Record[] makeRecords(XATTuple[] arrayTuple)
	{
		ByteArrayOutputStream baos = null;
		ObjectOutputStream os = null;
	
		Record[] arrayRecord = new Record[arrayTuple.length];		
																	
		for(int i=0;i<arrayTuple.length;i++)
		{		
			try {								
				baos = new ByteArrayOutputStream();
				os = new ObjectOutputStream(baos);
				
				//Writes the records to the object outputstream
												
				root.write(arrayTuple[i], os);
				os.flush();															
				byte[] buf = baos.toByteArray();
				//System.out.println("The length of the byte array written "+buf.length);
				arrayRecord[i] = new Record(baos.toByteArray());
																					
				buf = null;
				baos = null;
				os = null;	
				//os.close();
				//baos.close();							
			} catch (IOException e) {
				System.out.println("Exception "+e);
				e.printStackTrace();
			}
		}	
		return(arrayRecord);																	
	}

	
		
	/**
	 * See if you can get rid of this method. This method may cause
	 * Some extra overhead
	 * Method to covert an array of records to byte array
	 * @param Record[]
	 */
	public byte[] recordsToBytes(Record[] records)
	{
		ByteArrayOutputStream baos = new ByteArrayOutputStream();				
		//System.out.println("The length of records "+records.length);
		//System.out.println("The length of the 1st record "+records[0].length());
		for(int i=0;i < records.length;i++)
		{
			try {
				baos.write(records[i].getByteArray());				
			} catch (IOException e) {
				System.out.println("Exception "+e);
				e.printStackTrace();
			}
		}		
		//System.out.println("Size "+baos.size());
		byte[] buf = baos.toByteArray();
		//setting it to null because not going to use any further
		baos = null;
		return(buf);
	}		



				
}
