/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;

/**
 * DataStructure that stores the index Entry. Used by the 
 * Class DiskFile.
 * @author Nishant Mehta and Abhishek Kumar
 */

public class IndexEntry {
	/*Position in File with respect to ther records*/
	protected int position;
	/*Stores the offset for the 1st byte for the record in the File*/
	protected long startOffsetInFile;
	/*Stores the offset for the last byte of the record in the File*/
	protected long endOffsetInFile;	
	/*True if the record is marked deleted*/
	protected boolean deleted;
	
	/**
	 * Constructs a new Index Entry
	 * @param position
	 * @param startOffsetInFile
	 * @param endOffsetInFile
	 * @param deleted
	 */
	public IndexEntry(int position,long startOffsetInFile,long endOffsetInFile,boolean deleted) {
		
		this.position = position;
		this.startOffsetInFile = startOffsetInFile;
		this.endOffsetInFile = endOffsetInFile;
		this.deleted = deleted;		
	}
	
	
	/**
	 * Constructs a new Index Entry
	 * @param position
	 * @param byteOffsetInFile
	 * @param endOffsetInFile
	 */
	public IndexEntry(int position,long startOffsetInFile,long endOffsetInFile) {
		
		this.position = position;
		this.startOffsetInFile = startOffsetInFile;
		this.endOffsetInFile = endOffsetInFile;
		this.deleted = false;		
	}
	
	/**
	 * @see java.lang.Object#finalize()
	 */
	public void finalize()
	{}
	
	/**
	 * Returns the status of the index entry 
	 * True indicates that the entry is marked deleted
	 * @return boolean
	 */
	public boolean isDeleted() {
		return deleted;
	}

	/**
	 * Marks the index entry deleted
	 * @param deleted 
	 */
	public void setDeleted(boolean deleted) {
		this.deleted = deleted;
	}
	

	/**
	 * Returns the endOffsetInFile.
	 * @return long
	 */
	public long getEndOffsetInFile() {
		return endOffsetInFile;
	}

	/**
	 * Returns the position.
	 * @return int
	 */
	public int getPosition() {
		return position;
	}

	/**
	 * Returns the startOffsetInFile.
	 * @return long
	 */
	public long getStartOffsetInFile() {
		return startOffsetInFile;
	}
	
	/**
	 * Returns the size of the record in no Of bytes
	 * @return int
	 */
	public int size(){
		
	return((int) (this.endOffsetInFile - this.startOffsetInFile +1));
	
	}	
	
	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		return(new String (this.position+"  "+this.startOffsetInFile+"--"+this.endOffsetInFile));					
	}

}
