/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;

import java.io.FileWriter;
import java.io.IOException;

/**
 * A single unit of information used by the class DiskFile to
 * store and retreive data from the diskfile. 
 * Note the DiskFile will only store retrieve records.
 * Also a record is a byte representation of a tuple
 * @author Nishant Mehta and Abhishek Kumar
 */
public class Record {
	
	/*Stores a dynamic array and keeps adding bytes as and when needed*/	
	protected byte[] byteArray = null;
	/*Stores the index for this record*/
	protected int index; 
	
	/**
	 * Constructs a new record
	 * @param bt
	 * @param index
	 */
	public Record(byte[] bt,int index)
	{
		byteArray = bt;	
		this.index = index;
	}
	
	/**
	 * @see java.lang.Object#finalize()
	 */
	public void finalize()
	{
		byteArray = null;
	}	
	
	/**
	 * Constructs a new byte array using this byte array reference
	 * @param bt
	 */
	public Record(byte[] bt)
	{
		byteArray = bt;
	}
	
	/**
	 * Copies the value in between the startindex and endindex including
	 * the start and end index.
	 * @param buffer
	 * @param startIndex
	 * @param endIndex
	 */
	public Record(byte[] buffer, long startIndex, long endIndex, int fileIndex) {
		//Hack here too
		//endIndex = endIndex - startIndex;
		//startIndex = 0;
		
		if (endIndex > buffer.length - 1)
			endIndex = buffer.length - 1;
		
		if (startIndex > endIndex)
			startIndex = endIndex;
		
		//System.out.println(buffer.length+"  "+startIndex+"--"+endIndex);
		
		this.byteArray = new byte[(int)(endIndex - startIndex + 1)];
		try {
			
			System.arraycopy(buffer, (int)startIndex, byteArray, 0, byteArray.length);
			
		} catch (IndexOutOfBoundsException e) {
			try {
				FileWriter out = new FileWriter("record_exceptions.txt", true);
				out.write("Exception: " + e + "\n");
				out.write("Source: " + buffer.toString() + "\n");
				out.write("Source Position: " + (int)startIndex + "\n");
				out.write("Destination: " + byteArray.toString() + "\n");
				out.write("Destination Position: 0\n");
				out.write("Length: " + byteArray.length + "\n");
				out.write("srcPos + length (" + ((int)startIndex + byteArray.length) + ") >? " + buffer.length + "\n");
				out.write("destPos + length (" + (0 + byteArray.length) + " >? " + byteArray.length + "\n\n");
				out.close();				
			} catch (IOException ie) {
				
			}
		} catch (ArrayStoreException e) {
			try {
				FileWriter out = new FileWriter("record_exceptions.txt", true);
				out.write("Exception: " + e + "\n");
				out.write("Source: " + buffer.toString() + "\n");
				out.write("Source Position: " + (int)startIndex + "\n");
				out.write("Destination: " + byteArray.toString() + "\n");
				out.write("Destination Position: 0\n");
				out.write("Length: " + byteArray.length + "\n\n");
				out.close();				
			} catch (IOException ie) {
				
			}
		} catch (NullPointerException e) {
			try {
				FileWriter out = new FileWriter("record_exceptions.txt", true);
				out.write("Exception: " + e + "\n");
				out.write("Source: " + buffer.toString() + "\n");
				out.write("Source Position: " + (int)startIndex + "\n");
				out.write("Destination: " + byteArray.toString() + "\n");
				out.write("Destination Position: 0\n");
				out.write("Length: " + byteArray.length + "\n\n");
				out.close();				
			} catch (IOException ie) {
				
			}
		}
		
		//for(long i=0,j=startIndex;i<byteArray.length;i++,j++)
		//{
		//	this.byteArray[(int) i] = buffer[(int) j];
		//}
		this.index = fileIndex;
	}

		
	/**
	 * Gives as output in the form of byte array the current contents of the record
	 * @return byte[]
	 */
		
	public byte[] toByteArray()
	{
		return(null);
	}
	
	/**
	 * Used to append bytes to the byte array in the existing record
	 * @param arrayByte
	 */
	public void append(byte[] arrayByte)
	{
		
	}
	
	/**
	 * Returns the length of the record in Bytes
	 * @return long
	 */
	public long length()
	{
		return(byteArray.length);
	}
				
	/**
	 * Returns the byteArray.
	 * @return byte[]
	 */
	public byte[] getByteArray() {
		return byteArray;
	}
		

	/**
	 * Returns the index.
	 * @return int
	 */
	public int getIndex() {
		return index;
	}

	/**
	 * Method size.
	 */
	public int size() {
		return(this.byteArray.length);
		
	}

}
