/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;

//import java.io.BufferedWriter;
//import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidArgumentException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidIndexException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidPartitionException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.NotEnoughBytesException;
//import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.NotSupportedTypeException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.StorageManagerNotInitializedException;

/**
 * This class implements the IStore information.
 * 
 * @author Nishant Mehta and Abhishek Kumar; edited by Natasha Bogdanova (6/9/04)
 */
public class StorageManager implements IStore {
	
	/** Buffered writer used to write file sizes to the buffered writer. */
	//protected static BufferedWriter bw = null;

	/** Used to store the mapping from partitionNumber to DiskFile. */
	public HashMap PartitionToFileMapping;
	
	/** Stores the total number of tuples on the disk. */
	protected long totalSize = 0;

	/** A refernce to the new factory. */
	public Factory newFactory = null;
	
	/**
	 * Used to construct a new StorageManager. Initializes the object in
	 * the metainformation object and the factory object.
	 * 
	 * @param dummyInstance XATTupleImp
	 * @throws NotSupportedTypeException
	 */
	public StorageManager(XATTupleImp dummyInstance) {
		//throws NotSupportedTypeException {

		//try {
		//	if (bw == null)
		//		bw = new BufferedWriter(new FileWriter("data.txt"));
		//} catch (IOException e) {
		//	e.printStackTrace();
		//}

		PartitionToFileMapping = new HashMap();

		//This gets an instance of the factory 
		//and also will generate the complete symbol tree										
		this.newFactory = new Factory(dummyInstance);
		
		//System.out.println("Storage Manager initialized.");

		//System.out.println(MetaInformation.getMetaInformation().toString());
	}

	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IStore#retrieveFromDisk(int, boolean)
	 */
	public XATTupleImp[] retrieveFromDisk(int partitionNumber, boolean purge)
		throws InvalidPartitionException {

		IRecordStore df =
			(DiskFile) this.PartitionToFileMapping.get(
				new Integer(partitionNumber));
		if (df == null)
			throw new InvalidPartitionException();

		Record arrayRecord[] = df.retrieveFromDisk(purge);

		XATTupleImp[] indexTuple = null;
		
		indexTuple = this.newFactory.makeTuples(arrayRecord);
		//System.out.println("Retreived "+indexTuple.length+"  Records from bucket "+partitionNumber);				
		
		return (indexTuple);
	}
	
	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IStore#retrieveFromDisk(int, int, boolean)
	 */
	public XATTupleImp[] retrieveFromDisk(int partitionNumber, int noOfRecords, boolean purge)
		throws InvalidPartitionException, NotEnoughBytesException {

		IRecordStore df =
			(DiskFile) this.PartitionToFileMapping.get(
				new Integer(partitionNumber));
		if (df == null)
			throw new InvalidPartitionException();

		Record arrayRecord[] = df.retrieveFromDisk(noOfRecords, purge);

		XATTupleImp[] indexTuple = null;
		//for (int i = 0; i < arrayRecord.length; i++) {
		//	Record oneTuple[] = {arrayRecord[i]};
		//	System.out.println(i + "--> " + this.newFactory.makeTuples(oneTuple)[0].toString());
		//}
		
		indexTuple = this.newFactory.makeTuples(arrayRecord);
		//System.out.println("Retreived "+indexTuple.length+" Records from partition "+partitionNumber);				
		
		return (indexTuple);
	}


	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IStore#retrieveFromDisk(int, long, boolean)
	 */
	public XATTupleImp[] retrieveFromDisk(int partitionNumber, long noOfBytes, boolean purge)
		throws InvalidPartitionException, NotEnoughBytesException {

		IRecordStore df =
			(DiskFile) this.PartitionToFileMapping.get(
				new Integer(partitionNumber));
		if (df == null)
			throw new InvalidPartitionException();

		Record[] arrayRecord = df.retrieveFromDisk(noOfBytes, purge);
		
		return (newFactory.makeTuples(arrayRecord));
	}

	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IStore#retrieveFromDisk(int, int, long, boolean)
	 */
	public XATTupleImp[] retrieveFromDisk(int partitionNumber, int startIndex, 
										  long noOfBytes, boolean purge)
		throws InvalidPartitionException, InvalidIndexException, NotEnoughBytesException {

		IRecordStore df =
			(DiskFile) this.PartitionToFileMapping.get(
				new Integer(partitionNumber));
		if (df == null)
			throw new InvalidPartitionException();

		Record[] arrayRecord = df.retrieveFromDisk(startIndex, noOfBytes, purge);
		
		return (newFactory.makeTuples(arrayRecord));
	}

	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IStore#size(int)
	 */
	public long size(int partitionNumber) {
		IRecordStore df =
			(DiskFile) this.PartitionToFileMapping.get(
				new Integer(partitionNumber));
		if (df == null)
			return (0);
		else
			return (df.size());
	}

	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IStore#saveToDisk(int, XATTupleImp[])
	 * If the stoarage manager has not been initialized yet, it can now 
	 * initialize itself with the schema of the tuple using the 1st tuple
	 * in the array the 1st time the saveToDisk method is called.
	 */
	public void saveToDisk (int partitionNumber, XATTupleImp[] arrayTuple)
		throws IllegalArgumentException, StorageManagerNotInitializedException {
		
		if (this.newFactory == null) {
			throw (new StorageManagerNotInitializedException());
			//System.out.println(MetaInformation.getMetaInformation().toString());
		}
		
		IRecordStore irs =
			(IRecordStore) PartitionToFileMapping.get(
				new Integer(partitionNumber));
				
		if (irs == null) {
			IRecordStore newirs = new DiskFile("" + partitionNumber,this.newFactory);
			PartitionToFileMapping.put(new Integer(partitionNumber), newirs);
			irs = newirs;
		}
		//Make byte records from the tuples
		//		Record[] record = factory.makeRecords(arrayTuple);

		//Testing new Factory
		//System.out.println("Testing new factory");		
		Record[] record = newFactory.makeRecords(arrayTuple);

		//Save the records to disk.
		//System.out.println("---Got records from factory---");		

		try {
			irs.saveToDisk(record);

			//System.out.println("\n");			
			//System.out.println("Storing " +arrayTuple.length +" records in the bucket " + partitionNumber);								
			//System.out.println("The current size of the bucket "+irs.size());
			//System.out.println("\n");
			
			long sizeInBytes = 0;
			//Note this piece of code is to keep a trck of file size for evaluation		
			//Enumeration enumeration = this.PartitionToFileMapping.elements();

			//while (enumeration.hasMoreElements()) {
			//	IRecordStore  irsElement = (IRecordStore) enumeration.nextElement();
			//	sizeInBytes = sizeInBytes + irs.sizeInBytes();				
			//}
			
			//String size = Integer.toString((int)sizeInBytes)+" ";			
			//bw.write(size);
			//bw.flush();
			
			totalSize = totalSize + arrayTuple.length;
			
		} catch (InvalidArgumentException e) {
			throw(new IllegalArgumentException());			
		} catch (IOException e) {
			System.out.println("Exception " + e);
			e.printStackTrace();
		}
	}

	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IStore#purgeBucketFromDisk(int, int[])
	 */
	public void purgeBucketFromDisk(int partitionNumber, int[] tupleIndexes)
		throws
			NotEnoughBytesException,
			InvalidPartitionException,
			InvalidIndexException {
		IRecordStore irs =
			(DiskFile) this.PartitionToFileMapping.get(
				new Integer(partitionNumber));
		if (irs == null)
			throw new InvalidPartitionException();
		else
			irs.purgeFromDiskFile(tupleIndexes);
		totalSize = totalSize - tupleIndexes.length;	
	}

	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IStore#purgeBucketFromDisk(int)
	 */
	public void purgeBucketFromDisk(int partitionNumber)
		throws InvalidPartitionException {
		IRecordStore irs =
			(DiskFile) this.PartitionToFileMapping.get(
				new Integer(partitionNumber));
		if (irs == null)
			throw new InvalidPartitionException();
		else
		{		
			totalSize = totalSize - irs.size();
			irs.purgeDiskFile();
		}
			 
	}

	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IStore#cleanup()
	 * Cleans up all the files and open file handles.
	 */
	public void cleanup() {
		Object[] arrayObject = PartitionToFileMapping.values().toArray();
		for (int i = 0; i < arrayObject.length; i++) {
			((DiskFile) arrayObject[i]).cleanup();
		}
	}

	/**
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IStore#initialize()
	 */
	public void initialize() {}
	
	/**
	 *  @see edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.IStore#totalSize()
	 */
	public long totalSize() {		
		return(totalSize);		
	}

}
