/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.
package edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.InvalidPartitionException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.NotEnoughBytesException;
import edu.wpi.cs.dsrg.xmldb.xat.component.storagemanager.exceptions.StorageManagerNotInitializedException;
/**
 * @author natasha
 * 
 * Tests the storage manager.
 */
public class Test {
	public static void main(String[] args) {
		runReadWriteTest();
		//runPurgeTest();
	}
	public static XATTupleImp[] getTestArray()	
	{	
		Long ip_address_1 = new Long(1);	
		Long ip_address_2 = new Long(2);	
		XATDataValue[] dvs1 = {new XATDataValueImp(ip_address_1)};	
		XATDataValue[] dvs2 = {new XATDataValueImp(ip_address_2)};		
	
		XATTupleImp[] testTuples = new XATTupleImp[2];			
	
		testTuples[0] = new XATTupleImp(dvs1);	
		testTuples[1] = new XATTupleImp(dvs2);	
			
		return(testTuples);	
	}
	public static void runSimpleTest()
	{
		XATTupleImp[] testTupleArray = getTestArray();
		StorageManager sm1 = new StorageManager(testTupleArray[0]);		
		for(int i = 0; i< 3;i++)
		{
				try{
				sm1.saveToDisk(2, testTupleArray);		
				XATTupleImp[] readTuples2 = 	
							sm1.retrieveFromDisk(2, 1, true);			
	
				for (int j = 0; j < readTuples2.length; j++)	
					System.out.println("Tuple Value "+ readTuples2[j]);			
			}catch(StorageManagerNotInitializedException se){
				se.printStackTrace();
			} catch (InvalidPartitionException e) {
				e.printStackTrace();
			} catch (NotEnoughBytesException e) {
				e.printStackTrace();
			}
		} 
	}
	/**
	 * Runs a reading/writing from file test.
	 * Creation Date: (Jul 25, 2004 3:13:57 PM)
	 * @author Natasha Bogdanova
	 */
	public static void runReadWriteTest() {		
		//XATTupleImp[] tuples1 = getTupleArray1();
		XATTupleImp[] tuples2 = getTupleArray1();		
		//StorageManager sm1 = new StorageManager(tuples1[0]);
		StorageManager sm2 = new StorageManager(tuples2[0]);
				
		//XATTuple dummy1 = sm1.newFactory.dummyInstance;
		XATTuple dummy2 = sm2.newFactory.dummyInstance;				
		//System.out.println(sm2.newFactory.root.returnString() + "\n\n");
		//System.out.println(sm2.newFactory.root);
		//System.out.println(dummy1.getArrayData()[0].getValue().getClass());
		//System.out.println(dummy2.getArrayData()[0].getValue().getClass());
		for (int n = 0; n < 3; n++) {
			System.out.println(n);
			// write to disk
			//try {
			//	sm1.saveToDisk(1, tuples1);
			//} catch (StorageManagerNotInitializedException e) {
			//	System.out.println("sm1 not initialized.");
			//}
			try {
				sm2.saveToDisk(2, tuples2);
			} catch (StorageManagerNotInitializedException e) {
				System.out.println("sm2 not initialized.");
			}
			try {
				XATTupleImp[] readTuples2 = 
							sm2.retrieveFromDisk(2, 81, true);
				//System.out.println(readTuples2.length);
				for (int i = 0; i < readTuples2.length; i++)
					System.out.println(readTuples2[i]);
			} catch (Exception e) {
				System.out.println(e);
			}
		}
	}	
	/**
	 * Makes a sample array of tuples to be used for testing.
	 * Creation Date: (Jul 25, 2004 3:35:49 PM)
	 * @author Natasha Bogdanova
	 * @return XATTupleImp[]
	 */
	public static XATTupleImp[] getTupleArray1() {
		Long ip_address = new Long(2248710);
		String emptyAttr1 = "-";
		String emptyAttr2 = "-";
		String timestamp = "[04/May/1998:22:00:02 +0000]";
		String gmt = "GET /english/hosts/cfo/cfo.html HTTP/1.0";
		Integer objectID = new Integer(200);
		Integer clientID = new Integer(5029);				
		XATDataValue[] dvs1 = {new XATDataValueImp(ip_address),
							   new XATDataValueImp(emptyAttr1),
							   new XATDataValueImp(emptyAttr2),
							   new XATDataValueImp(timestamp),
							   new XATDataValueImp(gmt),
							   new XATDataValueImp(objectID),
							   new XATDataValueImp(clientID)};									   
		XATTupleImp[] tuples1 = new XATTupleImp[100];
				
		for (int i = 0; i < tuples1.length; i++) {
			tuples1[i] = new XATTupleImp(dvs1);
		}		
		return tuples1;
	}
	/**
	 * Makes a sample array of tuples to be used for testing.
	 * Creation Date: (Jul 25, 2004 3:36:13 PM)
	 * @author Natasha Bogdanova
	 * @return XATTupleImp[]
	 */
	public static XATTupleImp[] getTupleArray2() {
		String ip_address_2 = "2248710";
		String emptyAttr1_2 = "-";
		String emptyAttr2_2 = "-";
		String timestamp_2 = "[04/May/1998:22:00:02 +0000]";
		String gmt_2 = "GET /english/hosts/cfo/cfo.html HTTP/1.0";
		Integer objectID_2 = new Integer(200);
		Integer clientID_2 = new Integer(5029);									   
		XATDataValue[] dvs2 = {new XATDataValueImp(ip_address_2),
							   new XATDataValueImp(emptyAttr1_2),
							   new XATDataValueImp(emptyAttr2_2),
							   new XATDataValueImp(timestamp_2),
							   new XATDataValueImp(gmt_2),
							   new XATDataValueImp(objectID_2),
							   new XATDataValueImp(clientID_2)};
		XATTupleImp[] tuples2 = new XATTupleImp[200];				
		for (int i = 0; i < tuples2.length; i++) {
			tuples2[i] = new XATTupleImp(dvs2);
		}
		return tuples2;
	}
}