/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators;
import java.io.File;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
/**
 * 
 * @author Nishant Mehta
 * This class is mainly responsible for parsing the config
 * file. The parser validates the config file 
 * using the xml schema. This can be very useful for error handling
 * 
 */

public class Parser {
	/** All output will use this encoding */
	static final String outputEncoding = "UTF-8";
	/** Constants used for JAXP 1.2 */
	static final String JAXP_SCHEMA_LANGUAGE =
		"http://java.sun.com/xml/jaxp/properties/schemaLanguage";
	static final String W3C_XML_SCHEMA =
		"http://www.w3.org/2001/XMLSchema";
	static final String JAXP_SCHEMA_SOURCE =
		"http://java.sun.com/xml/jaxp/properties/schemaSource";
		
	/**Stores the name of the input file*/
	private String inputFileName = null;
	/**Stores the reference to the document tree*/	
	protected Document doc = null;

	/**
	 * @param inputFileName
	 * Creates a new xsd validating parser
	 */
	public Parser(String inputFileName) {
		this.inputFileName = inputFileName;
//		Step 1: create a DocumentBuilderFactory and configure it
			 DocumentBuilderFactory dbf =
				 DocumentBuilderFactory.newInstance();
		// Set namespaceAware to true to get a DOM Level 2 tree with nodes
		// containing namesapce information.  This is necessary because the
		// default value from JAXP 1.0 was defined to be false.
		//dbf.setNamespaceAware(true);
		//Sets the factory to validating
		//dbf.setValidating(true);
		//dbf.setAttribute(JAXP_SCHEMA_LANGUAGE, W3C_XML_SCHEMA);
		try{		
		DocumentBuilder db = dbf.newDocumentBuilder();

		// Set an ErrorHandler before parsing
		OutputStreamWriter errorWriter =
			new OutputStreamWriter(System.err, outputEncoding);
		db.setErrorHandler(
			new MyErrorHandler(new PrintWriter(errorWriter, true)));

		// Step 3: parse the input file
		doc = db.parse(new File(inputFileName));
										
		}catch (ParserConfigurationException e) {			
			e.printStackTrace();
		} catch (SAXException e) {	
			e.printStackTrace();
		} catch (IOException e) {			
			e.printStackTrace();
		}			
		
	}
	
	// Error handler to report errors and warnings
	private static class MyErrorHandler implements ErrorHandler {
		/** Error handler output goes here */
		private PrintWriter out;

		MyErrorHandler(PrintWriter out) {
			this.out = out;
		}

		/**
		 * Returns a string describing parse exception details
		 */
		private String getParseExceptionInfo(SAXParseException spe) {
			String systemId = spe.getSystemId();
			if (systemId == null) {
				systemId = "null";
			}
			String info = "URI=" + systemId +
				" Line=" + spe.getLineNumber() +
				": " + spe.getMessage();
			return info;
		}

		// The following methods are standard SAX ErrorHandler methods.
		// See SAX documentation for more info.

		public void warning(SAXParseException spe) throws SAXException {
			out.println("Warning: " + getParseExceptionInfo(spe));
		}
        
		public void error(SAXParseException spe) throws SAXException {
			String message = "Error: " + getParseExceptionInfo(spe);
			throw new SAXException(message);
		}

		public void fatalError(SAXParseException spe) throws SAXException {
			String message = "Fatal Error: " + getParseExceptionInfo(spe);
			throw new SAXException(message);
		}
	}
		
}
