/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.client;
import java.io.IOException;
import java.net.ConnectException;
import java.util.HashMap;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.Parser;

/**
 * @author Administrator 
 * This class parses the config file
 */
public class ClientConfigFileParser extends Parser {
	/**This table stores the stream name and its corresponding object reference*/	
	private HashMap streamTable;
	/**Stores the reference to the XATStream*/
	private Stream[] arrStream = null;
	/**Stores the reference to the RecvThreads*/
	private RecvThread[] servers = null;
	/**Stores the reference to the new packet pool*/	
	PacketPool pp = null;
	/**Stores the mapping of the streams to servers*/
	private HashMap serverTable = null;
	
	/**
	 * Constructs a new parser
	 * @param inputFileName
	 * @throws IOException
	 */
	public ClientConfigFileParser(String inputFileName) throws CannotServeStreamException, IOException, ServerNotFoundException {
		super(inputFileName);
		pp = new PacketPool();
		serverTable = new HashMap();
		streamTable = new HashMap();
		traverseTheTree(this.doc);		
	}

	/**
	 * This method traveserse the documet tree and creates an 
	 * array of XATStreams
	 * makes the creates the XATStreams given the specification
	 * of the document
	 * @param document
	 */
	private void traverseTheTree(Document document) throws CannotServeStreamException, IOException, ServerNotFoundException {		
		NodeList nodeList = doc.getElementsByTagName("stream");
		arrStream = new Stream[nodeList.getLength()];
		for(int i=0;i<nodeList.getLength();i++)
		{
			Node node = nodeList.item(i);
			arrStream[i] = makeStream(node);			
		}
		
		nodeList = doc.getElementsByTagName("server");
		
		servers = new RecvThread[nodeList.getLength()];
				
		System.out.println(nodeList.getLength());
		
		for(int i=0;i<nodeList.getLength();i++)
		{	
			Node node = nodeList.item(i);		
			
			servers[i] = makeServer(node);
			
		}	
	}
	
	/**
	 * Makes the recvThreads
	 * @param node
	 * @return
	 */
	private RecvThread makeServer(Node serverNode) throws CannotServeStreamException, IOException, ServerNotFoundException {
		
		NamedNodeMap attributes = serverNode.getAttributes();
		
		String serverName = attributes.getNamedItem("name").getNodeValue();
		 
		String ipAddress = attributes.getNamedItem("ip_address").getNodeValue();
		
		String port = attributes.getNamedItem("port").getNodeValue();
		
		String streamNames = (String) serverTable.get(serverName);
		RecvThread rt = null;				
		try{		
		rt = new RecvThread(pp,serverName,ipAddress,port,streamNames);
		}catch(ConnectException ce)
		{
			System.out.println("Server "+serverName+" does not exist");
			throw(new ServerNotFoundException(serverName));
		}
		return(rt);
	}

	/**
	* This method takes a single stream node and then makes an XATStream out 
	* using the parameters in the node 
	* @return XATStream
	*/
	public Stream makeStream(Node streamNode) throws IOException {
				
		NamedNodeMap attributes = streamNode.getAttributes();
		
		String streamName = attributes.getNamedItem("name").getNodeValue();
		
		String server= attributes.getNamedItem("server").getNodeValue();		
		
		String tempStreamName= "";
		if(this.serverTable.get(server) != null)
		{
			tempStreamName = (String) this.serverTable.get(server);
			tempStreamName = tempStreamName+","+streamName ;		
		}
		else
		 tempStreamName = streamName;
		serverTable.put(server,tempStreamName);
					
		Stream stream = new Stream(streamName);
		streamTable.put(streamName,stream);		
		return (stream);
	}
	
	/**
	 * This returns an array of XATStreams
	 * @return XATStream[]
	 * @throws IOException
	 */
	public HashMap getStreamTable()
	{
			return(streamTable);
	}		
	
	/**
	 * Returns the RecvThreads
	 * @return
	 */
	public RecvThread[] getRecvThreads() {
		return(this.servers);
	}

	/**
	 * @return
	 */
	public PacketPool getPacketPool() {
		return pp;
	}

}
