/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.client;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATMemoryQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPersistentQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObjectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;

/**
 * @author Administrator
 * This class mainly keeps the queue and the stream together
 * There is an instance of this class for each stream element
 * in the configuration file
 */
public class Stream extends XATQueryObjectImp{
	/**Stores the reference to the stream name*/
	private String streamName = null;
	/**Stores the reference to the XATQueue that belongs to this stream*/
	private XATQueue xatQueue = null;
	/**
	 * Constructs a new stream
	 * @param streamName
	 */
	public Stream(String streamName)
	{
		this.streamName = streamName;
		//this.xatQueue = new XATQueueImp(this);
		if(System.getProperty("DATA_MODEL").endsWith("XATMemoryQueueImp")) {
			this.xatQueue = new XATMemoryQueueImp(this);
		}
		else if (System.getProperty("DATA_MODEL").endsWith("XATPersistentQueueImp")) {
			this.xatQueue = new XATPersistentQueueImp(this);
		}
		else {
			System.out.println("!!! Error: Unknow type of DATA_MODEL !!!");
			System.exit(-1);
		}
	}
	
	/**
	 * This method adds the tuple to the XATQueue
	 * @param t
	 */
	public void addTuple(XATTuple t)
	{
		this.xatQueue.enqueue(t,this);
	}

	/**
	 * @return
	 */
	public XATQueue getStreamQueue() {		
		return this.xatQueue;
	}
	
	/**
	 * Sets the schema of the queue with the given schema
	 * @param schema
	 */
	public void setXATQueueSchema(XATQueueSchema schema)
	{
		this.xatQueue.setSchema(schema);
	}
	
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject#getLightWeightQueryObject()
	 */
	public LightWeightQueryObject getLightWeightQueryObject() {
		return new LightWeightQueryObject("Stream",this.getClass().getName(),true);
	}

}
