/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.client;
import java.util.HashMap;

//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.Packet;

/**
 * @author 
 * This class reads the packets from the packet pool and
 * feeds them to the approporite streams
 */
public class StreamFeeder implements Runnable{
	
	/**Stores the reference to the streamTable*/
	HashMap streamTable = null;
	/**Stores the reference to the packet pool*/
	Consumer consumer = null;
	/**Stores the reference to the thread*/
	Thread t= null;
	/**Stores the reference the boolean variable*/
	boolean enable = true;
	/**
	 * Constructs a new StreamFeeder
	 * @param streamTable
	 * @param consumer
	 */
	public StreamFeeder(HashMap streamTable, Consumer consumer) {
		if(streamTable == null || consumer == null)
			throw(new IllegalArgumentException());
		this.streamTable = streamTable;			
		this.consumer = consumer;
		t = new Thread(this,"Stream Feeder");	
		t.setDaemon(true);		
	}
	
	/**
	 * Sends the tuple to the stream it belongs to
	 * @see java.lang.Runnable#run()
	 */
	public void run()
	{
		try {
			while(enable)
			{			
				Packet p = consumer.removePacket();
				
				Stream s = (Stream) streamTable.get(p.getIdentifier());
				
				if(s == null)
				{
					throw(new NoSuchStreamException());
				}
				if(p.getControlMessage().compareTo("schema") == 0)
				{				
					System.out.println("Received the schema for the stream "+p.getIdentifier());
/** Bug: The schema passed from server is always empty, so for now comment out 
 *  the next line to prevent it from setting stream queue schema to be empty. */
//					s.setXATQueueSchema((XATQueueSchema) p.getObject());			
				}
				else
					s.addTuple((XATTuple) p.getObject());
			}
			
		} catch (InterruptedException e) {			
			System.out.println("------Shutting down the Client No Servers Alive----");
		}
	}

	/**
	 * Starts the XATStream Feeder thread 
	 */
	public void start() {	
		t.start();
	}
	
	/**
	 * Disables the thread
	 */
	public void disable()
	{
		this.enable = false;
		t.interrupt();						
	}	

}
