/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.client;

import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;

/**
 * Defines an interface to the stream controller
 * @author: Yali Zhu
 */
public interface XATStreamController extends Runnable, XATQueryObject{
/**
 * Returns the total number of streams
 * @return int
 */
int getStreamTotal();

/**
 * Returns the Queue given that stores the tuples
 * coming from the stream. Note it takes in the stream name
 * as the parameter this stream name should be the same
 * as the one in the stream config file
 * @param name
 * @return XATQueue
 */
public XATQueue getQueue(String streamName);

/**
 * This method connects initialized streams (more precisely: stream queues) with 
 * leave operators (skip source and navigate operators) of a given XATTree.
 *
 * *****************************************************
 * The method requires that a XATTree will have source operators as its leaves.
 * No navigation operators are in the XATTree.
 * *****************************************************
 *
 * @param queryTree
 */
public void connectStreamsWithTree(XATTree queryTree);

/**
 * Starts the thread that generates the streams and 
 * that is starts placing the tuples in the queue
 */
public void startGeneratingStreams();

/**
 * This method is used to add the listener to listen to the
 * end of stream event.
 * Note when the stream client finishes receving the streams.
 * It uses this listener interface to inform that the streams are over
 * @param tester
 */
void addListener(Listener l);

/**
 * Blocks until the stream generator finishes its task.
 */
public void join() throws InterruptedException;

}
