
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////



package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.Socket;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.*;

/**
 * @author 
 * This class stores all the properties of the client
 * like the socket and the object output stream etc...
 */

public class Client {	
	/**Stores the reference to the socket*/
	private Socket socket;
	/**Stores the reference to the objectoutput stream*/
	private ObjectOutputStream outputStream = null;
	private int counter;
	private int clientNumber;
	private NetworkPacket np;
	private long timeToStream;
	private long startTime;
	//private int npsize = 5;
	
	/**
	 * Constructs a new Client
	 * @param socket
	 */
	public Client(Socket socket, long length, int clientNumber) throws IOException
	{
		this.socket = socket;
		this.outputStream = new ObjectOutputStream(socket.getOutputStream());	
		np = new NetworkPacket(Integer.parseInt(System.getProperty("PACKET_SIZE")));	
		counter = 0;		
		timeToStream = length;
		startTime = System.currentTimeMillis();
		this.clientNumber = clientNumber;
		System.out.println("Will stream to " + socket + " for " + timeToStream + " milliseconds");
	}
	
	/**
	 * Sends the packet to the client
	 * @param packet
	 */
	public boolean send(Packet packet) throws IOException {
		
		long currentTime = System.currentTimeMillis();
		
		//Only stream for as long as we are supposed to.
		if(currentTime - startTime > timeToStream){
			this.outputStream.reset();
			this.outputStream.writeObject(np);
			//System.out.println("Sending " + np.size() + " tuples to " + this.socket.toString());
			System.out.println("Done Streaming to " + socket);
			return false;
		}
		
		if(packet.getControlMessage().equalsIgnoreCase("end")) {
		    System.out.println("End of the stream. Send remaining packet and unsubscribe the client");
		    this.outputStream.reset();
			this.outputStream.writeObject(np);
			//System.out.println("Sending " + np.size() + " tuples to " + this.socket.toString());
			return false;
		}
		// Fill up the NetworkPacket object
		if(counter < np.totalSize()){
			String id = packet.getIdentifier();
			Packet p = new Packet(id+clientNumber,packet.getObject());
			np.addPacket(p);
			counter++;
			return true;
		}
		//When it is full, send it and start collecting more packets.
		else{
			// take the following line to make gui faster. 
			//System.out.println("Sending " + np.size() + " tuples to " + this.socket.toString());

			this.outputStream.reset();
			this.outputStream.writeObject(np);
			this.outputStream.flush();
			np = null;
			np = new NetworkPacket(Integer.parseInt(System.getProperty("PACKET_SIZE")));
			counter=1;
			String id = packet.getIdentifier();
			Packet p = new Packet(id+clientNumber,packet.getObject());
			np.addPacket(p);

			return true;
		}
			
	}

	
	/**
	 * Closes the socket
	 */
	public void close(){
		try {
			np = null;
			counter = 0;
			socket.close();
			outputStream.close();
		} catch (IOException e) {
			//System.out.println("---Got Exception when trying to close the socket---");			
		}
				
	}
	
	public String toString(){
		String ret = new String();
		
		ret+= this.socket.getInetAddress() + ":" + this.socket.getPort();
		
		return ret;
	}

}
