
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////


package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValueImp;
import java.util.StringTokenizer;
import java.util.ArrayList;

/**
 * This stores the information about a particular column in the
 * the stream schema
 */
public class Column {
	/**Stores the reference to the coulmn name*/
	private String columnName = null;
	/**Stores the reference to the column type*/
	private String type = null;
	/**
	 * Constructs a new column
	 * @param id
	 * @param type
	 */
	public Column(String id,String type) {
		this.columnName = id;
		this.type = type;
	}
	


	/**
	 * @return
	 */
	public String getColumnName() {
		return columnName;
	}

	/**
	 * @return
	 */
	public String getType() {
		return type;
	}
	
	/**
	 * Makes a new XATDataValue depending on the type and then
	 * new the XATDataVAlue will help in making the tuples
	 * @param value
	 * @return
	 */
	public XATDataValue makeXATDataValue(String value) throws RecordParseException {	
		try{
			if (value.equals("*")) { //punctuation
				return new XATDataValueImp();
			} 
			
			if (value.startsWith("{") && value.endsWith("}")) {
				value = value.substring(1, value.length()-1);
				StringTokenizer st = new StringTokenizer(value, ",");
				ArrayList valueList = new ArrayList();
				
				if(type.toLowerCase().compareTo("string") == 0) {
					while (st.hasMoreElements()) {
						valueList.add(st.nextToken().toString());				
					}
				} else if(type.toLowerCase().compareTo("integer") == 0
                          || type.toLowerCase().compareTo("int") == 0) {
					while (st.hasMoreElements()) {
						valueList.add(new Integer(st.nextToken().toString()));
					}
				} else if(type.toLowerCase().compareTo("long") == 0) {
					while (st.hasMoreElements()) {
						valueList.add(new Long(st.nextToken().toString()));
					}
				} else if(type.toLowerCase().compareTo("float") == 0) {
					while (st.hasMoreElements()) {
						valueList.add(new Float(st.nextToken().toString()));
					}
				} else if(type.toLowerCase().compareTo("double") == 0) {
					while (st.hasMoreElements()) {
						valueList.add(new Double(st.nextToken().toString()));
					}
				}
				return new XATDataValueImp(valueList);
			}
			
			if(type.toLowerCase().compareTo("string") == 0)	{
				return new XATDataValueImp(value);				
			} else if(type.toLowerCase().compareTo("integer") == 0
                    || type.toLowerCase().compareTo("int") == 0) {
				return new XATDataValueImp(new Integer(value));
			} else if(type.toLowerCase().compareTo("long") == 0) {
				return new XATDataValueImp(new Long(value));
			} else if(type.toLowerCase().compareTo("float") == 0) {
				return new XATDataValueImp(new Float(value));
			} else if(type.toLowerCase().compareTo("double") == 0) {
				return new XATDataValueImp(new Double(value));
			}
			return new XATDataValueImp(value);
		} catch(Exception e) {
			//System.out.println("The value "+ value + " is not a valid" + this.type + ". Will store as a string.");
			return new XATDataValueImp(value);
			//throw(new RecordParseException(e));		
		}
	}

	/**
	 * @param value
	 * @return
	 * @throws RecordParseException
	 */
	public int makePunctType(String value) throws RecordParseException {	
		if (value.equals("*")) {
			return XATPunct.ALL_VALUE;
		} else if (value.startsWith("{") && value.endsWith("}")) {
			return XATPunct.SET_VALUE;
		} else if ((value.startsWith("(") || value.startsWith("[")) 
					&& (value.endsWith(")") || value.endsWith("]"))) {
			return XATPunct.RANGE_VALUE;
		} else {
			return XATPunct.SINGLE_VALUE;
		}
	}
	
	
	/**
	 * @param value
	 * @return
	 * @throws RecordParseException
	 */
	public int makeSecPunctType(String value) throws RecordParseException {	
		if (value.equals("*")) {
			return XATPunct.ALL_VALUE;
		} else if (value.startsWith("{") && value.endsWith("}")) {
			return XATPunct.SET_VALUE;
		} else if ((value.startsWith("(") || value.startsWith("[")) 
					&& (value.endsWith(")") || value.endsWith("]"))) {
			return XATPunct.RANGE_VALUE;
		} else {
			return XATPunct.SINGLE_VALUE;
		}
	}
	

}
