/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;

import java.lang.Math;
import java.util.ArrayList;
import java.util.Random;
/**
 * Simulating an Erlang distribution. 
 * The Erlang distribution was developed by A. K. Erlang to examine the 
 * number of telephone calls which might be made at the same time to the 
 * operators of the switching stations. This work on telephone traffic 
 * engineering has been expanded to consider waiting times in queueing 
 * systems in general. The distribution is now used in the field of 
 * stochastic processes. We decided to add it to CAPE as one of the possible
 * distributions for data streams
 *  
 * @author: Rimma V. Nehme
 */
public class ErlangRandomGenerator extends RandomGenerator 
{
	double mean;
	
	/* new pseudo random number generator */
	static Random generator = new Random();
		
	//Stores the parameter alpha which is used for generation of erlang distributed random numbers.
	private double alpha;

	//  Stores the parameter lambda which is used for generation of erlang distributed random numbers.
	private double lambda;

    private double helper1;
	
	/*
	public ErlangRandomGenerator()
	{
        this.alpha = 1;
        this.lambda = 1.0;
        this.UpdateHelpers();
	}
	*/
    
	public ErlangRandomGenerator(ArrayList vecIntervals, long seed) 
	{
		super(vecIntervals, seed);
		this.UpdateHelpers();
		generator = new Random(seed);
	}
	
	public ErlangRandomGenerator(ArrayList vecIntervals, double alphaVal, double lambdaVal, long seed)
	{
		super(vecIntervals, seed);
        this.alpha = alphaVal;
        this.lambda = lambdaVal;
        this.UpdateHelpers();
        generator = new Random(seed);
	}

	public double generateRandom() 
	{
	    return paretoRand();
	}
	
	public double paretoRand() 
	{
		return NextDouble();	
	}

	public static double getUniform(Random rgen) 
	{
	    /* returns random variable between 0 and 1*/
	    return rgen.nextDouble();
	}

	public boolean IsValidAlpha(double value)
	{
		return value > 0.0;
	}
	
	public boolean IsValidLambda(double value)
	{	
		return value > 0.0;
	}

    // Updates the helper variables that store intermediate results for generation of erlang distributed random numbers.
    private void UpdateHelpers()
    {
        this.helper1 = -1.0 / this.lambda;
    }

    //Gets the minimum possible value of erlang distributed random numbers.
	public double Minimum()
	{
		return 0.0;	
	}

	// Gets the maximum possible value of erlang distributed random numbers.
	public double Maximum()
	{
		return Double.MAX_VALUE;		
	}

	// Gets the mean value of erlang distributed random numbers.
	public double Mean()
	{
		 return this.alpha / this.lambda;		
	}
	
	// Gets the median of erlang distributed random numbers.
	public double Median()
	{
		return  Double.NaN;		
	}
	
	// Gets the variance of erlang distributed random numbers.
	public double Variance()
	{
		return this.alpha / Math.pow(this.lambda, 2.0);		
	}
	
	// Gets the variance of erlang distributed random numbers.
	public double[] Mode()
	{
		 return new double[] { (this.alpha - 1) / this.lambda };
	}
	
	// Returns an erlang distributed floating point random number.
	public double NextDouble()
	{        
        double product = 1.0;
        for (int i = 0; i < this.alpha; i++)
        {
            product *= getUniform(generator);
        }

        // Subtract product from 1.0 to avoid Math.Log(0.0)
        return this.helper1 * Math.log(1.0 - product);
	}
	
	public double generateRandom(long currentTimeStamp) 
	{
		//super.updateIntervals(currentTimeStamp);
		return NextDouble();
	}

	public void updateMeanInterval(int avgInterval) 
	{
		this.mean = avgInterval;		
	}		
}
