/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;

import java.util.ArrayList;
import java.util.Random;

/**
 * @author: Rimma V. Nehme
 */
public class GaussianRandomGenerator extends RandomGenerator 
{
    // Stores the parameter mu which is used for generation of normal distributed random numbers.
    private double mu = 1.0;
    
	// Stores the parameter sigma which is used for generation of normal distributed random numbers.
	private double sigma = 1.0;
		
    // Stores a precomputed normal distributed random number that will be returned the next time 
    /// Two new normal distributed random numbers are generated every other call to <see cref="NextDouble"/>.
    private double helper1;

    // Stores a value indicating whether <see cref="NextDouble"/> was called twice since last generation of 
    // normal distributed random numbers.
    private boolean helper2;
		
	/* new pseudo random number generator */
	static Random generator = new Random();
	
	/*
	public GaussianRandomGenerator()
	{
		this.mu = 1.0;
		this.sigma = 1.0;
		this.UpdateHelpers();
	}
	*/
    
	public GaussianRandomGenerator(ArrayList vecIntervals, long seed) 
	{
		super(vecIntervals, seed);
		this.UpdateHelpers();
		generator = new Random(seed);
	}
	
	public GaussianRandomGenerator(ArrayList vecIntervals, double muVal, double sigmaVal, long seed)
	{
		super(vecIntervals, seed);
		this.mu = muVal;
		this.sigma = sigmaVal;
        this.UpdateHelpers();
        generator = new Random(seed);
	}
	
	public double generateRandom(long currentTimeStamp) 
	{
		super.updateIntervals(currentTimeStamp);
	    return gaussianRand();
	}
	
	public double generateRandom() 
	{
	    return gaussianRand();
	}
	

	// Updates the helper variables that store intermediate results for generation of normal distributed random 
	// numbers.
	private void UpdateHelpers()
	{
		this.helper1 = 0.0;
		this.helper2 = false;
	}
	
	public double gaussianRand() 
	{
		return NextDouble();	
	}
	
	// Returns a normal distributed floating point random number.
    public double NextDouble()
    {
        if (this.helper2)
        {
            this.helper2 = false;

            return this.helper1;
        }
        else
        {
			this.helper2 = true;
			
			while (true)
            {
                double v1 = 2.0 * getUniform(generator) - 1.0;
                double v2 = 2.0 * getUniform(generator) - 1.0;
                double w = v1 * v1 + v2 * v2;

                if (w <= 1)
                {
                    double y = Math.sqrt(-2.0 * Math.log(w) / w) * this.sigma;
                    this.helper1 = v2 * y + this.mu;
                    return v1 * y + this.mu;
                }
            }
        }
    }
    
	public static double getUniform(Random rgen) 
	{
	    /* returns random variable between 0 and 1*/
	    return rgen.nextDouble();
	}
	
	public void updateMeanInterval(int meaninterval) 
	{	
		//for now not implemented as currently we assume a single interval - the entire stream
	}
}
