/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////


package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;
import java.util.EmptyStackException;

import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.exceptions.NoItemsInListException;

/**
 
 * 
 * * @author Luping Ding
 * A class that implements a linear link list. The list maintains
 * the objects within it in the ascending order of the key values
 * The main used methods will be the removeFirst and the insert methods
 * optimization could be that when insering nodes into the list.
 * We search for a position from the back of the list
 * That might take fewer number of comparisons
 */
public class Linearlist {
	/**Stores the referebce to the head of the list*/
	private Node head = null;
    private Node tail = null;
		
	/**
	 * Constructs a new Linear List
	 */
	public Linearlist() {		
		head = null;	
        tail = null;
	}
	
	/**
	 * Inserts the node into the link list
	 * key- XATStream
	 * Object- Tuple
	 * At the propert order
	 * @param key
	 * @param o
	 */
	public void insertIntoList(long key,Object o)
	{
		Node prev = head;
		Node current = head;
		
		//Insert into the first position of the List
		
		if( head == null)
		{
			Node temp = new Node(key,o,head);			
			head = temp;
            tail = temp;
		}
		else if(key < head.getKey())
			{		
				Node temp = new Node(key,o,head);							
				head = temp;			
			}		
		else
		{			
		//Insert the node in the middlef of the list
		while(current != null)
		{
			if(key >= current.getKey())
			{
			prev = current;
			current = current.getNext();
			}
			else
				break;
		}
			Node temp = new Node(key,o,current);
			prev.setNext(temp);
            if(current == null) tail = temp;
		}
	}
		
	
	/**
	 * Removes the object given as the parameter from the list
	 * @param o
	 * @return
	 */
	
	public Object removeNode(Object o)
	{
		Node prev = head;
		Node current = head;
		Node temp = null;
		
		if(o == head.getO())
		{
			temp = head;
            if(head == tail) tail = tail.getNext();
			head = head.getNext();				
		}
		else
		{
		
			while(current != null )
				{
					if(o != current.getO())
					{
					prev = current;
					current = current.getNext();					
					}
					else
						break;
				}
				if(current == null)
				{
					throw new IllegalArgumentException("No such value in list");		
				}
				
				temp = current;
				if (tail == current) tail = prev;
				prev.setNext(current.getNext());
		}			
		
			return(temp.getO());
	}
	
	/**
	 * Removes the first node from the list and returns that node
	 * @return
	 */
	public Object removeFirst() throws NoItemsInListException
	{		
		if(head == null)
		{
			throw new EmptyStackException();
		}			
		Node temp = null;					
		temp = head;
		if(head == tail) tail = tail.getNext();
		head = head.getNext();										
		return(temp.getO());		
	}
	
	/**
	 * If the list is empty this returns true or else returns false
	 * @return
	 */
	public boolean isEmpty()
	{
		if (head == null)
			return true;
		else
			return false;
	}


	public long getFirstKey()
	{
		return(head.getKey());
	}
    
    public long getLastKey()
    {
        return(tail.getKey());
    }
/**
 * @see java.lang.Object#toString()
 */
public String toString()
{
	Node temp = head;
	String newString = "";
	while(temp != null)
	{
		newString += temp.getKey()+" ";
	 	temp = temp.getNext();	 	
	}
	return(newString);
 }

/**
 * @return
 */
public Object getFirst() {
	return (head.getO());
 }

public Object getLast() {
    return (tail.getO());
}
}

class Node
{
	/**Stores the reference to the next node*/
	private Node next = null;
	/**Since it is a linear list need to store the key value on which 
	 * it is sorted*/
	private long key = 0;
	/**Stores the data item in the list*/
	private Object o = null;
	
	
	/**
	 * Constructs a new Node
	 * @param o
	 * @param next
	 */
	public Node(long key,Object o,Node next)
	{
		this.key = key;
		if(o == null)		
			throw(new IllegalArgumentException());		
		else
		{		
		this.o = o;
		this.next = next;
		}
	}
	
	/**
	 * Constructs a new Node
	 * @param key
	 * @param o
	 */
	public Node(long key,Object o)
	{
		this.key = key;
		if(o == null)
			throw(new IllegalArgumentException());
		else
		{		
		this.o = o;		
		}
	}
	
	/**
	 * @return
	 */
	public long getKey() {
		return key;
	}

	/**
	 * @return
	 */
	public Node getNext() {
		return next;
	}

	/**
	 * @return
	 */
	public Object getO() {
		return o;
	}

	/**
	 * @param node
	 */
	public void setNext(Node node) {
		next = node;
	}

}