/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;

import java.lang.Math;
import java.util.ArrayList;
import java.util.Random;
/**
 * Simulating a Pareto random variable. Used to simulate Internet traffic activity. 
 * @author: Rimma V. Nehme
 */
public class ParetoRandomGenerator extends RandomGenerator 
{
	double mean;
	
	/* new pseudo random number generator */
	static Random generator = new Random();
		
	// Stores the parameter alpha which is used for generation of pareto distributed random numbers.
	private double alpha;

	// Stores the parameter beta which is used for generation of pareto distributed random numbers.
	private double beta;

    // Stores an intermediate result for generation of pareto distributed random numbers.
    // Speeds up random number generation cause this value only depends on distribution parameters 
    // and therefor doesn't need to be recalculated in successive executions of NextDouble
    private double helper1;
	
	/*	
	public ParetoRandomGenerator()
	{
	    this.alpha = 1.0;
	    this.beta = 1.0;
	    this.UpdateHelpers();
	}	
	*/
    
	public ParetoRandomGenerator(ArrayList vecIntervals, long seed) 
	{
		super(vecIntervals);
		this.UpdateHelpers();
		generator = new Random(seed);
	}
	
	public ParetoRandomGenerator(ArrayList vecIntervals, double alphaVal, double betaVal, long seed)
	{
		super(vecIntervals);
	    this.alpha = alphaVal;
	    this.beta = betaVal;
        this.UpdateHelpers();
        generator = new Random(seed);
	}
	
	
	
	
	
	
	
	
	
	
	

	public double generateRandom() 
	{
	    return paretoRand();
	}
	
	public double paretoRand() 
	{
		return NextDouble();	
	}

	public static double getUniform(Random rgen) 
	{
	    /* returns random variable between 0 and 1*/
	    return rgen.nextDouble();
	}

	public boolean IsValidAlpha(double value)
	{
		return value > 0.0;
	}
	
	public boolean IsValidBeta(double value)
	{
		return value > 0.0;
	}

    // Updates the helper variables that store intermediate results for generation of pareto distributed random 
    private void UpdateHelpers()
    {
        this.helper1 = 1.0 / this.beta;
    }

    //Gets the minimum possible value of pareto distributed random numbers.
	public double Minimum()
	{
		return alpha;		
	}

	// Gets the maximum possible value of pareto distributed random numbers.
	public double Maximum()
	{
		return Double.MAX_VALUE;		
	}

	// Gets the mean value of pareto distributed random numbers.
	public double Mean()
	{
		if (this.beta > 1.0)
        {        
			return this.alpha * this.beta / (this.beta - 1.0);
        }
        else
        {        
        	return Double.NaN;
        }		
	}
	
	// Gets the median of pareto distributed random numbers.
	public double Median()
	{
		return alpha * Math.pow(2.0, 1.0 / beta);		
	}
	
	// Gets the variance of pareto distributed random numbers.
	public double Variance()
	{
		if (this.beta > 2.0)
        {             
			return this.beta * Math.pow(this.alpha, 2.0) / Math.pow(this.beta - 1.0, 2.0) / (this.beta - 2.0);
        }
        else
        {        
        	return Double.NaN;
        }		
	}
	
	// Gets the mode of pareto distributed random numbers.
	public double[] Mode()
	{
		return new double[] { this.alpha };
	}
	
	// Returns a pareto distributed floating point random number.
	public double NextDouble()
	{
        return this.alpha / Math.pow(1.0 - getUniform(generator), this.helper1);
	}
		
	public double generateRandom(long currentTimeStamp) 
	{
		//super.updateIntervals(currentTimeStamp);
		return NextDouble();
	}

	public void updateMeanInterval(int avgInterval) 
	{
		this.mean = avgInterval;		
	}
}
