/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////


package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;
import java.util.ArrayList;
/**
 * Insert the type's description here.
 * Creation date: (2/7/2003 3:44:59 PM)
 * @author: Yali Zhu
 */
public class PoissonRandomGenerator extends RandomGenerator 
{
    private double sq;
    private double alxm;
    private double g;
    private double oldm = -1;
    private double xm = 10.0;

	/**
	 * Insert the method's description here.
	 * Creation date: (2/8/2003 11:03:25 PM)
	 * @param xm double
	 */
	public PoissonRandomGenerator(ArrayList vecInterArrivals,double xm) 
	{
		super(vecInterArrivals);
	    this.xm = xm;
	}
	
	/**
	 * Insert the method's description here.
	 * Creation date: (5/7/2003 4:52:39 AM)
	 * @param newxm double
	 * @param newSeed long
	 */
	public PoissonRandomGenerator(ArrayList vecInterArrivals,double newxm, long newSeed) 
	{
		super(vecInterArrivals,newSeed);
	    xm = newxm;
	    seed = newSeed;
	}
	
	/**
	 * Insert the method's description here.
	 * Creation date: (2/7/2003 4:19:47 PM)
	 */
	public double generateRandom(long currentTimeStamp)
	{
		super.updateIntervals(currentTimeStamp);
	    return poissonRand(xm);
	}
	
	/**
	 * Insert the method's description here.
	 * Creation date: (2/7/2003 4:20:59 PM)
	 */
	public double poissonRand(double xm) 
	{
		//static float sq,alxm,g,oldm=(-1.0);  /* oldm is a flag for whether xm has changed */
	  double em,t,y;                        /* since last call. */
	
	  if (xm < 12.0)                       /* Use direct method. */
	  {
	     if (xm != oldm)
	     {
	        oldm = xm;
	        g = Math.exp(-xm);                    /* If xm is new, compute the exponential. */
	     }
	
	     em = -1;
	     t=1;
	
	     do                     /* Instead of adding exponential deviates it is equiv- */
	     {                      /* alent to multiply uniform deviates. We never */
	        ++em;               /* actually have to take the log, merely compare */
	        t *= uniformRand();    /* to the pre-computed exponential. */
	     } while (t > g);
	  }
	  else                      /* Use rejection method. */
	  {
	     if (xm != oldm)        /* If xm has changed since the last call, then pre- */
	     {                      /* compute some functions that occur below. */
	        oldm = xm;
	        sq = Math.sqrt(2.0*xm);
	        alxm = Math.log(xm);
	        g = xm*alxm - gammln(xm+1.0);
		/* The function gammln is the natural log of the gamma function */
	     }
	
	     do
	     {
	        do 	/* y is a deviate froma Lorentzian comparison function */
	        {
	            y = Math.tan(PI*uniformRand());
	
	            em = sq*y+xm; 	/* em is y, shifted and scaled. */
	        } while (em < 0.0); 	/* Reject if in regime of zero probability.*/
	
	        em = Math.floor(em);  	        /* The trick for integer-valued distributions.*/
	        t = 0.9*(1.0+y*y) * Math.exp(em*alxm-gammln(em+1.0)-g);
	        /* The ratio of the desired distribution to the comparison function; we accept or
	        reject by comparing it to another uniform deviate. The factor 0.9 is chosen so
	        that t never exceeds 1. */
	     } while (uniformRand() > t);
	  }
	
	  return em;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (2/10/2003 6:11:02 PM)
	 * @param meaninterval int
	 */
	public void updateMeanInterval(int meaninterval) 
	{
	    xm = meaninterval;
	}
}
