/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////



package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Vector;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTupleImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.RoutingToken;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.RoutingTokenImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.StqmPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.StqmPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.AccessControlModel;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.XATSecPunct;
import edu.wpi.cs.dsrg.xmldb.xat.common.security.XATSecPunctImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATDataValue;

/**
 * @author
 * This stores the complete information about the schema of the
 * stream
 */
public class Schema {

	//rimma: r-token "punctuation" tag for routing information in Query Meshes 
	public static final String STQM_PUNCT_TAG = "@@@";
	
	//rimma: r-token "punctuation" tag for routing information in Query Meshes 
	public static final String RTOKEN_PUNCT_TAG = "@@";
	
	//rimma: security punctuation tag
	public static final String SEC_PUNCT_TAG = "%%";
	
	public static final String PUNCT_TAG = "##";
	/**This stores an array of Columns
	 * Each cloumn contains a name and a type*/
	ArrayList columns = null;
	
	//?
	Object punctValue = null;
	String streamName;
	/**
	 * Makes the schema
	 */
	public Schema(String streamName) {
		columns =  new ArrayList();		
		this.streamName = streamName;
	}

	/**
	 * Adds a column to the schema
	 * @param a
	 */
	public void addColumn(Column a)
	{
		 columns.add(a);
	}
	
	/**
	 * Adds a column
	 * @param columnName
	 * @param type
	 */
	public void addColumn(String columnName,String type)
	{
		Column c = new Column(columnName,type);
		columns.add(c);
	}
	
	
	/**
	 * @param arrayValues
	 * @return
	 */
	private XATTuple makeTuple(String[] arrayValues, boolean punctuation) throws RecordParseException	
	{
		if(arrayValues == null || arrayValues.length != this.columns.size())
			throw(new IllegalArgumentException());
		 
		XATTuple xatTuple = null;
		if (punctuation) {
			xatTuple = new XATPunctImp();
			int[] punctTypeArray = new int[arrayValues.length];
			for(int i=0; i<arrayValues.length; i++) {
				int punctType = ((Column) this.columns.get(i)).makePunctType(arrayValues[i]);
				punctTypeArray[i] = punctType;
			}
			((XATPunct)xatTuple).setPunctTypeArray(punctTypeArray);
		} else {
			xatTuple = new XATTupleImp();
		}
		for(int i=0; i<arrayValues.length; i++) {
			XATDataValue xatDataValue = ((Column) this.columns.get(i)).makeXATDataValue(arrayValues[i]);
			xatTuple.addCell(xatDataValue);	
		}
		return xatTuple;
	}
	

	private XATTuple makeStqmPunctuationTuple(String[] arrayValues) throws RecordParseException	
	{		
		XATTuple xatTuple = new StqmPunctImp();
		
		for(int i=0; i<columns.size(); i++) 
		{
			XATDataValue xatDataValue = ((Column) this.columns.get(i)).makeXATDataValue(arrayValues[i]);
			xatTuple.addCell(xatDataValue);	
		}

		return xatTuple;
	}
	
	/**
	 * @param arrayValues
	 * @return
	 */
	private XATTuple makeRTokenPunctuationTuple(String[] arrayValues, ArrayList operatorStack) throws RecordParseException	
	{
		if(operatorStack == null)
			throw(new IllegalArgumentException());
		 		
		XATTuple xatTuple = null;		

		xatTuple = new RoutingTokenImp();
		int[] punctTypeArray = new int[columns.size()];
			
		((RoutingToken)xatTuple).setRTokenPunctTypeArray(punctTypeArray, operatorStack);					
		
		
		return xatTuple;
	}

	//------------------------------------------------------------------------------------
	//"old sp structure" - 1st ICDE paper
	private XATTuple makeSecurityPunctuationTuple(String[] arrayValues, 
												  AccessControlModel[] acmValues, 
												  String acmSign, 
												  boolean isImmutable) throws RecordParseException	
	{
		if(arrayValues == null || acmValues == null)
			throw(new IllegalArgumentException());
		 		
		XATTuple xatTuple = null;
		

		xatTuple = new XATSecPunctImp();
		int[] punctTypeArray = new int[columns.size()];
			
		for(int i=0; i<columns.size(); i++) 
		{
			int punctType = ((Column) this.columns.get(i)).makePunctType(arrayValues[i]);
			punctTypeArray[i] = punctType;
		}
			
		((XATSecPunct)xatTuple).setSecPunctTypeArray(punctTypeArray, 
													 acmValues, 
													 acmSign, 
													 isImmutable);					
		
		
		for(int i=0; i<columns.size(); i++) 
		{
			XATDataValue xatDataValue = ((Column) this.columns.get(i)).makeXATDataValue(arrayValues[i]);
			xatTuple.addCell(xatDataValue);	
		}
		return xatTuple;
	}

	//------------------------------------------------------------------------------------
	//"new sp structure"	
	private XATTuple makeSecurityPunctuationTuple(String[] arrayValues, 
												  AccessControlModel[] acmValues, 
												  String acmSign,
												  String spType,
												  String spEnforcement,
												  String spTimestamp) throws RecordParseException	
	{
		if(arrayValues == null || acmValues == null)
		{
			throw(new IllegalArgumentException());
		}
	
		XATTuple xatTuple = null;
		
		xatTuple = new XATSecPunctImp();
		int[] punctTypeArray = new int[columns.size()];
	
		for(int i = 0; i < columns.size(); i++) 
		{	
			int punctType = ((Column) this.columns.get(i)).makePunctType(arrayValues[i]);
			punctTypeArray[i] = punctType;
		}
				
		((XATSecPunct)xatTuple).setSecPunctTypeArray(punctTypeArray, 
													 acmValues, 
													 acmSign, 
													 spType,
													 spEnforcement, 
													 spTimestamp);					
		
		for(int i = 0; i < columns.size(); i++) 
		{	
			XATDataValue xatDataValue = ((Column) this.columns.get(i)).makeXATDataValue(arrayValues[i]);
			xatTuple.addCell(xatDataValue);	
		}
		return xatTuple;
	}
	
	//------------------------------------------------------------------------------------
	
	/**
	 * Converts a record read from the file to a Tuple using the
	 * schema
	 * @param record
	 * @param attrDelimeter
	 * @return
	 * @throws RecordParseException
	 */
	public XATTuple parse(String record,String attrDelimeter) throws RecordParseException
	{
		record = record.trim();
		String[] arrayStrings = new String[this.columns.size()];
		MyTokenizer st = null;
		String secDelimeter = "%%";
		
		//tokenizer for extracting security settings
		MyTokenizer stSec = null;
		
		//tokenizer for extracting routing settings
		MyTokenizer stRouting = null;

		boolean punctuation = false;
		boolean sec_punctuation = false;
		boolean rtoken_punctuation = false;
		boolean stqm_punctuation = false;

		if (record.startsWith(Schema.STQM_PUNCT_TAG)) 
		{
			stqm_punctuation = true;
			st = new MyTokenizer(record.substring(2, record.length()), attrDelimeter);			
			
			//re-initialize array of strings, add three more fields for access control fields
			arrayStrings = new String[this.columns.size()];
			
			//pattern fields
			for(int i = 0; i < this.columns.size();i++)
			{
				if (st.hasMoreTokens())
				{
					arrayStrings[i] = st.nextToken().trim();
				}
				else
				{
					arrayStrings[i] = "";
				}
			}
			ArrayList operatorStack = new ArrayList();
			return(makeStqmPunctuationTuple(arrayStrings));	
		}
		else if (record.startsWith(Schema.RTOKEN_PUNCT_TAG)) 
		{
			rtoken_punctuation = true;
			st = new MyTokenizer(record.substring(2, record.length()), attrDelimeter);
			stSec = new MyTokenizer(record.substring(2, record.length()), secDelimeter);
			stRouting = new MyTokenizer(record.substring(2, record.length()), secDelimeter);
			
			//re-initialize array of strings, add three more fields for access control fields
			arrayStrings = new String[this.columns.size()];
			
			//pattern fields
			for(int i=0; i < this.columns.size();i++)
			{
				arrayStrings[i] = st.nextToken().trim();
			}
			ArrayList operatorStack = new ArrayList();
			return(makeRTokenPunctuationTuple(arrayStrings, operatorStack));	
		}
		if (record.startsWith(Schema.SEC_PUNCT_TAG)) 
		{			
			sec_punctuation = true;
			st = new MyTokenizer(record.substring(2, record.length()), attrDelimeter);
			stSec = new MyTokenizer(record.substring(2, record.length()), secDelimeter);
			
			//re-initialize array of strings, add three more fields for access control fields
			arrayStrings = new String[this.columns.size()];
			
			//pattern fields
			for(int i = 0; i < this.columns.size(); i++)
			{
				arrayStrings[i] = st.nextToken().trim();
			}
			
			//extracting security restriction fields
			String spType = "";
			ArrayList srpFields = new ArrayList();
			String srpSign = "+"; //is it a positive or a negative authorization? (default: positive)
			boolean isImmutable = true; //by default SP is immutable (i.e., we do NOT allow to combine user SPs with server SPs, unless the user explicitly allowed so) 
			String spEnforcement = "";
			String spTimestamp = "";
			
			//there are FOUR parts to the (OLD) security punctuation:
			//OLD_SP_STRUCTURE: (1)DDP, (2)SRP, (3)Sign, (4)IsImmutable
			//NEW_SP_STRUCTURE: (1)DDP, (2)SRP, (3)Sign, (4) Enforcement, (5)Timestamp
			for(int i = 0; i < 6; i++) //OLD_SP_STRUCTURE: i = 4
			{
				if(i == 0)
				{
					//do nothing (it's a data description part)
					stSec.nextToken();
				}
				else if (i == 1)
				{
					spType = stSec.nextToken().trim();
				}
				else if (i == 2)
				{		
					//extract security restriction values (SRP part)
					String spAccessControlModelsValues = stSec.nextToken().trim();				
					MyTokenizer tempst = new MyTokenizer(spAccessControlModelsValues, attrDelimeter);
					
					for(int j = 0;tempst.hasMoreTokens();j++)
					{
						String currentACMValue = tempst.nextToken().trim();
						
						//discrete access control
						if (currentACMValue.startsWith("dac:"))
						{
							srpFields.add(new AccessControlModel(AccessControlModel.DAC_MODEL, currentACMValue.substring(5), this.streamName));
						}
						//mandatory access control
						else if (currentACMValue.startsWith("mac:"))
						{
							srpFields.add(new AccessControlModel(AccessControlModel.MAC_MODEL, currentACMValue.substring(5), this.streamName));
						}
						//flat role-based access control
						else if (currentACMValue.startsWith("frbac:"))
						{
							srpFields.add(new AccessControlModel(AccessControlModel.FLAT_RBAC_MODEL, currentACMValue.substring(6), this.streamName));
						}
						//flat purpose-based access control
						else if (currentACMValue.startsWith("fpbac"))
						{
							srpFields.add(new AccessControlModel(AccessControlModel.FLAT_PBAC_MODEL, currentACMValue.substring(6), this.streamName));
						}
						
					}
				}
				else if (i == 3)
				{
					//extract the sign of the security restriction
					srpSign = stSec.nextToken().trim();
				}
				else if (i == 4)
				{
					/*
					 * OLD_SP_STRUCTURE
					//check if it's an Immutable security punctuation (i.e., if it can be combined with server side SPs)
					String srpImmutable = stSec.nextToken().trim();
					if (srpImmutable.toUpperCase().startsWith("T"))
					{
						isImmutable = true;
					}
					else
					{
						isImmutable = false;
					}
					*/
					
					//NEW_SP_STRUCTURE
					spEnforcement = stSec.nextToken().trim();
				}
				else if (i == 5)
				{
					spTimestamp = stSec.nextToken().trim();
					spTimestamp = spTimestamp.replace("|","");
				}
				
			}//end for
			
			//assign security restrictions to the arrayAcmModels array			
			AccessControlModel[] arrayAcmModels = new AccessControlModel[srpFields.size()];
			for(int i = 0; i < srpFields.size(); i++)
			{
				arrayAcmModels[i] = (AccessControlModel)srpFields.get(i);
			}
			
			//OLD_SP_STRUCTURE
			//return(makeSecurityPunctuationTuple(arrayStrings, arrayAcmModels, srpSign, isImmutable));
			
			//NEW_SP_STRUCTURE
			return(makeSecurityPunctuationTuple(arrayStrings, 
												arrayAcmModels, 
												srpSign, 
												spType,
												spEnforcement,
												spTimestamp));
		} 
		else if (record.startsWith(Schema.PUNCT_TAG)) 
		{
			punctuation = true;
			st = new MyTokenizer(record.substring(2, record.length()),attrDelimeter);
		} 
		else 
		{
			st = new MyTokenizer(record.substring(0, record.length()),attrDelimeter);
		}
		
		try
		{
		
			for(int i = 0; st.hasMoreTokens(); i++)
			{
				arrayStrings[i] = st.nextToken().trim();
			}
		}
		catch(ArrayIndexOutOfBoundsException ae)
		{
			//System.out.println("Possible Corrupt Record: " + record +".");
		}
		return(makeTuple(arrayStrings, punctuation));				
	}

	/**
	 * @param columnName
	 * @return
	 */
	public ArrayList getColumns() 
	{
		return columns;
	}
	
	/**
	 * @param columnName
	 * @return
	 */
	public int getPosition(String columnName) {
		int i = 0;
		for(;i<columns.size();i++)
		{
			if(((Column) columns.get(i)).getColumnName().compareTo(columnName) == 0)
				break;
			
		}
		return(i);
	}
	
	/**
	 * @return
	 */
	public XATQueueSchema makeXATSchema()
	{
		Vector vecColNames = new Vector();
		for(int i=0;i<this.columns.size();i++)
		{
			vecColNames.add(((Column)this.columns.get(i)).getColumnName());
		}		
		XATQueueSchema schema = new XATQueueSchema(vecColNames);
		schema.setName(streamName);
		return(schema);
	}

}
/**
 * @author
 * This class uses the string tokenizer but considers the string within double quotes
 * the same token even if it has delimeters
 */
class MyTokenizer
{

	StringTokenizer st = null;
	/**
	 * @param record
	 * @param attrDelimeter
	 */
	public MyTokenizer(String record, String attrDelimeter) {		
		st = new StringTokenizer(record,attrDelimeter);
	}
	
		
	/**
	 * @return
	 */
	public boolean hasMoreTokens()
	{
		return(st.hasMoreTokens());
	}
	
	/**
	 * Returns the new token
	 * @return
	 */
	public String nextToken()
	{	
		String token = "";	
		String newToken = st.nextToken();		
		if(newToken.charAt(0) == '"')
		{	
			if(!(newToken.charAt((newToken.length()-1))== '"'))
			{
			
			while(true)
			{
				try
				{
					newToken = newToken + " "+token;
	
					if (token.equals(""))
					{
						break;
					}
					else if((token = st.nextToken()).endsWith("\""))
					{	newToken = newToken + " "+ token;
						break;
					}
					
				}
				catch(Exception ex)
				{
					System.out.println(ex.getMessage());
				}
			}
		}
	}
		return(newToken);		
}
	
	
	
}

