/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////


package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;
import java.io.IOException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.*;

/**
 * @author 
 *
 * This class sends the tuple to the clients who have
 * registered for a particular stream
 */
public class Sender 
{	
	/**This stores the referece to the client pool*/
	private Accessor accessor = null;
	
	/**
	 * Constructs a sender
	 * @param accessor
	 */
	public Sender(Accessor accessor)
	{
		if(accessor == null)
			throw(new IllegalArgumentException());
		this.accessor = accessor;
	}
	
	/**
	 * This method constructs a new packet with the tuple
	 * and the stream name ahd then sends the packets to the
	 * clients
	 * @param streamName
	 * @param tuple
	 */
	public void send(String streamName, XATTuple tuple) throws ClientDisconnectedException, NoClientSubscribedException 	
	{
		if(streamName == null || tuple == null)
		{
			//System.out.print(streamName);
			//System.out.print(tuple.toString());
			throw new IllegalArgumentException();
		}
		
		Client[] arrClient= this.accessor.getClients(streamName);
		
		Packet packet = new Packet(streamName, tuple, "tuple");
		if(arrClient.length != 0)
		{
			this.sendPacket(arrClient, packet); 	
			//System.out.println(packet.toString());		 
		} 
		else 
		{
		    //System.out.println("Client size = 0");
		    throw (new NoClientSubscribedException());
		}
	}

	/**
	 *  
	 */
	public void cleanup() 
	{
        //mjba
		this.accessor = null;
	}

	/**
	 * @param string
	 */
	public void sendEndNotification(String streamName)
	{		
		Client[] arrClient= this.accessor.getClients(streamName);
		Packet packet = new Packet(streamName, "end");
		if(arrClient != null)
		{
			try
			{
		        this.sendPacket(arrClient,packet);
		    } 
			catch(Exception e) 
			{
		        e.printStackTrace();
		    }
		}
	}
		
	
	
	/**
	 * Sends the packet to the other end
	 * @param arrClient
	 * @param packet
	 * @throws IOException
	 */
	private void sendPacket(Client[] arrClient,Packet packet) throws ClientDisconnectedException
	{
		boolean isOK = true;
	
		if(arrClient == null)
			throw(new IllegalArgumentException());
			for(int i = 0;i < arrClient.length; i++)
			{			
				try 
				{			
					isOK = arrClient[i].send(packet);
					
					if(!isOK)
					{
						accessor.removeClient(arrClient[i]);
					}
					
				} 
				catch (IOException e) 
				{
					accessor.removeClient(arrClient[i]);
					System.out.println("--Client " +arrClient[i].toString() +" disconnected from " + packet.getIdentifier() +" --");	
					throw(new ClientDisconnectedException());
				}							
			}
		}					
	}

