/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;
import java.io.IOException;
import java.util.ArrayList;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.*;

/**
 * @author Administrator 
 * This class parses the config file
 */
public class ServerConfigFileParser extends Parser 
{
	/**Stores the reference to the XATStream*/
	private XATStream[] arrXATStream = null;
	/**Stores the total amount of time*/
	private long totalTime = 0;
	/**Stores the reference to the factory*/
	private Factory factory = null;
		
	/**
	 * Constructs a new parser
	 * @param inputFileName
	 * @throws IOException
	 */
	public ServerConfigFileParser(String inputFileName) throws IOException, ParsingException 
	{
		super(inputFileName);
		factory = Factory.getInstance();
		traverseTheTree(this.doc);
	}

	/**
	 * This method traveserse the documet tree and creates an 
	 * array of XATStreams
	 * makes the creates the XATStreams given the specification
	 * of the document
	 * @param document
	 */
	private void traverseTheTree(Document document) throws IOException, ParsingException 
	{		
		NodeList nodeList = doc.getElementsByTagName("stream");
		arrXATStream = new XATStream[nodeList.getLength()];
		for(int i=0;i<nodeList.getLength();i++)
		{
			Node node = nodeList.item(i);
			arrXATStream[i] = makeStream(node,i);			
		}
		
		//the total streaming value for all the machines
		nodeList = doc.getElementsByTagName("total_time");
		totalTime = Long.parseLong(
								nodeList
								.item(0)
								.getAttributes()
								.getNamedItem("value")
								.getNodeValue());	
		
		nodeList = doc.getElementsByTagName("packet_size");
		if(nodeList.getLength() > 0)
		{
		    System.setProperty("PACKET_SIZE", nodeList
												.item(0)
												.getAttributes()
												.getNamedItem("value")
												.getNodeValue());
		}
		else
		{
			System.setProperty("PACKET_SIZE", "20");
		}
		
		//Control if data files are repeated or not when EOF is encountered.
		nodeList = doc.getElementsByTagName("repeat_file");
		if(nodeList.getLength() > 0)
		{
		    System.setProperty("REPEAT_FILE", nodeList
												.item(0)
												.getAttributes()
												.getNamedItem("value")
												.getNodeValue());
		}		
		else
		{
			System.setProperty("REPEAT_FILE", "TRUE");
		}		
	}
	
	/**
	* This method takes a single stream node and then makes an XATStream out 
	* using the parameters in the node 
	* @return XATStream
	*/
	public XATStream makeStream(Node streamNode,int streamNumber) throws IOException, ParsingException 
	{
		StreamTupleReader str = null;				
		NamedNodeMap attributes = streamNode.getAttributes();		
		String streamName = attributes.getNamedItem("name").getNodeValue();		
		String[] inputFileNames = null;
		
		Schema schema = null;
		NodeList nl = streamNode.getChildNodes();
		String attrDelimeter = null;
		char recDelimeter = (char) -1;
				
		for(int i = 0;i < nl.getLength();i++)
		{
			if(nl.item(i).getNodeName().compareTo("delimiter") == 0)
			{
				Node delimiterNode = nl.item(i);
				 attributes = delimiterNode.getAttributes();
				 attrDelimeter = attributes.getNamedItem("attribute").getNodeValue();
				 
				String strRecDelimeter = attributes.getNamedItem("record").getNodeValue();
				
				recDelimeter = convertToCharacter(strRecDelimeter);												 				
			}
			if(nl.item(i).getNodeName().compareTo("files") == 0)
			{				
				inputFileNames = getFileNames(nl.item(i));
			}
			if(nl.item(i).getNodeName().compareTo("schema") == 0)
			{
				schema = makeSchema(nl.item(i), streamName);
			}
			if(nl.item(i).getNodeName().compareTo("inter_arrival") == 0)
			{
				if(nl.item(i).hasChildNodes())
				{
					Node n = nl.item(i).getChildNodes().item(1);
					
					RandomGenerator rg = makeRandomGenerator(n);
					rg.setStreamName(streamName);
					
					str = factory.makeStreamTupleReader(streamName,
														inputFileNames,
														attrDelimeter,
														recDelimeter,
														schema,
														rg);
				}
				else
				{
					Node n = nl.item(i).getAttributes().getNamedItem("attribute");
					if(n == null)
					{
						throw (new ParsingException("The inter_arrival element has to have an attribute or a distribution as a sub element"));
					}
					String columnName = n.getNodeValue();
					
					n = nl.item(i).getAttributes().getNamedItem("format");
					if(n == null)
				   	{
						throw new ParsingException("The inter_arrival element must have the attribute format");	
				   	}
				   	String format = n.getNodeValue();
				   				
					str = factory.makeStreamTupleReader(streamName,
														inputFileNames,
														attrDelimeter,
														recDelimeter,
														schema,
														columnName,
														format);	
				}
			}			
		}						
		
		XATStream s = new XATStreamImp(streamName,streamNumber,inputFileNames,str);
		return(s);																
	}


	/**
	 * @param strRecDelimeter
	 * @return
	 */
	private char convertToCharacter(String strRecDelimeter) 
	{
		char result = (char) -1;
		if(strRecDelimeter.compareTo("\\n") == 0)
		{
			result = '\n';
		}
		else if(strRecDelimeter.compareTo("\\t") == 0)
		{
			result = '\t';
		}
		else 
			result = strRecDelimeter.charAt(0);
		return(result);
	}
	
	/**
	 * Creates a generator with the specifies stats values for the stream.
	 * 
	 * @param node
	 * @return
	 */
	private RandomGenerator makeRandomGenerator(Node distributionNode) throws ParsingException 
	{
		NamedNodeMap attr = distributionNode.getAttributes();
		String distribution = attr.getNamedItem("value").getNodeValue();
		int seed = Integer.parseInt(attr.getNamedItem("seed").getNodeValue());
		
		RandomGenerator rg = null;
				
		//Note making a new array list, passing the number of 		
		ArrayList vecIntervals = new ArrayList();
		
		NodeList intervals = distributionNode.getChildNodes();
		//Note the fist element of the array contains the 
		//starttime and the second element contains the mean
		int length = intervals.getLength();
		Node intervalNode = null;
		for (int i = 1; i < intervals.getLength()-1; i = i+2) 
		{						
			intervalNode = intervals.item(i);			 			
			double[] dblArray = getInterval(intervalNode);											
			vecIntervals.add(dblArray);
		}					
		//-------------------------------------------------------------------------
		// POISSON DISTRIBUTION
		//-------------------------------------------------------------------------
		if (distribution.compareTo("poisson") == 0) 
		{	
			System.out.println("-------------------------------------------------");
			System.out.println("Poisson distribution");
			if (seed != -1) 
				rg = new ExponentialRandomGenerator(vecIntervals, seed);				
		}
		//-------------------------------------------------------------------------
		// GAUSSIAN DISTRIBUTION
		//-------------------------------------------------------------------------
		else if (distribution.compareTo("gaussian") == 0) 
		{
			System.out.println("-------------------------------------------------");
			System.out.println("Gaussian distribution");
			int mu = Integer.parseInt(attr.getNamedItem("mu").getNodeValue());
			int sigma = Integer.parseInt(attr.getNamedItem("sigma").getNodeValue());
			rg = new GaussianRandomGenerator(vecIntervals, mu, sigma, seed);
		} 
		//-------------------------------------------------------------------------
		// FIXED RANDOM DISTRIBUTION
		//-------------------------------------------------------------------------
		else if (distribution.compareTo("fixed") == 0) 
		{
			System.out.println("-------------------------------------------------");
			System.out.println("Fixed distribution");
			rg = new FixedRandomGenerator(vecIntervals);
		}
		//-------------------------------------------------------------------------
		// PARETO RANDOM DISTRIBUTION
		//-------------------------------------------------------------------------
		else if (distribution.compareTo("pareto") == 0) 
		{
			System.out.println("-------------------------------------------------");
			System.out.println("Pareto distribution");
			int alpha = Integer.parseInt(attr.getNamedItem("alpha").getNodeValue());
			int beta = Integer.parseInt(attr.getNamedItem("beta").getNodeValue());
			rg = new ParetoRandomGenerator(vecIntervals, alpha, beta, seed);
		}
		//-------------------------------------------------------------------------
		// ERLANG RANDOM DISTRIBUTION
		//-------------------------------------------------------------------------
		else if (distribution.compareTo("erlang") == 0) 
		{
			System.out.println("-------------------------------------------------");
			System.out.println("Erlang distribution");
			int alpha = Integer.parseInt(attr.getNamedItem("alpha").getNodeValue());
			int lambda = Integer.parseInt(attr.getNamedItem("lambda").getNodeValue());
			rg = new ErlangRandomGenerator(vecIntervals, alpha, lambda, seed);
		}
		return(rg);
	}

	/**
	 * @param node
	 * @return
	 */
	private Schema makeSchema(Node schemaNode, String streamName) 
	{		
		Schema schema = new Schema(streamName);		
		NodeList nl = schemaNode.getChildNodes();
		for(int i = 0;i < nl.getLength(); i++)
		{
			Node node = null;
			if(nl.item(i).getNodeName().compareTo("attribute") == 0)
			{
				NamedNodeMap attributes = nl.item(i).getAttributes();
				String name = attributes.getNamedItem("name").getNodeValue();
				String type = attributes.getNamedItem("type").getNodeValue();
				schema.addColumn(name,type);				
			}
		}
		return(schema);	
	}
	
	/**
	 * @param node
	 * @return
	 */
	private String[] getFileNames(Node fileNode) 
	{		
		NodeList nl = fileNode.getChildNodes();
		String[] files = new String[nl.getLength()];
		int count = 0;
		for(int i = 0; i < nl.getLength(); i++)
		{
			Node node = null;
			if(nl.item(i).getNodeName().compareTo("file") == 0)
			{
				NamedNodeMap attributes = nl.item(i).getAttributes();
				String name = attributes.getNamedItem("name").getNodeValue();
				files[count] = name;
				count++;
			}
		}
		return files;	
	}
	

	/**
	 * This method parses a single interval node and then returns the 
	 * start time and the mean in a double array
	 * @param intervalNode
	 * @return
	 */
	private double[] getInterval(Node intervalNode) 
	{		
		double[] arrDouble = new double[2];		
		NamedNodeMap attributes  = intervalNode.getAttributes();		
		arrDouble[0] = Double.parseDouble(attributes.getNamedItem("start_time").getNodeValue());		
		arrDouble[1] = Double.parseDouble(attributes.getNamedItem("mean").getNodeValue());								
		return(arrDouble);		
	}

	/**
	 * This returns an array of XATStreams
	 * @return XATStream[]
	 * @throws IOException
	 */
	public XATStream[] getStreams() throws IOException
	{
		return(arrXATStream);
	}		
	
	/**
	 * Returns the total amount of time
	 * @return
	 */
	public long getTotalTime() 
	{
		return totalTime;
	}
}
