/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;
import java.io.BufferedInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueueSchema;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;

/**Note still left to code will code once we finish coding the new parser for the server config file

/**
 * A Parser that parses the file generated by the 
 * data generator
 * @author Luping Ding
 *
 */
public abstract class StreamTupleReaderTextImp  implements StreamTupleReader{
	/**Stores the name of the stream tuple reader*/
	private String identifier = null;
	/**Stores the name of the input File*/
	String[] inputFiles = null;
	/**Sores the attribute delimeter*/
	String attrDelimeter = null;
	/**Stores the record delimeter*/
	char recDelimeter;
	/**Stores the reference to the schema*/
	Schema schema  = null;
	/**Stores the reference to the buffered reader used to lines from the file*/
	BufferedInputStream in = null;		
		
	/**Maintains a count of the number of tuples read*/
	int count = 0;
	/**
	 * @param inputFileName
	 * @param attrDelimeter
	 * @param recDelimeter
	 * @param schema
	 *
	 */
	public StreamTupleReaderTextImp(String identifier,
									String[] inputFileNames, 
									String attrDelimeter,
									char recDelimeter,
									Schema schema) throws FileNotFoundException {
	this.identifier = identifier; //stream name 
	this.inputFiles = inputFileNames;
	this.attrDelimeter = attrDelimeter;
	this.recDelimeter = recDelimeter;
	this.schema = schema;
	
	//gets the input file
	File f = new File(inputFiles[0]);
	double size= f.length()/1048576.0;
	System.out.println("Now generating stream " + identifier + " from file " + inputFiles[0]+".");
	System.out.println("The size of this file is "+size+"MB.");
	
	//reads from the file
	in = new BufferedInputStream(new FileInputStream(inputFiles[0]));
		
	}
		
	
	/**
	 * Parses the input file and returns a new Tuple. Note 
	 * each time the parsing done is tuple by tuple
	 * @return XATStreamTuple
	 */
	
	public XATTuple getNextTuple(long currentTimeStamp) throws IOException, RecordParseException
	{
		XATTuple tuple = null;		
			char ch;
			String strRecord="";

			while((ch = (char) in.read()) != this.recDelimeter)
			{
			    //System.out.println("character read from " + inputFiles[0] + " is " + ch);
				if((int)ch == 65535)
				{ 
					// or -1
					//The file is ended
					//for now repeat the file
				    String repeat = System.getProperty("REPEAT_FILE");
				    if(repeat.compareToIgnoreCase("false") == 0) 
				    {
				        System.out.println("The file " + inputFiles[0] + " has no more entries.");
				        in.close();
				        throw (new EOFException());
				    } 
				    else 
				    {
				        System.out.println("Will start from beginning.");
				        //System.exit(1); //rimma remove
				        
						in.close();
						in = new BufferedInputStream(new FileInputStream(inputFiles[0]));
						strRecord = "";
				    }
				}
				else
				{	
					strRecord = strRecord + ch;
				}
			}
			tuple = schema.parse(strRecord,this.attrDelimeter);
			
			this.setTimeStamp(currentTimeStamp,tuple);
						
			count = count + 1;									
					
		return(tuple);
	}
	
	public abstract void  setTimeStamp(long currentTimeStamp,XATTuple tuple) throws RecordParseException ;
	
	/**
	 * Returns the XATSchema
	 * @see edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server.StreamTupleReader#getXATSchema()
	 */
	public XATQueueSchema getXATSchema() 
	{		
		return this.schema.makeXATSchema();
	}	
}
