/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow StreamGenerator 1.0.
//
//  Copyright 2003-2005, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Bin Liu, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang,Xin Zhang, Yali Zhu, Timothy Sutherland,
//  Bradley Momberger, Ming Li, and Rimma Nehme
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the DSRG Development Team: <dsrg@CS.WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////



package edu.wpi.cs.dsrg.xmldb.xat.component.streamgenerators.server;

import java.io.EOFException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.text.DecimalFormat;
//import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple;

/**
 * @author
 * This is the main class that instantiates all the other components
 * and then starts the streams once the components are ready
 */
public class XATStreamGenerator implements AccessStream, Runnable 
{	
	//------------------------------------------------------------------------------------
	
	/**Stores the reference to the list*/
	private Linearlist linearlist;
    /**Stores the lost of unconnected streams*/
    private LinkedList noClientList;
	/**Stores the name of the server config file Name*/
	private String configFileName = null;
	/**Stores the reference to the XATStreams*/
	private HashMap streams = null;
	/**Stores the reference to the client acceptor*/
	private ClientAcceptor clientAcceptor = null;
	/**Stores the reference to the Client pool*/
	private ClientPool clientPool = null;
	/**Stores the reference to the sender*/
	private Sender sender = null;
	/**Stores the total run time for the stream generator*/
	private long totalRunTime = -1;
	/**Stores the time at which the stream generator starts*/
	private long startTime = -1;
	private XATStream[] arrStreams;
	
	//------------------------------------------------------------------------------------
	
	/**
	 * Constructs a new XATStreamGenerator. 
	 * @param configFileName
	 * @throws IOException
	 */
	public XATStreamGenerator(int portNumber,String configFileName) throws DuplicateStreamNameException, IOException, ParsingException 
	{
		if(configFileName == null ||portNumber < 1)
		{
		 	throw new IllegalArgumentException();
		}
		
		this.configFileName = configFileName;
		streams = new HashMap();
		ServerConfigFileParser sc = new ServerConfigFileParser(configFileName);
		arrStreams = sc.getStreams();
		this.totalRunTime = sc.getTotalTime();
		//Instantiating the linear list
		this.linearlist = new Linearlist();	
        this.noClientList = new LinkedList();
		//Loading the hash table
		for(int i = 0; i < arrStreams.length; i++)
		{
			streams.put(arrStreams[i].getStreamName(),arrStreams[i]);
		}
		
		System.out.println("NO of Streams : "+streams.size());
		clientPool = new ClientPool();
		//Add the streams to the client pool
		for(int i = 0; i < arrStreams.length;i++)
		{
			clientPool.addStream(arrStreams[i].getStreamName());
		}
		
		sender = new Sender(clientPool);
	
		clientAcceptor = new ClientAcceptor(portNumber,clientPool,(AccessStream)this);		
				
		startTime = System.currentTimeMillis();
		
		//initializes the linear list
		makeList();	
		System.out.println("Made list");
		DecimalFormat df = new DecimalFormat("0.00#");
		Runtime rt = Runtime.getRuntime();
		System.out.println("Now generating streams on Port " + portNumber);
		System.out.println("Memory information before generating any streams");
		System.out.println("-------------Memory Information --------------");
		System.out.println("Free memory: " + df.format(((double)rt.freeMemory()/1024.0/1024.0)) + "MB");
		System.out.println("Total memory: " + df.format(((double)rt.totalMemory()/1024.0/1024.0)) + "MB");
		System.out.println("Max memory: " + df.format(((double)rt.maxMemory()/1024.0/1024.0)) + "MB");
		this.run();
//		this.clientAcceptor.disableAcceptor();
//		this.clientPool.cleanup();
//		this.sender.cleanup();
//		System.out.println("\n After cleaning up resources");
//		System.out.println("-------------Memory Information --------------");
//		System.out.println("Free memory: " + df.format(((double)rt.freeMemory()/1024.0/1024.0)) + "MB");
//		System.out.println("Total memory: " + df.format(((double)rt.totalMemory()/1024.0/1024.0)) + "MB");
//		System.out.println("Max memory: " + df.format(((double)rt.maxMemory()/1024.0/1024.0)) + "MB");
	}
	
	//------------------------------------------------------------------------------------
	
	private void printStreamStats(long[] numTuples) 
	{
		DecimalFormat df = new DecimalFormat("0.00#");
		Runtime rt = Runtime.getRuntime();
		
		long totalTuples = numTuples[0] + numTuples[1] + numTuples[2] + numTuples[3] + numTuples[4] + numTuples[5];
		
		System.out.println("------");
		System.out.println("Outputted " + totalTuples + " total tuples.");
		for(int j=0 ; j < streams.size() ; j++)
		{
			System.out.println("Stream " + arrStreams[j].getStreamName() +": "+numTuples[j]);
		}
		//System.out.println("The last tuple outputted was from " + currentStream.getStreamName() +": "+tuple);
		//time = System.currentTimeMillis();	
		
		System.out.println("-------------Memory Information --------------");
		System.out.println("Free memory: " + df.format(((double)rt.freeMemory()/1024.0/1024.0)) + "MB");
		System.out.println("Total memory: " + df.format(((double)rt.totalMemory()/1024.0/1024.0)) + "MB");
		System.out.println("Max memory: " + df.format(((double)rt.maxMemory()/1024.0/1024.0)) + "MB");
		System.gc();
		System.out.println("After calling garbage collection");
		System.out.println("Free memory: " + df.format(((double)rt.freeMemory()/1024.0/1024.0)) + "MB");
		System.out.println("Total memory: " + df.format(((double)rt.totalMemory()/1024.0/1024.0)) + "MB");
		System.out.println("Max memory: " + df.format(((double)rt.maxMemory()/1024.0/1024.0)) + "MB");	
		System.out.println("----------------------------------------------");
	}

	//------------------------------------------------------------------------------------
	

	/**
	 * Actually generates the tuples. Pools each of the
	 * streams to generate the new Tuples
	 */
	public void run() 
	{
		System.out.println("Run");
		long time = System.currentTimeMillis();
		long updatedTime = System.currentTimeMillis();
		long previousTimeStamp = 0;
		
		// Assume you stream no more than 4 streams.
		//counts the number of  tuples per stream
		long[] numTuples = new long[13];
		numTuples[0] = 0;
		numTuples[1] = 0;
		numTuples[2] = 0;
		numTuples[3] = 0;
		numTuples[4] = 0;
		numTuples[5] = 0;
		numTuples[6] = 0;
		numTuples[7] = 0;
		numTuples[8] = 0;
		numTuples[9] = 0;
		numTuples[10] = 0;
		numTuples[11] = 0;
		numTuples[12] = 0;
		
		
		// set debug flag of each stream to be the same as stream controller.
		//--long globaltime = System.currentTimeMillis() - startTime;	
		long clientupdatebase = System.currentTimeMillis();
	    
		while ((!(linearlist.isEmpty()) || !(noClientList.isEmpty())) && (!Thread.currentThread().isInterrupted())) 
		{		
			updatedTime = System.currentTimeMillis();
	        XATStream currentStream = null;				
	        //every .5 sec, check unsubscribed streams for clients.
	        if (updatedTime > clientupdatebase + 500)
	        {	            
	            LinkedList templist = new LinkedList();
	            
	            while(!noClientList.isEmpty())
	            {
	                Object[] orphantuple = (Object[])noClientList.removeFirst();
	                if (clientPool.getClients(((XATStream)orphantuple[0]).getStreamName()).length > 0)
	                {
	                    linearlist.insertIntoList(((Long)orphantuple[2]).longValue(), new Object[]{orphantuple[0], orphantuple[1]});
	                }
	                else
	                {
	                    templist.add(orphantuple);
	                }
	            }
	            noClientList = templist;
	            clientupdatebase = updatedTime;
	        }
	        try 
	        {	        
	        	//waiting for the stream to be scheduled
	        	try 
	        	{	            
	        		//Calculates the amount of time one can sleep                       	        
	        		//long margin = this.linearlist.getFirstKey() - (System.currentTimeMillis() - startTime);
	            
	        		long nextTimeStamp = this.linearlist.getFirstKey();	            
	        		long margin = nextTimeStamp - previousTimeStamp;	            
	        		previousTimeStamp = nextTimeStamp;
	            	            
	        		if (margin > 10) 
	        		{
	        			Thread.sleep(margin);	            
	        		}
	            	    	        
	        	} 
	        	catch (InterruptedException e1) 
	        	{
	        		e1.printStackTrace();	        
	        	} 
	        	catch (NullPointerException npe) 
	        	{	            
	        		throw new IndexOutOfBoundsException();	        
	        	}	
	        	long temptime = linearlist.getFirstKey();		 		
	
	        	//Object[] ob = (Object[]) linearlist.removeFirst();		
	        	Object[] ob = (Object[]) linearlist.getFirst();	
						
	        	currentStream = (XATStream) ob[0];	
	        	//DecimalFormat df = new DecimalFormat("0.00#");
	        	//Runtime rt = Runtime.getRuntime();					
	        	XATTuple tuple = (XATTuple) ob[1];						
										
	        	//send the new tuple			
	        	sender.send(currentStream.getStreamName(),tuple);

	        	//System.out.println("Stream: " + currentStream.getStreamName() + " Tuple: " + tuple);
				linearlist.removeFirst();	
				//System.out.println(tuple.toString());
				numTuples[currentStream.getStreamNumber()]++;
					
				//time was 600 000 before change mj				
				//print memory info every 5 min
				//System.out.println("tuple " + currentStream.getStreamName() + " " + tuple);
				if(updatedTime - time > 300000) 
				{ 				
					printStreamStats(numTuples);
					System.out.println("The last tuple outputted was from " + currentStream.getStreamName() +": "+tuple);
					time = System.currentTimeMillis();					
				}
				
				//next time the same tuple is to be scheduled
				long nextTime = -1;
				XATTuple newtuple = null;
				try 
				{																		
					//tuple = currentStream.makeNewTuple(0);
					newtuple = currentStream.makeNewTuple(Math.max(tuple.getMaxTimeStamp(), updatedTime - time));
	                previousTimeStamp = updatedTime = time;					
	                nextTime = newtuple.getMaxTimeStamp();												
				} 
				catch (RecordParseException e) 
				{				
					System.out.println("Record Parser Exception Removing the stream "+currentStream.getStreamName());											
					e.printStackTrace();								
					continue;										
				} 
				catch (NullPointerException e3)
				{
					// If the tuple comes back null, it means we are restarting
					// the file. So, we have to restart the timestamp as well. (Yali's comment: WHY????)
					//startTime = System.currentTimeMillis();
					try
					{								
						newtuple = currentStream.makeNewTuple(tuple.getMaxTimeStamp());		 					
						nextTime = newtuple.getMaxTimeStamp();							
					}
					catch(Exception e)
					{					
						System.out.println("We have a big problem!");					
					}					
				}
					 
				ob = new Object[2];
				ob[0] = currentStream;
				ob[1] = newtuple;
					 
				//insert this stream back into the list
				linearlist.insertIntoList(nextTime, ob);			
				ob = null;
				tuple = null;
				newtuple = null;
				currentStream = null;
	        }
	        catch(ClientDisconnectedException e) 
	        {			
	        	//System.out.println("Do nothing, just skip creating new tuples for this stream");				
	        }
	        catch(NoClientSubscribedException e) 
	        {				    
	        	//Clients haven't subscribed to the stream sources yet. 
				//Wait for some time and retry again. Don't remove tuple from linearlist.
				/* try{
				    Thread.sleep(500);
				} catch (Exception c) {
				    c.printStackTrace();
				} */
	                
	            // Previous exception handler blocked other streams.
	            // Instead, we'll rotate streams so that streams with subscribed clients are not blocked.				    
	        	Long timestamp = new Long(linearlist.getFirstKey());
	            Object[] item = (Object[])linearlist.removeFirst();
	            noClientList.add(new Object[] {item[0], item[1], timestamp });    	                				
	        }
	        catch (IOException e) 
	        {				
	        	if (e instanceof EOFException) 
	        	{					
	        		System.out.println("The " + currentStream.getStreamName() + " is empty.");
					System.out.println("Removing clients from stream " + currentStream.getStreamName());				
					sender.sendEndNotification(currentStream.getStreamName());
						
					//currentStream.cleanup();
					//sender.sendEndNotification(currentStream.getStreamName());
					//U don't need to remove the stream from the list.
					//just do not insert it back again
					//linearlist.removeFirst();					
	        	} 
	        	else 
	        	{				
	        		printStreamStats(numTuples);
					e.printStackTrace();				
	        	}	            
	        } 
	        catch (IndexOutOfBoundsException ioobe) 
	        {	        
	        	// Do nothing, just wait and continue when there are no streams with clients.	        
	        	try {Thread.sleep(500);}catch(InterruptedException ie)
	        	{
	        		
	        	}	            
	        }	            		
		}
		
		if (Thread.currentThread().isInterrupted()) 
		{		
			System.out.println("Streams interrupted.");
			printStreamStats(numTuples);
		} 
		else 
		{	
			printStreamStats(numTuples);
			this.cleanup();
			System.out.println("End of streams.");
		}	
		return;
	}

	
	//------------------------------------------------------------------------------------
	/**
	 * Cleans up all the components 
	 */
	private void cleanup() 
	{
		try 
		{
			this.clientPool.cleanup();		
			this.sender.cleanup();
			this.clientAcceptor.disableAcceptor();
		} 
		catch (IOException e) 
		{
			System.out.println("--Error when cleaning up--");					
			e.printStackTrace();
		}		
	}
	
	//------------------------------------------------------------------------------------

	private void makeList() throws IOException
	{		
		XATStream tempStream;	
		//long timer = System.currentTimeMillis() - startTime;
	
		//The first tuple from each stream has timestamp 0. 
		//This ensures that tuples have exactly the same timestamp from each run.
		long timer = 0; 	
	
		Collection collection = streams.values();	
		for (Iterator iter = collection.iterator(); iter.hasNext();) 
		{		
			tempStream = (XATStream) iter.next();					
			XATTuple tuple = null;
		
			long timeStamp = -1;
			try 
			{
				System.out.println("Initialize Stream " + tempStream.getStreamName());
				tuple = tempStream.makeNewTuple(timer);
				timeStamp = tuple.getMaxTimeStamp();
			} 
			catch (IOException e) 
			{
				System.out.println("Removing the current Stream,IOExeption  "+tempStream.getStreamName());						
			} 
			catch (RecordParseException e) 
			{			
				System.out.println("Removing the current Stream, Record Parse Exception "+tempStream.getStreamName());
				e.printStackTrace();		
			}		

			//Note change the list to use long from now on
			Object[] ob = new Object[3];
			ob[0] = tempStream;
			ob[1] = tuple;
			ob[2] = new Long(timeStamp);
			this.noClientList.add(ob);	
		}
	}

	//------------------------------------------------------------------------------------

	/**
	 * Returns the instance of the stream that has the stream name passed in as agrument
	 * @param streamName
	 * @return
	 */
	public XATStream getStream(String streamName)
	{
		return((XATStream)streams.get(streamName));		
	}

	//------------------------------------------------------------------------------------

	public static void main(String[] args) throws NumberFormatException, ParsingException 
	{
		if(args.length != 2)
		{
			System.out.println("Usage: XATStreamGenerator <portnumber> <configfile.xml>");
			System.exit(0);
		}
	
		try 
		{
			XATStreamGenerator streamGenerator = new XATStreamGenerator(Integer.parseInt(args[0]),args[1]);
			//streamGenerator.run();
			//System.out.println("Finished. Return.");
			//System.exit(0);
			return;	
		} 
		catch (IOException e) 
		{		
			e.printStackTrace();
		}
		catch(DuplicateStreamNameException e)
		{
			e.printStackTrace();
		}
	}
	
	//------------------------------------------------------------------------------------
}
