/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Query Mesh 1.0 - Adaptive Data Stream Query Processing.
//
//  Copyright 2010, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Chuan Lei, Justin Lutz, 
//	Guy D. Mann, Mukesh Mulchandani, Brian Murphy, Rimma Nehme, 
//	Bradford Pielech, Alex Proschitsky, Elke A. Rundensteiner, Ling Wang,
//	Song Wang, Karen Works, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/PROJECTS/QM/index.html
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the QueryMesh Development Team: <QueryMesh@WPI.EDU>.

package edu.wpi.cs.dsrg.STeM;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATMultiQueueStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.windowstreamoperator.XATMultiQueueWindowStreamOperatorImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Iterator;
import java.util.BitSet;

import edu.wpi.cs.dsrg.Mjoin.MJoinData;
import edu.wpi.cs.dsrg.eddy.datastructure.*;
import edu.wpi.cs.dsrg.eddy.operator.EddyOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.execution.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.OperatorIndividualSample;
import edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.selftuning.RouteSignatureStorage;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.DistributedExecutionController;

public class STeMJoinProbeOperatorImp extends XATMultiQueueWindowStreamOperatorImp implements STeMJoinProbeOperator
{	
	
	//-------------------------------------------------------------------------------------
	
	public int windowSize;
	String outFileJoinOverhead = "C:\\QueryMesh\\execution_stats\\JOIN_MONITORING.txt";
	//String outFileExeStatsOverhead = "C:\\Rimma\\Research\\QueryMesh\\execution_stats\\EXECUTION_STATS_MONITORING.txt";
	
	long numProcessedTuples;
	
	// is this an eddy op or not
	private boolean isEddyOp = false;
	
	//STQM: is self-tuning?
	private boolean isSelfTuning = false; //CHUAN LEI - TEST QM
	
	
	private Hashtable hashOldSTQMRusterExploitationStatisticsLocalStorage = new Hashtable(); //exploitation execution statistics used when QM was last adapted 
	private Hashtable hashOldSTQMRusterExplorationStatisticsLocalStorage = new Hashtable(); //exploration execution statistics used when QM was last adapted
	
	private Hashtable hashNewSTQMRusterExploitationStatisticsLocalStorage = new Hashtable();
	private Hashtable hashNewSTQMRusterExplorationStatisticsLocalStorage = new Hashtable();
	
	// index position of STem and tuple that are being compared
	
	// data set associated with STeM
	private STeMData STeMStreamData;
	
	//Info about probe
	private int ProbeSTREAMId;
	private ArrayList<Integer> NumTupleIndex;
	
	//Info about stored stem data
	private int DataSTREAMId;
	private ArrayList<Integer> NumSTeMIndex;
	private int STeMDataIndexArrayPos = 0;
	
	//operator bitmaps per STREAM
	ArrayList<BitSet> PreFilledPerStream;
	
	//used for QMesh to look up operators
	private OperatorLookUp OpLkUp;
	
	private int QMeshOperatorID;

	//used to determine when to create ruster
	int rusterSendOff;

	//-------------------------------------------------------------------------------------
	
	public STeMJoinProbeOperatorImp(XATNode init) 
	{
		super(init);
	}
		
	//-------------------------------------------------------------------------------------
	
	public boolean run() throws Exception 
	{		
		int totalNumberTupledEnq = 0;
		RoutingToken currentRToken = null;
		
		//STQM
		RoutingToken lastRToken = null;
		boolean bIsFirstRtoken = true;
		int lastRusterTuplesIn = 0;
		int lastRusterTuplesOut = 0;
		
		//input queue from STREAM
		XATQueueSchema inSchema;
		int inLength;
		
		//------------------------------------------------------------------------------
		// get input queue by operator id from the lookup table
		//------------------------------------------------------------------------------
		//XATQueue inputQueue = (XATQueue)this.OpLkUp.getOperatorQueue(this.getOperatorID());
		XATQueue inputQueue = (XATQueue)this.getInputQueue();
		
		if ((inSchema = inputQueue.getSchema()) == null)
		{
			System.out.println("input queue schema is NULL");
			return false;
		}
		
		inLength = inSchema.getLength();

		if (this.getOutputQueue().getSchema() == null) 
		{
			init();
		} 
				
		//int howMany = Integer.parseInt(this.getProperty("NUMBER_OF_TUPLES_TO_DEQUEUE"));
		int howMany  = 1000; //rimma temp "hack"

		/*
		if (howMany > 5000)
		{		  
			howMany *= .1;
		}
		*/
		
		int len = howMany;		
		XATTuple t = null;
		
		double timeExecutionStarted = System.currentTimeMillis();
				
		System.out.println("# of tuples in the queue: " + inputQueue.size());
		XATTuple[] tuples = inputQueue.dequeue(len, this);
		System.out.println("# of tuples dequeued: " + tuples.length);
		
		numProcessedTuples += tuples.length;
		
		//System.out.println("\t# of tuples in the STEM (to be probed) hash: " + STeMStreamData.TupleList.size());	
//		System.out.println("Dequeued: " + tuples.length + " tuples.");
				
		//update property for statistics monitoring
		this.setProperty("NUMBER_OF_TUPLES_TO_DEQUEUE", ""+tuples.length); 		
							
		//RoutingToken rt = null;
		boolean bSentRt  = false;			
		boolean bProbe  = false;
		boolean bWaitForRToken  = false;
		
		//ArrayList<Integer> CopyRusterOpStack = null;
		//ArrayList outTuples = new ArrayList();
		ArrayList outTuples = new ArrayList();
		int NumExpectedTuplesInRuster = 0;
		int NumReceivedTuplesAfterRToken = 0;
				
		int extraTupleCounter = 0;
		XATTuple[] tuplesUntilRToken = null;
		int i = -1;
		int tuplesToProcess = tuples.length;
		boolean bAlreadyDequeuedCutoffRusterTuples = false;
									
		
		//for(int i = 0; i < tuples.length;i++)
		while((i++ < tuplesToProcess) || (NumReceivedTuplesAfterRToken < NumExpectedTuplesInRuster))
		{									
			XATTuple ReceivedTuple;
			
			if (i < tuplesToProcess)
			{
				//System.out.println("Extracting tuple");
				ReceivedTuple = tuples[i];
			}
			else
			{	
				if (isEddyOp)
				{
					break;
				}
				
				if (!bAlreadyDequeuedCutoffRusterTuples)
				{
					int h = 0;
					if (inputQueue.getContents().size() > 0)
					{						
						String currTupleStr = inputQueue.getContents().get(h).getClass().toString();
						while(!currTupleStr.endsWith("RoutingTokenImp"))
						{
							h++;
							if (h == inputQueue.getContents().size())
							{
								break;
							}
							currTupleStr = inputQueue.getContents().get(h).getClass().toString();
						}
					}
										
					//int diffTupleToDequeueCount = NumExpectedTuplesInRuster - (NumReceivedTuplesAfterRToken);
					int actualTupleToDequeueCount = h;
					//System.out.println("diffTupleToDequeueCount: " + diffTupleToDequeueCount);
					//System.out.println("actualTupleToDequeueCount: " + actualTupleToDequeueCount);
					
					if (actualTupleToDequeueCount <= 0)
					{
						return true;
					}
					
					if (inputQueue.getContents().size() > 0)
					{
						if (inputQueue.getContents().get(0).getClass().toString().endsWith("RoutingTokenImp"))
						{
							return true;
						}
					}
					//dequeue the difference in the number of tuples (to make sure we have dequeued a complete ruster)
					//in this queue, we will dequeue tuples until the next r-token
					tuplesUntilRToken = inputQueue.dequeue(actualTupleToDequeueCount, this);
				
					if (tuplesUntilRToken == null)
					{
						return true;
					}
					else if (tuplesUntilRToken.length == 0)
					{ 
						//done (assume queue is empty)
						return true;	
					}										
					
					bAlreadyDequeuedCutoffRusterTuples = true; //already dequeued tuples, so just go and finish their processing
				}
				//System.out.println("extraTupleCounter: "+extraTupleCounter);
				if (extraTupleCounter == tuplesUntilRToken.length)
				{
					return true;
				}
				
				ReceivedTuple = tuplesUntilRToken[extraTupleCounter++];				
			}
						
			EddyTuple et = null, net = null;
			QMeshTuple qmt = null, result = null;
			
			// if this is an eddy op
			// parse the probe tuple from the eddy tuple
			if (isEddyOp)
			{	
			 	et = (EddyTuple)tuples[i];
			 	ReceivedTuple = et.getTuple(ProbeSTREAMId);
			 	bProbe  = true;
			 	
			 	isSelfTuning = false; //if executing eddy --> it cannot be a self-tuning QM
			}
			else // assume QMesh
			{
				if (((i == 0) || (bWaitForRToken)) && (!ReceivedTuple.isRoutingToken())) // is this a routing puncuation?
				{
					System.out.println("First tuple must always be an r-token!");
					bWaitForRToken  = true;
					System.out.println("Waiting for r-token");
					continue;
				}
								
				//-------------------------------------------------------------------
				// ROUTING PUNCTUATION
				//-------------------------------------------------------------------
				if (ReceivedTuple.isRoutingToken()) // is this a routing puncuation?
				{
					//System.out.println("Received punctuation");
					bWaitForRToken = false;
					currentRToken = (RoutingToken)ReceivedTuple;
					NumReceivedTuplesAfterRToken = currentRToken.GetNumTuples();
														
					NumExpectedTuplesInRuster = currentRToken.GetNumTuples();					
					//System.out.println("R-token: " + NumExpectedTuplesInRuster + " tuples are expected.");
					
					//System.out.println("Initialized NumExpectedTuplesInRuster: " + NumExpectedTuplesInRuster);
					NumReceivedTuplesAfterRToken = 0;
					
					//System.out.println("r-token: " + i);										
					
					//there are some "leftover" result tuples that are associated with the last r-token
					//so let's push them forward (send them to their operator)
					if(outTuples.size() > 1)
					{
						RoutingToken rtOut = (RoutingToken)(outTuples.get(0));
						
						//determine which queue to send tuples to 
						XATQueue outQ = null;
							
						if (rtOut.getOpStack().size() == 0)
						{
							outQ = OpLkUp.getOutputQueue();
						}
						else
						{
							int nextOp = rtOut.getTopOperator();							
							outQ = OpLkUp.getOperatorQueue(nextOp);
						}						
						rtOut.SetNumTuples(outTuples.size()-1);
						
						//enqueue only if there are tuples following r-token
						if (outTuples.size() > 1)
						{
							//System.out.println("route left: " + ((RoutingToken)(outTuples.get(0))).getStackString());
							outQ.enqueue(outTuples, this);
						}
						//System.out.println("Enqueued leftver " + outTuples.size() + "tuples");
						
						//increment output counter
						totalNumberTupledEnq += outTuples.size();
						//clear array of tuples and last r-token
						outTuples.clear();
					}														
					//check that the TOP operator has exactly the same id as the current operator
					
					if (currentRToken.getOpStack().size() == 0)
					{
						System.out.println("want to catch this");
					}
					
					int topOp = currentRToken.getTopOperator();
					
					//System.out.println("Current r-token: " + currentRToken.getStackString());
					//System.out.println("Current top operator in r-token: " + topOp);
					//System.out.println("Current operator id: " + this.getOperatorID());
					
					if (topOp != this.getOperatorID())
					{
						System.out.println("ROUTING MISMATH!!! topOperator in r-token is: " + topOp +"this operator is: " + this.getOperatorID());
					}
					
					//pop-off current operator from the routing stack
					currentRToken.popTopOperator();	
										
					//add r-token to the out tuples
					outTuples.clear();
					outTuples.add(currentRToken);	
					
					//check
					if (!outTuples.get(0).getClass().toString().endsWith("RoutingTokenImp"))
					{
						System.out.println("Something is not right! The first entry must be the r-token!");
					}					
					
					//System.out.println("Done processing r-token");
					bProbe = false;				
				}
				else
				{	
					//System.out.println("processin tuple");
					NumReceivedTuplesAfterRToken++;
					//System.out.println("Tuple: " + NumReceivedTuplesAfterRToken + " out of " + NumExpectedTuplesInRuster);
					//-------------------------------------------------------------------
					// QUERY MESH TUPLE
					//-------------------------------------------------------------------
					qmt = (QMeshTuple)ReceivedTuple;
					
					//System.out.println("Extracting tuple content by probe stream id");
					ReceivedTuple = qmt.getTuple(ProbeSTREAMId);
					
					bProbe  = true;
					if (ReceivedTuple == null)
					{
						System.out.println("want to catch this");
						bProbe = false;
					}
				}			
			}//end else query mesh (tuple or r-token)
			
			//----------------------------------------------------------------------------
			// GENERAL PROBE - FOR BOTH EDDY AND QUERY MESH
			//----------------------------------------------------------------------------
			// search the data for matches
			if (bProbe && (ReceivedTuple != null))
			{		
				//System.out.println("probing");
				ArrayList <XATTuple> results = STeMStreamData.Probe(ReceivedTuple,
																	STeMDataIndexArrayPos,
																	NumTupleIndex);				
				if (results == null)
				{				  
					ReceivedTuple = null;
				}
				else
				{
					//System.out.println("Results count: " + results.size());
					lastRusterTuplesOut += results.size();
					
					//----------------------------------------------------------------------
					// EDDY OPERATOR ANSWER
					//----------------------------------------------------------------------					
					if (isEddyOp)
					{
						//send Tup Found to eddy for
						// further processing
					
						// ready and done bits from tuple set to probe
						// with bits already set for this probe op
					
						for(int h = 0;h < this.EddyGetOpBitsPos().size();h++)
						{
							et.ClearReadyBit(this.EddyGetOpBitsPos().get(h));
							et.SetDoneBit(this.EddyGetOpBitsPos().get(h));
						}
				
						BitSet dbs = (BitSet)et.getDoneBitSet().getBits().clone();
						BitSet rbs = (BitSet)et.getReadyBitSet().getBits().clone();
								
						// get inverse of done bits
						// ie this represents ops not done
						BitSet InvDone = (BitSet)dbs.clone();
						InvDone.flip(1, dbs.size());			
					
						// and Inv of done bits with the rbs for the stream in the
						// solution
						// ie this represents all not done and ready
						BitSet rbsSolution = (BitSet)PreFilledPerStream.get(DataSTREAMId).clone();			
						rbsSolution.and(InvDone);
					
						// or the rbs from the solution with the normal rbs
						// ie this represents all not done by the added STREAM
						// and all that were not completed before this operator was called
						rbs.or(rbsSolution);
					
						for(int k = 0; k < results.size(); k++)
						{
							// create new eedy tuple for each intermediate answer
							net  = new EddyTupleImp(et.getNumTuplesInAnswer(), et.getSize());
							net.setAnswer(et.getAnswer());
							net.addTuple(DataSTREAMId, results.get(k));
							net.setDoneBitSet(dbs);
							net.setReadyBitSet(rbs);
							
							//increment counter
							totalNumberTupledEnq++;
							
							this.getOutputQueue().enqueue(net, this);
							net = null;
						}//end iterating through results
						et = null;
						dbs = null;
						rbs = null;
						InvDone = null;					
					}					
					else
					{						
						//----------------------------------------------------------------------
						// QUERY MESH ANSWER
						//----------------------------------------------------------------------
						//QMesh answer - check
						if (outTuples.size() > 1)
						{
							if (!outTuples.get(0).getClass().toString().endsWith("RoutingTokenImp"))
							{
								System.out.println("Something is not right! The first entry must be the r-token!");
							}
						}
				
						//System.out.println("Processing: " + results.size() + " results");
						for(int k = 0; k < results.size(); k++)
						{	
							// create new qmesh tuple for each intermediate answer
							result  = new QMeshTupleImp(qmt.getNumTuplesInAnswer());
							result.setAnswer(qmt.getAnswer());
							result.addTuple(DataSTREAMId, results.get(k));
							
							outTuples.add(result);
						}
						
						//System.out.println("# of outTuples " + outTuples.size());														
						if ((outTuples.size() - 1) >= rusterSendOff) //expect 1st tuple in the outTuples to be an r-token
						{															 							
							currentRToken.SetNumTuples(outTuples.size()-1);							
							//determine which queue to send tuples to 
							XATQueue outQ = null;
															
							if (currentRToken.OpStackSize() > 0)
							{
								int nextOp = currentRToken.getTopOperator();
								outQ = OpLkUp.getOperatorQueue(nextOp);
							}
							else 
							{
								outQ = OpLkUp.getOutputQueue();
							}
													
							//clear all tuple BUT the very first one which is the r-token
							RoutingToken currentRt = (RoutingToken)(outTuples.get(0));
							RoutingToken keepRt = currentRt.CreateClone();
														
							currentRt.SetNumTuples(outTuples.size() - 1);
							
							//System.out.println("Sending ruster with " + currentRt.GetNumTuples() + " tuples");
							
							if (!outTuples.get(0).getClass().toString().endsWith("RoutingTokenImp"))
							{
								System.out.println("Something is not right! The first entry must be the r-token!");
							}
							
							if (outTuples.size() > 1)
							{
								//System.out.println("route left: " + ((RoutingToken)(outTuples.get(0))).getStackString());
								outQ.enqueue(outTuples, this);
							}
							//System.out.println("Enqueued " + outTuples.size() + " tuples");
							
							totalNumberTupledEnq += outTuples.size();
							
							outTuples.clear();	
							
							//add r-token to as the first entry in the tupleOut array
							outTuples.add(keepRt);
							keepRt.SetNumTuples(0);
														
							//System.out.println("sending to op[" +nextOp+"] ruster, where # of ops in r-token: " + currentRToken.getOpStack().size()+", ["+currentRToken.getStackString()+"]" + ", # of tuples in ruster: " + currentRToken.GetNumTuples());							
						}
						//System.out.println("Done processing tuple");																															
					}//end else query mesh answer					
				}//end result is not empty
								
			}//end if probe
			
			//---------------------------------------------------------------
			// STQM EXECUTION STATISTICS
			//---------------------------------------------------------------
			//  Record statistics for the LAST ruster
			if (isSelfTuning)
			{	
				boolean bUpdateStatistics = false;
				if (i == (tuplesToProcess - 1))
				{
					bUpdateStatistics = true;
				}
				else if (i < (tuplesToProcess - 1))
				{
					XATTuple nextTuple = (XATTuple)tuples[i+1];
					if (nextTuple == null)
					{
						bUpdateStatistics = true;
					}
					else if (nextTuple.isRoutingToken())
					{
						bUpdateStatistics = true;
					}
				}
				//if the next tuple is an r-token, update the statistics for the current (just processed) ruster
				if ((bUpdateStatistics) && (currentRToken != null))
				{					
					double timeExecutionStatsUpdateStarted = System.currentTimeMillis();
					lastRToken = currentRToken;					
					
					System.out.println("STEM STREAM ID " + lastRToken.GetRTokenStreamId());
					
					System.out.println("stupid thing is " + DistributedExecutionController.hashSTQMRouteSignaturesGlobalStorage.isEmpty());
					
					//store statistics by route signature - first retrieve all routes for the stream 
					Hashtable hashStreamRoutesSignatures = 
					(Hashtable)DistributedExecutionController.hashSTQMRouteSignaturesGlobalStorage.get(lastRToken.GetRTokenStreamId());

					//check if there exists only a single route for the current stream
					RouteSignatureStorage rss = null;
					if (hashStreamRoutesSignatures.size() == 1)
					{						 
						Enumeration e = hashStreamRoutesSignatures.keys();
						String key = e.nextElement().toString(); //key - stream id (string)						
						rss = (RouteSignatureStorage)hashStreamRoutesSignatures.get(key);
					}
					else
					{
						//now retrieve the route based on the route id stored in the current r-token
						rss = (RouteSignatureStorage)hashStreamRoutesSignatures.get(lastRToken.GetRouteId());
					}
					
					if (rss == null)
					{
						//the rss might equal to null if the QM has changed (and the current route does NOT exist in the current QM)
						//in this case, DO NOT UPDATE STATS
						
						//do nothing
					}
					else
					{
						//now store the statistics in the global statistics storage based on the route signature (i.e., the string
						//representing the sequence of operators in the route - this is our identifier in the route statistics table)
						String globalStatsKey = lastRToken.GetRTokenStreamId() + "->" + rss.GetRouteSignature();
						
						lastRusterTuplesIn = NumReceivedTuplesAfterRToken;
						
						//globalStatsKey = stream id + operator list
						//System.out.println("UPDATING EXECUTION STATISTICS: " + globalStatsKey +", " + lastRusterTuplesIn+", " + lastRusterTuplesOut+", " + System.currentTimeMillis());
						
						if (hashNewSTQMRusterExploitationStatisticsLocalStorage.get(globalStatsKey) != null)
						{
							//update statistics for existing route
							OperatorIndividualSample ois = (OperatorIndividualSample)hashNewSTQMRusterExploitationStatisticsLocalStorage.get(globalStatsKey);
							ois.SetTuplesIn(lastRusterTuplesIn);
							ois.SetTuplesOut(lastRusterTuplesOut);
							ois.SetLastUpdateTime(System.currentTimeMillis());
						}
						else
						{
							//store statistics for brand new route
							OperatorIndividualSample ois = new OperatorIndividualSample(lastRusterTuplesIn, lastRusterTuplesOut, System.currentTimeMillis());
							hashNewSTQMRusterExploitationStatisticsLocalStorage.put(globalStatsKey, ois);
						}										
						
						double timeExecutionStatsUpdateEnded = System.currentTimeMillis();
						double timeExecutionStatsUpdateDuration = timeExecutionStatsUpdateEnded - timeExecutionStarted;
						double timePerTuple = timeExecutionStatsUpdateDuration / lastRusterTuplesIn;
						String outString = "EXECUTION STATS UPDATE OVERHEAD: # tuples: " +lastRusterTuplesIn + ", duration: " + timeExecutionStatsUpdateDuration + "per tuple overhead:\t " + timePerTuple;
					
						//WriteToFile(outFileExeStatsOverhead, outString);
						
						lastRusterTuplesIn = 0;
						lastRusterTuplesOut = 0;
					}
				}//end if (((XATTuple)tuples[i+1]).isRoutingToken())
			}//end if curr operator is self-tuning (i.e., it is a part of STQM)
		}//end while (iterating through all dequeued tuples)		
							
		//----------------------------------------------------------------------------
		// post-processing: send of ruster where number of tuples are < NumSendoff and
		// there are no tuples in the input queue
		//----------------------------------------------------------------------------		
		if(outTuples.size() > 1)
		{
			RoutingToken rtOut = (RoutingToken)(outTuples.get(0));
			
			//determine which queue to send tuples to 
			XATQueue outQ = null;
				
			if (rtOut.getOpStack().size() == 0)
			{
				outQ = OpLkUp.getOutputQueue();
			}
			else
			{
				int nextOp = rtOut.getTopOperator();							
				outQ = OpLkUp.getOperatorQueue(nextOp);
			}						
			rtOut.SetNumTuples(outTuples.size()-1);
			
			//enqueue only if there are tuples following r-token
			if (outTuples.size() > 1)
			{
				//System.out.println("route left: " + ((RoutingToken)(outTuples.get(0))).getStackString());
				outQ.enqueue(outTuples, this);
			}
			
			//increment output counter
			totalNumberTupledEnq += outTuples.size();
			//clear array of tuples and last r-token
			outTuples.clear();
		}		
		
		
		System.out.println("# of tuples enqueued: " + totalNumberTupledEnq);
					
		double timeExecutionEnded = System.currentTimeMillis();
		double timeExecutionDuration = timeExecutionEnded - timeExecutionStarted;
		double timePerTuple = timeExecutionDuration / tuples.length;
		String outString = "JOIN EXECUTION OVERHEAD: \t# tuples: \t" + tuples.length + "\t duration: \t" + timeExecutionDuration + "\t per tuple overhead:\t " + timePerTuple;
		WriteToFile(outFileJoinOverhead, outString);
				
		System.gc();
		return true;
	}
	
	//-------------------------------------------------------------------------------------

	public boolean cleanAccumulatedTuples(long maxtimestamp) 
	{
		return false;
	}
	
	//-------------------------------------------------------------------------------------

	public boolean cleanAllAccumulatedTuples() 
	{
		return false;
	}
	
	//-------------------------------------------------------------------------------------

	public boolean matchStates(XATStreamOperator op) 
	{
		return false;
	}

	//-------------------------------------------------------------------------------------
	
	public boolean hasUnmatchedState() 
	{
		return false;
	}

	public boolean recomputeUnmatchedState() 
	{
		return false;
	}

	//-------------------------------------------------------------------------------------
	
	public int setStateID(int startID) 
	{
		return 0;
	}

	//-------------------------------------------------------------------------------------
	
	public String toToolTipString() 
	{
		return "STeMJoinProbeOperator";
	}

	//-------------------------------------------------------------------------------------
	
	public boolean finalizeOperation() 
	{
		return false;
	}

	//-------------------------------------------------------------------------------------
	
	public void setSTeMStreamData(STeMData sd)
	{
		STeMStreamData = sd;
	}
	
	//-------------------------------------------------------------------------------------
	
	public void SetIsEddyOp(boolean b)
	{
		isEddyOp = b;
	}

	//-------------------------------------------------------------------------------------
	
	//STQM
	public void SetIsSelfTuning(boolean b)
	{
		isSelfTuning = b;
	}
	
	//-------------------------------------------------------------------------------------
	
	//STQM
	public boolean GetIsSelfTuning()
	{
		return isSelfTuning;
	}
	
	//-------------------------------------------------------------------------------------
	//Info about probe
	public void SetProbeSTREAMId(int i)
	{
		ProbeSTREAMId = i;
	}

	//-------------------------------------------------------------------------------------
	
	public int getProbeSTREAMId()
	{
		return ProbeSTREAMId;
	}

	//-------------------------------------------------------------------------------------
	
	public void SetNumTupleIndex(String val) 
	{
		NumTupleIndex = new ArrayList();
		StringTokenizer st = new StringTokenizer(val);
	    while (st.hasMoreElements()) 
	    {
	    	NumTupleIndex.add(Integer.parseInt(st.nextElement().toString()));
	    }	
	}

	//-------------------------------------------------------------------------------------
	
	public ArrayList<Integer> getNumTupleIndex()
	{
		return NumTupleIndex;
	}

	//-------------------------------------------------------------------------------------
	//Info about stored stem data
	
	public void SetDataSTREAMId(int i)
	{
		DataSTREAMId = i;
	}
	
	//-------------------------------------------------------------------------------------

	public int getDataSTREAMId()
	{
		return DataSTREAMId;
	}
	
	//-------------------------------------------------------------------------------------

	public void SetSTeMDataIndexArrayPos(int i)
	{
		STeMDataIndexArrayPos = i;
	}

	//-------------------------------------------------------------------------------------
	
	public int getSTeMDataIndexArrayPos()
	{
		return STeMDataIndexArrayPos;
	}
	
	//-------------------------------------------------------------------------------------
	
	public void SetNumSTeMIndex(String val) 
	{
		NumSTeMIndex = new ArrayList();
		StringTokenizer st = new StringTokenizer(val);
	    while (st.hasMoreElements()) 
	    {	    
	    	NumSTeMIndex.add(Integer.parseInt(st.nextElement().toString()));
	    }	
	}

	//-------------------------------------------------------------------------------------
	
	public ArrayList<Integer> getNumSTeMIndex()
	{
		return NumSTeMIndex;
	}

	//-------------------------------------------------------------------------------------
	
	public void setPreFilledPerStream(ArrayList<BitSet> p)
	{
		PreFilledPerStream = p;
	}	
	
	//-------------------------------------------------------------------------------------
	
	// table for look up operators used by QMesh
	public void setOperatorLookUp(OperatorLookUp z)
	{
		OpLkUp = z;
	}
	
	//-------------------------------------------------------------------------------------
	
	public void setQMeshOperatorID(int id) 
	{
		QMeshOperatorID = id;
	}
	
	//-------------------------------------------------------------------------------------
	
	public int getQMeshOperatorID() 
	{
		return QMeshOperatorID;
	}
	
	//-------------------------------------------------------------------------------------
	
	public void setSendOff(int so)
	{
		rusterSendOff = so;
	}	
	
	//-------------------------------------------------------------------------------------
	
	//used for experiments (otherwise comment it out)
	public void WriteToFile(String fileLocation, String stringToWrite)
	{
		try
		{
			//Create file 
			FileWriter fstream = new FileWriter(fileLocation, true);
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(stringToWrite);
			out.newLine();			
			//Close the output stream
			out.close();			    
		}
		catch (Exception e)
		{
			//Catch exception if any
			System.err.println("Error: " + e.getMessage());
		}
	}
	
	//-------------------------------------------------------------------------------------------
	
	//STQM
	public Hashtable GetNewOperatorExploitationStatisticsPerRuster()
	{
		return hashNewSTQMRusterExploitationStatisticsLocalStorage;
	}
	
	//-------------------------------------------------------------------------------------------
	
	//STQM
	public Hashtable GetOldOperatorExploitationStatisticsPerRuster()
	{
		return hashOldSTQMRusterExploitationStatisticsLocalStorage;
	}
	
	//-------------------------------------------------------------------------------------------
	
	//STQM
	public Hashtable GetNewOperatorExplorationStatisticsPerRuster()
	{
		return hashNewSTQMRusterExplorationStatisticsLocalStorage;
	}
	
	//-------------------------------------------------------------------------------------------
	
	//STQM
	public Hashtable GetOldOperatorExplorationStatisticsPerRuster()
	{
		return hashOldSTQMRusterExplorationStatisticsLocalStorage;
	}
	
	//-------------------------------------------------------------------------------------
	
	//STQM
	public void MakeNewStatisticsToBeOldStatistics()
	{
		//exploitation statistics
		hashOldSTQMRusterExploitationStatisticsLocalStorage = (Hashtable)hashNewSTQMRusterExploitationStatisticsLocalStorage.clone();
		hashNewSTQMRusterExploitationStatisticsLocalStorage.clear();
		
		//exploration statistics
		hashOldSTQMRusterExplorationStatisticsLocalStorage = (Hashtable)hashNewSTQMRusterExplorationStatisticsLocalStorage.clone();
		hashNewSTQMRusterExplorationStatisticsLocalStorage.clear();
	}
	
	//-------------------------------------------------------------------------------------------
}
