/*
 * Constraints.java
 *
 */
package edu.wpi.cs.dsrg.xmldb.xat.common.querymesh.optimization.JSetL;


import java.util.*;

/**
 * The <code>Constraints</code> class.
 *
 * @author  Elio Panegai
 * @author  Elisabetta Poleo
 * @version 1.3.0
 */
public class RewritingConstraintsRules implements java.io.Serializable {
    
    
    /* Member Vars */
    SolverClass Solver;
    
    
    /* Constructor */
    
    /**
     * Constructs an object.
     * @param s Reference to the solver.
     */
    public RewritingConstraintsRules(SolverClass s) {
        Solver = s;
    }
    
    
    protected void risElem(int k) throws Failure {
        Constraint s = (Constraint)Solver.get(k);
        if (Solver.ctx == 0) {   // Se ctx = 0 vengono risolti tutti i vincoli che non sono ancora
            if (!s.bool) {       // in forma risolta cio t.c. la loro variabile bool ha valore false.
                if (s.cons == Enviroment.eqCode) eq(s);                 // equality
                else if (s.cons == Enviroment.neqCode) neq(s);          // inequality
                else if (s.cons == Enviroment.neq_unionCode) neq_union(s);  // union's case (6) and (7)
                else if (s.cons == Enviroment.inCode) in(s);            // in
                else if (s.cons == Enviroment.ninCode) nin(s);          // not in
                else if (s.cons == Enviroment.selectCode) select(s);    // not deterministic choice
                else if (s.cons == Enviroment.disjCode) disj(s);        // disjunction
                else if (s.cons == Enviroment.ndisjCode) ndisj(s);      // not disjunction
                else if (s.cons == Enviroment.unionCode) union(s);      // union
                else if (s.cons == Enviroment.nunionCode) nunion(s);    // not union
                else if (s.cons == Enviroment.geCode) ge(s);            // greater or equal
                else if (s.cons == Enviroment.gtCode) gt(s);            // greater than
                else if (s.cons == Enviroment.leCode) le(s);            // less or equal
                else if (s.cons == Enviroment.ltCode) lt(s);            // less than
                else if (s.cons == Enviroment.sumCode) sum(s);          // sum
                else if (s.cons == Enviroment.subCode) sub(s);          // sub
                else if (s.cons == Enviroment.mulCode) mul(s);          // mul
                else if (s.cons == Enviroment.divCode) div(s);          // div
                else if (s.cons == Enviroment.modCode) mod(s);          // module
                else if (s.cons == Enviroment.forallCode) forall(s);    // forall
                else if (s.cons == Enviroment.intervCode) interv(s);    // interv
                else if (s.cons == Enviroment.cond_ifCode) cond_if(s);  // cond_if
                else if (s.cons == Enviroment.debugCode) debug(s);      // debug
                else {
                    s.bool = true;
                    /* If the constraint that must be resolved is not nobody of predefined constraints, the control goes
                     * to the user_code method of the class NewConstraints/LibConstraints because in this case
                     * we are treating a user defined constraint.
                     */
                    //try { Solver.N.user_code(s.cons, s); }
                    try {
                        if (Solver.N != null) Solver.N.user_code(s);
                        else throw new NotDefinedConstraint();
                    }
                    catch (NotDefinedConstraint e) { Solver.LC.user_code(s.cons, s); }
                }
            }
        }
        else {
            /* Se ctx (della classe Solver) non  0, vengono risolti solo i vincoli
               non ancora in forma risolta che hanno il campo ctx uquale alla
               variabile globale ctx della classe Solver. */
            if(!s.bool && s.ctx == Solver.ctx) {
                if (s.cons == Enviroment.eqCode) eq(s);         // equality
                else if (s.cons == Enviroment.neqCode) neq(s);  // inequality
                else if (s.cons == Enviroment.neq_unionCode) neq_union(s);  // union's case (6) and (7)
                else if (s.cons == Enviroment.inCode) in(s);  // in
                else if (s.cons == Enviroment.ninCode) nin(s);  // not in
                else if (s.cons == Enviroment.selectCode) select(s);  // not deterministic choice
                else if (s.cons == Enviroment.disjCode) disj(s);  // disjunction
                else if (s.cons == Enviroment.ndisjCode) ndisj(s);  // not disjunction
                else if (s.cons == Enviroment.unionCode) union(s);  // union
                else if (s.cons == Enviroment.nunionCode) nunion(s);  // not union
                else if (s.cons == Enviroment.geCode) ge(s);  // greater or equal
                else if (s.cons == Enviroment.gtCode) gt(s);  // greater than
                else if (s.cons == Enviroment.leCode) le(s);  // less or equal
                else if (s.cons == Enviroment.ltCode) lt(s);  // less than
                else if (s.cons == Enviroment.sumCode) sum(s);  // sum
                else if (s.cons == Enviroment.subCode) sub(s);  // sub
                else if (s.cons == Enviroment.mulCode) mul(s);  // mul
                else if (s.cons == Enviroment.divCode) div(s);  // div
                else if (s.cons == Enviroment.modCode) mod(s);  // module
                else if (s.cons == Enviroment.forallCode) forall(s);  // forall
                else if (s.cons == Enviroment.intervCode) interv(s);  // interv
                else if (s.cons == Enviroment.cond_ifCode) cond_if(s);  // cond_if
                else if (s.cons == Enviroment.debugCode) debug(s);  // debug
                else {
                    s.bool = true;
                    /* If the constraint that must be resolved is not nobody of predefined constraints, the control goes
                     * to the user_code method of the class NewConstraints/LibConstraints because in this case
                     * we are treating a user defined constraint.
                     */
                    try {
                        if (Solver.N != null) Solver.N.user_code(s);
                        else throw new NotDefinedConstraint();
                    }
                    catch (NotDefinedConstraint e) { Solver.LC.user_code(s.cons, s); }
                }
            }
        }
        return;
    }
    
    
    // costruttore del debug
    public static Constraint debug(String txt, Object toPrint) {
        return new Constraint(Enviroment.debugCode, txt, toPrint);
    }
    
    
    // cond_if con else
    public static Constraint cond_if(Constraint condizione, ConstraintsConjunction allora, ConstraintsConjunction altrimenti) {
        return new Constraint(Enviroment.cond_ifCode, condizione, allora, altrimenti);
    }
    
    
    public static Constraint cond_if(Constraint condizione, Constraint allora, ConstraintsConjunction altrimenti) {
        return new Constraint(Enviroment.cond_ifCode, condizione, allora, altrimenti);
    }
    
    
    public static Constraint cond_if(Constraint condizione, ConstraintsConjunction allora, Constraint altrimenti) {
        return new Constraint(Enviroment.cond_ifCode, condizione, allora, altrimenti);
    }
    
    
    public static Constraint cond_if(Constraint condizione, Constraint allora, Constraint altrimenti) {
        return new Constraint(Enviroment.cond_ifCode, condizione, allora, altrimenti);
    }
    
    
    // cond_if senza else
    public static Constraint cond_if(Constraint condizione, ConstraintsConjunction allora) {
        Constraint s = new Constraint(Enviroment.cond_ifCode, condizione, allora);
        return s;
    }
    
    
    public static Constraint cond_if(Constraint condizione, Constraint allora) {
        Constraint s = new Constraint(Enviroment.cond_ifCode, condizione, allora);
        return s;
    }
    
    /* Doesn't works for MultiSet*/
    protected Constraint getUnnestedConstraint(Constraint nested) throws Failure {
        Constraint notNested       = nested.cloneConstraint();
        Object firstArgConstraint  = (notNested.getMembers()).get(0);
        Object secondArgConstraint = (notNested.getMembers()).get(1);
        
        if (firstArgConstraint instanceof MultiSet ||
            secondArgConstraint instanceof MultiSet)  throw new Failure();
        
        
        if ((firstArgConstraint instanceof Lvar && ((Lvar)firstArgConstraint).equ  == null  &&
        secondArgConstraint instanceof Lvar && ((Lvar)secondArgConstraint).equ == null)
        ||
        (firstArgConstraint instanceof Lst  && ((Lst)firstArgConstraint).equ   == null  &&
        secondArgConstraint instanceof Lst  && ((Lst)secondArgConstraint).equ  == null)
        ||
        ((firstArgConstraint instanceof Set && ((Set)firstArgConstraint).equ   == null) &&
        (secondArgConstraint instanceof Set && ((Set)secondArgConstraint).equ  == null))
        ||
        (firstArgConstraint instanceof Lvar && ((Lvar)firstArgConstraint).equ  == null  &&
        secondArgConstraint instanceof Lst  && ((Lst)secondArgConstraint).equ  == null)
        ||
        (firstArgConstraint instanceof Lst  && ((Lst)firstArgConstraint).equ   == null  &&
        secondArgConstraint instanceof Lvar && ((Lvar)secondArgConstraint).equ == null)
        ||
        (firstArgConstraint instanceof Lvar && ((Lvar)firstArgConstraint).equ  == null  &&
        secondArgConstraint instanceof Set  && ((Set)secondArgConstraint).equ  == null)
        ||
        (firstArgConstraint instanceof Set  && ((Set)firstArgConstraint).equ   == null  &&
        secondArgConstraint instanceof Lvar && ((Lvar)secondArgConstraint).equ == null)
        ||
        (firstArgConstraint instanceof Lst  && ((Lst)firstArgConstraint).equ   == null  &&
        secondArgConstraint instanceof Set  && ((Set)secondArgConstraint).equ  == null)
        ||
        (firstArgConstraint instanceof Set  && ((Set)firstArgConstraint).equ   == null  &&
        secondArgConstraint instanceof Lst  && ((Lst)secondArgConstraint).equ  == null) ){
            // System.out.println("XELIO notNested");
            return notNested;
        }
        else if ((firstArgConstraint instanceof Lvar && ((Lvar)firstArgConstraint).equ == null) ||
        (firstArgConstraint instanceof Lst  && ((Lst)firstArgConstraint).equ  == null) ||
        (firstArgConstraint instanceof Set  && ((Set)firstArgConstraint).equ  == null)) {
            if (secondArgConstraint instanceof Lvar) notNested.arg2 = ((Lvar)secondArgConstraint).equ;
            else if (secondArgConstraint instanceof Lst) notNested.arg2 = ((Lst)secondArgConstraint).equ;
            else if (secondArgConstraint instanceof Set) notNested.arg2 = ((Set)secondArgConstraint).equ;
            getUnnestedConstraint(notNested);
        }
        else if ((secondArgConstraint instanceof Lvar && ((Lvar)secondArgConstraint).equ == null) ||
        (secondArgConstraint instanceof Lst  && ((Lst)secondArgConstraint).equ  == null) ||
        (secondArgConstraint instanceof Set  && ((Set)secondArgConstraint).equ  == null)) {
            if (firstArgConstraint instanceof Lvar) notNested.arg2 = ((Lvar)firstArgConstraint).equ;
            else if (firstArgConstraint instanceof Lst) notNested.arg2 = ((Lst)firstArgConstraint).equ;
            else if (firstArgConstraint instanceof Set) notNested.arg2 = ((Set)firstArgConstraint).equ;
            getUnnestedConstraint(notNested);
        }
        else {
            if (firstArgConstraint instanceof Lvar && secondArgConstraint instanceof Lvar) {
                notNested.arg1 = ((Lvar)firstArgConstraint).equ;
                notNested.arg2 = ((Lvar)secondArgConstraint).equ;
            }
            if (firstArgConstraint instanceof Lvar && secondArgConstraint instanceof Lst) {
                notNested.arg1 = ((Lvar)firstArgConstraint).equ;
                notNested.arg2 = ((Lst)secondArgConstraint).equ;
            }
            if (firstArgConstraint instanceof Lst  && secondArgConstraint instanceof Lvar) {
                notNested.arg1 = ((Lst)firstArgConstraint).equ;
                notNested.arg2 = ((Lvar)secondArgConstraint).equ;
            }
            if (firstArgConstraint instanceof Lvar && secondArgConstraint instanceof Set) {
                notNested.arg1 = ((Lvar)firstArgConstraint).equ;
                notNested.arg2 = ((Set)secondArgConstraint).equ;
            }
            if (firstArgConstraint instanceof Set  && secondArgConstraint instanceof Lvar) {
                notNested.arg1 = ((Set)firstArgConstraint).equ;
                notNested.arg2 = ((Lvar)secondArgConstraint).equ;
            }
            
            if (firstArgConstraint instanceof Set  && secondArgConstraint instanceof Set) {
                notNested.arg1 = ((Set)firstArgConstraint).equ;
                notNested.arg2 = ((Set)secondArgConstraint).equ;
            }
            if (firstArgConstraint instanceof Set  && secondArgConstraint instanceof Lst) {
                notNested.arg1 = ((Set)firstArgConstraint).equ;
                notNested.arg2 = ((Lst)secondArgConstraint).equ;
            }
            if (firstArgConstraint instanceof Lst  && secondArgConstraint instanceof Set) {
                notNested.arg1 = ((Lst)firstArgConstraint).equ;
                notNested.arg2 = ((Set)secondArgConstraint).equ;
            }
            if (firstArgConstraint instanceof Lst  && secondArgConstraint instanceof Lst) {
                notNested.arg1 = ((Lst)firstArgConstraint).equ;
                notNested.arg2 = ((Lst)secondArgConstraint).equ;
            }
            getUnnestedConstraint(notNested);
        }
        return notNested;
    }
    
    
    protected void eq(Constraint s) throws Failure {
        if(s.arg3 instanceof Set)
            eqSet((Set)s.arg1, (Set)s.arg2, (Set)s.arg3, s);
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar)
            eqLvar((Lvar)s.arg1, (Lvar)s.arg2, s);                     //  X.eq(Y)  entrambe Lvar
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Lst)
            eqLvarLst((Lvar)s.arg1, (Lst)s.arg2, s);                   //  X.eq([t1, .., tn])
        else if(s.arg1 instanceof Lst && s.arg2 instanceof Lvar)
            eqLvarLst((Lvar)s.arg2, (Lst)s.arg1, s);                   //  [t1, .., tn].eq(X)
        else if(s.arg1 instanceof Lst && s.arg2 instanceof Lst)
            eqLst((Lst)s.arg1, (Lst)s.arg2, s);                        //  [t1, .., tn].eq([s1, .., sn])
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Set)
            eqLvarSet((Lvar)s.arg1, (Set)s.arg2, s);                   //  X.eq( {...})
        else if(s.arg1 instanceof Set && s.arg2 instanceof Lvar)
            eqLvarSet((Lvar)s.arg2, (Set)s.arg1, s);                   //  {...}.eq(X)
        else if(s.arg1 instanceof Set && s.arg2 instanceof Set)
            eqSet((Set)s.arg1, (Set)s.arg2, s);                        //  {...}.eq( {...})
        //   else if(s.arg1 instanceof Set && s.arg2 instanceof Int)
        //        eqSetInt((Set)s.arg1, (Int)s.arg2, s);                   //  {...}.eq(Int(t1, t2))
        //   else if(s.arg1 instanceof Int && s.arg2 instanceof Set)
        //        eqSetInt((Set)s.arg2, (Int)s.arg1, s);                   //  Int(t1, t2)).eq( {...})
        else if(s.arg1 instanceof MultiSet && s.arg2 instanceof MultiSet)
            eqMultiSet((MultiSet)s.arg1, (MultiSet)s.arg2, s);         //  {...}.eq( {...})
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof MultiSet)
            eqLvarMultiSet((Lvar)s.arg1, (MultiSet)s.arg2, s);         //  X.eq( {...})
        else if( s.arg1 instanceof MultiSet && s.arg2 instanceof Lvar)
            eqLvarMultiSet((Lvar)s.arg2, (MultiSet)s.arg1, s);         //  {...}.eq( X)
        else if(s.arg1 instanceof Lvar)
            eqLvarObj((Lvar)s.arg1, s.arg2, s);                        //  X.eq(t)
        else if(s.arg2 instanceof Lvar)
            eqLvarObj((Lvar)s.arg2, s.arg1, s);                        //  t.eq(X)
        else if(s.arg1 instanceof Lst)
            eqLstObj((Lst)s.arg1, s.arg2, s);                          //  [t1, .., tn].eq(t)
        else if(s.arg2 instanceof Lst)
            eqLstObj((Lst)s.arg2, s.arg1, s);                          //  t.eq([t1, .., tn])
        else if(s.arg1 instanceof Set)
            eqSetObj((Set)s.arg1, s.arg2, s);                          //  {...}.eq(t)
        else if(s.arg2 instanceof Set)
            eqSetObj((Set)s.arg2, s.arg1, s);                          //  t.eq( {...})
        else if(s.arg1 instanceof MultiSet)
            eqMultiSetObj((MultiSet)s.arg1, s.arg2, s);                //  {...}.eq(t)
        else if(s.arg2 instanceof MultiSet)
            eqMultiSetObj((MultiSet)s.arg2, s.arg1, s);                //  t.eq( {...})
        else eqObj(s.arg1, s.arg2, s);                                  //  t.eq(s)
        return;
    }
    
    
    protected void neq(Constraint s) throws Failure {
        if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar)
            neqLvar((Lvar)s.arg1, (Lvar)s.arg2, s);                    //  X.neq(Y)  entrambe Lvar
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Lst)
            neqLvarLst((Lvar)s.arg1, (Lst)s.arg2, s);                  //  X.neq([t1, .., tn])
        else if(s.arg1 instanceof Lst && s.arg2 instanceof Lvar)
            neqLvarLst((Lvar)s.arg2, (Lst)s.arg1, s);                  //  [t1, .., tn].neq(X)
        else if(s.arg1 instanceof Lst && s.arg2 instanceof Lst)
            neqLst((Lst)s.arg1, (Lst)s.arg2, s);                       //  [t1, .., tn].neq([s1, .., sn])
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Set)
            neqLvarSet((Lvar)s.arg1, (Set)s.arg2, s);                  //  X.neq( {...})
        else if(s.arg1 instanceof Set && s.arg2 instanceof Lvar)
            neqLvarSet((Lvar)s.arg2, (Set)s.arg1, s);                  //  {...}.neq(X)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof MultiSet)
            neqLvarMultiSet((Lvar)s.arg1, (MultiSet)s.arg2, s);        //  X.neq( {...})
        else if(s.arg1 instanceof MultiSet && s.arg2 instanceof Lvar)
            neqLvarMultiSet((Lvar)s.arg2, (MultiSet)s.arg1, s);        //  {...}.neq(X)
        else if(s.arg1 instanceof Set && s.arg2 instanceof Set)
            neqSet((Set)s.arg1, (Set)s.arg2, s);                       //  {...}.neq( {...})
        //   else if(s.arg1 instanceof Set && s.arg2 instanceof Int)
        //        neqSetInt((Set)s.arg1, (Int)s.arg2, s);                  //  {...}.neq(Int(t1, t2))
        //   else if(s.arg1 instanceof Int && s.arg2 instanceof Set)
        //        neqSetInt((Set)s.arg2, (Int)s.arg1, s);                  //  Int(t1, t2)).neq( {...})
        else if(s.arg1 instanceof MultiSet && s.arg2 instanceof MultiSet)
            neqMultiSet((MultiSet)s.arg1, (MultiSet)s.arg2, s);        //  {...}.neq( {...})
        else if(s.arg1 instanceof Lvar)
            neqLvarObj((Lvar)s.arg1, s.arg2, s);                       //  X.neq(t)
        else if(s.arg2 instanceof Lvar)
            neqLvarObj((Lvar)s.arg2, s.arg1, s);                       //  t.neq(X)
        else if(s.arg1 instanceof Lst)
            neqLstObj((Lst)s.arg1, s.arg2, s);                         //  [t1, .., tn].neq(t)
        else if(s.arg2 instanceof Lst)
            neqLstObj((Lst)s.arg2, s.arg1, s);                         //  t.neq([t1, .., tn])
        else if(s.arg1 instanceof MultiSet)
            neqMultiSetObj((MultiSet)s.arg1, s.arg2, s);               //  {...}.neq(t)
        else if(s.arg2 instanceof MultiSet)
            neqMultiSetObj((MultiSet)s.arg2, s.arg1, s);               //  t.neq( {...})
        else if(s.arg1 instanceof Set)
            neqSetObj((Set)s.arg1, s.arg2, s);                         //  {...}.neq(t)
        else if(s.arg2 instanceof Set)
            neqSetObj((Set)s.arg2, s.arg1, s);                         //  t.neq( {...})
        else neqObj(s.arg1, s.arg2, s);                                //  t.neq(s)
        return;
    }
    
    
    protected void neq_union(Constraint s) throws Failure {
        if(s.arg1 instanceof Set && s.arg2 instanceof Set) neq_union((Set)s.arg1, (Set)s.arg2, s);
        return;
    }
    
    
    protected void in(Constraint s) throws Failure {
        if(s.arg1 instanceof Lvar && s.arg2 instanceof Set)
            inLvarSet((Lvar)s.arg1, (Set)s.arg2, s);                   //  X.in( {...})
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar)
            inLvarLvar((Lvar)s.arg1, (Lvar)s.arg2, s);                 //  X.in(Y)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof MultiSet)
            inLvarMultiSet((Lvar)s.arg1, (MultiSet)s.arg2, s);                 //  X.in(Y)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Int)
            inLvarInt((Lvar)s.arg1, (Int)s.arg2, s);                   //  X.in(Int(t1, t2))
        else if(s.arg2 instanceof Set)
            inObjSet(s.arg1, (Set)s.arg2, s);                          //  t.in( {...})
        else if(s.arg2 instanceof MultiSet)
            inObjMultiSet(s.arg1, (MultiSet)s.arg2, s);                //t.in( {...} )
        else if(s.arg2 instanceof Lvar)
            inObjLvar(s.arg1, (Lvar)s.arg2, s);                        //  t.in(X)
        else if(s.arg2 instanceof Int)
            inObjInt(s.arg1, (Int)s.arg2, s);                          //  t.in(Int(t1, t2))
        return;
    }
    
    
    /* Richiamata solo nel caso (10) della procedura di riscrittura
       dell'unificazione (si veda "Sets and Constraint Logic Programming").*/
    protected void select(Constraint s) throws Failure {
        select((Lvar)s.arg1, (Set)s.arg2, (Set)s.arg3, s);
        return;
    }
    
    
    protected void nin(Constraint s) throws Failure {
        if(s.arg1 instanceof Lvar && s.arg2 instanceof Set)
            ninLvarSet((Lvar)s.arg1, (Set)s.arg2, s);                  //  X.nin( {...})
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar)
            ninLvarLvar((Lvar)s.arg1, (Lvar)s.arg2, s);                //  X.nin(Y)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof MultiSet)
            ninLvarMultiSet((Lvar)s.arg1, (MultiSet)s.arg2, s);        //  X.nin( {...})
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Int)
            ninLvarInt((Lvar)s.arg1, (Int)s.arg2, s);                  //  X.nin(Int(t1, t2))
        else if(s.arg2 instanceof Set)
            ninObjSet(s.arg1, (Set)s.arg2, s);                         //  t.nin( {...})
        else if(s.arg2 instanceof MultiSet)
            ninObjMultiSet(s.arg1, (MultiSet)s.arg2, s);
        else if(s.arg2 instanceof Set)
            ninObjLvar(s.arg1, (Lvar)s.arg2, s);                       //  t.nin(X)
        else if(s.arg2 instanceof Int)
            ninObjInt(s.arg1, (Int)s.arg2, s);                         //  t.nin(Int(t1, t2))
        return;
    }
    
    
    protected void disj(Constraint s) throws Failure {
        if(s.arg1 instanceof Set && s.arg2 instanceof Set)
            disj((Set)s.arg1, (Set)s.arg2, s);                         //  {...}.disj( {...})
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Set)
            disjLvarSet((Lvar)s.arg1, (Set)s.arg2, s);                 //  X.disj( {...})
        else if(s.arg1 instanceof Set && s.arg2 instanceof Lvar)
            disjSetLvar((Set)s.arg1, (Lvar)s.arg2, s);                 //  {...}.disj(X)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar)
            disjLvarLvar((Lvar)s.arg1, (Lvar)s.arg2, s);               //  X.disj(Y)
        return;
    }
    
    
    protected void ndisj(Constraint s) throws Failure {
        if(s.arg1 instanceof Set && s.arg2 instanceof Set)
            ndisj((Set)s.arg1, (Set)s.arg2, s);                        //  {...}.ndisj( {...})
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Set)
            ndisjLvarSet((Lvar)s.arg1, (Set)s.arg2, s);                //  X.ndisj( {...})
        else if(s.arg1 instanceof Set && s.arg2 instanceof Lvar)
            ndisjSetLvar((Set)s.arg1, (Lvar)s.arg2, s);                //  {...}.ndisj(X)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar)
            ndisjLvarLvar((Lvar)s.arg1, (Lvar)s.arg2, s);              //  X.ndisj(Y)
        return;
    }
    
    
    protected void union(Constraint s) throws Failure {
        if(s.arg1 instanceof Set && s.arg2 instanceof Set && s.arg3 instanceof Set && s.arg4 instanceof Set)
            union((Set)s.arg1, (Set)s.arg2, (Set)s.arg3, (Set)s.arg4, s);//  union(s1, s2, s3, N)
        else if(s.arg1 instanceof Set && s.arg2 instanceof Set && s.arg3 instanceof Set && s.arg4 instanceof Lst)
            union((Set)s.arg1, (Set)s.arg2, (Set)s.arg3, (Set)((Lst)s.arg4).get(0), (Set)((Lst)s.arg4).get(1), s);
        //  union(s1, s2, s3, N, N1)
        else if(s.arg1 instanceof Set && s.arg2 instanceof Set && s.arg3 instanceof Set)
            union((Set)s.arg1, (Set)s.arg2, (Set)s.arg3, s);            //  union(s1, s2, s3)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Set && s.arg3 instanceof Set)
            union((Lvar)s.arg1, (Set)s.arg2, (Set)s.arg3, s);           //  union(X, s1, s2)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar && s.arg3 instanceof Set)
            union((Lvar)s.arg1, (Lvar)s.arg2, (Set)s.arg3, s);          //  union(X, Y, s)
        else if(s.arg1 instanceof Set && s.arg2 instanceof Set && s.arg3 instanceof Lvar)
            union((Set)s.arg1, (Set)s.arg2, (Lvar)s.arg3, s);           //  union(s1, s2, X)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Set && s.arg3 instanceof Lvar)
            union((Lvar)s.arg1, (Set)s.arg2, (Lvar)s.arg3, s);          //  union(X, s, Y)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar && s.arg3 instanceof Lvar)
            union((Lvar)s.arg1, (Lvar)s.arg2, (Lvar)s.arg3, s);         //  union(X, Y, Z)
        return;
    }
    
    
    protected void nunion(Constraint s) throws Failure {
        if(s.arg1 instanceof Set && s.arg2 instanceof Set && s.arg3 instanceof Set)
            nunion((Set)s.arg1, (Set)s.arg2, (Set)s.arg3, s);          //  nunion(s1, s2, s3)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Set && s.arg3 instanceof Set)
            nunion((Lvar)s.arg1, (Set)s.arg2, (Set)s.arg3, s);         //  nunion(X, s1, s2)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar && s.arg3 instanceof Set)
            nunion((Lvar)s.arg1, (Lvar)s.arg2, (Set)s.arg3, s);        //  nunion(X, Y, s2)
        else if(s.arg1 instanceof Set && s.arg2 instanceof Set && s.arg3 instanceof Lvar)
            nunion((Set)s.arg1, (Set)s.arg2, (Lvar)s.arg3, s);         //  nunion(s1, s2, X)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Set && s.arg3 instanceof Lvar)
            nunion((Lvar)s.arg1, (Set)s.arg2, (Lvar)s.arg3, s);        //  nunion(X, s2, Y)
        else if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar && s.arg3 instanceof Lvar)
            nunion((Lvar)s.arg1, (Lvar)s.arg2, (Lvar)s.arg3, s);       //  nunion(X, Y, Z)
        return;
    }
    
    
    protected void ge(Constraint s) throws Failure {
        if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar)
            ge((Lvar)s.arg1, (Lvar)s.arg2, s);      // X ge Y
        else if(s.arg1 instanceof Lvar)
            ge((Lvar)s.arg1, (Integer)s.arg2, s);   // X ge i
        return;
    }
    
    
    protected void gt(Constraint s)throws Failure {
        if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar)
            gt((Lvar)s.arg1, (Lvar)s.arg2, s);      // X gt Y
        else if(s.arg1 instanceof Lvar)
            gt((Lvar)s.arg1, (Integer)s.arg2, s);   // X gt i
        return;
    }
    
    
    protected void le(Constraint s) throws Failure {
        if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar)
            le((Lvar)s.arg1, (Lvar)s.arg2, s);      // X le Y
        else  if(s.arg1 instanceof Lvar)
            le((Lvar)s.arg1, (Integer)s.arg2, s);       // X le i
        return;
    }
    
    
    protected void lt(Constraint s) throws Failure {
        if(s.arg1 instanceof Lvar && s.arg2 instanceof Lvar)
            lt((Lvar)s.arg1, (Lvar)s.arg2, s);          // X lt Y
        else if(s.arg1 instanceof Lvar)
            lt((Lvar)s.arg1, (Integer)s.arg2, s);       // X lt i
        return;
    }
    
    
    protected void sum(Constraint s) throws Failure {
        if(s.arg2 instanceof Lvar && s.arg3 instanceof Lvar)
            sumLvar((Lvar)s.arg1, (Lvar)s.arg2, (Lvar)s.arg3, s);           // X = Y+Z
        if(s.arg2 instanceof Integer && s.arg3 instanceof Integer)
            sumInteger((Lvar)s.arg1, (Integer)s.arg2, (Integer)s.arg3, s);  // X = Integ1+Integ2
        else if(s.arg2 instanceof Lvar && s.arg3 instanceof Integer)
            sumLvarInteger((Lvar)s.arg1, (Lvar)s.arg2, (Integer)s.arg3, s); // X = Y+Integ
        else if(s.arg2 instanceof Integer && s.arg3 instanceof Lvar)
            sumLvarInteger((Lvar)s.arg1, (Lvar)s.arg3, (Integer)s.arg2, s); // X = Integ+Y
        return;
    }
    
    
    protected void sub(Constraint s) throws Failure {
        if(s.arg2 instanceof Lvar && s.arg3 instanceof Lvar)
            subLvar((Lvar)s.arg1, (Lvar)s.arg2, (Lvar)s.arg3, s);           // X = Y-Z
        if(s.arg2 instanceof Integer && s.arg3 instanceof Integer)
            subInteger((Lvar)s.arg1, (Integer)s.arg2, (Integer)s.arg3, s);  // X = Integ1-Integ2
        else if(s.arg2 instanceof Lvar && s.arg3 instanceof Integer)
            subLvarInteger((Lvar)s.arg1, (Lvar)s.arg2, (Integer)s.arg3, s); // X = Y-Integ
        else if(s.arg2 instanceof Integer && s.arg3 instanceof Lvar)
            subLvarInteger((Lvar)s.arg1, (Lvar)s.arg3, (Integer)s.arg2, s); // X = Integ-Y
        return;
    }
    
    
    protected void mul(Constraint s) throws Failure {
        if(s.arg2 instanceof Lvar && s.arg3 instanceof Lvar)
            mulLvar((Lvar)s.arg1, (Lvar)s.arg2, (Lvar)s.arg3, s);           // X = Y*Z
        if(s.arg2 instanceof Integer && s.arg3 instanceof Integer)
            mulInteger((Lvar)s.arg1, (Integer)s.arg2, (Integer)s.arg3, s);  // X = Integ1*Integ2
        else if(s.arg2 instanceof Lvar && s.arg3 instanceof Integer)
            mulLvarInteger((Lvar)s.arg1, (Lvar)s.arg2, (Integer)s.arg3, s); // X = Y*Integ
        else if(s.arg2 instanceof Integer && s.arg3 instanceof Lvar)
            mulLvarInteger((Lvar)s.arg1, (Lvar)s.arg3, (Integer)s.arg2, s); // X = Integ*Y
        return;
    }
    
    
    protected void div(Constraint s) throws Failure {
        if(s.arg2 instanceof Lvar && s.arg3 instanceof Lvar)
            divLvar((Lvar)s.arg1, (Lvar)s.arg2, (Lvar)s.arg3, s);           // X = Y/Z
        if(s.arg2 instanceof Integer && s.arg3 instanceof Integer)
            divInteger((Lvar)s.arg1, (Integer)s.arg2, (Integer)s.arg3, s);  // X = Integ1/Integ2
        else if(s.arg2 instanceof Lvar && s.arg3 instanceof Integer)
            divLvarInteger((Lvar)s.arg1, (Lvar)s.arg2, (Integer)s.arg3, s); // X = Y/Integ
        else if(s.arg2 instanceof Integer && s.arg3 instanceof Lvar)
            divLvarInteger((Lvar)s.arg1, (Lvar)s.arg3, (Integer)s.arg2, s); // X = Integ/Y
        return;
    }
    
    
    protected void mod(Constraint s) throws Failure {
        if(s.arg2 instanceof Lvar && s.arg3 instanceof Lvar)
            modLvar((Lvar)s.arg1, (Lvar)s.arg2, (Lvar)s.arg3, s);           // X = Y mod Z
        if(s.arg2 instanceof Integer && s.arg3 instanceof Integer)
            modInteger((Lvar)s.arg1, (Integer)s.arg2, (Integer)s.arg3, s);  // X = Integ1 mod Integ2
        else if(s.arg2 instanceof Lvar && s.arg3 instanceof Integer)
            modLvarInteger((Lvar)s.arg1, (Lvar)s.arg2, (Integer)s.arg3, s); // X = Y mod Integ
        else if(s.arg2 instanceof Integer && s.arg3 instanceof Lvar)
            modIntegerLvar((Lvar)s.arg1, (Integer)s.arg2, (Lvar)s.arg3, s); // X = Integ mod Y
        return;
    }
    
    
    protected void forall(Constraint s) throws Failure {
        // 22/04/04 Gestione (Lst di Lvar, Lst di Set, ConstraintsConjunction)
        if(s.arg1 instanceof Lst && s.arg2 instanceof Lst && s.arg3 instanceof ConstraintsConjunction) {
            forall((Lst)s.arg1, (Lst)s.arg2, (ConstraintsConjunction)s.arg3, s);
        }
        // 23/04/04 Gestione (Lst di Lvar, Lst di Set, Constraint)
        else if(s.arg1 instanceof Lst && s.arg2 instanceof Lst && s.arg3 instanceof Constraint) {
            forall((Lst)s.arg1, (Lst)s.arg2, (Constraint)s.arg3, s);
        }
        // 23/04/04 Gestione (Lst di Lvar, Lst di Set, Lst di Lvar esist., Constraint)
        else if(s.arg1 instanceof Lst && s.arg2 instanceof Lst && s.arg3 instanceof Lst && s.arg4 instanceof Constraint) {
            forall((Lst)s.arg1, (Lst)s.arg2, (Lst)s.arg3, (Constraint)s.arg4, s);
        }
        // 23/04/04 Gestione (Lst di Lvar, Lst di Set, Lst di Lvar esist., ConstraintsConjunction)
        else if(s.arg1 instanceof Lst && s.arg2 instanceof Lst && s.arg3 instanceof Lst && s.arg4 instanceof ConstraintsConjunction) {
            forall((Lst)s.arg1, (Lst)s.arg2, (Lst)s.arg3, (ConstraintsConjunction)s.arg4, s);
        }
        // forall(Lvar, Set, Constraint)
        else if((Lst)s.arg3 == Lst.empty && s.arg4 instanceof Constraint) {
            if(s.arg2 instanceof Set) forall((Lvar)s.arg1, (Set)s.arg2, (Constraint)s.arg4, s);
            else if(s.arg2 instanceof Int) forall((Lvar)s.arg1, (Int)s.arg2, (Constraint)s.arg4, s);
        }
        // forall(Lvar, Set, (Lst)VarEsist, Constraint)
        else if(s.arg4 instanceof Constraint) {
            if(s.arg2 instanceof Set) forall((Lvar)s.arg1, (Set)s.arg2, (Lst)s.arg3, (Constraint)s.arg4, s);
            else if(s.arg2 instanceof Int) forall((Lvar)s.arg1, (Int)s.arg2, (Lst)s.arg3, (Constraint)s.arg4, s);
        }
        // forall(Lvar, Set, ConstraintsConjunction)
        else if((Lst)s.arg3 == Lst.empty && s.arg4 instanceof ConstraintsConjunction) {
            if(s.arg2 instanceof Set) forall((Lvar)s.arg1, (Set)s.arg2, (ConstraintsConjunction)s.arg4, s);
            else if(s.arg2 instanceof Int) forall((Lvar)s.arg1, (Int)s.arg2, (ConstraintsConjunction)s.arg4, s);
        }
        // forall(Lvar, Set, (Lst)VarEsist, ConstraintsConjunction)
        else if(s.arg4 instanceof ConstraintsConjunction) {
            if(s.arg2 instanceof Set) forall((Lvar)s.arg1, (Set)s.arg2, (Lst)s.arg3, (ConstraintsConjunction)s.arg4, s);
            else if(s.arg2 instanceof Int) forall((Lvar)s.arg1, (Int)s.arg2, (Lst)s.arg3, (ConstraintsConjunction)s.arg4, s);
        }
        return;
    }
    
    
    protected void interv(Constraint s) throws Failure {
        if(s.arg2 instanceof Lvar && s.arg3 instanceof Lvar)
            interv((Set)s.arg1, (Lvar)s.arg2, (Lvar)s.arg3, s);
        else if(s.arg2 instanceof Lvar && s.arg3 instanceof Integer)
            interv((Set)s.arg1, (Lvar)s.arg2, (Integer)s.arg3, s);
        else if(s.arg2 instanceof Integer && s.arg3 instanceof Lvar)
            interv((Set)s.arg1, (Integer)s.arg2, (Lvar)s.arg3, s);
        return;
    }
    
    
    protected void cond_if(Constraint s) throws Failure {
        // cond_if senza else
        if(s.arg1 instanceof Constraint && s.arg2 instanceof Constraint && s.arg3 == null)
            condif((Constraint)s.arg1, (Constraint)s.arg2, s);
        else if(s.arg1 instanceof Constraint && s.arg2 instanceof ConstraintsConjunction && s.arg3 == null)
            condif((Constraint)s.arg1, (ConstraintsConjunction)s.arg2, s);
        // cond_if con else
        else if(s.arg1 instanceof Constraint && s.arg2 instanceof ConstraintsConjunction && s.arg3 instanceof ConstraintsConjunction)
            condif((Constraint)s.arg1, (ConstraintsConjunction)s.arg2, (ConstraintsConjunction)s.arg3, s);
        else if(s.arg1 instanceof Constraint && s.arg2 instanceof Constraint && s.arg3 instanceof ConstraintsConjunction)
            condif((Constraint)s.arg1, (Constraint)s.arg2, (ConstraintsConjunction)s.arg3, s);
        else if(s.arg1 instanceof Constraint && s.arg2 instanceof ConstraintsConjunction && s.arg3 instanceof Constraint)
            condif((Constraint)s.arg1, (ConstraintsConjunction)s.arg2, (Constraint)s.arg3, s);
        else if(s.arg1 instanceof Constraint && s.arg2 instanceof Constraint && s.arg3 instanceof Constraint)
            condif((Constraint)s.arg1, (Constraint)s.arg2, (Constraint)s.arg3, s);
        return;
    }
    
    
    protected void debug(Constraint s) throws Failure {
        // cond_if senza else
        if(s.arg1 instanceof String && s.arg2 instanceof Lvar && s.arg3 == null)
            debug((String)s.arg1, (Lvar)s.arg2, s);
        // else if(s.arg1 instanceof Constraint && s.arg2 instanceof ConstraintsConjunction && s.arg3 == null)
        return;
    }
    
    
    /*  x.eq(z) con x e z istanze di Lvar:
        Se entrambe sono non inizializzate, una resta invariata l'altra diventa
        inizializzata e il suo campo equ diventa un riferimento all'altra Lvar.
        Se una delle due  inizializzata, anche l'altra lo diventa e il suo campo
        equ diventa un riferimento all'altra Lvar.
     */
    protected void eqLvar(Lvar l1, Lvar l2, Constraint s) throws Failure {
        if(l1.equ == null && l2.equ == null) {
            if(l1 == l2) {
                s.bool = true;
                return;
            }
            else if(!l1.iniz && !l2.iniz) {
                if(l2.name.length()>5 && l2.name.substring(0, 5).equals("Lvar_")) {
                    l2.iniz = true;
                    l2.equ = l1;
                    s.arg2 = l1;
                    s.bool = true;
                }
                else {
                    l1.iniz = true;
                    l1.equ = l2;
                    s.arg1 = l2;
                    s.bool = true;
                }
                Solver.storeInvariato = false;
                return;
            }
            else if(l1.iniz && !l2.iniz) {
                l2.iniz = true;
                l2.equ = l1;
                s.arg2 = l1;
                s.bool = true;
                Solver.storeInvariato = false;
                return;
            }
            else if(l2.iniz && !l1.iniz) {
                l1.iniz = true;
                l1.equ = l2;
                s.arg1 = l2;
                s.bool = true;
                Solver.storeInvariato = false;
                return;
            }
            /* x.eq(z) con x e y entrambe inizializzate.
               Viene restituito il risultato di x.val.eq(y.val)
             */
            else {
                s.arg1 = l1.val;
                s.arg2 = l2.val;
                eq(s);
                return;
            }
        }
        else if(l2.equ == null) {
            s.arg1 = l1.equ;
            eqLvar(l1.equ, l2, s);
            return;
        }
        else if(l1.equ == null) {
            s.arg2 = l2.equ;
            eqLvar(l1, l2.equ, s);
            return;
        }
        else {
            s.arg1 = l1.equ;
            s.arg2 = l2.equ;
            eqLvar(l1.equ, l2.equ, s);
            return;
        }
    }
    
    
    /* x.eq(z) con x istanza di Lvar e z istanza di Lst:
       Se entrambe sono non inizializzate, z resta invariata, x diventa
       inizializzata e il suo campo val diventa un riferimento a z.
       Se solo x  inizializzata, vengono confrontate z e x.val.
       Se solo z  inizializzata: nel caso in cui x sia un elemento della
       lista z, si ha un fallimento; altrimenti x.val diventa un riferimento a z.*/
    protected void eqLvarLst(Lvar l1, Lst l2, Constraint s) throws Failure {
        if(l1.equ == null && l2.equ == null) {
            if(l1.val == l2) {
                s.bool = true;
                return;
            }
            else if(!l1.iniz && !l2.iniz) {
                l1.iniz = true;
                l1.val = l2;
                Solver.storeInvariato = false;
                s.bool = true;
                return;
            }
            else if(l1.iniz && !l2.iniz) {
                if(l1.val instanceof Lst) {
                    s.arg1 = (Lst)l1.val;
                    s.arg2 = l2;
                    eqLst((Lst)l1.val, l2, s);
                    Solver.storeInvariato = false;
                    return;
                }
                else {
                    Solver.fail(s);
                    return;
                }
            }
            else if(l2.iniz && !l1.iniz) {
                if(l2.occurs(l1)) {
                    Solver.fail(s);
                    return;
                }
                else {
                    l1.iniz = true;
                    l1.val = l2;
                    Solver.storeInvariato = false;
                    s.bool = true;
                    return;
                }
            }
            /* x.eq(y) con x istanza di Lvar, y istanza di Lst ed entrambe inizializzate.
               Viene restituito il risultato di y.eq(x.val) */
            else {
                if(l1.val instanceof Lst) {
                    s.arg1 = (Lst)l1.val;
                    s.arg2 = l2;
                    eqLst((Lst)l1.val, l2, s);
                    Solver.storeInvariato = false;
                    return;
                }
                else {
                    Solver.fail(s);
                    return;
                }
            }
        }
        else if(l2.equ == null) {
            s.arg1 = l1.equ;
            s.arg2 = l2;
            eqLvarLst(l1.equ, l2, s);
            return;
        }
        else if(l1.equ == null) {
            s.arg1 = l1;
            s.arg2 = l2.equ;
            eqLvarLst(l1, l2.equ, s);
            return;
        }
        else {
            s.arg1 = l1.equ;
            s.arg2 = l2.equ;
            eqLvarLst(l1.equ, l2.equ, s);
            return;
        }
    }
    
    
    /*  unificazione tra liste */
    protected void eqLst(Lst l1, Lst l2, Constraint s) throws Failure {
        if(l1.equ == null && l2.equ == null) {
            if(l1 == l2) {
                s.bool = true;
                return;
            }
            else if(!l1.iniz && !l2.iniz) {      // entrambe non inizializzate
                l2.iniz = true;
                l2.equ = l1;
                s.arg2 = l1;
                Solver.storeInvariato = false;
                s.bool = true;
                return;
            }
            else if(!l1.iniz) {                     // l2 inizializzata e l1 non inizializzata
                if(l2.occurs(l1) || (l2.ultimoResto() == l1 && l2.size_private()!= 0)) {
                    Solver.fail(s);
                    return;
                }
                else {
                    l1.iniz =true;
                    l1.equ = l2;
                    s.arg1 = l2;
                    Solver.storeInvariato = false;
                    s.bool = true;
                    return;
                }
            }
            else if(!l2.iniz) {                     // l1 inizializzata e l2 non inizializzata
                if(l1.occurs(l2) || (l1.ultimoResto() == l2 && l1.size_private()!= 0)) {
                    Solver.fail(s);
                    return;
                }
                else {
                    l2.iniz =true;
                    l2.equ = l1;
                    s.arg2 = l1;
                    Solver.storeInvariato = false;
                    s.bool = true;
                    return;
                }
            }
            else {                                   //entrambe inizializzate
                if(l1.occurs(l2) || l2.occurs(l1)) {
                    Solver.fail(s);
                    return;
                }
                else {
                    if(l1 == Lst.empty && l2 == Lst.empty)return;
                    else if(l1 == Lst.empty || l2 == Lst.empty) {
                        Solver.fail(s);
                        return;
                    }
                    else {
                        s.arg1 = l1.get(0);
                        s.arg2 = l2.get(0);
                        Constraint se0 = new Constraint(l1.sub(), Enviroment.eqCode, l2.sub());
                        Solver.add(Solver.indexOf(s)+1, se0);
                        eq(s);
                        Solver.storeInvariato = false;
                        return;
                    }
                }
            }
        }
        else if(l1.equ == null) {
            s.arg2 = l2.equ;
            eqLst(l1, l2.equ, s);
            return;
        }
        else if(l2.equ == null) {
            s.arg1 = l1.equ;
            eqLst(l1.equ, l2, s);
            return;
        }
        else {
            s.arg1 = l1.equ;
            s.arg2 = l2.equ;
            eqLst(l1.equ, l2.equ, s);
            return;
        }
    }
    
    
    /* Unificazione tra una Lvar e un Set
       x.eq(z) con x istanza di Lvar e z istanza di Set:
       Se entrambe sono non inizializzate, z resta invariata, x diventa
       inizializzata e il suo campo val diventa un riferimento a z.
       Se solo x  inizializzata, vengono confrontate z e x.val.
       Se solo z  inizializzata, x.val diventa un riferimento a z.*/
    protected void eqLvarSet(Lvar l, Set set, Constraint s) throws Failure {
        if(l.equ == null && set.equ ==  null) {
            if(l.val == set) {
                s.bool = true;
                return;
            }
            else if(!l.iniz && !set.iniz) {
                l.iniz = true;
                l.val = set;
                s.arg1 = set;
                s.arg2 = set;
                s.bool = true;
                Solver.storeInvariato = false;
                return;
            }
            else if(l.iniz && !set.iniz) {
                if(l.val instanceof Set) {
                    s.arg1 = (Set)l.val;
                    s.arg2 = set;
                    s.caseControl = 0;
                    eqSet((Set)l.val, set, s);
                    Solver.storeInvariato = false;
                    return;
                }
                else {
                    Solver.fail(s);
                    return;
                }
            }
            else if(set.iniz && !l.iniz) {
                if(set.occurs(l)) {
                    Solver.fail(s);
                    return;
                }
                else {
                    l.iniz = true;
                    l.val = set;
                    s.arg1 = set;
                    s.arg2 = set;
                    s.bool = true;
                    Solver.storeInvariato = false;
                    return;
                }
            }
            /* x.eq(z) con x istanza di Lvar, z istanza di Set ed entrambe inizializzate.
               Viene restituito il risultato di z.eq(x.val) */
            else {
                if(!(l.val instanceof Set)) {
                    Solver.fail(s);
                    return;
                }
                else {
                    s.arg1 = (Set)l.val;
                    s.arg2 = set;
                    s.caseControl = 0;
                    eqSet((Set)l.val, set, s);
                    return;
                }
            }
        }
        else if(set.equ == null) {
            s.arg1 = l.equ;
            s.arg2 = set;
            eqLvarSet(l.equ, set, s);
            return;
        }
        else if(l.equ == null) {
            s.arg1 = l;
            s.arg2 = set.equ;
            eqLvarSet(l, set.equ, s);
            return;
        }
        else {
            s.arg1 = l.equ;
            s.arg2 = set.equ;
            eqLvarSet(l.equ, set.equ, s);
            return;
        }
    }
    
    
    /*  Unificazione tra insiemi */
    protected void eqSet(Set set1, Set set2, Constraint s) throws Failure {
        if(set1.equ == null && set2.equ == null) {
            switch(s.caseControl) {
                case 0:                                            //  (1)
                    if(set1 == set2) {
                        s.bool = true;
                        return;
                    }
                    //           else if(set1.isInt) {
                    //              s.arg1 = (Int)set1.inter;
                    //              eq(s);
                    //              return;
                    //           }
                    //           else if(set2.isInt) {
                    //              s.arg2 = (Int)set2.inter;
                    //              eq(s);
                    //              return;
                    //           }
                    else if(!set1.iniz && !set2.iniz) {
                        set2.iniz = true;
                        set2.equ = set1;
                        s.arg2 = set1;
                        s.bool = true;
                        Solver.storeInvariato = false;
                        return;
                    }
                    else if(!set1.iniz) {                         //  (2)
                        if(set2 == Set.empty) {
                            set1.equ = set2;
                            Solver.storeInvariato = false;
                            return;
                        }
                        else if(set2.occurs(set1)) {
                            Solver.fail(s);
                            return;
                        }
                        else {                                     //  (6)
                            if(set2.ultimoResto() == set1 && !set2.isEmpty()) {  // X.eq( {t0, .., tn|X})
                                Set N = new Set();
                                Set M = new Set(set2.toVector(), N);  // X.eq( {t0, ..., tn|N})
                                s.arg2 = M;
                                eqSet(set1, M, s);
                                Solver.storeInvariato = false;
                                return;
                            }
                            else {                     // X.eq( {t0, .., tn|N})
                                set1.iniz = true;
                                set1.equ = set2;
                                s.bool = true;
                                Solver.storeInvariato = false;
                                return;
                            }
                        }
                    }
                    else if(!set2.iniz) {                         //  (2)
                        if(set1 == Set.empty) {
                            set2.equ = set1;
                            Solver.storeInvariato = false;
                            return;
                        }
                        else if(set1.occurs(set2)) {
                            Solver.fail(s);
                            return;
                        }
                        else {                                   //  (6)
                            if(set1.ultimoResto() == set2 && !set1.isEmpty()) {    // {t1, .., tn|X}.eq(X)
                                Set N = new Set();
                                Set M = new Set(set1.toVector(), N);
                                s.arg1 = M;
                                eqSet(M, set2, s);
                                Solver.storeInvariato = false;
                                return;
                            }
                            else {
                                set2.iniz =true;
                                set2.equ = set1;
                                Solver.storeInvariato = false;
                                return;
                            }
                        }
                    }
                    else {
                        if(set1.occurs(set2) || set2.occurs(set1)) {
                            Solver.fail(s);
                            return;
                        }
                        else {
                            if(set1.size() == 0 && set2.size() == 0) return;
                            else if(set1.size() == 0 || set2.size() == 0) {
                                Solver.fail(s);
                                return;
                            }                                         //  (9)
                            else if(set1.ultimoResto() != set2.ultimoResto() ||
                            (set1.ultimoResto().equals(Set.empty) && set2.ultimoResto().equals(Set.empty))) {   // {t|S}.eq( {t'|S'})
                                s.caseControl = 1;
                                eqSet(set1, set2, s);
                                return;
                            }                                         //  (10)
                            else if(set1.ultimoResto() == set2.ultimoResto()
                            && set1.ultimoResto() != Set.empty) {  // {t1, .., tn|X}.eq( {s1, .., sm|X})
                                s.caseControl = 5;
                                eqSet(set1, set2, s);
                                return;
                            }
                        }
                    }
                case 1:                                // {t|S}.eq( {t'|S'})    (9)(i)
                    Solver.B.addChoicePoint(2, (VarState)Solver.B.getVarState(),
                    (StoreState)Solver.B.getStoreState(s));
                    s.arg1 = set1.first();                  // t.eq(t')
                    s.arg2 = set2.first();
                    s.caseControl = 0;
                    Constraint se0 = new Constraint(set1.sub(), Enviroment.eqCode, set2.sub());
                    Solver.add(Solver.indexOf(s)+1, se0);      // S.eq(S')
                    eq(s);
                    Solver.storeInvariato = false;
                    return;
                case 2:                                // {t|S}.eq( {t'|S'})    (9)(ii)
                    Solver.B.addChoicePoint(3, (VarState)Solver.B.getVarState(),
                    (StoreState)Solver.B.getStoreState(s));
                    s.arg1 = set1.first();                  // t.eq(t')
                    s.arg2 = set2.first();
                    s.caseControl = 0;
                    Constraint se1 = new Constraint(set1, Enviroment.eqCode, set2.sub());
                    Solver.add(Solver.indexOf(s)+1, se1);
                    eq(s);
                    Solver.storeInvariato = false;
                    return;
                case 3:                                // {t|S}.eq( {t'|S'})    (9)(iii)
                    Solver.B.addChoicePoint(4, (VarState)Solver.B.getVarState(),
                    (StoreState)Solver.B.getStoreState(s));
                    s.arg1 = set1.first();                  // t.eq(t')
                    s.arg2 = set2.first();
                    s.caseControl = 0;
                    Constraint se2 = new Constraint(set1.sub(), Enviroment.eqCode, set2);
                    Solver.add(Solver.indexOf(s)+1, se2);
                    eq(s);
                    Solver.storeInvariato = false;
                    return;
                case 4:                                // {t|S}.eq( {t'|S'})    (9)(iv)
                    Set n = new Set();
                    s.arg1 = set1.sub();
                    s.arg2 = n.ins(set2.first());
                    s.caseControl = 0;
                    Constraint se3 = new Constraint(n.ins(set1.first()), Enviroment.eqCode, set2.sub());
                    Solver.add(Solver.indexOf(s)+1, se3);
                    eq(s);
                    Solver.storeInvariato = false;
                    return;
                case 5:                                // {t1, .., tn|X}.eq( {s1, .., sm|X})  (10) (i, ii, iii)
                    Solver.B.addChoicePoint(6, (VarState)Solver.B.getVarState(),
                    (StoreState)Solver.B.getStoreState(s));
                    Lvar set2_elem = new Lvar();
                    Set w = new Set();
                    Constraint s5 = new Constraint(Enviroment.selectCode, set2_elem, set2, w);         //select
                    Constraint ss5 = new Constraint(set2_elem, Enviroment.eqCode, set1.first());   //eq
                    Solver.add(Solver.indexOf(s), s5);
                    Solver.add(Solver.indexOf(s), ss5);
                    s.arg3 = w;
                    s.caseControl = 7;
                    Solver.storeInvariato = false;
                    return;
                case 6:                                // {t1, .., tn|X}.eq( {s1, .., sm|X})  (10) (iv)
                    Set N = new Set();
                    s.arg1 = set1.ultimoResto();
                    s.arg2 = N.ins(set1.first());
                    s.caseControl = 0;
                    Constraint s6 = new Constraint(set1.sub().append(N), Enviroment.eqCode, set2.append(N));
                    Solver.add(Solver.indexOf(s)+1, s6);
                    eq(s);
                    Solver.storeInvariato = false;
                    return;
            }
        }
        else if(set1.equ == null) {
            s.arg2 = set2.equ;
            eqSet(set1, set2.equ, s);
            return;
        }
        else if(set2.equ == null) {
            s.arg1 = set1.equ;
            eqSet(set1.equ, set2, s);
            return;
        }
        else {
            s.arg1 = set1.equ;
            s.arg2 = set2.equ;
            eqSet(set1.equ, set2.equ, s);
            return;
        }
    }
    
    
    
    /* unificazione tra insiemi */
    protected void eqSet(Set set1, Set set2, Set w, Constraint s) throws Failure {
        if(set1.equ == null && set2.equ == null && w.equ == null) {
            switch(s.caseControl) {
                case 7:                          // {t1, .., tn|X}.eq( {s1, .., sm|X})  (10) (i)
                    Solver.B.addChoicePoint(8, (VarState)Solver.B.getVarState(),
                    (StoreState)Solver.B.getStoreState(s));
                    s.arg1 = set1.sub();
                    s.arg2 = w;
                    s.arg3 = null;
                    s.caseControl = 0;
                    eqSet(set1.sub(), w, s);
                    return;
                case 8:                          // {t1, .., tn|X}.eq( {s1, .., sm|X})  (10) (ii)
                    Solver.B.addChoicePoint(9, (VarState)Solver.B.getVarState(),
                    (StoreState)Solver.B.getStoreState(s));
                    s.arg1 = set1;
                    s.arg2 = w;
                    s.arg3 = null;
                    s.caseControl = 0;
                    eqSet(set1, w, s);
                    return;
                case 9:                          // {t1, .., tn|X}.eq( {s1, .., sm|X})  (10) (iii)
                    s.arg1 = set1.sub();
                    s.arg2 = set2;
                    s.arg3 = null;
                    s.caseControl = 0;
                    eqSet(set1.sub(), set2, s);
                    return;
            }
        }
        else if(set1.equ != null) {
            s.arg1 = set1.equ;
            eqSet(set1.equ, set2, w, s);
            return;
        }
        else if(set2.equ != null) {
            s.arg2 = set2.equ;
            eqSet(set1, set2.equ, w, s);
            return;
        }
        else {
            s.arg3 = w.equ;
            eqSet(set1, set2, w.equ, s);
            return;
        }
    }
    
    /*  Unificazione tra multinsiemi */
    protected void eqMultiSet(MultiSet mset1, MultiSet mset2, Constraint s) throws Failure {
        if(mset1.equ == null && mset2.equ == null) {
            switch(s.caseControl) {
                case 0:                                            //  (1)
                    if(mset1 == mset2) {
                        s.bool = true;
                        return;
                    }
                    else if(!mset1.iniz && !mset2.iniz) {
                        mset2.iniz = true;
                        mset2.equ = mset1;
                        s.arg2 = mset1;
                        s.bool = true;
                        Solver.storeInvariato = false;
                        return;
                    }
                    else if(!mset1.iniz) {                         //  (2)
                        if(mset2 == MultiSet.empty) {
                            mset1.equ = mset2;
                            Solver.storeInvariato = false;
                            return;
                        }
                        else if(mset2.occurs(mset1)) {
                            Solver.fail(s);
                            return;
                        }
                        else {                                     //  (3)
                            if(mset2.ultimoResto() == mset1 && !mset2.isEmpty()) {  // X.eq( {t0, .., tn|X})
                                MultiSet N = new MultiSet();
                                MultiSet M = new MultiSet(mset2.toVector(), N);  // X.eq( {t0, ..., tn|N})
                                s.arg2 = M;
                                eqMultiSet(mset1, M, s);
                                Solver.storeInvariato = false;
                                return;
                            }
                            else {                     // X.eq( {t0, .., tn|N})
                                mset1.iniz = true;
                                mset1.equ = mset2;
                                s.bool = true;
                                Solver.storeInvariato = false;
                                return;
                            }
                        }
                    }
                    else if(!mset2.iniz) {                         //  (2)
                        if(mset1 == MultiSet.empty) {
                            mset2.equ = mset1;
                            Solver.storeInvariato = false;
                            return;
                        }
                        else if(mset1.occurs(mset2)) {
                            Solver.fail(s);
                            return;
                        }
                        else {                                   //  (3)
                            if(mset1.ultimoResto() == mset2 && !mset1.isEmpty()) {    // {t1, .., tn|X}.eq(X)
                                MultiSet N = new MultiSet();
                                MultiSet M = new MultiSet(mset1.toVector(), N);
                                s.arg1 = M;
                                eqMultiSet(M, mset2, s);
                                Solver.storeInvariato = false;
                                return;
                            }
                            else {
                                mset2.iniz =true;
                                mset2.equ = mset1;
                                Solver.storeInvariato = false;
                                return;
                            }
                        }
                    }
                    else {
                        if(mset1.occurs(mset2) || mset2.occurs(mset1)) {    //  (4)
                            Solver.fail(s);
                            return;
                        }
                        else {
                            if(mset1.size() == 0 && mset2.size() == 0) return;
                            else if(mset1.size() == 0 || mset2.size() == 0) {
                                Solver.fail(s);
                                return;
                            }                                         //  (7)
                            else if(mset1.ultimoResto() != mset2.ultimoResto() ||
                            (mset1.ultimoResto().equals(MultiSet.empty) && mset2.ultimoResto().equals(MultiSet.empty))) {   // {t|S}.eq( {t'|S'})
                                s.caseControl = 1;
                                eqMultiSet(mset1, mset2, s);
                                return;
                            }                                         //  (8)
                            else if(mset1.ultimoResto() == mset2.ultimoResto()
                            && mset1.ultimoResto() != MultiSet.empty) {  // {t1, .., tn|X}.eq( {s1, .., sm|X})
                                s.caseControl = 3;
                                eqMultiSet(mset1, mset2, s);
                                return;
                            }
                        }
                    }
                case 1:                                // {t|S}.eq( {t'|S'})    (7)(i)
                    Solver.B.addChoicePoint(2, (VarState)Solver.B.getVarState(),
                    (StoreState)Solver.B.getStoreState(s));
                    s.arg1 = mset1.first();                  // t.eq(t')
                    s.arg2 = mset2.first();
                    s.caseControl = 0;
                    Constraint se0 = new Constraint(mset1.sub(), Enviroment.eqCode, mset2.sub());
                    Solver.add(Solver.indexOf(s)+1, se0);      // S.eq(S')
                    eq(s);
                    Solver.storeInvariato = false;
                    return;
                case 2:                                // {t|S}.eq( {t'|S'})    (7)(ii)
                    MultiSet n = new MultiSet();
                    s.arg1 = mset1.sub();
                    s.arg2 = n.ins(mset2.first());
                    s.caseControl = 0;
                    Constraint se1 = new Constraint(n.ins(mset1.first()), Enviroment.eqCode, mset2.sub());
                    Solver.add(Solver.indexOf(s)+1, se1);
                    eq(s);
                    Solver.storeInvariato = false;
                    return;
                case 3:                                // {t1, .., tn|X}.eq( {s1, .., sm|X})  (8)
                    MultiSet w = mset1.notultimoResto();
                    MultiSet z = mset2.notultimoResto();
                    
                    Constraint ss5 = new Constraint(w, Enviroment.eqCode, z);   //eq
                    
                    Solver.add(Solver.indexOf(s), ss5);
                    //s.arg3 = w;
                    s.bool = true;
                    Solver.storeInvariato = false;
                    return;
            }
        }
        else if(mset1.equ == null) {
            s.arg2 = mset2.equ;
            eqMultiSet(mset1, mset2.equ, s);
	    return;
	}
	else if(mset2.equ == null) {
	    s.arg1 = mset1.equ;
	    eqMultiSet(mset1.equ, mset2, s);
	    return;
	}
	else {
	    s.arg1 = mset1.equ;
	    s.arg2 = mset2.equ;
	    eqMultiSet(mset1.equ, mset2.equ, s);
	    return;
	}
    }


    /* Unificazione tra una Lvar e un MultiSet
       x.eq(z) con x istanza di Lvar e z istanza di MultiSet:
       Se entrambe sono non inizializzate, z resta invariata, x diventa
       inizializzata e il suo campo val diventa un riferimento a z.
       Se solo x  inizializzata, vengono confrontate z e x.val.
       Se solo z  inizializzata, x.val diventa un riferimento a z.*/
    protected void eqLvarMultiSet(Lvar l, MultiSet mset, Constraint s) throws Failure {
	if(l.equ == null && mset.equ ==  null) {
	    if(l.val == mset) {
		s.bool = true;
		return;
	    }
	    else if(!l.iniz && !mset.iniz) {
		l.iniz = true;
		l.val = mset;
		s.arg1 = mset;
		s.arg2 = mset;
		s.bool = true;
		Solver.storeInvariato = false;
		return;
	    }
	    else if(l.iniz && !mset.iniz) {
		if(l.val instanceof MultiSet) {
		    s.arg1 = (MultiSet)l.val;
		    s.arg2 = mset;
		    s.caseControl = 0;
		    eqMultiSet((MultiSet)l.val, mset, s);
		    Solver.storeInvariato = false;
		    return;
		}
		else {
		    Solver.fail(s);
		    return;
		}
	    }
	    else if(mset.iniz && !l.iniz) {
		if(mset.occurs(l)) {
		    Solver.fail(s);
		    return;
		}
		else {
		    l.iniz = true;
		    l.val = mset;
		    s.arg1 = mset;
		    s.arg2 = mset;
		    s.bool = true;
		    Solver.storeInvariato = false;
		    return;
		}
	    }
	    /* x.eq(z) con x istanza di Lvar, z istanza di Set ed entrambe inizializzate.
	       Viene restituito il risultato di z.eq(x.val) */
	    else {
		if(!(l.val instanceof MultiSet)) {
		    Solver.fail(s);
		    return;
		}
		else {
		    s.arg1 = (MultiSet)l.val;
		    s.arg2 = mset;
		    s.caseControl = 0;
		    eqMultiSet((MultiSet)l.val, mset, s);
		    return;
		}
	    }
	}
	else if(mset.equ == null) {
	    s.arg1 = l.equ;
	    s.arg2 = mset;
	    eqLvarMultiSet(l.equ, mset, s);
	    return;
	}
	else if(l.equ == null) {
	    s.arg1 = l;
	    s.arg2 = mset.equ;
	    eqLvarMultiSet(l, mset.equ, s);
	    return;
	}
	else {
	    s.arg1 = l.equ;
	    s.arg2 = mset.equ;
	    eqLvarMultiSet(l.equ, mset.equ, s);
	    return;
	}
    }



    /* x.eq(z) con x istanza di Lvar e z oggetto diverso da Lvar, Lst e Set:
       Se x  non inizializzata, diventa inizializzata e il suo campo val
       diventa un riferimento all'oggetto z.
       Se x  inizializzata, se x.val S un riferimento ad un oggetto di tipo
       Lst o Set allora viene restituito false, altrimenti viene confrontato
       x.val con z. */
    protected void eqLvarObj(Lvar l, Object o, Constraint s) throws Failure {
	if(l.equ == null) {
	    if(!l.iniz) {
		l.iniz = true;
		l.val = o;
		s.arg1 = o;
		s.arg2 = o;
		s.bool = true;
		Solver.storeInvariato = false;
		return;
	    }
	    else {
		if(l.val instanceof Lst) {
		    Solver.fail(s);
		    return;
		}
		if(l.val instanceof Set) {
		    Solver.fail(s);
		    return;
		}
		else {
		    if(l.val.equals(o)) {
			s.bool = true;
			return;
		    }
		    else {
			Solver.fail(s);
			return;
		    }
		}
	    }
	}
	else {
	    s.arg1 = l.equ;
	    s.arg2 = o;
	    s.bool = true;
	    eqLvarObj(l.equ, o, s);
	    return;
	}
    }


    /* Il vincolo di uguaglianza tra una lista e un oggetto diverso da Lst o Lvar fallisce sempre. */
    protected void eqLstObj(Lst l, Object o, Constraint s) throws Failure {
	Solver.fail(s);
	return;
    }


    /* Il vincolo di uguaglianza tra un insieme e un oggetto diverso da Set o Lvar fallisce sempre. */
    protected void eqSetObj(Set l, Object o, Constraint s) throws Failure {
	Solver.fail(s);
	return;
    }


    /* Il vincolo di uguaglianza tra un multi-insieme e un oggetto diverso da MultiSet, Set o Lvar fallisce sempre. */
    protected void eqMultiSetObj(MultiSet l, Object o, Constraint s) throws Failure {
	Solver.fail(s);
	return;
    }


    /*  Uguaglianza tra oggetti diversi da Lvar Lst o Set */
    protected void eqObj(Object arg1, Object arg2, Constraint s) throws Failure {
	if(arg1.equals(arg2)) {
	    s.bool = true;
	    return;
	}
	else Solver.fail(s);
	return;
    }


    /* x.neq(y) con x e y enrambe istanze di Lvar */
    protected void neqLvar(Lvar l1, Lvar l2, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(l1 == l2) {                                   //  (3)
		Solver.fail(s);
		return;
	    }
	    else if(!l1.iniz || !l2.iniz)return;
	    else {                                           //  (4)
		s.arg1 = l1.val;
		s.arg2 = l2.val;
		Solver.storeInvariato = false;
		return;
	    }
	}
	else {
	    if(l1.equ == null) {
		s.arg2 = l2.equ;
		neq(s);
		return;
	    }
	    else if(l2.equ == null) {
		s.arg1 = l1.equ;
		neq(s);
		return;
	    }
	    else {
		s.arg1 = l1.equ;
		s.arg2 = l2.equ;
		neq(s);
		return;
	    }
	}
    }


    /* x.neq(y) con x istanza di Lvar, y istanza di Lst. Viene restituito il risultato di y.neq(x.val) */
    protected void neqLvarLst(Lvar l1, Lst l2, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(!l1.iniz || !l2.iniz) return;
	    else {
		if(l1.val == l2) {
		    Solver.fail(s);
		    return;
		}
		else if(l2.occurs(l1)) return;
		else {
		    s.arg1 = l1.val;
		    s.arg2 = l2;
		    neq(s);
		    return;
		}
	    }
	}
	else if(l1.equ == null) {
	    s.arg1 = l1;
	    s.arg2 = l2.equ;
	    neq(s);
	    return;
	}
	else if(l2.equ == null) {
	    s.arg1 = l1.equ;
	    s.arg2 = l2;
	    neq(s);
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    s.arg2 = l2.equ;
	    neq(s);
	    return;
	}
    }


    /*  x.neq(y) con x e y enrambe istanze di Lst. */
    protected void neqLst(Lst l1, Lst l2, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(l1 == l2) {
		Solver.fail(s);
		return;
	    }
	    else if(!l1.iniz || !l2.iniz) return;
	    else if(l1 == Lst.empty && l2 == Lst.empty) {
		Solver.fail(s);
		return;
	    }
	    else if(l1.size_private() == 0 && l2.size_private() == 0) {
		if(l1.ultimoResto() == l2.ultimoResto()) {
		    Solver.fail(s);
		    return;
		}
		else {
		    s.arg1 = l1.ultimoResto();
		    s.arg2 = l2.ultimoResto();
		    Solver.storeInvariato = false;
		    return;
		}
	    }
	    else if(l1.size_private() == 0 || l2.size_private() == 0) {
		s.bool = true;
		return;
	    }
	    else {
		if(l2.size_private() != l1.size_private()) {
		    if(l2.ultimoResto() == l1.ultimoResto()) {
			s.bool = true;
			return;
		    }
		    else return;
		}
		else if(l1.occurs(l2) || l2.occurs(l1)) {
		    s.bool = true;
		    return;
		}
		else {
		    switch(s.caseControl) {
		    case 0:
			VarState statoVarNeq = Solver.B.getVarState();
			StoreState statoStoreNeq = Solver.B.getStoreState(s);
			Solver.B.addChoicePoint(1, statoVarNeq, statoStoreNeq);
			s.arg1 = l1.first();
			s.arg2 = l2.first();
			s.cons = Enviroment.neqCode;
			Solver.storeInvariato = false;
			return;
		    case 1:
			s.arg1 = (Lst)l1.sub();
			s.arg2 = (Lst)l2.sub();
			s.cons = Enviroment.neqCode;
			s.caseControl = 0;
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }
	}
	else if(l1.equ == null) {
	    s.arg2 = l2.equ;
	    neq(s);
	    return;
	}
	else if(l2.equ == null) {
	    s.arg1 = l1.equ;
	    neq(s);
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    s.arg2 = l2.equ;
	    neq(s);
	    return;
	}
    }


    protected void neqLvarSet(Lvar l, Set set, Constraint s) throws Failure {
	if(l.equ == null && set.equ == null) {
	    if(!l.iniz || !set.iniz)return;
	    else {
		if(l.val instanceof Set && ((Set)l.val).isEmpty() && set.isEmpty()) {
		    Solver.fail(s);
		    return;
		}
		else if(set.isEmpty() || set.occurs(l) || !(l.val instanceof Set)) return;
		else if(l.val == set) {
		    Solver.fail(s);
		    return;
		}
		else {
		    s.arg1 = l.val;
		    s.arg2 = set;
		    neq(s);
		    return;
		}
	    }
	}
	else if(l.equ == null) {
	    s.arg1 = l;
	    s.arg2 = set.equ;
	    neq(s);
	    return;
	}
	else if(set.equ == null) {
	    s.arg1 = l.equ;
	    s.arg2 = set;
	    neq(s);
	    return;
	}
	else {
	    s.arg1 = l.equ;
	    s.arg2 = set.equ;
	    neq(s);
	    return;
	}
    }

    protected void neqLvarMultiSet(Lvar l, MultiSet mset, Constraint s) throws Failure {
	if(l.equ == null && mset.equ == null) {
	    if(!l.iniz || !mset.iniz)return;
	    else {
		if(l.val instanceof MultiSet && ((MultiSet)l.val).isEmpty() && mset.isEmpty()) {
		    Solver.fail(s);
		    return;
		}
		else if(mset.isEmpty() || mset.occurs(l) || !(l.val instanceof MultiSet)) return;
		else if(l.val == mset) {
		    Solver.fail(s);
		    return;
		}
		else {
		    s.arg1 = l.val;
		    s.arg2 = mset;
		    neq(s);
		    return;
		}
	    }
	}
	else if(l.equ == null) {
	    s.arg1 = l;
	    s.arg2 = mset.equ;
	    neq(s);
	    return;
	}
	else if(mset.equ == null) {
	    s.arg1 = l.equ;
	    s.arg2 = mset;
	    neq(s);
	    return;
	}
	else {
	    s.arg1 = l.equ;
	    s.arg2 = mset.equ;
	    neq(s);
	    return;
	}
    }


    /* x.neq(y) con x e y enrambe istanze di Set. */
    protected void neqSet(Set set1, Set set2, Constraint s) throws Failure {
	if(set1.equ == null && set2.equ == null) {
	    if(set1 == set2) {                                    //  (3)
		Solver.fail(s);
		return;
	    }
	    else if(set1.isInt) {
		s.arg1 = set2;
		s.arg2 = (Int)set1.inter;
		neq(s);
		return;
	    }
	    else if(set2.isInt) {
		s.arg2 = (Int)set2.inter;
		neq(s);
		return;
	    }                                                    //  (7)
	    else if(set2.iniz && set2.ultimoResto() == set1 && !set1.iniz) { // X.neq( {t1, t2, .., tn|X})
		if(set2.occurs(set1)) {
		    s.bool = true;
		    return;
		}
		else {
		    if (s.caseControl < 2) s.caseControl = 2;
		    switch(s.caseControl) {
		    case 2:                                       //  (7) (i)
			if(set2.size() > 0) {
			    VarState statoVarNeq2 = Solver.B.getVarState();
			    StoreState statoStoreNeq2 = Solver.B.getStoreState(s);
			    Solver.B.addChoicePoint(3, statoVarNeq2, statoStoreNeq2);
			}
			s.arg1 = set2.first();                     // t1.nin(X)
			s.arg2 = set1;
			s.cons = Enviroment.ninCode;
			s.caseControl = 0;
			Solver.storeInvariato = false;
			return;
		    case 3:                                       //  (7)
			s.arg1 = set1;
			s.arg2 = set2.sub();                       // X.neq( {t2, .., tn|X})
			s.cons = Enviroment.neqCode;
			s.caseControl = 2;
			neq(s);
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }                                                    //  (7)
	    else if(set1.iniz && set1.ultimoResto() == set2 && !set2.iniz) { // {t1, t2, .., tn|X}.neq(X)
		if(set1.occurs(set2)) return;
		else {
		    if(s.caseControl < 4)s.caseControl = 4;
		    switch(s.caseControl) {
		    case 4:                                       //  (7) (i)
			if(set1.size() > 0) {
			    VarState statoVarNeq2 = Solver.B.getVarState();
			    StoreState statoStoreNeq2 = Solver.B.getStoreState(s);
			    Solver.B.addChoicePoint(3, statoVarNeq2, statoStoreNeq2);
			}
			s.arg1 = set1.first();                     //t1.nin(X)
			s.arg2 = set2;
			s.cons = Enviroment.ninCode;
			s.caseControl = 0;
			Solver.storeInvariato = false;
			return;
		    case 5:                                       //  (7)
			s.arg1 = set1.sub();                       // {t2, .., tn|X}.neq(X)
			s.arg2 = set2;
			s.cons = Enviroment.neqCode;
			s.caseControl = 4;
			neq(s);
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }
	    else if(!set1.iniz || !set2.iniz) return;
	    else {                                           // entrambi inizializzati  (8)
		if(set1 == Set.empty && set2 == Set.empty) {
		    Solver.fail(s);
		    return;
		}
		else if(set1 == Set.empty || set2 == Set.empty) {
		    s.bool = true;
		    return;
		}
		else if(set1.occurs(set2) || set2.occurs(set1)) {
		    s.bool = true;
		    return;
		}
		else {
		    switch(s.caseControl) {
		    case 0:                                       // (8) (i)
			VarState statoVarNeq = Solver.B.getVarState();
			StoreState statoStoreNeq = Solver.B.getStoreState(s);
			Solver.B.addChoicePoint(1, statoVarNeq, statoStoreNeq);
			Lvar n = new Lvar();
			s.arg1 = n;                                // n.in(set1)
			s.arg2 = set1;
			s.cons = Enviroment.inCode;
			Constraint se0 = new Constraint(n, Enviroment.ninCode, set2); // n.nin(set2)
			Solver.add(se0);
			Solver.storeInvariato = false;
			return;
		    case 1:                                       // (8) (ii)
			Lvar m = new Lvar();
			s.arg1 = m;                                // n.nin(set1)
			s.arg2 = set1;
			s.cons = Enviroment.ninCode;
			s.caseControl = 0;
			Constraint se1 = new Constraint(m, Enviroment.inCode, set2); // n.in(set2)
			Solver.add(se1);
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }
	}
	else if(set1.equ == null) {
	    s.arg2 = set2.equ;
	    neq(s);
	    return;
	}
	else if(set2.equ == null) {
	    s.arg1 = set1.equ;
	    neq(s);
	    return;
	}
	else {
	    s.arg1 = set1.equ;
	    s.arg2 = set2.equ;
	    neq(s);
	    return;
	}
    }


    /* x.neq(y) con x e y enrambe istanze di MultiSet. */
    protected void neqMultiSet(MultiSet mset1, MultiSet mset2, Constraint s) throws Failure {
	if(mset1.equ == null && mset2.equ == null) {
	    if(mset1 == mset2) {
		Solver.fail(s);
		return;
	    }
            else if(mset2.iniz && !mset1.iniz) { // Caso 7:  X.neq( f(t1, t2, .., tn))
                // il controllo che X.in({variabili libere di f(t1, t2, .., tn)})  implicito
                // OSS. il controllo che non esistano vars quantificate usate esternamente ai
                // quantificatori viene fatto all'inizio della solve()
                if(mset1.occurs(mset2)) {
		    s.bool = true;
		    return;
		}
            }
            else if(mset1.iniz && !mset2.iniz) { // Caso 7:  X.neq( f(t1, t2, .., tn))
                // il controllo che X.in({variabili libere di f(t1, t2, .., tn)})  implicito
                // OSS. il controllo che non esistano vars quantificate usate esternamente ai
                // quantificatori viene fatto all'inizio della solve()
                if(mset2.occurs(mset1)) {
		    s.bool = true;
		    return;
		}
            }
	    else if(!mset1.iniz && !mset2.iniz) return;
//            else if(!mset1.iniz && !mset2.iniz && mset1.id == mset2.id) {
//                Solver.fail(s);
//                return;
//            }
	    else {
                 // Caso 6.1
		if(mset1 == MultiSet.empty && mset2 == MultiSet.empty) {
		    Solver.fail(s);
		    return;
		}
		else if(mset1 == MultiSet.empty || mset2 == MultiSet.empty) {
		    s.bool = true;
		    return;
		}
		else if(mset1.occurs(mset2) || mset2.occurs(mset1)) {
		    s.bool = true;
		    return;
		}
                // Gestione del caso 6.1   {t1|s1} != {t2|s2}  dove tail(s1) and tail(s2)
                // sono esattamente la stessa variabile non inizializzata
                else if( (mset1.ultimoResto() == mset2.ultimoResto()) &&
                         ((MultiSet)mset1.ultimoResto()) != null &&
                         ((MultiSet)mset1.ultimoResto()) != MultiSet.empty &&
                         !((MultiSet)mset1.ultimoResto()).iniz) {

                    MultiSet w = mset1.notultimoResto();
                    MultiSet z = mset2.notultimoResto();

                    Solver.add(Solver.indexOf(s), new Constraint(w, Enviroment.neqCode, z));
                    s.bool = true;
                    Solver.storeInvariato = false;
                    return;
                }
		else {  // {{t1|s1}}!={{t2|s2}} && s1!= s2
                    ((MultiSet)s.arg1).output();
                    ((MultiSet)s.arg2).output();
		    switch(s.caseControl) {
		    case 0:                                       // ex(8i). 6.2i t1!=t2 && t1 nin Mset2
			VarState statoVarNeq = Solver.B.getVarState();
			StoreState statoStoreNeq = Solver.B.getStoreState(s);
			Solver.B.addChoicePoint(1, statoVarNeq, statoStoreNeq);
			// Lvar n = new Lvar();
			s.arg1 = mset1.first();                 // n.in(set1)
			s.arg2 = mset2.first();
			s.cons = Enviroment.neqCode;
			Constraint se0 = new Constraint(s.arg1, Enviroment.ninCode, mset2.sub()); // n.nin(set2)
			Solver.add(se0);
			Solver.storeInvariato = false;
                        Solver.showStore();
			return;
		    case 1:                                       // (8) (ii)
			MultiSet N = new MultiSet();
			s.arg1 = N.ins(mset1.first());             // {{t1|N}}.eq(mset2)
			s.arg2 = mset2;
			s.cons = Enviroment.eqCode;
			s.caseControl = 0;
			Constraint se1 = new Constraint(mset1.sub(), Enviroment.neqCode, N); // s1.neq(N)
			Solver.add(se1);
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }
	}
	else if(mset1.equ == null) {
	    s.arg2 = mset2.equ;
	    neq(s);
	    return;
	}
	else if(mset2.equ == null) {
	    s.arg1 = mset1.equ;
	    neq(s);
	    return;
	}
	else {
	    s.arg1 = mset1.equ;
	    s.arg2 = mset2.equ;
	    neq(s);
	    return;
	}
    }



    /*  Disuguaglianza tra un insieme e un intervallo */
    protected void neqSetInt(Set set, Int inter, Constraint s) throws Failure {
	if(set.equ == null) {
	    switch(s.caseControl) {
	    case 0:
		if(!set.iniz) {
		    s.caseControl = 1;
		    neqSetInt(set, inter, s);
		    return;
		}
		else if(set.isInt) {
		    s.arg1 = (Int)set.inter;
		    s.arg2 = inter;
		    neqObj((Int)set.inter, inter, s);
		    return;
		}
		else if(set == Set.empty) {
		    if(((Integer)inter.sx).intValue() > ((Integer)inter.dx).intValue()) {
			Solver.fail(s);
			return;
		    }
		    else {
			s.bool = true;
			return;
		    }
		}
		else {
		    s.caseControl = 4;
		    neqSetInt(set, inter, s);
		    return;
		}
	    case 1:
		VarState statoVarNeq1 = Solver.B.getVarState();
		StoreState statoStoreNeq1 = Solver.B.getStoreState(s);
		Solver.B.addChoicePoint(2, statoVarNeq1, statoStoreNeq1);
		s.arg1 = (Integer)inter.sx;                     //t1.nin(X)
		s.arg2 = set;
		s.cons = Enviroment.ninCode;
		s.caseControl = 0;
		Solver.storeInvariato = false;
		return;
	    case 2:
		VarState statoVarNeq2 = Solver.B.getVarState();
		StoreState statoStoreNeq2 = Solver.B.getStoreState(s);
		Solver.B.addChoicePoint(3, statoVarNeq2, statoStoreNeq2);
		s.arg1 = (Integer)inter.dx;                     //t2.nin(X)
		s.arg2 = set;
		s.cons = Enviroment.ninCode;
		s.caseControl = 0;
		Solver.storeInvariato = false;
		return;
	    case 3:
		Set N = new Set();
		s.arg1 = set;                                   // X.eq( {t1, t2|N})
		s.arg2 = N.ins((Integer)inter.dx).ins((Integer)inter.sx);
		s.cons = Enviroment.eqCode;
		s.caseControl = 0;
		Int interNew = new Int(((Integer)inter.sx).intValue()+1, ((Integer)inter.dx).intValue()-1);
		Constraint s1 = new Constraint((Integer)inter.sx, Enviroment.ninCode, N);  // t1.nin(N)
		Constraint s2 = new Constraint((Integer)inter.dx, Enviroment.ninCode, N);  // t2.nin(N)
		Constraint s3 = new Constraint(N, Enviroment.neqCode, interNew);           // N.neq(int(t1+1, t2-1))
		Solver.add(s1);
		Solver.add(s2);
		Solver.add(s3);
		return;
	    case 4:
		VarState statoVarNeq4 = Solver.B.getVarState();
		StoreState statoStoreNeq4 = Solver.B.getStoreState(s);
		Solver.B.addChoicePoint(5, statoVarNeq4, statoStoreNeq4);
		s.arg1 = set.first();                                // t3.nin(int(t1, t2))
		s.arg2 = inter;
		s.cons = Enviroment.ninCode;
		s.caseControl = 0;
		Solver.storeInvariato = false;
		return;
	    case 5:
		VarState statoVarNeq5 = Solver.B.getVarState();
		StoreState statoStoreNeq5 = Solver.B.getStoreState(s);
		Solver.B.addChoicePoint(6, statoVarNeq5, statoStoreNeq5);
		s.arg1 = set.first();                                // t3.in(int(t1, t2))
		s.arg2 = inter;
		s.cons = Enviroment.inCode;
		s.caseControl = 0;
		Constraint s4 = new Constraint(set.sub(), Enviroment.neqCode, inter);     // S.neq(int(t1, t2))
		Solver.add(s4);
		Solver.storeInvariato = false;
		return;
	    case 6:
		VarState statoVarNeq6 = Solver.B.getVarState();
		StoreState statoStoreNeq6 = Solver.B.getStoreState(s);
		Solver.B.addChoicePoint(7, statoVarNeq6, statoStoreNeq6);
		s.arg1 = (Integer)inter.sx;                          // t1.nin( {t3|S})
		s.arg2 = set;
		s.cons = Enviroment.ninCode;
		s.caseControl = 0;
		Solver.storeInvariato = false;
		return;
	    case 7:
		Set M = new Set();
		Int I = new Int(((Integer)inter.sx).intValue()+1, (Integer)inter.dx);
		s.arg1 = set;                                        // {t3|S}.eq( {t1|M})
		s.arg2 = ((Set)M.ins((Integer)inter.sx));
		s.cons = Enviroment.ninCode;
		s.caseControl = 0;
		Constraint s5 = new Constraint((Integer)inter.sx, Enviroment.ninCode, M); // t1.nin(M)
		Constraint s6 = new Constraint(M, Enviroment.neqCode, I);                 // M.neq(int(t1+1, t2))
		Solver.add(s5);
		Solver.add(s6);
		Solver.storeInvariato = false;
		return;
	    }
	}
	else {
	    s.arg1 = set.equ;
	    s.arg2 = inter;
	    neqSetInt(set.equ, inter, s);
	    return;
	}
    }


    /* x.neq(y) con x istanza di Lvar e y oggetto diverso da Lvar, Lst e Set:
       se x S inizializzata, viene confrontato il valore di x con y, altrimenti
       resta memorizzato il vincolo in attesa di nuove informazioni su x. */
    protected void neqLvarObj(Lvar l, Object o, Constraint s) throws Failure {
	if(l.equ == null) {
	    if(l.iniz) {
		if(!l.val.equals(o)) {
		    s.bool = true;
		    return;
		}
		else {
		    Solver.fail(s);
		    return;
		}
	    }
	    else return;
	}
	else {
	    s.arg1 = l.equ;
	    s.arg2 = o;
	    neq(s);
	    return;
	}
    }


    protected void neqLstObj(Lst l, Object o, Constraint s) {
	s.bool = true;
	return;
    }


    protected void neqSetObj(Set set, Object o, Constraint s) {
	s.bool = true;
	return;
    }


    protected void neqMultiSetObj(MultiSet mset, Object o, Constraint s) {
	s.bool = true;
	return;
    }


    protected void neqObj(Object arg1, Object arg2, Constraint s) throws Failure {
	if(arg1.equals(arg2)) Solver.fail(s);
	else s.bool = true;
	return;
    }


    /* Appartenenza di una Lvar ad un Set */
    protected void inLvarSet(Lvar lvar, Set set, Constraint s) throws Failure {
	if(lvar.equ == null && set.equ == null) {
	    if(set.isInt) {
		s.arg2 = (Int)set.inter;
		inLvarInt(lvar, (Int)set.inter, s);
		return;
	    }
	    else if(set.iniz) {                      // (1)
		if(set == Set.empty) {
		    Solver.fail(s);
		    return;
		}
		else {                                // X.in( {t|R})
		    switch(s.caseControl) {
		    case 0:                           // (2) (i)
			VarState statoVarIn = Solver.B.getVarState();
			StoreState statoStoreIn = Solver.B.getStoreState(s);
			if(set.size() > 1 || !set.ultimoResto().iniz)
			    Solver.B.addChoicePoint(1, statoVarIn, statoStoreIn);
			s.arg1 = lvar;
			s.cons = Enviroment.eqCode;
			s.arg2 = set.first();          // X.eq(t)
			eq(s);
			Solver.storeInvariato = false;
			return;
		    case 1:                           // (2) (ii)
			s.caseControl = 0;
			s.arg2 = (Set)set.sub();       // X.in(R)
			in(s);
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }
	    else {                                   // set non iniz. (3)
		Set n = new Set();
		s.arg1 = set;
		s.arg2 = n.ins(lvar);               // set.eq( {X|N})  N Set non iniz.
		s.cons = Enviroment.eqCode;
		s.caseControl = 0;
		eq(s);
		Solver.storeInvariato = false;
		return;
	    }
	}
	else if(lvar.equ == null) {
	    s.arg2 = set.equ;
	    in(s);
	    return;
	}
	else if(set.equ == null) {
	    s.arg1 = lvar.equ;
	    in(s);
	    return;
	}
	else {
	    s.arg1 = lvar.equ;
	    s.arg2 = set.equ;
	    in(s);
	    return;
	}
    }


    /* Appartenenza di una Lvar ad un intervallo */
    protected void inLvarInt(Lvar lvar, Int inter, Constraint s) throws Failure {
	if(lvar.equ == null) {
	    if(((Integer)inter.sx).intValue() > ((Integer)inter.dx).intValue()) {
		Solver.fail(s);
		return;
	    }
	    else {
		if(lvar.iniz) {
		    s.arg1 = lvar.val;
		    inObjInt(lvar.val, inter, s);
		    return;
		}
		else {                                // lvar non iniz.
		    switch(s.caseControl) {
		    case 0:
			VarState statoVarIn = Solver.B.getVarState();
			StoreState statoStoreIn = Solver.B.getStoreState(s);
			Solver.B.addChoicePoint(1, statoVarIn, statoStoreIn);
			s.arg1 = lvar;
			s.cons = Enviroment.eqCode;
			s.arg2 = inter.sx;             // X.eq(sx)
			eq(s);
			Solver.storeInvariato = false;
			return;
		    case 1:
			Int interNew = new Int(((Integer)inter.sx).intValue()+1, (Integer)inter.dx);
			s.caseControl = 0;
			s.arg2 = interNew;
			in(s);
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }
	}
	else {
	    s.arg1 = lvar.equ;
	    inLvarInt(lvar.equ, inter, s);
	    return;
	}
    }


    /*   Appartenenza di una Lvar ad una Lvar */
    protected void inLvarLvar(Lvar l1, Lvar l2, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(l2.iniz) {
		if(l2.val instanceof Set) {
		    s.arg2 = (Set)l2.val;
		    Solver.storeInvariato = false;
		    return;
		}
                else if(l2.val instanceof MultiSet) { 
		    s.arg2 = (MultiSet)l2.val;
		    Solver.storeInvariato = false;
		    return;
		}                
		else {
		    Solver.fail(s);
		    return;
		}
	    }
	    else {
		Set n = new Set();
		l2.iniz = true;
		l2.val = n.ins(l1);
		s.arg2 = l2.val;
		Solver.storeInvariato = false;
		return;
	    }
	}
	else if(l1.equ == null) {
	    s.arg2 = l2.equ;
	    in(s);
	    return;
	}
	else if(l2.equ == null) {
	    s.arg1 = l1.equ;
	    in(s);
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    s.arg2 = l2.equ;
	    in(s);
	    return;
	}
    }

   protected void inLvarMultiSet(Lvar lvar, MultiSet multiset, Constraint s) throws Failure {
        if(lvar.equ == null && multiset.equ == null) {
          //mi chiedo se il Multiset  inizializzato
                if(multiset.iniz){
                  if(multiset == MultiSet.empty) {      //se il Multiset  vuoto fallisco
		    Solver.fail(s);
		    return;
		}
		 else {                                // X.in( {t|R})
		    switch(s.caseControl) {
		    case 0:                           // (2) (i)
			VarState statoVarIn = Solver.B.getVarState();
			StoreState statoStoreIn = Solver.B.getStoreState(s);
			if(multiset.size() > 1 || !multiset.ultimoResto().iniz)
			    Solver.B.addChoicePoint(1, statoVarIn, statoStoreIn);
			s.arg1 = lvar;
			s.cons = Enviroment.eqCode;
			s.arg2 = multiset.first();          // X.eq(t)
			eq(s);
			Solver.storeInvariato = false;
			return;
		    case 1:                           // (2) (ii)
			s.caseControl = 0;
			s.arg2 = (MultiSet)multiset.sub();       // X.in(R)
			in(s);
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }
	    else {                                   // multiset non iniz. (3)
		MultiSet n = new MultiSet();
		s.arg1 = multiset;
		s.arg2 = n.ins(lvar);               // multiset.eq( {X|N})  N MultiSet non iniz.
		s.cons = Enviroment.eqCode;
		s.caseControl = 0;
		eq(s);
		Solver.storeInvariato = false;
		return;
	    }
	}
        // Se il campo equ  diverso da null allora devo risolvere la catena di riferimenti
        // tra gli obj per entrambi i parametri.
        // x = y
        // y= z
        // z = w
        // w = 3
        // x in {1,2,3}
        // quando tratti l'in trovi la catena tra le Lvar, con questo ciclo la risolvi
	else if(lvar.equ == null) {
	    s.arg2 = multiset.equ;
            // Richiamo la procedura che studia i tipi di obj passati a in
	    in(s);
	    return;
	}
	else if(multiset.equ == null) {
	    s.arg1 = lvar.equ;
	    in(s);
	    return;
	}
	else {
	    s.arg1 = lvar.equ;
	    s.arg2 = multiset.equ;
	    in(s);
	    return;
	}

    }
   
   
   protected void inLvarMultiSetAR(Lvar lvar, MultiSet multiset, Lvar arL, Constraint s) throws Failure {
       if(lvar.equ == null && multiset.equ == null) {
           //mi chiedo se il Multiset  inizializzato
           if(multiset.iniz){
               if(multiset == MultiSet.empty) {      //se il Multiset  vuoto fallisco
                   Solver.fail(s);
                   return;
               }
               else {                                // X.in( {t|R})
                   switch(s.caseControl) {
                       case 0:                           // (2) (i)
                           VarState statoVarIn = Solver.B.getVarState();
                           StoreState statoStoreIn = Solver.B.getStoreState(s);
                           if(multiset.size() > 1 || !multiset.ultimoResto().iniz)
                               Solver.B.addChoicePoint(1, statoVarIn, statoStoreIn);
                           s.arg1 = lvar;
                           s.cons = Enviroment.eqCode;
                           s.arg2 = multiset.first();          // X.eq(t)
                           int c = 0; // numero di volte che lvar compare in multiset
                           Lvar count = new Lvar(c);
                           Constraint st2 = new Constraint(arL, Enviroment.eqCode, count);
                           eq(s);
                           Solver.storeInvariato = false;
                           return;
                       case 1:                           // (2) (ii)
                           s.caseControl = 0;
                           s.arg2 = (MultiSet)multiset.sub();       // X.in(R)
                           in(s);
                           Solver.storeInvariato = false;
                           return;
                   }
               }
           }
           else {                                   // multiset non iniz. (3)
               MultiSet n = new MultiSet();
               s.arg1 = multiset;
               s.arg2 = n.ins(lvar);               // multiset.eq( {X|N})  N MultiSet non iniz.
               s.cons = Enviroment.eqCode;
               s.caseControl = 0;
               eq(s);
               Solver.storeInvariato = false;
               return;
           }
       }
       // Se il campo equ  diverso da null allora devo risolvere la catena di riferimenti
       // tra gli obj per entrambi i parametri.
       // x = y
       // y= z
       // z = w
       // w = 3
       // x in {1,2,3}
       // quando tratti l'in trovi la catena tra le Lvar, con questo ciclo la risolvi
       else if(lvar.equ == null) {
           s.arg2 = multiset.equ;
           // Richiamo la procedura che studia i tipi di obj passati a in
           in(s);
           return;
       }
       else if(multiset.equ == null) {
           s.arg1 = lvar.equ;
           in(s);
           return;
       }
       else {
           s.arg1 = lvar.equ;
           s.arg2 = multiset.equ;
           in(s);
           return;
       }
   }
   


    /* Appartenenza di un oggetto diverso da Lvar, Lst e Set ad un Set */
    protected void inObjSet(Object obj, Set set, Constraint s) throws Failure {
	if(set.equ == null) {
	    if(set.isInt) {
		s.arg2 = (Int)set.inter;
		inObjInt(obj, (Int)set.inter, s);
		return;
	    }
	    else if(set.iniz) {
		if(set == Set.empty) {
		    Solver.fail(s);
		    return;
		}
		else {
		    switch(s.caseControl) {
		    case 0:                                         // (2) (i)
			VarState statoVarIn = Solver.B.getVarState();
			StoreState statoStoreIn = Solver.B.getStoreState(s);
			if(set.size() > 1 || !set.ultimoResto().iniz)
			    Solver.B.addChoicePoint(1, statoVarIn, statoStoreIn);
			s.arg1 = obj;
			s.cons = Enviroment.eqCode;
			s.arg2 = set.first();
			eq(s);
			Solver.storeInvariato = false;
			return;
		    case 1:                                         // (2) (ii)
			s.caseControl = 0;
			s.arg2 = (Set)set.sub();
			in(s);
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }
	    else {
		Set n = new Set();
		s.arg1 = set;
		s.cons = Enviroment.eqCode;
		s.arg2 = n.ins(obj);
		eq(s);
		Solver.storeInvariato = false;
		return;
	    }
	}
	else {
	    s.arg2 = set.equ;
	    in(s);
	    return;
	}
    }


    /* Appartenenza di un oggetto diverso da Lvar, Lst e Set ad una Lvar */
    protected void inObjLvar(Object obj, Lvar l, Constraint s) throws Failure {
	if(l.equ == null) {
	    if(l.iniz) {
		if(l.val instanceof Set) {
		    s.arg2 = (Set)l.val;
		    inObjSet(obj, (Set)l.val, s);
		    return;
		}
                else if(l.val instanceof MultiSet) {
		    s.arg2 = (MultiSet)l.val;
		    inObjMultiSet(obj, (MultiSet)l.val, s);
		    return;
		}                
		else {
		    Solver.fail(s);
		    return;
		}
	    }
	    else {
		Set n = new Set();
		s.arg1 = l;
		s.cons = Enviroment.eqCode;
		s.arg2 = n.ins(obj);
		eq(s);
		Solver.storeInvariato = false;
		return;
	    }
	}
	else {
	    s.arg2 = l.equ;
	    in(s);
	    return;
	}
    }

    /* Appartenenza di un oggetto diverso da Lvar, Lst e Set ad un MultiSet */
    protected void inObjMultiSet(Object obj, MultiSet multiset, Constraint s) throws Failure {
	if(multiset.equ == null) {
	     if(multiset.iniz) {
		if(multiset == MultiSet.empty) {
		    Solver.fail(s);
		    return;
		}
		else {
		    switch(s.caseControl) {
		    case 0:                                         // (2) (i)
			VarState statoVarIn = Solver.B.getVarState();
			StoreState statoStoreIn = Solver.B.getStoreState(s);
			if(multiset.size() > 1 || !multiset.ultimoResto().iniz)
			    Solver.B.addChoicePoint(1, statoVarIn, statoStoreIn);
			s.arg1 = obj;
			s.cons = Enviroment.eqCode;
			s.arg2 = multiset.first();
			eq(s);
			Solver.storeInvariato = false;
			return;
		    case 1:                                         // (2) (ii)
			s.caseControl = 0;
			s.arg2 = (MultiSet)multiset.sub();
			in(s);
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }
	    else {
		MultiSet n = new MultiSet();
		s.arg1 = multiset;
		s.cons = Enviroment.eqCode;
		s.arg2 = n.ins(obj);
		eq(s);
		Solver.storeInvariato = false;
		return;
	    }
	}
	else {
	    s.arg2 = multiset.equ;
	    in(s);
	    return;
	}
    }


    /* Appartenenza di un oggetto diverso da Lvar, Lst e Set ad un Internallo */
    protected void inObjInt(Object o, Int inter, Constraint s) throws Failure {
	if(((Integer)inter.sx).intValue() > ((Integer)inter.dx).intValue()) {
	    Solver.fail(s);
	    return;
	}
	else {
	    if(o instanceof Integer && ((Integer)inter.sx).intValue() <= ((Integer)o).intValue()
	       && ((Integer)o).intValue() <= ((Integer)inter.dx).intValue()) {
		s.bool = true;
		return;
	    }
	    else {
		Solver.fail(s);
		return;
	    }
	}
    }


    protected void select(Lvar lvar, Set set, Set rest, Constraint s) throws Failure {
	if(lvar.equ == null && set.equ == null && rest.equ == null) {
	    if(set.size() == 0) {
		Solver.fail(s);
		return;
	    }
	    else {
		switch(s.caseControl) {
		case 0:
		    /* Le prossime due righe servono per memorizzare le Lvar non inizializzate
		       e lo stato del constraint store (i vincoli).

		    */
		    VarState statoVarIn = Solver.B.getVarState();
		    StoreState statoStoreIn = Solver.B.getStoreState(s);
		    if(set.size() > 1)
			Solver.B.addChoicePoint(1, statoVarIn, statoStoreIn);
		    s.arg2 = set.first();                                // lvar.eq(set.first())
		    s.arg3 = null;
		    s.cons = Enviroment.eqCode;
		    eq(s);
		    Constraint s0 = new Constraint(rest, Enviroment.eqCode, (Set)set.sub()); // rest.eq(set.sub())
		    Solver.add(Solver.indexOf(s)+1, s0);
		    Solver.storeInvariato = false;
		    return;
		case 1:
		    Set v = new Set();
		    s.caseControl = 0;
		    s.arg2 = set.sub();
		    s.arg3 = v;
		    Constraint s1 = new Constraint(rest, Enviroment.eqCode, v.ins(set.first()));
		    Solver.add(Solver.indexOf(s), s1);
		    Solver.storeInvariato = false;
		    return;
		}
	    }
	}
	else if(lvar.equ != null) {
	    s.arg1 = lvar.equ;
	    select(s);
	    return;
	}
	else if(set.equ != null) {
	    s.arg2 = set.equ;
	    select(s);
	    return;
	}
	else {
	    s.arg3 = rest.equ;
	    select(s);
	    return;
	}
    }


    /*  Non appartenenza di una Lvar ad un Set */
    protected void ninLvarSet(Lvar lvar, Set set, Constraint s)throws Failure {
	if(lvar.equ == null && set.equ == null) {
	    if(set.iniz && set.isEmpty()) {                              // (1)
		s.bool = true;
		return;
	    }
	    else if(set.isInt) {
		s.arg2 = (Int)set.inter;
		ninLvarInt(lvar, (Int)set.inter, s);
		return;
	    }
	    else {
		if(!set.iniz)return;                                   // (3)
		else {
		    if(lvar.iniz && lvar.val instanceof Set) {
			s.arg1 = (Set)lvar.val;
			Solver.storeInvariato = false;
			return;
		    }
		    else {                                             // (2)
			s.arg2 = set.sub();
			Constraint st = new Constraint(lvar, Enviroment.neqCode, set.first());
			Solver.add(st);
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }
	}
	else if(lvar.equ == null) {
	    s.arg2 = set.equ;
	    nin(s);
	    return;
	}
	else if(set.equ == null) {
	    s.arg1 = lvar.equ;
	    nin(s);
	    return;
	}
	else {
	    s.arg1 = lvar.equ;
	    s.arg2 = set.equ;
	    nin(s);
	    return;
	}
    }


    /* Non appartenenza di una Lvar ad una Lvar */
    protected void ninLvarLvar(Lvar l1, Lvar l2, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(l2.val instanceof Set) {
		s.arg2 = (Set)l2.val;
		Solver.storeInvariato = false;
		return;
	    }
	    else return;
	}
	else if(l1.equ == null) {
	    s.arg2 = l2.equ;
	    nin(s);
	    return;
	}
	else if(l2.equ == null) {
	    s.arg1 = l1.equ;
	    nin(s);
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    s.arg2 = l2.equ;
	    nin(s);
	    return;
	}
    }


    /*  Non appartenenza di una Lvar ad un MultiSet */
    protected void ninLvarMultiSet(Lvar lvar, MultiSet mset, Constraint s)throws Failure {
	if(lvar.equ == null && mset.equ == null) {
	    if(mset.iniz && mset.isEmpty()) {                              // (1)
		s.bool = true;
		return;
	    }
	    else {
		if(!mset.iniz)return;                                   // (3)    r.nin{[X]}
		else {
		    if(lvar.iniz && lvar.val instanceof MultiSet) {
			s.arg1 = (MultiSet)lvar.val;
			Solver.storeInvariato = false;
			return;
		    }
		    else {                                             // (2)     r.nin{[t|s]}
			s.arg2 = mset.sub();
			Constraint st = new Constraint(lvar, Enviroment.neqCode, mset.first());  //r.neq(t)
			Constraint st2 = new Constraint(lvar, Enviroment.ninCode, s.arg2);         //r.nin(s)
                        Solver.add(st);
                        Solver.add(st2);
			Solver.storeInvariato = false;
			return;
		    }
		}
	    }
	}
	else if(lvar.equ == null) {
	    s.arg2 = mset.equ;
	    nin(s);
	    return;
	}
	else if(mset.equ == null) {
	    s.arg1 = lvar.equ;
	    nin(s);
	    return;
	}
	else {
	    s.arg1 = lvar.equ;
	    s.arg2 = mset.equ;
	    nin(s);
	    return;
	}
    }


    /* Non appartenenza di una Lvar ad un intervallo */
    protected void ninLvarInt(Lvar lvar, Int inter, Constraint s) throws Failure {
	if(lvar.equ == null) {
	    if(((Integer)inter.sx).intValue() > ((Integer)inter.dx).intValue()) {
		s.bool = true;
		return;
	    }
	    else {
		if(lvar.iniz) {
		    s.arg1 = lvar.val;
		    ninObjInt(lvar.val, inter, s);
		    return;
		}
		else return;
	    }
	}
	else {
	    s.arg1 = lvar.equ;
	    nin(s);
	    return;
	}
    }


    /* Non appartenenza di un oggetto diverso da Lvar Lst o Set ad un Set */
    protected void ninObjSet(Object obj, Set set, Constraint s) throws Failure {
	if(set.equ == null) {
	    if(!set.iniz)return;
	    else if(set.isInt) {
		s.arg2 = (Int)set.inter;
		ninObjInt(obj, (Int)set.inter, s);
		return;
	    }
	    else {
		if(set.iniz && set.isEmpty()) {
		    s.bool = true;
		    return;
		}
		else {
		    s.arg2 = set.sub();
		    Constraint st = new Constraint(obj, Enviroment.neqCode, set.first());
		    Solver.add(st);
		    Solver.storeInvariato = false;
		    return;
		}
	    }
	}
	else {
	    s.arg2 = set.equ;
	    nin(s);
	    return;
	}
    }


   /* Non appartenenza di un oggetto diverso da Lvar, Lst, Set, MultiSet ad un MultiSet */
    protected void ninObjMultiSet(Object obj, MultiSet mset, Constraint s) throws Failure {
	if(mset.equ == null) {
	    if(!mset.iniz) return;
	    else {
		if(mset.iniz && mset.isEmpty()) {
		    s.bool = true;
		    return;
		}
		else {
		    s.arg2 = mset.sub();
		    Constraint st = new Constraint(obj, Enviroment.neqCode, mset.first());
		    Solver.add(st);
		    Solver.storeInvariato = false;
		    return;
		}
	    }
	}
	else {
	    s.arg2 = mset.equ;
	    nin(s);
	    return;
	}
    }


    /* Non appartenenza di un oggetto diverso da Lvar Lst o Set ad una Lvar */
    protected void ninObjLvar(Object obj, Lvar l, Constraint s)throws Failure {
	if(l.equ == null) {
	    if(l.iniz) {
		if(l.val instanceof Set) {
		    s.arg2 = (Set)l.val;
		    ninObjSet(obj, (Set)l.val, s);
		    return;
		}
		else return;
	    }
	    else return;
	}
	else {
	    s.arg2 = l.equ;
	    nin(s);
	    return;
	}
    }


    /*  Non appartenenza di un oggetto diverso da Lvar Lst o Set ad un Intervallo */
    protected void ninObjInt(Object o, Int inter, Constraint s) throws Failure {
	if(((Integer)inter.sx).intValue() > ((Integer)inter.dx).intValue()) {
	    s.bool = true;
	    return;
	}
	else {
	    if(o instanceof Integer) {
		if(((Integer)inter.sx).intValue() > ((Integer)o).intValue() ||
		   ((Integer)o).intValue() > ((Integer)inter.dx).intValue()) {
		    s.bool = true;
		    return;
		}
		else {
		    Solver.fail(s);
		    return;
		}
	    }
	    else {
		s.bool = true;
		return;
	    }
	}
    }


    /* Disgiunzione tra due Set */
    protected void disj(Set set1, Set set2, Constraint s) throws Failure {
	if(set1.equ == null && set2.equ == null) {
	    if(!set1.iniz && !set2.iniz) {
		return;
	    }
	    else if(set1 == Set.empty || set2 == Set.empty) {                // (1)
		s.bool = true;
		return;
	    }
	    else if(set1 == set2) {                                          // (2)
		s.arg1 = set1;
		s.arg2 = Set.empty;
		s.arg3 = null;
		s.cons = Enviroment.eqCode;
		eq(s);
		return;
	    }
	    else if(set1.iniz && !set2.iniz) { //  {t1|s1} || X              // (3)
		s.arg1 = set1.sub();                                         // s1||X
		Constraint st = new Constraint(set1.first(), Enviroment.ninCode, set2);         // t1.nin(X)
		Solver.add(st);
		Solver.storeInvariato = false;
	    }
	    else if(!set1.iniz && set2.iniz) { //  X || {t2|s2}              // (3)
		s.arg2 = set2.sub();                                         // X||s2
		Constraint se = new Constraint(set2.first(), Enviroment.ninCode, set1);         // t2.nin(X)
		Solver.add(se);
		Solver.storeInvariato = false;
	    }
	    else {                                                           //  {t1|s1}|| {t2|s2} (4)
		s.arg1 = set1.sub();
		s.arg2 = set2.sub();                                         // s1||s2
		Constraint s1 = new Constraint(set1.first(), Enviroment.neqCode, set2.first()); // t1.neq(t2)
		Constraint s2 = new Constraint(set1.first(), Enviroment.ninCode, set2.sub());   // t1.nin(s2)
		Constraint s3 = new Constraint(set2.first(), Enviroment.ninCode, set1.sub());   // t2.nin(s1)
		Solver.add(s1);
		Solver.add(s2);
		Solver.add(s3);
		Solver.storeInvariato = false;
	    }
	}
	else if(set1.equ == null) {
	    s.arg2 = set2.equ;
	    disj(s);
	    return;
	}
	else if(set2.equ == null) {
	    s.arg1 = set1.equ;
	    disj(s);
	    return;
	}
	else {
	    s.arg1 = set1.equ;
	    s.arg2 = set2.equ;
	    disj(s);
	    return;
	}
    }


    /* Disgiunzione di una Lvar da un Set */
    protected void disjLvarSet(Lvar lv, Set set, Constraint s) throws Failure {
	if(set.equ == null) {
	    s.arg1 = (Set)lv.valSet();    // lv.valSet() restituisce il Set valore di lv
	    disj((Set)lv.valSet(), set, s); // se il valore di lv non  un set viene sollevata un eccezione NotSetException

	    //      Se la Lvar  inizializzata con un oggetto diverso da Set viene sollevata
	    //      un'eccezione di tipo NotSetException

	}
	else {
	    s.arg2 = set.equ;
	    disj(s);
	    return;
	}
    }


    protected void disjSetLvar(Set set, Lvar lv, Constraint s) throws Failure {
	if(set.equ == null) {
	    s.arg2 = (Set)lv.valSet();
	    disj(set, (Set)lv.valSet(), s);
	}
	else {
	    s.arg1 = set.equ;
	    disj(s);
	    return;
	}
    }


    /* Disgiunzione tra due Lvar */
    protected void disjLvarLvar(Lvar l1, Lvar l2, Constraint s) throws Failure {
	s.arg1 = (Set)l1.valSet();
	s.arg2 = (Set)l2.valSet();
	disj((Set)l1.valSet(), (Set)l2.valSet(), s);
    }


    /* Non disgiunzione tra due Set */
    protected void ndisj(Set set1, Set set2, Constraint s) throws Failure {
	if(set1.equ == null && set2.equ == null) {
	    Lvar n = new Lvar();
	    s.arg1 = n;
	    s.cons = Enviroment.inCode;
	    s.arg2 = set1;
	    Constraint s1 = new Constraint(n, Enviroment.inCode, set2);
	    Solver.add(s1);
	    Solver.storeInvariato = false;
	}
	else if(set1.equ == null) {
	    s.arg2 = set2.equ;
	    ndisj(s);
	    return;
	}
	else if(set2.equ == null) {
	    s.arg1 = set1.equ;
	    ndisj(s);
	    return;
	}
	else {
	    s.arg1 = set1.equ;
	    s.arg2 = set2.equ;
	    ndisj(s);
	    return;
	}
    }


    /* Non disgiunzione tra una Lvar e un Set */
    protected void ndisjLvarSet(Lvar lv, Set set, Constraint s) throws Failure {
	if(set.equ == null) {
	    s.arg1 = (Set)lv.valSet();
	    ndisj((Set)lv.valSet(), set, s);
	}
	else {
	    s.arg2 = set.equ;
	    ndisj(s);
	    return;
	}
    }


    protected void ndisjSetLvar(Set set, Lvar lv, Constraint s) throws Failure {
	if(set.equ == null) {
	    s.arg2 = (Set)lv.valSet();
	    ndisj(set, (Set)lv.valSet(), s);
	}
	else {
	    s.arg1 = set.equ;
	    ndisj(s);
	    return;
	}
    }


    /* Non disgiunzione tra due Lvar */
    protected void ndisjLvarLvar(Lvar l1, Lvar l2, Constraint s) throws Failure {
	s.arg1 = (Set)l1.valSet();
	s.arg2 = (Set)l2.valSet();
	ndisj((Set)l1.valSet(), (Set)l2.valSet(), s);
    }


    /* Unione */
    protected void union(Set set1, Set set2, Set set3, Constraint s) throws Failure {
	if(set1.equ == null && set2.equ == null && set3.equ == null) {
	    switch(s.caseControl) {
	    case 0:
		if(set1 == set2) {              // union(s, s, t) -> s.eq(t)           (1)
		    s.arg1 = set1;
		    s.arg2 = set3;
		    s.arg3 = null;
		    s.cons = Enviroment.eqCode;
		    eq(s);
		    return;
		}
		else if(set3 == Set.empty) {   // union(s, t, Ø) -> s.eq(Ø) , t.eq(Ø)  (2)
		    s.arg1 = set1;
		    s.arg2 = Set.empty;
		    s.arg3 = null;
		    s.cons = Enviroment.eqCode;
		    Constraint s1 = new Constraint(set2, Enviroment.eqCode, Set.empty);
		    Solver.add(Solver.indexOf(s)+1, s1);
		    eq(s);
		    return;
		}
		else if(set1 == Set.empty) {   // union(Ø, s, X) -> X.eq(s)             (3)
		    s.arg1 = set3;
		    s.arg2 = set2;
		    s.arg3 = null;
		    s.cons = Enviroment.eqCode;
		    eq(s);
		    return;
		}
		else if(set2 == Set.empty) {   // union(s, Ø, X) -> X.eq(s)             (3)
		    s.arg1 = set3;
		    s.arg2 = set1;
		    s.arg3 = null;
		    s.cons = Enviroment.eqCode;
		    eq(s);
		    return;
		}
		else if(set1.isGround() && set2.isGround()) {              // CASI GROUND
		    s.arg1 = set1.concat(set2);
		    s.arg2 = set3;
		    s.arg3 = null;
		    s.cons = Enviroment.eqCode;
		    eq(s);
		    return;
		}
		else if(set1.isGround()) {
		    s.arg1 = set1.appendGround(set2);
		    s.arg2 = set3;
		    s.arg3 = null;
		    s.cons = Enviroment.eqCode;
		    eq(s);
		    return;
		}
		else if(set2.isGround()) {
		    s.arg1 = set2.appendGround(set1);
		    s.arg2 = set3;
		    s.arg3 = null;
		    s.cons = Enviroment.eqCode;
		    eq(s);
		    return;
		}
		else if(set3.size() > 0) {            // union(s1, s2, {t1|t})               (4)
		    Set N = new Set();
		    s.arg1 = set3;
		    s.arg2 = N.ins(set3.first());
		    s.arg3 = null;
		    s.cons = Enviroment.eqCode;      // {t1|t}.eq( {t1|N})
		    Constraint s1 = new Constraint(set3.first(), Enviroment.ninCode, N);   // t1.nin(N)
		    Constraint s2 = new Constraint(Enviroment.unionCode, set1, set2, set3, N); // union
		    s2.caseControl = 1;
		    Solver.add(Solver.indexOf(s)+1, s1);
		    Solver.add(s2);
		    eq(s);
		    return;
		}
		else if(set1.size() > 0 && !set3.iniz) {     // union( {t1|t2}, t, X)        (5)
		    Set M = new Set();
		    Set M1 = new Set();
		    s.arg1 = set1;
		    s.arg2 = M1.ins(set1.first());
		    s.arg3 = null;
		    s.cons = Enviroment.eqCode;                                                 // {t1|t2}.eq( {t1|M1})
		    Constraint z1 = new Constraint(set1.first(), Enviroment.ninCode, M1);         // t1.nin(M1)
		    Constraint z2 = new Constraint(set3, Enviroment.eqCode, M.ins(set1.first())); // set3.eq(t1|M)
		    Constraint z3 = new Constraint(set1.first(), Enviroment.ninCode, M);          // t1.nin(M)
		    Constraint z4 = new Constraint(Enviroment.unionCode, set1, set2, set3, Lst.empty.ins1(M1).insn(M)); // union
		    z4.caseControl = 4;
		    Solver.add(Solver.indexOf(s)+1, z1);
		    Solver.add(Solver.indexOf(s)+1, z2);
		    Solver.add(Solver.indexOf(s)+1, z3);
		    Solver.add(z4);
		    eq(s);
		    return;
		}
		else if(set2.size() > 0 && !set3.iniz) { // union(t, {t1|t2}, X)            (5)
		    s.arg1 = set2;
		    s.arg2 = set1;
		    s.caseControl = 0;
		    union(set2, set1, set3, s);
		    return;
		}
		else {
		    return;
		}
	    }
	}
	else if(set1.equ != null) {
	    s.arg1 = set1.equ;
	    union(s);
	    return;
	}
	else if(set2.equ != null) {
	    s.arg2 = set2.equ;
	    union(s);
	    return;
	}
	else {
	    s.arg3 = set3.equ;
	    union(s);
	    return;
	}
    }


    protected void union(Set set1, Set set2, Set set3, Set N, Constraint s) throws Failure {
	if(set1.equ == null && set2.equ == null && set3.equ == null && N.equ == null) {
	    switch(s.caseControl) {
	    case 1:                                   // union(s1, s2, {t1|t})     (4) (i)
		VarState statoVarUnion1 = Solver.B.getVarState();
		StoreState statoStoreUnion1 = Solver.B.getStoreState(s);
		Solver.B.addChoicePoint(2, statoVarUnion1, statoStoreUnion1);
		Set N1 = new Set();
		s.arg1 = set1;                                       // set1.eq( {t1|N1})
		s.arg2 = N1.ins(set3.first());
		s.arg3 = null;
		s.arg4 = null;
		s.cons = Enviroment.eqCode;
		s.caseControl = 0;
		Constraint s4 = new Constraint(set3.first(), Enviroment.ninCode, N1);     // t1.nin(N1)
		Constraint s5 = new Constraint(Enviroment.unionCode, N1, set2, N);        // union(N1, set2, N)
		Solver.add(Solver.indexOf(s)+1, s4);
		Solver.add(s5);
		eq(s);
		return;
	    case 2:                                                      //       (4) (ii)
		VarState statoVarUnion2 = Solver.B.getVarState();
		StoreState statoStoreUnion2 = Solver.B.getStoreState(s);
		Solver.B.addChoicePoint(3, statoVarUnion2, statoStoreUnion2);
		Set NN1 = new Set();
		s.arg1 = set2;            // set2.eq( {t1|NN1})
		s.arg2 = NN1.ins(set3.first());
		s.arg3 = null;
		s.arg4 = null;
		s.cons = Enviroment.eqCode;
		s.caseControl = 0;
		Constraint s6 = new Constraint(set3.first(), Enviroment.ninCode, NN1);     // t1.nin(NN1)
		Constraint s7 = new Constraint(Enviroment.unionCode, set1, NN1, N);           // union(set1, NN1, N)
		Solver.add(Solver.indexOf(s)+1, s6);
		Solver.add(s7);
		eq(s);
		return;
	    case 3:                                                      //        (4) (iii)
		Set NNN1 = new Set();
		s.arg1 = set1;            // set1.eq( {t1|NNN1})
		s.arg2 = NNN1.ins(set3.first());
		s.arg3 = null;
		s.arg4 = null;
		s.cons = Enviroment.eqCode;
		s.caseControl = 0;
		Constraint s8 = new Constraint(set3.first(), Enviroment.ninCode, NNN1);     // t1.nin(NNN1)
		Set N2 = new Set();
		Constraint s9 = new Constraint(set2, Enviroment.eqCode, N2.ins(set3.first()));  // set1.eq( {t1|N2})
		Constraint s10 = new Constraint(set3.first(), Enviroment.ninCode, N2);             // t1.nin(N2)
		Constraint s11 = new Constraint(Enviroment.unionCode, NNN1, N2, N);                   // union(NNN1, N2, N)
		Solver.add(Solver.indexOf(s)+1, s8);
		Solver.add(Solver.indexOf(s)+1, s9);
		Solver.add(Solver.indexOf(s)+1, s10);
		Solver.add(s11);
		eq(s);
		return;
	    }
	}
	else if(set1.equ != null) {
	    s.arg1 = set1.equ;
	    union(set1.equ, set2, set3, N, s);
	    return;
	}
	else if(set2.equ != null) {
	    s.arg2 = set2.equ;
	    union(set1, set2.equ, set3, N, s);
	    return;
	}
	else if(set3.equ != null) {
	    s.arg3 = set3.equ;
	    union(set1, set2, set3.equ, N, s);
	    return;
	}
	else {
	    s.arg4 = N.equ;
	    union(set1, set2, set3, N.equ, s);
	    return;
	}
    }


    protected void union(Set set1, Set set2, Set set3, Set M1, Set M, Constraint s) throws Failure {
	if(set1.equ == null && set2.equ == null && set3.equ == null && M1.equ == null && M.equ == null) {
	    switch(s.caseControl) {
	    case 4:                                                   //  (5) (i)
		VarState statoVarUnion4 = Solver.B.getVarState();
		StoreState statoStoreUnion4 = Solver.B.getStoreState(s);
		Solver.B.addChoicePoint(5, statoVarUnion4, statoStoreUnion4);
		s.arg1 = set1.first();                             // t1.nin(set2)
		s.arg2 = set2;
		s.arg3 = null;
		s.arg4 = null;
		s.cons = Enviroment.ninCode;
		s.caseControl = 0;
		Constraint z5 = new Constraint(Enviroment.unionCode, M1, set2, M);         // union(M1, set2, M)
		Solver.add(z5);
		nin(s);
		return;
	    case 5:                                                   //  (5) (ii)
		Set M2 = new Set();
		s.arg1 = set2;                                     // set2.eq( {t1|M2})
		s.arg2 = M2.ins(set1.first());
		s.arg3 = null;
		s.arg4 = null;
		s.cons = Enviroment.eqCode;
		s.caseControl = 0;
		Constraint z6 = new Constraint(set1.first(), Enviroment.ninCode, M2);   // t1.nin(M2)
		Constraint z7 = new Constraint(Enviroment.unionCode, M1, M2, M);           // union(M1, M2, M)
		Solver.add(Solver.indexOf(s)+1, z6);
		Solver.add(z7);
		eq(s);
		return;
	    }
	}
	else if(set1.equ != null) {
	    s.arg1 = set1.equ;
	    union(set1.equ, set2, set3, M1, M, s);
	    return;
	}
	else if(set2.equ != null) {
	    s.arg2 = set2.equ;
	    union(set1, set2.equ, set3, M1, M, s);
	    return;
	}
	else if(set3.equ != null) {
	    s.arg3 = set3.equ;
	    union(set1, set2, set3.equ, M1, M, s);
	    return;
	}
	else if(M1.equ != null) {
	    union(set1, set2, set3, M1.equ, M, s);
	    return;
	}
	else {
	    union(set1, set2, set3, M1, M.equ, s);
	    return;
	}
    }


    protected void union(Lvar l, Set set1, Set set2, Constraint s) throws Failure {
	s.arg1 = (Set)l.valSet();
	union((Set)l.valSet(), set1, set2, s);
	return;
    }

    protected void union(Lvar l1, Lvar l2, Set set, Constraint s) throws Failure {
	s.arg1 = (Set)l1.valSet();
	s.arg2 = (Set)l2.valSet();
	union((Set)l1.valSet(), (Set)l2.valSet(), set, s);
	return;
    }


    protected void union(Set set1, Set set2, Lvar l, Constraint s) throws Failure {
	s.arg3 = (Set)l.valSet();
	union(set1, set2, (Set)l.valSet(), s);
	return;
    }

    protected void union(Lvar l1, Set set2, Lvar l2, Constraint s) throws Failure {
	s.arg1 = (Set)l1.valSet();
	s.arg3 = (Set)l2.valSet();
	union((Set)l1.valSet(), set2, (Set)l2.valSet(), s);
	return;
    }


    protected void union(Lvar l1, Lvar l2, Lvar l3, Constraint s) throws Failure {
	s.arg1 = (Set)l1.valSet();
	s.arg2 = (Set)l2.valSet();
	s.arg3 = (Set)l3.valSet();
	union((Set)l1.valSet(), (Set)l2.valSet(), (Set)l3.valSet(), s);
	return;
    }


    protected void neq_union(Set set1, Set set2, Constraint s) throws Failure {
	if(set1.equ == null && set2.equ == null) {
	    switch(s.caseControl) {
            case 0:                                        //  union(X, Y, Z)  (6)/(7) (i)
		Solver.B.addChoicePoint(1, (VarState)Solver.B.getVarState(),
					 (StoreState)Solver.B.getStoreState(s));
		Lvar N = new Lvar();
		s.arg1 = N;                                 // N in set1
		s.cons = Enviroment.inCode;
		s.arg2 = set1;
		Constraint s0 = new Constraint(N, Enviroment.ninCode, set2);     // N nin set2
		Solver.add(s0);
		Solver.storeInvariato = false;
		return;
            case 1:                                        //  (6)/(7) (ii)
		Solver.B.addChoicePoint(2, (VarState)Solver.B.getVarState(),
					 (StoreState)Solver.B.getStoreState(s));
		Lvar NN = new Lvar();
		s.arg1 = NN;
		s.cons = Enviroment.inCode;
		s.arg2 = set2;
		s.caseControl = 0;
		Constraint s1 = new Constraint(NN, Enviroment.ninCode, set1);
		Solver.add(s1);
		Solver.storeInvariato = false;
		return;
            case 2:                                        //  (6)/(7)  (iii)
		s.arg1 = set1;
		s.cons = Enviroment.eqCode;
		s.arg2 = Set.empty;
		s.caseControl = 0;
		Constraint s2 = new Constraint(set2, Enviroment.neqCode, Set.empty);
		Solver.add(s2);
		Solver.storeInvariato = false;
		return;
	    }
	}
	else if(set1.equ == null) {
	    s.arg2 = set2.equ;
	    neq_union(s);
	    return;
	}
	else {
	    s.arg1 = set1.equ;
	    neq_union(s);
	    return;
	}
    }


    /* Non Unione tra Set */
    protected void nunion(Set set1, Set set2, Set set3, Constraint s) throws Failure {
	if(set1.equ == null && set2.equ == null && set3.equ == null) {
	    switch(s.caseControl) {
            case 0:                                               // (1) (i)
		VarState statoVarNunion0 = Solver.B.getVarState();
		StoreState statoStoreNunion0 = Solver.B.getStoreState(s);
		Solver.B.addChoicePoint(1, statoVarNunion0, statoStoreNunion0);
		Lvar n = new Lvar();
		s.arg1 = n;
		s.arg2 = set3;
		s.cons = Enviroment.inCode;
		s.arg3 = null;
		Constraint s0 = new Constraint(n, Enviroment.ninCode, set1);
		Constraint s1 = new Constraint(n, Enviroment.ninCode, set2);
		Solver.add(s1);
		Solver.add(s0);
		Solver.storeInvariato = false;
		in(s);
		return;
            case 1:                                               // (1) (ii)
		VarState statoVarNunion1 = Solver.B.getVarState();
		StoreState statoStoreNunion1 = Solver.B.getStoreState(s);
		Solver.B.addChoicePoint(2, statoVarNunion1, statoStoreNunion1);
		Lvar m = new Lvar();
		s.arg1 = m;
		s.arg2 = set1;
		s.cons = Enviroment.inCode;
		s.caseControl = 0;
		Constraint s2 = new Constraint(m, Enviroment.ninCode, set3);
		Solver.add(s2);
		Solver.storeInvariato = false;
		in(s);
		return;
            case 2:                                               // (1) (iii)
		Lvar l = new Lvar();
		s.arg1 = l;
		s.arg2 = set2;
		s.cons = Enviroment.inCode;
		s.caseControl = 0;
		Constraint s3 = new Constraint(l, Enviroment.ninCode, set3);
		Solver.add(s3);
		Solver.storeInvariato = false;
		in(s);
		return;
	    }
	}
	else if(set1.equ != null) {
	    s.arg1 = set1.equ;
	    nunion(s);
	    return;
	}
	else if(set2.equ != null) {
	    s.arg2 = set2.equ;
	    nunion(s);
	    return;
	}
	else {
	    s.arg3 = set3.equ;
	    nunion(s);
	    return;
	}
    }


    protected void nunion(Lvar l, Set set1, Set set2, Constraint s) throws Failure {
	s.arg1 = (Set)l.valSet();
	nunion((Set)l.valSet(), set1, set2, s);
	return;
    }


    protected void nunion(Lvar l1, Lvar l2, Set set, Constraint s) throws Failure {
	s.arg1 = (Set)l1.valSet();
	s.arg2 = (Set)l2.valSet();
	nunion((Set)l1.valSet(), (Set)l2.valSet(), set, s);
	return;
    }


    protected void nunion(Set set1, Set set2, Lvar l, Constraint s) throws Failure {
	s.arg3 = (Set)l.valSet();
	nunion(set1, set2, (Set)l.valSet(), s);
	return;
    }


    protected void nunion(Lvar l1, Set set2, Lvar l2, Constraint s) throws Failure {
	s.arg1 = (Set)l1.valSet();
	s.arg3 = (Set)l2.valSet();
	nunion((Set)l1.valSet(), set2, (Set)l2.valSet(), s);
	return;
    }


    protected void nunion(Lvar l1, Lvar l2, Lvar l3, Constraint s) throws Failure {
	s.arg1 = (Set)l1.valSet();
	s.arg2 = (Set)l2.valSet();
	s.arg3 = (Set)l3.valSet();
	nunion((Set)l1.valSet(), (Set)l2.valSet(), (Set)l3.valSet(), s);
	return;
    }


    protected void ge(Lvar l1, Lvar l2, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(!l1.iniz || !l2.iniz)return;
	    else if((l1.val instanceof Integer) && (l2.val instanceof Integer)) {
		if(((Integer)l1.val).intValue() >= ((Integer)l2.val).intValue()) {
		    s.bool = true;
                    Solver.storeInvariato = false;
		    return;
		}
		else Solver.fail(s);
	    }
	    else throw new NotIntLvarException();
	}
	else if(l1.equ == null) {
	    s.arg2 = l2.equ;
	    ge(l1, l2.equ, s);
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    ge(l1.equ, l2, s);
	    return;
	}
	return;
    }


    protected void ge(Lvar l1, Integer i, Constraint s) throws Failure {
	if(l1.equ == null) {
	    if(!l1.iniz)return;
	    else if(l1.val instanceof Integer) {
		if(((Integer)l1.val).intValue() >= i.intValue()) {
		    s.bool = true;
                    Solver.storeInvariato = false;
		    return;
		}
		else Solver.fail(s);
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    s.arg1 = l1.equ;
	    ge(l1.equ, i, s);
	    return;
	}
	return;
    }


    protected void gt(Lvar l1, Lvar l2, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(!l1.iniz || !l2.iniz)return;
	    else if((l1.val instanceof Integer) && (l2.val instanceof Integer)) {
		if(((Integer)l1.val).intValue() > ((Integer)l2.val).intValue()) {
		    s.bool = true;
                    Solver.storeInvariato = false;
		    return;
		}
		else Solver.fail(s);
	    }
	    else throw new NotIntLvarException();
	}
	else if(l1.equ == null) {
	    s.arg2 = l2.equ;
	    gt(l1, l2.equ, s);
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    gt(l1.equ, l2, s);
	    return;
	}
	return;
    }


    protected void gt(Lvar l1, Integer i, Constraint s) throws Failure {
	if(l1.equ == null) {
	    if(!l1.iniz)return;
	    else if(l1.val instanceof Integer) {
		if(((Integer)l1.val).intValue() > i.intValue()) {
		    s.bool = true;
                    Solver.storeInvariato = false;
		    return;
		}
		else Solver.fail(s);
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    s.arg1 = l1.equ;
	    gt(l1.equ, i, s);
	    return;
	}
	return;
    }


    protected void le(Lvar l1, Lvar l2, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(!l1.iniz || !l2.iniz)return;
	    else if((l1.val instanceof Integer) && (l2.val instanceof Integer)) {
		if(((Integer)l1.val).intValue() <= ((Integer)l2.val).intValue()) {
		    s.bool = true;
                    Solver.storeInvariato = false;
		    return;
		}
		else Solver.fail(s);
	    }
	    else throw new NotIntLvarException();
	}
	else if(l1.equ == null) {
	    s.arg2 = l2.equ;
	    le(l1, l2.equ, s);
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    le(l1.equ, l2, s);
	    return;
	}
	return;
    }


    protected void le(Lvar l1, Integer i, Constraint s) throws Failure {
	if(l1.equ == null) {
	    if(!l1.iniz)return;
	    else if(l1.val instanceof Integer) {
		if(((Integer)l1.val).intValue() <= i.intValue()) {
		    s.bool = true;
                    Solver.storeInvariato = false;
		    return;
		}
		else Solver.fail(s);
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    s.arg1 = l1.equ;
	    le(l1.equ, i, s);
	    return;
	}
	return;
    }

    protected void lt(Lvar l1, Lvar l2, Constraint s)throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(!l1.iniz || !l2.iniz)return;
	    else if((l1.val instanceof Integer) && (l2.val instanceof Integer)) {
		if(((Integer)l1.val).intValue() < ((Integer)l2.val).intValue()) {
		    s.bool = true;
                    Solver.storeInvariato = false;
		    return;
		}
		else Solver.fail(s);
	    }
	    else throw new NotIntLvarException();
	}
	else if(l1.equ == null) {
	    s.arg2 = l2.equ;
	    lt(l1, l2.equ, s);
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    lt(l1.equ, l2, s);
	    return;
	}
	return;
    }


    protected void lt(Lvar l1, Integer i, Constraint s) throws Failure {
	if(l1.equ == null) {
	    if(!l1.iniz)return;
	    else if(l1.val instanceof Integer) {
		if(((Integer)l1.val).intValue() < i.intValue()) {
		    s.bool = true;
                    Solver.storeInvariato = false;
		    return;
		}
		else Solver.fail(s);
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    s.arg1 = l1.equ;
	    lt(l1.equ, i, s);
	    return;
	}
	return;
    }


    protected void sumLvar(Lvar l1, Lvar l2, Lvar l3, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null && l3.equ == null) {
	    if(!l2.iniz || !l3.iniz)return;
	    else if((l2.val instanceof Integer) && (l3.val instanceof Integer)) {
		Integer sum = new Integer(((Integer)l2.val).intValue() + ((Integer)l3.val).intValue());
		l1.iniz = true;
		l1.val = sum;
		s.bool = true;
                Solver.storeInvariato = false;
		return;
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    if(l1.equ != null) {
		s.arg1 = l1.equ;
		sum(s);
		return;
	    }
	    else if(l2.equ != null) {
		s.arg2 = l2.equ;
		sum(s);
		return;
	    }
	    else {
		s.arg3 = l3.equ;
		sum(s);
		return;
	    }
	}
    }


    protected void sumInteger(Lvar l1, Integer i1, Integer i2, Constraint s) throws Failure {
	if(l1.equ == null) {
	    Integer sum = new Integer(i1.intValue() + i2.intValue());
	    l1.iniz = true;
	    l1.val = sum;
	    s.bool = true;
            Solver.storeInvariato = false;
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    sum(s);
	    return;
	}
    }


    protected void sumLvarInteger(Lvar l1, Lvar l2, Integer i, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(!l2.iniz)return;
	    else if(l2.val instanceof Integer) {
		Integer sum = new Integer(((Integer)l2.val).intValue() + i.intValue());
		l1.iniz = true;
		l1.val = sum;
		s.bool = true;
                Solver.storeInvariato = false;
		return;
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    if(l1.equ != null) {
		s.arg1 = l1.equ;
		sum(s);
		return;
	    }
	    else {
		s.arg2 = l2.equ;
		s.arg3 = i;
		sum(s);
		return;
	    }
	}
    }


    protected void subLvar(Lvar l1, Lvar l2, Lvar l3, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null && l3.equ == null) {
	    if(!l2.iniz || !l3.iniz)return;
	    else if((l2.val instanceof Integer) && (l3.val instanceof Integer)) {
		Integer dif = new Integer(((Integer)l2.val).intValue() - ((Integer)l3.val).intValue());
		l1.iniz = true;
		l1.val = dif;
		s.bool = true;
                Solver.storeInvariato = false;
		return;
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    if(l1.equ != null) {
		s.arg1 = l1.equ;
		sub(s);
		return;
	    }
	    else if(l2.equ != null) {
		s.arg2 = l2.equ;
		sub(s);
		return;
	    }
	    else {
		s.arg3 = l3.equ;
		sub(s);
		return;
	    }
	}
    }


    protected void subInteger(Lvar l1, Integer i1, Integer i2, Constraint s) throws Failure {
	if(l1.equ == null) {
	    Integer dif = new Integer(i1.intValue() - i2.intValue());
	    l1.iniz = true;
	    l1.val = dif;
	    s.bool = true;
            Solver.storeInvariato = false;
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    sub(s);
	    return;
	}
    }


    protected void subLvarInteger(Lvar l1, Lvar l2, Integer i, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(!l2.iniz)return;
	    else if(l2.val instanceof Integer) {
		Integer dif = new Integer(((Integer)l2.val).intValue() - i.intValue());
		l1.iniz = true;
		l1.val = dif;
		s.bool = true;
                Solver.storeInvariato = false;
		return;
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    if(l1.equ != null) {
		s.arg1 = l1.equ;
		sub(s);
		return;
	    }
	    else {
		s.arg2 = l2.equ;
		s.arg3 = i;
		sub(s);
		return;
	    }
	}
    }


    protected void mulLvar(Lvar l1, Lvar l2, Lvar l3, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null && l3.equ == null) {
	    if(!l2.iniz || !l3.iniz)return;
	    else if((l2.val instanceof Integer) && (l3.val instanceof Integer)) {
		Integer mul = new Integer(((Integer)l2.val).intValue() * ((Integer)l3.val).intValue());
		l1.iniz = true;
		l1.val = mul;
		s.bool = true;
                Solver.storeInvariato = false;
		return;
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    if(l1.equ != null) {
		s.arg1 = l1.equ;
		mul(s);
		return;
	    }
	    else if(l2.equ != null) {
		s.arg2 = l2.equ;
		mul(s);
		return;
	    }
	    else {
		s.arg3 = l3.equ;
		mul(s);
		return;
	    }
	}
    }


    protected void mulInteger(Lvar l1, Integer i1, Integer i2, Constraint s) throws Failure {
	if(l1.equ == null) {
	    Integer mul = new Integer(i1.intValue() * i2.intValue());
	    l1.iniz = true;
	    l1.val = mul;
	    s.bool = true;
            Solver.storeInvariato = false;
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    mul(s);
	    return;
	}
    }


    protected void mulLvarInteger(Lvar l1, Lvar l2, Integer i, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(!l2.iniz)return;
	    else if(l2.val instanceof Integer) {
		Integer mul = new Integer(((Integer)l2.val).intValue() * i.intValue());
		l1.iniz = true;
		l1.val = mul;
		s.bool = true;
                Solver.storeInvariato = false;
		return;
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    if(l1.equ != null) {
		s.arg1 = l1.equ;
		mul(s);
		return;
	    }
	    else {
		s.arg2 = l2.equ;
		s.arg3 = i;
		mul(s);
		return;
	    }
	}
    }


    protected void divLvar(Lvar l1, Lvar l2, Lvar l3, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null && l3.equ == null) {
	    if(!l2.iniz || !l3.iniz)return;
	    else if((l2.val instanceof Integer) && (l3.val instanceof Integer)) {
		Float quo = new Float(((Integer)l2.val).intValue() / ((Integer)l3.val).intValue());
		l1.iniz = true;
		l1.val = quo;
		s.bool = true;
                Solver.storeInvariato = false;
		return;
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    if(l1.equ != null) {
		s.arg1 = l1.equ;
		div(s);
		return;
	    }
	    else if(l2.equ != null) {
		s.arg2 = l2.equ;
		div(s);
		return;
	    }
	    else {
		s.arg3 = l3.equ;
		div(s);
		return;
	    }
	}
    }


    protected void divInteger(Lvar l1, Integer i1, Integer i2, Constraint s) throws Failure {
	if(l1.equ == null) {
	    Float quo = new Float(i1.intValue() / i2.intValue());
	    l1.iniz = true;
	    l1.val = quo;
	    s.bool = true;
            Solver.storeInvariato = false;
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    div(s);
	    return;
	}
    }


    protected void divLvarInteger(Lvar l1, Lvar l2, Integer i, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(!l2.iniz)return;
	    else if(l2.val instanceof Integer) {
		Float quo = new Float(((Integer)l2.val).intValue() / i.intValue());
		l1.iniz = true;
		l1.val = quo;
		s.bool = true;
                Solver.storeInvariato = false;
		return;
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    if(l1.equ != null) {
		s.arg1 = l1.equ;
		div(s);
		return;
	    }
	    else {
		s.arg2 = l2.equ;
		s.arg3 = i;
		div(s);
		return;
	    }
	}
    }


    protected void modLvar(Lvar l1, Lvar l2, Lvar l3, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null && l3.equ == null) {
	    if(!l2.iniz || !l3.iniz)return;
	    else if((l2.val instanceof Integer) && (l3.val instanceof Integer)) {
		Integer mod = new Integer(((Integer)l2.val).intValue() % ((Integer)l3.val).intValue());
		l1.iniz = true;
		l1.val = mod;
		s.bool = true;
                Solver.storeInvariato = false;
		return;
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    if(l1.equ != null) {
		s.arg1 = l1.equ;
		mod(s);
		return;
	    }
	    else if(l2.equ != null) {
		s.arg2 = l2.equ;
		mod(s);
		return;
	    }
	    else {
		s.arg3 = l3.equ;
		mod(s);
		return;
	    }
	}
    }


    protected void modInteger(Lvar l1, Integer i1, Integer i2, Constraint s) throws Failure {
	if(l1.equ == null) {
	    Integer mod = new Integer(i1.intValue() % i2.intValue());
	    l1.iniz = true;
	    l1.val = mod;
	    s.bool = true;
            Solver.storeInvariato = false;
	    return;
	}
	else {
	    s.arg1 = l1.equ;
	    mod(s);
	    return;
	}
    }


    protected void modLvarInteger(Lvar l1, Lvar l2, Integer i, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(!l2.iniz)return;
	    else if(l2.val instanceof Integer) {
		Integer mod = new Integer(((Integer)l2.val).intValue() % i.intValue());
		l1.iniz = true;
		l1.val = mod;
		s.bool = true;
                Solver.storeInvariato = false;
		return;
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    if(l1.equ != null) {
		s.arg1 = l1.equ;
		mod(s);
		return;
	    }
	    else {
		s.arg2 = l2.equ;
		s.arg3 = i;
		mod(s);
		return;
	    }
	}
    }


    protected void modIntegerLvar(Lvar l1, Integer i, Lvar l2, Constraint s) throws Failure {
	if(l1.equ == null && l2.equ == null) {
	    if(!l2.iniz)return;
	    else if(l2.val instanceof Integer) {
		Integer mod = new Integer(i.intValue() % ((Integer)l2.val).intValue());
		l1.iniz = true;
		l1.val = mod;
		s.bool = true;
		return;
	    }
	    else throw new NotIntLvarException();
	}
	else {
	    if(l1.equ != null) {
		s.arg1 = l1.equ;
		mod(s);
		return;
	    }
	    else {
		s.arg3 = l2.equ;
		mod(s);
		return;
	    }
	}
    }


    /* Intervallo con estremi variabili */
    protected void interv(Set S, Lvar L1, Lvar L2, Constraint s) throws Failure {
	if(L1.equ == null && L2.equ == null) {
	    Vector v = new Vector();
	    for(int i = ((Integer)L1.value()).intValue();i <= ((Integer)L2.value()).intValue(); i++) {
		Integer ii = new Integer(i);
		v.add(ii);
	    }
	    S = new Set(v, Set.empty);
	    Solver.storeInvariato = false;
	    return;
	}
	else {
	    if(L1.equ != null) {
		s.arg2 = L1.equ;
		interv(s);
		return;
	    }
	    else {
		s.arg3 = L2.equ;
		interv(s);
		return;
	    }
	}
    }


    protected void interv(Set S, Integer n, Lvar L2, Constraint s) throws Failure {
	if(L2.equ == null) {
	    Vector v = new Vector();
	    for(int i = n.intValue() ;i <= ((Integer)L2.value()).intValue(); i++) {
		Integer ii = new Integer(i);
		v.add(ii);
	    }
	    S.iniz = true;
	    S.lista = v;
	    S.resto = Set.empty;
	    return;
	}
	else {
	    s.arg3 = L2.equ;
	    interv(s);
	    return;
	}
    }


    protected void interv(Set S, Lvar L1, Integer n, Constraint s) throws Failure {
	if(L1.equ == null) {
	    Vector v = new Vector();
	    for(int i = ((Integer)L1.value()).intValue();i <= n.intValue();i++) {
		Integer ii = new Integer(i);
		v.add(ii);
	    }
	    S = new Set(v, Set.empty);
	    return;
	}
	else {
	    s.arg2 = L1.equ;
	    interv(s);
	    return;
	}
    }


    protected boolean confrontaNeqLst(Lst l, Lst m) {
	Vector v = l.toVector();
	Vector w = m.toVector();
	Lst ll = l.ultimoResto();
	Lst mm = m.ultimoResto();
	if(v.size() == w.size()) {
	    for(int i = 0; i < v.size(); i++) {
		if(v.get(i) instanceof Lvar || v.get(i) instanceof Lst ||
		   w.get(i) instanceof Lvar || w.get(i) instanceof Lst )continue;
		else {
		    if(v.get(i) != w.get(i))return true;
		    else continue;
		}
	    }
	    return false;
	}
	else if(v.size() < w.size()) {
	    if(ll == mm)return true;
	    else {
		for(int i = 0; i < v.size(); i++) {
		    if(v.get(i) instanceof Lvar || v.get(i) instanceof Lvar ||
		       w.get(i) instanceof Lst || w.get(i) instanceof Lst ||
		       w.get(i) instanceof Set || w.get(i) instanceof Set )continue;
		    else {
			if(v.get(i) != w.get(i))return true;
			else continue;
		    }
		}
		return false;
	    }
	}
	else {                               //v.size() > w.size()
	    if(ll == mm)return true;
	    else {
		for(int i = 0; i < w.size(); i++) {
		    if(v.get(i) instanceof Lvar || v.get(i) instanceof Lvar ||
		       w.get(i) instanceof Lst || w.get(i) instanceof Lst ||
		       w.get(i) instanceof Set || w.get(i) instanceof Set )continue;
		    else {
			if(v.get(i) != w.get(i))return true;
			else continue;
		    }
		}
		return false;
	    }
	}
    }


    /* forall annidati */

    public void forall(Lst Y, Lst S, ConstraintsConjunction vs, Constraint s) throws Failure {
        if (Y.isAllUninitialized()) {
            if(Y.size_private() == 1) {
                s.bool = true;
                Solver.forall((Lvar)Y.get(0), (Set)S.get(0), vs);
                return;
                // Se Y e S hanno un solo elemento si ricade nel caso di forall pi semplice (senza annidamenti).
            }
            else {
                // Se  limitato (cio non della forma {x|R}) ed  diverso dal vuoto...
                if(!((Set)S.get(0)).isEmpty() && ((Set)S.get(0)).ultimoResto().iniz) {
                    s.bool = true;
                    Vector v = ((Set)S.get(0)).toVector();
                    for(int i = 0; i < v.size(); i++) {
                        Lst Ynew = new Lst(Lst.mkLst(Y.size_private()));
                        Solver.add(((Lvar)Ynew.get(0)).eq(v.get(i)));
                        Solver.forall(Ynew.ext1(), S.ext1(), vs.sostAll(Y, Ynew));
                    }
                    return;
                }
                else {
                    switch(s.caseControl) {
                    case 0:
                        s.bool = true;
                        Solver.B.addChoicePoint(s);
                        Solver.add(((Set)S.get(0)).eq(Set.empty));
                        return;
                    case 1:
                        s.bool = true;
                        Set R = new Set();
                        Lst Ynew = new Lst(Lst.mkLst(Y.size_private()));
                        Solver.add(((Set)S.get(0)).neq(Set.empty));
                        Solver.add(((Set)S.get(0)).less((Lvar)Ynew.get(0), R));

                        Solver.forall(Ynew.ext1(), S.ext1(), vs.sostAll(Y, Ynew));

                        Lst Snew = new Lst(S.ext1().ins1(R));

                        Solver.forall(Y, Snew, vs);

                        Solver.storeInvariato = false;
                        return;
                    }
                }
            }
        }
        else {throw new InitLvarException();}
    }


    // 26/04/04 Implementazione di un numero qualunque di forall annidati
    public void forall( Lst Y, Lst S, Constraint vs, Constraint s) throws Failure {
        if (Y.isAllUninitialized()) {
            if(Y.size_private() == 1) {
                s.bool = true;
                Solver.forall((Lvar)Y.get(0), (Set)S.get(0), vs);
                return;
                // Se Y e S hanno un solo elemento si ricade nel caso di forall pi semplice (senza annidamenti).
            }
            else {
                if(!((Set)S.get(0)).isEmpty() && ((Set)S.get(0)).ultimoResto().iniz) {
                    s.bool = true;
                    Vector v = ((Set)S.get(0)).toVector();
                    for(int i = 0; i < v.size(); i++) {
                        Lst Ynew = new Lst(Lst.mkLst(Y.size_private()));
                        Solver.add(((Lvar)Ynew.get(0)).eq(v.get(i)));
                        Solver.forall(Ynew.ext1(), S.ext1(), vs.sostAll(Y, Ynew));
                    }
                    return;
                }
                else {
                    switch(s.caseControl) {
                    case 0:
                        s.bool = true;
                        Solver.B.addChoicePoint(s);
                        Solver.add(((Set)S.get(0)).eq(Set.empty));
                        return;
                    case 1:
                        s.bool = true;
                        Set R = new Set();
                        Lst Ynew = new Lst(Lst.mkLst(Y.size_private()));
                        Solver.add(((Set)S.get(0)).neq(Set.empty));
                        Solver.add(((Set)S.get(0)).less((Lvar)Ynew.get(0), R));

                        Solver.forall(Ynew.ext1(), S.ext1(), vs.sostAll(Y, Ynew));

                        Lst Snew = new Lst(S.ext1().ins1(R));

                        Solver.forall(Y, Snew, vs);

                        Solver.storeInvariato = false;
                        return;
                    }
                }
            }
        }
        else { throw new InitLvarException(); }
    }


    // 27/04/04 Implementazione di un numero qualunque di forall annidati
    public void forall( Lst Y, Lst S, Lst l, Constraint vs, Constraint s) throws Failure {
        if (Y.isAllUninitialized() && l.isAllUninitialized()) {
            if(Y.size_private() == 1) {
                s.bool = true;
                Solver.forall((Lvar)Y.get(0), (Set)S.get(0), l, vs);
                return;
                // Se Y e S hanno un solo elemento si ricade nel caso di forall pi semplice (senza annidamenti).
            }
            else {
                if(!((Set)S.get(0)).isEmpty() && ((Set)S.get(0)).ultimoResto().iniz) {
                    s.bool = true;
                    Vector v = ((Set)S.get(0)).toVector();
                    for(int i = 0; i < v.size(); i++) {
                        Lst Ynew = new Lst(Lst.mkLst(Y.size_private()));
                        Lst L = new Lst(Lst.mkLst(l.size_private()));
                        Solver.add(((Lvar)Ynew.get(0)).eq(v.get(i)));
                        Solver.forall(Ynew.ext1(), S.ext1(), L, vs.sostAll(Y, Ynew).sostAll(l, L));
                    }
                    return;
                }
                else {
                    switch(s.caseControl) {
                    case 0:
                        s.bool = true;
                        Solver.B.addChoicePoint(s);
                        Solver.add(((Set)S.get(0)).eq(Set.empty));
                        return;
                    case 1:
                        s.bool = true;
                        Set R = new Set();
                        Lst Ynew = new Lst(Lst.mkLst(Y.size_private()));
                        Lst L = new Lst(Lst.mkLst(l.size_private()));
                        Solver.add(((Set)S.get(0)).neq(Set.empty));
                        Solver.add(((Set)S.get(0)).less((Lvar)Ynew.get(0), R));

                        Solver.forall(Ynew.ext1(), S.ext1(), L, vs.sostAll(Y, Ynew).sostAll(l, L));

                        Lst Snew = new Lst(S.ext1().ins1(R));

                        Solver.forall(Y, Snew, l, vs);

                        Solver.storeInvariato = false;
                        return;
                    }
                }
            }
        }
        else { throw new InitLvarException(); }
    }


    public void forall(Lst Y, Lst S, Lst l, ConstraintsConjunction vs, Constraint s) throws Failure {
        if (Y.isAllUninitialized() && l.isAllUninitialized()) {
            if(Y.size_private() == 1) {
                s.bool = true;
                Solver.forall((Lvar)Y.get(0), (Set)S.get(0), l, vs);
                return;
                // Se Y e S hanno un solo elemento si ricade nel caso di forall pi semplice (senza annidamenti).
            }
            else {
                if(!((Set)S.get(0)).isEmpty() && ((Set)S.get(0)).ultimoResto().iniz) {
                    s.bool = true;
                    Vector v = ((Set)S.get(0)).toVector();
                    for(int i = 0; i < v.size(); i++) {
                        Lst Ynew = new Lst(Lst.mkLst(Y.size_private()));
                        Lst L = new Lst(Lst.mkLst(l.size_private()));
                        Solver.add(((Lvar)Ynew.get(0)).eq(v.get(i)));
                        Solver.forall(Ynew.ext1(), S.ext1(), L, vs.sostAll(Y, Ynew).sostAll(l, L));
                    }
                    return;
                }
                else {
                    switch(s.caseControl) {
                    case 0:
                        s.bool = true;
                        Solver.B.addChoicePoint(s);
                        Solver.add(((Set)S.get(0)).eq(Set.empty));
                        return;
                    case 1:
                        s.bool = true;
                        Set R = new Set();
                        Lst Ynew = new Lst(Lst.mkLst(Y.size_private()));
                        Lst L = new Lst(Lst.mkLst(l.size_private()));
                        Solver.add(((Set)S.get(0)).neq(Set.empty));
                        Solver.add(((Set)S.get(0)).less((Lvar)Ynew.get(0), R));

                        Solver.forall(Ynew.ext1(), S.ext1(), L, vs.sostAll(Y, Ynew).sostAll(l, L));

                        Lst Snew = new Lst(S.ext1().ins1(R));

                        Solver.forall(Y, Snew, l, vs);

                        Solver.storeInvariato = false;
                        return;
                    }
                }
            }
        }
        else { throw new InitLvarException(); }
    }


    /* forall non annidati */

    public void forall(Lvar Y, Set S, Constraint s2, Constraint s) throws Failure {
        if (!Y.isGround()) {
            if(!S.isEmptySL() && S.ultimoResto().iniz) {
                s.bool = true;
                Vector v = S.toVector();
                for(int i = 0; i < v.size(); i++) {
                    Lvar X = new Lvar();
                    Solver.add(X.eq(v.get(i)));
                    Solver.add(s2.sost(Y, X));
                }
                return;
            }
            else {
                switch(s.caseControl) {
                case 0:
                    s.bool = true;
                    Solver.B.addChoicePoint(s);
                    Solver.add(S.eq(Set.empty));
                    return;
                case 1:
                    s.bool = true;
                    Set R = new Set();
                    Lvar X = new Lvar();

                    Solver.add(S.neq(Set.empty));
                    Solver.add(S.less(X, R));
                    Solver.add(s2.sost(Y, X));
                    Solver.forall(Y, R, s2);

                    Solver.storeInvariato = false;
                    return;
                }
            }
        }
        else { throw new InitLvarException(); }
    }


    public void forall(Lvar Y, Int I, Constraint s2, Constraint s) throws Failure {
        if (!Y.isGround()) {
            if(I.isVoid()) {
                s.bool = true;
                return;
            }
            else {
                s.bool = true;
                Solver.add(Y.eq((Integer)I.sx));
                Constraint snew = s2.cloneConstraint();
                Solver.add(snew);
                for(int i = ((Integer)I.sx).intValue()+1; i < ((Integer)I.dx).intValue()+1; i++) {
                    Lvar X = new Lvar();
                    Constraint snew1 = s2.cloneConstraint();
                    Constraint snew2 = new Constraint(snew1.sost(Y, X));
                    Solver.add(X.eq(i));
                    Solver.add(snew2);
                }
                return;
            }
        }
        else { throw new InitLvarException(); }
    }


    public void forall(Lvar Y, Set S, Lst l, Constraint s2, Constraint s) throws Failure {
        if (!(Y.isGround()) && l.isAllUninitialized()) {
            if(!S.isEmpty() && S.ultimoResto().iniz) {
                s.bool = true;
                Vector v = S.toVector();
                for(int i = 0; i < v.size(); i++) {
                    Lvar X = new Lvar();
                    Lst L = new Lst(Lst.mkLst(l.size_private()));

                    Solver.add(X.eq(v.get(i)));
                    Solver.add(s2.sost(Y, X).sost(l, L));
                }
                return;
            }
            else {

                switch(s.caseControl) {
                case 0:
                    s.bool = true;
                    Solver.B.addChoicePoint(s);
                    Solver.add(S.eq(Set.empty));
                    return;
                case 1:
                    s.bool = true;
                    Set R = new Set();
                    Lvar X = new Lvar();
                    Lst L = new Lst(Lst.mkLst(l.size_private()));

                    Solver.add(S.neq(Set.empty));
                    Solver.add(S.less(X, R));
                    Solver.add(s2.sost(Y, X).sost(l, L));
                    Solver.forall(Y, R, l, s2);

                    Solver.storeInvariato = false;
                    return;
                }
            }
        }
        else { throw new InitLvarException(); }
    }


    public void forall(Lvar Y, Int I, Lst l, Constraint s2, Constraint s) throws Failure {
        if (!(Y.isGround()) && l.isAllUninitialized()) {
            if(I.isVoid()) {
                s.bool = true;
                return;
            }
            else {
                s.bool = true;
                Solver.add(Y.eq((Integer)I.sx));
                Constraint snew = s2.cloneConstraint();
                Solver.add(snew);
                for(int i = ((Integer)I.sx).intValue()+1; i < ((Integer)I.dx).intValue()+1; i++) {
                    Lvar X = new Lvar();
                    Lst L = new Lst(Lst.mkLst(l.size_private()));
                    Constraint snew1 = s2.cloneConstraint();
                    Constraint snew2 = new Constraint(snew1.sost(Y, X).sost(l, L));
                    Solver.add(X.eq(i));
                    Solver.add(snew2);
                }
                return;
            }
        }
        else { throw new InitLvarException(); }
    }


    public void forall(Lvar Y, Set S, ConstraintsConjunction vs, Constraint s) throws Failure {
        if (!Y.isGround()) {
            if(!S.isEmpty() && S.ultimoResto().iniz) {
                s.bool = true;
                Vector v = S.toVector();
                for(int i = 0; i < v.size(); i++) {
                    Lvar X = new Lvar();

                    Solver.add(X.eq(v.get(i)));
                    Solver.add(vs.sost(Y, X));
                }
                return;
            }
            else {
                switch(s.caseControl) {
                case 0:
                    s.bool = true;
                    Solver.B.addChoicePoint(s);
                    Solver.add(S.eq(Set.empty));
                    return;
                case 1:
                    s.bool = true;
                    Set R = new Set();
                    Lvar X = new Lvar();

                    Solver.add(S.neq(Set.empty));
                    Solver.add(S.less(X, R));
                    Solver.add(vs.sost(Y, X));
                    Solver.forall(Y, R, vs);
                    Solver.storeInvariato = false;
                    return;
                }
            }
        }
        else { throw new InitLvarException(); }
    }


    public void forall(Lvar Y, Int I , ConstraintsConjunction vs, Constraint s)throws Failure {
        if (!(Y.isGround())) {
            if(I.isVoid()) {
                s.bool = true;
                return;
            }
            else {
                s.bool = true;
                Solver.add(Y.eq((Integer)I.sx));
                ConstraintsConjunction snew = vs.cloneConstraintsConjunction();
                Solver.add(snew);
                for(int i = ((Integer)I.sx).intValue()+1; i < ((Integer)I.dx).intValue()+1; i++) {
                    Lvar X = new Lvar();
                    ConstraintsConjunction snew1 = vs.cloneConstraintsConjunction();
                    ConstraintsConjunction snew2 = snew1.sost(Y, X);
                    Solver.add(X.eq(i));
                    Solver.add(snew2);
                }
                return;
            }
        }
        else { throw new InitLvarException(); }
    }


    // 21/04/04 Implementazione.
    public void forall(Lvar Y, Set S, Lst l, ConstraintsConjunction vs, Constraint s) throws Failure {
        if (!(Y.isGround()) && l.isAllUninitialized()) {
            if(!S.isEmpty() && S.ultimoResto().iniz) {
                s.bool = true;
                Vector v = S.toVector();
                for(int i = 0; i < v.size(); i++) {
                    Lvar X = new Lvar();
                    Lst L = new Lst(Lst.mkLst(l.size_private()));

                    Solver.add(X.eq(v.get(i)));
                    Solver.add(vs.sost(Y, X).sost(l, L));
                }
                return;
            }
            else {
                switch(s.caseControl) {
                case 0:
                    s.bool = true;
                    Solver.B.addChoicePoint(s);
                    Solver.add(S.eq(Set.empty));
                    return;
                case 1:
                    s.bool = true;
                    Set R = new Set();
                    Lvar X = new Lvar();
                    Lst L = new Lst(Lst.mkLst(l.size_private()));

                    Solver.add(S.neq(Set.empty));
                    Solver.add(S.less(X, R));
                    Solver.add(vs.sost(Y, X).sost(l, L));
                    Solver.forall(Y, R, l, vs);

                    Solver.storeInvariato = false;
                    return;
                }
            }
        }
        else { throw new InitLvarException(); }
    }


    public void forall(Lvar Y, Int I, Lst l, ConstraintsConjunction vs, Constraint s) throws Failure {
        if (!(Y.isGround()) && l.isAllUninitialized()) {
            if(I.isVoid()) {
                s.bool = true;
                return;
            }
            else {
                s.bool = true;
                Solver.add(Y.eq((Integer)I.sx));
                ConstraintsConjunction snew = vs.cloneConstraintsConjunction();
                Solver.add(snew);
                for(int i = ((Integer)I.sx).intValue()+1; i < ((Integer)I.dx).intValue()+1; i++) {
                    Lvar X = new Lvar();
                    Lst L = new Lst(Lst.mkLst(l.size_private()));
                    ConstraintsConjunction snew1 = vs.cloneConstraintsConjunction();
                    ConstraintsConjunction snew2 = snew1.sost(Y, X).sost(l, L);
                    Solver.add(X.eq(i));
                    Solver.add(snew2);
                }
                return;
            }
        }
        else { throw new InitLvarException(); }
    }


    protected Constraint not_cond(Constraint s) {
	Constraint snew = new Constraint(s);
        if (s.cons == Enviroment.eqCode) snew.cons = Enviroment.neqCode;            // !equality = inequality
        else if (s.cons == Enviroment.neqCode) snew.cons = Enviroment.eqCode;       // !inequality = equality
        else if (s.cons == Enviroment.inCode) snew.cons = Enviroment.ninCode;       // !in = nin
        else if (s.cons == Enviroment.ninCode) snew.cons = Enviroment.inCode;       // !nin = in
        else if (s.cons == Enviroment.disjCode) snew.cons = Enviroment.ndisjCode;   // ...
        else if (s.cons == Enviroment.ndisjCode) snew.cons = Enviroment.disjCode;
        else if (s.cons == Enviroment.unionCode) snew.cons = Enviroment.nunionCode;
        else if (s.cons == Enviroment.nunionCode) snew.cons = Enviroment.unionCode;
        else if (s.cons == Enviroment.geCode) snew.cons = Enviroment.ltCode;
        else if (s.cons == Enviroment.gtCode) snew.cons = Enviroment.leCode;
        else if (s.cons == Enviroment.leCode) snew.cons = Enviroment.gtCode;
        else if (s.cons == Enviroment.ltCode) snew.cons = Enviroment.geCode;
        else System.out.print("Negazione non gestita");
	return snew;
    }


    // cond_if senza else
    protected void condif(Constraint condizione, Constraint allora, Constraint s) throws Failure {
	switch(s.caseControl) {
	case 0:
	    s.bool = true;
	    Solver.B.addChoicePoint(s);
	    Solver.add(condizione.cloneConstraint());
	    Solver.add(allora.cloneConstraint());
	    break;
	case 1:
	    s.bool = true;
	    Solver.add(not_cond(condizione.cloneConstraint()));
	    break;
	}
	return;
    }


    protected void condif(Constraint condizione, ConstraintsConjunction allora, Constraint s) throws Failure {
	switch(s.caseControl) {
	case 0:
	    s.bool = true;
	    Solver.B.addChoicePoint(s);

	    Solver.add(condizione.cloneConstraint());
	    Solver.add(allora.cloneConstraintsConjunction());
	    break;
	case 1:
	    s.bool = true;
	    Solver.add(not_cond(condizione.cloneConstraint()));
	    break;
	}
	return;
    }


    // cond_if con else
    protected void condif(Constraint condizione, ConstraintsConjunction allora, ConstraintsConjunction altrimenti, Constraint s) throws Failure {
	switch(s.caseControl) {
	case 0:
	    s.bool = true;
	    Solver.B.addChoicePoint(s);

	    Solver.add(condizione.cloneConstraint());
	    Solver.add(allora.cloneConstraintsConjunction());
	    break;
	case 1:
	    s.bool = true;
	    Solver.add(not_cond(condizione.cloneConstraint()));
	    Solver.add(altrimenti.cloneConstraintsConjunction());
	    break;
	}
	return;
    }


    protected void condif(Constraint condizione, Constraint allora, ConstraintsConjunction altrimenti, Constraint s) throws Failure {
	switch(s.caseControl) {
	case 0:
	    s.bool = true;
	    Solver.B.addChoicePoint(s);

	    Solver.add(condizione.cloneConstraint());
	    Solver.add(allora.cloneConstraint());
	    break;
	case 1:
	    s.bool = true;
	    Solver.add(not_cond(condizione.cloneConstraint()));
	    Solver.add(altrimenti.cloneConstraintsConjunction());
	    break;
	}
	return;
    }


    protected void condif(Constraint condizione, ConstraintsConjunction allora, Constraint altrimenti, Constraint s) throws Failure {
	switch(s.caseControl) {
	case 0:
	    s.bool = true;
	    Solver.B.addChoicePoint(s);

	    Solver.add(condizione.cloneConstraint());
	    Solver.add(allora.cloneConstraintsConjunction());
	    break;
	case 1:
	    s.bool = true;
	    Solver.add(not_cond(condizione.cloneConstraint()));
	    Solver.add(altrimenti.cloneConstraint());
	    break;
	}
	return;
    }


    protected void condif(Constraint condizione, Constraint allora, Constraint altrimenti, Constraint s) throws Failure {
	switch(s.caseControl) {
	case 0:
	    s.bool = true;
	    Solver.B.addChoicePoint(s);

	    Solver.add(condizione.cloneConstraint());
	    Solver.add(allora.cloneConstraint());
	    break;
	case 1:
	    s.bool = true;
	    Solver.add(not_cond(condizione.cloneConstraint()));
	    Solver.add(altrimenti.cloneConstraint());
	    break;
	}
	return;
    }


    public void showStoreElem(Constraint s) {
	if(s.cons == Enviroment.eqCode || s.cons == Enviroment.neqCode || s.cons == Enviroment.inCode ||
           s.cons == Enviroment.ninCode || s.cons == Enviroment.disjCode || s.cons == Enviroment.ndisjCode ||
           s.cons == Enviroment.geCode || s.cons == Enviroment.gtCode || s.cons == Enviroment.leCode ||
           s.cons == Enviroment.ltCode) {
	    if(s.arg1 instanceof Lvar)((Lvar)s.arg1).print();
	    else if(s.arg1 instanceof Lst)((Lst)s.arg1).print();
	    else if(s.arg1 instanceof Set)((Set)s.arg1).print();
	    else System.out.print(s.arg1);
	    System.out.print(".");
            if (s.cons == Enviroment.eqCode) System.out.print("eq");
            else if (s.cons == Enviroment.neqCode) System.out.print("neq");
            else if (s.cons == Enviroment.inCode) System.out.print("in");
            else if (s.cons == Enviroment.ninCode) System.out.print("nin");
            else if (s.cons == Enviroment.disjCode) System.out.print("disj");
            else if (s.cons == Enviroment.ndisjCode) System.out.print("ndisj");
            else if (s.cons == Enviroment.geCode) System.out.print("ge");
            else if (s.cons == Enviroment.gtCode) System.out.print("gt");
            else if (s.cons == Enviroment.leCode) System.out.print("le");
            else if (s.cons == Enviroment.ltCode) System.out.print("lt");
            System.out.print("(");
            if(s.arg2 instanceof Lvar)((Lvar)s.arg2).print();
            else if(s.arg2 instanceof Lst)((Lst)s.arg2).print();
            else if(s.arg2 instanceof Set)((Set)s.arg2).print();
            else System.out.print(s.arg2);
            System.out.print(")");
	}
	else if(s.cons == Enviroment.sumCode || s.cons == Enviroment.subCode || s.cons == Enviroment.mulCode ||
		s.cons == Enviroment.divCode || s.cons == Enviroment.modCode) {
            if(s.arg1 instanceof Lvar)((Lvar)s.arg1).print();
            else if(s.arg1 instanceof Lst)((Lst)s.arg1).print();
            else if(s.arg1 instanceof Set)((Set)s.arg1).print();
            else System.out.print(s.arg1);
            System.out.print(".");
            if (s.cons == Enviroment.sumCode) System.out.print("sum");
            else if (s.cons == Enviroment.subCode) System.out.print("sub");
            else if (s.cons == Enviroment.mulCode) System.out.print("mul");
            else if (s.cons == Enviroment.divCode) System.out.print("div");
            else if (s.cons == Enviroment.modCode) System.out.print("mod");
            System.out.print("(");
            if(s.arg2 instanceof Lvar)((Lvar)s.arg2).print();
            else if(s.arg2 instanceof Lst)((Lst)s.arg2).print();
            else if(s.arg2 instanceof Set)((Set)s.arg2).print();
            else System.out.print(s.arg2);
            System.out.print(", ");
            if(s.arg3 instanceof Lvar)((Lvar)s.arg3).print();
            else if(s.arg3 instanceof Lst)((Lst)s.arg3).print();
            else if(s.arg3 instanceof Set)((Set)s.arg3).print();
            else System.out.print(s.arg3);
            System.out.print(")");
	}
	else if(s.cons == Enviroment.unionCode || s.cons == Enviroment.nunionCode) {
            if(s.arg3 instanceof Lvar)((Lvar)s.arg3).print();
            else if(s.arg3 instanceof Lst)((Lst)s.arg3).print();
            else if(s.arg3 instanceof Set)((Set)s.arg3).print();
            else System.out.print(s.arg3);
            System.out.print(".");
            if (s.cons == Enviroment.unionCode) System.out.print("union");
            else if (s.cons == Enviroment.nunionCode) System.out.print("nunion");
            System.out.print("(");
            if(s.arg1 instanceof Lvar)((Lvar)s.arg1).print();
            else if(s.arg1 instanceof Lst)((Lst)s.arg1).print();
            else if(s.arg1 instanceof Set)((Set)s.arg1).print();
            else System.out.print(s.arg1);
            System.out.print(", ");
            if(s.arg2 instanceof Lvar)((Lvar)s.arg2).print();
            else if(s.arg2 instanceof Lst)((Lst)s.arg2).print();
            else if(s.arg2 instanceof Set)((Set)s.arg2).print();
            else System.out.print(s.arg2);
            System.out.print(")");
	}
	else if(s.cons == Enviroment.forallCode) {
            System.out.print("forall");
            System.out.print("(");
            ((Lvar)s.arg1).print();
            System.out.print(", ");
            ((Set)s.arg2).print();
            System.out.print(", ");
            if(s.arg3 != Lst.empty) {
		((Lst)s.arg3).print();
		System.out.print(", ");
            }
            if(s.arg4 instanceof Constraint) showStoreElem((Constraint)s.arg4);
            else if(s.arg4 instanceof ConstraintsConjunction) {
		for(int i = 0; i < ((ConstraintsConjunction)s.arg4).size(); i++) {
		    showStoreElem((Constraint)((ConstraintsConjunction)s.arg4).get(i));
		}
            }
            System.out.print(")");
	}
	else if(s.cons == Enviroment.cond_ifCode) {
            System.out.print("condif");
            System.out.print("(");
            System.out.print("if ");
            showStoreElem((Constraint)s.arg1);
            System.out.print(", ");
            System.out.print("then ");
            showStoreElem((Constraint)s.arg2);
            if(s.arg3 != null) {
		System.out.print(", ");
		System.out.print("else");
		showStoreElem((Constraint)s.arg3);
            }
            System.out.print(")");
	}
	else System.out.print("const"+s.cons);
	return;
    }


    protected void debug(String txt, Lvar toPrint, Constraint s) throws Failure {
	s.bool = true;

	System.out.print("\n" + txt + " ");
	toPrint.output();
	System.out.print("\n");
    }

}

